!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief describes a generic linked list template.
!>      Linked list are supposed to always use pointers to the nodes for
!>      basically everything: a pointer to the node is a list, an element of
!>      the list, an iterator between the elment of the list.
!>      An empty list is represented by an unassociated pointer.
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      1.2002 created
!>      4.2002 changed most of the methods, by making access to the list
!>             always through pointers (identifying pointer, list and iterators)
!>      6.2004 removed %initialized from list elements
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_linked_list_real
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_linked_list_real'

!API type
   PUBLIC :: cp_sll_real_type, cp_sll_real_p_type
!API common methods
   PUBLIC :: cp_create, cp_dealloc, cp_next
!API special get
   PUBLIC :: cp_get_first_el, cp_get_rest, cp_get_length, cp_get_element_at, cp_to_array
!API special set
   PUBLIC :: cp_set_element_at
!API structure manipulation
   PUBLIC :: cp_insert, cp_remove_first_el, cp_remove_el, cp_remove_all

   !API underlying routines
   PUBLIC :: cp_sll_real_create, cp_sll_real_dealloc, &
             cp_sll_real_next, &
             cp_sll_real_get_length, &
             cp_sll_real_get_el_at, cp_sll_real_set_el_at, &
             cp_sll_real_get_first_el, cp_sll_real_get_rest, &
             cp_sll_real_insert_el, cp_sll_real_insert_el_at, &
             cp_sll_real_rm_first_el, cp_sll_real_rm_el_at, &
             cp_sll_real_rm_all_el, &
             cp_sll_real_to_array

! creation of an object (from a pointer)
   INTERFACE cp_create
      MODULE PROCEDURE cp_sll_real_create
   END INTERFACE
! destruction of an object (from a pointer)
   INTERFACE cp_dealloc
      MODULE PROCEDURE cp_sll_real_dealloc
   END INTERFACE
! iterates to the next element
   INTERFACE cp_next
      MODULE PROCEDURE cp_sll_real_next
   END INTERFACE
! returns the first element
   INTERFACE cp_get_first_el
      MODULE PROCEDURE cp_sll_real_get_first_el
   END INTERFACE
! returns the rest of the list
   INTERFACE cp_get_rest
      MODULE PROCEDURE cp_sll_real_get_rest
   END INTERFACE
! returns the length of the list
   INTERFACE cp_get_length
      MODULE PROCEDURE cp_sll_real_get_length
   END INTERFACE
! returns the element at the given position
   INTERFACE cp_get_element_at
      MODULE PROCEDURE cp_sll_real_get_el_at
   END INTERFACE
! sets the element at the given position
   INTERFACE cp_set_element_at
      MODULE PROCEDURE cp_sll_real_set_el_at
   END INTERFACE
! inserts one element call cp_insert(list,element,...)
   INTERFACE cp_insert
      MODULE PROCEDURE cp_sll_real_insert_el
   END INTERFACE
   INTERFACE cp_insert_at
      MODULE PROCEDURE cp_sll_real_insert_el_at
   END INTERFACE
! removes an element
   INTERFACE cp_remove_el
      MODULE PROCEDURE cp_sll_real_rm_first_el, &
         cp_sll_real_rm_el_at
   END INTERFACE
! removes the first el
   INTERFACE cp_remove_first_el
      MODULE PROCEDURE cp_sll_real_rm_first_el
   END INTERFACE
! remove all the elments
   INTERFACE cp_remove_all
      MODULE PROCEDURE cp_sll_real_rm_all_el
   END INTERFACE
! transorms the list in array
   INTERFACE cp_to_array
      MODULE PROCEDURE cp_sll_real_to_array
   END INTERFACE

! **************************************************************************************************
!> \brief represent a single linked list that stores pointers to the elements
!> \param first_el the element that is stored in this node.
!> \param rest the rest of the list
!>
!> \param empty true if the list pointer is not associated, if it points to
!>             to a not it is always false (as there is at least the
!>             first_el in the list)
!> \param length the number of elements in the list
!> \note
!>      List are alway accessed through pointers, so every node of the
!>      linked list can be seen as a list, its first element
!>      a pointer to the position before itself, in a very natural way:
!>      all the insertions take place before the actual element, and
!>      you still can insert an element at the end.
!>      This way I could identify nodes, lists and pointers between the
!>      elements of the list.
!>      Indexing is 1 based.
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_real_type
      REAL(kind=dp) :: first_el
      TYPE(cp_sll_real_type), POINTER :: rest
   END TYPE cp_sll_real_type

! **************************************************************************************************
!> \brief pointer to a linked list (to make arrays of pointers)
!> \param list the pointer to the list
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_real_p_type
      TYPE(cp_sll_real_type), POINTER :: list
   END TYPE cp_sll_real_p_type

CONTAINS

! =========== creation / distruction ========

! **************************************************************************************************
!> \brief allocates and initializes a single linked list
!> \param sll the single linked list to initialize
!> \param first_el the first element of this list
!> \param rest the following elements (if not given: empty)
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_real_create(sll, first_el, rest)
      TYPE(cp_sll_real_type), POINTER                    :: sll
      REAL(kind=dp), INTENT(in), OPTIONAL                :: first_el
      TYPE(cp_sll_real_type), OPTIONAL, POINTER          :: rest

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_create', &
         routineP = moduleN//':'//routineN

      IF (.NOT. PRESENT(first_el)) THEN
         NULLIFY (sll)
         IF (PRESENT(rest)) sll => rest
      ELSE
         ALLOCATE (sll)
         sll%first_el = first_el
         NULLIFY (sll%rest)
         IF (PRESENT(rest)) sll%rest => rest
      END IF
   END SUBROUTINE cp_sll_real_create

! **************************************************************************************************
!> \brief deallocates the singly linked list starting at sll.
!>      Does not work if loops are present!
!> \param sll the list to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      does not deallocate the elments that are stored in the list
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_real_dealloc(sll)
      TYPE(cp_sll_real_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_dealloc', &
         routineP = moduleN//':'//routineN

      CALL cp_sll_real_rm_all_el(sll)
   END SUBROUTINE cp_sll_real_dealloc

! * low-level *

! **************************************************************************************************
!> \brief deallocates a node of a singly linked list (low level)
!> \param sll the node to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_real_dealloc_node(sll)
      TYPE(cp_sll_real_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_dealloc_node', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (sll)
   END SUBROUTINE cp_sll_real_dealloc_node

! ============= get/set ============

! **************************************************************************************************
!> \brief returns the first element stored in the list
!> \param sll the single linked list to get the element from
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_real_get_first_el(sll) RESULT(res)
      TYPE(cp_sll_real_type), POINTER                    :: sll
      REAL(kind=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_get_first_el', &
         routineP = moduleN//':'//routineN

      res = sll%first_el
   END FUNCTION cp_sll_real_get_first_el

! **************************************************************************************************
!> \brief returns the rest of the list
!> \param sll the single linked list to get the rest from
!> \param iter how many times the call to rest should be iterated,
!>            defaults to 1; -1 means till end of the list.
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      split the case iter=1 to make it more optimized?
! **************************************************************************************************
   FUNCTION cp_sll_real_get_rest(sll, iter) RESULT(res)
      TYPE(cp_sll_real_type), POINTER                    :: sll
      INTEGER, OPTIONAL                                  :: iter
      TYPE(cp_sll_real_type), POINTER                    :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_get_rest', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(sll)) THEN
         NULLIFY (res)
      ELSE
         IF (PRESENT(iter)) THEN
            res => sll
            DO i = 1, iter
               IF (ASSOCIATED(res%rest)) THEN
                  res => res%rest
               ELSE
                  CALL cp__b("input/cp_linked_list_real.F",293,"tried to go past end")
               END IF
            END DO
            IF (iter == -1) THEN
               DO
                  IF (.NOT. ASSOCIATED(res%rest)) EXIT
                  res => res%rest
               END DO
            END IF
         ELSE
            res => sll%rest ! make the common case fast...
         END IF
      END IF
   END FUNCTION cp_sll_real_get_rest

! **************************************************************************************************
!> \brief returns the length of the list
!> \param sll the list you want to know the length of
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(n))
! **************************************************************************************************
   FUNCTION cp_sll_real_get_length(sll) RESULT(res)
      TYPE(cp_sll_real_type), POINTER                    :: sll
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_get_length', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_real_type), POINTER                    :: iterator

      res = 0
      iterator => sll
      DO
         IF (ASSOCIATED(iterator)) THEN
            res = res+1
            iterator => iterator%rest
         ELSE
            EXIT
         END IF
      END DO
   END FUNCTION cp_sll_real_get_length

! **************************************************************************************************
!> \brief returns the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   FUNCTION cp_sll_real_get_el_at(sll, index) RESULT(res)
      TYPE(cp_sll_real_type), POINTER                    :: sll
      INTEGER, INTENT(in)                                :: index
      REAL(kind=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_get_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_real_type), POINTER                    :: pos

      IF (index == -1) THEN
         pos => cp_sll_real_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_real_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_real.F",365)

      res = pos%first_el
   END FUNCTION cp_sll_real_get_el_at

! **************************************************************************************************
!> \brief sets the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!>             -1 means at the end
!> \param value the new element
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_real_set_el_at(sll, index, value)
      TYPE(cp_sll_real_type), POINTER                    :: sll
      INTEGER, INTENT(in)                                :: index
      REAL(kind=dp), INTENT(in)                          :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_set_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_real_type), POINTER                    :: pos

      IF (index == -1) THEN
         pos => cp_sll_real_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_real_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_real.F",397)

      pos%first_el = value
   END SUBROUTINE cp_sll_real_set_el_at

! * iteration *

! **************************************************************************************************
!> \brief returns true if the actual element is valid (i.e. iterator ont at end)
!>      moves the iterator to the next element
!> \param iterator iterator that moves along the list
!> \param el_att the actual element (valid only if the function returns true)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_real_next(iterator, el_att) RESULT(res)
      TYPE(cp_sll_real_type), POINTER                    :: iterator
      REAL(kind=dp), INTENT(out), OPTIONAL               :: el_att
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_next', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(iterator)) THEN
         res = .TRUE.
         IF (PRESENT(el_att)) el_att = iterator%first_el
         iterator => iterator%rest
      ELSE
         res = .FALSE.
      END IF
   END FUNCTION cp_sll_real_next

! ============ structure modifications ============

! **************************************************************************************************
!> \brief insert an element at the beginning of the list
!> \param sll the single linked list point at the beginning of which
!>           you want to add the element
!> \param el the element to add
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_real_insert_el(sll, el)
      TYPE(cp_sll_real_type), POINTER                    :: sll
      REAL(kind=dp), INTENT(in)                          :: el

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_insert_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_real_type), POINTER                    :: newSlot

      NULLIFY (newSlot)

      CALL cp_sll_real_create(newSlot, first_el=el, &
                              rest=sll)
      sll => newSlot
   END SUBROUTINE cp_sll_real_insert_el

! **************************************************************************************************
!> \brief remove the first element of the linked list
!> \param sll the list whose first element has to be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_real_rm_first_el(sll)
      TYPE(cp_sll_real_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_rm_first_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_real_type), POINTER                    :: node_to_rm

      node_to_rm => sll

      IF (ASSOCIATED(sll)) THEN
         sll => sll%rest
         CALL cp_sll_real_dealloc_node(node_to_rm)
      ELSE
         CALL cp__b("input/cp_linked_list_real.F",483,"tried to remove first el of an empty list")
      END IF
   END SUBROUTINE cp_sll_real_rm_first_el

! **************************************************************************************************
!> \brief inserts the element at the given index
!> \param sll the list you get the element from
!> \param el the new element
!> \param index the position of the element (stating at 1).
!>             If it is -1, it means at end
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_real_insert_el_at(sll, el, index)
      TYPE(cp_sll_real_type), POINTER                    :: sll
      REAL(kind=dp), INTENT(in)                          :: el
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_insert_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_real_type), POINTER                    :: pos

      IF (index == 1) THEN
         CALL cp_sll_real_insert_el(sll, el)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_real_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_real_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_real.F",517)
         CALL cp_sll_real_insert_el(pos%rest, el)
      END IF
   END SUBROUTINE cp_sll_real_insert_el_at

! **************************************************************************************************
!> \brief removes the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_real_rm_el_at(sll, index)
      TYPE(cp_sll_real_type), POINTER                    :: sll
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_rm_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_real_type), POINTER                    :: pos

      IF (index == 1) THEN
         CALL cp_sll_real_rm_first_el(sll)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_real_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_real_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_real.F",549)
         CALL cp_sll_real_rm_first_el(pos%rest)
      END IF
   END SUBROUTINE cp_sll_real_rm_el_at

! **************************************************************************************************
!> \brief removes all the elements from the list
!> \param sll the list that should be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_real_rm_all_el(sll)
      TYPE(cp_sll_real_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_rm_all_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_real_type), POINTER                    :: actual_node, next_node

      actual_node => sll
      DO
         IF (.NOT. ASSOCIATED(actual_node)) EXIT
         next_node => actual_node%rest
         CALL cp_sll_real_dealloc_node(actual_node)
         actual_node => next_node
      END DO
      NULLIFY (sll)
   END SUBROUTINE cp_sll_real_rm_all_el

! **************************************************************************************************
!> \brief returns a newly allocated array with the same contents as
!>      the linked list
!> \param sll the list to trasform in array
!> \retval res ...
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_real_to_array(sll) RESULT(res)
      TYPE(cp_sll_real_type), POINTER                    :: sll
      REAL(kind=dp), DIMENSION(:), POINTER               :: res

      INTEGER                                            :: i, len
      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_real_to_array', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_real_type), POINTER                    :: iter

      len = cp_sll_real_get_length(sll)
      ALLOCATE (res(len))
      iter => sll
      DO i = 1, len
         res(i) = iter%first_el
         IF(.NOT.(cp_sll_real_next(iter) .OR. i == len))CALL cp__a("input/cp_linked_list_real.F",605)
      END DO
   END FUNCTION cp_sll_real_to_array

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(defines,nametype1,type1,type1in,type1out,type1arrayEl,arrayEl,array=,=,USE,write_el,lessQ,equalQ,private_routines)]
! ARGS:
!  = = "="
!  USE = "USE kinds, only: dp"
!  array= = "="
!  arrayEl = ""
!  common_dir = "../common"
!  defines =
!    "#define CP_SLL_R_LESS_Q(el1,el2) ( el1 < el2 )
!     #define ( el1 == el2 ) ( el1 == el2 )
!     "
!  equalQ = "CP_SLL_R_EQUAL_Q"
!  lessQ = "CP_SLL_R_LESS_Q"
!  nametype1 = "real"
!  private_routines = ""
!  template = "../common/cp_linked_list__nametype1_.template"
!  type1 = "REAL(kind=dp)"
!  type1arrayEl = "REAL(kind=dp)"
!  type1in = "REAL(kind=dp), intent(in)"
!  type1out = "REAL(kind=dp), intent(out)"
!  write_el = ""

END MODULE cp_linked_list_real

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief describes a generic linked list template.
!>      Linked list are supposed to always use pointers to the nodes for
!>      basically everything: a pointer to the node is a list, an element of
!>      the list, an iterator between the elment of the list.
!>      An empty list is represented by an unassociated pointer.
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      1.2002 created
!>      4.2002 changed most of the methods, by making access to the list
!>             always through pointers (identifying pointer, list and iterators)
!>      6.2004 removed %initialized from list elements
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_linked_list_logical

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_linked_list_logical'

!API type
   PUBLIC :: cp_sll_logical_type, cp_sll_logical_p_type
!API common methods
   PUBLIC :: cp_create, cp_dealloc, cp_next
!API special get
   PUBLIC :: cp_get_first_el, cp_get_rest, cp_get_length, cp_get_element_at, cp_to_array
!API special set
   PUBLIC :: cp_set_element_at
!API structure manipulation
   PUBLIC :: cp_insert, cp_remove_first_el, cp_remove_el, cp_remove_all

   !API underlying routines
   PUBLIC :: cp_sll_logical_create, cp_sll_logical_dealloc, &
             cp_sll_logical_next, &
             cp_sll_logical_get_length, &
             cp_sll_logical_get_el_at, cp_sll_logical_set_el_at, &
             cp_sll_logical_get_first_el, cp_sll_logical_get_rest, &
             cp_sll_logical_insert_el, cp_sll_logical_insert_el_at, &
             cp_sll_logical_rm_first_el, cp_sll_logical_rm_el_at, &
             cp_sll_logical_rm_all_el, &
             cp_sll_logical_to_array

! creation of an object (from a pointer)
   INTERFACE cp_create
      MODULE PROCEDURE cp_sll_logical_create
   END INTERFACE
! destruction of an object (from a pointer)
   INTERFACE cp_dealloc
      MODULE PROCEDURE cp_sll_logical_dealloc
   END INTERFACE
! iterates to the next element
   INTERFACE cp_next
      MODULE PROCEDURE cp_sll_logical_next
   END INTERFACE
! returns the first element
   INTERFACE cp_get_first_el
      MODULE PROCEDURE cp_sll_logical_get_first_el
   END INTERFACE
! returns the rest of the list
   INTERFACE cp_get_rest
      MODULE PROCEDURE cp_sll_logical_get_rest
   END INTERFACE
! returns the length of the list
   INTERFACE cp_get_length
      MODULE PROCEDURE cp_sll_logical_get_length
   END INTERFACE
! returns the element at the given position
   INTERFACE cp_get_element_at
      MODULE PROCEDURE cp_sll_logical_get_el_at
   END INTERFACE
! sets the element at the given position
   INTERFACE cp_set_element_at
      MODULE PROCEDURE cp_sll_logical_set_el_at
   END INTERFACE
! inserts one element call cp_insert(list,element,...)
   INTERFACE cp_insert
      MODULE PROCEDURE cp_sll_logical_insert_el
   END INTERFACE
   INTERFACE cp_insert_at
      MODULE PROCEDURE cp_sll_logical_insert_el_at
   END INTERFACE
! removes an element
   INTERFACE cp_remove_el
      MODULE PROCEDURE cp_sll_logical_rm_first_el, &
         cp_sll_logical_rm_el_at
   END INTERFACE
! removes the first el
   INTERFACE cp_remove_first_el
      MODULE PROCEDURE cp_sll_logical_rm_first_el
   END INTERFACE
! remove all the elments
   INTERFACE cp_remove_all
      MODULE PROCEDURE cp_sll_logical_rm_all_el
   END INTERFACE
! transorms the list in array
   INTERFACE cp_to_array
      MODULE PROCEDURE cp_sll_logical_to_array
   END INTERFACE

! **************************************************************************************************
!> \brief represent a single linked list that stores pointers to the elements
!> \param first_el the element that is stored in this node.
!> \param rest the rest of the list
!>
!> \param empty true if the list pointer is not associated, if it points to
!>             to a not it is always false (as there is at least the
!>             first_el in the list)
!> \param length the number of elements in the list
!> \note
!>      List are alway accessed through pointers, so every node of the
!>      linked list can be seen as a list, its first element
!>      a pointer to the position before itself, in a very natural way:
!>      all the insertions take place before the actual element, and
!>      you still can insert an element at the end.
!>      This way I could identify nodes, lists and pointers between the
!>      elements of the list.
!>      Indexing is 1 based.
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_logical_type
      LOGICAL :: first_el
      TYPE(cp_sll_logical_type), POINTER :: rest
   END TYPE cp_sll_logical_type

! **************************************************************************************************
!> \brief pointer to a linked list (to make arrays of pointers)
!> \param list the pointer to the list
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_logical_p_type
      TYPE(cp_sll_logical_type), POINTER :: list
   END TYPE cp_sll_logical_p_type

CONTAINS

! =========== creation / distruction ========

! **************************************************************************************************
!> \brief allocates and initializes a single linked list
!> \param sll the single linked list to initialize
!> \param first_el the first element of this list
!> \param rest the following elements (if not given: empty)
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_logical_create(sll, first_el, rest)
      TYPE(cp_sll_logical_type), POINTER                 :: sll
      LOGICAL, INTENT(in), OPTIONAL                      :: first_el
      TYPE(cp_sll_logical_type), OPTIONAL, POINTER       :: rest

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_create', &
         routineP = moduleN//':'//routineN

      IF (.NOT. PRESENT(first_el)) THEN
         NULLIFY (sll)
         IF (PRESENT(rest)) sll => rest
      ELSE
         ALLOCATE (sll)
         sll%first_el = first_el
         NULLIFY (sll%rest)
         IF (PRESENT(rest)) sll%rest => rest
      END IF
   END SUBROUTINE cp_sll_logical_create

! **************************************************************************************************
!> \brief deallocates the singly linked list starting at sll.
!>      Does not work if loops are present!
!> \param sll the list to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      does not deallocate the elments that are stored in the list
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_logical_dealloc(sll)
      TYPE(cp_sll_logical_type), POINTER                 :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_dealloc', &
         routineP = moduleN//':'//routineN

      CALL cp_sll_logical_rm_all_el(sll)
   END SUBROUTINE cp_sll_logical_dealloc

! * low-level *

! **************************************************************************************************
!> \brief deallocates a node of a singly linked list (low level)
!> \param sll the node to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_logical_dealloc_node(sll)
      TYPE(cp_sll_logical_type), POINTER                 :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_dealloc_node', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (sll)
   END SUBROUTINE cp_sll_logical_dealloc_node

! ============= get/set ============

! **************************************************************************************************
!> \brief returns the first element stored in the list
!> \param sll the single linked list to get the element from
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_logical_get_first_el(sll) RESULT(res)
      TYPE(cp_sll_logical_type), POINTER                 :: sll
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_get_first_el', &
         routineP = moduleN//':'//routineN

      res = sll%first_el
   END FUNCTION cp_sll_logical_get_first_el

! **************************************************************************************************
!> \brief returns the rest of the list
!> \param sll the single linked list to get the rest from
!> \param iter how many times the call to rest should be iterated,
!>            defaults to 1; -1 means till end of the list.
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      split the case iter=1 to make it more optimized?
! **************************************************************************************************
   FUNCTION cp_sll_logical_get_rest(sll, iter) RESULT(res)
      TYPE(cp_sll_logical_type), POINTER                 :: sll
      INTEGER, OPTIONAL                                  :: iter
      TYPE(cp_sll_logical_type), POINTER                 :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_get_rest', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(sll)) THEN
         NULLIFY (res)
      ELSE
         IF (PRESENT(iter)) THEN
            res => sll
            DO i = 1, iter
               IF (ASSOCIATED(res%rest)) THEN
                  res => res%rest
               ELSE
                  CALL cp__b("input/cp_linked_list_logical.F",292,"tried to go past end")
               END IF
            END DO
            IF (iter == -1) THEN
               DO
                  IF (.NOT. ASSOCIATED(res%rest)) EXIT
                  res => res%rest
               END DO
            END IF
         ELSE
            res => sll%rest ! make the common case fast...
         END IF
      END IF
   END FUNCTION cp_sll_logical_get_rest

! **************************************************************************************************
!> \brief returns the length of the list
!> \param sll the list you want to know the length of
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(n))
! **************************************************************************************************
   FUNCTION cp_sll_logical_get_length(sll) RESULT(res)
      TYPE(cp_sll_logical_type), POINTER                 :: sll
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_get_length', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_logical_type), POINTER                 :: iterator

      res = 0
      iterator => sll
      DO
         IF (ASSOCIATED(iterator)) THEN
            res = res+1
            iterator => iterator%rest
         ELSE
            EXIT
         END IF
      END DO
   END FUNCTION cp_sll_logical_get_length

! **************************************************************************************************
!> \brief returns the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   FUNCTION cp_sll_logical_get_el_at(sll, index) RESULT(res)
      TYPE(cp_sll_logical_type), POINTER                 :: sll
      INTEGER, INTENT(in)                                :: index
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_get_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_logical_type), POINTER                 :: pos

      IF (index == -1) THEN
         pos => cp_sll_logical_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_logical_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_logical.F",364)

      res = pos%first_el
   END FUNCTION cp_sll_logical_get_el_at

! **************************************************************************************************
!> \brief sets the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!>             -1 means at the end
!> \param value the new element
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_logical_set_el_at(sll, index, value)
      TYPE(cp_sll_logical_type), POINTER                 :: sll
      INTEGER, INTENT(in)                                :: index
      LOGICAL, INTENT(in)                                :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_set_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_logical_type), POINTER                 :: pos

      IF (index == -1) THEN
         pos => cp_sll_logical_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_logical_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_logical.F",396)

      pos%first_el = value
   END SUBROUTINE cp_sll_logical_set_el_at

! * iteration *

! **************************************************************************************************
!> \brief returns true if the actual element is valid (i.e. iterator ont at end)
!>      moves the iterator to the next element
!> \param iterator iterator that moves along the list
!> \param el_att the actual element (valid only if the function returns true)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_logical_next(iterator, el_att) RESULT(res)
      TYPE(cp_sll_logical_type), POINTER                 :: iterator
      LOGICAL, INTENT(out), OPTIONAL                     :: el_att
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_next', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(iterator)) THEN
         res = .TRUE.
         IF (PRESENT(el_att)) el_att = iterator%first_el
         iterator => iterator%rest
      ELSE
         res = .FALSE.
      END IF
   END FUNCTION cp_sll_logical_next

! ============ structure modifications ============

! **************************************************************************************************
!> \brief insert an element at the beginning of the list
!> \param sll the single linked list point at the beginning of which
!>           you want to add the element
!> \param el the element to add
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_logical_insert_el(sll, el)
      TYPE(cp_sll_logical_type), POINTER                 :: sll
      LOGICAL, INTENT(in)                                :: el

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_insert_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_logical_type), POINTER                 :: newSlot

      NULLIFY (newSlot)

      CALL cp_sll_logical_create(newSlot, first_el=el, &
                                 rest=sll)
      sll => newSlot
   END SUBROUTINE cp_sll_logical_insert_el

! **************************************************************************************************
!> \brief remove the first element of the linked list
!> \param sll the list whose first element has to be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_logical_rm_first_el(sll)
      TYPE(cp_sll_logical_type), POINTER                 :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_rm_first_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_logical_type), POINTER                 :: node_to_rm

      node_to_rm => sll

      IF (ASSOCIATED(sll)) THEN
         sll => sll%rest
         CALL cp_sll_logical_dealloc_node(node_to_rm)
      ELSE
         CALL cp__b("input/cp_linked_list_logical.F",482,"tried to remove first el of an empty list")
      END IF
   END SUBROUTINE cp_sll_logical_rm_first_el

! **************************************************************************************************
!> \brief inserts the element at the given index
!> \param sll the list you get the element from
!> \param el the new element
!> \param index the position of the element (stating at 1).
!>             If it is -1, it means at end
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_logical_insert_el_at(sll, el, index)
      TYPE(cp_sll_logical_type), POINTER                 :: sll
      LOGICAL, INTENT(in)                                :: el
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_insert_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_logical_type), POINTER                 :: pos

      IF (index == 1) THEN
         CALL cp_sll_logical_insert_el(sll, el)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_logical_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_logical_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_logical.F",516)
         CALL cp_sll_logical_insert_el(pos%rest, el)
      END IF
   END SUBROUTINE cp_sll_logical_insert_el_at

! **************************************************************************************************
!> \brief removes the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_logical_rm_el_at(sll, index)
      TYPE(cp_sll_logical_type), POINTER                 :: sll
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_rm_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_logical_type), POINTER                 :: pos

      IF (index == 1) THEN
         CALL cp_sll_logical_rm_first_el(sll)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_logical_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_logical_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_logical.F",548)
         CALL cp_sll_logical_rm_first_el(pos%rest)
      END IF
   END SUBROUTINE cp_sll_logical_rm_el_at

! **************************************************************************************************
!> \brief removes all the elements from the list
!> \param sll the list that should be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_logical_rm_all_el(sll)
      TYPE(cp_sll_logical_type), POINTER                 :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_rm_all_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_logical_type), POINTER                 :: actual_node, next_node

      actual_node => sll
      DO
         IF (.NOT. ASSOCIATED(actual_node)) EXIT
         next_node => actual_node%rest
         CALL cp_sll_logical_dealloc_node(actual_node)
         actual_node => next_node
      END DO
      NULLIFY (sll)
   END SUBROUTINE cp_sll_logical_rm_all_el

! **************************************************************************************************
!> \brief returns a newly allocated array with the same contents as
!>      the linked list
!> \param sll the list to trasform in array
!> \retval res ...
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_logical_to_array(sll) RESULT(res)
      TYPE(cp_sll_logical_type), POINTER                 :: sll
      LOGICAL, DIMENSION(:), POINTER                     :: res

      INTEGER                                            :: i, len
      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_logical_to_array', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_logical_type), POINTER                 :: iter

      len = cp_sll_logical_get_length(sll)
      ALLOCATE (res(len))
      iter => sll
      DO i = 1, len
         res(i) = iter%first_el
         IF(.NOT.(cp_sll_logical_next(iter) .OR. i == len))CALL cp__a("input/cp_linked_list_logical.F",604)
      END DO
   END FUNCTION cp_sll_logical_to_array

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(defines,nametype1,type1,type1in,type1out,type1arrayEl,arrayEl,array=,=,USE,write_el,lessQ,equalQ,private_routines)]
! ARGS:
!  = = "="
!  USE = " "
!  array= = "="
!  arrayEl = ""
!  common_dir = "../common"
!  defines =
!    "#define CP_SLL_L_LESS_Q(el1,el2) ( .not.el1.and.el2 )
!     #define ( el1.EQV.el2 ) ( el1.EQV.el2 )
!     "
!  equalQ = "CP_SLL_L_EQUAL_Q"
!  lessQ = "CP_SLL_L_LESS_Q"
!  nametype1 = "logical"
!  private_routines = ""
!  template = "../common/cp_linked_list__nametype1_.template"
!  type1 = "logical"
!  type1arrayEl = "logical"
!  type1in = "logical, intent(in)"
!  type1out = "logical, intent(out)"
!  write_el = ""

END MODULE cp_linked_list_logical

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief describes a generic linked list template.
!>      Linked list are supposed to always use pointers to the nodes for
!>      basically everything: a pointer to the node is a list, an element of
!>      the list, an iterator between the elment of the list.
!>      An empty list is represented by an unassociated pointer.
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      1.2002 created
!>      4.2002 changed most of the methods, by making access to the list
!>             always through pointers (identifying pointer, list and iterators)
!>      6.2004 removed %initialized from list elements
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_linked_list_char
   USE kinds,                           ONLY: default_string_length
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_linked_list_char'

!API type
   PUBLIC :: cp_sll_char_type, cp_sll_char_p_type
!API common methods
   PUBLIC :: cp_create, cp_dealloc, cp_next
!API special get
   PUBLIC :: cp_get_first_el, cp_get_rest, cp_get_length, cp_get_element_at, cp_to_array
!API special set
   PUBLIC :: cp_set_element_at
!API structure manipulation
   PUBLIC :: cp_insert, cp_remove_first_el, cp_remove_el, cp_remove_all

   !API underlying routines
   PUBLIC :: cp_sll_char_create, cp_sll_char_dealloc, &
             cp_sll_char_next, &
             cp_sll_char_get_length, &
             cp_sll_char_get_el_at, cp_sll_char_set_el_at, &
             cp_sll_char_get_first_el, cp_sll_char_get_rest, &
             cp_sll_char_insert_el, cp_sll_char_insert_el_at, &
             cp_sll_char_rm_first_el, cp_sll_char_rm_el_at, &
             cp_sll_char_rm_all_el, &
             cp_sll_char_to_array

! creation of an object (from a pointer)
   INTERFACE cp_create
      MODULE PROCEDURE cp_sll_char_create
   END INTERFACE
! destruction of an object (from a pointer)
   INTERFACE cp_dealloc
      MODULE PROCEDURE cp_sll_char_dealloc
   END INTERFACE
! iterates to the next element
   INTERFACE cp_next
      MODULE PROCEDURE cp_sll_char_next
   END INTERFACE
! returns the first element
   INTERFACE cp_get_first_el
      MODULE PROCEDURE cp_sll_char_get_first_el
   END INTERFACE
! returns the rest of the list
   INTERFACE cp_get_rest
      MODULE PROCEDURE cp_sll_char_get_rest
   END INTERFACE
! returns the length of the list
   INTERFACE cp_get_length
      MODULE PROCEDURE cp_sll_char_get_length
   END INTERFACE
! returns the element at the given position
   INTERFACE cp_get_element_at
      MODULE PROCEDURE cp_sll_char_get_el_at
   END INTERFACE
! sets the element at the given position
   INTERFACE cp_set_element_at
      MODULE PROCEDURE cp_sll_char_set_el_at
   END INTERFACE
! inserts one element call cp_insert(list,element,...)
   INTERFACE cp_insert
      MODULE PROCEDURE cp_sll_char_insert_el
   END INTERFACE
   INTERFACE cp_insert_at
      MODULE PROCEDURE cp_sll_char_insert_el_at
   END INTERFACE
! removes an element
   INTERFACE cp_remove_el
      MODULE PROCEDURE cp_sll_char_rm_first_el, &
         cp_sll_char_rm_el_at
   END INTERFACE
! removes the first el
   INTERFACE cp_remove_first_el
      MODULE PROCEDURE cp_sll_char_rm_first_el
   END INTERFACE
! remove all the elments
   INTERFACE cp_remove_all
      MODULE PROCEDURE cp_sll_char_rm_all_el
   END INTERFACE
! transorms the list in array
   INTERFACE cp_to_array
      MODULE PROCEDURE cp_sll_char_to_array
   END INTERFACE

! **************************************************************************************************
!> \brief represent a single linked list that stores pointers to the elements
!> \param first_el the element that is stored in this node.
!> \param rest the rest of the list
!>
!> \param empty true if the list pointer is not associated, if it points to
!>             to a not it is always false (as there is at least the
!>             first_el in the list)
!> \param length the number of elements in the list
!> \note
!>      List are alway accessed through pointers, so every node of the
!>      linked list can be seen as a list, its first element
!>      a pointer to the position before itself, in a very natural way:
!>      all the insertions take place before the actual element, and
!>      you still can insert an element at the end.
!>      This way I could identify nodes, lists and pointers between the
!>      elements of the list.
!>      Indexing is 1 based.
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_char_type
      CHARACTER(len=default_string_length) :: first_el
      TYPE(cp_sll_char_type), POINTER :: rest
   END TYPE cp_sll_char_type

! **************************************************************************************************
!> \brief pointer to a linked list (to make arrays of pointers)
!> \param list the pointer to the list
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_char_p_type
      TYPE(cp_sll_char_type), POINTER :: list
   END TYPE cp_sll_char_p_type

CONTAINS

! =========== creation / distruction ========

! **************************************************************************************************
!> \brief allocates and initializes a single linked list
!> \param sll the single linked list to initialize
!> \param first_el the first element of this list
!> \param rest the following elements (if not given: empty)
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_char_create(sll, first_el, rest)
      TYPE(cp_sll_char_type), POINTER                    :: sll
      CHARACTER(len=default_string_length), INTENT(in), &
         OPTIONAL                                        :: first_el
      TYPE(cp_sll_char_type), OPTIONAL, POINTER          :: rest

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_create', &
         routineP = moduleN//':'//routineN

      IF (.NOT. PRESENT(first_el)) THEN
         NULLIFY (sll)
         IF (PRESENT(rest)) sll => rest
      ELSE
         ALLOCATE (sll)
         sll%first_el = first_el
         NULLIFY (sll%rest)
         IF (PRESENT(rest)) sll%rest => rest
      END IF
   END SUBROUTINE cp_sll_char_create

! **************************************************************************************************
!> \brief deallocates the singly linked list starting at sll.
!>      Does not work if loops are present!
!> \param sll the list to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      does not deallocate the elments that are stored in the list
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_char_dealloc(sll)
      TYPE(cp_sll_char_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_dealloc', &
         routineP = moduleN//':'//routineN

      CALL cp_sll_char_rm_all_el(sll)
   END SUBROUTINE cp_sll_char_dealloc

! * low-level *

! **************************************************************************************************
!> \brief deallocates a node of a singly linked list (low level)
!> \param sll the node to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_char_dealloc_node(sll)
      TYPE(cp_sll_char_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_dealloc_node', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (sll)
   END SUBROUTINE cp_sll_char_dealloc_node

! ============= get/set ============

! **************************************************************************************************
!> \brief returns the first element stored in the list
!> \param sll the single linked list to get the element from
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_char_get_first_el(sll) RESULT(res)
      TYPE(cp_sll_char_type), POINTER                    :: sll
      CHARACTER(len=default_string_length)               :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_get_first_el', &
         routineP = moduleN//':'//routineN

      res = sll%first_el
   END FUNCTION cp_sll_char_get_first_el

! **************************************************************************************************
!> \brief returns the rest of the list
!> \param sll the single linked list to get the rest from
!> \param iter how many times the call to rest should be iterated,
!>            defaults to 1; -1 means till end of the list.
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      split the case iter=1 to make it more optimized?
! **************************************************************************************************
   FUNCTION cp_sll_char_get_rest(sll, iter) RESULT(res)
      TYPE(cp_sll_char_type), POINTER                    :: sll
      INTEGER, OPTIONAL                                  :: iter
      TYPE(cp_sll_char_type), POINTER                    :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_get_rest', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(sll)) THEN
         NULLIFY (res)
      ELSE
         IF (PRESENT(iter)) THEN
            res => sll
            DO i = 1, iter
               IF (ASSOCIATED(res%rest)) THEN
                  res => res%rest
               ELSE
                  CALL cp__b("input/cp_linked_list_char.F",294,"tried to go past end")
               END IF
            END DO
            IF (iter == -1) THEN
               DO
                  IF (.NOT. ASSOCIATED(res%rest)) EXIT
                  res => res%rest
               END DO
            END IF
         ELSE
            res => sll%rest ! make the common case fast...
         END IF
      END IF
   END FUNCTION cp_sll_char_get_rest

! **************************************************************************************************
!> \brief returns the length of the list
!> \param sll the list you want to know the length of
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(n))
! **************************************************************************************************
   FUNCTION cp_sll_char_get_length(sll) RESULT(res)
      TYPE(cp_sll_char_type), POINTER                    :: sll
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_get_length', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_char_type), POINTER                    :: iterator

      res = 0
      iterator => sll
      DO
         IF (ASSOCIATED(iterator)) THEN
            res = res+1
            iterator => iterator%rest
         ELSE
            EXIT
         END IF
      END DO
   END FUNCTION cp_sll_char_get_length

! **************************************************************************************************
!> \brief returns the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   FUNCTION cp_sll_char_get_el_at(sll, index) RESULT(res)
      TYPE(cp_sll_char_type), POINTER                    :: sll
      INTEGER, INTENT(in)                                :: index
      CHARACTER(len=default_string_length)               :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_get_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_char_type), POINTER                    :: pos

      IF (index == -1) THEN
         pos => cp_sll_char_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_char_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_char.F",366)

      res = pos%first_el
   END FUNCTION cp_sll_char_get_el_at

! **************************************************************************************************
!> \brief sets the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!>             -1 means at the end
!> \param value the new element
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_char_set_el_at(sll, index, value)
      TYPE(cp_sll_char_type), POINTER                    :: sll
      INTEGER, INTENT(in)                                :: index
      CHARACTER(len=default_string_length), INTENT(in)   :: value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_set_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_char_type), POINTER                    :: pos

      IF (index == -1) THEN
         pos => cp_sll_char_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_char_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_char.F",398)

      pos%first_el = value
   END SUBROUTINE cp_sll_char_set_el_at

! * iteration *

! **************************************************************************************************
!> \brief returns true if the actual element is valid (i.e. iterator ont at end)
!>      moves the iterator to the next element
!> \param iterator iterator that moves along the list
!> \param el_att the actual element (valid only if the function returns true)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_char_next(iterator, el_att) RESULT(res)
      TYPE(cp_sll_char_type), POINTER                    :: iterator
      CHARACTER(len=default_string_length), &
         INTENT(out), OPTIONAL                           :: el_att
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_next', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(iterator)) THEN
         res = .TRUE.
         IF (PRESENT(el_att)) el_att = iterator%first_el
         iterator => iterator%rest
      ELSE
         res = .FALSE.
      END IF
   END FUNCTION cp_sll_char_next

! ============ structure modifications ============

! **************************************************************************************************
!> \brief insert an element at the beginning of the list
!> \param sll the single linked list point at the beginning of which
!>           you want to add the element
!> \param el the element to add
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_char_insert_el(sll, el)
      TYPE(cp_sll_char_type), POINTER                    :: sll
      CHARACTER(len=default_string_length), INTENT(in)   :: el

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_insert_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_char_type), POINTER                    :: newSlot

      NULLIFY (newSlot)

      CALL cp_sll_char_create(newSlot, first_el=el, &
                              rest=sll)
      sll => newSlot
   END SUBROUTINE cp_sll_char_insert_el

! **************************************************************************************************
!> \brief remove the first element of the linked list
!> \param sll the list whose first element has to be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_char_rm_first_el(sll)
      TYPE(cp_sll_char_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_rm_first_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_char_type), POINTER                    :: node_to_rm

      node_to_rm => sll

      IF (ASSOCIATED(sll)) THEN
         sll => sll%rest
         CALL cp_sll_char_dealloc_node(node_to_rm)
      ELSE
         CALL cp__b("input/cp_linked_list_char.F",485,"tried to remove first el of an empty list")
      END IF
   END SUBROUTINE cp_sll_char_rm_first_el

! **************************************************************************************************
!> \brief inserts the element at the given index
!> \param sll the list you get the element from
!> \param el the new element
!> \param index the position of the element (stating at 1).
!>             If it is -1, it means at end
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_char_insert_el_at(sll, el, index)
      TYPE(cp_sll_char_type), POINTER                    :: sll
      CHARACTER(len=default_string_length), INTENT(in)   :: el
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_insert_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_char_type), POINTER                    :: pos

      IF (index == 1) THEN
         CALL cp_sll_char_insert_el(sll, el)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_char_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_char_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_char.F",519)
         CALL cp_sll_char_insert_el(pos%rest, el)
      END IF
   END SUBROUTINE cp_sll_char_insert_el_at

! **************************************************************************************************
!> \brief removes the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_char_rm_el_at(sll, index)
      TYPE(cp_sll_char_type), POINTER                    :: sll
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_rm_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_char_type), POINTER                    :: pos

      IF (index == 1) THEN
         CALL cp_sll_char_rm_first_el(sll)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_char_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_char_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_char.F",551)
         CALL cp_sll_char_rm_first_el(pos%rest)
      END IF
   END SUBROUTINE cp_sll_char_rm_el_at

! **************************************************************************************************
!> \brief removes all the elements from the list
!> \param sll the list that should be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_char_rm_all_el(sll)
      TYPE(cp_sll_char_type), POINTER                    :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_rm_all_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_char_type), POINTER                    :: actual_node, next_node

      actual_node => sll
      DO
         IF (.NOT. ASSOCIATED(actual_node)) EXIT
         next_node => actual_node%rest
         CALL cp_sll_char_dealloc_node(actual_node)
         actual_node => next_node
      END DO
      NULLIFY (sll)
   END SUBROUTINE cp_sll_char_rm_all_el

! **************************************************************************************************
!> \brief returns a newly allocated array with the same contents as
!>      the linked list
!> \param sll the list to trasform in array
!> \retval res ...
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_char_to_array(sll) RESULT(res)
      TYPE(cp_sll_char_type), POINTER                    :: sll
      CHARACTER(len=default_string_length), &
         DIMENSION(:), POINTER                           :: res

      INTEGER                                            :: i, len
      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_char_to_array', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_char_type), POINTER                    :: iter

      len = cp_sll_char_get_length(sll)
      ALLOCATE (res(len))
      iter => sll
      DO i = 1, len
         res(i) = iter%first_el
         IF(.NOT.(cp_sll_char_next(iter) .OR. i == len))CALL cp__a("input/cp_linked_list_char.F",608)
      END DO
   END FUNCTION cp_sll_char_to_array

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(defines,nametype1,type1,type1in,type1out,type1arrayEl,arrayEl,array=,=,USE,write_el,lessQ,equalQ,private_routines)]
! ARGS:
!  = = "="
!  USE = "USE kinds, only: default_string_length"
!  array= = "="
!  arrayEl = ""
!  common_dir = "../common"
!  defines =
!    "#define CP_SLL_C_LESS_Q(el1,el2) ( el1 < el2 )
!     #define ( el1 == el2 ) ( el1 == el2 )
!     "
!  equalQ = "CP_SLL_C_EQUAL_Q"
!  lessQ = "CP_SLL_C_LESS_Q"
!  nametype1 = "char"
!  private_routines = ""
!  template = "../common/cp_linked_list__nametype1_.template"
!  type1 = "character(len=default_string_length)"
!  type1arrayEl = "character(len=default_string_length)"
!  type1in = "character(len=default_string_length), intent(in)"
!  type1out = "character(len=default_string_length), intent(out)"
!  write_el = ""

END MODULE cp_linked_list_char

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief describes a generic linked list template.
!>      Linked list are supposed to always use pointers to the nodes for
!>      basically everything: a pointer to the node is a list, an element of
!>      the list, an iterator between the elment of the list.
!>      An empty list is represented by an unassociated pointer.
!> \note
!>     ____              _ _     __  __           _ _  __         _____ _     _       _____ _ _      _
!>    |  _ \  ___  _ __ ( ) |_  |  \/  | ___   __| (_)/ _|_   _  |_   _| |__ (_)___  |  ___(_) | ___| |
!>    | | | |/ _ \| '_ \|/| __| | |\/| |/ _ \ / _` | | |_| | | |   | | | '_ \| / __| | |_  | | |/ _ \ |
!>    | |_| | (_) | | | | | |_  | |  | | (_) | (_| | |  _| |_| |   | | | | | | \__ \ |  _| | | |  __/_|
!>    |____/ \___/|_| |_|  \__| |_|  |_|\___/ \__,_|_|_|  \__, |   |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>                                                        |___/
!>      ____ _                  ___                              _ _       _       _
!>     / ___| | ___  ___  ___  |_ _|_ __ ___  _ __ ___   ___  __| (_) __ _| |_ ___| |_   _
!>    | |   | |/ _ \/ __|/ _ \  | || '_ ` _ \| '_ ` _ \ / _ \/ _` | |/ _` | __/ _ \ | | | |
!>    | |___| | (_) \__ \  __/  | || | | | | | | | | | |  __/ (_| | | (_| | ||  __/ | |_| |
!>     \____|_|\___/|___/\___| |___|_| |_| |_|_| |_| |_|\___|\__,_|_|\__,_|\__\___|_|\__, |
!>                                                                                   |___/
!>     _____ _     _       _____ _ _      _
!>    |_   _| |__ (_)___  |  ___(_) | ___| |
!>      | | | '_ \| / __| | |_  | | |/ _ \ |
!>      | | | | | | \__ \ |  _| | | |  __/_|
!>      |_| |_| |_|_|___/ |_|   |_|_|\___(_)
!>
!>      This is a template
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify the .template and .instantition instead
!> \par History
!>      1.2002 created
!>      4.2002 changed most of the methods, by making access to the list
!>             always through pointers (identifying pointer, list and iterators)
!>      6.2004 removed %initialized from list elements
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_linked_list_int

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_linked_list_int'

!API type
   PUBLIC :: cp_sll_int_type, cp_sll_int_p_type
!API common methods
   PUBLIC :: cp_create, cp_dealloc, cp_next
!API special get
   PUBLIC :: cp_get_first_el, cp_get_rest, cp_get_length, cp_get_element_at, cp_to_array
!API special set
   PUBLIC :: cp_set_element_at
!API structure manipulation
   PUBLIC :: cp_insert, cp_remove_first_el, cp_remove_el, cp_remove_all

   !API underlying routines
   PUBLIC :: cp_sll_int_create, cp_sll_int_dealloc, &
             cp_sll_int_next, &
             cp_sll_int_get_length, &
             cp_sll_int_get_el_at, cp_sll_int_set_el_at, &
             cp_sll_int_get_first_el, cp_sll_int_get_rest, &
             cp_sll_int_insert_el, cp_sll_int_insert_el_at, &
             cp_sll_int_rm_first_el, cp_sll_int_rm_el_at, &
             cp_sll_int_rm_all_el, &
             cp_sll_int_to_array

! creation of an object (from a pointer)
   INTERFACE cp_create
      MODULE PROCEDURE cp_sll_int_create
   END INTERFACE
! destruction of an object (from a pointer)
   INTERFACE cp_dealloc
      MODULE PROCEDURE cp_sll_int_dealloc
   END INTERFACE
! iterates to the next element
   INTERFACE cp_next
      MODULE PROCEDURE cp_sll_int_next
   END INTERFACE
! returns the first element
   INTERFACE cp_get_first_el
      MODULE PROCEDURE cp_sll_int_get_first_el
   END INTERFACE
! returns the rest of the list
   INTERFACE cp_get_rest
      MODULE PROCEDURE cp_sll_int_get_rest
   END INTERFACE
! returns the length of the list
   INTERFACE cp_get_length
      MODULE PROCEDURE cp_sll_int_get_length
   END INTERFACE
! returns the element at the given position
   INTERFACE cp_get_element_at
      MODULE PROCEDURE cp_sll_int_get_el_at
   END INTERFACE
! sets the element at the given position
   INTERFACE cp_set_element_at
      MODULE PROCEDURE cp_sll_int_set_el_at
   END INTERFACE
! inserts one element call cp_insert(list,element,...)
   INTERFACE cp_insert
      MODULE PROCEDURE cp_sll_int_insert_el
   END INTERFACE
   INTERFACE cp_insert_at
      MODULE PROCEDURE cp_sll_int_insert_el_at
   END INTERFACE
! removes an element
   INTERFACE cp_remove_el
      MODULE PROCEDURE cp_sll_int_rm_first_el, &
         cp_sll_int_rm_el_at
   END INTERFACE
! removes the first el
   INTERFACE cp_remove_first_el
      MODULE PROCEDURE cp_sll_int_rm_first_el
   END INTERFACE
! remove all the elments
   INTERFACE cp_remove_all
      MODULE PROCEDURE cp_sll_int_rm_all_el
   END INTERFACE
! transorms the list in array
   INTERFACE cp_to_array
      MODULE PROCEDURE cp_sll_int_to_array
   END INTERFACE

! **************************************************************************************************
!> \brief represent a single linked list that stores pointers to the elements
!> \param first_el the element that is stored in this node.
!> \param rest the rest of the list
!>
!> \param empty true if the list pointer is not associated, if it points to
!>             to a not it is always false (as there is at least the
!>             first_el in the list)
!> \param length the number of elements in the list
!> \note
!>      List are alway accessed through pointers, so every node of the
!>      linked list can be seen as a list, its first element
!>      a pointer to the position before itself, in a very natural way:
!>      all the insertions take place before the actual element, and
!>      you still can insert an element at the end.
!>      This way I could identify nodes, lists and pointers between the
!>      elements of the list.
!>      Indexing is 1 based.
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_int_type
      INTEGER :: first_el
      TYPE(cp_sll_int_type), POINTER :: rest
   END TYPE cp_sll_int_type

! **************************************************************************************************
!> \brief pointer to a linked list (to make arrays of pointers)
!> \param list the pointer to the list
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE cp_sll_int_p_type
      TYPE(cp_sll_int_type), POINTER :: list
   END TYPE cp_sll_int_p_type

CONTAINS

! =========== creation / distruction ========

! **************************************************************************************************
!> \brief allocates and initializes a single linked list
!> \param sll the single linked list to initialize
!> \param first_el the first element of this list
!> \param rest the following elements (if not given: empty)
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_int_create(sll, first_el, rest)
      TYPE(cp_sll_int_type), POINTER                     :: sll
      INTEGER, INTENT(in), OPTIONAL                      :: first_el
      TYPE(cp_sll_int_type), OPTIONAL, POINTER           :: rest

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_create', &
         routineP = moduleN//':'//routineN

      IF (.NOT. PRESENT(first_el)) THEN
         NULLIFY (sll)
         IF (PRESENT(rest)) sll => rest
      ELSE
         ALLOCATE (sll)
         sll%first_el = first_el
         NULLIFY (sll%rest)
         IF (PRESENT(rest)) sll%rest => rest
      END IF
   END SUBROUTINE cp_sll_int_create

! **************************************************************************************************
!> \brief deallocates the singly linked list starting at sll.
!>      Does not work if loops are present!
!> \param sll the list to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      does not deallocate the elments that are stored in the list
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_int_dealloc(sll)
      TYPE(cp_sll_int_type), POINTER                     :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_dealloc', &
         routineP = moduleN//':'//routineN

      CALL cp_sll_int_rm_all_el(sll)
   END SUBROUTINE cp_sll_int_dealloc

! * low-level *

! **************************************************************************************************
!> \brief deallocates a node of a singly linked list (low level)
!> \param sll the node to be deallocated
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_sll_int_dealloc_node(sll)
      TYPE(cp_sll_int_type), POINTER                     :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_dealloc_node', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (sll)
   END SUBROUTINE cp_sll_int_dealloc_node

! ============= get/set ============

! **************************************************************************************************
!> \brief returns the first element stored in the list
!> \param sll the single linked list to get the element from
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_int_get_first_el(sll) RESULT(res)
      TYPE(cp_sll_int_type), POINTER                     :: sll
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_get_first_el', &
         routineP = moduleN//':'//routineN

      res = sll%first_el
   END FUNCTION cp_sll_int_get_first_el

! **************************************************************************************************
!> \brief returns the rest of the list
!> \param sll the single linked list to get the rest from
!> \param iter how many times the call to rest should be iterated,
!>            defaults to 1; -1 means till end of the list.
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      split the case iter=1 to make it more optimized?
! **************************************************************************************************
   FUNCTION cp_sll_int_get_rest(sll, iter) RESULT(res)
      TYPE(cp_sll_int_type), POINTER                     :: sll
      INTEGER, OPTIONAL                                  :: iter
      TYPE(cp_sll_int_type), POINTER                     :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_get_rest', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (.NOT. ASSOCIATED(sll)) THEN
         NULLIFY (res)
      ELSE
         IF (PRESENT(iter)) THEN
            res => sll
            DO i = 1, iter
               IF (ASSOCIATED(res%rest)) THEN
                  res => res%rest
               ELSE
                  CALL cp__b("input/cp_linked_list_int.F",292,"tried to go past end")
               END IF
            END DO
            IF (iter == -1) THEN
               DO
                  IF (.NOT. ASSOCIATED(res%rest)) EXIT
                  res => res%rest
               END DO
            END IF
         ELSE
            res => sll%rest ! make the common case fast...
         END IF
      END IF
   END FUNCTION cp_sll_int_get_rest

! **************************************************************************************************
!> \brief returns the length of the list
!> \param sll the list you want to know the length of
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(n))
! **************************************************************************************************
   FUNCTION cp_sll_int_get_length(sll) RESULT(res)
      TYPE(cp_sll_int_type), POINTER                     :: sll
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_get_length', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_int_type), POINTER                     :: iterator

      res = 0
      iterator => sll
      DO
         IF (ASSOCIATED(iterator)) THEN
            res = res+1
            iterator => iterator%rest
         ELSE
            EXIT
         END IF
      END DO
   END FUNCTION cp_sll_int_get_length

! **************************************************************************************************
!> \brief returns the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   FUNCTION cp_sll_int_get_el_at(sll, index) RESULT(res)
      TYPE(cp_sll_int_type), POINTER                     :: sll
      INTEGER, INTENT(in)                                :: index
      INTEGER                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_get_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_int_type), POINTER                     :: pos

      IF (index == -1) THEN
         pos => cp_sll_int_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_int_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_int.F",364)

      res = pos%first_el
   END FUNCTION cp_sll_int_get_el_at

! **************************************************************************************************
!> \brief sets the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!>             -1 means at the end
!> \param value the new element
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_int_set_el_at(sll, index, value)
      TYPE(cp_sll_int_type), POINTER                     :: sll
      INTEGER, INTENT(in)                                :: index, value

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_set_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_int_type), POINTER                     :: pos

      IF (index == -1) THEN
         pos => cp_sll_int_get_rest(sll, iter=-1)
      ELSE
         pos => cp_sll_int_get_rest(sll, iter=index-1)
      END IF
      IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_int.F",395)

      pos%first_el = value
   END SUBROUTINE cp_sll_int_set_el_at

! * iteration *

! **************************************************************************************************
!> \brief returns true if the actual element is valid (i.e. iterator ont at end)
!>      moves the iterator to the next element
!> \param iterator iterator that moves along the list
!> \param el_att the actual element (valid only if the function returns true)
!> \retval res ...
!> \par History
!>      none
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_int_next(iterator, el_att) RESULT(res)
      TYPE(cp_sll_int_type), POINTER                     :: iterator
      INTEGER, INTENT(out), OPTIONAL                     :: el_att
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_next', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(iterator)) THEN
         res = .TRUE.
         IF (PRESENT(el_att)) el_att = iterator%first_el
         iterator => iterator%rest
      ELSE
         res = .FALSE.
      END IF
   END FUNCTION cp_sll_int_next

! ============ structure modifications ============

! **************************************************************************************************
!> \brief insert an element at the beginning of the list
!> \param sll the single linked list point at the beginning of which
!>           you want to add the element
!> \param el the element to add
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_int_insert_el(sll, el)
      TYPE(cp_sll_int_type), POINTER                     :: sll
      INTEGER, INTENT(in)                                :: el

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_insert_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_int_type), POINTER                     :: newSlot

      NULLIFY (newSlot)

      CALL cp_sll_int_create(newSlot, first_el=el, &
                             rest=sll)
      sll => newSlot
   END SUBROUTINE cp_sll_int_insert_el

! **************************************************************************************************
!> \brief remove the first element of the linked list
!> \param sll the list whose first element has to be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      fast (O(1))
! **************************************************************************************************
   SUBROUTINE cp_sll_int_rm_first_el(sll)
      TYPE(cp_sll_int_type), POINTER                     :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_rm_first_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_int_type), POINTER                     :: node_to_rm

      node_to_rm => sll

      IF (ASSOCIATED(sll)) THEN
         sll => sll%rest
         CALL cp_sll_int_dealloc_node(node_to_rm)
      ELSE
         CALL cp__b("input/cp_linked_list_int.F",481,"tried to remove first el of an empty list")
      END IF
   END SUBROUTINE cp_sll_int_rm_first_el

! **************************************************************************************************
!> \brief inserts the element at the given index
!> \param sll the list you get the element from
!> \param el the new element
!> \param index the position of the element (stating at 1).
!>             If it is -1, it means at end
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_int_insert_el_at(sll, el, index)
      TYPE(cp_sll_int_type), POINTER                     :: sll
      INTEGER, INTENT(in)                                :: el, index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_insert_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_int_type), POINTER                     :: pos

      IF (index == 1) THEN
         CALL cp_sll_int_insert_el(sll, el)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_int_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_int_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_int.F",514)
         CALL cp_sll_int_insert_el(pos%rest, el)
      END IF
   END SUBROUTINE cp_sll_int_insert_el_at

! **************************************************************************************************
!> \brief removes the element at the given index
!> \param sll the list you get the element from
!> \param index the position of the element (stating at 1)
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      slow (O(index))
! **************************************************************************************************
   SUBROUTINE cp_sll_int_rm_el_at(sll, index)
      TYPE(cp_sll_int_type), POINTER                     :: sll
      INTEGER, INTENT(in)                                :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_rm_el_at', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_int_type), POINTER                     :: pos

      IF (index == 1) THEN
         CALL cp_sll_int_rm_first_el(sll)
      ELSE
         IF (index == -1) THEN
            pos => cp_sll_int_get_rest(sll, iter=-1)
         ELSE
            pos => cp_sll_int_get_rest(sll, iter=index-2)
         END IF
         IF(.NOT.(ASSOCIATED(pos)))CALL cp__a("input/cp_linked_list_int.F",546)
         CALL cp_sll_int_rm_first_el(pos%rest)
      END IF
   END SUBROUTINE cp_sll_int_rm_el_at

! **************************************************************************************************
!> \brief removes all the elements from the list
!> \param sll the list that should be removed
!> \par History
!>      none
!> \author Fawzi Mohamed
!> \note
!>      check more?
! **************************************************************************************************
   SUBROUTINE cp_sll_int_rm_all_el(sll)
      TYPE(cp_sll_int_type), POINTER                     :: sll

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_rm_all_el', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_int_type), POINTER                     :: actual_node, next_node

      actual_node => sll
      DO
         IF (.NOT. ASSOCIATED(actual_node)) EXIT
         next_node => actual_node%rest
         CALL cp_sll_int_dealloc_node(actual_node)
         actual_node => next_node
      END DO
      NULLIFY (sll)
   END SUBROUTINE cp_sll_int_rm_all_el

! **************************************************************************************************
!> \brief returns a newly allocated array with the same contents as
!>      the linked list
!> \param sll the list to trasform in array
!> \retval res ...
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   FUNCTION cp_sll_int_to_array(sll) RESULT(res)
      TYPE(cp_sll_int_type), POINTER                     :: sll
      INTEGER, DIMENSION(:), POINTER                     :: res

      INTEGER                                            :: i, len
      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sll_int_to_array', &
         routineP = moduleN//':'//routineN

      TYPE(cp_sll_int_type), POINTER                     :: iter

      len = cp_sll_int_get_length(sll)
      ALLOCATE (res(len))
      iter => sll
      DO i = 1, len
         res(i) = iter%first_el
         IF(.NOT.(cp_sll_int_next(iter) .OR. i == len))CALL cp__a("input/cp_linked_list_int.F",602)
      END DO
   END FUNCTION cp_sll_int_to_array

! template def put here so that line numbers in template and derived
! files are almost the same (multi-line use change it a bit)
! [template(defines,nametype1,type1,type1in,type1out,type1arrayEl,arrayEl,array=,=,USE,write_el,lessQ,equalQ,private_routines)]
! ARGS:
!  = = "="
!  USE = " "
!  array= = "="
!  arrayEl = ""
!  common_dir = "../common"
!  defines =
!    "#define CP_SLL_I_LESS_Q(el1,el2) ( el1 < el2 )
!     #define ( el1 == el2 ) ( el1 == el2 )
!     "
!  equalQ = "CP_SLL_I_EQUAL_Q"
!  lessQ = "CP_SLL_I_LESS_Q"
!  nametype1 = "int"
!  private_routines = ""
!  template = "../common/cp_linked_list__nametype1_.template"
!  type1 = "integer"
!  type1arrayEl = "integer"
!  type1in = "integer, intent(in)"
!  type1out = "integer, intent(out)"
!  write_el = ""

END MODULE cp_linked_list_int

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines that parse the input
!> \par History
!>      06.2004 created
!> \author fawzi
! **************************************************************************************************
MODULE input_parsing
   USE cp_linked_list_char,             ONLY: cp_create,&
                                              cp_dealloc,&
                                              cp_sll_char_type,&
                                              cp_to_array
   USE cp_linked_list_int,              ONLY: cp_create,&
                                              cp_dealloc,&
                                              cp_sll_int_type,&
                                              cp_to_array
   USE cp_linked_list_logical,          ONLY: cp_create,&
                                              cp_dealloc,&
                                              cp_sll_logical_type,&
                                              cp_to_array
   USE cp_linked_list_real,             ONLY: cp_create,&
                                              cp_dealloc,&
                                              cp_sll_real_type,&
                                              cp_to_array
   USE cp_linked_list_val,              ONLY: cp_sll_val_create,&
                                              cp_sll_val_type
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit,&
                                              cp_to_string
   USE cp_parser_methods,               ONLY: parser_get_object,&
                                              parser_location,&
                                              parser_skip_space,&
                                              parser_test_next_token
   USE cp_parser_types,                 ONLY: cp_parser_type
   USE cp_units,                        ONLY: cp_unit_compatible,&
                                              cp_unit_create,&
                                              cp_unit_desc,&
                                              cp_unit_release,&
                                              cp_unit_set_type,&
                                              cp_unit_to_cp2k1,&
                                              cp_unit_type
   USE input_enumeration_types,         ONLY: enum_c2i,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_describe,&
                                              keyword_type
   USE input_section_types,             ONLY: &
        section_describe, section_get_keyword, section_get_keyword_index, &
        section_get_subsection_index, section_type, section_typo_match, section_vals_add_values, &
        section_vals_type, typo_match_section, typo_matching_line, typo_matching_rank
   USE input_val_types,                 ONLY: &
        char_t, enum_t, integer_t, lchar_t, logical_t, no_t, real_t, val_create, val_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              max_line_length
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_parsing'

   PUBLIC :: section_vals_parse
!***
CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param section_vals ...
!> \param parser ...
!> \param default_units ...
!> \param root_section if the root section should be parsed (defaults to true)
!> \author fawzi
! **************************************************************************************************
   RECURSIVE SUBROUTINE section_vals_parse(section_vals, parser, default_units, root_section)
      TYPE(section_vals_type), POINTER                   :: section_vals
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(cp_unit_set_type), POINTER                    :: default_units
      LOGICAL, INTENT(in), OPTIONAL                      :: root_section

      CHARACTER(len=*), PARAMETER :: routineN = 'section_vals_parse', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=max_line_length)                     :: token
      INTEGER                                            :: desc_level, handle, ik, imatch, irs, is, &
                                                            nsub, output_unit
      LOGICAL                                            :: at_end, compatible_end, root_sect, &
                                                            whole_section
      TYPE(cp_sll_val_type), POINTER                     :: last_val, new_val, previous_last, &
                                                            previous_list
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(val_type), POINTER                            :: el

      CALL timeset(routineN, handle)

      NULLIFY (previous_list, previous_last)

      root_sect = .TRUE.
      IF (PRESENT(root_section)) root_sect = root_section

      IF(.NOT.(ASSOCIATED(section_vals)))CALL cp__a("input/input_parsing.F",106)
      IF(.NOT.(ASSOCIATED(parser)))CALL cp__a("input/input_parsing.F",107)
      output_unit = cp_logger_get_default_io_unit()

      IF(.NOT.(section_vals%ref_count > 0))CALL cp__a("input/input_parsing.F",110)
      IF(.NOT.(parser%ref_count > 0))CALL cp__a("input/input_parsing.F",111)
      IF (root_sect .AND. parser%icol1 > parser%icol2) &
         CALL cp_abort(cp__l("input/input_parsing.F",113), &
                       "Error 1: this routine must be called just after having parsed the start of the section " &
                       //TRIM(parser_location(parser)))
      section => section_vals%section
      IF (root_sect) THEN
         token = parser%input_line(parser%icol1:parser%icol2)
         CALL uppercase(token)
         IF (token /= parser%section_character//section%name) &
            CALL cp_abort(cp__l("input/input_parsing.F",121), &
                          "Error 2: this routine must be called just after having parsed the start of the section " &
                          //TRIM(parser_location(parser)))
      END IF
      IF (.NOT. section%repeats .AND. SIZE(section_vals%values, 2) /= 0) &
         CALL cp_abort(cp__l("input/input_parsing.F",126), "Section "//TRIM(section%name)// &
                       " should not repeat "//TRIM(parser_location(parser)))
      CALL section_vals_add_values(section_vals)
      irs = SIZE(section_vals%values, 2)

      IF (ASSOCIATED(section%keywords(-1)%keyword)) THEN ! reads section params
         keyword => section%keywords(-1)%keyword
         NULLIFY (el)
         IF (keyword%type_of_var == lchar_t) CALL parser_skip_space(parser)
         CALL val_create_parsing(el, type_of_var=keyword%type_of_var, &
                                 n_var=keyword%n_var, default_value=keyword%lone_keyword_value, &
                                 enum=keyword%enum, unit=keyword%unit, &
                                 default_units=default_units, &
                                 parser=parser)
         NULLIFY (new_val)
         CALL cp_sll_val_create(new_val, el)
         section_vals%values(-1, irs)%list => new_val
         NULLIFY (el)
      END IF
      DO WHILE (.TRUE.)
         CALL parser_get_object(parser, token, newline=.TRUE., &
                                lower_to_upper=.TRUE., at_end=at_end)
         IF (at_end) THEN
            IF (root_sect) &
               CALL cp_abort(cp__l("input/input_parsing.F",150), &
                             "unexpected end of file while parsing section "// &
                             TRIM(section%name)//" "//TRIM(parser_location(parser)))
            EXIT
         END IF
         IF (token(1:1) == parser%section_character) THEN
            IF (token == "&END") THEN
               ! end of section
               compatible_end = .TRUE.
               IF (parser_test_next_token(parser) /= "EOL") THEN
                  CALL parser_get_object(parser, token, newline=.FALSE., &
                                         lower_to_upper=.TRUE.)
                  IF (token /= "SECTION" .AND. token /= section%name) THEN
                     compatible_end = .FALSE.
                  END IF
               END IF
               IF (parser_test_next_token(parser) /= "EOL") THEN
                  CALL parser_get_object(parser, token, newline=.FALSE., &
                                         lower_to_upper=.TRUE.)
                  IF (token /= section%name) THEN
                     PRINT*, TRIM(token), "/=", TRIM(section%name)
                     compatible_end = .FALSE.
                  END IF
               END IF
               IF (.NOT. compatible_end) THEN
                  CALL cp_abort(cp__l("input/input_parsing.F",175), &
                                "non-compatible end of section "//TRIM(section%name)//" "// &
                                TRIM(parser_location(parser)))
               END IF
               !              RETURN
               EXIT
            END IF
            is = section_get_subsection_index(section, token(2:))
            IF (is > 0) THEN
               CALL section_vals_parse(section_vals%subs_vals(is, irs)%section_vals, &
                                       default_units=default_units, parser=parser)
            ELSE
               ! unknown subsection
               IF (output_unit > 0) THEN
                  WRITE (output_unit, *)
                  WRITE (output_unit, '(T2,A)') "Possible matches for unknown subsection "
                  WRITE (output_unit, *)
                  WRITE (output_unit, '(T2,A)') TRIM(token(2:))
                  WRITE (output_unit, *)
                  CALL section_typo_match(typo_match_section, TRIM(section%name), TRIM(token(2:)), "", &
                                          typo_matching_rank, typo_matching_line, bonus=0)
                  DO imatch = 1, SIZE(typo_matching_rank)
                     WRITE (output_unit, '(T2,A,1X,I0)') TRIM(typo_matching_line(imatch))//" score: ", typo_matching_rank(imatch)
                  ENDDO
               ENDIF
               CALL cp_abort(cp__l("input/input_parsing.F",200), &
                             "unknown subsection "//TRIM(token(2:))//" of section " &
                             //TRIM(section%name))
               nSub = 1
               DO WHILE (nSub > 0)
                  CALL parser_get_object(parser, token, newline=.TRUE., &
                                         lower_to_upper=.TRUE.)
                  IF (token(1:1) == parser%section_character) THEN
                     IF (token == "&END") THEN
                        nSub = nSub-1
                     ELSE
                        nSub = nSub+1
                     END IF
                  END IF
               END DO
            END IF
         ELSE ! token is a keyword
            IF (token == "DESCRIBE") THEN
               IF (output_unit > 0) WRITE (output_unit, "(/,' ****** DESCRIPTION ******',/)")
               desc_level = 3
               IF (parser_test_next_token(parser) == "INT") THEN
                  CALL parser_get_object(parser, desc_level)
               END IF
               whole_section = .TRUE.
               DO WHILE (parser_test_next_token(parser) == "STR")
                  whole_section = .FALSE.
                  CALL parser_get_object(parser, token, newline=.FALSE., &
                                         lower_to_upper=.TRUE.)
                  keyword => section_get_keyword(section, token)
                  IF (.NOT. ASSOCIATED(keyword)) THEN
                     CALL cp_warn(cp__l("input/input_parsing.F",230), &
                                  "unknown keyword to describe "//TRIM(token)// &
                                  " in section "//TRIM(section%name))
                  ELSE
                     CALL keyword_describe(keyword, output_unit, desc_level)
                  END IF
               END DO
               IF (whole_section) THEN
                  CALL section_describe(section, output_unit, desc_level, hide_root=.NOT. root_sect)
               END IF
               IF (output_unit > 0) WRITE (output_unit, "(/,' ****** =========== ******',/)")

            ELSE ! token is a "normal" keyword
               ik = section_get_keyword_index(section, token)
               IF (ik < 1) THEN ! don't accept pseudo keyword names
                  parser%icol = parser%icol1-1 ! re-read also the actual token
                  ik = 0
                  IF (.NOT. ASSOCIATED(section%keywords(0)%keyword)) THEN
                     IF (output_unit > 0) THEN
                        WRITE (output_unit, *)
                        WRITE (output_unit, '(T2,A)') "Possible matches for unknown keyword "
                        WRITE (output_unit, *)
                        WRITE (output_unit, '(T2,A)') TRIM(token)
                        WRITE (output_unit, *)
                        CALL section_typo_match(typo_match_section, TRIM(section%name), TRIM(token), "", &
                                                typo_matching_rank, typo_matching_line, bonus=0)
                        DO imatch = 1, SIZE(typo_matching_rank)
                           WRITE (output_unit, '(T2,A,1X,I0)') &
                              TRIM(typo_matching_line(imatch))//" score: ", typo_matching_rank(imatch)
                        ENDDO
                     ENDIF
                     CALL cp_abort(cp__l("input/input_parsing.F",261), &
                                   "found an unknown keyword "//TRIM(token)// &
                                   " in section "//TRIM(section%name))
                  END IF
               END IF
               keyword => section%keywords(ik)%keyword
               IF (ASSOCIATED(keyword)) THEN
                  NULLIFY (el)
                  IF (ik /= 0 .AND. keyword%type_of_var == lchar_t) &
                     CALL parser_skip_space(parser)
                  CALL val_create_parsing(el, type_of_var=keyword%type_of_var, &
                                          n_var=keyword%n_var, default_value=keyword%lone_keyword_value, &
                                          enum=keyword%enum, unit=keyword%unit, &
                                          default_units=default_units, parser=parser)
                  IF (ASSOCIATED(el)) THEN
                     NULLIFY (new_val)
                     CALL cp_sll_val_create(new_val, el)
                     last_val => section_vals%values(ik, irs)%list
                     IF (.NOT. ASSOCIATED(last_val)) THEN
                        section_vals%values(ik, irs)%list => new_val
                     ELSE
                        IF (.NOT. keyword%repeats) &
                           CALL cp_abort(cp__l("input/input_parsing.F",283), &
                                         "Keyword "//TRIM(token)// &
                                         " in section "//TRIM(section%name)//" should not repeat.")
                        IF (ASSOCIATED(last_val, previous_list)) THEN
                           last_val => previous_last
                        ELSE
                           previous_list => last_val
                        ENDIF
                        DO WHILE (ASSOCIATED(last_val%rest))
                           last_val => last_val%rest
                        END DO
                        last_val%rest => new_val
                        previous_last => new_val
                     END IF
                  END IF
               END IF
            END IF
         END IF
      END DO
      CALL timestop(handle)
   END SUBROUTINE section_vals_parse

! **************************************************************************************************
!> \brief creates a val_type object by parsing the values
!> \param val the value that will be created
!> \param type_of_var type of the value to be created
!> \param n_var number of values to be parsed (-1: undefined)
!> \param enum ...
!> \param parser the parser from where the values should be read
!> \param unit ...
!> \param default_units ...
!> \param default_value a default value if nothing is found (can be null)
!> \author fawzi
!> \note
!>      - no_t does not create a value
! **************************************************************************************************
   SUBROUTINE val_create_parsing(val, type_of_var, n_var, enum, &
                                 parser, unit, default_units, default_value)
      TYPE(val_type), POINTER                            :: val
      INTEGER, INTENT(in)                                :: type_of_var, n_var
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(cp_unit_type), POINTER                        :: unit
      TYPE(cp_unit_set_type), POINTER                    :: default_units
      TYPE(val_type), OPTIONAL, POINTER                  :: default_value

      CHARACTER(len=*), PARAMETER :: routineN = 'val_create_parsing', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: c_val, info, location
      CHARACTER(len=default_string_length), &
         DIMENSION(:), POINTER                           :: c_val_p
      INTEGER                                            :: handle, i, i_val
      INTEGER, DIMENSION(:), POINTER                     :: i_val_p
      LOGICAL                                            :: check, l_val
      LOGICAL, DIMENSION(:), POINTER                     :: l_val_p
      REAL(kind=dp)                                      :: r_val
      REAL(kind=dp), DIMENSION(:), POINTER               :: r_val_p
      TYPE(cp_sll_char_type), POINTER                    :: c_first, c_last, c_new
      TYPE(cp_sll_int_type), POINTER                     :: i_first, i_last, i_new
      TYPE(cp_sll_logical_type), POINTER                 :: l_first, l_last, l_new
      TYPE(cp_sll_real_type), POINTER                    :: r_first, r_last, r_new

      CALL timeset(routineN, handle)

      IF(.NOT.(.NOT. ASSOCIATED(val)))CALL cp__a("input/input_parsing.F",348)
      SELECT CASE (type_of_var)
      CASE (no_t)
      CASE (logical_t)
         NULLIFY (l_val_p)
         IF (parser_test_next_token(parser) == "EOL") THEN
            IF (.NOT. ASSOCIATED(default_value)) THEN
               IF (n_var < 1) THEN
                  ALLOCATE (l_val_p(0))
                  CALL val_create(val, l_vals_ptr=l_val_p)
               ELSE
                  CALL cp_abort(cp__l("input/input_parsing.F",359), &
                                "no value was given and there is no default value"// &
                                TRIM(parser_location(parser)))
               END IF
            ELSE
               IF(.NOT.(ASSOCIATED(default_value%l_val)))CALL cp__a("input/input_parsing.F",364)
               CALL val_create(val, l_vals=default_value%l_val)
            END IF
         ELSE
            IF (n_var < 1) THEN
               NULLIFY (l_last, l_first)
               CALL parser_get_object(parser, l_val)
               CALL cp_create(l_first, l_val)
               l_last => l_first
               DO WHILE (parser_test_next_token(parser) /= "EOL")
                  CALL parser_get_object(parser, l_val)
                  CALL cp_create(l_new, l_val)
                  l_last%rest => l_new
                  l_last => l_new
               END DO
               l_val_p => cp_to_array(l_first)
               CALL cp_dealloc(l_first)
            ELSE
               ALLOCATE (l_val_p(n_var))
               DO i = 1, n_var
                  CALL parser_get_object(parser, l_val_p(i))
               END DO
            END IF
            IF (ASSOCIATED(l_val_p)) THEN
               CALL val_create(val, l_vals_ptr=l_val_p)
            END IF
         END IF
      CASE (integer_t)
         NULLIFY (i_val_p)
         IF (parser_test_next_token(parser) == "EOL") THEN
            IF (.NOT. ASSOCIATED(default_value)) THEN
               IF (n_var < 1) THEN
                  ALLOCATE (i_val_p(0))
                  CALL val_create(val, i_vals_ptr=i_val_p)
               ELSE
                  CALL cp_abort(cp__l("input/input_parsing.F",399), &
                                "no value was given and there is no default value"// &
                                TRIM(parser_location(parser)))
               END IF
            ELSE
               check = ASSOCIATED(default_value%i_val)
               IF(.NOT.(check))CALL cp__a("input/input_parsing.F",405)
               CALL val_create(val, i_vals=default_value%i_val)
            END IF
         ELSE
            IF (n_var < 1) THEN
               NULLIFY (i_last, i_first)
               CALL parser_get_object(parser, i_val)
               CALL cp_create(i_first, i_val)
               i_last => i_first
               DO WHILE (parser_test_next_token(parser) /= "EOL")
                  CALL parser_get_object(parser, i_val)
                  CALL cp_create(i_new, i_val)
                  i_last%rest => i_new
                  i_last => i_new
               END DO
               i_val_p => cp_to_array(i_first)
               CALL cp_dealloc(i_first)
            ELSE
               ALLOCATE (i_val_p(n_var))
               DO i = 1, n_var
                  CALL parser_get_object(parser, i_val_p(i))
               END DO
            END IF
            IF (ASSOCIATED(i_val_p)) THEN
               CALL val_create(val, i_vals_ptr=i_val_p)
            END IF
         END IF
      CASE (real_t)
         NULLIFY (r_val_p)
         IF (parser_test_next_token(parser) == "EOL") THEN
            IF (.NOT. ASSOCIATED(default_value)) THEN
               IF (n_var < 1) THEN
                  ALLOCATE (r_val_p(0))
                  CALL val_create(val, r_vals_ptr=r_val_p)
               ELSE
                  CALL cp_abort(cp__l("input/input_parsing.F",440), &
                                "no value was given and there is no default value"// &
                                TRIM(parser_location(parser)))
               END IF
            ELSE
               IF(.NOT.(ASSOCIATED(default_value%r_val)))CALL cp__a("input/input_parsing.F",445)
               CALL val_create(val, r_vals=default_value%r_val)
            END IF
         ELSE
            IF (n_var < 1) THEN
               NULLIFY (r_last, r_first)
               c_val = ""
               CALL get_r_val(r_val, parser, unit, default_units, c_val)
               CALL cp_create(r_first, r_val)
               r_last => r_first
               DO WHILE (parser_test_next_token(parser) /= "EOL")
                  CALL get_r_val(r_val, parser, unit, default_units, c_val)
                  CALL cp_create(r_new, r_val)
                  r_last%rest => r_new
                  r_last => r_new
               END DO
               NULLIFY (r_last)
               r_val_p => cp_to_array(r_first)
               CALL cp_dealloc(r_first)
            ELSE
               ALLOCATE (r_val_p(n_var))
               c_val = ""
               DO i = 1, n_var
                  CALL get_r_val(r_val_p(i), parser, unit, default_units, c_val)
               END DO
            END IF
            IF (ASSOCIATED(r_val_p)) THEN
               CALL val_create(val, r_vals_ptr=r_val_p)
            END IF
         END IF
      CASE (char_t)
         NULLIFY (c_val_p)
         IF (parser_test_next_token(parser) == "EOL") THEN
            IF (n_var < 1) THEN
               ALLOCATE (c_val_p(1))
               c_val_p(1) = ' '
               CALL val_create(val, c_vals_ptr=c_val_p)
            ELSE
               IF (.NOT. ASSOCIATED(default_value)) THEN
                  CALL cp_abort(cp__l("input/input_parsing.F",484), &
                                "no value was given and there is no default value"// &
                                TRIM(parser_location(parser)))
               ELSE
                  IF(.NOT.(ASSOCIATED(default_value%c_val)))CALL cp__a("input/input_parsing.F",488)
                  CALL val_create(val, c_vals=default_value%c_val)
               END IF
            END IF
         ELSE
            IF (n_var < 1) THEN
               IF(.NOT.(n_var == -1))CALL cp__a("input/input_parsing.F",494)
               NULLIFY (c_last, c_first)
               CALL parser_get_object(parser, c_val)
               CALL cp_create(c_first, c_val)
               c_last => c_first
               DO WHILE (parser_test_next_token(parser) /= "EOL")
                  CALL parser_get_object(parser, c_val)
                  CALL cp_create(c_new, c_val)
                  c_last%rest => c_new
                  c_last => c_new
               END DO
               c_val_p => cp_to_array(c_first)
               CALL cp_dealloc(c_first)
            ELSE
               ALLOCATE (c_val_p(n_var))
               DO i = 1, n_var
                  CALL parser_get_object(parser, c_val_p(i))
               END DO
            END IF
            IF (ASSOCIATED(c_val_p)) THEN
               CALL val_create(val, c_vals_ptr=c_val_p)
            END IF
         END IF
      CASE (lchar_t)
         IF (ASSOCIATED(default_value)) &
            CALL cp_abort(cp__l("input/input_parsing.F",519), &
                          "input variables of type lchar_t cannot have a lone keyword attribute,"// &
                          " no value is interpreted as empty string"// &
                          TRIM(parser_location(parser)))
         IF (n_var /= 1) &
            CALL cp_abort(cp__l("input/input_parsing.F",524), &
                          "input variables of type lchar_t cannot be repeated,"// &
                          " one always represent a whole line, till the end"// &
                          TRIM(parser_location(parser)))
         IF (parser_test_next_token(parser) == "EOL") THEN
            ALLOCATE (c_val_p(1))
            c_val_p(1) = ' '
         ELSE
            NULLIFY (c_last, c_first)
            CALL parser_get_object(parser, c_val, string_length=LEN(c_val))
            CALL cp_create(c_first, c_val)
            c_last => c_first
            DO WHILE (parser_test_next_token(parser) /= "EOL")
               CALL parser_get_object(parser, c_val, string_length=LEN(c_val))
               CALL cp_create(c_new, c_val)
               c_last%rest => c_new
               c_last => c_new
            END DO
            c_val_p => cp_to_array(c_first)
            CALL cp_dealloc(c_first)
         END IF
         IF(.NOT.(ASSOCIATED(c_val_p)))CALL cp__a("input/input_parsing.F",545)
         CALL val_create(val, lc_vals_ptr=c_val_p)
      CASE (enum_t)
         IF(.NOT.(ASSOCIATED(enum)))CALL cp__a("input/input_parsing.F",548)
         NULLIFY (i_val_p)
         IF (parser_test_next_token(parser) == "EOL") THEN
            IF (.NOT. ASSOCIATED(default_value)) THEN
               IF (n_var < 1) THEN
                  ALLOCATE (i_val_p(0))
                  CALL val_create(val, i_vals_ptr=i_val_p)
               ELSE
                  CALL cp_abort(cp__l("input/input_parsing.F",556), &
                                "no value was given and there is no default value"// &
                                TRIM(parser_location(parser)))
               END IF
            ELSE
               IF(.NOT.(ASSOCIATED(default_value%i_val)))CALL cp__a("input/input_parsing.F",561)
               CALL val_create(val, i_vals=default_value%i_val, &
                               enum=default_value%enum)
            END IF
         ELSE
            IF (n_var < 1) THEN
               NULLIFY (i_last, i_first)
               CALL parser_get_object(parser, c_val)
               CALL cp_create(i_first, enum_c2i(enum, c_val))
               i_last => i_first
               DO WHILE (parser_test_next_token(parser) /= "EOL")
                  CALL parser_get_object(parser, c_val)
                  CALL cp_create(i_new, enum_c2i(enum, c_val))
                  i_last%rest => i_new
                  i_last => i_new
               END DO
               i_val_p => cp_to_array(i_first)
               CALL cp_dealloc(i_first)
            ELSE
               ALLOCATE (i_val_p(n_var))
               DO i = 1, n_var
                  CALL parser_get_object(parser, c_val)
                  i_val_p(i) = enum_c2i(enum, c_val)
               END DO
            END IF
            IF (ASSOCIATED(i_val_p)) THEN
               CALL val_create(val, i_vals_ptr=i_val_p, enum=enum)
            END IF
         END IF
      CASE default
         CALL cp_abort(cp__l("input/input_parsing.F",591), &
                       "type "//cp_to_string(type_of_var)//"unknown to the parser")
      END SELECT
      IF (parser_test_next_token(parser) .NE. "EOL") THEN
         location = TRIM(parser_location(parser))
         CALL parser_get_object(parser, info)
         CALL cp_abort(cp__l("input/input_parsing.F",597), &
                       "found unexpected extra argument "//TRIM(info)//" at "//location)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE val_create_parsing

! **************************************************************************************************
!> \brief Reads and convert a real number from the input file
!> \param r_val ...
!> \param parser the parser from where the values should be read
!> \param unit ...
!> \param default_units ...
!> \param c_val ...
!> \author Teodoro Laino - 11.2007 [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE get_r_val(r_val, parser, unit, default_units, c_val)
      REAL(kind=dp), INTENT(OUT)                         :: r_val
      TYPE(cp_parser_type), POINTER                      :: parser
      TYPE(cp_unit_type), POINTER                        :: unit
      TYPE(cp_unit_set_type), POINTER                    :: default_units
      CHARACTER(len=default_string_length), &
         INTENT(INOUT)                                   :: c_val

      CHARACTER(len=*), PARAMETER :: routineN = 'get_r_val', routineP = moduleN//':'//routineN

      LOGICAL                                            :: check
      TYPE(cp_unit_type), POINTER                        :: my_unit

      NULLIFY (my_unit)
      IF (ASSOCIATED(unit)) THEN
         IF ('STR' == parser_test_next_token(parser)) THEN
            CALL parser_get_object(parser, c_val)
            check = c_val(1:1) == "["
            IF(.NOT.(check))CALL cp__a("input/input_parsing.F",632)
            check = c_val(LEN_TRIM(c_val):LEN_TRIM(c_val)) == "]"
            IF(.NOT.(check))CALL cp__a("input/input_parsing.F",634)
            CALL cp_unit_create(my_unit, c_val(2:LEN_TRIM(c_val)-1))
         ELSE
            IF (c_val /= "") THEN
               CALL cp_unit_create(my_unit, c_val(2:LEN_TRIM(c_val)-1))
            ELSE
               my_unit => unit
            END IF
         END IF
         IF (.NOT. cp_unit_compatible(unit, my_unit)) &
            CALL cp_abort(cp__l("input/input_parsing.F",644), &
                          "Incompatible units. Defined as ("// &
                          TRIM(cp_unit_desc(unit))//") specified in input as ("// &
                          TRIM(cp_unit_desc(my_unit))//"). These units are incompatible!")
      END IF
      CALL parser_get_object(parser, r_val)
      IF (ASSOCIATED(unit)) THEN
         r_val = cp_unit_to_cp2k1(r_val, my_unit, default_units)
         IF (.NOT. (ASSOCIATED(my_unit, unit))) CALL cp_unit_release(my_unit)
      END IF

   END SUBROUTINE get_r_val

END MODULE input_parsing
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief checks the input and perform some automatic "magic" on it
!> \par History
!>      01.2006 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE input_cp2k_check
   USE cp_log_handling,                 ONLY: cp_to_string
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_set_create,&
                                              cp_unit_set_release,&
                                              cp_unit_set_type
   USE input_constants,                 ONLY: &
        do_region_global, do_thermo_al, do_thermo_csvr, do_thermo_gle, do_thermo_nose, &
        do_thermo_same_as_part, npt_f_ensemble, npt_i_ensemble, vdw_nl_LMKLL, xc_funct_b3lyp, &
        xc_funct_beefvdw, xc_funct_blyp, xc_funct_bp, xc_funct_hcth120, xc_funct_no_shortcut, &
        xc_funct_olyp, xc_funct_pade, xc_funct_pbe, xc_funct_pbe0, xc_funct_tpss, xc_funct_xwpbe, &
        xc_none, xc_vdw_fun_nonloc
   USE input_keyword_types,             ONLY: keyword_type
   USE input_parsing,                   ONLY: section_vals_parse
   USE input_section_types,             ONLY: &
        section_type, section_vals_create, section_vals_get, section_vals_get_subs_vals, &
        section_vals_get_subs_vals3, section_vals_release, section_vals_remove_values, &
        section_vals_set_subs_vals, section_vals_type, section_vals_val_get, section_vals_val_set, &
        section_vals_val_unset
   USE input_val_types,                 ONLY: logical_t
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE xc_input_constants,              ONLY: do_vwn5
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_check'

   PUBLIC :: check_cp2k_input, xc_functionals_expand, remove_restart_info

CONTAINS

! **************************************************************************************************
!> \brief performs further checks on an input that parsed successfully
!> \param input_declaration ...
!> \param input_file the parsed input
!> \param para_env ...
!> \param output_unit ...
!> \author fawzi
!> \note
!>      at the moment does nothing
! **************************************************************************************************
   SUBROUTINE check_cp2k_input(input_declaration, input_file, para_env, output_unit)
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: input_file
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN), OPTIONAL                      :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'check_cp2k_input', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iforce_eval, nforce_eval
      LOGICAL                                            :: explicit, explicit_mix
      TYPE(section_vals_type), POINTER                   :: section, section1, section2, section3, &
                                                            sections

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(input_file)))CALL cp__a("input_cp2k_check.F",77)
      IF(.NOT.(input_file%ref_count > 0))CALL cp__a("input_cp2k_check.F",78)
      ! ext_restart
      IF (PRESENT(output_unit)) &
         CALL handle_ext_restart(input_declaration, input_file, para_env, output_unit)

      ! checks on force_eval section
      sections => section_vals_get_subs_vals(input_file, "FORCE_EVAL")
      CALL section_vals_get(sections, n_repetition=nforce_eval)

      ! multiple force_eval only if present RESPA or MIXED calculation is performed
      section2 => section_vals_get_subs_vals(input_file, "MOTION%MD%RESPA")
      CALL section_vals_get(section2, explicit=explicit)
      DO iforce_eval = 1, nforce_eval
         section3 => section_vals_get_subs_vals(sections, "MIXED", &
                                                i_rep_section=iforce_eval)
         CALL section_vals_get(section3, explicit=explicit_mix)
         IF (explicit_mix) EXIT
      END DO

      IF ((explicit .AND. (nforce_eval == 1)) .OR. (.NOT. explicit .AND. (nforce_eval > 1))) THEN
         IF ((explicit_mix .AND. (nforce_eval == 1)) .OR. (.NOT. explicit_mix .AND. (nforce_eval > 1))) THEN
            CALL cp_abort(cp__l("input_cp2k_check.F",99), &
                          "Error multiple force_env without RESPA or MIXED, or RESPA with one single "// &
                          " or MIXED with only two force_env section.")
         END IF
      END IF
      DO iforce_eval = 1, nforce_eval
         section => section_vals_get_subs_vals3(sections, "DFT", i_rep_section=iforce_eval)
         ! xc: expand and fix default for tddfpt
         section1 => section_vals_get_subs_vals(section, "XC")
         section2 => section_vals_get_subs_vals(section, "XC%XC_FUNCTIONAL")
         CALL xc_functionals_expand(section2, section1)
         section1 => section_vals_get_subs_vals(section, "TDDFPT%XC")
         section2 => section_vals_get_subs_vals(section, "TDDFPT%XC%XC_FUNCTIONAL")
         CALL section_vals_get(section2, explicit=explicit)
         IF (explicit) THEN
            CALL xc_functionals_expand(section2, section1)
         ELSE
            section2 => section_vals_get_subs_vals(section, "XC%XC_FUNCTIONAL")
            CALL section_vals_set_subs_vals(section, "TDDFPT%XC%XC_FUNCTIONAL", section2)
         END IF
      END DO

      CALL timestop(handle)
   END SUBROUTINE check_cp2k_input

! **************************************************************************************************
!> \brief expand a shortcutted functional section
!> \param functionals the functional section to expand
!> \param xc_section ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xc_functionals_expand(functionals, xc_section)
      TYPE(section_vals_type), POINTER                   :: functionals, xc_section

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_functionals_expand', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: shortcut

      CALL section_vals_val_get(functionals, "_SECTION_PARAMETERS_", &
                                i_val=shortcut)
      SELECT CASE (shortcut)
      CASE (xc_funct_no_shortcut, xc_none)
         ! nothing to expand
      CASE (xc_funct_pbe0)
         CALL section_vals_val_set(functionals, "PBE%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "PBE%SCALE_X", &
                                   r_val=0.75_dp)
         CALL section_vals_val_set(functionals, "PBE%SCALE_C", &
                                   r_val=1.0_dp)
         ! Hartree Fock Exact Exchange
         CALL section_vals_val_set(xc_section, "HF%FRACTION", &
                                   r_val=0.25_dp)
         CALL section_vals_val_set(functionals, "_SECTION_PARAMETERS_", &
                                   i_val=xc_funct_no_shortcut)
      CASE (xc_funct_beefvdw)
         CALL section_vals_val_set(functionals, "PBE%_SECTION_PARAMETERS_", & !40% PBEc
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "PBE%SCALE_C", &
                                   r_val=0.3998335231_dp)
         CALL section_vals_val_set(functionals, "PBE%SCALE_X", & !no PBEx
                                   r_val=0.0000000000_dp)

         !PW92 correlation functional from libxc is required.
         !The cp2k-native PW92 gives disagreeing results (in the 0.01E_H
         !decimal) and yields inconsistent forces in a DEBUG run.
         !(rk, 6.3.2014)
         CALL section_vals_val_set(functionals, "LIBXC%_SECTION_PARAMETERS_", & !60%LDA
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "LIBXC%FUNCTIONAL", &
                                   c_val="LDA_C_PW")
         CALL section_vals_val_set(functionals, "LIBXC%SCALE", &
                                   r_val=0.6001664769_dp)

         CALL section_vals_val_set(functionals, "BEEF%_SECTION_PARAMETERS_", & !BEEF exchange
                                   l_val=.TRUE.)

         !NONLOCAL, LMKLL.
         CALL section_vals_val_set(xc_section, "VDW_POTENTIAL%DISPERSION_FUNCTIONAL", &
                                   i_val=xc_vdw_fun_nonloc)
         CALL section_vals_val_set(xc_section, "VDW_POTENTIAL%NON_LOCAL%TYPE", &
                                   i_val=vdw_nl_LMKLL)
         CALL section_vals_val_set(functionals, "_SECTION_PARAMETERS_", &
                                   i_val=xc_funct_no_shortcut)
      CASE (xc_funct_b3lyp)
         CALL section_vals_val_set(functionals, "BECKE88%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "BECKE88%SCALE_X", &
                                   r_val=0.72_dp)
         CALL section_vals_val_set(functionals, "LYP%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "LYP%SCALE_C", &
                                   r_val=0.81_dp)
         CALL section_vals_val_set(functionals, "VWN%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "VWN%FUNCTIONAL_TYPE", &
                                   i_val=do_vwn5)
         CALL section_vals_val_set(functionals, "VWN%SCALE_C", &
                                   r_val=0.19_dp)
         CALL section_vals_val_set(functionals, "XALPHA%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "XALPHA%SCALE_X", &
                                   r_val=0.08_dp)
         ! Hartree Fock Exact Exchange
         CALL section_vals_val_set(xc_section, "HF%FRACTION", &
                                   r_val=0.20_dp)
         CALL section_vals_val_set(functionals, "_SECTION_PARAMETERS_", &
                                   i_val=xc_funct_no_shortcut)
      CASE (xc_funct_blyp)
         CALL section_vals_val_set(functionals, "BECKE88%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "LYP%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "_SECTION_PARAMETERS_", &
                                   i_val=xc_funct_no_shortcut)
      CASE (xc_funct_bp)
         CALL section_vals_val_set(functionals, "BECKE88%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "P86C%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "_SECTION_PARAMETERS_", &
                                   i_val=xc_funct_no_shortcut)
      CASE (xc_funct_pade)
         CALL section_vals_val_set(functionals, "PADE%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "_SECTION_PARAMETERS_", &
                                   i_val=xc_funct_no_shortcut)
      CASE (xc_funct_pbe)
         CALL section_vals_val_set(functionals, "PBE%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "_SECTION_PARAMETERS_", &
                                   i_val=xc_funct_no_shortcut)
      CASE (xc_funct_xwpbe)
         CALL section_vals_val_set(functionals, "XWPBE%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "_SECTION_PARAMETERS_", &
                                   i_val=xc_funct_no_shortcut)
      CASE (xc_funct_tpss)
         CALL section_vals_val_set(functionals, "TPSS%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "_SECTION_PARAMETERS_", &
                                   i_val=xc_funct_no_shortcut)
      CASE (xc_funct_olyp)
         CALL section_vals_val_set(functionals, "OPTX%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "LYP%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "_SECTION_PARAMETERS_", &
                                   i_val=xc_funct_no_shortcut)
      CASE (xc_funct_hcth120)
         CALL section_vals_val_set(functionals, "HCTH%_SECTION_PARAMETERS_", &
                                   l_val=.TRUE.)
         CALL section_vals_val_set(functionals, "HCTH%PARAMETER_SET", &
                                   i_val=120)
         CALL section_vals_val_set(functionals, "_SECTION_PARAMETERS_", &
                                   i_val=xc_funct_no_shortcut)
      CASE default
         CALL cp__b("input_cp2k_check.F",257,"unknown shortcut "//TRIM(ADJUSTL(cp_to_string(shortcut))))
      END SELECT
   END SUBROUTINE xc_functionals_expand

! **************************************************************************************************
!> \brief Replaces the requested sections in the input with those found
!>      in the external restart (EXT_RESTART%RESTART_FILE_NAME).
!> \param input_declaration ...
!> \param input_file the input file to initialize
!> \param para_env ...
!> \param output_unit ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE handle_ext_restart(input_declaration, input_file, para_env, output_unit)
      TYPE(section_type), POINTER                        :: input_declaration
      TYPE(section_vals_type), POINTER                   :: input_file
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'handle_ext_restart', &
         routineP = moduleN//':'//routineN

      CHARACTER(default_path_length)                     :: binary_restart_file, r_file_path
      CHARACTER(default_string_length)                   :: path
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: restarted_infos
      INTEGER                                            :: ensemble, handle, i_rep_val, &
                                                            iforce_eval, myi, n_rep_val, &
                                                            nforce_eval1, nforce_eval2
      INTEGER, DIMENSION(:), POINTER                     :: ivec, iwalkers_status, iwork, &
                                                            rwalkers_status
      LOGICAL                                            :: bsse_check, check, explicit1, explicit2, &
                                                            flag, flag2, qmmm_check, subsys_check
      REAL(KIND=dp)                                      :: myt
      REAL(KIND=dp), DIMENSION(:), POINTER               :: vec, work
      TYPE(cp_parser_type), POINTER                      :: cpparser
      TYPE(cp_unit_set_type), POINTER                    :: default_units
      TYPE(section_vals_type), POINTER                   :: r_section, rep_sections, restart_file, &
                                                            section, section1, section2, &
                                                            sections1, sections2

      NULLIFY (restarted_infos, iwalkers_status, rwalkers_status, vec, ivec, work, iwork)
      CALL timeset(routineN, handle)
      ! Handle restart file
      r_section => section_vals_get_subs_vals(input_file, "EXT_RESTART")
      CALL section_vals_val_get(r_section, "RESTART_FILE_NAME", c_val=r_file_path)
      CALL section_vals_val_get(r_section, "BINARY_RESTART_FILE_NAME", c_val=binary_restart_file)

      IF (r_file_path /= " ") THEN
         ! parse the input
         NULLIFY (default_units, restart_file)
         CALL section_vals_create(restart_file, input_declaration)
         NULLIFY (cpparser)
         CALL parser_create(cpparser, file_name=r_file_path, para_env=para_env)
         CALL cp_unit_set_create(default_units, "OUTPUT")
         CALL section_vals_parse(restart_file, cpparser, root_section=.FALSE., &
                                 default_units=default_units)
         CALL cp_unit_set_release(default_units)
         CALL parser_release(cpparser)

         ! Restart and input files same number of force_env sections
         sections1 => section_vals_get_subs_vals(restart_file, "FORCE_EVAL")
         CALL section_vals_get(sections1, n_repetition=nforce_eval1)
         sections2 => section_vals_get_subs_vals(input_file, "FORCE_EVAL")
         CALL section_vals_get(sections2, n_repetition=nforce_eval2)
         IF (nforce_eval1 /= nforce_eval2) THEN
            CALL cp__b("input_cp2k_check.F",323,"Restart and input file MUST have the number of force_env sections")
         END IF
         ! Handle default restarts
         CALL handle_defaults_restart(r_section)

         ! Real restart of force_evals
         DO iforce_eval = 1, nforce_eval1
            section1 => section_vals_get_subs_vals3(sections1, "SUBSYS", &
                                                    i_rep_section=iforce_eval)
            section2 => section_vals_get_subs_vals3(sections2, "SUBSYS", &
                                                    i_rep_section=iforce_eval)
            ! Some care needs to be handled when treating multiple force_eval
            ! Both subsys need to be consistently associated or not
            ! Mixed stuff will be rejected for safety reason..
            subsys_check = (ASSOCIATED(section1) .EQV. ASSOCIATED(section2))
            IF (subsys_check) THEN
               IF (ASSOCIATED(section1)) THEN
                  CALL section_vals_val_get(r_section, "RESTART_CELL", l_val=flag)
                  IF (flag) THEN
                     section => section_vals_get_subs_vals(section1, "CELL")
                     CALL section_vals_set_subs_vals(section2, "CELL", section)
                     CALL set_restart_info("CELL", restarted_infos)
                  END IF

                  CALL section_vals_val_get(r_section, "RESTART_POS", l_val=flag)
                  IF (flag) THEN
                     section => section_vals_get_subs_vals(section1, "COORD")
                     CALL section_vals_set_subs_vals(section2, "COORD", section)
                     CALL set_restart_info("COORDINATES", restarted_infos)
                     ! Copy over also the information on the multiple_unit_cell
                     CALL section_vals_val_get(section1, "TOPOLOGY%MULTIPLE_UNIT_CELL", i_vals=ivec)
                     ALLOCATE (iwork(3))
                     iwork = ivec
                     CALL section_vals_val_set(section2, "TOPOLOGY%MULTIPLE_UNIT_CELL", i_vals_ptr=iwork)
                  END IF

                  CALL section_vals_val_get(r_section, "RESTART_RANDOMG", l_val=flag)
                  IF (flag) THEN
                     section => section_vals_get_subs_vals(section1, "RNG_INIT")
                     CALL section_vals_set_subs_vals(section2, "RNG_INIT", section)
                     CALL set_restart_info("RANDOM NUMBER GENERATOR", restarted_infos)
                  END IF

                  CALL section_vals_val_get(r_section, "RESTART_VEL", l_val=flag)
                  IF (flag) THEN
                     section => section_vals_get_subs_vals(section1, "VELOCITY")
                     CALL section_vals_set_subs_vals(section2, "VELOCITY", section)
                     CALL set_restart_info("VELOCITIES", restarted_infos)
                  END IF

                  ! Core-Shell information "restarted" only when strictly necessary
                  CALL section_vals_val_get(r_section, "RESTART_SHELL_POS", l_val=flag)
                  IF (flag) THEN
                     section => section_vals_get_subs_vals(section1, "SHELL_COORD")
                     CALL section_vals_set_subs_vals(section2, "SHELL_COORD", section)
                     IF (check_restart(section1, section2, "SHELL_COORD")) &
                        CALL set_restart_info("SHELL COORDINATES", restarted_infos)
                  END IF
                  CALL section_vals_val_get(r_section, "RESTART_CORE_POS", l_val=flag)
                  IF (flag) THEN
                     section => section_vals_get_subs_vals(section1, "CORE_COORD")
                     CALL section_vals_set_subs_vals(section2, "CORE_COORD", section)
                     IF (check_restart(section1, section2, "CORE_COORD")) &
                        CALL set_restart_info("CORE COORDINATES", restarted_infos)
                  END IF
                  CALL section_vals_val_get(r_section, "RESTART_SHELL_VELOCITY", l_val=flag)
                  IF (flag) THEN
                     section => section_vals_get_subs_vals(section1, "SHELL_VELOCITY")
                     CALL section_vals_set_subs_vals(section2, "SHELL_VELOCITY", section)
                     IF (check_restart(section1, section2, "SHELL_VELOCITY")) &
                        CALL set_restart_info("SHELL VELOCITIES", restarted_infos)
                  END IF
                  CALL section_vals_val_get(r_section, "RESTART_CORE_VELOCITY", l_val=flag)
                  IF (flag) THEN
                     section => section_vals_get_subs_vals(section1, "CORE_VELOCITY")
                     CALL section_vals_set_subs_vals(section2, "CORE_VELOCITY", section)
                     IF (check_restart(section1, section2, "CORE_VELOCITY")) &
                        CALL set_restart_info("CORE VELOCITIES", restarted_infos)
                  END IF
               END IF
            ELSE
               CALL cp_abort(cp__l("input_cp2k_check.F",404), &
                             "Error while reading the restart file. Two force_eval have incompatible"// &
                             " subsys.One of them has an allocated subsys while the other has not! Check your"// &
                             " input file or whether the restart file is compatible with the input!")
            END IF
            ! QMMM restarts
            CALL section_vals_val_get(r_section, "RESTART_QMMM", l_val=flag)
            section1 => section_vals_get_subs_vals3(sections1, "QMMM", i_rep_section=iforce_eval)
            section2 => section_vals_get_subs_vals3(sections2, "QMMM", i_rep_section=iforce_eval)
            CALL section_vals_get(section1, explicit=explicit1)
            CALL section_vals_get(section2, explicit=explicit2)
            qmmm_check = (explicit1 .AND. explicit2)
            IF (flag .AND. qmmm_check) THEN
               CALL set_restart_info("QMMM TRANSLATION VECTOR", restarted_infos)
               CALL section_vals_val_get(section1, "INITIAL_TRANSLATION_VECTOR", r_vals=vec)
               ALLOCATE (work(3))
               work = vec
               CALL section_vals_val_set(section2, "INITIAL_TRANSLATION_VECTOR", r_vals_ptr=work)
            END IF
            ! BSSE restarts
            CALL section_vals_val_get(r_section, "RESTART_BSSE", l_val=flag)
            section1 => section_vals_get_subs_vals3(sections1, "BSSE", i_rep_section=iforce_eval)
            section2 => section_vals_get_subs_vals3(sections2, "BSSE", i_rep_section=iforce_eval)
            CALL section_vals_get(section1, explicit=explicit1)
            CALL section_vals_get(section2, explicit=explicit2)
            bsse_check = (explicit1 .AND. explicit2)
            IF (flag .AND. bsse_check) THEN
               section => section_vals_get_subs_vals(section1, "FRAGMENT_ENERGIES")
               CALL section_vals_set_subs_vals(section2, "FRAGMENT_ENERGIES", section)
               CALL set_restart_info("BSSE FRAGMENT ENERGIES", restarted_infos)
            END IF
         END DO

         CALL section_vals_val_get(r_section, "RESTART_COUNTERS", l_val=flag)
         IF (flag) THEN
            IF (check_restart(input_file, restart_file, "MOTION%MD")) THEN
               CALL section_vals_val_get(restart_file, "MOTION%MD%STEP_START_VAL", i_val=myi)
               CALL section_vals_val_set(input_file, "MOTION%MD%STEP_START_VAL", i_val=myi)
               CALL section_vals_val_get(restart_file, "MOTION%MD%TIME_START_VAL", r_val=myt)
               CALL section_vals_val_set(input_file, "MOTION%MD%TIME_START_VAL", r_val=myt)
               CALL section_vals_val_get(restart_file, "MOTION%MD%ECONS_START_VAL", r_val=myt)
               CALL section_vals_val_set(input_file, "MOTION%MD%ECONS_START_VAL", r_val=myt)
               CALL set_restart_info("MD COUNTERS", restarted_infos)
            END IF
            !
            IF (check_restart(input_file, restart_file, "MOTION%GEO_OPT")) THEN
               ! GEO_OPT
               CALL section_vals_val_get(restart_file, "MOTION%GEO_OPT%STEP_START_VAL", i_val=myi)
               CALL section_vals_val_set(input_file, "MOTION%GEO_OPT%STEP_START_VAL", i_val=myi)
               CALL set_restart_info("GEO_OPT COUNTERS", restarted_infos)
               ! ROT_OPT
               IF (check_restart(input_file, restart_file, "MOTION%GEO_OPT%TRANSITION_STATE%DIMER%ROT_OPT")) THEN
                  CALL section_vals_val_get(restart_file, "MOTION%GEO_OPT%TRANSITION_STATE%DIMER%ROT_OPT%STEP_START_VAL", &
                                            i_val=myi)
                  CALL section_vals_val_set(input_file, "MOTION%GEO_OPT%TRANSITION_STATE%DIMER%ROT_OPT%STEP_START_VAL", &
                                            i_val=myi)
                  CALL set_restart_info("ROT_OPT COUNTERS", restarted_infos)
               END IF
            END IF
            !
            IF (check_restart(input_file, restart_file, "MOTION%GEO_OPT")) THEN
               ! CELL_OPT
               CALL section_vals_val_get(restart_file, "MOTION%CELL_OPT%STEP_START_VAL", i_val=myi)
               CALL section_vals_val_set(input_file, "MOTION%CELL_OPT%STEP_START_VAL", i_val=myi)
               CALL set_restart_info("CELL_OPT COUNTERS", restarted_infos)
            END IF
            !
            IF (check_restart(input_file, restart_file, "OPTIMIZE_INPUT")) THEN
               CALL section_vals_val_get(restart_file, "OPTIMIZE_INPUT%ITER_START_VAL", i_val=myi)
               CALL section_vals_val_set(input_file, "OPTIMIZE_INPUT%ITER_START_VAL", i_val=myi)
               CALL set_restart_info("OPTIMIZE_INPUT ITERATION NUMBER", restarted_infos)
            END IF
            !
            IF (check_restart(input_file, restart_file, "MOTION%PINT")) THEN
               ! PINT
               CALL section_vals_val_get(restart_file, "MOTION%PINT%ITERATION", i_val=myi)
               CALL section_vals_val_set(input_file, "MOTION%PINT%ITERATION", i_val=myi)
               CALL set_restart_info("PINT ITERATION NUMBER", restarted_infos)
            END IF
            !
            CALL section_vals_val_get(r_section, "RESTART_METADYNAMICS", l_val=flag2)
            IF (flag2 .AND. check_restart(input_file, restart_file, "MOTION%FREE_ENERGY%METADYN")) THEN
               CALL section_vals_val_get(restart_file, &
                                         "MOTION%FREE_ENERGY%METADYN%STEP_START_VAL", i_val=myi)
               CALL section_vals_val_set(input_file, &
                                         "MOTION%FREE_ENERGY%METADYN%STEP_START_VAL", i_val=myi)
               CALL section_vals_val_get(restart_file, &
                                         "MOTION%FREE_ENERGY%METADYN%NHILLS_START_VAL", i_val=myi)
               CALL section_vals_val_set(input_file, &
                                         "MOTION%FREE_ENERGY%METADYN%NHILLS_START_VAL", i_val=myi)
               !RG Adaptive hills
               CALL section_vals_val_get(restart_file, &
                                         "MOTION%FREE_ENERGY%METADYN%OLD_HILL_NUMBER", i_val=myi)
               CALL section_vals_val_set(input_file, &
                                         "MOTION%FREE_ENERGY%METADYN%OLD_HILL_NUMBER", i_val=myi)
               CALL section_vals_val_get(restart_file, &
                                         "MOTION%FREE_ENERGY%METADYN%OLD_HILL_STEP", i_val=myi)
               CALL section_vals_val_set(input_file, &
                                         "MOTION%FREE_ENERGY%METADYN%OLD_HILL_STEP", i_val=myi)
               !RG Adaptive hills
               CALL set_restart_info("METADYNAMIC COUNTERS", restarted_infos)
            END IF
         END IF

         CALL section_vals_val_get(r_section, "RESTART_AVERAGES", l_val=flag)
         IF (flag) THEN
            IF (check_restart(input_file, restart_file, "MOTION%MD")) THEN
               rep_sections => section_vals_get_subs_vals(restart_file, "MOTION%MD%AVERAGES%RESTART_AVERAGES")
               CALL section_vals_set_subs_vals(input_file, "MOTION%MD%AVERAGES%RESTART_AVERAGES", rep_sections)
               CALL set_restart_info("MD AVERAGES", restarted_infos)
            END IF
         END IF

         CALL section_vals_val_get(r_section, "RESTART_BAND", l_val=flag)
         IF (flag .AND. check_restart(input_file, restart_file, "MOTION%BAND")) THEN
            rep_sections => section_vals_get_subs_vals(restart_file, "MOTION%BAND%REPLICA")
            CALL section_vals_set_subs_vals(input_file, "MOTION%BAND%REPLICA", rep_sections)
            CALL set_restart_info("BAND CALCULATION", restarted_infos)
         END IF

         CALL section_vals_val_get(r_section, "RESTART_OPTIMIZE_INPUT_VARIABLES", l_val=flag)
         IF (flag .AND. check_restart(input_file, restart_file, "OPTIMIZE_INPUT%VARIABLE")) THEN
            rep_sections => section_vals_get_subs_vals(restart_file, "OPTIMIZE_INPUT%VARIABLE")
            CALL section_vals_set_subs_vals(input_file, "OPTIMIZE_INPUT%VARIABLE", rep_sections)
            CALL set_restart_info("OPTIMIZE_INPUT: VARIABLES", restarted_infos)
         END IF

         CALL section_vals_val_get(r_section, "RESTART_BAROSTAT", l_val=flag)
         IF (flag .AND. check_restart(input_file, restart_file, "MOTION%MD%BAROSTAT")) THEN
            section => section_vals_get_subs_vals(restart_file, &
                                                  "MOTION%MD%BAROSTAT%MASS")
            CALL section_vals_set_subs_vals(input_file, "MOTION%MD%BAROSTAT%MASS", &
                                            section)
            section => section_vals_get_subs_vals(restart_file, &
                                                  "MOTION%MD%BAROSTAT%VELOCITY")
            CALL section_vals_set_subs_vals(input_file, "MOTION%MD%BAROSTAT%VELOCITY", &
                                            section)
            CALL set_restart_info("BAROSTAT", restarted_infos)
         END IF

         flag = check_restart(input_file, restart_file, "MOTION%MD")
         IF (flag) THEN
            CALL section_vals_val_get(input_file, "MOTION%MD%ENSEMBLE", i_val=ensemble)
            IF (ensemble == npt_i_ensemble .OR. ensemble == npt_f_ensemble) THEN
               CALL section_vals_val_get(r_section, "RESTART_BAROSTAT_THERMOSTAT", l_val=flag)
               check = check_restart(input_file, restart_file, "MOTION%MD%BAROSTAT")
               CALL restart_thermostat(flag, input_file, restart_file, "MOTION%MD%BAROSTAT%THERMOSTAT", &
                                       check=check)
               IF (flag .AND. check) CALL set_restart_info("THERMOSTAT OF BAROSTAT", restarted_infos)
            END IF
         END IF

         check = check_restart(input_file, restart_file, "MOTION%MD%SHELL")
         IF (check) THEN
            CALL section_vals_val_get(r_section, "RESTART_SHELL_THERMOSTAT", l_val=flag)
            CALL restart_thermostat(flag, input_file, restart_file, "MOTION%MD%SHELL%THERMOSTAT")
            CALL set_restart_info("SHELL THERMOSTAT", restarted_infos)
         END IF

         CALL section_vals_val_get(r_section, "RESTART_THERMOSTAT", l_val=flag)
         CALL restart_thermostat(flag, input_file, restart_file, "MOTION%MD%THERMOSTAT")
         IF (flag) CALL set_restart_info("PARTICLE THERMOSTAT", restarted_infos)

         CALL section_vals_val_get(r_section, "RESTART_CONSTRAINT", l_val=flag)
         IF (flag .AND. check_restart(input_file, restart_file, "MOTION%CONSTRAINT")) THEN
            section => section_vals_get_subs_vals(restart_file, "MOTION%CONSTRAINT")
            CALL section_vals_set_subs_vals(input_file, "MOTION%CONSTRAINT", section)
            CALL set_restart_info("CONSTRAINTS/RESTRAINTS", restarted_infos)
         END IF

         CALL section_vals_val_get(r_section, "RESTART_METADYNAMICS", l_val=flag)
         IF (flag .AND. check_restart(input_file, restart_file, "MOTION%FREE_ENERGY%METADYN")) THEN
            section => section_vals_get_subs_vals(restart_file, &
                                                  "MOTION%FREE_ENERGY%METADYN%SPAWNED_HILLS_POS")
            CALL section_vals_set_subs_vals(input_file, "MOTION%FREE_ENERGY%METADYN%SPAWNED_HILLS_POS", &
                                            section)
            section => section_vals_get_subs_vals(restart_file, &
                                                  "MOTION%FREE_ENERGY%METADYN%SPAWNED_HILLS_SCALE")
            CALL section_vals_set_subs_vals(input_file, "MOTION%FREE_ENERGY%METADYN%SPAWNED_HILLS_SCALE", &
                                            section)
            section => section_vals_get_subs_vals(restart_file, &
                                                  "MOTION%FREE_ENERGY%METADYN%SPAWNED_HILLS_HEIGHT")
            CALL section_vals_set_subs_vals(input_file, "MOTION%FREE_ENERGY%METADYN%SPAWNED_HILLS_HEIGHT", &
                                            section)
            section => section_vals_get_subs_vals(restart_file, &
                                                  "MOTION%FREE_ENERGY%METADYN%SPAWNED_HILLS_INVDT")
            CALL section_vals_set_subs_vals(input_file, "MOTION%FREE_ENERGY%METADYN%SPAWNED_HILLS_INVDT", &
                                            section)
            ! Extended Lagrangian
            section => section_vals_get_subs_vals(restart_file, &
                                                  "MOTION%FREE_ENERGY%METADYN%EXT_LAGRANGE_SS0")
            CALL section_vals_set_subs_vals(input_file, "MOTION%FREE_ENERGY%METADYN%EXT_LAGRANGE_SS0", &
                                            section)
            section => section_vals_get_subs_vals(restart_file, &
                                                  "MOTION%FREE_ENERGY%METADYN%EXT_LAGRANGE_VVP")
            CALL section_vals_set_subs_vals(input_file, "MOTION%FREE_ENERGY%METADYN%EXT_LAGRANGE_VVP", &
                                            section)
            section => section_vals_get_subs_vals(restart_file, &
                                                  "MOTION%FREE_ENERGY%METADYN%EXT_LAGRANGE_SS")
            CALL section_vals_set_subs_vals(input_file, "MOTION%FREE_ENERGY%METADYN%EXT_LAGRANGE_SS", &
                                            section)
            section => section_vals_get_subs_vals(restart_file, &
                                                  "MOTION%FREE_ENERGY%METADYN%EXT_LAGRANGE_FS")
            CALL section_vals_set_subs_vals(input_file, "MOTION%FREE_ENERGY%METADYN%EXT_LAGRANGE_FS", &
                                            section)
            CALL set_restart_info("METADYNAMICS", restarted_infos)
         END IF

         CALL section_vals_val_get(r_section, "RESTART_WALKERS", l_val=flag)
         IF (flag .AND. check_restart(input_file, restart_file, "MOTION%FREE_ENERGY%METADYN%MULTIPLE_WALKERS")) THEN
            CALL section_vals_val_get(restart_file, "MOTION%FREE_ENERGY%METADYN%MULTIPLE_WALKERS%WALKERS_STATUS", &
                                      i_vals=rwalkers_status)
            ALLOCATE (iwalkers_status(SIZE(rwalkers_status)))
            iwalkers_status = rwalkers_status
            CALL section_vals_val_set(input_file, "MOTION%FREE_ENERGY%METADYN%MULTIPLE_WALKERS%WALKERS_STATUS", &
                                      i_vals_ptr=iwalkers_status)
            CALL set_restart_info("WALKERS INFO", restarted_infos)
         END IF

         CALL section_vals_val_get(r_section, "RESTART_DIMER", l_val=flag)
         IF (flag .AND. check_restart(input_file, restart_file, "MOTION%GEO_OPT%TRANSITION_STATE%DIMER")) THEN
            section => section_vals_get_subs_vals(restart_file, &
                                                  "MOTION%GEO_OPT%TRANSITION_STATE%DIMER%DIMER_VECTOR")
            CALL section_vals_set_subs_vals(input_file, "MOTION%GEO_OPT%TRANSITION_STATE%DIMER%DIMER_VECTOR", &
                                            section)
            CALL set_restart_info("DIMER TRANSITION STATE SEARCH", restarted_infos)
         END IF

         CALL section_vals_val_get(r_section, "CUSTOM_PATH", n_rep_val=n_rep_val)
         DO i_rep_val = 1, n_rep_val
            CALL section_vals_val_get(r_section, "CUSTOM_PATH", i_rep_val=i_rep_val, c_val=path)
            IF (path /= " ") THEN
               section => section_vals_get_subs_vals(restart_file, path)
               CALL section_vals_set_subs_vals(input_file, path, section)
               CALL set_restart_info("USER RESTART: "//TRIM(path), restarted_infos)
            END IF
         END DO

         CALL section_vals_val_get(r_section, "RESTART_RTP", l_val=flag)
         !          IF(flag.AND.check_restart(input_file, restart_file, "FORCE_EVAL%DFT%REAL_TIME_PROPAGATION")) THEN
         IF (flag) THEN
            section => section_vals_get_subs_vals(restart_file, &
                                                  "FORCE_EVAL%DFT%REAL_TIME_PROPAGATION")
            CALL section_vals_val_get(section, "INITIAL_WFN", i_val=myi)
            CALL section_vals_val_set(input_file, "FORCE_EVAL%DFT%REAL_TIME_PROPAGATION%INITIAL_WFN", &
                                      i_val=myi)
            CALL set_restart_info("REAL TIME PROPAGATION", restarted_infos)
         END IF

         ! PIMD
         CALL section_vals_val_get(r_section, "RESTART_PINT_POS", l_val=flag)
         IF (flag) THEN
            section => section_vals_get_subs_vals(restart_file, "MOTION%PINT%BEADS%COORD")
            CALL section_vals_set_subs_vals(input_file, "MOTION%PINT%BEADS%COORD", section)
            CALL set_restart_info("PINT BEAD POSITIONS", restarted_infos)
         END IF
         CALL section_vals_val_get(r_section, "RESTART_PINT_VEL", l_val=flag)
         IF (flag) THEN
            section => section_vals_get_subs_vals(restart_file, "MOTION%PINT%BEADS%VELOCITY")
            CALL section_vals_set_subs_vals(input_file, "MOTION%PINT%BEADS%VELOCITY", section)
            CALL set_restart_info("PINT BEAD VELOCITIES", restarted_infos)
         END IF
         CALL section_vals_val_get(r_section, "RESTART_PINT_NOSE", l_val=flag)
         IF (flag) THEN
            section => section_vals_get_subs_vals(restart_file, "MOTION%PINT%NOSE%COORD")
            CALL section_vals_set_subs_vals(input_file, "MOTION%PINT%NOSE%COORD", section)
            section => section_vals_get_subs_vals(restart_file, "MOTION%PINT%NOSE%VELOCITY")
            CALL section_vals_set_subs_vals(input_file, "MOTION%PINT%NOSE%VELOCITY", section)
            CALL set_restart_info("PINT NOSE THERMOSTAT", restarted_infos)
         END IF
         CALL section_vals_val_get(r_section, "RESTART_PINT_GLE", l_val=flag)
         IF (flag) THEN
            section => section_vals_get_subs_vals(restart_file, "MOTION%PINT%GLE")
            CALL section_vals_set_subs_vals(input_file, "MOTION%PINT%GLE", section)
            CALL set_restart_info("PINT GLE THERMOSTAT", restarted_infos)
         END IF

         ! PIMC
         !
         CALL section_vals_val_get(r_section, "RESTART_HELIUM_POS", l_val=flag)
         IF (flag) THEN
            CALL section_vals_val_get(input_file, "MOTION%PINT%HELIUM%NUM_ENV", &
                                      explicit=explicit1)
            IF (.NOT. explicit1) THEN
               CALL section_vals_val_get(restart_file, "MOTION%PINT%HELIUM%NUM_ENV", i_val=myi)
               CALL section_vals_val_set(input_file, "MOTION%PINT%HELIUM%NUM_ENV", i_val=myi)
            END IF
            section => section_vals_get_subs_vals(restart_file, "MOTION%PINT%HELIUM%COORD")
            CALL section_vals_set_subs_vals(input_file, "MOTION%PINT%HELIUM%COORD", section)
            CALL set_restart_info("HELIUM BEAD POSITIONS", restarted_infos)
         END IF
         !
         CALL section_vals_val_get(r_section, "RESTART_HELIUM_PERMUTATION", l_val=flag)
         IF (flag) THEN
            CALL section_vals_val_get(input_file, "MOTION%PINT%HELIUM%NUM_ENV", &
                                      explicit=explicit1)
            IF (.NOT. explicit1) THEN
               CALL section_vals_val_get(restart_file, "MOTION%PINT%HELIUM%NUM_ENV", i_val=myi)
               CALL section_vals_val_set(input_file, "MOTION%PINT%HELIUM%NUM_ENV", i_val=myi)
            END IF
            section => section_vals_get_subs_vals(restart_file, "MOTION%PINT%HELIUM%PERM")
            CALL section_vals_set_subs_vals(input_file, "MOTION%PINT%HELIUM%PERM", section)
            CALL set_restart_info("HELIUM PERMUTATION STATE", restarted_infos)
         END IF
         !
         CALL section_vals_val_get(r_section, "RESTART_HELIUM_FORCE", l_val=flag)
         IF (flag) THEN
            CALL section_vals_val_get(input_file, "MOTION%PINT%HELIUM%NUM_ENV", &
                                      explicit=explicit1)
            IF (.NOT. explicit1) THEN
               CALL section_vals_val_get(restart_file, "MOTION%PINT%HELIUM%NUM_ENV", i_val=myi)
               CALL section_vals_val_set(input_file, "MOTION%PINT%HELIUM%NUM_ENV", i_val=myi)
            END IF
            section => section_vals_get_subs_vals(restart_file, "MOTION%PINT%HELIUM%FORCE")
            CALL section_vals_set_subs_vals(input_file, "MOTION%PINT%HELIUM%FORCE", section)
            CALL set_restart_info("HELIUM FORCES ON SOLUTE", restarted_infos)
         END IF
         !
         CALL section_vals_val_get(r_section, "RESTART_HELIUM_RNG", l_val=flag)
         IF (flag) THEN
            CALL section_vals_val_get(input_file, "MOTION%PINT%HELIUM%NUM_ENV", &
                                      explicit=explicit1)
            IF (.NOT. explicit1) THEN
               CALL section_vals_val_get(restart_file, "MOTION%PINT%HELIUM%NUM_ENV", i_val=myi)
               CALL section_vals_val_set(input_file, "MOTION%PINT%HELIUM%NUM_ENV", i_val=myi)
            END IF
            section => section_vals_get_subs_vals(restart_file, "MOTION%PINT%HELIUM%RNG_STATE")
            CALL section_vals_set_subs_vals(input_file, "MOTION%PINT%HELIUM%RNG_STATE", section)
            CALL set_restart_info("HELIUM RNG STATE", restarted_infos)
         END IF
         !
         !
         CALL section_vals_val_get(r_section, "RESTART_HELIUM_DENSITIES", l_val=flag)
         IF (flag) THEN
            CALL section_vals_val_get(input_file, "MOTION%PINT%HELIUM%NUM_ENV", &
                                      explicit=explicit1)
            IF (.NOT. explicit1) THEN
               CALL section_vals_val_get(restart_file, "MOTION%PINT%HELIUM%NUM_ENV", i_val=myi)
               CALL section_vals_val_set(input_file, "MOTION%PINT%HELIUM%NUM_ENV", i_val=myi)
            END IF
            section => section_vals_get_subs_vals(restart_file, "MOTION%PINT%HELIUM%RHO")
            CALL section_vals_set_subs_vals(input_file, "MOTION%PINT%HELIUM%RHO", section)
            CALL set_restart_info("HELIUM DENSITIES", restarted_infos)
         END IF
         !
         CALL section_vals_val_set(r_section, "RESTART_FILE_NAME", c_val=" ")
         CALL section_vals_release(restart_file)
         CALL release_restart_info(restarted_infos, r_file_path, binary_restart_file, &
                                   output_unit)
      END IF
      CALL timestop(handle)
   END SUBROUTINE handle_ext_restart

! **************************************************************************************************
!> \brief store information on the restarted quantities
!> \param label ...
!> \param restarted_infos ...
!> \author Teodoro Laino [tlaino] 09.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE set_restart_info(label, restarted_infos)

      CHARACTER(LEN=*), INTENT(IN)                       :: label
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: restarted_infos

      CHARACTER(len=*), PARAMETER :: routineN = 'set_restart_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: isize

      isize = 0
      IF (ASSOCIATED(restarted_infos)) isize = SIZE(restarted_infos)
      isize = isize+1
      CALL reallocate(restarted_infos, 1, isize)
      restarted_infos(isize) = TRIM(label)

   END SUBROUTINE set_restart_info

! **************************************************************************************************
!> \brief dumps on output the information on the information effectively restarted
!> \param restarted_infos ...
!> \param r_file_path ...
!> \param binary_restart_file ...
!> \param output_unit ...
!> \author Teodoro Laino [tlaino] 09.2008 - University of Zurich
! **************************************************************************************************
   SUBROUTINE release_restart_info(restarted_infos, r_file_path, &
                                   binary_restart_file, output_unit)
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: restarted_infos
      CHARACTER(LEN=*), INTENT(IN)                       :: r_file_path, binary_restart_file
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'release_restart_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      IF (output_unit > 0 .AND. ASSOCIATED(restarted_infos)) THEN
         WRITE (output_unit, '(1X,79("*"))')
         WRITE (output_unit, '(1X,"*",T30,A,T80,"*")') " RESTART INFORMATION "
         WRITE (output_unit, '(1X,79("*"))')
         WRITE (output_unit, '(1X,"*",T80,"*")')
         i = 1
         WRITE (output_unit, '(1X,"*",A,T26,A,T80,"*")') "    RESTART FILE NAME:  ", &
            r_file_path(53*(i-1)+1:53*i)
         DO i = 2, CEILING(REAL(LEN_TRIM(r_file_path), KIND=dp)/53.0_dp)
            WRITE (output_unit, '(T1,1X,"*",T26,A,T80,"*")') r_file_path(53*(i-1)+1:53*i)
         END DO
         IF (LEN_TRIM(binary_restart_file) > 0) THEN
            i = 1
            WRITE (output_unit, '(1X,"*",A,T26,A,T80,"*")') "  BINARY RESTART FILE:  ", &
               binary_restart_file(53*(i-1)+1:53*i)
            DO i = 2, CEILING(REAL(LEN_TRIM(binary_restart_file), KIND=dp)/53.0_dp)
               WRITE (output_unit, '(T1,1X,"*",T26,A,T80,"*")') binary_restart_file(53*(i-1)+1:53*i)
            END DO
         END IF
         WRITE (output_unit, '(1X,"*",T80,"*")')
         WRITE (output_unit, '(1X,"*", A,T80,"*")') " RESTARTED QUANTITIES:  "
         DO j = 1, SIZE(restarted_infos)
            DO i = 1, CEILING(REAL(LEN_TRIM(restarted_infos(j)), KIND=dp)/53.0_dp)
               WRITE (output_unit, '(T1,1X,"*",T26,A,T80,"*")') restarted_infos(j) (53*(i-1)+1:53*i)
            END DO
         END DO
         WRITE (output_unit, '(1X,79("*"),/)')
      END IF
      IF (ASSOCIATED(restarted_infos)) THEN
         DEALLOCATE (restarted_infos)
      END IF
   END SUBROUTINE release_restart_info

! **************************************************************************************************
!> \brief Possibly restart thermostats information
!> \param flag ...
!> \param input_file the input file to initialize
!> \param restart_file ...
!> \param path ...
!> \param check ...
!> \author Teodoro Laino [tlaino] 10.2007- University of Zurich
! **************************************************************************************************
   SUBROUTINE restart_thermostat(flag, input_file, restart_file, path, check)
      LOGICAL, INTENT(IN)                                :: flag
      TYPE(section_vals_type), POINTER                   :: input_file, restart_file
      CHARACTER(LEN=*), INTENT(IN)                       :: path
      LOGICAL, INTENT(IN), OPTIONAL                      :: check

      CHARACTER(len=*), PARAMETER :: routineN = 'restart_thermostat', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: input_region, input_type, &
                                                            restart_region, restart_type
      LOGICAL                                            :: check_loc, skip_other_checks
      TYPE(section_vals_type), POINTER                   :: section

      check_loc = check_restart(input_file, restart_file, TRIM(path))
      skip_other_checks = PRESENT(check)
      IF (skip_other_checks) check_loc = check
      IF (flag .AND. check_loc) THEN
         ! Let's check if the thermostat type is different otherwise it does not make any
         ! sense to do any kind of restart
         CALL section_vals_val_get(input_file, TRIM(path)//"%TYPE", i_val=input_type)
         CALL section_vals_val_get(restart_file, TRIM(path)//"%TYPE", i_val=restart_type)

         IF (input_type == do_thermo_same_as_part) THEN
            CALL section_vals_val_get(input_file, "MOTION%MD%THERMOSTAT%TYPE", i_val=input_type)
         END IF

         IF (skip_other_checks) THEN
            input_region = do_region_global
            restart_region = do_region_global
         ELSE
            ! Also the regions must be the same..
            CALL section_vals_val_get(input_file, TRIM(path)//"%REGION", i_val=input_region)
            CALL section_vals_val_get(restart_file, TRIM(path)//"%REGION", i_val=restart_region)
         END IF

         IF ((input_type == restart_type) .AND. (input_region == restart_region)) THEN
            SELECT CASE (input_type)
            CASE (do_thermo_nose)
               section => section_vals_get_subs_vals(restart_file, TRIM(path)//"%NOSE%COORD")
               CALL section_vals_set_subs_vals(input_file, TRIM(path)//"%NOSE%COORD", section)

               section => section_vals_get_subs_vals(restart_file, TRIM(path)//"%NOSE%VELOCITY")
               CALL section_vals_set_subs_vals(input_file, TRIM(path)//"%NOSE%VELOCITY", section)

               section => section_vals_get_subs_vals(restart_file, TRIM(path)//"%NOSE%MASS")
               CALL section_vals_set_subs_vals(input_file, TRIM(path)//"%NOSE%MASS", section)

               section => section_vals_get_subs_vals(restart_file, TRIM(path)//"%NOSE%FORCE")
               CALL section_vals_set_subs_vals(input_file, TRIM(path)//"%NOSE%FORCE", section)
            CASE (do_thermo_csvr)
               section => section_vals_get_subs_vals(restart_file, TRIM(path)//"%CSVR%THERMOSTAT_ENERGY")
               CALL section_vals_set_subs_vals(input_file, TRIM(path)//"%CSVR%THERMOSTAT_ENERGY", section)
               section => section_vals_get_subs_vals(restart_file, TRIM(path)//"%CSVR%RNG_INIT")
               CALL section_vals_set_subs_vals(input_file, TRIM(path)//"%CSVR%RNG_INIT", section)
            CASE (do_thermo_gle)
               section => section_vals_get_subs_vals(restart_file, TRIM(path)//"%GLE%THERMOSTAT_ENERGY")
               CALL section_vals_set_subs_vals(input_file, TRIM(path)//"%GLE%THERMOSTAT_ENERGY", section)
               section => section_vals_get_subs_vals(restart_file, TRIM(path)//"%GLE%RNG_INIT")
               CALL section_vals_set_subs_vals(input_file, TRIM(path)//"%GLE%RNG_INIT", section)
               section => section_vals_get_subs_vals(restart_file, TRIM(path)//"%GLE%S")
               CALL section_vals_set_subs_vals(input_file, TRIM(path)//"%GLE%S", section)
            CASE (do_thermo_al)
               section => section_vals_get_subs_vals(restart_file, TRIM(path)//"%AD_LANGEVIN%CHI")
               CALL section_vals_set_subs_vals(input_file, TRIM(path)//"%AD_LANGEVIN%CHI", section)
               section => section_vals_get_subs_vals(restart_file, TRIM(path)//"%AD_LANGEVIN%MASS")
               CALL section_vals_set_subs_vals(input_file, TRIM(path)//"%AD_LANGEVIN%MASS", section)
            END SELECT
         ELSE
            IF (input_type /= restart_type) &
               CALL cp_warn(cp__l("input_cp2k_check.F",914), &
                            "Requested to restart thermostat: "//TRIM(path)//". The thermostat "// &
                            "specified in the input file and the information present in the restart "// &
                            "file do not match the same type of thermostat! Restarting is not possible! "// &
                            "Thermostat will not be restarted! ")
            IF (input_region /= restart_region) &
               CALL cp_warn(cp__l("input_cp2k_check.F",920), &
                            "Requested to restart thermostat: "//TRIM(path)//". The thermostat "// &
                            "specified in the input file and the information present in the restart "// &
                            "file do not match the same type of REGION! Restarting is not possible! "// &
                            "Thermostat will not be restarted! ")
         END IF
      END IF
   END SUBROUTINE restart_thermostat

! **************************************************************************************************
!> \brief Checks if there are the proper conditions to do a restart
!> \param input_file the input file to initialize
!> \param restart_file ...
!> \param tag_section ...
!> \retval do_restart ...
!> \author teo
! **************************************************************************************************
   FUNCTION check_restart(input_file, restart_file, tag_section) RESULT(do_restart)
      TYPE(section_vals_type), POINTER                   :: input_file, restart_file
      CHARACTER(LEN=*), INTENT(IN)                       :: tag_section
      LOGICAL                                            :: do_restart

      CHARACTER(len=*), PARAMETER :: routineN = 'check_restart', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: explicit1, explicit2
      TYPE(section_vals_type), POINTER                   :: work_section

      CALL timeset(routineN, handle)
      NULLIFY (work_section)
      work_section => section_vals_get_subs_vals(input_file, TRIM(tag_section))
      CALL section_vals_get(work_section, explicit=explicit1)
      work_section => section_vals_get_subs_vals(restart_file, TRIM(tag_section))
      CALL section_vals_get(work_section, explicit=explicit2)

      do_restart = explicit1 .AND. explicit2
      CALL timestop(handle)
   END FUNCTION check_restart

! **************************************************************************************************
!> \brief Removes section used to restart a calculation from an
!>      input file in memory
!> \param input_file the input file to initialize
!> \author teo
! **************************************************************************************************
   SUBROUTINE remove_restart_info(input_file)
      TYPE(section_vals_type), POINTER                   :: input_file

      CHARACTER(len=*), PARAMETER :: routineN = 'remove_restart_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iforce_eval, nforce_eval1
      LOGICAL                                            :: explicit1
      TYPE(section_vals_type), POINTER                   :: md_section, motion_section, section1, &
                                                            section_to_delete, sections1, &
                                                            work_section

      CALL timeset(routineN, handle)

      NULLIFY (work_section)
      section_to_delete => section_vals_get_subs_vals(input_file, "EXT_RESTART")
      CALL section_vals_remove_values(section_to_delete)
      sections1 => section_vals_get_subs_vals(input_file, "FORCE_EVAL")
      CALL section_vals_get(sections1, n_repetition=nforce_eval1)

      DO iforce_eval = 1, nforce_eval1
         section1 => section_vals_get_subs_vals3(sections1, "SUBSYS", i_rep_section=iforce_eval)
         section_to_delete => section_vals_get_subs_vals(section1, "COORD")
         CALL section_vals_remove_values(section_to_delete)
         section_to_delete => section_vals_get_subs_vals(section1, "VELOCITY")
         CALL section_vals_remove_values(section_to_delete)
      END DO

      motion_section => section_vals_get_subs_vals(input_file, "MOTION")
      md_section => section_vals_get_subs_vals(motion_section, "MD")
      CALL section_vals_get(md_section, explicit=explicit1)
      IF (explicit1) THEN
         CALL section_vals_val_unset(md_section, "STEP_START_VAL")
         CALL section_vals_val_unset(md_section, "TIME_START_VAL")
         CALL section_vals_val_unset(md_section, "ECONS_START_VAL")
      END IF
      work_section => section_vals_get_subs_vals(motion_section, "FREE_ENERGY%METADYN")
      CALL section_vals_get(work_section, explicit=explicit1)
      IF (explicit1) THEN
         CALL section_vals_val_unset(motion_section, "FREE_ENERGY%METADYN%STEP_START_VAL")
         CALL section_vals_val_unset(motion_section, "FREE_ENERGY%METADYN%NHILLS_START_VAL")
      END IF
      section_to_delete => section_vals_get_subs_vals(motion_section, "BAND%REPLICA")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "AVERAGES%RESTART_AVERAGES")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "THERMOSTAT%NOSE%COORD")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "THERMOSTAT%NOSE%VELOCITY")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "THERMOSTAT%NOSE%MASS")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "THERMOSTAT%NOSE%FORCE")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "BAROSTAT%MASS")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "BAROSTAT%VELOCITY")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "BAROSTAT%THERMOSTAT%NOSE%COORD")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "BAROSTAT%THERMOSTAT%NOSE%VELOCITY")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "BAROSTAT%THERMOSTAT%NOSE%MASS")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "BAROSTAT%THERMOSTAT%NOSE%FORCE")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "SHELL%THERMOSTAT%NOSE%COORD")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "SHELL%THERMOSTAT%NOSE%VELOCITY")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "SHELL%THERMOSTAT%NOSE%MASS")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(md_section, "SHELL%THERMOSTAT%NOSE%FORCE")
      CALL section_vals_remove_values(section_to_delete)
      ! Constrained/Restrained section
      section_to_delete => section_vals_get_subs_vals(motion_section, "CONSTRAINT%FIX_ATOM_RESTART")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(motion_section, "CONSTRAINT%COLVAR_RESTART")
      CALL section_vals_remove_values(section_to_delete)
      ! Free energies restarts
      section_to_delete => section_vals_get_subs_vals(motion_section, "FREE_ENERGY%METADYN%SPAWNED_HILLS_POS")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(motion_section, "FREE_ENERGY%METADYN%SPAWNED_HILLS_SCALE")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(motion_section, "FREE_ENERGY%METADYN%SPAWNED_HILLS_HEIGHT")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(motion_section, "FREE_ENERGY%METADYN%SPAWNED_HILLS_INVDT")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(motion_section, "FREE_ENERGY%METADYN%EXT_LAGRANGE_SS0")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(motion_section, "FREE_ENERGY%METADYN%EXT_LAGRANGE_VVP")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(motion_section, "FREE_ENERGY%METADYN%EXT_LAGRANGE_SS")
      CALL section_vals_remove_values(section_to_delete)
      section_to_delete => section_vals_get_subs_vals(motion_section, "FREE_ENERGY%METADYN%EXT_LAGRANGE_FS")
      CALL section_vals_remove_values(section_to_delete)
      CALL timestop(handle)
   END SUBROUTINE remove_restart_info

! **************************************************************************************************
!> \brief This subroutine controls the defaults for the restartable quantities..
!> \param r_section ...
!> \author teo - University of Zurich - 09.2007 [tlaino]
! **************************************************************************************************
   SUBROUTINE handle_defaults_restart(r_section)
      TYPE(section_vals_type), POINTER                   :: r_section

      CHARACTER(len=*), PARAMETER :: routineN = 'handle_defaults_restart', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ik, nval
      LOGICAL                                            :: restart_default
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section

      CALL timeset(routineN, handle)
      NULLIFY (keyword, section)
      CALL section_vals_get(r_section, section=section)
      CALL section_vals_val_get(r_section, "RESTART_DEFAULT", l_val=restart_default)
      DO ik = -1, section%n_keywords
         keyword => section%keywords(ik)%keyword
         IF (ASSOCIATED(keyword)) THEN
            IF (keyword%type_of_var == logical_t .AND. keyword%names(1) (1:8) == "RESTART_") THEN
               IF (TRIM(keyword%names(1)) == "RESTART_DEFAULT") CYCLE
               CALL section_vals_val_get(r_section, keyword%names(1), n_rep_val=nval)
               IF (nval == 0) THEN
                  ! User didn't specify any value, use the value of the RESTART_DEFAULT keyword..
                  CALL section_vals_val_set(r_section, keyword%names(1), l_val=restart_default)
               END IF
            END IF
         END IF
      END DO
      CALL timestop(handle)

   END SUBROUTINE handle_defaults_restart

END MODULE input_cp2k_check
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines that build the integrals of the Vxc potential calculated
!>      for the atomic code
! **************************************************************************************************
MODULE atom_output
   USE atom_types,                      ONLY: &
        atom_basis_type, atom_gthpot_type, atom_potential_type, atom_state, atom_type, cgto_basis, &
        ecp_pseudo, gth_pseudo, gto_basis, no_pseudo, num_basis, sto_basis, upf_pseudo
   USE atom_utils,                      ONLY: get_maxl_occ,&
                                              get_maxn_occ,&
                                              get_rho0
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE input_constants,                 ONLY: &
        barrier_conf, do_dkh0_atom, do_dkh1_atom, do_dkh2_atom, do_dkh3_atom, do_dkh4_atom, &
        do_dkh5_atom, do_nonrel_atom, do_rhf_atom, do_rks_atom, do_rohf_atom, do_sczoramp_atom, &
        do_uhf_atom, do_uks_atom, do_zoramp_atom, poly_conf, xc_none
   USE input_cp2k_check,                ONLY: xc_functionals_expand
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_get_subs_vals2,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: pi
   USE periodic_table,                  ONLY: ptable
   USE physcon,                         ONLY: evolt
   USE xc_derivatives,                  ONLY: xc_functional_get_info
   USE xc_libxc,                        ONLY: libxc_version_info
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_output'

   PUBLIC :: atom_print_state, atom_print_energies, atom_print_iteration, &
             atom_print_basis, atom_print_method, atom_print_info, atom_print_potential, &
             atom_print_basis_file, atom_write_pseudo_param, atom_print_orbitals, &
             atom_print_zmp_iteration

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param zval ...
!> \param info ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_print_info(zval, info, iw)
      INTEGER, INTENT(IN)                                :: zval
      CHARACTER(len=*), INTENT(IN)                       :: info
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_print_info', &
         routineP = moduleN//':'//routineN

      WRITE (iw, '(/," ",A,T40,A," [",A,"]",T62,"Atomic number:",T78,I3,/)') &
         ADJUSTL(TRIM(info)), TRIM(ptable(zval)%name), TRIM(ptable(zval)%symbol), zval

   END SUBROUTINE atom_print_info
! **************************************************************************************************
!> \brief ...
!> \param state ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_print_state(state, iw)
      TYPE(atom_state)                                   :: state
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_print_state', &
         routineP = moduleN//':'//routineN
      CHARACTER(LEN=1), DIMENSION(0:3), PARAMETER        :: label = (/"S", "P", "D", "F"/)

      INTEGER                                            :: j, l, mc, mlc, mlo, mm(0:3), mo

      WRITE (iw, '(/,T2,A)') "Electronic structure"
      WRITE (iw, '(T5,A,T71,F10.2)') "Total number of core electrons", SUM(state%core)
      WRITE (iw, '(T5,A,T71,F10.2)') "Total number of valence electrons", SUM(state%occ)
      WRITE (iw, '(T5,A,T71,F10.2)') "Total number of electrons", SUM(state%occ+state%core)
      SELECT CASE (state%multiplicity)
      CASE (-1)
         WRITE (iw, '(T5,A,T68,A)') "Multiplicity", "not specified"
      CASE (-2)
         WRITE (iw, '(T5,A,T72,A)') "Multiplicity", "high spin"
      CASE (-3)
         WRITE (iw, '(T5,A,T73,A)') "Multiplicity", "low spin"
      CASE (1)
         WRITE (iw, '(T5,A,T74,A)') "Multiplicity", "singlet"
      CASE (2)
         WRITE (iw, '(T5,A,T74,A)') "Multiplicity", "doublet"
      CASE (3)
         WRITE (iw, '(T5,A,T74,A)') "Multiplicity", "triplet"
      CASE (4)
         WRITE (iw, '(T5,A,T74,A)') "Multiplicity", "quartet"
      CASE (5)
         WRITE (iw, '(T5,A,T74,A)') "Multiplicity", "quintet"
      CASE (6)
         WRITE (iw, '(T5,A,T75,A)') "Multiplicity", "sextet"
      CASE (7)
         WRITE (iw, '(T5,A,T75,A)') "Multiplicity", "septet"
      CASE DEFAULT
      END SELECT

      mlo = get_maxl_occ(state%occ)
      mlc = get_maxl_occ(state%core)
      mm = get_maxn_occ(state%core)

      IF (state%multiplicity == -1) THEN
         DO l = 0, MAX(mlo, mlc)
            mo = state%maxn_occ(l)
            IF (SUM(state%core(l, :)) == 0) THEN
               WRITE (iw, '(A5,T10,10F6.2)') label(l), (state%occ(l, j), j=1, mo)
            ELSE
               mc = mm(l)
               IF(.NOT.(SUM(state%occ(l, 1:mc)) == 0))CALL cp__a("atom_output.F",124)
               WRITE (iw, ADVANCE="no", FMT='(A5,T9,A1,10F6.2)') label(l), "[", (state%core(l, j), j=1, mc)
               WRITE (iw, FMT='(A1,F5.2,10F6.2)') "]", (state%occ(l, j), j=mc+1, mc+mo)
            END IF
         END DO
      ELSE
         WRITE (iw, '(T5,A)') "Alpha Electrons"
         DO l = 0, MAX(mlo, mlc)
            mo = state%maxn_occ(l)
            IF (SUM(state%core(l, :)) == 0) THEN
               WRITE (iw, '(A5,T10,10F6.2)') label(l), (state%occa(l, j), j=1, mo)
            ELSE
               mc = mm(l)
               WRITE (iw, ADVANCE="no", FMT='(A5,T9,A1,10F6.2)') label(l), "[", (0.5_dp*state%core(l, j), j=1, mc)
               WRITE (iw, FMT='(A1,F5.2,10F6.2)') "]", (state%occa(l, j), j=1, mo)
            END IF
         END DO
         WRITE (iw, '(T5,A)') "Beta Electrons"
         DO l = 0, MAX(mlo, mlc)
            mo = state%maxn_occ(l)
            IF (SUM(state%core(l, :)) == 0) THEN
               WRITE (iw, '(A5,T10,10F6.2)') label(l), (state%occb(l, j), j=1, mo)
            ELSE
               mc = mm(l)
               WRITE (iw, ADVANCE="no", FMT='(A5,T9,A1,10F6.2)') label(l), "[", (0.5_dp*state%core(l, j), j=1, mc)
               WRITE (iw, FMT='(A1,F5.2,10F6.2)') "]", (state%occb(l, j), j=1, mo)
            END IF
         END DO
      END IF
      WRITE (iw, *)

   END SUBROUTINE atom_print_state
! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_print_energies(atom, iw)
      TYPE(atom_type)                                    :: atom
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_print_energies', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, l, n
      REAL(KIND=dp)                                      :: drho

      WRITE (iw, '(/,A,T36,A,T61,F20.12)') " Energy components [Hartree]", &
         "    Total Energy ::", atom%energy%etot
      WRITE (iw, '(T36,A,T61,F20.12)') "     Band Energy ::", atom%energy%eband
      WRITE (iw, '(T36,A,T61,F20.12)') "  Kinetic Energy ::", atom%energy%ekin
      WRITE (iw, '(T36,A,T61,F20.12)') "Potential Energy ::", atom%energy%epot
      IF (atom%energy%ekin /= 0.0_dp) THEN
         WRITE (iw, '(T36,A,T61,F20.12)') "   Virial (-V/T) ::", -atom%energy%epot/atom%energy%ekin
      END IF
      WRITE (iw, '(T36,A,T61,F20.12)') "     Core Energy ::", atom%energy%ecore
      IF (atom%energy%exc /= 0._dp) &
         WRITE (iw, '(T36,A,T61,F20.12)') "       XC Energy ::", atom%energy%exc
      WRITE (iw, '(T36,A,T61,F20.12)') "  Coulomb Energy ::", atom%energy%ecoulomb
      IF (atom%energy%eexchange /= 0._dp) &
         WRITE (iw, '(T34,A,T61,F20.12)') "HF Exchange Energy ::", atom%energy%eexchange
      IF (atom%potential%ppot_type /= NO_PSEUDO) THEN
         WRITE (iw, '(T20,A,T61,F20.12)') "    Total Pseudopotential Energy ::", atom%energy%epseudo
         WRITE (iw, '(T20,A,T61,F20.12)') "    Local Pseudopotential Energy ::", atom%energy%eploc
         IF (atom%energy%elsd /= 0._dp) &
            WRITE (iw, '(T20,A,T61,F20.12)') "     Local Spin-potential Energy ::", atom%energy%elsd
         WRITE (iw, '(T20,A,T61,F20.12)') " Nonlocal Pseudopotential Energy ::", atom%energy%epnl
      END IF
      IF (atom%potential%confinement) THEN
         WRITE (iw, '(T36,A,T61,F20.12)') "     Confinement ::", atom%energy%econfinement
      END IF

      IF (atom%state%multiplicity == -1) THEN
         WRITE (iw, '(/,A,T20,A,T30,A,T36,A,T49,A,T71,A,/)') " Orbital energies", &
            "State", "L", "Occupation", "Energy[a.u.]", "Energy[eV]"
         DO l = 0, atom%state%maxl_calc
            n = atom%state%maxn_calc(l)
            DO i = 1, n
               WRITE (iw, '(T23,I2,T30,I1,T36,F10.3,T46,F15.6,T66,F15.6)') &
                  i, l, atom%state%occupation(l, i), atom%orbitals%ener(i, l), atom%orbitals%ener(i, l)*evolt
            END DO
            IF (n > 0) WRITE (iw, *)
         END DO
      ELSE
         WRITE (iw, '(/,A,T20,A,T30,A,T36,A,T42,A,T55,A,T71,A,/)') " Orbital energies", &
            "State", "Spin", "L", "Occupation", "Energy[a.u.]", "Energy[eV]"
         DO l = 0, atom%state%maxl_calc
            n = atom%state%maxn_calc(l)
            DO i = 1, n
               WRITE (iw, '(T23,I2,T29,A,T36,I1,T42,F10.3,T52,F15.6,T68,F13.6)') &
                  i, "alpha", l, atom%state%occa(l, i), atom%orbitals%enera(i, l), atom%orbitals%enera(i, l)*evolt
            END DO
            DO i = 1, n
               WRITE (iw, '(T23,I2,T29,A,T36,I1,T42,F10.3,T52,F15.6,T68,F13.6)') &
                  i, " beta", l, atom%state%occb(l, i), atom%orbitals%enerb(i, l), atom%orbitals%enerb(i, l)*evolt
            END DO
            IF (n > 0) WRITE (iw, *)
         END DO
      END IF

      CALL get_rho0(atom, drho)
      WRITE (iw, '(/,A,T66,F15.6)') " Total Electron Density at R=0: ", drho

   END SUBROUTINE atom_print_energies

! **************************************************************************************************
!> \brief Printing of the atomic iterations when ZMP is active
!>
!> \param iter ...
!> \param deps ...
!> \param atom ...
!> \param iw ...
!> \author D. Varsano [daniele.varsano@nano.cnr.it]
! **************************************************************************************************
   SUBROUTINE atom_print_zmp_iteration(iter, deps, atom, iw)
      INTEGER, INTENT(IN)                                :: iter
      REAL(dp), INTENT(IN)                               :: deps
      TYPE(atom_type), INTENT(IN)                        :: atom
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_print_zmp_iteration', &
         routineP = moduleN//':'//routineN

      IF (iter == 1) THEN
         WRITE (iw, '(/," ",79("*"),/,T33,"Integral",T48,"Integral",/,T3,A,T16,A,T33,A,T46,A,T69,A/," ",79("*"))') &
            "Iteration", "Convergence", "rho diff.", "rho*v_xc[au]", "Energy[au]"
      END IF
      WRITE (iw, '(T3,I9,T15,G13.6,T30,G13.6,T46,G13.6,T61,F20.12)') iter, deps, atom%rho_diff_integral, &
         atom%energy%exc, atom%energy%etot

   END SUBROUTINE atom_print_zmp_iteration

! **************************************************************************************************
!> \brief ...
!> \param iter ...
!> \param deps ...
!> \param etot ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_print_iteration(iter, deps, etot, iw)
      INTEGER, INTENT(IN)                                :: iter
      REAL(dp), INTENT(IN)                               :: deps, etot
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_print_iteration', &
         routineP = moduleN//':'//routineN

      IF (iter == 1) THEN
         WRITE (iw, '(/," ",79("*"),/,T19,A,T38,A,T70,A,/," ",79("*"))') &
            "Iteration", "Convergence", "Energy [au]"
      END IF
      WRITE (iw, '(T20,i8,T34,G14.6,T61,F20.12)') iter, deps, etot

   END SUBROUTINE atom_print_iteration
! **************************************************************************************************
!> \brief ...
!> \param atom_basis ...
!> \param iw ...
!> \param title ...
! **************************************************************************************************
   SUBROUTINE atom_print_basis(atom_basis, iw, title)
      TYPE(atom_basis_type)                              :: atom_basis
      INTEGER, INTENT(IN)                                :: iw
      CHARACTER(len=*)                                   :: title

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_print_basis', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, l

      WRITE (iw, '(/,A)') TRIM(title)
      SELECT CASE (atom_basis%basis_type)
      CASE (GTO_BASIS)
         IF (atom_basis%geometrical) THEN
            WRITE (iw, '(/," ",21("*"),A,22("*"))') " Geometrical Gaussian Type Orbitals "
            WRITE (iw, '(A,F15.8,T41,A,F15.8)') " Initial exponent: ", atom_basis%aval, &
               " Proportionality factor: ", atom_basis%cval
         ELSE
            WRITE (iw, '(/," ",21("*"),A,21("*"))') " Uncontracted Gaussian Type Orbitals "
         END IF
         IF (atom_basis%nbas(0) > 0) THEN
            WRITE (iw, '(/,T2,A,(T30,I5,T51,F30.8))') &
               "s Exponents: ", (i, atom_basis%am(i, 0), i=1, atom_basis%nbas(0))
         END IF
         IF (atom_basis%nbas(1) > 0) THEN
            WRITE (iw, '(/,T2,A,(T30,I5,T51,F30.8))') &
               "p Exponents: ", (i, atom_basis%am(i, 1), i=1, atom_basis%nbas(1))
         END IF
         IF (atom_basis%nbas(2) > 0) THEN
            WRITE (iw, '(/,T2,A,(T30,I5,T51,F30.8))') &
               "d Exponents: ", (i, atom_basis%am(i, 2), i=1, atom_basis%nbas(2))
         END IF
         IF (atom_basis%nbas(3) > 0) THEN
            WRITE (iw, '(/,T2,A,(T30,I5,T51,F30.8))') &
               "f Exponents: ", (i, atom_basis%am(i, 3), i=1, atom_basis%nbas(3))
         END IF
         WRITE (iw, '(" ",79("*"))')
      CASE (CGTO_BASIS)
         WRITE (iw, '(/," ",22("*"),A,22("*"))') " Contracted Gaussian Type Orbitals "
         DO l = 0, 3
            IF (atom_basis%nbas(l) > 0) THEN
               IF (l == 0) WRITE (iw, '(A)') " s Functions"
               IF (l == 1) WRITE (iw, '(A)') " p Functions"
               IF (l == 2) WRITE (iw, '(A)') " d Functions"
               IF (l == 3) WRITE (iw, '(A)') " f Functions"
               DO i = 1, atom_basis%nprim(l)
                  WRITE (iw, '(F15.6,5(T21,6F10.6,/))') &
                     atom_basis%am(i, l), (atom_basis%cm(i, j, l), j=1, atom_basis%nbas(l))
               END DO
            END IF
         END DO
         WRITE (iw, '(" ",79("*"))')
      CASE (STO_BASIS)
         WRITE (iw, '(/," ",28("*"),A,29("*"))') " Slater Type Orbitals "
         DO i = 1, atom_basis%nbas(0)
            WRITE (iw, '(T10,I1,A,T40,F25.12)') atom_basis%ns(i, 0), "S Exponent :", atom_basis%as(i, 0)
         END DO
         DO i = 1, atom_basis%nbas(1)
            WRITE (iw, '(T10,I1,A,T40,F25.12)') atom_basis%ns(i, 1), "P Exponent :", atom_basis%as(i, 1)
         END DO
         DO i = 1, atom_basis%nbas(2)
            WRITE (iw, '(T10,I1,A,T40,F25.12)') atom_basis%ns(i, 2), "D Exponent :", atom_basis%as(i, 2)
         END DO
         DO i = 1, atom_basis%nbas(3)
            WRITE (iw, '(T10,I1,A,T40,F25.12)') atom_basis%ns(i, 3), "F Exponent :", atom_basis%as(i, 3)
         END DO
         WRITE (iw, '(" ",79("*"))')
      CASE (NUM_BASIS)
         CALL cp__b("atom_output.F",352,"")
      CASE DEFAULT
         CALL cp__b("atom_output.F",354,"")
      END SELECT

   END SUBROUTINE atom_print_basis
! **************************************************************************************************
!> \brief ...
!> \param atom_basis ...
! **************************************************************************************************
   SUBROUTINE atom_print_basis_file(atom_basis)
      TYPE(atom_basis_type)                              :: atom_basis

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_print_basis_file', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iw

      CALL open_file(file_name="/mnt/sdcard/cp2k/work/OPT_BASIS", file_status="UNKNOWN", file_action="WRITE", unit_number=iw)
      SELECT CASE (atom_basis%basis_type)
      CASE (GTO_BASIS)
         IF (atom_basis%geometrical) THEN
            WRITE (iw, '(/," ",21("*"),A,22("*"))') " Geometrical Gaussian Type Orbitals "
            WRITE (iw, '(A,F15.8,T41,A,F15.8)') " Initial exponent: ", atom_basis%aval, &
               " Proportionality factor: ", atom_basis%cval
         ELSE
            WRITE (iw, '(A)') "      BASIS_TYPE GAUSSIAN"
         END IF
         IF (atom_basis%nbas(0) > 0) THEN
            WRITE (iw, '(T7,A,20F15.8)') &
               "S_EXPONENTS ", (atom_basis%am(i, 0), i=1, atom_basis%nbas(0))
         END IF
         IF (atom_basis%nbas(1) > 0) THEN
            WRITE (iw, '(T7,A,20F15.8)') &
               "P_EXPONENTS ", (atom_basis%am(i, 1), i=1, atom_basis%nbas(1))
         END IF
         IF (atom_basis%nbas(2) > 0) THEN
            WRITE (iw, '(T7,A,20F15.8)') &
               "D_EXPONENTS ", (atom_basis%am(i, 2), i=1, atom_basis%nbas(2))
         END IF
         IF (atom_basis%nbas(3) > 0) THEN
            WRITE (iw, '(T7,A,20F15.8)') &
               "F_EXPONENTS ", (atom_basis%am(i, 3), i=1, atom_basis%nbas(3))
         END IF
      CASE (CGTO_BASIS)
         CALL cp__b("atom_output.F",397,"")
      CASE (STO_BASIS)
         WRITE (iw, '(A)') "      BASIS_TYPE SLATER"
         IF (atom_basis%nbas(0) > 0) THEN
            WRITE (iw, '(T7,A,20F15.8)') &
               "S_EXPONENTS ", (atom_basis%as(i, 0), i=1, atom_basis%nbas(0))
            WRITE (iw, '(T7,A,20I5)') &
               "S_QUANTUM_NUMBERS ", (atom_basis%ns(i, 0), i=1, atom_basis%nbas(0))
         END IF
         IF (atom_basis%nbas(1) > 0) THEN
            WRITE (iw, '(T7,A,20F15.8)') &
               "P_EXPONENTS ", (atom_basis%as(i, 1), i=1, atom_basis%nbas(1))
            WRITE (iw, '(T7,A,20I5)') &
               "P_QUANTUM_NUMBERS ", (atom_basis%ns(i, 1), i=1, atom_basis%nbas(1))
         END IF
         IF (atom_basis%nbas(2) > 0) THEN
            WRITE (iw, '(T7,A,20F15.8)') &
               "D_EXPONENTS ", (atom_basis%as(i, 2), i=1, atom_basis%nbas(2))
            WRITE (iw, '(T7,A,20I5)') &
               "D_QUANTUM_NUMBERS ", (atom_basis%ns(i, 2), i=1, atom_basis%nbas(2))
         END IF
         IF (atom_basis%nbas(3) > 0) THEN
            WRITE (iw, '(T7,A,20F15.8)') &
               "F_EXPONENTS ", (atom_basis%as(i, 3), i=1, atom_basis%nbas(3))
            WRITE (iw, '(T7,A,20I5)') &
               "F_QUANTUM_NUMBERS ", (atom_basis%ns(i, 3), i=1, atom_basis%nbas(3))
         END IF
      CASE (NUM_BASIS)
         CALL cp__b("atom_output.F",425,"")
      CASE DEFAULT
         CALL cp__b("atom_output.F",427,"")
      END SELECT
      CALL close_file(unit_number=iw)

   END SUBROUTINE atom_print_basis_file
! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_print_method(atom, iw)
      TYPE(atom_type)                                    :: atom
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_print_method', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=10*default_string_length)            :: reference
      CHARACTER(len=160)                                 :: shortform
      CHARACTER(len=20)                                  :: tmpStr
      CHARACTER(len=80), DIMENSION(:), POINTER           :: func_name
      INTEGER                                            :: ifun, ifunc_name, il, meth, myfun, reltyp
      LOGICAL                                            :: lsd
      TYPE(section_vals_type), POINTER                   :: xc_fun, xc_fun_section, xc_section

      NULLIFY (xc_fun, xc_fun_section, xc_section)

      meth = atom%method_type

      xc_section => atom%xc_section
      xc_fun_section => section_vals_get_subs_vals(xc_section, "XC_FUNCTIONAL")
      SELECT CASE (meth)
      CASE DEFAULT
         CALL cp__b("atom_output.F",460,"")
      CASE (do_rks_atom)
         CALL section_vals_val_get(xc_fun_section, "_SECTION_PARAMETERS_", i_val=myfun)
      CASE (do_uks_atom)
         CALL section_vals_val_get(xc_fun_section, "_SECTION_PARAMETERS_", i_val=myfun)
      CASE (do_rhf_atom)
         myfun = xc_none
      CASE (do_uhf_atom)
         myfun = xc_none
      CASE (do_rohf_atom)
         myfun = xc_none
      END SELECT

      SELECT CASE (meth)
      CASE DEFAULT
         CALL cp__b("atom_output.F",475,"")
      CASE (do_rks_atom)
         IF (iw > 0) WRITE (iw, fmt="(/,' METHOD    | Restricted Kohn-Sham Calculation')")
      CASE (do_uks_atom)
         IF (iw > 0) WRITE (iw, fmt="(/,' METHOD    | Unrestricted Kohn-Sham Calculation')")
      CASE (do_rhf_atom)
         IF (iw > 0) WRITE (iw, fmt="(/,' METHOD    | Restricted Hartree-Fock Calculation')")
      CASE (do_uhf_atom)
         IF (iw > 0) WRITE (iw, fmt="(/,' METHOD    | Unrestricted Hartree-Fock Calculation')")
      CASE (do_rohf_atom)
         IF (iw > 0) WRITE (iw, fmt="(/,' METHOD    | Restricted Open-Shell Kohn-Sham Calculation')")
      END SELECT

      ! zmp
      IF (atom%do_zmp) THEN
         IF (iw > 0) WRITE (iw, fmt="(' ZMP       | Method on atomic radial density')")
         IF (iw > 0) WRITE (iw, fmt="(' ZMP       | Lambda : ',F5.1)") atom%lambda
         IF (iw > 0) WRITE (iw, fmt="(' ZMP       | Reading external density : ',A20)") atom%ext_file
         IF (atom%dm) THEN
            IF (iw > 0) WRITE (iw, fmt="(' ZMP       | The file is in the form of a density matrix')")
         ELSE
            IF (iw > 0) WRITE (iw, fmt="(' ZMP       | The file is in the form of a linear density')")
         END IF
         IF (atom%doread) THEN
            IF (iw > 0) WRITE (iw, fmt="(' ZMP       | Restarting calculation from ',A20,' file if present')") atom%zmp_restart_file
         END IF
      ELSE IF (atom%read_vxc) THEN
         IF (iw > 0) WRITE (iw, fmt="(' ZMP       | Calculating density from external V_xc')")
         IF (iw > 0) WRITE (iw, fmt="(' ZMP       | Reading external v_xc file : ',A20)") atom%ext_vxc_file
      END IF

      IF (atom%pp_calc) THEN
         IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Nonrelativistic Calculation')")
      ELSE
         reltyp = atom%relativistic

         SELECT CASE (reltyp)
         CASE DEFAULT
            CALL cp__b("atom_output.F",513,"")
         CASE (do_nonrel_atom)
            IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Nonrelativistic Calculation')")
         CASE (do_zoramp_atom)
            IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Relativistic Calculation using ZORA(MP)')")
         CASE (do_sczoramp_atom)
            IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Relativistic Calculation using scaled ZORA(MP)')")
         CASE (do_dkh0_atom)
            IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Relativistic Calculation using Douglas-Kroll 0th order')")
            IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Relativistic Calculation using kietic energy scaling')")
         CASE (do_dkh1_atom)
            IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Relativistic Calculation using Douglas-Kroll 1st order')")
            IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Relativistic Calculation using Foldy-Wouthuysen transformation')")
         CASE (do_dkh2_atom)
            IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Relativistic Calculation using Douglas-Kroll 2nd order')")
         CASE (do_dkh3_atom)
            IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Relativistic Calculation using Douglas-Kroll 3rd order')")
         CASE (do_dkh4_atom)
            IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Relativistic Calculation using Douglas-Kroll 4th order')")
         CASE (do_dkh5_atom)
            IF (iw > 0) WRITE (iw, fmt="(' METHOD    | Relativistic Calculation using Douglas-Kroll 5th order')")
         END SELECT
      END IF

      lsd = (meth == do_uks_atom)

      IF (myfun /= xc_none) THEN
         CALL section_vals_val_get(xc_section, "FUNCTIONAL_ROUTINE", c_val=tmpStr)
         IF (iw > 0) WRITE (iw, fmt="(' FUNCTIONAL| ROUTINE=',a)") TRIM(tmpStr)
         CALL xc_functionals_expand(xc_fun_section, xc_section)
         IF (iw > 0) THEN
            ifun = 0
            DO
               ifun = ifun+1
               xc_fun => section_vals_get_subs_vals2(xc_fun_section, i_section=ifun)
               IF (.NOT. ASSOCIATED(xc_fun)) EXIT
               CALL xc_functional_get_info(xc_fun, lsd=lsd, reference=reference, &
                                           shortform=shortform, ifunc_name=1)
               IF (TRIM(xc_fun%section%name) == "LIBXC") THEN
                  CALL libxc_version_info(tmpStr)
                  WRITE (iw, fmt="(A,A,A)") ' FUNCTIONAL| LIBXC Vers. ', TRIM(tmpStr(1:5)), &
                     ' (Marques, Oliveira, Burnus, CPC 183, 2272 (2012))'
                  WRITE (iw, fmt="(' FUNCTIONAL| ',a,':')") TRIM(shortform)
                  DO il = 1, LEN_TRIM(reference), 67
                     WRITE (iw, fmt="(' FUNCTIONAL| ',a67)") reference(il:)
                  END DO
                  CALL section_vals_val_get(xc_fun, "functional", c_vals=func_name)
                  DO ifunc_name = 2, SIZE(func_name)
                     CALL xc_functional_get_info(xc_fun, lsd=lsd, reference=reference, &
                                                 shortform=shortform, ifunc_name=ifunc_name)
                     WRITE (iw, fmt="(A,A,A)") ' FUNCTIONAL| LIBXC Vers. ', TRIM(tmpStr(1:5)), &
                        ' (Marques, Oliveira, Burnus, CPC 183, 2272 (2012))'
                     WRITE (iw, fmt="(' FUNCTIONAL| ',a,':')") TRIM(shortform)
                     DO il = 1, LEN_TRIM(reference), 67
                        WRITE (iw, fmt="(' FUNCTIONAL| ',a67)") reference(il:)
                     END DO
                  END DO
               ELSE
                  WRITE (iw, fmt="(' FUNCTIONAL| ',a,':')") &
                     TRIM(xc_fun%section%name)
                  DO il = 1, LEN_TRIM(reference), 67
                     WRITE (iw, fmt="(' FUNCTIONAL| ',a67)") reference(il:)
                  END DO
               END IF
            END DO
         END IF
      ELSE
         IF (iw > 0) WRITE (iw, fmt="(' FUNCTIONAL| NO EXCHANGE-CORRELATION FUNCTIONAL USED.')")
      END IF

   END SUBROUTINE atom_print_method
! **************************************************************************************************
!> \brief ...
!> \param potential ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_print_potential(potential, iw)
      TYPE(atom_potential_type)                          :: potential
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_print_potential', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=60)                                  :: pline
      INTEGER                                            :: i, j, k, l

      SELECT CASE (potential%ppot_type)
      CASE (no_pseudo)
         WRITE (iw, '(/," ",28("*"),A,27("*"))') " All Electron Potential "
      CASE (gth_pseudo)
         WRITE (iw, '(/," ",29("*"),A,29("*"))') " GTH Pseudopotential "
         WRITE (iw, '(T10,A,T76,F5.1)') " Core Charge ", potential%gth_pot%zion
         WRITE (iw, '(T10,A,T66,F15.6)') " Rc ", potential%gth_pot%rc
         WRITE (pline, '(5F12.6)') (potential%gth_pot%cl(i), i=1, potential%gth_pot%ncl)
         WRITE (iw, '(T10,A,T21,A60)') " C1 C2 ... ", ADJUSTR(pline)
         IF (potential%gth_pot%lpotextended) THEN
            DO k = 1, potential%gth_pot%nexp_lpot
               WRITE (iw, '(T10,A,F10.6,T38,A,4F10.6)') " LPot: rc=", potential%gth_pot%alpha_lpot(k), &
                  "CX=", (potential%gth_pot%cval_lpot(i, k), i=1, potential%gth_pot%nct_lpot(k))
            END DO
         END IF
         IF (potential%gth_pot%nlcc) THEN
            DO k = 1, potential%gth_pot%nexp_nlcc
               WRITE (iw, '(T10,A,F10.6,T38,A,4F10.6)') " LSDPot: rc=", potential%gth_pot%alpha_nlcc(k), &
                  "CX=", (potential%gth_pot%cval_nlcc(i, k)*4.0_dp*pi, i=1, potential%gth_pot%nct_nlcc(k))
            END DO
         END IF
         IF (potential%gth_pot%lsdpot) THEN
            DO k = 1, potential%gth_pot%nexp_lsd
               WRITE (iw, '(T10,A,F10.6,T38,A,4F10.6)') " LSDPot: rc=", potential%gth_pot%alpha_lsd(k), &
                  "CX=", (potential%gth_pot%cval_lsd(i, k), i=1, potential%gth_pot%nct_lsd(k))
            END DO
         END IF
         DO l = 0, 3
            IF (potential%gth_pot%nl(l) > 0) THEN
               WRITE (iw, '(T10,A,T76,I5)') " Angular momentum ", l
               WRITE (iw, '(T10,A,T66,F15.6)') " Rcnl ", potential%gth_pot%rcnl(l)
               WRITE (iw, '(T10,A,T76,I5)') " Nl ", potential%gth_pot%nl(l)
               WRITE (pline, '(5F12.6)') (potential%gth_pot%hnl(1, j, l), j=1, potential%gth_pot%nl(l))
               WRITE (iw, '(T10,A,T21,A60)') " Hnl ", ADJUSTR(pline)
               DO i = 2, potential%gth_pot%nl(l)
                  WRITE (pline, '(T21,5F12.6)') (potential%gth_pot%hnl(i, j, l), j=i, potential%gth_pot%nl(l))
                  WRITE (iw, '(T21,A60)') ADJUSTR(pline)
               END DO
            END IF
         END DO
      CASE (upf_pseudo)
         WRITE (iw, '(/," ",29("*"),A,29("*"))') " UPF Pseudopotential "
         DO k = 1, potential%upf_pot%maxinfo
            WRITE (iw, '(A80)') potential%upf_pot%info(k)
         END DO
      CASE (ecp_pseudo)
         WRITE (iw, '(/," ",26("*"),A,27("*"))') " Effective Core Potential "
         WRITE (iw, '(T10,A,T76,F5.1)') " Core Charge ", potential%ecp_pot%zion
         DO k = 1, potential%ecp_pot%nloc
            IF (k == 1) THEN
               WRITE (iw, '(T10,A,T40,I3,T49,2F16.8)') " Local Potential ", potential%ecp_pot%nrloc(k), &
                  potential%ecp_pot%bloc(k), potential%ecp_pot%aloc(k)
            ELSE
               WRITE (iw, '(T40,I3,T49,2F16.8)') potential%ecp_pot%nrloc(k), &
                  potential%ecp_pot%bloc(k), potential%ecp_pot%aloc(k)
            END IF
         END DO
         DO l = 0, potential%ecp_pot%lmax
            WRITE (iw, '(T10,A,I3)') " ECP l-value ", l
            DO k = 1, potential%ecp_pot%npot(l)
               WRITE (iw, '(T40,I3,T49,2F16.8)') potential%ecp_pot%nrpot(k, l), &
                  potential%ecp_pot%bpot(k, l), potential%ecp_pot%apot(k, l)
            END DO
         END DO
      CASE DEFAULT
         CALL cp__b("atom_output.F",664,"")
      END SELECT
      IF (potential%confinement) THEN
         IF (potential%conf_type == poly_conf) THEN
            WRITE (iw, '(/,T10,A,T51,F12.6," * (R /",F6.2,")**",F6.2)') &
               " Confinement Potential ", potential%acon, potential%rcon, potential%scon
         ELSE IF (potential%conf_type == barrier_conf) THEN
            WRITE (iw, '(/,T10,A)') " Confinement Potential s*F[(r-ron)/w] "
            WRITE (iw, '(T57,A,F12.6,A)') "s     =", potential%acon, "   Ha"
            WRITE (iw, '(T57,A,F12.6,A)') "w     =", potential%rcon, " Bohr"
            WRITE (iw, '(T57,A,F12.6,A)') "ron   =", potential%scon, " Bohr"
         ELSE
            CALL cp__b("atom_output.F",676,"")
         END IF
      ELSE
         WRITE (iw, '(/,T10,A)') " No Confinement Potential is applied "
      END IF
      WRITE (iw, '(" ",79("*"))')

   END SUBROUTINE atom_print_potential

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param gthpot ...
!> \param iunit ...
! **************************************************************************************************
   SUBROUTINE atom_write_pseudo_param(gthpot, iunit)
      TYPE(atom_gthpot_type), INTENT(INOUT)              :: gthpot
      INTEGER, INTENT(IN), OPTIONAL                      :: iunit

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_write_pseudo_param', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iw, j, k, n

      IF (PRESENT(iunit)) THEN
         iw = iunit
      ELSE
         CALL open_file(file_name="/mnt/sdcard/cp2k/work/GTH-PARAMETER", file_status="UNKNOWN", file_action="WRITE", unit_number=iw)
      END IF
      WRITE (iw, '(A2,A)') gthpot%symbol, ADJUSTL(TRIM(gthpot%pname))
      WRITE (iw, '(4I5)') gthpot%econf
      WRITE (iw, '(F18.12,I8,5F18.12)') gthpot%rc, gthpot%ncl, (gthpot%cl(i), i=1, gthpot%ncl)
      IF (gthpot%lpotextended) THEN
         WRITE (iw, '(A,I5)') "  LPOT", gthpot%nexp_lpot
         DO i = 1, gthpot%nexp_lpot
            WRITE (iw, '(F18.12,I8,5F18.12)') gthpot%alpha_lpot(i), gthpot%nct_lpot(i), &
               (gthpot%cval_lpot(j, i), j=1, gthpot%nct_lpot(i))
         END DO
      END IF
      IF (gthpot%lsdpot) THEN
         WRITE (iw, '(A,I5)') "  LSD ", gthpot%nexp_lsd
         DO i = 1, gthpot%nexp_lsd
            WRITE (iw, '(F18.12,I8,5F18.12)') gthpot%alpha_lsd(i), gthpot%nct_lsd(i), &
               (gthpot%cval_lsd(j, i), j=1, gthpot%nct_lsd(i))
         END DO
      END IF
      IF (gthpot%nlcc) THEN
         WRITE (iw, '(A,I5)') " NLCC ", gthpot%nexp_nlcc
         DO i = 1, gthpot%nexp_nlcc
            WRITE (iw, '(F18.12,I8,5F18.12)') gthpot%alpha_nlcc(i), gthpot%nct_nlcc(i), &
               (gthpot%cval_nlcc(j, i)*4.0_dp*pi, j=1, gthpot%nct_nlcc(i))
         END DO
      END IF
      n = 0
      DO i = 3, 0, -1
         IF (gthpot%nl(i) > 0) THEN
            n = i+1
            EXIT
         END IF
      END DO
      WRITE (iw, '(I8)') n
      DO i = 0, n-1
         WRITE (iw, '(F18.12,I8,5F18.12)') gthpot%rcnl(i), gthpot%nl(i), (gthpot%hnl(1, k, i), k=1, gthpot%nl(i))
         DO j = 2, gthpot%nl(i)
            WRITE (iw, '(T20,5F18.12)') (gthpot%hnl(j, k, i), k=j, gthpot%nl(i))
         END DO
      END DO
      IF (.NOT. PRESENT(iunit)) CALL close_file(unit_number=iw)

   END SUBROUTINE atom_write_pseudo_param

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_print_orbitals(atom, iw)
      TYPE(atom_type), POINTER                           :: atom
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_print_orbitals', &
         routineP = moduleN//':'//routineN

      SELECT CASE (atom%method_type)
      CASE DEFAULT
         CALL cp__b("atom_output.F",764,"")
      CASE (do_rks_atom)
         CALL atom_print_orbitals_helper(atom, atom%orbitals%wfn, "", iw)
      CASE (do_uks_atom)
         CALL atom_print_orbitals_helper(atom, atom%orbitals%wfna, "Alpha", iw)
         CALL atom_print_orbitals_helper(atom, atom%orbitals%wfnb, "Beta", iw)
      CASE (do_rhf_atom)
         CALL atom_print_orbitals_helper(atom, atom%orbitals%wfn, "", iw)
      CASE (do_uhf_atom)
         CALL atom_print_orbitals_helper(atom, atom%orbitals%wfna, "Alpha", iw)
         CALL atom_print_orbitals_helper(atom, atom%orbitals%wfnb, "Beta", iw)
      CASE (do_rohf_atom)
         CALL cp__b("atom_output.F",776,"")
      END SELECT

   END SUBROUTINE atom_print_orbitals

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param wfn ...
!> \param description ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE atom_print_orbitals_helper(atom, wfn, description, iw)
      TYPE(atom_type), POINTER                           :: atom
      REAL(KIND=dp), DIMENSION(:, :, 0:), INTENT(INOUT)  :: wfn
      CHARACTER(len=*), INTENT(IN)                       :: description
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_print_orbitals_helper', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: b, l, maxl, nb, nv, v

      WRITE (iw, '(/,A,A,A)') " Atomic orbital expansion coefficients [", description, "]"

      maxl = atom%state%maxl_calc
      DO l = 0, maxl

         nb = atom%basis%nbas(l)
         nv = atom%state%maxn_calc(l)
         IF (nb > 0 .AND. nv > 0) THEN
            nv = MIN(nv, SIZE(wfn, 2))
            DO v = 1, nv
               WRITE (iw, '(/,"    ORBITAL      L = ",I1,"      State = ",I3)') l, v
               DO b = 1, nb
                  WRITE (iw, '("      ",ES23.15)') wfn(b, v, l)
               END DO
            END DO
         END IF
      END DO
   END SUBROUTINE atom_print_orbitals_helper

! **************************************************************************************************

END MODULE atom_output

! **************************************************************************************************
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Optimizer for the atomic code
! **************************************************************************************************
MODULE atom_optimization
   USE atom_types,                      ONLY: atom_optimization_type
   USE kinds,                           ONLY: dp
   USE lapack,                          ONLY: lapack_sgelss
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_optimization'

   TYPE hmat_type
      REAL(dp), DIMENSION(:, :, :), POINTER     :: emat
      REAL(dp), DIMENSION(:, :, :), POINTER     :: fmat
   END TYPE hmat_type

   TYPE atom_history_type
      INTEGER                               :: max_history
      INTEGER                               :: hlen
      INTEGER                               :: hpos
      REAL(dp)                              :: damping
      REAL(dp)                              :: eps_diis
      REAL(dp), DIMENSION(:, :), POINTER       :: dmat
      TYPE(hmat_type), DIMENSION(:), POINTER  :: hmat
   END TYPE atom_history_type

   PUBLIC :: atom_opt, &
             atom_history_type, atom_history_init, atom_history_update, atom_history_release

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param history ...
!> \param optimization ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE atom_history_init(history, optimization, matrix)
      TYPE(atom_history_type), INTENT(INOUT)             :: history
      TYPE(atom_optimization_type), INTENT(IN)           :: optimization
      REAL(dp), DIMENSION(:, :, :), INTENT(IN)           :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_history_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n1, n2, n3, ndiis
      REAL(dp)                                           :: damp, eps

      ndiis = optimization%n_diis
      eps = optimization%eps_diis
      damp = optimization%damping

      CALL atom_history_release(history)

      history%max_history = ndiis
      history%hlen = 0
      history%hpos = 0
      history%damping = damp
      history%eps_diis = eps
      ALLOCATE (history%dmat(ndiis+1, ndiis+1))

      ALLOCATE (history%hmat(ndiis))
      n1 = SIZE(matrix, 1)
      n2 = SIZE(matrix, 2)
      n3 = SIZE(matrix, 3)
      DO i = 1, ndiis
         ALLOCATE (history%hmat(i)%emat(n1, n2, n3))
         ALLOCATE (history%hmat(i)%fmat(n1, n2, n3))
      END DO

   END SUBROUTINE atom_history_init
! **************************************************************************************************
!> \brief ...
!> \param history ...
!> \param fmat ...
!> \param emat ...
! **************************************************************************************************
   SUBROUTINE atom_history_update(history, fmat, emat)
      TYPE(atom_history_type), INTENT(INOUT)             :: history
      REAL(dp), DIMENSION(:, :, :), INTENT(IN)           :: fmat, emat

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_history_update', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nlen, nmax, nnow

      nmax = history%max_history
      nlen = MIN(history%hlen+1, nmax)
      nnow = history%hpos+1
      IF (nnow > nmax) nnow = 1

      history%hmat(nnow)%fmat = fmat
      history%hmat(nnow)%emat = emat

      history%hlen = nlen
      history%hpos = nnow

   END SUBROUTINE atom_history_update
! **************************************************************************************************
!> \brief ...
!> \param history ...
! **************************************************************************************************
   SUBROUTINE atom_history_release(history)
      TYPE(atom_history_type), INTENT(INOUT)             :: history

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_history_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      history%max_history = 0
      history%hlen = 0
      history%hpos = 0
      history%damping = 0._dp
      history%eps_diis = 0._dp
      IF (ASSOCIATED(history%dmat)) THEN
         DEALLOCATE (history%dmat)
      END IF
      IF (ASSOCIATED(history%hmat)) THEN
         DO i = 1, SIZE(history%hmat)
            IF (ASSOCIATED(history%hmat(i)%emat)) THEN
               DEALLOCATE (history%hmat(i)%emat)
            END IF
            IF (ASSOCIATED(history%hmat(i)%fmat)) THEN
               DEALLOCATE (history%hmat(i)%fmat)
            END IF
         END DO
         DEALLOCATE (history%hmat)
      END IF

   END SUBROUTINE atom_history_release
! **************************************************************************************************
!> \brief ...
!> \param fmat ...
!> \param history ...
!> \param err ...
! **************************************************************************************************
   SUBROUTINE atom_opt(fmat, history, err)
      REAL(dp), DIMENSION(:, :, :), INTENT(INOUT)        :: fmat
      TYPE(atom_history_type), INTENT(INOUT)             :: history
      REAL(dp), INTENT(IN)                               :: err

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_opt', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info, j, lwork, na, nb, nlen, nm, &
                                                            nmax, nnow, rank
      REAL(dp)                                           :: a, rcond, t
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: s, work
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: vec

      nmax = history%max_history
      nnow = history%hpos
      a = history%damping
      IF (history%hlen > 1) THEN
         IF (err < history%eps_diis) THEN
            ! DIIS
            rcond = 1.e-10_dp
            lwork = 25*nmax
            ALLOCATE (vec(nmax+1, 2), s(nmax+1), work(lwork))
            nlen = history%hlen
            vec = 0._dp
            vec(nlen+1, 1) = 1._dp
            history%dmat(1:nlen, nlen+1) = 1._dp
            history%dmat(nlen+1, 1:nlen) = 1._dp
            history%dmat(nlen+1, nlen+1) = 0._dp
            DO i = 1, nlen
               na = nnow+1-i
               IF (na < 1) na = nmax+na
               DO j = i, nlen
                  nb = nnow+1-j
                  IF (nb < 1) nb = nmax+nb
                  t = SUM(history%hmat(na)%emat*history%hmat(nb)%emat)
                  history%dmat(i, j) = t
                  history%dmat(j, i) = t
               END DO
            END DO
            CALL lapack_sgelss(nlen+1, nlen+1, 1, history%dmat, nmax+1, vec, nmax+1, s, &
                               rcond, rank, work, lwork, info)
            IF(.NOT.(info == 0))CALL cp__a("atom_optimization.F",190)
            fmat = 0._dp
            DO i = 1, nlen
               na = nnow+1-i
               IF (na < 1) na = nmax+na
               fmat = fmat+vec(i, 1)*history%hmat(na)%fmat
            END DO

            DEALLOCATE (vec, s, work)
         ELSE
            ! damping
            nm = nnow-1
            IF (nm < 1) nm = history%max_history
            fmat = a*history%hmat(nnow)%fmat+(1._dp-a)*history%hmat(nm)%fmat
         END IF
      ELSEIF (history%hlen == 1) THEN
         fmat = history%hmat(nnow)%fmat
      ELSE
         CALL cp__b("atom_optimization.F",208,"")
      END IF

   END SUBROUTINE atom_opt
! **************************************************************************************************

END MODULE atom_optimization
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Exchange and Correlation functional calculations
!> \par History
!>      (13-Feb-2001) JGH, based on earlier version of apsi
!>      02.2003 Many many changes [fawzi]
!>      03.2004 new xc interface [fawzi]
!>      04.2004 kinetic functionals [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE xc
   USE cp_array_utils_r,                ONLY: cp_3d_r_p_type
   USE cp_linked_list_xc_deriv,         ONLY: cp_sll_xc_deriv_next,&
                                              cp_sll_xc_deriv_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type,&
                                              cp_to_string
   USE input_section_types,             ONLY: section_get_ival,&
                                              section_get_rval,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: default_path_length,&
                                              dp
   USE message_passing,                 ONLY: mp_sum
   USE pw_grid_types,                   ONLY: PW_MODE_DISTRIBUTED,&
                                              pw_grid_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_derive,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_cr3d,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_spline_utils,                 ONLY: &
        nn10_coeffs, nn10_deriv_coeffs, nn50_coeffs, nn50_deriv_coeffs, pw_nn_deriv_r, &
        pw_nn_smear_r, pw_spline2_deriv_g, pw_spline2_interpolate_values_g, pw_spline3_deriv_g, &
        pw_spline3_interpolate_values_g, pw_spline_scale_deriv, spline2_coeffs, &
        spline2_deriv_coeffs, spline3_coeffs, spline3_deriv_coeffs
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_create,&
                                              pw_p_type,&
                                              pw_release,&
                                              pw_type
   USE xc_derivative_desc,              ONLY: MAX_DERIVATIVE_DESC_LENGTH,&
                                              MAX_LABEL_LENGTH
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_create,&
                                              xc_dset_get_derivative,&
                                              xc_dset_release
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_derivatives,                  ONLY: xc_functionals_eval,&
                                              xc_functionals_get_needs
   USE xc_input_constants,              ONLY: &
        xc_debug_new_routine, xc_deriv_nn10_smooth, xc_deriv_nn50_smooth, xc_deriv_pw, &
        xc_deriv_spline2, xc_deriv_spline2_smooth, xc_deriv_spline3, xc_deriv_spline3_smooth, &
        xc_new_f_routine, xc_rho_nn10, xc_rho_nn50, xc_rho_no_smooth, xc_rho_spline2_smooth, &
        xc_rho_spline3_smooth, xc_test_lsd_f_routine
   USE xc_rho_set_types,                ONLY: xc_rho_set_create,&
                                              xc_rho_set_get,&
                                              xc_rho_set_release,&
                                              xc_rho_set_type,&
                                              xc_rho_set_update
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: xc_vxc_pw_create1, xc_vxc_pw_create, &
             xc_exc_calc, &
             xc_calc_2nd_deriv, xc_prep_2nd_deriv, divide_by_norm_drho

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc'

CONTAINS

! **************************************************************************************************
!> \brief Exchange and Correlation functional calculations.
!>      depending on the selected functional_routine calls
!>      the correct routine
!> \param vxc_rho will contain the v_xc part that depend on rho
!>        (if one of the choosen xc functionals has it it is allocated and you
!>        are responsible for it)
!> \param vxc_tau will contain the kinetic (tau) part of v_xcthe functional
!>        (if one of the choosen xc functionals has it it is allocated and you
!>        are responsible for it)
!> \param rho_r the value of the density in the real space
!> \param rho_g value of the density in the g space (needs to be associated
!>        only for gradient corrections)
!> \param tau value of the kinetic density tau on the grid (can be null,
!>        used only with meta functionals)
!> \param exc the xc energy
!> \param xc_section parameters selecting the xc and the method used to
!>        calculate it
!> \param pw_pool the pool for the grids
!> \param compute_virial should the virial be computed... if so virial_xc must be present
!> \param virial_xc for calculating the GGA part of the stress
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE xc_vxc_pw_create1(vxc_rho, vxc_tau, rho_r, rho_g, tau, exc, xc_section, &
                                pw_pool, compute_virial, virial_xc)
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: vxc_rho, vxc_tau, rho_r, rho_g, tau
      REAL(KIND=dp), INTENT(out)                         :: exc
      TYPE(section_vals_type), POINTER                   :: xc_section
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      LOGICAL                                            :: compute_virial
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: virial_xc

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_vxc_pw_create1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: f_routine

      IF(.NOT.(ASSOCIATED(rho_r)))CALL cp__a("xc/xc.F",126)
      IF(.NOT.(ASSOCIATED(xc_section)))CALL cp__a("xc/xc.F",127)
      IF(.NOT.(.NOT. ASSOCIATED(vxc_rho)))CALL cp__a("xc/xc.F",128)
      IF(.NOT.(.NOT. ASSOCIATED(vxc_tau)))CALL cp__a("xc/xc.F",129)

      CALL section_vals_val_get(xc_section, "FUNCTIONAL_ROUTINE", &
                                i_val=f_routine)
      SELECT CASE (f_routine)
      CASE (xc_new_f_routine)
         CALL xc_vxc_pw_create(vxc_rho=vxc_rho, vxc_tau=vxc_tau, tau=tau, &
                               rho_r=rho_r, rho_g=rho_g, exc=exc, xc_section=xc_section, &
                               pw_pool=pw_pool, compute_virial=compute_virial, virial_xc=virial_xc)
      CASE (xc_debug_new_routine)
         CALL xc_vxc_pw_create_debug(vxc_rho=vxc_rho, vxc_tau=vxc_tau, tau=tau, &
                                     rho_r=rho_r, rho_g=rho_g, exc=exc, xc_section=xc_section, &
                                     pw_pool=pw_pool)
      CASE (xc_test_lsd_f_routine)
         CALL xc_vxc_pw_create_test_lsd(vxc_rho=vxc_rho, vxc_tau=vxc_tau, &
                                        tau=tau, rho_r=rho_r, rho_g=rho_g, exc=exc, &
                                        xc_section=xc_section, pw_pool=pw_pool)
      CASE default
      END SELECT

   END SUBROUTINE xc_vxc_pw_create1

! **************************************************************************************************
!> \brief calculates vxc using lsd with rhoa=rhob=0.5*rho and compares
!>      with the lda result
!> \param vxc_rho will contain the v_xc part that depend on rho
!>        (if one of the choosen xc functionals has it it is allocated and you
!>        are responsible for it)
!> \param vxc_tau will contain the kinetic (tau) part of v_xc
!>        (if one of the choosen xc functionals has it it is allocated and you
!>        are responsible for it)
!> \param rho_r the value of the density in the real space
!> \param rho_g value of the density in the g space (needs to be associated
!>        only for gradient corrections)
!> \param tau value of the kinetic density tau on the grid (can be null,
!>        used only with meta functionals)
!> \param exc the xc energy
!> \param xc_section which functional to calculate, and how
!> \param pw_pool the pool for the grids
!> \author Fawzi Mohamed
!> \note
!>      for debugging only: leaks, and non parallel
! **************************************************************************************************
   SUBROUTINE xc_vxc_pw_create_test_lsd(vxc_rho, vxc_tau, rho_r, rho_g, tau, &
                                        exc, xc_section, pw_pool)
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: vxc_rho, vxc_tau, rho_r, rho_g, tau
      REAL(KIND=dp), INTENT(out)                         :: exc
      TYPE(section_vals_type), POINTER                   :: xc_section
      TYPE(pw_pool_type), POINTER                        :: pw_pool

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_vxc_pw_create_test_lsd', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: compute_virial = .FALSE.

      CHARACTER(len=default_path_length)                 :: filename
      CHARACTER(len=MAX_LABEL_LENGTH), DIMENSION(:), &
         POINTER                                         :: split_desc
      INTEGER                                            :: i, ii, ispin, j, k
      INTEGER, DIMENSION(2, 3)                           :: bo
      REAL(kind=dp)                                      :: diff, exc2, maxdiff, tmp
      REAL(KIND=dp), DIMENSION(3, 3)                     :: virial_xc
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: pot, pot2, pot3
      TYPE(cp_sll_xc_deriv_type), POINTER                :: deriv_iter
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho2_g, rho2_r, tau2, vxc_rho2, vxc_tau2
      TYPE(xc_derivative_set_type), POINTER              :: dSet1, dSet2
      TYPE(xc_derivative_type), POINTER                  :: deriv, deriv2, deriv3
      TYPE(xc_rho_set_type), POINTER                     :: rho_set1, rho_set2

      NULLIFY (vxc_rho2, vxc_tau2, tau2, dSet1, dSet2, rho_set1, rho_set2, split_desc, pot, pot3, pot3, &
               deriv, deriv2, deriv3, rho2_g)

      bo = rho_r(1)%pw%pw_grid%bounds_local

      ALLOCATE (rho2_r(2))
      DO ispin = 1, 2
         NULLIFY (rho2_r(ispin)%pw)
         CALL pw_pool_create_pw(pw_pool, rho2_r(ispin)%pw, in_space=REALSPACE, &
                                use_data=REALDATA3D)
      END DO
      DO k = bo(1, 3), bo(2, 3)
         DO j = bo(1, 2), bo(2, 2)
            DO i = bo(1, 1), bo(2, 1)
               tmp = rho_r(1)%pw%cr3d(i, j, k)*0.5
               rho2_r(1)%pw%cr3d(i, j, k) = tmp
               rho2_r(2)%pw%cr3d(i, j, k) = tmp
            END DO
         END DO
      END DO

      IF (ASSOCIATED(tau)) THEN
         ALLOCATE (tau2(2))
         DO ispin = 1, 2
            NULLIFY (tau2(ispin)%pw)
            CALL pw_pool_create_pw(pw_pool, tau2(ispin)%pw, in_space=REALSPACE, &
                                   use_data=REALDATA3D)
         END DO

         DO k = bo(1, 3), bo(2, 3)
            DO j = bo(1, 2), bo(2, 2)
               DO i = bo(1, 1), bo(2, 1)
                  tmp = tau(1)%pw%cr3d(i, j, k)*0.5
                  tau2(1)%pw%cr3d(i, j, k) = tmp
                  tau2(2)%pw%cr3d(i, j, k) = tmp
               END DO
            END DO
         END DO
      END IF

      PRINT*, "about to calculate xc (lda)"
      CALL xc_rho_set_and_dset_create(rho_r=rho_r, rho_g=rho_g, &
                                      tau=tau, xc_section=xc_section, &
                                      pw_pool=pw_pool, rho_set=rho_set1, &
                                      deriv_set=dSet1, deriv_order=1, &
                                      needs_basic_components=.FALSE.)
      CALL xc_vxc_pw_create(rho_r=rho_r, rho_g=rho_g, tau=tau, &
                            vxc_rho=vxc_rho, vxc_tau=vxc_tau, exc=exc, xc_section=xc_section, &
                            pw_pool=pw_pool, &
                            compute_virial=compute_virial, virial_xc=virial_xc)
      PRINT*, "did calculate xc (lda)"
      PRINT*, "about to calculate xc (lsd)"
      CALL xc_rho_set_and_dset_create(rho_set=rho_set2, deriv_set=dSet2, &
                                      rho_r=rho2_r, rho_g=rho2_g, tau=tau2, xc_section=xc_section, &
                                      pw_pool=pw_pool, deriv_order=1, &
                                      needs_basic_components=.FALSE.)
      CALL xc_vxc_pw_create(rho_r=rho2_r, rho_g=rho2_g, tau=tau2, &
                            vxc_rho=vxc_rho2, vxc_tau=vxc_tau2, exc=exc2, xc_section=xc_section, &
                            pw_pool=pw_pool, compute_virial=compute_virial, virial_xc=virial_xc)
      PRINT*, "did calculate xc (new)"
      PRINT*, "at (0,0,0) rho_r=", rho_r(1)%pw%cr3d(0, 0, 0), &
         "rho2_r(1)=", rho2_r(1)%pw%cr3d(0, 0, 0), &
         "rho2_r(2)=", rho2_r(2)%pw%cr3d(0, 0, 0), &
         "rho_r_sm=", rho_set1%rho(0, 0, 0), "rhoa2_r_sm=", rho_set2%rhoa(0, 0, 0), &
         "rhob2_r_sm=", rho_set2%rhob(0, 0, 0)
      OPEN (unit=120, file="/mnt/sdcard/cp2k/work/rho.bindata", status="unknown", access='sequential', &
            form="unformatted", action="write")
      pot => rho_set1%rho
      WRITE (unit=120) pot(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(2, 3))
      CLOSE (unit=120)
      OPEN (unit=120, file="/mnt/sdcard/cp2k/work/rhoa.bindata", status="unknown", access='sequential', &
            form="unformatted", action="write")
      pot => rho_set2%rhoa
      WRITE (unit=120) pot(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(2, 3))
      CLOSE (unit=120)
      OPEN (unit=120, file="/mnt/sdcard/cp2k/work/rhob.bindata", status="unknown", access='sequential', &
            form="unformatted", action="write")
      pot => rho_set2%rhob
      WRITE (unit=120) pot(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(2, 3))
      CLOSE (unit=120)
      OPEN (unit=120, file="/mnt/sdcard/cp2k/work/ndrho.bindata", status="unknown", access='sequential', &
            form="unformatted", action="write")
      pot => rho_set1%norm_drho
      WRITE (unit=120) pot(:, :, bo(2, 3))
      CLOSE (unit=120)
      OPEN (unit=120, file="/mnt/sdcard/cp2k/work/ndrhoa.bindata", status="unknown", access='sequential', &
            form="unformatted", action="write")
      pot => rho_set2%norm_drhoa
      WRITE (unit=120) pot(:, :, bo(2, 3))
      CLOSE (unit=120)
      OPEN (unit=120, file="/mnt/sdcard/cp2k/work/ndrhob.bindata", status="unknown", access='sequential', &
            form="unformatted", action="write")
      pot => rho_set2%norm_drhob
      WRITE (unit=120) pot(:, :, bo(2, 3))
      CLOSE (unit=120)
      IF (rho_set1%has%tau) THEN
         OPEN (unit=120, file="/mnt/sdcard/cp2k/work/tau.bindata", status="unknown", access='sequential', &
               form="unformatted", action="write")
         pot => rho_set1%tau
         WRITE (unit=120) pot(:, :, bo(2, 3))
         CLOSE (unit=120)
      END IF
      IF (rho_set2%has%tau_spin) THEN
         OPEN (unit=120, file="/mnt/sdcard/cp2k/work/tau_a.bindata", status="unknown", access='sequential', &
               form="unformatted", action="write")
         pot => rho_set2%tau_a
         WRITE (unit=120) pot(:, :, bo(2, 3))
         CLOSE (unit=120)
         OPEN (unit=120, file="/mnt/sdcard/cp2k/work/tau_v.bindata", status="unknown", access='sequential', &
               form="unformatted", action="write")
         pot => rho_set2%tau_b
         WRITE (unit=120) pot(:, :, bo(2, 3))
         CLOSE (unit=120)
      END IF
      OPEN (unit=120, file="/mnt/sdcard/cp2k/work/vxc.bindata", status="unknown", access='sequential', &
            form="unformatted", action="write")
      pot => vxc_rho(1)%pw%cr3d
      WRITE (unit=120) pot(:, :, bo(2, 3))
      CLOSE (unit=120)
      OPEN (unit=120, file="/mnt/sdcard/cp2k/work/vxc2.bindata", status="unknown", access='sequential', &
            form="unformatted", action="write")
      pot => vxc_rho2(1)%pw%cr3d
      WRITE (unit=120) pot(:, :, bo(2, 3))
      CLOSE (unit=120)
      IF (ASSOCIATED(vxc_tau)) THEN
         OPEN (unit=120, file="/mnt/sdcard/cp2k/work/vxc_tau.bindata", status="unknown", access='sequential', &
               form="unformatted", action="write")
         pot => vxc_tau(1)%pw%cr3d
         WRITE (unit=120) pot(:, :, bo(2, 3))
         CLOSE (unit=120)
      END IF
      IF (ASSOCIATED(vxc_tau2)) THEN
         OPEN (unit=120, file="/mnt/sdcard/cp2k/work/vxc_tau2_a.bindata", status="unknown", access='sequential', &
               form="unformatted", action="write")
         pot => vxc_tau2(1)%pw%cr3d
         WRITE (unit=120) pot(:, :, bo(2, 3))
         CLOSE (unit=120)
         OPEN (unit=120, file="/mnt/sdcard/cp2k/work/vxc_tau2_b.bindata", status="unknown", access='sequential', &
               form="unformatted", action="write")
         pot => vxc_tau2(2)%pw%cr3d
         WRITE (unit=120) pot(:, :, bo(2, 3))
         CLOSE (unit=120)
      END IF

      PRINT*, "calc diff on vxc"
      maxDiff = 0.0_dp
      DO ispin = 1, 1
         ii = 0
         DO k = bo(1, 3), bo(2, 3)
            DO j = bo(1, 2), bo(2, 2)
               DO i = bo(1, 1), bo(2, 1)
                  ii = ii+1
                  diff = ABS(vxc_rho(ispin)%pw%cr3d(i, j, k)- &
                             vxc_rho2(ispin)%pw%cr3d(i, j, k))
                  IF (ii == 1) THEN
                     PRINT*, "vxc", ispin, "=", vxc_rho(ispin)%pw%cr3d(i, j, k), &
                        "vs", vxc_rho2(ispin)%pw%cr3d(i, j, k), "diff=", diff
                  END IF
                  IF (maxDiff < diff) THEN
                     maxDiff = diff
                     PRINT*, "diff=", diff, " at ", i, ",", j, ",", k, &
                        " spin=", ispin, "rho=", rho_set1%rho(i, j, k), &
                        " ndrho=", rho_set1%norm_drho(i, j, k)
                  END IF
               END DO
            END DO
         END DO
      END DO
      PRINT*, "diff exc=", ABS(exc-exc2), "diff vxc=", maxdiff
      ! IF(.NOT.(maxdiff<5.e-11))CALL cp__a("xc/xc.F",366)
      ! IF(.NOT.(ABS(exc-exc2)<1.e-14))CALL cp__a("xc/xc.F",367)

      IF (ASSOCIATED(vxc_tau)) THEN
         PRINT*, "calc diff on vxc_tau"
         maxDiff = 0.0_dp
         DO ispin = 1, 1
            ii = 0
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     ii = ii+1
                     diff = ABS(vxc_tau(ispin)%pw%cr3d(i, j, k)- &
                                vxc_tau2(ispin)%pw%cr3d(i, j, k))
                     IF (ii == 1) THEN
                        PRINT*, "vxc_tau", ispin, "=", vxc_tau(ispin)%pw%cr3d(i, j, k), &
                           "vs", vxc_tau2(ispin)%pw%cr3d(i, j, k), "diff=", diff
                     END IF
                     IF (maxDiff < diff) THEN
                        maxDiff = diff
                        PRINT*, "diff=", diff, " at ", i, ",", j, ",", k, &
                           " spin=", ispin, "rho=", rho_set1%rho(i, j, k), &
                           " ndrho=", rho_set1%norm_drho(i, j, k)
                     END IF
                  END DO
               END DO
            END DO
         END DO
         PRINT*, "diff exc=", ABS(exc-exc2), "diff vxc_tau=", maxdiff
      END IF
      deriv_iter => dSet1%derivs
      DO WHILE (cp_sll_xc_deriv_next(deriv_iter, el_att=deriv))
         CALL xc_derivative_get(deriv, &
                                split_desc=split_desc, deriv_data=pot)
         SELECT CASE (SIZE (split_desc))
         CASE (0)
            filename = "e_0.bindata"
            deriv2 => xc_dset_get_derivative(dSet2, "")
         CASE (1)
            filename = "e_"//TRIM(split_desc(1))//".bindata"
            IF (split_desc(1) == "rho") THEN
               deriv2 => xc_dset_get_derivative(dSet2, "(rhoa)")
            ELSEIF (split_desc(1) == "tau") THEN
               deriv2 => xc_dset_get_derivative(dSet2, "(tau_a)")
            ELSEIF (split_desc(1) == "norm_drho") THEN
               deriv2 => xc_dset_get_derivative(dSet2, "(norm_drhoa)")
               deriv3 => xc_dset_get_derivative(dSet2, "(norm_drho)")
               IF (ASSOCIATED(deriv3)) THEN
                  IF (ASSOCIATED(deriv2)) THEN
                     CALL xc_derivative_get(deriv2, &
                                            deriv_data=pot2)
                     CALL xc_derivative_get(deriv3, &
                                            deriv_data=pot3)
                     pot2 = pot2+pot3
                  ELSE
                     deriv2 => deriv3
                  END IF
                  NULLIFY (deriv3, pot2, pot3)
               END IF
            ELSE
               CALL cp__b("xc/xc.F",426,"")
            END IF
         CASE default
            CALL cp__b("xc/xc.F",429,"")
         END SELECT
         CALL xc_derivative_get(deriv2, &
                                deriv_data=pot2)
         PRINT*, "checking ", filename
         maxDiff = 0.0_dp
         DO k = bo(1, 3), bo(2, 3)
            DO j = bo(1, 2), bo(2, 2)
               DO i = bo(1, 1), bo(2, 1)
                  diff = ABS(pot(i, j, k)-pot2(i, j, k))
                  IF (maxDiff < diff) THEN
                     maxDiff = diff
                     PRINT*, "ediff(", i, j, k, ")=", maxDiff, &
                        "rho=", rho_set1%rho(i, j, k), &
                        "ndrho=", rho_set1%norm_drho(i, j, k)
                  END IF
               END DO
            END DO
         END DO
         PRINT*, "maxdiff ", filename, "=", maxDiff
         OPEN (unit=120, file=TRIM(filename), status="unknown", &
               access='sequential', &
               form="unformatted")
         WRITE (unit=120) pot(:, :, bo(2, 3))
         CLOSE (unit=120)
      END DO
      deriv_iter => dSet2%derivs
      DO WHILE (cp_sll_xc_deriv_next(deriv_iter, el_att=deriv))
         CALL xc_derivative_get(deriv, &
                                split_desc=split_desc, deriv_data=pot)
         SELECT CASE (SIZE (split_desc))
         CASE (0)
            filename = "e_0-2.bindata"
         CASE (1)
            filename = "e_"//TRIM(split_desc(1))//"-2.bindata"
         CASE default
            CALL cp__b("xc/xc.F",465,"")
         END SELECT
         OPEN (unit=120, file=TRIM(filename), status="unknown", &
               access='sequential', &
               form="unformatted")
         WRITE (unit=120) pot(:, :, bo(2, 3))
         CLOSE (unit=120)
      END DO
      CALL xc_rho_set_release(rho_set1)
      CALL xc_rho_set_release(rho_set2)
      CALL xc_dset_release(dSet2)
      CALL xc_dset_release(dSet1)
      DO ispin = 1, 2
         CALL pw_pool_give_back_pw(pw_pool, rho2_r(ispin)%pw)
         CALL pw_pool_give_back_pw(pw_pool, vxc_rho2(ispin)%pw)
         IF (ASSOCIATED(vxc_tau2)) THEN
            CALL pw_pool_give_back_pw(pw_pool, vxc_tau2(ispin)%pw)
         END IF
      END DO
      DEALLOCATE (vxc_rho2, rho2_r, rho2_g)
      IF (ASSOCIATED(vxc_tau2)) THEN
         DEALLOCATE (vxc_tau2)
      END IF

   END SUBROUTINE xc_vxc_pw_create_test_lsd

! **************************************************************************************************
!> \brief calculates vxc outputting the yz plane of rho, and of the various components
!>      of the the derivatives and of vxc
!> \param vxc_rho will contain the v_xc part that depend on rho
!>        (if one of the choosen xc functionals has it it is allocated and you
!>        are responsible for it)
!> \param vxc_tau will contain the kinetic (tau) part of v_xc
!>        (if one of the choosen xc functionals has it it is allocated and you
!>        are responsible for it)
!> \param rho_r the value of the density in the real space
!> \param rho_g value of the density in the g space (needs to be associated
!>        only for gradient corrections)
!> \param tau value of the kinetic density tau on the grid (can be null,
!>        used only with meta functionals)
!> \param exc the xc energy
!> \param xc_section which functional should be used, and how to do it
!> \param pw_pool the pool for the grids
!> \author Fawzi Mohamed
!> \note
!>      for debugging only.
! **************************************************************************************************
   SUBROUTINE xc_vxc_pw_create_debug(vxc_rho, vxc_tau, rho_r, rho_g, tau, exc, &
                                     xc_section, pw_pool)
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: vxc_rho, vxc_tau, rho_r, rho_g, tau
      REAL(KIND=dp), INTENT(out)                         :: exc
      TYPE(section_vals_type), POINTER                   :: xc_section
      TYPE(pw_pool_type), POINTER                        :: pw_pool

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_vxc_pw_create_debug', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: compute_virial = .FALSE.

      CHARACTER(len=default_path_length)                 :: filename
      CHARACTER(len=MAX_LABEL_LENGTH), DIMENSION(:), &
         POINTER                                         :: split_desc
      INTEGER                                            :: i, ispin, j, k
      INTEGER, DIMENSION(2, 3)                           :: bo
      REAL(KIND=dp), DIMENSION(3, 3)                     :: virial_xc
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: pot
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_sll_xc_deriv_type), POINTER                :: deriv_iter
      TYPE(xc_derivative_set_type), POINTER              :: dSet1
      TYPE(xc_derivative_type), POINTER                  :: deriv
      TYPE(xc_rho_set_type), POINTER                     :: rho_set1

      NULLIFY (dSet1, rho_set1, split_desc, pot, &
               deriv)
      logger => cp_get_default_logger()

      bo = rho_r(1)%pw%pw_grid%bounds_local

      CALL xc_rho_set_and_dset_create(rho_r=rho_r, rho_g=rho_g, &
                                      tau=tau, xc_section=xc_section, &
                                      pw_pool=pw_pool, rho_set=rho_set1, &
                                      deriv_set=dSet1, deriv_order=1, &
                                      needs_basic_components=.FALSE.)

      ! outputs 0,:,: plane
      IF (bo(1, 1) <= 0 .AND. 0 <= bo(2, 1)) THEN
         IF (rho_set1%has%rho_spin) THEN
            OPEN (unit=120, file="/mnt/sdcard/cp2k/work/rhoa.bindata", status="unknown", access='sequential', &
                  form="unformatted", action="write")
            pot => rho_set1%rhoa
            WRITE (unit=120) pot(0, :, :)
            CLOSE (unit=120)
            OPEN (unit=120, file="/mnt/sdcard/cp2k/work/rhob.bindata", status="unknown", access='sequential', &
                  form="unformatted", action="write")
            pot => rho_set1%rhob
            WRITE (unit=120) pot(0, :, :)
            CLOSE (unit=120)
         END IF
         IF (rho_set1%has%norm_drho) THEN
            OPEN (unit=120, file="/mnt/sdcard/cp2k/work/ndrho.bindata", status="unknown", access='sequential', &
                  form="unformatted", action="write")
            pot => rho_set1%norm_drho
            WRITE (unit=120) pot(0, :, :)
            CLOSE (unit=120)
         END IF
         IF (rho_set1%has%norm_drho_spin) THEN
            OPEN (unit=120, file="/mnt/sdcard/cp2k/work/ndrhoa.bindata", status="unknown", access='sequential', &
                  form="unformatted", action="write")
            pot => rho_set1%norm_drhoa
            WRITE (unit=120) pot(0, :, :)
            CLOSE (unit=120)
            OPEN (unit=120, file="/mnt/sdcard/cp2k/work/ndrhob.bindata", status="unknown", access='sequential', &
                  form="unformatted", action="write")
            pot => rho_set1%norm_drhob
            WRITE (unit=120) pot(0, :, :)
            CLOSE (unit=120)
         END IF
         IF (rho_set1%has%rho) THEN
            OPEN (unit=120, file="/mnt/sdcard/cp2k/work/rho.bindata", status="unknown", access='sequential', &
                  form="unformatted", action="write")
            pot => rho_set1%rho
            WRITE (unit=120) pot(0, :, :)
            CLOSE (unit=120)
         END IF
         IF (rho_set1%has%tau) THEN
            OPEN (unit=120, file="/mnt/sdcard/cp2k/work/tau.bindata", status="unknown", access='sequential', &
                  form="unformatted", action="write")
            pot => rho_set1%tau
            WRITE (unit=120) pot(0, :, :)
            CLOSE (unit=120)
         END IF
         IF (rho_set1%has%tau_spin) THEN
            OPEN (unit=120, file="/mnt/sdcard/cp2k/work/tau_a.bindata", status="unknown", access='sequential', &
                  form="unformatted", action="write")
            pot => rho_set1%tau_a
            WRITE (unit=120) pot(0, :, :)
            CLOSE (unit=120)
            OPEN (unit=120, file="/mnt/sdcard/cp2k/work/tau_b.bindata", status="unknown", access='sequential', &
                  form="unformatted", action="write")
            pot => rho_set1%tau_b
            WRITE (unit=120) pot(0, :, :)
            CLOSE (unit=120)
         END IF

         deriv_iter => dSet1%derivs
         DO WHILE (cp_sll_xc_deriv_next(deriv_iter, el_att=deriv))
            CALL xc_derivative_get(deriv, &
                                   split_desc=split_desc, deriv_data=pot)
            SELECT CASE (SIZE (split_desc))
            CASE (0)
               filename = "e_0.bindata"
            CASE (1)
               filename = "e_"//TRIM(split_desc(1))//".bindata"
            CASE default
               CALL cp__b("xc/xc.F",618,"")
            END SELECT
            OPEN (unit=120, file=TRIM(filename), status="unknown", &
                  access='sequential', &
                  form="unformatted")
            WRITE (unit=120) pot(0, :, :)
            CLOSE (unit=120)
         END DO
      END IF

      CALL xc_vxc_pw_create(vxc_rho=vxc_rho, vxc_tau=vxc_tau, &
                            rho_r=rho_r, rho_g=rho_g, tau=tau, &
                            exc=exc, xc_section=xc_section, &
                            pw_pool=pw_pool, &
                            compute_virial=compute_virial, virial_xc=virial_xc)

      ! outputs 0,:,: plane
      IF (bo(1, 1) <= 0 .AND. 0 <= bo(2, 1)) THEN
         IF (ASSOCIATED(vxc_rho)) THEN
            DO ispin = 1, SIZE(vxc_rho)
               WRITE (filename, "('vxc-',i1,'.bindata')") ispin
               OPEN (unit=120, file=filename, status="unknown", access='sequential', &
                     form="unformatted", action="write")
               pot => vxc_rho(ispin)%pw%cr3d
               WRITE (unit=120) pot(0, :, :)
               CLOSE (unit=120)

               pot => vxc_rho(ispin)%pw%cr3d
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        IF (ABS(pot(i, j, k)) > 10.0_dp) THEN
                           WRITE (cp_logger_get_default_unit_nr(logger), &
                                  "(' vxc_',i1,'(',i6,',',i6,',',i6,')=',e11.4)", &
                                  advance="no") ispin, i, j, k, pot(i, j, k)
                           IF (rho_set1%has%rho_spin) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' rho=(',e11.4,',',e11.4,')')", advance="no") &
                                 rho_set1%rhoa(i, j, k), rho_set1%rhob(i, j, k)
                           ELSE IF (rho_set1%has%rho) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' rho=',e11.4)", advance="no") rho_set1%rho(i, j, k)
                           END IF
                           IF (rho_set1%has%norm_drho_spin) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' ndrho=(',e11.4,',',e11.4,')')", advance="no") &
                                 rho_set1%norm_drhoa(i, j, k), &
                                 rho_set1%norm_drhob(i, j, k)
                           ELSE IF (rho_set1%has%norm_drho) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' ndrho=',e11.4)", advance="no") rho_set1%norm_drho(i, j, k)
                           END IF
                           IF (rho_set1%has%tau_spin) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' tau=(',e11.4,',',e11.4,')')", advance="no") &
                                 rho_set1%tau_a(i, j, k), &
                                 rho_set1%tau_b(i, j, k)
                           ELSE IF (rho_set1%has%tau) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' tau=',e11.4)", advance="no") rho_set1%tau(i, j, k)
                           END IF

                           deriv_iter => dSet1%derivs
                           DO WHILE (cp_sll_xc_deriv_next(deriv_iter, el_att=deriv))
                              CALL xc_derivative_get(deriv, &
                                                     split_desc=split_desc, deriv_data=pot)
                              SELECT CASE (SIZE (split_desc))
                              CASE (0)
                                 filename = " e_0"
                              CASE (1)
                                 filename = " e_"//TRIM(split_desc(1))
                              CASE default
                                 CALL cp__b("xc/xc.F",690,"")
                              END SELECT
                              WRITE (unit=cp_logger_get_default_unit_nr(logger), &
                                     fmt="(a,'=',e11.4)", advance="no") &
                                 TRIM(filename), pot(i, j, k)
                           END DO

                           WRITE (cp_logger_get_default_unit_nr(logger), &
                                  "()")
                        END IF
                     END DO
                  END DO
               END DO
            END DO
         END IF
         IF (ASSOCIATED(vxc_tau)) THEN
            DO ispin = 1, SIZE(vxc_tau)
               WRITE (filename, "('vxc_tau_',i1,'.bindata')") ispin
               OPEN (unit=120, file=filename, status="unknown", access='sequential', &
                     form="unformatted", action="write")
               pot => vxc_tau(ispin)%pw%cr3d
               WRITE (unit=120) pot(0, :, :)
               CLOSE (unit=120)

               pot => vxc_tau(ispin)%pw%cr3d
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        IF (ABS(pot(i, j, k)) > 10.0_dp) THEN
                           WRITE (cp_logger_get_default_unit_nr(logger), &
                                  "('vxc_tau_',i1,'(',i6,',',i6,',',i6,')=',e11.4)", &
                                  advance="no") ispin, i, j, k, pot(i, j, k)
                           IF (rho_set1%has%rho_spin) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' rho=(',e11.4,',',e11.4,')')", advance="no") &
                                 rho_set1%rhoa(i, j, k), rho_set1%rhob(i, j, k)
                           ELSE IF (rho_set1%has%rho) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' rho=',e11.4)", advance="no") rho_set1%rho(i, j, k)
                           END IF
                           IF (rho_set1%has%norm_drho_spin) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' ndrho=(',e11.4,',',e11.4,')')", advance="no") &
                                 rho_set1%norm_drhoa(i, j, k), &
                                 rho_set1%norm_drhob(i, j, k)
                           ELSE IF (rho_set1%has%norm_drho) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' ndrho=',e11.4)", advance="no") rho_set1%norm_drho(i, j, k)
                           END IF
                           IF (rho_set1%has%tau_spin) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' tau=(',e11.4,',',e11.4,')')", advance="no") &
                                 rho_set1%tau_a(i, j, k), &
                                 rho_set1%tau_b(i, j, k)
                           ELSE IF (rho_set1%has%tau) THEN
                              WRITE (cp_logger_get_default_unit_nr(logger), &
                                     "(' tau=',e11.4)", advance="no") rho_set1%tau(i, j, k)
                           END IF

                           deriv_iter => dSet1%derivs
                           DO WHILE (cp_sll_xc_deriv_next(deriv_iter, el_att=deriv))
                              CALL xc_derivative_get(deriv, &
                                                     split_desc=split_desc, deriv_data=pot)
                              SELECT CASE (SIZE (split_desc))
                              CASE (0)
                                 filename = " e_0"
                              CASE (1)
                                 filename = " e_"//TRIM(split_desc(1))
                              CASE default
                                 CALL cp__b("xc/xc.F",759,"")
                              END SELECT
                              WRITE (unit=cp_logger_get_default_unit_nr(logger), &
                                     fmt="(a,'=',e11.4)", advance="no") &
                                 TRIM(filename), pot(i, j, k)
                           END DO

                           WRITE (cp_logger_get_default_unit_nr(logger), "()")
                        END IF
                     END DO
                  END DO
               END DO
            END DO
         END IF
      END IF

      CALL xc_dset_release(dSet1)
      CALL xc_rho_set_release(rho_set1)
   END SUBROUTINE xc_vxc_pw_create_debug

! **************************************************************************************************
!> \brief creates a xc_rho_set and a derivative set containing the derivatives
!>      of the functionals with the given deriv_order.
!> \param rho_set will contain the rho set
!> \param deriv_set will contain the derivatives
!> \param deriv_order the order of the requested derivatives. If positive
!>        0:deriv_order are calculated, if negative only -deriv_order is
!>        guaranteed to be valid. Orders not requested might be present,
!>        but might contain garbage.
!> \param rho_r the value of the density in the real space
!> \param rho_g value of the density in the g space (can be null, used only
!>        without smoothing of rho or deriv)
!> \param tau value of the kinetic density tau on the grid (can be null,
!>        used only with meta functionals)
!> \param xc_section the section describing the functional to use
!> \param pw_pool the pool for the grids
!> \param needs_basic_components if the basic components of the arguments
!>        should be kept in rho set (a basic component is for example drho
!>        when with lda a functional needs norm_drho)
!> \author fawzi
!> \note
!>      if any of the functionals is gradient corrected the full gradient is
!>      added to the rho set
! **************************************************************************************************
   SUBROUTINE xc_rho_set_and_dset_create(rho_set, deriv_set, deriv_order, &
                                         rho_r, rho_g, tau, xc_section, pw_pool, &
                                         needs_basic_components)

      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: deriv_order
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r, rho_g, tau
      TYPE(section_vals_type), POINTER                   :: xc_section
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      LOGICAL, INTENT(in)                                :: needs_basic_components

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_rho_set_and_dset_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nspins
      LOGICAL                                            :: lsd
      TYPE(section_vals_type), POINTER                   :: xc_fun_sections

      CALL timeset(routineN, handle)

      IF(.NOT.(.NOT. ASSOCIATED(rho_set)))CALL cp__a("xc/xc.F",824)
      IF(.NOT.(.NOT. ASSOCIATED(deriv_set)))CALL cp__a("xc/xc.F",825)
      IF(.NOT.(ASSOCIATED(pw_pool)))CALL cp__a("xc/xc.F",826)

      nspins = SIZE(rho_r)
      lsd = (nspins /= 1)

      xc_fun_sections => section_vals_get_subs_vals(xc_section, "XC_FUNCTIONAL")

      CALL xc_dset_create(deriv_set, pw_pool)

      CALL xc_rho_set_create(rho_set, &
                             rho_r(1)%pw%pw_grid%bounds_local, &
                             rho_cutoff=section_get_rval(xc_section, "density_cutoff"), &
                             drho_cutoff=section_get_rval(xc_section, "gradient_cutoff"), &
                             tau_cutoff=section_get_rval(xc_section, "tau_cutoff"))

      CALL xc_rho_set_update(rho_set, rho_r, rho_g, tau, &
                             xc_functionals_get_needs(xc_fun_sections, lsd, needs_basic_components), &
                             section_get_ival(xc_section, "XC_GRID%XC_DERIV"), &
                             section_get_ival(xc_section, "XC_GRID%XC_SMOOTH_RHO"), &
                             pw_pool)

      CALL xc_functionals_eval(xc_fun_sections, &
                               lsd=lsd, &
                               rho_set=rho_set, &
                               deriv_set=deriv_set, &
                               deriv_order=deriv_order)

      CALL timestop(handle)

   END SUBROUTINE xc_rho_set_and_dset_create

! **************************************************************************************************
!> \brief smooths the cutoff on rho with a function smooth(rho) that is 0
!>      for rho<rho_cutoff and 1 for rho>rho_cutoff*rho_smooth_cutoff_range:
!>      E= integral e_0*smooth(rho) => dE/d...= de/d... * smooth,
!>      dE/drho = de/drho * smooth + e_0 * dsmooth/drho
!> \param pot the potential to smooth
!> \param rho , rhoa,rhob: the value of the density (used to apply the cutoff)
!> \param rhoa ...
!> \param rhob ...
!> \param rho_cutoff the vaule at whch the cutoff function must go to 0
!> \param rho_smooth_cutoff_range range of the smoothing
!> \param e_0 value of e_0, if given it is assumed that pot is the derivative
!>        wrt. to rho, and needs the dsmooth*e_0 contribution
!> \param e_0_scale_factor ...
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE smooth_cutoff(pot, rho, rhoa, rhob, rho_cutoff, &
                            rho_smooth_cutoff_range, e_0, e_0_scale_factor)
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: pot, rho, rhoa, rhob
      REAL(kind=dp), INTENT(in)                          :: rho_cutoff, rho_smooth_cutoff_range
      REAL(kind=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: e_0
      REAL(kind=dp), INTENT(in), OPTIONAL                :: e_0_scale_factor

      CHARACTER(len=*), PARAMETER :: routineN = 'smooth_cutoff', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k
      INTEGER, DIMENSION(2, 3)                           :: bo
      REAL(kind=dp) :: my_e_0_scale_factor, my_rho, my_rho_n, my_rho_n2, rho_smooth_cutoff, &
         rho_smooth_cutoff_2, rho_smooth_cutoff_range_2

      IF(.NOT.(ASSOCIATED(pot)))CALL cp__a("xc/xc.F",888)
      bo(1, :) = LBOUND(pot)
      bo(2, :) = UBOUND(pot)
      my_e_0_scale_factor = 1.0_dp
      IF (PRESENT(e_0_scale_factor)) my_e_0_scale_factor = e_0_scale_factor
      rho_smooth_cutoff = rho_cutoff*rho_smooth_cutoff_range
      rho_smooth_cutoff_2 = (rho_cutoff+rho_smooth_cutoff)/2
      rho_smooth_cutoff_range_2 = rho_smooth_cutoff_2-rho_cutoff

      IF (rho_smooth_cutoff_range > 0.0_dp) THEN
         IF (PRESENT(e_0)) THEN
            IF(.NOT.(ASSOCIATED(e_0)))CALL cp__a("xc/xc.F",899)
            IF (ASSOCIATED(rho)) THEN
!$OMP              PARALLEL DO DEFAULT(NONE) SHARED(bo,e_0,pot,rho,&
!$OMP                          rho_cutoff,rho_smooth_cutoff,rho_smooth_cutoff_2,&
!$OMP                          rho_smooth_cutoff_range_2,my_e_0_scale_factor)&
!$OMP                    PRIVATE(k,j,i,my_rho,my_rho_n,my_rho_n2)
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        my_rho = rho(i, j, k)
                        IF (my_rho < rho_smooth_cutoff) THEN
                           IF (my_rho < rho_cutoff) THEN
                              pot(i, j, k) = 0.0_dp
                           ELSEIF (my_rho < rho_smooth_cutoff_2) THEN
                              my_rho_n = (my_rho-rho_cutoff)/rho_smooth_cutoff_range_2
                              my_rho_n2 = my_rho_n*my_rho_n
                              pot(i, j, k) = pot(i, j, k)* &
                                             my_rho_n2*(my_rho_n-0.5_dp*my_rho_n2)+ &
                                             my_e_0_scale_factor*e_0(i, j, k)* &
                                             my_rho_n2*(3.0_dp-2.0_dp*my_rho_n) &
                                             /rho_smooth_cutoff_range_2
                           ELSE
                              my_rho_n = 2.0_dp-(my_rho-rho_cutoff)/rho_smooth_cutoff_range_2
                              my_rho_n2 = my_rho_n*my_rho_n
                              pot(i, j, k) = pot(i, j, k)* &
                                             (1.0_dp-my_rho_n2*(my_rho_n-0.5_dp*my_rho_n2)) &
                                             +my_e_0_scale_factor*e_0(i, j, k)* &
                                             my_rho_n2*(3.0_dp-2.0_dp*my_rho_n) &
                                             /rho_smooth_cutoff_range_2
                           END IF
                        END IF
                     END DO
                  END DO
               END DO
            ELSE
!$OMP              PARALLEL DO DEFAULT(NONE) SHARED(bo,pot,e_0,rhoa,rhob,&
!$OMP                          rho_cutoff,rho_smooth_cutoff,rho_smooth_cutoff_2,&
!$OMP                          rho_smooth_cutoff_range_2,my_e_0_scale_factor)&
!$OMP                    PRIVATE(k,j,i,my_rho,my_rho_n,my_rho_n2)
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        my_rho = rhoa(i, j, k)+rhob(i, j, k)
                        IF (my_rho < rho_smooth_cutoff) THEN
                           IF (my_rho < rho_cutoff) THEN
                              pot(i, j, k) = 0.0_dp
                           ELSEIF (my_rho < rho_smooth_cutoff_2) THEN
                              my_rho_n = (my_rho-rho_cutoff)/rho_smooth_cutoff_range_2
                              my_rho_n2 = my_rho_n*my_rho_n
                              pot(i, j, k) = pot(i, j, k)* &
                                             my_rho_n2*(my_rho_n-0.5_dp*my_rho_n2)+ &
                                             my_e_0_scale_factor*e_0(i, j, k)* &
                                             my_rho_n2*(3.0_dp-2.0_dp*my_rho_n) &
                                             /rho_smooth_cutoff_range_2
                           ELSE
                              my_rho_n = 2.0_dp-(my_rho-rho_cutoff)/rho_smooth_cutoff_range_2
                              my_rho_n2 = my_rho_n*my_rho_n
                              pot(i, j, k) = pot(i, j, k)* &
                                             (1.0_dp-my_rho_n2*(my_rho_n-0.5_dp*my_rho_n2)) &
                                             +my_e_0_scale_factor*e_0(i, j, k)* &
                                             my_rho_n2*(3.0_dp-2.0_dp*my_rho_n) &
                                             /rho_smooth_cutoff_range_2
                           END IF
                        END IF
                     END DO
                  END DO
               END DO
            END IF
         ELSE
            IF (ASSOCIATED(rho)) THEN
!$OMP              PARALLEL DO DEFAULT(NONE) SHARED(bo,pot,&
!$OMP                          rho_cutoff,rho_smooth_cutoff,rho_smooth_cutoff_2,&
!$OMP                          rho_smooth_cutoff_range_2,rho)&
!$OMP                    PRIVATE(k,j,i,my_rho,my_rho_n,my_rho_n2)
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        my_rho = rho(i, j, k)
                        IF (my_rho < rho_smooth_cutoff) THEN
                           IF (my_rho < rho_cutoff) THEN
                              pot(i, j, k) = 0.0_dp
                           ELSEIF (my_rho < rho_smooth_cutoff_2) THEN
                              my_rho_n = (my_rho-rho_cutoff)/rho_smooth_cutoff_range_2
                              my_rho_n2 = my_rho_n*my_rho_n
                              pot(i, j, k) = pot(i, j, k)* &
                                             my_rho_n2*(my_rho_n-0.5_dp*my_rho_n2)
                           ELSE
                              my_rho_n = 2.0_dp-(my_rho-rho_cutoff)/rho_smooth_cutoff_range_2
                              my_rho_n2 = my_rho_n*my_rho_n
                              pot(i, j, k) = pot(i, j, k)* &
                                             (1.0_dp-my_rho_n2*(my_rho_n-0.5_dp*my_rho_n2))
                           END IF
                        END IF
                     END DO
                  END DO
               END DO
            ELSE
               IF(.NOT.(ASSOCIATED(rhoa)))CALL cp__a("xc/xc.F",996)
               IF(.NOT.(ASSOCIATED(rhob)))CALL cp__a("xc/xc.F",997)
!$OMP              PARALLEL DO DEFAULT(NONE) SHARED(bo,pot,&
!$OMP                          rho_cutoff,rho_smooth_cutoff,rho_smooth_cutoff_2,&
!$OMP                          rho_smooth_cutoff_range_2,rhoa,rhob)&
!$OMP                    PRIVATE(k,j,i,my_rho,my_rho_n,my_rho_n2)
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        my_rho = rhoa(i, j, k)+rhob(i, j, k)
                        IF (my_rho < rho_smooth_cutoff) THEN
                           IF (my_rho < rho_cutoff) THEN
                              pot(i, j, k) = 0.0_dp
                           ELSEIF (my_rho < rho_smooth_cutoff_2) THEN
                              my_rho_n = (my_rho-rho_cutoff)/rho_smooth_cutoff_range_2
                              my_rho_n2 = my_rho_n*my_rho_n
                              pot(i, j, k) = pot(i, j, k)* &
                                             my_rho_n2*(my_rho_n-0.5_dp*my_rho_n2)
                           ELSE
                              my_rho_n = 2.0_dp-(my_rho-rho_cutoff)/rho_smooth_cutoff_range_2
                              my_rho_n2 = my_rho_n*my_rho_n
                              pot(i, j, k) = pot(i, j, k)* &
                                             (1.0_dp-my_rho_n2*(my_rho_n-0.5_dp*my_rho_n2))
                           END IF
                        END IF
                     END DO
                  END DO
               END DO
            END IF
         END IF
      END IF
   END SUBROUTINE smooth_cutoff

! **************************************************************************************************
!> \brief Exchange and Correlation functional calculations
!> \param vxc_rho will contain the v_xc part that depend on rho
!>        (if one of the choosen xc functionals has it it is allocated and you
!>        are responsible for it)
!> \param vxc_tau will contain the kinetic (tau) part of v_xc
!>        (if one of the choosen xc functionals has it it is allocated and you
!>        are responsible for it)
!> \param exc the xc energy
!> \param rho_r the value of the density in the real space
!> \param rho_g value of the density in the g space (needs to be associated
!>        only for gradient corrections)
!> \param tau value of the kinetic density tau on the grid (can be null,
!>        used only with meta functionals)
!> \param xc_section which functional to calculate, and how to do it
!> \param pw_pool the pool for the grids
!> \param compute_virial ...
!> \param virial_xc ...
!> \par History
!>      JGH (13-Jun-2002): adaptation to new functionals
!>      Fawzi (11.2002): drho_g(1:3)->drho_g
!>      Fawzi (1.2003). lsd version
!>      Fawzi (11.2003): version using the new xc interface
!>      Fawzi (03.2004): fft free for smoothed density and derivs, gga lsd
!>      Fawzi (04.2004): metafunctionals
!>      mguidon (12.2008) : laplace functionals
!> \author fawzi; based LDA version of JGH, based on earlier version of apsi
!> \note
!>      Beware: some really dirty pointer handling!
!>      energy should be kept consistent with xc_exc_calc
! **************************************************************************************************
   SUBROUTINE xc_vxc_pw_create(vxc_rho, vxc_tau, exc, rho_r, rho_g, tau, xc_section, &
                               pw_pool, compute_virial, virial_xc)
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: vxc_rho, vxc_tau
      REAL(KIND=dp), INTENT(out)                         :: exc
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r, rho_g, tau
      TYPE(section_vals_type), POINTER                   :: xc_section
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      LOGICAL                                            :: compute_virial
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: virial_xc

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_vxc_pw_create', &
         routineP = moduleN//':'//routineN
      CHARACTER(len=30), DIMENSION(2), PARAMETER :: &
         norm_drho_spin_name = (/"(norm_drhoa)", "(norm_drhob)"/)

      CHARACTER(len=MAX_DERIVATIVE_DESC_LENGTH)          :: desc
      INTEGER                                            :: handle, i, idir, ispin, j, jdir, k, &
                                                            n_deriv, npoints, nspins, order, &
                                                            xc_deriv_method_id, xc_rho_smooth_id
      INTEGER, DIMENSION(2, 3)                           :: bo
      INTEGER, DIMENSION(3, 3)                           :: nd, nd_laplace
      LOGICAL                                            :: dealloc_pw_to_deriv, has_laplace, &
                                                            has_tau, lsd, use_virial, zero_result
      REAL(KIND=dp)                                      :: density_smooth_cut_range, drho_cutoff, &
                                                            my_rho, ndr, rho_cutoff
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: deriv_data, norm_drho, norm_drho_spin, &
                                                            rho, rhoa, rhob, tmp_cr3d
      TYPE(cp_3d_r_p_type), DIMENSION(:), POINTER        :: drho, drho_spin, drhoa, drhob
      TYPE(cp_sll_xc_deriv_type), POINTER                :: pos
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_p_type), DIMENSION(2)                      :: vxc_to_deriv
      TYPE(pw_p_type), DIMENSION(3)                      :: pw_to_deriv, pw_to_deriv_rho
      TYPE(pw_type), POINTER                             :: tmp_g, tmp_r, virial_pw, vxc_g
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_derivative_type), POINTER                  :: deriv_att
      TYPE(xc_rho_set_type), POINTER                     :: rho_set

      CALL timeset(routineN, handle)
      NULLIFY (tmp_g, tmp_r, vxc_g, norm_drho_spin, norm_drho, drho_spin, drhoa, &
               drhob, pos, deriv_set, rho_set, virial_pw)
      nd = RESHAPE((/1, 0, 0, 0, 1, 0, 0, 0, 1/), (/3, 3/))
      DO idir = 1, 3
         NULLIFY (pw_to_deriv(idir)%pw, pw_to_deriv_rho(idir)%pw)
      END DO
      DO i = 1, 2
         NULLIFY (vxc_to_deriv(i)%pw)
      END DO

      pw_grid => rho_r(1)%pw%pw_grid

      IF(.NOT.(ASSOCIATED(xc_section)))CALL cp__a("xc/xc.F",1110)
      IF(.NOT.(ASSOCIATED(pw_pool)))CALL cp__a("xc/xc.F",1111)
      IF(.NOT.(.NOT. ASSOCIATED(vxc_rho)))CALL cp__a("xc/xc.F",1112)
      IF(.NOT.(.NOT. ASSOCIATED(vxc_tau)))CALL cp__a("xc/xc.F",1113)
      nspins = SIZE(rho_r)
      IF(.NOT.(ASSOCIATED(rho_r(SIZE(rho_r))%pw)))CALL cp__a("xc/xc.F",1115)
      lsd = (nspins /= 1)
      IF (lsd) THEN
         IF(.NOT.(nspins == 2))CALL cp__a("xc/xc.F",1118)
      END IF

      use_virial = compute_virial

      bo = rho_r(1)%pw%pw_grid%bounds_local
      npoints = (bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1)

      ! calculate the potential derivatives
      CALL xc_rho_set_and_dset_create(rho_set=rho_set, deriv_set=deriv_set, &
                                      deriv_order=1, rho_r=rho_r, rho_g=rho_g, tau=tau, &
                                      xc_section=xc_section, &
                                      pw_pool=pw_pool, &
                                      needs_basic_components=.TRUE.)

      CALL section_vals_val_get(xc_section, "XC_GRID%XC_DERIV", &
                                i_val=xc_deriv_method_id)
      CALL section_vals_val_get(xc_section, "XC_GRID%XC_SMOOTH_RHO", &
                                i_val=xc_rho_smooth_id)
      CALL section_vals_val_get(xc_section, "DENSITY_SMOOTH_CUTOFF_RANGE", &
                                r_val=density_smooth_cut_range)

      CALL xc_rho_set_get(rho_set, rho_cutoff=rho_cutoff, &
                          drho_cutoff=drho_cutoff)

      has_tau = .FALSE.
      has_laplace = .FALSE.
      ! check for unknown derivatives
      n_deriv = 0
      pos => deriv_set%derivs
      DO WHILE (cp_sll_xc_deriv_next(pos, el_att=deriv_att))
         CALL xc_derivative_get(deriv_att, order=order, &
                                desc=desc)
         IF (order == 1) THEN
            IF (lsd) THEN
               SELECT CASE (desc)
               CASE ("(rho)", "(rhoa)", "(rhob)", "(norm_drho)", "(norm_drhoa)", &
                     "(norm_drhob)")
                  n_deriv = n_deriv+1
               CASE ("(tau)", "(tau_a)", "(tau_b)")
                  has_tau = .TRUE.
                  n_deriv = n_deriv+1
               CASE ("(laplace_rhoa)", "(laplace_rhob)")
                  has_laplace = .TRUE.
                  n_deriv = n_deriv+1
               CASE default
                  !FM if you are looking at this error probably you are missing the
                  !FM cross term (drhoa_drhob), I never got round to implement it,
                  !FM in the functionals that I have implemented I use norm_drho
                  !FM instead, either do the same or implement it, it shouldn't be
                  !FM difficult (I am just lazy ;)
                  CALL cp__b("xc/xc.F",1169,"unknown functional derivative (LSD): '"//TRIM(desc)//"'")
               END SELECT
            ELSE
               SELECT CASE (desc)
               CASE ("(rho)", "(norm_drho)")
                  n_deriv = n_deriv+1
               CASE ("(tau)")
                  has_tau = .TRUE.
                  n_deriv = n_deriv+1
               CASE ("(laplace_rho)")
                  has_laplace = .TRUE.
                  n_deriv = n_deriv+1
               CASE default
                  CALL cp__b("xc/xc.F",1182,"unknown functional derivative (LDA): '"//TRIM(desc)//"'")
               END SELECT
            END IF
         END IF
      END DO

      ALLOCATE (vxc_rho(nspins))
      DO ispin = 1, nspins
         NULLIFY (vxc_rho(ispin)%pw)
      END DO

      CALL xc_rho_set_get(rho_set, rho=rho, rhoa=rhoa, rhob=rhob, &
                          can_return_null=.TRUE.)

      ! recover the vxc arrays
      IF (lsd) THEN
         deriv_att => xc_dset_get_derivative(deriv_set, "(rhoa)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL pw_create(vxc_rho(1)%pw, pw_grid=pw_grid, &
                           cr3d_ptr=deriv_att%deriv_data, &
                           use_data=REALDATA3D, in_space=REALSPACE)
            NULLIFY (deriv_att%deriv_data)
         ELSE
            CALL pw_pool_create_pw(pw_pool, vxc_rho(1)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(vxc_rho(1)%pw)
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(rhob)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL pw_create(vxc_rho(2)%pw, pw_grid=pw_grid, &
                           cr3d_ptr=deriv_att%deriv_data, &
                           use_data=REALDATA3D, in_space=REALSPACE)
            NULLIFY (deriv_att%deriv_data)
         ELSE
            CALL pw_pool_create_pw(pw_pool, vxc_rho(2)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(vxc_rho(2)%pw)
         END IF

      ELSE
         deriv_att => xc_dset_get_derivative(deriv_set, "(rho)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL pw_create(vxc_rho(1)%pw, pw_grid=pw_grid, &
                           cr3d_ptr=deriv_att%deriv_data, &
                           use_data=REALDATA3D, in_space=REALSPACE)
            NULLIFY (deriv_att%deriv_data)
         ELSE
            CALL pw_pool_create_pw(pw_pool, vxc_rho(1)%pw, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(vxc_rho(1)%pw)
         END IF
      END IF

      deriv_att => xc_dset_get_derivative(deriv_set, "(rho)")
      IF (ASSOCIATED(deriv_att)) THEN
         IF (lsd) THEN ! otherwise already taken care in vxc recovery
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO DEFAULT(NONE) SHARED(bo,vxc_rho,deriv_data)&
!$OMP         PRIVATE(k,j,i)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     vxc_rho(1)%pw%cr3d(i, j, k) = vxc_rho(1)%pw%cr3d(i, j, k)+deriv_data(i, j, k)
                     vxc_rho(2)%pw%cr3d(i, j, k) = vxc_rho(2)%pw%cr3d(i, j, k)+deriv_data(i, j, k)
                  END DO
               END DO
            END DO
            CALL pw_pool_give_back_cr3d(pw_pool, deriv_att%deriv_data)
            NULLIFY (deriv_att%deriv_data)
         END IF
      END IF

      ! rhoa,rhob already taken care of in vxc recovery

      deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drho)")
      IF (ASSOCIATED(deriv_att)) THEN
         CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)

         CALL xc_rho_set_get(rho_set, norm_drho=norm_drho, &
                             drho=drho, drhoa=drhoa, drhob=drhob, rho_cutoff=rho_cutoff, &
                             drho_cutoff=drho_cutoff, &
                             can_return_null=.TRUE.)

         IF (ASSOCIATED(norm_drho)) THEN
!$OMP         PARALLEL DO DEFAULT(NONE) SHARED(bo,deriv_data,norm_drho,drho_cutoff)&
!$OMP                     PRIVATE(k,j,i)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     deriv_data(i, j, k) = -deriv_data(i, j, k)/MAX(norm_drho(i, j, k), drho_cutoff)
                  END DO
               END DO
            END DO
         ELSE IF (ASSOCIATED(drho)) THEN
!$OMP         PARALLEL DO DEFAULT(NONE) SHARED(bo,deriv_data,drho,drho_cutoff)&
!$OMP                     PRIVATE(k,j,i,ndr)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     ndr = SQRT(drho(1)%array(i, j, k)**2+ &
                                drho(2)%array(i, j, k)**2+ &
                                drho(3)%array(i, j, k)**2)
                     deriv_data(i, j, k) = -deriv_data(i, j, k)/MAX(ndr, drho_cutoff)
                  END DO
               END DO
            END DO
         ELSE
            IF(.NOT.(ASSOCIATED(drhoa)))CALL cp__a("xc/xc.F",1290)
            IF(.NOT.(ASSOCIATED(drhob)))CALL cp__a("xc/xc.F",1291)
!$OMP         PARALLEL DO DEFAULT(NONE) SHARED(bo,deriv_data,drhoa,drhob,drho_cutoff)&
!$OMP                     PRIVATE(k,j,i,ndr)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     ndr = SQRT((drhoa(1)%array(i, j, k)+drhob(1)%array(i, j, k))**2+ &
                                (drhoa(2)%array(i, j, k)+drhob(2)%array(i, j, k))**2+ &
                                (drhoa(3)%array(i, j, k)+drhob(3)%array(i, j, k))**2)
                     deriv_data(i, j, k) = -deriv_data(i, j, k)/MAX(ndr, drho_cutoff)
                  END DO
               END DO
            END DO
         END IF

         IF (ASSOCIATED(drho) .AND. ASSOCIATED(drho(1)%array)) THEN
            IF(.NOT.(ASSOCIATED(deriv_data)))CALL cp__a("xc/xc.F",1307)
            IF (use_virial) THEN
               CALL pw_pool_create_pw(pw_pool, virial_pw, &
                                      use_data=REALDATA3D, &
                                      in_space=REALSPACE)
               CALL pw_zero(virial_pw)
               DO idir = 1, 3
                  DO k = bo(1, 3), bo(2, 3)
                     DO j = bo(1, 2), bo(2, 2)
                        DO i = bo(1, 1), bo(2, 1)
                           virial_pw%cr3d(i, j, k) = drho(idir)%array(i, j, k)*deriv_data(i, j, k)
                        END DO
                     END DO
                  END DO
                  DO jdir = 1, idir
                     virial_xc(idir, jdir) = pw_grid%dvol* &
                                             accurate_sum(virial_pw%cr3d(:, :, :)* &
                                                          drho(jdir)%array(:, :, :))
                     virial_xc(jdir, idir) = virial_xc(idir, jdir)
                  END DO
               END DO
               CALL pw_pool_give_back_pw(pw_pool, virial_pw)
            END IF ! use_virial
            DO idir = 1, 3
               CALL pw_create(pw_to_deriv_rho(idir)%pw, pw_grid=pw_grid, &
                              cr3d_ptr=drho(idir)%array, &
                              use_data=REALDATA3D, in_space=REALSPACE)
               IF(.NOT.(ASSOCIATED(drho(idir)%array)))CALL cp__a("xc/xc.F",1334)
!$OMP            PARALLEL DO DEFAULT(NONE) SHARED(bo,deriv_data,drho,idir)&
!$OMP                        PRIVATE(k,j,i)
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        drho(idir)%array(i, j, k) = drho(idir)%array(i, j, k)*deriv_data(i, j, k)
                     END DO
                  END DO
               END DO
               NULLIFY (drho(idir)%array)
            END DO
         ELSE
            IF (use_virial) THEN
               CALL pw_pool_create_pw(pw_pool, virial_pw, &
                                      use_data=REALDATA3D, &
                                      in_space=REALSPACE)
               CALL pw_zero(virial_pw)
               DO idir = 1, 3
                  DO k = bo(1, 3), bo(2, 3)
                     DO j = bo(1, 2), bo(2, 2)
                        DO i = bo(1, 1), bo(2, 1)
                           my_rho = drhoa(idir)%array(i, j, k)+drhob(idir)%array(i, j, k)
                           virial_pw%cr3d(i, j, k) = my_rho*deriv_data(i, j, k)
                        END DO
                     END DO
                  END DO
                  DO jdir = 1, idir
                     virial_xc(idir, jdir) = pw_grid%dvol*accurate_sum(virial_pw%cr3d(:, :, :)* &
                                                                       (drhoa(jdir)%array(:, :, :)+drhob(jdir)%array(:, :, :)))
                     virial_xc(jdir, idir) = virial_xc(idir, jdir)
                  END DO
               END DO
               CALL pw_pool_give_back_pw(pw_pool, virial_pw)
            END IF ! use_virial
            DO idir = 1, 3
               CALL pw_pool_create_pw(pw_pool, pw_to_deriv_rho(idir)%pw, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
!$OMP            PARALLEL DO DEFAULT(NONE) SHARED(bo,deriv_data,&
!$OMP                           pw_to_deriv_rho,drhoa,drhob,idir)&
!$OMP                        PRIVATE(k,j,i,my_rho)
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        my_rho = drhoa(idir)%array(i, j, k)+drhob(idir)%array(i, j, k)
                        pw_to_deriv_rho(idir)%pw%cr3d(i, j, k) = my_rho*deriv_data(i, j, k)
                     END DO
                  END DO
               END DO
            END DO
         END IF

         CALL pw_pool_give_back_cr3d(pw_pool, deriv_att%deriv_data)

      END IF

      DO ispin = 1, nspins

         IF (lsd) THEN
            IF (ispin == 1) THEN
               CALL xc_rho_set_get(rho_set, drhoa=drho_spin, &
                                   can_return_null=.TRUE.)
               CALL xc_rho_set_get(rho_set, norm_drhoa=norm_drho_spin, &
                                   can_return_null=.TRUE.)
            ELSE
               CALL xc_rho_set_get(rho_set, drhob=drho_spin, &
                                   can_return_null=.TRUE.)
               CALL xc_rho_set_get(rho_set, norm_drhob=norm_drho_spin, &
                                   can_return_null=.TRUE.)
            END IF

            deriv_att => xc_dset_get_derivative(deriv_set, norm_drho_spin_name(ispin))
            IF (ASSOCIATED(deriv_att)) THEN
               IF(.NOT.(lsd))CALL cp__a("xc/xc.F",1407)
               CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
               CALL pw_create(tmp_r, pw_grid, &
                              use_data=REALDATA3D, in_space=REALSPACE, &
                              cr3d_ptr=deriv_data)

               IF (ASSOCIATED(norm_drho_spin)) THEN
!$OMP               PARALLEL DO DEFAULT(NONE) SHARED(bo,deriv_data,norm_drho_spin,drho_cutoff)&
!$OMP                           PRIVATE(k,j,i)
                  DO k = bo(1, 3), bo(2, 3)
                     DO j = bo(1, 2), bo(2, 2)
                        DO i = bo(1, 1), bo(2, 1)
                           deriv_data(i, j, k) = -deriv_data(i, j, k)/ &
                                                 MAX(norm_drho_spin(i, j, k), drho_cutoff)
                        END DO
                     END DO
                  END DO
               ELSE
!$OMP               PARALLEL DO DEFAULT(NONE) SHARED(bo,deriv_data,drho_spin,drho_cutoff)&
!$OMP                           PRIVATE(k,j,i, ndr)
                  DO k = bo(1, 3), bo(2, 3)
                     DO j = bo(1, 2), bo(2, 2)
                        DO i = bo(1, 1), bo(2, 1)
                           ndr = SQRT(drho_spin(1)%array(i, j, k)**2+ &
                                      drho_spin(2)%array(i, j, k)**2+ &
                                      drho_spin(3)%array(i, j, k)**2)
                           deriv_data(i, j, k) = -deriv_data(i, j, k)/MAX(ndr, drho_cutoff)
                        END DO
                     END DO
                  END DO
               END IF

               IF (use_virial) THEN
                  CALL pw_pool_create_pw(pw_pool, virial_pw, &
                                         use_data=REALDATA3D, &
                                         in_space=REALSPACE)
                  CALL pw_zero(virial_pw)
                  DO idir = 1, 3
                     DO k = bo(1, 3), bo(2, 3)
                        DO j = bo(1, 2), bo(2, 2)
                           DO i = bo(1, 1), bo(2, 1)
                              virial_pw%cr3d(i, j, k) = drho_spin(idir)%array(i, j, k)*deriv_data(i, j, k)
                           END DO
                        END DO
                     END DO
                     DO jdir = 1, idir
                        virial_xc(idir, jdir) = virial_xc(idir, jdir)+pw_grid%dvol* &
                                                accurate_sum(virial_pw%cr3d(:, :, :)* &
                                                             drho_spin(jdir)%array(:, :, :))
                        virial_xc(jdir, idir) = virial_xc(idir, jdir)
                     END DO
                  END DO
                  CALL pw_pool_give_back_pw(pw_pool, virial_pw)
               END IF ! use_virial

               vxc_to_deriv(ispin)%pw => tmp_r
               NULLIFY (tmp_r, deriv_att%deriv_data)

               DO idir = 1, 3
                  IF(.NOT.(ASSOCIATED(drho_spin(idir)%array)))CALL cp__a("xc/xc.F",1466)
                  IF(.NOT.(ASSOCIATED(vxc_to_deriv(ispin)%pw%cr3d)))CALL cp__a("xc/xc.F",1467)
!$OMP               PARALLEL DO DEFAULT(NONE) SHARED(bo,deriv_data,drho_spin,&
!$OMP                           ispin,idir,vxc_to_deriv,drho_cutoff)&
!$OMP                           PRIVATE(k,j,i)
                  DO k = bo(1, 3), bo(2, 3)
                     DO j = bo(1, 2), bo(2, 2)
                        DO i = bo(1, 1), bo(2, 1)
                           drho_spin(idir)%array(i, j, k) = vxc_to_deriv(ispin)%pw%cr3d(i, j, k)* &
                                                            drho_spin(idir)%array(i, j, k)
                        END DO
                     END DO
                  END DO

                  CALL pw_create(pw_to_deriv(idir)%pw, pw_grid=pw_grid, &
                                 cr3d_ptr=drho_spin(idir)%array, &
                                 use_data=REALDATA3D, in_space=REALSPACE)
                  NULLIFY (drho_spin(idir)%array)
               END DO

               dealloc_pw_to_deriv = .TRUE.
               CALL pw_pool_give_back_pw(pw_pool, vxc_to_deriv(ispin)%pw)
            END IF ! deriv_att

         END IF ! LSD

         IF (ASSOCIATED(pw_to_deriv_rho(1)%pw)) THEN
            IF (.NOT. ASSOCIATED(pw_to_deriv(1)%pw)) THEN
               DO idir = 1, 3
                  pw_to_deriv(idir)%pw => pw_to_deriv_rho(idir)%pw
               END DO
               dealloc_pw_to_deriv = (.NOT. lsd) .OR. (ispin == 2)
               IF (dealloc_pw_to_deriv) THEN
                  DO idir = 1, 3
                     NULLIFY (pw_to_deriv_rho(idir)%pw)
                  END DO
               END IF
            ELSE
               DO idir = 1, 3
                  CALL pw_axpy(pw_to_deriv_rho(idir)%pw, pw_to_deriv(idir)%pw)
                  IF (ispin == 2) THEN
                     CALL pw_pool_give_back_pw(pw_pool, pw_to_deriv_rho(idir)%pw)
                  END IF
               END DO
            END IF
         END IF

         IF (ASSOCIATED(pw_to_deriv(1)%pw)) THEN
            ! partial integration
            IF (xc_deriv_method_id /= xc_deriv_pw) THEN
               CALL pw_spline_scale_deriv(pw_to_deriv, &
                                          transpose=.TRUE.)
            END IF

            IF (xc_deriv_method_id == xc_deriv_pw .OR. &
                xc_deriv_method_id == xc_deriv_spline3 .OR. &
                xc_deriv_method_id == xc_deriv_spline2) THEN

               IF (.NOT. ASSOCIATED(vxc_g)) THEN
                  CALL pw_pool_create_pw(pw_pool, vxc_g, &
                                         use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
                  zero_result = .TRUE.
               ELSE
                  zero_result = .FALSE.
               END IF

               DO idir = 1, 3
                  IF (zero_result .AND. idir == 1) THEN
                     tmp_g => vxc_g
                  ELSE
                     CALL pw_pool_create_pw(pw_pool, tmp_g, &
                                            use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
                  END IF

                  CALL pw_transfer(pw_to_deriv(idir)%pw, tmp_g)

                  SELECT CASE (xc_deriv_method_id)
                  CASE (xc_deriv_pw)
                     CALL pw_derive(tmp_g, nd(:, idir))
                  CASE (xc_deriv_spline2)
                     CALL pw_spline2_interpolate_values_g(tmp_g)
                     CALL pw_spline2_deriv_g(tmp_g, idir=idir)
                  CASE (xc_deriv_spline3)
                     CALL pw_spline3_interpolate_values_g(tmp_g)
                     CALL pw_spline3_deriv_g(tmp_g, idir=idir)
                  CASE default
                     CALL cp__b("xc/xc.F",1552,"")
                  END SELECT

                  IF (zero_result .AND. idir == 1) THEN
                     NULLIFY (tmp_g)
                  ELSE
                     CALL pw_axpy(tmp_g, vxc_g)
                     CALL pw_pool_give_back_pw(pw_pool, tmp_g)
                  END IF
                  IF (dealloc_pw_to_deriv) THEN
                     CALL pw_pool_give_back_pw(pw_pool, pw_to_deriv(idir)%pw)
                  END IF
               END DO
               ! transfer vxc in real space
               CALL pw_pool_create_pw(pw_pool, tmp_r, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_transfer(vxc_g, tmp_r)
               CALL pw_axpy(tmp_r, vxc_rho(ispin)%pw)
               CALL pw_pool_give_back_pw(pw_pool, tmp_r)
               CALL pw_pool_give_back_pw(pw_pool, vxc_g)
            ELSE
               tmp_r => vxc_rho(ispin)%pw
               DO idir = 1, 3
                  SELECT CASE (xc_deriv_method_id)
                  CASE (xc_deriv_spline2_smooth)
                     CALL pw_nn_deriv_r(pw_in=pw_to_deriv(idir)%pw, &
                                        pw_out=tmp_r, coeffs=spline2_deriv_coeffs, &
                                        idir=idir)
                  CASE (xc_deriv_spline3_smooth)
                     CALL pw_nn_deriv_r(pw_in=pw_to_deriv(idir)%pw, &
                                        pw_out=tmp_r, coeffs=spline3_deriv_coeffs, &
                                        idir=idir)
                  CASE (xc_deriv_nn10_smooth)
                     CALL pw_nn_deriv_r(pw_in=pw_to_deriv(idir)%pw, &
                                        pw_out=tmp_r, coeffs=nn10_deriv_coeffs, &
                                        idir=idir)
                  CASE (xc_deriv_nn50_smooth)
                     CALL pw_nn_deriv_r(pw_in=pw_to_deriv(idir)%pw, &
                                        pw_out=tmp_r, coeffs=nn50_deriv_coeffs, &
                                        idir=idir)
                  CASE default
                     CALL cp__b("xc/xc.F",1593,"")
                  END SELECT
                  IF (dealloc_pw_to_deriv) THEN
                     CALL pw_pool_give_back_pw(pw_pool, pw_to_deriv(idir)%pw)
                  END IF
               END DO
               NULLIFY (tmp_r)
            END IF
         END IF

         ! ** Add laplace part to vxc_rho
         IF (has_laplace) THEN
            IF (use_virial) THEN
               CALL cp__b("xc/xc.F",1606,"Stress tensor not available for Laplace functionals.")
            END IF
            nd_laplace = RESHAPE((/2, 0, 0, 0, 2, 0, 0, 0, 2/), (/3, 3/))
            IF (lsd) THEN
               IF (ispin == 1) THEN
                  deriv_att => xc_dset_get_derivative(deriv_set, "(laplace_rhoa)")
                  IF(.NOT.(ASSOCIATED(deriv_att)))CALL cp__a("xc/xc.F",1612)
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
               ELSE
                  deriv_att => xc_dset_get_derivative(deriv_set, "(laplace_rhob)")
                  IF(.NOT.(ASSOCIATED(deriv_att)))CALL cp__a("xc/xc.F",1616)
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
               END IF

            ELSE
               deriv_att => xc_dset_get_derivative(deriv_set, "(laplace_rho)")
               IF(.NOT.(ASSOCIATED(deriv_att)))CALL cp__a("xc/xc.F",1622)
               CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
            END IF
            DO idir = 1, 3
               CALL pw_pool_create_pw(pw_pool, pw_to_deriv(idir)%pw, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_zero(pw_to_deriv(idir)%pw)
!$OMP          PARALLEL DO DEFAULT(NONE) SHARED(bo,deriv_data,&
!$OMP                         pw_to_deriv,idir)&
!$OMP                      PRIVATE(k,j,i)
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        pw_to_deriv(idir)%pw%cr3d(i, j, k) = deriv_data(i, j, k)
                     END DO
                  END DO
               END DO
               CALL pw_pool_create_pw(pw_pool, tmp_g, &
                                      use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
               CALL pw_zero(tmp_g)
               CALL pw_transfer(pw_to_deriv(idir)%pw, tmp_g)

               SELECT CASE (xc_deriv_method_id)
               CASE (xc_deriv_pw)
                  CALL pw_derive(tmp_g, nd_laplace(:, idir))
               CASE default
                  CALL cp__b("xc/xc.F",1648,"")
               END SELECT
               ! Add this to the potential
               CALL pw_pool_create_pw(pw_pool, tmp_r, &
                                      use_data=REALDATA3D, in_space=REALSPACE)
               CALL pw_zero(tmp_r)
               CALL pw_transfer(tmp_g, tmp_r)

               CALL pw_axpy(tmp_r, vxc_rho(ispin)%pw)
               CALL pw_pool_give_back_pw(pw_pool, tmp_r)
               CALL pw_pool_give_back_pw(pw_pool, pw_to_deriv(idir)%pw)
               CALL pw_pool_give_back_pw(pw_pool, tmp_g)
            END DO
         END IF

         IF (pw_grid%spherical) THEN
            ! filter vxc
            CALL pw_pool_create_pw(pw_pool, vxc_g, &
                                   use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
            CALL pw_transfer(vxc_rho(ispin)%pw, vxc_g)
            CALL pw_transfer(vxc_g, vxc_rho(ispin)%pw)
            CALL pw_pool_give_back_pw(pw_pool, vxc_g)
         END IF
         CALL smooth_cutoff(pot=vxc_rho(ispin)%pw%cr3d, rho=rho, rhoa=rhoa, rhob=rhob, &
                            rho_cutoff=rho_cutoff*density_smooth_cut_range, &
                            rho_smooth_cutoff_range=density_smooth_cut_range)

         ! final smoothing if rho was smoothed
         IF (xc_rho_smooth_id /= xc_rho_no_smooth) THEN
            CALL pw_pool_create_pw(pw_pool, tmp_r, &
                                   use_data=REALDATA3D, in_space=REALSPACE)
            CALL pw_zero(tmp_r)
            SELECT CASE (xc_rho_smooth_id)
            CASE (xc_rho_spline2_smooth)
               CALL pw_nn_smear_r(pw_in=vxc_rho(ispin)%pw, pw_out=tmp_r, &
                                  coeffs=spline2_coeffs)
            CASE (xc_rho_spline3_smooth)
               CALL pw_nn_smear_r(pw_in=vxc_rho(ispin)%pw, pw_out=tmp_r, &
                                  coeffs=spline3_coeffs)
            CASE (xc_rho_nn10)
               CALL pw_nn_smear_r(pw_in=vxc_rho(ispin)%pw, pw_out=tmp_r, &
                                  coeffs=nn10_coeffs)
            CASE (xc_rho_nn50)
               CALL pw_nn_smear_r(pw_in=vxc_rho(ispin)%pw, pw_out=tmp_r, &
                                  coeffs=nn50_coeffs)
            CASE default
               CALL cp__b("xc/xc.F",1694,"")
            END SELECT
            deriv_data => vxc_rho(ispin)%pw%cr3d
            vxc_rho(ispin)%pw%cr3d => tmp_r%cr3d
            tmp_r%cr3d => deriv_data
            CALL pw_pool_give_back_pw(pw_pool, tmp_r)
         END IF
      END DO

      DO idir = 1, 3
         IF(.NOT.(.NOT. ASSOCIATED(pw_to_deriv(idir)%pw)))CALL cp__a("xc/xc.F",1704)
         IF(.NOT.(.NOT. ASSOCIATED(pw_to_deriv_rho(idir)%pw)))CALL cp__a("xc/xc.F",1705)
      END DO

      ! 0-deriv -> value of exc
      ! this has to be kept consistent with xc_exc_calc
      IF (n_deriv > 0) THEN
         deriv_att => xc_dset_get_derivative(deriv_set, "")
         IF(.NOT.(ASSOCIATED(deriv_att)))CALL cp__a("xc/xc.F",1712)
         CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)

         CALL pw_create(tmp_r, pw_grid, &
                        use_data=REALDATA3D, in_space=REALSPACE, &
                        cr3d_ptr=deriv_data)
         NULLIFY (tmp_r%cr3d)
         CALL pw_release(tmp_r)

         CALL smooth_cutoff(pot=deriv_data, rho=rho, rhoa=rhoa, rhob=rhob, &
                            rho_cutoff=rho_cutoff, &
                            rho_smooth_cutoff_range=density_smooth_cut_range)

         exc = accurate_sum(deriv_data)*pw_grid%dvol
         IF (pw_grid%para%mode == PW_MODE_DISTRIBUTED) THEN
            CALL mp_sum(exc, pw_grid%para%group)
         END IF
      ELSE
         exc = 0.0_dp
      END IF

      CALL xc_rho_set_release(rho_set, pw_pool=pw_pool)

      ! tau part
      IF (has_tau) THEN
         IF (use_virial) THEN
            CALL cp__b("xc/xc.F",1738,"Stress tensor not available for Meta-functionals.")
         END IF
         ALLOCATE (vxc_tau(nspins))
         DO ispin = 1, nspins
            NULLIFY (vxc_tau(ispin)%pw)
         END DO
         IF (lsd) THEN
            deriv_att => xc_dset_get_derivative(deriv_set, "(tau_a)")
            IF (ASSOCIATED(deriv_att)) THEN
               CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)

               CALL pw_create(vxc_tau(1)%pw, pw_grid, &
                              use_data=REALDATA3D, in_space=REALSPACE, &
                              cr3d_ptr=deriv_data)
               NULLIFY (deriv_att%deriv_data)
            END IF
            deriv_att => xc_dset_get_derivative(deriv_set, "(tau_b)")
            IF (ASSOCIATED(deriv_att)) THEN
               CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)

               CALL pw_create(vxc_tau(2)%pw, pw_grid, &
                              use_data=REALDATA3D, in_space=REALSPACE, &
                              cr3d_ptr=deriv_data)
               NULLIFY (deriv_att%deriv_data)
            END IF
            deriv_att => xc_dset_get_derivative(deriv_set, "(tau)")
            IF (ASSOCIATED(deriv_att)) THEN
               CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
               IF (ASSOCIATED(vxc_tau(1)%pw)) THEN
                  DO ispin = 1, 2
                     IF(.NOT.(ASSOCIATED(vxc_tau(ispin)%pw)))CALL cp__a("xc/xc.F",1768)
                     tmp_cr3d => vxc_tau(ispin)%pw%cr3d
                     CALL daxpy(npoints, 1.0_dp, deriv_data, 1, tmp_cr3d, 1)
                  END DO
               ELSE
                  CALL pw_create(vxc_tau(1)%pw, pw_grid, &
                                 use_data=REALDATA3D, in_space=REALSPACE, &
                                 cr3d_ptr=deriv_data)
                  NULLIFY (deriv_att%deriv_data)
                  CALL pw_pool_create_pw(pw_pool, vxc_tau(2)%pw, &
                                         use_data=REALDATA3D, in_space=REALSPACE)
                  CALL pw_copy(vxc_tau(1)%pw, vxc_tau(2)%pw)
               END IF
            END IF
         ELSE
            deriv_att => xc_dset_get_derivative(deriv_set, "(tau)")
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
            CALL pw_create(vxc_tau(1)%pw, pw_grid, &
                           use_data=REALDATA3D, in_space=REALSPACE, &
                           cr3d_ptr=deriv_data)
            NULLIFY (deriv_att%deriv_data)
         END IF
         DO ispin = 1, nspins
            IF(.NOT.(ASSOCIATED(vxc_tau(ispin)%pw)))CALL cp__a("xc/xc.F",1791)
         END DO
      END IF
      CALL xc_dset_release(deriv_set)

      CALL timestop(handle)

   END SUBROUTINE xc_vxc_pw_create

! **************************************************************************************************
!> \brief calculates just the exchange and correlation energy
!>      (no vxc)
!> \param rho_r ...
!> \param rho_g ...
!> \param tau ...
!> \param xc_section ...
!> \param pw_pool ...
!> \retval exc ...
!> \par History
!>      11.2003 created [fawzi]
!> \author fawzi
!> \note
!>      has to be kept consistent with xc_vxc_create
! **************************************************************************************************
   FUNCTION xc_exc_calc(rho_r, rho_g, tau, xc_section, pw_pool) &
      RESULT(exc)
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r, rho_g, tau
      TYPE(section_vals_type), POINTER                   :: xc_section
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      REAL(kind=dp)                                      :: exc

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_exc_calc', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(dp)                                           :: density_smooth_cut_range, rho_cutoff
      REAL(dp), DIMENSION(:, :, :), POINTER              :: e_0
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_derivative_type), POINTER                  :: deriv
      TYPE(xc_rho_set_type), POINTER                     :: rho_set

      CALL timeset(routineN, handle)
      NULLIFY (rho_set, deriv_set, deriv, e_0)
      exc = 0.0_dp

      ! this has to be consistent with what is done in xc_vxc_create
      CALL xc_rho_set_and_dset_create(rho_set=rho_set, &
                                      deriv_set=deriv_set, deriv_order=0, &
                                      rho_r=rho_r, rho_g=rho_g, tau=tau, xc_section=xc_section, &
                                      pw_pool=pw_pool, &
                                      needs_basic_components=.FALSE.)
      deriv => xc_dset_get_derivative(deriv_set, "")

      IF (ASSOCIATED(deriv)) THEN
         CALL xc_derivative_get(deriv, deriv_data=e_0)

         CALL section_vals_val_get(xc_section, "DENSITY_CUTOFF", &
                                   r_val=rho_cutoff)
         CALL section_vals_val_get(xc_section, "DENSITY_SMOOTH_CUTOFF_RANGE", &
                                   r_val=density_smooth_cut_range)
         CALL smooth_cutoff(pot=e_0, rho=rho_set%rho, &
                            rhoa=rho_set%rhoa, rhob=rho_set%rhob, &
                            rho_cutoff=rho_cutoff, &
                            rho_smooth_cutoff_range=density_smooth_cut_range)

         exc = accurate_sum(e_0)*rho_r(1)%pw%pw_grid%dvol
         IF (rho_r(1)%pw%pw_grid%para%mode == PW_MODE_DISTRIBUTED) THEN
            CALL mp_sum(exc, rho_r(1)%pw%pw_grid%para%group)
         END IF

         CALL xc_rho_set_release(rho_set, pw_pool=pw_pool)
         CALL xc_dset_release(deriv_set)
      END IF
      CALL timestop(handle)
   END FUNCTION xc_exc_calc

! **************************************************************************************************
!> \brief Calculates the second derivative of E_xc at rho in the direction
!>      rho1  (if you see the second derivative as bilinear form)
!>      partial_rho|_(rho=rho) partial_rho|_(rho=rho) E_xc drho(rho1)drho
!>      The other direction is still indetermined, thus it returns
!>      a potential (partial integration is performed to reduce it to
!>      function of rho, removing the dependence from its partial derivs)
!>      Has to be called after the setup by xc_prep_2nd_deriv.
!> \param v_xc ...
!> \param deriv_set object containing the potentials
!> \param rho_set object containing the density at which the derivatives were calculated
!> \param rho1_set object containing the density with which to fold
!> \param pw_pool the pool for the grids
!> \param xc_section ...
!> \param gapw ...
!> \param vxg ...
!> \param tddfpt_fac factor that multiplies the crossterms (tddfpt triplets
!>        on a closed shell sistem it should be -1, defaults to 1)
!> \note
!>      The old version of this routine was smarter: it handled split_desc(1)
!>      and split_desc(2) separatly, thus the code automatically handled all
!>      possible cross terms (you only had to check if it was diagonal to avoid
!>      double counting). I think that is the way to go if you want to add more
!>      terms (tau,rho in LSD,...). The problem with the old code was that it
!>      because of the old functional structure it sometime guessed wrongly
!>      which derivative was where. There were probably still bugs with gradient
!>      corrected functionals (never tested), and it didn't contain first
!>      derivatives with respect to drho (that contibute also to the second
!>      derivative wrt. rho).
!>      The code was a little complex because it really tried to handle any
!>      functional derivative in the most efficent way with the given contents of
!>      rho_set.
!>      Anyway I strongly encourage whoever wants to modify this code to give a
!>      look to the old version. [fawzi]
! **************************************************************************************************
   SUBROUTINE xc_calc_2nd_deriv(v_xc, deriv_set, rho_set, rho1_set, &
                                pw_pool, xc_section, gapw, vxg, tddfpt_fac)

      TYPE(pw_p_type), DIMENSION(:), POINTER             :: v_xc
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_rho_set_type), POINTER                     :: rho_set, rho1_set
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(section_vals_type), POINTER                   :: xc_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: gapw
      REAL(kind=dp), DIMENSION(:, :, :, :), OPTIONAL, &
         POINTER                                         :: vxg
      REAL(kind=dp), INTENT(in), OPTIONAL                :: tddfpt_fac

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_calc_2nd_deriv', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=MAX_DERIVATIVE_DESC_LENGTH)          :: desc
      CHARACTER(len=MAX_LABEL_LENGTH), DIMENSION(:), &
         POINTER                                         :: split_desc
      INTEGER                                            :: handle, i, ia, idir, ir, ispin, j, k, &
                                                            nspins, order, stat, xc_deriv_method_id
      INTEGER, DIMENSION(2, 3)                           :: bo
      LOGICAL                                            :: gradient_f, lsd, my_gapw, unknown_deriv
      REAL(KIND=dp)                                      :: dr1dr, fac, gradient_cut
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: deriv_data, e_drhoa, e_drhob, &
                                                            e_norm_drho, rho1, rho1a, rho1b
      TYPE(cp_3d_r_p_type), DIMENSION(:), POINTER        :: drho, drho1, drho1a, drho1b, drhoa, drhob
      TYPE(cp_sll_xc_deriv_type), POINTER                :: pos
      TYPE(pw_p_type)                                    :: v_drho
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: tmp_a, tmp_b, tmp_r, tmp_r2
      TYPE(xc_derivative_type), POINTER                  :: deriv_att

! PARAMETER

      CALL timeset(routineN, handle)

      NULLIFY (tmp_r, tmp_r2, tmp_a, tmp_b, e_drhoa, e_drhob, e_norm_drho, &
               split_desc)

      my_gapw = .FALSE.
      IF (PRESENT(gapw)) my_gapw = gapw

      IF(.NOT.(ASSOCIATED(rho_set)))CALL cp__a("xc/xc.F",1943)
      IF(.NOT.(ASSOCIATED(rho1_set)))CALL cp__a("xc/xc.F",1944)
      IF(.NOT.(ASSOCIATED(v_xc)))CALL cp__a("xc/xc.F",1945)
      IF(.NOT.(ASSOCIATED(xc_section)))CALL cp__a("xc/xc.F",1946)
      IF (my_gapw) THEN
         IF(.NOT.(PRESENT(vxg)))CALL cp__a("xc/xc.F",1948)
      END IF

      CALL section_vals_val_get(xc_section, "XC_GRID%XC_DERIV", &
                                i_val=xc_deriv_method_id)
      CALL xc_rho_set_get(rho_set, drho_cutoff=gradient_cut)
      nspins = SIZE(v_xc)
      lsd = ASSOCIATED(rho_set%rhoa)
      fac = 0.0_dp
      IF (PRESENT(tddfpt_fac)) fac = tddfpt_fac

      ALLOCATE (tmp_r(nspins), tmp_r2(nspins), tmp_a(nspins), tmp_b(nspins))
      DO ispin = 1, nspins
         NULLIFY (tmp_r(ispin)%pw, tmp_r2(ispin)%pw, tmp_a(ispin)%pw, tmp_b(nspins)%pw)
      END DO

      bo = rho_set%local_bounds

      gradient_f = .FALSE.
      ! check for unknown derivatives
      pos => deriv_set%derivs
      DO WHILE (cp_sll_xc_deriv_next(pos, el_att=deriv_att))
         unknown_deriv = .FALSE.
         CALL xc_derivative_get(deriv_att, order=order, &
                                desc=desc, split_desc=split_desc)
         SELECT CASE (order)
         CASE (1)
            IF (lsd) THEN
               SELECT CASE (split_desc (1))
               CASE ("rho", "rhoa", "rhob")
               CASE ("norm_drho", "norm_drhoa", "norm_drhob")
                  gradient_f = .TRUE.
               CASE default
                  unknown_deriv = .TRUE.
               END SELECT
            ELSE
               SELECT CASE (split_desc (1))
               CASE ("rho")
               CASE ("norm_drho")
                  gradient_f = .TRUE.
               CASE default
                  unknown_deriv = .TRUE.
               END SELECT
            END IF
         CASE (2)
            IF (lsd) THEN
               SELECT CASE (split_desc (1))
               CASE ("rhoa", "rhob")
                  SELECT CASE (split_desc (2))
                  CASE ("rhoa", "rhob")
                  CASE ("norm_drhoa", "norm_drhob", "norm_drho")
                     gradient_f = .TRUE.
                  CASE default
                     unknown_deriv = .TRUE.
                  END SELECT
               CASE ("norm_drho", "norm_drhoa", "norm_drhob")
                  gradient_f = .TRUE.
                  SELECT CASE (split_desc (2))
                  CASE ("rhoa", "rhob", "norm_drhoa", "norm_drhob", "norm_drho")
                  CASE default
                     unknown_deriv = .TRUE.
                  END SELECT
               CASE default
                  unknown_deriv = .TRUE.
               END SELECT
            ELSE
               SELECT CASE (split_desc (1))
               CASE ("rho")
                  SELECT CASE (split_desc (2))
                  CASE ("rho")
                  CASE ("norm_drho")
                     gradient_f = .TRUE.
                  CASE default
                     unknown_deriv = .TRUE.
                  END SELECT
               CASE ("norm_drho")
                  gradient_f = .TRUE.
                  SELECT CASE (split_desc (2))
                  CASE ("rho", "norm_drho")
                  CASE default
                     unknown_deriv = .TRUE.
                  END SELECT
               CASE default
                  unknown_deriv = .TRUE.
               END SELECT
            END IF
         END SELECT
         IF (unknown_deriv) &
            CALL cp_abort(cp__l("xc/xc.F",2036), &
                          "unknown functional derivative (LSD="//TRIM(cp_to_string(lsd))// &
                          "): '"//TRIM(desc)//"'")
      END DO

      IF (lsd) THEN

         !-------------------!
         ! UNrestricted case !
         !-------------------!

         CALL xc_rho_set_get(rho1_set, rhoa=rho1a, rhob=rho1b)

         IF (gradient_f) THEN
            CALL xc_rho_set_get(rho_set, drhoa=drhoa, drhob=drhob)
            CALL xc_rho_set_get(rho1_set, drhoa=drho1a, drhob=drho1b)
            DO ispin = 1, nspins
               IF (ASSOCIATED(pw_pool)) THEN
                  CALL pw_pool_create_pw(pw_pool, tmp_a(ispin)%pw, &
                                         use_data=REALDATA3D, &
                                         in_space=REALSPACE)
                  CALL pw_zero(tmp_a(ispin)%pw)
                  CALL pw_pool_create_pw(pw_pool, tmp_b(ispin)%pw, &
                                         use_data=REALDATA3D, &
                                         in_space=REALSPACE)
                  CALL pw_zero(tmp_b(ispin)%pw)
                  CALL pw_pool_create_pw(pw_pool, tmp_r(ispin)%pw, &
                                         use_data=REALDATA3D, &
                                         in_space=REALSPACE)
                  CALL pw_zero(tmp_r(ispin)%pw)
               ELSE
                  ALLOCATE (tmp_a(ispin)%pw, tmp_b(ispin)%pw, tmp_r(ispin)%pw)
                  ALLOCATE (tmp_a(ispin)%pw%cr3d(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)), &
                            tmp_b(ispin)%pw%cr3d(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)), &
                            tmp_r(ispin)%pw%cr3d(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)), &
                            stat=stat)
                  IF(.NOT.(stat == 0))CALL cp__a("xc/xc.F",2072)
                  tmp_a(ispin)%pw%cr3d = 0.0_dp
                  tmp_b(ispin)%pw%cr3d = 0.0_dp
               END IF
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i) SHARED(v_xc,bo,rho1a,deriv_data) DEFAULT(NONE)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     v_xc(1)%pw%cr3d(i, j, k) = v_xc(1)%pw%cr3d(i, j, k)+ &
                                                deriv_data(i, j, k)*rho1a(i, j, k)
                  END DO
               END DO
            END DO
         END IF

         IF (nspins /= 1) THEN
            deriv_att => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)")
            IF (ASSOCIATED(deriv_att)) THEN
               CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP            PARALLEL DO PRIVATE(k,j,i) DEFAULT(NONE) SHARED(bo,v_xc,deriv_data,rho1b)
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        v_xc(2)%pw%cr3d(i, j, k) = v_xc(2)%pw%cr3d(i, j, k)+ &
                                                   deriv_data(i, j, k)*rho1b(i, j, k)
                     END DO
                  END DO
               END DO
            END IF
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(rhoa)(rhob)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i) DEFAULT(NONE) SHARED(bo,v_xc,deriv_data,rho1b,rho1a,nspins,fac)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     IF (nspins /= 1) THEN
                        v_xc(1)%pw%cr3d(i, j, k) = v_xc(1)%pw%cr3d(i, j, k)+ &
                                                   deriv_data(i, j, k)*rho1b(i, j, k)
                        v_xc(2)%pw%cr3d(i, j, k) = v_xc(2)%pw%cr3d(i, j, k)+ &
                                                   deriv_data(i, j, k)*rho1a(i, j, k)
                     ELSE
                        v_xc(1)%pw%cr3d(i, j, k) = v_xc(1)%pw%cr3d(i, j, k)+ &
                                                   fac*deriv_data(i, j, k)*rho1b(i, j, k)
                     END IF
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(rhoa)(norm_drhoa)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,dr1dr,idir) DEFAULT(NONE) SHARED(drhoa,drho1a,v_xc,tmp_a,bo,deriv_data,rho1a)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     dr1dr = 0._dp
                     DO idir = 1, 3
                        dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                     END DO
                     v_xc(1)%pw%cr3d(i, j, k) = v_xc(1)%pw%cr3d(i, j, k)+ &
                                                deriv_data(i, j, k)*dr1dr
                     tmp_a(1)%pw%cr3d(i, j, k) = tmp_a(1)%pw%cr3d(i, j, k)- &
                                                 deriv_data(i, j, k)*rho1a(i, j, k)
                  END DO
               END DO
            END DO
         END IF

         IF (nspins /= 1) THEN
            deriv_att => xc_dset_get_derivative(deriv_set, "(rhob)(norm_drhob)")
            IF (ASSOCIATED(deriv_att)) THEN
               CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP            PARALLEL DO PRIVATE(k,j,i,dr1dr,idir) DEFAULT(NONE) SHARED(v_xc,tmp_b,deriv_data,rho1b,bo,drhob,drho1b)
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        v_xc(2)%pw%cr3d(i, j, k) = v_xc(2)%pw%cr3d(i, j, k)+ &
                                                   deriv_data(i, j, k)*dr1dr
                        tmp_b(2)%pw%cr3d(i, j, k) = tmp_b(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*rho1b(i, j, k)
                     END DO
                  END DO
               END DO
            END IF
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(rhoa)(norm_drhob)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,dr1dr,idir) DEFAULT(NONE) SHARED(bo,drhob,drho1b,v_xc,tmp_b,deriv_data,nspins,rho1a,fac)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     dr1dr = 0._dp
                     DO idir = 1, 3
                        dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                     END DO
                     IF (nspins /= 1) THEN
                        v_xc(1)%pw%cr3d(i, j, k) = v_xc(1)%pw%cr3d(i, j, k)+ &
                                                   deriv_data(i, j, k)*dr1dr
                        tmp_b(2)%pw%cr3d(i, j, k) = tmp_b(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*rho1a(i, j, k)
                     ELSE
                        v_xc(1)%pw%cr3d(i, j, k) = v_xc(1)%pw%cr3d(i, j, k)+ &
                                                   fac*deriv_data(i, j, k)*dr1dr
                     END IF
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(rhob)(norm_drhoa)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,dr1dr,idir) DEFAULT(NONE) SHARED(bo,nspins,drhoa,drho1a,v_xc,tmp_a,fac,deriv_data,rho1b)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     IF (nspins /= 1) THEN
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                        END DO
                        v_xc(2)%pw%cr3d(i, j, k) = v_xc(2)%pw%cr3d(i, j, k)+ &
                                                   deriv_data(i, j, k)*dr1dr
                        tmp_a(1)%pw%cr3d(i, j, k) = tmp_a(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*rho1b(i, j, k)
                     ELSE
                        tmp_a(1)%pw%cr3d(i, j, k) = tmp_a(1)%pw%cr3d(i, j, k)- &
                                                    fac*deriv_data(i, j, k)*rho1b(i, j, k)
                     END IF
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(rhoa)(norm_drho)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,dr1dr,idir) DEFAULT(NONE)&
!$OMP           SHARED(bo,drhob,drho1a,deriv_data,v_xc,drhoa,drho1b,tmp_a,tmp_b,nspins,rho1a,fac)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     IF (nspins /= 1) THEN
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                        END DO
                        v_xc(1)%pw%cr3d(i, j, k) = v_xc(1)%pw%cr3d(i, j, k)+ &
                                                   deriv_data(i, j, k)*dr1dr
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        v_xc(1)%pw%cr3d(i, j, k) = v_xc(1)%pw%cr3d(i, j, k)+ &
                                                   deriv_data(i, j, k)*dr1dr
                        tmp_b(1)%pw%cr3d(i, j, k) = tmp_b(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*rho1a(i, j, k)
                        tmp_a(2)%pw%cr3d(i, j, k) = tmp_a(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*rho1a(i, j, k)
                     ELSE
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)+ &
                                   fac*drhoa(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        v_xc(1)%pw%cr3d(i, j, k) = v_xc(1)%pw%cr3d(i, j, k)+ &
                                                   deriv_data(i, j, k)*dr1dr
                        tmp_b(1)%pw%cr3d(i, j, k) = tmp_b(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*rho1a(i, j, k)
                     END IF
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(norm_drho)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,dr1dr,idir) DEFAULT(NONE) &
!$OMP         SHARED(bo,nspins,drhob,drho1a,deriv_data,drhoa,drho1b,tmp_a,tmp_b,fac)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     IF (nspins /= 1) THEN
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                        END DO
                        tmp_a(1)%pw%cr3d(i, j, k) = tmp_a(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        tmp_a(1)%pw%cr3d(i, j, k) = tmp_a(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                        END DO
                        tmp_b(1)%pw%cr3d(i, j, k) = tmp_b(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                        END DO
                        tmp_a(2)%pw%cr3d(i, j, k) = tmp_a(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                     ELSE
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)+ &
                                   fac*drhoa(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        tmp_a(1)%pw%cr3d(i, j, k) = tmp_a(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                        END DO
                        tmp_b(1)%pw%cr3d(i, j, k) = tmp_b(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                     END IF
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(rhob)(norm_drho)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,dr1dr,idir)  DEFAULT(NONE) &
!$OMP         SHARED(bo,nspins,drhob,drho1a,deriv_data,drhoa,drho1b,tmp_a,tmp_b,fac,v_xc,rho1b)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     IF (nspins /= 1) THEN
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        v_xc(2)%pw%cr3d(i, j, k) = v_xc(2)%pw%cr3d(i, j, k)+ &
                                                   deriv_data(i, j, k)*dr1dr
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                        END DO
                        v_xc(2)%pw%cr3d(i, j, k) = v_xc(2)%pw%cr3d(i, j, k)+ &
                                                   deriv_data(i, j, k)*dr1dr
                        tmp_a(2)%pw%cr3d(i, j, k) = tmp_a(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*rho1b(i, j, k)
                        tmp_b(1)%pw%cr3d(i, j, k) = tmp_b(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*rho1b(i, j, k)
                     ELSE
                        tmp_b(1)%pw%cr3d(i, j, k) = tmp_b(1)%pw%cr3d(i, j, k)- &
                                                    fac*deriv_data(i, j, k)*rho1b(i, j, k)
                     END IF
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drhob)(norm_drho)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,dr1dr,idir)  DEFAULT(NONE) &
!$OMP         SHARED(bo,nspins,drhob,drho1a,deriv_data,drhoa,drho1b,tmp_a,tmp_b,fac)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     IF (nspins /= 1) THEN
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        tmp_b(2)%pw%cr3d(i, j, k) = tmp_b(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                        END DO
                        tmp_b(2)%pw%cr3d(i, j, k) = tmp_b(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        tmp_a(2)%pw%cr3d(i, j, k) = tmp_a(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        tmp_b(1)%pw%cr3d(i, j, k) = tmp_b(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                     ELSE
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        tmp_b(1)%pw%cr3d(i, j, k) = tmp_b(1)%pw%cr3d(i, j, k)- &
                                                    fac*deriv_data(i, j, k)*dr1dr
                     END IF
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(norm_drhoa)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,dr1dr,idir)  DEFAULT(NONE) &
!$OMP         SHARED(bo,nspins,drhoa,drho1a,deriv_data,tmp_a)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     dr1dr = 0._dp
                     DO idir = 1, 3
                        dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                     END DO
                     tmp_a(1)%pw%cr3d(i, j, k) = tmp_a(1)%pw%cr3d(i, j, k)- &
                                                 deriv_data(i, j, k)*dr1dr
                  END DO
               END DO
            END DO
         END IF

         IF (nspins /= 1) THEN
            deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drhob)(norm_drhob)")
            IF (ASSOCIATED(deriv_att)) THEN
               CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP            PARALLEL DO PRIVATE(k,j,i,dr1dr,idir) DEFAULT(NONE)&
!$OMP            SHARED(drhob,drho1b,tmp_b,deriv_data,bo)
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        tmp_b(2)%pw%cr3d(i, j, k) = tmp_b(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                     END DO
                  END DO
               END DO
            END IF
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drhoa)(norm_drhob)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,dr1dr,idir)  DEFAULT(NONE) &
!$OMP         SHARED(bo,nspins,drhob,drho1a,deriv_data,drhoa,drho1b,tmp_a,tmp_b,fac)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     dr1dr = 0._dp
                     DO idir = 1, 3
                        dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                     END DO
                     IF (nspins /= 1) THEN
                        tmp_a(1)%pw%cr3d(i, j, k) = tmp_a(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                        END DO
                        tmp_b(2)%pw%cr3d(i, j, k) = tmp_b(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                     ELSE
                        tmp_a(1)%pw%cr3d(i, j, k) = tmp_a(1)%pw%cr3d(i, j, k)- &
                                                    fac*deriv_data(i, j, k)*dr1dr
                     END IF
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drhoa)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
            CALL xc_derivative_get(deriv_att, deriv_data=e_drhoa)
!$OMP         PARALLEL DO PRIVATE(k,j,i,dr1dr,idir) DEFAULT(NONE)&
!$OMP         SHARED(bo,drhoa,drho1a,tmp_a,deriv_data)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     dr1dr = 0._dp
                     DO idir = 1, 3
                        dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                     END DO
                     tmp_a(1)%pw%cr3d(i, j, k) = tmp_a(1)%pw%cr3d(i, j, k)+ &
                                                 deriv_data(i, j, k)*dr1dr
                  END DO
               END DO
            END DO
         END IF

         IF (nspins /= 1) THEN
            deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drhob)")
            IF (ASSOCIATED(deriv_att)) THEN
               CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
               CALL xc_derivative_get(deriv_att, deriv_data=e_drhob)
!$OMP            PARALLEL DO PRIVATE(k,j,i,dr1dr,idir) DEFAULT(NONE) &
!$OMP            SHARED(bo,drhob,drho1b,tmp_b,deriv_data)
               DO k = bo(1, 3), bo(2, 3)
                  DO j = bo(1, 2), bo(2, 2)
                     DO i = bo(1, 1), bo(2, 1)
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        tmp_b(2)%pw%cr3d(i, j, k) = tmp_b(2)%pw%cr3d(i, j, k)+ &
                                                    deriv_data(i, j, k)*dr1dr
                     END DO
                  END DO
               END DO
            END IF
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drho)(norm_drho)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,dr1dr,idir) DEFAULT(NONE)&
!$OMP         SHARED(bo,nspins,drhob,drho1a,drhoa,drho1b,tmp_a,tmp_b,fac,deriv_data)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     IF (nspins /= 1) THEN
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)
                        END DO
                        tmp_b(1)%pw%cr3d(i, j, k) = tmp_b(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                        tmp_a(2)%pw%cr3d(i, j, k) = tmp_a(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhoa(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        tmp_b(1)%pw%cr3d(i, j, k) = tmp_b(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                        tmp_a(2)%pw%cr3d(i, j, k) = tmp_a(2)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                     ELSE
                        dr1dr = 0._dp
                        DO idir = 1, 3
                           dr1dr = dr1dr+drhob(idir)%array(i, j, k)*drho1a(idir)%array(i, j, k)+ &
                                   fac*drhoa(idir)%array(i, j, k)*drho1b(idir)%array(i, j, k)
                        END DO
                        tmp_b(1)%pw%cr3d(i, j, k) = tmp_b(1)%pw%cr3d(i, j, k)- &
                                                    deriv_data(i, j, k)*dr1dr
                     END IF
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drho)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=e_norm_drho)
         END IF

         IF (gradient_f) THEN

            IF (my_gapw) THEN
!$OMP            PARALLEL DO PRIVATE(ia,idir,ispin,ir) DEFAULT(NONE) &
!$OMP            SHARED(bo,nspins,vxg,drhoa,drhob,tmp_a,tmp_b,e_drhoa,e_drhob,e_norm_drho,drho1a,drho1b,fac)
               DO ir = bo(1, 2), bo(2, 2)
                  DO ia = bo(1, 1), bo(2, 1)
                     DO idir = 1, 3
                        DO ispin = 1, nspins
                           vxg(idir, ia, ir, ispin) = &
                              tmp_a(ispin)%pw%cr3d(ia, ir, 1)*drhoa(idir)%array(ia, ir, 1)+ &
                              tmp_b(ispin)%pw%cr3d(ia, ir, 1)*drhob(idir)%array(ia, ir, 1)
                        END DO
                        IF (ASSOCIATED(e_drhoa)) THEN
                           vxg(idir, ia, ir, 1) = vxg(idir, ia, ir, 1)- &
                                                  e_drhoa(ia, ir, 1)*drho1a(idir)%array(ia, ir, 1)
                        END IF
                        IF (nspins /= 1 .AND. ASSOCIATED(e_drhob)) THEN
                           vxg(idir, ia, ir, 2) = vxg(idir, ia, ir, 2)- &
                                                  e_drhob(ia, ir, 1)*drho1b(idir)%array(ia, ir, 1)
                        END IF
                        IF (ASSOCIATED(e_norm_drho)) THEN
                           IF (nspins /= 1) THEN
                              vxg(idir, ia, ir, 1) = vxg(idir, ia, ir, 1)- &
                                                     e_norm_drho(ia, ir, 1)*drho1b(idir)%array(ia, ir, 1)
                              vxg(idir, ia, ir, 2) = vxg(idir, ia, ir, 2)- &
                                                     e_norm_drho(ia, ir, 1)*drho1a(idir)%array(ia, ir, 1)
                           ELSE
                              vxg(idir, ia, ir, 1) = vxg(idir, ia, ir, 1)- &
                                                     fac*e_norm_drho(ia, ir, 1)*drho1b(idir)%array(ia, ir, 1)
                           END IF
                        END IF
                     END DO
                  END DO
               END DO
            ELSE

               ! partial integration
               DO idir = 1, 3

                  DO ispin = 1, nspins
!$OMP                  PARALLEL DO PRIVATE(k,j,i,dr1dr) DEFAULT(NONE) &
!$OMP                  SHARED(bo,tmp_r,tmp_a,tmp_b,drhoa,drhob,ispin,idir)
                     DO k = bo(1, 3), bo(2, 3)
                        DO j = bo(1, 2), bo(2, 2)
                           DO i = bo(1, 1), bo(2, 1)
                              tmp_r(ispin)%pw%cr3d(i, j, k) = &
                                 tmp_a(ispin)%pw%cr3d(i, j, k)*drhoa(idir)%array(i, j, k)+ &
                                 tmp_b(ispin)%pw%cr3d(i, j, k)*drhob(idir)%array(i, j, k)
                           END DO
                        END DO
                     END DO
                     IF (ASSOCIATED(e_drhoa)) THEN
!$OMP                     PARALLEL DO PRIVATE(k,j,i) DEFAULT(NONE) &
!$OMP                     SHARED(bo,tmp_r,e_drhoa,drho1a,idir)
                        DO k = bo(1, 3), bo(2, 3)
                           DO j = bo(1, 2), bo(2, 2)
                              DO i = bo(1, 1), bo(2, 1)
                                 tmp_r(1)%pw%cr3d(i, j, k) = tmp_r(1)%pw%cr3d(i, j, k)- &
                                                             e_drhoa(i, j, k)*drho1a(idir)%array(i, j, k)
                              END DO
                           END DO
                        END DO
                     END IF
                     IF (nspins /= 1 .AND. ASSOCIATED(e_drhob)) THEN
!$OMP                     PARALLEL DO PRIVATE(k,j,i) DEFAULT(NONE)&
!$OMP                     SHARED(bo,tmp_r,e_drhob,drho1b,idir)
                        DO k = bo(1, 3), bo(2, 3)
                           DO j = bo(1, 2), bo(2, 2)
                              DO i = bo(1, 1), bo(2, 1)
                                 tmp_r(2)%pw%cr3d(i, j, k) = tmp_r(2)%pw%cr3d(i, j, k)- &
                                                             e_drhob(i, j, k)*drho1b(idir)%array(i, j, k)
                              END DO
                           END DO
                        END DO
                     END IF
                     IF (ASSOCIATED(e_norm_drho)) THEN
!$OMP                     PARALLEL DO PRIVATE(k,j,i) DEFAULT(NONE)&
!$OMP                     SHARED(bo,tmp_r,e_norm_drho,drho1b,drho1a,fac,idir,nspins)
                        DO k = bo(1, 3), bo(2, 3)
                           DO j = bo(1, 2), bo(2, 2)
                              DO i = bo(1, 1), bo(2, 1)
                                 IF (nspins /= 1) THEN
                                    tmp_r(1)%pw%cr3d(i, j, k) = tmp_r(1)%pw%cr3d(i, j, k)- &
                                                                e_norm_drho(i, j, k)*drho1b(idir)%array(i, j, k)
                                    tmp_r(2)%pw%cr3d(i, j, k) = tmp_r(2)%pw%cr3d(i, j, k)- &
                                                                e_norm_drho(i, j, k)*drho1a(idir)%array(i, j, k)
                                 ELSE
                                    tmp_r(1)%pw%cr3d(i, j, k) = tmp_r(1)%pw%cr3d(i, j, k)- &
                                                                fac*e_norm_drho(i, j, k)*drho1b(idir)%array(i, j, k)
                                 END IF
                              END DO
                           END DO
                        END DO
                     END IF
                  END DO

                  DO ispin = 1, nspins
                     SELECT CASE (xc_deriv_method_id)
                     CASE (xc_deriv_spline2_smooth)
                        CALL pw_nn_deriv_r(pw_in=tmp_r(ispin)%pw, &
                                           pw_out=v_xc(ispin)%pw, coeffs=spline2_deriv_coeffs, &
                                           idir=idir)
                     CASE (xc_deriv_spline3_smooth)
                        CALL pw_nn_deriv_r(pw_in=tmp_r(ispin)%pw, &
                                           pw_out=v_xc(ispin)%pw, coeffs=spline3_deriv_coeffs, &
                                           idir=idir)
                     CASE (xc_deriv_nn10_smooth)
                        CALL pw_nn_deriv_r(pw_in=tmp_r(ispin)%pw, &
                                           pw_out=v_xc(ispin)%pw, coeffs=nn10_deriv_coeffs, &
                                           idir=idir)
                     CASE (xc_deriv_nn50_smooth)
                        CALL pw_nn_deriv_r(pw_in=tmp_r(ispin)%pw, &
                                           pw_out=v_xc(ispin)%pw, coeffs=nn50_deriv_coeffs, &
                                           idir=idir)
                     CASE default
                        CALL cp__b("xc/xc.F",2668,"XC_DERIV method not implemented for GPW-LSD!")
                     END SELECT
                  END DO ! ispin

               END DO ! idir

            END IF

            DO ispin = 1, nspins
               IF (ASSOCIATED(pw_pool)) THEN
                  CALL pw_pool_give_back_pw(pw_pool, tmp_a(ispin)%pw)
                  CALL pw_pool_give_back_pw(pw_pool, tmp_b(ispin)%pw)
                  CALL pw_pool_give_back_pw(pw_pool, tmp_r(ispin)%pw)
               ELSE
                  DEALLOCATE (tmp_a(ispin)%pw%cr3d, tmp_b(ispin)%pw%cr3d, tmp_r(ispin)%pw%cr3d)
                  DEALLOCATE (tmp_a(ispin)%pw, tmp_b(ispin)%pw, tmp_r(ispin)%pw)
               END IF
            END DO

         END IF ! gradient_f

      ELSE

         !-----------------!
         ! restricted case !
         !-----------------!

         CALL xc_rho_set_get(rho1_set, rho=rho1)

         IF (gradient_f) THEN
            CALL xc_rho_set_get(rho_set, drho=drho)
            CALL xc_rho_set_get(rho1_set, drho=drho1)
            IF (ASSOCIATED(pw_pool)) THEN
               CALL pw_pool_create_pw(pw_pool, v_drho%pw, &
                                      use_data=REALDATA3D, &
                                      in_space=REALSPACE)
               CALL pw_zero(v_drho%pw)
            ELSE
               ALLOCATE (v_drho%pw)
               ALLOCATE (v_drho%pw%cr3d(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
               v_drho%pw%cr3d = 0.0_dp
            END IF
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(rho)(rho)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i) DEFAULT(NONE) &
!$OMP         SHARED(bo,v_xc,deriv_data,rho1)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     v_xc(1)%pw%cr3d(i, j, k) = &
                        deriv_data(i, j, k)*rho1(i, j, k)
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(rho)(norm_drho)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,idir,dr1dr) DEFAULT(NONE)&
!$OMP         SHARED(bo,drho,drho1,v_xc,deriv_data,v_drho,rho1)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     dr1dr = 0._dp
                     DO idir = 1, 3
                        dr1dr = dr1dr+drho(idir)%array(i, j, k)*drho1(idir)%array(i, j, k)
                     END DO
                     v_xc(1)%pw%cr3d(i, j, k) = v_xc(1)%pw%cr3d(i, j, k)+ &
                                                deriv_data(i, j, k)*dr1dr
                     v_drho%pw%cr3d(i, j, k) = -1._dp*deriv_data(i, j, k)*rho1(i, j, k)
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drho)(norm_drho)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
!$OMP         PARALLEL DO PRIVATE(k,j,i,idir,dr1dr) DEFAULT(NONE)&
!$OMP         SHARED(bo,drho,drho1,v_drho,deriv_data)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     dr1dr = 0._dp
                     DO idir = 1, 3
                        dr1dr = dr1dr+drho(idir)%array(i, j, k)*drho1(idir)%array(i, j, k)
                     END DO
                     v_drho%pw%cr3d(i, j, k) = v_drho%pw%cr3d(i, j, k)-deriv_data(i, j, k)*dr1dr
                  END DO
               END DO
            END DO
         END IF

         deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drho)")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
            CALL xc_derivative_get(deriv_att, deriv_data=e_norm_drho)
!$OMP         PARALLEL DO PRIVATE(k,j,i,idir,dr1dr) DEFAULT(NONE)&
!$OMP         SHARED(bo,drho,drho1,gradient_cut,rho_set,v_drho,deriv_data)
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     dr1dr = 0._dp
                     DO idir = 1, 3
                        dr1dr = dr1dr+drho(idir)%array(i, j, k)*drho1(idir)%array(i, j, k)
                     END DO
                     IF (rho_set%norm_drho(i, j, k) > gradient_cut) THEN
                        dr1dr = dr1dr/(rho_set%norm_drho(i, j, k))**2
                        v_drho%pw%cr3d(i, j, k) = v_drho%pw%cr3d(i, j, k)+deriv_data(i, j, k)*dr1dr
                     END IF
                  END DO
               END DO
            END DO
         END IF

         IF (gradient_f) THEN

            IF (my_gapw) THEN

               DO idir = 1, 3
!$OMP               PARALLEL DO PRIVATE(ia,ir) DEFAULT(NONE) &
!$OMP               SHARED(bo,vxg,drho,v_drho,e_norm_drho,drho1,idir)
                  DO ia = bo(1, 1), bo(2, 1)
                     DO ir = bo(1, 2), bo(2, 2)
                        vxg(idir, ia, ir, 1) = drho(idir)%array(ia, ir, 1)*v_drho%pw%cr3d(ia, ir, 1)
                        IF (ASSOCIATED(e_norm_drho)) THEN
                           vxg(idir, ia, ir, 1) = vxg(idir, ia, ir, 1)-drho1(idir)%array(ia, ir, 1)*e_norm_drho(ia, ir, 1)
                        END IF
                     END DO
                  END DO
               END DO

            ELSE
               ! partial integration

               ! this does not work with non orthorombic cells
               ! (you will have to use a vector of pw with 3 components)
               IF (ASSOCIATED(pw_pool)) THEN
                  CALL pw_pool_create_pw(pw_pool, tmp_r(1)%pw, &
                                         use_data=REALDATA3D, &
                                         in_space=REALSPACE)
               ELSE
                  ALLOCATE (tmp_r(1)%pw)
                  ALLOCATE (tmp_r(1)%pw%cr3d(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)), &
                            stat=stat)
                  IF(.NOT.(stat == 0))CALL cp__a("xc/xc.F",2817)
               END IF

               DO idir = 1, 3
!$OMP               PARALLEL DO PRIVATE(k,j,i) DEFAULT(NONE)&
!$OMP               SHARED(bo,tmp_r,drho,v_drho,drho1,deriv_data,idir)
                  DO k = bo(1, 3), bo(2, 3)
                     DO j = bo(1, 2), bo(2, 2)
                        DO i = bo(1, 1), bo(2, 1)
                           tmp_r(1)%pw%cr3d(i, j, k) = drho(idir)%array(i, j, k)*v_drho%pw%cr3d(i, j, k)- &
                                                       drho1(idir)%array(i, j, k)*deriv_data(i, j, k)
                        END DO
                     END DO
                  END DO

                  SELECT CASE (xc_deriv_method_id)
                  CASE (xc_deriv_spline2_smooth)
                     CALL pw_nn_deriv_r(pw_in=tmp_r(1)%pw, &
                                        pw_out=v_xc(1)%pw, coeffs=spline2_deriv_coeffs, &
                                        idir=idir)
                  CASE (xc_deriv_spline3_smooth)
                     CALL pw_nn_deriv_r(pw_in=tmp_r(1)%pw, &
                                        pw_out=v_xc(1)%pw, coeffs=spline3_deriv_coeffs, &
                                        idir=idir)
                  CASE (xc_deriv_nn10_smooth)
                     CALL pw_nn_deriv_r(pw_in=tmp_r(1)%pw, &
                                        pw_out=v_xc(1)%pw, coeffs=nn10_deriv_coeffs, &
                                        idir=idir)
                  CASE (xc_deriv_nn50_smooth)
                     CALL pw_nn_deriv_r(pw_in=tmp_r(1)%pw, &
                                        pw_out=v_xc(1)%pw, coeffs=nn50_deriv_coeffs, &
                                        idir=idir)
                  CASE default
                     CALL cp__b("xc/xc.F",2850,"XC_DERIV method not implemented for GPW!")
                  END SELECT

               END DO

               IF (ASSOCIATED(pw_pool)) THEN
                  CALL pw_pool_give_back_pw(pw_pool, tmp_r(1)%pw)
               ELSE
                  DEALLOCATE (tmp_r(1)%pw%cr3d)
                  DEALLOCATE (tmp_r(1)%pw)
               END IF
            END IF
            IF (ASSOCIATED(pw_pool)) THEN
               CALL pw_pool_give_back_pw(pw_pool, v_drho%pw)
            ELSE
               DEALLOCATE (v_drho%pw%cr3d)
               DEALLOCATE (v_drho%pw)
            END IF

         END IF

      END IF

      DEALLOCATE (tmp_r, tmp_r2, tmp_a, tmp_b)

      CALL timestop(handle)
   END SUBROUTINE xc_calc_2nd_deriv

! **************************************************************************************************
!> \brief Does the first prepartions for the calculation of the 2nd deriv
!>      The calculation must then be performed with xc_calc_2nd_deriv
!>      Calculates the second derivative of E_xc at rho in the direction
!>      rho1  (if you see the second derivative as bilinear form)
!>      partial_rho|_(rho=rho) partial_rho|_(rho=rho) E_xc drho(rho1)drho
!>      The other direction is still indetermined, thus it returns
!>      a potential (partial integration is performed to reduce it to
!>      function of rho, removing the dependence from its partial derivs)
!> \param deriv_set object containing the potentials
!> \param rho_set object that will contain the density at which the
!>        derivatives were calculated
!> \param rho_r the place where you evaluate the derivative
!> \param pw_pool the pool for the grids
!> \param xc_section which functional should be used and how to calculate it
! **************************************************************************************************
   SUBROUTINE xc_prep_2nd_deriv(deriv_set, &
                                rho_set, rho_r, pw_pool, xc_section)

      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_r
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(section_vals_type), POINTER                   :: xc_section

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_prep_2nd_deriv', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspins
      LOGICAL                                            :: lsd
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g, rho_r_pw, tau

      CALL timeset(routineN, handle)

      IF(.NOT.(.NOT. ASSOCIATED(rho_set)))CALL cp__a("xc/xc.F",2912)
      IF(.NOT.(ASSOCIATED(xc_section)))CALL cp__a("xc/xc.F",2913)
      IF(.NOT.(ASSOCIATED(pw_pool)))CALL cp__a("xc/xc.F",2914)

      nspins = SIZE(rho_r)
      lsd = (nspins /= 1)

      ALLOCATE (rho_r_pw(nspins))
      DO ispin = 1, nspins
         rho_r_pw(ispin)%pw => rho_r(ispin)%pw
      END DO

      NULLIFY (rho_g, tau)
      CALL xc_rho_set_and_dset_create(rho_set, deriv_set, 2, &
                                      rho_r_pw, rho_g, tau, xc_section, pw_pool, &
                                      needs_basic_components=.TRUE.)

      DEALLOCATE (rho_r_pw)

      CALL divide_by_norm_drho(deriv_set, rho_set, lsd)

      CALL timestop(handle)

   END SUBROUTINE xc_prep_2nd_deriv

! **************************************************************************************************
!> \brief ...
!> \param deriv_set ...
!> \param rho_set ...
!> \param lsd ...
! **************************************************************************************************
   SUBROUTINE divide_by_norm_drho(deriv_set, rho_set, lsd)

      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      LOGICAL, INTENT(IN)                                :: lsd

      CHARACTER(len=*), PARAMETER :: routineN = 'divide_by_norm_drho', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=MAX_DERIVATIVE_DESC_LENGTH)          :: desc
      CHARACTER(len=MAX_LABEL_LENGTH), DIMENSION(:), &
         POINTER                                         :: split_desc
      INTEGER                                            :: i, idesc, j, k, order
      INTEGER, DIMENSION(2, 3)                           :: bo
      REAL(KIND=dp)                                      :: drho_cutoff
      TYPE(cp_sll_xc_deriv_type), POINTER                :: pos
      TYPE(xc_derivative_type), POINTER                  :: deriv_att

! check for unknown derivatives and divide by norm_drho where necessary

      bo = rho_set%local_bounds
      CALL xc_rho_set_get(rho_set, drho_cutoff=drho_cutoff)

      pos => deriv_set%derivs
      DO WHILE (cp_sll_xc_deriv_next(pos, el_att=deriv_att))
         CALL xc_derivative_get(deriv_att, order=order, &
                                desc=desc, split_desc=split_desc)
         IF (order == 1 .OR. order == 2) THEN
            DO idesc = 1, SIZE(split_desc)
               SELECT CASE (split_desc (idesc))
               CASE ("norm_drho")
!$OMP                 PARALLEL DO PRIVATE(i,j,k) DEFAULT(NONE) SHARED(bo,deriv_att,rho_set,drho_cutoff)
                  DO k = bo(1, 3), bo(2, 3)
                     DO j = bo(1, 2), bo(2, 2)
                        DO i = bo(1, 1), bo(2, 1)
                           deriv_att%deriv_data(i, j, k) = deriv_att%deriv_data(i, j, k)/ &
                                                           MAX(rho_set%norm_drho(i, j, k), drho_cutoff)
                        END DO
                     END DO
                  END DO
               CASE ("norm_drhoa")
!$OMP                 PARALLEL DO PRIVATE(i,j,k) DEFAULT(NONE) SHARED(bo,deriv_att,rho_set,drho_cutoff)
                  DO k = bo(1, 3), bo(2, 3)
                     DO j = bo(1, 2), bo(2, 2)
                        DO i = bo(1, 1), bo(2, 1)
                           deriv_att%deriv_data(i, j, k) = deriv_att%deriv_data(i, j, k)/ &
                                                           MAX(rho_set%norm_drhoa(i, j, k), drho_cutoff)
                        END DO
                     END DO
                  END DO
               CASE ("norm_drhob")
!$OMP                 PARALLEL DO PRIVATE(i,j,k) DEFAULT(NONE) SHARED(bo,deriv_att,rho_set,drho_cutoff)
                  DO k = bo(1, 3), bo(2, 3)
                     DO j = bo(1, 2), bo(2, 2)
                        DO i = bo(1, 1), bo(2, 1)
                           deriv_att%deriv_data(i, j, k) = deriv_att%deriv_data(i, j, k)/ &
                                                           MAX(rho_set%norm_drhob(i, j, k), drho_cutoff)
                        END DO
                     END DO
                  END DO
               CASE ("rho")
                  IF (lsd) &
                     CALL cp__b("xc/xc.F",3005,"rho not handled in lsd: '"//TRIM(desc)//"'")
               CASE ("rhoa", "rhob")
               CASE default
                  CALL cp_abort(cp__l("xc/xc.F",3008), &
                                "unhandled derivative: '"// &
                                TRIM(split_desc(idesc))//"' in '"// &
                                TRIM(desc)//"'")
               END SELECT
            END DO
         END IF
      END DO

   END SUBROUTINE divide_by_norm_drho

END MODULE xc

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE xc_atom

   USE cp_linked_list_xc_deriv,         ONLY: cp_sll_xc_deriv_next,&
                                              cp_sll_xc_deriv_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE pw_pool_types,                   ONLY: pw_pool_type
   USE pw_types,                        ONLY: pw_p_type
   USE xc,                              ONLY: divide_by_norm_drho,&
                                              xc_calc_2nd_deriv
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_get_derivative
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_derivatives,                  ONLY: xc_functionals_eval
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_get,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_atom'

   PUBLIC :: vxc_of_r_new, xc_rho_set_atom_update, xc_2nd_deriv_of_r, fill_rho_set

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param xc_fun_section ...
!> \param rho_set ...
!> \param deriv_set ...
!> \param deriv_order ...
!> \param needs ...
!> \param w ...
!> \param lsd ...
!> \param na ...
!> \param nr ...
!> \param exc ...
!> \param vxc ...
!> \param vxg ...
!> \param vtau ...
!> \param energy_only ...
!> \param epr_xc ...
!> \param adiabatic_rescale_factor ...
! **************************************************************************************************
   SUBROUTINE vxc_of_r_new(xc_fun_section, rho_set, deriv_set, deriv_order, needs, w, &
                           lsd, na, nr, exc, vxc, vxg, vtau, energy_only, epr_xc, adiabatic_rescale_factor)

! This routine updates rho_set by giving to it the rho and drho that are needed.
! Since for the local densities rho1_h and rho1_s local grids are used it is not possible
! to call xc_rho_set_update.
! As input of this routine one gets rho and drho on a one dimensional grid.
! The grid is the angular grid corresponding to a given point ir_pnt on the radial grid.
! The derivatives are calculated on this one dimensional grid, the results are stored in
! exc, vxc(1:na,ir_pnt,ispin), vxg(1:na,ir_pnt,ispin), vxg_cross(1:na,ir_pnt,ispin)
! Afterwords the arrays containing the derivatives are put to zero so that the routine
! can safely be called for the next radial point ir_pnt

      TYPE(section_vals_type), POINTER                   :: xc_fun_section
      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      INTEGER, INTENT(in)                                :: deriv_order
      TYPE(xc_rho_cflags_type), INTENT(in)               :: needs
      REAL(dp), DIMENSION(:, :), POINTER                 :: w
      LOGICAL, INTENT(IN)                                :: lsd
      INTEGER, INTENT(in)                                :: na, nr
      REAL(dp)                                           :: exc
      REAL(dp), DIMENSION(:, :, :), POINTER              :: vxc
      REAL(dp), DIMENSION(:, :, :, :), POINTER           :: vxg
      REAL(dp), DIMENSION(:, :, :), POINTER              :: vtau
      LOGICAL, INTENT(IN), OPTIONAL                      :: energy_only, epr_xc
      REAL(dp), INTENT(IN), OPTIONAL                     :: adiabatic_rescale_factor

      CHARACTER(LEN=*), PARAMETER :: routineN = 'vxc_of_r_new', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ia, idir, ir, my_deriv_order
      LOGICAL                                            :: gradient_f, my_epr_xc, my_only_energy
      REAL(dp)                                           :: my_adiabatic_rescale_factor
      REAL(dp), DIMENSION(:, :, :), POINTER              :: deriv_data
      REAL(KIND=dp)                                      :: drho_cutoff
      TYPE(xc_derivative_type), POINTER                  :: deriv_att

      CALL timeset(routineN, handle)
      my_only_energy = .FALSE.
      IF (PRESENT(energy_only)) my_only_energy = energy_only

      IF (PRESENT(adiabatic_rescale_factor)) THEN
         my_adiabatic_rescale_factor = adiabatic_rescale_factor
      ELSE
         my_adiabatic_rescale_factor = 1.0_dp
      END IF

      ! needed for the epr routines
      my_epr_xc = .FALSE.
      IF (PRESENT(epr_xc)) my_epr_xc = epr_xc
      my_deriv_order = deriv_order
      IF (my_epr_xc) my_deriv_order = 2

      gradient_f = (needs%drho_spin .OR. needs%norm_drho_spin .OR. &
                    needs%drhoa_drhob .OR. needs%drho .OR. needs%norm_drho)

!  Calculate the derivatives
      CALL xc_functionals_eval(xc_fun_section, &
                               lsd=lsd, &
                               rho_set=rho_set, &
                               deriv_set=deriv_set, &
                               deriv_order=my_deriv_order)

      CALL xc_rho_set_get(rho_set, drho_cutoff=drho_cutoff)

      NULLIFY (deriv_data)

      IF (my_epr_xc) THEN
         ! nabla v_xc (using the vxg arrays)
         ! there's no point doing this when lsd = false
         IF (lsd) THEN
            deriv_att => xc_dset_get_derivative(deriv_set, "(rhoa)(rhoa)")
            IF (ASSOCIATED(deriv_att)) THEN
               CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
               DO ir = 1, nr
                  DO ia = 1, na
                     DO idir = 1, 3
                        vxg(idir, ia, ir, 1) = rho_set%drhoa(idir)%array(ia, ir, 1)* &
                                               deriv_data(ia, ir, 1)
                     END DO !idir
                  END DO !ia
               END DO !ir
               NULLIFY (deriv_data)
            END IF
            deriv_att => xc_dset_get_derivative(deriv_set, "(rhob)(rhob)")
            IF (ASSOCIATED(deriv_att)) THEN
               CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
               DO ir = 1, nr
                  DO ia = 1, na
                     DO idir = 1, 3
                        vxg(idir, ia, ir, 2) = rho_set%drhob(idir)%array(ia, ir, 1)* &
                                               deriv_data(ia, ir, 1)
                     END DO !idir
                  END DO !ia
               END DO !ir
               NULLIFY (deriv_data)
            END IF
         END IF
         !  EXC energy ! is that needed for epr?
         deriv_att => xc_dset_get_derivative(deriv_set, "")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
            exc = 0.0_dp
            DO ir = 1, nr
               DO ia = 1, na
                  exc = exc+deriv_data(ia, ir, 1)*w(ia, ir)
               END DO
            END DO
            NULLIFY (deriv_data)
         END IF
      ELSE
!  EXC energy
         deriv_att => xc_dset_get_derivative(deriv_set, "")
         IF (ASSOCIATED(deriv_att)) THEN
            CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
            exc = 0.0_dp
            DO ir = 1, nr
               DO ia = 1, na
                  exc = exc+deriv_data(ia, ir, 1)*w(ia, ir)
               END DO
            END DO
            NULLIFY (deriv_data)
         END IF
         ! Calculate the potential only if needed
         IF (.NOT. my_only_energy) THEN
!  Derivative with respect to the density
            IF (lsd) THEN
               deriv_att => xc_dset_get_derivative(deriv_set, "(rhoa)")
               IF (ASSOCIATED(deriv_att)) THEN
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
                  vxc(:, :, 1) = deriv_data(:, :, 1)*w(:, :)*my_adiabatic_rescale_factor
                  NULLIFY (deriv_data)
               END IF
               deriv_att => xc_dset_get_derivative(deriv_set, "(rhob)")
               IF (ASSOCIATED(deriv_att)) THEN
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
                  vxc(:, :, 2) = deriv_data(:, :, 1)*w(:, :)*my_adiabatic_rescale_factor
                  NULLIFY (deriv_data)
               END IF
               deriv_att => xc_dset_get_derivative(deriv_set, "(rho)")
               IF (ASSOCIATED(deriv_att)) THEN
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
                  vxc(:, :, 1) = vxc(:, :, 1)+deriv_data(:, :, 1)*w(:, :)*my_adiabatic_rescale_factor
                  vxc(:, :, 2) = vxc(:, :, 2)+deriv_data(:, :, 1)*w(:, :)*my_adiabatic_rescale_factor
                  NULLIFY (deriv_data)
               END IF
            ELSE
               deriv_att => xc_dset_get_derivative(deriv_set, "(rho)")
               IF (ASSOCIATED(deriv_att)) THEN
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
                  vxc(:, :, 1) = deriv_data(:, :, 1)*w(:, :)*my_adiabatic_rescale_factor
                  NULLIFY (deriv_data)
               END IF
            END IF ! lsd

!  Derivatives with respect to the gradient
            IF (lsd) THEN
               deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drhoa)")
               IF (ASSOCIATED(deriv_att)) THEN
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
                  DO ir = 1, nr
                     DO ia = 1, na
                        DO idir = 1, 3
                           IF (rho_set%norm_drhoa(ia, ir, 1) > drho_cutoff) THEN
                              vxg(idir, ia, ir, 1) = rho_set%drhoa(idir)%array(ia, ir, 1)* &
                                                     deriv_data(ia, ir, 1)*w(ia, ir)/ &
                                                     rho_set%norm_drhoa(ia, ir, 1)*my_adiabatic_rescale_factor
                           ELSE
                              vxg(idir, ia, ir, 1) = 0.0_dp
                           END IF
                        END DO
                     END DO
                  END DO
                  NULLIFY (deriv_data)
               END IF
               deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drhob)")
               IF (ASSOCIATED(deriv_att)) THEN
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
                  DO ir = 1, nr
                     DO ia = 1, na
                        DO idir = 1, 3
                           IF (rho_set%norm_drhob(ia, ir, 1) > drho_cutoff) THEN
                              vxg(idir, ia, ir, 2) = rho_set%drhob(idir)%array(ia, ir, 1)* &
                                                     deriv_data(ia, ir, 1)*w(ia, ir)/ &
                                                     rho_set%norm_drhob(ia, ir, 1)*my_adiabatic_rescale_factor
                           ELSE
                              vxg(idir, ia, ir, 2) = 0.0_dp
                           END IF
                        END DO
                     END DO
                  END DO
                  NULLIFY (deriv_data)
               END IF
               ! Cross Terms
               deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drho)")
               IF (ASSOCIATED(deriv_att)) THEN
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
                  DO ir = 1, nr
                     DO ia = 1, na
                        DO idir = 1, 3
                           IF (rho_set%norm_drho(ia, ir, 1) > drho_cutoff) THEN
                              vxg(idir, ia, ir, 1:2) = &
                                 vxg(idir, ia, ir, 1:2)+( &
                                 rho_set%drhoa(idir)%array(ia, ir, 1)+ &
                                 rho_set%drhob(idir)%array(ia, ir, 1))* &
                                 deriv_data(ia, ir, 1)*w(ia, ir)/rho_set%norm_drho(ia, ir, 1)* &
                                 my_adiabatic_rescale_factor
                           END IF
                        END DO
                     END DO
                  END DO
                  NULLIFY (deriv_data)
               END IF
            ELSE
               deriv_att => xc_dset_get_derivative(deriv_set, "(norm_drho)")
               IF (ASSOCIATED(deriv_att)) THEN
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
                  DO ir = 1, nr
                     DO ia = 1, na
                        IF (rho_set%norm_drho(ia, ir, 1) > drho_cutoff) THEN
                           DO idir = 1, 3
                              vxg(idir, ia, ir, 1) = rho_set%drho(idir)%array(ia, ir, 1)* &
                                                     deriv_data(ia, ir, 1)*w(ia, ir)/ &
                                                     rho_set%norm_drho(ia, ir, 1)*my_adiabatic_rescale_factor
                           END DO
                        ELSE
                           vxg(1:3, ia, ir, 1) = 0.0_dp
                        END IF
                     END DO
                  END DO
                  NULLIFY (deriv_data)
               END IF
            END IF ! lsd
!  Derivative with respect to tau
            IF (lsd) THEN
               deriv_att => xc_dset_get_derivative(deriv_set, "(taua)")
               IF (ASSOCIATED(deriv_att)) THEN
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
                  vtau(:, :, 1) = deriv_data(:, :, 1)*w(:, :)*my_adiabatic_rescale_factor
                  NULLIFY (deriv_data)
               END IF
               deriv_att => xc_dset_get_derivative(deriv_set, "(taub)")
               IF (ASSOCIATED(deriv_att)) THEN
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
                  vtau(:, :, 2) = deriv_data(:, :, 1)*w(:, :)*my_adiabatic_rescale_factor
                  NULLIFY (deriv_data)
               END IF
            ELSE
               deriv_att => xc_dset_get_derivative(deriv_set, "(tau)")
               IF (ASSOCIATED(deriv_att)) THEN
                  CALL xc_derivative_get(deriv_att, deriv_data=deriv_data)
                  vtau(:, :, 1) = deriv_data(:, :, 1)*w(:, :)*my_adiabatic_rescale_factor
                  NULLIFY (deriv_data)
               END IF
            END IF ! lsd
         END IF ! only_energy
      END IF ! epr_xc

      CALL timestop(handle)

   END SUBROUTINE vxc_of_r_new

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param rho1_set ...
!> \param xc_section ...
!> \param deriv_set ...
!> \param w ...
!> \param vxc ...
!> \param vxg ...
! **************************************************************************************************
   SUBROUTINE xc_2nd_deriv_of_r(rho_set, rho1_set, xc_section, &
                                deriv_set, w, vxc, vxg)

! As input of this routine one gets rho and drho on a one dimensional grid.
! The grid is the angular grid corresponding to a given point ir on the radial grid.
! The derivatives are calculated on this one dimensional grid, the results are stored in
! vxc(1:na,ir,ispin), vxg(1:na,ir,ispin), vxg_cross(1:na,ir,ispin)
! Afterwords the arrays containing the derivatives are put to zero so that the routine
! can safely be called for the next radial point ir

      TYPE(xc_rho_set_type), POINTER                     :: rho_set, rho1_set
      TYPE(section_vals_type), POINTER                   :: xc_section
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      REAL(dp), DIMENSION(:, :), POINTER                 :: w
      REAL(dp), DIMENSION(:, :, :), POINTER              :: vxc
      REAL(dp), DIMENSION(:, :, :, :), POINTER           :: vxg

      CHARACTER(LEN=*), PARAMETER :: routineN = 'xc_2nd_deriv_of_r', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspins
      LOGICAL                                            :: lsd
      REAL(dp)                                           :: drho_cutoff
      TYPE(cp_sll_xc_deriv_type), POINTER                :: pos
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: vxc_pw
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(section_vals_type), POINTER                   :: xc_fun_section
      TYPE(xc_derivative_type), POINTER                  :: deriv_att

      CALL timeset(routineN, handle)

      nspins = SIZE(vxc, 3)
      lsd = (nspins == 2)
      IF (ASSOCIATED(rho_set%rhoa)) THEN
         lsd = .TRUE.
      END IF
      CALL xc_rho_set_get(rho_set, drho_cutoff=drho_cutoff)
      xc_fun_section => section_vals_get_subs_vals(xc_section, &
                                                   "XC_FUNCTIONAL")

      !  Calculate the derivatives
      CALL xc_functionals_eval(xc_fun_section, &
                               lsd=lsd, &
                               rho_set=rho_set, &
                               deriv_set=deriv_set, &
                               deriv_order=2)

      CALL divide_by_norm_drho(deriv_set, rho_set, lsd)

      ! multiply by -w
      pos => deriv_set%derivs
      DO WHILE (cp_sll_xc_deriv_next(pos, el_att=deriv_att))
         !deriv_att%deriv_data(:,:,1) = -w(:,:)*deriv_att%deriv_data(:,:,1)
         deriv_att%deriv_data(:, :, 1) = w(:, :)*deriv_att%deriv_data(:, :, 1)
      END DO

      NULLIFY (pw_pool)
      ALLOCATE (vxc_pw(nspins))
      DO ispin = 1, nspins
         ALLOCATE (vxc_pw(ispin)%pw)
         vxc_pw(ispin)%pw%cr3d => vxc(:, :, ispin:ispin)
      END DO

      CALL xc_calc_2nd_deriv(vxc_pw, deriv_set, rho_set, rho1_set, pw_pool, &
                             xc_section, gapw=.TRUE., vxg=vxg)

      DO ispin = 1, nspins
         DEALLOCATE (vxc_pw(ispin)%pw)
      END DO
      DEALLOCATE (vxc_pw)

      ! zero the derivative data for the next call
      pos => deriv_set%derivs
      DO WHILE (cp_sll_xc_deriv_next(pos, el_att=deriv_att))
         deriv_att%deriv_data = 0.0_dp
      END DO

      CALL timestop(handle)

   END SUBROUTINE xc_2nd_deriv_of_r

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param needs ...
!> \param nspins ...
!> \param bo ...
! **************************************************************************************************
   SUBROUTINE xc_rho_set_atom_update(rho_set, needs, nspins, bo)

!   This routine allocates the storage arrays for rho and drho
!   In calculate_vxc_atom this is called once for each atomic_kind,
!   After the loop over all the atoms of the kind and over all the points
!   of the radial grid for each atom, rho_set is deallocated.
!   Within the same kind, at each new point on the radial grid, the rho_set
!   arrays rho and drho are overwritten.

      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      TYPE(xc_rho_cflags_type)                           :: needs
      INTEGER, INTENT(IN)                                :: nspins
      INTEGER, DIMENSION(2, 3), INTENT(IN)               :: bo

      CHARACTER(LEN=*), PARAMETER :: routineN = 'xc_rho_set_atom_update', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: idir

      SELECT CASE (nspins)
      CASE (1)
!     What is this for?
         IF (needs%rho_1_3) THEN
            NULLIFY (rho_set%rho_1_3)
            ALLOCATE (rho_set%rho_1_3(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            rho_set%owns%rho_1_3 = .TRUE.
            rho_set%has%rho_1_3 = .FALSE.
         END IF
!     Allocate the storage space for the density
         IF (needs%rho) THEN
            NULLIFY (rho_set%rho)
            ALLOCATE (rho_set%rho(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            rho_set%owns%rho = .TRUE.
            rho_set%has%rho = .FALSE.
         END IF
!     Allocate the storage space for  the norm of the gradient of the density
         IF (needs%norm_drho) THEN
            NULLIFY (rho_set%norm_drho)
            ALLOCATE (rho_set%norm_drho(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            rho_set%owns%norm_drho = .TRUE.
            rho_set%has%norm_drho = .FALSE.
         END IF
!     Allocate the storage space for the three components of the gradient of the density
         IF (needs%drho) THEN
            DO idir = 1, 3
               NULLIFY (rho_set%drho(idir)%array)
               ALLOCATE (rho_set%drho(idir)%array(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            END DO
            rho_set%owns%drho = .TRUE.
            rho_set%has%drho = .FALSE.
         END IF
      CASE (2)
!     Allocate the storage space for the total density
         IF (needs%rho) THEN
            ! this should never be the case unless you use LDA functionals with LSD
            NULLIFY (rho_set%rho)
            ALLOCATE (rho_set%rho(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            rho_set%owns%rho = .TRUE.
            rho_set%has%rho = .FALSE.
         END IF
!     What is this for?
         IF (needs%rho_1_3) THEN
            NULLIFY (rho_set%rho_1_3)
            ALLOCATE (rho_set%rho_1_3(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            rho_set%owns%rho_1_3 = .TRUE.
            rho_set%has%rho_1_3 = .FALSE.
         END IF
!     What is this for?
         IF (needs%rho_spin_1_3) THEN
            NULLIFY (rho_set%rhoa_1_3, rho_set%rhob_1_3)
            ALLOCATE (rho_set%rhoa_1_3(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            ALLOCATE (rho_set%rhob_1_3(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            rho_set%owns%rho_spin_1_3 = .TRUE.
            rho_set%has%rho_spin_1_3 = .FALSE.
         END IF
!     Allocate the storage space for the spin densities rhoa and rhob
         IF (needs%rho_spin) THEN
            NULLIFY (rho_set%rhoa, rho_set%rhob)
            ALLOCATE (rho_set%rhoa(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            ALLOCATE (rho_set%rhob(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            rho_set%owns%rho_spin = .TRUE.
            rho_set%has%rho_spin = .FALSE.
         END IF
!     Allocate the storage space for the norm of the gradient of the total density
         IF (needs%norm_drho) THEN
            NULLIFY (rho_set%norm_drho)
            ALLOCATE (rho_set%norm_drho(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            rho_set%owns%norm_drho = .TRUE.
            rho_set%has%norm_drho = .FALSE.
         END IF
!     Allocate the storage space for the norm of the gradient of rhoa and of rhob separatedly
         IF (needs%norm_drho_spin) THEN
            NULLIFY (rho_set%norm_drhoa, rho_set%norm_drhob)
            ALLOCATE (rho_set%norm_drhoa(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            ALLOCATE (rho_set%norm_drhob(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            rho_set%owns%norm_drho_spin = .TRUE.
            rho_set%has%norm_drho_spin = .FALSE.
         END IF
!
         IF (needs%drhoa_drhob) THEN
            NULLIFY (rho_set%drhoa_drhob)
            ALLOCATE (rho_set%drhoa_drhob(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            rho_set%owns%drhoa_drhob = .TRUE.
            rho_set%has%drhoa_drhob = .FALSE.
         END IF
!     Allocate the storage space for the components of the gradient for the total rho
         IF (needs%drho) THEN
            DO idir = 1, 3
               NULLIFY (rho_set%drho(idir)%array)
               ALLOCATE (rho_set%drho(idir)%array(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            END DO
            rho_set%owns%drho = .TRUE.
            rho_set%has%drho = .FALSE.
         END IF
!     Allocate the storage space for the components of the gradient for rhoa and rhob
         IF (needs%drho_spin) THEN
            DO idir = 1, 3
               NULLIFY (rho_set%drhoa(idir)%array, rho_set%drhob(idir)%array)
               ALLOCATE (rho_set%drhoa(idir)%array(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
               ALLOCATE (rho_set%drhob(idir)%array(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
            END DO
            rho_set%owns%drho_spin = .TRUE.
            rho_set%has%drho_spin = .FALSE.
         END IF
!
      END SELECT

      ! tau part
      IF (needs%tau) THEN
         NULLIFY (rho_set%tau)
         ALLOCATE (rho_set%tau(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
         rho_set%owns%tau = .TRUE.
      END IF
      IF (needs%tau_spin) THEN
         NULLIFY (rho_set%tau_a, rho_set%tau_b)
         ALLOCATE (rho_set%tau_a(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
         ALLOCATE (rho_set%tau_b(bo(1, 1):bo(2, 1), bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
         rho_set%owns%tau_spin = .TRUE.
         rho_set%has%tau_spin = .FALSE.
      END IF

   END SUBROUTINE xc_rho_set_atom_update

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param lsd ...
!> \param nspins ...
!> \param needs ...
!> \param rho ...
!> \param drho ...
!> \param tau ...
!> \param na ...
!> \param ir ...
! **************************************************************************************************
   SUBROUTINE fill_rho_set(rho_set, lsd, nspins, needs, rho, drho, tau, na, ir)

      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      LOGICAL, INTENT(IN)                                :: lsd
      INTEGER, INTENT(IN)                                :: nspins
      TYPE(xc_rho_cflags_type), INTENT(in)               :: needs
      REAL(dp), DIMENSION(:, :), POINTER                 :: rho
      REAL(dp), DIMENSION(:, :, :, :), POINTER           :: drho
      REAL(dp), DIMENSION(:, :), POINTER                 :: tau
      INTEGER, INTENT(IN)                                :: na, ir

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fill_rho_set', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: f13 = (1.0_dp/3.0_dp)

      INTEGER                                            :: ia, idir, my_nspins
      LOGICAL                                            :: gradient_f, tddft_split

      my_nspins = nspins
      tddft_split = .FALSE.
      IF (lsd .AND. nspins == 1) THEN
         my_nspins = 2
         tddft_split = .TRUE.
      END IF

      ! some checks
      IF (lsd) THEN
      ELSE
         IF(.NOT.(SIZE(rho, 2) == 1))CALL cp__a("xc/xc_atom.F",599)
      END IF
      SELECT CASE (my_nspins)
      CASE (1)
         IF(.NOT.(.NOT. needs%rho_spin))CALL cp__a("xc/xc_atom.F",603)
         IF(.NOT.(.NOT. needs%drho_spin))CALL cp__a("xc/xc_atom.F",604)
         IF(.NOT.(.NOT. needs%norm_drho_spin))CALL cp__a("xc/xc_atom.F",605)
         IF(.NOT.(.NOT. needs%drhoa_drhob))CALL cp__a("xc/xc_atom.F",606)
         IF(.NOT.(.NOT. needs%rho_spin_1_3))CALL cp__a("xc/xc_atom.F",607)
      CASE (2)
      CASE default
         CALL cp__b("xc/xc_atom.F",610,"")
      END SELECT

      gradient_f = (needs%drho_spin .OR. needs%norm_drho_spin .OR. &
                    needs%drhoa_drhob .OR. needs%drho .OR. needs%norm_drho)

      SELECT CASE (my_nspins)
      CASE (1)
         ! Give rho to 1/3
         IF (needs%rho_1_3) THEN
            DO ia = 1, na
               rho_set%rho_1_3(ia, ir, 1) = MAX(rho(ia, 1), 0.0_dp)**f13
            END DO
            rho_set%owns%rho_1_3 = .TRUE.
            rho_set%has%rho_1_3 = .TRUE.
         END IF
         ! Give the density
         IF (needs%rho) THEN
            DO ia = 1, na
               rho_set%rho(ia, ir, 1) = rho(ia, 1)
            END DO
            rho_set%owns%rho = .TRUE.
            rho_set%has%rho = .TRUE.
         END IF
         ! Give the norm of the gradient of the density
         IF (needs%norm_drho) THEN
            DO ia = 1, na
               rho_set%norm_drho(ia, ir, 1) = drho(4, ia, ir, 1)
            END DO
            rho_set%owns%norm_drho = .TRUE.
            rho_set%has%norm_drho = .TRUE.
         END IF
         ! Give the three components of the gradient of the density
         IF (needs%drho) THEN
            DO idir = 1, 3
               DO ia = 1, na
                  rho_set%drho(idir)%array(ia, ir, 1) = drho(idir, ia, ir, 1)
               END DO
            END DO
            rho_set%owns%drho = .TRUE.
            rho_set%has%drho = .TRUE.
         END IF
      CASE (2)
         ! Give the total density
         IF (needs%rho) THEN
            ! this should never be the case unless you use LDA functionals with LSD
            IF (.NOT. tddft_split) THEN
               DO ia = 1, na
                  rho_set%rho(ia, ir, 1) = rho(ia, 1)+rho(ia, 2)
               END DO
            ELSE
               DO ia = 1, na
                  rho_set%rho(ia, ir, 1) = rho(ia, 1)
               END DO
            END IF
            rho_set%owns%rho = .TRUE.
            rho_set%has%rho = .TRUE.
         END IF
         ! Give the total density to 1/3
         IF (needs%rho_1_3) THEN
            IF (.NOT. tddft_split) THEN
               DO ia = 1, na
                  rho_set%rho_1_3(ia, ir, 1) = MAX(rho(ia, 1)+rho(ia, 2), 0.0_dp)**f13
               END DO
            ELSE
               DO ia = 1, na
                  rho_set%rho_1_3(ia, ir, 1) = MAX(rho(ia, 1), 0.0_dp)**f13
               END DO
            END IF
            rho_set%owns%rho_1_3 = .TRUE.
            rho_set%has%rho_1_3 = .TRUE.
         END IF
         ! Give the spin densities to 1/3
         IF (needs%rho_spin_1_3) THEN
            IF (.NOT. tddft_split) THEN
               DO ia = 1, na
                  rho_set%rhoa_1_3(ia, ir, 1) = MAX(rho(ia, 1), 0.0_dp)**f13
                  rho_set%rhob_1_3(ia, ir, 1) = MAX(rho(ia, 2), 0.0_dp)**f13
               END DO
            ELSE
               DO ia = 1, na
                  rho_set%rhoa_1_3(ia, ir, 1) = MAX(0.5_dp*rho(ia, 1), 0.0_dp)**f13
                  rho_set%rhob_1_3(ia, ir, 1) = rho_set%rhoa_1_3(ia, ir, 1)
               END DO
            END IF
            rho_set%owns%rho_spin_1_3 = .TRUE.
            rho_set%has%rho_spin_1_3 = .TRUE.
         END IF
         ! Give the spin densities rhoa and rhob
         IF (needs%rho_spin) THEN
            IF (.NOT. tddft_split) THEN
               DO ia = 1, na
                  rho_set%rhoa(ia, ir, 1) = rho(ia, 1)
                  rho_set%rhob(ia, ir, 1) = rho(ia, 2)
               END DO
            ELSE
               DO ia = 1, na
                  rho_set%rhoa(ia, ir, 1) = 0.5_dp*rho(ia, 1)
                  rho_set%rhob(ia, ir, 1) = rho_set%rhoa(ia, ir, 1)
               END DO
            END IF
            rho_set%owns%rho_spin = .TRUE.
            rho_set%has%rho_spin = .TRUE.
         END IF
         ! Give the norm of the gradient of the total density
         IF (needs%norm_drho) THEN
            IF (.NOT. tddft_split) THEN
               DO ia = 1, na
                  rho_set%norm_drho(ia, ir, 1) = SQRT( &
                                                 (drho(1, ia, ir, 1)+drho(1, ia, ir, 2))**2+ &
                                                 (drho(2, ia, ir, 1)+drho(2, ia, ir, 2))**2+ &
                                                 (drho(3, ia, ir, 1)+drho(3, ia, ir, 2))**2)
               END DO
            ELSE
               DO ia = 1, na
                  rho_set%norm_drho(ia, ir, 1) = drho(4, ia, ir, 1)
!!                rho_set%norm_drho(ia,ir,1) = SQRT(&
!!                     (drho(1,ia,ir,1))**2+&
!!                     (drho(2,ia,ir,1))**2+&
!!                     (drho(3,ia,ir,1))**2)
               END DO
            END IF
            rho_set%owns%norm_drho = .TRUE.
            rho_set%has%norm_drho = .TRUE.
         END IF
         ! Give the norm of the gradient of rhoa and of rhob separatedly
         IF (needs%norm_drho_spin) THEN
            IF (.NOT. tddft_split) THEN
               DO ia = 1, na
                  rho_set%norm_drhoa(ia, ir, 1) = drho(4, ia, ir, 1)
                  rho_set%norm_drhob(ia, ir, 1) = drho(4, ia, ir, 2)
               END DO
            ELSE
               DO ia = 1, na
                  rho_set%norm_drhoa(ia, ir, 1) = 0.5_dp*drho(4, ia, ir, 1)
                  rho_set%norm_drhob(ia, ir, 1) = rho_set%norm_drhoa(ia, ir, 1)
               END DO
            END IF
            rho_set%owns%norm_drho_spin = .TRUE.
            rho_set%has%norm_drho_spin = .TRUE.
         END IF
         !
         IF (needs%drhoa_drhob) THEN
            IF (.NOT. tddft_split) THEN
               DO ia = 1, na
                  rho_set%drhoa_drhob(ia, ir, 1) = &
                     (drho(1, ia, ir, 1)*drho(1, ia, ir, 2))+ &
                     (drho(2, ia, ir, 1)*drho(2, ia, ir, 2))+ &
                     (drho(3, ia, ir, 1)*drho(3, ia, ir, 2))
               END DO
            ELSE
               DO ia = 1, na
                  rho_set%drhoa_drhob(ia, ir, 1) = 0.25_dp*drho(4, ia, ir, 1)**2
!!                rho_set%drhoa_drhob(ia,ir,1) =&
!!                     (0.25_dp*drho(1,ia,ir,1)**2)+&
!!                     (0.25_dp*drho(2,ia,ir,1)**2)+&
!!                     (0.25_dp*drho(3,ia,ir,1)**2)
               END DO
            END IF
            rho_set%owns%drhoa_drhob = .TRUE.
            rho_set%has%drhoa_drhob = .TRUE.
         END IF
         ! Give the components of the gradient for the total rho
         IF (needs%drho) THEN
            IF (.NOT. tddft_split) THEN
               DO idir = 1, 3
                  DO ia = 1, na
                     rho_set%drho(idir)%array(ia, ir, 1) = drho(idir, ia, ir, 1)+drho(idir, ia, ir, 2)
                  END DO
               END DO
            ELSE
               DO idir = 1, 3
                  DO ia = 1, na
                     rho_set%drho(idir)%array(ia, ir, 1) = drho(idir, ia, ir, 1)
                  END DO
               END DO
            END IF
            rho_set%owns%drho = .TRUE.
            rho_set%has%drho = .TRUE.
         END IF
         ! Give the components of the gradient for rhoa and rhob
         IF (needs%drho_spin) THEN
            IF (.NOT. tddft_split) THEN
               DO idir = 1, 3
                  DO ia = 1, na
                     rho_set%drhoa(idir)%array(ia, ir, 1) = drho(idir, ia, ir, 1)
                     rho_set%drhob(idir)%array(ia, ir, 1) = drho(idir, ia, ir, 2)
                  END DO
               END DO
            ELSE
               DO idir = 1, 3
                  DO ia = 1, na
                     rho_set%drhoa(idir)%array(ia, ir, 1) = 0.5_dp*drho(idir, ia, ir, 1)
                     rho_set%drhob(idir)%array(ia, ir, 1) = rho_set%drhoa(idir)%array(ia, ir, 1)
                  END DO
               END DO
            END IF
            rho_set%owns%drho_spin = .TRUE.
            rho_set%has%drho_spin = .TRUE.
         END IF
         !
      END SELECT

      ! tau part
      IF (needs%tau .OR. needs%tau_spin) THEN
         IF(.NOT.(ASSOCIATED(tau)))CALL cp__a("xc/xc_atom.F",815)
         IF(.NOT.(SIZE(tau, 2) == my_nspins))CALL cp__a("xc/xc_atom.F",816)
      END IF
      IF (needs%tau) THEN
         IF (my_nspins == 2) THEN
            DO ia = 1, na
               rho_set%tau(ia, ir, 1) = tau(ia, 1)+tau(ia, 2)
            END DO
            rho_set%owns%tau = .TRUE.
            rho_set%has%tau = .TRUE.
         ELSE
            DO ia = 1, na
               rho_set%tau(ia, ir, 1) = tau(ia, 1)
            END DO
            rho_set%owns%tau = .TRUE.
            rho_set%has%tau = .TRUE.
         END IF
      END IF
      IF (needs%tau_spin) THEN
         DO ia = 1, na
            rho_set%tau_a(ia, ir, 1) = tau(ia, 1)
            rho_set%tau_b(ia, ir, 1) = tau(ia, 2)
         END DO
         rho_set%owns%tau_spin = .TRUE.
         rho_set%has%tau_spin = .TRUE.
      END IF

   END SUBROUTINE fill_rho_set

END MODULE xc_atom
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines that build the integrals of the Vxc potential calculated
!>      for the atomic code
! **************************************************************************************************
MODULE atom_xc

   USE atom_types,                      ONLY: atom_type,&
                                              gth_pseudo,&
                                              opmat_type,&
                                              upf_pseudo
! ZMP added coulomb_potential_numeric
   USE atom_utils,                      ONLY: atom_core_density,&
                                              atom_density,&
                                              coulomb_potential_numeric,&
                                              integrate_grid,&
                                              numpot_matrix
   USE cp_files,                        ONLY: close_file,&
                                              get_unit_number,&
                                              open_file
   USE input_constants,                 ONLY: xc_none
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi
   USE xc_atom,                         ONLY: xc_rho_set_atom_update
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_create,&
                                              xc_dset_get_derivative,&
                                              xc_dset_release,&
                                              xc_dset_zero_all
   USE xc_derivative_types,             ONLY: xc_derivative_get,&
                                              xc_derivative_type
   USE xc_derivatives,                  ONLY: xc_functionals_eval,&
                                              xc_functionals_get_needs
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_create,&
                                              xc_rho_set_release,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_xc'

! ZMP added public subroutines calculate_atom_zmp, calculate_atom_ext_vxc
   PUBLIC :: calculate_atom_vxc_lda, calculate_atom_vxc_lsd, &
             calculate_atom_zmp, calculate_atom_ext_vxc

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ZMP subroutine for the calculation of the effective constraint
!>        potential in the atomic code.
!>
!> \param ext_density ...
!> \param atom ...
!> \param lprint ...
!> \param xcmat ...
!> \author D. Varsano [daniele.varsano@nano.cnr.it]
!>
! **************************************************************************************************
   SUBROUTINE calculate_atom_zmp(ext_density, atom, lprint, xcmat)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: ext_density
      TYPE(atom_type), INTENT(INOUT)                     :: atom
      LOGICAL                                            :: lprint
      TYPE(opmat_type), OPTIONAL, POINTER                :: xcmat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_atom_zmp', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: extunit, handle, ir, nr, z
      REAL(KIND=dp)                                      :: int1, int2
      REAL(KIND=dp), DIMENSION(:), POINTER               :: deltarho, rho_dum, vxc, vxc1, vxc2
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rho

      CALL timeset(routineN, handle)

      nr = atom%basis%grid%nr
      z = atom%z
      ALLOCATE (rho(nr, 1), vxc(nr), vxc1(nr), vxc2(nr), rho_dum(nr), deltarho(nr))
      CALL atom_density(rho(:, 1), atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, typ="RHO")
      !Vxc1
      int1 = integrate_grid(ext_density, atom%basis%grid)
      int1 = fourpi*int1
      CALL coulomb_potential_numeric(vxc1, rho(:, 1), atom%basis%grid)

      vxc1 = -vxc1/z

      !Vxc2
      rho_dum = rho(:, 1)*int1/z
      deltarho = rho_dum-ext_density
      int2 = integrate_grid(deltarho, atom%basis%grid)
      CALL coulomb_potential_numeric(vxc2, deltarho, atom%basis%grid)

      vxc2 = vxc2*atom%lambda

      !Vxc
      vxc = vxc1+vxc2

      atom%energy%exc = fourpi*integrate_grid(vxc, rho(:, 1), atom%basis%grid)
      atom%rho_diff_integral = fourpi*int2

      IF (lprint) THEN
         extunit = get_unit_number()
         CALL open_file(file_name="/mnt/sdcard/cp2k/work/linear_potentials.dat", file_status="UNKNOWN", &
                        file_form="FORMATTED", file_action="WRITE", &
                        unit_number=extunit)

         WRITE (extunit, fmt='("#",T10,"R[bohr]",T36,"Rho[au]",T61,"DRho[au]",T86,"V_XC[au]",T111,"V_XC1[au]",T136,"V_XC2[au]")')
         DO ir = 1, nr
            WRITE (extunit, fmt='(T1,E24.15,T26,E24.15,T51,E24.15,T76,E24.15,T101,E24.15,T126,E24.15)') &
               atom%basis%grid%rad(ir), rho(ir, 1), deltarho(ir), vxc(ir), vxc1(ir), vxc2(ir)
         ENDDO
         CALL close_file(unit_number=extunit)
      ENDIF

      IF (PRESENT(xcmat)) CALL numpot_matrix(xcmat%op, vxc, atom%basis, 0)

      DEALLOCATE (rho, vxc, vxc1, vxc2, rho_dum, deltarho)

      CALL timestop(handle)

   END SUBROUTINE calculate_atom_zmp
! **************************************************************************************************
!> \brief ZMP subroutine for the scf external density from the external v_xc
!>
!> \param vxc ...
!> \param atom ...
!> \param lprint ...
!> \param xcmat ...
!> \author D. Varsano [daniele.varsano@nano.cnr.it]
!>
! **************************************************************************************************
   SUBROUTINE calculate_atom_ext_vxc(vxc, atom, lprint, xcmat)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: vxc
      TYPE(atom_type), INTENT(INOUT)                     :: atom
      LOGICAL, INTENT(in)                                :: lprint
      TYPE(opmat_type), INTENT(inout), OPTIONAL, POINTER :: xcmat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_atom_ext_vxc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: extunit, handle, ir, nr
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rho

      CALL timeset(routineN, handle)
      nr = atom%basis%grid%nr

      ALLOCATE (rho(nr, 1))

      CALL atom_density(rho(:, 1), atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, typ="RHO")

      IF (lprint) THEN
         extunit = get_unit_number()
         CALL open_file(file_name="/mnt/sdcard/cp2k/work/linear_potentials.dat", file_status="UNKNOWN", &
                        file_form="FORMATTED", file_action="WRITE", &
                        unit_number=extunit)

         WRITE (extunit, fmt='("#",T10,"R[bohr]",T36,"Rho[au]",T61,"V_XC[au]")')
         DO ir = 1, nr
            WRITE (extunit, fmt='(T1,E24.15,T26,E24.15,T51,E24.15)') &
               atom%basis%grid%rad(ir), rho(ir, 1), vxc(ir)
         ENDDO
         CALL close_file(unit_number=extunit)
      ENDIF

      atom%energy%exc = fourpi*integrate_grid(vxc, rho(:, 1), atom%basis%grid)
      IF (PRESENT(xcmat)) CALL numpot_matrix(xcmat%op, vxc, atom%basis, 0)

      DEALLOCATE (rho)

      CALL timestop(handle)

   END SUBROUTINE calculate_atom_ext_vxc
! **************************************************************************************************
!> \brief ...
!> \param xcmat ...
!> \param atom ...
!> \param xc_section ...
! **************************************************************************************************
   SUBROUTINE calculate_atom_vxc_lda(xcmat, atom, xc_section)
      TYPE(opmat_type), POINTER                          :: xcmat
      TYPE(atom_type), INTENT(INOUT)                     :: atom
      TYPE(section_vals_type), POINTER                   :: xc_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_atom_vxc_lda', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: deriv_order, handle, i, l, myfun, n1, &
                                                            n2, n3, nr, nspins, unit_nr
      INTEGER, DIMENSION(2, 3)                           :: bounds
      LOGICAL                                            :: lsd, nlcc
      REAL(KIND=dp)                                      :: density_cut, gradient_cut, tau_cut
      REAL(KIND=dp), DIMENSION(:), POINTER               :: exc, vxc
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: drho, lap, rho, tau
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: taumat, xcpot
      TYPE(section_vals_type), POINTER                   :: xc_fun_section
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_derivative_type), POINTER                  :: deriv
      TYPE(xc_rho_cflags_type)                           :: needs
      TYPE(xc_rho_set_type), POINTER                     :: rho_set

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      nlcc = .FALSE.
      IF (atom%potential%ppot_type == gth_pseudo) THEN
         nlcc = atom%potential%gth_pot%nlcc
      ELSE IF (atom%potential%ppot_type == upf_pseudo) THEN
         nlcc = atom%potential%upf_pot%core_correction
         IF(.NOT.(.NOT. nlcc))CALL cp__a("atom_xc.F",222)
      END IF

      IF (ASSOCIATED(xc_section)) THEN
         NULLIFY (rho_set)

         xc_fun_section => section_vals_get_subs_vals(xc_section, "XC_FUNCTIONAL")
         CALL section_vals_val_get(xc_fun_section, "_SECTION_PARAMETERS_", i_val=myfun)

         IF (myfun == xc_none) THEN
            atom%energy%exc = 0._dp
         ELSE
            CALL section_vals_val_get(xc_section, "DENSITY_CUTOFF", r_val=density_cut)
            CALL section_vals_val_get(xc_section, "GRADIENT_CUTOFF", r_val=gradient_cut)
            CALL section_vals_val_get(xc_section, "TAU_CUTOFF", r_val=tau_cut)

            lsd = .FALSE.
            nspins = 1
            needs = xc_functionals_get_needs(xc_fun_section, lsd=lsd, add_basic_components=.FALSE.)

            ! Prepare the structures needed to calculate and store the xc derivatives

            ! Array dimension: here anly one dimensional arrays are used,
            ! i.e. only the first column of deriv_data is read.
            ! The other to dimensions  are set to size equal 1
            nr = atom%basis%grid%nr
            bounds(1:2, 1:3) = 1
            bounds(2, 1) = nr

            ! create a place where to put the derivatives
            NULLIFY (deriv_set)
            CALL xc_dset_create(deriv_set, local_bounds=bounds)
            ! create the place where to store the argument for the functionals
            CALL xc_rho_set_create(rho_set, bounds, rho_cutoff=density_cut, &
                                   drho_cutoff=gradient_cut, tau_cutoff=tau_cut)
            ! allocate the required 3d arrays where to store rho and drho
            CALL xc_rho_set_atom_update(rho_set, needs, nspins, bounds)

            NULLIFY (rho, drho, tau)
            IF (needs%rho) THEN
               ALLOCATE (rho(nr, 1))
               CALL atom_density(rho(:, 1), atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, typ="RHO")
               IF (nlcc) THEN
                  CALL atom_core_density(rho(:, 1), atom%potential%gth_pot, typ="RHO", rr=atom%basis%grid%rad)
               END IF
            END IF
            IF (needs%norm_drho) THEN
               ALLOCATE (drho(nr, 1))
               CALL atom_density(drho(:, 1), atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, typ="DER")
               IF (nlcc) THEN
                  CALL atom_core_density(drho(:, 1), atom%potential%gth_pot, typ="DER", rr=atom%basis%grid%rad)
               END IF
            END IF
            IF (needs%tau) THEN
               ALLOCATE (tau(nr, 1))
               CALL atom_density(tau(:, 1), atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, &
                                 typ="KIN", rr=atom%basis%grid%rad2)
            END IF
            IF (needs%laplace_rho) THEN
               ALLOCATE (lap(nr, 1))
               CALL atom_density(lap(:, 1), atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, &
                                 typ="LAP", rr=atom%basis%grid%rad2)
               IF (nlcc) THEN
                  CALL atom_core_density(lap(:, 1), atom%potential%gth_pot, typ="LAP", rr=atom%basis%grid%rad)
               END IF
            END IF

            CALL fill_rho_set(rho_set, nspins, needs, rho, drho, tau, lap, nr)

            CALL xc_dset_zero_all(deriv_set)

            deriv_order = 1
            CALL xc_functionals_eval(xc_fun_section, lsd=lsd, rho_set=rho_set, deriv_set=deriv_set, &
                                     deriv_order=deriv_order)

            ! Integration to get the matrix elements and energy
            deriv => xc_dset_get_derivative(deriv_set, "", allocate_deriv=.FALSE.)
            CALL xc_derivative_get(deriv, deriv_data=xcpot)
            atom%energy%exc = fourpi*integrate_grid(xcpot(:, 1, 1), atom%basis%grid)

            ! dump grid density and xcpot (xc energy?)
            IF (.FALSE.) THEN
               CALL open_file(unit_number=unit_nr, file_name="/mnt/sdcard/cp2k/work/atom.dat")
               DO i = 1, SIZE(xcpot(:, 1, 1))
                  WRITE (unit_nr, *) atom%basis%grid%rad(i), rho(i, 1), xcpot(i, 1, 1)
               ENDDO
               CALL close_file(unit_nr)
            ENDIF

            IF (needs%rho) THEN
               deriv => xc_dset_get_derivative(deriv_set, "(rho)", allocate_deriv=.FALSE.)
               CALL xc_derivative_get(deriv, deriv_data=xcpot)
               CALL numpot_matrix(xcmat%op, xcpot(:, 1, 1), atom%basis, 0)
               DEALLOCATE (rho)
            END IF
            IF (needs%norm_drho) THEN
               deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)", allocate_deriv=.FALSE.)
               CALL xc_derivative_get(deriv, deriv_data=xcpot)
               CALL numpot_matrix(xcmat%op, xcpot(:, 1, 1), atom%basis, 1)
               DEALLOCATE (drho)
            END IF
            IF (needs%tau) THEN
               deriv => xc_dset_get_derivative(deriv_set, "(tau)", allocate_deriv=.FALSE.)
               CALL xc_derivative_get(deriv, deriv_data=xcpot)
               n1 = SIZE(xcmat%op, 1)
               n2 = SIZE(xcmat%op, 2)
               n3 = SIZE(xcmat%op, 3)
               ALLOCATE (taumat(n1, n2, 0:n3-1))
               taumat = 0._dp

               xcpot(:, 1, 1) = 0.5_dp*xcpot(:, 1, 1)
               CALL numpot_matrix(xcmat%op, xcpot(:, 1, 1), atom%basis, 2)
               xcpot(:, 1, 1) = xcpot(:, 1, 1)/atom%basis%grid%rad2(:)
               CALL numpot_matrix(taumat, xcpot(:, 1, 1), atom%basis, 0)
               DO l = 0, 3
                  xcmat%op(:, :, l) = xcmat%op(:, :, l)+REAL(l*(l+1), dp)*taumat(:, :, l)
               END DO

               DEALLOCATE (tau)
               DEALLOCATE (taumat)
            END IF

            ! Release the xc structure used to store the xc derivatives
            CALL xc_dset_release(deriv_set)
            CALL xc_rho_set_release(rho_set)

         END IF !xc_none

      ELSE

         ! we don't have an xc_section, use a default setup
         nr = atom%basis%grid%nr
         ALLOCATE (rho(nr, 1), exc(nr), vxc(nr))

         CALL atom_density(rho(:, 1), atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, typ="RHO")
         IF (nlcc) THEN
            CALL atom_core_density(rho(:, 1), atom%potential%gth_pot, typ="RHO", rr=atom%basis%grid%rad)
         END IF
         CALL lda_pade(rho(:, 1), exc, vxc)

         atom%energy%exc = fourpi*integrate_grid(exc, atom%basis%grid)
         CALL numpot_matrix(xcmat%op, vxc, atom%basis, 0)

         DEALLOCATE (rho, exc, vxc)

      END IF

      CALL timestop(handle)

   END SUBROUTINE calculate_atom_vxc_lda

! **************************************************************************************************
!> \brief ...
!> \param xcmata ...
!> \param xcmatb ...
!> \param atom ...
!> \param xc_section ...
! **************************************************************************************************
   SUBROUTINE calculate_atom_vxc_lsd(xcmata, xcmatb, atom, xc_section)
      TYPE(opmat_type), POINTER                          :: xcmata, xcmatb
      TYPE(atom_type), INTENT(INOUT)                     :: atom
      TYPE(section_vals_type), POINTER                   :: xc_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_atom_vxc_lsd', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: deriv_order, handle, l, myfun, n1, n2, &
                                                            n3, nr, nspins
      INTEGER, DIMENSION(2, 3)                           :: bounds
      LOGICAL                                            :: lsd, nlcc
      REAL(KIND=dp)                                      :: density_cut, gradient_cut, tau_cut
      REAL(KIND=dp), DIMENSION(:), POINTER               :: exc, vxca, vxcb, xfun
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: drho, lap, rho, tau
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: taumat, xcpot
      TYPE(section_vals_type), POINTER                   :: xc_fun_section
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_derivative_type), POINTER                  :: deriv
      TYPE(xc_rho_cflags_type)                           :: needs
      TYPE(xc_rho_set_type), POINTER                     :: rho_set

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      nlcc = .FALSE.
      IF (atom%potential%ppot_type == gth_pseudo) THEN
         nlcc = atom%potential%gth_pot%nlcc
      ELSE IF (atom%potential%ppot_type == upf_pseudo) THEN
         nlcc = atom%potential%upf_pot%core_correction
         IF(.NOT.(.NOT. nlcc))CALL cp__a("atom_xc.F",411)
      END IF

      IF (ASSOCIATED(xc_section)) THEN
         NULLIFY (rho_set)
         IF (nlcc) THEN
            nr = atom%basis%grid%nr
            ALLOCATE (xfun(nr))
         END IF

         xc_fun_section => section_vals_get_subs_vals(xc_section, "XC_FUNCTIONAL")
         CALL section_vals_val_get(xc_fun_section, "_SECTION_PARAMETERS_", i_val=myfun)

         IF (myfun == xc_none) THEN
            atom%energy%exc = 0._dp
         ELSE
            CALL section_vals_val_get(xc_section, "DENSITY_CUTOFF", r_val=density_cut)
            CALL section_vals_val_get(xc_section, "GRADIENT_CUTOFF", r_val=gradient_cut)
            CALL section_vals_val_get(xc_section, "TAU_CUTOFF", r_val=tau_cut)

            lsd = .TRUE.
            nspins = 2
            needs = xc_functionals_get_needs(xc_fun_section, lsd=lsd, add_basic_components=.FALSE.)

            ! Prepare the structures needed to calculate and store the xc derivatives

            ! Array dimension: here anly one dimensional arrays are used,
            ! i.e. only the first column of deriv_data is read.
            ! The other to dimensions  are set to size equal 1
            nr = atom%basis%grid%nr
            bounds(1:2, 1:3) = 1
            bounds(2, 1) = nr

            ! create a place where to put the derivatives
            NULLIFY (deriv_set)
            CALL xc_dset_create(deriv_set, local_bounds=bounds)
            ! create the place where to store the argument for the functionals
            CALL xc_rho_set_create(rho_set, bounds, rho_cutoff=density_cut, &
                                   drho_cutoff=gradient_cut, tau_cutoff=tau_cut)
            ! allocate the required 3d arrays where to store rho and drho
            CALL xc_rho_set_atom_update(rho_set, needs, nspins, bounds)

            NULLIFY (rho, drho, tau)
            IF (needs%rho_spin) THEN
               ALLOCATE (rho(nr, 2))
               CALL atom_density(rho(:, 1), atom%orbitals%pmata, atom%basis, atom%state%maxl_occ, typ="RHO")
               CALL atom_density(rho(:, 2), atom%orbitals%pmatb, atom%basis, atom%state%maxl_occ, typ="RHO")
               IF (nlcc) THEN
                  xfun = 0.0_dp
                  CALL atom_core_density(xfun(:), atom%potential%gth_pot, typ="RHO", rr=atom%basis%grid%rad)
                  rho(:, 1) = rho(:, 1)+0.5_dp*xfun(:)
                  rho(:, 2) = rho(:, 2)+0.5_dp*xfun(:)
               END IF
            END IF
            IF (needs%norm_drho_spin) THEN
               ALLOCATE (drho(nr, 2))
               CALL atom_density(drho(:, 1), atom%orbitals%pmata, atom%basis, atom%state%maxl_occ, typ="DER")
               CALL atom_density(drho(:, 2), atom%orbitals%pmatb, atom%basis, atom%state%maxl_occ, typ="DER")
               IF (nlcc) THEN
                  xfun = 0.0_dp
                  CALL atom_core_density(xfun(:), atom%potential%gth_pot, typ="DER", rr=atom%basis%grid%rad)
                  drho(:, 1) = drho(:, 1)+0.5_dp*xfun(:)
                  drho(:, 2) = drho(:, 2)+0.5_dp*xfun(:)
               END IF
            END IF
            IF (needs%tau_spin) THEN
               ALLOCATE (tau(nr, 2))
               CALL atom_density(tau(:, 1), atom%orbitals%pmata, atom%basis, atom%state%maxl_occ, &
                                 typ="KIN", rr=atom%basis%grid%rad2)
               CALL atom_density(tau(:, 2), atom%orbitals%pmatb, atom%basis, atom%state%maxl_occ, &
                                 typ="KIN", rr=atom%basis%grid%rad2)
            END IF
            IF (needs%laplace_rho_spin) THEN
               ALLOCATE (lap(nr, 2))
               CALL atom_density(lap(:, 1), atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, &
                                 typ="LAP", rr=atom%basis%grid%rad2)
               CALL atom_density(lap(:, 2), atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, &
                                 typ="LAP", rr=atom%basis%grid%rad2)
               IF (nlcc) THEN
                  xfun = 0.0_dp
                  CALL atom_core_density(xfun(:), atom%potential%gth_pot, typ="LAP", rr=atom%basis%grid%rad)
                  lap(:, 1) = lap(:, 1)+0.5_dp*xfun(:)
                  lap(:, 2) = lap(:, 2)+0.5_dp*xfun(:)
               END IF
            END IF

            CALL fill_rho_set(rho_set, nspins, needs, rho, drho, tau, lap, nr)

            CALL xc_dset_zero_all(deriv_set)

            deriv_order = 1
            CALL xc_functionals_eval(xc_fun_section, lsd=lsd, rho_set=rho_set, deriv_set=deriv_set, &
                                     deriv_order=deriv_order)

            ! Integration to get the matrix elements and energy
            deriv => xc_dset_get_derivative(deriv_set, "", allocate_deriv=.FALSE.)
            CALL xc_derivative_get(deriv, deriv_data=xcpot)
            atom%energy%exc = fourpi*integrate_grid(xcpot(:, 1, 1), atom%basis%grid)

            IF (needs%rho_spin) THEN
               deriv => xc_dset_get_derivative(deriv_set, "(rhoa)", allocate_deriv=.FALSE.)
               CALL xc_derivative_get(deriv, deriv_data=xcpot)
               CALL numpot_matrix(xcmata%op, xcpot(:, 1, 1), atom%basis, 0)
               deriv => xc_dset_get_derivative(deriv_set, "(rhob)", allocate_deriv=.FALSE.)
               CALL xc_derivative_get(deriv, deriv_data=xcpot)
               CALL numpot_matrix(xcmatb%op, xcpot(:, 1, 1), atom%basis, 0)
               DEALLOCATE (rho)
            END IF
            IF (needs%norm_drho_spin) THEN
               ! drhoa
               NULLIFY (deriv)
               deriv => xc_dset_get_derivative(deriv_set, "(norm_drhoa)", allocate_deriv=.FALSE.)
               CALL xc_derivative_get(deriv, deriv_data=xcpot)
               CALL numpot_matrix(xcmata%op, xcpot(:, 1, 1), atom%basis, 1)
               ! drhob
               NULLIFY (deriv)
               deriv => xc_dset_get_derivative(deriv_set, "(norm_drhob)", allocate_deriv=.FALSE.)
               CALL xc_derivative_get(deriv, deriv_data=xcpot)
               CALL numpot_matrix(xcmatb%op, xcpot(:, 1, 1), atom%basis, 1)
               ! Cross Terms
               NULLIFY (deriv)
               deriv => xc_dset_get_derivative(deriv_set, "(norm_drho)")
               IF (ASSOCIATED(deriv)) THEN
                  CALL xc_derivative_get(deriv, deriv_data=xcpot)
                  CALL numpot_matrix(xcmata%op, xcpot(:, 1, 1), atom%basis, 1)
                  CALL numpot_matrix(xcmatb%op, xcpot(:, 1, 1), atom%basis, 1)
               END IF
               DEALLOCATE (drho)
            END IF
            IF (needs%tau_spin) THEN
               n1 = SIZE(xcmata%op, 1)
               n2 = SIZE(xcmata%op, 2)
               n3 = SIZE(xcmata%op, 3)
               ALLOCATE (taumat(n1, n2, 0:n3-1))

               deriv => xc_dset_get_derivative(deriv_set, "(tau_a)", allocate_deriv=.FALSE.)
               CALL xc_derivative_get(deriv, deriv_data=xcpot)
               taumat = 0._dp
               xcpot(:, 1, 1) = 0.5_dp*xcpot(:, 1, 1)
               CALL numpot_matrix(xcmata%op, xcpot(:, 1, 1), atom%basis, 2)
               xcpot(:, 1, 1) = xcpot(:, 1, 1)/atom%basis%grid%rad2(:)
               CALL numpot_matrix(taumat, xcpot(:, 1, 1), atom%basis, 0)
               DO l = 0, 3
                  xcmata%op(:, :, l) = xcmata%op(:, :, l)+REAL(l*(l+1), dp)*taumat(:, :, l)
               END DO

               deriv => xc_dset_get_derivative(deriv_set, "(tau_b)", allocate_deriv=.FALSE.)
               CALL xc_derivative_get(deriv, deriv_data=xcpot)
               taumat = 0._dp
               xcpot(:, 1, 1) = 0.5_dp*xcpot(:, 1, 1)
               CALL numpot_matrix(xcmatb%op, xcpot(:, 1, 1), atom%basis, 2)
               xcpot(:, 1, 1) = xcpot(:, 1, 1)/atom%basis%grid%rad2(:)
               CALL numpot_matrix(taumat, xcpot(:, 1, 1), atom%basis, 0)
               DO l = 0, 3
                  xcmatb%op(:, :, l) = xcmatb%op(:, :, l)+REAL(l*(l+1), dp)*taumat(:, :, l)
               END DO

               DEALLOCATE (tau)
               DEALLOCATE (taumat)
            END IF

            ! Release the xc structure used to store the xc derivatives
            CALL xc_dset_release(deriv_set)
            CALL xc_rho_set_release(rho_set)

         END IF !xc_none

         IF (nlcc) DEALLOCATE (xfun)

      ELSE

         ! we don't have an xc_section, use a default setup
         nr = atom%basis%grid%nr
         ALLOCATE (rho(nr, 2), exc(nr), vxca(nr), vxcb(nr))
         IF (nlcc) ALLOCATE (xfun(nr))

         CALL atom_density(rho(:, 1), atom%orbitals%pmata, atom%basis, atom%state%maxl_occ, typ="RHO")
         CALL atom_density(rho(:, 2), atom%orbitals%pmatb, atom%basis, atom%state%maxl_occ, typ="RHO")
         IF (nlcc) THEN
            xfun(:) = 0.0_dp
            CALL atom_core_density(xfun(:), atom%potential%gth_pot, typ="RHO", rr=atom%basis%grid%rad)
            rho(:, 1) = rho(:, 1)+0.5_dp*xfun(:)
            rho(:, 2) = rho(:, 2)+0.5_dp*xfun(:)
         END IF
         CALL lsd_pade(rho(:, 1), rho(:, 2), exc, vxca, vxcb)

         atom%energy%exc = fourpi*integrate_grid(exc, atom%basis%grid)
         CALL numpot_matrix(xcmata%op, vxca, atom%basis, 0)
         CALL numpot_matrix(xcmatb%op, vxcb, atom%basis, 0)

         DEALLOCATE (rho, exc, vxca, vxcb)
         IF (nlcc) DEALLOCATE (xfun)

      END IF

      CALL timestop(handle)

   END SUBROUTINE calculate_atom_vxc_lsd

! **************************************************************************************************
!> \brief ...
!> \param rho_set ...
!> \param nspins ...
!> \param needs ...
!> \param rho ...
!> \param drho ...
!> \param tau ...
!> \param lap ...
!> \param na ...
! **************************************************************************************************
   SUBROUTINE fill_rho_set(rho_set, nspins, needs, rho, drho, tau, lap, na)

      TYPE(xc_rho_set_type), POINTER                     :: rho_set
      INTEGER, INTENT(IN)                                :: nspins
      TYPE(xc_rho_cflags_type), INTENT(in)               :: needs
      REAL(dp), DIMENSION(:, :), POINTER                 :: rho, drho, tau, lap
      INTEGER, INTENT(IN)                                :: na

      CHARACTER(LEN=*), PARAMETER :: routineN = 'fill_rho_set', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: f13 = (1.0_dp/3.0_dp)

      INTEGER                                            :: ia

      SELECT CASE (nspins)
      CASE (1)
         IF(.NOT.(.NOT. needs%rho_spin))CALL cp__a("atom_xc.F",636)
         IF(.NOT.(.NOT. needs%drho_spin))CALL cp__a("atom_xc.F",637)
         IF(.NOT.(.NOT. needs%norm_drho_spin))CALL cp__a("atom_xc.F",638)
         IF(.NOT.(.NOT. needs%drhoa_drhob))CALL cp__a("atom_xc.F",639)
         IF(.NOT.(.NOT. needs%rho_spin_1_3))CALL cp__a("atom_xc.F",640)
         IF(.NOT.(.NOT. needs%tau_spin))CALL cp__a("atom_xc.F",641)
         IF(.NOT.(.NOT. needs%drho))CALL cp__a("atom_xc.F",642)
         ! Give rho to 1/3
         IF (needs%rho_1_3) THEN
            DO ia = 1, na
               rho_set%rho_1_3(ia, 1, 1) = MAX(rho(ia, 1), 0.0_dp)**f13
            END DO
            rho_set%owns%rho_1_3 = .TRUE.
            rho_set%has%rho_1_3 = .TRUE.
         END IF
         ! Give the density
         IF (needs%rho) THEN
            DO ia = 1, na
               rho_set%rho(ia, 1, 1) = rho(ia, 1)
            END DO
            rho_set%owns%rho = .TRUE.
            rho_set%has%rho = .TRUE.
         END IF
         ! Give the norm of the gradient of the density
         IF (needs%norm_drho) THEN
            DO ia = 1, na
               rho_set%norm_drho(ia, 1, 1) = drho(ia, 1)
            END DO
            rho_set%owns%norm_drho = .TRUE.
            rho_set%has%norm_drho = .TRUE.
         END IF
      CASE (2)
         IF(.NOT.(.NOT. needs%drho))CALL cp__a("atom_xc.F",668)
         IF(.NOT.(.NOT. needs%drho_spin))CALL cp__a("atom_xc.F",669)
         IF(.NOT.(.NOT. needs%drhoa_drhob))CALL cp__a("atom_xc.F",670)
         ! Give the total density
         IF (needs%rho) THEN
            DO ia = 1, na
               rho_set%rho(ia, 1, 1) = rho(ia, 1)+rho(ia, 2)
            END DO
            rho_set%owns%rho = .TRUE.
            rho_set%has%rho = .TRUE.
         END IF
         ! Give the norm of the total gradient of the density
         IF (needs%norm_drho) THEN
            DO ia = 1, na
               rho_set%norm_drho(ia, 1, 1) = drho(ia, 1)+drho(ia, 2)
            END DO
            rho_set%owns%norm_drho = .TRUE.
            rho_set%has%norm_drho = .TRUE.
         END IF
         ! Give rho_spin
         IF (needs%rho_spin) THEN
            DO ia = 1, na
               rho_set%rhoa(ia, 1, 1) = rho(ia, 1)
               rho_set%rhob(ia, 1, 1) = rho(ia, 2)
            END DO
            rho_set%owns%rho_spin = .TRUE.
            rho_set%has%rho_spin = .TRUE.
         END IF
         ! Give rho_spin to 1/3
         IF (needs%rho_spin_1_3) THEN
            DO ia = 1, na
               rho_set%rhoa_1_3(ia, 1, 1) = MAX(rho(ia, 1), 0.0_dp)**f13
               rho_set%rhob_1_3(ia, 1, 1) = MAX(rho(ia, 2), 0.0_dp)**f13
            END DO
            rho_set%owns%rho_1_3 = .TRUE.
            rho_set%has%rho_1_3 = .TRUE.
         END IF
         ! Give the norm of the gradient of rhoa and of rhob separatedly
         IF (needs%norm_drho_spin) THEN
            DO ia = 1, na
               rho_set%norm_drhoa(ia, 1, 1) = drho(ia, 1)
               rho_set%norm_drhob(ia, 1, 1) = drho(ia, 2)
            END DO
            rho_set%owns%norm_drho_spin = .TRUE.
            rho_set%has%norm_drho_spin = .TRUE.
         END IF
         !
      END SELECT

      ! tau part
      IF (needs%tau) THEN
         IF (nspins == 2) THEN
            DO ia = 1, na
               rho_set%tau(ia, 1, 1) = tau(ia, 1)+tau(ia, 2)
            END DO
            rho_set%owns%tau = .TRUE.
            rho_set%has%tau = .TRUE.
         ELSE
            DO ia = 1, na
               rho_set%tau(ia, 1, 1) = tau(ia, 1)
            END DO
            rho_set%owns%tau = .TRUE.
            rho_set%has%tau = .TRUE.
         END IF
      END IF
      IF (needs%tau_spin) THEN
         IF(.NOT.(nspins == 2))CALL cp__a("atom_xc.F",734)
         DO ia = 1, na
            rho_set%tau_a(ia, 1, 1) = tau(ia, 1)
            rho_set%tau_b(ia, 1, 1) = tau(ia, 2)
         END DO
         rho_set%owns%tau_spin = .TRUE.
         rho_set%has%tau_spin = .TRUE.
      END IF

      ! Laplace
      IF (needs%laplace_rho) THEN
         IF (nspins == 2) THEN
            DO ia = 1, na
               rho_set%laplace_rho(ia, 1, 1) = lap(ia, 1)+lap(ia, 2)
            END DO
            rho_set%owns%laplace_rho = .TRUE.
            rho_set%has%laplace_rho = .TRUE.
         ELSE
            DO ia = 1, na
               rho_set%laplace_rho(ia, 1, 1) = lap(ia, 1)
            END DO
            rho_set%owns%laplace_rho = .TRUE.
            rho_set%has%laplace_rho = .TRUE.
         END IF
      END IF
      IF (needs%laplace_rho_spin) THEN
         IF(.NOT.(nspins == 2))CALL cp__a("atom_xc.F",760)
         DO ia = 1, na
            rho_set%laplace_rhoa(ia, 1, 1) = lap(ia, 1)
            rho_set%laplace_rhob(ia, 1, 1) = lap(ia, 2)
         END DO
         rho_set%owns%laplace_rho_spin = .TRUE.
         rho_set%has%laplace_rho_spin = .TRUE.
      END IF

   END SUBROUTINE fill_rho_set
! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param rho ...
!> \param exc ...
!> \param vxc ...
! **************************************************************************************************
   SUBROUTINE lda_pade(rho, exc, vxc)

      REAL(dp), DIMENSION(:)                             :: rho, exc, vxc

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lda_pade', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER :: a0 = 0.4581652932831429E+0_dp, a1 = 0.2217058676663745E+1_dp, &
         a2 = 0.7405551735357053E+0_dp, a3 = 0.1968227878617998E-1_dp, &
         b1 = 1.0000000000000000E+0_dp, b2 = 0.4504130959426697E+1_dp, &
         b3 = 0.1110667363742916E+1_dp, b4 = 0.2359291751427506E-1_dp, f13 = (1.0_dp/3.0_dp), &
         rsfac = 0.6203504908994000166680065_dp

      INTEGER                                            :: i, n
      REAL(KIND=dp)                                      :: depade, dpv, dq, epade, p, q, rs

      n = SIZE(rho)
      exc(1:n) = 0._dp
      vxc(1:n) = 0._dp

      DO i = 1, n
         IF (rho(i) > 1.e-20_dp) THEN
            rs = rsfac*rho(i)**(-f13)
            p = a0+(a1+(a2+a3*rs)*rs)*rs
            q = (b1+(b2+(b3+b4*rs)*rs)*rs)*rs
            epade = -p/q

            dpv = a1+(2.0_dp*a2+3.0_dp*a3*rs)*rs
            dq = b1+(2.0_dp*b2+(3.0_dp*b3+4.0_dp*b4*rs)*rs)*rs
            depade = f13*rs*(dpv*q-p*dq)/(q*q)

            exc(i) = epade*rho(i)
            vxc(i) = epade+depade
         END IF
      END DO

   END SUBROUTINE lda_pade
! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param rhoa ...
!> \param rhob ...
!> \param exc ...
!> \param vxca ...
!> \param vxcb ...
! **************************************************************************************************
   SUBROUTINE lsd_pade(rhoa, rhob, exc, vxca, vxcb)

      REAL(dp), DIMENSION(:)                             :: rhoa, rhob, exc, vxca, vxcb

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lsd_pade', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER :: a0 = 0.4581652932831429E+0_dp, a1 = 0.2217058676663745E+1_dp, &
         a2 = 0.7405551735357053E+0_dp, a3 = 0.1968227878617998E-1_dp, &
         b1 = 1.0000000000000000E+0_dp, b2 = 0.4504130959426697E+1_dp, &
         b3 = 0.1110667363742916E+1_dp, b4 = 0.2359291751427506E-1_dp, &
         da0 = 0.119086804055547E+0_dp, da1 = 0.6157402568883345E+0_dp, &
         da2 = 0.1574201515892867E+0_dp, da3 = 0.3532336663397157E-2_dp, &
         db1 = 0.0000000000000000E+0_dp, db2 = 0.2673612973836267E+0_dp, &
         db3 = 0.2052004607777787E+0_dp, db4 = 0.4200005045691381E-2_dp, f13 = (1.0_dp/3.0_dp), &
         f43 = (4.0_dp/3.0_dp)
      REAL(KIND=dp), PARAMETER :: fxfac = 1.923661050931536319759455_dp, &
         rsfac = 0.6203504908994000166680065_dp

      INTEGER                                            :: i, n
      REAL(KIND=dp)                                      :: dc, dpv, dq, dr, dx, fa0, fa1, fa2, fa3, &
                                                            fb1, fb2, fb3, fb4, fx1, fx2, p, q, &
                                                            rhoab, rs, x, xp, xq

! 1/(2^(4/3) - 2)

      n = SIZE(rhoa)
      exc(1:n) = 0._dp
      vxca(1:n) = 0._dp
      vxcb(1:n) = 0._dp

      DO i = 1, n
         rhoab = rhoa(i)+rhob(i)
         IF (rhoab > 1.e-20_dp) THEN
            rs = rsfac*rhoab**(-f13)

            x = (rhoa(i)-rhob(i))/rhoab
            IF (x < -1.0_dp) THEN
               fx1 = 1.0_dp
               fx2 = -f43*fxfac*2.0_dp**f13
            ELSE IF (x > 1.0_dp) THEN
               fx1 = 1.0_dp
               fx2 = f43*fxfac*2.0_dp**f13
            ELSE
               fx1 = ((1.0_dp+x)**f43+(1.0_dp-x)**f43-2.0_dp)*fxfac
               fx2 = ((1.0_dp+x)**f13-(1.0_dp-x)**f13)*fxfac*f43
            END IF

            fa0 = a0+fx1*da0
            fa1 = a1+fx1*da1
            fa2 = a2+fx1*da2
            fa3 = a3+fx1*da3
            fb1 = b1+fx1*db1
            fb2 = b2+fx1*db2
            fb3 = b3+fx1*db3
            fb4 = b4+fx1*db4

            p = fa0+(fa1+(fa2+fa3*rs)*rs)*rs
            q = (fb1+(fb2+(fb3+fb4*rs)*rs)*rs)*rs
            dpv = fa1+(2.0_dp*fa2+3.0_dp*fa3*rs)*rs
            dq = fb1+(2.0_dp*fb2+(3.0_dp*fb3+ &
                                  4.0_dp*fb4*rs)*rs)*rs
            xp = da0+(da1+(da2+da3*rs)*rs)*rs
            xq = (db1+(db2+(db3+db4*rs)*rs)*rs)*rs

            dr = (dpv*q-p*dq)/(q*q)
            dx = 2.0_dp*(xp*q-p*xq)/(q*q)*fx2/rhoab
            dc = f13*rs*dr-p/q

            exc(i) = -p/q*rhoab
            vxca(i) = dc-dx*rhob(i)
            vxcb(i) = dc+dx*rhoa(i)
         END IF
      END DO

   END SUBROUTINE lsd_pade

! **************************************************************************************************

END MODULE atom_xc
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE atom_electronic_structure
   USE atom_optimization,               ONLY: atom_history_init,&
                                              atom_history_release,&
                                              atom_history_type,&
                                              atom_history_update,&
                                              atom_opt
   USE atom_output,                     ONLY: atom_print_energies,&
                                              atom_print_iteration,&
                                              atom_print_state,&
                                              atom_print_zmp_iteration
   USE atom_types,                      ONLY: &
        atom_type, create_opgrid, create_opmat, ecp_pseudo, gth_pseudo, no_pseudo, opgrid_type, &
        opmat_type, release_opgrid, release_opmat, set_atom, upf_pseudo
   USE atom_utils,                      ONLY: &
        atom_denmat, atom_density, atom_read_external_density, atom_read_external_vxc, &
        atom_read_zmp_restart, atom_solve, atom_trace, ceri_contract, coulomb_potential_analytic, &
        coulomb_potential_numeric, eeri_contract, err_matrix, exchange_numeric, &
        exchange_semi_analytic, numpot_matrix, slater_density, wigner_slater_functional
! ZMP added calculate_atom_zmp, calculate_atom_ext_vxc
   USE atom_xc,                         ONLY: calculate_atom_ext_vxc,&
                                              calculate_atom_vxc_lda,&
                                              calculate_atom_vxc_lsd,&
                                              calculate_atom_zmp
   USE input_constants,                 ONLY: &
        do_analytic, do_dkh0_atom, do_dkh1_atom, do_dkh2_atom, do_dkh3_atom, do_dkh4_atom, &
        do_dkh5_atom, do_nonrel_atom, do_numeric, do_rhf_atom, do_rks_atom, do_rohf_atom, &
        do_sczoramp_atom, do_semi_analytic, do_uhf_atom, do_uks_atom, do_zoramp_atom
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC  :: calculate_atom

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_electronic_structure'

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param iw ...
!> \param noguess ...
! **************************************************************************************************
   SUBROUTINE calculate_atom(atom, iw, noguess)
      TYPE(atom_type), POINTER                           :: atom
      INTEGER, INTENT(IN)                                :: iw
      LOGICAL, INTENT(IN), OPTIONAL                      :: noguess

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_atom', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, method

      CALL timeset(routineN, handle)

      method = atom%method_type

      SELECT CASE (method)
      CASE (do_rks_atom, do_rhf_atom)
         CALL calculate_atom_restricted(atom, iw, noguess)
      CASE (do_uks_atom, do_uhf_atom)
         CALL calculate_atom_unrestricted(atom, iw, noguess)
      CASE (do_rohf_atom)
         CALL cp__b("atom_electronic_structure.F",76,"")
      CASE DEFAULT
         CALL cp__b("atom_electronic_structure.F",78,"")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE calculate_atom

! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param iw ...
!> \param noguess ...
! **************************************************************************************************
   SUBROUTINE calculate_atom_restricted(atom, iw, noguess)
      TYPE(atom_type), POINTER                           :: atom
      INTEGER, INTENT(IN)                                :: iw
      LOGICAL, INTENT(IN), OPTIONAL                      :: noguess

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_atom_restricted', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iter, l, max_iter, method, &
                                                            reltyp
      LOGICAL                                            :: do_hfx, doguess, need_vxc, need_x, &
                                                            need_xc, need_zmp
      REAL(KIND=dp)                                      :: deps, eps_scf, hf_frac
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ext_density, ext_vxc, tmp_dens
      TYPE(atom_history_type)                            :: history
      TYPE(opgrid_type), POINTER                         :: cpot, density
      TYPE(opmat_type), POINTER                          :: fmat, hcore, jmat, kmat, xcmat
      TYPE(section_vals_type), POINTER                   :: hfx_sections, xc_section

      CALL timeset(routineN, handle)

      IF (PRESENT(noguess)) THEN
         doguess = .NOT. noguess
      ELSE
         doguess = .TRUE.
      END IF

      hf_frac = 0._dp
      IF (ASSOCIATED(atom%xc_section)) THEN
         xc_section => atom%xc_section
         hfx_sections => section_vals_get_subs_vals(xc_section, "HF")
         CALL section_vals_get(hfx_sections, explicit=do_hfx)
         IF (do_hfx) THEN
            CALL section_vals_val_get(hfx_sections, "FRACTION", r_val=hf_frac)
         END IF
      ELSE
         NULLIFY (xc_section)
         do_hfx = .FALSE.
      END IF

      method = atom%method_type
      max_iter = atom%optimization%max_iter
      eps_scf = atom%optimization%eps_scf

      SELECT CASE (method)
      CASE DEFAULT
         CALL cp__b("atom_electronic_structure.F",137,"")
      CASE (do_rks_atom)
         need_x = do_hfx
         need_xc = .TRUE.
      CASE (do_uks_atom)
         CALL cp__b("atom_electronic_structure.F",142,"")
      CASE (do_rhf_atom)
         need_x = .TRUE.
         need_xc = .FALSE.
         hf_frac = 1._dp
      CASE (do_uhf_atom)
         CALL cp__b("atom_electronic_structure.F",148,"")
      CASE (do_rohf_atom)
         need_x = .TRUE.
         need_xc = .FALSE.
         hf_frac = 1._dp
         CALL cp__b("atom_electronic_structure.F",153,"")
      END SELECT

! ZMP starting to read external density for the zmp calculation
      need_zmp = atom%do_zmp
      IF (need_zmp) THEN
         need_x = .FALSE.
         need_xc = .FALSE.
         ALLOCATE (ext_density(atom%basis%grid%nr))
         ext_density = 0._dp
         CALL atom_read_external_density(ext_density, atom, iw)
      ENDIF

! ZMP starting to read external vxc for the zmp calculation
      need_vxc = atom%read_vxc

      IF (need_vxc) THEN
         need_x = .FALSE.
         need_xc = .FALSE.
         need_zmp = .FALSE.
         ALLOCATE (ext_vxc(atom%basis%grid%nr))
         ext_vxc = 0._dp
         CALL atom_read_external_vxc(ext_vxc, atom, iw)
      ENDIF

      ! check for relativistic method
      reltyp = atom%relativistic

      IF (iw > 0) CALL atom_print_state(atom%state, iw)

      NULLIFY (hcore)
      CALL create_opmat(hcore, atom%basis%nbas)
      ! Pseudopotentials
      SELECT CASE (atom%potential%ppot_type)
      CASE DEFAULT
         CALL cp__b("atom_electronic_structure.F",188,"")
      CASE (NO_PSEUDO)
         SELECT CASE (reltyp)
         CASE DEFAULT
            CALL cp__b("atom_electronic_structure.F",192,"")
         CASE (do_nonrel_atom)
            hcore%op = atom%integrals%kin-atom%zcore*atom%integrals%core
         CASE (do_zoramp_atom, do_sczoramp_atom)
            hcore%op = atom%integrals%kin+atom%integrals%tzora-atom%zcore*atom%integrals%core
         CASE (do_dkh0_atom, do_dkh1_atom, do_dkh2_atom, do_dkh3_atom, do_dkh4_atom, do_dkh5_atom)
            hcore%op = atom%integrals%hdkh
         END SELECT
      CASE (gth_pseudo, upf_pseudo, ecp_pseudo)
         hcore%op = atom%integrals%kin+atom%integrals%core+atom%integrals%hnl
      END SELECT
      ! add confinement potential (not included in relativistic transformations)
      IF (atom%potential%confinement) THEN
         hcore%op = hcore%op+atom%potential%acon*atom%integrals%conf
      END IF

      NULLIFY (fmat, jmat, kmat, xcmat)
      CALL create_opmat(fmat, atom%basis%nbas)
      CALL create_opmat(jmat, atom%basis%nbas)
      CALL create_opmat(kmat, atom%basis%nbas)
      CALL create_opmat(xcmat, atom%basis%nbas)

      NULLIFY (density, cpot)
      CALL create_opgrid(density, atom%basis%grid)
      CALL create_opgrid(cpot, atom%basis%grid)

      ! ZMP reading the file to restart
      IF (atom%doread) CALL atom_read_zmp_restart(atom, doguess, iw)

      IF (doguess) THEN
         ! initial guess
         ALLOCATE (tmp_dens(SIZE(density%op)))
         tmp_dens = 0
         CALL slater_density(density%op, tmp_dens, atom%z, atom%state, atom%basis%grid)
         density%op = density%op+tmp_dens
         DEALLOCATE (tmp_dens)
         CALL coulomb_potential_numeric(cpot%op, density%op, density%grid)
         CALL numpot_matrix(jmat%op, cpot%op, atom%basis, 0)
         CALL wigner_slater_functional(density%op, cpot%op)
         CALL numpot_matrix(xcmat%op, cpot%op, atom%basis, 0)
         fmat%op = hcore%op+jmat%op+xcmat%op
         CALL atom_solve(fmat%op, atom%integrals%utrans, atom%orbitals%wfn, atom%orbitals%ener, &
                         atom%basis%nbas, atom%integrals%nne, atom%state%maxl_calc)
      END IF
      CALL atom_denmat(atom%orbitals%pmat, atom%orbitals%wfn, atom%basis%nbas, atom%state%occupation, &
                       atom%state%maxl_occ, atom%state%maxn_occ)

      ! wavefunction history
      NULLIFY (history%dmat, history%hmat)
      CALL atom_history_init(history, atom%optimization, fmat%op)

      iter = 0
      DO !SCF Loop

         ! Kinetic energy
         atom%energy%ekin = atom_trace(atom%integrals%kin, atom%orbitals%pmat)

         ! Band energy
         atom%energy%eband = 0._dp
         DO l = 0, 3
            DO i = 1, MIN(SIZE(atom%state%occupation, 2), SIZE(atom%orbitals%ener, 1))
               atom%energy%eband = atom%energy%eband+atom%orbitals%ener(i, l)*atom%state%occupation(l, i)
            END DO
         END DO

         ! Pseudopotential energy
         SELECT CASE (atom%potential%ppot_type)
         CASE DEFAULT
            CALL cp__b("atom_electronic_structure.F",260,"")
         CASE (no_pseudo)
            atom%energy%eploc = 0._dp
            atom%energy%epnl = 0._dp
         CASE (gth_pseudo, upf_pseudo, ecp_pseudo)
            atom%energy%eploc = atom_trace(atom%integrals%core, atom%orbitals%pmat)
            atom%energy%epnl = atom_trace(atom%integrals%hnl, atom%orbitals%pmat)
         END SELECT
         atom%energy%epseudo = atom%energy%eploc+atom%energy%epnl

         ! Core energy
         atom%energy%ecore = atom_trace(hcore%op, atom%orbitals%pmat)

         ! Confinement energy
         IF (atom%potential%confinement) THEN
            atom%energy%econfinement = atom_trace(atom%integrals%conf, atom%orbitals%pmat)
         ELSE
            atom%energy%econfinement = 0._dp
         END IF

         ! Hartree Term
         jmat%op = 0._dp
         SELECT CASE (atom%coulomb_integral_type)
         CASE DEFAULT
            CALL cp__b("atom_electronic_structure.F",284,"")
         CASE (do_analytic)
            CALL ceri_contract(jmat%op, atom%integrals%ceri, atom%orbitals%pmat, atom%integrals%n)
         CASE (do_semi_analytic)
            CALL coulomb_potential_analytic(cpot%op, atom%orbitals%pmat, atom%basis, atom%basis%grid, &
                                            atom%state%maxl_occ)
            CALL numpot_matrix(jmat%op, cpot%op, atom%basis, 0)
         CASE (do_numeric)
            CALL atom_density(density%op, atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, typ="RHO")
            CALL coulomb_potential_numeric(cpot%op, density%op, density%grid)
            CALL numpot_matrix(jmat%op, cpot%op, atom%basis, 0)
         END SELECT
         atom%energy%ecoulomb = 0.5_dp*atom_trace(jmat%op, atom%orbitals%pmat)

         ! Exchange Term
         IF (need_x) THEN
            kmat%op = 0._dp
            SELECT CASE (atom%exchange_integral_type)
            CASE DEFAULT
               CALL cp__b("atom_electronic_structure.F",303,"")
            CASE (do_analytic)
               CALL eeri_contract(kmat%op, atom%integrals%eeri, atom%orbitals%pmat, atom%integrals%n)
            CASE (do_semi_analytic)
               CALL exchange_semi_analytic(kmat%op, atom%state, atom%state%occupation, atom%orbitals%wfn, atom%basis)
            CASE (do_numeric)
               CALL exchange_numeric(kmat%op, atom%state, atom%state%occupation, atom%orbitals%wfn, atom%basis)
            END SELECT
            atom%energy%eexchange = hf_frac*0.5_dp*atom_trace(kmat%op, atom%orbitals%pmat)
            kmat%op = hf_frac*kmat%op
         ELSE
            kmat%op = 0._dp
            atom%energy%eexchange = 0._dp
         END IF

         ! XC
         IF (need_xc) THEN
            xcmat%op = 0._dp
            CALL calculate_atom_vxc_lda(xcmat, atom, xc_section)
            ! ZMP added options for the zmp calculations, building external density and vxc potential
         ELSEIF (need_zmp) THEN
            xcmat%op = 0._dp
            CALL calculate_atom_zmp(ext_density=ext_density, atom=atom, lprint=.FALSE., xcmat=xcmat)
         ELSEIF (need_vxc) THEN
            xcmat%op = 0._dp
            CALL calculate_atom_ext_vxc(vxc=ext_vxc, atom=atom, lprint=.FALSE., xcmat=xcmat)
         ELSE
            xcmat%op = 0._dp
            atom%energy%exc = 0._dp
         END IF

         ! Zero this contribution
         atom%energy%elsd = 0._dp

         ! Total energy
         atom%energy%etot = atom%energy%ecore+atom%energy%ecoulomb+atom%energy%eexchange+atom%energy%exc

         ! Potential energy
         atom%energy%epot = atom%energy%etot-atom%energy%ekin

         ! Total HF/KS matrix
         fmat%op = hcore%op+jmat%op+kmat%op+xcmat%op

         ! calculate error matrix
         CALL err_matrix(jmat%op, deps, fmat%op, atom%orbitals%pmat, atom%integrals%utrans, &
                         atom%integrals%uptrans, atom%basis%nbas, atom%integrals%nne)

         iter = iter+1

         IF (iw > 0) THEN
            IF (need_zmp) THEN
               CALL atom_print_zmp_iteration(iter, deps, atom, iw)
            ELSE
               CALL atom_print_iteration(iter, deps, atom%energy%etot, iw)
            ENDIF
         END IF

         IF (deps < eps_scf) EXIT
         IF (iter >= max_iter) THEN
            IF (iw > 0) THEN
               WRITE (iw, "(A)") " No convergence within maximum number of iterations "
            END IF
            EXIT
         END IF

         ! update history container and extrapolate KS matrix
         CALL atom_history_update(history, fmat%op, jmat%op)
         CALL atom_opt(fmat%op, history, deps)

         ! Solve HF/KS equations
         CALL atom_solve(fmat%op, atom%integrals%utrans, atom%orbitals%wfn, atom%orbitals%ener, &
                         atom%basis%nbas, atom%integrals%nne, atom%state%maxl_calc)
         CALL atom_denmat(atom%orbitals%pmat, atom%orbitals%wfn, atom%basis%nbas, atom%state%occupation, &
                          atom%state%maxl_occ, atom%state%maxn_occ)

      END DO !SCF Loop

      IF (iw > 0) THEN
         CALL atom_print_energies(atom, iw)
      END IF

      CALL atom_history_release(history)

      ! conserve fmat within atom_type
      CALL set_atom(atom, fmat=fmat)
      CALL release_opmat(jmat)
      CALL release_opmat(kmat)
      CALL release_opmat(xcmat)
      CALL release_opmat(hcore)

      CALL release_opgrid(density)
      CALL release_opgrid(cpot)

      ! ZMP deallocating ext_density ext_vxc
      IF (need_zmp) DEALLOCATE (ext_density)
      IF (need_vxc) DEALLOCATE (ext_vxc)

      CALL timestop(handle)

   END SUBROUTINE calculate_atom_restricted

! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param iw ...
!> \param noguess ...
! **************************************************************************************************
   SUBROUTINE calculate_atom_unrestricted(atom, iw, noguess)
      TYPE(atom_type), POINTER                           :: atom
      INTEGER, INTENT(IN)                                :: iw
      LOGICAL, INTENT(IN), OPTIONAL                      :: noguess

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_atom_unrestricted', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iter, k, l, max_iter, method, &
                                                            reltyp
      LOGICAL                                            :: do_hfx, doguess, lsdpot, need_x, need_xc
      REAL(KIND=dp)                                      :: deps, depsa, depsb, eps_scf, hf_frac, &
                                                            ne, nm
      TYPE(atom_history_type)                            :: historya, historyb
      TYPE(opgrid_type), POINTER                         :: cpot, density, rhoa, rhob
      TYPE(opmat_type), POINTER                          :: fmata, fmatb, hcore, hlsd, jmat, kmata, &
                                                            kmatb, xcmata, xcmatb
      TYPE(section_vals_type), POINTER                   :: hfx_sections, xc_section

      CALL timeset(routineN, handle)

      IF (PRESENT(noguess)) THEN
         doguess = .NOT. noguess
      ELSE
         doguess = .TRUE.
      END IF

      hf_frac = 0._dp
      IF (ASSOCIATED(atom%xc_section)) THEN
         xc_section => atom%xc_section
         hfx_sections => section_vals_get_subs_vals(xc_section, "HF")
         CALL section_vals_get(hfx_sections, explicit=do_hfx)
         IF (do_hfx) THEN
            CALL section_vals_val_get(hfx_sections, "FRACTION", r_val=hf_frac)
         END IF
      ELSE
         NULLIFY (xc_section)
         do_hfx = .FALSE.
      END IF

      method = atom%method_type
      max_iter = atom%optimization%max_iter
      eps_scf = atom%optimization%eps_scf

      SELECT CASE (method)
      CASE DEFAULT
         CALL cp__b("atom_electronic_structure.F",456,"")
      CASE (do_rks_atom)
         CALL cp__b("atom_electronic_structure.F",458,"")
      CASE (do_uks_atom)
         need_x = do_hfx
         need_xc = .TRUE.
      CASE (do_rhf_atom)
         CALL cp__b("atom_electronic_structure.F",463,"")
      CASE (do_uhf_atom)
         need_x = .TRUE.
         need_xc = .FALSE.
         hf_frac = 1._dp
      CASE (do_rohf_atom)
         need_x = .TRUE.
         need_xc = .FALSE.
         hf_frac = 1._dp
         CALL cp__b("atom_electronic_structure.F",472,"")
      END SELECT

      ! set alpha and beta occupations
      atom%state%occa = 0._dp
      atom%state%occb = 0._dp
      DO l = 0, 3
         nm = REAL((2*l+1), KIND=dp)
         DO k = 1, 10
            ne = atom%state%occupation(l, k)
            IF (ne == 0._dp) THEN !empty shell
               EXIT !assume there are no holes
            ELSEIF (ne == 2._dp*nm) THEN !closed shell
               atom%state%occa(l, k) = nm
               atom%state%occb(l, k) = nm
            ELSEIF (atom%state%multiplicity == -2) THEN !High spin case
               atom%state%occa(l, k) = MIN(ne, nm)
               atom%state%occb(l, k) = MAX(0._dp, ne-nm)
            ELSE
               atom%state%occa(l, k) = 0.5_dp*(ne+atom%state%multiplicity-1._dp)
               atom%state%occb(l, k) = ne-atom%state%occa(l, k)
            END IF
         END DO
      END DO
      ! check for relativistic method
      reltyp = atom%relativistic

      IF (iw > 0) CALL atom_print_state(atom%state, iw)

      NULLIFY (hcore, hlsd)
      CALL create_opmat(hcore, atom%basis%nbas)
      CALL create_opmat(hlsd, atom%basis%nbas)
      hlsd%op = 0._dp
      ! Pseudopotentials
      lsdpot = .FALSE.
      SELECT CASE (atom%potential%ppot_type)
      CASE DEFAULT
         CALL cp__b("atom_electronic_structure.F",509,"")
      CASE (no_pseudo)
         SELECT CASE (reltyp)
         CASE DEFAULT
            CALL cp__b("atom_electronic_structure.F",513,"")
         CASE (do_nonrel_atom)
            hcore%op = atom%integrals%kin-atom%zcore*atom%integrals%core
         CASE (do_zoramp_atom, do_sczoramp_atom)
            hcore%op = atom%integrals%kin+atom%integrals%tzora-atom%zcore*atom%integrals%core
         CASE (do_dkh0_atom, do_dkh1_atom, do_dkh2_atom, do_dkh3_atom, do_dkh4_atom, do_dkh5_atom)
            hcore%op = atom%integrals%hdkh
         END SELECT
      CASE (gth_pseudo)
         hcore%op = atom%integrals%kin+atom%integrals%core+atom%integrals%hnl
         IF (atom%potential%gth_pot%lsdpot) THEN
            lsdpot = .TRUE.
            hlsd%op = atom%integrals%clsd
         END IF
      CASE (upf_pseudo, ecp_pseudo)
         hcore%op = atom%integrals%kin+atom%integrals%core+atom%integrals%hnl
      END SELECT
      ! add confinement potential (not included in relativistic transformations)
      IF (atom%potential%confinement) THEN
         hcore%op = hcore%op+atom%potential%acon*atom%integrals%conf
      END IF

      NULLIFY (fmata, fmatb, jmat, kmata, kmatb, xcmata, xcmatb)
      CALL create_opmat(fmata, atom%basis%nbas)
      CALL create_opmat(fmatb, atom%basis%nbas)
      CALL create_opmat(jmat, atom%basis%nbas)
      CALL create_opmat(kmata, atom%basis%nbas)
      CALL create_opmat(kmatb, atom%basis%nbas)
      CALL create_opmat(xcmata, atom%basis%nbas)
      CALL create_opmat(xcmatb, atom%basis%nbas)

      NULLIFY (density, rhoa, rhob, cpot)
      CALL create_opgrid(density, atom%basis%grid)
      CALL create_opgrid(rhoa, atom%basis%grid)
      CALL create_opgrid(rhob, atom%basis%grid)
      CALL create_opgrid(cpot, atom%basis%grid)

      IF (doguess) THEN
         ! initial guess
         CALL slater_density(rhoa%op, rhob%op, atom%z, atom%state, atom%basis%grid)
         density%op = rhoa%op+rhob%op
         CALL coulomb_potential_numeric(cpot%op, density%op, density%grid)
         CALL numpot_matrix(jmat%op, cpot%op, atom%basis, 0)
         ! alpha spin
         density%op = 2._dp*rhoa%op
         CALL wigner_slater_functional(density%op, cpot%op)
         CALL numpot_matrix(xcmata%op, cpot%op, atom%basis, 0)
         fmata%op = hcore%op+hlsd%op+jmat%op+xcmata%op
         CALL atom_solve(fmata%op, atom%integrals%utrans, atom%orbitals%wfna, atom%orbitals%enera, &
                         atom%basis%nbas, atom%integrals%nne, atom%state%maxl_calc)
         ! beta spin
         density%op = 2._dp*rhob%op
         CALL wigner_slater_functional(density%op, cpot%op)
         CALL numpot_matrix(xcmatb%op, cpot%op, atom%basis, 0)
         fmatb%op = hcore%op-hlsd%op+jmat%op+xcmatb%op
         CALL atom_solve(fmatb%op, atom%integrals%utrans, atom%orbitals%wfnb, atom%orbitals%enerb, &
                         atom%basis%nbas, atom%integrals%nne, atom%state%maxl_calc)
      END IF
      CALL atom_denmat(atom%orbitals%pmata, atom%orbitals%wfna, atom%basis%nbas, atom%state%occa, &
                       atom%state%maxl_occ, atom%state%maxn_occ)
      CALL atom_denmat(atom%orbitals%pmatb, atom%orbitals%wfnb, atom%basis%nbas, atom%state%occb, &
                       atom%state%maxl_occ, atom%state%maxn_occ)
      atom%orbitals%pmat = atom%orbitals%pmata+atom%orbitals%pmatb

      ! wavefunction history
      NULLIFY (historya%dmat, historya%hmat)
      CALL atom_history_init(historya, atom%optimization, fmata%op)
      NULLIFY (historyb%dmat, historyb%hmat)
      CALL atom_history_init(historyb, atom%optimization, fmatb%op)

      iter = 0
      DO !SCF Loop

         ! Kinetic energy
         atom%energy%ekin = atom_trace(atom%integrals%kin, atom%orbitals%pmat)

         ! Band energy
         atom%energy%eband = 0._dp
         DO l = 0, 3
            DO i = 1, MIN(SIZE(atom%state%occupation, 2), SIZE(atom%orbitals%ener, 1))
               atom%energy%eband = atom%energy%eband+atom%orbitals%enera(i, l)*atom%state%occa(l, i)
               atom%energy%eband = atom%energy%eband+atom%orbitals%enerb(i, l)*atom%state%occb(l, i)
            END DO
         END DO

         ! Pseudopotential energy
         SELECT CASE (atom%potential%ppot_type)
         CASE DEFAULT
            CALL cp__b("atom_electronic_structure.F",601,"")
         CASE (no_pseudo)
            atom%energy%eploc = 0._dp
            atom%energy%epnl = 0._dp
         CASE (gth_pseudo, upf_pseudo, ecp_pseudo)
            atom%energy%eploc = atom_trace(atom%integrals%core, atom%orbitals%pmat)
            atom%energy%epnl = atom_trace(atom%integrals%hnl, atom%orbitals%pmat)
         END SELECT
         atom%energy%epseudo = atom%energy%eploc+atom%energy%epnl

         ! Core energy
         atom%energy%ecore = atom_trace(hcore%op, atom%orbitals%pmat)

         ! Confinement energy
         IF (atom%potential%confinement) THEN
            atom%energy%econfinement = atom_trace(atom%integrals%conf, atom%orbitals%pmat)
         ELSE
            atom%energy%econfinement = 0._dp
         END IF

         ! Hartree Term
         jmat%op = 0._dp
         SELECT CASE (atom%coulomb_integral_type)
         CASE DEFAULT
            CALL cp__b("atom_electronic_structure.F",625,"")
         CASE (do_analytic)
            CALL ceri_contract(jmat%op, atom%integrals%ceri, atom%orbitals%pmat, atom%integrals%n)
         CASE (do_semi_analytic)
            CALL coulomb_potential_analytic(cpot%op, atom%orbitals%pmat, atom%basis, atom%basis%grid, &
                                            atom%state%maxl_occ)
            CALL numpot_matrix(jmat%op, cpot%op, atom%basis, 0)
         CASE (do_numeric)
            CALL atom_density(density%op, atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, typ="RHO")
            CALL coulomb_potential_numeric(cpot%op, density%op, density%grid)
            CALL numpot_matrix(jmat%op, cpot%op, atom%basis, 0)
         END SELECT
         atom%energy%ecoulomb = 0.5_dp*atom_trace(jmat%op, atom%orbitals%pmat)

         ! Exchange Term
         IF (need_x) THEN
            kmata%op = 0._dp
            kmatb%op = 0._dp
            SELECT CASE (atom%exchange_integral_type)
            CASE DEFAULT
               CALL cp__b("atom_electronic_structure.F",645,"")
            CASE (do_analytic)
               CALL eeri_contract(kmata%op, atom%integrals%eeri, atom%orbitals%pmata, atom%integrals%n)
               CALL eeri_contract(kmatb%op, atom%integrals%eeri, atom%orbitals%pmatb, atom%integrals%n)
            CASE (do_semi_analytic)
               CALL exchange_semi_analytic(kmata%op, atom%state, atom%state%occa, atom%orbitals%wfna, atom%basis)
               CALL exchange_semi_analytic(kmatb%op, atom%state, atom%state%occb, atom%orbitals%wfnb, atom%basis)
            CASE (do_numeric)
               CALL exchange_numeric(kmata%op, atom%state, atom%state%occa, atom%orbitals%wfna, atom%basis)
               CALL exchange_numeric(kmatb%op, atom%state, atom%state%occb, atom%orbitals%wfnb, atom%basis)
            END SELECT
            atom%energy%eexchange = hf_frac*(atom_trace(kmata%op, atom%orbitals%pmata)+ &
                                             atom_trace(kmatb%op, atom%orbitals%pmatb))
            kmata%op = 2._dp*hf_frac*kmata%op
            kmatb%op = 2._dp*hf_frac*kmatb%op
         ELSE
            kmata%op = 0._dp
            kmatb%op = 0._dp
            atom%energy%eexchange = 0._dp
         END IF

         ! XC
         IF (need_xc) THEN
            xcmata%op = 0._dp
            xcmatb%op = 0._dp
            CALL calculate_atom_vxc_lsd(xcmata, xcmatb, atom, xc_section)
         ELSE
            xcmata%op = 0._dp
            xcmatb%op = 0._dp
            atom%energy%exc = 0._dp
         END IF

         IF (lsdpot) THEN
            atom%energy%elsd = atom_trace(hlsd%op, atom%orbitals%pmata)- &
                               atom_trace(hlsd%op, atom%orbitals%pmatb)
            atom%energy%epseudo = atom%energy%epseudo+atom%energy%elsd
            atom%energy%ecore = atom%energy%ecore+atom%energy%elsd
         ELSE
            atom%energy%elsd = 0._dp
         END IF

         ! Total energy
         atom%energy%etot = atom%energy%ecore+atom%energy%ecoulomb+atom%energy%eexchange+atom%energy%exc

         ! Potential energy
         atom%energy%epot = atom%energy%etot-atom%energy%ekin

         ! Total HF/KS matrix
         fmata%op = hcore%op+hlsd%op+jmat%op+kmata%op+xcmata%op
         fmatb%op = hcore%op-hlsd%op+jmat%op+kmatb%op+xcmatb%op

         ! calculate error matrix
         CALL err_matrix(xcmata%op, depsa, fmata%op, atom%orbitals%pmata, atom%integrals%utrans, &
                         atom%integrals%uptrans, atom%basis%nbas, atom%integrals%nne)
         CALL err_matrix(xcmatb%op, depsb, fmatb%op, atom%orbitals%pmatb, atom%integrals%utrans, &
                         atom%integrals%uptrans, atom%basis%nbas, atom%integrals%nne)
         deps = 2._dp*MAX(depsa, depsb)

         iter = iter+1

         IF (iw > 0) THEN
            CALL atom_print_iteration(iter, deps, atom%energy%etot, iw)
         END IF

         IF (deps < eps_scf) EXIT
         IF (iter >= max_iter) THEN
            IF (iw > 0) THEN
               WRITE (iw, "(A)") " No convergence within maximum number of iterations "
            END IF
            EXIT
         END IF

         ! update history container and extrapolate KS matrix
         CALL atom_history_update(historya, fmata%op, xcmata%op)
         CALL atom_history_update(historyb, fmatb%op, xcmatb%op)
         CALL atom_opt(fmata%op, historya, depsa)
         CALL atom_opt(fmatb%op, historyb, depsb)

         ! Solve HF/KS equations
         CALL atom_solve(fmata%op, atom%integrals%utrans, atom%orbitals%wfna, atom%orbitals%enera, &
                         atom%basis%nbas, atom%integrals%nne, atom%state%maxl_calc)
         CALL atom_denmat(atom%orbitals%pmata, atom%orbitals%wfna, atom%basis%nbas, atom%state%occa, &
                          atom%state%maxl_occ, atom%state%maxn_occ)
         CALL atom_solve(fmatb%op, atom%integrals%utrans, atom%orbitals%wfnb, atom%orbitals%enerb, &
                         atom%basis%nbas, atom%integrals%nne, atom%state%maxl_calc)
         CALL atom_denmat(atom%orbitals%pmatb, atom%orbitals%wfnb, atom%basis%nbas, atom%state%occb, &
                          atom%state%maxl_occ, atom%state%maxn_occ)
         atom%orbitals%pmat = atom%orbitals%pmata+atom%orbitals%pmatb

      END DO !SCF Loop

      IF (iw > 0) THEN
         CALL atom_print_energies(atom, iw)
      END IF

      CALL atom_history_release(historya)
      CALL atom_history_release(historyb)

      CALL release_opgrid(density)
      CALL release_opgrid(rhoa)
      CALL release_opgrid(rhob)
      CALL release_opgrid(cpot)

      ! conserve fmata as fmat within atom_type.
      CALL set_atom(atom, fmat=fmata)
      CALL release_opmat(fmatb)
      CALL release_opmat(jmat)
      CALL release_opmat(kmata)
      CALL release_opmat(kmatb)
      CALL release_opmat(xcmata)
      CALL release_opmat(xcmatb)
      CALL release_opmat(hlsd)
      CALL release_opmat(hcore)

      CALL timestop(handle)

   END SUBROUTINE calculate_atom_unrestricted

! **************************************************************************************************

END MODULE atom_electronic_structure
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

!******************************************************************************
MODULE powell
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: twopi
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'powell'

  TYPE opt_state_type
    INTEGER            :: state
    INTEGER            :: nvar
    INTEGER            :: iprint
    INTEGER            :: unit
    INTEGER            :: maxfun
    REAL(dp)           :: rhobeg, rhoend
    REAL(dp), DIMENSION(:), POINTER  :: w
    REAL(dp), DIMENSION(:), POINTER  :: xopt
    ! local variables
    INTEGER            :: np, nh, nptm, nftest, idz, itest, nf, nfm, nfmm, &
                          nfsav, knew, kopt, ksave, ktemp
    REAL(dp)           :: rhosq, recip, reciq, fbeg, fopt, diffa, xoptsq, &
                          rho, delta, dsq, dnorm, ratio, temp, tempq, beta, &
                          dx, vquad, diff, diffc, diffb, fsave, detrat, hdiag, &
                          distsq, gisq, gqsq, f, bstep, alpha, dstep
  END TYPE opt_state_type

  PRIVATE
  PUBLIC      :: powell_optimize, opt_state_type

!******************************************************************************

CONTAINS

!******************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param x ...
!> \param optstate ...
! **************************************************************************************************
  SUBROUTINE powell_optimize (n,x,optstate)
      INTEGER                                            :: n
      REAL(dp), DIMENSION(*)                             :: x
      TYPE(opt_state_type)                               :: optstate

      CHARACTER(len=*), PARAMETER :: routineN = 'powell_optimize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npt

    CALL timeset(routineN,handle)

    SELECT CASE (optstate%state)
    CASE (0)
       npt=2*n+1
       ALLOCATE(optstate%w((npt+13)*(npt+n)+3*n*(n+3)/2))
       ALLOCATE(optstate%xopt(n))
       ! Initialize w
       optstate%w = 0.0_dp
       optstate%state = 1
       CALL newuoa (n, x, optstate)
    CASE (1,2)
       CALL newuoa (n, x, optstate)
    CASE (3)
       IF(optstate%unit > 0) THEN
         WRITE(optstate%unit,*) "POWELL| Exceeding maximum number of steps"
       ENDIF
       optstate%state = -1
    CASE (4)
       IF(optstate%unit > 0) THEN
         WRITE(optstate%unit,*) "POWELL| Error in trust region"
       ENDIF
       optstate%state = -1
    CASE (5)
       IF(optstate%unit > 0) THEN
         WRITE(optstate%unit,*) "POWELL| N out of range"
       ENDIF
       optstate%state = -1
    CASE (6,7)
       optstate%state = -1
    CASE (8)
       x(1:n) = optstate%xopt(1:n)
       DEALLOCATE(optstate%w)
       DEALLOCATE(optstate%xopt)
       optstate%state = -1
    CASE DEFAULT
       CALL cp__b("common/powell.F",94,"")
    END SELECT

    CALL timestop(handle)

  END SUBROUTINE powell_optimize
!******************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param x ...
!> \param optstate ...
! **************************************************************************************************
  SUBROUTINE newuoa (n,x,optstate)

      INTEGER                                            :: n
      REAL(dp), DIMENSION(*)                             :: x
      TYPE(opt_state_type)                               :: optstate

      INTEGER                                            :: ibmat, id, ifv, igq, ihq, ipq, ivl, iw, &
                                                            ixb, ixn, ixo, ixp, izmat, maxfun, &
                                                            ndim, np, npt, nptm
      REAL(dp)                                           :: rhobeg, rhoend

    maxfun = optstate%maxfun
    rhobeg = optstate%rhobeg
    rhoend = optstate%rhoend

    !
    !     This subroutine seeks the least value of a function of many variab
    !     by a trust region method that forms quadratic models by interpolat
    !     There can be some freedom in the interpolation conditions, which i
    !     taken up by minimizing the Frobenius norm of the change to the sec
    !     derivative of the quadratic model, beginning with a zero matrix. T
    !     arguments of the subroutine are as follows.
    !
    !     N must be set to the number of variables and must be at least two.
    !     NPT is the number of interpolation conditions. Its value must be i
    !       interval [N+2,(N+1)(N+2)/2].
    !     Initial values of the variables must be set in X(1),X(2),...,X(N).
    !       will be changed to the values that give the least calculated F.
    !     RHOBEG and RHOEND must be set to the initial and final values of a
    !       region radius, so both must be positive with RHOEND<=RHOBEG. Typ
    !       RHOBEG should be about one tenth of the greatest expected change
    !       variable, and RHOEND should indicate the accuracy that is requir
    !       the final values of the variables.
    !     The value of IPRINT should be set to 0, 1, 2 or 3, which controls
    !       amount of printing. Specifically, there is no output if IPRINT=0
    !       there is output only at the return if IPRINT=1. Otherwise, each
    !       value of RHO is printed, with the best vector of variables so fa
    !       the corresponding value of the objective function. Further, each
    !       value of F with its variables are output if IPRINT=3.
    !     MAXFUN must be set to an upper bound on the number of calls of CAL
    !     The array W will be used for working space. Its length must be at
    !     (NPT+13)*(NPT+N)+3*N*(N+3)/2.
    !
    !     SUBROUTINE CALFUN (N,X,F) must be provided by the user. It must se
    !     the value of the objective function for the variables X(1),X(2),..
    !
    !     Partition the working space array, so that different parts of it c
    !     treated separately by the subroutine that performs the main calcul
    !
    np=n+1
    npt=2*n+1
    nptm=npt-np
    IF (npt < n+2 .OR. npt > ((n+2)*np)/2) THEN
       optstate%state = 5
       RETURN
    END IF
    ndim=npt+n
    ixb=1
    ixo=ixb+n
    ixn=ixo+n
    ixp=ixn+n
    ifv=ixp+n*npt
    igq=ifv+npt
    ihq=igq+n
    ipq=ihq+(n*np)/2
    ibmat=ipq+npt
    izmat=ibmat+ndim*n
    id=izmat+npt*nptm
    ivl=id+n
    iw=ivl+ndim
    !
    !     The above settings provide a partition of W for subroutine NEWUOB.
    !     The partition requires the first NPT*(NPT+N)+5*N*(N+3)/2 elements
    !     W plus the space that is needed by the last array of NEWUOB.
    !
    CALL newuob (n,npt,x,rhobeg,rhoend,maxfun,optstate%w(ixb:),optstate%w(ixo:),&
         optstate%w(ixn:),optstate%w(ixp:),optstate%w(ifv:),optstate%w(igq:),optstate%w(ihq:),&
         optstate%w(ipq:),optstate%w(ibmat:),optstate%w(izmat:),ndim,optstate%w(id:),&
         optstate%w(ivl:),optstate%w(iw:),optstate)

    optstate%xopt(1:n) = optstate%w(ixb:ixb+n-1) + optstate%w(ixo:ixo+n-1)


  END SUBROUTINE newuoa

!******************************************************************************
! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param npt ...
!> \param x ...
!> \param rhobeg ...
!> \param rhoend ...
!> \param maxfun ...
!> \param xbase ...
!> \param xopt ...
!> \param xnew ...
!> \param xpt ...
!> \param fval ...
!> \param gq ...
!> \param hq ...
!> \param pq ...
!> \param bmat ...
!> \param zmat ...
!> \param ndim ...
!> \param d ...
!> \param vlag ...
!> \param w ...
!> \param opt ...
! **************************************************************************************************
  SUBROUTINE newuob (n,npt,x,rhobeg,rhoend,maxfun,xbase,&
       xopt,xnew,xpt,fval,gq,hq,pq,bmat,zmat,ndim,d,vlag,w,opt)

    INTEGER, INTENT(inout)                   :: n, npt
    REAL(dp), DIMENSION(1:n), INTENT(inout)  :: x
    REAL(dp), INTENT(inout)                  :: rhobeg, rhoend
    INTEGER, INTENT(inout)                   :: maxfun
    REAL(dp), DIMENSION(*), INTENT(inout)    :: xbase, xopt, xnew
    REAL(dp), DIMENSION(npt, *), &
      INTENT(inout)                          :: xpt
    REAL(dp), DIMENSION(*), INTENT(inout)    :: fval, gq, hq, pq
    INTEGER, INTENT(inout)                   :: ndim
    REAL(dp), DIMENSION(npt, *), &
      INTENT(inout)                          :: zmat
    REAL(dp), DIMENSION(ndim, *), &
      INTENT(inout)                          :: bmat
    REAL(dp), DIMENSION(*), INTENT(inout)    :: d, vlag, w
    TYPE(opt_state_type)                     :: opt

    INTEGER                                  :: i, idz, ih, ip, ipt, itemp, &
                                                itest, j, jp, jpt, k, knew, &
                                                kopt, ksave, ktemp, nf, nfm, &
                                                nfmm, nfsav, nftest, nh, np, &
                                                nptm
    REAL(dp) :: alpha, beta, bstep, bsum, crvmin, delta, detrat, diff, diffa, &
      diffb, diffc, distsq, dnorm, dsq, dstep, dx, f, fbeg, fopt, fsave, &
      gisq, gqsq, half, hdiag, one, ratio, recip, reciq, rho, rhosq, sum, &
      suma, sumb, sumz, temp, tempq, tenth, vquad, xipt, xjpt, xoptsq, zero

!
!     The arguments N, NPT, X, RHOBEG, RHOEND, IPRINT and MAXFUN are ide
!       to the corresponding arguments in SUBROUTINE NEWUOA.
!     XBASE will hold a shift of origin that should reduce the contribut
!       from rounding errors to values of the model and Lagrange functio
!     XOPT will be set to the displacement from XBASE of the vector of
!       variables that provides the least calculated F so far.
!     XNEW will be set to the displacement from XBASE of the vector of
!       variables for the current calculation of F.
!     XPT will contain the interpolation point coordinates relative to X
!     FVAL will hold the values of F at the interpolation points.
!     GQ will hold the gradient of the quadratic model at XBASE.
!     HQ will hold the explicit second derivatives of the quadratic mode
!     PQ will contain the parameters of the implicit second derivatives
!       the quadratic model.
!     BMAT will hold the last N columns of H.
!     ZMAT will hold the factorization of the leading NPT by NPT submatr
!       H, this factorization being ZMAT times Diag(DZ) times ZMAT^T, wh
!       the elements of DZ are plus or minus one, as specified by IDZ.
!     NDIM is the first dimension of BMAT and has the value NPT+N.
!     D is reserved for trial steps from XOPT.
!     VLAG will contain the values of the Lagrange functions at a new po
!       They are part of a product that requires VLAG to be of length ND
!     The array W will be used for working space. Its length must be at
!       10*NDIM = 10*(NPT+N).

    IF ( opt%state == 1 ) THEN
       ! initialize all variable that will be stored
       np           = 0
       nh           = 0
       nptm         = 0
       nftest       = 0
       idz          = 0
       itest        = 0
       nf           = 0
       nfm          = 0
       nfmm         = 0
       nfsav        = 0
       knew         = 0
       kopt         = 0
       ksave        = 0
       ktemp        = 0
       rhosq        = 0._dp
       recip        = 0._dp
       reciq        = 0._dp
       fbeg         = 0._dp
       fopt         = 0._dp
       diffa        = 0._dp
       xoptsq       = 0._dp
       rho          = 0._dp
       delta        = 0._dp
       dsq          = 0._dp
       dnorm        = 0._dp
       ratio        = 0._dp
       temp         = 0._dp
       tempq        = 0._dp
       beta         = 0._dp
       dx           = 0._dp
       vquad        = 0._dp
       diff         = 0._dp
       diffc        = 0._dp
       diffb        = 0._dp
       fsave        = 0._dp
       detrat       = 0._dp
       hdiag        = 0._dp
       distsq       = 0._dp
       gisq         = 0._dp
       gqsq         = 0._dp
       f            = 0._dp
       bstep        = 0._dp
       alpha        = 0._dp
       dstep        = 0._dp
       !
    END IF

    ipt          = 0
    jpt          = 0
    xipt         = 0._dp
    xjpt         = 0._dp

    half=0.5_dp
    one=1.0_dp
    tenth=0.1_dp
    zero=0.0_dp
    np=n+1
    nh=(n*np)/2
    nptm=npt-np
    nftest=MAX(maxfun,1)

    IF ( opt%state == 2 ) GOTO 1000
    !
    !     Set the initial elements of XPT, BMAT, HQ, PQ and ZMAT to zero.
    !
    DO j=1,n
       xbase(j)=x(j)
       DO k=1,npt
          xpt(k,j)=zero
       END DO
       DO i=1,ndim
          bmat(i,j)=zero
       END DO
    END DO
    DO ih=1,nh
       hq(ih)=zero
    END DO
    DO k=1,npt
       pq(k)=zero
       DO j=1,nptm
          zmat(k,j)=zero
       END DO
    END DO
    !
    !     Begin the initialization procedure. NF becomes one more than the n
    !     of function values so far. The coordinates of the displacement of
    !     next initial interpolation point from XBASE are set in XPT(NF,.).
    !
    rhosq=rhobeg*rhobeg
    recip=one/rhosq
    reciq=SQRT(half)/rhosq
    nf=0
50  nfm=nf
    nfmm=nf-n
    nf=nf+1
    IF (nfm <= 2*n) THEN
       IF (nfm >= 1 .AND. nfm <= N) THEN
          xpt(nf,nfm)=rhobeg
       ELSE IF (nfm > n) THEN
          xpt(nf,nfmm)=-rhobeg
       END IF
    ELSE
       itemp=(nfmm-1)/n
       jpt=nfm-itemp*n-n
       ipt=jpt+itemp
       IF (ipt > n) THEN
          itemp=jpt
          jpt=ipt-n
          ipt=itemp
       END IF
       xipt=rhobeg
       IF (fval(ipt+np) < fval(ipt+1)) xipt=-xipt
       XJPT=RHOBEG
       IF (fval(jpt+np) < fval(jpt+1)) xjpt=-xjpt
       xpt(nf,ipt)=xipt
       xpt(nf,jpt)=xjpt
    END IF
    !
    !     Calculate the next value of F, label 70 being reached immediately
    !     after this calculation. The least function value so far and its in
    !     are required.
    !
    DO j=1,n
       x(j)=xpt(nf,j)+xbase(j)
    END DO
    GOTO 310
70  fval(nf)=f
    IF (nf == 1) THEN
       fbeg=f
       fopt=f
       kopt=1
    ELSE IF (f < fopt) THEN
       fopt=f
       kopt=nf
    END IF
    !
    !     Set the nonzero initial elements of BMAT and the quadratic model i
    !     the cases when NF is at most 2*N+1.
    !
    IF (NFM <= 2*N) THEN
       IF (nfm >= 1 .AND. nfm <= n) THEN
          gq(nfm)=(f-fbeg)/rhobeg
          IF (npt < nf+n) THEN
             bmat(1,nfm)=-one/rhobeg
             bmat(nf,nfm)=one/rhobeg
             bmat(npt+nfm,nfm)=-half*rhosq
          END IF
       ELSE IF (nfm > n) THEN
          bmat(nf-n,nfmm)=half/rhobeg
          bmat(nf,nfmm)=-half/rhobeg
          zmat(1,nfmm)=-reciq-reciq
          zmat(nf-n,nfmm)=reciq
          zmat(nf,nfmm)=reciq
          ih=(nfmm*(nfmm+1))/2
          temp=(fbeg-f)/rhobeg
          hq(ih)=(gq(nfmm)-temp)/rhobeg
          gq(nfmm)=half*(gq(nfmm)+temp)
       END IF
       !
       !     Set the off-diagonal second derivatives of the Lagrange functions
       !     the initial quadratic model.
       !
    ELSE
       ih=(ipt*(ipt-1))/2+jpt
       IF (xipt < zero) ipt=ipt+n
       IF (xjpt < zero) jpt=jpt+n
       zmat(1,nfmm)=recip
       zmat(nf,nfmm)=recip
       zmat(ipt+1,nfmm)=-recip
       zmat(jpt+1,nfmm)=-recip
       hq(ih)=(fbeg-fval(ipt+1)-fval(jpt+1)+f)/(xipt*xjpt)
    END IF
    IF (nf < npt) GOTO 50
    !
    !     Begin the iterative procedure, because the initial model is comple
    !
    rho=rhobeg
    delta=rho
    idz=1
    diffa=zero
    diffb=zero
    itest=0
    xoptsq=zero
    DO i=1,n
       xopt(i)=xpt(kopt,i)
       xoptsq=xoptsq+xopt(i)**2
    END DO
90  nfsav=nf
    !
    !     Generate the next trust region step and test its length. Set KNEW
    !     to -1 if the purpose of the next F will be to improve the model.
    !
100 knew=0
    CALL trsapp (n,npt,xopt,xpt,gq,hq,pq,delta,d,w,w(np),w(np+n),w(np+2*n),crvmin)
    dsq=zero
    DO i=1,n
       dsq=dsq+d(i)**2
    END DO
    dnorm=MIN(delta,SQRT(dsq))
    IF (dnorm < half*rho) THEN
       knew=-1
       delta=tenth*delta
       ratio=-1.0_dp
       IF (delta <= 1.5_dp*rho) delta=rho
       IF (nf <= nfsav+2) GOTO 460
       temp=0.125_dp*crvmin*rho*rho
       IF (temp <= MAX(diffa,diffb,diffc)) GOTO 460
       GOTO 490
    END IF
    !
    !     Shift XBASE if XOPT may be too far from XBASE. First make the chan
    !     to BMAT that do not depend on ZMAT.
    !
120 IF (dsq <= 1.0e-3_dp*xoptsq) THEN
       tempq=0.25_dp*xoptsq
       DO k=1,npt
          sum=zero
          DO i=1,n
             sum=sum+xpt(k,i)*xopt(i)
          END DO
          temp=pq(k)*sum
          sum=sum-half*xoptsq
          w(npt+k)=sum
          DO i=1,n
             gq(i)=gq(i)+temp*xpt(k,i)
             xpt(k,i)=xpt(k,i)-half*xopt(i)
             vlag(i)=bmat(k,i)
             w(i)=sum*xpt(k,i)+tempq*xopt(i)
             ip=npt+i
             DO j=1,i
                bmat(ip,j)=bmat(ip,j)+vlag(i)*w(j)+w(i)*vlag(j)
             END DO
          END DO
       END DO
       !
       !     Then the revisions of BMAT that depend on ZMAT are calculated.
       !
       DO k=1,nptm
          sumz=zero
          DO i=1,npt
             sumz=sumz+zmat(i,k)
             w(i)=w(npt+i)*zmat(i,k)
          END DO
          DO j=1,n
             sum=tempq*sumz*xopt(j)
             DO i=1,npt
                sum=sum+w(i)*xpt(i,j)
                vlag(j)=sum
                IF (k < idz) sum=-sum
             END DO
             DO i=1,npt
                bmat(i,j)=bmat(i,j)+sum*zmat(i,k)
             END DO
          END DO
          DO i=1,n
             ip=i+npt
             temp=vlag(i)
             IF (k < idz) temp=-temp
             DO j=1,i
                bmat(ip,j)=bmat(ip,j)+temp*vlag(j)
             END DO
          END DO
       END DO
       !
       !     The following instructions complete the shift of XBASE, including
       !     the changes to the parameters of the quadratic model.
       !
       ih=0
       DO j=1,n
          w(j)=zero
          DO k=1,npt
             w(j)=w(j)+pq(k)*xpt(k,j)
             xpt(k,j)=xpt(k,j)-half*xopt(j)
          END DO
          DO i=1,j
             ih=ih+1
             IF (i < j) gq(j)=gq(j)+hq(ih)*xopt(i)
             gq(i)=gq(i)+hq(ih)*xopt(j)
             hq(ih)=hq(ih)+w(i)*xopt(j)+xopt(i)*w(j)
             bmat(npt+i,j)=bmat(npt+j,i)
          END DO
       END DO
       DO j=1,n
          xbase(j)=xbase(j)+xopt(j)
          xopt(j)=zero
       END DO
       xoptsq=zero
    END IF
    !
    !     Pick the model step if KNEW is positive. A different choice of D
    !     may be made later, if the choice of D by BIGLAG causes substantial
    !     cancellation in DENOM.
    !
    IF (knew > 0) THEN
       CALL biglag (n,npt,xopt,xpt,bmat,zmat,idz,ndim,knew,dstep,    &
                   d,alpha,vlag,vlag(npt+1),w,w(np),w(np+n))
    END IF
    !
    !     Calculate VLAG and BETA for the current choice of D. The first NPT
    !     components of W_check will be held in W.
    !
    DO k=1,npt
       suma=zero
       sumb=zero
       sum=zero
       DO j=1,n
          suma=suma+xpt(k,j)*d(j)
          sumb=sumb+xpt(k,j)*xopt(j)
          sum=sum+bmat(k,j)*d(j)
       END DO
       w(k)=suma*(half*suma+sumb)
       vlag(k)=sum
    END DO
    beta=zero
    DO k=1,nptm
       sum=zero
       DO i=1,npt
          sum=sum+zmat(i,k)*w(i)
       END DO
       IF (k < idz) THEN
          beta=beta+sum*sum
          sum=-sum
       ELSE
          beta=beta-sum*sum
       END IF
       DO i=1,npt
          vlag(i)=vlag(i)+sum*zmat(i,k)
       END DO
    END DO
    bsum=zero
    dx=zero
    DO j=1,n
       sum=zero
       DO i=1,npt
          sum=sum+w(i)*bmat(i,j)
       END DO
       bsum=bsum+sum*d(j)
       jp=npt+j
       DO k=1,n
          sum=sum+bmat(jp,k)*d(k)
       END DO
       vlag(jp)=sum
       bsum=bsum+sum*d(j)
       dx=dx+d(j)*xopt(j)
    END DO
    beta=dx*dx+dsq*(xoptsq+dx+dx+half*dsq)+beta-bsum
    vlag(kopt)=vlag(kopt)+one
    !
    !     If KNEW is positive and if the cancellation in DENOM is unacceptab
    !     then BIGDEN calculates an alternative model step, XNEW being used
    !     working space.
    !
    IF (knew > 0) THEN
       temp=one+alpha*beta/vlag(knew)**2
       IF (ABS(temp) <= 0.8_dp) THEN
          CALL bigden (n,npt,xopt,xpt,bmat,zmat,idz,ndim,kopt,      &
                        knew,d,w,vlag,beta,xnew,w(ndim+1),w(6*ndim+1))
       END IF
    END IF
    !
    !     Calculate the next value of the objective function.
    !
290 DO i=1,n
       xnew(i)=xopt(i)+d(i)
       x(i)=xbase(i)+xnew(i)
    END DO
    nf=nf+1
310 IF (nf > nftest) THEN
       !         return to many steps
       nf=nf-1
       opt%state = 3
       CALL get_state
       GOTO 530
    END IF

    CALL get_state

    opt%state = 2

    RETURN

1000 CONTINUE

    CALL set_state

    IF (nf <= npt) GOTO 70
    IF (knew == -1) THEN
       opt%state = 6
       CALL get_state
       GOTO 530
    END IF
    !
    !     Use the quadratic model to predict the change in F due to the step
    !     and set DIFF to the error of this prediction.
    !
    vquad=zero
    ih=0
    DO j=1,n
       vquad=vquad+d(j)*gq(j)
       DO i=1,j
          ih=ih+1
          temp=d(i)*xnew(j)+d(j)*xopt(i)
          IF (i == j) temp=half*temp
          vquad=vquad+temp*hq(ih)
       END DO
    END DO
    DO k=1,npt
       vquad=vquad+pq(k)*w(k)
    END DO
    diff=f-fopt-vquad
    diffc=diffb
    diffb=diffa
    diffa=ABS(diff)
    IF (dnorm > rho) nfsav=nf
    !
    !     Update FOPT and XOPT if the new F is the least value of the object
    !     function so far. The branch when KNEW is positive occurs if D is n
    !     a trust region step.
    !
    fsave=fopt
    IF (f < fopt) THEN
       fopt=f
       xoptsq=zero
       DO i=1,n
          xopt(i)=xnew(i)
          xoptsq=xoptsq+xopt(i)**2
       END DO
    END IF
    ksave=knew
    IF (knew > 0) GOTO 410
    !
    !     Pick the next value of DELTA after a trust region step.
    !
    IF (vquad >= zero) THEN
       ! Return because a trust region step has failed to reduce Q
       opt%state = 4
       CALL get_state
       GOTO 530
    END IF
    ratio=(f-fsave)/vquad
    IF (ratio <= tenth) THEN
       delta=half*dnorm
    ELSE IF (ratio <= 0.7_dp) THEN
       delta=MAX(half*delta,dnorm)
    ELSE
       delta=MAX(half*delta,dnorm+dnorm)
    END IF
    IF (delta <= 1.5_dp*rho) delta=rho
    !
    !     Set KNEW to the index of the next interpolation point to be delete
    !
    rhosq=MAX(tenth*delta,rho)**2
    ktemp=0
    detrat=zero
    IF (f >= fsave) THEN
       ktemp=kopt
       detrat=one
    END IF
    DO k=1,npt
       hdiag=zero
       DO j=1,nptm
          temp=one
          IF (j < idz) temp=-one
          hdiag=hdiag+temp*zmat(k,j)**2
       END DO
       temp=ABS(beta*hdiag+vlag(k)**2)
       distsq=zero
       DO j=1,n
          distsq=distsq+(xpt(k,j)-xopt(j))**2
       END DO
       IF (distsq > rhosq) temp=temp*(distsq/rhosq)**3
       IF (temp > detrat .AND. k /= ktemp) THEN
          detrat=temp
          knew=k
       END IF
    END DO
    IF (knew == 0) GOTO 460
    !
    !     Update BMAT, ZMAT and IDZ, so that the KNEW-th interpolation point
    !     can be moved. Begin the updating of the quadratic model, starting
    !     with the explicit second derivative term.
    !
410 CALL update (n,npt,bmat,zmat,idz,ndim,vlag,beta,knew,w)
    fval(knew)=f
    ih=0
    DO i=1,n
       temp=pq(knew)*xpt(knew,i)
       DO j=1,i
          ih=ih+1
          hq(ih)=hq(ih)+temp*xpt(knew,j)
       END DO
    END DO
    pq(knew)=zero
    !
    !     Update the other second derivative parameters, and then the gradie
    !     vector of the model. Also include the new interpolation point.
    !
    DO j=1,nptm
       temp=diff*zmat(knew,j)
       IF (j < idz) temp=-temp
       DO k=1,npt
          pq(k)=pq(k)+temp*zmat(k,j)
       END DO
    END DO
    gqsq=zero
    DO i=1,n
       gq(i)=gq(i)+diff*bmat(knew,i)
       gqsq=gqsq+gq(i)**2
       xpt(knew,i)=xnew(i)
    END DO
    !
    !     If a trust region step makes a small change to the objective funct
    !     then calculate the gradient of the least Frobenius norm interpolan
    !     XBASE, and store it in W, using VLAG for a vector of right hand si
    !
    IF (ksave == 0 .AND. delta == rho) THEN
       IF (ABS(ratio) > 1.0e-2_dp) THEN
          itest=0
       ELSE
          DO k=1,npt
             vlag(k)=fval(k)-fval(kopt)
          END DO
          gisq=zero
          DO i=1,n
             sum=zero
             DO k=1,npt
                sum=sum+bmat(k,i)*vlag(k)
             END DO
             gisq=gisq+sum*sum
             w(i)=sum
          END DO
          !
          !     Test whether to replace the new quadratic model by the least Frobe
          !     norm interpolant, making the replacement if the test is satisfied.
          !
          itest=itest+1
          IF (gqsq < 1.0e2_dp*gisq) itest=0
          IF (itest >= 3) THEN
             DO i=1,n
                gq(i)=w(i)
             END DO
             DO ih=1,nh
                hq(ih)=zero
             END DO
             DO j=1,nptm
                w(j)=zero
                DO k=1,npt
                   w(j)=w(j)+vlag(k)*zmat(k,j)
                END DO
                IF (j < idz) w(j)=-w(j)
             END DO
             DO k=1,npt
                pq(k)=zero
                DO j=1,nptm
                   pq(k)=pq(k)+zmat(k,j)*w(j)
                END DO
             END DO
             itest=0
          END IF
       END IF
    END IF
    IF (f < fsave) kopt=knew
    !
    !     If a trust region step has provided a sufficient decrease in F, th
    !     branch for another trust region calculation. The case KSAVE>0 occu
    !     when the new function value was calculated by a model step.
    !
    IF (f <= fsave+tenth*vquad) GOTO 100
    IF (ksave > 0) GOTO 100
    !
    !     Alternatively, find out if the interpolation points are close enou
    !     to the best point so far.
    !
    knew=0
460 distsq=4.0_dp*delta*delta
    DO k=1,npt
       sum=zero
       DO j=1,n
          sum=sum+(xpt(k,j)-xopt(j))**2
       END DO
       IF (sum > distsq) THEN
          knew=k
          distsq=sum
       END IF
    END DO
    !
    !     If KNEW is positive, then set DSTEP, and branch back for the next
    !     iteration, which will generate a "model step".
    !
    IF (knew > 0) THEN
       dstep=MAX(MIN(tenth*SQRT(distsq),half*delta),rho)
       dsq=dstep*dstep
       GOTO 120
    END IF
    IF (ratio > zero) GOTO 100
    IF (MAX(delta,dnorm) > rho) GOTO 100
    !
    !     The calculations with the current value of RHO are complete. Pick
    !     next values of RHO and DELTA.
    !
490 IF (rho > rhoend) THEN
       delta=half*rho
       ratio=rho/rhoend
       IF (ratio <= 16.0_dp) THEN
          rho=rhoend
       ELSE IF (ratio <= 250.0_dp) THEN
          rho=SQRT(ratio)*rhoend
       ELSE
          rho=tenth*rho
       END IF
       delta=MAX(delta,rho)
       GOTO 90
    END IF
    !
    !     Return from the calculation, after another Newton-Raphson step, if
    !     it is too short to have been tried before.
    !
    IF (knew == -1) GOTO 290
    opt%state = 7
    CALL get_state
530 IF (fopt <= f) THEN
       DO i=1,n
          x(i)=xbase(i)+xopt(i)
       END DO
       f=fopt
    END IF

    CALL get_state

    !******************************************************************************
  CONTAINS
    !******************************************************************************
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE get_state()
      opt%np       = np
      opt%nh       = nh
      opt%nptm     = nptm
      opt%nftest   = nftest
      opt%idz      = idz
      opt%itest    = itest
      opt%nf       = nf
      opt%nfm      = nfm
      opt%nfmm     = nfmm
      opt%nfsav    = nfsav
      opt%knew     = knew
      opt%kopt     = kopt
      opt%ksave    = ksave
      opt%ktemp    = ktemp
      opt%rhosq    = rhosq
      opt%recip    = recip
      opt%reciq    = reciq
      opt%fbeg     = fbeg
      opt%fopt     = fopt
      opt%diffa    = diffa
      opt%xoptsq   = xoptsq
      opt%rho      = rho
      opt%delta    = delta
      opt%dsq      = dsq
      opt%dnorm    = dnorm
      opt%ratio    = ratio
      opt%temp     = temp
      opt%tempq    = tempq
      opt%beta     = beta
      opt%dx       = dx
      opt%vquad    = vquad
      opt%diff     = diff
      opt%diffc    = diffc
      opt%diffb    = diffb
      opt%fsave    = fsave
      opt%detrat   = detrat
      opt%hdiag    = hdiag
      opt%distsq   = distsq
      opt%gisq     = gisq
      opt%gqsq     = gqsq
      opt%f        = f
      opt%bstep    = bstep
      opt%alpha    = alpha
      opt%dstep    = dstep
    END SUBROUTINE get_state

    !******************************************************************************
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE set_state()
      np           = opt%np
      nh           = opt%nh
      nptm         = opt%nptm
      nftest       = opt%nftest
      idz          = opt%idz
      itest        = opt%itest
      nf           = opt%nf
      nfm          = opt%nfm
      nfmm         = opt%nfmm
      nfsav        = opt%nfsav
      knew         = opt%knew
      kopt         = opt%kopt
      ksave        = opt%ksave
      ktemp        = opt%ktemp
      rhosq        = opt%rhosq
      recip        = opt%recip
      reciq        = opt%reciq
      fbeg         = opt%fbeg
      fopt         = opt%fopt
      diffa        = opt%diffa
      xoptsq       = opt%xoptsq
      rho          = opt%rho
      delta        = opt%delta
      dsq          = opt%dsq
      dnorm        = opt%dnorm
      ratio        = opt%ratio
      temp         = opt%temp
      tempq        = opt%tempq
      beta         = opt%beta
      dx           = opt%dx
      vquad        = opt%vquad
      diff         = opt%diff
      diffc        = opt%diffc
      diffb        = opt%diffb
      fsave        = opt%fsave
      detrat       = opt%detrat
      hdiag        = opt%hdiag
      distsq       = opt%distsq
      gisq         = opt%gisq
      gqsq         = opt%gqsq
      f            = opt%f
      bstep        = opt%bstep
      alpha        = opt%alpha
      dstep        = opt%dstep
    END SUBROUTINE set_state

  END SUBROUTINE newuob

!******************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param npt ...
!> \param xopt ...
!> \param xpt ...
!> \param bmat ...
!> \param zmat ...
!> \param idz ...
!> \param ndim ...
!> \param kopt ...
!> \param knew ...
!> \param d ...
!> \param w ...
!> \param vlag ...
!> \param beta ...
!> \param s ...
!> \param wvec ...
!> \param prod ...
! **************************************************************************************************
  SUBROUTINE bigden (n,npt,xopt,xpt,bmat,zmat,idz,ndim,kopt,&
       knew,d,w,vlag,beta,s,wvec,prod)

      INTEGER, INTENT(inout)                             :: n, npt
      REAL(dp), DIMENSION(*), INTENT(inout)              :: xopt
      REAL(dp), DIMENSION(npt, *), INTENT(inout)         :: xpt
      INTEGER, INTENT(inout)                             :: ndim, idz
      REAL(dp), DIMENSION(npt, *), INTENT(inout)         :: zmat
      REAL(dp), DIMENSION(ndim, *), INTENT(inout)        :: bmat
      INTEGER, INTENT(inout)                             :: kopt, knew
      REAL(dp), DIMENSION(*), INTENT(inout)              :: d, w, vlag
      REAL(dp), INTENT(inout)                            :: beta
      REAL(dp), DIMENSION(*), INTENT(inout)              :: s
      REAL(dp), DIMENSION(ndim, *), INTENT(inout)        :: wvec, prod

      REAL(dp), PARAMETER                                :: half = 0.5_dp, one = 1._dp, &
                                                            quart = 0.25_dp, two = 2._dp, &
                                                            zero = 0._dp

      INTEGER                                            :: i, ip, isave, iterc, iu, j, jc, k, ksav, &
                                                            nptm, nw
      REAL(dp) :: alpha, angle, dd, denmax, denold, densav, diff, ds, dstemp, dtest, ss, ssden, &
         sstemp, step, sum, sumold, tau, temp, tempa, tempb, tempc, xoptd, xopts, xoptsq
      REAL(dp), DIMENSION(9)                             :: den, denex, par

!
!     N is the number of variables.
!     NPT is the number of interpolation equations.
!     XOPT is the best interpolation point so far.
!     XPT contains the coordinates of the current interpolation points.
!     BMAT provides the last N columns of H.
!     ZMAT and IDZ give a factorization of the first NPT by NPT submatri
!     NDIM is the first dimension of BMAT and has the value NPT+N.
!     KOPT is the index of the optimal interpolation point.
!     KNEW is the index of the interpolation point that is going to be m
!     D will be set to the step from XOPT to the new point, and on entry
!       should be the D that was calculated by the last call of BIGLAG.
!       length of the initial D provides a trust region bound on the fin
!     W will be set to Wcheck for the final choice of D.
!     VLAG will be set to Theta*Wcheck+e_b for the final choice of D.
!     BETA will be set to the value that will occur in the updating form
!       when the KNEW-th interpolation point is moved to its new positio
!     S, WVEC, PROD and the private arrays DEN, DENEX and PAR will be us
!       for working space.
!
!     D is calculated in a way that should provide a denominator with a
!     modulus in the updating formula when the KNEW-th interpolation poi
!     shifted to the new position XOPT+D.
!

    nptm=npt-n-1
    !
    !     Store the first NPT elements of the KNEW-th column of H in W(N+1)
    !     to W(N+NPT).
    !
    DO k=1,npt
       w(n+k)=zero
    END DO
    DO j=1,nptm
       temp=zmat(knew,j)
       IF (j < idz) temp=-temp
       DO k=1,npt
          w(n+k)=w(n+k)+temp*zmat(k,j)
       END DO
    END DO
    alpha=w(n+knew)
    !
    !     The initial search direction D is taken from the last call of BIGL
    !     and the initial S is set below, usually to the direction from X_OP
    !     to X_KNEW, but a different direction to an interpolation point may
    !     be chosen, in order to prevent S from being nearly parallel to D.
    !
    dd=zero
    ds=zero
    ss=zero
    xoptsq=zero
    DO i=1,n
       dd=dd+d(i)**2
       s(i)=xpt(knew,i)-xopt(i)
       ds=ds+d(i)*s(i)
       ss=ss+s(i)**2
       xoptsq=xoptsq+xopt(i)**2
    END DO
    IF (ds*ds > 0.99_dp*dd*ss) THEN
       ksav=knew
       dtest=ds*ds/ss
       DO k=1,npt
          IF (k /= kopt) THEN
             dstemp=zero
             sstemp=zero
             DO i=1,n
                diff=xpt(k,i)-xopt(i)
                dstemp=dstemp+d(i)*diff
                sstemp=sstemp+diff*diff
             END DO
             IF (dstemp*dstemp/sstemp < dtest) THEN
                ksav=k
                dtest=dstemp*dstemp/sstemp
                ds=dstemp
                ss=sstemp
             END IF
          END IF
       END DO
       DO i=1,n
          s(i)=xpt(ksav,i)-xopt(i)
       END DO
    END IF
    ssden=dd*ss-ds*ds
    iterc=0
    densav=zero
    !
    !     Begin the iteration by overwriting S with a vector that has the
    !     required length and direction.
    !
    mainloop : DO
       iterc=iterc+1
       temp=one/SQRT(ssden)
       xoptd=zero
       xopts=zero
       DO i=1,n
          s(i)=temp*(dd*s(i)-ds*d(i))
          xoptd=xoptd+xopt(i)*d(i)
          xopts=xopts+xopt(i)*s(i)
       END DO
       !
       !     Set the coefficients of the first two terms of BETA.
       !
       tempa=half*xoptd*xoptd
       tempb=half*xopts*xopts
       den(1)=dd*(xoptsq+half*dd)+tempa+tempb
       den(2)=two*xoptd*dd
       den(3)=two*xopts*dd
       den(4)=tempa-tempb
       den(5)=xoptd*xopts
       DO i=6,9
          den(i)=zero
       END DO
       !
       !     Put the coefficients of Wcheck in WVEC.
       !
       DO k=1,npt
          tempa=zero
          tempb=zero
          tempc=zero
          DO i=1,n
             tempa=tempa+xpt(k,i)*d(i)
             tempb=tempb+xpt(k,i)*s(i)
             tempc=tempc+xpt(k,i)*xopt(i)
          END DO
          wvec(k,1)=quart*(tempa*tempa+tempb*tempb)
          wvec(k,2)=tempa*tempc
          wvec(k,3)=tempb*tempc
          wvec(k,4)=quart*(tempa*tempa-tempb*tempb)
          wvec(k,5)=half*tempa*tempb
       END DO
       DO i=1,n
          ip=i+npt
          wvec(ip,1)=zero
          wvec(ip,2)=d(i)
          wvec(ip,3)=s(i)
          wvec(ip,4)=zero
          wvec(ip,5)=zero
       END DO
       !
       !     Put the coefficents of THETA*Wcheck in PROD.
       !
       DO jc=1,5
          nw=npt
          IF (jc == 2 .OR. jc == 3) nw=ndim
          DO k=1,npt
             prod(k,jc)=zero
          END DO
          DO j=1,nptm
             sum=zero
             DO k=1,npt
                sum=sum+zmat(k,j)*wvec(k,jc)
             END DO
             IF (j < idz) sum=-sum
             DO k=1,npt
                prod(k,jc)=prod(k,jc)+sum*zmat(k,j)
             END DO
          END DO
          IF (nw == ndim) THEN
             DO k=1,npt
                sum=zero
                DO j=1,n
                   sum=sum+bmat(k,j)*wvec(npt+j,jc)
                END DO
                prod(k,jc)=prod(k,jc)+sum
             END DO
          END IF
          DO j=1,n
             sum=zero
             DO i=1,nw
                sum=sum+bmat(i,j)*wvec(i,jc)
             END DO
             prod(npt+j,jc)=sum
          END DO
       END DO
       !
       !     Include in DEN the part of BETA that depends on THETA.
       !
       DO k=1,ndim
          sum=zero
          DO I=1,5
             par(i)=half*prod(k,i)*wvec(k,i)
             sum=sum+par(i)
          END DO
          den(1)=den(1)-par(1)-sum
          tempa=prod(k,1)*wvec(k,2)+prod(k,2)*wvec(k,1)
          tempb=prod(k,2)*wvec(k,4)+prod(k,4)*wvec(k,2)
          tempc=prod(k,3)*wvec(k,5)+prod(k,5)*wvec(k,3)
          den(2)=den(2)-tempa-half*(tempb+tempc)
          den(6)=den(6)-half*(tempb-tempc)
          tempa=prod(k,1)*wvec(k,3)+prod(k,3)*wvec(k,1)
          tempb=prod(k,2)*wvec(k,5)+prod(k,5)*wvec(k,2)
          tempc=prod(k,3)*wvec(k,4)+prod(k,4)*wvec(k,3)
          den(3)=den(3)-tempa-half*(tempb-tempc)
          den(7)=den(7)-half*(tempb+tempc)
          tempa=prod(k,1)*wvec(k,4)+prod(k,4)*wvec(k,1)
          den(4)=den(4)-tempa-par(2)+par(3)
          tempa=prod(k,1)*wvec(k,5)+prod(k,5)*wvec(k,1)
          tempb=prod(k,2)*wvec(k,3)+prod(k,3)*wvec(k,2)
          den(5)=den(5)-tempa-half*tempb
          den(8)=den(8)-par(4)+par(5)
          tempa=prod(k,4)*wvec(k,5)+prod(k,5)*wvec(k,4)
          den(9)=den(9)-half*tempa
       END DO
       !
       !     Extend DEN so that it holds all the coefficients of DENOM.
       !
       sum=zero
       DO i=1,5
          par(i)=half*prod(knew,i)**2
          sum=sum+par(i)
       END DO
       denex(1)=alpha*den(1)+par(1)+sum
       tempa=two*prod(knew,1)*prod(knew,2)
       tempb=prod(knew,2)*prod(knew,4)
       tempc=prod(knew,3)*prod(knew,5)
       denex(2)=alpha*den(2)+tempa+tempb+tempc
       denex(6)=alpha*den(6)+tempb-tempc
       tempa=two*prod(knew,1)*prod(knew,3)
       tempb=prod(knew,2)*prod(knew,5)
       tempc=prod(knew,3)*prod(knew,4)
       denex(3)=alpha*den(3)+tempa+tempb-tempc
       denex(7)=alpha*den(7)+tempb+tempc
       tempa=two*prod(knew,1)*prod(knew,4)
       denex(4)=alpha*den(4)+tempa+par(2)-par(3)
       tempa=two*prod(knew,1)*prod(knew,5)
       denex(5)=alpha*den(5)+tempa+prod(knew,2)*prod(knew,3)
       denex(8)=alpha*den(8)+par(4)-par(5)
       denex(9)=alpha*den(9)+prod(knew,4)*prod(knew,5)
       !
       !     Seek the value of the angle that maximizes the modulus of DENOM.
       !
       sum=denex(1)+denex(2)+denex(4)+denex(6)+denex(8)
       denold=sum
       denmax=sum
       isave=0
       iu=49
       temp=twopi/REAL(IU+1,dp)
       par(1)=one
       DO i=1,iu
          angle=REAL(i,dp)*temp
          par(2)=COS(angle)
          par(3)=SIN(angle)
          DO j=4,8,2
             par(j)=par(2)*par(j-2)-par(3)*par(j-1)
             par(j+1)=par(2)*par(j-1)+par(3)*par(j-2)
          END DO
          sumold=sum
          sum=zero
          DO j=1,9
             sum=sum+denex(j)*par(j)
          END DO
          IF (ABS(sum) > ABS(denmax)) THEN
             denmax=sum
             isave=i
             tempa=sumold
          ELSE IF (i == isave+1) THEN
             tempb=sum
          END IF
       END DO
       IF (isave == 0) tempa=sum
       IF (isave == iu) tempb=denold
       step=zero
       IF (tempa /= tempb) THEN
          tempa=tempa-denmax
          tempb=tempb-denmax
          step=half*(tempa-tempb)/(tempa+tempb)
       END IF
       angle=temp*(REAL(isave,dp)+step)
       !
       !     Calculate the new parameters of the denominator, the new VLAG vect
       !     and the new D. Then test for convergence.
       !
       par(2)=COS(angle)
       par(3)=SIN(angle)
       DO j=4,8,2
          par(j)=par(2)*par(j-2)-par(3)*par(j-1)
          par(j+1)=par(2)*par(j-1)+par(3)*par(j-2)
       END DO
       beta=zero
       denmax=zero
       DO j=1,9
          beta=beta+den(j)*par(j)
          denmax=denmax+denex(j)*par(j)
       END DO
       DO k=1,ndim
          vlag(k)=zero
          DO j=1,5
             vlag(k)=vlag(k)+prod(k,j)*par(j)
          END DO
       END DO
       tau=vlag(knew)
       dd=zero
       tempa=zero
       tempb=zero
       DO i=1,n
          d(i)=par(2)*d(i)+par(3)*s(i)
          w(i)=xopt(i)+d(i)
          dd=dd+d(i)**2
          tempa=tempa+d(i)*w(i)
          tempb=tempb+w(i)*w(i)
       END DO
       IF (iterc >= n) EXIT mainloop
       IF (iterc >= 1) densav=MAX(densav,denold)
       IF (ABS(denmax) <= 1.1_dp*ABS(densav)) EXIT mainloop
       densav=denmax
       !
       !     Set S to half the gradient of the denominator with respect to D.
       !     Then branch for the next iteration.
       !
       DO i=1,n
          temp=tempa*xopt(i)+tempb*d(i)-vlag(npt+i)
          s(i)=tau*bmat(knew,i)+alpha*temp
       END DO
       DO k=1,npt
          sum=zero
          DO j=1,n
             sum=sum+xpt(k,j)*w(j)
          END DO
          temp=(tau*w(n+k)-alpha*vlag(k))*sum
          DO i=1,n
             s(i)=s(i)+temp*xpt(k,i)
          END DO
       END DO
       ss=zero
       ds=zero
       DO i=1,n
          ss=ss+s(i)**2
          ds=ds+d(i)*s(i)
       END DO
       ssden=dd*ss-ds*ds
       IF (ssden < 1.0e-8_dp*dd*ss) EXIT mainloop
    END DO mainloop
    !
    !     Set the vector W before the RETURN from the subroutine.
    !
    DO k=1,ndim
       w(k)=zero
       DO j=1,5
          w(k)=w(k)+wvec(k,j)*par(j)
       END DO
    END DO
    vlag(kopt)=vlag(kopt)+one

  END SUBROUTINE bigden
!******************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param npt ...
!> \param xopt ...
!> \param xpt ...
!> \param bmat ...
!> \param zmat ...
!> \param idz ...
!> \param ndim ...
!> \param knew ...
!> \param delta ...
!> \param d ...
!> \param alpha ...
!> \param hcol ...
!> \param gc ...
!> \param gd ...
!> \param s ...
!> \param w ...
! **************************************************************************************************
  SUBROUTINE biglag (n,npt,xopt,xpt,bmat,zmat,idz,ndim,knew,&
       delta,d,alpha,hcol,gc,gd,s,w)
      INTEGER, INTENT(inout)                             :: n, npt
      REAL(dp), DIMENSION(*), INTENT(inout)              :: xopt
      REAL(dp), DIMENSION(npt, *), INTENT(inout)         :: xpt
      INTEGER, INTENT(inout)                             :: ndim, idz
      REAL(dp), DIMENSION(npt, *), INTENT(inout)         :: zmat
      REAL(dp), DIMENSION(ndim, *), INTENT(inout)        :: bmat
      INTEGER, INTENT(inout)                             :: knew
      REAL(dp), INTENT(inout)                            :: delta
      REAL(dp), DIMENSION(*), INTENT(inout)              :: d
      REAL(dp), INTENT(inout)                            :: alpha
      REAL(dp), DIMENSION(*), INTENT(inout)              :: hcol, gc, gd, s, w

      REAL(dp), PARAMETER                                :: half = 0.5_dp, one = 1._dp, zero = 0._dp

      INTEGER                                            :: i, isave, iterc, iu, j, k, nptm
      REAL(dp)                                           :: angle, cf1, cf2, cf3, cf4, cf5, cth, dd, &
                                                            delsq, denom, dhd, gg, scale, sp, ss, &
                                                            step, sth, sum, tau, taubeg, taumax, &
                                                            tauold, temp, tempa, tempb

!
!
!     N is the number of variables.
!     NPT is the number of interpolation equations.
!     XOPT is the best interpolation point so far.
!     XPT contains the coordinates of the current interpolation points.
!     BMAT provides the last N columns of H.
!     ZMAT and IDZ give a factorization of the first NPT by NPT submatrix
!     NDIM is the first dimension of BMAT and has the value NPT+N.
!     KNEW is the index of the interpolation point that is going to be m
!     DELTA is the current trust region bound.
!     D will be set to the step from XOPT to the new point.
!     ALPHA will be set to the KNEW-th diagonal element of the H matrix.
!     HCOL, GC, GD, S and W will be used for working space.
!
!     The step D is calculated in a way that attempts to maximize the mo
!     of LFUNC(XOPT+D), subject to the bound ||D|| .LE. DELTA, where LFU
!     the KNEW-th Lagrange function.
!

    delsq=delta*delta
    nptm=npt-n-1
    !
    !     Set the first NPT components of HCOL to the leading elements of th
    !     KNEW-th column of H.
    !
    iterc=0
    DO k=1,npt
       hcol(k)=zero
    END DO
    DO j=1,nptm
       temp=zmat(knew,j)
       IF (j < idz) temp=-temp
       DO k=1,npt
          hcol(k)=hcol(k)+temp*zmat(k,j)
       END DO
    END DO
    alpha=hcol(knew)
    !
    !     Set the unscaled initial direction D. Form the gradient of LFUNC a
    !     XOPT, and multiply D by the second derivative matrix of LFUNC.
    !
    dd=zero
    DO i=1,n
       d(i)=xpt(knew,i)-xopt(i)
       gc(i)=bmat(knew,i)
       gd(i)=zero
       dd=dd+d(i)**2
    END DO
    DO k=1,npt
       temp=zero
       sum=zero
       DO j=1,n
          temp=temp+xpt(k,j)*xopt(j)
          sum=sum+xpt(k,j)*d(j)
       END DO
       temp=hcol(k)*temp
       sum=hcol(k)*sum
       DO i=1,n
          gc(i)=gc(i)+temp*xpt(k,i)
          gd(i)=gd(i)+sum*xpt(k,i)
       END DO
    END DO
    !
    !     Scale D and GD, with a sign change if required. Set S to another
    !     vector in the initial two dimensional subspace.
    !
    gg=zero
    sp=zero
    dhd=zero
    DO i=1,n
       gg=gg+gc(i)**2
       sp=sp+d(i)*gc(i)
       dhd=dhd+d(i)*gd(i)
    END DO
    scale=delta/SQRT(dd)
    IF (sp*dhd < zero) scale=-scale
    temp=zero
    IF (sp*sp > 0.99_dp*dd*gg) temp=one
    tau=scale*(ABS(sp)+half*scale*ABS(dhd))
    IF (gg*delsq < 0.01_dp*tau*tau) temp=one
    DO i=1,n
       d(i)=scale*d(i)
       gd(i)=scale*gd(i)
       s(i)=gc(i)+temp*gd(i)
    END DO
    !
    !     Begin the iteration by overwriting S with a vector that has the
    !     required length and direction, except that termination occurs if
    !     the given D and S are nearly parallel.
    !
    mainloop : DO
       iterc=iterc+1
       dd=zero
       sp=zero
       ss=zero
       DO i=1,n
          dd=dd+d(i)**2
          sp=sp+d(i)*s(i)
          ss=ss+s(i)**2
       END DO
       temp=dd*ss-sp*sp
       IF (temp <= 1.0e-8_dp*dd*ss) EXIT mainloop
       denom=SQRT(temp)
       DO i=1,n
          s(i)=(dd*s(i)-sp*d(i))/denom
          w(i)=zero
       END DO
       !
       !     Calculate the coefficients of the objective function on the circle
       !     beginning with the multiplication of S by the second derivative ma
       !
       DO k=1,npt
          sum=zero
          DO j=1,n
             sum=sum+xpt(k,j)*s(j)
          END DO
          sum=hcol(k)*sum
          DO i=1,n
             w(i)=w(i)+sum*xpt(k,i)
          END DO
       END DO
       cf1=zero
       cf2=zero
       cf3=zero
       cf4=zero
       cf5=zero
       DO i=1,n
          cf1=cf1+s(i)*w(i)
          cf2=cf2+d(i)*gc(i)
          cf3=cf3+s(i)*gc(i)
          cf4=cf4+d(i)*gd(i)
          cf5=cf5+s(i)*gd(i)
       END DO
       cf1=half*cf1
       cf4=half*cf4-cf1
       !
       !     Seek the value of the angle that maximizes the modulus of TAU.
       !
       taubeg=cf1+cf2+cf4
       taumax=taubeg
       tauold=taubeg
       isave=0
       iu=49
       temp=twopi/REAL(iu+1,DP)
       DO i=1,iu
          angle=REAL(i,dp)*temp
          cth=COS(angle)
          sth=SIN(angle)
          tau=cf1+(cf2+cf4*cth)*cth+(cf3+cf5*cth)*sth
          IF (ABS(tau) > ABS(taumax)) THEN
             taumax=tau
             isave=i
             tempa=tauold
          ELSE IF (i == isave+1) THEN
             tempb=taU
          END IF
          tauold=tau
       END DO
       IF (isave == 0) tempa=tau
       IF (isave == iu) tempb=taubeg
       step=zero
       IF (tempa /= tempb) THEN
          tempa=tempa-taumax
          tempb=tempb-taumax
          step=half*(tempa-tempb)/(tempa+tempb)
       END IF
       angle=temp*(REAL(isave,DP)+step)
       !
       !     Calculate the new D and GD. Then test for convergence.
       !
       cth=COS(angle)
       sth=SIN(angle)
       tau=cf1+(cf2+cf4*cth)*cth+(cf3+cf5*cth)*sth
       DO i=1,n
          d(i)=cth*d(i)+sth*s(i)
          gd(i)=cth*gd(i)+sth*w(i)
          s(i)=gc(i)+gd(i)
       END DO
       IF (ABS(tau) <= 1.1_dp*ABS(taubeg)) EXIT mainloop
       IF (iterc >= n) EXIT mainloop
    END DO mainloop

  END SUBROUTINE biglag
!******************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param npt ...
!> \param xopt ...
!> \param xpt ...
!> \param gq ...
!> \param hq ...
!> \param pq ...
!> \param delta ...
!> \param step ...
!> \param d ...
!> \param g ...
!> \param hd ...
!> \param hs ...
!> \param crvmin ...
! **************************************************************************************************
  SUBROUTINE trsapp (n,npt,xopt,xpt,gq,hq,pq,delta,step,d,g,hd,hs,crvmin)

    INTEGER, INTENT(INOUT)                   :: n, npt
    REAL(dp), DIMENSION(*), INTENT(INOUT)    :: xopt
    REAL(dp), DIMENSION(npt, *), &
      INTENT(INOUT)                          :: xpt
    REAL(dp), DIMENSION(*), INTENT(INOUT)    :: gq, hq, pq
    REAL(dp), INTENT(INOUT)                  :: delta
    REAL(dp), DIMENSION(*), INTENT(INOUT)    :: step, d, g, hd, hs
    REAL(dp), INTENT(INOUT)                  :: crvmin

    REAL(dp), PARAMETER                      :: half = 0.5_dp, zero = 0.0_dp

    INTEGER                                  :: i, isave, iterc, itermax, &
                                                itersw, iu, j
    LOGICAL                                  :: jump1, jump2
    REAL(dp) :: alpha, angle, angtest, bstep, cf, cth, dd, delsq, dg, dhd, &
      dhs, ds, gg, ggbeg, ggsav, qadd, qbeg, qmin, qnew, qred, qsav, ratio, &
      reduc, sg, sgk, shs, ss, sth, temp, tempa, tempb

!
!   N is the number of variables of a quadratic objective function, Q
!   The arguments NPT, XOPT, XPT, GQ, HQ and PQ have their usual meani
!     in order to define the current quadratic model Q.
!   DELTA is the trust region radius, and has to be positive.
!   STEP will be set to the calculated trial step.
!   The arrays D, G, HD and HS will be used for working space.
!   CRVMIN will be set to the least curvature of H along the conjugate
!     directions that occur, except that it is set to zero if STEP goe
!     all the way to the trust region boundary.
!
!   The calculation of STEP begins with the truncated conjugate gradient
!   method. If the boundary of the trust region is reached, then further
!   changes to STEP may be made, each one being in the 2D space spanned
!   by the current STEP and the corresponding gradient of Q. Thus STEP
!   should provide a substantial reduction to Q within the trust region
!
!   Initialization, which includes setting HD to H times XOPT.
!

    delsq=delta*delta
    iterc=0
    itermax=n
    itersw=itermax
    DO i=1,n
       d(i)=xopt(i)
    END DO
    CALL updatehd
    !
    !   Prepare for the first line search.
    !
    qred=zero
    dd=zero
    DO i=1,n
       step(i)=zero
       hs(i)=zero
       g(i)=gq(i)+hd(i)
       d(i)=-g(i)
       dd=dd+d(i)**2
    END DO
    crvmin=zero
    IF (dd == zero) RETURN
    ds=zero
    ss=zero
    gg=dd
    ggbeg=gg
    !
    !   Calculate the step to the trust region boundary and the product HD
    !
    jump1 = .FALSE.
    jump2 = .FALSE.
    mainloop : DO
       IF ( .NOT. jump2 ) THEN
          IF ( .NOT. jump1 ) THEN
             iterc=iterc+1
             temp=delsq-ss
             bstep=temp/(ds+SQRT(ds*ds+dd*temp))
             CALL updatehd
          END IF
          jump1 = .FALSE.
          IF (iterc <= itersw) THEN
             dhd=zero
             DO j=1,n
                dhd=dhd+d(j)*hd(j)
             END DO
             !
             !     Update CRVMIN and set the step-length ALPHA.
             !
             alpha=bstep
             IF (dhd > zero) THEN
                temp=dhd/dd
                IF (iterc == 1) crvmin=temp
                crvmin=MIN(crvmin,temp)
                alpha=MIN(alpha,gg/dhd)
             END IF
             qadd=alpha*(gg-half*alpha*dhd)
             qred=qred+qadd
             !
             !     Update STEP and HS.
             !
             ggsav=gg
             gg=zero
             DO i=1,n
                step(i)=step(i)+alpha*d(i)
                hs(i)=hs(i)+alpha*hd(i)
                gg=gg+(g(i)+hs(i))**2
             END DO
             !
             !     Begin another conjugate direction iteration if required.
             !
             IF (alpha < bstep) THEN
                IF (qadd <= 0.01_dp*qred) EXIT mainloop
                IF (gg <= 1.0e-4_dp*ggbeg) EXIT mainloop
                IF (iterc == itermax) EXIT mainloop
                temp=gg/ggsav
                dd=zero
                ds=zero
                ss=zero
                DO i=1,n
                   d(i)=temp*d(i)-g(i)-hs(i)
                   dd=dd+d(i)**2
                   ds=ds+d(i)*step(I)
                   ss=ss+step(i)**2
                END DO
                IF (ds <= zero) EXIT mainloop
                IF (ss < delsq) CYCLE mainloop
             END IF
             crvmin=zero
             itersw=iterc
             jump2 = .TRUE.
             IF (gg <= 1.0e-4_dp*ggbeg) EXIT mainloop
          ELSE
             jump2 = .FALSE.
          END IF
       END IF
       !
       !     Test whether an alternative iteration is required.
       !
!!!!  IF (gg <= 1.0e-4_dp*ggbeg) EXIT mainloop
       IF (jump2) THEN
          sg=zero
          shs=zero
          DO i=1,n
             sg=sg+step(i)*g(i)
             shs=shs+step(i)*hs(i)
          END DO
          sgk=sg+shs
          angtest=sgk/SQRT(gg*delsq)
          IF (angtest <= -0.99_dp) EXIT mainloop
          !
          !     Begin the alternative iteration by calculating D and HD and some
          !     scalar products.
          !
          iterc=iterc+1
          temp=SQRT(delsq*gg-sgk*sgk)
          tempa=delsq/temp
          tempb=sgk/temp
          DO i=1,n
             d(i)=tempa*(g(i)+hs(i))-tempb*step(i)
          END DO
          CALL updatehd
          IF (iterc <= itersw) THEN
             jump1 = .TRUE.
             CYCLE mainloop
          END IF
       END IF
       dg=zero
       dhd=zero
       dhs=zero
       DO i=1,n
          dg=dg+d(i)*g(i)
          dhd=dhd+hd(i)*d(i)
          dhs=dhs+hd(i)*step(i)
       END DO
       !
       !     Seek the value of the angle that minimizes Q.
       !
       cf=half*(shs-dhd)
       qbeg=sg+cf
       qsav=qbeg
       qmin=qbeg
       isave=0
       iu=49
       temp=twopi/REAL(iu+1,dp)
       DO i=1,iu
          angle=REAL(i,dp)*temp
          cth=COS(angle)
          sth=SIN(angle)
          qnew=(sg+cf*cth)*cth+(dg+dhs*cth)*sth
          IF (qnew < qmin) THEN
             qmin=qnew
             isave=i
             tempa=qsav
          ELSE IF (i == isave+1) THEN
             tempb=qnew
          END IF
          qsav=qnew
       END DO
       IF (isave == zero) tempa=qnew
       IF (isave == iu) tempb=qbeg
       angle=zero
       IF (tempa /= tempb) THEN
          tempa=tempa-qmin
          tempb=tempb-qmin
          angle=half*(tempa-tempb)/(tempa+tempb)
       END IF
       angle=temp*(REAL(isave,DP)+angle)
       !
       !     Calculate the new STEP and HS. Then test for convergence.
       !
       cth=COS(angle)
       sth=SIN(angle)
       reduc=qbeg-(sg+cf*cth)*cth-(dg+dhs*cth)*sth
       gg=zero
       DO i=1,n
          step(i)=cth*step(i)+sth*d(i)
          hs(i)=cth*hs(i)+sth*hd(i)
          gg=gg+(g(i)+hs(i))**2
       END DO
       qred=qred+reduc
       ratio=reduc/qred
       IF (iterc < itermax .AND. ratio > 0.01_dp) THEN
          jump2 = .TRUE.
       ELSE
          EXIT mainloop
       END IF

       IF (gg <= 1.0e-4_dp*ggbeg) EXIT mainloop

    END DO mainloop

    !*******************************************************************************
  CONTAINS
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
    SUBROUTINE updatehd
      INTEGER                                            :: i, ih, j, k

      DO i=1,n
         hd(i)=zero
      END DO
      DO k=1,npt
         temp=zero
         DO j=1,n
            temp=temp+xpt(k,j)*d(j)
         END DO
         temp=temp*pq(k)
         DO i=1,n
            hd(i)=hd(i)+temp*xpt(k,i)
         END DO
      END DO
      ih=0
      DO j=1,n
         DO i=1,j
            ih=ih+1
            IF (i < j) hd(j)=hd(j)+hq(ih)*d(i)
            hd(i)=hd(i)+hq(ih)*d(j)
         END DO
      END DO
    END SUBROUTINE updatehd

  END SUBROUTINE trsapp
!******************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param npt ...
!> \param bmat ...
!> \param zmat ...
!> \param idz ...
!> \param ndim ...
!> \param vlag ...
!> \param beta ...
!> \param knew ...
!> \param w ...
! **************************************************************************************************
  SUBROUTINE update (n,npt,bmat,zmat,idz,ndim,vlag,beta,knew,w)

      INTEGER, INTENT(IN)                                :: n, npt, ndim
      INTEGER, INTENT(INOUT)                             :: idz
      REAL(dp), DIMENSION(npt, *), INTENT(INOUT)         :: zmat
      REAL(dp), DIMENSION(ndim, *), INTENT(INOUT)        :: bmat
      REAL(dp), DIMENSION(*), INTENT(INOUT)              :: vlag
      REAL(dp), INTENT(INOUT)                            :: beta
      INTEGER, INTENT(INOUT)                             :: knew
      REAL(dp), DIMENSION(*), INTENT(INOUT)              :: w

      REAL(dp), PARAMETER                                :: one = 1.0_dp, zero = 0.0_dp

      INTEGER                                            :: i, iflag, j, ja, jb, jl, jp, nptm
      REAL(dp)                                           :: alpha, denom, scala, scalb, tau, tausq, &
                                                            temp, tempa, tempb

!   The arrays BMAT and ZMAT with IDZ are updated, in order to shift the
!   interpolation point that has index KNEW. On entry, VLAG contains the
!   components of the vector Theta*Wcheck+e_b of the updating formula
!   (6.11), and BETA holds the value of the parameter that has this na
!   The vector W is used for working space.
!

    nptm=npt-n-1
    !
    !     Apply the rotations that put zeros in the KNEW-th row of ZMAT.
    !
    jl=1
    DO j=2,nptm
       IF (j == idz) THEN
          jl=idz
       ELSE IF (zmat(knew,j) /= zero) THEN
          temp=SQRT(zmat(knew,jl)**2+zmat(knew,j)**2)
          tempa=zmat(knew,jl)/temp
          tempb=zmat(knew,j)/temp
          DO  I=1,NPT
             temp=tempa*zmat(i,jl)+tempb*zmat(i,j)
             zmat(i,j)=tempa*zmat(i,j)-tempb*zmat(i,jl)
             zmat(i,jl)=temp
          END DO
          zmat(knew,j)=zero
       END IF
    END DO
    !
    !   Put the first NPT components of the KNEW-th column of HLAG into W,
    !   and calculate the parameters of the updating formula.
    !
    tempa=zmat(knew,1)
    IF (idz >= 2) tempa=-tempa
    IF (jl > 1) tempb=zmat(knew,jl)
    DO i=1,npt
       w(i)=tempa*zmat(i,1)
       IF (jl > 1) w(i)=w(i)+tempb*zmat(i,jl)
    END DO
    alpha=w(knew)
    tau=vlag(knew)
    tausq=tau*tau
    denom=alpha*beta+tausq
    vlag(knew)=vlag(knew)-one
    !
    !   Complete the updating of ZMAT when there is only one nonzero eleme
    !   in the KNEW-th row of the new matrix ZMAT, but, if IFLAG is set to
    !   then the first column of ZMAT will be exchanged with another one l
    !
    iflag=0
    IF (JL == 1) THEN
       temp=SQRT(ABS(denom))
       tempb=tempa/temp
       tempa=tau/temp
       DO i=1,npt
          zmat(i,1)=tempa*zmat(i,1)-tempb*vlag(i)
       END DO
       IF (idz == 1 .AND. temp < zero) idz=2
       IF (idz >= 2 .AND. temp >= zero) iflag=1
    ELSE
       !
       !   Complete the updating of ZMAT in the alternative case.
       !
       ja=1
       IF (beta >= zero) ja=jl
       jb=jl+1-ja
       temp=zmat(knew,jb)/denom
       tempa=temp*beta
       tempb=temp*tau
       temp=zmat(knew,ja)
       scala=one/SQRT(ABS(beta)*temp*temp+tausq)
       scalb=scala*SQRT(ABS(denom))
       DO i=1,npt
          zmat(i,ja)=scala*(tau*zmat(i,ja)-temp*vlag(i))
          zmat(i,jb)=scalb*(zmat(i,jb)-tempa*w(i)-tempb*vlag(i))
       END DO
       IF (denom <= zero) THEN
          IF (beta < zero) idz=idz+1
          IF (beta >= zero) iflag=1
       END IF
    END IF
    !
    !   IDZ is reduced in the following case, and usually the first column
    !   of ZMAT is exchanged with a later one.
    !
    IF (iflag == 1) THEN
       idz=idz-1
       DO i=1,npt
          temp=zmat(i,1)
          zmat(i,1)=zmat(i,idz)
          zmat(i,idz)=temp
       END DO
    END IF
    !
    !   Finally, update the matrix BMAT.
    !
    DO j=1,n
       jp=npt+j
       w(jp)=bmat(knew,j)
       tempa=(alpha*vlag(jp)-tau*w(jp))/denom
       tempb=(-beta*w(jp)-tau*vlag(jp))/denom
       DO i=1,jp
          bmat(i,j)=bmat(i,j)+tempa*vlag(i)+tempb*w(i)
          IF (i > npt) bmat(jp,i-npt)=bmat(i,j)
       END DO
    END DO

  END SUBROUTINE update

END MODULE powell

!******************************************************************************
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE dkh_main
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale,&
                                              cp_fm_scale_and_add,&
                                              cp_fm_schur_product,&
                                              cp_fm_syrk,&
                                              cp_fm_transpose,&
                                              cp_fm_triangular_multiply,&
                                              cp_fm_upper_to_full
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose,&
                                              cp_fm_cholesky_reduce
   USE cp_fm_diag,                      ONLY: cp_fm_syevd
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_get,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE kinds,                           ONLY: dp
   USE physcon,                         ONLY: a_fine
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dkh_main'

   PUBLIC                               :: dkh_atom_transformation

CONTAINS

! **************************************************************************************************
!> \brief 2th order DKH calculations
!>
!> \param qs_env ...
!> \param matrix_s ...
!> \param matrix_v ...
!> \param matrix_t ...
!> \param matrix_pVp ...
!> \param n ...
!> \param dkh_order ...
!> \par Literature
!>  M. Reiher, A. Wolf, J. Chem. Phys. 121 (2004) 10944-10956
!>  A. Wolf, M. Reiher, B. A. Hess, J. Chem. Phys. 117 (2002) 9215-9226
!>
!>\par Note
!>      based on subroutines for DKH1 to DKH5 by
!>       A. Wolf, M. Reiher, B. A. Hess
!>
!>  INPUT:
!>    qs_env (:)        The quickstep environment
!>    n                 Number of primitive gaussians
!>    matrix_s    (:,:) overlap matrix
!>    matrix_pVp  (:,:) pVp matrix
!>
!>  IN_OUT:
!>    matrix_v    (:,:) input: nonrelativistic potential energy matrix
!>                      output: (ev1+ev2)
!>    matrix_t    (:,:) input: kinetic energy matrix
!>                      output: kinetic part of hamiltonian
!>                      in position space
!>
!>  INTERNAL
!>    sinv (:,:) inverted, orthogonalized overlap matrix
!>    ev0t (:)   DKH-even0 matrix in T-basis
!>    e    (:)   e=SQRT(p^2c^2+c^4)
!>    eig  (:,:) eigenvectors of sinv' h sinv
!>    tt   (:)   eigenvalues of sinv' h sinv
!>    revt (:,:) reverse transformation matrix T-basis -> position space
!>    aa   (:)   kinematical factors f. DKH SQRT((c^2+e(i))/(2.0*e(i)))
!>    rr   (:)   kinematical factors f. DKH c/(c^2+e(i))
!>    vt   (:,:) non relativistic potential matrix in T-basis
!>    pvpt (:,:) pvp integral matrix in T-basis
!>    ev1t (:,:) DKH-even1 matrix in T-basis
!>    evt2 (:,:) DKH-even2 matrix in T-basis
!>    ev1  (:,:) DKH-even1 matrix in position space
!>    ev2  (:,:) DKH-even2 matrix in position space
!>    ove (:,:) scratch
!>    aux (:,:) scratch
!>    velit  velocity of light 137 a.u.
!>    prea   prefactor, 1/137^2
!>    con2   prefactor, 2/137^2
!>    con    prefactor, 137^2
!> \author
!>     Jens Thar, Barbara Kirchner: Uni Bonn (09/2006)
!>     Markus Reiher: ETH Zurich (09/2006)
!>
! **************************************************************************************************
   SUBROUTINE DKH_full_transformation(qs_env, matrix_s, matrix_v, matrix_t, matrix_pVp, n, dkh_order)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_fm_type), POINTER                          :: matrix_s, matrix_v, matrix_t, matrix_pVp
      INTEGER, INTENT(IN)                                :: n, dkh_order

      CHARACTER(LEN=*), PARAMETER :: routineN = 'DKH_full_transformation', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i
      REAL(KIND=dp)                                      :: velit
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: aa, e, ev0t, rr, tt
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_full
      TYPE(cp_fm_type), POINTER :: matrix_aux, matrix_aux2, matrix_eig, matrix_ev1, matrix_ev2, &
         matrix_ev3, matrix_ev4, matrix_pe1p, matrix_rev, matrix_se, matrix_sinv

      CALL timeset(routineN, handle)
      NULLIFY (blacs_env)
      !-----------------------------------------------------------------------
      !     Define velocity of light
      !-----------------------------------------------------------------------

      !     velit = 137.0359895_dp   ! 1/a_fine  physcon
      velit = 1._dp/a_fine

      !-----------------------------------------------------------------------
      !     Construct the matrix structure
      !-----------------------------------------------------------------------
      CALL get_qs_env(qs_env, blacs_env=blacs_env)
      CALL cp_fm_struct_create(fmstruct=matrix_full, &
                               context=blacs_env, &
                               nrow_global=n, &
                               ncol_global=n)

      !-----------------------------------------------------------------------
      !     Allocate some matrices
      !-----------------------------------------------------------------------

      ALLOCATE (e(n))
      ALLOCATE (aa(n))
      ALLOCATE (rr(n))
      ALLOCATE (tt(n))
      ALLOCATE (ev0t(n))

      NULLIFY (matrix_eig)
      NULLIFY (matrix_aux)
      NULLIFY (matrix_aux2)
      NULLIFY (matrix_rev)
      NULLIFY (matrix_se)
      NULLIFY (matrix_ev1)
      NULLIFY (matrix_ev2)
      NULLIFY (matrix_ev3)
      NULLIFY (matrix_ev4)
      NULLIFY (matrix_sinv)
      NULLIFY (matrix_pe1p)

      CALL cp_fm_create(matrix_eig, matrix_full)
      CALL cp_fm_create(matrix_aux, matrix_full)
      CALL cp_fm_create(matrix_aux2, matrix_full)
      CALL cp_fm_create(matrix_rev, matrix_full)
      CALL cp_fm_create(matrix_se, matrix_full)
      CALL cp_fm_create(matrix_ev1, matrix_full)
      CALL cp_fm_create(matrix_ev2, matrix_full)
      CALL cp_fm_create(matrix_sinv, matrix_full)
      CALL cp_fm_create(matrix_ev3, matrix_full)
      CALL cp_fm_create(matrix_ev4, matrix_full)
      CALL cp_fm_create(matrix_pe1p, matrix_full)

      !-----------------------------------------------------------------------
      !     Now with Cholesky decomposition
      !-----------------------------------------------------------------------

      CALL cp_fm_to_fm(matrix_s, matrix_sinv)
      CALL cp_fm_cholesky_decompose(matrix_sinv, n)

      !-----------------------------------------------------------------------
      !     Calculate matrix representation from nonrelativistic T matrix
      !-----------------------------------------------------------------------

      CALL cp_fm_cholesky_reduce(matrix_t, matrix_sinv)
      CALL cp_fm_syevd(matrix_t, matrix_eig, tt)

      !-----------------------------------------------------------------------
      !     Calculate kinetic part of Hamiltonian in T-basis
      !-----------------------------------------------------------------------

      CALL kintegral(n, ev0t, tt, e, velit)

      !-----------------------------------------------------------------------
      !     Calculate reverse transformation matrix revt
      !-----------------------------------------------------------------------

      CALL cp_fm_to_fm(matrix_eig, matrix_rev)
      CALL cp_fm_triangular_multiply(matrix_sinv, matrix_rev, transpose_tr=.TRUE.)

      !-----------------------------------------------------------------------
      !     Calculate kinetic part of the Hamiltonian
      !-----------------------------------------------------------------------

      CALL cp_fm_to_fm(matrix_rev, matrix_aux)
      CALL cp_fm_column_scale(matrix_aux, ev0t)
      CALL cp_gemm("N", "T", n, n, n, 1.0_dp, matrix_rev, matrix_aux, 0.0_dp, matrix_t)

      !-----------------------------------------------------------------------
      !     Calculate kinematical factors for DKH
      !     only vectors present - will be done on every CPU
      !-----------------------------------------------------------------------

      DO i = 1, n
         aa(i) = SQRT((velit*velit+e(i))/(2.0_dp*e(i)))
         rr(i) = SQRT(velit*velit)/(velit*velit+e(i))
      END DO

      !-----------------------------------------------------------------------
      !     Transform v integrals to T-basis (v -> v(t))
      !-----------------------------------------------------------------------

      CALL cp_fm_cholesky_reduce(matrix_v, matrix_sinv)
      CALL cp_fm_upper_to_full(matrix_v, matrix_aux)
      CALL cp_gemm("T", "N", n, n, n, 1.0_dp, matrix_eig, matrix_v, 0.0_dp, matrix_aux)
      CALL cp_gemm("N", "N", n, n, n, 1.0_dp, matrix_aux, matrix_eig, 0.0_dp, matrix_v)

      !-----------------------------------------------------------------------
      !     Transform pVp integrals to T-basis (pVp -> pVp(t))
      !-----------------------------------------------------------------------

      CALL cp_fm_cholesky_reduce(matrix_pVp, matrix_sinv)
      CALL cp_fm_upper_to_full(matrix_pVp, matrix_aux)
      CALL cp_gemm("T", "N", n, n, n, 1.0_dp, matrix_eig, matrix_pVp, 0.0_dp, matrix_aux)
      CALL cp_gemm("N", "N", n, n, n, 1.0_dp, matrix_aux, matrix_eig, 0.0_dp, matrix_pVp)

      !-----------------------------------------------------------------------
      !     Calculate even1 in T-basis
      !-----------------------------------------------------------------------

      CALL even1(matrix_ev1, matrix_v, matrix_pvp, aa, rr, matrix_aux, matrix_aux2)

      !-----------------------------------------------------------------------
      !     Calculate even2 in T-basis
      !-----------------------------------------------------------------------

      CALL even2c(n, matrix_ev2, matrix_v, matrix_pVp, aa, rr, tt, e, matrix_aux)

      !-----------------------------------------------------------------------
      !     Calculate even3 in T-basis, only if requested
      !-----------------------------------------------------------------------

      IF (dkh_order .GE. 3) THEN
         CALL peven1p(n, matrix_pe1p, matrix_v, matrix_pvp, matrix_aux, matrix_aux2, aa, rr, tt)
         CALL even3b(n, matrix_ev3, matrix_ev1, matrix_pe1p, matrix_v, matrix_pvp, aa, rr, tt, e, matrix_aux)

         !-----------------------------------------------------------------------
         !     Transform even3 back to position space
         !-----------------------------------------------------------------------

         CALL cp_gemm("N", "N", n, n, n, 1.0_dp, matrix_rev, matrix_ev3, 0.0_dp, matrix_aux)
         CALL cp_gemm("N", "T", n, n, n, 1.0_dp, matrix_aux, matrix_rev, 0.0_dp, matrix_ev3)

         !-----------------------------------------------------------------------
         !     Calculate even4 in T-basis, only if requested
         !-----------------------------------------------------------------------

         IF (dkh_order .GE. 4) THEN
            CALL cp__b("dkh_main.F",262,"DKH order greater than 3 not yet available")
            !          CALL even4a(n,matrix_ev4%local_data,matrix_ev2%local_data,matrix_pe1p%local_data,matrix_v%local_data,&
            !                      matrix_pvp%local_data,aa,rr,tt,e)

            !-----------------------------------------------------------------------
            !     Transform even4 back to position space
            !-----------------------------------------------------------------------

            !        CALL cp_gemm("N","N",n,n,n,1.0_dp,matrix_rev,matrix_ev4,0.0_dp,matrix_aux)
            !        CALL cp_gemm("N","T",n,n,n,1.0_dp,matrix_aux,matrix_rev,0.0_dp,matrix_ev4)

         END IF
      END IF

      !----------------------------------------------------------------------
      !     Transform even1 back to position space
      !----------------------------------------------------------------------

      CALL cp_gemm("N", "N", n, n, n, 1.0_dp, matrix_rev, matrix_ev1, 0.0_dp, matrix_aux)
      CALL cp_gemm("N", "T", n, n, n, 1.0_dp, matrix_aux, matrix_rev, 0.0_dp, matrix_ev1)

      !-----------------------------------------------------------------------
      !     Transform even2 back to position space
      !-----------------------------------------------------------------------

      CALL cp_gemm("N", "N", n, n, n, 1.0_dp, matrix_rev, matrix_ev2, 0.0_dp, matrix_aux)
      CALL cp_gemm("N", "T", n, n, n, 1.0_dp, matrix_aux, matrix_rev, 0.0_dp, matrix_ev2)

      !-----------------------------------------------------------------------
      !     Calculate v in position space
      !-----------------------------------------------------------------------
      !
      CALL cp_fm_scale_and_add(1.0_dp, matrix_ev1, 1.0_dp, matrix_ev2)
      CALL cp_fm_upper_to_full(matrix_ev1, matrix_aux)
      CALL cp_fm_to_fm(matrix_ev1, matrix_v)
      IF (dkh_order .GE. 3) THEN
         CALL cp_fm_scale_and_add(1.0_dp, matrix_v, 1.0_dp, matrix_ev3)
         IF (dkh_order .GE. 4) THEN
            CALL cp_fm_scale_and_add(1.0_dp, matrix_v, 1.0_dp, matrix_ev4)
         END IF
      END IF

      !-----------------------------------------------------------------------

      CALL cp_fm_release(matrix_eig)
      CALL cp_fm_release(matrix_aux)
      CALL cp_fm_release(matrix_aux2)
      CALL cp_fm_release(matrix_rev)
      CALL cp_fm_release(matrix_se)
      CALL cp_fm_release(matrix_ev1)
      CALL cp_fm_release(matrix_ev2)
      CALL cp_fm_release(matrix_sinv)
      CALL cp_fm_release(matrix_ev3)
      CALL cp_fm_release(matrix_ev4)
      CALL cp_fm_release(matrix_pe1p)

      CALL cp_fm_struct_release(matrix_full)

      DEALLOCATE (ev0t, e, aa, rr, tt)

      CALL timestop(handle)

   END SUBROUTINE DKH_full_transformation

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param ev0t ...
!> \param tt ...
!> \param e ...
!> \param velit ...
! **************************************************************************************************
   SUBROUTINE kintegral(n, ev0t, tt, e, velit)
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: ev0t
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: tt
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: e
      REAL(KIND=dp), INTENT(IN)                          :: velit

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: con, con2, prea, ratio, tv1, tv2, tv3, &
                                                            tv4

      prea = 1/(velit*velit)
      con2 = prea+prea
      con = 1.0_dp/prea

      DO i = 1, n
         IF (tt(i) .LT. 0.0_dp) THEN
            WRITE (*, *) ' dkh_main.F | tt(', i, ') = ', tt(i)
         END IF

         !       If T is sufficiently small, use series expansion to avoid
         !       cancellation, otherwise calculate SQRT directly

         ev0t(i) = tt(i)
         ratio = tt(i)/velit
         IF (ratio .LE. 0.02_dp) THEN
            tv1 = tt(i)
            tv2 = -tv1*tt(i)*prea*0.5_dp
            tv3 = -tv2*tt(i)*prea
            tv4 = -tv3*tt(i)*prea*1.25_dp
            ev0t(i) = tv1+tv2+tv3+tv4
         ELSE
            ev0t(i) = con*(SQRT(1.0_dp+con2*tt(i))-1.0_dp)
         END IF
         e(i) = ev0t(i)+con
      END DO

      RETURN
   END SUBROUTINE kintegral

! **************************************************************************************************
!> \brief 1st order DKH-approximation
!> \param matrix_ev1 even1 output matrix
!> \param matrix_v potential matrix v in T-space
!> \param matrix_pvp pvp matrix in T-space
!> \param aa A-factors (diagonal)
!> \param rr R-factors (diagonal)
!> \param matrix_aux ...
!> \param matrix_aux2 ...
! **************************************************************************************************
   SUBROUTINE even1(matrix_ev1, matrix_v, matrix_pvp, aa, rr, matrix_aux, matrix_aux2)
      TYPE(cp_fm_type), POINTER                          :: matrix_ev1, matrix_v, matrix_pVp
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: aa, rr
      TYPE(cp_fm_type), POINTER                          :: matrix_aux, matrix_aux2

      CALL cp_fm_to_fm(matrix_v, matrix_aux)
      CALL cp_fm_column_scale(matrix_aux, aa)
      CALL cp_fm_transpose(matrix_aux, matrix_ev1)
      CALL cp_fm_column_scale(matrix_ev1, aa)

      CALL cp_fm_to_fm(matrix_pVp, matrix_aux)
      CALL cp_fm_column_scale(matrix_aux, aa)
      CALL cp_fm_column_scale(matrix_aux, rr)
      CALL cp_fm_transpose(matrix_aux, matrix_aux2)
      CALL cp_fm_column_scale(matrix_aux2, aa)
      CALL cp_fm_column_scale(matrix_aux2, rr)

      CALL cp_fm_scale_and_add(1.0_dp, matrix_ev1, 1.0_dp, matrix_aux2)

      RETURN
   END SUBROUTINE even1

! **************************************************************************************************
!> \brief 1st order DKH-approximation
!> \param n dimension of matrices
!> \param matrix_pe1p peven1p output matrix
!> \param matrix_v potential matrix v in T-space
!> \param matrix_pvp pvp matrix in T-space
!> \param matrix_aux ...
!> \param matrix_aux2 ...
!> \param aa A-factors (diagonal)
!> \param rr R-factors (diagonal)
!> \param tt ...
! **************************************************************************************************
   SUBROUTINE peven1p(n, matrix_pe1p, matrix_v, matrix_pvp, matrix_aux, matrix_aux2, aa, rr, tt)

      INTEGER, INTENT(IN)                                :: n
      TYPE(cp_fm_type), POINTER                          :: matrix_pe1p, matrix_v, matrix_pvp, &
                                                            matrix_aux, matrix_aux2
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: aa, rr, tt

      INTEGER                                            :: i, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: row_indices
      REAL(KIND=dp), DIMENSION(n)                        :: vec_ar, vec_arrt
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_fm_struct_type), POINTER                   :: vec_full
      TYPE(cp_fm_type), POINTER                          :: vec_a

      DO i = 1, n
         vec_ar(i) = aa(i)*rr(i)
         vec_arrt(i) = vec_ar(i)*rr(i)*tt(i)
      END DO

      CALL cp_fm_struct_get(matrix_v%matrix_struct, context=context)
      CALL cp_fm_struct_create(fmstruct=vec_full, &
                               context=context, &
                               nrow_global=n, &
                               ncol_global=1)

      NULLIFY (vec_a)
      CALL cp_fm_create(vec_a, vec_full)

      CALL cp_fm_get_info(matrix_v, nrow_local=nrow_local, &
                          row_indices=row_indices)

      DO i = 1, nrow_local
         vec_a%local_data(i, 1) = vec_arrt(row_indices(i))
      END DO

      CALL cp_fm_syrk('U', 'N', 1, 1.0_dp, vec_a, 1, 1, 0.0_dp, matrix_aux)
      CALL cp_fm_upper_to_full(matrix_aux, matrix_aux2)
      CALL cp_fm_schur_product(matrix_v, matrix_aux, matrix_pe1p)

      DO i = 1, nrow_local
         vec_a%local_data(i, 1) = vec_ar(row_indices(i))
      END DO

      CALL cp_fm_syrk('U', 'N', 1, 1.0_dp, vec_a, 1, 1, 0.0_dp, matrix_aux)
      CALL cp_fm_upper_to_full(matrix_aux, matrix_aux2)
      CALL cp_fm_schur_product(matrix_pvp, matrix_aux, matrix_aux2)

      CALL cp_fm_scale_and_add(4.0_dp, matrix_pe1p, 1.0_dp, matrix_aux2)

      CALL cp_fm_release(vec_a)
      CALL cp_fm_struct_release(vec_full)

      RETURN
   END SUBROUTINE peven1p

   !-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief 2nd order DK-approximation (original DK-transformation with U = SQRT(1+W^2) + W)
!> \param n Dimension of matrices
!> \param matrix_ev2 even2 output matrix = final result
!> \param matrix_v symmetric (n x n)-matrix containing (A V A)
!> \param matrix_pVp symmetric (n x n)-matrix containing (A P V P A)
!> \param aa A-Factors (DIAGONAL
!> \param rr R-Factors (DIAGONAL)
!> \param tt Nonrel. kinetic Energy (DIAGONAL)
!> \param e Rel. Energy = SQRT(p^2*c^2 + c^4)  (DIAGONAL)
!> \param matrix_aux ...
! **************************************************************************************************
   SUBROUTINE even2c(n, matrix_ev2, matrix_v, matrix_pVp, aa, rr, tt, e, matrix_aux)

      !***********************************************************************
      !                                                                      *
      !     Alexander Wolf, last modified: 20.02.2002 - DKH2                 *
      !                                                                      *
      !                                                                      *
      !     Version: 1.1  (20.2.2002) :  Usage of SR mat_add included        *
      !              1.0  (6.2.2002)                                         *
      !     Modification history:                                            *
      !     30.09.2006 Jens Thar: deleted obsolete F77 memory manager        *
      !     2008       Jens Thar: transfer to CP2K                           *
      !                                                                      *
      !     ev2 = 1/2 [W1,O1]                                                *
      !                                                                      *
      !***********************************************************************

      INTEGER, INTENT(IN)                                :: n
      TYPE(cp_fm_type), POINTER                          :: matrix_ev2, matrix_v, matrix_pVp
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: aa, rr, tt, e
      TYPE(cp_fm_type), POINTER                          :: matrix_aux

      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_full
      TYPE(cp_fm_type), POINTER                          :: matrix_apVpa, matrix_apVVpa, &
                                                            matrix_aux2, matrix_ava, matrix_avva

!     result  intermediate result of even2-calculation
!-----------------------------------------------------------------------
!     1.   General Structures and Patterns for DKH2
!-----------------------------------------------------------------------

      CALL cp_fm_struct_get(matrix_v%matrix_struct, context=context)
      CALL cp_fm_struct_create(fmstruct=matrix_full, &
                               context=context, &
                               nrow_global=n, &
                               ncol_global=n)

      NULLIFY (matrix_aux2)
      NULLIFY (matrix_ava)
      NULLIFY (matrix_avva)
      NULLIFY (matrix_apVpa)
      NULLIFY (matrix_apVVpa)

      CALL cp_fm_create(matrix_aux2, matrix_full)
      CALL cp_fm_create(matrix_ava, matrix_full)
      CALL cp_fm_create(matrix_avva, matrix_full)
      CALL cp_fm_create(matrix_apVpa, matrix_full)
      CALL cp_fm_create(matrix_apVVpa, matrix_full)

      CALL cp_fm_to_fm(matrix_v, matrix_ava)
      CALL cp_fm_to_fm(matrix_v, matrix_avva)
      CALL cp_fm_to_fm(matrix_pVp, matrix_apVpa)
      CALL cp_fm_to_fm(matrix_pVp, matrix_apVVpa)

      !  Calculate  v = A V A:

      CALL mat_axa(matrix_v, matrix_ava, n, aa, matrix_aux)

      !  Calculate  pvp = A P V P A:

      CALL mat_arxra(matrix_pVp, matrix_apVpa, n, aa, rr, matrix_aux)

      !  Calculate  vh = A V~ A:

      CALL mat_1_over_h(matrix_v, matrix_avva, e, matrix_aux)
      CALL cp_fm_to_fm(matrix_avva, matrix_aux2)
      CALL mat_axa(matrix_aux2, matrix_avva, n, aa, matrix_aux)

      !  Calculate  pvph = A P V~ P A:

      CALL mat_1_over_h(matrix_pVp, matrix_apVVpa, e, matrix_aux)
      CALL cp_fm_to_fm(matrix_apVVpa, matrix_aux2)
      CALL mat_arxra(matrix_aux2, matrix_apVVpa, n, aa, rr, matrix_aux)

      !  Calculate w1o1:

      CALL cp_gemm("N", "N", n, n, n, -1.0_dp, matrix_apVVpa, matrix_ava, 0.0_dp, matrix_aux2)
      CALL mat_muld(matrix_aux2, matrix_apVVpa, matrix_apVpa, n, 1.0_dp, 1.0_dp, tt, rr, matrix_aux)
      CALL mat_mulm(matrix_aux2, matrix_avva, matrix_ava, n, 1.0_dp, 1.0_dp, tt, rr, matrix_aux)
      CALL cp_gemm("N", "N", n, n, n, -1.0_dp, matrix_avva, matrix_apVpa, 1.0_dp, matrix_aux2)

      !  Calculate o1w1 (already stored in ev2):

      CALL cp_gemm("N", "N", n, n, n, 1.0_dp, matrix_apVpa, matrix_avva, 0.0_dp, matrix_ev2)
      CALL mat_muld(matrix_ev2, matrix_apVpa, matrix_apVVpa, n, -1.0_dp, 1.0_dp, tt, rr, matrix_aux)
      CALL mat_mulm(matrix_ev2, matrix_ava, matrix_avva, n, -1.0_dp, 1.0_dp, tt, rr, matrix_aux)
      CALL cp_gemm("N", "N", n, n, n, 1.0_dp, matrix_ava, matrix_apVVpa, 1.0_dp, matrix_ev2)

      !-----------------------------------------------------------------------
      !     2.   1/2 [W1,O1] = 1/2 W1O1 -  1/2 O1W1
      !-----------------------------------------------------------------------

      CALL cp_fm_scale_and_add(-0.5_dp, matrix_ev2, 0.5_dp, matrix_aux2)

      !-----------------------------------------------------------------------
      !     3.   Finish up the stuff!!
      !-----------------------------------------------------------------------

      CALL cp_fm_release(matrix_aux2)
      CALL cp_fm_release(matrix_ava)
      CALL cp_fm_release(matrix_avva)
      CALL cp_fm_release(matrix_apVpa)
      CALL cp_fm_release(matrix_apVVpa)

      CALL cp_fm_struct_release(matrix_full)

!    WRITE (*,*) "CAW:  DKH2 with even2c (Alex)"
!    WRITE (*,*) "JT:  Now available in cp2k"
      RETURN
   END SUBROUTINE even2c

   !-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param matrix_ev3 ...
!> \param matrix_ev1 ...
!> \param matrix_pe1p ...
!> \param matrix_v ...
!> \param matrix_pVp ...
!> \param aa ...
!> \param rr ...
!> \param tt ...
!> \param e ...
!> \param matrix_aux ...
! **************************************************************************************************
   SUBROUTINE even3b(n, matrix_ev3, matrix_ev1, matrix_pe1p, matrix_v, matrix_pVp, aa, rr, tt, e, matrix_aux)

      !***********************************************************************
      !                                                                      *
      !     Alexander Wolf, last modified:  20.2.2002 - DKH3                 *
      !                                                                      *
      !     3rd order DK-approximation (generalised DK-transformation)       *
      !                                                                      *
      !     Version: 1.1  (20.2.2002) :  Usage of SR mat_add included        *
      !              1.0  (7.2.2002)                                         *
      !                                                                      *
      !     ev3 = 1/2 [W1,[W1,E1]]                                           *
      !                                                                      *
      !     Modification history:                                            *
      !     30.09.2006 Jens Thar: deleted obsolete F77 memory manager        *
      !                                                                      *
      !         ----  Meaning of Parameters  ----                            *
      !                                                                      *
      !     n       in   Dimension of matrices                               *
      !     ev3     out  even3 output matrix = final result                  *
      !     e1      in   E1 = even1-operator                                 *
      !     pe1p    in   pE1p                                                *
      !     vv      in   potential v                                         *
      !     gg      in   pvp                                                 *
      !     aa      in   A-Factors (DIAGONAL)                                *
      !     rr      in   R-Factors (DIAGONAL)                                *
      !     tt      in   Nonrel. kinetic Energy (DIAGONAL)                   *
      !     e       in   Rel. Energy = SQRT(p^2*c^2 + c^4)  (DIAGONAL)       *
      !     result  intermediate result of even2-calculation
      !     vh      symmetric (n x n)-matrix containing (A V~ A)
      !     pvph    symmetric (n x n)-matrix containing (A P V~ P A)
      !     e1      E1
      !     pe1p    pE1p
      !     w1w1    (W1)^2
      !     w1e1w1  W1*E1*W1
      !     scr_i   temporary (n x n)-scratch-matrices (i=1,2)
      !                                                                      *
      !***********************************************************************

      INTEGER, INTENT(IN)                                :: n
      TYPE(cp_fm_type), POINTER                          :: matrix_ev3, matrix_ev1, matrix_pe1p, &
                                                            matrix_v, matrix_pVp
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: aa, rr, tt, e
      TYPE(cp_fm_type), POINTER                          :: matrix_aux

      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_full
      TYPE(cp_fm_type), POINTER                          :: matrix_apVVpa, matrix_aux2, matrix_avva, &
                                                            matrix_w1e1w1, matrix_w1w1

!-----------------------------------------------------------------------
!     1.   General Structures and Patterns for DKH3
!-----------------------------------------------------------------------

      CALL cp_fm_struct_get(matrix_v%matrix_struct, context=context)
      CALL cp_fm_struct_create(fmstruct=matrix_full, &
                               context=context, &
                               nrow_global=n, &
                               ncol_global=n)

      NULLIFY (matrix_aux2)
      NULLIFY (matrix_w1w1)
      NULLIFY (matrix_w1e1w1)
      NULLIFY (matrix_avva)
      NULLIFY (matrix_apVVpa)

      CALL cp_fm_create(matrix_aux2, matrix_full)
      CALL cp_fm_create(matrix_w1w1, matrix_full)
      CALL cp_fm_create(matrix_w1e1w1, matrix_full)
      CALL cp_fm_create(matrix_avva, matrix_full)
      CALL cp_fm_create(matrix_apVVpa, matrix_full)

      CALL cp_fm_to_fm(matrix_v, matrix_avva)
      CALL cp_fm_to_fm(matrix_pVp, matrix_apVVpa)

      !  Calculate  vh = A V~ A:

      CALL mat_1_over_h(matrix_v, matrix_avva, e, matrix_aux)
      CALL cp_fm_to_fm(matrix_avva, matrix_aux2)
      CALL mat_axa(matrix_aux2, matrix_avva, n, aa, matrix_aux)

      !  Calculate  pvph = A P V~ P A:

      CALL mat_1_over_h(matrix_pVp, matrix_apVVpa, e, matrix_aux)
      CALL cp_fm_to_fm(matrix_apVVpa, matrix_aux2)
      CALL mat_arxra(matrix_aux2, matrix_apVVpa, n, aa, rr, matrix_aux)

      !  Calculate w1w1:

      CALL cp_gemm("N", "N", n, n, n, 1.0_dp, matrix_apVVpa, matrix_avva, 0.0_dp, matrix_w1w1)
      CALL mat_muld(matrix_w1w1, matrix_apVVpa, matrix_apVVpa, n, -1.0_dp, 1.0_dp, tt, rr, matrix_aux2)
      CALL mat_mulm(matrix_w1w1, matrix_avva, matrix_avva, n, -1.0_dp, 1.0_dp, tt, rr, matrix_aux2)
      CALL cp_gemm("N", "N", n, n, n, 1.0_dp, matrix_avva, matrix_apVVpa, 1.0_dp, matrix_w1w1)

      !  Calculate w1e1w1: (warning: ev3 is scratch array)

      CALL mat_muld(matrix_aux, matrix_apVVpa, matrix_pe1p, n, 1.0_dp, 0.0_dp, tt, rr, matrix_aux2)
      CALL cp_gemm("N", "N", n, n, n, 1.0_dp, matrix_avva, matrix_pe1p, 0.0_dp, matrix_aux2)
      CALL cp_gemm("N", "N", n, n, n, 1.0_dp, matrix_aux, matrix_avva, 0.0_dp, matrix_w1e1w1)
      CALL mat_muld(matrix_w1e1w1, matrix_aux, matrix_apVVpa, n, -1.0_dp, 1.0_dp, tt, rr, matrix_ev3)
      CALL cp_gemm("N", "N", n, n, n, -1.0_dp, matrix_aux2, matrix_avva, 1.0_dp, matrix_w1e1w1)
      CALL mat_muld(matrix_w1e1w1, matrix_aux2, matrix_apVVpa, n, 1.0_dp, 1.0_dp, tt, rr, matrix_ev3)

      !-----------------------------------------------------------------------
      !     2.   ev3 = 1/2 (W1^2)E1 + 1/2 E1(W1^2) - W1E1W1
      !-----------------------------------------------------------------------

      CALL cp_gemm("N", "N", n, n, n, 0.5_dp, matrix_w1w1, matrix_ev1, 0.0_dp, matrix_ev3)
      CALL cp_gemm("N", "N", n, n, n, 0.5_dp, matrix_ev1, matrix_w1w1, 1.0_dp, matrix_ev3)
      CALL cp_fm_scale_and_add(1.0_dp, matrix_ev3, -1.0_dp, matrix_w1e1w1)

      !-----------------------------------------------------------------------
      !     3.   Finish up the stuff!!
      !-----------------------------------------------------------------------

      CALL cp_fm_release(matrix_aux2)
      CALL cp_fm_release(matrix_avva)
      CALL cp_fm_release(matrix_apVVpa)
      CALL cp_fm_release(matrix_w1w1)
      CALL cp_fm_release(matrix_w1e1w1)

      CALL cp_fm_struct_release(matrix_full)

!    WRITE (*,*) "CAW:  DKH3 with even3b (Alex)"
!    WRITE (*,*) "JT:  Now available in cp2k"
      RETURN
   END SUBROUTINE even3b

   !-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param ev4 ...
!> \param e1 ...
!> \param pe1p ...
!> \param vv ...
!> \param gg ...
! **************************************************************************************************
   SUBROUTINE even4a(n, ev4, e1, pe1p, vv, gg)

      !***********************************************************************
      !                                                                      *
      !     Alexander Wolf,   last modified: 25.02.2002   --   DKH4          *
      !                                                                      *
      !     4th order DK-approximation (scalar = spin-free)                  *
      !                                                                      *
      !     Version: 1.2  (25.2.2002) :  Elegant (short) way of calculation  *
      !                                  included                            *
      !              1.1  (20.2.2002) :  Usage of SR mat_add included        *
      !              1.0  (8.2.2002)                                         *
      !                                                                      *
      !     ev4  =  1/2 [W2,[W1,E1]] + 1/8 [W1,[W1,[W1,O1]]]  =              *
      !                                                                      *
      !          =      sum_1        +         sum_2                         *
      !                                                                      *
      !                                                                      *
      !     Modification history:                                            *
      !     30.09.2006 Jens Thar: deleted obsolete F77 memory manager        *
      !                (not working yet)                                     *
      !                                                                      *
      !         ----  Meaning of Parameters  ----                            *
      !                                                                      *
      !     n       in   Dimension of matrices                               *
      !     ev4     out  even4 output matrix = final result                  *
      !     e1     in   E1                                                   *
      !     pe1p   in   p(E1)p                                               *
      !     vv      in   potential v                                         *
      !     gg      in   pvp                                                 *
      !     aa      in   A-Factors (DIAGONAL)                                *
      !     rr      in   R-Factors (DIAGONAL)                                *
      !     tt      in   Nonrel. kinetic Energy (DIAGONAL)                   *
      !     e       in   Rel. Energy = SQRT(p^2*c^2 + c^4)  (DIAGONAL)       *
      !     v       symmetric (n x n)-matrix containing (A V A)              *
      !     pvp     symmetric (n x n)-matrix containing (A P V P A)          *
      !     vh      symmetric (n x n)-matrix containing (A V~ A)             *
      !     pvph    symmetric (n x n)-matrix containing (A P V~ P A)         *
      !     w1w1    (W1)^2                                                   *
      !     w1o1    W1*O1      (2-component formulation)                     *
      !     o1w1    O1*W1      (2-component formulation)                     *
      !     e1      symmetric (n x n)-matrix containing E1                   *
      !     pe1p    symmetric (n x n)-matrix containing p(E1)p               *
      !     sum_i   2 addends defined above  (i=1,2)                         *
      !     scr_i   temporary (n x n)-scratch-matrices (i=1,..,4)            *
      !     scrh_i  temp. (n x n)-scr.-mat. with energy-denom. (i=1,..,4)    *
      !                                                                      *
      !***********************************************************************

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: ev4
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: e1, pe1p, vv, gg

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: o1w1, pvp, pvph, scr_1, scr_2, scr_3, &
                                                            scr_4, scrh_1, scrh_2, scrh_3, scrh_4, &
                                                            sum_1, sum_2, v, vh, w1o1, w1w1

!C-----------------------------------------------------------------------
!C     1.   General Structures and Patterns for DKH4
!C-----------------------------------------------------------------------

      ALLOCATE (v(n, n))
      ALLOCATE (pVp(n, n))
      ALLOCATE (vh(n, n))
      ALLOCATE (pVph(n, n))
      v = 0.0_dp
      pVp = 0.0_dp
      vh = 0.0_dp
      pVph = 0.0_dp
      v(1:n, 1:n) = vv(1:n, 1:n)
      vh(1:n, 1:n) = vv(1:n, 1:n)
      pvp(1:n, 1:n) = gg(1:n, 1:n)
      pvph(1:n, 1:n) = gg(1:n, 1:n)

      ev4 = 0.0_dp
      !  Calculate  v = A V A:

      !     CALL mat_axa(v,n,aa)

      !  Calculate  pvp = A P V P A:

      !     CALL mat_arxra(pvp,n,aa,rr)

      !  Calculate  vh = A V~ A:

      !     CALL mat_1_over_h(vh,n,e)
      !     CALL mat_axa(vh,n,aa)

      !  Calculate  pvph = A P V~ P A:

      !     CALL mat_1_over_h(pvph,n,e)
      !     CALL mat_arxra(pvph,n,aa,rr)

      !  Create/Initialize necessary matrices:
      ALLOCATE (w1w1(n, n))
      w1w1 = 0.0_dp
      ALLOCATE (w1o1(n, n))
      w1o1 = 0.0_dp
      ALLOCATE (o1w1(n, n))
      o1w1 = 0.0_dp
      ALLOCATE (sum_1(n, n))
      sum_1 = 0.0_dp
      ALLOCATE (sum_2(n, n))
      sum_2 = 0.0_dp
      ALLOCATE (scr_1(n, n))
      scr_1 = 0.0_dp
      ALLOCATE (scr_2(n, n))
      scr_2 = 0.0_dp
      ALLOCATE (scr_3(n, n))
      scr_3 = 0.0_dp
      ALLOCATE (scr_4(n, n))
      scr_4 = 0.0_dp
      ALLOCATE (scrh_1(n, n))
      scrh_1 = 0.0_dp
      ALLOCATE (scrh_2(n, n))
      scrh_2 = 0.0_dp
      ALLOCATE (scrh_3(n, n))
      scrh_3 = 0.0_dp
      ALLOCATE (scrh_4(n, n))
      scrh_4 = 0.0_dp

      !  Calculate w1w1:
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pvph, n, vh, n, 0.0_dp, w1w1, n)
      !      CALL mat_muld(w1w1,pvph,pvph,n, -1.0_dp,1.0_dp,tt,rr)
      !      CALL mat_mulm(w1w1,vh,  vh,n,   -1.0_dp,1.0_dp,tt,rr)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, pvph, n, 1.0_dp, w1w1, n)

      !  Calculate w1o1:
      CALL dgemm("N", "N", n, n, n, -1.0_dp, pvph, n, v, n, 0.0_dp, w1o1, n)
      !      CALL mat_muld(w1o1,pvph,pvp,n,  1.0_dp,1.0_dp,tt,rr)
      !      CALL mat_mulm(w1o1,vh,  v,n,    1.0_dp,1.0_dp,tt,rr)
      CALL dgemm("N", "N", n, n, n, -1.0_dp, vh, n, pvp, n, 1.0_dp, w1o1, n)
      !  Calculate o1w1:
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pvp, n, vh, n, 0.0_dp, o1w1, n)
      !      CALL mat_muld(o1w1,pvp,pvph,n,  -1.0_dp,1.0_dp,tt,rr)
      !      CALL mat_mulm(o1w1,v,  vh,n,    -1.0_dp,1.0_dp,tt,rr)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, v, n, pvph, n, 1.0_dp, o1w1, n)

      !-----------------------------------------------------------------------
      !   2. sum_1 = 1/2 [W2,[W1,E1]] = 1/2 (W2W1E1 - W2E1W1 - W1E1W2 + E1W1W2)
      !-----------------------------------------------------------------------

      !  scr_i & scrh_i  for steps 2a (W2W1E1)  and 2b (W2E1W1):

      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, e1, n, 0.0_dp, scr_1, n)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pvph, n, e1, n, 0.0_dp, scr_2, n)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pe1p, n, vh, n, 0.0_dp, scr_3, n)
      !      CALL mat_muld(scr_4, pe1p,pvph,n,1.0_dp,0.0_dp,tt,rr)

      !      CALL mat_muld(scrh_1,pvph,pe1p,n,1.0_dp,0.0_dp,tt,rr)
      !      CALL mat_1_over_h(scrh_1,n,e)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, pe1p, n, 0.0_dp, scrh_2, n)
      !      CALL mat_1_over_h(scrh_2,n,e)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, e1, n, pvph, n, 0.0_dp, scrh_3, n)
      !      CALL mat_1_over_h(scrh_3,n,e)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, e1, n, vh, n, 0.0_dp, scrh_4, n)
      !      CALL mat_1_over_h(scrh_4,n,e)

      !  2a)  sum_1 = 1/2 W2W1E1               ( [1]-[8] )

      CALL dgemm("N", "N", n, n, n, 0.5_dp, scrh_1, n, scr_1, n, 0.0_dp, sum_1, n)
      !      CALL mat_muld(sum_1,scrh_1,scr_2,n,-0.5_dp,1.0_dp,tt,rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scrh_2, n, scr_1, n, 1.0_dp, sum_1, n)
      !      CALL mat_muld(sum_1,scrh_2,scr_2,n, 0.5_dp,1.0_dp,tt,rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scrh_3, n, scr_1, n, 1.0_dp, sum_1, n)
      !      CALL mat_muld(sum_1,scrh_3,scr_2,n, 0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_mulm(sum_1,scrh_4,scr_1,n, 0.5_dp,1.0_dp,tt,rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scrh_4, n, scr_2, n, 1.0_dp, sum_1, n)

      !  2b)  sum_1 = - 1/2 W2E1W1 (+ sum_1)   ( [9]-[16] )

      !      CALL mat_muld(sum_1,scrh_1,scr_3,n,-0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_muld(sum_1,scrh_1,scr_4,n, 0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_muld(sum_1,scrh_2,scr_3,n, 0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_muld(sum_1,scrh_2,scr_4,n,-0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_muld(sum_1,scrh_3,scr_3,n, 0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_muld(sum_1,scrh_3,scr_4,n,-0.5_dp,1.0_dp,tt,rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scrh_4, n, scr_3, n, 1.0_dp, sum_1, n)
      CALL dgemm("N", "N", n, n, n, 0.5_dp, scrh_4, n, scr_4, n, 1.0_dp, sum_1, n)

      !  scr_i & scrh_i  for steps 2c (W1E1W2)  and 2d (E1W1W2):

      !      CALL mat_muld(scr_1, pvph,pe1p,n,1.0_dp,0.0_dp,tt,rr)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, pe1p, n, 0.0_dp, scr_2, n)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, e1, n, pvph, n, 0.0_dp, scr_3, n)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, e1, n, vh, n, 0.0_dp, scr_4, n)

      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, e1, n, 0.0_dp, scrh_1, n)
      !      CALL mat_1_over_h(scrh_1,n,e)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pvph, n, e1, n, 0.0_dp, scrh_2, n)
      !      CALL mat_1_over_h(scrh_2,n,e)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pe1p, n, vh, n, 0.0_dp, scr_3, n)
      !      CALL mat_1_over_h(scrh_3,n,e)
      !      CALL mat_muld(scrh_4,pe1p,pvph,n,1.0_dp,0.0_dp,tt,rr)
      !      CALL mat_1_over_h(scrh_4,n,e)

      !  2c)  sum_1 = - 1/2 W1E1W2 (+ sum_1)   ( [17]-[24] )

      CALL dgemm("N", "N", n, n, n, 0.5_dp, scr_1, n, scrh_1, n, 0.0_dp, sum_1, n)
      !      CALL mat_muld(sum_1,scr_1,scrh_2,n,-0.5_dp,1.0_dp,tt,rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scr_2, n, scrh_1, n, 1.0_dp, sum_1, n)
      !      CALL mat_muld(sum_1,scr_2,scrh_2,n, 0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_muld(sum_1,scr_1,scrh_3,n,-0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_muld(sum_1,scr_1,scrh_4,n, 0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_muld(sum_1,scr_2,scrh_3,n, 0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_muld(sum_1,scr_2,scrh_4,n,-0.5_dp,1.0_dp,tt,rr)

      !  2d)  sum_1 = 1/2 E1W1W2 (+ sum_1)     ( [25]-[32] )

      CALL dgemm("N", "N", n, n, n, -0.5_dp, scr_3, n, scrh_1, n, 0.0_dp, sum_1, n)
      !      CALL mat_muld(sum_1,scr_3,scrh_2,n, 0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_mulm(sum_1,scr_4,scrh_1,n, 0.5_dp,1.0_dp,tt,rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scr_4, n, scrh_2, n, 1.0_dp, sum_1, n)
      !      CALL mat_muld(sum_1,scr_3,scrh_3,n, 0.5_dp,1.0_dp,tt,rr)
      !      CALL mat_muld(sum_1,scr_3,scrh_4,n,-0.5_dp,1.0_dp,tt,rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scr_4, n, scrh_3, n, 1.0_dp, sum_1, n)
      CALL dgemm("N", "N", n, n, n, 0.5_dp, scr_4, n, scrh_4, n, 1.0_dp, sum_1, n)

      !-----------------------------------------------------------------------
      !   3.  sum_2 = 1/8 [W1,[W1,[W1,O1]]] =
      !
      !             = 1/8 ( (W1^3)O1 - 3(W1^2)O1W1 + 3 W1O1(W1^2) - O1(W1^3) )
      !-----------------------------------------------------------------------

      CALL dgemm("N", "N", n, n, n, 0.125_dp, w1w1, n, w1o1, n, 0.0_dp, sum_2, n)
      CALL dgemm("N", "N", n, n, n, -0.375_dp, w1w1, n, o1w1, n, 1.0_dp, sum_2, n)
      CALL dgemm("N", "N", n, n, n, 0.375_dp, w1o1, n, w1w1, n, 1.0_dp, sum_2, n)
      CALL dgemm("N", "N", n, n, n, -0.125_dp, o1w1, n, w1w1, n, 1.0_dp, sum_2, n)

      !-----------------------------------------------------------------------
      !   4.  result = sum_1 + sum_2
      !-----------------------------------------------------------------------

      CALL mat_add(ev4, 1.0_dp, sum_1, 1.0_dp, sum_2, n)

      !-----------------------------------------------------------------------
      !   5. Finish up the stuff!!
      !-----------------------------------------------------------------------

      DEALLOCATE (v, pvp, vh, pvph, w1w1, w1o1, o1w1, sum_1, sum_2)
      DEALLOCATE (scr_1, scr_2, scr_3, scr_4, scrh_1, scrh_2, scrh_3, scrh_4)

!    WRITE (*,*) "CAW:  DKH4 with even4a (Alex)"
!    WRITE (*,*) "JT:   Now available in cp2k"

      RETURN
   END SUBROUTINE even4a

   !-----------------------------------------------------------------------
   !                                                                      -
   !     Matrix routines for DKH-procedure                                -
   !     Alexander Wolf                                                   -
   !     modifed: Jens Thar: Mem manager deleted                          -
   !     This file contains the                                           -
   !      following subroutines:                                          -
   !                                 1. mat_1_over_h                      -
   !                                 2. mat_axa                           -
   !                                 3. mat_arxra                         -
   !                                 4. mat_mulm                          -
   !                                 5. mat_muld                          -
   !                                 6. mat_add                           -
   !                                                                      -
   !-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param matrix_p ...
!> \param matrix_pp ...
!> \param e ...
!> \param matrix_aux ...
! **************************************************************************************************
   SUBROUTINE mat_1_over_h(matrix_p, matrix_pp, e, matrix_aux)

      !***********************************************************************
      !                                                                      *
      !   2. SR mat_1_over_h: Transform matrix p into matrix p/(e(i)+e(j))   *
      !                                                                      *
      !   p    in  REAL(:,:) :   matrix p                                    *
      !   e    in  REAL(:)   :   rel. energy (diagonal)                      *
      !                                                                      *
      !***********************************************************************

      TYPE(cp_fm_type), POINTER                          :: matrix_p, matrix_pp
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: e
      TYPE(cp_fm_type), POINTER                          :: matrix_aux

      INTEGER                                            :: i, j, ncol_local, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices

      CALL cp_fm_get_info(matrix_aux, nrow_local=nrow_local, ncol_local=ncol_local, &
                          row_indices=row_indices, col_indices=col_indices)

      DO j = 1, ncol_local
         DO i = 1, nrow_local
            matrix_aux%local_data(i, j) = 1/(e(row_indices(i))+e(col_indices(j)))
         ENDDO
      ENDDO

      CALL cp_fm_schur_product(matrix_p, matrix_aux, matrix_pp)

      RETURN

   END SUBROUTINE mat_1_over_h
   !C-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param matrix_x ...
!> \param matrix_axa ...
!> \param n ...
!> \param a ...
!> \param matrix_aux ...
! **************************************************************************************************
   SUBROUTINE mat_axa(matrix_x, matrix_axa, n, a, matrix_aux)

      !C***********************************************************************
      !C                                                                      *
      !C   3. SR mat_axa: Transform matrix p into matrix  a*p*a               *
      !C                                                                      *
      !C   p    in  REAL(:,:):   matrix p                                     *
      !C   a    in  REAL(:)  :   A-factors (diagonal)                         *
      !CJT n    in  INTEGER  :   dimension of matrix p                        *
      !C                                                                      *
      !C***********************************************************************

      TYPE(cp_fm_type), POINTER                          :: matrix_x, matrix_axa
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: a
      TYPE(cp_fm_type), POINTER                          :: matrix_aux

      INTEGER                                            :: i, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: row_indices
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_fm_struct_type), POINTER                   :: vec_full
      TYPE(cp_fm_type), POINTER                          :: vec_a

      CALL cp_fm_struct_get(matrix_x%matrix_struct, context=context)
      CALL cp_fm_struct_create(fmstruct=vec_full, &
                               context=context, &
                               nrow_global=n, &
                               ncol_global=1)

      NULLIFY (vec_a)
      CALL cp_fm_create(vec_a, vec_full)

      CALL cp_fm_get_info(matrix_x, nrow_local=nrow_local, &
                          row_indices=row_indices)

      DO i = 1, nrow_local
         vec_a%local_data(i, 1) = a(row_indices(i))
      END DO

      CALL cp_fm_syrk('U', 'N', 1, 1.0_dp, vec_a, 1, 1, 0.0_dp, matrix_aux)
      CALL cp_fm_upper_to_full(matrix_aux, matrix_axa)
      CALL cp_fm_schur_product(matrix_x, matrix_aux, matrix_axa)

      !     DO i=1,n
      !       DO j=1,n
      !          p(i,j)=p(i,j)*a(i)*a(j)
      !       ENDDO
      !     ENDDO

      CALL cp_fm_release(vec_a)
      CALL cp_fm_struct_release(vec_full)

      RETURN
   END SUBROUTINE mat_axa
   !C-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param matrix_x ...
!> \param matrix_axa ...
!> \param n ...
!> \param a ...
!> \param r ...
!> \param matrix_aux ...
! **************************************************************************************************
   SUBROUTINE mat_arxra(matrix_x, matrix_axa, n, a, r, matrix_aux)

      !C***********************************************************************
      !C                                                                      *
      !C   4. SR mat_arxra: Transform matrix p into matrix  a*r*p*r*a         *
      !C                                                                      *
      !C   p    in  REAL(:,:) :   matrix p                                    *
      !C   a    in  REAL(:)   :   A-factors (diagonal)                        *
      !C   r    in  REAL(:)   :   R-factors (diagonal)                        *
      !C   n    in  INTEGER   :   dimension of matrix p                       *
      !C                                                                      *
      !C***********************************************************************

      TYPE(cp_fm_type), POINTER                          :: matrix_x, matrix_axa
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: a, r
      TYPE(cp_fm_type), POINTER                          :: matrix_aux

      INTEGER                                            :: i, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: row_indices
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_fm_struct_type), POINTER                   :: vec_full
      TYPE(cp_fm_type), POINTER                          :: vec_a

      CALL cp_fm_struct_get(matrix_x%matrix_struct, context=context)
      CALL cp_fm_struct_create(fmstruct=vec_full, &
                               context=context, &
                               nrow_global=n, &
                               ncol_global=1)

      CALL cp_fm_get_info(matrix_aux, nrow_local=nrow_local, &
                          row_indices=row_indices)

      NULLIFY (vec_a)
      CALL cp_fm_create(vec_a, vec_full)

      DO i = 1, nrow_local
         vec_a%local_data(i, 1) = a(row_indices(i))*r(row_indices(i))
      END DO

      CALL cp_fm_syrk('U', 'N', 1, 1.0_dp, vec_a, 1, 1, 0.0_dp, matrix_aux)
      CALL cp_fm_upper_to_full(matrix_aux, matrix_axa)
      CALL cp_fm_schur_product(matrix_x, matrix_aux, matrix_axa)

      CALL cp_fm_release(vec_a)
      CALL cp_fm_struct_release(vec_full)

      RETURN
   END SUBROUTINE mat_arxra
   !C-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param matrix_p ...
!> \param matrix_q ...
!> \param matrix_r ...
!> \param n ...
!> \param alpha ...
!> \param beta ...
!> \param t ...
!> \param rr ...
!> \param matrix_aux ...
! **************************************************************************************************
   SUBROUTINE mat_mulm(matrix_p, matrix_q, matrix_r, n, alpha, beta, t, rr, matrix_aux)

      !C***********************************************************************
      !C                                                                      *
      !C   5. SR mat_mulm:  Multiply matrices according to:                   *
      !C                                                                      *
      !C                      p = alpha*q*(..P^2..)*r + beta*p                *
      !C                                                                      *
      !C   p      out  REAL(:,:):   matrix p                                  *
      !C   q      in   REAL(:,:):   matrix q                                  *
      !C   r      in   REAL(:,.):   matrix r                                  *
      !C   n      in   INTEGER  :   dimension n of matrices                   *
      !C   alpha  in   REAL(dp) :                                             *
      !C   beta   in   REAL(dp) :                                             *
      !C   t      in   REAL(:)  :   non-rel. kinetic energy  (diagonal)       *
      !C   rr     in   REAL(:)  :   R-factors  (diagonal)                     *
      !C                                                                      *
      !C***********************************************************************

      TYPE(cp_fm_type), POINTER                          :: matrix_p, matrix_q, matrix_r
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: alpha, beta
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: t, rr
      TYPE(cp_fm_type), POINTER                          :: matrix_aux

      INTEGER                                            :: i
      REAL(KIND=dp), DIMENSION(n)                        :: vec

      CALL cp_fm_to_fm(matrix_q, matrix_aux)

      DO i = 1, n
         vec(i) = 2.0_dp*t(i)*rr(i)*rr(i)
      END DO
      CALL cp_fm_column_scale(matrix_aux, vec)

      CALL cp_gemm("N", "N", n, n, n, alpha, matrix_aux, matrix_r, beta, matrix_p)
      RETURN

   END SUBROUTINE mat_mulm
   !C-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param matrix_p ...
!> \param matrix_q ...
!> \param matrix_r ...
!> \param n ...
!> \param alpha ...
!> \param beta ...
!> \param t ...
!> \param rr ...
!> \param matrix_aux ...
! **************************************************************************************************
   SUBROUTINE mat_muld(matrix_p, matrix_q, matrix_r, n, alpha, beta, t, rr, matrix_aux)

      !C***********************************************************************
      !C                                                                      *
      !C   16. SR mat_muld:  Multiply matrices according to:                  *
      !C                                                                      *
      !C                      p = alpha*q*(..1/P^2..)*r + beta*p              *
      !C                                                                      *
      !C   p      out  REAL(:,:):   matrix p                                  *
      !C   q      in   REAL(:,:):   matrix q                                  *
      !C   r      in   REAL(:,:):   matrix r                                  *
      !C   n      in   INTEGER  :   Dimension of all matrices                 *
      !C   alpha  in   REAL(dp) :                                             *
      !C   beta   in   REAL(dp) :                                             *
      !C   t      in   REAL(:)  :   non-rel. kinetic energy  (diagonal)       *
      !C   rr     in   REAL(:)  :   R-factors  (diagonal)                     *
      !C                                                                      *
      !C***********************************************************************

      TYPE(cp_fm_type), POINTER                          :: matrix_p, matrix_q, matrix_r
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: alpha, beta
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: t, rr
      TYPE(cp_fm_type), POINTER                          :: matrix_aux

      INTEGER                                            :: i
      REAL(KIND=dp), DIMENSION(n)                        :: vec

      CALL cp_fm_to_fm(matrix_q, matrix_aux)

      DO i = 1, n
         vec(i) = 0.5_dp/(t(i)*rr(i)*rr(i))
      END DO

      CALL cp_fm_column_scale(matrix_aux, vec)

      CALL cp_gemm("N", "N", n, n, n, alpha, matrix_aux, matrix_r, beta, matrix_p)

      RETURN

   END SUBROUTINE mat_muld
   !C-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param s ...
!> \param v ...
!> \param h ...
!> \param pVp ...
!> \param n ...
!> \param dkh_order ...
! **************************************************************************************************
   SUBROUTINE DKH_atom_transformation(s, v, h, pVp, n, dkh_order)

      !-----------------------------------------------------------------------
      !                                                                      *
      !  INPUT:                                                              *
      !    n          Number of primitive gaussians                          *
      !    s    (:,:) overlap matrix                                         *
      !    pVp  (:,:) pVp matrix                                             *
      !                                                                      *
      !  IN_OUT:                                                             *
      !    v    (:,:) input: nonrelativistic potential energy matrix         *
      !               output: (ev1+ev2)                                      *
      !    h    (:,:) input: kinetic energy matrix                           *
      !               output: kinetic part of hamiltonian in position space  *
      !                                                                      *
      !  INTERNAL                                                            *
      !    sinv (:,:) inverted, orthogonalized overlap matrix                *
      !    ev0t (:)   DKH-even0 matrix in T-basis                            *
      !    e    (:)   e=SQRT(p^2c^2+c^4)                                     *
      !    eig  (:,:) eigenvectors of sinv' h sinv                           *
      !    tt   (:)   eigenvalues of sinv' h sinv                            *
      !    revt (:,:) reverse transformation matrix T-basis -> position space*
      !    aa   (:)   kinematical factors f. DKH SQRT((c^2+e(i))/(2.0*e(i))) *
      !    rr   (:)   kinematical factors f. DKH c/(c^2+e(i))                *
      !    vt   (:,:) non relativistic potential matrix in T-basis           *
      !    pvpt (:,:) pvp integral matrix in T-basis                         *
      !    ev1t (:,:) DKH-even1 matrix in T-basis                            *
      !    evt2 (:,:) DKH-even2 matrix in T-basis                            *
      !    ev1  (:,:) DKH-even1 matrix in position space                     *
      !    ev2  (:,:) DKH-even2 matrix in position space                     *
      !    ove (:,:) scratch                                                 *
      !    aux (:,:) scratch                                                 *
      !    velit  velocity of light 137 a.u.                                 *
      !    prea   prefactor, 1/137^2                                         *
      !    con2   prefactor, 2/137^2                                         *
      !    con    prefactor, 137^2                                           *
      !-----------------------------------------------------------------------

      !-----------------------------------------------------------------------

      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: s, v, h, pVp
      INTEGER, INTENT(IN)                                :: n, dkh_order

      CHARACTER(len=*), PARAMETER :: routineN = 'DKH_atom_transformation', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k
      REAL(KIND=dp)                                      :: velit
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: aa, e, ev0t, rr, tt
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: aux, eig, ev1, ev1t, ev2, ev2t, ev3, &
                                                            ev3t, ev4, ev4t, ove, pev1tp, pVpt, &
                                                            revt, sinv, vt

      IF (dkh_order .LT. 0) RETURN

      !CAW  pp: p^2-values (in momentum-space), stored as matrix!!
      !-----------------------------------------------------------------------
      !     Define velocity of light
      !-----------------------------------------------------------------------
      !      velit = 137.0359895_dp

      velit = 1._dp/a_fine

      !-----------------------------------------------------------------------
      !     Allocate some matrices
      !-----------------------------------------------------------------------

      ALLOCATE (eig(n, n))
      ALLOCATE (sinv(n, n))
      ALLOCATE (revt(n, n))
      ALLOCATE (aux(n, n))
      ALLOCATE (ove(n, n))
      ALLOCATE (ev0t(n))
      ALLOCATE (e(n))
      ALLOCATE (aa(n))
      ALLOCATE (rr(n))
      ALLOCATE (tt(n))
      ALLOCATE (ev1t(n, n))
      ALLOCATE (ev2t(n, n))
      ALLOCATE (ev3t(n, n))
      ALLOCATE (ev4t(n, n))
      ALLOCATE (vt(n, n))
      ALLOCATE (pVpt(n, n))
      ALLOCATE (pev1tp(n, n))
      ALLOCATE (ev1(n, n))
      ALLOCATE (ev2(n, n))
      ALLOCATE (ev3(n, n))
      ALLOCATE (ev4(n, n))

      !-----------------------------------------------------------------------
      !     Schmidt-orthogonalize overlap matrix
      !-----------------------------------------------------------------------

      CALL sog(n, s, sinv)

      !-----------------------------------------------------------------------
      !     Calculate matrix representation from nonrelativistic T matrix
      !-----------------------------------------------------------------------

      CALL dkh_diag(h, n, eig, tt, sinv, aux, 0)

      !-----------------------------------------------------------------------
      !     Calculate kinetic part of Hamiltonian in T-basis
      !-----------------------------------------------------------------------

      CALL kintegral_a(n, ev0t, tt, e, velit)

      !-----------------------------------------------------------------------
      !     Calculate reverse transformation matrix revt
      !-----------------------------------------------------------------------

      CALL dgemm("N", "N", n, n, n, 1.0_dp, sinv, n, eig, n, 0.0_dp, aux, n)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, s, n, aux, n, 0.0_dp, revt, n)

      !-----------------------------------------------------------------------
      !     Calculate kinetic part of the Hamiltonian
      !-----------------------------------------------------------------------

      h = 0.0_dp
      DO i = 1, n
         DO j = 1, i
            DO k = 1, n
               h(i, j) = h(i, j)+revt(i, k)*revt(j, k)*ev0t(k)
               h(j, i) = h(i, j)
            END DO
         END DO
      END DO

      !-----------------------------------------------------------------------
      !     Calculate kinematical factors for DKH
      !-----------------------------------------------------------------------

      DO i = 1, n
         aa(i) = SQRT((velit*velit+e(i))/(2.0_dp*e(i)))
         rr(i) = SQRT(velit*velit)/(velit*velit+e(i))
      END DO

      !-----------------------------------------------------------------------
      !     Transform v integrals to T-basis (v -> vt)
      !-----------------------------------------------------------------------

      CALL trsm(v, sinv, ove, n, aux)
      CALL trsm(ove, eig, vt, n, aux)

      !-----------------------------------------------------------------------
      !     Transform pVp integrals to T-basis (pVp -> pVpt)
      !-----------------------------------------------------------------------

      CALL trsm(pVp, sinv, ove, n, aux)
      CALL trsm(ove, eig, pVpt, n, aux)

      !-----------------------------------------------------------------------
      !     Calculate even1 in T-basis
      !-----------------------------------------------------------------------

      IF (dkh_order .GE. 1) THEN
         CALL even1_a(n, ev1t, vt, pvpt, aa, rr)

         !----------------------------------------------------------------------
         !     Transform even1 back to position space
         !----------------------------------------------------------------------

         CALL dgemm("N", "N", n, n, n, 1.0_dp, revt, n, ev1t, n, 0.0_dp, aux, n)
         CALL dgemm("N", "T", n, n, n, 1.0_dp, aux, n, revt, n, 0.0_dp, ev1, n)
      END IF

      !-----------------------------------------------------------------------
      !     Calculate even2 in T-basis
      !-----------------------------------------------------------------------

      IF (dkh_order .GE. 2) THEN
         CALL even2c_a(n, ev2t, vt, pvpt, aa, rr, tt, e)

         !-----------------------------------------------------------------------
         !     Transform even2 back to position space
         !-----------------------------------------------------------------------

         aux = 0.0_dp
         CALL dgemm("N", "N", n, n, n, 1.0_dp, revt, n, ev2t, n, 0.0_dp, aux, n)
         CALL dgemm("N", "T", n, n, n, 1.0_dp, aux, n, revt, n, 0.0_dp, ev2, n)
      END IF

      !-----------------------------------------------------------------------
      !     Calculate even3 in T-basis, only if requested
      !-----------------------------------------------------------------------

      IF (dkh_order .GE. 3) THEN
         CALL peven1p_a(n, pev1tp, vt, pvpt, aa, rr, tt)
         CALL even3b_a(n, ev3t, ev1t, pev1tp, vt, pvpt, aa, rr, tt, e)

         !-----------------------------------------------------------------------
         !     Transform even3 back to position space
         !-----------------------------------------------------------------------
         aux = 0.0_dp
         CALL dgemm("N", "N", n, n, n, 1.0_dp, revt, n, ev3t, n, 0.0_dp, aux, n)
         CALL dgemm("N", "T", n, n, n, 1.0_dp, aux, n, revt, n, 0.0_dp, ev3, n)

         !-----------------------------------------------------------------------
         !     Calculate even4 in T-basis, only if requested
         !-----------------------------------------------------------------------

         IF (dkh_order .GE. 4) THEN
            CALL even4a_a(n, ev4t, ev1t, pev1tp, vt, pvpt, aa, rr, tt, e)

            !-----------------------------------------------------------------------
            !     Transform even4 back to position space
            !-----------------------------------------------------------------------
            aux = 0.0_dp
            CALL dgemm("N", "N", n, n, n, 1.0_dp, revt, n, ev4t, n, 0.0_dp, aux, n)
            CALL dgemm("N", "T", n, n, n, 1.0_dp, aux, n, revt, n, 0.0_dp, ev4, n)
         END IF
      END IF

      IF (dkh_order .GE. 4) THEN
         CALL cp__b("dkh_main.F",1514,"DKH 4")
      END IF
      !-----------------------------------------------------------------------
      !     Calculate v in position space
      !-----------------------------------------------------------------------

      IF (dkh_order .GE. 1) THEN
         CALL mat_add2(v, 0.0_dp, 1.0_dp, ev1, n)
      END IF
      IF (dkh_order .GE. 2) THEN
         CALL mat_add2(v, 1.0_dp, 1.0_dp, ev2, n)
      END IF
      IF (dkh_order .GE. 3) THEN
         CALL mat_add2(v, 1.0_dp, 1.0_dp, ev3, n)
      END IF
      IF (dkh_order .GE. 4) THEN
         CALL mat_add2(v, 1.0_dp, 1.0_dp, ev4, n)
      END IF

      !-----------------------------------------------------------------------

      DEALLOCATE (eig, sinv, revt, ove, aux, vt, pVpt, ev1, ev2, ev3, ev4, ev1t, ev2t, ev3t, ev4t, pev1tp)
      DEALLOCATE (ev0t, e, aa, rr, tt)

      RETURN
   END SUBROUTINE dkh_atom_transformation

   !-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param ev0t ...
!> \param tt ...
!> \param e ...
!> \param velit ...
! **************************************************************************************************
   SUBROUTINE kintegral_a(n, ev0t, tt, e, velit)

      !-----------------------------------------------------------------------

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: ev0t
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: tt
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT)           :: e
      REAL(KIND=dp), INTENT(IN)                          :: velit

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: con, con2, prea, ratio, tv1, tv2, tv3, &
                                                            tv4

!-----------------------------------------------------------------------

      DO i = 1, n
         IF (tt(i) .LT. 0.0_dp) THEN
            WRITE (*, *) ' dkh_main.F | tt(', i, ') = ', tt(i)
         END IF

         !       Calculate some constants

         prea = 1/(velit*velit)
         con2 = prea+prea
         con = 1.0_dp/prea

         !       If T is sufficiently small, use series expansion to avoid
         !       cancellation, otherwise calculate SQRT directly

         ev0t(i) = tt(i)
         ratio = tt(i)/velit
         IF (ratio .LE. 0.02_dp) THEN
            tv1 = tt(i)
            tv2 = -tv1*tt(i)*prea/2.0_dp
            tv3 = -tv2*tt(i)*prea
            tv4 = -tv3*tt(i)*prea*1.25_dp
            ev0t(i) = tv1+tv2+tv3+tv4
         ELSE
            ev0t(i) = con*(SQRT(1.0_dp+con2*tt(i))-1.0_dp)
         END IF
         e(i) = ev0t(i)+con
      END DO

      RETURN
   END SUBROUTINE kintegral_a

   !-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param ev1t ...
!> \param vt ...
!> \param pvpt ...
!> \param aa ...
!> \param rr ...
! **************************************************************************************************
   SUBROUTINE even1_a(n, ev1t, vt, pvpt, aa, rr)

      !-----------------------------------------------------------------------
      !                                                                      -
      !     1st order DKH-approximation                                      -
      !                                                                      -
      !     n    in   dimension of matrices                                  -
      !     ev1t out  even1 output matrix                                    -
      !     vt   in   potential matrix v in T-space                          -
      !     pvpt in   pvp matrix in T-space                                  -
      !     aa   in   A-factors (diagonal)                                   -
      !     rr   in   R-factors (diagonal)                                   -
      !                                                                      -
      !-----------------------------------------------------------------------

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: ev1t
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: vt, pvpt
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: aa, rr

      INTEGER                                            :: i, j

!-----------------------------------------------------------------------

      DO i = 1, n
         DO j = 1, i
            ev1t(i, j) = vt(i, j)*aa(i)*aa(j)+pVpt(i, j)*aa(i)*rr(i)*aa(j)*rr(j)
            ev1t(j, i) = ev1t(i, j)
         END DO
      END DO

      RETURN
   END SUBROUTINE even1_a

   !-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param pev1tp ...
!> \param vt ...
!> \param pvpt ...
!> \param aa ...
!> \param rr ...
!> \param tt ...
! **************************************************************************************************
   SUBROUTINE peven1p_a(n, pev1tp, vt, pvpt, aa, rr, tt)

      !-----------------------------------------------------------------------
      !                                                                      -
      !     1st order DKH-approximation                                      -
      !                                                                      -
      !     n      in   dimension of matrices                                -
      !     pev1tp out  peven1p output matrix                                -
      !     vt     in   potential matrix v in T-space                        -
      !     pvpt   in   pvp matrix in T-space                                -
      !     aa     in   A-factors (diagonal)                                 -
      !     rr     in   R-factors (diagonal)                                 -
      !                                                                      -
      !-----------------------------------------------------------------------

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: pev1tp
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: vt, pvpt
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: aa, rr, tt

      INTEGER                                            :: i, j

!-----------------------------------------------------------------------

      DO i = 1, n
         DO j = 1, i
            pev1tp(i, j) = 4.0_dp*vt(i, j)*aa(i)*aa(j)*rr(i)*rr(i)*rr(j)*rr(j)*tt(i)*tt(j)+ &
                           pVpt(i, j)*aa(i)*rr(i)*aa(j)*rr(j)
            pev1tp(j, i) = pev1tp(i, j)
         END DO
      END DO

      RETURN
   END SUBROUTINE peven1p_a

   !-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param ev2 ...
!> \param vv ...
!> \param gg ...
!> \param aa ...
!> \param rr ...
!> \param tt ...
!> \param e ...
! **************************************************************************************************
   SUBROUTINE even2c_a(n, ev2, vv, gg, aa, rr, tt, e)

      !***********************************************************************
      !                                                                      *
      !     Alexander Wolf, last modified: 20.02.2002 - DKH2                 *
      !                                                                      *
      !     2nd order DK-approximation ( original DK-transformation with     *
      !                                       U = SQRT(1+W^2) + W        )   *
      !                                                                      *
      !     Version: 1.1  (20.2.2002) :  Usage of SR mat_add included        *
      !              1.0  (6.2.2002)                                         *
      !     Modification history:                                            *
      !     30.09.2006 Jens Thar: deleted obsolete F77 memory manager        *
      !                                                                      *
      !     ev2 = 1/2 [W1,O1]                                                *
      !                                                                      *
      !         ----  Meaning of Parameters  ----                            *
      !                                                                      *
      !     n       in   Dimension of matrices                               *
      !     ev2     out  even2 output matrix = final result                  *
      !     vv      in   potential v                                         *
      !     gg      in   pvp                                                 *
      !     aa      in   A-Factors (DIAGONAL)                                *
      !     rr      in   R-Factors (DIAGONAL)                                *
      !     tt      in   Nonrel. kinetic Energy (DIAGONAL)                   *
      !     e       in   Rel. Energy = SQRT(p^2*c^2 + c^4)  (DIAGONAL)       *
      !     result  intermediate result of even2-calculation                 *
      !     v       symmetric (n x n)-matrix containing (A V A)              *
      !     pvp     symmetric (n x n)-matrix containing (A P V P A)          *
      !     vh      symmetric (n x n)-matrix containing (A V~ A)             *
      !     pvph    symmetric (n x n)-matrix containing (A P V~ P A)         *
      !     w1o1    W1*O1 (2-component form)                                 *
      !     o1w1    O1*W1 (2-component form)                                 *
      !                                                                      *
      !***********************************************************************

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: ev2
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: vv, gg
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: aa, rr, tt, e

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: o1w1, pvp, pvph, v, vh, w1o1

!-----------------------------------------------------------------------
!     1.   General Structures and Patterns for DKH2
!-----------------------------------------------------------------------

      ALLOCATE (v(n, n))
      ALLOCATE (pVp(n, n))
      ALLOCATE (vh(n, n))
      ALLOCATE (pVph(n, n))
      v = 0.0_dp
      pVp = 0.0_dp
      vh = 0.0_dp
      pVph = 0.0_dp
      v(1:n, 1:n) = vv(1:n, 1:n)
      vh(1:n, 1:n) = vv(1:n, 1:n)
      pvp(1:n, 1:n) = gg(1:n, 1:n)
      pvph(1:n, 1:n) = gg(1:n, 1:n)

      ev2 = 0.0_dp
      !  Calculate  v = A V A:

      CALL mat_axa_a(v, n, aa)

      !  Calculate  pvp = A P V P A:

      CALL mat_arxra_a(pvp, n, aa, rr)

      !  Calculate  vh = A V~ A:

      CALL mat_1_over_h_a(vh, n, e)
      CALL mat_axa_a(vh, n, aa)

      !  Calculate  pvph = A P V~ P A:

      CALL mat_1_over_h_a(pvph, n, e)
      CALL mat_arxra_a(pvph, n, aa, rr)

      !  Create/Initialize necessary matrices:
      ALLOCATE (w1o1(n, n))
      ALLOCATE (o1w1(n, n))
      w1o1 = 0.0_dp
      o1w1 = 0.0_dp

      !  Calculate w1o1:
      CALL dgemm("N", "N", n, n, n, -1.0_dp, pvph, n, v, n, 0.0_dp, w1o1, n)
      CALL mat_muld_a(w1o1, pvph, pvp, n, 1.0_dp, 1.0_dp, tt, rr)
      CALL mat_mulm_a(w1o1, vh, v, n, 1.0_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, -1.0_dp, vh, n, pvp, n, 1.0_dp, w1o1, n)
      !  Calculate o1w1:
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pvp, n, vh, n, 0.0_dp, o1w1, n)
      CALL mat_muld_a(o1w1, pvp, pvph, n, -1.0_dp, 1.0_dp, tt, rr)
      CALL mat_mulm_a(o1w1, v, vh, n, -1.0_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, v, n, pvph, n, 1.0_dp, o1w1, n)
      !  Calculate in symmetric pakets

      !-----------------------------------------------------------------------
      !     2.   1/2 [W1,O1] = 1/2 W1O1 -  1/2 O1W1
      !-----------------------------------------------------------------------

      CALL mat_add(ev2, 0.5_dp, w1o1, -0.5_dp, o1w1, n)

      !-----------------------------------------------------------------------
      !     3.   Finish up the stuff!!
      !-----------------------------------------------------------------------

      DEALLOCATE (v, vh, pvp, pvph, w1o1, o1w1)

!    WRITE (*,*) "CAW:  DKH2 with even2c (Alex)"
!    WRITE (*,*) "!JT:  Now available in cp2k"
      RETURN
   END SUBROUTINE even2c_a

   !-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param ev3 ...
!> \param e1 ...
!> \param pe1p ...
!> \param vv ...
!> \param gg ...
!> \param aa ...
!> \param rr ...
!> \param tt ...
!> \param e ...
! **************************************************************************************************
   SUBROUTINE even3b_a(n, ev3, e1, pe1p, vv, gg, aa, rr, tt, e)

      !***********************************************************************
      !                                                                      *
      !     Alexander Wolf, last modified:  20.2.2002 - DKH3                 *
      !                                                                      *
      !     3rd order DK-approximation (generalised DK-transformation)       *
      !                                                                      *
      !     Version: 1.1  (20.2.2002) :  Usage of SR mat_add included        *
      !              1.0  (7.2.2002)                                         *
      !                                                                      *
      !     ev3 = 1/2 [W1,[W1,E1]]                                           *
      !                                                                      *
      !     Modification history:                                            *
      !     30.09.2006 Jens Thar: deleted obsolete F77 memory manager        *
      !                                                                      *
      !         ----  Meaning of Parameters  ----                            *
      !                                                                      *
      !     n       in   Dimension of matrices                               *
      !     ev3     out  even3 output matrix = final result                  *
      !     e1      in   E1 = even1-operator                                 *
      !     pe1p    in   pE1p                                                *
      !     vv      in   potential v                                         *
      !     gg      in   pvp                                                 *
      !     aa      in   A-Factors (DIAGONAL)                                *
      !     rr      in   R-Factors (DIAGONAL)                                *
      !     tt      in   Nonrel. kinetic Energy (DIAGONAL)                   *
      !     e       in   Rel. Energy = SQRT(p^2*c^2 + c^4)  (DIAGONAL)       *
      !     result  intermediate result of even2-calculation                 *
      !     vh      symmetric (n x n)-matrix containing (A V~ A)             *
      !     pvph    symmetric (n x n)-matrix containing (A P V~ P A)         *
      !     e1      E1                                                       *
      !     pe1p    pE1p                                                     *
      !     w1w1    (W1)^2                                                   *
      !     w1e1w1  W1*E1*W1                                                 *
      !     scr_i   temporary (n x n)-scratch-matrices (i=1,2)               *
      !                                                                      *
      !***********************************************************************

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: ev3
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: e1, pe1p, vv, gg
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: aa, rr, tt, e

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: pvph, scr_1, scr_2, vh, w1e1w1, w1w1

!-----------------------------------------------------------------------
!     1.   General Structures and Patterns for DKH3
!-----------------------------------------------------------------------

      ALLOCATE (vh(n, n))
      ALLOCATE (pVph(n, n))
      vh = 0.0_dp
      pVph = 0.0_dp
      vh(1:n, 1:n) = vv(1:n, 1:n)
      pvph(1:n, 1:n) = gg(1:n, 1:n)

      ev3 = 0.0_dp

      !  Calculate  vh = A V~ A:

      CALL mat_1_over_h_a(vh, n, e)
      CALL mat_axa_a(vh, n, aa)

      !  Calculate  pvph = A P V~ P A:

      CALL mat_1_over_h_a(pvph, n, e)
      CALL mat_arxra_a(pvph, n, aa, rr)

      !  Create/Initialize necessary matrices:
      ALLOCATE (w1w1(n, n))
      ALLOCATE (w1e1w1(n, n))
      ALLOCATE (scr_1(n, n))
      ALLOCATE (scr_2(n, n))
      w1w1 = 0.0_dp
      w1e1w1 = 0.0_dp
      scr_1 = 0.0_dp
      scr_2 = 0.0_dp

      !  Calculate w1w1:
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pvph, n, vh, n, 0.0_dp, w1w1, n)
      CALL mat_muld_a(w1w1, pvph, pvph, n, -1.0_dp, 1.0_dp, tt, rr)
      CALL mat_mulm_a(w1w1, vh, vh, n, -1.0_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, pvph, n, 1.0_dp, w1w1, n)

      !  Calculate w1e1w1:
      CALL mat_muld_a(scr_1, pvph, pe1p, n, 1.0_dp, 0.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, pe1p, n, 0.0_dp, scr_2, n)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, scr_1, n, vh, n, 0.0_dp, w1e1w1, n)
      CALL mat_muld_a(w1e1w1, scr_1, pvph, n, -1.0_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, -1.0_dp, scr_2, n, vh, n, 1.0_dp, w1e1w1, n)
      CALL mat_muld_a(w1e1w1, scr_2, pvph, n, 1.0_dp, 1.0_dp, tt, rr)

      !-----------------------------------------------------------------------
      !     2.   ev3 = 1/2 (W1^2)E1 + 1/2 E1(W1^2) - W1E1W1
      !-----------------------------------------------------------------------

      CALL dgemm("N", "N", n, n, n, 0.5_dp, w1w1, n, e1, n, 0.0_dp, ev3, n)
      CALL dgemm("N", "N", n, n, n, 0.5_dp, e1, n, w1w1, n, 1.0_dp, ev3, n)
      CALL mat_add2(ev3, 1.0_dp, -1.0_dp, w1e1w1, n)

      !-----------------------------------------------------------------------
      !     3.   Finish up the stuff!!
      !-----------------------------------------------------------------------

      DEALLOCATE (vh, pvph, w1w1, w1e1w1, scr_1, scr_2)

!    WRITE (*,*) "CAW:  DKH3 with even3b (Alex)"
!    WRITE (*,*) "JT:  Now available in cp2k"
      RETURN
   END SUBROUTINE even3b_a

   !-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param ev4 ...
!> \param e1 ...
!> \param pe1p ...
!> \param vv ...
!> \param gg ...
!> \param aa ...
!> \param rr ...
!> \param tt ...
!> \param e ...
! **************************************************************************************************
   SUBROUTINE even4a_a(n, ev4, e1, pe1p, vv, gg, aa, rr, tt, e)

      !***********************************************************************
      !                                                                      *
      !     Alexander Wolf,   last modified: 25.02.2002   --   DKH4          *
      !                                                                      *
      !     4th order DK-approximation (scalar = spin-free)                  *
      !                                                                      *
      !     Version: 1.2  (25.2.2002) :  Elegant (short) way of calculation  *
      !                                  included                            *
      !              1.1  (20.2.2002) :  Usage of SR mat_add included        *
      !              1.0  (8.2.2002)                                         *
      !                                                                      *
      !     ev4  =  1/2 [W2,[W1,E1]] + 1/8 [W1,[W1,[W1,O1]]]  =              *
      !                                                                      *
      !          =      sum_1        +         sum_2                         *
      !                                                                      *
      !                                                                      *
      !     Modification history:                                            *
      !     30.09.2006 Jens Thar: deleted obsolete F77 memory manager        *
      !                                                                      *
      !         ----  Meaning of Parameters  ----                            *
      !                                                                      *
      !     n       in   Dimension of matrices                               *
      !     ev4     out  even4 output matrix = final result                  *
      !     e1     in   E1                                                   *
      !     pe1p   in   p(E1)p                                               *
      !     vv      in   potential v                                         *
      !     gg      in   pvp                                                 *
      !     aa      in   A-Factors (DIAGONAL)                                *
      !     rr      in   R-Factors (DIAGONAL)                                *
      !     tt      in   Nonrel. kinetic Energy (DIAGONAL)                   *
      !     e       in   Rel. Energy = SQRT(p^2*c^2 + c^4)  (DIAGONAL)       *
      !     v       symmetric (n x n)-matrix containing (A V A)              *
      !     pvp     symmetric (n x n)-matrix containing (A P V P A)          *
      !     vh      symmetric (n x n)-matrix containing (A V~ A)             *
      !     pvph    symmetric (n x n)-matrix containing (A P V~ P A)         *
      !     w1w1    (W1)^2                                                   *
      !     w1o1    W1*O1      (2-component formulation)                     *
      !     o1w1    O1*W1      (2-component formulation)                     *
      !     e1      symmetric (n x n)-matrix containing E1                   *
      !     pe1p    symmetric (n x n)-matrix containing p(E1)p               *
      !     sum_i   2 addends defined above  (i=1,2)                         *
      !     scr_i   temporary (n x n)-scratch-matrices (i=1,..,4)            *
      !     scrh_i  temp. (n x n)-scr.-mat. with energy-denom. (i=1,..,4)    *
      !                                                                      *
      !***********************************************************************

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: ev4
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: e1, pe1p, vv, gg
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: aa, rr, tt, e

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: o1w1, pvp, pvph, scr_1, scr_2, scr_3, &
                                                            scr_4, scrh_1, scrh_2, scrh_3, scrh_4, &
                                                            sum_1, sum_2, v, vh, w1o1, w1w1

!C-----------------------------------------------------------------------
!C     1.   General Structures and Patterns for DKH4
!C-----------------------------------------------------------------------

      ALLOCATE (v(n, n))
      ALLOCATE (pVp(n, n))
      ALLOCATE (vh(n, n))
      ALLOCATE (pVph(n, n))
      v = 0.0_dp
      pVp = 0.0_dp
      vh = 0.0_dp
      pVph = 0.0_dp
      v(1:n, 1:n) = vv(1:n, 1:n)
      vh(1:n, 1:n) = vv(1:n, 1:n)
      pvp(1:n, 1:n) = gg(1:n, 1:n)
      pvph(1:n, 1:n) = gg(1:n, 1:n)

      ev4 = 0.0_dp
      !  Calculate  v = A V A:

      CALL mat_axa_a(v, n, aa)

      !  Calculate  pvp = A P V P A:

      CALL mat_arxra_a(pvp, n, aa, rr)

      !  Calculate  vh = A V~ A:

      CALL mat_1_over_h_a(vh, n, e)
      CALL mat_axa_a(vh, n, aa)

      !  Calculate  pvph = A P V~ P A:

      CALL mat_1_over_h_a(pvph, n, e)
      CALL mat_arxra_a(pvph, n, aa, rr)

      !  Create/Initialize necessary matrices:
      ALLOCATE (w1w1(n, n))
      w1w1 = 0.0_dp
      ALLOCATE (w1o1(n, n))
      w1o1 = 0.0_dp
      ALLOCATE (o1w1(n, n))
      o1w1 = 0.0_dp
      ALLOCATE (sum_1(n, n))
      sum_1 = 0.0_dp
      ALLOCATE (sum_2(n, n))
      sum_2 = 0.0_dp
      ALLOCATE (scr_1(n, n))
      scr_1 = 0.0_dp
      ALLOCATE (scr_2(n, n))
      scr_2 = 0.0_dp
      ALLOCATE (scr_3(n, n))
      scr_3 = 0.0_dp
      ALLOCATE (scr_4(n, n))
      scr_4 = 0.0_dp
      ALLOCATE (scrh_1(n, n))
      scrh_1 = 0.0_dp
      ALLOCATE (scrh_2(n, n))
      scrh_2 = 0.0_dp
      ALLOCATE (scrh_3(n, n))
      scrh_3 = 0.0_dp
      ALLOCATE (scrh_4(n, n))
      scrh_4 = 0.0_dp

      !  Calculate w1w1:
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pvph, n, vh, n, 0.0_dp, w1w1, n)
      CALL mat_muld_a(w1w1, pvph, pvph, n, -1.0_dp, 1.0_dp, tt, rr)
      CALL mat_mulm_a(w1w1, vh, vh, n, -1.0_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, pvph, n, 1.0_dp, w1w1, n)

      !  Calculate w1o1:
      CALL dgemm("N", "N", n, n, n, -1.0_dp, pvph, n, v, n, 0.0_dp, w1o1, n)
      CALL mat_muld_a(w1o1, pvph, pvp, n, 1.0_dp, 1.0_dp, tt, rr)
      CALL mat_mulm_a(w1o1, vh, v, n, 1.0_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, -1.0_dp, vh, n, pvp, n, 1.0_dp, w1o1, n)
      !  Calculate o1w1:
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pvp, n, vh, n, 0.0_dp, o1w1, n)
      CALL mat_muld_a(o1w1, pvp, pvph, n, -1.0_dp, 1.0_dp, tt, rr)
      CALL mat_mulm_a(o1w1, v, vh, n, -1.0_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, v, n, pvph, n, 1.0_dp, o1w1, n)

      !-----------------------------------------------------------------------
      !   2. sum_1 = 1/2 [W2,[W1,E1]] = 1/2 (W2W1E1 - W2E1W1 - W1E1W2 + E1W1W2)
      !-----------------------------------------------------------------------

      !  scr_i & scrh_i  for steps 2a (W2W1E1)  and 2b (W2E1W1):

      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, e1, n, 0.0_dp, scr_1, n)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pvph, n, e1, n, 0.0_dp, scr_2, n)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pe1p, n, vh, n, 0.0_dp, scr_3, n)
      CALL mat_muld_a(scr_4, pe1p, pvph, n, 1.0_dp, 0.0_dp, tt, rr)

      CALL mat_muld_a(scrh_1, pvph, pe1p, n, 1.0_dp, 0.0_dp, tt, rr)
      CALL mat_1_over_h_a(scrh_1, n, e)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, pe1p, n, 0.0_dp, scrh_2, n)
      CALL mat_1_over_h_a(scrh_2, n, e)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, e1, n, pvph, n, 0.0_dp, scrh_3, n)
      CALL mat_1_over_h_a(scrh_3, n, e)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, e1, n, vh, n, 0.0_dp, scrh_4, n)
      CALL mat_1_over_h_a(scrh_4, n, e)

      !  2a)  sum_1 = 1/2 W2W1E1               ( [1]-[8] )

      CALL dgemm("N", "N", n, n, n, 0.5_dp, scrh_1, n, scr_1, n, 0.0_dp, sum_1, n)
      CALL mat_muld_a(sum_1, scrh_1, scr_2, n, -0.5_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scrh_2, n, scr_1, n, 1.0_dp, sum_1, n)
      CALL mat_muld_a(sum_1, scrh_2, scr_2, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scrh_3, n, scr_1, n, 1.0_dp, sum_1, n)
      CALL mat_muld_a(sum_1, scrh_3, scr_2, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL mat_mulm_a(sum_1, scrh_4, scr_1, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scrh_4, n, scr_2, n, 1.0_dp, sum_1, n)

      !  2b)  sum_1 = - 1/2 W2E1W1 (+ sum_1)   ( [9]-[16] )

      CALL mat_muld_a(sum_1, scrh_1, scr_3, n, -0.5_dp, 1.0_dp, tt, rr)
      CALL mat_muld_a(sum_1, scrh_1, scr_4, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL mat_muld_a(sum_1, scrh_2, scr_3, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL mat_muld_a(sum_1, scrh_2, scr_4, n, -0.5_dp, 1.0_dp, tt, rr)
      CALL mat_muld_a(sum_1, scrh_3, scr_3, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL mat_muld_a(sum_1, scrh_3, scr_4, n, -0.5_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scrh_4, n, scr_3, n, 1.0_dp, sum_1, n)
      CALL dgemm("N", "N", n, n, n, 0.5_dp, scrh_4, n, scr_4, n, 1.0_dp, sum_1, n)

      !  scr_i & scrh_i  for steps 2c (W1E1W2)  and 2d (E1W1W2):

      CALL mat_muld_a(scr_1, pvph, pe1p, n, 1.0_dp, 0.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, pe1p, n, 0.0_dp, scr_2, n)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, e1, n, pvph, n, 0.0_dp, scr_3, n)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, e1, n, vh, n, 0.0_dp, scr_4, n)

      CALL dgemm("N", "N", n, n, n, 1.0_dp, vh, n, e1, n, 0.0_dp, scrh_1, n)
      CALL mat_1_over_h_a(scrh_1, n, e)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pvph, n, e1, n, 0.0_dp, scrh_2, n)
      CALL mat_1_over_h_a(scrh_2, n, e)
      CALL dgemm("N", "N", n, n, n, 1.0_dp, pe1p, n, vh, n, 0.0_dp, scr_3, n)
      CALL mat_1_over_h_a(scrh_3, n, e)
      CALL mat_muld_a(scrh_4, pe1p, pvph, n, 1.0_dp, 0.0_dp, tt, rr)
      CALL mat_1_over_h_a(scrh_4, n, e)

      !  2c)  sum_1 = - 1/2 W1E1W2 (+ sum_1)   ( [17]-[24] )

      CALL dgemm("N", "N", n, n, n, 0.5_dp, scr_1, n, scrh_1, n, 0.0_dp, sum_1, n)
      CALL mat_muld_a(sum_1, scr_1, scrh_2, n, -0.5_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scr_2, n, scrh_1, n, 1.0_dp, sum_1, n)
      CALL mat_muld_a(sum_1, scr_2, scrh_2, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL mat_muld_a(sum_1, scr_1, scrh_3, n, -0.5_dp, 1.0_dp, tt, rr)
      CALL mat_muld_a(sum_1, scr_1, scrh_4, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL mat_muld_a(sum_1, scr_2, scrh_3, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL mat_muld_a(sum_1, scr_2, scrh_4, n, -0.5_dp, 1.0_dp, tt, rr)

      !  2d)  sum_1 = 1/2 E1W1W2 (+ sum_1)     ( [25]-[32] )

      CALL dgemm("N", "N", n, n, n, -0.5_dp, scr_3, n, scrh_1, n, 0.0_dp, sum_1, n)
      CALL mat_muld_a(sum_1, scr_3, scrh_2, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL mat_mulm_a(sum_1, scr_4, scrh_1, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scr_4, n, scrh_2, n, 1.0_dp, sum_1, n)
      CALL mat_muld_a(sum_1, scr_3, scrh_3, n, 0.5_dp, 1.0_dp, tt, rr)
      CALL mat_muld_a(sum_1, scr_3, scrh_4, n, -0.5_dp, 1.0_dp, tt, rr)
      CALL dgemm("N", "N", n, n, n, -0.5_dp, scr_4, n, scrh_3, n, 1.0_dp, sum_1, n)
      CALL dgemm("N", "N", n, n, n, 0.5_dp, scr_4, n, scrh_4, n, 1.0_dp, sum_1, n)

      !-----------------------------------------------------------------------
      !   3.  sum_2 = 1/8 [W1,[W1,[W1,O1]]] =
      !
      !             = 1/8 ( (W1^3)O1 - 3(W1^2)O1W1 + 3 W1O1(W1^2) - O1(W1^3) )
      !-----------------------------------------------------------------------

      CALL dgemm("N", "N", n, n, n, 0.125_dp, w1w1, n, w1o1, n, 0.0_dp, sum_2, n)
      CALL dgemm("N", "N", n, n, n, -0.375_dp, w1w1, n, o1w1, n, 1.0_dp, sum_2, n)
      CALL dgemm("N", "N", n, n, n, 0.375_dp, w1o1, n, w1w1, n, 1.0_dp, sum_2, n)
      CALL dgemm("N", "N", n, n, n, -0.125_dp, o1w1, n, w1w1, n, 1.0_dp, sum_2, n)

      !-----------------------------------------------------------------------
      !   4.  result = sum_1 + sum_2
      !-----------------------------------------------------------------------

      CALL mat_add(ev4, 1.0_dp, sum_1, 1.0_dp, sum_2, n)

      !-----------------------------------------------------------------------
      !   5. Finish up the stuff!!
      !-----------------------------------------------------------------------

      DEALLOCATE (v, pvp, vh, pvph, w1w1, w1o1, o1w1, sum_1, sum_2)
      DEALLOCATE (scr_1, scr_2, scr_3, scr_4, scrh_1, scrh_2, scrh_3, scrh_4)

!    WRITE (*,*) "CAW:  DKH4 with even4a (Alex)"
!    WRITE (*,*) "JT:   Now available in cp2k"

      RETURN
   END SUBROUTINE even4a_a

   !-----------------------------------------------------------------------
   !                                                                      -
   !     Matrix routines for DKH-procedure                                -
   !     Alexander Wolf                                                   -
   !     modifed: Jens Thar: Mem manager deleted                          -
   !     This file contains the                                           -
   !      following subroutines:                                          -
   !                                 1. mat_1_over_h                      -
   !                                 2. mat_axa                           -
   !                                 3. mat_arxra                         -
   !                                 4. mat_mulm                          -
   !                                 5. mat_muld                          -
   !                                 6. mat_add                           -
   !                                                                      -
   !-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param p ...
!> \param n ...
!> \param e ...
! **************************************************************************************************
   SUBROUTINE mat_1_over_h_a(p, n, e)

      !***********************************************************************
      !                                                                      *
      !   2. SR mat_1_over_h: Transform matrix p into matrix p/(e(i)+e(j))   *
      !                                                                      *
      !   p    in  REAL(:,:) :   matrix p                                    *
      !   e    in  REAL(:)   :   rel. energy (diagonal)                      *
      !   n    in  INTEGER                                                   *
      !                                                                      *
      !***********************************************************************

      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: p
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: e

      INTEGER                                            :: i, j

      DO i = 1, n
         DO j = 1, n
            p(i, j) = p(i, j)/(e(i)+e(j))
         ENDDO
      ENDDO

      RETURN
   END SUBROUTINE mat_1_over_h_a
   !C-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param p ...
!> \param n ...
!> \param a ...
! **************************************************************************************************
   SUBROUTINE mat_axa_a(p, n, a)

      !C***********************************************************************
      !C                                                                      *
      !C   3. SR mat_axa: Transform matrix p into matrix  a*p*a               *
      !C                                                                      *
      !C   p    in  REAL(:,:):   matrix p                                     *
      !C   a    in  REAL(:)  :   A-factors (diagonal)                         *
      !CJT n    in  INTEGER  :   dimension of matrix p                        *
      !C                                                                      *
      !C***********************************************************************

      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: p
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: a

      INTEGER                                            :: i, j

      DO i = 1, n
         DO j = 1, n
            p(i, j) = p(i, j)*a(i)*a(j)
         ENDDO
      ENDDO

      RETURN
   END SUBROUTINE mat_axa_a
   !C-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param p ...
!> \param n ...
!> \param a ...
!> \param r ...
! **************************************************************************************************
   SUBROUTINE mat_arxra_a(p, n, a, r)

      !C***********************************************************************
      !C                                                                      *
      !C   4. SR mat_arxra: Transform matrix p into matrix  a*r*p*r*a         *
      !C                                                                      *
      !C   p    in  REAL(:,:) :   matrix p                                    *
      !C   a    in  REAL(:)   :   A-factors (diagonal)                        *
      !C   r    in  REAL(:)   :   R-factors (diagonal)                        *
      !C   n    in  INTEGER   :   dimension of matrix p                       *
      !C                                                                      *
      !C***********************************************************************

      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: p
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: a, r

      INTEGER                                            :: i, j

      DO i = 1, n
         DO j = 1, n
            p(i, j) = p(i, j)*a(i)*a(j)*r(i)*r(j)
         ENDDO
      ENDDO

      RETURN
   END SUBROUTINE mat_arxra_a
   !C-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param p ...
!> \param q ...
!> \param r ...
!> \param n ...
!> \param alpha ...
!> \param beta ...
!> \param t ...
!> \param rr ...
! **************************************************************************************************
   SUBROUTINE mat_mulm_a(p, q, r, n, alpha, beta, t, rr)

      !C***********************************************************************
      !C                                                                      *
      !C   5. SR mat_mulm:  Multiply matrices according to:                   *
      !C                                                                      *
      !C                      p = alpha*q*(..P^2..)*r + beta*p                *
      !C                                                                      *
      !C   p      out  REAL(:,:):   matrix p                                  *
      !C   q      in   REAL(:,:):   matrix q                                  *
      !C   r      in   REAL(:,.):   matrix r                                  *
      !C   n      in   INTEGER  :   dimension n of matrices                   *
      !C   alpha  in   REAL(dp) :                                             *
      !C   beta   in   REAL(dp) :                                             *
      !C   t      in   REAL(:)  :   non-rel. kinetic energy  (diagonal)       *
      !C   rr     in   REAL(:)  :   R-factors  (diagonal)                     *
      !C                                                                      *
      !C***********************************************************************

      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: p, q, r
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: alpha, beta
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: t, rr

      INTEGER                                            :: i, j
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: qtemp

      ALLOCATE (qtemp(n, n))

      DO i = 1, n
         DO j = 1, n
            qtemp(i, j) = q(i, j)*2.0_dp*t(j)*rr(j)*rr(j)
         ENDDO
      ENDDO

      CALL dgemm("N", "N", n, n, n, alpha, qtemp, n, r, n, beta, p, n)
      RETURN
      DEALLOCATE (qtemp)
   END SUBROUTINE mat_mulm_a
   !C-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param p ...
!> \param q ...
!> \param r ...
!> \param n ...
!> \param alpha ...
!> \param beta ...
!> \param t ...
!> \param rr ...
! **************************************************************************************************
   SUBROUTINE mat_muld_a(p, q, r, n, alpha, beta, t, rr)

      !C***********************************************************************
      !C                                                                      *
      !C   16. SR mat_muld:  Multiply matrices according to:                  *
      !C                                                                      *
      !C                      p = alpha*q*(..1/P^2..)*r + beta*p              *
      !C                                                                      *
      !C   p      out  REAL(:,:):   matrix p                                  *
      !C   q      in   REAL(:,:):   matrix q                                  *
      !C   r      in   REAL(:,:):   matrix r                                  *
      !C   n      in   INTEGER  :   Dimension of all matrices                 *
      !C   alpha  in   REAL(dp) :                                             *
      !C   beta   in   REAL(dp) :                                             *
      !C   t      in   REAL(:)  :   non-rel. kinetic energy  (diagonal)       *
      !C   rr     in   REAL(:)  :   R-factors  (diagonal)                     *
      !C                                                                      *
      !C***********************************************************************

      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: p, q, r
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), INTENT(IN)                          :: alpha, beta
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: t, rr

      INTEGER                                            :: i, j
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: qtemp

      ALLOCATE (qtemp(n, n))

      DO i = 1, n
         DO j = 1, n
            qtemp(i, j) = q(i, j)*0.5_dp/(t(j)*rr(j)*rr(j))
         ENDDO
      ENDDO

      CALL dgemm("N", "N", n, n, n, alpha, qtemp, n, r, n, beta, p, n)

      RETURN
      DEALLOCATE (qtemp)
   END SUBROUTINE mat_muld_a
   !C-----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param p ...
!> \param alpha ...
!> \param beta ...
!> \param r ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE mat_add2(p, alpha, beta, r, n)

      !C***********************************************************************
      !C                                                                      *
      !C   19. SR mat_add:  Add two matrices of the same size according to:   *
      !C                                                                      *
      !C                            p = alpha*p + beta*r                      *
      !C                                                                      *
      !C                    and store them in the first                       *
      !C   p      out  REAL(:,:)  :   matrix p                                *
      !C   r      in   REAL(:,:)  :   matrix r                                *
      !C   alpha  in   REAL(dp)                                               *
      !C   beta   in   REAL(dp)                                               *
      !C                                                                      *
      !C   Matrix p must already exist before calling this SR!!               *
      !C                                                                      *
      !C  [written by: Alexander Wolf,  20.2.2002,  v1.0]                     *
      !C                                                                      *
      !C***********************************************************************

      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: p
      REAL(KIND=dp), INTENT(IN)                          :: alpha, beta
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r
      INTEGER, INTENT(IN)                                :: n

      INTEGER                                            :: i, j

!C  Add matrices:

      DO i = 1, n
         DO j = 1, n
            p(i, j) = alpha*p(i, j)+beta*r(i, j)
         ENDDO
      ENDDO

      RETURN
   END SUBROUTINE mat_add2

   !CAW---------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param p ...
!> \param alpha ...
!> \param q ...
!> \param beta ...
!> \param r ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE mat_add(p, alpha, q, beta, r, n)

      !C***********************************************************************
      !C                                                                      *
      !C   19. SR mat_add:  Add two matrices of the same size according to:   *
      !C                                                                      *
      !C                            p = alpha*q + beta*r                      *
      !C                                                                      *
      !C   p      out  REAL(:,:)  :   matrix p                                *
      !C   q      in   REAL(:,:)  :   matrix q                                *
      !C   r      in   REAL(:,:)  :   matrix r                                *
      !C   alpha  in   REAL(dp)                                               *
      !C   beta   in   REAL(dp)                                               *
      !C                                                                      *
      !C   Matrix p must already exist before calling this SR!!               *
      !C                                                                      *
      !C  [written by: Alexander Wolf,  20.2.2002,  v1.0]                     *
      !C                                                                      *
      !C***********************************************************************

      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: p
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: q
      REAL(KIND=dp), INTENT(IN)                          :: beta
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: r
      INTEGER, INTENT(IN)                                :: n

      INTEGER                                            :: i, j

!C  Add matrices:

      DO i = 1, n
         DO j = 1, n
            p(i, j) = alpha*q(i, j)+beta*r(i, j)
         ENDDO
      ENDDO

      RETURN
   END SUBROUTINE mat_add

   !CAW---------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param W ...
!> \param B ...
!> \param C ...
!> \param N ...
!> \param H ...
! **************************************************************************************************
   SUBROUTINE TRSM(W, B, C, N, H)

      REAL(KIND=dp), DIMENSION(:, :)                     :: W, B, C
      INTEGER                                            :: N
      REAL(KIND=dp), DIMENSION(:, :)                     :: H

      INTEGER                                            :: I, IJ, J, K, L

!C
!C     TRANSFORM SYMMETRIC matrix A by UNITARY TRANSFORMATION
!C     IN B. RESULT IS IN C
!C
!CAW      C = B^{dagger} * A * B

      IJ = 0
      DO I = 1, N
         DO J = 1, I
            IJ = IJ+1
            C(I, J) = 0.0_dp
            C(J, I) = 0.0_dp
            H(I, J) = 0.0_dp
            H(J, I) = 0.0_dp
         END DO
      END DO
      DO I = 1, N
         DO L = 1, N
            DO K = 1, N
               H(I, L) = B(K, I)*W(K, L)+H(I, L)
            END DO
         END DO
      END DO

      IJ = 0
      DO I = 1, N
         DO J = 1, I
            IJ = IJ+1
            DO L = 1, N
               C(I, J) = H(I, L)*B(L, J)+C(I, J)
               C(J, I) = C(I, J)
            END DO
         END DO
      END DO

      RETURN
   END SUBROUTINE TRSM

   !JT----------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param matrix_t_pgf ...
!> \param n ...
!> \param eig ...
!> \param ew ...
!> \param matrix_sinv_pgf ...
!> \param aux ...
!> \param ic ...
! **************************************************************************************************
   SUBROUTINE dkh_diag(matrix_t_pgf, n, eig, ew, matrix_sinv_pgf, aux, ic)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: matrix_t_pgf
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: eig
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: ew
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: matrix_sinv_pgf
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: aux
      INTEGER                                            :: ic

      INTEGER                                            :: n2

      eig = 0.0_dp
      aux = 0.0_dp

      CALL dgemm("N", "N", n, n, n, 1.0_dp, matrix_t_pgf, n, matrix_sinv_pgf, n, 0.0_dp, eig, n)

      aux = 0.0_dp

      CALL dgemm("T", "N", n, n, n, 1.0_dp, matrix_sinv_pgf, n, eig, n, 0.0_dp, aux, n)

      n2 = 3*n-1

      CALL JACOB2(AUX, EIG, EW, N, IC)
      RETURN
   END SUBROUTINE dkh_diag

   !CAW---------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param sogt ...
!> \param eigv ...
!> \param eigw ...
!> \param n ...
!> \param ic ...
! **************************************************************************************************
   SUBROUTINE JACOB2(sogt, eigv, eigw, n, ic)

      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(n), INTENT(OUT)           :: eigw
      REAL(KIND=dp), DIMENSION(n, n), INTENT(OUT)        :: eigv
      REAL(KIND=dp), DIMENSION(n, n), INTENT(INOUT)      :: sogt
      INTEGER, INTENT(IN)                                :: ic

      INTEGER                                            :: i, il, im, ind, j, k, l, ll, m, mm
      REAL(KIND=dp)                                      :: cost, cost2, ext_norm, sincs, sint, &
                                                            sint2, thr, thr_min, tol, u1, x, xy, y

      tol = 1.0E-15
      ext_norm = 0.0_dp
      u1 = REAL(n)
      DO i = 1, n
         eigv(i, i) = 1.0_dp
         eigw(i) = sogt(i, i)
         DO j = 1, i
            IF (i .NE. j) THEN
               eigv(i, j) = 0.0_dp
               eigv(j, i) = 0.0_dp
               ext_norm = ext_norm+sogt(i, j)*sogt(i, j)
            END IF
         END DO
      END DO

      IF (ext_norm .GT. 0.0_dp) THEN
         ext_norm = SQRT(2.0_dp*ext_norm)
         thr_min = ext_norm*tol/u1
         ind = 0
         thr = ext_norm

         DO
            thr = thr/u1
            DO
               l = 1
               DO
                  m = l+1
                  DO
                     IF ((ABS(sogt(m, l))-thr) .GE. 0.0_dp) THEN
                        ind = 1
                        x = 0.5_dp*(eigw(l)-eigw(m))
                        y = -sogt(m, l)/SQRT(sogt(m, l)*sogt(m, l)+x*x)
                        IF (x .LT. 0.0_dp) y = -y

                        IF (y .GT. 1.0_dp) y = 1.0_dp
                        IF (y .LT. -1.0_dp) y = -1.0_dp
                        xy = 1.0_dp-y*y
                        sint = y/SQRT(2.0_dp*(1.0_dp+SQRT(xy)))
                        sint2 = sint*sint
                        cost2 = 1.0_dp-sint2
                        cost = SQRT(cost2)
                        sincs = sint*cost

                        DO i = 1, n
                           IF ((i-m) .NE. 0) THEN
                              IF ((i-m) .LT. 0) THEN
                                 im = m
                                 mm = i
                              ELSE
                                 im = i
                                 mm = m
                              END IF
                              IF ((i-l) .NE. 0) THEN
                                 IF ((i-l) .LT. 0) THEN
                                    il = l
                                    ll = i
                                 ELSE
                                    il = i
                                    ll = l
                                 END IF
                                 x = sogt(il, ll)*cost-sogt(im, mm)*sint
                                 sogt(im, mm) = sogt(il, ll)*sint+sogt(im, mm)*cost
                                 sogt(il, ll) = x
                              END IF
                           END IF

                           x = eigv(i, l)*cost-eigv(i, m)*sint
                           eigv(i, m) = eigv(i, l)*sint+eigv(i, m)*cost
                           eigv(i, l) = x
                        END DO

                        x = 2.0_dp*sogt(m, l)*sincs
                        y = eigw(l)*cost2+eigw(m)*sint2-x
                        x = eigw(l)*sint2+eigw(m)*cost2+x
                        sogt(m, l) = (eigw(l)-eigw(m))*sincs+sogt(m, l)*(cost2-sint2)
                        eigw(l) = y
                        eigw(m) = x
                     END IF
                     IF ((m-n) .EQ. 0) EXIT
                     m = m+1
                  END DO
                  IF ((l-m+1) .EQ. 0) EXIT
                  l = l+1
               END DO
               IF ((ind-1) .NE. 0.0_dp) EXIT
               ind = 0
            END DO
            IF ((thr-thr_min) .LE. 0.0_dp) EXIT
         END DO
      END IF

      IF (ic .NE. 0) THEN
         DO i = 1, n
            DO j = 1, n
               IF ((eigw(i)-eigw(j)) .GT. 0.0_dp) THEN
                  x = eigw(i)
                  eigw(i) = eigw(j)
                  eigw(j) = x
                  DO k = 1, n
                     y = eigv(k, i)
                     eigv(k, i) = eigv(k, j)
                     eigv(k, j) = y
                  END DO
               END IF
            END DO
         END DO

      END IF
      RETURN
   END SUBROUTINE JACOB2

   !JT---------------------------------------------------------------------

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param matrix_s_pgf ...
!> \param matrix_sinv_pgf ...
! **************************************************************************************************
   SUBROUTINE SOG(n, matrix_s_pgf, matrix_sinv_pgf)

      INTEGER                                            :: n
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: matrix_s_pgf
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: matrix_sinv_pgf

      INTEGER                                            :: i, j, jn, k
      REAL(KIND=dp)                                      :: diag_s, row_sum, scalar
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: a
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: g

!
!     SUBROUTINE TO CALCULATE TRANSFORMATION TO SCHMIDT-
!     ORTHOGONALIZED BASIS.
!     sinv-1*matrix_s_pgf*sinv = "orthogonal matrix"
!     n              dimension of matrices
!     matrix_s_pgf   original overlap matrix
!     matrix_sinv_pgf new overlap matrix
!     g              scratch
!     a              scratch
!

      ALLOCATE (a(n))
      ALLOCATE (g(n, n))

      DO jn = 1, n
         diag_s = matrix_s_pgf(jn, jn)
         g(jn, jn) = 1.0_dp

         IF (jn .NE. 1) THEN
            DO j = 1, jn-1
               scalar = 0.0_dp
               DO i = 1, j
                  scalar = scalar+matrix_s_pgf(i, jn)*g(i, j)
               END DO
               diag_s = diag_s-scalar*scalar
               a(j) = scalar
            END DO

            DO j = 1, jn-1
               row_sum = 0.0_dp
               DO k = j, jn-1
                  row_sum = row_sum+a(k)*g(j, k)
               END DO
               g(j, jn) = -row_sum
            END DO
         END IF

         diag_s = 1.0_dp/SQRT(diag_s)
         DO i = 1, jn
            g(i, jn) = g(i, jn)*diag_s
         END DO
      END DO

      DO j = 1, n
         DO i = 1, j
            matrix_sinv_pgf(j, i) = 0.0_dp
            matrix_sinv_pgf(i, j) = g(i, j)
         END DO
      END DO
      DEALLOCATE (a, g)

      RETURN

   END SUBROUTINE SOG

END MODULE dkh_main
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Calculate the atomic operator matrices
!> \author  jgh
!> \date    03.03.2008
!> \version 1.0
!>
! **************************************************************************************************
MODULE atom_operators
   USE ai_onecenter,                    ONLY: &
        sg_coulomb, sg_erf, sg_erfc, sg_exchange, sg_gpot, sg_kinetic, sg_kinnuc, sg_nuclear, &
        sg_overlap, sg_proj_ol, sto_kinetic, sto_nuclear, sto_overlap
   USE atom_types,                      ONLY: &
        atom_basis_gridrep, atom_basis_type, atom_integrals, atom_potential_type, atom_state, &
        cgto_basis, ecp_pseudo, gth_pseudo, gto_basis, no_pseudo, num_basis, release_atom_basis, &
        sto_basis, upf_pseudo
   USE atom_utils,                      ONLY: &
        atom_solve, contract2, contract2add, contract4, coulomb_potential_numeric, integrate_grid, &
        numpot_matrix, slater_density, wigner_slater_functional
   USE dkh_main,                        ONLY: dkh_atom_transformation
   USE erf_fn,                          ONLY: erfc
   USE input_constants,                 ONLY: &
        barrier_conf, do_dkh0_atom, do_dkh1_atom, do_dkh2_atom, do_dkh3_atom, do_dkh4_atom, &
        do_dkh5_atom, do_nonrel_atom, do_sczoramp_atom, do_zoramp_atom, poly_conf
   USE kinds,                           ONLY: dp
   USE lapack,                          ONLY: lapack_sgesv
   USE mathconstants,                   ONLY: gamma1,&
                                              sqrt2
   USE mathlib,                         ONLY: jacobi
   USE periodic_table,                  ONLY: ptable
   USE physcon,                         ONLY: a_fine
   USE qs_grid_atom,                    ONLY: grid_atom_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_operators'

   PUBLIC :: atom_int_setup, atom_ppint_setup, atom_int_release, atom_ppint_release
   PUBLIC :: atom_relint_setup, atom_relint_release
   PUBLIC :: contract2

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param integrals ...
!> \param basis ...
!> \param potential ...
!> \param eri_coulomb ...
!> \param eri_exchange ...
!> \param all_nu ...
! **************************************************************************************************
   SUBROUTINE atom_int_setup(integrals, basis, potential, &
                             eri_coulomb, eri_exchange, all_nu)
      TYPE(atom_integrals), INTENT(INOUT)                :: integrals
      TYPE(atom_basis_type), INTENT(INOUT)               :: basis
      TYPE(atom_potential_type), INTENT(IN), OPTIONAL    :: potential
      LOGICAL, INTENT(IN), OPTIONAL                      :: eri_coulomb, eri_exchange, all_nu

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_int_setup', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ii, info, ipiv(1000), l, l1, &
                                                            l2, ll, lwork, m, m1, m2, mm1, mm2, n, &
                                                            n1, n2, nn1, nn2, nu, nx
      REAL(KIND=dp)                                      :: om, rc, ron, sc, x
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cpot, w, work
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: omat, vmat
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: eri

      CALL timeset(routineN, handle)

      IF (integrals%status == 0) THEN
         n = MAXVAL(basis%nbas)
         integrals%n = basis%nbas

         IF (PRESENT(eri_coulomb)) THEN
            integrals%eri_coulomb = eri_coulomb
         ELSE
            integrals%eri_coulomb = .FALSE.
         END IF
         IF (PRESENT(eri_exchange)) THEN
            integrals%eri_exchange = eri_exchange
         ELSE
            integrals%eri_exchange = .FALSE.
         END IF
         IF (PRESENT(all_nu)) THEN
            integrals%all_nu = all_nu
         ELSE
            integrals%all_nu = .FALSE.
         END IF

         NULLIFY (integrals%ovlp, integrals%kin, integrals%core, integrals%conf)
         DO ll = 1, SIZE(integrals%ceri)
            NULLIFY (integrals%ceri(ll)%int, integrals%eeri(ll)%int)
         END DO

         ALLOCATE (integrals%ovlp(n, n, 0:3))
         integrals%ovlp = 0._dp

         ALLOCATE (integrals%kin(n, n, 0:3))
         integrals%kin = 0._dp

         integrals%status = 1

         IF (PRESENT(potential)) THEN
            IF (potential%confinement) THEN
               ALLOCATE (integrals%conf(n, n, 0:3))
               integrals%conf = 0._dp
               m = basis%grid%nr
               ALLOCATE (cpot(1:m))
               IF (potential%conf_type == poly_conf) THEN
                  rc = potential%rcon
                  sc = potential%scon
                  cpot(1:m) = (basis%grid%rad(1:m)/rc)**sc
               ELSEIF (potential%conf_type == barrier_conf) THEN
                  om = potential%rcon
                  ron = potential%scon
                  rc = ron+om
                  DO i = 1, m
                     IF (basis%grid%rad(i) < ron) THEN
                        cpot(i) = 0.0_dp
                     ELSEIF (basis%grid%rad(i) < rc) THEN
                        x = (basis%grid%rad(i)-ron)/om
                        x = 1._dp-x
                        cpot(i) = -6._dp*x**5+15._dp*x**4-10._dp*x**3+1._dp
                     ELSE
                        cpot(i) = 1.0_dp
                     END IF
                  END DO
               ELSE
                  CALL cp__b("atom_operators.F",140,"")
               END IF
               CALL numpot_matrix(integrals%conf, cpot, basis, 0)
               DEALLOCATE (cpot)
            END IF
         END IF

         SELECT CASE (basis%basis_type)
         CASE DEFAULT
            CALL cp__b("atom_operators.F",149,"")
         CASE (GTO_BASIS)
            DO l = 0, 3
               n = integrals%n(l)
               CALL sg_overlap(integrals%ovlp(1:n, 1:n, l), l, basis%am(1:n, l), basis%am(1:n, l))
               CALL sg_kinetic(integrals%kin(1:n, 1:n, l), l, basis%am(1:n, l), basis%am(1:n, l))
            END DO
            IF (integrals%eri_coulomb) THEN
               ll = 0
               DO l1 = 0, 3
                  n1 = integrals%n(l1)
                  nn1 = (n1*(n1+1))/2
                  DO l2 = 0, l1
                     n2 = integrals%n(l2)
                     nn2 = (n2*(n2+1))/2
                     IF (integrals%all_nu) THEN
                        nx = MIN(2*l1, 2*l2)
                     ELSE
                        nx = 0
                     END IF
                     DO nu = 0, nx, 2
                        ll = ll+1
                        ALLOCATE (integrals%ceri(ll)%int(nn1, nn2))
                        integrals%ceri(ll)%int = 0._dp
                        eri => integrals%ceri(ll)%int
                        CALL sg_coulomb(eri, nu, basis%am(1:n1, l1), l1, basis%am(1:n2, l2), l2)
                     END DO
                  END DO
               END DO
            END IF
            IF (integrals%eri_exchange) THEN
               ll = 0
               DO l1 = 0, 3
                  n1 = integrals%n(l1)
                  nn1 = (n1*(n1+1))/2
                  DO l2 = 0, l1
                     n2 = integrals%n(l2)
                     nn2 = (n2*(n2+1))/2
                     DO nu = ABS(l1-l2), l1+l2, 2
                        ll = ll+1
                        ALLOCATE (integrals%eeri(ll)%int(nn1, nn2))
                        integrals%eeri(ll)%int = 0._dp
                        eri => integrals%eeri(ll)%int
                        CALL sg_exchange(eri, nu, basis%am(1:n1, l1), l1, basis%am(1:n2, l2), l2)
                     END DO
                  END DO
               END DO
            END IF
         CASE (CGTO_BASIS)
            DO l = 0, 3
               n = integrals%n(l)
               m = basis%nprim(l)
               IF (n > 0 .AND. m > 0) THEN
                  ALLOCATE (omat(m, m))

                  CALL sg_overlap(omat(1:m, 1:m), l, basis%am(1:m, l), basis%am(1:m, l))
                  CALL contract2(integrals%ovlp(1:n, 1:n, l), omat(1:m, 1:m), basis%cm(1:m, 1:n, l))
                  CALL sg_kinetic(omat(1:m, 1:m), l, basis%am(1:m, l), basis%am(1:m, l))
                  CALL contract2(integrals%kin(1:n, 1:n, l), omat(1:m, 1:m), basis%cm(1:m, 1:n, l))
                  DEALLOCATE (omat)
               END IF
            END DO
            IF (integrals%eri_coulomb) THEN
               ll = 0
               DO l1 = 0, 3
                  n1 = integrals%n(l1)
                  nn1 = (n1*(n1+1))/2
                  m1 = basis%nprim(l1)
                  mm1 = (m1*(m1+1))/2
                  DO l2 = 0, l1
                     n2 = integrals%n(l2)
                     nn2 = (n2*(n2+1))/2
                     m2 = basis%nprim(l2)
                     mm2 = (m2*(m2+1))/2
                     IF (integrals%all_nu) THEN
                        nx = MIN(2*l1, 2*l2)
                     ELSE
                        nx = 0
                     END IF
                     DO nu = 0, nx, 2
                        ll = ll+1
                        ALLOCATE (integrals%ceri(ll)%int(nn1, nn2))
                        integrals%ceri(ll)%int = 0._dp
                        ALLOCATE (omat(mm1, mm2))

                        eri => integrals%ceri(ll)%int
                        CALL sg_coulomb(omat, nu, basis%am(1:m1, l1), l1, basis%am(1:m2, l2), l2)
                        CALL contract4(eri, omat, basis%cm(1:m1, 1:n1, l1), basis%cm(1:m2, 1:n2, l2))

                        DEALLOCATE (omat)
                     END DO
                  END DO
               END DO
            END IF
            IF (integrals%eri_exchange) THEN
               ll = 0
               DO l1 = 0, 3
                  n1 = integrals%n(l1)
                  nn1 = (n1*(n1+1))/2
                  m1 = basis%nprim(l1)
                  mm1 = (m1*(m1+1))/2
                  DO l2 = 0, l1
                     n2 = integrals%n(l2)
                     nn2 = (n2*(n2+1))/2
                     m2 = basis%nprim(l2)
                     mm2 = (m2*(m2+1))/2
                     DO nu = ABS(l1-l2), l1+l2, 2
                        ll = ll+1
                        ALLOCATE (integrals%eeri(ll)%int(nn1, nn2))
                        integrals%eeri(ll)%int = 0._dp
                        ALLOCATE (omat(mm1, mm2))

                        eri => integrals%eeri(ll)%int
                        CALL sg_exchange(omat, nu, basis%am(1:m1, l1), l1, basis%am(1:m2, l2), l2)
                        CALL contract4(eri, omat, basis%cm(1:m1, 1:n1, l1), basis%cm(1:m2, 1:n2, l2))

                        DEALLOCATE (omat)
                     END DO
                  END DO
               END DO
            END IF
         CASE (STO_BASIS)
            DO l = 0, 3
               n = integrals%n(l)
               CALL sto_overlap(integrals%ovlp(1:n, 1:n, l), basis%ns(1:n, l), basis%as(1:n, l), &
                                basis%ns(1:n, l), basis%as(1:n, l))
               CALL sto_kinetic(integrals%kin(1:n, 1:n, l), l, basis%ns(1:n, l), basis%as(1:n, l), &
                                basis%ns(1:n, l), basis%as(1:n, l))
            END DO
            IF(.NOT.(.NOT. integrals%eri_coulomb))CALL cp__a("atom_operators.F",278)
            IF(.NOT.(.NOT. integrals%eri_exchange))CALL cp__a("atom_operators.F",279)
         CASE (NUM_BASIS)
            CALL cp__b("atom_operators.F",281,"")
         END SELECT

         ! setup transformation matrix to get an orthogonal basis, remove linear dependencies
         NULLIFY (integrals%utrans, integrals%uptrans)
         n = MAXVAL(basis%nbas)
         ALLOCATE (integrals%utrans(n, n, 0:3), integrals%uptrans(n, n, 0:3))
         integrals%utrans = 0._dp
         integrals%uptrans = 0._dp
         integrals%nne = integrals%n
         lwork = 10*n
         ALLOCATE (omat(n, n), w(n), vmat(n, n), work(lwork))
         DO l = 0, 3
            n = integrals%n(l)
            IF (n > 0) THEN
               omat(1:n, 1:n) = integrals%ovlp(1:n, 1:n, l)
               CALL jacobi(omat(1:n, 1:n), w(1:n), vmat(1:n, 1:n))
               omat(1:n, 1:n) = vmat(1:n, 1:n)
               ii = 0
               DO i = 1, n
                  IF (w(i) > basis%eps_eig) THEN
                     ii = ii+1
                     integrals%utrans(1:n, ii, l) = omat(1:n, i)/SQRT(w(i))
                  END IF
               END DO
               integrals%nne(l) = ii
               omat(1:ii, 1:ii) = MATMUL(TRANSPOSE(integrals%utrans(1:n, 1:ii, l)), integrals%utrans(1:n, 1:ii, l))
               DO i = 1, ii
                  integrals%uptrans(i, i, l) = 1._dp
               ENDDO
               CALL lapack_sgesv(ii, ii, omat(1:ii, 1:ii), ii, ipiv, integrals%uptrans(1:ii, 1:ii, l), ii, info)
               IF(.NOT.(info == 0))CALL cp__a("atom_operators.F",312)
            END IF
         END DO
         DEALLOCATE (omat, vmat, w, work)

      END IF

      CALL timestop(handle)

   END SUBROUTINE atom_int_setup
! **************************************************************************************************
!> \brief ...
!> \param integrals ...
!> \param basis ...
!> \param potential ...
! **************************************************************************************************
   SUBROUTINE atom_ppint_setup(integrals, basis, potential)
      TYPE(atom_integrals), INTENT(INOUT)                :: integrals
      TYPE(atom_basis_type), INTENT(INOUT)               :: basis
      TYPE(atom_potential_type), INTENT(IN)              :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_ppint_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ii, j, k, l, m, n
      REAL(KIND=dp)                                      :: al, alpha, rc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cpot, xmat
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: omat, spmat
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rad

      CALL timeset(routineN, handle)

      IF (integrals%ppstat == 0) THEN
         n = MAXVAL(basis%nbas)
         integrals%n = basis%nbas

         NULLIFY (integrals%core, integrals%hnl)

         ALLOCATE (integrals%hnl(n, n, 0:3))
         integrals%hnl = 0._dp

         ALLOCATE (integrals%core(n, n, 0:3))
         integrals%core = 0._dp

         ALLOCATE (integrals%clsd(n, n, 0:3))
         integrals%clsd = 0._dp

         integrals%ppstat = 1

         SELECT CASE (basis%basis_type)
         CASE DEFAULT
            CALL cp__b("atom_operators.F",363,"")
         CASE (GTO_BASIS)

            SELECT CASE (potential%ppot_type)
            CASE (no_pseudo, ecp_pseudo)
               DO l = 0, 3
                  n = integrals%n(l)
                  CALL sg_nuclear(integrals%core(1:n, 1:n, l), l, basis%am(1:n, l), basis%am(1:n, l))
               END DO
            CASE (gth_pseudo)
               alpha = 1._dp/potential%gth_pot%rc/SQRT(2._dp)
               DO l = 0, 3
                  n = integrals%n(l)
                  ALLOCATE (omat(n, n), spmat(n, 5))

                  omat = 0._dp
                  CALL sg_erf(omat(1:n, 1:n), l, alpha, basis%am(1:n, l), basis%am(1:n, l))
                  integrals%core(1:n, 1:n, l) = -potential%gth_pot%zion*omat(1:n, 1:n)
                  DO i = 1, potential%gth_pot%ncl
                     omat = 0._dp
                     CALL sg_gpot(omat(1:n, 1:n), i-1, potential%gth_pot%rc, l, basis%am(1:n, l), basis%am(1:n, l))
                     integrals%core(1:n, 1:n, l) = integrals%core(1:n, 1:n, l)+ &
                                                   potential%gth_pot%cl(i)*omat(1:n, 1:n)
                  END DO
                  IF (potential%gth_pot%lpotextended) THEN
                     DO k = 1, potential%gth_pot%nexp_lpot
                        DO i = 1, potential%gth_pot%nct_lpot(k)
                           omat = 0._dp
                           CALL sg_gpot(omat(1:n, 1:n), i-1, potential%gth_pot%alpha_lpot(k), l, &
                                        basis%am(1:n, l), basis%am(1:n, l))
                           integrals%core(1:n, 1:n, l) = integrals%core(1:n, 1:n, l)+ &
                                                         potential%gth_pot%cval_lpot(i, k)*omat(1:n, 1:n)
                        END DO
                     END DO
                  END IF
                  IF (potential%gth_pot%lsdpot) THEN
                     DO k = 1, potential%gth_pot%nexp_lsd
                        DO i = 1, potential%gth_pot%nct_lsd(k)
                           omat = 0._dp
                           CALL sg_gpot(omat(1:n, 1:n), i-1, potential%gth_pot%alpha_lsd(k), l, &
                                        basis%am(1:n, l), basis%am(1:n, l))
                           integrals%clsd(1:n, 1:n, l) = integrals%clsd(1:n, 1:n, l)+ &
                                                         potential%gth_pot%cval_lsd(i, k)*omat(1:n, 1:n)
                        END DO
                     END DO
                  END IF

                  spmat = 0._dp
                  m = potential%gth_pot%nl(l)
                  DO i = 1, m
                     CALL sg_proj_ol(spmat(1:n, i), l, basis%am(1:n, l), i-1, potential%gth_pot%rcnl(l))
                  END DO
                  integrals%hnl(1:n, 1:n, l) = MATMUL(spmat(1:n, 1:m), &
                                                      MATMUL(potential%gth_pot%hnl(1:m, 1:m, l), TRANSPOSE(spmat(1:n, 1:m))))

                  DEALLOCATE (omat, spmat)
               END DO
            CASE (upf_pseudo)
               CALL upfint_setup(integrals, basis, potential)
            CASE DEFAULT
               CALL cp__b("atom_operators.F",423,"")
            END SELECT

         CASE (CGTO_BASIS)

            SELECT CASE (potential%ppot_type)
            CASE (no_pseudo, ecp_pseudo)
               DO l = 0, 3
                  n = integrals%n(l)
                  m = basis%nprim(l)
                  ALLOCATE (omat(m, m))

                  CALL sg_nuclear(omat(1:m, 1:m), l, basis%am(1:m, l), basis%am(1:m, l))
                  CALL contract2(integrals%core(1:n, 1:n, l), omat(1:m, 1:m), basis%cm(1:m, 1:n, l))

                  DEALLOCATE (omat)
               END DO
            CASE (gth_pseudo)
               alpha = 1._dp/potential%gth_pot%rc/SQRT(2._dp)
               DO l = 0, 3
                  n = integrals%n(l)
                  m = basis%nprim(l)
                  IF (n > 0 .AND. m > 0) THEN
                     ALLOCATE (omat(m, m), spmat(n, 5), xmat(m))

                     omat = 0._dp
                     CALL sg_erf(omat(1:m, 1:m), l, alpha, basis%am(1:m, l), basis%am(1:m, l))
                     omat(1:m, 1:m) = -potential%gth_pot%zion*omat(1:m, 1:m)
                     CALL contract2(integrals%core(1:n, 1:n, l), omat(1:m, 1:m), basis%cm(1:m, 1:n, l))
                     DO i = 1, potential%gth_pot%ncl
                        omat = 0._dp
                        CALL sg_gpot(omat(1:m, 1:m), i-1, potential%gth_pot%rc, l, basis%am(1:m, l), basis%am(1:m, l))
                        omat(1:m, 1:m) = potential%gth_pot%cl(i)*omat(1:m, 1:m)
                        CALL contract2add(integrals%core(1:n, 1:n, l), omat(1:m, 1:m), basis%cm(1:m, 1:n, l))
                     END DO
                     IF (potential%gth_pot%lpotextended) THEN
                        DO k = 1, potential%gth_pot%nexp_lpot
                           DO i = 1, potential%gth_pot%nct_lpot(k)
                              omat = 0._dp
                              CALL sg_gpot(omat(1:m, 1:m), i-1, potential%gth_pot%alpha_lpot(k), l, &
                                           basis%am(1:m, l), basis%am(1:m, l))
                              omat(1:m, 1:m) = potential%gth_pot%cval_lpot(i, k)*omat(1:m, 1:m)
                              CALL contract2add(integrals%core(1:n, 1:n, l), omat(1:m, 1:m), basis%cm(1:m, 1:n, l))
                           END DO
                        END DO
                     END IF
                     IF (potential%gth_pot%lsdpot) THEN
                        DO k = 1, potential%gth_pot%nexp_lsd
                           DO i = 1, potential%gth_pot%nct_lsd(k)
                              omat = 0._dp
                              CALL sg_gpot(omat(1:m, 1:m), i-1, potential%gth_pot%alpha_lsd(k), l, &
                                           basis%am(1:m, l), basis%am(1:m, l))
                              omat(1:m, 1:m) = potential%gth_pot%cval_lsd(i, k)*omat(1:m, 1:m)
                              CALL contract2add(integrals%clsd(1:n, 1:n, l), omat(1:m, 1:m), basis%cm(1:m, 1:n, l))
                           END DO
                        END DO
                     END IF

                     spmat = 0._dp
                     k = potential%gth_pot%nl(l)
                     DO i = 1, k
                        CALL sg_proj_ol(xmat(1:m), l, basis%am(1:m, l), i-1, potential%gth_pot%rcnl(l))
                        spmat(1:n, i) = MATMUL(TRANSPOSE(basis%cm(1:m, 1:n, l)), xmat(1:m))
                     END DO
                     IF (k > 0) THEN
                        integrals%hnl(1:n, 1:n, l) = MATMUL(spmat(1:n, 1:k), &
                                                            MATMUL(potential%gth_pot%hnl(1:k, 1:k, l), TRANSPOSE(spmat(1:n, 1:k))))
                     END IF

                     DEALLOCATE (omat, spmat, xmat)
                  END IF
               END DO
            CASE (upf_pseudo)
               CALL upfint_setup(integrals, basis, potential)
            CASE DEFAULT
               CALL cp__b("atom_operators.F",498,"")
            END SELECT

         CASE (STO_BASIS)

            SELECT CASE (potential%ppot_type)
            CASE (no_pseudo, ecp_pseudo)
               DO l = 0, 3
                  n = integrals%n(l)
                  CALL sto_nuclear(integrals%core(1:n, 1:n, l), basis%ns(1:n, l), basis%as(1:n, l), &
                                   basis%ns(1:n, l), basis%as(1:n, l))
               END DO
            CASE (gth_pseudo)
               rad => basis%grid%rad
               m = basis%grid%nr
               ALLOCATE (cpot(1:m))
               rc = potential%gth_pot%rc
               alpha = 1._dp/rc/SQRT(2._dp)

               ! local pseudopotential, we use erf = 1/r - erfc
               integrals%core = 0._dp
               DO i = 1, m
                  cpot(i) = potential%gth_pot%zion*erfc(alpha*rad(i))/rad(i)
               END DO
               DO i = 1, potential%gth_pot%ncl
                  ii = 2*(i-1)
                  cpot(1:m) = cpot(1:m)+potential%gth_pot%cl(i)*(rad/rc)**ii*EXP(-0.5_dp*(rad/rc)**2)
               END DO
               IF (potential%gth_pot%lpotextended) THEN
                  DO k = 1, potential%gth_pot%nexp_lpot
                     al = potential%gth_pot%alpha_lpot(k)
                     DO i = 1, potential%gth_pot%nct_lpot(k)
                        ii = 2*(i-1)
                        cpot(1:m) = cpot(1:m)+potential%gth_pot%cval_lpot(i, k)*(rad/al)**ii*EXP(-0.5_dp*(rad/al)**2)
                     END DO
                  END DO
               END IF
               CALL numpot_matrix(integrals%core, cpot, basis, 0)
               DO l = 0, 3
                  n = integrals%n(l)
                  ALLOCATE (omat(n, n))
                  omat = 0._dp
                  CALL sto_nuclear(omat(1:n, 1:n), basis%ns(1:n, l), basis%as(1:n, l), &
                                   basis%ns(1:n, l), basis%as(1:n, l))
                  integrals%core(1:n, 1:n, l) = integrals%core(1:n, 1:n, l)-potential%gth_pot%zion*omat(1:n, 1:n)
                  DEALLOCATE (omat)
               END DO

               IF (potential%gth_pot%lsdpot) THEN
                  cpot = 0._dp
                  DO k = 1, potential%gth_pot%nexp_lsd
                     al = potential%gth_pot%alpha_lsd(k)
                     DO i = 1, potential%gth_pot%nct_lsd(k)
                        ii = 2*(i-1)
                        cpot(:) = cpot+potential%gth_pot%cval_lsd(i, k)*(rad/al)**ii*EXP(-0.5_dp*(rad/al)**2)
                     END DO
                  END DO
                  CALL numpot_matrix(integrals%clsd, cpot, basis, 0)
               END IF

               DO l = 0, 3
                  n = integrals%n(l)
                  ! non local pseudopotential
                  ALLOCATE (spmat(n, 5))
                  spmat = 0._dp
                  k = potential%gth_pot%nl(l)
                  DO i = 1, k
                     rc = potential%gth_pot%rcnl(l)
                     cpot(:) = sqrt2/SQRT(gamma1(l+2*i-1))*rad**(l+2*i-2)*EXP(-0.5_dp*(rad/rc)**2)/rc**(l+2*i-0.5_dp)
                     DO j = 1, basis%nbas(l)
                        spmat(j, i) = integrate_grid(cpot, basis%bf(:, j, l), basis%grid)
                     END DO
                  END DO
                  integrals%hnl(1:n, 1:n, l) = MATMUL(spmat(1:n, 1:k), &
                                                      MATMUL(potential%gth_pot%hnl(1:k, 1:k, l), TRANSPOSE(spmat(1:n, 1:k))))

                  DEALLOCATE (spmat)

               END DO

               DEALLOCATE (cpot)

            CASE (upf_pseudo)
               CALL upfint_setup(integrals, basis, potential)
            CASE DEFAULT
               CALL cp__b("atom_operators.F",583,"")
            END SELECT

         CASE (NUM_BASIS)
            CALL cp__b("atom_operators.F",587,"")
         END SELECT

         ! add ecp_pseudo using numerical representation of basis
         IF (potential%ppot_type == ecp_pseudo) THEN
            ! scale 1/r potential
            integrals%core = -potential%ecp_pot%zion*integrals%core
            ! local potential
            m = basis%grid%nr
            rad => basis%grid%rad
            ALLOCATE (cpot(1:m))
            cpot = 0._dp
            DO k = 1, potential%ecp_pot%nloc
               n = potential%ecp_pot%nrloc(k)
               alpha = potential%ecp_pot%bloc(k)
               cpot(:) = cpot+potential%ecp_pot%aloc(k)*rad**(n-2)*EXP(-alpha*rad**2)
            END DO
            CALL numpot_matrix(integrals%core, cpot, basis, 0)
            ! non local pseudopotential
            DO l = 0, MIN(potential%ecp_pot%lmax, 3)
               cpot = 0._dp
               DO k = 1, potential%ecp_pot%npot(l)
                  n = potential%ecp_pot%nrpot(k, l)
                  alpha = potential%ecp_pot%bpot(k, l)
                  cpot(:) = cpot+potential%ecp_pot%apot(k, l)*rad**(n-2)*EXP(-alpha*rad**2)
               END DO
               DO i = 1, basis%nbas(l)
                  DO j = i, basis%nbas(l)
                     integrals%hnl(i, j, l) = integrate_grid(cpot, basis%bf(:, i, l), basis%bf(:, j, l), basis%grid)
                     integrals%hnl(j, i, l) = integrals%hnl(i, j, l)
                  END DO
               END DO
            END DO
            DEALLOCATE (cpot)
         END IF

      END IF

      CALL timestop(handle)

   END SUBROUTINE atom_ppint_setup
! **************************************************************************************************
!> \brief ...
!> \param integrals ...
!> \param basis ...
!> \param potential ...
! **************************************************************************************************
   SUBROUTINE upfint_setup(integrals, basis, potential)
      TYPE(atom_integrals), INTENT(INOUT)                :: integrals
      TYPE(atom_basis_type), INTENT(INOUT)               :: basis
      TYPE(atom_potential_type), INTENT(IN)              :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'upfint_setup', routineP = moduleN//':'//routineN

      CHARACTER(len=4)                                   :: ptype
      INTEGER                                            :: i, j, k1, k2, la, lb, m, n
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: spmat
      TYPE(atom_basis_type)                              :: gbasis

! get basis representation on UPF grid

      CALL atom_basis_gridrep(basis, gbasis, potential%upf_pot%r, potential%upf_pot%rab)

      ! local pseudopotential
      integrals%core = 0._dp
      CALL numpot_matrix(integrals%core, potential%upf_pot%vlocal, gbasis, 0)

      ptype = ADJUSTL(TRIM(potential%upf_pot%pseudo_type))
      IF (ptype(1:2) == "NC") THEN
         ! non local pseudopotential
         n = MAXVAL(integrals%n(:))
         m = potential%upf_pot%number_of_proj
         ALLOCATE (spmat(n, m))
         spmat = 0.0_dp
         DO i = 1, m
            la = potential%upf_pot%lbeta(i)
            DO j = 1, gbasis%nbas(la)
               spmat(j, i) = integrate_grid(potential%upf_pot%beta(:, i), gbasis%bf(:, j, la), gbasis%grid)
            END DO
         END DO
         DO i = 1, m
            la = potential%upf_pot%lbeta(i)
            DO j = 1, m
               lb = potential%upf_pot%lbeta(j)
               IF (la == lb) THEN
                  DO k1 = 1, gbasis%nbas(la)
                     DO k2 = 1, gbasis%nbas(la)
                        integrals%hnl(k1, k2, la) = integrals%hnl(k1, k2, la)+ &
                                                    spmat(k1, i)*potential%upf_pot%dion(i, j)*spmat(k2, j)
                     END DO
                  END DO
               END IF
            END DO
         END DO
         DEALLOCATE (spmat)
      ELSE IF (ptype(1:2) == "SL") THEN
         DO la = 0, potential%upf_pot%l_max
            CALL numpot_matrix(integrals%core, potential%upf_pot%vsemi(:, la+1), gbasis, 0)
         END DO
      ELSE
         CALL cp__b("atom_operators.F",687,"")
      END IF

      ! release basis representation on UPF grid
      CALL release_atom_basis(gbasis)

   END SUBROUTINE upfint_setup
! **************************************************************************************************
!> \brief ...
!> \param integrals ...
!> \param basis ...
!> \param reltyp ...
!> \param zcore ...
!> \param alpha ...
! **************************************************************************************************
   SUBROUTINE atom_relint_setup(integrals, basis, reltyp, zcore, alpha)
      TYPE(atom_integrals), INTENT(INOUT)                :: integrals
      TYPE(atom_basis_type), INTENT(INOUT)               :: basis
      INTEGER, INTENT(IN)                                :: reltyp
      REAL(dp), INTENT(IN)                               :: zcore
      REAL(dp), INTENT(IN), OPTIONAL                     :: alpha

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_relint_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dkhorder, handle, i, k1, k2, l, m, n, nl
      REAL(dp)                                           :: ascal, cspeed
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: cpot
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: modpot
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: ener, sps
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: hmat, pvp, sp, tp, vp, wfn

      CALL timeset(routineN, handle)

      cspeed = 1._dp/a_fine

      SELECT CASE (reltyp)
      CASE DEFAULT
         CALL cp__b("atom_operators.F",725,"")
      CASE (do_nonrel_atom, do_zoramp_atom, do_sczoramp_atom)
         dkhorder = -1
      CASE (do_dkh0_atom)
         dkhorder = 0
      CASE (do_dkh1_atom)
         dkhorder = 1
      CASE (do_dkh2_atom)
         dkhorder = 2
      CASE (do_dkh3_atom)
         dkhorder = 3
      CASE (do_dkh4_atom)
         dkhorder = 4
      CASE (do_dkh5_atom)
         dkhorder = 5
      END SELECT

      SELECT CASE (reltyp)
      CASE DEFAULT
         CALL cp__b("atom_operators.F",744,"")
      CASE (do_nonrel_atom)
         ! nothing to do
         NULLIFY (integrals%tzora, integrals%hdkh)
      CASE (do_zoramp_atom, do_sczoramp_atom)

         NULLIFY (integrals%hdkh)

         IF (integrals%zorastat == 0) THEN
            n = MAXVAL(basis%nbas)
            ALLOCATE (integrals%tzora(n, n, 0:3))
            integrals%tzora = 0._dp
            m = basis%grid%nr
            ALLOCATE (modpot(1:m), cpot(1:m))
            CALL calculate_model_potential(modpot, basis%grid, zcore)
            ! Zora potential
            cpot(1:m) = modpot(1:m)/(4._dp*cspeed*cspeed-2._dp*modpot(1:m))
            cpot(1:m) = cpot(1:m)/basis%grid%rad2(1:m)
            CALL numpot_matrix(integrals%tzora, cpot, basis, 0)
            DO l = 0, 3
               nl = basis%nbas(l)
               integrals%tzora(1:nl, 1:nl, l) = REAL(l*(l+1), dp)*integrals%tzora(1:nl, 1:nl, l)
            END DO
            cpot(1:m) = cpot(1:m)*basis%grid%rad2(1:m)
            CALL numpot_matrix(integrals%tzora, cpot, basis, 2)
            !
            ! scaled ZORA
            IF (reltyp == do_sczoramp_atom) THEN
               ALLOCATE (hmat(n, n, 0:3), wfn(n, n, 0:3), ener(n, 0:3), pvp(n, n, 0:3), sps(n, n))
               hmat(:, :, :) = integrals%kin+integrals%tzora
               ! model potential
               CALL numpot_matrix(hmat, modpot, basis, 0)
               ! eigenvalues and eigenvectors
               CALL atom_solve(hmat, integrals%utrans, wfn, ener, basis%nbas, integrals%nne, 3)
               ! relativistic kinetic energy
               cpot(1:m) = cspeed*cspeed/(2._dp*cspeed*cspeed-modpot(1:m))**2
               cpot(1:m) = cpot(1:m)/basis%grid%rad2(1:m)
               pvp = 0.0_dp
               CALL numpot_matrix(pvp, cpot, basis, 0)
               DO l = 0, 3
                  nl = basis%nbas(l)
                  pvp(1:nl, 1:nl, l) = REAL(l*(l+1), dp)*pvp(1:nl, 1:nl, l)
               END DO
               cpot(1:m) = cpot(1:m)*basis%grid%rad2(1:m)
               CALL numpot_matrix(pvp, cpot, basis, 2)
               ! calculate psi*pvp*psi and the scaled orbital energies
               ! actually, we directly calculate the energy difference
               DO l = 0, 3
                  nl = basis%nbas(l)
                  DO i = 1, integrals%nne(l)
                     IF (ener(i, l) < 0._dp) THEN
                        ascal = SUM(wfn(1:nl, i, l)*MATMUL(pvp(1:nl, 1:nl, l), wfn(1:nl, i, l)))
                        ener(i, l) = ener(i, l)*ascal/(1.0_dp+ascal)
                     ELSE
                        ener(i, l) = 0.0_dp
                     END IF
                  END DO
               END DO
               ! correction term is calculated as a projector
               hmat = 0.0_dp
               DO l = 0, 3
                  nl = basis%nbas(l)
                  DO i = 1, integrals%nne(l)
                     DO k1 = 1, nl
                        DO k2 = 1, nl
                           hmat(k1, k2, l) = hmat(k1, k2, l)+ener(i, l)*wfn(k1, i, l)*wfn(k2, i, l)
                        END DO
                     END DO
                  END DO
                  ! transform with overlap matrix
                  sps(1:nl, 1:nl) = MATMUL(integrals%ovlp(1:nl, 1:nl, l), &
                                           MATMUL(hmat(1:nl, 1:nl, l), integrals%ovlp(1:nl, 1:nl, l)))
                  ! add scaling correction to tzora
                  integrals%tzora(1:nl, 1:nl, l) = integrals%tzora(1:nl, 1:nl, l)-sps(1:nl, 1:nl)
               END DO

               DEALLOCATE (hmat, wfn, ener, pvp, sps)
            END IF
            !
            DEALLOCATE (modpot, cpot)

            integrals%zorastat = 1

         END IF

      CASE (do_dkh0_atom, do_dkh1_atom, do_dkh2_atom, do_dkh3_atom, do_dkh4_atom, do_dkh5_atom)

         NULLIFY (integrals%tzora)

         IF (integrals%dkhstat == 0) THEN
            n = MAXVAL(basis%nbas)
            ALLOCATE (integrals%hdkh(n, n, 0:3))
            integrals%hdkh = 0._dp

            m = MAXVAL(basis%nprim)
            ALLOCATE (tp(m, m, 0:3), sp(m, m, 0:3), vp(m, m, 0:3), pvp(m, m, 0:3))
            tp = 0._dp
            sp = 0._dp
            vp = 0._dp
            pvp = 0._dp

            SELECT CASE (basis%basis_type)
            CASE DEFAULT
               CALL cp__b("atom_operators.F",847,"")
            CASE (GTO_BASIS, CGTO_BASIS)

               DO l = 0, 3
                  m = basis%nprim(l)
                  IF (m > 0) THEN
                     CALL sg_kinetic(tp(1:m, 1:m, l), l, basis%am(1:m, l), basis%am(1:m, l))
                     CALL sg_overlap(sp(1:m, 1:m, l), l, basis%am(1:m, l), basis%am(1:m, l))
                     IF (PRESENT(alpha)) THEN
                        CALL sg_erfc(vp(1:m, 1:m, l), l, alpha, basis%am(1:m, l), basis%am(1:m, l))
                     ELSE
                        CALL sg_nuclear(vp(1:m, 1:m, l), l, basis%am(1:m, l), basis%am(1:m, l))
                     END IF
                     CALL sg_kinnuc(pvp(1:m, 1:m, l), l, basis%am(1:m, l), basis%am(1:m, l))
                     vp(1:m, 1:m, l) = -zcore*vp(1:m, 1:m, l)
                     pvp(1:m, 1:m, l) = -zcore*pvp(1:m, 1:m, l)
                  END IF
               END DO

            CASE (STO_BASIS)
               CALL cp__b("atom_operators.F",867,"")
            CASE (NUM_BASIS)
               CALL cp__b("atom_operators.F",869,"")
            END SELECT

            CALL dkh_integrals(integrals, basis, dkhorder, sp, tp, vp, pvp)

            integrals%dkhstat = 1

            DEALLOCATE (tp, sp, vp, pvp)

         ELSE
            IF(.NOT.(ASSOCIATED(integrals%hdkh)))CALL cp__a("atom_operators.F",879)
         END IF

      END SELECT

      CALL timestop(handle)

   END SUBROUTINE atom_relint_setup
! **************************************************************************************************
!> \brief ...
!> \param integrals ...
!> \param basis ...
!> \param order ...
!> \param sp ...
!> \param tp ...
!> \param vp ...
!> \param pvp ...
! **************************************************************************************************
   SUBROUTINE dkh_integrals(integrals, basis, order, sp, tp, vp, pvp)
      TYPE(atom_integrals), INTENT(INOUT)                :: integrals
      TYPE(atom_basis_type), INTENT(INOUT)               :: basis
      INTEGER, INTENT(IN)                                :: order
      REAL(dp), DIMENSION(:, :, 0:)                      :: sp, tp, vp, pvp

      CHARACTER(len=*), PARAMETER :: routineN = 'dkh_integrals', routineP = moduleN//':'//routineN

      INTEGER                                            :: l, m, n
      REAL(dp), DIMENSION(:, :, :), POINTER              :: hdkh

      IF(.NOT.(order >= 0))CALL cp__a("atom_operators.F",908)

      hdkh => integrals%hdkh

      DO l = 0, 3
         n = integrals%n(l)
         m = basis%nprim(l)
         IF (n > 0) THEN
            CALL dkh_atom_transformation(sp(1:m, 1:m, l), vp(1:m, 1:m, l), tp(1:m, 1:m, l), pvp(1:m, 1:m, l), m, order)
            SELECT CASE (basis%basis_type)
            CASE DEFAULT
               CALL cp__b("atom_operators.F",919,"")
            CASE (GTO_BASIS)
               IF(.NOT.(n == m))CALL cp__a("atom_operators.F",921)
               integrals%hdkh(1:n, 1:n, l) = tp(1:n, 1:n, l)+vp(1:n, 1:n, l)
            CASE (CGTO_BASIS)
               CALL contract2(integrals%hdkh(1:n, 1:n, l), tp(1:m, 1:m, l), basis%cm(1:m, 1:n, l))
               CALL contract2add(integrals%hdkh(1:n, 1:n, l), vp(1:m, 1:m, l), basis%cm(1:m, 1:n, l))
            CASE (STO_BASIS)
               CALL cp__b("atom_operators.F",927,"")
            CASE (NUM_BASIS)
               CALL cp__b("atom_operators.F",929,"")
            END SELECT
         ELSE
            integrals%hdkh(1:n, 1:n, l) = 0._dp
         END IF
      END DO

   END SUBROUTINE dkh_integrals
! **************************************************************************************************
!> \brief ...
!> \param integrals ...
! **************************************************************************************************
   SUBROUTINE atom_int_release(integrals)
      TYPE(atom_integrals), INTENT(INOUT)                :: integrals

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_int_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ll

      IF (ASSOCIATED(integrals%ovlp)) THEN
         DEALLOCATE (integrals%ovlp)
      END IF
      IF (ASSOCIATED(integrals%kin)) THEN
         DEALLOCATE (integrals%kin)
      END IF
      IF (ASSOCIATED(integrals%conf)) THEN
         DEALLOCATE (integrals%conf)
      END IF
      DO ll = 1, SIZE(integrals%ceri)
         IF (ASSOCIATED(integrals%ceri(ll)%int)) THEN
            DEALLOCATE (integrals%ceri(ll)%int)
         END IF
         IF (ASSOCIATED(integrals%eeri(ll)%int)) THEN
            DEALLOCATE (integrals%eeri(ll)%int)
         END IF
      END DO
      IF (ASSOCIATED(integrals%utrans)) THEN
         DEALLOCATE (integrals%utrans)
      END IF
      IF (ASSOCIATED(integrals%uptrans)) THEN
         DEALLOCATE (integrals%uptrans)
      END IF

      integrals%status = 0

   END SUBROUTINE atom_int_release
! **************************************************************************************************
!> \brief ...
!> \param integrals ...
! **************************************************************************************************
   SUBROUTINE atom_ppint_release(integrals)
      TYPE(atom_integrals), INTENT(INOUT)                :: integrals

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_ppint_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(integrals%hnl)) THEN
         DEALLOCATE (integrals%hnl)
      END IF
      IF (ASSOCIATED(integrals%core)) THEN
         DEALLOCATE (integrals%core)
      END IF
      IF (ASSOCIATED(integrals%clsd)) THEN
         DEALLOCATE (integrals%clsd)
      END IF

      integrals%ppstat = 0

   END SUBROUTINE atom_ppint_release
! **************************************************************************************************
!> \brief ...
!> \param integrals ...
! **************************************************************************************************
   SUBROUTINE atom_relint_release(integrals)
      TYPE(atom_integrals), INTENT(INOUT)                :: integrals

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_relint_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(integrals%tzora)) THEN
         DEALLOCATE (integrals%tzora)
      END IF
      IF (ASSOCIATED(integrals%hdkh)) THEN
         DEALLOCATE (integrals%hdkh)
      END IF

      integrals%zorastat = 0
      integrals%dkhstat = 0

   END SUBROUTINE atom_relint_release
! **************************************************************************************************
!> \brief ...
!> \param modpot ...
!> \param grid ...
!> \param zcore ...
! **************************************************************************************************
   SUBROUTINE calculate_model_potential(modpot, grid, zcore)
      REAL(dp), DIMENSION(:), INTENT(INOUT)              :: modpot
      TYPE(grid_atom_type), INTENT(IN)                   :: grid
      REAL(dp), INTENT(IN)                               :: zcore

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_model_potential', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, l, ll, n, z
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: pot, rho
      TYPE(atom_state)                                   :: state

      n = SIZE(modpot)
      ALLOCATE (rho(n), pot(n))

      ! fill default occupation
      state%core = 0._dp
      state%occ = 0._dp
      state%multiplicity = -1
      z = NINT(zcore)
      DO l = 0, 3
         IF (ptable(z)%e_conv(l) /= 0) THEN
            state%maxl_occ = l
            ll = 2*(2*l+1)
            DO i = 1, SIZE(state%occ, 2)
               ii = ptable(z)%e_conv(l)-(i-1)*ll
               IF (ii <= ll) THEN
                  state%occ(l, i) = ii
                  EXIT
               ELSE
                  state%occ(l, i) = ll
               END IF
            END DO
         END IF
      END DO

      modpot = -zcore/grid%rad(:)

      ! Coulomb potential
      CALL slater_density(rho, pot, NINT(zcore), state, grid)
      CALL coulomb_potential_numeric(pot, rho, grid)
      modpot = modpot+pot

      ! XC potential
      CALL wigner_slater_functional(rho, pot)
      modpot = modpot+pot

      DEALLOCATE (rho, pot)

   END SUBROUTINE calculate_model_potential
! **************************************************************************************************

END MODULE atom_operators
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines that fit parameters for /from atomic calculations
! **************************************************************************************************
MODULE atom_fit
   USE atom_electronic_structure,       ONLY: calculate_atom
   USE atom_operators,                  ONLY: atom_int_release,&
                                              atom_int_setup,&
                                              atom_ppint_release,&
                                              atom_ppint_setup,&
                                              atom_relint_release,&
                                              atom_relint_setup
   USE atom_output,                     ONLY: atom_print_basis,&
                                              atom_print_basis_file,&
                                              atom_write_pseudo_param
   USE atom_types,                      ONLY: &
        GTO_BASIS, STO_BASIS, atom_basis_type, atom_gthpot_type, atom_integrals, atom_p_type, &
        atom_potential_type, atom_type, create_opgrid, opgrid_type, release_opgrid, set_atom
   USE atom_utils,                      ONLY: &
        atom_completeness, atom_condnumber, atom_consistent_method, atom_denmat, atom_density, &
        atom_orbital_charge, atom_orbital_max, atom_orbital_nodes, atom_wfnr0, get_maxn_occ, &
        integrate_grid
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE input_constants,                 ONLY: do_analytic
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE lapack,                          ONLY: lapack_sgesv
   USE mathconstants,                   ONLY: fac,&
                                              fourpi,&
                                              pi
   USE periodic_table,                  ONLY: ptable
   USE physcon,                         ONLY: bohr,&
                                              evolt
   USE powell,                          ONLY: opt_state_type,&
                                              powell_optimize
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_fit'

   PUBLIC :: atom_fit_density, atom_fit_basis, atom_fit_pseudo, atom_fit_kgpot

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param num_gto ...
!> \param norder ...
!> \param iunit ...
!> \param powell_section ...
!> \param results ...
! **************************************************************************************************
   SUBROUTINE atom_fit_density(atom, num_gto, norder, iunit, powell_section, results)
      TYPE(atom_type), POINTER                           :: atom
      INTEGER, INTENT(IN)                                :: num_gto, norder, iunit
      TYPE(section_vals_type), OPTIONAL, POINTER         :: powell_section
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: results

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_fit_density', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: n10
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: co
      REAL(KIND=dp), DIMENSION(2)                        :: x
      TYPE(opgrid_type), POINTER                         :: density
      TYPE(opt_state_type)                               :: ostate

      ALLOCATE (co(num_gto))
      co = 0._dp
      NULLIFY (density)
      CALL create_opgrid(density, atom%basis%grid)
      CALL atom_denmat(atom%orbitals%pmat, atom%orbitals%wfn, atom%basis%nbas, atom%state%occupation, &
                       atom%state%maxl_occ, atom%state%maxn_occ)
      CALL atom_density(density%op, atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, &
                        typ="RHO")
      density%op = fourpi*density%op
      IF (norder /= 0) THEN
         density%op = density%op*atom%basis%grid%rad**norder
      END IF

      ostate%nf = 0
      ostate%nvar = 2
      x(1) = 0.10_dp !starting point of geometric series
      x(2) = 2.00_dp !factor of series
      IF (PRESENT(powell_section)) THEN
         CALL section_vals_val_get(powell_section, "ACCURACY", r_val=ostate%rhoend)
         CALL section_vals_val_get(powell_section, "STEP_SIZE", r_val=ostate%rhobeg)
         CALL section_vals_val_get(powell_section, "MAX_FUN", i_val=ostate%maxfun)
      ELSE
         ostate%rhoend = 1.e-8_dp
         ostate%rhobeg = 5.e-2_dp
         ostate%maxfun = 1000
      END IF
      ostate%iprint = 1
      ostate%unit = iunit

      ostate%state = 0
      IF (iunit > 0) THEN
         WRITE (iunit, '(/," POWELL| Start optimization procedure")')
      END IF
      n10 = ostate%maxfun/10

      DO

         IF (ostate%state == 2) THEN
            CALL density_fit(density, atom, num_gto, x(1), x(2), co, ostate%f)
         END IF

         IF (ostate%state == -1) EXIT

         CALL powell_optimize(ostate%nvar, x, ostate)

         IF (MOD(ostate%nf, n10) == 0 .AND. iunit > 0) THEN
            WRITE (iunit, '(" POWELL| Reached",i4,"% of maximal function calls")') &
               INT(REAL(ostate%nf, dp)/REAL(ostate%maxfun, dp)*100._dp)
         END IF

      END DO

      ostate%state = 8
      CALL powell_optimize(ostate%nvar, x, ostate)

      CALL release_opgrid(density)

      IF (iunit > 0) THEN
         WRITE (iunit, '(" POWELL| Number of function evaluations",T71,I10)') ostate%nf
         WRITE (iunit, '(" POWELL| Final value of function",T61,G20.10)') ostate%fopt
         WRITE (iunit, '(" Optimized representation of density using a Geometrical Gaussian basis")')
         WRITE (iunit, '(A,I3,/,T10,A,F10.6,T48,A,F10.6)') " Number of Gaussians:", num_gto, &
            "Starting exponent:", x(1), "Proportionality factor:", x(2)
         WRITE (iunit, '(A)') " Expansion coefficients"
         WRITE (iunit, '(4F20.10)') co(1:num_gto)
      END IF

      IF (PRESENT(results)) THEN
         IF(.NOT.(SIZE(results) >= num_gto+2))CALL cp__a("atom_fit.F",148)
         results(1) = x(1)
         results(2) = x(2)
         results(3:2+num_gto) = co(1:num_gto)
      END IF

      DEALLOCATE (co)

   END SUBROUTINE atom_fit_density
! **************************************************************************************************
!> \brief ...
!> \param atom_info ...
!> \param basis ...
!> \param pptype ...
!> \param iunit ...
!> \param powell_section ...
! **************************************************************************************************
   SUBROUTINE atom_fit_basis(atom_info, basis, pptype, iunit, powell_section)
      TYPE(atom_p_type), DIMENSION(:, :), POINTER        :: atom_info
      TYPE(atom_basis_type), POINTER                     :: basis
      LOGICAL, INTENT(IN)                                :: pptype
      INTEGER, INTENT(IN)                                :: iunit
      TYPE(section_vals_type), POINTER                   :: powell_section

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_fit_basis', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, l, ll, m, n, n10, nl, nr, zval
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: xtob
      LOGICAL                                            :: explicit, mult
      REAL(KIND=dp)                                      :: al, crad, ear, fopt, pf, rk
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: x
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: wem
      REAL(KIND=dp), DIMENSION(:), POINTER               :: w
      TYPE(opt_state_type)                               :: ostate

      SELECT CASE (basis%basis_type)
      CASE DEFAULT
         CALL cp__b("atom_fit.F",185,"")
      CASE (GTO_BASIS)
         IF (basis%geometrical) THEN
            ostate%nvar = 2
            ALLOCATE (x(2))
            x(1) = SQRT(basis%aval)
            x(2) = SQRT(basis%cval)
         ELSE
            ll = MAXVAL(basis%nprim(:))
            ALLOCATE (xtob(ll, 0:3))
            xtob = 0
            ll = SUM(basis%nprim(:))
            ALLOCATE (x(ll))
            x = 0._dp
            ll = 0
            DO l = 0, 3
               DO i = 1, basis%nprim(l)
                  mult = .FALSE.
                  DO k = 1, ll
                     IF (ABS(basis%am(i, l)-x(k)) < 1.e-6_dp) THEN
                        mult = .TRUE.
                        xtob(i, l) = k
                     END IF
                  END DO
                  IF (.NOT. mult) THEN
                     ll = ll+1
                     x(ll) = basis%am(i, l)
                     xtob(i, l) = ll
                  END IF
               END DO
            END DO
            ostate%nvar = ll
            DO i = 1, ostate%nvar
               x(i) = SQRT(LOG(1._dp+x(i)))
            END DO
         END IF
      CASE (STO_BASIS)
         ll = MAXVAL(basis%nbas(:))
         ALLOCATE (xtob(ll, 0:3))
         xtob = 0
         ll = SUM(basis%nbas(:))
         ALLOCATE (x(ll))
         x = 0._dp
         ll = 0
         DO l = 0, 3
            DO i = 1, basis%nbas(l)
               ll = ll+1
               x(ll) = basis%as(i, l)
               xtob(i, l) = ll
            END DO
         END DO
         ostate%nvar = ll
         DO i = 1, ostate%nvar
            x(i) = SQRT(LOG(1._dp+x(i)))
         END DO
      END SELECT

      CALL section_vals_val_get(powell_section, "ACCURACY", r_val=ostate%rhoend)
      CALL section_vals_val_get(powell_section, "STEP_SIZE", r_val=ostate%rhobeg)
      CALL section_vals_val_get(powell_section, "MAX_FUN", i_val=ostate%maxfun)

      n = SIZE(atom_info, 1)
      m = SIZE(atom_info, 2)
      ALLOCATE (wem(n, m))
      wem = 1._dp
      CALL section_vals_val_get(powell_section, "WEIGHT_ELECTRON_CONFIGURATION", explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(powell_section, "WEIGHT_ELECTRON_CONFIGURATION", r_vals=w)
         DO i = 1, MIN(SIZE(w), n)
            wem(i, :) = w(i)*wem(i, :)
         END DO
      END IF
      CALL section_vals_val_get(powell_section, "WEIGHT_METHOD", explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(powell_section, "WEIGHT_METHOD", r_vals=w)
         DO i = 1, MIN(SIZE(w), m)
            wem(:, i) = w(i)*wem(:, i)
         END DO
      END IF

      DO i = 1, SIZE(atom_info, 1)
         DO j = 1, SIZE(atom_info, 2)
            atom_info(i, j)%atom%weight = wem(i, j)
         END DO
      END DO
      DEALLOCATE (wem)

      ostate%nf = 0
      ostate%iprint = 1
      ostate%unit = iunit

      ostate%state = 0
      IF (iunit > 0) THEN
         WRITE (iunit, '(/," POWELL| Start optimization procedure")')
         WRITE (iunit, '(/," POWELL| Total number of parameters in optimization",T71,I10)') ostate%nvar
      END IF
      n10 = MAX(ostate%maxfun/100, 1)

      fopt = HUGE(0._dp)

      DO

         IF (ostate%state == 2) THEN
            SELECT CASE (basis%basis_type)
            CASE DEFAULT
               CALL cp__b("atom_fit.F",290,"")
            CASE (GTO_BASIS)
               IF (basis%geometrical) THEN
                  basis%am = 0._dp
                  DO l = 0, 3
                     DO i = 1, basis%nbas(l)
                        ll = i-1+basis%start(l)
                        basis%am(i, l) = x(1)*x(1)*(x(2)*x(2))**(ll)
                     END DO
                  END DO
                  basis%aval = x(1)*x(1)
                  basis%cval = x(2)*x(2)
               ELSE
                  DO l = 0, 3
                     DO i = 1, basis%nprim(l)
                        al = x(xtob(i, l))**2
                        basis%am(i, l) = EXP(al)-1._dp
                     END DO
                  END DO
               END IF
               basis%bf = 0._dp
               basis%dbf = 0._dp
               nr = basis%grid%nr
               DO l = 0, 3
                  DO i = 1, basis%nbas(l)
                     al = basis%am(i, l)
                     DO k = 1, nr
                        rk = basis%grid%rad(k)
                        ear = EXP(-al*basis%grid%rad(k)**2)
                        basis%bf(k, i, l) = rk**l*ear
                        basis%dbf(k, i, l) = (REAL(l, dp)*rk**(l-1)-2._dp*al*rk**(l+1))*ear
                     END DO
                  END DO
               END DO
            CASE (STO_BASIS)
               DO l = 0, 3
                  DO i = 1, basis%nbas(l)
                     al = x(xtob(i, l))**2
                     basis%as(i, l) = EXP(al)-1._dp
                  END DO
               END DO
               basis%bf = 0._dp
               basis%dbf = 0._dp
               nr = basis%grid%nr
               DO l = 0, 3
                  DO i = 1, basis%nbas(l)
                     al = basis%as(i, l)
                     nl = basis%ns(i, l)
                     pf = (2._dp*al)**nl*SQRT(2._dp*al/fac(2*nl))
                     DO k = 1, nr
                        rk = basis%grid%rad(k)
                        ear = EXP(-al*basis%grid%rad(k))
                        basis%bf(k, i, l) = pf*rk**(nl-1)*ear
                        basis%dbf(k, i, l) = pf*(REAL(nl-1, dp)/rk-al)*rk**(nl-1)*ear
                     END DO
                  END DO
               END DO
            END SELECT
            CALL basis_fit(atom_info, basis, pptype, ostate%f, 0)
            fopt = MIN(fopt, ostate%f)
         END IF

         IF (ostate%state == -1) EXIT

         CALL powell_optimize(ostate%nvar, x, ostate)

         IF (ostate%nf == 2 .AND. iunit > 0) THEN
            WRITE (iunit, '(" POWELL| Inital value of function",T61,F20.10)') ostate%f
         END IF
         IF (MOD(ostate%nf, n10) == 0 .AND. iunit > 0) THEN
            WRITE (iunit, '(" POWELL| Reached",i4,"% of maximal function calls",T61,F20.10)') &
               INT(REAL(ostate%nf, dp)/REAL(ostate%maxfun, dp)*100._dp), fopt
         END IF

      END DO

      ostate%state = 8
      CALL powell_optimize(ostate%nvar, x, ostate)

      IF (iunit > 0) THEN
         WRITE (iunit, '(" POWELL| Number of function evaluations",T71,I10)') ostate%nf
         WRITE (iunit, '(" POWELL| Final value of function",T61,F20.10)') ostate%fopt
         ! x->basis
         SELECT CASE (basis%basis_type)
         CASE DEFAULT
            CALL cp__b("atom_fit.F",375,"")
         CASE (GTO_BASIS)
            IF (basis%geometrical) THEN
               basis%am = 0._dp
               DO l = 0, 3
                  DO i = 1, basis%nbas(l)
                     ll = i-1+basis%start(l)
                     basis%am(i, l) = x(1)*x(1)*(x(2)*x(2))**(ll)
                  END DO
               END DO
               basis%aval = x(1)*x(1)
               basis%cval = x(2)*x(2)
            ELSE
               DO l = 0, 3
                  DO i = 1, basis%nprim(l)
                     al = x(xtob(i, l))**2
                     basis%am(i, l) = EXP(al)-1._dp
                  END DO
               END DO
            END IF
            basis%bf = 0._dp
            basis%dbf = 0._dp
            nr = basis%grid%nr
            DO l = 0, 3
               DO i = 1, basis%nbas(l)
                  al = basis%am(i, l)
                  DO k = 1, nr
                     rk = basis%grid%rad(k)
                     ear = EXP(-al*basis%grid%rad(k)**2)
                     basis%bf(k, i, l) = rk**l*ear
                     basis%dbf(k, i, l) = (REAL(l, dp)*rk**(l-1)-2._dp*al*rk**(l+1))*ear
                  END DO
               END DO
            END DO
         CASE (STO_BASIS)
            DO l = 0, 3
               DO i = 1, basis%nprim(l)
                  al = x(xtob(i, l))**2
                  basis%as(i, l) = EXP(al)-1._dp
               END DO
            END DO
            basis%bf = 0._dp
            basis%dbf = 0._dp
            nr = basis%grid%nr
            DO l = 0, 3
               DO i = 1, basis%nbas(l)
                  al = basis%as(i, l)
                  nl = basis%ns(i, l)
                  pf = (2._dp*al)**nl*SQRT(2._dp*al/fac(2*nl))
                  DO k = 1, nr
                     rk = basis%grid%rad(k)
                     ear = EXP(-al*basis%grid%rad(k))
                     basis%bf(k, i, l) = pf*rk**(nl-1)*ear
                     basis%dbf(k, i, l) = pf*(REAL(nl-1, dp)/rk-al)*rk**(nl-1)*ear
                  END DO
               END DO
            END DO
         END SELECT
         CALL atom_print_basis(basis, iunit, " Optimized Basis")
         CALL atom_print_basis_file(basis)
      END IF

      DEALLOCATE (x)

      IF (ALLOCATED(xtob)) THEN
         DEALLOCATE (xtob)
      END IF

      SELECT CASE (basis%basis_type)
      CASE DEFAULT
         CALL cp__b("atom_fit.F",445,"")
      CASE (GTO_BASIS)
         zval = atom_info(1, 1)%atom%z
         crad = ptable(zval)%covalent_radius*bohr
         IF (iunit > 0) THEN
            CALL atom_condnumber(basis, crad, iunit)
            CALL atom_completeness(basis, zval, iunit)
         END IF
      CASE (STO_BASIS)
         ! no basis test available
      END SELECT

   END SUBROUTINE atom_fit_basis
! **************************************************************************************************
!> \brief ...
!> \param atom_info ...
!> \param atom_refs ...
!> \param ppot ...
!> \param iunit ...
!> \param powell_section ...
! **************************************************************************************************
   SUBROUTINE atom_fit_pseudo(atom_info, atom_refs, ppot, iunit, powell_section)
      TYPE(atom_p_type), DIMENSION(:, :), POINTER        :: atom_info, atom_refs
      TYPE(atom_potential_type), POINTER                 :: ppot
      INTEGER, INTENT(IN)                                :: iunit
      TYPE(section_vals_type), POINTER                   :: powell_section

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_fit_pseudo', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=2)                                   :: pc1, pc2
      INTEGER                                            :: i, i1, i2, iw, j, j1, j2, k, l, m, n, &
                                                            n10, np, nre, nreinit, ntarget
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: xtob
      INTEGER, DIMENSION(0:3)                            :: ncore
      LOGICAL                                            :: explicit
      REAL(KIND=dp) :: afun, charge, de, deig, drho, dx, eig, fopt, oc, pchg, peig, pv, rcm, rcov, &
         rmax, t_ener, t_semi, t_valence, t_virt, w_ener, w_node, w_psir0, w_semi, w_valence, &
         w_virt, wtot
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: x, xi
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: wem
      REAL(KIND=dp), ALLOCATABLE, &
         DIMENSION(:, :, :, :, :)                        :: dener, pval
      REAL(KIND=dp), DIMENSION(:), POINTER               :: w
      TYPE(atom_type), POINTER                           :: atom
      TYPE(opt_state_type)                               :: ostate

      ! weights for the optimization
      CALL section_vals_val_get(powell_section, "ACCURACY", r_val=ostate%rhoend)
      CALL section_vals_val_get(powell_section, "STEP_SIZE", r_val=ostate%rhobeg)
      CALL section_vals_val_get(powell_section, "MAX_FUN", i_val=ostate%maxfun)
      CALL section_vals_val_get(powell_section, "MAX_INIT", i_val=nreinit)

      CALL section_vals_val_get(powell_section, "WEIGHT_POT_VALENCE", r_val=w_valence)
      CALL section_vals_val_get(powell_section, "WEIGHT_POT_VIRTUAL", r_val=w_virt)
      CALL section_vals_val_get(powell_section, "WEIGHT_POT_SEMICORE", r_val=w_semi)
      CALL section_vals_val_get(powell_section, "WEIGHT_POT_NODE", r_val=w_node)
      CALL section_vals_val_get(powell_section, "WEIGHT_DELTA_ENERGY", r_val=w_ener)

      CALL section_vals_val_get(powell_section, "WEIGHT_PSIR0", r_val=w_psir0)
      CALL section_vals_val_get(powell_section, "RCOV_MULTIPLICATION", r_val=rcm)

      CALL section_vals_val_get(powell_section, "TARGET_POT_VALENCE", r_val=t_valence)
      CALL section_vals_val_get(powell_section, "TARGET_POT_VIRTUAL", r_val=t_virt)
      CALL section_vals_val_get(powell_section, "TARGET_POT_SEMICORE", r_val=t_semi)
      CALL section_vals_val_get(powell_section, "TARGET_DELTA_ENERGY", r_val=t_ener)

      n = SIZE(atom_info, 1)
      m = SIZE(atom_info, 2)
      ALLOCATE (wem(n, m))
      wem = 1._dp
      ALLOCATE (pval(8, 10, 0:3, m, n))
      ALLOCATE (dener(2, m, m, n, n))
      dener = 0.0_dp

      CALL section_vals_val_get(powell_section, "WEIGHT_ELECTRON_CONFIGURATION", explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(powell_section, "WEIGHT_ELECTRON_CONFIGURATION", r_vals=w)
         DO i = 1, MIN(SIZE(w), n)
            wem(i, :) = w(i)*wem(i, :)
         END DO
      END IF
      CALL section_vals_val_get(powell_section, "WEIGHT_METHOD", explicit=explicit)
      IF (explicit) THEN
         CALL section_vals_val_get(powell_section, "WEIGHT_METHOD", r_vals=w)
         DO i = 1, MIN(SIZE(w), m)
            wem(:, i) = w(i)*wem(:, i)
         END DO
      END IF

      CALL open_file(file_name="/mnt/sdcard/cp2k/work/POWELL_RESULT", file_status="UNKNOWN", file_action="WRITE", unit_number=iw)

      ALLOCATE (xi(200))
      !decide here what to optimize
      CALL get_pseudo_param(xi, ostate%nvar, ppot%gth_pot)
      ALLOCATE (x(ostate%nvar))
      x(1:ostate%nvar) = xi(1:ostate%nvar)
      DEALLOCATE (xi)

      ostate%nf = 0
      ostate%iprint = 1
      ostate%unit = iunit

      ostate%state = 0
      IF (iunit > 0) THEN
         WRITE (iunit, '(/," POWELL| Start optimization procedure")')
         WRITE (iunit, '(/," POWELL| Total number of parameters in optimization",T71,I10)') ostate%nvar
      END IF
      n10 = MAX(ostate%maxfun/100, 1)

      rcov = ptable(atom_refs(1, 1)%atom%z)%covalent_radius*bohr*rcm
      !set all reference values
      ntarget = 0
      wtot = 0.0_dp
      DO i = 1, SIZE(atom_info, 1)
         DO j = 1, SIZE(atom_info, 2)
            atom => atom_info(i, j)%atom
            IF (atom_consistent_method(atom%method_type, atom%state%multiplicity)) THEN
               dener(2, j, j, i, i) = atom_refs(i, j)%atom%energy%etot
               IF (atom%state%multiplicity == -1) THEN
                  ! no spin polarization
                  atom%weight = wem(i, j)
                  ncore = get_maxn_occ(atom_info(i, j)%atom%state%core)
                  DO l = 0, 3
                     np = atom%state%maxn_calc(l)
                     DO k = 1, np
                        CALL atom_orbital_max(rmax, atom_refs(i, j)%atom%orbitals%wfn(:, ncore(l)+k, l), &
                                              rcov, l, atom_refs(i, j)%atom%basis)
                        atom%orbitals%rcmax(k, l, 1) = MAX(rcov, rmax)
                        CALL atom_orbital_charge(charge, atom_refs(i, j)%atom%orbitals%wfn(:, ncore(l)+k, l), &
                                                 atom%orbitals%rcmax(k, l, 1), l, atom_refs(i, j)%atom%basis)
                        atom%orbitals%refene(k, l, 1) = atom_refs(i, j)%atom%orbitals%ener(ncore(l)+k, l)
                        atom%orbitals%refchg(k, l, 1) = charge
                        IF (k > atom%state%maxn_occ(l)) THEN
                           IF (k <= atom%state%maxn_occ(l)+1) THEN
                              atom%orbitals%wrefene(k, l, 1) = w_virt
                              atom%orbitals%wrefchg(k, l, 1) = w_virt/100._dp
                              atom%orbitals%crefene(k, l, 1) = t_virt
                              ntarget = ntarget+2
                              wtot = wtot+atom%weight*(w_virt+w_virt/100._dp)
                           ELSE
                              atom%orbitals%wrefene(k, l, 1) = w_virt/100._dp
                              atom%orbitals%wrefchg(k, l, 1) = 0._dp
                              atom%orbitals%crefene(k, l, 1) = t_virt*10._dp
                              ntarget = ntarget+1
                              wtot = wtot+atom%weight*w_virt
                           END IF
                        ELSEIF (k < atom%state%maxn_occ(l)) THEN
                           atom%orbitals%wrefene(k, l, 1) = w_semi
                           atom%orbitals%wrefchg(k, l, 1) = w_semi/100._dp
                           atom%orbitals%crefene(k, l, 1) = t_semi
                           ntarget = ntarget+2
                           wtot = wtot+atom%weight*(w_semi+w_semi/100._dp)
                        ELSE
                           IF (ABS(atom%state%occupation(l, k)-REAL(4*l+2, KIND=dp)) < 0.01_dp) THEN
                              ! full shell
                              atom%orbitals%wrefene(k, l, 1) = w_semi
                              atom%orbitals%wrefchg(k, l, 1) = w_semi/100._dp
                              atom%orbitals%crefene(k, l, 1) = t_semi
                              wtot = wtot+atom%weight*(w_semi+w_semi/100._dp)
                           ELSE
                              atom%orbitals%wrefene(k, l, 1) = w_valence
                              atom%orbitals%wrefchg(k, l, 1) = w_valence/100._dp
                              atom%orbitals%crefene(k, l, 1) = t_valence
                              wtot = wtot+atom%weight*(w_valence+w_valence/100._dp)
                           END IF
                           IF (l == 0) THEN
                              atom%orbitals%wpsir0(k, 1) = w_psir0
                              wtot = wtot+atom%weight*w_psir0
                           END IF
                           ntarget = ntarget+2
                        END IF
                     END DO
                     DO k = 1, np
                        atom%orbitals%refnod(k, l, 1) = REAL(k-1, KIND=dp)
                        ! we only enforce 0-nodes for the first state
                        IF (k == 1 .AND. atom%state%occupation(l, k) /= 0._dp) THEN
                           atom%orbitals%wrefnod(k, l, 1) = w_node
                           wtot = wtot+atom%weight*w_node
                        END IF
                     END DO
                  END DO
               ELSE
                  ! spin polarization
                  atom%weight = wem(i, j)
                  ncore = get_maxn_occ(atom_info(i, j)%atom%state%core)
                  DO l = 0, 3
                     np = atom%state%maxn_calc(l)
                     DO k = 1, np
                        CALL atom_orbital_max(rmax, atom_refs(i, j)%atom%orbitals%wfna(:, ncore(l)+k, l), &
                                              rcov, l, atom_refs(i, j)%atom%basis)
                        atom%orbitals%rcmax(k, l, 1) = MAX(rcov, rmax)
                        CALL atom_orbital_max(rmax, atom_refs(i, j)%atom%orbitals%wfnb(:, ncore(l)+k, l), &
                                              rcov, l, atom_refs(i, j)%atom%basis)
                        atom%orbitals%rcmax(k, l, 2) = MAX(rcov, rmax)
                        CALL atom_orbital_charge(charge, atom_refs(i, j)%atom%orbitals%wfna(:, ncore(l)+k, l), &
                                                 atom%orbitals%rcmax(k, l, 1), l, atom_refs(i, j)%atom%basis)
                        atom%orbitals%refene(k, l, 1) = atom_refs(i, j)%atom%orbitals%enera(ncore(l)+k, l)
                        atom%orbitals%refchg(k, l, 1) = charge
                        CALL atom_orbital_charge(charge, atom_refs(i, j)%atom%orbitals%wfnb(:, ncore(l)+k, l), &
                                                 atom%orbitals%rcmax(k, l, 1), l, atom_refs(i, j)%atom%basis)
                        atom%orbitals%refene(k, l, 2) = atom_refs(i, j)%atom%orbitals%enerb(ncore(l)+k, l)
                        atom%orbitals%refchg(k, l, 2) = charge
                        ! the following assignments could be further specialized
                        IF (k > atom%state%maxn_occ(l)) THEN
                           IF (k <= atom%state%maxn_occ(l)+1) THEN
                              atom%orbitals%wrefene(k, l, 1:2) = w_virt
                              atom%orbitals%wrefchg(k, l, 1:2) = w_virt/100._dp
                              atom%orbitals%crefene(k, l, 1:2) = t_virt
                              ntarget = ntarget+4
                              wtot = wtot+atom%weight*2._dp*(w_virt+w_virt/100._dp)
                           ELSE
                              atom%orbitals%wrefene(k, l, 1:2) = w_virt/100._dp
                              atom%orbitals%wrefchg(k, l, 1:2) = 0._dp
                              atom%orbitals%crefene(k, l, 1:2) = t_virt*10.0_dp
                              wtot = wtot+atom%weight*2._dp*w_virt
                              ntarget = ntarget+2
                           END IF
                        ELSEIF (k < atom%state%maxn_occ(l)) THEN
                           atom%orbitals%wrefene(k, l, 1:2) = w_semi
                           atom%orbitals%wrefchg(k, l, 1:2) = w_semi/100._dp
                           atom%orbitals%crefene(k, l, 1:2) = t_semi
                           ntarget = ntarget+4
                           wtot = wtot+atom%weight*2._dp*(w_semi+w_semi/100._dp)
                        ELSE
                           IF (ABS(atom%state%occupation(l, k)-REAL(2*l+1, KIND=dp)) < 0.01_dp) THEN
                              atom%orbitals%wrefene(k, l, 1:2) = w_semi
                              atom%orbitals%wrefchg(k, l, 1:2) = w_semi/100._dp
                              atom%orbitals%crefene(k, l, 1:2) = t_semi
                              wtot = wtot+atom%weight*2._dp*(w_semi+w_semi/100._dp)
                           ELSE
                              atom%orbitals%wrefene(k, l, 1:2) = w_valence
                              atom%orbitals%wrefchg(k, l, 1:2) = w_valence/100._dp
                              atom%orbitals%crefene(k, l, 1:2) = t_valence
                              wtot = wtot+atom%weight*2._dp*(w_valence+w_valence/100._dp)
                           END IF
                           ntarget = ntarget+4
                           IF (l == 0) THEN
                              atom%orbitals%wpsir0(k, 1:2) = w_psir0
                              wtot = wtot+atom%weight*2._dp*w_psir0
                           END IF
                        END IF
                     END DO
                     DO k = 1, np
                        atom%orbitals%refnod(k, l, 1:2) = REAL(k-1, KIND=dp)
                        ! we only enforce 0-nodes for the first state
                        IF (k == 1 .AND. atom%state%occa(l, k) /= 0._dp) THEN
                           atom%orbitals%wrefnod(k, l, 1) = w_node
                           wtot = wtot+atom%weight*w_node
                        END IF
                        IF (k == 1 .AND. atom%state%occb(l, k) /= 0._dp) THEN
                           atom%orbitals%wrefnod(k, l, 2) = w_node
                           wtot = wtot+atom%weight*w_node
                        END IF
                     END DO
                  END DO
               END IF
               CALL calculate_atom(atom, 0)
            END IF
         END DO
      END DO
      ! energy differences
      DO j1 = 1, SIZE(atom_info, 2)
         DO j2 = 1, SIZE(atom_info, 2)
            DO i1 = 1, SIZE(atom_info, 1)
               DO i2 = 1, SIZE(atom_info, 1)
                  IF ((j1 > j2) .OR. (j1 == j2 .AND. i1 >= i2)) CYCLE
                  dener(2, j1, j2, i1, i2) = dener(2, j1, j1, i1, i1)-dener(2, j2, j2, i2, i2)
                  wtot = wtot+w_ener
               END DO
            END DO
         END DO
      END DO

      DEALLOCATE (wem)

      ALLOCATE (xi(ostate%nvar))
      DO nre = 1, nreinit
         xi(:) = x(:)
         CALL put_pseudo_param(x, ppot%gth_pot)
         CALL pseudo_fit(atom_info, ppot, ostate%f, wtot, pval, dener, w_ener, t_ener)
         IF (nre == 1) THEN
            WRITE (iunit, '(/," POWELL| Initial errors of target values")')
            afun = ostate%f*wtot
            DO i = 1, SIZE(atom_info, 1)
               DO j = 1, SIZE(atom_info, 2)
                  atom => atom_info(i, j)%atom
                  IF (atom_consistent_method(atom%method_type, atom%state%multiplicity)) THEN
                     WRITE (iunit, '(/," Reference configuration  ",T31,i5,T50," Method number ",T76,i5)') i, j
                     IF (atom%state%multiplicity == -1) THEN
                        ! no spin polarization
                        WRITE (iunit, '("    L    N    Occupation      Eigenvalue [eV]           dE [eV]          dCharge ")')
                        DO l = 0, 3
                           np = atom%state%maxn_calc(l)
                           IF (np > 0) THEN
                              DO k = 1, np
                                 oc = atom%state%occupation(l, k)
                                 eig = atom%orbitals%ener(k, l)
                                 deig = eig-atom%orbitals%refene(k, l, 1)
                                 peig = pval(1, k, l, j, i)/afun*100._dp
                                 IF (pval(5, k, l, j, i) > 0.5_dp) THEN
                                    pc1 = " X"
                                 ELSE
                                    WRITE (pc1, "(I2)") NINT(peig)
                                 END IF
                                 CALL atom_orbital_charge(charge, atom%orbitals%wfn(:, k, l), &
                                                          atom%orbitals%rcmax(k, l, 1), l, atom%basis)
                                 drho = charge-atom%orbitals%refchg(k, l, 1)
                                 pchg = pval(2, k, l, j, i)/afun*100._dp
                                 IF (pval(6, k, l, j, i) > 0.5_dp) THEN
                                    pc2 = " X"
                                 ELSE
                                    WRITE (pc2, "(I2)") MIN(NINT(pchg), 99)
                                 END IF
                                 WRITE (iunit, '(I5,I5,F14.2,F21.10,F14.6,"[",A2,"]",F13.6,"[",A2,"]")') &
                                    l, k, oc, eig*evolt, deig*evolt, pc1, drho, pc2
                              END DO
                           END IF
                        END DO
                     ELSE
                        ! spin polarization
                        WRITE (iunit, '("    L    N   Spin Occupation    Eigenvalue [eV]          dE [eV]         dCharge ")')
                        DO l = 0, 3
                           np = atom%state%maxn_calc(l)
                           IF (np > 0) THEN
                              DO k = 1, np
                                 oc = atom%state%occa(l, k)
                                 eig = atom%orbitals%enera(k, l)
                                 deig = eig-atom%orbitals%refene(k, l, 1)
                                 peig = pval(1, k, l, j, i)/afun*100._dp
                                 IF (pval(5, k, l, j, i) > 0.5_dp) THEN
                                    pc1 = " X"
                                 ELSE
                                    WRITE (pc1, "(I2)") NINT(peig)
                                 END IF
                                 CALL atom_orbital_charge( &
                                    charge, atom%orbitals%wfna(:, k, l), atom%orbitals%rcmax(k, l, 1), l, atom%basis)
                                 drho = charge-atom%orbitals%refchg(k, l, 1)
                                 pchg = pval(2, k, l, j, i)/afun*100._dp
                                 IF (pval(6, k, l, j, i) > 0.5_dp) THEN
                                    pc2 = " X"
                                 ELSE
                                    WRITE (pc2, "(I2)") MIN(NINT(pchg), 99)
                                 END IF
                                 WRITE (iunit, '(I5,I5,2X,A5,F11.2,F19.10,F13.6,"[",A2,"]",F12.6,"[",A2,"]")') &
                                    l, k, "alpha", oc, eig*evolt, deig*evolt, pc1, drho, pc2
                                 oc = atom%state%occb(l, k)
                                 eig = atom%orbitals%enerb(k, l)
                                 deig = eig-atom%orbitals%refene(k, l, 2)
                                 peig = pval(3, k, l, j, i)/afun*100._dp
                                 IF (pval(7, k, l, j, i) > 0.5_dp) THEN
                                    pc1 = " X"
                                 ELSE
                                    WRITE (pc1, "(I2)") NINT(peig)
                                 END IF
                                 CALL atom_orbital_charge( &
                                    charge, atom%orbitals%wfnb(:, k, l), atom%orbitals%rcmax(k, l, 2), l, atom%basis)
                                 drho = charge-atom%orbitals%refchg(k, l, 2)
                                 pchg = pval(4, k, l, j, i)/afun*100._dp
                                 IF (pval(8, k, l, j, i) > 0.5_dp) THEN
                                    pc2 = " X"
                                 ELSE
                                    WRITE (pc2, "(I2)") MIN(NINT(pchg), 99)
                                 END IF
                                 WRITE (iunit, '(I5,I5,2X,A5,F11.2,F19.10,F13.6,"[",A2,"]",F12.6,"[",A2,"]")') &
                                    l, k, " beta", oc, eig*evolt, deig*evolt, pc1, drho, pc2
                              END DO
                           END IF
                        END DO
                     END IF
                  END IF
               END DO
               WRITE (iunit, *)
            END DO
            ! energy differences
            IF (n*m > 1) THEN
               WRITE (iunit, '(" Method/Econf 1 "," Method/Econf 2","      Delta Energy ","      Error Energy ","     Target")')
               DO j1 = 1, SIZE(atom_info, 2)
                  DO j2 = 1, SIZE(atom_info, 2)
                     DO i1 = 1, SIZE(atom_info, 1)
                        DO i2 = 1, SIZE(atom_info, 1)
                           IF ((j1 > j2) .OR. (j1 == j2 .AND. i1 >= i2)) CYCLE
                           IF (ABS(dener(1, j1, j1, i1, i1)) < 0.000001_dp) CYCLE
                           IF (ABS(dener(2, j1, j1, i1, i1)) < 0.000001_dp) CYCLE
                           IF (ABS(dener(1, j2, j2, i2, i2)) < 0.000001_dp) CYCLE
                           IF (ABS(dener(2, j2, j2, i2, i2)) < 0.000001_dp) CYCLE
                           de = dener(2, j1, j2, i1, i2)-dener(1, j1, j2, i1, i2)
                           WRITE (iunit, '(i6,i6,i10,i6,5X,F16.6,F19.6,F12.6)') &
                              j1, i1, j2, i2, dener(2, j1, j2, i1, i2), de, t_ener
                        END DO
                     END DO
                  END DO
               END DO
               WRITE (iunit, *)
            END IF

            WRITE (iunit, '(" Number of target values reached: ",T69,i5," of ",i3)') &
               INT(SUM(pval(5:8, :, :, :, :))), ntarget
            WRITE (iunit, *)

         END IF

         fopt = HUGE(0._dp)
         ostate%state = 0

         DO

            IF (ostate%state == 2) THEN
               CALL put_pseudo_param(x, ppot%gth_pot)
               CALL pseudo_fit(atom_info, ppot, ostate%f, wtot, pval, dener, w_ener, t_ener)
               fopt = MIN(fopt, ostate%f)
            END IF

            IF (ostate%state == -1) EXIT

            CALL powell_optimize(ostate%nvar, x, ostate)

            IF (ostate%nf == 2 .AND. iunit > 0) THEN
               WRITE (iunit, '(" POWELL| Inital value of function",T61,F20.10)') ostate%f
            END IF
            IF (MOD(ostate%nf, n10) == 0 .AND. iunit > 0 .AND. ostate%nf > 2) THEN
               WRITE (iunit, '(" POWELL| Reached",i4,"% of maximal function calls",T61,F20.10)') &
                  INT(REAL(ostate%nf, dp)/REAL(ostate%maxfun, dp)*100._dp), fopt
               CALL put_pseudo_param(ostate%xopt, ppot%gth_pot)
               CALL atom_write_pseudo_param(ppot%gth_pot)
            END IF

            WRITE (iw, *) ostate%nf, ostate%f, x(1:ostate%nvar)

         END DO

         dx = SQRT(SUM((ostate%xopt(:)-xi(:))**2)/REAL(ostate%nvar, KIND=dp))
         IF (iunit > 0) THEN
            WRITE (iunit, '(" POWELL| RMS average of variables",T69,F12.10)') dx
         END IF
         ostate%state = 8
         CALL powell_optimize(ostate%nvar, x, ostate)
         CALL put_pseudo_param(x, ppot%gth_pot)
         CALL atom_write_pseudo_param(ppot%gth_pot)
         IF (dx < SQRT(ostate%rhoend)) EXIT
         ostate%rhobeg = 0.25_dp*ostate%rhobeg
      END DO
      DEALLOCATE (xi)

      IF (iunit > 0) THEN
         WRITE (iunit, '(" POWELL| Number of function evaluations",T71,I10)') ostate%nf
         WRITE (iunit, '(" POWELL| Final value of function",T61,F20.10)') ostate%fopt

         CALL put_pseudo_param(x, ppot%gth_pot)
         CALL pseudo_fit(atom_info, ppot, ostate%f, wtot, pval, dener, w_ener, t_ener)
         afun = wtot*ostate%f

         WRITE (iunit, '(/," POWELL| Final errors of target values")')
         DO i = 1, SIZE(atom_info, 1)
            DO j = 1, SIZE(atom_info, 2)
               atom => atom_info(i, j)%atom
               IF (atom_consistent_method(atom%method_type, atom%state%multiplicity)) THEN
                  WRITE (iunit, '(/," Reference configuration  ",T31,i5,T50," Method number ",T76,i5)') i, j
                  IF (atom%state%multiplicity == -1) THEN
                     !no spin polarization
                     WRITE (iunit, '("    L    N    Occupation      Eigenvalue [eV]           dE [eV]          dCharge ")')
                     DO l = 0, 3
                        np = atom%state%maxn_calc(l)
                        IF (np > 0) THEN
                           DO k = 1, np
                              oc = atom%state%occupation(l, k)
                              eig = atom%orbitals%ener(k, l)
                              deig = eig-atom%orbitals%refene(k, l, 1)
                              peig = pval(1, k, l, j, i)/afun*100._dp
                              IF (pval(5, k, l, j, i) > 0.5_dp) THEN
                                 pc1 = " X"
                              ELSE
                                 WRITE (pc1, "(I2)") NINT(peig)
                              END IF
                              CALL atom_orbital_charge( &
                                 charge, atom%orbitals%wfn(:, k, l), atom%orbitals%rcmax(k, l, 1), l, atom%basis)
                              drho = charge-atom%orbitals%refchg(k, l, 1)
                              pchg = pval(2, k, l, j, i)/afun*100._dp
                              IF (pval(6, k, l, j, i) > 0.5_dp) THEN
                                 pc2 = " X"
                              ELSE
                                 WRITE (pc2, "(I2)") MIN(NINT(pchg), 99)
                              END IF
                              WRITE (iunit, '(I5,I5,F14.2,F21.10,F14.6,"[",A2,"]",F13.6,"[",A2,"]")') &
                                 l, k, oc, eig*evolt, deig*evolt, pc1, drho, pc2
                           END DO
                        END IF
                     END DO
                     np = atom%state%maxn_calc(0)
                     DO k = 1, np
                        CALL atom_wfnr0(pv, atom%orbitals%wfn(:, k, 0), atom%basis)
                        pchg = atom%weight*atom%orbitals%wpsir0(k, 1)*pv*pv/afun*100._dp
                        WRITE (iunit, '("    s-states"," N=",I5,T40,"Wavefunction at r=0:",T64,F13.6,"[",I2,"]")') k, pv, NINT(pchg)
                     END DO
                  ELSE
                     !spin polarization
                     WRITE (iunit, '("    L    N   Spin Occupation     Eigenvalue [eV]          dE [eV]        dCharge ")')
                     DO l = 0, 3
                        np = atom%state%maxn_calc(l)
                        IF (np > 0) THEN
                           DO k = 1, np
                              oc = atom%state%occa(l, k)
                              eig = atom%orbitals%enera(k, l)
                              deig = eig-atom%orbitals%refene(k, l, 1)
                              peig = pval(1, k, l, j, i)/afun*100._dp
                              IF (pval(5, k, l, j, i) > 0.5_dp) THEN
                                 pc1 = " X"
                              ELSE
                                 WRITE (pc1, "(I2)") NINT(peig)
                              END IF
                              CALL atom_orbital_charge( &
                                 charge, atom%orbitals%wfna(:, k, l), atom%orbitals%rcmax(k, l, 1), l, atom%basis)
                              drho = charge-atom%orbitals%refchg(k, l, 1)
                              pchg = pval(2, k, l, j, i)/afun*100._dp
                              IF (pval(6, k, l, j, i) > 0.5_dp) THEN
                                 pc2 = " X"
                              ELSE
                                 WRITE (pc2, "(I2)") MIN(NINT(pchg), 99)
                              END IF
                              WRITE (iunit, '(I5,I5,A,F11.2,F20.10,F13.6,"[",A2,"]",F11.6,"[",A2,"]")') &
                                 l, k, "  alpha", oc, eig*evolt, deig*evolt, pc1, drho, pc2
                              oc = atom%state%occb(l, k)
                              eig = atom%orbitals%enerb(k, l)
                              deig = eig-atom%orbitals%refene(k, l, 2)
                              peig = pval(3, k, l, j, i)/afun*100._dp
                              IF (pval(7, k, l, j, i) > 0.5_dp) THEN
                                 pc1 = " X"
                              ELSE
                                 WRITE (pc1, "(I2)") NINT(peig)
                              END IF
                              CALL atom_orbital_charge( &
                                 charge, atom%orbitals%wfnb(:, k, l), atom%orbitals%rcmax(k, l, 2), l, atom%basis)
                              drho = charge-atom%orbitals%refchg(k, l, 2)
                              pchg = pval(4, k, l, j, i)/afun*100._dp
                              IF (pval(8, k, l, j, i) > 0.5_dp) THEN
                                 pc2 = " X"
                              ELSE
                                 WRITE (pc2, "(I2)") MIN(NINT(pchg), 99)
                              END IF
                              WRITE (iunit, '(I5,I5,A,F11.2,F20.10,F13.6,"[",A2,"]",F11.6,"[",A2,"]")') &
                                 l, k, "   beta", oc, eig*evolt, deig*evolt, pc1, drho, pc2
                           END DO
                        END IF
                     END DO
                     np = atom%state%maxn_calc(0)
                     DO k = 1, np
                        CALL atom_wfnr0(pv, atom%orbitals%wfna(:, k, 0), atom%basis)
                        pchg = atom%weight*atom%orbitals%wpsir0(k, 1)*pv*pv/afun*100._dp
                        WRITE (iunit, '("    s-states"," N=",I5,T35,"Alpha Wavefunction at r=0:",T64,F13.6,"[",I2,"]")') &
                           k, pv, NINT(pchg)
                        CALL atom_wfnr0(pv, atom%orbitals%wfnb(:, k, 0), atom%basis)
                        pchg = atom%weight*atom%orbitals%wpsir0(k, 2)*pv*pv/afun*100._dp
                        WRITE (iunit, '("    s-states"," N=",I5,T36,"Beta Wavefunction at r=0:",T64,F13.6,"[",I2,"]")') &
                           k, pv, NINT(pchg)
                     END DO
                  END IF
               END IF
            END DO
         END DO
         ! energy differences
         IF (n*m > 1) THEN
            WRITE (iunit, *)
            WRITE (iunit, '(" Method/Econf 1 "," Method/Econf 2","      Delta Energy ","      Error Energy ","     Target")')
            DO j1 = 1, SIZE(atom_info, 2)
               DO j2 = 1, SIZE(atom_info, 2)
                  DO i1 = 1, SIZE(atom_info, 1)
                     DO i2 = 1, SIZE(atom_info, 1)
                        IF ((j1 > j2) .OR. (j1 == j2 .AND. i1 >= i2)) CYCLE
                        IF (ABS(dener(1, j1, j1, i1, i1)) < 0.000001_dp) CYCLE
                        IF (ABS(dener(2, j1, j1, i1, i1)) < 0.000001_dp) CYCLE
                        IF (ABS(dener(1, j2, j2, i2, i2)) < 0.000001_dp) CYCLE
                        IF (ABS(dener(2, j2, j2, i2, i2)) < 0.000001_dp) CYCLE
                        de = dener(2, j1, j2, i1, i2)-dener(1, j1, j2, i1, i2)
                        WRITE (iunit, '(i6,i6,i10,i6,5X,F16.6,F19.6,F12.6)') j1, i1, j2, i2, dener(2, j1, j2, i1, i2), de, t_ener
                     END DO
                  END DO
               END DO
            END DO
            WRITE (iunit, *)
         END IF

         WRITE (iunit, '(/," Number of target values reached: ",T69,i5," of ",i3)') INT(SUM(pval(5:8, :, :, :, :))), ntarget
         WRITE (iunit, *)
      END IF

      DEALLOCATE (x, pval, dener)

      IF (ALLOCATED(xtob)) THEN
         DEALLOCATE (xtob)
      END IF

      CALL close_file(unit_number=iw)

   END SUBROUTINE atom_fit_pseudo
! **************************************************************************************************
!> \brief ...
!> \param density ...
!> \param atom ...
!> \param n ...
!> \param aval ...
!> \param cval ...
!> \param co ...
!> \param aerr ...
! **************************************************************************************************
   SUBROUTINE density_fit(density, atom, n, aval, cval, co, aerr)
      TYPE(opgrid_type), POINTER                         :: density
      TYPE(atom_type), POINTER                           :: atom
      INTEGER, INTENT(IN)                                :: n
      REAL(dp), INTENT(IN)                               :: aval, cval
      REAL(dp), DIMENSION(:), INTENT(INOUT)              :: co
      REAL(dp), INTENT(OUT)                              :: aerr

      CHARACTER(len=*), PARAMETER :: routineN = 'density_fit', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, info, ip, ipiv(1000), iq, k, nr
      REAL(dp)                                           :: a, rk, zval
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: den, pe, uval
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: bf, smat, tval

      ALLOCATE (pe(n))
      nr = atom%basis%grid%nr
      ALLOCATE (bf(nr, n), den(nr))
      bf = 0._dp

      DO i = 1, n
         pe(i) = aval*cval**i
         a = pe(i)
         DO k = 1, nr
            rk = atom%basis%grid%rad(k)
            bf(k, i) = EXP(-a*rk**2)
         END DO
      END DO

      ! total charge
      zval = integrate_grid(density%op, atom%basis%grid)

      ! allocate vectors and matrices for overlaps
      ALLOCATE (tval(n+1, 1), uval(n), smat(n+1, n+1))
      DO i = 1, n
         uval(i) = (pi/pe(i))**1.5_dp
         tval(i, 1) = integrate_grid(density%op, bf(:, i), atom%basis%grid)
      END DO
      tval(n+1, 1) = zval

      DO iq = 1, n
         DO ip = 1, n
            smat(ip, iq) = (pi/(pe(ip)+pe(iq)))**1.5_dp
         END DO
      END DO
      smat(1:n, n+1) = uval(1:n)
      smat(n+1, 1:n) = uval(1:n)
      smat(n+1, n+1) = 0._dp

      CALL lapack_sgesv(n+1, 1, smat, n+1, ipiv, tval, n+1, info)
      IF(.NOT.(info == 0))CALL cp__a("atom_fit.F",1099)
      co(1:n) = tval(1:n, 1)

      ! calculate density
      den(:) = 0._dp
      DO i = 1, n
         den(:) = den(:)+co(i)*bf(:, i)
      END DO
      den(:) = den(:)*fourpi
      den(:) = (den(:)-density%op(:))**2
      aerr = SQRT(integrate_grid(den, atom%basis%grid))

      DEALLOCATE (pe, bf, den)

      DEALLOCATE (tval, uval, smat)

   END SUBROUTINE density_fit
! **************************************************************************************************
!> \brief ...
!> \param atom_info ...
!> \param basis ...
!> \param pptype ...
!> \param afun ...
!> \param iw ...
! **************************************************************************************************
   SUBROUTINE basis_fit(atom_info, basis, pptype, afun, iw)
      TYPE(atom_p_type), DIMENSION(:, :), POINTER        :: atom_info
      TYPE(atom_basis_type), POINTER                     :: basis
      LOGICAL, INTENT(IN)                                :: pptype
      REAL(dp), INTENT(OUT)                              :: afun
      INTEGER, INTENT(IN)                                :: iw

      CHARACTER(len=*), PARAMETER :: routineN = 'basis_fit', routineP = moduleN//':'//routineN

      INTEGER                                            :: do_eric, do_erie, im, in, nm, nn, &
                                                            reltyp, zval
      LOGICAL                                            :: eri_c, eri_e
      TYPE(atom_integrals), POINTER                      :: atint
      TYPE(atom_potential_type), POINTER                 :: pot
      TYPE(atom_type), POINTER                           :: atom

      ALLOCATE (atint)

      nn = SIZE(atom_info, 1)
      nm = SIZE(atom_info, 2)

      ! calculate integrals
      NULLIFY (pot)
      zval = 0
      eri_c = .FALSE.
      eri_e = .FALSE.
      DO in = 1, nn
         DO im = 1, nm
            atom => atom_info(in, im)%atom
            IF (pptype .EQV. atom%pp_calc) THEN
               pot => atom%potential
               zval = atom_info(in, im)%atom%z
               reltyp = atom_info(in, im)%atom%relativistic
               do_eric = atom_info(in, im)%atom%coulomb_integral_type
               do_erie = atom_info(in, im)%atom%exchange_integral_type
               IF (do_eric == do_analytic) eri_c = .TRUE.
               IF (do_erie == do_analytic) eri_e = .TRUE.
               EXIT
            END IF
         END DO
         IF (ASSOCIATED(pot)) EXIT
      END DO
      ! general integrals
      CALL atom_int_setup(atint, basis, potential=pot, eri_coulomb=eri_c, eri_exchange=eri_e)
      ! potential
      CALL atom_ppint_setup(atint, basis, potential=pot)
      IF (pptype) THEN
         NULLIFY (atint%tzora, atint%hdkh)
      ELSE
         ! relativistic correction terms
         CALL atom_relint_setup(atint, basis, reltyp, zcore=REAL(zval, dp))
      END IF

      afun = 0._dp

      DO in = 1, nn
         DO im = 1, nm
            atom => atom_info(in, im)%atom
            IF (atom_consistent_method(atom%method_type, atom%state%multiplicity)) THEN
               IF (pptype .EQV. atom%pp_calc) THEN
                  CALL set_atom(atom, basis=basis)
                  CALL set_atom(atom, integrals=atint)
                  CALL calculate_atom(atom, iw)
                  afun = afun+atom%energy%etot*atom%weight
               END IF
            END IF
         END DO
      END DO

      CALL atom_int_release(atint)
      CALL atom_ppint_release(atint)
      CALL atom_relint_release(atint)

      DEALLOCATE (atint)

   END SUBROUTINE basis_fit
! **************************************************************************************************
!> \brief ...
!> \param atom_info ...
!> \param ppot ...
!> \param afun ...
!> \param wtot ...
!> \param pval ...
!> \param dener ...
!> \param wen ...
!> \param ten ...
! **************************************************************************************************
   SUBROUTINE pseudo_fit(atom_info, ppot, afun, wtot, pval, dener, wen, ten)
      TYPE(atom_p_type), DIMENSION(:, :), POINTER        :: atom_info
      TYPE(atom_potential_type), POINTER                 :: ppot
      REAL(dp), INTENT(OUT)                              :: afun
      REAL(dp), INTENT(IN)                               :: wtot
      REAL(dp), DIMENSION(:, :, 0:, :, :), INTENT(INOUT) :: pval
      REAL(dp), DIMENSION(:, :, :, :, :), INTENT(INOUT)  :: dener
      REAL(dp), INTENT(IN)                               :: wen, ten

      CHARACTER(len=*), PARAMETER :: routineN = 'pseudo_fit', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, i1, i2, j, j1, j2, k, l, n, node
      REAL(KIND=dp)                                      :: charge, de, fde, pv, rcov, rcov1, rcov2, &
                                                            tv
      TYPE(atom_integrals), POINTER                      :: pp_int
      TYPE(atom_type), POINTER                           :: atom

      afun = 0._dp
      pval = 0._dp
      dener(1, :, :, :, :) = 0._dp

      pp_int => atom_info(1, 1)%atom%integrals
      CALL atom_ppint_release(pp_int)
      CALL atom_ppint_setup(pp_int, atom_info(1, 1)%atom%basis, potential=ppot)

      DO i = 1, SIZE(atom_info, 1)
         DO j = 1, SIZE(atom_info, 2)
            atom => atom_info(i, j)%atom
            IF (atom_consistent_method(atom%method_type, atom%state%multiplicity)) THEN
               CALL set_atom(atom, integrals=pp_int, potential=ppot)
               CALL calculate_atom(atom, 0, noguess=.TRUE.)
               dener(1, j, j, i, i) = atom%energy%etot
               DO l = 0, atom%state%maxl_calc
                  n = atom%state%maxn_calc(l)
                  DO k = 1, n
                     IF (atom%state%multiplicity == -1) THEN
                        !no spin polarization
                        rcov = atom%orbitals%rcmax(k, l, 1)
                        tv = atom%orbitals%crefene(k, l, 1)
                        de = ABS(atom%orbitals%ener(k, l)-atom%orbitals%refene(k, l, 1))
                        fde = get_error_value(de, tv)
                        IF (fde < 1.e-8) pval(5, k, l, j, i) = 1._dp
                        pv = atom%weight*atom%orbitals%wrefene(k, l, 1)*fde
                        afun = afun+pv
                        pval(1, k, l, j, i) = pv
                        IF (atom%orbitals%wrefchg(k, l, 1) > 0._dp) THEN
                           CALL atom_orbital_charge(charge, atom%orbitals%wfn(:, k, l), rcov, l, atom%basis)
                           de = ABS(charge-atom%orbitals%refchg(k, l, 1))
                           fde = get_error_value(de, 25._dp*tv)
                           IF (fde < 1.e-8) pval(6, k, l, j, i) = 1._dp
                           pv = atom%weight*atom%orbitals%wrefchg(k, l, 1)*fde
                           afun = afun+pv
                           pval(2, k, l, j, i) = pv
                        END IF
                        IF (atom%orbitals%wrefnod(k, l, 1) > 0._dp) THEN
                           CALL atom_orbital_nodes(node, atom%orbitals%wfn(:, k, l), 2._dp*rcov, l, atom%basis)
                           afun = afun+atom%weight*atom%orbitals%wrefnod(k, l, 1)* &
                                  ABS(REAL(node, dp)-atom%orbitals%refnod(k, l, 1))
                        END IF
                        IF (l == 0) THEN
                           IF (atom%orbitals%wpsir0(k, 1) > 0._dp) THEN
                              CALL atom_wfnr0(pv, atom%orbitals%wfn(:, k, l), atom%basis)
                              pv = atom%weight*atom%orbitals%wpsir0(k, 1)*pv*pv
                              afun = afun+pv
                           END IF
                        END IF
                     ELSE
                        !spin polarization
                        rcov1 = atom%orbitals%rcmax(k, l, 1)
                        rcov2 = atom%orbitals%rcmax(k, l, 2)
                        tv = atom%orbitals%crefene(k, l, 1)
                        de = ABS(atom%orbitals%enera(k, l)-atom%orbitals%refene(k, l, 1))
                        fde = get_error_value(de, tv)
                        IF (fde < 1.e-8) pval(5, k, l, j, i) = 1._dp
                        pv = atom%weight*atom%orbitals%wrefene(k, l, 1)*fde
                        afun = afun+pv
                        pval(1, k, l, j, i) = pv
                        tv = atom%orbitals%crefene(k, l, 2)
                        de = ABS(atom%orbitals%enerb(k, l)-atom%orbitals%refene(k, l, 2))
                        fde = get_error_value(de, tv)
                        IF (fde < 1.e-8) pval(7, k, l, j, i) = 1._dp
                        pv = atom%weight*atom%orbitals%wrefene(k, l, 2)*fde
                        afun = afun+pv
                        pval(3, k, l, j, i) = pv
                        IF (atom%orbitals%wrefchg(k, l, 1) > 0._dp) THEN
                           CALL atom_orbital_charge(charge, atom%orbitals%wfna(:, k, l), rcov1, l, atom%basis)
                           de = ABS(charge-atom%orbitals%refchg(k, l, 1))
                           fde = get_error_value(de, 25._dp*tv)
                           IF (fde < 1.e-8) pval(6, k, l, j, i) = 1._dp
                           pv = atom%weight*atom%orbitals%wrefchg(k, l, 1)*fde
                           afun = afun+pv
                           pval(2, k, l, j, i) = pv
                        END IF
                        IF (atom%orbitals%wrefchg(k, l, 2) > 0._dp) THEN
                           CALL atom_orbital_charge(charge, atom%orbitals%wfnb(:, k, l), rcov2, l, atom%basis)
                           de = ABS(charge-atom%orbitals%refchg(k, l, 2))
                           fde = get_error_value(de, 25._dp*tv)
                           IF (fde < 1.e-8) pval(8, k, l, j, i) = 1._dp
                           pv = atom%weight*atom%orbitals%wrefchg(k, l, 2)*fde
                           afun = afun+pv
                           pval(4, k, l, j, i) = pv
                        END IF
                        IF (atom%orbitals%wrefnod(k, l, 1) > 0._dp) THEN
                           CALL atom_orbital_nodes(node, atom%orbitals%wfna(:, k, l), 2._dp*rcov1, l, atom%basis)
                           afun = afun+atom%weight*atom%orbitals%wrefnod(k, l, 1)* &
                                  ABS(REAL(node, dp)-atom%orbitals%refnod(k, l, 1))
                        END IF
                        IF (atom%orbitals%wrefnod(k, l, 2) > 0._dp) THEN
                           CALL atom_orbital_nodes(node, atom%orbitals%wfnb(:, k, l), 2._dp*rcov2, l, atom%basis)
                           afun = afun+atom%weight*atom%orbitals%wrefnod(k, l, 2)* &
                                  ABS(REAL(node, dp)-atom%orbitals%refnod(k, l, 2))
                        END IF
                        IF (l == 0) THEN
                           IF (atom%orbitals%wpsir0(k, 1) > 0._dp) THEN
                              CALL atom_wfnr0(pv, atom%orbitals%wfna(:, k, l), atom%basis)
                              pv = atom%weight*atom%orbitals%wpsir0(k, 1)*pv*pv
                              afun = afun+pv
                           END IF
                           IF (atom%orbitals%wpsir0(k, 2) > 0._dp) THEN
                              CALL atom_wfnr0(pv, atom%orbitals%wfnb(:, k, l), atom%basis)
                              pv = atom%weight*atom%orbitals%wpsir0(k, 2)*pv*pv
                              afun = afun+pv
                           END IF
                        END IF
                     ENDIF
                  END DO
               END DO
            END IF
         END DO
      END DO

      ! energy differences
      DO j1 = 1, SIZE(atom_info, 2)
         DO j2 = 1, SIZE(atom_info, 2)
            DO i1 = 1, SIZE(atom_info, 1)
               DO i2 = i1+1, SIZE(atom_info, 1)
                  IF ((j1 > j2) .OR. (j1 == j2 .AND. i1 >= i2)) CYCLE
                  dener(1, j1, j2, i1, i2) = dener(1, j1, j1, i1, i1)-dener(1, j2, j2, i2, i2)
                  de = ABS(dener(2, j1, j2, i1, i2)-dener(1, j1, j2, i1, i2))
                  fde = get_error_value(de, ten)
                  afun = afun+wen*fde
               END DO
            END DO
         END DO
      END DO

      ! scaling
      afun = afun/wtot

   END SUBROUTINE pseudo_fit
! **************************************************************************************************
!> \brief ...
!> \param fval ...
!> \param ftarget ...
!> \retval errval ...
! **************************************************************************************************
   FUNCTION get_error_value(fval, ftarget) RESULT(errval)
      REAL(KIND=dp), INTENT(in)                          :: fval, ftarget
      REAL(KIND=dp)                                      :: errval

      IF(.NOT.(fval >= 0.0_dp))CALL cp__a("atom_fit.F",1371)

      IF (fval <= ftarget) THEN
         errval = 0.0_dp
      ELSE
         errval = -(LOG(fval)-LOG(ftarget))/LOG(ftarget)
         errval = SQRT(errval)
      END IF

   END FUNCTION get_error_value
! **************************************************************************************************
!> \brief ...
!> \param pvec ...
!> \param nval ...
!> \param gthpot ...
! **************************************************************************************************
   SUBROUTINE get_pseudo_param(pvec, nval, gthpot)
      REAL(KIND=dp), DIMENSION(:), INTENT(out)           :: pvec
      INTEGER, INTENT(out)                               :: nval
      TYPE(atom_gthpot_type), INTENT(in)                 :: gthpot

      CHARACTER(len=*), PARAMETER :: routineN = 'get_pseudo_param', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ival, j, l, n

      IF (gthpot%lsdpot) THEN
         pvec = 0
         ival = 0
         DO j = 1, gthpot%nexp_lsd
            ival = ival+1
            pvec(ival) = rcpro(-1, gthpot%alpha_lsd(j))
            DO i = 1, gthpot%nct_lsd(j)
               ival = ival+1
               pvec(ival) = gthpot%cval_lsd(i, j)
            END DO
         END DO
      ELSE
         pvec = 0
         ival = 1
         pvec(ival) = rcpro(-1, gthpot%rc)
         DO i = 1, gthpot%ncl
            ival = ival+1
            pvec(ival) = gthpot%cl(i)
         END DO
         IF (gthpot%lpotextended) THEN
            DO j = 1, gthpot%nexp_lpot
               ival = ival+1
               pvec(ival) = rcpro(-1, gthpot%alpha_lpot(j))
               DO i = 1, gthpot%nct_lpot(j)
                  ival = ival+1
                  pvec(ival) = gthpot%cval_lpot(i, j)
               END DO
            END DO
         END IF
         IF (gthpot%nlcc) THEN
            DO j = 1, gthpot%nexp_nlcc
               ival = ival+1
               pvec(ival) = rcpro(-1, gthpot%alpha_nlcc(j))
               DO i = 1, gthpot%nct_nlcc(j)
                  ival = ival+1
                  pvec(ival) = gthpot%cval_nlcc(i, j)
               END DO
            END DO
         END IF
         DO l = 0, 3
            IF (gthpot%nl(l) > 0) THEN
               ival = ival+1
               pvec(ival) = rcpro(-1, gthpot%rcnl(l))
            END IF
         END DO
         DO l = 0, 3
            IF (gthpot%nl(l) > 0) THEN
               n = gthpot%nl(l)
               DO i = 1, n
                  DO j = i, n
                     ival = ival+1
                     pvec(ival) = gthpot%hnl(i, j, l)
                  END DO
               END DO
            END IF
         END DO
      END IF
      nval = ival

   END SUBROUTINE get_pseudo_param

! **************************************************************************************************
!> \brief ...
!> \param pvec ...
!> \param gthpot ...
! **************************************************************************************************
   SUBROUTINE put_pseudo_param(pvec, gthpot)
      REAL(KIND=dp), DIMENSION(:), INTENT(in)            :: pvec
      TYPE(atom_gthpot_type), INTENT(inout)              :: gthpot

      CHARACTER(len=*), PARAMETER :: routineN = 'put_pseudo_param', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ival, j, l, n

      IF (gthpot%lsdpot) THEN
         ival = 0
         DO j = 1, gthpot%nexp_lsd
            ival = ival+1
            gthpot%alpha_lsd(j) = rcpro(1, pvec(ival))
            DO i = 1, gthpot%nct_lsd(j)
               ival = ival+1
               gthpot%cval_lsd(i, j) = pvec(ival)
            END DO
         END DO
      ELSE
         ival = 1
         gthpot%rc = rcpro(1, pvec(ival))
         DO i = 1, gthpot%ncl
            ival = ival+1
            gthpot%cl(i) = pvec(ival)
         END DO
         IF (gthpot%lpotextended) THEN
            DO j = 1, gthpot%nexp_lpot
               ival = ival+1
               gthpot%alpha_lpot(j) = rcpro(1, pvec(ival))
               DO i = 1, gthpot%nct_lpot(j)
                  ival = ival+1
                  gthpot%cval_lpot(i, j) = pvec(ival)
               END DO
            END DO
         END IF
         IF (gthpot%nlcc) THEN
            DO j = 1, gthpot%nexp_nlcc
               ival = ival+1
               gthpot%alpha_nlcc(j) = rcpro(1, pvec(ival))
               DO i = 1, gthpot%nct_nlcc(j)
                  ival = ival+1
                  gthpot%cval_nlcc(i, j) = pvec(ival)
               END DO
            END DO
         END IF
         DO l = 0, 3
            IF (gthpot%nl(l) > 0) THEN
               ival = ival+1
               gthpot%rcnl(l) = rcpro(1, pvec(ival))
            END IF
         END DO
         DO l = 0, 3
            IF (gthpot%nl(l) > 0) THEN
               n = gthpot%nl(l)
               DO i = 1, n
                  DO j = i, n
                     ival = ival+1
                     gthpot%hnl(i, j, l) = pvec(ival)
                  END DO
               END DO
            END IF
         END DO
      END IF

   END SUBROUTINE put_pseudo_param

! **************************************************************************************************
!> \brief ...
!> \param id ...
!> \param xval ...
!> \retval yval ...
! **************************************************************************************************
   FUNCTION rcpro(id, xval) RESULT(yval)
      INTEGER, INTENT(IN)                                :: id
      REAL(dp), INTENT(IN)                               :: xval
      REAL(dp)                                           :: yval

      CHARACTER(len=*), PARAMETER :: routineN = 'rcpro', routineP = moduleN//':'//routineN

      REAL(dp)                                           :: x1, x2

      IF (id == 1) THEN
         yval = 2.0_dp*TANH(0.1_dp*xval)**2
      ELSE IF (id == -1) THEN
         x1 = SQRT(xval/2.0_dp)
         IF(.NOT.(x1 <= 1._dp))CALL cp__a("atom_fit.F",1549)
         x2 = 0.5_dp*LOG((1._dp+x1)/(1._dp-x1))
         yval = x2/0.1_dp
      ELSE
         CALL cp__b("atom_fit.F",1553,"wrong id")
      END IF
   END FUNCTION rcpro

! **************************************************************************************************
!> \brief ...
!> \param atom ...
!> \param num_gau ...
!> \param num_pol ...
!> \param iunit ...
!> \param powell_section ...
!> \param results ...
! **************************************************************************************************
   SUBROUTINE atom_fit_kgpot(atom, num_gau, num_pol, iunit, powell_section, results)
      TYPE(atom_type), POINTER                           :: atom
      INTEGER, INTENT(IN)                                :: num_gau, num_pol, iunit
      TYPE(section_vals_type), OPTIONAL, POINTER         :: powell_section
      REAL(KIND=dp), DIMENSION(:), OPTIONAL              :: results

      CHARACTER(len=*), PARAMETER :: routineN = 'atom_fit_kgpot', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: t23 = 2._dp/3._dp

      INTEGER                                            :: i, ig, ip, iw, j, n10
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: x
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: co
      TYPE(opgrid_type), POINTER                         :: density
      TYPE(opt_state_type)                               :: ostate

! at least one parameter to be optimized

      IF(.NOT.(num_pol*num_gau > 0))CALL cp__a("atom_fit.F",1583)

      ALLOCATE (co(num_pol+1, num_gau), x(num_pol*num_gau+num_gau))
      co = 0._dp

      ! calculate density
      NULLIFY (density)
      CALL create_opgrid(density, atom%basis%grid)
      CALL atom_denmat(atom%orbitals%pmat, atom%orbitals%wfn, atom%basis%nbas, atom%state%occupation, &
                       atom%state%maxl_occ, atom%state%maxn_occ)
      CALL atom_density(density%op, atom%orbitals%pmat, atom%basis, atom%state%maxl_occ, typ="RHO")
      ! target functional
      density%op = t23*(0.3_dp*(3.0_dp*pi*pi)**t23)*density%op**t23

      ! initiallize parameter
      ostate%nf = 0
      ostate%nvar = num_pol*num_gau+num_gau
      DO i = 1, num_gau
         co(1, i) = 0.5_dp+REAL(i-1, KIND=dp)
         co(2, i) = 1.0_dp
         DO j = 3, num_pol+1
            co(j, i) = 0.1_dp
         END DO
      END DO
      CALL putvar(x, co, num_pol, num_gau)

      IF (PRESENT(powell_section)) THEN
         CALL section_vals_val_get(powell_section, "ACCURACY", r_val=ostate%rhoend)
         CALL section_vals_val_get(powell_section, "STEP_SIZE", r_val=ostate%rhobeg)
         CALL section_vals_val_get(powell_section, "MAX_FUN", i_val=ostate%maxfun)
      ELSE
         ostate%rhoend = 1.e-8_dp
         ostate%rhobeg = 5.e-2_dp
         ostate%maxfun = 1000
      END IF
      ostate%iprint = 1
      ostate%unit = iunit

      ostate%state = 0
      IF (iunit > 0) THEN
         WRITE (iunit, '(/," ",13("*")," Approximated Nonadditive Kinetic Energy Functional ",14("*"))')
         WRITE (iunit, '(" POWELL| Start optimization procedure")')
      END IF
      n10 = 50

      DO

         IF (ostate%state == 2) THEN
            CALL getvar(x, co, num_pol, num_gau)
            CALL kgpot_fit(density, num_gau, num_pol, co, ostate%f)
         END IF

         IF (ostate%state == -1) EXIT

         CALL powell_optimize(ostate%nvar, x, ostate)

         IF (MOD(ostate%nf, n10) == 0 .AND. iunit > 0) THEN
            WRITE (iunit, '(" POWELL| Reached",i4,"% of maximal function calls",T66,G15.7)') &
               INT(REAL(ostate%nf, dp)/REAL(ostate%maxfun, dp)*100._dp), ostate%fopt
         END IF

      END DO

      ostate%state = 8
      CALL powell_optimize(ostate%nvar, x, ostate)
      CALL getvar(x, co, num_pol, num_gau)

      CALL release_opgrid(density)

      IF (iunit > 0) THEN
         WRITE (iunit, '(" POWELL| Number of function evaluations",T71,I10)') ostate%nf
         WRITE (iunit, '(" POWELL| Final value of function",T61,G20.10)') ostate%fopt
         WRITE (iunit, '(" Optimized local potential of approximated nonadditive kinetic energy functional")')
         DO ig = 1, num_gau
            WRITE (iunit, '(I2,T15,"Gaussian polynomial expansion",T66,"Rc=",F12.4)') ig, co(1, ig)
            WRITE (iunit, '(T15,"Coefficients",T33,4F12.4)') (co(1+ip, ig), ip=1, num_pol)
         END DO
      END IF

      CALL open_file(file_name="/mnt/sdcard/cp2k/work/FIT_RESULT", file_status="UNKNOWN", file_action="WRITE", unit_number=iw)
      WRITE (iw, *) ptable(atom%z)%symbol
      WRITE (iw, *) num_gau, num_pol
      DO ig = 1, num_gau
         WRITE (iw, '(T10,F12.4,6X,4F12.4)') (co(ip, ig), ip=1, num_pol+1)
      END DO
      CALL close_file(unit_number=iw)

      IF (PRESENT(results)) THEN
         IF(.NOT.(SIZE(results) >= SIZE(x)))CALL cp__a("atom_fit.F",1671)
         results = x
      END IF

      DEALLOCATE (co, x)

   END SUBROUTINE atom_fit_kgpot

! **************************************************************************************************
!> \brief ...
!> \param kgpot ...
!> \param ng ...
!> \param np ...
!> \param cval ...
!> \param aerr ...
! **************************************************************************************************
   SUBROUTINE kgpot_fit(kgpot, ng, np, cval, aerr)
      TYPE(opgrid_type), POINTER                         :: kgpot
      INTEGER, INTENT(IN)                                :: ng, np
      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: cval
      REAL(dp), INTENT(OUT)                              :: aerr

      CHARACTER(len=*), PARAMETER :: routineN = 'kgpot_fit', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ig, ip, n
      REAL(KIND=dp)                                      :: pc, rc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: pval

      n = kgpot%grid%nr
      ALLOCATE (pval(n))
      pval = 0.0_dp
      DO i = 1, n
         DO ig = 1, ng
            rc = kgpot%grid%rad(i)/cval(1, ig)
            pc = 0.0_dp
            DO ip = 1, np
               pc = pc+cval(ip+1, ig)*rc**(2*ip-2)
            END DO
            pval(i) = pval(i)+pc*EXP(-0.5_dp*rc*rc)
         END DO
      END DO
      pval(1:n) = (pval(1:n)-kgpot%op(1:n))**2
      aerr = fourpi*SUM(pval(1:n)*kgpot%grid%wr(1:n))

      DEALLOCATE (pval)

   END SUBROUTINE kgpot_fit

! **************************************************************************************************
!> \brief ...
!> \param xvar ...
!> \param cvar ...
!> \param np ...
!> \param ng ...
! **************************************************************************************************
   SUBROUTINE getvar(xvar, cvar, np, ng)
      REAL(KIND=dp), DIMENSION(:), INTENT(in)            :: xvar
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: cvar
      INTEGER                                            :: np, ng

      INTEGER                                            :: ig, ii, ip

      ii = 0
      DO ig = 1, ng
         ii = ii+1
         cvar(1, ig) = xvar(ii)
         DO ip = 1, np
            ii = ii+1
            cvar(ip+1, ig) = xvar(ii)**2
         END DO
      END DO

   END SUBROUTINE getvar

! **************************************************************************************************
!> \brief ...
!> \param xvar ...
!> \param cvar ...
!> \param np ...
!> \param ng ...
! **************************************************************************************************
   SUBROUTINE putvar(xvar, cvar, np, ng)
      REAL(KIND=dp), DIMENSION(:), INTENT(inout)         :: xvar
      REAL(KIND=dp), DIMENSION(:, :), INTENT(in)         :: cvar
      INTEGER                                            :: np, ng

      INTEGER                                            :: ig, ii, ip

      ii = 0
      DO ig = 1, ng
         ii = ii+1
         xvar(ii) = cvar(1, ig)
         DO ip = 1, np
            ii = ii+1
            xvar(ii) = SQRT(ABS(cvar(ip+1, ig)))
         END DO
      END DO

   END SUBROUTINE putvar
! **************************************************************************************************

END MODULE atom_fit
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief calculate the orbitals for a given atomic kind type
! **************************************************************************************************
MODULE atom_kind_orbitals
   USE ai_onecenter,                    ONLY: sg_erfc
   USE atom_electronic_structure,       ONLY: calculate_atom
   USE atom_fit,                        ONLY: atom_fit_density
   USE atom_operators,                  ONLY: atom_int_release,&
                                              atom_int_setup,&
                                              atom_ppint_release,&
                                              atom_ppint_setup,&
                                              atom_relint_release,&
                                              atom_relint_setup
   USE atom_types,                      ONLY: &
        CGTO_BASIS, Clementi_geobas, GTO_BASIS, atom_basis_type, atom_integrals, atom_orbitals, &
        atom_potential_type, atom_type, create_atom_orbs, create_atom_type, release_atom_basis, &
        release_atom_potential, release_atom_type, set_atom
   USE atom_utils,                      ONLY: get_maxl_occ,&
                                              get_maxn_occ
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE external_potential_types,        ONLY: all_potential_type,&
                                              get_potential,&
                                              gth_potential_type
   USE input_constants,                 ONLY: &
        barrier_conf, do_analytic, do_dkh0_atom, do_dkh1_atom, do_dkh2_atom, do_dkh3_atom, &
        do_dkh4_atom, do_dkh5_atom, do_gapw_log, do_nonrel_atom, do_numeric, do_rks_atom, &
        do_sczoramp_atom, do_zoramp_atom, gth_pseudo, no_pseudo, poly_conf, rel_dkh, rel_none, &
        rel_sczora_mp, rel_zora, rel_zora_full, rel_zora_mp
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: dfac,&
                                              pi
   USE periodic_table,                  ONLY: ptable
   USE physcon,                         ONLY: bohr
   USE qs_grid_atom,                    ONLY: allocate_grid_atom,&
                                              create_grid_atom,&
                                              grid_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              init_atom_electronic_state,&
                                              qs_kind_type
   USE rel_control_types,               ONLY: rel_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'atom_kind_orbitals'

   PUBLIC :: calculate_atomic_orbitals, calculate_atomic_density, &
             calculate_atomic_relkin

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param atomic_kind ...
!> \param qs_kind ...
!> \param iunit ...
!> \param pmat ...
!> \param fmat ...
!> \param ispin ...
!> \param confine ...
!> \param xc_section ...
! **************************************************************************************************
   SUBROUTINE calculate_atomic_orbitals(atomic_kind, qs_kind, iunit, pmat, fmat, &
                                        ispin, confine, xc_section)
      TYPE(atomic_kind_type), INTENT(IN)                 :: atomic_kind
      TYPE(qs_kind_type), INTENT(IN)                     :: qs_kind
      INTEGER, INTENT(IN), OPTIONAL                      :: iunit
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: pmat, fmat
      INTEGER, INTENT(IN)                                :: ispin
      LOGICAL, INTENT(IN), OPTIONAL                      :: confine
      TYPE(section_vals_type), OPTIONAL, POINTER         :: xc_section

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_atomic_orbitals', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, ipgf, j, k, k1, k2, l, ll, lm, m, &
                                                            mb, mo, n, ne, nexp_lpot, nexp_lsd, &
                                                            nexp_nlcc, ngp, nj, nr, ns, nset, &
                                                            nsgf, quadtype, z
      INTEGER, DIMENSION(0:3, 10)                        :: ncalc, ncore, nelem
      INTEGER, DIMENSION(0:3, 100)                       :: set_index, shell_index
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, nct_lpot, nct_lsd, nct_nlcc, &
                                                            npgf, nppnl, nshell, ppeconf
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgf, last_sgf, ls
      LOGICAL                                            :: lpot_present, lsd_present, nlcc_present
      REAL(KIND=dp)                                      :: ac, al, ear, rk, scal, zeff
      REAL(KIND=dp), DIMENSION(:), POINTER               :: alpha_lpot, alpha_lsd, alpha_nlcc, ap, ce
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cval_lpot, cval_lsd, cval_nlcc, zet
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc, hp
      TYPE(all_potential_type), POINTER                  :: all_potential
      TYPE(atom_basis_type), POINTER                     :: basis
      TYPE(atom_integrals), POINTER                      :: integrals
      TYPE(atom_orbitals), POINTER                       :: orbitals
      TYPE(atom_potential_type), POINTER                 :: potential
      TYPE(atom_type), POINTER                           :: atom
      TYPE(grid_atom_type), POINTER                      :: grid
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set

      NULLIFY (atom)
      CALL create_atom_type(atom)

      IF (PRESENT(xc_section)) THEN
         atom%xc_section => xc_section
      ELSE
         NULLIFY (atom%xc_section)
      END IF

      CALL get_atomic_kind(atomic_kind, z=z)
      NULLIFY (all_potential, gth_potential, orb_basis_set)
      CALL get_qs_kind(qs_kind, zeff=zeff, &
                       basis_set=orb_basis_set, &
                       all_potential=all_potential, &
                       gth_potential=gth_potential)

      IF(.NOT.(ASSOCIATED(orb_basis_set)))CALL cp__a("atom_kind_orbitals.F",129)

      atom%z = z
      CALL set_atom(atom, &
                    pp_calc=ASSOCIATED(gth_potential), &
                    do_zmp=.FALSE., &
                    doread=.FALSE., &
                    read_vxc=.FALSE., &
                    method_type=do_rks_atom, &
                    relativistic=do_nonrel_atom, &
                    coulomb_integral_type=do_numeric, &
                    exchange_integral_type=do_numeric)

      ALLOCATE (potential, basis, integrals)

      IF (ASSOCIATED(gth_potential)) THEN
         potential%ppot_type = gth_pseudo
         IF (PRESENT(confine)) THEN
            potential%confinement = confine
         ELSE
            potential%confinement = .TRUE.
         END IF
         potential%conf_type = poly_conf
         potential%acon = 0.1_dp
         potential%rcon = 2.0_dp*ptable(z)%vdw_radius*bohr
         potential%scon = 2.0_dp

         CALL get_potential(gth_potential, &
                            zeff=zeff, &
                            elec_conf=ppeconf, &
                            alpha_core_charge=ac, &
                            nexp_ppl=ne, &
                            cexp_ppl=ce, &
                            lppnl=lm, &
                            nprj_ppnl=nppnl, &
                            alpha_ppnl=ap, &
                            hprj_ppnl=hp)

         potential%gth_pot%zion = zeff
         potential%gth_pot%rc = SQRT(0.5_dp/ac)
         potential%gth_pot%ncl = ne
         potential%gth_pot%cl(:) = 0._dp
         IF (ac > 0._dp) THEN
            DO i = 1, ne
               potential%gth_pot%cl(i) = ce(i)/(2._dp*ac)**(i-1)
            END DO
         END IF
         !extended type
         potential%gth_pot%lpotextended = .FALSE.
         potential%gth_pot%lsdpot = .FALSE.
         potential%gth_pot%nlcc = .FALSE.
         potential%gth_pot%nexp_lpot = 0
         potential%gth_pot%nexp_lsd = 0
         potential%gth_pot%nexp_nlcc = 0
         CALL get_potential(gth_potential, &
                            lpot_present=lpot_present, &
                            lsd_present=lsd_present, &
                            nlcc_present=nlcc_present)
         IF (lpot_present) THEN
            CALL get_potential(gth_potential, &
                               nexp_lpot=nexp_lpot, &
                               alpha_lpot=alpha_lpot, &
                               nct_lpot=nct_lpot, &
                               cval_lpot=cval_lpot)
            potential%gth_pot%lpotextended = .TRUE.
            potential%gth_pot%nexp_lpot = nexp_lpot
            potential%gth_pot%alpha_lpot(1:nexp_lpot) = SQRT(0.5_dp/alpha_lpot(1:nexp_lpot))
            potential%gth_pot%nct_lpot(1:nexp_lpot) = nct_lpot(1:nexp_lpot)
            DO j = 1, nexp_lpot
               ac = alpha_lpot(j)
               DO i = 1, 4
                  potential%gth_pot%cval_lpot(i, j) = cval_lpot(i, j)/(2._dp*ac)**(i-1)
               END DO
            END DO
         END IF
         IF (lsd_present) THEN
            CALL get_potential(gth_potential, &
                               nexp_lsd=nexp_lsd, &
                               alpha_lsd=alpha_lsd, &
                               nct_lsd=nct_lsd, &
                               cval_lsd=cval_lsd)
            potential%gth_pot%lsdpot = .TRUE.
            potential%gth_pot%nexp_lsd = nexp_lsd
            potential%gth_pot%alpha_lsd(1:nexp_lsd) = SQRT(0.5_dp/alpha_lsd(1:nexp_lsd))
            potential%gth_pot%nct_lsd(1:nexp_lsd) = nct_lsd(1:nexp_lsd)
            DO j = 1, nexp_lpot
               ac = alpha_lsd(j)
               DO i = 1, 4
                  potential%gth_pot%cval_lsd(i, j) = cval_lsd(i, j)/(2._dp*ac)**(i-1)
               END DO
            END DO
         END IF

         ! nonlocal part
         potential%gth_pot%nl(:) = 0
         potential%gth_pot%rcnl(:) = 0._dp
         potential%gth_pot%hnl(:, :, :) = 0._dp
         DO l = 0, lm
            n = nppnl(l)
            potential%gth_pot%nl(l) = n
            potential%gth_pot%rcnl(l) = SQRT(0.5_dp/ap(l))
            potential%gth_pot%hnl(1:n, 1:n, l) = hp(1:n, 1:n, l)
         END DO

         IF (nlcc_present) THEN
            CALL get_potential(gth_potential, &
                               nexp_nlcc=nexp_nlcc, &
                               alpha_nlcc=alpha_nlcc, &
                               nct_nlcc=nct_nlcc, &
                               cval_nlcc=cval_nlcc)
            potential%gth_pot%nlcc = .TRUE.
            potential%gth_pot%nexp_nlcc = nexp_nlcc
            potential%gth_pot%alpha_nlcc(1:nexp_nlcc) = alpha_nlcc(1:nexp_nlcc)
            potential%gth_pot%nct_nlcc(1:nexp_nlcc) = nct_nlcc(1:nexp_nlcc)
            potential%gth_pot%cval_nlcc(1:4, 1:nexp_nlcc) = cval_nlcc(1:4, 1:nexp_nlcc)
         END IF

         CALL set_atom(atom, zcore=NINT(zeff), potential=potential)
      ELSE
         potential%ppot_type = no_pseudo
         IF (PRESENT(confine)) THEN
            potential%confinement = confine
         ELSE
            potential%confinement = .FALSE.
         END IF
         potential%acon = 0.1_dp
         potential%rcon = 2.0_dp*ptable(z)%vdw_radius*bohr
         potential%scon = 2.0_dp
         CALL set_atom(atom, zcore=z, potential=potential)
      END IF

      CALL get_gto_basis_set(orb_basis_set, &
                             nset=nset, nshell=nshell, npgf=npgf, lmin=lmin, lmax=lmax, l=ls, nsgf=nsgf, zet=zet, gcc=gcc, &
                             first_sgf=first_sgf, last_sgf=last_sgf)

      NULLIFY (grid)
      ngp = 400
      quadtype = do_gapw_log
      CALL allocate_grid_atom(grid)
      CALL create_grid_atom(grid, ngp, 1, 1, 0, quadtype)
      grid%nr = ngp
      basis%grid => grid

      NULLIFY (basis%am, basis%cm, basis%as, basis%ns, basis%bf, basis%dbf, basis%ddbf)
      basis%basis_type = CGTO_BASIS
      basis%eps_eig = 1.e-12_dp

      ! fill in the basis data structures
      set_index = 0
      shell_index = 0
      basis%nprim = 0
      basis%nbas = 0
      DO i = 1, nset
         DO j = lmin(i), MIN(lmax(i), 3)
            basis%nprim(j) = basis%nprim(j)+npgf(i)
         END DO
         DO j = 1, nshell(i)
            l = ls(j, i)
            IF (l <= 3) THEN
               basis%nbas(l) = basis%nbas(l)+1
               k = basis%nbas(l)
               IF(.NOT.(k <= 100))CALL cp__a("atom_kind_orbitals.F",290)
               set_index(l, k) = i
               shell_index(l, k) = j
            END IF
         END DO
      END DO

      nj = MAXVAL(basis%nprim)
      ns = MAXVAL(basis%nbas)
      ALLOCATE (basis%am(nj, 0:3))
      basis%am = 0._dp
      ALLOCATE (basis%cm(nj, ns, 0:3))
      basis%cm = 0._dp
      DO j = 0, 3
         nj = 0
         ns = 0
         DO i = 1, nset
            IF (j >= lmin(i) .AND. j <= lmax(i)) THEN
               DO ipgf = 1, npgf(i)
                  basis%am(nj+ipgf, j) = zet(ipgf, i)
               END DO
               DO ii = 1, nshell(i)
                  IF (ls(ii, i) == j) THEN
                     ns = ns+1
                     DO ipgf = 1, npgf(i)
                        basis%cm(nj+ipgf, ns, j) = gcc(ipgf, ii, i)
                     END DO
                  END IF
               END DO
               nj = nj+npgf(i)
            END IF
         END DO
      END DO

      ! initialize basis function on a radial grid
      nr = basis%grid%nr
      m = MAXVAL(basis%nbas)
      ALLOCATE (basis%bf(nr, m, 0:3))
      ALLOCATE (basis%dbf(nr, m, 0:3))
      ALLOCATE (basis%ddbf(nr, m, 0:3))

      basis%bf = 0._dp
      basis%dbf = 0._dp
      basis%ddbf = 0._dp
      DO l = 0, 3
         DO i = 1, basis%nprim(l)
            al = basis%am(i, l)
            DO k = 1, nr
               rk = basis%grid%rad(k)
               ear = EXP(-al*basis%grid%rad(k)**2)
               DO j = 1, basis%nbas(l)
                  basis%bf(k, j, l) = basis%bf(k, j, l)+rk**l*ear*basis%cm(i, j, l)
                  basis%dbf(k, j, l) = basis%dbf(k, j, l) &
                                       +(REAL(l, dp)*rk**(l-1)-2._dp*al*rk**(l+1))*ear*basis%cm(i, j, l)
                  basis%ddbf(k, j, l) = basis%ddbf(k, j, l)+ &
                                        (REAL(l*(l-1), dp)*rk**(l-2)-2._dp*al*REAL(2*l+1, dp)*rk**(l)+4._dp*al*rk**(l+2))* &
                                        ear*basis%cm(i, j, l)
               END DO
            END DO
         END DO
      END DO

      CALL set_atom(atom, basis=basis)

      ! optimization defaults
      atom%optimization%damping = 0.2_dp
      atom%optimization%eps_scf = 1.e-6_dp
      atom%optimization%eps_diis = 100._dp
      atom%optimization%max_iter = 50
      atom%optimization%n_diis = 5

      ! set up the electronic state
      CALL init_atom_electronic_state(atomic_kind=atomic_kind, &
                                      qs_kind=qs_kind, &
                                      ispin=ispin, &
                                      ncalc=ncalc, &
                                      ncore=ncore, &
                                      nelem=nelem)

      ALLOCATE (atom%state)

      atom%state%core = 0._dp
      atom%state%core(0:3, 1:7) = REAL(ncore(0:3, 1:7), dp)
      atom%state%occ = 0._dp
      atom%state%occ(0:3, 1:7) = REAL(ncalc(0:3, 1:7), dp)
      atom%state%occupation = 0._dp
      atom%state%multiplicity = -1
      DO l = 0, 3
         k = 0
         DO i = 1, 7
            IF (ncalc(l, i) > 0) THEN
               k = k+1
               atom%state%occupation(l, k) = REAL(ncalc(l, i), dp)
            END IF
         END DO
      END DO

      atom%state%maxl_occ = get_maxl_occ(atom%state%occupation)
      atom%state%maxn_occ = get_maxn_occ(atom%state%occupation)
      atom%state%maxl_calc = atom%state%maxl_occ
      atom%state%maxn_calc = atom%state%maxn_occ

      ! calculate integrals
      ! general integrals
      CALL atom_int_setup(integrals, basis, potential=atom%potential, &
                          eri_coulomb=(atom%coulomb_integral_type == do_analytic), &
                          eri_exchange=(atom%exchange_integral_type == do_analytic))
      ! potential
      CALL atom_ppint_setup(integrals, basis, potential=atom%potential)
      ! relativistic correction terms
      NULLIFY (integrals%tzora, integrals%hdkh)
      CALL atom_relint_setup(integrals, basis, atom%relativistic, zcore=REAL(atom%zcore, dp))
      CALL set_atom(atom, integrals=integrals)

      NULLIFY (orbitals)
      mo = MAXVAL(atom%state%maxn_calc)
      mb = MAXVAL(atom%basis%nbas)
      CALL create_atom_orbs(orbitals, mb, mo)
      CALL set_atom(atom, orbitals=orbitals)

      IF (PRESENT(iunit)) THEN
         CALL calculate_atom(atom, iunit)
      ELSE
         CALL calculate_atom(atom, -1)
      END IF

      IF (PRESENT(pmat)) THEN
         ! recover density matrix in CP2K/GPW order and normalization
         IF (ASSOCIATED(pmat)) THEN
            DEALLOCATE (pmat)
         END IF
         ALLOCATE (pmat(nsgf, nsgf))
         pmat = 0._dp
         DO l = 0, 3
            ll = 2*l
            DO k1 = 1, atom%basis%nbas(l)
               DO k2 = 1, atom%basis%nbas(l)
                  scal = SQRT(atom%integrals%ovlp(k1, k1, l)*atom%integrals%ovlp(k2, k2, l))/REAL(2*l+1, KIND=dp)
                  i = first_sgf(shell_index(l, k1), set_index(l, k1))
                  j = first_sgf(shell_index(l, k2), set_index(l, k2))
                  DO m = 0, ll
                     pmat(i+m, j+m) = atom%orbitals%pmat(k1, k2, l)*scal
                  END DO
               END DO
            END DO
         ENDDO
      END IF

      IF (PRESENT(fmat)) THEN
         ! recover fock matrix in CP2K/GPW order.
         ! Caution: Normalization is not take care of, so it's probably weird.
         IF (ispin /= 1) CALL cp__b("atom_kind_orbitals.F",441,"calculate_atomic_orbitals: only ispin==1 is implemented")
         IF (ASSOCIATED(fmat)) CALL cp__b("atom_kind_orbitals.F",442,"fmat already associated")
         IF (.NOT. ASSOCIATED(atom%fmat)) CALL cp__b("atom_kind_orbitals.F",443,"atom%fmat not associated")
         ALLOCATE (fmat(nsgf, nsgf))
         fmat = 0.0_dp
         DO l = 0, 3
            ll = 2*l
            DO k1 = 1, atom%basis%nbas(l)
            DO k2 = 1, atom%basis%nbas(l)
               scal = SQRT(atom%integrals%ovlp(k1, k1, l)*atom%integrals%ovlp(k2, k2, l))
               i = first_sgf(shell_index(l, k1), set_index(l, k1))
               j = first_sgf(shell_index(l, k2), set_index(l, k2))
               DO m = 0, ll
                  fmat(i+m, j+m) = atom%fmat%op(k1, k2, l)/scal
               END DO
            END DO
            END DO
         ENDDO
      END IF

      ! clean up
      CALL atom_int_release(integrals)
      CALL atom_ppint_release(integrals)
      CALL atom_relint_release(integrals)
      CALL release_atom_basis(basis)
      CALL release_atom_potential(potential)
      CALL release_atom_type(atom)

      DEALLOCATE (potential, basis, integrals)

   END SUBROUTINE calculate_atomic_orbitals

! **************************************************************************************************
!> \brief ...
!> \param density ...
!> \param atomic_kind ...
!> \param qs_kind ...
!> \param ngto ...
!> \param iunit ...
!> \param allelectron ...
!> \param confine ...
! **************************************************************************************************
   SUBROUTINE calculate_atomic_density(density, atomic_kind, qs_kind, ngto, iunit, allelectron, confine)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: density
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      INTEGER, INTENT(IN)                                :: ngto
      INTEGER, INTENT(IN), OPTIONAL                      :: iunit
      LOGICAL, INTENT(IN), OPTIONAL                      :: allelectron, confine

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_atomic_density', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: num_gto = 40

      INTEGER                                            :: i, ii, j, k, l, ll, lm, m, mb, mo, n, &
                                                            ne, nexp_lpot, nexp_lsd, nexp_nlcc, &
                                                            ngp, nn, nr, quadtype, relativistic, z
      INTEGER, DIMENSION(0:3)                            :: econfx, starti
      INTEGER, DIMENSION(0:3, 10)                        :: ncalc, ncore, nelem
      INTEGER, DIMENSION(:), POINTER                     :: econf, nct_lpot, nct_lsd, nct_nlcc, &
                                                            nppnl, ppeconf
      LOGICAL                                            :: lpot_present, lsd_present, nlcc_present
      REAL(KIND=dp)                                      :: ac, al, aval, cc, cval, ear, rk, xx, zeff
      REAL(KIND=dp), DIMENSION(:), POINTER               :: alpha_lpot, alpha_lsd, alpha_nlcc, ap, ce
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cval_lpot, cval_lsd, cval_nlcc
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: hp
      REAL(KIND=dp), DIMENSION(num_gto+2)                :: results
      TYPE(all_potential_type), POINTER                  :: all_potential
      TYPE(atom_basis_type), POINTER                     :: basis
      TYPE(atom_integrals), POINTER                      :: integrals
      TYPE(atom_orbitals), POINTER                       :: orbitals
      TYPE(atom_potential_type), POINTER                 :: potential
      TYPE(atom_type), POINTER                           :: atom
      TYPE(grid_atom_type), POINTER                      :: grid
      TYPE(gth_potential_type), POINTER                  :: gth_potential

      NULLIFY (atom)
      CALL create_atom_type(atom)

      CALL get_atomic_kind(atomic_kind, z=z)
      NULLIFY (all_potential, gth_potential)
      CALL get_qs_kind(qs_kind, zeff=zeff, &
                       all_potential=all_potential, &
                       gth_potential=gth_potential)

      IF (PRESENT(allelectron)) THEN
         IF (allelectron) THEN
            NULLIFY (gth_potential)
            zeff = z
         END IF
      END IF

      IF(.NOT.(ngto <= num_gto))CALL cp__a("atom_kind_orbitals.F",533)

      IF (ASSOCIATED(gth_potential)) THEN
         ! PP calculation are non-relativistic
         relativistic = do_nonrel_atom
      ELSE
         ! AE calculations use DKH2
         relativistic = do_dkh2_atom
      END IF

      atom%z = z
      CALL set_atom(atom, &
                    pp_calc=ASSOCIATED(gth_potential), &
                    method_type=do_rks_atom, &
                    relativistic=relativistic, &
                    coulomb_integral_type=do_numeric, &
                    exchange_integral_type=do_numeric)

      ALLOCATE (potential, basis, integrals)

      IF (ASSOCIATED(gth_potential)) THEN
         potential%ppot_type = gth_pseudo
         IF (PRESENT(confine)) THEN
            potential%confinement = confine
         ELSE
            potential%confinement = .TRUE.
         END IF
         potential%conf_type = barrier_conf
         potential%acon = 200._dp
         potential%rcon = 4.0_dp
         potential%scon = 8.0_dp

         CALL get_potential(gth_potential, &
                            zeff=zeff, &
                            elec_conf=ppeconf, &
                            alpha_core_charge=ac, &
                            nexp_ppl=ne, &
                            cexp_ppl=ce, &
                            lppnl=lm, &
                            nprj_ppnl=nppnl, &
                            alpha_ppnl=ap, &
                            hprj_ppnl=hp)

         potential%gth_pot%zion = zeff
         potential%gth_pot%rc = SQRT(0.5_dp/ac)
         potential%gth_pot%ncl = ne
         potential%gth_pot%cl(:) = 0._dp
         IF (ac > 0._dp) THEN
            DO i = 1, ne
               potential%gth_pot%cl(i) = ce(i)/(2._dp*ac)**(i-1)
            END DO
         END IF
         !extended type
         potential%gth_pot%lpotextended = .FALSE.
         potential%gth_pot%lsdpot = .FALSE.
         potential%gth_pot%nlcc = .FALSE.
         potential%gth_pot%nexp_lpot = 0
         potential%gth_pot%nexp_lsd = 0
         potential%gth_pot%nexp_nlcc = 0
         CALL get_potential(gth_potential, &
                            lpot_present=lpot_present, &
                            lsd_present=lsd_present, &
                            nlcc_present=nlcc_present)
         IF (lpot_present) THEN
            CALL get_potential(gth_potential, &
                               nexp_lpot=nexp_lpot, &
                               alpha_lpot=alpha_lpot, &
                               nct_lpot=nct_lpot, &
                               cval_lpot=cval_lpot)
            potential%gth_pot%lpotextended = .TRUE.
            potential%gth_pot%nexp_lpot = nexp_lpot
            potential%gth_pot%alpha_lpot(1:nexp_lpot) = SQRT(0.5_dp/alpha_lpot(1:nexp_lpot))
            potential%gth_pot%nct_lpot(1:nexp_lpot) = nct_lpot(1:nexp_lpot)
            DO j = 1, nexp_lpot
               ac = alpha_lpot(j)
               DO i = 1, 4
                  potential%gth_pot%cval_lpot(i, j) = cval_lpot(i, j)/(2._dp*ac)**(i-1)
               END DO
            END DO
         END IF
         IF (lsd_present) THEN
            CALL get_potential(gth_potential, &
                               nexp_lsd=nexp_lsd, &
                               alpha_lsd=alpha_lsd, &
                               nct_lsd=nct_lsd, &
                               cval_lsd=cval_lsd)
            potential%gth_pot%lsdpot = .TRUE.
            potential%gth_pot%nexp_lsd = nexp_lsd
            potential%gth_pot%alpha_lsd(1:nexp_lsd) = SQRT(0.5_dp/alpha_lsd(1:nexp_lsd))
            potential%gth_pot%nct_lsd(1:nexp_lsd) = nct_lsd(1:nexp_lsd)
            DO j = 1, nexp_lpot
               ac = alpha_lsd(j)
               DO i = 1, 4
                  potential%gth_pot%cval_lsd(i, j) = cval_lsd(i, j)/(2._dp*ac)**(i-1)
               END DO
            END DO
         END IF

         ! nonlocal part
         potential%gth_pot%nl(:) = 0
         potential%gth_pot%rcnl(:) = 0._dp
         potential%gth_pot%hnl(:, :, :) = 0._dp
         DO l = 0, lm
            n = nppnl(l)
            potential%gth_pot%nl(l) = n
            potential%gth_pot%rcnl(l) = SQRT(0.5_dp/ap(l))
            potential%gth_pot%hnl(1:n, 1:n, l) = hp(1:n, 1:n, l)
         END DO

         IF (nlcc_present) THEN
            CALL get_potential(gth_potential, &
                               nexp_nlcc=nexp_nlcc, &
                               alpha_nlcc=alpha_nlcc, &
                               nct_nlcc=nct_nlcc, &
                               cval_nlcc=cval_nlcc)
            potential%gth_pot%nlcc = .TRUE.
            potential%gth_pot%nexp_nlcc = nexp_nlcc
            potential%gth_pot%alpha_nlcc(1:nexp_nlcc) = alpha_nlcc(1:nexp_nlcc)
            potential%gth_pot%nct_nlcc(1:nexp_nlcc) = nct_nlcc(1:nexp_nlcc)
            potential%gth_pot%cval_nlcc(1:4, 1:nexp_nlcc) = cval_nlcc(1:4, 1:nexp_nlcc)
         END IF

         CALL set_atom(atom, zcore=NINT(zeff), potential=potential)
      ELSE
         potential%ppot_type = no_pseudo
         IF (PRESENT(confine)) THEN
            potential%confinement = confine
         ELSE
            potential%confinement = .FALSE.
         END IF
         potential%conf_type = barrier_conf
         potential%acon = 200._dp
         potential%rcon = 4.0_dp
         potential%scon = 8.0_dp
         CALL set_atom(atom, zcore=z, potential=potential)
         CALL set_atom(atom, zcore=z, potential=potential)
      END IF

      ! atomic grid
      NULLIFY (grid)
      ngp = 400
      quadtype = do_gapw_log
      CALL allocate_grid_atom(grid)
      CALL create_grid_atom(grid, ngp, 1, 1, 0, quadtype)
      grid%nr = ngp
      basis%grid => grid

      NULLIFY (basis%am, basis%cm, basis%as, basis%ns, basis%bf, basis%dbf, basis%ddbf)

      ! fill in the basis data structures
      basis%eps_eig = 1.e-12_dp
      basis%basis_type = GTO_BASIS
      CALL Clementi_geobas(z, cval, aval, basis%nbas, starti)
      basis%nprim = basis%nbas
      m = MAXVAL(basis%nbas)
      ALLOCATE (basis%am(m, 0:3))
      basis%am = 0._dp
      DO l = 0, 3
         DO i = 1, basis%nbas(l)
            ll = i-1+starti(l)
            basis%am(i, l) = aval*cval**(ll)
         END DO
      END DO

      basis%geometrical = .TRUE.
      basis%aval = aval
      basis%cval = cval
      basis%start = starti

      ! initialize basis function on a radial grid
      nr = basis%grid%nr
      m = MAXVAL(basis%nbas)
      ALLOCATE (basis%bf(nr, m, 0:3))
      ALLOCATE (basis%dbf(nr, m, 0:3))
      ALLOCATE (basis%ddbf(nr, m, 0:3))
      basis%bf = 0._dp
      basis%dbf = 0._dp
      basis%ddbf = 0._dp
      DO l = 0, 3
         DO i = 1, basis%nbas(l)
            al = basis%am(i, l)
            DO k = 1, nr
               rk = basis%grid%rad(k)
               ear = EXP(-al*basis%grid%rad(k)**2)
               basis%bf(k, i, l) = rk**l*ear
               basis%dbf(k, i, l) = (REAL(l, dp)*rk**(l-1)-2._dp*al*rk**(l+1))*ear
               basis%ddbf(k, i, l) = (REAL(l*(l-1), dp)*rk**(l-2)- &
                                      2._dp*al*REAL(2*l+1, dp)*rk**(l)+4._dp*al*rk**(l+2))*ear
            END DO
         END DO
      END DO

      CALL set_atom(atom, basis=basis)

      ! optimization defaults
      atom%optimization%damping = 0.2_dp
      atom%optimization%eps_scf = 1.e-6_dp
      atom%optimization%eps_diis = 100._dp
      atom%optimization%max_iter = 50
      atom%optimization%n_diis = 5

      nelem = 0
      ncore = 0
      ncalc = 0
      NULLIFY (econf)
      IF (ASSOCIATED(gth_potential)) THEN
         CALL get_potential(gth_potential, elec_conf=econf)
         econfx = 0
         econfx(0:SIZE(econf)-1) = econf
         IF (SUM(econf) >= 0) THEN
            DO l = 0, 3
               ll = 2*(2*l+1)
               nn = ptable(z)%e_conv(l)-econfx(l)
               ii = 0
               DO
                  ii = ii+1
                  IF (nn <= ll) THEN
                     ncore(l, ii) = nn
                     EXIT
                  ELSE
                     ncore(l, ii) = ll
                     nn = nn-ll
                  END IF
               END DO
            END DO
            DO l = 0, 3
               ll = 2*(2*l+1)
               nn = ptable(z)%e_conv(l)
               ii = 0
               DO
                  ii = ii+1
                  IF (nn <= ll) THEN
                     nelem(l, ii) = nn
                     EXIT
                  ELSE
                     nelem(l, ii) = ll
                     nn = nn-ll
                  END IF
               END DO
            END DO
            ncalc = nelem-ncore
         ELSE
            ncore = 0
            ncalc = 0
            DO l = 0, 3
               ll = 2*(2*l+1)
               nn = ABS(econfx(l))
               ii = 0
               DO
                  ii = ii+1
                  IF (nn <= ll) THEN
                     ncalc(l, ii) = -nn
                     EXIT
                  ELSE
                     ncalc(l, ii) = -ll
                     nn = nn-ll
                  END IF
               END DO
            END DO
         END IF
      ELSE
         DO l = 0, 3
            ll = 2*(2*l+1)
            nn = ptable(z)%e_conv(l)
            ii = 0
            DO
               ii = ii+1
               IF (nn <= ll) THEN
                  nelem(l, ii) = nn
                  EXIT
               ELSE
                  nelem(l, ii) = ll
                  nn = nn-ll
               END IF
            END DO
         END DO
         ncalc = nelem-ncore
      END IF

      IF (qs_kind%ghost) THEN
         nelem = 0
         ncore = 0
         ncalc = 0
      END IF

      ALLOCATE (atom%state)

      atom%state%core = 0._dp
      atom%state%core(0:3, 1:7) = REAL(ncore(0:3, 1:7), dp)
      atom%state%occ = 0._dp
      atom%state%occ(0:3, 1:7) = REAL(ncalc(0:3, 1:7), dp)
      atom%state%occupation = 0._dp
      atom%state%multiplicity = -1
      DO l = 0, 3
         k = 0
         DO i = 1, 7
            IF (ncalc(l, i) > 0) THEN
               k = k+1
               atom%state%occupation(l, k) = REAL(ncalc(l, i), dp)
            END IF
         END DO
      END DO

      atom%state%maxl_occ = get_maxl_occ(atom%state%occupation)
      atom%state%maxn_occ = get_maxn_occ(atom%state%occupation)
      atom%state%maxl_calc = atom%state%maxl_occ
      atom%state%maxn_calc = atom%state%maxn_occ

      ! calculate integrals
      ! general integrals
      CALL atom_int_setup(integrals, basis, potential=atom%potential, &
                          eri_coulomb=(atom%coulomb_integral_type == do_analytic), &
                          eri_exchange=(atom%exchange_integral_type == do_analytic))
      ! potential
      CALL atom_ppint_setup(integrals, basis, potential=atom%potential)
      ! relativistic correction terms
      NULLIFY (integrals%tzora, integrals%hdkh)
      CALL atom_relint_setup(integrals, basis, atom%relativistic, zcore=REAL(atom%zcore, dp))
      CALL set_atom(atom, integrals=integrals)

      NULLIFY (orbitals)
      mo = MAXVAL(atom%state%maxn_calc)
      mb = MAXVAL(atom%basis%nbas)
      CALL create_atom_orbs(orbitals, mb, mo)
      CALL set_atom(atom, orbitals=orbitals)

      IF (PRESENT(iunit)) THEN
         CALL calculate_atom(atom, iunit)
         CALL atom_fit_density(atom, ngto, 0, iunit, results=results)
      ELSE
         CALL calculate_atom(atom, -1)
         CALL atom_fit_density(atom, ngto, 0, -1, results=results)
      END IF

      xx = results(1)
      cc = results(2)
      DO i = 1, ngto
         density(i, 1) = xx*cc**i
         density(i, 2) = results(2+i)
      END DO

      ! clean up
      CALL atom_int_release(integrals)
      CALL atom_ppint_release(integrals)
      CALL atom_relint_release(integrals)
      CALL release_atom_basis(basis)
      CALL release_atom_potential(potential)
      CALL release_atom_type(atom)

      DEALLOCATE (potential, basis, integrals)

   END SUBROUTINE calculate_atomic_density

! **************************************************************************************************
!> \brief ...
!> \param atomic_kind ...
!> \param qs_kind ...
!> \param rel_control ...
!> \param rtmat ...
! **************************************************************************************************
   SUBROUTINE calculate_atomic_relkin(atomic_kind, qs_kind, rel_control, rtmat)
      TYPE(atomic_kind_type), INTENT(IN)                 :: atomic_kind
      TYPE(qs_kind_type), INTENT(IN)                     :: qs_kind
      TYPE(rel_control_type), POINTER                    :: rel_control
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rtmat

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_atomic_relkin', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ii, ipgf, j, k, k1, k2, l, ll, m, n, &
                                                            ngp, nj, nn, nr, ns, nset, nsgf, &
                                                            quadtype, relativistic, z
      INTEGER, DIMENSION(0:3, 10)                        :: ncalc, ncore, nelem
      INTEGER, DIMENSION(0:3, 100)                       :: set_index, shell_index
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgf, last_sgf, ls
      REAL(KIND=dp)                                      :: al, alpha, ear, prefac, rk, zeff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: omat
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: zet
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: gcc
      TYPE(all_potential_type), POINTER                  :: all_potential
      TYPE(atom_basis_type), POINTER                     :: basis
      TYPE(atom_integrals), POINTER                      :: integrals
      TYPE(atom_potential_type), POINTER                 :: potential
      TYPE(atom_type), POINTER                           :: atom
      TYPE(grid_atom_type), POINTER                      :: grid
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set

      IF (rel_control%rel_method == rel_none) RETURN

      NULLIFY (all_potential, orb_basis_set)
      CALL get_qs_kind(qs_kind, basis_set=orb_basis_set, all_potential=all_potential)

      IF(.NOT.(ASSOCIATED(orb_basis_set)))CALL cp__a("atom_kind_orbitals.F",926)

      IF (ASSOCIATED(all_potential)) THEN
         ! only all electron atoms will get the relativistic correction

         CALL get_atomic_kind(atomic_kind, z=z)
         CALL get_qs_kind(qs_kind, zeff=zeff)
         NULLIFY (atom)
         CALL create_atom_type(atom)
         NULLIFY (atom%xc_section)
         NULLIFY (atom%orbitals)
         atom%z = z
         alpha = SQRT(all_potential%alpha_core_charge)

         ! set the method flag
         SELECT CASE (rel_control%rel_method)
         CASE DEFAULT
            CALL cp__b("atom_kind_orbitals.F",943,"")
         CASE (rel_dkh)
            SELECT CASE (rel_control%rel_DKH_order)
            CASE DEFAULT
               CALL cp__b("atom_kind_orbitals.F",947,"")
            CASE (0)
               relativistic = do_dkh0_atom
            CASE (1)
               relativistic = do_dkh1_atom
            CASE (2)
               relativistic = do_dkh2_atom
            CASE (3)
               relativistic = do_dkh3_atom
            CASE (4)
               relativistic = do_dkh4_atom
            CASE (5)
               relativistic = do_dkh5_atom
            END SELECT
         CASE (rel_zora)
            SELECT CASE (rel_control%rel_zora_type)
            CASE DEFAULT
               CALL cp__b("atom_kind_orbitals.F",964,"")
            CASE (rel_zora_full)
               CALL cp__b("atom_kind_orbitals.F",966,"")
            CASE (rel_zora_mp)
               relativistic = do_zoramp_atom
            CASE (rel_sczora_mp)
               relativistic = do_sczoramp_atom
            END SELECT
         END SELECT

         CALL set_atom(atom, &
                       pp_calc=.FALSE., &
                       method_type=do_rks_atom, &
                       relativistic=relativistic, &
                       coulomb_integral_type=do_numeric, &
                       exchange_integral_type=do_numeric)

         ALLOCATE (potential, basis, integrals)

         potential%ppot_type = no_pseudo
         CALL set_atom(atom, zcore=z, potential=potential)

         CALL get_gto_basis_set(orb_basis_set, &
                                nset=nset, nshell=nshell, npgf=npgf, lmin=lmin, lmax=lmax, l=ls, nsgf=nsgf, zet=zet, gcc=gcc, &
                                first_sgf=first_sgf, last_sgf=last_sgf)

         NULLIFY (grid)
         ngp = 400
         quadtype = do_gapw_log
         CALL allocate_grid_atom(grid)
         CALL create_grid_atom(grid, ngp, 1, 1, 0, quadtype)
         grid%nr = ngp
         basis%grid => grid

         NULLIFY (basis%am, basis%cm, basis%as, basis%ns, basis%bf, basis%dbf, basis%ddbf)
         basis%basis_type = CGTO_BASIS
         basis%eps_eig = 1.e-12_dp

         ! fill in the basis data structures
         set_index = 0
         shell_index = 0
         basis%nprim = 0
         basis%nbas = 0
         DO i = 1, nset
            DO j = lmin(i), MIN(lmax(i), 3)
               basis%nprim(j) = basis%nprim(j)+npgf(i)
            END DO
            DO j = 1, nshell(i)
               l = ls(j, i)
               IF (l <= 3) THEN
                  basis%nbas(l) = basis%nbas(l)+1
                  k = basis%nbas(l)
                  IF(.NOT.(k <= 100))CALL cp__a("atom_kind_orbitals.F",1016)
                  set_index(l, k) = i
                  shell_index(l, k) = j
               END IF
            END DO
         END DO

         nj = MAXVAL(basis%nprim)
         ns = MAXVAL(basis%nbas)
         ALLOCATE (basis%am(nj, 0:3))
         basis%am = 0._dp
         ALLOCATE (basis%cm(nj, ns, 0:3))
         basis%cm = 0._dp
         DO j = 0, 3
            nj = 0
            ns = 0
            DO i = 1, nset
               IF (j >= lmin(i) .AND. j <= lmax(i)) THEN
                  DO ipgf = 1, npgf(i)
                     basis%am(nj+ipgf, j) = zet(ipgf, i)
                  END DO
                  DO ii = 1, nshell(i)
                     IF (ls(ii, i) == j) THEN
                        ns = ns+1
                        DO ipgf = 1, npgf(i)
                           basis%cm(nj+ipgf, ns, j) = gcc(ipgf, ii, i)
                        END DO
                     END IF
                  END DO
                  nj = nj+npgf(i)
               END IF
            END DO
         END DO

         ! Normalization as used in the atomic code
         ! We have to undo the Quickstep normalization
         DO j = 0, 3
            prefac = 2.0_dp*SQRT(pi/dfac(2*j+1))
            DO ipgf = 1, basis%nprim(j)
               DO ii = 1, basis%nbas(j)
                  basis%cm(ipgf, ii, j) = prefac*basis%cm(ipgf, ii, j)
               END DO
            END DO
         END DO

         ! initialize basis function on a radial grid
         nr = basis%grid%nr
         m = MAXVAL(basis%nbas)
         ALLOCATE (basis%bf(nr, m, 0:3))
         ALLOCATE (basis%dbf(nr, m, 0:3))
         ALLOCATE (basis%ddbf(nr, m, 0:3))

         basis%bf = 0._dp
         basis%dbf = 0._dp
         basis%ddbf = 0._dp
         DO l = 0, 3
            DO i = 1, basis%nprim(l)
               al = basis%am(i, l)
               DO k = 1, nr
                  rk = basis%grid%rad(k)
                  ear = EXP(-al*basis%grid%rad(k)**2)
                  DO j = 1, basis%nbas(l)
                     basis%bf(k, j, l) = basis%bf(k, j, l)+rk**l*ear*basis%cm(i, j, l)
                     basis%dbf(k, j, l) = basis%dbf(k, j, l) &
                                          +(REAL(l, dp)*rk**(l-1)-2._dp*al*rk**(l+1))*ear*basis%cm(i, j, l)
                     basis%ddbf(k, j, l) = basis%ddbf(k, j, l)+ &
                                           (REAL(l*(l-1), dp)*rk**(l-2)-2._dp*al*REAL(2*l+1, dp)*rk**(l)+4._dp*al*rk**(l+2))* &
                                           ear*basis%cm(i, j, l)
                  END DO
               END DO
            END DO
         END DO

         CALL set_atom(atom, basis=basis)

         ! optimization defaults
         atom%optimization%damping = 0.2_dp
         atom%optimization%eps_scf = 1.e-6_dp
         atom%optimization%eps_diis = 100._dp
         atom%optimization%max_iter = 50
         atom%optimization%n_diis = 5

         ! electronic state
         nelem = 0
         ncore = 0
         ncalc = 0
         DO l = 0, 3
            ll = 2*(2*l+1)
            nn = ptable(z)%e_conv(l)
            ii = 0
            DO
               ii = ii+1
               IF (nn <= ll) THEN
                  nelem(l, ii) = nn
                  EXIT
               ELSE
                  nelem(l, ii) = ll
                  nn = nn-ll
               END IF
            END DO
         END DO
         ncalc = nelem-ncore

         IF (qs_kind%ghost) THEN
            nelem = 0
            ncore = 0
            ncalc = 0
         END IF

         ALLOCATE (atom%state)

         atom%state%core = 0._dp
         atom%state%core(0:3, 1:7) = REAL(ncore(0:3, 1:7), dp)
         atom%state%occ = 0._dp
         atom%state%occ(0:3, 1:7) = REAL(ncalc(0:3, 1:7), dp)
         atom%state%occupation = 0._dp
         atom%state%multiplicity = -1
         DO l = 0, 3
            k = 0
            DO i = 1, 7
               IF (ncalc(l, i) > 0) THEN
                  k = k+1
                  atom%state%occupation(l, k) = REAL(ncalc(l, i), dp)
               END IF
            END DO
         END DO

         atom%state%maxl_occ = get_maxl_occ(atom%state%occupation)
         atom%state%maxn_occ = get_maxn_occ(atom%state%occupation)
         atom%state%maxl_calc = atom%state%maxl_occ
         atom%state%maxn_calc = atom%state%maxn_occ

         ! calculate integrals
         ! general integrals
         CALL atom_int_setup(integrals, basis)
         ! potential
         CALL atom_ppint_setup(integrals, basis, potential=atom%potential)
         ! relativistic correction terms
         NULLIFY (integrals%tzora, integrals%hdkh)
         CALL atom_relint_setup(integrals, basis, atom%relativistic, zcore=REAL(atom%zcore, dp), &
                                alpha=alpha)
         CALL set_atom(atom, integrals=integrals)

         ! for DKH we need erfc integrals to correct non-relativistic
         integrals%core = 0.0_dp
         DO l = 0, 3
            n = integrals%n(l)
            m = basis%nprim(l)
            ALLOCATE (omat(m, m))

            CALL sg_erfc(omat(1:m, 1:m), l, alpha, basis%am(1:m, l), basis%am(1:m, l))
            integrals%core(1:n, 1:n, l) = MATMUL(TRANSPOSE(basis%cm(1:m, 1:n, l)), &
                                                 MATMUL(omat(1:m, 1:m), basis%cm(1:m, 1:n, l)))

            DEALLOCATE (omat)
         END DO

         ! recover relativistic kinetic matrix in CP2K/GPW order and normalization
         IF (ASSOCIATED(rtmat)) THEN
            DEALLOCATE (rtmat)
         END IF
         ALLOCATE (rtmat(nsgf, nsgf))
         rtmat = 0._dp
         DO l = 0, 3
            ll = 2*l
            DO k1 = 1, basis%nbas(l)
               DO k2 = 1, basis%nbas(l)
                  i = first_sgf(shell_index(l, k1), set_index(l, k1))
                  j = first_sgf(shell_index(l, k2), set_index(l, k2))
                  SELECT CASE (atom%relativistic)
                  CASE DEFAULT
                     CALL cp__b("atom_kind_orbitals.F",1187,"")
                  CASE (do_zoramp_atom, do_sczoramp_atom)
                     DO m = 0, ll
                        rtmat(i+m, j+m) = integrals%tzora(k1, k2, l)
                     END DO
                  CASE (do_dkh0_atom, do_dkh1_atom, do_dkh2_atom, do_dkh3_atom, do_dkh4_atom, do_dkh5_atom)
                     DO m = 0, ll
                        rtmat(i+m, j+m) = integrals%hdkh(k1, k2, l)-integrals%kin(k1, k2, l)+atom%zcore*integrals%core(k1, k2, l)
                     END DO
                  END SELECT
               END DO
            END DO
         ENDDO
         DO k1 = 1, nsgf
            DO k2 = k1, nsgf
               rtmat(k1, k2) = 0.5_dp*(rtmat(k1, k2)+rtmat(k2, k1))
               rtmat(k2, k1) = rtmat(k1, k2)
            END DO
         END DO

         ! clean up
         CALL atom_int_release(integrals)
         CALL atom_ppint_release(integrals)
         CALL atom_relint_release(integrals)
         CALL release_atom_basis(basis)
         CALL release_atom_potential(potential)
         CALL release_atom_type(atom)

         DEALLOCATE (potential, basis, integrals)

      ELSE

         IF (ASSOCIATED(rtmat)) THEN
            DEALLOCATE (rtmat)
         END IF
         NULLIFY (rtmat)

      END IF

   END SUBROUTINE calculate_atomic_relkin

! **************************************************************************************************

END MODULE atom_kind_orbitals
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief initializes the environment for lri
!>        lri : local resolution of the identity
!> \par History
!>      created [06.2015]
!> \author Dorothea Golze
! **************************************************************************************************
MODULE lri_environment_init
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE basis_set_types,                 ONLY: copy_gto_basis_set,&
                                              gto_basis_set_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE lri_environment_types,           ONLY: deallocate_bas_properties,&
                                              lri_env_create,&
                                              lri_environment_type
   USE mathconstants,                   ONLY: fac,&
                                              pi
   USE memory_utilities,                ONLY: reallocate
   USE orbital_pointers,                ONLY: indso,&
                                              nsoset
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE spherical_harmonics,             ONLY: clebsch_gordon,&
                                              clebsch_gordon_deallocate,&
                                              clebsch_gordon_init
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lri_environment_init'

   PUBLIC :: lri_env_init, lri_basis_init

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief initializes the lri env
!> \param qs_env ...
!> \param force_env_section ...
!> \param qs_kind_set ...
! **************************************************************************************************
   SUBROUTINE lri_env_init(qs_env, force_env_section, qs_kind_set)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: force_env_section
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CHARACTER(len=*), PARAMETER :: routineN = 'lri_env_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, lmax_ikind_orb, lmax_ikind_ri, &
                                                            maxl_orb, maxl_ri, nkind
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(gto_basis_set_type), POINTER                  :: lri_basis_set, orb_basis_set
      TYPE(lri_environment_type), POINTER                :: lri_env

      NULLIFY (atomic_kind_set, lri_basis_set, lri_env, orb_basis_set)
      CALL lri_env_create(lri_env)

      ! initialize the basic basis sets (orb and ri)
      CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set)
      nkind = SIZE(atomic_kind_set)
      ALLOCATE (lri_env%orb_basis(nkind), lri_env%ri_basis(nkind))
      maxl_orb = 0
      maxl_ri = 0
      DO ikind = 1, nkind
         NULLIFY (orb_basis_set, lri_basis_set)
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set, basis_type="ORB")
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=lri_basis_set, basis_type="LRI")
         NULLIFY (lri_env%orb_basis(ikind)%gto_basis_set)
         NULLIFY (lri_env%ri_basis(ikind)%gto_basis_set)
         IF (ASSOCIATED(orb_basis_set)) THEN
            CALL copy_gto_basis_set(orb_basis_set, lri_env%orb_basis(ikind)%gto_basis_set)
            CALL copy_gto_basis_set(lri_basis_set, lri_env%ri_basis(ikind)%gto_basis_set)
         END IF
         lmax_ikind_orb = MAXVAL(lri_env%orb_basis(ikind)%gto_basis_set%lmax)
         lmax_ikind_ri = MAXVAL(lri_env%ri_basis(ikind)%gto_basis_set%lmax)
         maxl_orb = MAX(maxl_orb, lmax_ikind_orb)
         maxl_ri = MAX(maxl_ri, lmax_ikind_ri)
      END DO
      CALL lri_basis_init(lri_env)

      ! CG coefficients needed for lri integrals
      IF (ASSOCIATED(lri_env%cg_shg)) THEN
         CALL get_clebsch_gordon_coefficients(lri_env%cg_shg%cg_coeff, &
                                              lri_env%cg_shg%cg_none0_list, &
                                              lri_env%cg_shg%ncg_none0, &
                                              maxl_orb, maxl_ri)
      ENDIF
      ! init keywords

      ! check for debug (OS scheme)
      CALL section_vals_val_get(force_env_section, "DFT%QS%LRIGPW%DEBUG_LRI_INTEGRALS", &
                                l_val=lri_env%debug)
      ! integrals based on solid harmonic Gaussians
      CALL section_vals_val_get(force_env_section, "DFT%QS%LRIGPW%SHG_LRI_INTEGRALS", &
                                l_val=lri_env%use_shg_integrals)
      ! how to calculate inverse/pseuodinverse of overlap
      CALL section_vals_val_get(force_env_section, "DFT%QS%LRIGPW%LRI_OVERLAP_MATRIX", &
                                i_val=lri_env%lri_overlap_inv)
      CALL section_vals_val_get(force_env_section, "DFT%QS%LRIGPW%MAX_CONDITION_NUM", &
                                r_val=lri_env%cond_max)
      !
      CALL set_qs_env(qs_env, lri_env=lri_env)

   END SUBROUTINE lri_env_init

! **************************************************************************************************
!> \brief initializes the lri basis: calculates the norm, self-overlap
!>        and integral of the ri basis
!> \param lri_env ...
! **************************************************************************************************
   SUBROUTINE lri_basis_init(lri_env)
      TYPE(lri_environment_type), POINTER                :: lri_env

      CHARACTER(len=*), PARAMETER :: routineN = 'lri_basis_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, nkind
      REAL(KIND=dp), DIMENSION(:), POINTER               :: orb_norm_r, ri_int_fbas, ri_norm_r, &
                                                            ri_norm_s
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: orb_ovlp, ri_ovlp
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: scon_orb, scon_ri
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis, ri_basis

      IF (ASSOCIATED(lri_env)) THEN
         IF (ASSOCIATED(lri_env%orb_basis)) THEN
            IF(.NOT.(ASSOCIATED(lri_env%ri_basis)))CALL cp__a("lri_environment_init.F",142)
            nkind = SIZE(lri_env%orb_basis)
            CALL deallocate_bas_properties(lri_env)
            ALLOCATE (lri_env%bas_prop(nkind))
            DO ikind = 1, nkind
               NULLIFY (orb_basis, ri_basis)
               orb_basis => lri_env%orb_basis(ikind)%gto_basis_set
               IF (ASSOCIATED(orb_basis)) THEN
                  ri_basis => lri_env%ri_basis(ikind)%gto_basis_set
                  IF(.NOT.(ASSOCIATED(ri_basis)))CALL cp__a("lri_environment_init.F",151)
                  NULLIFY (ri_norm_r)
                  CALL basis_norm_radial(ri_basis, ri_norm_r)
                  NULLIFY (orb_norm_r)
                  CALL basis_norm_radial(orb_basis, orb_norm_r)
                  NULLIFY (ri_norm_s)
                  CALL basis_norm_s_func(ri_basis, ri_norm_s)
                  NULLIFY (ri_int_fbas)
                  CALL basis_int(ri_basis, ri_int_fbas, ri_norm_s)
                  lri_env%bas_prop(ikind)%int_fbas => ri_int_fbas
                  NULLIFY (ri_ovlp)
                  CALL basis_ovlp(ri_basis, ri_ovlp, ri_norm_r)
                  lri_env%bas_prop(ikind)%ri_ovlp => ri_ovlp
                  NULLIFY (orb_ovlp)
                  CALL basis_ovlp(orb_basis, orb_ovlp, orb_norm_r)
                  lri_env%bas_prop(ikind)%orb_ovlp => orb_ovlp
                  NULLIFY (scon_ri)
                  CALL contraction_matrix_shg(ri_basis, scon_ri)
                  lri_env%bas_prop(ikind)%scon_ri => scon_ri
                  NULLIFY (scon_orb)
                  CALL contraction_matrix_shg(orb_basis, scon_orb)
                  lri_env%bas_prop(ikind)%scon_orb => scon_orb
                  DEALLOCATE (orb_norm_r, ri_norm_r, ri_norm_s)
               END IF
            END DO
         END IF
      END IF

   END SUBROUTINE lri_basis_init

!*****************************************************************************
!> \brief normalization for a contracted Gaussian s-function,
!>        spherical = cartesian Gaussian for s-functions
!> \param basis ...
!> \param norm ...
! **************************************************************************************************
   SUBROUTINE basis_norm_s_func(basis, norm)

      TYPE(gto_basis_set_type), POINTER                  :: basis
      REAL(dp), DIMENSION(:), POINTER                    :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'basis_norm_s_func', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ipgf, iset, isgf, ishell, jpgf, l, nbas
      REAL(KIND=dp)                                      :: aai, aaj, cci, ccj, expa, ppl

      NULLIFY (norm)

      nbas = basis%nsgf
      ALLOCATE (norm(nbas))
      norm = 0._dp

      DO iset = 1, basis%nset
         DO ishell = 1, basis%nshell(iset)
            l = basis%l(ishell, iset)
            IF (l /= 0) CYCLE
            expa = 0.5_dp*REAL(2*l+3, dp)
            ppl = pi**(3._dp/2._dp)
            DO isgf = basis%first_sgf(ishell, iset), basis%last_sgf(ishell, iset)
               DO ipgf = 1, basis%npgf(iset)
                  cci = basis%gcc(ipgf, ishell, iset)
                  aai = basis%zet(ipgf, iset)
                  DO jpgf = 1, basis%npgf(iset)
                     ccj = basis%gcc(jpgf, ishell, iset)
                     aaj = basis%zet(jpgf, iset)
                     norm(isgf) = norm(isgf)+cci*ccj*ppl/(aai+aaj)**expa
                  END DO
               END DO
               norm(isgf) = 1.0_dp/SQRT(norm(isgf))
            END DO
         END DO
      END DO

   END SUBROUTINE basis_norm_s_func

!*****************************************************************************
!> \brief normalization for radial part of contracted spherical Gaussian
!>        functions
!> \param basis ...
!> \param norm ...
! **************************************************************************************************
   SUBROUTINE basis_norm_radial(basis, norm)

      TYPE(gto_basis_set_type), POINTER                  :: basis
      REAL(dp), DIMENSION(:), POINTER                    :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'basis_norm_radial', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ipgf, iset, isgf, ishell, jpgf, l, nbas
      REAL(KIND=dp)                                      :: aai, aaj, cci, ccj, expa, ppl

      NULLIFY (norm)

      nbas = basis%nsgf
      ALLOCATE (norm(nbas))
      norm = 0._dp

      DO iset = 1, basis%nset
         DO ishell = 1, basis%nshell(iset)
            l = basis%l(ishell, iset)
            expa = 0.5_dp*REAL(2*l+3, dp)
            ppl = fac(2*l+2)*SQRT(pi)/2._dp**REAL(2*l+3, dp)/fac(l+1)
            DO isgf = basis%first_sgf(ishell, iset), basis%last_sgf(ishell, iset)
               DO ipgf = 1, basis%npgf(iset)
                  cci = basis%gcc(ipgf, ishell, iset)
                  aai = basis%zet(ipgf, iset)
                  DO jpgf = 1, basis%npgf(iset)
                     ccj = basis%gcc(jpgf, ishell, iset)
                     aaj = basis%zet(jpgf, iset)
                     norm(isgf) = norm(isgf)+cci*ccj*ppl/(aai+aaj)**expa
                  END DO
               END DO
               norm(isgf) = 1.0_dp/SQRT(norm(isgf))
            END DO
         END DO
      END DO

   END SUBROUTINE basis_norm_radial

!*****************************************************************************
!> \brief normalization solid harmonic Gaussians (SHG)
!> \param basis ...
!> \param norm ...
! **************************************************************************************************
   SUBROUTINE basis_norm_shg(basis, norm)

      TYPE(gto_basis_set_type), POINTER                  :: basis
      REAL(dp), DIMENSION(:, :), POINTER                 :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'basis_norm_shg', routineP = moduleN//':'//routineN

      INTEGER                                            :: ipgf, iset, ishell, jpgf, l
      REAL(KIND=dp)                                      :: aai, aaj, cci, ccj, expa, ppl

      norm = 0._dp

      DO iset = 1, basis%nset
         DO ishell = 1, basis%nshell(iset)
            l = basis%l(ishell, iset)
            expa = 0.5_dp*REAL(2*l+3, dp)
            ppl = fac(2*l+2)*pi**(1.5_dp)/fac(l+1)
            ppl = ppl/(2._dp**REAL(2*l+1, dp))
            ppl = ppl/REAL(2*l+1, dp)
            DO ipgf = 1, basis%npgf(iset)
               cci = basis%gcc(ipgf, ishell, iset)
               aai = basis%zet(ipgf, iset)
               DO jpgf = 1, basis%npgf(iset)
                  ccj = basis%gcc(jpgf, ishell, iset)
                  aaj = basis%zet(jpgf, iset)
                  norm(iset, ishell) = norm(iset, ishell)+cci*ccj*ppl/(aai+aaj)**expa
               END DO
            END DO
            norm(iset, ishell) = 1.0_dp/SQRT(norm(iset, ishell))
         END DO
      END DO

   END SUBROUTINE basis_norm_shg

!*****************************************************************************
!> \brief integral over a single (contracted) lri auxiliary basis function,
!>        integral is zero for all but s-functions
!> \param basis ...
!> \param int_aux ...
!> \param norm ...
! **************************************************************************************************
   SUBROUTINE basis_int(basis, int_aux, norm)

      TYPE(gto_basis_set_type), POINTER                  :: basis
      REAL(dp), DIMENSION(:), POINTER                    :: int_aux, norm

      CHARACTER(len=*), PARAMETER :: routineN = 'basis_int', routineP = moduleN//':'//routineN

      INTEGER                                            :: ipgf, iset, isgf, ishell, l, nbas
      REAL(KIND=dp)                                      :: aa, cc, pp

      nbas = basis%nsgf
      ALLOCATE (int_aux(nbas))
      int_aux = 0._dp

      DO iset = 1, basis%nset
         DO ishell = 1, basis%nshell(iset)
            l = basis%l(ishell, iset)
            IF (l /= 0) CYCLE
            DO isgf = basis%first_sgf(ishell, iset), basis%last_sgf(ishell, iset)
               DO ipgf = 1, basis%npgf(iset)
                  cc = basis%gcc(ipgf, ishell, iset)
                  aa = basis%zet(ipgf, iset)
                  pp = (pi/aa)**(3._dp/2._dp)
                  int_aux(isgf) = int_aux(isgf)+norm(isgf)*cc*pp
               END DO
            END DO
         END DO
      END DO

   END SUBROUTINE basis_int

!*****************************************************************************
!> \brief self-overlap of lri basis for contracted spherical Gaussians.
!>        Overlap of radial part. Norm contains only normalization of radial
!>        part. Norm and overlap of spherical harmonics not explicitly
!>        calculated since this cancels for the self-overlap anyway.
!> \param basis ...
!> \param ovlp ...
!> \param norm ...
! **************************************************************************************************
   SUBROUTINE basis_ovlp(basis, ovlp, norm)

      TYPE(gto_basis_set_type), POINTER                  :: basis
      REAL(dp), DIMENSION(:, :), POINTER                 :: ovlp
      REAL(dp), DIMENSION(:), POINTER                    :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'basis_ovlp', routineP = moduleN//':'//routineN

      INTEGER                                            :: ipgf, iset, isgf, ishell, jpgf, jset, &
                                                            jsgf, jshell, l, li, lj, nbas
      REAL(KIND=dp)                                      :: aai, aaj, cci, ccj, expa, m_i, m_j, &
                                                            norm_i, norm_j, oo, ppl

      nbas = basis%nsgf
      ALLOCATE (ovlp(nbas, nbas))
      ovlp = 0._dp

      DO iset = 1, basis%nset
         DO ishell = 1, basis%nshell(iset)
            li = basis%l(ishell, iset)
            DO jset = 1, basis%nset
               DO jshell = 1, basis%nshell(jset)
                  lj = basis%l(jshell, jset)
                  IF (li == lj) THEN
                     l = li
                     expa = 0.5_dp*REAL(2*l+3, dp)
                     ppl = fac(2*l+2)*SQRT(pi)/2._dp**REAL(2*l+3, dp)/fac(l+1)
                     DO isgf = basis%first_sgf(ishell, iset), basis%last_sgf(ishell, iset)
                        m_i = basis%m(isgf)
                        DO jsgf = basis%first_sgf(jshell, jset), basis%last_sgf(jshell, jset)
                           m_j = basis%m(jsgf)
                           IF (m_i == m_j) THEN
                              DO ipgf = 1, basis%npgf(iset)
                                 cci = basis%gcc(ipgf, ishell, iset)
                                 aai = basis%zet(ipgf, iset)
                                 norm_i = norm(isgf)
                                 DO jpgf = 1, basis%npgf(jset)
                                    ccj = basis%gcc(jpgf, jshell, jset)
                                    aaj = basis%zet(jpgf, jset)
                                    oo = 1._dp/(aai+aaj)**expa
                                    norm_j = norm(jsgf)
                                    ovlp(isgf, jsgf) = ovlp(isgf, jsgf)+norm_i*norm_j*ppl*cci*ccj*oo
                                 END DO
                              END DO
                           ENDIF
                        END DO
                     END DO
                  END IF
               END DO
            END DO
         END DO
      END DO

   END SUBROUTINE basis_ovlp

!*****************************************************************************
!> \brief contraction matrix for SHG integrals
!> \param basis ...
!> \param scon_shg contraction matrix
! **************************************************************************************************
   SUBROUTINE contraction_matrix_shg(basis, scon_shg)

      TYPE(gto_basis_set_type), POINTER                  :: basis
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: scon_shg

      CHARACTER(len=*), PARAMETER :: routineN = 'contraction_matrix_shg', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ipgf, iset, ishell, l, maxpgf, maxshell, &
                                                            nset
      INTEGER, DIMENSION(:), POINTER                     :: npgf, nshell
      REAL(KIND=dp)                                      :: gcc, norm_w
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: norm, zet

      nset = basis%nset
      npgf => basis%npgf
      nshell => basis%nshell
      zet => basis%zet

      maxpgf = SIZE(basis%gcc, 1)
      maxshell = SIZE(basis%gcc, 2)
      ALLOCATE (norm(basis%nset, maxshell))
      ALLOCATE (scon_shg(maxpgf, maxshell, nset))

      CALL basis_norm_shg(basis, norm)

      DO iset = 1, nset
         DO ishell = 1, nshell(iset)
            l = basis%l(ishell, iset)
            norm_w = norm(iset, ishell)
            DO ipgf = 1, npgf(iset)
               gcc = basis%gcc(ipgf, ishell, iset)
               scon_shg(ipgf, ishell, iset) = norm(iset, ishell)*gcc
            END DO
         END DO
      END DO

      DEALLOCATE (norm)

   END SUBROUTINE contraction_matrix_shg

!*****************************************************************************
!> \brief calculate the Clebsch-Gordon (CG) coefficients for expansion of the
!>        product of two spherical harmonic Gaussians
!> \param my_cg matrix storing CG coefficients
!> \param cg_none0_list list of none-zero CG coefficients
!> \param ncg_none0 number of none-zero CG coefficients
!> \param maxl1 maximal l quantum number of 1st spherical function
!> \param maxl2 maximal l quantum number of 2nd spherical function
! **************************************************************************************************
   SUBROUTINE get_clebsch_gordon_coefficients(my_cg, cg_none0_list, ncg_none0, &
                                              maxl1, maxl2)

      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: my_cg
      INTEGER, DIMENSION(:, :, :), POINTER               :: cg_none0_list
      INTEGER, DIMENSION(:, :), POINTER                  :: ncg_none0
      INTEGER, INTENT(IN)                                :: maxl1, maxl2

      CHARACTER(len=*), PARAMETER :: routineN = 'get_clebsch_gordon_coefficients', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: il, ilist, iso, iso1, iso2, l1, l1l2, &
                                                            l2, lc1, lc2, lp, m1, m2, maxl, mm, &
                                                            mp, nlist, nlist_max, nsfunc, nsfunc1, &
                                                            nsfunc2
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: rga

      nlist_max = 6
      nsfunc1 = nsoset(maxl1)
      nsfunc2 = nsoset(maxl2)
      maxl = maxl1+maxl2
      nsfunc = nsoset(maxl)

      CALL clebsch_gordon_init(maxl)

      ALLOCATE (my_cg(nsfunc1, nsfunc2, nsfunc))
      my_cg = 0.0_dp
      ALLOCATE (ncg_none0(nsfunc1, nsfunc2))
      ncg_none0 = 0
      ALLOCATE (cg_none0_list(nsfunc1, nsfunc2, nlist_max))
      cg_none0_list = 0

      ALLOCATE (rga(maxl, 2))
      rga = 0.0_dp
      DO lc1 = 0, maxl1
         DO iso1 = nsoset(lc1-1)+1, nsoset(lc1)
            l1 = indso(1, iso1)
            m1 = indso(2, iso1)
            DO lc2 = 0, maxl2
               DO iso2 = nsoset(lc2-1)+1, nsoset(lc2)
                  l2 = indso(1, iso2)
                  m2 = indso(2, iso2)
                  CALL clebsch_gordon(l1, m1, l2, m2, rga)
                  l1l2 = l1+l2
                  mp = m1+m2
                  mm = m1-m2
                  IF (m1*m2 < 0 .OR. (m1*m2 == 0 .AND. (m1 < 0 .OR. m2 < 0))) THEN
                     mp = -ABS(mp)
                     mm = -ABS(mm)
                  ELSE
                     mp = ABS(mp)
                     mm = ABS(mm)
                  END IF
                  DO lp = MOD(l1+l2, 2), l1l2, 2
                     il = lp/2+1
                     IF (ABS(mp) <= lp) THEN
                        IF (mp >= 0) THEN
                           iso = nsoset(lp-1)+lp+1+mp
                        ELSE
                           iso = nsoset(lp-1)+lp+1-ABS(mp)
                        END IF
                        my_cg(iso1, iso2, iso) = rga(il, 1)
                     ENDIF
                     IF (mp /= mm .AND. ABS(mm) <= lp) THEN
                        IF (mm >= 0) THEN
                           iso = nsoset(lp-1)+lp+1+mm
                        ELSE
                           iso = nsoset(lp-1)+lp+1-ABS(mm)
                        END IF
                        my_cg(iso1, iso2, iso) = rga(il, 2)
                     ENDIF
                  END DO
                  nlist = 0
                  DO ilist = 1, nsfunc
                     IF (ABS(my_cg(iso1, iso2, ilist)) > 1.E-8_dp) THEN
                        nlist = nlist+1
                        IF (nlist > nlist_max) THEN
                           CALL reallocate(cg_none0_list, 1, nsfunc1, 1, nsfunc2, 1, nlist)
                           nlist_max = nlist
                        ENDIF
                        cg_none0_list(iso1, iso2, nlist) = ilist
                     ENDIF
                  ENDDO
                  ncg_none0(iso1, iso2) = nlist
               ENDDO ! iso2
            ENDDO ! lc2
         ENDDO ! iso1
      ENDDO ! lc1

      DEALLOCATE (rga)
      CALL clebsch_gordon_deallocate()

   END SUBROUTINE get_clebsch_gordon_coefficients

END MODULE lri_environment_init
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines for DFT+NEGF calculations (coupling with the quantum transport code OMEN)
!> \par History
!>       12.2012 created external_scf_method [Hossein Bani-Hashemian]
!>       05.2013 created rotines to work with C-interoperable matrices [Hossein Bani-Hashemian]
!>       07.2013 created transport_env routines [Hossein Bani-Hashemian]
!>       11.2014 switch to CSR matrices [Hossein Bani-Hashemian]
!>       12.2014 merged [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
MODULE transport
   USE ISO_C_BINDING,                   ONLY: C_ASSOCIATED,&
                                              C_DOUBLE,&
                                              C_F_PROCPOINTER,&
                                              C_INT,&
                                              C_LOC,&
                                              C_NULL_PTR,&
                                              C_PTR
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE bibliography,                    ONLY: Bruck2014,&
                                              cite_reference
   USE cp_dbcsr_cp2k_link,              ONLY: cp_dbcsr_to_csr_screening
   USE cp_dbcsr_interface,              ONLY: &
        cp_convert_csr_to_dbcsr, cp_convert_dbcsr_to_csr, cp_csr_create, cp_csr_create_from_dbcsr, &
        cp_csr_print_sparsity, cp_dbcsr_copy, cp_dbcsr_copy_into_existing, cp_dbcsr_create, &
        cp_dbcsr_desymmetrize, cp_dbcsr_has_symmetry, cp_dbcsr_init, cp_dbcsr_set, cp_dbcsr_type, &
        csr_eqrow_ceil_dist, csr_type, dbcsr_type_real_8
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: transport_diagscalapack
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp,&
                                              int_1
   USE message_passing,                 ONLY: mp_allgather,&
                                              mp_bcast,&
                                              mp_environ
   USE particle_methods,                ONLY: get_particle_set
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: evolt
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_integral_utils,               ONLY: basis_set_list_setup
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_env_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              get_neighbor_list_set_p,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE transport_env_types,             ONLY: cp2k_csr_interop_type,&
                                              cp2k_transport_parameters,&
                                              csr_interop_matrix_get_info,&
                                              csr_interop_nullify,&
                                              transport_env_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'transport'

   PUBLIC :: transport_env_create, transport_initialize, external_scf_method

!> interface between C/C++ and FORTRAN
   INTERFACE c_func_interface
! **************************************************************************************************
!> \brief C routine that takes the S and H matrices as input and outputs a P matrix
!> \param cp2k_transport_params transport parameters read form a CP2K input file
!> \param s_mat  C-interoperable overlap matrix
!> \param ks_mat C-interoperable Kohn-Sham matrix
!> \param p_mat  C-interoperable density matrix
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
      SUBROUTINE c_scf_routine(cp2k_transport_params, s_mat, ks_mat, p_mat) BIND(C)
         IMPORT :: C_INT, C_PTR, cp2k_csr_interop_type, cp2k_transport_parameters
         IMPLICIT NONE
         TYPE(cp2k_transport_parameters), VALUE, INTENT(IN) :: cp2k_transport_params
         TYPE(cp2k_csr_interop_type), VALUE, INTENT(IN)     :: s_mat
         TYPE(cp2k_csr_interop_type), VALUE, INTENT(IN)     :: ks_mat
         TYPE(cp2k_csr_interop_type), INTENT(INOUT)         :: p_mat
      END SUBROUTINE c_scf_routine
   END INTERFACE c_func_interface

CONTAINS

! **************************************************************************************************
!> \brief creates the transport environment
!> \param[inout] qs_env the qs_env containing the transport_env
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE transport_env_create(qs_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'transport_env_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(transport_env_type), POINTER                  :: transport_env

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env, transport_env=transport_env, input=input)

      IF(.NOT.(.NOT. ASSOCIATED(transport_env)))CALL cp__a("transport.F",125)

      ALLOCATE (transport_env)
      CALL transport_init_read_input(input, transport_env)
      CALL transport_set_contact_params(qs_env, transport_env)
      CALL cp_dbcsr_init(transport_env%template_matrix_sym)
      CALL cp_dbcsr_init(transport_env%template_matrix_nosym)
      CALL cp_dbcsr_init(transport_env%csr_sparsity)
      CALL set_qs_env(qs_env, transport_env=transport_env)

      CALL timestop(handle)

   END SUBROUTINE transport_env_create

! **************************************************************************************************
!> \brief intitializes all fields of transport_env using the parameters read from
!>        the corresponding input section
!> \param[inout] input         the input file
!> \param[inout] transport_env the transport_env to be initialized
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE transport_init_read_input(input, transport_env)
      TYPE(section_vals_type), POINTER                   :: input
      TYPE(transport_env_type), INTENT(INOUT)            :: transport_env

      CHARACTER(len=*), PARAMETER :: routineN = 'transport_init_read_input', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: contact_bandwidth, contact_injsign, &
                                                            contact_natoms, contact_start, handle, &
                                                            i, n_contacts, stride_contacts
      INTEGER, DIMENSION(:), POINTER                     :: i_vals
      LOGICAL                                            :: contact_explicit, injecting_contact, &
                                                            one_circle
      TYPE(section_vals_type), POINTER                   :: beyn_section, contact_section, &
                                                            transport_section

      CALL timeset(routineN, handle)

      transport_section => section_vals_get_subs_vals(input, "DFT%TRANSPORT")
      contact_section => section_vals_get_subs_vals(transport_section, "CONTACT")
      beyn_section => section_vals_get_subs_vals(transport_section, "BEYN")
      CALL section_vals_get(contact_section, explicit=contact_explicit, n_repetition=n_contacts)

      NULLIFY (i_vals)
! read from input
      CALL section_vals_val_get(transport_section, "TRANSPORT_METHOD", i_val=transport_env%params%method)
      CALL section_vals_val_get(transport_section, "INJECTION_METHOD", i_val=transport_env%params%injection_method)
      CALL section_vals_val_get(transport_section, "REAL_AXIS_INTEGRATION_METHOD", &
                                i_val=transport_env%params%rlaxis_integration_method)
      CALL section_vals_val_get(transport_section, "LINEAR_SOLVER", i_val=transport_env%params%linear_solver)
      CALL section_vals_val_get(transport_section, "N_ABSCISSAE", i_val=transport_env%params%n_abscissae)
      CALL section_vals_val_get(transport_section, "N_KPOINTS", i_val=transport_env%params%n_kpoint)
      CALL section_vals_val_get(transport_section, "NUM_INTERVAL", i_val=transport_env%params%num_interval)
      CALL section_vals_val_get(transport_section, "TASKS_PER_POINT", i_val=transport_env%params%tasks_per_point)
      CALL section_vals_val_get(transport_section, "GPUS_PER_POINT", i_val=transport_env%params%gpus_per_point)
      CALL section_vals_val_get(transport_section, "COLZERO_THRESHOLD", r_val=transport_env%params%colzero_threshold)
      CALL section_vals_val_get(transport_section, "EPS_LIMIT", r_val=transport_env%params%eps_limit)
      CALL section_vals_val_get(transport_section, "EPS_LIMIT_CC", r_val=transport_env%params%eps_limit_cc)
      CALL section_vals_val_get(transport_section, "EPS_DECAY", r_val=transport_env%params%eps_decay)
      CALL section_vals_val_get(transport_section, "EPS_SINGULARITY_CURVATURES", &
                                r_val=transport_env%params%eps_singularity_curvatures)
      CALL section_vals_val_get(transport_section, "EPS_MU", r_val=transport_env%params%eps_mu)
      CALL section_vals_val_get(transport_section, "EPS_EIGVAL_DEGEN", r_val=transport_env%params%eps_eigval_degen)
      CALL section_vals_val_get(transport_section, "EPS_FERMI", r_val=transport_env%params%eps_fermi)
      CALL section_vals_val_get(transport_section, "ENERGY_INTERVAL", r_val=transport_env%params%energy_interval)
      CALL section_vals_val_get(transport_section, "MIN_INTERVAL", r_val=transport_env%params%min_interval)
      CALL section_vals_val_get(transport_section, "TEMPERATURE", r_val=transport_env%params%temperature)
      CALL section_vals_val_get(transport_section, "CSR_SCREENING", l_val=transport_env%csr_screening)

      CALL section_vals_val_get(transport_section, "CUTOUT", i_vals=i_vals)
      transport_env%params%cutout = i_vals

      CALL section_vals_val_get(beyn_section, "N_POINTS_BEYN", i_val=transport_env%params%n_points_beyn)
      CALL section_vals_val_get(beyn_section, "N_RAND", r_val=transport_env%params%n_rand_beyn)
      CALL section_vals_val_get(beyn_section, "N_RAND_CC", r_val=transport_env%params%n_rand_cc_beyn)
      CALL section_vals_val_get(beyn_section, "SVD_CUTOFF", r_val=transport_env%params%svd_cutoff)
      CALL section_vals_val_get(beyn_section, "ONE_CIRCLE", l_val=one_circle)
      IF (one_circle) THEN
         transport_env%params%ncrc_beyn = 1
      ELSE
         transport_env%params%ncrc_beyn = 2
      END IF

      IF (contact_explicit) THEN
         transport_env%params%num_contacts = n_contacts
         stride_contacts = 6; 
         ALLOCATE (transport_env%contacts_data(stride_contacts*n_contacts))

         DO i = 1, n_contacts
            CALL section_vals_val_get(contact_section, "BANDWIDTH", i_rep_section=i, i_val=contact_bandwidth)
            CALL section_vals_val_get(contact_section, "START", i_rep_section=i, i_val=contact_start)
            CALL section_vals_val_get(contact_section, "N_ATOMS", i_rep_section=i, i_val=contact_natoms)
            CALL section_vals_val_get(contact_section, "INJECTION_SIGN", i_rep_section=i, i_val=contact_injsign)
            CALL section_vals_val_get(contact_section, "INJECTING_CONTACT", i_rep_section=i, l_val=injecting_contact)

            IF (contact_natoms .LE. 0) CALL cp__b("transport.F",221,"Number of atoms in contact region needs to be defined.")

            transport_env%contacts_data((i-1)*stride_contacts+1) = contact_bandwidth
            transport_env%contacts_data((i-1)*stride_contacts+2) = contact_start-1 ! C indexing
            transport_env%contacts_data((i-1)*stride_contacts+3) = contact_natoms
            transport_env%contacts_data((i-1)*stride_contacts+4) = contact_injsign

            IF (injecting_contact) THEN
               transport_env%contacts_data((i-1)*stride_contacts+5) = 1
            ELSE
               transport_env%contacts_data((i-1)*stride_contacts+5) = 0
            END IF
            transport_env%contacts_data((i-1)*stride_contacts+6) = contact_bandwidth*contact_natoms
         END DO
         transport_env%params%contacts_data = C_LOC(transport_env%contacts_data(1))
      ELSE
         IF (transport_env%params%method .EQ. transport_diagscalapack) THEN
            transport_env%params%num_contacts = 0
            transport_env%params%contacts_data = C_NULL_PTR
         ELSE
            CALL cp__b("transport.F",241,"No contact region is defined.")
         END IF
      END IF

      CALL timestop(handle)

   END SUBROUTINE transport_init_read_input

! **************************************************************************************************
!> \brief initializes the transport environment
!> \param ks_env ...
!> \param[inout] transport_env the transport env to be initialized
!> \param[in]    template_matrix   template matrix to keep the sparsity of matrices fixed
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE transport_initialize(ks_env, transport_env, template_matrix)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(transport_env_type), INTENT(INOUT)            :: transport_env
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: template_matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'transport_initialize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, mp_group, mynode, numnodes, &
                                                            unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      CALL cite_reference(Bruck2014)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      mp_group = logger%para_env%group
      CALL mp_environ(numnodes, mynode, mp_group)

      ! Use all tasks if user input for TASKS_PER_POINT is 0 or greater than MPI size.
      IF ((transport_env%params%tasks_per_point .GT. numnodes) &
          .OR. (transport_env%params%tasks_per_point .EQ. 0)) THEN
         transport_env%params%tasks_per_point = numnodes
      ENDIF

      ! Determine valid TASKS_PER_POINT from user input
      DO WHILE (MOD(numnodes, transport_env%params%tasks_per_point) .NE. 0)
         transport_env%params%tasks_per_point = transport_env%params%tasks_per_point+1
      ENDDO

      IF (cp_dbcsr_has_symmetry(template_matrix)) THEN
         CALL cp_dbcsr_copy(transport_env%template_matrix_sym, template_matrix)
         CALL cp_dbcsr_desymmetrize(transport_env%template_matrix_sym, transport_env%template_matrix_nosym)
      ELSE
         CALL cp_dbcsr_copy(transport_env%template_matrix_nosym, template_matrix)
         CALL cp_dbcsr_copy(transport_env%template_matrix_sym, template_matrix)
      END IF

      CALL cp_dbcsr_create(transport_env%csr_sparsity, "CSR sparsity", &
                           template=transport_env%template_matrix_sym, &
                           data_type=dbcsr_type_real_8)
      CALL cp_dbcsr_copy(transport_env%csr_sparsity, transport_env%template_matrix_sym)

      CALL cp_dbcsr_to_csr_screening(ks_env, transport_env%csr_sparsity)

      IF (.NOT. transport_env%csr_screening) CALL cp_dbcsr_set(transport_env%csr_sparsity, 1.0)
      CALL cp_csr_create_from_dbcsr(transport_env%template_matrix_nosym, &
                                    transport_env%s_matrix, &
                                    csr_eqrow_ceil_dist, &
                                    csr_sparsity=transport_env%csr_sparsity, &
                                    numnodes=numnodes)

      CALL cp_csr_print_sparsity(transport_env%s_matrix, unit_nr)

      CALL cp_convert_dbcsr_to_csr(transport_env%template_matrix_nosym, transport_env%s_matrix)

      CALL cp_csr_create(transport_env%ks_matrix, transport_env%s_matrix)
      CALL cp_csr_create(transport_env%p_matrix, transport_env%s_matrix)

      CALL transport_eval_blocks(ks_env, transport_env)

      CALL timestop(handle)

   END SUBROUTINE transport_initialize

! **************************************************************************************************
!> \brief SCF calcualtion with an externally evaluated density matrix
!> \param[inout] transport_env  transport environment
!> \param[in]    matrix_s       DBCSR overlap matrix
!> \param[in]    matrix_ks      DBCSR Kohn-Sham matrix
!> \param[inout] matrix_p       DBCSR density matrix
!> \param[in]    nelectron_spin number of electrons
!> \param[in]    natoms         number of atoms
!> \param[in]    energy_diff    scf energy difference
!> \param[in]    extra_scf      whether or not an extra scf step will be performed
!> \par History
!>       12.2012 created [Hossein Bani-Hashemian]
!>       12.2014 revised [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE external_scf_method(transport_env, matrix_s, matrix_ks, matrix_p, &
                                  nelectron_spin, natoms, energy_diff, extra_scf)

      TYPE(transport_env_type), INTENT(INOUT)            :: transport_env
      TYPE(cp_dbcsr_type), INTENT(IN)                    :: matrix_s, matrix_ks
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_p
      INTEGER, INTENT(IN)                                :: nelectron_spin, natoms
      REAL(dp), INTENT(IN)                               :: energy_diff
      LOGICAL, INTENT(IN)                                :: extra_scf

      CHARACTER(len=*), PARAMETER :: routineN = 'external_scf_method', &
         routineP = moduleN//':'//routineN

      TYPE(cp2k_csr_interop_type)                        :: ks_mat, p_mat, s_mat

      PROCEDURE(c_scf_routine), POINTER        :: c_method
      INTEGER                                  :: handle

      CALL timeset(routineN, handle)

      CALL C_F_PROCPOINTER(transport_env%ext_c_method_ptr, c_method)
      IF (.NOT. C_ASSOCIATED(transport_env%ext_c_method_ptr)) &
         CALL cp_abort(cp__l("transport.F",365), &
                       "MISSING C/C++ ROUTINE: The TRANSPORT section is meant to be used together with an external "// &
                       "program, e.g. the quantum transport code OMEN, that provides CP2K with a density matrix.")

      transport_env%params%n_occ = nelectron_spin
      transport_env%params%n_atoms = natoms
      transport_env%params%energy_diff = energy_diff
      transport_env%params%evoltfactor = evolt
      transport_env%params%extra_scf = extra_scf

      CALL csr_interop_nullify(s_mat)
      CALL csr_interop_nullify(ks_mat)
      CALL csr_interop_nullify(p_mat)

      CALL cp_dbcsr_copy_into_existing(transport_env%template_matrix_sym, matrix_s)
      CALL convert_dbcsr_to_csr_interop(transport_env%template_matrix_sym, transport_env%s_matrix, s_mat)

      CALL cp_dbcsr_copy_into_existing(transport_env%template_matrix_sym, matrix_ks)
      CALL convert_dbcsr_to_csr_interop(transport_env%template_matrix_sym, transport_env%ks_matrix, ks_mat)

      CALL cp_dbcsr_copy_into_existing(transport_env%template_matrix_sym, matrix_p)
      CALL convert_dbcsr_to_csr_interop(transport_env%template_matrix_sym, transport_env%p_matrix, p_mat)

      CALL c_method(transport_env%params, s_mat, ks_mat, p_mat)

      CALL convert_csr_interop_to_dbcsr(p_mat, transport_env%p_matrix, transport_env%template_matrix_nosym)
      CALL cp_dbcsr_copy(matrix_p, transport_env%template_matrix_nosym)

      CALL timestop(handle)

   END SUBROUTINE external_scf_method

! **************************************************************************************************
!> \brief converts a DBCSR matrix to a C-interoperable CSR matrix
!> \param[in]    dbcsr_mat  DBCSR matrix to be converted
!> \param[inout] csr_mat    auxiliary CSR matrix
!> \param[inout] csr_interop_mat C-interoperable CSR matrix
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE convert_dbcsr_to_csr_interop(dbcsr_mat, csr_mat, csr_interop_mat)

      TYPE(cp_dbcsr_type), INTENT(IN)          :: dbcsr_mat
      TYPE(csr_type), INTENT(INOUT)            :: csr_mat
      TYPE(cp2k_csr_interop_type), INTENT(INOUT)      :: csr_interop_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'convert_dbcsr_to_csr_interop', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, mepos, num_pe, mp_group
      INTEGER, ALLOCATABLE, DIMENSION(:)       :: nrows_local_all, first_row_all
      INTEGER(C_INT), DIMENSION(:), POINTER    :: colind_local, rowptr_local, nzerow_local
      REAL(C_DOUBLE), DIMENSION(:), POINTER    :: nzvals_local
      TYPE(cp_logger_type), POINTER            :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      mp_group = logger%para_env%group
      mepos = logger%para_env%mepos
      num_pe = logger%para_env%num_pe

! dbcsr to csr
      CALL cp_convert_dbcsr_to_csr(dbcsr_mat, csr_mat)

! csr to csr_interop
      rowptr_local => csr_mat%rowptr_local
      colind_local => csr_mat%colind_local
      nzerow_local => csr_mat%nzerow_local
      nzvals_local => csr_mat%nzval_local%r_dp ! support real double percision for now

      IF (SIZE(rowptr_local) .EQ. 0) THEN
         csr_interop_mat%rowptr_local = C_NULL_PTR
      ELSE
         csr_interop_mat%rowptr_local = C_LOC(rowptr_local(1))
      END IF

      IF (SIZE(colind_local) .EQ. 0) THEN
         csr_interop_mat%colind_local = C_NULL_PTR
      ELSE
         csr_interop_mat%colind_local = C_LOC(colind_local(1))
      END IF

      IF (SIZE(nzerow_local) .EQ. 0) THEN
         csr_interop_mat%nzerow_local = C_NULL_PTR
      ELSE
         csr_interop_mat%nzerow_local = C_LOC(nzerow_local(1))
      END IF

      IF (SIZE(nzvals_local) .EQ. 0) THEN
         csr_interop_mat%nzvals_local = C_NULL_PTR
      ELSE
         csr_interop_mat%nzvals_local = C_LOC(nzvals_local(1))
      END IF

      ALLOCATE (nrows_local_all(0:num_pe-1), first_row_all(0:num_pe-1))
      CALL mp_allgather(csr_mat%nrows_local, nrows_local_all, mp_group)
      CALL cumsum_i(nrows_local_all, first_row_all)

      IF (mepos .EQ. 0) THEN
         csr_interop_mat%first_row = 0
      ELSE
         csr_interop_mat%first_row = first_row_all(mepos-1)
      END IF
      csr_interop_mat%nrows_total = csr_mat%nrows_total
      csr_interop_mat%ncols_total = csr_mat%ncols_total
      csr_interop_mat%nze_local = csr_mat%nze_local
      IF (csr_mat%nze_total > HUGE(csr_interop_mat%nze_total)) THEN
         CALL cp__b("transport.F",472,"overflow in nze")
      ENDIF
      csr_interop_mat%nze_total = INT(csr_mat%nze_total, KIND=KIND(csr_interop_mat%nze_total))
      csr_interop_mat%nrows_local = csr_mat%nrows_local
      csr_interop_mat%data_type = csr_mat%nzval_local%data_type

      CALL timestop(handle)

   CONTAINS
! **************************************************************************************************
!> \brief cumulative sum of a 1d array of integers
!> \param[in]  arr    input array
!> \param[out] cumsum cumulative sum of the input array
! **************************************************************************************************
      SUBROUTINE cumsum_i(arr, cumsum)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: arr
      INTEGER, DIMENSION(SIZE(arr)), INTENT(OUT)         :: cumsum

      INTEGER                                            :: i

         cumsum(1) = arr(1)
         DO i = 2, SIZE(arr)
            cumsum(i) = cumsum(i-1)+arr(i)
         END DO
      END SUBROUTINE cumsum_i

   END SUBROUTINE convert_dbcsr_to_csr_interop

! **************************************************************************************************
!> \brief converts a C-interoperable CSR matrix to a DBCSR matrix
!> \param[in] csr_interop_mat C-interoperable CSR matrix
!> \param[inout] csr_mat         auxiliary CSR matrix
!> \param[inout] dbcsr_mat       DBCSR matrix
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE convert_csr_interop_to_dbcsr(csr_interop_mat, csr_mat, dbcsr_mat)

      TYPE(cp2k_csr_interop_type), INTENT(IN)            :: csr_interop_mat
      TYPE(csr_type), INTENT(INOUT)                      :: csr_mat
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: dbcsr_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'convert_csr_interop_to_dbcsr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_type, handle, ncols_total, &
                                                            nrows_local, nrows_total, nze_local, &
                                                            nze_total
      INTEGER, DIMENSION(:), POINTER                     :: colind_local, nzerow_local, rowptr_local
      REAL(dp), DIMENSION(:), POINTER                    :: nzvals_local

      CALL timeset(routineN, handle)

! csr_interop to csr
      CALL csr_interop_matrix_get_info(csr_interop_mat, &
                                       nrows_total=nrows_total, ncols_total=ncols_total, nze_local=nze_local, &
                                       nze_total=nze_total, nrows_local=nrows_local, data_type=data_type, &
                                       rowptr_local=rowptr_local, colind_local=colind_local, &
                                       nzerow_local=nzerow_local, nzvals_local=nzvals_local)

      csr_mat%nrows_total = nrows_total
      csr_mat%ncols_total = ncols_total
      csr_mat%nze_local = nze_local
      csr_mat%nze_total = nze_total
      csr_mat%nrows_local = nrows_local
      csr_mat%nzval_local%data_type = data_type

      csr_mat%rowptr_local = rowptr_local
      csr_mat%colind_local = colind_local
      csr_mat%nzerow_local = nzerow_local
      csr_mat%nzval_local%r_dp = nzvals_local

! csr to dbcsr
      CALL cp_convert_csr_to_dbcsr(dbcsr_mat, csr_mat)

      CALL timestop(handle)

   END SUBROUTINE convert_csr_interop_to_dbcsr

! **************************************************************************************************
!> \brief extraxts zeff (effective nuclear charges per atom) and nsgf (the size
!>   of spherical Gaussian basis functions per atom) from qs_env and intializes
!>   the corresponding arrays in transport_env%params
!> \param[in] qs_env qs environment
!> \param[inout] transport_env transport environment
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE transport_set_contact_params(qs_env, transport_env)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(transport_env_type), INTENT(INOUT)            :: transport_env

      CHARACTER(len=*), PARAMETER :: routineN = 'transport_set_contact_params', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iat, ikind, natom, nkind, &
                                                            stride_contacts
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      REAL(KIND=dp)                                      :: zeff
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

      CALL get_qs_env(qs_env, nkind=nkind, natom=natom)
      CALL get_qs_env(qs_env, particle_set=particle_set, &
                      qs_kind_set=qs_kind_set, &
                      atomic_kind_set=atomic_kind_set)

      ALLOCATE (transport_env%nsgf(natom))
      ALLOCATE (transport_env%zeff(natom))
      CALL get_particle_set(particle_set, qs_kind_set, nsgf=transport_env%nsgf)

      ! reference charges
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), zeff=zeff)
         atomic_kind => atomic_kind_set(ikind)
         CALL get_atomic_kind(atomic_kind, atom_list=atom_list)
         DO iat = 1, SIZE(atom_list)
            i = atom_list(iat)
            transport_env%zeff(i) = zeff
         END DO
      END DO

      transport_env%params%nsgf = C_LOC(transport_env%nsgf(1))
      transport_env%params%zeff = C_LOC(transport_env%zeff(1))

      stride_contacts = 6
      DO i = 1, transport_env%params%num_contacts
         IF (transport_env%contacts_data((i-1)*stride_contacts+2) .LT. 0) THEN
            IF (transport_env%contacts_data((i-1)*stride_contacts+5) .EQ. 0) THEN
               transport_env%contacts_data((i-1)*stride_contacts+2) = natom/2; 
            ELSE IF (transport_env%contacts_data((i-1)*stride_contacts+4) .EQ. -1) THEN
               transport_env%contacts_data((i-1)*stride_contacts+2) = &
                  natom-transport_env%contacts_data((i-1)*stride_contacts+3)-transport_env%params%cutout(2)
            ELSE
               transport_env%contacts_data((i-1)*stride_contacts+2) = &
                  transport_env%params%cutout(1)
            END IF
         END IF
      END DO

   END SUBROUTINE transport_set_contact_params

! **************************************************************************************************
!> \brief determine the interaction range of the contact unit cells, the size of the self energy
!>        matrices and the adjacency matrix to partition the system into tridiagonal blocks
!> \param[inout] ks_env ks environment
!> \param[inout] transport_env transport environment
!> \author Sascha Brueck
! **************************************************************************************************
   SUBROUTINE transport_eval_blocks(ks_env, transport_env)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(transport_env_type), INTENT(INOUT)            :: transport_env

      CHARACTER(len=*), PARAMETER :: routineN = 'transport_eval_blocks', &
         routineP = moduleN//':'//routineN

      INTEGER :: colend, contact_bandwidth, contact_injsign, contact_natoms, contact_start, &
         handle, i, iatom, icol, icol_end, icol_start, ikind, irow, jatom, jkind, maxcol, mincol, &
         natom, nkind, num_glo_atom_pairs, num_loc_atom_pairs, num_tridiag_blocks, sigmasize, &
         stride_contacts, tridiag_end, tridiag_start
      INTEGER(KIND=int_1), ALLOCATABLE, DIMENSION(:, :)  :: fulladjmat
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind, glo_atom_pairs, &
                                                            loc_atom_pairs, loc_atom_pairs_big, &
                                                            rcount, rdispl, tridiag_blocks_start
      LOGICAL                                            :: block_exists, do_symmetric
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_list_a, basis_set_list_b
      TYPE(gto_basis_set_type), POINTER                  :: basis_set_a, basis_set_b
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: neighbour_list
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set

! fulladjmat could probably be int_8 but is defined int_1 to save memory, just to be on the safe side

      NULLIFY (atomic_kind_set, qs_kind_set, nl_iterator)
      NULLIFY (basis_set_list_a, basis_set_list_b, basis_set_a, basis_set_b)

      CALL timeset(routineN, handle)

      CALL get_ks_env(ks_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      sab_orb=neighbour_list, &
                      natom=natom, &
                      para_env=para_env)

      ALLOCATE (loc_atom_pairs_big(2*natom*natom))
      num_loc_atom_pairs = 0

      nkind = SIZE(qs_kind_set)
      ALLOCATE (atom_of_kind(natom))
      CALL get_atomic_kind_set(atomic_kind_set, atom_of_kind=atom_of_kind)
      IF(.NOT.(SIZE(neighbour_list) > 0))CALL cp__a("transport.F",667)
      CALL get_neighbor_list_set_p(neighbor_list_sets=neighbour_list, symmetric=do_symmetric)
      IF(.NOT.(do_symmetric))CALL cp__a("transport.F",669)
      ALLOCATE (basis_set_list_a(nkind), basis_set_list_b(nkind))
      CALL basis_set_list_setup(basis_set_list_a, "ORB", qs_kind_set)
      CALL basis_set_list_setup(basis_set_list_b, "ORB", qs_kind_set)

      NULLIFY (nl_iterator)
      CALL neighbor_list_iterator_create(nl_iterator, neighbour_list)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, &
                                ikind=ikind, jkind=jkind, &
                                iatom=iatom, jatom=jatom)

         basis_set_a => basis_set_list_a(ikind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_a)) CYCLE
         basis_set_b => basis_set_list_b(jkind)%gto_basis_set
         IF (.NOT. ASSOCIATED(basis_set_b)) CYCLE

         loc_atom_pairs_big(num_loc_atom_pairs+1) = iatom
         loc_atom_pairs_big(num_loc_atom_pairs+2) = jatom
         num_loc_atom_pairs = num_loc_atom_pairs+2
      ENDDO

      ALLOCATE (loc_atom_pairs(num_loc_atom_pairs))
      DO i = 1, num_loc_atom_pairs
         loc_atom_pairs(i) = loc_atom_pairs_big(i)
      END DO
      DEALLOCATE (loc_atom_pairs_big)
      CALL neighbor_list_iterator_release(nl_iterator)
      DEALLOCATE (basis_set_list_a, basis_set_list_b)

      ALLOCATE (rcount(para_env%num_pe), rdispl(para_env%num_pe))
      CALL mp_allgather(num_loc_atom_pairs, rcount, para_env%group)
      num_glo_atom_pairs = SUM(rcount)
      rdispl(1) = 0
      DO i = 2, para_env%num_pe
         rdispl(i) = rdispl(i-1)+rcount(i-1)
      ENDDO

      ALLOCATE (glo_atom_pairs(num_glo_atom_pairs))
      CALL mp_allgather(loc_atom_pairs, glo_atom_pairs, rcount, rdispl, para_env%group)
      DEALLOCATE (loc_atom_pairs, rcount, rdispl)

      IF (para_env%mepos == para_env%source) THEN
         ALLOCATE (fulladjmat(natom, natom))
         fulladjmat = 0
         DO i = 1, num_glo_atom_pairs, 2
            irow = glo_atom_pairs(i)
            icol = glo_atom_pairs(i+1)
            fulladjmat(irow, icol) = 1
            fulladjmat(icol, irow) = 1
         END DO
         DO i = 1, transport_env%params%num_contacts
            stride_contacts = 6
            contact_bandwidth = transport_env%contacts_data((i-1)*stride_contacts+1)
            contact_start = transport_env%contacts_data((i-1)*stride_contacts+2)
            contact_natoms = transport_env%contacts_data((i-1)*stride_contacts+3)
            contact_injsign = transport_env%contacts_data((i-1)*stride_contacts+4)
            IF (contact_bandwidth .LE. 0) THEN
               contact_bandwidth = 0
               block_exists = .TRUE.
               DO WHILE (block_exists)
                  block_exists = .FALSE.
                  DO irow = contact_start+1, contact_start+contact_natoms
                     icol_start = contact_start+1+contact_injsign*(contact_bandwidth+1)*contact_natoms
                     icol_end = contact_start+contact_natoms+contact_injsign*(contact_bandwidth+1)*contact_natoms
                     DO icol = icol_start, icol_end
                        IF (fulladjmat(irow, icol) .EQ. 1) THEN
                           block_exists = .TRUE.
                           contact_bandwidth = contact_bandwidth+1
                           EXIT
                        END IF
                     END DO
                     IF (block_exists) EXIT
                  END DO
               END DO
               transport_env%contacts_data((i-1)*stride_contacts+1) = contact_bandwidth
            END IF
            tridiag_start = contact_start
            IF (contact_injsign .EQ. -1) tridiag_start = contact_start-(contact_bandwidth-1)*contact_natoms
            mincol = natom
            maxcol = 0
            DO irow = tridiag_start+1, tridiag_start+contact_bandwidth*contact_natoms
               icol_start = tridiag_start+1+contact_injsign*contact_bandwidth*contact_natoms
               icol_end = tridiag_start+contact_bandwidth*contact_natoms+contact_injsign*contact_bandwidth*contact_natoms
               DO icol = icol_start, icol_end
                  IF (fulladjmat(irow, icol) .EQ. 1) THEN
                     IF (icol < mincol) mincol = icol; 
                     IF (icol > maxcol) maxcol = icol; 
                  END IF
               END DO
            END DO
            sigmasize = maxcol-(tridiag_start+contact_bandwidth*contact_natoms)
            IF (contact_injsign .EQ. -1) sigmasize = tridiag_start+1-mincol
            transport_env%contacts_data((i-1)*stride_contacts+6) = sigmasize
         END DO
         ALLOCATE (tridiag_blocks_start(MAX(3, natom)))
         num_tridiag_blocks = 1
         tridiag_blocks_start(1) = transport_env%params%cutout(1)
         tridiag_blocks_start(2) = transport_env%contacts_data(2)+transport_env%contacts_data(6)
         tridiag_end = transport_env%contacts_data(stride_contacts+2)+ &
                       transport_env%contacts_data(stride_contacts+3)- &
                       transport_env%contacts_data(stride_contacts+6)
         maxcol = 0
         DO WHILE (tridiag_blocks_start(num_tridiag_blocks+1) < tridiag_end)
            DO irow = tridiag_blocks_start(num_tridiag_blocks)+1, tridiag_blocks_start(num_tridiag_blocks+1)
               colend = natom-transport_env%params%cutout(2)
               IF (irow .LE. transport_env%contacts_data(1)*transport_env%contacts_data(3)) THEN
                  colend = transport_env%contacts_data(2)+2*transport_env%contacts_data(1)*transport_env%contacts_data(3)
               END IF
               DO icol = irow, colend
                  IF (fulladjmat(irow, icol) .EQ. 1) THEN
                     IF (icol > maxcol) maxcol = icol; 
                  END IF
               END DO
            END DO
            num_tridiag_blocks = num_tridiag_blocks+1
            tridiag_blocks_start(num_tridiag_blocks+1) = maxcol
         END DO
         tridiag_blocks_start(1) = 0
         tridiag_blocks_start(num_tridiag_blocks+1) = natom
         ALLOCATE (transport_env%tridiag_blocks(num_tridiag_blocks))
         DO i = 1, num_tridiag_blocks
            transport_env%tridiag_blocks(i) = tridiag_blocks_start(i+1)-tridiag_blocks_start(i)
         END DO
         DEALLOCATE (tridiag_blocks_start)
         DEALLOCATE (fulladjmat)
      END IF
      DEALLOCATE (glo_atom_pairs)

      CALL mp_bcast(num_tridiag_blocks, para_env%source, para_env%group)
      transport_env%params%n_blocks = num_tridiag_blocks
      IF (para_env%mepos .NE. para_env%source) ALLOCATE (transport_env%tridiag_blocks(num_tridiag_blocks))
      CALL mp_bcast(transport_env%tridiag_blocks, para_env%source, para_env%group)
      transport_env%params%tridiag_blocks = C_LOC(transport_env%tridiag_blocks(1))

      DO i = 1, transport_env%params%num_contacts
         stride_contacts = 6
         CALL mp_bcast(transport_env%contacts_data((i-1)*stride_contacts+1), para_env%source, para_env%group)
         CALL mp_bcast(transport_env%contacts_data((i-1)*stride_contacts+6), para_env%source, para_env%group)
      END DO

      CALL timestop(handle)
   END SUBROUTINE transport_eval_blocks

END MODULE transport

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief K-points and crystal symmetry routines
!> \author jgh
! **************************************************************************************************
MODULE k290
  
   USE kinds,                           ONLY: dp
   USE mathlib,                         ONLY: invmat
   USE string_utilities,                ONLY: xstring
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  IMPLICIT NONE
  PRIVATE
  PUBLIC :: csym_type, pgrp, pgrd
  PUBLIC :: kp_sym_gen, release_csym_type, print_crys_symmetry

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'k290'

! **************************************************************************************************
!> \brief CSM type
!> \par   Content:
!>       PLEVEL . Output level
!>       PUNIT .. Output unit
!>       NTVEC .. Number of primitive cells?
!>       DELTA .. Accuracy in determining symmetry elements
!>       IHG .... Point group of the primitive lattice, holohedral
!>       IHC .... Code distinguishing between hexagonal and cubic groups
!>       ISY .... Code indicating whether the space group is symmorphic
!>       LI ..... Inversions symmetry
!>       NC ..... Total number of elements in the point group
!>       INDPG .. Point group index
!>       IB ..... List of the rotations constituting the point group
!>       V ...... Nonprimitive translations (for nonsymmorphic groups)
!>       XB ..... Cartesian standard origin or coordinates
!>       ORIGIN . Crystal standard origin or coordinates
!>       R ...... List of the 3 x 3 rotation matrices
!>
! **************************************************************************************************
  TYPE csym_type
      INTEGER                                     :: plevel
      INTEGER                                     :: istriz = -1
      INTEGER                                     :: punit
      INTEGER                                     :: nat, nsp, iq1, iq2, iq3, nkpoint, ntvec
      REAL(KIND=dp), DIMENSION (:,:), POINTER     :: xkappa
      INTEGER, DIMENSION (:), POINTER             :: ty
      REAL(KIND=dp)                               :: a1(3), a2(3), a3(3), alat
      REAL(KIND=dp)                               :: strain(3,3) = 0.0_dp
      REAL(KIND=dp)                               :: wvk0(3) = 0.0_dp
      REAL(KIND=dp)                               :: delta, origin(3), xb(3)
      INTEGER                                     :: ihg, ihc, isy, li, nc, indpg
      INTEGER, DIMENSION(48)                      :: ib
      REAL(KIND=dp), DIMENSION(3,48)              :: v
      REAL(KIND=dp), DIMENSION(3,3,48)            :: r
  END TYPE csym_type

!------------------------------------------------------------------------------!
!     Constants
!     Name of 48 rotations (convention Warren-Worlton)
!------------------------------------------------------------------------------!
  CHARACTER(len=10) :: rname_cubic(48) = (/ ' 1        ', ' 2[ 1 0 0]', &
      ' 2[ 0 1 0]', ' 2[ 0 0 1]', ' 3[-1-1-1]', ' 3[ 1 1-1]', ' 3[-1 1 1]', &
      ' 3[ 1-1 1]', ' 3[ 1 1 1]', ' 3[-1 1-1]', ' 3[-1-1 1]', ' 3[ 1-1-1]', &
      ' 2[-1 1 0]', ' 4[ 0 0 1]', ' 4[ 0 0-1]', ' 2[ 1 1 0]', ' 2[ 0-1 1]', &
      ' 2[ 0 1 1]', ' 4[ 1 0 0]', ' 4[-1 0 0]', ' 2[-1 0 1]', ' 4[ 0-1 0]', &
      ' 2[ 1 0 1]', ' 4[ 0 1 0]', '-1        ', '-2[ 1 0 0]', '-2[ 0 1 0]', &
      '-2[ 0 0 1]', '-3[-1-1-1]', '-3[ 1 1-1]', '-3[-1 1 1]', '-3[ 1-1 1]', &
      '-3[ 1 1 1]', '-3[-1 1-1]', '-3[-1-1 1]', '-3[ 1-1-1]', '-2[-1 1 0]', &
      '-4[ 0 0 1]', '-4[ 0 0-1]', '-2[ 1 1 0]', '-2[ 0-1 1]', '-2[ 0 1 1]', &
      '-4[ 1 0 0]', '-4[-1 0 0]', '-2[-1 0 1]', '-4[ 0-1 0]', '-2[ 1 0 1]', &
      '-4[ 0 1 0]'/)
  CHARACTER(len=11) :: rname_hexa(24) = (/ ' 1         ', ' 6[ 0 0. 1]', &
      ' 3[ 0 0. 1]', ' 2[ 0 0. 1]', ' 3[ 0 0.-1]', ' 6[ 0 0.-1]', &
      ' 2[ 0 1. 0]', ' 2[-1 1. 0]', ' 2[ 1 0. 0]', ' 2[ 2 1. 0]', &
      ' 2[ 1 1. 0]', ' 2[ 1 2. 0]', '-1         ', '-6[ 0 0. 1]', &
      '-3[ 0 0. 1]', '-2[ 0 0. 1]', '-3[ 0 0.-1]', '-6[ 0 0.-1]', &
      '-2[ 0 1. 0]', '-2[-1 1. 0]', '-2[ 1 0. 0]', '-2[ 2 1. 0]', &
      '-2[ 1 1. 0]', '-2[ 1 2. 0]'/)
  CHARACTER(len=12) :: icst(7) = (/ '   TRICLINIC', '  MONOCLINIC', &
      'ORTHORHOMBIC', '  TETRAGONAL', '       CUBIC', '    TRIGONAL', &
      '   HEXAGONAL'/)

  CHARACTER(len=5)                         :: pgrp(32)
  DATA pgrp/'    1', '  <1>', '    2', '    m', '  2/m', '    3', &
   '  <3>', '   32', '   3m', ' <3>m', '    4', '  <4>', '  4/m', &
   '  422', '  4mm', '<4>2m', '4/mmm', '    6', '  <6>', '  6/m', &
   '  622', '  6mm', '<6>m2', '6/mmm', '  222', '  mm2', '  mmm', &
   '   23', ' m<3>', '  432', '<4>3m', 'm<3>m'/

  CHARACTER(len=3)                         :: pgrd(32)
  DATA pgrd/'C1 ', 'Ci ', 'C2 ', 'Cs ', 'C2h', 'C3 ', 'S6 ', 'D3 ', &
     'C3v', 'D3d', 'C4 ', 'S4 ', 'C4h', 'D4 ', 'C4v', 'D2d', 'D4h', &
     'C6 ', 'C3h', 'C6h', 'D6 ', 'C6v', 'D3h', 'D6h', 'D2 ', 'C2v', &
     'D2h', 'T  ', 'Th ', 'O  ', 'Td ', 'Oh '/

!------------------------------------------------------------------------------!
!     INDPG group   indpg   group    indpg  group     indpg   group
!        1    1 (C1)    9    3m (C3v)   17  4/mmm(D4h)  25   222(D2)
!        2   <1>(Ci)   10   <3>m(D3d)   18     6 (C6)   26   mm2(C2v)
!        3    2 (C2)   11     4 (C4)    19    <6>(C3h)  27   mmm(D2h)
!        4    m (Cs)   12    <4>(S4)    20    6/m(C6h)  28   23 (T)
!        5   2/m(C2h)  13    4/m(C4h)   21    622(D6)   29  m<3>(Th)
!        6    3 (C3)   14    422(D4)    22    6mm(C6v)  30   432(O)
!        7   <3>(S6)   15    4mm(C4v)   23  <6>m2(D3h)  31 <4>3m(Td)
!        8   32 (D3)   16  <4>2m(D2d)   24  6/mmm(D6h)  32 m<3>m(Oh)
!------------------------------------------------------------------------------!
CONTAINS

! **************************************************************************************************
!> \brief Release the CSYM type
!> \param csym  The CSYM type
! **************************************************************************************************
  SUBROUTINE release_csym_type(csym)
      TYPE(csym_type), POINTER                           :: csym

      CHARACTER(LEN=*), PARAMETER :: routineN = 'release_csym_type', &
         routineP = moduleN//':'//routineN

    IF(.NOT.(ASSOCIATED(csym)))CALL cp__a("k290.F",123)

    IF ( ASSOCIATED(csym%xkappa) ) THEN
       DEALLOCATE(csym%xkappa)
    END IF
    IF ( ASSOCIATED(csym%ty) ) THEN
       DEALLOCATE(csym%ty)
    END IF

    DEALLOCATE(csym)

  END SUBROUTINE release_csym_type

! **************************************************************************************************
!> \brief ...
!> \param csym ...
!> \param coor ...
!> \param types ...
!> \param hmat ...
!> \param nk ...
!> \param shift ...
!> \param stress ...
!> \param symm ...
!> \param delta ...
!> \param wkpoint ...
!> \param xkpoint ...
!> \param iounit ...
! **************************************************************************************************
  SUBROUTINE kp_sym_gen(csym,coor,types,hmat,nk,shift,stress,symm,delta,&
                        wkpoint,xkpoint,iounit)
      TYPE(csym_type), POINTER                           :: csym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: coor
      INTEGER, DIMENSION(:), INTENT(IN)                  :: types
      REAL(KIND=dp), INTENT(IN)                          :: hmat(3,3)
      INTEGER, INTENT(IN), OPTIONAL                      :: nk(3)
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: shift(3), stress(3,3)
      LOGICAL, INTENT(IN), OPTIONAL                      :: symm
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: delta
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: wkpoint
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: xkpoint
      INTEGER, INTENT(IN), OPTIONAL                      :: iounit

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kp_sym_gen', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ik, isum, n, nat, nhash, nkk, &
                                                            nkpts, nsp
      INTEGER, ALLOCATABLE                               :: f0(:,:), includ(:), isc(:), list(:), &
                                                            lrot(:,:), lwght(:)
      REAL(KIND=dp)                                      :: wsum
      REAL(KIND=dp), ALLOCATABLE                         :: rlist(:,:), rx(:,:), tvec(:,:), wvkl(:,:)

    CALL timeset(routineN,handle)

    IF(.NOT.(.NOT.ASSOCIATED(csym)))CALL cp__a("k290.F",176)
    ALLOCATE(csym)

    !..total number of atoms
    nat = SIZE(coor,2)
    !..allocate arrays for coordinates and atom types
    ALLOCATE (csym%xkappa(3,nat))
    ALLOCATE (csym%ty(nat))
    !..count number of atom types
    csym%ty = types
    nsp = 0
    DO
       n = MAXVAL(csym%ty)
       IF (n/=-100) THEN
          nsp = nsp + 1
          WHERE (csym%ty==n) csym%ty = -100
       ELSE
          EXIT
       END IF
    END DO

    !..copy coordinates and atom types
    csym%xkappa = coor
    csym%ty = types
    csym%nat = nat
    csym%nsp = nsp

    !..set values
    csym%a1(:) = hmat(:,1)
    csym%a2(:) = hmat(:,2)
    csym%a3(:) = hmat(:,3)
    IF ( ABS(csym%a1(1)) > 1.e-10_dp ) THEN
       csym%alat = csym%a1(1)
    ELSE
       csym%alat = SQRT(SUM(csym%a1(:)**2))
    END IF
    IF (PRESENT(nk)) THEN
       csym%iq1 = nk(1)
       csym%iq2 = nk(2)
       csym%iq3 = nk(3)
    ELSE
       csym%iq1 = 0
       csym%iq2 = 0
       csym%iq3 = 0
    END IF
    IF (PRESENT(shift)) csym%wvk0 = shift
    IF (PRESENT(stress)) csym%strain = stress
    IF (PRESENT(symm)) THEN
       csym%istriz = -1
       IF (symm) csym%istriz = 1
    END IF
    csym%nkpoint = csym%iq1*csym%iq2*csym%iq3 * 8
    nkk = csym%nkpoint
    !..allocate intermediate arrays
    ALLOCATE (rx(3,nat))
    ALLOCATE (tvec(3,nat))
    ALLOCATE (isc(nat))
    ALLOCATE (f0(49,nat))
    ALLOCATE (wvkl(3,nkk))
    ALLOCATE (lwght(nkk))
    ALLOCATE (lrot(48,nkk))
    ALLOCATE (includ(nkk))
    nhash = MAX(2000,nkk/10)
    ALLOCATE (list(nkk+nhash))
    ALLOCATE (rlist(3,nkk))

    IF ( PRESENT(iounit) ) THEN
       csym%punit = iounit
    ELSE
       csym%punit = -1
    END IF
    IF ( PRESENT(delta) ) THEN
       csym%delta = delta
    ELSE
       csym%delta = 1.e-6_dp
    END IF
    !..calculate symmetry operations and generate kpoints
    CALL k290prg(csym%punit,csym%nat,csym%nkpoint,csym%iq1,csym%iq2,csym%iq3,&
         csym%istriz,csym%a1,csym%a2,csym%a3,csym%alat,csym%strain,csym%xkappa,&
         rx,tvec,csym%ty,isc,f0,csym%ntvec,csym%wvk0,csym%ihg,csym%ihc,csym%isy,&
         csym%li,csym%nc,csym%indpg,csym%ib,csym%v,csym%r,csym%origin,csym%xb,&
         wvkl,lwght,lrot,nhash,includ,list,rlist,csym%delta)
    IF ( PRESENT(xkpoint) ) THEN
       !..return kpoints and weights
       IF(.NOT.(PRESENT(wkpoint)))CALL cp__a("k290.F",260)
       isum=0
       nkpts=0
       DO ik=1,nkk
          IF(lwght(ik) == 0) EXIT
          nkpts=ik
          isum = isum + lwght(nkpts)
       END DO
       wsum=REAL(isum,KIND=dp)
       ALLOCATE (xkpoint(3,nkpts),wkpoint(nkpts))
       DO ik=1,nkpts
          wkpoint(ik) = REAL(lwght(ik),KIND=dp)/wsum
          xkpoint(1:3,ik) = wvkl(1:3,ik)
       END DO
    END IF
    !..deallocate intermediate arrays
    DEALLOCATE (rx,tvec,isc,f0,wvkl,lwght,lrot,includ,list,rlist)

    CALL timestop(handle)

  END SUBROUTINE kp_sym_gen

! **************************************************************************************************
!> \brief ...
!> \param csym ...
! **************************************************************************************************
  SUBROUTINE print_crys_symmetry(csym)
      TYPE(csym_type), POINTER                           :: csym

      CHARACTER(LEN=*), PARAMETER :: routineN = 'print_crys_symmetry', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, indpg, j, k, l, n, plevel, unit
      REAL(KIND=dp)                                      :: vs, vv0(3)

    unit = csym%punit
    IF(unit < 0) RETURN
    plevel = csym%plevel
    indpg=csym%indpg

    vs = 0.0_dp
    DO n = 1, csym%nc
       DO i = 1, 3
          vs = vs + ABS(csym%v(i,n))
       END DO
    END DO

    ! do be determined
    csym%xb=0._dp

    WRITE (unit,"(/,T2,A)") "Crystal Symmetry Information"

    WRITE (unit, '(A,T62,A,A)') "       Point group of the primitive lattice: ",&
         ADJUSTR(icst(csym%ihg))," system"
    IF (csym%isy==0) THEN
       WRITE (unit,'(T62,"nonsymmorphic group")')
    ELSE IF (csym%isy==1) THEN
       WRITE (unit,'(T65,"symmorphic group")')
    ELSE IF (csym%isy==-1) THEN
       WRITE (unit,'(T40,"symmorphic group with non-standard origin")')
       WRITE (unit,'(T7,A,A,/,T18,3F10.6,3X,3F10.6)') &
            ' The standard origin of coordinates is:   ', '[CARTESIAN]   [CRYSTAL]', csym%xb, csym%origin
    ELSE IF (csym%isy==-2) THEN
       WRITE (unit,'(T24,A)') ' Cannot determine if the space group is symmorphic or not'
       WRITE (unit,'(T7,A,/,T7,A,/,T7,A,F15.6,A)') ' The space group is non-symmorphic,', &
            ' or else a non standard origin of coordinates was used.',' (sum of translation vectors=', vs, ')'
    END IF
    IF (csym%li==0) THEN
       WRITE (unit,'(T60,"no inversion symmetry")')
    ELSE IF (csym%li>0) THEN
       WRITE (unit,'(T63,"inversion symmetry")')
    END IF

    IF (indpg>0) THEN
       CALL xstring(pgrp(indpg),i,j)
       CALL xstring(pgrd(indpg),k,l)
       WRITE (unit,'(T7,A,T60,"[INDEX=",I2,"]",T76,A5)') &
            ' The point group of the crystal is [Hermann-Mauguin] ',indpg,ADJUSTL(pgrp(indpg)(i:j))
       WRITE (unit,'(T41,A,T78,A3)') ' [Schoenflies] ',ADJUSTL(pgrd(indpg)(k:l))
    ELSE
       CALL xstring(pgrp(-indpg),i,j)
       CALL xstring(pgrd(-indpg),k,l)
       WRITE (unit,'(T7,A,I2,T40,A,A,"(",A,")",T71,"[INDEX=",I2,"]")') &
            '  Group order=', csym%nc, '    Subgroup of ', &
            pgrp(-indpg) (i:j), pgrd(-indpg) (k:l), -indpg
    END IF
    IF (csym%ntvec>1) THEN
       WRITE (unit,'(T7,A,T75,I6)') ' Number of primitive cells:',csym%ntvec
    END IF

    WRITE (unit,'(7X,A,T78,I3)') &
         "Total number of elements in the point group:",csym%nc
    WRITE (unit,'(7X,"to sum up: (",T64,I1,5I3,")")') &
         csym%ihg, csym%ihc, csym%isy, csym%li, csym%nc, csym%indpg
    WRITE (unit,'(/,7X,"List of the rotations:")')
    WRITE (unit,'(7X,T33,12I4)') (csym%ib(i),i=1,csym%nc)

    IF (csym%isy<=0) THEN
       WRITE (unit,'(7X,"Nonprimitive translations:")')
       WRITE (unit,'(T7,A,A)') ' ROT    V  in the basis A1, A2, A3      ', &
            'V  in cartesian coordinates'
       ! Cartesian components of nonprimitive translation.
       DO i = 1, csym%nc
          DO j = 1, 3
             vv0(j) = csym%v(1,i)*csym%a1(j) + csym%v(2,i)*csym%a2(j) + csym%v(3,i)*csym%a3(j)
          END DO
          WRITE (unit,'(7X,I6,3F10.5,7X,3F10.5)') csym%ib(i), (csym%v(j,i),j=1,3), vv0
       END DO
    END IF

    IF ( MOD(plevel,10) /= 0 ) THEN
       WRITE (unit,'(7X,"List of the 3 X 3 rotation matrices:")')
       IF (csym%ihc==0) THEN
          DO k = 1, csym%nc
             WRITE (unit, '(17X,I3," (",I2,": ",A11,")",2(3F14.6,/,38X),3F14.6)') k, &
                  csym%ib(k), rname_hexa(csym%ib(k)), ((csym%r(i,j,k),j=1,3),i=1,3)
          END DO
       ELSE
          DO k = 1, csym%nc
             WRITE (unit, '(17X,I3," (",I2,": ",A10,") ",2(3F14.6,/,38X),3F14.6)') k, &
                  csym%ib(k), rname_cubic(csym%ib(k)), ((csym%r(i,j,k),j=1,3),i=1,3)
          END DO
       END IF
    END IF

  END SUBROUTINE print_crys_symmetry

! **************************************************************************************************
!> \brief ...
!> \param iout ...
!> \param nat ...
!> \param nkpoint ...
!> \param iq1 ...
!> \param iq2 ...
!> \param iq3 ...
!> \param istriz ...
!> \param a1 ...
!> \param a2 ...
!> \param a3 ...
!> \param alat ...
!> \param strain ...
!> \param xkapa ...
!> \param rx ...
!> \param tvec ...
!> \param ty ...
!> \param isc ...
!> \param f0 ...
!> \param ntvec ...
!> \param wvk0 ...
!> \param ihg ...
!> \param ihc ...
!> \param isy ...
!> \param li ...
!> \param nc ...
!> \param indpg ...
!> \param ib ...
!> \param v ...
!> \param r ...
!> \param xb ...
!> \param origin ...
!> \param wvkl ...
!> \param lwght ...
!> \param lrot ...
!> \param nhash ...
!> \param includ ...
!> \param list ...
!> \param rlist ...
!> \param delta ...
!> \par Original Header:
!> ---------------------------------------------------------------------------!
!>      WRITTEN ON SEPTEMBER 12TH, 1979.
!>      IBM-RETOUCHED ON OCTOBER 27TH, 1980.
!>      GENERATION OF SPECIAL POINTS MODIFIED ON 26-MAY-82 BY OHN.
!>      RETOUCHED ON JANUARY 8TH, 1997
!>      INTEGRATION IN CPMD-FEMD PROGRAM BY THIERRY DEUTSCH
!> ---------------------------------------------------------------------------!
!>      PLAYING WITH SPECIAL POINTS AND CREATION OF 'CRYSTALLOGRAPHIC'
!>      FILE FOR BAND STRUCTURE CALCULATIONS.
!>      GENERATION OF SPECIAL POINTS IN K-SPACE FOR AN ARBITRARY LATTICE,
!>      FOLLOWING THE METHOD MONKHORST,PACK, PHYS. REV. B13 (1976) 5188
!>      MODIFIED BY MACDONALD, PHYS. REV. B18 (1978) 5897
!>      MODIFIED ALSO BY OLE HOLM NIELSEN ("SYMMETRIZATION")
!> ---------------------------------------------------------------------------!
!>      TESTING THEIR EFFICIENCY AND PREPARATION OF THE
!>      "STRUCTURAL" FILE FOR RUNNING THE
!>      SELF-CONSISTENT BAND STRUCTURE PROGRAMS.
!>      IN THE CASES WHERE THE POINT GROUP OF THE CRYSTAL DOES NOT
!>      CONTAIN INVERSION, THE LATTER IS ARTIFICIALLY ADDED, IN ORDER
!>      TO MAKE USE OF THE HERMITICITY OF THE HAMILTONIAN
!> ---------------------------------------------------------------------------!
!>         INPUT:
!>           IOUT    LOGIC FILE NUMBER
!>           NAT     NUMBER OF ATOMS
!>           NKPOINT MAXIMAL NUMBER OF K POINTS
!>           NSP     NUMBER OF SPECIES
!>           IQ1,IQ2,IQ3 THE MONKHORST-PACK MESH PARAMETERS
!>           ISTRIZ  SWITCH FOR SYMMETRIZATION
!>           A1(3),A2(3),A3(3) LATTICE VECTORS
!>           ALAT    LATTICE CONSTANT
!>           STRAIN(3,3)  STRAIN APPLIED TO LATTICE IN ORDER
!>                   TO HAVE K POINTS WITH SYMMETRY OF STRAINED LATTICE
!>           XKAPA(3,NAT)   ATOMS COORDINATES
!>           TY(NAT)      TYPES OF ATOMS
!>           WVK0(3) SHIFT FOR K POINTS MESh (MACDONALD ARTICLE)
!>           NHASH   SIZE OF THE HASH TABLES (LIST)
!>           DELTA   REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
!>                   K-VECTOR < DELTA IS CONSIDERED ZERO
!>         OUTPUT:
!>           RX(3,NAT) SCRATCH ARRAY USED BY GROUP1 ROUTINE
!>           TVEC(1:3,1:NTVEC) TRANSLATION VECTORS (SEE NTVEC)
!>           ISC(NAT)  SCRATCH ARRAY USED BY GROUP1 ROUTINE
!>           F0(49,NAT) ATOM TRANSFORMATION TABLE
!>                      IF NTVEC/=1 THE 49TH GIVES INEQUIVALENT ATOMS
!>           NTVEC NUMBER OF TRANSLATION VECTORS (IF NOT PRIMITIVE CELL)
!>           WVKL(3,NKPOINT) SPECIAL KPOINTS GENERATED
!>           LWGHT(NKPOINT)  WEIGHT FOR EACH K POINT
!>           LROT(48,NKPOINT) SYMMETRY OPERATION FOR EACh K POINTS
!>           INCLUD(NKPOINT)  SCRATCH ARRAY USED BY SPPT2
!>           LIST(NKPOINT+NHASH) HASH TABLE USED BY SPPT2
!>           RLIST(3,NKPOINT) SCRATCH ARRAY USED BY SPPT2
!> ---------------------------------------------------------------------------!
!>      SUBROUTINES NEEDED:
!>         SPPT2, GROUP1, PGL1, ATFTM1, ROT1, STRUCT,
!>         BZRDUC, INBZ, MESH, BZDEFI
!>      (GROUP1, PGL1, ATFTM1, ROT1 FROM THE
!>       "COMPUTER PHYSICS COMMUNICATIONS" PACKAGE "ACMI" - (1971,1974)
!>       WORLTON-WARREN).
! **************************************************************************************************
  SUBROUTINE k290prg(iout,nat,nkpoint,iq1,iq2,iq3,istriz,a1,a2,a3, &
       alat,strain,xkapa,rx,tvec,ty,isc,f0,ntvec,wvk0,ihg,ihc,isy,li,nc,indpg,ib,v,r,&
       xb,origin,wvkl,lwght,lrot,nhash,includ,list,rlist,delta)
      INTEGER                                            :: iout, nat, nkpoint, iq1, iq2, iq3, istriz
      REAL(KIND=dp)                                      :: a1(3), a2(3), a3(3), alat, strain(6), &
                                                            xkapa(3,nat), rx(3,nat), tvec(3,nat)
      INTEGER                                            :: ty(nat), isc(nat), f0(49,nat), ntvec
      REAL(KIND=dp)                                      :: wvk0(3)
      INTEGER                                            :: ihg, ihc, isy, li, nc, indpg, ib(48)
      REAL(KIND=dp)                                      :: v(3,48), r(3,3,48), xb(3), origin(3), &
                                                            wvkl(3,nkpoint)
      INTEGER                                            :: lwght(nkpoint), lrot(48,nkpoint), nhash, &
                                                            includ(nkpoint), list(nkpoint+nhash)
      REAL(KIND=dp)                                      :: rlist(3,nkpoint), delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'k290prg', routineP = moduleN//':'//routineN

      INTEGER                                            :: f00(49,1), i, ib0(48), ihc0, ihg0, &
                                                            indpg0, invadd, istrin, iswght, isy0, &
                                                            j, l, li0, lmax, n, nc0, ntot, ntvec0
      REAL(KIND=dp) :: a01(3), a02(3), a03(3), b01(3), b02(3), b03(3), b1(3), b2(3), b3(3), &
         dtotstr, origin0(3), proj1, proj2, proj3, r0(3,3,48), totstr, tvec0(3,1), v0(3,48), &
         volum, x0(3,1), xb0(3)

    v(:,:) = 0.0_dp
    f00 = 0
    x0 = 0.0_dp
    v0 = 0.0_dp
    !----------------------------------------------------------------------------!
    !   read in lattice structure
    !----------------------------------------------------------------------------!
    DO i = 1, 3
       a01(i) = a1(i)/alat
       a02(i) = a2(i)/alat
       a03(i) = a3(i)/alat
    END DO
    !----------------------------------------------------------------------------!
    !   is the strain significant ?
    !----------------------------------------------------------------------------!
    dtotstr = delta*delta
    totstr = 0.0_dp
    istrin = 0
    DO i = 1, 6
       totstr = totstr + ABS(strain(i))
    END DO
    IF (totstr>dtotstr) istrin = 1
    !----------------------------------------------------------------------------!
    !   Volume of the cell.
    volum = a1(1)*a2(2)*a3(3) + a2(1)*a3(2)*a1(3) + a3(1)*a1(2)*a2(3) - &
         a1(3)*a2(2)*a3(1) - a2(3)*a3(2)*a1(1) - a3(3)*a1(2)*a2(1)
    volum = ABS(volum)
    b1(1) = (a2(2)*a3(3)-a2(3)*a3(2))/volum
    b1(2) = (a2(3)*a3(1)-a2(1)*a3(3))/volum
    b1(3) = (a2(1)*a3(2)-a2(2)*a3(1))/volum
    b2(1) = (a3(2)*a1(3)-a3(3)*a1(2))/volum
    b2(2) = (a3(3)*a1(1)-a3(1)*a1(3))/volum
    b2(3) = (a3(1)*a1(2)-a3(2)*a1(1))/volum
    b3(1) = (a1(2)*a2(3)-a1(3)*a2(2))/volum
    b3(2) = (a1(3)*a2(1)-a1(1)*a2(3))/volum
    b3(3) = (a1(1)*a2(2)-a1(2)*a2(1))/volum
    !----------------------------------------------------------------------------!
    !      GROUP-THEORY ANALYSIS OF LATTICE
    !----------------------------------------------------------------------------!
    CALL group1(a1,a2,a3,nat,ty,xkapa,b1,b2,b3,ihg,ihc,isy,li,nc, &
         indpg,ib,ntvec,v,f0,xb,r,tvec,origin,rx,isc,delta)
    !----------------------------------------------------------------------------!
    DO n = nc + 1, 48
       ib(n) = 0
    END DO
    !----------------------------------------------------------------------------!
    IF (ABS(iq1)+ABS(iq2)+ABS(iq3)/=0) THEN

       invadd = 0
       IF (li==0) invadd = 1
       IF (li==0 .AND. iout>0) THEN
          WRITE (iout,'(A)') &
               ' K290| Although the point group of the crystal does not'
          WRITE (iout,'(A,A)') ' K290| contain inversion, ', &
               'the special point generation algorithm'
          WRITE (iout,'(A)') ' K290| will consider it as a symmetry operation'
          invadd = 1
       END IF
       !----------------------------------------------------------------------------!
       !   GENERATE THE BRAVAIS LATTICE
       !----------------------------------------------------------------------------!
       IF(iout>0) WRITE (iout,'(/,1X,20("-"),A,20("-"),/,A)') &
                      ' The (unstrained) bravais lattice ', &
                      ' (used for generating the largest possible mesh in the B.Z.)'
       !----------------------------------------------------------------------------!
       CALL group1(a01,a02,a03,1,ty,x0,b01,b02,b03,ihg0,ihc0,isy0,li0, &
            nc0,indpg0,ib0,ntvec0,v0,f00,xb0,r0,tvec0,origin0,rx,isc,delta)
       !----------------------------------------------------------------------------!
       !   It is assumed that the same 'type' of symmetry operations
       !   (cubic/hexagonal) will apply to the crystal as well as the Bravais
       !   lattice.
       !----------------------------------------------------------------------------!
       IF(iout>0) THEN
          WRITE (iout,'(/,1X,19("*"),A,25("*"))') &
            ' Generation of special points '
          !   Parameter Q of Monkhorst and Pack, generalized for 3 axes B1,2,3
          WRITE (iout,'(A,/,1X,3I5)') &
            ' Monkhorst-Pack parameters (generalized) IQ1,IQ2,IQ3:', iq1, iq2, &
            iq3
          !   WVK0 is the shift of the whole mesh (see Macdonald)
          WRITE (iout,'(A,/,1X,3F10.5)') &
            ' constant vector shift (MacDonald) of this mesh:', wvk0
          IF (ABS(istriz)/=1) THEN
             WRITE (iout,'(" invalid switch for symmetrization",I10)') istriz
             CALL cp__b("k290.F",596,"")
          END IF
          WRITE (iout,'(" symmetrization switch: ",I3)',ADVANCE="NO") istriz
          IF (istriz==1) THEN
             WRITE (iout,'(" (symmetrization of Monkhorst-Pack mesh)")')
          ELSE
             WRITE (iout,'(" (no symmetrization of Monkhorst-Pack mesh)")')
          END IF
       END IF
       !   Set to 0.
       DO i = 1, nkpoint
          lwght(i) = 0
       END DO
       !----------------------------------------------------------------------------!
       !      Generation of the points (they are not multiplied
       !      by 2*Pi because B1,2,3  were not,either)
       !----------------------------------------------------------------------------!
       IF (nc>nc0) THEN
          !     Due to non-use of primitive cell, the crystal has more
          !     rotations than Bravais lattice.
          !     We use only the rotations for Bravais lattices
          IF (ntvec==1 .AND. iout>0) THEN
             WRITE (iout,*) ' K290| Number of rotations for bravais lattice', &
                  nc0
             WRITE (iout,*) ' K290| Number of rotations for crystal lattice', &
                  nc
             WRITE (iout,*) ' K290| No duplication found'
             CALL cp__b("k290.F",623,"")
          END IF
          nc = nc0
          DO i = 1, nc0
             ib(i) = ib0(i)
          END DO
          IF(iout>0) THEN
             WRITE (iout,'(/,1X,20("!"),"WARNING",20("!"))')
             WRITE (iout,'(A)') &
                  ' K290| The crystal has more symmetry than the bravais lattice'
             WRITE (iout,'(A)') ' K290| Because this is not a primitive cell'
             WRITE (iout,'(A)') ' K290| Use only symmetry from bravais lattice'
             WRITE (iout,'(1X,20("!"),"WARNING",20("!"),/)')
          END IF
       END IF
       CALL sppt2(iout,iq1,iq2,iq3,wvk0,nkpoint,a01,a02,a03,b01,b02,b03, &
            invadd,nc,ib,r,ntot,wvkl,lwght,lrot,nc0,ib0,istriz,nhash,includ, &
            list,rlist,delta)
       !----------------------------------------------------------------------------!
       !      Check on error signals
       !----------------------------------------------------------------------------!
       IF(iout>0) WRITE (iout,'(/,1X,I5," Special points generated")') ntot
       IF (ntot/=0) THEN
          IF (ntot<0) THEN
             IF(iout>0) WRITE (iout,'(A,I5,/,A,/,A)') ' Dimension nkpoint =', nkpoint, &
                  ' insufficient for accommodating all the special points', &
                  ' what follows is an incomplete list'
             ntot = iabs(ntot)
          END IF
          !   Before using the list WVKL as wave vectors, they have to be
          !   multiplied by 2*Pi
          !   The list of weights LWGHT is not normalized
          iswght = 0
          DO i = 1, ntot
             iswght = iswght + lwght(i)
          END DO
          IF(iout>0) WRITE (iout,'(8X,A,T33,A,4X,A)') 'WAVEVECTOR K', 'WEIGHT', &
               'UNFOLDING ROTATIONS'
          !   Set near-zeroes equal to zero:
          DO l = 1, ntot
             DO i = 1, 3
                IF (ABS(wvkl(i,l))<delta) wvkl(i,l) = 0.0_dp
             END DO
             IF (istrin/=0) THEN
                !       Express special points in (unstrained) basis.
                proj1 = 0.0_dp
                proj2 = 0.0_dp
                proj3 = 0.0_dp
                DO i = 1, 3
                   proj1 = proj1 + wvkl(i,l)*a01(i)
                   proj2 = proj2 + wvkl(i,l)*a02(i)
                   proj3 = proj3 + wvkl(i,l)*a03(i)
                END DO
                DO i = 1, 3
                   wvkl(i,l) = proj1*b1(i) + proj2*b2(i) + proj3*b3(i)
                END DO
             END IF
             lmax = lwght(l)
             IF(iout>0) THEN
                WRITE (iout,fmt='(1X,I5,3F8.4,I8,T42,12I3)') l, &
                     (wvkl(i,l),i=1,3), lwght(l), (lrot(i,l),i=1,MIN(lmax,12))
                DO j = 13, lmax, 12
                   WRITE (iout,fmt='(T42,12I3)') (lrot(i,l),i=j,MIN(lmax,j-1+12))
                END DO
             END IF
          END DO
          IF(iout>0) WRITE (iout,'(24X,"TOTAL:",I8)') iswght
       END IF
    END IF
  END SUBROUTINE k290prg

! **************************************************************************************************
!> \brief ...
!> \param a1 ...
!> \param a2 ...
!> \param a3 ...
!> \param nat ...
!> \param ty ...
!> \param x ...
!> \param b1 ...
!> \param b2 ...
!> \param b3 ...
!> \param ihg ...
!> \param ihc ...
!> \param isy ...
!> \param li ...
!> \param nc ...
!> \param indpg ...
!> \param ib ...
!> \param ntvec ...
!> \param v ...
!> \param f0 ...
!> \param xb ...
!> \param r ...
!> \param tvec ...
!> \param origin ...
!> \param rx ...
!> \param isc ...
!> \param delta ...
!> \par Original Header:
!> ---------------------------------------------------------------------------!
!>       WRITTEN ON SEPTEMBER 10TH - FROM THE ACMI COMPLEX
!>       (WORLTON AND WARREN, COMPUT.PHYS.COMMUN. 8,71-84 (1974))
!>       (AND 3,88-117 (1972))
!>       BASIC CRYSTALLOGRAPHIC INFORMATION
!>       ABOUT A GIVEN CRYSTAL STRUCTURE.
!>       SUBROUTINES NEEDED: PGL1,ATFTM1,ROT1,RLV3
!> ---------------------------------------------------------------------------!
!>       INPUT DATA:
!>       IOUT ... NUMBER OF THE OUTPUT UNIT FOR ON-LINE PRINTING
!>                OF VARIOUS MESSAGES
!>                IF IOUT.LE.0 NO MESSAGE
!>       A1,A2,A3 .. ELEMENTARY TRANSLATIONS OF THE LATTICE, IN SOME
!>                UNIT OF LENGTH
!>       NAT .... NUMBER OF ATOMS IN THE UNIT CELL
!>                ALL THE DIMENSIONS ARE SET FOR NAT .LE. 20
!>       TY ..... INTEGERS DISTINGUISHING BETWEEN THE ATOMS OF
!>                DIFFERENT TYPE. TY(I) IS THE TYPE OF THE I-TH ATOM
!>                OF THE BASIS
!>       X ...... CARTESIAN COORDINATES OF THE NAT ATOMS OF THE BASIS
!>       DELTA... REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
!> ---------------------------------------------------------------------------!
!>       OUTPUT DATA:
!>       B1,B2,B3 .. RECIPROCAL LATTICE VECTORS, NOT MULTIPLIED BY
!>                ANY 2PI, IN UNITS RECIPROCAL TO THOSE OF A1,A2,A3
!>       IHG .... POINT GROUP OF THE PRIMITIVE LATTICE, HOLOHEDRAL
!>                GROUP NUMBER:
!>                IHG=1 STANDS FOR TRICLINIC    SYSTEM
!>                IHG=2 STANDS FOR MONOCLINIC   SYSTEM
!>                IHG=3 STANDS FOR ORTHORHOMBIC SYSTEM
!>                IHG=4 STANDS FOR TETRAGONAL   SYSTEM
!>                IHG=5 STANDS FOR CUBIC        SYSTEM
!>                IHG=6 STANDS FOR TRIGONAL     SYSTEM
!>                IHG=7 STANDS FOR HEXAGONAL    SYSTEM
!>       IHC .... CODE DISTINGUISHING BETWEEN HEXAGONAL AND CUBIC
!>                GROUPS
!>                IHC=0 STANDS FOR HEXAGONAL GROUPS
!>                IHC=1 STANDS FOR CUBIC GROUPS
!>       ISY .... CODE INDICATING WHETHER THE SPACE GROUP IS
!>                SYMMORPHIC OR NONSYMMORPHIC
!>                ISY= 0 NONSYMMORPHIC GROUP
!>                ISY= 1 SYMMORPHIC GROUP
!>                ISY=-1 SYMMORPHIC GROUP WITH NON-STANDARD ORIGIN
!>                ISY=-2 UNDETERMINED (NORMALLY NEVER)
!>                THE GROUP IS CONSIDERED SYMMORPHIC IF FOR EACH
!>                OPERATION OF THE POINT GROUP THE SUM OF THE 3
!>                COMPONENTS OF ABS(V(N)) (NONPRIMITIVE TRANSLATION,
!>                SEE BELOW) IS LT. 0.0001
!>       ORIGIN   STANDARD ORIGIN IF SYMMORPHIC (CRYSTAL COORDINATES)
!>       LI ..... CODE INDICATING WHETHER THE POINT GROUP
!>                OF THE CRYSTAL CONTAINS INVERSION OR NOT
!>                (OPERATIONS 13 OR 25 IN RESPECTIVELY HEXAGONAL
!>                OR CUBIC GROUPS).
!>                LI=0 MEANS: DOES NOT CONTAIN INVERSION
!>                LI.GT.0 MEANS: THERE IS INVERSION IN THE POINT
!>                               GROUP OF THE CRYSTAL
!>       NC ..... TOTAL NUMBER OF ELEMENTS IN THE POINT GROUP OF THE
!>                CRYSTAL
!>       INDPG .. POINT GROUP INDEX (DETERMINED IF SYMMORPHIC GROUP)
!>       IB ..... LIST OF THE ROTATIONS CONSTITUTING THE POINT GROUP
!>                OF THE CRYSTAL. THE NUMBERING IS THAT DEFINED IN
!>                WORLTON AND WARREN, I.E. THE ONE MATERIALIZED IN THE
!>                ARRAY R (SEE BELOW)
!>                ONLY THE FIRST NC ELEMENTS OF THE ARRAY IB ARE
!>                MEANINGFUL
!>       NTVEC .. NUMBER OF TRANSLATIONAL VECTORS
!>                ASSOCIATED WITH IDENTITY OPERATOR I.E.
!>                GIVES THE NUMBER OF IDENTICAL PRIMITIVE CELLS
!>       V ...... NONPRIMITIVE TRANSLATIONS (IN THE CASE OF NONSYMMOR-
!>                PHIC GROUPS). V(I,N) IS THE I-TH COMPONENT
!>                OF THE TRANSLATION CONNECTED WITH THE N-TH ELEMENT
!>                OF THE POINT GROUP (I.E. WITH THE ROTATION
!>                NUMBER IB(N) ).
!>                ATTENTION: V(I) ARE NOT CARTESIAN COMPONENTS,
!>                THEY REFER TO THE SYSTEM A1,A2,A3.
!>       F0 ..... THE FUNCTION DEFINED IN MARADUDIN, IPATOVA BY
!>                EQ. (3.2.12): ATOM TRANSFORMATION TABLE.
!>                THE ELEMENT F0(N,KAPA) MEANS THAT THE N-TH
!>                OPERATION OF THE SPACE GROUP (I.E. OPERATION NUMBER
!>                IB(N), TOGETHER WITH AN EVENTUAL NONPRIMITIVE
!>                TRANSLATION  V(N)) TRANSFERS THE ATOM KAPA INTO THE
!>                ATOM F0(N,KAPA).
!>                THE 49TH LINE GIVES EQUIVALENT ATOMS FOR
!>                FRACTIONAl TRANSLATIONS ASSOCIATED WITH IDENTITY
!>       R ...... LIST OF THE 3 X 3 ROTATION MATRICES
!>                (XYZ REPRESENTATION OF THE O(H) OR D(6)H GROUPS)
!>                ALL 48 OR 24 MATRICES ARE LISTED.
!>                FOLLOW NOTATION OF WORLTON-WARREN(1972)
!>       TVEC  .. LIST OF NTVEC TRANSLATIONAL VECTORS
!>                ASSOCIATED WITH IDENTITY OPERATOR
!>                TVEC(1:3,1) = \(0,0,0\)
!>                (CRYSTAL COORDINATES)
!>       RX ..... SCRATCH ARRAY
!>       ISC .... SCRATCH ARRAY
!> ---------------------------------------------------------------------------!
!>       PRINTED OUTPUT:
!>       PROGRAM PRINTS THE TYPE OF THE LATTICE (IHG, IN WORDS),
!>       LISTS THE OPERATIONS OF THE  POINT GROUP OF THE
!>       CRYSTAL, INDICATES WHETHER THE SPACE GROUP IS SYMMORPHIC OR
!>       NONSYMMORPHIC AND WHETHER THE POINT GROUP OF THE CRYSTAL
!>       CONTAINS INVERSION.
! **************************************************************************************************
  SUBROUTINE group1(a1,a2,a3,nat,ty,x,b1,b2,b3,ihg,ihc,isy,li,nc, &
       indpg,ib,ntvec,v,f0,xb,r,tvec,origin,rx,isc,delta)
      REAL(KIND=dp)                                      :: a1(3), a2(3), a3(3)
      INTEGER                                            :: nat, ty(nat)
      REAL(KIND=dp)                                      :: x(3,nat), b1(3), b2(3), b3(3)
      INTEGER                                            :: ihg, ihc, isy, li, nc, indpg, ib(48), &
                                                            ntvec
      REAL(KIND=dp)                                      :: v(3,48)
      INTEGER                                            :: f0(49,nat)
      REAL(KIND=dp)                                      :: xb(3), r(3,3,48), tvec(3,nat), &
                                                            origin(3), rx(3,nat)
      INTEGER                                            :: isc(nat)
      REAL(KIND=dp)                                      :: delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'group1', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ncprim
      REAL(KIND=dp)                                      :: a(3,3), ai(3,3), ap(3,3), api(3,3)

    DO i = 1, 3
       a(i,1) = a1(i)
       a(i,2) = a2(i)
       a(i,3) = a3(i)
    END DO
    !------------------------------------------------------------------------------!
    !        A(I,J) IS THE I-TH CARTESIAN COMPONENT OF THE J-TH PRIMITIVE
    !        TRANSLATION VECTOR OF THE DIRECT LATTICE
    !        TY(I) IS AN INTEGER DISTINGUISHING ATOMS OF DIFFERENT TYPE,
    !        I.E., DIFFERENT ATOMIC SPECIES
    !        X(J,I) IS THE J-TH CARTESIAN COMPONENT OF THE POSITION
    !        VECTOR FOR THE I-TH ATOM IN THE UNIT CELL.
    !------------------------------------------------------------------------------!
    !     DETERMINE PRIMITIVE LATTICE VECTORS FOR THE RECIPROCAL LATTICE
    !------------------------------------------------------------------------------!

    CALL calbrec(a,ai)
    DO i = 1, 3
       b1(i) = ai(1,i)
       b2(i) = ai(2,i)
       b3(i) = ai(3,i)
    END DO
    !------------------------------------------------------------------------------!
    !     Determination of the translation vectors associated with
    !     the Identity matrix i.e. if the cell is duplicated
    !     Give also the ``primitive lattice''
    CALL primlatt(a,ai,ap,api,nat,ty,x,ntvec,tvec,f0,isc,delta)
    !------------------------------------------------------------------------------!
    !     Determination of the holohedral group (and crystal system)
    CALL pgl1(ap,api,ihc,nc,ib,ihg,r,delta)
    IF (ntvec>1) THEN
       !       All rotations found by PGL1 have axes in x, y or z cart. axis
       !       So we have too check if we do not loose symmetry
       ncprim = nc
       !       The hexagonal system is found if the z axis is the sixfold axis
       CALL pgl1(a,ai,ihc,nc,ib,ihg,r,delta)
       IF (ncprim>nc) THEN
          !         More symmetry with
          CALL pgl1(ap,api,ihc,nc,ib,ihg,r,delta)
       END IF
    END IF
    !------------------------------------------------------------------------------!
    !     Determination of the space group
    CALL atftm1(r,v,x,f0,xb,origin,ib,ty,nat,ihg,rx,nc,indpg,ntvec,a,ai, &
         li,isy,isc,delta)
    !------------------------------------------------------------------------------!
  END SUBROUTINE group1

! **************************************************************************************************
!> \brief CALCULATE RECIPROCAL VECTOR BASIS (AI(1:3,1:3))
!> \param a ...
!> \param ai ...
!> \par   Note
!>             INPUT:
!>              A(3,3) A(I,J) IS THE I-TH CARTESIAN COMPONENT
!>                OF THE J-TH PRIMITIVE TRANSLATION VECTOR OF
!>                THE DIRECT LATTICE
!>             OUTPUT:
!>              AI(3,3) RECIPROCAL VECTOR BASIS
! **************************************************************************************************
  SUBROUTINE calbrec(a,ai)
      REAL(KIND=dp)                                      :: a(3,3), ai(3,3)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calbrec', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, il, iu, j, jl, ju
      REAL(KIND=dp)                                      :: det

    det = a(1,1)*a(2,2)*a(3,3) + a(2,1)*a(1,3)*a(3,2) + &
         a(3,1)*a(1,2)*a(2,3) - a(1,1)*a(2,3)*a(3,2) - a(2,1)*a(1,2)*a(3,3) - &
         a(3,1)*a(1,3)*a(2,2)
    det = 1.E0_dp/det
    DO i = 1, 3
       il = 1
       iu = 3
       IF (i==1) il = 2
       IF (i==3) iu = 2
       DO j = 1, 3
          jl = 1
          ju = 3
          IF (j==1) jl = 2
          IF (j==3) ju = 2
          ai(j,i) = (-1.0_dp)**(i+j)*det*(a(il,jl)*a(iu,ju)-a(il,ju)*a(iu,jl &
               ))
       END DO
    END DO
  END SUBROUTINE calbrec

! **************************************************************************************************
!> \brief DETERMINATION OF THE TRANSLATION VECTORS ASSOCIATED WITH
!>        THE IDENTITY SYMMETRY I.E. IF THE CELL IS DUPLICATED
!>        GIVE ALSO THE PRIMITIVE DIRECT AND RECIPROCAL LATTICE VECTOR
!> \param a ...
!> \param ai ...
!> \param ap ...
!> \param api ...
!> \param nat ...
!> \param ty ...
!> \param x ...
!> \param ntvec ...
!> \param tvec ...
!> \param f0 ...
!> \param isc ...
!> \param delta ...
!> \par Note
!>        INPUT:
!>          A(3,3)   A(I,J) IS THE I-TH CARTESIAN COMPONENT
!>                   OF THE J-TH TRANSLATION VECTOR OF
!>                   THE DIRECT LATTICE
!>          AI(3,3)  RECIPROCAL VECTOR BASIS (CARTESIAN)
!>          NAT      NUMBER OF ATOMS
!>          TY(NAT)  TYPE OF ATOMS
!>          X(3,NAT) ATOMIC COORDINATES IN CARTESIAN COORDINATES
!>          DELTA    REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
!>        OUTPUT:
!>          AP(3,3)  COMPONENTS OF THE PRIMITIVE TRANSLATION VECTORS
!>          API(3,3) PRIMITIVE RECIPROCAL BASIS VECTORS
!>                   BOTH BAISI ARE IN CARTESIAN COORDINATES
!>          NTVEC    NUMBER OF TRANSLATION VECTORS (FRACTIONNAL)
!>          TVEC(3,NTVEC) COMPONENTS OF TRANSLATIONAL VECTORS
!>                        (CRYSTAL COORDINATES)
!>          F0(49,NAT)    GIVES INEQUIVALENT ATOM FOR EACH ATOM
!>                        THE 49-TH LINE
!>          ISC(NAT)      SCRATCH ARRAY
! **************************************************************************************************
  SUBROUTINE primlatt(a,ai,ap,api,nat,ty,x,ntvec,tvec,f0,isc,delta)
      REAL(KIND=dp)                                      :: a(3,3), ai(3,3), ap(3,3), api(3,3)
      INTEGER                                            :: nat, ty(nat)
      REAL(KIND=dp)                                      :: x(3,nat)
      INTEGER                                            :: ntvec
      REAL(KIND=dp)                                      :: tvec(3,nat)
      INTEGER                                            :: f0(49,nat), isc(nat)
      REAL(KIND=dp)                                      :: delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'primlatt', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, il, iv, j, k2
      LOGICAL                                            :: oksym
      REAL(KIND=dp)                                      :: vr(3), xb(3)

    ntvec = 1
    tvec(1,1) = 0.0_dp
    tvec(2,1) = 0.0_dp
    tvec(3,1) = 0.0_dp
    !     First we check if there exist fractional translational vectors
    !     associated with Identity operation i.e.
    !     if the cell is duplicated or not.
    DO i = 1, nat
       f0(49,i) = i
    END DO
    DO k2 = 2, nat
       IF (ty(1)/=ty(k2)) CYCLE
       DO i = 1, 3
          xb(i) = x(i,k2) - x(i,1)
       END DO
       !       A fractional translation vector VR is defined.
       CALL rlv3(ai,xb,vr,il,delta)
       CALL checkrlv3(1,nat,ty,x,x,vr,f0,ai,isc,.TRUE.,oksym,delta)
       IF (oksym) THEN
          !         A fractional translational vector is found
          ntvec = ntvec + 1
          !         F0(49,1:NAT) gives number of equivalent atoms
          !         and has atom indexes of inequivalent atoms (for translation)
          DO i = 1, nat
             IF (f0(49,i)>f0(1,i)) f0(49,i) = f0(1,i)
          END DO
          DO i = 1, 3
             tvec(i,ntvec) = vr(i)
          END DO
       END IF
    END DO
    !------------------------------------------------------------------------------!
    DO i = 1, 3
       DO il = 1, 3
          ap(il,i) = a(il,i)
       END DO
    END DO
    DO j = 1, 3
       DO i = 1, 3
          api(i,j) = ai(i,j)
       END DO
    END DO
    IF (ntvec==1) THEN
       !       The current cell is definitely a primitive one
       !       Copy A and AI to AP and API
    ELSE
       !       We are looking for the primitive lattice vector basis set
       !       AP is our current lattice vector basis
       DO iv = 2, ntvec
          !         TVEC in cartesian coordinates
          DO i = 1, 3
             xb(i) = tvec(1,iv)*a(i,1) + tvec(2,iv)*a(i,2) + &
                  tvec(3,iv)*a(i,3)
          END DO
          !         We calculare TVEC in AP basis
          CALL rlv3(api,xb,vr,il,delta)
          DO i = 1, 3
             IF (ABS(vr(i))>delta) THEN
                il = NINT(1.0_dp/ABS(vr(i)))
                IF (il>1) THEN
                   !               We replace AP(1:3,I) by TVEC(1:3,IV)
                   DO j = 1, 3
                      ap(j,i) = xb(j)
                   END DO
                   !               Calculate new API
                   CALL calbrec(ap,api)
                   EXIT
                END IF
             END IF
          END DO
       END DO
    END IF
  END SUBROUTINE primlatt

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param ai ...
!> \param ihc ...
!> \param nc ...
!> \param ib ...
!> \param ihg ...
!> \param r ...
!> \param delta ...
!> \par Original Header
!> ---------------------------------------------------------------------------!
!>         WRITTEN ON SEPTEMBER 11TH, 1979 - FROM ACMI COMPLEX
!>         AUXILIARY SUBROUTINE TO GROUP1
!>         SUBROUTINE PGL DETERMINES THE POINT GROUP OF THE LATTICE
!>         AND THE CRYSTAL SYSTEM.
!>         SUBROUTINES NEEDED: ROT1, RLV3
!> ---------------------------------------------------------------------------!
!>         WARNING: FOR THE HEXAGONAL SYSTEM, THE 3RD AXIS SUPPOSE
!>                  TO BE THE SIX-FOLD AXIS
!> ---------------------------------------------------------------------------!
!>         INPUT:
!>         A ..... DIRECT LATTICE VECTORS
!>         AI .... RECIPROCAL LATTICE VECTORS
!>         DELTA.. REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
!> ---------------------------------------------------------------------------!
!>         OUTPUT:
!>         IHC .... CODE DISTINGUISHING BETWEEN HEXAGONAL AND CUBIC
!>                  GROUPS
!>                  IHC=0 STANDS FOR HEXAGONAL GROUPS
!>                  IHC=1 STANDS FOR CUBIC GROUPS
!>         NC .... NUMBER OF ROTATIONS IN THE POINT GROUP
!>         IB .... SET OF ROTATION
!>         IHG .... POINT GROUP OF THE PRIMITIVE LATTICE, HOLOHEDRAL
!>                  GROUP NUMBER:
!>                  IHG=1 STANDS FOR TRICLINIC    SYSTEM
!>                  IHG=2 STANDS FOR MONOCLINIC   SYSTEM
!>                  IHG=3 STANDS FOR ORTHORHOMBIC SYSTEM
!>                  IHG=4 STANDS FOR TETRAGONAL   SYSTEM
!>                  IHG=5 STANDS FOR CUBIC        SYSTEM
!>                  IHG=6 STANDS FOR TRIGONAL     SYSTEM
!>                  IHG=7 STANDS FOR HEXAGONAL    SYSTEM
!>         R ...... LIST OF THE 3 X 3 ROTATION MATRICES
!>                  (XYZ REPRESENTATION OF THE O(H) OR D(6)H GROUPS)
!>                  ALL 48 OR 24 MATRICES ARE LISTED.
!>                  FOLLOW NOTATION OF WORLTON-WARREN(1972)
! **************************************************************************************************
  SUBROUTINE pgl1(a,ai,ihc,nc,ib,ihg,r,delta)
      REAL(KIND=dp)                                      :: a(3,3), ai(3,3)
      INTEGER                                            :: ihc, nc, ib(48), ihg
      REAL(KIND=dp)                                      :: r(3,3,48), delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'pgl1', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, lx, n, nr
      REAL(KIND=dp)                                      :: tr, vr(3), xa(3)

    DO ihc = 0, 1
       !       IHC is 0 for hexagonal groups and 1 for cubic groups.
       IF (ihc==0) THEN
          nr = 24
       ELSE
          nr = 48
       END IF
       nc = 0
       !       Constructs rotation operations.
       CALL rot1(ihc,r)
       nloop: DO n = 1, nr
          ib(n) = 0
          !         Rotate the A1,2,3 vectors by rotation No. N
          DO k = 1, 3
             DO i = 1, 3
                xa(i) = 0.0_dp
                DO j = 1, 3
                   xa(i) = xa(i) + r(i,j,n)*a(j,k)
                END DO
             END DO
             CALL rlv3(ai,xa,vr,lx,delta)
             tr = 0.0_dp
             DO i = 1, 3
                tr = tr + ABS(vr(i))
             END DO
             !           If VR.ne.0, then XA cannot be a multiple of a lattice vector
             IF (tr>delta) CYCLE nloop
          END DO
          nc = nc + 1
          ib(nc) = n
       END DO nloop
       !------------------------------------------------------------------------------!
       !       IHG stands for holohedral group number.
       IF (ihc==0) THEN
          !         Hexagonal group:
          IF (nc==12) ihg = 6
          IF (nc>12) ihg = 7
          IF (nc>=12) RETURN
          !         Too few operations, try cubic group: (IHC=1,NR=48)
       ELSE
          !         Cubic group:
          IF (nc<4) ihg = 1
          IF (nc==4) ihg = 2
          IF (nc>4) ihg = 3
          IF (nc==16) ihg = 4
          IF (nc>16) ihg = 5
          RETURN
       END IF
    END DO
  END SUBROUTINE pgl1

! **************************************************************************************************
!> \brief ...
!> \param ai ...
!> \param xb ...
!> \param vr ...
!> \param il ...
!> \param delta ...
!> \par Original Header
!> ---------------------------------------------------------------------------!
!>         WRITTEN ON SEPTEMBER 11TH, 1979 - FROM ACMI COMPLEX
!>         AUXILIARY SUBROUTINE TO GROUP1
!>         SUBROUTINE RLV REMOVES A DIRECT LATTICE VECTOR
!>         FROM XB LEAVING THE REMAINDER IN VR.
!>         IF A NONZERO LATTICE VECTOR WAS REMOVED, IL IS MADE NONZERO.
!>         VR STANDS FOR V-REFERENCE.
!> ---------------------------------------------------------------------------!
!>         INPUT:
!>            AI(I,J) ARE THE RECIPROCAL LATTICE VECTORS,
!>                    B(I) = AI(I,J),J=1,2,3
!>            XB(1:3) VECTOR IN CARTESIAN COORDINATES
!>         DELTA      REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
!>         OUTPUT:
!>            VR IS NOT GIVEN IN CARTESIAN COORDINATES BUT
!>               IN THE SYSTEM A1,A2,A3 (CRYSTAL COORDINATES)
!>               AND BETWEEN -1/2 AND 1/2
!>            IL ABS OF VR
!>         K.K., 23.10.1979
! **************************************************************************************************
  SUBROUTINE rlv3(ai,xb,vr,il,delta)

      REAL(KIND=dp)                                      :: ai(3,3), xb(3), vr(3)
      INTEGER                                            :: il
      REAL(KIND=dp)                                      :: delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rlv3', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      REAL(KIND=dp)                                      :: ts

    il = 0
    DO i = 1, 3
       vr(i) = 0.0_dp
    END DO
    ts = ABS(xb(1)) + ABS(xb(2)) + ABS(xb(3))
    IF (ts<=delta) RETURN
    DO i = 1, 3
       DO j = 1, 3
          vr(i) = vr(i) + ai(i,j)*xb(j)
       END DO
       il = il + NINT(ABS(vr(i)))
       !       Change in order to have correct determination of origin and
       !       symmorphic group (T.D 30/03/98)
       !        VR(I) = - MOD(REAL(VR(I),KIND=dp),1.0_dp)
       vr(i) = NINT(vr(i)) - vr(i)
    END DO
  END SUBROUTINE rlv3

! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param v ...
!> \param x ...
!> \param f0 ...
!> \param xb ...
!> \param origin ...
!> \param ib ...
!> \param ty ...
!> \param nat ...
!> \param ihg ...
!> \param rx ...
!> \param nc ...
!> \param indpg ...
!> \param ntvec ...
!> \param a ...
!> \param ai ...
!> \param li ...
!> \param isy ...
!> \param isc ...
!> \param delta ...
!> \par Original Header
!> ---------------------------------------------------------------------------!
!>         WRITTEN ON SEPTEMBER 11TH, 1979 - FROM ACMI COMPLEX
!>         AUXILIARY SUBROUTINE TO GROUP1
!>         SUBROUTINE ATFTMT DETERMINES
!>                    THE POINT GROUP OF THE CRYSTAL,
!>                    THE ATOM TRANSFORMATION TABLE,F0,
!>                    THE FRACTIONAL TRANSLATIONS,V,
!>                    ASSOCIATED WITH EACH ROTATION.
!>         SUBROUTINES NEEDED: RLV3 CHECKRLV3 SYMMORPHIC STOPGM XSTRING
!>         MAY 14TH,1998: A LOT OF CHANGES (ARGUMENTS)
!>                        BETTER DETERMINATION OF V
!>         SEP 15TH,1998: DETERMINATION OF FRACTIONAL TRANSLATIONAL VEC.
!> ---------------------------------------------------------------------------!
!>         INPUT:
!>           IOUT Logical file number (output)
!>                If IOUT.LE.0 no message
!>           IHG  Holohedral group number (determined by PGL1)
!>           NC   Number of rotation operations
!>           NAT  Number of atoms (used in the routine)
!>           X    Coordinates of atoms (cartesian)
!>           TY   Type of atoms
!>           R    Sets of transformation operations (cartesian)
!>           IB   Index giving NC operations in R
!>           AI   Reciprocal lattice vectors
!>           NTVEC Number of translational vectors
!>                associated with Identity
!>                if primitive cell NTVEC=1, TVEC=(0,0,0)
!>         DELTA  REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
!>         OUTPUT:
!>           RX(3,NAT) Scratch array
!>           ISC(NAT)  Scratch array
!>           NC        is modified (number of symmetry operations)
!>           INDPG     Point group index
!>           V(3,48)   The fractional translations associated
!>                     with each rotation (crystal coordinates)
!>           F0(1:48,NAT)
!>                The atom transformation table for rotation (48,NAT)
!>           ORIGIN Standard origin if symmorphic (crystal coordinates)
!>           ISY  = 1 Isommorphic group
!>                =-1 Isommorphic group with non-standard origin
!>                = 0 Non-Isommorphic group
!>                =-2 Undetermined (normally never)
!>         LI ..... Code indicating whether the point group
!>                  of the crystal contains inversion or not
!>                  (operations 13 or 25 in respectively hexagonal
!>                  or cubic groups).
!>                  LI=0    : does not contain inversion
!>                  LI.GT.0 : there is inversion in the point
!>                            group of the crystal
! **************************************************************************************************
  SUBROUTINE atftm1(r,v,x,f0,xb,origin,ib,ty,nat,ihg,rx,nc,indpg,ntvec, &
       a,ai,li,isy,isc,delta)
      REAL(KIND=dp)                                      :: r(3,3,48), v(3,48), xb(3), origin(3)
      INTEGER                                            :: ib(48), nat, ty(nat), f0(49,nat)
      REAL(KIND=dp)                                      :: x(3,nat)
      INTEGER                                            :: ihg
      REAL(KIND=dp)                                      :: rx(3,nat)
      INTEGER                                            :: nc, indpg, ntvec
      REAL(KIND=dp)                                      :: a(3,3), ai(3,3)
      INTEGER                                            :: li, isy, isc(nat)
      REAL(KIND=dp)                                      :: delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'atftm1', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iis(48), il, info, j, k, k2, l, n, &
                                                            nca, ni
      LOGICAL                                            :: nodupli, oksym
      REAL(KIND=dp)                                      :: vc(3,48), vr(3), vs

    nodupli = ntvec == 1
    nca = 0
    DO n = 1, 48
       iis(n) = 0
    END DO
    !     Calculate translational vector for each operation
    !     and atom transformation table.
    DO n = 1, nc
       l = ib(n)
       iis(l) = 1
       DO k = 1, nat
          DO i = 1, 3
             rx(i,k) = 0.0_dp
             DO j = 1, 3
                rx(i,k) = rx(i,k) + r(i,j,l)*x(j,k)
             END DO
          END DO
       END DO
       DO k = 1, 3
          vr(k) = 0.0_dp
       END DO
       !       First we determine for VR=(/0,0,0/)
       !       IMPORTANT IF NOT UNIQUE ATOMS FOR DETERMINATION OF SYMMORPHIC
       CALL checkrlv3(n,nat,ty,rx,x,vr,f0,ai,isc,nodupli,oksym,delta)
       IF (oksym) THEN
         nca = nca + 1
         DO i = 1, 3
           v(i,nca) = vr(i)
         END DO
         CYCLE
       END IF
       !       Now we try other possible VR
       !       F0(49,1:NAT) has only inequivalent atom indexes for translation
       DO k2 = 1, nat
          IF (f0(49,k2)<k2.OR.ty(1)/=ty(k2)) CYCLE
          DO i = 1, 3
             xb(i) = rx(i,1) - x(i,k2)
          END DO
          !         A translation vector VR is defined.
          CALL rlv3(ai,xb,vr,il,delta)
          !------------------------------------------------------------------------------!
          !            SUBROUTINE RLV3 REMOVES A DIRECT LATTICE VECTOR FROM XB
          !            LEAVING THE REMAINDER IN VR. IF A NONZERO LATTICE
          !            VECTOR WAS REMOVED, IL IS MADE NONZERO.
          !            VR STANDS FOR V-REFERENCE.
          !            VR IS NOT GIVEN IN CARTESIAN COORDINATES BUT
          !            IN THE SYSTEM A1,A2,A3.     K.K., 23.10.1979
          !------------------------------------------------------------------------------!
          CALL checkrlv3(n,nat,ty,rx,x,vr,f0,ai,isc,nodupli,oksym,delta)
          IF (oksym) THEN
            nca = nca + 1
            DO i = 1, 3
              v(i,nca) = vr(i)
            END DO
            EXIT
          END IF
       END DO
       IF (oksym) CYCLE
       iis(l) = 0
       !------------------------------------------------------------------------------!
       !          V(I,N) IS THE I-TH COMPONENT OF THE FRACTIONAL
       !          TRANSLATION ASSOCIATED WITH THE ROTATION N.
       !          ATTENTION: V(I) ARE NOT CARTESIAN COMPONENTS, THEY ARE
       !          GIVEN IN THE SYSTEM A1,A2,A3.
       !          K.K., 23. 10. 1979
       !------------------------------------------------------------------------------!
    END DO
    !     Remove unused operations
    i = 0
    ni = 13
    IF (ihg<6) ni = 25
    li = 0
    DO n = 1, nc
       l = ib(n)
       IF (iis(l)==0) CYCLE
       i = i + 1
       ib(i) = ib(n)
       IF (ib(i)==ni) li = i
       DO k = 1, nat
          f0(i,k) = f0(n,k)
       END DO
    END DO
    !------------------------------------------------------------------------------!
    nc = i
    vs = 0.0_dp
    DO n = 1, nc
       DO i = 1, 3
          vs = vs + ABS(v(i,n))
       END DO
    END DO
    !     THE ORIGINAL VALUE DELTA=0.0001 WAS MODIFIED
    !     BY K.K. , SEPTEMBER 1979 TO 0.0005
    !     AND RETURNED TO 0.0001 BY RJN OCT 1987
    IF (vs>delta) THEN
       isy = 0
    ELSE
       isy = 1
    END IF
    !------------------------------------------------------------------------------!
    !     Determination of the point group
    !     (Thierry Deutsch - 1998 [Maybe not complete!!])
    IF (ihg<6) THEN
       IF (nc==0) THEN
          WRITE (*,'(" ATFTM1! IHG=",A," NC=",I2)') icst(ihg), nc
          CALL cp__b("k290.F",1419,"")
          !       Triclinic system
       ELSE IF (nc==1) THEN
          !         IB=1
          indpg = & !1 (c1)
               1
       ELSE IF (nc==2 .AND. ib(2)==25) THEN
          !         IB=1 25
          indpg = & !<1>(ci)
               2
       ELSE IF (nc==2 .AND. (ib(2)==4 .OR. & !         Monoclinic system
            ib(2)==2 .OR. & !         IB=1 4 (z-axis) OR
            ib(2)==3)) & !         IB=1 2 (x-axis) OR
            THEN
          !         IB=1 3 (y-axis)
          !2[001]
          !2[100]
          !2[010]
          indpg = & !2 (c2)
               3
       ELSE IF (nc==2 .AND. (ib(2)==28 .OR. ib(2)==26 .OR. ib(2)==27)) THEN
          !         IB=1 28 (z-axis) OR
          !         IB=1 26 (x-axis) OR
          !         IB=1 27 (y-axis)
          indpg = & !m (c1h)
               4
       ELSE IF (nc==4 .AND. (ib(4)==28 .OR. & !         IB=1  4 25 28 (z-axis)  OR
            ib(4)==27 .OR. & !         IB=1  2 25 26 (x-axis)  OR
            ib(4)==26 .OR. & !         IB=1  3 25 27 (y-axis)  OR
            ib(4)==37 .OR. & !         IB=1 13 25 37 (-xy-axis)OR
            ib(4)==40)) & !         IB=1 16 25 40 (xy-axis)
            THEN
          !2[001]
          !2[010]
          !2[100]
          !-2[-110]
          !2[110]
          indpg = & !2/m(c2h)
               5
       ELSE IF (nc==4 .AND. (ib(4)==15 .OR. ib(4)==20 .OR. ib(4)==24)) THEN
          !         Tetragonal system
          !         IB=1 4 14 15 (z-axis) OR
          !         IB=1 2 19 20 (x-axis) OR
          !         IB=1 3 22 24 (y-axis)
          indpg = & !4 (c4)
               11
       ELSE IF (nc==4 .AND. (ib(4)==39 .OR. ib(4)==44 .OR. ib(4)==48)) THEN
          !         IB=1 4 38 39 (z-axis) OR
          !         IB=1 2 43 44 (x-axis) OR
          !         IB=1 3 46 48 (y-axis)
          indpg = & !<4>(s4)
               12
       ELSE IF (nc==8 .AND. ((ib(3)==14 .AND. ib(8)==39) .OR. (ib(3)==19 &
            .AND. ib(8)==44) .OR. (ib(3)==22 .AND. ib(8)==48))) THEN
          !         IB=1 4 14 15 28 25 38 39 (z-axis) OR
          !         IB=1 2 19 20 26 25 43 44 (x-axis) OR
          !         IB=1 3 22 24 27 25 46 48 (y-axis)
          indpg = & !422(d4)
               13
       ELSE IF (nc==8 .AND. ib(4)==4 .AND. (ib(8)==16 .OR. ib( &
            8)==20 .OR. ib(8)==24)) THEN
          !         IB=1 2 3 4 13 14 15 16 (z-axis) OR
          !         IB=1 2 3 4 17 19 20 18 (x-axis) OR
          !         IB=1 2 3 4 21 22 24 23 (y-axis)
          indpg = & !4/m(c4h)
               14
       ELSE IF (nc==8 .AND. (ib(8)==40 .OR. ib(8)==42 .OR. ib(8)==47)) THEN
          !         IB=1 4 14 15 26 27 37 40 (z-axis) OR
          !         IB=1 2 19 20 28 27 41 42 (x-axis) OR
          !         IB=1 3 22 24 26 28 45 47 (y-axis)
          indpg = & !4mm(c4v)
               15
       ELSE IF (nc==8 .AND. ((ib(3)==13 .AND. ib(8)==39) .OR. (ib(3)==17 &
            .AND. ib(8)==44) .OR. (ib(3)==21 .AND. ib(8)==48))) THEN
          !         IB=1 4 13 16 26 27 38 39 (z-axis) OR
          !         IB=1 2 17 18 28 27 43 44 (x-axis) OR
          !         IB=1 3 21 23 26 28 46 48 (y-axis)
          indpg = & !<4>2m(d2d)
               16
       ELSE IF (nc==16 .AND. (ib(16)==40 .OR. ib(16)==44 .OR. ib(16)==48)) &
            THEN
          !         IB=1 2 3 4 13 14 15 16 25 26 27 28 37 38 39 40 (z-axis) OR
          !         IB=1 2 3 4 17 19 20 18 25 26 27 28 41 43 44 42 (x-axis) OR
          !         IB=1 2 3 4 21 22 24 23 25 26 27 28 45 46 48 47 (y-axis)
          indpg = & !4/mmm(d4h)
               17
       ELSE IF (nc==4 .AND. (ib(4)==4)) THEN
          !         Orthorhombic system
          !         IB=1 2  3  4
          indpg = & !222(d2)
               25
       ELSE IF (nc==4 .AND. (ib(4)==27 .OR. ib(4)==28)) THEN
          !         IB=1 3 26 27 (z-axis) OR
          !         IB=1 2 27 28 (x-axis) OR
          !         IB=1 4 26 28 (y-axis) OR
          indpg = & !mm2(c2v)
               26
       ELSE IF (nc==8) THEN
          !         IB=1 2 3 4 25 26 27 28
          indpg = & !mmm(d2h)
               27
       ELSE IF (nc==12 .AND. (ib(12)==12 .OR. ib(12)==47 .OR. ib(12)==45)) &
            THEN
          !         Cubic system
          !         IB=1 2  3  4  5  6  7  8  9 10 11 12 OR
          !         IB=1 5 11 13 18 23 25 30 35 37 42 47 OR
          !         IB=1 8 10 16 18 21 25 32 34 40 42 45
          indpg = & !23 (t)
               28
       ELSE IF (nc==24 .AND. ib(24)==36) THEN
          !         IB= 1  2  3  4  5  6  7  8  9 10 11 12
          !            25 26 27 28 29 30 31 32 33 34 35 36
          indpg = & !m3 (th)
               29
       ELSE IF (nc==24 .AND. ib(24)==24) THEN
          !         IB=1 2 3 4 5 6 7 8 9 10 11 12
          !           13 14 15 16 17 18 19 20 21 22 23 24
          indpg = & !432 (o)
               30
       ELSE IF (nc==24 .AND. ib(24)==48) THEN
          !         IB=1 2 3 4 5 6 7 8 9 10 11 12
          !           37 38 39 40 41 42 43 45 46 47 48
          indpg = & !<4>3m(td)
               31
       ELSE IF (nc==48) THEN
          !         IB=1..48
          indpg = & !m3m(oh)
               32
       ELSE
          !          WRITE(*,'(" ATFTM1! IHG=",A," NC=",I2)') ICST(IHG),NC
          !          WRITE(*,'(" ATFTM1!",19I3)') (IB(I),I=1,NC)
          !          WRITE(*,'(" ATFTM1! THIS CASE IS UNKNOWN IN THE DATABASE")')
          !         Probably a sub-group of 32
          indpg = -32
       END IF
    ELSE IF (ihg>=6) THEN
       IF (nc==0) THEN
          WRITE (*,'(" ATFTM1! IHG=",A," NC=",I2)') icst(ihg), nc
          CALL cp__b("k290.F",1557,"")
          !       Triclinic system
       ELSE IF (nc==1) THEN
          !         IB=1
          indpg = & !1 (c1)
               1
       ELSE IF (nc==2 .AND. ib(2)==13) THEN
          !         IB=1 13
          indpg = & !<1>(ci)
               2
       ELSE IF (nc==2 .AND. (ib(2)==4)) & !         Monoclinic system
            THEN
          !         IB=1 4
          !2[001]
          indpg = & !2 (c2)
               3
       ELSE IF (nc==2 .AND. (ib(2)==16)) THEN
          !         IB=1 16
          indpg = & !m (c1h)
               4
       ELSE IF (nc==4 .AND. (ib(4)==24 .OR. ib(4)==20)) THEN
          !         IB=1 12 13 24 OR
          !         IB=1  8 13 20
          indpg = & !2/m(c2h)
               5
       ELSE IF (nc==3 .AND. ib(3)==5) THEN
          !         Trigonal system
          !         IB=1 3 5
          indpg = & !3 (c3)
               6
       ELSE IF (nc==6 .AND. ib(6)==17) THEN
          !         IB=1 13 15 17 35
          indpg = & !<3>(c3i)
               7
       ELSE IF (nc==6 .AND. ib(6)==11) THEN
          !         IB=1 7 9 11 35
          indpg = & !32 (d3)
               8
       ELSE IF (nc==6 .AND. ib(6)==23) THEN
          !         IB=1 3 5 19 21 23
          indpg = & !3m (c3v)
               9
       ELSE IF (nc==12 .AND. ib(12)==23) THEN
          !         IB=1 3 5 7 9 11 13 15 17 19 21 23
          indpg = & !<3>m(d3d)
               10
       ELSE IF (nc==6 .AND. ib(6)==6) THEN
          !         Hexagonal system
          !         IB=1 2 3 4 5 6
          indpg = & !6 (c6)
               18
       ELSE IF (nc==6 .AND. ib(6)==18) THEN
          !         IB=1 3 5 14 16 18
          indpg = & !<6>(c3h)
               19
       ELSE IF (nc==12 .AND. ib(12)==18) THEN
          !         IB=1 2 3 4 5 6 13 14 15 16 17 18
          indpg = & !6/m(c6h)
               20
       ELSE IF (nc==12 .AND. ib(12)==12) THEN
          !         IB=1 2 3 4 5 6 7 8 9 10 11 12
          indpg = & !622(d6)
               21
       ELSE IF (nc==12 .AND. ib(2)==2 .AND. ib(12)==24) THEN
          !         IB=1 2 3 4 5 6 19 20 21 22 23 24
          indpg = & !6mm(c6v)
               22
       ELSE IF (nc==12 .AND. ib(2)==3 .AND. ib(12)==24) THEN
          !         IB=1 3 5 7 9 11 14 16 18 20 22 24
          indpg = & !<6>m2(d3h)
               23
       ELSE IF (nc==24) THEN
          !         IB=1..24
          indpg = & !6/mmm(d6h)
               24
       ELSE
          !         Probably a sub-group of 24
          !          WRITE(*,'(" ATFTM1! IHG=",A," NC=",I2)') ICST(IHG),NC
          !          WRITE(*,'(" ATFTM1!",48I3)') (IB(I),I=1,NC)
          !          WRITE(*,'(" ATFTM1! THIS CASE IS UNKNOWN IN THE DATABASE")')
          indpg = -24
       END IF
    END IF
    !------------------------------------------------------------------------------!
    !        Determination if the space group is symmorphic or not
    !------------------------------------------------------------------------------!
    IF (isy/=1) THEN
       !       Transform V in cartesian coordinates
       DO n = 1, nc
          DO i = 1, 3
             vc(i,n) = a(i,1)*v(1,n) + a(i,2)*v(2,n) + a(i,3)*v(3,n)
          END DO
       END DO
       CALL symmorphic(nc,ib,r,vc,ai,info,origin,delta)
       IF (info==1) THEN
          CALL rlv3(ai,origin,xb,il,delta)
          !         !!!RLV3 determines -XB in crystal coordinates
          !         !!We want between 0.0 and 1.0
          DO i = 1, 3
             IF (-xb(i)>=0.0_dp) THEN
                origin(i) = -xb(i)
             ELSE
                origin(i) = 1.0_dp - xb(i)
             END IF
          END DO
          DO i = 1, 3
             xb(i) = a(i,1)*origin(1) + a(i,2)*origin(2) + a(i,3)*origin(3)
          END DO
          isy = -1
       ELSE IF (info==0) THEN
          isy = 0
       ELSE
          isy = -2
       END IF
    ELSE
       DO i = 1, 3
          origin(i) = 0.0_dp
       END DO
    END IF
  END SUBROUTINE atftm1
! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param nat ...
!> \param ty ...
!> \param rx ...
!> \param x ...
!> \param vr ...
!> \param f0 ...
!> \param ai ...
!> \param isc ...
!> \param nodupli ...
!> \param oksym ...
!> \param delta ...
!> \par Original Header
!> ------------------------------------------------------------------------------!
!>         WRITTEN IN MAY 14TH, 1998 (T.D.)
!>         CHECK IF RX+VR GIVES THE SAME LATTICE AS X
!>         BUILD THE ATOM TRANSFORMATION TABLE
!> ------------------------------------------------------------------------------!
!>         INPUT:
!>           N   ROTATION NUMBER (INDEX USED IN F0 BETWEEN 1 AND 48)
!>           NAT           NUMBER OF ATOMS
!>           TY(1:NAT)     TYPE OF ATOMS
!>           RX(1:3,1:NAT) ATOMIC COORDINATES FROM Nth ROTATION (CART.)
!>           X(1:3,1:NAT)  ATOMIC COORDINATES (CARTESIAN)
!>           VR(1:3)       TRANSLATION VECTOR (CRYSTAL COOR.)
!>           AI(1:3,1:3)   LATTICE RECIPROCAL VECTORS
!>           NODUPLI       .TRUE., THE CELL IS A PRIMITIVE ONE
!>                         WE CAN SPEED UP
!>         DELTA           REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
!>         OUTPUT:
!>           F0(1:49,1:NAT) ATOM TRANSFORMATION TABLE
!>              F0 IS THE FUNCTION DEFINED IN MARADUDIN AND VOSK0
!>              BY EQ.(2.35).
!>              IT DEFINES THE ATOM TRANSFORMATION TABLE
!>           OKSYM          TRUE IF RX+VR = X
!>           ISC(1:NAT)     SCRATCH ARRAY
!>                          USED TO SPEED UP THE ROUTINE
!>                          EACH ATOM IS ONLY ONCE AN IMAGE
!>                          IF NO DUPLICATION OF THE CELL
! **************************************************************************************************
  SUBROUTINE checkrlv3(n,nat,ty,rx,x,vr,f0,ai,isc,nodupli,oksym,delta)

      INTEGER                                            :: n, nat, ty(nat)
      REAL(KIND=dp)                                      :: rx(3,nat), x(3,nat), vr(3)
      INTEGER                                            :: f0(49,nat)
      REAL(KIND=dp)                                      :: ai(3,3)
      INTEGER                                            :: isc(nat)
      LOGICAL                                            :: nodupli, oksym
      REAL(KIND=dp)                                      :: delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'checkrlv3', routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, ib, il
      REAL(KIND=dp)                                      :: vt(3), xb(3)

    DO ia = 1, nat
       isc(ia) = 0
    END DO
    !     Now we check if ROT(N)+VR gives a correct symmetry.
    ialoop: DO ia = 1, nat
       DO ib = 1, nat
          IF (ty(ia)==ty(ib) .AND. isc(ib)==0) THEN
             xb(1) = rx(1,ia) - x(1,ib)
             xb(2) = rx(2,ia) - x(2,ib)
             xb(3) = rx(3,ia) - x(3,ib)
             CALL rlv3(ai,xb,vt,il,delta)
             !           VT STANDS FOR V-TEST
             oksym = (ABS((vr(1)-vt(1))-NINT(vr(1)-vt(1)))<delta) .AND. &
                  (ABS((vr(2)-vt(2))-NINT(vr(2)-vt(2)))<delta) .AND. &
                  (ABS((vr(3)-vt(3))-NINT(vr(3)-vt(3)))<delta)
             IF (oksym) THEN
                IF (nodupli) isc(ib) = 1
                f0(n,ia) = ib
                !             IR+VR is the good one: another symmetry operation
                !             Next atom
                CYCLE ialoop
             END IF
          END IF
       END DO
       !       VR is not the correct translation vector
       RETURN
    END DO ialoop
  END SUBROUTINE checkrlv3
! **************************************************************************************************
!> \brief ...
!> \param nc ...
!> \param ib ...
!> \param r ...
!> \param v ...
!> \param ai ...
!> \param info ...
!> \param origin ...
!> \param delta ...
!> \par Original Header
!> ------------------------------------------------------------------------------!
!>         Check if the group is symmorphic with a non-standard origin
!>         WARNING: If there are equivalent atoms, this routine could
!>         not determine if the space group is symmorphic
!>         So you have to check if the solution V=0 works (see ATFTM1)
!> ------------------------------------------------------------------------------!
!>         INPUT:
!>           NC Number of operations
!>           IB(NC) Index of operation in R
!>           R(3,3,48) Rotations
!>           V(3,NC) Fractional translations related to R(3,3,IB(NC))
!>                   R AND V ARE IN CARTESIAN COORDINATES
!>           AI(I,J) ARE THE RECIPROCAL LATTICE VECTORS,
!>                   B(I) = AI(I,J),J=1,2,3
!>         DELTA     REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
!>
!>         OUTPUT:
!>           ORIGIN(1:3) Give standard origin (cartesian coordinates)
!>                    Give the standard origin with smallest coordinates
!>                    if NTVEC /= 1
!>           INFO = 1 The group is symmorphic
!>           INFO = 0 The group is not symmorphic
!>           INFO =-1 The routine cannot determine
! **************************************************************************************************
  SUBROUTINE symmorphic(nc,ib,r,v,ai,info,origin,delta)
      INTEGER                                            :: nc, ib(nc)
      REAL(KIND=dp)                                      :: r(3,3,48), v(3,nc), ai(3,3)
      INTEGER                                            :: info
      REAL(KIND=dp)                                      :: origin(3), delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'symmorphic', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, i1, ierror, igood(3), il, imissing2, &
                                                            imissing3, iok(3), ionly, ir, j, j1
      REAL(KIND=dp)                                      :: diag, dif, r2(2,2), r3(3,3), vr(3), xb(3)

    DO i = 1, 3
       iok(i) = 0
    END DO
    DO i = 1, 3
       origin(i) = 0.0_dp
    END DO
    !     Find a point A / V_R = (1-R).OA
    DO ir = 1, nc
       dif = v(1,ir)*v(1,ir) + v(2,ir)*v(2,ir) + v(3,ir)*v(3,ir)
       IF (dif>delta*delta) THEN
          DO i = 1, 3
             igood(i) = 1
          END DO
          !         V is non-zero. Construct matrix 1-R
          DO i = 1, 3
             DO j = 1, 3
                r3(i,j) = -r(i,j,ib(ir))
             END DO
             r3(i,i) = 1 + r3(i,i)
          END DO
          CALL invmat(r3(1:3,1:3),ierror)
          IF (ierror==0) THEN
             !           The matrix 3x3 has an inverse.
             DO i = 1, 3
                vr(i) = r3(i,1)*v(1,ir) + r3(i,2)*v(2,ir) + r3(i,3)*v(3,ir)
             END DO
          ELSE
             !           IERROR gives the column which causes some trouble
             !           Construct matrix 1-R with 2x2
             igood(ierror) = 0
             imissing3 = ierror
             i1 = 0
             DO i = 1, 3
                IF (i/=ierror) THEN
                   i1 = i1 + 1
                   j1 = 0
                   DO j = 1, 3
                      IF (j/=ierror) THEN
                         j1 = j1 + 1
                         r2(i1,j1) = -r(i,j,ib(ir))
                      END IF
                   END DO
                   r2(i1,i1) = 1 + r2(i1,i1)
                END IF
             END DO
             CALL invmat(r2(1:2,1:2),ierror)
             IF (ierror==0) THEN
                !             The matrix 2X2 has an inverse.
                !             Solve Vxy = (1-R).OAxy + OAz R3z (z is IMISSING3)
                i1 = 0
                DO i = 1, 3
                   IF (igood(i)==1) THEN
                      i1 = i1 + 1
                      vr(i) = 0.0_dp
                      j1 = 0
                      DO j = 1, 3
                         IF (igood(j)==1) THEN
                            j1 = j1 + 1
                            vr(i) = vr(i) + r2(i1,j1)*(v(j,ir)+origin(imissing3)*r &
                                 (j,imissing3,ib(ir)))
                         END IF
                      END DO
                   ELSE
                      vr(i) = origin(i)
                   END IF
                END DO
             ELSE
                !             Construct matrix 1-R with 1x1
                i1 = 0
                DO i = 1, 3
                   IF (i/=imissing3) THEN
                      i1 = i1 + 1
                      IF (i1==ierror) THEN
                         igood(i) = 0
                         imissing2 = i
                      ELSE
                         ionly = i
                      END IF
                   END IF
                END DO
                diag = (1-r(ionly,ionly,ib(ir)))
                IF (ABS(diag)>delta) THEN
                   vr(ionly) = 1.0_dp/diag*(v(ionly,ir)+origin(imissing3)*r( &
                        ionly,imissing3,ib(ir))+origin(imissing2)*r(ionly, &
                        imissing2,ib(ir)))
                ELSE
                   vr(ionly) = origin(ionly)
                   igood(ionly) = 0
                END IF
                vr(imissing3) = origin(imissing3)
                vr(imissing2) = origin(imissing2)
             END IF
          END IF
          !------------------------------------------------------------------------------!
          !         Compare VR with ORIGIN
          dif = 0.0_dp
          !         If NTVEC /=1 there are NTVEC possible standard origins
          DO i = 1, 3
             IF (iok(i)==1) THEN
                dif = dif + ABS(origin(i)-vr(i))
             END IF
          END DO
          IF (dif>delta) THEN
             !           Non-symmorphic
             info = 0
             RETURN
          ELSE
             DO i = 1, 3
                IF (iok(i)/=1 .AND. igood(i)==1) THEN
                   iok(i) = 1
                   origin(i) = vr(i)
                END IF
             END DO
          END IF
       END IF
    END DO
    !------------------------------------------------------------------------------!
    IF (iok(1)==0 .AND. iok(2)==0 .AND. iok(3)==0) THEN
       !       Cannot not determine
       info = -1
       RETURN
    END IF
    !     The group is symmorphic
    info = 1
    !     Check
    DO ir = 1, nc
       DO i = 1, 3
          vr(i) = r(i,1,ib(ir))*origin(1) + r(i,2,ib(ir))*origin(2) + &
               r(i,3,ib(ir))*origin(3)
          vr(i) = (origin(i)-vr(i)) - v(i,ir)
       END DO
       CALL rlv3(ai,vr,xb,il,delta)
       dif = ABS(xb(1)) + ABS(xb(2)) + ABS(xb(3))
       IF (dif>delta) THEN
          !         Non-symmorphic
          info = 0
          RETURN
       END IF
    END DO
  END SUBROUTINE symmorphic

! **************************************************************************************************
!> \brief ...
!> \param ihc ...
!> \param r ...
!> \par Original Header
!> ------------------------------------------------------------------------------!
!>         WRITTEN ON FEBRUARY 17TH, 1976
!>         GENERATION OF THE X,Y,Z-TRANSFORMATION MATRICES 3X3
!>         FOR HEXAGONAL AND CUBIC GROUPS
!>         SUBROUTINES NEEDED -- NONE
!> ------------------------------------------------------------------------------!
!>         THIS IS IDENTICAL WITH THE SUBROUTINE ROT OF WORLTON-WARREN
!>         (IN THE AC-COMPLEX), ONLY THE WAY OF TRANSFERRING THE DATA
!>         WAS CHANGED
!> ------------------------------------------------------------------------------!
!>         INPUT DATA:
!>         IHC SWITCH DETERMINING IF WE DESIRE
!>             THE HEXAGONAL GROUP(IHC=0) OR THE CUBIC GROUP (IHC=1)
!>         OUTPUT DATA:
!>         R...THE 3X3 MATRICES OF THE DESIRED COORDINATE REPRESENTATION
!>             THEIR NUMBERING CORRESPONDS TO THE SYMMETRY ELEMENTS AS
!>             LISTE IN WORLTON-WARREN
!>                      (COMPUT. PHYS. COMM. 3(1972) 88--117)
!>         FOR IHC=0 THE FIRST 24 MATRICES OF THE ARRAY R REPRESENT
!>                   THE FULL HEXAGONAL GROUP D(6H)
!>         FOR IHC=1 THE FIRST 48 MATRICES OF THE ARRAY R REPRESENT
!>                   THE FULL CUBIC GROUP O(H)
! **************************************************************************************************
  SUBROUTINE rot1(ihc,r)
      INTEGER                                            :: ihc
      REAL(KIND=dp)                                      :: r(3,3,48)

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rot1', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k, n, nv
      REAL(KIND=dp)                                      :: c, s

!     Variables
!------------------------------------------------------------------------------!

    DO j = 1, 3
       DO i = 1, 3
          DO n = 1, 48
             r(i,j,n) = 0.0_dp
          END DO
       END DO
    END DO
    IF (ihc==0) THEN
       !------------------------------------------------------------------------------!
       !       DEFINE THE GENERATORS FOR THE ROTATION MATRICES--HEXAGONAL GROUP
       !------------------------------------------------------------------------------!
       c = 0.5E0_dp
       s = 0.5E0_dp*SQRT(3.0E0_dp)
       r(1,1,2) = c
       r(1,2,2) = -s
       r(2,1,2) = s
       r(2,2,2) = c
       r(1,1,7) = -c
       r(1,2,7) = -s
       r(2,1,7) = -s
       r(2,2,7) = c
       DO n = 1, 6
          r(3,3,n) = 1.0_dp
          r(3,3,n+18) = 1.0_dp
          r(3,3,n+6) = -1.0_dp
          r(3,3,n+12) = -1.0_dp
       END DO
       !------------------------------------------------------------------------------!
       !          GENERATE THE REST OF THE ROTATION MATRICES
       !------------------------------------------------------------------------------!
       DO i = 1, 2
          r(i,i,1) = 1.0_dp
          DO j = 1, 2
             r(i,j,6) = r(j,i,2)
             DO k = 1, 2
                r(i,j,3) = r(i,j,3) + r(i,k,2)*r(k,j,2)
                r(i,j,8) = r(i,j,8) + r(i,k,2)*r(k,j,7)
                r(i,j,12) = r(i,j,12) + r(i,k,7)*r(k,j,2)
             END DO
          END DO
       END DO
       DO i = 1, 2
          DO j = 1, 2
             r(i,j,5) = r(j,i,3)
             DO k = 1, 2
                r(i,j,4) = r(i,j,4) + r(i,k,2)*r(k,j,3)
                r(i,j,9) = r(i,j,9) + r(i,k,2)*r(k,j,8)
                r(i,j,10) = r(i,j,10) + r(i,k,12)*r(k,j,3)
                r(i,j,11) = r(i,j,11) + r(i,k,12)*r(k,j,2)
             END DO
          END DO
       END DO
       DO n = 1, 12
          nv = n + 12
          DO i = 1, 2
             DO j = 1, 2
                r(i,j,nv) = -r(i,j,n)
             END DO
          END DO
       END DO
    ELSE
       !------------------------------------------------------------------------------!
       !          DEFINE THE GENERATORS FOR THE ROTATION MATRICES-CUBIC GROUP
       !------------------------------------------------------------------------------!
       r(1,3,9) = 1.0_dp
       r(2,1,9) = 1.0_dp
       r(3,2,9) = 1.0_dp
       r(1,1,19) = 1.0_dp
       r(2,3,19) = -1.0_dp
       r(3,2,19) = 1.0_dp
       DO i = 1, 3
          r(i,i,1) = 1.0_dp
          DO j = 1, 3
             r(i,j,20) = r(j,i,19)
             r(i,j,5) = r(j,i,9)
             DO k = 1, 3
                r(i,j,2) = r(i,j,2) + r(i,k,19)*r(k,j,19)
                r(i,j,16) = r(i,j,16) + r(i,k,9)*r(k,j,19)
                r(i,j,23) = r(i,j,23) + r(i,k,19)*r(k,j,9)
             END DO
          END DO
       END DO
       DO i = 1, 3
          DO j = 1, 3
             DO k = 1, 3
                r(i,j,6) = r(i,j,6) + r(i,k,2)*r(k,j,5)
                r(i,j,7) = r(i,j,7) + r(i,k,16)*r(k,j,23)
                r(i,j,8) = r(i,j,8) + r(i,k,5)*r(k,j,2)
                r(i,j,10) = r(i,j,10) + r(i,k,2)*r(k,j,9)
                r(i,j,11) = r(i,j,11) + r(i,k,9)*r(k,j,2)
                r(i,j,12) = r(i,j,12) + r(i,k,23)*r(k,j,16)
                r(i,j,14) = r(i,j,14) + r(i,k,16)*r(k,j,2)
                r(i,j,15) = r(i,j,15) + r(i,k,2)*r(k,j,16)
                r(i,j,22) = r(i,j,22) + r(i,k,23)*r(k,j,2)
                r(i,j,24) = r(i,j,24) + r(i,k,2)*r(k,j,23)
             END DO
          END DO
       END DO
       DO i = 1, 3
          DO j = 1, 3
             DO k = 1, 3
                r(i,j,3) = r(i,j,3) + r(i,k,5)*r(k,j,12)
                r(i,j,4) = r(i,j,4) + r(i,k,5)*r(k,j,10)
                r(i,j,13) = r(i,j,13) + r(i,k,23)*r(k,j,11)
                r(i,j,17) = r(i,j,17) + r(i,k,16)*r(k,j,12)
                r(i,j,18) = r(i,j,18) + r(i,k,16)*r(k,j,10)
                r(i,j,21) = r(i,j,21) + r(i,k,12)*r(k,j,15)
             END DO
          END DO
       END DO
       DO n = 1, 24
          nv = n + 24
          DO i = 1, 3
             DO j = 1, 3
                r(i,j,nv) = -r(i,j,n)
             END DO
          END DO
       END DO
    END IF
  END SUBROUTINE rot1
! **************************************************************************************************
!> \brief ...
!> \param iout ...
!> \param iq1 ...
!> \param iq2 ...
!> \param iq3 ...
!> \param wvk0 ...
!> \param nkpoint ...
!> \param a1 ...
!> \param a2 ...
!> \param a3 ...
!> \param b1 ...
!> \param b2 ...
!> \param b3 ...
!> \param inv ...
!> \param nc ...
!> \param ib ...
!> \param r ...
!> \param ntot ...
!> \param wvkl ...
!> \param lwght ...
!> \param lrot ...
!> \param ncbrav ...
!> \param ibrav ...
!> \param istriz ...
!> \param nhash ...
!> \param includ ...
!> \param list ...
!> \param rlist ...
!> \param delta ...
!> \par Original Header
!> ------------------------------------------------------------------------------!
!>         WRITTEN ON SEPTEMBER 12-20TH, 1979 BY K.K.
!>         MODIFIED 26-MAY-82 BY OLE HOLM NIELSEN
!>         GENERATION OF SPECIAL POINTS FOR AN ARBITRARY LATTICE,
!>         FOLLOWING THE METHOD MONKHORST,PACK,
!>         PHYS. REV. B13 (1976) 5188
!>         MODIFIED BY MACDONALD, PHYS. REV. B18 (1978) 5897
!>         THE SUBROUTINE IS WRITTEN ASSUMING THAT THE POINTS ARE
!>         GENERATED IN THE RECIPROCAL SPACE.
!>         IF, HOWEVER, THE B1,B2,B3 ARE REPLACED BY A1,A2,A3, THEN
!>         SPECIAL POINTS IN THE DIRECT SPACE CAN BE PRODUCED, AS WELL.
!>         (NO MULTIPLICATION BY 2PI IS THEN NECESSARY.)
!>         IN THE CASE OF NONSYMMORPHIC GROUPS, THE APPLICATION IN THE
!>         DIRECT SPACE WOULD PROBABLY REQUIRE A CERTAIN CAUTION.
!>         SUBROUTINES NEEDED: BZDEFI,BZRDUC,INBZ,MESH
!>         IN THE CASES WHERE THE POINT GROUP OF THE CRYSTAL DOES NOT
!>         CONTAIN INVERSION. THE LATTER MAY BE ADDED IF WE WISH
!>         (SEE COMMENT TO THE SWITCH INV).
!>         REDUCTION TO THE 1ST BRILLOUIN ZONE IS DONE
!>         BY ADDING G-VECTORS TO FIND THE SHORTEST WAVE-VECTOR.
!>         THE ROTATIONS OF THE BRAVAIS LATTICE ARE APPLIED TO THE
!>         MONKHORST/PACK MESH IN ORDER TO FIND ALL K-POINTS
!>         THAT ARE RELATED BY SYMMETRY. (OLE HOLM NIELSEN)
!> ------------------------------------------------------------------------------!
!>         INPUT DATA:
!>         IOUT:    LOGICAL UNIT FOR OUTPUT
!>                  IF (IOUT.LE.0) NO MESSAGE
!>         IQ1,IQ2,IQ3 .. PARAMETER Q OF MONKHORST AND PACK,
!>                  GENERALIZED AND DIFFERENT FOR THE 3 DIRECTIONS B1,
!>                  B2 AND B3
!>         WVK0 ... THE 'ARBITRARY' SHIFT OF THE WHOLE MESH, DENOTED K0
!>                  IN MACDONALD. WVK0 = 0 CORRESPONDS TO THE ORIGINAL
!>                  SCHEME OF MONKHORST AND PACK.
!>                  UNITS: 2PI/(UNITS OF LENGTH  USED IN A1, A2, A3),
!>                  I.E. THE SAME  UNITS AS THE GENERATED SPECIAL POINTS
!>         NKPOINT .. VARIABLE DIMENSION OF THE (OUTPUT) ARRAYS WVKL,
!>                  LWGHT,LROT, I.E. SPACE RESERVED FOR THE SPECIAL
!>                  POINTS AND ACCESSORIES.
!>                  NKPOINT HAS TO BE .GE. NTOT (TOTAL NUMBER OF SPECIAL
!>                  POINTS. THIS IS CHECKED BY THE SUBROUTINE.
!>         ISTRIZ . INDICATES WHETHER ADDITIONAL MESH POINTS SHOULD BE
!>                  GENERATED BY APPLYING GROUP OPERATIONS TO THE MESH.
!>                  ISTRIZ=+1 MEANS SYMMETRIZE
!>                  ISTRIZ=-1 MEANS DO NOT SYMMETRIZE
!>         THE FOLLOWING INPUT DATA MAY BE OBTAINED FROM THE SBRT.
!>         B1,B2,B3 .. RECIPROCAL LATTICE VECTORS, NOT MULTIPLIED BY
!>                  GROUP1: ANY 2PI (IN UNITS RECIPROCAL TO THOSE
!>                          OF A1,A2,A3)
!>         INV .... CODE INDICATING WHETHER WE WISH TO ADD THE INVERSION
!>                  TO THE POINT GROUP OF THE CRYSTAL OR NOT (IN THE
!>                  CASE THAT THE POINT GROUP DOES NOT CONTAIN ANY).
!>                  INV=0 MEANS: DO NOT ADD INVERSION
!>                  INV.NE.0 MEANS: ADD THE INVERSION
!>                  INV.NE.0 SHOULD BE THE STANDARD CHOICE WHEN SPPT2
!>                  IS USED IN RECIPROCAL SPACE - IN ORDER TO MAKE
!>                  USE OF THE HERMITICITY OF HAMILTONIAN.
!>                  WHEN USED IN DIRECT SPACE, THE RIGHT CHOICE OF INV
!>                  WILL DEPEND ON THE NATURE OF THE PHYSICAL PROBLEM.
!>                  IN THE CASES WHERE THE INVERSION IS ADDED BY THE
!>                  SWITCH INV, THE LIST IB WILL NOT BE MODIFIED BUT IN
!>                  THE OUTPUT LIST LROT SOME OF THE OPERATIONS WILL
!>                  APPEAR WITH NEGATIVE SIGN; THIS MEANS THAT THEY HAVE
!>                  TO BE APPLIED MULTIPLIED BY INVERSION.
!>         NC ..... TOTAL NUMBER OF ELEMENTS IN THE POINT GROUP OF THE
!>                  CRYSTAL
!>         IB ..... LIST OF THE ROTATIONS CONSTITUTING THE POINT GROUP
!>                  OF THE CRYSTAL. THE NUMBERING IS THAT DEFINED IN
!>                  WORLTON AND WARREN, I.E. THE ONE MATERIALIZED IN THE
!>                  ARRAY R (SEE BELOW)
!>                  ONLY THE FIRST NC ELEMENTS OF THE ARRAY IB ARE
!>                  MEANINGFUL
!>         R ...... LIST OF THE 3 X 3 ROTATION MATRICES
!>                  (XYZ REPRESENTATION OF THE O(H) OR D(6)H GROUPS)
!>                  ALL 48 OR 24 MATRICES ARE LISTED.
!>         NCBRAV . TOTAL NUMBER OF ELEMENTS IN RBRAV
!>         IBRAV .. LIST OF NCBRAV OPERATIONS OF THE BRAVAIS LATTICE
!>         DELTA    REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
!> ------------------------------------------------------------------------------!
!>         OUTPUT DATA:
!>         NTOT ... TOTAL NUMBER OF SPECIAL POINTS
!>                  IF NTOT APPEARS NEGATIVE, THIS IS AN ERROR SIGNAL
!>                  WHICH MEANS THAT THE DIMENSION NKPOINT WAS CHOSEN
!>                  TOO SMALL SO THAT THE ARRAYS WVKL ETC. CANNOT
!>                  ACCOMODATE ALL THE GENERATED SPECIAL POINTS.
!>                  IN THIS CASE THE ARRAYS WILL BE FILLED UP TO NKPOINT
!>                  AND FURTHER GENERATION OF NEW POINTS WILL BE
!>                  INTERRUPTED.
!>         WVKL ... LIST OF SPECIAL POINTS.
!>                  CARTESIAN COORDINATES AND NOT MULTIPLIED BY 2*PI.
!>                  ONLY THE FIRST NTOT VECTORS ARE MEANINGFUL
!>                  ALTHOUGH NO 2 POINTS FROM THE LIST ARE EQUIVALENT
!>                  BY SYMMETRY, THIS SUBROUTINE STILL HAS A KIND OF
!>                  'BEAUTY DEFECT': THE POINTS FINALLY
!>                  SELECTED ARE NOT NECESSARILY SITUATED IN A
!>                  'COMPACT' IRREDUCIBLE BRILL.ZONE; THEY MIGHT LIE IN
!>                  DIFFERENT IRREDUCIBLE PARTS OF THE B.Z. - BUT THEY
!>                  DO REPRESENT AN IRREDUCIBLE SET FOR INTEGRATION
!>                  OVER THE ENTIRE B.Z.
!>         LWGHT ... THE LIST OF WEIGHTS OF THE CORRESPONDING POINTS.
!>                  THESE WEIGHTS ARE NOT NORMALIZED (JUST INTEGERS)
!>         LROT ... FOR EACH SPECIAL POINT THE 'UNFOLDING ROTATIONS'
!>                  ARE LISTED. IF E.G. THE WEIGHT OF THE I-TH SPECIAL
!>                  POINT IS LWGHT(I), THEN THE ROTATIONS WITH NUMBERS
!>                  LROT(J,I), J=1,2,...,LWGHT(I) WILL 'SPREAD' THIS
!>                  SINGLE POINT FROM THE IRREDUCIBLE PART OF B.Z. INTO
!>                  SEVERAL POINTS IN AN ELEMENTARY UNIT CELL
!>                  (PARALLELOPIPED) OF THE RECIPROCAL SPACE.
!>                  SOME OPERATION NUMBERS IN THE LIST LROT MAY APPEAR
!>                  NEGATIVE, THIS MEANS THAT THE CORRESPONDING ROTATION
!>                  HAS TO BE APPLIED WITH INVERSION (THE LATTER HAVING
!>                  BEEN ARTIFICIALLY ADDED AS SYMMETRY OPERATION IN
!>                  CASE INV.NE.0).NO OTHER EFFORT WAS TAKEN,TO RENUMBER
!>                  THE ROTATIONS WITH MINUS SIGN OR TO EXTEND THE
!>                  LIST OF THE POINT-GROUP OPERATIONS IN THE LIST NB.
!>         INCLUD ... INTEGER ARRAY USED BY SPPT2 INCLUD(NKPOINT)
!>                  THE FIRST BIT (0) IS USED BY THE ROUTINE.
!>                  THE OTHER BITS GIVE THE K-POINT INDEX IN
!>                  THE SPECIAL K-POINT TABLE.
!> ------------------------------------------------------------------------------!
!>         NHASH    USED BY MESH ROUTINE
!>         LIST     INTEGER ARRAY USED BY MESH  LIST(NHASH+NKPOINT)
!>         RLIST    REAL(KIND=dp)  ARRAY USED BY MESH  RLIST(3,NKPOINT)
!> ------------------------------------------------------------------------------!
!>         Use bit manipulations functions
!>          IBSET(I,POS) sets the bit POS to 1 in I integer
!>          IBCLR(I,POS) clears the bit POS to 1 in I integer
!>          BTEST(I,POS) .TRUE. if bit POS is 1 in I integer
! **************************************************************************************************
  SUBROUTINE sppt2(iout,iq1,iq2,iq3,wvk0,nkpoint,a1,a2,a3,b1,b2,b3,inv,nc, &
       ib,r,ntot,wvkl,lwght,lrot,ncbrav,ibrav,istriz,nhash,includ,list, &
       rlist,delta)
      INTEGER                                            :: iout, iq1, iq2, iq3
      REAL(KIND=dp)                                      :: wvk0(3)
      INTEGER                                            :: nkpoint
      REAL(KIND=dp)                                      :: a1(3), a2(3), a3(3), b1(3), b2(3), b3(3)
      INTEGER                                            :: inv, nc, ib(48)
      REAL(KIND=dp)                                      :: r(3,3,48)
      INTEGER                                            :: ntot
      REAL(KIND=dp)                                      :: wvkl(3,nkpoint)
      INTEGER                                            :: lwght(nkpoint), lrot(48,nkpoint), &
                                                            ncbrav, ibrav(48), istriz, nhash, &
                                                            includ(nkpoint), list(nkpoint+nhash)
      REAL(KIND=dp)                                      :: rlist(3,nkpoint), delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'sppt2', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: no = 0, nrsdir = 100

      INTEGER                                            :: i, i1, i2, i3, ibsign, igarb0, igarbage, &
                                                            igarbg, imesh, iop, iplace, iremov, &
                                                            iwvk, j, jplace, k, n, nplane
      REAL(KIND=dp)                                      :: diff, proja(3), projb(3), &
                                                            rsdir(4,nrsdir), ur1, ur2, ur3, &
                                                            wva(3), wvk(3)

!------------------------------------------------------------------------------!

    ntot = 0
    DO i = 1, nkpoint
       lrot(1,i) = 1
       DO j = 2, 48
          lrot(j,i) = 0
       END DO
    END DO
    DO i = 1, nkpoint
       includ(i) = no
    END DO
    DO i = 1, 3
       wva(i) = 0.0_dp
    END DO
    !------------------------------------------------------------------------------!
    !        DEFINE THE 1ST BRILLOUIN ZONE
    !------------------------------------------------------------------------------!
    CALL bzdefi(iout,b1,b2,b3,rsdir,nrsdir,nplane,delta)
    !------------------------------------------------------------------------------!
    !        Generation of the mesh (they are not multiplied by 2*pi) by
    !        the Monkhorst/Pack algorithm, supplemented by all rotations
    !------------------------------------------------------------------------------!
    !     Initialize the list of vectors
    iplace = -2
    CALL mesh(iout,wva,iplace,igarb0,igarbg,nkpoint,nhash,list,rlist, &
         delta)
    imesh = 0
    DO i1 = 1, iq1
       DO i2 = 1, iq2
          DO i3 = 1, iq3
             ur1 = REAL(1+iq1-2*i1,KIND=dp)/REAL(2*iq1,KIND=dp)
             ur2 = REAL(1+iq2-2*i2,KIND=dp)/REAL(2*iq2,KIND=dp)
             ur3 = REAL(1+iq3-2*i3,KIND=dp)/REAL(2*iq3,KIND=dp)
             DO i = 1, 3
                wvk(i) = ur1*b1(i) + ur2*b2(i) + ur3*b3(i) + wvk0(i)
             END DO
             !           Reduce WVK to the 1st Brillouin zone
             CALL bzrduc(wvk,a1,a2,a3,b1,b2,b3,rsdir,nrsdir,nplane,delta)
             IF (istriz==1) THEN
                !             Symmetrization of the k-points mesh.
                !             Apply all the Bravais lattice operations to WVK
                DO iop = 1, ncbrav
                   DO i = 1, 3
                      wva(i) = 0.0_dp
                      DO j = 1, 3
                         wva(i) = wva(i) + r(i,j,ibrav(iop))*wvk(j)
                      END DO
                   END DO
                   !               Check that WVA is inside the 1 Bz.
                   IF (inbz(wva,rsdir,nrsdir,nplane,delta)==no) GO TO 450
                   !               Place WVA in list
                   iplace = 0
                   CALL mesh(iout,wva,iplace,igarb0,igarbg,nkpoint,nhash,list, &
                        rlist,delta)
                   !               If WVA was new (and therefore inserted),
                   !               IPLACE is the number.
                   IF (iplace>0) imesh = iplace
                   IF (iplace>nkpoint) GO TO 470
                END DO
             ELSE
                !             Place WVK in list
                iplace = 0
                CALL mesh(iout,wvk,iplace,igarb0,igarbg,nkpoint,nhash,list, &
                     rlist,delta)
                imesh = iplace
                IF (iplace>nkpoint) GO TO 470
             END IF
          END DO
       END DO
    END DO
    IF (iout>0) THEN
       !       IMESH: Number of k points in the mesh.
       WRITE (iout,'(" K290| The wavevector mesh contains ",i5," points")') &
            imesh
       WRITE (iout,'(" K290| The points are:")')
       DO i = 1, imesh
          CALL mesh(iout,wva,i,igarb0,igarbg,nkpoint,nhash,list,rlist,delta)
          IF (MOD(i,2)==1) THEN
             WRITE (iout,'(1X,I5,3F10.4)',ADVANCE="NO") i, wva
          ELSE
             WRITE (iout,'(1X,I5,3F10.4)') i, wva
          END IF
       END DO
       WRITE (iout,*)
    END IF
    !------------------------------------------------------------------------------!
    IF (istriz==1) THEN
       !       Now figure out if any special point difference (K - K'') is an
       !       integral multiple of a reciprocal-space vector
       iremov = 0
       DO i = 1, (imesh-1)
          iplace = i
          CALL mesh(iout,wva,iplace,igarb0,igarbg,nkpoint,nhash,list,rlist, &
               delta)
          !         Project WVA onto B1,2,3:
          proja(1) = 0.0_dp
          proja(2) = 0.0_dp
          proja(3) = 0.0_dp
          DO k = 1, 3
             proja(1) = proja(1) + wva(k)*a1(k)
             proja(2) = proja(2) + wva(k)*a2(k)
             proja(3) = proja(3) + wva(k)*a3(k)
          END DO
          !         Now loop over all the rest of the mesh points
          jloop: DO j = (i+1), imesh
             jplace = j
             CALL mesh(iout,wvk,jplace,igarb0,igarbg,nkpoint,nhash,list, &
                  rlist,delta)
             !           Project WVK onto B1,2,3:
             projb(1) = 0.0_dp
             projb(2) = 0.0_dp
             projb(3) = 0.0_dp
             DO k = 1, 3
                projb(1) = projb(1) + wvk(k)*a1(k)
                projb(2) = projb(2) + wvk(k)*a2(k)
                projb(3) = projb(3) + wvk(k)*a3(k)
             END DO
             !           Check (PROJA - PROJB): Is it integral ?
             DO k = 1, 3
                diff = proja(k) - projb(k)
                IF (ABS(REAL(NINT(diff),KIND=dp)-diff)>delta) CYCLE jloop
             END DO
             !           DIFF is integral: remove WVK from mesh:
             CALL remove(wvk,jplace,igarb0,igarbg,nkpoint,nhash,list,rlist, &
                  delta)
             !           If WVK actually removed, increment IREMOV
             IF (jplace>0) iremov = iremov + 1
          END DO jloop
       END DO
       IF (iremov>0 .AND. iout>0) WRITE (iout,'(A,A,/,A,I6,A,/)') &
            ' K290| Some of these mesh points are related by lattice ', &
            'translation vectors', ' K290| ', iremov, &
            ' of the mesh points removed.'
    END IF
    !------------------------------------------------------------------------------!
    !        IN THE MESH OF WAVEVECTORS, NOW SEARCH FOR EQUIVALENT POINTS:
    !        THE INVERSION (TIME REVERSAL !) MAY BE USED.
    !------------------------------------------------------------------------------!
    DO iwvk = 1, imesh
       !        IF(INCLUD(IWVK) .EQ. YES) GOTO 350
       IF (BTEST(includ(iwvk),0)) CYCLE
       !       IWVK has not been encountered previously: new special point,
       !       (only if WVK is not a garbage vector, however.)
       !        INCLUD(IWVK) = YES
       includ(iwvk) = IBSET(includ(iwvk),0)
       iplace = iwvk
       CALL mesh(iout,wvk,iplace,igarb0,igarbg,nkpoint,nhash,list,rlist, &
            delta)
       !       Find out whether Wvk is in the garbage list
       CALL garbag(wvk,igarbage,igarb0,nkpoint,nhash,list,rlist,delta)
       IF (igarbage>0) CYCLE
       ntot = ntot + 1
       !       Give the index in the special k points table.
       includ(iwvk) = includ(iwvk) + ntot*2
       DO i = 1, 3
          wvkl(i,ntot) = wvk(i)
       END DO
       lwght(ntot) = 1
       !------------------------------------------------------------------------------!
       !       Find all the equivalent points (symmetry given by atoms)
       DO n = 1, nc
          !         Rotate:
          DO i = 1, 3
             wva(i) = 0.0_dp
             DO j = 1, 3
                wva(i) = wva(i) + r(i,j,ib(n))*wvk(j)
             END DO
          END DO
          ibsign = + 1
363       CONTINUE
          !         Find WVA in the list
          iplace = -1
          CALL mesh(iout,wva,iplace,igarb0,igarbg,nkpoint,nhash,list,rlist, &
               delta)
          IF (iplace==0) THEN
             IF (istriz==-1) THEN
                !             No symmetrisation -> WVA not in the list
                GO TO 364
             ELSE
               ! Find out whether WVA is in the garbage list
               CALL garbag(wva,igarbage,igarb0,nkpoint,nhash,list,rlist,delta)
               IF (igarbage==0) THEN
                 ! I think this case is impossible (NC <= NCBRAV)
                 ! Error message
                 GOTO 490
               END IF
             END IF
          END IF
          !         Find out whether WVA is in the garbage list
          CALL garbag(wva,igarbage,igarb0,nkpoint,nhash,list,rlist,delta)
          IF (igarbage>0) CYCLE
          !         Was WVA encountered before ?
          !          IF(INCLUD(IPLACE) .EQ. YES) GOTO 364
          IF (BTEST(includ(iplace),0)) GO TO 364
          !         Increment weight.
          lwght(ntot) = lwght(ntot) + 1
          lrot(lwght(ntot),ntot) = ib(n)*ibsign
          !          INCLUD(IPLACE) = YES
          includ(iplace) = IBSET(includ(iplace),0)
          !         This k-point is an image of a special k-point.
          !         Put the index of the special k-point.
          includ(iplace) = includ(iplace) + ntot*2
364       CONTINUE
          IF (ibsign==-1 .OR. inv==0) CYCLE
          !         The case where we also apply the inversion to WVA
          !         Repeat the search, but for -WVA
          ibsign = -1
          DO i = 1, 3
             wva(i) = -wva(i)
          END DO
          GO TO 363
       END DO
    END DO
    !------------------------------------------------------------------------------!
    !        TOTAL NUMBER OF SPECIAL POINTS: NTOT
    !        BEFORE USING THE LIST WVKL AS WAVE VECTORS, THEY HAVE TO BE
    !        MULTIPLIED BY 2*PI
    !        THE LIST OF WEIGHTS LWGHT IS NOT NORMALIZED
    !------------------------------------------------------------------------------!
    IF (ntot>nkpoint .AND. iout>0) THEN
       WRITE (iout,*) ' K290| In sppt2 number of special points = ', ntot
       WRITE (iout,*) ' K290| but nkpoint = ', nkpoint
       ntot = -1
    END IF
    IF (iout>0) THEN
       !       Write the index table relating k points in the mesh
       !       with special k points
       WRITE (iout,'(/,A)') &
            ' K290| Cross table relating mesh points with special points:'
       WRITE (iout,'(5(4X,"IK -> SK"))')
       DO i = 1, imesh
          iplace = includ(i)/2
          WRITE (iout,'(1X,I5,1X,I5)',ADVANCE="NO") i, iplace
          IF (MOD(i,5)==0) WRITE (iout,*)
       END DO
       IF (MOD(j-1,5)/=0) WRITE (iout,*)
    END IF
    RETURN
    !------------------------------------------------------------------------------!
    !        ERROR MESSAGES
    !------------------------------------------------------------------------------!
450 CONTINUE
    WRITE (*,'(A,/)') ' SUBROUTINE SPPT2 *** FATAL ERROR ***'
    WRITE (*,'(A,3F10.4,/,A,3F10.4,A,/,A,I3,A)') ' THE VECTOR     ', &
         wva, ' GENERATED FROM ', wvk, ' IN THE BASIC MESH', &
         ' BY ROTATION NO. ', ibrav(iop), ' IS OUTSIDE THE 1BZ'
    CALL cp__b("k290.F",2542,"")
    !------------------------------------------------------------------------------!
470 CONTINUE
    WRITE (*,'(A,/)') ' SUBROUTINE SPPT2 *** FATAL ERROR ***'
    WRITE (*,*) 'MESH SIZE EXCEEDS NKPOINT=', nkpoint
    CALL cp__b("k290.F",2547,"")
    !------------------------------------------------------------------------------!
490 CONTINUE
    WRITE (*,'(A,/)') ' SUBROUTINE SPPT2 *** FATAL ERROR ***'
    WRITE (*,'(A,3F10.4,/,A,3F10.4,A,/,A,I3,A)') ' THE VECTOR     ', &
         wva, ' GENERATED FROM ', wvk, ' IN THE BASIC MESH', &
         ' BY ROTATION NO. ', ib(n), ' IS NOT IN THE LIST'
    CALL cp__b("k290.F",2554,"")
  END SUBROUTINE sppt2
! **************************************************************************************************
!> \brief ...
!> \param iout ...
!> \param wvk ...
!> \param iplace ...
!> \param igarb0 ...
!> \param igarbg ...
!> \param nmesh ...
!> \param nhash ...
!> \param list ...
!> \param rlist ...
!> \param delta ...
!> \par Original Header
!>         MESH MAINTAINS A LIST OF VECTORS FOR PLACEMENT AND/OR LOOKUP
!>
!>         ADDITIONAL ENTRY POINTS: REMOVE .... REMOVE VECTOR FROM LIST
!>                                  GARBAG .... WAS VECTOR REMOVED ?
!>
!>         WVK ....... VECTOR
!>         IPLACE .... ON INPUT: -2 MEANS: INITIALIZE  THE LIST
!>                                         (AND RETURN)
!>                               -1 MEANS: FIND WVK IN THE LIST
!>                                0 MEANS: ADD  WVK TO THE LIST
!>                               >0 MEANS: RETURN WVK NO. IPLACE
!>                    ON OUTPUT: THE POSITION ASSIGNED TO WVK
!>                               (=0 IF WVK IS NOT IN THE LIST)
!>         DELTA      REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
! **************************************************************************************************
  SUBROUTINE mesh(iout,wvk,iplace,igarb0,igarbg,nmesh,nhash,list,rlist, &
       delta)

      INTEGER                                            :: iout
      REAL(KIND=dp)                                      :: wvk(3)
      INTEGER                                            :: iplace, igarb0, igarbg, nmesh, nhash, &
                                                            list(nhash+nmesh)
      REAL(KIND=dp)                                      :: rlist(3,nmesh), delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mesh', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nil = 0

      INTEGER                                            :: i, ihash, ipoint, j
      INTEGER, SAVE                                      :: istore = 0
      REAL(KIND=dp)                                      :: delta1, rhash

    delta1 = 10.0_dp*delta
    IF (iplace<=-2) THEN
       DO i = 1, nhash + nmesh
          list(i) = nil
       END DO
       istore = 1
       !       IGARB0 points to a linked list of removed WVKS (the garbage).
       igarb0 = 0
       igarbg = 0
       RETURN
       !------------------------------------------------------------------------------!
    ELSE IF ((iplace>-2) .AND. (iplace<=0)) THEN
       !       The particular HASH function used in this case:
       rhash = 0.7890E0_dp*wvk(1) + 0.6810E0_dp*wvk(2) + 0.5811E0_dp*wvk(3) + delta
       ihash = INT(ABS(rhash)*REAL(nhash,KIND=dp))
       ihash = MOD(ihash,nhash) + nmesh + 1
       !       Search for WVK in linked list
       ipoint = list(ihash)
       DO i = 1, 100
          !         List exhausted
          IF (ipoint==nil) GO TO 130
          !         Compare WVK with this element
          DO j = 1, 3
             IF (ABS(wvk(j)-rlist(j,ipoint))>delta1) GO TO 115
          END DO
          !         WVK located
          GO TO 160
          !         Next element of list
115       CONTINUE
          ihash = ipoint
          ipoint = list(ihash)
       END DO
       !       List too long
       WRITE (*,'(2A,/,A)') &
            ' SUBROUTINE MESH *** FATAL ERROR *** LINKED LIST', &
            ' TOO LONG ***', ' CHOOSE A BETTER HASH-FUNCTION'
       CALL cp__b("k290.F",2636,"")
       !       WVK was not found
130    CONTINUE
       IF (iplace==-1) THEN
          !         IPLACE=-1 : search for WVK unsuccessful
          iplace = 0
          RETURN
       ELSE
          !         IPLACE=0: add WVK to the list
          list(ihash) = istore
          IF (istore>nmesh) THEN
             WRITE (*,'(A)') 'SUBROUTINE MESH *** FATAL ERROR ***'
             WRITE (*,'(A,I10,A,/,A,3F10.5)') ' ISTORE=', istore, &
                  ' EXCEEDS DIMENSIONS', ' WVK = ', wvk
             CALL cp__b("k290.F",2650,"")
          END IF
          list(istore) = nil
          DO i = 1, 3
             rlist(i,istore) = wvk(i)
          END DO
          istore = istore + 1
          iplace = istore - 1
          RETURN
       END IF
       !       WVK was found
160    CONTINUE
       IF (iplace==0) RETURN
       !       IPLACE=-1
       iplace = ipoint
       RETURN
    ELSE
       !------------------------------------------------------------------------------!
       !        Return a wavevector (IPLACE > 0)
       !------------------------------------------------------------------------------!
       ipoint = iplace
       IF (ipoint<istore) THEN
          DO i = 1, 3
             wvk(i) = rlist(i,ipoint)
          END DO
          RETURN
       ENDIF
    END IF
    !------------------------------------------------------------------------------!
    !        Error - beyond list
    !------------------------------------------------------------------------------!
    IF (iout>0) WRITE (iout,'(A,/,A,I5,A,/)') &
         ' SUBROUTINE MESH *** WARNING ***', ' IPLACE = ', iplace, &
         ' IS BEYOND THE LISTS - WVK SET TO 1.0E38'
    DO i = 1, 3
       wvk(i) = 1.0E38_dp
    END DO
  END SUBROUTINE mesh

! **************************************************************************************************
!> \brief ...
!> \param wvk ...
!> \param iplace ...
!> \param igarb0 ...
!> \param igarbg ...
!> \param nmesh ...
!> \param nhash ...
!> \param list ...
!> \param rlist ...
!> \param delta ...
!> \par Original Header
!>         ENTRY POINT FOR REMOVING A WAVEVECTOR
!>
!>         INPUT:
!>           WVK(3)
!>           DELTA   REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
!>         OUTPUT:
!>           IPLACE ..... 1 IF WVK WAS REMOVED
!>                        0 IF WVK WAS NOT REMOVED
!>                          (WVK NOT IN THE LINKED LISTS)
! **************************************************************************************************
  SUBROUTINE remove(wvk,iplace,igarb0,igarbg,nmesh,nhash,list,rlist,delta)

      REAL(KIND=dp)                                      :: wvk(3)
      INTEGER                                            :: iplace, igarb0, igarbg, nmesh, nhash, &
                                                            list(nhash+nmesh)
      REAL(KIND=dp)                                      :: rlist(3,nmesh), delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'remove', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nil = 0

      INTEGER                                            :: i, ihash, ipoint, j
      LOGICAL                                            :: lskip
      REAL(KIND=dp)                                      :: delta1, rhash

    delta1 = 10.0_dp*delta
    !     The particular hash function used in this case:
    rhash = 0.7890E0_dp*wvk(1) + 0.6810E0_dp*wvk(2) + 0.5811E0_dp*wvk(3) + delta
    ihash = INT(ABS(rhash)*REAL(nhash,KIND=dp))
    ihash = MOD(ihash,nhash) + nmesh + 1
    !     Search for WVK in linked list
    ipoint = list(ihash)
    DO i = 1, 100
       !       List exhausted
       IF (ipoint==nil) THEN
          !         WVK was not found in the mesh:
          iplace = 0
          RETURN
       END IF
       !       Compare WVK with this element
       lskip = .FALSE.
       DO j = 1, 3
          IF (ABS(wvk(j)-rlist(j,ipoint))>delta1) lskip=.TRUE.
       END DO
       IF (lskip) THEN
         !       Next element of list
         ihash = ipoint
         ipoint = list(ihash)
       ELSE
         !       WVK located, now remove it from the list:
         list(ihash) = list(ipoint)
         !       LIST(IHASH) now points to the next element in the list,
         !       and the present WVK has become garbage.
         !       Add WVK to the list of garbage:
         IF (igarb0==0) THEN
            !         Start up the garbage list:
            igarb0 = ipoint
         ELSE
            list(igarbg) = ipoint
         END IF
         igarbg = ipoint
         list(igarbg) = nil
         iplace = 1
         RETURN
       END IF
    END DO
    !     List too long
    CALL cp__b("k290.F",2767,"")
  END SUBROUTINE remove

! **************************************************************************************************
!> \brief ...
!> \param wvk ...
!> \param iplace ...
!> \param igarb0 ...
!> \param nmesh ...
!> \param nhash ...
!> \param list ...
!> \param rlist ...
!> \param delta ...
!> \par Original Header
!>         ENTRY POINT FOR CHECKING IF A WAVEVECTOR
!>                     IS IN THE GARBAGE LIST
!>         INPUT:
!>           WVK(3)
!>           DELTA      REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
!>
!>         OUTPUT:
!>           IPLACE  ..... I > 0 IS THE PLACE IN THE GARBAGE LIST
!>                             0 IF WVK NOT AMONG THE GARBAGE
! **************************************************************************************************
  SUBROUTINE garbag(wvk,iplace,igarb0,nmesh,nhash,list,rlist,delta)
      REAL(KIND=dp)                                      :: wvk(3)
      INTEGER                                            :: iplace, igarb0, nmesh, nhash, &
                                                            list(nhash+nmesh)
      REAL(KIND=dp)                                      :: rlist(3,nmesh), delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'garbag', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nil = 0

      INTEGER                                            :: i, ihash, ipoint, j
      REAL(KIND=dp)                                      :: delta1

    delta1 = 10.0_dp*delta
    !     Search for WVK in linked list
    !     Point to the garbage list
    ipoint = igarb0
    nloop: DO i = 1, nmesh
       !       LIST EXHAUSTED
       IF (ipoint==nil) THEN
          !         WVK was not found in the mesh:
          iplace = 0
          RETURN
       END IF
       !       Compare WVK with this element
       DO j = 1, 3
          IF (ABS(wvk(j)-rlist(j,ipoint))>delta1) THEN
             !       Next element of list
             ihash = ipoint
             ipoint = list(ihash)
             CYCLE nloop
          ENDIF
       END DO
       !       WVK was located in the garbage list
       iplace = i
       RETURN

    END DO nloop
    !     List too long
    CALL cp__b("k290.F",2829,"")
  END SUBROUTINE garbag

! **************************************************************************************************
!> \brief ...
!> \param wvk ...
!> \param a1 ...
!> \param a2 ...
!> \param a3 ...
!> \param b1 ...
!> \param b2 ...
!> \param b3 ...
!> \param rsdir ...
!> \param nrsdir ...
!> \param nplane ...
!> \param delta ...
!> \par Original Header
!>         REDUCE WVK TO LIE ENTIRELY WITHIN THE 1ST BRILLOUIN ZONE
!>         BY ADDING B-VECTORS
!>         DELTA      REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
! **************************************************************************************************
  SUBROUTINE bzrduc(wvk,a1,a2,a3,b1,b2,b3,rsdir,nrsdir,nplane,delta)
      REAL(KIND=dp)                                      :: wvk(3), a1(3), a2(3), a3(3), b1(3), &
                                                            b2(3), b3(3)
      INTEGER                                            :: nrsdir
      REAL(KIND=dp)                                      :: rsdir(4,nrsdir)
      INTEGER                                            :: nplane
      REAL(KIND=dp)                                      :: delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'bzrduc', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nzones = 4, nnn = 2*nzones + 1, &
                                                            nn = nzones + 1, yes = 1

      INTEGER                                            :: i, i1, i2, i3, n1, n2, n3, nn1, nn2, nn3
      REAL(KIND=dp)                                      :: wb(3), wva(3)

    IF (inbz(wvk,rsdir,nrsdir,nplane,delta)==yes) RETURN
    !------------------------------------------------------------------------------!
    !     Look around +/- "NZONES" to locate vector
    !     NZONES may need to be increased for very anisotropic zones
    !------------------------------------------------------------------------------!
    !     WVK already inside 1Bz
    !------------------------------------------------------------------------------!
    !     Express WVK in the basis of B1,2,3.
    !     This permits an estimate of how far WVK is from the 1Bz.
    wb(1) = wvk(1)*a1(1) + wvk(2)*a1(2) + wvk(3)*a1(3)
    wb(2) = wvk(1)*a2(1) + wvk(2)*a2(2) + wvk(3)*a2(3)
    wb(3) = wvk(1)*a3(1) + wvk(2)*a3(2) + wvk(3)*a3(3)
    nn1 = NINT(wb(1))
    nn2 = NINT(wb(2))
    nn3 = NINT(wb(3))
    !     Look around the estimated vector for the one truly inside the 1Bz
    DO n1 = 1, nnn
       i1 = nn - n1 - nn1
       DO n2 = 1, nnn
          i2 = nn - n2 - nn2
          DO n3 = 1, nnn
             i3 = nn - n3 - nn3
             DO i = 1, 3
                wva(i) = wvk(i) + REAL(i1,KIND=dp)*b1(i) + REAL(i2,KIND=dp)*b2(i) + &
                     REAL(i3,kind=dp)*b3(i)
             END DO
             IF (inbz(wva,rsdir,nrsdir,nplane,delta)==yes) THEN
             !------------------------------------------------------------------------------!
             !     The reduced vector
                DO i = 1, 3
                   wvk(i) = wva(i)
                END DO
                RETURN
             ENDIF
          END DO
       END DO
    END DO
    !------------------------------------------------------------------------------!
    !     Fatal error
    WRITE (*,'(A,/,A,3F10.4,A)') ' SUBROUTINE BZRDUC *** FATAL ERROR ***', &
         ' WAVEVECTOR ', wvk, ' COULD NOT BE REDUCED TO THE 1BZ'
    CALL cp__b("k290.F",2906,"")

  END SUBROUTINE bzrduc
! **************************************************************************************************
!> \brief ...
!> \param wvk ...
!> \param rsdir ...
!> \param nrsdir ...
!> \param nplane ...
!> \param delta ...
!> \retval inbz ...
!> \par Original Header
!>         IS WVK IN THE 1ST BRILLOUIN ZONE ?
!>         CHECK WHETHER WVK LIES INSIDE ALL THE PLANES
!>         THAT DEFINE THE 1BZ.
!>         DELTA      REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
! **************************************************************************************************
  FUNCTION inbz(wvk,rsdir,nrsdir,nplane,delta)
      REAL(KIND=dp)                                      :: wvk(3)
      INTEGER                                            :: nrsdir
      REAL(KIND=dp)                                      :: rsdir(4,nrsdir)
      INTEGER                                            :: nplane
      REAL(KIND=dp)                                      :: delta
      INTEGER                                            :: inbz

      INTEGER, PARAMETER                                 :: no = 0, yes = 1

      INTEGER                                            :: n
      REAL(KIND=dp)                                      :: projct

!------------------------------------------------------------------------------!

    inbz = no
    DO n = 1, nplane
       projct = (rsdir(1,n)*wvk(1)+rsdir(2,n)*wvk(2)+rsdir(3,n)*wvk(3))/ &
            rsdir(4,n)
       !       WVK is outside the Bz
       IF (ABS(projct)>0.5E0_dp+delta) RETURN
    END DO
    inbz = yes
  END FUNCTION inbz

! **************************************************************************************************
!> \brief ...
!> \param iout ...
!> \param b1 ...
!> \param b2 ...
!> \param b3 ...
!> \param rsdir ...
!> \param nrsdir ...
!> \param nplane ...
!> \param delta ...
!> \par Original Header
!>         FIND THE VECTORS WHOSE HALVES DEFINE THE 1ST BRILLOUIN ZONE
!>         OUTPUT:
!>         NPLANE TELLS HOW MANY ELEMENTS OF RSDIR CONTAIN
!>                NORMAL VECTORS DEFINING THE PLANES
!>         METHOD: STARTING WITH THE PARALLELOPIPED SPANNED BY B1,2,3
!>         AROUND THE ORIGIN, VECTORS INSIDE A SUFFICIENTLY LARGE
!>         SPHERE ARE TESTED TO SEE WHETHER THE PLANES AT 1/2*B WILL
!>         FURTHER CONFINE THE 1BZ.
!>         THE RESULTING VECTORS ARE NOT CLEANED TO AVOID REDUNDANT
!>         PLANES.
!>         DELTA      REQUIRED ACCURACY (1.0E-6_dp IS A GOOD VALUE)
! **************************************************************************************************
  SUBROUTINE bzdefi(iout,b1,b2,b3,rsdir,nrsdir,nplane,delta)
      INTEGER                                            :: iout
      REAL(KIND=dp)                                      :: b1(3), b2(3), b3(3)
      INTEGER                                            :: nrsdir
      REAL(KIND=dp)                                      :: rsdir(4,nrsdir)
      INTEGER                                            :: nplane
      REAL(KIND=dp)                                      :: delta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'bzdefi', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, i1, i2, i3, j, n, n1, n2, n3, nb1, &
                                                            nb2, nb3, nnb1, nnb2, nnb3
      REAL(KIND=dp)                                      :: b1len, b2len, b3len, bmax, bvec(3), &
                                                            projct

!------------------------------------------------------------------------------!

    b1len = b1(1)**2 + b1(2)**2 + b1(3)**2
    b2len = b2(1)**2 + b2(2)**2 + b2(3)**2
    b3len = b3(1)**2 + b3(2)**2 + b3(3)**2
    !     Lattice containing entirely the brillouin zone
    bmax = b1len + b2len + b3len
    nb1 = INT(SQRT(bmax/b1len)+delta) + 1
    nb2 = INT(SQRT(bmax/b2len)+delta) + 1
    nb3 = INT(SQRT(bmax/b3len)+delta) + 1
    !     PRINT *,'NB1,2,3 = ',NB1,NB2,NB3
    DO i = 1, nrsdir
       DO j = 1, 4
          rsdir(j,i) = 0.0_dp
       END DO
    END DO
    !     1Bz is certainly confined inside the 1/2(B1,B2,B3) parallelopiped
    DO i = 1, 3
       rsdir(i,1) = b1(i)
       rsdir(i,2) = b2(i)
       rsdir(i,3) = b3(i)
    END DO
    rsdir(4,1) = b1len
    rsdir(4,2) = b2len
    rsdir(4,3) = b3len
    !     Starting confinement: 3 planes
    nplane = 3
    nnb1 = 2*nb1 + 1
    nnb2 = 2*nb2 + 1
    nnb3 = 2*nb3 + 1
    DO n1 = 1, nnb1
       i1 = nb1 + 1 - n1
       DO n2 = 1, nnb2
          i2 = nb2 + 1 - n2
          n3loop: DO n3 = 1, nnb3
             i3 = nb3 + 1 - n3
             IF (i1==0 .AND. i2==0 .AND. i3==0) CYCLE n3loop
             DO i = 1, 3
                bvec(i) = REAL(i1,KIND=dp)*b1(i) + REAL(i2,KIND=dp)*b2(i) + &
                     REAL(i3,KIND=dp)*b3(i)
             END DO
             !           Does the plane of 1/2*BVEC narrow down the 1Bz ?
             DO n = 1, nplane
                projct = 0.5E0_dp*(rsdir(1,n)*bvec(1)+rsdir(2,n)*bvec(2)+rsdir(3, &
                     n)*bvec(3))/rsdir(4,n)
                !             1/2*BVEC is outside the Bz - skip this direction
                !             The 1.0E-6_dp takes care of single points touching the Bz,
                !             and of the -(plane)
                IF (ABS(projct)>0.5E0_dp-delta) CYCLE n3loop
             END DO
             !           1/2*BVEC further confines the 1Bz - include into RSDIR
             nplane = nplane + 1
             !           PRINT *,NPLANE,' PLANE INCLUDED, I1,2,3 = ',I1,I2,I3
             IF (nplane>nrsdir) THEN
                !------------------------------------------------------------------------------!
                !     Error messages
                IF (iout>0) THEN
                   WRITE (iout,'(A)') ' SUBROUTINE BZDEFI *** FATAL ERROR ***'
                   WRITE (iout,'(" TOO MANY PLANES, NRSDIR = ",I5)') nrsdir
                END IF
                CALL cp__b("k290.F",3046,"")
             ENDIF
             DO i = 1, 3
                rsdir(i,nplane) = bvec(i)
             END DO
             !           Length squared
             rsdir(4,nplane) = bvec(1)**2 + bvec(2)**2 + bvec(3)**2
          END DO n3loop
       END DO
    END DO
    !------------------------------------------------------------------------------!
    !     Print information
    IF (iout>0) WRITE (iout,'(A,I3,A,/,A,/,100(1X,3F10.4,/))') &
         ' THE 1ST BRILLOUIN ZONE IS CONFINED BY (AT MOST)', nplane, &
         ' PLANES', ' AS DEFINED BY THE +/- HALVES OF THE VECTORS:', &
         ((rsdir(i,n),i=1,3),n=1,nplane)

  END SUBROUTINE bzdefi

END MODULE k290
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Molecular symmetry routines
!> \par History
!>      2008 adapted from older routines by Matthias Krack
!> \author jgh
! **************************************************************************************************
MODULE molsym

   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE mathlib,                         ONLY: angle,&
                                              build_rotmat,&
                                              jacobi,&
                                              reflect_vector,&
                                              rotate_vector,&
                                              unit_matrix,&
                                              vector_product
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'molsym'

   INTEGER, PARAMETER :: maxcn = 20, &
                         maxsec = maxcn+1, &
                         maxses = 2*maxcn+1, &
                         maxsig = maxcn+1, &
                         maxsn = 2*maxcn

   PUBLIC :: molsym_type
   PUBLIC :: release_molsym, molecular_symmetry, print_symmetry

!***

! **************************************************************************************************
!> \brief Container for information about molecular symmetry
!> \param ain               : Atomic identification numbers (symmetry code).
!> \param aw                : Atomic weights for the symmetry analysis.
!> \param eps_geo           : Accuracy requested for analysis
!> \param inptostd          : Transformation matrix for input to standard orientation.
!> \param point_group_symbol: Point group symbol.
!> \param rotmat            : Rotation matrix.
!> \param sec               : List of C axes
!> \param                     (sec(:,i,j) => x,y,z of the ith j-fold C axis).
!> \param ses               : List of S axes
!> \param                     (ses(:,i,j) => x,y,z of the ith j-fold S axis).
!> \param sig               : List of mirror planes
!> \param                     (sig(:,i) => x,y,z of the ith mirror plane).
!> \param center_of_mass    : Shift vector for the center of mass.
!> \param tenmat            : Molecular tensor of inertia.
!> \param tenval            : Eigenvalues of the molecular tensor of inertia.
!> \param tenvec            : Eigenvectors of the molecular tensor of inertia.
!> \param group_of          : Group of equivalent atom.
!> \param llequatom         : Lower limit of a group in symequ_list.
!> \param ncn               : Degree of the C axis with highest degree.
!> \param ndod              : Number of substituted dodecahedral angles.
!> \param nequatom          : Number of equivalent atoms.
!> \param ngroup            : Number of groups of equivalent atoms.
!> \param nico              : Number of substituted icosahedral angles.
!> \param nlin              : Number of substituted angles of 180 degrees.
!> \param nsec              : Number of C axes.
!> \param nses              : Number of S axes.
!> \param nsig              : Number of mirror planes.
!> \param nsn               : Degree of the S axis with highest degree.
!> \param ntet              : Number of substituted tetrahedral angles.
!> \param point_group_order : Group order.
!> \param symequ_list       : List of all atoms ordered in groups of equivalent atoms.
!> \param ulequatom         : Upper limit of a group in symequ_list.
!> \param cubic : .TRUE., if a cubic point group was found (T,Th,Td,O,Oh,I,Ih).
!> \param dgroup: .TRUE., if a point group of D symmetry was found (Dn,Dnh,Dnd)
!> \param igroup: .TRUE., if a point group of icosahedral symmetry was found (I,Ih).
!> \param invers: .TRUE., if the molecule has a center of inversion.
!> \param linear: .TRUE., if the molecule is linear.
!> \param maxis : .TRUE., if the molecule has a main axis.
!> \param ogroup: .TRUE., if a point group of octahedral symmetry was found (O,Oh)
!> \param sgroup: .TRUE., if a point group of S symmetry was found.
!> \param sigmad: .TRUE., if there is a sigma_d mirror plane.
!> \param sigmah: .TRUE., if there is a sigma_h mirror plane.
!> \param sigmav: .TRUE., if there is a sigma_v mirror plane.
!> \param tgroup: .TRUE., if a point group of tetrahedral symmetry was found (T,Th,Td).
!> \par History
!>      05.2008 created
!> \author jgh
! **************************************************************************************************
   TYPE molsym_type
      CHARACTER(4)                               :: point_group_symbol
      INTEGER                                    :: point_group_order
      INTEGER                                    :: ncn, ndod, ngroup, nico, nlin, nsig, nsn, ntet
      LOGICAL                                    :: cubic, dgroup, igroup, invers, linear, maxis, &
                                                    ogroup, sgroup, sigmad, sigmah, sigmav, tgroup
      REAL(KIND=dp)                              :: eps_geo
      REAL(KIND=dp), DIMENSION(3)                :: center_of_mass, tenval
      REAL(KIND=dp), DIMENSION(3)                :: x_axis, y_axis, z_axis
      REAL(KIND=dp), DIMENSION(:), POINTER       :: ain, aw
      REAL(KIND=dp), DIMENSION(3, 3)              :: inptostd, rotmat, tenmat, tenvec
      REAL(KIND=dp), DIMENSION(3, maxsig)         :: sig
      REAL(KIND=dp), DIMENSION(3, maxsec, 2:maxcn) :: sec
      REAL(KIND=dp), DIMENSION(3, maxses, 2:maxsn) :: ses
      INTEGER, DIMENSION(maxcn)                  :: nsec
      INTEGER, DIMENSION(maxsn)                  :: nses
      INTEGER, DIMENSION(:), POINTER             :: group_of, llequatom, nequatom, &
                                                    symequ_list, ulequatom
   END TYPE molsym_type

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief  Create an object of molsym type
!> \param sym ...
!> \param natoms ...
!> \author jgh
! **************************************************************************************************
   SUBROUTINE create_molsym(sym, natoms)
      TYPE(molsym_type), POINTER                         :: sym
      INTEGER, INTENT(IN)                                :: natoms

      CHARACTER(len=*), PARAMETER :: routineN = 'create_molsym', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(sym)) CALL release_molsym(sym)

      ALLOCATE (sym)

      ALLOCATE (sym%ain(natoms), sym%aw(natoms), sym%group_of(natoms), sym%llequatom(natoms), &
                sym%nequatom(natoms), sym%symequ_list(natoms), sym%ulequatom(natoms))

   END SUBROUTINE create_molsym

! **************************************************************************************************
!> \brief  release an object of molsym type
!> \param sym ...
!> \author jgh
! **************************************************************************************************
   SUBROUTINE release_molsym(sym)
      TYPE(molsym_type), POINTER                         :: sym

      CHARACTER(len=*), PARAMETER :: routineN = 'release_molsym', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(sym)))CALL cp__a("molsym.F",147)

      IF (ASSOCIATED(sym%aw)) THEN
         DEALLOCATE (sym%aw)
      END IF
      IF (ASSOCIATED(sym%ain)) THEN
         DEALLOCATE (sym%ain)
      END IF
      IF (ASSOCIATED(sym%group_of)) THEN
         DEALLOCATE (sym%group_of)
      END IF
      IF (ASSOCIATED(sym%llequatom)) THEN
         DEALLOCATE (sym%llequatom)
      END IF
      IF (ASSOCIATED(sym%nequatom)) THEN
         DEALLOCATE (sym%nequatom)
      END IF
      IF (ASSOCIATED(sym%symequ_list)) THEN
         DEALLOCATE (sym%symequ_list)
      END IF
      IF (ASSOCIATED(sym%ulequatom)) THEN
         DEALLOCATE (sym%ulequatom)
      END IF

      DEALLOCATE (sym)

   END SUBROUTINE release_molsym

! **************************************************************************************************

! **************************************************************************************************
!> \brief Add a new C_n axis to the list sec, but first check, if the
!>         Cn axis is already in the list.
!> \param n ...
!> \param a ...
!> \param sym ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE addsec(n, a, sym)
      INTEGER, INTENT(IN)                                :: n
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: a
      TYPE(molsym_type), INTENT(inout)                   :: sym

      CHARACTER(len=*), PARAMETER :: routineN = 'addsec', routineP = moduleN//':'//routineN

      INTEGER                                            :: isec
      REAL(dp)                                           :: length_of_a, scapro
      REAL(dp), DIMENSION(3)                             :: d

      length_of_a = SQRT(a(1)*a(1)+a(2)*a(2)+a(3)*a(3))
      d(:) = a(:)/length_of_a

      ! Check, if the current Cn axis is already in the list
      DO isec = 1, sym%nsec(n)
         scapro = sym%sec(1, isec, n)*d(1)+sym%sec(2, isec, n)*d(2)+sym%sec(3, isec, n)*d(3)
         IF (ABS(ABS(scapro)-1.0_dp) < sym%eps_geo) RETURN
      END DO
      sym%ncn = MAX(sym%ncn, n)

      ! Add the new Cn axis to the list sec
      IF(.NOT.(sym%nsec(n) < maxsec))CALL cp__a("molsym.F",209)
      sym%nsec(1) = sym%nsec(1)+1
      sym%nsec(n) = sym%nsec(n)+1
      sym%sec(:, sym%nsec(n), n) = d(:)

   END SUBROUTINE addsec

! **************************************************************************************************
!> \brief  Add a new Sn axis to the list ses, but first check, if the
!>         Sn axis is already in the list.
!> \param n ...
!> \param a ...
!> \param sym ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE addses(n, a, sym)
      INTEGER, INTENT(IN)                                :: n
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: a
      TYPE(molsym_type), INTENT(inout)                   :: sym

      CHARACTER(len=*), PARAMETER :: routineN = 'addses', routineP = moduleN//':'//routineN

      INTEGER                                            :: ises
      REAL(dp)                                           :: length_of_a, scapro
      REAL(dp), DIMENSION(3)                             :: d

      length_of_a = SQRT(a(1)*a(1)+a(2)*a(2)+a(3)*a(3))
      d(:) = a(:)/length_of_a

      ! Check, if the current Sn axis is already in the list
      DO ises = 1, sym%nses(n)
         scapro = sym%ses(1, ises, n)*d(1)+sym%ses(2, ises, n)*d(2)+sym%ses(3, ises, n)*d(3)
         IF (ABS(ABS(scapro)-1.0_dp) < sym%eps_geo) RETURN
      END DO
      sym%nsn = MAX(sym%nsn, n)

      ! Add the new Sn axis to the list ses
      IF(.NOT.(sym%nses(n) < maxses))CALL cp__a("molsym.F",248)
      sym%nses(1) = sym%nses(1)+1
      sym%nses(n) = sym%nses(n)+1
      sym%ses(:, sym%nses(n), n) = d(:)

   END SUBROUTINE addses

! **************************************************************************************************
!> \brief  Add a new mirror plane to the list sig, but first check, if the
!>         normal vector of the mirror plane is already in the list.
!> \param a ...
!> \param sym ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE addsig(a, sym)
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: a
      TYPE(molsym_type), INTENT(inout)                   :: sym

      CHARACTER(len=*), PARAMETER :: routineN = 'addsig', routineP = moduleN//':'//routineN

      INTEGER                                            :: isig
      REAL(dp)                                           :: length_of_a, scapro
      REAL(dp), DIMENSION(3)                             :: d

      length_of_a = SQRT(a(1)*a(1)+a(2)*a(2)+a(3)*a(3))
      d(:) = a(:)/length_of_a

      ! Check, if the normal vector of the current mirror plane is already in the list
      DO isig = 1, sym%nsig
         scapro = sym%sig(1, isig)*d(1)+sym%sig(2, isig)*d(2)+sym%sig(3, isig)*d(3)
         IF (ABS(ABS(scapro)-1.0_dp) < sym%eps_geo) RETURN
      END DO

      ! Add the normal vector of the new mirror plane to the list sig
      IF(.NOT.(sym%nsig < maxsig))CALL cp__a("molsym.F",284)
      sym%nsig = sym%nsig+1
      sym%sig(:, sym%nsig) = d(:)

   END SUBROUTINE addsig

! **************************************************************************************************
!> \brief  Symmetry handling for only one atom.
!> \param sym ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE atomsym(sym)
      TYPE(molsym_type), INTENT(inout)                   :: sym

      CHARACTER(len=*), PARAMETER :: routineN = 'atomsym', routineP = moduleN//':'//routineN

! Set point group symbol

      sym%point_group_symbol = "R(3)"

      ! Set variables like D*h
      sym%linear = .TRUE.
      sym%invers = .TRUE.
      sym%dgroup = .TRUE.
      sym%sigmah = .TRUE.

   END SUBROUTINE atomsym

! **************************************************************************************************
!> \brief Search for point groups with AXial SYMmetry (Cn,Cnh,Cnv,Dn,Dnh,Dnd,S2n).
!> \param coord ...
!> \param sym ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE axsym(coord, sym)
      REAL(Kind=dp), DIMENSION(:, :), INTENT(inout)      :: coord
      TYPE(molsym_type), INTENT(inout)                   :: sym

      CHARACTER(len=*), PARAMETER :: routineN = 'axsym', routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, isig, jatom, m, n, natoms, nx, &
                                                            ny, nz
      REAL(dp)                                           :: phi
      REAL(dp), DIMENSION(3)                             :: a, b, d

! Find the correct main axis and rotate main axis to z axis

      IF ((sym%ncn == 2) .AND. (sym%nses(2) == 3)) THEN
         IF (sym%nsn == 4) THEN
            ! Special case: D2d
            phi = angle(sym%ses(:, 1, sym%nsn), sym%z_axis(:))
            d(:) = vector_product(sym%ses(:, 1, sym%nsn), sym%z_axis(:))
            CALL rotate_molecule(phi, d(:), sym, coord)
         ELSE
            ! Special cases: D2 and D2h
            phi = 0.5_dp*pi
            nx = naxis(sym%x_axis(:), coord, sym)
            ny = naxis(sym%y_axis(:), coord, sym)
            nz = naxis(sym%z_axis(:), coord, sym)
            IF ((nx > ny) .AND. (nx > nz)) THEN
               CALL rotate_molecule(-phi, sym%y_axis(:), sym, coord)
            ELSE IF ((ny > nz) .AND. (ny > nx)) THEN
               CALL rotate_molecule(phi, sym%x_axis(:), sym, coord)
            END IF
         END IF
      ELSE
         phi = angle(sym%sec(:, 1, sym%ncn), sym%z_axis(:))
         d(:) = vector_product(sym%sec(:, 1, sym%ncn), sym%z_axis(:))
         CALL rotate_molecule(phi, d(:), sym, coord)
      END IF

      ! Search for C2 axes perpendicular to the main axis
      ! Loop over all pairs of atoms (except on the z axis)
      natoms = SIZE(coord, 2)
      DO iatom = 1, natoms
         a(:) = coord(:, iatom)
         IF ((ABS(a(1)) > sym%eps_geo) .OR. (ABS(a(2)) > sym%eps_geo)) THEN
            a(3) = 0.0_dp
            IF (caxis(2, a(:), sym, coord)) CALL addsec(2, a(:), sym)
            d(:) = vector_product(a(:), sym%z_axis(:))
            IF (sigma(d(:), sym, coord)) CALL addsig(d(:), sym)
            DO jatom = iatom+1, natoms
               b(:) = coord(:, jatom)
               IF ((ABS(b(1)) > sym%eps_geo) .OR. (ABS(b(2)) > sym%eps_geo)) THEN
                  b(3) = 0.0_dp
                  phi = 0.5_dp*angle(a(:), b(:))
                  d(:) = vector_product(a(:), b(:))
                  b(:) = rotate_vector(a(:), phi, d(:))
                  IF (caxis(2, b(:), sym, coord)) CALL addsec(2, b(:), sym)
                  d(:) = vector_product(b(:), sym%z_axis)
                  IF (sigma(d(:), sym, coord)) CALL addsig(d(:), sym)
               END IF
            END DO
         END IF
      END DO

      ! Check the xy plane for mirror plane
      IF (sigma(sym%z_axis(:), sym, coord)) THEN
         CALL addsig(sym%z_axis(:), sym)
         sym%sigmah = .TRUE.
      END IF

      ! Set logical variables for point group determination ***
      IF (sym%nsec(2) > 1) THEN
         sym%dgroup = .TRUE.
         IF ((.NOT. sym%sigmah) .AND. (sym%nsig > 0)) sym%sigmad = .TRUE.
      ELSE
         IF ((.NOT. sym%sigmah) .AND. (sym%nsig > 0)) sym%sigmav = .TRUE.
         ! Cnh groups with n>3 were wrong
         ! IF (sym%nsn > 3) sym%sgroup = .TRUE.
         IF ((.NOT. sym%sigmah) .AND. (sym%nsn > 3)) sym%sgroup = .TRUE.
      END IF

      ! Rotate to standard orientation
      n = 0
      m = 0
      IF ((sym%dgroup .AND. sym%sigmah) .OR. (sym%dgroup .AND. sym%sigmad) .OR. sym%sigmav) THEN
         ! Dnh, Dnd or Cnv
         DO isig = 1, sym%nsig
            IF (ABS(sym%sig(3, isig)) < sym%eps_geo) THEN
               n = nsigma(sym%sig(:, isig), sym, coord)
               IF (n > m) THEN
                  m = n
                  a(:) = sym%sig(:, isig)
               END IF
            END IF
         END DO
         IF (m > 0) THEN
            ! Check for special case: C2v with all atoms in a plane
            IF (sym%sigmav .AND. (m == natoms)) THEN
               phi = angle(a(:), sym%x_axis(:))
               d(:) = vector_product(a(:), sym%x_axis(:))
            ELSE
               phi = angle(a(:), sym%y_axis(:))
               d(:) = vector_product(a(:), sym%y_axis(:))
            END IF
            CALL rotate_molecule(phi, d(:), sym, coord)
         END IF
      ELSE IF (sym%sigmah) THEN
         ! Cnh
         DO iatom = 1, natoms
            IF (ABS(coord(3, iatom)) < sym%eps_geo) THEN
               n = naxis(coord(:, iatom), coord, sym)
               IF (n > m) THEN
                  m = n
                  a(:) = coord(:, iatom)
               END IF
            END IF
         END DO
         IF (m > 0) THEN
            phi = angle(a(:), sym%x_axis(:))
            d(:) = vector_product(a(:), sym%x_axis(:))
            CALL rotate_molecule(phi, d(:), sym, coord)
         END IF
         ! No action for Dn, Cn or S2n ***
      END IF

   END SUBROUTINE axsym

! **************************************************************************************************
!> \brief Generate a symmetry list to identify equivalent atoms.
!> \param sym ...
!> \param coord ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE build_symequ_list(sym, coord)
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(Kind=dp), DIMENSION(:, :), INTENT(inout)      :: coord

      CHARACTER(len=*), PARAMETER :: routineN = 'build_symequ_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iatom, icn, iequatom, incr, isec, &
                                                            ises, isig, isn, jatom, jcn, jsn, &
                                                            natoms
      REAL(KIND=dp)                                      :: phi
      REAL(KIND=dp), DIMENSION(3)                        :: a

      natoms = SIZE(coord, 2)

      IF (sym%sigmah) THEN
         incr = 1
      ELSE
         incr = 2
      END IF

      ! Initialize the atom and the group counter
      iatom = 1
      sym%ngroup = 1

      loop: DO

         ! Loop over all mirror planes
         DO isig = 1, sym%nsig
            a(:) = reflect_vector(coord(:, iatom), sym%sig(:, isig))
            iequatom = equatom(iatom, a(:), sym, coord)
            IF ((iequatom > 0) .AND. (.NOT. in_symequ_list(iequatom, sym))) THEN
               sym%ulequatom(sym%ngroup) = sym%ulequatom(sym%ngroup)+1
               sym%symequ_list(sym%ulequatom(sym%ngroup)) = iequatom
               sym%nequatom(sym%ngroup) = sym%nequatom(sym%ngroup)+1
            END IF
         END DO

         ! Loop over all Cn axes
         DO icn = 2, sym%ncn
            DO isec = 1, sym%nsec(icn)
               DO jcn = 1, icn-1
                  IF (newse(jcn, icn)) THEN
                     phi = 2.0_dp*pi*REAL(jcn, KIND=dp)/REAL(icn, KIND=dp)
                     a(:) = rotate_vector(coord(:, iatom), phi, sym%sec(:, isec, icn))
                     iequatom = equatom(iatom, a(:), sym, coord)
                     IF ((iequatom > 0) .AND. (.NOT. in_symequ_list(iequatom, sym))) THEN
                        sym%ulequatom(sym%ngroup) = sym%ulequatom(sym%ngroup)+1
                        sym%symequ_list(sym%ulequatom(sym%ngroup)) = iequatom
                        sym%nequatom(sym%ngroup) = sym%nequatom(sym%ngroup)+1
                     END IF
                  END IF
               END DO
            END DO
         END DO

         ! Loop over all Sn axes
         DO isn = 2, sym%nsn
            DO ises = 1, sym%nses(isn)
               DO jsn = 1, isn-1, incr
                  IF (newse(jsn, isn)) THEN
                     phi = 2.0_dp*pi*REAL(jsn, KIND=dp)/REAL(isn, KIND=dp)
                     a(:) = rotate_vector(coord(:, iatom), phi, sym%ses(:, ises, isn))
                     a(:) = reflect_vector(a(:), sym%ses(:, ises, isn))
                     iequatom = equatom(iatom, a(:), sym, coord)
                     IF ((iequatom > 0) .AND. (.NOT. in_symequ_list(iequatom, sym))) THEN
                        sym%ulequatom(sym%ngroup) = sym%ulequatom(sym%ngroup)+1
                        sym%symequ_list(sym%ulequatom(sym%ngroup)) = iequatom
                        sym%nequatom(sym%ngroup) = sym%nequatom(sym%ngroup)+1
                     END IF
                  END IF
               END DO
            END DO
         END DO

         ! Exit loop, if all atoms are in the list
         IF (sym%ulequatom(sym%ngroup) == natoms) EXIT

         ! Search for the next atom which is not in the list yet
         DO jatom = 2, natoms
            IF (.NOT. in_symequ_list(jatom, sym)) THEN
               iatom = jatom
               sym%ngroup = sym%ngroup+1
               sym%llequatom(sym%ngroup) = sym%ulequatom(sym%ngroup-1)+1
               sym%ulequatom(sym%ngroup) = sym%llequatom(sym%ngroup)
               sym%symequ_list(sym%llequatom(sym%ngroup)) = iatom
               CYCLE loop
            END IF
         END DO

      END DO loop

      ! Generate list vector group_of
      DO i = 1, sym%ngroup
         DO iequatom = sym%llequatom(i), sym%ulequatom(i)
            sym%group_of(sym%symequ_list(iequatom)) = i
         END DO
      END DO

   END SUBROUTINE build_symequ_list

! **************************************************************************************************
!> \brief Rotate the molecule about a n-fold axis defined by the vector a
!>        using a rotation angle of 2*pi/n. Check, if the axis is a Cn axis.
!> \param n ...
!> \param a ...
!> \param sym ...
!> \param coord ...
!> \retval caxis ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   FUNCTION caxis(n, a, sym, coord)
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(Kind=dp), DIMENSION(:, :), INTENT(inout)      :: coord
      LOGICAL                                            :: caxis

      INTEGER                                            :: iatom, natoms
      REAL(KIND=dp)                                      :: length_of_a, phi
      REAL(KIND=dp), DIMENSION(3)                        :: b

      caxis = .FALSE.
      length_of_a = SQRT(a(1)*a(1)+a(2)*a(2)+a(3)*a(3))

      ! Check the length of the axis vector a
      natoms = SIZE(coord, 2)
      IF (length_of_a > sym%eps_geo) THEN
         ! Calculate the rotation angle phi and build up the rotation matrix rotmat
         phi = 2.0_dp*pi/REAL(n, dp)
         CALL build_rotmat(phi, a(:), sym%rotmat(:, :))
         ! Check all atoms
         DO iatom = 1, natoms
            ! Rotate the actual atom by 2*pi/n about a
            b(:) = MATMUL(sym%rotmat(:, :), coord(:, iatom))
            ! Check if the coordinates of the rotated atom
            ! are in the coordinate set of the molecule
            IF (equatom(iatom, b(:), sym, coord) == 0) RETURN
         END DO
         ! The axis defined by vector a is a Cn axis, thus return with caxis = .TRUE.
         caxis = .TRUE.
      END IF

   END FUNCTION caxis

! **************************************************************************************************
!> \brief Check for a point group of cubic symmetry (I,Ih,O,Oh,T,Th,Td).
!> \param sym ...
!> \param coord ...
!> \param failed ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE cubsym(sym, coord, failed)
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord
      LOGICAL, INTENT(INOUT)                             :: failed

      INTEGER                                            :: i, iatom, iax, ic3, isec, jatom, jc3, &
                                                            jsec, katom, natoms, nc3
      REAL(KIND=dp)                                      :: phi1, phi2, phidd, phidi, phiii
      REAL(KIND=dp), DIMENSION(3)                        :: a, b, c, d

! Angle between two adjacent atoms of the dodecahedron => <(C3,C3)

      phidd = ATAN(0.4_dp*SQRT(5.0_dp))
      ! Angle between two adjacent atoms of the icosahedron and the dodecahedron => <(C5,C3)
      phidi = ATAN(3.0_dp-SQRT(5.0_dp))
      ! Angle between two adjacent atoms of the icosahedron <(C5,C5)
      phiii = ATAN(2.0_dp)

      ! Find a pair of C3 axes
      natoms = SIZE(coord, 2)
      DO iatom = 1, natoms
         ! Check all atomic vectors for C3 axis
         IF (caxis(3, coord(:, iatom), sym, coord)) THEN
            CALL addsec(3, coord(:, iatom), sym)
            IF (sym%nsec(3) > 1) EXIT
         END IF
      END DO

      ! Check the center of mass vector of a triangle
      ! defined by three atomic vectors for C3 axis
      IF (sym%nsec(3) < 2) THEN

         loop: DO iatom = 1, natoms
            a(:) = coord(:, iatom)
            DO jatom = iatom+1, natoms
               DO katom = jatom+1, natoms
                  IF ((ABS(dist(coord(:, iatom), coord(:, jatom)) &
                           -dist(coord(:, iatom), coord(:, katom))) < sym%eps_geo) .AND. &
                      (ABS(dist(coord(:, iatom), coord(:, jatom)) &
                           -dist(coord(:, jatom), coord(:, katom))) < sym%eps_geo)) THEN
                     b(:) = a(:)+coord(:, jatom)+coord(:, katom)
                     IF (caxis(3, b(:), sym, coord)) THEN
                        CALL addsec(3, b(:), sym)
                        IF (sym%nsec(3) > 1) EXIT loop
                     END IF
                  END IF
               END DO
            END DO
         END DO loop

      END IF

      ! Return after unsuccessful search for a pair of C3 axes
      IF (sym%nsec(3) < 2) THEN
         failed = .TRUE.
         RETURN
      END IF

      ! Generate the remaining C3 axes
      nc3 = 0
      DO
         nc3 = sym%nsec(3)
         DO ic3 = 1, nc3
            a(:) = sym%sec(:, ic3, 3)
            DO jc3 = 1, nc3
               IF (ic3 /= jc3) THEN
                  d(:) = sym%sec(:, jc3, 3)
                  DO i = 1, 2
                     phi1 = 2.0_dp*REAL(i, KIND=dp)*pi/3.0_dp
                     b(:) = rotate_vector(a(:), phi1, d(:))
                     CALL addsec(3, b(:), sym)
                  END DO
               END IF
            END DO
         END DO

         IF (sym%nsec(3) > nc3) THEN
            ! New C3 axes were found
            CYCLE
         ELSE
            ! In the case of I or Ih there have to be more C3 axes
            IF (sym%nsec(3) == 4) THEN
               a(:) = sym%sec(:, 1, 3)
               b(:) = sym%sec(:, 2, 3)
               phi1 = 0.5_dp*angle(a(:), b(:))
               IF (phi1 < 0.5_dp*pi) phi1 = phi1-0.5_dp*pi
               d(:) = vector_product(a(:), b(:))
               b(:) = rotate_vector(a(:), phi1, d(:))
               c(:) = sym%sec(:, 3, 3)
               phi1 = 0.5_dp*angle(a(:), c(:))
               IF (phi1 < 0.5_dp*pi) phi1 = phi1-0.5_dp*pi
               d(:) = vector_product(a(:), c(:))
               c(:) = rotate_vector(a(:), phi1, d(:))
               d(:) = vector_product(b(:), c(:))
               phi1 = 0.5_dp*phidd
               a(:) = rotate_vector(b(:), phi1, d(:))
               IF (caxis(3, a(:), sym, coord)) THEN
                  CALL addsec(3, a(:), sym)
               ELSE
                  phi2 = 0.5_dp*pi-phi1
                  a(:) = rotate_vector(b(:), phi2, d(:))
                  IF (caxis(3, a(:), sym, coord)) CALL addsec(3, a(:), sym)
               END IF

               IF (sym%nsec(3) > 4) CYCLE

            ELSE IF (sym%nsec(3) /= 10) THEN

               !  C3 axes found, but only 4 or 10 are possible
               failed = .TRUE.
               RETURN

            END IF

            ! Exit loop, if 4 or 10 C3 axes were found
            EXIT

         END IF

      END DO

      ! Loop over all pairs of C3 axes to find all C5, C4 and C2 axes
      DO isec = 1, sym%nsec(3)

         a(:) = sym%sec(:, isec, 3)
         DO jsec = isec+1, sym%nsec(3)
            phi1 = 0.5_dp*angle(a(:), sym%sec(:, jsec, 3))
            d(:) = vector_product(a(:), sym%sec(:, jsec, 3))

            ! Check for C5 axis (I,Ih)
            IF (sym%nsec(3) == 10) THEN
               b(:) = rotate_vector(a(:), phidi, d(:))
               IF (caxis(5, b(:), sym, coord)) THEN
                  CALL addsec(5, b(:), sym)
                  phi1 = phidi+phiii
                  b(:) = rotate_vector(a(:), phi1, d(:))
                  IF (caxis(5, b(:), sym, coord)) CALL addsec(5, b(:), sym)
               END IF
            END IF

            ! Check for C4 (O,Oh), C2 and S2 (center of inversion) axis
            DO i = 0, 1
               phi2 = phi1-0.5_dp*REAL(i, KIND=dp)*pi
               b(:) = rotate_vector(a(:), phi2, d(:))
               IF (caxis(2, b(:), sym, coord)) THEN
                  CALL addsec(2, b(:), sym)
                  IF (sym%nsec(3) == 4) THEN
                     IF (caxis(4, b(:), sym, coord)) CALL addsec(4, b(:), sym)
                  END IF
                  IF (saxis(2, b(:), sym, coord)) THEN
                     CALL addses(2, b(:), sym)
                     sym%invers = .TRUE.
                  END IF
               END IF
               IF (sigma(b(:), sym, coord)) CALL addsig(b(:), sym)
            END DO

            ! Check for mirror plane
            IF (sigma(d(:), sym, coord)) CALL addsig(d(:), sym)

         END DO

      END DO

      ! Set the logical variables for point group determination
      iax = sym%ncn

      IF ((sym%ncn == 5) .AND. (sym%nsec(5) > 1)) THEN
         sym%igroup = .TRUE.
      ELSE IF ((sym%ncn == 4) .AND. (sym%nsec(4) > 1)) THEN
         sym%ogroup = .TRUE.
      ELSE IF ((sym%ncn == 3) .AND. (sym%nsec(3) > 1)) THEN
         sym%tgroup = .TRUE.
         IF ((.NOT. sym%invers) .AND. (sym%nsig > 0)) sym%sigmad = .TRUE.
         iax = 2
      ELSE
         ! No main axis found
         failed = .TRUE.
         RETURN
      END IF

      ! Rotate molecule to standard orientation
      phi1 = angle(sym%sec(:, 1, iax), sym%z_axis(:))
      d(:) = vector_product(sym%sec(:, 1, iax), sym%z_axis(:))
      CALL rotate_molecule(phi1, d(:), sym, coord)

      a(:) = sym%sec(:, 2, iax)
      a(3) = 0.0_dp
      phi2 = angle(a(:), sym%x_axis(:))
      d(:) = vector_product(a(:), sym%x_axis(:))
      CALL rotate_molecule(phi2, d(:), sym, coord)

   END SUBROUTINE cubsym

! **************************************************************************************************
!> \brief The number of a equivalent atom to atoma is returned. If there
!>        is no equivalent atom, then zero is returned. The cartesian
!>        coordinates of the equivalent atom are defined by the vector a.
!> \param atoma ...
!> \param a ...
!> \param sym ...
!> \param coord ...
!> \retval equatom ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   FUNCTION equatom(atoma, a, sym, coord)
      INTEGER, INTENT(IN)                                :: atoma
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord
      INTEGER                                            :: equatom

      INTEGER                                            :: iatom, natoms

      equatom = 0
      natoms = SIZE(coord, 2)
      DO iatom = 1, natoms
         IF ((ABS(sym%ain(iatom)-sym%ain(atoma)) < TINY(0.0_dp)) .AND. &
             (ABS(a(1)-coord(1, iatom)) < sym%eps_geo) .AND. &
             (ABS(a(2)-coord(2, iatom)) < sym%eps_geo) .AND. &
             (ABS(a(3)-coord(3, iatom)) < sym%eps_geo)) THEN
            equatom = iatom
            RETURN
         END IF
      END DO

   END FUNCTION equatom

! **************************************************************************************************
!> \brief Calculate the order of the point group.
!> \param sym ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE get_point_group_order(sym)
      TYPE(molsym_type), INTENT(inout)                   :: sym

      INTEGER                                            :: icn, incr, isec, ises, isn, jcn, jsn

! Count all symmetry elements of the symmetry group
! First E and all mirror planes

      sym%point_group_order = 1+sym%nsig

      ! Loop over all C axes
      DO icn = 2, sym%ncn
         DO isec = 1, sym%nsec(icn)
            DO jcn = 1, icn-1
               IF (newse(jcn, icn)) sym%point_group_order = sym%point_group_order+1
            END DO
         END DO
      END DO

      ! Loop over all S axes
      IF (sym%sigmah) THEN
         incr = 1
      ELSE
         incr = 2
      END IF

      DO isn = 2, sym%nsn
         DO ises = 1, sym%nses(isn)
            DO jsn = 1, isn-1, incr
               IF (newse(jsn, isn)) sym%point_group_order = sym%point_group_order+1
            END DO
         END DO
      END DO

   END SUBROUTINE get_point_group_order

! **************************************************************************************************
!> \brief Get the point group symbol.
!> \param sym ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE get_point_group_symbol(sym)
      TYPE(molsym_type), INTENT(inout)                   :: sym

      CHARACTER(4)                                       :: degree

      IF (sym%linear) THEN
         IF (sym%invers) THEN
            sym%point_group_symbol = "D*h"
         ELSE
            sym%point_group_symbol = "C*v"
         END IF
      ELSE IF (sym%cubic) THEN
         IF (sym%igroup) THEN
            sym%point_group_symbol = "I"
         ELSE IF (sym%ogroup) THEN
            sym%point_group_symbol = "O"
         ELSE IF (sym%tgroup) THEN
            sym%point_group_symbol = "T"
         END IF
         IF (sym%invers) THEN
            sym%point_group_symbol = TRIM(sym%point_group_symbol)//"h"
         ELSE IF (sym%sigmad) THEN
            sym%point_group_symbol = TRIM(sym%point_group_symbol)//"d"
         END IF
      ELSE IF (sym%maxis) THEN
         IF (sym%dgroup) THEN
            WRITE (degree, "(I3)") sym%ncn
            sym%point_group_symbol = "D"//TRIM(ADJUSTL(degree))
            IF (sym%sigmah) THEN
               sym%point_group_symbol = TRIM(sym%point_group_symbol)//"h"
            ELSE IF (sym%sigmad) THEN
               sym%point_group_symbol = TRIM(sym%point_group_symbol)//"d"
            END IF
         ELSE IF (sym%sgroup) THEN
            WRITE (degree, "(I3)") sym%nsn
            sym%point_group_symbol = "S"//TRIM(ADJUSTL(degree))
         ELSE
            WRITE (degree, "(I3)") sym%ncn
            sym%point_group_symbol = "C"//TRIM(ADJUSTL(degree))
            IF (sym%sigmah) THEN
               sym%point_group_symbol = TRIM(sym%point_group_symbol)//"h"
            ELSE IF (sym%sigmav) THEN
               sym%point_group_symbol = TRIM(sym%point_group_symbol)//"v"
            END IF
         END IF
      ELSE
         IF (sym%invers) THEN
            sym%point_group_symbol = "Ci"
         ELSE IF (sym%sigmah) THEN
            sym%point_group_symbol = "Cs"
         ELSE
            sym%point_group_symbol = "C1"
         END IF
      ENDIF

   END SUBROUTINE get_point_group_symbol

! **************************************************************************************************
!> \brief Initialization of the global variables of module symmetry.
!> \param sym ...
!> \param atype ...
!> \param weight ...
!> \par History
!>        Creation (19.10.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE init_symmetry(sym, atype, weight)
      TYPE(molsym_type), INTENT(inout)                   :: sym
      INTEGER, DIMENSION(:), INTENT(IN)                  :: atype
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: weight

      CHARACTER(len=*), PARAMETER :: routineN = 'init_symmetry', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iatom, natoms

! Define the Cartesian axis vectors

      sym%x_axis = (/1.0_dp, 0.0_dp, 0.0_dp/)
      sym%y_axis = (/0.0_dp, 1.0_dp, 0.0_dp/)
      sym%z_axis = (/0.0_dp, 0.0_dp, 1.0_dp/)

      ! Initialize lists for symmetry elements
      sym%sec(:, :, :) = 0.0_dp
      sym%ses(:, :, :) = 0.0_dp
      sym%sig(:, :) = 0.0_dp

      sym%center_of_mass(:) = 0.0_dp

      ! Initialize symmetry element counters
      sym%ncn = 1
      sym%nsn = 1
      sym%nsec(:) = 0
      sym%nses(:) = 0
      sym%nsig = 0

      ! Initialize logical variables
      sym%cubic = .FALSE.
      sym%dgroup = .FALSE.
      sym%igroup = .FALSE.
      sym%invers = .FALSE.
      sym%linear = .FALSE.
      sym%maxis = .FALSE.
      sym%ogroup = .FALSE.
      sym%sgroup = .FALSE.
      sym%sigmad = .FALSE.
      sym%sigmah = .FALSE.
      sym%sigmav = .FALSE.
      sym%tgroup = .FALSE.

      ! Initialize list of equivalent atoms (C1 symmetry)
      natoms = SIZE(sym%nequatom)
      sym%ngroup = natoms
      sym%nequatom(:) = 1
      DO i = 1, sym%ngroup
         sym%group_of(i) = i
         sym%llequatom(i) = i
         sym%symequ_list(i) = i
         sym%ulequatom(i) = i
      END DO

      sym%point_group_order = -1

      DO iatom = 1, natoms
         sym%aw(iatom) = weight(iatom)
      END DO

      ! Generate atomic identification numbers (symmetry code) ***
      sym%ain(:) = REAL(atype(:), KIND=dp)+sym%aw(:)

      ! Initialize the transformation matrix for input orientation -> standard orientation
      CALL unit_matrix(sym%inptostd(:, :))

   END SUBROUTINE init_symmetry

! **************************************************************************************************
!> \brief  Return .TRUE., if the atom atoma is in the list symequ_list.
!> \param atoma ...
!> \param sym ...
!> \retval in_symequ_list ...
!> \par History
!>        Creation (21.04.95, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   FUNCTION in_symequ_list(atoma, sym)
      INTEGER, INTENT(IN)                                :: atoma
      TYPE(molsym_type), INTENT(inout)                   :: sym
      LOGICAL                                            :: in_symequ_list

      INTEGER                                            :: iatom

      in_symequ_list = .FALSE.

      DO iatom = 1, sym%ulequatom(sym%ngroup)
         IF (sym%symequ_list(iatom) == atoma) THEN
            in_symequ_list = .TRUE.
            RETURN
         END IF
      END DO

   END FUNCTION in_symequ_list

! **************************************************************************************************
!> \brief Search for point groups of LOW SYMmetry (Ci,Cs).
!> \param sym ...
!> \param coord ...
!> \par History
!>        Creation (21.04.95, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE lowsym(sym, coord)
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord

      CHARACTER(len=*), PARAMETER :: routineN = 'lowsym', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: phi
      REAL(KIND=dp), DIMENSION(3)                        :: d

      IF (sym%nsn == 2) THEN

         ! Ci
         sym%invers = .TRUE.
         phi = angle(sym%ses(:, 1, 2), sym%z_axis(:))
         d(:) = vector_product(sym%ses(:, 1, 2), sym%z_axis(:))
         CALL rotate_molecule(phi, d(:), sym, coord)

      ELSE IF (sym%nsig == 1) THEN

         ! Cs
         sym%sigmah = .TRUE.
         phi = angle(sym%sig(:, 1), sym%z_axis(:))
         d(:) = vector_product(sym%sig(:, 1), sym%z_axis(:))
         CALL rotate_molecule(phi, d(:), sym, coord)

      END IF

   END SUBROUTINE lowsym

! **************************************************************************************************
!> \brief Main program for symmetry analysis.
!> \param sym ...
!> \param eps_geo ...
!> \param coord ...
!> \param atype ...
!> \param weight ...
!> \par History
!>        Creation (14.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE molecular_symmetry(sym, eps_geo, coord, atype, weight)
      TYPE(molsym_type), POINTER                         :: sym
      REAL(KIND=dp), INTENT(IN)                          :: eps_geo
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord
      INTEGER, DIMENSION(:), INTENT(IN)                  :: atype
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: weight

      CHARACTER(len=*), PARAMETER :: routineN = 'molecular_symmetry', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, natoms

      CALL timeset(routineN, handle)

      ! Perform memory allocation for the symmetry analysis
      natoms = SIZE(coord, 2)
      CALL create_molsym(sym, natoms)
      sym%eps_geo = eps_geo

      ! Initialization of symmetry analysis
      CALL init_symmetry(sym, atype, weight)

      IF (natoms == 1) THEN
         ! Special case: only one atom
         CALL atomsym(sym)
      ELSE
         ! Find molecular symmetry
         CALL moleculesym(sym, coord)
         ! Get point group and load point group table
         CALL get_point_group_symbol(sym)
      END IF

      ! Calculate group order
      IF (.NOT. sym%linear) CALL get_point_group_order(sym)

      ! Generate a list of equivalent atoms
      CALL build_symequ_list(sym, coord)

      CALL timestop(handle)

   END SUBROUTINE molecular_symmetry

! **************************************************************************************************
!> \brief Find the molecular symmetry.
!> \param sym ...
!> \param coord ...
!> \par History
!>        Creation (14.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE moleculesym(sym, coord)
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord

      CHARACTER(len=*), PARAMETER :: routineN = 'moleculesym', routineP = moduleN//':'//routineN

      INTEGER                                            :: icn, isec, isn
      LOGICAL                                            :: failed
      REAL(KIND=dp)                                      :: eps_tenval

! Tolerance of degenerate eigenvalues for the molecular tensor of inertia

      eps_tenval = 0.01_dp*sym%eps_geo

      ! Calculate the molecular tensor of inertia
      CALL tensor(sym, coord)
      ! Use symmetry information from the eigenvalues of the molecular tensor of inertia
      IF ((sym%tenval(3)-sym%tenval(1)) < eps_tenval) THEN
         ! 0 < tenval(1) = tenval(2) = tenval(3)
         sym%cubic = .TRUE.
      ELSE IF ((sym%tenval(3)-sym%tenval(2)) < eps_tenval) THEN
         ! 0 < tenval(1) < tenval(2) = tenval(3)
         ! special case: 0 = tenval(1) < tenval(2) = tenval(3)
         IF (sym%tenval(1) < eps_tenval) sym%linear = .TRUE.
         CALL tracar(sym, coord, (/3, 1, 2/))
      ELSE
         ! 0 < tenval(1) = tenval(2) < tenval(3) or 0 < tenval(1) < tenval(2) < tenval(3)
         CALL tracar(sym, coord, (/1, 2, 3/))
      END IF

      ! Continue with the new coordinate axes
      DO
         failed = .FALSE.
         IF (sym%cubic) THEN
            CALL cubsym(sym, coord, failed)
            IF (failed) THEN
               sym%cubic = .FALSE.
               CYCLE
            END IF

         ELSE IF (sym%linear) THEN

            ! Linear molecule
            IF (saxis(2, sym%z_axis(:), sym, coord)) THEN
               sym%invers = .TRUE.
               sym%dgroup = .TRUE.
               sym%sigmah = .TRUE.
            END IF

         ELSE

            ! Check the new coordinate axes for Cn axes
            DO icn = 2, maxcn
               IF (caxis(icn, sym%z_axis(:), sym, coord)) CALL addsec(icn, sym%z_axis(:), sym)
               IF (caxis(icn, sym%x_axis(:), sym, coord)) CALL addsec(icn, sym%x_axis(:), sym)
               IF (caxis(icn, sym%y_axis(:), sym, coord)) CALL addsec(icn, sym%y_axis(:), sym)
            END DO

            ! Check the new coordinate axes for Sn axes
            DO isn = 2, maxsn
               IF (saxis(isn, sym%z_axis(:), sym, coord)) CALL addses(isn, sym%z_axis(:), sym)
               IF (saxis(isn, sym%x_axis(:), sym, coord)) CALL addses(isn, sym%x_axis(:), sym)
               IF (saxis(isn, sym%y_axis(:), sym, coord)) CALL addses(isn, sym%y_axis(:), sym)
            END DO

            ! Check the new coordinate planes for mirror planes
            IF (sigma(sym%z_axis(:), sym, coord)) CALL addsig(sym%z_axis(:), sym)
            IF (sigma(sym%x_axis(:), sym, coord)) CALL addsig(sym%x_axis(:), sym)
            IF (sigma(sym%y_axis(:), sym, coord)) CALL addsig(sym%y_axis(:), sym)

            ! There is a main axis (MAXIS = .TRUE.)
            IF ((sym%ncn > 1) .OR. (sym%nsn > 3)) THEN
               sym%maxis = .TRUE.
               CALL axsym(coord, sym)
            ELSE
               ! Only low or no symmetry (Ci, Cs or C1)
               CALL lowsym(sym, coord)
            END IF

         END IF

         IF (.NOT. failed) EXIT

      END DO

      ! Find the remaining S axes
      IF (.NOT. sym%linear) THEN
         DO icn = 2, sym%ncn
            DO isec = 1, sym%nsec(icn)
               IF (saxis(icn, sym%sec(:, isec, icn), sym, coord)) &
                  CALL addses(icn, sym%sec(:, isec, icn), sym)
               IF (saxis(2*icn, sym%sec(:, isec, icn), sym, coord)) &
                  CALL addses(2*icn, sym%sec(:, isec, icn), sym)
            END DO
         END DO
      END IF

      ! Remove redundant S2 axes
      IF (sym%nses(2) > 0) THEN
         sym%nses(1) = sym%nses(1)-sym%nses(2)
         sym%nses(2) = 0
         CALL addses(2, sym%z_axis(:), sym)
      END IF

   END SUBROUTINE moleculesym

! **************************************************************************************************
!> \brief The number of atoms which are placed on an axis defined by the vector a is returned.
!> \param a ...
!> \param coord ...
!> \param sym ...
!> \retval naxis ...
!> \par History
!>        Creation (16.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   FUNCTION naxis(a, coord, sym)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord
      TYPE(molsym_type), INTENT(inout)                   :: sym
      INTEGER                                            :: naxis

      INTEGER                                            :: iatom, natoms
      REAL(KIND=dp)                                      :: length_of_a, length_of_b, scapro
      REAL(KIND=dp), DIMENSION(3)                        :: a_norm, b, b_norm

      naxis = 0
      natoms = SIZE(coord, 2)

      length_of_a = SQRT(a(1)*a(1)+a(2)*a(2)+a(3)*a(3))

      ! Check the length of vector a
      IF (length_of_a > sym%eps_geo) THEN

         DO iatom = 1, natoms
            b(:) = coord(:, iatom)
            length_of_b = SQRT(b(1)*b(1)+b(2)*b(2)+b(3)*b(3))
            ! An atom in the origin counts for each axis
            IF (length_of_b < sym%eps_geo) THEN
               naxis = naxis+1
            ELSE
               a_norm = a(:)/length_of_a
               b_norm = b(:)/length_of_b
               scapro = a_norm(1)*b_norm(1)+a_norm(2)*b_norm(2)+a_norm(3)*b_norm(3)
               IF (ABS(ABS(scapro)-1.0_dp) < sym%eps_geo) naxis = naxis+1
            END IF
         END DO

      END IF

   END FUNCTION naxis

! **************************************************************************************************
!> \brief Return .FALSE., if the symmetry elements se1 and se2 are a pair of
!>         redundant symmetry elements.
!> \param se1 ...
!> \param se2 ...
!> \retval newse ...
!> \par History
!>        Creation (16.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   FUNCTION newse(se1, se2)
      INTEGER, INTENT(IN)                                :: se1, se2
      LOGICAL                                            :: newse

      INTEGER                                            :: ise

      newse = .TRUE.

      DO ise = 2, MIN(se1, se2)
         IF ((MODULO(se1, ise) == 0) .AND. (MODULO(se2, ise) == 0)) THEN
            newse = .FALSE.
            RETURN
         END IF
      END DO

   END FUNCTION newse

! **************************************************************************************************
!> \brief The number of atoms which are placed in a plane defined by the
!>         the normal vector a is returned.
!> \param a ...
!> \param sym ...
!> \param coord ...
!> \retval nsigma ...
!> \par History
!>        Creation (16.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   FUNCTION nsigma(a, sym, coord)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord
      INTEGER                                            :: nsigma

      INTEGER                                            :: iatom, natoms
      REAL(KIND=dp)                                      :: length_of_a, length_of_b, scapro
      REAL(KIND=dp), DIMENSION(3)                        :: a_norm, b, b_norm

      nsigma = 0

      length_of_a = SQRT(a(1)*a(1)+a(2)*a(2)+a(3)*a(3))

      ! Check the length of vector a
      IF (length_of_a > sym%eps_geo) THEN
         natoms = SIZE(coord, 2)
         DO iatom = 1, natoms
            b(:) = coord(:, iatom)
            length_of_b = SQRT(b(1)*b(1)+b(2)*b(2)+b(3)*b(3))
            ! An atom in the origin counts for each mirror plane
            IF (length_of_b < sym%eps_geo) THEN
               nsigma = nsigma+1
            ELSE
               a_norm = a(:)/length_of_a
               b_norm = b(:)/length_of_b
               scapro = a_norm(1)*b_norm(1)+a_norm(2)*b_norm(2)+a_norm(3)*b_norm(3)
               IF (ABS(scapro) < sym%eps_geo) nsigma = nsigma+1
            END IF
         END DO
      END IF

   END FUNCTION nsigma

! **************************************************************************************************
!> \brief Style the output of the symmetry elements.
!> \param se ...
!> \param eps ...
!> \par History
!>        Creation (16.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE outse(se, eps)
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: se
      REAL(KIND=dp), INTENT(IN)                          :: eps

! Positive z component for all vectors

      IF (ABS(se(3)) < eps) THEN
         IF (ABS(se(1)) < eps) THEN
            se(2) = -se(2)
         ELSE
            IF (se(1) < 0.0_dp) se(1:2) = -se(1:2)
         END IF
      ELSE
         IF (se(3) < 0.0_dp) se(:) = -se(:)
      END IF

   END SUBROUTINE outse

! **************************************************************************************************
!> \brief Print the output of the symmetry analysis.
!> \param sym ...
!> \param coord ...
!> \param atype ...
!> \param element ...
!> \param z ...
!> \param weight ...
!> \param iw ...
!> \param plevel ...
!> \par History
!>        Creation (16.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE print_symmetry(sym, coord, atype, element, z, weight, iw, plevel)
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(in)         :: coord
      INTEGER, DIMENSION(:), INTENT(in)                  :: atype
      CHARACTER(LEN=*), DIMENSION(:), INTENT(in)         :: element
      INTEGER, DIMENSION(:), INTENT(in)                  :: z
      REAL(KIND=dp), DIMENSION(:), INTENT(in)            :: weight
      INTEGER, INTENT(IN)                                :: iw, plevel

      CHARACTER(len=*), PARAMETER :: routineN = 'print_symmetry', routineP = moduleN//':'//routineN

      CHARACTER(3)                                       :: string
      INTEGER                                            :: i, icn, iequatom, isec, ises, isig, isn, &
                                                            j, secount
      REAL(KIND=dp)                                      :: null

! Print point group symbol and point group order

      IF (sym%point_group_order == -1) THEN
         WRITE (iw, "(/,T2,A,T40,A,T77,A4)") "Molecular Symmetry Information", &
            "Point Group: ", ADJUSTR(sym%point_group_symbol)
      ELSE
         WRITE (iw, "(/,T2,A,T40,A,T77,A4,/,T40,A,T77,I4)") &
            "Molecular Symmetry Information", &
            "Point Group: ", ADJUSTR(sym%point_group_symbol), &
            "Group Order: ", sym%point_group_order
      END IF

      IF (MOD(plevel, 10) == 1) THEN
         ! Print the Cartesian coordinates of the standard orientation
         CALL write_coordinates(coord, atype, element, z, weight, iw)

         WRITE (iw, "(/,T3,A,(T41,10I4))") "Group Number:      1   Group Members:", &
            (sym%symequ_list(iequatom), iequatom=sym%llequatom(1), sym%ulequatom(1))
         DO i = 2, sym%ngroup
            WRITE (iw, "(T15,I8,(T41,10I4))") &
               i, (sym%symequ_list(iequatom), iequatom=sym%llequatom(i), sym%ulequatom(i))
         END DO
      END IF

      IF (MOD(plevel/100, 10) == 1) THEN
         !   *** Print all symmetry elements ***
         WRITE (iw, "(/,T24,A,9X,A,12X,A,12X,A)") "Symmetry Elements", "X", "Y", "Z"

         secount = 1
         null = 0._dp
         WRITE (iw, "(T24,2I5,2X,A1,5x,3F13.6)") secount, secount, "E", null, null, null

         string = "@  "
         DO isig = 1, sym%nsig
            secount = secount+1
            CALL outse(sym%sig(:, isig), sym%eps_geo)
            WRITE (iw, "(T24,2I5,2X,A3,3X,3F13.6)") &
               secount, isig, string, (sym%sig(i, isig), i=1, 3)
         END DO

         DO icn = 2, sym%ncn
            IF (icn < 10) THEN
               WRITE (string, "(A1,I1)") "C", icn
            ELSE
               WRITE (string, "(A1,I2)") "C", icn
            END IF
            DO isec = 1, sym%nsec(icn)
               secount = secount+1
               CALL outse(sym%sec(:, isec, icn), sym%eps_geo)
               WRITE (iw, "(T24,2I5,2X,A3,3X,3F13.6)") &
                  secount, isec, string, (sym%sec(i, isec, icn), i=1, 3)
            END DO
         END DO

         DO isn = 2, sym%nsn
            IF (isn == 2) THEN
               WRITE (string, "(A3)") "i  "
            ELSE IF (isn < 10) THEN
               WRITE (string, "(A1,I1)") "S", isn
            ELSE
               WRITE (string, "(A1,I2)") "S", isn
            END IF
            DO ises = 1, sym%nses(isn)
               secount = secount+1
               CALL outse(sym%ses(:, ises, isn), sym%eps_geo)
               WRITE (iw, "(T24,2I5,2X,A3,3X,3F13.6)") &
                  secount, ises, string, (sym%ses(i, ises, icn), i=1, 3)
            END DO
         END DO
      END IF

      IF (MOD(plevel/10, 10) == 1 .AND. SIZE(coord, 2) > 1) THEN
         ! Print the moments of the molecular inertia tensor
         WRITE (iw, "(/,T24,A,/,T40,A,2(16X,A),3(/,T25,A,3F18.5))") &
            "Moments of the Molecular Inertia Tensor", "Ix", "Iy", "Iz", &
            "Ix", (sym%tenmat(1, i), i=1, 3), &
            "Iy", (sym%tenmat(2, i), i=1, 3), &
            "Iz", (sym%tenmat(3, i), i=1, 3)
         WRITE (iw, "(/,T24,A,/,T27,3F18.5,/,(T27,3F18.5))") &
            "Principal Moments and Axes of Inertia", (sym%tenval(i), i=1, 3), &
            ((sym%tenvec(i, j), j=1, 3), i=1, 3)
      END IF

   END SUBROUTINE print_symmetry

! **************************************************************************************************
!> \brief Rotate the molecule about an axis defined by the vector a. The
!>        rotation angle is phi (radians).
!> \param phi ...
!> \param a ...
!> \param sym ...
!> \param coord ...
!> \par History
!>        Creation (16.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE rotate_molecule(phi, a, sym, coord)
      REAL(KIND=dp), INTENT(IN)                          :: phi
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord

      CHARACTER(len=*), PARAMETER :: routineN = 'rotate_molecule', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: length_of_a

! Check the length of vector a

      length_of_a = SQRT(a(1)*a(1)+a(2)*a(2)+a(3)*a(3))
      IF (length_of_a > sym%eps_geo) THEN

         ! Build up the rotation matrix
         CALL build_rotmat(phi, a(:), sym%rotmat(:, :))

         ! Rotate the molecule by phi around a
         coord(:, :) = MATMUL(sym%rotmat(:, :), coord(:, :))

         ! Rotate the normal vectors of the mirror planes which are already found
         sym%sig(:, 1:sym%nsig) = MATMUL(sym%rotmat(:, :), sym%sig(:, 1:sym%nsig))

         ! Rotate the Cn axes which are already found
         DO i = 2, sym%ncn
            sym%sec(:, 1:sym%nsec(i), i) = MATMUL(sym%rotmat(:, :), sym%sec(:, 1:sym%nsec(i), i))
         END DO

         ! Rotate the Sn axes which are already found
         DO i = 2, sym%nsn
            sym%ses(:, 1:sym%nses(i), i) = MATMUL(sym%rotmat(:, :), sym%ses(:, 1:sym%nses(i), i))
         END DO

         ! Store current rotation
         sym%inptostd(:, :) = MATMUL(sym%rotmat(:, :), sym%inptostd(:, :))

      END IF

   END SUBROUTINE rotate_molecule

! **************************************************************************************************
!> \brief Rotate the molecule about a n-fold axis defined by the vector a
!>        using a rotation angle of 2*pi/n. Check, if the axis is a Cn axis.
!> \param n ...
!> \param a ...
!> \param sym ...
!> \param coord ...
!> \retval saxis ...
!> \par History
!>        Creation (16.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   FUNCTION saxis(n, a, sym, coord)
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord
      LOGICAL                                            :: saxis

      INTEGER                                            :: iatom, natoms
      REAL(KIND=dp)                                      :: length_of_a, phi
      REAL(KIND=dp), DIMENSION(3)                        :: b

      saxis = .FALSE.

      length_of_a = SQRT(a(1)*a(1)+a(2)*a(2)+a(3)*a(3))

      natoms = SIZE(coord, 2)

      ! Check the length of the axis vector a
      IF (length_of_a > sym%eps_geo) THEN
         ! Calculate the rotation angle phi and build up the rotation matrix rotmat
         phi = 2.0_dp*pi/REAL(n, KIND=dp)
         CALL build_rotmat(phi, a(:), sym%rotmat(:, :))
         ! Check all atoms
         DO iatom = 1, natoms
            ! Rotate the actual atom by 2*pi/n about a
            b(:) = MATMUL(sym%rotmat(:, :), coord(:, iatom))
            ! Reflect the coordinates of the rotated atom
            b(:) = reflect_vector(b(:), a(:))
            ! Check if the coordinates of the rotated atom are in the coordinate set of the molecule
            IF (equatom(iatom, b(:), sym, coord) == 0) RETURN
         END DO
         ! The axis defined by a is a Sn axis, thus return with saxis = .TRUE.
         saxis = .TRUE.
      END IF

   END FUNCTION saxis

! **************************************************************************************************
!> \brief Reflect all atoms of the molecule through a mirror plane defined
!>        by the normal vector a.
!> \param a ...
!> \param sym ...
!> \param coord ...
!> \retval sigma ...
!> \par History
!>        Creation (16.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   FUNCTION sigma(a, sym, coord)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord
      LOGICAL                                            :: sigma

      INTEGER                                            :: iatom, natoms
      REAL(KIND=dp)                                      :: length_of_a
      REAL(KIND=dp), DIMENSION(3)                        :: b

      sigma = .FALSE.

      length_of_a = SQRT(a(1)*a(1)+a(2)*a(2)+a(3)*a(3))

      ! Check the length of vector a
      IF (length_of_a > sym%eps_geo) THEN
         natoms = SIZE(coord, 2)
         DO iatom = 1, natoms
            ! Reflect the actual atom
            b(:) = reflect_vector(coord(:, iatom), a(:))
            ! Check if the coordinates of the reflected atom are in the coordinate set of the molecule
            IF (equatom(iatom, b(:), sym, coord) == 0) RETURN
         END DO
         ! The plane defined by a is a mirror plane, thus return with sigma = .TRUE.
         sigma = .TRUE.
      END IF

   END FUNCTION sigma

! **************************************************************************************************
!> \brief Calculate the molecular tensor of inertia and the vector to the
!>        center of mass of the molecule.
!> \param sym ...
!> \param coord ...
!> \par History
!>        Creation (16.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE tensor(sym, coord)
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord

      CHARACTER(len=*), PARAMETER :: routineN = 'tensor', routineP = moduleN//':'//routineN

      INTEGER                                            :: natoms
      REAL(KIND=dp)                                      :: tt

! Find the vector center_of_mass to molecular center of mass

      natoms = SIZE(coord, 2)
      sym%center_of_mass(:) = MATMUL(coord(:, 1:natoms), sym%aw(1:natoms))/SUM(sym%aw(1:natoms))

      ! Translate the center of mass of the molecule to the origin
      coord(:, 1:natoms) = coord(:, 1:natoms)-SPREAD(sym%center_of_mass(:), 2, natoms)

      ! Build up the molecular tensor of inertia

      sym%tenmat(1, 1) = DOT_PRODUCT(sym%aw(1:natoms), (coord(2, 1:natoms)**2+coord(3, 1:natoms)**2))
      sym%tenmat(2, 2) = DOT_PRODUCT(sym%aw(1:natoms), (coord(3, 1:natoms)**2+coord(1, 1:natoms)**2))
      sym%tenmat(3, 3) = DOT_PRODUCT(sym%aw(1:natoms), (coord(1, 1:natoms)**2+coord(2, 1:natoms)**2))

      sym%tenmat(1, 2) = -DOT_PRODUCT(sym%aw(1:natoms), (coord(1, 1:natoms)*coord(2, 1:natoms)))
      sym%tenmat(1, 3) = -DOT_PRODUCT(sym%aw(1:natoms), (coord(1, 1:natoms)*coord(3, 1:natoms)))
      sym%tenmat(2, 3) = -DOT_PRODUCT(sym%aw(1:natoms), (coord(2, 1:natoms)*coord(3, 1:natoms)))

      ! Symmetrize tensor matrix
      sym%tenmat(2, 1) = sym%tenmat(1, 2)
      sym%tenmat(3, 1) = sym%tenmat(1, 3)
      sym%tenmat(3, 2) = sym%tenmat(2, 3)

      ! Diagonalize the molecular tensor of inertia
      CALL jacobi(sym%tenmat(:, :), sym%tenval(:), sym%tenvec(:, :))

      ! Secure that the principal axes are right-handed
      sym%tenvec(:, 3) = vector_product(sym%tenvec(:, 1), sym%tenvec(:, 2))

      tt = SQRT(sym%tenval(1)**2+sym%tenval(2)**2+sym%tenval(3)**2)
      IF(.NOT.(tt /= 0))CALL cp__a("molsym.F",1706)

   END SUBROUTINE tensor

! **************************************************************************************************
!> \brief Transformation the Cartesian coodinates with the matrix tenvec.
!>        The coordinate axes can be switched according to the index
!>        vector idx. If idx(1) is equal to 3 instead to 1, then the first
!>        eigenvector (smallest eigenvalue) of the molecular tensor of
!>        inertia is connected to the z axis instead of the x axis. In
!>        addition the local atomic coordinate systems are initialized,
!>        if the symmetry is turned off.
!> \param sym ...
!> \param coord ...
!> \param idx ...
!> \par History
!>        Creation (16.11.98, Matthias Krack)
!> \author jgh
! **************************************************************************************************
   SUBROUTINE tracar(sym, coord, idx)
      TYPE(molsym_type), INTENT(inout)                   :: sym
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: coord
      INTEGER, DIMENSION(3), INTENT(IN)                  :: idx

! Transformation of the atomic coordinates  ***

      coord(idx, :) = MATMUL(TRANSPOSE(sym%tenvec(:, :)), coord(:, :))

      ! Modify the transformation matrix for input orientation -> standard orientation
      sym%inptostd(idx, :) = TRANSPOSE(sym%tenvec(:, :))

   END SUBROUTINE tracar

! **************************************************************************************************
!> \brief Distance between two points
!> \param a ...
!> \param b ...
!> \retval d ...
!> \author jgh
! **************************************************************************************************
   FUNCTION dist(a, b) RESULT(d)
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: a, b
      REAL(KIND=dp)                                      :: d

      d = SQRT(SUM((a-b)**2))

   END FUNCTION
! **************************************************************************************************
!> \brief   Write atomic coordinates to output unit iw.
!> \param coord ...
!> \param atype ...
!> \param element ...
!> \param z ...
!> \param weight ...
!> \param iw ...
!> \date    08.05.2008
!> \author  JGH
! **************************************************************************************************
   SUBROUTINE write_coordinates(coord, atype, element, z, weight, iw)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(in)         :: coord
      INTEGER, DIMENSION(:), INTENT(in)                  :: atype
      CHARACTER(LEN=*), DIMENSION(:), INTENT(in)         :: element
      INTEGER, DIMENSION(:), INTENT(in)                  :: z
      REAL(KIND=dp), DIMENSION(:), INTENT(in)            :: weight
      INTEGER, INTENT(IN)                                :: iw

      INTEGER                                            :: iatom, natom

      IF (iw > 0) THEN

         WRITE (UNIT=iw, FMT="(/,T2,A)") &
            "Atomic Coordinates [Bohr] in Standard Orientation"
         WRITE (UNIT=iw, &
                FMT="(/,T3,A,7X,2(A1,11X),A1,6X,A10,4X,A4,/)") &
            "Atom  Kind  Element", "X", "Y", "Z", "Atomic-No.", "Mass"

         natom = SIZE(coord, 2)
         DO iatom = 1, natom
            WRITE (UNIT=iw, FMT="(T2,I5,1X,I4,3X,A2,5X,3F12.6,4X,I4,2X,F11.4)") &
               iatom, atype(iatom), element(iatom), coord(1:3, iatom), z(iatom), weight(iatom)
         END DO

      END IF

   END SUBROUTINE write_coordinates
! **************************************************************************************************

END MODULE molsym
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Work with symmetry
!> \par History
!> \author jgh
! **************************************************************************************************
MODULE cp_symmetry
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE k290,                            ONLY: csym_type,&
                                              kp_sym_gen,&
                                              pgrd,&
                                              pgrp,&
                                              print_crys_symmetry,&
                                              release_csym_type
   USE kinds,                           ONLY: dp
   USE molsym,                          ONLY: molecular_symmetry,&
                                              molsym_type,&
                                              print_symmetry,&
                                              release_molsym
   USE particle_types,                  ONLY: particle_type
   USE physcon,                         ONLY: massunit
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Global parameters (in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_symmetry'

   PUBLIC :: write_symmetry

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief Write symmetry information to output
!> \param particle_set  Atom coordinates and types
!> \param cell          Cell information
!> \param input_section Input
!> \par History
!> \author jgh
! **************************************************************************************************
   SUBROUTINE write_symmetry(particle_set, cell, input_section)
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: input_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_symmetry', routineP = moduleN//':'//routineN

      CHARACTER(LEN=2), ALLOCATABLE, DIMENSION(:)        :: element
      CHARACTER(LEN=8)                                   :: csym1, csym2, csymm, esymm
      INTEGER                                            :: handle, i, iw, natom, plevel
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atype, z
      LOGICAL                                            :: check, molecular, pall, pcoor, pinertia, &
                                                            prmat, psymmele
      REAL(KIND=dp)                                      :: eps_geo
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: weight
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: coord
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(csym_type), POINTER                           :: crys_sym
      TYPE(molsym_type), POINTER                         :: mol_sym
      TYPE(section_vals_type), POINTER                   :: section

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      NULLIFY (section)

      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger=logger, &
                                basis_section=input_section, &
                                print_key_path="PRINT%SYMMETRY", &
                                extension=".symLog")

      IF (iw > 0) THEN
         section => section_vals_get_subs_vals(section_vals=input_section, &
                                               subsection_name="PRINT%SYMMETRY")
         CALL section_vals_val_get(section_vals=section, &
                                   keyword_name="MOLECULE", l_val=molecular)
         CALL section_vals_val_get(section_vals=section, &
                                   keyword_name="EPS_GEO", r_val=eps_geo)
         IF (molecular) THEN

            NULLIFY (mol_sym)

            natom = SIZE(particle_set)
            ALLOCATE (coord(3, natom), z(natom), weight(natom), atype(natom), element(natom))

            DO i = 1, natom
               CALL get_atomic_kind(particle_set(i)%atomic_kind, z=z(i))
               CALL get_atomic_kind(atomic_kind=particle_set(i)%atomic_kind, &
                                    kind_number=atype(i), element_symbol=element(i), mass=weight(i))
               coord(1:3, i) = particle_set(i)%r(1:3)
            END DO
            weight(:) = weight(:)/massunit

            CALL molecular_symmetry(mol_sym, eps_geo, coord, atype, weight)

            CALL section_vals_val_get(section_vals=section, &
                                      keyword_name="STANDARD_ORIENTATION", l_val=pcoor)
            CALL section_vals_val_get(section_vals=section, &
                                      keyword_name="INERTIA", l_val=pinertia)
            CALL section_vals_val_get(section_vals=section, &
                                      keyword_name="SYMMETRY_ELEMENTS", l_val=psymmele)
            CALL section_vals_val_get(section_vals=section, &
                                      keyword_name="ALL", l_val=pall)
            plevel = 0
            IF (pcoor) plevel = plevel+1
            IF (pinertia) plevel = plevel+10
            IF (psymmele) plevel = plevel+100
            IF (pall) plevel = 1111111111

            CALL print_symmetry(mol_sym, coord, atype, element, z, weight, iw, plevel)

            CALL section_vals_val_get(section_vals=section, &
                                      keyword_name="CHECK_SYMMETRY", c_val=esymm)
            CALL uppercase(esymm)
            IF (TRIM(esymm) /= "NONE") THEN
               csymm = mol_sym%point_group_symbol
               CALL uppercase(csymm)
               check = TRIM(ADJUSTL(csymm)) == TRIM(ADJUSTL(esymm))
               IF(.NOT.(check))CALL cp__a("cp_symmetry.F",138)
            END IF

            DEALLOCATE (coord, z, weight, atype, element)

            CALL release_molsym(mol_sym)

         ELSE
            ! Crystal symmetry

            NULLIFY (crys_sym)

            natom = SIZE(particle_set)
            ALLOCATE (coord(3, natom), z(natom), weight(natom), atype(natom), element(natom))

            DO i = 1, natom
               CALL get_atomic_kind(particle_set(i)%atomic_kind, z=z(i))
               CALL get_atomic_kind(atomic_kind=particle_set(i)%atomic_kind, &
                                    kind_number=atype(i), element_symbol=element(i), mass=weight(i))
               coord(1:3, i) = particle_set(i)%r(1:3)
            END DO
            weight(:) = weight(:)/massunit

            CALL kp_sym_gen(crys_sym, coord, atype, cell%hmat, delta=eps_geo, iounit=iw)

            CALL section_vals_val_get(section_vals=section, &
                                      keyword_name="ROTATION_MATRICES", l_val=prmat)
            CALL section_vals_val_get(section_vals=section, &
                                      keyword_name="ALL", l_val=pall)
            plevel = 0
            IF (prmat) plevel = plevel+1
            IF (pall) plevel = 1111111111
            crys_sym%plevel = plevel

            CALL print_crys_symmetry(crys_sym)

            CALL section_vals_val_get(section_vals=section, &
                                      keyword_name="CHECK_SYMMETRY", c_val=esymm)
            CALL uppercase(esymm)
            IF (TRIM(esymm) /= "NONE") THEN
               csym1 = pgrp(crys_sym%indpg)
               CALL uppercase(csym1)
               csym2 = pgrd(crys_sym%indpg)
               CALL uppercase(csym2)
               check = (TRIM(ADJUSTL(csym1)) == TRIM(ADJUSTL(esymm))) .OR. (TRIM(ADJUSTL(csym2)) == TRIM(ADJUSTL(esymm)))
               IF(.NOT.(check))CALL cp__a("cp_symmetry.F",183)
            END IF

            CALL release_csym_type(crys_sym)

         END IF

      END IF
      CALL cp_print_key_finished_output(iw, logger, input_section, "PRINT%SYMMETRY")

      CALL timestop(handle)

   END SUBROUTINE write_symmetry

! **************************************************************************************************

END MODULE cp_symmetry
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE kg_min_heap
   USE kinds,                           ONLY: int_4,&
                                              int_8

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: heap_t, keyt, valt
   PUBLIC :: heap_pop, heap_reset_node, heap_fill
   PUBLIC :: heap_new, heap_release

   ! Sets the types
   INTEGER, PARAMETER         :: keyt = int_4
   INTEGER, PARAMETER         :: valt = int_8

   TYPE heap_node
      INTEGER(KIND=keyt) :: key
      INTEGER(KIND=valt) :: value
   END TYPE heap_node

   TYPE heap_node_e
      TYPE(heap_node) :: node
   END TYPE heap_node_e

   TYPE heap_t
      INTEGER :: n
      INTEGER, DIMENSION(:), POINTER           :: index
      TYPE(heap_node_e), DIMENSION(:), POINTER :: nodes
   END TYPE heap_t

   INTERFACE node_set
      MODULE PROCEDURE node_set_node, node_set_heap
   END INTERFACE

   INTERFACE node_get
      MODULE PROCEDURE node_get_node, node_get_heap
   END INTERFACE

CONTAINS

   ! Lookup functions

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \retval parent ...
! **************************************************************************************************
   ELEMENTAL FUNCTION get_parent(n) RESULT(parent)
      INTEGER, INTENT(IN)                                :: n
      INTEGER                                            :: parent

      parent = INT(n/2)
   END FUNCTION get_parent

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \retval child ...
! **************************************************************************************************
   ELEMENTAL FUNCTION get_left_child(n) RESULT(child)
      INTEGER, INTENT(IN)                                :: n
      INTEGER                                            :: child

      child = 2*n
   END FUNCTION get_left_child

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \retval child ...
! **************************************************************************************************
   ELEMENTAL FUNCTION get_right_child(n) RESULT(child)
      INTEGER, INTENT(IN)                                :: n
      INTEGER                                            :: child

      child = 2*n+1
   END FUNCTION get_right_child

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param n ...
!> \retval value ...
! **************************************************************************************************
   ELEMENTAL FUNCTION get_value(heap, n) RESULT(value)
      TYPE(heap_t), INTENT(IN)                           :: heap
      INTEGER, INTENT(IN)                                :: n
      INTEGER(KIND=valt)                                 :: value

      value = heap%nodes(n)%node%value
   END FUNCTION get_value

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param key ...
!> \retval value ...
! **************************************************************************************************
   ELEMENTAL FUNCTION get_value_by_key(heap, key) RESULT(value)
      TYPE(heap_t), INTENT(IN)                           :: heap
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER(KIND=valt)                                 :: value

      INTEGER                                            :: n

      n = heap%index(key)
      value = get_value(heap, n)
   END FUNCTION get_value_by_key

   ! Initialization functions

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE heap_new(heap, n)
      TYPE(heap_t), INTENT(OUT)                          :: heap
      INTEGER, INTENT(IN)                                :: n

      heap%n = n
      ALLOCATE (heap%index(n))
      ALLOCATE (heap%nodes(n))
   END SUBROUTINE heap_new

! **************************************************************************************************
!> \brief ...
!> \param heap ...
! **************************************************************************************************
   SUBROUTINE heap_release(heap)
      TYPE(heap_t), INTENT(INOUT)                        :: heap

      DEALLOCATE (heap%index)
      DEALLOCATE (heap%nodes)
      heap%n = 0
   END SUBROUTINE heap_release

!> \brief Fill heap with given values
!> \param heap ...
!> \param values ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE heap_fill(heap, values, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER(KIND=valt), DIMENSION(:), INTENT(IN)       :: values
      LOGICAL, INTENT(OUT)                               :: error

      INTEGER                                            :: first, i, n

      error = .FALSE.
      n = SIZE(values)
      IF (heap%n .LT. n) THEN
         error = .TRUE.
      ELSE
         DO i = 1, n
            !   call node_set(heap, i, i, values(i))
            heap%index(i) = i
            heap%nodes(i)%node%key = i
            heap%nodes(i)%node%value = values(i)
         END DO
         ! Sort from the last full subtree
         first = get_parent(n)
         DO i = first, 1, -1
            CALL bubble_down(heap, i, error)
         ENDDO
      ENDIF

   END SUBROUTINE heap_fill

! **************************************************************************************************
!> \brief Returns the first heap element without removing it.
!> \param heap ...
!> \param key ...
!> \param value ...
!> \param found ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE heap_get_first(heap, key, value, found, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER(KIND=keyt), INTENT(OUT)                    :: key
      INTEGER(KIND=valt), INTENT(OUT)                    :: value
      LOGICAL, INTENT(OUT)                               :: found, error

!

      error = .FALSE.
      IF (heap%n .LT. 1) THEN
         found = .FALSE.
      ELSE
         found = .TRUE.
         CALL node_get(heap, 1, key, value)
      ENDIF
   END SUBROUTINE heap_get_first

! **************************************************************************************************
!> \brief Returns and removes the first heap element and rebalances
!>        the heap.
!> \param heap ...
!> \param key ...
!> \param value ...
!> \param found ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE heap_pop(heap, key, value, found, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER(KIND=keyt), INTENT(OUT)                    :: key
      INTEGER(KIND=valt), INTENT(OUT)                    :: value
      LOGICAL, INTENT(OUT)                               :: found, error

!

      CALL heap_get_first(heap, key, value, found, error)
      IF (.NOT. error .AND. found) THEN
         IF (heap%n .GT. 1) THEN
            CALL heap_copy_node(heap, 1, heap%n)
            heap%n = heap%n-1
            CALL bubble_down(heap, 1, error)
         ELSE
            heap%n = heap%n-1
         ENDIF
      ENDIF
   END SUBROUTINE heap_pop

! **************************************************************************************************
!> \brief Changes the value of the heap element with given key and
!>        rebalances the heap.
!> \param heap ...
!> \param key ...
!> \param value ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE heap_reset_node(heap, key, value, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER(KIND=valt), INTENT(IN)                     :: value
      LOGICAL, INTENT(OUT)                               :: error

      INTEGER                                            :: n, new_pos

!

      error = .FALSE.
      IF (heap%n .LT. 1) THEN
         error = .TRUE.
      ELSE
         n = heap%index(key)
         heap%nodes(n)%node%value = value
         CALL bubble_up(heap, n, new_pos, error)
         CALL bubble_down(heap, new_pos, error)
      ENDIF
   END SUBROUTINE heap_reset_node

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param e1 ...
!> \param e2 ...
! **************************************************************************************************
   PURE SUBROUTINE heap_swap(heap, e1, e2)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER, INTENT(IN)                                :: e1, e2

      INTEGER(KIND=keyt)                                 :: key1, key2
      TYPE(heap_node)                                    :: tmp_node

!
!node_get_heap (heap, e1, key1, value)
!node_get_heap (heap, e2, key2, value)

      key1 = heap%nodes(e1)%node%key
      key2 = heap%nodes(e2)%node%key

      tmp_node = heap%nodes(e1)%node
      heap%nodes(e1)%node = heap%nodes(e2)%node
      heap%nodes(e2)%node = tmp_node

      heap%index(key1) = e2
      heap%index(key2) = e1
   END SUBROUTINE heap_swap

! **************************************************************************************************
!> \brief Sets node e1 to e2
!> \param heap ...
!> \param e1 ...
!> \param e2 ...
! **************************************************************************************************
   PURE SUBROUTINE heap_copy_node(heap, e1, e2)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER, INTENT(IN)                                :: e1, e2

      INTEGER(KIND=keyt)                                 :: key1, key2

!
!node_get_heap (heap, e1, key1, value)
!node_get_heap (heap, e2, key2, value)

      key1 = heap%nodes(e1)%node%key
      key2 = heap%nodes(e2)%node%key

      heap%nodes(e1)%node = heap%nodes(e2)%node

      heap%index(key1) = 0
      heap%index(key2) = e1
   END SUBROUTINE heap_copy_node

! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   PURE SUBROUTINE node_get_node(node, key, value)
      TYPE(heap_node), INTENT(IN)                        :: node
      INTEGER(KIND=keyt), INTENT(OUT)                    :: key
      INTEGER(KIND=valt), INTENT(OUT)                    :: value

!

      key = node%key
      value = node%value
   END SUBROUTINE node_get_node
! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param n ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   PURE SUBROUTINE node_get_heap(heap, n, key, value)
      TYPE(heap_t), INTENT(IN)                           :: heap
      INTEGER, INTENT(IN)                                :: n
      INTEGER(KIND=keyt), INTENT(OUT)                    :: key
      INTEGER(KIND=valt), INTENT(OUT)                    :: value

!

      CALL node_get_node(heap%nodes(n)%node, key, value)
   END SUBROUTINE node_get_heap
! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   PURE SUBROUTINE node_set_node(node, key, value)
      TYPE(heap_node), INTENT(OUT)                       :: node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER(KIND=valt), INTENT(IN)                     :: value

!

      node%key = key
      node%value = value
   END SUBROUTINE node_set_node
! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param n ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   PURE SUBROUTINE node_set_heap(heap, n, key, value)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER, INTENT(IN)                                :: n
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER(KIND=valt), INTENT(IN)                     :: value

!

      CALL node_set_node(heap%nodes(n)%node, key, value)
      heap%index(key) = n
   END SUBROUTINE node_set_heap

! **************************************************************************************************
!> \brief Balances a heap by bubbling down from the given element.
!> \param heap ...
!> \param first ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE bubble_down(heap, first, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER, INTENT(IN)                                :: first
      LOGICAL, INTENT(OUT)                               :: error

      INTEGER                                            :: e, left_child, right_child, smallest
      INTEGER(kind=valt)                                 :: left_child_value, min_value, &
                                                            right_child_value
      LOGICAL                                            :: all_done

!

      IF (first .GT. heap%n .OR. first .LT. 1) THEN
         error = .TRUE.
         RETURN
      ENDIF
      error = .FALSE.
      e = first
      all_done = .FALSE.
      ! Check whether we are finished, i.e,. whether the element to
      ! bubble down is childless.
      DO WHILE (e .LE. get_parent(heap%n) .AND. .NOT. all_done)
         ! Determines which node (current, left, or right child) has the
         ! smallest value.
         smallest = e
         min_value = get_value(heap, e)
         left_child = get_left_child(e)
         IF (left_child .LE. heap%n) THEN
            left_child_value = get_value(heap, left_child)
            IF (left_child_value .LT. min_value) THEN
               min_value = left_child_value
               smallest = left_child
            ENDIF
         ENDIF
         right_child = left_child+1
         IF (right_child .LE. heap%n) THEN
            right_child_value = get_value(heap, right_child)
            IF (right_child_value .LT. min_value) THEN
               min_value = right_child_value
               smallest = right_child
            ENDIF
         ENDIF
         !
         CALL heap_swap(heap, e, smallest)
         IF (smallest .EQ. e) THEN
            all_done = .TRUE.
         ELSE
            e = smallest
         ENDIF
      ENDDO
   END SUBROUTINE bubble_down

! **************************************************************************************************
!> \brief Balances a heap by bubbling up from the given element.
!> \param heap ...
!> \param first ...
!> \param new_pos ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE bubble_up(heap, first, new_pos, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER, INTENT(IN)                                :: first
      INTEGER, INTENT(OUT)                               :: new_pos
      LOGICAL, INTENT(OUT)                               :: error

      INTEGER                                            :: e, parent
      INTEGER(kind=valt)                                 :: my_value, parent_value
      LOGICAL                                            :: all_done

!

      IF (first .GT. heap%n .OR. first .LT. 1) THEN
         error = .TRUE.
         RETURN
      ENDIF
      error = .FALSE.
      e = first
      all_done = .FALSE.
      IF (e .GT. 1) THEN
         my_value = get_value(heap, e)
      ENDIF
      ! Check whether we are finished, i.e,. whether the element to
      ! bubble up is an orphan.
      new_pos = e
      DO WHILE (e .GT. 1 .AND. .NOT. all_done)
         ! Switches the parent and the current element if the current
         ! element's value is greater than the parent's value.
         parent = get_parent(e)
         parent_value = get_value(heap, parent)
         IF (my_value .LT. parent_value) THEN
            CALL heap_swap(heap, e, parent)
            e = parent
         ELSE
            all_done = .TRUE.
         ENDIF
      ENDDO
      new_pos = e
   END SUBROUTINE bubble_up

END MODULE kg_min_heap
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for a Kim-Gordon-like partitioning into molecular subunits
!>        unsing a vertex coloring algorithm (DSATUR) to find non-interating
!>        subsets, such that two molecules within the same subset have
!>        small/zero overlap (in other words: this molecular pair is not present in
!>        the neighborlist sab_orb for the current value of EPS_DEFAULT)
!> \par History
!>       2012.07 created [Martin Haeufel]
!>         2013.11 Added pair switching and revised Dsatur [Samuel Andermatt]
!> \author Martin Haeufel
! **************************************************************************************************
MODULE kg_vertex_coloring_methods
   USE bibliography,                    ONLY: Brelaz1979,&
                                              cite_reference
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE input_constants,                 ONLY: kg_color_dsatur,&
                                              kg_color_greedy
   USE kg_environment_types,            ONLY: kg_environment_type
   USE kg_min_heap,                     ONLY: heap_fill,&
                                              heap_new,&
                                              heap_pop,&
                                              heap_release,&
                                              heap_reset_node,&
                                              heap_t,&
                                              keyt,&
                                              valt
   USE kinds,                           ONLY: int_4,&
                                              int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   TYPE vertex
      INTEGER :: id
      INTEGER :: color
      INTEGER :: degree ! degree (number of neighbors)
      INTEGER :: dsat ! degree of saturation
      LOGICAL, ALLOCATABLE, DIMENSION(:)       :: color_present ! the colors present on neighbors
      TYPE(vertex_p_type), DIMENSION(:), POINTER :: neighbors => NULL()
   END TYPE vertex

   TYPE vertex_p_type
      TYPE(vertex), POINTER :: vertex
   END TYPE vertex_p_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'kg_vertex_coloring_methods'

   PUBLIC :: kg_vertex_coloring

CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param kg_env ...
!> \param pairs ...
!> \param graph ...
! **************************************************************************************************
   SUBROUTINE kg_create_graph(kg_env, pairs, graph)
      TYPE(kg_environment_type), POINTER                 :: kg_env
      INTEGER(KIND=int_4), ALLOCATABLE, &
         DIMENSION(:, :), INTENT(IN)                     :: pairs
      TYPE(vertex_p_type), DIMENSION(:), POINTER         :: graph

      CHARACTER(len=*), PARAMETER :: routineN = 'kg_create_graph', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, imol, ineighbor, jmol, kmol, &
                                                            maxdegree, nneighbors, nnodes

      CALL timeset(routineN, handle)

! The array pairs contains all interacting (overlapping) pairs of molecules.
! It is assumed to be ordered in the following way:
! (1,2), (1,3), (1,4), ..., (2,3), (2,4), ...
! There are no entries (i,i)
! get the number of nodes = total number of molecules

      nnodes = SIZE(kg_env%molecule_set)

      NULLIFY (graph)
      ALLOCATE (graph(nnodes))

      ! allocate and initialize all vertices
      DO i = 1, nnodes
         ALLOCATE (graph(i)%vertex)
         graph(i)%vertex%id = i ! id = imol (molecular index)
         graph(i)%vertex%color = 0 ! means vertex is not colored yet
         graph(i)%vertex%dsat = 0 ! no colored neighbors yet
         graph(i)%vertex%degree = 0 ! as needed for maxdegree....
      END DO

      ! allocate the neighbor lists
      imol = 0

      maxdegree = 0

      DO i = 1, SIZE(pairs, 2)
         jmol = pairs(1, i)
         ! counting loop
         IF (jmol .NE. imol) THEN
            IF (imol .NE. 0) THEN
               ALLOCATE (graph(imol)%vertex%neighbors(nneighbors))
               graph(imol)%vertex%degree = nneighbors
               IF (nneighbors .GT. maxdegree) maxdegree = nneighbors
            END IF
            imol = jmol
            nneighbors = 0
         END IF
         nneighbors = nneighbors+1
      END DO

      IF (imol .NE. 0) THEN
         ALLOCATE (graph(imol)%vertex%neighbors(nneighbors))
         graph(imol)%vertex%degree = nneighbors
         IF (nneighbors .GT. maxdegree) maxdegree = nneighbors
      END IF

      kg_env%maxdegree = maxdegree

      ! there can be now some nodes that have no neighbors, thus vertex%neighbors
      ! is NOT ASSOCIATED

      ! now add the neighbors - if there are any
      imol = 0
      ineighbor = 0

      DO i = 1, SIZE(pairs, 2)
         jmol = pairs(1, i)
         IF (jmol .NE. imol) THEN
            ineighbor = 0
            imol = jmol
         END IF
         ineighbor = ineighbor+1
         kmol = pairs(2, i)
         graph(imol)%vertex%neighbors(ineighbor)%vertex => graph(kmol)%vertex
      END DO

      DO i = 1, SIZE(graph)
         IF (graph(i)%vertex%degree > 0) THEN
            ALLOCATE (graph(i)%vertex%color_present(100))
            graph(i)%vertex%color_present(:) = .FALSE.
         END IF
      END DO

      CALL timestop(handle)

   END SUBROUTINE

   ! greedy algorithm
! **************************************************************************************************
!> \brief ...
!> \param graph ...
!> \param maxdegree ...
!> \param ncolors ...
! **************************************************************************************************
   SUBROUTINE color_graph_greedy(graph, maxdegree, ncolors)
      TYPE(vertex_p_type), DIMENSION(:), POINTER         :: graph
      INTEGER, INTENT(in)                                :: maxdegree
      INTEGER, INTENT(out)                               :: ncolors

      CHARACTER(len=*), PARAMETER :: routineN = 'color_graph_greedy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: color, handle, i, j, newcolor, &
                                                            nneighbors, nnodes
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: color_present

      CALL timeset(routineN, handle)

      ncolors = 0

      nnodes = SIZE(graph)

      ALLOCATE (color_present(maxdegree+1))

      DO i = 1, nnodes
         color_present(:) = .FALSE.
         IF (ASSOCIATED(graph(i)%vertex%neighbors)) THEN
            nneighbors = SIZE(graph(i)%vertex%neighbors)
            DO j = 1, nneighbors
               color = graph(i)%vertex%neighbors(j)%vertex%color
               IF (color .NE. 0) color_present(color) = .TRUE.
            END DO
         END IF
         DO j = 1, maxdegree+1 !nnodes
            IF (color_present(j) .EQV. .FALSE.) THEN
               newcolor = j
               EXIT
            END IF
         END DO
         IF (newcolor .GT. ncolors) ncolors = newcolor
         graph(i)%vertex%color = newcolor ! smallest possible
      END DO

      DEALLOCATE (color_present)

      CALL timestop(handle)

   END SUBROUTINE

   ! prints the subset info to the screen - useful for vmd visualization
   ! note that the index starts with '0' and not with '1'
! **************************************************************************************************
!> \brief ...
!> \param graph ...
!> \param ncolors ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE print_subsets(graph, ncolors, unit_nr)
      TYPE(vertex_p_type), DIMENSION(:), POINTER         :: graph
      INTEGER, INTENT(IN)                                :: ncolors, unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'print_subsets', routineP = moduleN//':'//routineN

      INTEGER                                            :: counter, handle, i, j, nnodes

      CALL timeset(routineN, handle)

      IF (unit_nr > 0) THEN

         WRITE (unit_nr, '(T2,A,T10,A)') "Color |", "Molecules in this subset (IDs start from 0)"

         nnodes = SIZE(graph)

         DO i = 1, ncolors
            WRITE (unit_nr, '(T2,I5,1X,A)', ADVANCE='NO') i, "|"
            counter = 0
            DO j = 1, nnodes
               IF (graph(j)%vertex%color .EQ. i) THEN
                  counter = counter+1
                  IF (MOD(counter, 13) .EQ. 0) THEN
                     counter = counter+1
                     WRITE (unit_nr, '()') ! line break
                     WRITE (unit_nr, '(6X,A)', ADVANCE='NO') " |" ! indent next line
                  END IF
                  WRITE (unit_nr, '(I5,1X)', ADVANCE='NO') graph(j)%vertex%id-1
               END IF
            END DO
            WRITE (unit_nr, '()')
         END DO

      END IF

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param dsat ...
!> \param degree ...
!> \retval value ...
! **************************************************************************************************
   ELEMENTAL FUNCTION kg_get_value(dsat, degree) RESULT(value)
      INTEGER, INTENT(IN)                                :: dsat, degree
      INTEGER(KIND=valt)                                 :: value

      INTEGER, PARAMETER                                 :: huge_4 = 2_int_4**30

!   INTEGER, PARAMETER                       :: huge_4 = HUGE(0_int_4) ! PR67219 workaround
! we actually need a max_heap

      value = (huge_4-INT(dsat, KIND=int_8))*huge_4+huge_4-degree

   END FUNCTION

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param graph ...
! **************************************************************************************************
   SUBROUTINE kg_heap_fill(heap, graph)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      TYPE(vertex_p_type), DIMENSION(:), INTENT(IN), &
         POINTER                                         :: graph

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kg_heap_fill', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, nnodes
      INTEGER(kind=valt), ALLOCATABLE, DIMENSION(:)      :: values
      LOGICAL                                            :: heap_error

      nnodes = SIZE(graph)

      ALLOCATE (values(nnodes))

      DO i = 1, nnodes
         values(i) = kg_get_value(0, graph(i)%vertex%degree)
      END DO

      ! fill the heap
      CALL heap_fill(heap, values, heap_error)
      IF (heap_error) &
         CALL cp__w("kg_vertex_coloring_methods.F",302,"Error initially filling heap.")

      DEALLOCATE (values)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param key ...
! **************************************************************************************************
   SUBROUTINE kg_heap_pop(heap, key)
      TYPE(heap_t)                                       :: heap
      INTEGER, INTENT(OUT)                               :: key

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kg_heap_pop', routineP = moduleN//':'//routineN

      INTEGER(KIND=valt)                                 :: value
      LOGICAL                                            :: found, heap_error

      CALL heap_pop(heap, key, value, found, heap_error)
      IF (heap_error) &
         CALL cp__w("kg_vertex_coloring_methods.F",324,"Error getting topmost heap element.")

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   SUBROUTINE kg_heap_reset_node(heap, key, value)
      TYPE(heap_t)                                       :: heap
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER(KIND=valt), INTENT(IN)                     :: value

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kg_heap_reset_node', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: heap_error

      CALL heap_reset_node(heap, key, value, heap_error)
      IF (heap_error) &
         CALL cp__w("kg_vertex_coloring_methods.F",346,"Error changing value of heap element.")

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param node ...
! **************************************************************************************************
   SUBROUTINE kg_update_node(heap, node)
      TYPE(heap_t)                                       :: heap
      TYPE(vertex), INTENT(IN), POINTER                  :: node

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kg_update_node', routineP = moduleN//':'//routineN

      INTEGER                                            :: degree, dsat, id
      INTEGER(KIND=valt)                                 :: value

      id = node%id

      ! only update node if not yet colored
      IF (heap%index(id) > 0) THEN

         degree = node%degree
         dsat = node%dsat

         value = kg_get_value(dsat, degree)

         CALL kg_heap_reset_node(heap, id, value)

      END IF

   END SUBROUTINE

   ! Subroutine revised by Samuel Andermatt (2013.11)
! **************************************************************************************************
!> \brief ...
!> \param kg_env ...
!> \param graph ...
!> \param ncolors ...
! **************************************************************************************************
   SUBROUTINE kg_dsatur(kg_env, graph, ncolors)
      TYPE(kg_environment_type), POINTER                 :: kg_env
      TYPE(vertex_p_type), DIMENSION(:), POINTER         :: graph
      INTEGER(KIND=int_4), INTENT(OUT)                   :: ncolors

      CHARACTER(len=*), PARAMETER :: routineN = 'kg_dsatur', routineP = moduleN//':'//routineN

      INTEGER                                            :: color_limit, handle, i, j, key, &
                                                            maxdegree, nneighbors, nnodes
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: color_present
      TYPE(heap_t)                                       :: heap
      TYPE(vertex), POINTER                              :: neighbor, this

      CALL timeset(routineN, handle)

      CALL cite_reference(Brelaz1979)

      ncolors = 0
      color_limit = 100
      maxdegree = kg_env%maxdegree
      nnodes = SIZE(graph)

      IF (kg_env%maxdegree .EQ. 0) THEN
         ! all nodes are disconnected

         ncolors = 1

         DO i = 1, nnodes
            ! only one color needed to color the graph
            graph(i)%vertex%color = 1
         END DO

      ELSE

         ! allocate and fill the heap
         CALL heap_new(heap, nnodes)

         CALL kg_heap_fill(heap, graph)

         DO WHILE (heap%n > 0)

            CALL kg_heap_pop(heap, key)

            this => graph(key)%vertex

            nneighbors = 0

            IF (ASSOCIATED(this%neighbors)) THEN
               nneighbors = SIZE(this%neighbors)
            ELSE !node is isolated
               this%color = 1
               CYCLE
            END IF
            DO i = 1, this%degree+1
               IF (this%color_present(i) .EQV. .FALSE.) THEN
                  this%color = i ! smallest possible
                  EXIT
               END IF
            END DO
            IF (this%color .GT. ncolors) ncolors = this%color

            ! update all neighboring nodes
            DO i = 1, nneighbors
               neighbor => this%neighbors(i)%vertex
               IF (neighbor%color_present(this%color)) CYCLE
               neighbor%color_present(this%color) = .TRUE.
               neighbor%dsat = neighbor%dsat+1
               IF (neighbor%color /= 0) CYCLE
               CALL kg_update_node(heap, neighbor)

            END DO

            IF (this%color == color_limit) THEN !resize all color_present arrays
               ALLOCATE (color_present(color_limit))
               DO i = 1, nnodes
                  IF (graph(i)%vertex%degree == 0) CYCLE
                  color_present(:) = graph(i)%vertex%color_present
                  DEALLOCATE (graph(i)%vertex%color_present)
                  ALLOCATE (graph(i)%vertex%color_present(color_limit*2))
                  DO j = 1, color_limit
                     graph(i)%vertex%color_present(j) = color_present(j)
                  END DO
                  DO j = color_limit+1, 2*color_limit
                     graph(i)%vertex%color_present(j) = .FALSE.
                  END DO
               END DO
               DEALLOCATE (color_present)
               color_limit = color_limit*2
            END IF

         END DO

         ! release the heap
         CALL heap_release(heap)

      END IF

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief Checks if the color of two nodes can be exchanged legally
!> \param this ...
!> \param neighbor ...
!> \param low_col ...
!> \param switchable ...
!> \param ncolors ...
!> \param color_present ...
!> \par History
!>       2013.11 created [Samuel Andermatt]
!> \author Samuel Andermatt
! **************************************************************************************************
   SUBROUTINE kg_check_switch(this, neighbor, low_col, switchable, ncolors, color_present)

      TYPE(vertex), POINTER                              :: this, neighbor
      INTEGER, INTENT(OUT)                               :: low_col
      LOGICAL                                            :: switchable
      INTEGER                                            :: ncolors
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: color_present

      INTEGER                                            :: i

      switchable = .TRUE.
      low_col = ncolors+1

      DO i = 1, this%degree
         IF (this%neighbors(i)%vertex%id == neighbor%id) CYCLE
         IF (this%neighbors(i)%vertex%color == neighbor%color) THEN
            switchable = .FALSE.
            EXIT
         END IF
      END DO
      IF (.NOT. switchable) RETURN

      color_present(:) = .FALSE.
      color_present(neighbor%color) = .TRUE.
      DO i = 1, neighbor%degree
         IF (neighbor%neighbors(i)%vertex%id == this%id) CYCLE
         color_present(neighbor%neighbors(i)%vertex%color) = .TRUE.
      END DO
      DO i = 1, this%color
         IF (.NOT. color_present(i)) THEN
            low_col = i
            EXIT
         END IF
      END DO

   END SUBROUTINE

! **************************************************************************************************
!> \brief An algorithm that works on an already colored graph and tries to
!>          reduce the amount of colors by switching the colors of
!>          neighboring vertices
!> \param graph ...
!> \param ncolors ...
!> \par History
!>       2013.11 created [Samuel Andermatt]
!> \author Samuel Andermatt
! **************************************************************************************************
   SUBROUTINE kg_pair_switching(graph, ncolors)
      TYPE(vertex_p_type), DIMENSION(:), POINTER         :: graph
      INTEGER                                            :: ncolors

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kg_pair_switching', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: depth, handle, i, iter, j, low_col, &
                                                            low_col_neigh, maxdepth, &
                                                            maxiterations, nnodes, partner
      LOGICAL                                            :: switchable
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: color_present
      TYPE(vertex), POINTER                              :: neighbor, this

      CALL timeset(routineN, handle)
      nnodes = SIZE(graph)
      ALLOCATE (color_present(ncolors))

      !Some default values that should work well
      maxdepth = 4
      maxiterations = 20

      DO iter = 1, maxiterations
         DO depth = 1, maxdepth !First the nodes with larges colors are attempted to be switched and reduced
            !Go trough all nodes and try if you can reduce their color by switching with a neighbour
            DO i = 1, nnodes
               this => graph(i)%vertex
               IF (.NOT. ASSOCIATED(this%neighbors)) CYCLE
               IF (graph(i)%vertex%color < ncolors-depth+1) CYCLE !Node already has a low color

               partner = 0
               low_col = this%color+1

               DO j = 1, this%degree
                  neighbor => this%neighbors(j)%vertex
                  IF (neighbor%color > this%color) CYCLE
                  CALL kg_check_switch(this, neighbor, low_col_neigh, switchable, ncolors, color_present)
                  IF (switchable .AND. low_col_neigh < low_col) THEN
                     partner = j
                     low_col = low_col_neigh
                  END IF
               END DO
               IF (partner == 0) CYCLE !Cannot switch favourably with anybody
               !Switch the nodes
               this%color = this%neighbors(partner)%vertex%color
               this%neighbors(partner)%vertex%color = low_col
            END DO

            ncolors = 0
            DO j = 1, nnodes
               IF (graph(j)%vertex%color > ncolors) THEN
                  ncolors = graph(j)%vertex%color
               END IF
            END DO
         END DO
      END DO

      DEALLOCATE (color_present)
      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param graph ...
!> \param valid ...
! **************************************************************************************************
   SUBROUTINE check_coloring(graph, valid)
      TYPE(vertex_p_type), DIMENSION(:), INTENT(in), &
         POINTER                                         :: graph
      LOGICAL, INTENT(out)                               :: valid

      CHARACTER(len=*), PARAMETER :: routineN = 'check_coloring', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, nneighbors, nnodes
      TYPE(vertex), POINTER                              :: neighbor, node

      CALL timeset(routineN, handle)

      valid = .TRUE.
      nnodes = SIZE(graph)

      DO i = 1, nnodes
         node => graph(i)%vertex
         IF (ASSOCIATED(node%neighbors)) THEN
            nneighbors = SIZE(node%neighbors)
            DO j = 1, nneighbors
               neighbor => node%neighbors(j)%vertex
               IF (neighbor%color == node%color) THEN
                  valid = .FALSE.
                  RETURN
               END IF
            END DO
         END IF
      END DO

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param graph ...
! **************************************************************************************************
   SUBROUTINE deallocate_graph(graph)
      TYPE(vertex_p_type), DIMENSION(:), POINTER         :: graph

      INTEGER                                            :: i, nnodes

      nnodes = SIZE(graph)

      DO i = 1, nnodes
         IF (ASSOCIATED(graph(i)%vertex%neighbors)) DEALLOCATE (graph(i)%vertex%neighbors)
         DEALLOCATE (graph(i)%vertex)
      END DO
      DEALLOCATE (graph)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param kg_env ...
!> \param pairs ...
!> \param ncolors ...
!> \param color_of_node ...
! **************************************************************************************************
   SUBROUTINE kg_vertex_coloring(kg_env, pairs, ncolors, color_of_node)
      TYPE(kg_environment_type), POINTER                 :: kg_env
      INTEGER(KIND=int_4), ALLOCATABLE, &
         DIMENSION(:, :), INTENT(IN)                     :: pairs
      INTEGER(KIND=int_4), INTENT(OUT)                   :: ncolors
      INTEGER(KIND=int_4), ALLOCATABLE, DIMENSION(:), &
         INTENT(out)                                     :: color_of_node

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kg_vertex_coloring', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: color, handle, i, nnodes, unit_nr
      LOGICAL                                            :: valid
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(vertex_p_type), DIMENSION(:), POINTER         :: graph

! get a useful output_unit

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      CALL timeset(routineN, handle)

      CALL kg_create_graph(kg_env, pairs, graph)

      SELECT CASE (kg_env%coloring_method)
      CASE (kg_color_greedy)
         ! simple greedy algorithm
         CALL color_graph_greedy(graph, kg_env%maxdegree, ncolors)
      CASE (kg_color_dsatur)
         ! color with max degree of saturation
         CALL kg_dsatur(kg_env, graph, ncolors)
      CASE DEFAULT
         CALL cp__b("kg_vertex_coloring_methods.F",710,"Coloring method not known.")
      END SELECT

      CALL kg_pair_switching(graph, ncolors)

      valid = .FALSE.
      CALL check_coloring(graph, valid)
      IF (.NOT. valid) &
         CALL cp__b("kg_vertex_coloring_methods.F",718,"Coloring not valid.")

      nnodes = SIZE(kg_env%molecule_set)

      ALLOCATE (color_of_node(nnodes))

      ! gather the subset info in a simple integer array
      DO i = 1, nnodes
         color = graph(i)%vertex%color
         color_of_node(i) = color
      END DO

      IF (unit_nr > 0) THEN

         WRITE (unit_nr, '(T2,A,A,A)') REPEAT("-", 30), " KG coloring result ", REPEAT("-", 29)
         IF (.FALSE.) THEN ! more output for VMD
            WRITE (unit_nr, '()')
            CALL print_subsets(graph, ncolors, unit_nr)
            WRITE (unit_nr, '()')
         ENDIF
         WRITE (unit_nr, '(T2, A16,50X,I6,1X,A6)') 'Number of colors', ncolors, 'colors'
         IF (valid) WRITE (unit_nr, '(T2, A17,59X,A3)') 'Coloring verified', 'yes'
         WRITE (unit_nr, '(T2,A)') REPEAT("-", 79)

      END IF

      CALL deallocate_graph(graph)

      CALL timestop(handle)

   END SUBROUTINE

END MODULE
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types to handle submatrices
!> \par History
!>       2013.01 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE domain_submatrix_types
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'domain_submatrix_types'

   INTEGER, PARAMETER, PUBLIC           :: select_row_col = 1
   INTEGER, PARAMETER, PUBLIC           :: select_row = 2

   PUBLIC :: domain_submatrix_type, domain_map_type

   ! submatrix storage with the meta-data necessary to convert
   ! the submatrix into the DBCSR format
   TYPE domain_submatrix_type
      INTEGER                                       :: domain
      REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE    :: mdata
      INTEGER                                       :: nbrows
      INTEGER                                       :: nbcols
      INTEGER                                       :: nrows
      INTEGER                                       :: ncols
      INTEGER, DIMENSION(:), ALLOCATABLE            :: dbcsr_row
      INTEGER, DIMENSION(:), ALLOCATABLE            :: dbcsr_col
      INTEGER, DIMENSION(:), ALLOCATABLE            :: size_brow
      INTEGER, DIMENSION(:), ALLOCATABLE            :: size_bcol
      INTEGER                                       :: nnodes
      INTEGER                                       :: groupid
   END TYPE domain_submatrix_type

   TYPE domain_map_type
      INTEGER, DIMENSION(:), ALLOCATABLE     :: index1
      INTEGER, DIMENSION(:, :), ALLOCATABLE   :: pairs
   END TYPE domain_map_type

END MODULE domain_submatrix_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types for all ALMO-based methods
!> \par History
!>       2011.05 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE almo_scf_types
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE domain_submatrix_types,          ONLY: domain_map_type,&
                                              domain_submatrix_type
   USE input_constants,                 ONLY: &
        cg_dai_yuan, cg_fletcher, cg_fletcher_reeves, cg_hager_zhang, cg_hestenes_stiefel, &
        cg_liu_storey, cg_polak_ribiere, cg_zero, optimizer_diis, optimizer_pcg, prec_default, &
        prec_zero
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'almo_scf_types'

   INTEGER, PARAMETER, PUBLIC   :: almo_mat_dim_aobasis = 1, &
                                   almo_mat_dim_occ = 2, &
                                   almo_mat_dim_virt = 3, &
                                   almo_mat_dim_virt_full = 4, &
                                   almo_mat_dim_domains = 5, &
                                   almo_mat_dim_virt_disc = 6
   REAL(KIND=dp), PARAMETER, PUBLIC :: almo_max_cutoff_multiplier = 2.2_dp

   PUBLIC :: almo_scf_env_type, optimizer_options_type, &
             print_optimizer_options

   TYPE optimizer_options_type

      ! SCF procedure for the block-diagonal ALMOs
      !REAL(KIND=dp)         :: eps_scf_bd -> eps_error
      !INTEGER               :: max_scf_bd -> max_iter
      !INTEGER               :: ndiis_bd -> ndiis

      ! SCF procedure for the quenched ALMOs (xALMOs)
      !REAL(KIND=dp)         :: eps_scf_q -> eps_error
      !REAL(KIND=dp)         :: eps_lin_search ->  eps_lin_search_error
      !INTEGER               :: max_scf_q -> max_iter
      !INTEGER               :: scf_conjugator -> conjugator
      !INTEGER               :: outer_max_scf_q -> max_iter for a separate optimizer
      !INTEGER               :: ndiis_q -> ndiis

      REAL(KIND=dp)  :: eps_error, &
                        lin_search_eps_error, &
                        lin_search_step_size_guess

      INTEGER        :: optimizer_type ! diis, pcg, etc.

      INTEGER        :: preconditioner, & ! preconditioner type
                        conjugator, & ! conjugator type
                        max_iter, &
                        max_iter_outer_loop, &
                        ndiis ! diis history length

   END TYPE optimizer_options_type

   ! the structure contains general info about the system
   TYPE almo_scf_env_type

      TYPE(cp_para_env_type), POINTER  :: para_env
      TYPE(cp_blacs_env_type), POINTER  :: blacs_env

      INTEGER :: nspins, nelectrons_total, naos
      INTEGER :: natoms, nmolecules
      INTEGER, DIMENSION(2) :: nelectrons_spin

      ! Definitions:
      ! I.  Domain - a subset of basis functions (e.g. AOs),
      ! II. Group  - a subset of electrons delocalized within a domain.
      !
      ! The following variables specify the group-domain structure
      ! of the system. Several rules must be obeyed:
      ! 1. There should be no zero domains (i.e. domain contains at least one AO).
      ! 2. There should be no empty domains (i.e. all domains must be populated
      !     by at least one electron).
      ! 3. If two groups are localized within the same domain they are combined
      ! It follows that the number of domains is equal to the number of groups
      !
      ! Number of domains
      INTEGER :: ndomains

      ! List of atoms, whose basis functions are included into the domain.
      ! It is assumed that:
      !   (a) basis functions are localized and atom-labeled,
      !   (b) basis functions are grouped into atomic sets (i.e. if a basis
      !       function on an atom is in domain A then all basis functions on
      !       this atom are in domain A)
      !TYPE(domain_list_type), DIMENSION(:), ALLOCATABLE   :: atom_list_of_domain
      ! List of basis functions included into the domain
      !TYPE(domain_list_type), DIMENSION(:), ALLOCATABLE   :: basis_list_of_domain

      ! Number of electrons of each spin for a given domain (second dim is spin).
      ! Note that some domains can be populated only with alpha or beta electrons.
      INTEGER, DIMENSION(:, :), ALLOCATABLE                :: nocc_of_domain
      ! Number of basis functions for a given domain
      INTEGER, DIMENSION(:), ALLOCATABLE                  :: nbasis_of_domain
      ! Define number of virtuals for a given domain: nvirt = nbasis - nocc
      INTEGER, DIMENSION(:, :), ALLOCATABLE                :: nvirt_full_of_domain
      ! Define the dimension of truncated virtual subspace for a given domain:
      INTEGER, DIMENSION(:, :), ALLOCATABLE                :: nvirt_of_domain
      ! Define the dimension of discarded virtual subspace for a given domain:
      INTEGER, DIMENSION(:, :), ALLOCATABLE                :: nvirt_disc_of_domain
      ! Each domain has its own mu - "fermi" level
      REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE          :: mu_of_domain
      INTEGER, DIMENSION(:), ALLOCATABLE                  :: first_atom_of_domain
      INTEGER, DIMENSION(:), ALLOCATABLE                  :: last_atom_of_domain
      ! The following arrays are useful only with non-overlapping domains
      ! RZK-warning generalization is required
      INTEGER, DIMENSION(:), ALLOCATABLE        :: domain_index_of_ao
      INTEGER, DIMENSION(:), ALLOCATABLE        :: domain_index_of_atom

      ! Charge of domains
      INTEGER, DIMENSION(:), ALLOCATABLE                  :: charge_of_domain
      ! Charge of domains
      INTEGER, DIMENSION(:), ALLOCATABLE                  :: multiplicity_of_domain

      ! The matrix contains information about the delocalization of
      ! alpha and beta electrons.
      ! Rows denote basis function, columns denote electrons.
      ! Non-zero (j,i) entry means that electron j can delocalize over
      ! basis function i. 0.0 means no delocalization
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE   :: quench_t
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE   :: quench_t_blk
      ! Local array for a compact description of quench_t
      TYPE(domain_map_type), DIMENSION(:), ALLOCATABLE :: domain_map

      ! Several special cases for the structure of the group-domain matrix:
      ! 1. The basis functions can be grouped into:
      !    a. molecular sets
      !    b. atomic sets
      ! 2. Electrons can be grouped into:
      !    a. molecular sets
      !    b. atomic sets
      INTEGER :: domain_layout_mos, domain_layout_aos
      ! ALMO  constraint type.
      INTEGER :: constraint_type

      ! decomposition analysis (energy da, charge da)
      INTEGER :: almo_eda

      ! Desciptors of molecules
      !INTEGER, DIMENSION(:), ALLOCATABLE        :: molecule_index_of_atom
      !INTEGER, DIMENSION(:), ALLOCATABLE        :: first_atom_of_molecule
      !INTEGER, DIMENSION(:), ALLOCATABLE        :: nbasis_of_molecule
      !INTEGER, DIMENSION(:,:), ALLOCATABLE      :: nocc_of_molecule
      !INTEGER, DIMENSION(:,:), ALLOCATABLE      :: nvirt_of_molecule
      !REAL(KIND=dp),DIMENSION(:,:), ALLOCATABLE :: mu_of_molecule

      ! Descriptors of atoms
      !INTEGER, DIMENSION(:), ALLOCATABLE        :: nbasis_of_atom
      !INTEGER, DIMENSION(:,:), ALLOCATABLE      :: nocc_of_atom
      !INTEGER, DIMENSION(:,:), ALLOCATABLE      :: nvirt_of_atom
      !REAL(KIND=dp),DIMENSION(:,:), ALLOCATABLE :: mu_of_atom

      ! All AO and MO matrices are distributed for parallel computations.
      ! The following flags specify what constitues a block for a parallel
      ! distribution. Both AOs and MOs can be divided into atomic or
      ! molecular blocks. Domain blocks should be equal or larger than
      ! the distribution blocks (otherwise retain_sparsity does not work).
      ! Possible values: almo_mat_distr_atomic, almo_mat_distr_molecular
      INTEGER :: mat_distr_aos, mat_distr_mos
      ! Define mappping from a distribution block to a domain
      INTEGER, DIMENSION(:), ALLOCATABLE :: domain_index_of_ao_block
      INTEGER, DIMENSION(:), ALLOCATABLE :: domain_index_of_mo_block

      LOGICAL              :: need_previous_ks
      LOGICAL              :: need_virtuals
      LOGICAL              :: need_orbital_energies
      LOGICAL              :: s_inv_done
      LOGICAL              :: s_sqrt_done
      REAL(KIND=dp)        :: almo_scf_energy
      LOGICAL               :: orthogonal_basis, fixed_mu

      ! Controls for the SCF procedure
      REAL(KIND=dp)         :: eps_filter
      INTEGER               :: almo_scf_guess
      REAL(KIND=dp)         :: eps_prev_guess
      INTEGER               :: order_lanczos
      REAL(KIND=dp)         :: eps_lanczos
      INTEGER               :: max_iter_lanczos
      REAL(KIND=dp)         :: mixing_fraction
      REAL(KIND=dp)         :: mu
      ! SCF procedure for the block-diagonal ALMOs
      INTEGER               :: almo_update_algorithm
      ! SCF procedure for the quenched ALMOs (xALMOs)
      INTEGER               :: xalmo_update_algorithm

      ! ALMO SCF delocalization control
      LOGICAL               :: perturbative_delocalization
      INTEGER               :: quencher_radius_type
      REAL(KIND=dp)         :: quencher_r0_factor, &
                               quencher_r1_factor, &
                               !quencher_r0_shift,&
                               !quencher_r1_shift,&
                               quencher_s0, &
                               quencher_s1, &
                               envelope_amplitude

      !!!!!!!!!!!!!!!!!!!!!!!
      !!!!!! MATRICES !!!!!!!
      !!!!!!!!!!!!!!!!!!!!!!!

      ! AO overlap NxN
      TYPE(cp_dbcsr_type), DIMENSION(1)   :: matrix_s
      TYPE(cp_dbcsr_type), DIMENSION(1)   :: matrix_s_inv
      TYPE(cp_dbcsr_type), DIMENSION(1)   :: matrix_s_sqrt
      TYPE(cp_dbcsr_type), DIMENSION(1)   :: matrix_s_sqrt_inv
      ! block-diagonal AO overlap NxN
      TYPE(cp_dbcsr_type), DIMENSION(1)   :: matrix_s_blk
      TYPE(cp_dbcsr_type), DIMENSION(1)   :: matrix_s_blk_inv
      TYPE(cp_dbcsr_type), DIMENSION(1)   :: matrix_s_blk_sqrt
      TYPE(cp_dbcsr_type), DIMENSION(1)   :: matrix_s_blk_sqrt_inv

      ! occupied ALMO coeff NxOCC (alpha,beta - if necessary)
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_t_blk
      ! occupied MO coeff NxOCC (alpha,beta - if necessary)
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_t
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_t_tr
      ! MO overlap OCCxOCC and its inverse (alpha, beta - if necessary)
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_sigma, &
                                                        matrix_sigma_inv, &
                                                        matrix_sigma_sqrt, &
                                                        matrix_sigma_sqrt_inv, &
                                                        matrix_sigma_blk

      ! error vector (alpha,beta - if necessary)
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_err_blk
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_err_xx

      ! MO overlap VIRTxVIRT and its derivatives
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_sigma_vv, &
                                                        matrix_sigma_vv_blk, &
                                                        matrix_sigma_vv_sqrt, &
                                                        matrix_sigma_vv_sqrt_inv

      ! template of various VIRT x VIR matrices
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_vv_full_blk, &
                                                        matrix_vv_disc_blk, &
                                                        matrix_vv_disc

      ! VIRT-OCC MO overlap
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_vo, matrix_ov
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_ov_full, &
                                                        matrix_ov_disc
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_x

      ! VIRT_DISC x VIRT_RETAINED
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_k_blk
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_k_tr
      ! matrix_k_blk_ones is blocked with all elements equal to 1.0
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_k_blk_ones

      ! virtual ALMO coeff NxV
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_v_blk, &
                                                        matrix_v, &
                                                        matrix_v_full_blk, &
                                                        matrix_v_disc, &
                                                        matrix_v_disc_blk

      ! kohn-sham matrix (alpha,beta - if necessary)
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_ks
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_ks_blk
      ! density NxN (alpha,beta - if necessary)
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_p
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_p_blk

      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_eoo
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_evv_full
      TYPE(cp_dbcsr_type), DIMENSION(:), &
         ALLOCATABLE                  :: matrix_ks_almo_scf_converged

      ! preconditioner for k-optimization
      ! RZK-warning: do they have to be stored?
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: opt_k_t_rr, &
                                                        opt_k_t_dd, &
                                                        opt_k_denom

      ! second dimension is spin
      TYPE(domain_submatrix_type), DIMENSION(:, :), ALLOCATABLE :: domain_preconditioner
      TYPE(domain_submatrix_type), DIMENSION(:, :), ALLOCATABLE :: domain_s_inv
      TYPE(domain_submatrix_type), DIMENSION(:, :), ALLOCATABLE :: domain_s_sqrt
      TYPE(domain_submatrix_type), DIMENSION(:, :), ALLOCATABLE :: domain_s_sqrt_inv
      TYPE(domain_submatrix_type), DIMENSION(:, :), ALLOCATABLE :: domain_ks_xx
      TYPE(domain_submatrix_type), DIMENSION(:, :), ALLOCATABLE :: domain_t
      TYPE(domain_submatrix_type), DIMENSION(:, :), ALLOCATABLE :: domain_err
      TYPE(domain_submatrix_type), DIMENSION(:, :), ALLOCATABLE :: domain_r_down_up

      INTEGER, DIMENSION(:), ALLOCATABLE                       :: cpu_of_domain

      ! Options for various optimizers collected neatly
      TYPE(optimizer_options_type)                   :: opt_block_diag_diis
      TYPE(optimizer_options_type)                   :: opt_block_diag_pcg
      TYPE(optimizer_options_type)                   :: opt_xalmo_diis
      TYPE(optimizer_options_type)                   :: opt_xalmo_pcg
      TYPE(optimizer_options_type)                   :: opt_k_pcg

      ! keywords that control electron delocalization treatment
      ! RZK-warning: many of these varibles should be collected
      !  into an optimizer_options_type variable
      INTEGER         :: deloc_method
      LOGICAL         :: deloc_use_occ_orbs
      LOGICAL         :: deloc_cayley_use_virt_orbs
      INTEGER         :: deloc_cayley_tensor_type
      LOGICAL         :: deloc_cayley_linear
      INTEGER         :: deloc_cayley_conjugator
      REAL(KIND=dp)   :: deloc_cayley_eps_convergence
      INTEGER         :: deloc_cayley_max_iter
      INTEGER         :: deloc_truncate_virt
      INTEGER         :: deloc_virt_per_domain
      LOGICAL         :: deloc_cayley_occ_precond
      LOGICAL         :: deloc_cayley_vir_precond

      !! keywords that control optimization of retained orbitals
      INTEGER         :: opt_k_conjugator !-> conjugartor
      REAL(KIND=dp)   :: opt_k_eps_convergence !-> eps_error
      REAL(KIND=dp)   :: opt_k_trial_step_size !-> lin_search_step_size_guess
      INTEGER         :: opt_k_max_iter !-> max_iter
      INTEGER         :: opt_k_outer_max_iter !-> max_iter for a separate 'outer' optimizer
      REAL(KIND=dp)   :: opt_k_trial_step_size_multiplier !-> ?
      INTEGER         :: opt_k_conj_iter_start !-> ?
      INTEGER         :: opt_k_prec_iter_start !-> ?
      INTEGER         :: opt_k_conj_iter_freq !-> ?
      INTEGER         :: opt_k_prec_iter_freq !-> ?

      ! development keywords
      INTEGER         :: integer01
      INTEGER         :: integer02
      INTEGER         :: integer03
      INTEGER         :: integer04
      INTEGER         :: integer05
      REAL(KIND=dp)   :: real01
      REAL(KIND=dp)   :: real02
      REAL(KIND=dp)   :: real03
      REAL(KIND=dp)   :: real04
      REAL(KIND=dp)   :: real05
      LOGICAL         :: logical01
      LOGICAL         :: logical02
      LOGICAL         :: logical03
      LOGICAL         :: logical04
      LOGICAL         :: logical05

   END TYPE almo_scf_env_type

CONTAINS

! **************************************************************************************************
!> \brief Prints out the options of an optimizer
!> \param optimizer - options to print
!> \param unit_nr - output stream
!> \par History
!>       2014.10 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE print_optimizer_options(optimizer, unit_nr)

      TYPE(optimizer_options_type), INTENT(IN)           :: optimizer
      INTEGER, INTENT(IN)                                :: unit_nr

      CHARACTER(len=*), PARAMETER :: routineN = 'print_optimizer_options', &
         routineP = moduleN//':'//routineN

      CHARACTER(33)                                      :: conj_string, prec_string, type_string

      IF (unit_nr .GT. 0) THEN

         SELECT CASE (optimizer%optimizer_type)
         CASE (optimizer_diis)
            type_string = "DIIS"
         CASE (optimizer_pcg)
            type_string = "PCG"
         END SELECT

         WRITE (unit_nr, '(T4,A,T48,A33)') "optimizer type:", TRIM(type_string)
         WRITE (unit_nr, '(T4,A,T48,I33)') "maximum iterations:", optimizer%max_iter
         WRITE (unit_nr, '(T4,A,T48,E33.3)') "target error:", optimizer%eps_error

         SELECT CASE (optimizer%optimizer_type)
         CASE (optimizer_diis)

            WRITE (unit_nr, '(T4,A,T48,I33)') "maximum DIIS history:", optimizer%ndiis

         CASE (optimizer_pcg)

            WRITE (unit_nr, '(T4,A,T48,E33.3)') "line search step size guess:", &
               optimizer%lin_search_step_size_guess
            WRITE (unit_nr, '(T4,A,T48,E33.3)') "line search target error:", &
               optimizer%lin_search_eps_error
            WRITE (unit_nr, '(T4,A,T48,I33)') "maximum outer loop iterations:", &
               optimizer%max_iter_outer_loop

            SELECT CASE (optimizer%preconditioner)
            CASE (prec_zero)
               prec_string = "NONE"
            CASE (prec_default)
               prec_string = "0.5 H + 0.5 S"
            END SELECT
            WRITE (unit_nr, '(T4,A,T48,A33)') "preconditioner:", TRIM(prec_string)

            SELECT CASE (optimizer%conjugator)
            CASE (cg_zero)
               conj_string = "Steepest descent"
            CASE (cg_polak_ribiere)
               conj_string = "Polak-Ribiere"
            CASE (cg_fletcher_reeves)
               conj_string = "Fletcher-Reeves"
            CASE (cg_hestenes_stiefel)
               conj_string = "Hestenes-Stiefel"
            CASE (cg_fletcher)
               conj_string = "Fletcher"
            CASE (cg_liu_storey)
               conj_string = "Liu-Storey"
            CASE (cg_dai_yuan)
               conj_string = "Dai-Yuan"
            CASE (cg_hager_zhang)
               conj_string = "Hager-Zhang"
            END SELECT
            WRITE (unit_nr, '(T4,A,T48,A33)') "conjugator:", TRIM(conj_string)

         END SELECT !optimizer type

      ENDIF

   END SUBROUTINE print_optimizer_options

END MODULE almo_scf_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Generate the atomic neighbor lists.
!> \par History
!>      - List rebuild for sab_orb neighbor list (10.09.2002,MK)
!>      - List rebuild for all lists (25.09.2002,MK)
!>      - Row-wise parallelized version (16.06.2003,MK)
!>      - Row- and column-wise parallelized version (19.07.2003,MK)
!>      - bug fix for non-periodic case (23.02.06,MK)
!>      - major refactoring (25.07.10,jhu)
!> \author Matthias Krack (08.10.1999,26.03.2002,16.06.2003)
! **************************************************************************************************
MODULE qs_neighbor_lists
   USE almo_scf_types,                  ONLY: almo_max_cutoff_multiplier
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell,&
                                              pbc,&
                                              plane_distance,&
                                              real_to_scaled,&
                                              scaled_to_real
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE distribution_2d_types,           ONLY: distribution_2d_type
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE external_potential_types,        ONLY: all_potential_type,&
                                              get_potential,&
                                              gth_potential_type
   USE input_constants,                 ONLY: dispersion_uff,&
                                              do_method_lrigpw,&
                                              do_se_IS_slater,&
                                              vdw_pairpot_dftd3,&
                                              vdw_pairpot_dftd3bj,&
                                              xc_vdw_fun_pairpot
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8
   USE kpoint_types,                    ONLY: kpoint_type
   USE message_passing,                 ONLY: mp_max,&
                                              mp_sum
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_types,                  ONLY: particle_type
   USE paw_proj_set_types,              ONLY: get_paw_proj_set,&
                                              paw_proj_set_type
   USE periodic_table,                  ONLY: ptable
   USE physcon,                         ONLY: bohr
   USE qs_dftb_types,                   ONLY: qs_dftb_atom_type
   USE qs_dftb_utils,                   ONLY: get_dftb_atom_param
   USE qs_dispersion_types,             ONLY: qs_dispersion_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_env_type,&
                                              set_ks_env
   USE qs_neighbor_list_types,          ONLY: &
        add_neighbor_list, add_neighbor_node, allocate_neighbor_list_set, &
        deallocate_neighbor_list_set, get_iterator_info, neighbor_list_iterate, &
        neighbor_list_iterator_create, neighbor_list_iterator_p_type, &
        neighbor_list_iterator_release, neighbor_list_p_type, neighbor_list_set_p_type, &
        neighbor_list_set_type
   USE scptb_types,                     ONLY: get_scptb_parameter,&
                                              scptb_parameter_type
   USE string_utilities,                ONLY: compress
   USE subcell_types,                   ONLY: allocate_subcell,&
                                              deallocate_subcell,&
                                              give_ijk_subcell,&
                                              subcell_type
   USE util,                            ONLY: locate,&
                                              sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************
   TYPE local_atoms_type
      INTEGER, DIMENSION(:), POINTER                   :: list, &
                                                          list_local_a_index, &
                                                          list_local_b_index, &
                                                          list_1d, &
                                                          list_a_mol, &
                                                          list_b_mol
   END TYPE local_atoms_type
! **************************************************************************************************

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_neighbor_lists'

   ! private counter, used to version qs neighbor lists
   INTEGER, SAVE, PRIVATE :: last_qs_neighbor_list_id_nr = 0

   ! Public subroutines
   PUBLIC :: build_qs_neighbor_lists, local_atoms_type, atom2d_cleanup, &
             atom2d_build, build_neighbor_lists, pair_radius_setup, &
             setup_neighbor_list, write_neighbor_lists
CONTAINS

! **************************************************************************************************
!> \brief   free the internals of atom2d
!> \param atom2d ...
!> \param
! **************************************************************************************************
   SUBROUTINE atom2d_cleanup(atom2d)
      TYPE(local_atoms_type), DIMENSION(:)               :: atom2d

      CHARACTER(len=*), PARAMETER :: routineN = 'atom2d_cleanup', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind

      CALL timeset(routineN, handle)
      DO ikind = 1, SIZE(atom2d)
         NULLIFY (atom2d(ikind)%list)
         IF (ASSOCIATED(atom2d(ikind)%list_local_a_index)) THEN
            DEALLOCATE (atom2d(ikind)%list_local_a_index)
         END IF
         IF (ASSOCIATED(atom2d(ikind)%list_local_b_index)) THEN
            DEALLOCATE (atom2d(ikind)%list_local_b_index)
         END IF
         IF (ASSOCIATED(atom2d(ikind)%list_a_mol)) THEN
            DEALLOCATE (atom2d(ikind)%list_a_mol)
         END IF
         IF (ASSOCIATED(atom2d(ikind)%list_b_mol)) THEN
            DEALLOCATE (atom2d(ikind)%list_b_mol)
         END IF
         IF (ASSOCIATED(atom2d(ikind)%list_1d)) THEN
            DEALLOCATE (atom2d(ikind)%list_1d)
         END IF
      END DO
      CALL timestop(handle)

   END SUBROUTINE atom2d_cleanup

! **************************************************************************************************
!> \brief   Build some distribution structure of atoms, refactored from build_qs_neighbor_lists
!> \param atom2d output
!> \param distribution_1d ...
!> \param distribution_2d ...
!> \param atomic_kind_set ...
!> \param molecule_set ...
!> \param molecule_only ...
!> \param particle_set ...
!> \author  JH
! **************************************************************************************************
   SUBROUTINE atom2d_build(atom2d, distribution_1d, distribution_2d, &
                           atomic_kind_set, molecule_set, molecule_only, particle_set)
      TYPE(local_atoms_type), DIMENSION(:)               :: atom2d
      TYPE(distribution_1d_type), POINTER                :: distribution_1d
      TYPE(distribution_2d_type), POINTER                :: distribution_2d
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      LOGICAL                                            :: molecule_only
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      CHARACTER(len=*), PARAMETER :: routineN = 'atom2d_build', routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, ia, iat, iatom, &
                                                            iatom_local, ikind, imol, natom, &
                                                            natom_a, natom_local_a, natom_local_b, &
                                                            nel, nkind
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom2mol, listindex, listsort
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind, local_cols_array, &
                                                            local_rows_array

      CALL timeset(routineN, handle)

      nkind = SIZE(atomic_kind_set)
      natom = SIZE(particle_set)
      ALLOCATE (atom_of_kind(natom))
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                               atom_of_kind=atom_of_kind)

      IF (molecule_only) THEN
         ALLOCATE (atom2mol(natom))
         DO imol = 1, SIZE(molecule_set)
            DO iat = molecule_set(imol)%first_atom, molecule_set(imol)%last_atom
               atom2mol(iat) = imol
            ENDDO
         END DO
      ENDIF

      DO ikind = 1, nkind
         NULLIFY (atom2d(ikind)%list)
         NULLIFY (atom2d(ikind)%list_local_a_index)
         NULLIFY (atom2d(ikind)%list_local_b_index)
         NULLIFY (atom2d(ikind)%list_1d)
         NULLIFY (atom2d(ikind)%list_a_mol)
         NULLIFY (atom2d(ikind)%list_b_mol)

         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom2d(ikind)%list)

         natom_a = SIZE(atom2d(ikind)%list)

         natom_local_a = distribution_2d%n_local_rows(ikind)
         natom_local_b = distribution_2d%n_local_cols(ikind)
         local_rows_array => distribution_2d%local_rows(ikind)%array
         local_cols_array => distribution_2d%local_cols(ikind)%array

         nel = distribution_1d%n_el(ikind)
         ALLOCATE (atom2d(ikind)%list_1d(nel))
         DO iat = 1, nel
            ia = distribution_1d%list(ikind)%array(iat)
            atom2d(ikind)%list_1d(iat) = atom_of_kind(ia)
         END DO

         ALLOCATE (listsort(natom_a), listindex(natom_a))
         listsort(1:natom_a) = atom2d(ikind)%list(1:natom_a)
         CALL sort(listsort, natom_a, listindex)
         ! Block rows
         IF (natom_local_a > 0) THEN
            ALLOCATE (atom2d(ikind)%list_local_a_index(natom_local_a))
            ALLOCATE (atom2d(ikind)%list_a_mol(natom_local_a))
            atom2d(ikind)%list_a_mol(:) = 0

            ! Build index vector for mapping
            DO iatom_local = 1, natom_local_a
               atom_a = local_rows_array(iatom_local)
               iatom = locate(listsort, atom_a)
               atom2d(ikind)%list_local_a_index(iatom_local) = listindex(iatom)
               IF (molecule_only) atom2d(ikind)%list_a_mol(iatom_local) = atom2mol(atom_a)
            END DO

         END IF

         ! Block columns
         IF (natom_local_b > 0) THEN

            ALLOCATE (atom2d(ikind)%list_local_b_index(natom_local_b))
            ALLOCATE (atom2d(ikind)%list_b_mol(natom_local_b))
            atom2d(ikind)%list_b_mol(:) = 0

            ! Build index vector for mapping
            DO iatom_local = 1, natom_local_b
               atom_a = local_cols_array(iatom_local)
               iatom = locate(listsort, atom_a)
               atom2d(ikind)%list_local_b_index(iatom_local) = listindex(iatom)
               IF (molecule_only) atom2d(ikind)%list_b_mol(iatom_local) = atom2mol(atom_a)
            END DO

         END IF

         DEALLOCATE (listsort, listindex)

      ENDDO

      DEALLOCATE (atom_of_kind)

      CALL timestop(handle)

   END SUBROUTINE atom2d_build

! **************************************************************************************************
!> \brief   Build all the required neighbor lists for Quickstep.
!> \param qs_env ...
!> \param para_env ...
!> \param molecular ...
!> \param force_env_section ...
!> \date    28.08.2000
!> \par History
!>          - Major refactoring (25.07.2010,jhu)
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE build_qs_neighbor_lists(qs_env, para_env, molecular, force_env_section)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      LOGICAL, OPTIONAL                                  :: molecular
      TYPE(section_vals_type), POINTER                   :: force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'build_qs_neighbor_lists', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: print_key_path
      INTEGER                                            :: handle, ikind, iw, maxatom, nkind, &
                                                            numshells, zat
      LOGICAL :: all_potential_present, almo, dftb, do_hfx, dokp, explicit, gth_potential_present, &
         lri_optbas, lrigpw, mic, molecule_only, nddo, paw_atom, paw_atom_present, scptb
      LOGICAL, ALLOCATABLE, DIMENSION(:) :: all_present, aux_fit_present, aux_present, &
         core_present, default_present, oce_present, orb_present, ppl_present, ppnl_present
      REAL(dp)                                           :: almo_rcov, almo_rvdw, alpha, roperator, &
                                                            subcells
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: all_pot_rad, aux_fit_radius, c_radius, &
                                                            calpha, core_radius, oce_radius, &
                                                            orb_radius, ppl_radius, ppnl_radius, &
                                                            zeff
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: pair_radius
      TYPE(all_potential_type), POINTER                  :: all_potential
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(distribution_1d_type), POINTER                :: distribution_1d
      TYPE(distribution_2d_type), POINTER                :: distribution_2d
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(gto_basis_set_type), POINTER                  :: aux_basis_set, aux_fit_basis_set, &
                                                            orb_basis_set
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(local_atoms_type), ALLOCATABLE, DIMENSION(:)  :: atom2d
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(neighbor_list_set_p_type), DIMENSION(:), POINTER :: sab_all, sab_almo, sab_aux_fit, &
         sab_aux_fit_asymm, sab_aux_fit_vs_orb, sab_cn, sab_core, sab_kp, sab_lrc, sab_orb, &
         sab_scp, sab_se, sab_tbe, sab_vdw, sac_ae, sac_ppl, sap_oce, sap_ppnl, soo_list
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(paw_proj_set_type), POINTER                   :: paw_proj
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_atom
      TYPE(qs_dispersion_type), POINTER                  :: dispersion_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(scptb_parameter_type), POINTER                :: scptb_kind
      TYPE(section_vals_type), POINTER                   :: hfx_sections, neighbor_list_section

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()

      NULLIFY (atomic_kind_set, qs_kind_set, cell, neighbor_list_section, &
               distribution_1d, distribution_2d, gth_potential, orb_basis_set, &
               particle_set, molecule_set, dft_control, ks_env)

      NULLIFY (sab_orb)
      NULLIFY (sac_ae)
      NULLIFY (sac_ppl)
      NULLIFY (sap_ppnl)
      NULLIFY (sap_oce)
      NULLIFY (sab_se)
      NULLIFY (sab_lrc)
      NULLIFY (sab_tbe)
      NULLIFY (sab_core)
      NULLIFY (sab_all)
      NULLIFY (sab_vdw)
      NULLIFY (sab_cn)
      NULLIFY (sab_aux_fit)
      NULLIFY (sab_aux_fit_vs_orb)
      NULLIFY (soo_list)
      NULLIFY (sab_scp)
      NULLIFY (sab_almo)
      NULLIFY (sab_kp)

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      kpoints=kpoints, &
                      distribution_2d=distribution_2d, &
                      local_particles=distribution_1d, &
                      particle_set=particle_set, &
                      molecule_set=molecule_set, &
                      dft_control=dft_control)

      neighbor_list_section => section_vals_get_subs_vals(force_env_section, "DFT%PRINT%NEIGHBOR_LISTS")

      molecule_only = .FALSE.
      IF (PRESENT(molecular)) molecule_only = molecular

      ! This sets the id number of the qs neighbor lists, new lists, means new version
      ! new version implies new sparsity of the matrices

      last_qs_neighbor_list_id_nr = last_qs_neighbor_list_id_nr+1
      CALL set_ks_env(ks_env=ks_env, neighbor_list_id=last_qs_neighbor_list_id_nr)

      CALL get_ks_env(ks_env=ks_env, &
                      sab_orb=sab_orb, &
                      sab_aux_fit=sab_aux_fit, &
                      sab_aux_fit_vs_orb=sab_aux_fit_vs_orb, &
                      sab_aux_fit_asymm=sab_aux_fit_asymm, &
                      sac_ae=sac_ae, &
                      sac_ppl=sac_ppl, &
                      sab_vdw=sab_vdw, &
                      sap_ppnl=sap_ppnl, &
                      sap_oce=sap_oce, &
                      sab_se=sab_se, &
                      sab_lrc=sab_lrc, &
                      sab_tbe=sab_tbe, &
                      sab_core=sab_core, &
                      sab_scp=sab_scp, &
                      sab_all=sab_all, &
                      sab_almo=sab_almo, &
                      sab_kp=sab_kp)

      dokp = (kpoints%nkp > 0)
      nddo = dft_control%qs_control%semi_empirical
      dftb = dft_control%qs_control%dftb
      scptb = dft_control%qs_control%scptb
      almo = dft_control%qs_control%do_almo_scf
      lrigpw = (dft_control%qs_control%method_id == do_method_lrigpw)
      lri_optbas = dft_control%qs_control%lri_optbas

      hfx_sections => section_vals_get_subs_vals(qs_env%input, "DFT%XC%HF")
      CALL section_vals_get(hfx_sections, explicit=do_hfx)

      CALL get_atomic_kind_set(atomic_kind_set, maxatom=maxatom)
      CALL get_qs_kind_set(qs_kind_set, paw_atom_present=paw_atom_present, &
                           gth_potential_present=gth_potential_present, &
                           all_potential_present=all_potential_present)

      CALL section_vals_val_get(qs_env%input, "DFT%SUBCELLS", r_val=subcells)

      ! Allocate work storage

      nkind = SIZE(atomic_kind_set)

      ALLOCATE (orb_present(nkind), aux_fit_present(nkind), aux_present(nkind), &
                default_present(nkind), core_present(nkind))

      ALLOCATE (orb_radius(nkind), aux_fit_radius(nkind), c_radius(nkind), &
                core_radius(nkind), calpha(nkind), zeff(nkind))
      orb_radius(:) = 0.0_dp
      aux_fit_radius(:) = 0.0_dp
      c_radius(:) = 0.0_dp
      core_radius(:) = 0.0_dp
      calpha(:) = 0.0_dp
      zeff(:) = 0.0_dp

      ALLOCATE (pair_radius(nkind, nkind))

      IF (gth_potential_present) THEN
         ALLOCATE (ppl_present(nkind), ppl_radius(nkind))
         ppl_radius = 0.0_dp
         ALLOCATE (ppnl_present(nkind), ppnl_radius(nkind))
         ppnl_radius = 0.0_dp
      END IF

      IF (paw_atom_present) THEN
         ALLOCATE (oce_present(nkind), oce_radius(nkind))
         oce_radius = 0.0_dp
      END IF

      IF (all_potential_present) THEN
         ALLOCATE (all_present(nkind), all_pot_rad(nkind))
         all_pot_rad = 0.0_dp
      END IF

      ! Initialize the local data structures
      ALLOCATE (atom2d(nkind))

      CALL atom2d_build(atom2d, distribution_1d, distribution_2d, atomic_kind_set, &
                        molecule_set, molecule_only, particle_set=particle_set)

      DO ikind = 1, nkind

         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom2d(ikind)%list)

         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set, basis_type="ORB")
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=aux_basis_set, basis_type="AUX")
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=aux_fit_basis_set, basis_type="AUX_FIT")

         CALL get_qs_kind(qs_kind_set(ikind), &
                          paw_proj_set=paw_proj, &
                          paw_atom=paw_atom, &
                          all_potential=all_potential, &
                          gth_potential=gth_potential)

         IF (dftb) THEN
            ! Set the interaction radius for the neighbor lists (DFTB case)
            ! This includes all interactions (orbitals and short range pair potential) except vdW
            CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_atom)
            CALL get_dftb_atom_param(dftb_parameter=dftb_atom, &
                                     cutoff=orb_radius(ikind), &
                                     defined=orb_present(ikind))
         ELSE
            IF (ASSOCIATED(orb_basis_set)) THEN
               orb_present(ikind) = .TRUE.
               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, kind_radius=orb_radius(ikind))
            ELSE
               orb_present(ikind) = .FALSE.
            ENDIF
         ENDIF

         IF (ASSOCIATED(aux_basis_set)) THEN
            aux_present(ikind) = .TRUE.
         ELSE
            aux_present(ikind) = .FALSE.
         ENDIF

         IF (ASSOCIATED(aux_fit_basis_set)) THEN
            aux_fit_present(ikind) = .TRUE.
            CALL get_gto_basis_set(gto_basis_set=aux_fit_basis_set, kind_radius=aux_fit_radius(ikind))
         ELSE
            aux_fit_present(ikind) = .FALSE.
         END IF

         ! core overlap
         CALL get_qs_kind(qs_kind_set(ikind), &
                          alpha_core_charge=calpha(ikind), &
                          core_charge_radius=core_radius(ikind), &
                          zeff=zeff(ikind))
         IF (zeff(ikind) /= 0._dp .AND. calpha(ikind) /= 0._dp) THEN
            core_present(ikind) = .TRUE.
         ELSE
            core_present(ikind) = .FALSE.
         END IF

         ! Pseudopotentials
         IF (gth_potential_present) THEN
            IF (ASSOCIATED(gth_potential)) THEN
               CALL get_potential(potential=gth_potential, &
                                  ppl_present=ppl_present(ikind), &
                                  ppl_radius=ppl_radius(ikind), &
                                  ppnl_present=ppnl_present(ikind), &
                                  ppnl_radius=ppnl_radius(ikind))
            ELSE
               ppl_present(ikind) = .FALSE.
               ppnl_present(ikind) = .FALSE.
            END IF
         END IF

         ! GAPW
         IF (paw_atom_present) THEN
            IF (paw_atom) THEN
               oce_present(ikind) = .TRUE.
               CALL get_paw_proj_set(paw_proj_set=paw_proj, rcprj=oce_radius(ikind))
            ELSE
               oce_present(ikind) = .FALSE.
            END IF
         END IF

         ! Check the presence of an all electron potential
         IF (all_potential_present) THEN
            IF (ASSOCIATED(all_potential)) THEN
               all_present(ikind) = .TRUE.
               CALL get_potential(potential=all_potential, core_charge_radius=all_pot_rad(ikind))
            ELSE
               all_present(ikind) = .FALSE.
               all_pot_rad(ikind) = 0.0_dp
            END IF
         END IF

      END DO

      IF (dokp) THEN
         ! no MIC for kpoints
         mic = .FALSE.
      ELSE
         ! enforce MIC for interaction lists in SE and DFTB
         IF (nddo .OR. dftb .OR. scptb) THEN
            mic = .TRUE.
         ELSE
            mic = .FALSE.
         END IF
      END IF

      ! Build the orbital-orbital overlap neighbor lists
      CALL pair_radius_setup(orb_present, orb_present, orb_radius, orb_radius, pair_radius)
      CALL build_neighbor_lists(sab_orb, particle_set, atom2d, cell, pair_radius, &
                                mic=mic, subcells=subcells, molecular=molecule_only, name="sab_orb")
      CALL set_ks_env(ks_env=ks_env, sab_orb=sab_orb)
      CALL write_neighbor_lists(sab_orb, particle_set, cell, para_env, neighbor_list_section, &
                                "/SAB_ORB", "sab_orb", "ORBITAL ORBITAL")

      ! Build orbital-orbital list containing all the pairs, to be used with
      ! non-symmetric operators. Beware: the cutoff of the orbital-orbital overlap
      ! might be not optimal. It should be verified for each operator.
      IF (.NOT. (nddo .OR. dftb .OR. scptb)) THEN
         CALL build_neighbor_lists(sab_all, particle_set, atom2d, cell, pair_radius, &
                                   mic=mic, symmetric=.FALSE., subcells=subcells, molecular=molecule_only, name="sab_all")
         CALL set_ks_env(ks_env=ks_env, sab_all=sab_all)
      ENDIF

      ! Build the core-core overlap neighbor lists
      IF (.NOT. (nddo .OR. dftb .OR. scptb)) THEN
         CALL pair_radius_setup(core_present, core_present, core_radius, core_radius, pair_radius)
         CALL build_neighbor_lists(sab_core, particle_set, atom2d, cell, pair_radius, mic=mic, subcells=subcells, &
                                   operator_type="PP", name="sab_core")
         CALL set_ks_env(ks_env=ks_env, sab_core=sab_core)
         CALL write_neighbor_lists(sab_core, particle_set, cell, para_env, neighbor_list_section, &
                                   "/SAB_CORE", "sab_core", "CORE CORE")
      ENDIF

      IF (dft_control%do_admm) THEN
         CALL pair_radius_setup(aux_fit_present, aux_fit_present, aux_fit_radius, aux_fit_radius, pair_radius)
         CALL build_neighbor_lists(sab_aux_fit, particle_set, atom2d, cell, pair_radius, &
                                   mic=mic, molecular=molecule_only, subcells=subcells, name="sab_aux_fit")
         CALL build_neighbor_lists(sab_aux_fit_asymm, particle_set, atom2d, cell, pair_radius, &
                                   mic=mic, symmetric=.FALSE., molecular=molecule_only, subcells=subcells, &
                                   name="sab_aux_fit_asymm")
         CALL pair_radius_setup(aux_fit_present, orb_present, aux_fit_radius, orb_radius, pair_radius)
         CALL build_neighbor_lists(sab_aux_fit_vs_orb, particle_set, atom2d, cell, pair_radius, &
                                   mic=mic, symmetric=.FALSE., molecular=molecule_only, subcells=subcells, &
                                   name="sab_aux_fit_vs_orb")

         CALL set_ks_env(ks_env=ks_env, sab_aux_fit=sab_aux_fit)
         CALL set_ks_env(ks_env=ks_env, sab_aux_fit_vs_orb=sab_aux_fit_vs_orb)
         CALL set_ks_env(ks_env=ks_env, sab_aux_fit_asymm=sab_aux_fit_asymm)

         CALL write_neighbor_lists(sab_aux_fit, particle_set, cell, para_env, neighbor_list_section, &
                                   "/SAB_AUX_FIT", "sab_aux_fit", "AUX_FIT_ORBITAL AUX_FIT_ORBITAL")
         CALL write_neighbor_lists(sab_aux_fit_vs_orb, particle_set, cell, para_env, neighbor_list_section, &
                                   "/SAB_AUX_FIT_VS_ORB", "sab_aux_fit_vs_orb", "ORBITAL AUX_FIT_ORBITAL")
      END IF

      IF (dokp) THEN
         ! We try to guess an integration radius for K-points
         ! For non-HFX calculations we use the overlap list
         ! For HFX we use the interaction radius of kinds (ORB or ADMM basis)
         ! plus a range for the operator (some arbitrary guess, should be revisited!)
         IF (do_hfx) THEN
            roperator = 0.0_dp
            CALL section_vals_val_get(hfx_sections, "INTERACTION_POTENTIAL%CUTOFF_RADIUS", &
                                      explicit=explicit)
            IF (explicit) THEN
               CALL section_vals_val_get(hfx_sections, "INTERACTION_POTENTIAL%CUTOFF_RADIUS", &
                                         r_val=roperator)
            ELSE
               CALL section_vals_val_get(hfx_sections, "PERIODIC%NUMBER_OF_SHELLS", &
                                         i_val=numshells)
               numshells = MAX(1, numshells)
               roperator = MAX(plane_distance(1, 0, 0, cell), &
                               plane_distance(0, 1, 0, cell), &
                               plane_distance(0, 0, 1, cell))*numshells
            END IF
            IF (dft_control%do_admm) THEN
               CALL pair_radius_setup(aux_fit_present, aux_fit_present, aux_fit_radius, aux_fit_radius, &
                                      pair_radius)
            ELSE
               CALL pair_radius_setup(orb_present, orb_present, orb_radius, orb_radius, pair_radius)
            END IF
            pair_radius = pair_radius+roperator
         ELSE
            CALL pair_radius_setup(orb_present, orb_present, orb_radius, orb_radius, pair_radius)
         END IF
         CALL build_neighbor_lists(sab_kp, particle_set, atom2d, cell, pair_radius, &
                                   mic=mic, subcells=subcells, molecular=molecule_only, name="sab_kp")
         CALL set_ks_env(ks_env=ks_env, sab_kp=sab_kp)
      END IF

      ! Build orbital GTH-PPL operator overlap list
      IF (gth_potential_present) THEN
         IF (ANY(ppl_present)) THEN
            CALL pair_radius_setup(orb_present, ppl_present, orb_radius, ppl_radius, pair_radius)
            CALL build_neighbor_lists(sac_ppl, particle_set, atom2d, cell, pair_radius, &
                                      subcells=subcells, operator_type="ABC", name="sac_ppl")
            CALL set_ks_env(ks_env=ks_env, sac_ppl=sac_ppl)
            CALL write_neighbor_lists(sac_ppl, particle_set, cell, para_env, neighbor_list_section, &
                                      "/SAC_PPL", "sac_ppl", "ORBITAL GTH-PPL")
         END IF

         IF (ANY(ppnl_present)) THEN
            CALL pair_radius_setup(orb_present, ppnl_present, orb_radius, ppnl_radius, pair_radius)
            CALL build_neighbor_lists(sap_ppnl, particle_set, atom2d, cell, pair_radius, &
                                      subcells=subcells, operator_type="ABBA", name="sap_ppnl")
            CALL set_ks_env(ks_env=ks_env, sap_ppnl=sap_ppnl)
            CALL write_neighbor_lists(sap_ppnl, particle_set, cell, para_env, neighbor_list_section, &
                                      "/SAP_PPNL", "sap_ppnl", "ORBITAL GTH-PPNL")
         END IF
      END IF

      IF (paw_atom_present) THEN
         ! Build orbital-GAPW projector overlap list
         IF (ANY(oce_present)) THEN
            CALL pair_radius_setup(orb_present, oce_present, orb_radius, oce_radius, pair_radius)
            CALL build_neighbor_lists(sap_oce, particle_set, atom2d, cell, pair_radius, &
                                      subcells=subcells, operator_type="ABBA", name="sap_oce")
            CALL set_ks_env(ks_env=ks_env, sap_oce=sap_oce)
            CALL write_neighbor_lists(sap_oce, particle_set, cell, para_env, neighbor_list_section, &
                                      "/SAP_OCE", "sap_oce", "ORBITAL(A) PAW-PRJ")
         END IF
      END IF

      ! Build orbital-ERFC potential list
      IF (all_potential_present .AND. .NOT. (nddo .OR. dftb .OR. scptb)) THEN
         CALL pair_radius_setup(orb_present, all_present, orb_radius, all_pot_rad, pair_radius)
         CALL build_neighbor_lists(sac_ae, particle_set, atom2d, cell, pair_radius, &
                                   subcells=subcells, operator_type="ABC", name="sac_ae")
         CALL set_ks_env(ks_env=ks_env, sac_ae=sac_ae)
         CALL write_neighbor_lists(sac_ae, particle_set, cell, para_env, neighbor_list_section, &
                                   "/SAC_AE", "sac_ae", "ORBITAL ERFC POTENTIAL")
      END IF

      IF (nddo) THEN
         ! Semi-empirical neighbor lists
         default_present = .TRUE.
         c_radius = dft_control%qs_control%se_control%cutoff_cou
         ! Build the neighbor lists for the Hartree terms
         CALL pair_radius_setup(default_present, default_present, c_radius, c_radius, pair_radius)
         IF (dft_control%qs_control%se_control%do_ewald_gks) THEN
            ! Use MIC for the periodic code of GKS
            CALL build_neighbor_lists(sab_se, particle_set, atom2d, cell, pair_radius, mic=mic, &
                                      subcells=subcells, name="sab_se")
         ELSE
            CALL build_neighbor_lists(sab_se, particle_set, atom2d, cell, pair_radius, &
                                      subcells=subcells, name="sab_se")
         END IF
         CALL set_ks_env(ks_env=ks_env, sab_se=sab_se)
         CALL write_neighbor_lists(sab_se, particle_set, cell, para_env, neighbor_list_section, &
                                   "/SAB_SE", "sab_se", "HARTREE INTERACTIONS")

         ! If requested build the SE long-range correction neighbor list
         IF ((dft_control%qs_control%se_control%do_ewald) .AND. &
             (dft_control%qs_control%se_control%integral_screening /= do_se_IS_slater)) THEN
            c_radius = dft_control%qs_control%se_control%cutoff_lrc
            CALL pair_radius_setup(default_present, default_present, c_radius, c_radius, pair_radius)
            CALL build_neighbor_lists(sab_lrc, particle_set, atom2d, cell, pair_radius, &
                                      subcells=subcells, name="sab_lrc")
            CALL set_ks_env(ks_env=ks_env, sab_lrc=sab_lrc)
            CALL write_neighbor_lists(sab_lrc, particle_set, cell, para_env, neighbor_list_section, &
                                      "/SAB_LRC", "sab_lrc", "SE LONG-RANGE CORRECTION")
         END IF
      END IF

      IF (dftb) THEN
         ! Build the neighbor lists for the DFTB Ewald methods
         IF (dft_control%qs_control%dftb_control%do_ewald) THEN
            CALL get_qs_env(qs_env=qs_env, ewald_env=ewald_env)
            CALL ewald_env_get(ewald_env, alpha=alpha)
            c_radius = 0.5_dp*SQRT(-LOG(3.5_dp*alpha**3*1.e-12_dp))/alpha
            CALL pair_radius_setup(orb_present, orb_present, c_radius, c_radius, pair_radius)
            CALL build_neighbor_lists(sab_tbe, particle_set, atom2d, cell, pair_radius, mic=mic, &
                                      subcells=subcells, name="sab_tbe")
            CALL set_ks_env(ks_env=ks_env, sab_tbe=sab_tbe)
         END IF

         ! Build the neighbor lists for the DFTB vdW pair potential
         IF (dft_control%qs_control%dftb_control%dispersion) THEN
            IF (dft_control%qs_control%dftb_control%dispersion_type == dispersion_uff) THEN
               DO ikind = 1, nkind
                  CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_atom)
                  CALL get_dftb_atom_param(dftb_parameter=dftb_atom, &
                                           rcdisp=c_radius(ikind))
               END DO
               default_present = .TRUE.
               CALL pair_radius_setup(default_present, default_present, c_radius, c_radius, pair_radius)
               CALL build_neighbor_lists(sab_vdw, particle_set, atom2d, cell, pair_radius, &
                                         subcells=subcells, name="sab_vdw")
               CALL set_ks_env(ks_env=ks_env, sab_vdw=sab_vdw)
            END IF
         END IF
      END IF

      ! Build the neighbor lists for the vdW pair potential
      CALL get_qs_env(qs_env=qs_env, dispersion_env=dispersion_env)
      sab_vdw => dispersion_env%sab_vdw
      sab_cn => dispersion_env%sab_cn
      IF (dispersion_env%type == xc_vdw_fun_pairpot) THEN
         c_radius(:) = dispersion_env%rc_disp
         default_present = .TRUE. !include all atoms in vdW (even without basis)
         CALL pair_radius_setup(default_present, default_present, c_radius, c_radius, pair_radius)
         CALL build_neighbor_lists(sab_vdw, particle_set, atom2d, cell, pair_radius, &
                                   subcells=subcells, operator_type="PP", name="sab_vdw")
         dispersion_env%sab_vdw => sab_vdw

         IF (dispersion_env%pp_type == vdw_pairpot_dftd3 .OR. &
             dispersion_env%pp_type == vdw_pairpot_dftd3bj) THEN
            ! Build the neighbor lists for coordination numbers as needed by the DFT-D3 method
            DO ikind = 1, nkind
               CALL get_atomic_kind(atomic_kind_set(ikind), z=zat)
               c_radius(ikind) = 4._dp*ptable(zat)%covalent_radius*bohr
            END DO
            CALL pair_radius_setup(default_present, default_present, c_radius, c_radius, pair_radius)
            CALL build_neighbor_lists(sab_cn, particle_set, atom2d, cell, pair_radius, &
                                      subcells=subcells, operator_type="PP", name="sab_cn")
            dispersion_env%sab_cn => sab_cn
         END IF
      END IF

      IF (scptb) THEN
         ! Build the neighbor lists for the Core term
         default_present = .FALSE.
         DO ikind = 1, nkind
            CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
            CALL get_scptb_parameter(scptb_kind, rcpair=c_radius(ikind), defined=default_present(ikind))
         END DO
         CALL pair_radius_setup(default_present, default_present, c_radius, c_radius, pair_radius)
         CALL build_neighbor_lists(sab_core, particle_set, atom2d, cell, pair_radius, mic=mic, &
                                   subcells=subcells, name="sab_core")
         CALL set_ks_env(ks_env=ks_env, sab_core=sab_core)
         CALL write_neighbor_lists(sab_core, particle_set, cell, para_env, neighbor_list_section, &
                                   "/SAB_CORE", "sab_core", "CORE INTERACTIONS")
         ! Build the neighbor lists for the SCPTB Ewald methods
         IF (dft_control%qs_control%scptb_control%do_ewald) THEN
            ! Ewald with alpha
            CALL get_qs_env(qs_env=qs_env, ewald_env=ewald_env)
            CALL ewald_env_get(ewald_env, alpha=alpha)
            c_radius = 0.5_dp*SQRT(-LOG(3.5_dp*alpha**3*1.e-12_dp))/alpha
            CALL pair_radius_setup(default_present, default_present, c_radius, c_radius, pair_radius)
            CALL build_neighbor_lists(sab_tbe, particle_set, atom2d, cell, pair_radius, mic=mic, &
                                      subcells=subcells, name="sab_tbe")
            CALL set_ks_env(ks_env=ks_env, sab_tbe=sab_tbe)
            ! Correction term for Gaussians
            DO ikind = 1, nkind
               CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
               CALL get_scptb_parameter(scptb_kind, ag=alpha)
               alpha = SQRT(alpha)
               c_radius(ikind) = 0.5_dp*SQRT(-LOG(3.5_dp*alpha**3*1.e-12_dp))/alpha
            END DO
            CALL pair_radius_setup(default_present, default_present, c_radius, c_radius, pair_radius)
            CALL build_neighbor_lists(sab_scp, particle_set, atom2d, cell, pair_radius, mic=mic, &
                                      subcells=subcells, name="sab_scp")
            CALL set_ks_env(ks_env=ks_env, sab_scp=sab_scp)
         END IF
      END IF
      IF (lrigpw .OR. lri_optbas) THEN
         ! set neighborlists in lri_env environment
         CALL pair_radius_setup(orb_present, orb_present, orb_radius, orb_radius, pair_radius)
         soo_list => qs_env%lri_env%soo_list
         CALL build_neighbor_lists(soo_list, particle_set, atom2d, cell, pair_radius, &
                                   subcells=subcells, name="soo_list")
         qs_env%lri_env%soo_list => soo_list
         CALL write_neighbor_lists(soo_list, particle_set, cell, para_env, neighbor_list_section, &
                                   "/SOO_LIST", "soo_list", "ORBITAL ORBITAL (RI)")
      END IF

      ! Build the neighbor lists for the ALMO delocalization
      IF (almo) THEN
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), rcov=almo_rcov, rvdw=almo_rvdw)
            ! multiply the radius by some hard-coded number
            c_radius(ikind) = MAX(almo_rcov, almo_rvdw)*bohr* &
                              almo_max_cutoff_multiplier
         END DO
         default_present = .TRUE. !include all atoms (even without basis)
         CALL pair_radius_setup(default_present, default_present, c_radius, c_radius, pair_radius)
         CALL build_neighbor_lists(sab_almo, particle_set, atom2d, cell, pair_radius, &
                                   subcells=subcells, operator_type="PP", name="sab_almo")
         CALL set_ks_env(ks_env=ks_env, sab_almo=sab_almo)
      ENDIF

      ! Print particle distribution
      print_key_path = "PRINT%DISTRIBUTION"
      IF (BTEST(cp_print_key_should_output(logger%iter_info, force_env_section, &
                                           print_key_path), &
                cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger=logger, &
                                   basis_section=force_env_section, &
                                   print_key_path=print_key_path, &
                                   extension=".out")
         CALL write_neighbor_distribution(sab_orb, qs_kind_set, iw, para_env)
         CALL cp_print_key_finished_output(unit_nr=iw, &
                                           logger=logger, &
                                           basis_section=force_env_section, &
                                           print_key_path=print_key_path)
      END IF

      ! Release work storage
      CALL atom2d_cleanup(atom2d)

      DEALLOCATE (atom2d)
      DEALLOCATE (orb_present, default_present, core_present)
      DEALLOCATE (orb_radius, aux_fit_radius, c_radius, core_radius)
      DEALLOCATE (calpha, zeff)
      DEALLOCATE (pair_radius)
      IF (gth_potential_present) THEN
         DEALLOCATE (ppl_present, ppl_radius)
         DEALLOCATE (ppnl_present, ppnl_radius)
      END IF
      IF (paw_atom_present) THEN
         DEALLOCATE (oce_present, oce_radius)
      ENDIF
      IF (all_potential_present) THEN
         DEALLOCATE (all_present, all_pot_rad)
      END IF

      CALL timestop(handle)

   END SUBROUTINE build_qs_neighbor_lists

! **************************************************************************************************
!> \brief   Build simple pair neighbor lists.
!> \param ab_list ...
!> \param particle_set ...
!> \param atom ...
!> \param cell ...
!> \param pair_radius ...
!> \param subcells ...
!> \param mic ...
!> \param symmetric ...
!> \param molecular ...
!> \param subset_of_mol ...
!> \param current_subset ...
!> \param operator_type ...
!> \param name ...
!> \date    20.03.2002
!> \par History
!>          - Major refactoring (25.07.2010,jhu)
!> \author  MK
!> \version 2.0
! **************************************************************************************************
   SUBROUTINE build_neighbor_lists(ab_list, particle_set, atom, cell, pair_radius, subcells, &
                                   mic, symmetric, molecular, subset_of_mol, current_subset, &
                                   operator_type, name)

      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: ab_list
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(local_atoms_type), DIMENSION(:), INTENT(IN)   :: atom
      TYPE(cell_type), POINTER                           :: cell
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: pair_radius
      REAL(dp), INTENT(IN)                               :: subcells
      LOGICAL, INTENT(IN), OPTIONAL                      :: mic, symmetric, molecular
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: subset_of_mol
      INTEGER, OPTIONAL                                  :: current_subset
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: operator_type
      CHARACTER(LEN=*), INTENT(IN)                       :: name

      CHARACTER(len=*), PARAMETER :: routineN = 'build_neighbor_lists', &
         routineP = moduleN//':'//routineN

      TYPE local_lists
         INTEGER, DIMENSION(:), POINTER           :: list
      END TYPE local_lists
      INTEGER :: atom_a, atom_b, handle, i, iab, iatom, iatom_local, &
                 iatom_subcell, icell, ikind, j, jatom, jatom_local, jcell, jkind, k, &
                 kcell, maxat, mol_a, mol_b, nkind, otype, natom
      INTEGER, DIMENSION(3)                    :: cell_b, ncell, &
                                                  nsubcell, periodic
      INTEGER, DIMENSION(:), POINTER           :: index_list
      LOGICAL                                  :: include_ab, my_mic, &
                                                  my_molecular, my_symmetric
      LOGICAL, ALLOCATABLE, DIMENSION(:)       :: pres_a, pres_b
      REAL(dp)                                 :: rab2, rab2_max, rab_max, rabm, deth, subcell_scale
      REAL(dp), DIMENSION(3)                   :: r, rab, ra, rb, sab_max, sb, &
                                                  sb_pbc, sb_min, sb_max, rab_pbc, pd, sab_max_guard
      INTEGER, ALLOCATABLE, DIMENSION(:)       :: nlista, nlistb
      TYPE(local_lists), DIMENSION(:), POINTER :: lista, listb
      TYPE(neighbor_list_p_type), &
         ALLOCATABLE, DIMENSION(:)              :: kind_a
      TYPE(neighbor_list_set_type), POINTER    :: neighbor_list_set
      TYPE(subcell_type), DIMENSION(:, :, :), &
         POINTER                                :: subcell
      REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE :: r_pbc

      CALL timeset(routineN//"_"//TRIM(name), handle)

      ! input options
      my_mic = .FALSE.
      IF (PRESENT(mic)) my_mic = mic
      my_symmetric = .TRUE.
      IF (PRESENT(symmetric)) my_symmetric = symmetric
      my_molecular = .FALSE.
      ! if we have a molecular NL, MIC has to be used
      IF (PRESENT(molecular)) my_molecular = molecular
      ! check for operator types
      IF (PRESENT(operator_type)) THEN
         SELECT CASE (operator_type)
         CASE ("AB")
            otype = 1 ! simple overlap
         CASE ("ABC")
            otype = 2 ! for three center operators
            IF(.NOT.(.NOT. my_molecular))CALL cp__a("qs_neighbor_lists.F",964)
            my_symmetric = .FALSE.
         CASE ("ABBA")
            otype = 3 ! for separable nonlocal operators
            my_symmetric = .FALSE.
         CASE ("PP")
            otype = 4 ! simple atomic pair potential list
         CASE default
            CALL cp__b("qs_neighbor_lists.F",972,"")
         END SELECT
      ELSE
         ! default is a simple AB neighbor list
         otype = 1
      END IF

      ! Deallocate the old neighbor list structure
      IF (ASSOCIATED(ab_list)) THEN
         DO iab = 1, SIZE(ab_list)
            CALL deallocate_neighbor_list_set(ab_list(iab)%neighbor_list_set)
         END DO
         DEALLOCATE (ab_list)
      END IF
      nkind = SIZE(atom)
      ! Allocate and initialize the new neighbor list structure
      ALLOCATE (ab_list(nkind*nkind))
      DO iab = 1, SIZE(ab_list)
         NULLIFY (ab_list(iab)%neighbor_list_set)
      END DO

      ! Allocate and initialize the kind availability
      ALLOCATE (pres_a(nkind), pres_b(nkind))
      DO ikind = 1, nkind
         pres_a(ikind) = ANY(pair_radius(ikind, :) > 0._dp)
         pres_b(ikind) = ANY(pair_radius(:, ikind) > 0._dp)
      END DO

      ! create a copy of the pbc'ed coordinates
      natom = SIZE(particle_set)
      ALLOCATE (r_pbc(3, natom))
      DO i = 1, natom
         r_pbc(1:3, i) = pbc(particle_set(i)%r(1:3), cell)
      ENDDO

      ! setup the local lists of atoms
      maxat = 0
      DO ikind = 1, nkind
         maxat = MAX(maxat, SIZE(atom(ikind)%list))
      END DO
      ALLOCATE (index_list(maxat))
      DO i = 1, maxat
         index_list(i) = i
      END DO
      ALLOCATE (lista(nkind), listb(nkind), nlista(nkind), nlistb(nkind))
      nlista = 0
      nlistb = 0
      DO ikind = 1, nkind
         NULLIFY (lista(ikind)%list, listb(ikind)%list)
         SELECT CASE (otype)
         CASE (1)
            IF (ASSOCIATED(atom(ikind)%list_local_a_index)) THEN
               lista(ikind)%list => atom(ikind)%list_local_a_index
               nlista(ikind) = SIZE(lista(ikind)%list)
            END IF
            IF (ASSOCIATED(atom(ikind)%list_local_b_index)) THEN
               listb(ikind)%list => atom(ikind)%list_local_b_index
               nlistb(ikind) = SIZE(listb(ikind)%list)
            END IF
         CASE (2)
            IF (ASSOCIATED(atom(ikind)%list_local_a_index)) THEN
               lista(ikind)%list => atom(ikind)%list_local_a_index
               nlista(ikind) = SIZE(lista(ikind)%list)
            END IF
            nlistb(ikind) = SIZE(atom(ikind)%list)
            listb(ikind)%list => index_list
         CASE (3)
            CALL combine_lists(lista(ikind)%list, nlista(ikind), ikind, atom)
            nlistb(ikind) = SIZE(atom(ikind)%list)
            listb(ikind)%list => index_list
         CASE (4)
            nlista(ikind) = SIZE(atom(ikind)%list_1d)
            lista(ikind)%list => atom(ikind)%list_1d
            nlistb(ikind) = SIZE(atom(ikind)%list)
            listb(ikind)%list => index_list
         CASE default
            CALL cp__b("qs_neighbor_lists.F",1048,"")
         END SELECT
      END DO

      ! Determine max. number of local atoms
      maxat = 0
      DO ikind = 1, nkind
         maxat = MAX(maxat, nlista(ikind), nlistb(ikind))
      END DO
      ALLOCATE (kind_a(2*maxat))

      ! Load informations about the simulation cell
      CALL get_cell(cell=cell, periodic=periodic, deth=deth)

      ! Loop over all atomic kind pairs
      DO ikind = 1, nkind
         IF (.NOT. pres_a(ikind)) CYCLE

         DO jkind = 1, nkind
            IF (.NOT. pres_b(jkind)) CYCLE

            iab = ikind+nkind*(jkind-1)

            ! Calculate the square of the maximum interaction distance
            IF (pair_radius(ikind, jkind) <= 0._dp) CYCLE
            rab_max = pair_radius(ikind, jkind)
            IF (otype == 3) THEN
               ! Calculate the square of the maximum interaction distance
               ! for sac_max / ncell this must be the maximum over all kinds
               ! to be correct for three center terms involving different kinds
               rabm = MAXVAL(pair_radius(:, jkind))
            ELSE
               rabm = rab_max
            END IF
            rab2_max = rabm*rabm

            pd(1) = plane_distance(1, 0, 0, cell)
            pd(2) = plane_distance(0, 1, 0, cell)
            pd(3) = plane_distance(0, 0, 1, cell)

            sab_max = rabm/pd
            sab_max_guard = 15.0_dp/pd

            ! It makes sense to have fewer subcells for larger systems
            subcell_scale = ((125.0_dp**3)/deth)**(1.0_dp/6.0_dp)

            ! guess the number of subcells for optimal performance, guard against crazy stuff triggered by very small rabm
            nsubcell(:) = INT(MAX(1.0_dp, MIN(0.5_dp*subcells*subcell_scale/sab_max(:), &
                                              0.5_dp*subcells*subcell_scale/sab_max_guard(:))))

            ! number of image cells to be considered
            ncell(:) = (INT(sab_max(:))+1)*periodic(:)

            CALL allocate_neighbor_list_set(neighbor_list_set=ab_list(iab)%neighbor_list_set, &
                                            r_max=rabm, symmetric=my_symmetric, mic=my_mic, molecular=my_molecular)
            neighbor_list_set => ab_list(iab)%neighbor_list_set

            DO iatom_local = 1, nlista(ikind)
               iatom = lista(ikind)%list(iatom_local)
               atom_a = atom(ikind)%list(iatom)
               CALL add_neighbor_list(neighbor_list_set=neighbor_list_set, &
                                      atom=atom_a, &
                                      neighbor_list=kind_a(iatom_local)%neighbor_list)
            END DO

            CALL allocate_subcell(subcell, nsubcell)
            DO iatom_local = 1, nlista(ikind)
               iatom = lista(ikind)%list(iatom_local)
               atom_a = atom(ikind)%list(iatom)
               r = r_pbc(:, atom_a)
               CALL give_ijk_subcell(r, i, j, k, cell, nsubcell)
               subcell(i, j, k)%natom = subcell(i, j, k)%natom+1
            END DO
            DO k = 1, nsubcell(3)
               DO j = 1, nsubcell(2)
                  DO i = 1, nsubcell(1)
                     maxat = subcell(i, j, k)%natom+subcell(i, j, k)%natom/10
                     ALLOCATE (subcell(i, j, k)%atom_list(maxat))
                     subcell(i, j, k)%natom = 0
                  END DO
               END DO
            END DO
            DO iatom_local = 1, nlista(ikind)
               iatom = lista(ikind)%list(iatom_local)
               atom_a = atom(ikind)%list(iatom)
               r = r_pbc(:, atom_a)
               CALL give_ijk_subcell(r, i, j, k, cell, nsubcell)
               subcell(i, j, k)%natom = subcell(i, j, k)%natom+1
               subcell(i, j, k)%atom_list(subcell(i, j, k)%natom) = iatom_local
            END DO

            DO jatom_local = 1, nlistb(jkind)
               jatom = listb(jkind)%list(jatom_local)
               atom_b = atom(jkind)%list(jatom)
               IF (my_molecular) THEN
                  mol_b = atom(jkind)%list_b_mol(jatom_local)
                  IF (PRESENT(subset_of_mol)) THEN
                     IF (subset_of_mol(mol_b) .NE. current_subset) CYCLE
                  END IF
               END IF
               r = r_pbc(:, atom_b)
               CALL real_to_scaled(sb_pbc(:), r(:), cell)

               loop2_kcell: DO kcell = -ncell(3), ncell(3)
                  sb(3) = sb_pbc(3)+REAL(kcell, dp)
                  sb_min(3) = sb(3)-sab_max(3)
                  sb_max(3) = sb(3)+sab_max(3)
                  IF (periodic(3) /= 0) THEN
                     IF (sb_min(3) >= 0.5_dp) EXIT loop2_kcell
                     IF (sb_max(3) < -0.5_dp) CYCLE loop2_kcell
                  END IF
                  cell_b(3) = kcell

                  loop2_jcell: DO jcell = -ncell(2), ncell(2)
                     sb(2) = sb_pbc(2)+REAL(jcell, dp)
                     sb_min(2) = sb(2)-sab_max(2)
                     sb_max(2) = sb(2)+sab_max(2)
                     IF (periodic(2) /= 0) THEN
                        IF (sb_min(2) >= 0.5_dp) EXIT loop2_jcell
                        IF (sb_max(2) < -0.5_dp) CYCLE loop2_jcell
                     END IF
                     cell_b(2) = jcell

                     loop2_icell: DO icell = -ncell(1), ncell(1)
                        sb(1) = sb_pbc(1)+REAL(icell, dp)
                        sb_min(1) = sb(1)-sab_max(1)
                        sb_max(1) = sb(1)+sab_max(1)
                        IF (periodic(1) /= 0) THEN
                           IF (sb_min(1) >= 0.5_dp) EXIT loop2_icell
                           IF (sb_max(1) < -0.5_dp) CYCLE loop2_icell
                        END IF
                        cell_b(1) = icell

                        CALL scaled_to_real(rb, sb, cell)

                        loop_k: DO k = 1, nsubcell(3)
                           loop_j: DO j = 1, nsubcell(2)
                              loop_i: DO i = 1, nsubcell(1)

                                 ! FIXME for non-periodic systems, the whole subcell trick is skipped
                                 ! yielding a Natom**2 pair list build.
                                 IF (periodic(3) /= 0) THEN
                                    IF (sb_max(3) < subcell(i, j, k)%s_min(3)) EXIT loop_k
                                    IF (sb_min(3) >= subcell(i, j, k)%s_max(3)) CYCLE loop_k
                                 END IF

                                 IF (periodic(2) /= 0) THEN
                                    IF (sb_max(2) < subcell(i, j, k)%s_min(2)) EXIT loop_j
                                    IF (sb_min(2) >= subcell(i, j, k)%s_max(2)) CYCLE loop_j
                                 END IF

                                 IF (periodic(1) /= 0) THEN
                                    IF (sb_max(1) < subcell(i, j, k)%s_min(1)) EXIT loop_i
                                    IF (sb_min(1) >= subcell(i, j, k)%s_max(1)) CYCLE loop_i
                                 END IF

                                 IF (subcell(i, j, k)%natom == 0) CYCLE

                                 DO iatom_subcell = 1, subcell(i, j, k)%natom
                                    iatom_local = subcell(i, j, k)%atom_list(iatom_subcell)
                                    iatom = lista(ikind)%list(iatom_local)
                                    atom_a = atom(ikind)%list(iatom)
                                    IF (my_molecular) THEN
                                       mol_a = atom(ikind)%list_a_mol(iatom_local)
                                       IF (mol_a /= mol_b) CYCLE
                                    END IF
                                    IF (my_symmetric) THEN
                                       IF (atom_a > atom_b) THEN
                                          include_ab = (MODULO(atom_a+atom_b, 2) /= 0)
                                       ELSE
                                          include_ab = (MODULO(atom_a+atom_b, 2) == 0)
                                       END IF
                                    ELSE
                                       include_ab = .TRUE.
                                    END IF
                                    IF (include_ab) THEN
                                       ra(:) = r_pbc(:, atom_a)
                                       rab(:) = rb(:)-ra(:)
                                       rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
                                       IF (rab2 < rab2_max) THEN
                                          include_ab = .TRUE.
                                          IF (my_mic) THEN
                                             ! only if rab is minimum image the pair will be included
                                             ! ideally the range of the pair list is < L/2 so that this never triggers
                                             rab_pbc(:) = pbc(rab(:), cell)
                                             IF (SUM((rab_pbc-rab)**2) > EPSILON(1.0_dp)) THEN
                                                include_ab = .FALSE.
                                             ENDIF
                                          ENDIF
                                          IF (include_ab) THEN
                                             CALL add_neighbor_node( &
                                                neighbor_list=kind_a(iatom_local)%neighbor_list, &
                                                neighbor=atom_b, &
                                                cell=cell_b, &
                                                r=rab, &
                                                nkind=nkind)
                                          ENDIF
                                       END IF
                                    END IF
                                 END DO

                              END DO loop_i
                           END DO loop_j
                        END DO loop_k

                     END DO loop2_icell
                  END DO loop2_jcell
               END DO loop2_kcell

            END DO

            CALL deallocate_subcell(subcell)

         END DO
      END DO

      SELECT CASE (otype)
      CASE (1:2, 4)
      CASE (3)
         DO ikind = 1, nkind
            DEALLOCATE (lista(ikind)%list)
         END DO
      CASE default
         CALL cp__b("qs_neighbor_lists.F",1271,"")
      END SELECT
      DEALLOCATE (kind_a, pres_a, pres_b, lista, listb, nlista, nlistb)
      DEALLOCATE (index_list)
      DEALLOCATE (r_pbc)

      CALL timestop(handle)

   END SUBROUTINE build_neighbor_lists

! **************************************************************************************************
!> \brief Build a neighborlist
!> \param ab_list ...
!> \param basis_set_a ...
!> \param basis_set_b ...
!> \param qs_env ...
!> \param mic ...
!> \param symmetric ...
!> \param molecular ...
!> \param operator_type ...
!> \date    14.03.2016
!> \author  JGH
! **************************************************************************************************
   SUBROUTINE setup_neighbor_list(ab_list, basis_set_a, basis_set_b, qs_env, &
                                  mic, symmetric, molecular, operator_type)

      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: ab_list
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_set_a
      TYPE(gto_basis_set_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: basis_set_b
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN), OPTIONAL                      :: mic, symmetric, molecular
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: operator_type

      CHARACTER(len=*), PARAMETER :: routineN = 'setup_neighbor_list', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=4)                                   :: otype
      INTEGER                                            :: ikind, nkind
      LOGICAL                                            :: my_mic, my_molecular, my_symmetric
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: a_present, b_present
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: a_radius, b_radius
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: pair_radius
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(distribution_1d_type), POINTER                :: distribution_1d
      TYPE(distribution_2d_type), POINTER                :: distribution_2d
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER  :: basis_a, basis_b
      TYPE(gto_basis_set_type), POINTER                  :: abas, bbas
      TYPE(local_atoms_type), ALLOCATABLE, DIMENSION(:)  :: atom2d
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set

      basis_a => basis_set_a
      IF (PRESENT(basis_set_b)) THEN
         basis_b => basis_set_b
         my_symmetric = .FALSE.
      ELSE
         basis_b => basis_set_a
         my_symmetric = .TRUE.
      END IF
      IF (PRESENT(symmetric)) my_symmetric = symmetric

      IF (PRESENT(mic)) THEN
         my_mic = mic
      ELSE
         my_mic = .FALSE.
      END IF

      IF (PRESENT(molecular)) THEN
         my_molecular = molecular
      ELSE
         my_molecular = .FALSE.
      END IF

      IF (PRESENT(operator_type)) THEN
         otype = operator_type
      ELSE
         ! default is a simple AB neighbor list
         otype = "AB"
      END IF

      nkind = SIZE(basis_a)
      ALLOCATE (a_present(nkind), b_present(nkind))
      a_present = .FALSE.
      b_present = .FALSE.
      ALLOCATE (a_radius(nkind), b_radius(nkind))
      a_radius = 0.0_dp
      b_radius = 0.0_dp
      DO ikind = 1, nkind
         IF (ASSOCIATED(basis_a(ikind)%gto_basis_set)) THEN
            a_present(ikind) = .TRUE.
            abas => basis_a(ikind)%gto_basis_set
            CALL get_gto_basis_set(gto_basis_set=abas, kind_radius=a_radius(ikind))
         END IF
         IF (ASSOCIATED(basis_b(ikind)%gto_basis_set)) THEN
            b_present(ikind) = .TRUE.
            bbas => basis_b(ikind)%gto_basis_set
            CALL get_gto_basis_set(gto_basis_set=bbas, kind_radius=b_radius(ikind))
         END IF
      END DO

      ALLOCATE (pair_radius(nkind, nkind))
      pair_radius = 0.0_dp
      CALL pair_radius_setup(a_present, b_present, a_radius, b_radius, pair_radius)

      CALL get_qs_env(qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      cell=cell, &
                      distribution_2d=distribution_2d, &
                      local_particles=distribution_1d, &
                      particle_set=particle_set, &
                      molecule_set=molecule_set)

      ALLOCATE (atom2d(nkind))
      CALL atom2d_build(atom2d, distribution_1d, distribution_2d, atomic_kind_set, &
                        molecule_set, my_molecular, particle_set=particle_set)
      CALL build_neighbor_lists(ab_list, particle_set, atom2d, cell, pair_radius, &
                                mic=my_mic, symmetric=my_symmetric, molecular=my_molecular, &
                                subcells=2.0_dp, name="AUX_NL")

      CALL atom2d_cleanup(atom2d)

      DEALLOCATE (a_present, b_present, a_radius, b_radius, pair_radius, atom2d)

   END SUBROUTINE setup_neighbor_list

! **************************************************************************************************
!> \brief ...
!> \param list ...
!> \param n ...
!> \param ikind ...
!> \param atom ...
! **************************************************************************************************
   SUBROUTINE combine_lists(list, n, ikind, atom)
      INTEGER, DIMENSION(:), POINTER                     :: list
      INTEGER, INTENT(OUT)                               :: n
      INTEGER, INTENT(IN)                                :: ikind
      TYPE(local_atoms_type), DIMENSION(:), INTENT(IN)   :: atom

      CHARACTER(len=*), PARAMETER :: routineN = 'combine_lists', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ib, na, nb
      INTEGER, DIMENSION(:), POINTER                     :: lista, listb

      IF(.NOT.(.NOT. ASSOCIATED(list)))CALL cp__a("qs_neighbor_lists.F",1417)

      lista => atom(ikind)%list_local_a_index
      listb => atom(ikind)%list_local_b_index

      IF (ASSOCIATED(lista)) THEN
         na = SIZE(lista)
      ELSE
         na = 0
      END IF

      IF (ASSOCIATED(listb)) THEN
         nb = SIZE(listb)
      ELSE
         nb = 0
      END IF

      ALLOCATE (list(na+nb))

      n = na
      IF (na .GT. 0) list(1:na) = lista(1:na)
      IF (nb .GT. 0) THEN
         loopb: DO ib = 1, nb
            DO i = 1, na
               IF (listb(ib) == list(i)) CYCLE loopb
            END DO
            n = n+1
            list(n) = listb(ib)
         END DO loopb
      ENDIF
   END SUBROUTINE combine_lists

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param present_a ...
!> \param present_b ...
!> \param radius_a ...
!> \param radius_b ...
!> \param pair_radius ...
! **************************************************************************************************
   SUBROUTINE pair_radius_setup(present_a, present_b, radius_a, radius_b, pair_radius)
      LOGICAL, DIMENSION(:), INTENT(IN)                  :: present_a, present_b
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: radius_a, radius_b
      REAL(dp), DIMENSION(:, :), INTENT(OUT)             :: pair_radius

      CHARACTER(len=*), PARAMETER :: routineN = 'pair_radius_setup', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, nkind

      nkind = SIZE(present_a)

      pair_radius = 0._dp

      DO i = 1, nkind
         IF (.NOT. present_a(i)) CYCLE
         DO j = 1, nkind
            IF (.NOT. present_b(j)) CYCLE
            pair_radius(i, j) = radius_a(i)+radius_b(j)
         END DO
      END DO

   END SUBROUTINE pair_radius_setup

! **************************************************************************************************
!> \brief   Print the distribution of the simple pair neighbor list.
!> \param ab ...
!> \param qs_kind_set ...
!> \param output_unit ...
!> \param para_env ...
!> \date    19.06.2003
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_neighbor_distribution(ab, qs_kind_set, output_unit, para_env)
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: ab
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER, INTENT(in)                                :: output_unit
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'write_neighbor_distribution', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: full_output = .FALSE.

      INTEGER                                            :: group, handle, ikind, inode, ipe, jkind, &
                                                            mype, n, nkind, nnode, npe
      INTEGER(int_8)                                     :: nblock_max, nblock_sum, nelement_max, &
                                                            nelement_sum, tmp(2)
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: nblock, nelement, nnsgf
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator

      CALL timeset(routineN, handle)
      group = para_env%group
      mype = para_env%mepos+1
      npe = para_env%num_pe

      ! Allocate work storage
      ALLOCATE (nblock(npe), nelement(npe))
      nblock(:) = 0
      nelement(:) = 0
      nkind = SIZE(qs_kind_set)
      ALLOCATE (nnsgf(nkind))
      nnsgf = 1
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
         IF (ASSOCIATED(orb_basis_set)) THEN
            CALL get_gto_basis_set(gto_basis_set=orb_basis_set, nsgf=nnsgf(ikind))
         END IF
      END DO

      CALL neighbor_list_iterator_create(nl_iterator, ab)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, inode=inode, nnode=nnode)
         IF (inode == 1) THEN
            n = nnsgf(ikind)*nnsgf(jkind)
            nblock(mype) = nblock(mype)+nnode
            nelement(mype) = nelement(mype)+n*nnode
         END IF
      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      IF (full_output) THEN
         ! XXXXXXXX should gather/scatter this on ionode
         CALL mp_sum(nblock, group)
         CALL mp_sum(nelement, group)

         nblock_sum = SUM(INT(nblock, KIND=int_8))
         nelement_sum = SUM(INT(nelement, KIND=int_8))
      ELSE
         nblock_sum = nblock(mype)
         nblock_max = nblock(mype)
         nelement_sum = nelement(mype)
         nelement_max = nelement(mype)
         tmp = (/nblock_sum, nelement_sum/)
         CALL mp_sum(tmp, group)
         nblock_sum = tmp(1); nelement_sum = tmp(2)
         tmp = (/nblock_max, nelement_max/)
         CALL mp_max(tmp, group)
         nblock_max = tmp(1); nelement_max = tmp(2)
      ENDIF

      IF (output_unit > 0) THEN
         IF (full_output) THEN
            WRITE (UNIT=output_unit, &
                   FMT="(/,/,T2,A,/,/,T3,A,/,/,(T4,I6,T27,I10,T55,I10))") &
               "DISTRIBUTION OF THE NEIGHBOR LISTS", &
               "Process   Number of particle pairs   Number of matrix elements", &
               (ipe-1, nblock(ipe), nelement(ipe), ipe=1, npe)
            WRITE (UNIT=output_unit, FMT="(/,T7,A3,T27,I10,T55,I10)") &
               "Sum", SUM(nblock), SUM(nelement)
         ELSE
            WRITE (UNIT=output_unit, FMT="(/,T2,A)") "DISTRIBUTION OF THE NEIGHBOR LISTS"
            WRITE (UNIT=output_unit, FMT="(T15,A,T68,I13)") "Total number of particle pairs:", nblock_sum
            WRITE (UNIT=output_unit, FMT="(T15,A,T68,I13)") "Total number of matrix elements:", nelement_sum
            WRITE (UNIT=output_unit, FMT="(T15,A,T68,I13)") "Average number of particle pairs:", (nblock_sum+npe-1)/npe
            WRITE (UNIT=output_unit, FMT="(T15,A,T68,I13)") "Maximum number of particle pairs:", nblock_max
            WRITE (UNIT=output_unit, FMT="(T15,A,T68,I13)") "Average number of matrix element:", (nelement_sum+npe-1)/npe
            WRITE (UNIT=output_unit, FMT="(T15,A,T68,I13)") "Maximum number of matrix elements:", nelement_max
         ENDIF
      END IF

      ! Release work storage

      DEALLOCATE (nblock, nelement, nnsgf)

      CALL timestop(handle)

   END SUBROUTINE write_neighbor_distribution

! **************************************************************************************************
!> \brief   Write a set of neighbor lists to the output unit.
!> \param ab ...
!> \param particle_set ...
!> \param cell ...
!> \param para_env ...
!> \param neighbor_list_section ...
!> \param nl_type ...
!> \param middle_name ...
!> \param name ...
!> \date    04.03.2002
!> \par History
!>       - Adapted to the new parallelized neighbor list version
!>         (26.06.2003,MK)
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE write_neighbor_lists(ab, particle_set, cell, para_env, neighbor_list_section, &
                                   nl_type, middle_name, name)

      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: ab
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: neighbor_list_section
      CHARACTER(LEN=*), INTENT(IN)                       :: nl_type, middle_name, name

      CHARACTER(LEN=default_string_length)               :: string, unit_str
      INTEGER                                            :: iatom, inode, iw, jatom, mype, &
                                                            nneighbor, nnode
      INTEGER, DIMENSION(3)                              :: cell_b
      REAL(dp)                                           :: dab, unit_conv
      REAL(dp), DIMENSION(3)                             :: ra, rab, rb
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator

      NULLIFY (logger)
      logger => cp_get_default_logger()
      IF (BTEST(cp_print_key_should_output(logger%iter_info, neighbor_list_section, &
                                           TRIM(nl_type)), &
                cp_p_file)) THEN
         iw = cp_print_key_unit_nr(logger=logger, &
                                   basis_section=neighbor_list_section, &
                                   print_key_path=TRIM(nl_type), &
                                   extension=".out", &
                                   middle_name=TRIM(middle_name), &
                                   local=.TRUE., &
                                   log_filename=.FALSE., &
                                   file_position="REWIND")
         mype = para_env%mepos
         CALL section_vals_val_get(neighbor_list_section, "UNIT", c_val=unit_str)
         unit_conv = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))

         ! Print headline
         string = ""
         WRITE (UNIT=string, FMT="(A,I5,A)") &
            TRIM(name)//" IN "//TRIM(unit_str)//" (PROCESS", mype, ")"
         CALL compress(string)
         IF (iw > 0) WRITE (UNIT=iw, FMT="(/,/,T2,A)") TRIM(string)

         nneighbor = 0

         CALL neighbor_list_iterator_create(nl_iterator, ab)
         DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
            CALL get_iterator_info(nl_iterator, inode=inode, nnode=nnode, &
                                   iatom=iatom, jatom=jatom, cell=cell_b, r=rab)
            nneighbor = nneighbor+1
            ra(:) = pbc(particle_set(iatom)%r, cell)
            rb(:) = ra(:)+rab(:)
            dab = SQRT(rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3))
            IF (iw > 0) THEN
               IF (inode == 1) THEN
                  WRITE (UNIT=iw, FMT="(/,T2,I5,3X,I6,3X,3F12.6)") &
                     iatom, nnode, ra(1:3)*unit_conv
               END IF
               WRITE (UNIT=iw, FMT="(T10,I6,3X,3I4,3F12.6,2X,F12.6)") &
                  jatom, cell_b(1:3), rb(1:3)*unit_conv, dab*unit_conv
            END IF
         END DO
         CALL neighbor_list_iterator_release(nl_iterator)

         string = ""
         WRITE (UNIT=string, FMT="(A,I12,A,I12)") &
            "Total number of neighbor interactions for process", mype, ":", &
            nneighbor
         CALL compress(string)
         IF (iw > 0) WRITE (UNIT=iw, FMT="(/,T2,A)") TRIM(string)
         CALL cp_print_key_finished_output(unit_nr=iw, &
                                           logger=logger, &
                                           basis_section=neighbor_list_section, &
                                           print_key_path=TRIM(nl_type), &
                                           local=.TRUE.)
      END IF

   END SUBROUTINE write_neighbor_lists

END MODULE qs_neighbor_lists
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for a Kim-Gordon-like partitioning into molecular subunits
!> \par History
!>       2012.07 created [Martin Haeufel]
!> \author Martin Haeufel
! **************************************************************************************************
MODULE kg_environment
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE distribution_2d_types,           ONLY: distribution_2d_type
   USE external_potential_types,        ONLY: get_potential,&
                                              local_potential_type
   USE input_constants,                 ONLY: kg_tnadd_atomic,&
                                              kg_tnadd_embed,&
                                              xc_funct_no_shortcut
   USE input_section_types,             ONLY: &
        section_vals_create, section_vals_duplicate, section_vals_get, section_vals_get_subs_vals, &
        section_vals_release, section_vals_set_subs_vals, section_vals_type, section_vals_val_get, &
        section_vals_val_set
   USE kg_environment_types,            ONLY: kg_environment_type
   USE kg_vertex_coloring_methods,      ONLY: kg_vertex_coloring
   USE kinds,                           ONLY: dp,&
                                              int_4,&
                                              int_4_size,&
                                              int_8
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_gather,&
                                              mp_max
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_types,                  ONLY: particle_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: deallocate_neighbor_list_set,&
                                              get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_neighbor_lists,               ONLY: atom2d_build,&
                                              atom2d_cleanup,&
                                              build_neighbor_lists,&
                                              local_atoms_type,&
                                              pair_radius_setup,&
                                              write_neighbor_lists
   USE task_list_types,                 ONLY: deallocate_task_list
   USE util,                            ONLY: sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'kg_environment'

   PUBLIC :: kg_env_create, kg_build_neighborlist, kg_build_subsets

CONTAINS

! **************************************************************************************************
!> \brief Allocates and intitializes kg_env
!> \param kg_env the object to create
!> \param input ...
!> \par History
!>       2012.07 created [Martin Haeufel]
!> \author Martin Haeufel
! **************************************************************************************************
   SUBROUTINE kg_env_create(kg_env, input)
      TYPE(kg_environment_type), POINTER                 :: kg_env
      TYPE(section_vals_type), OPTIONAL, POINTER         :: input

      CHARACTER(len=*), PARAMETER :: routineN = 'kg_env_create', routineP = moduleN//':'//routineN

      ALLOCATE (kg_env)
      CALL init_kg_env(kg_env, input)
   END SUBROUTINE kg_env_create

! **************************************************************************************************
!> \brief Initializes kg_env
!> \param kg_env ...
!> \param input ...
!> \par History
!>       2012.07 created [Martin Haeufel]
!> \author Martin Haeufel
! **************************************************************************************************
   SUBROUTINE init_kg_env(kg_env, input)
      TYPE(kg_environment_type), POINTER                 :: kg_env
      TYPE(section_vals_type), OPTIONAL, POINTER         :: input

      CHARACTER(LEN=*), PARAMETER :: routineN = 'init_kg_env', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ke_sections
      LOGICAL                                            :: is_set
      TYPE(section_vals_type), POINTER                   :: kef_section, xc_fun_section, &
                                                            xc_fun_section_kg, xc_section, &
                                                            xc_section_kg

      CALL timeset(routineN, handle)

      NULLIFY (kg_env%sab_orb_full)
      NULLIFY (kg_env%sac_kin)
      NULLIFY (kg_env%subset_of_mol)
      NULLIFY (kg_env%subset)
      NULLIFY (kg_env%tnadd_mat)

      kg_env%nsubsets = 0

      ! get coloring method settings
      CALL section_vals_val_get(input, "DFT%KG_METHOD%COLORING_METHOD", i_val=kg_env%coloring_method)
      ! get method for nonadditive kinetic energy embedding potential
      CALL section_vals_val_get(input, "DFT%KG_METHOD%TNADD_METHOD", i_val=kg_env%tnadd_method)

      IF (kg_env%tnadd_method == kg_tnadd_embed) THEN
         ! generate a new XC section with only the KE functional
         NULLIFY (xc_fun_section_kg, xc_section_kg)

         xc_section => section_vals_get_subs_vals(input, "DFT%XC")
         xc_fun_section => section_vals_get_subs_vals(xc_section, "XC_FUNCTIONAL")

         CALL section_vals_duplicate(xc_section, xc_section_kg)

         ! test for TF, TFW, KE_GGA, and KE_LIBXC sections in turn
         ke_sections = 0
         NULLIFY (kef_section)
         kef_section => section_vals_get_subs_vals(xc_fun_section, "TF")
         CALL section_vals_get(kef_section, explicit=is_set)
         IF (is_set) THEN
            NULLIFY (xc_fun_section_kg)
            CALL section_vals_create(xc_fun_section_kg, xc_fun_section%section)
            CALL section_vals_val_set(xc_fun_section_kg, "_SECTION_PARAMETERS_", &
                                      i_val=xc_funct_no_shortcut)
            CALL section_vals_set_subs_vals(xc_fun_section_kg, "TF", &
                                            kef_section)
            CALL section_vals_set_subs_vals(xc_section_kg, "XC_FUNCTIONAL", &
                                            xc_fun_section_kg)
            ke_sections = ke_sections+1
         END IF

         NULLIFY (kef_section)
         kef_section => section_vals_get_subs_vals(xc_fun_section, "TFW")
         CALL section_vals_get(kef_section, explicit=is_set)
         IF (is_set) THEN
            NULLIFY (xc_fun_section_kg)
            CALL section_vals_create(xc_fun_section_kg, xc_fun_section%section)
            CALL section_vals_val_set(xc_fun_section_kg, "_SECTION_PARAMETERS_", &
                                      i_val=xc_funct_no_shortcut)
            CALL section_vals_set_subs_vals(xc_fun_section_kg, "TFW", &
                                            kef_section)
            CALL section_vals_set_subs_vals(xc_section_kg, "XC_FUNCTIONAL", &
                                            xc_fun_section_kg)
            ke_sections = ke_sections+1
         END IF

         NULLIFY (kef_section)
         kef_section => section_vals_get_subs_vals(xc_fun_section, "KE_GGA")
         CALL section_vals_get(kef_section, explicit=is_set)
         IF (is_set) THEN
            NULLIFY (xc_fun_section_kg)
            CALL section_vals_create(xc_fun_section_kg, xc_fun_section%section)
            CALL section_vals_val_set(xc_fun_section_kg, "_SECTION_PARAMETERS_", &
                                      i_val=xc_funct_no_shortcut)
            CALL section_vals_set_subs_vals(xc_fun_section_kg, "KE_GGA", &
                                            kef_section)
            CALL section_vals_set_subs_vals(xc_section_kg, "XC_FUNCTIONAL", &
                                            xc_fun_section_kg)
            ke_sections = ke_sections+1
         END IF

         NULLIFY (kef_section)
         kef_section => section_vals_get_subs_vals(xc_fun_section, "KE_LIBXC")
         CALL section_vals_get(kef_section, explicit=is_set)
         IF (is_set) THEN
            NULLIFY (xc_fun_section_kg)
            CALL section_vals_create(xc_fun_section_kg, xc_fun_section%section)
            CALL section_vals_val_set(xc_fun_section_kg, "_SECTION_PARAMETERS_", &
                                      i_val=xc_funct_no_shortcut)
            CALL section_vals_set_subs_vals(xc_fun_section_kg, "KE_LIBXC", &
                                            kef_section)
            CALL section_vals_set_subs_vals(xc_section_kg, "XC_FUNCTIONAL", &
                                            xc_fun_section_kg)
            ke_sections = ke_sections+1
         END IF

         ! stop if there is more than one kinetic energy functional section present
         !CPPostcondition(ke_sections==1,cp_failure_level,routineP,failure)
         IF (ke_sections /= 1) &
            CALL cp_abort(cp__l("kg_environment.F",201), &
                          "KG runs require a kinetic energy functional. Exactly one of the following "// &
                          "subsections of XC_FUNCTIONAL needs to be present: KE_GGA, TF, TFW, or KE_LIBXC.")

         kg_env%xc_section_kg => xc_section_kg

         NULLIFY (kef_section, xc_fun_section)

         CALL section_vals_release(xc_fun_section_kg)

      ELSEIF (kg_env%tnadd_method == kg_tnadd_atomic) THEN
         NULLIFY (kg_env%xc_section_kg)
      ELSE
         CALL cp__b("kg_environment.F",214,"")
      END IF

      CALL timestop(handle)

   END SUBROUTINE init_kg_env

! **************************************************************************************************
!> \brief builds either the full neighborlist or neighborlists of molecular
!> \brief subsets, depending on parameter values
!> \param qs_env ...
!> \param sab_orb the return type, a neighborlist
!> \param sac_kin ...
!> \param molecular if false, the full neighborlist is build
!> \param subset_of_mol the molecular subsets
!> \param current_subset the subset of which the neighborlist is to be build
!> \par History
!>       2012.07 created [Martin Haeufel]
!> \author Martin Haeufel
! **************************************************************************************************
   SUBROUTINE kg_build_neighborlist(qs_env, sab_orb, sac_kin, &
                                    molecular, subset_of_mol, current_subset)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: sab_orb, sac_kin
      LOGICAL, OPTIONAL                                  :: molecular
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: subset_of_mol
      INTEGER, OPTIONAL                                  :: current_subset

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kg_build_neighborlist', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind, nkind
      LOGICAL                                            :: molecule_only
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: orb_present, tpot_present
      REAL(dp)                                           :: subcells
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: orb_radius, tpot_radius
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: pair_radius
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(distribution_1d_type), POINTER                :: distribution_1d
      TYPE(distribution_2d_type), POINTER                :: distribution_2d
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(local_atoms_type), ALLOCATABLE, DIMENSION(:)  :: atom2d
      TYPE(local_potential_type), POINTER                :: tnadd_potential
      TYPE(molecule_type), DIMENSION(:), POINTER         :: molecule_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: neighbor_list_section

      CALL timeset(routineN, handle)
      NULLIFY (para_env)

      molecule_only = .FALSE.
      IF (PRESENT(molecular)) molecule_only = molecular

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      distribution_2d=distribution_2d, &
                      molecule_set=molecule_set, &
                      local_particles=distribution_1d, &
                      particle_set=particle_set, &
                      para_env=para_env)

      CALL section_vals_val_get(qs_env%input, "DFT%SUBCELLS", r_val=subcells)

      ! Allocate work storage
      nkind = SIZE(atomic_kind_set)
      ALLOCATE (orb_radius(nkind), tpot_radius(nkind))
      orb_radius(:) = 0.0_dp
      tpot_radius(:) = 0.0_dp
      ALLOCATE (orb_present(nkind), tpot_present(nkind))
      ALLOCATE (pair_radius(nkind, nkind))
      ALLOCATE (atom2d(nkind))

      CALL atom2d_build(atom2d, distribution_1d, distribution_2d, atomic_kind_set, &
                        molecule_set, molecule_only, particle_set=particle_set)

      DO ikind = 1, nkind
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom2d(ikind)%list)
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_set)
         IF (ASSOCIATED(orb_basis_set)) THEN
            orb_present(ikind) = .TRUE.
            IF (PRESENT(subset_of_mol)) THEN
               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, kind_radius=orb_radius(ikind))
            ELSE
               CALL get_gto_basis_set(gto_basis_set=orb_basis_set, short_kind_radius=orb_radius(ikind))
            END IF
         ELSE
            orb_present(ikind) = .FALSE.
            orb_radius(ikind) = 0.0_dp
         END IF
      END DO

      IF (PRESENT(sab_orb)) THEN

         ! Build the orbital-orbital overlap neighbor list
         CALL pair_radius_setup(orb_present, orb_present, orb_radius, orb_radius, pair_radius)

         IF (PRESENT(subset_of_mol)) THEN
            CALL build_neighbor_lists(sab_orb, particle_set, atom2d, cell, pair_radius, &
                                      mic=.FALSE., subcells=subcells, molecular=molecule_only, subset_of_mol=subset_of_mol, &
                                      current_subset=current_subset, name="sab_orb")
         ELSE
            CALL build_neighbor_lists(sab_orb, particle_set, atom2d, cell, pair_radius, &
                                      mic=.FALSE., subcells=subcells, molecular=molecule_only, name="sab_orb")
         END IF

         ! Print out the neighborlist
         neighbor_list_section => section_vals_get_subs_vals(qs_env%input, "DFT%KG_METHOD%PRINT%NEIGHBOR_LISTS")

         IF (molecule_only) THEN
            CALL write_neighbor_lists(sab_orb, particle_set, cell, para_env, neighbor_list_section, &
                                      "/SAB_ORB_MOLECULAR", "sab_orb", "MOLECULAR SUBSET NEIGHBORLIST")
         ELSE
            CALL write_neighbor_lists(sab_orb, particle_set, cell, para_env, neighbor_list_section, &
                                      "/SAB_ORB_FULL", "sab_orb", "FULL NEIGHBORLIST")
         END IF

      END IF

      IF (PRESENT(sac_kin)) THEN
         DO ikind = 1, nkind
            tpot_present(ikind) = .FALSE.
            CALL get_qs_kind(qs_kind_set(ikind), tnadd_potential=tnadd_potential)
            IF (ASSOCIATED(tnadd_potential)) THEN
               CALL get_potential(potential=tnadd_potential, radius=tpot_radius(ikind))
               tpot_present(ikind) = .TRUE.
            END IF
         END DO
         CALL pair_radius_setup(orb_present, tpot_present, orb_radius, tpot_radius, pair_radius)
         CALL build_neighbor_lists(sac_kin, particle_set, atom2d, cell, pair_radius, &
                                   subcells=subcells, operator_type="ABC", name="sac_kin")
         neighbor_list_section => section_vals_get_subs_vals(qs_env%input, &
                                                             "DFT%KG_METHOD%PRINT%NEIGHBOR_LISTS")
         CALL write_neighbor_lists(sac_kin, particle_set, cell, para_env, neighbor_list_section, &
                                   "/SAC_KIN", "sac_kin", "ORBITAL kin energy potential")
      END IF

      ! Release work storage
      CALL atom2d_cleanup(atom2d)
      DEALLOCATE (atom2d)
      DEALLOCATE (orb_present, tpot_present)
      DEALLOCATE (orb_radius, tpot_radius)
      DEALLOCATE (pair_radius)

      CALL timestop(handle)

   END SUBROUTINE kg_build_neighborlist

! **************************************************************************************************
!> \brief Removes all replicated pairs from a 2d integer buffer array
!> \param pairs_buffer the array, assumed to have the shape (2,:)
!> \param n number of pairs (in), number of disjunct pairs (out)
!> \par History
!>       2012.07 created [Martin Haeufel]
!>       2014.11 simplified [Ole Schuett]
!> \author Martin Haeufel
! **************************************************************************************************
   SUBROUTINE kg_remove_duplicates(pairs_buffer, n)
      INTEGER(KIND=int_4), DIMENSION(:, :), &
         INTENT(INOUT)                                   :: pairs_buffer
      INTEGER, INTENT(INOUT)                             :: n

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kg_remove_duplicates', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, npairs
      INTEGER, DIMENSION(n)                              :: ind
      INTEGER(KIND=int_8), DIMENSION(n)                  :: sort_keys
      INTEGER(KIND=int_4), DIMENSION(2, n)               :: work

      CALL timeset(routineN, handle)

      IF (n > 0) THEN
         ! represent a pair of int_4 as a single int_8 number, simplifies sorting.
         sort_keys(1:n) = ISHFT(INT(pairs_buffer(1, 1:n), KIND=int_8), 8*int_4_size)
         sort_keys(1:n) = sort_keys(1:n)+pairs_buffer(2, 1:n) !upper + lower bytes
         CALL sort(sort_keys, n, ind)

         ! add first pair, the case npairs==0 was excluded above
         npairs = 1
         work(:, 1) = pairs_buffer(:, ind(1))

         ! remove duplicates from the sorted list
         DO i = 2, n
            IF (sort_keys(i) /= sort_keys(i-1)) THEN
               npairs = npairs+1
               work(:, npairs) = pairs_buffer(:, ind(i))
            END IF
         END DO

         n = npairs
         pairs_buffer(:, :n) = work(:, :n)
      END IF

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief writes the graph to file using the DIMACS standard format
!>        for a definition of the file format see
!>        mat.gsia.cmu.edu?COLOR/general/ccformat.ps
!>        c comment line
!>        p edge NODES EDGES
!>        with NODES - number of nodes
!>        EDGES - numer of edges
!>        e W V
!>        ...
!>        there is one edge descriptor line for each edge in the graph
!>        for an edge (w,v) the fields W and V specify its endpoints
!> \param pairs ...
!> \param nnodes the total number of nodes
! **************************************************************************************************
   SUBROUTINE write_to_file(pairs, nnodes)
      INTEGER(KIND=int_4), ALLOCATABLE, &
         DIMENSION(:, :), INTENT(IN)                     :: pairs
      INTEGER, INTENT(IN)                                :: nnodes

      CHARACTER(LEN=*), PARAMETER :: routineN = 'write_to_file', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, imol, jmol, npairs, unit_nr
      INTEGER(KIND=int_4), ALLOCATABLE, DIMENSION(:, :)  :: sorted_pairs

      CALL timeset(routineN, handle)

      ! get the number of disjunct pairs
      npairs = SIZE(pairs, 2)

      ALLOCATE (sorted_pairs(2, npairs))

      ! reorder pairs such that pairs(1,*) < pairs(2,*)
      DO i = 1, npairs
         ! get molecular ids
         imol = pairs(1, i)
         jmol = pairs(2, i)
         IF (imol > jmol) THEN
            ! switch pair and store
            sorted_pairs(1, i) = jmol
            sorted_pairs(2, i) = imol
         ELSE
            ! keep ordering just copy
            sorted_pairs(1, i) = imol
            sorted_pairs(2, i) = jmol
         END IF
      END DO

      ! remove duplicates and get the number of disjunct pairs (number of edges)
      CALL kg_remove_duplicates(sorted_pairs, npairs)

      ! should now be half as much pairs as before
      IF(.NOT.(npairs == SIZE(pairs, 2)/2))CALL cp__a("kg_environment.F",469)

      CALL open_file(unit_number=unit_nr, file_name="/mnt/sdcard/cp2k/work/graph.col")

      WRITE (unit_nr, '(A6,1X,I8,1X,I8)') "p edge", nnodes, npairs

      ! only write out the first npairs entries
      DO i = 1, npairs
         WRITE (unit_nr, '(A1,1X,I8,1X,I8)') "e", sorted_pairs(1, i), sorted_pairs(2, i)
      END DO

      CALL close_file(unit_nr)

      DEALLOCATE (sorted_pairs)

      CALL timestop(handle)

   END SUBROUTINE

! **************************************************************************************************
!> \brief ...
!> \param kg_env ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE kg_build_subsets(kg_env, para_env)
      TYPE(kg_environment_type), POINTER                 :: kg_env
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kg_build_subsets', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: color, handle, i, iab, iatom, imol, &
                                                            isub, jatom, jmol, nmol, npairs, &
                                                            npairs_local
      INTEGER(KIND=int_4)                                :: ncolors
      INTEGER(KIND=int_4), ALLOCATABLE, DIMENSION(:)     :: color_of_node
      INTEGER(KIND=int_4), ALLOCATABLE, DIMENSION(:, :)  :: msg_gather, pairs, pairs_buffer
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: index_of_color, nnodes_of_color
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator

! first: get a (local) list of pairs from the (local) neighbor list data

      CALL timeset(routineN, handle)

      nmol = SIZE(kg_env%molecule_set)

      npairs = 0
      CALL neighbor_list_iterator_create(nl_iterator, kg_env%sab_orb_full)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, iatom=iatom, jatom=jatom)

         imol = kg_env%atom_to_molecule(iatom)
         jmol = kg_env%atom_to_molecule(jatom)

         !IF (imol<jmol) THEN
         IF (imol .NE. jmol) THEN

            npairs = npairs+2

         END IF

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      ALLOCATE (pairs_buffer(2, npairs))

      npairs = 0
      CALL neighbor_list_iterator_create(nl_iterator, kg_env%sab_orb_full)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, iatom=iatom, jatom=jatom)

         imol = kg_env%atom_to_molecule(iatom)
         jmol = kg_env%atom_to_molecule(jatom)

         IF (imol .NE. jmol) THEN

            ! add pair to the local list

            ! add both orderings - makes it easier to build the neighborlist
            npairs = npairs+1

            pairs_buffer(1, npairs) = imol
            pairs_buffer(2, npairs) = jmol

            npairs = npairs+1

            pairs_buffer(2, npairs) = imol
            pairs_buffer(1, npairs) = jmol

         END IF

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      ! remove duplicates
      CALL kg_remove_duplicates(pairs_buffer, npairs)

      ! get the maximum number of local pairs on all nodes (size of the mssg)
      ! remember how many pairs we have local
      npairs_local = npairs
      CALL mp_max(npairs, para_env%group)

      ! allocate message
      ALLOCATE (pairs(2, npairs))

      pairs(:, 1:npairs_local) = pairs_buffer(:, 1:npairs_local)
      pairs(:, npairs_local+1:) = 0

      DEALLOCATE (pairs_buffer)

      ! second: gather all data on the master node
      ! better would be a scheme where duplicates are removed in a tree-like reduction scheme.
      ! this step can be needlessly memory intensive in the current implementation.

      IF (para_env%source .EQ. para_env%mepos) THEN
         ALLOCATE (msg_gather(2, npairs*para_env%num_pe))
      ELSE
         ALLOCATE (msg_gather(2, 1))
      ENDIF

      msg_gather = 0

      CALL mp_gather(pairs, msg_gather, para_env%source, para_env%group)

      DEALLOCATE (pairs)

      IF (para_env%source .EQ. para_env%mepos) THEN

         ! shift all non-zero entries to the beginning of the array and count the number of actual pairs
         npairs = 0

         DO i = 1, SIZE(msg_gather, 2)
            IF (msg_gather(1, i) .NE. 0) THEN
               npairs = npairs+1
               msg_gather(:, npairs) = msg_gather(:, i)
            END IF
         END DO

         ! remove duplicates
         CALL kg_remove_duplicates(msg_gather, npairs)

         ALLOCATE (pairs(2, npairs))

         pairs(:, 1:npairs) = msg_gather(:, 1:npairs)

         DEALLOCATE (msg_gather)

         !WRITE(*,'(A48,5X,I10,4X,A2,1X,I10)') " KG| Total number of overlapping molecular pairs",npairs/2,"of",nmol*(nmol-1)/2

         ! write to file, nnodes = number of molecules
         IF (.FALSE.) THEN
            CALL write_to_file(pairs, SIZE(kg_env%molecule_set))
         ENDIF

         ! vertex coloring algorithm
         CALL kg_vertex_coloring(kg_env, pairs, ncolors, color_of_node)

         DEALLOCATE (pairs)

      ELSE

         DEALLOCATE (msg_gather)

      END IF

      !WRITE(*,'(A27,40X,I6,1X,A6)') ' KG| Vertex coloring result', ncolors, 'colors'

      ! broadcast the number of colors to all nodes
      CALL mp_bcast(ncolors, para_env%source, para_env%group)

      IF (.NOT. ALLOCATED(color_of_node)) ALLOCATE (color_of_node(nmol))

      ! broadcast the resulting coloring to all nodes.....
      CALL mp_bcast(color_of_node, para_env%source, para_env%group)

      IF ((kg_env%nsubsets .NE. 0) .AND. (ncolors .NE. kg_env%nsubsets)) THEN
         ! number of subsets has changed

         ! deallocate stuff if necessary
         IF (ASSOCIATED(kg_env%subset)) THEN

            DO isub = 1, kg_env%nsubsets

               DO iab = 1, SIZE(kg_env%subset(isub)%sab_orb)

                  CALL deallocate_neighbor_list_set(kg_env%subset(isub)%sab_orb(iab)%neighbor_list_set)

               END DO

               DEALLOCATE (kg_env%subset(isub)%sab_orb)

               CALL deallocate_task_list(kg_env%subset(isub)%task_list)

            END DO

            DEALLOCATE (kg_env%subset)

            NULLIFY (kg_env%subset)

         END IF

      END IF

      ! allocate and nullify some stuff
      IF (.NOT. ASSOCIATED(kg_env%subset)) THEN

         ALLOCATE (kg_env%subset(ncolors))

         DO i = 1, ncolors
            NULLIFY (kg_env%subset(i)%sab_orb)
            NULLIFY (kg_env%subset(i)%task_list)
         END DO
      END IF

      ! set the number of subsets
      kg_env%nsubsets = ncolors

      ! counting loop
      ALLOCATE (nnodes_of_color(ncolors))
      nnodes_of_color = 0
      DO i = 1, nmol ! nmol=nnodes
         color = color_of_node(i)
         kg_env%subset_of_mol(i) = color
         nnodes_of_color(color) = nnodes_of_color(color)+1
      END DO

      DEALLOCATE (nnodes_of_color)

      ! store the subset information
      ALLOCATE (index_of_color(ncolors))

      index_of_color = 0
      DO i = 1, nmol
         color = color_of_node(i)
         index_of_color(color) = index_of_color(color)+1
      END DO
      DEALLOCATE (index_of_color)

      DEALLOCATE (color_of_node)

      CALL timestop(handle)

   END SUBROUTINE

END MODULE kg_environment
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief basis types for the calculation of the perturbation of density theory.
!> \par History
!>      4.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qs_kpp1_env_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_p_type
   USE pw_types,                        ONLY: pw_p_type,&
                                              pw_release
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_release
   USE xc_rho_set_types,                ONLY: xc_rho_set_release,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: qs_kpp1_env_type
   PUBLIC :: kpp1_release

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_kpp1_env_types'

! **************************************************************************************************
!> \brief environement that keeps the informations and temporary
!>      val to build the kpp1 kernel matrix
!> \param ref_count reference count (how may objects are sharing this one)
!> \param v_rspace potential in r space. This is used to do an update only
!>        of what has changed. Useful if P1 converges to some density
!>        (you spare in the grid-ao conversion).
!>        With LSD contains the various orbitals.
!> \param v_ao the potential in the ao basis (used togheter with v_rspace
!>        to update only what changed
!> \param id_nr identification number, unique for each kpp1 env
!> \param print_count counter to create unique filename
!> \param iter number of iterations
!> \param drho_r (idir,ispin): the derivative of rho wrt. x,y,z in the real space
!> \param deriv_xc (ii,ipot): the second derivative of the xc potential at psi0
!>        (qs_env%c), if grad pot is true it should already be divised
!>        by the gradient
!> \param spin_pot (1:2,ipot): information about wrt. to which spins the
!>         corresponding component of deriv_xc was derived (see
!>         xc_create_2nd_deriv_info)
!> \param grad_pot (1:2,ipot): if the derivative spin_pot was wrt. to
!>        the gradient (see xc_create_2nd_deriv_info)
!> \param ndiag_term (ipot): it the term is an off diagonal term (see
!>        xc_create_2nd_deriv_info)
! **************************************************************************************************
   TYPE qs_kpp1_env_type
      INTEGER :: ref_count, id_nr, print_count, iter
      TYPE(pw_p_type), DIMENSION(:), POINTER :: v_rspace
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: v_ao
      TYPE(pw_p_type), DIMENSION(:, :), POINTER :: drho_r
      TYPE(xc_derivative_set_type), POINTER :: deriv_set
      TYPE(xc_rho_set_type), POINTER        :: rho_set
      INTEGER, DIMENSION(:, :), POINTER :: spin_pot
      LOGICAL, DIMENSION(:, :), POINTER :: grad_pot
      LOGICAL, DIMENSION(:), POINTER :: ndiag_term
   END TYPE qs_kpp1_env_type

! **************************************************************************************************
!> \brief just to build array of pointers
!> \param kpp1_env: the pointer to the kpp1_env
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE qs_kpp1_env_p_type
      TYPE(qs_kpp1_env_type), POINTER :: kpp1_env
   END TYPE qs_kpp1_env_p_type

CONTAINS

! **************************************************************************************************
!> \brief releases a kpp1_env (see doc/ReferenceCounting.html)
!> \param kpp1_env the environement to release
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE kpp1_release(kpp1_env)
      TYPE(qs_kpp1_env_type), POINTER                    :: kpp1_env

      CHARACTER(len=*), PARAMETER :: routineN = 'kpp1_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin

      IF (ASSOCIATED(kpp1_env)) THEN
         IF(.NOT.(kpp1_env%ref_count > 0))CALL cp__a("qs_kpp1_env_types.F",96)
         kpp1_env%ref_count = kpp1_env%ref_count-1
         IF (kpp1_env%ref_count < 1) THEN
            IF (ASSOCIATED(kpp1_env%v_rspace)) THEN
               DO ispin = 1, SIZE(kpp1_env%v_rspace)
                  CALL pw_release(kpp1_env%v_rspace(ispin)%pw)
               END DO
               DEALLOCATE (kpp1_env%v_rspace)
            END IF
            IF (ASSOCIATED(kpp1_env%v_ao)) THEN
               DO ispin = 1, SIZE(kpp1_env%v_ao)
                  IF (ASSOCIATED(kpp1_env%v_ao(ispin)%matrix)) THEN
                     CALL cp_dbcsr_deallocate_matrix(kpp1_env%v_ao(ispin)%matrix)
                  END IF
               END DO
               DEALLOCATE (kpp1_env%v_ao)
            END IF
            IF (ASSOCIATED(kpp1_env%drho_r)) THEN
               DEALLOCATE (kpp1_env%drho_r)
            END IF
            IF (ASSOCIATED(kpp1_env%deriv_set)) THEN
               CALL xc_dset_release(kpp1_env%deriv_set)
               NULLIFY (kpp1_env%deriv_set)
            END IF
            IF (ASSOCIATED(kpp1_env%rho_set)) THEN
               CALL xc_rho_set_release(kpp1_env%rho_set)
               NULLIFY (kpp1_env%rho_set)
            END IF
            IF (ASSOCIATED(kpp1_env%spin_pot)) THEN
               DEALLOCATE (kpp1_env%spin_pot)
            END IF
            IF (ASSOCIATED(kpp1_env%grad_pot)) THEN
               DEALLOCATE (kpp1_env%grad_pot)
            END IF
            IF (ASSOCIATED(kpp1_env%ndiag_term)) THEN
               DEALLOCATE (kpp1_env%ndiag_term)
            END IF
            DEALLOCATE (kpp1_env)
         END IF
      END IF
      NULLIFY (kpp1_env)
   END SUBROUTINE kpp1_release

! **************************************************************************************************
!> \brief retains a kpp1_env (see doc/ReferenceCounting.html)
!> \param kpp1_env the environement to retain
!> \par History
!>      11.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE kpp1_retain(kpp1_env)
      TYPE(qs_kpp1_env_type), POINTER                    :: kpp1_env

      CHARACTER(len=*), PARAMETER :: routineN = 'kpp1_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(kpp1_env)))CALL cp__a("qs_kpp1_env_types.F",151)
      IF(.NOT.(kpp1_env%ref_count > 0))CALL cp__a("qs_kpp1_env_types.F",152)
      kpp1_env%ref_count = kpp1_env%ref_count+1
   END SUBROUTINE kpp1_retain

END MODULE qs_kpp1_env_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief basis types for the calculation of the perturbation of density theory.
!> \par History
!>      4.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qs_p_env_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_p_type
   USE cp_fm_types,                     ONLY: cp_fm_p_type
   USE cp_fm_vect,                      ONLY: cp_fm_vect_dealloc
   USE hartree_local_types,             ONLY: hartree_local_release,&
                                              hartree_local_type
   USE kinds,                           ONLY: dp
   USE preconditioner_types,            ONLY: destroy_preconditioner,&
                                              preconditioner_type
   USE qs_kpp1_env_types,               ONLY: kpp1_release,&
                                              qs_kpp1_env_type
   USE qs_local_rho_types,              ONLY: local_rho_set_release,&
                                              local_rho_type
   USE qs_rho_types,                    ONLY: qs_rho_release,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: qs_p_env_type
   PUBLIC :: p_env_release

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_p_env_types'

! **************************************************************************************************
!> \brief Represent a qs system that is perturbed.
!>      Can calculate the linear operator and the rhs of the system
!>      of equations that needs to be solved for the perturbation.
!> \param orthogonal_orbitals if the orbitals are orthogonal
!> \param id_nr identification number (unique between p_env_types)
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param iter number of iterations
!> \param kpp 1: the kpp1 matrix (make it temporary?)
!> \param m_epsilon minus epsilon: -<psi0d|H_rho|psi0d>
!> \param psi 0d: the dual basis of psi0: psi0 (psi0^T S psi0)^-1
!> \param S_psi 0: S times psi0, cached for performace reasons
!> \param Smo_inv inverse of the mo overlap: (psi0^T S psi0)^-1
!> \param kpp 1_env: environment for the calculation of kpp1
!> \param rho 1: the density rho1
!> \param rho 1: the soft density rho1 for gapw_xc
!> \param n_mo cached number of mo: n_mo(i)=qs_env%c(i)%nmo
!> \param n_ao cached number of ao: n_ao(i)=qs_env%c(i)%nao
!> \note
!>      for the moment no smearing of the orbitals.
! **************************************************************************************************
   TYPE qs_p_env_type

      LOGICAL :: orthogonal_orbitals
      INTEGER :: id_nr, ref_count, iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: kpp1, p1
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER :: m_epsilon, &
                                                   psi0d, S_psi0, Smo_inv
      TYPE(qs_kpp1_env_type), POINTER :: kpp1_env
      TYPE(qs_rho_type), POINTER :: rho1
      TYPE(qs_rho_type), POINTER :: rho1_xc
      INTEGER, DIMENSION(2) :: n_mo, & ! no of molecular orbitals
                               n_ao ! no of basis functions
      ! GAPW stuff
      TYPE(hartree_local_type), POINTER                 :: hartree_local
      TYPE(local_rho_type), POINTER                      :: local_rho_set

      ! Linear Response Modules
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER  :: PS_psi0
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER  :: ev_h0

      ! preconditioner matrix should be symmetric and positive definite
      TYPE(preconditioner_type), DIMENSION(:), POINTER :: preconditioner
      LOGICAL :: new_preconditioner

      !factors
      REAL(KIND=dp)  :: delta, gnorm, gnorm_cross, gnorm_old, etotal, gradient
      !line search
      INTEGER :: ls_count
      REAL(KIND=dp)  :: ls_pos(53), ls_energy(53), ls_grad(53)
      LOGICAL :: only_energy, os_valid

   END TYPE qs_p_env_type

! **************************************************************************************************
!> \brief to have arrays of pointers
!> \param p_env the pointer to the p_env
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE qs_p_env_p_type
      TYPE(qs_p_env_type), POINTER :: p_env
   END TYPE qs_p_env_p_type

CONTAINS

! **************************************************************************************************
!> \brief retains the given p_env (see doc/ReferenceCounting.html)
!> \param p_env the p_env to retain
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE p_env_retain(p_env)
      TYPE(qs_p_env_type), POINTER                       :: p_env

      CHARACTER(len=*), PARAMETER :: routineN = 'p_env_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(p_env)))CALL cp__a("qs_p_env_types.F",117)
      IF(.NOT.(p_env%ref_count > 0))CALL cp__a("qs_p_env_types.F",118)
      p_env%ref_count = p_env%ref_count+1
   END SUBROUTINE p_env_retain

! **************************************************************************************************
!> \brief relases the given p_env (see doc/ReferenceCounting.html)
!> \param p_env the environment to release
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE p_env_release(p_env)

      TYPE(qs_p_env_type), POINTER                       :: p_env

      CHARACTER(len=*), PARAMETER :: routineN = 'p_env_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: ip

      IF (ASSOCIATED(p_env)) THEN
         IF(.NOT.(p_env%ref_count > 0))CALL cp__a("qs_p_env_types.F",138)
         p_env%ref_count = p_env%ref_count-1
         IF (p_env%ref_count < 1) THEN
            CALL kpp1_release(p_env%kpp1_env)
            CALL cp_fm_vect_dealloc(p_env%S_psi0)
            CALL cp_fm_vect_dealloc(p_env%m_epsilon)
            CALL cp_fm_vect_dealloc(p_env%psi0d)
            CALL cp_fm_vect_dealloc(p_env%Smo_inv)
            IF (ASSOCIATED(p_env%rho1_xc)) THEN
               CALL qs_rho_release(p_env%rho1_xc)
            ENDIF
            CALL qs_rho_release(p_env%rho1)
            IF (ASSOCIATED(p_env%kpp1)) CALL cp_dbcsr_deallocate_matrix_set(p_env%kpp1)
            IF (ASSOCIATED(p_env%p1)) CALL cp_dbcsr_deallocate_matrix_set(p_env%p1)
            IF (ASSOCIATED(p_env%local_rho_set)) THEN
               CALL local_rho_set_release(p_env%local_rho_set)
            END IF
            IF (ASSOCIATED(p_env%hartree_local)) THEN
               CALL hartree_local_release(p_env%hartree_local)
            END IF
            IF (ASSOCIATED(p_env%PS_psi0)) THEN
               CALL cp_fm_vect_dealloc(p_env%PS_psi0)
            END IF
            IF (ASSOCIATED(p_env%ev_h0)) THEN
               DO ip = 1, SIZE(p_env%ev_h0, 1)
                  NULLIFY (p_env%ev_h0(ip)%matrix)
               END DO
               DEALLOCATE (p_env%ev_h0)
            END IF
            IF (ASSOCIATED(p_env%preconditioner)) THEN
               DO ip = 1, SIZE(p_env%preconditioner, 1)
                  CALL destroy_preconditioner(p_env%preconditioner(ip))
               END DO
               DEALLOCATE (p_env%preconditioner)
            END IF
         END IF
         DEALLOCATE (p_env)
      END IF
      NULLIFY (p_env)
   END SUBROUTINE p_env_release

END MODULE qs_p_env_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
MODULE hartree_local_methods
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE atprop_types,                    ONLY: atprop_array_init,&
                                              atprop_type
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE hartree_local_types,             ONLY: allocate_ecoul_1center,&
                                              ecoul_1center_type,&
                                              hartree_local_type,&
                                              set_ecoul_1c
   USE input_constants,                 ONLY: tddfpt_singlet
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi,&
                                              pi
   USE message_passing,                 ONLY: mp_sum
   USE orbital_pointers,                ONLY: indso,&
                                              nsoset
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_poisson_types,                ONLY: pw_poisson_periodic,&
                                              pw_poisson_type
   USE qs_charges_types,                ONLY: qs_charges_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: get_none0_cg_list,&
                                              harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE qs_local_rho_types,              ONLY: rhoz_type
   USE qs_p_env_types,                  ONLY: qs_p_env_type
   USE qs_rho0_types,                   ONLY: get_rho0_mpole,&
                                              rho0_atom_type,&
                                              rho0_mpole_type
   USE qs_rho_atom_types,               ONLY: get_rho_atom,&
                                              rho_atom_coeff,&
                                              rho_atom_type
   USE util,                            ONLY: get_limit
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'hartree_local_methods'

   ! Public Subroutine

   PUBLIC :: init_coulomb_local, calculate_Vh_1center, Vh_1c_gg_integrals

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param hartree_local ...
!> \param natom ...
! **************************************************************************************************
   SUBROUTINE init_coulomb_local(hartree_local, natom)

      TYPE(hartree_local_type), POINTER                  :: hartree_local
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'init_coulomb_local', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(ecoul_1center_type), DIMENSION(:), POINTER    :: ecoul_1c

      CALL timeset(routineN, handle)

      NULLIFY (ecoul_1c)
      !   Allocate and Initialize 1-center Potentials and Integrals
      CALL allocate_ecoul_1center(ecoul_1c, natom)
      hartree_local%ecoul_1c => ecoul_1c

      CALL timestop(handle)

   END SUBROUTINE init_coulomb_local

! **************************************************************************************************
!> \brief Calculates Hartree potential for hard and soft densities (including
!>        nuclear charge and compensation charges) using numerical integration
!> \param vrad_h ...
!> \param vrad_s ...
!> \param rrad_h ...
!> \param rrad_s ...
!> \param rrad_0 ...
!> \param rrad_z ...
!> \param grid_atom ...
!> \par History
!>      05.2012 JGH refactoring
!> \author ??
! **************************************************************************************************
   SUBROUTINE calculate_Vh_1center(vrad_h, vrad_s, rrad_h, rrad_s, rrad_0, rrad_z, grid_atom)

      REAL(dp), DIMENSION(:, :), INTENT(INOUT)           :: vrad_h, vrad_s
      TYPE(rho_atom_coeff), DIMENSION(:), INTENT(IN)     :: rrad_h, rrad_s
      REAL(dp), DIMENSION(:, :), INTENT(IN)              :: rrad_0
      REAL(dp), DIMENSION(:), INTENT(IN)                 :: rrad_z
      TYPE(grid_atom_type), POINTER                      :: grid_atom

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_Vh_1center', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ir, iso, ispin, l_ang, &
                                                            max_s_harm, nchannels, nr, nspins
      REAL(dp)                                           :: I1_down, I1_up, I2_down, I2_up, prefactor
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: rho_1, rho_2
      REAL(dp), DIMENSION(:), POINTER                    :: wr
      REAL(dp), DIMENSION(:, :), POINTER                 :: oor2l, r2l

      CALL timeset(routineN, handle)

      nr = grid_atom%nr
      max_s_harm = SIZE(vrad_h, 2)
      nspins = SIZE(rrad_h, 1)
      nchannels = SIZE(rrad_0, 2)

      r2l => grid_atom%rad2l
      oor2l => grid_atom%oorad2l
      wr => grid_atom%wr

      ALLOCATE (rho_1(nr, max_s_harm), rho_2(nr, max_s_harm))
      rho_1 = 0.0_dp
      rho_2 = 0.0_dp

      !   Case lm = 0
      rho_1(:, 1) = rrad_z(:)
      rho_2(:, 1) = rrad_0(:, 1)

      DO iso = 2, nchannels
         rho_2(:, iso) = rrad_0(:, iso)
      END DO

      DO iso = 1, max_s_harm
         DO ispin = 1, nspins
            rho_1(:, iso) = rho_1(:, iso)+rrad_h(ispin)%r_coef(:, iso)
            rho_2(:, iso) = rho_2(:, iso)+rrad_s(ispin)%r_coef(:, iso)
         END DO

         l_ang = indso(1, iso)
         prefactor = fourpi/(2._dp*l_ang+1._dp)

         rho_1(:, iso) = rho_1(:, iso)*wr(:)
         rho_2(:, iso) = rho_2(:, iso)*wr(:)

         I1_up = 0.0_dp
         I1_down = 0.0_dp
         I2_up = 0.0_dp
         I2_down = 0.0_dp

         I1_up = r2l(nr, l_ang)*rho_1(nr, iso)
         I2_up = r2l(nr, l_ang)*rho_2(nr, iso)

         DO ir = nr-1, 1, -1
            I1_down = I1_down+oor2l(ir, l_ang+1)*rho_1(ir, iso)
            I2_down = I2_down+oor2l(ir, l_ang+1)*rho_2(ir, iso)
         END DO

         vrad_h(nr, iso) = vrad_h(nr, iso)+prefactor* &
                           (oor2l(nr, l_ang+1)*I1_up+r2l(nr, l_ang)*I1_down)
         vrad_s(nr, iso) = vrad_s(nr, iso)+prefactor* &
                           (oor2l(nr, l_ang+1)*I2_up+r2l(nr, l_ang)*I2_down)

         DO ir = nr-1, 1, -1
            I1_up = I1_up+r2l(ir, l_ang)*rho_1(ir, iso)
            I1_down = I1_down-oor2l(ir, l_ang+1)*rho_1(ir, iso)
            I2_up = I2_up+r2l(ir, l_ang)*rho_2(ir, iso)
            I2_down = I2_down-oor2l(ir, l_ang+1)*rho_2(ir, iso)

            vrad_h(ir, iso) = vrad_h(ir, iso)+prefactor* &
                              (oor2l(ir, l_ang+1)*I1_up+r2l(ir, l_ang)*I1_down)
            vrad_s(ir, iso) = vrad_s(ir, iso)+prefactor* &
                              (oor2l(ir, l_ang+1)*I2_up+r2l(ir, l_ang)*I2_down)

         END DO

      END DO

      DEALLOCATE (rho_1, rho_2)

      CALL timestop(handle)

   END SUBROUTINE calculate_Vh_1center

! **************************************************************************************************
!> \brief Calculates one center GAPW Hartree energies and matrix elements
!>        Hartree potentials are input
!>        Takes possible background charge into account
!>        Special case for densities without core charge
!> \param qs_env ...
!> \param energy_hartree_1c ...
!> \param tddft ...
!> \param do_triplet ...
!> \param p_env ...
!> \par History
!>      05.2012 JGH refactoring
!> \author ??
! **************************************************************************************************
   SUBROUTINE Vh_1c_gg_integrals(qs_env, energy_hartree_1c, tddft, do_triplet, p_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      REAL(kind=dp), INTENT(out)                         :: energy_hartree_1c
      LOGICAL, INTENT(IN), OPTIONAL                      :: tddft, do_triplet
      TYPE(qs_p_env_type), OPTIONAL, POINTER             :: p_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'Vh_1c_gg_integrals', &
         routineP = moduleN//':'//routineN

      INTEGER :: bo(2), handle, iat, iatom, ikind, ipgf1, is1, iset1, iso, l_ang, llmax, lmax0, &
         lmax_0, m1, max_iso, max_iso_not0, max_s_harm, maxl, maxso, mepos, n1, nat, natom, &
         nchan_0, nkind, nr, nset, nsotot, nspins, num_pe
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cg_n_list
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: cg_list
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, lmax, lmin, npgf
      LOGICAL                                            :: atenergy, core_charge, my_periodic, &
                                                            my_tddft, paw_atom
      REAL(dp)                                           :: back_ch, factor
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: gexp, sqrtwr
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: aVh1b_00, aVh1b_hh, aVh1b_ss, g0_h_w
      REAL(dp), DIMENSION(:), POINTER                    :: rrad_z, vrrad_z
      REAL(dp), DIMENSION(:, :), POINTER                 :: g0_h, gsph, rrad_0, Vh1_h, Vh1_s, &
                                                            vrrad_0, zet
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_CG, Qlm_gg
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(ecoul_1center_type), DIMENSION(:), POINTER    :: ecoul_1c
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(qs_charges_type), POINTER                     :: qs_charges
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(rho0_atom_type), DIMENSION(:), POINTER        :: rho0_atom_set
      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      TYPE(rhoz_type), DIMENSION(:), POINTER             :: rhoz_set

      CALL timeset(routineN, handle)

      NULLIFY (cell, dft_control, para_env, poisson_env, pw_env, qs_charges)
      NULLIFY (atomic_kind_set, qs_kind_set, rho_atom_set, rho0_atom_set)
      NULLIFY (rho0_mpole, rhoz_set, ecoul_1c)
      NULLIFY (atom_list, grid_atom, harmonics)
      NULLIFY (orb_basis, lmin, lmax, npgf, zet)
      NULLIFY (gsph, atprop)

      my_tddft = .FALSE.
      IF (PRESENT(tddft)) my_tddft = tddft
      core_charge = .NOT. my_tddft

      CALL get_qs_env(qs_env=qs_env, &
                      cell=cell, dft_control=dft_control, &
                      para_env=para_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, atprop=atprop, &
                      pw_env=pw_env, qs_charges=qs_charges)

      CALL pw_env_get(pw_env, poisson_env=poisson_env)
      my_periodic = (poisson_env%method == pw_poisson_periodic)

      back_ch = qs_charges%background*cell%deth

      IF (my_tddft) THEN
         rho_atom_set => p_env%local_rho_set%rho_atom_set
         rho0_atom_set => p_env%local_rho_set%rho0_atom_set
         rho0_mpole => p_env%local_rho_set%rho0_mpole
         ecoul_1c => p_env%hartree_local%ecoul_1c
      ELSE
         CALL get_qs_env(qs_env=qs_env, &
                         rho_atom_set=rho_atom_set, &
                         rho0_atom_set=rho0_atom_set, &
                         rho0_mpole=rho0_mpole, &
                         rhoz_set=rhoz_set, &
                         ecoul_1c=ecoul_1c)
      END IF

      nkind = SIZE(atomic_kind_set, 1)
      nspins = dft_control%nspins

      IF (my_tddft) THEN
         IF (PRESENT(do_triplet)) THEN
            IF (nspins == 1 .AND. do_triplet) RETURN
         ELSE
            IF (nspins == 1 .AND. dft_control%tddfpt_control%res_etype /= tddfpt_singlet) RETURN
         ENDIF
      END IF

      CALL get_qs_kind_set(qs_kind_set, maxg_iso_not0=max_iso)
      CALL get_rho0_mpole(rho0_mpole=rho0_mpole, lmax_0=lmax_0)

      atenergy = .FALSE.
      IF (ASSOCIATED(atprop)) THEN
         atenergy = atprop%energy
         IF (atenergy) THEN
            CALL get_qs_env(qs_env=qs_env, natom=natom)
            CALL atprop_array_init(atprop%ate1c, natom)
         END IF
      END IF

      !   Put to 0 the local hartree energy contribution from 1 center integrals
      energy_hartree_1c = 0.0_dp

      !   Here starts the loop over all the atoms
      DO ikind = 1, nkind

         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=nat)
         CALL get_qs_kind(qs_kind_set(ikind), &
                          basis_set=orb_basis, &
                          grid_atom=grid_atom, &
                          harmonics=harmonics, ngrid_rad=nr, &
                          max_iso_not0=max_iso_not0, paw_atom=paw_atom)

         IF (paw_atom) THEN
!===========    PAW   ===============
            CALL get_gto_basis_set(gto_basis_set=orb_basis, lmax=lmax, lmin=lmin, &
                                   maxso=maxso, npgf=npgf, maxl=maxl, &
                                   nset=nset, zet=zet)

            max_s_harm = harmonics%max_s_harm
            llmax = harmonics%llmax

            nsotot = maxso*nset
            ALLOCATE (gsph(nr, nsotot))
            ALLOCATE (gexp(nr))
            ALLOCATE (sqrtwr(nr), g0_h_w(nr, 0:lmax_0))

            NULLIFY (Vh1_h, Vh1_s)
            ALLOCATE (Vh1_h(nr, max_iso_not0))
            ALLOCATE (Vh1_s(nr, max_iso_not0))

            ALLOCATE (aVh1b_hh(nsotot, nsotot))
            ALLOCATE (aVh1b_ss(nsotot, nsotot))
            ALLOCATE (aVh1b_00(nsotot, nsotot))
            ALLOCATE (cg_list(2, nsoset(maxl)**2, max_s_harm), cg_n_list(max_s_harm))

            NULLIFY (Qlm_gg, g0_h)
            CALL get_rho0_mpole(rho0_mpole=rho0_mpole, ikind=ikind, &
                                l0_ikind=lmax0, &
                                Qlm_gg=Qlm_gg, g0_h=g0_h)

            nchan_0 = nsoset(lmax0)

            IF (nchan_0 > max_iso_not0) CALL cp__b("hartree_local_methods.F",359,"channels for rho0 > # max of spherical harmonics")

            NULLIFY (rrad_z, my_CG)
            my_CG => harmonics%my_CG

            !     set to zero temporary arrays
            sqrtwr = 0.0_dp
            g0_h_w = 0.0_dp
            gexp = 0.0_dp
            gsph = 0.0_dp

            sqrtwr(1:nr) = SQRT(grid_atom%wr(1:nr))
            DO l_ang = 0, lmax0
               g0_h_w(1:nr, l_ang) = g0_h(1:nr, l_ang)*grid_atom%wr(1:nr)
            END DO

            m1 = 0
            DO iset1 = 1, nset
               n1 = nsoset(lmax(iset1))
               DO ipgf1 = 1, npgf(iset1)
                  gexp(1:nr) = EXP(-zet(ipgf1, iset1)*grid_atom%rad2(1:nr))*sqrtwr(1:nr)
                  DO is1 = nsoset(lmin(iset1)-1)+1, nsoset(lmax(iset1))
                     iso = is1+(ipgf1-1)*n1+m1
                     l_ang = indso(1, is1)
                     gsph(1:nr, iso) = grid_atom%rad2l(1:nr, l_ang)*gexp(1:nr)
                  END DO ! is1
               END DO ! ipgf1
               m1 = m1+maxso
            END DO ! iset1

            !     Distribute the atoms of this kind
            num_pe = para_env%num_pe
            mepos = para_env%mepos
            bo = get_limit(nat, num_pe, mepos)

            DO iat = bo(1), bo(2) !1,nat
               iatom = atom_list(iat)
               rho_atom => rho_atom_set(iatom)

               NULLIFY (rrad_z, vrrad_z, rrad_0, vrrad_0)
               IF (core_charge) THEN
                  rrad_z => rhoz_set(ikind)%r_coef
                  vrrad_z => rhoz_set(ikind)%vr_coef
               END IF
               rrad_0 => rho0_atom_set(iatom)%rho0_rad_h%r_coef
               vrrad_0 => rho0_atom_set(iatom)%vrho0_rad_h%r_coef
               IF (my_periodic .AND. back_ch .GT. 1.E-3_dp) THEN
                  factor = -2.0_dp*pi/3.0_dp*SQRT(fourpi)*qs_charges%background
               ELSE
                  factor = 0._dp
               END IF

               CALL Vh_1c_atom_potential(rho_atom, vrrad_0, &
                                         grid_atom, core_charge, vrrad_z, Vh1_h, Vh1_s, &
                                         nchan_0, nspins, max_iso_not0, factor)

               CALL Vh_1c_atom_energy(energy_hartree_1c, ecoul_1c, rho_atom, rrad_0, &
                                      grid_atom, iatom, core_charge, rrad_z, Vh1_h, Vh1_s, &
                                      nchan_0, nspins, max_iso_not0, atenergy, atprop%ate1c)

               CALL Vh_1c_atom_integrals(rho_atom, &
                                         aVh1b_hh, aVh1b_ss, aVh1b_00, Vh1_h, Vh1_s, max_iso_not0, &
                                         max_s_harm, llmax, cg_list, cg_n_list, &
                                         nset, npgf, lmin, lmax, nsotot, maxso, nspins, nchan_0, gsph, g0_h_w, my_CG, Qlm_gg)

            END DO ! iat

            DEALLOCATE (aVh1b_hh)
            DEALLOCATE (aVh1b_ss)
            DEALLOCATE (aVh1b_00)
            DEALLOCATE (Vh1_h, Vh1_s)
            DEALLOCATE (cg_list, cg_n_list)
            DEALLOCATE (gsph)
            DEALLOCATE (gexp)
            DEALLOCATE (sqrtwr, g0_h_w)

         ELSE
!===========   NO  PAW   ===============
!  This term is taken care of using the core density as in GPW
            CYCLE
         END IF ! paw
      END DO ! ikind

      CALL mp_sum(energy_hartree_1c, para_env%group)

      CALL timestop(handle)

   END SUBROUTINE Vh_1c_gg_integrals

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param rho_atom ...
!> \param vrrad_0 ...
!> \param grid_atom ...
!> \param core_charge ...
!> \param vrrad_z ...
!> \param Vh1_h ...
!> \param Vh1_s ...
!> \param nchan_0 ...
!> \param nspins ...
!> \param max_iso_not0 ...
!> \param bfactor ...
! **************************************************************************************************
   SUBROUTINE Vh_1c_atom_potential(rho_atom, vrrad_0, &
                                   grid_atom, core_charge, vrrad_z, Vh1_h, Vh1_s, &
                                   nchan_0, nspins, max_iso_not0, bfactor)

      TYPE(rho_atom_type), POINTER                       :: rho_atom
      REAL(dp), DIMENSION(:, :), POINTER                 :: vrrad_0
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      LOGICAL, INTENT(IN)                                :: core_charge
      REAL(dp), DIMENSION(:), POINTER                    :: vrrad_z
      REAL(dp), DIMENSION(:, :), POINTER                 :: Vh1_h, Vh1_s
      INTEGER, INTENT(IN)                                :: nchan_0, nspins, max_iso_not0
      REAL(dp), INTENT(IN)                               :: bfactor

      CHARACTER(LEN=*), PARAMETER :: routineN = 'Vh_1c_atom_potential', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ir, iso, ispin, nr
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: vr_h, vr_s

      nr = grid_atom%nr

      NULLIFY (vr_h, vr_s)
      CALL get_rho_atom(rho_atom=rho_atom, vrho_rad_h=vr_h, vrho_rad_s=vr_s)

      Vh1_h = 0.0_dp
      Vh1_s = 0.0_dp

      IF (core_charge) Vh1_h(:, 1) = vrrad_z(:)

      DO iso = 1, nchan_0
         Vh1_s(:, iso) = vrrad_0(:, iso)
      END DO

      DO ispin = 1, nspins
         DO iso = 1, max_iso_not0
            Vh1_h(:, iso) = Vh1_h(:, iso)+vr_h(ispin)%r_coef(:, iso)
            Vh1_s(:, iso) = Vh1_s(:, iso)+vr_s(ispin)%r_coef(:, iso)
         END DO
      END DO

      IF (bfactor /= 0._dp) THEN
         DO ir = 1, nr
            Vh1_h(ir, 1) = Vh1_h(ir, 1)+bfactor*grid_atom%rad2(ir)*grid_atom%wr(ir)
            Vh1_s(ir, 1) = Vh1_s(ir, 1)+bfactor*grid_atom%rad2(ir)*grid_atom%wr(ir)
         END DO
      END IF

   END SUBROUTINE Vh_1c_atom_potential

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param energy_hartree_1c ...
!> \param ecoul_1c ...
!> \param rho_atom ...
!> \param rrad_0 ...
!> \param grid_atom ...
!> \param iatom ...
!> \param core_charge ...
!> \param rrad_z ...
!> \param Vh1_h ...
!> \param Vh1_s ...
!> \param nchan_0 ...
!> \param nspins ...
!> \param max_iso_not0 ...
!> \param atenergy ...
!> \param ate1c ...
! **************************************************************************************************
   SUBROUTINE Vh_1c_atom_energy(energy_hartree_1c, ecoul_1c, rho_atom, rrad_0, &
                                grid_atom, iatom, core_charge, rrad_z, Vh1_h, Vh1_s, &
                                nchan_0, nspins, max_iso_not0, atenergy, ate1c)

      REAL(dp), INTENT(INOUT)                            :: energy_hartree_1c
      TYPE(ecoul_1center_type), DIMENSION(:), POINTER    :: ecoul_1c
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      REAL(dp), DIMENSION(:, :), POINTER                 :: rrad_0
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      INTEGER, INTENT(IN)                                :: iatom
      LOGICAL, INTENT(IN)                                :: core_charge
      REAL(dp), DIMENSION(:), POINTER                    :: rrad_z
      REAL(dp), DIMENSION(:, :), POINTER                 :: Vh1_h, Vh1_s
      INTEGER, INTENT(IN)                                :: nchan_0, nspins, max_iso_not0
      LOGICAL, INTENT(IN)                                :: atenergy
      REAL(dp), DIMENSION(:), POINTER                    :: ate1c

      CHARACTER(LEN=*), PARAMETER :: routineN = 'Vh_1c_atom_energy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iso, ispin, nr
      REAL(dp)                                           :: ecoul_1_0, ecoul_1_h, ecoul_1_s, &
                                                            ecoul_1_z
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: r_h, r_s

      nr = grid_atom%nr

      NULLIFY (r_h, r_s)
      CALL get_rho_atom(rho_atom=rho_atom, rho_rad_h=r_h, rho_rad_s=r_s)

      !       Calculate the contributions to Ecoul coming from Vh1_h*rhoz
      ecoul_1_z = 0.0_dp
      IF (core_charge) THEN
         ecoul_1_z = 0.5_dp*SUM(Vh1_h(:, 1)*rrad_z(:)*grid_atom%wr(:))
      END IF

      !       Calculate the contributions to Ecoul coming from  Vh1_s*rho0
      ecoul_1_0 = 0.0_dp
      DO iso = 1, nchan_0
         ecoul_1_0 = ecoul_1_0+0.5_dp*SUM(Vh1_s(:, iso)*rrad_0(:, iso)*grid_atom%wr(:))
      END DO

      !       Calculate the contributions to Ecoul coming from Vh1_h*rho1_h and Vh1_s*rho1_s
      ecoul_1_s = 0.0_dp
      ecoul_1_h = 0.0_dp
      DO ispin = 1, nspins
         DO iso = 1, max_iso_not0
            ecoul_1_s = ecoul_1_s+0.5_dp*SUM(Vh1_s(:, iso)*r_s(ispin)%r_coef(:, iso)*grid_atom%wr(:))
            ecoul_1_h = ecoul_1_h+0.5_dp*SUM(Vh1_h(:, iso)*r_h(ispin)%r_coef(:, iso)*grid_atom%wr(:))
         END DO
      END DO

      CALL set_ecoul_1c(ecoul_1c, iatom, ecoul_1_z=ecoul_1_z, ecoul_1_0=ecoul_1_0)
      CALL set_ecoul_1c(ecoul_1c=ecoul_1c, iatom=iatom, ecoul_1_h=ecoul_1_h, ecoul_1_s=ecoul_1_s)

      energy_hartree_1c = energy_hartree_1c+ecoul_1_z-ecoul_1_0
      energy_hartree_1c = energy_hartree_1c+ecoul_1_h-ecoul_1_s

      ! atomic energy contribution
      IF (atenergy) THEN
         ate1c(iatom) = ate1c(iatom)+ecoul_1_z-ecoul_1_0
      END IF

   END SUBROUTINE Vh_1c_atom_energy

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

! **************************************************************************************************
!> \brief ...
!> \param rho_atom ...
!> \param aVh1b_hh ...
!> \param aVh1b_ss ...
!> \param aVh1b_00 ...
!> \param Vh1_h ...
!> \param Vh1_s ...
!> \param max_iso_not0 ...
!> \param max_s_harm ...
!> \param llmax ...
!> \param cg_list ...
!> \param cg_n_list ...
!> \param nset ...
!> \param npgf ...
!> \param lmin ...
!> \param lmax ...
!> \param nsotot ...
!> \param maxso ...
!> \param nspins ...
!> \param nchan_0 ...
!> \param gsph ...
!> \param g0_h_w ...
!> \param my_CG ...
!> \param Qlm_gg ...
! **************************************************************************************************
   SUBROUTINE Vh_1c_atom_integrals(rho_atom, &
                                   aVh1b_hh, aVh1b_ss, aVh1b_00, Vh1_h, Vh1_s, max_iso_not0, &
                                   max_s_harm, llmax, cg_list, cg_n_list, &
                                   nset, npgf, lmin, lmax, nsotot, maxso, nspins, nchan_0, gsph, g0_h_w, my_CG, Qlm_gg)

      TYPE(rho_atom_type), POINTER                       :: rho_atom
      REAL(dp), DIMENSION(:, :)                          :: aVh1b_hh, aVh1b_ss, aVh1b_00
      REAL(dp), DIMENSION(:, :), POINTER                 :: Vh1_h, Vh1_s
      INTEGER, INTENT(IN)                                :: max_iso_not0, max_s_harm, llmax
      INTEGER, DIMENSION(:, :, :)                        :: cg_list
      INTEGER, DIMENSION(:)                              :: cg_n_list
      INTEGER, INTENT(IN)                                :: nset
      INTEGER, DIMENSION(:), POINTER                     :: npgf, lmin, lmax
      INTEGER, INTENT(IN)                                :: nsotot, maxso, nspins, nchan_0
      REAL(dp), DIMENSION(:, :), POINTER                 :: gsph
      REAL(dp), DIMENSION(:, 0:)                         :: g0_h_w
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_CG, Qlm_gg

      CHARACTER(LEN=*), PARAMETER :: routineN = 'Vh_1c_atom_integrals', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: icg, ipgf1, ipgf2, ir, is1, is2, iset1, &
                                                            iset2, iso, iso1, iso2, ispin, l_ang, &
                                                            m1, m2, max_iso_not0_local, n1, n2, nr
      REAL(dp)                                           :: gVg_0, gVg_h, gVg_s
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: int_local_h, int_local_s

      NULLIFY (int_local_h, int_local_s)
      CALL get_rho_atom(rho_atom=rho_atom, &
                        ga_Vlocal_gb_h=int_local_h, &
                        ga_Vlocal_gb_s=int_local_s)

      !       Calculate the integrals of the potential with 2 primitives
      aVh1b_hh = 0.0_dp
      aVh1b_ss = 0.0_dp
      aVh1b_00 = 0.0_dp

      nr = SIZE(gsph, 1)

      m1 = 0
      DO iset1 = 1, nset
         m2 = 0
         DO iset2 = 1, nset
            CALL get_none0_cg_list(my_CG, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                   max_s_harm, llmax, cg_list, cg_n_list, max_iso_not0_local)

            n1 = nsoset(lmax(iset1))
            DO ipgf1 = 1, npgf(iset1)
               n2 = nsoset(lmax(iset2))
               DO ipgf2 = 1, npgf(iset2)
                  !               with contributions to  V1_s*rho0
                  DO iso = 1, nchan_0
                     l_ang = indso(1, iso)
                     gVg_0 = SUM(Vh1_s(:, iso)*g0_h_w(:, l_ang))
                     DO icg = 1, cg_n_list(iso)
                        is1 = cg_list(1, icg, iso)
                        is2 = cg_list(2, icg, iso)

                        iso1 = is1+n1*(ipgf1-1)+m1
                        iso2 = is2+n2*(ipgf2-1)+m2
                        gVg_h = 0.0_dp
                        gVg_s = 0.0_dp

                        DO ir = 1, nr
                           gVg_h = gVg_h+gsph(ir, iso1)*gsph(ir, iso2)*Vh1_h(ir, iso)
                           gVg_s = gVg_s+gsph(ir, iso1)*gsph(ir, iso2)*Vh1_s(ir, iso)
                        END DO ! ir

                        aVh1b_hh(iso1, iso2) = aVh1b_hh(iso1, iso2)+gVg_h*my_CG(is1, is2, iso)
                        aVh1b_ss(iso1, iso2) = aVh1b_ss(iso1, iso2)+gVg_s*my_CG(is1, is2, iso)
                        aVh1b_00(iso1, iso2) = aVh1b_00(iso1, iso2)+gVg_0*Qlm_gg(iso1, iso2, iso)

                     END DO !icg
                  END DO ! iso
                  !               without contributions to  V1_s*rho0
                  DO iso = nchan_0+1, max_iso_not0
                     DO icg = 1, cg_n_list(iso)
                        is1 = cg_list(1, icg, iso)
                        is2 = cg_list(2, icg, iso)

                        iso1 = is1+n1*(ipgf1-1)+m1
                        iso2 = is2+n2*(ipgf2-1)+m2
                        gVg_h = 0.0_dp
                        gVg_s = 0.0_dp

                        DO ir = 1, nr
                           gVg_h = gVg_h+gsph(ir, iso1)*gsph(ir, iso2)*Vh1_h(ir, iso)
                           gVg_s = gVg_s+gsph(ir, iso1)*gsph(ir, iso2)*Vh1_s(ir, iso)
                        END DO ! ir

                        aVh1b_hh(iso1, iso2) = aVh1b_hh(iso1, iso2)+gVg_h*my_CG(is1, is2, iso)
                        aVh1b_ss(iso1, iso2) = aVh1b_ss(iso1, iso2)+gVg_s*my_CG(is1, is2, iso)

                     END DO !icg
                  END DO ! iso
               END DO ! ipgf2
            END DO ! ipgf1
            m2 = m2+maxso
         END DO ! iset2
         m1 = m1+maxso
      END DO !iset1

      DO ispin = 1, nspins
         CALL daxpy(nsotot*nsotot, 1.0_dp, aVh1b_hh, 1, int_local_h(ispin)%r_coef, 1)
         CALL daxpy(nsotot*nsotot, 1.0_dp, aVh1b_ss, 1, int_local_s(ispin)%r_coef, 1)
         CALL daxpy(nsotot*nsotot, -1.0_dp, aVh1b_00, 1, int_local_h(ispin)%r_coef, 1)
         CALL daxpy(nsotot*nsotot, -1.0_dp, aVh1b_00, 1, int_local_s(ispin)%r_coef, 1)
      END DO ! ispin

   END SUBROUTINE Vh_1c_atom_integrals

!%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

END MODULE hartree_local_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Default parameter sets for SCPTB method
!>      key, atomname            key, atomic name
!>      z, zeff                  atomic number, core charge
!>      lmaxorb, lmaxscp         max l basis, max l polarization
!>      norb                     number of orbitals (0,1,2,3)
!>      zeta                     exponents of orbitals (0,1,2,3)
!>      hcore                    two body terms
!>      occupation               orbital occupation in free atom
!>      energy                   total energy of free atom
!>      crep                     core repulsion terms (1=a,2=c,3=d)
!>      pol                      polarization constants (1,2,3)
!>      ag                       Gaussian exponents
!> \par History
!> \author JGH (12.2012)
! **************************************************************************************************
MODULE scptb_parameters

   USE kinds,                           ONLY: dp
   USE periodic_table,                  ONLY: ptable
   USE scptb_types,                     ONLY: scptb_parameter_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'scptb_parameters'

   INTEGER, PARAMETER, PRIVATE :: nelem = 106

   PUBLIC :: scptb_default_parameter

CONTAINS

! **************************************************************************************************
!> \brief Default parameter sets for SCPTB method
!> \param param ...
! **************************************************************************************************
   SUBROUTINE scptb_default_parameter(param)
      TYPE(scptb_parameter_type), INTENT(INOUT)          :: param

      CHARACTER(len=*), PARAMETER :: routineN = 'scptb_default_parameter', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, l

      param%parameterization = "UZH.00.00"
      SELECT CASE (TRIM (param%key))
      CASE DEFAULT
         CALL cp__b("scptb_parameters.F",56,"")
      CASE ("XX")
         param%defined = .TRUE.
      CASE ("H")
         !Hydrogen
         param%defined = .TRUE.
         param%key = "H"
         param%z = 1
         param%zeff = 1
         param%lmaxorb = 0
         param%lmaxscp = 1
         param%norb(0:3) = (/1, 0, 0, 0/)
         param%nqm(1, 0) = 1
         param%zeta(1, 0) = 1.2_dp
         param%hcore(1, 0) = 1.0_dp
         param%occupation(1, 0) = 1
         param%crep(1:3) = (/1.0_dp, 1.0_dp, 1.0_dp/)
         param%pol(1:3) = (/1.0_dp, 1.0_dp, 1.0_dp/)
         param%ag = 4.0_dp
         param%energy = -0.5_dp
!   CASE (2)
!      !Helium
!   CASE (3)
!      !Lithium
!   CASE (4)
!      !Beryllium
!   CASE (5)
!      !Boron
!   CASE (6)
!      !Carbon
!   CASE (7)
!      !Nitrogen
      CASE ("O")
         !Oxygen
         param%defined = .TRUE.
         param%key = "O"
         param%z = 8
         param%zeff = 6
         param%lmaxorb = 1
         param%lmaxscp = 2
         param%norb(0:3) = (/1, 1, 0, 0/)
         param%nqm(1, 0:1) = (/2, 2/)
         param%zeta(1, 0:1) = (/1.8_dp, 1.8_dp/)
         param%hcore(1, 0:1) = (/1.0_dp, 1.0_dp/)
         param%occupation(1, 0:1) = (/2, 4/)
         param%crep(1:3) = (/1.0_dp, 1.0_dp, 1.0_dp/)
         param%pol(1:3) = (/1.0_dp, 1.0_dp, 1.0_dp/)
         param%ag = 1.0_dp
         param%energy = -0.5_dp
      CASE ("F")
         !Fluorine
         param%defined = .TRUE.
         param%key = "F"
         param%z = 9
         param%zeff = 7
         param%lmaxorb = 1
         param%lmaxscp = 2
         param%norb(0:3) = (/1, 1, 0, 0/)
         param%nqm(1, 0:1) = (/2, 2/)
         param%zeta(1, 0:1) = (/1.8_dp, 1.8_dp/)
         param%hcore(1, 0:1) = (/1.0_dp, 1.0_dp/)
         param%occupation(1, 0:1) = (/2, 5/)
         param%crep(1:3) = (/1.0_dp, 1.0_dp, 1.0_dp/)
         param%pol(1:3) = (/1.0_dp, 1.0_dp, 1.0_dp/)
         param%ag = 4.0_dp
         param%energy = -0.5_dp
!   CASE (10)
!      !Neon
!   CASE (11)
!      !Sodium
!   CASE (12)
!      !Magnesium
!   CASE (13)
!      !Aluminium
!   CASE (14)
!      !Silicon
!   CASE (15)
!      !Phosphorus
!   CASE (16)
!      !Sulphur
!   CASE (17)
!      !Chlorine
!   CASE (18)
!      !Argon
!   CASE (19)
!      !Potassium
!   CASE (20)
!      !Calcium
!   CASE (21)
!   CASE (22)
!   CASE (23)
!   CASE (24)
!   CASE (25)
!   CASE (26)
!   CASE (27)
!   CASE (28)
!   CASE (29)
!   CASE (30)
!      !Zinc
!   CASE (31)
!      !Gallium
!   CASE (32)
!      !Germanium
!   CASE (33)
!      !Arsenic
!   CASE (34)
!      !Selenium
!   CASE (35)
!      !Bromine
!   CASE (36)
!   CASE (37)
!      !Rubidium
!   CASE (38)
!      !Strontium
!   CASE (39)
!   CASE (40)
!   CASE (41)
!   CASE (42)
!   CASE (43)
!   CASE (44)
!   CASE (45)
!   CASE (46)
!   CASE (47)
!   CASE (48)
!      !Cadmium
!   CASE (49)
!      !Indium
!   CASE (50)
!      !Tin
!   CASE (51)
!      !Antimony
!   CASE (52)
!      !Tellurium
!   CASE (53)
!      !Iodine
!   CASE (54)
!   CASE (55)
!      !Cesium
!   CASE (56)
!      !Barium
!   CASE (57)
!   CASE (58)
!   CASE (59)
!   CASE (60)
!   CASE (61)
!   CASE (62)
!   CASE (63)
!   CASE (64)
!   CASE (65)
!   CASE (66)
!   CASE (67)
!   CASE (68)
!   CASE (69)
!   CASE (70)
!   CASE (71)
!   CASE (72)
!   CASE (73)
!   CASE (74)
!   CASE (75)
!   CASE (76)
!   CASE (77)
!   CASE (78)
!   CASE (79)
!   CASE (80)
!      !Mercury
!   CASE (81)
!      !Thallium
!   CASE (82)
!      !Lead
!   CASE (83)
!      !Bismuth
!   CASE (84)
!   CASE (85)
!   CASE (86)
!   CASE (87)
!   CASE (88)
!   CASE (89)
!   CASE (90)
!   CASE (91)
!   CASE (92)
!   CASE (93)
!   CASE (94)
!   CASE (95)
!   CASE (96)
!   CASE (97)
!   CASE (98)
!   CASE (99:nelem)
      END SELECT

      IF (param%defined) THEN
         param%natorb = 0
         DO l = 0, 3
            param%natorb = param%natorb+((l+1)*(l+2))/2*param%norb(l)
         END DO
         ! check atomic name
         i = param%z
         param%atomname = ptable(i)%symbol
      END IF

   END SUBROUTINE scptb_default_parameter

END MODULE scptb_parameters

! xmlparse.f90 - Simple, limited XML parser in Fortran
!
! $Id: xmlparse.f90,v 1.14 2007/12/07 10:10:19 arjenmarkus Exp $
!
! Arjen Markus
!
! General information:
! The module reads XML files by:
! - Identifying the tag and all attributes and data belonging
!   to the tag.
! - Returning to the calling subprogram to let it take care of
!   the tag, attributes and data.
! - If the tag is actually an ending tag, then this is flagged
!   too.
! - Handling all the data is left to the calling subprogram,
!   the module merely facilitates in the parsing.
!
! Note:
! The module in its current version has a number of limitations:
! - It does not handle escape sequences (like &gt. to signify
!   a ">" sign)
! - It does not handle tags with attributes that are spread
!   over more than one line
! - The maximum length of a line is 1000 characters
! - It may report too many lines of data (empty lines)
! - No DOM support nor support for an object tree
! - It is probably not very robust in detecting malformed XML files
!
! Some questions:
! - What to do with leading blanks?
!
! Update - several ideas:
! - Introduce at least two options (via xml_options):
!   - ignore_whitespace  - remove leading blanks and leading and trailing
!                          empty lines from the PCDATA
!   - no_data_truncation - consider truncation of data (more
!                          attributes or lines of character data than
!                          can be stored) a read error
! - Introduce convenience functions and subroutines:
!   - xml_ok()           - all is well, reading can continue
!   - xml_data_trunc()   - was there truncation of the data?
!   - xml_find_attrib()  - find an attribute by name
!
! Further ideas:
!   - simple checking via a table: parent, tag, id, min, max
!
MODULE xml_parser
   USE cp_files,                        ONLY: close_file,&
                                              file_exists,&
                                              open_file

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: XML_PARSE, xml_close, xml_get, xml_ok, xml_open

   INTEGER, PARAMETER :: XML_BUFFER_LENGTH = 1000

   ! Define the data type that holds the parser information

   TYPE XML_PARSE
      INTEGER          :: lun ! LU-number of the XML-file
      INTEGER          :: level ! Indentation level (output)
      INTEGER          :: lineno ! Line in file
      LOGICAL          :: ignore_whitespace ! Ignore leading blanks etc.
      LOGICAL          :: no_data_truncation ! Do not allow data truncation
      LOGICAL          :: too_many_attribs ! More attributes than could be stored?
      LOGICAL          :: too_many_data ! More lines of data than could be stored?
      LOGICAL          :: eof ! End of file?
      LOGICAL          :: error ! Invalid XML file or other error?
      CHARACTER(len=XML_BUFFER_LENGTH) :: line ! Buffer
   END TYPE XML_PARSE

   ! Global options

   INTEGER, PARAMETER    :: XML_STDOUT = -1
   INTEGER, PRIVATE      :: report_lun_ = XML_STDOUT
   LOGICAL, PRIVATE      :: report_errors_ = .FALSE.
   LOGICAL, PRIVATE      :: report_details_ = .FALSE.

   ! Global data (the ampersand must come first)

   CHARACTER(len=10), DIMENSION(2, 3), SAVE, PRIVATE :: entities = &
                                                        RESHAPE((/'&    ', '&amp;', &
                                                                  '>    ', '&gt; ', &
                                                                  '<    ', '&lt; '/), (/2, 3/))

   ! Auxiliary routines - private

   PRIVATE               :: xml_compress_
   PRIVATE               :: xml_put_open_tag_
   PRIVATE               :: xml_put_element_
   PRIVATE               :: xml_put_close_tag_
   PRIVATE               :: xml_replace_entities_
   !
   ! Interfaces to reporting routines
   !
   PRIVATE               :: xml_report_details_int_
   PRIVATE               :: xml_report_details_string_
   PRIVATE               :: xml_report_errors_int_
   PRIVATE               :: xml_report_errors_string_

   INTERFACE xml_report_details
      MODULE PROCEDURE xml_report_details_int_
      MODULE PROCEDURE xml_report_details_string_
   END INTERFACE
   INTERFACE xml_report_errors
      MODULE PROCEDURE xml_report_errors_int_
      MODULE PROCEDURE xml_report_errors_string_
      MODULE PROCEDURE xml_report_errors_extern_
   END INTERFACE

CONTAINS

! xml_report_details_int_ --
!    Routine to write a text with an integer value
! Arguments:
!    text        Text to be written
!    int         Integer value to be added
!
! **************************************************************************************************
!> \brief ...
!> \param text ...
!> \param int ...
! **************************************************************************************************
   SUBROUTINE xml_report_details_int_(text, int)
      CHARACTER(len=*), INTENT(in)                       :: text
      INTEGER, INTENT(in)                                :: int

      IF (report_details_) THEN
         IF (report_lun_ .EQ. XML_STDOUT) THEN
            WRITE (*, *) TRIM(text), int
         ELSE
            WRITE (report_lun_, *) TRIM(text), int
         ENDIF
      ENDIF
   END SUBROUTINE xml_report_details_int_

! xml_report_details_string_ --
!    Routine to write a text with a string value
! Arguments:
!    text        Text to be written
!    string      String to be added
!
! **************************************************************************************************
!> \brief ...
!> \param text ...
!> \param string ...
! **************************************************************************************************
   SUBROUTINE xml_report_details_string_(text, string)
      CHARACTER(len=*), INTENT(in)                       :: text, string

      IF (report_details_) THEN
         IF (report_lun_ .EQ. XML_STDOUT) THEN
            WRITE (*, *) TRIM(text), ' ', TRIM(string)
         ELSE
            WRITE (report_lun_, *) TRIM(text), ' ', TRIM(string)
         ENDIF
      ENDIF
   END SUBROUTINE xml_report_details_string_

! xml_report_errors_string_ --
!    Routine to write an error message text with an integer value
! Arguments:
!    text        Text to be written
!    int         Integer value to be added
!    lineno      Line number in the file
!
! **************************************************************************************************
!> \brief ...
!> \param text ...
!> \param int ...
!> \param lineno ...
! **************************************************************************************************
   SUBROUTINE xml_report_errors_int_(text, int, lineno)
      CHARACTER(len=*), INTENT(in)                       :: text
      INTEGER, INTENT(in)                                :: int
      INTEGER, INTENT(in), OPTIONAL                      :: lineno

      IF (report_errors_ .OR. report_details_) THEN
         IF (report_lun_ .EQ. XML_STDOUT) THEN
            WRITE (*, *) TRIM(text), int
            IF (PRESENT(lineno)) THEN
               WRITE (*, *) '   At or near line', lineno
            ENDIF
         ELSE
            WRITE (report_lun_, *) TRIM(text), int
            IF (PRESENT(lineno)) THEN
               WRITE (report_lun_, *) '   At or near line', lineno
            ENDIF
         ENDIF
      ENDIF
   END SUBROUTINE xml_report_errors_int_

! xml_report_errors_string_ --
!    Routine to write an error message text with a string value
! Arguments:
!    text        Text to be written
!    string      String to be added
!    lineno      Line number in the file
!
! **************************************************************************************************
!> \brief ...
!> \param text ...
!> \param string ...
!> \param lineno ...
! **************************************************************************************************
   SUBROUTINE xml_report_errors_string_(text, string, lineno)
      CHARACTER(len=*), INTENT(in)                       :: text, string
      INTEGER, INTENT(in), OPTIONAL                      :: lineno

      IF (report_errors_ .OR. report_details_) THEN
         IF (report_lun_ .EQ. XML_STDOUT) THEN
            WRITE (*, *) TRIM(text), ' ', TRIM(string)
            IF (PRESENT(lineno)) THEN
               WRITE (*, *) '   At or near line', lineno
            ENDIF
         ELSE
            WRITE (report_lun_, *) TRIM(text), ' ', TRIM(string)
            IF (PRESENT(lineno)) THEN
               WRITE (report_lun_, *) '   At or near line', lineno
            ENDIF
         ENDIF
      ENDIF
   END SUBROUTINE xml_report_errors_string_

! xml_report_errors_extern_ --
!    Routine to write an error message text with a string value
! Arguments:
!    info        Structure holding information on the XML-file
!    text        Text to be written
! Note:
!    This routine is meant for use by routines outside
!    this module
!
! **************************************************************************************************
!> \brief ...
!> \param info ...
!> \param text ...
! **************************************************************************************************
   SUBROUTINE xml_report_errors_extern_(info, text)
      TYPE(XML_PARSE), INTENT(in)                        :: info
      CHARACTER(len=*), INTENT(in)                       :: text

      IF (report_lun_ .EQ. XML_STDOUT) THEN
         WRITE (*, *) TRIM(text), ' - at or near line', info%lineno
      ELSE
         WRITE (report_lun_, *) TRIM(text), ' - at or near line', info%lineno
      ENDIF
   END SUBROUTINE xml_report_errors_extern_

! xml_open --
!    Routine to open an XML file for reading or writing
! Arguments:
!    info        Structure holding information on the XML-file
!    fname       Name of the file
!    mustread    The file will be read (.true.) or written (.false.)
!
! **************************************************************************************************
!> \brief ...
!> \param info ...
!> \param fname ...
!> \param mustread ...
! **************************************************************************************************
   SUBROUTINE xml_open(info, fname, mustread)
      TYPE(XML_PARSE), INTENT(out)                       :: info
      CHARACTER(len=*), INTENT(in)                       :: fname
      LOGICAL, INTENT(in)                                :: mustread

      INTEGER                                            :: ierr, k, kend

      info%lun = 10
      info%ignore_whitespace = .FALSE.
      info%no_data_truncation = .FALSE.
      info%too_many_attribs = .FALSE.
      info%too_many_data = .FALSE.
      info%eof = .FALSE.
      info%error = .FALSE.
      info%level = -1
      info%lineno = 0

      IF (.NOT. file_exists(fname) .AND. mustread) THEN
         CALL xml_report_errors('XML_OPEN: file does not exist:', TRIM(fname))
         info%lun = -1
         info%error = .TRUE.
      ELSE
         CALL open_file(file_name=fname, unit_number=info%lun)
         CALL xml_report_details('XML_OPEN: opened file ', TRIM(fname))
         CALL xml_report_details('at LU-number: ', info%lun)
      ENDIF

      IF (.NOT. info%error .AND. mustread) THEN
         k = 1
         DO WHILE (k .GE. 1)
            READ (info%lun, '(a)', iostat=ierr) info%line
            IF (ierr .EQ. 0) THEN
               info%line = ADJUSTL(info%line)
               k = INDEX(info%line, '<?')
               !
               ! Assume (for now at least) that <?xml ... ?> appears on a single line!
               !
               IF (k .GE. 1) THEN
                  kend = INDEX(info%line, '?>')
                  IF (kend .LE. 0) THEN
                     CALL xml_report_errors('XML_OPEN: error reading file with LU-number: ', info%lun)
                     CALL xml_report_errors('Line starting with "<?xml" should end with "?>"', ' ')
                     info%error = .TRUE.
                     EXIT
                  ENDIF
               ENDIF
            ELSE
               CALL xml_report_errors('XML_OPEN: error reading file with LU-number: ', info%lun)
               CALL xml_report_errors('Possibly no line starting with "<?xml"', ' ')
               CALL xml_close(info)
               info%error = .TRUE.
               EXIT
            ENDIF
         ENDDO
      ENDIF
      IF (.NOT. info%error .AND. .NOT. mustread) THEN
         WRITE (info%lun, '(a)') '<?xml version="1.0"?>'
      ENDIF
   END SUBROUTINE xml_open

! xml_close --
!    Routine to close an XML file
! Arguments:
!    info        Structure holding information on the XML-file
!
! **************************************************************************************************
!> \brief ...
!> \param info ...
! **************************************************************************************************
   SUBROUTINE xml_close(info)
      TYPE(XML_PARSE), INTENT(inout)                     :: info

      CALL close_file(info%lun)

      !
      ! Only clean up the LU-number, so that the calling program
      ! can examine the last condition
      !
      CALL xml_report_details('XML_CLOSE: Closing file with LU-number ', info%lun)
      info%lun = -1
   END SUBROUTINE xml_close

! xml_get --
!    Routine to get the next bit of information from an XML file
! Arguments:
!    info        Structure holding information on the XML-file
!    tag         Tag that was encountered
!    endtag      Whether the end of the element was encountered
!    attribs     List of attribute-value pairs
!    no_attribs  Number of pairs in the list
!    data        Lines of character data found
!    no_data     Number of lines of character data
!
! **************************************************************************************************
!> \brief ...
!> \param info ...
!> \param tag ...
!> \param endtag ...
!> \param attribs ...
!> \param no_attribs ...
!> \param DATA ...
!> \param no_data ...
! **************************************************************************************************
   SUBROUTINE xml_get(info, tag, endtag, attribs, no_attribs, &
                      DATA, no_data)
      TYPE(XML_PARSE), INTENT(inout)                     :: info
      CHARACTER(len=*), INTENT(out)                      :: tag
      LOGICAL, INTENT(out)                               :: endtag
      CHARACTER(len=*), DIMENSION(:, :), INTENT(out)     :: attribs
      INTEGER, INTENT(out)                               :: no_attribs
      CHARACTER(len=*), DIMENSION(:), INTENT(out)        :: DATA
      INTEGER, INTENT(out)                               :: no_data

      CHARACTER(len=XML_BUFFER_LENGTH)                   :: nextline
      INTEGER                                            :: idxat, idxdat, ierr, kend, keq, kfirst, &
                                                            ksecond, kspace
      LOGICAL                                            :: close_bracket, comment_tag

!
! Initialise the output
!

      endtag = .FALSE.
      no_attribs = 0
      no_data = 0

      info%too_many_attribs = .FALSE.
      info%too_many_data = .FALSE.

      IF (info%lun .LT. 0) THEN
         CALL xml_report_details('XML_GET on closed file ', ' ')
         RETURN
      ENDIF

      !
      ! From the previous call or the call to xmlopen we have
      ! the line that we need to parse already in memory:
      ! <tag attrib1="..." attrib2="..." />
      !
      comment_tag = .FALSE.
      close_bracket = .FALSE.
      kspace = INDEX(info%line, ' ')
      kend = INDEX(info%line, '>')
      DO WHILE (kend .LE. 0)
         info%lineno = info%lineno+1
         READ (info%lun, '(a)', iostat=ierr) nextline
         IF (ierr .EQ. 0) THEN
            info%line = TRIM(info%line)//' '//ADJUSTL(nextline)
         ELSE
            info%error = .TRUE.
            CALL xml_report_errors('XML_GET - end of tag not found ', &
                                   '(buffer too small?)', info%lineno)
            CALL xml_close(info)
            RETURN
         ENDIF
         kend = INDEX(info%line, '>')
      ENDDO
      IF (kend .GT. kspace) THEN
         kend = kspace
      ELSE
         close_bracket = .TRUE.
      ENDIF

      !
      ! Check for the end of an ordianry tag and of
      ! a comment tag
      !
      IF (info%line(1:3) .EQ. '-->') THEN
         endtag = .TRUE.
         tag = info%line(4:kend-1)
      ELSE IF (info%line(1:2) .EQ. '</') THEN
         endtag = .TRUE.
         tag = info%line(3:kend-1)
      ELSE
         IF (info%line(1:1) .EQ. '<') THEN
            tag = info%line(2:kend-1)
            CALL xml_report_details('XML_GET - tag found: ', TRIM(tag))
         ELSE
            kend = 0 ! Beginning of data!
         ENDIF
      ENDIF

      info%line = ADJUSTL(info%line(kend+1:))

      idxat = 0
      idxdat = 0

      IF (tag(1:3) .EQ. '!--') comment_tag = .TRUE.

      DO WHILE (info%line .NE. ' ' .AND. .NOT. close_bracket .AND. .NOT. comment_tag)

         keq = INDEX(info%line, '=')
         kend = INDEX(info%line, '>')
         IF (keq .GT. kend) keq = 0 ! Guard against multiple tags
         ! with attributes on one line

         !
         ! No attributes any more?
         !
         IF (keq .LT. 1) THEN
            kend = INDEX(info%line, '/>')
            IF (kend .GE. 1) THEN
               kend = kend+1 ! To go beyond the ">" character
               endtag = .TRUE.
            ELSE
               kend = INDEX(info%line, '>')
               IF (kend .LT. 1) THEN
                  CALL xml_report_errors('XML_GET - wrong ending of tag ', &
                                         TRIM(info%line), info%lineno)
                  info%error = .TRUE. ! Wrong ending of line!
                  CALL xml_close(info)
                  RETURN
               ELSE
                  close_bracket = .TRUE.
               ENDIF
            ENDIF
            IF (kend .GE. 1) THEN
               info%line = ADJUSTL(info%line(kend+1:))
            ENDIF
            EXIT
         ENDIF

         idxat = idxat+1
         IF (idxat .LE. SIZE(attribs, 2)) THEN
            no_attribs = idxat
            attribs(1, idxat) = ADJUSTL(info%line(1:keq-1)) ! Use adjustl() to avoid
            ! multiple spaces, etc
            info%line = ADJUSTL(info%line(keq+1:))

            !
            ! We have almost found the start of the attribute's value
            !
            kfirst = INDEX(info%line, '"')
            IF (kfirst .LT. 1) THEN
               CALL xml_report_errors('XML_GET - malformed attribute-value pair: ', &
                                      TRIM(info%line), info%lineno)
               info%error = .TRUE. ! Wrong form of attribute-value pair
               CALL xml_close(info)
               RETURN
            ENDIF

            ksecond = INDEX(info%line(kfirst+1:), '"')+kfirst
            IF (ksecond .LT. 1) THEN
               CALL xml_report_errors('XML_GET - malformed attribute-value pair: ', &
                                      TRIM(info%line), info%lineno)
               info%error = .TRUE. ! Wrong form of attribute-value pair
               CALL xml_close(info)
               RETURN
            ENDIF

            attribs(2, idxat) = info%line(kfirst+1:ksecond-1)
            info%line = ADJUSTL(info%line(ksecond+1:))
         ENDIF

         IF (idxat .GT. SIZE(attribs, 2)) THEN
            CALL xml_report_errors('XML_GET - more attributes than could be stored: ', &
                                   TRIM(info%line), info%lineno)
            info%too_many_attribs = .TRUE.
            info%line = ' '
            EXIT
         ENDIF
      ENDDO

      !
      ! Now read the data associated with the current tag
      ! - all the way to the next "<" character
      !
      ! To do: reduce the number of data lines - empty ones
      ! at the end should not count.
      !
      DO
         IF (comment_tag) THEN
            kend = INDEX(info%line, '-->')
         ELSE
            kend = INDEX(info%line, '<')
         ENDIF
         idxdat = idxdat+1
         IF (idxdat .LE. SIZE(DATA)) THEN
            no_data = idxdat
            IF (kend .GE. 1) THEN
               DATA(idxdat) = info%line(1:kend-1)
               info%line = info%line(kend:)
            ELSE
               DATA(idxdat) = info%line
            ENDIF
         ELSE
            CALL xml_report_errors('XML_GET - more data lines than could be stored: ', &
                                   TRIM(info%line), info%lineno)
            info%too_many_data = .TRUE.
            EXIT
         ENDIF

         !
         ! No more data? Otherwise, read on
         !
         IF (kend .GE. 1) THEN
            EXIT
         ELSE
            info%lineno = info%lineno+1
            READ (info%lun, '(a)', iostat=ierr) info%line
            IF (ierr .LT. 0) THEN
               CALL xml_report_details('XML_GET - end of file found - LU-number: ', &
                                       info%lun)
               info%eof = .TRUE.
            ELSEIF (ierr .GT. 0) THEN
               CALL xml_report_errors('XML_GET - error reading file with LU-number ', &
                                      info%lun, info%lineno)
               info%error = .TRUE.
            ENDIF
            IF (ierr .NE. 0) THEN
               EXIT
            ENDIF
         ENDIF
      ENDDO

      !
      ! Compress the data?
      !
      IF (info%ignore_whitespace) THEN
         CALL xml_compress_(DATA, no_data)
      ENDIF

      !
      ! Replace the entities, if any
      !
      CALL xml_replace_entities_(DATA, no_data)

      CALL xml_report_details('XML_GET - number of attributes: ', no_attribs)
      CALL xml_report_details('XML_GET - number of data lines: ', no_data)

   END SUBROUTINE xml_get

! **************************************************************************************************
!> \brief Routine to write a tag with the associated data to an XML file
!> \param info Structure holding information on the XML-file
!> \param tag Tag that was encountered
!> \param attribs List of attribute-value pairs
!> \param no_attribs Number of pairs in the list
!> \param DATA Lines of character data found
!> \param no_data Number of lines of character data
!> \param TYPE ype of action:
!>                open - just the opening tag with attributes
!>                elem - complete element
!>                close - just the closing tag
! **************************************************************************************************
   SUBROUTINE xml_put(info, tag, attribs, no_attribs, &
                      DATA, no_data, TYPE)

      TYPE(XML_PARSE), INTENT(inout)                     :: info
      CHARACTER(len=*), INTENT(in)                       :: tag
      CHARACTER(len=*), DIMENSION(:, :), INTENT(in)      :: attribs
      INTEGER, INTENT(in)                                :: no_attribs
      CHARACTER(len=*), DIMENSION(:), INTENT(in)         :: DATA
      INTEGER, INTENT(in)                                :: no_data
      CHARACTER(len=*)                                   :: TYPE

      SELECT CASE (TYPE)
      CASE ('open')
         CALL xml_put_open_tag_(info, tag, attribs, no_attribs)
      CASE ('elem')
         CALL xml_put_element_(info, tag, attribs, no_attribs, &
                               DATA, no_data)
      CASE ('close')
         CALL xml_put_close_tag_(info, tag)
      END SELECT

   END SUBROUTINE xml_put

! **************************************************************************************************
!> \brief Routine to write the opening tag with the attributes
!> \param info Structure holding information on the XML-file
!> \param tag Tag that was encountered
!> \param attribs List of attribute-value pairs
!> \param no_attribs Number of pairs in the list
! **************************************************************************************************
   SUBROUTINE xml_put_open_tag_(info, tag, attribs, no_attribs)

      TYPE(XML_PARSE), INTENT(inout)                     :: info
      CHARACTER(len=*), INTENT(in)                       :: tag
      CHARACTER(len=*), DIMENSION(:, :), INTENT(in)      :: attribs
      INTEGER, INTENT(in)                                :: no_attribs

      CHARACTER(len=300), PARAMETER                      :: indent = ' '

      INTEGER                                            :: i

      WRITE (info%lun, '(3a)', advance='no') &
         indent(1:3*info%level), '<', ADJUSTL(tag)
      DO i = 1, no_attribs
         IF (attribs(2, i) .NE. '') THEN
            WRITE (info%lun, '(5a)', advance='no') &
               ' ', TRIM(attribs(1, i)), '="', TRIM(attribs(2, i)), '"'
         ENDIF
      ENDDO
      WRITE (info%lun, '(a)') '>'
      info%level = info%level+1

   END SUBROUTINE xml_put_open_tag_

! **************************************************************************************************
!> \brief Routine to write the complete element
!> \param info Structure holding information on the XML-file
!> \param tag Tag that was encountered
!> \param attribs List of attribute-value pairs
!> \param no_attribs Number of pairs in the list
!> \param DATA Lines of character data found
!> \param no_data Number of lines of character data
! **************************************************************************************************
   SUBROUTINE xml_put_element_(info, tag, attribs, no_attribs, &
                               DATA, no_data)

      TYPE(XML_PARSE), INTENT(inout)                     :: info
      CHARACTER(len=*), INTENT(in)                       :: tag
      CHARACTER(len=*), DIMENSION(:, :), INTENT(in)      :: attribs
      INTEGER, INTENT(in)                                :: no_attribs
      CHARACTER(len=*), DIMENSION(:), INTENT(in)         :: DATA
      INTEGER, INTENT(in)                                :: no_data

      CHARACTER(len=300), PARAMETER                      :: indent = ' '

      CHARACTER(len=1)                                   :: aa
      INTEGER                                            :: i, ii
      LOGICAL                                            :: logic

      IF ((no_attribs .EQ. 0 .AND. no_data .EQ. 0)) THEN
         RETURN
      ELSE
         logic = .TRUE.
         DO ii = 1, no_attribs
            logic = logic .AND. (attribs(2, ii) .EQ. '')
         ENDDO
         DO ii = 1, no_data
            logic = logic .AND. (DATA(ii) .EQ. '')
         ENDDO
         IF (logic) THEN
            RETURN
         ELSE
            WRITE (info%lun, '(3a)', advance='no') &
               indent(1:3*info%level), '<', ADJUSTL(tag)
            DO i = 1, no_attribs
               IF (attribs(2, i) .NE. '') THEN
                  WRITE (info%lun, '(5a)', advance='no') &
                     ' ', TRIM(attribs(1, i)), '="', TRIM(attribs(2, i)), '"'
               ENDIF
            ENDDO
            IF (no_attribs .GT. 0 .AND. no_data .EQ. 0) THEN
               aa = 'a'
            ELSEIF ((no_attribs .GT. 0 .AND. no_data .GT. 0) .OR. &
                    (no_attribs .EQ. 0 .AND. no_data .GT. 0)) THEN
               aa = 'b'
            ELSE
               WRITE (*, *) no_attribs, no_data
            ENDIF
         ENDIF
      ENDIF

      SELECT CASE (aa)
      CASE ('a')
         WRITE (info%lun, '(a)') '/>'
      CASE ('b')
         WRITE (info%lun, '(a)', advance='no') '>'
         WRITE (info%lun, '(2a)', advance='no') &
            (' ', TRIM(DATA(i)), i=1, no_data)
         WRITE (info%lun, '(4a)') ' ', '</', tag, '>'
      END SELECT

   END SUBROUTINE xml_put_element_

! **************************************************************************************************
!> \brief Routine to write the closing tag
!> \param info Structure holding information on the XML-file
!> \param tag Tag that was encountered
! **************************************************************************************************
   SUBROUTINE xml_put_close_tag_(info, tag)

      TYPE(XML_PARSE), INTENT(inout)                     :: info
      CHARACTER(len=*), INTENT(in)                       :: tag

      CHARACTER(len=300), PARAMETER                      :: indent = ' '

      info%level = info%level-1
      WRITE (info%lun, '(4a)') &
         indent(1:3*info%level), '</', ADJUSTL(tag), '>'

   END SUBROUTINE xml_put_close_tag_

! **************************************************************************************************
!> \brief Routine to remove empty lines from the character data
!> \param DATA Lines of character data found
!> \param no_data (Nett) number of lines of character data
! **************************************************************************************************
   SUBROUTINE xml_compress_(DATA, no_data)
      CHARACTER(len=*), DIMENSION(:), INTENT(inout)      :: DATA
      INTEGER, INTENT(inout)                             :: no_data

      INTEGER                                            :: i, j
      LOGICAL                                            :: empty

      j = 0
      empty = .TRUE.
      DO i = 1, no_data
         IF (LEN_TRIM(DATA(i)) .NE. 0 .OR. .NOT. empty) THEN
            j = j+1
            DATA(j) = ADJUSTL(DATA(i))
            empty = .FALSE.
         ENDIF
      ENDDO

      no_data = j

      DO i = no_data, 1, -1
         IF (LEN_TRIM(DATA(i)) .NE. 0) THEN
            EXIT
         ELSE
            no_data = no_data-1
         ENDIF
      ENDDO

   END SUBROUTINE xml_compress_

! **************************************************************************************************
!> \brief Routine to replace entities such as &gt; by their
!>        proper character representation
!> \param DATA Lines of character data found
!> \param no_data (Nett) number of lines of character data
! **************************************************************************************************
   SUBROUTINE xml_replace_entities_(DATA, no_data)
      CHARACTER(len=*), DIMENSION(:), INTENT(inout)      :: DATA
      INTEGER, INTENT(inout)                             :: no_data

      INTEGER                                            :: i, j, j2, k, pos
      LOGICAL                                            :: found

      DO i = 1, no_data
         j = 1
         DO
            DO k = 1, SIZE(entities, 2)
               found = .FALSE.
               pos = INDEX(DATA(i) (j:), TRIM(entities(2, k)))
               IF (pos .GT. 0) THEN
                  found = .TRUE.
                  j = j+pos-1
                  j2 = j+LEN_TRIM(entities(2, k))
                  DATA(i) (j:) = TRIM(entities(1, k))//DATA(i) (j2:)
                  j = j2
               ENDIF
            ENDDO
            IF (.NOT. found) EXIT
         ENDDO
      ENDDO

   END SUBROUTINE xml_replace_entities_

! **************************************************************************************************
!> \brief Routine to handle the parser options
!> \param info Structure holding information on the XML-file
!> \param ignore_whitespace Ignore whitespace (leading blanks, empty lines) or not
!> \param no_data_truncation Consider truncation of strings an error or not
!> \param report_lun LU-number for reporting information
!> \param report_errors Write messages about errors or not
!> \param report_details Write messages about all kinds of actions or not
! **************************************************************************************************
   SUBROUTINE xml_options(info, ignore_whitespace, no_data_truncation, &
                          report_lun, report_errors, &
                          report_details)
      TYPE(XML_PARSE), INTENT(inout)                     :: info
      LOGICAL, INTENT(in), OPTIONAL                      :: ignore_whitespace, no_data_truncation
      INTEGER, INTENT(in), OPTIONAL                      :: report_lun
      LOGICAL, INTENT(in), OPTIONAL                      :: report_errors, report_details

      IF (PRESENT(ignore_whitespace)) THEN
         info%ignore_whitespace = ignore_whitespace
      ENDIF
      IF (PRESENT(no_data_truncation)) THEN
         info%no_data_truncation = no_data_truncation
      ENDIF
      IF (PRESENT(report_lun)) THEN
         report_lun_ = report_lun
      ENDIF
      IF (PRESENT(report_errors)) THEN
         report_errors_ = report_errors
      ENDIF
      IF (PRESENT(report_details)) THEN
         report_details_ = report_details
      ENDIF
   END SUBROUTINE xml_options

! **************************************************************************************************
!> \brief Function that returns whether all was okay or not
!> \param info Structure holding information on the XML-file
!> \retval xml_ok .true. if there was no error, .false. otherwise
! **************************************************************************************************
   LOGICAL FUNCTION xml_ok(info)
      TYPE(XML_PARSE), INTENT(in)               :: info

      xml_ok = info%eof .OR. info%error .OR. &
               (info%no_data_truncation .AND. &
                (info%too_many_attribs .OR. info%too_many_data))
      xml_ok = .NOT. xml_ok
   END FUNCTION xml_ok

! **************************************************************************************************
!> \brief Function that returns whether there was an error
!> \param info Structure holding information on the XML-file
!> \retval xml_error .true. if there was an error, .false. if there was none
! **************************************************************************************************
   LOGICAL FUNCTION xml_error(info)
      TYPE(XML_PARSE), INTENT(in)               :: info

      xml_error = info%error .OR. &
                  (info%no_data_truncation .AND. &
                   (info%too_many_attribs .OR. info%too_many_data))
   END FUNCTION xml_error

! **************************************************************************************************
!> \brief Function that returns whether data were truncated or not
!> \param info Structure holding information on the XML-file
!> \retval xml_data_trunc .true. if data were truncated, .false. otherwise
! **************************************************************************************************
   LOGICAL FUNCTION xml_data_trunc(info)
      TYPE(XML_PARSE), INTENT(in)               :: info

      xml_data_trunc = info%too_many_attribs .OR. info%too_many_data
   END FUNCTION xml_data_trunc

! **************************************************************************************************
!> \brief ...
!> \param attribs ...
!> \param no_attribs ...
!> \param name ...
!> \param value ...
!> \retval xml_find_attrib ...
! **************************************************************************************************
   INTEGER FUNCTION xml_find_attrib(attribs, no_attribs, name, value)
      CHARACTER(len=*), DIMENSION(:, :)  :: attribs
      INTEGER                           :: no_attribs
      CHARACTER(len=*)                  :: name
      CHARACTER(len=*)                  :: value

      INTEGER :: i

      xml_find_attrib = -1
      DO i = 1, no_attribs
         IF (name .EQ. attribs(1, i)) THEN
            value = attribs(2, i)
            xml_find_attrib = i
            EXIT
         ENDIF
      ENDDO

   END FUNCTION xml_find_attrib

! **************************************************************************************************
!> \brief Routine to read the XML file as a whole and distribute processing
!>        the contents over three user-defined subroutines
!> \param filename Name of the file to process
!> \param attribs Array for holding the attributes
!> \param DATA Array for holding the character data
!> \param startfunc Subroutine to handle the start of elements
!> \param datafunc Subroutine to handle the character data
!> \param endfunc Subroutine to handle the end of elements
!> \param lunrep ...
!> \param error Indicates if there was an error or not
! **************************************************************************************************
   RECURSIVE SUBROUTINE xml_process(filename, attribs, DATA, startfunc, datafunc, endfunc, lunrep, error)
      CHARACTER(len=*)                                   :: filename
      CHARACTER(len=*), DIMENSION(:, :)                  :: attribs
      CHARACTER(len=*), DIMENSION(:)                     :: DATA
      INTERFACE
         RECURSIVE SUBROUTINE startfunc(tag, attribs, error)
            CHARACTER(len=*)                  :: tag
            CHARACTER(len=*), DIMENSION(:, :)  :: attribs
            LOGICAL                           :: error
         END SUBROUTINE
      END INTERFACE
      INTERFACE
         RECURSIVE SUBROUTINE datafunc(tag, DATA, error)
            CHARACTER(len=*)                  :: tag
            CHARACTER(len=*), DIMENSION(:)    :: DATA
            LOGICAL                           :: error
         END SUBROUTINE
      END INTERFACE
      INTERFACE
         RECURSIVE SUBROUTINE endfunc(tag, error)
            CHARACTER(len=*)                  :: tag
            LOGICAL                           :: error
         END SUBROUTINE
      END INTERFACE
      INTEGER                                            :: lunrep
      LOGICAL                                            :: error

      CHARACTER(len=80)                                  :: tag
      INTEGER                                            :: noattribs, nodata
      LOGICAL                                            :: endtag
      TYPE(XML_PARSE)                                    :: info

      CALL xml_options(info, report_lun=lunrep, report_details=.FALSE.)
      CALL xml_open(info, filename, .TRUE.)

      error = .FALSE.
      DO
         CALL xml_get(info, tag, endtag, attribs, noattribs, DATA, nodata)
         IF (.NOT. xml_ok(info)) THEN
            EXIT
         ENDIF

         IF (xml_error(info)) THEN
            WRITE (lunrep, *) 'Error reading XML file!'
            error = .TRUE.
            EXIT
         ENDIF

         IF (.NOT. endtag .OR. noattribs .NE. 0) THEN
            CALL startfunc(tag, attribs(:, 1:noattribs), error)
            IF (error) EXIT

            CALL datafunc(tag, DATA(1:nodata), error)
            IF (error) EXIT
         ENDIF

         IF (endtag) THEN
            CALL endfunc(tag, error)
            IF (error) EXIT
         ENDIF
      ENDDO
      CALL xml_close(info)
   END SUBROUTINE xml_process

END MODULE xml_parser
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Routines for SCPTB
!> \author JGH (12.2011), RK (07.2012)
! **************************************************************************************************
MODULE scptb_utils

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_container_types,       ONLY: add_basis_set_to_container
   USE basis_set_types,                 ONLY: allocate_sto_basis_set,&
                                              create_gto_from_sto_basis,&
                                              deallocate_sto_basis_set,&
                                              gto_basis_set_type,&
                                              set_sto_basis_set,&
                                              sto_basis_set_type
   USE cp_control_types,                ONLY: scptb_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE external_potential_types,        ONLY: set_potential
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE message_passing,                 ONLY: mp_bcast
   USE qs_kind_types,                   ONLY: qs_kind_type,&
                                              set_qs_kind
   USE scptb_parameters,                ONLY: scptb_default_parameter
   USE scptb_types,                     ONLY: allocate_scptb_parameter,&
                                              scptb_parameter_type,&
                                              set_scptb_parameter
   USE string_utilities,                ONLY: uppercase
   USE xml_parser,                      ONLY: XML_PARSE,&
                                              xml_close,&
                                              xml_get,&
                                              xml_ok,&
                                              xml_open
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'scptb_utils'

   PUBLIC :: scptb_parameter_init

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param scptb_control ...
!> \param print_section ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE scptb_parameter_init(atomic_kind_set, qs_kind_set, scptb_control, print_section, para_env)

      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(scptb_control_type), INTENT(INOUT)            :: scptb_control
      TYPE(section_vals_type), POINTER                   :: print_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scptb_parameter_init', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=6), DIMENSION(:), POINTER            :: symbol
      CHARACTER(LEN=default_string_length)               :: iname
      INTEGER                                            :: i, ikind, is, l, nkind, nshell, &
                                                            output_unit
      INTEGER, DIMENSION(:), POINTER                     :: lq, nq
      LOGICAL                                            :: print_info
      REAL(KIND=dp)                                      :: a, c, d, eff, r, rmax, rmin
      REAL(KIND=dp), DIMENSION(:), POINTER               :: zet
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(gto_basis_set_type), POINTER                  :: gto_basis
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(scptb_parameter_type), POINTER                :: scptb_parameter
      TYPE(sto_basis_set_type), POINTER                  :: sto_basis

      output_unit = -1
      NULLIFY (logger)
      logger => cp_get_default_logger()
      print_info = (BTEST(cp_print_key_should_output(logger%iter_info, print_section, &
                                                     "KINDS/BASIS_SET"), cp_p_file))
      IF (print_info) THEN
         output_unit = cp_print_key_unit_nr(logger, print_section, "KINDS", extension=".Log")
         IF (output_unit > 0) THEN
            WRITE (output_unit, "(/,A)") " SCPTB| A set of SCPTB "// &
               "parameters for material sciences."
            WRITE (output_unit, "(A)") " SCPTB| J. Hutter, Y. Misteli, R. Koitz"
            WRITE (output_unit, "(A)") " SCPTB| University of Zurich, 2012"
         END IF
         CALL cp_print_key_finished_output(output_unit, logger, print_section, "KINDS")
      END IF

      ! read global parameters from xml file
      IF (para_env%ionode) CALL scptb_global_parameter_from_file(scptb_control)
      CALL scptb_global_parameter_broadcast(scptb_control, para_env)

      nkind = SIZE(atomic_kind_set)
      DO ikind = 1, nkind
         CALL get_atomic_kind(atomic_kind_set(ikind), name=iname)
         CALL uppercase(iname)
         NULLIFY (scptb_parameter)
         CALL allocate_scptb_parameter(scptb_parameter)
         CALL set_scptb_parameter(scptb_parameter, key=iname, defined=.FALSE.)

         ! only for ionode
         IF (para_env%ionode) CALL scptb_parameter_from_file(scptb_parameter, scptb_control)
         CALL scptb_parameter_broadcast(scptb_parameter, para_env)

         IF (scptb_parameter%defined .EQV. .FALSE.) THEN
            output_unit = cp_print_key_unit_nr(logger, print_section, "KINDS", extension=".Log")
            IF (output_unit > 0) THEN
               WRITE (output_unit, "(A,I3,2X,A,A)") " SCPTB| Kind ", ikind, TRIM(iname), &
                  ": No parameter xml given or element undefined. Using defaults."
            END IF
            !defaults, but undefined for most elements atm
            CALL scptb_default_parameter(scptb_parameter)
            CALL cp_print_key_finished_output(output_unit, logger, print_section, "KINDS")
         END IF

         !initialization

         IF (scptb_parameter%defined) THEN
            CALL set_potential(potential=qs_kind_set(ikind)%all_potential, zeff=scptb_parameter%zeff, &
                               zeff_correction=0.0_dp)
            ! basis set
            NULLIFY (sto_basis)
            CALL allocate_sto_basis_set(sto_basis)
            nshell = SUM(scptb_parameter%norb)
            ALLOCATE (nq(nshell), lq(nshell), zet(nshell))
            nq = 0
            lq = 0
            zet = 0._dp
            ALLOCATE (symbol(nshell))
            symbol = ""
            is = 0
            !Aggregate all the electron shell parameters into arrays
            DO l = 0, scptb_parameter%lmaxorb
               DO i = 1, scptb_parameter%norb(l)
                  is = is+1
                  zet(is) = scptb_parameter%zeta(i, l)
                  lq(is) = l
                  nq(is) = scptb_parameter%nqm(i, l)
                  IF (l == 0) WRITE (symbol(is), '(I1,A1)') nq(is), "S"
                  IF (l == 1) WRITE (symbol(is), '(I1,A1)') nq(is), "P"
                  IF (l == 2) WRITE (symbol(is), '(I1,A1)') nq(is), "D"
                  IF (l == 3) WRITE (symbol(is), '(I1,A1)') nq(is), "F"
               END DO
            END DO

            IF (nshell > 0) THEN
               CALL set_sto_basis_set(sto_basis, name=scptb_parameter%atomname, &
                                      nshell=nshell, symbol=symbol, nq=nq, lq=lq, zet=zet)
               qs_kind => qs_kind_set(ikind)
               NULLIFY (gto_basis)
               CALL create_gto_from_sto_basis(sto_basis, gto_basis, scptb_control%sto_ng)
               CALL add_basis_set_to_container(qs_kind%basis_sets, gto_basis, "ORB")
            END IF

            DEALLOCATE (nq, lq, zet, symbol)

            CALL deallocate_sto_basis_set(sto_basis)
            !throw away the STO basis set, since from now on we use the GTO

            a = scptb_parameter%crep(1)
            c = scptb_parameter%crep(2)
            d = scptb_parameter%crep(3)

            rmin = 1._dp
            rmax = 100._dp
            DO
               r = 0.5_dp*(rmax+rmin)
               eff = (c+d*r**2)*EXP(-(a*r))
               IF (eff < scptb_control%epspair) THEN
                  rmax = r
               ELSE
                  rmin = r
               END IF
               IF ((rmax-rmin) < 1.e-2) EXIT
            END DO
            scptb_parameter%rcpair = 0.5_dp*rmax

            !lmaxscp: make sure it is also consistent with pure SCC-TB
            IF (scptb_parameter%lmaxscp < 0) THEN
               IF (scptb_parameter%ag > 0.0_dp) THEN
                  scptb_parameter%lmaxscp = 0
               END IF
            END IF

         END IF

         CALL set_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_parameter)

      END DO

   END SUBROUTINE scptb_parameter_init

! **************************************************************************************************
!> \brief ...
!> \param scptb_control ...
! **************************************************************************************************
   SUBROUTINE scptb_global_parameter_from_file(scptb_control)

      TYPE(scptb_control_type), INTENT(INOUT)            :: scptb_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scptb_global_parameter_from_file', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: attribs(2, 10), author, DATA(10), fname, &
                                                            found_key, tag, vers
      INTEGER                                            :: no_attribs, no_data, tag_count, &
                                                            unrecognized_tags
      LOGICAL                                            :: endtag, has_A, has_ag, has_B, has_C, &
                                                            has_D, has_energy, has_eta, has_hcore, &
                                                            has_nqm, has_occ, has_z, has_zeff, &
                                                            tag_over
      TYPE(XML_PARSE)                                    :: xml_info

      DATA has_z, has_zeff, has_ag, has_energy, has_hcore, has_occ, &
         has_nqm, has_eta, has_A, has_B, has_C, has_D/12*.FALSE./

      fname = scptb_control%parameter_file
      IF (fname /= "") THEN
         CALL xml_open(xml_info, fname, .TRUE.)

         tag_count = 0
         tag = ""
         DATA = ""
         attribs = ""
         unrecognized_tags = 0
         tag_over = .FALSE.
         found_key = ""
         author = "unknown author"
         vers = " v 0.0"

         DO WHILE (xml_ok(xml_info)) !Until the file is read entirely
            CALL xml_get(xml_info, tag, endtag, attribs, no_attribs, DATA, no_data)
            CALL uppercase(tag)
            SELECT CASE (TRIM (tag)) !Decide on the first-level tag
            CASE ("")
               !just garbage from the parser
            CASE ("NM12_PARAMETERS")
               !just the header, do nothing
            CASE ("META_INFO")
               tag_over = .FALSE.
               DO WHILE (tag_over .EQV. .FALSE.)
                  CALL xml_get(xml_info, tag, endtag, attribs, no_attribs, DATA, no_data)
                  CALL uppercase(tag)
                  SELECT CASE (tag)
                  CASE ("META_INFO")
                     IF (endtag .EQV. .TRUE.) tag_over = .TRUE.
                  CASE ("VERSION")
                     IF (endtag .EQV. .FALSE.) vers = DATA(1)
                  CASE ("AUTHOR")
                     IF (endtag .EQV. .FALSE.) author = DATA(1)
                  CASE ("REF")
                     !IF (endtag .eqv. .FALSE.)  !do nothing
                  CASE ("COMMENT")
                     !IF (endtag .eqv. .FALSE.)  !do nothing
                  CASE DEFAULT
                     unrecognized_tags = unrecognized_tags+1
                  END SELECT
               END DO

            CASE ("GENERAL_PARAMETERS") !general params
               tag_over = .FALSE.
               DO WHILE (tag_over .EQV. .FALSE.)
                  CALL xml_get(xml_info, tag, endtag, attribs, no_attribs, DATA, no_data)
                  CALL uppercase(tag)
                  SELECT CASE (tag)
                  CASE ("GENERAL_PARAMETERS")
                     IF (endtag .EQV. .TRUE.) tag_over = .TRUE.
                  CASE ("D3_S6")
                     IF (endtag .EQV. .FALSE.) READ (DATA(1), *) scptb_control%sd3(1)
                  CASE ("D3_SR6")
                     IF (endtag .EQV. .FALSE.) READ (DATA(1), *) scptb_control%sd3(2)
                  CASE ("D3_S8")
                     IF (endtag .EQV. .FALSE.) READ (DATA(1), *) scptb_control%sd3(3)
                  CASE DEFAULT
                     unrecognized_tags = unrecognized_tags+1
                  END SELECT
               END DO

            CASE ("ELEMENT") !now find the element we are looking for
               tag_count = tag_count+1
               tag_over = .FALSE.

            CASE DEFAULT
               unrecognized_tags = unrecognized_tags+1
            END SELECT

         END DO

         CALL xml_close(xml_info) !That's it, close the file.

      END IF

   END SUBROUTINE scptb_global_parameter_from_file

! **************************************************************************************************
!> \brief ...
!> \param scptb_parameter ...
!> \param scptb_control ...
! **************************************************************************************************
   SUBROUTINE scptb_parameter_from_file(scptb_parameter, scptb_control)

      TYPE(scptb_parameter_type), INTENT(INOUT)          :: scptb_parameter
      TYPE(scptb_control_type), INTENT(INOUT)            :: scptb_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scptb_parameter_from_file', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: attribs(2, 10), author, DATA(10), fname, &
                                                            found_key, search_key, tag, vers
      INTEGER                                            :: j, l, n, no_attribs, no_data, tag_count, &
                                                            unrecognized_tags
      LOGICAL :: endtag, has_A, has_ag, has_B, has_C, has_D, has_energy, has_eta, has_hcore, &
         has_nqm, has_occ, has_z, has_zeff, tag_over, tag_over2
      TYPE(XML_PARSE)                                    :: xml_info

      DATA has_z, has_zeff, has_ag, has_energy, has_hcore, has_occ, &
         has_nqm, has_eta, has_A, has_B, has_C, has_D/12*.FALSE./

      fname = scptb_control%parameter_file
      IF (fname /= "") THEN
         CALL xml_open(xml_info, fname, .TRUE.)

         tag_count = 0
         tag = ""
         DATA = ""
         attribs = ""
         unrecognized_tags = 0
         tag_over = .FALSE.
         found_key = ""
         author = "unknown author"
         vers = " v 0.0"

         search_key = TRIM(scptb_parameter%key)
         CALL uppercase(search_key)

         DO WHILE (xml_ok(xml_info)) !Until the file is read entirely
            CALL xml_get(xml_info, tag, endtag, attribs, no_attribs, DATA, no_data)
            CALL uppercase(tag)
            SELECT CASE (TRIM (tag)) !Decide on the first-level tag
            CASE ("")
               !just garbage from the parser
            CASE ("NM12_PARAMETERS")
               !just the header, do nothing
            CASE ("META_INFO")
               tag_over = .FALSE.
               DO WHILE (tag_over .EQV. .FALSE.)
                  CALL xml_get(xml_info, tag, endtag, attribs, no_attribs, DATA, no_data)
                  CALL uppercase(tag)
                  SELECT CASE (tag)
                  CASE ("META_INFO")
                     IF (endtag .EQV. .TRUE.) tag_over = .TRUE.
                  CASE ("VERSION")
                     IF (endtag .EQV. .FALSE.) vers = DATA(1)
                  CASE ("AUTHOR")
                     IF (endtag .EQV. .FALSE.) author = DATA(1)
                  CASE ("REF")
                     !IF (endtag .eqv. .FALSE.)  !do nothing
                  CASE ("COMMENT")
                     !IF (endtag .eqv. .FALSE.)  !do nothing
                  CASE DEFAULT
                     unrecognized_tags = unrecognized_tags+1
                  END SELECT
               END DO

               WRITE (scptb_parameter%parameterization, '(A,A,A)') TRIM(author), ", ", TRIM(vers)

            CASE ("GENERAL_PARAMETERS") !general params
               tag_over = .FALSE.
               DO WHILE (tag_over .EQV. .FALSE.)
                  CALL xml_get(xml_info, tag, endtag, attribs, no_attribs, DATA, no_data)
                  CALL uppercase(tag)
                  SELECT CASE (tag)
                  CASE ("GENERAL_PARAMETERS")
                     IF (endtag .EQV. .TRUE.) tag_over = .TRUE.
                  CASE ("D3_S6")
                     ! do nothing in this sweep of data
                  CASE ("D3_SR6")
                     ! do nothing in this sweep of data
                  CASE ("D3_S8")
                     ! do nothing in this sweep of data
                  CASE DEFAULT
                     unrecognized_tags = unrecognized_tags+1
                  END SELECT
               END DO

            CASE ("ELEMENT") !now find the element we are looking for
               tag_count = tag_count+1
               tag_over = .FALSE.
               found_key = TRIM(attribs(2, 1))
               CALL uppercase(found_key)

               IF (found_key == search_key) THEN
                  ! read all the values

                  scptb_parameter%atomname = search_key

                  DO WHILE (tag_over .EQV. .FALSE.)
                     CALL xml_get(xml_info, tag, endtag, attribs, no_attribs, DATA, no_data)
                     CALL uppercase(tag)
                     SELECT CASE (tag)
                     CASE ("ELEMENT")
                        IF (endtag .EQV. .TRUE.) tag_over = .TRUE.
                     CASE ("Z")
                        IF (endtag .EQV. .FALSE.) READ (DATA(1), '(I4)') scptb_parameter%z; has_z = .TRUE.
                     CASE ("ZEFF")
                        IF (endtag .EQV. .FALSE.) READ (DATA(1), *) scptb_parameter%zeff; has_zeff = .TRUE.
                     CASE ("CHARGE_WIDTH")
                        IF (endtag .EQV. .FALSE.) READ (DATA(1), *) scptb_parameter%ag; has_ag = .TRUE.
                     CASE ("ENERGY")
                        IF (endtag .EQV. .FALSE.) READ (DATA(1), *) scptb_parameter%energy; has_energy = .TRUE.

                     CASE ("ELECTRON_SHELL")
                        !first deal with the attributes
                        DO j = 1, no_attribs
                           IF (attribs(1, j) == "n") READ (attribs(2, j), *) n
                           IF (attribs(1, j) == "l") READ (attribs(2, j), *) l
                        END DO

                        !now deal with the sub tags
                        tag_over2 = .FALSE.
                        DO WHILE (tag_over2 .EQV. .FALSE.)
                           CALL xml_get(xml_info, tag, endtag, attribs, no_attribs, DATA, no_data)
                           CALL uppercase(tag)
                           SELECT CASE (tag)
                           CASE ("ELECTRON_SHELL")
                              IF (endtag .EQV. .TRUE.) tag_over2 = .TRUE.
                           CASE ("ZETA")
                              !deal with the values
                              IF (endtag .EQV. .FALSE.) THEN
                                 READ (DATA(1), *) scptb_parameter%zeta(n, l)
                                 scptb_parameter%natorb = scptb_parameter%natorb+1
                                 scptb_parameter%norb(l) = scptb_parameter%norb(l)+1
                                 IF (l > scptb_parameter%lmaxorb) scptb_parameter%lmaxorb = l
                              END IF
                           CASE ("H_CORE")
                              IF (endtag .EQV. .FALSE.) THEN
                                 READ (DATA(1), *) scptb_parameter%hcore(n, l)
                                 has_hcore = .TRUE.
                              END IF
                           CASE ("OCC")
                              IF (endtag .EQV. .FALSE.) &
                                 READ (DATA(1), *) scptb_parameter%occupation(n, l); has_occ = .TRUE.
                           CASE ("N")
                              IF (endtag .EQV. .FALSE.) &
                                 READ (DATA(1), *) scptb_parameter%nqm(n, l); has_nqm = .TRUE.
                           CASE DEFAULT
                              IF (endtag .EQV. .FALSE.) unrecognized_tags = unrecognized_tags+1
                           END SELECT
                        END DO

                     CASE ("SCP")
                        !deal with the sub tags
                        tag_over2 = .FALSE.
                        DO WHILE (tag_over2 .EQV. .FALSE.)
                           CALL xml_get(xml_info, tag, endtag, attribs, no_attribs, DATA, no_data)
                           CALL uppercase(tag)
                           SELECT CASE (tag)
                           CASE ("SCP")
                              IF (endtag .EQV. .TRUE.) tag_over2 = .TRUE.
                           CASE ("ETA")
                              !first deal with the attributes
                              DO j = 1, no_attribs
                                 IF (attribs(1, j) == "l") READ (attribs(2, j), *) l; has_eta = .TRUE.
                              END DO
                              IF (endtag .EQV. .FALSE.) THEN
                                 IF (l > 0) READ (DATA(1), *) scptb_parameter%pol(l)
                                 IF (l > scptb_parameter%lmaxscp) scptb_parameter%lmaxscp = l
                              END IF
                           CASE DEFAULT
                              IF (endtag .EQV. .FALSE.) unrecognized_tags = unrecognized_tags+1
                           END SELECT
                        END DO

                     CASE ("REPULSIVE")
                        !deal with the sub tags
                        tag_over2 = .FALSE.
                        DO WHILE (tag_over2 .EQV. .FALSE.)
                           CALL xml_get(xml_info, tag, endtag, attribs, no_attribs, DATA, no_data)
                           CALL uppercase(tag)
                           SELECT CASE (tag)
                           CASE ("REPULSIVE")
                              IF (endtag .EQV. .TRUE.) tag_over2 = .TRUE.
                           CASE ("A")
                              IF (endtag .EQV. .FALSE.) READ (DATA(1), *) scptb_parameter%crep(1); has_A = .TRUE.
                           CASE ("C")
                              IF (endtag .EQV. .FALSE.) READ (DATA(1), *) scptb_parameter%crep(2); has_C = .TRUE.
                           CASE ("D")
                              IF (endtag .EQV. .FALSE.) READ (DATA(1), *) scptb_parameter%crep(3); has_D = .TRUE.
                           CASE DEFAULT
                              IF (endtag .EQV. .FALSE.) unrecognized_tags = unrecognized_tags+1
                           END SELECT
                        END DO
                     CASE DEFAULT
                        IF (endtag .EQV. .FALSE.) unrecognized_tags = unrecognized_tags+1
                     END SELECT
                  END DO
               ELSE
                  !when the current element is not the one we are looking for,
                  !just keep running through the file until the next element.
                  DO WHILE (tag_over .EQV. .FALSE.)
                     CALL xml_get(xml_info, tag, endtag, attribs, no_attribs, DATA, no_data)
                     CALL uppercase(tag)
                     IF (tag == "ELEMENT" .AND. endtag .EQV. .TRUE.) tag_over = .TRUE.
                  END DO
               END IF

            CASE DEFAULT
               unrecognized_tags = unrecognized_tags+1
            END SELECT

         END DO

         CALL xml_close(xml_info) !That's it, close the file.

         !An element is only considered properly defined if all non-optional parameters were
         !specified. That doesn't safeguard against incomplete or meaningless input.
         IF (has_z .AND. has_zeff .AND. has_ag .AND. has_hcore .AND. &
             has_nqm .AND. has_A .AND. has_C .AND. has_D) THEN
            scptb_parameter%defined = .TRUE.
         ELSE
            scptb_parameter%defined = .FALSE.
         END IF

      END IF

   END SUBROUTINE scptb_parameter_from_file

! **************************************************************************************************
!> \brief ...
!> \param scptb_control ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE scptb_global_parameter_broadcast(scptb_control, para_env)

      TYPE(scptb_control_type), INTENT(INOUT)            :: scptb_control
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scptb_global_parameter_broadcast', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: group, source

      source = para_env%source
      group = para_env%group

      IF (para_env%ionode) THEN
         IF(.NOT.(source == para_env%mepos))CALL cp__a("scptb_utils.F",569)
      ELSE
         IF(.NOT.(source /= para_env%mepos))CALL cp__a("scptb_utils.F",571)
      END IF

      CALL mp_bcast(scptb_control%sd3, source, group)

   END SUBROUTINE scptb_global_parameter_broadcast

! **************************************************************************************************
!> \brief ...
!> \param scptb_parameter ...
!> \param para_env ...
! **************************************************************************************************
   SUBROUTINE scptb_parameter_broadcast(scptb_parameter, para_env)

      TYPE(scptb_parameter_type), INTENT(INOUT)          :: scptb_parameter
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scptb_parameter_broadcast', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: group, source

      source = para_env%source
      group = para_env%group

      IF (para_env%ionode) THEN
         IF(.NOT.(source == para_env%mepos))CALL cp__a("scptb_utils.F",597)
      ELSE
         IF(.NOT.(source /= para_env%mepos))CALL cp__a("scptb_utils.F",599)
      END IF

      CALL mp_bcast(scptb_parameter%key, source, group)
      CALL mp_bcast(scptb_parameter%parameterization, source, group)
      CALL mp_bcast(scptb_parameter%atomname, source, group)
      CALL mp_bcast(scptb_parameter%defined, source, group)
      CALL mp_bcast(scptb_parameter%z, source, group)
      CALL mp_bcast(scptb_parameter%zeff, source, group)
      CALL mp_bcast(scptb_parameter%natorb, source, group)
      CALL mp_bcast(scptb_parameter%lmaxorb, source, group)
      CALL mp_bcast(scptb_parameter%lmaxscp, source, group)
      CALL mp_bcast(scptb_parameter%norb, source, group)
      CALL mp_bcast(scptb_parameter%nqm, source, group)
      CALL mp_bcast(scptb_parameter%zeta, source, group)
      CALL mp_bcast(scptb_parameter%hcore, source, group)
      CALL mp_bcast(scptb_parameter%occupation, source, group)
      CALL mp_bcast(scptb_parameter%energy, source, group)
      CALL mp_bcast(scptb_parameter%crep, source, group)
      CALL mp_bcast(scptb_parameter%pol, source, group)
      CALL mp_bcast(scptb_parameter%ag, source, group)
      CALL mp_bcast(scptb_parameter%rcpair, source, group)

   END SUBROUTINE scptb_parameter_broadcast

END MODULE scptb_utils

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Writes information on XC functionals to output
! **************************************************************************************************
MODULE xc_write_output

   USE input_constants,                 ONLY: xc_none
   USE input_cp2k_check,                ONLY: xc_functionals_expand
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_get_subs_vals2,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length
   USE xc_derivatives,                  ONLY: xc_functional_get_info
   USE xc_libxc,                        ONLY: libxc_version_info
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'xc_write_output'

   PUBLIC :: xc_write

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param output_unit ...
!> \param xc_section ...
!> \param lsd ...
! **************************************************************************************************
   SUBROUTINE xc_write(output_unit, xc_section, lsd)
      INTEGER, INTENT(IN)                                :: output_unit
      TYPE(section_vals_type), POINTER                   :: xc_section
      LOGICAL, INTENT(IN)                                :: lsd

      CHARACTER(len=*), PARAMETER :: routineN = 'xc_write', routineP = moduleN//':'//routineN

      CHARACTER(LEN=10*default_string_length)            :: reference
      CHARACTER(LEN=2*default_string_length)             :: shortform
      CHARACTER(LEN=20)                                  :: tmpStr
      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: func_name
      INTEGER                                            :: ifun, ifunc_name, il, myfun
      TYPE(section_vals_type), POINTER                   :: xc_fun, xc_fun_section

      IF (output_unit > 0) THEN

         xc_fun_section => section_vals_get_subs_vals(xc_section, &
                                                      "XC_FUNCTIONAL")
         CALL section_vals_val_get(xc_fun_section, "_SECTION_PARAMETERS_", i_val=myfun)
         IF (myfun /= xc_none) THEN
            CALL section_vals_val_get(xc_section, "FUNCTIONAL_ROUTINE", &
                                      c_val=tmpStr)
            WRITE (output_unit, fmt="(' FUNCTIONAL| ROUTINE=',a)") &
               TRIM(tmpStr)
            CALL xc_functionals_expand(xc_fun_section, xc_section)
            ifun = 0
            DO
               ifun = ifun+1
               xc_fun => section_vals_get_subs_vals2(xc_fun_section, i_section=ifun)
               IF (.NOT. ASSOCIATED(xc_fun)) EXIT
               CALL xc_functional_get_info(xc_fun, &
                                           lsd=lsd, &
                                           reference=reference, &
                                           shortform=shortform, &
                                           ifunc_name=1)
               IF (TRIM(xc_fun%section%name) == "LIBXC") THEN
                  CALL libxc_version_info(tmpStr)
                  WRITE (output_unit, fmt="(A,A,A)") ' FUNCTIONAL| LIBXC Vers. ', TRIM(tmpStr(1:5)), &
                     ' (Marques, Oliveira, Burnus, CPC 183, 2272 (2012))'
                  WRITE (output_unit, fmt="(' FUNCTIONAL| ',a,':')") TRIM(shortform)
                  DO il = 1, LEN_TRIM(reference), 67
                     WRITE (output_unit, fmt="(' FUNCTIONAL| ',a67)") reference(il:)
                  END DO
                  CALL section_vals_val_get(xc_fun, "functional", c_vals=func_name)
                  DO ifunc_name = 2, SIZE(func_name)
                     CALL xc_functional_get_info(xc_fun, &
                                                 lsd=lsd, &
                                                 reference=reference, &
                                                 shortform=shortform, &
                                                 ifunc_name=ifunc_name)
                     WRITE (output_unit, fmt="(A,A,A)") ' FUNCTIONAL| LIBXC Vers. ', TRIM(tmpStr(1:5)), &
                        ' (Marques, Oliveira, Burnus, CPC 183, 2272 (2012))'
                     WRITE (output_unit, fmt="(' FUNCTIONAL| ',a,':')") TRIM(shortform)
                     DO il = 1, LEN_TRIM(reference), 67
                        WRITE (output_unit, fmt="(' FUNCTIONAL| ',a67)") reference(il:)
                     END DO
                  END DO
               ELSE
                  WRITE (output_unit, fmt="(' FUNCTIONAL| ',a,':')") &
                     TRIM(xc_fun%section%name)
                  DO il = 1, LEN_TRIM(reference), 67
                     WRITE (output_unit, fmt="(' FUNCTIONAL| ',a67)") reference(il:)
                  END DO
               END IF
            END DO
         ELSE
            WRITE (output_unit, fmt="(' FUNCTIONAL| NO EXCHANGE-CORRELATION FUNCTIONAL USED.')")
         END IF
      END IF

   END SUBROUTINE xc_write

END MODULE xc_write_output
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utilities to set up the control types
! **************************************************************************************************
MODULE cp_control_utils
   USE bibliography,                    ONLY: &
        Andreussi2012, Dewar1977, Dewar1985, Elstner1998, Fattebert2002, Hu2007, Krack2000, &
        Lippert1997, Lippert1999, Porezag1995, Repasky2002, Rocha2006, Schenter2008, Seifert1996, &
        Souza2002, Stengel2009, Stewart1989, Stewart2007, Thiel1992, Umari2002, VandeVondele2005a, &
        VandeVondele2005b, Zhechkov2005, cite_reference
   USE cp_control_types,                ONLY: &
        admm_control_create, admm_control_type, ddapc_control_create, ddapc_restraint_type, &
        dft_control_create, dft_control_type, efield_type, qs_control_type, sccs_control_create, &
        tddfpt_control_create, tddfpt_control_type
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_units,                        ONLY: cp_unit_from_cp2k,&
                                              cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        constant_env, custom_env, do_admm_basis_projection, do_admm_blocked_projection, &
        do_admm_blocking_purify_full, do_admm_charge_constrained_projection, &
        do_admm_exch_scaling_merlot, do_admm_purify_mcweeny, do_admm_purify_mo_diag, &
        do_admm_purify_mo_no_diag, do_admm_purify_none, do_admm_purify_none_dm, &
        do_ddapc_constraint, do_ddapc_restraint, do_method_am1, do_method_dftb, do_method_gapw, &
        do_method_gapw_xc, do_method_gpw, do_method_lrigpw, do_method_mndo, do_method_mndod, &
        do_method_ofgpw, do_method_pdg, do_method_pm3, do_method_pm6, do_method_pnnl, &
        do_method_rm1, do_method_scptb, do_pwgrid_ns_fullspace, do_pwgrid_ns_halfspace, &
        do_pwgrid_spherical, do_s2_constraint, do_s2_restraint, do_se_is_kdso, do_se_is_kdso_d, &
        do_se_is_slater, do_se_lr_ewald, do_se_lr_ewald_gks, do_se_lr_ewald_r3, do_se_lr_none, &
        gaussian_env, numerical, ramp_env, real_time_propagation, sccs_andreussi, &
        sccs_derivative_cd3, sccs_derivative_cd5, sccs_derivative_cd7, sccs_derivative_fft, &
        sccs_fattebert_gygi, sic_ad, sic_eo, sic_list_all, sic_list_unpaired, sic_mauri_spz, &
        sic_mauri_us, sic_none, slater, tddfpt_excitations
   USE input_cp2k_dft,                  ONLY: create_dft_section,&
                                              create_qs_section
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: &
        section_get_ival, section_get_keyword, section_release, section_type, section_vals_get, &
        section_vals_get_subs_vals, section_vals_type, section_vals_val_get, section_vals_val_set
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE xc_derivatives,                  ONLY: xc_functionals_get_needs
   USE xc_input_constants,              ONLY: xc_deriv_collocate
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_write_output,                 ONLY: xc_write
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_control_utils'

   PUBLIC :: read_dft_control, &
             read_mgrid_section, &
             read_qs_section, &
             read_tddfpt_control, &
             write_dft_control, &
             write_qs_control, &
             read_ddapc_section, &
             read_becke_section

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param dft_control ...
!> \param dft_section ...
! **************************************************************************************************
   SUBROUTINE read_dft_control(dft_control, dft_section)
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_dft_control', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: basis_set_file_name, potential_file_name
      INTEGER                                            :: excitations, i, method_id, nrep, stat, &
                                                            xc_deriv_method_id
      LOGICAL                                            :: do_ot, do_rtp, is_present, l_param, &
                                                            not_SE, was_present
      REAL(KIND=dp)                                      :: density_cut, gradient_cut, tau_cut
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pol
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: sccs_section, scf_section, tmp_section, &
                                                            xc_fun_section, xc_section
      TYPE(xc_rho_cflags_type)                           :: needs

      was_present = .FALSE.

      logger => cp_get_default_logger()

      NULLIFY (tmp_section, xc_fun_section, xc_section)
      CALL dft_control_create(dft_control)
      ! determine wheather this is a semiempirical or DFTB run
      ! --> (no XC section needs to be provided)
      not_SE = .TRUE.
      CALL section_vals_val_get(dft_section, "QS%METHOD", i_val=method_id)
      SELECT CASE (method_id)
      CASE (do_method_dftb, do_method_scptb, do_method_mndo, do_method_am1, do_method_pm3, do_method_pnnl, &
            do_method_pm6, do_method_pdg, do_method_rm1, do_method_mndod)
         not_SE = .FALSE.
      END SELECT
      ! Check for XC section and XC_FUNCTIONAL section
      xc_section => section_vals_get_subs_vals(dft_section, "XC")
      CALL section_vals_get(xc_section, explicit=is_present)
      IF (.NOT. is_present .AND. not_SE) THEN
         CALL cp__b("cp_control_utils.F",120,"XC section missing.")
      END IF
      IF (is_present) THEN
         CALL section_vals_val_get(xc_section, "density_cutoff", r_val=density_cut)
         CALL section_vals_val_get(xc_section, "gradient_cutoff", r_val=gradient_cut)
         CALL section_vals_val_get(xc_section, "tau_cutoff", r_val=tau_cut)
         ! Perform numerical stability checks and possibly correct the issues
         IF (density_cut <= EPSILON(0.0_dp)*100.0_dp) &
            CALL cp_warn(cp__l("cp_control_utils.F",128), &
                         "DENSITY_CUTOFF lower than 100*EPSILON, where EPSILON is the machine precision. "// &
                         "This may lead to numerical problems. Setting up shake_tol to 100*EPSILON! ")
         density_cut = MAX(EPSILON(0.0_dp)*100.0_dp, density_cut)
         IF (gradient_cut <= EPSILON(0.0_dp)*100.0_dp) &
            CALL cp_warn(cp__l("cp_control_utils.F",133), &
                         "GRADIENT_CUTOFF lower than 100*EPSILON, where EPSILON is the machine precision. "// &
                         "This may lead to numerical problems. Setting up shake_tol to 100*EPSILON! ")
         gradient_cut = MAX(EPSILON(0.0_dp)*100.0_dp, gradient_cut)
         IF (tau_cut <= EPSILON(0.0_dp)*100.0_dp) &
            CALL cp_warn(cp__l("cp_control_utils.F",138), &
                         "TAU_CUTOFF lower than 100*EPSILON, where EPSILON is the machine precision. "// &
                         "This may lead to numerical problems. Setting up shake_tol to 100*EPSILON! ")
         tau_cut = MAX(EPSILON(0.0_dp)*100.0_dp, tau_cut)
         CALL section_vals_val_set(xc_section, "density_cutoff", r_val=density_cut)
         CALL section_vals_val_set(xc_section, "gradient_cutoff", r_val=gradient_cut)
         CALL section_vals_val_set(xc_section, "tau_cutoff", r_val=tau_cut)
      END IF
      xc_fun_section => section_vals_get_subs_vals(xc_section, "XC_FUNCTIONAL")
      CALL section_vals_get(xc_fun_section, explicit=is_present)
      IF (.NOT. is_present .AND. not_SE) THEN
         CALL cp__b("cp_control_utils.F",149,"XC_FUNCTIONAL section missing.")
      END IF
      scf_section => section_vals_get_subs_vals(dft_section, "SCF")
      CALL section_vals_val_get(dft_section, "UKS", l_val=dft_control%uks)
      CALL section_vals_val_get(dft_section, "ROKS", l_val=dft_control%roks)
      IF (dft_control%uks .OR. dft_control%roks) THEN
         dft_control%nspins = 2
      ELSE
         dft_control%nspins = 1
      END IF

      dft_control%lsd = (dft_control%nspins > 1)
      needs = xc_functionals_get_needs(xc_fun_section, &
                                       lsd=dft_control%lsd, &
                                       add_basic_components=.TRUE.)
      dft_control%use_kinetic_energy_density = (needs%tau_spin .OR. needs%tau)

      xc_deriv_method_id = section_get_ival(xc_section, "XC_GRID%XC_DERIV")
      dft_control%drho_by_collocation = (needs%norm_drho .AND. (xc_deriv_method_id == xc_deriv_collocate))
      IF (dft_control%drho_by_collocation) THEN
         CALL cp__b("cp_control_utils.F",169,"derivatives by collocation not implemented")
      END IF

      !! check if we do wavefunction fitting
      tmp_section => section_vals_get_subs_vals(dft_section, "AUXILIARY_DENSITY_MATRIX_METHOD")
      CALL section_vals_get(tmp_section, explicit=is_present)
      dft_control%do_admm = is_present
      dft_control%do_admm_mo = .FALSE.
      dft_control%do_admm_dm = .FALSE.
      IF (is_present) THEN
         do_ot = .FALSE.
         CALL section_vals_val_get(scf_section, "OT%_SECTION_PARAMETERS_", l_val=do_ot)
         CALL admm_control_create(dft_control%admm_control)

         CALL section_vals_val_get(dft_section, "AUXILIARY_DENSITY_MATRIX_METHOD%EPS_FILTER", &
                                   r_val=dft_control%admm_control%eps_filter)
         CALL section_vals_val_get(dft_section, "AUXILIARY_DENSITY_MATRIX_METHOD%ADMM_PURIFICATION_METHOD", i_val=method_id)
         dft_control%admm_control%purification_method = method_id

         CALL section_vals_val_get(dft_section, "AUXILIARY_DENSITY_MATRIX_METHOD%METHOD", i_val=method_id)
         dft_control%admm_control%method = method_id

         CALL section_vals_val_get(dft_section, "AUXILIARY_DENSITY_MATRIX_METHOD%EXCH_SCALING_MODEL", i_val=method_id)
         dft_control%admm_control%scaling_model = method_id

         CALL section_vals_val_get(dft_section, "AUXILIARY_DENSITY_MATRIX_METHOD%EXCH_CORRECTION_FUNC", i_val=method_id)
         dft_control%admm_control%aux_exch_func = method_id

         CALL read_admm_block_list(dft_control%admm_control, dft_section)

         !    In the case of charge-constrained projection (e.g. according to Merlot),
         !    there is no purification needed and hence, do_admm_purify_none has to be set.

         IF ((dft_control%admm_control%method == do_admm_blocking_purify_full .OR. &
              dft_control%admm_control%method == do_admm_blocked_projection) &
             .AND. dft_control%admm_control%scaling_model == do_admm_exch_scaling_merlot) THEN
            CALL cp__b("cp_control_utils.F",205,"ADMM: Blocking and Merlot scaling are mutually exclusive.")
         END IF

         IF (dft_control%admm_control%method == do_admm_charge_constrained_projection .AND. &
             dft_control%admm_control%purification_method /= do_admm_purify_none) THEN
            CALL cp_abort(cp__l("cp_control_utils.F",210), &
                          "ADMM: In the case of METHOD=CHARGE_CONSTRAINED_PROJECTION, "// &
                          "ADMM_PURIFICATION_METHOD=NONE has to be set.")
         END IF

         IF (dft_control%admm_control%purification_method == do_admm_purify_mo_diag .OR. &
             dft_control%admm_control%purification_method == do_admm_purify_mo_no_diag) THEN
            IF (dft_control%admm_control%method /= do_admm_basis_projection) &
               CALL cp__b("cp_control_utils.F",218,"ADMM: Chosen purification requires BASIS_PROJECTION")

            IF (.NOT. do_ot) CALL cp__b("cp_control_utils.F",220,"ADMM: MO-based purification requires OT.")
         END IF

         IF (dft_control%admm_control%purification_method == do_admm_purify_none_dm .OR. &
             dft_control%admm_control%purification_method == do_admm_purify_mcweeny) THEN
            dft_control%do_admm_dm = .TRUE.
         ELSE
            dft_control%do_admm_mo = .TRUE.
         ENDIF
      END IF

      ! Set restricted to true, if both OT and ROKS are requested
      !MK in principle dft_control%restricted could be dropped completely like the
      !MK input key by using only dft_control%roks now
      CALL section_vals_val_get(scf_section, "OT%_SECTION_PARAMETERS_", l_val=l_param)
      dft_control%restricted = (dft_control%roks .AND. l_param)

      CALL section_vals_val_get(dft_section, "CHARGE", i_val=dft_control%charge)
      CALL section_vals_val_get(dft_section, "MULTIPLICITY", i_val=dft_control%multiplicity)
      CALL section_vals_val_get(dft_section, "RELAX_MULTIPLICITY", r_val=dft_control%relax_multiplicity)
      IF (dft_control%relax_multiplicity > 0.0_dp) THEN
         IF (.NOT. dft_control%uks) &
            CALL cp_abort(cp__l("cp_control_utils.F",242), "The option RELAX_MULTIPLICITY is only valid for "// &
                          "unrestricted Kohn-Sham (UKS) calculations")
      END IF

      ! check for the presence of the low spin roks section
      tmp_section => section_vals_get_subs_vals(dft_section, "LOW_SPIN_ROKS")
      CALL section_vals_get(tmp_section, explicit=dft_control%low_spin_roks)

      dft_control%sic_method_id = sic_none
      dft_control%sic_scaling_a = 1.0_dp
      dft_control%sic_scaling_b = 1.0_dp

      ! DFT+U
      dft_control%dft_plus_u = .FALSE.
      CALL section_vals_val_get(dft_section, "PLUS_U_METHOD", i_val=method_id)
      dft_control%plus_u_method_id = method_id

      ! Smearing in use
      dft_control%smear = .FALSE.

      ! Surface dipole correction
      dft_control%correct_surf_dip = .FALSE.
      CALL section_vals_val_get(dft_section, "SURFACE_DIPOLE_CORRECTION", l_val=dft_control%correct_surf_dip)
      CALL section_vals_val_get(dft_section, "SURF_DIP_DIR", i_val=dft_control%dir_surf_dip)

      CALL section_vals_val_get(dft_section, "BASIS_SET_FILE_NAME", &
                                c_val=basis_set_file_name)
      CALL section_vals_val_get(dft_section, "POTENTIAL_FILE_NAME", &
                                c_val=potential_file_name)

      ! Read the input section
      tmp_section => section_vals_get_subs_vals(dft_section, "sic")
      CALL section_vals_val_get(tmp_section, "SIC_METHOD", &
                                i_val=dft_control%sic_method_id)
      CALL section_vals_val_get(tmp_section, "ORBITAL_SET", &
                                i_val=dft_control%sic_list_id)
      CALL section_vals_val_get(tmp_section, "SIC_SCALING_A", &
                                r_val=dft_control%sic_scaling_a)
      CALL section_vals_val_get(tmp_section, "SIC_SCALING_B", &
                                r_val=dft_control%sic_scaling_b)

      ! Determine if this is a TDDFPT run
      CALL section_vals_val_get(dft_section, "EXCITATIONS", i_val=excitations)
      dft_control%do_tddfpt_calculation = (excitations == tddfpt_excitations)
      IF (dft_control%do_tddfpt_calculation) THEN
         CALL tddfpt_control_create(dft_control%tddfpt_control)
      END IF

      do_rtp = .FALSE.
      tmp_section => section_vals_get_subs_vals(dft_section, "REAL_TIME_PROPAGATION")
      CALL section_vals_get(tmp_section, explicit=is_present)
      IF (is_present) THEN
         CALL read_rtp_section(dft_control, tmp_section)
         do_rtp = .TRUE.
      END IF

      ! Read the input section
      tmp_section => section_vals_get_subs_vals(dft_section, "XAS")
      CALL section_vals_get(tmp_section, explicit=dft_control%do_xas_calculation)
      IF (dft_control%do_xas_calculation) THEN
         ! Override with section parameter
         CALL section_vals_val_get(tmp_section, "_SECTION_PARAMETERS_", &
                                   l_val=dft_control%do_xas_calculation)
      END IF

      ! Read the finite field input section
      dft_control%apply_efield = .FALSE.
      dft_control%apply_efield_field = .FALSE. !this is for RTP
      tmp_section => section_vals_get_subs_vals(dft_section, "EFIELD")
      CALL section_vals_get(tmp_section, n_repetition=nrep, explicit=is_present)
      IF (is_present) THEN
         ALLOCATE (dft_control%efield_fields(nrep))
         DO i = 1, nrep
            CALL read_efield_sections(dft_control, tmp_section)
         END DO
         IF (do_rtp) THEN
            dft_control%apply_efield_field = .TRUE.
         ELSE
            dft_control%apply_efield = .TRUE.
            IF(.NOT.(nrep == 1))CALL cp__a("cp_control_utils.F",321)
         END IF
      END IF

      ! Read the finite field input section for periodic fields
      tmp_section => section_vals_get_subs_vals(dft_section, "PERIODIC_EFIELD")
      CALL section_vals_get(tmp_section, explicit=dft_control%apply_period_efield)
      IF (dft_control%apply_period_efield) THEN
         ALLOCATE (dft_control%period_efield, stat=stat)
         ALLOCATE (dft_control%period_efield%polarisation(3), stat=stat)
         CALL section_vals_val_get(tmp_section, "POLARISATION", &
                                   r_vals=pol)
         dft_control%period_efield%polarisation = pol
         CALL section_vals_val_get(tmp_section, "INTENSITY", &
                                   r_val=dft_control%period_efield%strength)
         dft_control%period_efield%displacement_field = .FALSE.
         CALL section_vals_val_get(tmp_section, "DISPLACEMENT_FIELD", &
                                   l_val=dft_control%period_efield%displacement_field)
         ! periodic fields don't work with RTP
         IF(.NOT.(.NOT. do_rtp))CALL cp__a("cp_control_utils.F",340)
         IF (dft_control%period_efield%displacement_field) THEN
            CALL cite_reference(Stengel2009)
         ELSE
            CALL cite_reference(Souza2002)
            CALL cite_reference(Umari2002)
         END IF
      END IF

      ! Read the external potential input section
      tmp_section => section_vals_get_subs_vals(dft_section, "EXTERNAL_POTENTIAL")
      CALL section_vals_get(tmp_section, explicit=dft_control%apply_external_potential)

      !MK Read the SCCS input section if present
      sccs_section => section_vals_get_subs_vals(dft_section, "SCCS")
      CALL section_vals_get(sccs_section, explicit=is_present)
      IF (is_present) THEN
         ! Check section parameter if SCCS is activated
         CALL section_vals_val_get(sccs_section, "_SECTION_PARAMETERS_", &
                                   l_val=dft_control%do_sccs)
         IF (dft_control%do_sccs) THEN
            CALL sccs_control_create(dft_control%sccs_control)
            CALL section_vals_val_get(sccs_section, "ALPHA", &
                                      r_val=dft_control%sccs_control%alpha_solvent)
            CALL section_vals_val_get(sccs_section, "BETA", &
                                      r_val=dft_control%sccs_control%beta_solvent)
            CALL section_vals_val_get(sccs_section, "DELTA_RHO", &
                                      r_val=dft_control%sccs_control%delta_rho)
            CALL section_vals_val_get(sccs_section, "DERIVATIVE_METHOD", &
                                      i_val=dft_control%sccs_control%derivative_method)
            CALL section_vals_val_get(sccs_section, "METHOD", &
                                      i_val=dft_control%sccs_control%method_id)
            CALL section_vals_val_get(sccs_section, "DIELECTRIC_CONSTANT", &
                                      r_val=dft_control%sccs_control%epsilon_solvent)
            CALL section_vals_val_get(sccs_section, "EPS_SCCS", &
                                      r_val=dft_control%sccs_control%eps_sccs)
            CALL section_vals_val_get(sccs_section, "EPS_SCF", &
                                      r_val=dft_control%sccs_control%eps_scf)
            CALL section_vals_val_get(sccs_section, "GAMMA", &
                                      r_val=dft_control%sccs_control%gamma_solvent)
            CALL section_vals_val_get(sccs_section, "MAX_ITER", &
                                      i_val=dft_control%sccs_control%max_iter)
            CALL section_vals_val_get(sccs_section, "MIXING", &
                                      r_val=dft_control%sccs_control%mixing)
            SELECT CASE (dft_control%sccs_control%method_id)
            CASE (sccs_andreussi)
               tmp_section => section_vals_get_subs_vals(sccs_section, "ANDREUSSI")
               CALL section_vals_val_get(tmp_section, "RHO_MAX", &
                                         r_val=dft_control%sccs_control%rho_max)
               CALL section_vals_val_get(tmp_section, "RHO_MIN", &
                                         r_val=dft_control%sccs_control%rho_min)
               IF (dft_control%sccs_control%rho_max < dft_control%sccs_control%rho_min) THEN
                  CALL cp_abort(cp__l("cp_control_utils.F",392), &
                                "The SCCS parameter RHO_MAX is smaller than RHO_MIN. "// &
                                "Please, check your input!")
               END IF
            CASE (sccs_fattebert_gygi)
               tmp_section => section_vals_get_subs_vals(sccs_section, "FATTEBERT-GYGI")
               CALL section_vals_val_get(tmp_section, "BETA", &
                                         r_val=dft_control%sccs_control%beta)
               IF (dft_control%sccs_control%beta < 0.5_dp) THEN
                  CALL cp_abort(cp__l("cp_control_utils.F",401), &
                                "A value smaller than 0.5 for the SCCS parameter beta "// &
                                "causes numerical problems. Please, check your input!")
               END IF
               CALL section_vals_val_get(tmp_section, "RHO_ZERO", &
                                         r_val=dft_control%sccs_control%rho_zero)
            CASE DEFAULT
               CALL cp__b("cp_control_utils.F",408,"Invalid SCCS model specified. Please, check your input!")
            END SELECT
            CALL cite_reference(Fattebert2002)
            CALL cite_reference(Andreussi2012)
         END IF
      END IF

      ! ZMP added input sections
      ! Read the external density input section
      tmp_section => section_vals_get_subs_vals(dft_section, "EXTERNAL_DENSITY")
      CALL section_vals_get(tmp_section, explicit=dft_control%apply_external_density)

      ! Read the external vxc input section
      tmp_section => section_vals_get_subs_vals(dft_section, "EXTERNAL_VXC")
      CALL section_vals_get(tmp_section, explicit=dft_control%apply_external_vxc)

   END SUBROUTINE read_dft_control

! **************************************************************************************************
!> \brief ...
!> \param qs_control ...
!> \param dft_section ...
! **************************************************************************************************
   SUBROUTINE read_mgrid_section(qs_control, dft_section)

      TYPE(qs_control_type), INTENT(INOUT)               :: qs_control
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_mgrid_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, igrid_level, ngrid_level
      LOGICAL                                            :: explicit, multigrid_set
      REAL(dp)                                           :: cutoff
      REAL(dp), DIMENSION(:), POINTER                    :: cutofflist
      TYPE(section_vals_type), POINTER                   :: mgrid_section

      CALL timeset(routineN, handle)

      NULLIFY (mgrid_section, cutofflist)
      mgrid_section => section_vals_get_subs_vals(dft_section, "MGRID")

      CALL section_vals_val_get(mgrid_section, "NGRIDS", i_val=ngrid_level)
      CALL section_vals_val_get(mgrid_section, "MULTIGRID_SET", l_val=multigrid_set)
      CALL section_vals_val_get(mgrid_section, "CUTOFF", r_val=cutoff)
      CALL section_vals_val_get(mgrid_section, "PROGRESSION_FACTOR", r_val=qs_control%progression_factor)
      CALL section_vals_val_get(mgrid_section, "COMMENSURATE", l_val=qs_control%commensurate_mgrids)
      CALL section_vals_val_get(mgrid_section, "REALSPACE", l_val=qs_control%realspace_mgrids)
      CALL section_vals_val_get(mgrid_section, "REL_CUTOFF", r_val=qs_control%relative_cutoff)
      CALL section_vals_val_get(mgrid_section, "SKIP_LOAD_BALANCE_DISTRIBUTED", &
                                l_val=qs_control%skip_load_balance_distributed)

      ! For SE and DFTB possibly override with new defaults
      IF (qs_control%semi_empirical .OR. qs_control%dftb) THEN
         ngrid_level = 1
         multigrid_set = .FALSE.
         ! Override default cutoff value unless user specified an explicit argument..
         CALL section_vals_val_get(mgrid_section, "CUTOFF", explicit=explicit, r_val=cutoff)
         IF (.NOT. explicit) cutoff = 1.0_dp
      END IF
      IF (qs_control%scptb) THEN
         ngrid_level = 1
         multigrid_set = .FALSE.
         ! Override default cutoff value unless user specified an explicit argument..
         CALL section_vals_val_get(mgrid_section, "CUTOFF", explicit=explicit, r_val=cutoff)
         IF (.NOT. explicit) cutoff = 45.0_dp
      END IF

      ALLOCATE (qs_control%e_cutoff(ngrid_level))
      qs_control%cutoff = cutoff

      IF (multigrid_set) THEN
         ! Read the values from input
         IF (qs_control%commensurate_mgrids) THEN
            CALL cp__b("cp_control_utils.F",482,"Do not specify cutoffs for the commensurate grids (NYI)")
         END IF

         CALL section_vals_val_get(mgrid_section, "MULTIGRID_CUTOFF", r_vals=cutofflist)
         IF (ASSOCIATED(cutofflist)) THEN
            IF (SIZE(cutofflist, 1) /= ngrid_level) &
               CALL cp__b("cp_control_utils.F",488,"Inconsistent values for number of multi grids")
            DO igrid_level = 1, ngrid_level
               qs_control%e_cutoff(igrid_level) = cutofflist(igrid_level)*0.5_dp
            ENDDO
         END IF
         ! set cutoff to smallest value in multgrid available with >= cutoff
         DO igrid_level = ngrid_level, 1, -1
            IF (qs_control%cutoff <= qs_control%e_cutoff(igrid_level)) THEN
               qs_control%cutoff = qs_control%e_cutoff(igrid_level)
               EXIT
            END IF
            ! set largest grid value to cutoff
            IF (igrid_level == 1) THEN
               qs_control%cutoff = qs_control%e_cutoff(1)
            END IF
         END DO
      ELSE
         IF (qs_control%commensurate_mgrids) qs_control%progression_factor = 4.0_dp
         qs_control%e_cutoff(1) = qs_control%cutoff
         DO igrid_level = 2, ngrid_level
            qs_control%e_cutoff(igrid_level) = qs_control%e_cutoff(igrid_level-1) &
                                               /qs_control%progression_factor
         END DO
      END IF
      ! check that multigrids are ordered
      DO igrid_level = 2, ngrid_level
         IF (qs_control%e_cutoff(igrid_level) > &
             qs_control%e_cutoff(igrid_level-1)) THEN
            CALL cp__b("cp_control_utils.F",516,"Multi-grids not ordered")
         END IF
      END DO
      CALL timestop(handle)
   END SUBROUTINE read_mgrid_section

! **************************************************************************************************
!> \brief ...
!> \param qs_control ...
!> \param qs_section ...
! **************************************************************************************************
   SUBROUTINE read_qs_section(qs_control, qs_section)

      TYPE(qs_control_type), INTENT(INOUT)               :: qs_control
      TYPE(section_vals_type), POINTER                   :: qs_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_qs_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: clist
      INTEGER                                            :: handle, itmp, j, jj, k, n_rep, n_var, &
                                                            nrep
      INTEGER, DIMENSION(:), POINTER                     :: tmplist
      LOGICAL                                            :: explicit, was_present
      REAL(dp)                                           :: tmp, tmpsqrt, value
      REAL(dp), POINTER                                  :: scal(:)
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER :: becke_restraint_section, ddapc_restraint_section, &
         dftb_parameter, dftb_section, lri_optbas_section, mull_section, s2_restraint_section, &
         scptb_section, se_section

      CALL timeset(routineN, handle)

      was_present = .FALSE.
      NULLIFY (mull_section, se_section, dftb_section, scptb_section, lri_optbas_section)

      mull_section => section_vals_get_subs_vals(qs_section, "MULLIKEN_RESTRAINT")
      ddapc_restraint_section => section_vals_get_subs_vals(qs_section, "DDAPC_RESTRAINT")
      s2_restraint_section => section_vals_get_subs_vals(qs_section, "S2_RESTRAINT")
      becke_restraint_section => section_vals_get_subs_vals(qs_section, "BECKE_RESTRAINT")
      se_section => section_vals_get_subs_vals(qs_section, "SE")
      dftb_section => section_vals_get_subs_vals(qs_section, "DFTB")
      dftb_parameter => section_vals_get_subs_vals(dftb_section, "PARAMETER")
      scptb_section => section_vals_get_subs_vals(qs_section, "SCPTB")
      lri_optbas_section => section_vals_get_subs_vals(qs_section, "OPTIMIZE_LRI_BASIS")

      ! Setup all defaults values and overwrite input parameters
      ! EPS_DEFAULT should set the target accuracy in the total energy (~per electron) or a closely related value
      CALL section_vals_val_get(qs_section, "EPS_DEFAULT", r_val=value)
      tmpsqrt = SQRT(value) ! a trick to work around a NAG 5.1 optimizer bug

      ! random choice ?
      qs_control%eps_core_charge = value/100.0_dp
      ! correct if all Gaussians would have the same radius (overlap will be smaller than eps_pgf_orb**2).
      ! Can be significantly in error if not... requires fully new screening/pairlist procedures
      qs_control%eps_pgf_orb = tmpsqrt
      qs_control%eps_kg_orb = qs_control%eps_pgf_orb
      ! consistent since also a kind of overlap
      qs_control%eps_ppnl = qs_control%eps_pgf_orb/100.0_dp
      ! accuracy is basically set by the overlap, this sets an empirical shift
      qs_control%eps_ppl = 1.0E-2_dp
      !
      qs_control%gapw_control%eps_cpc = value
      ! expexted error in the density
      qs_control%eps_rho_gspace = value
      qs_control%eps_rho_rspace = value
      ! error in the gradient, can be the sqrt of the error in the energy, ignored if map_consistent
      qs_control%eps_gvg_rspace = tmpsqrt
      !
      CALL section_vals_val_get(qs_section, "EPS_CORE_CHARGE", n_rep_val=n_rep)
      IF (n_rep /= 0) THEN
         CALL section_vals_val_get(qs_section, "EPS_CORE_CHARGE", r_val=qs_control%eps_core_charge)
      END IF
      CALL section_vals_val_get(qs_section, "EPS_GVG_RSPACE", n_rep_val=n_rep)
      IF (n_rep /= 0) THEN
         CALL section_vals_val_get(qs_section, "EPS_GVG_RSPACE", r_val=qs_control%eps_gvg_rspace)
      END IF
      CALL section_vals_val_get(qs_section, "EPS_PGF_ORB", n_rep_val=n_rep)
      IF (n_rep /= 0) THEN
         CALL section_vals_val_get(qs_section, "EPS_PGF_ORB", r_val=qs_control%eps_pgf_orb)
      END IF
      CALL section_vals_val_get(qs_section, "EPS_KG_ORB", n_rep_val=n_rep)
      IF (n_rep /= 0) THEN
         CALL section_vals_val_get(qs_section, "EPS_KG_ORB", r_val=tmp)
         qs_control%eps_kg_orb = SQRT(tmp)
      END IF
      CALL section_vals_val_get(qs_section, "EPS_PPL", n_rep_val=n_rep)
      IF (n_rep /= 0) THEN
         CALL section_vals_val_get(qs_section, "EPS_PPL", r_val=qs_control%eps_ppl)
      END IF
      CALL section_vals_val_get(qs_section, "EPS_PPNL", n_rep_val=n_rep)
      IF (n_rep /= 0) THEN
         CALL section_vals_val_get(qs_section, "EPS_PPNL", r_val=qs_control%eps_ppnl)
      END IF
      CALL section_vals_val_get(qs_section, "EPS_RHO", n_rep_val=n_rep)
      IF (n_rep /= 0) THEN
         CALL section_vals_val_get(qs_section, "EPS_RHO", r_val=qs_control%eps_rho_gspace)
         qs_control%eps_rho_rspace = qs_control%eps_rho_gspace
      END IF
      CALL section_vals_val_get(qs_section, "EPS_RHO_RSPACE", n_rep_val=n_rep)
      IF (n_rep /= 0) THEN
         CALL section_vals_val_get(qs_section, "EPS_RHO_RSPACE", r_val=qs_control%eps_rho_rspace)
      END IF
      CALL section_vals_val_get(qs_section, "EPS_RHO_GSPACE", n_rep_val=n_rep)
      IF (n_rep /= 0) THEN
         CALL section_vals_val_get(qs_section, "EPS_RHO_GSPACE", r_val=qs_control%eps_rho_gspace)
      END IF
      CALL section_vals_val_get(qs_section, "EPS_FILTER_MATRIX", n_rep_val=n_rep)
      IF (n_rep /= 0) THEN
         CALL section_vals_val_get(qs_section, "EPS_FILTER_MATRIX", r_val=qs_control%eps_filter_matrix)
      END IF
      CALL section_vals_val_get(qs_section, "EPS_CPC", n_rep_val=n_rep)
      IF (n_rep /= 0) THEN
         CALL section_vals_val_get(qs_section, "EPS_CPC", r_val=qs_control%gapw_control%eps_cpc)
      END IF

      CALL section_vals_val_get(qs_section, "EPSFIT", r_val=qs_control%gapw_control%eps_fit)
      CALL section_vals_val_get(qs_section, "EPSISO", r_val=qs_control%gapw_control%eps_iso)
      CALL section_vals_val_get(qs_section, "EPSSVD", r_val=qs_control%gapw_control%eps_svd)
      CALL section_vals_val_get(qs_section, "EPSRHO0", r_val=qs_control%gapw_control%eps_Vrho0)
      CALL section_vals_val_get(qs_section, "ALPHA0_HARD", r_val=qs_control%gapw_control%alpha0_hard)
      qs_control%gapw_control%lrho1_eq_lrho0 = .FALSE.
      qs_control%gapw_control%alpha0_hard_from_input = .FALSE.
      IF (qs_control%gapw_control%alpha0_hard /= 0.0_dp) qs_control%gapw_control%alpha0_hard_from_input = .TRUE.
      CALL section_vals_val_get(qs_section, "FORCE_PAW", l_val=qs_control%gapw_control%force_paw)
      CALL section_vals_val_get(qs_section, "MAX_RAD_LOCAL", r_val=qs_control%gapw_control%max_rad_local)

      CALL section_vals_val_get(qs_section, "LS_SCF", l_val=qs_control%do_ls_scf)
      CALL section_vals_val_get(qs_section, "ALMO_SCF", l_val=qs_control%do_almo_scf)
      CALL section_vals_val_get(qs_section, "KG_METHOD", l_val=qs_control%do_kg)

      ! Logicals
      CALL section_vals_val_get(qs_section, "MAP_CONSISTENT", l_val=qs_control%map_consistent)

      ! Integers gapw
      CALL section_vals_val_get(qs_section, "LMAXN1", i_val=qs_control%gapw_control%lmax_sphere)
      CALL section_vals_val_get(qs_section, "LMAXN0", i_val=qs_control%gapw_control%lmax_rho0)
      CALL section_vals_val_get(qs_section, "LADDN0", i_val=qs_control%gapw_control%ladd_rho0)
      CALL section_vals_val_get(qs_section, "QUADRATURE", i_val=qs_control%gapw_control%quadrature)

      ! Integers grids
      CALL section_vals_val_get(qs_section, "PW_GRID", i_val=itmp)
      SELECT CASE (itmp)
      CASE (do_pwgrid_spherical)
         qs_control%pw_grid_opt%spherical = .TRUE.
         qs_control%pw_grid_opt%fullspace = .FALSE.
      CASE (do_pwgrid_ns_fullspace)
         qs_control%pw_grid_opt%spherical = .FALSE.
         qs_control%pw_grid_opt%fullspace = .TRUE.
      CASE (do_pwgrid_ns_halfspace)
         qs_control%pw_grid_opt%spherical = .FALSE.
         qs_control%pw_grid_opt%fullspace = .FALSE.
      END SELECT

      !   Method for PPL calculation
      CALL section_vals_val_get(qs_section, "CORE_PPL", i_val=itmp)
      qs_control%do_ppl_method = itmp

      CALL section_vals_val_get(qs_section, "PW_GRID_LAYOUT", i_vals=tmplist)
      qs_control%pw_grid_opt%distribution_layout = tmplist
      CALL section_vals_val_get(qs_section, "PW_GRID_BLOCKED", i_val=qs_control%pw_grid_opt%blocked)

      !Integers extrapolation
      CALL section_vals_val_get(qs_section, "EXTRAPOLATION", i_val=qs_control%wf_interpolation_method_nr)
      CALL section_vals_val_get(qs_section, "EXTRAPOLATION_ORDER", i_val=qs_control%wf_extrapolation_order)

      !Method
      CALL section_vals_val_get(qs_section, "METHOD", i_val=qs_control%method_id)
      NULLIFY (section, keyword, enum)
      CALL create_qs_section(section)
      keyword => section_get_keyword(section, "METHOD")
      CALL keyword_get(keyword, enum=enum)
      qs_control%method = enum_i2c(enum, qs_control%method_id)
      CALL section_release(section)
      qs_control%gapw = .FALSE.
      qs_control%gapw_xc = .FALSE.
      qs_control%gpw = .FALSE.
      qs_control%pao = .FALSE.
      qs_control%dftb = .FALSE.
      qs_control%scptb = .FALSE.
      qs_control%semi_empirical = .FALSE.
      qs_control%ofgpw = .FALSE.
      qs_control%lrigpw = .FALSE.
      SELECT CASE (qs_control%method_id)
      CASE (do_method_gapw)
         CALL cite_reference(Lippert1999)
         CALL cite_reference(Krack2000)
         qs_control%gapw = .TRUE.
      CASE (do_method_gapw_xc)
         qs_control%gapw_xc = .TRUE.
      CASE (do_method_gpw)
         CALL cite_reference(Lippert1997)
         CALL cite_reference(VandeVondele2005a)
         qs_control%gpw = .TRUE.
      CASE (do_method_ofgpw)
         qs_control%ofgpw = .TRUE.
      CASE (do_method_lrigpw)
         qs_control%lrigpw = .TRUE.
      CASE (do_method_dftb)
         qs_control%dftb = .TRUE.
         CALL cite_reference(Porezag1995)
         CALL cite_reference(Seifert1996)
      CASE (do_method_scptb)
         qs_control%scptb = .TRUE.
      CASE (do_method_mndo)
         CALL cite_reference(Dewar1977)
         qs_control%semi_empirical = .TRUE.
      CASE (do_method_am1)
         CALL cite_reference(Dewar1985)
         qs_control%semi_empirical = .TRUE.
      CASE (do_method_pm3)
         CALL cite_reference(Stewart1989)
         qs_control%semi_empirical = .TRUE.
      CASE (do_method_pnnl)
         CALL cite_reference(Schenter2008)
         qs_control%semi_empirical = .TRUE.
      CASE (do_method_pm6)
         CALL cite_reference(Stewart2007)
         qs_control%semi_empirical = .TRUE.
      CASE (do_method_pdg)
         CALL cite_reference(Repasky2002)
         qs_control%semi_empirical = .TRUE.
      CASE (do_method_rm1)
         CALL cite_reference(Rocha2006)
         qs_control%semi_empirical = .TRUE.
      CASE (do_method_mndod)
         CALL cite_reference(Dewar1977)
         CALL cite_reference(Thiel1992)
         qs_control%semi_empirical = .TRUE.
      END SELECT

      CALL section_vals_get(mull_section, explicit=qs_control%mulliken_restraint)

      IF (qs_control%mulliken_restraint) THEN
         CALL section_vals_val_get(mull_section, "STRENGTH", r_val=qs_control%mulliken_restraint_control%strength)
         CALL section_vals_val_get(mull_section, "TARGET", r_val=qs_control%mulliken_restraint_control%target)
         CALL section_vals_val_get(mull_section, "ATOMS", n_rep_val=n_rep)
         jj = 0
         DO k = 1, n_rep
            CALL section_vals_val_get(mull_section, "ATOMS", i_rep_val=k, i_vals=tmplist)
            jj = jj+SIZE(tmplist)
         END DO
         qs_control%mulliken_restraint_control%natoms = jj
         IF (qs_control%mulliken_restraint_control%natoms < 1) &
            CALL cp__b("cp_control_utils.F",763,"Need at least 1 atom to use mulliken contraints")
         ALLOCATE (qs_control%mulliken_restraint_control%atoms(qs_control%mulliken_restraint_control%natoms))
         jj = 0
         DO k = 1, n_rep
            CALL section_vals_val_get(mull_section, "ATOMS", i_rep_val=k, i_vals=tmplist)
            DO j = 1, SIZE(tmplist)
               jj = jj+1
               qs_control%mulliken_restraint_control%atoms(jj) = tmplist(j)
            END DO
         END DO
      ENDIF
      CALL section_vals_get(ddapc_restraint_section, n_repetition=nrep, explicit=qs_control%ddapc_restraint)
      IF (qs_control%ddapc_restraint) THEN
         ALLOCATE (qs_control%ddapc_restraint_control(nrep))
         CALL read_ddapc_section(qs_control, qs_section=qs_section)
         qs_control%ddapc_restraint_is_spin = .FALSE.
         qs_control%ddapc_explicit_potential = .FALSE.
      ENDIF

      CALL section_vals_get(s2_restraint_section, explicit=qs_control%s2_restraint)
      IF (qs_control%s2_restraint) THEN
         CALL section_vals_val_get(s2_restraint_section, "STRENGTH", &
                                   r_val=qs_control%s2_restraint_control%strength)
         CALL section_vals_val_get(s2_restraint_section, "TARGET", &
                                   r_val=qs_control%s2_restraint_control%target)
         CALL section_vals_val_get(s2_restraint_section, "FUNCTIONAL_FORM", &
                                   i_val=qs_control%s2_restraint_control%functional_form)
      ENDIF

      CALL section_vals_get(becke_restraint_section, explicit=qs_control%becke_restraint)

      IF (qs_control%becke_restraint) THEN
         CALL read_becke_section(qs_control, becke_restraint_section)
      ENDIF

      ! Semi-empirical code
      IF (qs_control%semi_empirical) THEN
         CALL section_vals_val_get(se_section, "ORTHOGONAL_BASIS", &
                                   l_val=qs_control%se_control%orthogonal_basis)
         CALL section_vals_val_get(se_section, "DELTA", &
                                   r_val=qs_control%se_control%delta)
         CALL section_vals_val_get(se_section, "ANALYTICAL_GRADIENTS", &
                                   l_val=qs_control%se_control%analytical_gradients)
         CALL section_vals_val_get(se_section, "FORCE_KDSO-D_EXCHANGE", &
                                   l_val=qs_control%se_control%force_kdsod_EX)
         ! Integral Screening
         CALL section_vals_val_get(se_section, "INTEGRAL_SCREENING", &
                                   i_val=qs_control%se_control%integral_screening)
         IF (qs_control%method_id == do_method_pnnl) THEN
            IF (qs_control%se_control%integral_screening /= do_se_IS_slater) &
               CALL cp_warn(cp__l("cp_control_utils.F",813), &
                            "PNNL semi-empirical parameterization supports only the Slater type "// &
                            "integral scheme. Revert to Slater and continue the calculation.")
            qs_control%se_control%integral_screening = do_se_IS_slater
         END IF
         ! Global Arrays variable
         CALL section_vals_val_get(se_section, "GA%NCELLS", &
                                   i_val=qs_control%se_control%ga_ncells)
         ! Long-Range correction
         CALL section_vals_val_get(se_section, "LR_CORRECTION%CUTOFF", &
                                   r_val=qs_control%se_control%cutoff_lrc)
         qs_control%se_control%taper_lrc = qs_control%se_control%cutoff_lrc
         CALL section_vals_val_get(se_section, "LR_CORRECTION%RC_TAPER", &
                                   explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(se_section, "LR_CORRECTION%RC_TAPER", &
                                      r_val=qs_control%se_control%taper_lrc)
         END IF
         CALL section_vals_val_get(se_section, "LR_CORRECTION%RC_RANGE", &
                                   r_val=qs_control%se_control%range_lrc)
         ! Coulomb
         CALL section_vals_val_get(se_section, "COULOMB%CUTOFF", &
                                   r_val=qs_control%se_control%cutoff_cou)
         qs_control%se_control%taper_cou = qs_control%se_control%cutoff_cou
         CALL section_vals_val_get(se_section, "COULOMB%RC_TAPER", &
                                   explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(se_section, "COULOMB%RC_TAPER", &
                                      r_val=qs_control%se_control%taper_cou)
         END IF
         CALL section_vals_val_get(se_section, "COULOMB%RC_RANGE", &
                                   r_val=qs_control%se_control%range_cou)
         ! Exchange
         CALL section_vals_val_get(se_section, "EXCHANGE%CUTOFF", &
                                   r_val=qs_control%se_control%cutoff_exc)
         qs_control%se_control%taper_exc = qs_control%se_control%cutoff_exc
         CALL section_vals_val_get(se_section, "EXCHANGE%RC_TAPER", &
                                   explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(se_section, "EXCHANGE%RC_TAPER", &
                                      r_val=qs_control%se_control%taper_exc)
         END IF
         CALL section_vals_val_get(se_section, "EXCHANGE%RC_RANGE", &
                                   r_val=qs_control%se_control%range_exc)
         ! Screening (only if the integral scheme is of dumped type)
         IF (qs_control%se_control%integral_screening == do_se_IS_kdso_d) THEN
            CALL section_vals_val_get(se_section, "SCREENING%RC_TAPER", &
                                      r_val=qs_control%se_control%taper_scr)
            CALL section_vals_val_get(se_section, "SCREENING%RC_RANGE", &
                                      r_val=qs_control%se_control%range_scr)
         END IF
         ! Periodic Type Calculation
         CALL section_vals_val_get(se_section, "PERIODIC", &
                                   i_val=qs_control%se_control%periodic_type)
         SELECT CASE (qs_control%se_control%periodic_type)
         CASE (do_se_lr_none)
            qs_control%se_control%do_ewald = .FALSE.
            qs_control%se_control%do_ewald_r3 = .FALSE.
            qs_control%se_control%do_ewald_gks = .FALSE.
         CASE (do_se_lr_ewald)
            qs_control%se_control%do_ewald = .TRUE.
            qs_control%se_control%do_ewald_r3 = .FALSE.
            qs_control%se_control%do_ewald_gks = .FALSE.
         CASE (do_se_lr_ewald_gks)
            qs_control%se_control%do_ewald = .FALSE.
            qs_control%se_control%do_ewald_r3 = .FALSE.
            qs_control%se_control%do_ewald_gks = .TRUE.
            IF (qs_control%method_id /= do_method_pnnl) &
               CALL cp_abort(cp__l("cp_control_utils.F",881), &
                             "A periodic semi-empirical calculation was requested with a long-range  "// &
                             "summation on the single integral evaluation. This scheme is supported  "// &
                             "only by the PNNL parameterization.")
         CASE (do_se_lr_ewald_r3)
            qs_control%se_control%do_ewald = .TRUE.
            qs_control%se_control%do_ewald_r3 = .TRUE.
            qs_control%se_control%do_ewald_gks = .FALSE.
            IF (qs_control%se_control%integral_screening /= do_se_IS_kdso) &
               CALL cp_abort(cp__l("cp_control_utils.F",890), &
                             "A periodic semi-empirical calculation was requested with a long-range  "// &
                             "summation for the slowly convergent part 1/R^3, which is not congruent "// &
                             "with the integral screening chosen. The only integral screening supported "// &
                             "by this periodic type calculation is the standard Klopman-Dewar-Sabelli-Ohno.")
         END SELECT

         ! dispersion pair potentials
         CALL section_vals_val_get(se_section, "DISPERSION", &
                                   l_val=qs_control%se_control%dispersion)
         CALL section_vals_val_get(se_section, "DISPERSION_RADIUS", &
                                   r_val=qs_control%se_control%rcdisp)
         CALL section_vals_val_get(se_section, "COORDINATION_CUTOFF", &
                                   r_val=qs_control%se_control%epscn)
         CALL section_vals_val_get(se_section, "D3_SCALING", r_vals=scal)
         qs_control%se_control%sd3(1) = scal(1)
         qs_control%se_control%sd3(2) = scal(2)
         qs_control%se_control%sd3(3) = scal(3)
         CALL section_vals_val_get(se_section, "DISPERSION_PARAMETER_FILE", &
                                   c_val=qs_control%se_control%dispersion_parameter_file)

         ! Stop the execution for non-implemented features
         IF (qs_control%se_control%periodic_type == do_se_lr_ewald_r3) THEN
            CALL cp__b("cp_control_utils.F",913,"EWALD_R3 not implemented yet!")
         END IF

         IF (qs_control%method_id == do_method_mndo .OR. &
             qs_control%method_id == do_method_am1 .OR. &
             qs_control%method_id == do_method_mndod .OR. &
             qs_control%method_id == do_method_pdg .OR. &
             qs_control%method_id == do_method_pm3 .OR. &
             qs_control%method_id == do_method_pm6 .OR. &
             qs_control%method_id == do_method_pnnl .OR. &
             qs_control%method_id == do_method_rm1) THEN
            qs_control%se_control%orthogonal_basis = .TRUE.
         END IF
      END IF

      ! DFTB code
      IF (qs_control%dftb) THEN
         CALL section_vals_val_get(dftb_section, "ORTHOGONAL_BASIS", &
                                   l_val=qs_control%dftb_control%orthogonal_basis)
         CALL section_vals_val_get(dftb_section, "SELF_CONSISTENT", &
                                   l_val=qs_control%dftb_control%self_consistent)
         CALL section_vals_val_get(dftb_section, "DISPERSION", &
                                   l_val=qs_control%dftb_control%dispersion)
         CALL section_vals_val_get(dftb_section, "DIAGONAL_DFTB3", &
                                   l_val=qs_control%dftb_control%dftb3_diagonal)
         CALL section_vals_val_get(dftb_section, "HB_SR_GAMMA", &
                                   l_val=qs_control%dftb_control%hb_sr_damp)
         CALL section_vals_val_get(dftb_section, "EPS_DISP", &
                                   r_val=qs_control%dftb_control%eps_disp)
         CALL section_vals_val_get(dftb_section, "DO_EWALD", &
                                   l_val=qs_control%dftb_control%do_ewald)
         CALL section_vals_val_get(dftb_parameter, "PARAM_FILE_PATH", &
                                   c_val=qs_control%dftb_control%sk_file_path)
         CALL section_vals_val_get(dftb_parameter, "PARAM_FILE_NAME", &
                                   c_val=qs_control%dftb_control%sk_file_list)
         CALL section_vals_val_get(dftb_parameter, "HB_SR_PARAM", &
                                   r_val=qs_control%dftb_control%hb_sr_para)
         CALL section_vals_val_get(dftb_parameter, "SK_FILE", n_rep_val=n_var)
         ALLOCATE (qs_control%dftb_control%sk_pair_list(3, n_var))
         DO k = 1, n_var
            CALL section_vals_val_get(dftb_parameter, "SK_FILE", i_rep_val=k, &
                                      c_vals=clist)
            qs_control%dftb_control%sk_pair_list(1:3, k) = clist(1:3)
         END DO
         ! Dispersion type
         CALL section_vals_val_get(dftb_parameter, "DISPERSION_TYPE", &
                                   i_val=qs_control%dftb_control%dispersion_type)
         CALL section_vals_val_get(dftb_parameter, "UFF_FORCE_FIELD", &
                                   c_val=qs_control%dftb_control%uff_force_field)
         ! D3 Dispersion
         CALL section_vals_val_get(dftb_parameter, "DISPERSION_RADIUS", &
                                   r_val=qs_control%dftb_control%rcdisp)
         CALL section_vals_val_get(dftb_parameter, "COORDINATION_CUTOFF", &
                                   r_val=qs_control%dftb_control%epscn)
         CALL section_vals_val_get(dftb_parameter, "D3_SCALING", r_vals=scal)
         qs_control%dftb_control%sd3(1) = scal(1)
         qs_control%dftb_control%sd3(2) = scal(2)
         qs_control%dftb_control%sd3(3) = scal(3)
         CALL section_vals_val_get(dftb_parameter, "DISPERSION_PARAMETER_FILE", &
                                   c_val=qs_control%dftb_control%dispersion_parameter_file)

         IF (qs_control%dftb_control%dispersion) CALL cite_reference(Zhechkov2005)
         IF (qs_control%dftb_control%self_consistent) CALL cite_reference(Elstner1998)
         IF (qs_control%dftb_control%hb_sr_damp) CALL cite_reference(Hu2007)
      END IF

      ! SCPTB code
      IF (qs_control%scptb) THEN
         CALL section_vals_val_get(scptb_section, "DISPERSION", &
                                   l_val=qs_control%scptb_control%dispersion)
         CALL section_vals_val_get(scptb_section, "PARAMETER_FILE_NAME", &
                                   c_val=qs_control%scptb_control%parameter_file)
         CALL section_vals_val_get(scptb_section, "DISPERSION_RADIUS", &
                                   r_val=qs_control%scptb_control%rcdisp)
         CALL section_vals_val_get(scptb_section, "COORDINATION_CUTOFF", &
                                   r_val=qs_control%scptb_control%epscn)
         CALL section_vals_val_get(scptb_section, "D3_SCALING", r_vals=scal)
         qs_control%scptb_control%sd3(1) = scal(1)
         qs_control%scptb_control%sd3(2) = scal(2)
         qs_control%scptb_control%sd3(3) = scal(3)
         CALL section_vals_val_get(scptb_section, "STO_NG", &
                                   i_val=qs_control%scptb_control%sto_ng)
         CALL section_vals_val_get(scptb_section, "DISPERSION_PARAMETER_FILE", &
                                   c_val=qs_control%scptb_control%dispersion_parameter_file)
         CALL section_vals_val_get(scptb_section, "PAIR_CUTOFF", &
                                   r_val=qs_control%scptb_control%epspair)
         CALL section_vals_val_get(scptb_section, "DO_EWALD", &
                                   l_val=qs_control%scptb_control%do_ewald)
         CALL section_vals_val_get(scptb_section, "DO_SCP", &
                                   l_val=qs_control%scptb_control%do_scp)
         CALL section_vals_val_get(scptb_section, "DO_SCC", &
                                   l_val=qs_control%scptb_control%do_scc)
      END IF

      ! Optimize LRI basis set
      CALL section_vals_get(lri_optbas_section, explicit=qs_control%lri_optbas)

      CALL timestop(handle)
   END SUBROUTINE read_qs_section

! **************************************************************************************************
!> \brief ...
!> \param t_control ...
!> \param dft_section ...
! **************************************************************************************************
   SUBROUTINE read_tddfpt_control(t_control, dft_section)
      TYPE(tddfpt_control_type)                          :: t_control
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_tddfpt_control', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: kenergy_den
      TYPE(section_vals_type), POINTER                   :: sic_section, t_section

      kenergy_den = .FALSE.
      NULLIFY (sic_section, t_section)
      t_section => section_vals_get_subs_vals(dft_section, "TDDFPT")

      CALL section_vals_val_get(t_section, "CONVERGENCE", r_val=t_control%tolerance)
      CALL section_vals_val_get(t_section, "NEV", i_val=t_control%n_ev)
      CALL section_vals_val_get(t_section, "MAX_KV", i_val=t_control%max_kv)
      CALL section_vals_val_get(t_section, "RESTARTS", i_val=t_control%n_restarts)
      CALL section_vals_val_get(t_section, "NREORTHO", i_val=t_control%n_reortho)
      CALL section_vals_val_get(t_section, "RES_ETYPE", i_val=t_control%res_etype)
      CALL section_vals_val_get(t_section, "DIAG_METHOD", i_val=t_control%diag_method)
      CALL section_vals_val_get(t_section, "KERNEL", l_val=t_control%do_kernel)
      CALL section_vals_val_get(t_section, "LSD_SINGLETS", l_val=t_control%lsd_singlets)
      CALL section_vals_val_get(t_section, "INVERT_S", l_val=t_control%invert_S)
      CALL section_vals_val_get(t_section, "PRECOND", l_val=t_control%precond)
      CALL section_vals_val_get(t_section, "OE_CORR", i_val=t_control%oe_corr)

      t_control%use_kinetic_energy_density = .FALSE.
      sic_section => section_vals_get_subs_vals(t_section, "SIC")
      CALL section_vals_val_get(sic_section, "SIC_METHOD", i_val=t_control%sic_method_id)
      CALL section_vals_val_get(sic_section, "ORBITAL_SET", i_val=t_control%sic_list_id)
      CALL section_vals_val_get(sic_section, "SIC_SCALING_A", r_val=t_control%sic_scaling_a)
      CALL section_vals_val_get(sic_section, "SIC_SCALING_B", r_val=t_control%sic_scaling_b)

   END SUBROUTINE read_tddfpt_control

! **************************************************************************************************
!> \brief Write the DFT control parameters to the output unit.
!> \param dft_control ...
!> \param dft_section ...
! **************************************************************************************************
   SUBROUTINE write_dft_control(dft_control, dft_section)
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_dft_control', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=20)                                  :: tmpStr
      INTEGER                                            :: handle, output_unit
      REAL(kind=dp)                                      :: density_cut, density_smooth_cut_range, &
                                                            gradient_cut, tau_cut
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: xc_section

      IF (dft_control%qs_control%semi_empirical) RETURN
      IF (dft_control%qs_control%dftb) RETURN
      IF (dft_control%qs_control%scptb) RETURN
      CALL timeset(routineN, handle)

      NULLIFY (logger)
      logger => cp_get_default_logger()

      output_unit = cp_print_key_unit_nr(logger, dft_section, &
                                         "PRINT%DFT_CONTROL_PARAMETERS", extension=".Log")

      IF (output_unit > 0) THEN

         xc_section => section_vals_get_subs_vals(dft_section, "XC")

         IF (dft_control%uks) THEN
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T78,A)") &
               "DFT| Spin unrestricted (spin-polarized) Kohn-Sham calculation", "UKS"
         ELSE IF (dft_control%roks) THEN
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T77,A)") &
               "DFT| Spin restricted open Kohn-Sham calculation", "ROKS"
         ELSE
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T78,A)") &
               "DFT| Spin restricted Kohn-Sham (RKS) calculation", "RKS"
         END IF

         WRITE (UNIT=output_unit, FMT="(T2,A,T76,I5)") &
            "DFT| Multiplicity", dft_control%multiplicity
         WRITE (UNIT=output_unit, FMT="(T2,A,T76,I5)") &
            "DFT| Number of spin states", dft_control%nspins

         WRITE (UNIT=output_unit, FMT="(T2,A,T76,I5)") &
            "DFT| Charge", dft_control%charge

         IF (dft_control%sic_method_id .NE. sic_none) CALL cite_reference(VandeVondele2005b)
         SELECT CASE (dft_control%sic_method_id)
         CASE (sic_none)
            tmpstr = "NO"
         CASE (sic_mauri_spz)
            tmpstr = "SPZ/MAURI SIC"
         CASE (sic_mauri_us)
            tmpstr = "US/MAURI SIC"
         CASE (sic_ad)
            tmpstr = "AD SIC"
         CASE (sic_eo)
            tmpstr = "Explicit Orbital SIC"
         CASE DEFAULT
            ! fix throughout the cp2k for this option
            CALL cp__b("cp_control_utils.F",1124,"SIC option unknown")
         END SELECT

         WRITE (UNIT=output_unit, FMT="(T2,A,T61,A20)") &
            "DFT| Self-interaction correction (SIC)", ADJUSTR(TRIM(tmpstr))

         IF (dft_control%sic_method_id /= sic_none) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T66,ES15.6)") &
               "DFT| SIC scaling parameter a", dft_control%sic_scaling_a, &
               "DFT| SIC scaling parameter b", dft_control%sic_scaling_b
         END IF

         IF (dft_control%sic_method_id == sic_eo) THEN
            IF (dft_control%sic_list_id == sic_list_all) THEN
               WRITE (UNIT=output_unit, FMT="(T2,A,T66,A)") &
                  "DFT| SIC orbitals", "ALL"
            ENDIF
            IF (dft_control%sic_list_id == sic_list_unpaired) THEN
               WRITE (UNIT=output_unit, FMT="(T2,A,T66,A)") &
                  "DFT| SIC orbitals", "UNPAIRED"
            ENDIF
         END IF

         CALL section_vals_val_get(xc_section, "density_cutoff", r_val=density_cut)
         CALL section_vals_val_get(xc_section, "gradient_cutoff", r_val=gradient_cut)
         CALL section_vals_val_get(xc_section, "tau_cutoff", r_val=tau_cut)
         CALL section_vals_val_get(xc_section, "density_smooth_cutoff_range", r_val=density_smooth_cut_range)

         WRITE (UNIT=output_unit, FMT="(T2,A,T66,ES15.6)") &
            "DFT| Cutoffs: density ", density_cut, &
            "DFT|          gradient", gradient_cut, &
            "DFT|          tau     ", tau_cut, &
            "DFT|          cutoff_smoothing_range", density_smooth_cut_range
         CALL section_vals_val_get(xc_section, "XC_GRID%XC_SMOOTH_RHO", &
                                   c_val=tmpStr)
         WRITE (output_unit, '( A, T61, A )') &
            " DFT| XC density smoothing ", ADJUSTR(tmpStr)
         CALL section_vals_val_get(xc_section, "XC_GRID%XC_DERIV", &
                                   c_val=tmpStr)
         WRITE (output_unit, '( A, T61, A )') &
            " DFT| XC derivatives ", ADJUSTR(tmpStr)
         IF (dft_control%dft_plus_u) THEN
            NULLIFY (enum, keyword, section)
            CALL create_dft_section(section)
            keyword => section_get_keyword(section, "PLUS_U_METHOD")
            CALL keyword_get(keyword, enum=enum)
            WRITE (UNIT=output_unit, FMT="(/,T2,A,T41,A40)") &
               "DFT+U| Method", ADJUSTR(TRIM(enum_i2c(enum, dft_control%plus_u_method_id)))
            WRITE (UNIT=output_unit, FMT="(T2,A)") &
               "DFT+U| Check atomic kind information for details"
            CALL section_release(section)
         END IF

         CALL xc_write(output_unit, xc_section, dft_control%lsd)

         IF (dft_control%do_sccs) THEN
            IF (dft_control%qs_control%gapw) THEN
               CALL cp__b("cp_control_utils.F",1181,"SCCS is not yet implemented with GAPW")
            END IF
            WRITE (UNIT=output_unit, FMT="(/,T2,A)") &
               "SCCS| Self-consistent continuum solvation model"
            SELECT CASE (dft_control%sccs_control%method_id)
            CASE (sccs_andreussi)
               WRITE (UNIT=output_unit, FMT="(T2,A,/,(T2,A,T69,ES12.3))") &
                  "SCCS| Dielectric function proposed by Andreussi et al.", &
                  "SCCS|  rho_max", dft_control%sccs_control%rho_max, &
                  "SCCS|  rho_min", dft_control%sccs_control%rho_min
            CASE (sccs_fattebert_gygi)
               WRITE (UNIT=output_unit, FMT="(T2,A,/,(T2,A,T69,ES12.3))") &
                  "SCCS| Dielectric function proposed by Fattebert and Gygi", &
                  "SCCS|  beta", dft_control%sccs_control%beta, &
                  "SCCS|  rho_zero", dft_control%sccs_control%rho_zero
            CASE DEFAULT
               CALL cp__b("cp_control_utils.F",1197,"Invalid SCCS model specified. Please, check your input!")
            END SELECT
            SELECT CASE (dft_control%sccs_control%derivative_method)
            CASE (sccs_derivative_fft)
               WRITE (UNIT=output_unit, FMT="(T2,A,T46,A35)") "SCCS| Numerical derivative calculation", &
                  ADJUSTR("FFT")
            CASE (sccs_derivative_cd3)
               WRITE (UNIT=output_unit, FMT="(T2,A,T46,A35)") "SCCS| Numerical derivative calculation", &
                  ADJUSTR("3-point stencil central differences")
            CASE (sccs_derivative_cd5)
               WRITE (UNIT=output_unit, FMT="(T2,A,T46,A35)") "SCCS| Numerical derivative calculation", &
                  ADJUSTR("5-point stencil central differences")
            CASE (sccs_derivative_cd7)
               WRITE (UNIT=output_unit, FMT="(T2,A,T46,A35)") "SCCS| Numerical derivative calculation", &
                  ADJUSTR("7-point stencil central differences")
            CASE DEFAULT
               CALL cp_abort(cp__l("cp_control_utils.F",1213), &
                             "Invalid derivative method specified for SCCS model. "// &
                             "Please, check your input!")
            END SELECT
            WRITE (UNIT=output_unit, FMT="(T2,A,T69,F12.3)") &
               "SCCS| Repulsion parameter alpha [mN/m] = [dyn/cm]", &
               cp_unit_from_cp2k(dft_control%sccs_control%alpha_solvent, "mN/m")
            WRITE (UNIT=output_unit, FMT="(T2,A,T69,F12.3)") &
               "SCCS| Dispersion parameter beta [GPa]", &
               cp_unit_from_cp2k(dft_control%sccs_control%beta_solvent, "GPa")
            WRITE (UNIT=output_unit, FMT="(T2,A,T69,F12.3)") &
               "SCCS| Surface tension gamma [mN/m] = [dyn/cm]", &
               cp_unit_from_cp2k(dft_control%sccs_control%gamma_solvent, "mN/m")
            WRITE (UNIT=output_unit, FMT="(T2,A,T69,F12.3)") &
               "SCCS| Mixing parameter applied during the iteration cycle", &
               dft_control%sccs_control%mixing
            WRITE (UNIT=output_unit, FMT="(T2,A,T69,ES12.3)") &
               "SCCS| Tolerance for the convergence of the SCCS iteration cycle", &
               dft_control%sccs_control%eps_sccs
            WRITE (UNIT=output_unit, FMT="(T2,A,T69,I12)") &
               "SCCS| Maximum number of iteration steps", &
               dft_control%sccs_control%max_iter
            WRITE (UNIT=output_unit, FMT="(T2,A,T69,ES12.3)") &
               "SCCS| SCF convergence threshold for starting the SCCS iteration", &
               dft_control%sccs_control%eps_scf
            WRITE (UNIT=output_unit, FMT="(T2,A,T69,ES12.3)") &
               "SCCS| Numerical increment for the cavity surface calculation", &
               dft_control%sccs_control%delta_rho
         END IF

      END IF

      CALL cp_print_key_finished_output(output_unit, logger, dft_section, &
                                        "PRINT%DFT_CONTROL_PARAMETERS")

      CALL timestop(handle)

   END SUBROUTINE write_dft_control

! **************************************************************************************************
!> \brief Purpose: Write the QS control parameters to the output unit.
!> \param qs_control ...
!> \param dft_section ...
! **************************************************************************************************
   SUBROUTINE write_qs_control(qs_control, dft_section)
      TYPE(qs_control_type), INTENT(IN)                  :: qs_control
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_qs_control', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, igrid_level, ngrid_level, &
                                                            output_unit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(ddapc_restraint_type), POINTER                :: ddapc_restraint_control
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section

      IF (qs_control%semi_empirical) RETURN
      IF (qs_control%dftb) RETURN
      IF (qs_control%scptb) RETURN
      CALL timeset(routineN, handle)
      NULLIFY (logger, enum, keyword, section)
      logger => cp_get_default_logger()

      CALL create_qs_section(section)
      keyword => section_get_keyword(section, "QUADRATURE")
      CALL keyword_get(keyword, enum=enum)

      output_unit = cp_print_key_unit_nr(logger, dft_section, &
                                         "PRINT%DFT_CONTROL_PARAMETERS", extension=".Log")
      IF (output_unit > 0) THEN
         ngrid_level = SIZE(qs_control%e_cutoff)
         WRITE (UNIT=output_unit, FMT="(/,T2,A,T71,A)") &
            "QS| Method:", ADJUSTR(qs_control%method)
         IF (qs_control%pw_grid_opt%spherical) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T61,A)") &
               "QS| Density plane wave grid type", " SPHERICAL HALFSPACE"
         ELSE IF (qs_control%pw_grid_opt%fullspace) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T57,A)") &
               "QS| Density plane wave grid type", " NON-SPHERICAL FULLSPACE"
         ELSE
            WRITE (UNIT=output_unit, FMT="(T2,A,T57,A)") &
               "QS| Density plane wave grid type", " NON-SPHERICAL HALFSPACE"
         END IF
         WRITE (UNIT=output_unit, FMT="(T2,A,T71,I10)") &
            "QS| Number of grid levels:", SIZE(qs_control%e_cutoff)
         IF (ngrid_level == 1) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T71,F10.1)") &
               "QS| Density cutoff [a.u.]:", qs_control%e_cutoff(1)
         ELSE
            WRITE (UNIT=output_unit, FMT="(T2,A,T71,F10.1)") &
               "QS| Density cutoff [a.u.]:", qs_control%cutoff
            IF (qs_control%commensurate_mgrids) &
               WRITE (UNIT=output_unit, FMT="(T2,A)") "QS| Using commensurate multigrids"
            WRITE (UNIT=output_unit, FMT="(T2,A,T71,F10.1)") &
               "QS| Multi grid cutoff [a.u.]: 1) grid level", qs_control%e_cutoff(1)
            WRITE (UNIT=output_unit, FMT="(T2,A,I3,A,T71,F10.1)") &
               ("QS|                         ", igrid_level, ") grid level", &
                qs_control%e_cutoff(igrid_level), &
                igrid_level=2, SIZE(qs_control%e_cutoff))
         END IF
         IF (qs_control%pao) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A)") "QS| PAO active"
         END IF
         WRITE (UNIT=output_unit, FMT="(T2,A,T71,F10.1)") &
            "QS| Grid level progression factor:", qs_control%progression_factor
         WRITE (UNIT=output_unit, FMT="(T2,A,T71,F10.1)") &
            "QS| Relative density cutoff [a.u.]:", qs_control%relative_cutoff
         IF (qs_control%map_consistent) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A)") &
               "QS| Consistent realspace mapping and integration "
         ENDIF
         WRITE (UNIT=output_unit, FMT="(T2,A,T73,ES8.1)") &
            "QS| Interaction thresholds: eps_pgf_orb:", &
            qs_control%eps_pgf_orb, &
            "QS|                         eps_filter_matrix:", &
            qs_control%eps_filter_matrix, &
            "QS|                         eps_core_charge:", &
            qs_control%eps_core_charge, &
            "QS|                         eps_rho_gspace:", &
            qs_control%eps_rho_gspace, &
            "QS|                         eps_rho_rspace:", &
            qs_control%eps_rho_rspace, &
            "QS|                         eps_gvg_rspace:", &
            qs_control%eps_gvg_rspace, &
            "QS|                         eps_ppl:", &
            qs_control%eps_ppl, &
            "QS|                         eps_ppnl:", &
            qs_control%eps_ppnl
         IF (qs_control%gapw) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T73,ES8.1)") &
               "QS| GAPW|                   eps_fit:", &
               qs_control%gapw_control%eps_fit, &
               "QS| GAPW|                   eps_iso:", &
               qs_control%gapw_control%eps_iso, &
               "QS| GAPW|                   eps_svd:", &
               qs_control%gapw_control%eps_svd, &
               "QS| GAPW|                   eps_cpc:", &
               qs_control%gapw_control%eps_cpc
            WRITE (UNIT=output_unit, FMT="(T2,A,T55,A30)") &
               "QS| GAPW|   atom-r-grid: quadrature:", &
               enum_i2c(enum, qs_control%gapw_control%quadrature)
            WRITE (UNIT=output_unit, FMT="(T2,A,T71,I10)") &
               "QS| GAPW|      atom-s-grid:  max l :", &
               qs_control%gapw_control%lmax_sphere, &
               "QS| GAPW|      max_l_rho0 :", &
               qs_control%gapw_control%lmax_rho0
            IF (qs_control%gapw_control%non_paw_atoms) THEN
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "QS| GAPW|      At least one kind is NOT PAW, i.e. it has only soft AO "
            END IF
            IF (qs_control%gapw_control%nopaw_as_gpw) THEN
               WRITE (UNIT=output_unit, FMT="(T2,A)") &
                  "QS| GAPW|      The NOT PAW atoms are treated fully GPW"
            END IF
         END IF
         IF (qs_control%gapw_xc) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T73,ES8.1)") &
               "QS| GAPW_XC|                eps_fit:", &
               qs_control%gapw_control%eps_fit, &
               "QS| GAPW_XC|                eps_iso:", &
               qs_control%gapw_control%eps_iso, &
               "QS| GAPW_XC|                eps_svd:", &
               qs_control%gapw_control%eps_svd
            WRITE (UNIT=output_unit, FMT="(T2,A,T55,A30)") &
               "QS| GAPW_XC|atom-r-grid: quadrature:", &
               enum_i2c(enum, qs_control%gapw_control%quadrature)
            WRITE (UNIT=output_unit, FMT="(T2,A,T71,I10)") &
               "QS| GAPW_XC|   atom-s-grid:  max l :", &
               qs_control%gapw_control%lmax_sphere
         END IF
         IF (qs_control%mulliken_restraint) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T73,ES8.1)") &
               "QS| Mulliken restraint target", qs_control%mulliken_restraint_control%target
            WRITE (UNIT=output_unit, FMT="(T2,A,T73,ES8.1)") &
               "QS| Mulliken restraint strength", qs_control%mulliken_restraint_control%strength
            WRITE (UNIT=output_unit, FMT="(T2,A,T73,I8)") &
               "QS| Mulliken restraint atoms: ", qs_control%mulliken_restraint_control%natoms
            WRITE (UNIT=output_unit, FMT="(5I8)") qs_control%mulliken_restraint_control%atoms
         END IF
         IF (qs_control%ddapc_restraint) THEN
            DO i = 1, SIZE(qs_control%ddapc_restraint_control)
               ddapc_restraint_control => qs_control%ddapc_restraint_control(i)%ddapc_restraint_control
               IF (SIZE(qs_control%ddapc_restraint_control) .GT. 1) &
                  WRITE (UNIT=output_unit, FMT="(T2,A,T3,I8)") &
                  "QS| parameters for DDAPC restraint number", i
               WRITE (UNIT=output_unit, FMT="(T2,A,T73,ES8.1)") &
                  "QS| ddapc restraint target", ddapc_restraint_control%target
               WRITE (UNIT=output_unit, FMT="(T2,A,T73,ES8.1)") &
                  "QS| ddapc restraint strength", ddapc_restraint_control%strength
               WRITE (UNIT=output_unit, FMT="(T2,A,T73,I8)") &
                  "QS| ddapc restraint atoms: ", ddapc_restraint_control%natoms
               WRITE (UNIT=output_unit, FMT="(5I8)") ddapc_restraint_control%atoms
               WRITE (UNIT=output_unit, FMT="(T2,A)") "Coefficients:"
               WRITE (UNIT=output_unit, FMT="(5F6.2)") ddapc_restraint_control%coeff
               SELECT CASE (ddapc_restraint_control%functional_form)
               CASE (do_ddapc_restraint)
                  WRITE (UNIT=output_unit, FMT="(T2,A,T61,A20)") &
                     "QS| ddapc restraint functional form :", "RESTRAINT"
               CASE (do_ddapc_constraint)
                  WRITE (UNIT=output_unit, FMT="(T2,A,T61,A20)") &
                     "QS| ddapc restraint functional form :", "CONSTRAINT"
               CASE DEFAULT
                  CALL cp__b("cp_control_utils.F",1418,"Unknown ddapc restraint")
               END SELECT
            END DO
         END IF
         IF (qs_control%s2_restraint) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T73,ES8.1)") &
               "QS| s2 restraint target", qs_control%s2_restraint_control%target
            WRITE (UNIT=output_unit, FMT="(T2,A,T73,ES8.1)") &
               "QS| s2 restraint strength", qs_control%s2_restraint_control%strength
            SELECT CASE (qs_control%s2_restraint_control%functional_form)
            CASE (do_s2_restraint)
               WRITE (UNIT=output_unit, FMT="(T2,A,T61,A20)") &
                  "QS| s2 restraint functional form :", "RESTRAINT"
               CALL cp__b("cp_control_utils.F",1431,"Not yet implemented")
            CASE (do_s2_constraint)
               WRITE (UNIT=output_unit, FMT="(T2,A,T61,A20)") &
                  "QS| s2 restraint functional form :", "CONSTRAINT"
            CASE DEFAULT
               CALL cp__b("cp_control_utils.F",1436,"Unknown ddapc restraint")
            END SELECT
         END IF
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, dft_section, &
                                        "PRINT%DFT_CONTROL_PARAMETERS")
      CALL section_release(section)

      CALL timestop(handle)

   END SUBROUTINE write_qs_control

! **************************************************************************************************
!> \brief reads the input parameters needed for ddapc.
!> \param qs_control ...
!> \param qs_section ...
!> \param ddapc_restraint_section ...
!> \author fschiff
!> \note
!>      either reads DFT%QS%DDAPC_RESTRAINT or PROPERTIES%ET_coupling
!>      if(qs_section is present the DFT part is read, if ddapc_restraint_section
!>      is present ET_COUPLING is read. Avoid having both!!!
! **************************************************************************************************
   SUBROUTINE read_ddapc_section(qs_control, qs_section, ddapc_restraint_section)

      TYPE(qs_control_type), INTENT(INOUT)               :: qs_control
      TYPE(section_vals_type), OPTIONAL, POINTER         :: qs_section, ddapc_restraint_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_ddapc_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, jj, k, n_rep
      INTEGER, DIMENSION(:), POINTER                     :: tmplist
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rtmplist
      TYPE(ddapc_restraint_type), POINTER                :: ddapc_restraint_control
      TYPE(section_vals_type), POINTER                   :: ddapc_section

      IF (PRESENT(ddapc_restraint_section)) THEN
         IF (ASSOCIATED(qs_control%ddapc_restraint_control)) THEN
            IF (SIZE(qs_control%ddapc_restraint_control) .GE. 2) &
               CALL cp__b("cp_control_utils.F",1476,"ET_COUPLING cannot be used in combination with a normal restraint")
         ELSE
            ddapc_section => ddapc_restraint_section
            ALLOCATE (qs_control%ddapc_restraint_control(1))
         END IF
      END IF

      IF (PRESENT(qs_section)) THEN
         NULLIFY (ddapc_section)
         ddapc_section => section_vals_get_subs_vals(qs_section, &
                                                     "DDAPC_RESTRAINT")
      END IF

      DO i = 1, SIZE(qs_control%ddapc_restraint_control)

         NULLIFY (qs_control%ddapc_restraint_control(i)%ddapc_restraint_control)
         CALL ddapc_control_create(qs_control%ddapc_restraint_control(i)%ddapc_restraint_control)
         ddapc_restraint_control => qs_control%ddapc_restraint_control(i)%ddapc_restraint_control

         CALL section_vals_val_get(ddapc_section, "STRENGTH", i_rep_section=i, &
                                   r_val=ddapc_restraint_control%strength)
         CALL section_vals_val_get(ddapc_section, "TARGET", i_rep_section=i, &
                                   r_val=ddapc_restraint_control%target)
         CALL section_vals_val_get(ddapc_section, "FUNCTIONAL_FORM", i_rep_section=i, &
                                   i_val=ddapc_restraint_control%functional_form)
         CALL section_vals_val_get(ddapc_section, "ATOMS", i_rep_section=i, &
                                   n_rep_val=n_rep)
         CALL section_vals_val_get(ddapc_section, "TYPE_OF_DENSITY", i_rep_section=i, &
                                   i_val=ddapc_restraint_control%density_type)

         jj = 0
         DO k = 1, n_rep
            CALL section_vals_val_get(ddapc_section, "ATOMS", i_rep_section=i, &
                                      i_rep_val=k, i_vals=tmplist)
            DO j = 1, SIZE(tmplist)
               jj = jj+1
            END DO
         END DO
         IF (jj < 1) CALL cp__b("cp_control_utils.F",1514,"Need at least 1 atom to use ddapc contraints")
         ddapc_restraint_control%natoms = jj
         IF (ASSOCIATED(ddapc_restraint_control%atoms)) &
            DEALLOCATE (ddapc_restraint_control%atoms)
         ALLOCATE (ddapc_restraint_control%atoms(ddapc_restraint_control%natoms))
         jj = 0
         DO k = 1, n_rep
            CALL section_vals_val_get(ddapc_section, "ATOMS", i_rep_section=i, &
                                      i_rep_val=k, i_vals=tmplist)
            DO j = 1, SIZE(tmplist)
               jj = jj+1
               ddapc_restraint_control%atoms(jj) = tmplist(j)
            END DO
         END DO

         IF (ASSOCIATED(ddapc_restraint_control%coeff)) &
            DEALLOCATE (ddapc_restraint_control%coeff)
         ALLOCATE (ddapc_restraint_control%coeff(ddapc_restraint_control%natoms))
         ddapc_restraint_control%coeff = 1.0_dp

         CALL section_vals_val_get(ddapc_section, "COEFF", i_rep_section=i, &
                                   n_rep_val=n_rep)
         jj = 0
         DO k = 1, n_rep
            CALL section_vals_val_get(ddapc_section, "COEFF", i_rep_section=i, &
                                      i_rep_val=k, r_vals=rtmplist)
            DO j = 1, SIZE(rtmplist)
               jj = jj+1
               IF (jj > ddapc_restraint_control%natoms) &
                  CALL cp__b("cp_control_utils.F",1543,"Need the same number of coeff as there are atoms ")
               ddapc_restraint_control%coeff(jj) = rtmplist(j)
            END DO
         END DO
         IF (jj < ddapc_restraint_control%natoms .AND. jj .NE. 0) &
            CALL cp__b("cp_control_utils.F",1548,"Need no or the same number of coeff as there are atoms.")
      END DO
      k = 0
      DO i = 1, SIZE(qs_control%ddapc_restraint_control)
         IF (qs_control%ddapc_restraint_control(i)%ddapc_restraint_control%functional_form == &
             do_ddapc_constraint) k = k+1
      END DO
      IF (k == 2) CALL cp_abort(cp__l("cp_control_utils.F",1555), &
                                "Only a single constraint possible yet, try to use restraints instead ")

   END SUBROUTINE read_ddapc_section

! **************************************************************************************************
!> \brief reads the input parameters needed for evaluating a becke weight population constraint
!> \param qs_control ...
!> \param becke_section ...
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE read_becke_section(qs_control, becke_section)

      TYPE(qs_control_type), INTENT(INOUT)               :: qs_control
      TYPE(section_vals_type), POINTER                   :: becke_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_becke_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: j, jj, k, n_rep
      INTEGER, DIMENSION(:), POINTER                     :: tmplist
      REAL(KIND=dp), DIMENSION(:), POINTER               :: rtmplist

      CALL section_vals_val_get(becke_section, "STRENGTH", &
                                r_val=qs_control%becke_control%strength)
      CALL section_vals_val_get(becke_section, "TARGET", &
                                r_val=qs_control%becke_control%target)
      CALL section_vals_val_get(becke_section, "TYPE_OF_DENSITY", &
                                i_val=qs_control%becke_control%density_type)
      CALL section_vals_val_get(becke_section, "ATOMS", &
                                n_rep_val=n_rep)
      jj = 0
      DO k = 1, n_rep
         CALL section_vals_val_get(becke_section, "ATOMS", i_rep_val=k, i_vals=tmplist)
         DO j = 1, SIZE(tmplist)
            jj = jj+1
         END DO
      END DO
      IF (jj < 1) CALL cp__b("cp_control_utils.F",1593,"Need at least 1 atom to use ddapc contraints.")
      qs_control%becke_control%natoms = jj
      IF (ASSOCIATED(qs_control%becke_control%atoms)) DEALLOCATE (qs_control%becke_control%atoms)
      ALLOCATE (qs_control%becke_control%atoms(qs_control%becke_control%natoms))
      jj = 0
      DO k = 1, n_rep
         CALL section_vals_val_get(becke_section, "ATOMS", i_rep_val=k, i_vals=tmplist)
         DO j = 1, SIZE(tmplist)
            jj = jj+1
            qs_control%becke_control%atoms(jj) = tmplist(j)
         END DO
      END DO

      IF (ASSOCIATED(qs_control%becke_control%coeff)) &
         DEALLOCATE (qs_control%becke_control%coeff)
      ALLOCATE (qs_control%becke_control%coeff(qs_control%becke_control%natoms))
      qs_control%becke_control%coeff = 1.0_dp

      CALL section_vals_val_get(becke_section, "COEFF", &
                                n_rep_val=n_rep)
      jj = 0
      DO k = 1, n_rep
         CALL section_vals_val_get(becke_section, "COEFF", i_rep_val=k, r_vals=rtmplist)
         DO j = 1, SIZE(rtmplist)
            jj = jj+1
            IF (jj > qs_control%becke_control%natoms) &
               CALL cp__b("cp_control_utils.F",1619,"Need the same number of coeff as there are atoms.")
            qs_control%becke_control%coeff(jj) = rtmplist(j)
         END DO
      END DO
      IF (jj < qs_control%becke_control%natoms .AND. jj .NE. 0) &
         CALL cp__b("cp_control_utils.F",1624,"Need no or the same number of coeff as there are atoms.")
   END SUBROUTINE read_becke_section

! **************************************************************************************************
!> \brief ...
!> \param dft_control ...
!> \param efield_section ...
! **************************************************************************************************
   SUBROUTINE read_efield_sections(dft_control, efield_section)
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(section_vals_type), POINTER                   :: efield_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_efield_sections', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_path_length)                 :: file_name
      INTEGER                                            :: i, io, j, n, unit_nr
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tmp_vals
      TYPE(efield_type), POINTER                         :: efield
      TYPE(section_vals_type), POINTER                   :: tmp_section

      DO i = 1, SIZE(dft_control%efield_fields)
         NULLIFY (dft_control%efield_fields(i)%efield)
         ALLOCATE (dft_control%efield_fields(i)%efield)
         efield => dft_control%efield_fields(i)%efield
         NULLIFY (efield%envelop_i_vars, efield%envelop_r_vars)
         CALL section_vals_val_get(efield_section, "INTENSITY", i_rep_section=i, &
                                   r_val=efield%strength)

         CALL section_vals_val_get(efield_section, "POLARISATION", i_rep_section=i, &
                                   r_vals=tmp_vals)
         ALLOCATE (efield%polarisation(SIZE(tmp_vals)))
         efield%polarisation = tmp_vals
         CALL section_vals_val_get(efield_section, "PHASE", i_rep_section=i, &
                                   r_val=efield%phase_offset)
         CALL section_vals_val_get(efield_section, "ENVELOP", i_rep_section=i, &
                                   i_val=efield%envelop_id)
         CALL section_vals_val_get(efield_section, "WAVELENGTH", i_rep_section=i, &
                                   r_val=efield%wavelength)

         IF (efield%envelop_id == constant_env) THEN
            ALLOCATE (efield%envelop_i_vars(2))
            tmp_section => section_vals_get_subs_vals(efield_section, "CONSTANT_ENV", i_rep_section=i)
            CALL section_vals_val_get(tmp_section, "START_STEP", &
                                      i_val=efield%envelop_i_vars(1))
            CALL section_vals_val_get(tmp_section, "END_STEP", &
                                      i_val=efield%envelop_i_vars(2))
         ELSE IF (efield%envelop_id == gaussian_env) THEN
            ALLOCATE (efield%envelop_r_vars(2))
            tmp_section => section_vals_get_subs_vals(efield_section, "GAUSSIAN_ENV", i_rep_section=i)
            CALL section_vals_val_get(tmp_section, "T0", &
                                      r_val=efield%envelop_r_vars(1))
            CALL section_vals_val_get(tmp_section, "SIGMA", &
                                      r_val=efield%envelop_r_vars(2))
         ELSE IF (efield%envelop_id == ramp_env) THEN
            ALLOCATE (efield%envelop_i_vars(4))
            tmp_section => section_vals_get_subs_vals(efield_section, "RAMP_ENV", i_rep_section=i)
            CALL section_vals_val_get(tmp_section, "START_STEP_IN", &
                                      i_val=efield%envelop_i_vars(1))
            CALL section_vals_val_get(tmp_section, "END_STEP_IN", &
                                      i_val=efield%envelop_i_vars(2))
            CALL section_vals_val_get(tmp_section, "START_STEP_OUT", &
                                      i_val=efield%envelop_i_vars(3))
            CALL section_vals_val_get(tmp_section, "END_STEP_OUT", &
                                      i_val=efield%envelop_i_vars(4))
         ELSE IF (efield%envelop_id == custom_env) THEN
            tmp_section => section_vals_get_subs_vals(efield_section, "CUSTOM_ENV", i_rep_section=i)
            CALL section_vals_val_get(tmp_section, "EFIELD_FILE_NAME", c_val=file_name)
            CALL open_file(file_name=TRIM(file_name), file_action="READ", file_status="OLD", unit_number=unit_nr)
            !Determine the number of lines in file
            n = 0
            DO WHILE (.TRUE.)
               READ (unit_nr, *, iostat=io)
               IF (io /= 0) EXIT
               n = n+1
            END DO
            REWIND (unit_nr)
            ALLOCATE (efield%envelop_r_vars(n+1))
            !Store the timestep of the list in the first entry of the r_vars
            CALL section_vals_val_get(tmp_section, "TIMESTEP", r_val=efield%envelop_r_vars(1))
            !Read the file
            DO j = 2, n+1
               READ (unit_nr, *) efield%envelop_r_vars(j)
               efield%envelop_r_vars(j) = cp_unit_to_cp2k(efield%envelop_r_vars(j), "volt/m")
            END DO
            CALL close_file(unit_nr)
         END IF
      END DO
   END SUBROUTINE read_efield_sections

! **************************************************************************************************
!> \brief reads the input parameters needed real time propagation
!> \param dft_control ...
!> \param rtp_section ...
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE read_rtp_section(dft_control, rtp_section)

      TYPE(dft_control_type), INTENT(INOUT)              :: dft_control
      TYPE(section_vals_type), POINTER                   :: rtp_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_rtp_section', &
         routineP = moduleN//':'//routineN

      INTEGER, DIMENSION(:), POINTER                     :: tmp

      ALLOCATE (dft_control%rtp_control)
      CALL section_vals_val_get(rtp_section, "MAX_ITER", &
                                i_val=dft_control%rtp_control%max_iter)
      CALL section_vals_val_get(rtp_section, "MAT_EXP", &
                                i_val=dft_control%rtp_control%mat_exp)
      CALL section_vals_val_get(rtp_section, "ASPC_ORDER", &
                                i_val=dft_control%rtp_control%aspc_order)
      CALL section_vals_val_get(rtp_section, "EXP_ACCURACY", &
                                r_val=dft_control%rtp_control%eps_exp)
      CALL section_vals_val_get(rtp_section, "PROPAGATOR", &
                                i_val=dft_control%rtp_control%propagator)
      CALL section_vals_val_get(rtp_section, "EPS_ITER", &
                                r_val=dft_control%rtp_control%eps_ener)
      CALL section_vals_val_get(rtp_section, "INITIAL_WFN", &
                                i_val=dft_control%rtp_control%initial_wfn)
      CALL section_vals_val_get(rtp_section, "HFX_BALANCE_IN_CORE", &
                                l_val=dft_control%rtp_control%hfx_redistribute)
      CALL section_vals_val_get(rtp_section, "APPLY_DELTA_PULSE", &
                                l_val=dft_control%rtp_control%apply_delta_pulse)
      CALL section_vals_val_get(rtp_section, "PERIODIC", &
                                l_val=dft_control%rtp_control%periodic)
      CALL section_vals_val_get(rtp_section, "DENSITY_PROPAGATION", &
                                l_val=dft_control%rtp_control%linear_scaling)
      CALL section_vals_val_get(rtp_section, "ORTHONORMAL", &
                                l_val=dft_control%rtp_control%orthonormal)
      CALL section_vals_val_get(rtp_section, "MCWEENY_MAX_ITER", &
                                i_val=dft_control%rtp_control%mcweeny_max_iter)
      CALL section_vals_val_get(rtp_section, "ACCURACY_REFINEMENT", &
                                i_val=dft_control%rtp_control%acc_ref)
      CALL section_vals_val_get(rtp_section, "MCWEENY_EPS", &
                                r_val=dft_control%rtp_control%mcweeny_eps)
      CALL section_vals_val_get(rtp_section, "DELTA_PULSE_SCALE", &
                                r_val=dft_control%rtp_control%delta_pulse_scale)
      CALL section_vals_val_get(rtp_section, "DELTA_PULSE_DIRECTION", &
                                i_vals=tmp)
      dft_control%rtp_control%delta_pulse_direction = tmp
      CALL section_vals_val_get(rtp_section, "SC_CHECK_START", &
                                i_val=dft_control%rtp_control%sc_check_start)

   END SUBROUTINE read_rtp_section

! **************************************************************************************************
!> \brief Parses the BLOCK_LIST keywords from the ADMM section
!> \param admm_control ...
!> \param dft_section ...
! **************************************************************************************************
   SUBROUTINE read_admm_block_list(admm_control, dft_section)
      TYPE(admm_control_type), POINTER                   :: admm_control
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'read_admm_block_list', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: irep, list_size, n_rep
      INTEGER, DIMENSION(:), POINTER                     :: tmplist

      NULLIFY (tmplist)

      CALL section_vals_val_get(dft_section, "AUXILIARY_DENSITY_MATRIX_METHOD%BLOCK_LIST", &
                                n_rep_val=n_rep)

      ALLOCATE (admm_control%blocks(n_rep))

      DO irep = 1, n_rep
         CALL section_vals_val_get(dft_section, "AUXILIARY_DENSITY_MATRIX_METHOD%BLOCK_LIST", &
                                   i_rep_val=irep, i_vals=tmplist)
         list_size = SIZE(tmplist)
         ALLOCATE (admm_control%blocks(irep)%list(list_size))
         admm_control%blocks(irep)%list(:) = tmplist(:)
      END DO

   END SUBROUTINE read_admm_block_list

END MODULE cp_control_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Generate Gaussian cube files
! **************************************************************************************************
MODULE realspace_grid_cube
   USE cp_files,                        ONLY: close_file,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_maxloc,&
                                              mp_recv,&
                                              mp_send,&
                                              mp_sum,&
                                              mp_sync
   USE pw_grid_types,                   ONLY: PW_MODE_LOCAL
   USE pw_types,                        ONLY: pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: pw_to_cube, cube_to_pw

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'realspace_grid_cube'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param pw ...
!> \param unit_nr ...
!> \param title ...
!> \param particles_r ...
!> \param particles_z ...
!> \param stride ...
!> \param zero_tails ...
! **************************************************************************************************
   SUBROUTINE pw_to_cube(pw, unit_nr, title, particles_r, particles_z, stride, zero_tails)
      TYPE(pw_type), POINTER                             :: pw
      INTEGER, INTENT(IN)                                :: unit_nr
      CHARACTER(*), INTENT(IN), OPTIONAL                 :: title
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN), &
         OPTIONAL                                        :: particles_r
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: particles_z
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: stride
      LOGICAL, INTENT(IN), OPTIONAL                      :: zero_tails

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_to_cube', routineP = moduleN//':'//routineN

      INTEGER                                            :: checksum, dest, gid, handle, i, I1, I2, &
                                                            I3, iat, ip, L1, L2, L3, my_rank, &
                                                            my_stride(3), np, num_pe, rank(2), &
                                                            source, tag, U1, U2, U3
      LOGICAL                                            :: my_zero_tails
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: buf

      CALL timeset(routineN, handle)

      my_zero_tails = .FALSE.
      IF (PRESENT(zero_tails)) my_zero_tails = zero_tails
      my_stride = 1
      IF (PRESENT(stride)) THEN
         IF (SIZE(stride) /= 1 .AND. SIZE(stride) /= 3) &
            CALL cp_abort(cp__l("pw/realspace_grid_cube.F",69), "STRIDE keyword can accept only 1 "// &
                          "(the same for X,Y,Z) or 3 values. Correct your input file.")
         IF (SIZE(stride) == 1) THEN
            DO i = 1, 3
               my_stride(i) = stride(1)
            END DO
         ELSE
            my_stride = stride(1:3)
         END IF
         IF(.NOT.(my_stride(1) > 0))CALL cp__a("pw/realspace_grid_cube.F",78)
         IF(.NOT.(my_stride(2) > 0))CALL cp__a("pw/realspace_grid_cube.F",79)
         IF(.NOT.(my_stride(3) > 0))CALL cp__a("pw/realspace_grid_cube.F",80)
      END IF

      IF (unit_nr > 0) THEN
         ! this format seems to work for e.g. molekel and gOpenmol
         ! latest version of VMD can read non orthorhombic cells
         WRITE (unit_nr, '(a11)') "-Quickstep-"
         IF (PRESENT(title)) THEN
            WRITE (unit_nr, *) TRIM(title)
         ELSE
            WRITE (unit_nr, *) "No Title"
         ENDIF

         IF(.NOT.(PRESENT(particles_z) .EQV. PRESENT(particles_r)))CALL cp__a("pw/realspace_grid_cube.F",93)
         np = 0
         IF (PRESENT(particles_z)) THEN
            IF(.NOT.(SIZE(particles_z) == SIZE(particles_r, dim=2)))CALL cp__a("pw/realspace_grid_cube.F",96)
            ! cube files can only be written for 99999 particles due to a format limitation (I5)
            ! so we limit the number of particles written.
            np = MIN(99999, SIZE(particles_z))
         END IF

         WRITE (unit_nr, '(I5,3f12.6)') np, 0.0_dp, 0._dp, 0._dp !start of cube

         WRITE (unit_nr, '(I5,3f12.6)') (pw%pw_grid%npts(1)+my_stride(1)-1)/my_stride(1), &
            pw%pw_grid%dh(1, 1)*REAL(my_stride(1), dp), pw%pw_grid%dh(2, 1)*REAL(my_stride(1), dp), &
            pw%pw_grid%dh(3, 1)*REAL(my_stride(1), dp)
         WRITE (unit_nr, '(I5,3f12.6)') (pw%pw_grid%npts(2)+my_stride(2)-1)/my_stride(2), &
            pw%pw_grid%dh(1, 2)*REAL(my_stride(2), dp), pw%pw_grid%dh(2, 2)*REAL(my_stride(2), dp), &
            pw%pw_grid%dh(3, 2)*REAL(my_stride(2), dp)
         WRITE (unit_nr, '(I5,3f12.6)') (pw%pw_grid%npts(3)+my_stride(3)-1)/my_stride(3), &
            pw%pw_grid%dh(1, 3)*REAL(my_stride(3), dp), pw%pw_grid%dh(2, 3)*REAL(my_stride(3), dp), &
            pw%pw_grid%dh(3, 3)*REAL(my_stride(3), dp)

         IF (PRESENT(particles_z)) THEN
            DO iat = 1, np
               WRITE (unit_nr, '(I5,4f12.6)') particles_z(iat), 0._dp, particles_r(:, iat)
            END DO
         END IF
      ENDIF

      ! shortcut
      L1 = pw%pw_grid%bounds(1, 1)
      L2 = pw%pw_grid%bounds(1, 2)
      L3 = pw%pw_grid%bounds(1, 3)
      U1 = pw%pw_grid%bounds(2, 1)
      U2 = pw%pw_grid%bounds(2, 2)
      U3 = pw%pw_grid%bounds(2, 3)

      ALLOCATE (buf(L3:U3))

      my_rank = pw%pw_grid%para%my_pos
      gid = pw%pw_grid%para%group
      num_pe = pw%pw_grid%para%group_size
      tag = 1

      rank(1) = unit_nr
      rank(2) = my_rank
      checksum = 0
      IF (unit_nr > 0) checksum = 1

      CALL mp_sum(checksum, gid)
      IF(.NOT.(checksum == 1))CALL cp__a("pw/realspace_grid_cube.F",142)

      CALL mp_maxloc(rank, gid)
      IF(.NOT.(rank(1) > 0))CALL cp__a("pw/realspace_grid_cube.F",145)

      dest = rank(2)
      DO I1 = L1, U1, my_stride(1)
         DO I2 = L2, U2, my_stride(2)

            ! cycling through the CPUs, check if the current ray (I1,I2) is local to that CPU
            IF (pw%pw_grid%para%mode .NE. PW_MODE_LOCAL) THEN
               DO ip = 0, num_pe-1
                  IF (pw%pw_grid%para%bo(1, 1, ip, 1) <= I1-L1+1 .AND. pw%pw_grid%para%bo(2, 1, ip, 1) >= I1-L1+1 .AND. &
                      pw%pw_grid%para%bo(1, 2, ip, 1) <= I2-L2+1 .AND. pw%pw_grid%para%bo(2, 2, ip, 1) >= I2-L2+1) THEN
                     source = ip
                  ENDIF
               ENDDO
            ELSE
               source = dest
            ENDIF

            IF (source == dest) THEN
               IF (my_rank == source) THEN
                  buf(:) = pw%cr3d(I1, I2, :)
               ENDIF
            ELSE
               IF (my_rank == source) THEN
                  buf(:) = pw%cr3d(I1, I2, :)
                  CALL mp_send(buf, dest, tag, gid)
               ENDIF
               IF (my_rank == dest) THEN
                  CALL mp_recv(buf, source, tag, gid)
               ENDIF
            ENDIF

            IF (unit_nr > 0) THEN
               IF (my_zero_tails) THEN
                  DO I3 = L3, U3
                     IF (buf(I3) < 1.E-7_dp) buf(I3) = 0.0_dp
                  END DO
               END IF
               WRITE (unit_nr, '(6E13.5)') (buf(I3), I3=L3, U3, my_stride(3))
            ENDIF

            ! this double loop generates so many messages that it can overload
            ! the message passing system, e.g. on XT3
            ! we therefore put a barrier here that limits the amount of message
            ! that flies around at any given time.
            ! if ever this routine becomes a bottleneck, we should go for a
            ! more complicated rewrite
            CALL mp_sync(gid)

         ENDDO
      ENDDO

      DEALLOCATE (buf)

      CALL timestop(handle)

   END SUBROUTINE pw_to_cube

! **************************************************************************************************
!> \brief  Computes the external density on the grid
!>         hacked from external_read_density
!> \param grid     pw to read from cube file
!> \param filename name of cube file
!> \param scaling  scale values before storing
!> \date   01.2014
!> \author M.Watkins
! **************************************************************************************************
   SUBROUTINE cube_to_pw(grid, filename, scaling)

      TYPE(pw_type), POINTER                             :: grid
      CHARACTER(len=*), INTENT(in)                       :: filename
      REAL(kind=dp), INTENT(in)                          :: scaling

      CHARACTER(len=*), PARAMETER :: routineN = 'cube_to_pw', routineP = moduleN//':'//routineN

      INTEGER                                            :: extunit, gid, handle, i, ip, j, k, &
                                                            master, my_rank, nat, ndum, num_pe, &
                                                            output_unit, tag
      INTEGER, DIMENSION(3)                              :: lbounds, lbounds_local, npoints, &
                                                            npoints_local, ubounds, ubounds_local
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: buffer
      REAL(kind=dp), DIMENSION(3)                        :: dr, rdum

      output_unit = cp_logger_get_default_io_unit()

      CALL timeset(routineN, handle)

      !get rs grids and parallel envnment
      gid = grid%pw_grid%para%group
      my_rank = grid%pw_grid%para%my_pos
      num_pe = grid%pw_grid%para%group_size
      tag = 1

      DO i = 1, 3
         dr(i) = grid%pw_grid%dh(i, i)
      END DO

      npoints = grid%pw_grid%npts
      lbounds = grid%pw_grid%bounds(1, :)
      ubounds = grid%pw_grid%bounds(2, :)

      npoints_local = grid%pw_grid%npts_local
      lbounds_local = grid%pw_grid%bounds_local(1, :)
      ubounds_local = grid%pw_grid%bounds_local(2, :)

      !pw grids at most pencils - all processors have a full set of z data for x,y
      ALLOCATE (buffer(lbounds(3):ubounds(3)))

      IF (my_rank == 0) THEN
         IF (output_unit > 0) THEN
            WRITE (output_unit, FMT="(/,(T3,A,T61,A30))") "Reading the cube file:     ", filename
         END IF

         CALL open_file(file_name=filename, &
                        file_status="OLD", &
                        file_form="FORMATTED", &
                        file_action="READ", &
                        unit_number=extunit)

         !skip header comments
         DO i = 1, 2
            READ (extunit, *)
         ENDDO
         READ (extunit, *) nat, rdum
         DO i = 1, 3
            READ (extunit, *) ndum, rdum
            IF ((ndum /= npoints(i) .OR. (ABS(rdum(i)-dr(i)) > 1e-4)) .AND. &
                output_unit > 0) THEN
               WRITE (output_unit, *) "Restart from density | ERROR! | CUBE FILE NOT COINCIDENT WITH INTERNAL GRID ", i
               WRITE (output_unit, *) "Restart from density | ", ndum, " DIFFERS FROM ", npoints(i)
               WRITE (output_unit, *) "Restart from density | ", rdum, " DIFFERS FROM ", dr(i)
            ENDIF
         ENDDO
         !ignore atomic postion data - read from coord or topology instead
         DO i = 1, nat
            READ (extunit, *)
         ENDDO
      ENDIF

      !master sends all data to everyone
      DO i = lbounds(1), ubounds(1)
         DO j = lbounds(2), ubounds(2)
            IF (my_rank .EQ. 0) THEN
               READ (extunit, *) (buffer(k), k=lbounds(3), ubounds(3))
               IF (num_pe .GT. 1) THEN
                  DO ip = 1, num_pe-1
                     CALL mp_send(buffer(lbounds(3):ubounds(3)), ip, tag, gid)
                  ENDDO
               ENDIF
            ELSE
               master = 0
               CALL mp_recv(buffer(lbounds(3):ubounds(3)), master, tag, gid)
            ENDIF
            CALL mp_sync(gid)

            !only use data that is local to me - i.e. in slice of pencil I own
            IF ((lbounds_local(1) .LE. i) .AND. (i .LE. ubounds_local(1)) .AND. (lbounds_local(2) .LE. j) &
                .AND. (j .LE. ubounds_local(2))) THEN
               !allow scaling of external potential values by factor 'scaling' (SCALING_FACTOR in input file)
               grid%cr3d(i, j, lbounds(3):ubounds(3)) = buffer(lbounds(3):ubounds(3))*scaling
            ENDIF

         ENDDO
      ENDDO

      IF (my_rank == 0) CALL close_file(unit_number=extunit)

      CALL mp_sync(gid)

      CALL timestop(handle)

   END SUBROUTINE cube_to_pw

END MODULE realspace_grid_cube
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A wrapper around pw_to_cube() which accepts particle_list_type
!> \author Ole Schuett
! **************************************************************************************************
MODULE cp_realspace_grid_cube
   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE kinds,                           ONLY: dp
   USE particle_list_types,             ONLY: particle_list_type
   USE pw_types,                        ONLY: pw_type
   USE realspace_grid_cube,             ONLY: pw_to_cube
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC :: cp_pw_to_cube

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_realspace_grid_cube'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param pw ...
!> \param unit_nr ...
!> \param title ...
!> \param particles ...
!> \param stride ...
!> \param zero_tails ...
! **************************************************************************************************
   SUBROUTINE cp_pw_to_cube(pw, unit_nr, title, particles, stride, zero_tails)
      TYPE(pw_type), POINTER                             :: pw
      INTEGER, INTENT(IN)                                :: unit_nr
      CHARACTER(*), INTENT(IN), OPTIONAL                 :: title
      TYPE(particle_list_type), POINTER                  :: particles
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: stride
      LOGICAL, INTENT(IN), OPTIONAL                      :: zero_tails

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_pw_to_cube', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: particles_z
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: particles_r
      TYPE(particle_list_type), POINTER                  :: my_particles

      NULLIFY (my_particles)
      my_particles => particles
      IF (ASSOCIATED(my_particles)) THEN
         n = my_particles%n_els
         ALLOCATE (particles_z(n))
         ALLOCATE (particles_r(3, n))
         DO i = 1, n
            CALL get_atomic_kind(my_particles%els(i)%atomic_kind, z=particles_z(i))
            particles_r(:, i) = my_particles%els(i)%r(:)
         END DO

         CALL pw_to_cube(pw=pw, unit_nr=unit_nr, title=title, &
                         particles_z=particles_z, particles_r=particles_r, &
                         stride=stride, zero_tails=zero_tails)
      ELSE
         CALL pw_to_cube(pw=pw, unit_nr=unit_nr, title=title, &
                         stride=stride, zero_tails=zero_tails)
      END IF

   END SUBROUTINE cp_pw_to_cube

END MODULE cp_realspace_grid_cube
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief groups fairly general SCF methods, so that modules other than qs_scf can use them too
!>        split off from qs_scf to reduce dependencies
!> \par History
!>      - Joost VandeVondele (03.2006)
!>      - combine_ks_matrices added (05.04.06,MK)
!>      - second ROKS scheme added (15.04.06,MK)
!>      - MO occupation management moved (29.08.2008,MK)
!>      - correct_mo_eigenvalues was moved from qs_mo_types;
!>        new subroutine shift_unocc_mos (03.2016, Sergey Chulkov)
! **************************************************************************************************
MODULE qs_scf_methods

   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_desymmetrize, cp_dbcsr_get_block_p, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale,&
                                              cp_fm_symm,&
                                              cp_fm_triangular_multiply,&
                                              cp_fm_upper_to_full
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_reduce,&
                                              cp_fm_cholesky_restore
   USE cp_fm_diag,                      ONLY: choose_eigv_solver,&
                                              cp_fm_block_jacobi
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_equivalent,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: cholesky_inverse,&
                                              cholesky_off,&
                                              cholesky_reduce,&
                                              cholesky_restore
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_max
   USE qs_density_mixing_types,         ONLY: mixing_storage_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_methods'
   REAL(KIND=dp), PARAMETER    :: ratio = 0.25_dp

   PUBLIC :: combine_ks_matrices, &
             cp_sm_mix, &
             eigensolver, &
             eigensolver_dbcsr, &
             eigensolver_symm, &
             eigensolver_simple, &
             scf_env_density_mixing

   PRIVATE :: correct_mo_eigenvalues, shift_unocc_mos

   INTERFACE combine_ks_matrices
      MODULE PROCEDURE combine_ks_matrices_1, &
         combine_ks_matrices_2
   END INTERFACE combine_ks_matrices

CONTAINS

! **************************************************************************************************
!> \brief perform (if requested) a density mixing
!> \param p_mix_new    New density matrices
!> \param mixing_store ...
!> \param rho_ao       Density environment
!> \param para_env ...
!> \param iter_delta ...
!> \param iter_count ...
!> \param diis ...
!> \param invert       Invert mixing
!> \par History
!>      02.2003 created [fawzi]
!>      08.2014 adapted for kpoints [JGH]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE scf_env_density_mixing(p_mix_new, mixing_store, rho_ao, para_env, &
                                     iter_delta, iter_count, diis, invert)
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: p_mix_new
      TYPE(mixing_storage_type), POINTER                 :: mixing_store
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: rho_ao
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), INTENT(INOUT)                       :: iter_delta
      INTEGER, INTENT(IN)                                :: iter_count
      LOGICAL, INTENT(in), OPTIONAL                      :: diis, invert

      CHARACTER(len=*), PARAMETER :: routineN = 'scf_env_density_mixing', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ic, ispin
      LOGICAL                                            :: my_diis, my_invert
      REAL(KIND=dp)                                      :: my_p_mix, tmp

      CALL timeset(routineN, handle)

      my_diis = .FALSE.
      IF (PRESENT(diis)) my_diis = diis
      my_invert = .FALSE.
      IF (PRESENT(invert)) my_invert = invert
      my_p_mix = mixing_store%alpha
      IF (my_diis .OR. iter_count < mixing_store%nskip_mixing) THEN
         my_p_mix = 1.0_dp
      END IF

      iter_delta = 0.0_dp
      IF(.NOT.(ASSOCIATED(p_mix_new)))CALL cp__a("qs_scf_methods.F",123)
      DO ic = 1, SIZE(p_mix_new, 2)
         DO ispin = 1, SIZE(p_mix_new, 1)
            IF (my_invert) THEN
               IF(.NOT.(my_p_mix /= 0.0_dp))CALL cp__a("qs_scf_methods.F",127)
               IF (my_p_mix /= 1.0_dp) THEN
                  CALL cp_dbcsr_add(matrix_a=p_mix_new(ispin, ic)%matrix, &
                                    alpha_scalar=1.0_dp/my_p_mix, &
                                    matrix_b=rho_ao(ispin, ic)%matrix, &
                                    beta_scalar=(my_p_mix-1.0_dp)/my_p_mix)
               END IF
            ELSE
               CALL cp_sm_mix(m1=p_mix_new(ispin, ic)%matrix, &
                              m2=rho_ao(ispin, ic)%matrix, &
                              p_mix=my_p_mix, &
                              delta=tmp, &
                              para_env=para_env)
               iter_delta = MAX(iter_delta, tmp)
            END IF
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE scf_env_density_mixing

! **************************************************************************************************
!> \brief   Diagonalise the Kohn-Sham matrix to get a new set of MO eigen-
!>          vectors and MO eigenvalues. ks will be modified
!> \param matrix_ks_fm ...
!> \param mo_set ...
!> \param ortho ...
!> \param work ...
!> \param cholesky_method ...
!> \param do_level_shift activate the level shifting technique
!> \param level_shift    amount of shift applied (in a.u.)
!> \param matrix_u_fm    matrix U : S (overlap matrix) = U^T * U
!> \param use_jacobi ...
!> \date    01.05.2001
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE eigensolver(matrix_ks_fm, mo_set, ortho, work, &
                          cholesky_method, do_level_shift, &
                          level_shift, matrix_u_fm, use_jacobi)
      TYPE(cp_fm_type), POINTER                          :: matrix_ks_fm
      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_fm_type), POINTER                          :: ortho, work
      INTEGER, INTENT(inout)                             :: cholesky_method
      LOGICAL, INTENT(in)                                :: do_level_shift
      REAL(KIND=dp), INTENT(in)                          :: level_shift
      TYPE(cp_fm_type), POINTER                          :: matrix_u_fm
      LOGICAL, INTENT(in)                                :: use_jacobi

      CHARACTER(len=*), PARAMETER :: routineN = 'eigensolver', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, homo, nao, nmo
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues
      TYPE(cp_fm_type), POINTER                          :: mo_coeff

      CALL timeset(routineN, handle)

      NULLIFY (mo_coeff)
      NULLIFY (mo_eigenvalues)

      ! Diagonalise the Kohn-Sham matrix

      CALL get_mo_set(mo_set=mo_set, &
                      nao=nao, &
                      nmo=nmo, &
                      homo=homo, &
                      eigenvalues=mo_eigenvalues, &
                      mo_coeff=mo_coeff)

      SELECT CASE (cholesky_method)
      CASE (cholesky_reduce)
         CALL cp_fm_cholesky_reduce(matrix_ks_fm, ortho)

         IF (do_level_shift) &
            CALL shift_unocc_mos(matrix_ks_fm=matrix_ks_fm, mo_coeff=mo_coeff, homo=homo, nmo=nmo, nao=nao, &
                                 level_shift=level_shift, is_triangular=.TRUE., matrix_u_fm=matrix_u_fm)

         CALL choose_eigv_solver(matrix_ks_fm, work, mo_eigenvalues)
         CALL cp_fm_cholesky_restore(work, nmo, ortho, mo_coeff, "SOLVE")
         IF (do_level_shift) &
            CALL correct_mo_eigenvalues(mo_eigenvalues, homo, nmo, level_shift)

      CASE (cholesky_restore)
         CALL cp_fm_upper_to_full(matrix_ks_fm, work)
         CALL cp_fm_cholesky_restore(matrix_ks_fm, nao, ortho, work, &
                                     "SOLVE", pos="RIGHT")
         CALL cp_fm_cholesky_restore(work, nao, ortho, matrix_ks_fm, &
                                     "SOLVE", pos="LEFT", transa="T")

         IF (do_level_shift) &
            CALL shift_unocc_mos(matrix_ks_fm=matrix_ks_fm, mo_coeff=mo_coeff, homo=homo, nmo=nmo, nao=nao, &
                                 level_shift=level_shift, is_triangular=.TRUE., matrix_u_fm=matrix_u_fm)

         CALL choose_eigv_solver(matrix_ks_fm, work, mo_eigenvalues)
         CALL cp_fm_cholesky_restore(work, nmo, ortho, mo_coeff, "SOLVE")

         IF (do_level_shift) &
            CALL correct_mo_eigenvalues(mo_eigenvalues, homo, nmo, level_shift)

      CASE (cholesky_inverse)
         CALL cp_fm_upper_to_full(matrix_ks_fm, work)

         CALL cp_fm_triangular_multiply(ortho, matrix_ks_fm, side="R", transpose_tr=.FALSE., &
                                        invert_tr=.FALSE., uplo_tr="U", n_rows=nao, n_cols=nao, alpha=1.0_dp)
         CALL cp_fm_triangular_multiply(ortho, matrix_ks_fm, side="L", transpose_tr=.TRUE., &
                                        invert_tr=.FALSE., uplo_tr="U", n_rows=nao, n_cols=nao, alpha=1.0_dp)

         IF (do_level_shift) &
            CALL shift_unocc_mos(matrix_ks_fm=matrix_ks_fm, mo_coeff=mo_coeff, homo=homo, nmo=nmo, nao=nao, &
                                 level_shift=level_shift, is_triangular=.TRUE., matrix_u_fm=matrix_u_fm)

         CALL choose_eigv_solver(matrix_ks_fm, work, mo_eigenvalues)
         CALL cp_fm_triangular_multiply(ortho, work, side="L", transpose_tr=.FALSE., &
                                        invert_tr=.FALSE., uplo_tr="U", n_rows=nao, n_cols=nmo, alpha=1.0_dp)
         CALL cp_fm_to_fm(work, mo_coeff, nmo, 1, 1)

         IF (do_level_shift) &
            CALL correct_mo_eigenvalues(mo_eigenvalues, homo, nmo, level_shift)

      END SELECT

      IF (use_jacobi) THEN
         CALL cp_fm_to_fm(mo_coeff, ortho)
         cholesky_method = cholesky_off
      END IF

      CALL timestop(handle)

   END SUBROUTINE eigensolver

! **************************************************************************************************
!> \brief ...
!> \param matrix_ks ...
!> \param matrix_ks_fm ...
!> \param mo_set ...
!> \param ortho_dbcsr ...
!> \param ksbuf1 ...
!> \param ksbuf2 ...
!> \param work ...
! **************************************************************************************************
   SUBROUTINE eigensolver_dbcsr(matrix_ks, matrix_ks_fm, mo_set, ortho_dbcsr, ksbuf1, ksbuf2, work)
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_ks
      TYPE(cp_fm_type), POINTER                          :: matrix_ks_fm
      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_dbcsr_type), POINTER                       :: ortho_dbcsr, ksbuf1, ksbuf2
      TYPE(cp_fm_type), POINTER                          :: work

      CHARACTER(len=*), PARAMETER :: routineN = 'eigensolver_dbcsr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nao, nmo
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues
      TYPE(cp_fm_type), POINTER                          :: mo_coeff

      CALL timeset(routineN, handle)

      NULLIFY (mo_coeff)
      NULLIFY (mo_eigenvalues)

      CALL get_mo_set(mo_set=mo_set, &
                      nao=nao, &
                      nmo=nmo, &
                      eigenvalues=mo_eigenvalues, &
                      mo_coeff=mo_coeff)

!    Reduce KS matrix
      CALL cp_dbcsr_desymmetrize(matrix_ks, ksbuf2)
      CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, ksbuf2, ortho_dbcsr, 0.0_dp, ksbuf1)
      CALL cp_dbcsr_multiply('T', 'N', 1.0_dp, ortho_dbcsr, ksbuf1, 0.0_dp, ksbuf2)

!    Solve the eigenvalue problem
      CALL copy_dbcsr_to_fm(ksbuf2, matrix_ks_fm)
      CALL choose_eigv_solver(matrix_ks_fm, work, mo_eigenvalues)

!    Restore the eigenvector of the general eig. problem
      CALL cp_dbcsr_sm_fm_multiply(ortho_dbcsr, work, mo_coeff, nmo)

      CALL timestop(handle)

   END SUBROUTINE eigensolver_dbcsr

! **************************************************************************************************
!> \brief ...
!> \param matrix_ks_fm ...
!> \param mo_set ...
!> \param ortho ...
!> \param work ...
!> \param do_level_shift activate the level shifting technique
!> \param level_shift    amount of shift applied (in a.u.)
!> \param matrix_u_fm    matrix U : S (overlap matrix) = U^T * U
!> \param use_jacobi ...
!> \param jacobi_threshold ...
! **************************************************************************************************
   SUBROUTINE eigensolver_symm(matrix_ks_fm, mo_set, ortho, work, do_level_shift, &
                               level_shift, matrix_u_fm, use_jacobi, jacobi_threshold)
      TYPE(cp_fm_type), POINTER                          :: matrix_ks_fm
      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_fm_type), POINTER                          :: ortho, work
      LOGICAL, INTENT(IN)                                :: do_level_shift
      REAL(KIND=dp), INTENT(IN)                          :: level_shift
      TYPE(cp_fm_type), POINTER                          :: matrix_u_fm
      LOGICAL, INTENT(IN)                                :: use_jacobi
      REAL(KIND=dp), INTENT(IN)                          :: jacobi_threshold

      CHARACTER(len=*), PARAMETER :: routineN = 'eigensolver_symm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, homo, nao, nelectron, nmo
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues
      TYPE(cp_fm_type), POINTER                          :: mo_coeff

      CALL timeset(routineN, handle)

      NULLIFY (mo_coeff)
      NULLIFY (mo_eigenvalues)

      ! Diagonalise the Kohn-Sham matrix

      CALL get_mo_set(mo_set=mo_set, &
                      nao=nao, &
                      nmo=nmo, &
                      homo=homo, &
                      nelectron=nelectron, &
                      eigenvalues=mo_eigenvalues, &
                      mo_coeff=mo_coeff)

      IF (use_jacobi) THEN

         CALL cp_fm_symm("L", "U", nao, homo, 1.0_dp, matrix_ks_fm, mo_coeff, 0.0_dp, work)
         CALL cp_gemm("T", "N", homo, nao-homo, nao, 1.0_dp, work, mo_coeff, &
                      0.0_dp, matrix_ks_fm, b_first_col=homo+1, c_first_col=homo+1)

         ! Block Jacobi (pseudo-diagonalization, only one sweep)
         CALL cp_fm_block_jacobi(matrix_ks_fm, mo_coeff, mo_eigenvalues, &
                                 jacobi_threshold, homo+1)

      ELSE ! full S^(-1/2) has been computed

         CALL cp_fm_symm("L", "U", nao, nao, 1.0_dp, matrix_ks_fm, ortho, 0.0_dp, work)
         CALL cp_gemm("T", "N", nao, nao, nao, 1.0_dp, ortho, work, 0.0_dp, matrix_ks_fm)

         IF (do_level_shift) &
            CALL shift_unocc_mos(matrix_ks_fm=matrix_ks_fm, mo_coeff=mo_coeff, homo=homo, nmo=nmo, nao=nao, &
                                 level_shift=level_shift, is_triangular=.FALSE., matrix_u_fm=matrix_u_fm)

         CALL choose_eigv_solver(matrix_ks_fm, work, mo_eigenvalues)

         CALL cp_gemm("N", "N", nao, nmo, nao, 1.0_dp, ortho, work, 0.0_dp, &
                      mo_coeff)

         IF (do_level_shift) &
            CALL correct_mo_eigenvalues(mo_eigenvalues, homo, nmo, level_shift)

      END IF

      CALL timestop(handle)

   END SUBROUTINE eigensolver_symm

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param matrix_ks ...
!> \param mo_set ...
!> \param work ...
!> \param do_level_shift activate the level shifting technique
!> \param level_shift    amount of shift applied (in a.u.)
!> \param use_jacobi ...
!> \param jacobi_threshold ...
! **************************************************************************************************
   SUBROUTINE eigensolver_simple(matrix_ks, mo_set, work, do_level_shift, &
                                 level_shift, use_jacobi, jacobi_threshold)

      TYPE(cp_fm_type), POINTER                          :: matrix_ks
      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_fm_type), POINTER                          :: work
      LOGICAL, INTENT(IN)                                :: do_level_shift
      REAL(KIND=dp), INTENT(IN)                          :: level_shift
      LOGICAL, INTENT(IN)                                :: use_jacobi
      REAL(KIND=dp), INTENT(IN)                          :: jacobi_threshold

      CHARACTER(len=*), PARAMETER :: routineN = 'eigensolver_simple', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, homo, nao, nelectron, nmo
      REAL(KIND=dp), DIMENSION(:), POINTER               :: mo_eigenvalues
      TYPE(cp_fm_type), POINTER                          :: mo_coeff

      CALL timeset(routineN, handle)

      NULLIFY (mo_coeff)
      NULLIFY (mo_eigenvalues)

      CALL get_mo_set(mo_set=mo_set, &
                      nao=nao, &
                      nmo=nmo, &
                      homo=homo, &
                      nelectron=nelectron, &
                      eigenvalues=mo_eigenvalues, &
                      mo_coeff=mo_coeff)

      IF (do_level_shift) THEN
         ! matrix_u_fm is simply an identity matrix, so we omit it here
         CALL shift_unocc_mos(matrix_ks_fm=matrix_ks, mo_coeff=mo_coeff, homo=homo, nmo=nmo, nao=nao, &
                              level_shift=level_shift, is_triangular=.FALSE.)
      END IF

      IF (use_jacobi) THEN
         CALL cp_fm_symm("L", "U", nao, homo, 1.0_dp, matrix_ks, mo_coeff, 0.0_dp, work)
         CALL cp_gemm("T", "N", homo, nao-homo, nao, 1.0_dp, work, mo_coeff, &
                      0.0_dp, matrix_ks, b_first_col=homo+1, c_first_col=homo+1)
         ! Block Jacobi (pseudo-diagonalization, only one sweep)
         CALL cp_fm_block_jacobi(matrix_ks, mo_coeff, mo_eigenvalues, jacobi_threshold, homo+1)
      ELSE

         CALL choose_eigv_solver(matrix_ks, work, mo_eigenvalues)

         CALL cp_fm_to_fm(work, mo_coeff, nmo, 1, 1)

      END IF

      IF (do_level_shift) &
         CALL correct_mo_eigenvalues(mo_eigenvalues, homo, nmo, level_shift)

      CALL timestop(handle)

   END SUBROUTINE eigensolver_simple

! **************************************************************************************************
!> \brief Perform a mixing of the given matrixes into the first matrix
!>      m1 = m2 + p_mix (m1-m2)
!> \param m1 first (new) matrix, is modified
!> \param m2 the second (old) matrix
!> \param p_mix how much m1 is conserved (0: none, 1: all)
!> \param delta maximum norm of m1-m2
!> \param para_env ...
!> \param m3 ...
!> \par History
!>      02.2003 rewamped [fawzi]
!> \author fawzi
!> \note
!>      if you what to store the result in m2 swap m1 and m2 an use
!>      (1-pmix) as pmix
!>      para_env should be removed (embedded in matrix)
! **************************************************************************************************
   SUBROUTINE cp_sm_mix(m1, m2, p_mix, delta, para_env, m3)

      TYPE(cp_dbcsr_type), POINTER                       :: m1, m2
      REAL(KIND=dp), INTENT(IN)                          :: p_mix
      REAL(KIND=dp), INTENT(OUT)                         :: delta
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: m3

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_sm_mix', routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, handle, i, iblock_col, iblock_row, j
      LOGICAL                                            :: found
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: p_delta_block, p_new_block, p_old_block
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL timeset(routineN, handle)
      delta = 0.0_dp

      CALL cp_dbcsr_iterator_start(iter, m1)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, p_new_block, blk)
         CALL cp_dbcsr_get_block_p(matrix=m2, row=iblock_row, col=iblock_col, &
                                   BLOCK=p_old_block, found=found)
         IF(.NOT.(ASSOCIATED(p_old_block)))CALL cp__a("qs_scf_methods.F",497)
         IF (PRESENT(m3)) THEN
            CALL cp_dbcsr_get_block_p(matrix=m3, row=iblock_row, col=iblock_col, &
                                      BLOCK=p_delta_block, found=found)
            IF(.NOT.(ASSOCIATED(p_delta_block)))CALL cp__a("qs_scf_methods.F",501)

            DO j = 1, SIZE(p_new_block, 2)
               DO i = 1, SIZE(p_new_block, 1)
                  p_delta_block(i, j) = p_new_block(i, j)-p_old_block(i, j)
                  delta = MAX(delta, ABS(p_delta_block(i, j)))
               END DO
            END DO
         ELSE
            DO j = 1, SIZE(p_new_block, 2)
               DO i = 1, SIZE(p_new_block, 1)
                  p_new_block(i, j) = p_new_block(i, j)-p_old_block(i, j)
                  delta = MAX(delta, ABS(p_new_block(i, j)))
                  p_new_block(i, j) = p_old_block(i, j)+p_mix*p_new_block(i, j)
               END DO
            END DO
         ENDIF
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL mp_max(delta, para_env%group)

      CALL timestop(handle)

   END SUBROUTINE cp_sm_mix

! **************************************************************************************************
!> \brief ...
!> \param ksa ...
!> \param ksb ...
!> \param occa ...
!> \param occb ...
!> \param roks_parameter ...
! **************************************************************************************************
   SUBROUTINE combine_ks_matrices_1(ksa, ksb, occa, occb, roks_parameter)

      ! Combine the alpha and beta Kohn-Sham matrices during a restricted open
      ! Kohn-Sham (ROKS) calculation
      ! On input ksa and ksb contain the alpha and beta Kohn-Sham matrices,
      ! respectively. occa and occb contain the corresponding MO occupation
      ! numbers. On output the combined ROKS operator matrix is returned in ksa.

      ! Literature: - C. C. J. Roothaan, Rev. Mod. Phys. 32, 179 (1960)
      !             - M. F. Guest and V. R. Saunders, Mol. Phys. 28(3), 819 (1974)

      TYPE(cp_fm_type), POINTER                          :: ksa, ksb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: occa, occb
      REAL(KIND=dp), DIMENSION(0:2, 0:2, 1:2), &
         INTENT(IN)                                      :: roks_parameter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'combine_ks_matrices_1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, icol_global, icol_local, &
                                                            irow_global, irow_local, j, &
                                                            ncol_local, nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: compatible_matrices
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: fa, fb
      TYPE(cp_fm_struct_type), POINTER                   :: ksa_struct, ksb_struct

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(ksa)))CALL cp__a("qs_scf_methods.F",566)
      IF(.NOT.(ASSOCIATED(ksb)))CALL cp__a("qs_scf_methods.F",567)

      CALL cp_fm_get_info(matrix=ksa, &
                          matrix_struct=ksa_struct, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices, &
                          local_data=fa)

      CALL cp_fm_get_info(matrix=ksb, &
                          matrix_struct=ksb_struct, &
                          local_data=fb)

      compatible_matrices = cp_fm_struct_equivalent(ksa_struct, ksb_struct)
      IF(.NOT.(compatible_matrices))CALL cp__a("qs_scf_methods.F",582)

      IF (SUM(occb) == 0.0_dp) fb = 0.0_dp

      DO icol_local = 1, ncol_local
         icol_global = col_indices(icol_local)
         j = INT(occa(icol_global))+INT(occb(icol_global))
         DO irow_local = 1, nrow_local
            irow_global = row_indices(irow_local)
            i = INT(occa(irow_global))+INT(occb(irow_global))
            fa(irow_local, icol_local) = &
               roks_parameter(i, j, 1)*fa(irow_local, icol_local)+ &
               roks_parameter(i, j, 2)*fb(irow_local, icol_local)
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE combine_ks_matrices_1

! **************************************************************************************************
!> \brief ...
!> \param ksa ...
!> \param ksb ...
!> \param occa ...
!> \param occb ...
!> \param f ...
!> \param nalpha ...
!> \param nbeta ...
! **************************************************************************************************
   SUBROUTINE combine_ks_matrices_2(ksa, ksb, occa, occb, f, nalpha, nbeta)

      ! Combine the alpha and beta Kohn-Sham matrices during a restricted open
      ! Kohn-Sham (ROKS) calculation
      ! On input ksa and ksb contain the alpha and beta Kohn-Sham matrices,
      ! respectively. occa and occb contain the corresponding MO occupation
      ! numbers. On output the combined ROKS operator matrix is returned in ksa.

      ! Literature: - C. C. J. Roothaan, Rev. Mod. Phys. 32, 179 (1960)
      !             - M. Filatov and S. Shaik, Chem. Phys. Lett. 288, 689 (1998)

      TYPE(cp_fm_type), POINTER                          :: ksa, ksb
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: occa, occb
      REAL(KIND=dp), INTENT(IN)                          :: f
      INTEGER, INTENT(IN)                                :: nalpha, nbeta

      CHARACTER(LEN=*), PARAMETER :: routineN = 'combine_ks_matrices_2', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, icol_global, icol_local, &
                                                            irow_global, irow_local, ncol_local, &
                                                            nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      LOGICAL                                            :: compatible_matrices
      REAL(KIND=dp)                                      :: beta, t1, t2, ta, tb
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: fa, fb
      TYPE(cp_fm_struct_type), POINTER                   :: ksa_struct, ksb_struct

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(ksa)))CALL cp__a("qs_scf_methods.F",644)
      IF(.NOT.(ASSOCIATED(ksb)))CALL cp__a("qs_scf_methods.F",645)

      CALL cp_fm_get_info(matrix=ksa, &
                          matrix_struct=ksa_struct, &
                          nrow_local=nrow_local, &
                          ncol_local=ncol_local, &
                          row_indices=row_indices, &
                          col_indices=col_indices, &
                          local_data=fa)

      CALL cp_fm_get_info(matrix=ksb, &
                          matrix_struct=ksb_struct, &
                          local_data=fb)

      compatible_matrices = cp_fm_struct_equivalent(ksa_struct, ksb_struct)
      IF(.NOT.(compatible_matrices))CALL cp__a("qs_scf_methods.F",660)

      beta = 1.0_dp/(1.0_dp-f)

      DO icol_local = 1, ncol_local

         icol_global = col_indices(icol_local)

         DO irow_local = 1, nrow_local

            irow_global = row_indices(irow_local)

            t1 = 0.5_dp*(fa(irow_local, icol_local)+fb(irow_local, icol_local))

            IF ((0 < irow_global) .AND. (irow_global <= nbeta)) THEN
               IF ((0 < icol_global) .AND. (icol_global <= nbeta)) THEN
                  ! closed-closed
                  fa(irow_local, icol_local) = t1
               ELSE IF ((nbeta < icol_global) .AND. (icol_global <= nalpha)) THEN
                  ! closed-open
                  ta = 0.5_dp*(f-REAL(occa(icol_global), KIND=dp))/f
                  tb = 0.5_dp*(f-REAL(occb(icol_global), KIND=dp))/f
                  t2 = ta*fa(irow_local, icol_local)+tb*fb(irow_local, icol_local)
                  fa(irow_local, icol_local) = t1+(beta-1.0_dp)*t2
               ELSE
                  ! closed-virtual
                  fa(irow_local, icol_local) = t1
               END IF
            ELSE IF ((nbeta < irow_global) .AND. (irow_global <= nalpha)) THEN
               IF ((0 < irow_global) .AND. (irow_global <= nbeta)) THEN
                  ! open-closed
                  ta = 0.5_dp*(f-REAL(occa(irow_global), KIND=dp))/f
                  tb = 0.5_dp*(f-REAL(occb(irow_global), KIND=dp))/f
                  t2 = ta*fa(irow_local, icol_local)+tb*fb(irow_local, icol_local)
                  fa(irow_local, icol_local) = t1+(beta-1.0_dp)*t2
               ELSE IF ((nbeta < icol_global) .AND. (icol_global <= nalpha)) THEN
                  ! open-open
                  ta = 0.5_dp*(f-REAL(occa(icol_global), KIND=dp))/f
                  tb = 0.5_dp*(f-REAL(occb(icol_global), KIND=dp))/f
                  t2 = ta*fa(irow_local, icol_local)+tb*fb(irow_local, icol_local)
                  IF (irow_global == icol_global) THEN
                     fa(irow_local, icol_local) = t1-t2
                  ELSE
                     fa(irow_local, icol_local) = t1-0.5_dp*t2
                  END IF
               ELSE
                  ! open-virtual
                  ta = 0.5_dp*(f-REAL(occa(irow_global), KIND=dp))/f
                  tb = 0.5_dp*(f-REAL(occb(irow_global), KIND=dp))/f
                  t2 = ta*fa(irow_local, icol_local)+tb*fb(irow_local, icol_local)
                  fa(irow_local, icol_local) = t1-t2
               END IF
            ELSE
               IF ((0 < irow_global) .AND. (irow_global < nbeta)) THEN
                  ! virtual-closed
                  fa(irow_local, icol_local) = t1
               ELSE IF ((nbeta < icol_global) .AND. (icol_global <= nalpha)) THEN
                  ! virtual-open
                  ta = 0.5_dp*(f-REAL(occa(icol_global), KIND=dp))/f
                  tb = 0.5_dp*(f-REAL(occb(icol_global), KIND=dp))/f
                  t2 = ta*fa(irow_local, icol_local)+tb*fb(irow_local, icol_local)
                  fa(irow_local, icol_local) = t1-t2
               ELSE
                  ! virtual-virtual
                  fa(irow_local, icol_local) = t1
               END IF
            END IF

         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE combine_ks_matrices_2

! **************************************************************************************************
!> \brief   Correct MO eigenvalues after MO level shifting.
!> \param mo_eigenvalues vector of eigenvalues
!> \param homo index of the highest occupied molecular orbital
!> \param nmo  number of molecular orbitals
!> \param level_shift amount of applied level shifting (in a.u.)
!> \date    19.04.2002
!> \par History
!>      - correct_mo_eigenvalues added (18.04.02,MK)
!>      - moved from module qs_mo_types, revised interface (03.2016, Sergey Chulkov)
!> \author  MK
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE correct_mo_eigenvalues(mo_eigenvalues, homo, nmo, level_shift)

      REAL(kind=dp), DIMENSION(:), INTENT(inout)         :: mo_eigenvalues
      INTEGER, INTENT(in)                                :: homo, nmo
      REAL(kind=dp), INTENT(in)                          :: level_shift

      CHARACTER(len=*), PARAMETER :: routineN = 'correct_mo_eigenvalues', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imo

      DO imo = homo+1, nmo
         mo_eigenvalues(imo) = mo_eigenvalues(imo)-level_shift
      END DO

   END SUBROUTINE correct_mo_eigenvalues

! **************************************************************************************************
!> \brief Adjust the Kohn-Sham matrix by shifting the orbital energies of all
!>        unoccupied molecular orbitals
!> \param matrix_ks_fm   transformed Kohn-Sham matrix = U^{-1,T} * KS * U^{-1}
!> \param mo_coeff       matrix of molecular orbitals (C)
!> \param homo           number of occupied molecular orbitals
!> \param nmo            number of molecular orbitals
!> \param nao            number of atomic orbitals
!> \param level_shift    amount of shift applying (in a.u.)
!> \param is_triangular  indicates that matrix_u_fm contains an upper triangular matrix
!> \param matrix_u_fm    matrix U: S (overlap matrix) = U^T * U;
!>                       assume an identity matrix if omitted
!> \par History
!>      03.2016 created [Sergey Chulkov]
! **************************************************************************************************
   SUBROUTINE shift_unocc_mos(matrix_ks_fm, mo_coeff, homo, nmo, nao, &
                              level_shift, is_triangular, matrix_u_fm)

      TYPE(cp_fm_type), POINTER                          :: matrix_ks_fm, mo_coeff
      INTEGER, INTENT(in)                                :: homo, nmo, nao
      REAL(kind=dp), INTENT(in)                          :: level_shift
      LOGICAL, INTENT(in)                                :: is_triangular
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: matrix_u_fm

      CHARACTER(len=*), PARAMETER :: routineN = 'shift_unocc_mos', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: weights
      TYPE(cp_fm_struct_type), POINTER                   :: ao_mo_fmstruct
      TYPE(cp_fm_type), POINTER                          :: u_mo, u_mo_scaled

      CALL timeset(routineN, handle)

      NULLIFY (ao_mo_fmstruct, u_mo, u_mo_scaled)
      CALL cp_fm_struct_create(ao_mo_fmstruct, nrow_global=nao, ncol_global=nmo, &
                               para_env=mo_coeff%matrix_struct%para_env, context=mo_coeff%matrix_struct%context)

      CALL cp_fm_create(u_mo, ao_mo_fmstruct)
      CALL cp_fm_create(u_mo_scaled, ao_mo_fmstruct)

      CALL cp_fm_struct_release(ao_mo_fmstruct)

      ! U * C
      IF (PRESENT(matrix_u_fm)) THEN
         IF (is_triangular) THEN
            CALL cp_fm_to_fm(mo_coeff, u_mo)
            CALL cp_fm_triangular_multiply(matrix_u_fm, u_mo, side="L", transpose_tr=.FALSE., &
                                           invert_tr=.FALSE., uplo_tr="U", n_rows=nao, n_cols=nmo, alpha=1.0_dp)
         ELSE
            CALL cp_gemm("N", "N", nao, nmo, nao, 1.0_dp, matrix_u_fm, mo_coeff, 0.0_dp, u_mo)
         END IF
      ELSE
         ! assume U is an identity matrix
         CALL cp_fm_to_fm(mo_coeff, u_mo)
      END IF

      CALL cp_fm_to_fm(u_mo, u_mo_scaled)

      ! up-shift all unoccupied molecular orbitals by the amount of 'level_shift'
      ! weight = diag(DELTA) = (0, ... 0, level_shift, ..., level_shift)
      !             MO index :  1 .. homo   homo+1     ...  nmo
      ALLOCATE (weights(nmo))
      weights(1:homo) = 0.0_dp
      weights(homo+1:nmo) = level_shift
      ! DELTA * U * C
      ! DELTA is a diagonal matrix, so simply scale all the columns of (U * C) by weights(:)
      CALL cp_fm_column_scale(u_mo_scaled, weights)
      DEALLOCATE (weights)

      ! NewKS = U^{-1,T} * KS * U^{-1} + (U * C) * DELTA * (U * C)^T
      CALL cp_gemm("N", "T", nao, nao, nmo, 1.0_dp, u_mo, u_mo_scaled, 1.0_dp, matrix_ks_fm)

      CALL cp_fm_release(u_mo_scaled)
      CALL cp_fm_release(u_mo)

      CALL timestop(handle)

   END SUBROUTINE shift_unocc_mos

END MODULE qs_scf_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Apply the direct inversion in the iterative subspace (DIIS) of Pulay
!>      in the framework of an SCF iteration for convergence acceleration
!> \par Literature
!>      - P. Pulay, Chem. Phys. Lett. 73, 393 (1980)
!>      - P. Pulay, J. Comput. Chem. 3, 556 (1982)
!> \par History
!>      - Changed to BLACS matrix usage (08.06.2001,MK)
!>      - rewritten to include LSD (1st attempt) (01.2003, Joost VandeVondele)
!>      - DIIS for ROKS (05.04.06,MK)
!> \author Matthias Krack (28.06.2000)
! **************************************************************************************************
MODULE qs_diis
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_init, cp_dbcsr_maxabs, &
        cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_set, cp_dbcsr_trace, &
        cp_dbcsr_transposed, cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale,&
                                              cp_fm_scale_and_add,&
                                              cp_fm_symm,&
                                              cp_fm_trace
   USE cp_fm_struct,                    ONLY: cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_maxabsval,&
                                              cp_fm_p_type,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE dm_ls_scf_types,                 ONLY: ls_scf_env_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              dp_size
   USE mathlib,                         ONLY: diamat_all
   USE qs_diis_types,                   ONLY: qs_diis_buffer_type,&
                                              qs_diis_buffer_type_sparse
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE string_utilities,                ONLY: compress
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_diis'
   INTEGER, SAVE               :: last_diis_b_id = 0

   ! Public subroutines

   PUBLIC :: qs_diis_b_clear, &
             qs_diis_b_create, &
             qs_diis_b_step
   PUBLIC :: qs_diis_b_clear_sparse, &
             qs_diis_b_create_sparse, &
             qs_diis_b_step_4lscf

CONTAINS

! **************************************************************************************************
!> \brief Allocates an SCF DIIS buffer
!> \param diis_buffer the buffer to create
!> \param nbuffer ...
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE qs_diis_b_create(diis_buffer, nbuffer)

      TYPE(qs_diis_buffer_type), POINTER                 :: diis_buffer
      INTEGER, INTENT(in)                                :: nbuffer

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_diis_b_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      ALLOCATE (diis_buffer)

      NULLIFY (diis_buffer%b_matrix)
      NULLIFY (diis_buffer%error)
      NULLIFY (diis_buffer%parameter)
      diis_buffer%nbuffer = nbuffer
      diis_buffer%ncall = 0
      last_diis_b_id = last_diis_b_id+1
      diis_buffer%id_nr = last_diis_b_id
      diis_buffer%ref_count = 1

      CALL timestop(handle)

   END SUBROUTINE qs_diis_b_create

! **************************************************************************************************
!> \brief Allocate and initialize a DIIS buffer for nao*nao parameter
!>      variables and with a buffer size of nbuffer.
!> \param diis_buffer the buffer to initialize
!> \param matrix_struct the structure for the matrix of the buffer
!> \param nspin ...
!> \param scf_section ...
!> \par History
!>      - Creation (07.05.2001, Matthias Krack)
!>      - Changed to BLACS matrix usage (08.06.2001,MK)
!>      - DIIS for ROKS (05.04.06,MK)
!> \author Matthias Krack
!> \note
!>      check to allocate matrixes only when needed, using a linked list?
! **************************************************************************************************
   SUBROUTINE qs_diis_b_check_i_alloc(diis_buffer, matrix_struct, nspin, &
                                      scf_section)

      TYPE(qs_diis_buffer_type), POINTER                 :: diis_buffer
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_struct
      INTEGER, INTENT(IN)                                :: nspin
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_diis_b_check_i_alloc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ibuffer, ispin, istat, nbuffer, &
                                                            output_unit
      TYPE(cp_logger_type), POINTER                      :: logger

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()

      IF(.NOT.(ASSOCIATED(diis_buffer)))CALL cp__a("qs_diis.F",150)
      IF(.NOT.(diis_buffer%ref_count > 0))CALL cp__a("qs_diis.F",151)

      nbuffer = diis_buffer%nbuffer

      IF (.NOT. ASSOCIATED(diis_buffer%error)) THEN
         ALLOCATE (diis_buffer%error(nbuffer, nspin), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 157, &
                                          "diis_buffer%error", 0)

         DO ispin = 1, nspin
            DO ibuffer = 1, nbuffer
               NULLIFY (diis_buffer%error(ibuffer, ispin)%matrix)
               CALL cp_fm_create(diis_buffer%error(ibuffer, ispin)%matrix, &
                                 name="qs_diis_b"// &
                                 TRIM(ADJUSTL(cp_to_string(diis_buffer%id_nr)))// &
                                 "%error("// &
                                 TRIM(ADJUSTL(cp_to_string(ibuffer)))//","// &
                                 TRIM(ADJUSTL(cp_to_string(ibuffer)))//")", &
                                 matrix_struct=matrix_struct)
            END DO
         END DO
      END IF

      IF (.NOT. ASSOCIATED(diis_buffer%parameter)) THEN
         ALLOCATE (diis_buffer%parameter(nbuffer, nspin), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 176, &
                                          "diis_buffer%parameter", 0)

         DO ispin = 1, nspin
            DO ibuffer = 1, nbuffer
               NULLIFY (diis_buffer%parameter(ibuffer, ispin)%matrix)
               CALL cp_fm_create(diis_buffer%parameter(ibuffer, ispin)%matrix, &
                                 name="qs_diis_b"// &
                                 TRIM(ADJUSTL(cp_to_string(diis_buffer%id_nr)))// &
                                 "%parameter("// &
                                 TRIM(ADJUSTL(cp_to_string(ibuffer)))//","// &
                                 TRIM(ADJUSTL(cp_to_string(ibuffer)))//")", &
                                 matrix_struct=matrix_struct)
            END DO
         END DO
      END IF

      IF (.NOT. ASSOCIATED(diis_buffer%b_matrix)) THEN
         ALLOCATE (diis_buffer%b_matrix(nbuffer+1, nbuffer+1), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 195, &
                                          "diis_buffer%b_matrix", &
                                          (nbuffer+1)**2*dp_size)
         diis_buffer%b_matrix = 0.0_dp
         output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%DIIS_INFO", &
                                            extension=".scfLog")
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, FMT="(/,T9,A)") &
               "DIIS | The SCF DIIS buffer was allocated and initialized"
         END IF
         CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                           "PRINT%DIIS_INFO")
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_diis_b_check_i_alloc

! **************************************************************************************************
!> \brief Update the SCF DIIS buffer, and if appropriate does a diis step.
!> \param diis_buffer ...
!> \param mo_array ...
!> \param kc ...
!> \param sc ...
!> \param delta ...
!> \param error_max ...
!> \param diis_step ...
!> \param eps_diis ...
!> \param nmixing ...
!> \param s_matrix ...
!> \param scf_section ...
!> \param roks ...
!> \par History
!>      - Creation (07.05.2001, Matthias Krack)
!>      - Changed to BLACS matrix usage (08.06.2001, MK)
!>      - 03.2003 rewamped [fawzi]
!>      - Adapted for high-spin ROKS (08.04.06,MK)
!> \author Matthias Krack
! **************************************************************************************************
   SUBROUTINE qs_diis_b_step(diis_buffer, mo_array, kc, sc, delta, error_max, &
                             diis_step, eps_diis, nmixing, s_matrix, scf_section, roks)

      TYPE(qs_diis_buffer_type), POINTER                 :: diis_buffer
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: kc
      TYPE(cp_fm_type), POINTER                          :: sc
      REAL(KIND=dp), INTENT(IN)                          :: delta
      REAL(KIND=dp), INTENT(OUT)                         :: error_max
      LOGICAL, INTENT(OUT)                               :: diis_step
      REAL(KIND=dp), INTENT(IN)                          :: eps_diis
      INTEGER, INTENT(IN), OPTIONAL                      :: nmixing
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: s_matrix
      TYPE(section_vals_type), POINTER                   :: scf_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: roks

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_diis_b_step', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER :: eigenvalue_threshold = 1.0E-12_dp

      CHARACTER(LEN=2*default_string_length)             :: message
      INTEGER                                            :: handle, homo, ib, imo, ispin, istat, jb, &
                                                            my_nmixing, nao, nb, nb1, nmo, nspin, &
                                                            output_unit
      LOGICAL                                            :: eigenvectors_discarded, mo_uocc, my_roks
      REAL(KIND=dp)                                      :: maxocc, tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ev, occ
      REAL(KIND=dp), DIMENSION(:), POINTER               :: occa, occb
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: a, b
      TYPE(cp_fm_struct_type), POINTER                   :: matrix_struct
      TYPE(cp_fm_type), POINTER                          :: c, new_errors, old_errors, parameters
      TYPE(cp_logger_type), POINTER                      :: logger

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      nspin = SIZE(mo_array)
      diis_step = .FALSE.

      IF (PRESENT(roks)) THEN
         my_roks = .TRUE.
         nspin = 1
      ELSE
         my_roks = .FALSE.
      END IF

      my_nmixing = 2
      IF (PRESENT(nmixing)) my_nmixing = nmixing

      NULLIFY (c, new_errors, old_errors, parameters, matrix_struct, a, b, occa, occb)
      logger => cp_get_default_logger()

      ! Quick return, if no DIIS is requested

      IF (diis_buffer%nbuffer < 1) THEN
         CALL timestop(handle)
         RETURN
      END IF

      CALL cp_fm_get_info(kc(1)%matrix, &
                          matrix_struct=matrix_struct)
      CALL qs_diis_b_check_i_alloc(diis_buffer, &
                                   matrix_struct=matrix_struct, &
                                   nspin=nspin, &
                                   scf_section=scf_section)

      error_max = 0.0_dp

      ib = MODULO(diis_buffer%ncall, diis_buffer%nbuffer)+1
      diis_buffer%ncall = diis_buffer%ncall+1
      nb = MIN(diis_buffer%ncall, diis_buffer%nbuffer)

      DO ispin = 1, nspin

         CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, &
                         nao=nao, &
                         nmo=nmo, &
                         homo=homo, &
                         mo_coeff=c, &
                         occupation_numbers=occa, &
                         uniform_occupation=mo_uocc, &
                         maxocc=maxocc)

         new_errors => diis_buffer%error(ib, ispin)%matrix
         parameters => diis_buffer%parameter(ib, ispin)%matrix

         ! Copy the Kohn-Sham matrix K to the DIIS buffer

         CALL cp_fm_to_fm(kc(ispin)%matrix, parameters)

         IF (my_roks) THEN

            ALLOCATE (occ(nmo), STAT=istat)
            IF (istat /= 0) CALL stop_memory(routineN, moduleN, 328, &
                                             "occ", nmo*dp_size)

            CALL get_mo_set(mo_set=mo_array(2)%mo_set, &
                            occupation_numbers=occb)

            DO imo = 1, nmo
               occ(imo) = SQRT(occa(imo)+occb(imo))
            END DO

            CALL cp_fm_to_fm(c, sc)
            CALL cp_fm_column_scale(sc, occ(1:homo))

            ! KC <- K*C
            CALL cp_fm_symm("L", "U", nao, homo, 1.0_dp, parameters, sc, 0.0_dp, kc(ispin)%matrix)

            IF (PRESENT(s_matrix)) THEN
               CALL copy_dbcsr_to_fm(s_matrix(1)%matrix, new_errors)
               ! SC <- S*C
               CALL cp_fm_symm("L", "U", nao, homo, 1.0_dp, new_errors, c, 0.0_dp, sc)
               CALL cp_fm_column_scale(sc, occ(1:homo))
            END IF

            ! new_errors <- KC*(SC)^T - (SC)*(KC)^T = K*P*S - S*P*K
            ! or for an orthogonal basis
            ! new_errors <- KC*C^T - C*(KC)^T = K*P - P*K with S = I
            CALL cp_gemm("N", "T", nao, nao, homo, 1.0_dp, sc, kc(ispin)%matrix, 0.0_dp, new_errors)
            CALL cp_gemm("N", "T", nao, nao, homo, 1.0_dp, kc(ispin)%matrix, sc, -1.0_dp, new_errors)

            DEALLOCATE (occ)

         ELSE

            ! KC <- K*C
            CALL cp_fm_symm("L", "U", nao, homo, maxocc, parameters, c, 0.0_dp, kc(ispin)%matrix)

            IF (PRESENT(s_matrix)) THEN
               ! I guess that this copy can be avoided for LSD
               CALL copy_dbcsr_to_fm(s_matrix(1)%matrix, new_errors)
               ! sc <- S*C
               CALL cp_fm_symm("L", "U", nao, homo, 2.0_dp, new_errors, c, 0.0_dp, sc)
               ! new_errors <- KC*(SC)^T - (SC)*(KC)^T = K*P*S - S*P*K
               CALL cp_gemm("N", "T", nao, nao, homo, 1.0_dp, sc, kc(ispin)%matrix, 0.0_dp, new_errors)
               CALL cp_gemm("N", "T", nao, nao, homo, 1.0_dp, kc(ispin)%matrix, sc, -1.0_dp, new_errors)
            ELSE
               ! new_errors <- KC*(C)^T - C*(KC)^T = K*P - P*K
               CALL cp_gemm("N", "T", nao, nao, homo, 1.0_dp, c, kc(ispin)%matrix, 0.0_dp, new_errors)
               CALL cp_gemm("N", "T", nao, nao, homo, 1.0_dp, kc(ispin)%matrix, c, -1.0_dp, new_errors)
            END IF

         END IF

         CALL cp_fm_maxabsval(new_errors, tmp)
         error_max = MAX(error_max, tmp)

      END DO

      ! Check, if a DIIS step is appropiate

      diis_step = ((diis_buffer%ncall >= my_nmixing) .AND. (delta < eps_diis))

      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%DIIS_INFO", &
                                         extension=".scfLog")
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,T9,A,I4,/,(T9,A,ES12.3))") &
            "DIIS | Current SCF DIIS buffer size:         ", nb, &
            "DIIS | Maximum SCF DIIS error vector element:", error_max, &
            "DIIS | Current SCF convergence:              ", delta, &
            "DIIS | Threshold value for a DIIS step:      ", eps_diis
         IF (error_max < eps_diis) THEN
            WRITE (UNIT=output_unit, FMT="(T9,A)") &
               "DIIS | => The SCF DIIS buffer will be updated"
         ELSE
            WRITE (UNIT=output_unit, FMT="(T9,A)") &
               "DIIS | => No update of the SCF DIIS buffer"
         END IF
         IF (diis_step .AND. (error_max < eps_diis)) THEN
            WRITE (UNIT=output_unit, FMT="(T9,A,/)") &
               "DIIS | => A SCF DIIS step will be performed"
         ELSE
            WRITE (UNIT=output_unit, FMT="(T9,A,/)") &
               "DIIS | => No SCF DIIS step will be performed"
         END IF
      END IF

      ! Update the SCF DIIS buffer

      IF (error_max < eps_diis) THEN

         b => diis_buffer%b_matrix

         DO jb = 1, nb
            b(jb, ib) = 0.0_dp
            DO ispin = 1, nspin
               old_errors => diis_buffer%error(jb, ispin)%matrix
               new_errors => diis_buffer%error(ib, ispin)%matrix
               CALL cp_fm_trace(old_errors, new_errors, tmp)
               b(jb, ib) = b(jb, ib)+tmp
            END DO
            b(ib, jb) = b(jb, ib)
         END DO

      ELSE

         diis_step = .FALSE.

      END IF

      ! Perform DIIS step

      IF (diis_step) THEN

         nb1 = nb+1

         ALLOCATE (a(nb1, nb1), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 443, &
                                          "a", nb1*nb1*dp_size)
         ALLOCATE (b(nb1, nb1), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 446, &
                                          "b", nb1*nb1*dp_size)
         ALLOCATE (ev(nb1), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 449, &
                                          "ev", nb1*dp_size)

         ! Set up the linear DIIS equation system

         b(1:nb, 1:nb) = diis_buffer%b_matrix(1:nb, 1:nb)

         b(1:nb, nb1) = -1.0_dp
         b(nb1, 1:nb) = -1.0_dp
         b(nb1, nb1) = 0.0_dp

         ! Solve the linear DIIS equation system

         CALL diamat_all(b(1:nb1, 1:nb1), ev(1:nb1))

         a(1:nb1, 1:nb1) = b(1:nb1, 1:nb1)

         eigenvectors_discarded = .FALSE.

         DO jb = 1, nb1
            IF (ABS(ev(jb)) < eigenvalue_threshold) THEN
               IF (output_unit > 0) THEN
                  IF (.NOT. eigenvectors_discarded) THEN
                     WRITE (UNIT=output_unit, FMT="(T9,A)") &
                        "DIIS | Checking eigenvalues of the DIIS error matrix"
                  END IF
                  WRITE (UNIT=message, FMT="(T9,A,I6,A,ES10.1,A,ES10.1)") &
                     "DIIS | Eigenvalue ", jb, " = ", ev(jb), " is smaller than "// &
                     "threshold ", eigenvalue_threshold
                  CALL compress(message)
                  WRITE (UNIT=output_unit, FMT="(T9,A)") TRIM(message)
                  eigenvectors_discarded = .TRUE.
               END IF
               a(1:nb1, jb) = 0.0_dp
            ELSE
               a(1:nb1, jb) = a(1:nb1, jb)/ev(jb)
            END IF
         END DO

         IF ((output_unit > 0) .AND. eigenvectors_discarded) THEN
            WRITE (UNIT=output_unit, FMT="(T9,A,/)") &
               "DIIS | The corresponding eigenvectors were discarded"
         END IF

         ev(1:nb) = MATMUL(a(1:nb, 1:nb1), b(nb1, 1:nb1))

         ! Update Kohn-Sham matrix

         DO ispin = 1, nspin
            CALL cp_fm_set_all(kc(ispin)%matrix, 0.0_dp)
            DO jb = 1, nb
               parameters => diis_buffer%parameter(jb, ispin)%matrix
               CALL cp_fm_scale_and_add(1.0_dp, kc(ispin)%matrix, -ev(jb), parameters)
            END DO
         END DO

         DEALLOCATE (a)
         DEALLOCATE (b)
         DEALLOCATE (ev)

      ELSE

         DO ispin = 1, nspin
            parameters => diis_buffer%parameter(ib, ispin)%matrix
            CALL cp_fm_to_fm(parameters, kc(ispin)%matrix)
         END DO

      END IF

      CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                        "PRINT%DIIS_INFO")

      CALL timestop(handle)

   END SUBROUTINE qs_diis_b_step

! **************************************************************************************************
!> \brief clears the buffer
!> \param diis_buffer the buffer to clear
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE qs_diis_b_clear(diis_buffer)

      TYPE(qs_diis_buffer_type), POINTER                 :: diis_buffer

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_diis_b_clear', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(diis_buffer)))CALL cp__a("qs_diis.F",545)
      IF(.NOT.(diis_buffer%ref_count > 0))CALL cp__a("qs_diis.F",546)

      diis_buffer%ncall = 0

      CALL timestop(handle)

   END SUBROUTINE qs_diis_b_clear

! **************************************************************************************************
!> \brief Update the SCF DIIS buffer in linear scaling SCF (LS-SCF),
!>        and if appropriate does a diis step.
!> \param diis_buffer ...
!> \param qs_env ...
!> \param ls_scf_env ...
!> \param unit_nr ...
!> \param iscf ...
!> \param diis_step ...
!> \param eps_diis ...
!> \param nmixing ...
!> \param s_matrix ...
!> \param threshold ...
!> \par History
!>      - Adapted for LS-SCF (10-11-14) from qs_diis_b_step
!> \author Fredy W. Aquino
! **************************************************************************************************

   SUBROUTINE qs_diis_b_step_4lscf(diis_buffer, qs_env, ls_scf_env, unit_nr, iscf, &
                                   diis_step, eps_diis, nmixing, s_matrix, threshold)
! Note.- Input: ls_scf_env%matrix_p(ispin) , Density Matrix
!               matrix_ks (from qs_env)    , Kohn-Sham Matrix  (IN/OUT)

      TYPE(qs_diis_buffer_type_sparse), POINTER          :: diis_buffer
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(ls_scf_env_type)                              :: ls_scf_env
      INTEGER, INTENT(IN)                                :: unit_nr, iscf
      LOGICAL, INTENT(OUT)                               :: diis_step
      REAL(KIND=dp), INTENT(IN)                          :: eps_diis
      INTEGER, INTENT(IN), OPTIONAL                      :: nmixing
      TYPE(cp_dbcsr_type), OPTIONAL                      :: s_matrix
      REAL(KIND=dp), INTENT(IN)                          :: threshold

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_diis_b_step_4lscf', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER :: eigenvalue_threshold = 1.0E-12_dp

      INTEGER                                            :: handle, ib, ispin, istat, jb, &
                                                            my_nmixing, nb, nb1, nspin
      REAL(KIND=dp)                                      :: error_max, tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: ev
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: a, b
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(cp_dbcsr_type)                                :: matrix_KSerr_t, matrix_tmp
      TYPE(cp_dbcsr_type), POINTER                       :: new_errors, old_errors, parameters
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL timeset(routineN, handle)
      nspin = ls_scf_env%nspins
      diis_step = .FALSE.
      my_nmixing = 2
      IF (PRESENT(nmixing)) my_nmixing = nmixing
      NULLIFY (new_errors, old_errors, parameters, a, b)
      logger => cp_get_default_logger()
      ! Quick return, if no DIIS is requested
      IF (diis_buffer%nbuffer < 1) THEN
         CALL timestop(handle)
         RETURN
      END IF

! Getting current Kohn-Sham matrix from qs_env
      CALL get_qs_env(qs_env, &
                      para_env=para_env, &
                      matrix_ks=matrix_ks)
      CALL qs_diis_b_check_i_alloc_sparse( &
         diis_buffer, &
         ls_scf_env, &
         nspin)
      error_max = 0.0_dp

      ib = MODULO(diis_buffer%ncall, diis_buffer%nbuffer)+1
      diis_buffer%ncall = diis_buffer%ncall+1
      nb = MIN(diis_buffer%ncall, diis_buffer%nbuffer)
! Create scratch arrays
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, &
                           template=ls_scf_env%matrix_ks(1), &
                           matrix_type='N')
      CALL cp_dbcsr_set(matrix_tmp, 0.0_dp) ! reset matrix
      CALL cp_dbcsr_init(matrix_KSerr_t)
      CALL cp_dbcsr_create(matrix_KSerr_t, &
                           template=ls_scf_env%matrix_ks(1), &
                           matrix_type='N')
      CALL cp_dbcsr_set(matrix_KSerr_t, 0.0_dp) ! reset matrix

      DO ispin = 1, nspin ! ------ Loop-ispin----START

         new_errors => diis_buffer%error(ib, ispin)%matrix
         parameters => diis_buffer%parameter(ib, ispin)%matrix
         ! Copy the Kohn-Sham matrix K to the DIIS buffer
         CALL cp_dbcsr_copy(parameters, & ! out
                            matrix_ks(ispin)%matrix) ! in

         IF (PRESENT(s_matrix)) THEN ! if-s_matrix ---------- START
! Calculate Kohn-Sham error (non-orthogonal)= K*P*S-(K*P*S)^T
! matrix_tmp = P*S
            CALL cp_dbcsr_multiply("N", "N", &
                                   1.0_dp, ls_scf_env%matrix_p(ispin), &
                                   s_matrix, &
                                   0.0_dp, matrix_tmp, &
                                   filter_eps=threshold)
! new_errors= K*P*S
            CALL cp_dbcsr_multiply("N", "N", &
                                   1.0_dp, matrix_ks(ispin)%matrix, &
                                   matrix_tmp, &
                                   0.0_dp, new_errors, &
                                   filter_eps=threshold)
! matrix_KSerr_t= transpose(K*P*S)
            CALL cp_dbcsr_transposed(matrix_KSerr_t, &
                                     new_errors)
! new_errors=K*P*S-transpose(K*P*S)
            CALL cp_dbcsr_add(new_errors, &
                              matrix_KSerr_t, &
                              1.0_dp, -1.0_dp)
         ELSE ! if-s_matrix ---------- MID
! Calculate Kohn-Sham error (orthogonal)= K*P - P*K
! new_errors=K*P
            CALL cp_dbcsr_multiply("N", "N", &
                                   1.0_dp, matrix_ks(ispin)%matrix, &
                                   ls_scf_env%matrix_p(ispin), &
                                   0.0_dp, new_errors, &
                                   filter_eps=threshold)
! matrix_KSerr_t= transpose(K*P)
            CALL cp_dbcsr_transposed(matrix_KSerr_t, &
                                     new_errors)
! new_errors=K*P-transpose(K*P)
            CALL cp_dbcsr_add(new_errors, &
                              matrix_KSerr_t, &
                              1.0_dp, -1.0_dp)
         END IF ! if-s_matrix ---------- END

         tmp = cp_dbcsr_maxabs(new_errors)
         error_max = MAX(error_max, tmp)

      END DO ! ------ Loop-ispin----END

      ! Check, if a DIIS step is appropiate

      diis_step = (diis_buffer%ncall >= my_nmixing)

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(A29,I3,A3,4(I3,A1))') &
            "DIIS: (ncall,nbuffer,ib,nb)=(", iscf, ")=(", &
            diis_buffer%ncall, ",", diis_buffer%nbuffer, ",", ib, ",", nb, ")"
         WRITE (unit_nr, '(A57,I3,A3,L1,A1,F10.8,A1,F4.2,A1,L1,A1)') &
            "DIIS: (diis_step,error_max,eps_diis,error_max<eps_diis)=(", &
            iscf, ")=(", diis_step, ",", error_max, ",", eps_diis, ",", &
            (error_max < eps_diis), ")"
         WRITE (unit_nr, '(A75)') &
            "DIIS: diis_step=T : Perform DIIS  error_max<eps_diis=T : Update DIIS buffer"
      ENDIF

      ! Update the SCF DIIS buffer
      IF (error_max < eps_diis) THEN
         b => diis_buffer%b_matrix
         DO jb = 1, nb
            b(jb, ib) = 0.0_dp
            DO ispin = 1, nspin
               old_errors => diis_buffer%error(jb, ispin)%matrix
               new_errors => diis_buffer%error(ib, ispin)%matrix
               CALL cp_dbcsr_trace(old_errors, &
                                   new_errors, &
                                   tmp) ! out : < f_i | f_j >
               b(jb, ib) = b(jb, ib)+tmp
            END DO ! end-loop-ispin
            b(ib, jb) = b(jb, ib)
         END DO ! end-loop-jb
      ELSE
         diis_step = .FALSE.
      END IF

      ! Perform DIIS step
      IF (diis_step) THEN
         nb1 = nb+1
         ALLOCATE (a(nb1, nb1), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 730, &
                                          "a", nb1*nb1*dp_size)
         ALLOCATE (b(nb1, nb1), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 733, &
                                          "b", nb1*nb1*dp_size)
         ALLOCATE (ev(nb1), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 736, &
                                          "ev", nb1*dp_size)
         ! Set up the linear DIIS equation system
         b(1:nb, 1:nb) = diis_buffer%b_matrix(1:nb, 1:nb)
         b(1:nb, nb1) = -1.0_dp
         b(nb1, 1:nb) = -1.0_dp
         b(nb1, nb1) = 0.0_dp
         ! Solve the linear DIIS equation system
         CALL diamat_all(b(1:nb1, 1:nb1), ev(1:nb1))
         a(1:nb1, 1:nb1) = b(1:nb1, 1:nb1)
         DO jb = 1, nb1
            IF (ABS(ev(jb)) < eigenvalue_threshold) THEN
               a(1:nb1, jb) = 0.0_dp
            ELSE
               a(1:nb1, jb) = a(1:nb1, jb)/ev(jb)
            END IF
         END DO ! end-loop-jb

         ev(1:nb) = MATMUL(a(1:nb, 1:nb1), b(nb1, 1:nb1))

         ! Update Kohn-Sham matrix
         IF (iscf .GE. ls_scf_env%iter_ini_diis) THEN ! if-iscf-to-updateKS------ START

            IF (unit_nr > 0) THEN
               WRITE (unit_nr, '(A40,I3)') 'DIIS: Updating Kohn-Sham matrix at iscf=', iscf
            ENDIF

            DO ispin = 1, nspin
               CALL cp_dbcsr_set(matrix_ks(ispin)%matrix, & ! reset matrix
                                 0.0_dp)
               DO jb = 1, nb
                  parameters => diis_buffer%parameter(jb, ispin)%matrix
                  CALL cp_dbcsr_add(matrix_ks(ispin)%matrix, parameters, &
                                    1.0_dp, -ev(jb))
               END DO ! end-loop-jb
            END DO ! end-loop-ispin
         ENDIF ! if-iscf-to-updateKS------ END

         DEALLOCATE (a)
         DEALLOCATE (b)
         DEALLOCATE (ev)

      ELSE
         DO ispin = 1, nspin
            parameters => diis_buffer%parameter(ib, ispin)%matrix
            CALL cp_dbcsr_copy(parameters, & ! out
                               matrix_ks(ispin)%matrix) ! in
         ENDDO ! end-loop-ispin
      END IF
      CALL cp_dbcsr_release(matrix_tmp)
      CALL cp_dbcsr_release(matrix_KSerr_t)
      CALL timestop(handle)

   END SUBROUTINE qs_diis_b_step_4lscf

! **************************************************************************************************
!> \brief Allocate and initialize a DIIS buffer with a buffer size of nbuffer.
!> \param diis_buffer the buffer to initialize
!> \param ls_scf_env ...
!> \param nspin ...
!> \par History
!>      - Adapted from qs_diis_b_check_i_alloc for sparse matrices and
!>        used in LS-SCF module (ls_scf_main) (10-11-14)
!> \author Fredy W. Aquino
!> \note
!>      check to allocate matrices only when needed
! **************************************************************************************************

   SUBROUTINE qs_diis_b_check_i_alloc_sparse(diis_buffer, ls_scf_env, &
                                             nspin)

      TYPE(qs_diis_buffer_type_sparse), POINTER          :: diis_buffer
      TYPE(ls_scf_env_type)                              :: ls_scf_env
      INTEGER, INTENT(IN)                                :: nspin

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_diis_b_check_i_alloc_sparse', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ibuffer, ispin, istat, nbuffer
      TYPE(cp_logger_type), POINTER                      :: logger

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()

      IF(.NOT.(ASSOCIATED(diis_buffer)))CALL cp__a("qs_diis.F",823)
      IF(.NOT.(diis_buffer%ref_count > 0))CALL cp__a("qs_diis.F",824)

      nbuffer = diis_buffer%nbuffer

      IF (.NOT. ASSOCIATED(diis_buffer%error)) THEN
         ALLOCATE (diis_buffer%error(nbuffer, nspin), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 830, &
                                          "diis_buffer%error", 0)

         DO ispin = 1, nspin
            DO ibuffer = 1, nbuffer
               ALLOCATE (diis_buffer%error(ibuffer, ispin)%matrix)
               CALL cp_dbcsr_init(diis_buffer%error(ibuffer, ispin)%matrix)

               CALL cp_dbcsr_create(diis_buffer%error(ibuffer, ispin)%matrix, &
                                    template=ls_scf_env%matrix_ks(1), &
                                    matrix_type='N')
            END DO
         END DO
      END IF

      IF (.NOT. ASSOCIATED(diis_buffer%parameter)) THEN
         ALLOCATE (diis_buffer%parameter(nbuffer, nspin), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 847, &
                                          "diis_buffer%parameter", 0)

         DO ispin = 1, nspin
            DO ibuffer = 1, nbuffer
               ALLOCATE (diis_buffer%parameter(ibuffer, ispin)%matrix)
               CALL cp_dbcsr_init(diis_buffer%parameter(ibuffer, ispin)%matrix)
               CALL cp_dbcsr_create(diis_buffer%parameter(ibuffer, ispin)%matrix, &
                                    template=ls_scf_env%matrix_ks(1), &
                                    matrix_type='N')
            END DO
         END DO
      END IF

      IF (.NOT. ASSOCIATED(diis_buffer%b_matrix)) THEN
         ALLOCATE (diis_buffer%b_matrix(nbuffer+1, nbuffer+1), STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 863, &
                                          "diis_buffer%b_matrix", &
                                          (nbuffer+1)**2*dp_size)

         diis_buffer%b_matrix = 0.0_dp
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_diis_b_check_i_alloc_sparse

! **************************************************************************************************
!> \brief clears the DIIS buffer in LS-SCF calculation
!> \param diis_buffer the buffer to clear
!> \par History
!>      10-11-14 created [FA] modified from qs_diis_b_clear
!> \author Fredy W. Aquino
! **************************************************************************************************

   SUBROUTINE qs_diis_b_clear_sparse(diis_buffer)

      TYPE(qs_diis_buffer_type_sparse), POINTER          :: diis_buffer

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_diis_b_clear_sparse', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(diis_buffer)))CALL cp__a("qs_diis.F",895)
      IF(.NOT.(diis_buffer%ref_count > 0))CALL cp__a("qs_diis.F",896)

      diis_buffer%ncall = 0

      CALL timestop(handle)

   END SUBROUTINE qs_diis_b_clear_sparse

! **************************************************************************************************
!> \brief Allocates an SCF DIIS buffer for LS-SCF calculation
!> \param diis_buffer the buffer to create
!> \param nbuffer ...
!> \par History
!>      10-11-14 created [FA] modified from qs_diis_b_create
!> \author Fredy W. Aquino
! **************************************************************************************************
   SUBROUTINE qs_diis_b_create_sparse(diis_buffer, nbuffer)

      TYPE(qs_diis_buffer_type_sparse), POINTER          :: diis_buffer
      INTEGER, INTENT(in)                                :: nbuffer

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_diis_b_create_sparse', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      ALLOCATE (diis_buffer)

      NULLIFY (diis_buffer%b_matrix)
      NULLIFY (diis_buffer%error)
      NULLIFY (diis_buffer%parameter)
      diis_buffer%nbuffer = nbuffer
      diis_buffer%ncall = 0
      last_diis_b_id = last_diis_b_id+1
      diis_buffer%id_nr = last_diis_b_id
      diis_buffer%ref_count = 1

      CALL timestop(handle)

   END SUBROUTINE qs_diis_b_create_sparse

END MODULE qs_diis
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief computes preconditioners, and implements methods to apply them
!>      currently used in qs_ot
!> \par History
!>      - [UB] 2009-05-13 Adding stable approximate inverse (full and sparse)
!> \author Joost VandeVondele (09.2002)
! **************************************************************************************************
MODULE preconditioner_makes
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_add_on_diag, cp_dbcsr_arnoldi_ev, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_get_info, cp_dbcsr_init, cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_release, &
        cp_dbcsr_setup_arnoldi_data, cp_dbcsr_type, cp_set_arnoldi_initial_vector, &
        dbcsr_arnoldi_data, dbcsr_type_symmetric, deallocate_arnoldi_data, get_selected_ritz_val, &
        get_selected_ritz_vec
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              cp_dbcsr_m_by_n_from_template,&
                                              cp_dbcsr_sm_fm_multiply,&
                                              cp_fm_to_dbcsr_row_template
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale,&
                                              cp_fm_triangular_invert,&
                                              cp_fm_triangular_multiply,&
                                              cp_fm_upper_to_full
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose,&
                                              cp_fm_cholesky_reduce,&
                                              cp_fm_cholesky_restore
   USE cp_fm_diag,                      ONLY: choose_eigv_solver
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_diag,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_to_fm,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE input_constants,                 ONLY: &
        cholesky_inverse, cholesky_reduce, ot_precond_full_all, ot_precond_full_kinetic, &
        ot_precond_full_single, ot_precond_full_single_inverse, ot_precond_s_inverse, &
        ot_precond_solver_default, ot_precond_solver_inv_chol
   USE kinds,                           ONLY: dp
   USE preconditioner_types,            ONLY: preconditioner_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'preconditioner_makes'

   PUBLIC :: make_preconditioner_matrix

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param preconditioner_env ...
!> \param matrix_h ...
!> \param matrix_s ...
!> \param matrix_t ...
!> \param mo_coeff ...
!> \param energy_homo ...
!> \param eigenvalues_ot ...
!> \param energy_gap ...
!> \param my_solver_type ...
! **************************************************************************************************
   SUBROUTINE make_preconditioner_matrix(preconditioner_env, matrix_h, matrix_s, matrix_t, mo_coeff, &
                                         energy_homo, eigenvalues_ot, energy_gap, &
                                         my_solver_type)
      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: matrix_s, matrix_t
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      REAL(KIND=dp)                                      :: energy_homo
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigenvalues_ot
      REAL(KIND=dp)                                      :: energy_gap
      INTEGER                                            :: my_solver_type

      CHARACTER(len=*), PARAMETER :: routineN = 'make_preconditioner_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: precon_type

      precon_type = preconditioner_env%in_use
      SELECT CASE (precon_type)
      CASE (ot_precond_full_single)
         IF (my_solver_type .NE. ot_precond_solver_default) &
            CALL cp__b("preconditioner_makes.F",93,"Only PRECOND_SOLVER DEFAULT for the moment")
         IF (PRESENT(matrix_s)) THEN
            CALL make_full_single(preconditioner_env, preconditioner_env%fm, &
                                  matrix_h, matrix_s, energy_homo, energy_gap)
         ELSE
            CALL make_full_single_ortho(preconditioner_env, preconditioner_env%fm, &
                                        matrix_h, energy_homo, energy_gap)
         END IF

      CASE (ot_precond_s_inverse)
         IF (my_solver_type .EQ. ot_precond_solver_default) my_solver_type = ot_precond_solver_inv_chol
         IF (.NOT. PRESENT(matrix_s)) &
            CALL cp__b("preconditioner_makes.F",105,"Type for S=1 not implemented")
         CALL make_full_s_inverse(preconditioner_env, matrix_s)

      CASE (ot_precond_full_kinetic)
         IF (my_solver_type .EQ. ot_precond_solver_default) my_solver_type = ot_precond_solver_inv_chol
         IF (.NOT. (PRESENT(matrix_s) .AND. PRESENT(matrix_t))) &
            CALL cp__b("preconditioner_makes.F",111,"Type for S=1 not implemented")
         CALL make_full_kinetic(preconditioner_env, matrix_t, matrix_s, energy_gap)
      CASE (ot_precond_full_single_inverse)
         IF (my_solver_type .EQ. ot_precond_solver_default) my_solver_type = ot_precond_solver_inv_chol
         CALL make_full_single_inverse(preconditioner_env, mo_coeff, matrix_h, energy_gap, &
                                       matrix_s=matrix_s)
      CASE (ot_precond_full_all)
         IF (my_solver_type .NE. ot_precond_solver_default) THEN
            CALL cp__b("preconditioner_makes.F",119,"Only PRECOND_SOLVER DEFAULT for the moment")
         ENDIF
         IF (PRESENT(matrix_s)) THEN
            CALL make_full_all(preconditioner_env, mo_coeff, matrix_h, matrix_s, &
                               eigenvalues_ot, energy_gap)
         ELSE
            CALL make_full_all_ortho(preconditioner_env, mo_coeff, matrix_h, &
                                     eigenvalues_ot, energy_gap)
         END IF

      CASE DEFAULT
         CALL cp__b("preconditioner_makes.F",130,"Type not implemented")
      END SELECT

   END SUBROUTINE make_preconditioner_matrix

! **************************************************************************************************
!> \brief Simply takes the overlap matrix as preconditioner
!> \param preconditioner_env ...
!> \param matrix_s ...
! **************************************************************************************************
   SUBROUTINE make_full_s_inverse(preconditioner_env, matrix_s)
      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_s

      CHARACTER(len=*), PARAMETER :: routineN = 'make_full_s_inverse', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(matrix_s)))CALL cp__a("preconditioner_makes.F",151)

      IF (.NOT. ASSOCIATED(preconditioner_env%sparse_matrix)) THEN
         ALLOCATE (preconditioner_env%sparse_matrix)
         CALL cp_dbcsr_init(preconditioner_env%sparse_matrix)
      END IF
      CALL cp_dbcsr_copy(preconditioner_env%sparse_matrix, matrix_s, name="full_kinetic")

      CALL timestop(handle)

   END SUBROUTINE make_full_s_inverse

! **************************************************************************************************
!> \brief kinetic matrix+shift*overlap as preconditioner. Cheap but could
!>        be better
!> \param preconditioner_env ...
!> \param matrix_t ...
!> \param matrix_s ...
!> \param energy_gap ...
! **************************************************************************************************
   SUBROUTINE make_full_kinetic(preconditioner_env, matrix_t, matrix_s, &
                                energy_gap)
      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_t, matrix_s
      REAL(KIND=dp)                                      :: energy_gap

      CHARACTER(len=*), PARAMETER :: routineN = 'make_full_kinetic', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: shift

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(matrix_t)))CALL cp__a("preconditioner_makes.F",185)
      IF(.NOT.(ASSOCIATED(matrix_s)))CALL cp__a("preconditioner_makes.F",186)

      IF (.NOT. ASSOCIATED(preconditioner_env%sparse_matrix)) THEN
         ALLOCATE (preconditioner_env%sparse_matrix)
         CALL cp_dbcsr_init(preconditioner_env%sparse_matrix)
      END IF
      CALL cp_dbcsr_copy(preconditioner_env%sparse_matrix, matrix_t, name="full_kinetic")

      shift = MAX(0.0_dp, energy_gap)

      CALL cp_dbcsr_add(preconditioner_env%sparse_matrix, matrix_s, &
                        alpha_scalar=1.0_dp, beta_scalar=shift)

      CALL timestop(handle)

   END SUBROUTINE make_full_kinetic

! **************************************************************************************************
!> \brief full_single_preconditioner
!> \param preconditioner_env ...
!> \param fm ...
!> \param matrix_h ...
!> \param matrix_s ...
!> \param energy_homo ...
!> \param energy_gap ...
! **************************************************************************************************
   SUBROUTINE make_full_single(preconditioner_env, fm, matrix_h, matrix_s, &
                               energy_homo, energy_gap)
      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_fm_type), POINTER                          :: fm
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h, matrix_s
      REAL(KIND=dp)                                      :: energy_homo, energy_gap

      CHARACTER(len=*), PARAMETER :: routineN = 'make_full_single', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, n
      REAL(KIND=dp), DIMENSION(:), POINTER               :: evals
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: fm_h, fm_s

      CALL timeset(routineN, handle)

      NULLIFY (fm_h, fm_s, fm_struct_tmp, evals)

      IF (ASSOCIATED(fm)) THEN
         CALL cp_fm_release(fm)
      ENDIF
      CALL cp_dbcsr_get_info(matrix_h, nfullrows_total=n)
      ALLOCATE (evals(n))

      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=n, ncol_global=n, &
                               context=preconditioner_env%ctxt, &
                               para_env=preconditioner_env%para_env)
      CALL cp_fm_create(fm, fm_struct_tmp, name="preconditioner")
      CALL cp_fm_create(fm_h, fm_struct_tmp, name="fm_h")
      CALL cp_fm_create(fm_s, fm_struct_tmp, name="fm_s")
      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL copy_dbcsr_to_fm(matrix_h, fm_h)
      CALL copy_dbcsr_to_fm(matrix_s, fm_s)
      CALL cp_fm_cholesky_decompose(fm_s)

      SELECT CASE (preconditioner_env%cholesky_use)
      CASE (cholesky_inverse)
! if cho inverse
         CALL cp_fm_triangular_invert(fm_s)
         CALL cp_fm_upper_to_full(fm_h, fm)

         CALL cp_fm_triangular_multiply(fm_s, fm_h, side="R", transpose_tr=.FALSE., &
                                        invert_tr=.FALSE., uplo_tr="U", n_rows=n, n_cols=n, alpha=1.0_dp)
         CALL cp_fm_triangular_multiply(fm_s, fm_h, side="L", transpose_tr=.TRUE., &
                                        invert_tr=.FALSE., uplo_tr="U", n_rows=n, n_cols=n, alpha=1.0_dp)
      CASE (cholesky_reduce)
         CALL cp_fm_cholesky_reduce(fm_h, fm_s)
      CASE DEFAULT
         CALL cp__b("preconditioner_makes.F",262,"cholesky type not implemented")
      END SELECT

      CALL choose_eigv_solver(fm_h, fm, evals)

      SELECT CASE (preconditioner_env%cholesky_use)
      CASE (cholesky_inverse)
         CALL cp_fm_triangular_multiply(fm_s, fm, side="L", transpose_tr=.FALSE., &
                                        invert_tr=.FALSE., uplo_tr="U", n_rows=n, n_cols=n, alpha=1.0_dp)
         DO i = 1, n
            evals(i) = 1.0_dp/MAX(evals(i)-energy_homo, energy_gap)
         ENDDO
         CALL cp_fm_to_fm(fm, fm_h)
      CASE (cholesky_reduce)
         CALL cp_fm_cholesky_restore(fm, n, fm_s, fm_h, "SOLVE")
         DO i = 1, n
            evals(i) = 1.0_dp/MAX(evals(i)-energy_homo, energy_gap)
         ENDDO
         CALL cp_fm_to_fm(fm_h, fm)
      END SELECT

      CALL cp_fm_column_scale(fm, evals)
      CALL cp_gemm('N', 'T', n, n, n, 1.0_dp, fm, fm_h, 0.0_dp, fm_s)
      CALL cp_fm_to_fm(fm_s, fm)

      DEALLOCATE (evals)
      CALL cp_fm_release(fm_h)
      CALL cp_fm_release(fm_s)

      CALL timestop(handle)

   END SUBROUTINE make_full_single

! **************************************************************************************************
!> \brief full single in the orthonormal basis
!> \param preconditioner_env ...
!> \param fm ...
!> \param matrix_h ...
!> \param energy_homo ...
!> \param energy_gap ...
! **************************************************************************************************
   SUBROUTINE make_full_single_ortho(preconditioner_env, fm, matrix_h, &
                                     energy_homo, energy_gap)
      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_fm_type), POINTER                          :: fm
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h
      REAL(KIND=dp)                                      :: energy_homo, energy_gap

      CHARACTER(len=*), PARAMETER :: routineN = 'make_full_single_ortho', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, n
      REAL(KIND=dp), DIMENSION(:), POINTER               :: evals
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: fm_h, fm_s

      CALL timeset(routineN, handle)
      NULLIFY (fm_h, fm_s, fm_struct_tmp, evals)

      IF (ASSOCIATED(fm)) THEN
         CALL cp_fm_release(fm)
      ENDIF
      CALL cp_dbcsr_get_info(matrix_h, nfullrows_total=n)
      ALLOCATE (evals(n))

      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=n, ncol_global=n, &
                               context=preconditioner_env%ctxt, &
                               para_env=preconditioner_env%para_env)
      CALL cp_fm_create(fm, fm_struct_tmp, name="preconditioner")
      CALL cp_fm_create(fm_h, fm_struct_tmp, name="fm_h")
      CALL cp_fm_create(fm_s, fm_struct_tmp, name="fm_s")
      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL copy_dbcsr_to_fm(matrix_h, fm_h)

      CALL choose_eigv_solver(fm_h, fm, evals)
      DO i = 1, n
         evals(i) = 1.0_dp/MAX(evals(i)-energy_homo, energy_gap)
      ENDDO
      CALL cp_fm_to_fm(fm, fm_h)
      CALL cp_fm_column_scale(fm, evals)
      CALL cp_gemm('N', 'T', n, n, n, 1.0_dp, fm, fm_h, 0.0_dp, fm_s)
      CALL cp_fm_to_fm(fm_s, fm)

      DEALLOCATE (evals)
      CALL cp_fm_release(fm_h)
      CALL cp_fm_release(fm_s)

      CALL timestop(handle)

   END SUBROUTINE make_full_single_ortho

! **************************************************************************************************
!> \brief generates a state by state preconditioner based on the full hamiltonian matrix
!> \param preconditioner_env ...
!> \param matrix_c0 ...
!> \param matrix_h ...
!> \param matrix_s ...
!> \param c0_evals ...
!> \param energy_gap should be a slight underestimate of the physical energy gap for almost all systems
!>      the c0 are already ritz states of (h,s)
!> \par History
!>      10.2006 made more stable [Joost VandeVondele]
!> \note
!>      includes error estimate on the hamiltonian matrix to result in a stable preconditioner
!>      a preconditioner for each eigenstate i is generated by keeping the factorized form
!>      U diag( something i ) U^T. It is important to only precondition in the subspace orthogonal to c0.
!>      not only is it the only part that matters, it also simplifies the computation of
!>      the lagrangian multipliers in the OT minimization  (i.e. if the c0 here is different
!>      from the c0 used in the OT setup, there will be a bug).
! **************************************************************************************************
   SUBROUTINE make_full_all(preconditioner_env, matrix_c0, matrix_h, matrix_s, c0_evals, energy_gap)
      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_fm_type), POINTER                          :: matrix_c0
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h, matrix_s
      REAL(KIND=dp), DIMENSION(:), POINTER               :: c0_evals
      REAL(KIND=dp)                                      :: energy_gap

      CHARACTER(len=*), PARAMETER :: routineN = 'make_full_all', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: fudge_factor = 0.25_dp, &
                                                            lambda_base = 10.0_dp

      INTEGER                                            :: handle, k, n
      REAL(KIND=dp)                                      :: error_estimate, lambda
      REAL(KIND=dp), DIMENSION(:), POINTER               :: diag, norms, shifted_evals
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: matrix_hc0, matrix_left, matrix_pre, &
                                                            matrix_s1, matrix_s2, matrix_sc0, &
                                                            matrix_shc0, matrix_tmp, ortho

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(preconditioner_env%fm)) CALL cp_fm_release(preconditioner_env%fm)
      CALL cp_fm_get_info(matrix_c0, nrow_global=n, ncol_global=k)
      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=n, ncol_global=n, &
                               context=preconditioner_env%ctxt, &
                               para_env=preconditioner_env%para_env)
      CALL cp_fm_create(preconditioner_env%fm, fm_struct_tmp, name="preconditioner_env%fm")
      matrix_pre => preconditioner_env%fm
      CALL cp_fm_create(ortho, fm_struct_tmp, name="ortho")
      CALL cp_fm_create(matrix_tmp, fm_struct_tmp, name="matrix_tmp")
      CALL cp_fm_struct_release(fm_struct_tmp)
      ALLOCATE (preconditioner_env%full_evals(n))
      ALLOCATE (preconditioner_env%occ_evals(k))

      ! 0) cholesky decompose the overlap matrix, if this fails the basis is singular,
      !    more than EPS_DEFAULT
      CALL copy_dbcsr_to_fm(matrix_s, ortho)
      CALL cp_fm_cholesky_decompose(ortho)
! if cho inverse
      IF (preconditioner_env%cholesky_use == cholesky_inverse) THEN
         CALL cp_fm_triangular_invert(ortho)
      END IF
      ! 1) Construct a new H matrix, which has the current C0 as eigenvectors,
      !    possibly shifted by an amount lambda,
      !    and the same spectrum as the original H matrix in the space orthogonal to the C0
      !    with P=C0 C0 ^ T
      !    (1 - PS)^T H (1-PS) + (PS)^T (H - lambda S ) (PS)
      !    we exploit that the C0 are already the ritz states of H
      CALL cp_fm_create(matrix_sc0, matrix_c0%matrix_struct, name="sc0")
      CALL cp_dbcsr_sm_fm_multiply(matrix_s, matrix_c0, matrix_sc0, k)
      CALL cp_fm_create(matrix_hc0, matrix_c0%matrix_struct, name="hc0")
      CALL cp_dbcsr_sm_fm_multiply(matrix_h, matrix_c0, matrix_hc0, k)

      ! An aside, try to estimate the error on the ritz values, we'll need it later on
      CALL cp_fm_create(matrix_shc0, matrix_c0%matrix_struct, name="shc0")

      SELECT CASE (preconditioner_env%cholesky_use)
      CASE (cholesky_inverse)
! if cho inverse
         CALL cp_fm_to_fm(matrix_hc0, matrix_shc0)
         CALL cp_fm_triangular_multiply(ortho, matrix_shc0, side="L", transpose_tr=.TRUE., &
                                        invert_tr=.FALSE., uplo_tr="U", n_rows=n, n_cols=k, alpha=1.0_dp)
      CASE (cholesky_reduce)
         CALL cp_fm_cholesky_restore(matrix_hc0, k, ortho, matrix_shc0, "SOLVE", transa="T")
      CASE DEFAULT
         CALL cp__b("preconditioner_makes.F",438,"cholesky type not implemented")
      END SELECT
      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=k, ncol_global=k, &
                               context=preconditioner_env%ctxt, &
                               para_env=preconditioner_env%para_env)
      CALL cp_fm_create(matrix_s1, fm_struct_tmp, name="matrix_s1")
      CALL cp_fm_struct_release(fm_struct_tmp)
      ! since we only use diagonal elements this is a bit of a waste
      CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, matrix_shc0, matrix_shc0, 0.0_dp, matrix_s1)
      ALLOCATE (diag(k))
      CALL cp_fm_get_diag(matrix_s1, diag)
      error_estimate = MAXVAL(SQRT(ABS(diag-c0_evals**2)))
      DEALLOCATE (diag)
      CALL cp_fm_release(matrix_s1)
      CALL cp_fm_release(matrix_shc0)
      ! we'll only use the energy gap, if our estimate of the error on the eigenvalues
      ! is small enough. A large error combined with a small energy gap would otherwise lead to
      ! an aggressive but bad preconditioner. Only when the error is small (MD), we can precondition
      ! aggressively
      preconditioner_env%energy_gap = MAX(energy_gap, error_estimate*fudge_factor)
      CALL copy_dbcsr_to_fm(matrix_h, matrix_tmp)
      CALL cp_fm_upper_to_full(matrix_tmp, matrix_pre)
      ! tmp = H ( 1 - PS )
      CALL cp_gemm('N', 'T', n, n, k, -1.0_dp, matrix_hc0, matrix_sc0, 1.0_dp, matrix_tmp)

      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=k, ncol_global=n, &
                               context=preconditioner_env%ctxt, &
                               para_env=preconditioner_env%para_env)
      CALL cp_fm_create(matrix_left, fm_struct_tmp, name="matrix_left")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_gemm('T', 'N', k, n, n, 1.0_dp, matrix_c0, matrix_tmp, 0.0_dp, matrix_left)
      ! tmp = (1 - PS)^T H (1-PS)
      CALL cp_gemm('N', 'N', n, n, k, -1.0_dp, matrix_sc0, matrix_left, 1.0_dp, matrix_tmp)
      CALL cp_fm_release(matrix_left)

      ALLOCATE (shifted_evals(k))
      lambda = lambda_base+error_estimate
      shifted_evals = c0_evals-lambda
      CALL cp_fm_to_fm(matrix_sc0, matrix_hc0)
      CALL cp_fm_column_scale(matrix_hc0, shifted_evals)
      CALL cp_gemm('N', 'T', n, n, k, 1.0_dp, matrix_hc0, matrix_sc0, 1.0_dp, matrix_tmp)

      ! 2) diagonalize this operator
      SELECT CASE (preconditioner_env%cholesky_use)
      CASE (cholesky_inverse)
         CALL cp_fm_triangular_multiply(ortho, matrix_tmp, side="R", transpose_tr=.FALSE., &
                                        invert_tr=.FALSE., uplo_tr="U", n_rows=n, n_cols=n, alpha=1.0_dp)
         CALL cp_fm_triangular_multiply(ortho, matrix_tmp, side="L", transpose_tr=.TRUE., &
                                        invert_tr=.FALSE., uplo_tr="U", n_rows=n, n_cols=n, alpha=1.0_dp)
      CASE (cholesky_reduce)
         CALL cp_fm_cholesky_reduce(matrix_tmp, ortho)
      END SELECT
      CALL choose_eigv_solver(matrix_tmp, matrix_pre, preconditioner_env%full_evals)
      SELECT CASE (preconditioner_env%cholesky_use)
      CASE (cholesky_inverse)
         CALL cp_fm_triangular_multiply(ortho, matrix_pre, side="L", transpose_tr=.FALSE., &
                                        invert_tr=.FALSE., uplo_tr="U", n_rows=n, n_cols=n, alpha=1.0_dp)
         CALL cp_fm_to_fm(matrix_pre, matrix_tmp)
      CASE (cholesky_reduce)
         CALL cp_fm_cholesky_restore(matrix_pre, n, ortho, matrix_tmp, "SOLVE")
         CALL cp_fm_to_fm(matrix_tmp, matrix_pre)
      END SELECT

      ! test that the subspace remained conserved
      IF (.FALSE.) THEN
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=k, ncol_global=k, &
                                  context=preconditioner_env%ctxt, &
                                  para_env=preconditioner_env%para_env)
         CALL cp_fm_create(matrix_s1, fm_struct_tmp, name="matrix_s1")
         CALL cp_fm_create(matrix_s2, fm_struct_tmp, name="matrix_s2")
         CALL cp_fm_struct_release(fm_struct_tmp)
         ALLOCATE (norms(k))
         CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, matrix_sc0, matrix_tmp, 0.0_dp, matrix_s1)
         CALL choose_eigv_solver(matrix_s1, matrix_s2, norms)
         WRITE (*, *) "matrix norm deviation (should be close to zero): ", MAXVAL(ABS(ABS(norms)-1.0_dp))
         DEALLOCATE (norms)
         CALL cp_fm_release(matrix_s1)
         CALL cp_fm_release(matrix_s2)
      ENDIF

      ! 3) replace the lowest k evals and evecs with what they should be
      preconditioner_env%occ_evals = c0_evals
      ! notice, this choice causes the preconditioner to be constant when applied to sc0 (see apply_full_all)
      preconditioner_env%full_evals(1:k) = c0_evals
      CALL cp_fm_to_fm(matrix_c0, matrix_pre, k, 1, 1)

      CALL cp_fm_release(matrix_sc0)
      CALL cp_fm_release(matrix_hc0)
      CALL cp_fm_release(ortho)
      CALL cp_fm_release(matrix_tmp)
      DEALLOCATE (shifted_evals)
      CALL timestop(handle)

   END SUBROUTINE make_full_all

! **************************************************************************************************
!> \brief full all in the orthonormal basis
!> \param preconditioner_env ...
!> \param matrix_c0 ...
!> \param matrix_h ...
!> \param c0_evals ...
!> \param energy_gap ...
! **************************************************************************************************
   SUBROUTINE make_full_all_ortho(preconditioner_env, matrix_c0, matrix_h, c0_evals, energy_gap)

      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_fm_type), POINTER                          :: matrix_c0
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h
      REAL(KIND=dp), DIMENSION(:), POINTER               :: c0_evals
      REAL(KIND=dp)                                      :: energy_gap

      CHARACTER(len=*), PARAMETER :: routineN = 'make_full_all_ortho', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: fudge_factor = 0.25_dp, &
                                                            lambda_base = 10.0_dp

      INTEGER                                            :: handle, k, n
      REAL(KIND=dp)                                      :: error_estimate, lambda
      REAL(KIND=dp), DIMENSION(:), POINTER               :: diag, norms, shifted_evals
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: matrix_hc0, matrix_left, matrix_pre, &
                                                            matrix_s1, matrix_s2, matrix_sc0, &
                                                            matrix_tmp

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(preconditioner_env%fm)) CALL cp_fm_release(preconditioner_env%fm)
      CALL cp_fm_get_info(matrix_c0, nrow_global=n, ncol_global=k)
      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=n, ncol_global=n, &
                               context=preconditioner_env%ctxt, &
                               para_env=preconditioner_env%para_env)
      CALL cp_fm_create(preconditioner_env%fm, fm_struct_tmp, name="preconditioner_env%fm")
      matrix_pre => preconditioner_env%fm
      CALL cp_fm_create(matrix_tmp, fm_struct_tmp, name="matrix_tmp")
      CALL cp_fm_struct_release(fm_struct_tmp)
      ALLOCATE (preconditioner_env%full_evals(n))
      ALLOCATE (preconditioner_env%occ_evals(k))

      ! 1) Construct a new H matrix, which has the current C0 as eigenvectors,
      !    possibly shifted by an amount lambda,
      !    and the same spectrum as the original H matrix in the space orthogonal to the C0
      !    with P=C0 C0 ^ T
      !    (1 - PS)^T H (1-PS) + (PS)^T (H - lambda S ) (PS)
      !    we exploit that the C0 are already the ritz states of H
      CALL cp_fm_create(matrix_sc0, matrix_c0%matrix_struct, name="sc0")
      CALL cp_fm_to_fm(matrix_c0, matrix_sc0)
      CALL cp_fm_create(matrix_hc0, matrix_c0%matrix_struct, name="hc0")
      CALL cp_dbcsr_sm_fm_multiply(matrix_h, matrix_c0, matrix_hc0, k)

      ! An aside, try to estimate the error on the ritz values, we'll need it later on
      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=k, ncol_global=k, &
                               context=preconditioner_env%ctxt, &
                               para_env=preconditioner_env%para_env)
      CALL cp_fm_create(matrix_s1, fm_struct_tmp, name="matrix_s1")
      CALL cp_fm_struct_release(fm_struct_tmp)
      ! since we only use diagonal elements this is a bit of a waste
      CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, matrix_hc0, matrix_hc0, 0.0_dp, matrix_s1)
      ALLOCATE (diag(k))
      CALL cp_fm_get_diag(matrix_s1, diag)
      error_estimate = MAXVAL(SQRT(ABS(diag-c0_evals**2)))
      DEALLOCATE (diag)
      CALL cp_fm_release(matrix_s1)
      ! we'll only use the energy gap, if our estimate of the error on the eigenvalues
      ! is small enough. A large error combined with a small energy gap would otherwise lead to
      ! an aggressive but bad preconditioner. Only when the error is small (MD), we can precondition
      ! aggressively
      preconditioner_env%energy_gap = MAX(energy_gap, error_estimate*fudge_factor)

      CALL copy_dbcsr_to_fm(matrix_h, matrix_tmp)
      CALL cp_fm_upper_to_full(matrix_tmp, matrix_pre)
      ! tmp = H ( 1 - PS )
      CALL cp_gemm('N', 'T', n, n, k, -1.0_dp, matrix_hc0, matrix_sc0, 1.0_dp, matrix_tmp)

      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=k, ncol_global=n, &
                               context=preconditioner_env%ctxt, &
                               para_env=preconditioner_env%para_env)
      CALL cp_fm_create(matrix_left, fm_struct_tmp, name="matrix_left")
      CALL cp_fm_struct_release(fm_struct_tmp)
      CALL cp_gemm('T', 'N', k, n, n, 1.0_dp, matrix_c0, matrix_tmp, 0.0_dp, matrix_left)
      ! tmp = (1 - PS)^T H (1-PS)
      CALL cp_gemm('N', 'N', n, n, k, -1.0_dp, matrix_sc0, matrix_left, 1.0_dp, matrix_tmp)
      CALL cp_fm_release(matrix_left)

      ALLOCATE (shifted_evals(k))
      lambda = lambda_base+error_estimate
      shifted_evals = c0_evals-lambda
      CALL cp_fm_to_fm(matrix_sc0, matrix_hc0)
      CALL cp_fm_column_scale(matrix_hc0, shifted_evals)
      CALL cp_gemm('N', 'T', n, n, k, 1.0_dp, matrix_hc0, matrix_sc0, 1.0_dp, matrix_tmp)

      ! 2) diagonalize this operator
      CALL choose_eigv_solver(matrix_tmp, matrix_pre, preconditioner_env%full_evals)

      ! test that the subspace remained conserved
      IF (.FALSE.) THEN
         CALL cp_fm_to_fm(matrix_pre, matrix_tmp)
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=k, ncol_global=k, &
                                  context=preconditioner_env%ctxt, &
                                  para_env=preconditioner_env%para_env)
         CALL cp_fm_create(matrix_s1, fm_struct_tmp, name="matrix_s1")
         CALL cp_fm_create(matrix_s2, fm_struct_tmp, name="matrix_s2")
         CALL cp_fm_struct_release(fm_struct_tmp)
         ALLOCATE (norms(k))
         CALL cp_gemm('T', 'N', k, k, n, 1.0_dp, matrix_sc0, matrix_tmp, 0.0_dp, matrix_s1)
         CALL choose_eigv_solver(matrix_s1, matrix_s2, norms)

         WRITE (*, *) "matrix norm deviation (should be close to zero): ", MAXVAL(ABS(ABS(norms)-1.0_dp))
         DEALLOCATE (norms)
         CALL cp_fm_release(matrix_s1)
         CALL cp_fm_release(matrix_s2)
      ENDIF

      ! 3) replace the lowest k evals and evecs with what they should be
      preconditioner_env%occ_evals = c0_evals
      ! notice, this choice causes the preconditioner to be constant when applied to sc0 (see apply_full_all)
      preconditioner_env%full_evals(1:k) = c0_evals
      CALL cp_fm_to_fm(matrix_c0, matrix_pre, k, 1, 1)

      CALL cp_fm_release(matrix_sc0)
      CALL cp_fm_release(matrix_hc0)
      CALL cp_fm_release(matrix_tmp)
      DEALLOCATE (shifted_evals)

      CALL timestop(handle)

   END SUBROUTINE make_full_all_ortho

! **************************************************************************************************
!> \brief generates a preconditioner matrix H-lambda S+(SC)(2.0*CT*H*C+delta)(SC)^T
!>        for later inversion.
!>        H is the Kohn Sham matrix
!>        lambda*S shifts the spectrum of the generalized form up by lambda
!>        the last term only shifts the occupied space (reversing them in energy order)
!>        This form is implicitely multiplied from both sides by S^0.5
!>        This ensures we precondition the correct quantity
!>        Before this reads S^-0.5 H S^-0.5 + lambda + (S^0.5 C)shifts(S^0.5 C)T
!>        which might be a bit more obvious
!>        Replaced the old full_single_inverse at revision 14616
!> \param preconditioner_env the preconditioner env
!> \param matrix_c0 the MO coefficient matrix (fm)
!> \param matrix_h Kohn-Sham matrix (dbcsr)
!> \param energy_gap an additional shift in lambda=-E_homo+energy_gap
!> \param matrix_s the overlap matrix if not orthonormal (dbcsr, optional)
! **************************************************************************************************
   SUBROUTINE make_full_single_inverse(preconditioner_env, matrix_c0, matrix_h, energy_gap, matrix_s)
      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_fm_type), POINTER                          :: matrix_c0
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h
      REAL(KIND=dp)                                      :: energy_gap
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: matrix_s

      CHARACTER(len=*), PARAMETER :: routineN = 'make_full_single_inverse', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, k, n
      REAL(KIND=dp)                                      :: max_ev, min_ev, pre_shift
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrices
      TYPE(cp_dbcsr_type), TARGET                        :: dbcsr_cThc, dbcsr_hc, dbcsr_sc, mo_dbcsr
      TYPE(dbcsr_arnoldi_data)                           :: my_arnoldi

      CALL timeset(routineN, handle)

      ! Allocate all working matrices needed
      CALL cp_fm_get_info(matrix_c0, nrow_global=n, ncol_global=k)
      CALL cp_dbcsr_init(mo_dbcsr)
      ! copy the fm MO's to a sparse matrix, can be solved better if the sparse version is already present
      ! but for the time beeing this will do
      CALL cp_fm_to_dbcsr_row_template(mo_dbcsr, matrix_c0, matrix_h)
      CALL cp_dbcsr_init(dbcsr_sc)
      CALL cp_dbcsr_create(dbcsr_sc, template=mo_dbcsr)
      CALL cp_dbcsr_init(dbcsr_hc)
      CALL cp_dbcsr_create(dbcsr_hc, template=mo_dbcsr)
      CALL cp_dbcsr_init(dbcsr_cThc)
      CALL cp_dbcsr_m_by_n_from_template(dbcsr_cThc, matrix_h, k, k, sym=dbcsr_type_symmetric)

      ! Check whether the output matrix was already created, if not do it now
      IF (.NOT. ASSOCIATED(preconditioner_env%sparse_matrix)) THEN
         ALLOCATE (preconditioner_env%sparse_matrix)
         CALL cp_dbcsr_init(preconditioner_env%sparse_matrix)
      END IF

      ! Put the first term of the preconditioner (H) into the output matrix
      CALL cp_dbcsr_copy(preconditioner_env%sparse_matrix, matrix_h)

      ! Precompute some matrices
      ! S*C, if orthonormal this will be simply C so a copy will do
      IF (PRESENT(matrix_s)) THEN
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_s, mo_dbcsr, 0.0_dp, dbcsr_sc)
      ELSE
         CALL cp_dbcsr_copy(dbcsr_sc, mo_dbcsr)
      END IF

!----------------------------compute the occupied subspace and shift it ------------------------------------
      ! cT*H*C which will be used to shift the occupied states to 0
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_h, mo_dbcsr, 0.0_dp, dbcsr_hc)
      CALL cp_dbcsr_multiply("T", "N", 1.0_dp, mo_dbcsr, dbcsr_hc, 0.0_dp, dbcsr_cThc)

      ! Compute the Energy of the HOMO. We will use this as a reference energy
      ALLOCATE (matrices(1))
      matrices(1)%matrix => dbcsr_cThc
      CALL cp_dbcsr_setup_arnoldi_data(my_arnoldi, matrices, max_iter=20, threshold=1.0E-3_dp, selection_crit=2, &
                                       nval_request=1, nrestarts=8, generalized_ev=.FALSE., iram=.FALSE.)
      IF (ASSOCIATED(preconditioner_env%max_ev_vector)) &
         CALL cp_set_arnoldi_initial_vector(my_arnoldi, preconditioner_env%max_ev_vector)
      CALL cp_dbcsr_arnoldi_ev(matrices, my_arnoldi)
      max_ev = REAL(get_selected_ritz_val(my_arnoldi, 1), dp)

      ! save the ev as guess for the next time
      IF (.NOT. ASSOCIATED(preconditioner_env%max_ev_vector)) ALLOCATE (preconditioner_env%max_ev_vector)
      CALL get_selected_ritz_vec(my_arnoldi, 1, matrices(1)%matrix, preconditioner_env%max_ev_vector)
      CALL deallocate_arnoldi_data(my_arnoldi)
      DEALLOCATE (matrices)

      ! Lets shift the occupied states a bit further up, -1.0 because we gonna subtract it from H
      CALL cp_dbcsr_add_on_diag(dbcsr_cThc, -0.5_dp)
      ! Get the AO representation of the shift (see above why S is needed), W-matrix like object
      CALL cp_dbcsr_multiply("N", "N", 2.0_dp, dbcsr_sc, dbcsr_cThc, 0.0_dp, dbcsr_hc)
      CALL cp_dbcsr_multiply("N", "T", -1.0_dp, dbcsr_hc, dbcsr_sc, 1.0_dp, preconditioner_env%sparse_matrix)

!-------------------------------------compute eigenvalues of H ----------------------------------------------
      ! Setup the arnoldi procedure to compute the lowest ev. if S is present this has to be the generalized ev
      IF (PRESENT(matrix_s)) THEN
         ALLOCATE (matrices(2))
         matrices(1)%matrix => preconditioner_env%sparse_matrix
         matrices(2)%matrix => matrix_s
         CALL cp_dbcsr_setup_arnoldi_data(my_arnoldi, matrices, max_iter=20, threshold=2.0E-2_dp, selection_crit=3, &
                                          nval_request=1, nrestarts=21, generalized_ev=.TRUE., iram=.FALSE.)
      ELSE
         ALLOCATE (matrices(1))
         matrices(1)%matrix => preconditioner_env%sparse_matrix
         CALL cp_dbcsr_setup_arnoldi_data(my_arnoldi, matrices, max_iter=20, threshold=2.0E-2_dp, selection_crit=3, &
                                          nval_request=1, nrestarts=8, generalized_ev=.FALSE., iram=.FALSE.)
      END IF
      IF (ASSOCIATED(preconditioner_env%min_ev_vector)) &
         CALL cp_set_arnoldi_initial_vector(my_arnoldi, preconditioner_env%min_ev_vector)

      ! compute the LUMO energy
      CALL cp_dbcsr_arnoldi_ev(matrices, my_arnoldi)
      min_eV = REAL(get_selected_ritz_val(my_arnoldi, 1), dp)

      ! save the lumo vector for restarting in the next step
      IF (.NOT. ASSOCIATED(preconditioner_env%min_ev_vector)) ALLOCATE (preconditioner_env%min_ev_vector)
      CALL get_selected_ritz_vec(my_arnoldi, 1, matrices(1)%matrix, preconditioner_env%min_ev_vector)
      CALL deallocate_arnoldi_data(my_arnoldi)
      DEALLOCATE (matrices)

!-------------------------------------compute eigenvalues of H ----------------------------------------------
      ! Shift the Lumo to the 1.5*the computed energy_gap or the external energy gap value
      ! The factor 1.5 is determined by trying. If the LUMO is positive, enough, just leave it alone
      pre_shift = MAX(1.5_dp*(min_ev-max_ev), energy_gap)
      IF (min_ev .LT. pre_shift) THEN
         pre_shift = pre_shift-min_ev
      ELSE
         pre_shift = 0.0_dp
      END IF
      IF (PRESENT(matrix_s)) THEN
         CALL cp_dbcsr_add(preconditioner_env%sparse_matrix, matrix_s, 1.0_dp, pre_shift)
      ELSE
         CALL cp_dbcsr_add_on_diag(preconditioner_env%sparse_matrix, pre_shift)
      END IF

      CALL cp_dbcsr_release(mo_dbcsr)
      CALL cp_dbcsr_release(dbcsr_hc)
      CALL cp_dbcsr_release(dbcsr_sc)
      CALL cp_dbcsr_release(dbcsr_cThc)

      CALL timestop(handle)

   END SUBROUTINE make_full_single_inverse

END MODULE preconditioner_makes

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief Routines useful for iterative matrix calculations
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE iterate_matrix
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_add_on_diag, cp_dbcsr_arnoldi_ev, cp_dbcsr_arnoldi_extremal, &
        cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_filter, cp_dbcsr_frobenius_norm, &
        cp_dbcsr_gershgorin_norm, cp_dbcsr_get_info, cp_dbcsr_get_matrix_type, &
        cp_dbcsr_get_occupation, cp_dbcsr_init, cp_dbcsr_multiply, cp_dbcsr_norm, cp_dbcsr_p_type, &
        cp_dbcsr_release, cp_dbcsr_scale, cp_dbcsr_set, cp_dbcsr_setup_arnoldi_data, &
        cp_dbcsr_transposed, cp_dbcsr_type, dbcsr_arnoldi_data, dbcsr_norm_maxabsnorm, &
        dbcsr_type_no_symmetry, deallocate_arnoldi_data, get_selected_ritz_val
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE machine,                         ONLY: m_flush,&
                                              m_walltime
   USE mathconstants,                   ONLY: ifac
   USE mathlib,                         ONLY: abnormal_value
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'iterate_matrix'

   INTERFACE purify_mcweeny
      MODULE PROCEDURE purify_mcweeny_orth, purify_mcweeny_nonorth
   END INTERFACE

   PUBLIC :: invert_Hotelling, matrix_sign_Newton_Schulz, matrix_sqrt_Newton_Schulz, &
             purify_mcweeny

CONTAINS

! **************************************************************************************************
!> \brief invert a symmetric positive definite matrix by Hotelling's method
!>        explicit symmetrization makes this code not suitable for other matrix types
!>        Currently a bit messy with the options, to to be cleaned soon
!> \param matrix_inverse ...
!> \param matrix ...
!> \param threshold convergence threshold nased on the max abs
!> \param use_inv_as_guess logical whether input can be used as guess for inverse
!> \param norm_convergence convergence threshold for the 2-norm, useful for approximate solutions
!> \param filter_eps filter_eps for matrix multiplications, if not passed nothing is filteres
!> \param silent ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!>       2011.10 guess option added [Rustam Z Khaliullin]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE invert_Hotelling(matrix_inverse, matrix, threshold, use_inv_as_guess, &
                               norm_convergence, filter_eps, silent)

      TYPE(cp_dbcsr_type), INTENT(INOUT), TARGET         :: matrix_inverse, matrix
      REAL(KIND=dp), INTENT(IN)                          :: threshold
      LOGICAL, INTENT(IN), OPTIONAL                      :: use_inv_as_guess
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: norm_convergence, filter_eps
      LOGICAL, INTENT(IN), OPTIONAL                      :: silent

      CHARACTER(LEN=*), PARAMETER :: routineN = 'invert_Hotelling', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, nrow, unit_nr
      INTEGER(KIND=int_8)                                :: flop1, flop2
      LOGICAL                                            :: use_inv_guess
      REAL(KIND=dp)                                      :: convergence, frob_matrix, &
                                                            gershgorin_norm, max_ev, &
                                                            maxnorm_matrix, min_ev, occ_matrix, &
                                                            t1, t2
      TYPE(cp_dbcsr_p_type), DIMENSION(1)                :: mymat
      TYPE(cp_dbcsr_type), TARGET                        :: tmp1, tmp2
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dbcsr_arnoldi_data)                           :: my_arnoldi

! turn this off for the time being

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF
      IF (PRESENT(silent)) THEN
         IF (silent) unit_nr = -1
      END IF

      convergence = threshold
      IF (PRESENT(norm_convergence)) convergence = norm_convergence

      use_inv_guess = .FALSE.
      IF (PRESENT(use_inv_as_guess)) use_inv_guess = use_inv_as_guess
      IF (.NOT. use_inv_guess) THEN
         ! initialize matrix to unity and use arnoldi to scale it into the convergence range
         gershgorin_norm = cp_dbcsr_gershgorin_norm(matrix)
         frob_matrix = cp_dbcsr_frobenius_norm(matrix)
         CALL cp_dbcsr_set(matrix_inverse, 0.0_dp)
         CALL cp_dbcsr_add_on_diag(matrix_inverse, 1.0_dp)
         ! everything commutes, therefor our all products will be symmetric
         CALL cp_dbcsr_init(tmp1)
         CALL cp_dbcsr_create(tmp1, template=matrix_inverse)
      ELSE
         ! It is unlikely that our guess will commute with the matrix, therefore the first product will
         ! be non symmetric
         CALL cp_dbcsr_init(tmp1)
         CALL cp_dbcsr_create(tmp1, template=matrix_inverse, matrix_type=dbcsr_type_no_symmetry)
      ENDIF

      CALL cp_dbcsr_get_info(matrix, nfullrows_total=nrow)
      CALL cp_dbcsr_init(tmp2)
      CALL cp_dbcsr_create(tmp2, template=matrix_inverse)

      IF (unit_nr > 0) WRITE (unit_nr, *)

      ! scale the approximate inverse to be within the convergence radius
      t1 = m_walltime()
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_inverse, matrix, &
                             0.0_dp, tmp1, flop=flop1, filter_eps=filter_eps)

      mymat(1)%matrix => tmp1
      CALL cp_dbcsr_setup_arnoldi_data(my_arnoldi, mymat, max_iter=30, threshold=1.0E-3_dp, selection_crit=1, &
                                       nval_request=2, nrestarts=2, generalized_ev=.FALSE., iram=.TRUE.)
      CALL cp_dbcsr_arnoldi_ev(mymat, my_arnoldi)
      max_eV = REAL(get_selected_ritz_val(my_arnoldi, 2), dp)
      min_eV = REAL(get_selected_ritz_val(my_arnoldi, 1), dp)
      CALL deallocate_arnoldi_data(my_arnoldi)

      occ_matrix = cp_dbcsr_get_occupation(matrix_inverse)
      ! 2.0 would be the correct scaling howver, we should make sure here, that we are in the convergence radius
      CALL cp_dbcsr_scale(tmp1, 1.9_dp/(min_ev+max_ev))
      CALL cp_dbcsr_scale(matrix_inverse, 1.9_dp/(min_ev+max_ev))
      min_ev = min_ev*1.9_dp/(min_ev+max_ev)

      DO i = 1, 100

         ! tmp1 = S^-1 S

         ! for the convergence check
         !frob_matrix_base=cp_dbcsr_frobenius_norm(tmp1)
         CALL cp_dbcsr_add_on_diag(tmp1, -1.0_dp)
         frob_matrix = cp_dbcsr_frobenius_norm(tmp1)

         CALL cp_dbcsr_norm(tmp1, &
                            dbcsr_norm_maxabsnorm, norm_scalar=maxnorm_matrix)

         CALL cp_dbcsr_add_on_diag(tmp1, +1.0_dp)

         ! tmp2 = S^-1 S S^-1
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmp1, matrix_inverse, 0.0_dp, tmp2, &
                                flop=flop2, filter_eps=filter_eps)
         ! S^-1_{n+1} = 2 S^-1 - S^-1 S S^-1
         CALL cp_dbcsr_add(matrix_inverse, tmp2, 2.0_dp, -1.0_dp)

         CALL cp_dbcsr_filter(matrix_inverse, threshold)
         t2 = m_walltime()
         occ_matrix = cp_dbcsr_get_occupation(matrix_inverse)

         ! use the scalar form of the algorithm to trace the EV
         min_ev = min_ev*(2.0_dp-min_ev)
         IF (PRESENT(norm_convergence)) maxnorm_matrix = ABS(min_eV-1.0_dp)

         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(T6,A,1X,I3,1X,F10.8,E12.3,F12.3,F13.3)') "Hotelling iter", i, occ_matrix, &
               maxnorm_matrix, t2-t1, &
               (flop1+flop2)/(1.0E6_dp*MAX(0.001_dp, t2-t1))
            CALL m_flush(unit_nr)
         ENDIF

         IF (maxnorm_matrix < convergence) EXIT

         ! scale the matrix for improved convergence
         min_ev = min_ev*2.0_dp/(min_ev+1.0_dp)
         CALL cp_dbcsr_scale(matrix_inverse, 2.0_dp/(min_ev+1.0_dp))

         t1 = m_walltime()
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_inverse, matrix, &
                                0.0_dp, tmp1, flop=flop1, filter_eps=filter_eps)

      ENDDO

      ! try to symmetrize the output matrix
      IF (cp_dbcsr_get_matrix_type(matrix_inverse) == dbcsr_type_no_symmetry) THEN
         CALL cp_dbcsr_transposed(tmp2, matrix_inverse)
         CALL cp_dbcsr_add(matrix_inverse, tmp2, 0.5_dp, 0.5_dp)
      END IF

      IF (unit_nr > 0) THEN
         !   WRITE(unit_nr,'(T6,A,1X,I3,1X,F10.8,E12.3)') "Final Hotelling ",i,occ_matrix,&
         !      !frob_matrix/frob_matrix_base
         !      maxnorm_matrix
         WRITE (unit_nr, '()')
         CALL m_flush(unit_nr)
      ENDIF

      CALL cp_dbcsr_release(tmp1)
      CALL cp_dbcsr_release(tmp2)

      CALL timestop(handle)

   END SUBROUTINE invert_Hotelling

! **************************************************************************************************
!> \brief compute the sign a matrix using Newton-Schulz iterations
!> \param matrix_sign ...
!> \param matrix ...
!> \param threshold ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE matrix_sign_Newton_Schulz(matrix_sign, matrix, threshold)

      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_sign, matrix
      REAL(KIND=dp), INTENT(IN)                          :: threshold

      CHARACTER(LEN=*), PARAMETER :: routineN = 'matrix_sign_Newton_Schulz', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: count, handle, i, unit_nr
      INTEGER(KIND=int_8)                                :: flop1, flop2
      REAL(KIND=dp)                                      :: frob_matrix, frob_matrix_base, &
                                                            gersh_matrix, occ_matrix, t1, t2
      TYPE(cp_dbcsr_type)                                :: tmp1, tmp2
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      CALL cp_dbcsr_init(tmp1)
      CALL cp_dbcsr_create(tmp1, template=matrix_sign)

      CALL cp_dbcsr_init(tmp2)
      CALL cp_dbcsr_create(tmp2, template=matrix_sign)

      CALL cp_dbcsr_copy(matrix_sign, matrix)
      CALL cp_dbcsr_filter(matrix_sign, threshold)

      ! scale the matrix to get into the convergence range
      frob_matrix = cp_dbcsr_frobenius_norm(matrix_sign)
      gersh_matrix = cp_dbcsr_gershgorin_norm(matrix_sign)
      CALL cp_dbcsr_scale(matrix_sign, 1/MIN(frob_matrix, gersh_matrix))

      IF (unit_nr > 0) WRITE (unit_nr, *)

      count = 0
      DO i = 1, 100

         t1 = m_walltime()
         ! tmp1 = X * X
         CALL cp_dbcsr_multiply("N", "N", -1.0_dp, matrix_sign, matrix_sign, 0.0_dp, tmp1, &
                                filter_eps=threshold, flop=flop1)

         ! check convergence (frob norm of what should be the identity matrix minus identity matrix)
         frob_matrix_base = cp_dbcsr_frobenius_norm(tmp1)
         CALL cp_dbcsr_add_on_diag(tmp1, +1.0_dp)
         frob_matrix = cp_dbcsr_frobenius_norm(tmp1)

         ! update the above to 3*I-X*X
         CALL cp_dbcsr_add_on_diag(tmp1, +2.0_dp)
         occ_matrix = cp_dbcsr_get_occupation(matrix_sign)

         ! tmp2 = 0.5 * X * (3*I-X*X)
         CALL cp_dbcsr_multiply("N", "N", 0.5_dp, matrix_sign, tmp1, 0.0_dp, tmp2, &
                                filter_eps=threshold, flop=flop2)

         ! done iterating
         ! CALL cp_dbcsr_filter(tmp2,threshold)
         CALL cp_dbcsr_copy(matrix_sign, tmp2)
         t2 = m_walltime()

         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(T6,A,1X,I3,1X,F10.8,E12.3,F12.3,F13.3)') "NS sign iter ", i, occ_matrix, &
               frob_matrix/frob_matrix_base, t2-t1, &
               (flop1+flop2)/(1.0E6_dp*MAX(0.001_dp, t2-t1))
            CALL m_flush(unit_nr)
         ENDIF

         IF (frob_matrix/frob_matrix_base < SQRT(threshold)) EXIT

      ENDDO

      ! this check is not really needed
      CALL cp_dbcsr_multiply("N", "N", +1.0_dp, matrix_sign, matrix_sign, 0.0_dp, tmp1, &
                             filter_eps=threshold)
      frob_matrix_base = cp_dbcsr_frobenius_norm(tmp1)
      CALL cp_dbcsr_add_on_diag(tmp1, -1.0_dp)
      frob_matrix = cp_dbcsr_frobenius_norm(tmp1)
      occ_matrix = cp_dbcsr_get_occupation(matrix_sign)
      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(T6,A,1X,I3,1X,F10.8,E12.3)') "Final NS sign iter", i, occ_matrix, &
            frob_matrix/frob_matrix_base
         WRITE (unit_nr, '()')
         CALL m_flush(unit_nr)
      ENDIF

      CALL cp_dbcsr_release(tmp1)
      CALL cp_dbcsr_release(tmp2)

      CALL timestop(handle)

   END SUBROUTINE matrix_sign_Newton_Schulz

! **************************************************************************************************
!> \brief compute the sqrt of a matrix via the sign function and the corresponding Newton-Schulz iterations
!>        the order of the algorithm should be 2..5, 3 or 5 is recommended
!> \param matrix_sqrt ...
!> \param matrix_sqrt_inv ...
!> \param matrix ...
!> \param threshold ...
!> \param order ...
!> \param eps_lanczos ...
!> \param max_iter_lanczos ...
!> \param converged ...
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE matrix_sqrt_Newton_Schulz(matrix_sqrt, matrix_sqrt_inv, matrix, threshold, order, &
                                        eps_lanczos, max_iter_lanczos, converged)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_sqrt, matrix_sqrt_inv, matrix
      REAL(KIND=dp), INTENT(IN)                          :: threshold
      INTEGER, INTENT(IN)                                :: order
      REAL(KIND=dp), INTENT(IN)                          :: eps_lanczos
      INTEGER, INTENT(IN)                                :: max_iter_lanczos
      LOGICAL, OPTIONAL                                  :: converged

      CHARACTER(LEN=*), PARAMETER :: routineN = 'matrix_sqrt_Newton_Schulz', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, unit_nr
      INTEGER(KIND=int_8)                                :: flop1, flop2, flop3, flop4, flop5
      LOGICAL                                            :: arnoldi_converged
      REAL(KIND=dp)                                      :: a, b, c, conv, d, frob_matrix, &
                                                            frob_matrix_base, max_ev, min_ev, oa, &
                                                            ob, oc, occ_matrix, od, scaling, t1, t2
      TYPE(cp_dbcsr_type)                                :: tmp1, tmp2, tmp3
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      IF (PRESENT(converged)) converged = .FALSE.

      ! for stability symmetry can not be assumed
      CALL cp_dbcsr_init(tmp1)
      CALL cp_dbcsr_create(tmp1, template=matrix, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(tmp2)
      CALL cp_dbcsr_create(tmp2, template=matrix, matrix_type=dbcsr_type_no_symmetry)
      IF (order .GE. 4) THEN
         CALL cp_dbcsr_init(tmp3)
         CALL cp_dbcsr_create(tmp3, template=matrix, matrix_type=dbcsr_type_no_symmetry)
      ENDIF

      CALL cp_dbcsr_set(matrix_sqrt_inv, 0.0_dp)
      CALL cp_dbcsr_add_on_diag(matrix_sqrt_inv, 1.0_dp)
      CALL cp_dbcsr_filter(matrix_sqrt_inv, threshold)
      CALL cp_dbcsr_copy(matrix_sqrt, matrix)

      ! scale the matrix to get into the convergence range
      CALL cp_dbcsr_arnoldi_extremal(matrix_sqrt, max_ev, min_ev, threshold=eps_lanczos, &
                                     max_iter=max_iter_lanczos, converged=arnoldi_converged)
      IF (unit_nr > 0) THEN
         WRITE (unit_nr, *)
         WRITE (unit_nr, '(T6,A,1X,L1,A,E12.3)') "Lanczos converged: ", arnoldi_converged, " threshold:", eps_lanczos
         WRITE (unit_nr, '(T6,A,1X,E12.3,E12.3)') "Est. extremal eigenvalues:", max_ev, min_ev
         WRITE (unit_nr, '(T6,A,1X,E12.3)') "Est. condition number :", max_ev/MAX(min_ev, EPSILON(min_ev))
      ENDIF
      ! conservatively assume we get a relatively large error (100*threshold_lanczos) in the estimates
      ! and adjust the scaling to be on the safe side
      scaling = 2/(max_ev+min_ev+100*eps_lanczos)

      CALL cp_dbcsr_scale(matrix_sqrt, scaling)
      CALL cp_dbcsr_filter(matrix_sqrt, threshold)

      DO i = 1, 100

         t1 = m_walltime()

         ! tmp1 = Zk * Yk - I
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_sqrt_inv, matrix_sqrt, 0.0_dp, tmp1, &
                                filter_eps=threshold, flop=flop1)
         frob_matrix_base = cp_dbcsr_frobenius_norm(tmp1)
         CALL cp_dbcsr_add_on_diag(tmp1, -1.0_dp)

         ! check convergence (frob norm of what should be the identity matrix minus identity matrix)
         frob_matrix = cp_dbcsr_frobenius_norm(tmp1)

         flop4 = 0; flop5 = 0
         SELECT CASE (order)
         CASE (2)
            ! update the above to 0.5*(3*I-Zk*Yk)
            CALL cp_dbcsr_add_on_diag(tmp1, -2.0_dp)
            CALL cp_dbcsr_scale(tmp1, -0.5_dp)
         CASE (3)
            ! tmp2 = tmp1 ** 2
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmp1, tmp1, 0.0_dp, tmp2, &
                                   filter_eps=threshold, flop=flop4)
            ! tmp1 = 1/16 * (16*I-8*tmp1+6*tmp1**2-5*tmp1**3)
            CALL cp_dbcsr_add(tmp1, tmp2, -4.0_dp, 3.0_dp)
            CALL cp_dbcsr_add_on_diag(tmp1, 8.0_dp)
            CALL cp_dbcsr_scale(tmp1, 0.125_dp)
         CASE (4) ! as expensive as case(5), so little need to use it
            ! tmp2 = tmp1 ** 2
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmp1, tmp1, 0.0_dp, tmp2, &
                                   filter_eps=threshold, flop=flop4)
            ! tmp3 = tmp2 * tmp1
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmp2, tmp1, 0.0_dp, tmp3, &
                                   filter_eps=threshold, flop=flop5)
            CALL cp_dbcsr_scale(tmp1, -8.0_dp)
            CALL cp_dbcsr_add_on_diag(tmp1, 16.0_dp)
            CALL cp_dbcsr_add(tmp1, tmp2, 1.0_dp, 6.0_dp)
            CALL cp_dbcsr_add(tmp1, tmp3, 1.0_dp, -5.0_dp)
            CALL cp_dbcsr_scale(tmp1, 1/16.0_dp)
         CASE (5)
            ! Knuth's reformulation to evaluate the polynomial of 4th degree in 2 multiplications
            ! p = y4+A*y3+B*y2+C*y+D
            ! z := y * (y+a); P := (z+y+b) * (z+c) + d.
            ! a=(A-1)/2 ; b=B*(a+1)-C-a*(a+1)*(a+1)
            ! c=B-b-a*(a+1)
            ! d=D-bc
            oa = -40.0_dp/35.0_dp
            ob = 48.0_dp/35.0_dp
            oc = -64.0_dp/35.0_dp
            od = 128.0_dp/35.0_dp
            a = (oa-1)/2
            b = ob*(a+1)-oc-a*(a+1)**2
            c = ob-b-a*(a+1)
            d = od-b*c
            ! tmp2 = tmp1 ** 2 + a * tmp1
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmp1, tmp1, 0.0_dp, tmp2, &
                                   filter_eps=threshold, flop=flop4)
            CALL cp_dbcsr_add(tmp2, tmp1, 1.0_dp, a)
            ! tmp3 = tmp2 + tmp1 + b
            CALL cp_dbcsr_copy(tmp3, tmp2)
            CALL cp_dbcsr_add(tmp3, tmp1, 1.0_dp, 1.0_dp)
            CALL cp_dbcsr_add_on_diag(tmp3, b)
            ! tmp2 = tmp2 + c
            CALL cp_dbcsr_add_on_diag(tmp2, c)
            ! tmp1 = tmp2 * tmp3
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmp2, tmp3, 0.0_dp, tmp1, &
                                   filter_eps=threshold, flop=flop5)
            ! tmp1 = tmp1 + d
            CALL cp_dbcsr_add_on_diag(tmp1, d)
            ! final scale
            CALL cp_dbcsr_scale(tmp1, 35.0_dp/128.0_dp)
         CASE DEFAULT
            CALL cp__b("iterate_matrix.F",471,"")
         END SELECT

         ! tmp2 = Yk * tmp1 = Y(k+1)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_sqrt, tmp1, 0.0_dp, tmp2, &
                                filter_eps=threshold, flop=flop2)
         ! CALL cp_dbcsr_filter(tmp2,threshold)
         CALL cp_dbcsr_copy(matrix_sqrt, tmp2)

         ! tmp2 = tmp1 * Zk = Z(k+1)
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, tmp1, matrix_sqrt_inv, 0.0_dp, tmp2, &
                                filter_eps=threshold, flop=flop3)
         ! CALL cp_dbcsr_filter(tmp2,threshold)
         CALL cp_dbcsr_copy(matrix_sqrt_inv, tmp2)

         occ_matrix = cp_dbcsr_get_occupation(matrix_sqrt_inv)

         ! done iterating
         t2 = m_walltime()

         conv = frob_matrix/frob_matrix_base

         IF (unit_nr > 0) THEN
            WRITE (unit_nr, '(T6,A,1X,I3,1X,F10.8,E12.3,F12.3,F13.3)') "NS sqrt iter ", i, occ_matrix, &
               conv, t2-t1, &
               (flop1+flop2+flop3+flop4+flop5)/(1.0E6_dp*MAX(0.001_dp, t2-t1))
            CALL m_flush(unit_nr)
         ENDIF

         IF (abnormal_value(conv)) &
            CALL cp__b("iterate_matrix.F",501,"conv is an abnormal value (NaN/Inf).")

         IF (conv < SQRT(threshold)) THEN
            IF (PRESENT(converged)) converged = .TRUE.
            EXIT
         ENDIF

      ENDDO

      ! symmetrize the matrices as this is not guaranteed by the algorithm
      CALL cp_dbcsr_transposed(tmp1, matrix_sqrt_inv)
      CALL cp_dbcsr_add(matrix_sqrt_inv, tmp1, 0.5_dp, 0.5_dp)
      CALL cp_dbcsr_transposed(tmp1, matrix_sqrt)
      CALL cp_dbcsr_add(matrix_sqrt, tmp1, 0.5_dp, 0.5_dp)

      ! this check is not really needed
      CALL cp_dbcsr_multiply("N", "N", +1.0_dp, matrix_sqrt_inv, matrix_sqrt, 0.0_dp, tmp1, &
                             filter_eps=threshold)
      frob_matrix_base = cp_dbcsr_frobenius_norm(tmp1)
      CALL cp_dbcsr_add_on_diag(tmp1, -1.0_dp)
      frob_matrix = cp_dbcsr_frobenius_norm(tmp1)
      occ_matrix = cp_dbcsr_get_occupation(matrix_sqrt_inv)
      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(T6,A,1X,I3,1X,F10.8,E12.3)') "Final NS sqrt iter ", i, occ_matrix, &
            frob_matrix/frob_matrix_base
         WRITE (unit_nr, '()')
         CALL m_flush(unit_nr)
      ENDIF

      ! scale to proper end results
      CALL cp_dbcsr_scale(matrix_sqrt, 1/SQRT(scaling))
      CALL cp_dbcsr_scale(matrix_sqrt_inv, SQRT(scaling))

      CALL cp_dbcsr_release(tmp1)
      CALL cp_dbcsr_release(tmp2)
      IF (order .GE. 4) THEN
         CALL cp_dbcsr_release(tmp3)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE matrix_sqrt_Newton_Schulz

! **************************************************************************************************
!> \brief ...
!> \param matrix_exp ...
!> \param matrix ...
!> \param omega ...
!> \param alpha ...
!> \param threshold ...
! **************************************************************************************************
   SUBROUTINE matrix_exponential(matrix_exp, matrix, omega, alpha, threshold)
      ! compute matrix_exp=omega*exp(alpha*matrix)
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_exp, matrix
      REAL(KIND=dp), INTENT(IN)                          :: omega, alpha, threshold

      CHARACTER(LEN=*), PARAMETER :: routineN = 'matrix_exponential', &
         routineP = moduleN//':'//routineN
      REAL(dp), PARAMETER                                :: one = 1.0_dp, toll = 1.E-17_dp, &
                                                            zero = 0.0_dp

      INTEGER                                            :: handle, i, k, unit_nr
      REAL(dp)                                           :: factorial, norm_C, norm_D, norm_scalar
      TYPE(cp_dbcsr_type)                                :: B, B_square, C, D, D_product
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      ! Calculate the norm of the matrix alpha*matrix, and scale it until it is less than 1.0
      norm_scalar = ABS(alpha)*cp_dbcsr_frobenius_norm(matrix)

      ! k=scaling parameter
      k = 1
      DO
         IF ((norm_scalar/2.0_dp**k) <= one) EXIT
         k = k+1
      END DO

      ! copy and scale the input matrix in matrix C and in matrix D
      CALL cp_dbcsr_init(C)
      CALL cp_dbcsr_create(C, template=matrix, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_copy(C, matrix)
      CALL cp_dbcsr_scale(C, alpha_scalar=alpha/2.0_dp**k)

      CALL cp_dbcsr_init(D)
      CALL cp_dbcsr_create(D, template=matrix, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_copy(D, C)

      !   write(*,*)
      !   write(*,*)
      !   CALL cp_dbcsr_print(D, nodata=.FALSE., matlab_format=.TRUE., variable_name="D", unit_nr=6)

      ! set the B matrix as B=Identity+D
      CALL cp_dbcsr_init(B)
      CALL cp_dbcsr_create(B, template=matrix, matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_copy(B, D)
      CALL cp_dbcsr_add_on_diag(B, alpha_scalar=one)

      !   CALL cp_dbcsr_print(B, nodata=.FALSE., matlab_format=.TRUE., variable_name="B", unit_nr=6)

      ! Calculate the norm of C and moltiply by toll to be used as a threshold
      norm_C = toll*cp_dbcsr_frobenius_norm(matrix)

      ! iteration for the trucated taylor series expansion
      CALL cp_dbcsr_init(D_product)
      CALL cp_dbcsr_create(D_product, template=matrix, matrix_type=dbcsr_type_no_symmetry)
      i = 1
      DO
         i = i+1
         ! compute D_product=D*C
         CALL cp_dbcsr_multiply("N", "N", one, D, C, &
                                zero, D_product, filter_eps=threshold)

         ! copy D_product in D
         CALL cp_dbcsr_copy(D, D_product)

         ! calculate B=B+D_product/fat(i)
         factorial = ifac(i)
         CALL cp_dbcsr_add(B, D_product, one, factorial)

         ! check for convergence using the norm of D (copy of the matrix D_product) and C
         norm_D = factorial*cp_dbcsr_frobenius_norm(D)
         IF (norm_D < norm_C) EXIT
      END DO

      ! start the k iteration for the squaring of the matrix
      CALL cp_dbcsr_init(B_square)
      CALL cp_dbcsr_create(B_square, template=matrix, matrix_type=dbcsr_type_no_symmetry)
      DO i = 1, k
         !compute B_square=B*B
         CALL cp_dbcsr_multiply("N", "N", one, B, B, &
                                zero, B_square, filter_eps=threshold)
         ! copy Bsquare in B to iterate
         CALL cp_dbcsr_copy(B, B_square)
      END DO

      ! copy B_square in matrix_exp and
      CALL cp_dbcsr_copy(matrix_exp, B_square)

      ! scale matrix_exp by omega, matrix_exp=omega*B_square
      CALL cp_dbcsr_scale(matrix_exp, alpha_scalar=omega)
      ! write(*,*) alpha,omega

      CALL timestop(handle)

   END SUBROUTINE matrix_exponential

! **************************************************************************************************
!> \brief McWeeny purification of a matrix in the orthonormal basis
!> \param matrix_p Matrix to purify (needs to be almost idempotent already)
!> \param threshold Threshold used as filter_eps and convergence criteria
!> \param max_steps Max number of iterations
!> \par History
!>       2013.01 created [Florian Schiffmann]
!>       2014.07 slightly refactored [Ole Schuett]
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE purify_mcweeny_orth(matrix_p, threshold, max_steps)
      TYPE(cp_dbcsr_type), DIMENSION(:)                  :: matrix_p
      REAL(KIND=dp)                                      :: threshold
      INTEGER                                            :: max_steps

      CHARACTER(LEN=*), PARAMETER :: routineN = 'purify_mcweeny_orth', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ispin, unit_nr
      REAL(KIND=dp)                                      :: frob_norm
      TYPE(cp_dbcsr_type)                                :: matrix_pp, matrix_tmp
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)
      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      CALL cp_dbcsr_init(matrix_pp)
      CALL cp_dbcsr_create(matrix_pp, template=matrix_p(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=matrix_p(1), matrix_type=dbcsr_type_no_symmetry)

      DO ispin = 1, SIZE(matrix_p)
         DO i = 1, max_steps
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_p(ispin), matrix_p(ispin), &
                                   0.0_dp, matrix_pp, filter_eps=threshold)

            ! test convergence
            CALL cp_dbcsr_copy(matrix_tmp, matrix_pp)
            CALL cp_dbcsr_add(matrix_tmp, matrix_p(ispin), 1.0_dp, -1.0_dp)
            frob_norm = cp_dbcsr_frobenius_norm(matrix_tmp) ! tmp = PP - P
            IF (unit_nr > 0) WRITE (unit_nr, '(t3,a,f16.8)') "McWeeny: Deviation of idempotency", frob_norm
            IF (frob_norm < 1000_dp*threshold .AND. i > 1) EXIT

            ! construct new P
            CALL cp_dbcsr_copy(matrix_tmp, matrix_pp)
            CALL cp_dbcsr_multiply("N", "N", -2.0_dp, matrix_pp, matrix_p(ispin), &
                                   3.0_dp, matrix_tmp, filter_eps=threshold)
            CALL cp_dbcsr_copy(matrix_p(ispin), matrix_tmp) ! tmp = 3PP - 2PPP
         END DO
      END DO

      CALL cp_dbcsr_release(matrix_pp)
      CALL cp_dbcsr_release(matrix_tmp)
      CALL timestop(handle)
   END SUBROUTINE purify_mcweeny_orth

! **************************************************************************************************
!> \brief McWeeny purification of a matrix in the non-orthonormal basis
!> \param matrix_p Matrix to purify (needs to be almost idempotent already)
!> \param matrix_s Overlap-Matrix
!> \param threshold Threshold used as filter_eps and convergence criteria
!> \param max_steps Max number of iterations
!> \par History
!>       2013.01 created [Florian Schiffmann]
!>       2014.07 slightly refactored [Ole Schuett]
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE purify_mcweeny_nonorth(matrix_p, matrix_s, threshold, max_steps)
      TYPE(cp_dbcsr_type), DIMENSION(:)                  :: matrix_p
      TYPE(cp_dbcsr_type)                                :: matrix_s
      REAL(KIND=dp)                                      :: threshold
      INTEGER                                            :: max_steps

      CHARACTER(LEN=*), PARAMETER :: routineN = 'purify_mcweeny_nonorth', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ispin, unit_nr
      REAL(KIND=dp)                                      :: frob_norm
      TYPE(cp_dbcsr_type)                                :: matrix_ps, matrix_psp, matrix_test
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      CALL cp_dbcsr_init(matrix_ps)
      CALL cp_dbcsr_create(matrix_ps, template=matrix_p(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_psp)
      CALL cp_dbcsr_create(matrix_psp, template=matrix_p(1), matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_init(matrix_test)
      CALL cp_dbcsr_create(matrix_test, template=matrix_p(1), matrix_type=dbcsr_type_no_symmetry)

      DO ispin = 1, SIZE(matrix_p)
         DO i = 1, max_steps
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_p(ispin), matrix_s, &
                                   0.0_dp, matrix_ps, filter_eps=threshold)
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_ps, matrix_p(ispin), &
                                   0.0_dp, matrix_psp, filter_eps=threshold)

            ! test convergence
            CALL cp_dbcsr_copy(matrix_test, matrix_psp)
            CALL cp_dbcsr_add(matrix_test, matrix_p(ispin), 1.0_dp, -1.0_dp)
            frob_norm = cp_dbcsr_frobenius_norm(matrix_test) ! test = PSP - P
            IF (unit_nr > 0) WRITE (unit_nr, '(t3,a,2f16.8)') "McWeeny: Deviation of idempotency", frob_norm
            IF (frob_norm < 1000_dp*threshold .AND. i > 1) EXIT

            ! construct new P
            CALL cp_dbcsr_copy(matrix_p(ispin), matrix_psp)
            CALL cp_dbcsr_multiply("N", "N", -2.0_dp, matrix_ps, matrix_psp, &
                                   3.0_dp, matrix_p(ispin), filter_eps=threshold)
         END DO
      END DO

      CALL cp_dbcsr_release(matrix_ps)
      CALL cp_dbcsr_release(matrix_psp)
      CALL cp_dbcsr_release(matrix_test)
      CALL timestop(handle)
   END SUBROUTINE purify_mcweeny_nonorth

END MODULE iterate_matrix
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief solves the preconditioner, contains to utility function for
!>        fm<->dbcsr transfers, should be moved soon
!> \par History
!>      - [UB] 2009-05-13 Adding stable approximate inverse (full and sparse)
!> \author Joost VandeVondele (09.2002)
! **************************************************************************************************
MODULE preconditioner_solvers
   USE bibliography,                    ONLY: Schiffmann2015,&
                                              cite_reference
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_arnoldi_ev, cp_dbcsr_col_block_sizes, cp_dbcsr_create, cp_dbcsr_distribution, &
        cp_dbcsr_filter, cp_dbcsr_get_info, cp_dbcsr_get_occupation, cp_dbcsr_init_p, &
        cp_dbcsr_p_type, cp_dbcsr_release, cp_dbcsr_row_block_sizes, cp_dbcsr_setup_arnoldi_data, &
        cp_dbcsr_type, dbcsr_arnoldi_data, dbcsr_type_no_symmetry, dbcsr_type_real_4, &
        dbcsr_type_real_default, deallocate_arnoldi_data, get_selected_ritz_val
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr
   USE cp_fm_basic_linalg,              ONLY: cp_fm_upper_to_full
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose,&
                                              cp_fm_cholesky_invert
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: ot_precond_solver_default,&
                                              ot_precond_solver_direct,&
                                              ot_precond_solver_inv_chol,&
                                              ot_precond_solver_update
   USE iterate_matrix,                  ONLY: invert_Hotelling
   USE kinds,                           ONLY: dp
   USE preconditioner_types,            ONLY: preconditioner_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'preconditioner_solvers'

   PUBLIC :: solve_preconditioner, transfer_fm_to_dbcsr, transfer_dbcsr_to_fm

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param my_solver_type ...
!> \param preconditioner_env ...
!> \param matrix_s ...
!> \param matrix_h ...
!> \param my_mixed_precision ...
! **************************************************************************************************
   SUBROUTINE solve_preconditioner(my_solver_type, preconditioner_env, matrix_s, &
                                   matrix_h, my_mixed_precision)
      INTEGER                                            :: my_solver_type
      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: matrix_s
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h
      LOGICAL                                            :: my_mixed_precision

      CHARACTER(len=*), PARAMETER :: routineN = 'solve_preconditioner', &
         routineP = moduleN//':'//routineN

      REAL(dp)                                           :: occ_matrix

! here comes the solver

      SELECT CASE (my_solver_type)
      CASE (ot_precond_solver_inv_chol)
         !
         ! compute the full inverse
         preconditioner_env%solver = ot_precond_solver_inv_chol
         CALL make_full_inverse_cholesky(preconditioner_env, matrix_s, my_mixed_precision)
      CASE (ot_precond_solver_direct)
         !
         ! prepare for the direct solver
         preconditioner_env%solver = ot_precond_solver_direct
         CALL make_full_fact_cholesky(preconditioner_env, matrix_s, my_mixed_precision)
      CASE (ot_precond_solver_update)
         !
         ! uses an update of the full inverse (needs to be computed the first time)
         ! make sure preconditioner_env is not destroyed in between
         occ_matrix = 1.0_dp
         IF (ASSOCIATED(preconditioner_env%sparse_matrix)) THEN
            IF (preconditioner_env%condition_num < 0.0_dp) &
               CALL estimate_cond_num(preconditioner_env%sparse_matrix, preconditioner_env%condition_num)
            CALL cp_dbcsr_filter(preconditioner_env%sparse_matrix, &
                                 1.0_dp/preconditioner_env%condition_num*0.01_dp)
            occ_matrix = cp_dbcsr_get_occupation(preconditioner_env%sparse_matrix)
         END IF
         ! check whether we are in the first step and if it is a good idea to use cholesky (matrix sparsity)
         IF (preconditioner_env%solver .NE. ot_precond_solver_update .AND. occ_matrix > 0.5_dp) THEN
            preconditioner_env%solver = ot_precond_solver_update
            CALL make_full_inverse_cholesky(preconditioner_env, matrix_s, my_mixed_precision)
         ELSE
            preconditioner_env%solver = ot_precond_solver_update
            CALL make_inverse_update(preconditioner_env, matrix_h, my_mixed_precision)
         END IF
      CASE (ot_precond_solver_default)
         preconditioner_env%solver = ot_precond_solver_default
      CASE DEFAULT
         !
         CALL cp__b("preconditioner_solvers.F",113,"Doesn't know this type of solver")
      END SELECT

   END SUBROUTINE solve_preconditioner

! **************************************************************************************************
!> \brief Compute the inverse using cholseky factorization
!> \param preconditioner_env ...
!> \param matrix_s ...
!> \param mixed_precision ...
! **************************************************************************************************
   SUBROUTINE make_full_inverse_cholesky(preconditioner_env, matrix_s, mixed_precision)

      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: matrix_s
      LOGICAL, INTENT(IN)                                :: mixed_precision

      CHARACTER(len=*), PARAMETER :: routineN = 'make_full_inverse_cholesky', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, info
      TYPE(cp_fm_type), POINTER                          :: fm, fm_work

      CALL timeset(routineN, handle)

      ! Maybe we will get a sparse Cholesky at a given point then this can go,
      ! if stuff was stored in fm anyway this simple returns
      CALL transfer_dbcsr_to_fm(preconditioner_env%sparse_matrix, preconditioner_env%fm, &
                                preconditioner_env%para_env, preconditioner_env%ctxt, &
                                mixed_precision)
      fm => preconditioner_env%fm

      NULLIFY (fm_work)

      CALL cp_fm_create(fm_work, fm%matrix_struct, name="fm_work", use_sp=mixed_precision)
      !
      ! compute the inverse of SPD matrix fm using the Cholesky factorization
      CALL cp_fm_cholesky_decompose(fm, info_out=info)

      !
      ! if fm not SPD we go with the overlap matrix
      IF (info /= 0) THEN
         !
         ! just the overlap matrix
         IF (PRESENT(matrix_s)) THEN
            CALL copy_dbcsr_to_fm(matrix_s, fm)
            CALL cp_fm_cholesky_decompose(fm)
         ELSE
            CALL cp_fm_set_all(fm, alpha=0._dp, beta=1._dp)
         ENDIF
      ENDIF
      CALL cp_fm_cholesky_invert(fm)

      CALL cp_fm_upper_to_full(fm, fm_work)
      CALL cp_fm_release(fm_work)

      CALL timestop(handle)

   END SUBROUTINE make_full_inverse_cholesky

! **************************************************************************************************
!> \brief Only perform the factorization, can be used later to solve the linear
!>        system on the fly
!> \param preconditioner_env ...
!> \param matrix_s ...
!> \param mixed_precision ...
! **************************************************************************************************
   SUBROUTINE make_full_fact_cholesky(preconditioner_env, matrix_s, mixed_precision)

      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: matrix_s
      LOGICAL, INTENT(IN)                                :: mixed_precision

      CHARACTER(len=*), PARAMETER :: routineN = 'make_full_fact_cholesky', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, info_out
      TYPE(cp_fm_type), POINTER                          :: fm

      CALL timeset(routineN, handle)

      ! Maybe we will get a sparse Cholesky at a given point then this can go,
      ! if stuff was stored in fm anyway this simple returns
      CALL transfer_dbcsr_to_fm(preconditioner_env%sparse_matrix, preconditioner_env%fm, &
                                preconditioner_env%para_env, preconditioner_env%ctxt, &
                                mixed_precision)

      fm => preconditioner_env%fm
      !
      ! compute the inverse of SPD matrix fm using the Cholesky factorization
      CALL cp_fm_cholesky_decompose(fm, info_out=info_out)
      !
      ! if fm not SPD we go with the overlap matrix
      IF (info_out .NE. 0) THEN
         !
         ! just the overlap matrix
         IF (PRESENT(matrix_s)) THEN
            CALL copy_dbcsr_to_fm(matrix_s, fm)
            CALL cp_fm_cholesky_decompose(fm)
         ELSE
            CALL cp_fm_set_all(fm, alpha=0._dp, beta=1._dp)
         ENDIF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE make_full_fact_cholesky

! **************************************************************************************************
!> \brief computes an approximate inverse using Hotelling iterations
!> \param preconditioner_env ...
!> \param matrix_h as S is not always present this is a safe template for the transfer
!> \param mixed_precision ...
! **************************************************************************************************
   SUBROUTINE make_inverse_update(preconditioner_env, matrix_h, mixed_precision)
      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h
      LOGICAL, INTENT(IN)                                :: mixed_precision

      CHARACTER(len=*), PARAMETER :: routineN = 'make_inverse_update', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: use_guess
      REAL(KIND=dp)                                      :: filter_eps

      CALL timeset(routineN, handle)
      use_guess = .TRUE.
      !
      ! uses an update of the full inverse (needs to be computed the first time)
      ! make sure preconditioner_env is not destroyed in between

      CALL cite_reference(Schiffmann2015)

      ! Maybe I gonna add a fm Hotelling, ... for now the same as above make sure we are dbcsr
      CALL transfer_fm_to_dbcsr(preconditioner_env%fm, preconditioner_env%sparse_matrix, &
                                matrix_h, mixed_precision)
      IF (.NOT. ASSOCIATED(preconditioner_env%dbcsr_matrix)) THEN
         use_guess = .FALSE.
         CALL cp_dbcsr_init_p(preconditioner_env%dbcsr_matrix)
         CALL cp_dbcsr_create(preconditioner_env%dbcsr_matrix, "prec_dbcsr", &
                              template=matrix_h, matrix_type=dbcsr_type_no_symmetry)
      END IF

      ! Try to get a reasonbale guess for the filtering threshold
      filter_eps = 1.0_dp/preconditioner_env%condition_num*0.1_dp

      ! Agressive filtering on the initial guess is needed to avoid fill ins and retain sparsity
      CALL cp_dbcsr_filter(preconditioner_env%dbcsr_matrix, filter_eps*100.0_dp)
      ! We don't need a high accuracy for the inverse so 0.4 is reasonable for convergence
      CALL invert_Hotelling(preconditioner_env%dbcsr_matrix, preconditioner_env%sparse_matrix, filter_eps*10.0_dp, &
                            use_inv_as_guess=use_guess, norm_convergence=0.4_dp, filter_eps=filter_eps)

      CALL timestop(handle)

   END SUBROUTINE make_inverse_update

! **************************************************************************************************
!> \brief computes an approximation to the condition number of a matrix using
!>        arnoldi iterations
!> \param matrix ...
!> \param cond_num ...
! **************************************************************************************************
   SUBROUTINE estimate_cond_num(matrix, cond_num)
      TYPE(cp_dbcsr_type), POINTER                       :: matrix
      REAL(KIND=dp)                                      :: cond_num

      CHARACTER(len=*), PARAMETER :: routineN = 'estimate_cond_num', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: max_ev, min_ev
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrices
      TYPE(dbcsr_arnoldi_data)                           :: my_arnoldi

      CALL timeset(routineN, handle)

      ! its better to do 2 calculations as the maximum should quickly converge and the minimum won't need iram
      ALLOCATE (matrices(1))
      matrices(1)%matrix => matrix
      ! compute the minimum ev
      CALL cp_dbcsr_setup_arnoldi_data(my_arnoldi, matrices, max_iter=20, threshold=5.0E-4_dp, selection_crit=2, &
                                       nval_request=1, nrestarts=15, generalized_ev=.FALSE., iram=.FALSE.)
      CALL cp_dbcsr_arnoldi_ev(matrices, my_arnoldi)
      max_ev = REAL(get_selected_ritz_val(my_arnoldi, 1), dp)
      CALL deallocate_arnoldi_data(my_arnoldi)

      CALL cp_dbcsr_setup_arnoldi_data(my_arnoldi, matrices, max_iter=20, threshold=5.0E-4_dp, selection_crit=3, &
                                       nval_request=1, nrestarts=15, generalized_ev=.FALSE., iram=.FALSE.)
      CALL cp_dbcsr_arnoldi_ev(matrices, my_arnoldi)
      min_ev = REAL(get_selected_ritz_val(my_arnoldi, 1), dp)
      CALL deallocate_arnoldi_data(my_arnoldi)

      cond_num = max_ev/min_ev
      DEALLOCATE (matrices)

      CALL timestop(handle)
   END SUBROUTINE estimate_cond_num

!   ***************************************************************************
!> \brief transfers a full matrix to a dbcsr
!> \param fm_matrix a full matrix gets deallocated in the end
!> \param dbcsr_matrix a dbcsr matrix, gets create from a template
!> \param template_mat the template which is used for the structure
!> \param mixed_precision  whether to use single or double precision
! **************************************************************************************************
   SUBROUTINE transfer_fm_to_dbcsr(fm_matrix, dbcsr_matrix, template_mat, mixed_precision)

      TYPE(cp_fm_type), POINTER                          :: fm_matrix
      TYPE(cp_dbcsr_type), POINTER                       :: dbcsr_matrix, template_mat
      LOGICAL                                            :: mixed_precision

      CHARACTER(len=*), PARAMETER :: routineN = 'transfer_fm_to_dbcsr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      IF (ASSOCIATED(fm_matrix)) THEN
         IF (.NOT. ASSOCIATED(dbcsr_matrix)) THEN
            CALL cp_dbcsr_init_p(dbcsr_matrix)
            IF (mixed_precision) THEN
               CALL cp_dbcsr_create(dbcsr_matrix, "transfered_matrix", &
                                    cp_dbcsr_distribution(template_mat), dbcsr_type_no_symmetry, &
                                    cp_dbcsr_row_block_sizes(template_mat), cp_dbcsr_col_block_sizes(template_mat), &
                                    nze=0, data_type=dbcsr_type_real_4)
            ELSE
               CALL cp_dbcsr_create(dbcsr_matrix, "preconditioner_env%dbcsr_matrix", &
                                    cp_dbcsr_distribution(template_mat), dbcsr_type_no_symmetry, &
                                    cp_dbcsr_row_block_sizes(template_mat), cp_dbcsr_col_block_sizes(template_mat), &
                                    nze=0, data_type=dbcsr_type_real_default)
            ENDIF
         ENDIF
!       CALL cp_fm_create(fm_tmp, matrix_struct=fm_matrix%matrix_struct)
!       CALL cp_fm_upper_to_full(fm_matrix, fm_tmp)
         CALL copy_fm_to_dbcsr(fm_matrix, dbcsr_matrix)
!       CALL cp_fm_release(fm_tmp)
         CALL cp_fm_release(fm_matrix)
      END IF

      CALL timestop(handle)

   END SUBROUTINE transfer_fm_to_dbcsr

! **************************************************************************************************
!> \brief transfers a dbcsr to a full matrix
!> \param dbcsr_matrix a dbcsr matrix, gets deallocated at the end
!> \param fm_matrix a full matrix gets created if not yet done
!> \param para_env the para_env
!> \param context the blacs context
!> \param mixed_precision  whether to use single or double precision
! **************************************************************************************************
   SUBROUTINE transfer_dbcsr_to_fm(dbcsr_matrix, fm_matrix, para_env, context, mixed_precision)

      TYPE(cp_dbcsr_type), POINTER                       :: dbcsr_matrix
      TYPE(cp_fm_type), POINTER                          :: fm_matrix
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_blacs_env_type), POINTER                   :: context
      LOGICAL                                            :: mixed_precision

      CHARACTER(len=*), PARAMETER :: routineN = 'transfer_dbcsr_to_fm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, n
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp

      CALL timeset(routineN, handle)
      IF (ASSOCIATED(dbcsr_matrix)) THEN
         NULLIFY (fm_struct_tmp)

         IF (ASSOCIATED(fm_matrix)) THEN
            CALL cp_fm_release(fm_matrix)
         ENDIF

         CALL cp_dbcsr_get_info(dbcsr_matrix, nfullrows_total=n)
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=n, ncol_global=n, &
                                  context=context, para_env=para_env)
         CALL cp_fm_create(fm_matrix, fm_struct_tmp, use_sp=mixed_precision)
         CALL cp_fm_struct_release(fm_struct_tmp)

         CALL copy_dbcsr_to_fm(dbcsr_matrix, fm_matrix)
         CALL cp_dbcsr_release(dbcsr_matrix)
         DEALLOCATE (dbcsr_matrix)
      END IF

      CALL timestop(handle)

   END SUBROUTINE transfer_dbcsr_to_fm

END MODULE preconditioner_solvers
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief computes preconditioners, and implements methods to apply them
!>      currently used in qs_ot
!> \par History
!>      - [UB] 2009-05-13 Adding stable approximate inverse (full and sparse)
!> \author Joost VandeVondele (09.2002)
! **************************************************************************************************
MODULE preconditioner_apply
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_copy, cp_dbcsr_init, cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, &
        cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, &
        cp_dbcsr_multiply, cp_dbcsr_release, cp_dbcsr_type
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_restore
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_info,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE input_constants,                 ONLY: ot_precond_full_all,&
                                              ot_precond_full_kinetic,&
                                              ot_precond_full_single,&
                                              ot_precond_full_single_inverse,&
                                              ot_precond_s_inverse,&
                                              ot_precond_solver_direct,&
                                              ot_precond_solver_inv_chol,&
                                              ot_precond_solver_update
   USE kinds,                           ONLY: dp
   USE preconditioner_types,            ONLY: preconditioner_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'preconditioner_apply'

   PUBLIC :: apply_preconditioner_fm, apply_preconditioner_dbcsr

CONTAINS

! **************************************************************************************************
!> \brief applies a previously created preconditioner to a full matrix
!> \param preconditioner_env ...
!> \param matrix_in ...
!> \param matrix_out ...
! **************************************************************************************************
   SUBROUTINE apply_preconditioner_fm(preconditioner_env, matrix_in, matrix_out)

      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_fm_type), POINTER                          :: matrix_in, matrix_out

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_preconditioner_fm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      SELECT CASE (preconditioner_env%in_use)
      CASE (0)
         CALL cp__b("preconditioner_apply.F",65,"No preconditioner in use")
      CASE (ot_precond_full_single)
         CALL apply_full_single(preconditioner_env, matrix_in, matrix_out)
      CASE (ot_precond_full_all)
         CALL apply_full_all(preconditioner_env, matrix_in, matrix_out)
      CASE (ot_precond_full_kinetic, ot_precond_full_single_inverse, ot_precond_s_inverse)
         SELECT CASE (preconditioner_env%solver)
         CASE (ot_precond_solver_inv_chol, ot_precond_solver_update)
            CALL apply_full_single(preconditioner_env, matrix_in, matrix_out)
         CASE (ot_precond_solver_direct)
            CALL apply_full_direct(preconditioner_env, matrix_in, matrix_out)
         CASE DEFAULT
            CALL cp__b("preconditioner_apply.F",77,"Solver not implemented")
         END SELECT
      CASE DEFAULT
         CALL cp__b("preconditioner_apply.F",80,"Unknown preconditioner")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE apply_preconditioner_fm

! **************************************************************************************************
!> \brief ...
!> \param preconditioner_env ...
!> \param matrix_in ...
!> \param matrix_out ...
! **************************************************************************************************
   SUBROUTINE apply_preconditioner_dbcsr(preconditioner_env, matrix_in, matrix_out)

      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_dbcsr_type)                                :: matrix_in, matrix_out

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_preconditioner_dbcsr', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      SELECT CASE (preconditioner_env%in_use)
      CASE (0)
         CALL cp__b("preconditioner_apply.F",107,"No preconditioner in use")
      CASE (ot_precond_full_single)
         CALL apply_single(preconditioner_env, matrix_in, matrix_out)
      CASE (ot_precond_full_all)
         CALL apply_all(preconditioner_env, matrix_in, matrix_out)
      CASE (ot_precond_full_kinetic, ot_precond_full_single_inverse, ot_precond_s_inverse)
         SELECT CASE (preconditioner_env%solver)
         CASE (ot_precond_solver_inv_chol, ot_precond_solver_update)
            CALL apply_single(preconditioner_env, matrix_in, matrix_out)
         CASE (ot_precond_solver_direct)
            CALL cp__b("preconditioner_apply.F",117,"Apply_full_direct not supported with ot")
            !CALL apply_full_direct(preconditioner_env, matrix_in, matrix_out)
         CASE DEFAULT
            CALL cp__b("preconditioner_apply.F",120,"Wrong solver")
         END SELECT
      CASE DEFAULT
         CALL cp__b("preconditioner_apply.F",123,"Wrong preconditioner")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE apply_preconditioner_dbcsr

! **************************************************************************************************
!> \brief apply to full matrix, complete inversion has already been done
!> \param preconditioner_env ...
!> \param matrix_in ...
!> \param matrix_out ...
! **************************************************************************************************
   SUBROUTINE apply_full_single(preconditioner_env, matrix_in, matrix_out)

      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_fm_type), POINTER                          :: matrix_in, matrix_out

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_full_single', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, k, n

      CALL timeset(routineN, handle)

      CALL cp_fm_get_info(matrix_in, nrow_global=n, ncol_global=k)
      CALL cp_gemm('N', 'N', n, k, n, 1.0_dp, preconditioner_env%fm, &
                   matrix_in, 0.0_dp, matrix_out)
      CALL timestop(handle)

   END SUBROUTINE apply_full_single

! **************************************************************************************************
!> \brief apply to dbcsr matrix, complete inversion has already been done
!> \param preconditioner_env ...
!> \param matrix_in ...
!> \param matrix_out ...
! **************************************************************************************************
   SUBROUTINE apply_single(preconditioner_env, matrix_in, matrix_out)

      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_dbcsr_type)                                :: matrix_in, matrix_out

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_single', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (.NOT. ASSOCIATED(preconditioner_env%dbcsr_matrix)) &
         CALL cp__b("preconditioner_apply.F",173,"NOT ASSOCIATED preconditioner_env%dbcsr_matrix")
      CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, preconditioner_env%dbcsr_matrix, matrix_in, &
                             0.0_dp, matrix_out)

      CALL timestop(handle)

   END SUBROUTINE apply_single

! **************************************************************************************************
!> \brief preconditioner contains the factorization, application done by
!>        solving the linear system
!> \param preconditioner_env ...
!> \param matrix_in ...
!> \param matrix_out ...
! **************************************************************************************************
   SUBROUTINE apply_full_direct(preconditioner_env, matrix_in, matrix_out)

      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_fm_type), POINTER                          :: matrix_in, matrix_out

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_full_direct', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, k, n
      TYPE(cp_fm_type), POINTER                          :: work

      CALL timeset(routineN, handle)

      CALL cp_fm_get_info(matrix_in, nrow_global=n, ncol_global=k)
      CALL cp_fm_create(work, matrix_in%matrix_struct, name="apply_full_single", &
                        use_sp=matrix_in%use_sp)
      CALL cp_fm_cholesky_restore(matrix_in, k, preconditioner_env%fm, work,&
           &                      "SOLVE", transa="T")
      CALL cp_fm_cholesky_restore(work, k, preconditioner_env%fm, matrix_out,&
           &                      "SOLVE", transa="N")
      CALL cp_fm_release(work)

      CALL timestop(handle)

   END SUBROUTINE apply_full_direct

! **************************************************************************************************
!> \brief full all to a full matrix
!> \param preconditioner_env ...
!> \param matrix_in ...
!> \param matrix_out ...
! **************************************************************************************************
   SUBROUTINE apply_full_all(preconditioner_env, matrix_in, matrix_out)

      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_fm_type), POINTER                          :: matrix_in, matrix_out

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_full_all', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, j, k, n, ncol_local, &
                                                            nrow_local
      INTEGER, DIMENSION(:), POINTER                     :: col_indices, row_indices
      REAL(KIND=dp)                                      :: dum
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: local_data
      TYPE(cp_fm_type), POINTER                          :: matrix_tmp

      CALL timeset(routineN, handle)

      CALL cp_fm_get_info(matrix_in, nrow_global=n, ncol_global=k)

      CALL cp_fm_create(matrix_tmp, matrix_in%matrix_struct, name="apply_full_all")
      CALL cp_fm_get_info(matrix_tmp, nrow_local=nrow_local, ncol_local=ncol_local, &
                          row_indices=row_indices, col_indices=col_indices, local_data=local_data)

      !
      CALL cp_gemm('T', 'N', n, k, n, 1.0_dp, preconditioner_env%fm, &
                   matrix_in, 0.0_dp, matrix_tmp)

      ! do the right scaling
      DO j = 1, ncol_local
      DO i = 1, nrow_local
         dum = 1.0_dp/MAX(preconditioner_env%energy_gap, &
                          preconditioner_env%full_evals(row_indices(i))-preconditioner_env%occ_evals(col_indices(j)))
         local_data(i, j) = local_data(i, j)*dum
      ENDDO
      ENDDO

      ! mult back
      CALL cp_gemm('N', 'N', n, k, n, 1.0_dp, preconditioner_env%fm, &
                   matrix_tmp, 0.0_dp, matrix_out)

      CALL cp_fm_release(matrix_tmp)

      CALL timestop(handle)

   END SUBROUTINE apply_full_all

! **************************************************************************************************
!> \brief full all to a dbcsr matrix
!> \param preconditioner_env ...
!> \param matrix_in ...
!> \param matrix_out ...
! **************************************************************************************************
   SUBROUTINE apply_all(preconditioner_env, matrix_in, matrix_out)

      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_dbcsr_type)                                :: matrix_in, matrix_out

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_all', routineP = moduleN//':'//routineN

      INTEGER                                            :: col, col_offset, col_size, handle, i, j, &
                                                            row, row_offset, row_size
      REAL(KIND=dp)                                      :: dum
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: DATA
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_type)                                :: matrix_tmp

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_copy(matrix_tmp, matrix_in, name="apply_full_all")
      CALL cp_dbcsr_multiply('T', 'N', 1.0_dp, preconditioner_env%dbcsr_matrix, &
                             matrix_in, 0.0_dp, matrix_tmp)
      ! do the right scaling
      CALL cp_dbcsr_iterator_start(iter, matrix_tmp)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, DATA, &
                                           row_size=row_size, col_size=col_size, &
                                           row_offset=row_offset, col_offset=col_offset)
         DO j = 1, col_size
         DO i = 1, row_size
            dum = 1.0_dp/MAX(preconditioner_env%energy_gap, &
                             preconditioner_env%full_evals(row_offset+i-1) &
                             -preconditioner_env%occ_evals(col_offset+j-1))
            DATA(i, j) = DATA(i, j)*dum
         ENDDO
         ENDDO
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      ! mult back
      CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, preconditioner_env%dbcsr_matrix, &
                             matrix_tmp, 0.0_dp, matrix_out)
      CALL cp_dbcsr_release(matrix_tmp)
      CALL timestop(handle)

   END SUBROUTINE apply_all

END MODULE preconditioner_apply
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief computes preconditioners, and implements methods to apply them
!>      currently used in qs_ot
!> \par History
!>      - [UB] 2009-05-13 Adding stable approximate inverse (full and sparse)
!> \author Joost VandeVondele (09.2002)
! **************************************************************************************************
MODULE preconditioner
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm
   USE cp_fm_types,                     ONLY: cp_fm_get_info,&
                                              cp_fm_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: cholesky_reduce,&
                                              ot_precond_full_all,&
                                              ot_precond_full_kinetic,&
                                              ot_precond_full_single,&
                                              ot_precond_full_single_inverse,&
                                              ot_precond_none,&
                                              ot_precond_s_inverse,&
                                              ot_precond_solver_update
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE preconditioner_apply,            ONLY: apply_preconditioner_dbcsr,&
                                              apply_preconditioner_fm
   USE preconditioner_makes,            ONLY: make_preconditioner_matrix
   USE preconditioner_solvers,          ONLY: solve_preconditioner,&
                                              transfer_dbcsr_to_fm,&
                                              transfer_fm_to_dbcsr
   USE preconditioner_types,            ONLY: destroy_preconditioner,&
                                              init_preconditioner,&
                                              preconditioner_p_type,&
                                              preconditioner_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_mo_methods,                   ONLY: calculate_subspace_eigenvalues
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type,&
                                              mo_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'preconditioner'

   PUBLIC :: make_preconditioner, restart_preconditioner
   PUBLIC :: apply_preconditioner, prepare_preconditioner

! The public interface for apply preconditioner, the routines can be found in preconditioner_apply.F
   INTERFACE apply_preconditioner
      MODULE PROCEDURE apply_preconditioner_dbcsr
      MODULE PROCEDURE apply_preconditioner_fm
   END INTERFACE

! **************************************************************************************************

CONTAINS

! **************************************************************************************************

! creates a preconditioner for the system (H-energy_homo S)
! this preconditioner is (must be) symmetric positive definite.
! currently uses a atom-block-diagonal form
! each block will be  ....
! might overwrite matrix_h, matrix_t

! **************************************************************************************************
!> \brief ...
!> \param preconditioner_env ...
!> \param precon_type ...
!> \param solver_type ...
!> \param matrix_h ...
!> \param matrix_s ...
!> \param matrix_t ...
!> \param mo_set ...
!> \param energy_gap ...
!> \param mixed_precision ...
!> \param convert_precond_to_dbcsr ...
!> \param chol_type ...
!> \par History
!>      09.2014 removed some unused or unfinished methods
!>              removed sparse preconditioners and the
!>              sparse approximate inverse at rev 14341 [Florian Schiffmann]
! **************************************************************************************************
   SUBROUTINE make_preconditioner(preconditioner_env, precon_type, solver_type, matrix_h, matrix_s, &
                                  matrix_t, mo_set, energy_gap, mixed_precision, convert_precond_to_dbcsr, chol_type)

      TYPE(preconditioner_type)                          :: preconditioner_env
      INTEGER, INTENT(IN)                                :: precon_type, solver_type
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: matrix_s, matrix_t
      TYPE(mo_set_type), POINTER                         :: mo_set
      REAL(KIND=dp)                                      :: energy_gap
      LOGICAL, INTENT(IN), OPTIONAL                      :: mixed_precision, convert_precond_to_dbcsr
      INTEGER, INTENT(IN), OPTIONAL                      :: chol_type

      CHARACTER(len=*), PARAMETER :: routineN = 'make_preconditioner', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, k, my_solver_type
      LOGICAL                                            :: my_convert_precond_to_dbcsr, &
                                                            my_mixed_precision, &
                                                            needs_full_spectrum, needs_homo, &
                                                            use_mo_coeff_b
      REAL(KIND=dp)                                      :: energy_homo
      REAL(KIND=dp), DIMENSION(:), POINTER               :: eigenvalues_ot
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_b
      TYPE(cp_fm_type), POINTER                          :: mo_coeff

      CALL timeset(routineN, handle)

      CALL get_mo_set(mo_set=mo_set, mo_coeff=mo_coeff, mo_coeff_b=mo_coeff_b)
      use_mo_coeff_b = mo_set%use_mo_coeff_b
      CALL cp_fm_get_info(mo_coeff, ncol_global=k)

      ! Starting some matrix mess, check where to store the result in preconditioner_env, fm or dbcsr_matrix
      my_convert_precond_to_dbcsr = .FALSE.
      IF (PRESENT(convert_precond_to_dbcsr)) my_convert_precond_to_dbcsr = convert_precond_to_dbcsr

      ! Check whether we perform everything in single precision
      my_mixed_precision = .FALSE.
      IF (PRESENT(mixed_precision)) my_mixed_precision = mixed_precision
      IF (my_mixed_precision) THEN
         SELECT CASE (precon_type)
         CASE (ot_precond_full_kinetic, ot_precond_full_single_inverse)
            !supported
         CASE DEFAULT
            CALL cp__b("preconditioner.F",138,"This precond with mixed precision is not supported yet")
         END SELECT
      ENDIF

      ! Thanks to the mess with the matrices we need to make sure in this case that the
      ! Previous inverse is properly stored as a sparse matrix, fm gets deallocated here
      ! if it wasn't anyway
      IF (preconditioner_env%solver == ot_precond_solver_update) &
         CALL transfer_fm_to_dbcsr(preconditioner_env%fm, preconditioner_env%dbcsr_matrix, &
                                   matrix_h, my_mixed_precision)

      needs_full_spectrum = .FALSE.
      needs_homo = .FALSE.

      SELECT CASE (precon_type)
      CASE (ot_precond_full_all)
         needs_full_spectrum = .TRUE.
         ! both of them need the coefficients as fm's, more matrix mess
         IF (use_mo_coeff_b) THEN
            CALL copy_dbcsr_to_fm(mo_coeff_b, mo_coeff)
         ENDIF
      CASE (ot_precond_full_single)
         needs_homo = .TRUE.
         ! XXXX to be removed if homo estimate only is implemented
         needs_full_spectrum = .TRUE.
      CASE (ot_precond_full_kinetic, ot_precond_s_inverse, ot_precond_full_single_inverse)
         ! these should be happy without an estimate for the homo energy
         ! preconditioning can  not depend on an absolute eigenvalue, only on eigenvalue differences
      CASE DEFAULT
         CALL cp__b("preconditioner.F",167,"The preconditioner is unknown ...")
      END SELECT

      energy_homo = 0.0_dp
      IF (needs_full_spectrum) THEN
         ALLOCATE (eigenvalues_ot(k))
         ! XXXXXXXXXXXXXXXX do not touch the initial MOs, could be harmful for either
         !                  the case of non-equivalent MOs but also for the derivate
         ! we could already have all eigenvalues e.g. full_all and we could skip this
         ! to be optimised later.
         ! one flaw is that not all SCF methods (i.e. that go over mo_derivs directly)
         ! have a 'valid' matrix_h... (we even don't know what evals are in that case)
         IF (use_mo_coeff_b) THEN
            CALL calculate_subspace_eigenvalues(mo_coeff_b, matrix_h, &
                                                eigenvalues_ot, do_rotation=.FALSE., &
                                                para_env=mo_coeff%matrix_struct%para_env, &
                                                blacs_env=mo_coeff%matrix_struct%context)
         ELSE
            CALL calculate_subspace_eigenvalues(mo_coeff, matrix_h, &
                                                eigenvalues_ot, do_rotation=.FALSE.)
         ENDIF
         IF (k > 0) energy_homo = eigenvalues_ot(k)
      ELSE
         IF (needs_homo) THEN
            CALL cp__b("preconditioner.F",191,"Not yet implemented")
         ENDIF
      ENDIF

      ! After all bits and pieces of checking and initialization, here comes the
      ! part where the preconditioner matrix gets created and solved.
      ! This will give the matrices for later use
      my_solver_type = solver_type
      preconditioner_env%in_use = precon_type
      preconditioner_env%cholesky_use = cholesky_reduce
!dbg
!      write(*,*) PRESENT(chol_type)
!      IF(PRESENT(chol_type))  write(*,*) chol_type, cholesky_reduce
!      stop
!dbg
      IF (PRESENT(chol_type)) preconditioner_env%cholesky_use = chol_type
      preconditioner_env%in_use = precon_type
      CALL make_preconditioner_matrix(preconditioner_env, matrix_h, matrix_s, matrix_t, mo_coeff, &
                                      energy_homo, eigenvalues_ot, energy_gap, &
                                      my_solver_type)

      CALL solve_preconditioner(my_solver_type, preconditioner_env, matrix_s, matrix_h, my_mixed_precision)

      ! Here comes more matrix mess, make sure to output the correct matrix format,
      ! A bit pointless to convert the cholesky factorized version as it doesn't work in
      ! dbcsr form and will crash later,...
      IF (my_convert_precond_to_dbcsr) THEN
         CALL transfer_fm_to_dbcsr(preconditioner_env%fm, preconditioner_env%dbcsr_matrix, &
                                   matrix_h, my_mixed_precision)
      ELSE
         CALL transfer_dbcsr_to_fm(preconditioner_env%dbcsr_matrix, preconditioner_env%fm, &
                                   preconditioner_env%para_env, preconditioner_env%ctxt, &
                                   my_mixed_precision)
      ENDIF

      IF (needs_full_spectrum) THEN
         DEALLOCATE (eigenvalues_ot)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE make_preconditioner

! **************************************************************************************************
!> \brief Allows for a restart of the preconditioner
!>        depending on the method it purges all arrays or keeps them
!> \param qs_env ...
!> \param preconditioner ...
!> \param prec_type ...
!> \param nspins ...
! **************************************************************************************************
   SUBROUTINE restart_preconditioner(qs_env, preconditioner, prec_type, nspins)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(preconditioner_p_type), DIMENSION(:), POINTER :: preconditioner
      INTEGER, INTENT(IN)                                :: prec_type, nspins

      CHARACTER(LEN=*), PARAMETER :: routineN = 'restart_preconditioner', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env

      NULLIFY (para_env, blacs_env)
      CALL get_qs_env(qs_env, para_env=para_env, blacs_env=blacs_env)

      IF (ASSOCIATED(preconditioner)) THEN
         SELECT CASE (prec_type)
         CASE (ot_precond_full_all, ot_precond_full_single) ! these depend on the ks matrix
            DO ispin = 1, SIZE(preconditioner)
               CALL destroy_preconditioner(preconditioner(ispin)%preconditioner)
               DEALLOCATE (preconditioner(ispin)%preconditioner)
            ENDDO
            DEALLOCATE (preconditioner)
         CASE (ot_precond_none, ot_precond_full_kinetic, ot_precond_s_inverse, ot_precond_full_single_inverse) ! these are 'independent'
            ! do nothing
         CASE DEFAULT
            CALL cp__b("preconditioner.F",269,"")
         END SELECT
      END IF

      ! add an OT preconditioner if none is present
      IF (.NOT. ASSOCIATED(preconditioner)) THEN
         SELECT CASE (prec_type)
         CASE (ot_precond_full_all, ot_precond_full_single_inverse)
            ALLOCATE (preconditioner(nspins))
         CASE DEFAULT
            ALLOCATE (preconditioner(1))
         END SELECT
         DO ispin = 1, SIZE(preconditioner)
            ALLOCATE (preconditioner(ispin)%preconditioner)
            CALL init_preconditioner(preconditioner(ispin)%preconditioner, &
                                     para_env=para_env, &
                                     blacs_env=blacs_env)
         ENDDO
      END IF

   END SUBROUTINE restart_preconditioner

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param mos ...
!> \param matrix_ks ...
!> \param matrix_s ...
!> \param ot_preconditioner ...
!> \param prec_type ...
!> \param solver_type ...
!> \param energy_gap ...
!> \param nspins ...
!> \param has_unit_metric ...
!> \param mixed_precision ...
!> \param convert_to_dbcsr ...
!> \param chol_type ...
! **************************************************************************************************
   SUBROUTINE prepare_preconditioner(qs_env, mos, matrix_ks, matrix_s, &
                                     ot_preconditioner, prec_type, solver_type, &
                                     energy_gap, nspins, has_unit_metric, mixed_precision, &
                                     convert_to_dbcsr, chol_type)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_s
      TYPE(preconditioner_p_type), DIMENSION(:), POINTER :: ot_preconditioner
      INTEGER, INTENT(IN)                                :: prec_type, solver_type
      REAL(dp), INTENT(IN)                               :: energy_gap
      INTEGER, INTENT(IN)                                :: nspins
      LOGICAL, INTENT(IN), OPTIONAL                      :: has_unit_metric, mixed_precision, &
                                                            convert_to_dbcsr
      INTEGER, INTENT(IN), OPTIONAL                      :: chol_type

      CHARACTER(LEN=*), PARAMETER :: routineN = 'prepare_preconditioner', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: msg
      INTEGER                                            :: handle, icall, ispin, n_loops
      LOGICAL                                            :: do_co_rotate, my_convert_to_dbcsr, &
                                                            my_has_unit_metric, &
                                                            my_mixed_precision, use_mo_coeff_b
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: kinetic
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_t, mo_coeff_b
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control

      CALL timeset(routineN, handle)
      NULLIFY (matrix_t, mo_coeff_b, mo_coeff, kinetic, dft_control, para_env, blacs_env)
      my_has_unit_metric = .FALSE.
      IF (PRESENT(has_unit_metric)) my_has_unit_metric = has_unit_metric
      my_mixed_precision = .FALSE.
      IF (PRESENT(mixed_precision)) my_mixed_precision = mixed_precision
      my_convert_to_dbcsr = .TRUE.
      IF (PRESENT(convert_to_dbcsr)) my_convert_to_dbcsr = convert_to_dbcsr

!dbg
!    write(*,*) PRESENT(chol_type), "chol_type", chol_type
!    stop
!dbg
      CALL get_qs_env(qs_env, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      blacs_env=blacs_env)

      IF (dft_control%qs_control%semi_empirical .OR. dft_control%qs_control%dftb) THEN
         IF (prec_type == ot_precond_full_kinetic) THEN
            msg = "Full_kinetic not available for semi-empirical methods"
            CALL cp__b("preconditioner.F",359,TRIM(msg))
         END IF
         matrix_t => matrix_s(1)%matrix
      ELSE
         IF(.NOT.(.NOT. my_has_unit_metric))CALL cp__a("preconditioner.F",363)
         CALL get_qs_env(qs_env, kinetic=kinetic)
         matrix_t => kinetic(1)%matrix
      END IF

      !determins how often make preconditioner is called, spin dependent methods have to be called twice
      n_loops = 1
      IF (prec_type == ot_precond_full_single_inverse) n_loops = nspins
      ! check whether we need the ev and rotate the MOs
      SELECT CASE (prec_type)
      CASE (ot_precond_full_all)
         ! if one of these preconditioners is used every spin needs to call make_preconditioner
         n_loops = nspins

         do_co_rotate = ASSOCIATED(qs_env%mo_derivs)
         DO ispin = 1, nspins
            CALL get_mo_set(mo_set=mos(ispin)%mo_set, mo_coeff_b=mo_coeff_b, mo_coeff=mo_coeff)
            use_mo_coeff_b = mos(ispin)%mo_set%use_mo_coeff_b
            IF (use_mo_coeff_b .AND. do_co_rotate) THEN
               CALL calculate_subspace_eigenvalues(mo_coeff_b, matrix_ks(ispin)%matrix, &
                                                   do_rotation=.TRUE., &
                                                   co_rotate=qs_env%mo_derivs(ispin)%matrix, &
                                                   para_env=para_env, &
                                                   blacs_env=blacs_env)
            ELSEIF (use_mo_coeff_b) THEN
               CALL calculate_subspace_eigenvalues(mo_coeff_b, matrix_ks(ispin)%matrix, &
                                                   do_rotation=.TRUE., &
                                                   para_env=para_env, &
                                                   blacs_env=blacs_env)
            ELSE
               CALL calculate_subspace_eigenvalues(mo_coeff, matrix_ks(ispin)%matrix, &
                                                   do_rotation=.TRUE.)
            END IF
         END DO
      CASE DEFAULT
         ! No need to rotate the MOs
      END SELECT

      ! check whether we have a preconditioner
      SELECT CASE (prec_type)
      CASE (ot_precond_none)
         DO ispin = 1, SIZE(ot_preconditioner)
            ot_preconditioner(ispin)%preconditioner%in_use = 0
         END DO
      CASE DEFAULT
         DO icall = 1, n_loops
            IF (my_has_unit_metric) THEN
               CALL make_preconditioner(ot_preconditioner(icall)%preconditioner, &
                                        prec_type, &
                                        solver_type, &
                                        matrix_h=matrix_ks(icall)%matrix, &
                                        mo_set=mos(icall)%mo_set, &
                                        energy_gap=energy_gap, &
                                        convert_precond_to_dbcsr=my_convert_to_dbcsr)
            ELSE
               CALL make_preconditioner(ot_preconditioner(icall)%preconditioner, &
                                        prec_type, &
                                        solver_type, &
                                        matrix_h=matrix_ks(icall)%matrix, &
                                        matrix_s=matrix_s(1)%matrix, &
                                        matrix_t=matrix_t, &
                                        mo_set=mos(icall)%mo_set, &
                                        energy_gap=energy_gap, &
                                        mixed_precision=my_mixed_precision, &
                                        convert_precond_to_dbcsr=my_convert_to_dbcsr, chol_type=chol_type)
            END IF
         END DO
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE prepare_preconditioner

END MODULE preconditioner

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module that contains the algorithms to perform an itrative
!>         diagonalization by the block-Davidson approach
!>         P. Blaha, et al J. Comp. Physics, 229, (2010), 453-460
!>         Iterative diagonalization in augmented plane wave based
!>         methods in electronic structure calculations
!> \par History
!>      05.2011 created [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_scf_block_davidson

   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_col_block_sizes, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_distribution, cp_dbcsr_get_diag, cp_dbcsr_get_info, cp_dbcsr_init_p, &
        cp_dbcsr_multiply, cp_dbcsr_norm, cp_dbcsr_release_p, cp_dbcsr_row_block_sizes, &
        cp_dbcsr_scale_by_vector, cp_dbcsr_type, dbcsr_norm_column, dbcsr_type_no_symmetry, &
        dbcsr_type_real_default, dbcsr_type_symmetric
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr,&
                                              cp_dbcsr_m_by_n_from_row_template,&
                                              cp_dbcsr_m_by_n_from_template,&
                                              cp_dbcsr_sm_fm_multiply
   USE cp_fm_basic_linalg,              ONLY: cp_fm_column_scale,&
                                              cp_fm_scale_and_add,&
                                              cp_fm_symm,&
                                              cp_fm_transpose,&
                                              cp_fm_triangular_invert,&
                                              cp_fm_upper_to_full
   USE cp_fm_cholesky,                  ONLY: cp_fm_cholesky_decompose,&
                                              cp_fm_cholesky_restore
   USE cp_fm_diag,                      ONLY: choose_eigv_solver,&
                                              cp_fm_power
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_get_diag,&
                                              cp_fm_release,&
                                              cp_fm_set_all,&
                                              cp_fm_to_fm,&
                                              cp_fm_to_fm_submat,&
                                              cp_fm_type,&
                                              cp_fm_vectorsnorm
   USE cp_gemm_interface,               ONLY: cp_gemm
   USE kinds,                           ONLY: dp
   USE machine,                         ONLY: m_walltime
   USE message_passing,                 ONLY: mp_sum
   USE preconditioner,                  ONLY: apply_preconditioner
   USE preconditioner_types,            ONLY: preconditioner_type
   USE qs_block_davidson_types,         ONLY: davidson_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_block_davidson'

   PUBLIC :: generate_extended_space, generate_extended_space_sparse

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param bdav_env ...
!> \param mo_set ...
!> \param matrix_h ...
!> \param matrix_s ...
!> \param output_unit ...
!> \param preconditioner ...
! **************************************************************************************************
   SUBROUTINE generate_extended_space(bdav_env, mo_set, matrix_h, matrix_s, output_unit, &
                                      preconditioner)

      TYPE(davidson_type)                                :: bdav_env
      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h, matrix_s
      INTEGER, INTENT(IN)                                :: output_unit
      TYPE(preconditioner_type), OPTIONAL, POINTER       :: preconditioner

      CHARACTER(len=*), PARAMETER :: routineN = 'generate_extended_space', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, homo, i_first, i_last, imo, iter, j, jj, max_iter, n, nao, nmat, nmat2, &
         nmo, nmo_converged, nmo_not_converged, nset, nset_conv, nset_not_conv
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iconv, inotconv
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: iconv_set, inotconv_set
      LOGICAL                                            :: converged, do_apply_preconditioner
      REAL(dp)                                           :: lambda, max_norm, min_norm, t1, t2
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: ritz_coeff, vnorm
      REAL(dp), DIMENSION(:), POINTER                    :: eig_not_conv, eigenvalues, evals
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff_b
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER                          :: c_conv, c_notconv, c_out, c_pz, c_z, &
                                                            h_block, h_fm, m_hc, m_sc, m_tmp, &
                                                            mo_coeff, mt_tmp, s_block, s_fm, &
                                                            v_block, w_block

      CALL timeset(routineN, handle)

      NULLIFY (mo_coeff, mo_coeff_b, eigenvalues)

      do_apply_preconditioner = .FALSE.
      IF (PRESENT(preconditioner)) do_apply_preconditioner = .TRUE.
      CALL get_mo_set(mo_set=mo_set, mo_coeff=mo_coeff, mo_coeff_b=mo_coeff_b, eigenvalues=eigenvalues, &
                      nao=nao, nmo=nmo, homo=homo)
      IF (do_apply_preconditioner) THEN
         max_iter = bdav_env%max_iter
      ELSE
         max_iter = 1
      END IF

      NULLIFY (c_conv, c_notconv, c_out, c_z, c_pz, m_hc, m_sc, m_tmp, mt_tmp)
      NULLIFY (h_block, h_fm, s_block, s_fm, v_block, w_block)
      NULLIFY (evals, eig_not_conv)
      t1 = m_walltime()
      IF (output_unit > 0) THEN
         WRITE (output_unit, "(T15,A,T23,A,T36,A,T49,A,T60,A,/,T8,A)") &
            " Cycle ", " conv. MOS ", " B2MAX ", " B2MIN ", " Time", REPEAT("-", 60)
      END IF

      ALLOCATE (iconv(nmo))
      ALLOCATE (inotconv(nmo))
      ALLOCATE (ritz_coeff(nmo))
      ALLOCATE (vnorm(nmo))

      converged = .FALSE.
      DO iter = 1, max_iter

         ! compute Ritz values
         ritz_coeff = 0.0_dp
         CALL cp_fm_create(m_hc, mo_coeff%matrix_struct, name="hc")
         CALL cp_dbcsr_sm_fm_multiply(matrix_h, mo_coeff, m_hc, nmo)
         CALL cp_fm_create(m_sc, mo_coeff%matrix_struct, name="sc")
         CALL cp_dbcsr_sm_fm_multiply(matrix_s, mo_coeff, m_sc, nmo)

         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nmo, ncol_global=nmo, &
                                  context=mo_coeff%matrix_struct%context, &
                                  para_env=mo_coeff%matrix_struct%para_env)
         CALL cp_fm_create(m_tmp, fm_struct_tmp, name="matrix_tmp")
         CALL cp_fm_struct_release(fm_struct_tmp)

         CALL cp_gemm('T', 'N', nmo, nmo, nao, 1.0_dp, mo_coeff, m_hc, 0.0_dp, m_tmp)
         CALL cp_fm_get_diag(m_tmp, ritz_coeff)
         CALL cp_fm_release(m_tmp)

         ! Check for converged eigenvectors
!      CALL cp_fm_create(c_z,mo_coeff%matrix_struct,name="tmp")
         c_z => bdav_env%matrix_z
         c_pz => bdav_env%matrix_pz
         CALL cp_fm_to_fm(m_sc, c_z)
         CALL cp_fm_column_scale(c_z, ritz_coeff)
         CALL cp_fm_scale_and_add(-1.0_dp, c_z, 1.0_dp, m_hc)
         CALL cp_fm_vectorsnorm(c_z, vnorm)

         nmo_converged = 0
         nmo_not_converged = 0
         max_norm = 0.0_dp
         min_norm = 1.e10_dp
         DO imo = 1, nmo
            max_norm = MAX(max_norm, vnorm(imo))
            min_norm = MIN(min_norm, vnorm(imo))
         END DO
         iconv = 0
         inotconv = 0
         DO imo = 1, nmo
            IF (vnorm(imo) <= bdav_env%eps_iter) THEN
               nmo_converged = nmo_converged+1
               iconv(nmo_converged) = imo
            ELSE
               nmo_not_converged = nmo_not_converged+1
               inotconv(nmo_not_converged) = imo
            END IF
         END DO

         IF (nmo_converged > 0) THEN
            ALLOCATE (iconv_set(nmo_converged, 2))
            ALLOCATE (inotconv_set(nmo_not_converged, 2))
            i_last = iconv(1)
            nset = 0
            DO j = 1, nmo_converged
               imo = iconv(j)

               IF (imo == i_last+1) THEN
                  i_last = imo
                  iconv_set(nset, 2) = imo
               ELSE
                  i_last = imo
                  nset = nset+1
                  iconv_set(nset, 1) = imo
                  iconv_set(nset, 2) = imo
               END IF
            END DO
            nset_conv = nset

            i_last = inotconv(1)
            nset = 0
            DO j = 1, nmo_not_converged
               imo = inotconv(j)

               IF (imo == i_last+1) THEN
                  i_last = imo
                  inotconv_set(nset, 2) = imo
               ELSE
                  i_last = imo
                  nset = nset+1
                  inotconv_set(nset, 1) = imo
                  inotconv_set(nset, 2) = imo
               END IF
            END DO
            nset_not_conv = nset
            CALL cp_fm_release(m_sc)
            CALL cp_fm_release(m_hc)
            NULLIFY (c_z, c_pz)
         END IF

         IF (REAL(nmo_converged, dp)/REAL(nmo, dp) > bdav_env%conv_percent) THEN
            converged = .TRUE.
            DEALLOCATE (iconv_set)
            DEALLOCATE (inotconv_set)
            t2 = m_walltime()
            IF (output_unit > 0) THEN
               WRITE (output_unit, '(T16,I5,T24,I6,T33,E12.4,2x,E12.4,T60,F8.3)') &
                  iter, nmo_converged, max_norm, min_norm, t2-t1

               WRITE (output_unit, *) " Reached convergence in ", iter, &
                  " Davidson iterations"
            END IF

            EXIT
         END IF

         IF (nmo_converged > 0) THEN
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nao, &
                                     context=mo_coeff%matrix_struct%context, &
                                     para_env=mo_coeff%matrix_struct%para_env)
            !allocate h_fm
            CALL cp_fm_create(h_fm, fm_struct_tmp, name="matrix_tmp")
            !allocate s_fm
            CALL cp_fm_create(s_fm, fm_struct_tmp, name="matrix_tmp")
            !copy matrix_h in h_fm
            CALL copy_dbcsr_to_fm(matrix_h, h_fm)
            CALL cp_fm_upper_to_full(h_fm, s_fm)

            !copy matrix_s in s_fm
!        CALL cp_fm_set_all(s_fm,0.0_dp)
            CALL copy_dbcsr_to_fm(matrix_s, s_fm)

            !allocate c_out
            CALL cp_fm_create(c_out, fm_struct_tmp, name="matrix_tmp")
            ! set c_out to zero
            CALL cp_fm_set_all(c_out, 0.0_dp)
            CALL cp_fm_struct_release(fm_struct_tmp)

            !allocate c_conv
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nmo_converged, &
                                     context=mo_coeff%matrix_struct%context, &
                                     para_env=mo_coeff%matrix_struct%para_env)
            CALL cp_fm_create(c_conv, fm_struct_tmp, name="c_conv")
            CALL cp_fm_set_all(c_conv, 0.0_dp)
            !allocate m_tmp
            CALL cp_fm_create(m_tmp, fm_struct_tmp, name="m_tmp_nxmc")
            CALL cp_fm_struct_release(fm_struct_tmp)
         END IF

         !allocate c_notconv
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nmo_not_converged, &
                                  context=mo_coeff%matrix_struct%context, &
                                  para_env=mo_coeff%matrix_struct%para_env)
         CALL cp_fm_create(c_notconv, fm_struct_tmp, name="c_notconv")
         CALL cp_fm_set_all(c_notconv, 0.0_dp)
         IF (nmo_converged > 0) THEN
            CALL cp_fm_create(m_hc, fm_struct_tmp, name="m_hc")
            CALL cp_fm_create(m_sc, fm_struct_tmp, name="m_sc")
            !allocate c_z
            CALL cp_fm_create(c_z, fm_struct_tmp, name="c_z")
            CALL cp_fm_create(c_pz, fm_struct_tmp, name="c_pz")
            CALL cp_fm_set_all(c_z, 0.0_dp)

            ! sum contributions to c_out
            jj = 1
            DO j = 1, nset_conv
               i_first = iconv_set(j, 1)
               i_last = iconv_set(j, 2)
               n = i_last-i_first+1
               CALL cp_fm_to_fm_submat(mo_coeff, c_conv, nao, n, 1, i_first, 1, jj)
               jj = jj+n
            END DO
            CALL cp_fm_symm('L', 'U', nao, nmo_converged, 1.0_dp, s_fm, c_conv, 0.0_dp, m_tmp)
            CALL cp_gemm('N', 'T', nao, nao, nmo_converged, 1.0_dp, m_tmp, m_tmp, 0.0_dp, c_out)

            ! project c_out out of H
            lambda = 100.0_dp*ABS(eigenvalues(homo))
            CALL cp_fm_scale_and_add(lambda, c_out, 1.0_dp, h_fm)
            CALL cp_fm_release(m_tmp)
            CALL cp_fm_release(h_fm)

         END IF

         !allocate m_tmp
         CALL cp_fm_create(m_tmp, fm_struct_tmp, name="m_tmp_nxm")
         CALL cp_fm_struct_release(fm_struct_tmp)
         IF (nmo_converged > 0) THEN
            ALLOCATE (eig_not_conv(nmo_not_converged))
            jj = 1
            DO j = 1, nset_not_conv
               i_first = inotconv_set(j, 1)
               i_last = inotconv_set(j, 2)
               n = i_last-i_first+1
               CALL cp_fm_to_fm_submat(mo_coeff, c_notconv, nao, n, 1, i_first, 1, jj)
               eig_not_conv(jj:jj+n-1) = ritz_coeff(i_first:i_last)
               jj = jj+n
            END DO
            CALL cp_gemm('N', 'N', nao, nmo_not_converged, nao, 1.0_dp, c_out, c_notconv, 0.0_dp, m_hc)
            CALL cp_fm_symm('L', 'U', nao, nmo_not_converged, 1.0_dp, s_fm, c_notconv, 0.0_dp, m_sc)
            ! extend suspace using only the not converged vectors
            CALL cp_fm_to_fm(m_sc, m_tmp)
            CALL cp_fm_column_scale(m_tmp, eig_not_conv)
            CALL cp_fm_scale_and_add(-1.0_dp, m_tmp, 1.0_dp, m_hc)
            DEALLOCATE (eig_not_conv)
            CALL cp_fm_to_fm(m_tmp, c_z)
         ELSE
            CALL cp_fm_to_fm(mo_coeff, c_notconv)
         END IF

         !preconditioner
         IF (do_apply_preconditioner) THEN
            IF (preconditioner%in_use /= 0) THEN
               CALL apply_preconditioner(preconditioner, c_z, c_pz)
            ELSE
               CALL cp_fm_to_fm(c_z, c_pz)
            ENDIF
         ELSE
            CALL cp_fm_to_fm(c_z, c_pz)
         END IF
         CALL cp_fm_release(m_tmp)

         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nmo_not_converged, ncol_global=nmo_not_converged, &
                                  context=mo_coeff%matrix_struct%context, &
                                  para_env=mo_coeff%matrix_struct%para_env)

         CALL cp_fm_create(m_tmp, fm_struct_tmp, name="m_tmp_mxm")
         CALL cp_fm_create(mt_tmp, fm_struct_tmp, name="mt_tmp_mxm")
         CALL cp_fm_struct_release(fm_struct_tmp)

         nmat = nmo_not_converged
         nmat2 = 2*nmo_not_converged
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nmat2, ncol_global=nmat2, &
                                  context=mo_coeff%matrix_struct%context, &
                                  para_env=mo_coeff%matrix_struct%para_env)

         CALL cp_fm_create(s_block, fm_struct_tmp, name="sb")
         CALL cp_fm_create(h_block, fm_struct_tmp, name="hb")
         CALL cp_fm_create(v_block, fm_struct_tmp, name="vb")
         CALL cp_fm_create(w_block, fm_struct_tmp, name="wb")
         ALLOCATE (evals(nmat2))

         CALL cp_fm_struct_release(fm_struct_tmp)

         ! compute CSC
         CALL cp_fm_set_all(s_block, 0.0_dp, 1.0_dp)

         ! compute CHC
         CALL cp_gemm('T', 'N', nmat, nmat, nao, 1.0_dp, c_notconv, m_hc, 0.0_dp, m_tmp)
         CALL cp_fm_to_fm_submat(m_tmp, h_block, nmat, nmat, 1, 1, 1, 1)

         ! compute ZSC
         CALL cp_gemm('T', 'N', nmat, nmat, nao, 1.0_dp, c_pz, m_sc, 0.0_dp, m_tmp)
         CALL cp_fm_to_fm_submat(m_tmp, s_block, nmat, nmat, 1, 1, 1+nmat, 1)
         CALL cp_fm_transpose(m_tmp, mt_tmp)
         CALL cp_fm_to_fm_submat(mt_tmp, s_block, nmat, nmat, 1, 1, 1, 1+nmat)
         ! compute ZHC
         CALL cp_gemm('T', 'N', nmat, nmat, nao, 1.0_dp, c_pz, m_hc, 0.0_dp, m_tmp)
         CALL cp_fm_to_fm_submat(m_tmp, h_block, nmat, nmat, 1, 1, 1+nmat, 1)
         CALL cp_fm_transpose(m_tmp, mt_tmp)
         CALL cp_fm_to_fm_submat(mt_tmp, h_block, nmat, nmat, 1, 1, 1, 1+nmat)

         CALL cp_fm_release(mt_tmp)

         ! reuse m_sc and m_hc to computr HZ and SZ
         IF (nmo_converged > 0) THEN
            CALL cp_gemm('N', 'N', nao, nmat, nao, 1.0_dp, c_out, c_pz, 0.0_dp, m_hc)
            CALL cp_fm_symm('L', 'U', nao, nmo_not_converged, 1.0_dp, s_fm, c_pz, 0.0_dp, m_sc)

            CALL cp_fm_release(c_out)
            CALL cp_fm_release(c_conv)
            CALL cp_fm_release(s_fm)
         ELSE
            CALL cp_dbcsr_sm_fm_multiply(matrix_h, c_pz, m_hc, nmo)
            CALL cp_dbcsr_sm_fm_multiply(matrix_s, c_pz, m_sc, nmo)
         END IF

         ! compute ZSZ
         CALL cp_gemm('T', 'N', nmat, nmat, nao, 1.0_dp, c_pz, m_sc, 0.0_dp, m_tmp)
         CALL cp_fm_to_fm_submat(m_tmp, s_block, nmat, nmat, 1, 1, 1+nmat, 1+nmat)
         ! compute ZHZ
         CALL cp_gemm('T', 'N', nmat, nmat, nao, 1.0_dp, c_pz, m_hc, 0.0_dp, m_tmp)
         CALL cp_fm_to_fm_submat(m_tmp, h_block, nmat, nmat, 1, 1, 1+nmat, 1+nmat)

         CALL cp_fm_release(m_sc)

         ! solution of the reduced eigenvalues problem
         CALL reduce_extended_space(s_block, h_block, v_block, w_block, evals, nmat2)

         ! extract egenvectors
         CALL cp_fm_to_fm_submat(v_block, m_tmp, nmat, nmat, 1, 1, 1, 1)
         CALL cp_gemm('N', 'N', nao, nmat, nmat, 1.0_dp, c_notconv, m_tmp, 0.0_dp, m_hc)
         CALL cp_fm_to_fm_submat(v_block, m_tmp, nmat, nmat, 1+nmat, 1, 1, 1)
         CALL cp_gemm('N', 'N', nao, nmat, nmat, 1.0_dp, c_pz, m_tmp, 1.0_dp, m_hc)

         CALL cp_fm_release(m_tmp)

         CALL cp_fm_release(c_notconv)
         CALL cp_fm_release(s_block)
         CALL cp_fm_release(h_block)
         CALL cp_fm_release(w_block)
         CALL cp_fm_release(v_block)

         IF (nmo_converged > 0) THEN
            CALL cp_fm_release(c_z)
            CALL cp_fm_release(c_pz)
            jj = 1
            DO j = 1, nset_not_conv
               i_first = inotconv_set(j, 1)
               i_last = inotconv_set(j, 2)
               n = i_last-i_first+1
               CALL cp_fm_to_fm_submat(m_hc, mo_coeff, nao, n, 1, jj, 1, i_first)
               eigenvalues(i_first:i_last) = evals(jj:jj+n-1)
               jj = jj+n
            END DO
            DEALLOCATE (iconv_set)
            DEALLOCATE (inotconv_set)
         ELSE
            CALL cp_fm_to_fm(m_hc, mo_coeff)
            eigenvalues(1:nmo) = evals(1:nmo)
         END IF
         DEALLOCATE (evals)

         CALL cp_fm_release(m_hc)

         CALL copy_fm_to_dbcsr(mo_coeff, mo_coeff_b) !fm->dbcsr

         t2 = m_walltime()
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(T16,I5,T24,I6,T33,E12.4,2x,E12.4,T60,F8.3)') &
               iter, nmo_converged, max_norm, min_norm, t2-t1
         END IF
         t1 = m_walltime()

      END DO ! iter

      DEALLOCATE (iconv)
      DEALLOCATE (inotconv)
      DEALLOCATE (ritz_coeff)
      DEALLOCATE (vnorm)

      CALL timestop(handle)
   END SUBROUTINE generate_extended_space

! **************************************************************************************************
!> \brief ...
!> \param bdav_env ...
!> \param mo_set ...
!> \param matrix_h ...
!> \param matrix_s ...
!> \param output_unit ...
!> \param preconditioner ...
! **************************************************************************************************
   SUBROUTINE generate_extended_space_sparse(bdav_env, mo_set, matrix_h, matrix_s, output_unit, &
                                             preconditioner)

      TYPE(davidson_type)                                :: bdav_env
      TYPE(mo_set_type), POINTER                         :: mo_set
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h, matrix_s
      INTEGER, INTENT(IN)                                :: output_unit
      TYPE(preconditioner_type), OPTIONAL, POINTER       :: preconditioner

      CHARACTER(len=*), PARAMETER :: routineN = 'generate_extended_space_sparse', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, homo, i_first, i_last, imo, iter, j, jj, k, max_iter, n, nao, nmat, &
         nmat2, nmo, nmo_converged, nmo_not_converged, nset, nset_conv, nset_not_conv
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: iconv, inotconv
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: iconv_set, inotconv_set
      LOGICAL                                            :: converged, do_apply_preconditioner
      REAL(dp)                                           :: lambda, max_norm, min_norm, t1, t2
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: eig_not_conv, evals, ritz_coeff, vnorm
      REAL(dp), DIMENSION(:), POINTER                    :: eigenvalues
      TYPE(cp_dbcsr_type), POINTER                       :: c_out, matrix_hc, matrix_mm, matrix_pz, &
                                                            matrix_sc, matrix_z, mo_coeff_b, &
                                                            mo_conv, mo_notconv, smo_conv
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp
      TYPE(cp_fm_type), POINTER :: h_block, matrix_mm_fm, matrix_mmt_fm, matrix_nm_fm, &
         matrix_z_fm, mo_coeff, mo_conv_fm, mo_notconv_fm, s_block, v_block, w_block

      CALL timeset(routineN, handle)

      do_apply_preconditioner = .FALSE.
      IF (PRESENT(preconditioner)) do_apply_preconditioner = .TRUE.

      NULLIFY (mo_coeff, mo_coeff_b, matrix_hc, matrix_sc, matrix_z, matrix_pz, matrix_mm)
      NULLIFY (mo_conv_fm, mo_notconv_fm, mo_conv, mo_notconv, smo_conv, c_out)
      NULLIFY (matrix_mm_fm, matrix_mmt_fm, mo_coeff, matrix_nm_fm, matrix_z_fm)
      NULLIFY (h_block, s_block, v_block, w_block)
      NULLIFY (fm_struct_tmp)
      CALL get_mo_set(mo_set=mo_set, mo_coeff=mo_coeff, mo_coeff_b=mo_coeff_b, &
                      eigenvalues=eigenvalues, homo=homo, nao=nao, nmo=nmo)
      IF (do_apply_preconditioner) THEN
         max_iter = bdav_env%max_iter
      ELSE
         max_iter = 1
      END IF

      t1 = m_walltime()
      IF (output_unit > 0) THEN
         WRITE (output_unit, "(T15,A,T23,A,T36,A,T49,A,T60,A,/,T8,A)") &
            " Cycle ", " conv. MOS ", " B2MAX ", " B2MIN ", " Time", REPEAT("-", 60)
      END IF

      ! Allocate array for Ritz values
      ALLOCATE (ritz_coeff(nmo))
      ALLOCATE (iconv(nmo))
      ALLOCATE (inotconv(nmo))
      ALLOCATE (vnorm(nmo))

      converged = .FALSE.
      DO iter = 1, max_iter
         NULLIFY (c_out, mo_conv, mo_notconv_fm, mo_notconv)
         ! Prepare HC and SC, using mo_coeff_b (sparse), these are still sparse
         CALL cp_dbcsr_init_p(matrix_hc)
         CALL cp_dbcsr_create(matrix_hc, "matrix_hc", cp_dbcsr_distribution(mo_coeff_b), &
                              dbcsr_type_no_symmetry, cp_dbcsr_row_block_sizes(mo_coeff_b), &
                              cp_dbcsr_col_block_sizes(mo_coeff_b), nze=0, data_type=dbcsr_type_real_default)
         CALL cp_dbcsr_init_p(matrix_sc)
         CALL cp_dbcsr_create(matrix_sc, "matrix_sc", cp_dbcsr_distribution(mo_coeff_b), &
                              dbcsr_type_no_symmetry, cp_dbcsr_row_block_sizes(mo_coeff_b), &
                              cp_dbcsr_col_block_sizes(mo_coeff_b), nze=0, data_type=dbcsr_type_real_default)

         CALL cp_dbcsr_get_info(mo_coeff_b, nfullrows_total=n, nfullcols_total=k)
         CALL cp_dbcsr_multiply('n', 'n', 1.0_dp, matrix_h, mo_coeff_b, 0.0_dp, matrix_hc, last_column=k)
         CALL cp_dbcsr_multiply('n', 'n', 1.0_dp, matrix_s, mo_coeff_b, 0.0_dp, matrix_sc, last_column=k)

         ! compute Ritz values
         ritz_coeff = 0.0_dp
         ! Allocate Sparse matrices: nmoxnmo
         ! matrix_mm

         CALL cp_dbcsr_init_p(matrix_mm)
         CALL cp_dbcsr_m_by_n_from_template(matrix_mm, template=matrix_s, m=nmo, n=nmo, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_multiply('t', 'n', 1.0_dp, mo_coeff_b, matrix_hc, 0.0_dp, matrix_mm, last_column=k)
         CALL cp_dbcsr_get_diag(matrix_mm, ritz_coeff)
         CALL mp_sum(ritz_coeff, mo_coeff%matrix_struct%para_env%group)

         ! extended subspace P Z = P [H - theta S]C  this ia another matrix of type and size as mo_coeff_b
         CALL cp_dbcsr_init_p(matrix_z)
         CALL cp_dbcsr_create(matrix_z, "matrix_z", cp_dbcsr_distribution(mo_coeff_b), &
                              dbcsr_type_no_symmetry, cp_dbcsr_row_block_sizes(mo_coeff_b), &
                              cp_dbcsr_col_block_sizes(mo_coeff_b), nze=0, data_type=dbcsr_type_real_default)
         CALL cp_dbcsr_copy(matrix_z, matrix_sc)
         CALL cp_dbcsr_scale_by_vector(matrix_z, ritz_coeff, side='right')
         CALL cp_dbcsr_add(matrix_z, matrix_hc, -1.0_dp, 1.0_dp)

         ! Check for converged eigenvectors
         vnorm = 0.0_dp
         CALL cp_dbcsr_norm(matrix_z, which_norm=dbcsr_norm_column, norm_vector=vnorm)
         nmo_converged = 0
         nmo_not_converged = 0
         max_norm = 0.0_dp
         min_norm = 1.e10_dp
         DO imo = 1, nmo
            max_norm = MAX(max_norm, vnorm(imo))
            min_norm = MIN(min_norm, vnorm(imo))
         END DO
         iconv = 0
         inotconv = 0

         DO imo = 1, nmo
            IF (vnorm(imo) <= bdav_env%eps_iter) THEN
               nmo_converged = nmo_converged+1
               iconv(nmo_converged) = imo
            ELSE
               nmo_not_converged = nmo_not_converged+1
               inotconv(nmo_not_converged) = imo
            END IF
         END DO

         IF (nmo_converged > 0) THEN
            ALLOCATE (iconv_set(nmo_converged, 2))
            ALLOCATE (inotconv_set(nmo_not_converged, 2))
            i_last = iconv(1)
            nset = 0
            DO j = 1, nmo_converged
               imo = iconv(j)

               IF (imo == i_last+1) THEN
                  i_last = imo
                  iconv_set(nset, 2) = imo
               ELSE
                  i_last = imo
                  nset = nset+1
                  iconv_set(nset, 1) = imo
                  iconv_set(nset, 2) = imo
               END IF
            END DO
            nset_conv = nset

            i_last = inotconv(1)
            nset = 0
            DO j = 1, nmo_not_converged
               imo = inotconv(j)

               IF (imo == i_last+1) THEN
                  i_last = imo
                  inotconv_set(nset, 2) = imo
               ELSE
                  i_last = imo
                  nset = nset+1
                  inotconv_set(nset, 1) = imo
                  inotconv_set(nset, 2) = imo
               END IF
            END DO
            nset_not_conv = nset

            CALL cp_dbcsr_release_p(matrix_hc)
            CALL cp_dbcsr_release_p(matrix_sc)
            CALL cp_dbcsr_release_p(matrix_z)
            CALL cp_dbcsr_release_p(matrix_mm)
         END IF

         IF (REAL(nmo_converged, dp)/REAL(nmo, dp) > bdav_env%conv_percent) THEN
            DEALLOCATE (iconv_set)

            DEALLOCATE (inotconv_set)

            converged = .TRUE.
            t2 = m_walltime()
            IF (output_unit > 0) THEN
               WRITE (output_unit, '(T16,I5,T24,I6,T33,E12.4,2x,E12.4,T60,F8.3)') &
                  iter, nmo_converged, max_norm, min_norm, t2-t1

               WRITE (output_unit, *) " Reached convergence in ", iter, &
                  " Davidson iterations"
            END IF

            EXIT
         END IF

         IF (nmo_converged > 0) THEN

            !allocate mo_conv_fm
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nmo_converged, &
                                     context=mo_coeff%matrix_struct%context, &
                                     para_env=mo_coeff%matrix_struct%para_env)
            CALL cp_fm_create(mo_conv_fm, fm_struct_tmp, name="mo_conv_fm")

            CALL cp_fm_struct_release(fm_struct_tmp)

            ! extract mo_conv from mo_coeff full matrix
            jj = 1
            DO j = 1, nset_conv
               i_first = iconv_set(j, 1)
               i_last = iconv_set(j, 2)
               n = i_last-i_first+1
               CALL cp_fm_to_fm_submat(mo_coeff, mo_conv_fm, nao, n, 1, i_first, 1, jj)
               jj = jj+n
            END DO

            ! allocate c_out sparse matrix, to project out the converged MOS
            CALL cp_dbcsr_init_p(c_out)
            CALL cp_dbcsr_create(c_out, "c_out", cp_dbcsr_distribution(matrix_s), &
                                 dbcsr_type_symmetric, cp_dbcsr_row_block_sizes(matrix_s), &
                                 cp_dbcsr_col_block_sizes(matrix_s), nze=0, data_type=dbcsr_type_real_default)

            ! allocate mo_conv sparse
            CALL cp_dbcsr_init_p(mo_conv)
            CALL cp_dbcsr_m_by_n_from_row_template(mo_conv, template=matrix_s, n=nmo_converged, &
                                                   sym=dbcsr_type_no_symmetry)

            CALL cp_dbcsr_init_p(smo_conv)
            CALL cp_dbcsr_m_by_n_from_row_template(smo_conv, template=matrix_s, n=nmo_converged, &
                                                   sym=dbcsr_type_no_symmetry)

            CALL copy_fm_to_dbcsr(mo_conv_fm, mo_conv) !fm->dbcsr

            CALL cp_dbcsr_multiply('n', 'n', 1.0_dp, matrix_s, mo_conv, 0.0_dp, smo_conv, last_column=nmo_converged)
            CALL cp_dbcsr_multiply('n', 't', 1.0_dp, smo_conv, smo_conv, 0.0_dp, c_out, last_column=nao)
            ! project c_out out of H
            lambda = 100.0_dp*ABS(eigenvalues(homo))
            CALL cp_dbcsr_add(c_out, matrix_h, lambda, 1.0_dp)

            CALL cp_dbcsr_release_p(mo_conv)
            CALL cp_dbcsr_release_p(smo_conv)
            CALL cp_fm_release(mo_conv_fm)

            !allocate c_notconv_fm
            CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nmo_not_converged, &
                                     context=mo_coeff%matrix_struct%context, &
                                     para_env=mo_coeff%matrix_struct%para_env)
            CALL cp_fm_create(mo_notconv_fm, fm_struct_tmp, name="mo_notconv_fm")
            CALL cp_fm_struct_release(fm_struct_tmp)

            ! extract mo_notconv from mo_coeff full matrix
            ALLOCATE (eig_not_conv(nmo_not_converged))

            jj = 1
            DO j = 1, nset_not_conv
               i_first = inotconv_set(j, 1)
               i_last = inotconv_set(j, 2)
               n = i_last-i_first+1
               CALL cp_fm_to_fm_submat(mo_coeff, mo_notconv_fm, nao, n, 1, i_first, 1, jj)
               eig_not_conv(jj:jj+n-1) = ritz_coeff(i_first:i_last)
               jj = jj+n
            END DO

            ! allocate mo_conv sparse
            CALL cp_dbcsr_init_p(mo_notconv)
            CALL cp_dbcsr_m_by_n_from_row_template(mo_notconv, template=matrix_s, n=nmo_not_converged, &
                                                   sym=dbcsr_type_no_symmetry)

            CALL cp_dbcsr_init_p(matrix_hc)
            CALL cp_dbcsr_m_by_n_from_row_template(matrix_hc, template=matrix_s, n=nmo_not_converged, &
                                                   sym=dbcsr_type_no_symmetry)

            CALL cp_dbcsr_init_p(matrix_sc)
            CALL cp_dbcsr_m_by_n_from_row_template(matrix_sc, template=matrix_s, n=nmo_not_converged, &
                                                   sym=dbcsr_type_no_symmetry)

            CALL cp_dbcsr_init_p(matrix_z)
            CALL cp_dbcsr_m_by_n_from_row_template(matrix_z, template=matrix_s, n=nmo_not_converged, &
                                                   sym=dbcsr_type_no_symmetry)

            CALL copy_fm_to_dbcsr(mo_notconv_fm, mo_notconv) !fm->dbcsr

            CALL cp_dbcsr_multiply('n', 'n', 1.0_dp, c_out, mo_notconv, 0.0_dp, matrix_hc, &
                                   last_column=nmo_not_converged)
            CALL cp_dbcsr_multiply('n', 'n', 1.0_dp, matrix_s, mo_notconv, 0.0_dp, matrix_sc, &
                                   last_column=nmo_not_converged)

            CALL cp_dbcsr_copy(matrix_z, matrix_sc)
            CALL cp_dbcsr_scale_by_vector(matrix_z, eig_not_conv, side='right')
            CALL cp_dbcsr_add(matrix_z, matrix_hc, -1.0_dp, 1.0_dp)

            DEALLOCATE (eig_not_conv)

            ! matrix_mm
            CALL cp_dbcsr_init_p(matrix_mm)
            CALL cp_dbcsr_m_by_n_from_template(matrix_mm, template=matrix_s, m=nmo_not_converged, n=nmo_not_converged, &
                                               sym=dbcsr_type_no_symmetry)

            CALL cp_dbcsr_multiply('t', 'n', 1.0_dp, mo_notconv, matrix_hc, 0.0_dp, matrix_mm, &
                                   last_column=nmo_not_converged)

         ELSE
            mo_notconv => mo_coeff_b
            mo_notconv_fm => mo_coeff
            c_out => matrix_h
         END IF

         ! allocate matrix_pz using as template matrix_z
         CALL cp_dbcsr_init_p(matrix_pz)
         CALL cp_dbcsr_create(matrix_pz, "matrix_pz", cp_dbcsr_distribution(matrix_z), &
                              dbcsr_type_no_symmetry, cp_dbcsr_row_block_sizes(matrix_z), &
                              cp_dbcsr_col_block_sizes(matrix_z), nze=0, data_type=dbcsr_type_real_default)

         IF (do_apply_preconditioner) THEN
            IF (preconditioner%in_use /= 0) THEN
               CALL apply_preconditioner(preconditioner, matrix_z, matrix_pz)
            ELSE
               CALL cp_dbcsr_copy(matrix_pz, matrix_z)
            END IF
         ELSE
            CALL cp_dbcsr_copy(matrix_pz, matrix_z)
         END IF

         !allocate NMOxNMO  full matrices
         nmat = nmo_not_converged
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nmat, ncol_global=nmat, &
                                  context=mo_coeff%matrix_struct%context, &
                                  para_env=mo_coeff%matrix_struct%para_env)
         CALL cp_fm_create(matrix_mm_fm, fm_struct_tmp, name="m_tmp_mxm")
         CALL cp_fm_create(matrix_mmt_fm, fm_struct_tmp, name="mt_tmp_mxm")
         CALL cp_fm_struct_release(fm_struct_tmp)

         !allocate 2NMOx2NMO full matrices
         nmat2 = 2*nmo_not_converged
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nmat2, ncol_global=nmat2, &
                                  context=mo_coeff%matrix_struct%context, &
                                  para_env=mo_coeff%matrix_struct%para_env)

         CALL cp_fm_create(s_block, fm_struct_tmp, name="sb")
         CALL cp_fm_create(h_block, fm_struct_tmp, name="hb")
         CALL cp_fm_create(v_block, fm_struct_tmp, name="vb")
         CALL cp_fm_create(w_block, fm_struct_tmp, name="wb")
         ALLOCATE (evals(nmat2))
         CALL cp_fm_struct_release(fm_struct_tmp)

         ! compute CSC
         CALL cp_fm_set_all(s_block, 0.0_dp, 1.0_dp)
         ! compute CHC
         CALL copy_dbcsr_to_fm(matrix_mm, matrix_mm_fm)
         CALL cp_fm_to_fm_submat(matrix_mm_fm, h_block, nmat, nmat, 1, 1, 1, 1)

         ! compute the bottom left  ZSC (top right is transpose)
         CALL cp_dbcsr_multiply('t', 'n', 1.0_dp, matrix_pz, matrix_sc, 0.0_dp, matrix_mm, last_column=nmat)
         !  set the bottom left part of S[C,Z] block matrix  ZSC
         !copy sparse to full
         CALL copy_dbcsr_to_fm(matrix_mm, matrix_mm_fm)
         CALL cp_fm_to_fm_submat(matrix_mm_fm, s_block, nmat, nmat, 1, 1, 1+nmat, 1)
         CALL cp_fm_transpose(matrix_mm_fm, matrix_mmt_fm)
         CALL cp_fm_to_fm_submat(matrix_mmt_fm, s_block, nmat, nmat, 1, 1, 1, 1+nmat)

         ! compute the bottom left  ZHC (top right is transpose)
         CALL cp_dbcsr_multiply('t', 'n', 1.0_dp, matrix_pz, matrix_hc, 0.0_dp, matrix_mm, last_column=nmat)
         ! set the bottom left part of S[C,Z] block matrix  ZHC
         CALL copy_dbcsr_to_fm(matrix_mm, matrix_mm_fm)
         CALL cp_fm_to_fm_submat(matrix_mm_fm, h_block, nmat, nmat, 1, 1, 1+nmat, 1)
         CALL cp_fm_transpose(matrix_mm_fm, matrix_mmt_fm)
         CALL cp_fm_to_fm_submat(matrix_mmt_fm, h_block, nmat, nmat, 1, 1, 1, 1+nmat)

         CALL cp_fm_release(matrix_mmt_fm)

         ! (reuse matrix_sc and matrix_hc to computr HZ and SZ)
         CALL cp_dbcsr_get_info(matrix_pz, nfullrows_total=n, nfullcols_total=k)
         CALL cp_dbcsr_multiply('n', 'n', 1.0_dp, c_out, matrix_pz, 0.0_dp, matrix_hc, last_column=k)
         CALL cp_dbcsr_multiply('n', 'n', 1.0_dp, matrix_s, matrix_pz, 0.0_dp, matrix_sc, last_column=k)

         ! compute the bottom right  ZSZ
         CALL cp_dbcsr_multiply('t', 'n', 1.0_dp, matrix_pz, matrix_sc, 0.0_dp, matrix_mm, last_column=k)
         ! set the bottom right part of S[C,Z] block matrix  ZSZ
         CALL copy_dbcsr_to_fm(matrix_mm, matrix_mm_fm)
         CALL cp_fm_to_fm_submat(matrix_mm_fm, s_block, nmat, nmat, 1, 1, 1+nmat, 1+nmat)

         ! compute the bottom right  ZHZ
         CALL cp_dbcsr_multiply('t', 'n', 1.0_dp, matrix_pz, matrix_hc, 0.0_dp, matrix_mm, last_column=k)
         ! set the bottom right part of H[C,Z] block matrix  ZHZ
         CALL copy_dbcsr_to_fm(matrix_mm, matrix_mm_fm)
         CALL cp_fm_to_fm_submat(matrix_mm_fm, h_block, nmat, nmat, 1, 1, 1+nmat, 1+nmat)

         CALL cp_dbcsr_release_p(matrix_mm)
         CALL cp_dbcsr_release_p(matrix_sc)
         CALL cp_dbcsr_release_p(matrix_hc)

         CALL reduce_extended_space(s_block, h_block, v_block, w_block, evals, nmat2)

         ! allocate two (nao x nmat) full matrix
         CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nmat, &
                                  context=mo_coeff%matrix_struct%context, &
                                  para_env=mo_coeff%matrix_struct%para_env)
         CALL cp_fm_create(matrix_nm_fm, fm_struct_tmp, name="m_nxm")
         CALL cp_fm_create(matrix_z_fm, fm_struct_tmp, name="m_nxm")
         CALL cp_fm_struct_release(fm_struct_tmp)

         CALL copy_dbcsr_to_fm(matrix_pz, matrix_z_fm)
         ! extract egenvectors
         CALL cp_fm_to_fm_submat(v_block, matrix_mm_fm, nmat, nmat, 1, 1, 1, 1)
         CALL cp_gemm('N', 'N', nao, nmat, nmat, 1.0_dp, mo_notconv_fm, matrix_mm_fm, 0.0_dp, matrix_nm_fm)
         CALL cp_fm_to_fm_submat(v_block, matrix_mm_fm, nmat, nmat, 1+nmat, 1, 1, 1)
         CALL cp_gemm('N', 'N', nao, nmat, nmat, 1.0_dp, matrix_z_fm, matrix_mm_fm, 1.0_dp, matrix_nm_fm)

         CALL cp_dbcsr_release_p(matrix_z)
         CALL cp_dbcsr_release_p(matrix_pz)
         CALL cp_fm_release(matrix_z_fm)
         CALL cp_fm_release(s_block)
         CALL cp_fm_release(h_block)
         CALL cp_fm_release(w_block)
         CALL cp_fm_release(v_block)
         CALL cp_fm_release(matrix_mm_fm)

         ! in case some vector are already converged only a subset of vectors are copied in the MOS
         IF (nmo_converged > 0) THEN
            jj = 1
            DO j = 1, nset_not_conv
               i_first = inotconv_set(j, 1)
               i_last = inotconv_set(j, 2)
               n = i_last-i_first+1
               CALL cp_fm_to_fm_submat(matrix_nm_fm, mo_coeff, nao, n, 1, jj, 1, i_first)
               eigenvalues(i_first:i_last) = evals(jj:jj+n-1)
               jj = jj+n
            END DO
            DEALLOCATE (iconv_set)
            DEALLOCATE (inotconv_set)

            CALL cp_dbcsr_release_p(mo_notconv)
            CALL cp_dbcsr_release_p(c_out)
            CALL cp_fm_release(mo_notconv_fm)
         ELSE
            CALL cp_fm_to_fm(matrix_nm_fm, mo_coeff)
            eigenvalues(1:nmo) = evals(1:nmo)
         END IF
         DEALLOCATE (evals)

         CALL cp_fm_release(matrix_nm_fm)
         CALL copy_fm_to_dbcsr(mo_coeff, mo_coeff_b) !fm->dbcsr

         t2 = m_walltime()
         IF (output_unit > 0) THEN
            WRITE (output_unit, '(T16,I5,T24,I6,T33,E12.4,2x,E12.4,T60,F8.3)') &
               iter, nmo_converged, max_norm, min_norm, t2-t1
         END IF
         t1 = m_walltime()

      END DO ! iter

      DEALLOCATE (ritz_coeff)
      DEALLOCATE (iconv)
      DEALLOCATE (inotconv)
      DEALLOCATE (vnorm)

      CALL timestop(handle)

   END SUBROUTINE generate_extended_space_sparse

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param s_block ...
!> \param h_block ...
!> \param v_block ...
!> \param w_block ...
!> \param evals ...
!> \param ndim ...
! **************************************************************************************************
   SUBROUTINE reduce_extended_space(s_block, h_block, v_block, w_block, evals, ndim)

      TYPE(cp_fm_type), POINTER                          :: s_block, h_block, v_block, w_block
      REAL(dp), DIMENSION(:)                             :: evals
      INTEGER                                            :: ndim

      CHARACTER(len=*), PARAMETER :: routineN = 'reduce_extended_space', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, info

      CALL timeset(routineN, handle)

      CALL cp_fm_to_fm(s_block, w_block)
      CALL cp_fm_cholesky_decompose(s_block, info_out=info)
      IF (info == 0) THEN
         CALL cp_fm_triangular_invert(s_block)
         CALL cp_fm_cholesky_restore(H_block, ndim, S_block, w_block, "MULTIPLY", pos="RIGHT")
         CALL cp_fm_cholesky_restore(w_block, ndim, S_block, H_block, "MULTIPLY", pos="LEFT", transa="T")
         CALL choose_eigv_solver(H_block, w_block, evals)
         CALL cp_fm_cholesky_restore(w_block, ndim, S_block, v_block, "MULTIPLY")
      ELSE
! S^(-1/2)
         CALL cp_fm_power(w_block, s_block, -0.5_dp, 1.0E-5_dp, info)
         CALL cp_fm_to_fm(w_block, s_block)
         CALL cp_gemm('N', 'N', ndim, ndim, ndim, 1.0_dp, H_block, s_block, 0.0_dp, w_block)
         CALL cp_gemm('N', 'N', ndim, ndim, ndim, 1.0_dp, s_block, w_block, 0.0_dp, H_block)
         CALL choose_eigv_solver(H_block, w_block, evals)
         CALL cp_gemm('N', 'N', ndim, ndim, ndim, 1.0_dp, s_block, w_block, 0.0_dp, v_block)
      END IF

      CALL timestop(handle)

   END SUBROUTINE reduce_extended_space

END MODULE qs_scf_block_davidson
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Interface to (sca)lapack for the Cholesky based procedures
!> \author  VW
!> \date    2009-09-08
!> \version 0.8
!>
!> <b>Modification history:</b>
!> - Created 2009-09-08
! **************************************************************************************************
MODULE cp_dbcsr_cholesky
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_get_info,&
                                              cp_dbcsr_type
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              copy_fm_to_dbcsr
   USE cp_fm_basic_linalg,              ONLY: cp_fm_upper_to_full
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp,&
                                              sp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_dbcsr_cholesky'

   PUBLIC :: cp_dbcsr_cholesky_decompose, cp_dbcsr_cholesky_invert, &
             cp_dbcsr_cholesky_restore

   PRIVATE

CONTAINS

! **************************************************************************************************
!> \brief used to replace a symmetric positive def. matrix M with its cholesky
!>      decomposition U: M = U^T * U, with U upper triangular
!> \param matrix the matrix to replace with its cholesky decomposition
!> \param n the number of row (and columns) of the matrix &
!>        (defaults to the min(size(matrix)))
!> \param para_env ...
!> \param blacs_env ...
!> \par History
!>      05.2002 created [JVdV]
!>      12.2002 updated, added n optional parm [fawzi]
!> \author Joost
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_cholesky_decompose(matrix, n, para_env, blacs_env)
      TYPE(cp_dbcsr_type)                      :: matrix
      INTEGER, INTENT(in), OPTIONAL            :: n
      TYPE(cp_para_env_type), POINTER          :: para_env
      TYPE(cp_blacs_env_type), POINTER         :: blacs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_dbcsr_cholesky_decompose', &
                                     routineP = moduleN//':'//routineN

      INTEGER                                  :: handle, info, my_n, &
                                                  nfullcols_total, &
                                                  nfullrows_total
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a
      REAL(KIND=sp), DIMENSION(:, :), POINTER  :: a_sp
      TYPE(cp_fm_struct_type), POINTER         :: fm_struct
      TYPE(cp_fm_type), POINTER                :: fm_matrix

      CALL timeset(routineN, handle)

      NULLIFY (fm_matrix, fm_struct)
      CALL cp_dbcsr_get_info(matrix, nfullrows_total=nfullrows_total, nfullcols_total=nfullcols_total)

      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=nfullrows_total, &
                               ncol_global=nfullcols_total, para_env=para_env)
      CALL cp_fm_create(fm_matrix, fm_struct, name="fm_matrix")
      CALL cp_fm_struct_release(fm_struct)

      CALL copy_dbcsr_to_fm(matrix, fm_matrix)

      my_n = MIN(fm_matrix%matrix_struct%nrow_global, &
                 fm_matrix%matrix_struct%ncol_global)
      IF (PRESENT(n)) THEN
         IF(.NOT.(n <= my_n))CALL cp__a("cp_dbcsr_cholesky.F",92)
         my_n = n
      END IF

      a => fm_matrix%local_data
      a_sp => fm_matrix%local_data_sp


      IF (fm_matrix%use_sp) THEN
         CALL spotrf('U', my_n, a_sp(1, 1), SIZE(a_sp, 1), info)
      ELSE
         CALL dpotrf('U', my_n, a(1, 1), SIZE(a, 1), info)
      ENDIF


      IF (info /= 0) &
         CALL cp__b("cp_dbcsr_cholesky.F",119,"Cholesky decomposition failed. Matrix ill conditioned ?")

      CALL copy_fm_to_dbcsr(fm_matrix, matrix)

      CALL cp_fm_release(fm_matrix)

      CALL timestop(handle)

   END SUBROUTINE cp_dbcsr_cholesky_decompose

! **************************************************************************************************
!> \brief used to replace the cholesky decomposition by the inverse
!> \param matrix the matrix to invert (must be an upper triangular matrix)
!> \param n size of the matrix to invert (defaults to the min(size(matrix)))
!> \param para_env ...
!> \param blacs_env ...
!> \param upper_to_full ...
!> \par History
!>      05.2002 created [JVdV]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_cholesky_invert(matrix, n, para_env, blacs_env, upper_to_full)
      TYPE(cp_dbcsr_type)                           :: matrix
      INTEGER, INTENT(in), OPTIONAL             :: n
      TYPE(cp_para_env_type), POINTER           :: para_env
      TYPE(cp_blacs_env_type), POINTER          :: blacs_env
      LOGICAL, INTENT(IN)                       :: upper_to_full

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_cholesky_invert', &
                                     routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a
      REAL(KIND=sp), DIMENSION(:, :), POINTER  :: a_sp
      INTEGER                                   :: info, handle
      INTEGER                                   :: my_n, nfullrows_total, nfullcols_total
      TYPE(cp_fm_type), POINTER                 :: fm_matrix, fm_matrix_tmp
      TYPE(cp_fm_struct_type), POINTER          :: fm_struct

      CALL timeset(routineN, handle)

      NULLIFY (fm_matrix, fm_struct)
      CALL cp_dbcsr_get_info(matrix, nfullrows_total=nfullrows_total, nfullcols_total=nfullcols_total)

      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=nfullrows_total, &
                               ncol_global=nfullrows_total, para_env=para_env)
      CALL cp_fm_create(fm_matrix, fm_struct, name="fm_matrix")
      CALL cp_fm_struct_release(fm_struct)

      CALL copy_dbcsr_to_fm(matrix, fm_matrix)

      my_n = MIN(fm_matrix%matrix_struct%nrow_global, &
                 fm_matrix%matrix_struct%ncol_global)
      IF (PRESENT(n)) THEN
         IF(.NOT.(n <= my_n))CALL cp__a("cp_dbcsr_cholesky.F",175)
         my_n = n
      END IF

      a => fm_matrix%local_data
      a_sp => fm_matrix%local_data_sp


      IF (fm_matrix%use_sp) THEN
         CALL spotri('U', my_n, a_sp(1, 1), SIZE(a_sp, 1), info)
      ELSE
         CALL dpotri('U', my_n, a(1, 1), SIZE(a, 1), info)
      ENDIF


      IF(.NOT.(info == 0))CALL cp__a("cp_dbcsr_cholesky.F",202)

      IF (upper_to_full) THEN
         CALL cp_fm_create(fm_matrix_tmp, fm_matrix%matrix_struct, name="fm_matrix_tmp")
         CALL cp_fm_upper_to_full(fm_matrix, fm_matrix_tmp)
         CALL cp_fm_release(fm_matrix_tmp)
      ENDIF

      CALL copy_fm_to_dbcsr(fm_matrix, matrix)

      CALL cp_fm_release(fm_matrix)

      CALL timestop(handle)

   END SUBROUTINE cp_dbcsr_cholesky_invert

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param neig ...
!> \param matrixb ...
!> \param matrixout ...
!> \param op ...
!> \param pos ...
!> \param transa ...
!> \param para_env ...
!> \param blacs_env ...
! **************************************************************************************************
   SUBROUTINE cp_dbcsr_cholesky_restore(matrix, neig, matrixb, matrixout, op, pos, transa, &
                                        para_env, blacs_env)
      TYPE(cp_dbcsr_type)                                :: matrix, matrixb, matrixout
      INTEGER, INTENT(IN)                            :: neig
      CHARACTER(LEN=*), INTENT(IN)           :: op
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL :: pos
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL :: transa
      TYPE(cp_para_env_type), POINTER                :: para_env
      TYPE(cp_blacs_env_type), POINTER               :: blacs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_cholesky_restore', &
                                     routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: a, b, out
      REAL(KIND=sp), DIMENSION(:, :), POINTER  :: a_sp, b_sp, out_sp
      INTEGER                                   :: itype, handle
      INTEGER                                   :: n
      REAL(KIND=dp)                           :: alpha
      INTEGER                                   :: myprow, mypcol, nfullrows_total, &
                                                   nfullcols_total
      TYPE(cp_blacs_env_type), POINTER          :: context
      CHARACTER                                 :: chol_pos, chol_transa
      TYPE(cp_fm_type), POINTER                 :: fm_matrix, fm_matrixb, fm_matrixout
      TYPE(cp_fm_struct_type), POINTER          :: fm_struct

      CALL timeset(routineN, handle)

      NULLIFY (fm_matrix, fm_matrixb, fm_matrixout, fm_struct)

      CALL cp_dbcsr_get_info(matrix, nfullrows_total=nfullrows_total, nfullcols_total=nfullcols_total)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=nfullrows_total, &
                               ncol_global=nfullcols_total, para_env=para_env)
      CALL cp_fm_create(fm_matrix, fm_struct, name="fm_matrix")
      CALL cp_fm_struct_release(fm_struct)

      CALL cp_dbcsr_get_info(matrixb, nfullrows_total=nfullrows_total, nfullcols_total=nfullcols_total)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=nfullrows_total, &
                               ncol_global=nfullcols_total, para_env=para_env)
      CALL cp_fm_create(fm_matrixb, fm_struct, name="fm_matrixb")
      CALL cp_fm_struct_release(fm_struct)

      CALL cp_dbcsr_get_info(matrixout, nfullrows_total=nfullrows_total, nfullcols_total=nfullcols_total)
      CALL cp_fm_struct_create(fm_struct, context=blacs_env, nrow_global=nfullrows_total, &
                               ncol_global=nfullcols_total, para_env=para_env)
      CALL cp_fm_create(fm_matrixout, fm_struct, name="fm_matrixout")
      CALL cp_fm_struct_release(fm_struct)

      CALL copy_dbcsr_to_fm(matrix, fm_matrix)
      CALL copy_dbcsr_to_fm(matrixb, fm_matrixb)
      !CALL copy_dbcsr_to_fm(matrixout, fm_matrixout)

      context => fm_matrix%matrix_struct%context
      myprow = context%mepos(1)
      mypcol = context%mepos(2)
      n = fm_matrix%matrix_struct%nrow_global
      itype = 1
      IF (op /= "SOLVE" .AND. op /= "MULTIPLY") &
         CALL cp__b("cp_dbcsr_cholesky.F",291,"wrong argument op")

      IF (PRESENT(pos)) THEN
         SELECT CASE (pos)
         CASE ("LEFT")
            chol_pos = 'L'
         CASE ("RIGHT")
            chol_pos = 'R'
         CASE DEFAULT
            CALL cp__b("cp_dbcsr_cholesky.F",300,"wrong argument pos")
         END SELECT
      ELSE
         chol_pos = 'L'
      ENDIF

      chol_transa = 'N'
      IF (PRESENT(transa)) chol_transa = transa

      IF ((fm_matrix%use_sp .NEQV. fm_matrixb%use_sp) .OR. (fm_matrix%use_sp .NEQV. fm_matrixout%use_sp)) &
         CALL cp__b("cp_dbcsr_cholesky.F",310,"not the same precision")

      ! notice b is the cholesky guy
      a => fm_matrix%local_data
      b => fm_matrixb%local_data
      out => fm_matrixout%local_data
      a_sp => fm_matrix%local_data_sp
      b_sp => fm_matrixb%local_data_sp
      out_sp => fm_matrixout%local_data_sp


      alpha = 1.0_dp
      IF (fm_matrix%use_sp) THEN
         CALL scopy(neig*n, a_sp(1, 1), 1, out_sp(1, 1), 1)
      ELSE
         CALL dcopy(neig*n, a(1, 1), 1, out(1, 1), 1)
      ENDIF
      IF (op .EQ. "SOLVE") THEN
         IF (fm_matrix%use_sp) THEN
            CALL strsm(chol_pos, 'U', chol_transa, 'N', n, neig, REAL(alpha, sp), b_sp(1, 1), SIZE(b_sp, 1), out_sp(1, 1), n)
         ELSE
            CALL dtrsm(chol_pos, 'U', chol_transa, 'N', n, neig, alpha, b(1, 1), SIZE(b, 1), out(1, 1), n)
         ENDIF
      ELSE
         IF (fm_matrix%use_sp) THEN
            CALL strmm(chol_pos, 'U', chol_transa, 'N', n, neig, REAL(alpha, sp), b_sp(1, 1), n, out_sp(1, 1), n)
         ELSE
            CALL dtrmm(chol_pos, 'U', chol_transa, 'N', n, neig, alpha, b(1, 1), n, out(1, 1), n)
         ENDIF
      ENDIF


      CALL copy_fm_to_dbcsr(fm_matrixout, matrixout)

      CALL cp_fm_release(fm_matrix)
      CALL cp_fm_release(fm_matrixb)
      CALL cp_fm_release(fm_matrixout)

      CALL timestop(handle)

   END SUBROUTINE cp_dbcsr_cholesky_restore

END MODULE cp_dbcsr_cholesky

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief orbital transformations
!> \par History
!>      Added Taylor expansion based computation of the matrix functions (01.2004)
!>      added additional rotation variables for non-equivalent occupied orbs (08.2004)
!> \author Joost VandeVondele (06.2002)
! **************************************************************************************************
MODULE qs_ot
   USE cp_dbcsr_cholesky,               ONLY: cp_dbcsr_cholesky_decompose,&
                                              cp_dbcsr_cholesky_invert,&
                                              cp_dbcsr_cholesky_restore
   USE cp_dbcsr_diag,                   ONLY: cp_dbcsr_heevd,&
                                              cp_dbcsr_syevd
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_add_on_diag, cp_dbcsr_arnoldi_extremal, cp_dbcsr_copy, &
        cp_dbcsr_distribution, cp_dbcsr_filter, cp_dbcsr_frobenius_norm, cp_dbcsr_gershgorin_norm, &
        cp_dbcsr_get_block_p, cp_dbcsr_get_info, cp_dbcsr_get_occupation, &
        cp_dbcsr_hadamard_product, cp_dbcsr_init, cp_dbcsr_init_p, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_release, cp_dbcsr_release_p, &
        cp_dbcsr_scale, cp_dbcsr_scale_by_vector, cp_dbcsr_set, cp_dbcsr_transposed, &
        cp_dbcsr_type, dbcsr_distribution_mp, dbcsr_mp_group
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE preconditioner,                  ONLY: apply_preconditioner
   USE preconditioner_types,            ONLY: preconditioner_type
   USE qs_ot_types,                     ONLY: qs_ot_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC  :: qs_ot_get_p
   PUBLIC  :: qs_ot_get_orbitals
   PUBLIC  :: qs_ot_get_derivative
   PUBLIC  :: qs_ot_get_orbitals_ref
   PUBLIC  :: qs_ot_get_derivative_ref
   PUBLIC  :: qs_ot_new_preconditioner
   PRIVATE :: qs_ot_p2m_diag
   PRIVATE :: qs_ot_sinc
   PRIVATE :: qs_ot_ref_poly
   PRIVATE :: qs_ot_ref_chol
   PRIVATE :: qs_ot_ref_lwdn
   PRIVATE :: qs_ot_ref_decide
   PRIVATE :: qs_ot_ref_update
   PRIVATE :: qs_ot_refine
   PRIVATE :: qs_ot_on_the_fly_localize

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_ot'

CONTAINS

   ! gets ready to use the preconditioner/ or renew the preconditioner
   ! only keeps a pointer to the preconditioner.
   ! If you change the preconditioner, you have to call this routine
   ! you remain responsible of proper deallocate of your preconditioner
   ! (or you can reuse it on the next step of the computation)
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
!> \param preconditioner ...
! **************************************************************************************************
   SUBROUTINE qs_ot_new_preconditioner(qs_ot_env, preconditioner)
      TYPE(qs_ot_type)                                   :: qs_ot_env
      TYPE(preconditioner_type), POINTER                 :: preconditioner

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_new_preconditioner', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ncoef
      LOGICAL                                            :: mixed_precision

      qs_ot_env%preconditioner => preconditioner
      qs_ot_env%os_valid = .FALSE.
      IF (.NOT. ASSOCIATED(qs_ot_env%matrix_psc0)) THEN
         CALL cp_dbcsr_init_p(qs_ot_env%matrix_psc0)
         CALL cp_dbcsr_copy(qs_ot_env%matrix_psc0, qs_ot_env%matrix_sc0, 'matrix_psc0')
      ENDIF

      mixed_precision = qs_ot_env%settings%mixed_precision

      IF (.NOT. qs_ot_env%use_dx) THEN
         qs_ot_env%use_dx = .TRUE.
         CALL cp_dbcsr_init_p(qs_ot_env%matrix_dx)
         CALL cp_dbcsr_copy(qs_ot_env%matrix_dx, qs_ot_env%matrix_gx, 'matrix_dx')
         IF (qs_ot_env%settings%do_rotation) THEN
            CALL cp_dbcsr_init_p(qs_ot_env%rot_mat_dx)
            CALL cp_dbcsr_copy(qs_ot_env%rot_mat_dx, qs_ot_env%rot_mat_gx, 'rot_mat_dx')
         ENDIF
         IF (qs_ot_env%settings%do_ener) THEN
            ncoef = SIZE(qs_ot_env%ener_gx)
            ALLOCATE (qs_ot_env%ener_dx(ncoef))
            qs_ot_env%ener_dx = 0.0_dp
         ENDIF
      ENDIF

   END SUBROUTINE qs_ot_new_preconditioner

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
!> \param C_NEW ...
!> \param SC ...
!> \param G_OLD ...
!> \param D ...
! **************************************************************************************************
   SUBROUTINE qs_ot_on_the_fly_localize(qs_ot_env, C_NEW, SC, G_OLD, D)
      !
      TYPE(qs_ot_type)                                   :: qs_ot_env
      TYPE(cp_dbcsr_type), POINTER                       :: C_NEW, SC, G_OLD, D

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_on_the_fly_localize', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: taylor_order = 50
      REAL(KIND=dp), PARAMETER                           :: alpha = 0.1_dp, f2_eps = 0.01_dp

      INTEGER                                            :: blk, col, col_size, handle, i, k, n, &
                                                            output_unit, p, row, row_size
      REAL(dp), DIMENSION(:, :), POINTER                 :: DATA
      REAL(KIND=dp)                                      :: expfactor, f2, norm_fro, norm_gct, tmp
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_type), POINTER                       :: C, Gp1, Gp2, GU, U

      CALL timeset(routineN, handle)
      !
      !
      CALL cp_dbcsr_get_info(C_NEW, nfullrows_total=n, nfullcols_total=k)
      !
      ! C = C*expm(-G)
      GU => qs_ot_env%buf1_k_k_nosym ! a buffer
      U => qs_ot_env%buf2_k_k_nosym ! a buffer
      Gp1 => qs_ot_env%buf3_k_k_nosym ! a buffer
      Gp2 => qs_ot_env%buf4_k_k_nosym ! a buffer
      C => qs_ot_env%buf1_n_k ! a buffer
      !
      ! compute the derivative of the norm
      !-------------------------------------------------------------------
      ! (x^2+eps)^1/2
      f2 = 0.0_dp
      CALL cp_dbcsr_copy(C, C_NEW)
      CALL cp_dbcsr_iterator_start(iter, C)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, DATA, blk, &
                                           row_size=row_size, col_size=col_size)
         DO p = 1, col_size ! p
         DO i = 1, row_size ! i
            tmp = SQRT(DATA(i, p)**2+f2_eps)
            f2 = f2+tmp
            DATA(i, p) = DATA(i, p)/tmp
         ENDDO
         ENDDO
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)
      CALL mp_sum(f2, dbcsr_mp_group(dbcsr_distribution_mp(cp_dbcsr_distribution(C))))
      output_unit = cp_logger_get_default_io_unit()
      IF (output_unit > 0) WRITE (output_unit, *) routineN//' f2 =', f2
      !
      !
      CALL cp_dbcsr_multiply('T', 'N', 1.0_dp, C, C_NEW, 0.0_dp, GU)
      !
      ! antisymetrize
      CALL cp_dbcsr_transposed(U, GU, shallow_data_copy=.FALSE., &
                               use_distribution=cp_dbcsr_distribution(GU), &
                               transpose_distribution=.FALSE.)
      CALL cp_dbcsr_add(GU, U, alpha_scalar=-0.5_dp, beta_scalar=0.5_dp)
      !-------------------------------------------------------------------
      !
      norm_fro = cp_dbcsr_frobenius_norm(GU)
      norm_gct = cp_dbcsr_gershgorin_norm(GU)
      !write(*,*) 'qs_ot_localize: ||P-I||_f=',norm_fro,' ||P-I||_GCT=',norm_gct
      !
      !kscale = CEILING(LOG(MIN(norm_fro,norm_gct))/LOG(2.0_dp))
      !scale  = LOG(MIN(norm_fro,norm_gct))/LOG(2.0_dp)
      !write(*,*) 'qs_ot_localize: scale=',scale,' kscale=',kscale
      !
      ! rescale for steepest descent
      CALL cp_dbcsr_scale(GU, -alpha)
      !
      ! compute unitary transform
      ! zeroth and first order
      expfactor = 1.0_dp
      CALL cp_dbcsr_copy(U, GU)
      CALL cp_dbcsr_scale(U, expfactor)
      CALL cp_dbcsr_add_on_diag(U, 1.0_dp)
      ! other orders
      CALL cp_dbcsr_copy(Gp1, GU)
      DO i = 2, taylor_order
         ! new power of G
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, GU, Gp1, 0.0_dp, Gp2)
         CALL cp_dbcsr_copy(Gp1, Gp2)
         ! add to the taylor expansion so far
         expfactor = expfactor/REAL(i, KIND=dp)
         CALL cp_dbcsr_add(U, Gp1, alpha_scalar=1.0_dp, beta_scalar=expfactor)
         norm_fro = cp_dbcsr_frobenius_norm(Gp1)
         !write(*,*) 'Taylor expansion i=',i,' norm(X^i)/i!=',norm_fro*expfactor
         IF (norm_fro*expfactor .LT. 1.0E-10_dp) EXIT
      ENDDO
      !
      ! rotate MOs
      CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, C_NEW, U, 0.0_dp, C)
      CALL cp_dbcsr_copy(C_NEW, C)
      !
      ! rotate SC
      CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, SC, U, 0.0_dp, C)
      CALL cp_dbcsr_copy(SC, C)
      !
      ! rotate D_i
      CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, D, U, 0.0_dp, C)
      CALL cp_dbcsr_copy(D, C)
      !
      ! rotate G_i-1
      IF (ASSOCIATED(G_OLD)) THEN
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, G_OLD, U, 0.0_dp, C)
         CALL cp_dbcsr_copy(G_OLD, C)
      ENDIF
      !
      CALL timestop(handle)
   END SUBROUTINE qs_ot_on_the_fly_localize

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
!> \param C_OLD ...
!> \param C_TMP ...
!> \param C_NEW ...
!> \param P ...
!> \param SC ...
!> \param update ...
! **************************************************************************************************
   SUBROUTINE qs_ot_ref_chol(qs_ot_env, C_OLD, C_TMP, C_NEW, P, SC, update)
      !
      TYPE(qs_ot_type)                                   :: qs_ot_env
      TYPE(cp_dbcsr_type)                                :: C_OLD, C_TMP, C_NEW, P, SC
      LOGICAL, INTENT(IN)                                :: update

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_ref_chol', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, k, n

      CALL timeset(routineN, handle)
      !
      CALL cp_dbcsr_get_info(C_NEW, nfullrows_total=n, nfullcols_total=k)
      !
      ! P = U'*U
      CALL cp_dbcsr_cholesky_decompose(P, k, qs_ot_env%para_env, qs_ot_env%blacs_env)
      !
      ! C_NEW = C_OLD*inv(U)
      CALL cp_dbcsr_cholesky_restore(C_OLD, k, P, C_NEW, op="SOLVE", pos="RIGHT", &
                                     transa="N", para_env=qs_ot_env%para_env, blacs_env=qs_ot_env%blacs_env)
      !
      ! Update SC if needed
      IF (update) THEN
         CALL cp_dbcsr_cholesky_restore(SC, k, P, C_TMP, op="SOLVE", pos="RIGHT", &
                                        transa="N", para_env=qs_ot_env%para_env, blacs_env=qs_ot_env%blacs_env)
         CALL cp_dbcsr_copy(SC, C_TMP)
      ENDIF
      !
      CALL timestop(handle)
   END SUBROUTINE qs_ot_ref_chol

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
!> \param C_OLD ...
!> \param C_TMP ...
!> \param C_NEW ...
!> \param P ...
!> \param SC ...
!> \param update ...
! **************************************************************************************************
   SUBROUTINE qs_ot_ref_lwdn(qs_ot_env, C_OLD, C_TMP, C_NEW, P, SC, update)
      !
      TYPE(qs_ot_type)                                   :: qs_ot_env
      TYPE(cp_dbcsr_type)                                :: C_OLD, C_TMP, C_NEW, P, SC
      LOGICAL, INTENT(IN)                                :: update

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_ref_lwdn', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, k, n
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: eig, fun
      TYPE(cp_dbcsr_type), POINTER                       :: V, W

      CALL timeset(routineN, handle)
      !
      CALL cp_dbcsr_get_info(C_NEW, nfullrows_total=n, nfullcols_total=k)
      !
      V => qs_ot_env%buf1_k_k_nosym ! a buffer
      W => qs_ot_env%buf2_k_k_nosym ! a buffer
      ALLOCATE (eig(k), fun(k))
      !
      CALL cp_dbcsr_syevd(P, V, eig, qs_ot_env%para_env, qs_ot_env%blacs_env)
      !
      ! compute the P^(-1/2)
      DO i = 1, k
         IF (eig(i) .LE. 0.0_dp) &
            CALL cp__b("qs_ot.F",302,"P not positive definite")
         IF (eig(i) .LT. 1.0E-8_dp) THEN
            fun(i) = 0.0_dp
         ELSE
            fun(i) = 1.0_dp/SQRT(eig(i))
         ENDIF
      ENDDO
      CALL cp_dbcsr_copy(W, V)
      CALL cp_dbcsr_scale_by_vector(V, alpha=fun, side='right')
      CALL cp_dbcsr_multiply('N', 'T', 1.0_dp, W, V, 0.0_dp, P)
      !
      ! Update C
      CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, C_OLD, P, 0.0_dp, C_NEW)
      !
      ! Update SC if needed
      IF (update) THEN
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, SC, P, 0.0_dp, C_TMP)
         CALL cp_dbcsr_copy(SC, C_TMP)
      ENDIF
      !
      DEALLOCATE (eig, fun)
      !
      CALL timestop(handle)
   END SUBROUTINE qs_ot_ref_lwdn

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
!> \param C_OLD ...
!> \param C_TMP ...
!> \param C_NEW ...
!> \param P ...
!> \param SC ...
!> \param norm_in ...
!> \param update ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE qs_ot_ref_poly(qs_ot_env, C_OLD, C_TMP, C_NEW, P, SC, norm_in, update, output_unit)
      !
      TYPE(qs_ot_type)                                   :: qs_ot_env
      TYPE(cp_dbcsr_type), POINTER                       :: C_OLD, C_TMP, C_NEW, P
      TYPE(cp_dbcsr_type)                                :: SC
      REAL(dp), INTENT(IN)                               :: norm_in
      LOGICAL, INTENT(IN)                                :: update
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_ref_poly', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, irefine, k, n
      LOGICAL                                            :: quick_exit
      REAL(dp)                                           :: norm, norm_fro, norm_gct, occ_in, &
                                                            occ_out, rescale
      TYPE(cp_dbcsr_type), POINTER                       :: BUF1, BUF2, BUF_NOSYM, FT, FY

      CALL timeset(routineN, handle)
      !
      CALL cp_dbcsr_get_info(C_NEW, nfullrows_total=n, nfullcols_total=k)
      !
      BUF_NOSYM => qs_ot_env%buf1_k_k_nosym ! a buffer
      BUF1 => qs_ot_env%buf1_k_k_sym ! a buffer
      BUF2 => qs_ot_env%buf2_k_k_sym ! a buffer
      FY => qs_ot_env%buf3_k_k_sym ! a buffer
      FT => qs_ot_env%buf4_k_k_sym ! a buffer
      !
      ! initialize the norm (already computed in qs_ot_get_orbitals_ref)
      norm = norm_in
      !
      ! can we do a quick exit?
      quick_exit = .FALSE.
      IF (norm .LT. qs_ot_env%settings%eps_irac_quick_exit) quick_exit = .TRUE.
      !
      ! lets refine
      rescale = 1.0_dp
      DO irefine = 1, qs_ot_env%settings%max_irac
         !
         ! rescaling
         IF (norm .GT. 1.0_dp) THEN
            IF (output_unit > 0) WRITE (output_unit, '(A,I3,A)') &
               routineN, irefine, ': we rescale (C+a*D)'
            CALL cp_dbcsr_scale(P, 1.0_dp/norm)
            rescale = rescale/SQRT(norm)
         ENDIF
         !
         ! get the refinement polynomial
         CALL qs_ot_refine(P, FY, BUF1, BUF2, qs_ot_env%settings%irac_degree, &
                           qs_ot_env%settings%eps_irac_filter_matrix, output_unit)
         !
         ! collect the transformation
         IF (irefine .EQ. 1) THEN
            CALL cp_dbcsr_copy(FT, FY, name='FT')
         ELSE
            CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, FT, FY, 0.0_dp, BUF1)
            IF (qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) THEN
               occ_in = cp_dbcsr_get_occupation(buf1)
               CALL cp_dbcsr_filter(buf1, qs_ot_env%settings%eps_irac_filter_matrix)
               occ_out = cp_dbcsr_get_occupation(buf1)
               IF (output_unit > 0 .AND. qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
                  WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(BUF1): occ_in', occ_in, ' occ_out', occ_out
            ENDIF
            CALL cp_dbcsr_copy(FT, BUF1, name='FT')
         ENDIF
         !
         ! quick exit if possible
         IF (quick_exit) THEN
            IF (output_unit > 0) WRITE (output_unit, '(A,I3,A)') &
               routineN, irefine, ': quick exit!'
            EXIT
         ENDIF
         !
         ! P = FY^T * P * FY
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, P, FY, 0.0_dp, BUF_NOSYM)
         IF (qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(buf_nosym)
            CALL cp_dbcsr_filter(buf_nosym, qs_ot_env%settings%eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(buf_nosym)
            IF (output_unit > 0 .AND. qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(BUF_NOSYM): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, FY, BUF_NOSYM, 0.0_dp, P)
         IF (qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(p)
            CALL cp_dbcsr_filter(p, qs_ot_env%settings%eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(p)
            IF (output_unit > 0 .AND. qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(P): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         !
         ! check ||P-1||_gct
         CALL cp_dbcsr_add_on_diag(P, -1.0_dp)
         norm_fro = cp_dbcsr_frobenius_norm(P)
         norm_gct = cp_dbcsr_gershgorin_norm(P)
         CALL cp_dbcsr_add_on_diag(P, 1.0_dp)
         norm = MIN(norm_gct, norm_fro)
         !
         ! printing
         IF (output_unit > 0) WRITE (output_unit, '(A,I3,A,E12.5)') &
            routineN, irefine, ': ||P-I||=', norm
         !
         ! blows up
         IF (norm > 1.0E10_dp) THEN
            CALL cp_abort(cp__l("qs_ot.F",442), &
                          "Refinement blows up! "// &
                          "We need you to improve the code, please post your input on "// &
                          "the forum https://www.cp2k.org/")
         END IF
         !
         ! can we do a quick exit next step?
         IF (norm .LT. qs_ot_env%settings%eps_irac_quick_exit) quick_exit = .TRUE.
         !
         ! are we done?
         IF (norm .LT. qs_ot_env%settings%eps_irac) EXIT
         !
      ENDDO
      !
      ! C_NEW = C_NEW * FT * rescale
      CALL cp_dbcsr_multiply('N', 'N', rescale, C_OLD, FT, 0.0_dp, C_NEW)
      IF (qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) THEN
         occ_in = cp_dbcsr_get_occupation(c_new)
         CALL cp_dbcsr_filter(c_new, qs_ot_env%settings%eps_irac_filter_matrix)
         occ_out = cp_dbcsr_get_occupation(c_new)
         IF (output_unit > 0 .AND. qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
            WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(C_NEW): occ_in', occ_in, ' occ_out', occ_out
      ENDIF
      !
      ! update SC = SC * FY * rescale
      IF (update) THEN
         CALL cp_dbcsr_multiply('N', 'N', rescale, SC, FT, 0.0_dp, C_TMP)
         IF (qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(c_tmp)
            CALL cp_dbcsr_filter(c_tmp, qs_ot_env%settings%eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(c_tmp)
            IF (output_unit > 0 .AND. qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(C_TMP): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         CALL cp_dbcsr_copy(SC, C_TMP)
      ENDIF
      !
      CALL timestop(handle)
   END SUBROUTINE qs_ot_ref_poly

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env1 ...
!> \retval update ...
! **************************************************************************************************
   FUNCTION qs_ot_ref_update(qs_ot_env1) RESULT(update)
      !
      TYPE(qs_ot_type)                                   :: qs_ot_env1
      LOGICAL                                            :: update

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_ref_update', &
         routineP = moduleN//':'//routineN

      update = .FALSE.
      SELECT CASE (qs_ot_env1%settings%ot_method)
      CASE ("CG")
         SELECT CASE (qs_ot_env1%settings%line_search_method)
         CASE ("2PNT")
            IF (qs_ot_env1%line_search_count .EQ. 2) update = .TRUE.
         CASE DEFAULT
            CALL cp__b("qs_ot.F",502,"NYI")
         END SELECT
      CASE ("DIIS")
         update = .TRUE.
      CASE DEFAULT
         CALL cp__b("qs_ot.F",507,"NYI")
      END SELECT
   END FUNCTION qs_ot_ref_update

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env1 ...
!> \param norm_in ...
!> \param ortho_irac ...
! **************************************************************************************************
   SUBROUTINE qs_ot_ref_decide(qs_ot_env1, norm_in, ortho_irac)
      !
      TYPE(qs_ot_type)                                   :: qs_ot_env1
      REAL(dp), INTENT(IN)                               :: norm_in
      CHARACTER(LEN=*), INTENT(INOUT)                    :: ortho_irac

      ortho_irac = qs_ot_env1%settings%ortho_irac
      IF (norm_in .LT. qs_ot_env1%settings%eps_irac_switch) ortho_irac = "POLY"
   END SUBROUTINE qs_ot_ref_decide

! **************************************************************************************************
!> \brief ...
!> \param matrix_c ...
!> \param matrix_s ...
!> \param matrix_x ...
!> \param matrix_sx ...
!> \param matrix_gx_old ...
!> \param matrix_dx ...
!> \param qs_ot_env ...
!> \param qs_ot_env1 ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE qs_ot_get_orbitals_ref(matrix_c, matrix_s, matrix_x, matrix_sx, &
                                     matrix_gx_old, matrix_dx, qs_ot_env, &
                                     qs_ot_env1, output_unit)
      !
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_c, matrix_s, matrix_x, matrix_sx, &
                                                            matrix_gx_old, matrix_dx
      TYPE(qs_ot_type)                                   :: qs_ot_env, qs_ot_env1
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_get_orbitals_ref', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=4)                                   :: ortho_irac
      INTEGER                                            :: handle, k, n
      LOGICAL                                            :: on_the_fly_loc, update
      REAL(dp)                                           :: norm, norm_fro, norm_gct, occ_in, occ_out
      TYPE(cp_dbcsr_type), POINTER                       :: C_NEW, C_OLD, C_TMP, D, G_OLD, P, S, SC

      CALL timeset(routineN, handle)

      IF (output_unit > 0 .AND. qs_ot_env1%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
         WRITE (output_unit, *) routinen//' eps_irac_filter_matrix=', &
         qs_ot_env1%settings%eps_irac_filter_matrix

      CALL cp_dbcsr_get_info(matrix_c, nfullrows_total=n, nfullcols_total=k)
      !
      C_NEW => matrix_c
      C_OLD => matrix_x ! need to be carefully updated for the gradient !
      SC => matrix_sx ! need to be carefully updated for the gradient !
      G_OLD => matrix_gx_old ! need to be carefully updated for localization !
      D => matrix_dx ! need to be carefully updated for localization !
      S => matrix_s

      P => qs_ot_env%p_k_k_sym ! a buffer
      C_TMP => qs_ot_env%buf1_n_k ! a buffer
      !
      ! do we need to update C_OLD and SC?
      update = qs_ot_ref_update(qs_ot_env1)
      !
      ! do we want to on the fly localize?
      ! for the moment this is set from the input,
      ! later we might want to localize every n-step or
      ! when the sparsity increases...
      on_the_fly_loc = qs_ot_env1%settings%on_the_fly_loc
      !
      ! compute SC = S*C
      IF (ASSOCIATED(S)) THEN
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, S, C_OLD, 0.0_dp, SC)
         IF (qs_ot_env1%settings%eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(sc)
            CALL cp_dbcsr_filter(sc, qs_ot_env1%settings%eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(sc)
            IF (output_unit > 0 .AND. qs_ot_env1%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(SC): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
      ELSE
         CALL cp_dbcsr_copy(SC, C_OLD)
      ENDIF
      !
      ! compute P = C'*SC
      CALL cp_dbcsr_multiply('T', 'N', 1.0_dp, C_OLD, SC, 0.0_dp, P)
      IF (qs_ot_env1%settings%eps_irac_filter_matrix .GT. 0.0_dp) THEN
         occ_in = cp_dbcsr_get_occupation(p)
         CALL cp_dbcsr_filter(p, qs_ot_env1%settings%eps_irac_filter_matrix)
         occ_out = cp_dbcsr_get_occupation(p)
         IF (output_unit > 0 .AND. qs_ot_env1%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
            WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(P): occ_in', occ_in, ' occ_out', occ_out
      ENDIF
      !
      ! check ||P-1||_f and ||P-1||_gct
      CALL cp_dbcsr_add_on_diag(P, -1.0_dp)
      norm_fro = cp_dbcsr_frobenius_norm(P)
      norm_gct = cp_dbcsr_gershgorin_norm(P)
      CALL cp_dbcsr_add_on_diag(P, 1.0_dp)
      norm = MIN(norm_gct, norm_fro)
      CALL qs_ot_ref_decide(qs_ot_env1, norm, ortho_irac)
      IF (output_unit > 0) WRITE (output_unit, '(A,I3,A,E12.5,A)') &
             routineN, 0, ': ||P-I||=', norm,&
           & ', ortho_irac = '//ortho_irac
      !
      ! select the orthogonality method
      SELECT CASE (ortho_irac)
      CASE ("CHOL")
         CALL qs_ot_ref_chol(qs_ot_env, C_OLD, C_TMP, C_NEW, P, SC, update)
      CASE ("LWDN")
         CALL qs_ot_ref_lwdn(qs_ot_env, C_OLD, C_TMP, C_NEW, P, SC, update)
      CASE ("POLY")
         CALL qs_ot_ref_poly(qs_ot_env, C_OLD, C_TMP, C_NEW, P, SC, norm, update, output_unit)
      CASE DEFAULT
         CALL cp__b("qs_ot.F",628,"Wrong argument")
      END SELECT
      !
      ! We update the C_i+1 and localization
      IF (update) THEN
         IF (on_the_fly_loc) THEN
            IF (output_unit > 0) WRITE (output_unit, '(A)') &
               routineN//' we localize C'
            CALL qs_ot_on_the_fly_localize(qs_ot_env, C_NEW, SC, G_OLD, D)
         ENDIF
         CALL cp_dbcsr_copy(C_OLD, C_NEW)
      ENDIF
      !
      CALL timestop(handle)
   END SUBROUTINE qs_ot_get_orbitals_ref

! **************************************************************************************************
!> \brief ...
!> \param P ...
!> \param FY ...
!> \param P2 ...
!> \param T ...
!> \param irac_degree ...
!> \param eps_irac_filter_matrix ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE qs_ot_refine(P, FY, P2, T, irac_degree, eps_irac_filter_matrix, output_unit)
      !----------------------------------------------------------------------
      ! refinement polynomial of degree 2,3 and 4 (PRB 70, 193102 (2004))
      !----------------------------------------------------------------------

      TYPE(cp_dbcsr_type), INTENT(inout)                 :: P, FY, P2, T
      INTEGER, INTENT(in)                                :: irac_degree
      REAL(dp), INTENT(in)                               :: eps_irac_filter_matrix
      INTEGER, INTENT(in)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_refine', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, k
      REAL(dp)                                           :: occ_in, occ_out, r

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_get_info(P, nfullcols_total=k)
      SELECT CASE (irac_degree)
      CASE (2)
         ! C_out = C_in * ( 15/8 * I - 10/8 * P + 3/8 * P^2)
         r = 3.0_dp/8.0_dp
         CALL cp_dbcsr_multiply('N', 'N', r, P, P, 0.0_dp, FY)
         IF (eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(fy)
            CALL cp_dbcsr_filter(fy, eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(fy)
            IF (output_unit > 0 .AND. eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(FY): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         r = -10.0_dp/8.0_dp
         CALL cp_dbcsr_add(FY, P, alpha_scalar=1.0_dp, beta_scalar=r)
         r = 15.0_dp/8.0_dp
         CALL cp_dbcsr_add_on_diag(FY, alpha_scalar=r)
      CASE (3)
         ! C_out = C_in * ( 35/16 * I - 35/16 * P + 21/16 * P^2 - 5/16 P^3)
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, P, P, 0.0_dp, P2)
         IF (eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(p2)
            CALL cp_dbcsr_filter(p2, eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(p2)
            IF (output_unit > 0 .AND. eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(P2): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         r = -5.0_dp/16.0_dp
         CALL cp_dbcsr_multiply('N', 'N', r, P2, P, 0.0_dp, FY)
         IF (eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(fy)
            CALL cp_dbcsr_filter(fy, eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(fy)
            IF (output_unit > 0 .AND. eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(FY): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         r = 21.0_dp/16.0_dp
         CALL cp_dbcsr_add(FY, P2, alpha_scalar=1.0_dp, beta_scalar=r)
         r = -35.0_dp/16.0_dp
         CALL cp_dbcsr_add(FY, P, alpha_scalar=1.0_dp, beta_scalar=r)
         r = 35.0_dp/16.0_dp
         CALL cp_dbcsr_add_on_diag(FY, alpha_scalar=r)
      CASE (4)
         ! C_out = C_in * ( 315/128 * I - 420/128 * P + 378/128 * P^2 - 180/128 P^3 + 35/128 P^4 )
         !       = C_in * ( 315/128 * I - 420/128 * P + 378/128 * P^2 + ( - 180/128 * P + 35/128 * P^2 ) * P^2 )
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, P, P, 0.0_dp, P2) ! P^2
         IF (eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(p2)
            CALL cp_dbcsr_filter(p2, eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(p2)
            IF (output_unit > 0 .AND. eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(P2): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         r = -180.0_dp/128.0_dp
         CALL cp_dbcsr_add(T, P, alpha_scalar=0.0_dp, beta_scalar=r) ! T=-180/128*P
         r = 35.0_dp/128.0_dp
         CALL cp_dbcsr_add(T, P2, alpha_scalar=1.0_dp, beta_scalar=r) ! T=T+35/128*P^2
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, T, P2, 0.0_dp, FY) ! Y=T*P^2
         IF (eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(fy)
            CALL cp_dbcsr_filter(fy, eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(fy)
            IF (output_unit > 0 .AND. eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(FY): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         r = 378.0_dp/128.0_dp
         CALL cp_dbcsr_add(FY, P2, alpha_scalar=1.0_dp, beta_scalar=r) ! Y=Y+378/128*P^2
         r = -420.0_dp/128.0_dp
         CALL cp_dbcsr_add(FY, P, alpha_scalar=1.0_dp, beta_scalar=r) ! Y=Y-420/128*P
         r = 315.0_dp/128.0_dp
         CALL cp_dbcsr_add_on_diag(FY, alpha_scalar=r) ! Y=Y+315/128*I
      CASE DEFAULT
         CALL cp__b("qs_ot.F",743,"This irac_order NYI")
      END SELECT
      CALL timestop(handle)
   END SUBROUTINE qs_ot_refine

! **************************************************************************************************
!> \brief ...
!> \param matrix_hc ...
!> \param matrix_x ...
!> \param matrix_sx ...
!> \param matrix_gx ...
!> \param qs_ot_env ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE qs_ot_get_derivative_ref(matrix_hc, matrix_x, matrix_sx, matrix_gx, &
                                       qs_ot_env, output_unit)
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_hc, matrix_x, matrix_sx, matrix_gx
      TYPE(qs_ot_type)                                   :: qs_ot_env
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_get_derivative_ref', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, k, n
      LOGICAL                                            :: mixed_precision
      REAL(dp)                                           :: occ_in, occ_out
      TYPE(cp_dbcsr_type), POINTER                       :: C, CHC, G, G_dp, HC, SC

      CALL timeset(routineN, handle)

      mixed_precision = qs_ot_env%settings%mixed_precision
      CALL cp_dbcsr_get_info(matrix_x, nfullrows_total=n, nfullcols_total=k)
      !
      C => matrix_x ! NBsf*NOcc
      SC => matrix_sx ! NBsf*NOcc need to be up2date
      HC => matrix_hc ! NBsf*NOcc
      G => matrix_gx ! NBsf*NOcc
      CHC => qs_ot_env%buf1_k_k_sym ! buffer
      G_dp => qs_ot_env%buf1_n_k_dp ! buffer dp

      IF (mixed_precision) THEN
         ! C'*(H*C)
         CALL cp_dbcsr_multiply('T', 'N', 1.0_dp, C, HC, 0.0_dp, CHC)
         IF (qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(chc)
            CALL cp_dbcsr_filter(chc, qs_ot_env%settings%eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(chc)
            IF (output_unit > 0 .AND. qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(CHC): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         ! (S*C)*(C'*H*C)
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, SC, CHC, 0.0_dp, G_dp)
         IF (qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(g_dp)
            CALL cp_dbcsr_filter(g_dp, qs_ot_env%settings%eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(g_dp)
            IF (output_unit > 0 .AND. qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(G_dp): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         ! G = 2*(1-S*C*C')*H*C
         CALL cp_dbcsr_add(G_dp, HC, alpha_scalar=-1.0_dp, beta_scalar=1.0_dp)
         CALL cp_dbcsr_copy(G, G_dp)
      ELSE
         ! C'*(H*C)
         CALL cp_dbcsr_multiply('T', 'N', 1.0_dp, C, HC, 0.0_dp, CHC)
         IF (qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(chc)
            CALL cp_dbcsr_filter(chc, qs_ot_env%settings%eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(chc)
            IF (output_unit > 0 .AND. qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(CHC): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         ! (S*C)*(C'*H*C)
         CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, SC, CHC, 0.0_dp, G)
         IF (qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) THEN
            occ_in = cp_dbcsr_get_occupation(g)
            CALL cp_dbcsr_filter(g, qs_ot_env%settings%eps_irac_filter_matrix)
            occ_out = cp_dbcsr_get_occupation(g)
            IF (output_unit > 0 .AND. qs_ot_env%settings%eps_irac_filter_matrix .GT. 0.0_dp) &
               WRITE (output_unit, '(2(A,F8.5))') routinen//' filter(G): occ_in', occ_in, ' occ_out', occ_out
         ENDIF
         ! G = 2*(1-S*C*C')*H*C
         CALL cp_dbcsr_add(G, HC, alpha_scalar=-1.0_dp, beta_scalar=1.0_dp)
      ENDIF
      !
      CALL timestop(handle)
   END SUBROUTINE qs_ot_get_derivative_ref

   ! computes p=x*S*x and the matrix functionals related matrices
! **************************************************************************************************
!> \brief ...
!> \param matrix_x ...
!> \param matrix_sx ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE qs_ot_get_p(matrix_x, matrix_sx, qs_ot_env)

      TYPE(cp_dbcsr_type), POINTER                       :: matrix_x, matrix_sx
      TYPE(qs_ot_type)                                   :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_get_p', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, k, max_iter, n
      LOGICAL                                            :: converged
      REAL(KIND=dp)                                      :: max_ev, min_ev, threshold

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_get_info(matrix_x, nfullrows_total=n, nfullcols_total=k)

      ! get the overlap
      CALL cp_dbcsr_multiply('T', 'N', rone, matrix_x, matrix_sx, rzero, &
                             qs_ot_env%matrix_p)

      ! get an upper bound for the largest eigenvalue
      ! try using lancos first and fall back to gershgorin norm if it fails
      max_iter = 30; threshold = 1.0E-03_dp
      CALL cp_dbcsr_arnoldi_extremal(qs_ot_env%matrix_p, max_ev, min_ev, converged, threshold, max_iter)
      qs_ot_env%largest_eval_upper_bound = MAX(max_ev, ABS(min_ev))

      IF (.NOT. converged) qs_ot_env%largest_eval_upper_bound = cp_dbcsr_gershgorin_norm(qs_ot_env%matrix_p)
      CALL decide_strategy(qs_ot_env)
      IF (qs_ot_env%do_taylor) THEN
         CALL qs_ot_p2m_taylor(qs_ot_env)
      ELSE
         CALL qs_ot_p2m_diag(qs_ot_env)
      ENDIF

      IF (qs_ot_env%settings%do_rotation) THEN
         CALL qs_ot_generate_rotation(qs_ot_env)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE qs_ot_get_p

! **************************************************************************************************
!> \brief computes the rotation matrix rot_mat_u that is associated to a given
!>      rot_mat_x using rot_mat_u=exp(rot_mat_x)
!> \param qs_ot_env a valid qs_ot_env
!> \par History
!>      08.2004 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE qs_ot_generate_rotation(qs_ot_env)

      TYPE(qs_ot_type)                                   :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_generate_rotation', &
         routineP = moduleN//':'//routineN
      COMPLEX(KIND=dp), PARAMETER                        :: cone = (1.0_dp, 0.0_dp), &
                                                            czero = (0.0_dp, 0.0_dp)

      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: evals_exp
      COMPLEX(KIND=dp), DIMENSION(:), POINTER            :: data_z
      INTEGER                                            :: blk, col, handle, k, row
      LOGICAL                                            :: found
      REAL(KIND=dp), DIMENSION(:), POINTER               :: data_d
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_type)                                :: cmat_u, cmat_x

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_get_info(qs_ot_env%rot_mat_x, nfullrows_total=k)

      IF (k /= 0) THEN
         CALL cp_dbcsr_init(cmat_x)
         CALL cp_dbcsr_init(cmat_u)
         CALL cp_dbcsr_copy(cmat_x, qs_ot_env%rot_mat_evec, name='cmat_x')
         CALL cp_dbcsr_copy(cmat_u, qs_ot_env%rot_mat_evec, name='cmat_u')
         ALLOCATE (evals_exp(k))

         ! rot_mat_u = exp(rot_mat_x)
         ! i rot_mat_x is hermitian, so go over the complex variables for diag
         !vwCALL cp_cfm_get_info(cmat_x,local_data=local_data_c)
         !vwCALL cp_fm_get_info(qs_ot_env%rot_mat_x,local_data=local_data_r)
         !vwlocal_data_c=CMPLX(0.0_dp,local_data_r,KIND=dp)
         CALL cp_dbcsr_iterator_start(iter, cmat_x)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row, col, data_z, blk)
            CALL cp_dbcsr_get_block_p(qs_ot_env%rot_mat_x, row, col, data_d, found)
            IF (.NOT. found) THEN
               WRITE (*, *) routineN//' .NOT.found'
               !stop
            ELSE
               data_z = CMPLX(0.0_dp, data_d, KIND=dp)
            ENDIF
         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)

         CALL cp_dbcsr_heevd(cmat_x, qs_ot_env%rot_mat_evec, qs_ot_env%rot_mat_evals, &
                             qs_ot_env%para_env, qs_ot_env%blacs_env)
         evals_exp(:) = EXP((0.0_dp, -1.0_dp)*qs_ot_env%rot_mat_evals(:))
         CALL cp_dbcsr_copy(cmat_x, qs_ot_env%rot_mat_evec)
         CALL cp_dbcsr_scale_by_vector(cmat_x, alpha=evals_exp, side='right')
         CALL cp_dbcsr_multiply('N', 'C', cone, cmat_x, qs_ot_env%rot_mat_evec, czero, cmat_u)
         CALL cp_dbcsr_copy(qs_ot_env%rot_mat_u, cmat_u, keep_imaginary=.FALSE.)
         CALL cp_dbcsr_release(cmat_x)
         CALL cp_dbcsr_release(cmat_u)
         DEALLOCATE (evals_exp)
      END IF

      CALL timestop(handle)

   END SUBROUTINE qs_ot_generate_rotation

! **************************************************************************************************
!> \brief computes the derivative fields with respect to rot_mat_x
!> \param qs_ot_env valid qs_ot_env. In particular qs_ot_generate_rotation has to be called before
!>                        and the rot_mat_dedu matrix has to be up to date
!> \par History
!>      08.2004 created [ Joost VandeVondele ]
! **************************************************************************************************
   SUBROUTINE qs_ot_rot_mat_derivative(qs_ot_env)
      TYPE(qs_ot_type)                         :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_rot_mat_derivative', &
                                     routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp), PARAMETER              :: cone = (1.0_dp, 0.0_dp), &
                                                  czero = (0.0_dp, 0.0_dp)

      INTEGER                                  :: handle, i, j, k
      REAL(KIND=dp)                            :: e1, e2
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: data_d
      TYPE(cp_dbcsr_type)                          :: cmat_buf1, cmat_buf2
      TYPE(cp_dbcsr_iterator)                     :: iter
      COMPLEX(dp), DIMENSION(:, :), POINTER :: data_z
      INTEGER::row, col, blk, row_offset, col_offset, row_size, col_size
      LOGICAL::found

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_get_info(qs_ot_env%rot_mat_u, nfullrows_total=k)
      IF (k /= 0) THEN
         CALL cp_dbcsr_copy(qs_ot_env%matrix_buf1, qs_ot_env%rot_mat_dedu)
         ! now we get to the derivative wrt the antisymmetric matrix rot_mat_x
         CALL cp_dbcsr_init(cmat_buf1)
         CALL cp_dbcsr_init(cmat_buf2)
         CALL cp_dbcsr_copy(cmat_buf1, qs_ot_env%rot_mat_evec, "cmat_buf1")
         CALL cp_dbcsr_copy(cmat_buf2, qs_ot_env%rot_mat_evec, "cmat_buf2")

         ! init cmat_buf1
         !CALL cp_fm_get_info(qs_ot_env%matrix_buf1,matrix_struct=fm_struct, local_data=local_data_r)
         !CALL cp_cfm_get_info(cmat_buf1, nrow_local=nrow_local,   ncol_local=ncol_local, &
         !     row_indices=row_indices, col_indices=col_indices, &
         !     local_data=local_data_c)
         !local_data_c=local_data_r

         CALL cp_dbcsr_iterator_start(iter, cmat_buf1)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row, col, data_z, blk)
            CALL cp_dbcsr_get_block_p(qs_ot_env%matrix_buf1, row, col, data_d, found)
            data_z = data_d
         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)

         CALL cp_dbcsr_multiply('T', 'N', cone, cmat_buf1, qs_ot_env%rot_mat_evec, &
                                czero, cmat_buf2)
         CALL cp_dbcsr_multiply('C', 'N', cone, qs_ot_env%rot_mat_evec, cmat_buf2, &
                                czero, cmat_buf1)

         CALL cp_dbcsr_iterator_start(iter, cmat_buf1)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, row, col, data_z, blk, &
                                              row_size=row_size, col_size=col_size, &
                                              row_offset=row_offset, col_offset=col_offset)
            DO j = 1, col_size
            DO i = 1, row_size
               e1 = qs_ot_env%rot_mat_evals(row_offset+i-1)
               e2 = qs_ot_env%rot_mat_evals(col_offset+j-1)
               data_z(i, j) = data_z(i, j)*cint(e1, e2)
            ENDDO
            ENDDO
         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)

         CALL cp_dbcsr_multiply('N', 'N', cone, qs_ot_env%rot_mat_evec, cmat_buf1, &
                                czero, cmat_buf2)
         CALL cp_dbcsr_multiply('N', 'C', cone, cmat_buf2, qs_ot_env%rot_mat_evec, &
                                czero, cmat_buf1)

         CALL cp_dbcsr_copy(qs_ot_env%matrix_buf1, cmat_buf1)

         CALL cp_dbcsr_transposed(qs_ot_env%matrix_buf2, qs_ot_env%matrix_buf1, &
                                  shallow_data_copy=.FALSE., use_distribution=cp_dbcsr_distribution(qs_ot_env%matrix_buf3), &
                                  transpose_distribution=.FALSE.)
         CALL cp_dbcsr_add(qs_ot_env%matrix_buf1, qs_ot_env%matrix_buf2, &
                           alpha_scalar=-1.0_dp, beta_scalar=+1.0_dp)
         CALL cp_dbcsr_copy(qs_ot_env%rot_mat_gx, qs_ot_env%matrix_buf1)
         CALL cp_dbcsr_release(cmat_buf1)
         CALL cp_dbcsr_release(cmat_buf2)
      END IF
      CALL timestop(handle)
   CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param e1 ...
!> \param e2 ...
!> \retval cint ...
! **************************************************************************************************
      FUNCTION cint(e1, e2)
      REAL(KIND=dp)                                      :: e1, e2
      COMPLEX(KIND=dp)                                   :: cint

      COMPLEX(KIND=dp)                                   :: l1, l2, x
      INTEGER                                            :: I

         l1 = (0.0_dp, -1.0_dp)*e1
         l2 = (0.0_dp, -1.0_dp)*e2
         IF (ABS(l1-l2) .GT. 0.5_dp) THEN
            cint = (EXP(l1)-EXP(l2))/(l1-l2)
         ELSE
            x = 1.0_dp
            cint = 0.0_dp
            DO I = 1, 16
               cint = cint+x
               x = x*(l1-l2)/REAL(I+1, KIND=dp)
            ENDDO
            cint = cint*EXP(l2)
         ENDIF
      END FUNCTION cint
   END SUBROUTINE qs_ot_rot_mat_derivative

   !
   ! decide strategy
   ! tries to decide if the taylor expansion of cos(sqrt(xsx)) converges rapidly enough
   ! to make a taylor expansion of the functions cos(sqrt(xsx)) and sin(sqrt(xsx))/sqrt(xsx)
   ! and their derivatives faster than their computation based on diagonalization
   ! since xsx can be very small, especially during dynamics, only a few terms might indeed be needed
   ! we find the necessary order N to have largest_eval_upper_bound**(N+1)/(2(N+1))! < eps_taylor
   !
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE decide_strategy(qs_ot_env)
      TYPE(qs_ot_type)                                   :: qs_ot_env

      INTEGER                                            :: N
      REAL(KIND=dp)                                      :: num_error

      qs_ot_env%do_taylor = .FALSE.
      N = 0
      num_error = qs_ot_env%largest_eval_upper_bound/(2.0_dp)
      DO WHILE (num_error > qs_ot_env%settings%eps_taylor .AND. N <= qs_ot_env%settings%max_taylor)
         N = N+1
         num_error = num_error*qs_ot_env%largest_eval_upper_bound/REAL((2*N+1)*(2*N+2), KIND=dp)
      END DO
      qs_ot_env%taylor_order = N
      IF (qs_ot_env%taylor_order <= qs_ot_env%settings%max_taylor) THEN
         qs_ot_env%do_taylor = .TRUE.
      ENDIF

   END SUBROUTINE decide_strategy

   ! c=(c0*cos(p^0.5)+x*sin(p^0.5)*p^(-0.5)) x rot_mat_u
   ! this assumes that x is already ortho to S*C0, and that p is x*S*x
   ! rot_mat_u is an optional rotation matrix
! **************************************************************************************************
!> \brief ...
!> \param matrix_c ...
!> \param matrix_x ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE qs_ot_get_orbitals(matrix_c, matrix_x, qs_ot_env)

      TYPE(cp_dbcsr_type), POINTER                       :: matrix_c, matrix_x
      TYPE(qs_ot_type)                                   :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_get_orbitals', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, k, n
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_kk

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_get_info(matrix_x, nfullrows_total=n, nfullcols_total=k)

      ! rotate the multiplying matrices cosp and sinp instead of the result,
      ! this should be cheaper for large basis sets
      IF (qs_ot_env%settings%do_rotation) THEN
         matrix_kk => qs_ot_env%matrix_buf1
         CALL cp_dbcsr_multiply('N', 'N', rone, qs_ot_env%matrix_cosp, &
                                qs_ot_env%rot_mat_u, rzero, matrix_kk)
      ELSE
         matrix_kk => qs_ot_env%matrix_cosp
      ENDIF

      CALL cp_dbcsr_multiply('N', 'N', rone, qs_ot_env%matrix_c0, matrix_kk, &
                             rzero, matrix_c)

      IF (qs_ot_env%settings%do_rotation) THEN
         matrix_kk => qs_ot_env%matrix_buf1
         CALL cp_dbcsr_multiply('N', 'N', rone, qs_ot_env%matrix_sinp, &
                                qs_ot_env%rot_mat_u, rzero, matrix_kk)
      ELSE
         matrix_kk => qs_ot_env%matrix_sinp
      ENDIF
      CALL cp_dbcsr_multiply('N', 'N', rone, matrix_x, matrix_kk, &
                             rone, matrix_c)

      CALL timestop(handle)

   END SUBROUTINE qs_ot_get_orbitals

! **************************************************************************************************
   ! this routines computes dE/dx=dx, with dx ortho to sc0
   ! needs dE/dC=hc,C0,X,SX,p
   ! if preconditioned it will not be the derivative, but the lagrangian multiplier
   ! is changed so that P*dE/dx is the right derivative (i.e. in the allowed subspace)
! **************************************************************************************************
!> \brief ...
!> \param matrix_hc ...
!> \param matrix_x ...
!> \param matrix_sx ...
!> \param matrix_gx ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE qs_ot_get_derivative(matrix_hc, matrix_x, matrix_sx, matrix_gx, &
                                   qs_ot_env)
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_hc, matrix_x, matrix_sx, matrix_gx
      TYPE(qs_ot_type)                                   :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_get_derivative', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, k, n, ortho_k
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_hc_local, matrix_target

      CALL timeset(routineN, handle)

      NULLIFY (matrix_hc_local)

      CALL cp_dbcsr_get_info(matrix_x, nfullrows_total=n, nfullcols_total=k)

      ! could in principle be taken inside qs_ot_get_derivative_* for increased efficiency
      ! create a local rotated version of matrix_hc leaving matrix_hc untouched (needed
      ! for lagrangian multipliers)
      IF (qs_ot_env%settings%do_rotation) THEN
         CALL cp_dbcsr_copy(matrix_gx, matrix_hc) ! use gx as temporary
         CALL cp_dbcsr_init_p(matrix_hc_local)
         CALL cp_dbcsr_copy(matrix_hc_local, matrix_hc, name='matrix_hc_local')
         CALL cp_dbcsr_set(matrix_hc_local, 0.0_dp)
         CALL cp_dbcsr_multiply('N', 'T', rone, matrix_gx, qs_ot_env%rot_mat_u, rzero, matrix_hc_local)
      ELSE
         matrix_hc_local => matrix_hc
      ENDIF

      IF (qs_ot_env%do_taylor) THEN
         CALL qs_ot_get_derivative_taylor(matrix_hc_local, matrix_x, matrix_sx, matrix_gx, qs_ot_env)
      ELSE
         CALL qs_ot_get_derivative_diag(matrix_hc_local, matrix_x, matrix_sx, matrix_gx, qs_ot_env)
      ENDIF

      ! and make it orthogonal
      CALL cp_dbcsr_get_info(qs_ot_env%matrix_sc0, nfullcols_total=ortho_k)

      IF (ASSOCIATED(qs_ot_env%preconditioner)) THEN
         matrix_target => qs_ot_env%matrix_psc0
      ELSE
         matrix_target => qs_ot_env%matrix_sc0
      ENDIF
      ! first make the matrix os if not yet valid
      IF (.NOT. qs_ot_env%os_valid) THEN
         ! this assumes that the preconditioner is a single matrix
         ! that maps sc0 onto psc0

         IF (ASSOCIATED(qs_ot_env%preconditioner)) THEN
            CALL apply_preconditioner(qs_ot_env%preconditioner, qs_ot_env%matrix_sc0, &
                                      qs_ot_env%matrix_psc0)
         ENDIF
         CALL cp_dbcsr_multiply('T', 'N', rone, &
                                qs_ot_env%matrix_sc0, matrix_target, &
                                rzero, qs_ot_env%matrix_os)
         CALL cp_dbcsr_cholesky_decompose(qs_ot_env%matrix_os, &
                                          para_env=qs_ot_env%para_env, blacs_env=qs_ot_env%blacs_env)
         CALL cp_dbcsr_cholesky_invert(qs_ot_env%matrix_os, &
                                       para_env=qs_ot_env%para_env, blacs_env=qs_ot_env%blacs_env, &
                                       upper_to_full=.TRUE.)
         qs_ot_env%os_valid = .TRUE.
      ENDIF
      CALL cp_dbcsr_multiply('T', 'N', rone, matrix_target, matrix_gx, &
                             rzero, qs_ot_env%matrix_buf1_ortho)
      CALL cp_dbcsr_multiply('N', 'N', rone, qs_ot_env%matrix_os, &
                             qs_ot_env%matrix_buf1_ortho, rzero, qs_ot_env%matrix_buf2_ortho)
      CALL cp_dbcsr_multiply('N', 'N', -rone, qs_ot_env%matrix_sc0, &
                             qs_ot_env%matrix_buf2_ortho, rone, matrix_gx)
      ! also treat the rot_mat gradient here
      IF (qs_ot_env%settings%do_rotation) THEN
         CALL qs_ot_rot_mat_derivative(qs_ot_env)
      ENDIF

      IF (qs_ot_env%settings%do_rotation) THEN
         CALL cp_dbcsr_release_p(matrix_hc_local)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE qs_ot_get_derivative

! **************************************************************************************************
!> \brief ...
!> \param matrix_hc ...
!> \param matrix_x ...
!> \param matrix_sx ...
!> \param matrix_gx ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE qs_ot_get_derivative_diag(matrix_hc, matrix_x, matrix_sx, matrix_gx, &
                                        qs_ot_env)

      TYPE(cp_dbcsr_type), POINTER                       :: matrix_hc, matrix_x, matrix_sx, matrix_gx
      TYPE(qs_ot_type)                                   :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_get_derivative_diag', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, k, n

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_get_info(matrix_x, nfullrows_total=n, nfullcols_total=k)

      ! go for the derivative now
      ! this de/dc*(dX/dx)*sinp
      CALL cp_dbcsr_multiply('N', 'N', rone, matrix_hc, qs_ot_env%matrix_sinp, rzero, matrix_gx)
      ! overlap hc*x
      CALL cp_dbcsr_multiply('T', 'N', rone, matrix_hc, matrix_x, rzero, qs_ot_env%matrix_buf2)
      ! get it in the basis of the eigenvectors
      CALL cp_dbcsr_multiply('N', 'N', rone, qs_ot_env%matrix_buf2, qs_ot_env%matrix_r, &
                             rzero, qs_ot_env%matrix_buf1)
      CALL cp_dbcsr_multiply('T', 'N', rone, qs_ot_env%matrix_r, qs_ot_env%matrix_buf1, &
                             rzero, qs_ot_env%matrix_buf2)

      ! get the schur product of O_uv*B_uv
      CALL cp_dbcsr_hadamard_product(qs_ot_env%matrix_buf2, qs_ot_env%matrix_sinp_b, &
                                     qs_ot_env%matrix_buf3)

      ! overlap hc*c0
      CALL cp_dbcsr_multiply('T', 'N', rone, matrix_hc, qs_ot_env%matrix_c0, rzero, &
                             qs_ot_env%matrix_buf2)
      ! get it in the basis of the eigenvectors
      CALL cp_dbcsr_multiply('N', 'N', rone, qs_ot_env%matrix_buf2, qs_ot_env%matrix_r, &
                             rzero, qs_ot_env%matrix_buf1)
      CALL cp_dbcsr_multiply('T', 'N', rone, qs_ot_env%matrix_r, qs_ot_env%matrix_buf1, &
                             rzero, qs_ot_env%matrix_buf2)

      CALL cp_dbcsr_hadamard_product(qs_ot_env%matrix_buf2, qs_ot_env%matrix_cosp_b, &
                                     qs_ot_env%matrix_buf4)

      ! add the two bs and compute b+b^T
      CALL cp_dbcsr_add(qs_ot_env%matrix_buf3, qs_ot_env%matrix_buf4, &
                        alpha_scalar=rone, beta_scalar=rone)

      ! get the b in the eigenvector basis
      CALL cp_dbcsr_multiply('N', 'T', rone, qs_ot_env%matrix_buf3, qs_ot_env%matrix_r, &
                             rzero, qs_ot_env%matrix_buf1)
      CALL cp_dbcsr_multiply('N', 'N', rone, qs_ot_env%matrix_r, qs_ot_env%matrix_buf1, &
                             rzero, qs_ot_env%matrix_buf3)
      CALL cp_dbcsr_transposed(qs_ot_env%matrix_buf1, qs_ot_env%matrix_buf3, &
                               shallow_data_copy=.FALSE., use_distribution=cp_dbcsr_distribution(qs_ot_env%matrix_buf3), &
                               transpose_distribution=.FALSE.)
      CALL cp_dbcsr_add(qs_ot_env%matrix_buf3, qs_ot_env%matrix_buf1, &
                        alpha_scalar=rone, beta_scalar=rone)

      ! and add to the derivative
      CALL cp_dbcsr_multiply('N', 'N', rone, matrix_sx, qs_ot_env%matrix_buf3, &
                             rone, matrix_gx)
      CALL timestop(handle)

   END SUBROUTINE qs_ot_get_derivative_diag

   ! compute the derivative of the taylor expansion below
! **************************************************************************************************
!> \brief ...
!> \param matrix_hc ...
!> \param matrix_x ...
!> \param matrix_sx ...
!> \param matrix_gx ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE qs_ot_get_derivative_taylor(matrix_hc, matrix_x, matrix_sx, matrix_gx, &
                                          qs_ot_env)

      TYPE(cp_dbcsr_type), POINTER                       :: matrix_hc, matrix_x, matrix_sx, matrix_gx
      TYPE(qs_ot_type)                                   :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_get_derivative_taylor', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, i, k, n
      REAL(KIND=dp)                                      :: cosfactor, sinfactor
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_left, matrix_right

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_get_info(matrix_x, nfullrows_total=n, nfullcols_total=k)

      ! go for the derivative now
      ! this de/dc*(dX/dx)*sinp i.e. zeroth order
      CALL cp_dbcsr_multiply('N', 'N', rone, matrix_hc, qs_ot_env%matrix_sinp, rzero, matrix_gx)

      IF (qs_ot_env%taylor_order .LE. 0) THEN
         CALL timestop(handle)
         RETURN
      ENDIF

      ! we store the matrix that will multiply sx in matrix_r
      CALL cp_dbcsr_set(qs_ot_env%matrix_r, rzero)

      ! just better names for matrix_cosp_b and matrix_sinp_b (they are buffer space here)
      matrix_left => qs_ot_env%matrix_cosp_b
      matrix_right => qs_ot_env%matrix_sinp_b

      ! overlap hc*x and add its transpose to matrix_left
      CALL cp_dbcsr_multiply('T', 'N', rone, matrix_hc, matrix_x, rzero, matrix_left)
      CALL cp_dbcsr_transposed(qs_ot_env%matrix_buf1, matrix_left, &
                               shallow_data_copy=.FALSE., use_distribution=cp_dbcsr_distribution(matrix_left), &
                               transpose_distribution=.FALSE.)
      CALL cp_dbcsr_add(matrix_left, qs_ot_env%matrix_buf1, &
                        alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
      CALL cp_dbcsr_copy(matrix_right, matrix_left)

      ! first order
      sinfactor = -1.0_dp/(2.0_dp*3.0_dp)
      CALL cp_dbcsr_add(qs_ot_env%matrix_r, matrix_left, &
                        alpha_scalar=1.0_dp, beta_scalar=sinfactor)

      !      M
      !    OM+MO
      ! OOM+OMO+MOO
      !   ...
      DO i = 2, qs_ot_env%taylor_order
         sinfactor = sinfactor*(-1.0_dp)/REAL(2*i*(2*i+1), KIND=dp)
         CALL cp_dbcsr_multiply('N', 'N', rone, qs_ot_env%matrix_p, matrix_left, rzero, qs_ot_env%matrix_buf1)
         CALL cp_dbcsr_multiply('N', 'N', rone, matrix_right, qs_ot_env%matrix_p, rzero, matrix_left)
         CALL cp_dbcsr_copy(matrix_right, matrix_left)
         CALL cp_dbcsr_add(matrix_left, qs_ot_env%matrix_buf1, &
                           1.0_dp, 1.0_dp)
         CALL cp_dbcsr_add(qs_ot_env%matrix_r, matrix_left, &
                           alpha_scalar=1.0_dp, beta_scalar=sinfactor)
      ENDDO

      ! overlap hc*c0 and add its transpose to matrix_left
      CALL cp_dbcsr_multiply('T', 'N', rone, matrix_hc, qs_ot_env%matrix_c0, rzero, matrix_left)
      CALL cp_dbcsr_transposed(qs_ot_env%matrix_buf1, matrix_left, &
                               shallow_data_copy=.FALSE., use_distribution=cp_dbcsr_distribution(matrix_left), &
                               transpose_distribution=.FALSE.)
      CALL cp_dbcsr_add(matrix_left, qs_ot_env%matrix_buf1, 1.0_dp, 1.0_dp)
      CALL cp_dbcsr_copy(matrix_right, matrix_left)

      ! first order
      cosfactor = -1.0_dp/(1.0_dp*2.0_dp)
      CALL cp_dbcsr_add(qs_ot_env%matrix_r, matrix_left, &
                        alpha_scalar=1.0_dp, beta_scalar=cosfactor)

      !      M
      !    OM+MO
      ! OOM+OMO+MOO
      !   ...
      DO i = 2, qs_ot_env%taylor_order
         cosfactor = cosfactor*(-1.0_dp)/REAL(2*i*(2*i-1), KIND=dp)
         CALL cp_dbcsr_multiply('N', 'N', rone, qs_ot_env%matrix_p, matrix_left, rzero, qs_ot_env%matrix_buf1)
         CALL cp_dbcsr_multiply('N', 'N', rone, matrix_right, qs_ot_env%matrix_p, rzero, matrix_left)
         CALL cp_dbcsr_copy(matrix_right, matrix_left)
         CALL cp_dbcsr_add(matrix_left, qs_ot_env%matrix_buf1, 1.0_dp, 1.0_dp)
         CALL cp_dbcsr_add(qs_ot_env%matrix_r, matrix_left, &
                           alpha_scalar=1.0_dp, beta_scalar=cosfactor)
      ENDDO

      ! and add to the derivative
      CALL cp_dbcsr_multiply('N', 'N', rone, matrix_sx, qs_ot_env%matrix_r, rone, matrix_gx)

      CALL timestop(handle)

   END SUBROUTINE qs_ot_get_derivative_taylor

   ! computes a taylor expansion.
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE qs_ot_p2m_taylor(qs_ot_env)
      TYPE(qs_ot_type)                                   :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_p2m_taylor', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, i, k
      REAL(KIND=dp)                                      :: cosfactor, sinfactor

      CALL timeset(routineN, handle)

      ! zeroth order
      CALL cp_dbcsr_set(qs_ot_env%matrix_cosp, rzero)
      CALL cp_dbcsr_set(qs_ot_env%matrix_sinp, rzero)
      CALL cp_dbcsr_add_on_diag(qs_ot_env%matrix_cosp, rone)
      CALL cp_dbcsr_add_on_diag(qs_ot_env%matrix_sinp, rone)

      IF (qs_ot_env%taylor_order .LE. 0) THEN
         CALL timestop(handle)
         RETURN
      ENDIF

      ! first order
      cosfactor = -1.0_dp/(1.0_dp*2.0_dp)
      sinfactor = -1.0_dp/(2.0_dp*3.0_dp)
      CALL cp_dbcsr_add(qs_ot_env%matrix_cosp, qs_ot_env%matrix_p, alpha_scalar=1.0_dp, beta_scalar=cosfactor)
      CALL cp_dbcsr_add(qs_ot_env%matrix_sinp, qs_ot_env%matrix_p, alpha_scalar=1.0_dp, beta_scalar=sinfactor)
      IF (qs_ot_env%taylor_order .LE. 1) THEN
         CALL timestop(handle)
         RETURN
      ENDIF

      ! other orders
      CALL cp_dbcsr_get_info(qs_ot_env%matrix_p, nfullrows_total=k)
      CALL cp_dbcsr_copy(qs_ot_env%matrix_r, qs_ot_env%matrix_p)

      DO i = 2, qs_ot_env%taylor_order
         ! new power of p
         CALL cp_dbcsr_multiply('N', 'N', rone, qs_ot_env%matrix_p, qs_ot_env%matrix_r, &
                                rzero, qs_ot_env%matrix_buf1)
         CALL cp_dbcsr_copy(qs_ot_env%matrix_r, qs_ot_env%matrix_buf1)
         ! add to the taylor expansion so far
         cosfactor = cosfactor*(-1.0_dp)/REAL(2*i*(2*i-1), KIND=dp)
         sinfactor = sinfactor*(-1.0_dp)/REAL(2*i*(2*i+1), KIND=dp)
         CALL cp_dbcsr_add(qs_ot_env%matrix_cosp, qs_ot_env%matrix_r, &
                           alpha_scalar=1.0_dp, beta_scalar=cosfactor)
         CALL cp_dbcsr_add(qs_ot_env%matrix_sinp, qs_ot_env%matrix_r, &
                           alpha_scalar=1.0_dp, beta_scalar=sinfactor)
      ENDDO

      CALL timestop(handle)

   END SUBROUTINE qs_ot_p2m_taylor

   ! given p, computes  - eigenstuff (matrix_r,evals)
   !                    - cos(p^0.5),p^(-0.5)*sin(p^0.5)
   !                    - the real b matrices, needed for the derivatives of these guys
   !                    cosp_b_ij=(1/(2pii) * int(cos(z^1/2)/((z-eval(i))*(z-eval(j))))
   !                    sinp_b_ij=(1/(2pii) * int(z^(-1/2)*sin(z^1/2)/((z-eval(i))*(z-eval(j))))
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE qs_ot_p2m_diag(qs_ot_env)

      TYPE(qs_ot_type)                                   :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_p2m_diag', routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: blk, col, col_offset, col_size, handle, &
                                                            i, j, k, row, row_offset, row_size
      REAL(dp), DIMENSION(:, :), POINTER                 :: DATA
      REAL(KIND=dp)                                      :: a, b
      TYPE(cp_dbcsr_iterator)                            :: iter

      CALL timeset(routineN, handle)

      CALL cp_dbcsr_get_info(qs_ot_env%matrix_p, nfullrows_total=k)
      CALL cp_dbcsr_copy(qs_ot_env%matrix_buf1, qs_ot_env%matrix_p)
      CALL cp_dbcsr_syevd(qs_ot_env%matrix_buf1, qs_ot_env%matrix_r, qs_ot_env%evals, &
                          qs_ot_env%para_env, qs_ot_env%blacs_env)
      DO i = 1, k
         qs_ot_env%evals(i) = MAX(0.0_dp, qs_ot_env%evals(i))
      ENDDO

!$OMP     PARALLEL DO DEFAULT(NONE) PRIVATE(i) SHARED(k,qs_ot_env)
      DO i = 1, k
         qs_ot_env%dum(i) = COS(SQRT(qs_ot_env%evals(i)))
      ENDDO
      CALL cp_dbcsr_copy(qs_ot_env%matrix_buf1, qs_ot_env%matrix_r)
      CALL cp_dbcsr_scale_by_vector(qs_ot_env%matrix_buf1, alpha=qs_ot_env%dum, side='right')
      CALL cp_dbcsr_multiply('N', 'T', rone, qs_ot_env%matrix_r, qs_ot_env%matrix_buf1, &
                             rzero, qs_ot_env%matrix_cosp)

!$OMP     PARALLEL DO DEFAULT(NONE) PRIVATE(i) SHARED(k,qs_ot_env)
      DO i = 1, k
         qs_ot_env%dum(i) = qs_ot_sinc(SQRT(qs_ot_env%evals(i)))
      ENDDO
      CALL cp_dbcsr_copy(qs_ot_env%matrix_buf1, qs_ot_env%matrix_r)
      CALL cp_dbcsr_scale_by_vector(qs_ot_env%matrix_buf1, alpha=qs_ot_env%dum, side='right')
      CALL cp_dbcsr_multiply('N', 'T', rone, qs_ot_env%matrix_r, qs_ot_env%matrix_buf1, &
                             rzero, qs_ot_env%matrix_sinp)

      CALL cp_dbcsr_copy(qs_ot_env%matrix_cosp_b, qs_ot_env%matrix_cosp)
      CALL cp_dbcsr_iterator_start(iter, qs_ot_env%matrix_cosp_b)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, DATA, &
                                           block_number=blk, row_size=row_size, col_size=col_size, &
                                           row_offset=row_offset, col_offset=col_offset)
         DO j = 1, col_size
         DO i = 1, row_size
            a = (SQRT(qs_ot_env%evals(row_offset+i-1)) &
                 -SQRT(qs_ot_env%evals(col_offset+j-1)))/2.0_dp
            b = (SQRT(qs_ot_env%evals(row_offset+i-1)) &
                 +SQRT(qs_ot_env%evals(col_offset+j-1)))/2.0_dp
            DATA(i, j) = -0.5_dp*qs_ot_sinc(a)*qs_ot_sinc(b)
         ENDDO
         ENDDO
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL cp_dbcsr_copy(qs_ot_env%matrix_sinp_b, qs_ot_env%matrix_sinp)
      CALL cp_dbcsr_iterator_start(iter, qs_ot_env%matrix_sinp_b)
      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
         CALL cp_dbcsr_iterator_next_block(iter, row, col, DATA, &
                                           block_number=blk, row_size=row_size, col_size=col_size, &
                                           row_offset=row_offset, col_offset=col_offset)
         DO j = 1, col_size
         DO i = 1, row_size
            a = SQRT(qs_ot_env%evals(row_offset+i-1))
            b = SQRT(qs_ot_env%evals(col_offset+j-1))
            DATA(i, j) = qs_ot_sincf(a, b)
         ENDDO
         ENDDO
      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL timestop(handle)

   END SUBROUTINE qs_ot_p2m_diag

   ! computes sin(x)/x for all values of the argument
! **************************************************************************************************
!> \brief ...
!> \param x ...
!> \retval qs_ot_sinc ...
! **************************************************************************************************
   FUNCTION qs_ot_sinc(x)

      REAL(KIND=dp), INTENT(IN)                          :: x
      REAL(KIND=dp)                                      :: qs_ot_sinc

      REAL(KIND=dp), PARAMETER :: q1 = 1.0_dp, q2 = -q1/(2.0_dp*3.0_dp), q3 = -q2/(4.0_dp*5.0_dp), &
         q4 = -q3/(6.0_dp*7.0_dp), q5 = -q4/(8.0_dp*9.0_dp), q6 = -q5/(10.0_dp*11.0_dp), &
         q7 = -q6/(12.0_dp*13.0_dp), q8 = -q7/(14.0_dp*15.0_dp), q9 = -q8/(16.0_dp*17.0_dp), &
         q10 = -q9/(18.0_dp*19.0_dp)

      REAL(KIND=dp)                                      :: y

      IF (ABS(x) > 0.5_dp) THEN
         qs_ot_sinc = SIN(x)/x
      ELSE
         y = x*x
         qs_ot_sinc = q1+y*(q2+y*(q3+y*(q4+y*(q5+y*(q6+y*(q7+y*(q8+y*(q9+y*(q10)))))))))
      ENDIF
   END FUNCTION qs_ot_sinc
   ! computes (1/(x^2-y^2))*(sinc(x)-sinc(y)) for all positive values of the arguments
! **************************************************************************************************
!> \brief ...
!> \param xa ...
!> \param ya ...
!> \retval qs_ot_sincf ...
! **************************************************************************************************
   FUNCTION qs_ot_sincf(xa, ya)

      REAL(KIND=dp), INTENT(IN)                          :: xa, ya
      REAL(KIND=dp)                                      :: qs_ot_sincf

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_sincf', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: a, b, rs, sf, x, xs, y, ybx, ybxs

! this is currently a limit of the routine, could be removed rather easily

      IF (xa .LT. 0) CALL cp__b("qs_ot.F",1618,"x is negative")
      IF (ya .LT. 0) CALL cp__b("qs_ot.F",1619,"y is negative")

      IF (xa .LT. ya) THEN
         x = ya
         y = xa
      ELSE
         x = xa
         y = ya
      ENDIF

      IF (x .LT. 0.5_dp) THEN ! use series, keeping in mind that x,y,x+y,x-y can all be zero

         qs_ot_sincf = 0.0_dp
         IF (x .GT. 0.0_dp) THEN
            ybx = y/x
         ELSE ! should be irrelevant  !?
            ybx = 0.0_dp
         ENDIF

         sf = -1.0_dp/((1.0_dp+ybx)*6.0_dp)
         rs = 1.0_dp
         ybxs = ybx
         xs = 1.0_dp

         DO i = 1, 10
            qs_ot_sincf = qs_ot_sincf+sf*rs*xs*(1.0_dp+ybxs)
            sf = -sf/(REAL((2*i+2), dp)*REAL((2*i+3), dp))
            rs = rs+ybxs
            ybxs = ybxs*ybx
            xs = xs*x*x
         ENDDO

      ELSE ! no series expansion
         IF (x-y .GT. 0.1_dp) THEN ! safe to use the normal form
            qs_ot_sincf = (qs_ot_sinc(x)-qs_ot_sinc(y))/((x+y)*(x-y))
         ELSE
            a = (x+y)/2.0_dp
            b = (x-y)/2.0_dp ! might be close to zero
            ! y (=(a-b)) can not be close to zero since it is close to x>0.5
            qs_ot_sincf = (qs_ot_sinc(b)*COS(a)-qs_ot_sinc(a)*COS(b))/(2*x*y)
         ENDIF
      ENDIF

   END FUNCTION qs_ot_sincf

END MODULE qs_ot
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief orbital transformations
!> \par History
!>      None
!> \author Joost VandeVondele (09.2002)
! **************************************************************************************************
MODULE qs_ot_minimizer

   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_add,&
                                              cp_dbcsr_copy,&
                                              cp_dbcsr_get_info,&
                                              cp_dbcsr_init_random,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_scale,&
                                              cp_dbcsr_set,&
                                              cp_dbcsr_trace
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE mathlib,                         ONLY: diamat_all
   USE preconditioner,                  ONLY: apply_preconditioner
   USE qs_ot,                           ONLY: qs_ot_get_derivative,&
                                              qs_ot_get_derivative_ref
   USE qs_ot_types,                     ONLY: qs_ot_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC  :: ot_mini

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_ot_minimizer'

CONTAINS
!
! the minimizer interface
! should present all possible modes of minimization
! these include CG SD DIIS
!
!
! IN the case of nspin != 1 we have a gradient that is distributed over different qs_ot_env.
! still things remain basically the same, since there are no constraints between the different qs_ot_env
! we only should take care that the various scalar products are taken over the full vectors.
! all the information needed and collected can be stored in the fist qs_ot_env only
! (indicating that the data type for the gradient/position and minization should be separated)
!
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
!> \param matrix_hc ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE ot_mini(qs_ot_env, matrix_hc, output_unit)
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_hc
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'ot_mini', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspin
      LOGICAL                                            :: do_ener, do_ks
      REAL(KIND=dp)                                      :: tmp

      CALL timeset(routineN, handle)

      nspin = SIZE(qs_ot_env)

      do_ks = qs_ot_env(1)%settings%ks
      do_ener = qs_ot_env(1)%settings%do_ener

      qs_ot_env(1)%OT_METHOD_FULL = ""

      ! compute the gradient for the variables x
      IF (.NOT. qs_ot_env(1)%energy_only) THEN
         qs_ot_env(1)%gradient = 0.0_dp
         DO ispin = 1, nspin
            IF (do_ks) THEN
               SELECT CASE (qs_ot_env (1)%settings%ot_algorithm)
               CASE ("TOD")
                  CALL qs_ot_get_derivative(matrix_hc(ispin)%matrix, qs_ot_env(ispin)%matrix_x, &
                                            qs_ot_env(ispin)%matrix_sx, &
                                            qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin))
               CASE ("REF")
                  CALL qs_ot_get_derivative_ref(matrix_hc(ispin)%matrix, &
                                                qs_ot_env(ispin)%matrix_x, qs_ot_env(ispin)%matrix_sx, &
                                                qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin), output_unit)
               CASE DEFAULT
                  CALL cp__b("qs_ot_minimizer.F",96,"ALGORITHM NYI")
               END SELECT
            END IF
            ! and also the gradient along the direction
            IF (qs_ot_env(1)%use_dx) THEN
               IF (do_ks) THEN
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_dx, tmp)
                  qs_ot_env(1)%gradient = qs_ot_env(1)%gradient+tmp
                  IF (qs_ot_env(1)%settings%do_rotation) THEN
                     CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_gx, qs_ot_env(ispin)%rot_mat_dx, tmp)
                     qs_ot_env(1)%gradient = qs_ot_env(1)%gradient+0.5_dp*tmp
                  ENDIF
               END IF
               IF (do_ener) THEN
                  tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_gx, qs_ot_env(ispin)%ener_dx)
                  qs_ot_env(1)%gradient = qs_ot_env(1)%gradient+tmp
               ENDIF
            ELSE
               IF (do_ks) THEN
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_gx, tmp)
                  qs_ot_env(1)%gradient = qs_ot_env(1)%gradient-tmp
                  IF (qs_ot_env(1)%settings%do_rotation) THEN
                     CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_gx, qs_ot_env(ispin)%rot_mat_gx, tmp)
                     qs_ot_env(1)%gradient = qs_ot_env(1)%gradient-0.5_dp*tmp
                  ENDIF
               ENDIF
               IF (do_ener) THEN
                  tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_gx, qs_ot_env(ispin)%ener_gx)
                  qs_ot_env(1)%gradient = qs_ot_env(1)%gradient-tmp
               ENDIF
            ENDIF
         ENDDO
      ENDIF

      SELECT CASE (qs_ot_env (1)%settings%OT_METHOD)
      CASE ("CG")
         IF (current_point_is_fine(qs_ot_env)) THEN
            qs_ot_env(1)%OT_METHOD_FULL = "OT CG"
            CALL ot_new_cg_direction(qs_ot_env)
            qs_ot_env(1)%line_search_count = 0
         ELSE
            qs_ot_env(1)%OT_METHOD_FULL = "OT LS"
         ENDIF
         CALL do_line_search(qs_ot_env)
      CASE ("SD")
         IF (current_point_is_fine(qs_ot_env)) THEN
            qs_ot_env(1)%OT_METHOD_FULL = "OT SD"
            CALL ot_new_sd_direction(qs_ot_env)
            qs_ot_env(1)%line_search_count = 0
         ELSE
            qs_ot_env(1)%OT_METHOD_FULL = "OT LS"
         ENDIF
         CALL do_line_search(qs_ot_env)
      CASE ("DIIS")
         qs_ot_env(1)%OT_METHOD_FULL = "OT DIIS"
         CALL ot_diis_step(qs_ot_env)
      CASE ("BROY")
         qs_ot_env(1)%OT_METHOD_FULL = "OT BROY"
         CALL ot_broyden_step(qs_ot_env)
      CASE DEFAULT
         CALL cp__b("qs_ot_minimizer.F",156,"OT_METHOD NYI")
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE ot_mini

!
! checks if the current point is a good point for finding a new direction
! or if we should improve the line_search, if it is used
!
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION current_point_is_fine(qs_ot_env) RESULT(res)
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env
      LOGICAL                                            :: res

      res = .FALSE.

      ! only if we have a gradient it can be fine
      IF (.NOT. qs_ot_env(1)%energy_only) THEN

         ! we have not yet started with the line search
         IF (qs_ot_env(1)%line_search_count .EQ. 0) THEN
            res = .TRUE.
            RETURN
         ENDIF

         IF (qs_ot_env(1)%line_search_might_be_done) THEN
            ! here we put the more complicated logic later
            res = .TRUE.
            RETURN
         ENDIF

      ENDIF

   END FUNCTION current_point_is_fine

!
! performs various kinds of line searches
!
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE do_line_search(qs_ot_env)
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'do_line_search', routineP = moduleN//':'//routineN

      SELECT CASE (qs_ot_env (1)%settings%line_search_method)
      CASE ("GOLD")
         CALL do_line_search_gold(qs_ot_env)
      CASE ("3PNT")
         CALL do_line_search_3pnt(qs_ot_env)
      CASE ("2PNT")
         CALL do_line_search_2pnt(qs_ot_env)
      CASE ("NONE")
         CALL do_line_search_none(qs_ot_env)
      CASE DEFAULT
         CALL cp__b("qs_ot_minimizer.F",219,"NYI")
      END SELECT
   END SUBROUTINE do_line_search

! **************************************************************************************************
!> \brief moves x adding the right amount (ds) of the gradient or search direction
!> \param ds ...
!> \param qs_ot_env ...
!> \par History
!>      08.2004 created [ Joost VandeVondele ] copied here from a larger number of subroutines
! **************************************************************************************************
   SUBROUTINE take_step(ds, qs_ot_env)
      REAL(KIND=dp)                                      :: ds
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'take_step', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspin
      LOGICAL                                            :: do_ener, do_ks

      CALL timeset(routineN, handle)

      nspin = SIZE(qs_ot_env)

      do_ks = qs_ot_env(1)%settings%ks
      do_ener = qs_ot_env(1)%settings%do_ener

      ! now update x to take into account this new step
      ! either dx or -gx is the direction to use
      IF (qs_ot_env(1)%use_dx) THEN
         IF (do_ks) THEN
            DO ispin = 1, nspin
               CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, qs_ot_env(ispin)%matrix_dx, &
                                 alpha_scalar=1.0_dp, beta_scalar=ds)
               IF (qs_ot_env(ispin)%settings%do_rotation) THEN
                  CALL cp_dbcsr_add(qs_ot_env(ispin)%rot_mat_x, qs_ot_env(ispin)%rot_mat_dx, &
                                    alpha_scalar=1.0_dp, beta_scalar=ds)
               ENDIF
            ENDDO
         END IF
         IF (do_ener) THEN
            DO ispin = 1, nspin
               qs_ot_env(ispin)%ener_x = qs_ot_env(ispin)%ener_x+ds*qs_ot_env(ispin)%ener_dx
            ENDDO
         ENDIF
      ELSE
         IF (do_ks) THEN
            DO ispin = 1, nspin
               CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, qs_ot_env(ispin)%matrix_gx, &
                                 alpha_scalar=1.0_dp, beta_scalar=-ds)
               IF (qs_ot_env(ispin)%settings%do_rotation) THEN
                  CALL cp_dbcsr_add(qs_ot_env(ispin)%rot_mat_x, qs_ot_env(ispin)%rot_mat_gx, &
                                    alpha_scalar=1.0_dp, beta_scalar=-ds)
               ENDIF
            ENDDO
         ENDIF
         IF (do_ener) THEN
            DO ispin = 1, nspin
               qs_ot_env(ispin)%ener_x = qs_ot_env(ispin)%ener_x-ds*qs_ot_env(ispin)%ener_gx
            ENDDO
         ENDIF
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE take_step

! implements a golden ratio search as a robust way of minimizing
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE do_line_search_gold(qs_ot_env)

      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'do_line_search_gold', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: gold_sec = 0.3819_dp

      INTEGER                                            :: count, handle
      REAL(KIND=dp)                                      :: ds

      CALL timeset(routineN, handle)

      qs_ot_env(1)%line_search_count = qs_ot_env(1)%line_search_count+1
      count = qs_ot_env(1)%line_search_count
      qs_ot_env(1)%line_search_might_be_done = .FALSE.
      qs_ot_env(1)%energy_only = .TRUE.

      IF (count+1 .GT. SIZE(qs_ot_env(1)%OT_pos)) THEN
         ! should not happen, we pass with a warning first
         ! you can increase the size of OT_pos and the like in qs_ot_env
         CALL cp__b("qs_ot_minimizer.F",310,"MAX ITER EXCEEDED : FATAL")
      ENDIF

      IF (qs_ot_env(1)%line_search_count .EQ. 1) THEN
         qs_ot_env(1)%line_search_left = 1
         qs_ot_env(1)%line_search_right = 0
         qs_ot_env(1)%line_search_mid = 1
         qs_ot_env(1)%ot_pos(1) = 0.0_dp
         qs_ot_env(1)%ot_energy(1) = qs_ot_env(1)%etotal
         qs_ot_env(1)%ot_pos(2) = qs_ot_env(1)%ds_min/gold_sec
      ELSE
         qs_ot_env(1)%ot_energy(count) = qs_ot_env(1)%etotal
         ! it's essentially a book keeping game.
         ! keep left on the left, keep (bring) right on the right
         ! and mid in between these two
         IF (qs_ot_env(1)%line_search_right .EQ. 0) THEN ! we do not yet have the right bracket
            IF (qs_ot_env(1)%ot_energy(count-1) .LT. qs_ot_env(1)%ot_energy(count)) THEN
               qs_ot_env(1)%line_search_right = count
               qs_ot_env(1)%ot_pos(count+1) = qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_mid)+ &
                                              (qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_right)- &
                                               qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_mid))*gold_sec
            ELSE
               qs_ot_env(1)%line_search_left = qs_ot_env(1)%line_search_mid
               qs_ot_env(1)%line_search_mid = count
               qs_ot_env(1)%ot_pos(count+1) = qs_ot_env(1)%ot_pos(count)/gold_sec ! expand
            ENDIF
         ELSE
            ! first determine where we are and construct the new triplet
            IF (qs_ot_env(1)%ot_pos(count) .LT. qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_mid)) THEN
               IF (qs_ot_env(1)%ot_energy(count) .LT. qs_ot_env(1)%ot_energy(qs_ot_env(1)%line_search_mid)) THEN
                  qs_ot_env(1)%line_search_right = qs_ot_env(1)%line_search_mid
                  qs_ot_env(1)%line_search_mid = count
               ELSE
                  qs_ot_env(1)%line_search_left = count
               ENDIF
            ELSE
               IF (qs_ot_env(1)%ot_energy(count) .LT. qs_ot_env(1)%ot_energy(qs_ot_env(1)%line_search_mid)) THEN
                  qs_ot_env(1)%line_search_left = qs_ot_env(1)%line_search_mid
                  qs_ot_env(1)%line_search_mid = count
               ELSE
                  qs_ot_env(1)%line_search_right = count
               ENDIF
            ENDIF
            ! now find the new point in the largest section
            IF ((qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_right) &
                 -qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_mid)) .GT. &
                (qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_mid) &
                 -qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_left))) THEN
               qs_ot_env(1)%ot_pos(count+1) = &
                  qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_mid)+ &
                  gold_sec*(qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_right) &
                            -qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_mid))
            ELSE
               qs_ot_env(1)%ot_pos(count+1) = &
                  qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_left)+ &
                  gold_sec*(qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_mid) &
                            -qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_left))
            ENDIF
            ! check for termination
            IF (((qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_right) &
                  -qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_mid)) .LT. &
                 qs_ot_env(1)%ds_min*qs_ot_env(1)%settings%gold_target) .AND. &
                ((qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_mid) &
                  -qs_ot_env(1)%ot_pos(qs_ot_env(1)%line_search_left)) .LT. &
                 qs_ot_env(1)%ds_min*qs_ot_env(1)%settings%gold_target)) THEN
               qs_ot_env(1)%energy_only = .FALSE.
               qs_ot_env(1)%line_search_might_be_done = .TRUE.
            ENDIF
         ENDIF
      ENDIF
      ds = qs_ot_env(1)%OT_pos(count+1)-qs_ot_env(1)%OT_pos(count)
      qs_ot_env(1)%ds_min = qs_ot_env(1)%OT_pos(count+1)

      CALL take_step(ds, qs_ot_env)

      CALL timestop(handle)

   END SUBROUTINE do_line_search_gold

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE do_line_search_3pnt(qs_ot_env)

      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'do_line_search_3pnt', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: count, handle
      REAL(KIND=dp)                                      :: denom, ds, fa, fb, fc, nom, pos, val, &
                                                            xa, xb, xc

      CALL timeset(routineN, handle)

      qs_ot_env(1)%line_search_might_be_done = .FALSE.
      qs_ot_env(1)%energy_only = .TRUE.

      ! a three point interpolation based on the energy
      qs_ot_env(1)%line_search_count = qs_ot_env(1)%line_search_count+1
      count = qs_ot_env(1)%line_search_count
      qs_ot_env(1)%ot_energy(count) = qs_ot_env(1)%etotal
      SELECT CASE (count)
      CASE (1)
         qs_ot_env(1)%ot_pos(count) = 0.0_dp
         qs_ot_env(1)%ot_pos(count+1) = qs_ot_env(1)%ds_min*0.8_dp
      CASE (2)
         IF (qs_ot_env(1)%OT_energy(count) .GT. qs_ot_env(1)%OT_energy(count-1)) THEN
            qs_ot_env(1)%OT_pos(count+1) = qs_ot_env(1)%ds_min*0.5_dp
         ELSE
            qs_ot_env(1)%OT_pos(count+1) = qs_ot_env(1)%ds_min*1.4_dp
         ENDIF
      CASE (3)
         xa = qs_ot_env(1)%OT_pos(1)
         xb = qs_ot_env(1)%OT_pos(2)
         xc = qs_ot_env(1)%OT_pos(3)
         fa = qs_ot_env(1)%OT_energy(1)
         fb = qs_ot_env(1)%OT_energy(2)
         fc = qs_ot_env(1)%OT_energy(3)
         nom = (xb-xa)**2*(fb-fc)-(xb-xc)**2*(fb-fa)
         denom = (xb-xa)*(fb-fc)-(xb-xc)*(fb-fa)
         IF (ABS(denom) .LE. 1.0E-18_dp*MAX(ABS(fb-fc), ABS(fb-fa))) THEN
            pos = xb
         ELSE
            pos = xb-0.5_dp*nom/denom ! position of the stationary point
         ENDIF
         val = (pos-xa)*(pos-xb)*fc/((xc-xa)*(xc-xb))+ &
               (pos-xb)*(pos-xc)*fa/((xa-xb)*(xa-xc))+ &
               (pos-xc)*(pos-xa)*fb/((xb-xc)*(xb-xa))
         IF (val .LT. fa .AND. val .LE. fb .AND. val .LE. fc) THEN ! OK, we go to a minimum
            ! we take a guard against too large steps
            qs_ot_env(1)%OT_pos(count+1) = MAX(MAXVAL(qs_ot_env(1)%OT_pos(1:3))*0.01_dp, &
                                               MIN(pos, MAXVAL(qs_ot_env(1)%OT_pos(1:3))*4.0_dp))
         ELSE ! just take an extended step
            qs_ot_env(1)%OT_pos(count+1) = MAXVAL(qs_ot_env(1)%OT_pos(1:3))*2.0_dp
         ENDIF
         qs_ot_env(1)%energy_only = .FALSE.
         qs_ot_env(1)%line_search_might_be_done = .TRUE.
      CASE DEFAULT
         CALL cp__b("qs_ot_minimizer.F",450,"NYI")
      END SELECT
      ds = qs_ot_env(1)%OT_pos(count+1)-qs_ot_env(1)%OT_pos(count)
      qs_ot_env(1)%ds_min = qs_ot_env(1)%OT_pos(count+1)

      CALL take_step(ds, qs_ot_env)

      CALL timestop(handle)

   END SUBROUTINE do_line_search_3pnt

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE do_line_search_2pnt(qs_ot_env)

      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'do_line_search_2pnt', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: count, handle
      REAL(KIND=dp)                                      :: a, b, c, ds, pos, val, x0, x1

      CALL timeset(routineN, handle)

      qs_ot_env(1)%line_search_might_be_done = .FALSE.
      qs_ot_env(1)%energy_only = .TRUE.

      ! a three point interpolation based on the energy
      qs_ot_env(1)%line_search_count = qs_ot_env(1)%line_search_count+1
      count = qs_ot_env(1)%line_search_count
      qs_ot_env(1)%ot_energy(count) = qs_ot_env(1)%etotal
      SELECT CASE (count)
      CASE (1)
         qs_ot_env(1)%ot_pos(count) = 0.0_dp
         qs_ot_env(1)%ot_grad(count) = qs_ot_env(1)%gradient
         qs_ot_env(1)%ot_pos(count+1) = qs_ot_env(1)%ds_min*1.0_dp
      CASE (2)
         x0 = 0.0_dp
         c = qs_ot_env(1)%ot_energy(1)
         b = qs_ot_env(1)%ot_grad(1)
         x1 = qs_ot_env(1)%ot_pos(2)
         a = (qs_ot_env(1)%ot_energy(2)-b*x1-c)/(x1**2)
         IF (a .LE. 0.0_dp) a = 1.0E-15_dp
         pos = -b/(2.0_dp*a)
         val = a*pos**2+b*pos+c
         qs_ot_env(1)%energy_only = .FALSE.
         qs_ot_env(1)%line_search_might_be_done = .TRUE.
         IF (val .LT. qs_ot_env(1)%ot_energy(1) .AND. val .LE. qs_ot_env(1)%ot_energy(2)) THEN
            ! we go to a minimum, but ...
            ! we take a guard against too large steps
            qs_ot_env(1)%OT_pos(count+1) = MAX(MAXVAL(qs_ot_env(1)%OT_pos(1:2))*0.01_dp, &
                                               MIN(pos, MAXVAL(qs_ot_env(1)%OT_pos(1:2))*4.0_dp))
         ELSE ! just take an extended step
            qs_ot_env(1)%OT_pos(count+1) = MAXVAL(qs_ot_env(1)%OT_pos(1:2))*2.0_dp
         ENDIF
      CASE DEFAULT
         CALL cp__b("qs_ot_minimizer.F",509,"NYI")
      END SELECT
      ds = qs_ot_env(1)%OT_pos(count+1)-qs_ot_env(1)%OT_pos(count)
      qs_ot_env(1)%ds_min = qs_ot_env(1)%OT_pos(count+1)

      CALL take_step(ds, qs_ot_env)

      CALL timestop(handle)

   END SUBROUTINE do_line_search_2pnt

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE do_line_search_none(qs_ot_env)
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env

      CALL take_step(qs_ot_env(1)%ds_min, qs_ot_env)

   END SUBROUTINE do_line_search_none

!
! creates a new SD direction, using the preconditioner if associated
! also updates the gradient for line search
!

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE ot_new_sd_direction(qs_ot_env)
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ot_new_sd_direction', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, itmp, k, n, nener, nspin
      LOGICAL                                            :: do_ener, do_ks
      REAL(KIND=dp)                                      :: tmp
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

!***SCP

      nspin = SIZE(qs_ot_env)
      logger => cp_get_default_logger()
      do_ks = qs_ot_env(1)%settings%ks
      do_ener = qs_ot_env(1)%settings%do_ener

      IF (ASSOCIATED(qs_ot_env(1)%preconditioner)) THEN
         IF (.NOT. qs_ot_env(1)%use_dx) CALL cp__b("qs_ot_minimizer.F",561,"use dx")
         qs_ot_env(1)%gnorm = 0.0_dp
         IF (do_ks) THEN
            DO ispin = 1, nspin
               CALL apply_preconditioner(qs_ot_env(ispin)%preconditioner, &
                                         qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_dx)
               CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_dx, tmp)
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
            ENDDO
            IF (qs_ot_env(1)%gnorm .LT. 0.0_dp) THEN
               logger => cp_get_default_logger()
               WRITE (cp_logger_get_default_unit_nr(logger), *) "WARNING Preconditioner not positive definite !"
            ENDIF
            DO ispin = 1, nspin
               CALL cp_dbcsr_scale(qs_ot_env(ispin)%matrix_dx, -1.0_dp)
            ENDDO
            IF (qs_ot_env(1)%settings%do_rotation) THEN
               DO ispin = 1, nspin
                  ! right now no preconditioner yet
                  CALL cp_dbcsr_copy(qs_ot_env(ispin)%rot_mat_dx, qs_ot_env(ispin)%rot_mat_gx)
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_gx, qs_ot_env(ispin)%rot_mat_dx, tmp)
                  ! added 0.5, because we have (antisymmetry) only half the number of variables
                  qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+0.5_dp*tmp
               ENDDO
               DO ispin = 1, nspin
                  CALL cp_dbcsr_scale(qs_ot_env(ispin)%rot_mat_dx, -1.0_dp)
               ENDDO
            ENDIF
         ENDIF
         IF (do_ener) THEN
            DO ispin = 1, nspin
               qs_ot_env(ispin)%ener_dx = qs_ot_env(ispin)%ener_gx
               tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_dx, qs_ot_env(ispin)%ener_gx)
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
               qs_ot_env(ispin)%ener_dx = -qs_ot_env(ispin)%ener_dx
            ENDDO
         ENDIF
      ELSE
         qs_ot_env(1)%gnorm = 0.0_dp
         IF (do_ks) THEN
            DO ispin = 1, nspin
               CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_gx, tmp)
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
            ENDDO
            IF (qs_ot_env(1)%settings%do_rotation) THEN
               DO ispin = 1, nspin
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_gx, qs_ot_env(ispin)%rot_mat_gx, tmp)
                  ! added 0.5, because we have (antisymmetry) only half the number of variables
                  qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+0.5_dp*tmp
               ENDDO
            ENDIF
         ENDIF
         IF (do_ener) THEN
            DO ispin = 1, nspin
               tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_gx, qs_ot_env(ispin)%ener_gx)
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
            ENDDO
         ENDIF
      ENDIF

      k = 0
      n = 0
      nener = 0
      IF (do_ks) THEN
         CALL cp_dbcsr_get_info(qs_ot_env(1)%matrix_x, nfullrows_total=n)
         DO ispin = 1, nspin
            CALL cp_dbcsr_get_info(qs_ot_env(ispin)%matrix_x, nfullcols_total=itmp)
            k = k+itmp
         ENDDO
      ENDIF
      IF (do_ener) THEN
         DO ispin = 1, nspin
            nener = nener+SIZE(qs_ot_env(ispin)%ener_x)
         ENDDO
      ENDIF
      ! Handling the case of no free variables to optimize
      IF (INT(n, KIND=int_8)*INT(k, KIND=int_8)+nener /= 0) THEN
         qs_ot_env(1)%delta = SQRT(ABS(qs_ot_env(1)%gnorm)/(INT(n, KIND=int_8)*INT(k, KIND=int_8)+nener))
         qs_ot_env(1)%gradient = -qs_ot_env(1)%gnorm
      ELSE
         qs_ot_env(1)%delta = 0.0_dp
         qs_ot_env(1)%gradient = 0.0_dp
      END IF

      CALL timestop(handle)

   END SUBROUTINE ot_new_sd_direction

!
! creates a new CG direction. Implements Polak-Ribierre variant
! using the preconditioner if associated
! also updates the gradient for line search
!
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE ot_new_cg_direction(qs_ot_env)
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ot_new_cg_direction', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, itmp, k, n, nener, nspin
      LOGICAL                                            :: do_ener, do_ks
      REAL(KIND=dp)                                      :: beta_pr, gnorm_cross, test_down, tmp
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      nspin = SIZE(qs_ot_env)
      logger => cp_get_default_logger()

      do_ks = qs_ot_env(1)%settings%ks
      do_ener = qs_ot_env(1)%settings%do_ener

      gnorm_cross = 0.0_dp
      IF (do_ks) THEN
         DO ispin = 1, nspin
            CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_gx_old, tmp)
            gnorm_cross = gnorm_cross+tmp
         ENDDO
         IF (qs_ot_env(1)%settings%do_rotation) THEN
            DO ispin = 1, nspin
               CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_gx, qs_ot_env(ispin)%rot_mat_gx_old, tmp)
               ! added 0.5, because we have (antisymmetry) only half the number of variables
               gnorm_cross = gnorm_cross+0.5_dp*tmp
            ENDDO
         ENDIF
      END IF
      IF (do_ener) THEN
         DO ispin = 1, nspin
            tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_gx, qs_ot_env(ispin)%ener_gx_old)
            gnorm_cross = gnorm_cross+tmp
         ENDDO
      END IF

      IF (ASSOCIATED(qs_ot_env(1)%preconditioner)) THEN

         DO ispin = 1, nspin
            CALL apply_preconditioner(qs_ot_env(ispin)%preconditioner, &
                                      qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_gx_old)
         ENDDO
         qs_ot_env(1)%gnorm = 0.0_dp
         IF (do_ks) THEN
            DO ispin = 1, nspin
               CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_gx_old, tmp)
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
            ENDDO
            IF (qs_ot_env(1)%gnorm .LT. 0.0_dp) THEN
               WRITE (cp_logger_get_default_unit_nr(logger), *) "WARNING Preconditioner not positive definite !"
            ENDIF
            DO ispin = 1, nspin
               CALL cp_dbcsr_copy(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_gx_old)
            ENDDO
            IF (qs_ot_env(1)%settings%do_rotation) THEN
               DO ispin = 1, nspin
                  ! right now no preconditioner yet
                  CALL cp_dbcsr_copy(qs_ot_env(ispin)%rot_mat_gx_old, qs_ot_env(ispin)%rot_mat_gx)
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_gx, qs_ot_env(ispin)%rot_mat_gx_old, tmp)
                  ! added 0.5, because we have (antisymmetry) only half the number of variables
                  qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+0.5_dp*tmp
               ENDDO
               DO ispin = 1, nspin
                  CALL cp_dbcsr_copy(qs_ot_env(ispin)%rot_mat_gx, qs_ot_env(ispin)%rot_mat_gx_old)
               ENDDO
            ENDIF
         END IF
         IF (do_ener) THEN
            DO ispin = 1, nspin
               qs_ot_env(ispin)%ener_gx_old = qs_ot_env(ispin)%ener_gx
               tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_gx, qs_ot_env(ispin)%ener_gx_old)
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
               qs_ot_env(ispin)%ener_gx = qs_ot_env(ispin)%ener_gx_old
            ENDDO
         END IF
      ELSE
         IF (do_ks) THEN
            qs_ot_env(1)%gnorm = 0.0_dp
            DO ispin = 1, nspin
               CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_gx, tmp)
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
               CALL cp_dbcsr_copy(qs_ot_env(ispin)%matrix_gx_old, qs_ot_env(ispin)%matrix_gx)
            ENDDO
            IF (qs_ot_env(1)%settings%do_rotation) THEN
               DO ispin = 1, nspin
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_gx, qs_ot_env(ispin)%rot_mat_gx, tmp)
                  ! added 0.5, because we have (antisymmetry) only half the number of variables
                  qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+0.5_dp*tmp
                  CALL cp_dbcsr_copy(qs_ot_env(ispin)%rot_mat_gx_old, qs_ot_env(ispin)%rot_mat_gx)
               ENDDO
            ENDIF
         ENDIF
         IF (do_ener) THEN
            DO ispin = 1, nspin
               tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_gx, qs_ot_env(ispin)%ener_gx)
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
               qs_ot_env(ispin)%ener_gx_old = qs_ot_env(ispin)%ener_gx
            ENDDO
         ENDIF
      ENDIF

      k = 0
      n = 0
      nener = 0
      IF (do_ks) THEN
         CALL cp_dbcsr_get_info(qs_ot_env(1)%matrix_x, nfullrows_total=n)
         DO ispin = 1, nspin
            CALL cp_dbcsr_get_info(qs_ot_env(ispin)%matrix_x, nfullcols_total=itmp)
            k = k+itmp
         ENDDO
      END IF
      IF (do_ener) THEN
         DO ispin = 1, nspin
            nener = nener+SIZE(qs_ot_env(ispin)%ener_x)
         ENDDO
      ENDIF
      ! Handling the case of no free variables to optimize
      IF (INT(n, KIND=int_8)*INT(k, KIND=int_8)+nener /= 0) THEN
         qs_ot_env(1)%delta = SQRT(ABS(qs_ot_env(1)%gnorm)/(INT(n, KIND=int_8)*INT(k, KIND=int_8)+nener))
         beta_pr = (qs_ot_env(1)%gnorm-gnorm_cross)/qs_ot_env(1)%gnorm_old
      ELSE
         qs_ot_env(1)%delta = 0.0_dp
         beta_pr = 0.0_dp
      END IF
      beta_pr = MAX(beta_pr, 0.0_dp) ! reset to SD

      test_down = 0.0_dp
      IF (do_ks) THEN
         DO ispin = 1, nspin
            CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_dx, qs_ot_env(ispin)%matrix_gx, &
                              alpha_scalar=beta_pr, beta_scalar=-1.0_dp)
            CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_dx, tmp)
            test_down = test_down+tmp
            IF (qs_ot_env(1)%settings%do_rotation) THEN
               CALL cp_dbcsr_add(qs_ot_env(ispin)%rot_mat_dx, qs_ot_env(ispin)%rot_mat_gx, &
                                 alpha_scalar=beta_pr, beta_scalar=-1.0_dp)
               CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_gx, qs_ot_env(ispin)%rot_mat_dx, tmp)
               test_down = test_down+0.5_dp*tmp
            ENDIF
         ENDDO
      END IF
      IF (do_ener) THEN
         DO ispin = 1, nspin
            qs_ot_env(ispin)%ener_dx = beta_pr*qs_ot_env(ispin)%ener_dx-qs_ot_env(ispin)%ener_gx
            tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_gx, qs_ot_env(ispin)%ener_dx)
            test_down = test_down+tmp
         ENDDO
      ENDIF

      IF (test_down .GE. 0.0_dp) THEN ! reset to SD
         beta_pr = 0.0_dp
         IF (do_ks) THEN
            DO ispin = 1, nspin
               CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_dx, qs_ot_env(ispin)%matrix_gx, &
                                 alpha_scalar=beta_pr, beta_scalar=-1.0_dp)
               IF (qs_ot_env(1)%settings%do_rotation) THEN
                  CALL cp_dbcsr_add(qs_ot_env(ispin)%rot_mat_dx, &
                                    qs_ot_env(ispin)%rot_mat_gx, alpha_scalar=beta_pr, beta_scalar=-1.0_dp)
               ENDIF
            ENDDO
         END IF
         IF (do_ener) THEN
            DO ispin = 1, nspin
               qs_ot_env(ispin)%ener_dx = beta_pr*qs_ot_env(ispin)%ener_dx-qs_ot_env(ispin)%ener_gx
            ENDDO
         ENDIF
      ENDIF
      ! since we change the direction we have to adjust the gradient
      qs_ot_env(1)%gradient = beta_pr*qs_ot_env(1)%gradient-qs_ot_env(1)%gnorm
      qs_ot_env(1)%gnorm_old = qs_ot_env(1)%gnorm

      CALL timestop(handle)

   END SUBROUTINE ot_new_cg_direction

! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE ot_diis_step(qs_ot_env)
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ot_diis_step', routineP = moduleN//':'//routineN

      INTEGER                                            :: diis_bound, diis_m, handle, i, info, &
                                                            ispin, itmp, j, k, n, nener, nspin
      LOGICAL                                            :: do_ener, do_ks, do_ot_sd
      REAL(KIND=dp)                                      :: overlap, tmp, tr_xnew_gx, tr_xold_gx
      TYPE(cp_logger_type), POINTER                      :: logger

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()

      do_ks = qs_ot_env(1)%settings%ks
      do_ener = qs_ot_env(1)%settings%do_ener
      nspin = SIZE(qs_ot_env)

      diis_m = qs_ot_env(1)%settings%diis_m

      IF (qs_ot_env(1)%diis_iter .LT. diis_m) THEN
         diis_bound = qs_ot_env(1)%diis_iter+1
      ELSE
         diis_bound = diis_m
      ENDIF

      j = MOD(qs_ot_env(1)%diis_iter, diis_m)+1 ! index in the circular array

      ! copy the position and the error vector in the diis buffers

      IF (do_ks) THEN
         DO ispin = 1, nspin
            CALL cp_dbcsr_copy(qs_ot_env(ispin)%matrix_h_x(j)%matrix, qs_ot_env(ispin)%matrix_x)
            IF (qs_ot_env(ispin)%settings%do_rotation) THEN
               CALL cp_dbcsr_copy(qs_ot_env(ispin)%rot_mat_h_x(j)%matrix, qs_ot_env(ispin)%rot_mat_x)
            ENDIF
         ENDDO
      END IF
      IF (do_ener) THEN
         DO ispin = 1, nspin
            qs_ot_env(ispin)%ener_h_x(j, :) = qs_ot_env(ispin)%ener_x(:)
         ENDDO
      END IF
      IF (ASSOCIATED(qs_ot_env(1)%preconditioner)) THEN
         qs_ot_env(1)%gnorm = 0.0_dp
         IF (do_ks) THEN
            DO ispin = 1, nspin
               CALL apply_preconditioner(qs_ot_env(ispin)%preconditioner, &
                                         qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_h_e(j)%matrix)
               CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_h_e(j)%matrix, &
                                   tmp)
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
            ENDDO
            IF (qs_ot_env(1)%gnorm .LT. 0.0_dp) THEN
               WRITE (cp_logger_get_default_unit_nr(logger), *) "WARNING Preconditioner not positive definite !"
            ENDIF
            DO ispin = 1, nspin
               CALL cp_dbcsr_scale(qs_ot_env(ispin)%matrix_h_e(j)%matrix, -qs_ot_env(1)%ds_min)
            ENDDO
            IF (qs_ot_env(1)%settings%do_rotation) THEN
               DO ispin = 1, nspin
                  CALL cp_dbcsr_copy(qs_ot_env(ispin)%rot_mat_h_e(j)%matrix, qs_ot_env(ispin)%rot_mat_gx)
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_gx, qs_ot_env(ispin)%rot_mat_h_e(j)%matrix, &
                                      tmp)
                  qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+0.5_dp*tmp
               ENDDO
               DO ispin = 1, nspin
                  CALL cp_dbcsr_scale(qs_ot_env(ispin)%rot_mat_h_e(j)%matrix, -qs_ot_env(1)%ds_min)
               ENDDO
            ENDIF
         END IF
         IF (do_ener) THEN
            DO ispin = 1, nspin
               qs_ot_env(ispin)%ener_h_e(j, :) = qs_ot_env(ispin)%ener_gx(:)
               tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_h_e(j, :), qs_ot_env(ispin)%ener_gx(:))
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
               qs_ot_env(ispin)%ener_h_e(j, :) = -qs_ot_env(1)%ds_min*qs_ot_env(ispin)%ener_h_e(j, :)
            ENDDO
         ENDIF
      ELSE
         qs_ot_env(1)%gnorm = 0.0_dp
         IF (do_ks) THEN
            DO ispin = 1, nspin
               CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_gx, tmp)
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
               CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_h_e(j)%matrix, &
                                 qs_ot_env(ispin)%matrix_gx, alpha_scalar=0.0_dp, beta_scalar=-qs_ot_env(1)%ds_min)
            ENDDO
            IF (qs_ot_env(1)%settings%do_rotation) THEN
               DO ispin = 1, nspin
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_gx, qs_ot_env(ispin)%rot_mat_gx, tmp)
                  qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+0.5_dp*tmp
                  CALL cp_dbcsr_add(qs_ot_env(ispin)%rot_mat_h_e(j)%matrix, &
                                    qs_ot_env(ispin)%rot_mat_gx, alpha_scalar=0.0_dp, beta_scalar=-qs_ot_env(1)%ds_min)
               ENDDO
            ENDIF
         END IF
         IF (do_ener) THEN
            DO ispin = 1, nspin
               tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_gx(:), qs_ot_env(ispin)%ener_gx(:))
               qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
               qs_ot_env(ispin)%ener_h_e(j, :) = -qs_ot_env(1)%ds_min*qs_ot_env(ispin)%ener_gx(:)
            ENDDO
         END IF
      ENDIF
      k = 0
      n = 0
      nener = 0
      IF (do_ks) THEN
         CALL cp_dbcsr_get_info(qs_ot_env(1)%matrix_x, nfullrows_total=n)
         DO ispin = 1, nspin
            CALL cp_dbcsr_get_info(qs_ot_env(ispin)%matrix_x, nfullcols_total=itmp)
            k = k+itmp
         ENDDO
      END IF
      IF (do_ener) THEN
         DO ispin = 1, nspin
            nener = nener+SIZE(qs_ot_env(ispin)%ener_x)
         ENDDO
      ENDIF
      ! Handling the case of no free variables to optimize
      IF (INT(n, KIND=int_8)*INT(k, KIND=int_8)+nener /= 0) THEN
         qs_ot_env(1)%delta = SQRT(ABS(qs_ot_env(1)%gnorm)/(INT(n, KIND=int_8)*INT(k, KIND=int_8)+nener))
         qs_ot_env(1)%gradient = -qs_ot_env(1)%gnorm
      ELSE
         qs_ot_env(1)%delta = 0.0_dp
         qs_ot_env(1)%gradient = 0.0_dp
      END IF

      ! make the diis matrix and solve it
      DO i = 1, diis_bound
         ! I think there are two possible options, with and without preconditioner
         ! as a metric
         ! the second option seems most logical to me, and it seems marginally faster
         ! in some of the tests
         IF (.FALSE.) THEN
            qs_ot_env(1)%ls_diis(i, j) = 0.0_dp
            IF (do_ks) THEN
               DO ispin = 1, nspin
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_h_e(j)%matrix, &
                                      qs_ot_env(ispin)%matrix_h_e(i)%matrix, &
                                      tmp)
                  qs_ot_env(1)%ls_diis(i, j) = qs_ot_env(1)%ls_diis(i, j)+tmp
                  IF (qs_ot_env(ispin)%settings%do_rotation) THEN
                     CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_h_e(j)%matrix, &
                                         qs_ot_env(ispin)%rot_mat_h_e(i)%matrix, &
                                         tmp)
                     qs_ot_env(1)%ls_diis(i, j) = qs_ot_env(1)%ls_diis(i, j)+0.5_dp*tmp
                  ENDIF
               ENDDO
            END IF
            IF (do_ener) THEN
               DO ispin = 1, nspin
                  tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_h_e(j, :), qs_ot_env(ispin)%ener_h_e(i, :))
                  qs_ot_env(1)%ls_diis(i, j) = qs_ot_env(1)%ls_diis(i, j)+tmp
               ENDDO
            END IF
         ELSE
            qs_ot_env(1)%ls_diis(i, j) = 0.0_dp
            IF (do_ks) THEN
               DO ispin = 1, nspin
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, &
                                      qs_ot_env(ispin)%matrix_h_e(i)%matrix, &
                                      tmp)
                  qs_ot_env(1)%ls_diis(i, j) = qs_ot_env(1)%ls_diis(i, j)-qs_ot_env(1)%ds_min*tmp
                  IF (qs_ot_env(ispin)%settings%do_rotation) THEN
                     CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_gx, &
                                         qs_ot_env(ispin)%rot_mat_h_e(i)%matrix, &
                                         tmp)
                     qs_ot_env(1)%ls_diis(i, j) = qs_ot_env(1)%ls_diis(i, j)-qs_ot_env(1)%ds_min*0.5_dp*tmp
                  ENDIF
               ENDDO
            END IF
            IF (do_ener) THEN
               DO ispin = 1, nspin
                  tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_gx(:), qs_ot_env(ispin)%ener_h_e(i, :))
                  qs_ot_env(1)%ls_diis(i, j) = qs_ot_env(1)%ls_diis(i, j)-qs_ot_env(1)%ds_min*tmp
               ENDDO
            END IF
         ENDIF
         qs_ot_env(1)%ls_diis(j, i) = qs_ot_env(1)%ls_diis(i, j)
         qs_ot_env(1)%ls_diis(i, diis_bound+1) = 1.0_dp
         qs_ot_env(1)%ls_diis(diis_bound+1, i) = 1.0_dp
         qs_ot_env(1)%c_diis(i) = 0.0_dp
      ENDDO
      qs_ot_env(1)%ls_diis(diis_bound+1, diis_bound+1) = 0.0_dp
      qs_ot_env(1)%c_diis(diis_bound+1) = 1.0_dp
      ! put in buffer, dgesv destroys
      qs_ot_env(1)%lss_diis = qs_ot_env(1)%ls_diis

      CALL DGESV(diis_bound+1, 1, qs_ot_env(1)%lss_diis, diis_m+1, qs_ot_env(1)%ipivot, &
                 qs_ot_env(1)%c_diis, diis_m+1, info)

      IF (info .NE. 0) THEN
         do_ot_sd = .TRUE.
         WRITE (cp_logger_get_default_unit_nr(logger), *) "Singular DIIS matrix"
      ELSE
         do_ot_sd = .FALSE.
         IF (do_ks) THEN
            DO ispin = 1, nspin
               ! OK, add the vectors now
               CALL cp_dbcsr_set(qs_ot_env(ispin)%matrix_x, 0.0_dp)
               DO i = 1, diis_bound
                  CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, &
                                    qs_ot_env(ispin)%matrix_h_e(i)%matrix, &
                                    alpha_scalar=1.0_dp, beta_scalar=qs_ot_env(1)%c_diis(i))
               ENDDO
               DO i = 1, diis_bound
                  CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, &
                                    qs_ot_env(ispin)%matrix_h_x(i)%matrix, &
                                    alpha_scalar=1.0_dp, beta_scalar=qs_ot_env(1)%c_diis(i))
               ENDDO
               IF (qs_ot_env(ispin)%settings%do_rotation) THEN
                  CALL cp_dbcsr_set(qs_ot_env(ispin)%rot_mat_x, 0.0_dp)
                  DO i = 1, diis_bound
                     CALL cp_dbcsr_add(qs_ot_env(ispin)%rot_mat_x, &
                                       qs_ot_env(ispin)%rot_mat_h_e(i)%matrix, &
                                       alpha_scalar=1.0_dp, beta_scalar=qs_ot_env(1)%c_diis(i))
                  ENDDO
                  DO i = 1, diis_bound
                     CALL cp_dbcsr_add(qs_ot_env(ispin)%rot_mat_x, &
                                       qs_ot_env(ispin)%rot_mat_h_x(i)%matrix, &
                                       alpha_scalar=1.0_dp, beta_scalar=qs_ot_env(1)%c_diis(i))
                  ENDDO
               ENDIF
            ENDDO
         END IF
         IF (do_ener) THEN
            DO ispin = 1, nspin
               qs_ot_env(ispin)%ener_x(:) = 0.0_dp
               DO i = 1, diis_bound
                  qs_ot_env(ispin)%ener_x(:) = qs_ot_env(ispin)%ener_x(:) &
                                               +qs_ot_env(1)%c_diis(i)*qs_ot_env(ispin)%ener_h_e(i, :)
               END DO
               DO i = 1, diis_bound
                  qs_ot_env(ispin)%ener_x(:) = qs_ot_env(ispin)%ener_x(:) &
                                               +qs_ot_env(1)%c_diis(i)*qs_ot_env(ispin)%ener_h_x(i, :)
               END DO
            ENDDO
         END IF
         qs_ot_env(1)%diis_iter = qs_ot_env(1)%diis_iter+1
         IF (qs_ot_env(1)%settings%safer_diis) THEN
            ! now, final check, is the step in fact in the direction of the -gradient ?
            ! if not we're walking towards a sadle point, and should avoid that
            ! the direction of the step is x_new-x_old
            tr_xold_gx = 0.0_dp
            tr_xnew_gx = 0.0_dp
            IF (do_ks) THEN
               DO ispin = 1, nspin
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_h_x(j)%matrix, &
                                      qs_ot_env(ispin)%matrix_gx, tmp)
                  tr_xold_gx = tr_xold_gx+tmp
                  CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_x, &
                                      qs_ot_env(ispin)%matrix_gx, tmp)
                  tr_xnew_gx = tr_xnew_gx+tmp
                  IF (qs_ot_env(ispin)%settings%do_rotation) THEN
                     CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_h_x(j)%matrix, &
                                         qs_ot_env(ispin)%rot_mat_gx, tmp)
                     tr_xold_gx = tr_xold_gx+0.5_dp*tmp
                     CALL cp_dbcsr_trace(qs_ot_env(ispin)%rot_mat_x, &
                                         qs_ot_env(ispin)%rot_mat_gx, tmp)
                     tr_xnew_gx = tr_xnew_gx+0.5_dp*tmp
                  ENDIF
               ENDDO
            END IF
            IF (do_ener) THEN
               DO ispin = 1, nspin
                  tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_h_x(j, :), qs_ot_env(ispin)%ener_gx(:))
                  tr_xold_gx = tr_xold_gx+tmp
                  tmp = DOT_PRODUCT(qs_ot_env(ispin)%ener_x(:), qs_ot_env(ispin)%ener_gx(:))
                  tr_xnew_gx = tr_xnew_gx+tmp
               ENDDO
            END IF
            overlap = (tr_xnew_gx-tr_xold_gx)
            ! OK, bad luck, take a SD step along the preconditioned gradient
            IF (overlap .GT. 0.0_dp) THEN
               do_ot_sd = .TRUE.
            ENDIF
         ENDIF
      ENDIF

      IF (do_ot_sd) THEN
         qs_ot_env(1)%OT_METHOD_FULL = "OT SD"
         IF (do_ks) THEN
            DO ispin = 1, nspin
               CALL cp_dbcsr_set(qs_ot_env(ispin)%matrix_x, 0.0_dp)
               CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, &
                                 qs_ot_env(ispin)%matrix_h_e(j)%matrix, &
                                 1.0_dp, 1.0_dp)
               CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, &
                                 qs_ot_env(ispin)%matrix_h_x(j)%matrix, &
                                 1.0_dp, 1.0_dp)
               IF (qs_ot_env(ispin)%settings%do_rotation) THEN
                  CALL cp_dbcsr_set(qs_ot_env(ispin)%rot_mat_x, 0.0_dp)
                  CALL cp_dbcsr_add(qs_ot_env(ispin)%rot_mat_x, &
                                    qs_ot_env(ispin)%rot_mat_h_e(j)%matrix, &
                                    1.0_dp, 1.0_dp)
                  CALL cp_dbcsr_add(qs_ot_env(ispin)%rot_mat_x, &
                                    qs_ot_env(ispin)%rot_mat_h_x(j)%matrix, &
                                    1.0_dp, 1.0_dp)
               ENDIF
            ENDDO
         END IF
         IF (do_ener) THEN
            DO ispin = 1, nspin
               qs_ot_env(ispin)%ener_x(:) = 0._dp
               qs_ot_env(ispin)%ener_x(:) = qs_ot_env(ispin)%ener_x(:)+qs_ot_env(ispin)%ener_h_e(j, :)
               qs_ot_env(ispin)%ener_x(:) = qs_ot_env(ispin)%ener_x(:)+qs_ot_env(ispin)%ener_h_x(j, :)
            ENDDO
         END IF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE ot_diis_step

! **************************************************************************************************
!> \brief Energy minimizer by Broyden's method
!> \param qs_ot_env variable to control minimizer behaviour
!> \author Kurt Baarman (09.2010)
! **************************************************************************************************
   SUBROUTINE ot_broyden_step(qs_ot_env)
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ot_broyden_step', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: diis_bound, diis_m, i, ispin, itmp, j, &
                                                            k, n, nspin
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: circ_index
      LOGICAL                                            :: adaptive_sigma, do_ener, do_ks, &
                                                            enable_flip, forget_history
      REAL(KIND=dp)                                      :: beta, eta, gamma, omega, sigma, &
                                                            sigma_dec, sigma_min, tmp, tmp2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: f, x
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: G, S
      TYPE(cp_logger_type), POINTER                      :: logger

      eta = qs_ot_env(1)%settings%broyden_eta
      omega = qs_ot_env(1)%settings%broyden_omega
      sigma_dec = qs_ot_env(1)%settings%broyden_sigma_decrease
      sigma_min = qs_ot_env(1)%settings%broyden_sigma_min
      forget_history = qs_ot_env(1)%settings%broyden_forget_history
      adaptive_sigma = qs_ot_env(1)%settings%broyden_adaptive_sigma
      enable_flip = qs_ot_env(1)%settings%broyden_enable_flip

      do_ks = qs_ot_env(1)%settings%ks
      do_ener = qs_ot_env(1)%settings%do_ener

      beta = qs_ot_env(1)%settings%broyden_beta
      gamma = qs_ot_env(1)%settings%broyden_gamma
      IF (adaptive_sigma) THEN
         IF (qs_ot_env(1)%broyden_adaptive_sigma .LT. 0.0_dp) THEN
            sigma = qs_ot_env(1)%settings%broyden_sigma
         ELSE
            sigma = qs_ot_env(1)%broyden_adaptive_sigma
         ENDIF
      ELSE
         sigma = qs_ot_env(1)%settings%broyden_sigma
      ENDIF

      ! simplify our life....
      IF (do_ener .OR. .NOT. do_ks .OR. qs_ot_env(1)%settings%do_rotation) THEN
         CALL cp__b("qs_ot_minimizer.F",1205,"Not yet implemented")
      ENDIF
      !
      nspin = SIZE(qs_ot_env)

      diis_m = qs_ot_env(1)%settings%diis_m

      IF (qs_ot_env(1)%diis_iter .LT. diis_m) THEN
         diis_bound = qs_ot_env(1)%diis_iter+1
      ELSE
         diis_bound = diis_m
      ENDIF

      ! We want x:s, f:s and one random vector
      k = 2*diis_bound+1
      ALLOCATE (S(k, k))
      ALLOCATE (G(k, k))
      ALLOCATE (f(k))
      ALLOCATE (x(k))
      ALLOCATE (circ_index(diis_bound))
      G = 0.0
      DO i = 1, k
         G(i, i) = sigma
      ENDDO
      S = 0.0

      j = MOD(qs_ot_env(1)%diis_iter, diis_m)+1 ! index in the circular array

      DO ispin = 1, nspin
         CALL cp_dbcsr_copy(qs_ot_env(ispin)%matrix_h_x(j)%matrix, qs_ot_env(ispin)%matrix_x)
      ENDDO

      IF (ASSOCIATED(qs_ot_env(1)%preconditioner)) THEN
         qs_ot_env(1)%gnorm = 0.0_dp
         DO ispin = 1, nspin
            CALL apply_preconditioner(qs_ot_env(ispin)%preconditioner, &
                                      qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_h_e(j)%matrix)
            CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_h_e(j)%matrix, &
                                tmp)
            qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
         ENDDO
         IF (qs_ot_env(1)%gnorm .LT. 0.0_dp) THEN
            WRITE (cp_logger_get_default_unit_nr(logger), *) "WARNING Preconditioner not positive definite !"
         ENDIF
         DO ispin = 1, nspin
            CALL cp_dbcsr_scale(qs_ot_env(ispin)%matrix_h_e(j)%matrix, -1.0_dp)
         ENDDO
      ELSE
         qs_ot_env(1)%gnorm = 0.0_dp
         DO ispin = 1, nspin
            CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, qs_ot_env(ispin)%matrix_gx, tmp)
            qs_ot_env(1)%gnorm = qs_ot_env(1)%gnorm+tmp
            CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_h_e(j)%matrix, &
                              qs_ot_env(ispin)%matrix_gx, alpha_scalar=0.0_dp, beta_scalar=-1.0_dp)
         ENDDO
      ENDIF

      k = 0
      n = 0
      CALL cp_dbcsr_get_info(qs_ot_env(1)%matrix_x, nfullrows_total=n)
      DO ispin = 1, nspin
         CALL cp_dbcsr_get_info(qs_ot_env(ispin)%matrix_x, nfullcols_total=itmp)
         k = k+itmp
      ENDDO

      ! Handling the case of no free variables to optimize
      IF (INT(n, KIND=int_8)*INT(k, KIND=int_8) /= 0) THEN
         qs_ot_env(1)%delta = SQRT(ABS(qs_ot_env(1)%gnorm)/(INT(n, KIND=int_8)*INT(k, KIND=int_8)))
         qs_ot_env(1)%gradient = -qs_ot_env(1)%gnorm
      ELSE
         qs_ot_env(1)%delta = 0.0_dp
         qs_ot_env(1)%gradient = 0.0_dp
      END IF

      IF (diis_bound == diis_m) THEN
         DO i = 1, diis_bound
            circ_index(i) = MOD(j+i-1, diis_m)+1
         ENDDO
      ELSE
         DO i = 1, diis_bound
            circ_index(i) = i
         ENDDO
      ENDIF

      S = 0.0_dp
      DO ispin = 1, nspin
         CALL cp_dbcsr_init_random(qs_ot_env(ispin)%matrix_x)
         DO i = 1, diis_bound
            CALL cp_dbcsr_trace( &
               qs_ot_env(ispin)%matrix_h_x(circ_index(i))%matrix, &
               qs_ot_env(ispin)%matrix_h_x(circ_index(i))%matrix, tmp)
            S(i, i) = S(i, i)+tmp
            CALL cp_dbcsr_trace( &
               qs_ot_env(ispin)%matrix_h_e(circ_index(i))%matrix, &
               qs_ot_env(ispin)%matrix_h_e(circ_index(i))%matrix, tmp)
            S(i+diis_bound, i+diis_bound) = S(i+diis_bound, i+diis_bound)+tmp
            CALL cp_dbcsr_trace( &
               qs_ot_env(ispin)%matrix_h_x(circ_index(i))%matrix, &
               qs_ot_env(ispin)%matrix_x, tmp)
            S(i, 2*diis_bound+1) = S(i, 2*diis_bound+1)+tmp
            S(i, 2*diis_bound+1) = S(2*diis_bound+1, i)
            CALL cp_dbcsr_trace( &
               qs_ot_env(ispin)%matrix_h_e(circ_index(i))%matrix, &
               qs_ot_env(ispin)%matrix_x, tmp)
            S(i+diis_bound, 2*diis_bound+1) = S(i+diis_bound, 2*diis_bound+1)+tmp
            S(i+diis_bound, 2*diis_bound+1) = S(2*diis_bound+1, diis_bound+i)
            DO k = (i+1), diis_bound
               CALL cp_dbcsr_trace( &
                  qs_ot_env(ispin)%matrix_h_x(circ_index(i))%matrix, &
                  qs_ot_env(ispin)%matrix_h_x(circ_index(k))%matrix, &
                  tmp)
               S(i, k) = S(i, k)+tmp
               S(k, i) = S(i, k)
               CALL cp_dbcsr_trace( &
                  qs_ot_env(ispin)%matrix_h_e(circ_index(i))%matrix, &
                  qs_ot_env(ispin)%matrix_h_e(circ_index(k))%matrix, &
                  tmp)
               S(diis_bound+i, diis_bound+k) = S(diis_bound+i, diis_bound+k)+tmp
               S(diis_bound+k, diis_bound+i) = S(diis_bound+i, diis_bound+k)
            ENDDO
            DO k = 1, diis_bound
               CALL cp_dbcsr_trace( &
                  qs_ot_env(ispin)%matrix_h_x(circ_index(i))%matrix, &
                  qs_ot_env(ispin)%matrix_h_e(circ_index(k))%matrix, tmp)
               S(i, k+diis_bound) = S(i, k+diis_bound)+tmp
               S(k+diis_bound, i) = S(i, k+diis_bound)
            ENDDO
         ENDDO
         CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_x, qs_ot_env(ispin)%matrix_x, tmp)
         S(2*diis_bound+1, 2*diis_bound+1) = S(2*diis_bound+1, 2*diis_bound+1)+tmp
      ENDDO

      ! normalize
      k = 2*diis_bound+1
      tmp = SQRT(S(k, k))
      S(k, :) = S(k, :)/tmp
      S(:, k) = S(:, k)/tmp

      IF (diis_bound .GT. 1) THEN
         tmp = 0.0_dp
         tmp2 = 0.0_dp
         i = diis_bound
         DO ispin = 1, nspin
            ! dot product of differences
            CALL cp_dbcsr_trace( &
               qs_ot_env(ispin)%matrix_h_x(circ_index(i))%matrix, &
               qs_ot_env(ispin)%matrix_h_e(circ_index(i))%matrix, &
               tmp)
            tmp2 = tmp2+tmp
            CALL cp_dbcsr_trace( &
               qs_ot_env(ispin)%matrix_h_x(circ_index(i-1))%matrix, &
               qs_ot_env(ispin)%matrix_h_e(circ_index(i))%matrix, &
               tmp)
            tmp2 = tmp2-tmp
            CALL cp_dbcsr_trace( &
               qs_ot_env(ispin)%matrix_h_x(circ_index(i))%matrix, &
               qs_ot_env(ispin)%matrix_h_e(circ_index(i-1))%matrix, &
               tmp)
            tmp2 = tmp2-tmp
            CALL cp_dbcsr_trace( &
               qs_ot_env(ispin)%matrix_h_x(circ_index(i-1))%matrix, &
               qs_ot_env(ispin)%matrix_h_e(circ_index(i-1))%matrix, &
               tmp)
            tmp2 = tmp2+tmp
         ENDDO
         qs_ot_env(1)%c_broy(i-1) = tmp2
      ENDIF

      qs_ot_env(1)%energy_h(j) = qs_ot_env(1)%etotal

      ! If we went uphill, do backtracking line search
      i = MINLOC(qs_ot_env(1)%energy_h(1:diis_bound), dim=1)
      IF (i .NE. j) THEN
         sigma = sigma_dec*sigma
         qs_ot_env(1)%OT_METHOD_FULL = "OT BTRK"
         DO ispin = 1, nspin
            CALL cp_dbcsr_set(qs_ot_env(ispin)%matrix_x, 0.0_dp)
            CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, &
                              qs_ot_env(ispin)%matrix_h_x(i)%matrix, &
                              alpha_scalar=1.0_dp, beta_scalar=(1.0_dp-gamma))
            CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, &
                              qs_ot_env(ispin)%matrix_h_x(circ_index(diis_bound))%matrix, &
                              alpha_scalar=1.0_dp, beta_scalar=gamma)
         ENDDO
      ELSE
         ! Construct G
         DO i = 2, diis_bound
            f = 0.0
            x = 0.0
            ! f is df_i
            x(i) = 1.0
            x(i-1) = -1.0
            ! x is dx_i
            f(diis_bound+i) = 1.0
            f(diis_bound+i-1) = -1.0
            tmp = 1.0_dp
            ! We want a pos def Hessian
            IF (enable_flip) THEN
               IF (qs_ot_env(1)%c_broy(i-1) .GT. 0) THEN
                  !qs_ot_env(1)%OT_METHOD_FULL="OT FLIP"
                  tmp = -1.0_dp
               ENDIF
            ENDIF

            ! get dx-Gdf
            x(:) = tmp*x-MATMUL(G, f)
            ! dfSdf
            ! we calculate matmul(S, f) twice. They're small...
            tmp = DOT_PRODUCT(f, MATMUL(S, f))
            ! NOTE THAT S IS SYMMETRIC !!!
            f(:) = MATMUL(S, f)/tmp
            ! the spread is an outer vector product
            G(:, :) = G+SPREAD(x, dim=2, ncopies=SIZE(f))*SPREAD(f, dim=1, ncopies=SIZE(x))
         ENDDO
         f = 0.0_dp
         f(2*diis_bound) = 1.0_dp
         x(:) = -beta*MATMUL(G, f)

         ! OK, add the vectors now, this sums up to the proposed step
         DO ispin = 1, nspin
            CALL cp_dbcsr_set(qs_ot_env(ispin)%matrix_x, 0.0_dp)
            DO i = 1, diis_bound
               CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, &
                                 qs_ot_env(ispin)%matrix_h_e(circ_index(i))%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=-x(i+diis_bound))
            ENDDO
            DO i = 1, diis_bound
               CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, &
                                 qs_ot_env(ispin)%matrix_h_x(circ_index(i))%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=x(i))
            ENDDO
         ENDDO

         IF (adaptive_sigma) THEN
            tmp = new_sigma(G, S, diis_bound)
            !tmp = tmp * qs_ot_env ( 1 ) % settings % broyden_sigma
            tmp = tmp*eta
            sigma = MIN(omega*sigma, tmp)
         ENDIF

         ! compute the inner product of direction of the step and gradient
         tmp = 0.0_dp
         DO ispin = 1, nspin
            CALL cp_dbcsr_trace(qs_ot_env(ispin)%matrix_gx, &
                                qs_ot_env(ispin)%matrix_x, &
                                tmp2)
            tmp = tmp+tmp2
         ENDDO

         DO ispin = 1, nspin
            ! if the direction of the step is not in direction of the gradient,
            ! change step sign
            IF (tmp .GE. 0.0_dp) THEN
               qs_ot_env(1)%OT_METHOD_FULL = "OT TURN"
               IF (forget_history) THEN
                  qs_ot_env(1)%diis_iter = 0
               ENDIF
               sigma = sigma*sigma_dec
               CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, &
                                 qs_ot_env(ispin)%matrix_h_x(circ_index(diis_bound))%matrix, &
                                 alpha_scalar=-1.0_dp, beta_scalar=1.0_dp)
            ELSE
               CALL cp_dbcsr_add(qs_ot_env(ispin)%matrix_x, &
                                 qs_ot_env(ispin)%matrix_h_x(circ_index(diis_bound))%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
            ENDIF
         ENDDO
      ENDIF

      ! get rid of S, G, f, x, circ_index for next round
      DEALLOCATE (S, G, f, x, circ_index)

      ! update for next round
      qs_ot_env(1)%diis_iter = qs_ot_env(1)%diis_iter+1
      qs_ot_env(1)%broyden_adaptive_sigma = MAX(sigma, sigma_min)

   END SUBROUTINE ot_broyden_step

! **************************************************************************************************
!> \brief ...
!> \param G ...
!> \param S ...
!> \param n ...
!> \retval sigma ...
! **************************************************************************************************
   FUNCTION new_sigma(G, S, n) RESULT(sigma)
!
! Calculate new sigma from eigenvalues of full size G by Arnoldi.
!
! **************************************************************************************************

      REAL(KIND=dp), DIMENSION(:, :)                     :: G, S
      INTEGER                                            :: n
      REAL(KIND=dp)                                      :: sigma

      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: eigv
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: H

      ALLOCATE (H(n, n))
      CALL hess_G(G, S, H, n)
      ALLOCATE (eigv(n))
      CALL diamat_all(H(1:n, 1:n), eigv)

      SELECT CASE (1)
      CASE (1)
         ! This estimator seems to work well. No theory.
         sigma = SUM(ABS(eigv**2))/SUM(ABS(eigv))
      CASE (2)
         ! Estimator based on Frobenius norm minimizer
         sigma = SUM(ABS(eigv))/MAX(1, SIZE(eigv))
      CASE (3)
         ! Estimator based on induced 2-norm
         sigma = (MAXVAL(ABS(eigv))+MINVAL(ABS(eigv)))*0.5_dp
      END SELECT

      DEALLOCATE (H, eigv)
   END FUNCTION new_sigma

! **************************************************************************************************
!> \brief ...
!> \param G ...
!> \param S ...
!> \param H ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE hess_G(G, S, H, n)
!
! Make a hessenberg out of G into H. Cf Arnoldi.
! Inner product is weighted by S.
! Possible lucky breakdown at n.
!
! **************************************************************************************************
      REAL(KIND=dp), DIMENSION(:, :)                     :: G, S, H
      INTEGER                                            :: n

      INTEGER                                            :: i, j, k
      REAL(KIND=dp)                                      :: tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: v
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Q

      i = SIZE(G, 1)
      k = SIZE(H, 1)
      ALLOCATE (Q(i, k))
      ALLOCATE (v(i))
      H = 0.0_dp
      Q = 0.0_dp

      Q(:, 1) = 1.0_dp
      tmp = SQRT(DOT_PRODUCT(Q(:, 1), MATMUL(S, Q(:, 1))))
      Q(:, :) = Q(:, :)/tmp

      DO i = 1, k
         v(:) = MATMUL(G, Q(:, i))
         DO j = 1, i
            H(j, i) = DOT_PRODUCT(Q(:, j), MATMUL(S, v))
            v(:) = v-H(j, i)*Q(:, j)
         ENDDO
         IF (i .LT. k) THEN
            tmp = DOT_PRODUCT(v, MATMUL(S, v))
            IF (tmp .LE. 0.0_dp) THEN
               n = i
               EXIT
            ENDIF
            tmp = SQRT(tmp)
            ! Lucky breakdown
            IF (ABS(tmp) .LT. 1e-9_dp) THEN
               n = i
               EXIT
            ENDIF
            H(i+1, i) = tmp
            Q(:, i+1) = v/H(i+1, i)
         ENDIF
      ENDDO

      DEALLOCATE (Q, v)
   END SUBROUTINE hess_G

END MODULE qs_ot_minimizer
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief an eigen-space solver for the generalised symmetric eigenvalue problem
!>      for sparse matrices, needing only multiplications
!> \author Joost VandeVondele (25.08.2002)
! **************************************************************************************************
MODULE qs_ot_eigensolver
   USE cp_dbcsr_cholesky,               ONLY: cp_dbcsr_cholesky_decompose,&
                                              cp_dbcsr_cholesky_invert
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_copy, cp_dbcsr_init_p, cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_release_p, &
        cp_dbcsr_scale, cp_dbcsr_set, cp_dbcsr_trace, cp_dbcsr_type, dbcsr_type_no_symmetry
   USE cp_dbcsr_operations,             ONLY: copy_dbcsr_to_fm,&
                                              cp_dbcsr_copy_columns_hack,&
                                              cp_dbcsr_m_by_n_from_template,&
                                              cp_fm_to_dbcsr_row_template
   USE cp_fm_types,                     ONLY: cp_fm_get_info,&
                                              cp_fm_type
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE kinds,                           ONLY: dp
   USE preconditioner_types,            ONLY: preconditioner_in_use,&
                                              preconditioner_type
   USE qs_mo_methods,                   ONLY: make_basis_sv
   USE qs_ot,                           ONLY: qs_ot_get_orbitals,&
                                              qs_ot_get_p,&
                                              qs_ot_new_preconditioner
   USE qs_ot_minimizer,                 ONLY: ot_mini
   USE qs_ot_types,                     ONLY: qs_ot_allocate,&
                                              qs_ot_destroy,&
                                              qs_ot_init,&
                                              qs_ot_settings_init,&
                                              qs_ot_settings_type,&
                                              qs_ot_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_ot_eigensolver'

! *** Public subroutines ***

   PUBLIC :: ot_eigensolver

CONTAINS

! on input c contains the initial guess (should not be zero !)
! on output c spans the subspace
! **************************************************************************************************
!> \brief ...
!> \param matrix_h ...
!> \param matrix_s ...
!> \param matrix_orthogonal_space_fm ...
!> \param matrix_c_fm ...
!> \param preconditioner ...
!> \param eps_gradient ...
!> \param iter_max ...
!> \param size_ortho_space ...
!> \param silent ...
!> \param ot_settings ...
! **************************************************************************************************
   SUBROUTINE ot_eigensolver(matrix_h, matrix_s, matrix_orthogonal_space_fm, &
                             matrix_c_fm, preconditioner, eps_gradient, &
                             iter_max, size_ortho_space, silent, ot_settings)

      TYPE(cp_dbcsr_type), POINTER                       :: matrix_h, matrix_s
      TYPE(cp_fm_type), OPTIONAL, POINTER                :: matrix_orthogonal_space_fm
      TYPE(cp_fm_type), POINTER                          :: matrix_c_fm
      TYPE(preconditioner_type), OPTIONAL, POINTER       :: preconditioner
      REAL(KIND=dp)                                      :: eps_gradient
      INTEGER, INTENT(IN)                                :: iter_max
      INTEGER, INTENT(IN), OPTIONAL                      :: size_ortho_space
      LOGICAL, INTENT(IN), OPTIONAL                      :: silent
      TYPE(qs_ot_settings_type), INTENT(IN), OPTIONAL    :: ot_settings

      CHARACTER(len=*), PARAMETER :: routineN = 'ot_eigensolver', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: max_iter_inner_loop = 40
      REAL(KIND=dp), PARAMETER                           :: rone = 1.0_dp, rzero = 0.0_dp

      INTEGER                                            :: handle, ieigensolver, iter_total, k, n, &
                                                            ortho_k, ortho_space_k, output_unit
      LOGICAL                                            :: energy_only, my_silent, ortho
      REAL(KIND=dp)                                      :: delta, energy
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_hc
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_buf1_ortho, matrix_buf2_ortho, &
                                                            matrix_c, matrix_orthogonal_space, &
                                                            matrix_os_ortho, matrix_s_ortho
      TYPE(qs_ot_type), DIMENSION(:), POINTER            :: qs_ot_env

      CALL timeset(routineN, handle)

      output_unit = cp_logger_get_default_io_unit()

      IF (PRESENT(silent)) THEN
         my_silent = silent
      ELSE
         my_silent = .FALSE.
      END IF

      NULLIFY (matrix_c) ! fm->dbcsr
      !XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXx

      CALL cp_fm_get_info(matrix_c_fm, nrow_global=n, ncol_global=k) ! fm->dbcsr
      ALLOCATE (matrix_c)
      CALL cp_fm_to_dbcsr_row_template(matrix_c, fm_in=matrix_c_fm, template=matrix_h)

      iter_total = 0

      outer_scf: DO

         NULLIFY (qs_ot_env)

         NULLIFY (matrix_s_ortho)
         NULLIFY (matrix_os_ortho)
         NULLIFY (matrix_buf1_ortho)
         NULLIFY (matrix_buf2_ortho)
         NULLIFY (matrix_orthogonal_space)

         ALLOCATE (qs_ot_env(1))
         ALLOCATE (matrix_hc(1))
         NULLIFY (matrix_hc(1)%matrix)
         CALL cp_dbcsr_init_p(matrix_hc(1)%matrix)
         CALL cp_dbcsr_copy(matrix_hc(1)%matrix, matrix_c, 'matrix_hc')

         ortho = .FALSE.
         IF (PRESENT(matrix_orthogonal_space_fm)) ortho = .TRUE.

         ! decide settings
         IF (PRESENT(ot_settings)) THEN
            qs_ot_env(1)%settings = ot_settings
         ELSE
            CALL qs_ot_settings_init(qs_ot_env(1)%settings)
            ! overwrite defaults
            qs_ot_env(1)%settings%ds_min = 0.10_dp
         END IF

         IF (ortho) THEN
            ALLOCATE (matrix_orthogonal_space)
            CALL cp_fm_to_dbcsr_row_template(matrix_orthogonal_space, fm_in=matrix_orthogonal_space_fm, template=matrix_h)
            CALL cp_fm_get_info(matrix_orthogonal_space_fm, ncol_global=ortho_space_k)

            IF (PRESENT(size_ortho_space)) ortho_space_k = size_ortho_space
            ortho_k = ortho_space_k+k
         ELSE
            ortho_k = k
         ENDIF

         ! allocate
         CALL qs_ot_allocate(qs_ot_env(1), matrix_s, matrix_c_fm%matrix_struct, ortho_k=ortho_k)

         IF (ortho) THEN
            ! construct an initial guess that is orthogonal to matrix_orthogonal_space

            CALL cp_dbcsr_init_p(matrix_s_ortho)
            CALL cp_dbcsr_copy(matrix_s_ortho, matrix_orthogonal_space, name="matrix_s_ortho")

            CALL cp_dbcsr_init_p(matrix_os_ortho)
            CALL cp_dbcsr_m_by_n_from_template(matrix_os_ortho, template=matrix_h, m=ortho_space_k, n=ortho_space_k, &
                                               sym=dbcsr_type_no_symmetry)

            CALL cp_dbcsr_init_p(matrix_buf1_ortho)
            CALL cp_dbcsr_m_by_n_from_template(matrix_buf1_ortho, template=matrix_h, m=ortho_space_k, n=k, &
                                               sym=dbcsr_type_no_symmetry)

            CALL cp_dbcsr_init_p(matrix_buf2_ortho)
            CALL cp_dbcsr_m_by_n_from_template(matrix_buf2_ortho, template=matrix_h, m=ortho_space_k, n=k, &
                                               sym=dbcsr_type_no_symmetry)

            CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, matrix_s, matrix_orthogonal_space, &
                                   0.0_dp, matrix_s_ortho)
            CALL cp_dbcsr_multiply('T', 'N', rone, matrix_s_ortho, matrix_s_ortho, &
                                   rzero, matrix_os_ortho)

            CALL cp_dbcsr_cholesky_decompose(matrix_os_ortho, &
                                             para_env=qs_ot_env(1)%para_env, blacs_env=qs_ot_env(1)%blacs_env)
            CALL cp_dbcsr_cholesky_invert(matrix_os_ortho, &
                                          para_env=qs_ot_env(1)%para_env, blacs_env=qs_ot_env(1)%blacs_env, &
                                          upper_to_full=.TRUE.)

            CALL cp_dbcsr_multiply('T', 'N', rone, matrix_s_ortho, matrix_c, &
                                   rzero, matrix_buf1_ortho)
            CALL cp_dbcsr_multiply('N', 'N', rone, matrix_os_ortho, matrix_buf1_ortho, &
                                   rzero, matrix_buf2_ortho)
            CALL cp_dbcsr_multiply('N', 'N', -rone, matrix_s_ortho, matrix_buf2_ortho, &
                                   rone, matrix_c)

            ! make matrix_c0 an orthogonal basis, matrix_c contains sc0
            CALL cp_dbcsr_copy(qs_ot_env(1)%matrix_c0, matrix_c)
            CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, matrix_s, qs_ot_env(1)%matrix_c0, &
                                   0.0_dp, matrix_c)

            CALL make_basis_sv(qs_ot_env(1)%matrix_c0, k, matrix_c, &
                               qs_ot_env(1)%para_env, qs_ot_env(1)%blacs_env)

            ! copy sc0 and matrix_s_ortho in qs_ot_env(1)%matrix_sc0
            !CALL cp_dbcsr_copy_columns(qs_ot_env(1)%matrix_sc0,matrix_s_ortho,ortho_space_k,1,1)
            CALL cp_dbcsr_copy_columns_hack(qs_ot_env(1)%matrix_sc0, matrix_s_ortho, ortho_space_k, 1, 1, &
                                            para_env=qs_ot_env(1)%para_env, blacs_env=qs_ot_env(1)%blacs_env)
            !CALL cp_dbcsr_copy_columns(qs_ot_env(1)%matrix_sc0,matrix_c,k,1,ortho_space_k+1)
            CALL cp_dbcsr_copy_columns_hack(qs_ot_env(1)%matrix_sc0, matrix_c, k, 1, ortho_space_k+1, &
                                            para_env=qs_ot_env(1)%para_env, blacs_env=qs_ot_env(1)%blacs_env)

            CALL cp_dbcsr_release_p(matrix_buf1_ortho)
            CALL cp_dbcsr_release_p(matrix_buf2_ortho)
            CALL cp_dbcsr_release_p(matrix_os_ortho)
            CALL cp_dbcsr_release_p(matrix_s_ortho)

         ELSE

            ! set c0,sc0
            CALL cp_dbcsr_copy(qs_ot_env(1)%matrix_c0, matrix_c)
            CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, matrix_s, qs_ot_env(1)%matrix_c0, &
                                   0.0_dp, qs_ot_env(1)%matrix_sc0)

            CALL make_basis_sv(qs_ot_env(1)%matrix_c0, k, qs_ot_env(1)%matrix_sc0, &
                               qs_ot_env(1)%para_env, qs_ot_env(1)%blacs_env)
         ENDIF

         ! init
         CALL qs_ot_init(qs_ot_env(1))
         energy_only = qs_ot_env(1)%energy_only

         ! set x
         CALL cp_dbcsr_set(qs_ot_env(1)%matrix_x, 0.0_dp)
         CALL cp_dbcsr_set(qs_ot_env(1)%matrix_sx, 0.0_dp)

         ! get c
         CALL qs_ot_get_p(qs_ot_env(1)%matrix_x, qs_ot_env(1)%matrix_sx, qs_ot_env(1))
         CALL qs_ot_get_orbitals(matrix_c, qs_ot_env(1)%matrix_x, qs_ot_env(1))

         ! if present preconditioner, use it

         IF (PRESENT(preconditioner)) THEN
            IF (ASSOCIATED(preconditioner)) THEN
               IF (preconditioner_in_use(preconditioner)) THEN
                  CALL qs_ot_new_preconditioner(qs_ot_env(1), preconditioner)
               ELSE
                  ! we should presumably make one
               END IF
            ENDIF
         ENDIF

!     *** Eigensolver loop ***
         ieigensolver = 0
         eigensolver_loop: DO

            ieigensolver = ieigensolver+1
            iter_total = iter_total+1

            ! the energy is cHc, the gradient is 2*H*c
            CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, matrix_h, matrix_c, &
                                   0.0_dp, matrix_hc(1)%matrix)
            CALL cp_dbcsr_trace(matrix_c, matrix_hc(1)%matrix, energy)
            IF (.NOT. energy_only) THEN
               CALL cp_dbcsr_scale(matrix_hc(1)%matrix, 2.0_dp)
            ENDIF

            qs_ot_env(1)%etotal = energy
            CALL ot_mini(qs_ot_env, matrix_hc, output_unit=0)
            delta = qs_ot_env(1)%delta
            energy_only = qs_ot_env(1)%energy_only

            CALL cp_dbcsr_multiply('N', 'N', 1.0_dp, matrix_s, qs_ot_env(1)%matrix_x, &
                                   0.0_dp, qs_ot_env(1)%matrix_sx)

            CALL qs_ot_get_p(qs_ot_env(1)%matrix_x, qs_ot_env(1)%matrix_sx, qs_ot_env(1))
            CALL qs_ot_get_orbitals(matrix_c, qs_ot_env(1)%matrix_x, qs_ot_env(1))

            ! exit on convergence or if maximum of inner loop  cycles is reached
            IF (delta < eps_gradient .OR. ieigensolver >= max_iter_inner_loop) EXIT eigensolver_loop
            ! exit if total number of steps is reached, but not during a line search step
            IF (iter_total >= iter_max .AND. qs_ot_env(1)%OT_METHOD_FULL /= "OT LS") EXIT eigensolver_loop

         END DO eigensolver_loop

         CALL qs_ot_destroy(qs_ot_env(1))
         DEALLOCATE (qs_ot_env)
         CALL cp_dbcsr_release_p(matrix_hc(1)%matrix)
         DEALLOCATE (matrix_hc)
         CALL cp_dbcsr_release_p(matrix_orthogonal_space)

         IF (delta < eps_gradient) THEN
            IF ((output_unit > 0) .AND. .NOT. my_silent) THEN
               WRITE (output_unit, *) " Reached convergence in ", iter_total, " iterations "
            ENDIF
            EXIT outer_scf
         ENDIF
         IF (iter_total >= iter_max) THEN
            IF (output_unit > 0) THEN
               IF (my_silent) THEN
                  WRITE (output_unit, "(A,T60,E20.10)") "  WARNING OT eigensolver did not converge: current gradient", delta
               ELSE
                  WRITE (output_unit, *) "WARNING : did not converge in ot_eigensolver"
                  WRITE (output_unit, *) "number of iterations ", iter_total, " exceeded maximum"
                  WRITE (output_unit, *) "current gradient / target gradient", delta, " / ", eps_gradient
               END IF
            ENDIF
            EXIT outer_scf
         ENDIF

      ENDDO outer_scf

      CALL copy_dbcsr_to_fm(matrix_c, matrix_c_fm) ! fm->dbcsr
      CALL cp_dbcsr_release_p(matrix_c) ! fm->dbcsr

      CALL timestop(handle)

   END SUBROUTINE ot_eigensolver

END MODULE qs_ot_eigensolver
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
MODULE qs_integrate_potential_low
   USE ao_util,                         ONLY: exp_radius_very_extended
   USE cell_types,                      ONLY: cell_type
   USE cube_utils,                      ONLY: compute_cube_center,&
                                              cube_info_type,&
                                              return_cube,&
                                              return_cube_nonortho
   USE d3_poly,                         ONLY: poly_d32cp2k
   USE gauss_colloc,                    ONLY: integrateGaussFull
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE mathconstants,                   ONLY: fac
   USE orbital_pointers,                ONLY: coset,&
                                              current_maxl,&
                                              ncoset
   USE realspace_grid_types,            ONLY: realspace_grid_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_integrate_potential_low'

! *** Don't include this routines directly, use the interface to
! *** qs_integrate_potential

   PUBLIC :: integrate_pgf_product_rspace

CONTAINS

! **************************************************************************************************
!> \brief low level function to compute matrix elements of primitive gaussian functions
!> \param la_max ...
!> \param zeta ...
!> \param la_min ...
!> \param lb_max ...
!> \param zetb ...
!> \param lb_min ...
!> \param ra ...
!> \param rab ...
!> \param rab2 ...
!> \param rsgrid ...
!> \param cell ...
!> \param cube_info ...
!> \param hab ...
!> \param pab ...
!> \param o1 ...
!> \param o2 ...
!> \param eps_gvg_rspace ...
!> \param calculate_forces ...
!> \param hdab ...
!> \param hadb ...
!> \param force_a ...
!> \param force_b ...
!> \param compute_tau ...
!> \param map_consistent ...
!> \param collocate_rho0 ...
!> \param rpgf0_s ...
!> \param use_virial ...
!> \param my_virial_a ...
!> \param my_virial_b ...
!> \param a_hdab ...
!> \param use_subpatch ...
!> \param subpatch_pattern ...
! **************************************************************************************************
   SUBROUTINE integrate_pgf_product_rspace(la_max, zeta, la_min, &
                                           lb_max, zetb, lb_min, &
                                           ra, rab, rab2, rsgrid, cell, &
                                           cube_info, hab, pab, o1, o2, &
                                           eps_gvg_rspace, &
                                           calculate_forces, hdab, hadb, force_a, force_b, &
                                           compute_tau, map_consistent, &
                                           collocate_rho0, rpgf0_s, use_virial, my_virial_a, &
                                           my_virial_b, a_hdab, use_subpatch, subpatch_pattern)

      INTEGER, INTENT(IN)                      :: la_max
      REAL(KIND=dp), INTENT(IN)                :: zeta
      INTEGER, INTENT(IN)                      :: la_min, lb_max
      REAL(KIND=dp), INTENT(IN)                :: zetb
      INTEGER, INTENT(IN)                      :: lb_min
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)  :: ra, rab
      REAL(KIND=dp), INTENT(IN)                :: rab2
      TYPE(realspace_grid_type), POINTER       :: rsgrid
      TYPE(cell_type), POINTER                 :: cell
      TYPE(cube_info_type), INTENT(IN)         :: cube_info
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: hab
      REAL(KIND=dp), DIMENSION(:, :), &
         OPTIONAL, POINTER                      :: pab
      INTEGER, INTENT(IN)                      :: o1, o2
      REAL(KIND=dp), INTENT(IN)                :: eps_gvg_rspace
      LOGICAL, INTENT(IN)                      :: calculate_forces
      REAL(KIND=dp), DIMENSION(:, :, :), &
         OPTIONAL, POINTER                      :: hdab, hadb
      REAL(KIND=dp), DIMENSION(3), &
         INTENT(INOUT), OPTIONAL                :: force_a, force_b
      LOGICAL, INTENT(IN), OPTIONAL            :: compute_tau, map_consistent, &
                                                  collocate_rho0
      REAL(dp), INTENT(IN), OPTIONAL           :: rpgf0_s
      LOGICAL, INTENT(IN), OPTIONAL            :: use_virial
      REAL(KIND=dp), DIMENSION(3, 3), OPTIONAL  :: my_virial_a, my_virial_b
      REAL(KIND=dp), DIMENSION(:, :, :, :), OPTIONAL, POINTER  :: a_hdab
      LOGICAL, OPTIONAL                        :: use_subpatch
      INTEGER(KIND=int_8), INTENT(IN), OPTIONAL :: subpatch_pattern

      CHARACTER(len=*), PARAMETER :: routineN = 'integrate_pgf_product_rspace', &
                                     routineP = moduleN//':'//routineN

      INTEGER :: ax, ay, az, bx, by, bz, cmax, coef_max, gridbounds(2, 3), i, &
                 ico, icoef, ig, j, jco, la, la_max_local, la_min_local, lb, &
                 lb_cube_min, lb_max_local, lb_min_local, length, lxa, lxb, lxy, &
                 lxyz, lya, lyb, lza, lzb, offset, start, ub_cube_max
      INTEGER, DIMENSION(3)                    :: cubecenter, lb_cube, ng, &
                                                  ub_cube
      INTEGER, DIMENSION(:), POINTER           :: sphere_bounds
      LOGICAL                                  :: my_collocate_rho0, &
                                                  my_compute_tau, &
                                                  my_map_consistent, &
                                                  my_use_virial, &
                                                  subpatch_integrate
      REAL(KIND=dp) :: axpm0, cutoff, &
                       der_a(3), der_b(3), f, ftza, ftzb, pabval, pg, &
                       prefactor, radius, rpg, zetp
      REAL(KIND=dp), DIMENSION(3)              :: dr, rap, rb, rbp, roffset, rp
      REAL(KIND=dp), DIMENSION(:, :, :), &
         POINTER                                :: grid

      INTEGER :: lxp, lyp, lzp, lp, iaxis
      INTEGER, ALLOCATABLE, DIMENSION(:, :) :: map
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:) :: coef_xyz
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:) :: coef_xyt
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:) :: coef_xtt
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :, :) :: pol_z
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :, :) :: pol_y
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :) :: pol_x
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:, :) :: vab

      REAL(KIND=dp) :: t_exp_1, t_exp_2, t_exp_min_1, t_exp_min_2, t_exp_plus_1, t_exp_plus_2

      subpatch_integrate = .FALSE.

      IF (PRESENT(use_subpatch)) THEN
         IF (use_subpatch) THEN
            subpatch_integrate = .TRUE.
            IF(.NOT.(PRESENT(subpatch_pattern)))CALL cp__a("qs_integrate_potential_low.F",151)
         ENDIF
      ENDIF

      IF (PRESENT(use_virial)) THEN
         my_use_virial = use_virial
      ELSE
         my_use_virial = .FALSE.
      ENDIF

      ! my_compute_tau defaults to .FALSE.
      ! IF (.true.) it will compute 0.5 * (nabla x_a).(v(r) nabla x_b)
      IF (PRESENT(compute_tau)) THEN
         my_compute_tau = compute_tau
      ELSE
         my_compute_tau = .FALSE.
      ENDIF

      ! use identical radii for integrate and collocate ?
      IF (PRESENT(map_consistent)) THEN
         my_map_consistent = map_consistent
      ELSE
         my_map_consistent = .FALSE.
      ENDIF

      IF (PRESENT(collocate_rho0) .AND. PRESENT(rpgf0_s)) THEN
         my_collocate_rho0 = collocate_rho0
      ELSE
         my_collocate_rho0 = .FALSE.
      END IF

      IF (calculate_forces) THEN
         la_max_local = la_max+1 ! needed for the derivative of the gaussian, unimportant which one
         la_min_local = MAX(la_min-1, 0) ! just in case the la_min,lb_min is not zero
         lb_min_local = MAX(lb_min-1, 0)
         lb_max_local = lb_max
         IF (my_use_virial) THEN
            la_max_local = la_max_local+1
            lb_max_local = lb_max_local+1
         ENDIF
      ELSE
         la_max_local = la_max
         la_min_local = la_min
         lb_min_local = lb_min
         lb_max_local = lb_max
      END IF

      IF (my_compute_tau) THEN
         la_max_local = la_max_local+1
         lb_max_local = lb_max_local+1
         la_min_local = MAX(la_min_local-1, 0)
         lb_min_local = MAX(lb_min_local-1, 0)
      ENDIF

      coef_max = la_max_local+lb_max_local+1
      zetp = zeta+zetb
      f = zetb/zetp
      prefactor = EXP(-zeta*f*rab2)
!   *** position of the gaussian product
      rap(:) = f*rab(:)
      rbp(:) = rap(:)-rab(:)
      rp(:) = ra(:)+rap(:) ! this is the gaussian center in real coordinates
      rb(:) = ra(:)+rab(:)

      IF (my_map_consistent) THEN ! still assumes that eps_gvg_rspace=eps_rho_rspace
         cutoff = 1.0_dp
         radius = exp_radius_very_extended(la_min, la_max, lb_min, lb_max, ra=ra, rb=rb, rp=rp, &
                                           zetp=zetp, eps=eps_gvg_rspace, prefactor=prefactor, cutoff=cutoff)
      ELSE IF (my_collocate_rho0) THEN
         cutoff = 0.0_dp
         prefactor = 1.0_dp
         radius = rpgf0_s
      ELSE
         cutoff = 1.0_dp
         IF (PRESENT(pab)) THEN
            radius = exp_radius_very_extended(la_min, la_max, lb_min, lb_max, pab, o1, o2, ra, rb, rp, &
                                              zetp, eps_gvg_rspace, prefactor, cutoff)
         ELSE
            radius = exp_radius_very_extended(la_min, la_max, lb_min, lb_max, ra=ra, rb=rb, rp=rp, &
                                              zetp=zetp, eps=eps_gvg_rspace, prefactor=prefactor, cutoff=cutoff)
         ENDIF
      ENDIF

      IF (radius == 0.0_dp) THEN
         RETURN
      ENDIF

      ng(:) = rsgrid%desc%npts(:)
      grid => rsgrid%r(:, :, :)
      ALLOCATE (vab(ncoset(la_max_local), ncoset(lb_max_local)))
      vab = 0.0_dp

      ! the likely call is integrate_ortho
      IF (.NOT. subpatch_integrate) THEN
         IF (rsgrid%desc%orthorhombic) THEN
            CALL integrate_ortho()
         ELSE
            CALL integrate_general_wings()
         END IF
      ELSE
         CALL integrate_general_subpatch()
      END IF

!   *** vab contains all the information needed to find the elements of hab
!   *** and optionally of derivatives of these elements

      ftza = 2.0_dp*zeta
      ftzb = 2.0_dp*zetb

      DO la = la_min, la_max
         DO ax = 0, la
            DO ay = 0, la-ax
               az = la-ax-ay
               ico = coset(ax, ay, az)
               DO lb = lb_min, lb_max
                  DO bx = 0, lb
                     DO by = 0, lb-bx
                        bz = lb-bx-by
                        jco = coset(bx, by, bz)
                        IF (.NOT. my_compute_tau) THEN
                           axpm0 = vab(ico, jco)
                        ELSE
                           axpm0 = 0.5_dp*(ax*bx*vab(coset(MAX(ax-1, 0), ay, az), coset(MAX(bx-1, 0), by, bz))+ &
                                           ay*by*vab(coset(ax, MAX(ay-1, 0), az), coset(bx, MAX(by-1, 0), bz))+ &
                                           az*bz*vab(coset(ax, ay, MAX(az-1, 0)), coset(bx, by, MAX(bz-1, 0))) &
                                           -ftza*bx*vab(coset(ax+1, ay, az), coset(MAX(bx-1, 0), by, bz)) &
                                           -ftza*by*vab(coset(ax, ay+1, az), coset(bx, MAX(by-1, 0), bz)) &
                                           -ftza*bz*vab(coset(ax, ay, az+1), coset(bx, by, MAX(bz-1, 0))) &
                                           -ax*ftzb*vab(coset(MAX(ax-1, 0), ay, az), coset(bx+1, by, bz)) &
                                           -ay*ftzb*vab(coset(ax, MAX(ay-1, 0), az), coset(bx, by+1, bz)) &
                                           -az*ftzb*vab(coset(ax, ay, MAX(az-1, 0)), coset(bx, by, bz+1))+ &
                                           ftza*ftzb*vab(coset(ax+1, ay, az), coset(bx+1, by, bz))+ &
                                           ftza*ftzb*vab(coset(ax, ay+1, az), coset(bx, by+1, bz))+ &
                                           ftza*ftzb*vab(coset(ax, ay, az+1), coset(bx, by, bz+1)))
                        ENDIF
                        hab(o1+ico, o2+jco) = hab(o1+ico, o2+jco)+axpm0
                        IF (calculate_forces .AND. PRESENT(force_a)) THEN
                           IF (my_compute_tau) THEN
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*ax*bx
                              CALL force_update( &
                                 force_a, force_b, rab, pabval, ftza, ftzb, MAX(ax-1, 0), ay, az, MAX(bx-1, 0), by, bz, vab)
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*ay*by
                              CALL force_update( &
                                 force_a, force_b, rab, pabval, ftza, ftzb, ax, MAX(ay-1, 0), az, bx, MAX(by-1, 0), bz, vab)
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*az*bz
                              CALL force_update( &
                                 force_a, force_b, rab, pabval, ftza, ftzb, ax, ay, MAX(az-1, 0), bx, by, MAX(bz-1, 0), vab)
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*(-ftza*bx)
                              CALL force_update(force_a, force_b, rab, pabval, ftza, ftzb, ax+1, ay, az, MAX(bx-1, 0), by, bz, vab)
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*(-ftza*by)
                              CALL force_update(force_a, force_b, rab, pabval, ftza, ftzb, ax, ay+1, az, bx, MAX(by-1, 0), bz, vab)
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*(-ftza*bz)
                              CALL force_update(force_a, force_b, rab, pabval, ftza, ftzb, ax, ay, az+1, bx, by, MAX(bz-1, 0), vab)
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*(-ax*ftzb)
                              CALL force_update(force_a, force_b, rab, pabval, ftza, ftzb, MAX(ax-1, 0), ay, az, bx+1, by, bz, vab)
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*(-ay*ftzb)
                              CALL force_update(force_a, force_b, rab, pabval, ftza, ftzb, ax, MAX(ay-1, 0), az, bx, by+1, bz, vab)
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*(-az*ftzb)
                              CALL force_update(force_a, force_b, rab, pabval, ftza, ftzb, ax, ay, MAX(az-1, 0), bx, by, bz+1, vab)
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*(ftza*ftzb)
                              CALL force_update(force_a, force_b, rab, pabval, ftza, ftzb, ax+1, ay, az, bx+1, by, bz, vab)
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*(ftza*ftzb)
                              CALL force_update(force_a, force_b, rab, pabval, ftza, ftzb, ax, ay+1, az, bx, by+1, bz, vab)
                              pabval = pab(o1+ico, o2+jco)*0.5_dp*(ftza*ftzb)
                              CALL force_update(force_a, force_b, rab, pabval, ftza, ftzb, ax, ay, az+1, bx, by, bz+1, vab)
                           ELSE
                              pabval = pab(o1+ico, o2+jco)
                              CALL force_update(force_a, force_b, rab, pabval, ftza, ftzb, ax, ay, az, bx, by, bz, vab)
                              IF (my_use_virial) THEN
                                 CALL virial_update(my_virial_a, my_virial_b, rab, pabval, ftza, ftzb, ax, ay, az, bx, by, bz, vab)
                              ENDIF
                           ENDIF
                        END IF
                        IF (calculate_forces .AND. PRESENT(hdab)) THEN
                           der_a(1:3) = 0.0_dp
                           der_b(1:3) = 0.0_dp
                           CALL hab_derivatives(der_a, der_b, rab, ftza, ftzb, ax, ay, az, bx, by, bz, vab)
                           hdab(1:3, o1+ico, o2+jco) = der_a(1:3)
                           hadb(1:3, o1+ico, o2+jco) = der_b(1:3)
                           pabval = 1.0_dp
                           IF (my_use_virial .AND. PRESENT(a_hdab)) THEN
                              my_virial_a = 0.0_dp
                              my_virial_b = 0.0_dp
                              CALL virial_update(my_virial_a, my_virial_b, rab, pabval, ftza, ftzb, ax, ay, az, bx, by, bz, vab)
                              DO j = 1, 3
                                 a_hdab(1:3, j, o1+ico, o2+jco) = a_hdab(1:3, j, o1+ico, o2+jco)+my_virial_a(1:3, j)
                              END DO
                           ENDIF
                        END IF
                     END DO
                  END DO
               END DO
            END DO
         END DO
      END DO
      DEALLOCATE (vab)

   CONTAINS

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE integrate_ortho()

         CALL return_cube(cube_info, radius, lb_cube, ub_cube, sphere_bounds)
         cmax = MAXVAL(ub_cube)

         dr(1) = rsgrid%desc%dh(1, 1)
         dr(2) = rsgrid%desc%dh(2, 2)
         dr(3) = rsgrid%desc%dh(3, 3)

         gridbounds(1, 1) = LBOUND(GRID, 1)
         gridbounds(2, 1) = UBOUND(GRID, 1)
         gridbounds(1, 2) = LBOUND(GRID, 2)
         gridbounds(2, 2) = UBOUND(GRID, 2)
         gridbounds(1, 3) = LBOUND(GRID, 3)
         gridbounds(2, 3) = UBOUND(GRID, 3)

         CALL compute_cube_center(cubecenter, rsgrid%desc, zeta, zetb, ra, rab)
         roffset(:) = rp(:)-REAL(cubecenter(:), dp)*dr(:)
         lb_cube_min = MINVAL(lb_cube(:))
         ub_cube_max = MAXVAL(ub_cube(:))

!   *** a mapping so that the ig corresponds to the right grid point, also with pbc
         ALLOCATE (map(-cmax:cmax, 3))
         DO i = 1, 3
            IF (rsgrid%desc%perd(i) == 1) THEN
               start = lb_cube(i)
               DO
                  offset = MODULO(cubecenter(i)+start, ng(i))+1-start
                  length = MIN(ub_cube(i), ng(i)-offset)-start
                  DO ig = start, start+length
                     map(ig, i) = ig+offset
                  END DO
                  IF (start+length .GE. ub_cube(i)) EXIT
                  start = start+length+1
               END DO
            ELSE
               ! this takes partial grid + border regions into account
               offset = MODULO(cubecenter(i)+lb_cube(i)+rsgrid%desc%lb(i)-rsgrid%lb_local(i), ng(i))+1-lb_cube(i)
               ! check for out of bounds
               IF (ub_cube(i)+offset > UBOUND(grid, i) .OR. lb_cube(i)+offset < LBOUND(grid, i)) THEN
                  CALL cp__b("qs_integrate_potential_low.F",393,"")
               ENDIF
               DO ig = lb_cube(i), ub_cube(i)
                  map(ig, i) = ig+offset
               END DO
            END IF
         ENDDO

         lp = la_max_local+lb_max_local
         ALLOCATE (coef_xyz(((lp+1)*(lp+2)*(lp+3))/6))
         ALLOCATE (pol_z(1:2, 0:lp, -cmax:0))
         ALLOCATE (pol_y(1:2, 0:lp, -cmax:0))
         ALLOCATE (pol_x(0:lp, -cmax:cmax))

!
!   compute the values of all (x-xp)**lp*exp(..)
!
! still requires the old trick:
!  new trick to avoid to many exps (reuse the result from the previous gridpoint):
!  exp( -a*(x+d)**2)=exp(-a*x**2)*(-2*a*x*d)*exp(-a*d**2)
!  exp(-2*a*(x+d)*d)=exp(-2*a*x*d)*exp(-2*a*d**2)

      iaxis=3
      t_exp_1=EXP(-zetp*dr(iaxis)**2)
      t_exp_2=t_exp_1**2
      t_exp_min_1=EXP(-zetp*(+dr(iaxis)- roffset(iaxis))**2)
      t_exp_min_2=EXP(-2*zetp*(+dr(iaxis)- roffset(iaxis))*(-dr(iaxis)))
      t_exp_plus_1=EXP(-zetp*( - roffset(iaxis))**2)
      t_exp_plus_2=EXP(-2*zetp*( - roffset(iaxis))*(+dr(iaxis)))
      DO ig=0,lb_cube(iaxis),-1
        rpg = REAL(ig,dp)*dr(iaxis) - roffset(iaxis)
        t_exp_min_1=t_exp_min_1*t_exp_min_2*t_exp_1
        t_exp_min_2=t_exp_min_2*t_exp_2
        pg = t_exp_min_1
        ! pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_z(1,icoef,ig)=pg
           pg=pg*(rpg)
        ENDDO

        rpg = REAL(1-ig,dp)*dr(iaxis) - roffset(iaxis)
        t_exp_plus_1=t_exp_plus_1*t_exp_plus_2*t_exp_1
        t_exp_plus_2=t_exp_plus_2*t_exp_2
        pg = t_exp_plus_1
        ! pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_z(2,icoef,ig)=pg
           pg=pg*(rpg)
        ENDDO
      ENDDO

      iaxis=2
      t_exp_1=EXP(-zetp*dr(iaxis)**2)
      t_exp_2=t_exp_1**2
      t_exp_min_1=EXP(-zetp*(+dr(iaxis)- roffset(iaxis))**2)
      t_exp_min_2=EXP(-2*zetp*(+dr(iaxis)- roffset(iaxis))*(-dr(iaxis)))
      t_exp_plus_1=EXP(-zetp*( - roffset(iaxis))**2)
      t_exp_plus_2=EXP(-2*zetp*( - roffset(iaxis))*(+dr(iaxis)))
      DO ig=0,lb_cube(iaxis),-1
        rpg = REAL(ig,dp)*dr(iaxis) - roffset(iaxis)
        t_exp_min_1=t_exp_min_1*t_exp_min_2*t_exp_1
        t_exp_min_2=t_exp_min_2*t_exp_2
        pg = t_exp_min_1
        ! pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_y(1,icoef,ig)=pg
           pg=pg*(rpg)
        ENDDO

        rpg = REAL(1-ig,dp)*dr(iaxis) - roffset(iaxis)
        t_exp_plus_1=t_exp_plus_1*t_exp_plus_2*t_exp_1
        t_exp_plus_2=t_exp_plus_2*t_exp_2
        pg = t_exp_plus_1
        ! pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_y(2,icoef,ig)=pg
           pg=pg*(rpg)
        ENDDO
      ENDDO

      iaxis=1
      t_exp_1=EXP(-zetp*dr(iaxis)**2)
      t_exp_2=t_exp_1**2
      t_exp_min_1=EXP(-zetp*(+dr(iaxis)- roffset(iaxis))**2)
      t_exp_min_2=EXP(-2*zetp*(+dr(iaxis)- roffset(iaxis))*(-dr(iaxis)))
      t_exp_plus_1=EXP(-zetp*( - roffset(iaxis))**2)
      t_exp_plus_2=EXP(-2*zetp*( - roffset(iaxis))*(+dr(iaxis)))
      DO ig=0,lb_cube(1),-1

        rpg = REAL(ig,dp)*dr(1) - roffset(1)
        t_exp_min_1=t_exp_min_1*t_exp_min_2*t_exp_1
        t_exp_min_2=t_exp_min_2*t_exp_2
        pg = t_exp_min_1
        !pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_x(icoef,ig)=pg
           pg=pg*(rpg)
        ENDDO

        rpg = REAL(1-ig,dp)*dr(1) - roffset(1)
        t_exp_plus_1=t_exp_plus_1*t_exp_plus_2*t_exp_1
        t_exp_plus_2=t_exp_plus_2*t_exp_2
        pg = t_exp_plus_1
        ! pg  = EXP(-zetp*rpg**2)
        DO icoef=0,lp
           pol_x(icoef,1-ig)=pg
           pg=pg*(rpg)
        ENDDO
      ENDDO



         CALL call_integrate()

         CALL call_to_xyz_to_vab(prefactor, coef_xyz, lp, la_max_local, lb_max_local, &
                                 rp, ra, rab, vab, coset, la_min_local, &
                                 lb_min_local, current_maxl, ncoset(la_max_local), ncoset(lb_max_local))

         DEALLOCATE (coef_xyz)
         DEALLOCATE (pol_z)
         DEALLOCATE (pol_y)
         DEALLOCATE (pol_x)
         DEALLOCATE (map)
      END SUBROUTINE integrate_ortho

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE call_integrate()

         SELECT CASE (lp)
         CASE (0)
            CALL integrate_core_0(grid(1, 1, 1), coef_xyz(1), pol_x(0, -cmax), pol_y(1, 0, -cmax), pol_z(1, 0, -cmax), &
                                  map(-cmax, 1), sphere_bounds(1), cmax, gridbounds(1, 1))
         CASE (1)
            CALL integrate_core_1(grid(1, 1, 1), coef_xyz(1), pol_x(0, -cmax), pol_y(1, 0, -cmax), pol_z(1, 0, -cmax), &
                                  map(-cmax, 1), sphere_bounds(1), cmax, gridbounds(1, 1))
         CASE (2)
            CALL integrate_core_2(grid(1, 1, 1), coef_xyz(1), pol_x(0, -cmax), pol_y(1, 0, -cmax), pol_z(1, 0, -cmax), &
                                  map(-cmax, 1), sphere_bounds(1), cmax, gridbounds(1, 1))
         CASE (3)
            CALL integrate_core_3(grid(1, 1, 1), coef_xyz(1), pol_x(0, -cmax), pol_y(1, 0, -cmax), pol_z(1, 0, -cmax), &
                                  map(-cmax, 1), sphere_bounds(1), cmax, gridbounds(1, 1))
         CASE (4)
            CALL integrate_core_4(grid(1, 1, 1), coef_xyz(1), pol_x(0, -cmax), pol_y(1, 0, -cmax), pol_z(1, 0, -cmax), &
                                  map(-cmax, 1), sphere_bounds(1), cmax, gridbounds(1, 1))
         CASE (5)
            CALL integrate_core_5(grid(1, 1, 1), coef_xyz(1), pol_x(0, -cmax), pol_y(1, 0, -cmax), pol_z(1, 0, -cmax), &
                                  map(-cmax, 1), sphere_bounds(1), cmax, gridbounds(1, 1))
         CASE (6)
            CALL integrate_core_6(grid(1, 1, 1), coef_xyz(1), pol_x(0, -cmax), pol_y(1, 0, -cmax), pol_z(1, 0, -cmax), &
                                  map(-cmax, 1), sphere_bounds(1), cmax, gridbounds(1, 1))
         CASE (7)
            CALL integrate_core_7(grid(1, 1, 1), coef_xyz(1), pol_x(0, -cmax), pol_y(1, 0, -cmax), pol_z(1, 0, -cmax), &
                                  map(-cmax, 1), sphere_bounds(1), cmax, gridbounds(1, 1))
         CASE (8)
            CALL integrate_core_8(grid(1, 1, 1), coef_xyz(1), pol_x(0, -cmax), pol_y(1, 0, -cmax), pol_z(1, 0, -cmax), &
                                  map(-cmax, 1), sphere_bounds(1), cmax, gridbounds(1, 1))
         CASE (9)
            CALL integrate_core_9(grid(1, 1, 1), coef_xyz(1), pol_x(0, -cmax), pol_y(1, 0, -cmax), pol_z(1, 0, -cmax), &
                                  map(-cmax, 1), sphere_bounds(1), cmax, gridbounds(1, 1))
         CASE DEFAULT
            CALL integrate_core_default(grid(1, 1, 1), coef_xyz(1), pol_x(0, -cmax), pol_y(1, 0, -cmax), pol_z(1, 0, -cmax), &
                                        map(-cmax, 1), sphere_bounds(1), lp, cmax, gridbounds(1, 1))
         END SELECT

      END SUBROUTINE call_integrate

! **************************************************************************************************
! general optimized routine, not called for orthogonal cells
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE integrate_general_opt()
      INTEGER :: i, i_index, il, ilx, ily, ilz, index_max(3), index_min(3), ismax, ismin, j, &
         j_index, jl, jlx, jly, jlz, k, k_index, kl, klx, kly, klz, lpx, lpy, lpz, lx, ly, lz, &
         offset(3)
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: grid_map
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: coef_map
      REAL(KIND=dp)                                      :: a, b, c, d, di, dip, dj, djp, dk, dkp, &
                                                            exp0i, exp1i, exp2i, gp(3), gridval, &
                                                            hmatgrid(3, 3), pointj(3), pointk(3), &
                                                            v(3)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: coef_ijk
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: hmatgridp

!
! transform P_{lxp,lyp,lzp} into a P_{lip,ljp,lkp} such that
! sum_{lxp,lyp,lzp} P_{lxp,lyp,lzp} (x-x_p)**lxp (y-y_p)**lyp (z-z_p)**lzp =
! sum_{lip,ljp,lkp} P_{lip,ljp,lkp} (i-i_p)**lip (j-j_p)**ljp (k-k_p)**lkp
!

         lp = la_max_local+lb_max_local
         ALLOCATE (coef_xyz(((lp+1)*(lp+2)*(lp+3))/6))
         ALLOCATE (coef_ijk(((lp+1)*(lp+2)*(lp+3))/6))
         ALLOCATE (coef_xyt(((lp+1)*(lp+2))/2))
         ALLOCATE (coef_xtt(0:lp))

         ! aux mapping array to simplify life
         ALLOCATE (coef_map(0:lp, 0:lp, 0:lp))
         coef_map = HUGE(coef_map)
         lxyz = 0
         DO lzp = 0, lp
         DO lyp = 0, lp-lzp
         DO lxp = 0, lp-lzp-lyp
            lxyz = lxyz+1
            coef_map(lxp, lyp, lzp) = lxyz
         ENDDO
         ENDDO
         ENDDO

         ! cell hmat in grid points
         hmatgrid(:, 1) = cell%hmat(:, 1)/ng(1)
         hmatgrid(:, 2) = cell%hmat(:, 2)/ng(2)
         hmatgrid(:, 3) = cell%hmat(:, 3)/ng(3)

         ! center in grid coords
         gp = MATMUL(cell%h_inv, rp)*ng

         ! added bt matt
         cubecenter(:) = FLOOR(gp)

         !t2=nanotime_ia32()
         !write(*,*) t2-t1
         !t1=nanotime_ia32()

         CALL return_cube_nonortho(cube_info, radius, index_min, index_max, rp)

         offset(:) = MODULO(index_min(:)+rsgrid%desc%lb(:)-rsgrid%lb_local(:), ng(:))+1

         ALLOCATE (grid_map(index_min(1):index_max(1)))
         DO i = index_min(1), index_max(1)
            grid_map(i) = MODULO(i, ng(1))+1
            IF (rsgrid%desc%perd(1) == 1) THEN
               grid_map(i) = MODULO(i, ng(1))+1
            ELSE
               grid_map(i) = i-index_min(1)+offset(1)
            ENDIF
         ENDDO

         coef_ijk = 0.0_dp

         ! go over the grid, but cycle if the point is not within the radius
         DO k = index_min(3), index_max(3)
            dk = k-gp(3)
            pointk = hmatgrid(:, 3)*dk

            ! allow for generalised rs grids
            IF (rsgrid%desc%perd(3) == 1) THEN
               k_index = MODULO(k, ng(3))+1
            ELSE
               k_index = k-index_min(3)+offset(3)
            ENDIF

            coef_xyt = 0.0_dp

            DO j = index_min(2), index_max(2)
               dj = j-gp(2)
               pointj = pointk+hmatgrid(:, 2)*dj
               IF (rsgrid%desc%perd(2) == 1) THEN
                  j_index = MODULO(j, ng(2))+1
               ELSE
                  j_index = j-index_min(2)+offset(2)
               ENDIF

               coef_xtt = 0.0_dp

               ! find bounds for the inner loop
               ! based on a quadratic equation in i
               ! a*i**2+b*i+c=radius**2
               v = pointj-gp(1)*hmatgrid(:, 1)
               a = DOT_PRODUCT(hmatgrid(:, 1), hmatgrid(:, 1))
               b = 2*DOT_PRODUCT(v, hmatgrid(:, 1))
               c = DOT_PRODUCT(v, v)
               d = b*b-4*a*(c-radius**2)

               IF (d < 0) THEN
                  CYCLE
               ELSE
                  d = SQRT(d)
                  ismin = CEILING((-b-d)/(2*a))
                  ismax = FLOOR((-b+d)/(2*a))
               ENDIF
               ! prepare for computing -zetp*rsq
               a = -zetp*a
               b = -zetp*b
               c = -zetp*c
               i = ismin-1
               exp2i = EXP((a*i+b)*i+c)
               exp1i = EXP(2*a*i+a+b)
               exp0i = EXP(2*a)

               coef_xtt = 0.0_dp

               DO i = ismin, ismax
                  di = i-gp(1)

                  exp2i = exp2i*exp1i
                  exp1i = exp1i*exp0i

                  i_index = grid_map(i)
                  gridval = grid(i_index, j_index, k_index)*exp2i

                  dip = 1.0_dp
                  DO il = 0, lp
                     coef_xtt(il) = coef_xtt(il)+gridval*dip
                     dip = dip*di
                  ENDDO
               ENDDO

               lxy = 0
               djp = 1.0_dp
               DO jl = 0, lp
                  DO il = 0, lp-jl
                     lxy = lxy+1
                     coef_xyt(lxy) = coef_xyt(lxy)+coef_xtt(il)*djp
                  ENDDO
                  djp = djp*dj
               ENDDO

            ENDDO

            lxyz = 0
            dkp = 1.0_dp
            DO kl = 0, lp
               lxy = 0
               DO jl = 0, lp-kl
                  DO il = 0, lp-kl-jl
                     lxyz = lxyz+1; lxy = lxy+1
                     coef_ijk(lxyz) = coef_ijk(lxyz)+dkp*coef_xyt(lxy)
                  ENDDO
                  lxy = lxy+kl
               ENDDO
               dkp = dkp*dk
            ENDDO

         ENDDO

         ! transform using multinomials
         ALLOCATE (hmatgridp(3, 3, 0:lp))
         hmatgridp(:, :, 0) = 1.0_dp
         DO k = 1, lp
            hmatgridp(:, :, k) = hmatgridp(:, :, k-1)*hmatgrid(:, :)
         ENDDO

         coef_xyz = 0.0_dp
         lpx = lp
         DO klx = 0, lpx
         DO jlx = 0, lpx-klx
         DO ilx = 0, lpx-klx-jlx
            lx = ilx+jlx+klx
            lpy = lp-lx
            DO kly = 0, lpy
            DO jly = 0, lpy-kly
            DO ily = 0, lpy-kly-jly
               ly = ily+jly+kly
               lpz = lp-lx-ly
               DO klz = 0, lpz
               DO jlz = 0, lpz-klz
               DO ilz = 0, lpz-klz-jlz
                  lz = ilz+jlz+klz

                  il = ilx+ily+ilz
                  jl = jlx+jly+jlz
                  kl = klx+kly+klz
                  coef_xyz(coef_map(lx, ly, lz)) = &
                     coef_xyz(coef_map(lx, ly, lz))+coef_ijk(coef_map(il, jl, kl))* &
                     hmatgridp(1, 1, ilx)*hmatgridp(1, 2, jlx)*hmatgridp(1, 3, klx)* &
                     hmatgridp(2, 1, ily)*hmatgridp(2, 2, jly)*hmatgridp(2, 3, kly)* &
                     hmatgridp(3, 1, ilz)*hmatgridp(3, 2, jlz)*hmatgridp(3, 3, klz)* &
                     fac(lx)*fac(ly)*fac(lz)/ &
                     (fac(ilx)*fac(ily)*fac(ilz)*fac(jlx)*fac(jly)*fac(jlz)*fac(klx)*fac(kly)*fac(klz))
               ENDDO
               ENDDO
               ENDDO
            ENDDO
            ENDDO
            ENDDO
         ENDDO
         ENDDO
         ENDDO

         CALL call_to_xyz_to_vab(prefactor, coef_xyz, lp, la_max_local, lb_max_local, &
                                 rp, ra, rab, vab, coset, la_min_local, &
                                 lb_min_local, current_maxl, ncoset(la_max_local), &
                                 ncoset(lb_max_local))

         ! deallocation needed to pass around a pgi bug..
         DEALLOCATE (hmatgridp)
         DEALLOCATE (grid_map)
         DEALLOCATE (coef_map)
         DEALLOCATE (coef_xtt)
         DEALLOCATE (coef_xyt)
         DEALLOCATE (coef_ijk)
         DEALLOCATE (coef_xyz)
      END SUBROUTINE integrate_general_opt

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE integrate_general_subpatch()
      INTEGER, DIMENSION(2, 3)                           :: local_b
      INTEGER, DIMENSION(3)                              :: local_s, periodic
      REAL(dp), DIMENSION((la_max_local+lb_max_local+1)*&
         (la_max_local+lb_max_local+2)*(&
         la_max_local+lb_max_local+3)/6)                 :: poly_d3

         periodic = 1 ! cell%perd
         lp = la_max_local+lb_max_local
         local_b(1, :) = rsgrid%lb_real-rsgrid%desc%lb
         local_b(2, :) = rsgrid%ub_real-rsgrid%desc%lb
         local_s = rsgrid%lb_real-rsgrid%lb_local
         IF (BTEST(subpatch_pattern, 0)) local_b(1, 1) = local_b(1, 1)-rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 1)) local_b(2, 1) = local_b(2, 1)+rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 2)) local_b(1, 2) = local_b(1, 2)-rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 3)) local_b(2, 2) = local_b(2, 2)+rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 4)) local_b(1, 3) = local_b(1, 3)-rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 5)) local_b(2, 3) = local_b(2, 3)+rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 0)) local_s(1) = local_s(1)-rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 2)) local_s(2) = local_s(2)-rsgrid%desc%border
         IF (BTEST(subpatch_pattern, 4)) local_s(3) = local_s(3)-rsgrid%desc%border
         CALL integrateGaussFull(h=cell%hmat, h_inv=cell%h_inv, &
                                 grid=grid, poly=poly_d3, alphai=zetp, posi=rp, max_r2=radius*radius, &
                                 periodic=periodic, gdim=ng, local_bounds=local_b, local_shift=local_s, &
                                 scale=rsgrid%desc%ngpts/ABS(cell%deth))
         ! defaults: local_shift=(/0,0,0/),poly_shift=(/0.0_dp,0.0_dp,0.0_dp/),scale=1.0_dp,
         ALLOCATE (coef_xyz(((lp+1)*(lp+2)*(lp+3))/6))
         CALL poly_d32cp2k(coef_xyz, lp, poly_d3)
         CALL call_to_xyz_to_vab(prefactor, coef_xyz, lp, la_max_local, lb_max_local, &
                                 rp, ra, rab, vab, coset, la_min_local, lb_min_local, &
                                 current_maxl, ncoset(la_max_local), &
                                 ncoset(lb_max_local))

         DEALLOCATE (coef_xyz)
      END SUBROUTINE

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE integrate_general_wings()
      INTEGER                                            :: periodic(3)
      INTEGER, DIMENSION(2, 3)                           :: local_b
      REAL(dp), DIMENSION((la_max_local+lb_max_local+1)*&
         (la_max_local+lb_max_local+2)*(&
         la_max_local+lb_max_local+3)/6)                 :: poly_d3
      REAL(dp), DIMENSION(3)                             :: local_shift, rShifted

         periodic = 1 ! cell%perd
         local_b(1, :) = 0
         local_b(2, :) = MIN(rsgrid%desc%npts-1, rsgrid%ub_local-rsgrid%lb_local)
         local_shift = REAL(rsgrid%desc%lb-rsgrid%lb_local, dp)/REAL(rsgrid%desc%npts, dp)
         rShifted(1) = rp(1)+cell%hmat(1, 1)*local_shift(1) &
                       +cell%hmat(1, 2)*local_shift(2) &
                       +cell%hmat(1, 3)*local_shift(3)
         rShifted(2) = rp(2)+cell%hmat(2, 1)*local_shift(1) &
                       +cell%hmat(2, 2)*local_shift(2) &
                       +cell%hmat(2, 3)*local_shift(3)
         rShifted(3) = rp(3)+cell%hmat(3, 1)*local_shift(1) &
                       +cell%hmat(3, 2)*local_shift(2) &
                       +cell%hmat(3, 3)*local_shift(3)
         lp = la_max_local+lb_max_local
         CALL integrateGaussFull(h=cell%hmat, h_inv=cell%h_inv, &
                                 grid=grid, poly=poly_d3, alphai=zetp, posi=rShifted, &
                                 max_r2=radius*radius, &
                                 periodic=periodic, gdim=ng, local_bounds=local_b, &
                                 scale=rsgrid%desc%ngpts/ABS(cell%deth))
         ! defaults: local_shift=(/0,0,0/),poly_shift=(/0.0_dp,0.0_dp,0.0_dp/),scale=1.0_dp,
         ALLOCATE (coef_xyz(((lp+1)*(lp+2)*(lp+3))/6))
         CALL poly_d32cp2k(coef_xyz, lp, poly_d3)
         CALL call_to_xyz_to_vab(prefactor, coef_xyz, lp, la_max_local, lb_max_local, &
                                 rp, ra, rab, vab, coset, la_min_local, lb_min_local, &
                                 current_maxl, ncoset(la_max_local), ncoset(lb_max_local))

         DEALLOCATE (coef_xyz)
      END SUBROUTINE

! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
      SUBROUTINE integrate_general()
      INTEGER                                            :: i, index_max(3), index_min(3), &
                                                            ipoint(3), j, k
      REAL(KIND=dp)                                      :: gridval, point(3)

         CALL return_cube_nonortho(cube_info, radius, index_min, index_max, rp)

         ! go over the grid, but cycle if the point is not within the radius
         DO k = index_min(3), index_max(3)
         DO j = index_min(2), index_max(2)
         DO i = index_min(1), index_max(1)
            ! point in real space
            point = MATMUL(cell%hmat, REAL((/i, j, k/), KIND=dp)/ng)
            ! skip if outside of the sphere
            IF (SUM((point-rp)**2) > radius**2) CYCLE
            ! point on the grid (including pbc)
            ipoint = MODULO((/i, j, k/), ng)+1
            ! integrate on the grid
            gridval = grid(ipoint(1), ipoint(2), ipoint(3))
            CALL primitive_integrate(point, gridval)
         ENDDO
         ENDDO
         ENDDO
      END SUBROUTINE integrate_general

! **************************************************************************************************
!> \brief ...
!> \param point ...
!> \param gridval ...
! **************************************************************************************************
      SUBROUTINE primitive_integrate(point, gridval)
      REAL(KIND=dp)                                      :: point(3), gridval

      REAL(KIND=dp)                                      :: dra(3), drap(3), drb(3), drbp(3), myexp

         myexp = EXP(-zetp*SUM((point-rp)**2))*prefactor*gridval
         dra = point-ra
         drb = point-rb
         drap(1) = 1.0_dp
         DO lxa = 0, la_max_local
            drbp(1) = 1.0_dp
            DO lxb = 0, lb_max_local
               drap(2) = 1.0_dp
               DO lya = 0, la_max_local-lxa
                  drbp(2) = 1.0_dp
                  DO lyb = 0, lb_max_local-lxb
                     drap(3) = 1.0_dp
                     DO lza = 1, MAX(la_min_local-lxa-lya, 0)
                        drap(3) = drap(3)*dra(3)
                     ENDDO
                     DO lza = MAX(la_min_local-lxa-lya, 0), la_max_local-lxa-lya
                        drbp(3) = 1.0_dp
                        DO lzb = 1, MAX(lb_min_local-lxb-lyb, 0)
                           drbp(3) = drbp(3)*drb(3)
                        ENDDO
                        DO lzb = MAX(lb_min_local-lxb-lyb, 0), lb_max_local-lxb-lyb
                           ico = coset(lxa, lya, lza)
                           jco = coset(lxb, lyb, lzb)
                           vab(ico, jco) = vab(ico, jco)+myexp*PRODUCT(drap)*PRODUCT(drbp)
                           drbp(3) = drbp(3)*drb(3)
                        ENDDO
                        drap(3) = drap(3)*dra(3)
                     ENDDO
                     drbp(2) = drbp(2)*drb(2)
                  ENDDO
                  drap(2) = drap(2)*dra(2)
               ENDDO
               drbp(1) = drbp(1)*drb(1)
            ENDDO
            drap(1) = drap(1)*dra(1)
         ENDDO

      END SUBROUTINE

   END SUBROUTINE integrate_pgf_product_rspace

! **************************************************************************************************
!> \brief given a set of matrix elements, perform the correct contraction to obtain the virial
!> \param my_virial_a ...
!> \param my_virial_b ...
!> \param rab ...
!> \param pab ...
!> \param ftza ...
!> \param ftzb ...
!> \param ax ...
!> \param ay ...
!> \param az ...
!> \param bx ...
!> \param by ...
!> \param bz ...
!> \param vab ...
! **************************************************************************************************
   SUBROUTINE virial_update(my_virial_a, my_virial_b, rab, pab, &
                            ftza, ftzb, ax, ay, az, bx, by, bz, vab)
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(INOUT)      :: my_virial_a, my_virial_b
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), INTENT(IN)                          :: pab, ftza, ftzb
      INTEGER, INTENT(IN)                                :: ax, ay, az, bx, by, bz
      REAL(KIND=dp)                                      :: vab(:, :)

      my_virial_a(1, 1) = my_virial_a(1, 1) &
                          +pab*ftza*vab(coset(ax+2, ay, az), coset(bx, by, bz)) &
                          -pab*REAL(ax, dp)*vab(coset(MAX(0, ax-1)+1, ay, az), coset(bx, by, bz))
      my_virial_a(1, 2) = my_virial_a(1, 2) &
                          +pab*ftza*vab(coset(ax+1, ay+1, az), coset(bx, by, bz)) &
                          -pab*REAL(ax, dp)*vab(coset(MAX(0, ax-1), ay+1, az), coset(bx, by, bz))
      my_virial_a(1, 3) = my_virial_a(1, 3) &
                          +pab*ftza*vab(coset(ax+1, ay, az+1), coset(bx, by, bz)) &
                          -pab*REAL(ax, dp)*vab(coset(MAX(0, ax-1), ay, az+1), coset(bx, by, bz))
      my_virial_a(2, 1) = my_virial_a(2, 1) &
                          +pab*ftza*vab(coset(ax+1, ay+1, az), coset(bx, by, bz)) &
                          -pab*REAL(ay, dp)*vab(coset(ax+1, MAX(0, ay-1), az), coset(bx, by, bz))
      my_virial_a(2, 2) = my_virial_a(2, 2) &
                          +pab*ftza*vab(coset(ax, ay+2, az), coset(bx, by, bz)) &
                          -pab*REAL(ay, dp)*vab(coset(ax, MAX(0, ay-1)+1, az), coset(bx, by, bz))
      my_virial_a(2, 3) = my_virial_a(2, 3) &
                          +pab*ftza*vab(coset(ax, ay+1, az+1), coset(bx, by, bz)) &
                          -pab*REAL(ay, dp)*vab(coset(ax, MAX(0, ay-1), az+1), coset(bx, by, bz))
      my_virial_a(3, 1) = my_virial_a(3, 1) &
                          +pab*ftza*vab(coset(ax+1, ay, az+1), coset(bx, by, bz)) &
                          -pab*REAL(az, dp)*vab(coset(ax+1, ay, MAX(0, az-1)), coset(bx, by, bz))
      my_virial_a(3, 2) = my_virial_a(3, 2) &
                          +pab*ftza*vab(coset(ax, ay+1, az+1), coset(bx, by, bz)) &
                          -pab*REAL(az, dp)*vab(coset(ax, ay+1, MAX(0, az-1)), coset(bx, by, bz))
      my_virial_a(3, 3) = my_virial_a(3, 3) &
                          +pab*ftza*vab(coset(ax, ay, az+2), coset(bx, by, bz)) &
                          -pab*REAL(az, dp)*vab(coset(ax, ay, MAX(0, az-1)+1), coset(bx, by, bz))

      my_virial_b(1, 1) = my_virial_b(1, 1)+pab*ftzb*( &
                          vab(coset(ax+2, ay, az), coset(bx, by, bz)) &
                          -vab(coset(ax+1, ay, az), coset(bx, by, bz))*rab(1) &
                          -vab(coset(ax+1, ay, az), coset(bx, by, bz))*rab(1) &
                          +vab(coset(ax, ay, az), coset(bx, by, bz))*rab(1)*rab(1)) &
                          -pab*REAL(bx, dp)*vab(coset(ax, ay, az), coset(MAX(0, bx-1)+1, by, bz))
      my_virial_b(1, 2) = my_virial_b(1, 2)+pab*ftzb*( &
                          vab(coset(ax+1, ay+1, az), coset(bx, by, bz)) &
                          -vab(coset(ax, ay+1, az), coset(bx, by, bz))*rab(1) &
                          -vab(coset(ax+1, ay, az), coset(bx, by, bz))*rab(2) &
                          +vab(coset(ax, ay, az), coset(bx, by, bz))*rab(1)*rab(2)) &
                          -pab*REAL(bx, dp)*vab(coset(ax, ay, az), coset(MAX(0, bx-1), by+1, bz))
      my_virial_b(1, 3) = my_virial_b(1, 3)+pab*ftzb*( &
                          vab(coset(ax+1, ay, az+1), coset(bx, by, bz)) &
                          -vab(coset(ax, ay, az+1), coset(bx, by, bz))*rab(1) &
                          -vab(coset(ax+1, ay, az), coset(bx, by, bz))*rab(3) &
                          +vab(coset(ax, ay, az), coset(bx, by, bz))*rab(1)*rab(3)) &
                          -pab*REAL(bx, dp)*vab(coset(ax, ay, az), coset(MAX(0, bx-1), by, bz+1))
      my_virial_b(2, 1) = my_virial_b(2, 1)+pab*ftzb*( &
                          vab(coset(ax+1, ay+1, az), coset(bx, by, bz)) &
                          -vab(coset(ax+1, ay, az), coset(bx, by, bz))*rab(2) &
                          -vab(coset(ax, ay+1, az), coset(bx, by, bz))*rab(1) &
                          +vab(coset(ax, ay, az), coset(bx, by, bz))*rab(2)*rab(1)) &
                          -pab*REAL(by, dp)*vab(coset(ax, ay, az), coset(bx+1, MAX(0, by-1), bz))
      my_virial_b(2, 2) = my_virial_b(2, 2)+pab*ftzb*( &
                          vab(coset(ax, ay+2, az), coset(bx, by, bz)) &
                          -vab(coset(ax, ay+1, az), coset(bx, by, bz))*rab(2) &
                          -vab(coset(ax, ay+1, az), coset(bx, by, bz))*rab(2) &
                          +vab(coset(ax, ay, az), coset(bx, by, bz))*rab(2)*rab(2)) &
                          -pab*REAL(by, dp)*vab(coset(ax, ay, az), coset(bx, MAX(0, by-1)+1, bz))
      my_virial_b(2, 3) = my_virial_b(2, 3)+pab*ftzb*( &
                          vab(coset(ax, ay+1, az+1), coset(bx, by, bz)) &
                          -vab(coset(ax, ay, az+1), coset(bx, by, bz))*rab(2) &
                          -vab(coset(ax, ay+1, az), coset(bx, by, bz))*rab(3) &
                          +vab(coset(ax, ay, az), coset(bx, by, bz))*rab(2)*rab(3)) &
                          -pab*REAL(by, dp)*vab(coset(ax, ay, az), coset(bx, MAX(0, by-1), bz+1))
      my_virial_b(3, 1) = my_virial_b(3, 1)+pab*ftzb*( &
                          vab(coset(ax+1, ay, az+1), coset(bx, by, bz)) &
                          -vab(coset(ax+1, ay, az), coset(bx, by, bz))*rab(3) &
                          -vab(coset(ax, ay, az+1), coset(bx, by, bz))*rab(1) &
                          +vab(coset(ax, ay, az), coset(bx, by, bz))*rab(3)*rab(1)) &
                          -pab*REAL(bz, dp)*vab(coset(ax, ay, az), coset(bx+1, by, MAX(0, bz-1)))
      my_virial_b(3, 2) = my_virial_b(3, 2)+pab*ftzb*( &
                          vab(coset(ax, ay+1, az+1), coset(bx, by, bz)) &
                          -vab(coset(ax, ay+1, az), coset(bx, by, bz))*rab(3) &
                          -vab(coset(ax, ay, az+1), coset(bx, by, bz))*rab(2) &
                          +vab(coset(ax, ay, az), coset(bx, by, bz))*rab(3)*rab(2)) &
                          -pab*REAL(bz, dp)*vab(coset(ax, ay, az), coset(bx, by+1, MAX(0, bz-1)))
      my_virial_b(3, 3) = my_virial_b(3, 3)+pab*ftzb*( &
                          vab(coset(ax, ay, az+2), coset(bx, by, bz)) &
                          -vab(coset(ax, ay, az+1), coset(bx, by, bz))*rab(3) &
                          -vab(coset(ax, ay, az+1), coset(bx, by, bz))*rab(3) &
                          +vab(coset(ax, ay, az), coset(bx, by, bz))*rab(3)*rab(3)) &
                          -pab*REAL(bz, dp)*vab(coset(ax, ay, az), coset(bx, by, MAX(0, bz-1)+1))

   END SUBROUTINE virial_update

! **************************************************************************************************
!> \brief given a bunch of matrix elements, performe the right contractions to obtain the forces
!> \param force_a ...
!> \param force_b ...
!> \param rab ...
!> \param pab ...
!> \param ftza ...
!> \param ftzb ...
!> \param ax ...
!> \param ay ...
!> \param az ...
!> \param bx ...
!> \param by ...
!> \param bz ...
!> \param vab ...
! **************************************************************************************************
   SUBROUTINE force_update(force_a, force_b, rab, pab, ftza, ftzb, ax, ay, az, bx, by, bz, vab)
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: force_a, force_b
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), INTENT(IN)                          :: pab, ftza, ftzb
      INTEGER, INTENT(IN)                                :: ax, ay, az, bx, by, bz
      REAL(KIND=dp)                                      :: vab(:, :)

      REAL(KIND=dp)                                      :: axm1, axp1, axpm0, aym1, ayp1, azm1, &
                                                            azp1, bxm1, bym1, bzm1

      axpm0 = vab(coset(ax, ay, az), coset(bx, by, bz))
      axp1 = vab(coset(ax+1, ay, az), coset(bx, by, bz))
      axm1 = vab(coset(MAX(0, ax-1), ay, az), coset(bx, by, bz))
      ayp1 = vab(coset(ax, ay+1, az), coset(bx, by, bz))
      aym1 = vab(coset(ax, MAX(0, ay-1), az), coset(bx, by, bz))
      azp1 = vab(coset(ax, ay, az+1), coset(bx, by, bz))
      azm1 = vab(coset(ax, ay, MAX(0, az-1)), coset(bx, by, bz))
      bxm1 = vab(coset(ax, ay, az), coset(MAX(0, bx-1), by, bz))
      bym1 = vab(coset(ax, ay, az), coset(bx, MAX(0, by-1), bz))
      bzm1 = vab(coset(ax, ay, az), coset(bx, by, MAX(0, bz-1)))
      force_a(1) = force_a(1)+pab*(ftza*axp1-REAL(ax, dp)*axm1)
      force_a(2) = force_a(2)+pab*(ftza*ayp1-REAL(ay, dp)*aym1)
      force_a(3) = force_a(3)+pab*(ftza*azp1-REAL(az, dp)*azm1)
      force_b(1) = force_b(1)+pab*(ftzb*(axp1-rab(1)*axpm0)-REAL(bx, dp)*bxm1)
      force_b(2) = force_b(2)+pab*(ftzb*(ayp1-rab(2)*axpm0)-REAL(by, dp)*bym1)
      force_b(3) = force_b(3)+pab*(ftzb*(azp1-rab(3)*axpm0)-REAL(bz, dp)*bzm1)

   END SUBROUTINE force_update

! **************************************************************************************************
!> \brief given a bunch of matrix elements perform the right contractions to obtain the
!>      derivatives of the hab matirx
!> \param der_a ...
!> \param der_b ...
!> \param rab ...
!> \param ftza ...
!> \param ftzb ...
!> \param ax ...
!> \param ay ...
!> \param az ...
!> \param bx ...
!> \param by ...
!> \param bz ...
!> \param vab ...
! **************************************************************************************************
   SUBROUTINE hab_derivatives(der_a, der_b, rab, ftza, ftzb, ax, ay, az, bx, by, bz, vab)
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: der_a, der_b
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), INTENT(IN)                          :: ftza, ftzb
      INTEGER, INTENT(IN)                                :: ax, ay, az, bx, by, bz
      REAL(KIND=dp)                                      :: vab(:, :)

      REAL(KIND=dp)                                      :: axm1, axp1, axpm0, aym1, ayp1, azm1, &
                                                            azp1, bxm1, bym1, bzm1

      axpm0 = vab(coset(ax, ay, az), coset(bx, by, bz))
      axp1 = vab(coset(ax+1, ay, az), coset(bx, by, bz))
      axm1 = vab(coset(MAX(0, ax-1), ay, az), coset(bx, by, bz))
      ayp1 = vab(coset(ax, ay+1, az), coset(bx, by, bz))
      aym1 = vab(coset(ax, MAX(0, ay-1), az), coset(bx, by, bz))
      azp1 = vab(coset(ax, ay, az+1), coset(bx, by, bz))
      azm1 = vab(coset(ax, ay, MAX(0, az-1)), coset(bx, by, bz))
      bxm1 = vab(coset(ax, ay, az), coset(MAX(0, bx-1), by, bz))
      bym1 = vab(coset(ax, ay, az), coset(bx, MAX(0, by-1), bz))
      bzm1 = vab(coset(ax, ay, az), coset(bx, by, MAX(0, bz-1)))
      der_a(1) = (ftza*axp1-REAL(ax, dp)*axm1)
      der_a(2) = (ftza*ayp1-REAL(ay, dp)*aym1)
      der_a(3) = (ftza*azp1-REAL(az, dp)*azm1)
      der_b(1) = (ftzb*(axp1-rab(1)*axpm0)-REAL(bx, dp)*bxm1)
      der_b(2) = (ftzb*(ayp1-rab(2)*axpm0)-REAL(by, dp)*bym1)
      der_b(3) = (ftzb*(azp1-rab(3)*axpm0)-REAL(bz, dp)*bzm1)

   END SUBROUTINE hab_derivatives

END MODULE qs_integrate_potential_low
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Build up the plane wave density by collocating the primitive Gaussian
!>      functions (pgf).
!> \par History
!>      Joost VandeVondele (02.2002)
!>            1) rewrote collocate_pgf for increased accuracy and speed
!>            2) collocate_core hack for PGI compiler
!>            3) added multiple grid feature
!>            4) new way to go over the grid
!>      Joost VandeVondele (05.2002)
!>            1) prelim. introduction of the real space grid type
!>      JGH [30.08.02] multigrid arrays independent from potential
!>      JGH [17.07.03] distributed real space code
!>      JGH [23.11.03] refactoring and new loop ordering
!>      JGH [04.12.03] OpneMP parallelization of main loops
!>      Joost VandeVondele (12.2003)
!>           1) modified to compute tau
!>      Joost removed incremental build feature
!>      Joost introduced map consistent
!>      Rewrote grid integration/collocation routines, [Joost VandeVondele,03.2007]
!>      JGH [26.06.15] modification to allow for k-points
!> \author Matthias Krack (03.04.2001)
! **************************************************************************************************
MODULE qs_integrate_potential_product
   USE admm_types,                      ONLY: admm_type
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add_block_node, cp_dbcsr_col_block_sizes, cp_dbcsr_copy, cp_dbcsr_create, &
        cp_dbcsr_deallocate_matrix_set, cp_dbcsr_distribution, cp_dbcsr_finalize, &
        cp_dbcsr_get_block_p, cp_dbcsr_get_data_size, cp_dbcsr_get_matrix_type, cp_dbcsr_init, &
        cp_dbcsr_p_type, cp_dbcsr_row_block_sizes, cp_dbcsr_type, cp_dbcsr_work_create, &
        dbcsr_distribution_has_threads
   USE cube_utils,                      ONLY: cube_info_type
   USE gaussian_gridlevels,             ONLY: gridlevel_info_type
   USE input_constants,                 ONLY: do_admm_exch_scaling_merlot
   USE kinds,                           ONLY: default_string_length,&
                                              dp,&
                                              int_8
   USE memory_utilities,                ONLY: reallocate
   USE orbital_pointers,                ONLY: ncoset
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_types,                        ONLY: pw_p_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_integrate_potential_low,      ONLY: integrate_pgf_product_rspace
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE realspace_grid_types,            ONLY: realspace_grid_desc_p_type,&
                                              realspace_grid_p_type,&
                                              rs_grid_release,&
                                              rs_grid_retain
   USE rs_pw_interface,                 ONLY: potential_pw2rs
   USE task_list_methods,               ONLY: int2pair,&
                                              rs_distribute_matrix
   USE task_list_types,                 ONLY: task_list_type
   USE virial_types,                    ONLY: virial_type

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   INTEGER :: debug_count = 0

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_integrate_potential_product'

! *** Public subroutines ***
! *** Don't include this routines directly, use the interface to
! *** qs_integrate_potential

   PUBLIC :: integrate_v_rspace

CONTAINS

! **************************************************************************************************
!> \brief computes matrix elements corresponding to a given potential
!> \param v_rspace ...
!> \param hmat ...
!> \param hmat_kp ...
!> \param pmat ...
!> \param pmat_kp ...
!> \param qs_env ...
!> \param calculate_forces ...
!> \param force_adm ... whether force of in aux. dens. matrix is calculated
!> \param ispin ...
!> \param compute_tau ...
!> \param gapw ...
!> \param basis_type ...
!> \param pw_env_external ...
!> \param task_list_external ...
!> \par History
!>      IAB (29-Apr-2010): Added OpenMP parallelisation to task loop
!>                         (c) The Numerical Algorithms Group (NAG) Ltd, 2010 on behalf of the HECToR project
!>      Some refactoring, get priorities for options correct (JGH, 04.2014)
!>      Added options to allow for k-points
!>      For a smooth transition we allow for old and new (vector) matrices (hmat, pmat) (JGH, 06.2015)
!> \note
!>     integrates a given potential (or other object on a real
!>     space grid) = v_rspace using a multi grid technique (mgrid_*)
!>     over the basis set producing a number for every element of h
!>     (should have the same sparsity structure of S)
!>     additional screening is available using the magnitude of the
!>     elements in p (? I'm not sure this is a very good idea)
!>     this argument is optional
!>     derivatives of these matrix elements with respect to the ionic
!>     coordinates can be computed as well
! **************************************************************************************************
   SUBROUTINE integrate_v_rspace(v_rspace, hmat, hmat_kp, pmat, pmat_kp, &
                                 qs_env, calculate_forces, force_adm, ispin, &
                                 compute_tau, gapw, basis_type, pw_env_external, task_list_external)

      TYPE(pw_p_type)                                    :: v_rspace
      TYPE(cp_dbcsr_p_type), INTENT(INOUT), OPTIONAL     :: hmat
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: hmat_kp
      TYPE(cp_dbcsr_p_type), INTENT(IN), OPTIONAL        :: pmat
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: pmat_kp
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: calculate_forces
      LOGICAL, INTENT(IN), OPTIONAL                      :: force_adm
      INTEGER, INTENT(IN), OPTIONAL                      :: ispin
      LOGICAL, INTENT(IN), OPTIONAL                      :: compute_tau, gapw
      CHARACTER(len=*), INTENT(IN), OPTIONAL             :: basis_type
      TYPE(pw_env_type), OPTIONAL, POINTER               :: pw_env_external
      TYPE(task_list_type), OPTIONAL, POINTER            :: task_list_external

      CHARACTER(len=*), PARAMETER :: routineN = 'integrate_v_rspace', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=default_string_length)               :: my_basis_type
      INTEGER :: atom_a, atom_b, bcol, brow, handle, i, iatom, igrid_level, ikind, ikind_old, &
         ilevel, img, ipair, ipgf, ipgf_new, iset, iset_new, iset_old, itask, ithread, jatom, &
         jkind, jkind_old, jpgf, jpgf_new, jset, jset_new, jset_old, maxco, maxpgf, maxset, &
         maxsgf_set, na1, na2, natom, nb1, nb2, ncoa, ncob, nimages, nkind, nseta, nsetb, nthread, &
         offs_dv, sgfa, sgfb
      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: atom_pair_recv, atom_pair_send
      INTEGER(kind=int_8), DIMENSION(:, :), POINTER      :: tasks
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: block_touched
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, lb_max, lb_min, npgfa, &
                                                            npgfb, nsgfa, nsgfb
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, first_sgfb
      LOGICAL :: atom_pair_changed, atom_pair_done, distributed_grids, do_kp, found, h_duplicated, &
         map_consistent, my_compute_tau, my_force_adm, my_gapw, new_set_pair_coming, p_duplicated, &
         pab_required, scatter, use_subpatch, use_virial
      REAL(KIND=dp)                                      :: admm_scal_fac, dab, eps_gvg_rspace, &
                                                            rab2, scalef, zetp
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, force_b, ra, rab, rab_inv, rb
      REAL(KIND=dp), DIMENSION(3, 3)                     :: my_virial_a, my_virial_b
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a, set_radius_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dist_ab, h_block, hab, p_block, pab, &
                                                            rpgfa, rpgfb, sphi_a, sphi_b, work, &
                                                            zeta, zetb
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: habt, hadb, hdab, pabt, workt
      REAL(kind=dp), DIMENSION(:, :, :, :), POINTER      :: hadbt, hdabt
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: deltap, dhmat, htemp
      TYPE(cp_dbcsr_type), POINTER                       :: href
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_v
      TYPE(task_list_type), POINTER                      :: task_list, task_list_soft
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      ! we test here if the provided operator matrices are consistent
      IF(.NOT.(PRESENT(hmat) .OR. PRESENT(hmat_kp)))CALL cp__a("qs_integrate_potential_product.F",199)
      do_kp = .FALSE.
      IF (PRESENT(hmat_kp)) do_kp = .TRUE.
      IF (PRESENT(pmat)) THEN
         IF(.NOT.(PRESENT(hmat)))CALL cp__a("qs_integrate_potential_product.F",203)
      ELSE IF (PRESENT(pmat_kp)) THEN
         IF(.NOT.(PRESENT(hmat_kp)))CALL cp__a("qs_integrate_potential_product.F",205)
      END IF

      NULLIFY (pw_env, rs_descs, tasks, dist_ab, admm_env)

      debug_count = debug_count+1

      offs_dv = 0

      ! this routine works in two modes:
      ! normal mode : <a| V | b>
      ! tau mode    : < nabla a| V | nabla b>
      my_compute_tau = .FALSE.
      IF (PRESENT(compute_tau)) my_compute_tau = compute_tau

      my_force_adm = .FALSE.
      IF (PRESENT(force_adm)) my_force_adm = force_adm

      ! this sets the basis set to be used. GAPW(==soft basis) overwrites basis_type
      ! default is "ORB"
      my_gapw = .FALSE.
      IF (PRESENT(gapw)) my_gapw = gapw
      IF (PRESENT(basis_type)) THEN
         my_basis_type = basis_type
      ELSE
         my_basis_type = "ORB"
      END IF

      ! get the task lists
      ! task lists have to be in sync with basis sets
      ! there is an option to provide the task list from outside (not through qs_env)
      ! outside option has highest priority
      SELECT CASE (my_basis_type)
      CASE ("ORB")
         CALL get_qs_env(qs_env=qs_env, &
                         task_list=task_list, &
                         task_list_soft=task_list_soft)
      CASE ("AUX_FIT")
         CALL get_qs_env(qs_env=qs_env, &
                         task_list_aux_fit=task_list, &
                         task_list_soft=task_list_soft)
      END SELECT
      IF (my_gapw) task_list => task_list_soft
      IF (PRESENT(task_list_external)) task_list => task_list_external
      IF(.NOT.(ASSOCIATED(task_list)))CALL cp__a("qs_integrate_potential_product.F",249)

      ! the information on the grids is provided through pw_env
      ! pw_env has to be the parent env for the potential grid (input)
      ! there is an option to provide an external grid
      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
      IF (PRESENT(pw_env_external)) pw_env => pw_env_external

      ! get all the general information on the system we are working on
      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      force=force, &
                      virial=virial)

      admm_scal_fac = 1.0_dp
      IF (my_force_adm) THEN
         CALL get_qs_env(qs_env=qs_env, admm_env=admm_env)
         ! Calculate bare scaling of force according to Merlot, 1. IF: ADMMP, 2. IF: ADMMS,
         IF ((.NOT. admm_env%charge_constrain) .AND. &
             (admm_env%scaling_model == do_admm_exch_scaling_merlot)) THEN
            admm_scal_fac = admm_env%gsi(ispin)**2
         ELSE IF (admm_env%charge_constrain .AND. &
                  (admm_env%scaling_model == do_admm_exch_scaling_merlot)) THEN
            admm_scal_fac = (admm_env%gsi(ispin))**(2.0_dp/3.0_dp)
         END IF
      END IF

      ! short cuts to task list variables
      tasks => task_list%tasks
      dist_ab => task_list%dist_ab
      atom_pair_send => task_list%atom_pair_send
      atom_pair_recv => task_list%atom_pair_recv

      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_integrate_potential_product.F",286)
      CALL pw_env_get(pw_env, rs_descs=rs_descs, rs_grids=rs_v)
      DO i = 1, SIZE(rs_v)
         CALL rs_grid_retain(rs_v(i)%rs_grid)
      END DO

      ! assign from pw_env
      gridlevel_info => pw_env%gridlevel_info
      cube_info => pw_env%cube_info

      ! transform the potential on the rs_multigrids
      CALL potential_pw2rs(rs_v, v_rspace, pw_env)

      nimages = dft_control%nimages
      IF (nimages > 1) THEN
         IF(.NOT.(do_kp))CALL cp__a("qs_integrate_potential_product.F",301)
      END IF
      nkind = SIZE(qs_kind_set)
      natom = SIZE(particle_set)
      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      IF (calculate_forces) THEN
         ALLOCATE (atom_of_kind(natom))
         CALL get_atomic_kind_set(atomic_kind_set, atom_of_kind=atom_of_kind)
      END IF

      map_consistent = dft_control%qs_control%map_consistent
      IF (map_consistent) THEN
         ! needs to be consistent with rho_rspace
         eps_gvg_rspace = dft_control%qs_control%eps_rho_rspace
      ELSE
         eps_gvg_rspace = dft_control%qs_control%eps_gvg_rspace
      ENDIF

      pab_required = (PRESENT(pmat) .OR. PRESENT(pmat_kp)) &
                     .AND. (calculate_forces .OR. .NOT. map_consistent)

      CALL get_qs_kind_set(qs_kind_set=qs_kind_set, &
                           maxco=maxco, &
                           maxsgf_set=maxsgf_set, &
                           basis_type=my_basis_type)

      distributed_grids = .FALSE.
      DO igrid_level = 1, gridlevel_info%ngrid_levels
         IF (rs_v(igrid_level)%rs_grid%desc%distributed) THEN
            distributed_grids = .TRUE.
         ENDIF
      ENDDO

      ! initialize the working hmat structures
      h_duplicated = .FALSE.
      ALLOCATE (dhmat(nimages))
      IF (do_kp) THEN
         DO img = 1, nimages
            dhmat(img)%matrix => hmat_kp(img)%matrix
         END DO
      ELSE
         dhmat(1)%matrix => hmat%matrix
      END IF
      IF (distributed_grids) THEN
         h_duplicated = .TRUE.
         href => dhmat(1)%matrix
         DO img = 1, nimages
            NULLIFY (dhmat(img)%matrix)
            ALLOCATE (dhmat(img)%matrix)
            CALL cp_dbcsr_init(dhmat(img)%matrix)
            CALL cp_dbcsr_create(dhmat(img)%matrix, 'LocalH', &
                                 cp_dbcsr_distribution(href), &
                                 cp_dbcsr_get_matrix_type(href), cp_dbcsr_row_block_sizes(href), &
                                 cp_dbcsr_col_block_sizes(href), cp_dbcsr_get_data_size(href))
         END DO
      END IF

      p_duplicated = .FALSE.
      IF (pab_required) THEN
         ! initialize the working pmat structures
         ALLOCATE (deltap(nimages))
         IF (do_kp) THEN
            DO img = 1, nimages
               deltap(img)%matrix => pmat_kp(img)%matrix
            END DO
         ELSE
            deltap(1)%matrix => pmat%matrix
         END IF
         IF (distributed_grids) THEN
            p_duplicated = .TRUE.
            DO img = 1, nimages
               NULLIFY (deltap(img)%matrix)
               ALLOCATE (deltap(img)%matrix)
               CALL cp_dbcsr_init(deltap(img)%matrix)
            END DO
            IF (do_kp) THEN
               DO img = 1, nimages
                  CALL cp_dbcsr_copy(deltap(img)%matrix, pmat_kp(img)%matrix, name="LocalP")
               END DO
            ELSE
               CALL cp_dbcsr_copy(deltap(1)%matrix, pmat%matrix, name="LocalP")
            END IF
         END IF
      END IF

      nthread = 1
!$    nthread = omp_get_max_threads()

      !   *** Allocate work storage ***
      NULLIFY (pabt, habt, workt)
      CALL reallocate(habt, 1, maxco, 1, maxco, 0, nthread)
      CALL reallocate(workt, 1, maxco, 1, maxsgf_set, 0, nthread)
      IF (pab_required) THEN
         CALL reallocate(pabt, 1, maxco, 1, maxco, 0, nthread)
      ELSE
         IF(.NOT.(.NOT. calculate_forces))CALL cp__a("qs_integrate_potential_product.F",397)
      END IF

      NULLIFY (hdabt, hadbt, hdab, hadb)

      !   get maximum numbers
      natom = SIZE(particle_set)
      maxset = 0
      maxpgf = 0
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), &
                          softb=my_gapw, &
                          basis_set=orb_basis_set, basis_type=my_basis_type)

         IF (.NOT. ASSOCIATED(orb_basis_set)) CYCLE

         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                npgf=npgfa, nset=nseta)

         maxset = MAX(nseta, maxset)
         maxpgf = MAX(MAXVAL(npgfa), maxpgf)
      END DO

      IF (distributed_grids .AND. pab_required) THEN
         CALL rs_distribute_matrix(rs_descs, deltap, atom_pair_send, atom_pair_recv, &
                                   natom, nimages, scatter=.TRUE.)
      ENDIF

      IF (debug_this_module) THEN
         ALLOCATE (block_touched(natom, natom))
      END IF

!$OMP PARALLEL DEFAULT(NONE), &
!$OMP SHARED(workt,habt,hdabt,hadbt,pabt,tasks,particle_set,natom,maxset), &
!$OMP SHARED(maxpgf,my_basis_type,my_gapw,dhmat,deltap,use_virial,admm_scal_fac), &
!$OMP SHARED(pab_required,calculate_forces,ncoset,rs_v,cube_info,my_compute_tau), &
!$OMP SHARED(map_consistent,eps_gvg_rspace,force,virial,cell,atom_of_kind,dist_ab), &
!$OMP SHARED(gridlevel_info,task_list,block_touched,nthread,qs_kind_set), &
!$OMP SHARED(nimages,do_kp), &
!$OMP PRIVATE(ithread,work,hab,hdab,hadb,pab,iset_old,jset_old), &
!$OMP PRIVATE(ikind_old,jkind_old,iatom,jatom,iset,jset,ikind,jkind,ilevel,ipgf,jpgf), &
!$OMP PRIVATE(img,brow,bcol,orb_basis_set,first_sgfa,la_max,la_min,npgfa,nseta,nsgfa), &
!$OMP PRIVATE(rpgfa,set_radius_a,sphi_a,zeta,first_sgfb,lb_max,lb_min,npgfb), &
!$OMP PRIVATE(nsetb,nsgfb,rpgfb,set_radius_b,sphi_b,zetb,found,atom_a,atom_b), &
!$OMP PRIVATE(force_a,force_b,my_virial_a,my_virial_b,atom_pair_changed,h_block), &
!$OMP PRIVATE(p_block,ncoa,sgfa,ncob,sgfb,rab,rab2,ra,rb,zetp,dab,igrid_level), &
!$OMP PRIVATE(na1,na2,nb1,nb2,use_subpatch,rab_inv,new_set_pair_coming,atom_pair_done), &
!$OMP PRIVATE(iset_new,jset_new,ipgf_new,jpgf_new,scalef), &
!$OMP PRIVATE(itask)

      ithread = 0
!$    ithread = omp_get_thread_num()
      work => workt(:, :, ithread)
      hab => habt(:, :, ithread)
      IF (pab_required) THEN
         pab => pabt(:, :, ithread)
      END IF

      iset_old = -1; jset_old = -1
      ikind_old = -1; jkind_old = -1

      ! Here we loop over gridlevels first, finalising the matrix after each grid level is
      ! completed.  On each grid level, we loop over atom pairs, which will only access
      ! a single block of each matrix, so with OpenMP, each matrix block is only touched
      ! by a single thread for each grid level
      loop_gridlevels: DO igrid_level = 1, gridlevel_info%ngrid_levels

         DO img = 1, nimages
            CALL cp_dbcsr_work_create(dhmat(img)%matrix, work_mutable=.TRUE., n=nthread)
!$          IF (.NOT. dbcsr_distribution_has_threads(cp_dbcsr_distribution(dhmat(img)%matrix))) &
!$             CALL cp__b("qs_integrate_potential_product.F",467,"No thread distribution defined.")
         END DO
!$OMP BARRIER

         IF (debug_this_module) THEN
!$OMP SINGLE
            block_touched = -1
!$OMP END SINGLE
!$OMP FLUSH
         END IF

!$OMP DO schedule (dynamic, MAX(1,task_list%npairs(igrid_level)/(nthread*50)))
         loop_pairs: DO ipair = 1, task_list%npairs(igrid_level)
         loop_tasks: DO itask = task_list%taskstart(ipair, igrid_level), task_list%taskstop(ipair, igrid_level)

            CALL int2pair(tasks(3, itask), ilevel, img, iatom, jatom, iset, jset, ipgf, jpgf, &
                          nimages, natom, maxset, maxpgf)
            IF(.NOT.(img == 1 .OR. do_kp))CALL cp__a("qs_integrate_potential_product.F",484)

            ! At the start of a block of tasks, get atom data (and kind data, if needed)
            IF (itask .EQ. task_list%taskstart(ipair, igrid_level)) THEN

               ikind = particle_set(iatom)%atomic_kind%kind_number
               jkind = particle_set(jatom)%atomic_kind%kind_number

               ra(:) = pbc(particle_set(iatom)%r, cell)

               IF (iatom <= jatom) THEN
                  brow = iatom
                  bcol = jatom
               ELSE
                  brow = jatom
                  bcol = iatom
               END IF

               IF (ikind .NE. ikind_old) THEN
                  CALL get_qs_kind(qs_kind_set(ikind), &
                                   softb=my_gapw, &
                                   basis_set=orb_basis_set, basis_type=my_basis_type)

                  CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                         first_sgf=first_sgfa, &
                                         lmax=la_max, &
                                         lmin=la_min, &
                                         npgf=npgfa, &
                                         nset=nseta, &
                                         nsgf_set=nsgfa, &
                                         pgf_radius=rpgfa, &
                                         set_radius=set_radius_a, &
                                         sphi=sphi_a, &
                                         zet=zeta)
               ENDIF

               IF (jkind .NE. jkind_old) THEN
                  CALL get_qs_kind(qs_kind_set(jkind), &
                                   softb=my_gapw, &
                                   basis_set=orb_basis_set, basis_type=my_basis_type)
                  CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                         first_sgf=first_sgfb, &
                                         lmax=lb_max, &
                                         lmin=lb_min, &
                                         npgf=npgfb, &
                                         nset=nsetb, &
                                         nsgf_set=nsgfb, &
                                         pgf_radius=rpgfb, &
                                         set_radius=set_radius_b, &
                                         sphi=sphi_b, &
                                         zet=zetb)

               ENDIF

               IF (debug_this_module) THEN
!$OMP CRITICAL (block_touched_critical)
                  IF ((block_touched(brow, bcol) .NE. ithread) .AND. (block_touched(brow, bcol) .NE. -1)) THEN
                     CALL cp__b("qs_integrate_potential_product.F",541,"Block has been modified by another thread")
                  END IF
                  block_touched(brow, bcol) = ithread
!$OMP END CRITICAL (block_touched_critical)
               END IF

               NULLIFY (h_block)
               CALL cp_dbcsr_get_block_p(dhmat(img)%matrix, brow, bcol, h_block, found)
               IF (.NOT. ASSOCIATED(h_block)) THEN
                  CALL cp_dbcsr_add_block_node(dhmat(img)%matrix, brow, bcol, h_block)
               END IF

               IF (pab_required) THEN
                  CALL cp_dbcsr_get_block_p(matrix=deltap(img)%matrix, &
                                            row=brow, col=bcol, BLOCK=p_block, found=found)
                  IF(.NOT.(found))CALL cp__a("qs_integrate_potential_product.F",556)
               END IF

               IF (calculate_forces) THEN
                  atom_a = atom_of_kind(iatom)
                  atom_b = atom_of_kind(jatom)
                  force_a(:) = 0.0_dp
                  force_b(:) = 0.0_dp
               ENDIF
               IF (use_virial) THEN
                  my_virial_a = 0.0_dp
                  my_virial_b = 0.0_dp
               ENDIF

               ikind_old = ikind
               jkind_old = jkind

               atom_pair_changed = .TRUE.

            ELSE

               atom_pair_changed = .FALSE.

            ENDIF

            IF (atom_pair_changed .OR. iset_old .NE. iset .OR. jset_old .NE. jset) THEN

               ncoa = npgfa(iset)*ncoset(la_max(iset))
               sgfa = first_sgfa(1, iset)
               ncob = npgfb(jset)*ncoset(lb_max(jset))
               sgfb = first_sgfb(1, jset)
               IF (pab_required) THEN
                  IF (iatom <= jatom) THEN
                     CALL dgemm("N", "N", ncoa, nsgfb(jset), nsgfa(iset), &
                                1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                p_block(sgfa, sgfb), SIZE(p_block, 1), &
                                0.0_dp, work(1, 1), SIZE(work, 1))
                     CALL dgemm("N", "T", ncoa, ncob, nsgfb(jset), &
                                1.0_dp, work(1, 1), SIZE(work, 1), &
                                sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                                0.0_dp, pab(1, 1), SIZE(pab, 1))
                  ELSE
                     CALL dgemm("N", "N", ncob, nsgfa(iset), nsgfb(jset), &
                                1.0_dp, sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                                p_block(sgfb, sgfa), SIZE(p_block, 1), &
                                0.0_dp, work(1, 1), SIZE(work, 1))
                     CALL dgemm("N", "T", ncob, ncoa, nsgfa(iset), &
                                1.0_dp, work(1, 1), SIZE(work, 1), &
                                sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                                0.0_dp, pab(1, 1), SIZE(pab, 1))
                  END IF
               END IF

               IF (iatom <= jatom) THEN
                  hab(1:ncoa, 1:ncob) = 0._dp
               ELSE
                  hab(1:ncob, 1:ncoa) = 0._dp
               ENDIF

               iset_old = iset
               jset_old = jset

            ENDIF

            rab(1) = dist_ab(1, itask)
            rab(2) = dist_ab(2, itask)
            rab(3) = dist_ab(3, itask)
            rab2 = rab(1)*rab(1)+rab(2)*rab(2)+rab(3)*rab(3)
            rb(1) = ra(1)+rab(1)
            rb(2) = ra(2)+rab(2)
            rb(3) = ra(3)+rab(3)
            zetp = zeta(ipgf, iset)+zetb(jpgf, jset)
            dab = SQRT(rab2)

            na1 = (ipgf-1)*ncoset(la_max(iset))+1
            na2 = ipgf*ncoset(la_max(iset))
            nb1 = (jpgf-1)*ncoset(lb_max(jset))+1
            nb2 = jpgf*ncoset(lb_max(jset))

            ! check whether we need to use fawzi's generalised collocation scheme
            IF (rs_v(igrid_level)%rs_grid%desc%distributed) THEN
               !tasks(4,:) is 0 for replicated, 1 for distributed 2 for exceptional distributed tasks
               IF (tasks(4, itask) .EQ. 2) THEN
                  use_subpatch = .TRUE.
               ELSE
                  use_subpatch = .FALSE.
               ENDIF
            ELSE
               use_subpatch = .FALSE.
            ENDIF

            IF (pab_required) THEN
               IF (iatom <= jatom) THEN
                  CALL integrate_pgf_product_rspace( &
                     la_max(iset), zeta(ipgf, iset), la_min(iset), &
                     lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                     ra, rab, rab2, rs_v(igrid_level)%rs_grid, cell, &
                     cube_info(igrid_level), &
                     hab, pab=pab, o1=na1-1, o2=nb1-1, &
                     eps_gvg_rspace=eps_gvg_rspace, &
                     calculate_forces=calculate_forces, &
                     force_a=force_a, force_b=force_b, &
                     compute_tau=my_compute_tau, map_consistent=map_consistent, &
                     use_virial=use_virial, my_virial_a=my_virial_a, &
                     my_virial_b=my_virial_b, use_subpatch=use_subpatch, subpatch_pattern=tasks(6, itask))
               ELSE
                  rab_inv = -rab
                  CALL integrate_pgf_product_rspace( &
                     lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                     la_max(iset), zeta(ipgf, iset), la_min(iset), &
                     rb, rab_inv, rab2, rs_v(igrid_level)%rs_grid, cell, &
                     cube_info(igrid_level), &
                     hab, pab=pab, o1=nb1-1, o2=na1-1, &
                     eps_gvg_rspace=eps_gvg_rspace, &
                     calculate_forces=calculate_forces, &
                     force_a=force_b, force_b=force_a, &
                     compute_tau=my_compute_tau, map_consistent=map_consistent, &
                     use_virial=use_virial, my_virial_a=my_virial_b, &
                     my_virial_b=my_virial_a, use_subpatch=use_subpatch, subpatch_pattern=tasks(6, itask))
               END IF
            ELSE
               IF (iatom <= jatom) THEN
                  CALL integrate_pgf_product_rspace( &
                     la_max(iset), zeta(ipgf, iset), la_min(iset), &
                     lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                     ra, rab, rab2, rs_v(igrid_level)%rs_grid, cell, &
                     cube_info(igrid_level), &
                     hab, o1=na1-1, o2=nb1-1, &
                     eps_gvg_rspace=eps_gvg_rspace, &
                     calculate_forces=calculate_forces, &
                     force_a=force_a, force_b=force_b, &
                     compute_tau=my_compute_tau, &
                     map_consistent=map_consistent, use_subpatch=use_subpatch, subpatch_pattern=tasks(6, itask))
               ELSE
                  rab_inv = -rab
                  CALL integrate_pgf_product_rspace( &
                     lb_max(jset), zetb(jpgf, jset), lb_min(jset), &
                     la_max(iset), zeta(ipgf, iset), la_min(iset), &
                     rb, rab_inv, rab2, rs_v(igrid_level)%rs_grid, cell, &
                     cube_info(igrid_level), &
                     hab, o1=nb1-1, o2=na1-1, &
                     eps_gvg_rspace=eps_gvg_rspace, &
                     calculate_forces=calculate_forces, &
                     force_a=force_b, force_b=force_a, &
                     compute_tau=my_compute_tau, &
                     map_consistent=map_consistent, use_subpatch=use_subpatch, subpatch_pattern=tasks(6, itask))
               END IF
            END IF

            new_set_pair_coming = .FALSE.
            atom_pair_done = .FALSE.
            IF (itask < task_list%taskstop(ipair, igrid_level)) THEN
               CALL int2pair(tasks(3, itask+1), ilevel, img, iatom, jatom, iset_new, jset_new, ipgf_new, jpgf_new, &
                             nimages, natom, maxset, maxpgf)
               IF (iset_new .NE. iset .OR. jset_new .NE. jset) THEN
                  new_set_pair_coming = .TRUE.
               ENDIF
            ELSE
               ! do not forget the last block
               new_set_pair_coming = .TRUE.
               atom_pair_done = .TRUE.
            ENDIF

            ! contract the block into h if we're done with the current set pair
            IF (new_set_pair_coming) THEN
               IF (iatom <= jatom) THEN
                  CALL dgemm("N", "N", ncoa, nsgfb(jset), ncob, &
                             1.0_dp, hab(1, 1), SIZE(hab, 1), &
                             sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                             0.0_dp, work(1, 1), SIZE(work, 1))
                  CALL dgemm("T", "N", nsgfa(iset), nsgfb(jset), ncoa, &
                             1.0_dp, sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             work(1, 1), SIZE(work, 1), &
                             1.0_dp, h_block(sgfa, sgfb), SIZE(h_block, 1))
               ELSE
                  CALL dgemm("N", "N", ncob, nsgfa(iset), ncoa, &
                             1.0_dp, hab(1, 1), SIZE(hab, 1), &
                             sphi_a(1, sgfa), SIZE(sphi_a, 1), &
                             0.0_dp, work(1, 1), SIZE(work, 1))
                  CALL dgemm("T", "N", nsgfb(jset), nsgfa(iset), ncob, &
                             1.0_dp, sphi_b(1, sgfb), SIZE(sphi_b, 1), &
                             work(1, 1), SIZE(work, 1), &
                             1.0_dp, h_block(sgfb, sgfa), SIZE(h_block, 1))
               END IF
            END IF

            IF (atom_pair_done) THEN
!$OMP CRITICAL(force_critical)
               IF (iatom == jatom) THEN
                  scalef = 1.0_dp
               ELSE
                  scalef = 2.0_dp
               END IF
               IF (calculate_forces) THEN
                  force(ikind)%rho_elec(:, atom_a) = &
                     force(ikind)%rho_elec(:, atom_a)+scalef*admm_scal_fac*force_a(:)
                  force(jkind)%rho_elec(:, atom_b) = &
                     force(jkind)%rho_elec(:, atom_b)+scalef*admm_scal_fac*force_b(:)
               ENDIF
               IF (use_virial) THEN
                  IF (use_virial .AND. calculate_forces) THEN
                     virial%pv_virial = virial%pv_virial+scalef*admm_scal_fac*my_virial_a
                     virial%pv_virial = virial%pv_virial+scalef*admm_scal_fac*my_virial_b
                  END IF
               END IF
!$OMP END CRITICAL (force_critical)
            ENDIF
         END DO loop_tasks
         END DO loop_pairs
!$OMP END DO

         DO img = 1, nimages
            CALL cp_dbcsr_finalize(dhmat(img)%matrix)
         END DO

      END DO loop_gridlevels

!$OMP END PARALLEL

      IF (debug_this_module) THEN
         DEALLOCATE (block_touched)
      END IF

      IF (h_duplicated) THEN
         ! Reconstruct H matrix if using distributed RS grids
         ! note send and recv direction reversed WRT collocate
         scatter = .FALSE.
         IF (do_kp) THEN
            CALL rs_distribute_matrix(rs_descs, dhmat, atom_pair_recv, atom_pair_send, &
                                      natom, nimages, scatter, hmats=hmat_kp)
         ELSE
            ALLOCATE (htemp(1))
            htemp(1)%matrix => hmat%matrix

            CALL rs_distribute_matrix(rs_descs, dhmat, atom_pair_recv, atom_pair_send, &
                                      natom, nimages, scatter, hmats=htemp)

            NULLIFY (htemp(1)%matrix)
            DEALLOCATE (htemp)
         END IF
         CALL cp_dbcsr_deallocate_matrix_set(dhmat)
      ELSE
         DO img = 1, nimages
            NULLIFY (dhmat(img)%matrix)
         END DO
         DEALLOCATE (dhmat)
      END IF

      IF (pab_required) THEN
         IF (p_duplicated) THEN
            CALL cp_dbcsr_deallocate_matrix_set(deltap)
         ELSE
            DO img = 1, nimages
               NULLIFY (deltap(img)%matrix)
            END DO
            DEALLOCATE (deltap)
         END IF
      END IF

      !   *** Release work storage ***

      DEALLOCATE (habt, workt)

      IF (pab_required) THEN
         DEALLOCATE (pabt)
      END IF

      IF (ASSOCIATED(rs_v)) THEN
         DO i = 1, SIZE(rs_v)
            CALL rs_grid_release(rs_v(i)%rs_grid)
         END DO
      END IF

      IF (calculate_forces) THEN
         DEALLOCATE (atom_of_kind)
      END IF

      CALL timestop(handle)

   END SUBROUTINE integrate_v_rspace

END MODULE qs_integrate_potential_product
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Build up the plane wave density by collocating the primitive Gaussian
!>      functions (pgf).
!> \par History
!>      Joost VandeVondele (02.2002)
!>            1) rewrote collocate_pgf for increased accuracy and speed
!>            2) collocate_core hack for PGI compiler
!>            3) added multiple grid feature
!>            4) new way to go over the grid
!>      Joost VandeVondele (05.2002)
!>            1) prelim. introduction of the real space grid type
!>      JGH [30.08.02] multigrid arrays independent from potential
!>      JGH [17.07.03] distributed real space code
!>      JGH [23.11.03] refactoring and new loop ordering
!>      JGH [04.12.03] OpneMP parallelization of main loops
!>      Joost VandeVondele (12.2003)
!>           1) modified to compute tau
!>      Joost removed incremental build feature
!>      Joost introduced map consistent
!>      Rewrote grid integration/collocation routines, [Joost VandeVondele,03.2007]
!> \author Matthias Krack (03.04.2001)
! **************************************************************************************************
MODULE qs_integrate_potential_single
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE atprop_types,                    ONLY: atprop_array_init,&
                                              atprop_type
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE external_potential_types,        ONLY: get_potential,&
                                              gth_potential_type
   USE gaussian_gridlevels,             ONLY: gaussian_gridlevel,&
                                              gridlevel_info_type
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE lri_environment_types,           ONLY: lri_kind_type
   USE mathconstants,                   ONLY: dfac,&
                                              pi
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_sum
   USE orbital_pointers,                ONLY: coset,&
                                              nco,&
                                              ncoset,&
                                              nso,&
                                              nsoset
   USE orbital_transformation_matrices, ONLY: orbtramat
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_types,                        ONLY: pw_p_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_integrate_potential_low,      ONLY: integrate_pgf_product_rspace
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE realspace_grid_types,            ONLY: pw2rs,&
                                              realspace_grid_p_type,&
                                              realspace_grid_type,&
                                              rs_grid_release,&
                                              rs_grid_retain,&
                                              rs_grid_zero,&
                                              rs_pw_transfer
   USE rs_pw_interface,                 ONLY: potential_pw2rs
   USE scptb_types,                     ONLY: get_scptb_parameter,&
                                              scp_vector_type,&
                                              scptb_parameter_type
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_integrate_potential_single'

! *** Public subroutines ***
! *** Don't include this routines directly, use the interface to
! *** qs_integrate_potential

   PUBLIC :: integrate_v_rspace_one_center, &
             integrate_v_core_rspace, &
             integrate_ppl_rspace, &
             integrate_scp_rspace, &
             integrate_rho_nlcc

CONTAINS

! **************************************************************************************************
!> \brief integrates the SCP charge functions on the grid potential
!> \param scp_pot ...
!> \param qs_env ...
!> \param scpv ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE integrate_scp_rspace(scp_pot, qs_env, scpv, calculate_forces)

      TYPE(pw_p_type), INTENT(INOUT)                     :: scp_pot
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(scp_vector_type), POINTER                     :: scpv
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'integrate_scp_rspace', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, i, iatom, ii, ikind, j, &
                                                            jj, l, lmaxscp, natom_of_kind, ni, nj, &
                                                            npme
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, cores
      LOGICAL                                            :: defined, use_virial
      REAL(KIND=dp)                                      :: alpha, dvol, eps_rho_rspace, norm, pp
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, force_b, ra
      REAL(KIND=dp), DIMENSION(3, 3)                     :: my_virial_a, my_virial_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: hab, pab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_type), POINTER                 :: rs_v
      TYPE(scptb_parameter_type), POINTER                :: scptb_kind
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (pw_env, cores)

      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
      CALL pw_env_get(pw_env=pw_env, auxbas_rs_grid=rs_v)
      CALL rs_grid_retain(rs_v)

      CALL rs_pw_transfer(rs_v, scp_pot%pw, pw2rs)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      pw_env=pw_env, &
                      force=force, virial=virial)

      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer) .AND. calculate_forces

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
      dvol = scp_pot%pw%pw_grid%dvol

      DO ikind = 1, SIZE(atomic_kind_set)

         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom_of_kind, atom_list=atom_list)
         CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
         CALL get_scptb_parameter(scptb_kind, defined=defined, lmaxscp=lmaxscp, ag=alpha)
         IF (.NOT. defined) CYCLE

         ni = ncoset(lmaxscp)
         ALLOCATE (hab(ni, 1), pab(ni, 1))
         pab = 0._dp
         hab = 0._dp

         ALLOCATE (cores(natom_of_kind))
         cores = 0
         npme = 0

         DO iatom = 1, natom_of_kind
            atom_a = atom_list(iatom)
            ra(:) = pbc(particle_set(atom_a)%r, cell)
            IF (rs_v%desc%parallel .AND. .NOT. rs_v%desc%distributed) THEN
               ! replicated realspace grid, split the atoms up between procs
               IF (MODULO(iatom, rs_v%desc%group_size) == rs_v%desc%my_pos) THEN
                  npme = npme+1
                  cores(npme) = iatom
               ENDIF
            ELSE
               npme = npme+1
               cores(npme) = iatom
            ENDIF
         END DO

         DO nj = 1, npme

            iatom = cores(nj)
            atom_a = atom_list(iatom)
            ra(:) = pbc(particle_set(atom_a)%r, cell)
            hab(:, 1) = 0.0_dp
            IF (calculate_forces) THEN
               force_a(:) = 0.0_dp
               force_b(:) = 0.0_dp
               IF (use_virial) THEN
                  my_virial_a = 0.0_dp
                  my_virial_b = 0.0_dp
               END IF
               pab = 0._dp
               DO l = 0, lmaxscp
                  pp = (2._dp*l+3._dp)/2._dp
                  norm = 2._dp**(l+2)/SQRT(pi)/dfac(2*l+1)
                  norm = SQRT(0.25_dp*dfac(2*l+1)/pi)*norm*alpha**pp
                  DO jj = 1, nco(l)
                     j = ncoset(l-1)+jj
                     DO ii = 1, nso(l)
                        i = nsoset(l-1)+ii
                        pab(j, 1) = pab(j, 1)+orbtramat(l)%s2c(ii, jj)*norm
                     END DO
                  END DO
               END DO
            END IF

            CALL integrate_pgf_product_rspace(lmaxscp, alpha, 0, &
                                              0, 0.0_dp, 0, ra, (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, &
                                              rs_v, cell, pw_env%cube_info(1), hab, pab=pab, o1=0, o2=0, &
                                              eps_gvg_rspace=eps_rho_rspace, &
                                              calculate_forces=calculate_forces, force_a=force_a, &
                                              force_b=force_b, use_virial=use_virial, my_virial_a=my_virial_a, &
                                              my_virial_b=my_virial_b, use_subpatch=.TRUE., subpatch_pattern=0_int_8)

            DO l = 0, lmaxscp
               pp = (2._dp*l+3._dp)/2._dp
               norm = 2._dp**(l+2)/SQRT(pi)/dfac(2*l+1)*dvol
               norm = SQRT(0.25_dp*dfac(2*l+1)/pi)*norm*alpha**pp
               DO jj = 1, nco(l)
                  j = ncoset(l-1)+jj
                  DO ii = 1, nso(l)
                     i = nsoset(l-1)+ii
                     scpv%vector(ikind)%vmat(i, iatom) = scpv%vector(ikind)%vmat(i, iatom)+ &
                                                         hab(j, 1)*orbtramat(l)%c2s(ii, jj)*norm
                  END DO
               END DO
            END DO

            IF (calculate_forces) THEN
               force(ikind)%rho_elec(:, iatom) = force(ikind)%rho_elec(:, iatom)+force_a(:)*dvol
               IF (use_virial) THEN
                  virial%pv_virial = virial%pv_virial+my_virial_a*dvol
               END IF
            END IF
         END DO

         DEALLOCATE (cores, hab, pab)

      END DO

      CALL rs_grid_release(rs_v)

      CALL timestop(handle)

   END SUBROUTINE integrate_scp_rspace
! **************************************************************************************************
!> \brief computes the forces/virial due to the local pseudopotential
!> \param rho_rspace ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE integrate_ppl_rspace(rho_rspace, qs_env)
      TYPE(pw_p_type), INTENT(INOUT)                     :: rho_rspace
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'integrate_ppl_rspace', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, ikind, j, lppl, &
                                                            n, natom_of_kind, ni, npme
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, cores
      LOGICAL                                            :: use_virial
      REAL(KIND=dp)                                      :: alpha, eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, force_b, ra
      REAL(KIND=dp), DIMENSION(3, 3)                     :: my_virial_a, my_virial_b
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cexp_ppl
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: hab, pab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_type), POINTER                 :: rs_v
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (pw_env, cores)

      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
      CALL pw_env_get(pw_env=pw_env, auxbas_rs_grid=rs_v)
      CALL rs_grid_retain(rs_v)

      CALL rs_pw_transfer(rs_v, rho_rspace%pw, pw2rs)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      pw_env=pw_env, &
                      force=force, virial=virial)

      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace

      DO ikind = 1, SIZE(atomic_kind_set)

         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom_of_kind, atom_list=atom_list)
         CALL get_qs_kind(qs_kind_set(ikind), gth_potential=gth_potential)

         IF (.NOT. ASSOCIATED(gth_potential)) CYCLE
         CALL get_potential(potential=gth_potential, alpha_ppl=alpha, nexp_ppl=lppl, cexp_ppl=cexp_ppl)

         IF (lppl <= 0) CYCLE

         ni = ncoset(2*lppl-2)
         ALLOCATE (hab(ni, 1), pab(ni, 1))
         pab = 0._dp

         CALL reallocate(cores, 1, natom_of_kind)
         npme = 0
         cores = 0

         ! prepare core function
         DO j = 1, lppl
            SELECT CASE (j)
            CASE (1)
               pab(1, 1) = cexp_ppl(1)
            CASE (2)
               n = coset(2, 0, 0)
               pab(n, 1) = cexp_ppl(2)
               n = coset(0, 2, 0)
               pab(n, 1) = cexp_ppl(2)
               n = coset(0, 0, 2)
               pab(n, 1) = cexp_ppl(2)
            CASE (3)
               n = coset(4, 0, 0)
               pab(n, 1) = cexp_ppl(3)
               n = coset(0, 4, 0)
               pab(n, 1) = cexp_ppl(3)
               n = coset(0, 0, 4)
               pab(n, 1) = cexp_ppl(3)
               n = coset(2, 2, 0)
               pab(n, 1) = 2._dp*cexp_ppl(3)
               n = coset(2, 0, 2)
               pab(n, 1) = 2._dp*cexp_ppl(3)
               n = coset(0, 2, 2)
               pab(n, 1) = 2._dp*cexp_ppl(3)
            CASE (4)
               n = coset(6, 0, 0)
               pab(n, 1) = cexp_ppl(4)
               n = coset(0, 6, 0)
               pab(n, 1) = cexp_ppl(4)
               n = coset(0, 0, 6)
               pab(n, 1) = cexp_ppl(4)
               n = coset(4, 2, 0)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(4, 0, 2)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(2, 4, 0)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(2, 0, 4)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(0, 4, 2)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(0, 2, 4)
               pab(n, 1) = 3._dp*cexp_ppl(4)
               n = coset(2, 2, 2)
               pab(n, 1) = 6._dp*cexp_ppl(4)
            CASE DEFAULT
               CALL cp__b("qs_integrate_potential_single.F",379,"")
            END SELECT
         END DO

         DO iatom = 1, natom_of_kind
            atom_a = atom_list(iatom)
            ra(:) = pbc(particle_set(atom_a)%r, cell)
            IF (rs_v%desc%parallel .AND. .NOT. rs_v%desc%distributed) THEN
               ! replicated realspace grid, split the atoms up between procs
               IF (MODULO(iatom, rs_v%desc%group_size) == rs_v%desc%my_pos) THEN
                  npme = npme+1
                  cores(npme) = iatom
               ENDIF
            ELSE
               npme = npme+1
               cores(npme) = iatom
            ENDIF
         END DO

         DO j = 1, npme

            iatom = cores(j)
            atom_a = atom_list(iatom)
            ra(:) = pbc(particle_set(atom_a)%r, cell)
            hab(:, 1) = 0.0_dp
            force_a(:) = 0.0_dp
            force_b(:) = 0.0_dp
            IF (use_virial) THEN
               my_virial_a = 0.0_dp
               my_virial_b = 0.0_dp
            END IF
            ni = 2*lppl-2

            CALL integrate_pgf_product_rspace(ni, alpha, 0, &
                                              0, 0.0_dp, 0, ra, (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, &
                                              rs_v, cell, pw_env%cube_info(1), hab, pab=pab, o1=0, o2=0, &
                                              eps_gvg_rspace=eps_rho_rspace, &
                                              calculate_forces=.TRUE., force_a=force_a, &
                                              force_b=force_b, use_virial=use_virial, my_virial_a=my_virial_a, &
                                              my_virial_b=my_virial_b, use_subpatch=.TRUE., subpatch_pattern=0_int_8)

            force(ikind)%gth_ppl(:, iatom) = &
               force(ikind)%gth_ppl(:, iatom)+force_a(:)*rho_rspace%pw%pw_grid%dvol

            IF (use_virial) THEN
               virial%pv_virial = virial%pv_virial+my_virial_a*rho_rspace%pw%pw_grid%dvol
               CALL cp__b("qs_integrate_potential_single.F",425,"Virial not debuged for CORE_PPL")
            END IF
         END DO

         DEALLOCATE (hab, pab)

      END DO

      CALL rs_grid_release(rs_v)

      DEALLOCATE (cores)

      CALL timestop(handle)

   END SUBROUTINE integrate_ppl_rspace

! **************************************************************************************************
!> \brief computes the forces/virial due to the nlcc pseudopotential
!> \param rho_rspace ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE integrate_rho_nlcc(rho_rspace, qs_env)
      TYPE(pw_p_type), INTENT(INOUT)                     :: rho_rspace
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'integrate_rho_nlcc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, iexp_nlcc, ikind, &
                                                            ithread, j, n, natom, nc, nexp_nlcc, &
                                                            ni, npme, nthread
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, cores, nct_nlcc
      LOGICAL                                            :: nlcc, use_virial
      REAL(KIND=dp)                                      :: alpha, eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, force_b, ra
      REAL(KIND=dp), DIMENSION(3, 3)                     :: my_virial_a, my_virial_b
      REAL(KIND=dp), DIMENSION(:), POINTER               :: alpha_nlcc
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cval_nlcc, hab, pab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gth_potential_type), POINTER                  :: gth_potential
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_type), POINTER                 :: rs_v
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (pw_env, cores)

      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
      CALL pw_env_get(pw_env=pw_env, auxbas_rs_grid=rs_v)
      CALL rs_grid_retain(rs_v)

      CALL rs_pw_transfer(rs_v, rho_rspace%pw, pw2rs)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      particle_set=particle_set, &
                      pw_env=pw_env, &
                      force=force, virial=virial)

      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace

      DO ikind = 1, SIZE(atomic_kind_set)

         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom, atom_list=atom_list)
         CALL get_qs_kind(qs_kind_set(ikind), gth_potential=gth_potential)

         IF (.NOT. ASSOCIATED(gth_potential)) CYCLE
         CALL get_potential(potential=gth_potential, nlcc_present=nlcc, nexp_nlcc=nexp_nlcc, &
                            alpha_nlcc=alpha_nlcc, nct_nlcc=nct_nlcc, cval_nlcc=cval_nlcc)

         IF (.NOT. nlcc) CYCLE

         DO iexp_nlcc = 1, nexp_nlcc

            alpha = alpha_nlcc(iexp_nlcc)
            nc = nct_nlcc(iexp_nlcc)

            ni = ncoset(2*nc-2)

            nthread = 1
            ithread = 0

            ALLOCATE (hab(ni, 1), pab(ni, 1))
            pab = 0._dp

            CALL reallocate(cores, 1, natom)
            npme = 0
            cores = 0

            ! prepare core function
            DO j = 1, nc
               SELECT CASE (j)
               CASE (1)
                  pab(1, 1) = cval_nlcc(1, iexp_nlcc)
               CASE (2)
                  n = coset(2, 0, 0)
                  pab(n, 1) = cval_nlcc(2, iexp_nlcc)/alpha**2
                  n = coset(0, 2, 0)
                  pab(n, 1) = cval_nlcc(2, iexp_nlcc)/alpha**2
                  n = coset(0, 0, 2)
                  pab(n, 1) = cval_nlcc(2, iexp_nlcc)/alpha**2
               CASE (3)
                  n = coset(4, 0, 0)
                  pab(n, 1) = cval_nlcc(3, iexp_nlcc)/alpha**4
                  n = coset(0, 4, 0)
                  pab(n, 1) = cval_nlcc(3, iexp_nlcc)/alpha**4
                  n = coset(0, 0, 4)
                  pab(n, 1) = cval_nlcc(3, iexp_nlcc)/alpha**4
                  n = coset(2, 2, 0)
                  pab(n, 1) = 2._dp*cval_nlcc(3, iexp_nlcc)/alpha**4
                  n = coset(2, 0, 2)
                  pab(n, 1) = 2._dp*cval_nlcc(3, iexp_nlcc)/alpha**4
                  n = coset(0, 2, 2)
                  pab(n, 1) = 2._dp*cval_nlcc(3, iexp_nlcc)/alpha**4
               CASE (4)
                  n = coset(6, 0, 0)
                  pab(n, 1) = cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(0, 6, 0)
                  pab(n, 1) = cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(0, 0, 6)
                  pab(n, 1) = cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(4, 2, 0)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(4, 0, 2)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(2, 4, 0)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(2, 0, 4)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(0, 4, 2)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(0, 2, 4)
                  pab(n, 1) = 3._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
                  n = coset(2, 2, 2)
                  pab(n, 1) = 6._dp*cval_nlcc(4, iexp_nlcc)/alpha**6
               CASE DEFAULT
                  CALL cp__b("qs_integrate_potential_single.F",572,"")
               END SELECT
            END DO
            IF (dft_control%nspins == 2) pab = pab*0.5_dp

            DO iatom = 1, natom
               atom_a = atom_list(iatom)
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               IF (rs_v%desc%parallel .AND. .NOT. rs_v%desc%distributed) THEN
                  ! replicated realspace grid, split the atoms up between procs
                  IF (MODULO(iatom, rs_v%desc%group_size) == rs_v%desc%my_pos) THEN
                     npme = npme+1
                     cores(npme) = iatom
                  ENDIF
               ELSE
                  npme = npme+1
                  cores(npme) = iatom
               ENDIF
            END DO

            DO j = 1, npme

               iatom = cores(j)
               atom_a = atom_list(iatom)
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               hab(:, 1) = 0.0_dp
               force_a(:) = 0.0_dp
               force_b(:) = 0.0_dp
               IF (use_virial) THEN
                  my_virial_a = 0.0_dp
                  my_virial_b = 0.0_dp
               END IF
               ni = 2*nc-2

               CALL integrate_pgf_product_rspace(ni, 1/(2*alpha**2), 0, &
                                                 0, 0.0_dp, 0, ra, (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, &
                                                 rs_v, cell, pw_env%cube_info(1), hab, pab=pab, o1=0, o2=0, &
                                                 eps_gvg_rspace=eps_rho_rspace, &
                                                 calculate_forces=.TRUE., force_a=force_a, &
                                                 force_b=force_b, use_virial=use_virial, my_virial_a=my_virial_a, &
                                                 my_virial_b=my_virial_b, use_subpatch=.TRUE., subpatch_pattern=0_int_8)

               force(ikind)%gth_nlcc(:, iatom) = &
                  force(ikind)%gth_nlcc(:, iatom)+force_a(:)*rho_rspace%pw%pw_grid%dvol

               IF (use_virial) THEN
                  virial%pv_virial = virial%pv_virial+my_virial_a*rho_rspace%pw%pw_grid%dvol
               END IF
            END DO

            DEALLOCATE (hab, pab)

         END DO

      END DO

      CALL rs_grid_release(rs_v)

      DEALLOCATE (cores)

      CALL timestop(handle)

   END SUBROUTINE integrate_rho_nlcc

! **************************************************************************************************
!> \brief computes the forces/virial due to the ionic cores with a potential on
!>      grid
!> \param v_rspace ...
!> \param qs_env ...
! **************************************************************************************************
   SUBROUTINE integrate_v_core_rspace(v_rspace, qs_env)
      TYPE(pw_p_type), INTENT(INOUT)                     :: v_rspace
      TYPE(qs_environment_type), POINTER                 :: qs_env

      CHARACTER(len=*), PARAMETER :: routineN = 'integrate_v_core_rspace', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, ikind, j, natom, &
                                                            natom_of_kind, npme
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, cores
      LOGICAL                                            :: paw_atom, skip_fcore, use_virial
      REAL(KIND=dp)                                      :: alpha_core_charge, ccore_charge, &
                                                            eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, force_b, ra
      REAL(KIND=dp), DIMENSION(3, 3)                     :: my_virial_a, my_virial_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: hab, pab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_type), POINTER                 :: rs_v
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)
      NULLIFY (virial, atprop, dft_control)

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control)

      !If gapw, check for gpw kinds
      skip_fcore = .FALSE.
      IF (dft_control%qs_control%gapw) THEN
         IF (.NOT. dft_control%qs_control%gapw_control%nopaw_as_gpw) skip_fcore = .TRUE.
      END IF

      IF (.NOT. skip_fcore) THEN
         NULLIFY (pw_env)
         ALLOCATE (cores(1))
         ALLOCATE (hab(1, 1))
         ALLOCATE (pab(1, 1))

         CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)
         CALL pw_env_get(pw_env=pw_env, auxbas_rs_grid=rs_v)
         CALL rs_grid_retain(rs_v)

         CALL rs_pw_transfer(rs_v, v_rspace%pw, pw2rs)

         CALL get_qs_env(qs_env=qs_env, &
                         atomic_kind_set=atomic_kind_set, &
                         qs_kind_set=qs_kind_set, &
                         cell=cell, &
                         dft_control=dft_control, &
                         particle_set=particle_set, &
                         pw_env=pw_env, &
                         force=force, &
                         virial=virial, &
                         atprop=atprop)

         ! atomic energy contributions
         IF (ASSOCIATED(atprop)) THEN
            natom = SIZE(particle_set)
            CALL atprop_array_init(atprop%ateb, natom)
         END IF

         use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

         eps_rho_rspace = dft_control%qs_control%eps_rho_rspace

         DO ikind = 1, SIZE(atomic_kind_set)

            CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom_of_kind, atom_list=atom_list)
            CALL get_qs_kind(qs_kind_set(ikind), paw_atom=paw_atom, &
                             alpha_core_charge=alpha_core_charge, &
                             ccore_charge=ccore_charge)

            IF (paw_atom) THEN
               force(ikind)%rho_core(:, :) = 0.0_dp
               CYCLE
            END IF
            pab(1, 1) = -ccore_charge

            IF (ASSOCIATED(force)) THEN
               force(ikind)%rho_core = 0.0_dp
            ENDIF

            IF (alpha_core_charge == 0.0_dp .OR. pab(1, 1) == 0.0_dp) CYCLE

            CALL reallocate(cores, 1, natom_of_kind)
            npme = 0
            cores = 0

            DO iatom = 1, natom_of_kind
               atom_a = atom_list(iatom)
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               IF (rs_v%desc%parallel .AND. .NOT. rs_v%desc%distributed) THEN
                  ! replicated realspace grid, split the atoms up between procs
                  IF (MODULO(iatom, rs_v%desc%group_size) == rs_v%desc%my_pos) THEN
                     npme = npme+1
                     cores(npme) = iatom
                  ENDIF
               ELSE
                  npme = npme+1
                  cores(npme) = iatom
               ENDIF
            END DO

            DO j = 1, npme

               iatom = cores(j)
               atom_a = atom_list(iatom)
               ra(:) = pbc(particle_set(atom_a)%r, cell)
               hab(1, 1) = 0.0_dp
               force_a(:) = 0.0_dp
               force_b(:) = 0.0_dp
               IF (use_virial) THEN
                  my_virial_a = 0.0_dp
                  my_virial_b = 0.0_dp
               END IF

               CALL integrate_pgf_product_rspace(0, alpha_core_charge, 0, &
                                                 0, 0.0_dp, 0, ra, (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, &
                                                 rs_v, cell, pw_env%cube_info(1), hab, pab=pab, o1=0, o2=0, &
                                                 eps_gvg_rspace=eps_rho_rspace, &
                                                 calculate_forces=.TRUE., force_a=force_a, &
                                                 force_b=force_b, use_virial=use_virial, my_virial_a=my_virial_a, &
                                                 my_virial_b=my_virial_b, use_subpatch=.TRUE., subpatch_pattern=0_int_8)

               IF (ASSOCIATED(force)) THEN
                  force(ikind)%rho_core(:, iatom) = &
                     force(ikind)%rho_core(:, iatom)+force_a(:)
               ENDIF

               IF (use_virial) THEN
                  virial%pv_virial = virial%pv_virial+my_virial_a
               END IF
               IF (ASSOCIATED(atprop)) THEN
                  atprop%ateb(atom_a) = atprop%ateb(atom_a)+0.5_dp*hab(1, 1)*pab(1, 1)
               END IF

            END DO

         END DO

         CALL rs_grid_release(rs_v)

         DEALLOCATE (hab, pab, cores)

      END IF

      CALL timestop(handle)

   END SUBROUTINE integrate_v_core_rspace

! **************************************************************************************************
!> \brief computes integrals of product of v_rspace times a one-center function
!>        required for LRIGPW
!> \param v_rspace ...
!> \param qs_env ...
!> \param int_res ...
!> \param calculate_forces ...
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE integrate_v_rspace_one_center(v_rspace, qs_env, int_res, &
                                            calculate_forces)
      TYPE(pw_p_type), INTENT(IN)                        :: v_rspace
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(lri_kind_type), DIMENSION(:), POINTER         :: int_res
      LOGICAL, INTENT(IN)                                :: calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'integrate_v_rspace_one_center', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, dir, group_size, handle, i, iatom, igrid_level, ikind, ipgf, iset, m1, &
         maxco, maxsgf_set, my_pos, na1, natom_of_kind, ncoa, nkind, nseta, offset, sgfa
      INTEGER, DIMENSION(3)                              :: lb, location, tp, ub
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, la_max, la_min, npgfa, &
                                                            nsgf_seta
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa
      LOGICAL                                            :: map_consistent, map_it_here, use_virial
      REAL(KIND=dp)                                      :: eps_gvg_rspace, eps_rho_rspace
      REAL(KIND=dp), DIMENSION(3)                        :: force_a, force_b, ra
      REAL(KIND=dp), DIMENSION(3, 3)                     :: my_virial_a, my_virial_b
      REAL(KIND=dp), DIMENSION(:), POINTER               :: set_radius_a
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: hab, pab, rpgfa, sphi_a, work_f, work_i, &
                                                            zeta
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gridlevel_info_type), POINTER                 :: gridlevel_info
      TYPE(gto_basis_set_type), POINTER                  :: lri_basis_set
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_v
      TYPE(realspace_grid_type), POINTER                 :: rs_grid
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set, qs_kind_set, atom_list, cell, dft_control, &
               first_sgfa, gridlevel_info, hab, la_max, la_min, lri_basis_set, &
               npgfa, nsgf_seta, pab, para_env, particle_set, pw_env, rpgfa, &
               rs_grid, rs_v, virial, set_radius_a, sphi_a, work_f, &
               work_i, zeta)

      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env)

      CALL pw_env_get(pw_env, rs_grids=rs_v)
      DO i = 1, SIZE(rs_v)
         CALL rs_grid_retain(rs_v(i)%rs_grid)
         CALL rs_grid_zero(rs_v(i)%rs_grid)
      END DO

      gridlevel_info => pw_env%gridlevel_info

      CALL potential_pw2rs(rs_v, v_rspace, pw_env)

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      nkind=nkind, &
                      particle_set=particle_set, &
                      para_env=para_env, pw_env=pw_env, &
                      virial=virial)

      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace
      map_consistent = dft_control%qs_control%map_consistent
      IF (map_consistent) THEN
         eps_gvg_rspace = dft_control%qs_control%eps_rho_rspace
      ELSE
         eps_gvg_rspace = dft_control%qs_control%eps_gvg_rspace
      ENDIF

      offset = 0
      my_pos = v_rspace%pw%pw_grid%para%my_pos
      group_size = v_rspace%pw%pw_grid%para%group_size

      DO ikind = 1, nkind

         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom_of_kind, atom_list=atom_list)
         CALL get_qs_kind(qs_kind_set(ikind), basis_set=lri_basis_set, basis_type="LRI")
         CALL get_gto_basis_set(gto_basis_set=lri_basis_set, &
                                first_sgf=first_sgfa, &
                                lmax=la_max, &
                                lmin=la_min, &
                                maxco=maxco, &
                                maxsgf_set=maxsgf_set, &
                                npgf=npgfa, &
                                nset=nseta, &
                                nsgf_set=nsgf_seta, &
                                pgf_radius=rpgfa, &
                                set_radius=set_radius_a, &
                                sphi=sphi_a, &
                                zet=zeta)

         CALL reallocate(hab, 1, maxco, 1, 1)
         CALL reallocate(pab, 1, maxco, 1, 1)
         int_res(ikind)%v_int = 0.0_dp
         hab = 0._dp
         pab(:, 1) = 0._dp

         DO iatom = 1, natom_of_kind

            atom_a = atom_list(iatom)
            ra(:) = pbc(particle_set(atom_a)%r, cell)
            force_a(:) = 0._dp
            force_b(:) = 0._dp
            my_virial_a(:, :) = 0._dp
            my_virial_b(:, :) = 0._dp

            DO iset = 1, nseta
               sgfa = first_sgfa(1, iset)
               ncoa = npgfa(iset)*ncoset(la_max(iset))
               hab(:, 1) = 0._dp
               CALL reallocate(work_i, 1, nsgf_seta(iset), 1, 1)
               work_i = 0.0_dp

               ! get fit coefficients for forces
               IF (calculate_forces) THEN
                  m1 = sgfa+nsgf_seta(iset)-1
                  CALL reallocate(work_f, 1, nsgf_seta(iset), 1, 1)
                  work_f(1:nsgf_seta(iset), 1) = int_res(ikind)%acoef(iatom, sgfa:m1)
                  CALL dgemm("N", "N", ncoa, 1, nsgf_seta(iset), 1.0_dp, sphi_a(1, sgfa), &
                             SIZE(sphi_a, 1), work_f(1, 1), SIZE(work_f, 1), 0.0_dp, pab(1, 1), &
                             SIZE(pab, 1))
               ENDIF

               DO ipgf = 1, npgfa(iset)
                  na1 = (ipgf-1)*ncoset(la_max(iset))
                  igrid_level = gaussian_gridlevel(gridlevel_info, zeta(ipgf, iset))
                  rs_grid => rs_v(igrid_level)%rs_grid

                  map_it_here = .FALSE.

                  IF (.NOT. ALL(rs_grid%desc%perd == 1)) THEN
                     DO dir = 1, 3
                        ! bounds of local grid (i.e. removing the 'wings'), if periodic
                        tp(dir) = FLOOR(DOT_PRODUCT(cell%h_inv(dir, :), ra)*rs_grid%desc%npts(dir))
                        tp(dir) = MODULO(tp(dir), rs_grid%desc%npts(dir))
                        IF (rs_grid%desc%perd(dir) .NE. 1) THEN
                           lb(dir) = rs_grid%lb_local(dir)+rs_grid%desc%border
                           ub(dir) = rs_grid%ub_local(dir)-rs_grid%desc%border
                        ELSE
                           lb(dir) = rs_grid%lb_local(dir)
                           ub(dir) = rs_grid%ub_local(dir)
                        ENDIF
                        ! distributed grid, only map if it is local to the grid
                        location(dir) = tp(dir)+rs_grid%desc%lb(dir)
                     ENDDO
                     IF (lb(1) <= location(1) .AND. location(1) <= ub(1) .AND. &
                         lb(2) <= location(2) .AND. location(2) <= ub(2) .AND. &
                         lb(3) <= location(3) .AND. location(3) <= ub(3)) THEN
                        map_it_here = .TRUE.
                     ENDIF
                  ELSE
                     ! not distributed, just a round-robin distribution over the full set of CPUs
                     IF (MODULO(offset, group_size) == my_pos) map_it_here = .TRUE.
                  ENDIF

                  IF (map_it_here) THEN
                     IF (.NOT. calculate_forces) THEN
                        CALL integrate_pgf_product_rspace(la_max=la_max(iset), &
                                                          zeta=zeta(ipgf, iset), la_min=la_min(iset), &
                                                          lb_max=0, zetb=0.0_dp, lb_min=0, &
                                                          ra=ra, rab=(/0.0_dp, 0.0_dp, 0.0_dp/), rab2=0.0_dp, &
                                                          rsgrid=rs_grid, cell=cell, &
                                                          cube_info=pw_env%cube_info(igrid_level), &
                                                          hab=hab, o1=na1, o2=0, eps_gvg_rspace=eps_rho_rspace, &
                                                          calculate_forces=calculate_forces, &
                                                          map_consistent=map_consistent)
                     ELSE
                        CALL integrate_pgf_product_rspace(la_max=la_max(iset), &
                                                          zeta=zeta(ipgf, iset), la_min=la_min(iset), &
                                                          lb_max=0, zetb=0.0_dp, lb_min=0, &
                                                          ra=ra, rab=(/0.0_dp, 0.0_dp, 0.0_dp/), rab2=0.0_dp, &
                                                          rsgrid=rs_grid, cell=cell, &
                                                          cube_info=pw_env%cube_info(igrid_level), &
                                                          hab=hab, pab=pab, o1=na1, o2=0, eps_gvg_rspace=eps_rho_rspace, &
                                                          calculate_forces=calculate_forces, &
                                                          force_a=force_a, force_b=force_b, &
                                                          use_virial=use_virial, &
                                                          my_virial_a=my_virial_a, my_virial_b=my_virial_b, &
                                                          map_consistent=map_consistent)
                     ENDIF
                  ENDIF
               ENDDO
               ! contract hab
               CALL dgemm("T", "N", nsgf_seta(iset), 1, ncoa, 1.0_dp, sphi_a(1, sgfa), &
                          SIZE(sphi_a, 1), hab(1, 1), SIZE(hab, 1), 0.0_dp, work_i(1, 1), SIZE(work_i, 1))

               int_res(ikind)%v_int(iatom, sgfa:sgfa-1+nsgf_seta(iset)) = work_i(1:nsgf_seta(iset), 1)

               offset = offset+nsgf_seta(iset)

            ENDDO

            IF (calculate_forces) THEN
               int_res(ikind)%v_dfdr(iatom, :) = force_a(:)
               IF (use_virial) THEN
                  virial%pv_virial = virial%pv_virial+my_virial_a
               ENDIF
            ENDIF

            DEALLOCATE (work_i)
            IF (calculate_forces) THEN
               DEALLOCATE (work_f)
            ENDIF

         END DO

         CALL mp_sum(int_res(ikind)%v_int, para_env%group)

         DEALLOCATE (hab, pab)
      END DO

      DO i = 1, SIZE(rs_v)
         CALL rs_grid_release(rs_v(i)%rs_grid)
      END DO

      CALL timestop(handle)

   END SUBROUTINE integrate_v_rspace_one_center

END MODULE qs_integrate_potential_single
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Integrate single or product functions over a potential on a RS grid
!> \par History
!>      Refactored from earlier versions by Joost VandeVondele (2002,2003,2007)
!> \author JGH [04.2014]
! **************************************************************************************************
!
! This module acts as a common container for the routines from the low level
! modules
!           qs_integrate_potential_product
!           qs_integrate_potential_single
!           qs_integrate_potential_low
!
! **************************************************************************************************
MODULE qs_integrate_potential
   USE qs_integrate_potential_low,      ONLY: integrate_pgf_product_rspace
   USE qs_integrate_potential_product,  ONLY: integrate_v_rspace
   USE qs_integrate_potential_single,   ONLY: integrate_ppl_rspace,&
                                              integrate_rho_nlcc,&
                                              integrate_scp_rspace,&
                                              integrate_v_core_rspace,&
                                              integrate_v_rspace_one_center
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_integrate_potential'

! *** Public subroutines ***

   ! included from qs_integrate_potential_product
   PUBLIC :: integrate_v_rspace

   ! included from qs_integrate_potential_single
   PUBLIC :: integrate_v_rspace_one_center, &
             integrate_v_core_rspace, &
             integrate_ppl_rspace, &
             integrate_scp_rspace, &
             integrate_rho_nlcc

   ! included from qs_integrate_potential_low
   PUBLIC :: integrate_pgf_product_rspace

END MODULE qs_integrate_potential
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
MODULE qs_rho0_ggrid

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cube_utils,                      ONLY: cube_info_type
   USE input_constants,                 ONLY: tddfpt_singlet
   USE kinds,                           ONLY: dp,&
                                              dp_size,&
                                              int_8
   USE mathconstants,                   ONLY: dfac,&
                                              fourpi
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_sum
   USE orbital_pointers,                ONLY: indco,&
                                              nco,&
                                              ncoset,&
                                              nso,&
                                              nsoset
   USE orbital_transformation_matrices, ONLY: orbtramat
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_integrate_function,&
                                              pw_transfer,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_release
   USE qs_collocate_density,            ONLY: collocate_pgf_product_rspace
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_harmonics_atom,               ONLY: get_none0_cg_list,&
                                              harmonics_atom_type
   USE qs_integrate_potential,          ONLY: integrate_pgf_product_rspace
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_p_env_types,                  ONLY: qs_p_env_type
   USE qs_rho0_types,                   ONLY: get_rho0_mpole,&
                                              rho0_mpole_type
   USE qs_rho_atom_types,               ONLY: get_rho_atom,&
                                              rho_atom_coeff,&
                                              rho_atom_type
   USE realspace_grid_types,            ONLY: &
        pw2rs, realspace_grid_desc_p_type, realspace_grid_desc_type, realspace_grid_p_type, &
        realspace_grid_type, rs2pw, rs_grid_create, rs_grid_release, rs_grid_retain, rs_grid_zero, &
        rs_pw_transfer
   USE termination,                     ONLY: stop_memory
   USE util,                            ONLY: get_limit
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_rho0_ggrid'

! *** Public subroutines ***

   PUBLIC :: put_rho0_on_grid, rho0_s_grid_create, integrate_vhg0_rspace

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param rho0 ...
!> \param tot_rs_int ...
! **************************************************************************************************
   SUBROUTINE put_rho0_on_grid(qs_env, rho0, tot_rs_int)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(rho0_mpole_type), POINTER                     :: rho0
      REAL(dp), INTENT(OUT)                              :: tot_rs_int

      CHARACTER(LEN=*), PARAMETER :: routineN = 'put_rho0_on_grid', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: auxbas_grid, handle, iat, iatom, igrid, &
                                                            ikind, istat, ithread, j, l0_ikind, &
                                                            lmax0, lmax_global, nat, nch_ik, &
                                                            nch_max, npme
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, cores
      LOGICAL                                            :: paw_atom
      REAL(KIND=dp)                                      :: eps_rho_rspace, rpgf0, zet0
      REAL(KIND=dp), DIMENSION(3)                        :: ra
      REAL(KIND=dp), DIMENSION(:), POINTER               :: Qlm_c
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: pab
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: rho0_r_tmp
      TYPE(pw_p_type), POINTER                           :: coeff_gspace, coeff_rspace, rho0_s_gs, &
                                                            rho0_s_rs
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: descs
      TYPE(realspace_grid_desc_type), POINTER            :: desc
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: grids
      TYPE(realspace_grid_type), POINTER                 :: rs_grid

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set, qs_kind_set, cores, pab, Qlm_c)

      NULLIFY (dft_control, pw_env, particle_set, para_env, cell, rho0_s_gs, rho0_s_rs)
      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, &
                      particle_set=particle_set, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      para_env=para_env, &
                      pw_env=pw_env, cell=cell)
      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace

      NULLIFY (descs, pw_pools)
      CALL pw_env_get(pw_env=pw_env, rs_descs=descs, rs_grids=grids, pw_pools=pw_pools)
      cube_info => pw_env%cube_info
      auxbas_grid = pw_env%auxbas_grid

      NULLIFY (rho0_s_gs, rho0_s_rs)
      CALL get_rho0_mpole(rho0_mpole=rho0, lmax_0=lmax0, &
                          zet0_h=zet0, igrid_zet0_s=igrid, &
                          rho0_s_gs=rho0_s_gs, &
                          rho0_s_rs=rho0_s_rs)

      ! *** set up the rs grid at level igrid
      NULLIFY (rs_grid, desc, pw_pool, coeff_rspace, coeff_gspace)
      desc => descs(igrid)%rs_desc
      rs_grid => grids(igrid)%rs_grid
      pw_pool => pw_pools(igrid)%pool

      IF(.NOT.(ASSOCIATED(desc)))CALL cp__a("qs_rho0_ggrid.F",161)
      IF(.NOT.(ASSOCIATED(pw_pool)))CALL cp__a("qs_rho0_ggrid.F",162)

      IF (igrid /= auxbas_grid) THEN
         ALLOCATE (coeff_rspace, STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 166, &
                                          "coeff_rspace", 0)
         ALLOCATE (coeff_gspace, STAT=istat)
         IF (istat /= 0) CALL stop_memory(routineN, moduleN, 169, &
                                          "coeff_gspace", 0)

         CALL pw_pool_create_pw(pw_pool, coeff_rspace%pw, use_data=REALDATA3D, &
                                in_space=REALSPACE)
         CALL pw_pool_create_pw(pw_pool, coeff_gspace%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
      END IF
      CALL rs_grid_retain(rs_grid)
      CALL rs_grid_zero(rs_grid)

      nch_max = ncoset(lmax0)

      ALLOCATE (pab(nch_max, 1), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 184, &
                                       "pab", dp_size*nch_max)

      !Compute global lmax for use in collocate_pgf_product_rspace
      lmax_global = 0
      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_rho0_mpole(rho0_mpole=rho0, ikind=ikind, l0_ikind=l0_ikind)
         lmax_global = MAX(lmax_global, l0_ikind)
      END DO

      DO ikind = 1, SIZE(atomic_kind_set)
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=nat)
         CALL get_qs_kind(qs_kind_set(ikind), paw_atom=paw_atom)

         IF (.NOT. paw_atom .AND. dft_control%qs_control%gapw_control%nopaw_as_gpw) CYCLE

         CALL get_rho0_mpole(rho0_mpole=rho0, ikind=ikind, l0_ikind=l0_ikind, &
                             rpgf0_s=rpgf0)

         nch_ik = ncoset(l0_ikind)
         pab = 0.0_dp

         CALL reallocate(cores, 1, nat)
         npme = 0
         cores = 0

         DO iat = 1, nat
            iatom = atom_list(iat)
            ra(:) = pbc(particle_set(iatom)%r, cell)
            IF (rs_grid%desc%parallel .AND. .NOT. rs_grid%desc%distributed) THEN
               ! replicated realspace grid, split the atoms up between procs
               IF (MODULO(nat, rs_grid%desc%group_size) == rs_grid%desc%my_pos) THEN
                  npme = npme+1
                  cores(npme) = iat
               ENDIF
            ELSE
               npme = npme+1
               cores(npme) = iat
            ENDIF

         END DO

         ithread = 0
         DO j = 1, npme

            iat = cores(j)
            iatom = atom_list(iat)

            CALL get_rho0_mpole(rho0_mpole=rho0, iat=iatom, Qlm_car=Qlm_c)

            pab(1:nch_ik, 1) = Qlm_c(1:nch_ik)

            ra(:) = pbc(particle_set(iatom)%r, cell)

            CALL collocate_pgf_product_rspace( &
               l0_ikind, zet0, 0, 0, 0.0_dp, 0, &
               ra, (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, 1.0_dp, pab, 0, 0, &
               rs_grid, cell, cube_info(igrid), eps_rho_rspace, ga_gb_function=401, &
               ithread=ithread, collocate_rho0=.TRUE., rpgf0_s=rpgf0, &
               use_subpatch=.TRUE., subpatch_pattern=0_int_8, lmax_global=lmax_global)

         END DO !  j

      END DO ! ikind

      IF (ASSOCIATED(cores)) THEN
         DEALLOCATE (cores)
      END IF

      DEALLOCATE (pab)

      IF (igrid /= auxbas_grid) THEN
         CALL rs_pw_transfer(rs_grid, coeff_rspace%pw, rs2pw)
         CALL rs_grid_release(rs_grid)
         CALL pw_zero(rho0_s_gs%pw)
         CALL pw_transfer(coeff_rspace%pw, coeff_gspace%pw)
         CALL pw_axpy(coeff_gspace%pw, rho0_s_gs%pw)

         tot_rs_int = pw_integrate_function(coeff_rspace%pw, isign=-1)

         CALL pw_pool_give_back_pw(pw_pool, coeff_rspace%pw)
         CALL pw_pool_give_back_pw(pw_pool, coeff_gspace%pw)

         DEALLOCATE (coeff_rspace, coeff_gspace)

      ELSE

         CALL pw_pool_create_pw(pw_pool, rho0_r_tmp%pw, &
                                use_data=REALDATA3D, in_space=REALSPACE)

         CALL rs_pw_transfer(rs_grid, rho0_r_tmp%pw, rs2pw)
         CALL rs_grid_release(rs_grid)

         tot_rs_int = pw_integrate_function(rho0_r_tmp%pw, isign=-1)

         CALL pw_transfer(rho0_r_tmp%pw, rho0_s_rs%pw)
         CALL pw_pool_give_back_pw(pw_pool, rho0_r_tmp%pw)

         CALL pw_zero(rho0_s_gs%pw)
         CALL pw_transfer(rho0_s_rs%pw, rho0_s_gs%pw)
      END IF
      CALL timestop(handle)

   END SUBROUTINE put_rho0_on_grid

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param rho0_mpole ...
!> \param tddft ...
! **************************************************************************************************
   SUBROUTINE rho0_s_grid_create(qs_env, rho0_mpole, tddft)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole
      LOGICAL, INTENT(IN), OPTIONAL                      :: tddft

      CHARACTER(len=*), PARAMETER :: routineN = 'rho0_s_grid_create', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_tddft
      TYPE(pw_env_type), POINTER                         :: new_pw_env
      TYPE(pw_p_type), POINTER                           :: rho_gs, rho_rs
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool

      my_tddft = .FALSE.
      IF (PRESENT(tddft)) my_tddft = tddft

      NULLIFY (new_pw_env)
      CALL get_qs_env(qs_env, pw_env=new_pw_env)
      IF(.NOT.(ASSOCIATED(new_pw_env)))CALL cp__a("qs_rho0_ggrid.F",314)

      NULLIFY (auxbas_pw_pool)
      CALL pw_env_get(new_pw_env, auxbas_pw_pool=auxbas_pw_pool)
      IF(.NOT.(ASSOCIATED(auxbas_pw_pool)))CALL cp__a("qs_rho0_ggrid.F",318)

      ! reallocate rho0 on the global grid in real and reciprocal space
      NULLIFY (rho_rs, rho_gs)
      IF(.NOT.(ASSOCIATED(rho0_mpole)))CALL cp__a("qs_rho0_ggrid.F",322)
      rho_rs => rho0_mpole%rho0_s_rs
      rho_gs => rho0_mpole%rho0_s_gs

      ! rho0 density in real space
      IF (ASSOCIATED(rho_rs)) THEN
         CALL pw_release(rho_rs%pw)
         DEALLOCATE (rho_rs)
      END IF
      ALLOCATE (rho_rs)
      CALL pw_pool_create_pw(auxbas_pw_pool, rho_rs%pw, &
                             use_data=REALDATA3D, in_space=REALSPACE)
      rho0_mpole%rho0_s_rs => rho_rs

      ! rho0 density in reciprocal space
      IF (ASSOCIATED(rho_gs)) THEN
         CALL pw_release(rho_gs%pw)
         DEALLOCATE (rho_gs)
      END IF
      ALLOCATE (rho_gs)
      CALL pw_pool_create_pw(auxbas_pw_pool, rho_gs%pw, &
                             use_data=COMPLEXDATA1D)
      rho_gs%pw%in_space = RECIPROCALSPACE
      rho0_mpole%rho0_s_gs => rho_gs

      IF (my_tddft) THEN
         rho0_mpole%igrid_zet0_s = qs_env%local_rho_set%rho0_mpole%igrid_zet0_s
      END IF

   END SUBROUTINE rho0_s_grid_create

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param v_rspace ...
!> \param calculate_forces ...
!> \param tddft ...
!> \param do_triplet ...
!> \param p_env ...
! **************************************************************************************************
   SUBROUTINE integrate_vhg0_rspace(qs_env, v_rspace, calculate_forces, &
                                    tddft, do_triplet, p_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(pw_p_type)                                    :: v_rspace
      LOGICAL, INTENT(IN)                                :: calculate_forces
      LOGICAL, INTENT(IN), OPTIONAL                      :: tddft, do_triplet
      TYPE(qs_p_env_type), OPTIONAL, POINTER             :: p_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'integrate_vhg0_rspace', &
         routineP = moduleN//':'//routineN

      INTEGER :: auxbas_grid, bo(2), handle, i, iat, iatom, ic, icg, ico, ig1, ig2, igrid, ii, &
         ikind, ipgf1, ipgf2, is, iset1, iset2, iso, iso1, iso2, ispin, j, l0_ikind, llmax, lmax0, &
         lshell, lx, ly, lz, m1, m2, max_iso_not0_local, max_s_harm, maxl, maxso, mepos, n1, n2, &
         nat, nch_ik, nch_max, ncurr, nset, nsotot, nspins, num_pe
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cg_n_list
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: cg_list
      INTEGER, DIMENSION(:), POINTER                     :: atom_list, lmax, lmin, npgf
      LOGICAL                                            :: grid_distributed, my_tddft, paw_atom, &
                                                            use_virial
      REAL(dp)                                           :: c4pi, eps_rho_rspace, force_tmp(3), &
                                                            ra(3), rpgf0, scale, zet0
      REAL(dp), DIMENSION(:), POINTER                    :: hab_sph, norm_l, Qlm
      REAL(dp), DIMENSION(:, :), POINTER                 :: hab, hdab_sph, intloc, pab
      REAL(dp), DIMENSION(:, :, :), POINTER              :: a_hdab_sph, hadb, hdab, Qlm_gg
      REAL(dp), DIMENSION(:, :, :, :), POINTER           :: a_hdab
      REAL(KIND=dp), DIMENSION(3, 3)                     :: my_virial_a, my_virial_b
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cube_info_type), DIMENSION(:), POINTER        :: cube_info
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type), POINTER                           :: coeff_gaux, coeff_gspace, coeff_raux, &
                                                            coeff_rspace
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: pw_aux, pw_pool
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(realspace_grid_desc_p_type), DIMENSION(:), &
         POINTER                                         :: rs_descs
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_type), POINTER                 :: rs_v
      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: int_local_h, int_local_s
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      TYPE(virial_type), POINTER                         :: virial

      c4pi = fourpi

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set, qs_kind_set, dft_control, para_env, particle_set)
      NULLIFY (cell, force, pw_env, rho0_mpole, rho_atom_set)

      my_tddft = .FALSE.
      IF (PRESENT(tddft)) my_tddft = tddft

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      cell=cell, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      force=force, pw_env=pw_env, &
                      rho0_mpole=rho0_mpole, &
                      rho_atom_set=rho_atom_set, &
                      particle_set=particle_set, &
                      virial=virial)

      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      nspins = dft_control%nspins

      IF (my_tddft) THEN
         IF (PRESENT(do_triplet)) THEN
            IF (nspins == 1 .AND. do_triplet) RETURN
         ELSE
            IF (nspins == 1 .AND. dft_control%tddfpt_control%res_etype /= tddfpt_singlet) RETURN
         ENDIF
      END IF

      IF (my_tddft) THEN
         rho0_mpole => p_env%local_rho_set%rho0_mpole
         rho_atom_set => p_env%local_rho_set%rho_atom_set
      END IF

      CALL get_rho0_mpole(rho0_mpole=rho0_mpole, lmax_0=lmax0, &
                          zet0_h=zet0, igrid_zet0_s=igrid, &
                          norm_g0l_h=norm_l)

      ! *** set up of the potential on the multigrids
      NULLIFY (rs_descs, pw_pools)
      IF(.NOT.(ASSOCIATED(pw_env)))CALL cp__a("qs_rho0_ggrid.F",460)
      CALL pw_env_get(pw_env, rs_descs=rs_descs, pw_pools=pw_pools)

      ! *** assign from pw_env
      auxbas_grid = pw_env%auxbas_grid
      cube_info => pw_env%cube_info

      ! *** Get the potential on the right grid
      NULLIFY (rs_v, rs_desc, pw_pool, pw_aux, coeff_rspace, coeff_gspace)
      rs_desc => rs_descs(igrid)%rs_desc
      pw_pool => pw_pools(igrid)%pool

      ALLOCATE (coeff_gspace)
      ALLOCATE (coeff_rspace)
      CALL pw_pool_create_pw(pw_pool, coeff_gspace%pw, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)

      CALL pw_pool_create_pw(pw_pool, coeff_rspace%pw, use_data=REALDATA3D, &
                             in_space=REALSPACE)

      IF (igrid /= auxbas_grid) THEN
         pw_aux => pw_pools(auxbas_grid)%pool
         ALLOCATE (coeff_gaux)
         CALL pw_pool_create_pw(pw_aux, coeff_gaux%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
         CALL pw_transfer(v_rspace%pw, coeff_gaux%pw)
         CALL pw_copy(coeff_gaux%pw, coeff_gspace%pw)
         CALL pw_transfer(coeff_gspace%pw, coeff_rspace%pw)
         CALL pw_pool_give_back_pw(pw_aux, coeff_gaux%pw)
         DEALLOCATE (coeff_gaux)
         ALLOCATE (coeff_raux)
         CALL pw_pool_create_pw(pw_aux, coeff_raux%pw, use_data=REALDATA3D, &
                                in_space=REALSPACE)
         scale = coeff_rspace%pw%pw_grid%dvol/coeff_raux%pw%pw_grid%dvol
         coeff_rspace%pw%cr3d = scale*coeff_rspace%pw%cr3d
         CALL pw_pool_give_back_pw(pw_aux, coeff_raux%pw)
         DEALLOCATE (coeff_raux)
      ELSE

         IF (coeff_gspace%pw%pw_grid%spherical) THEN
            CALL pw_transfer(v_rspace%pw, coeff_gspace%pw)
            CALL pw_transfer(coeff_gspace%pw, coeff_rspace%pw)
         ELSE
            CALL pw_copy(v_rspace%pw, coeff_rspace%pw)
         END IF
      END IF
      CALL pw_pool_give_back_pw(pw_pool, coeff_gspace%pw)
      DEALLOCATE (coeff_gspace)

      ! *** set up the rs grid at level igrid
      CALL rs_grid_create(rs_v, rs_desc)
      CALL rs_grid_zero(rs_v)
      CALL rs_pw_transfer(rs_v, coeff_rspace%pw, pw2rs)

      CALL pw_pool_give_back_pw(pw_pool, coeff_rspace%pw)
      DEALLOCATE (coeff_rspace)

      ! *** Now the potential is on the right grid => integration

      eps_rho_rspace = dft_control%qs_control%eps_rho_rspace

      !   *** Allocate work storage ***

      NULLIFY (hab, hab_sph, hdab, hdab_sph, hadb, pab, a_hdab, a_hdab_sph)
      nch_max = ncoset(lmax0)
      CALL reallocate(hab, 1, nch_max, 1, 1)
      CALL reallocate(hab_sph, 1, nch_max)
      CALL reallocate(hdab, 1, 3, 1, nch_max, 1, 1)
      CALL reallocate(hadb, 1, 3, 1, nch_max, 1, 1)
      CALL reallocate(hdab_sph, 1, 3, 1, nch_max)
      CALL reallocate(a_hdab, 1, 3, 1, 3, 1, nch_max, 1, 1)
      CALL reallocate(a_hdab_sph, 1, 3, 1, 3, 1, nch_max)
      CALL reallocate(pab, 1, nch_max, 1, 1)

      ncurr = -1

      grid_distributed = rs_v%desc%distributed

      DO ikind = 1, SIZE(atomic_kind_set, 1)
         NULLIFY (orb_basis_set, atom_list, harmonics)
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=nat)
         CALL get_qs_kind(qs_kind_set(ikind), &
                          basis_set=orb_basis_set, &
                          paw_atom=paw_atom, &
                          harmonics=harmonics)

         IF (.NOT. paw_atom) CYCLE

         NULLIFY (Qlm_gg, lmax, npgf)
         CALL get_rho0_mpole(rho0_mpole=rho0_mpole, ikind=ikind, &
                             l0_ikind=l0_ikind, Qlm_gg=Qlm_gg, &
                             rpgf0_s=rpgf0)

         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                lmax=lmax, lmin=lmin, &
                                maxso=maxso, maxl=maxl, &
                                nset=nset, npgf=npgf)

         nsotot = maxso*nset
         ALLOCATE (intloc(nsotot, nsotot))

         ! Initialize the local KS integrals

         nch_ik = ncoset(l0_ikind)
         pab = 1.0_dp
         max_s_harm = harmonics%max_s_harm
         llmax = harmonics%llmax

         ALLOCATE (cg_list(2, nsoset(maxl)**2, max_s_harm), cg_n_list(max_s_harm))

         num_pe = para_env%num_pe
         mepos = para_env%mepos
         DO j = 0, num_pe-1
            bo = get_limit(nat, num_pe, j)
            IF (.NOT. grid_distributed .AND. j /= mepos) CYCLE

            DO iat = bo(1), bo(2)
               iatom = atom_list(iat)
               ra(:) = pbc(particle_set(iatom)%r, cell)

               NULLIFY (Qlm)
               CALL get_rho0_mpole(rho0_mpole=rho0_mpole, iat=iatom, Qlm_tot=Qlm)

               hab = 0.0_dp
               hdab = 0.0_dp
               intloc = 0._dp
               IF (use_virial) THEN
                  my_virial_a = 0.0_dp
                  my_virial_b = 0.0_dp
                  a_hdab = 0.0_dp
               END IF

               CALL integrate_pgf_product_rspace( &
                  l0_ikind, zet0, 0, 0, 0.0_dp, 0, &
                  ra, (/0.0_dp, 0.0_dp, 0.0_dp/), 0.0_dp, rs_v, cell, &
                  cube_info(igrid), hab, pab, o1=0, o2=0, &
                  eps_gvg_rspace=eps_rho_rspace, &
                  calculate_forces=calculate_forces, &
                  hdab=hdab, hadb=hadb, &
                  collocate_rho0=.TRUE., rpgf0_s=rpgf0, &
                  use_virial=use_virial, my_virial_a=my_virial_a, my_virial_b=my_virial_b, &
                  a_hdab=a_hdab, use_subpatch=.TRUE., subpatch_pattern=0_int_8)

               ! Convert from cartesian to spherical
               DO lshell = 0, l0_ikind
                  DO is = 1, nso(lshell)
                     iso = is+nsoset(lshell-1)
                     hab_sph(iso) = 0.0_dp
                     hdab_sph(1:3, iso) = 0.0_dp
                     a_hdab_sph(1:3, 1:3, iso) = 0.0_dp
                     DO ic = 1, nco(lshell)
                        ico = ic+ncoset(lshell-1)
                        lx = indco(1, ico)
                        ly = indco(2, ico)
                        lz = indco(3, ico)

                        hab_sph(iso) = hab_sph(iso)+ &
                                       orbtramat(lshell)%c2s(is, ic)*hab(ico, 1)* &
                                       norm_l(lshell)/ &
                                       SQRT(c4pi*dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1)/ &
                                            dfac(2*lshell+1))

                        IF (calculate_forces) THEN
                           hdab_sph(1:3, iso) = hdab_sph(1:3, iso)+ &
                                                orbtramat(lshell)%c2s(is, ic)*hdab(1:3, ico, 1)* &
                                                norm_l(lshell)/ &
                                                SQRT(c4pi*dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1)/ &
                                                     dfac(2*lshell+1))
                        END IF
                        IF (use_virial) THEN
                           DO ii = 1, 3
                           DO i = 1, 3
                              a_hdab_sph(i, ii, iso) = a_hdab_sph(i, ii, iso)+ &
                                                       orbtramat(lshell)%c2s(is, ic)*a_hdab(i, ii, ico, 1)* &
                                                       norm_l(lshell)/ &
                                                       SQRT(c4pi*dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1)/ &
                                                            dfac(2*lshell+1))
                           END DO
                           END DO
                        END IF

                     END DO ! ic
                  END DO ! is
               END DO ! lshell

               m1 = 0
               DO iset1 = 1, nset

                  m2 = 0
                  DO iset2 = 1, nset
                     CALL get_none0_cg_list(harmonics%my_CG, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                            max_s_harm, llmax, cg_list, cg_n_list, max_iso_not0_local)
                     n1 = nsoset(lmax(iset1))
                     DO ipgf1 = 1, npgf(iset1)
                        n2 = nsoset(lmax(iset2))
                        DO ipgf2 = 1, npgf(iset2)

                           DO iso = 1, MIN(nsoset(l0_ikind), max_iso_not0_local)
                              DO icg = 1, cg_n_list(iso)
                                 iso1 = cg_list(1, icg, iso)
                                 iso2 = cg_list(2, icg, iso)

                                 ig1 = iso1+n1*(ipgf1-1)+m1
                                 ig2 = iso2+n2*(ipgf2-1)+m2

                                 intloc(ig1, ig2) = intloc(ig1, ig2)+Qlm_gg(ig1, ig2, iso)*hab_sph(iso)

                              END DO ! icg
                           END DO ! iso

                        END DO ! ipgf2
                     END DO ! ipgf1
                     m2 = m2+maxso
                  END DO ! iset2
                  m1 = m1+maxso
               END DO ! iset1

               IF (grid_distributed) THEN
                  ! sum result over all processors
                  CALL mp_sum(intloc, para_env%group)
               END IF

               IF (j == mepos) THEN
                  rho_atom => rho_atom_set(iatom)
                  CALL get_rho_atom(rho_atom=rho_atom, ga_Vlocal_gb_h=int_local_h, ga_Vlocal_gb_s=int_local_s)
                  DO ispin = 1, nspins
                     int_local_h(ispin)%r_coef = int_local_h(ispin)%r_coef+intloc
                     int_local_s(ispin)%r_coef = int_local_s(ispin)%r_coef+intloc
                  END DO
               END IF

               IF (calculate_forces) THEN
                  force_tmp(1:3) = 0.0_dp
                  DO iso = 1, nsoset(l0_ikind)
                     force_tmp(1) = force_tmp(1)+Qlm(iso)*hdab_sph(1, iso)
                     force_tmp(2) = force_tmp(2)+Qlm(iso)*hdab_sph(2, iso)
                     force_tmp(3) = force_tmp(3)+Qlm(iso)*hdab_sph(3, iso)
                  END DO
                  force(ikind)%g0s_Vh_elec(1:3, iat) = force(ikind)%g0s_Vh_elec(1:3, iat)+force_tmp(1:3)
               END IF
               IF (use_virial) THEN
                  my_virial_a = 0.0_dp
                  DO iso = 1, nsoset(l0_ikind)
                     DO ii = 1, 3
                     DO i = 1, 3
                        virial%pv_virial(i, ii) = virial%pv_virial(i, ii)+Qlm(iso)*a_hdab_sph(i, ii, iso)
                     END DO
                     END DO
                  END DO
               END IF

            END DO
         END DO

         DEALLOCATE (intloc)
         DEALLOCATE (cg_list, cg_n_list)

      END DO ! ikind

      CALL rs_grid_release(rs_v)

      DEALLOCATE (hab, hdab, hadb, hab_sph, hdab_sph, pab, a_hdab, a_hdab_sph)

      CALL timestop(handle)

   END SUBROUTINE integrate_vhg0_rspace

END MODULE qs_rho0_ggrid
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE qs_rho0_methods

   USE ao_util,                         ONLY: exp_radius,&
                                              gaussint_sph,&
                                              trace_r_AxB
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cp_control_types,                ONLY: dft_control_type,&
                                              gapw_control_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: dfac,&
                                              fourpi
   USE memory_utilities,                ONLY: reallocate
   USE orbital_pointers,                ONLY: indco,&
                                              indso,&
                                              nco,&
                                              ncoset,&
                                              nso,&
                                              nsoset
   USE orbital_transformation_matrices, ONLY: orbtramat
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type,&
                                              set_qs_env
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: get_none0_cg_list,&
                                              harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type,&
                                              set_qs_kind
   USE qs_local_rho_types,              ONLY: allocate_rhoz,&
                                              calculate_rhoz,&
                                              local_rho_type,&
                                              rhoz_type
   USE qs_oce_methods,                  ONLY: prj_scatter
   USE qs_rho0_ggrid,                   ONLY: rho0_s_grid_create
   USE qs_rho0_types,                   ONLY: &
        allocate_multipoles, allocate_rho0_atom, allocate_rho0_atom_rad, allocate_rho0_mpole, &
        calculate_g0, get_rho0_mpole, initialize_mpole_rho, mpole_gau_overlap, mpole_rho_atom, &
        rho0_atom_type, rho0_mpole_type, write_rho0_info
   USE qs_rho_atom_types,               ONLY: get_rho_atom,&
                                              rho_atom_coeff,&
                                              rho_atom_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_rho0_methods'

! *** Public subroutines ***

   PUBLIC :: calculate_rho0_atom, init_rho0

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param mp_gau ...
!> \param orb_basis ...
!> \param harmonics ...
!> \param nchannels ...
!> \param nsotot ...
! **************************************************************************************************
   SUBROUTINE calculate_mpole_gau(mp_gau, orb_basis, harmonics, nchannels, nsotot)

      TYPE(mpole_gau_overlap)                            :: mp_gau
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      INTEGER, INTENT(IN)                                :: nchannels, nsotot

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_mpole_gau', &
         routineP = moduleN//':'//routineN

      INTEGER :: handle, icg, ig1, ig2, ipgf1, ipgf2, iset1, iset2, iso, iso1, iso2, l, l1, l2, &
         llmax, m1, m2, max_iso_not0_local, max_s_harm, maxl, maxso, n1, n2, nset
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cg_n_list
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: cg_list
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf
      REAL(dp)                                           :: zet1, zet2
      REAL(dp), DIMENSION(:, :), POINTER                 :: zet
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_CG

      CALL timeset(routineN, handle)

      NULLIFY (lmax, lmin, npgf, my_CG, zet)

      CALL reallocate(mp_gau%Qlm_gg, 1, nsotot, 1, nsotot, 1, nchannels)

      CALL get_gto_basis_set(gto_basis_set=orb_basis, &
                             lmax=lmax, lmin=lmin, maxso=maxso, &
                             npgf=npgf, nset=nset, zet=zet, maxl=maxl)

      max_s_harm = harmonics%max_s_harm
      llmax = harmonics%llmax

      ALLOCATE (cg_list(2, nsoset(maxl)**2, max_s_harm), cg_n_list(max_s_harm))

      my_CG => harmonics%my_CG

      m1 = 0
      DO iset1 = 1, nset
         m2 = 0
         DO iset2 = 1, nset

            CALL get_none0_cg_list(my_CG, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                   max_s_harm, llmax, cg_list, cg_n_list, max_iso_not0_local)

            n1 = nsoset(lmax(iset1))
            DO ipgf1 = 1, npgf(iset1)
               zet1 = zet(ipgf1, iset1)

               n2 = nsoset(lmax(iset2))
               DO ipgf2 = 1, npgf(iset2)
                  zet2 = zet(ipgf2, iset2)

                  DO iso = 1, MIN(nchannels, max_iso_not0_local)
                     l = indso(1, iso)
                     DO icg = 1, cg_n_list(iso)
                        iso1 = cg_list(1, icg, iso)
                        iso2 = cg_list(2, icg, iso)

                        l1 = indso(1, iso1)
                        l2 = indso(1, iso2)
                        ig1 = iso1+n1*(ipgf1-1)+m1
                        ig2 = iso2+n2*(ipgf2-1)+m2

                        mp_gau%Qlm_gg(ig1, ig2, iso) = fourpi/(2._dp*l+1._dp)* &
                                                       my_CG(iso1, iso2, iso)*gaussint_sph(zet1+zet2, l+l1+l2)
                     END DO ! icg
                  END DO ! iso

               END DO ! ipgf2
            END DO ! ipgf1
            m2 = m2+maxso
         END DO ! iset2
         m1 = m1+maxso
      END DO ! iset1

      DEALLOCATE (cg_list, cg_n_list)

      CALL timestop(handle)
   END SUBROUTINE calculate_mpole_gau

! **************************************************************************************************
!> \brief ...
!> \param gapw_control ...
!> \param rho_atom_set ...
!> \param rho0_atom_set ...
!> \param rho0_mp ...
!> \param a_list ...
!> \param g_atom ...
!> \param paw_atom ...
!> \param natom ...
!> \param ikind ...
!> \param qs_kind ...
!> \param harmonics ...
!> \param rho0_h_tot ...
! **************************************************************************************************
   SUBROUTINE calculate_rho0_atom(gapw_control, rho_atom_set, rho0_atom_set, &
                                  rho0_mp, a_list, g_atom, &
                                  paw_atom, natom, ikind, qs_kind, harmonics, &
                                  rho0_h_tot)

      TYPE(gapw_control_type), POINTER                   :: gapw_control
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(rho0_atom_type), DIMENSION(:), POINTER        :: rho0_atom_set
      TYPE(rho0_mpole_type), POINTER                     :: rho0_mp
      INTEGER, DIMENSION(:), INTENT(IN)                  :: a_list
      TYPE(grid_atom_type), INTENT(IN)                   :: g_atom
      LOGICAL, INTENT(IN)                                :: paw_atom
      INTEGER, INTENT(IN)                                :: natom, ikind
      TYPE(qs_kind_type), INTENT(IN)                     :: qs_kind
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      REAL(dp), INTENT(INOUT)                            :: rho0_h_tot

      CHARACTER(len=*), PARAMETER :: routineN = 'calculate_rho0_atom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iat, iatom, ic, ico, ir, is, &
                                                            iso, ispin, l, lmax0, lshell, lx, ly, &
                                                            lz, nr, nsotot, nspins
      REAL(dp)                                           :: sum1
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: cpc_ah, cpc_as
      REAL(dp), DIMENSION(:), POINTER                    :: norm_g0l_h
      REAL(dp), DIMENSION(:, :), POINTER                 :: g0_h, vg0_h
      TYPE(mpole_gau_overlap), POINTER                   :: mpole_gau
      TYPE(mpole_rho_atom), POINTER                      :: mpole_rho
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: cpc_h, cpc_s
      TYPE(rho_atom_type), POINTER                       :: rho_atom

      CALL timeset(routineN, handle)

      NULLIFY (mpole_gau)
      NULLIFY (mpole_rho)
      NULLIFY (g0_h, vg0_h)
      NULLIFY (norm_g0l_h)

      CALL get_rho0_mpole(rho0_mpole=rho0_mp, ikind=ikind, &
                          l0_ikind=lmax0, mp_gau_ikind=mpole_gau, &
                          g0_h=g0_h, &
                          vg0_h=vg0_h, &
                          norm_g0l_h=norm_g0l_h)

      nr = g_atom%nr

! Set density coefficient to zero befor the calculation
      DO iat = 1, natom
         iatom = a_list(iat)
         rho0_atom_set(iatom)%rho0_rad_h%r_coef = 0.0_dp
         rho0_mp%mp_rho(iatom)%Qlm_tot = 0.0_dp
         rho0_mp%mp_rho(iatom)%Qlm_tot(1) = rho0_mp%mp_rho(iatom)%Qlm_z
         rho0_mp%mp_rho(iatom)%Q0 = 0.0_dp
      ENDDO

      IF (.NOT. (.NOT. paw_atom .AND. gapw_control%nopaw_as_gpw)) THEN
         DO iat = 1, natom
            iatom = a_list(iat)
            mpole_rho => rho0_mp%mp_rho(iatom)
            rho_atom => rho_atom_set(iatom)

            IF (paw_atom) THEN
               NULLIFY (cpc_h, cpc_s)
               CALL get_rho_atom(rho_atom=rho_atom, cpc_h=cpc_h, cpc_s=cpc_s)
               nspins = SIZE(cpc_h)
               nsotot = SIZE(mpole_gau%Qlm_gg, 1)
               ALLOCATE (cpc_ah(nsotot, nsotot, nspins))
               cpc_ah = 0._dp
               ALLOCATE (cpc_as(nsotot, nsotot, nspins))
               cpc_as = 0._dp
               DO ispin = 1, nspins
                  CALL prj_scatter(cpc_h(ispin)%r_coef, cpc_ah(:, :, ispin), qs_kind)
                  CALL prj_scatter(cpc_s(ispin)%r_coef, cpc_as(:, :, ispin), qs_kind)
               END DO
            END IF

            ! Total charge (hard-soft) at atom
            IF (paw_atom) THEN
               DO ispin = 1, nspins
                  mpole_rho%Q0(ispin) = (trace_r_AxB(mpole_gau%Qlm_gg(:, :, 1), nsotot, &
                                                     cpc_ah(:, :, ispin), nsotot, nsotot, nsotot) &
                                         -trace_r_AxB(mpole_gau%Qlm_gg(:, :, 1), nsotot, &
                                                      cpc_as(:, :, ispin), nsotot, nsotot, nsotot))/SQRT(fourpi)
               END DO
            END IF
            ! Multipoles of local charge distribution
            DO iso = 1, nsoset(lmax0)
               l = indso(1, iso)
               IF (paw_atom) THEN
                  mpole_rho%Qlm_h(iso) = 0.0_dp
                  mpole_rho%Qlm_s(iso) = 0.0_dp

                  DO ispin = 1, nspins
                     mpole_rho%Qlm_h(iso) = mpole_rho%Qlm_h(iso)+ &
                                            trace_r_AxB(mpole_gau%Qlm_gg(:, :, iso), nsotot, &
                                                        cpc_ah(:, :, ispin), nsotot, nsotot, nsotot)
                     mpole_rho%Qlm_s(iso) = mpole_rho%Qlm_s(iso)+ &
                                            trace_r_AxB(mpole_gau%Qlm_gg(:, :, iso), nsotot, &
                                                        cpc_as(:, :, ispin), nsotot, nsotot, nsotot)
                  END DO ! ispin

                  mpole_rho%Qlm_tot(iso) = mpole_rho%Qlm_tot(iso)+ &
                                           mpole_rho%Qlm_h(iso)-mpole_rho%Qlm_s(iso)
               END IF

               rho0_atom_set(iatom)%rho0_rad_h%r_coef(1:nr, iso) = &
                  g0_h(1:nr, l)*mpole_rho%Qlm_tot(iso)
               rho0_atom_set(iatom)%vrho0_rad_h%r_coef(1:nr, iso) = &
                  vg0_h(1:nr, l)*mpole_rho%Qlm_tot(iso)

               sum1 = 0.0_dp
               DO ir = 1, nr
                  sum1 = sum1+g_atom%wr(ir)* &
                         rho0_atom_set(iatom)%rho0_rad_h%r_coef(ir, iso)
               ENDDO
               rho0_h_tot = rho0_h_tot+sum1*harmonics%slm_int(iso)
            END DO ! iso
            IF (paw_atom) THEN
               DEALLOCATE (cpc_ah, cpc_as)
            END IF
         END DO ! iat
      END IF

!   Transform the coefficinets from spherical to cartesian
      IF (.NOT. paw_atom .AND. gapw_control%nopaw_as_gpw) THEN
         DO iat = 1, natom
            iatom = a_list(iat)
            mpole_rho => rho0_mp%mp_rho(iatom)

            DO lshell = 0, lmax0
               DO ic = 1, nco(lshell)
                  ico = ic+ncoset(lshell-1)
                  mpole_rho%Qlm_car(ico) = 0.0_dp
               END DO
            END DO
         END DO
      ELSE
         DO iat = 1, natom
            iatom = a_list(iat)
            mpole_rho => rho0_mp%mp_rho(iatom)

            DO lshell = 0, lmax0
               DO ic = 1, nco(lshell)
                  ico = ic+ncoset(lshell-1)
                  mpole_rho%Qlm_car(ico) = 0.0_dp
                  lx = indco(1, ico)
                  ly = indco(2, ico)
                  lz = indco(3, ico)

                  DO is = 1, nso(lshell)
                     iso = is+nsoset(lshell-1)

                     mpole_rho%Qlm_car(ico) = mpole_rho%Qlm_car(ico)+ &
                                              orbtramat(lshell)%c2s(is, ic)*mpole_rho%Qlm_tot(iso)* &
                                              norm_g0l_h(lshell) &
                                              /SQRT(dfac(2*lx-1)*dfac(2*ly-1)*dfac(2*lz-1)*fourpi/dfac(2*lshell+1))

                  END DO
               END DO
            END DO ! lshell
         END DO ! iat
      END IF
!MI get rid of full gapw

      CALL timestop(handle)

   END SUBROUTINE calculate_rho0_atom

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param gapw_control ...
!> \param tddft ...
!> \param tddft_local_rho_set ...
! **************************************************************************************************
   SUBROUTINE init_rho0(qs_env, gapw_control, &
                        tddft, tddft_local_rho_set)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(gapw_control_type), POINTER                   :: gapw_control
      LOGICAL, INTENT(IN), OPTIONAL                      :: tddft
      TYPE(local_rho_type), OPTIONAL, POINTER            :: tddft_local_rho_set

      CHARACTER(len=*), PARAMETER :: routineN = 'init_rho0', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: unit_str
      INTEGER :: handle, iat, iatom, ikind, l, l_rho1_max, laddg, lmaxg, maxl, maxnset, maxso, &
         nat, natom, nchan_c, nchan_s, nkind, nr, nset, nsotot, output_unit
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: my_tddft, paw_atom
      REAL(dp)                                           :: alpha_core, eps_Vrho0, max_rpgf0_s, &
                                                            radius, rc_min, rc_orb, &
                                                            total_rho_core_rspace
      REAL(KIND=dp)                                      :: zeff
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(rho0_atom_type), DIMENSION(:), POINTER        :: rho0_atom_set
      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole
      TYPE(rhoz_type), DIMENSION(:), POINTER             :: rhoz_set
      TYPE(section_vals_type), POINTER                   :: dft_section

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      logger => cp_get_default_logger()

      NULLIFY (qs_kind_set)
      NULLIFY (atomic_kind_set)
      NULLIFY (dft_control)
      NULLIFY (harmonics)
      NULLIFY (orb_basis)
      NULLIFY (rho0_mpole)
      NULLIFY (rho0_atom_set)
      NULLIFY (rhoz_set)

      my_tddft = .FALSE.
      IF (PRESENT(tddft)) my_tddft = tddft

      CALL get_qs_env(qs_env=qs_env, qs_kind_set=qs_kind_set, &
                      atomic_kind_set=atomic_kind_set, &
                      dft_control=dft_control)

      nkind = SIZE(atomic_kind_set)
      eps_Vrho0 = gapw_control%eps_Vrho0

!   Initialize rhoz total to zero
!   in gapw rhoz is calculated on local the lebedev grids
      total_rho_core_rspace = 0.0_dp

      CALL get_atomic_kind_set(atomic_kind_set, natom=natom)

!   Initialize the multipole and the compensation charge type
      CALL allocate_rho0_mpole(rho0_mpole)
      CALL allocate_rho0_atom(rho0_atom_set, natom)

!   Allocate the multipole set
      CALL allocate_multipoles(rho0_mpole%mp_rho, natom, rho0_mpole%mp_gau, nkind)

!   Allocate the core density on the radial grid for each kind: rhoz_set
      CALL allocate_rhoz(rhoz_set, nkind)

!   For each kind, determine the max l for the compensation charge density
      lmaxg = gapw_control%lmax_rho0
      laddg = gapw_control%ladd_rho0

      CALL reallocate(rho0_mpole%lmax0_kind, 1, nkind)

      rho0_mpole%lmax_0 = 0
      rc_min = 100.0_dp
      maxnset = 0
      DO ikind = 1, nkind
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=nat)
         CALL get_qs_kind(qs_kind_set(ikind), &
                          basis_set=orb_basis, &
                          ngrid_rad=nr, &
                          grid_atom=grid_atom, &
                          harmonics=harmonics, &
                          paw_atom=paw_atom, &
                          hard0_radius=rc_orb, &
                          zeff=zeff, &
                          alpha_core_charge=alpha_core)

         CALL get_gto_basis_set(gto_basis_set=orb_basis, &
                                maxl=maxl, &
                                maxso=maxso, nset=nset)

         maxnset = MAX(maxnset, nset)

         l_rho1_max = indso(1, harmonics%max_iso_not0)
         IF (paw_atom) THEN
            rho0_mpole%lmax0_kind(ikind) = MIN(2*maxl, l_rho1_max, maxl+laddg, lmaxg)
         ELSE
            rho0_mpole%lmax0_kind(ikind) = 0
         END IF

         CALL set_qs_kind(qs_kind_set(ikind), lmax_rho0=rho0_mpole%lmax0_kind(ikind))

         IF (gapw_control%lrho1_eq_lrho0) harmonics%max_iso_not0 = &
            nsoset(rho0_mpole%lmax0_kind(ikind))

         rho0_mpole%lmax_0 = MAX(rho0_mpole%lmax_0, rho0_mpole%lmax0_kind(ikind))
         rc_min = MIN(rc_min, rc_orb)

         nchan_s = nsoset(rho0_mpole%lmax0_kind(ikind))
         nchan_c = ncoset(rho0_mpole%lmax0_kind(ikind))
         nsotot = maxso*nset

         DO iat = 1, nat
            iatom = atom_list(iat)
!        Allocate the multipole for rho1_h rho1_s and rho_z
            CALL initialize_mpole_rho(rho0_mpole%mp_rho(iatom), nchan_s, nchan_c, zeff, my_tddft)
!        Allocate the radial part of rho0_h and rho0_s
!        This is calculated on the radial grid centered at the atomic position
            CALL allocate_rho0_atom_rad(rho0_atom_set(iatom), nr, nchan_s)
         END DO
!
         IF (paw_atom) THEN
!        Calculate multipoles given by the product of 2 primitives Qlm_gg
            CALL calculate_mpole_gau(rho0_mpole%mp_gau(ikind), &
                                     orb_basis, harmonics, nchan_s, nsotot)
         END IF

!     Calculate the core density rhoz
!                  exp(-alpha_c**2 r**2)Z(alpha_c**2/pi)**(3/2)
!     on the logarithmic radial grid
!     WARNING: alpha_core_charge = alpha_c**2
         CALL calculate_rhoz(rhoz_set(ikind), grid_atom, alpha_core, zeff, &
                             nat, total_rho_core_rspace, harmonics)
      END DO ! ikind
      total_rho_core_rspace = -total_rho_core_rspace

      IF (gapw_control%alpha0_hard_from_input) THEN
!   The Exponent for the compensation charge rho0_hard is read from input
         rho0_mpole%zet0_h = gapw_control%alpha0_hard
      ELSE
!   Calculate the exponent for the compensation charge rho0_hard
         rho0_mpole%zet0_h = 0.1_dp
         DO
            radius = exp_radius(rho0_mpole%lmax_0, rho0_mpole%zet0_h, eps_Vrho0, 1.0_dp)
            IF (radius <= rc_min) EXIT
            rho0_mpole%zet0_h = rho0_mpole%zet0_h+0.1_dp
         END DO

      END IF

!   Allocate and calculate the normalization factors for g0_lm_h and g0_lm_s
      CALL reallocate(rho0_mpole%norm_g0l_h, 0, rho0_mpole%lmax_0)
      DO l = 0, rho0_mpole%lmax_0
         rho0_mpole%norm_g0l_h(l) = (2._dp*l+1._dp)/ &
                                    (fourpi*gaussint_sph(rho0_mpole%zet0_h, 2*l))
      END DO

!   Allocate and Initialize the g0 gaussians used to build the compensation density
!   and calculate the interaction radii
      max_rpgf0_s = 0.0_dp
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), grid_atom=grid_atom)
         CALL calculate_g0(rho0_mpole, grid_atom, ikind)
         CALL interaction_radii_g0(rho0_mpole, ikind, eps_Vrho0, max_rpgf0_s)
      END DO
      rho0_mpole%max_rpgf0_s = max_rpgf0_s

      IF (.NOT. my_tddft) THEN
         CALL set_qs_env(qs_env=qs_env, rho0_mpole=rho0_mpole, &
                         rho0_atom_set=rho0_atom_set, &
                         rhoz_set=rhoz_set, &
                         rhoz_tot=total_rho_core_rspace)
      ELSE
         tddft_local_rho_set%rho0_mpole => rho0_mpole
         tddft_local_rho_set%rho0_atom_set => rho0_atom_set
         tddft_local_rho_set%rhoz_set => rhoz_set
         tddft_local_rho_set%rhoz_tot = total_rho_core_rspace
         CALL rho0_s_grid_create(qs_env, rho0_mpole, .TRUE.)
      END IF

      dft_section => section_vals_get_subs_vals(qs_env%input, "DFT")
      output_unit = cp_print_key_unit_nr(logger, dft_section, "PRINT%GAPW%RHO0_INFORMATION", &
                                         extension=".Log")
      CALL section_vals_val_get(dft_section, "PRINT%GAPW%RHO0_INFORMATION%UNIT", c_val=unit_str)
      IF (output_unit > 0) THEN
         CALL write_rho0_info(rho0_mpole, unit_str, output_unit)
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, dft_section, &
                                        "PRINT%GAPW%RHO0_INFORMATION")

      CALL timestop(handle)

   END SUBROUTINE init_rho0

! **************************************************************************************************
!> \brief ...
!> \param rho0_mpole ...
!> \param ik ...
!> \param eps_Vrho0 ...
!> \param max_rpgf0_s ...
! **************************************************************************************************
   SUBROUTINE interaction_radii_g0(rho0_mpole, ik, eps_Vrho0, max_rpgf0_s)

      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole
      INTEGER, INTENT(IN)                                :: ik
      REAL(dp), INTENT(IN)                               :: eps_Vrho0
      REAL(dp), INTENT(INOUT)                            :: max_rpgf0_s

      CHARACTER(len=*), PARAMETER :: routineN = 'interaction_radii_g0', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: l, lmax
      REAL(dp)                                           :: r_h, z0_h
      REAL(dp), DIMENSION(:), POINTER                    :: ng0_h

      CALL get_rho0_mpole(rho0_mpole, ikind=ik, l0_ikind=lmax, &
                          zet0_h=z0_h, norm_g0l_h=ng0_h)
      r_h = 0.0_dp
      DO l = 0, lmax
         r_h = MAX(r_h, exp_radius(l, z0_h, eps_Vrho0, ng0_h(l)))
      END DO

      rho0_mpole%mp_gau(ik)%rpgf0_h = r_h
      rho0_mpole%mp_gau(ik)%rpgf0_s = r_h
      max_rpgf0_s = MAX(max_rpgf0_s, r_h)

   END SUBROUTINE interaction_radii_g0

! **************************************************************************************************

END MODULE qs_rho0_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
MODULE qs_gapw_densities
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cp_control_types,                ONLY: dft_control_type,&
                                              gapw_control_type
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE qs_charges_types,                ONLY: qs_charges_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_local_rho_types,              ONLY: local_rho_type
   USE qs_rho0_ggrid,                   ONLY: put_rho0_on_grid
   USE qs_rho0_methods,                 ONLY: calculate_rho0_atom
   USE qs_rho0_types,                   ONLY: rho0_atom_type,&
                                              rho0_mpole_type
   USE qs_rho_atom_methods,             ONLY: calculate_rho_atom
   USE qs_rho_atom_types,               ONLY: rho_atom_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_gapw_densities'

   PUBLIC :: prepare_gapw_den

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param local_rho_set ...
!> \param do_rho0 ...
! **************************************************************************************************
   SUBROUTINE prepare_gapw_den(qs_env, local_rho_set, do_rho0)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(local_rho_type), OPTIONAL, POINTER            :: local_rho_set
      LOGICAL, INTENT(IN), OPTIONAL                      :: do_rho0

      CHARACTER(len=*), PARAMETER :: routineN = 'prepare_gapw_den', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ikind, ispin, natom, nspins, &
                                                            output_unit
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: my_do_rho0, paw_atom
      REAL(dp)                                           :: rho0_h_tot, tot_rs_int
      REAL(dp), DIMENSION(:), POINTER                    :: rho1_h_tot, rho1_s_tot
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(gapw_control_type), POINTER                   :: gapw_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(qs_charges_type), POINTER                     :: qs_charges
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(rho0_atom_type), DIMENSION(:), POINTER        :: rho0_atom_set
      TYPE(rho0_mpole_type), POINTER                     :: rho0_mpole
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set

      CALL timeset(routineN, handle)

      NULLIFY (atomic_kind_set)
      NULLIFY (qs_kind_set)
      NULLIFY (dft_control)
      NULLIFY (gapw_control)
      NULLIFY (para_env)
      NULLIFY (atom_list)
      NULLIFY (grid_atom)
      NULLIFY (rho0_mpole)
      NULLIFY (qs_charges)
      NULLIFY (rho1_h_tot, rho1_s_tot)
      NULLIFY (rho_atom_set)
      NULLIFY (rho0_atom_set)

      my_do_rho0 = .TRUE.
      IF (PRESENT(do_rho0)) my_do_rho0 = do_rho0

      output_unit = cp_logger_get_default_io_unit()

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, &
                      para_env=para_env, &
                      qs_charges=qs_charges, &
                      qs_kind_set=qs_kind_set, &
                      atomic_kind_set=atomic_kind_set, &
                      rho0_mpole=rho0_mpole, &
                      rho_atom_set=rho_atom_set, &
                      rho0_atom_set=rho0_atom_set)

      gapw_control => dft_control%qs_control%gapw_control

      IF (PRESENT(local_rho_set)) THEN
         rho_atom_set => local_rho_set%rho_atom_set
         IF (my_do_rho0) THEN
            rho0_mpole => local_rho_set%rho0_mpole
            rho0_atom_set => local_rho_set%rho0_atom_set
         END IF
      END IF

      nspins = dft_control%nspins

      rho0_h_tot = 0.0_dp
      ALLOCATE (rho1_h_tot(1:nspins), rho1_s_tot(1:nspins))
      rho1_h_tot = 0.0_dp
      rho1_s_tot = 0.0_dp

      DO ikind = 1, SIZE(atomic_kind_set)
         NULLIFY (grid_atom, harmonics)
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=natom)
         CALL get_qs_kind(qs_kind_set(ikind), grid_atom=grid_atom, &
                          paw_atom=paw_atom, harmonics=harmonics)

!     Calculate rho1_h and rho1_s on the radial grids centered on the atomic position
         IF (paw_atom) &
            CALL calculate_rho_atom(para_env, rho_atom_set, qs_kind_set(ikind), &
                                    atom_list, grid_atom, natom, nspins, rho1_h_tot, rho1_s_tot)

!     Calculate rho0_h and rho0_s on the radial grids centered on the atomic position
         IF (my_do_rho0) &
            CALL calculate_rho0_atom(gapw_control, rho_atom_set, rho0_atom_set, rho0_mpole, &
                                     atom_list, grid_atom, paw_atom, natom, ikind, qs_kind_set(ikind), &
                                     harmonics, rho0_h_tot)

      ENDDO

      CALL mp_sum(rho1_h_tot, para_env%group)
      CALL mp_sum(rho1_s_tot, para_env%group)
      DO ispin = 1, nspins
         qs_charges%total_rho1_hard(ispin) = -rho1_h_tot(ispin)
         qs_charges%total_rho1_soft(ispin) = -rho1_s_tot(ispin)
      END DO

      IF (my_do_rho0) THEN
         rho0_mpole%total_rho0_h = -rho0_h_tot
!      Put the rho0_soft on the global grid
         CALL put_rho0_on_grid(qs_env, rho0_mpole, tot_rs_int)
         IF (ABS(rho0_h_tot) .GE. 1.0E-5_dp) THEN
            IF (ABS(1.0_dp-ABS(tot_rs_int/rho0_h_tot)) .GT. 1.0E-3_dp) THEN
               IF (output_unit > 0) THEN
                  WRITE (output_unit, '(/,72("*"))')
                  WRITE (output_unit, '(T2,A,T66,1E20.8)') &
                     "WARNING: rho0 calculated on the local grid is  :", -rho0_h_tot, &
                     "         rho0 calculated on the global grid is :", tot_rs_int
                  WRITE (output_unit, '(T2,A)') &
                     "         bad integration"
                  WRITE (output_unit, '(72("*"),/)')
               END IF
            END IF
         END IF
         qs_charges%total_rho0_soft_rspace = tot_rs_int
         qs_charges%total_rho0_hard_lebedev = rho0_h_tot
      ELSE
         qs_charges%total_rho0_hard_lebedev = 0.0_dp
      END IF

      DEALLOCATE (rho1_h_tot, rho1_s_tot)

      CALL timestop(handle)

   END SUBROUTINE prepare_gapw_den

END MODULE qs_gapw_densities
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Contains ADMM methods which only require the density matrix
!> \par History
!>      11.2014 created [Ole Schuett]
!> \author Ole Schuett
! **************************************************************************************************
MODULE admm_dm_methods
   USE admm_dm_types,                   ONLY: admm_dm_type,&
                                              mcweeny_history_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_copy, cp_dbcsr_create, cp_dbcsr_deallocate_matrix_set, &
        cp_dbcsr_frobenius_norm, cp_dbcsr_get_block_p, cp_dbcsr_init, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_release, &
        cp_dbcsr_scale, cp_dbcsr_set, cp_dbcsr_type
   USE cp_log_handling,                 ONLY: cp_logger_get_default_unit_nr
   USE input_constants,                 ONLY: do_admm_basis_projection,&
                                              do_admm_blocked_projection
   USE iterate_matrix,                  ONLY: invert_Hotelling
   USE kinds,                           ONLY: dp
   USE pw_types,                        ONLY: pw_p_type
   USE qs_collocate_density,            ONLY: calculate_rho_elec
   USE qs_ks_types,                     ONLY: get_ks_env,&
                                              qs_ks_env_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_set,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: admm_dm_calc_rho_aux, admm_dm_merge_ks_matrix

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'admm_dm_methods'

CONTAINS

! **************************************************************************************************
!> \brief Entry methods: Calculates auxilary density matrix from primary one.
!> \param ks_env ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE admm_dm_calc_rho_aux(ks_env)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      CHARACTER(len=*), PARAMETER :: routineN = 'admm_dm_calc_rho_aux', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(admm_dm_type), POINTER                        :: admm_dm

      NULLIFY (admm_dm)
      CALL timeset(routineN, handle)
      CALL get_ks_env(ks_env, admm_dm=admm_dm)

      SELECT CASE (admm_dm%method)
      CASE (do_admm_basis_projection)
         CALL map_dm_projection(ks_env)

      CASE (do_admm_blocked_projection)
         CALL map_dm_blocked(ks_env)

      CASE DEFAULT
         CALL cp__b("admm_dm_methods.F",71,"admm_dm_calc_rho_aux: unknown method")
      END SELECT

      IF (admm_dm%purify) &
         CALL purify_mcweeny(ks_env)

      CALL update_rho_aux(ks_env)

      CALL timestop(handle)
   END SUBROUTINE admm_dm_calc_rho_aux

! **************************************************************************************************
!> \brief Entry methods: Merges auxilary Kohn-Sham matrix into primary one.
!> \param ks_env ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE admm_dm_merge_ks_matrix(ks_env)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'admm_dm_merge_ks_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(admm_dm_type), POINTER                        :: admm_dm
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks_merge

      CALL timeset(routineN, handle)
      NULLIFY (admm_dm, matrix_ks_merge)

      CALL get_ks_env(ks_env, admm_dm=admm_dm)

      IF (admm_dm%purify) THEN
         CALL revert_purify_mcweeny(ks_env, matrix_ks_merge)
      ELSE
         CALL get_ks_env(ks_env, matrix_ks_aux_fit=matrix_ks_merge)
      ENDIF

      SELECT CASE (admm_dm%method)
      CASE (do_admm_basis_projection)
         CALL merge_dm_projection(ks_env, matrix_ks_merge)

      CASE (do_admm_blocked_projection)
         CALL merge_dm_blocked(ks_env, matrix_ks_merge)

      CASE DEFAULT
         CALL cp__b("admm_dm_methods.F",116,"admm_dm_merge_ks_matrix: unknown method")
      END SELECT

      IF (admm_dm%purify) &
         CALL cp_dbcsr_deallocate_matrix_set(matrix_ks_merge)

      CALL timestop(handle)

   END SUBROUTINE admm_dm_merge_ks_matrix

! **************************************************************************************************
!> \brief Calculates auxilary density matrix via basis projection.
!> \param ks_env ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE map_dm_projection(ks_env)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      CHARACTER(len=*), PARAMETER :: routineN = 'map_dm_projection', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin
      LOGICAL                                            :: s_mstruct_changed
      TYPE(admm_dm_type), POINTER                        :: admm_dm
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s_aux, matrix_s_mixed, rho_ao, &
                                                            rho_ao_aux
      TYPE(cp_dbcsr_type)                                :: matrix_s_aux_inv, matrix_tmp
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_rho_type), POINTER                         :: rho, rho_aux

      NULLIFY (dft_control, admm_dm, matrix_s_aux, matrix_s_mixed, rho, rho_aux)
      NULLIFY (rho_ao, rho_ao_aux)

      CALL get_ks_env(ks_env, &
                      admm_dm=admm_dm, &
                      dft_control=dft_control, &
                      matrix_s_aux_fit=matrix_s_aux, &
                      matrix_s_aux_fit_vs_orb=matrix_s_mixed, &
                      s_mstruct_changed=s_mstruct_changed, &
                      rho=rho, &
                      rho_aux_fit=rho_aux)

      CALL qs_rho_get(rho, rho_ao=rho_ao)
      CALL qs_rho_get(rho_aux, rho_ao=rho_ao_aux)

      IF (s_mstruct_changed) THEN
         ! Calculate A = S_aux^(-1) * S_mixed
         CALL cp_dbcsr_init(matrix_s_aux_inv)
         CALL cp_dbcsr_create(matrix_s_aux_inv, template=matrix_s_aux(1)%matrix, matrix_type="N")
         CALL invert_Hotelling(matrix_s_aux_inv, matrix_s_aux(1)%matrix, admm_dm%eps_filter)

         IF (.NOT. ASSOCIATED(admm_dm%matrix_A)) THEN
            ALLOCATE (admm_dm%matrix_A)
            CALL cp_dbcsr_init(admm_dm%matrix_A)
            CALL cp_dbcsr_create(admm_dm%matrix_A, template=matrix_s_mixed(1)%matrix, matrix_type="N")
         ENDIF
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_s_aux_inv, matrix_s_mixed(1)%matrix, &
                                0.0_dp, admm_dm%matrix_A)
         CALL cp_dbcsr_release(matrix_s_aux_inv)
      ENDIF

      ! Calculate P_aux = A * P * A^T
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=admm_dm%matrix_A)
      DO ispin = 1, dft_control%nspins
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, admm_dm%matrix_A, rho_ao(ispin)%matrix, &
                                0.0_dp, matrix_tmp)
         CALL cp_dbcsr_multiply("N", "T", 1.0_dp, matrix_tmp, admm_dm%matrix_A, &
                                0.0_dp, rho_ao_aux(ispin)%matrix)
      END DO
      CALL cp_dbcsr_release(matrix_tmp)

   END SUBROUTINE map_dm_projection

! **************************************************************************************************
!> \brief Calculates auxilary density matrix via blocking.
!> \param ks_env ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE map_dm_blocked(ks_env)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      INTEGER                                            :: blk, iatom, ispin, jatom
      LOGICAL                                            :: found
      REAL(dp), DIMENSION(:, :), POINTER                 :: sparse_block, sparse_block_aux
      TYPE(admm_dm_type), POINTER                        :: admm_dm
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao, rho_ao_aux
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_rho_type), POINTER                         :: rho, rho_aux

      NULLIFY (dft_control, admm_dm, rho, rho_aux, rho_ao, rho_ao_aux)

      CALL get_ks_env(ks_env, &
                      admm_dm=admm_dm, &
                      dft_control=dft_control, &
                      rho=rho, &
                      rho_aux_fit=rho_aux)

      CALL qs_rho_get(rho, rho_ao=rho_ao)
      CALL qs_rho_get(rho_aux, rho_ao=rho_ao_aux)

      ! ** set blocked density matrix to 0
      DO ispin = 1, dft_control%nspins
         CALL cp_dbcsr_set(rho_ao_aux(ispin)%matrix, 0.0_dp)
         ! ** now loop through the list and copy corresponding blocks
         CALL cp_dbcsr_iterator_start(iter, rho_ao(ispin)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, iatom, jatom, sparse_block, blk)
            IF (admm_dm%block_map(iatom, jatom) == 1) THEN
               CALL cp_dbcsr_get_block_p(rho_ao_aux(ispin)%matrix, &
                                         row=iatom, col=jatom, BLOCK=sparse_block_aux, found=found)
               IF (found) &
                  sparse_block_aux = sparse_block
            END IF
         END DO
         CALL cp_dbcsr_iterator_stop(iter)
      ENDDO

   END SUBROUTINE map_dm_blocked

! **************************************************************************************************
!> \brief Call calculate_rho_elec() for auxilary density
!> \param ks_env ...
! **************************************************************************************************
   SUBROUTINE update_rho_aux(ks_env)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      INTEGER                                            :: ispin
      REAL(KIND=dp), DIMENSION(:), POINTER               :: tot_rho_r_aux
      TYPE(admm_dm_type), POINTER                        :: admm_dm
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: rho_ao_aux
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(pw_p_type), DIMENSION(:), POINTER             :: rho_g_aux, rho_r_aux
      TYPE(qs_rho_type), POINTER                         :: rho_aux

      NULLIFY (dft_control, admm_dm, rho_aux, rho_ao_aux, rho_r_aux, rho_g_aux, tot_rho_r_aux)

      CALL get_ks_env(ks_env, &
                      admm_dm=admm_dm, &
                      dft_control=dft_control, &
                      rho_aux_fit=rho_aux)

      CALL qs_rho_get(rho_aux, &
                      rho_ao=rho_ao_aux, &
                      rho_r=rho_r_aux, &
                      rho_g=rho_g_aux, &
                      tot_rho_r=tot_rho_r_aux)

      DO ispin = 1, dft_control%nspins
         CALL calculate_rho_elec(ks_env=ks_env, &
                                 matrix_p=rho_ao_aux(ispin)%matrix, &
                                 rho=rho_r_aux(ispin), &
                                 rho_gspace=rho_g_aux(ispin), &
                                 total_rho=tot_rho_r_aux(ispin), &
                                 soft_valid=.FALSE., &
                                 basis_type="AUX_FIT")
      END DO

      CALL qs_rho_set(rho_aux, rho_r_valid=.TRUE., rho_g_valid=.TRUE.)

   END SUBROUTINE update_rho_aux

! **************************************************************************************************
!> \brief Merges auxilary Kohn-Sham matrix via basis projection.
!> \param ks_env ...
!> \param matrix_ks_merge Input: The KS matrix to be merged
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE merge_dm_projection(ks_env, matrix_ks_merge)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks_merge

      INTEGER                                            :: ispin
      TYPE(admm_dm_type), POINTER                        :: admm_dm
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(cp_dbcsr_type)                                :: matrix_tmp
      TYPE(dft_control_type), POINTER                    :: dft_control

      NULLIFY (admm_dm, dft_control, matrix_ks)

      CALL get_ks_env(ks_env, &
                      admm_dm=admm_dm, &
                      dft_control=dft_control, &
                      matrix_ks=matrix_ks)

      ! Calculate K += A^T * K_aux * A
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=admm_dm%matrix_A, matrix_type="N")

      DO ispin = 1, dft_control%nspins
         CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_ks_merge(ispin)%matrix, admm_dm%matrix_A, &
                                0.0_dp, matrix_tmp)
         CALL cp_dbcsr_multiply("T", "N", 1.0_dp, admm_dm%matrix_A, matrix_tmp, &
                                1.0_dp, matrix_ks(ispin)%matrix)
      END DO

      CALL cp_dbcsr_release(matrix_tmp)

   END SUBROUTINE merge_dm_projection

! **************************************************************************************************
!> \brief Merges auxilary Kohn-Sham matrix via blocking.
!> \param ks_env ...
!> \param matrix_ks_merge Input: The KS matrix to be merged
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE merge_dm_blocked(ks_env, matrix_ks_merge)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks_merge

      INTEGER                                            :: blk, iatom, ispin, jatom
      REAL(dp), DIMENSION(:, :), POINTER                 :: sparse_block
      TYPE(admm_dm_type), POINTER                        :: admm_dm
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks
      TYPE(dft_control_type), POINTER                    :: dft_control

      NULLIFY (admm_dm, dft_control, matrix_ks)

      CALL get_ks_env(ks_env, &
                      admm_dm=admm_dm, &
                      dft_control=dft_control, &
                      matrix_ks=matrix_ks)

      DO ispin = 1, dft_control%nspins
         CALL cp_dbcsr_iterator_start(iter, matrix_ks_merge(ispin)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, iatom, jatom, sparse_block, blk)
            IF (admm_dm%block_map(iatom, jatom) == 0) &
               sparse_block = 0.0_dp
         END DO
         CALL cp_dbcsr_iterator_stop(iter)
         CALL cp_dbcsr_add(matrix_ks(ispin)%matrix, matrix_ks_merge(ispin)%matrix, 1.0_dp, 1.0_dp)
      ENDDO

   END SUBROUTINE merge_dm_blocked

! **************************************************************************************************
!> \brief Apply McWeeny purification to auxilary density matrix
!> \param ks_env ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE purify_mcweeny(ks_env)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'purify_mcweeny', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, istep, nspins, unit_nr
      REAL(KIND=dp)                                      :: frob_norm
      TYPE(admm_dm_type), POINTER                        :: admm_dm
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_s_aux_fit, rho_ao_aux
      TYPE(cp_dbcsr_type)                                :: matrix_ps, matrix_psp, matrix_test
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_p, matrix_s
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mcweeny_history_type), POINTER                :: history, new_hist_entry
      TYPE(qs_rho_type), POINTER                         :: rho_aux_fit

      CALL timeset(routineN, handle)
      NULLIFY (dft_control, admm_dm, matrix_s_aux_fit, rho_aux_fit, new_hist_entry, &
               matrix_p, matrix_s, rho_ao_aux)

      unit_nr = cp_logger_get_default_unit_nr()
      CALL get_ks_env(ks_env, &
                      dft_control=dft_control, &
                      admm_dm=admm_dm, &
                      matrix_s_aux_fit=matrix_s_aux_fit, &
                      rho_aux_fit=rho_aux_fit)

      CALL qs_rho_get(rho_aux_fit, rho_ao=rho_ao_aux)

      matrix_p => rho_ao_aux(1)%matrix
      CALL cp_dbcsr_init(matrix_PS)
      CALL cp_dbcsr_create(matrix_PS, template=matrix_p, matrix_type="N")
      CALL cp_dbcsr_init(matrix_PSP)
      CALL cp_dbcsr_create(matrix_PSP, template=matrix_p, matrix_type="S")
      CALL cp_dbcsr_init(matrix_test)
      CALL cp_dbcsr_create(matrix_test, template=matrix_p, matrix_type="S")

      nspins = dft_control%nspins
      DO ispin = 1, nspins
         matrix_p => rho_ao_aux(ispin)%matrix
         matrix_s => matrix_s_aux_fit(1)%matrix
         history => admm_dm%mcweeny_history(ispin)%p
         IF (ASSOCIATED(history)) CALL cp__b("admm_dm_methods.F",400,"purify_dm_mcweeny: history already associated")
         IF (nspins == 1) CALL cp_dbcsr_scale(matrix_p, 0.5_dp)

         DO istep = 1, admm_dm%mcweeny_max_steps
            ! allocate new element in linked list
            ALLOCATE (new_hist_entry)
            new_hist_entry%next => history
            history => new_hist_entry
            history%count = istep
            NULLIFY (new_hist_entry)
            CALL cp_dbcsr_init(history%m)
            CALL cp_dbcsr_create(history%m, template=matrix_p, matrix_type="N")
            CALL cp_dbcsr_copy(history%m, matrix_p, name="P from McWeeny")

            ! calc PS and PSP
            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_p, matrix_s, &
                                   0.0_dp, matrix_ps)

            CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_ps, matrix_p, &
                                   0.0_dp, matrix_psp)

            !test convergence
            CALL cp_dbcsr_copy(matrix_test, matrix_psp)
            CALL cp_dbcsr_add(matrix_test, matrix_p, 1.0_dp, -1.0_dp)
            frob_norm = cp_dbcsr_frobenius_norm(matrix_test)
            IF (unit_nr > 0) WRITE (unit_nr, '(t3,a,i5,a,f16.8)') "McWeeny-Step", istep, &
               ": Deviation of idempotency", frob_norm
            IF (frob_norm < 1000_dp*admm_dm%eps_filter .AND. istep > 1) EXIT

            ! build next P matrix
            CALL cp_dbcsr_copy(matrix_p, matrix_PSP, name="P from McWeeny")
            CALL cp_dbcsr_multiply("N", "N", -2.0_dp, matrix_PS, matrix_PSP, &
                                   3.0_dp, matrix_p)
         END DO
         admm_dm%mcweeny_history(ispin)%p => history
         IF (nspins == 1) CALL cp_dbcsr_scale(matrix_p, 2.0_dp)
      END DO

      ! clean up
      CALL cp_dbcsr_release(matrix_PS)
      CALL cp_dbcsr_release(matrix_PSP)
      CALL cp_dbcsr_release(matrix_test)
      CALL timestop(handle)
   END SUBROUTINE purify_mcweeny

! **************************************************************************************************
!> \brief Prepare auxilary KS-matrix for merge using reverse McWeeny
!> \param ks_env ...
!> \param matrix_ks_merge Output: The KS matrix for the merge
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE revert_purify_mcweeny(ks_env, matrix_ks_merge)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks_merge

      CHARACTER(LEN=*), PARAMETER :: routineN = 'revert_purify_mcweeny', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, nspins, unit_nr
      TYPE(admm_dm_type), POINTER                        :: admm_dm
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: matrix_ks, matrix_ks_aux_fit, &
                                                            matrix_s_aux_fit, &
                                                            matrix_s_aux_fit_vs_orb
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_k
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mcweeny_history_type), POINTER                :: history_curr, history_next

      CALL timeset(routineN, handle)
      unit_nr = cp_logger_get_default_unit_nr()
      NULLIFY (admm_dm, dft_control, matrix_ks, matrix_ks_aux_fit, &
               matrix_s_aux_fit, matrix_s_aux_fit_vs_orb, &
               history_next, history_curr, matrix_k)

      CALL get_ks_env(ks_env, &
                      admm_dm=admm_dm, &
                      dft_control=dft_control, &
                      matrix_ks=matrix_ks, &
                      matrix_ks_aux_fit=matrix_ks_aux_fit, &
                      matrix_s_aux_fit=matrix_s_aux_fit, &
                      matrix_s_aux_fit_vs_orb=matrix_s_aux_fit_vs_orb)

      nspins = dft_control%nspins
      ALLOCATE (matrix_ks_merge(nspins))

      DO ispin = 1, nspins
         ALLOCATE (matrix_ks_merge(ispin)%matrix)
         matrix_k => matrix_ks_merge(ispin)%matrix
         CALL cp_dbcsr_init(matrix_k)
         CALL cp_dbcsr_copy(matrix_k, matrix_ks_aux_fit(ispin)%matrix, name="K")
         history_curr => admm_dm%mcweeny_history(ispin)%p
         NULLIFY (admm_dm%mcweeny_history(ispin)%p)

         ! reverse McWeeny iteration
         DO WHILE (ASSOCIATED(history_curr))
            IF (unit_nr > 0) WRITE (unit_nr, '(t3,a,i5)') "Reverse McWeeny-Step ", history_curr%count
            CALL reverse_mcweeny_step(matrix_k=matrix_k, &
                                      matrix_s=matrix_s_aux_fit(1)%matrix, &
                                      matrix_p=history_curr%m)
            CALL cp_dbcsr_release(history_curr%m)
            history_next => history_curr%next
            DEALLOCATE (history_curr)
            history_curr => history_next
            NULLIFY (history_next)
         END DO

      END DO

      ! clean up
      CALL timestop(handle)

   END SUBROUTINE revert_purify_mcweeny

! **************************************************************************************************
!> \brief Multiply matrix_k with partial derivative of McWeeny by reversing it.
!> \param matrix_k ...
!> \param matrix_s ...
!> \param matrix_p ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE reverse_mcweeny_step(matrix_k, matrix_s, matrix_p)
      TYPE(cp_dbcsr_type)                                :: matrix_k, matrix_s, matrix_p

      CHARACTER(LEN=*), PARAMETER :: routineN = 'reverse_mcweeny_step', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(cp_dbcsr_type)                                :: matrix_ps, matrix_sp, matrix_sum, &
                                                            matrix_tmp

      CALL timeset(routineN, handle)
      CALL cp_dbcsr_init(matrix_ps)
      CALL cp_dbcsr_create(matrix_ps, template=matrix_p, matrix_type="N")
      CALL cp_dbcsr_init(matrix_sp)
      CALL cp_dbcsr_create(matrix_sp, template=matrix_p, matrix_type="N")
      CALL cp_dbcsr_init(matrix_tmp)
      CALL cp_dbcsr_create(matrix_tmp, template=matrix_p, matrix_type="N")
      CALL cp_dbcsr_init(matrix_sum)
      CALL cp_dbcsr_create(matrix_sum, template=matrix_p, matrix_type="N")

      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_p, matrix_s, &
                             0.0_dp, matrix_ps)
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_s, matrix_p, &
                             0.0_dp, matrix_sp)

      !TODO: can we exploid more symmetry?
      CALL cp_dbcsr_multiply("N", "N", 3.0_dp, matrix_k, matrix_ps, &
                             0.0_dp, matrix_sum)
      CALL cp_dbcsr_multiply("N", "N", 3.0_dp, matrix_sp, matrix_k, &
                             1.0_dp, matrix_sum)

      !matrix_tmp = KPS
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_k, matrix_ps, &
                             0.0_dp, matrix_tmp)
      CALL cp_dbcsr_multiply("N", "N", -2.0_dp, matrix_tmp, matrix_ps, &
                             1.0_dp, matrix_sum)
      CALL cp_dbcsr_multiply("N", "N", -2.0_dp, matrix_sp, matrix_tmp, &
                             1.0_dp, matrix_sum)

      !matrix_tmp = SPK
      CALL cp_dbcsr_multiply("N", "N", 1.0_dp, matrix_sp, matrix_k, &
                             0.0_dp, matrix_tmp)
      CALL cp_dbcsr_multiply("N", "N", -2.0_dp, matrix_sp, matrix_tmp, &
                             1.0_dp, matrix_sum)

      ! overwrite matrix_k
      CALL cp_dbcsr_copy(matrix_k, matrix_sum, name="K from reverse McWeeny")

      ! clean up
      CALL cp_dbcsr_release(matrix_sum)
      CALL cp_dbcsr_release(matrix_tmp)
      CALL cp_dbcsr_release(matrix_ps)
      CALL cp_dbcsr_release(matrix_sp)
      CALL timestop(handle)
   END SUBROUTINE reverse_mcweeny_step

END MODULE admm_dm_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routines that build the integrals of the Vxc potential calculated
!>      for the atomic density in the basis set of spherical primitives
! **************************************************************************************************
MODULE qs_vxc_atom
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: tddfpt_excitations,&
                                              tddfpt_triplet,&
                                              xc_none
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp,&
                                              dp_size
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_sum
   USE orbital_pointers,                ONLY: indso,&
                                              nsoset,&
                                              soset
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: get_none0_cg_list,&
                                              harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_linres_types,                 ONLY: nablavks_atom_type
   USE qs_p_env_types,                  ONLY: qs_p_env_type
   USE qs_rho_atom_types,               ONLY: get_rho_atom,&
                                              rho_atom_coeff,&
                                              rho_atom_type
   USE termination,                     ONLY: stop_memory
   USE util,                            ONLY: get_limit
   USE xc_atom,                         ONLY: fill_rho_set,&
                                              vxc_of_r_new,&
                                              xc_2nd_deriv_of_r,&
                                              xc_rho_set_atom_update
   USE xc_derivative_set_types,         ONLY: xc_derivative_set_type,&
                                              xc_dset_create,&
                                              xc_dset_release,&
                                              xc_dset_zero_all
   USE xc_derivatives,                  ONLY: xc_functionals_get_needs
   USE xc_rho_cflags_types,             ONLY: xc_rho_cflags_type
   USE xc_rho_set_types,                ONLY: xc_rho_set_create,&
                                              xc_rho_set_release,&
                                              xc_rho_set_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_vxc_atom'

   PUBLIC :: calculate_vxc_atom, &
             calculate_xc_2nd_deriv_atom, &
             calc_rho_angular, &
             gaVxcgb_noGC

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param energy_only ...
!> \param gradient_atom_set ...
!> \param adiabatic_rescale_factor ...
! **************************************************************************************************
   SUBROUTINE calculate_vxc_atom(qs_env, energy_only, gradient_atom_set, adiabatic_rescale_factor)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(IN)                                :: energy_only
      TYPE(nablavks_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: gradient_atom_set
      REAL(dp), INTENT(IN), OPTIONAL                     :: adiabatic_rescale_factor

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_vxc_atom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bo(2), handle, ia, iat, iatom, idir, &
                                                            ikind, ir, ispin, mepos, myfun, na, &
                                                            natom, nr, nspins, num_pe
      INTEGER, DIMENSION(2, 3)                           :: bounds
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: epr_xc, gradient_f, lsd, paw_atom, tau_f
      REAL(dp)                                           :: density_cut, exc_h, exc_s, gradient_cut, &
                                                            my_adiabatic_rescale_factor, tau_cut
      REAL(dp), DIMENSION(:, :), POINTER                 :: rho_h, rho_s, tau_h, tau_s, weight
      REAL(dp), DIMENSION(:, :, :), POINTER              :: vtau_h, vtau_s, vxc_h, vxc_s
      REAL(dp), DIMENSION(:, :, :, :), POINTER           :: drho_h, drho_s, vxg_h, vxg_s
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_kind_type), POINTER                        :: qs_kind
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: dr_h, dr_s, r_h, r_s
      TYPE(rho_atom_coeff), DIMENSION(:, :), POINTER     :: r_h_d, r_s_d, trho_h, trho_s
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      TYPE(section_vals_type), POINTER                   :: input, xc_fun_section, xc_section
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_rho_cflags_type)                           :: needs
      TYPE(xc_rho_set_type), POINTER                     :: rho_set_h, rho_set_s

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (atom_list)
      NULLIFY (qs_kind_set)
      NULLIFY (atomic_kind_set)
      NULLIFY (qs_kind)
      NULLIFY (deriv_set)
      NULLIFY (energy)
      NULLIFY (grid_atom)
      NULLIFY (harmonics)
      NULLIFY (input)
      NULLIFY (para_env)
      NULLIFY (rho_atom)
      NULLIFY (rho_atom_set)
      NULLIFY (rho_set_h)
      NULLIFY (rho_set_s)

      epr_xc = .FALSE.
      IF (PRESENT(gradient_atom_set)) THEN
         epr_xc = .TRUE.
      END IF

      IF (PRESENT(adiabatic_rescale_factor)) THEN
         my_adiabatic_rescale_factor = adiabatic_rescale_factor
      ELSE
         my_adiabatic_rescale_factor = 1.0_dp
      END IF

      CALL get_qs_env(qs_env=qs_env, &
                      dft_control=dft_control, &
                      para_env=para_env, &
                      energy=energy, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      input=input, &
                      rho_atom_set=rho_atom_set)

      IF (epr_xc) THEN
         xc_section => section_vals_get_subs_vals(input, &
                                                  "PROPERTIES%LINRES%EPR%PRINT%G_TENSOR%XC")
      ELSE
         xc_section => section_vals_get_subs_vals(input, "DFT%XC")
      END IF
      xc_fun_section => section_vals_get_subs_vals(xc_section, "XC_FUNCTIONAL")
      CALL section_vals_val_get(xc_fun_section, "_SECTION_PARAMETERS_", &
                                i_val=myfun)

      IF (myfun == xc_none) THEN
         energy%exc1 = 0.0_dp
         rho_atom_set(:)%exc_h = 0.0_dp
         rho_atom_set(:)%exc_s = 0.0_dp
      ELSE
         CALL section_vals_val_get(xc_section, "DENSITY_CUTOFF", &
                                   r_val=density_cut)
         CALL section_vals_val_get(xc_section, "GRADIENT_CUTOFF", &
                                   r_val=gradient_cut)
         CALL section_vals_val_get(xc_section, "TAU_CUTOFF", &
                                   r_val=tau_cut)

         lsd = dft_control%lsd
         nspins = dft_control%nspins
         needs = xc_functionals_get_needs(xc_fun_section, &
                                          lsd=lsd, &
                                          add_basic_components=.TRUE.)

         ! whatever the xc, if epr_xc, drho_spin is needed
         IF (epr_xc) needs%drho_spin = .TRUE.

         gradient_f = (needs%drho .OR. needs%drho_spin)
         tau_f = (needs%tau .OR. needs%tau_spin)

         ! Initialize energy contribution from the one center XC terms to zero
         energy%exc1 = 0.0_dp

         ! Nullify some pointers for work-arrays
         NULLIFY (rho_h, drho_h, rho_s, drho_s, weight)
         NULLIFY (vxc_h, vxc_s, vxg_h, vxg_s)
         NULLIFY (tau_h, tau_s)
         NULLIFY (vtau_h, vtau_s)

         ! Here starts the loop over all the atoms

         DO ikind = 1, SIZE(atomic_kind_set)
            CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=natom)
            CALL get_qs_kind(qs_kind_set(ikind), paw_atom=paw_atom, &
                             harmonics=harmonics, grid_atom=grid_atom)

            IF (.NOT. paw_atom) CYCLE

            nr = grid_atom%nr
            na = grid_atom%ng_sphere

            ! Prepare the structures needed to calculate and store the xc derivatives

            ! Array dimension: here anly one dimensional arrays are used,
            ! i.e. only the first column of deriv_data is read.
            ! The other to dimensions  are set to size equal 1
            bounds(1:2, 1:3) = 1
            bounds(2, 1) = na
            bounds(2, 2) = nr

            ! create a place where to put the derivatives
            CALL xc_dset_create(deriv_set, local_bounds=bounds)
            ! create the place where to store the argument for the functionals
            CALL xc_rho_set_create(rho_set_h, bounds, rho_cutoff=density_cut, &
                                   drho_cutoff=gradient_cut, tau_cutoff=tau_cut)
            CALL xc_rho_set_create(rho_set_s, bounds, rho_cutoff=density_cut, &
                                   drho_cutoff=gradient_cut, tau_cutoff=tau_cut)

            ! allocate the required 3d arrays where to store rho and drho
            CALL xc_rho_set_atom_update(rho_set_h, needs, nspins, bounds)
            CALL xc_rho_set_atom_update(rho_set_s, needs, nspins, bounds)

            CALL reallocate(rho_h, 1, na, 1, nspins)
            CALL reallocate(rho_s, 1, na, 1, nspins)
            weight => grid_atom%weight
            CALL reallocate(vxc_h, 1, na, 1, nr, 1, nspins)
            CALL reallocate(vxc_s, 1, na, 1, nr, 1, nspins)

            IF (gradient_f) THEN
               CALL reallocate(drho_h, 1, 4, 1, na, 1, nr, 1, nspins)
               CALL reallocate(drho_s, 1, 4, 1, na, 1, nr, 1, nspins)
               CALL reallocate(vxg_h, 1, 3, 1, na, 1, nr, 1, nspins)
               CALL reallocate(vxg_s, 1, 3, 1, na, 1, nr, 1, nspins)
            END IF

            IF (tau_f) THEN
               CALL reallocate(tau_h, 1, na, 1, nspins)
               CALL reallocate(tau_s, 1, na, 1, nspins)
               CALL reallocate(vtau_h, 1, na, 1, nr, 1, nspins)
               CALL reallocate(vtau_s, 1, na, 1, nr, 1, nspins)
            END IF

            ! Distribute the atoms of this kind

            num_pe = para_env%num_pe
            mepos = para_env%mepos
            bo = get_limit(natom, num_pe, mepos)

            DO iat = bo(1), bo(2)
               iatom = atom_list(iat)

               rho_atom_set(iatom)%exc_h = 0.0_dp
               rho_atom_set(iatom)%exc_s = 0.0_dp

               rho_atom => rho_atom_set(iatom)
               IF (gradient_f) THEN
                  NULLIFY (r_h, r_s, dr_h, dr_s, r_h_d, r_s_d)
                  CALL get_rho_atom(rho_atom=rho_atom, rho_rad_h=r_h, &
                                    rho_rad_s=r_s, drho_rad_h=dr_h, &
                                    drho_rad_s=dr_s, rho_rad_h_d=r_h_d, &
                                    rho_rad_s_d=r_s_d)
                  drho_h = 0.0_dp
                  drho_s = 0.0_dp
               ELSE
                  NULLIFY (r_h, r_s)
                  CALL get_rho_atom(rho_atom=rho_atom, rho_rad_h=r_h, rho_rad_s=r_s)
               END IF
               IF (tau_f) THEN
                  NULLIFY (trho_h, trho_s)
                  CALL get_rho_atom(rho_atom=rho_atom, trho_rad_h=trho_h, trho_rad_s=trho_s)
               END IF

               DO ir = 1, nr
                  CALL calc_rho_angular(grid_atom, harmonics, nspins, gradient_f, &
                                        ir, r_h, r_s, rho_h, rho_s, dr_h, dr_s, &
                                        r_h_d, r_s_d, drho_h, drho_s)
                  IF (tau_f) THEN
                     CALL calc_tau_angular(grid_atom, harmonics, nspins, ir, &
                                           trho_h, trho_s, tau_h, tau_s)
                  END IF
                  CALL fill_rho_set(rho_set_h, lsd, nspins, needs, rho_h, drho_h, tau_h, na, ir)
                  CALL fill_rho_set(rho_set_s, lsd, nspins, needs, rho_s, drho_s, tau_s, na, ir)
               END DO

               !-------------------!
               ! hard atom density !
               !-------------------!
               CALL xc_dset_zero_all(deriv_set)
               CALL vxc_of_r_new(xc_fun_section, rho_set_h, deriv_set, 1, needs, weight, &
                                 lsd, na, nr, exc_h, vxc_h, vxg_h, vtau_h, energy_only=energy_only, &
                                 epr_xc=epr_xc, adiabatic_rescale_factor=my_adiabatic_rescale_factor)
               rho_atom%exc_h = rho_atom%exc_h+exc_h

               !-------------------!
               ! soft atom density !
               !-------------------!
               CALL xc_dset_zero_all(deriv_set)
               CALL vxc_of_r_new(xc_fun_section, rho_set_s, deriv_set, 1, needs, weight, &
                                 lsd, na, nr, exc_s, vxc_s, vxg_s, vtau_s, energy_only=energy_only, &
                                 epr_xc=epr_xc, adiabatic_rescale_factor=my_adiabatic_rescale_factor)
               rho_atom%exc_s = rho_atom%exc_s+exc_s

               IF (epr_xc) THEN
                  DO ispin = 1, nspins
                     DO idir = 1, 3
                        DO ir = 1, nr
                           DO ia = 1, na
                              gradient_atom_set(iatom)%nablavks_vec_rad_h(idir, ispin)%r_coef(ir, ia) = &
                                 gradient_atom_set(iatom)%nablavks_vec_rad_h(idir, ispin)%r_coef(ir, ia) &
                                 +vxg_h(idir, ia, ir, ispin)
                              gradient_atom_set(iatom)%nablavks_vec_rad_s(idir, ispin)%r_coef(ir, ia) = &
                                 gradient_atom_set(iatom)%nablavks_vec_rad_s(idir, ispin)%r_coef(ir, ia) &
                                 +vxg_s(idir, ia, ir, ispin)
                           END DO ! ia
                        END DO ! ir
                     END DO ! idir
                  END DO ! ispin
               END IF

               ! Add contributions to the exc energy

               energy%exc1 = energy%exc1+rho_atom%exc_h-rho_atom%exc_s

               ! Integration to get the matrix elements relative to the vxc_atom
               ! here the products with the primitives is done: gaVxcgb
               ! internal transformation to get the integral in cartesian Gaussians

               IF (.NOT. energy_only) THEN
                  qs_kind => qs_kind_set(ikind)
                  IF (gradient_f) THEN
                     CALL gaVxcgb_GC(vxc_h, vxc_s, vxg_h, vxg_s, qs_kind, &
                                     rho_atom, nspins)
                  ELSE
                     CALL gaVxcgb_noGC(vxc_h, vxc_s, qs_kind, rho_atom, nspins)
                  END IF
                  IF (tau_f) THEN
                     CALL dgaVtaudgb(vtau_h, vtau_s, qs_kind, rho_atom, nspins)
                  END IF
               END IF ! energy_only
               NULLIFY (r_h, r_s, dr_h, dr_s)
            END DO ! iat

            ! Release the xc structure used to store the xc derivatives
            CALL xc_dset_release(deriv_set)
            CALL xc_rho_set_release(rho_set_h)
            CALL xc_rho_set_release(rho_set_s)

         END DO ! ikind

         CALL mp_sum(energy%exc1, para_env%group)

         IF (ASSOCIATED(rho_h)) THEN
            DEALLOCATE (rho_h)
         END IF

         IF (ASSOCIATED(rho_s)) THEN
            DEALLOCATE (rho_s)
         END IF

         IF (ASSOCIATED(vxc_h)) THEN
            DEALLOCATE (vxc_h)
         END IF

         IF (ASSOCIATED(vxc_s)) THEN
            DEALLOCATE (vxc_s)
         END IF

         IF (gradient_f) THEN

            IF (ASSOCIATED(drho_h)) THEN
               DEALLOCATE (drho_h)
            END IF

            IF (ASSOCIATED(drho_s)) THEN
               DEALLOCATE (drho_s)
            END IF

            IF (ASSOCIATED(vxg_h)) THEN
               DEALLOCATE (vxg_h)
            END IF

            IF (ASSOCIATED(vxg_s)) THEN
               DEALLOCATE (vxg_s)
            END IF

         END IF

         IF (tau_f) THEN

            IF (ASSOCIATED(tau_h)) THEN
               DEALLOCATE (tau_h)
            END IF

            IF (ASSOCIATED(tau_s)) THEN
               DEALLOCATE (tau_s)
            END IF

            IF (ASSOCIATED(vtau_h)) THEN
               DEALLOCATE (vtau_h)
            END IF

            IF (ASSOCIATED(vtau_s)) THEN
               DEALLOCATE (vtau_s)
            END IF

         END IF

      END IF !xc_none

      CALL timestop(handle)

   END SUBROUTINE calculate_vxc_atom

! **************************************************************************************************
!> \brief ...
!> \param p_env ...
!> \param qs_env ...
!> \param xc_section ...
!> \param do_tddft ...
!> \param do_triplet ...
! **************************************************************************************************
   SUBROUTINE calculate_xc_2nd_deriv_atom(p_env, qs_env, xc_section, do_tddft, do_triplet)

      TYPE(qs_p_env_type), POINTER                       :: p_env
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(section_vals_type), POINTER                   :: xc_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: do_tddft, do_triplet

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calculate_xc_2nd_deriv_atom', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom, excitations, handle, iatom, ikind, &
                                                            ir, na, natom, nr, nspins, res_etype
      INTEGER, DIMENSION(2)                              :: local_loop_limit
      INTEGER, DIMENSION(2, 3)                           :: bounds
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      LOGICAL                                            :: gradient_functional, lsd, lsd_singlets, &
                                                            my_do_tddft, paw_atom, tau_f
      REAL(KIND=dp)                                      :: density_cut, gradient_cut, rtot, tau_cut
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rho1_h, rho1_s, rho_h, rho_s, tau1_h, &
                                                            tau1_s, tau_h, tau_s, weight
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: vxc_h, vxc_s
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: drho1_h, drho1_s, drho_h, drho_s, vxg_h, &
                                                            vxg_s
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: dr1_h, dr1_s, dr_h, dr_s, r1_h, r1_s, &
                                                            r_h, r_s
      TYPE(rho_atom_coeff), DIMENSION(:, :), POINTER     :: r1_h_d, r1_s_d, r_h_d, r_s_d
      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho1_atom_set, rho_atom_set
      TYPE(rho_atom_type), POINTER                       :: rho1_atom, rho_atom
      TYPE(section_vals_type), POINTER                   :: input, xc_fun_section
      TYPE(xc_derivative_set_type), POINTER              :: deriv_set
      TYPE(xc_rho_cflags_type)                           :: needs
      TYPE(xc_rho_set_type), POINTER                     :: rho1_set_h, rho1_set_s, rho_set_h, &
                                                            rho_set_s

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (qs_kind_set, energy, para_env)
      NULLIFY (rho_h, rho_s, drho_h, drho_s, weight)
      NULLIFY (rho1_h, rho1_s, drho1_h, drho1_s)
      NULLIFY (vxc_h, vxc_s, vxg_h, vxg_s)
      NULLIFY (tau_h, tau_s, tau1_h, tau1_s)

      my_do_tddft = .TRUE.
      IF (PRESENT(do_tddft)) my_do_tddft = do_tddft

      CALL get_qs_env(qs_env=qs_env, &
                      para_env=para_env, &
                      energy=energy, &
                      input=input, &
                      qs_kind_set=qs_kind_set, &
                      atomic_kind_set=atomic_kind_set, &
                      rho_atom_set=rho_atom_set)

      rho1_atom_set => p_env%local_rho_set%rho_atom_set

      CALL section_vals_val_get(input, "DFT%LSD", l_val=lsd)
      CALL section_vals_val_get(xc_section, "DENSITY_CUTOFF", &
                                r_val=density_cut)
      CALL section_vals_val_get(xc_section, "GRADIENT_CUTOFF", &
                                r_val=gradient_cut)
      CALL section_vals_val_get(xc_section, "TAU_CUTOFF", &
                                r_val=tau_cut)
      IF (my_do_tddft) THEN
         CALL section_vals_val_get(input, "DFT%EXCITATIONS", &
                                   i_val=excitations)
         CALL section_vals_val_get(input, "DFT%TDDFPT%LSD_SINGLETS", &
                                   l_val=lsd_singlets)
         CALL section_vals_val_get(input, "DFT%TDDFPT%RES_ETYPE", &
                                   i_val=res_etype)
      ENDIF
      xc_fun_section => section_vals_get_subs_vals(xc_section, &
                                                   "XC_FUNCTIONAL")
      IF (lsd) THEN
         nspins = 2
      ELSE
         nspins = 1
      END IF
      needs = xc_functionals_get_needs(xc_fun_section, lsd=lsd, &
                                       add_basic_components=.TRUE.)
      gradient_functional = needs%drho .OR. needs%drho_spin
      tau_f = (needs%tau .OR. needs%tau_spin)
      IF (tau_f) THEN
         CALL cp__b("qs_vxc_atom.F",523,"Tau functionals not implemented for GAPW 2nd derivatives")
      END IF

      IF (my_do_tddft) THEN
         IF (excitations == tddfpt_excitations) THEN
            IF (nspins == 1 .AND. (lsd_singlets .OR. res_etype == tddfpt_triplet)) THEN
               lsd = .TRUE.
            END IF
         END IF
      ELSE
         IF (PRESENT(do_triplet)) THEN
            IF (nspins == 1 .AND. do_triplet) THEN
               lsd = .TRUE.
            ENDIF
         ENDIF
      ENDIF

!  Here starts the loop over all the atoms
      DO ikind = 1, SIZE(atomic_kind_set)

         NULLIFY (atom_list, harmonics, grid_atom)
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=natom)
         CALL get_qs_kind(qs_kind_set(ikind), paw_atom=paw_atom, &
                          harmonics=harmonics, grid_atom=grid_atom)
         IF (.NOT. paw_atom) CYCLE

         nr = grid_atom%nr
         na = grid_atom%ng_sphere

         ! Array dimension: here anly one dimensional arrays are used,
         ! i.e. only the first column of deriv_data is read.
         ! The other to dimensions  are set to size equal 1.
         bounds(1:2, 1:3) = 1
         bounds(2, 1) = na
         bounds(2, 2) = nr

         NULLIFY (deriv_set, rho_set_h, rho_set_s, rho1_set_h, rho1_set_s)
         CALL xc_dset_create(deriv_set, local_bounds=bounds)
         CALL xc_rho_set_create(rho_set_h, bounds, rho_cutoff=density_cut, &
                                drho_cutoff=gradient_cut, tau_cutoff=tau_cut)
         CALL xc_rho_set_create(rho_set_s, bounds, rho_cutoff=density_cut, &
                                drho_cutoff=gradient_cut, tau_cutoff=tau_cut)
         CALL xc_rho_set_create(rho1_set_h, bounds, rho_cutoff=density_cut, &
                                drho_cutoff=gradient_cut, tau_cutoff=tau_cut)
         CALL xc_rho_set_create(rho1_set_s, bounds, rho_cutoff=density_cut, &
                                drho_cutoff=gradient_cut, tau_cutoff=tau_cut)

         ! allocate the required 3d arrays where to store rho and drho
         IF (nspins == 1 .AND. .NOT. lsd) THEN
            CALL xc_rho_set_atom_update(rho_set_h, needs, 1, bounds)
            CALL xc_rho_set_atom_update(rho1_set_h, needs, 1, bounds)
            CALL xc_rho_set_atom_update(rho_set_s, needs, 1, bounds)
            CALL xc_rho_set_atom_update(rho1_set_s, needs, 1, bounds)
         ELSE
            CALL xc_rho_set_atom_update(rho_set_h, needs, 2, bounds)
            CALL xc_rho_set_atom_update(rho1_set_h, needs, 2, bounds)
            CALL xc_rho_set_atom_update(rho_set_s, needs, 2, bounds)
            CALL xc_rho_set_atom_update(rho1_set_s, needs, 2, bounds)
         END IF

         ALLOCATE (rho_h(1:na, 1:nspins), rho1_h(1:na, 1:nspins), &
                   rho_s(1:na, 1:nspins), rho1_s(1:na, 1:nspins))

         ALLOCATE (vxc_h(1:na, 1:nr, 1:nspins), vxc_s(1:na, 1:nr, 1:nspins))
         vxc_h = 0.0_dp
         vxc_s = 0.0_dp

         weight => grid_atom%weight

         IF (gradient_functional) THEN
            ALLOCATE (drho_h(1:4, 1:na, 1:nr, 1:nspins), drho1_h(1:4, 1:na, 1:nr, 1:nspins), &
                      drho_s(1:4, 1:na, 1:nr, 1:nspins), drho1_s(1:4, 1:na, 1:nr, 1:nspins))

            ALLOCATE (vxg_h(1:3, 1:na, 1:nr, 1:nspins), vxg_s(1:3, 1:na, 1:nr, 1:nspins))
            vxg_h = 0.0_dp
            vxg_s = 0.0_dp
         END IF

         ! parallelization
         local_loop_limit = get_limit(natom, para_env%num_pe, para_env%mepos)

         DO iatom = local_loop_limit(1), local_loop_limit(2) !1,natom
            atom = atom_list(iatom)

            rho_atom_set(atom)%exc_h = 0.0_dp
            rho_atom_set(atom)%exc_s = 0.0_dp
            rho1_atom_set(atom)%exc_h = 0.0_dp
            rho1_atom_set(atom)%exc_s = 0.0_dp

            rho_atom => rho_atom_set(atom)
            rho1_atom => rho1_atom_set(atom)
            NULLIFY (r_h, r_s, dr_h, dr_s, r_h_d, r_s_d)
            NULLIFY (r1_h, r1_s, dr1_h, dr1_s, r1_h_d, r1_s_d)
            IF (gradient_functional) THEN
               CALL get_rho_atom(rho_atom=rho_atom, &
                                 rho_rad_h=r_h, rho_rad_s=r_s, &
                                 drho_rad_h=dr_h, drho_rad_s=dr_s, &
                                 rho_rad_h_d=r_h_d, rho_rad_s_d=r_s_d)
               CALL get_rho_atom(rho_atom=rho1_atom, &
                                 rho_rad_h=r1_h, rho_rad_s=r1_s, &
                                 drho_rad_h=dr1_h, drho_rad_s=dr1_s, &
                                 rho_rad_h_d=r1_h_d, rho_rad_s_d=r1_s_d)
               drho_h = 0.0_dp; drho_s = 0.0_dp
               drho1_h = 0.0_dp; drho1_s = 0.0_dp
            ELSE
               CALL get_rho_atom(rho_atom=rho_atom, &
                                 rho_rad_h=r_h, rho_rad_s=r_s)
               CALL get_rho_atom(rho_atom=rho1_atom, &
                                 rho_rad_h=r1_h, rho_rad_s=r1_s)
            END IF

            rtot = 0.0_dp

            DO ir = 1, nr

               CALL calc_rho_angular(grid_atom, harmonics, nspins, gradient_functional, &
                                     ir, r_h, r_s, rho_h, rho_s, dr_h, dr_s, r_h_d, r_s_d, &
                                     drho_h, drho_s)
               CALL calc_rho_angular(grid_atom, harmonics, nspins, gradient_functional, &
                                     ir, r1_h, r1_s, rho1_h, rho1_s, dr1_h, dr1_s, r1_h_d, r1_s_d, &
                                     drho1_h, drho1_s)

               CALL fill_rho_set(rho_set_h, lsd, nspins, needs, rho_h, drho_h, tau_h, na, ir)
               CALL fill_rho_set(rho1_set_h, lsd, nspins, needs, rho1_h, drho1_h, tau1_h, na, ir)
               CALL fill_rho_set(rho_set_s, lsd, nspins, needs, rho_s, drho_s, tau_s, na, ir)
               CALL fill_rho_set(rho1_set_s, lsd, nspins, needs, rho1_s, drho1_s, tau1_s, na, ir)

            END DO

            CALL xc_2nd_deriv_of_r(xc_section=xc_section, &
                                   rho_set=rho_set_h, rho1_set=rho1_set_h, &
                                   deriv_set=deriv_set, &
                                   w=weight, vxc=vxc_h, vxg=vxg_h)
            CALL xc_2nd_deriv_of_r(xc_section=xc_section, &
                                   rho_set=rho_set_s, rho1_set=rho1_set_s, &
                                   deriv_set=deriv_set, &
                                   w=weight, vxc=vxc_s, vxg=vxg_s)

            IF (gradient_functional) THEN
               CALL gaVxcgb_GC(vxc_h, vxc_s, vxg_h, vxg_s, qs_kind_set(ikind), &
                               rho1_atom, nspins)
            ELSE
               CALL gaVxcgb_noGC(vxc_h, vxc_s, qs_kind_set(ikind), &
                                 rho1_atom, nspins)
            ENDIF

            NULLIFY (r_h, r_s, dr_h, dr_s)

         END DO

         ! some cleanup
         NULLIFY (weight)
         DEALLOCATE (rho_h, rho_s, rho1_h, rho1_s, vxc_h, vxc_s)
         IF (gradient_functional) THEN
            DEALLOCATE (drho_h, drho_s, vxg_h, vxg_s)
            DEALLOCATE (drho1_h, drho1_s)
         END IF

         CALL xc_dset_release(deriv_set)
         CALL xc_rho_set_release(rho_set_h)
         CALL xc_rho_set_release(rho1_set_h)
         CALL xc_rho_set_release(rho_set_s)
         CALL xc_rho_set_release(rho1_set_s)

      END DO

      CALL timestop(handle)

   END SUBROUTINE calculate_xc_2nd_deriv_atom

! **************************************************************************************************
!> \brief ...
!> \param grid_atom ...
!> \param harmonics ...
!> \param nspins ...
!> \param grad_func ...
!> \param ir ...
!> \param r_h ...
!> \param r_s ...
!> \param rho_h ...
!> \param rho_s ...
!> \param dr_h ...
!> \param dr_s ...
!> \param r_h_d ...
!> \param r_s_d ...
!> \param drho_h ...
!> \param drho_s ...
! **************************************************************************************************
   SUBROUTINE calc_rho_angular(grid_atom, harmonics, nspins, grad_func, &
                               ir, r_h, r_s, rho_h, rho_s, &
                               dr_h, dr_s, r_h_d, r_s_d, drho_h, drho_s)

      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      INTEGER, INTENT(IN)                                :: nspins
      LOGICAL, INTENT(IN)                                :: grad_func
      INTEGER, INTENT(IN)                                :: ir
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: r_h, r_s
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rho_h, rho_s
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: dr_h, dr_s
      TYPE(rho_atom_coeff), DIMENSION(:, :), POINTER     :: r_h_d, r_s_d
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER      :: drho_h, drho_s

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_rho_angular', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, iso, ispin, na
      REAL(KIND=dp)                                      :: rad, urad

      IF(.NOT.(ASSOCIATED(r_h)))CALL cp__a("qs_vxc_atom.F",732)
      IF(.NOT.(ASSOCIATED(r_s)))CALL cp__a("qs_vxc_atom.F",733)
      IF(.NOT.(ASSOCIATED(rho_h)))CALL cp__a("qs_vxc_atom.F",734)
      IF(.NOT.(ASSOCIATED(rho_s)))CALL cp__a("qs_vxc_atom.F",735)
      IF (grad_func) THEN
         IF(.NOT.(ASSOCIATED(dr_h)))CALL cp__a("qs_vxc_atom.F",737)
         IF(.NOT.(ASSOCIATED(dr_s)))CALL cp__a("qs_vxc_atom.F",738)
         IF(.NOT.(ASSOCIATED(r_h_d)))CALL cp__a("qs_vxc_atom.F",739)
         IF(.NOT.(ASSOCIATED(r_s_d)))CALL cp__a("qs_vxc_atom.F",740)
         IF(.NOT.(ASSOCIATED(drho_h)))CALL cp__a("qs_vxc_atom.F",741)
         IF(.NOT.(ASSOCIATED(drho_s)))CALL cp__a("qs_vxc_atom.F",742)
      END IF

      na = grid_atom%ng_sphere
      rad = grid_atom%rad(ir)
      urad = grid_atom%oorad2l(ir, 1)
      rho_h = 0.0_dp; rho_s = 0.0_dp
      DO ispin = 1, nspins
         DO iso = 1, harmonics%max_iso_not0
            DO ia = 1, na
               rho_h(ia, ispin) = rho_h(ia, ispin)+ &
                                  r_h(ispin)%r_coef(ir, iso)* &
                                  harmonics%slm(ia, iso)
               rho_s(ia, ispin) = rho_s(ia, ispin)+ &
                                  r_s(ispin)%r_coef(ir, iso)* &
                                  harmonics%slm(ia, iso)
            END DO ! ia
         END DO ! iso
      END DO ! ispin

      IF (grad_func) THEN
         DO ispin = 1, nspins
            DO iso = 1, harmonics%max_iso_not0
               DO ia = 1, na

!              components of the gradient of rho1 hard
                  drho_h(1, ia, ir, ispin) = drho_h(1, ia, ir, ispin)+ &
                                             dr_h(ispin)%r_coef(ir, iso)* &
                                             harmonics%a(1, ia)*harmonics%slm(ia, iso)+ &
                                             r_h_d(1, ispin)%r_coef(ir, iso)* &
                                             harmonics%slm(ia, iso)

                  drho_h(2, ia, ir, ispin) = drho_h(2, ia, ir, ispin)+ &
                                             dr_h(ispin)%r_coef(ir, iso)* &
                                             harmonics%a(2, ia)*harmonics%slm(ia, iso)+ &
                                             r_h_d(2, ispin)%r_coef(ir, iso)* &
                                             harmonics%slm(ia, iso)

                  drho_h(3, ia, ir, ispin) = drho_h(3, ia, ir, ispin)+ &
                                             dr_h(ispin)%r_coef(ir, iso)* &
                                             harmonics%a(3, ia)*harmonics%slm(ia, iso)+ &
                                             r_h_d(3, ispin)%r_coef(ir, iso)* &
                                             harmonics%slm(ia, iso)

!              components of the gradient of rho1 soft
                  drho_s(1, ia, ir, ispin) = drho_s(1, ia, ir, ispin)+ &
                                             dr_s(ispin)%r_coef(ir, iso)* &
                                             harmonics%a(1, ia)*harmonics%slm(ia, iso)+ &
                                             r_s_d(1, ispin)%r_coef(ir, iso)* &
                                             harmonics%slm(ia, iso)

                  drho_s(2, ia, ir, ispin) = drho_s(2, ia, ir, ispin)+ &
                                             dr_s(ispin)%r_coef(ir, iso)* &
                                             harmonics%a(2, ia)*harmonics%slm(ia, iso)+ &
                                             r_s_d(2, ispin)%r_coef(ir, iso)* &
                                             harmonics%slm(ia, iso)

                  drho_s(3, ia, ir, ispin) = drho_s(3, ia, ir, ispin)+ &
                                             dr_s(ispin)%r_coef(ir, iso)* &
                                             harmonics%a(3, ia)*harmonics%slm(ia, iso)+ &
                                             r_s_d(3, ispin)%r_coef(ir, iso)* &
                                             harmonics%slm(ia, iso)

                  drho_h(4, ia, ir, ispin) = SQRT( &
                                             drho_h(1, ia, ir, ispin)*drho_h(1, ia, ir, ispin)+ &
                                             drho_h(2, ia, ir, ispin)*drho_h(2, ia, ir, ispin)+ &
                                             drho_h(3, ia, ir, ispin)*drho_h(3, ia, ir, ispin))

                  drho_s(4, ia, ir, ispin) = SQRT( &
                                             drho_s(1, ia, ir, ispin)*drho_s(1, ia, ir, ispin)+ &
                                             drho_s(2, ia, ir, ispin)*drho_s(2, ia, ir, ispin)+ &
                                             drho_s(3, ia, ir, ispin)*drho_s(3, ia, ir, ispin))

               END DO ! ia
            END DO ! iso
         END DO ! ispin
      END IF

   END SUBROUTINE calc_rho_angular
! **************************************************************************************************
!> \brief ...
!> \param grid_atom ...
!> \param harmonics ...
!> \param nspins ...
!> \param ir ...
!> \param trho_h ...
!> \param trho_s ...
!> \param tau_h ...
!> \param tau_s ...
! **************************************************************************************************
   SUBROUTINE calc_tau_angular(grid_atom, harmonics, nspins, ir, &
                               trho_h, trho_s, tau_h, tau_s)

      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      INTEGER, INTENT(IN)                                :: nspins, ir
      TYPE(rho_atom_coeff), DIMENSION(:, :), POINTER     :: trho_h, trho_s
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: tau_h, tau_s

      CHARACTER(LEN=*), PARAMETER :: routineN = 'calc_tau_angular', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ia, iso, ispin, na

      IF(.NOT.(ASSOCIATED(trho_h)))CALL cp__a("qs_vxc_atom.F",846)
      IF(.NOT.(ASSOCIATED(trho_s)))CALL cp__a("qs_vxc_atom.F",847)
      IF(.NOT.(ASSOCIATED(tau_h)))CALL cp__a("qs_vxc_atom.F",848)
      IF(.NOT.(ASSOCIATED(tau_s)))CALL cp__a("qs_vxc_atom.F",849)

      na = grid_atom%ng_sphere
      tau_h = 0.0_dp
      tau_s = 0.0_dp

      DO ispin = 1, nspins
         DO iso = 1, harmonics%max_iso_not0
            DO ia = 1, na
               tau_h(ia, ispin) = tau_h(ia, ispin)+ &
                                  trho_h(1, ispin)%r_coef(ir, iso)*harmonics%slm(ia, iso)
               tau_h(ia, ispin) = tau_h(ia, ispin)+ &
                                  trho_h(3, ispin)%r_coef(ir, iso)*harmonics%slm(ia, iso)
               tau_s(ia, ispin) = tau_s(ia, ispin)+ &
                                  trho_s(1, ispin)%r_coef(ir, iso)*harmonics%slm(ia, iso)
               tau_s(ia, ispin) = tau_s(ia, ispin)+ &
                                  trho_s(3, ispin)%r_coef(ir, iso)*harmonics%slm(ia, iso)
            END DO ! ia
         END DO ! iso
         DO iso = 1, harmonics%max_iso_not0
            DO ia = 1, na
               tau_h(ia, ispin) = tau_h(ia, ispin)+ &
                                  trho_h(2, ispin)%r_coef(ir, iso)*harmonics%slm(ia, iso)*grid_atom%usin_azi(ia)**2
               tau_s(ia, ispin) = tau_s(ia, ispin)+ &
                                  trho_s(2, ispin)%r_coef(ir, iso)*harmonics%slm(ia, iso)*grid_atom%usin_azi(ia)**2
            END DO ! ia
         END DO ! iso
      END DO ! ispin

   END SUBROUTINE calc_tau_angular

! **************************************************************************************************
!> \brief ...
!> \param vxc_h ...
!> \param vxc_s ...
!> \param qs_kind ...
!> \param rho_atom ...
!> \param nspins ...
! **************************************************************************************************
   SUBROUTINE gaVxcgb_noGC(vxc_h, vxc_s, qs_kind, rho_atom, nspins)

      REAL(dp), DIMENSION(:, :, :), POINTER              :: vxc_h, vxc_s
      TYPE(qs_kind_type), INTENT(IN)                     :: qs_kind
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      INTEGER, INTENT(IN)                                :: nspins

      CHARACTER(len=*), PARAMETER :: routineN = 'gaVxcgb_noGC', routineP = moduleN//':'//routineN

      INTEGER :: handle, ia, ic, icg, ipgf1, ipgf2, ir, iset1, iset2, iso, iso1, iso2, ispin, &
         istat, l, ld, lmax12, lmax_expansion, lmin12, m1, m2, max_iso_not0, max_iso_not0_local, &
         max_s_harm, maxl, maxso, n1, n2, na, ngau1, ngau2, nngau1, nr, nset, size1
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cg_n_list
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: cg_list
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: g1, g2
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: gg, gVg_h, gVg_s, matso_h, matso_s, vx
      REAL(dp), DIMENSION(:, :), POINTER                 :: zet
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_CG
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: int_hh, int_ss

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (lmin, lmax, npgf, zet, my_CG, harmonics, grid_atom)

      CALL get_qs_kind(qs_kind, basis_set=orb_basis, &
                       harmonics=harmonics, grid_atom=grid_atom)

      CALL get_gto_basis_set(gto_basis_set=orb_basis, lmax=lmax, lmin=lmin, &
                             maxso=maxso, maxl=maxl, npgf=npgf, &
                             nset=nset, zet=zet)

      nr = grid_atom%nr
      na = grid_atom%ng_sphere
      my_CG => harmonics%my_CG
      max_iso_not0 = harmonics%max_iso_not0
      lmax_expansion = indso(1, max_iso_not0)
      max_s_harm = harmonics%max_s_harm

      ALLOCATE (g1(nr), g2(nr), gg(nr, 0:2*maxl), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 933, &
                                       "g1,g2,gg", (2*nr+2*maxl*nr)*dp_size)

      ALLOCATE (gVg_h(na, 0:2*maxl), gVg_s(na, 0:2*maxl), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 937, &
                                       "gVg_h,gVg_s", 4*na*maxl*dp_size)

      ALLOCATE (matso_h(nsoset(maxl), nsoset(maxl)), &
                matso_s(nsoset(maxl), nsoset(maxl)), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 942, &
                                       "matso_h,matso_s", &
                                       2*nsoset(maxl)*nsoset(maxl)*dp_size)

      ALLOCATE (vx(na, nr), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 947, &
                                       "vx", na*nr*dp_size)

      ALLOCATE (cg_list(2, nsoset(maxl)**2, max_s_harm), cg_n_list(max_s_harm))

      NULLIFY (int_hh, int_ss)

      CALL get_rho_atom(rho_atom=rho_atom, ga_Vlocal_gb_h=int_hh, ga_Vlocal_gb_s=int_ss)

      g1 = 0.0_dp
      g2 = 0.0_dp
      m1 = 0
      DO iset1 = 1, nset
         n1 = nsoset(lmax(iset1))
         m2 = 0
         DO iset2 = 1, nset
            CALL get_none0_cg_list(my_CG, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                   max_s_harm, lmax_expansion, cg_list, cg_n_list, max_iso_not0_local)
            IF(.NOT.(max_iso_not0_local .LE. max_iso_not0))CALL cp__a("qs_vxc_atom.F",965)

            n2 = nsoset(lmax(iset2))
            DO ipgf1 = 1, npgf(iset1)
               ngau1 = n1*(ipgf1-1)+m1
               size1 = nsoset(lmax(iset1))-nsoset(lmin(iset1)-1)
               nngau1 = nsoset(lmin(iset1)-1)+ngau1

               g1(1:nr) = EXP(-zet(ipgf1, iset1)*grid_atom%rad2(1:nr))
               DO ipgf2 = 1, npgf(iset2)
                  ngau2 = n2*(ipgf2-1)+m2

                  g2(1:nr) = EXP(-zet(ipgf2, iset2)*grid_atom%rad2(1:nr))
                  lmin12 = lmin(iset1)+lmin(iset2)
                  lmax12 = lmax(iset1)+lmax(iset2)

                  ! reduce expansion local densities
                  IF (lmin12 .LE. lmax_expansion) THEN

                     gg = 0.0_dp
                     IF (lmin12 == 0) THEN
                        gg(1:nr, lmin12) = g1(1:nr)*g2(1:nr)
                     ELSE
                        gg(1:nr, lmin12) = grid_atom%rad2l(1:nr, lmin12)*g1(1:nr)*g2(1:nr)
                     END IF

                     ! limit the expansion of the local densities to a max L
                     IF (lmax12 .GT. lmax_expansion) lmax12 = lmax_expansion

                     DO l = lmin12+1, lmax12
                        gg(1:nr, l) = grid_atom%rad(1:nr)*gg(:, l-1)
                     END DO

                     DO ispin = 1, nspins
                        ld = lmax12+1
                        DO ir = 1, nr
                           vx(1:na, ir) = vxc_h(1:na, ir, ispin)
                        END DO
                        CALL dgemm('N', 'N', na, ld, nr, 1.0_dp, vx(1:na, 1:nr), na, &
                                   gg(1:nr, 0:lmax12), nr, 0.0_dp, gVg_h(1:na, 0:lmax12), na)
                        DO ir = 1, nr
                           vx(1:na, ir) = vxc_s(1:na, ir, ispin)
                        END DO
                        CALL dgemm('N', 'N', na, ld, nr, 1.0_dp, vx(1:na, 1:nr), na, &
                                   gg(1:nr, 0:lmax12), nr, 0.0_dp, gVg_s(1:na, 0:lmax12), na)

                        matso_h = 0.0_dp
                        matso_s = 0.0_dp
                        DO iso = 1, max_iso_not0_local
                           DO icg = 1, cg_n_list(iso)
                              iso1 = cg_list(1, icg, iso)
                              iso2 = cg_list(2, icg, iso)
                              l = indso(1, iso1)+indso(1, iso2)

                              IF(.NOT.(l <= lmax_expansion))CALL cp__a("qs_vxc_atom.F",1019)
                              DO ia = 1, na
                                 matso_h(iso1, iso2) = matso_h(iso1, iso2)+ &
                                                       gVg_h(ia, l)* &
                                                       my_CG(iso1, iso2, iso)* &
                                                       harmonics%slm(ia, iso)
                                 matso_s(iso1, iso2) = matso_s(iso1, iso2)+ &
                                                       gVg_s(ia, l)* &
                                                       my_CG(iso1, iso2, iso)* &
                                                       harmonics%slm(ia, iso)
                              END DO
                           END DO
                        END DO

!             Write in the global matrix
                        DO ic = nsoset(lmin(iset2)-1)+1, nsoset(lmax(iset2))
                           iso1 = nsoset(lmin(iset1)-1)+1
                           iso2 = ngau2+ic
                           CALL daxpy(size1, 1.0_dp, matso_h(iso1, ic), 1, &
                                      int_hh(ispin)%r_coef(nngau1+1, iso2), 1)
                           CALL daxpy(size1, 1.0_dp, matso_s(iso1, ic), 1, &
                                      int_ss(ispin)%r_coef(nngau1+1, iso2), 1)
                        END DO

                     END DO ! ispin

                  END IF ! lmax_expansion

               END DO ! ipfg2
            END DO ! ipfg1
            m2 = m2+maxso
         END DO ! iset2
         m1 = m1+maxso
      END DO ! iset1

      DEALLOCATE (g1, g2, gg, matso_h, matso_s, gVg_s, gVg_h, vx)

      DEALLOCATE (cg_list, cg_n_list)

      CALL timestop(handle)

   END SUBROUTINE gaVxcgb_noGC

! **************************************************************************************************
!> \brief ...
!> \param vxc_h ...
!> \param vxc_s ...
!> \param vxg_h ...
!> \param vxg_s ...
!> \param qs_kind ...
!> \param rho_atom ...
!> \param nspins ...
! **************************************************************************************************
   SUBROUTINE gaVxcgb_GC(vxc_h, vxc_s, vxg_h, vxg_s, qs_kind, rho_atom, &
                         nspins)

      REAL(dp), DIMENSION(:, :, :), POINTER              :: vxc_h, vxc_s
      REAL(dp), DIMENSION(:, :, :, :), POINTER           :: vxg_h, vxg_s
      TYPE(qs_kind_type), INTENT(IN)                     :: qs_kind
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      INTEGER, INTENT(IN)                                :: nspins

      CHARACTER(len=*), PARAMETER :: routineN = 'gaVxcgb_GC', routineP = moduleN//':'//routineN

      INTEGER :: dmax_iso_not0, dmax_iso_not0_local, handle, ia, ic, icg, ipgf1, ipgf2, ir, iset1, &
         iset2, iso, iso1, iso2, ispin, istat, l, lmax12, lmax_expansion, lmin12, m1, m2, &
         max_iso_not0, max_iso_not0_local, max_s_harm, maxl, maxso, n1, n2, na, ngau1, ngau2, &
         nngau1, nr, nset, size1
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cg_n_list, dcg_n_list
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: cg_list, dcg_list
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf
      REAL(dp)                                           :: urad
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: g1, g2
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: dgg, gg, gVXCg_h, gVXCg_s, matso_h, &
                                                            matso_s
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: gVXGg_h, gVXGg_s
      REAL(dp), DIMENSION(:, :), POINTER                 :: zet
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_CG
      REAL(dp), DIMENSION(:, :, :, :), POINTER           :: my_CG_dxyz
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: int_hh, int_ss

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (lmin, lmax, npgf, zet, my_CG, my_CG_dxyz, harmonics, grid_atom)

      CALL get_qs_kind(qs_kind, basis_set=orb_basis, &
                       harmonics=harmonics, grid_atom=grid_atom)

      CALL get_gto_basis_set(gto_basis_set=orb_basis, lmax=lmax, lmin=lmin, &
                             maxso=maxso, maxl=maxl, npgf=npgf, &
                             nset=nset, zet=zet)

      nr = grid_atom%nr
      na = grid_atom%ng_sphere
      my_CG => harmonics%my_CG
      my_CG_dxyz => harmonics%my_CG_dxyz
      max_iso_not0 = harmonics%max_iso_not0
      dmax_iso_not0 = harmonics%dmax_iso_not0
      lmax_expansion = indso(1, max_iso_not0)
      max_s_harm = harmonics%max_s_harm

      ALLOCATE (g1(nr), g2(nr), gg(nr, 0:2*maxl), dgg(nr, 0:2*maxl), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1126, &
                                       "g1,g2,gg,dgg", (2*nr+4*nr*maxl)*dp_size)

      ALLOCATE (gVXCg_h(na, 0:2*maxl), gVXCg_s(na, 0:2*maxl), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1130, &
                                       "gVXCg_h,gVXCg_s", 4*na*maxl*dp_size)
      ALLOCATE (gVXGg_h(3, na, 0:2*maxl), gVXGg_s(3, na, 0:2*maxl), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1133, &
                                       "gVXCg_h,gVXCg_s", 12*na*maxl*dp_size)
      ALLOCATE (cg_list(2, nsoset(maxl)**2, max_s_harm), cg_n_list(max_s_harm), &
                dcg_list(2, nsoset(maxl)**2, max_s_harm), dcg_n_list(max_s_harm))

      NULLIFY (int_hh, int_ss)

      ALLOCATE (matso_h(nsoset(maxl), nsoset(maxl)), &
                matso_s(nsoset(maxl), nsoset(maxl)), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 1142, &
                                       "matso_h,matso_s", 2*nsoset(maxl)*nsoset(maxl)*dp_size)

      CALL get_rho_atom(rho_atom=rho_atom, ga_Vlocal_gb_h=int_hh, ga_Vlocal_gb_s=int_ss)

      DO ispin = 1, nspins

         g1 = 0.0_dp
         g2 = 0.0_dp
         m1 = 0
         DO iset1 = 1, nset
            n1 = nsoset(lmax(iset1))
            m2 = 0
            DO iset2 = 1, nset
               CALL get_none0_cg_list(my_CG, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                      max_s_harm, lmax_expansion, cg_list, cg_n_list, max_iso_not0_local)
               IF(.NOT.(max_iso_not0_local .LE. max_iso_not0))CALL cp__a("qs_vxc_atom.F",1158)
               CALL get_none0_cg_list(my_CG_dxyz, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                      max_s_harm, lmax_expansion, dcg_list, dcg_n_list, dmax_iso_not0_local)
               !IF(.NOT.(dmax_iso_not0_local.LE.dmax_iso_not0))CALL cp__a("qs_vxc_atom.F",1161)

               n2 = nsoset(lmax(iset2))
               DO ipgf1 = 1, npgf(iset1)
                  ngau1 = n1*(ipgf1-1)+m1
                  size1 = nsoset(lmax(iset1))-nsoset(lmin(iset1)-1)
                  nngau1 = nsoset(lmin(iset1)-1)+ngau1

                  g1(1:nr) = EXP(-zet(ipgf1, iset1)*grid_atom%rad2(1:nr))
                  DO ipgf2 = 1, npgf(iset2)
                     ngau2 = n2*(ipgf2-1)+m2

                     g2(1:nr) = EXP(-zet(ipgf2, iset2)*grid_atom%rad2(1:nr))
                     lmin12 = lmin(iset1)+lmin(iset2)
                     lmax12 = lmax(iset1)+lmax(iset2)

                     !test reduce expansion local densities
                     IF (lmin12 .LE. lmax_expansion) THEN

                        gg = 0.0_dp
                        dgg = 0.0_dp

                        IF (lmin12 == 0) THEN
                           gg(1:nr, lmin12) = g1(1:nr)*g2(1:nr)
                        ELSE
                           gg(1:nr, lmin12) = grid_atom%rad2l(1:nr, lmin12)*g1(1:nr)*g2(1:nr)
                        END IF

                        !test reduce expansion local densities
                        IF (lmax12 .GT. lmax_expansion) lmax12 = lmax_expansion

                        DO l = lmin12+1, lmax12
                           gg(1:nr, l) = grid_atom%rad(1:nr)*gg(:, l-1)
                           dgg(1:nr, l-1) = dgg(1:nr, l-1)-2.0_dp*(zet(ipgf1, iset1)+ &
                                                                   zet(ipgf2, iset2))*gg(1:nr, l)
                        END DO
                        dgg(1:nr, lmax12) = dgg(1:nr, lmax12)-2.0_dp*(zet(ipgf1, iset1)+ &
                                                                      zet(ipgf2, iset2))*grid_atom%rad(1:nr)* &
                                            gg(1:nr, lmax12)

                        gVXCg_h = 0.0_dp
                        gVXCg_s = 0.0_dp
                        gVXGg_h = 0.0_dp
                        gVXGg_s = 0.0_dp

                        ! Cross Term
                        DO l = lmin12, lmax12
                           DO ia = 1, na
                              DO ir = 1, nr
                                 gVXCg_h(ia, l) = gVXCg_h(ia, l)+ &
                                                  gg(ir, l)*vxc_h(ia, ir, ispin)+ &
                                                  dgg(ir, l)* &
                                                  (vxg_h(1, ia, ir, ispin)*harmonics%a(1, ia)+ &
                                                   vxg_h(2, ia, ir, ispin)*harmonics%a(2, ia)+ &
                                                   vxg_h(3, ia, ir, ispin)*harmonics%a(3, ia))

                                 gVXCg_s(ia, l) = gVXCg_s(ia, l)+ &
                                                  gg(ir, l)*vxc_s(ia, ir, ispin)+ &
                                                  dgg(ir, l)* &
                                                  (vxg_s(1, ia, ir, ispin)*harmonics%a(1, ia)+ &
                                                   vxg_s(2, ia, ir, ispin)*harmonics%a(2, ia)+ &
                                                   vxg_s(3, ia, ir, ispin)*harmonics%a(3, ia))

                                 urad = grid_atom%oorad2l(ir, 1)

                                 gVXGg_h(1, ia, l) = gVXGg_h(1, ia, l)+ &
                                                     vxg_h(1, ia, ir, ispin)* &
                                                     gg(ir, l)*urad

                                 gVXGg_h(2, ia, l) = gVXGg_h(2, ia, l)+ &
                                                     vxg_h(2, ia, ir, ispin)* &
                                                     gg(ir, l)*urad

                                 gVXGg_h(3, ia, l) = gVXGg_h(3, ia, l)+ &
                                                     vxg_h(3, ia, ir, ispin)* &
                                                     gg(ir, l)*urad

                                 gVXGg_s(1, ia, l) = gVXGg_s(1, ia, l)+ &
                                                     vxg_s(1, ia, ir, ispin)* &
                                                     gg(ir, l)*urad

                                 gVXGg_s(2, ia, l) = gVXGg_s(2, ia, l)+ &
                                                     vxg_s(2, ia, ir, ispin)* &
                                                     gg(ir, l)*urad

                                 gVXGg_s(3, ia, l) = gVXGg_s(3, ia, l)+ &
                                                     vxg_s(3, ia, ir, ispin)* &
                                                     gg(ir, l)*urad

                              ENDDO ! ir
                           ENDDO ! ia
                        ENDDO ! l

                        matso_h = 0.0_dp
                        matso_s = 0.0_dp
                        DO iso = 1, max_iso_not0_local
                           DO icg = 1, cg_n_list(iso)
                              iso1 = cg_list(1, icg, iso)
                              iso2 = cg_list(2, icg, iso)

                              l = indso(1, iso1)+indso(1, iso2)

                              !test reduce expansion local densities
                              IF(.NOT.(l <= lmax_expansion))CALL cp__a("qs_vxc_atom.F",1264)
                              DO ia = 1, na
                                 matso_h(iso1, iso2) = matso_h(iso1, iso2)+ &
                                                       gVXCg_h(ia, l)* &
                                                       harmonics%slm(ia, iso)* &
                                                       my_CG(iso1, iso2, iso)
                                 matso_s(iso1, iso2) = matso_s(iso1, iso2)+ &
                                                       gVXCg_s(ia, l)* &
                                                       harmonics%slm(ia, iso)* &
                                                       my_CG(iso1, iso2, iso)
                              END DO ! ia

                              !test reduce expansion local densities

                           END DO

                        END DO ! iso

                        DO iso = 1, dmax_iso_not0_local
                           DO icg = 1, dcg_n_list(iso)
                              iso1 = dcg_list(1, icg, iso)
                              iso2 = dcg_list(2, icg, iso)

                              l = indso(1, iso1)+indso(1, iso2)
                              !test reduce expansion local densities
                              IF(.NOT.(l <= lmax_expansion))CALL cp__a("qs_vxc_atom.F",1289)
                              DO ia = 1, na
                                 matso_h(iso1, iso2) = matso_h(iso1, iso2)+ &
                                                       (gVXGg_h(1, ia, l)*my_CG_dxyz(1, iso1, iso2, iso)+ &
                                                        gVXGg_h(2, ia, l)*my_CG_dxyz(2, iso1, iso2, iso)+ &
                                                        gVXGg_h(3, ia, l)*my_CG_dxyz(3, iso1, iso2, iso))* &
                                                       harmonics%slm(ia, iso)

                                 matso_s(iso1, iso2) = matso_s(iso1, iso2)+ &
                                                       (gVXGg_s(1, ia, l)*my_CG_dxyz(1, iso1, iso2, iso)+ &
                                                        gVXGg_s(2, ia, l)*my_CG_dxyz(2, iso1, iso2, iso)+ &
                                                        gVXGg_s(3, ia, l)*my_CG_dxyz(3, iso1, iso2, iso))* &
                                                       harmonics%slm(ia, iso)

                              END DO ! ia

                              !test reduce expansion local densities

                           END DO ! icg
                        END DO ! iso
                        !test reduce expansion local densities
                     END IF ! lmax_expansion

!             Write in the global matrix
                     DO ic = nsoset(lmin(iset2)-1)+1, nsoset(lmax(iset2))
                        iso1 = nsoset(lmin(iset1)-1)+1
                        iso2 = ngau2+ic
                        CALL daxpy(size1, 1.0_dp, matso_h(iso1, ic), 1, &
                                   int_hh(ispin)%r_coef(nngau1+1, iso2), 1)
                        CALL daxpy(size1, 1.0_dp, matso_s(iso1, ic), 1, &
                                   int_ss(ispin)%r_coef(nngau1+1, iso2), 1)
                     END DO

                  END DO ! ipfg2
               END DO ! ipfg1
               m2 = m2+maxso
            END DO ! iset2
            m1 = m1+maxso
         END DO ! iset1
      END DO ! ispin

      DEALLOCATE (g1, g2, gg, dgg, matso_h, matso_s, gVXCg_h, gVXCg_s, gVXGg_h, gVXGg_s)
      DEALLOCATE (cg_list, cg_n_list, dcg_list, dcg_n_list)

      CALL timestop(handle)

   END SUBROUTINE gaVxcgb_GC

! **************************************************************************************************
!> \brief ...
!> \param vtau_h ...
!> \param vtau_s ...
!> \param qs_kind ...
!> \param rho_atom ...
!> \param nspins ...
! **************************************************************************************************
   SUBROUTINE dgaVtaudgb(vtau_h, vtau_s, qs_kind, rho_atom, nspins)

      REAL(dp), DIMENSION(:, :, :), POINTER              :: vtau_h, vtau_s
      TYPE(qs_kind_type), INTENT(IN)                     :: qs_kind
      TYPE(rho_atom_type), POINTER                       :: rho_atom
      INTEGER, INTENT(IN)                                :: nspins

      CHARACTER(len=*), PARAMETER :: routineN = 'dgaVtaudgb', routineP = moduleN//':'//routineN

      INTEGER :: dmax_iso_not0_local, handle, ic, icg, ipgf1, ipgf2, iset1, iset2, iso, iso1, &
         iso2, isom1, isom2, ispin, l1, l2, lmax12, lmax_expansion, lmin12, m1, m2, max_iso_not0, &
         max_iso_not0_local, max_s_harm, maxiso, maxl, maxso, mm1, mm2, n1, n2, na, ngau1, ngau2, &
         nngau1, nr, nset, size1
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cg_n_list, dcg_n_list
      INTEGER, ALLOCATABLE, DIMENSION(:, :, :)           :: cg_list, dcg_list
      INTEGER, DIMENSION(:), POINTER                     :: lmax, lmin, npgf
      REAL(dp), ALLOCATABLE, DIMENSION(:)                :: dd, g1, g2, gg, gr, ww
      REAL(dp), ALLOCATABLE, DIMENSION(:, :)             :: dgr, dvth, dvts, matso_h, matso_s, vth, &
                                                            vts
      REAL(dp), ALLOCATABLE, DIMENSION(:, :, :)          :: fgr
      REAL(dp), DIMENSION(:, :), POINTER                 :: zet
      REAL(dp), DIMENSION(:, :, :), POINTER              :: my_CG, my_dCG
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis
      TYPE(harmonics_atom_type), POINTER                 :: harmonics
      TYPE(rho_atom_coeff), DIMENSION(:), POINTER        :: int_hh, int_ss

! -------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      NULLIFY (harmonics, grid_atom)
      CALL get_qs_kind(qs_kind, basis_set=orb_basis, &
                       harmonics=harmonics, grid_atom=grid_atom)

      NULLIFY (lmin, lmax, npgf, zet)
      CALL get_gto_basis_set(gto_basis_set=orb_basis, lmax=lmax, lmin=lmin, &
                             maxso=maxso, maxl=maxl, npgf=npgf, &
                             nset=nset, zet=zet)

      my_CG => harmonics%my_CG
      my_dCG => harmonics%my_dCG
      max_iso_not0 = harmonics%max_iso_not0
      lmax_expansion = indso(1, max_iso_not0)
      max_s_harm = harmonics%max_s_harm

      nr = grid_atom%nr
      na = grid_atom%ng_sphere
      ALLOCATE (g1(nr), g2(nr), gg(nr), dd(nr), gr(nr), ww(na))
      ALLOCATE (fgr(nr, 0:maxl, 0:maxl), dgr(nr, 0:lmax_expansion))

      ALLOCATE (cg_list(2, nsoset(maxl)**2, max_s_harm), cg_n_list(max_s_harm), &
                dcg_list(2, nsoset(maxl)**2, max_s_harm), dcg_n_list(max_s_harm))

      ALLOCATE (matso_h(nsoset(maxl), nsoset(maxl)), &
                matso_s(nsoset(maxl), nsoset(maxl)))

      NULLIFY (int_hh, int_ss)
      CALL get_rho_atom(rho_atom=rho_atom, ga_Vlocal_gb_h=int_hh, ga_Vlocal_gb_s=int_ss)

      vtau_h = 0.5_dp*vtau_h
      vtau_s = 0.5_dp*vtau_s

      DO ispin = 1, nspins
         maxiso = SIZE(harmonics%slm, 2)
         ALLOCATE (vth(nr, maxiso), vts(nr, maxiso))
         ALLOCATE (dvth(nr, maxiso), dvts(nr, maxiso))
         DO iso = 1, maxiso
            ww(1:na) = harmonics%slm(1:na, iso)
            vth(1:nr, iso) = MATMUL(ww(1:na), vtau_h(1:na, 1:nr, ispin))
            vts(1:nr, iso) = MATMUL(ww(1:na), vtau_s(1:na, 1:nr, ispin))
            ww(1:na) = harmonics%slm(1:na, iso)*grid_atom%usin_azi(1:na)**2
            dvth(1:nr, iso) = MATMUL(ww(1:na), vtau_h(1:na, 1:nr, ispin))
            dvts(1:nr, iso) = MATMUL(ww(1:na), vtau_s(1:na, 1:nr, ispin))
         END DO
         m1 = 0
         DO iset1 = 1, nset
            n1 = nsoset(lmax(iset1))
            m2 = 0
            DO iset2 = 1, nset
               CALL get_none0_cg_list(my_CG, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                      max_s_harm, lmax_expansion, cg_list, cg_n_list, max_iso_not0_local)
               IF(.NOT.(max_iso_not0_local .LE. max_iso_not0))CALL cp__a("qs_vxc_atom.F",1427)
               CALL get_none0_cg_list(my_dCG, lmin(iset1), lmax(iset1), lmin(iset2), lmax(iset2), &
                                      max_s_harm, lmax_expansion, dcg_list, dcg_n_list, dmax_iso_not0_local)

               n2 = nsoset(lmax(iset2))
               DO ipgf1 = 1, npgf(iset1)
                  ngau1 = n1*(ipgf1-1)+m1
                  size1 = nsoset(lmax(iset1))-nsoset(lmin(iset1)-1)
                  nngau1 = nsoset(lmin(iset1)-1)+ngau1
                  g1(1:nr) = EXP(-zet(ipgf1, iset1)*grid_atom%rad2(1:nr))
                  DO ipgf2 = 1, npgf(iset2)
                     ngau2 = n2*(ipgf2-1)+m2
                     g2(1:nr) = EXP(-zet(ipgf2, iset2)*grid_atom%rad2(1:nr))
                     lmin12 = lmin(iset1)+lmin(iset2)
                     lmax12 = lmax(iset1)+lmax(iset2)

                     IF (lmin12 .LE. lmax_expansion) THEN
                        fgr = 0._dp
                        dgr = 0._dp
                        DO l1 = 0, maxl
                           DO l2 = 0, maxl
                              IF (l1+l2 > lmax_expansion) CYCLE
                              IF (l1+l2 > 0) THEN
                                 gg(1:nr) = g1(1:nr)*g2(1:nr)*grid_atom%rad2l(1:nr, l1+l2)
                              ELSE
                                 gg(1:nr) = g1(1:nr)*g2(1:nr)
                              END IF
                              dd(1:nr) = REAL(l1*l2, dp)/grid_atom%rad2(1:nr)- &
                                         2._dp*REAL(l1, dp)*zet(ipgf2, iset2)- &
                                         2._dp*REAL(l2, dp)*zet(ipgf1, iset1)+ &
                                         4._dp*zet(ipgf1, iset1)*zet(ipgf2, iset2)*grid_atom%rad2(1:nr)
                              fgr(1:nr, l1, l2) = dd(1:nr)*gg(1:nr)
                              dgr(1:nr, l1+l2) = gg(1:nr)/grid_atom%rad2(1:nr)
                           END DO
                        END DO
                     END IF

                     matso_h = 0.0_dp
                     matso_s = 0.0_dp
                     IF (lmin12 .LE. lmax_expansion) THEN
                        DO iso = 1, max_iso_not0_local
                           DO icg = 1, cg_n_list(iso)
                              iso1 = cg_list(1, icg, iso)
                              iso2 = cg_list(2, icg, iso)

                              l1 = indso(1, iso1)
                              l2 = indso(1, iso2)
                              mm1 = indso(2, iso1)
                              mm2 = indso(2, iso2)
                              isom1 = soset(l1, -mm1)
                              isom2 = soset(l2, -mm2)

                              IF (l1+l2 > lmax_expansion) CYCLE

                              matso_h(iso1, iso2) = matso_h(iso1, iso2)+my_CG(iso1, iso2, iso)* &
                                                    DOT_PRODUCT(vth(1:nr, ispin), fgr(1:nr, l1, l2))
                              matso_s(iso1, iso2) = matso_s(iso1, iso2)+my_CG(iso1, iso2, iso)* &
                                                    DOT_PRODUCT(vts(1:nr, ispin), fgr(1:nr, l1, l2))
                              ! d azimuthal
                              matso_h(iso1, iso2) = matso_h(iso1, iso2)+REAL(mm1*mm2, dp)*my_CG(isom1, isom2, iso)* &
                                                    DOT_PRODUCT(vth(1:nr, ispin), dgr(1:nr, l1+l2))
                              matso_s(iso1, iso2) = matso_s(iso1, iso2)+REAL(mm1*mm2, dp)*my_CG(isom1, isom2, iso)* &
                                                    DOT_PRODUCT(vts(1:nr, ispin), dgr(1:nr, l1+l2))
                           END DO
                        END DO
                     END IF
                     ! d polar
                     DO iso = 1, dmax_iso_not0_local
                        DO icg = 1, dcg_n_list(iso)
                           iso1 = dcg_list(1, icg, iso)
                           iso2 = dcg_list(2, icg, iso)

                           l1 = indso(1, iso1)
                           l2 = indso(1, iso2)
                           matso_h(iso1, iso2) = matso_h(iso1, iso2)+my_dCG(iso1, iso2, iso)* &
                                                 DOT_PRODUCT(dvth(1:nr, ispin), dgr(1:nr, l1+l2))
                           matso_s(iso1, iso2) = matso_s(iso1, iso2)+my_dCG(iso1, iso2, iso)* &
                                                 DOT_PRODUCT(dvts(1:nr, ispin), dgr(1:nr, l1+l2))
                        END DO
                     END DO

                     ! Write in the global matrix
                     DO ic = nsoset(lmin(iset2)-1)+1, nsoset(lmax(iset2))
                        iso1 = nsoset(lmin(iset1)-1)+1
                        iso2 = ngau2+ic
                        CALL daxpy(size1, 1.0_dp, matso_h(iso1, ic), 1, int_hh(ispin)%r_coef(nngau1+1, iso2), 1)
                        CALL daxpy(size1, 1.0_dp, matso_s(iso1, ic), 1, int_ss(ispin)%r_coef(nngau1+1, iso2), 1)
                     END DO

                  END DO ! ipfg2
               END DO ! ipfg1
               m2 = m2+maxso
            END DO ! iset2
            m1 = m1+maxso
         END DO ! iset1
         DEALLOCATE (vth, vts, dvth, dvts)
      END DO ! ispin

      DEALLOCATE (matso_h, matso_s)
      DEALLOCATE (g1, g2, gg, dd, gr, ww)
      DEALLOCATE (fgr, dgr)
      DEALLOCATE (cg_list, cg_n_list, dcg_list, dcg_n_list)

      vtau_h = 2._dp*vtau_h
      vtau_s = 2._dp*vtau_s

      CALL timestop(handle)

   END SUBROUTINE dgaVtaudgb

END MODULE qs_vxc_atom
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Tools common both to PME and SPME
!> \par History
!>      JGH (03-May-2001) : first correctly working version
!>      teo (Feb-2007)    : Merging common routines to spme and pme
!> \author JGH (21-Mar-2001)
! **************************************************************************************************
MODULE pme_tools

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              real_to_scaled
   USE kinds,                           ONLY: dp
   USE particle_types,                  ONLY: particle_type
   USE realspace_grid_types,            ONLY: realspace_grid_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE
   PUBLIC :: get_center, set_list

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pme_tools'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param part ...
!> \param npart ...
!> \param center ...
!> \param p1 ...
!> \param rs ...
!> \param ipart ...
!> \param core_center ...
! **************************************************************************************************
   SUBROUTINE set_list(part, npart, center, p1, rs, ipart, core_center)

      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: part
      INTEGER, INTENT(IN)                                :: npart
      INTEGER, DIMENSION(:, :), INTENT(IN)               :: center
      INTEGER, INTENT(OUT)                               :: p1
      TYPE(realspace_grid_type), POINTER                 :: rs
      INTEGER, INTENT(INOUT)                             :: ipart
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: core_center

      INTEGER                                            :: ndim, npos
      INTEGER, DIMENSION(3)                              :: lb, ub
      REAL(KIND=dp)                                      :: charge
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      p1 = 0
      lb = rs%lb_real
      ub = rs%ub_real

      DO
         ipart = ipart+1
         IF (ipart > npart) EXIT
         atomic_kind => part(ipart)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, qeff=charge)
         IF (charge == 0.0_dp .AND. part(ipart)%shell_index == 0) CYCLE
         IF (rs%desc%parallel) THEN
            ! check if the rs grid is distributed or not
            IF (ALL(rs%desc%group_dim == 1)) THEN
               ndim = rs%desc%group_size
               npos = rs%desc%my_pos
               ! All processors work on the same grid
               IF (MOD(ipart, ndim) == npos) THEN
                  p1 = ipart
                  EXIT
               END IF
            ELSE
               ! First check if this atom is on my grid
               IF (part(ipart)%shell_index /= 0 .AND. PRESENT(core_center)) THEN
                  IF (in_slice(core_center(:, part(ipart)%shell_index), lb, ub)) THEN
                     p1 = ipart
                  ENDIF
               ELSE
                  IF (in_slice(center(:, ipart), lb, ub)) THEN
                     p1 = ipart
                     EXIT
                  END IF
               END IF
            ENDIF
         ELSE
            p1 = ipart
            EXIT
         END IF
      END DO

   END SUBROUTINE set_list

! **************************************************************************************************
!> \brief ...
!> \param pos ...
!> \param lb ...
!> \param ub ...
!> \retval internal ...
! **************************************************************************************************
   FUNCTION in_slice(pos, lb, ub) RESULT(internal)

      INTEGER, DIMENSION(3), INTENT(IN)                  :: pos, lb, ub
      LOGICAL                                            :: internal

      IF (ALL(pos >= lb) .AND. ALL(pos <= ub)) THEN
         internal = .TRUE.
      ELSE
         internal = .FALSE.
      END IF

   END FUNCTION in_slice

! **************************************************************************************************
!> \brief ...
!> \param part ...
!> \param box ...
!> \param center ...
!> \param npts ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE get_center(part, box, center, npts, n)

      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: part
      TYPE(cell_type), POINTER                           :: box
      INTEGER, DIMENSION(:, :), INTENT(OUT)              :: center
      INTEGER, DIMENSION(:), INTENT(IN)                  :: npts
      INTEGER, INTENT(IN)                                :: n

      INTEGER                                            :: ipart, mp
      REAL(KIND=dp)                                      :: rmp
      REAL(KIND=dp), DIMENSION(3)                        :: gp, s

      mp = MAXVAL(npts(:))
      rmp = REAL(mp, KIND=dp)
      DO ipart = 1, SIZE(part)
         ! compute the scaled coordinate of atom ipart
         CALL real_to_scaled(s, part(ipart)%r, box)
         s = s-NINT(s)
         gp = REAL(npts, KIND=dp)*s
         ! find the closest grid point (on big grid)
         IF (MOD(n, 2) == 0) THEN
            center(:, ipart) = INT(gp+rmp)-mp
         ELSE
            center(:, ipart) = NINT(gp)
         END IF
         center(:, ipart) = center(:, ipart)+npts(:)/2
         center(:, ipart) = MODULO(center(:, ipart), npts(:))
         center(:, ipart) = center(:, ipart)-npts(:)/2
      END DO

   END SUBROUTINE get_center

END MODULE pme_tools

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculate the electrostatic energy by the Smooth Particle Ewald method
!> \par History
!>      JGH (03-May-2001) : first correctly working version
!> \author JGH (21-Mar-2001)
! **************************************************************************************************
MODULE spme

   USE atomic_kind_types,               ONLY: get_atomic_kind
   USE atprop_types,                    ONLY: atprop_type
   USE bibliography,                    ONLY: Essmann1995,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type,&
                                              real_to_scaled
   USE dgs,                             ONLY: dg_sum_patch,&
                                              dg_sum_patch_force_1d,&
                                              dg_sum_patch_force_3d
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_get,&
                                              ewald_pw_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi
   USE particle_types,                  ONLY: particle_type
   USE pme_tools,                       ONLY: get_center,&
                                              set_list
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_grids,                        ONLY: get_pw_grid_info
   USE pw_methods,                      ONLY: pw_copy,&
                                              pw_derive,&
                                              pw_integral_a2b,&
                                              pw_transfer
   USE pw_poisson_methods,              ONLY: pw_poisson_rebuild,&
                                              pw_poisson_solve
   USE pw_poisson_types,                ONLY: greens_fn_type,&
                                              pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE realspace_grid_types,            ONLY: &
        pw2rs, realspace_grid_desc_type, realspace_grid_p_type, realspace_grid_type, rs2pw, &
        rs_grid_create, rs_grid_release, rs_grid_set_box, rs_grid_zero, rs_pw_transfer
   USE shell_potential_types,           ONLY: shell_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'spme'

   PRIVATE
   PUBLIC :: spme_evaluate, spme_potential, spme_forces, get_patch

   INTERFACE get_patch
      MODULE PROCEDURE get_patch_a, get_patch_b
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param box ...
!> \param particle_set ...
!> \param fg_coulomb ...
!> \param vg_coulomb ...
!> \param pv_g ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param fgshell_coulomb ...
!> \param fgcore_coulomb ...
!> \param use_virial ...
!> \param charges ...
!> \param atprop ...
!> \par History
!>      JGH (03-May-2001) : SPME with charge definition
!> \author JGH (21-Mar-2001)
! **************************************************************************************************
   SUBROUTINE spme_evaluate(ewald_env, ewald_pw, box, particle_set, &
                            fg_coulomb, vg_coulomb, pv_g, shell_particle_set, core_particle_set, &
                            fgshell_coulomb, fgcore_coulomb, use_virial, charges, atprop)

      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(cell_type), POINTER                           :: box
      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: particle_set
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: fg_coulomb
      REAL(KIND=dp), INTENT(OUT)                         :: vg_coulomb
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: pv_g
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: shell_particle_set, core_particle_set
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT), &
         OPTIONAL                                        :: fgshell_coulomb, fgcore_coulomb
      LOGICAL, INTENT(IN)                                :: use_virial
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges
      TYPE(atprop_type), POINTER                         :: atprop

      CHARACTER(len=*), PARAMETER :: routineN = 'spme_evaluate', routineP = moduleN//':'//routineN

      INTEGER                                            :: group, handle, i, ig, ipart, j, n, &
                                                            ncore, npart, nshell, o_spline, p1, &
                                                            p1_shell
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: center, core_center, shell_center
      INTEGER, DIMENSION(3)                              :: nd, npts
      LOGICAL                                            :: do_shell
      REAL(KIND=dp)                                      :: alpha, dvols, fat1, ffa, ffb
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rhos
      REAL(KIND=dp), DIMENSION(3)                        :: fat
      REAL(KIND=dp), DIMENSION(3, 3)                     :: f_stress, h_stress
      TYPE(greens_fn_type), POINTER                      :: green
      TYPE(pw_grid_type), POINTER                        :: grid_spme
      TYPE(pw_p_type), DIMENSION(3)                      :: dphi_g
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(pw_type), POINTER                             :: phi_g, rhob_g, rhob_r
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: drpot
      TYPE(realspace_grid_type), POINTER                 :: rden, rpot

      NULLIFY (drpot, grid_spme, green, poisson_env, phi_g, rhob_g, rhob_r, pw_pool, rden, rpot)
      CALL timeset(routineN, handle)
      CALL cite_reference(Essmann1995)

      !-------------- INITIALISATION ---------------------
      CALL ewald_env_get(ewald_env, alpha=alpha, o_spline=o_spline, &
                         group=group)
      CALL ewald_pw_get(ewald_pw, pw_big_pool=pw_pool, rs_desc=rs_desc, &
                        poisson_env=poisson_env)
      CALL pw_poisson_rebuild(poisson_env)
      green => poisson_env%green_fft
      grid_spme => pw_pool%pw_grid

      npart = SIZE(particle_set)

      CALL get_pw_grid_info(grid_spme, npts=npts, dvol=dvols)

      n = o_spline
      ALLOCATE (rhos(n, n, n))
      CALL rs_grid_create(rden, rs_desc)
      CALL rs_grid_set_box(grid_spme, rs=rden)
      CALL rs_grid_zero(rden)

      ALLOCATE (center(3, npart))
      CALL get_center(particle_set, box, center, npts, n)
      IF (PRESENT(shell_particle_set) .AND. (PRESENT(core_particle_set))) THEN
         IF(.NOT.(ASSOCIATED(shell_particle_set)))CALL cp__a("spme.F",157)
         IF(.NOT.(ASSOCIATED(core_particle_set)))CALL cp__a("spme.F",158)
         nshell = SIZE(shell_particle_set)
         ncore = SIZE(core_particle_set)
         IF(.NOT.(nshell == ncore))CALL cp__a("spme.F",161)
         ALLOCATE (shell_center(3, nshell))
         CALL get_center(shell_particle_set, box, shell_center, npts, n)
         ALLOCATE (core_center(3, nshell))
         CALL get_center(core_particle_set, box, core_center, npts, n)
      END IF

      !-------------- DENSITY CALCULATION ----------------
      ipart = 0
      ! Particles
      DO
         CALL set_list(particle_set, npart, center, p1, rden, ipart)
         IF (p1 == 0) EXIT

         do_shell = (particle_set(p1)%shell_index /= 0)
         IF (do_shell) CYCLE
         ! calculate function on small boxes
         CALL get_patch(particle_set, box, green, npts, p1, rhos, is_core=.FALSE., &
                        is_shell=.FALSE., unit_charge=.FALSE., charges=charges)

         ! add boxes to real space grid (big box)
         CALL dg_sum_patch(rden, rhos, center(:, p1))
      END DO
      ! Shell-Model
      IF (PRESENT(shell_particle_set) .AND. PRESENT(core_particle_set)) THEN
         ipart = 0
         DO
            ! calculate function on small boxes
            CALL set_list(shell_particle_set, nshell, shell_center, p1_shell, &
                          rden, ipart)
            IF (p1_shell == 0) EXIT
            CALL get_patch(shell_particle_set, box, green, npts, p1_shell, rhos, &
                           is_core=.FALSE., is_shell=.TRUE., unit_charge=.FALSE.)

            ! add boxes to real space grid (big box)
            CALL dg_sum_patch(rden, rhos, shell_center(:, p1_shell))
         END DO
         ipart = 0
         DO
            ! calculate function on small boxes
            CALL set_list(core_particle_set, nshell, core_center, p1_shell, &
                          rden, ipart)
            IF (p1_shell == 0) EXIT
            CALL get_patch(core_particle_set, box, green, npts, p1_shell, rhos, &
                           is_core=.TRUE., is_shell=.FALSE., unit_charge=.FALSE.)

            ! add boxes to real space grid (big box)
            CALL dg_sum_patch(rden, rhos, core_center(:, p1_shell))
         END DO
      END IF
      !----------- END OF DENSITY CALCULATION -------------

      CALL pw_pool_create_pw(pw_pool, rhob_r, use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL rs_pw_transfer(rden, rhob_r, rs2pw)
      ! transform density to G space and add charge function
      CALL pw_pool_create_pw(pw_pool, rhob_g, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_transfer(rhob_r, rhob_g)
      ! update charge function
      rhob_g%cc = rhob_g%cc*green%p3m_charge%cr

      !-------------- ELECTROSTATIC CALCULATION -----------
      ! allocate intermediate arrays
      DO i = 1, 3
         NULLIFY (dphi_g(i)%pw)
         CALL pw_pool_create_pw(pw_pool, dphi_g(i)%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
      END DO
      CALL pw_pool_create_pw(pw_pool, phi_g, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_poisson_solve(poisson_env, rhob_g, vg_coulomb, phi_g, dphi_g, &
                            h_stress)
      !---------- END OF ELECTROSTATIC CALCULATION --------

      ! Atomic Energy and Stress
      IF (atprop%energy .OR. atprop%stress) THEN
         CALL rs_grid_create(rpot, rs_desc)
         CALL rs_grid_set_box(grid_spme, rs=rpot)
         phi_g%cc = phi_g%cc*green%p3m_charge%cr
         CALL pw_transfer(phi_g, rhob_r)
         CALL rs_pw_transfer(rpot, rhob_r, pw2rs)
         ipart = 0
         DO
            CALL set_list(particle_set, npart, center, p1, rden, ipart)
            IF (p1 == 0) EXIT
            do_shell = (particle_set(p1)%shell_index /= 0)
            IF (do_shell) CYCLE
            ! calculate function on small boxes
            CALL get_patch(particle_set, box, green, grid_spme%npts, p1, rhos, is_core=.FALSE., &
                           is_shell=.FALSE., unit_charge=.FALSE., charges=charges)
            ! integrate box and potential
            CALL dg_sum_patch_force_1d(rpot, rhos, center(:, p1), fat1)
            IF (atprop%energy) THEN
               atprop%atener(p1) = atprop%atener(p1)+0.5_dp*fat1*dvols
            END IF
            IF (atprop%stress) THEN
               atprop%atstress(1, 1, p1) = atprop%atstress(1, 1, p1)+0.5_dp*fat1*dvols
               atprop%atstress(2, 2, p1) = atprop%atstress(2, 2, p1)+0.5_dp*fat1*dvols
               atprop%atstress(3, 3, p1) = atprop%atstress(3, 3, p1)+0.5_dp*fat1*dvols
            END IF
         END DO
         ! Core-shell model
         IF (PRESENT(shell_particle_set) .AND. PRESENT(core_particle_set)) THEN
            ipart = 0
            DO
               CALL set_list(shell_particle_set, nshell, shell_center, p1_shell, rden, ipart)
               IF (p1_shell == 0) EXIT
               CALL get_patch(shell_particle_set, box, green, npts, p1_shell, rhos, &
                              is_core=.FALSE., is_shell=.TRUE., unit_charge=.FALSE.)
               CALL dg_sum_patch_force_1d(rpot, rhos, shell_center(:, p1_shell), fat1)
               p1 = shell_particle_set(p1_shell)%atom_index
               IF (atprop%energy) THEN
                  atprop%atener(p1) = atprop%atener(p1)+0.5_dp*fat1*dvols
               END IF
            END DO
            ipart = 0
            DO
               CALL set_list(core_particle_set, nshell, core_center, p1_shell, rden, ipart)
               IF (p1_shell == 0) EXIT
               CALL get_patch(core_particle_set, box, green, npts, p1_shell, rhos, &
                              is_core=.TRUE., is_shell=.FALSE., unit_charge=.FALSE.)
               CALL dg_sum_patch_force_1d(rpot, rhos, core_center(:, p1_shell), fat1)
               p1 = core_particle_set(p1_shell)%atom_index
               IF (atprop%energy) THEN
                  atprop%atener(p1) = atprop%atener(p1)+0.5_dp*fat1*dvols
               END IF
            END DO
         END IF
         IF (atprop%stress) THEN
            ffa = (0.5_dp/alpha)**2
            ffb = 1.0_dp/fourpi
            DO i = 1, 3
               DO ig = grid_spme%first_gne0, grid_spme%ngpts_cut_local
                  phi_g%cc(ig) = ffb*dphi_g(i)%pw%cc(ig)*(ffa*grid_spme%gsq(ig)+1.0_dp)
                  phi_g%cc(ig) = phi_g%cc(ig)*poisson_env%green_fft%influence_fn%cc(ig)
               END DO
               IF (grid_spme%have_g0) phi_g%cc(1) = 0.0_dp
               DO j = 1, i
                  nd = 0
                  nd(j) = 1
                  CALL pw_copy(phi_g, rhob_g)
                  CALL pw_derive(rhob_g, nd)
                  rhob_g%cc = rhob_g%cc*green%p3m_charge%cr
                  CALL pw_transfer(rhob_g, rhob_r)
                  CALL rs_pw_transfer(rpot, rhob_r, pw2rs)

                  ipart = 0
                  DO
                     CALL set_list(particle_set, npart, center, p1, rden, ipart)
                     IF (p1 == 0) EXIT
                     ! calculate function on small boxes
                     CALL get_patch(particle_set, box, green, grid_spme%npts, p1, rhos, &
                                    is_core=.FALSE., is_shell=.FALSE., unit_charge=.FALSE., charges=charges)
                     ! integrate box and potential
                     CALL dg_sum_patch_force_1d(rpot, rhos, center(:, p1), fat1)
                     atprop%atstress(i, j, p1) = atprop%atstress(i, j, p1)+fat1*dvols
                     IF (i /= j) atprop%atstress(j, i, p1) = atprop%atstress(j, i, p1)+fat1*dvols
                  END DO

               END DO
            END DO
         END IF
         CALL rs_grid_release(rpot)
      END IF

      CALL pw_pool_give_back_pw(pw_pool, phi_g)
      CALL pw_pool_give_back_pw(pw_pool, rhob_g)

      !------------- STRESS TENSOR CALCULATION ------------
      IF (use_virial) THEN
         DO i = 1, 3
            DO j = i, 3
               f_stress(i, j) = pw_integral_a2b(dphi_g(i)%pw, dphi_g(j)%pw)
               f_stress(j, i) = f_stress(i, j)
            END DO
         END DO
         ffa = (1.0_dp/fourpi)*(0.5_dp/alpha)**2
         f_stress = -ffa*f_stress
         pv_g = h_stress+f_stress
      END IF
      !--------END OF STRESS TENSOR CALCULATION -----------
      ! move derivative of potential to real space grid and
      ! multiply by charge function in g-space
      ALLOCATE (drpot(1:3))
      DO i = 1, 3
         CALL rs_grid_create(drpot(i)%rs_grid, rs_desc)
         CALL rs_grid_set_box(grid_spme, rs=drpot(i)%rs_grid)
         dphi_g(i)%pw%cc = dphi_g(i)%pw%cc*green%p3m_charge%cr
         CALL pw_transfer(dphi_g(i)%pw, rhob_r)
         CALL pw_pool_give_back_pw(pw_pool, dphi_g(i)%pw)
         CALL rs_pw_transfer(drpot(i)%rs_grid, rhob_r, pw2rs)
      END DO

      CALL pw_pool_give_back_pw(pw_pool, rhob_r)
      !----------------- FORCE CALCULATION ----------------
      ! initialize the forces
      fg_coulomb = 0.0_dp
      ! Particles
      ipart = 0
      DO
         CALL set_list(particle_set, npart, center, p1, rden, ipart)
         IF (p1 == 0) EXIT

         do_shell = (particle_set(p1)%shell_index /= 0)
         IF (do_shell) CYCLE
         ! calculate function on small boxes
         CALL get_patch(particle_set, box, green, grid_spme%npts, p1, rhos, is_core=.FALSE., &
                        is_shell=.FALSE., unit_charge=.FALSE., charges=charges)

         ! add boxes to real space grid (big box)
         CALL dg_sum_patch_force_3d(drpot, rhos, center(:, p1), fat)
         fg_coulomb(1, p1) = fg_coulomb(1, p1)-fat(1)*dvols
         fg_coulomb(2, p1) = fg_coulomb(2, p1)-fat(2)*dvols
         fg_coulomb(3, p1) = fg_coulomb(3, p1)-fat(3)*dvols
      END DO
      ! Shell-Model
      IF (PRESENT(shell_particle_set) .AND. (PRESENT(core_particle_set))) THEN
         IF (PRESENT(fgshell_coulomb)) THEN
            ipart = 0
            fgshell_coulomb = 0.0_dp
            DO
               ! calculate function on small boxes
               CALL set_list(shell_particle_set, nshell, shell_center, p1_shell, &
                             rden, ipart)
               IF (p1_shell == 0) EXIT

               CALL get_patch(shell_particle_set, box, green, grid_spme%npts, &
                              p1_shell, rhos, is_core=.FALSE., is_shell=.TRUE., unit_charge=.FALSE.)

               ! add boxes to real space grid (big box)
               CALL dg_sum_patch_force_3d(drpot, rhos, shell_center(:, p1_shell), fat)
               fgshell_coulomb(1, p1_shell) = fgshell_coulomb(1, p1_shell)-fat(1)*dvols
               fgshell_coulomb(2, p1_shell) = fgshell_coulomb(2, p1_shell)-fat(2)*dvols
               fgshell_coulomb(3, p1_shell) = fgshell_coulomb(3, p1_shell)-fat(3)*dvols

            END DO
         END IF
         IF (PRESENT(fgcore_coulomb)) THEN
            ipart = 0
            fgcore_coulomb = 0.0_dp
            DO
               ! calculate function on small boxes
               CALL set_list(core_particle_set, nshell, core_center, p1_shell, &
                             rden, ipart)
               IF (p1_shell == 0) EXIT

               CALL get_patch(core_particle_set, box, green, grid_spme%npts, &
                              p1_shell, rhos, is_core=.TRUE., is_shell=.FALSE., unit_charge=.FALSE.)

               ! add boxes to real space grid (big box)
               CALL dg_sum_patch_force_3d(drpot, rhos, core_center(:, p1_shell), fat)
               fgcore_coulomb(1, p1_shell) = fgcore_coulomb(1, p1_shell)-fat(1)*dvols
               fgcore_coulomb(2, p1_shell) = fgcore_coulomb(2, p1_shell)-fat(2)*dvols
               fgcore_coulomb(3, p1_shell) = fgcore_coulomb(3, p1_shell)-fat(3)*dvols
            END DO
         END IF

      END IF
      !--------------END OF FORCE CALCULATION -------------

      !------------------CLEANING UP ----------------------
      CALL rs_grid_release(rden)
      IF (ASSOCIATED(drpot)) THEN
         DO i = 1, 3
            CALL rs_grid_release(drpot(i)%rs_grid)
         END DO
         DEALLOCATE (drpot)
      END IF

      DEALLOCATE (rhos)
      DEALLOCATE (center)
      IF (ALLOCATED(shell_center)) THEN
         DEALLOCATE (shell_center)
      END IF
      IF (ALLOCATED(core_center)) THEN
         DEALLOCATE (core_center)
      END IF
      CALL timestop(handle)

   END SUBROUTINE spme_evaluate

! **************************************************************************************************
!> \brief Calculate the electrostatic potential from particles A (charge A) at
!>        positions of particles B
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param box ...
!> \param particle_set_a ...
!> \param charges_a ...
!> \param particle_set_b ...
!> \param potential ...
!> \par History
!>      JGH (23-Oct-2014) : adapted from SPME evaluate
!> \author JGH (23-Oct-2014)
! **************************************************************************************************
   SUBROUTINE spme_potential(ewald_env, ewald_pw, box, particle_set_a, charges_a, &
                             particle_set_b, potential)

      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(cell_type), POINTER                           :: box
      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: particle_set_a
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges_a
      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: particle_set_b
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: potential

      CHARACTER(len=*), PARAMETER :: routineN = 'spme_potential', routineP = moduleN//':'//routineN

      INTEGER                                            :: group, handle, ipart, n, npart_a, &
                                                            npart_b, o_spline, p1
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: center
      INTEGER, DIMENSION(3)                              :: npts
      REAL(KIND=dp)                                      :: alpha, dvols, fat1
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rhos
      TYPE(greens_fn_type), POINTER                      :: green
      TYPE(pw_grid_type), POINTER                        :: grid_spme
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(pw_type), POINTER                             :: phi_g, rhob_g, rhob_r
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_type), POINTER                 :: rden, rpot

      NULLIFY (grid_spme, green, poisson_env, phi_g, rhob_g, rhob_r, pw_pool, &
               rden, rpot)
      CALL timeset(routineN, handle)
      CALL cite_reference(Essmann1995)

      !-------------- INITIALISATION ---------------------
      CALL ewald_env_get(ewald_env, alpha=alpha, o_spline=o_spline, group=group)
      CALL ewald_pw_get(ewald_pw, pw_big_pool=pw_pool, rs_desc=rs_desc, poisson_env=poisson_env)
      CALL pw_poisson_rebuild(poisson_env)
      green => poisson_env%green_fft
      grid_spme => pw_pool%pw_grid

      npart_a = SIZE(particle_set_a)
      npart_b = SIZE(particle_set_b)

      CALL get_pw_grid_info(grid_spme, npts=npts, dvol=dvols)

      n = o_spline
      ALLOCATE (rhos(n, n, n))
      CALL rs_grid_create(rden, rs_desc)
      CALL rs_grid_set_box(grid_spme, rs=rden)
      CALL rs_grid_zero(rden)

      ALLOCATE (center(3, npart_a))
      CALL get_center(particle_set_a, box, center, npts, n)

      !-------------- DENSITY CALCULATION ----------------
      ipart = 0
      ! Particles
      DO
         CALL set_list(particle_set_a, npart_a, center, p1, rden, ipart)
         IF (p1 == 0) EXIT

         ! calculate function on small boxes
         CALL get_patch(particle_set_a, box, green, npts, p1, rhos, charges_a)

         ! add boxes to real space grid (big box)
         CALL dg_sum_patch(rden, rhos, center(:, p1))
      END DO
      DEALLOCATE (center)
      !----------- END OF DENSITY CALCULATION -------------

      CALL pw_pool_create_pw(pw_pool, rhob_r, use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL rs_pw_transfer(rden, rhob_r, rs2pw)
      ! transform density to G space and add charge function
      CALL pw_pool_create_pw(pw_pool, rhob_g, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_transfer(rhob_r, rhob_g)
      ! update charge function
      rhob_g%cc = rhob_g%cc*green%p3m_charge%cr

      !-------------- ELECTROSTATIC CALCULATION -----------
      ! allocate intermediate arrays
      CALL pw_pool_create_pw(pw_pool, phi_g, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_poisson_solve(poisson_env, density=rhob_g, vhartree=phi_g)
      !---------- END OF ELECTROSTATIC CALCULATION --------

      !-------------- POTENTAIL AT ATOMIC POSITIONS -------
      ALLOCATE (center(3, npart_b))
      CALL get_center(particle_set_b, box, center, npts, n)
      rpot => rden
      phi_g%cc = phi_g%cc*green%p3m_charge%cr
      CALL pw_transfer(phi_g, rhob_r)
      CALL rs_pw_transfer(rpot, rhob_r, pw2rs)
      ipart = 0
      DO
         CALL set_list(particle_set_b, npart_b, center, p1, rden, ipart)
         IF (p1 == 0) EXIT
         ! calculate function on small boxes
         CALL get_patch(particle_set_b, box, green, grid_spme%npts, p1, rhos, &
                        is_core=.FALSE., is_shell=.FALSE., unit_charge=.TRUE.)
         ! integrate box and potential
         CALL dg_sum_patch_force_1d(rpot, rhos, center(:, p1), fat1)
         potential(p1) = potential(p1)+fat1*dvols
      END DO

      !------------------CLEANING UP ----------------------
      CALL pw_pool_give_back_pw(pw_pool, phi_g)
      CALL pw_pool_give_back_pw(pw_pool, rhob_g)
      CALL pw_pool_give_back_pw(pw_pool, rhob_r)
      CALL rs_grid_release(rden)

      DEALLOCATE (rhos)
      DEALLOCATE (center)
      CALL timestop(handle)

   END SUBROUTINE spme_potential

! **************************************************************************************************
!> \brief Calculate the forces on particles B for the electrostatic interaction
!>        betrween particles A and B
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param box ...
!> \param particle_set_a ...
!> \param charges_a ...
!> \param particle_set_b ...
!> \param charges_b ...
!> \param forces_b ...
!> \par History
!>      JGH (27-Oct-2014) : adapted from SPME evaluate
!> \author JGH (23-Oct-2014)
! **************************************************************************************************
   SUBROUTINE spme_forces(ewald_env, ewald_pw, box, particle_set_a, charges_a, &
                          particle_set_b, charges_b, forces_b)

      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(cell_type), POINTER                           :: box
      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: particle_set_a
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges_a
      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: particle_set_b
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges_b
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: forces_b

      CHARACTER(len=*), PARAMETER :: routineN = 'spme_forces', routineP = moduleN//':'//routineN

      INTEGER                                            :: group, handle, i, ipart, n, npart_a, &
                                                            npart_b, o_spline, p1
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: center
      INTEGER, DIMENSION(3)                              :: npts
      REAL(KIND=dp)                                      :: alpha, dvols
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rhos
      REAL(KIND=dp), DIMENSION(3)                        :: fat
      TYPE(greens_fn_type), POINTER                      :: green
      TYPE(pw_grid_type), POINTER                        :: grid_spme
      TYPE(pw_p_type), DIMENSION(3)                      :: dphi_g
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(pw_type), POINTER                             :: phi_g, rhob_g, rhob_r
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: drpot
      TYPE(realspace_grid_type), POINTER                 :: rden

      NULLIFY (drpot, grid_spme, green, poisson_env, phi_g, rhob_g, rhob_r, &
               pw_pool, rden)
      CALL timeset(routineN, handle)
      CALL cite_reference(Essmann1995)

      !-------------- INITIALISATION ---------------------
      CALL ewald_env_get(ewald_env, alpha=alpha, o_spline=o_spline, group=group)
      CALL ewald_pw_get(ewald_pw, pw_big_pool=pw_pool, rs_desc=rs_desc, poisson_env=poisson_env)
      CALL pw_poisson_rebuild(poisson_env)
      green => poisson_env%green_fft
      grid_spme => pw_pool%pw_grid

      npart_a = SIZE(particle_set_a)
      npart_b = SIZE(particle_set_b)

      CALL get_pw_grid_info(grid_spme, npts=npts, dvol=dvols)

      n = o_spline
      ALLOCATE (rhos(n, n, n))
      CALL rs_grid_create(rden, rs_desc)
      CALL rs_grid_set_box(grid_spme, rs=rden)
      CALL rs_grid_zero(rden)

      ALLOCATE (center(3, npart_a))
      CALL get_center(particle_set_a, box, center, npts, n)

      !-------------- DENSITY CALCULATION ----------------
      ipart = 0
      ! Particles
      DO
         CALL set_list(particle_set_a, npart_a, center, p1, rden, ipart)
         IF (p1 == 0) EXIT

         ! calculate function on small boxes
         CALL get_patch(particle_set_a, box, green, npts, p1, rhos, charges_a)

         ! add boxes to real space grid (big box)
         CALL dg_sum_patch(rden, rhos, center(:, p1))
      END DO
      DEALLOCATE (center)
      !----------- END OF DENSITY CALCULATION -------------

      CALL pw_pool_create_pw(pw_pool, rhob_r, use_data=REALDATA3D, &
                             in_space=REALSPACE)
      CALL rs_pw_transfer(rden, rhob_r, rs2pw)
      ! transform density to G space and add charge function
      CALL pw_pool_create_pw(pw_pool, rhob_g, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_transfer(rhob_r, rhob_g)
      ! update charge function
      rhob_g%cc = rhob_g%cc*green%p3m_charge%cr

      !-------------- ELECTROSTATIC CALCULATION -----------
      ! allocate intermediate arrays
      DO i = 1, 3
         NULLIFY (dphi_g(i)%pw)
         CALL pw_pool_create_pw(pw_pool, dphi_g(i)%pw, &
                                use_data=COMPLEXDATA1D, &
                                in_space=RECIPROCALSPACE)
      END DO
      CALL pw_pool_create_pw(pw_pool, phi_g, &
                             use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_poisson_solve(poisson_env, density=rhob_g, vhartree=phi_g, &
                            dvhartree=dphi_g)
      !---------- END OF ELECTROSTATIC CALCULATION --------
      ! move derivative of potential to real space grid and
      ! multiply by charge function in g-space
      ALLOCATE (drpot(1:3))
      DO i = 1, 3
         CALL rs_grid_create(drpot(i)%rs_grid, rs_desc)
         CALL rs_grid_set_box(grid_spme, rs=drpot(i)%rs_grid)
         dphi_g(i)%pw%cc = dphi_g(i)%pw%cc*green%p3m_charge%cr
         CALL pw_transfer(dphi_g(i)%pw, rhob_r)
         CALL pw_pool_give_back_pw(pw_pool, dphi_g(i)%pw)
         CALL rs_pw_transfer(drpot(i)%rs_grid, rhob_r, pw2rs)
      END DO
      !----------------- FORCE CALCULATION ----------------
      ALLOCATE (center(3, npart_b))
      CALL get_center(particle_set_b, box, center, npts, n)
      ipart = 0
      DO
         CALL set_list(particle_set_b, npart_b, center, p1, rden, ipart)
         IF (p1 == 0) EXIT
         ! calculate function on small boxes
         CALL get_patch(particle_set_b, box, green, grid_spme%npts, p1, rhos, &
                        is_core=.FALSE., is_shell=.FALSE., unit_charge=.FALSE., charges=charges_b)
         ! add boxes to real space grid (big box)
         CALL dg_sum_patch_force_3d(drpot, rhos, center(:, p1), fat)
         forces_b(1, p1) = forces_b(1, p1)-fat(1)*dvols
         forces_b(2, p1) = forces_b(2, p1)-fat(2)*dvols
         forces_b(3, p1) = forces_b(3, p1)-fat(3)*dvols
      END DO
      !------------------CLEANING UP ----------------------
      IF (ASSOCIATED(drpot)) THEN
         DO i = 1, 3
            CALL rs_grid_release(drpot(i)%rs_grid)
         END DO
         DEALLOCATE (drpot)
      END IF
      CALL pw_pool_give_back_pw(pw_pool, phi_g)
      CALL pw_pool_give_back_pw(pw_pool, rhob_g)
      CALL pw_pool_give_back_pw(pw_pool, rhob_r)
      CALL rs_grid_release(rden)

      DEALLOCATE (rhos)
      DEALLOCATE (center)
      CALL timestop(handle)

   END SUBROUTINE spme_forces

! **************************************************************************************************
!> \brief Calculates local density in a small box
!> \param part ...
!> \param box ...
!> \param green ...
!> \param npts ...
!> \param p ...
!> \param rhos ...
!> \param is_core ...
!> \param is_shell ...
!> \param unit_charge ...
!> \param charges ...
!> \par History
!>      none
!> \author JGH (21-Mar-2001)
! **************************************************************************************************
   SUBROUTINE get_patch_a(part, box, green, npts, p, rhos, is_core, is_shell, &
                          unit_charge, charges)

      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: part
      TYPE(cell_type), POINTER                           :: box
      TYPE(greens_fn_type), POINTER                      :: green
      INTEGER, DIMENSION(3), INTENT(IN)                  :: npts
      INTEGER, INTENT(IN)                                :: p
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(OUT)     :: rhos
      LOGICAL, INTENT(IN)                                :: is_core, is_shell, unit_charge
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges

      CHARACTER(len=*), PARAMETER :: routineN = 'get_patch_a', routineP = moduleN//':'//routineN

      INTEGER                                            :: nbox
      LOGICAL                                            :: use_charge_array
      REAL(KIND=dp)                                      :: q
      REAL(KIND=dp), DIMENSION(3)                        :: delta, r
      TYPE(shell_kind_type), POINTER                     :: shell

      NULLIFY (shell)
      use_charge_array = PRESENT(charges)
      IF (use_charge_array) use_charge_array = ASSOCIATED(charges)
      IF (is_core .AND. is_shell) THEN
         CALL cp__b("spme.F",777,"Shell-model: cannot be core and shell simultaneously")
      END IF

      nbox = SIZE(rhos, 1)
      r = part(p)%r
      q = 1.0_dp
      IF (.NOT. unit_charge) THEN
         IF (is_core) THEN
            CALL get_atomic_kind(atomic_kind=part(p)%atomic_kind, shell=shell)
            q = shell%charge_core
         ELSE IF (is_shell) THEN
            CALL get_atomic_kind(atomic_kind=part(p)%atomic_kind, shell=shell)
            q = shell%charge_shell
         ELSE
            CALL get_atomic_kind(atomic_kind=part(p)%atomic_kind, qeff=q)
         END IF
         IF (use_charge_array) q = charges(p)
      END IF
      CALL get_delta(box, r, npts, delta, nbox)
      CALL spme_get_patch(rhos, nbox, delta, q, green%p3m_coeff)

   END SUBROUTINE get_patch_a

! **************************************************************************************************
!> \brief Calculates local density in a small box
!> \param part ...
!> \param box ...
!> \param green ...
!> \param npts ...
!> \param p ...
!> \param rhos ...
!> \param charges ...
!> \par History
!>      none
!> \author JGH (21-Mar-2001)
! **************************************************************************************************
   SUBROUTINE get_patch_b(part, box, green, npts, p, rhos, charges)

      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: part
      TYPE(cell_type), POINTER                           :: box
      TYPE(greens_fn_type), POINTER                      :: green
      INTEGER, DIMENSION(3), INTENT(IN)                  :: npts
      INTEGER, INTENT(IN)                                :: p
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(OUT)     :: rhos
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges

      CHARACTER(len=*), PARAMETER :: routineN = 'get_patch_b', routineP = moduleN//':'//routineN

      INTEGER                                            :: nbox
      REAL(KIND=dp)                                      :: q
      REAL(KIND=dp), DIMENSION(3)                        :: delta, r

      nbox = SIZE(rhos, 1)
      r = part(p)%r
      q = charges(p)
      CALL get_delta(box, r, npts, delta, nbox)
      CALL spme_get_patch(rhos, nbox, delta, q, green%p3m_coeff)

   END SUBROUTINE get_patch_b

! **************************************************************************************************
!> \brief Calculates SPME charge assignment
!> \param rhos ...
!> \param n ...
!> \param delta ...
!> \param q ...
!> \param coeff ...
!> \par History
!>      DG (29-Mar-2001) : code implemented
!> \author JGH (22-Mar-2001)
! **************************************************************************************************
   SUBROUTINE spme_get_patch(rhos, n, delta, q, coeff)

      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(OUT)     :: rhos
      INTEGER, INTENT(IN)                                :: n
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: delta
      REAL(KIND=dp), INTENT(IN)                          :: q
      REAL(KIND=dp), DIMENSION(-(n-1):n-1, 0:n-1), &
         INTENT(IN)                                      :: coeff

      CHARACTER(len=*), PARAMETER :: routineN = 'spme_get_patch', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: nmax = 12

      INTEGER                                            :: i, i1, i2, i3, j, l
      REAL(KIND=dp)                                      :: r2, r3
      REAL(KIND=dp), DIMENSION(3, -nmax:nmax)            :: w_assign
      REAL(KIND=dp), DIMENSION(3, 0:nmax-1)              :: deltal
      REAL(KIND=dp), DIMENSION(3, 1:nmax)                :: f_assign

      IF (n > nmax) THEN
         CALL cp__b("spme.F",867,"nmax value too small")
      END IF
      ! calculate the assignment function values and
      ! the charges on the grid (small box)

      deltal(1, 0) = 1.0_dp
      deltal(2, 0) = 1.0_dp
      deltal(3, 0) = 1.0_dp
      DO l = 1, n-1
         deltal(1, l) = deltal(1, l-1)*delta(1)
         deltal(2, l) = deltal(2, l-1)*delta(2)
         deltal(3, l) = deltal(3, l-1)*delta(3)
      END DO

      w_assign = 0.0_dp
      DO j = -(n-1), n-1, 2
         DO l = 0, n-1
            w_assign(1, j) = w_assign(1, j)+coeff(j, l)*deltal(1, l)
            w_assign(2, j) = w_assign(2, j)+coeff(j, l)*deltal(2, l)
            w_assign(3, j) = w_assign(3, j)+coeff(j, l)*deltal(3, l)
         END DO
      END DO
      DO i = 1, n
         j = n+1-2*i
         f_assign(1, i) = w_assign(1, j)
         f_assign(2, i) = w_assign(2, j)
         f_assign(3, i) = w_assign(3, j)
      END DO

      DO i3 = 1, n
         r3 = q*f_assign(3, i3)
         DO i2 = 1, n
            r2 = r3*f_assign(2, i2)
            DO i1 = 1, n
               rhos(i1, i2, i3) = r2*f_assign(1, i1)
            END DO
         END DO
      END DO

   END SUBROUTINE spme_get_patch

! **************************************************************************************************
!> \brief ...
!> \param box ...
!> \param r ...
!> \param npts ...
!> \param delta ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE get_delta(box, r, npts, delta, n)

      TYPE(cell_type), POINTER                           :: box
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: r
      INTEGER, DIMENSION(3), INTENT(IN)                  :: npts
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: delta
      INTEGER, INTENT(IN)                                :: n

      CHARACTER(len=*), PARAMETER :: routineN = 'get_delta', routineP = moduleN//':'//routineN

      INTEGER                                            :: mp
      INTEGER, DIMENSION(3)                              :: center
      REAL(KIND=dp)                                      :: rmp
      REAL(KIND=dp), DIMENSION(3)                        :: ca, grid_i, s

      mp = MAXVAL(npts(:))
      rmp = REAL(mp, KIND=dp)
      ! compute the scaled coordinate of atomi
      CALL real_to_scaled(s, r, box)
      s = s-REAL(NINT(s), KIND=dp)

      ! find the continuous ``grid'' point
      grid_i(1:3) = REAL(npts(1:3), KIND=dp)*s(1:3)

      ! find the closest grid point

      IF (MOD(n, 2) == 0) THEN
         center(:) = INT(grid_i(:)+rmp)-mp
         ca(:) = REAL(center(:), KIND=dp)+0.5_dp
      ELSE
         center(:) = NINT(grid_i(:))
         ca(:) = REAL(center(:), KIND=dp)
      END IF

      ! find the distance vector
      delta(:) = grid_i(:)-ca(:)

   END SUBROUTINE get_delta

END MODULE spme

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of DFTB3 Terms
!> \author JGH
! **************************************************************************************************
MODULE qs_dftb3_methods
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE atprop_types,                    ONLY: atprop_type
   USE cell_types,                      ONLY: cell_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_add,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_p_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE kinds,                           ONLY: dp
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_type
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE qs_dftb_types,                   ONLY: qs_dftb_atom_type
   USE qs_dftb_utils,                   ONLY: get_dftb_atom_param
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_dftb3_methods'

   PUBLIC :: build_dftb3_diagonal

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param ks_matrix ...
!> \param rho ...
!> \param mcharge ...
!> \param energy ...
!> \param calculate_forces ...
!> \param just_energy ...
! **************************************************************************************************
   SUBROUTINE build_dftb3_diagonal(qs_env, ks_matrix, rho, mcharge, energy, &
                                   calculate_forces, just_energy)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ks_matrix
      TYPE(qs_rho_type), POINTER                         :: rho
      REAL(dp), DIMENSION(:)                             :: mcharge
      TYPE(qs_energy_type), POINTER                      :: energy
      LOGICAL, INTENT(in)                                :: calculate_forces, just_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'build_dftb3_diagonal', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_i, atom_j, handle, i, ia, iatom, &
                                                            ic, icol, ikind, irow, is, jatom, &
                                                            jkind, natom, nimg
      INTEGER, DIMENSION(3)                              :: cellind
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind, kind_of
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: found, use_virial
      REAL(KIND=dp)                                      :: dr, eb3, eloc, fi, gmij, ua, ui, uj, zeff
      REAL(KIND=dp), DIMENSION(3)                        :: fij, rij
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dsblock, ksblock, pblock, sblock
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_p, matrix_s
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: n_list
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_kind
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)
      NULLIFY (atprop)

      ! Energy
      CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, atprop=atprop)

      eb3 = 0.0_dp
      CALL get_qs_env(qs_env=qs_env, local_particles=local_particles)
      DO ikind = 1, SIZE(local_particles%n_el)
         CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_kind)
         CALL get_dftb_atom_param(dftb_kind, dudq=ua, zeff=zeff)
         DO ia = 1, local_particles%n_el(ikind)
            iatom = local_particles%list(ikind)%array(ia)
            eloc = -1.0_dp/6.0_dp*ua*mcharge(iatom)**3
            eb3 = eb3+eloc
            IF (atprop%energy) THEN
               ! we have to add the part not covered by 0.5*Tr(FP)
               eloc = -0.5_dp*eloc-0.25_dp*ua*zeff*mcharge(iatom)**2
               atprop%atecoul(iatom) = atprop%atecoul(iatom)+eloc
            END IF
         END DO
      END DO
      CALL get_qs_env(qs_env=qs_env, para_env=para_env)
      CALL mp_sum(eb3, para_env%group)
      energy%dftb3 = eb3

      ! Forces and Virial
      IF (calculate_forces) THEN
         CALL get_qs_env(qs_env=qs_env, matrix_s_kp=matrix_s, natom=natom, force=force, &
                         cell=cell, virial=virial, particle_set=particle_set)
         ! virial
         use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

         ALLOCATE (atom_of_kind(natom), kind_of(natom))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                                  kind_of=kind_of, atom_of_kind=atom_of_kind)
         CALL qs_rho_get(rho, rho_ao_kp=matrix_p)
         IF (SIZE(matrix_p, 1) == 2) THEN
            DO ic = 1, SIZE(matrix_p, 2)
               CALL cp_dbcsr_add(matrix_p(1, ic)%matrix, matrix_p(2, ic)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF
         !
         nimg = SIZE(matrix_p, 2)
         NULLIFY (cell_to_index)
         IF (nimg > 1) THEN
            NULLIFY (kpoints)
            CALL get_qs_env(qs_env=qs_env, kpoints=kpoints)
            CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
         END IF
         NULLIFY (n_list)
         CALL get_qs_env(qs_env=qs_env, sab_orb=n_list)
         CALL neighbor_list_iterator_create(nl_iterator, n_list)
         DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
            CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                   iatom=iatom, jatom=jatom, r=rij, cell=cellind)

            dr = SQRT(SUM(rij**2))
            IF (iatom == jatom .AND. dr < 1.0e-6_dp) CYCLE

            icol = MAX(iatom, jatom)
            irow = MIN(iatom, jatom)

            IF (nimg == 1) THEN
               ic = 1
            ELSE
               ic = cell_to_index(cellind(1), cellind(2), cellind(3))
               IF(.NOT.(ic > 0))CALL cp__a("qs_dftb3_methods.F",174)
            END IF

            ikind = kind_of(iatom)
            atom_i = atom_of_kind(iatom)
            CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_kind)
            CALL get_dftb_atom_param(dftb_kind, dudq=ui)
            jkind = kind_of(jatom)
            atom_j = atom_of_kind(jatom)
            CALL get_qs_kind(qs_kind_set(jkind), dftb_parameter=dftb_kind)
            CALL get_dftb_atom_param(dftb_kind, dudq=uj)
            !
            gmij = -0.5_dp*(ui*mcharge(iatom)**2+uj*mcharge(jatom)**2)
            !
            NULLIFY (pblock)
            CALL cp_dbcsr_get_block_p(matrix=matrix_p(1, ic)%matrix, &
                                      row=irow, col=icol, block=pblock, found=found)
            IF(.NOT.(found))CALL cp__a("qs_dftb3_methods.F",191)
            DO i = 1, 3
               NULLIFY (dsblock)
               CALL cp_dbcsr_get_block_p(matrix=matrix_s(1+i, ic)%matrix, &
                                         row=irow, col=icol, block=dsblock, found=found)
               IF(.NOT.(found))CALL cp__a("qs_dftb3_methods.F",196)
               IF (irow == iatom) THEN
                  fi = -gmij*SUM(pblock*dsblock)
               ELSE
                  fi = gmij*SUM(pblock*dsblock)
               END IF
               force(ikind)%rho_elec(i, atom_i) = force(ikind)%rho_elec(i, atom_i)+fi
               force(jkind)%rho_elec(i, atom_j) = force(jkind)%rho_elec(i, atom_j)-fi
               fij(i) = fi
            END DO
            IF (use_virial) THEN
               CALL virial_pair_force(virial%pv_virial, 1._dp, fij, rij)
               IF (atprop%stress) THEN
                  CALL virial_pair_force(atprop%atstress(:, :, iatom), 0.5_dp, fij, rij)
                  CALL virial_pair_force(atprop%atstress(:, :, jatom), 0.5_dp, fij, rij)
               END IF
            END IF

         END DO
         CALL neighbor_list_iterator_release(nl_iterator)
         !
         DEALLOCATE (atom_of_kind, kind_of)
         IF (SIZE(matrix_p, 1) == 2) THEN
            DO ic = 1, SIZE(matrix_p, 2)
               CALL cp_dbcsr_add(matrix_p(1, ic)%matrix, matrix_p(2, ic)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=-1.0_dp)
            END DO
         END IF
      END IF

      ! KS matrix
      IF (.NOT. just_energy) THEN
         CALL get_qs_env(qs_env=qs_env, matrix_s_kp=matrix_s, natom=natom)
         ALLOCATE (kind_of(natom))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, kind_of=kind_of)
         !
         nimg = SIZE(ks_matrix, 2)
         NULLIFY (cell_to_index)
         IF (nimg > 1) THEN
            NULLIFY (kpoints)
            CALL get_qs_env(qs_env=qs_env, kpoints=kpoints)
            CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
         END IF

         NULLIFY (n_list)
         CALL get_qs_env(qs_env=qs_env, sab_orb=n_list)
         CALL neighbor_list_iterator_create(nl_iterator, n_list)
         DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
            CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                   iatom=iatom, jatom=jatom, r=rij, cell=cellind)

            icol = MAX(iatom, jatom)
            irow = MIN(iatom, jatom)

            IF (nimg == 1) THEN
               ic = 1
            ELSE
               ic = cell_to_index(cellind(1), cellind(2), cellind(3))
               IF(.NOT.(ic > 0))CALL cp__a("qs_dftb3_methods.F",254)
            END IF

            ikind = kind_of(iatom)
            CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_kind)
            CALL get_dftb_atom_param(dftb_kind, dudq=ui)
            jkind = kind_of(jatom)
            CALL get_qs_kind(qs_kind_set(jkind), dftb_parameter=dftb_kind)
            CALL get_dftb_atom_param(dftb_kind, dudq=uj)
            gmij = -0.5_dp*(ui*mcharge(iatom)**2+uj*mcharge(jatom)**2)
            !
            NULLIFY (sblock)
            CALL cp_dbcsr_get_block_p(matrix=matrix_s(1, ic)%matrix, &
                                      row=irow, col=icol, block=sblock, found=found)
            IF(.NOT.(found))CALL cp__a("qs_dftb3_methods.F",268)
            DO is = 1, SIZE(ks_matrix, 1)
               NULLIFY (ksblock)
               CALL cp_dbcsr_get_block_p(matrix=ks_matrix(is, ic)%matrix, &
                                         row=irow, col=icol, block=ksblock, found=found)
               IF(.NOT.(found))CALL cp__a("qs_dftb3_methods.F",273)
               ksblock = ksblock-0.5_dp*gmij*sblock
            END DO

         END DO
         CALL neighbor_list_iterator_release(nl_iterator)
         !
         DEALLOCATE (kind_of)
      END IF

      CALL timestop(handle)

   END SUBROUTINE build_dftb3_diagonal

END MODULE qs_dftb3_methods

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of Coulomb contributions in DFTB
!> \author JGH
! **************************************************************************************************
MODULE qs_dftb_coulomb
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE atprop_types,                    ONLY: atprop_array_init,&
                                              atprop_type
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_add,&
                                              cp_dbcsr_get_block_p,&
                                              cp_dbcsr_iterator,&
                                              cp_dbcsr_iterator_blocks_left,&
                                              cp_dbcsr_iterator_next_block,&
                                              cp_dbcsr_iterator_start,&
                                              cp_dbcsr_iterator_stop,&
                                              cp_dbcsr_p_type
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_multiply_local
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE dgs,                             ONLY: dg_sum_patch,&
                                              dg_sum_patch_force_1d,&
                                              dg_sum_patch_force_3d
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE erf_fn,                          ONLY: erfc
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_get,&
                                              ewald_pw_type
   USE kinds,                           ONLY: dp
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_type
   USE mathconstants,                   ONLY: fourpi,&
                                              oorootpi,&
                                              pi
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE pme_tools,                       ONLY: get_center,&
                                              set_list
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_grids,                        ONLY: get_pw_grid_info
   USE pw_methods,                      ONLY: pw_copy,&
                                              pw_derive,&
                                              pw_integral_a2b,&
                                              pw_transfer
   USE pw_poisson_methods,              ONLY: pw_poisson_rebuild,&
                                              pw_poisson_solve
   USE pw_poisson_types,                ONLY: do_ewald_ewald,&
                                              do_ewald_none,&
                                              do_ewald_pme,&
                                              do_ewald_spme,&
                                              greens_fn_type,&
                                              pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
   USE qmmm_types_low,                  ONLY: qmmm_env_qm_type
   USE qs_dftb3_methods,                ONLY: build_dftb3_diagonal
   USE qs_dftb_types,                   ONLY: qs_dftb_atom_type
   USE qs_dftb_utils,                   ONLY: get_dftb_atom_param
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE realspace_grid_types,            ONLY: &
        pw2rs, realspace_grid_desc_type, realspace_grid_p_type, realspace_grid_type, rs2pw, &
        rs_grid_create, rs_grid_release, rs_grid_set_box, rs_grid_zero, rs_pw_transfer
   USE spme,                            ONLY: get_patch
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_dftb_coulomb'

   PUBLIC :: build_dftb_coulomb
   PUBLIC :: dftb_spme_evaluate, dftb_ewald_overlap

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param ks_matrix ...
!> \param rho ...
!> \param mcharge ...
!> \param energy ...
!> \param calculate_forces ...
!> \param just_energy ...
! **************************************************************************************************
   SUBROUTINE build_dftb_coulomb(qs_env, ks_matrix, rho, mcharge, energy, &
                                 calculate_forces, just_energy)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ks_matrix
      TYPE(qs_rho_type), POINTER                         :: rho
      REAL(dp), DIMENSION(:)                             :: mcharge
      TYPE(qs_energy_type), POINTER                      :: energy
      LOGICAL, INTENT(in)                                :: calculate_forces, just_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'build_dftb_coulomb', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_i, atom_j, blk, ewald_type, handle, &
                                                            i, ia, iatom, ic, icol, ikind, img, &
                                                            irow, is, jatom, jkind, natom, nimg, &
                                                            nmat
      INTEGER, DIMENSION(3)                              :: cellind, periodic
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind, kind_of
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: do_ewald, found, use_virial
      REAL(KIND=dp)                                      :: alpha, deth, dr, fi, gmij, zeff
      REAL(KIND=dp), DIMENSION(3)                        :: fij, rij
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dsblock, gblock, gmcharge, ksblock, &
                                                            mcin, mcout, pblock, sblock
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: gamma_matrix
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_p, matrix_s
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: n_list
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_dftb_atom_type), POINTER                   :: dftb_kind
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (gamma_matrix, matrix_p, matrix_s, virial, atprop, dft_control)

      use_virial = .FALSE.

      IF (calculate_forces) THEN
         nmat = 4
      ELSE
         nmat = 1
      END IF

      natom = SIZE(mcharge)
      ALLOCATE (gmcharge(natom, nmat))
      gmcharge = 0._dp

      CALL get_qs_env(qs_env, &
                      gamma_matrix=gamma_matrix, &
                      particle_set=particle_set, &
                      cell=cell, &
                      virial=virial, &
                      atprop=atprop, &
                      dft_control=dft_control)

      IF (calculate_forces) THEN
         use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
      END IF

      ALLOCATE (mcin(natom, 1), mcout(natom, 1))
      mcin(:, 1) = mcharge(:)

      DO i = 1, nmat
         mcout = 0._dp
         CALL cp_dbcsr_multiply_local(gamma_matrix(i)%matrix, mcin, mcout, ncol=1)
         gmcharge(:, i) = mcout(:, 1)
      END DO

      DEALLOCATE (mcin, mcout)

      IF (calculate_forces .AND. use_virial) THEN
         CALL cp_dbcsr_iterator_start(iter, gamma_matrix(1)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            NULLIFY (gblock)
            CALL cp_dbcsr_iterator_next_block(iter, iatom, jatom, gblock, blk)
            DO i = 1, 3
               NULLIFY (gblock)
               CALL cp_dbcsr_get_block_p(matrix=gamma_matrix(i+1)%matrix, &
                                         row=iatom, col=jatom, block=gblock, found=found)
               fij(i) = gblock(1, 1)*mcharge(iatom)*mcharge(jatom)
            END DO
            rij = particle_set(iatom)%r-particle_set(jatom)%r
            rij = pbc(rij, cell)
            CALL virial_pair_force(virial%pv_virial, 1._dp, fij, rij)
            IF (atprop%stress) THEN
               CALL virial_pair_force(atprop%atstress(:, :, iatom), 0.5_dp, fij, rij)
               CALL virial_pair_force(atprop%atstress(:, :, jatom), 0.5_dp, fij, rij)
            END IF
         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)
      END IF

      IF (atprop%energy) THEN
         CALL get_qs_env(qs_env=qs_env, particle_set=particle_set)
         natom = SIZE(particle_set)
         CALL atprop_array_init(atprop%atecoul, natom)
      END IF

      ! 1/R contribution
      do_ewald = dft_control%qs_control%dftb_control%do_ewald
      IF (do_ewald) THEN
         ! Ewald sum
         NULLIFY (ewald_env, ewald_pw)
         CALL get_qs_env(qs_env=qs_env, &
                         ewald_env=ewald_env, ewald_pw=ewald_pw)
         CALL get_cell(cell=cell, periodic=periodic, deth=deth)
         CALL ewald_env_get(ewald_env, alpha=alpha, ewald_type=ewald_type)
         CALL get_qs_env(qs_env=qs_env, sab_tbe=n_list)
         CALL dftb_ewald_overlap(gmcharge, mcharge, alpha, n_list, &
                                 virial, use_virial, atprop)
         SELECT CASE (ewald_type)
         CASE DEFAULT
            CALL cp__b("qs_dftb_coulomb.F",246,"Invalid Ewald type")
         CASE (do_ewald_none)
            CALL cp__b("qs_dftb_coulomb.F",248,"Not allowed with DFTB")
         CASE (do_ewald_ewald)
            CALL cp__b("qs_dftb_coulomb.F",250,"Standard Ewald not implemented in DFTB")
         CASE (do_ewald_pme)
            CALL cp__b("qs_dftb_coulomb.F",252,"PME not implemented in DFTB")
         CASE (do_ewald_spme)
            CALL dftb_spme_evaluate(ewald_env, ewald_pw, particle_set, cell, &
                                    gmcharge, mcharge, calculate_forces, virial, use_virial, atprop)
         END SELECT
      ELSE
         ! direct sum
         CALL get_qs_env(qs_env=qs_env, &
                         local_particles=local_particles)
         DO ikind = 1, SIZE(local_particles%n_el)
            DO ia = 1, local_particles%n_el(ikind)
               iatom = local_particles%list(ikind)%array(ia)
               DO jatom = 1, iatom-1
                  rij = particle_set(iatom)%r-particle_set(jatom)%r
                  rij = pbc(rij, cell)
                  dr = SQRT(SUM(rij(:)**2))
                  gmcharge(iatom, 1) = gmcharge(iatom, 1)+mcharge(jatom)/dr
                  gmcharge(jatom, 1) = gmcharge(jatom, 1)+mcharge(iatom)/dr
                  DO i = 2, nmat
                     gmcharge(iatom, i) = gmcharge(iatom, i)+rij(i-1)*mcharge(jatom)/dr**3
                     gmcharge(jatom, i) = gmcharge(jatom, i)-rij(i-1)*mcharge(iatom)/dr**3
                  END DO
               END DO
            END DO
         END DO
         IF(.NOT.(.NOT. use_virial))CALL cp__a("qs_dftb_coulomb.F",277)
      END IF

      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      force=force, para_env=para_env)
      CALL mp_sum(gmcharge(:, 1), para_env%group)

      IF (do_ewald) THEN
         ! add self charge interaction and background charge contribution
         gmcharge(:, 1) = gmcharge(:, 1)-2._dp*alpha*oorootpi*mcharge(:)
         IF (ANY(periodic(:) == 1)) THEN
            gmcharge(:, 1) = gmcharge(:, 1)-pi/alpha**2/deth
         END IF
      END IF

      energy%hartree = energy%hartree+0.5_dp*SUM(mcharge(:)*gmcharge(:, 1))
      IF (atprop%energy) THEN
         CALL get_qs_env(qs_env=qs_env, &
                         local_particles=local_particles)
         DO ikind = 1, SIZE(local_particles%n_el)
            CALL get_qs_kind(qs_kind_set(ikind), dftb_parameter=dftb_kind)
            CALL get_dftb_atom_param(dftb_kind, zeff=zeff)
            DO ia = 1, local_particles%n_el(ikind)
               iatom = local_particles%list(ikind)%array(ia)
               atprop%atecoul(iatom) = atprop%atecoul(iatom)+ &
                                       0.5_dp*zeff*gmcharge(iatom, 1)
            END DO
         END DO
      END IF

      IF (calculate_forces) THEN
         ALLOCATE (atom_of_kind(natom), kind_of(natom))

         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                                  kind_of=kind_of, &
                                  atom_of_kind=atom_of_kind)

         gmcharge(:, 2) = gmcharge(:, 2)*mcharge(:)
         gmcharge(:, 3) = gmcharge(:, 3)*mcharge(:)
         gmcharge(:, 4) = gmcharge(:, 4)*mcharge(:)
         DO iatom = 1, natom
            ikind = kind_of(iatom)
            atom_i = atom_of_kind(iatom)
            force(ikind)%rho_elec(1, atom_i) = force(ikind)%rho_elec(1, atom_i)-gmcharge(iatom, 2)
            force(ikind)%rho_elec(2, atom_i) = force(ikind)%rho_elec(2, atom_i)-gmcharge(iatom, 3)
            force(ikind)%rho_elec(3, atom_i) = force(ikind)%rho_elec(3, atom_i)-gmcharge(iatom, 4)
         END DO
      END IF

      IF (.NOT. just_energy) THEN
         CALL get_qs_env(qs_env=qs_env, matrix_s_kp=matrix_s)
         CALL qs_rho_get(rho, rho_ao_kp=matrix_p)

         nimg = dft_control%nimages
         NULLIFY (cell_to_index)
         IF (nimg > 1) THEN
            NULLIFY (kpoints)
            CALL get_qs_env(qs_env=qs_env, kpoints=kpoints)
            CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
         END IF

         IF (calculate_forces .AND. SIZE(matrix_p, 1) == 2) THEN
            DO img = 1, nimg
               CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
            END DO
         END IF

         NULLIFY (n_list)
         CALL get_qs_env(qs_env=qs_env, sab_orb=n_list)
         CALL neighbor_list_iterator_create(nl_iterator, n_list)
         DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
            CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                   iatom=iatom, jatom=jatom, r=rij, cell=cellind)

            icol = MAX(iatom, jatom)
            irow = MIN(iatom, jatom)

            IF (nimg == 1) THEN
               ic = 1
            ELSE
               ic = cell_to_index(cellind(1), cellind(2), cellind(3))
               IF(.NOT.(ic > 0))CALL cp__a("qs_dftb_coulomb.F",361)
            END IF

            gmij = 0.5_dp*(gmcharge(iatom, 1)+gmcharge(jatom, 1))

            NULLIFY (sblock)
            CALL cp_dbcsr_get_block_p(matrix=matrix_s(1, ic)%matrix, &
                                      row=irow, col=icol, block=sblock, found=found)
            IF(.NOT.(found))CALL cp__a("qs_dftb_coulomb.F",369)
            DO is = 1, SIZE(ks_matrix, 1)
               NULLIFY (ksblock)
               CALL cp_dbcsr_get_block_p(matrix=ks_matrix(is, ic)%matrix, &
                                         row=irow, col=icol, block=ksblock, found=found)
               IF(.NOT.(found))CALL cp__a("qs_dftb_coulomb.F",374)
               ksblock = ksblock-gmij*sblock
            END DO

            IF (calculate_forces) THEN
               ikind = kind_of(iatom)
               atom_i = atom_of_kind(iatom)
               jkind = kind_of(jatom)
               atom_j = atom_of_kind(jatom)
               IF (irow == jatom) gmij = -gmij
               NULLIFY (pblock)
               CALL cp_dbcsr_get_block_p(matrix=matrix_p(1, ic)%matrix, &
                                         row=irow, col=icol, block=pblock, found=found)
               IF(.NOT.(found))CALL cp__a("qs_dftb_coulomb.F",387)
               DO i = 1, 3
                  NULLIFY (dsblock)
                  CALL cp_dbcsr_get_block_p(matrix=matrix_s(1+i, ic)%matrix, &
                                            row=irow, col=icol, block=dsblock, found=found)
                  IF(.NOT.(found))CALL cp__a("qs_dftb_coulomb.F",392)
                  fi = -2.0_dp*gmij*SUM(pblock*dsblock)
                  force(ikind)%rho_elec(i, atom_i) = force(ikind)%rho_elec(i, atom_i)+fi
                  force(jkind)%rho_elec(i, atom_j) = force(jkind)%rho_elec(i, atom_j)-fi
                  fij(i) = fi
               END DO
               IF (use_virial) THEN
                  CALL virial_pair_force(virial%pv_virial, 1._dp, fij, rij)
                  IF (atprop%stress) THEN
                     CALL virial_pair_force(atprop%atstress(:, :, iatom), 0.5_dp, fij, rij)
                     CALL virial_pair_force(atprop%atstress(:, :, jatom), 0.5_dp, fij, rij)
                  END IF
               END IF
            END IF

         END DO
         CALL neighbor_list_iterator_release(nl_iterator)

         IF (calculate_forces .AND. SIZE(matrix_p, 1) == 2) THEN
            DO img = 1, nimg
               CALL cp_dbcsr_add(matrix_p(1, img)%matrix, matrix_p(2, img)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=-1.0_dp)
            END DO
         END IF
      END IF

      IF (dft_control%qs_control%dftb_control%dftb3_diagonal) THEN
         ! Diagonal 3rd order correction (DFTB3)
         CALL build_dftb3_diagonal(qs_env, ks_matrix, rho, mcharge, energy, &
                                   calculate_forces, just_energy)
      END IF

      ! QMMM
      IF (qs_env%qmmm .AND. qs_env%qmmm_periodic) THEN
         CALL build_dftb_coulomb_qmqm(qs_env, ks_matrix, rho, mcharge, energy, &
                                      calculate_forces, just_energy)
      END IF

      IF (calculate_forces) THEN
         DEALLOCATE (atom_of_kind, kind_of)
      END IF
      DEALLOCATE (gmcharge)

      CALL timestop(handle)

   END SUBROUTINE build_dftb_coulomb

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param ks_matrix ...
!> \param rho ...
!> \param mcharge ...
!> \param energy ...
!> \param calculate_forces ...
!> \param just_energy ...
! **************************************************************************************************
   SUBROUTINE build_dftb_coulomb_qmqm(qs_env, ks_matrix, rho, mcharge, energy, &
                                      calculate_forces, just_energy)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ks_matrix
      TYPE(qs_rho_type), POINTER                         :: rho
      REAL(dp), DIMENSION(:)                             :: mcharge
      TYPE(qs_energy_type), POINTER                      :: energy
      LOGICAL, INTENT(in)                                :: calculate_forces, just_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'build_dftb_coulomb_qmqm', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_i, atom_j, blk, ewald_type, handle, &
                                                            i, ia, iatom, ikind, jatom, jkind, &
                                                            natom, nmat
      INTEGER, DIMENSION(3)                              :: periodic
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind, kind_of
      LOGICAL                                            :: found, use_virial
      REAL(KIND=dp)                                      :: alpha, deth, dfr, dr, fi, fr, gmij
      REAL(KIND=dp), DIMENSION(3)                        :: fij, rij
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dsblock, gmcharge, ksblock, ksblock_2, &
                                                            pblock, sblock
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell, mm_cell
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: gamma_matrix, matrix_p, matrix_s
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qmmm_env_qm_type), POINTER                    :: qmmm_env_qm
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (gamma_matrix, matrix_p, matrix_s, virial, atprop, dft_control)

      use_virial = .FALSE.

      IF (calculate_forces) THEN
         nmat = 4
      ELSE
         nmat = 1
      END IF

      natom = SIZE(mcharge)
      ALLOCATE (gmcharge(natom, nmat))
      gmcharge = 0._dp

      CALL get_qs_env(qs_env, &
                      gamma_matrix=gamma_matrix, &
                      particle_set=particle_set, &
                      cell=cell, &
                      virial=virial, &
                      atprop=atprop, &
                      dft_control=dft_control)

      IF (calculate_forces) THEN
         use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
      END IF

      ! Qm-QM long range correction for QMMM calculations
      ! no atomic energy evaluation
      IF(.NOT.(.NOT. atprop%energy))CALL cp__a("qs_dftb_coulomb.F",517)
      ! no stress tensor possible for QMMM
      IF(.NOT.(.NOT. use_virial))CALL cp__a("qs_dftb_coulomb.F",519)
      qmmm_env_qm => qs_env%qmmm_env_qm
      ewald_env => qmmm_env_qm%ewald_env
      ewald_pw => qmmm_env_qm%ewald_pw
      CALL get_qs_env(qs_env=qs_env, super_cell=mm_cell)
      CALL get_cell(cell=mm_cell, periodic=periodic, deth=deth)
      CALL ewald_env_get(ewald_env, alpha=alpha, ewald_type=ewald_type)
      gmcharge = 0.0_dp
      ! direct sum for overlap and local correction
      CALL get_qs_env(qs_env=qs_env, &
                      atomic_kind_set=atomic_kind_set, &
                      local_particles=local_particles, &
                      force=force, para_env=para_env)
      DO ikind = 1, SIZE(local_particles%n_el)
         DO ia = 1, local_particles%n_el(ikind)
            iatom = local_particles%list(ikind)%array(ia)
            DO jatom = 1, iatom-1
               rij = particle_set(iatom)%r-particle_set(jatom)%r
               ! no pbc(rij,mm_cell) at this point, we assume that QM particles are
               ! inside QM box and QM box << MM box
               dr = SQRT(SUM(rij(:)**2))
               ! local (unit cell) correction 1/R
               gmcharge(iatom, 1) = gmcharge(iatom, 1)-mcharge(jatom)/dr
               gmcharge(jatom, 1) = gmcharge(jatom, 1)-mcharge(iatom)/dr
               DO i = 2, nmat
                  gmcharge(iatom, i) = gmcharge(iatom, i)-rij(i-1)*mcharge(jatom)/dr**3
                  gmcharge(jatom, i) = gmcharge(jatom, i)+rij(i-1)*mcharge(iatom)/dr**3
               END DO
               ! overlap correction
               fr = erfc(alpha*dr)/dr
               gmcharge(iatom, 1) = gmcharge(iatom, 1)+mcharge(jatom)*fr
               gmcharge(jatom, 1) = gmcharge(jatom, 1)+mcharge(iatom)*fr
               IF (nmat > 1) THEN
                  dfr = -2._dp*alpha*EXP(-alpha*alpha*dr*dr)*oorootpi/dr-fr/dr
                  dfr = -dfr/dr
                  DO i = 2, nmat
                     gmcharge(iatom, i) = gmcharge(iatom, i)-rij(i-1)*mcharge(jatom)*dfr
                     gmcharge(jatom, i) = gmcharge(jatom, i)+rij(i-1)*mcharge(iatom)*dfr
                  END DO
               END IF
            END DO
         END DO
      END DO

      SELECT CASE (ewald_type)
      CASE DEFAULT
         CALL cp__b("qs_dftb_coulomb.F",565,"Invalid Ewald type")
      CASE (do_ewald_none)
         CALL cp__b("qs_dftb_coulomb.F",567,"Not allowed with DFTB")
      CASE (do_ewald_ewald)
         CALL cp__b("qs_dftb_coulomb.F",569,"Standard Ewald not implemented in DFTB")
      CASE (do_ewald_pme)
         CALL cp__b("qs_dftb_coulomb.F",571,"PME not implemented in DFTB")
      CASE (do_ewald_spme)
         CALL dftb_spme_evaluate(ewald_env, ewald_pw, particle_set, mm_cell, &
                                 gmcharge, mcharge, calculate_forces, virial, use_virial, atprop)
      END SELECT
      !
      CALL mp_sum(gmcharge(:, 1), para_env%group)
      !
      ! add self charge interaction and background charge contribution
      gmcharge(:, 1) = gmcharge(:, 1)-2._dp*alpha*oorootpi*mcharge(:)
      IF (ANY(periodic(:) == 1)) THEN
         gmcharge(:, 1) = gmcharge(:, 1)-pi/alpha**2/deth
      END IF
      !
      energy%qmmm_el = energy%qmmm_el+0.5_dp*SUM(mcharge(:)*gmcharge(:, 1))
      !
      IF (calculate_forces) THEN
         ALLOCATE (atom_of_kind(natom), kind_of(natom))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                                  kind_of=kind_of, &
                                  atom_of_kind=atom_of_kind)
      END IF
      !
      IF (.NOT. just_energy) THEN
         CALL get_qs_env(qs_env=qs_env, matrix_s=matrix_s)
         CALL qs_rho_get(rho, rho_ao=matrix_p)

         IF (calculate_forces .AND. SIZE(matrix_p) == 2) THEN
            CALL cp_dbcsr_add(matrix_p(1)%matrix, matrix_p(2)%matrix, &
                              alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
         END IF

         CALL cp_dbcsr_iterator_start(iter, ks_matrix(1, 1)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, iatom, jatom, ksblock, blk)
            NULLIFY (sblock, ksblock_2)
            IF (SIZE(ks_matrix, 1) > 1) CALL cp_dbcsr_get_block_p(matrix=ks_matrix(2, 1)%matrix, &
                                                                  row=iatom, col=jatom, block=ksblock_2, found=found)
            CALL cp_dbcsr_get_block_p(matrix=matrix_s(1)%matrix, &
                                      row=iatom, col=jatom, block=sblock, found=found)
            gmij = 0.5_dp*(gmcharge(iatom, 1)+gmcharge(jatom, 1))
            ksblock = ksblock-gmij*sblock
            IF (SIZE(ks_matrix, 1) > 1) ksblock_2 = ksblock_2-gmij*sblock
            IF (calculate_forces) THEN
               ikind = kind_of(iatom)
               atom_i = atom_of_kind(iatom)
               jkind = kind_of(jatom)
               atom_j = atom_of_kind(jatom)
               NULLIFY (pblock)
               CALL cp_dbcsr_get_block_p(matrix=matrix_p(1)%matrix, &
                                         row=iatom, col=jatom, block=pblock, found=found)
               DO i = 1, 3
                  NULLIFY (dsblock)
                  CALL cp_dbcsr_get_block_p(matrix=matrix_s(1+i)%matrix, &
                                            row=iatom, col=jatom, block=dsblock, found=found)
                  fi = -2.0_dp*gmij*SUM(pblock*dsblock)
                  force(ikind)%rho_elec(i, atom_i) = force(ikind)%rho_elec(i, atom_i)+fi
                  force(jkind)%rho_elec(i, atom_j) = force(jkind)%rho_elec(i, atom_j)-fi
                  fij(i) = fi
               END DO
            END IF
         END DO
         CALL cp_dbcsr_iterator_stop(iter)
         IF (calculate_forces .AND. SIZE(matrix_p) == 2) THEN
            CALL cp_dbcsr_add(matrix_p(1)%matrix, matrix_p(2)%matrix, &
                              alpha_scalar=1.0_dp, beta_scalar=-1.0_dp)
         END IF
      END IF

      IF (calculate_forces) THEN
         DEALLOCATE (atom_of_kind, kind_of)
      END IF
      DEALLOCATE (gmcharge)

      CALL timestop(handle)

   END SUBROUTINE build_dftb_coulomb_qmqm

! **************************************************************************************************
!> \brief ...
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param particle_set ...
!> \param box ...
!> \param gmcharge ...
!> \param mcharge ...
!> \param calculate_forces ...
!> \param virial ...
!> \param use_virial ...
!> \param atprop ...
! **************************************************************************************************
   SUBROUTINE dftb_spme_evaluate(ewald_env, ewald_pw, particle_set, box, &
                                 gmcharge, mcharge, calculate_forces, virial, use_virial, atprop)

      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: particle_set
      TYPE(cell_type), POINTER                           :: box
      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: gmcharge
      REAL(KIND=dp), DIMENSION(:), INTENT(inout)         :: mcharge
      LOGICAL, INTENT(in)                                :: calculate_forces
      TYPE(virial_type), POINTER                         :: virial
      LOGICAL, INTENT(in)                                :: use_virial
      TYPE(atprop_type), POINTER                         :: atprop

      CHARACTER(len=*), PARAMETER :: routineN = 'dftb_spme_evaluate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: group, handle, i, ig, ipart, j, n, &
                                                            npart, o_spline, p1
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: center
      INTEGER, DIMENSION(3)                              :: nd, npts
      REAL(KIND=dp)                                      :: alpha, dvols, fat(3), ffa, ffb, fint, vgc
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: rhos
      REAL(KIND=dp), DIMENSION(3, 3)                     :: f_stress, h_stress
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(greens_fn_type), POINTER                      :: green
      TYPE(pw_grid_type), POINTER                        :: grid_spme
      TYPE(pw_p_type), DIMENSION(3)                      :: dphi_g
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(pw_type), POINTER                             :: phi_g, phib_g, rhob_g, rhob_r
      TYPE(realspace_grid_desc_type), POINTER            :: rs_desc
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: drpot
      TYPE(realspace_grid_type), POINTER                 :: rden, rpot

      CALL timeset(routineN, handle)
      !-------------- INITIALISATION ---------------------
      CALL ewald_env_get(ewald_env, alpha=alpha, o_spline=o_spline, group=group, &
                         para_env=para_env)
      NULLIFY (green, poisson_env, pw_pool)
      CALL ewald_pw_get(ewald_pw, pw_big_pool=pw_pool, rs_desc=rs_desc, &
                        poisson_env=poisson_env)
      CALL pw_poisson_rebuild(poisson_env)
      green => poisson_env%green_fft
      grid_spme => pw_pool%pw_grid

      CALL get_pw_grid_info(grid_spme, dvol=dvols, npts=npts)

      npart = SIZE(particle_set)

      n = o_spline
      ALLOCATE (rhos(n, n, n))

      CALL rs_grid_create(rden, rs_desc)
      CALL rs_grid_set_box(grid_spme, rs=rden)
      CALL rs_grid_zero(rden)

      ALLOCATE (center(3, npart))
      CALL get_center(particle_set, box, center, npts, n)

      !-------------- DENSITY CALCULATION ----------------
      ipart = 0
      DO
         CALL set_list(particle_set, npart, center, p1, rden, ipart)
         IF (p1 == 0) EXIT

         ! calculate function on small boxes
         CALL get_patch(particle_set, box, green, npts, p1, rhos, is_core=.FALSE., &
                        is_shell=.FALSE., unit_charge=.TRUE.)
         rhos(:, :, :) = rhos(:, :, :)*mcharge(p1)

         ! add boxes to real space grid (big box)
         CALL dg_sum_patch(rden, rhos, center(:, p1))
      END DO

      CALL pw_pool_create_pw(pw_pool, rhob_r, use_data=REALDATA3D, &
                             in_space=REALSPACE)

      CALL rs_pw_transfer(rden, rhob_r, rs2pw)

      ! transform density to G space and add charge function
      CALL pw_pool_create_pw(pw_pool, rhob_g, use_data=COMPLEXDATA1D, &
                             in_space=RECIPROCALSPACE)
      CALL pw_transfer(rhob_r, rhob_g)
      ! update charge function
      rhob_g%cc = rhob_g%cc*green%p3m_charge%cr

      !-------------- ELECTROSTATIC CALCULATION -----------

      ! allocate intermediate arrays
      DO i = 1, 3
         NULLIFY (dphi_g(i)%pw)
         CALL pw_pool_create_pw(pw_pool, dphi_g(i)%pw, &
                                use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      END DO
      CALL pw_pool_create_pw(pw_pool, phi_g, &
                             use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
      IF (use_virial) THEN
         CALL pw_poisson_solve(poisson_env, rhob_g, vgc, phi_g, dphi_g, h_stress=h_stress)
      ELSE
         CALL pw_poisson_solve(poisson_env, rhob_g, vgc, phi_g, dphi_g)
      END IF

      CALL rs_grid_create(rpot, rs_desc)
      CALL rs_grid_set_box(grid_spme, rs=rpot)

      ! Atomic Stress
      IF (atprop%stress .AND. use_virial) THEN

         CALL rs_grid_zero(rpot)
         rhob_g%cc = phi_g%cc*green%p3m_charge%cr
         CALL pw_transfer(rhob_g, rhob_r)
         CALL rs_pw_transfer(rpot, rhob_r, pw2rs)
         ipart = 0
         DO
            CALL set_list(particle_set, npart, center, p1, rden, ipart)
            IF (p1 == 0) EXIT
            ! calculate function on small boxes
            CALL get_patch(particle_set, box, green, npts, p1, rhos, is_core=.FALSE., &
                           is_shell=.FALSE., unit_charge=.TRUE.)
            CALL dg_sum_patch_force_1d(rpot, rhos, center(:, p1), fint)
            atprop%atstress(1, 1, p1) = atprop%atstress(1, 1, p1)-0.5_dp*mcharge(p1)*fint*dvols
            atprop%atstress(2, 2, p1) = atprop%atstress(2, 2, p1)-0.5_dp*mcharge(p1)*fint*dvols
            atprop%atstress(3, 3, p1) = atprop%atstress(3, 3, p1)-0.5_dp*mcharge(p1)*fint*dvols
         END DO

         CALL pw_pool_create_pw(pw_pool, phib_g, &
                                use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
         ffa = (0.5_dp/alpha)**2
         ffb = 1.0_dp/fourpi
         DO i = 1, 3
            DO ig = grid_spme%first_gne0, grid_spme%ngpts_cut_local
               phib_g%cc(ig) = ffb*dphi_g(i)%pw%cc(ig)*(1.0_dp-ffa*grid_spme%gsq(ig))
               phib_g%cc(ig) = phib_g%cc(ig)*green%influence_fn%cc(ig)
            END DO
            IF (grid_spme%have_g0) phib_g%cc(1) = 0.0_dp
            DO j = 1, i
               nd = 0
               nd(j) = 1
               CALL pw_copy(phib_g, rhob_g)
               CALL pw_derive(rhob_g, nd)
               rhob_g%cc = rhob_g%cc*green%p3m_charge%cr
               CALL pw_transfer(rhob_g, rhob_r)
               CALL rs_pw_transfer(rpot, rhob_r, pw2rs)

               ipart = 0
               DO
                  CALL set_list(particle_set, npart, center, p1, rden, ipart)
                  IF (p1 == 0) EXIT
                  ! calculate function on small boxes
                  CALL get_patch(particle_set, box, green, npts, p1, rhos, &
                                 is_core=.FALSE., is_shell=.FALSE., unit_charge=.TRUE.)
                  ! integrate box and potential
                  CALL dg_sum_patch_force_1d(rpot, rhos, center(:, p1), fint)
                  atprop%atstress(i, j, p1) = atprop%atstress(i, j, p1)-fint*dvols*mcharge(p1)
                  IF (i /= j) atprop%atstress(j, i, p1) = atprop%atstress(j, i, p1)-fint*dvols*mcharge(p1)
               END DO

            END DO
         END DO
         CALL pw_pool_give_back_pw(pw_pool, phib_g)

      END IF

      CALL pw_pool_give_back_pw(pw_pool, rhob_g)

      CALL rs_grid_zero(rpot)
      phi_g%cc = phi_g%cc*green%p3m_charge%cr
      CALL pw_transfer(phi_g, rhob_r)
      CALL pw_pool_give_back_pw(pw_pool, phi_g)
      CALL rs_pw_transfer(rpot, rhob_r, pw2rs)

      !---------- END OF ELECTROSTATIC CALCULATION --------

      !------------- STRESS TENSOR CALCULATION ------------

      IF (use_virial) THEN
         DO i = 1, 3
            DO j = i, 3
               f_stress(i, j) = pw_integral_a2b(dphi_g(i)%pw, dphi_g(j)%pw)
               f_stress(j, i) = f_stress(i, j)
            END DO
         END DO
         ffa = (1.0_dp/fourpi)*(0.5_dp/alpha)**2
         virial%pv_virial = virial%pv_virial-(ffa*f_stress+h_stress)/REAL(para_env%num_pe, dp)
      END IF

      !--------END OF STRESS TENSOR CALCULATION -----------

      IF (calculate_forces) THEN
         ! move derivative of potential to real space grid and
         ! multiply by charge function in g-space
         ALLOCATE (drpot(1:3))
         DO i = 1, 3
            CALL rs_grid_create(drpot(i)%rs_grid, rs_desc)
            CALL rs_grid_set_box(grid_spme, rs=drpot(i)%rs_grid)
            dphi_g(i)%pw%cc = dphi_g(i)%pw%cc*green%p3m_charge%cr
            CALL pw_transfer(dphi_g(i)%pw, rhob_r)
            CALL pw_pool_give_back_pw(pw_pool, dphi_g(i)%pw)
            CALL rs_pw_transfer(drpot(i)%rs_grid, rhob_r, pw2rs)
         END DO
      ELSE
         DO i = 1, 3
            CALL pw_pool_give_back_pw(pw_pool, dphi_g(i)%pw)
         END DO
      END IF
      CALL pw_pool_give_back_pw(pw_pool, rhob_r)

      !----------------- FORCE CALCULATION ----------------

      ipart = 0
      DO

         CALL set_list(particle_set, npart, center, p1, rden, ipart)
         IF (p1 == 0) EXIT

         ! calculate function on small boxes
         CALL get_patch(particle_set, box, green, npts, p1, rhos, is_core=.FALSE., &
                        is_shell=.FALSE., unit_charge=.TRUE.)

         CALL dg_sum_patch_force_1d(rpot, rhos, center(:, p1), fint)
         gmcharge(p1, 1) = gmcharge(p1, 1)+fint*dvols

         IF (calculate_forces) THEN
            CALL dg_sum_patch_force_3d(drpot, rhos, center(:, p1), fat)
            gmcharge(p1, 2) = gmcharge(p1, 2)-fat(1)*dvols
            gmcharge(p1, 3) = gmcharge(p1, 3)-fat(2)*dvols
            gmcharge(p1, 4) = gmcharge(p1, 4)-fat(3)*dvols
         END IF

      END DO

      !--------------END OF FORCE CALCULATION -------------

      !------------------CLEANING UP ----------------------

      CALL rs_grid_release(rden)
      CALL rs_grid_release(rpot)
      IF (calculate_forces) THEN
         IF (ASSOCIATED(drpot)) THEN
            DO i = 1, 3
               CALL rs_grid_release(drpot(i)%rs_grid)
            END DO
            DEALLOCATE (drpot)
         END IF
      END IF
      DEALLOCATE (rhos)
      DEALLOCATE (center)

      CALL timestop(handle)

   END SUBROUTINE dftb_spme_evaluate

! **************************************************************************************************
!> \brief ...
!> \param gmcharge ...
!> \param mcharge ...
!> \param alpha ...
!> \param n_list ...
!> \param virial ...
!> \param use_virial ...
!> \param atprop ...
! **************************************************************************************************
   SUBROUTINE dftb_ewald_overlap(gmcharge, mcharge, alpha, n_list, virial, use_virial, atprop)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: gmcharge
      REAL(KIND=dp), DIMENSION(:), INTENT(in)            :: mcharge
      REAL(KIND=dp), INTENT(in)                          :: alpha
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: n_list
      TYPE(virial_type), POINTER                         :: virial
      LOGICAL, INTENT(IN)                                :: use_virial
      TYPE(atprop_type), POINTER                         :: atprop

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dftb_ewald_overlap', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iatom, jatom, nmat
      REAL(KIND=dp)                                      :: dfr, dr, fr, pfr, rij(3)
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator

      CALL timeset(routineN, handle)

      nmat = SIZE(gmcharge, 2)

      CALL neighbor_list_iterator_create(nl_iterator, n_list)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, iatom=iatom, jatom=jatom, r=rij)

         dr = SQRT(SUM(rij(:)**2))
         IF (dr > 1.e-10) THEN
            fr = erfc(alpha*dr)/dr
            gmcharge(iatom, 1) = gmcharge(iatom, 1)+mcharge(jatom)*fr
            gmcharge(jatom, 1) = gmcharge(jatom, 1)+mcharge(iatom)*fr
            IF (nmat > 1) THEN
               dfr = -2._dp*alpha*EXP(-alpha*alpha*dr*dr)*oorootpi/dr-fr/dr
               dfr = -dfr/dr
               DO i = 2, nmat
                  gmcharge(iatom, i) = gmcharge(iatom, i)-rij(i-1)*mcharge(jatom)*dfr
                  gmcharge(jatom, i) = gmcharge(jatom, i)+rij(i-1)*mcharge(iatom)*dfr
               END DO
            END IF
            IF (use_virial) THEN
               pfr = -dfr*mcharge(iatom)*mcharge(jatom)
               CALL virial_pair_force(virial%pv_virial, -pfr, rij, rij)
               IF (atprop%stress) THEN
                  CALL virial_pair_force(atprop%atstress(:, :, iatom), -0.5_dp*pfr, rij, rij)
                  CALL virial_pair_force(atprop%atstress(:, :, jatom), -0.5_dp*pfr, rij, rij)
               END IF
            END IF
         END IF

      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      CALL timestop(handle)

   END SUBROUTINE dftb_ewald_overlap

END MODULE qs_dftb_coulomb

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Provides interfaces to LAPACK routines for factorisation and
!>      linear system solving
!> \note
!>      We are using LAPACK interfaces, so please make sure in IBM/AIX you have
!>      the lapack library before essl: "xlf90 ... -llapack -lessl" !!!
!> \par History
!>      none
!> \author JGH (30-5-2001)
! **************************************************************************************************
MODULE linear_systems

   USE kinds,                           ONLY: dp
   USE lapack,                          ONLY: lapack_sgesv
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'linear_systems'

   PUBLIC :: solve_system

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param mysize ...
!> \param eigenvectors ...
! **************************************************************************************************
   SUBROUTINE solve_system(matrix, mysize, eigenvectors)

      REAL(KIND=dp), INTENT(INOUT)                       :: matrix(:, :)
      INTEGER, INTENT(IN)                                :: mysize
      REAL(KIND=dp), INTENT(INOUT)                       :: eigenvectors(:, :)

      CHARACTER(len=*), PARAMETER :: routineN = 'solve_system', routineP = moduleN//':'//routineN

      INTEGER                                            :: info, lda, ldb, nrhs, ipiv(mysize)

      lda = SIZE(matrix, 1)
      ldb = SIZE(eigenvectors, 1)
      nrhs = SIZE(eigenvectors, 2)

      CALL lapack_sgesv(mysize, nrhs, matrix, lda, ipiv, &
                        eigenvectors, ldb, info)
      IF (info /= 0) THEN
         CALL cp__b("common/linear_systems.F",55,"Error in inversion")
      END IF

   END SUBROUTINE solve_system

END MODULE linear_systems

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief compute mulliken charges
!>      we (currently) define them as c_i = 1/2 [ (PS)_{ii} + (SP)_{ii} ]
!> \author Joost VandeVondele March 2003
! **************************************************************************************************
MODULE mulliken
   USE atomic_charges,                  ONLY: print_atomic_charges
   USE cp_control_types,                ONLY: mulliken_restraint_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_get_block_p, cp_dbcsr_get_info, cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, &
        cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, &
        cp_dbcsr_p_type, cp_dbcsr_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE qs_kind_types,                   ONLY: qs_kind_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mulliken'

! *** Public subroutines ***

   PUBLIC :: mulliken_charges, mulliken_restraint

   INTERFACE mulliken_charges
      MODULE PROCEDURE mulliken_charges_a, mulliken_charges_b, mulliken_charges_c, &
         mulliken_charges_akp, mulliken_charges_bkp, mulliken_charges_ckp
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief computes the energy and density matrix derivate of a constraint on the
!>      mulliken charges
!>
!>      optional outputs:
!>      computes energy (added)
!>      contribution to KS matrix (added)
!>      contribution to W  matrix (added)
!> \param mulliken_restraint_control additional parameters needed to control the restraint
!> \param para_env para_env of the matrices
!> \param s_matrix ,p_matrix : containing the respective quantities
!> \param p_matrix ...
!> \param energy ...
!> \param order_p ...
!> \param ks_matrix ...
!> \param w_matrix ...
!> \par History
!>      06.2004 created [Joost VandeVondele]
!> \note
!>      contribution to the KS matrix is derivative wrt P
!>      contribution to the W matrix is derivate wrt S (sign?)
!>      needed for orbital and ionic forces respectively
! **************************************************************************************************
   SUBROUTINE mulliken_restraint(mulliken_restraint_control, para_env, &
                                 s_matrix, p_matrix, energy, order_p, ks_matrix, w_matrix)
      TYPE(mulliken_restraint_type), INTENT(IN)          :: mulliken_restraint_control
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_dbcsr_type), POINTER                       :: s_matrix
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p_matrix
      REAL(KIND=dp), OPTIONAL                            :: energy, order_p
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: ks_matrix, w_matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'mulliken_restraint', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, iblock_col, iblock_row, ispin, &
                                                            nblock, nspin
      LOGICAL                                            :: found
      REAL(kind=dp)                                      :: mult, my_energy, my_order_p
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: charges, charges_deriv, ks_block, &
                                                            p_block, s_block, w_block
      TYPE(cp_dbcsr_iterator)                            :: iter

! here we get the numbers for charges

      nspin = SIZE(p_matrix)
      CALL cp_dbcsr_get_info(s_matrix, nblkrows_total=nblock)

      ALLOCATE (charges(nblock, nspin))
      ALLOCATE (charges_deriv(nblock, nspin))
      CALL compute_charges(p_matrix, s_matrix, charges, para_env)
      !
      ! this can be used to check the correct implementation of the derivative
      ! CALL rf_deriv_check(mulliken_restraint_control,charges)
      !
      CALL restraint_functional(mulliken_restraint_control, &
                                charges, charges_deriv, my_energy, my_order_p)

      IF (PRESENT(order_p)) THEN
         order_p = my_order_p
      ENDIF
      IF (PRESENT(energy)) THEN
         energy = my_energy
      ENDIF

      IF (PRESENT(ks_matrix)) THEN

         DO ispin = 1, nspin
            CALL cp_dbcsr_iterator_start(iter, s_matrix)
            DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
               CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, s_block, blk)
               CALL cp_dbcsr_get_block_p(matrix=ks_matrix(ispin)%matrix, &
                                         row=iblock_row, col=iblock_col, BLOCK=ks_block, found=found)

               IF (.NOT. (ASSOCIATED(s_block) .AND. ASSOCIATED(ks_block))) THEN
                  CALL cp__b("mulliken.F",118,"Unexpected s / ks structure")
               END IF
               mult = 0.5_dp*charges_deriv(iblock_row, ispin)+ &
                      0.5_dp*charges_deriv(iblock_col, ispin)

               ks_block = ks_block+mult*s_block

            ENDDO
            CALL cp_dbcsr_iterator_stop(iter)
         ENDDO

      ENDIF

      IF (PRESENT(w_matrix)) THEN

         DO ispin = 1, nspin
            CALL cp_dbcsr_iterator_start(iter, p_matrix(ispin)%matrix)
            DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
               CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, p_block, blk)
               CALL cp_dbcsr_get_block_p(matrix=w_matrix(ispin)%matrix, &
                                         row=iblock_row, col=iblock_col, BLOCK=w_block, found=found)

               ! we can cycle if a block is not present
               IF (.NOT. (ASSOCIATED(w_block) .AND. ASSOCIATED(p_block))) CYCLE

               ! minus sign relates to convention for W
               mult = -0.5_dp*charges_deriv(iblock_row, ispin) &
                      -0.5_dp*charges_deriv(iblock_col, ispin)

               w_block = w_block+mult*p_block

            END DO
            CALL cp_dbcsr_iterator_stop(iter)
         ENDDO

      ENDIF

      DEALLOCATE (charges)
      DEALLOCATE (charges_deriv)

   END SUBROUTINE mulliken_restraint

! **************************************************************************************************
!> \brief computes energy and derivatives given a set of charges
!>       this implementation uses the spin density on a number of atoms
!>       as a penalty function
!> \param mulliken_restraint_control ...
!> \param charges (nblock,nspin)
!> \param charges_deriv derivate wrt the corresponding charge entry
!> \param energy ...
!> \param order_p ...
!> \par History
!>      06.2004 created [Joost VandeVondele]
!>      02.2005 added more general form [Joost VandeVondele]
!> \note
!>       should be easy to adapt for other specialized cases
! **************************************************************************************************
   SUBROUTINE restraint_functional(mulliken_restraint_control, charges, &
                                   charges_deriv, energy, order_p)
      TYPE(mulliken_restraint_type), INTENT(IN)          :: mulliken_restraint_control
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: charges, charges_deriv
      REAL(KIND=dp), INTENT(OUT)                         :: energy, order_p

      INTEGER                                            :: I
      REAL(KIND=dp)                                      :: dum

      charges_deriv = 0.0_dp
      order_p = 0.0_dp

      DO I = 1, mulliken_restraint_control%natoms
         order_p = order_p+charges(mulliken_restraint_control%atoms(I), 1) &
                   -charges(mulliken_restraint_control%atoms(I), 2) ! spin density on the relevant atoms
      ENDDO
      ! energy
      energy = mulliken_restraint_control%strength*(order_p-mulliken_restraint_control%target)**2
      ! derivative
      dum = 2*mulliken_restraint_control%strength*(order_p-mulliken_restraint_control%target)
      DO I = 1, mulliken_restraint_control%natoms
         charges_deriv(mulliken_restraint_control%atoms(I), 1) = dum
         charges_deriv(mulliken_restraint_control%atoms(I), 2) = -dum
      ENDDO
   END SUBROUTINE restraint_functional

! **************************************************************************************************
!> \brief compute the mulliken charges
!> \param p_matrix , s_matrix, para_env
!> \param s_matrix ...
!> \param charges previously allocated with the right size (natom,nspin)
!> \param para_env ...
!> \par History
!>      06.2004 created [Joost VandeVondele]
!> \note
!>      charges are computed per spin in the LSD case
! **************************************************************************************************
   SUBROUTINE compute_charges(p_matrix, s_matrix, charges, para_env)
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p_matrix
      TYPE(cp_dbcsr_type), POINTER                       :: s_matrix
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: charges
      TYPE(cp_para_env_type), POINTER                    :: para_env

      INTEGER                                            :: blk, iblock_col, iblock_row, ispin, nspin
      LOGICAL                                            :: found
      REAL(kind=dp)                                      :: mult
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: p_block, s_block
      TYPE(cp_dbcsr_iterator)                            :: iter

      nspin = SIZE(p_matrix)

      charges = 0.0_dp
      DO ispin = 1, nspin
         CALL cp_dbcsr_iterator_start(iter, s_matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            NULLIFY (s_block, p_block)
            CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, s_block, blk)
            CALL cp_dbcsr_get_block_p(matrix=p_matrix(ispin)%matrix, &
                                      row=iblock_row, col=iblock_col, BLOCK=p_block, found=found)

            ! we can cycle if a block is not present
            IF (.NOT. found) CYCLE
            IF (.NOT. (ASSOCIATED(s_block) .AND. ASSOCIATED(p_block))) CYCLE

            IF (iblock_row .EQ. iblock_col) THEN
               mult = 0.5_dp ! avoid double counting of diagonal blocks
            ELSE
               mult = 1.0_dp
            ENDIF
            charges(iblock_row, ispin) = charges(iblock_row, ispin)+ &
                                         mult*SUM(p_block*s_block)
            charges(iblock_col, ispin) = charges(iblock_col, ispin)+ &
                                         mult*SUM(p_block*s_block)

         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)
      ENDDO
      CALL mp_sum(charges, para_env%group)

   END SUBROUTINE compute_charges

! **************************************************************************************************
!> \brief compute the mulliken charge derivatives
!> \param p_matrix , s_matrix, para_env
!> \param s_matrix ...
!> \param charges ...
!> \param dcharges previously allocated with the right size (natom,3)
!> \param para_env ...
!> \par History
!>      01.2012 created [JHU]
! **************************************************************************************************
   SUBROUTINE compute_dcharges(p_matrix, s_matrix, charges, dcharges, para_env)
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p_matrix, s_matrix
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: charges, dcharges
      TYPE(cp_para_env_type), POINTER                    :: para_env

      INTEGER                                            :: blk, iblock_col, iblock_row, ider, &
                                                            ispin, nspin
      LOGICAL                                            :: found
      REAL(kind=dp)                                      :: mult
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ds_block, p_block, s_block
      TYPE(cp_dbcsr_iterator)                            :: iter

      nspin = SIZE(p_matrix)

      charges = 0.0_dp
      dcharges = 0.0_dp
      DO ispin = 1, nspin
         CALL cp_dbcsr_iterator_start(iter, s_matrix(1)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            NULLIFY (s_block, p_block)
            CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, s_block, blk)
            CALL cp_dbcsr_get_block_p(matrix=p_matrix(ispin)%matrix, &
                                      row=iblock_row, col=iblock_col, BLOCK=p_block, found=found)

            ! we can cycle if a block is not present
            IF (.NOT. found) CYCLE
            IF (.NOT. (ASSOCIATED(s_block) .AND. ASSOCIATED(p_block))) CYCLE

            IF (iblock_row .EQ. iblock_col) THEN
               mult = 0.5_dp ! avoid double counting of diagonal blocks
            ELSE
               mult = 1.0_dp
            ENDIF
            charges(iblock_row, ispin) = charges(iblock_row, ispin)+mult*SUM(p_block*s_block)
            charges(iblock_col, ispin) = charges(iblock_col, ispin)+mult*SUM(p_block*s_block)
            DO ider = 1, 3
               CALL cp_dbcsr_get_block_p(matrix=s_matrix(ider+1)%matrix, &
                                         row=iblock_row, col=iblock_col, BLOCK=ds_block, found=found)
               dcharges(iblock_row, ider) = dcharges(iblock_row, ider)+mult*SUM(p_block*ds_block)
               dcharges(iblock_col, ider) = dcharges(iblock_col, ider)+mult*SUM(p_block*ds_block)
            END DO

         ENDDO
         CALL cp_dbcsr_iterator_stop(iter)
      ENDDO
      CALL mp_sum(charges, para_env%group)
      CALL mp_sum(dcharges, para_env%group)

   END SUBROUTINE compute_dcharges

! **************************************************************************************************
!> \brief print the mulliken charges to scr on ionode
!> \param p_matrix , s_matrix, para_env
!> \param s_matrix ...
!> \param para_env ...
!> \param particle_set (needed for Z)
!> \param qs_kind_set ...
!> \param scr unit for output
!> \param title ...
!> \par History
!>      06.2004 adapted to remove explicit matrix multiply [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE mulliken_charges_a(p_matrix, s_matrix, para_env, particle_set, &
                                 qs_kind_set, scr, title)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p_matrix
      TYPE(cp_dbcsr_type), POINTER                       :: s_matrix
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER                                            :: scr
      CHARACTER(LEN=*)                                   :: title

      CHARACTER(len=*), PARAMETER :: routineN = 'mulliken_charges_a', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nblock, nspin
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: charges

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(p_matrix)))CALL cp__a("mulliken.F",347)
      IF(.NOT.(ASSOCIATED(s_matrix)))CALL cp__a("mulliken.F",348)
      ! here we get the numbers for charges
      nspin = SIZE(p_matrix)
      CALL cp_dbcsr_get_info(s_matrix, nblkrows_total=nblock)
      ALLOCATE (charges(nblock, nspin))

      CALL compute_charges(p_matrix, s_matrix, charges, para_env)

      CALL print_atomic_charges(particle_set, qs_kind_set, scr, title, electronic_charges=charges)

      DEALLOCATE (charges)

      CALL timestop(handle)

   END SUBROUTINE mulliken_charges_a

! **************************************************************************************************
!> \brief ...
!> \param p_matrix ...
!> \param s_matrix ...
!> \param para_env ...
!> \param mcharge ...
! **************************************************************************************************
   SUBROUTINE mulliken_charges_b(p_matrix, s_matrix, para_env, mcharge)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p_matrix
      TYPE(cp_dbcsr_type), POINTER                       :: s_matrix
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mcharge

      CHARACTER(len=*), PARAMETER :: routineN = 'mulliken_charges_b', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(p_matrix) .AND. ASSOCIATED(s_matrix)) THEN
         CALL compute_charges(p_matrix, s_matrix, mcharge, para_env)
      END IF

      CALL timestop(handle)

   END SUBROUTINE mulliken_charges_b

! **************************************************************************************************
!> \brief ...
!> \param p_matrix ...
!> \param s_matrix ...
!> \param para_env ...
!> \param mcharge ...
!> \param dmcharge ...
! **************************************************************************************************
   SUBROUTINE mulliken_charges_c(p_matrix, s_matrix, para_env, mcharge, dmcharge)

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p_matrix, s_matrix
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mcharge, dmcharge

      CHARACTER(len=*), PARAMETER :: routineN = 'mulliken_charges_c', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(p_matrix) .AND. ASSOCIATED(s_matrix)) THEN
         CALL compute_dcharges(p_matrix, s_matrix, mcharge, dmcharge, para_env)
      END IF

      CALL timestop(handle)

   END SUBROUTINE mulliken_charges_c

! **************************************************************************************************
!> \brief print the mulliken charges to scr on ionode
!> \param p_matrix_kp ...
!> \param s_matrix_kp ...
!> \param para_env ...
!> \param particle_set (needed for Z)
!> \param qs_kind_set ...
!> \param scr unit for output
!> \param title ...
!> \par History
!>      06.2004 adapted to remove explicit matrix multiply [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE mulliken_charges_akp(p_matrix_kp, s_matrix_kp, para_env, particle_set, &
                                   qs_kind_set, scr, title)

      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: p_matrix_kp, s_matrix_kp
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER                                            :: scr
      CHARACTER(LEN=*)                                   :: title

      CHARACTER(len=*), PARAMETER :: routineN = 'mulliken_charges_akp', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ic, nblock, nspin
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: charges, charges_im
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p_matrix
      TYPE(cp_dbcsr_type), POINTER                       :: s_matrix

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(p_matrix_kp)))CALL cp__a("mulliken.F",454)
      IF(.NOT.(ASSOCIATED(s_matrix_kp)))CALL cp__a("mulliken.F",455)

      nspin = SIZE(p_matrix)
      CALL cp_dbcsr_get_info(s_matrix, nblkrows_total=nblock)
      ALLOCATE (charges(nblock, nspin), charges_im(nblock, nspin))
      charges = 0.0_dp

      DO ic = 1, SIZE(s_matrix_kp, 2)
         NULLIFY (p_matrix, s_matrix)
         p_matrix => p_matrix_kp(:, ic)
         s_matrix => s_matrix_kp(1, ic)%matrix
         charges_im = 0.0_dp
         CALL compute_charges(p_matrix, s_matrix, charges_im, para_env)
         charges(:, :) = charges(:, :)+charges_im(:, :)
      END DO

      CALL print_atomic_charges(particle_set, qs_kind_set, scr, title, electronic_charges=charges)

      DEALLOCATE (charges, charges_im)

      CALL timestop(handle)

   END SUBROUTINE mulliken_charges_akp

! **************************************************************************************************
!> \brief ...
!> \param p_matrix_kp ...
!> \param s_matrix_kp ...
!> \param para_env ...
!> \param mcharge ...
! **************************************************************************************************
   SUBROUTINE mulliken_charges_bkp(p_matrix_kp, s_matrix_kp, para_env, mcharge)

      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: p_matrix_kp, s_matrix_kp
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mcharge

      CHARACTER(len=*), PARAMETER :: routineN = 'mulliken_charges_bkp', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ic, natom, nspin
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mcharge_im
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p_matrix
      TYPE(cp_dbcsr_type), POINTER                       :: s_matrix

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(p_matrix_kp) .AND. ASSOCIATED(s_matrix_kp)) THEN

         mcharge = 0.0_dp
         natom = SIZE(mcharge, 1)
         nspin = SIZE(mcharge, 2)
         ALLOCATE (mcharge_im(natom, nspin))

         DO ic = 1, SIZE(s_matrix_kp, 2)
            NULLIFY (p_matrix, s_matrix)
            p_matrix => p_matrix_kp(:, ic)
            s_matrix => s_matrix_kp(1, ic)%matrix
            IF (ASSOCIATED(p_matrix) .AND. ASSOCIATED(s_matrix)) THEN
               CALL compute_charges(p_matrix, s_matrix, mcharge_im, para_env)
               mcharge(:, :) = mcharge(:, :)+mcharge_im(:, :)
            END IF
         END DO

         DEALLOCATE (mcharge_im)

      END IF

      CALL timestop(handle)

   END SUBROUTINE mulliken_charges_bkp

! **************************************************************************************************
!> \brief ...
!> \param p_matrix_kp ...
!> \param s_matrix_kp ...
!> \param para_env ...
!> \param mcharge ...
!> \param dmcharge ...
! **************************************************************************************************
   SUBROUTINE mulliken_charges_ckp(p_matrix_kp, s_matrix_kp, para_env, &
                                   mcharge, dmcharge)

      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: p_matrix_kp, s_matrix_kp
      TYPE(cp_para_env_type), POINTER                    :: para_env
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: mcharge, dmcharge

      CHARACTER(len=*), PARAMETER :: routineN = 'mulliken_charges_ckp', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ic, natom, nder, nspin
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dmcharge_im, mcharge_im
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: p_matrix, s_matrix

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(p_matrix_kp) .AND. ASSOCIATED(s_matrix_kp)) THEN

         mcharge = 0.0_dp
         dmcharge = 0.0_dp
         natom = SIZE(mcharge, 1)
         nspin = SIZE(mcharge, 2)
         nder = SIZE(dmcharge, 2)
         ALLOCATE (mcharge_im(natom, nspin), dmcharge_im(natom, nder))

         DO ic = 1, SIZE(s_matrix_kp, 2)
            NULLIFY (p_matrix, s_matrix)
            p_matrix => p_matrix_kp(:, ic)
            s_matrix => s_matrix_kp(:, ic)
            IF (ASSOCIATED(p_matrix) .AND. ASSOCIATED(s_matrix)) THEN
               CALL compute_dcharges(p_matrix, s_matrix, mcharge_im, dmcharge_im, para_env)
               mcharge(:, :) = mcharge(:, :)+mcharge_im(:, :)
               dmcharge(:, :) = dmcharge(:, :)+dmcharge_im(:, :)
            END IF
         END DO

         DEALLOCATE (mcharge_im, dmcharge_im)

      END IF

      CALL timestop(handle)

   END SUBROUTINE mulliken_charges_ckp

! **************************************************************************************************

END MODULE mulliken
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Calculation of Hamiltonian matrices in SCPTB
!> \author JGH
! **************************************************************************************************
MODULE scptb_ks_matrix
   USE ai_coulomb,                      ONLY: coulomb2_new
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE atprop_types,                    ONLY: atprop_type
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell,&
                                              pbc
   USE cp_control_types,                ONLY: dft_control_type,&
                                              scptb_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_copy, cp_dbcsr_get_block_p, cp_dbcsr_iterator, &
        cp_dbcsr_iterator_blocks_left, cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, &
        cp_dbcsr_iterator_stop, cp_dbcsr_multiply, cp_dbcsr_p_type, cp_dbcsr_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_type
   USE gamma,                           ONLY: fgamma_0
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE kpoint_types,                    ONLY: get_kpoint_info,&
                                              kpoint_type
   USE linear_systems,                  ONLY: solve_system
   USE mathconstants,                   ONLY: dfac,&
                                              fourpi,&
                                              oorootpi,&
                                              pi
   USE message_passing,                 ONLY: mp_sum
   USE mulliken,                        ONLY: mulliken_charges
   USE orbital_pointers,                ONLY: indso,&
                                              ncoset,&
                                              nsoset
   USE orbital_transformation_matrices, ONLY: c2s_tramat
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_get,&
                                              pw_env_type
   USE pw_methods,                      ONLY: pw_transfer
   USE pw_poisson_methods,              ONLY: pw_poisson_solve
   USE pw_poisson_types,                ONLY: do_ewald_ewald,&
                                              do_ewald_none,&
                                              do_ewald_pme,&
                                              do_ewald_spme,&
                                              pw_poisson_type
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_p_type,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type
   USE qs_collocate_density,            ONLY: calculate_scp_charge
   USE qs_dftb_coulomb,                 ONLY: dftb_ewald_overlap,&
                                              dftb_spme_evaluate
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_integrate_potential,          ONLY: integrate_scp_rspace
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_ks_types,                     ONLY: qs_ks_env_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_get,&
                                              qs_rho_type
   USE scptb_types,                     ONLY: &
        get_scptb_parameter, scp_vector_add, scp_vector_copy, scp_vector_create, scp_vector_dot, &
        scp_vector_mult, scp_vector_p_type, scp_vector_release, scp_vector_set, scp_vector_type, &
        scptb_parameter_type
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'scptb_ks_matrix'

   PUBLIC :: build_scptb_ks_matrix

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param calculate_forces ...
!> \param just_energy ...
! **************************************************************************************************
   SUBROUTINE build_scptb_ks_matrix(qs_env, calculate_forces, just_energy)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(in)                                :: calculate_forces, just_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'build_scptb_ks_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: atom_a, handle, iatom, ic, ikind, ispin, &
                                                            lmaxscp, natom, natom_kind, nkind, &
                                                            nspins, output_unit
      LOGICAL                                            :: do_es, do_scp
      REAL(KIND=dp)                                      :: zeff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: mcharge
      REAL(KIND=dp), DIMENSION(:), POINTER               :: occupation_numbers
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: charges, dmcharge
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: mo_derivs
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ks_matrixkp, matrixkp_h, matrixkp_p, &
                                                            matrixkp_s
      TYPE(cp_dbcsr_type), POINTER                       :: mo_coeff
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mo_array
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(qs_rho_type), POINTER                         :: rho
      TYPE(scptb_control_type), POINTER                  :: scptb_control
      TYPE(scptb_parameter_type), POINTER                :: scptb_kind
      TYPE(section_vals_type), POINTER                   :: scf_section

      CALL timeset(routineN, handle)
      NULLIFY (dft_control, logger, scf_section)
      NULLIFY (particle_set, ks_env, ks_matrixkp, rho, energy)
      IF(.NOT.(ASSOCIATED(qs_env)))CALL cp__a("scptb_ks_matrix.F",151)
      logger => cp_get_default_logger()

      CALL get_qs_env(qs_env, &
                      ks_env=ks_env, &
                      dft_control=dft_control, &
                      atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, &
                      para_env=para_env, &
                      rho=rho, &
                      energy=energy, &
                      matrix_h_kp=matrixkp_h, &
                      matrix_ks_kp=ks_matrixkp)

      IF(.NOT.(SIZE(ks_matrixkp, 1) > 0))CALL cp__a("scptb_ks_matrix.F",165)
      IF(.NOT.(SIZE(ks_matrixkp, 2) > 0))CALL cp__a("scptb_ks_matrix.F",166)

      scf_section => section_vals_get_subs_vals(qs_env%input, "DFT%SCF")
      nspins = dft_control%nspins
      IF(.NOT.(ASSOCIATED(rho)))CALL cp__a("scptb_ks_matrix.F",170)

      do_es = .FALSE.
      do_scp = .FALSE.
      nkind = SIZE(atomic_kind_set)
      DO ikind = 1, nkind
         CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom_kind)
         CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
         CALL get_scptb_parameter(scptb_kind, lmaxscp=lmaxscp)
         IF (lmaxscp > 0) do_scp = .TRUE.
         IF (lmaxscp >= 0) do_es = .TRUE.
      END DO
      scptb_control => dft_control%qs_control%scptb_control
      IF (.NOT. scptb_control%do_scp) do_scp = .FALSE.
      IF (.NOT. scptb_control%do_scc) do_es = .FALSE.
      ! SCP requires self-consistent charge method
      IF (do_scp) do_es = .TRUE.

      ! copy the core matrix into the fock matrix
      IF(.NOT.(ASSOCIATED(matrixkp_h)))CALL cp__a("scptb_ks_matrix.F",189)
      DO ispin = 1, nspins
         DO ic = 1, SIZE(matrixkp_h, 2)
            CALL cp_dbcsr_copy(ks_matrixkp(ispin, ic)%matrix, matrixkp_h(1, ic)%matrix)
         END DO
      END DO

      IF (do_es) THEN
         ! Mulliken charges
         CALL get_qs_env(qs_env, &
                         particle_set=particle_set, &
                         matrix_s_kp=matrixkp_s)
         CALL qs_rho_get(rho, rho_ao_kp=matrixkp_p)
         natom = SIZE(particle_set)
         ALLOCATE (charges(natom, nspins))
         ALLOCATE (mcharge(natom))
         ALLOCATE (dmcharge(natom, 3))
         IF (calculate_forces) THEN
            CALL mulliken_charges(matrixkp_p, matrixkp_s, para_env, charges, dmcharge)
         ELSE
            charges = 0._dp
            CALL mulliken_charges(matrixkp_p, matrixkp_s, para_env, charges)
         END IF
         nkind = SIZE(atomic_kind_set)
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom_kind)
            CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
            CALL get_scptb_parameter(scptb_kind, zeff=zeff)
            DO iatom = 1, natom_kind
               atom_a = atomic_kind_set(ikind)%atom_list(iatom)
               mcharge(atom_a) = zeff-SUM(charges(atom_a, 1:nspins))
            END DO
         END DO
         DEALLOCATE (charges)

         IF (do_scp) THEN
            CALL scp_coulomb(qs_env, ks_matrixkp, matrixkp_s, rho, mcharge, calculate_forces, just_energy)
         ELSE
            CALL tb_coulomb(qs_env, ks_matrixkp, matrixkp_s, rho, mcharge, calculate_forces, just_energy)
         END IF

         DEALLOCATE (mcharge)
         DEALLOCATE (dmcharge)
      ELSE
         energy%hartree = 0._dp
      END IF

      energy%total = energy%core+energy%hartree+energy%qmmm_el+energy%repulsive+ &
                     energy%dispersion

      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%DETAILED_ENERGY", &
                                         extension=".scfLog")
      IF (output_unit > 0) THEN
         WRITE (UNIT=output_unit, FMT="(/,(T9,A,T60,F20.10))") &
            "Repulsive pair potential energy:               ", energy%repulsive, &
            "Zeroth order Hamiltonian energy:               ", energy%core, &
            "Total kinetic energy:                          ", energy%kinetic, &
            "Charge fluctuation energy:                     ", energy%hartree, &
            "London dispersion energy:                      ", energy%dispersion
         IF (qs_env%qmmm) THEN
            WRITE (UNIT=output_unit, FMT="(T3,A,T60,F20.10)") &
               "QM/MM Electrostatic energy:                    ", energy%qmmm_el
         END IF
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                        "PRINT%DETAILED_ENERGY")
      ! here we compute dE/dC if needed. Assumes dE/dC is H_{ks}C (plus occupation numbers)
      IF (qs_env%requires_mo_derivs .AND. .NOT. just_energy) THEN
         CALL get_qs_env(qs_env, mo_derivs=mo_derivs, mos=mo_array)
         DO ispin = 1, SIZE(mo_derivs)
            CALL get_mo_set(mo_set=mo_array(ispin)%mo_set, &
                            mo_coeff_b=mo_coeff, occupation_numbers=occupation_numbers)
            IF (.NOT. mo_array(ispin)%mo_set%use_mo_coeff_b) THEN
               CALL cp__b("scptb_ks_matrix.F",262,"")
            ENDIF
            CALL cp_dbcsr_multiply('n', 'n', 1.0_dp, ks_matrixkp(1, ispin)%matrix, mo_coeff, &
                                   0.0_dp, mo_derivs(ispin)%matrix)
         ENDDO
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE build_scptb_ks_matrix

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param ks_matrix ...
!> \param matrix_s ...
!> \param rho ...
!> \param mcharge ...
!> \param calculate_forces ...
!> \param just_energy ...
! **************************************************************************************************
   SUBROUTINE tb_coulomb(qs_env, ks_matrix, matrix_s, rho, mcharge, calculate_forces, just_energy)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ks_matrix, matrix_s
      TYPE(qs_rho_type), POINTER                         :: rho
      REAL(KIND=dp), DIMENSION(:), INTENT(inout)         :: mcharge
      LOGICAL, INTENT(in)                                :: calculate_forces, just_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'tb_coulomb', routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, ewald_type, i, ia, iat, iatom, ic, &
                                                            icol, ikind, irow, is, ja, jat, jatom, &
                                                            jkind, lmaxscp, nai, naj, natom, ndim, &
                                                            nimg, nkind, nsize
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind, kind_of
      INTEGER, DIMENSION(3)                              :: cellind, periodic
      INTEGER, DIMENSION(:), POINTER                     :: atomi_list, atomj_list
      INTEGER, DIMENSION(:, :, :), POINTER               :: cell_to_index
      LOGICAL                                            :: defined, found, use_virial
      REAL(KIND=dp)                                      :: alpha, beta, chabs, deth, dq, dr, dr2, &
                                                            ehartree, eta, f0, ff, fi, gmij, noa, &
                                                            nob, t
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: qcharge
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: drcharge, gmcharge
      REAL(KIND=dp), DIMENSION(0:1)                      :: f
      REAL(KIND=dp), DIMENSION(3)                        :: fij, rij
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dsblock, ksblock, pblock, sblock
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: matrix_p
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(kpoint_type), POINTER                         :: kpoints
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: n_list
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(scptb_control_type), POINTER                  :: scptb_control
      TYPE(scptb_parameter_type), POINTER                :: scptb_kind
      TYPE(virial_type), POINTER                         :: virial

      NULLIFY (virial, atprop)
      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, virial=virial, atprop=atprop, &
                      energy=energy, para_env=para_env, particle_set=particle_set, cell=cell)
      CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set, &
                      local_particles=local_particles, natom=natom)

      scptb_control => dft_control%qs_control%scptb_control
      use_virial = .FALSE.
      ndim = 1
      IF (calculate_forces) THEN
         CALL get_qs_env(qs_env=qs_env, force=force)
         ndim = 4
         use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer) .AND. scptb_control%do_ewald
      END IF

      nimg = dft_control%nimages
      NULLIFY (cell_to_index)
      IF (nimg > 1) THEN
         NULLIFY (kpoints)
         CALL get_qs_env(qs_env=qs_env, kpoints=kpoints)
         CALL get_kpoint_info(kpoint=kpoints, cell_to_index=cell_to_index)
      END IF

      energy%hartree = 0._dp

      chabs = SUM(ABS(mcharge))
      CALL mp_sum(chabs, para_env%group)
      IF (chabs > 1.e-8_dp) THEN
         IF (scptb_control%do_ewald) THEN
            ALLOCATE (gmcharge(natom, ndim))
            gmcharge = 0._dp
            ! Ewald sum
            NULLIFY (ewald_env, ewald_pw)
            CALL get_qs_env(qs_env=qs_env, ewald_env=ewald_env, ewald_pw=ewald_pw)
            CALL get_cell(cell=cell, periodic=periodic, deth=deth)
            CALL ewald_env_get(ewald_env, alpha=alpha, ewald_type=ewald_type)
            CALL get_qs_env(qs_env=qs_env, sab_tbe=n_list)
            CALL dftb_ewald_overlap(gmcharge, mcharge, alpha, n_list, &
                                    virial, use_virial, atprop)
            SELECT CASE (ewald_type)
            CASE DEFAULT
               CALL cp__b("scptb_ks_matrix.F",374,"Invalid Ewald type")
            CASE (do_ewald_none)
               CALL cp__b("scptb_ks_matrix.F",376,"Not allowed with SCPTB")
            CASE (do_ewald_ewald)
               CALL cp__b("scptb_ks_matrix.F",378,"Standard Ewald not implemented in SCPTB")
            CASE (do_ewald_pme)
               CALL cp__b("scptb_ks_matrix.F",380,"PME not implemented in SCPTB")
            CASE (do_ewald_spme)
               CALL dftb_spme_evaluate(ewald_env, ewald_pw, particle_set, cell, &
                                       gmcharge, mcharge, calculate_forces, virial, use_virial, atprop)
            END SELECT

            ! add self charge interaction and background charge contribution
            CALL mp_sum(gmcharge(:, 1), para_env%group)
            gmcharge(:, 1) = gmcharge(:, 1)-2._dp*alpha*oorootpi*mcharge(:)
            IF (ANY(periodic(:) == 1)) THEN
               gmcharge(:, 1) = gmcharge(:, 1)-pi/alpha**2/deth
            END IF
            energy%hartree = energy%hartree+0.5_dp*SUM(mcharge(:)*gmcharge(:, 1))
            IF (calculate_forces) THEN
               ALLOCATE (atom_of_kind(natom), kind_of(natom))

               CALL get_atomic_kind_set(atomic_kind_set, kind_of=kind_of, atom_of_kind=atom_of_kind)

               gmcharge(:, 2) = gmcharge(:, 2)*mcharge(:)
               gmcharge(:, 3) = gmcharge(:, 3)*mcharge(:)
               gmcharge(:, 4) = gmcharge(:, 4)*mcharge(:)
               DO iatom = 1, natom
                  ikind = kind_of(iatom)
                  iat = atom_of_kind(iatom)
                  force(ikind)%rho_elec(1, iat) = force(ikind)%rho_elec(1, iat)-gmcharge(iatom, 2)
                  force(ikind)%rho_elec(2, iat) = force(ikind)%rho_elec(2, iat)-gmcharge(iatom, 3)
                  force(ikind)%rho_elec(3, iat) = force(ikind)%rho_elec(3, iat)-gmcharge(iatom, 4)
               END DO
            END IF
            IF (.NOT. just_energy) THEN
               CALL qs_rho_get(rho, rho_ao_kp=matrix_p)
               IF (calculate_forces .AND. SIZE(matrix_p, 1) == 2) THEN
                  DO ic = 1, nimg
                     CALL cp_dbcsr_add(matrix_p(1, ic)%matrix, matrix_p(2, ic)%matrix, &
                                       alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
                  END DO
               END IF

               nsize = SIZE(ks_matrix, 1)
               NULLIFY (n_list)
               CALL get_qs_env(qs_env=qs_env, sab_orb=n_list)
               CALL neighbor_list_iterator_create(nl_iterator, n_list)
               DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
                  CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                         iatom=iatom, jatom=jatom, r=rij, cell=cellind)

                  icol = MAX(iatom, jatom)
                  irow = MIN(iatom, jatom)

                  IF (nimg == 1) THEN
                     ic = 1
                  ELSE
                     ic = cell_to_index(cellind(1), cellind(2), cellind(3))
                     IF(.NOT.(ic > 0))CALL cp__a("scptb_ks_matrix.F",433)
                  END IF

                  gmij = 0.5_dp*(gmcharge(iatom, 1)+gmcharge(jatom, 1))

                  CALL cp_dbcsr_get_block_p(matrix=matrix_s(1, ic)%matrix, &
                                            row=irow, col=icol, block=sblock, found=found)
                  IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",440)
                  DO is = 1, nsize
                     NULLIFY (ksblock)
                     CALL cp_dbcsr_get_block_p(matrix=ks_matrix(is, ic)%matrix, &
                                               row=irow, col=icol, block=ksblock, found=found)
                     IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",445)
                     ksblock = ksblock-gmij*sblock
                  END DO
                  IF (calculate_forces .AND. iatom /= jatom) THEN
                     ikind = kind_of(iatom)
                     iat = atom_of_kind(iatom)
                     jkind = kind_of(jatom)
                     jat = atom_of_kind(jatom)
                     NULLIFY (pblock)
                     CALL cp_dbcsr_get_block_p(matrix=matrix_p(1, ic)%matrix, &
                                               row=irow, col=icol, block=pblock, found=found)
                     IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",456)
                     DO i = 1, 3
                        NULLIFY (dsblock)
                        CALL cp_dbcsr_get_block_p(matrix=matrix_s(1+i, ic)%matrix, &
                                                  row=irow, col=icol, block=dsblock, found=found)
                        IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",461)
                        fi = -gmij*SUM(pblock*dsblock)*2.0_dp
                        force(ikind)%rho_elec(i, iat) = force(ikind)%rho_elec(i, iat)+fi
                        force(jkind)%rho_elec(i, jat) = force(jkind)%rho_elec(i, jat)-fi
                        fij(i) = fi
                     END DO
                     IF (use_virial) THEN
                        CALL virial_pair_force(virial%pv_virial, 1._dp, fij, rij)
                        IF (atprop%stress) THEN
                           CALL virial_pair_force(atprop%atstress(:, :, iatom), 0.5_dp, fij, rij)
                           CALL virial_pair_force(atprop%atstress(:, :, jatom), 0.5_dp, fij, rij)
                        END IF
                     END IF
                  END IF

               END DO
               CALL neighbor_list_iterator_release(nl_iterator)

               IF (calculate_forces .AND. SIZE(matrix_p, 1) == 2) THEN
                  DO ic = 1, nimg
                     CALL cp_dbcsr_add(matrix_p(1, ic)%matrix, matrix_p(2, ic)%matrix, &
                                       alpha_scalar=1.0_dp, beta_scalar=-1.0_dp)
                  END DO
               END IF
            END IF

            ! direct sum correction for Gaussians
            ehartree = 0._dp
            IF (.NOT. just_energy .OR. calculate_forces) THEN
               ALLOCATE (qcharge(natom))
               qcharge = 0._dp
            END IF
            IF (calculate_forces) THEN
               ALLOCATE (drcharge(3, natom))
               drcharge = 0._dp
            END IF
            NULLIFY (n_list)
            CALL get_qs_env(qs_env=qs_env, sab_scp=n_list)
            CALL neighbor_list_iterator_create(nl_iterator, n_list)
            DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
               CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, jatom=jatom, r=rij)

               dr2 = SUM(rij(:)**2)
               dr = SQRT(dr2)
               IF (dr > 1.e-10) THEN
                  CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
                  CALL get_scptb_parameter(scptb_kind, lmaxscp=lmaxscp, ag=alpha)
                  IF (lmaxscp < 0) CYCLE
                  noa = (alpha/pi)**1.5_dp
                  CALL get_qs_kind(qs_kind_set(jkind), scptb_parameter=scptb_kind)
                  CALL get_scptb_parameter(scptb_kind, lmaxscp=lmaxscp, ag=beta)
                  IF (lmaxscp < 0) CYCLE
                  eta = alpha*beta/(alpha+beta)
                  f0 = 2.0_dp*SQRT(pi**5/(alpha+beta))/(alpha*beta)
                  nob = (beta/pi)**1.5_dp
                  t = eta*dr2
                  CALL fgamma_0(0, t, f)
                  ff = noa*nob*f(0)*f0
                  ehartree = ehartree+mcharge(iatom)*mcharge(jatom)*ff
                  ehartree = ehartree-mcharge(iatom)*mcharge(jatom)/dr
                  IF (.NOT. just_energy .OR. calculate_forces) THEN
                     qcharge(iatom) = qcharge(iatom)+mcharge(jatom)*ff
                     qcharge(jatom) = qcharge(jatom)+mcharge(iatom)*ff
                     qcharge(iatom) = qcharge(iatom)-mcharge(jatom)/dr
                     qcharge(jatom) = qcharge(jatom)-mcharge(iatom)/dr
                  END IF
                  IF (calculate_forces) THEN
                     ff = noa*nob*f0*(EXP(-t)-f(0))/dr2
                     drcharge(:, iatom) = drcharge(:, iatom)+ff*rij(:)*mcharge(iatom)*mcharge(jatom)
                     drcharge(:, jatom) = drcharge(:, jatom)-ff*rij(:)*mcharge(iatom)*mcharge(jatom)
                     drcharge(:, iatom) = drcharge(:, iatom)+rij(:)/(dr*dr2)*mcharge(iatom)*mcharge(jatom)
                     drcharge(:, jatom) = drcharge(:, jatom)-rij(:)/(dr*dr2)*mcharge(iatom)*mcharge(jatom)
                     IF (use_virial) THEN
                        fij(:) = ff*rij(:)*mcharge(iatom)*mcharge(jatom)+ &
                                 rij(:)/(dr*dr2)*mcharge(iatom)*mcharge(jatom)
                        CALL virial_pair_force(virial%pv_virial, -1._dp, fij, rij)
                     END IF
                  END IF

               END IF

            END DO
            CALL neighbor_list_iterator_release(nl_iterator)
            CALL mp_sum(ehartree, para_env%group)
            energy%hartree = energy%hartree+ehartree

            IF (.NOT. just_energy .OR. calculate_forces) THEN
               IF (calculate_forces) THEN
                  CALL qs_rho_get(rho, rho_ao_kp=matrix_p)
                  IF (SIZE(matrix_p, 1) == 2) THEN
                     DO ic = 1, nimg
                        CALL cp_dbcsr_add(matrix_p(1, ic)%matrix, matrix_p(2, ic)%matrix, &
                                          alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
                     END DO
                  END IF
               END IF
               ! calculate KS matrix
               CALL mp_sum(qcharge, para_env%group)

               NULLIFY (n_list)
               CALL get_qs_env(qs_env=qs_env, sab_orb=n_list)
               CALL neighbor_list_iterator_create(nl_iterator, n_list)
               DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
                  CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                         iatom=iatom, jatom=jatom, r=rij, cell=cellind)

                  icol = MAX(iatom, jatom)
                  irow = MIN(iatom, jatom)

                  IF (nimg == 1) THEN
                     ic = 1
                  ELSE
                     ic = cell_to_index(cellind(1), cellind(2), cellind(3))
                     IF(.NOT.(ic > 0))CALL cp__a("scptb_ks_matrix.F",574)
                  END IF

                  dq = qcharge(iatom)+qcharge(jatom)
                  IF (.NOT. just_energy) THEN
                     CALL cp_dbcsr_get_block_p(matrix=matrix_s(1, ic)%matrix, &
                                               row=irow, col=icol, block=sblock, found=found)
                     IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",581)
                     DO is = 1, nsize
                        NULLIFY (ksblock)
                        CALL cp_dbcsr_get_block_p(matrix=ks_matrix(is, ic)%matrix, &
                                                  row=irow, col=icol, block=ksblock, found=found)
                        IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",586)
                        ksblock = ksblock-0.5_dp*dq*sblock
                     END DO
                  END IF
                  IF (calculate_forces .AND. iatom /= jatom) THEN
                     ikind = kind_of(iatom)
                     iat = atom_of_kind(iatom)
                     jkind = kind_of(jatom)
                     jat = atom_of_kind(jatom)
                     NULLIFY (pblock)
                     CALL cp_dbcsr_get_block_p(matrix=matrix_p(1, ic)%matrix, &
                                               row=irow, col=icol, block=pblock, found=found)
                     IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",598)
                     DO i = 1, 3
                        NULLIFY (dsblock)
                        CALL cp_dbcsr_get_block_p(matrix=matrix_s(1+i, ic)%matrix, &
                                                  row=irow, col=icol, block=dsblock, found=found)
                        IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",603)
                        fij(i) = dq*SUM(pblock*dsblock)
                        force(ikind)%rho_elec(i, iat) = force(ikind)%rho_elec(i, iat)-fij(i)
                        force(jkind)%rho_elec(i, jat) = force(jkind)%rho_elec(i, jat)+fij(i)
                     END DO
                     IF (use_virial) THEN
                        CALL virial_pair_force(virial%pv_virial, -1._dp, fij, rij)
                     END IF
                  END IF

               END DO
               CALL neighbor_list_iterator_release(nl_iterator)

               IF (calculate_forces) THEN
                  DO iatom = 1, natom
                     ikind = kind_of(iatom)
                     iat = atom_of_kind(iatom)
                     force(ikind)%rho_elec(1, iat) = force(ikind)%rho_elec(1, iat)-drcharge(1, iatom)
                     force(ikind)%rho_elec(2, iat) = force(ikind)%rho_elec(2, iat)-drcharge(2, iatom)
                     force(ikind)%rho_elec(3, iat) = force(ikind)%rho_elec(3, iat)-drcharge(3, iatom)
                  END DO
                  IF (SIZE(matrix_p, 1) == 2) THEN
                     DO ic = 1, nimg
                        CALL cp_dbcsr_add(matrix_p(1, ic)%matrix, matrix_p(2, ic)%matrix, &
                                          alpha_scalar=1.0_dp, beta_scalar=-1.0_dp)
                     END DO
                  END IF
               END IF
            END IF

            DEALLOCATE (gmcharge)
            IF (.NOT. just_energy .OR. calculate_forces) THEN
               DEALLOCATE (qcharge)
            END IF
            IF (calculate_forces) THEN
               DEALLOCATE (drcharge)
               DEALLOCATE (atom_of_kind, kind_of)
            END IF
         ELSE
            ! direct sum
            IF (nimg > 1) THEN
               CALL cp__b("scptb_ks_matrix.F",644,"SCPTB direct sum electrostatic cannot be used with k-points.")
            END IF
            IF (.NOT. just_energy .OR. calculate_forces) THEN
               ALLOCATE (qcharge(natom))
               qcharge = 0._dp
            END IF
            IF (calculate_forces) THEN
               ALLOCATE (drcharge(3, natom))
               drcharge = 0._dp
            END IF
            nkind = SIZE(atomic_kind_set)
            DO ikind = 1, nkind
               CALL get_atomic_kind(atomic_kind_set(ikind), natom=nai, atom_list=atomi_list)
               CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
               CALL get_scptb_parameter(scptb_kind, defined=defined, lmaxscp=lmaxscp, ag=alpha)
               IF (.NOT. defined) CYCLE
               IF (lmaxscp < 0) CYCLE
               noa = (alpha/pi)**1.5_dp
               DO jkind = 1, ikind
                  CALL get_atomic_kind(atomic_kind_set(jkind), natom=naj, atom_list=atomj_list)
                  CALL get_qs_kind(qs_kind_set(jkind), scptb_parameter=scptb_kind)
                  CALL get_scptb_parameter(scptb_kind, defined=defined, lmaxscp=lmaxscp, ag=beta)
                  eta = alpha*beta/(alpha+beta)
                  f0 = 2.0_dp*SQRT(pi**5/(alpha+beta))/(alpha*beta)
                  IF (.NOT. defined) CYCLE
                  IF (lmaxscp < 0) CYCLE
                  nob = (beta/pi)**1.5_dp
                  DO ia = 1, nai
                     iatom = atomi_list(ia)
                     DO ja = 1, local_particles%n_el(jkind)
                        jatom = local_particles%list(jkind)%array(ja)
                        IF (ikind == jkind .AND. jatom >= iatom) CYCLE
                        rij = particle_set(iatom)%r-particle_set(jatom)%r
                        rij = pbc(rij, cell)
                        dr2 = SUM(rij(:)**2)
                        t = eta*dr2
                        CALL fgamma_0(0, t, f)
                        ff = noa*nob*f(0)*f0
                        energy%hartree = energy%hartree+mcharge(iatom)*mcharge(jatom)*ff
                        IF (.NOT. just_energy .OR. calculate_forces) THEN
                           qcharge(iatom) = qcharge(iatom)+mcharge(jatom)*ff
                           qcharge(jatom) = qcharge(jatom)+mcharge(iatom)*ff
                        END IF
                        IF (calculate_forces) THEN
                           ff = noa*nob*f0*(EXP(-t)-f(0))/dr2
                           drcharge(:, iatom) = drcharge(:, iatom)+ff*rij(:)*mcharge(iatom)*mcharge(jatom)
                           drcharge(:, jatom) = drcharge(:, jatom)-ff*rij(:)*mcharge(iatom)*mcharge(jatom)
                        END IF
                     END DO
                  END DO
               END DO
            END DO
            CALL mp_sum(energy%hartree, para_env%group)
            IF (.NOT. just_energy .OR. calculate_forces) THEN
               CALL mp_sum(qcharge, para_env%group)
               ALLOCATE (atom_of_kind(natom), kind_of(natom))
               CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                                        kind_of=kind_of, atom_of_kind=atom_of_kind)
               IF (calculate_forces) THEN
                  CALL qs_rho_get(rho, rho_ao_kp=matrix_p)
                  IF (SIZE(matrix_p, 1) == 2) THEN
                     DO ic = 1, nimg
                        CALL cp_dbcsr_add(matrix_p(1, ic)%matrix, matrix_p(2, ic)%matrix, &
                                          alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
                     END DO
                  END IF
               END IF

               ! calculate KS matrix
               nsize = SIZE(ks_matrix)
               CALL cp_dbcsr_iterator_start(iter, matrix_s(1, 1)%matrix)
               DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
                  CALL cp_dbcsr_iterator_next_block(iter, iatom, jatom, sblock, blk)
                  dq = qcharge(iatom)+qcharge(jatom)
                  IF (.NOT. just_energy) THEN
                     DO is = 1, nsize
                        NULLIFY (ksblock)
                        CALL cp_dbcsr_get_block_p(matrix=ks_matrix(is, 1)%matrix, &
                                                  row=iatom, col=jatom, block=ksblock, found=found)
                        IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",723)
                        ksblock = ksblock-0.5_dp*dq*sblock
                     END DO
                  END IF
                  IF (calculate_forces) THEN
                     ikind = kind_of(iatom)
                     iat = atom_of_kind(iatom)
                     jkind = kind_of(jatom)
                     jat = atom_of_kind(jatom)
                     NULLIFY (pblock)
                     CALL cp_dbcsr_get_block_p(matrix=matrix_p(1, 1)%matrix, &
                                               row=iatom, col=jatom, block=pblock, found=found)
                     IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",735)
                     DO i = 1, 3
                        NULLIFY (dsblock)
                        CALL cp_dbcsr_get_block_p(matrix=matrix_s(1+i, 1)%matrix, &
                                                  row=iatom, col=jatom, block=dsblock, found=found)
                        IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",740)
                        fi = dq*SUM(pblock*dsblock)
                        force(ikind)%rho_elec(i, iat) = force(ikind)%rho_elec(i, iat)-fi
                        force(jkind)%rho_elec(i, jat) = force(jkind)%rho_elec(i, jat)+fi
                     END DO
                  END IF
               END DO
               CALL cp_dbcsr_iterator_stop(iter)

               IF (calculate_forces) THEN
                  DO iatom = 1, natom
                     ikind = kind_of(iatom)
                     iat = atom_of_kind(iatom)
                     force(ikind)%rho_elec(1, iat) = force(ikind)%rho_elec(1, iat)+drcharge(1, iatom)
                     force(ikind)%rho_elec(2, iat) = force(ikind)%rho_elec(2, iat)+drcharge(2, iatom)
                     force(ikind)%rho_elec(3, iat) = force(ikind)%rho_elec(3, iat)+drcharge(3, iatom)
                  END DO
                  DEALLOCATE (drcharge)
                  IF (SIZE(matrix_p, 1) == 2) THEN
                     DO ic = 1, nimg
                        CALL cp_dbcsr_add(matrix_p(1, ic)%matrix, matrix_p(2, ic)%matrix, &
                                          alpha_scalar=1.0_dp, beta_scalar=-1.0_dp)
                     END DO
                  END IF
               END IF

               DEALLOCATE (atom_of_kind, kind_of)
               DEALLOCATE (qcharge)
            END IF

         END IF
      END IF

   END SUBROUTINE tb_coulomb

! **************************************************************************************************
!> \brief ...
!> \param qs_env ...
!> \param ks_matrixkp ...
!> \param matrixkp_s ...
!> \param rho ...
!> \param mcharge ...
!> \param calculate_forces ...
!> \param just_energy ...
! **************************************************************************************************
   SUBROUTINE scp_coulomb(qs_env, ks_matrixkp, matrixkp_s, rho, mcharge, &
                          calculate_forces, just_energy)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ks_matrixkp, matrixkp_s
      TYPE(qs_rho_type), POINTER                         :: rho
      REAL(KIND=dp), DIMENSION(:), INTENT(inout)         :: mcharge
      LOGICAL, INTENT(in)                                :: calculate_forces, just_energy

      CHARACTER(len=*), PARAMETER :: routineN = 'scp_coulomb', routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: ndiis = 5
      REAL(KIND=dp), PARAMETER                           :: epsrel = 1.e-8_dp

      INTEGER                                            :: blk, i, iat, iatom, ibas, ikind, is, j, &
                                                            jat, jatom, jkind, kint, kmax, l, &
                                                            lmaxscp, mdiis, mpos, nat, natom, &
                                                            nkind, norb, nsize
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: atom_of_kind, kind_of, lscp, natoms, &
                                                            nbasis
      LOGICAL                                            :: defined, do_kpoints, found
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: adef
      REAL(KIND=dp)                                      :: ak, alpha, dq, dqv, fi, rhom1
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: charges, zeta
      REAL(KIND=dp), DIMENSION(1:3)                      :: pol
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dsblock, ksblock, pblock, sblock
      REAL(KIND=dp), DIMENSION(ndiis+1, 1)               :: vdiis
      REAL(KIND=dp), DIMENSION(ndiis+1, ndiis+1)         :: bdiis
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_matrix, matrix_p, matrix_s
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(qs_energy_type), POINTER                      :: energy
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(scp_vector_p_type), DIMENSION(ndiis)          :: cres, cval
      TYPE(scp_vector_type), POINTER                     :: cp, cpol, cself, cvec, cw
      TYPE(scptb_control_type), POINTER                  :: scptb_control
      TYPE(scptb_parameter_type), POINTER                :: scptb_kind

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, &
                      do_kpoints=do_kpoints, &
                      energy=energy, force=force, para_env=para_env)

      IF (do_kpoints) THEN
         CALL cp__b("scptb_ks_matrix.F",830,"SCPTB not implemented with k-points.")
      END IF

      NULLIFY (ks_matrix, matrix_s)
      ks_matrix => ks_matrixkp(:, 1)
      matrix_s => matrixkp_s(:, 1)

      energy%hartree = 0._dp

      natom = SIZE(mcharge, 1)
      ALLOCATE (charges(natom))
      dqv = 1._dp/SQRT(4._dp*pi)
      charges(:) = mcharge*dqv
      scptb_control => dft_control%qs_control%scptb_control

      NULLIFY (cvec, cpol, cself, cp, cw)

      ! Create scp vectors
      CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set)
      nkind = SIZE(atomic_kind_set)
      ALLOCATE (nbasis(nkind), natoms(nkind), lscp(nkind), zeta(nkind), adef(nkind))
      DO ikind = 1, nkind
         CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
         CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
         CALL get_scptb_parameter(scptb_kind, defined=defined, lmaxscp=lmaxscp, ag=alpha)
         IF (defined) THEN
            norb = (lmaxscp+1)**2
         ELSE
            norb = 0
         END IF
         natoms(ikind) = nat
         nbasis(ikind) = norb
         lscp(ikind) = lmaxscp
         zeta(ikind) = alpha
         adef(ikind) = defined
      END DO
      NULLIFY (cvec, cself, cpol, cp, cw)
      CALL scp_vector_create(cvec, nkind, natoms, nbasis)
      CALL scp_vector_set(cvec, 0._dp)
      CALL scp_vector_create(cself, nkind, natoms, nbasis)
      CALL scp_vector_set(cself, 0._dp)
      CALL scp_vector_create(cpol, nkind, natoms, nbasis)
      CALL scp_vector_set(cpol, 0._dp)
      ! set vector of multipoles
      CALL scp_set_charge(cvec, charges, adef, natoms, atomic_kind_set)
      DO ikind = 1, nkind
         IF (adef(ikind)) THEN
            CALL get_atomic_kind(atomic_kind_set(ikind), natom=nat)
            CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
            CALL get_scptb_parameter(scptb_kind, pol=pol)
            alpha = zeta(ikind)
            DO iat = 1, natoms(ikind)
               DO ibas = 1, nbasis(ikind)
                  l = indso(1, ibas)
                  IF (l > 0) THEN
                     cpol%vector(ikind)%vmat(ibas, iat) = 1._dp/pol(l)
                  END IF
                  cself%vector(ikind)%vmat(ibas, iat) = &
                     4._dp/(2*l+1)/dfac(2*l+1)*SQRT(2._dp*pi)*alpha**(l+0.5_dp)
               END DO
            END DO
         END IF
      END DO

      ! Calclulate scp vector
      DO iat = 1, ndiis
         NULLIFY (cval(iat)%vec, cres(iat)%vec)
         CALL scp_vector_create(cval(iat)%vec, nkind, natoms, nbasis)
         CALL scp_vector_create(cres(iat)%vec, nkind, natoms, nbasis)
      END DO
      CALL scp_vector_create(cp, nkind, natoms, nbasis)
      CALL scp_vector_create(cw, nkind, natoms, nbasis)

      kmax = MIN(SUM(nbasis), 100)*10
      mdiis = 0
      DO kint = 1, kmax
         mdiis = MIN(mdiis+1, ndiis)
         mpos = MOD(kint-1, ndiis)+1
         CALL scp_vector_set(cw, 0._dp)
         CALL apply_scp_vector(cvec, cw, cself, cpol, cp, qs_env, scptb_control%do_ewald, &
                               calculate_forces=.FALSE.)
         rhom1 = scp_dot(cw, cw)
         IF (SQRT(rhom1) < epsrel) EXIT
         CALL scp_vector_copy(cvec, cval(mpos)%vec)
         CALL scp_vector_copy(cw, cres(mpos)%vec)
         IF (mdiis > 100) THEN
            vdiis = 0._dp
            vdiis(mdiis+1, 1) = 1._dp
            bdiis = 0._dp
            bdiis(mdiis+1, 1:mdiis) = 1._dp
            bdiis(1:mdiis, mdiis+1) = 1._dp
            DO i = 1, mdiis
               DO j = 1, i
                  bdiis(i, j) = scp_dot(cres(i)%vec, cres(j)%vec)
                  bdiis(j, i) = bdiis(i, j)
               END DO
            END DO
            CALL solve_system(bdiis, mdiis+1, vdiis)
            CALL scp_vector_set(cvec, 0._dp)
            DO i = 1, mdiis
               CALL scp_vector_add(vdiis(i, 1), cval(i)%vec, cvec)
            END DO
         ELSE
            CALL scp_vector_add(-0.1_dp, cw, cvec)
         END IF
         CALL scp_set_charge(cvec, charges, adef, natoms, atomic_kind_set)
      END DO
      DO iat = 1, ndiis
         CALL scp_vector_release(cval(iat)%vec)
         CALL scp_vector_release(cres(iat)%vec)
      END DO
      ! SCP energy
      CALL scp_vector_set(cw, 0._dp)
      IF (calculate_forces) THEN
         CALL apply_scp_vector(cvec, cw, cself, cpol, cp, qs_env, scptb_control%do_ewald, &
                               calculate_forces=.TRUE.)
      ELSE
         CALL apply_scp_vector(cvec, cw, cself, cpol, cp, qs_env, scptb_control%do_ewald, &
                               calculate_forces=.FALSE.)
      END IF
      CALL scp_vector_dot(ak, cw, cvec)
      energy%hartree = 0.5_dp*ak

      IF (.NOT. just_energy) THEN

         ALLOCATE (atom_of_kind(natom), kind_of(natom))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, kind_of=kind_of, atom_of_kind=atom_of_kind)
         IF (calculate_forces) THEN
            CALL qs_rho_get(rho, rho_ao=matrix_p)
            IF (SIZE(matrix_p) == 2) THEN
               CALL cp_dbcsr_add(matrix_p(1)%matrix, matrix_p(2)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=1.0_dp)
            END IF
         END IF

         ! calculate KS matrix
         nsize = SIZE(ks_matrix)
         CALL cp_dbcsr_iterator_start(iter, matrix_s(1)%matrix)
         DO WHILE (cp_dbcsr_iterator_blocks_left(iter))
            CALL cp_dbcsr_iterator_next_block(iter, iatom, jatom, sblock, blk)
            ikind = kind_of(iatom)
            iat = atom_of_kind(iatom)
            jkind = kind_of(jatom)
            jat = atom_of_kind(jatom)
            dq = 0.5_dp*(cw%vector(ikind)%vmat(1, iat)+cw%vector(jkind)%vmat(1, jat))/SQRT(fourpi)
            DO is = 1, nsize
               NULLIFY (ksblock)
               CALL cp_dbcsr_get_block_p(matrix=ks_matrix(is)%matrix, &
                                         row=iatom, col=jatom, block=ksblock, found=found)
               ksblock = ksblock-dq*sblock
            END DO
            IF (calculate_forces .AND. iatom /= jatom) THEN
               NULLIFY (pblock)
               CALL cp_dbcsr_get_block_p(matrix=matrix_p(1)%matrix, &
                                         row=iatom, col=jatom, block=pblock, found=found)
               IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",985)
               dq = (cw%vector(ikind)%vmat(1, iat)+ &
                     cw%vector(ikind)%vmat(1, iat)*cself%vector(ikind)%vmat(1, iat)+ &
                     cw%vector(jkind)%vmat(1, jat)+ &
                     cw%vector(jkind)%vmat(1, jat)*cself%vector(jkind)%vmat(1, jat))/SQRT(fourpi)
               DO i = 1, 3
                  NULLIFY (dsblock)
                  CALL cp_dbcsr_get_block_p(matrix=matrix_s(1+i)%matrix, &
                                            row=iatom, col=jatom, block=dsblock, found=found)
                  IF(.NOT.(found))CALL cp__a("scptb_ks_matrix.F",994)
                  fi = dq*SUM(pblock*dsblock)
!deb what about these forces?
!               force(ikind)%rho_elec(i,iat) = force(ikind)%rho_elec(i,iat) + fi
!               force(jkind)%rho_elec(i,jat) = force(jkind)%rho_elec(i,jat) - fi
               END DO
            END IF
         END DO
         CALL cp_dbcsr_iterator_stop(iter)

         IF (calculate_forces) THEN
            IF (SIZE(matrix_p) == 2) THEN
               CALL cp_dbcsr_add(matrix_p(1)%matrix, matrix_p(2)%matrix, &
                                 alpha_scalar=1.0_dp, beta_scalar=-1.0_dp)
            END IF
         END IF

      END IF

      CALL scp_vector_release(cself)
      CALL scp_vector_release(cvec)
      CALL scp_vector_release(cpol)
      CALL scp_vector_release(cw)
      CALL scp_vector_release(cp)
      DEALLOCATE (nbasis, natoms, lscp, zeta, adef, charges)

   END SUBROUTINE scp_coulomb

! **************************************************************************************************
!> \brief ...
!> \param vec1 ...
!> \param vec2 ...
!> \retval res ...
! **************************************************************************************************
   FUNCTION scp_dot(vec1, vec2) RESULT(res)
      TYPE(scp_vector_type)                              :: vec1, vec2
      REAL(KIND=dp)                                      :: res

      INTEGER                                            :: i

! dot product without l=0 term

      res = 0._dp
      DO i = 1, SIZE(vec1%vector)
         res = res+SUM(vec1%vector(i)%vmat(2:, :)*vec2%vector(i)%vmat(2:, :))
      END DO

   END FUNCTION scp_dot

! **************************************************************************************************
!> \brief ...
!> \param vec ...
!> \param charge ...
!> \param adef ...
!> \param natoms ...
!> \param atomic_kind_set ...
! **************************************************************************************************
   SUBROUTINE scp_set_charge(vec, charge, adef, natoms, atomic_kind_set)
      TYPE(scp_vector_type)                              :: vec
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: charge
      LOGICAL, DIMENSION(:), INTENT(IN)                  :: adef
      INTEGER, DIMENSION(:), INTENT(IN)                  :: natoms
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set

      INTEGER                                            :: iat, iatom, ikind, nkind
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      nkind = SIZE(adef)
      DO ikind = 1, nkind
         IF (adef(ikind)) THEN
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind=atomic_kind, atom_list=atom_list)
            DO iat = 1, natoms(ikind)
               iatom = atom_list(iat)
               vec%vector(ikind)%vmat(1, iat) = charge(iatom)
            END DO
         END IF
      END DO

   END SUBROUTINE scp_set_charge

! **************************************************************************************************
!> \brief ...
!> \param cin ...
!> \param cout ...
!> \param cself ...
!> \param cpol ...
!> \param cdum ...
!> \param qs_env ...
!> \param do_ewald ...
!> \param calculate_forces ...
! **************************************************************************************************
   SUBROUTINE apply_scp_vector(cin, cout, cself, cpol, cdum, qs_env, do_ewald, calculate_forces)

      TYPE(scp_vector_type), POINTER                     :: cin, cout, cself, cpol, cdum
      TYPE(qs_environment_type), POINTER                 :: qs_env
      LOGICAL, INTENT(in)                                :: do_ewald, calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'apply_scp_vector', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: dr = 0.001_dp

      INTEGER                                            :: i, ia, iatom, ikind, j, ja, jatom, &
                                                            jkind, li, lj, lmaxscp, nai, naj, ni, &
                                                            nj, nkind
      INTEGER, DIMENSION(:), POINTER                     :: atomi_list, atomj_list
      LOGICAL                                            :: defined, use_virial
      REAL(KIND=dp)                                      :: alpha, beta, ehartree
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: aintegral, maint, paint
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: daintegral
      REAL(KIND=dp), DIMENSION(16)                       :: normi, normj
      REAL(KIND=dp), DIMENSION(3)                        :: drij, fij, rij
      REAL(KIND=dp), DIMENSION(3, 3)                     :: h_stress
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cii, cij, coi, coj
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(pw_env_type), POINTER                         :: pw_env
      TYPE(pw_p_type)                                    :: scp_pot, scp_rho, scp_rho_g
      TYPE(pw_poisson_type), POINTER                     :: poisson_env
      TYPE(pw_pool_p_type), DIMENSION(:), POINTER        :: pw_pools
      TYPE(pw_pool_type), POINTER                        :: auxbas_pw_pool
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(scptb_parameter_type), POINTER                :: scptb_kind
      TYPE(virial_type), POINTER                         :: virial

      CALL get_qs_env(qs_env=qs_env, pw_env=pw_env, force=force, virial=virial, para_env=para_env)
      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      IF (do_ewald) THEN
         CALL pw_env_get(pw_env, auxbas_pw_pool=auxbas_pw_pool, &
                         pw_pools=pw_pools, poisson_env=poisson_env)
         CALL pw_pool_create_pw(auxbas_pw_pool, scp_rho%pw, use_data=REALDATA3D, in_space=REALSPACE)
         CALL pw_pool_create_pw(auxbas_pw_pool, scp_pot%pw, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)
         CALL pw_pool_create_pw(auxbas_pw_pool, scp_rho_g%pw, use_data=COMPLEXDATA1D, in_space=RECIPROCALSPACE)

         CALL calculate_scp_charge(scp_rho, qs_env, cin)
         CALL pw_transfer(scp_rho%pw, scp_rho_g%pw)

         ! Getting the Hartree energy and Hartree potential.  Also getting the stress tensor
         ! from the Hartree term if needed.
         IF (use_virial .AND. calculate_forces) THEN
            h_stress(:, :) = 0.0_dp
            CALL pw_poisson_solve(poisson_env, scp_rho_g%pw, ehartree, scp_pot%pw, h_stress=h_stress)
            virial%pv_virial = virial%pv_virial+h_stress/REAL(para_env%num_pe, dp)
         ELSE
            CALL pw_poisson_solve(poisson_env, scp_rho_g%pw, ehartree, scp_pot%pw)
         END IF
         CALL pw_pool_give_back_pw(auxbas_pw_pool, scp_rho_g%pw)

         CALL pw_transfer(scp_pot%pw, scp_rho%pw)
         CALL integrate_scp_rspace(scp_rho, qs_env, cout, calculate_forces)

         CALL pw_pool_give_back_pw(auxbas_pw_pool, scp_rho%pw)
         CALL pw_pool_give_back_pw(auxbas_pw_pool, scp_pot%pw)
      ELSE
         ! direct sum
         NULLIFY (atomic_kind_set, qs_kind_set, particle_set, cell)
         CALL get_qs_env(qs_env=qs_env, atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set, &
                         particle_set=particle_set, cell=cell)
         nkind = SIZE(atomic_kind_set)
         DO ikind = 1, nkind
            CALL get_atomic_kind(atomic_kind_set(ikind), natom=nai, atom_list=atomi_list)
            CALL get_qs_kind(qs_kind_set(ikind), scptb_parameter=scptb_kind)
            CALL get_scptb_parameter(scptb_kind, defined=defined, lmaxscp=lmaxscp, ag=alpha)
            IF (.NOT. defined) CYCLE
            IF (lmaxscp < 0) CYCLE
            li = lmaxscp
            ni = (li+1)**2
            coi => cout%vector(ikind)%vmat
            cii => cin%vector(ikind)%vmat
            CALL calc_norm(normi, li, alpha)
            DO jkind = 1, ikind
               CALL get_atomic_kind(atomic_kind_set(jkind), natom=naj, atom_list=atomj_list)
               CALL get_qs_kind(qs_kind_set(jkind), scptb_parameter=scptb_kind)
               CALL get_scptb_parameter(scptb_kind, defined=defined, lmaxscp=lmaxscp, ag=beta)
               IF (.NOT. defined) CYCLE
               IF (lmaxscp < 0) CYCLE
               lj = lmaxscp
               nj = (lj+1)**2
               coj => cout%vector(jkind)%vmat
               cij => cin%vector(jkind)%vmat
               ALLOCATE (aintegral(ni, nj))
               IF (calculate_forces) THEN
                  ALLOCATE (daintegral(ni, nj, 3), paint(ni, nj), maint(ni, nj))
               END IF
               CALL calc_norm(normj, lj, beta)
               DO iatom = 1, nai
                  ia = atomic_kind_set(ikind)%atom_list(iatom)
                  DO jatom = 1, naj
                     ja = atomic_kind_set(jkind)%atom_list(jatom)
                     IF (ikind == jkind .AND. jatom > iatom) CYCLE
                     rij = particle_set(ia)%r-particle_set(ja)%r
                     rij = pbc(rij, cell)
                     CALL calc_int(aintegral, rij, alpha, li, beta, lj)
                     ! add normalization constants to integrals
                     DO j = 1, nj
                        DO i = 1, ni
                           aintegral(i, j) = aintegral(i, j)*normi(i)*normj(j)
                        END DO
                     END DO
                     ! add contribution to vector
                     coi(1:ni, iatom) = coi(1:ni, iatom)+MATMUL(aintegral(1:ni, 1:nj), cij(1:nj, jatom))
                     IF (ikind == jkind .AND. jatom == iatom) CYCLE
                     coj(1:nj, jatom) = coj(1:nj, jatom)+MATMUL(cii(1:ni, iatom), aintegral(1:ni, 1:nj))
                     IF (calculate_forces) THEN
                        DO i = 1, 3
                           drij = rij
                           drij(i) = drij(i)+dr
                           CALL calc_int(paint, drij, alpha, li, beta, lj)
                           drij(i) = drij(i)-2.0_dp*dr
                           CALL calc_int(maint, drij, alpha, li, beta, lj)
                           daintegral(:, :, i) = (paint-maint)/(2.0_dp*dr)
                        END DO
                        DO j = 1, nj
                           DO i = 1, ni
                              daintegral(i, j, 1) = daintegral(i, j, 1)*normi(i)*normj(j)
                              daintegral(i, j, 2) = daintegral(i, j, 2)*normi(i)*normj(j)
                              daintegral(i, j, 3) = daintegral(i, j, 3)*normi(i)*normj(j)
                           END DO
                        END DO
                        fij(1) = SUM(cii(1:ni, iatom)*MATMUL(daintegral(1:ni, 1:nj, 1), cij(1:nj, jatom)))
                        fij(2) = SUM(cii(1:ni, iatom)*MATMUL(daintegral(1:ni, 1:nj, 2), cij(1:nj, jatom)))
                        fij(3) = SUM(cii(1:ni, iatom)*MATMUL(daintegral(1:ni, 1:nj, 3), cij(1:nj, jatom)))
                        force(ikind)%rho_elec(:, iatom) = force(ikind)%rho_elec(:, iatom)+fij
                        force(jkind)%rho_elec(:, jatom) = force(jkind)%rho_elec(:, jatom)-fij
                     END IF
                  END DO
               END DO
               DEALLOCATE (aintegral)
               IF (calculate_forces) THEN
                  DEALLOCATE (daintegral, paint, maint)
               END IF
            END DO
         END DO

      END IF

      ! remove self interaction
      CALL scp_vector_copy(cself, cdum)
      CALL scp_vector_mult(cin, cdum)
      CALL scp_vector_add(-1._dp, cdum, cout)
      ! add scp polarization
      CALL scp_vector_copy(cpol, cdum)
      CALL scp_vector_mult(cin, cdum)
      CALL scp_vector_add(1._dp, cdum, cout)

   END SUBROUTINE apply_scp_vector

! **************************************************************************************************
!> \brief ...
!> \param aintegral ...
!> \param rij ...
!> \param alpha ...
!> \param li ...
!> \param beta ...
!> \param lj ...
! **************************************************************************************************
   SUBROUTINE calc_int(aintegral, rij, alpha, li, beta, lj)

      REAL(KIND=dp), DIMENSION(:, :), INTENT(inout)      :: aintegral
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rij
      REAL(KIND=dp), INTENT(IN)                          :: alpha
      INTEGER, INTENT(IN)                                :: li
      REAL(KIND=dp), INTENT(IN)                          :: beta
      INTEGER, INTENT(IN)                                :: lj

      CHARACTER(len=*), PARAMETER :: routineN = 'calc_int', routineP = moduleN//':'//routineN

      INTEGER                                            :: mi, mj, ni, nj, nl
      REAL(KIND=dp)                                      :: dr2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: vij
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: v
      REAL(KIND=dp), DIMENSION(0:9)                      :: f
      REAL(KIND=dp), DIMENSION(1)                        :: zi, zj

      dr2 = SUM(rij(:)**2)
      zi(1) = alpha
      zj(1) = beta
      nl = li+lj+1
      ni = ncoset(li)
      nj = ncoset(lj)
      ALLOCATE (vij(ni, nj), v(ni, nj, nl))

      ! cartesian integrals
      CALL coulomb2_new(li, 1, zi, 0, lj, 1, zj, 0, rij, dr2, vij, v, f)
      ! transform to spherical
      mi = nsoset(li)
      mj = nsoset(lj)
      v(1:ni, 1:mj, 1) = MATMUL(vij(1:ni, 1:nj), TRANSPOSE(c2s_tramat(1:mj, 1:nj)))
      aintegral(1:mi, 1:mj) = MATMUL(c2s_tramat(1:mi, 1:ni), v(1:ni, 1:mj, 1))

      DEALLOCATE (vij, v)

   END SUBROUTINE calc_int

! **************************************************************************************************
!> \brief ...
!> \param anorm ...
!> \param lm ...
!> \param alpha ...
! **************************************************************************************************
   SUBROUTINE calc_norm(anorm, lm, alpha)

      REAL(KIND=dp), DIMENSION(:), INTENT(out)           :: anorm
      INTEGER, INTENT(IN)                                :: lm
      REAL(KIND=dp), INTENT(IN)                          :: alpha

      CHARACTER(len=*), PARAMETER :: routineN = 'calc_norm', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, il, l

      anorm = 0.0_dp
      i = 0
      DO l = 0, lm
         DO il = 1, 2*l+1
            i = i+1
            anorm(i) = 2.0_dp**(l+2)*oorootpi/dfac(2*l+1)*alpha**(l+1.5_dp)
         END DO
      END DO

   END SUBROUTINE calc_norm

! **************************************************************************************************

END MODULE scptb_ks_matrix

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Treats the electrostatic for multipoles (up to quadrupoles)
!> \author Teodoro Laino [tlaino] - 12.2007 - University of Zurich
!> inclusion of optional electric field damping for the polarizable atoms
!> Rodolphe Vuilleumier and Mathieu Salanne - 12.2009
! **************************************************************************************************
MODULE ewalds_multipole
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE bibliography,                    ONLY: Aguado2003,&
                                              Laino2008,&
                                              cite_reference
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE damping_dipole_types,            ONLY: damping_type,&
                                              no_damping,&
                                              tang_toennies
   USE dg_rho0_types,                   ONLY: dg_rho0_type
   USE dg_types,                        ONLY: dg_get,&
                                              dg_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE erf_fn,                          ONLY: erf,&
                                              erfc
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_get,&
                                              ewald_pw_type
   USE fist_neighbor_list_types,        ONLY: fist_neighbor_type,&
                                              neighbor_kind_pairs_type
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_get,&
                                              fist_nonbond_env_type,&
                                              pos_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi,&
                                              oorootpi,&
                                              pi,&
                                              sqrthalf
   USE mathlib,                         ONLY: matvec_3x3
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_pool_types,                   ONLY: pw_pool_type
   USE structure_factor_types,          ONLY: structure_factor_type
   USE structure_factors,               ONLY: structure_factor_allocate,&
                                              structure_factor_deallocate,&
                                              structure_factor_evaluate
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2016  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief  Interfaces for MULTIPOLES debug routines
!> \author Teodoro Laino [tlaino] - University of Zurich - 05.2008
!> \date   05.2008
! *****************************************************************************
INTERFACE
  SUBROUTINE debug_ewald_multipoles(ewald_env, ewald_pw, nonbond_env, cell, &
       particle_set, local_particles, iw, debug_r_space)
    USE cell_types,                      ONLY: cell_type
    USE distribution_1d_types,           ONLY: distribution_1d_type
    USE ewald_environment_types,         ONLY: ewald_environment_type
    USE ewald_pw_types,                  ONLY: ewald_pw_type
    USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_type
    USE particle_types,                  ONLY: particle_type
    TYPE(ewald_environment_type), POINTER    :: ewald_env
    TYPE(ewald_pw_type), POINTER             :: ewald_pw
    TYPE(fist_nonbond_env_type), POINTER     :: nonbond_env
    TYPE(cell_type), POINTER                 :: cell
    TYPE(particle_type), DIMENSION(:), &
      POINTER                                :: particle_set
    TYPE(distribution_1d_type), POINTER      :: local_particles
    INTEGER, INTENT(IN)                      :: iw
    LOGICAL, INTENT(IN)                      :: debug_r_space

  END SUBROUTINE debug_ewald_multipoles
END INTERFACE

INTERFACE
  SUBROUTINE debug_ewald_multipoles_fields(ewald_env, ewald_pw, nonbond_env, cell,&
       particle_set, local_particles, radii, charges, dipoles, quadrupoles, task, iw,&
       atomic_kind_set, force_env_section)
    USE atomic_kind_types,               ONLY: atomic_kind_type
    USE cell_types,                      ONLY: cell_type
    USE distribution_1d_types,           ONLY: distribution_1d_type
    USE ewald_environment_types,         ONLY: ewald_environment_type
    USE ewald_pw_types,                  ONLY: ewald_pw_type
    USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_type
    USE input_section_types,             ONLY: section_vals_type
    USE kinds,                           ONLY: dp
    USE particle_types,                  ONLY: particle_type
    TYPE(ewald_environment_type), POINTER    :: ewald_env
    TYPE(ewald_pw_type), POINTER             :: ewald_pw
    TYPE(fist_nonbond_env_type), POINTER     :: nonbond_env
    TYPE(cell_type), POINTER                 :: cell
    TYPE(particle_type), POINTER             :: particle_set(:)
    TYPE(distribution_1d_type), POINTER      :: local_particles
    REAL(KIND=dp), DIMENSION(:), &
         POINTER, OPTIONAL                   :: radii, charges
    REAL(KIND=dp), DIMENSION(:, :), &
         POINTER, OPTIONAL                   :: dipoles
    REAL(KIND=dp), DIMENSION(:, :, :), &
         POINTER, OPTIONAL                   :: quadrupoles
    LOGICAL, DIMENSION(3), INTENT(IN)        :: task
    INTEGER, INTENT(IN)                      :: iw
    TYPE(atomic_kind_type), POINTER          :: atomic_kind_set( : )
    TYPE(section_vals_type), POINTER         :: force_env_section

  END SUBROUTINE debug_ewald_multipoles_fields
END INTERFACE

INTERFACE
  SUBROUTINE debug_ewald_multipoles_fields2(ewald_env, ewald_pw, nonbond_env, cell,&
       particle_set, local_particles, radii, charges, dipoles, quadrupoles, task, iw)
    USE cell_types,                      ONLY: cell_type
    USE distribution_1d_types,           ONLY: distribution_1d_type
    USE ewald_environment_types,         ONLY: ewald_environment_type
    USE ewald_pw_types,                  ONLY: ewald_pw_type
    USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_type
    USE kinds,                           ONLY: dp
    USE particle_types,                  ONLY: particle_type
    TYPE(ewald_environment_type), POINTER    :: ewald_env
    TYPE(ewald_pw_type), POINTER             :: ewald_pw
    TYPE(fist_nonbond_env_type), POINTER     :: nonbond_env
    TYPE(cell_type), POINTER                 :: cell
    TYPE(particle_type), POINTER             :: particle_set(:)
    TYPE(distribution_1d_type), POINTER      :: local_particles
    REAL(KIND=dp), DIMENSION(:), &
         POINTER, OPTIONAL                   :: radii, charges
    REAL(KIND=dp), DIMENSION(:, :), &
         POINTER, OPTIONAL                   :: dipoles
    REAL(KIND=dp), DIMENSION(:, :, :), &
         POINTER, OPTIONAL                   :: quadrupoles
    LOGICAL, DIMENSION(3), INTENT(IN)        :: task
    INTEGER, INTENT(IN)                      :: iw

  END SUBROUTINE debug_ewald_multipoles_fields2
END INTERFACE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   LOGICAL, PRIVATE, PARAMETER :: debug_r_space = .FALSE.
   LOGICAL, PRIVATE, PARAMETER :: debug_g_space = .FALSE.
   LOGICAL, PRIVATE, PARAMETER :: debug_e_field = .FALSE.
   LOGICAL, PRIVATE, PARAMETER :: debug_e_field_en = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ewalds_multipole'

   PUBLIC :: ewald_multipole_evaluate

CONTAINS

! **************************************************************************************************
!> \brief  Computes the potential and the force for a lattice sum of multipoles
!>      (up to quadrupole)
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param nonbond_env ...
!> \param cell ...
!> \param particle_set ...
!> \param local_particles ...
!> \param energy_local ...
!> \param energy_glob ...
!> \param e_neut ...
!> \param e_self ...
!> \param task ...
!> \param do_correction_bonded ...
!> \param do_forces ...
!> \param do_stress ...
!> \param do_efield ...
!> \param radii ...
!> \param charges ...
!> \param dipoles ...
!> \param quadrupoles ...
!> \param forces_local ...
!> \param forces_glob ...
!> \param pv_local ...
!> \param pv_glob ...
!> \param efield0 ...
!> \param efield1 ...
!> \param efield2 ...
!> \param iw ...
!> \param do_debug ...
!> \param atomic_kind_set ...
!> \param mm_section ...
!> \par    Note
!>         atomic_kind_set and mm_section are between the arguments only
!>         for debug purpose (therefore optional) and can be avoided when this
!>         function is called in other part of the program
!> \par    Note
!>         When a gaussian multipole is used instead of point multipole, i.e.
!>         when radii(i)>0, the electrostatic fields (efield0, efield1, efield2)
!>         become derivatives of the electrostatic potential energy towards
!>         these gaussian multipoles.
!> \author Teodoro Laino [tlaino] - 12.2007 - University of Zurich
! **************************************************************************************************
   RECURSIVE SUBROUTINE ewald_multipole_evaluate(ewald_env, ewald_pw, nonbond_env, &
                                                 cell, particle_set, local_particles, energy_local, energy_glob, e_neut, e_self, &
                                                 task, do_correction_bonded, do_forces, do_stress, &
                                                 do_efield, radii, charges, dipoles, &
                                                 quadrupoles, forces_local, forces_glob, pv_local, pv_glob, efield0, efield1, &
                                                 efield2, iw, do_debug, atomic_kind_set, mm_section)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(fist_nonbond_env_type), POINTER               :: nonbond_env
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      REAL(KIND=dp), INTENT(INOUT)                       :: energy_local, energy_glob
      REAL(KIND=dp), INTENT(OUT)                         :: e_neut, e_self
      LOGICAL, DIMENSION(3), INTENT(IN)                  :: task
      LOGICAL, INTENT(IN)                                :: do_correction_bonded, do_forces, &
                                                            do_stress, do_efield
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: radii, charges
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: dipoles
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: quadrupoles
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: forces_local, forces_glob, pv_local, &
                                                            pv_glob
      REAL(KIND=dp), DIMENSION(:), INTENT(OUT), OPTIONAL :: efield0
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT), &
         OPTIONAL                                        :: efield1, efield2
      INTEGER, INTENT(IN)                                :: iw
      LOGICAL, INTENT(IN)                                :: do_debug
      TYPE(atomic_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: atomic_kind_set
      TYPE(section_vals_type), OPTIONAL, POINTER         :: mm_section

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_multipole_evaluate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: group, handle, i, j, size1, size2
      LOGICAL                                            :: check_debug, check_efield, check_forces, &
                                                            do_task(3)
      LOGICAL, DIMENSION(3, 3)                           :: my_task
      REAL(KIND=dp)                                      :: e_bonded, e_bonded_t, e_rspace, &
                                                            e_rspace_t, energy_glob_t
      REAL(KIND=dp), DIMENSION(:), POINTER               :: efield0_lr, efield0_sr
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: efield1_lr, efield1_sr, efield2_lr, &
                                                            efield2_sr

      CALL cite_reference(Aguado2003)
      CALL cite_reference(Laino2008)
      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(nonbond_env)))CALL cp__a("ewalds_multipole.F",162)
      check_debug = (debug_this_module .OR. debug_r_space .OR. debug_g_space .OR. debug_e_field .OR. debug_e_field_en) &
                    .EQV. debug_this_module
      IF(.NOT.(check_debug))CALL cp__a("ewalds_multipole.F",165)
      check_forces = do_forces .EQV. (PRESENT(forces_local) .AND. PRESENT(forces_glob))
      IF(.NOT.(check_forces))CALL cp__a("ewalds_multipole.F",167)
      check_efield = do_efield .EQV. (PRESENT(efield0) .OR. PRESENT(efield1) .OR. PRESENT(efield2))
      IF(.NOT.(check_efield))CALL cp__a("ewalds_multipole.F",169)
      ! Debugging this module
      IF (debug_this_module .AND. do_debug) THEN
         ! Debug specifically real space part
         IF (debug_r_space) THEN
            CALL debug_ewald_multipoles(ewald_env, ewald_pw, nonbond_env, cell, &
                                        particle_set, local_particles, iw, debug_r_space)
            CALL cp__b("ewalds_multipole.F",176,"Debug Multipole Requested:  Real Part!")
         END IF
         ! Debug electric fields and gradients as pure derivatives
         IF (debug_e_field) THEN
            IF(.NOT.(PRESENT(atomic_kind_set)))CALL cp__a("ewalds_multipole.F",180)
            IF(.NOT.(PRESENT(mm_section)))CALL cp__a("ewalds_multipole.F",181)
            CALL debug_ewald_multipoles_fields(ewald_env, ewald_pw, nonbond_env, &
                                               cell, particle_set, local_particles, radii, charges, dipoles, &
                                               quadrupoles, task, iw, atomic_kind_set, mm_section)
            CALL cp__b("ewalds_multipole.F",185,"Debug Multipole Requested:  POT+EFIELDS+GRAD!")
         END IF
         ! Debug the potential, electric fields and electric fields gradient in oder
         ! to retrieve the correct energy
         IF (debug_e_field_en) THEN
            CALL debug_ewald_multipoles_fields2(ewald_env, ewald_pw, nonbond_env, &
                                                cell, particle_set, local_particles, radii, charges, dipoles, &
                                                quadrupoles, task, iw)
            CALL cp__b("ewalds_multipole.F",193,"Debug Multipole Requested:  POT+EFIELDS+GRAD to give the correct energy!!")
         END IF
      END IF

      ! Setup the tasks (needed to skip useless parts in the real-space term)
      do_task = task
      DO i = 1, 3
         IF (do_task(i)) THEN
            SELECT CASE (i)
            CASE (1)
               do_task(1) = ANY(charges /= 0.0_dp)
            CASE (2)
               do_task(2) = ANY(dipoles /= 0.0_dp)
            CASE (3)
               do_task(3) = ANY(quadrupoles /= 0.0_dp)
            END SELECT
         END IF
      END DO
      DO i = 1, 3
         DO j = i, 3
            my_task(j, i) = do_task(i) .AND. do_task(j)
            my_task(i, j) = my_task(j, i)
         END DO
      END DO

      ! Allocate arrays for the evaluation of the potential, fields and electrostatic field gradients
      NULLIFY (efield0_sr, efield0_lr, efield1_sr, efield1_lr, efield2_sr, efield2_lr)
      IF (do_efield) THEN
         IF (PRESENT(efield0)) THEN
            size1 = SIZE(efield0)
            ALLOCATE (efield0_sr(size1))
            ALLOCATE (efield0_lr(size1))
            efield0_sr = 0.0_dp
            efield0_lr = 0.0_dp
         END IF
         IF (PRESENT(efield1)) THEN
            size1 = SIZE(efield1, 1)
            size2 = SIZE(efield1, 2)
            ALLOCATE (efield1_sr(size1, size2))
            ALLOCATE (efield1_lr(size1, size2))
            efield1_sr = 0.0_dp
            efield1_lr = 0.0_dp
         END IF
         IF (PRESENT(efield2)) THEN
            size1 = SIZE(efield2, 1)
            size2 = SIZE(efield2, 2)
            ALLOCATE (efield2_sr(size1, size2))
            ALLOCATE (efield2_lr(size1, size2))
            efield2_sr = 0.0_dp
            efield2_lr = 0.0_dp
         END IF
      END IF

      e_rspace = 0.0_dp
      e_bonded = 0.0_dp
      IF ((.NOT. debug_g_space) .AND. (nonbond_env%do_nonbonded)) THEN
         ! Compute the Real Space (Short-Range) part of the Ewald sum.
         ! This contribution is only added when the nonbonded flag in the input
         ! is set, because these contributions depend. the neighborlists.
         CALL ewald_multipole_SR(nonbond_env, ewald_env, atomic_kind_set, &
                                 particle_set, cell, e_rspace, my_task, &
                                 do_forces, do_efield, do_stress, radii, charges, dipoles, quadrupoles, &
                                 forces_glob, pv_glob, efield0_sr, efield1_sr, efield2_sr)
         energy_glob = energy_glob+e_rspace

         IF (do_correction_bonded) THEN
            ! The corrections for bonded interactions are stored in the Real Space
            ! (Short-Range) part of the fields array.
            CALL ewald_multipole_bonded(nonbond_env, particle_set, ewald_env, &
                                        cell, e_bonded, my_task, do_forces, do_efield, do_stress, &
                                        charges, dipoles, quadrupoles, forces_glob, pv_glob, &
                                        efield0_sr, efield1_sr, efield2_sr)
            energy_glob = energy_glob+e_bonded
         END IF
      END IF

      e_neut = 0.0_dp
      e_self = 0.0_dp
      energy_local = 0.0_dp
      IF (.NOT. debug_r_space) THEN
         ! Compute the Reciprocal Space (Long-Range) part of the Ewald sum
         CALL ewald_multipole_LR(ewald_env, ewald_pw, cell, particle_set, &
                                 local_particles, energy_local, my_task, do_forces, do_efield, do_stress, &
                                 charges, dipoles, quadrupoles, forces_local, pv_local, efield0_lr, efield1_lr, &
                                 efield2_lr)

         ! Self-Interactions corrections
         CALL ewald_multipole_self(ewald_env, cell, local_particles, e_self, &
                                   e_neut, my_task, do_efield, radii, charges, dipoles, quadrupoles, &
                                   efield0_lr, efield1_lr, efield2_lr)
      END IF

      ! Sumup energy contributions for possible IO
      CALL ewald_env_get(ewald_env, group=group)
      energy_glob_t = energy_glob
      e_rspace_t = e_rspace
      e_bonded_t = e_bonded
      CALL mp_sum(energy_glob_t, group)
      CALL mp_sum(e_rspace_t, group)
      CALL mp_sum(e_bonded_t, group)
      ! Print some info about energetics
      CALL ewald_multipole_print(iw, energy_local, e_rspace_t, e_bonded_t, e_self, e_neut)

      ! Gather the components of the potential, fields and electrostatic field gradients
      IF (do_efield) THEN
         IF (PRESENT(efield0)) THEN
            efield0 = efield0_sr+efield0_lr
            CALL mp_sum(efield0, group)
            DEALLOCATE (efield0_sr)
            DEALLOCATE (efield0_lr)
         END IF
         IF (PRESENT(efield1)) THEN
            efield1 = efield1_sr+efield1_lr
            CALL mp_sum(efield1, group)
            DEALLOCATE (efield1_sr)
            DEALLOCATE (efield1_lr)
         END IF
         IF (PRESENT(efield2)) THEN
            efield2 = efield2_sr+efield2_lr
            CALL mp_sum(efield2, group)
            DEALLOCATE (efield2_sr)
            DEALLOCATE (efield2_lr)
         END IF
      END IF
      CALL timestop(handle)
   END SUBROUTINE ewald_multipole_evaluate

! **************************************************************************************************
!> \brief computes the potential and the force for a lattice sum of multipoles
!>      up to quadrupole - Short Range (Real Space) Term
!> \param nonbond_env ...
!> \param ewald_env ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param cell ...
!> \param energy ...
!> \param task ...
!> \param do_forces ...
!> \param do_efield ...
!> \param do_stress ...
!> \param radii ...
!> \param charges ...
!> \param dipoles ...
!> \param quadrupoles ...
!> \param forces ...
!> \param pv ...
!> \param efield0 ...
!> \param efield1 ...
!> \param efield2 ...
!> \author Teodoro Laino [tlaino] - 12.2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE ewald_multipole_SR(nonbond_env, ewald_env, atomic_kind_set, &
                                 particle_set, cell, energy, task, &
                                 do_forces, do_efield, do_stress, radii, charges, dipoles, quadrupoles, &
                                 forces, pv, efield0, efield1, efield2)
      TYPE(fist_nonbond_env_type), POINTER               :: nonbond_env
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(atomic_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: atomic_kind_set
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      LOGICAL, DIMENSION(3, 3), INTENT(IN)               :: task
      LOGICAL, INTENT(IN)                                :: do_forces, do_efield, do_stress
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: radii, charges
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: dipoles
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: quadrupoles
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: forces, pv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: efield0
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: efield1, efield2

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_multipole_SR', &
         routineP = moduleN//':'//routineN

      INTEGER :: a, atom_a, atom_b, b, c, d, e, handle, i, iend, igrp, ikind, ilist, ipair, &
         istart, itype_ij, itype_ji, jkind, k, kind_a, kind_b, kk, nkdamp_ij, nkdamp_ji, nkinds, &
         npairs
      INTEGER, DIMENSION(:, :), POINTER                  :: list
      LOGICAL                                            :: do_efield0, do_efield1, do_efield2, &
                                                            force_eval
      REAL(KIND=dp) :: alpha, beta, ch_i, ch_j, dampa_ij, dampa_ji, dampaexpi, dampaexpj, &
         dampfac_ij, dampfac_ji, dampfuncdiffi, dampfuncdiffj, dampfunci, dampfuncj, dampsumfi, &
         dampsumfj, ef0_i, ef0_j, eloc, fac, fac_ij, factorial, ir, irab2, ptens11, ptens12, &
         ptens13, ptens21, ptens22, ptens23, ptens31, ptens32, ptens33, r, rab2, rab2_max, radius, &
         rcut, tij, tmp, tmp1, tmp11, tmp12, tmp13, tmp2, tmp21, tmp22, tmp23, tmp31, tmp32, &
         tmp33, tmp_ij, tmp_ji, xf
      REAL(KIND=dp), DIMENSION(0:5)                      :: f
      REAL(KIND=dp), DIMENSION(3)                        :: cell_v, cvi, damptij_a, damptji_a, dp_i, &
                                                            dp_j, ef1_i, ef1_j, fr, rab, tij_a
      REAL(KIND=dp), DIMENSION(3, 3)                     :: damptij_ab, damptji_ab, ef2_i, ef2_j, &
                                                            qp_i, qp_j, tij_ab
      REAL(KIND=dp), DIMENSION(3, 3, 3)                  :: tij_abc
      REAL(KIND=dp), DIMENSION(3, 3, 3, 3)               :: tij_abcd
      REAL(KIND=dp), DIMENSION(3, 3, 3, 3, 3)            :: tij_abcde
      TYPE(damping_type)                                 :: damping_ij, damping_ji
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update, r_last_update_pbc

      CALL timeset(routineN, handle)
      NULLIFY (nonbonded, r_last_update, r_last_update_pbc)
      do_efield0 = do_efield .AND. ASSOCIATED(efield0)
      do_efield1 = do_efield .AND. ASSOCIATED(efield1)
      do_efield2 = do_efield .AND. ASSOCIATED(efield2)
      IF (do_stress) THEN
         ptens11 = 0.0_dp; ptens12 = 0.0_dp; ptens13 = 0.0_dp
         ptens21 = 0.0_dp; ptens22 = 0.0_dp; ptens23 = 0.0_dp
         ptens31 = 0.0_dp; ptens32 = 0.0_dp; ptens33 = 0.0_dp
      END IF
      ! Get nonbond_env info
      CALL fist_nonbond_env_get(nonbond_env, nonbonded=nonbonded, natom_types=nkinds, &
                                r_last_update=r_last_update, r_last_update_pbc=r_last_update_pbc)
      CALL ewald_env_get(ewald_env, alpha=alpha, rcut=rcut)
      rab2_max = rcut**2
      IF (debug_r_space) THEN
         rab2_max = HUGE(0.0_dp)
      END IF
      ! Starting the force loop
      Lists: DO ilist = 1, nonbonded%nlists
         neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
         npairs = neighbor_kind_pair%npairs
         IF (npairs == 0) CYCLE
         list => neighbor_kind_pair%list
         cvi = neighbor_kind_pair%cell_vector
         CALL matvec_3x3(cell_v, cell%hmat, cvi)
         Kind_Group_Loop: DO igrp = 1, neighbor_kind_pair%ngrp_kind
            istart = neighbor_kind_pair%grp_kind_start(igrp)
            iend = neighbor_kind_pair%grp_kind_end(igrp)
            ikind = neighbor_kind_pair%ij_kind(1, igrp)
            jkind = neighbor_kind_pair%ij_kind(2, igrp)

            itype_ij = no_damping
            nkdamp_ij = 1
            dampa_ij = 1.0_dp
            dampfac_ij = 0.0_dp

            itype_ji = no_damping
            nkdamp_ji = 1
            dampa_ji = 1.0_dp
            dampfac_ji = 0.0_dp
            IF (PRESENT(atomic_kind_set)) THEN
               IF (ASSOCIATED(atomic_kind_set(jkind)%damping)) THEN
                  damping_ij = atomic_kind_set(jkind)%damping%damp(ikind)
                  itype_ij = damping_ij%itype
                  nkdamp_ij = damping_ij%order
                  dampa_ij = damping_ij%bij
                  dampfac_ij = damping_ij%cij
               END IF

               IF (ASSOCIATED(atomic_kind_set(ikind)%damping)) THEN
                  damping_ji = atomic_kind_set(ikind)%damping%damp(jkind)
                  itype_ji = damping_ji%itype
                  nkdamp_ji = damping_ji%order
                  dampa_ji = damping_ji%bij
                  dampfac_ji = damping_ji%cij
               END IF
            END IF

            Pairs: DO ipair = istart, iend
               IF (ipair <= neighbor_kind_pair%nscale) THEN
                  ! scale the electrostatic interaction if needed
                  ! (most often scaled to zero)
                  fac_ij = neighbor_kind_pair%ei_scale(ipair)
                  IF (fac_ij <= 0) CYCLE
               ELSE
                  fac_ij = 1.0_dp
               END IF
               atom_a = list(1, ipair)
               atom_b = list(2, ipair)
               kind_a = particle_set(atom_a)%atomic_kind%kind_number
               kind_b = particle_set(atom_b)%atomic_kind%kind_number
               IF (atom_a == atom_b) fac_ij = 0.5_dp
               rab = r_last_update_pbc(atom_b)%r-r_last_update_pbc(atom_a)%r
               rab = rab+cell_v
               rab2 = rab(1)**2+rab(2)**2+rab(3)**2
               IF (rab2 <= rab2_max) THEN
                  IF (PRESENT(radii)) THEN
                     radius = SQRT(radii(atom_a)*radii(atom_a)+radii(atom_b)*radii(atom_b))
                  ELSE
                     radius = 0.0_dp
                  END IF
                  IF (radius > 0.0_dp) THEN
                     beta = sqrthalf/radius
                ! Compute the Short Range constribution according the task
                IF (debug_this_module) THEN
                   f         = HUGE(0.0_dp)
                   tij       = HUGE(0.0_dp)
                   tij_a     = HUGE(0.0_dp)
                   tij_ab    = HUGE(0.0_dp)
                   tij_abc   = HUGE(0.0_dp)
                   tij_abcd  = HUGE(0.0_dp)
                   tij_abcde = HUGE(0.0_dp)
                END IF
                r     = SQRT(rab2)
                irab2 = 1.0_dp/rab2
                ir    = 1.0_dp/r

                ! Compute the radial function
                ! code for gaussian multipole with screening
                IF (debug_this_module.AND.debug_r_space.AND.(.NOT.debug_g_space)) THEN
                   f(0)  = ir
                   tmp1   = 0.0_dp
                   tmp2   = 0.0_dp
                ELSE
                   f(0)  = erf(beta*r)*ir - erf(alpha*r)*ir
                   tmp1   = EXP(-alpha**2*rab2)*oorootpi
                   tmp2   = EXP(-beta**2*rab2)*oorootpi
                END IF
                fac = 1.0_dp
                DO i = 1, 5
                   fac  = fac*REAL(2*i-1,KIND=dp)
                   f(i) = irab2*(f(i-1) + tmp1*((2.0_dp*alpha**2)**i)/(fac*alpha) - tmp2*((2.0_dp*beta**2)**i)/(fac*beta))
                END DO
                ! Compute the Tensor components
                force_eval = do_stress
                IF (task(1,1)) THEN
                   tij         = f(0)*fac_ij
                                                 force_eval = do_forces .OR.do_efield1
                END IF
                IF (task(2,2))                   force_eval = force_eval.OR.do_efield0
                IF (task(1,2).OR.force_eval) THEN
                   force_eval = do_stress
                   tij_a    = - rab*f(1)*fac_ij
                   IF (task(1,2))                force_eval = force_eval.OR. do_forces
                END IF
                IF (task(1,1))                   force_eval = force_eval.OR.do_efield2
                IF (task(3,3))                   force_eval = force_eval.OR.do_efield0
                IF (task(2,2).OR.task(3,1).OR.force_eval) THEN
                   force_eval = do_stress
                   DO b = 1,3
                      DO a = 1,3
                         tmp = rab(a)*rab(b)*fac_ij
                         tij_ab(a,b) = 3.0_dp*tmp*f(2)
                         IF (a==b) tij_ab(a,b) = tij_ab(a,b) - f(1)*fac_ij
                      END DO
                   END DO
                   IF (task(2,2).OR.task(3,1))   force_eval = force_eval.OR. do_forces
                END IF
                IF (task(2,2))                   force_eval = force_eval.OR.do_efield2
                IF (task(3,3))                   force_eval = force_eval.OR.do_efield1
                IF (task(3,2).OR.force_eval) THEN
                   force_eval = do_stress
                   DO c = 1, 3
                      DO b = 1, 3
                         DO a = 1, 3
                            tmp = rab(a)*rab(b)*rab(c)*fac_ij
                            tij_abc(a,b,c) = - 15.0_dp*tmp*f(3)
                            tmp = 3.0_dp*f(2)*fac_ij
                            IF (a==b) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(c)
                            IF (a==c) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(b)
                            IF (b==c) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(a)
                         END DO
                      END DO
                   END DO
                   IF (task(3,2))                force_eval = force_eval.OR. do_forces
                END IF
                IF (task(3,3).OR.force_eval) THEN
                   force_eval = do_stress
                   DO d = 1, 3
                      DO c = 1, 3
                         DO b = 1, 3
                            DO a = 1, 3
                               tmp = rab(a)*rab(b)*rab(c)*rab(d)*fac_ij
                               tij_abcd(a,b,c,d) = 105.0_dp*tmp*f(4)
                               tmp1 = 15.0_dp*f(3)*fac_ij
                               tmp2 =  3.0_dp*f(2)*fac_ij
                               IF (a==b) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(c)*rab(d)
                                  IF (c==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (a==c) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(b)*rab(d)
                                  IF (b==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (a==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(b)*rab(c)
                               IF (b==c) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(d)
                                  IF (a==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (b==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(c)
                               IF (c==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(b)
                            END DO
                         END DO
                      END DO
                   END DO
                   IF (task(3,3))                force_eval = force_eval.OR. do_forces
                END IF
                IF (force_eval) THEN
                   force_eval = do_stress
                   DO e = 1, 3
                      DO d = 1, 3
                         DO c = 1, 3
                            DO b = 1, 3
                               DO a = 1, 3
                                  tmp = rab(a)*rab(b)*rab(c)*rab(d)*rab(e)*fac_ij
                                  tij_abcde(a,b,c,d,e) = -945.0_dp*tmp*f(5)
                                  tmp1 = 105.0_dp*f(4)*fac_ij
                                  tmp2 =  15.0_dp*f(3)*fac_ij
                                  IF (a==b) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(c)*rab(d)*rab(e)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                  END IF
                                  IF (a==c) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(d)*rab(e)
                                     IF (b==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (b==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (a==d) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(c)*rab(e)
                                     IF (b==c) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (b==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (a==e) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(c)*rab(d)
                                     IF (b==c) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (b==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (b==c) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(d)*rab(e)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (b==d) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(c)*rab(e)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (b==e) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(c)*rab(d)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (c==d)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(e)
                                  IF (c==e)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(d)
                                  IF (d==e)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(c)
                               END DO
                            END DO
                         END DO
                      END DO
                   END DO
                END IF
                eloc  = 0.0_dp
                fr    = 0.0_dp
                ef0_i = 0.0_dp
                ef0_j = 0.0_dp
                ef1_j = 0.0_dp
                ef1_i = 0.0_dp
                ef2_j = 0.0_dp
                ef2_i = 0.0_dp


                ! Initialize the charge, dipole and quadrupole for atom A and B
                IF (debug_this_module) THEN
                   ch_j  = HUGE(0.0_dp)
                   ch_i  = HUGE(0.0_dp)
                   dp_j  = HUGE(0.0_dp)
                   dp_i  = HUGE(0.0_dp)
                   qp_j  = HUGE(0.0_dp)
                   qp_i  = HUGE(0.0_dp)
                END IF
                IF (ANY(task(1,:))) THEN
                   ch_j  = charges(atom_a)
                   ch_i  = charges(atom_b)
                END IF
                IF (ANY(task(2,:))) THEN
                   dp_j  = dipoles(:,atom_a)
                   dp_i  = dipoles(:,atom_b)
                END IF
                IF (ANY(task(3,:))) THEN
                   qp_j  = quadrupoles(:,:,atom_a)
                   qp_i  = quadrupoles(:,:,atom_b)
                END IF
                IF (task(1,1)) THEN
                   ! Charge - Charge
                   eloc = eloc + ch_i*tij*ch_j
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      fr(1) = fr(1) - ch_j * tij_a(1) * ch_i
                      fr(2) = fr(2) - ch_j * tij_a(2) * ch_i
                      fr(3) = fr(3) - ch_j * tij_a(3) * ch_i
                   END IF
                   ! Electric fields
                   IF (do_efield) THEN
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i + tij * ch_j

                         ef0_j = ef0_j + tij * ch_i
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) - tij_a(1) * ch_j
                         ef1_i(2) = ef1_i(2) - tij_a(2) * ch_j
                         ef1_i(3) = ef1_i(3) - tij_a(3) * ch_j

                         ef1_j(1) = ef1_j(1) + tij_a(1) * ch_i
                         ef1_j(2) = ef1_j(2) + tij_a(2) * ch_i
                         ef1_j(3) = ef1_j(3) + tij_a(3) * ch_i


                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         ef2_i(1,1) = ef2_i(1,1) - tij_ab(1,1) * ch_j
                         ef2_i(2,1) = ef2_i(2,1) - tij_ab(2,1) * ch_j
                         ef2_i(3,1) = ef2_i(3,1) - tij_ab(3,1) * ch_j
                         ef2_i(1,2) = ef2_i(1,2) - tij_ab(1,2) * ch_j
                         ef2_i(2,2) = ef2_i(2,2) - tij_ab(2,2) * ch_j
                         ef2_i(3,2) = ef2_i(3,2) - tij_ab(3,2) * ch_j
                         ef2_i(1,3) = ef2_i(1,3) - tij_ab(1,3) * ch_j
                         ef2_i(2,3) = ef2_i(2,3) - tij_ab(2,3) * ch_j
                         ef2_i(3,3) = ef2_i(3,3) - tij_ab(3,3) * ch_j

                         ef2_j(1,1) = ef2_j(1,1) - tij_ab(1,1) * ch_i
                         ef2_j(2,1) = ef2_j(2,1) - tij_ab(2,1) * ch_i
                         ef2_j(3,1) = ef2_j(3,1) - tij_ab(3,1) * ch_i
                         ef2_j(1,2) = ef2_j(1,2) - tij_ab(1,2) * ch_i
                         ef2_j(2,2) = ef2_j(2,2) - tij_ab(2,2) * ch_i
                         ef2_j(3,2) = ef2_j(3,2) - tij_ab(3,2) * ch_i
                         ef2_j(1,3) = ef2_j(1,3) - tij_ab(1,3) * ch_i
                         ef2_j(2,3) = ef2_j(2,3) - tij_ab(2,3) * ch_i
                         ef2_j(3,3) = ef2_j(3,3) - tij_ab(3,3) * ch_i
                      END IF
                   END IF
                END IF
                IF (task(2,2)) THEN
                   ! Dipole - Dipole
                   tmp= - (dp_i(1)*(tij_ab(1,1)*dp_j(1)+&
                                    tij_ab(2,1)*dp_j(2)+&
                                    tij_ab(3,1)*dp_j(3))+&
                           dp_i(2)*(tij_ab(1,2)*dp_j(1)+&
                                    tij_ab(2,2)*dp_j(2)+&
                                    tij_ab(3,2)*dp_j(3))+&
                           dp_i(3)*(tij_ab(1,3)*dp_j(1)+&
                                    tij_ab(2,3)*dp_j(2)+&
                                    tij_ab(3,3)*dp_j(3)))
                   eloc = eloc + tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         fr(k) = fr(k) +  dp_i(1)*(tij_abc(1,1,k)*dp_j(1)+&
                                                   tij_abc(2,1,k)*dp_j(2)+&
                                                   tij_abc(3,1,k)*dp_j(3))&
                                       +  dp_i(2)*(tij_abc(1,2,k)*dp_j(1)+&
                                                   tij_abc(2,2,k)*dp_j(2)+&
                                                   tij_abc(3,2,k)*dp_j(3))&
                                       +  dp_i(3)*(tij_abc(1,3,k)*dp_j(1)+&
                                                   tij_abc(2,3,k)*dp_j(2)+&
                                                   tij_abc(3,3,k)*dp_j(3))
                      END DO
                   END IF
                   ! Electric fields
                   IF (do_efield) THEN
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i - (tij_a(1)*dp_j(1)+&
                                          tij_a(2)*dp_j(2)+&
                                          tij_a(3)*dp_j(3))

                         ef0_j = ef0_j + (tij_a(1)*dp_i(1)+&
                                          tij_a(2)*dp_i(2)+&
                                          tij_a(3)*dp_i(3))
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) + (tij_ab(1,1)*dp_j(1)+&
                                                tij_ab(2,1)*dp_j(2)+&
                                                tij_ab(3,1)*dp_j(3))
                         ef1_i(2) = ef1_i(2) + (tij_ab(1,2)*dp_j(1)+&
                                                tij_ab(2,2)*dp_j(2)+&
                                                tij_ab(3,2)*dp_j(3))
                         ef1_i(3) = ef1_i(3) + (tij_ab(1,3)*dp_j(1)+&
                                                tij_ab(2,3)*dp_j(2)+&
                                                tij_ab(3,3)*dp_j(3))

                         ef1_j(1) = ef1_j(1) + (tij_ab(1,1)*dp_i(1)+&
                                                tij_ab(2,1)*dp_i(2)+&
                                                tij_ab(3,1)*dp_i(3))
                         ef1_j(2) = ef1_j(2) + (tij_ab(1,2)*dp_i(1)+&
                                                tij_ab(2,2)*dp_i(2)+&
                                                tij_ab(3,2)*dp_i(3))
                         ef1_j(3) = ef1_j(3) + (tij_ab(1,3)*dp_i(1)+&
                                                tij_ab(2,3)*dp_i(2)+&
                                                tij_ab(3,3)*dp_i(3))
                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         ef2_i(1,1) = ef2_i(1,1) + (tij_abc(1,1,1)*dp_j(1)+&
                                                    tij_abc(2,1,1)*dp_j(2)+&
                                                    tij_abc(3,1,1)*dp_j(3))
                         ef2_i(1,2) = ef2_i(1,2) + (tij_abc(1,1,2)*dp_j(1)+&
                                                    tij_abc(2,1,2)*dp_j(2)+&
                                                    tij_abc(3,1,2)*dp_j(3))
                         ef2_i(1,3) = ef2_i(1,3) + (tij_abc(1,1,3)*dp_j(1)+&
                                                    tij_abc(2,1,3)*dp_j(2)+&
                                                    tij_abc(3,1,3)*dp_j(3))
                         ef2_i(2,1) = ef2_i(2,1) + (tij_abc(1,2,1)*dp_j(1)+&
                                                    tij_abc(2,2,1)*dp_j(2)+&
                                                    tij_abc(3,2,1)*dp_j(3))
                         ef2_i(2,2) = ef2_i(2,2) + (tij_abc(1,2,2)*dp_j(1)+&
                                                    tij_abc(2,2,2)*dp_j(2)+&
                                                    tij_abc(3,2,2)*dp_j(3))
                         ef2_i(2,3) = ef2_i(2,3) + (tij_abc(1,2,3)*dp_j(1)+&
                                                    tij_abc(2,2,3)*dp_j(2)+&
                                                    tij_abc(3,2,3)*dp_j(3))
                         ef2_i(3,1) = ef2_i(3,1) + (tij_abc(1,3,1)*dp_j(1)+&
                                                    tij_abc(2,3,1)*dp_j(2)+&
                                                    tij_abc(3,3,1)*dp_j(3))
                         ef2_i(3,2) = ef2_i(3,2) + (tij_abc(1,3,2)*dp_j(1)+&
                                                    tij_abc(2,3,2)*dp_j(2)+&
                                                    tij_abc(3,3,2)*dp_j(3))
                         ef2_i(3,3) = ef2_i(3,3) + (tij_abc(1,3,3)*dp_j(1)+&
                                                    tij_abc(2,3,3)*dp_j(2)+&
                                                    tij_abc(3,3,3)*dp_j(3))

                         ef2_j(1,1) = ef2_j(1,1) - (tij_abc(1,1,1)*dp_i(1)+&
                                                    tij_abc(2,1,1)*dp_i(2)+&
                                                    tij_abc(3,1,1)*dp_i(3))
                         ef2_j(1,2) = ef2_j(1,2) - (tij_abc(1,1,2)*dp_i(1)+&
                                                    tij_abc(2,1,2)*dp_i(2)+&
                                                    tij_abc(3,1,2)*dp_i(3))
                         ef2_j(1,3) = ef2_j(1,3) - (tij_abc(1,1,3)*dp_i(1)+&
                                                    tij_abc(2,1,3)*dp_i(2)+&
                                                    tij_abc(3,1,3)*dp_i(3))
                         ef2_j(2,1) = ef2_j(2,1) - (tij_abc(1,2,1)*dp_i(1)+&
                                                    tij_abc(2,2,1)*dp_i(2)+&
                                                    tij_abc(3,2,1)*dp_i(3))
                         ef2_j(2,2) = ef2_j(2,2) - (tij_abc(1,2,2)*dp_i(1)+&
                                                    tij_abc(2,2,2)*dp_i(2)+&
                                                    tij_abc(3,2,2)*dp_i(3))
                         ef2_j(2,3) = ef2_j(2,3) - (tij_abc(1,2,3)*dp_i(1)+&
                                                    tij_abc(2,2,3)*dp_i(2)+&
                                                    tij_abc(3,2,3)*dp_i(3))
                         ef2_j(3,1) = ef2_j(3,1) - (tij_abc(1,3,1)*dp_i(1)+&
                                                    tij_abc(2,3,1)*dp_i(2)+&
                                                    tij_abc(3,3,1)*dp_i(3))
                         ef2_j(3,2) = ef2_j(3,2) - (tij_abc(1,3,2)*dp_i(1)+&
                                                    tij_abc(2,3,2)*dp_i(2)+&
                                                    tij_abc(3,3,2)*dp_i(3))
                         ef2_j(3,3) = ef2_j(3,3) - (tij_abc(1,3,3)*dp_i(1)+&
                                                    tij_abc(2,3,3)*dp_i(2)+&
                                                    tij_abc(3,3,3)*dp_i(3))
                      END IF
                   END IF
                END IF
                IF (task(2,1)) THEN
                   ! Dipole - Charge
                   tmp=   ch_j*(tij_a(1)*dp_i(1)+&
                                tij_a(2)*dp_i(2)+&
                                tij_a(3)*dp_i(3))&
                        - ch_i*(tij_a(1)*dp_j(1)+&
                                tij_a(2)*dp_j(2)+&
                                tij_a(3)*dp_j(3))
                   eloc = eloc + tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         fr(k) = fr(k) -  ch_j *(tij_ab(1,k)*dp_i(1)+&
                                                 tij_ab(2,k)*dp_i(2)+&
                                                 tij_ab(3,k)*dp_i(3))&
                                       +  ch_i *(tij_ab(1,k)*dp_j(1)+&
                                                 tij_ab(2,k)*dp_j(2)+&
                                                 tij_ab(3,k)*dp_j(3))
                      END DO
                   END IF
                END IF
                IF (task(3,3)) THEN
                   ! Quadrupole - Quadrupole
                   fac  = 1.0_dp/9.0_dp
                   tmp11 = qp_i(1,1)*(tij_abcd(1,1,1,1)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,1)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,1)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,1)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,1)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,1)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,1)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,1)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,1)*qp_j(3,3))
                   tmp21 = qp_i(2,1)*(tij_abcd(1,1,1,2)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,2)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,2)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,2)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,2)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,2)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,2)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,2)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,2)*qp_j(3,3))
                   tmp31 = qp_i(3,1)*(tij_abcd(1,1,1,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,3)*qp_j(3,3))
                   tmp22 = qp_i(2,2)*(tij_abcd(1,1,2,2)*qp_j(1,1)+&
                                      tij_abcd(2,1,2,2)*qp_j(2,1)+&
                                      tij_abcd(3,1,2,2)*qp_j(3,1)+&
                                      tij_abcd(1,2,2,2)*qp_j(1,2)+&
                                      tij_abcd(2,2,2,2)*qp_j(2,2)+&
                                      tij_abcd(3,2,2,2)*qp_j(3,2)+&
                                      tij_abcd(1,3,2,2)*qp_j(1,3)+&
                                      tij_abcd(2,3,2,2)*qp_j(2,3)+&
                                      tij_abcd(3,3,2,2)*qp_j(3,3))
                   tmp32 = qp_i(3,2)*(tij_abcd(1,1,2,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,2,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,2,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,2,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,2,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,2,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,2,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,2,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,2,3)*qp_j(3,3))
                   tmp33 = qp_i(3,3)*(tij_abcd(1,1,3,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,3,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,3,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,3,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,3,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,3,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,3,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,3,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,3,3)*qp_j(3,3))
                   tmp12 = tmp21
                   tmp13 = tmp31
                   tmp23 = tmp32
                   tmp   = tmp11 + tmp12 + tmp13 + &
                           tmp21 + tmp22 + tmp23 + &
                           tmp31 + tmp32 + tmp33

                   eloc = eloc + fac*tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         tmp11 = qp_i(1,1)*(tij_abcde(1,1,1,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,1,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,1,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,1,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,1,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,1,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,1,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,1,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,1,1,k)*qp_j(3,3))
                         tmp21 = qp_i(2,1)*(tij_abcde(1,1,2,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,2,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,2,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,2,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,2,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,2,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,2,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,2,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,2,1,k)*qp_j(3,3))
                         tmp31 = qp_i(3,1)*(tij_abcde(1,1,3,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,1,k)*qp_j(3,3))
                         tmp22 = qp_i(2,2)*(tij_abcde(1,1,2,2,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,2,2,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,2,2,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,2,2,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,2,2,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,2,2,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,2,2,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,2,2,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,2,2,k)*qp_j(3,3))
                         tmp32 = qp_i(3,2)*(tij_abcde(1,1,3,2,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,2,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,2,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,2,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,2,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,2,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,2,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,2,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,2,k)*qp_j(3,3))
                         tmp33 = qp_i(3,3)*(tij_abcde(1,1,3,3,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,3,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,3,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,3,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,3,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,3,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,3,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,3,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,3,k)*qp_j(3,3))
                         tmp12 = tmp21
                         tmp13 = tmp31
                         tmp23 = tmp32
                         fr(k) = fr(k) - fac * ( tmp11 + tmp12 + tmp13 +&
                                                 tmp21 + tmp22 + tmp23 +&
                                                 tmp31 + tmp32 + tmp33  )
                      END DO
                   END IF
                   ! Electric field
                   IF (do_efield) THEN
                      fac = 1.0_dp/3.0_dp
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i + fac*(tij_ab(1,1)*qp_j(1,1)+&
                                              tij_ab(2,1)*qp_j(2,1)+&
                                              tij_ab(3,1)*qp_j(3,1)+&
                                              tij_ab(1,2)*qp_j(1,2)+&
                                              tij_ab(2,2)*qp_j(2,2)+&
                                              tij_ab(3,2)*qp_j(3,2)+&
                                              tij_ab(1,3)*qp_j(1,3)+&
                                              tij_ab(2,3)*qp_j(2,3)+&
                                              tij_ab(3,3)*qp_j(3,3))

                         ef0_j = ef0_j + fac*(tij_ab(1,1)*qp_i(1,1)+&
                                              tij_ab(2,1)*qp_i(2,1)+&
                                              tij_ab(3,1)*qp_i(3,1)+&
                                              tij_ab(1,2)*qp_i(1,2)+&
                                              tij_ab(2,2)*qp_i(2,2)+&
                                              tij_ab(3,2)*qp_i(3,2)+&
                                              tij_ab(1,3)*qp_i(1,3)+&
                                              tij_ab(2,3)*qp_i(2,3)+&
                                              tij_ab(3,3)*qp_i(3,3))
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) - fac*(tij_abc(1,1,1)*qp_j(1,1)+&
                                                    tij_abc(2,1,1)*qp_j(2,1)+&
                                                    tij_abc(3,1,1)*qp_j(3,1)+&
                                                    tij_abc(1,2,1)*qp_j(1,2)+&
                                                    tij_abc(2,2,1)*qp_j(2,2)+&
                                                    tij_abc(3,2,1)*qp_j(3,2)+&
                                                    tij_abc(1,3,1)*qp_j(1,3)+&
                                                    tij_abc(2,3,1)*qp_j(2,3)+&
                                                    tij_abc(3,3,1)*qp_j(3,3))
                         ef1_i(2) = ef1_i(2) - fac*(tij_abc(1,1,2)*qp_j(1,1)+&
                                                    tij_abc(2,1,2)*qp_j(2,1)+&
                                                    tij_abc(3,1,2)*qp_j(3,1)+&
                                                    tij_abc(1,2,2)*qp_j(1,2)+&
                                                    tij_abc(2,2,2)*qp_j(2,2)+&
                                                    tij_abc(3,2,2)*qp_j(3,2)+&
                                                    tij_abc(1,3,2)*qp_j(1,3)+&
                                                    tij_abc(2,3,2)*qp_j(2,3)+&
                                                    tij_abc(3,3,2)*qp_j(3,3))
                         ef1_i(3) = ef1_i(3) - fac*(tij_abc(1,1,3)*qp_j(1,1)+&
                                                    tij_abc(2,1,3)*qp_j(2,1)+&
                                                    tij_abc(3,1,3)*qp_j(3,1)+&
                                                    tij_abc(1,2,3)*qp_j(1,2)+&
                                                    tij_abc(2,2,3)*qp_j(2,2)+&
                                                    tij_abc(3,2,3)*qp_j(3,2)+&
                                                    tij_abc(1,3,3)*qp_j(1,3)+&
                                                    tij_abc(2,3,3)*qp_j(2,3)+&
                                                    tij_abc(3,3,3)*qp_j(3,3))

                         ef1_j(1) = ef1_j(1) + fac*(tij_abc(1,1,1)*qp_i(1,1)+&
                                                    tij_abc(2,1,1)*qp_i(2,1)+&
                                                    tij_abc(3,1,1)*qp_i(3,1)+&
                                                    tij_abc(1,2,1)*qp_i(1,2)+&
                                                    tij_abc(2,2,1)*qp_i(2,2)+&
                                                    tij_abc(3,2,1)*qp_i(3,2)+&
                                                    tij_abc(1,3,1)*qp_i(1,3)+&
                                                    tij_abc(2,3,1)*qp_i(2,3)+&
                                                    tij_abc(3,3,1)*qp_i(3,3))
                         ef1_j(2) = ef1_j(2) + fac*(tij_abc(1,1,2)*qp_i(1,1)+&
                                                    tij_abc(2,1,2)*qp_i(2,1)+&
                                                    tij_abc(3,1,2)*qp_i(3,1)+&
                                                    tij_abc(1,2,2)*qp_i(1,2)+&
                                                    tij_abc(2,2,2)*qp_i(2,2)+&
                                                    tij_abc(3,2,2)*qp_i(3,2)+&
                                                    tij_abc(1,3,2)*qp_i(1,3)+&
                                                    tij_abc(2,3,2)*qp_i(2,3)+&
                                                    tij_abc(3,3,2)*qp_i(3,3))
                         ef1_j(3) = ef1_j(3) + fac*(tij_abc(1,1,3)*qp_i(1,1)+&
                                                    tij_abc(2,1,3)*qp_i(2,1)+&
                                                    tij_abc(3,1,3)*qp_i(3,1)+&
                                                    tij_abc(1,2,3)*qp_i(1,2)+&
                                                    tij_abc(2,2,3)*qp_i(2,2)+&
                                                    tij_abc(3,2,3)*qp_i(3,2)+&
                                                    tij_abc(1,3,3)*qp_i(1,3)+&
                                                    tij_abc(2,3,3)*qp_i(2,3)+&
                                                    tij_abc(3,3,3)*qp_i(3,3))
                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         tmp11 =   fac *(tij_abcd(1,1,1,1)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,1)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,1)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,1)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,1)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,1)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,1)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,1)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,1)*qp_j(3,3))
                         tmp12 =   fac *(tij_abcd(1,1,1,2)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,2)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,2)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,2)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,2)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,2)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,2)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,2)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,2)*qp_j(3,3))
                         tmp13 =   fac *(tij_abcd(1,1,1,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,3)*qp_j(3,3))
                         tmp22 =   fac *(tij_abcd(1,1,2,2)*qp_j(1,1)+&
                                         tij_abcd(2,1,2,2)*qp_j(2,1)+&
                                         tij_abcd(3,1,2,2)*qp_j(3,1)+&
                                         tij_abcd(1,2,2,2)*qp_j(1,2)+&
                                         tij_abcd(2,2,2,2)*qp_j(2,2)+&
                                         tij_abcd(3,2,2,2)*qp_j(3,2)+&
                                         tij_abcd(1,3,2,2)*qp_j(1,3)+&
                                         tij_abcd(2,3,2,2)*qp_j(2,3)+&
                                         tij_abcd(3,3,2,2)*qp_j(3,3))
                         tmp23 =   fac *(tij_abcd(1,1,2,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,2,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,2,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,2,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,2,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,2,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,2,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,2,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,2,3)*qp_j(3,3))
                         tmp33 =   fac *(tij_abcd(1,1,3,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,3,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,3,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,3,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,3,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,3,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,3,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,3,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,3,3)*qp_j(3,3))

                         ef2_i(1,1) = ef2_i(1,1) - tmp11
                         ef2_i(1,2) = ef2_i(1,2) - tmp12
                         ef2_i(1,3) = ef2_i(1,3) - tmp13
                         ef2_i(2,1) = ef2_i(2,1) - tmp12
                         ef2_i(2,2) = ef2_i(2,2) - tmp22
                         ef2_i(2,3) = ef2_i(2,3) - tmp23
                         ef2_i(3,1) = ef2_i(3,1) - tmp13
                         ef2_i(3,2) = ef2_i(3,2) - tmp23
                         ef2_i(3,3) = ef2_i(3,3) - tmp33

                         tmp11 =   fac *(tij_abcd(1,1,1,1)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,1)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,1)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,1)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,1)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,1)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,1)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,1)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,1)*qp_i(3,3))
                         tmp12 =   fac *(tij_abcd(1,1,1,2)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,2)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,2)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,2)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,2)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,2)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,2)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,2)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,2)*qp_i(3,3))
                         tmp13 =   fac *(tij_abcd(1,1,1,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,3)*qp_i(3,3))
                         tmp22 =   fac *(tij_abcd(1,1,2,2)*qp_i(1,1)+&
                                         tij_abcd(2,1,2,2)*qp_i(2,1)+&
                                         tij_abcd(3,1,2,2)*qp_i(3,1)+&
                                         tij_abcd(1,2,2,2)*qp_i(1,2)+&
                                         tij_abcd(2,2,2,2)*qp_i(2,2)+&
                                         tij_abcd(3,2,2,2)*qp_i(3,2)+&
                                         tij_abcd(1,3,2,2)*qp_i(1,3)+&
                                         tij_abcd(2,3,2,2)*qp_i(2,3)+&
                                         tij_abcd(3,3,2,2)*qp_i(3,3))
                         tmp23 =   fac *(tij_abcd(1,1,2,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,2,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,2,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,2,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,2,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,2,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,2,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,2,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,2,3)*qp_i(3,3))
                         tmp33 =   fac *(tij_abcd(1,1,3,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,3,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,3,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,3,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,3,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,3,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,3,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,3,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,3,3)*qp_i(3,3))

                         ef2_j(1,1) = ef2_j(1,1) - tmp11
                         ef2_j(1,2) = ef2_j(1,2) - tmp12
                         ef2_j(1,3) = ef2_j(1,3) - tmp13
                         ef2_j(2,1) = ef2_j(2,1) - tmp12
                         ef2_j(2,2) = ef2_j(2,2) - tmp22
                         ef2_j(2,3) = ef2_j(2,3) - tmp23
                         ef2_j(3,1) = ef2_j(3,1) - tmp13
                         ef2_j(3,2) = ef2_j(3,2) - tmp23
                         ef2_j(3,3) = ef2_j(3,3) - tmp33
                      END IF
                   END IF
                END IF
                IF (task(3,2)) THEN
                   ! Quadrupole - Dipole
                   fac = 1.0_dp/3.0_dp
                   ! Dipole i (locally B) - Quadrupole j (locally A)
                   tmp_ij = dp_i(1)*(tij_abc(1,1,1)*qp_j(1,1)+&
                                     tij_abc(2,1,1)*qp_j(2,1)+&
                                     tij_abc(3,1,1)*qp_j(3,1)+&
                                     tij_abc(1,2,1)*qp_j(1,2)+&
                                     tij_abc(2,2,1)*qp_j(2,2)+&
                                     tij_abc(3,2,1)*qp_j(3,2)+&
                                     tij_abc(1,3,1)*qp_j(1,3)+&
                                     tij_abc(2,3,1)*qp_j(2,3)+&
                                     tij_abc(3,3,1)*qp_j(3,3))+&
                            dp_i(2)*(tij_abc(1,1,2)*qp_j(1,1)+&
                                     tij_abc(2,1,2)*qp_j(2,1)+&
                                     tij_abc(3,1,2)*qp_j(3,1)+&
                                     tij_abc(1,2,2)*qp_j(1,2)+&
                                     tij_abc(2,2,2)*qp_j(2,2)+&
                                     tij_abc(3,2,2)*qp_j(3,2)+&
                                     tij_abc(1,3,2)*qp_j(1,3)+&
                                     tij_abc(2,3,2)*qp_j(2,3)+&
                                     tij_abc(3,3,2)*qp_j(3,3))+&
                            dp_i(3)*(tij_abc(1,1,3)*qp_j(1,1)+&
                                     tij_abc(2,1,3)*qp_j(2,1)+&
                                     tij_abc(3,1,3)*qp_j(3,1)+&
                                     tij_abc(1,2,3)*qp_j(1,2)+&
                                     tij_abc(2,2,3)*qp_j(2,2)+&
                                     tij_abc(3,2,3)*qp_j(3,2)+&
                                     tij_abc(1,3,3)*qp_j(1,3)+&
                                     tij_abc(2,3,3)*qp_j(2,3)+&
                                     tij_abc(3,3,3)*qp_j(3,3))

                   ! Dipole j (locally A) - Quadrupole i (locally B)
                   tmp_ji = dp_j(1)*(tij_abc(1,1,1)*qp_i(1,1)+&
                                     tij_abc(2,1,1)*qp_i(2,1)+&
                                     tij_abc(3,1,1)*qp_i(3,1)+&
                                     tij_abc(1,2,1)*qp_i(1,2)+&
                                     tij_abc(2,2,1)*qp_i(2,2)+&
                                     tij_abc(3,2,1)*qp_i(3,2)+&
                                     tij_abc(1,3,1)*qp_i(1,3)+&
                                     tij_abc(2,3,1)*qp_i(2,3)+&
                                     tij_abc(3,3,1)*qp_i(3,3))+&
                            dp_j(2)*(tij_abc(1,1,2)*qp_i(1,1)+&
                                     tij_abc(2,1,2)*qp_i(2,1)+&
                                     tij_abc(3,1,2)*qp_i(3,1)+&
                                     tij_abc(1,2,2)*qp_i(1,2)+&
                                     tij_abc(2,2,2)*qp_i(2,2)+&
                                     tij_abc(3,2,2)*qp_i(3,2)+&
                                     tij_abc(1,3,2)*qp_i(1,3)+&
                                     tij_abc(2,3,2)*qp_i(2,3)+&
                                     tij_abc(3,3,2)*qp_i(3,3))+&
                            dp_j(3)*(tij_abc(1,1,3)*qp_i(1,1)+&
                                     tij_abc(2,1,3)*qp_i(2,1)+&
                                     tij_abc(3,1,3)*qp_i(3,1)+&
                                     tij_abc(1,2,3)*qp_i(1,2)+&
                                     tij_abc(2,2,3)*qp_i(2,2)+&
                                     tij_abc(3,2,3)*qp_i(3,2)+&
                                     tij_abc(1,3,3)*qp_i(1,3)+&
                                     tij_abc(2,3,3)*qp_i(2,3)+&
                                     tij_abc(3,3,3)*qp_i(3,3))

                   tmp= fac * (tmp_ij - tmp_ji)
                   eloc = eloc + tmp
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         ! Dipole i (locally B) - Quadrupole j (locally A)
                         tmp_ij = dp_i(1)*(tij_abcd(1,1,1,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,1,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,1,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,1,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,1,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,1,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,1,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,1,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,1,k)*qp_j(3,3))+&
                                  dp_i(2)*(tij_abcd(1,1,2,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,2,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,2,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,2,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,2,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,2,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,2,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,2,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,2,k)*qp_j(3,3))+&
                                  dp_i(3)*(tij_abcd(1,1,3,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,3,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,3,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,3,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,3,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,3,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,3,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,3,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,3,k)*qp_j(3,3))

                         ! Dipole j (locally A) - Quadrupole i (locally B)
                         tmp_ji = dp_j(1)*(tij_abcd(1,1,1,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,1,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,1,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,1,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,1,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,1,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,1,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,1,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,1,k)*qp_i(3,3))+&
                                  dp_j(2)*(tij_abcd(1,1,2,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,2,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,2,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,2,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,2,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,2,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,2,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,2,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,2,k)*qp_i(3,3))+&
                                  dp_j(3)*(tij_abcd(1,1,3,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,3,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,3,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,3,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,3,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,3,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,3,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,3,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,3,k)*qp_i(3,3))

                         fr(k) = fr(k) - fac * (tmp_ij - tmp_ji)
                      END DO
                   END IF
                END IF
                IF (task(3,1)) THEN
                   ! Quadrupole - Charge
                   fac = 1.0_dp/3.0_dp

                   ! Quadrupole j (locally A) - Charge j (locally B)
                   tmp_ij = ch_i * (tij_ab(1,1)*qp_j(1,1)+&
                                    tij_ab(2,1)*qp_j(2,1)+&
                                    tij_ab(3,1)*qp_j(3,1)+&
                                    tij_ab(1,2)*qp_j(1,2)+&
                                    tij_ab(2,2)*qp_j(2,2)+&
                                    tij_ab(3,2)*qp_j(3,2)+&
                                    tij_ab(1,3)*qp_j(1,3)+&
                                    tij_ab(2,3)*qp_j(2,3)+&
                                    tij_ab(3,3)*qp_j(3,3))

                   ! Quadrupole i (locally B) - Charge j (locally A)
                   tmp_ji = ch_j * (tij_ab(1,1)*qp_i(1,1)+&
                                    tij_ab(2,1)*qp_i(2,1)+&
                                    tij_ab(3,1)*qp_i(3,1)+&
                                    tij_ab(1,2)*qp_i(1,2)+&
                                    tij_ab(2,2)*qp_i(2,2)+&
                                    tij_ab(3,2)*qp_i(3,2)+&
                                    tij_ab(1,3)*qp_i(1,3)+&
                                    tij_ab(2,3)*qp_i(2,3)+&
                                    tij_ab(3,3)*qp_i(3,3))

                   eloc = eloc + fac*(tmp_ij+tmp_ji)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         ! Quadrupole j (locally A) - Charge i (locally B)
                         tmp_ij = ch_i * (tij_abc(1,1,k)*qp_j(1,1)+&
                                          tij_abc(2,1,k)*qp_j(2,1)+&
                                          tij_abc(3,1,k)*qp_j(3,1)+&
                                          tij_abc(1,2,k)*qp_j(1,2)+&
                                          tij_abc(2,2,k)*qp_j(2,2)+&
                                          tij_abc(3,2,k)*qp_j(3,2)+&
                                          tij_abc(1,3,k)*qp_j(1,3)+&
                                          tij_abc(2,3,k)*qp_j(2,3)+&
                                          tij_abc(3,3,k)*qp_j(3,3))

                         ! Quadrupole i (locally B) - Charge j (locally A)
                         tmp_ji = ch_j * (tij_abc(1,1,k)*qp_i(1,1)+&
                                          tij_abc(2,1,k)*qp_i(2,1)+&
                                          tij_abc(3,1,k)*qp_i(3,1)+&
                                          tij_abc(1,2,k)*qp_i(1,2)+&
                                          tij_abc(2,2,k)*qp_i(2,2)+&
                                          tij_abc(3,2,k)*qp_i(3,2)+&
                                          tij_abc(1,3,k)*qp_i(1,3)+&
                                          tij_abc(2,3,k)*qp_i(2,3)+&
                                          tij_abc(3,3,k)*qp_i(3,3))

                         fr(k) = fr(k) - fac *(tmp_ij + tmp_ji)
                      END DO
                   END IF
                END IF
                energy = energy + eloc
                IF (do_forces) THEN
                   forces(1,atom_a) = forces(1,atom_a) - fr(1)
                   forces(2,atom_a) = forces(2,atom_a) - fr(2)
                   forces(3,atom_a) = forces(3,atom_a) - fr(3)
                   forces(1,atom_b) = forces(1,atom_b) + fr(1)
                   forces(2,atom_b) = forces(2,atom_b) + fr(2)
                   forces(3,atom_b) = forces(3,atom_b) + fr(3)
                END IF
                ! Electric fields
                IF (do_efield) THEN
                   ! Potential
                   IF (do_efield0) THEN
                      efield0(  atom_a) = efield0(  atom_a) + ef0_j

                      efield0(  atom_b) = efield0(  atom_b) + ef0_i
                   END IF
                   ! Electric field
                   IF (do_efield1) THEN
                      efield1(1,atom_a) = efield1(1,atom_a) + ef1_j(1)
                      efield1(2,atom_a) = efield1(2,atom_a) + ef1_j(2)
                      efield1(3,atom_a) = efield1(3,atom_a) + ef1_j(3)

                      efield1(1,atom_b) = efield1(1,atom_b) + ef1_i(1)
                      efield1(2,atom_b) = efield1(2,atom_b) + ef1_i(2)
                      efield1(3,atom_b) = efield1(3,atom_b) + ef1_i(3)
                   END IF
                   ! Electric field gradient
                   IF (do_efield2) THEN
                      efield2(1,atom_a) = efield2(1,atom_a) + ef2_j(1,1)
                      efield2(2,atom_a) = efield2(2,atom_a) + ef2_j(1,2)
                      efield2(3,atom_a) = efield2(3,atom_a) + ef2_j(1,3)
                      efield2(4,atom_a) = efield2(4,atom_a) + ef2_j(2,1)
                      efield2(5,atom_a) = efield2(5,atom_a) + ef2_j(2,2)
                      efield2(6,atom_a) = efield2(6,atom_a) + ef2_j(2,3)
                      efield2(7,atom_a) = efield2(7,atom_a) + ef2_j(3,1)
                      efield2(8,atom_a) = efield2(8,atom_a) + ef2_j(3,2)
                      efield2(9,atom_a) = efield2(9,atom_a) + ef2_j(3,3)

                      efield2(1,atom_b) = efield2(1,atom_b) + ef2_i(1,1)
                      efield2(2,atom_b) = efield2(2,atom_b) + ef2_i(1,2)
                      efield2(3,atom_b) = efield2(3,atom_b) + ef2_i(1,3)
                      efield2(4,atom_b) = efield2(4,atom_b) + ef2_i(2,1)
                      efield2(5,atom_b) = efield2(5,atom_b) + ef2_i(2,2)
                      efield2(6,atom_b) = efield2(6,atom_b) + ef2_i(2,3)
                      efield2(7,atom_b) = efield2(7,atom_b) + ef2_i(3,1)
                      efield2(8,atom_b) = efield2(8,atom_b) + ef2_i(3,2)
                      efield2(9,atom_b) = efield2(9,atom_b) + ef2_i(3,3)
                   END IF
                END IF
                IF (do_stress) THEN
                   ptens11 = ptens11 + rab(1) * fr(1)
                   ptens21 = ptens21 + rab(2) * fr(1)
                   ptens31 = ptens31 + rab(3) * fr(1)
                   ptens12 = ptens12 + rab(1) * fr(2)
                   ptens22 = ptens22 + rab(2) * fr(2)
                   ptens32 = ptens32 + rab(3) * fr(2)
                   ptens13 = ptens13 + rab(1) * fr(3)
                   ptens23 = ptens23 + rab(2) * fr(3)
                   ptens33 = ptens33 + rab(3) * fr(3)
                END IF
                  ELSE
                ! Compute the Short Range constribution according the task
                IF (debug_this_module) THEN
                   f         = HUGE(0.0_dp)
                   tij       = HUGE(0.0_dp)
                   tij_a     = HUGE(0.0_dp)
                   tij_ab    = HUGE(0.0_dp)
                   tij_abc   = HUGE(0.0_dp)
                   tij_abcd  = HUGE(0.0_dp)
                   tij_abcde = HUGE(0.0_dp)
                END IF
                r     = SQRT(rab2)
                irab2 = 1.0_dp/rab2
                ir    = 1.0_dp/r

                ! Compute the radial function
                ! code for point multipole with screening
                IF (debug_this_module.AND.debug_r_space.AND.(.NOT.debug_g_space)) THEN
                   f(0)  = ir
                   tmp   = 0.0_dp
                ELSE
                   f(0)  = erfc(alpha*r)*ir
                   tmp   = EXP(-alpha**2*rab2)*oorootpi
                END IF
                fac = 1.0_dp
                DO i = 1, 5
                   fac  = fac*REAL(2*i-1,KIND=dp)
                   f(i) = irab2*(f(i-1)+ tmp*((2.0_dp*alpha**2)**i)/(fac*alpha))
                END DO
                ! Compute the Tensor components
                force_eval = do_stress
                IF (task(1,1)) THEN
                   tij         = f(0)*fac_ij
                                                 force_eval = do_forces .OR.do_efield1
                END IF
                IF (task(2,2))                   force_eval = force_eval.OR.do_efield0
                IF (task(1,2).OR.force_eval) THEN
                   force_eval = do_stress
                   tij_a    = - rab*f(1)*fac_ij
                   IF (task(1,2))                force_eval = force_eval.OR. do_forces
                END IF
                IF (task(1,1))                   force_eval = force_eval.OR.do_efield2
                IF (task(3,3))                   force_eval = force_eval.OR.do_efield0
                IF (task(2,2).OR.task(3,1).OR.force_eval) THEN
                   force_eval = do_stress
                   DO b = 1,3
                      DO a = 1,3
                         tmp = rab(a)*rab(b)*fac_ij
                         tij_ab(a,b) = 3.0_dp*tmp*f(2)
                         IF (a==b) tij_ab(a,b) = tij_ab(a,b) - f(1)*fac_ij
                      END DO
                   END DO
                   IF (task(2,2).OR.task(3,1))   force_eval = force_eval.OR. do_forces
                END IF
                IF (task(2,2))                   force_eval = force_eval.OR.do_efield2
                IF (task(3,3))                   force_eval = force_eval.OR.do_efield1
                IF (task(3,2).OR.force_eval) THEN
                   force_eval = do_stress
                   DO c = 1, 3
                      DO b = 1, 3
                         DO a = 1, 3
                            tmp = rab(a)*rab(b)*rab(c)*fac_ij
                            tij_abc(a,b,c) = - 15.0_dp*tmp*f(3)
                            tmp = 3.0_dp*f(2)*fac_ij
                            IF (a==b) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(c)
                            IF (a==c) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(b)
                            IF (b==c) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(a)
                         END DO
                      END DO
                   END DO
                   IF (task(3,2))                force_eval = force_eval.OR. do_forces
                END IF
                IF (task(3,3).OR.force_eval) THEN
                   force_eval = do_stress
                   DO d = 1, 3
                      DO c = 1, 3
                         DO b = 1, 3
                            DO a = 1, 3
                               tmp = rab(a)*rab(b)*rab(c)*rab(d)*fac_ij
                               tij_abcd(a,b,c,d) = 105.0_dp*tmp*f(4)
                               tmp1 = 15.0_dp*f(3)*fac_ij
                               tmp2 =  3.0_dp*f(2)*fac_ij
                               IF (a==b) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(c)*rab(d)
                                  IF (c==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (a==c) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(b)*rab(d)
                                  IF (b==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (a==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(b)*rab(c)
                               IF (b==c) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(d)
                                  IF (a==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (b==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(c)
                               IF (c==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(b)
                            END DO
                         END DO
                      END DO
                   END DO
                   IF (task(3,3))                force_eval = force_eval.OR. do_forces
                END IF
                IF (force_eval) THEN
                   force_eval = do_stress
                   DO e = 1, 3
                      DO d = 1, 3
                         DO c = 1, 3
                            DO b = 1, 3
                               DO a = 1, 3
                                  tmp = rab(a)*rab(b)*rab(c)*rab(d)*rab(e)*fac_ij
                                  tij_abcde(a,b,c,d,e) = -945.0_dp*tmp*f(5)
                                  tmp1 = 105.0_dp*f(4)*fac_ij
                                  tmp2 =  15.0_dp*f(3)*fac_ij
                                  IF (a==b) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(c)*rab(d)*rab(e)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                  END IF
                                  IF (a==c) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(d)*rab(e)
                                     IF (b==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (b==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (a==d) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(c)*rab(e)
                                     IF (b==c) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (b==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (a==e) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(c)*rab(d)
                                     IF (b==c) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (b==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (b==c) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(d)*rab(e)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (b==d) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(c)*rab(e)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (b==e) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(c)*rab(d)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (c==d)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(e)
                                  IF (c==e)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(d)
                                  IF (d==e)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(c)
                               END DO
                            END DO
                         END DO
                      END DO
                   END DO
                END IF
                eloc  = 0.0_dp
                fr    = 0.0_dp
                ef0_i = 0.0_dp
                ef0_j = 0.0_dp
                ef1_j = 0.0_dp
                ef1_i = 0.0_dp
                ef2_j = 0.0_dp
                ef2_i = 0.0_dp


                ! Initialize the damping function.
                IF (kind_a==ikind) THEN
                   ! for atom i
                   SELECT CASE (itype_ij)
                   CASE (tang_toennies)
                      dampsumfi = 1.0_dp
                      xf = 1.0_dp
                      factorial = 1.0_dp
                      DO kk = 1, nkdamp_ij
                         xf = xf*dampa_ij*r
                         factorial = factorial * float(kk)
                         dampsumfi = dampsumfi + (xf/factorial)
                      END DO
                      dampaexpi = dexp(-dampa_ij * r)
                      dampfunci = dampsumfi * dampaexpi * dampfac_ij
                      dampfuncdiffi = -dampa_ij * dampaexpi * &
                                      dampfac_ij * (((dampa_ij * r) ** nkdamp_ij) / &
                                      factorial)
                   CASE DEFAULT
                      dampfunci=0.0_dp
                      dampfuncdiffi=0.0_dp
                   END SELECT

                   ! for atom j
                   SELECT CASE (itype_ji)
                   CASE (tang_toennies)
                      dampsumfj = 1.0_dp
                      xf = 1.0_dp
                      factorial = 1.0_dp
                      DO kk = 1, nkdamp_ji
                         xf = xf*dampa_ji*r
                         factorial = factorial * float(kk)
                         dampsumfj = dampsumfj + (xf/factorial)
                      END DO
                      dampaexpj = dexp(-dampa_ji * r)
                      dampfuncj = dampsumfj * dampaexpj * dampfac_ji
                      dampfuncdiffj = -dampa_ji * dampaexpj * &
                                      dampfac_ji * (((dampa_ji * r) ** nkdamp_ji) / &
                                      factorial)
                   CASE DEFAULT
                      dampfuncj = 0.0_dp
                      dampfuncdiffj = 0.0_dp
                   END SELECT
                ELSE
                   SELECT CASE (itype_ij)
                   CASE(tang_toennies)
                      dampsumfj = 1.0_dp
                      xf = 1.0_dp
                      factorial = 1.0_dp
                      DO kk = 1, nkdamp_ij
                         xf = xf*dampa_ij*r
                         factorial = factorial * float(kk)
                         dampsumfj = dampsumfj + (xf/factorial)
                      END DO
                      dampaexpj = dexp(-dampa_ij * r)
                      dampfuncj = dampsumfj * dampaexpj * dampfac_ij
                      dampfuncdiffj = -dampa_ij * dampaexpj * &
                                      dampfac_ij * (((dampa_ij * r) ** nkdamp_ij) / &
                                      factorial)
                   CASE DEFAULT
                      dampfuncj=0.0_dp
                      dampfuncdiffj=0.0_dp
                   END SELECT

                   !for j
                   SELECT CASE (itype_ji)
                   CASE (tang_toennies)
                      dampsumfi = 1.0_dp
                      xf = 1.0_dp
                      factorial = 1.0_dp
                      DO kk = 1, nkdamp_ji
                         xf = xf*dampa_ji*r
                         factorial = factorial * float(kk)
                         dampsumfi = dampsumfi + (xf/factorial)
                      END DO
                      dampaexpi = dexp(-dampa_ji * r)
                      dampfunci = dampsumfi * dampaexpi * dampfac_ji
                      dampfuncdiffi = -dampa_ji * dampaexpi * &
                                      dampfac_ji * (((dampa_ji * r) ** nkdamp_ji) / &
                                      factorial)
                   CASE DEFAULT
                      dampfunci = 0.0_dp
                      dampfuncdiffi = 0.0_dp
                   END SELECT
                END IF

                damptij_a = -rab*dampfunci*fac_ij*irab2*ir
                damptji_a = -rab*dampfuncj*fac_ij*irab2*ir
                DO b = 1,3
                   DO a = 1,3
                      tmp = rab(a)*rab(b)*fac_ij
                      damptij_ab(a,b) = tmp*(-dampfuncdiffi*irab2*irab2+3.0_dp*dampfunci*irab2*irab2*ir)
                      damptji_ab(a,b) = tmp*(-dampfuncdiffj*irab2*irab2+3.0_dp*dampfuncj*irab2*irab2*ir)
                      IF (a==b) damptij_ab(a,b) = damptij_ab(a,b) - dampfunci*fac_ij*irab2*ir
                      IF (a==b) damptji_ab(a,b) = damptji_ab(a,b) - dampfuncj*fac_ij*irab2*ir
                   END DO
                END DO


                ! Initialize the charge, dipole and quadrupole for atom A and B
                IF (debug_this_module) THEN
                   ch_j  = HUGE(0.0_dp)
                   ch_i  = HUGE(0.0_dp)
                   dp_j  = HUGE(0.0_dp)
                   dp_i  = HUGE(0.0_dp)
                   qp_j  = HUGE(0.0_dp)
                   qp_i  = HUGE(0.0_dp)
                END IF
                IF (ANY(task(1,:))) THEN
                   ch_j  = charges(atom_a)
                   ch_i  = charges(atom_b)
                END IF
                IF (ANY(task(2,:))) THEN
                   dp_j  = dipoles(:,atom_a)
                   dp_i  = dipoles(:,atom_b)
                END IF
                IF (ANY(task(3,:))) THEN
                   qp_j  = quadrupoles(:,:,atom_a)
                   qp_i  = quadrupoles(:,:,atom_b)
                END IF
                IF (task(1,1)) THEN
                   ! Charge - Charge
                   eloc = eloc + ch_i*tij*ch_j
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      fr(1) = fr(1) - ch_j * tij_a(1) * ch_i
                      fr(2) = fr(2) - ch_j * tij_a(2) * ch_i
                      fr(3) = fr(3) - ch_j * tij_a(3) * ch_i
                   END IF
                   ! Electric fields
                   IF (do_efield) THEN
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i + tij * ch_j

                         ef0_j = ef0_j + tij * ch_i
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) - tij_a(1) * ch_j
                         ef1_i(2) = ef1_i(2) - tij_a(2) * ch_j
                         ef1_i(3) = ef1_i(3) - tij_a(3) * ch_j

                         ef1_j(1) = ef1_j(1) + tij_a(1) * ch_i
                         ef1_j(2) = ef1_j(2) + tij_a(2) * ch_i
                         ef1_j(3) = ef1_j(3) + tij_a(3) * ch_i

                         ef1_i(1) = ef1_i(1) + damptij_a(1) * ch_j
                         ef1_i(2) = ef1_i(2) + damptij_a(2) * ch_j
                         ef1_i(3) = ef1_i(3) + damptij_a(3) * ch_j

                         ef1_j(1) = ef1_j(1) - damptji_a(1) * ch_i
                         ef1_j(2) = ef1_j(2) - damptji_a(2) * ch_i
                         ef1_j(3) = ef1_j(3) - damptji_a(3) * ch_i

                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         ef2_i(1,1) = ef2_i(1,1) - tij_ab(1,1) * ch_j
                         ef2_i(2,1) = ef2_i(2,1) - tij_ab(2,1) * ch_j
                         ef2_i(3,1) = ef2_i(3,1) - tij_ab(3,1) * ch_j
                         ef2_i(1,2) = ef2_i(1,2) - tij_ab(1,2) * ch_j
                         ef2_i(2,2) = ef2_i(2,2) - tij_ab(2,2) * ch_j
                         ef2_i(3,2) = ef2_i(3,2) - tij_ab(3,2) * ch_j
                         ef2_i(1,3) = ef2_i(1,3) - tij_ab(1,3) * ch_j
                         ef2_i(2,3) = ef2_i(2,3) - tij_ab(2,3) * ch_j
                         ef2_i(3,3) = ef2_i(3,3) - tij_ab(3,3) * ch_j

                         ef2_j(1,1) = ef2_j(1,1) - tij_ab(1,1) * ch_i
                         ef2_j(2,1) = ef2_j(2,1) - tij_ab(2,1) * ch_i
                         ef2_j(3,1) = ef2_j(3,1) - tij_ab(3,1) * ch_i
                         ef2_j(1,2) = ef2_j(1,2) - tij_ab(1,2) * ch_i
                         ef2_j(2,2) = ef2_j(2,2) - tij_ab(2,2) * ch_i
                         ef2_j(3,2) = ef2_j(3,2) - tij_ab(3,2) * ch_i
                         ef2_j(1,3) = ef2_j(1,3) - tij_ab(1,3) * ch_i
                         ef2_j(2,3) = ef2_j(2,3) - tij_ab(2,3) * ch_i
                         ef2_j(3,3) = ef2_j(3,3) - tij_ab(3,3) * ch_i
                      END IF
                   END IF
                END IF
                IF (task(2,2)) THEN
                   ! Dipole - Dipole
                   tmp= - (dp_i(1)*(tij_ab(1,1)*dp_j(1)+&
                                    tij_ab(2,1)*dp_j(2)+&
                                    tij_ab(3,1)*dp_j(3))+&
                           dp_i(2)*(tij_ab(1,2)*dp_j(1)+&
                                    tij_ab(2,2)*dp_j(2)+&
                                    tij_ab(3,2)*dp_j(3))+&
                           dp_i(3)*(tij_ab(1,3)*dp_j(1)+&
                                    tij_ab(2,3)*dp_j(2)+&
                                    tij_ab(3,3)*dp_j(3)))
                   eloc = eloc + tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         fr(k) = fr(k) +  dp_i(1)*(tij_abc(1,1,k)*dp_j(1)+&
                                                   tij_abc(2,1,k)*dp_j(2)+&
                                                   tij_abc(3,1,k)*dp_j(3))&
                                       +  dp_i(2)*(tij_abc(1,2,k)*dp_j(1)+&
                                                   tij_abc(2,2,k)*dp_j(2)+&
                                                   tij_abc(3,2,k)*dp_j(3))&
                                       +  dp_i(3)*(tij_abc(1,3,k)*dp_j(1)+&
                                                   tij_abc(2,3,k)*dp_j(2)+&
                                                   tij_abc(3,3,k)*dp_j(3))
                      END DO
                   END IF
                   ! Electric fields
                   IF (do_efield) THEN
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i - (tij_a(1)*dp_j(1)+&
                                          tij_a(2)*dp_j(2)+&
                                          tij_a(3)*dp_j(3))

                         ef0_j = ef0_j + (tij_a(1)*dp_i(1)+&
                                          tij_a(2)*dp_i(2)+&
                                          tij_a(3)*dp_i(3))
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) + (tij_ab(1,1)*dp_j(1)+&
                                                tij_ab(2,1)*dp_j(2)+&
                                                tij_ab(3,1)*dp_j(3))
                         ef1_i(2) = ef1_i(2) + (tij_ab(1,2)*dp_j(1)+&
                                                tij_ab(2,2)*dp_j(2)+&
                                                tij_ab(3,2)*dp_j(3))
                         ef1_i(3) = ef1_i(3) + (tij_ab(1,3)*dp_j(1)+&
                                                tij_ab(2,3)*dp_j(2)+&
                                                tij_ab(3,3)*dp_j(3))

                         ef1_j(1) = ef1_j(1) + (tij_ab(1,1)*dp_i(1)+&
                                                tij_ab(2,1)*dp_i(2)+&
                                                tij_ab(3,1)*dp_i(3))
                         ef1_j(2) = ef1_j(2) + (tij_ab(1,2)*dp_i(1)+&
                                                tij_ab(2,2)*dp_i(2)+&
                                                tij_ab(3,2)*dp_i(3))
                         ef1_j(3) = ef1_j(3) + (tij_ab(1,3)*dp_i(1)+&
                                                tij_ab(2,3)*dp_i(2)+&
                                                tij_ab(3,3)*dp_i(3))
                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         ef2_i(1,1) = ef2_i(1,1) + (tij_abc(1,1,1)*dp_j(1)+&
                                                    tij_abc(2,1,1)*dp_j(2)+&
                                                    tij_abc(3,1,1)*dp_j(3))
                         ef2_i(1,2) = ef2_i(1,2) + (tij_abc(1,1,2)*dp_j(1)+&
                                                    tij_abc(2,1,2)*dp_j(2)+&
                                                    tij_abc(3,1,2)*dp_j(3))
                         ef2_i(1,3) = ef2_i(1,3) + (tij_abc(1,1,3)*dp_j(1)+&
                                                    tij_abc(2,1,3)*dp_j(2)+&
                                                    tij_abc(3,1,3)*dp_j(3))
                         ef2_i(2,1) = ef2_i(2,1) + (tij_abc(1,2,1)*dp_j(1)+&
                                                    tij_abc(2,2,1)*dp_j(2)+&
                                                    tij_abc(3,2,1)*dp_j(3))
                         ef2_i(2,2) = ef2_i(2,2) + (tij_abc(1,2,2)*dp_j(1)+&
                                                    tij_abc(2,2,2)*dp_j(2)+&
                                                    tij_abc(3,2,2)*dp_j(3))
                         ef2_i(2,3) = ef2_i(2,3) + (tij_abc(1,2,3)*dp_j(1)+&
                                                    tij_abc(2,2,3)*dp_j(2)+&
                                                    tij_abc(3,2,3)*dp_j(3))
                         ef2_i(3,1) = ef2_i(3,1) + (tij_abc(1,3,1)*dp_j(1)+&
                                                    tij_abc(2,3,1)*dp_j(2)+&
                                                    tij_abc(3,3,1)*dp_j(3))
                         ef2_i(3,2) = ef2_i(3,2) + (tij_abc(1,3,2)*dp_j(1)+&
                                                    tij_abc(2,3,2)*dp_j(2)+&
                                                    tij_abc(3,3,2)*dp_j(3))
                         ef2_i(3,3) = ef2_i(3,3) + (tij_abc(1,3,3)*dp_j(1)+&
                                                    tij_abc(2,3,3)*dp_j(2)+&
                                                    tij_abc(3,3,3)*dp_j(3))

                         ef2_j(1,1) = ef2_j(1,1) - (tij_abc(1,1,1)*dp_i(1)+&
                                                    tij_abc(2,1,1)*dp_i(2)+&
                                                    tij_abc(3,1,1)*dp_i(3))
                         ef2_j(1,2) = ef2_j(1,2) - (tij_abc(1,1,2)*dp_i(1)+&
                                                    tij_abc(2,1,2)*dp_i(2)+&
                                                    tij_abc(3,1,2)*dp_i(3))
                         ef2_j(1,3) = ef2_j(1,3) - (tij_abc(1,1,3)*dp_i(1)+&
                                                    tij_abc(2,1,3)*dp_i(2)+&
                                                    tij_abc(3,1,3)*dp_i(3))
                         ef2_j(2,1) = ef2_j(2,1) - (tij_abc(1,2,1)*dp_i(1)+&
                                                    tij_abc(2,2,1)*dp_i(2)+&
                                                    tij_abc(3,2,1)*dp_i(3))
                         ef2_j(2,2) = ef2_j(2,2) - (tij_abc(1,2,2)*dp_i(1)+&
                                                    tij_abc(2,2,2)*dp_i(2)+&
                                                    tij_abc(3,2,2)*dp_i(3))
                         ef2_j(2,3) = ef2_j(2,3) - (tij_abc(1,2,3)*dp_i(1)+&
                                                    tij_abc(2,2,3)*dp_i(2)+&
                                                    tij_abc(3,2,3)*dp_i(3))
                         ef2_j(3,1) = ef2_j(3,1) - (tij_abc(1,3,1)*dp_i(1)+&
                                                    tij_abc(2,3,1)*dp_i(2)+&
                                                    tij_abc(3,3,1)*dp_i(3))
                         ef2_j(3,2) = ef2_j(3,2) - (tij_abc(1,3,2)*dp_i(1)+&
                                                    tij_abc(2,3,2)*dp_i(2)+&
                                                    tij_abc(3,3,2)*dp_i(3))
                         ef2_j(3,3) = ef2_j(3,3) - (tij_abc(1,3,3)*dp_i(1)+&
                                                    tij_abc(2,3,3)*dp_i(2)+&
                                                    tij_abc(3,3,3)*dp_i(3))
                      END IF
                   END IF
                END IF
                IF (task(2,1)) THEN
                   ! Dipole - Charge
                   tmp=   ch_j*(tij_a(1)*dp_i(1)+&
                                tij_a(2)*dp_i(2)+&
                                tij_a(3)*dp_i(3))&
                        - ch_i*(tij_a(1)*dp_j(1)+&
                                tij_a(2)*dp_j(2)+&
                                tij_a(3)*dp_j(3))
                   tmp=  tmp- ch_j*(damptij_a(1)*dp_i(1)+&
                                damptij_a(2)*dp_i(2)+&
                                damptij_a(3)*dp_i(3))&
                        + ch_i*(damptji_a(1)*dp_j(1)+&
                                damptji_a(2)*dp_j(2)+&
                                damptji_a(3)*dp_j(3))
                   eloc = eloc + tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         fr(k) = fr(k) -  ch_j *(tij_ab(1,k)*dp_i(1)+&
                                                 tij_ab(2,k)*dp_i(2)+&
                                                 tij_ab(3,k)*dp_i(3))&
                                       +  ch_i *(tij_ab(1,k)*dp_j(1)+&
                                                 tij_ab(2,k)*dp_j(2)+&
                                                 tij_ab(3,k)*dp_j(3))
                         fr(k) = fr(k) +  ch_j *(damptij_ab(1,k)*dp_i(1)+&
                                                 damptij_ab(2,k)*dp_i(2)+&
                                                 damptij_ab(3,k)*dp_i(3))&
                                       -  ch_i *(damptji_ab(1,k)*dp_j(1)+&
                                                 damptji_ab(2,k)*dp_j(2)+&
                                                 damptji_ab(3,k)*dp_j(3))
                      END DO
                   END IF
                END IF
                IF (task(3,3)) THEN
                   ! Quadrupole - Quadrupole
                   fac  = 1.0_dp/9.0_dp
                   tmp11 = qp_i(1,1)*(tij_abcd(1,1,1,1)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,1)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,1)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,1)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,1)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,1)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,1)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,1)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,1)*qp_j(3,3))
                   tmp21 = qp_i(2,1)*(tij_abcd(1,1,1,2)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,2)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,2)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,2)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,2)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,2)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,2)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,2)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,2)*qp_j(3,3))
                   tmp31 = qp_i(3,1)*(tij_abcd(1,1,1,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,3)*qp_j(3,3))
                   tmp22 = qp_i(2,2)*(tij_abcd(1,1,2,2)*qp_j(1,1)+&
                                      tij_abcd(2,1,2,2)*qp_j(2,1)+&
                                      tij_abcd(3,1,2,2)*qp_j(3,1)+&
                                      tij_abcd(1,2,2,2)*qp_j(1,2)+&
                                      tij_abcd(2,2,2,2)*qp_j(2,2)+&
                                      tij_abcd(3,2,2,2)*qp_j(3,2)+&
                                      tij_abcd(1,3,2,2)*qp_j(1,3)+&
                                      tij_abcd(2,3,2,2)*qp_j(2,3)+&
                                      tij_abcd(3,3,2,2)*qp_j(3,3))
                   tmp32 = qp_i(3,2)*(tij_abcd(1,1,2,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,2,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,2,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,2,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,2,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,2,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,2,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,2,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,2,3)*qp_j(3,3))
                   tmp33 = qp_i(3,3)*(tij_abcd(1,1,3,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,3,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,3,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,3,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,3,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,3,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,3,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,3,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,3,3)*qp_j(3,3))
                   tmp12 = tmp21
                   tmp13 = tmp31
                   tmp23 = tmp32
                   tmp   = tmp11 + tmp12 + tmp13 + &
                           tmp21 + tmp22 + tmp23 + &
                           tmp31 + tmp32 + tmp33

                   eloc = eloc + fac*tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         tmp11 = qp_i(1,1)*(tij_abcde(1,1,1,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,1,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,1,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,1,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,1,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,1,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,1,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,1,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,1,1,k)*qp_j(3,3))
                         tmp21 = qp_i(2,1)*(tij_abcde(1,1,2,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,2,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,2,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,2,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,2,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,2,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,2,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,2,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,2,1,k)*qp_j(3,3))
                         tmp31 = qp_i(3,1)*(tij_abcde(1,1,3,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,1,k)*qp_j(3,3))
                         tmp22 = qp_i(2,2)*(tij_abcde(1,1,2,2,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,2,2,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,2,2,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,2,2,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,2,2,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,2,2,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,2,2,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,2,2,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,2,2,k)*qp_j(3,3))
                         tmp32 = qp_i(3,2)*(tij_abcde(1,1,3,2,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,2,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,2,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,2,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,2,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,2,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,2,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,2,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,2,k)*qp_j(3,3))
                         tmp33 = qp_i(3,3)*(tij_abcde(1,1,3,3,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,3,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,3,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,3,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,3,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,3,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,3,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,3,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,3,k)*qp_j(3,3))
                         tmp12 = tmp21
                         tmp13 = tmp31
                         tmp23 = tmp32
                         fr(k) = fr(k) - fac * ( tmp11 + tmp12 + tmp13 +&
                                                 tmp21 + tmp22 + tmp23 +&
                                                 tmp31 + tmp32 + tmp33  )
                      END DO
                   END IF
                   ! Electric field
                   IF (do_efield) THEN
                      fac = 1.0_dp/3.0_dp
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i + fac*(tij_ab(1,1)*qp_j(1,1)+&
                                              tij_ab(2,1)*qp_j(2,1)+&
                                              tij_ab(3,1)*qp_j(3,1)+&
                                              tij_ab(1,2)*qp_j(1,2)+&
                                              tij_ab(2,2)*qp_j(2,2)+&
                                              tij_ab(3,2)*qp_j(3,2)+&
                                              tij_ab(1,3)*qp_j(1,3)+&
                                              tij_ab(2,3)*qp_j(2,3)+&
                                              tij_ab(3,3)*qp_j(3,3))

                         ef0_j = ef0_j + fac*(tij_ab(1,1)*qp_i(1,1)+&
                                              tij_ab(2,1)*qp_i(2,1)+&
                                              tij_ab(3,1)*qp_i(3,1)+&
                                              tij_ab(1,2)*qp_i(1,2)+&
                                              tij_ab(2,2)*qp_i(2,2)+&
                                              tij_ab(3,2)*qp_i(3,2)+&
                                              tij_ab(1,3)*qp_i(1,3)+&
                                              tij_ab(2,3)*qp_i(2,3)+&
                                              tij_ab(3,3)*qp_i(3,3))
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) - fac*(tij_abc(1,1,1)*qp_j(1,1)+&
                                                    tij_abc(2,1,1)*qp_j(2,1)+&
                                                    tij_abc(3,1,1)*qp_j(3,1)+&
                                                    tij_abc(1,2,1)*qp_j(1,2)+&
                                                    tij_abc(2,2,1)*qp_j(2,2)+&
                                                    tij_abc(3,2,1)*qp_j(3,2)+&
                                                    tij_abc(1,3,1)*qp_j(1,3)+&
                                                    tij_abc(2,3,1)*qp_j(2,3)+&
                                                    tij_abc(3,3,1)*qp_j(3,3))
                         ef1_i(2) = ef1_i(2) - fac*(tij_abc(1,1,2)*qp_j(1,1)+&
                                                    tij_abc(2,1,2)*qp_j(2,1)+&
                                                    tij_abc(3,1,2)*qp_j(3,1)+&
                                                    tij_abc(1,2,2)*qp_j(1,2)+&
                                                    tij_abc(2,2,2)*qp_j(2,2)+&
                                                    tij_abc(3,2,2)*qp_j(3,2)+&
                                                    tij_abc(1,3,2)*qp_j(1,3)+&
                                                    tij_abc(2,3,2)*qp_j(2,3)+&
                                                    tij_abc(3,3,2)*qp_j(3,3))
                         ef1_i(3) = ef1_i(3) - fac*(tij_abc(1,1,3)*qp_j(1,1)+&
                                                    tij_abc(2,1,3)*qp_j(2,1)+&
                                                    tij_abc(3,1,3)*qp_j(3,1)+&
                                                    tij_abc(1,2,3)*qp_j(1,2)+&
                                                    tij_abc(2,2,3)*qp_j(2,2)+&
                                                    tij_abc(3,2,3)*qp_j(3,2)+&
                                                    tij_abc(1,3,3)*qp_j(1,3)+&
                                                    tij_abc(2,3,3)*qp_j(2,3)+&
                                                    tij_abc(3,3,3)*qp_j(3,3))

                         ef1_j(1) = ef1_j(1) + fac*(tij_abc(1,1,1)*qp_i(1,1)+&
                                                    tij_abc(2,1,1)*qp_i(2,1)+&
                                                    tij_abc(3,1,1)*qp_i(3,1)+&
                                                    tij_abc(1,2,1)*qp_i(1,2)+&
                                                    tij_abc(2,2,1)*qp_i(2,2)+&
                                                    tij_abc(3,2,1)*qp_i(3,2)+&
                                                    tij_abc(1,3,1)*qp_i(1,3)+&
                                                    tij_abc(2,3,1)*qp_i(2,3)+&
                                                    tij_abc(3,3,1)*qp_i(3,3))
                         ef1_j(2) = ef1_j(2) + fac*(tij_abc(1,1,2)*qp_i(1,1)+&
                                                    tij_abc(2,1,2)*qp_i(2,1)+&
                                                    tij_abc(3,1,2)*qp_i(3,1)+&
                                                    tij_abc(1,2,2)*qp_i(1,2)+&
                                                    tij_abc(2,2,2)*qp_i(2,2)+&
                                                    tij_abc(3,2,2)*qp_i(3,2)+&
                                                    tij_abc(1,3,2)*qp_i(1,3)+&
                                                    tij_abc(2,3,2)*qp_i(2,3)+&
                                                    tij_abc(3,3,2)*qp_i(3,3))
                         ef1_j(3) = ef1_j(3) + fac*(tij_abc(1,1,3)*qp_i(1,1)+&
                                                    tij_abc(2,1,3)*qp_i(2,1)+&
                                                    tij_abc(3,1,3)*qp_i(3,1)+&
                                                    tij_abc(1,2,3)*qp_i(1,2)+&
                                                    tij_abc(2,2,3)*qp_i(2,2)+&
                                                    tij_abc(3,2,3)*qp_i(3,2)+&
                                                    tij_abc(1,3,3)*qp_i(1,3)+&
                                                    tij_abc(2,3,3)*qp_i(2,3)+&
                                                    tij_abc(3,3,3)*qp_i(3,3))
                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         tmp11 =   fac *(tij_abcd(1,1,1,1)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,1)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,1)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,1)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,1)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,1)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,1)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,1)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,1)*qp_j(3,3))
                         tmp12 =   fac *(tij_abcd(1,1,1,2)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,2)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,2)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,2)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,2)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,2)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,2)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,2)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,2)*qp_j(3,3))
                         tmp13 =   fac *(tij_abcd(1,1,1,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,3)*qp_j(3,3))
                         tmp22 =   fac *(tij_abcd(1,1,2,2)*qp_j(1,1)+&
                                         tij_abcd(2,1,2,2)*qp_j(2,1)+&
                                         tij_abcd(3,1,2,2)*qp_j(3,1)+&
                                         tij_abcd(1,2,2,2)*qp_j(1,2)+&
                                         tij_abcd(2,2,2,2)*qp_j(2,2)+&
                                         tij_abcd(3,2,2,2)*qp_j(3,2)+&
                                         tij_abcd(1,3,2,2)*qp_j(1,3)+&
                                         tij_abcd(2,3,2,2)*qp_j(2,3)+&
                                         tij_abcd(3,3,2,2)*qp_j(3,3))
                         tmp23 =   fac *(tij_abcd(1,1,2,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,2,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,2,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,2,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,2,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,2,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,2,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,2,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,2,3)*qp_j(3,3))
                         tmp33 =   fac *(tij_abcd(1,1,3,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,3,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,3,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,3,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,3,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,3,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,3,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,3,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,3,3)*qp_j(3,3))

                         ef2_i(1,1) = ef2_i(1,1) - tmp11
                         ef2_i(1,2) = ef2_i(1,2) - tmp12
                         ef2_i(1,3) = ef2_i(1,3) - tmp13
                         ef2_i(2,1) = ef2_i(2,1) - tmp12
                         ef2_i(2,2) = ef2_i(2,2) - tmp22
                         ef2_i(2,3) = ef2_i(2,3) - tmp23
                         ef2_i(3,1) = ef2_i(3,1) - tmp13
                         ef2_i(3,2) = ef2_i(3,2) - tmp23
                         ef2_i(3,3) = ef2_i(3,3) - tmp33

                         tmp11 =   fac *(tij_abcd(1,1,1,1)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,1)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,1)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,1)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,1)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,1)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,1)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,1)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,1)*qp_i(3,3))
                         tmp12 =   fac *(tij_abcd(1,1,1,2)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,2)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,2)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,2)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,2)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,2)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,2)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,2)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,2)*qp_i(3,3))
                         tmp13 =   fac *(tij_abcd(1,1,1,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,3)*qp_i(3,3))
                         tmp22 =   fac *(tij_abcd(1,1,2,2)*qp_i(1,1)+&
                                         tij_abcd(2,1,2,2)*qp_i(2,1)+&
                                         tij_abcd(3,1,2,2)*qp_i(3,1)+&
                                         tij_abcd(1,2,2,2)*qp_i(1,2)+&
                                         tij_abcd(2,2,2,2)*qp_i(2,2)+&
                                         tij_abcd(3,2,2,2)*qp_i(3,2)+&
                                         tij_abcd(1,3,2,2)*qp_i(1,3)+&
                                         tij_abcd(2,3,2,2)*qp_i(2,3)+&
                                         tij_abcd(3,3,2,2)*qp_i(3,3))
                         tmp23 =   fac *(tij_abcd(1,1,2,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,2,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,2,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,2,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,2,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,2,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,2,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,2,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,2,3)*qp_i(3,3))
                         tmp33 =   fac *(tij_abcd(1,1,3,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,3,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,3,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,3,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,3,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,3,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,3,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,3,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,3,3)*qp_i(3,3))

                         ef2_j(1,1) = ef2_j(1,1) - tmp11
                         ef2_j(1,2) = ef2_j(1,2) - tmp12
                         ef2_j(1,3) = ef2_j(1,3) - tmp13
                         ef2_j(2,1) = ef2_j(2,1) - tmp12
                         ef2_j(2,2) = ef2_j(2,2) - tmp22
                         ef2_j(2,3) = ef2_j(2,3) - tmp23
                         ef2_j(3,1) = ef2_j(3,1) - tmp13
                         ef2_j(3,2) = ef2_j(3,2) - tmp23
                         ef2_j(3,3) = ef2_j(3,3) - tmp33
                      END IF
                   END IF
                END IF
                IF (task(3,2)) THEN
                   ! Quadrupole - Dipole
                   fac = 1.0_dp/3.0_dp
                   ! Dipole i (locally B) - Quadrupole j (locally A)
                   tmp_ij = dp_i(1)*(tij_abc(1,1,1)*qp_j(1,1)+&
                                     tij_abc(2,1,1)*qp_j(2,1)+&
                                     tij_abc(3,1,1)*qp_j(3,1)+&
                                     tij_abc(1,2,1)*qp_j(1,2)+&
                                     tij_abc(2,2,1)*qp_j(2,2)+&
                                     tij_abc(3,2,1)*qp_j(3,2)+&
                                     tij_abc(1,3,1)*qp_j(1,3)+&
                                     tij_abc(2,3,1)*qp_j(2,3)+&
                                     tij_abc(3,3,1)*qp_j(3,3))+&
                            dp_i(2)*(tij_abc(1,1,2)*qp_j(1,1)+&
                                     tij_abc(2,1,2)*qp_j(2,1)+&
                                     tij_abc(3,1,2)*qp_j(3,1)+&
                                     tij_abc(1,2,2)*qp_j(1,2)+&
                                     tij_abc(2,2,2)*qp_j(2,2)+&
                                     tij_abc(3,2,2)*qp_j(3,2)+&
                                     tij_abc(1,3,2)*qp_j(1,3)+&
                                     tij_abc(2,3,2)*qp_j(2,3)+&
                                     tij_abc(3,3,2)*qp_j(3,3))+&
                            dp_i(3)*(tij_abc(1,1,3)*qp_j(1,1)+&
                                     tij_abc(2,1,3)*qp_j(2,1)+&
                                     tij_abc(3,1,3)*qp_j(3,1)+&
                                     tij_abc(1,2,3)*qp_j(1,2)+&
                                     tij_abc(2,2,3)*qp_j(2,2)+&
                                     tij_abc(3,2,3)*qp_j(3,2)+&
                                     tij_abc(1,3,3)*qp_j(1,3)+&
                                     tij_abc(2,3,3)*qp_j(2,3)+&
                                     tij_abc(3,3,3)*qp_j(3,3))

                   ! Dipole j (locally A) - Quadrupole i (locally B)
                   tmp_ji = dp_j(1)*(tij_abc(1,1,1)*qp_i(1,1)+&
                                     tij_abc(2,1,1)*qp_i(2,1)+&
                                     tij_abc(3,1,1)*qp_i(3,1)+&
                                     tij_abc(1,2,1)*qp_i(1,2)+&
                                     tij_abc(2,2,1)*qp_i(2,2)+&
                                     tij_abc(3,2,1)*qp_i(3,2)+&
                                     tij_abc(1,3,1)*qp_i(1,3)+&
                                     tij_abc(2,3,1)*qp_i(2,3)+&
                                     tij_abc(3,3,1)*qp_i(3,3))+&
                            dp_j(2)*(tij_abc(1,1,2)*qp_i(1,1)+&
                                     tij_abc(2,1,2)*qp_i(2,1)+&
                                     tij_abc(3,1,2)*qp_i(3,1)+&
                                     tij_abc(1,2,2)*qp_i(1,2)+&
                                     tij_abc(2,2,2)*qp_i(2,2)+&
                                     tij_abc(3,2,2)*qp_i(3,2)+&
                                     tij_abc(1,3,2)*qp_i(1,3)+&
                                     tij_abc(2,3,2)*qp_i(2,3)+&
                                     tij_abc(3,3,2)*qp_i(3,3))+&
                            dp_j(3)*(tij_abc(1,1,3)*qp_i(1,1)+&
                                     tij_abc(2,1,3)*qp_i(2,1)+&
                                     tij_abc(3,1,3)*qp_i(3,1)+&
                                     tij_abc(1,2,3)*qp_i(1,2)+&
                                     tij_abc(2,2,3)*qp_i(2,2)+&
                                     tij_abc(3,2,3)*qp_i(3,2)+&
                                     tij_abc(1,3,3)*qp_i(1,3)+&
                                     tij_abc(2,3,3)*qp_i(2,3)+&
                                     tij_abc(3,3,3)*qp_i(3,3))

                   tmp= fac * (tmp_ij - tmp_ji)
                   eloc = eloc + tmp
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         ! Dipole i (locally B) - Quadrupole j (locally A)
                         tmp_ij = dp_i(1)*(tij_abcd(1,1,1,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,1,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,1,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,1,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,1,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,1,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,1,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,1,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,1,k)*qp_j(3,3))+&
                                  dp_i(2)*(tij_abcd(1,1,2,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,2,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,2,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,2,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,2,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,2,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,2,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,2,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,2,k)*qp_j(3,3))+&
                                  dp_i(3)*(tij_abcd(1,1,3,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,3,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,3,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,3,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,3,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,3,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,3,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,3,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,3,k)*qp_j(3,3))

                         ! Dipole j (locally A) - Quadrupole i (locally B)
                         tmp_ji = dp_j(1)*(tij_abcd(1,1,1,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,1,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,1,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,1,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,1,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,1,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,1,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,1,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,1,k)*qp_i(3,3))+&
                                  dp_j(2)*(tij_abcd(1,1,2,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,2,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,2,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,2,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,2,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,2,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,2,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,2,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,2,k)*qp_i(3,3))+&
                                  dp_j(3)*(tij_abcd(1,1,3,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,3,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,3,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,3,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,3,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,3,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,3,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,3,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,3,k)*qp_i(3,3))

                         fr(k) = fr(k) - fac * (tmp_ij - tmp_ji)
                      END DO
                   END IF
                END IF
                IF (task(3,1)) THEN
                   ! Quadrupole - Charge
                   fac = 1.0_dp/3.0_dp

                   ! Quadrupole j (locally A) - Charge j (locally B)
                   tmp_ij = ch_i * (tij_ab(1,1)*qp_j(1,1)+&
                                    tij_ab(2,1)*qp_j(2,1)+&
                                    tij_ab(3,1)*qp_j(3,1)+&
                                    tij_ab(1,2)*qp_j(1,2)+&
                                    tij_ab(2,2)*qp_j(2,2)+&
                                    tij_ab(3,2)*qp_j(3,2)+&
                                    tij_ab(1,3)*qp_j(1,3)+&
                                    tij_ab(2,3)*qp_j(2,3)+&
                                    tij_ab(3,3)*qp_j(3,3))

                   ! Quadrupole i (locally B) - Charge j (locally A)
                   tmp_ji = ch_j * (tij_ab(1,1)*qp_i(1,1)+&
                                    tij_ab(2,1)*qp_i(2,1)+&
                                    tij_ab(3,1)*qp_i(3,1)+&
                                    tij_ab(1,2)*qp_i(1,2)+&
                                    tij_ab(2,2)*qp_i(2,2)+&
                                    tij_ab(3,2)*qp_i(3,2)+&
                                    tij_ab(1,3)*qp_i(1,3)+&
                                    tij_ab(2,3)*qp_i(2,3)+&
                                    tij_ab(3,3)*qp_i(3,3))

                   eloc = eloc + fac*(tmp_ij+tmp_ji)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         ! Quadrupole j (locally A) - Charge i (locally B)
                         tmp_ij = ch_i * (tij_abc(1,1,k)*qp_j(1,1)+&
                                          tij_abc(2,1,k)*qp_j(2,1)+&
                                          tij_abc(3,1,k)*qp_j(3,1)+&
                                          tij_abc(1,2,k)*qp_j(1,2)+&
                                          tij_abc(2,2,k)*qp_j(2,2)+&
                                          tij_abc(3,2,k)*qp_j(3,2)+&
                                          tij_abc(1,3,k)*qp_j(1,3)+&
                                          tij_abc(2,3,k)*qp_j(2,3)+&
                                          tij_abc(3,3,k)*qp_j(3,3))

                         ! Quadrupole i (locally B) - Charge j (locally A)
                         tmp_ji = ch_j * (tij_abc(1,1,k)*qp_i(1,1)+&
                                          tij_abc(2,1,k)*qp_i(2,1)+&
                                          tij_abc(3,1,k)*qp_i(3,1)+&
                                          tij_abc(1,2,k)*qp_i(1,2)+&
                                          tij_abc(2,2,k)*qp_i(2,2)+&
                                          tij_abc(3,2,k)*qp_i(3,2)+&
                                          tij_abc(1,3,k)*qp_i(1,3)+&
                                          tij_abc(2,3,k)*qp_i(2,3)+&
                                          tij_abc(3,3,k)*qp_i(3,3))

                         fr(k) = fr(k) - fac *(tmp_ij + tmp_ji)
                      END DO
                   END IF
                END IF
                energy = energy + eloc
                IF (do_forces) THEN
                   forces(1,atom_a) = forces(1,atom_a) - fr(1)
                   forces(2,atom_a) = forces(2,atom_a) - fr(2)
                   forces(3,atom_a) = forces(3,atom_a) - fr(3)
                   forces(1,atom_b) = forces(1,atom_b) + fr(1)
                   forces(2,atom_b) = forces(2,atom_b) + fr(2)
                   forces(3,atom_b) = forces(3,atom_b) + fr(3)
                END IF
                ! Electric fields
                IF (do_efield) THEN
                   ! Potential
                   IF (do_efield0) THEN
                      efield0(  atom_a) = efield0(  atom_a) + ef0_j

                      efield0(  atom_b) = efield0(  atom_b) + ef0_i
                   END IF
                   ! Electric field
                   IF (do_efield1) THEN
                      efield1(1,atom_a) = efield1(1,atom_a) + ef1_j(1)
                      efield1(2,atom_a) = efield1(2,atom_a) + ef1_j(2)
                      efield1(3,atom_a) = efield1(3,atom_a) + ef1_j(3)

                      efield1(1,atom_b) = efield1(1,atom_b) + ef1_i(1)
                      efield1(2,atom_b) = efield1(2,atom_b) + ef1_i(2)
                      efield1(3,atom_b) = efield1(3,atom_b) + ef1_i(3)
                   END IF
                   ! Electric field gradient
                   IF (do_efield2) THEN
                      efield2(1,atom_a) = efield2(1,atom_a) + ef2_j(1,1)
                      efield2(2,atom_a) = efield2(2,atom_a) + ef2_j(1,2)
                      efield2(3,atom_a) = efield2(3,atom_a) + ef2_j(1,3)
                      efield2(4,atom_a) = efield2(4,atom_a) + ef2_j(2,1)
                      efield2(5,atom_a) = efield2(5,atom_a) + ef2_j(2,2)
                      efield2(6,atom_a) = efield2(6,atom_a) + ef2_j(2,3)
                      efield2(7,atom_a) = efield2(7,atom_a) + ef2_j(3,1)
                      efield2(8,atom_a) = efield2(8,atom_a) + ef2_j(3,2)
                      efield2(9,atom_a) = efield2(9,atom_a) + ef2_j(3,3)

                      efield2(1,atom_b) = efield2(1,atom_b) + ef2_i(1,1)
                      efield2(2,atom_b) = efield2(2,atom_b) + ef2_i(1,2)
                      efield2(3,atom_b) = efield2(3,atom_b) + ef2_i(1,3)
                      efield2(4,atom_b) = efield2(4,atom_b) + ef2_i(2,1)
                      efield2(5,atom_b) = efield2(5,atom_b) + ef2_i(2,2)
                      efield2(6,atom_b) = efield2(6,atom_b) + ef2_i(2,3)
                      efield2(7,atom_b) = efield2(7,atom_b) + ef2_i(3,1)
                      efield2(8,atom_b) = efield2(8,atom_b) + ef2_i(3,2)
                      efield2(9,atom_b) = efield2(9,atom_b) + ef2_i(3,3)
                   END IF
                END IF
                IF (do_stress) THEN
                   ptens11 = ptens11 + rab(1) * fr(1)
                   ptens21 = ptens21 + rab(2) * fr(1)
                   ptens31 = ptens31 + rab(3) * fr(1)
                   ptens12 = ptens12 + rab(1) * fr(2)
                   ptens22 = ptens22 + rab(2) * fr(2)
                   ptens32 = ptens32 + rab(3) * fr(2)
                   ptens13 = ptens13 + rab(1) * fr(3)
                   ptens23 = ptens23 + rab(2) * fr(3)
                   ptens33 = ptens33 + rab(3) * fr(3)
                END IF
                  END IF
               END IF
            END DO Pairs
         END DO Kind_Group_Loop
      END DO Lists
      IF (do_stress) THEN
         pv(1, 1) = pv(1, 1)+ptens11
         pv(1, 2) = pv(1, 2)+(ptens12+ptens21)*0.5_dp
         pv(1, 3) = pv(1, 3)+(ptens13+ptens31)*0.5_dp
         pv(2, 1) = pv(1, 2)
         pv(2, 2) = pv(2, 2)+ptens22
         pv(2, 3) = pv(2, 3)+(ptens23+ptens32)*0.5_dp
         pv(3, 1) = pv(1, 3)
         pv(3, 2) = pv(2, 3)
         pv(3, 3) = pv(3, 3)+ptens33
      END IF

      CALL timestop(handle)
   END SUBROUTINE ewald_multipole_SR

! **************************************************************************************************
!> \brief computes the bonded correction for the potential and the force for a
!>        lattice sum of multipoles up to quadrupole
!> \param nonbond_env ...
!> \param particle_set ...
!> \param ewald_env ...
!> \param cell ...
!> \param energy ...
!> \param task ...
!> \param do_forces ...
!> \param do_efield ...
!> \param do_stress ...
!> \param charges ...
!> \param dipoles ...
!> \param quadrupoles ...
!> \param forces ...
!> \param pv ...
!> \param efield0 ...
!> \param efield1 ...
!> \param efield2 ...
!> \author Teodoro Laino [tlaino] - 05.2009
! **************************************************************************************************
   SUBROUTINE ewald_multipole_bonded(nonbond_env, particle_set, ewald_env, &
                                     cell, energy, task, do_forces, do_efield, do_stress, charges, &
                                     dipoles, quadrupoles, forces, pv, efield0, efield1, efield2)

      TYPE(fist_nonbond_env_type), POINTER               :: nonbond_env
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      LOGICAL, DIMENSION(3, 3), INTENT(IN)               :: task
      LOGICAL, INTENT(IN)                                :: do_forces, do_efield, do_stress
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: dipoles
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: quadrupoles
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: forces, pv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: efield0
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: efield1, efield2

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_multipole_bonded', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: a, atom_a, atom_b, b, c, d, e, handle, &
                                                            i, iend, igrp, ilist, ipair, istart, &
                                                            k, nscale
      INTEGER, DIMENSION(:, :), POINTER                  :: list
      LOGICAL                                            :: do_efield0, do_efield1, do_efield2, &
                                                            force_eval
      REAL(KIND=dp) :: alpha, ch_i, ch_j, ef0_i, ef0_j, eloc, fac, fac_ij, ir, irab2, ptens11, &
         ptens12, ptens13, ptens21, ptens22, ptens23, ptens31, ptens32, ptens33, r, rab2, tij, &
         tmp, tmp1, tmp11, tmp12, tmp13, tmp2, tmp21, tmp22, tmp23, tmp31, tmp32, tmp33, tmp_ij, &
         tmp_ji
      REAL(KIND=dp), DIMENSION(0:5)                      :: f
      REAL(KIND=dp), DIMENSION(3)                        :: dp_i, dp_j, ef1_i, ef1_j, fr, rab, tij_a
      REAL(KIND=dp), DIMENSION(3, 3)                     :: ef2_i, ef2_j, qp_i, qp_j, tij_ab
      REAL(KIND=dp), DIMENSION(3, 3, 3)                  :: tij_abc
      REAL(KIND=dp), DIMENSION(3, 3, 3, 3)               :: tij_abcd
      REAL(KIND=dp), DIMENSION(3, 3, 3, 3, 3)            :: tij_abcde
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(neighbor_kind_pairs_type), POINTER            :: neighbor_kind_pair

      CALL timeset(routineN, handle)
      do_efield0 = do_efield .AND. ASSOCIATED(efield0)
      do_efield1 = do_efield .AND. ASSOCIATED(efield1)
      do_efield2 = do_efield .AND. ASSOCIATED(efield2)
      IF (do_stress) THEN
         ptens11 = 0.0_dp; ptens12 = 0.0_dp; ptens13 = 0.0_dp
         ptens21 = 0.0_dp; ptens22 = 0.0_dp; ptens23 = 0.0_dp
         ptens31 = 0.0_dp; ptens32 = 0.0_dp; ptens33 = 0.0_dp
      END IF
      CALL ewald_env_get(ewald_env, alpha=alpha)
      CALL fist_nonbond_env_get(nonbond_env, nonbonded=nonbonded)

      ! Starting the force loop
      Lists: DO ilist = 1, nonbonded%nlists
         neighbor_kind_pair => nonbonded%neighbor_kind_pairs(ilist)
         nscale = neighbor_kind_pair%nscale
         IF (nscale == 0) CYCLE
         list => neighbor_kind_pair%list
         Kind_Group_Loop: DO igrp = 1, neighbor_kind_pair%ngrp_kind
            istart = neighbor_kind_pair%grp_kind_start(igrp)
            IF (istart > nscale) CYCLE
            iend = MIN(neighbor_kind_pair%grp_kind_end(igrp), nscale)
            Pairs: DO ipair = istart, iend
               ! only use pairs that are (partially) excluded for electrostatics
               fac_ij = -1.0_dp+neighbor_kind_pair%ei_scale(ipair)
               IF (fac_ij >= 0) CYCLE

               atom_a = list(1, ipair)
               atom_b = list(2, ipair)

               rab = particle_set(atom_b)%r-particle_set(atom_a)%r
               rab = pbc(rab, cell)
               rab2 = rab(1)**2+rab(2)**2+rab(3)**2
                ! Compute the Short Range constribution according the task
                IF (debug_this_module) THEN
                   f         = HUGE(0.0_dp)
                   tij       = HUGE(0.0_dp)
                   tij_a     = HUGE(0.0_dp)
                   tij_ab    = HUGE(0.0_dp)
                   tij_abc   = HUGE(0.0_dp)
                   tij_abcd  = HUGE(0.0_dp)
                   tij_abcde = HUGE(0.0_dp)
                END IF
                r     = SQRT(rab2)
                irab2 = 1.0_dp/rab2
                ir    = 1.0_dp/r

                ! Compute the radial function
                IF (debug_this_module.AND.debug_r_space.AND.(.NOT.debug_g_space)) THEN
                   f(0)  = ir
                   tmp   = 0.0_dp
                ELSE
                   f(0)  = erf(alpha*r)*ir
                   tmp   = EXP(-alpha**2*rab2)*oorootpi
                END IF
                fac = 1.0_dp
                DO i = 1, 5
                   fac  = fac*REAL(2*i-1,KIND=dp)
                   f(i) = irab2*(f(i-1) - tmp*((2.0_dp*alpha**2)**i)/(fac*alpha))
                END DO
                ! Compute the Tensor components
                force_eval = do_stress
                IF (task(1,1)) THEN
                   tij         = f(0)*fac_ij
                                                 force_eval = do_forces .OR.do_efield1
                END IF
                IF (task(2,2))                   force_eval = force_eval.OR.do_efield0
                IF (task(1,2).OR.force_eval) THEN
                   force_eval = do_stress
                   tij_a    = - rab*f(1)*fac_ij
                   IF (task(1,2))                force_eval = force_eval.OR. do_forces
                END IF
                IF (task(1,1))                   force_eval = force_eval.OR.do_efield2
                IF (task(3,3))                   force_eval = force_eval.OR.do_efield0
                IF (task(2,2).OR.task(3,1).OR.force_eval) THEN
                   force_eval = do_stress
                   DO b = 1,3
                      DO a = 1,3
                         tmp = rab(a)*rab(b)*fac_ij
                         tij_ab(a,b) = 3.0_dp*tmp*f(2)
                         IF (a==b) tij_ab(a,b) = tij_ab(a,b) - f(1)*fac_ij
                      END DO
                   END DO
                   IF (task(2,2).OR.task(3,1))   force_eval = force_eval.OR. do_forces
                END IF
                IF (task(2,2))                   force_eval = force_eval.OR.do_efield2
                IF (task(3,3))                   force_eval = force_eval.OR.do_efield1
                IF (task(3,2).OR.force_eval) THEN
                   force_eval = do_stress
                   DO c = 1, 3
                      DO b = 1, 3
                         DO a = 1, 3
                            tmp = rab(a)*rab(b)*rab(c)*fac_ij
                            tij_abc(a,b,c) = - 15.0_dp*tmp*f(3)
                            tmp = 3.0_dp*f(2)*fac_ij
                            IF (a==b) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(c)
                            IF (a==c) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(b)
                            IF (b==c) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(a)
                         END DO
                      END DO
                   END DO
                   IF (task(3,2))                force_eval = force_eval.OR. do_forces
                END IF
                IF (task(3,3).OR.force_eval) THEN
                   force_eval = do_stress
                   DO d = 1, 3
                      DO c = 1, 3
                         DO b = 1, 3
                            DO a = 1, 3
                               tmp = rab(a)*rab(b)*rab(c)*rab(d)*fac_ij
                               tij_abcd(a,b,c,d) = 105.0_dp*tmp*f(4)
                               tmp1 = 15.0_dp*f(3)*fac_ij
                               tmp2 =  3.0_dp*f(2)*fac_ij
                               IF (a==b) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(c)*rab(d)
                                  IF (c==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (a==c) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(b)*rab(d)
                                  IF (b==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (a==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(b)*rab(c)
                               IF (b==c) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(d)
                                  IF (a==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (b==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(c)
                               IF (c==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(b)
                            END DO
                         END DO
                      END DO
                   END DO
                   IF (task(3,3))                force_eval = force_eval.OR. do_forces
                END IF
                IF (force_eval) THEN
                   force_eval = do_stress
                   DO e = 1, 3
                      DO d = 1, 3
                         DO c = 1, 3
                            DO b = 1, 3
                               DO a = 1, 3
                                  tmp = rab(a)*rab(b)*rab(c)*rab(d)*rab(e)*fac_ij
                                  tij_abcde(a,b,c,d,e) = -945.0_dp*tmp*f(5)
                                  tmp1 = 105.0_dp*f(4)*fac_ij
                                  tmp2 =  15.0_dp*f(3)*fac_ij
                                  IF (a==b) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(c)*rab(d)*rab(e)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                  END IF
                                  IF (a==c) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(d)*rab(e)
                                     IF (b==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (b==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (a==d) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(c)*rab(e)
                                     IF (b==c) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (b==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (a==e) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(c)*rab(d)
                                     IF (b==c) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (b==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (b==c) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(d)*rab(e)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (b==d) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(c)*rab(e)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (b==e) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(c)*rab(d)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (c==d)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(e)
                                  IF (c==e)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(d)
                                  IF (d==e)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(c)
                               END DO
                            END DO
                         END DO
                      END DO
                   END DO
                END IF
                eloc  = 0.0_dp
                fr    = 0.0_dp
                ef0_i = 0.0_dp
                ef0_j = 0.0_dp
                ef1_j = 0.0_dp
                ef1_i = 0.0_dp
                ef2_j = 0.0_dp
                ef2_i = 0.0_dp


                ! Initialize the charge, dipole and quadrupole for atom A and B
                IF (debug_this_module) THEN
                   ch_j  = HUGE(0.0_dp)
                   ch_i  = HUGE(0.0_dp)
                   dp_j  = HUGE(0.0_dp)
                   dp_i  = HUGE(0.0_dp)
                   qp_j  = HUGE(0.0_dp)
                   qp_i  = HUGE(0.0_dp)
                END IF
                IF (ANY(task(1,:))) THEN
                   ch_j  = charges(atom_a)
                   ch_i  = charges(atom_b)
                END IF
                IF (ANY(task(2,:))) THEN
                   dp_j  = dipoles(:,atom_a)
                   dp_i  = dipoles(:,atom_b)
                END IF
                IF (ANY(task(3,:))) THEN
                   qp_j  = quadrupoles(:,:,atom_a)
                   qp_i  = quadrupoles(:,:,atom_b)
                END IF
                IF (task(1,1)) THEN
                   ! Charge - Charge
                   eloc = eloc + ch_i*tij*ch_j
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      fr(1) = fr(1) - ch_j * tij_a(1) * ch_i
                      fr(2) = fr(2) - ch_j * tij_a(2) * ch_i
                      fr(3) = fr(3) - ch_j * tij_a(3) * ch_i
                   END IF
                   ! Electric fields
                   IF (do_efield) THEN
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i + tij * ch_j

                         ef0_j = ef0_j + tij * ch_i
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) - tij_a(1) * ch_j
                         ef1_i(2) = ef1_i(2) - tij_a(2) * ch_j
                         ef1_i(3) = ef1_i(3) - tij_a(3) * ch_j

                         ef1_j(1) = ef1_j(1) + tij_a(1) * ch_i
                         ef1_j(2) = ef1_j(2) + tij_a(2) * ch_i
                         ef1_j(3) = ef1_j(3) + tij_a(3) * ch_i


                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         ef2_i(1,1) = ef2_i(1,1) - tij_ab(1,1) * ch_j
                         ef2_i(2,1) = ef2_i(2,1) - tij_ab(2,1) * ch_j
                         ef2_i(3,1) = ef2_i(3,1) - tij_ab(3,1) * ch_j
                         ef2_i(1,2) = ef2_i(1,2) - tij_ab(1,2) * ch_j
                         ef2_i(2,2) = ef2_i(2,2) - tij_ab(2,2) * ch_j
                         ef2_i(3,2) = ef2_i(3,2) - tij_ab(3,2) * ch_j
                         ef2_i(1,3) = ef2_i(1,3) - tij_ab(1,3) * ch_j
                         ef2_i(2,3) = ef2_i(2,3) - tij_ab(2,3) * ch_j
                         ef2_i(3,3) = ef2_i(3,3) - tij_ab(3,3) * ch_j

                         ef2_j(1,1) = ef2_j(1,1) - tij_ab(1,1) * ch_i
                         ef2_j(2,1) = ef2_j(2,1) - tij_ab(2,1) * ch_i
                         ef2_j(3,1) = ef2_j(3,1) - tij_ab(3,1) * ch_i
                         ef2_j(1,2) = ef2_j(1,2) - tij_ab(1,2) * ch_i
                         ef2_j(2,2) = ef2_j(2,2) - tij_ab(2,2) * ch_i
                         ef2_j(3,2) = ef2_j(3,2) - tij_ab(3,2) * ch_i
                         ef2_j(1,3) = ef2_j(1,3) - tij_ab(1,3) * ch_i
                         ef2_j(2,3) = ef2_j(2,3) - tij_ab(2,3) * ch_i
                         ef2_j(3,3) = ef2_j(3,3) - tij_ab(3,3) * ch_i
                      END IF
                   END IF
                END IF
                IF (task(2,2)) THEN
                   ! Dipole - Dipole
                   tmp= - (dp_i(1)*(tij_ab(1,1)*dp_j(1)+&
                                    tij_ab(2,1)*dp_j(2)+&
                                    tij_ab(3,1)*dp_j(3))+&
                           dp_i(2)*(tij_ab(1,2)*dp_j(1)+&
                                    tij_ab(2,2)*dp_j(2)+&
                                    tij_ab(3,2)*dp_j(3))+&
                           dp_i(3)*(tij_ab(1,3)*dp_j(1)+&
                                    tij_ab(2,3)*dp_j(2)+&
                                    tij_ab(3,3)*dp_j(3)))
                   eloc = eloc + tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         fr(k) = fr(k) +  dp_i(1)*(tij_abc(1,1,k)*dp_j(1)+&
                                                   tij_abc(2,1,k)*dp_j(2)+&
                                                   tij_abc(3,1,k)*dp_j(3))&
                                       +  dp_i(2)*(tij_abc(1,2,k)*dp_j(1)+&
                                                   tij_abc(2,2,k)*dp_j(2)+&
                                                   tij_abc(3,2,k)*dp_j(3))&
                                       +  dp_i(3)*(tij_abc(1,3,k)*dp_j(1)+&
                                                   tij_abc(2,3,k)*dp_j(2)+&
                                                   tij_abc(3,3,k)*dp_j(3))
                      END DO
                   END IF
                   ! Electric fields
                   IF (do_efield) THEN
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i - (tij_a(1)*dp_j(1)+&
                                          tij_a(2)*dp_j(2)+&
                                          tij_a(3)*dp_j(3))

                         ef0_j = ef0_j + (tij_a(1)*dp_i(1)+&
                                          tij_a(2)*dp_i(2)+&
                                          tij_a(3)*dp_i(3))
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) + (tij_ab(1,1)*dp_j(1)+&
                                                tij_ab(2,1)*dp_j(2)+&
                                                tij_ab(3,1)*dp_j(3))
                         ef1_i(2) = ef1_i(2) + (tij_ab(1,2)*dp_j(1)+&
                                                tij_ab(2,2)*dp_j(2)+&
                                                tij_ab(3,2)*dp_j(3))
                         ef1_i(3) = ef1_i(3) + (tij_ab(1,3)*dp_j(1)+&
                                                tij_ab(2,3)*dp_j(2)+&
                                                tij_ab(3,3)*dp_j(3))

                         ef1_j(1) = ef1_j(1) + (tij_ab(1,1)*dp_i(1)+&
                                                tij_ab(2,1)*dp_i(2)+&
                                                tij_ab(3,1)*dp_i(3))
                         ef1_j(2) = ef1_j(2) + (tij_ab(1,2)*dp_i(1)+&
                                                tij_ab(2,2)*dp_i(2)+&
                                                tij_ab(3,2)*dp_i(3))
                         ef1_j(3) = ef1_j(3) + (tij_ab(1,3)*dp_i(1)+&
                                                tij_ab(2,3)*dp_i(2)+&
                                                tij_ab(3,3)*dp_i(3))
                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         ef2_i(1,1) = ef2_i(1,1) + (tij_abc(1,1,1)*dp_j(1)+&
                                                    tij_abc(2,1,1)*dp_j(2)+&
                                                    tij_abc(3,1,1)*dp_j(3))
                         ef2_i(1,2) = ef2_i(1,2) + (tij_abc(1,1,2)*dp_j(1)+&
                                                    tij_abc(2,1,2)*dp_j(2)+&
                                                    tij_abc(3,1,2)*dp_j(3))
                         ef2_i(1,3) = ef2_i(1,3) + (tij_abc(1,1,3)*dp_j(1)+&
                                                    tij_abc(2,1,3)*dp_j(2)+&
                                                    tij_abc(3,1,3)*dp_j(3))
                         ef2_i(2,1) = ef2_i(2,1) + (tij_abc(1,2,1)*dp_j(1)+&
                                                    tij_abc(2,2,1)*dp_j(2)+&
                                                    tij_abc(3,2,1)*dp_j(3))
                         ef2_i(2,2) = ef2_i(2,2) + (tij_abc(1,2,2)*dp_j(1)+&
                                                    tij_abc(2,2,2)*dp_j(2)+&
                                                    tij_abc(3,2,2)*dp_j(3))
                         ef2_i(2,3) = ef2_i(2,3) + (tij_abc(1,2,3)*dp_j(1)+&
                                                    tij_abc(2,2,3)*dp_j(2)+&
                                                    tij_abc(3,2,3)*dp_j(3))
                         ef2_i(3,1) = ef2_i(3,1) + (tij_abc(1,3,1)*dp_j(1)+&
                                                    tij_abc(2,3,1)*dp_j(2)+&
                                                    tij_abc(3,3,1)*dp_j(3))
                         ef2_i(3,2) = ef2_i(3,2) + (tij_abc(1,3,2)*dp_j(1)+&
                                                    tij_abc(2,3,2)*dp_j(2)+&
                                                    tij_abc(3,3,2)*dp_j(3))
                         ef2_i(3,3) = ef2_i(3,3) + (tij_abc(1,3,3)*dp_j(1)+&
                                                    tij_abc(2,3,3)*dp_j(2)+&
                                                    tij_abc(3,3,3)*dp_j(3))

                         ef2_j(1,1) = ef2_j(1,1) - (tij_abc(1,1,1)*dp_i(1)+&
                                                    tij_abc(2,1,1)*dp_i(2)+&
                                                    tij_abc(3,1,1)*dp_i(3))
                         ef2_j(1,2) = ef2_j(1,2) - (tij_abc(1,1,2)*dp_i(1)+&
                                                    tij_abc(2,1,2)*dp_i(2)+&
                                                    tij_abc(3,1,2)*dp_i(3))
                         ef2_j(1,3) = ef2_j(1,3) - (tij_abc(1,1,3)*dp_i(1)+&
                                                    tij_abc(2,1,3)*dp_i(2)+&
                                                    tij_abc(3,1,3)*dp_i(3))
                         ef2_j(2,1) = ef2_j(2,1) - (tij_abc(1,2,1)*dp_i(1)+&
                                                    tij_abc(2,2,1)*dp_i(2)+&
                                                    tij_abc(3,2,1)*dp_i(3))
                         ef2_j(2,2) = ef2_j(2,2) - (tij_abc(1,2,2)*dp_i(1)+&
                                                    tij_abc(2,2,2)*dp_i(2)+&
                                                    tij_abc(3,2,2)*dp_i(3))
                         ef2_j(2,3) = ef2_j(2,3) - (tij_abc(1,2,3)*dp_i(1)+&
                                                    tij_abc(2,2,3)*dp_i(2)+&
                                                    tij_abc(3,2,3)*dp_i(3))
                         ef2_j(3,1) = ef2_j(3,1) - (tij_abc(1,3,1)*dp_i(1)+&
                                                    tij_abc(2,3,1)*dp_i(2)+&
                                                    tij_abc(3,3,1)*dp_i(3))
                         ef2_j(3,2) = ef2_j(3,2) - (tij_abc(1,3,2)*dp_i(1)+&
                                                    tij_abc(2,3,2)*dp_i(2)+&
                                                    tij_abc(3,3,2)*dp_i(3))
                         ef2_j(3,3) = ef2_j(3,3) - (tij_abc(1,3,3)*dp_i(1)+&
                                                    tij_abc(2,3,3)*dp_i(2)+&
                                                    tij_abc(3,3,3)*dp_i(3))
                      END IF
                   END IF
                END IF
                IF (task(2,1)) THEN
                   ! Dipole - Charge
                   tmp=   ch_j*(tij_a(1)*dp_i(1)+&
                                tij_a(2)*dp_i(2)+&
                                tij_a(3)*dp_i(3))&
                        - ch_i*(tij_a(1)*dp_j(1)+&
                                tij_a(2)*dp_j(2)+&
                                tij_a(3)*dp_j(3))
                   eloc = eloc + tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         fr(k) = fr(k) -  ch_j *(tij_ab(1,k)*dp_i(1)+&
                                                 tij_ab(2,k)*dp_i(2)+&
                                                 tij_ab(3,k)*dp_i(3))&
                                       +  ch_i *(tij_ab(1,k)*dp_j(1)+&
                                                 tij_ab(2,k)*dp_j(2)+&
                                                 tij_ab(3,k)*dp_j(3))
                      END DO
                   END IF
                END IF
                IF (task(3,3)) THEN
                   ! Quadrupole - Quadrupole
                   fac  = 1.0_dp/9.0_dp
                   tmp11 = qp_i(1,1)*(tij_abcd(1,1,1,1)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,1)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,1)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,1)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,1)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,1)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,1)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,1)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,1)*qp_j(3,3))
                   tmp21 = qp_i(2,1)*(tij_abcd(1,1,1,2)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,2)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,2)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,2)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,2)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,2)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,2)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,2)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,2)*qp_j(3,3))
                   tmp31 = qp_i(3,1)*(tij_abcd(1,1,1,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,3)*qp_j(3,3))
                   tmp22 = qp_i(2,2)*(tij_abcd(1,1,2,2)*qp_j(1,1)+&
                                      tij_abcd(2,1,2,2)*qp_j(2,1)+&
                                      tij_abcd(3,1,2,2)*qp_j(3,1)+&
                                      tij_abcd(1,2,2,2)*qp_j(1,2)+&
                                      tij_abcd(2,2,2,2)*qp_j(2,2)+&
                                      tij_abcd(3,2,2,2)*qp_j(3,2)+&
                                      tij_abcd(1,3,2,2)*qp_j(1,3)+&
                                      tij_abcd(2,3,2,2)*qp_j(2,3)+&
                                      tij_abcd(3,3,2,2)*qp_j(3,3))
                   tmp32 = qp_i(3,2)*(tij_abcd(1,1,2,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,2,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,2,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,2,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,2,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,2,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,2,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,2,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,2,3)*qp_j(3,3))
                   tmp33 = qp_i(3,3)*(tij_abcd(1,1,3,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,3,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,3,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,3,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,3,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,3,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,3,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,3,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,3,3)*qp_j(3,3))
                   tmp12 = tmp21
                   tmp13 = tmp31
                   tmp23 = tmp32
                   tmp   = tmp11 + tmp12 + tmp13 + &
                           tmp21 + tmp22 + tmp23 + &
                           tmp31 + tmp32 + tmp33

                   eloc = eloc + fac*tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         tmp11 = qp_i(1,1)*(tij_abcde(1,1,1,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,1,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,1,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,1,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,1,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,1,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,1,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,1,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,1,1,k)*qp_j(3,3))
                         tmp21 = qp_i(2,1)*(tij_abcde(1,1,2,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,2,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,2,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,2,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,2,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,2,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,2,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,2,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,2,1,k)*qp_j(3,3))
                         tmp31 = qp_i(3,1)*(tij_abcde(1,1,3,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,1,k)*qp_j(3,3))
                         tmp22 = qp_i(2,2)*(tij_abcde(1,1,2,2,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,2,2,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,2,2,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,2,2,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,2,2,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,2,2,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,2,2,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,2,2,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,2,2,k)*qp_j(3,3))
                         tmp32 = qp_i(3,2)*(tij_abcde(1,1,3,2,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,2,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,2,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,2,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,2,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,2,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,2,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,2,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,2,k)*qp_j(3,3))
                         tmp33 = qp_i(3,3)*(tij_abcde(1,1,3,3,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,3,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,3,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,3,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,3,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,3,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,3,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,3,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,3,k)*qp_j(3,3))
                         tmp12 = tmp21
                         tmp13 = tmp31
                         tmp23 = tmp32
                         fr(k) = fr(k) - fac * ( tmp11 + tmp12 + tmp13 +&
                                                 tmp21 + tmp22 + tmp23 +&
                                                 tmp31 + tmp32 + tmp33  )
                      END DO
                   END IF
                   ! Electric field
                   IF (do_efield) THEN
                      fac = 1.0_dp/3.0_dp
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i + fac*(tij_ab(1,1)*qp_j(1,1)+&
                                              tij_ab(2,1)*qp_j(2,1)+&
                                              tij_ab(3,1)*qp_j(3,1)+&
                                              tij_ab(1,2)*qp_j(1,2)+&
                                              tij_ab(2,2)*qp_j(2,2)+&
                                              tij_ab(3,2)*qp_j(3,2)+&
                                              tij_ab(1,3)*qp_j(1,3)+&
                                              tij_ab(2,3)*qp_j(2,3)+&
                                              tij_ab(3,3)*qp_j(3,3))

                         ef0_j = ef0_j + fac*(tij_ab(1,1)*qp_i(1,1)+&
                                              tij_ab(2,1)*qp_i(2,1)+&
                                              tij_ab(3,1)*qp_i(3,1)+&
                                              tij_ab(1,2)*qp_i(1,2)+&
                                              tij_ab(2,2)*qp_i(2,2)+&
                                              tij_ab(3,2)*qp_i(3,2)+&
                                              tij_ab(1,3)*qp_i(1,3)+&
                                              tij_ab(2,3)*qp_i(2,3)+&
                                              tij_ab(3,3)*qp_i(3,3))
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) - fac*(tij_abc(1,1,1)*qp_j(1,1)+&
                                                    tij_abc(2,1,1)*qp_j(2,1)+&
                                                    tij_abc(3,1,1)*qp_j(3,1)+&
                                                    tij_abc(1,2,1)*qp_j(1,2)+&
                                                    tij_abc(2,2,1)*qp_j(2,2)+&
                                                    tij_abc(3,2,1)*qp_j(3,2)+&
                                                    tij_abc(1,3,1)*qp_j(1,3)+&
                                                    tij_abc(2,3,1)*qp_j(2,3)+&
                                                    tij_abc(3,3,1)*qp_j(3,3))
                         ef1_i(2) = ef1_i(2) - fac*(tij_abc(1,1,2)*qp_j(1,1)+&
                                                    tij_abc(2,1,2)*qp_j(2,1)+&
                                                    tij_abc(3,1,2)*qp_j(3,1)+&
                                                    tij_abc(1,2,2)*qp_j(1,2)+&
                                                    tij_abc(2,2,2)*qp_j(2,2)+&
                                                    tij_abc(3,2,2)*qp_j(3,2)+&
                                                    tij_abc(1,3,2)*qp_j(1,3)+&
                                                    tij_abc(2,3,2)*qp_j(2,3)+&
                                                    tij_abc(3,3,2)*qp_j(3,3))
                         ef1_i(3) = ef1_i(3) - fac*(tij_abc(1,1,3)*qp_j(1,1)+&
                                                    tij_abc(2,1,3)*qp_j(2,1)+&
                                                    tij_abc(3,1,3)*qp_j(3,1)+&
                                                    tij_abc(1,2,3)*qp_j(1,2)+&
                                                    tij_abc(2,2,3)*qp_j(2,2)+&
                                                    tij_abc(3,2,3)*qp_j(3,2)+&
                                                    tij_abc(1,3,3)*qp_j(1,3)+&
                                                    tij_abc(2,3,3)*qp_j(2,3)+&
                                                    tij_abc(3,3,3)*qp_j(3,3))

                         ef1_j(1) = ef1_j(1) + fac*(tij_abc(1,1,1)*qp_i(1,1)+&
                                                    tij_abc(2,1,1)*qp_i(2,1)+&
                                                    tij_abc(3,1,1)*qp_i(3,1)+&
                                                    tij_abc(1,2,1)*qp_i(1,2)+&
                                                    tij_abc(2,2,1)*qp_i(2,2)+&
                                                    tij_abc(3,2,1)*qp_i(3,2)+&
                                                    tij_abc(1,3,1)*qp_i(1,3)+&
                                                    tij_abc(2,3,1)*qp_i(2,3)+&
                                                    tij_abc(3,3,1)*qp_i(3,3))
                         ef1_j(2) = ef1_j(2) + fac*(tij_abc(1,1,2)*qp_i(1,1)+&
                                                    tij_abc(2,1,2)*qp_i(2,1)+&
                                                    tij_abc(3,1,2)*qp_i(3,1)+&
                                                    tij_abc(1,2,2)*qp_i(1,2)+&
                                                    tij_abc(2,2,2)*qp_i(2,2)+&
                                                    tij_abc(3,2,2)*qp_i(3,2)+&
                                                    tij_abc(1,3,2)*qp_i(1,3)+&
                                                    tij_abc(2,3,2)*qp_i(2,3)+&
                                                    tij_abc(3,3,2)*qp_i(3,3))
                         ef1_j(3) = ef1_j(3) + fac*(tij_abc(1,1,3)*qp_i(1,1)+&
                                                    tij_abc(2,1,3)*qp_i(2,1)+&
                                                    tij_abc(3,1,3)*qp_i(3,1)+&
                                                    tij_abc(1,2,3)*qp_i(1,2)+&
                                                    tij_abc(2,2,3)*qp_i(2,2)+&
                                                    tij_abc(3,2,3)*qp_i(3,2)+&
                                                    tij_abc(1,3,3)*qp_i(1,3)+&
                                                    tij_abc(2,3,3)*qp_i(2,3)+&
                                                    tij_abc(3,3,3)*qp_i(3,3))
                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         tmp11 =   fac *(tij_abcd(1,1,1,1)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,1)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,1)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,1)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,1)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,1)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,1)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,1)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,1)*qp_j(3,3))
                         tmp12 =   fac *(tij_abcd(1,1,1,2)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,2)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,2)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,2)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,2)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,2)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,2)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,2)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,2)*qp_j(3,3))
                         tmp13 =   fac *(tij_abcd(1,1,1,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,3)*qp_j(3,3))
                         tmp22 =   fac *(tij_abcd(1,1,2,2)*qp_j(1,1)+&
                                         tij_abcd(2,1,2,2)*qp_j(2,1)+&
                                         tij_abcd(3,1,2,2)*qp_j(3,1)+&
                                         tij_abcd(1,2,2,2)*qp_j(1,2)+&
                                         tij_abcd(2,2,2,2)*qp_j(2,2)+&
                                         tij_abcd(3,2,2,2)*qp_j(3,2)+&
                                         tij_abcd(1,3,2,2)*qp_j(1,3)+&
                                         tij_abcd(2,3,2,2)*qp_j(2,3)+&
                                         tij_abcd(3,3,2,2)*qp_j(3,3))
                         tmp23 =   fac *(tij_abcd(1,1,2,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,2,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,2,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,2,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,2,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,2,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,2,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,2,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,2,3)*qp_j(3,3))
                         tmp33 =   fac *(tij_abcd(1,1,3,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,3,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,3,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,3,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,3,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,3,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,3,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,3,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,3,3)*qp_j(3,3))

                         ef2_i(1,1) = ef2_i(1,1) - tmp11
                         ef2_i(1,2) = ef2_i(1,2) - tmp12
                         ef2_i(1,3) = ef2_i(1,3) - tmp13
                         ef2_i(2,1) = ef2_i(2,1) - tmp12
                         ef2_i(2,2) = ef2_i(2,2) - tmp22
                         ef2_i(2,3) = ef2_i(2,3) - tmp23
                         ef2_i(3,1) = ef2_i(3,1) - tmp13
                         ef2_i(3,2) = ef2_i(3,2) - tmp23
                         ef2_i(3,3) = ef2_i(3,3) - tmp33

                         tmp11 =   fac *(tij_abcd(1,1,1,1)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,1)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,1)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,1)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,1)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,1)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,1)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,1)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,1)*qp_i(3,3))
                         tmp12 =   fac *(tij_abcd(1,1,1,2)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,2)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,2)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,2)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,2)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,2)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,2)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,2)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,2)*qp_i(3,3))
                         tmp13 =   fac *(tij_abcd(1,1,1,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,3)*qp_i(3,3))
                         tmp22 =   fac *(tij_abcd(1,1,2,2)*qp_i(1,1)+&
                                         tij_abcd(2,1,2,2)*qp_i(2,1)+&
                                         tij_abcd(3,1,2,2)*qp_i(3,1)+&
                                         tij_abcd(1,2,2,2)*qp_i(1,2)+&
                                         tij_abcd(2,2,2,2)*qp_i(2,2)+&
                                         tij_abcd(3,2,2,2)*qp_i(3,2)+&
                                         tij_abcd(1,3,2,2)*qp_i(1,3)+&
                                         tij_abcd(2,3,2,2)*qp_i(2,3)+&
                                         tij_abcd(3,3,2,2)*qp_i(3,3))
                         tmp23 =   fac *(tij_abcd(1,1,2,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,2,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,2,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,2,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,2,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,2,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,2,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,2,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,2,3)*qp_i(3,3))
                         tmp33 =   fac *(tij_abcd(1,1,3,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,3,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,3,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,3,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,3,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,3,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,3,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,3,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,3,3)*qp_i(3,3))

                         ef2_j(1,1) = ef2_j(1,1) - tmp11
                         ef2_j(1,2) = ef2_j(1,2) - tmp12
                         ef2_j(1,3) = ef2_j(1,3) - tmp13
                         ef2_j(2,1) = ef2_j(2,1) - tmp12
                         ef2_j(2,2) = ef2_j(2,2) - tmp22
                         ef2_j(2,3) = ef2_j(2,3) - tmp23
                         ef2_j(3,1) = ef2_j(3,1) - tmp13
                         ef2_j(3,2) = ef2_j(3,2) - tmp23
                         ef2_j(3,3) = ef2_j(3,3) - tmp33
                      END IF
                   END IF
                END IF
                IF (task(3,2)) THEN
                   ! Quadrupole - Dipole
                   fac = 1.0_dp/3.0_dp
                   ! Dipole i (locally B) - Quadrupole j (locally A)
                   tmp_ij = dp_i(1)*(tij_abc(1,1,1)*qp_j(1,1)+&
                                     tij_abc(2,1,1)*qp_j(2,1)+&
                                     tij_abc(3,1,1)*qp_j(3,1)+&
                                     tij_abc(1,2,1)*qp_j(1,2)+&
                                     tij_abc(2,2,1)*qp_j(2,2)+&
                                     tij_abc(3,2,1)*qp_j(3,2)+&
                                     tij_abc(1,3,1)*qp_j(1,3)+&
                                     tij_abc(2,3,1)*qp_j(2,3)+&
                                     tij_abc(3,3,1)*qp_j(3,3))+&
                            dp_i(2)*(tij_abc(1,1,2)*qp_j(1,1)+&
                                     tij_abc(2,1,2)*qp_j(2,1)+&
                                     tij_abc(3,1,2)*qp_j(3,1)+&
                                     tij_abc(1,2,2)*qp_j(1,2)+&
                                     tij_abc(2,2,2)*qp_j(2,2)+&
                                     tij_abc(3,2,2)*qp_j(3,2)+&
                                     tij_abc(1,3,2)*qp_j(1,3)+&
                                     tij_abc(2,3,2)*qp_j(2,3)+&
                                     tij_abc(3,3,2)*qp_j(3,3))+&
                            dp_i(3)*(tij_abc(1,1,3)*qp_j(1,1)+&
                                     tij_abc(2,1,3)*qp_j(2,1)+&
                                     tij_abc(3,1,3)*qp_j(3,1)+&
                                     tij_abc(1,2,3)*qp_j(1,2)+&
                                     tij_abc(2,2,3)*qp_j(2,2)+&
                                     tij_abc(3,2,3)*qp_j(3,2)+&
                                     tij_abc(1,3,3)*qp_j(1,3)+&
                                     tij_abc(2,3,3)*qp_j(2,3)+&
                                     tij_abc(3,3,3)*qp_j(3,3))

                   ! Dipole j (locally A) - Quadrupole i (locally B)
                   tmp_ji = dp_j(1)*(tij_abc(1,1,1)*qp_i(1,1)+&
                                     tij_abc(2,1,1)*qp_i(2,1)+&
                                     tij_abc(3,1,1)*qp_i(3,1)+&
                                     tij_abc(1,2,1)*qp_i(1,2)+&
                                     tij_abc(2,2,1)*qp_i(2,2)+&
                                     tij_abc(3,2,1)*qp_i(3,2)+&
                                     tij_abc(1,3,1)*qp_i(1,3)+&
                                     tij_abc(2,3,1)*qp_i(2,3)+&
                                     tij_abc(3,3,1)*qp_i(3,3))+&
                            dp_j(2)*(tij_abc(1,1,2)*qp_i(1,1)+&
                                     tij_abc(2,1,2)*qp_i(2,1)+&
                                     tij_abc(3,1,2)*qp_i(3,1)+&
                                     tij_abc(1,2,2)*qp_i(1,2)+&
                                     tij_abc(2,2,2)*qp_i(2,2)+&
                                     tij_abc(3,2,2)*qp_i(3,2)+&
                                     tij_abc(1,3,2)*qp_i(1,3)+&
                                     tij_abc(2,3,2)*qp_i(2,3)+&
                                     tij_abc(3,3,2)*qp_i(3,3))+&
                            dp_j(3)*(tij_abc(1,1,3)*qp_i(1,1)+&
                                     tij_abc(2,1,3)*qp_i(2,1)+&
                                     tij_abc(3,1,3)*qp_i(3,1)+&
                                     tij_abc(1,2,3)*qp_i(1,2)+&
                                     tij_abc(2,2,3)*qp_i(2,2)+&
                                     tij_abc(3,2,3)*qp_i(3,2)+&
                                     tij_abc(1,3,3)*qp_i(1,3)+&
                                     tij_abc(2,3,3)*qp_i(2,3)+&
                                     tij_abc(3,3,3)*qp_i(3,3))

                   tmp= fac * (tmp_ij - tmp_ji)
                   eloc = eloc + tmp
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         ! Dipole i (locally B) - Quadrupole j (locally A)
                         tmp_ij = dp_i(1)*(tij_abcd(1,1,1,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,1,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,1,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,1,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,1,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,1,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,1,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,1,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,1,k)*qp_j(3,3))+&
                                  dp_i(2)*(tij_abcd(1,1,2,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,2,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,2,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,2,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,2,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,2,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,2,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,2,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,2,k)*qp_j(3,3))+&
                                  dp_i(3)*(tij_abcd(1,1,3,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,3,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,3,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,3,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,3,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,3,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,3,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,3,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,3,k)*qp_j(3,3))

                         ! Dipole j (locally A) - Quadrupole i (locally B)
                         tmp_ji = dp_j(1)*(tij_abcd(1,1,1,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,1,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,1,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,1,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,1,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,1,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,1,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,1,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,1,k)*qp_i(3,3))+&
                                  dp_j(2)*(tij_abcd(1,1,2,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,2,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,2,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,2,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,2,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,2,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,2,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,2,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,2,k)*qp_i(3,3))+&
                                  dp_j(3)*(tij_abcd(1,1,3,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,3,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,3,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,3,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,3,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,3,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,3,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,3,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,3,k)*qp_i(3,3))

                         fr(k) = fr(k) - fac * (tmp_ij - tmp_ji)
                      END DO
                   END IF
                END IF
                IF (task(3,1)) THEN
                   ! Quadrupole - Charge
                   fac = 1.0_dp/3.0_dp

                   ! Quadrupole j (locally A) - Charge j (locally B)
                   tmp_ij = ch_i * (tij_ab(1,1)*qp_j(1,1)+&
                                    tij_ab(2,1)*qp_j(2,1)+&
                                    tij_ab(3,1)*qp_j(3,1)+&
                                    tij_ab(1,2)*qp_j(1,2)+&
                                    tij_ab(2,2)*qp_j(2,2)+&
                                    tij_ab(3,2)*qp_j(3,2)+&
                                    tij_ab(1,3)*qp_j(1,3)+&
                                    tij_ab(2,3)*qp_j(2,3)+&
                                    tij_ab(3,3)*qp_j(3,3))

                   ! Quadrupole i (locally B) - Charge j (locally A)
                   tmp_ji = ch_j * (tij_ab(1,1)*qp_i(1,1)+&
                                    tij_ab(2,1)*qp_i(2,1)+&
                                    tij_ab(3,1)*qp_i(3,1)+&
                                    tij_ab(1,2)*qp_i(1,2)+&
                                    tij_ab(2,2)*qp_i(2,2)+&
                                    tij_ab(3,2)*qp_i(3,2)+&
                                    tij_ab(1,3)*qp_i(1,3)+&
                                    tij_ab(2,3)*qp_i(2,3)+&
                                    tij_ab(3,3)*qp_i(3,3))

                   eloc = eloc + fac*(tmp_ij+tmp_ji)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         ! Quadrupole j (locally A) - Charge i (locally B)
                         tmp_ij = ch_i * (tij_abc(1,1,k)*qp_j(1,1)+&
                                          tij_abc(2,1,k)*qp_j(2,1)+&
                                          tij_abc(3,1,k)*qp_j(3,1)+&
                                          tij_abc(1,2,k)*qp_j(1,2)+&
                                          tij_abc(2,2,k)*qp_j(2,2)+&
                                          tij_abc(3,2,k)*qp_j(3,2)+&
                                          tij_abc(1,3,k)*qp_j(1,3)+&
                                          tij_abc(2,3,k)*qp_j(2,3)+&
                                          tij_abc(3,3,k)*qp_j(3,3))

                         ! Quadrupole i (locally B) - Charge j (locally A)
                         tmp_ji = ch_j * (tij_abc(1,1,k)*qp_i(1,1)+&
                                          tij_abc(2,1,k)*qp_i(2,1)+&
                                          tij_abc(3,1,k)*qp_i(3,1)+&
                                          tij_abc(1,2,k)*qp_i(1,2)+&
                                          tij_abc(2,2,k)*qp_i(2,2)+&
                                          tij_abc(3,2,k)*qp_i(3,2)+&
                                          tij_abc(1,3,k)*qp_i(1,3)+&
                                          tij_abc(2,3,k)*qp_i(2,3)+&
                                          tij_abc(3,3,k)*qp_i(3,3))

                         fr(k) = fr(k) - fac *(tmp_ij + tmp_ji)
                      END DO
                   END IF
                END IF
                energy = energy + eloc
                IF (do_forces) THEN
                   forces(1,atom_a) = forces(1,atom_a) - fr(1)
                   forces(2,atom_a) = forces(2,atom_a) - fr(2)
                   forces(3,atom_a) = forces(3,atom_a) - fr(3)
                   forces(1,atom_b) = forces(1,atom_b) + fr(1)
                   forces(2,atom_b) = forces(2,atom_b) + fr(2)
                   forces(3,atom_b) = forces(3,atom_b) + fr(3)
                END IF
                ! Electric fields
                IF (do_efield) THEN
                   ! Potential
                   IF (do_efield0) THEN
                      efield0(  atom_a) = efield0(  atom_a) + ef0_j

                      efield0(  atom_b) = efield0(  atom_b) + ef0_i
                   END IF
                   ! Electric field
                   IF (do_efield1) THEN
                      efield1(1,atom_a) = efield1(1,atom_a) + ef1_j(1)
                      efield1(2,atom_a) = efield1(2,atom_a) + ef1_j(2)
                      efield1(3,atom_a) = efield1(3,atom_a) + ef1_j(3)

                      efield1(1,atom_b) = efield1(1,atom_b) + ef1_i(1)
                      efield1(2,atom_b) = efield1(2,atom_b) + ef1_i(2)
                      efield1(3,atom_b) = efield1(3,atom_b) + ef1_i(3)
                   END IF
                   ! Electric field gradient
                   IF (do_efield2) THEN
                      efield2(1,atom_a) = efield2(1,atom_a) + ef2_j(1,1)
                      efield2(2,atom_a) = efield2(2,atom_a) + ef2_j(1,2)
                      efield2(3,atom_a) = efield2(3,atom_a) + ef2_j(1,3)
                      efield2(4,atom_a) = efield2(4,atom_a) + ef2_j(2,1)
                      efield2(5,atom_a) = efield2(5,atom_a) + ef2_j(2,2)
                      efield2(6,atom_a) = efield2(6,atom_a) + ef2_j(2,3)
                      efield2(7,atom_a) = efield2(7,atom_a) + ef2_j(3,1)
                      efield2(8,atom_a) = efield2(8,atom_a) + ef2_j(3,2)
                      efield2(9,atom_a) = efield2(9,atom_a) + ef2_j(3,3)

                      efield2(1,atom_b) = efield2(1,atom_b) + ef2_i(1,1)
                      efield2(2,atom_b) = efield2(2,atom_b) + ef2_i(1,2)
                      efield2(3,atom_b) = efield2(3,atom_b) + ef2_i(1,3)
                      efield2(4,atom_b) = efield2(4,atom_b) + ef2_i(2,1)
                      efield2(5,atom_b) = efield2(5,atom_b) + ef2_i(2,2)
                      efield2(6,atom_b) = efield2(6,atom_b) + ef2_i(2,3)
                      efield2(7,atom_b) = efield2(7,atom_b) + ef2_i(3,1)
                      efield2(8,atom_b) = efield2(8,atom_b) + ef2_i(3,2)
                      efield2(9,atom_b) = efield2(9,atom_b) + ef2_i(3,3)
                   END IF
                END IF
                IF (do_stress) THEN
                   ptens11 = ptens11 + rab(1) * fr(1)
                   ptens21 = ptens21 + rab(2) * fr(1)
                   ptens31 = ptens31 + rab(3) * fr(1)
                   ptens12 = ptens12 + rab(1) * fr(2)
                   ptens22 = ptens22 + rab(2) * fr(2)
                   ptens32 = ptens32 + rab(3) * fr(2)
                   ptens13 = ptens13 + rab(1) * fr(3)
                   ptens23 = ptens23 + rab(2) * fr(3)
                   ptens33 = ptens33 + rab(3) * fr(3)
                END IF
            END DO Pairs
         END DO Kind_Group_Loop
      END DO Lists
      IF (do_stress) THEN
         pv(1, 1) = pv(1, 1)+ptens11
         pv(1, 2) = pv(1, 2)+(ptens12+ptens21)*0.5_dp
         pv(1, 3) = pv(1, 3)+(ptens13+ptens31)*0.5_dp
         pv(2, 1) = pv(1, 2)
         pv(2, 2) = pv(2, 2)+ptens22
         pv(2, 3) = pv(2, 3)+(ptens23+ptens32)*0.5_dp
         pv(3, 1) = pv(1, 3)
         pv(3, 2) = pv(2, 3)
         pv(3, 3) = pv(3, 3)+ptens33
      END IF

      CALL timestop(handle)
   END SUBROUTINE ewald_multipole_bonded

! **************************************************************************************************
!> \brief computes the potential and the force for a lattice sum of multipoles
!>      up to quadrupole - Long Range (Reciprocal Space) Term
!> \param ewald_env ...
!> \param ewald_pw ...
!> \param cell ...
!> \param particle_set ...
!> \param local_particles ...
!> \param energy ...
!> \param task ...
!> \param do_forces ...
!> \param do_efield ...
!> \param do_stress ...
!> \param charges ...
!> \param dipoles ...
!> \param quadrupoles ...
!> \param forces ...
!> \param pv ...
!> \param efield0 ...
!> \param efield1 ...
!> \param efield2 ...
!> \author Teodoro Laino [tlaino] - 12.2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE ewald_multipole_LR(ewald_env, ewald_pw, cell, particle_set, &
                                 local_particles, energy, task, do_forces, do_efield, do_stress, &
                                 charges, dipoles, quadrupoles, forces, pv, efield0, efield1, efield2)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(cell_type), POINTER                           :: cell
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      REAL(KIND=dp), INTENT(INOUT)                       :: energy
      LOGICAL, DIMENSION(3, 3), INTENT(IN)               :: task
      LOGICAL, INTENT(IN)                                :: do_forces, do_efield, do_stress
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: dipoles
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: quadrupoles
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: forces, pv
      REAL(KIND=dp), DIMENSION(:), POINTER               :: efield0
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: efield1, efield2

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_multipole_LR', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: atm_factor, atm_factor_st(3), cnjg_fac, &
                                                            fac, summe_tmp
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:)        :: summe_ef
      COMPLEX(KIND=dp), ALLOCATABLE, DIMENSION(:, :)     :: summe_st
      INTEGER :: gpt, group, handle, iparticle, iparticle_kind, iparticle_local, lp, mp, nnodes, &
         node, np, nparticle_kind, nparticle_local
      INTEGER, DIMENSION(:, :), POINTER                  :: bds
      LOGICAL                                            :: do_efield0, do_efield1, do_efield2
      REAL(KIND=dp)                                      :: alpha, denom, dipole_t(3), f0, factor, &
                                                            four_alpha_sq, gauss, pref, q_t, tmp, &
                                                            trq_t
      REAL(KIND=dp), DIMENSION(3)                        :: tmp_v, vec
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv_tmp
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: rho0
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0
      TYPE(dg_type), POINTER                             :: dg
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(structure_factor_type)                        :: exp_igr

      CALL timeset(routineN, handle)
      do_efield0 = do_efield .AND. ASSOCIATED(efield0)
      do_efield1 = do_efield .AND. ASSOCIATED(efield1)
      do_efield2 = do_efield .AND. ASSOCIATED(efield2)

      ! Gathering data from the ewald environment
      CALL ewald_env_get(ewald_env, alpha=alpha, group=group)
      CALL ewald_pw_get(ewald_pw, pw_big_pool=pw_pool, dg=dg)
      CALL dg_get(dg, dg_rho0=dg_rho0)
      rho0 => dg_rho0%density%pw%cr3d
      pw_grid => pw_pool%pw_grid
      bds => pw_grid%bounds

      ! Allocation of working arrays
      nparticle_kind = SIZE(local_particles%n_el)
      nnodes = 0
      DO iparticle_kind = 1, nparticle_kind
         nnodes = nnodes+local_particles%n_el(iparticle_kind)
      ENDDO
      CALL structure_factor_allocate(pw_grid%bounds, nnodes, exp_igr)

      ALLOCATE (summe_ef(1:pw_grid%ngpts_cut))
      summe_ef = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
      ! Stress Tensor
      IF (do_stress) THEN
         pv_tmp = 0.0_dp
         ALLOCATE (summe_st(3, 1:pw_grid%ngpts_cut))
         summe_st = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
      END IF

      ! Defining four_alpha_sq
      four_alpha_sq = 4.0_dp*alpha**2
      dipole_t = 0.0_dp
      q_t = 0.0_dp
      trq_t = 0.0_dp
      ! Zero node count
      node = 0
      DO iparticle_kind = 1, nparticle_kind
         nparticle_local = local_particles%n_el(iparticle_kind)
         DO iparticle_local = 1, nparticle_local
            node = node+1
            iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
            CALL matvec_3x3(vec, cell%h_inv, particle_set(iparticle)%r)
            CALL structure_factor_evaluate(vec, exp_igr%lb, &
                                           exp_igr%ex(:, node), exp_igr%ey(:, node), exp_igr%ez(:, node))

            ! Computing the total charge, dipole and quadrupole trace (if any)
            IF (ANY(task(1, :))) THEN
               q_t = q_t+charges(iparticle)
            END IF
            IF (ANY(task(2, :))) THEN
               dipole_t = dipole_t+dipoles(:, iparticle)
            END IF
            IF (ANY(task(3, :))) THEN
               trq_t = trq_t+quadrupoles(1, 1, iparticle)+ &
                       quadrupoles(2, 2, iparticle)+ &
                       quadrupoles(3, 3, iparticle)
            END IF
         END DO
      END DO

      ! Looping over the positive g-vectors
      DO gpt = 1, pw_grid%ngpts_cut_local
         lp = pw_grid%mapl%pos(pw_grid%g_hat(1, gpt))
         mp = pw_grid%mapm%pos(pw_grid%g_hat(2, gpt))
         np = pw_grid%mapn%pos(pw_grid%g_hat(3, gpt))

         lp = lp+bds(1, 1)
         mp = mp+bds(1, 2)
         np = np+bds(1, 3)

         ! Initializing sum to be used in the energy and force
         node = 0
         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               node = node+1
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               ! Density for energy and forces
               CALL get_atom_factor(atm_factor, pw_grid, gpt, iparticle, task, charges, &
                                    dipoles, quadrupoles)
               summe_tmp = exp_igr%ex(lp, node)*exp_igr%ey(mp, node)*exp_igr%ez(np, node)
               summe_ef(gpt) = summe_ef(gpt)+atm_factor*summe_tmp

               ! Precompute pseudo-density for stress tensor calculation
               IF (do_stress) THEN
                  CALL get_atom_factor_stress(atm_factor_st, pw_grid, gpt, iparticle, task, &
                                              dipoles, quadrupoles)
                  summe_st(1:3, gpt) = summe_st(1:3, gpt)+atm_factor_st(1:3)*summe_tmp
               END IF
            END DO
         END DO
      END DO
      CALL mp_sum(q_t, group)
      CALL mp_sum(dipole_t, group)
      CALL mp_sum(trq_t, group)
      CALL mp_sum(summe_ef, group)
      IF (do_stress) CALL mp_sum(summe_st, group)

      ! Looping over the positive g-vectors
      DO gpt = 1, pw_grid%ngpts_cut_local
         ! computing the potential energy
         lp = pw_grid%mapl%pos(pw_grid%g_hat(1, gpt))
         mp = pw_grid%mapm%pos(pw_grid%g_hat(2, gpt))
         np = pw_grid%mapn%pos(pw_grid%g_hat(3, gpt))

         lp = lp+bds(1, 1)
         mp = mp+bds(1, 2)
         np = np+bds(1, 3)

         IF (pw_grid%gsq(gpt) == 0.0_dp) THEN
            ! G=0 vector for dipole-dipole and charge-quadrupole
            energy = energy+(1.0_dp/6.0_dp)*DOT_PRODUCT(dipole_t, dipole_t) &
                     -(1.0_dp/9.0_dp)*q_t*trq_t
            ! Stress tensor
            IF (do_stress) THEN
               pv_tmp(1, 1) = pv_tmp(1, 1)+(1.0_dp/6.0_dp)*DOT_PRODUCT(dipole_t, dipole_t)
               pv_tmp(2, 2) = pv_tmp(2, 2)+(1.0_dp/6.0_dp)*DOT_PRODUCT(dipole_t, dipole_t)
               pv_tmp(3, 3) = pv_tmp(3, 3)+(1.0_dp/6.0_dp)*DOT_PRODUCT(dipole_t, dipole_t)
            END IF
            ! Corrections for G=0 to potential, field and field gradient
            IF (do_efield .AND. (debug_e_field_en .OR. (.NOT. debug_this_module))) THEN
               ! This term is important and may give problems if one is debugging
               ! VS finite differences since it comes from a residual integral in
               ! the complex plane (cannot be reproduced with finite differences)
               node = 0
               DO iparticle_kind = 1, nparticle_kind
                  nparticle_local = local_particles%n_el(iparticle_kind)
                  DO iparticle_local = 1, nparticle_local
                     node = node+1
                     iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)

                     ! Potential
                     IF (do_efield0) THEN
                        efield0(iparticle) = efield0(iparticle)
                     END IF
                     ! Electrostatic field
                     IF (do_efield1) THEN
                        efield1(1:3, iparticle) = efield1(1:3, iparticle)-(1.0_dp/6.0_dp)*dipole_t(1:3)
                     END IF
                     ! Electrostatic field gradients
                     IF (do_efield2) THEN
                        efield2(1, iparticle) = efield2(1, iparticle)-(1.0_dp/(18.0_dp))*q_t
                        efield2(5, iparticle) = efield2(5, iparticle)-(1.0_dp/(18.0_dp))*q_t
                        efield2(9, iparticle) = efield2(9, iparticle)-(1.0_dp/(18.0_dp))*q_t
                     END IF
                  END DO
               END DO
            END IF
            CYCLE
         END IF
         gauss = (rho0(lp, mp, np)*pw_grid%vol)**2/pw_grid%gsq(gpt)
         factor = gauss*REAL(summe_ef(gpt)*CONJG(summe_ef(gpt)), KIND=dp)
         energy = energy+factor

         IF (do_forces .OR. do_efield) THEN
            node = 0
            DO iparticle_kind = 1, nparticle_kind
               nparticle_local = local_particles%n_el(iparticle_kind)
               DO iparticle_local = 1, nparticle_local
                  node = node+1
                  iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
                  fac = exp_igr%ex(lp, node)*exp_igr%ey(mp, node)*exp_igr%ez(np, node)
                  cnjg_fac = CONJG(fac)

                  ! Forces
                  IF (do_forces) THEN
                     CALL get_atom_factor(atm_factor, pw_grid, gpt, iparticle, task, charges, &
                                          dipoles, quadrupoles)

                     tmp = gauss*AIMAG(summe_ef(gpt)*(cnjg_fac*CONJG(atm_factor)))
                     forces(1, node) = forces(1, node)+tmp*pw_grid%g(1, gpt)
                     forces(2, node) = forces(2, node)+tmp*pw_grid%g(2, gpt)
                     forces(3, node) = forces(3, node)+tmp*pw_grid%g(3, gpt)
                  END IF

                  ! Electric field
                  IF (do_efield) THEN
                     ! Potential
                     IF (do_efield0) THEN
                        efield0(iparticle) = efield0(iparticle)+gauss*REAL(fac*CONJG(summe_ef(gpt)), KIND=dp)
                     END IF
                     ! Electric field
                     IF (do_efield1) THEN
                        tmp = AIMAG(fac*CONJG(summe_ef(gpt)))*gauss
                        efield1(1, iparticle) = efield1(1, iparticle)-tmp*pw_grid%g(1, gpt)
                        efield1(2, iparticle) = efield1(2, iparticle)-tmp*pw_grid%g(2, gpt)
                        efield1(3, iparticle) = efield1(3, iparticle)-tmp*pw_grid%g(3, gpt)
                     END IF
                     ! Electric field gradient
                     IF (do_efield2) THEN
                        tmp_v(1) = REAL(fac*CONJG(summe_ef(gpt)), KIND=dp)*pw_grid%g(1, gpt)*gauss
                        tmp_v(2) = REAL(fac*CONJG(summe_ef(gpt)), KIND=dp)*pw_grid%g(2, gpt)*gauss
                        tmp_v(3) = REAL(fac*CONJG(summe_ef(gpt)), KIND=dp)*pw_grid%g(3, gpt)*gauss

                        efield2(1, iparticle) = efield2(1, iparticle)+tmp_v(1)*pw_grid%g(1, gpt)
                        efield2(2, iparticle) = efield2(2, iparticle)+tmp_v(1)*pw_grid%g(2, gpt)
                        efield2(3, iparticle) = efield2(3, iparticle)+tmp_v(1)*pw_grid%g(3, gpt)
                        efield2(4, iparticle) = efield2(4, iparticle)+tmp_v(2)*pw_grid%g(1, gpt)
                        efield2(5, iparticle) = efield2(5, iparticle)+tmp_v(2)*pw_grid%g(2, gpt)
                        efield2(6, iparticle) = efield2(6, iparticle)+tmp_v(2)*pw_grid%g(3, gpt)
                        efield2(7, iparticle) = efield2(7, iparticle)+tmp_v(3)*pw_grid%g(1, gpt)
                        efield2(8, iparticle) = efield2(8, iparticle)+tmp_v(3)*pw_grid%g(2, gpt)
                        efield2(9, iparticle) = efield2(9, iparticle)+tmp_v(3)*pw_grid%g(3, gpt)
                     END IF
                  END IF
               END DO
            END DO
         END IF

         ! Compute the virial P*V
         IF (do_stress) THEN
            ! The Stress Tensor can be decomposed in two main components.
            ! The first one is just a normal ewald component for reciprocal space
            denom = 1.0_dp/four_alpha_sq+1.0_dp/pw_grid%gsq(gpt)
            pv_tmp(1, 1) = pv_tmp(1, 1)+factor*(1.0_dp-2.0_dp*pw_grid%g(1, gpt)*pw_grid%g(1, gpt)*denom)
            pv_tmp(1, 2) = pv_tmp(1, 2)-factor*(2.0_dp*pw_grid%g(1, gpt)*pw_grid%g(2, gpt)*denom)
            pv_tmp(1, 3) = pv_tmp(1, 3)-factor*(2.0_dp*pw_grid%g(1, gpt)*pw_grid%g(3, gpt)*denom)
            pv_tmp(2, 1) = pv_tmp(2, 1)-factor*(2.0_dp*pw_grid%g(2, gpt)*pw_grid%g(1, gpt)*denom)
            pv_tmp(2, 2) = pv_tmp(2, 2)+factor*(1.0_dp-2.0_dp*pw_grid%g(2, gpt)*pw_grid%g(2, gpt)*denom)
            pv_tmp(2, 3) = pv_tmp(2, 3)-factor*(2.0_dp*pw_grid%g(2, gpt)*pw_grid%g(3, gpt)*denom)
            pv_tmp(3, 1) = pv_tmp(3, 1)-factor*(2.0_dp*pw_grid%g(3, gpt)*pw_grid%g(1, gpt)*denom)
            pv_tmp(3, 2) = pv_tmp(3, 2)-factor*(2.0_dp*pw_grid%g(3, gpt)*pw_grid%g(2, gpt)*denom)
            pv_tmp(3, 3) = pv_tmp(3, 3)+factor*(1.0_dp-2.0_dp*pw_grid%g(3, gpt)*pw_grid%g(3, gpt)*denom)
            ! The second one can be written in the following way
            f0 = 2.0_dp*gauss
            pv_tmp(1, 1) = pv_tmp(1, 1)+f0*pw_grid%g(1, gpt)*REAL(summe_st(1, gpt)*CONJG(summe_ef(gpt)), KIND=dp)
            pv_tmp(1, 2) = pv_tmp(1, 2)+f0*pw_grid%g(1, gpt)*REAL(summe_st(2, gpt)*CONJG(summe_ef(gpt)), KIND=dp)
            pv_tmp(1, 3) = pv_tmp(1, 3)+f0*pw_grid%g(1, gpt)*REAL(summe_st(3, gpt)*CONJG(summe_ef(gpt)), KIND=dp)
            pv_tmp(2, 1) = pv_tmp(2, 1)+f0*pw_grid%g(2, gpt)*REAL(summe_st(1, gpt)*CONJG(summe_ef(gpt)), KIND=dp)
            pv_tmp(2, 2) = pv_tmp(2, 2)+f0*pw_grid%g(2, gpt)*REAL(summe_st(2, gpt)*CONJG(summe_ef(gpt)), KIND=dp)
            pv_tmp(2, 3) = pv_tmp(2, 3)+f0*pw_grid%g(2, gpt)*REAL(summe_st(3, gpt)*CONJG(summe_ef(gpt)), KIND=dp)
            pv_tmp(3, 1) = pv_tmp(3, 1)+f0*pw_grid%g(3, gpt)*REAL(summe_st(1, gpt)*CONJG(summe_ef(gpt)), KIND=dp)
            pv_tmp(3, 2) = pv_tmp(3, 2)+f0*pw_grid%g(3, gpt)*REAL(summe_st(2, gpt)*CONJG(summe_ef(gpt)), KIND=dp)
            pv_tmp(3, 3) = pv_tmp(3, 3)+f0*pw_grid%g(3, gpt)*REAL(summe_st(3, gpt)*CONJG(summe_ef(gpt)), KIND=dp)
         END IF
      END DO
      pref = fourpi/pw_grid%vol
      energy = energy*pref

      CALL structure_factor_deallocate(exp_igr)
      DEALLOCATE (summe_ef)
      IF (do_stress) THEN
         pv_tmp = pv_tmp*pref
         ! Symmetrize the tensor
         pv(1, 1) = pv(1, 1)+pv_tmp(1, 1)
         pv(1, 2) = pv(1, 2)+(pv_tmp(1, 2)+pv_tmp(2, 1))*0.5_dp
         pv(1, 3) = pv(1, 3)+(pv_tmp(1, 3)+pv_tmp(3, 1))*0.5_dp
         pv(2, 1) = pv(1, 2)
         pv(2, 2) = pv(2, 2)+pv_tmp(2, 2)
         pv(2, 3) = pv(2, 3)+(pv_tmp(2, 3)+pv_tmp(3, 2))*0.5_dp
         pv(3, 1) = pv(1, 3)
         pv(3, 2) = pv(2, 3)
         pv(3, 3) = pv(3, 3)+pv_tmp(3, 3)
         DEALLOCATE (summe_st)
      END IF
      IF (do_forces) THEN
         forces = 2.0_dp*forces*pref
      END IF
      IF (do_efield0) THEN
         efield0 = 2.0_dp*efield0*pref
      END IF
      IF (do_efield1) THEN
         efield1 = 2.0_dp*efield1*pref
      END IF
      IF (do_efield2) THEN
         efield2 = 2.0_dp*efield2*pref
      END IF
      CALL timestop(handle)

   END SUBROUTINE ewald_multipole_LR

! **************************************************************************************************
!> \brief Computes the atom factor including charge, dipole and quadrupole
!> \param atm_factor ...
!> \param pw_grid ...
!> \param gpt ...
!> \param iparticle ...
!> \param task ...
!> \param charges ...
!> \param dipoles ...
!> \param quadrupoles ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino] - 12.2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE get_atom_factor(atm_factor, pw_grid, gpt, iparticle, task, charges, &
                              dipoles, quadrupoles)
      COMPLEX(KIND=dp), INTENT(OUT)                      :: atm_factor
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, INTENT(IN)                                :: gpt
      INTEGER                                            :: iparticle
      LOGICAL, DIMENSION(3, 3), INTENT(IN)               :: task
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: charges
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: dipoles
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: quadrupoles

      CHARACTER(len=*), PARAMETER :: routineN = 'get_atom_factor', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: tmp
      INTEGER                                            :: i, j

      atm_factor = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
      IF (task(1, 1)) THEN
         ! Charge
         atm_factor = atm_factor+charges(iparticle)
      END IF
      IF (task(2, 2)) THEN
         ! Dipole
         tmp = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
         DO i = 1, 3
            tmp = tmp+dipoles(i, iparticle)*pw_grid%g(i, gpt)
         END DO
         atm_factor = atm_factor+tmp*CMPLX(0.0_dp, -1.0_dp, KIND=dp)
      END IF
      IF (task(3, 3)) THEN
         ! Quadrupole
         tmp = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
         DO i = 1, 3
            DO j = 1, 3
               tmp = tmp+quadrupoles(j, i, iparticle)*pw_grid%g(j, gpt)*pw_grid%g(i, gpt)
            END DO
         END DO
         atm_factor = atm_factor-1.0_dp/3.0_dp*tmp
      END IF

   END SUBROUTINE get_atom_factor

! **************************************************************************************************
!> \brief Computes the atom factor including charge, dipole and quadrupole
!> \param atm_factor ...
!> \param pw_grid ...
!> \param gpt ...
!> \param iparticle ...
!> \param task ...
!> \param dipoles ...
!> \param quadrupoles ...
!> \par History
!>      none
!> \author Teodoro Laino [tlaino] - 12.2007 - University of Zurich
! **************************************************************************************************
   SUBROUTINE get_atom_factor_stress(atm_factor, pw_grid, gpt, iparticle, task, &
                                     dipoles, quadrupoles)
      COMPLEX(KIND=dp), INTENT(OUT)                      :: atm_factor(3)
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      INTEGER, INTENT(IN)                                :: gpt
      INTEGER                                            :: iparticle
      LOGICAL, DIMENSION(3, 3), INTENT(IN)               :: task
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: dipoles
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: quadrupoles

      CHARACTER(len=*), PARAMETER :: routineN = 'get_atom_factor_stress', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      atm_factor = CMPLX(0.0_dp, 0.0_dp, KIND=dp)
      IF (ANY(task(2, :))) THEN
         ! Dipole
         atm_factor = dipoles(:, iparticle)*CMPLX(0.0_dp, -1.0_dp, KIND=dp)
      END IF
      IF (ANY(task(3, :))) THEN
         ! Quadrupole
         DO i = 1, 3
            atm_factor(1) = atm_factor(1)-1.0_dp/3.0_dp* &
                            (quadrupoles(1, i, iparticle)*pw_grid%g(i, gpt)+ &
                             quadrupoles(i, 1, iparticle)*pw_grid%g(i, gpt))
            atm_factor(2) = atm_factor(2)-1.0_dp/3.0_dp* &
                            (quadrupoles(2, i, iparticle)*pw_grid%g(i, gpt)+ &
                             quadrupoles(i, 2, iparticle)*pw_grid%g(i, gpt))
            atm_factor(3) = atm_factor(3)-1.0_dp/3.0_dp* &
                            (quadrupoles(3, i, iparticle)*pw_grid%g(i, gpt)+ &
                             quadrupoles(i, 3, iparticle)*pw_grid%g(i, gpt))
         END DO
      END IF

   END SUBROUTINE get_atom_factor_stress

! **************************************************************************************************
!> \brief Computes the self interaction from g-space and the neutralizing background
!>     when using multipoles
!> \param ewald_env ...
!> \param cell ...
!> \param local_particles ...
!> \param e_self ...
!> \param e_neut ...
!> \param task ...
!> \param do_efield ...
!> \param radii ...
!> \param charges ...
!> \param dipoles ...
!> \param quadrupoles ...
!> \param efield0 ...
!> \param efield1 ...
!> \param efield2 ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 12.2007
! **************************************************************************************************
   SUBROUTINE ewald_multipole_self(ewald_env, cell, local_particles, e_self, &
                                   e_neut, task, do_efield, radii, charges, dipoles, quadrupoles, efield0, &
                                   efield1, efield2)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(cell_type), INTENT(IN)                        :: cell
      TYPE(distribution_1d_type), POINTER                :: local_particles
      REAL(KIND=dp), INTENT(OUT)                         :: e_self, e_neut
      LOGICAL, DIMENSION(3, 3), INTENT(IN)               :: task
      LOGICAL, INTENT(IN)                                :: do_efield
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: radii, charges
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: dipoles
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: quadrupoles
      REAL(KIND=dp), DIMENSION(:), POINTER               :: efield0
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: efield1, efield2

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_multipole_self', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: f23 = 2.0_dp/3.0_dp, &
                                                            f415 = 4.0_dp/15.0_dp

      INTEGER                                            :: ewald_type, group, i, iparticle, &
                                                            iparticle_kind, iparticle_local, j, &
                                                            nparticle_local
      LOGICAL                                            :: do_efield0, do_efield1, do_efield2, &
                                                            lradii
      REAL(KIND=dp)                                      :: alpha, ch_qu_self, ch_qu_self_tmp, &
                                                            dipole_self, fac1, fac2, fac3, fac4, &
                                                            q, q_neutg, q_self, q_sum, qu_qu_self, &
                                                            radius

      CALL ewald_env_get(ewald_env, ewald_type=ewald_type, alpha=alpha, &
                         group=group)

      do_efield0 = do_efield .AND. ASSOCIATED(efield0)
      do_efield1 = do_efield .AND. ASSOCIATED(efield1)
      do_efield2 = do_efield .AND. ASSOCIATED(efield2)
      q_self = 0.0_dp
      q_sum = 0.0_dp
      dipole_self = 0.0_dp
      ch_qu_self = 0.0_dp
      qu_qu_self = 0.0_dp
      fac1 = 2.0_dp*alpha*oorootpi
      fac2 = 6.0_dp*(f23**2)*(alpha**3)*oorootpi
      fac3 = (2.0_dp*oorootpi)*f23*alpha**3
      fac4 = (4.0_dp*oorootpi)*f415*alpha**5
      lradii = PRESENT(radii)
      radius = 0.0_dp
      q_neutg = 0.0_dp
      DO iparticle_kind = 1, SIZE(local_particles%n_el)
         nparticle_local = local_particles%n_el(iparticle_kind)
         DO iparticle_local = 1, nparticle_local
            iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
            IF (ANY(task(1, :))) THEN
               ! Charge - Charge
               q = charges(iparticle)
               IF (lradii) radius = radii(iparticle)
               IF (radius > 0) THEN
                  q_neutg = q_neutg+2.0_dp*q*radius**2
               END IF
               q_self = q_self+q*q
               q_sum = q_sum+q
               ! Potential
               IF (do_efield0) THEN
                  efield0(iparticle) = efield0(iparticle)-q*fac1
               END IF

               IF (task(1, 3)) THEN
                  ! Charge - Quadrupole
                  ch_qu_self_tmp = 0.0_dp
                  DO i = 1, 3
                     ch_qu_self_tmp = ch_qu_self_tmp+quadrupoles(i, i, iparticle)*q
                  END DO
                  ch_qu_self = ch_qu_self+ch_qu_self_tmp
                  ! Electric Field Gradient
                  IF (do_efield2) THEN
                     efield2(1, iparticle) = efield2(1, iparticle)+fac2*q
                     efield2(5, iparticle) = efield2(5, iparticle)+fac2*q
                     efield2(9, iparticle) = efield2(9, iparticle)+fac2*q
                  END IF
               END IF
            END IF
            IF (ANY(task(2, :))) THEN
               ! Dipole - Dipole
               DO i = 1, 3
                  dipole_self = dipole_self+dipoles(i, iparticle)**2
               END DO
               ! Electric Field
               IF (do_efield1) THEN
                  efield1(1, iparticle) = efield1(1, iparticle)+fac3*dipoles(1, iparticle)
                  efield1(2, iparticle) = efield1(2, iparticle)+fac3*dipoles(2, iparticle)
                  efield1(3, iparticle) = efield1(3, iparticle)+fac3*dipoles(3, iparticle)
               END IF
            END IF
            IF (ANY(task(3, :))) THEN
               ! Quadrupole - Quadrupole
               DO i = 1, 3
                  DO j = 1, 3
                     qu_qu_self = qu_qu_self+quadrupoles(j, i, iparticle)**2
                  END DO
               END DO
               ! Electric Field Gradient
               IF (do_efield2) THEN
                  efield2(1, iparticle) = efield2(1, iparticle)+fac4*quadrupoles(1, 1, iparticle)
                  efield2(2, iparticle) = efield2(2, iparticle)+fac4*quadrupoles(2, 1, iparticle)
                  efield2(3, iparticle) = efield2(3, iparticle)+fac4*quadrupoles(3, 1, iparticle)
                  efield2(4, iparticle) = efield2(4, iparticle)+fac4*quadrupoles(1, 2, iparticle)
                  efield2(5, iparticle) = efield2(5, iparticle)+fac4*quadrupoles(2, 2, iparticle)
                  efield2(6, iparticle) = efield2(6, iparticle)+fac4*quadrupoles(3, 2, iparticle)
                  efield2(7, iparticle) = efield2(7, iparticle)+fac4*quadrupoles(1, 3, iparticle)
                  efield2(8, iparticle) = efield2(8, iparticle)+fac4*quadrupoles(2, 3, iparticle)
                  efield2(9, iparticle) = efield2(9, iparticle)+fac4*quadrupoles(3, 3, iparticle)
               END IF
            END IF
         END DO
      END DO

      CALL mp_sum(q_neutg, group)
      CALL mp_sum(q_self, group)
      CALL mp_sum(q_sum, group)
      CALL mp_sum(dipole_self, group)
      CALL mp_sum(ch_qu_self, group)
      CALL mp_sum(qu_qu_self, group)

      e_self = -(q_self+f23*(dipole_self-f23*ch_qu_self+f415*qu_qu_self*alpha**2)*alpha**2)*alpha*oorootpi
      fac1 = pi/(2.0_dp*cell%deth)
      e_neut = -q_sum*fac1*(q_sum/alpha**2-q_neutg)

      ! Correcting Potential for the neutralizing background charge
      DO iparticle_kind = 1, SIZE(local_particles%n_el)
         nparticle_local = local_particles%n_el(iparticle_kind)
         DO iparticle_local = 1, nparticle_local
            iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
            IF (ANY(task(1, :))) THEN
               ! Potential energy
               IF (do_efield0) THEN
                  efield0(iparticle) = efield0(iparticle)-q_sum*2.0_dp*fac1/alpha**2
                  IF (lradii) radius = radii(iparticle)
                  IF (radius > 0) THEN
                     q = charges(iparticle)
                     efield0(iparticle) = efield0(iparticle)+fac1*radius**2*(q_sum+q)
                  END IF
               END IF
            END IF
         END DO
      END DO
   END SUBROUTINE ewald_multipole_self

! **************************************************************************************************
!> \brief ...
!> \param iw ...
!> \param e_gspace ...
!> \param e_rspace ...
!> \param e_bonded ...
!> \param e_self ...
!> \param e_neut ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 12.2007
! **************************************************************************************************
   SUBROUTINE ewald_multipole_print(iw, e_gspace, e_rspace, e_bonded, e_self, e_neut)

      INTEGER, INTENT(IN)                                :: iw
      REAL(KIND=dp), INTENT(IN)                          :: e_gspace, e_rspace, e_bonded, e_self, &
                                                            e_neut

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_multipole_print', &
         routineP = moduleN//':'//routineN

      IF (iw > 0) THEN
         WRITE (iw, '( A, A )') ' *********************************', &
            '**********************************************'
         WRITE (iw, '( A, A, T35, A, T56, E25.15 )') ' INITIAL GSPACE ENERGY', &
            '[hartree]', '= ', e_gspace
         WRITE (iw, '( A, A, T35, A, T56, E25.15 )') ' INITIAL RSPACE ENERGY', &
            '[hartree]', '= ', e_rspace
         WRITE (iw, '( A, A, T35, A, T56, E25.15 )') ' BONDED CORRECTION', &
            '[hartree]', '= ', e_bonded
         WRITE (iw, '( A, A, T35, A, T56, E25.15 )') ' SELF ENERGY CORRECTION', &
            '[hartree]', '= ', e_self
         WRITE (iw, '( A, A, T35, A, T56, E25.15 )') ' NEUTRALIZ. BCKGR. ENERGY', &
            '[hartree]', '= ', e_neut
         WRITE (iw, '( A, A, T35, A, T56, E25.15 )') ' TOTAL ELECTROSTATIC EN.', &
            '[hartree]', '= ', e_rspace+e_bonded+e_gspace+e_self+e_neut
         WRITE (iw, '( A, A )') ' *********************************', &
            '**********************************************'
      END IF
   END SUBROUTINE ewald_multipole_print

END MODULE ewalds_multipole
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      Harald Forbert (Dec-2000): Changes for multiple linked lists
!>                                 linklist_internal_data_type
!>      07.02.2005: using real coordinates for r_last_update; cleaned (MK)
!> \author CJM,MK
! **************************************************************************************************
MODULE fist_neighbor_list_control

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE cell_types,                      ONLY: cell_clone,&
                                              cell_create,&
                                              cell_release,&
                                              cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE exclusion_types,                 ONLY: exclusion_type
   USE fist_neighbor_list_types,        ONLY: fist_neighbor_type
   USE fist_neighbor_lists,             ONLY: build_fist_neighbor_lists
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_get,&
                                              fist_nonbond_env_set,&
                                              fist_nonbond_env_type,&
                                              pos_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_max
   USE pair_potential_types,            ONLY: pair_potential_pp_type,&
                                              siepmann_type,&
                                              tersoff_type
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'fist_neighbor_list_control'

   PUBLIC :: list_control

!***

CONTAINS

! to decide whether the neighbor list is to be updated or not
! based on a displacement criterion;
! if any particle has moved by 0.5*verlet_skin from the previous
! list update, then the list routine is called.

! **************************************************************************************************
!> \brief ...
!> \param atomic_kind_set ...
!> \param particle_set ...
!> \param local_particles ...
!> \param cell ...
!> \param fist_nonbond_env ...
!> \param para_env ...
!> \param mm_section ...
!> \param shell_particle_set ...
!> \param core_particle_set ...
!> \param force_update ...
!> \param exclusions ...
! **************************************************************************************************
   SUBROUTINE list_control(atomic_kind_set, particle_set, local_particles, &
                           cell, fist_nonbond_env, para_env, mm_section, shell_particle_set, &
                           core_particle_set, force_update, exclusions)

      TYPE(atomic_kind_type), POINTER                    :: atomic_kind_set(:)
      TYPE(particle_type), POINTER                       :: particle_set(:)
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(cell_type), POINTER                           :: cell
      TYPE(fist_nonbond_env_type), POINTER               :: fist_nonbond_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: mm_section
      TYPE(particle_type), OPTIONAL, POINTER             :: shell_particle_set(:), &
                                                            core_particle_set(:)
      LOGICAL, INTENT(IN), OPTIONAL                      :: force_update
      TYPE(exclusion_type), DIMENSION(:), OPTIONAL       :: exclusions

      CHARACTER(LEN=*), PARAMETER :: routineN = 'list_control', routineP = moduleN//':'//routineN

      INTEGER :: counter, handle, ikind, iparticle, iparticle_kind, iparticle_local, ishell, &
         jkind, last_update, nparticle, nparticle_kind, nparticle_local, nshell, num_update, &
         output_unit
      LOGICAL                                            :: build_from_scratch, geo_check, &
                                                            shell_adiabatic, shell_present, &
                                                            update_neighbor_lists
      LOGICAL, DIMENSION(:, :), POINTER                  :: full_nl
      REAL(KIND=dp)                                      :: aup, dr2, dr2_max, ei_scale14, lup, &
                                                            vdw_scale14, verlet_skin
      REAL(KIND=dp), DIMENSION(3)                        :: dr, rab, s, s2r
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: rlist_cut, rlist_lowsq
      TYPE(cell_type), POINTER                           :: cell_last_update
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(fist_neighbor_type), POINTER                  :: nonbonded
      TYPE(pair_potential_pp_type), POINTER              :: potparm
      TYPE(pos_type), DIMENSION(:), POINTER              :: r_last_update, r_last_update_pbc, &
                                                            rcore_last_update_pbc, &
                                                            rshell_last_update_pbc

      CALL timeset(routineN, handle)
      NULLIFY (logger)
      logger => cp_get_default_logger()

      ! *** Assigning local pointers ***
      CALL fist_nonbond_env_get(fist_nonbond_env, &
                                nonbonded=nonbonded, &
                                rlist_cut=rlist_cut, &
                                rlist_lowsq=rlist_lowsq, &
                                aup=aup, &
                                lup=lup, &
                                ei_scale14=ei_scale14, &
                                vdw_scale14=vdw_scale14, &
                                counter=counter, &
                                r_last_update=r_last_update, &
                                r_last_update_pbc=r_last_update_pbc, &
                                rshell_last_update_pbc=rshell_last_update_pbc, &
                                rcore_last_update_pbc=rcore_last_update_pbc, &
                                cell_last_update=cell_last_update, &
                                num_update=num_update, &
                                potparm=potparm, &
                                last_update=last_update)

      nparticle = SIZE(particle_set)
      nparticle_kind = SIZE(atomic_kind_set)
      nshell = 0
      CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, &
                               shell_present=shell_present, shell_adiabatic=shell_adiabatic)
      IF (shell_present) THEN
         nshell = SIZE(shell_particle_set)
      END IF

      ! *** Check, if the neighbor lists have to be built or updated ***
      update_neighbor_lists = .FALSE.
      CALL section_vals_val_get(mm_section, "NEIGHBOR_LISTS%NEIGHBOR_LISTS_FROM_SCRATCH", &
                                l_val=build_from_scratch)
      CALL section_vals_val_get(mm_section, "NEIGHBOR_LISTS%GEO_CHECK", &
                                l_val=geo_check)
      IF (ASSOCIATED(r_last_update)) THEN
         ! Determine the maximum of the squared displacement, compared to
         ! r_last_update.
         CALL section_vals_val_get(mm_section, "NEIGHBOR_LISTS%VERLET_SKIN", &
                                   r_val=verlet_skin)
         dr2_max = 0.0_dp
         DO iparticle_kind = 1, nparticle_kind
            nparticle_local = local_particles%n_el(iparticle_kind)
            DO iparticle_local = 1, nparticle_local
               iparticle = local_particles%list(iparticle_kind)%array(iparticle_local)
               s2r = r_last_update(iparticle)%r
               s = particle_set(iparticle)%r(:)
               dr(:) = s2r-s
               dr2 = dr(1)*dr(1)+dr(2)*dr(2)+dr(3)*dr(3)
               dr2_max = MAX(dr2_max, dr2)
            END DO
         END DO

         CALL mp_max(dr2_max, para_env%group)

         ! If the maximum distplacement is too large, ...
         IF (dr2_max > 0.25_dp*verlet_skin**2 .OR. build_from_scratch) THEN
            DO iparticle = 1, nparticle
               r_last_update(iparticle)%r = particle_set(iparticle)%r(:)
            END DO
            update_neighbor_lists = .TRUE.
         END IF
      ELSE
         ! There is no r_last_update to compare with. Neighbor lists from scratch.
         ALLOCATE (r_last_update(nparticle))
         DO iparticle = 1, nparticle
            r_last_update(iparticle)%r = particle_set(iparticle)%r(:)
         END DO

         update_neighbor_lists = .TRUE.
         build_from_scratch = .TRUE.
      END IF
      ! Force Update
      IF (PRESENT(force_update)) THEN
         IF (force_update) update_neighbor_lists = .TRUE.
      END IF

      ! Allocate the r_last_update_pbc, rshell_last_update_pbc, rcore_last_update_pbc
      IF (.NOT. ASSOCIATED(r_last_update_pbc)) THEN
         ALLOCATE (r_last_update_pbc(nparticle))
      END IF
      IF (shell_present .AND. .NOT. ASSOCIATED(rshell_last_update_pbc)) THEN
         ALLOCATE (rshell_last_update_pbc(nshell))
      END IF
      IF (shell_present .AND. .NOT. ASSOCIATED(rcore_last_update_pbc)) THEN
         ALLOCATE (rcore_last_update_pbc(nshell))
      END IF

      ! update the neighbor lists
      IF (update_neighbor_lists) THEN
         ! determine which pairs of atom kinds need full neighbor lists. Full
         ! means that atom a is in the neighbor list of atom b and vice versa.
         ALLOCATE (full_nl(nparticle_kind, nparticle_kind))
         IF (ASSOCIATED(potparm)) THEN
            DO ikind = 1, nparticle_kind
               DO jkind = ikind, nparticle_kind
                  full_nl(ikind, jkind) = .FALSE.
                  IF (ANY(potparm%pot(ikind, jkind)%pot%type == tersoff_type)) THEN
                     full_nl(ikind, jkind) = .TRUE.
                  END IF
                  IF (ANY(potparm%pot(ikind, jkind)%pot%type == siepmann_type)) THEN
                     full_nl(ikind, jkind) = .TRUE.
                  END IF
                  full_nl(jkind, ikind) = full_nl(ikind, jkind)
               END DO
            END DO
         ELSE
            full_nl = .FALSE.
         END IF
         CALL build_fist_neighbor_lists(atomic_kind_set, particle_set, &
                                        local_particles, cell, rlist_cut, rlist_lowsq, ei_scale14, &
                                        vdw_scale14, nonbonded, para_env, &
                                        build_from_scratch=build_from_scratch, geo_check=geo_check, &
                                        mm_section=mm_section, full_nl=full_nl, &
                                        exclusions=exclusions)

         CALL cell_release(cell_last_update)
         CALL cell_create(cell_last_update)
         CALL cell_clone(cell, cell_last_update)

         IF (counter > 0) THEN
            num_update = num_update+1
            lup = counter+1-last_update
            last_update = counter+1
            aup = aup+(lup-aup)/REAL(num_update, KIND=dp)
         ELSE
            num_update = 0
            lup = 0
            last_update = 1
            aup = 0.0_dp
         END IF

         CALL fist_nonbond_env_set(fist_nonbond_env, &
                                   lup=lup, &
                                   aup=aup, &
                                   r_last_update=r_last_update, &
                                   r_last_update_pbc=r_last_update_pbc, &
                                   rshell_last_update_pbc=rshell_last_update_pbc, &
                                   rcore_last_update_pbc=rcore_last_update_pbc, &
                                   nonbonded=nonbonded, &
                                   num_update=num_update, &
                                   last_update=last_update, &
                                   cell_last_update=cell_last_update)

         output_unit = cp_print_key_unit_nr(logger, mm_section, "PRINT%NEIGHBOR_LISTS", &
                                            extension=".mmLog")
         IF (output_unit > 0) THEN
            WRITE (UNIT=output_unit, &
                   FMT="(/,T2,A,/,T52,A,/,A,T31,A,T49,2(1X,F15.2),/,T2,A,/)") &
               REPEAT("*", 79), "INSTANTANEOUS        AVERAGES", &
               " LIST UPDATES[steps]", "= ", lup, aup, REPEAT("*", 79)
         END IF
         CALL cp_print_key_finished_output(output_unit, logger, mm_section, &
                                           "PRINT%NEIGHBOR_LISTS")
         DEALLOCATE (full_nl)
      END IF

      ! Store particle positions after the last update, translated to the
      ! primitive cell, in r_last_update_pbc.
      DO iparticle = 1, nparticle
         rab = r_last_update(iparticle)%r
         IF (cell%orthorhombic) THEN
            rab(1) = -cell%hmat(1, 1)*cell%perd(1)*ANINT(cell_last_update%h_inv(1, 1)*rab(1))
            rab(2) = -cell%hmat(2, 2)*cell%perd(2)*ANINT(cell_last_update%h_inv(2, 2)*rab(2))
            rab(3) = -cell%hmat(3, 3)*cell%perd(3)*ANINT(cell_last_update%h_inv(3, 3)*rab(3))
         ELSE
            s(1) = cell_last_update%h_inv(1, 1)*rab(1)+cell_last_update%h_inv(1, 2)*rab(2)+ &
                   cell_last_update%h_inv(1, 3)*rab(3)
            s(2) = cell_last_update%h_inv(2, 1)*rab(1)+cell_last_update%h_inv(2, 2)*rab(2)+ &
                   cell_last_update%h_inv(2, 3)*rab(3)
            s(3) = cell_last_update%h_inv(3, 1)*rab(1)+cell_last_update%h_inv(3, 2)*rab(2)+ &
                   cell_last_update%h_inv(3, 3)*rab(3)
            s(1) = -cell%perd(1)*ANINT(s(1))
            s(2) = -cell%perd(2)*ANINT(s(2))
            s(3) = -cell%perd(3)*ANINT(s(3))
            rab(1) = +cell%hmat(1, 1)*s(1)+cell%hmat(1, 2)*s(2)+cell%hmat(1, 3)*s(3)
            rab(2) = +cell%hmat(2, 1)*s(1)+cell%hmat(2, 2)*s(2)+cell%hmat(2, 3)*s(3)
            rab(3) = +cell%hmat(3, 1)*s(1)+cell%hmat(3, 2)*s(2)+cell%hmat(3, 3)*s(3)
         END IF
         r_last_update_pbc(iparticle)%r = particle_set(iparticle)%r+rab
         ! Use the same translation for core and shell.
         ishell = particle_set(iparticle)%shell_index
         IF (ishell /= 0) THEN
            rshell_last_update_pbc(ishell)%r = rab+shell_particle_set(ishell)%r(:)
            IF (shell_adiabatic) THEN
               rcore_last_update_pbc(ishell)%r = rab+core_particle_set(ishell)%r(:)
            ELSE
               rcore_last_update_pbc(ishell)%r = r_last_update_pbc(iparticle)%r(:)
            END IF
         END IF
      END DO

      counter = counter+1
      CALL fist_nonbond_env_set(fist_nonbond_env, counter=counter)
      CALL timestop(handle)

   END SUBROUTINE list_control

END MODULE fist_neighbor_list_control
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Utilities for Integrals for semi-empiric methods
!> \author Teodoro Laino (03.2008) [tlaino] - University of Zurich
! **************************************************************************************************
MODULE semi_empirical_int_utils

   USE input_constants,                 ONLY: do_method_pchg,&
                                              do_se_IS_kdso_d
   USE kinds,                           ONLY: dp
   USE semi_empirical_int3_utils,       ONLY: charg_int_3,&
                                              dcharg_int_3,&
                                              ijkl_low_3
   USE semi_empirical_int_arrays,       ONLY: &
        CLMp, CLMxx, CLMxy, CLMyy, CLMz, CLMzp, CLMzz, clm_d, clm_sp, ijkl_ind, indexa, indexb, &
        int2c_type
   USE semi_empirical_types,            ONLY: rotmat_type,&
                                              se_int_control_type,&
                                              se_int_screen_type,&
                                              se_taper_type,&
                                              semi_empirical_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2016  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief Debug the derivatives of the the rotational matrices
!>
!> \author Teodoro Laino [tlaino] - University of Zurich
!> \date 04.2008 [tlaino]
! *****************************************************************************
INTERFACE 
   SUBROUTINE check_rotmat_der( sepi, sepj, rjiv, ij_matrix, do_invert)
     USE kinds,                           ONLY: dp
     USE semi_empirical_types,            ONLY: rotmat_type,&
                                                semi_empirical_type
     IMPLICIT NONE
     TYPE(semi_empirical_type), POINTER       :: sepi, sepj
     REAL(KIND=dp), DIMENSION(3), INTENT(IN)  :: rjiv
     TYPE(rotmat_type), POINTER               :: ij_matrix
     LOGICAL, INTENT(IN)                      :: do_invert
   END SUBROUTINE check_rotmat_der
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical NUCINT ssss
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE
  SUBROUTINE check_dssss_nucint_ana (sepi,sepj,r,dssss,itype,se_int_control,&
       se_taper)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), INTENT(IN)                     :: r
    REAL(dp), INTENT(IN)                     :: dssss
    INTEGER, INTENT(IN)                      :: itype
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
  END SUBROUTINE check_dssss_nucint_ana
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical NUCINT core
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
  SUBROUTINE check_dcore_nucint_ana (sepi,sepj,r,dcore,itype,se_int_control,&
       se_taper)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), INTENT(IN)                     :: r
    REAL(dp), DIMENSION(10, 2), INTENT(IN)   :: dcore
    INTEGER, INTENT(IN)                      :: itype
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
  END SUBROUTINE check_dcore_nucint_ana
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical ROTNUC
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
   SUBROUTINE check_drotnuc_ana(sepi, sepj, rijv, itype, se_int_control, se_taper,&
        e1b, e2a, de1b, de2a)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), DIMENSION(3), INTENT(IN)       :: rijv
    INTEGER, INTENT(IN)                      :: itype
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
    REAL(dp), DIMENSION(45), INTENT(IN), &
      OPTIONAL                               :: e1b, e2a
    REAL(dp), DIMENSION(45, 3), &
      INTENT(IN), OPTIONAL                   :: de1b, de2a
   END SUBROUTINE check_drotnuc_ana
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical CORECORE
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
  SUBROUTINE check_dcorecore_ana(sepi, sepj, rijv, itype,se_int_control,&
       se_taper, enuc, denuc)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), DIMENSION(3), INTENT(IN)       :: rijv
    INTEGER, INTENT(IN)                      :: itype
    REAL(dp), INTENT(IN), OPTIONAL           :: enuc
    REAL(dp), DIMENSION(3), INTENT(IN), &
         OPTIONAL                            :: denuc
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
  END SUBROUTINE check_dcorecore_ana

END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical rot_2el_2c_first
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
  SUBROUTINE rot_2el_2c_first_debug(sepi, sepj, rijv, se_int_control, se_taper,&
       invert, ii, kk, v_d)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(KIND=dp), DIMENSION(3), INTENT(IN)  :: rijv
    LOGICAL, INTENT(IN)                      :: invert
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
    INTEGER, INTENT(IN)                      :: ii, kk
    REAL(KIND=dp), DIMENSION(45, 45, 3), &
      INTENT(IN)                             :: v_d
  END SUBROUTINE rot_2el_2c_first_debug
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical check_dterep_ana
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
  SUBROUTINE check_dterep_ana (sepi,sepj,r,ri,dri,se_int_control,se_taper,lgrad)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), INTENT(IN)                     :: r
    REAL(dp), DIMENSION(491), INTENT(IN)     :: ri, dri
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    LOGICAL, INTENT(IN)                      :: lgrad
    TYPE(se_taper_type), POINTER             :: se_taper
  END SUBROUTINE check_dterep_ana
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical check_rotint_ana
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
  SUBROUTINE check_rotint_ana(sepi,sepj,rijv,w,dw,se_int_control,se_taper)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), DIMENSION(3), INTENT(IN)       :: rijv
    REAL(dp), DIMENSION(2025), INTENT(IN), &
      OPTIONAL                               :: w
    REAL(dp), DIMENSION(2025, 3), &
      INTENT(IN), OPTIONAL                   :: dw
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
  END SUBROUTINE check_rotint_ana
END INTERFACE 

   PRIVATE
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_int_utils'

   PUBLIC ::   ijkl_sp, ijkl_d, rotmat, rot_2el_2c_first, store_2el_2c_diag, &
             d_ijkl_sp, d_ijkl_d

CONTAINS

! **************************************************************************************************
!> \brief General driver for computing semi-empirical integrals <ij|kl> with
!>        sp basis set. This code uses the old definitions of quadrupoles and
!>        therefore cannot be used for integrals involving d-orbitals (which
!>        require a definition of quadrupoles based on the rotational invariant
!>        property)
!>
!> \param sepi ...
!> \param sepj ...
!> \param ij ...
!> \param kl ...
!> \param li ...
!> \param lj ...
!> \param lk ...
!> \param ll ...
!> \param ic ...
!> \param r ...
!> \param se_int_control ...
!> \param se_int_screen ...
!> \param itype ...
!> \retval res ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION ijkl_sp(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_control, &
                    se_int_screen, itype) RESULT(res)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      INTEGER, INTENT(IN)                                :: ij, kl, li, lj, lk, ll, ic
      REAL(KIND=dp), INTENT(IN)                          :: r
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen
      INTEGER, INTENT(IN)                                :: itype
      REAL(KIND=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'ijkl_sp', routineP = moduleN//':'//routineN

      res = ijkl_sp_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                        se_int_control%integral_screening, se_int_control%shortrange, &
                        se_int_control%pc_coulomb_int, se_int_control%max_multipole, &
                        itype, charg_int_nri)

      ! If only the shortrange component is requested we can skip the rest
      IF ((.NOT. se_int_control%pc_coulomb_int) .AND. (itype /= do_method_pchg)) THEN
         ! Handle the 1/r^3 term, this term is ALWAYS false for KDSO-D integrals
         IF (se_int_control%shortrange .AND. se_int_control%do_ewald_r3) THEN
            res = res-ijkl_low_3(sepi, sepj, ij, kl, li, lj, lk, ll, ic, &
                                 itype, charg_int_3)
         END IF
      END IF
   END FUNCTION ijkl_sp

! **************************************************************************************************
!> \brief General driver for computing derivatives of semi-empirical integrals
!>        <ij|kl> with sp basis set.
!>        This code uses the old definitions of quadrupoles and therefore
!>        cannot be used for integrals involving d-orbitals (which requires a
!>        definition of quadrupoles based on the rotational invariant property)
!>
!> \param sepi ...
!> \param sepj ...
!> \param ij ...
!> \param kl ...
!> \param li ...
!> \param lj ...
!> \param lk ...
!> \param ll ...
!> \param ic ...
!> \param r ...
!> \param se_int_control ...
!> \param se_int_screen ...
!> \param itype ...
!> \retval res ...
!> \date 05.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION d_ijkl_sp(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_control, &
                      se_int_screen, itype) RESULT(res)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      INTEGER, INTENT(IN)                                :: ij, kl, li, lj, lk, ll, ic
      REAL(KIND=dp), INTENT(IN)                          :: r
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen
      INTEGER, INTENT(IN)                                :: itype
      REAL(KIND=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'd_ijkl_sp', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: dfs, srd

      IF (se_int_control%integral_screening == do_se_IS_kdso_d) THEN
         res = ijkl_sp_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                           se_int_control%integral_screening, .FALSE., &
                           se_int_control%pc_coulomb_int, se_int_control%max_multipole, &
                           itype, dcharg_int_nri)

         IF (.NOT. se_int_control%pc_coulomb_int) THEN
            dfs = ijkl_sp_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                              se_int_control%integral_screening, .FALSE., .FALSE., &
                              se_int_control%max_multipole, itype, dcharg_int_nri_fs)
            res = res+dfs*se_int_screen%dft

            ! In case we need the shortrange part we have to evaluate an additional derivative
            ! to handle the derivative of the Tapering term
            IF (se_int_control%shortrange) THEN
               srd = ijkl_sp_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                                 se_int_control%integral_screening, .FALSE., .TRUE., &
                                 se_int_control%max_multipole, itype, dcharg_int_nri)
               res = res-srd
            END IF
         END IF
      ELSE
         res = ijkl_sp_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                           se_int_control%integral_screening, se_int_control%shortrange, &
                           se_int_control%pc_coulomb_int, se_int_control%max_multipole, &
                           itype, dcharg_int_nri)
      END IF

      ! If only the shortrange component is requested we can skip the rest
      IF ((.NOT. se_int_control%pc_coulomb_int) .AND. (itype /= do_method_pchg)) THEN
         ! Handle the 1/r^3 term, this term is ALWAYS false for KDSO-D integrals
         IF (se_int_control%shortrange .AND. se_int_control%do_ewald_r3) THEN
            res = res-ijkl_low_3(sepi, sepj, ij, kl, li, lj, lk, ll, ic, &
                                 itype, dcharg_int_3)
         END IF
      END IF

   END FUNCTION d_ijkl_sp

! **************************************************************************************************
!> \brief Low level general driver for computing semi-empirical integrals
!>        <ij|kl> and their derivatives with sp basis set only.
!>        This code uses the old definitions of quadrupoles and
!>        therefore cannot be used for integrals involving d-orbitals (which
!>        require a definition of quadrupoles based on the rotational invariant
!>        property)
!>
!> \param sepi ...
!> \param sepj ...
!> \param ij ...
!> \param kl ...
!> \param li ...
!> \param lj ...
!> \param lk ...
!> \param ll ...
!> \param ic ...
!> \param r ...
!> \param se_int_screen ...
!> \param iscreen ...
!> \param shortrange ...
!> \param pc_coulomb_int ...
!> \param max_multipole ...
!> \param itype ...
!> \param eval a function without explicit interface
!> \retval res ...
!> \date 05.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION ijkl_sp_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                        iscreen, shortrange, pc_coulomb_int, max_multipole, itype, eval) RESULT(res)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      INTEGER, INTENT(IN)                                :: ij, kl, li, lj, lk, ll, ic
      REAL(KIND=dp), INTENT(IN)                          :: r
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen
      INTEGER, INTENT(IN)                                :: iscreen
      LOGICAL, INTENT(IN)                                :: shortrange, pc_coulomb_int
      INTEGER, INTENT(IN)                                :: max_multipole, itype
      REAL(KIND=dp)                                      :: eval, res

      CHARACTER(len=*), PARAMETER :: routineN = 'ijkl_sp_low', routineP = moduleN//':'//routineN

      INTEGER                                            :: ccc, l1, l1max, l1min, l2, l2max, l2min, &
                                                            lij, lkl, lmin, m
      REAL(KIND=dp)                                      :: add, chrg, dij, dkl, fact_ij, fact_kl, &
                                                            fact_screen, pij, pkl, s1, sum

      l1min = ABS(li-lj)
      l1max = li+lj
      lij = indexb(li+1, lj+1)
      l2min = ABS(lk-ll)
      l2max = lk+ll
      lkl = indexb(lk+1, ll+1)

      l1max = MIN(l1max, 2)
      l1min = MIN(l1min, 2)
      l2max = MIN(l2max, 2)
      l2min = MIN(l2min, 2)
      sum = 0.0_dp
      dij = 0.0_dp
      dkl = 0.0_dp
      fact_ij = 1.0_dp
      fact_kl = 1.0_dp
      fact_screen = 1.0_dp
      IF (lij == 3) fact_ij = SQRT(2.0_dp)
      IF (lkl == 3) fact_kl = SQRT(2.0_dp)
      IF (.NOT. pc_coulomb_int) THEN
         IF (iscreen == do_se_IS_kdso_d) fact_screen = se_int_screen%ft
         ! Standard value of the integral
         DO l1 = l1min, l1max
            IF (l1 == 0) THEN
               IF (lij == 1) THEN
                  pij = sepi%ko(1)
                  IF (ic == -1 .OR. ic == 1) THEN
                     pij = sepi%ko(9)
                  END IF
               ELSE IF (lij == 3) THEN
                  pij = sepi%ko(7)
               END IF
            ELSE
               dij = sepi%cs(lij)*fact_ij
               pij = sepi%ko(lij)
            END IF
            !
            DO l2 = l2min, l2max
               IF (l2 == 0) THEN
                  IF (lkl == 1) THEN
                     pkl = sepj%ko(1)
                     IF (ic == -1 .OR. ic == 2) THEN
                        pkl = sepj%ko(9)
                     END IF
                  ELSE IF (lkl == 3) THEN
                     pkl = sepj%ko(7)
                  END IF
               ELSE
                  dkl = sepj%cs(lkl)*fact_kl
                  pkl = sepj%ko(lkl)
               END IF
               IF (itype == do_method_pchg) THEN
                  add = 0.0_dp
               ELSE
                  add = (pij+pkl)**2
               END IF
               lmin = MAX(l1, l2)
               s1 = 0.0_dp
               DO m = -lmin, lmin
                  ccc = clm_sp(ij, l1, m)*clm_sp(kl, l2, m)
                  IF (ABS(ccc) > EPSILON(0.0_dp)) THEN
                     chrg = eval(r, l1, l2, clm_sp(ij, l1, m), clm_sp(kl, l2, m), dij, dkl, add, fact_screen)
                     s1 = s1+chrg
                  END IF
               END DO
               sum = sum+s1
            END DO
         END DO
         res = sum
      END IF
      ! Shortrange: Possibly computes pure Coulomb and subtract from the original integral value
      IF (shortrange .OR. pc_coulomb_int) THEN
         sum = 0.0_dp
         dij = 0.0_dp
         dkl = 0.0_dp
         add = 0.0_dp
         fact_screen = 0.0_dp
         DO l1 = l1min, l1max
            IF (l1 > max_multipole) CYCLE
            IF (l1 /= 0) THEN
               dij = sepi%cs(lij)*fact_ij
            END IF
            !
            DO l2 = l2min, l2max
               IF (l2 > max_multipole) CYCLE
               IF (l2 /= 0) THEN
                  dkl = sepj%cs(lkl)*fact_kl
               END IF
               lmin = MAX(l1, l2)
               s1 = 0.0_dp
               DO m = -lmin, lmin
                  ccc = clm_sp(ij, l1, m)*clm_sp(kl, l2, m)
                  IF (ABS(ccc) > EPSILON(0.0_dp)) THEN
                     chrg = eval(r, l1, l2, clm_sp(ij, l1, m), clm_sp(kl, l2, m), dij, dkl, add, fact_screen)
                     s1 = s1+chrg
                  END IF
               END DO
               sum = sum+s1
            END DO
         END DO
         IF (pc_coulomb_int) res = sum
         IF (shortrange) res = res-sum
      END IF
   END FUNCTION ijkl_sp_low

! **************************************************************************************************
!> \brief Interaction function between two point-charge configurations NDDO sp-code
!>        Non-Rotational Invariant definition of quadrupoles
!>        r    -  Distance r12
!>        l1,m -  Quantum numbers for multipole of configuration 1
!>        l2,m -  Quantum numbers for multipole of configuration 2
!>        da   -  charge separation of configuration 1
!>        db   -  charge separation of configuration 2
!>        add  -  additive term
!>
!> \param r ...
!> \param l1_i ...
!> \param l2_i ...
!> \param m1_i ...
!> \param m2_i ...
!> \param da_i ...
!> \param db_i ...
!> \param add0 ...
!> \param fact_screen ...
!> \retval charg ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION charg_int_nri(r, l1_i, l2_i, m1_i, m2_i, da_i, db_i, add0, fact_screen) RESULT(charg)
      REAL(KIND=dp), INTENT(in)                          :: r
      INTEGER, INTENT(in)                                :: l1_i, l2_i, m1_i, m2_i
      REAL(KIND=dp), INTENT(in)                          :: da_i, db_i, add0, fact_screen
      REAL(KIND=dp)                                      :: charg

      CHARACTER(len=*), PARAMETER :: routineN = 'charg_int_nri', routineP = moduleN//':'//routineN

      INTEGER                                            :: l1, l2, m1, m2
      REAL(KIND=dp)                                      :: add, da, db, dxdx, dxqxz, dzdz, dzqxx, &
                                                            dzqzz, fact, qqxx, qqzz, qxxqxx, &
                                                            qxxqyy, qxzqxz, xyxy, zzzz

! Computing only Integral Values

      IF (l1_i < l2_i) THEN
         l1 = l1_i
         l2 = l2_i
         m1 = m1_i
         m2 = m2_i
         da = da_i
         db = db_i
         fact = 1.0_dp
      ELSE IF (l1_i > l2_i) THEN
         l1 = l2_i
         l2 = l1_i
         m1 = m2_i
         m2 = m1_i
         da = db_i
         db = da_i
         fact = (-1.0_dp)**(l1+l2)
      ELSE IF (l1_i == l2_i) THEN
         l1 = l1_i
         l2 = l2_i
         IF (m1_i <= m2_i) THEN
            m1 = m1_i
            m2 = m2_i
            da = da_i
            db = db_i
         ELSE
            m1 = m2_i
            m2 = m1_i
            da = db_i
            db = da_i
         END IF
         fact = 1.0_dp
      END IF
      add = add0*fact_screen
      charg = 0.0_dp
      ! Q - Q.
      IF (l1 == 0 .AND. l2 == 0) THEN
         charg = fact/SQRT(r**2+add)
         RETURN
      END IF
      ! Q - Z.
      IF (l1 == 0 .AND. l2 == 1 .AND. m2 == CLMz) THEN
         charg = 1.0_dp/SQRT((r+db)**2+add)-1.0_dp/SQRT((r-db)**2+add)
         charg = charg*0.5_dp*fact
         RETURN
      END IF
      ! Z - Z.
      IF (l1 == 1 .AND. l2 == 1 .AND. m1 == CLMz .AND. m2 == CLMz) THEN
         dzdz = &
            +1.0_dp/SQRT((r+da-db)**2+add)+1.0_dp/SQRT((r-da+db)**2+add) &
            -1.0_dp/SQRT((r-da-db)**2+add)-1.0_dp/SQRT((r+da+db)**2+add)
         charg = dzdz*0.25_dp*fact
         RETURN
      END IF
      ! X - X
      IF (l1 == 1 .AND. l2 == 1 .AND. m1 == CLMp .AND. m2 == CLMp) THEN
         dxdx = 2.0_dp/SQRT(r**2+(da-db)**2+add)-2.0_dp/SQRT(r**2+(da+db)**2+add)
         charg = dxdx*0.25_dp*fact
         RETURN
      END IF
      ! Q - ZZ
      IF (l1 == 0 .AND. l2 == 2 .AND. m2 == CLMzz) THEN
         qqzz = 1.0_dp/SQRT((r-db)**2+add)-2.0_dp/SQRT(r**2+add)+1.0_dp/SQRT((r+db)**2+add)
         charg = qqzz*0.25_dp*fact
         RETURN
      END IF
      ! Q - XX
      IF (l1 == 0 .AND. l2 == 2 .AND. (m2 == CLMyy .OR. m2 == CLMxx)) THEN
         qqxx = -1.0_dp/SQRT(r**2+add)+1.0_dp/SQRT(r**2+add+db**2)
         charg = qqxx*0.5_dp*fact
         RETURN
      END IF
      ! Z - ZZ
      IF (l1 == 1 .AND. l2 == 2 .AND. m1 == CLMz .AND. m2 == CLMzz) THEN
         dzqzz = &
            +1.0_dp/SQRT((r-da-db)**2+add)-2.0_dp/SQRT((r-da)**2+add) &
            +1.0_dp/SQRT((r-da+db)**2+add)-1.0_dp/SQRT((r+da-db)**2+add) &
            +2.0_dp/SQRT((r+da)**2+add)-1.0_dp/SQRT((r+da+db)**2+add)
         charg = dzqzz*0.125_dp*fact
         RETURN
      END IF
      ! Z - XX
      IF (l1 == 1 .AND. l2 == 2 .AND. m1 == CLMz .AND. (m2 == CLMyy .OR. m2 == CLMxx)) THEN
         dzqxx = &
            +1.0_dp/SQRT((r+da)**2+add)-1.0_dp/SQRT((r+da)**2+add+db**2) &
            -1.0_dp/SQRT((r-da)**2+add)+1.0_dp/SQRT((r-da)**2+add+db**2)
         charg = dzqxx*0.25_dp*fact
         RETURN
      END IF
      ! ZZ - ZZ
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMzz .AND. m2 == CLMzz) THEN
         zzzz = &
            +1.0_dp/SQRT((r-da-db)**2+add)+1.0_dp/SQRT((r+da+db)**2+add) &
            +1.0_dp/SQRT((r-da+db)**2+add)+1.0_dp/SQRT((r+da-db)**2+add)
         xyxy = &
            +1.0_dp/SQRT((r-da)**2+add)+1.0_dp/SQRT((r+da)**2+add) &
            +1.0_dp/SQRT((r-db)**2+add)+1.0_dp/SQRT((r+db)**2+add) &
            -2.0_dp/SQRT(r**2+add)
         charg = (zzzz*0.0625_dp-xyxy*0.125_dp)*fact
         RETURN
      END IF
      ! ZZ - XX
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMzz .AND. (m2 == CLMxx .OR. m2 == CLMyy)) THEN
         zzzz = &
            -1.0_dp/SQRT((r+da)**2+add)+1.0_dp/SQRT((r+da)**2+db**2+add) &
            -1.0_dp/SQRT((r-da)**2+add)+1.0_dp/SQRT((r-da)**2+db**2+add)
         xyxy = &
            +1.0_dp/SQRT(r**2+db**2+add)-1.0_dp/SQRT(r**2+add)
         charg = (zzzz*0.125_dp-xyxy*0.25_dp)*fact
         RETURN
      END IF
      ! X - ZX
      IF (l1 == 1 .AND. l2 == 2 .AND. m1 == CLMp .AND. m2 == CLMzp) THEN
         db = db/2.0_dp
         dxqxz = &
            -1.0_dp/SQRT((r-db)**2+(da-db)**2+add)+1.0_dp/SQRT((r+db)**2+(da-db)**2+add) &
            +1.0_dp/SQRT((r-db)**2+(da+db)**2+add)-1.0_dp/SQRT((r+db)**2+(da+db)**2+add)
         charg = dxqxz*0.25_dp*fact
         RETURN
      END IF
      ! ZX - ZX
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMzp .AND. m2 == CLMzp) THEN
         da = da/2.0_dp
         db = db/2.0_dp
         qxzqxz = &
            +1.0_dp/SQRT((r+da-db)**2+(da-db)**2+add)-1.0_dp/SQRT((r+da+db)**2+(da-db)**2+add) &
            -1.0_dp/SQRT((r-da-db)**2+(da-db)**2+add)+1.0_dp/SQRT((r-da+db)**2+(da-db)**2+add) &
            -1.0_dp/SQRT((r+da-db)**2+(da+db)**2+add)+1.0_dp/SQRT((r+da+db)**2+(da+db)**2+add) &
            +1.0_dp/SQRT((r-da-db)**2+(da+db)**2+add)-1.0_dp/SQRT((r-da+db)**2+(da+db)**2+add)
         charg = qxzqxz*0.125_dp*fact
         RETURN
      END IF
      ! XX - XX
      IF (l1 == 2 .AND. l2 == 2 .AND. (((m1 == CLMyy) .AND. (m2 == CLMyy)) .OR. ((m1 == CLMxx) .AND. (m2 == CLMxx)))) THEN
         qxxqxx = &
            +2.0_dp/SQRT(r**2+add)+1.0_dp/SQRT(r**2+(da-db)**2+add) &
            +1.0_dp/SQRT(r**2+(da+db)**2+add)-2.0_dp/SQRT(r**2+da**2+add) &
            -2.0_dp/SQRT(r**2+db**2+add)
         charg = qxxqxx*0.125_dp*fact
         RETURN
      END IF
      ! XX - YY
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMyy .AND. m2 == CLMxx) THEN
         qxxqyy = &
            +1.0_dp/SQRT(r**2+add)-1.0_dp/SQRT(r**2+da**2+add) &
            -1.0_dp/SQRT(r**2+db**2+add)+1.0_dp/SQRT(r**2+da**2+db**2+add)
         charg = qxxqyy*0.25_dp*fact
         RETURN
      END IF
      ! XY - XY
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMxy .AND. m2 == CLMxy) THEN
         qxxqxx = &
            +2.0_dp/SQRT(r**2+add)+1.0_dp/SQRT(r**2+(da-db)**2+add) &
            +1.0_dp/SQRT(r**2+(da+db)**2+add)-2.0_dp/SQRT(r**2+da**2+add) &
            -2.0_dp/SQRT(r**2+db**2+add)
         qxxqyy = &
            +1.0_dp/SQRT(r**2+add)-1.0_dp/SQRT(r**2+da**2+add) &
            -1.0_dp/SQRT(r**2+db**2+add)+1.0_dp/SQRT(r**2+da**2+db**2+add)
         charg = 0.5_dp*(qxxqxx*0.125_dp-qxxqyy*0.25_dp)*fact
         RETURN
      END IF
      ! We should NEVER reach this point
      CALL cp__b("semi_empirical_int_utils.F",519,"")
   END FUNCTION charg_int_nri

! **************************************************************************************************
!> \brief Derivatives of interaction function between two point-charge
!>        configurations NDDO sp-code.
!>        Non-Rotational Invariant definition of quadrupoles
!>
!>        r    -  Distance r12
!>        l1,m -  Quantum numbers for multipole of configuration 1
!>        l2,m -  Quantum numbers for multipole of configuration 2
!>        da   -  charge separation of configuration 1
!>        db   -  charge separation of configuration 2
!>        add  -  additive term
!>
!> \param r ...
!> \param l1_i ...
!> \param l2_i ...
!> \param m1_i ...
!> \param m2_i ...
!> \param da_i ...
!> \param db_i ...
!> \param add0 ...
!> \param fact_screen ...
!> \retval charg ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION dcharg_int_nri(r, l1_i, l2_i, m1_i, m2_i, da_i, db_i, add0, fact_screen) RESULT(charg)
      REAL(KIND=dp), INTENT(in)                          :: r
      INTEGER, INTENT(in)                                :: l1_i, l2_i, m1_i, m2_i
      REAL(KIND=dp), INTENT(in)                          :: da_i, db_i, add0, fact_screen
      REAL(KIND=dp)                                      :: charg

      CHARACTER(len=*), PARAMETER :: routineN = 'dcharg_int_nri', routineP = moduleN//':'//routineN

      INTEGER                                            :: l1, l2, m1, m2
      REAL(KIND=dp)                                      :: add, da, db, dxdx, dxqxz, dzdz, dzqxx, &
                                                            dzqzz, fact, qqxx, qqzz, qxxqxx, &
                                                            qxxqyy, qxzqxz, xyxy, zzzz

! Computing only Integral Derivatives

      IF (l1_i < l2_i) THEN
         l1 = l1_i
         l2 = l2_i
         m1 = m1_i
         m2 = m2_i
         da = da_i
         db = db_i
         fact = 1.0_dp
      ELSE IF (l1_i > l2_i) THEN
         l1 = l2_i
         l2 = l1_i
         m1 = m2_i
         m2 = m1_i
         da = db_i
         db = da_i
         fact = (-1.0_dp)**(l1+l2)
      ELSE IF (l1_i == l2_i) THEN
         l1 = l1_i
         l2 = l2_i
         IF (m1_i <= m2_i) THEN
            m1 = m1_i
            m2 = m2_i
            da = da_i
            db = db_i
         ELSE
            m1 = m2_i
            m2 = m1_i
            da = db_i
            db = da_i
         END IF
         fact = 1.0_dp
      END IF
      charg = 0.0_dp
      add = add0*fact_screen
      ! Q - Q.
      IF (l1 == 0 .AND. l2 == 0) THEN
         charg = r/SQRT(r**2+add)**3
         charg = -charg*fact
         RETURN
      END IF
      ! Q - Z.
      IF (l1 == 0 .AND. l2 == 1 .AND. m2 == CLMz) THEN
         charg = (r+db)/SQRT((r+db)**2+add)**3-(r-db)/SQRT((r-db)**2+add)**3
         charg = -charg*0.5_dp*fact
         RETURN
      END IF
      ! Z - Z.
      IF (l1 == 1 .AND. l2 == 1 .AND. m1 == CLMz .AND. m2 == CLMz) THEN
         dzdz = &
            +(r+da-db)/SQRT((r+da-db)**2+add)**3+(r-da+db)/SQRT((r-da+db)**2+add)**3 &
            -(r-da-db)/SQRT((r-da-db)**2+add)**3-(r+da+db)/SQRT((r+da+db)**2+add)**3
         charg = -dzdz*0.25_dp*fact
         RETURN
      END IF
      ! X - X
      IF (l1 == 1 .AND. l2 == 1 .AND. m1 == CLMp .AND. m2 == CLMp) THEN
         dxdx = 2.0_dp*r/SQRT(r**2+(da-db)**2+add)**3-2.0_dp*r/SQRT(r**2+(da+db)**2+add)**3
         charg = -dxdx*0.25_dp*fact
         RETURN
      END IF
      ! Q - ZZ
      IF (l1 == 0 .AND. l2 == 2 .AND. m2 == CLMzz) THEN
         qqzz = (r-db)/SQRT((r-db)**2+add)**3-2.0_dp*r/SQRT(r**2+add)**3+(r+db)/SQRT((r+db)**2+add)**3
         charg = -qqzz*0.25_dp*fact
         RETURN
      END IF
      ! Q - XX
      IF (l1 == 0 .AND. l2 == 2 .AND. (m2 == CLMyy .OR. m2 == CLMxx)) THEN
         qqxx = -r/SQRT(r**2+add)**3+r/SQRT(r**2+add+db**2)**3
         charg = -qqxx*0.5_dp*fact
         RETURN
      END IF
      ! Z - ZZ
      IF (l1 == 1 .AND. l2 == 2 .AND. m1 == CLMz .AND. m2 == CLMzz) THEN
         dzqzz = &
            +(r-da-db)/SQRT((r-da-db)**2+add)**3-2.0_dp*(r-da)/SQRT((r-da)**2+add)**3 &
            +(r-da+db)/SQRT((r-da+db)**2+add)**3-(r+da-db)/SQRT((r+da-db)**2+add)**3 &
            +2.0_dp*(r+da)/SQRT((r+da)**2+add)**3-(r+da+db)/SQRT((r+da+db)**2+add)**3
         charg = -dzqzz*0.125_dp*fact
         RETURN
      END IF
      ! Z - XX
      IF (l1 == 1 .AND. l2 == 2 .AND. m1 == CLMz .AND. (m2 == CLMyy .OR. m2 == CLMxx)) THEN
         dzqxx = &
            +(r+da)/SQRT((r+da)**2+add)**3-(r+da)/SQRT((r+da)**2+add+db**2)**3 &
            -(r-da)/SQRT((r-da)**2+add)**3+(r-da)/SQRT((r-da)**2+add+db**2)**3
         charg = -dzqxx*0.25_dp*fact
         RETURN
      END IF
      ! ZZ - ZZ
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMzz .AND. m2 == CLMzz) THEN
         zzzz = &
            +(r-da-db)/SQRT((r-da-db)**2+add)**3+(r+da+db)/SQRT((r+da+db)**2+add)**3 &
            +(r-da+db)/SQRT((r-da+db)**2+add)**3+(r+da-db)/SQRT((r+da-db)**2+add)**3
         xyxy = &
            +(r-da)/SQRT((r-da)**2+add)**3+(r+da)/SQRT((r+da)**2+add)**3 &
            +(r-db)/SQRT((r-db)**2+add)**3+(r+db)/SQRT((r+db)**2+add)**3 &
            -2.0_dp*r/SQRT(r**2+add)**3
         charg = -(zzzz*0.0625_dp-xyxy*0.125_dp)*fact
         RETURN
      END IF
      ! ZZ - XX
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMzz .AND. (m2 == CLMxx .OR. m2 == CLMyy)) THEN
         zzzz = &
            -(r+da)/SQRT((r+da)**2+add)**3+(r+da)/SQRT((r+da)**2+db**2+add)**3 &
            -(r-da)/SQRT((r-da)**2+add)**3+(r-da)/SQRT((r-da)**2+db**2+add)**3
         xyxy = r/SQRT(r**2+db**2+add)**3-r/SQRT(r**2+add)**3
         charg = -(zzzz*0.125_dp-xyxy*0.25_dp)*fact
         RETURN
      END IF
      ! X - ZX
      IF (l1 == 1 .AND. l2 == 2 .AND. m1 == CLMp .AND. m2 == CLMzp) THEN
         db = db/2.0_dp
         dxqxz = &
            -(r-db)/SQRT((r-db)**2+(da-db)**2+add)**3+(r+db)/SQRT((r+db)**2+(da-db)**2+add)**3 &
            +(r-db)/SQRT((r-db)**2+(da+db)**2+add)**3-(r+db)/SQRT((r+db)**2+(da+db)**2+add)**3
         charg = -dxqxz*0.25_dp*fact
         RETURN
      END IF
      ! ZX - ZX
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMzp .AND. m2 == CLMzp) THEN
         da = da/2.0_dp
         db = db/2.0_dp
         qxzqxz = &
            +(r+da-db)/SQRT((r+da-db)**2+(da-db)**2+add)**3-(r+da+db)/SQRT((r+da+db)**2+(da-db)**2+add)**3 &
            -(r-da-db)/SQRT((r-da-db)**2+(da-db)**2+add)**3+(r-da+db)/SQRT((r-da+db)**2+(da-db)**2+add)**3 &
            -(r+da-db)/SQRT((r+da-db)**2+(da+db)**2+add)**3+(r+da+db)/SQRT((r+da+db)**2+(da+db)**2+add)**3 &
            +(r-da-db)/SQRT((r-da-db)**2+(da+db)**2+add)**3-(r-da+db)/SQRT((r-da+db)**2+(da+db)**2+add)**3
         charg = -qxzqxz*0.125_dp*fact
         RETURN
      END IF
      ! XX - XX
      IF (l1 == 2 .AND. l2 == 2 .AND. (((m1 == CLMyy) .AND. (m2 == CLMyy)) .OR. ((m1 == CLMxx) .AND. (m2 == CLMxx)))) THEN
         qxxqxx = &
            +2.0_dp*r/SQRT(r**2+add)**3+r/SQRT(r**2+(da-db)**2+add)**3 &
            +r/SQRT(r**2+(da+db)**2+add)**3-2.0_dp*r/SQRT(r**2+da**2+add)**3 &
            -2.0_dp*r/SQRT(r**2+db**2+add)**3
         charg = -qxxqxx*0.125_dp*fact
         RETURN
      END IF
      ! XX - YY
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMyy .AND. m2 == CLMxx) THEN
         qxxqyy = &
            +r/SQRT(r**2+add)**3-r/SQRT(r**2+da**2+add)**3 &
            -r/SQRT(r**2+db**2+add)**3+r/SQRT(r**2+da**2+db**2+add)**3
         charg = -qxxqyy*0.25_dp*fact
         RETURN
      END IF
      ! XY - XY
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMxy .AND. m2 == CLMxy) THEN
         qxxqxx = &
            +2.0_dp*r/SQRT(r**2+add)**3+r/SQRT(r**2+(da-db)**2+add)**3 &
            +r/SQRT(r**2+(da+db)**2+add)**3-2.0_dp*r/SQRT(r**2+da**2+add)**3 &
            -2.0_dp*r/SQRT(r**2+db**2+add)**3
         qxxqyy = &
            +r/SQRT(r**2+add)**3-r/SQRT(r**2+da**2+add)**3 &
            -r/SQRT(r**2+db**2+add)**3+r/SQRT(r**2+da**2+db**2+add)**3
         charg = -0.5_dp*(qxxqxx*0.125_dp-qxxqyy*0.25_dp)*fact
         RETURN
      END IF
      ! We should NEVER reach this point
      CALL cp__b("semi_empirical_int_utils.F",723,"")
   END FUNCTION dcharg_int_nri

! **************************************************************************************************
!> \brief Derivatives of interaction function between two point-charge
!>        configurations NDDO sp-code. The derivative takes care of the screening
!>        term only.
!>        Non-Rotational Invariant definition of quadrupoles
!>
!>        r    -  Distance r12
!>        l1,m -  Quantum numbers for multipole of configuration 1
!>        l2,m -  Quantum numbers for multipole of configuration 2
!>        da   -  charge separation of configuration 1
!>        db   -  charge separation of configuration 2
!>        add  -  additive term
!>
!> \param r ...
!> \param l1_i ...
!> \param l2_i ...
!> \param m1_i ...
!> \param m2_i ...
!> \param da_i ...
!> \param db_i ...
!> \param add0 ...
!> \param fact_screen ...
!> \retval charg ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION dcharg_int_nri_fs(r, l1_i, l2_i, m1_i, m2_i, da_i, db_i, add0, fact_screen) RESULT(charg)
      REAL(KIND=dp), INTENT(in)                          :: r
      INTEGER, INTENT(in)                                :: l1_i, l2_i, m1_i, m2_i
      REAL(KIND=dp), INTENT(in)                          :: da_i, db_i, add0, fact_screen
      REAL(KIND=dp)                                      :: charg

      CHARACTER(len=*), PARAMETER :: routineN = 'dcharg_int_nri_fs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: l1, l2, m1, m2
      REAL(KIND=dp)                                      :: add, da, db, dxdx, dxqxz, dzdz, dzqxx, &
                                                            dzqzz, fact, qqxx, qqzz, qxxqxx, &
                                                            qxxqyy, qxzqxz, xyxy, zzzz

! Computing only Integral Derivatives

      IF (l1_i < l2_i) THEN
         l1 = l1_i
         l2 = l2_i
         m1 = m1_i
         m2 = m2_i
         da = da_i
         db = db_i
         fact = 1.0_dp
      ELSE IF (l1_i > l2_i) THEN
         l1 = l2_i
         l2 = l1_i
         m1 = m2_i
         m2 = m1_i
         da = db_i
         db = da_i
         fact = (-1.0_dp)**(l1+l2)
      ELSE IF (l1_i == l2_i) THEN
         l1 = l1_i
         l2 = l2_i
         IF (m1_i <= m2_i) THEN
            m1 = m1_i
            m2 = m2_i
            da = da_i
            db = db_i
         ELSE
            m1 = m2_i
            m2 = m1_i
            da = db_i
            db = da_i
         END IF
         fact = 1.0_dp
      END IF
      charg = 0.0_dp
      add = add0*fact_screen
      ! The 0.5 factor handles the derivative of the SQRT
      fact = fact*0.5_dp
      ! Q - Q.
      IF (l1 == 0 .AND. l2 == 0) THEN
         charg = add0/SQRT(r**2+add)**3
         charg = -charg*fact
         RETURN
      END IF
      ! Q - Z.
      IF (l1 == 0 .AND. l2 == 1 .AND. m2 == CLMz) THEN
         charg = add0/SQRT((r+db)**2+add)**3-add0/SQRT((r-db)**2+add)**3
         charg = -charg*0.5_dp*fact
         RETURN
      END IF
      ! Z - Z.
      IF (l1 == 1 .AND. l2 == 1 .AND. m1 == CLMz .AND. m2 == CLMz) THEN
         dzdz = &
            +add0/SQRT((r+da-db)**2+add)**3+add0/SQRT((r-da+db)**2+add)**3 &
            -add0/SQRT((r-da-db)**2+add)**3-add0/SQRT((r+da+db)**2+add)**3
         charg = -dzdz*0.25_dp*fact
         RETURN
      END IF
      ! X - X
      IF (l1 == 1 .AND. l2 == 1 .AND. m1 == CLMp .AND. m2 == CLMp) THEN
         dxdx = 2.0_dp*add0/SQRT(r**2+(da-db)**2+add)**3-2.0_dp*add0/SQRT(r**2+(da+db)**2+add)**3
         charg = -dxdx*0.25_dp*fact
         RETURN
      END IF
      ! Q - ZZ
      IF (l1 == 0 .AND. l2 == 2 .AND. m2 == CLMzz) THEN
         qqzz = add0/SQRT((r-db)**2+add)**3-2.0_dp*add0/SQRT(r**2+add)**3+add0/SQRT((r+db)**2+add)**3
         charg = -qqzz*0.25_dp*fact
         RETURN
      END IF
      ! Q - XX
      IF (l1 == 0 .AND. l2 == 2 .AND. (m2 == CLMyy .OR. m2 == CLMxx)) THEN
         qqxx = -add0/SQRT(r**2+add)**3+add0/SQRT(r**2+add+db**2)**3
         charg = -qqxx*0.5_dp*fact
         RETURN
      END IF
      ! Z - ZZ
      IF (l1 == 1 .AND. l2 == 2 .AND. m1 == CLMz .AND. m2 == CLMzz) THEN
         dzqzz = &
            +add0/SQRT((r-da-db)**2+add)**3-2.0_dp*add0/SQRT((r-da)**2+add)**3 &
            +add0/SQRT((r-da+db)**2+add)**3-add0/SQRT((r+da-db)**2+add)**3 &
            +2.0_dp*add0/SQRT((r+da)**2+add)**3-add0/SQRT((r+da+db)**2+add)**3
         charg = -dzqzz*0.125_dp*fact
         RETURN
      END IF
      ! Z - XX
      IF (l1 == 1 .AND. l2 == 2 .AND. m1 == CLMz .AND. (m2 == CLMyy .OR. m2 == CLMxx)) THEN
         dzqxx = &
            +add0/SQRT((r+da)**2+add)**3-add0/SQRT((r+da)**2+add+db**2)**3 &
            -add0/SQRT((r-da)**2+add)**3+add0/SQRT((r-da)**2+add+db**2)**3
         charg = -dzqxx*0.25_dp*fact
         RETURN
      END IF
      ! ZZ - ZZ
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMzz .AND. m2 == CLMzz) THEN
         zzzz = &
            +add0/SQRT((r-da-db)**2+add)**3+add0/SQRT((r+da+db)**2+add)**3 &
            +add0/SQRT((r-da+db)**2+add)**3+add0/SQRT((r+da-db)**2+add)**3
         xyxy = &
            +add0/SQRT((r-da)**2+add)**3+add0/SQRT((r+da)**2+add)**3 &
            +add0/SQRT((r-db)**2+add)**3+add0/SQRT((r+db)**2+add)**3 &
            -2.0_dp*add0/SQRT(r**2+add)**3
         charg = -(zzzz*0.0625_dp-xyxy*0.125_dp)*fact
         RETURN
      END IF
      ! ZZ - XX
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMzz .AND. (m2 == CLMxx .OR. m2 == CLMyy)) THEN
         zzzz = &
            -add0/SQRT((r+da)**2+add)**3+add0/SQRT((r+da)**2+db**2+add)**3 &
            -add0/SQRT((r-da)**2+add)**3+add0/SQRT((r-da)**2+db**2+add)**3
         xyxy = add0/SQRT(r**2+db**2+add)**3-add0/SQRT(r**2+add)**3
         charg = -(zzzz*0.125_dp-xyxy*0.25_dp)*fact
         RETURN
      END IF
      ! X - ZX
      IF (l1 == 1 .AND. l2 == 2 .AND. m1 == CLMp .AND. m2 == CLMzp) THEN
         db = db/2.0_dp
         dxqxz = &
            -add0/SQRT((r-db)**2+(da-db)**2+add)**3+add0/SQRT((r+db)**2+(da-db)**2+add)**3 &
            +add0/SQRT((r-db)**2+(da+db)**2+add)**3-add0/SQRT((r+db)**2+(da+db)**2+add)**3
         charg = -dxqxz*0.25_dp*fact
         RETURN
      END IF
      ! ZX - ZX
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMzp .AND. m2 == CLMzp) THEN
         da = da/2.0_dp
         db = db/2.0_dp
         qxzqxz = &
            +add0/SQRT((r+da-db)**2+(da-db)**2+add)**3-add0/SQRT((r+da+db)**2+(da-db)**2+add)**3 &
            -add0/SQRT((r-da-db)**2+(da-db)**2+add)**3+add0/SQRT((r-da+db)**2+(da-db)**2+add)**3 &
            -add0/SQRT((r+da-db)**2+(da+db)**2+add)**3+add0/SQRT((r+da+db)**2+(da+db)**2+add)**3 &
            +add0/SQRT((r-da-db)**2+(da+db)**2+add)**3-add0/SQRT((r-da+db)**2+(da+db)**2+add)**3
         charg = -qxzqxz*0.125_dp*fact
         RETURN
      END IF
      ! XX - XX
      IF (l1 == 2 .AND. l2 == 2 .AND. (((m1 == CLMyy) .AND. (m2 == CLMyy)) .OR. ((m1 == CLMxx) .AND. (m2 == CLMxx)))) THEN
         qxxqxx = &
            +2.0_dp*add0/SQRT(r**2+add)**3+add0/SQRT(r**2+(da-db)**2+add)**3 &
            +add0/SQRT(r**2+(da+db)**2+add)**3-2.0_dp*add0/SQRT(r**2+da**2+add)**3 &
            -2.0_dp*add0/SQRT(r**2+db**2+add)**3
         charg = -qxxqxx*0.125_dp*fact
         RETURN
      END IF
      ! XX - YY
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMyy .AND. m2 == CLMxx) THEN
         qxxqyy = &
            +add0/SQRT(r**2+add)**3-add0/SQRT(r**2+da**2+add)**3 &
            -add0/SQRT(r**2+db**2+add)**3+add0/SQRT(r**2+da**2+db**2+add)**3
         charg = -qxxqyy*0.25_dp*fact
         RETURN
      END IF
      ! XY - XY
      IF (l1 == 2 .AND. l2 == 2 .AND. m1 == CLMxy .AND. m2 == CLMxy) THEN
         qxxqxx = &
            +2.0_dp*add0/SQRT(r**2+add)**3+add0/SQRT(r**2+(da-db)**2+add)**3 &
            +add0/SQRT(r**2+(da+db)**2+add)**3-2.0_dp*add0/SQRT(r**2+da**2+add)**3 &
            -2.0_dp*add0/SQRT(r**2+db**2+add)**3
         qxxqyy = &
            +add0/SQRT(r**2+add)**3-add0/SQRT(r**2+da**2+add)**3 &
            -add0/SQRT(r**2+db**2+add)**3+add0/SQRT(r**2+da**2+db**2+add)**3
         charg = -0.5_dp*(qxxqxx*0.125_dp-qxxqyy*0.25_dp)*fact
         RETURN
      END IF
      ! We should NEVER reach this point
      CALL cp__b("semi_empirical_int_utils.F",931,"")
   END FUNCTION dcharg_int_nri_fs

! **************************************************************************************************
!> \brief General driver for computing semi-empirical integrals <ij|kl>
!>        involving d-orbitals.
!>        The choice of the linear quadrupole was REALLY unhappy
!>        in the first development of the NDDO codes. That choice makes
!>        impossible the merging of the integral code with or without d-orbitals
!>        unless a reparametrization of all NDDO codes for s and p orbitals will
!>        be performed.. more over the choice of the linear quadrupole does not make
!>        calculations rotational invariants (of course the rotational invariant
!>        can be forced). The definitions of quadrupoles for d-orbitals is the
!>        correct one in order to have the rotational invariant property by
!>        construction..
!>
!> \param sepi ...
!> \param sepj ...
!> \param ij ...
!> \param kl ...
!> \param li ...
!> \param lj ...
!> \param lk ...
!> \param ll ...
!> \param ic ...
!> \param r ...
!> \param se_int_control ...
!> \param se_int_screen ...
!> \param itype ...
!> \retval res ...
!> \date 03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION ijkl_d(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_control, &
                   se_int_screen, itype) RESULT(res)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      INTEGER, INTENT(IN)                                :: ij, kl, li, lj, lk, ll, ic
      REAL(KIND=dp), INTENT(IN)                          :: r
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen
      INTEGER, INTENT(IN)                                :: itype
      REAL(KIND=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'ijkl_d', routineP = moduleN//':'//routineN

      res = ijkl_d_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                       se_int_control%integral_screening, se_int_control%shortrange, &
                       se_int_control%pc_coulomb_int, se_int_control%max_multipole, &
                       itype, charg_int_ri)

      ! If only the shortrange component is requested we can skip the rest
      IF ((.NOT. se_int_control%pc_coulomb_int) .AND. (itype /= do_method_pchg)) THEN
         ! Handle the 1/r^3 term, this term is ALWAYS false for KDSO-D integrals
         IF (se_int_control%shortrange .AND. se_int_control%do_ewald_r3) THEN
            res = res-ijkl_low_3(sepi, sepj, ij, kl, li, lj, lk, ll, ic, &
                                 itype, charg_int_3)
         END IF
      END IF
   END FUNCTION ijkl_d

! **************************************************************************************************
!> \brief General driver for computing the derivatives of semi-empirical integrals <ij|kl>
!>        involving d-orbitals.
!>        The choice of the linear quadrupole was REALLY unhappy
!>        in the first development of the NDDO codes. That choice makes
!>        impossible the merging of the integral code with or without d-orbitals
!>        unless a reparametrization of all NDDO codes for s and p orbitals will
!>        be performed.. more over the choice of the linear quadrupole does not make
!>        calculations rotational invariants (of course the rotational invariant
!>        can be forced). The definitions of quadrupoles for d-orbitals is the
!>        correct one in order to have the rotational invariant property by
!>        construction..
!>
!> \param sepi ...
!> \param sepj ...
!> \param ij ...
!> \param kl ...
!> \param li ...
!> \param lj ...
!> \param lk ...
!> \param ll ...
!> \param ic ...
!> \param r ...
!> \param se_int_control ...
!> \param se_int_screen ...
!> \param itype ...
!> \retval res ...
!> \date 03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION d_ijkl_d(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_control, &
                     se_int_screen, itype) RESULT(res)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      INTEGER, INTENT(IN)                                :: ij, kl, li, lj, lk, ll, ic
      REAL(KIND=dp), INTENT(IN)                          :: r
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen
      INTEGER, INTENT(IN)                                :: itype
      REAL(KIND=dp)                                      :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'd_ijkl_d', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: dfs, srd

      IF (se_int_control%integral_screening == do_se_IS_kdso_d) THEN
         res = ijkl_d_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                          se_int_control%integral_screening, .FALSE., &
                          se_int_control%pc_coulomb_int, se_int_control%max_multipole, &
                          itype, dcharg_int_ri)

         IF (.NOT. se_int_control%pc_coulomb_int) THEN
            dfs = ijkl_d_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                             se_int_control%integral_screening, .FALSE., .FALSE., &
                             se_int_control%max_multipole, itype, dcharg_int_ri_fs)
            res = res+dfs*se_int_screen%dft

            ! In case we need the shortrange part we have to evaluate an additional derivative
            ! to handle the derivative of the Tapering term
            IF (se_int_control%shortrange) THEN
               srd = ijkl_d_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                                se_int_control%integral_screening, .FALSE., .TRUE., &
                                se_int_control%max_multipole, itype, dcharg_int_ri)
               res = res-srd
            END IF
         END IF
      ELSE
         res = ijkl_d_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                          se_int_control%integral_screening, se_int_control%shortrange, &
                          se_int_control%pc_coulomb_int, se_int_control%max_multipole, &
                          itype, dcharg_int_ri)
      END IF

      ! If only the shortrange component is requested we can skip the rest
      IF ((.NOT. se_int_control%pc_coulomb_int) .AND. (itype /= do_method_pchg)) THEN
         ! Handle the 1/r^3 term, this term is ALWAYS false for KDSO-D integrals
         IF (se_int_control%shortrange .AND. se_int_control%do_ewald_r3) THEN
            res = res-ijkl_low_3(sepi, sepj, ij, kl, li, lj, lk, ll, ic, &
                                 itype, dcharg_int_3)
         END IF
      END IF

   END FUNCTION d_ijkl_d

! **************************************************************************************************
!> \brief Low level general driver for computing semi-empirical integrals <ij|kl>
!>        and their derivatives involving d-orbitals.
!>        The choice of the linear quadrupole was REALLY unhappy
!>        in the first development of the NDDO codes. That choice makes
!>        impossible the merging of the integral code with or without d-orbitals
!>        unless a reparametrization of all NDDO codes for s and p orbitals will
!>        be performed.. more over the choice of the linear quadrupole does not make
!>        calculations rotational invariants (of course the rotational invariant
!>        can be forced). The definitions of quadrupoles for d-orbitals is the
!>        correct one in order to have the rotational invariant property by
!>        construction..
!>
!> \param sepi ...
!> \param sepj ...
!> \param ij ...
!> \param kl ...
!> \param li ...
!> \param lj ...
!> \param lk ...
!> \param ll ...
!> \param ic ...
!> \param r ...
!> \param se_int_screen ...
!> \param iscreen ...
!> \param shortrange ...
!> \param pc_coulomb_int ...
!> \param max_multipole ...
!> \param itype ...
!> \param eval a function without explicit interface
!> \retval res ...
!> \date 03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION ijkl_d_low(sepi, sepj, ij, kl, li, lj, lk, ll, ic, r, se_int_screen, &
                       iscreen, shortrange, pc_coulomb_int, max_multipole, itype, eval) RESULT(res)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      INTEGER, INTENT(IN)                                :: ij, kl, li, lj, lk, ll, ic
      REAL(KIND=dp), INTENT(IN)                          :: r
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen
      INTEGER, INTENT(IN)                                :: iscreen
      LOGICAL, INTENT(IN)                                :: shortrange, pc_coulomb_int
      INTEGER, INTENT(IN)                                :: max_multipole, itype
      REAL(KIND=dp)                                      :: eval, res

      CHARACTER(len=*), PARAMETER :: routineN = 'ijkl_d_low', routineP = moduleN//':'//routineN

      INTEGER                                            :: l1, l1max, l1min, l2, l2max, l2min, lij, &
                                                            lkl, lmin, m, mm
      REAL(KIND=dp)                                      :: add, ccc, chrg, dij, dkl, fact_screen, &
                                                            pij, pkl, s1, sum

      l1min = ABS(li-lj)
      l1max = li+lj
      lij = indexb(li+1, lj+1)
      l2min = ABS(lk-ll)
      l2max = lk+ll
      lkl = indexb(lk+1, ll+1)

      l1max = MIN(l1max, 2)
      l1min = MIN(l1min, 2)
      l2max = MIN(l2max, 2)
      l2min = MIN(l2min, 2)
      sum = 0.0_dp
      dij = 0.0_dp
      dkl = 0.0_dp
      fact_screen = 1.0_dp
      IF (.NOT. pc_coulomb_int) THEN
         IF (iscreen == do_se_IS_kdso_d) fact_screen = se_int_screen%ft
         ! Standard value of the integral
         DO l1 = l1min, l1max
            IF (l1 == 0) THEN
               IF (lij == 1) THEN
                  pij = sepi%ko(1)
                  IF (ic == 1) THEN
                     pij = sepi%ko(9)
                  END IF
               ELSE IF (lij == 3) THEN
                  pij = sepi%ko(7)
               ELSE IF (lij == 6) THEN
                  pij = sepi%ko(8)
               END IF
            ELSE
               dij = sepi%cs(lij)
               pij = sepi%ko(lij)
            END IF
            !
            DO l2 = l2min, l2max
               IF (l2 == 0) THEN
                  IF (lkl == 1) THEN
                     pkl = sepj%ko(1)
                     IF (ic == 2) THEN
                        pkl = sepj%ko(9)
                     END IF
                  ELSE IF (lkl == 3) THEN
                     pkl = sepj%ko(7)
                  ELSE IF (lkl == 6) THEN
                     pkl = sepj%ko(8)
                  END IF
               ELSE
                  dkl = sepj%cs(lkl)
                  pkl = sepj%ko(lkl)
               END IF
               IF (itype == do_method_pchg) THEN
                  add = 0.0_dp
               ELSE
                  add = (pij+pkl)**2
               END IF
               lmin = MIN(l1, l2)
               s1 = 0.0_dp
               DO m = -lmin, lmin
                  ccc = clm_d(ij, l1, m)*clm_d(kl, l2, m)
                  IF (ABS(ccc) > EPSILON(0.0_dp)) THEN
                     mm = ABS(m)
                     chrg = eval(r, l1, l2, mm, dij, dkl, add, fact_screen)
                     s1 = s1+chrg*ccc
                  END IF
               END DO
               sum = sum+s1
            END DO
         END DO
         res = sum
      END IF
      ! Shortrange: Possibly computes pure Coulomb and subtract from the original integral valeu
      IF (shortrange .OR. pc_coulomb_int) THEN
         sum = 0.0_dp
         dij = 0.0_dp
         dkl = 0.0_dp
         add = 0.0_dp
         fact_screen = 0.0_dp
         DO l1 = l1min, l1max
            IF (l1 > max_multipole) CYCLE
            IF (l1 /= 0) THEN
               dij = sepi%cs(lij)
            END IF
            !
            DO l2 = l2min, l2max
               IF (l2 > max_multipole) CYCLE
               IF (l2 /= 0) THEN
                  dkl = sepj%cs(lkl)
               END IF
               lmin = MIN(l1, l2)
               s1 = 0.0_dp
               DO m = -lmin, lmin
                  ccc = clm_d(ij, l1, m)*clm_d(kl, l2, m)
                  IF (ABS(ccc) > EPSILON(0.0_dp)) THEN
                     mm = ABS(m)
                     chrg = eval(r, l1, l2, mm, dij, dkl, add, fact_screen)
                     s1 = s1+chrg*ccc
                  END IF
               END DO
               sum = sum+s1
            END DO
         END DO
         IF (pc_coulomb_int) res = sum
         IF (shortrange) res = res-sum
      END IF
   END FUNCTION ijkl_d_low

! **************************************************************************************************
!> \brief Interaction function between two point-charge configurations (MNDO/d)
!>        Rotational invariant property built-in in the quadrupole definition
!>        r    -  Distance r12
!>        l1,m -  Quantum numbers for multipole of configuration 1
!>        l2,m -  Quantum numbers for multipole of configuration 2
!>        da   -  charge separation of configuration 1
!>        db   -  charge separation of configuration 2
!>        add  -  additive term
!>
!> \param r ...
!> \param l1_i ...
!> \param l2_i ...
!> \param m ...
!> \param da_i ...
!> \param db_i ...
!> \param add0 ...
!> \param fact_screen ...
!> \retval charg ...
!> \date 03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION charg_int_ri(r, l1_i, l2_i, m, da_i, db_i, add0, fact_screen) RESULT(charg)
      REAL(KIND=dp), INTENT(in)                          :: r
      INTEGER, INTENT(in)                                :: l1_i, l2_i, m
      REAL(KIND=dp), INTENT(in)                          :: da_i, db_i, add0, fact_screen
      REAL(KIND=dp)                                      :: charg

      CHARACTER(len=*), PARAMETER :: routineN = 'charg_int_ri', routineP = moduleN//':'//routineN

      INTEGER                                            :: l1, l2
      REAL(KIND=dp)                                      :: aa, ab, add, da, db, dxdx, dxqxz, dzdz, &
                                                            dzqzz, fact, qqzz, qxzqxz, xyxy, zzzz

      IF (l1_i <= l2_i) THEN
         l1 = l1_i
         l2 = l2_i
         da = da_i
         db = db_i
         fact = 1.0_dp
      ELSE IF (l1_i > l2_i) THEN
         l1 = l2_i
         l2 = l1_i
         da = db_i
         db = da_i
         fact = (-1.0_dp)**(l1+l2)
      END IF
      charg = 0.0_dp
      add = add0*fact_screen
      ! Q - Q.
      IF (l1 == 0 .AND. l2 == 0) THEN
         charg = fact/SQRT(r**2+add)
         RETURN
      END IF
      ! Q - Z.
      IF (l1 == 0 .AND. l2 == 1) THEN
         charg = 1.0_dp/SQRT((r+db)**2+add)-1.0_dp/SQRT((r-db)**2+add)
         charg = charg*0.5_dp*fact
         RETURN
      END IF
      ! Z - Z.
      IF (l1 == 1 .AND. l2 == 1 .AND. m == 0) THEN
         dzdz = &
            +1.0_dp/SQRT((r+da-db)**2+add)+1.0_dp/SQRT((r-da+db)**2+add) &
            -1.0_dp/SQRT((r-da-db)**2+add)-1.0_dp/SQRT((r+da+db)**2+add)
         charg = dzdz*0.25_dp*fact
         RETURN
      END IF
      ! X - X
      IF (l1 == 1 .AND. l2 == 1 .AND. m == 1) THEN
         dxdx = 2.0_dp/SQRT(r**2+(da-db)**2+add)-2.0_dp/SQRT(r**2+(da+db)**2+add)
         charg = dxdx*0.25_dp*fact
         RETURN
      END IF
      ! Q - ZZ
      IF (l1 == 0 .AND. l2 == 2) THEN
         qqzz = 1.0_dp/SQRT((r-db)**2+add)-2.0_dp/SQRT(r**2+db**2+add)+1.0_dp/SQRT((r+db)**2+add)
         charg = qqzz*0.25_dp*fact
         RETURN
      END IF
      ! Z - ZZ
      IF (l1 == 1 .AND. l2 == 2 .AND. m == 0) THEN
         dzqzz = &
            +1.0_dp/SQRT((r-da-db)**2+add)-2.0_dp/SQRT((r-da)**2+db**2+add) &
            +1.0_dp/SQRT((r+db-da)**2+add)-1.0_dp/SQRT((r-db+da)**2+add) &
            +2.0_dp/SQRT((r+da)**2+db**2+add)-1.0_dp/SQRT((r+da+db)**2+add)
         charg = dzqzz*0.125_dp*fact
         RETURN
      END IF
      ! ZZ - ZZ
      IF (l1 == 2 .AND. l2 == 2 .AND. m == 0) THEN
         zzzz = &
            +1.0_dp/SQRT((r-da-db)**2+add)+1.0_dp/SQRT((r+da+db)**2+add) &
            +1.0_dp/SQRT((r-da+db)**2+add)+1.0_dp/SQRT((r+da-db)**2+add) &
            -2.0_dp/SQRT((r-da)**2+db**2+add)-2.0_dp/SQRT((r-db)**2+da**2+add) &
            -2.0_dp/SQRT((r+da)**2+db**2+add)-2.0_dp/SQRT((r+db)**2+da**2+add) &
            +2.0_dp/SQRT(r**2+(da-db)**2+add)+2.0_dp/SQRT(r**2+(da+db)**2+add)
         xyxy = &
            +4.0_dp/SQRT(r**2+(da-db)**2+add)+4.0_dp/SQRT(r**2+(da+db)**2+add) &
            -8.0_dp/SQRT(r**2+da**2+db**2+add)
         charg = (zzzz*0.0625_dp-xyxy*0.015625_dp)*fact
         RETURN
      END IF
      ! X - ZX
      IF (l1 == 1 .AND. l2 == 2 .AND. m == 1) THEN
         ab = db/SQRT(2.0_dp)
         dxqxz = &
            -2.0_dp/SQRT((r-ab)**2+(da-ab)**2+add)+2.0_dp/SQRT((r+ab)**2+(da-ab)**2+add) &
            +2.0_dp/SQRT((r-ab)**2+(da+ab)**2+add)-2.0_dp/SQRT((r+ab)**2+(da+ab)**2+add)
         charg = dxqxz*0.125_dp*fact
         RETURN
      END IF
      ! ZX - ZX
      IF (l1 == 2 .AND. l2 == 2 .AND. m == 1) THEN
         aa = da/SQRT(2.0_dp)
         ab = db/SQRT(2.0_dp)
         qxzqxz = &
            +2.0_dp/SQRT((r+aa-ab)**2+(aa-ab)**2+add)-2.0_dp/SQRT((r+aa+ab)**2+(aa-ab)**2+add) &
            -2.0_dp/SQRT((r-aa-ab)**2+(aa-ab)**2+add)+2.0_dp/SQRT((r-aa+ab)**2+(aa-ab)**2+add) &
            -2.0_dp/SQRT((r+aa-ab)**2+(aa+ab)**2+add)+2.0_dp/SQRT((r+aa+ab)**2+(aa+ab)**2+add) &
            +2.0_dp/SQRT((r-aa-ab)**2+(aa+ab)**2+add)-2.0_dp/SQRT((r-aa+ab)**2+(aa+ab)**2+add)
         charg = qxzqxz*0.0625_dp*fact
         RETURN
      END IF
      ! XX - XX
      IF (l1 == 2 .AND. l2 == 2 .AND. m == 2) THEN
         xyxy = 4.0_dp/SQRT(r**2+(da-db)**2+add)+4.0_dp/SQRT(r**2+(da+db)**2+add)-8.0_dp/SQRT(r**2+da**2+db**2+add)
         charg = xyxy*0.0625_dp*fact
         RETURN
      END IF
      ! We should NEVER reach this point
      CALL cp__b("semi_empirical_int_utils.F",1364,"")
   END FUNCTION charg_int_ri

! **************************************************************************************************
!> \brief Derivatives of the interaction function between two point-charge
!>        configurations (MNDO/d)
!>        Rotational invariant property built-in in the quadrupole definition
!>        r    -  Distance r12
!>        l1,m -  Quantum numbers for multipole of configuration 1
!>        l2,m -  Quantum numbers for multipole of configuration 2
!>        da   -  charge separation of configuration 1
!>        db   -  charge separation of configuration 2
!>        add  -  additive term
!>
!> \param r ...
!> \param l1_i ...
!> \param l2_i ...
!> \param m ...
!> \param da_i ...
!> \param db_i ...
!> \param add0 ...
!> \param fact_screen ...
!> \retval charg ...
!> \date 03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION dcharg_int_ri(r, l1_i, l2_i, m, da_i, db_i, add0, fact_screen) RESULT(charg)
      REAL(KIND=dp), INTENT(in)                          :: r
      INTEGER, INTENT(in)                                :: l1_i, l2_i, m
      REAL(KIND=dp), INTENT(in)                          :: da_i, db_i, add0, fact_screen
      REAL(KIND=dp)                                      :: charg

      CHARACTER(len=*), PARAMETER :: routineN = 'dcharg_int_ri', routineP = moduleN//':'//routineN

      INTEGER                                            :: l1, l2
      REAL(KIND=dp)                                      :: aa, ab, add, da, db, dxdx, dxqxz, dzdz, &
                                                            dzqzz, fact, qqzz, qxzqxz, xyxy, zzzz

      IF (l1_i <= l2_i) THEN
         l1 = l1_i
         l2 = l2_i
         da = da_i
         db = db_i
         fact = 1.0_dp
      ELSE IF (l1_i > l2_i) THEN
         l1 = l2_i
         l2 = l1_i
         da = db_i
         db = da_i
         fact = (-1.0_dp)**(l1+l2)
      END IF
      charg = 0.0_dp
      add = add0*fact_screen
      ! Q - Q.
      IF (l1 == 0 .AND. l2 == 0) THEN
         charg = r/SQRT(r**2+add)**3
         charg = -fact*charg
         RETURN
      END IF
      ! Q - Z.
      IF (l1 == 0 .AND. l2 == 1) THEN
         charg = (r+db)/SQRT((r+db)**2+add)**3-(r-db)/SQRT((r-db)**2+add)**3
         charg = -charg*0.5_dp*fact
         RETURN
      END IF
      ! Z - Z.
      IF (l1 == 1 .AND. l2 == 1 .AND. m == 0) THEN
         dzdz = &
            +(r+da-db)/SQRT((r+da-db)**2+add)**3+(r-da+db)/SQRT((r-da+db)**2+add)**3 &
            -(r-da-db)/SQRT((r-da-db)**2+add)**3-(r+da+db)/SQRT((r+da+db)**2+add)**3
         charg = -dzdz*0.25_dp*fact
         RETURN
      END IF
      ! X - X
      IF (l1 == 1 .AND. l2 == 1 .AND. m == 1) THEN
         dxdx = 2.0_dp*r/SQRT(r**2+(da-db)**2+add)**3-2.0_dp*r/SQRT(r**2+(da+db)**2+add)**3
         charg = -dxdx*0.25_dp*fact
         RETURN
      END IF
      ! Q - ZZ
      IF (l1 == 0 .AND. l2 == 2) THEN
         qqzz = (r-db)/SQRT((r-db)**2+add)**3-2.0_dp*r/SQRT(r**2+db**2+add)**3+(r+db)/SQRT((r+db)**2+add)**3
         charg = -qqzz*0.25_dp*fact
         RETURN
      END IF
      ! Z - ZZ
      IF (l1 == 1 .AND. l2 == 2 .AND. m == 0) THEN
         dzqzz = &
            +(r-da-db)/SQRT((r-da-db)**2+add)**3-2.0_dp*(r-da)/SQRT((r-da)**2+db**2+add)**3 &
            +(r+db-da)/SQRT((r+db-da)**2+add)**3-(r-db+da)/SQRT((r-db+da)**2+add)**3 &
            +2.0_dp*(r+da)/SQRT((r+da)**2+db**2+add)**3-(r+da+db)/SQRT((r+da+db)**2+add)**3
         charg = -dzqzz*0.125_dp*fact
         RETURN
      END IF
      ! ZZ - ZZ
      IF (l1 == 2 .AND. l2 == 2 .AND. m == 0) THEN
         zzzz = &
            +(r-da-db)/SQRT((r-da-db)**2+add)**3+(r+da+db)/SQRT((r+da+db)**2+add)**3 &
            +(r-da+db)/SQRT((r-da+db)**2+add)**3+(r+da-db)/SQRT((r+da-db)**2+add)**3 &
            -2.0_dp*(r-da)/SQRT((r-da)**2+db**2+add)**3-2.0_dp*(r-db)/SQRT((r-db)**2+da**2+add)**3 &
            -2.0_dp*(r+da)/SQRT((r+da)**2+db**2+add)**3-2.0_dp*(r+db)/SQRT((r+db)**2+da**2+add)**3 &
            +2.0_dp*r/SQRT(r**2+(da-db)**2+add)**3+2.0_dp*r/SQRT(r**2+(da+db)**2+add)**3
         xyxy = &
            +4.0_dp*r/SQRT(r**2+(da-db)**2+add)**3+4.0_dp*r/SQRT(r**2+(da+db)**2+add)**3 &
            -8.0_dp*r/SQRT(r**2+da**2+db**2+add)**3
         charg = -(zzzz*0.0625_dp-xyxy*0.015625_dp)*fact
         RETURN
      END IF
      ! X - ZX
      IF (l1 == 1 .AND. l2 == 2 .AND. m == 1) THEN
         ab = db/SQRT(2.0_dp)
         dxqxz = &
            -2.0_dp*(r-ab)/SQRT((r-ab)**2+(da-ab)**2+add)**3+2.0_dp*(r+ab)/SQRT((r+ab)**2+(da-ab)**2+add)**3 &
            +2.0_dp*(r-ab)/SQRT((r-ab)**2+(da+ab)**2+add)**3-2.0_dp*(r+ab)/SQRT((r+ab)**2+(da+ab)**2+add)**3
         charg = -dxqxz*0.125_dp*fact
         RETURN
      END IF
      ! ZX - ZX
      IF (l1 == 2 .AND. l2 == 2 .AND. m == 1) THEN
         aa = da/SQRT(2.0_dp)
         ab = db/SQRT(2.0_dp)
         qxzqxz = &
            +2.0_dp*(r+aa-ab)/SQRT((r+aa-ab)**2+(aa-ab)**2+add)**3-2.0_dp*(r+aa+ab)/SQRT((r+aa+ab)**2+(aa-ab)**2+add)**3 &
            -2.0_dp*(r-aa-ab)/SQRT((r-aa-ab)**2+(aa-ab)**2+add)**3+2.0_dp*(r-aa+ab)/SQRT((r-aa+ab)**2+(aa-ab)**2+add)**3 &
            -2.0_dp*(r+aa-ab)/SQRT((r+aa-ab)**2+(aa+ab)**2+add)**3+2.0_dp*(r+aa+ab)/SQRT((r+aa+ab)**2+(aa+ab)**2+add)**3 &
            +2.0_dp*(r-aa-ab)/SQRT((r-aa-ab)**2+(aa+ab)**2+add)**3-2.0_dp*(r-aa+ab)/SQRT((r-aa+ab)**2+(aa+ab)**2+add)**3
         charg = -qxzqxz*0.0625_dp*fact
         RETURN
      END IF
      ! XX - XX
      IF (l1 == 2 .AND. l2 == 2 .AND. m == 2) THEN
         xyxy = 4.0_dp*r/SQRT(r**2+(da-db)**2+add)**3+4.0_dp*r/SQRT(r**2+(da+db)**2+add)**3-8.0_dp*r/SQRT(r**2+da**2+db**2+add)**3
         charg = -xyxy*0.0625_dp*fact
         RETURN
      END IF
      ! We should NEVER reach this point
      CALL cp__b("semi_empirical_int_utils.F",1500,"")
   END FUNCTION dcharg_int_ri

! **************************************************************************************************
!> \brief Derivatives of the interaction function between two point-charge
!>        configurations (MNDO/d). This derivative takes into account only the
!>        tapering term
!>        Rotational invariant property built-in in the quadrupole definition
!>        r    -  Distance r12
!>        l1,m -  Quantum numbers for multipole of configuration 1
!>        l2,m -  Quantum numbers for multipole of configuration 2
!>        da   -  charge separation of configuration 1
!>        db   -  charge separation of configuration 2
!>        add  -  additive term
!>
!> \param r ...
!> \param l1_i ...
!> \param l2_i ...
!> \param m ...
!> \param da_i ...
!> \param db_i ...
!> \param add0 ...
!> \param fact_screen ...
!> \retval charg ...
!> \date 03.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   FUNCTION dcharg_int_ri_fs(r, l1_i, l2_i, m, da_i, db_i, add0, fact_screen) RESULT(charg)
      REAL(KIND=dp), INTENT(in)                          :: r
      INTEGER, INTENT(in)                                :: l1_i, l2_i, m
      REAL(KIND=dp), INTENT(in)                          :: da_i, db_i, add0, fact_screen
      REAL(KIND=dp)                                      :: charg

      CHARACTER(len=*), PARAMETER :: routineN = 'dcharg_int_ri_fs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: l1, l2
      REAL(KIND=dp)                                      :: aa, ab, add, da, db, dxdx, dxqxz, dzdz, &
                                                            dzqzz, fact, qqzz, qxzqxz, xyxy, zzzz

      IF (l1_i <= l2_i) THEN
         l1 = l1_i
         l2 = l2_i
         da = da_i
         db = db_i
         fact = 1.0_dp
      ELSE IF (l1_i > l2_i) THEN
         l1 = l2_i
         l2 = l1_i
         da = db_i
         db = da_i
         fact = (-1.0_dp)**(l1+l2)
      END IF
      charg = 0.0_dp
      add = add0*fact_screen
      ! The 0.5 factor handles the derivative of the SQRT
      fact = fact*0.5_dp
      ! Q - Q.
      IF (l1 == 0 .AND. l2 == 0) THEN
         charg = add0/SQRT(r**2+add)**3
         charg = -fact*charg
         RETURN
      END IF
      ! Q - Z.
      IF (l1 == 0 .AND. l2 == 1) THEN
         charg = add0/SQRT((r+db)**2+add)**3-add0/SQRT((r-db)**2+add)**3
         charg = -charg*0.5_dp*fact
         RETURN
      END IF
      ! Z - Z.
      IF (l1 == 1 .AND. l2 == 1 .AND. m == 0) THEN
         dzdz = &
            +add0/SQRT((r+da-db)**2+add)**3+add0/SQRT((r-da+db)**2+add)**3 &
            -add0/SQRT((r-da-db)**2+add)**3-add0/SQRT((r+da+db)**2+add)**3
         charg = -dzdz*0.25_dp*fact
         RETURN
      END IF
      ! X - X
      IF (l1 == 1 .AND. l2 == 1 .AND. m == 1) THEN
         dxdx = 2.0_dp*add0/SQRT(r**2+(da-db)**2+add)**3-2.0_dp*add0/SQRT(r**2+(da+db)**2+add)**3
         charg = -dxdx*0.25_dp*fact
         RETURN
      END IF
      ! Q - ZZ
      IF (l1 == 0 .AND. l2 == 2) THEN
         qqzz = add0/SQRT((r-db)**2+add)**3-2.0_dp*add0/SQRT(r**2+db**2+add)**3+add0/SQRT((r+db)**2+add)**3
         charg = -qqzz*0.25_dp*fact
         RETURN
      END IF
      ! Z - ZZ
      IF (l1 == 1 .AND. l2 == 2 .AND. m == 0) THEN
         dzqzz = &
            +add0/SQRT((r-da-db)**2+add)**3-2.0_dp*add0/SQRT((r-da)**2+db**2+add)**3 &
            +add0/SQRT((r+db-da)**2+add)**3-add0/SQRT((r-db+da)**2+add)**3 &
            +2.0_dp*add0/SQRT((r+da)**2+db**2+add)**3-add0/SQRT((r+da+db)**2+add)**3
         charg = -dzqzz*0.125_dp*fact
         RETURN
      END IF
      ! ZZ - ZZ
      IF (l1 == 2 .AND. l2 == 2 .AND. m == 0) THEN
         zzzz = &
            +add0/SQRT((r-da-db)**2+add)**3+add0/SQRT((r+da+db)**2+add)**3 &
            +add0/SQRT((r-da+db)**2+add)**3+add0/SQRT((r+da-db)**2+add)**3 &
            -2.0_dp*add0/SQRT((r-da)**2+db**2+add)**3-2.0_dp*add0/SQRT((r-db)**2+da**2+add)**3 &
            -2.0_dp*add0/SQRT((r+da)**2+db**2+add)**3-2.0_dp*add0/SQRT((r+db)**2+da**2+add)**3 &
            +2.0_dp*add0/SQRT(r**2+(da-db)**2+add)**3+2.0_dp*add0/SQRT(r**2+(da+db)**2+add)**3
         xyxy = &
            +4.0_dp*add0/SQRT(r**2+(da-db)**2+add)**3+4.0_dp*add0/SQRT(r**2+(da+db)**2+add)**3 &
            -8.0_dp*add0/SQRT(r**2+da**2+db**2+add)**3
         charg = -(zzzz*0.0625_dp-xyxy*0.015625_dp)*fact
         RETURN
      END IF
      ! X - ZX
      IF (l1 == 1 .AND. l2 == 2 .AND. m == 1) THEN
         ab = db/SQRT(2.0_dp)
         dxqxz = &
            -2.0_dp*add0/SQRT((r-ab)**2+(da-ab)**2+add)**3+2.0_dp*add0/SQRT((r+ab)**2+(da-ab)**2+add)**3 &
            +2.0_dp*add0/SQRT((r-ab)**2+(da+ab)**2+add)**3-2.0_dp*add0/SQRT((r+ab)**2+(da+ab)**2+add)**3
         charg = -dxqxz*0.125_dp*fact
         RETURN
      END IF
      ! ZX - ZX
      IF (l1 == 2 .AND. l2 == 2 .AND. m == 1) THEN
         aa = da/SQRT(2.0_dp)
         ab = db/SQRT(2.0_dp)
         qxzqxz = &
            +2.0_dp*add0/SQRT((r+aa-ab)**2+(aa-ab)**2+add)**3-2.0_dp*add0/SQRT((r+aa+ab)**2+(aa-ab)**2+add)**3 &
            -2.0_dp*add0/SQRT((r-aa-ab)**2+(aa-ab)**2+add)**3+2.0_dp*add0/SQRT((r-aa+ab)**2+(aa-ab)**2+add)**3 &
            -2.0_dp*add0/SQRT((r+aa-ab)**2+(aa+ab)**2+add)**3+2.0_dp*add0/SQRT((r+aa+ab)**2+(aa+ab)**2+add)**3 &
            +2.0_dp*add0/SQRT((r-aa-ab)**2+(aa+ab)**2+add)**3-2.0_dp*add0/SQRT((r-aa+ab)**2+(aa+ab)**2+add)**3
         charg = -qxzqxz*0.0625_dp*fact
         RETURN
      END IF
      ! XX - XX
      IF (l1 == 2 .AND. l2 == 2 .AND. m == 2) THEN
         xyxy = 4.0_dp*add0/SQRT(r**2+(da-db)**2+add)**3+4.0_dp*add0/SQRT(r**2+(da+db)**2+add)**3- &
                8.0_dp*add0/SQRT(r**2+da**2+db**2+add)**3
         charg = -xyxy*0.0625_dp*fact
         RETURN
      END IF
      ! We should NEVER reach this point
      CALL cp__b("semi_empirical_int_utils.F",1641,"")
   END FUNCTION dcharg_int_ri_fs

! **************************************************************************************************
!> \brief Computes the general rotation matrices for the integrals
!>        between I and J (J>=I)
!>
!> \param sepi ...
!> \param sepj ...
!> \param rjiv ...
!> \param r ...
!> \param ij_matrix ...
!> \param do_derivatives ...
!> \param do_invert ...
!> \param debug_invert ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   RECURSIVE SUBROUTINE rotmat(sepi, sepj, rjiv, r, ij_matrix, do_derivatives, &
                               do_invert, debug_invert)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rjiv
      REAL(KIND=dp), INTENT(IN)                          :: r
      TYPE(rotmat_type), POINTER                         :: ij_matrix
      LOGICAL, INTENT(IN)                                :: do_derivatives
      LOGICAL, INTENT(OUT), OPTIONAL                     :: do_invert
      LOGICAL, INTENT(IN), OPTIONAL                      :: debug_invert

      CHARACTER(len=*), PARAMETER :: routineN = 'rotmat', routineP = moduleN//':'//routineN

      INTEGER                                            :: imap(3), k, l
      LOGICAL                                            :: dbg_inv, eval
      REAL(KIND=dp)                                      :: b, c2a, c2b, ca, ca2, cb, cb2, check, &
                                                            pt5sq3, r2i, s2a, s2b, sa, sb, sb2, &
                                                            sqb, sqb2i, x11, x22, x33
      REAL(KIND=dp), DIMENSION(3)                        :: b_d, c2a_d, c2b_d, ca2_d, ca_d, cb2_d, &
                                                            cb_d, r_d, s2a_d, s2b_d, sa_d, sb2_d, &
                                                            sb_d, sqb_d, x11_d, x22_d, x33_d
      REAL(KIND=dp), DIMENSION(3, 3)                     :: p
      REAL(KIND=dp), DIMENSION(3, 3, 3)                  :: p_d
      REAL(KIND=dp), DIMENSION(3, 5, 5)                  :: d_d
      REAL(KIND=dp), DIMENSION(5, 5)                     :: d

      IF(.NOT.(ASSOCIATED(ij_matrix)))CALL cp__a("semi_empirical_int_utils.F",1684)
      IF (PRESENT(do_invert)) do_invert = .FALSE.
      IF ((sepi%natorb > 1) .OR. (sepj%natorb > 1)) THEN
         ! Compute Geomtric data and interatomic distance
         !     CA  = COS(PHI)    , SA  = SIN(PHI)
         !     CB  = COS(THETA)  , SB  = SIN(THETA)
         !     C2A = COS(2*PHI)  , S2A = SIN(2*PHI)
         !     C2B = COS(2*THETA), S2B = SIN(2*PHI)
         eval = .TRUE.
         dbg_inv = .FALSE.
         pt5sq3 = 0.5_dp*SQRT(3.0_dp)
         imap(1) = 1
         imap(2) = 2
         imap(3) = 3
         check = rjiv(3)/r
         IF (PRESENT(debug_invert)) dbg_inv = debug_invert
         ! Check for special case: both atoms lie on the Z Axis
         IF (ABS(check) > 0.99999999_dp) THEN
            IF (PRESENT(do_invert)) THEN
               imap(1) = 3
               imap(2) = 2
               imap(3) = 1
               eval = .TRUE.
               do_invert = .TRUE.
            ELSE
               sa = 0.0_dp
               sb = 0.0_dp
               IF (check < 0.0_dp) THEN
                  ca = -1.0_dp
                  cb = -1.0_dp
               ELSE IF (check > 0.0_dp) THEN
                  ca = 1.0_dp
                  cb = 1.0_dp
               ELSE
                  ca = 0.0_dp
                  cb = 0.0_dp
               END IF
               eval = .FALSE.
            END IF
         END IF
         IF (dbg_inv) THEN
            ! When debugging the derivatives of the rotation matrix we must possibly force
            ! the inversion of the reference frame
            IF(.NOT.(.NOT. do_derivatives))CALL cp__a("semi_empirical_int_utils.F",1727)
            imap(1) = 3
            imap(2) = 2
            imap(3) = 1
            eval = .TRUE.
         END IF
         IF (eval) THEN
            x11 = rjiv(imap(1))
            x22 = rjiv(imap(2))
            x33 = rjiv(imap(3))
            cb = x33/r
            b = x11**2+x22**2
            sqb = SQRT(b)
            ca = x11/sqb
            sa = x22/sqb
            sb = sqb/r
         END IF
         ! Calculate rotation matrix elements
         p(1, 1) = ca*sb
         p(2, 1) = ca*cb
         p(3, 1) = -sa
         p(1, 2) = sa*sb
         p(2, 2) = sa*cb
         p(3, 2) = ca
         p(1, 3) = cb
         p(2, 3) = -sb
         p(3, 3) = 0.0_dp
         IF (sepi%dorb .OR. sepj%dorb) THEN
            ca2 = ca**2
            cb2 = cb**2
            sb2 = sb**2
            c2a = 2.0_dp*ca2-1.0_dp
            c2b = 2.0_dp*cb2-1.0_dp
            s2a = 2.0_dp*sa*ca
            s2b = 2.0_dp*sb*cb
            d(1, 1) = pt5sq3*c2a*sb2
            d(2, 1) = 0.5_dp*c2a*s2b
            d(3, 1) = -s2a*sb
            d(4, 1) = c2a*(cb2+0.5_dp*sb2)
            d(5, 1) = -s2a*cb
            d(1, 2) = pt5sq3*ca*s2b
            d(2, 2) = ca*c2b
            d(3, 2) = -sa*cb
            d(4, 2) = -0.5_dp*ca*s2b
            d(5, 2) = sa*sb
            d(1, 3) = cb2-0.5_dp*sb2
            d(2, 3) = -pt5sq3*s2b
            d(3, 3) = 0.0_dp
            d(4, 3) = pt5sq3*sb2
            d(5, 3) = 0.0_dp
            d(1, 4) = pt5sq3*sa*s2b
            d(2, 4) = sa*c2b
            d(3, 4) = ca*cb
            d(4, 4) = -0.5_dp*sa*s2b
            d(5, 4) = -ca*sb
            d(1, 5) = pt5sq3*s2a*sb2
            d(2, 5) = 0.5_dp*s2a*s2b
            d(3, 5) = c2a*sb
            d(4, 5) = s2a*(cb2+0.5_dp*sb2)
            d(5, 5) = c2a*cb
         END IF
         !  Rotation Elements for : S-P
         DO k = 1, 3
            DO l = 1, 3
               ij_matrix%sp(k, l) = p(k, l)
            END DO
         END DO
         !  Rotation Elements for : P-P
         DO k = 1, 3
            ij_matrix%pp(1, k, k) = p(k, 1)*p(k, 1)
            ij_matrix%pp(2, k, k) = p(k, 2)*p(k, 2)
            ij_matrix%pp(3, k, k) = p(k, 3)*p(k, 3)
            ij_matrix%pp(4, k, k) = p(k, 1)*p(k, 2)
            ij_matrix%pp(5, k, k) = p(k, 1)*p(k, 3)
            ij_matrix%pp(6, k, k) = p(k, 2)*p(k, 3)
            IF (k /= 1) THEN
               DO l = 1, k-1
                  ij_matrix%pp(1, k, l) = 2.0_dp*p(k, 1)*p(l, 1)
                  ij_matrix%pp(2, k, l) = 2.0_dp*p(k, 2)*p(l, 2)
                  ij_matrix%pp(3, k, l) = 2.0_dp*p(k, 3)*p(l, 3)
                  ij_matrix%pp(4, k, l) = p(k, 1)*p(l, 2)+p(k, 2)*p(l, 1)
                  ij_matrix%pp(5, k, l) = p(k, 1)*p(l, 3)+p(k, 3)*p(l, 1)
                  ij_matrix%pp(6, k, l) = p(k, 2)*p(l, 3)+p(k, 3)*p(l, 2)
               END DO
            END IF
         END DO
         IF (sepi%dorb .OR. sepj%dorb) THEN
            !  Rotation Elements for : S-D
            DO k = 1, 5
               DO l = 1, 5
                  ij_matrix%sd(k, l) = d(k, l)
               END DO
            END DO
            !  Rotation Elements for : D-P
            DO k = 1, 5
               DO l = 1, 3
                  ij_matrix%pd(1, k, l) = d(k, 1)*p(l, 1)
                  ij_matrix%pd(2, k, l) = d(k, 1)*p(l, 2)
                  ij_matrix%pd(3, k, l) = d(k, 1)*p(l, 3)
                  ij_matrix%pd(4, k, l) = d(k, 2)*p(l, 1)
                  ij_matrix%pd(5, k, l) = d(k, 2)*p(l, 2)
                  ij_matrix%pd(6, k, l) = d(k, 2)*p(l, 3)
                  ij_matrix%pd(7, k, l) = d(k, 3)*p(l, 1)
                  ij_matrix%pd(8, k, l) = d(k, 3)*p(l, 2)
                  ij_matrix%pd(9, k, l) = d(k, 3)*p(l, 3)
                  ij_matrix%pd(10, k, l) = d(k, 4)*p(l, 1)
                  ij_matrix%pd(11, k, l) = d(k, 4)*p(l, 2)
                  ij_matrix%pd(12, k, l) = d(k, 4)*p(l, 3)
                  ij_matrix%pd(13, k, l) = d(k, 5)*p(l, 1)
                  ij_matrix%pd(14, k, l) = d(k, 5)*p(l, 2)
                  ij_matrix%pd(15, k, l) = d(k, 5)*p(l, 3)
               END DO
            END DO
            !  Rotation Elements for : D-D
            DO k = 1, 5
               ij_matrix%dd(1, k, k) = d(k, 1)*d(k, 1)
               ij_matrix%dd(2, k, k) = d(k, 2)*d(k, 2)
               ij_matrix%dd(3, k, k) = d(k, 3)*d(k, 3)
               ij_matrix%dd(4, k, k) = d(k, 4)*d(k, 4)
               ij_matrix%dd(5, k, k) = d(k, 5)*d(k, 5)
               ij_matrix%dd(6, k, k) = d(k, 1)*d(k, 2)
               ij_matrix%dd(7, k, k) = d(k, 1)*d(k, 3)
               ij_matrix%dd(8, k, k) = d(k, 2)*d(k, 3)
               ij_matrix%dd(9, k, k) = d(k, 1)*d(k, 4)
               ij_matrix%dd(10, k, k) = d(k, 2)*d(k, 4)
               ij_matrix%dd(11, k, k) = d(k, 3)*d(k, 4)
               ij_matrix%dd(12, k, k) = d(k, 1)*d(k, 5)
               ij_matrix%dd(13, k, k) = d(k, 2)*d(k, 5)
               ij_matrix%dd(14, k, k) = d(k, 3)*d(k, 5)
               ij_matrix%dd(15, k, k) = d(k, 4)*d(k, 5)
               IF (k /= 1) THEN
                  DO l = 1, k-1
                     ij_matrix%dd(1, k, l) = 2.0_dp*d(k, 1)*d(l, 1)
                     ij_matrix%dd(2, k, l) = 2.0_dp*d(k, 2)*d(l, 2)
                     ij_matrix%dd(3, k, l) = 2.0_dp*d(k, 3)*d(l, 3)
                     ij_matrix%dd(4, k, l) = 2.0_dp*d(k, 4)*d(l, 4)
                     ij_matrix%dd(5, k, l) = 2.0_dp*d(k, 5)*d(l, 5)
                     ij_matrix%dd(6, k, l) = d(k, 1)*d(l, 2)+d(k, 2)*d(l, 1)
                     ij_matrix%dd(7, k, l) = d(k, 1)*d(l, 3)+d(k, 3)*d(l, 1)
                     ij_matrix%dd(8, k, l) = d(k, 2)*d(l, 3)+d(k, 3)*d(l, 2)
                     ij_matrix%dd(9, k, l) = d(k, 1)*d(l, 4)+d(k, 4)*d(l, 1)
                     ij_matrix%dd(10, k, l) = d(k, 2)*d(l, 4)+d(k, 4)*d(l, 2)
                     ij_matrix%dd(11, k, l) = d(k, 3)*d(l, 4)+d(k, 4)*d(l, 3)
                     ij_matrix%dd(12, k, l) = d(k, 1)*d(l, 5)+d(k, 5)*d(l, 1)
                     ij_matrix%dd(13, k, l) = d(k, 2)*d(l, 5)+d(k, 5)*d(l, 2)
                     ij_matrix%dd(14, k, l) = d(k, 3)*d(l, 5)+d(k, 5)*d(l, 3)
                     ij_matrix%dd(15, k, l) = d(k, 4)*d(l, 5)+d(k, 5)*d(l, 4)
                  END DO
               END IF
            END DO
         END IF
         ! Evaluate Derivatives if required
         IF (do_derivatives) THEN
            ! This condition is necessary because derivatives uses the invertion of the
            ! axis for treating the divergence point along z-axis
            IF(.NOT.(eval))CALL cp__a("semi_empirical_int_utils.F",1882)
            x11_d = 0.0_dp; x11_d(1) = 1.0_dp
            x22_d = 0.0_dp; x22_d(2) = 1.0_dp
            x33_d = 0.0_dp; x33_d(3) = 1.0_dp
            r_d = (/x11, x22, x33/)/r
            b_d = 2.0_dp*(x11*x11_d+x22*x22_d)
            sqb_d = (0.5_dp/sqb)*b_d
            r2i = 1.0_dp/r**2
            sqb2i = 1.0_dp/sqb**2
            cb_d = (x33_d*r-x33*r_d)*r2i
            ca_d = (x11_d*sqb-x11*sqb_d)*sqb2i
            sa_d = (x22_d*sqb-x22*sqb_d)*sqb2i
            sb_d = (sqb_d*r-sqb*r_d)*r2i
            ! Calculate derivatives of rotation matrix elements
            p_d(:, 1, 1) = ca_d*sb+ca*sb_d
            p_d(:, 2, 1) = ca_d*cb+ca*cb_d
            p_d(:, 3, 1) = -sa_d
            p_d(:, 1, 2) = sa_d*sb+sa*sb_d
            p_d(:, 2, 2) = sa_d*cb+sa*cb_d
            p_d(:, 3, 2) = ca_d
            p_d(:, 1, 3) = cb_d
            p_d(:, 2, 3) = -sb_d
            p_d(:, 3, 3) = 0.0_dp
            IF (sepi%dorb .OR. sepj%dorb) THEN
               ca2_d = 2.0_dp*ca*ca_d
               cb2_d = 2.0_dp*cb*cb_d
               sb2_d = 2.0_dp*sb*sb_d
               c2a_d = 2.0_dp*ca2_d
               c2b_d = 2.0_dp*cb2_d
               s2a_d = 2.0_dp*(sa_d*ca+sa*ca_d)
               s2b_d = 2.0_dp*(sb_d*cb+sb*cb_d)
               d_d(:, 1, 1) = pt5sq3*(c2a_d*sb2+c2a*sb2_d)
               d_d(:, 2, 1) = 0.5_dp*(c2a_d*s2b+c2a*s2b_d)
               d_d(:, 3, 1) = -s2a_d*sb-s2a*sb_d
               d_d(:, 4, 1) = c2a_d*(cb2+0.5_dp*sb2)+c2a*(cb2_d+0.5_dp*sb2_d)
               d_d(:, 5, 1) = -s2a_d*cb-s2a*cb_d
               d_d(:, 1, 2) = pt5sq3*(ca_d*s2b+ca*s2b_d)
               d_d(:, 2, 2) = ca_d*c2b+ca*c2b_d
               d_d(:, 3, 2) = -sa_d*cb-sa*cb_d
               d_d(:, 4, 2) = -0.5_dp*(ca_d*s2b+ca*s2b_d)
               d_d(:, 5, 2) = sa_d*sb+sa*sb_d
               d_d(:, 1, 3) = cb2_d-0.5_dp*sb2_d
               d_d(:, 2, 3) = -pt5sq3*s2b_d
               d_d(:, 3, 3) = 0.0_dp
               d_d(:, 4, 3) = pt5sq3*sb2_d
               d_d(:, 5, 3) = 0.0_dp
               d_d(:, 1, 4) = pt5sq3*(sa_d*s2b+sa*s2b_d)
               d_d(:, 2, 4) = sa_d*c2b+sa*c2b_d
               d_d(:, 3, 4) = ca_d*cb+ca*cb_d
               d_d(:, 4, 4) = -0.5_dp*(sa_d*s2b+sa*s2b_d)
               d_d(:, 5, 4) = -ca_d*sb-ca*sb_d
               d_d(:, 1, 5) = pt5sq3*(s2a_d*sb2+s2a*sb2_d)
               d_d(:, 2, 5) = 0.5_dp*(s2a_d*s2b+s2a*s2b_d)
               d_d(:, 3, 5) = c2a_d*sb+c2a*sb_d
               d_d(:, 4, 5) = s2a_d*(cb2+0.5_dp*sb2)+s2a*(cb2_d+0.5_dp*sb2_d)
               d_d(:, 5, 5) = c2a_d*cb+c2a*cb_d
            END IF
            !  Derivative for Rotation Elements for : S-P
            DO k = 1, 3
               DO l = 1, 3
                  ij_matrix%sp_d(:, k, l) = p_d(:, k, l)
               END DO
            END DO
            !  Derivative for Rotation Elements for : P-P
            DO k = 1, 3
               ij_matrix%pp_d(:, 1, k, k) = p_d(:, k, 1)*p(k, 1)+p(k, 1)*p_d(:, k, 1)
               ij_matrix%pp_d(:, 2, k, k) = p_d(:, k, 2)*p(k, 2)+p(k, 2)*p_d(:, k, 2)
               ij_matrix%pp_d(:, 3, k, k) = p_d(:, k, 3)*p(k, 3)+p(k, 3)*p_d(:, k, 3)
               ij_matrix%pp_d(:, 4, k, k) = p_d(:, k, 1)*p(k, 2)+p(k, 1)*p_d(:, k, 2)
               ij_matrix%pp_d(:, 5, k, k) = p_d(:, k, 1)*p(k, 3)+p(k, 1)*p_d(:, k, 3)
               ij_matrix%pp_d(:, 6, k, k) = p_d(:, k, 2)*p(k, 3)+p(k, 2)*p_d(:, k, 3)
               IF (k /= 1) THEN
                  DO l = 1, k-1
                     ij_matrix%pp_d(:, 1, k, l) = 2.0_dp*(p_d(:, k, 1)*p(l, 1)+p(k, 1)*p_d(:, l, 1))
                     ij_matrix%pp_d(:, 2, k, l) = 2.0_dp*(p_d(:, k, 2)*p(l, 2)+p(k, 2)*p_d(:, l, 2))
                     ij_matrix%pp_d(:, 3, k, l) = 2.0_dp*(p_d(:, k, 3)*p(l, 3)+p(k, 3)*p_d(:, l, 3))
                     ij_matrix%pp_d(:, 4, k, l) = (p_d(:, k, 1)*p(l, 2)+p(k, 1)*p_d(:, l, 2))+ &
                                                  (p_d(:, k, 2)*p(l, 1)+p(k, 2)*p_d(:, l, 1))
                     ij_matrix%pp_d(:, 5, k, l) = (p_d(:, k, 1)*p(l, 3)+p(k, 1)*p_d(:, l, 3))+ &
                                                  (p_d(:, k, 3)*p(l, 1)+p(k, 3)*p_d(:, l, 1))
                     ij_matrix%pp_d(:, 6, k, l) = (p_d(:, k, 2)*p(l, 3)+p(k, 2)*p_d(:, l, 3))+ &
                                                  (p_d(:, k, 3)*p(l, 2)+p(k, 3)*p_d(:, l, 2))
                  END DO
               END IF
            END DO
            IF (sepi%dorb .OR. sepj%dorb) THEN
               !  Rotation Elements for : S-D
               DO k = 1, 5
                  DO l = 1, 5
                     ij_matrix%sd_d(:, k, l) = d_d(:, k, l)
                  END DO
               END DO
               !  Rotation Elements for : D-P
               DO k = 1, 5
                  DO l = 1, 3
                     ij_matrix%pd_d(:, 1, k, l) = (d_d(:, k, 1)*p(l, 1)+d(k, 1)*p_d(:, l, 1))
                     ij_matrix%pd_d(:, 2, k, l) = (d_d(:, k, 1)*p(l, 2)+d(k, 1)*p_d(:, l, 2))
                     ij_matrix%pd_d(:, 3, k, l) = (d_d(:, k, 1)*p(l, 3)+d(k, 1)*p_d(:, l, 3))
                     ij_matrix%pd_d(:, 4, k, l) = (d_d(:, k, 2)*p(l, 1)+d(k, 2)*p_d(:, l, 1))
                     ij_matrix%pd_d(:, 5, k, l) = (d_d(:, k, 2)*p(l, 2)+d(k, 2)*p_d(:, l, 2))
                     ij_matrix%pd_d(:, 6, k, l) = (d_d(:, k, 2)*p(l, 3)+d(k, 2)*p_d(:, l, 3))
                     ij_matrix%pd_d(:, 7, k, l) = (d_d(:, k, 3)*p(l, 1)+d(k, 3)*p_d(:, l, 1))
                     ij_matrix%pd_d(:, 8, k, l) = (d_d(:, k, 3)*p(l, 2)+d(k, 3)*p_d(:, l, 2))
                     ij_matrix%pd_d(:, 9, k, l) = (d_d(:, k, 3)*p(l, 3)+d(k, 3)*p_d(:, l, 3))
                     ij_matrix%pd_d(:, 10, k, l) = (d_d(:, k, 4)*p(l, 1)+d(k, 4)*p_d(:, l, 1))
                     ij_matrix%pd_d(:, 11, k, l) = (d_d(:, k, 4)*p(l, 2)+d(k, 4)*p_d(:, l, 2))
                     ij_matrix%pd_d(:, 12, k, l) = (d_d(:, k, 4)*p(l, 3)+d(k, 4)*p_d(:, l, 3))
                     ij_matrix%pd_d(:, 13, k, l) = (d_d(:, k, 5)*p(l, 1)+d(k, 5)*p_d(:, l, 1))
                     ij_matrix%pd_d(:, 14, k, l) = (d_d(:, k, 5)*p(l, 2)+d(k, 5)*p_d(:, l, 2))
                     ij_matrix%pd_d(:, 15, k, l) = (d_d(:, k, 5)*p(l, 3)+d(k, 5)*p_d(:, l, 3))
                  END DO
               END DO
               !  Rotation Elements for : D-D
               DO k = 1, 5
                  ij_matrix%dd_d(:, 1, k, k) = (d_d(:, k, 1)*d(k, 1)+d(k, 1)*d_d(:, k, 1))
                  ij_matrix%dd_d(:, 2, k, k) = (d_d(:, k, 2)*d(k, 2)+d(k, 2)*d_d(:, k, 2))
                  ij_matrix%dd_d(:, 3, k, k) = (d_d(:, k, 3)*d(k, 3)+d(k, 3)*d_d(:, k, 3))
                  ij_matrix%dd_d(:, 4, k, k) = (d_d(:, k, 4)*d(k, 4)+d(k, 4)*d_d(:, k, 4))
                  ij_matrix%dd_d(:, 5, k, k) = (d_d(:, k, 5)*d(k, 5)+d(k, 5)*d_d(:, k, 5))
                  ij_matrix%dd_d(:, 6, k, k) = (d_d(:, k, 1)*d(k, 2)+d(k, 1)*d_d(:, k, 2))
                  ij_matrix%dd_d(:, 7, k, k) = (d_d(:, k, 1)*d(k, 3)+d(k, 1)*d_d(:, k, 3))
                  ij_matrix%dd_d(:, 8, k, k) = (d_d(:, k, 2)*d(k, 3)+d(k, 2)*d_d(:, k, 3))
                  ij_matrix%dd_d(:, 9, k, k) = (d_d(:, k, 1)*d(k, 4)+d(k, 1)*d_d(:, k, 4))
                  ij_matrix%dd_d(:, 10, k, k) = (d_d(:, k, 2)*d(k, 4)+d(k, 2)*d_d(:, k, 4))
                  ij_matrix%dd_d(:, 11, k, k) = (d_d(:, k, 3)*d(k, 4)+d(k, 3)*d_d(:, k, 4))
                  ij_matrix%dd_d(:, 12, k, k) = (d_d(:, k, 1)*d(k, 5)+d(k, 1)*d_d(:, k, 5))
                  ij_matrix%dd_d(:, 13, k, k) = (d_d(:, k, 2)*d(k, 5)+d(k, 2)*d_d(:, k, 5))
                  ij_matrix%dd_d(:, 14, k, k) = (d_d(:, k, 3)*d(k, 5)+d(k, 3)*d_d(:, k, 5))
                  ij_matrix%dd_d(:, 15, k, k) = (d_d(:, k, 4)*d(k, 5)+d(k, 4)*d_d(:, k, 5))
                  IF (k /= 1) THEN
                     DO l = 1, k-1
                        ij_matrix%dd_d(:, 1, k, l) = 2.0_dp*(d_d(:, k, 1)*d(l, 1)+d(k, 1)*d_d(:, l, 1))
                        ij_matrix%dd_d(:, 2, k, l) = 2.0_dp*(d_d(:, k, 2)*d(l, 2)+d(k, 2)*d_d(:, l, 2))
                        ij_matrix%dd_d(:, 3, k, l) = 2.0_dp*(d_d(:, k, 3)*d(l, 3)+d(k, 3)*d_d(:, l, 3))
                        ij_matrix%dd_d(:, 4, k, l) = 2.0_dp*(d_d(:, k, 4)*d(l, 4)+d(k, 4)*d_d(:, l, 4))
                        ij_matrix%dd_d(:, 5, k, l) = 2.0_dp*(d_d(:, k, 5)*d(l, 5)+d(k, 5)*d_d(:, l, 5))
                        ij_matrix%dd_d(:, 6, k, l) = (d_d(:, k, 1)*d(l, 2)+d(k, 1)*d_d(:, l, 2))+ &
                                                     (d_d(:, k, 2)*d(l, 1)+d(k, 2)*d_d(:, l, 1))
                        ij_matrix%dd_d(:, 7, k, l) = (d_d(:, k, 1)*d(l, 3)+d(k, 1)*d_d(:, l, 3))+ &
                                                     (d_d(:, k, 3)*d(l, 1)+d(k, 3)*d_d(:, l, 1))
                        ij_matrix%dd_d(:, 8, k, l) = (d_d(:, k, 2)*d(l, 3)+d(k, 2)*d_d(:, l, 3))+ &
                                                     (d_d(:, k, 3)*d(l, 2)+d(k, 3)*d_d(:, l, 2))
                        ij_matrix%dd_d(:, 9, k, l) = (d_d(:, k, 1)*d(l, 4)+d(k, 1)*d_d(:, l, 4))+ &
                                                     (d_d(:, k, 4)*d(l, 1)+d(k, 4)*d_d(:, l, 1))
                        ij_matrix%dd_d(:, 10, k, l) = (d_d(:, k, 2)*d(l, 4)+d(k, 2)*d_d(:, l, 4))+ &
                                                      (d_d(:, k, 4)*d(l, 2)+d(k, 4)*d_d(:, l, 2))
                        ij_matrix%dd_d(:, 11, k, l) = (d_d(:, k, 3)*d(l, 4)+d(k, 3)*d_d(:, l, 4))+ &
                                                      (d_d(:, k, 4)*d(l, 3)+d(k, 4)*d_d(:, l, 3))
                        ij_matrix%dd_d(:, 12, k, l) = (d_d(:, k, 1)*d(l, 5)+d(k, 1)*d_d(:, l, 5))+ &
                                                      (d_d(:, k, 5)*d(l, 1)+d(k, 5)*d_d(:, l, 1))
                        ij_matrix%dd_d(:, 13, k, l) = (d_d(:, k, 2)*d(l, 5)+d(k, 2)*d_d(:, l, 5))+ &
                                                      (d_d(:, k, 5)*d(l, 2)+d(k, 5)*d_d(:, l, 2))
                        ij_matrix%dd_d(:, 14, k, l) = (d_d(:, k, 3)*d(l, 5)+d(k, 3)*d_d(:, l, 5))+ &
                                                      (d_d(:, k, 5)*d(l, 3)+d(k, 5)*d_d(:, l, 3))
                        ij_matrix%dd_d(:, 15, k, l) = (d_d(:, k, 4)*d(l, 5)+d(k, 4)*d_d(:, l, 5))+ &
                                                      (d_d(:, k, 5)*d(l, 4)+d(k, 5)*d_d(:, l, 4))
                     END DO
                  END IF
               END DO
            END IF
            IF (debug_this_module) THEN
               CALL check_rotmat_der(sepi, sepj, rjiv, ij_matrix, do_invert=do_invert)
            END IF
         END IF
      END IF
   END SUBROUTINE rotmat

! **************************************************************************************************
!> \brief First Step of the rotation of the two-electron two-center integrals in
!>        SPD basis
!>
!> \param sepi ...
!> \param sepj ...
!> \param rijv ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param invert ...
!> \param ii ...
!> \param kk ...
!> \param rep ...
!> \param logv ...
!> \param ij_matrix ...
!> \param v ...
!> \param lgrad ...
!> \param rep_d ...
!> \param v_d ...
!> \param logv_d ...
!> \param drij ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   RECURSIVE SUBROUTINE rot_2el_2c_first(sepi, sepj, rijv, se_int_control, se_taper, &
                                         invert, ii, kk, rep, logv, ij_matrix, v, lgrad, rep_d, v_d, logv_d, drij)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rijv
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      LOGICAL, INTENT(IN)                                :: invert
      INTEGER, INTENT(IN)                                :: ii, kk
      REAL(KIND=dp), DIMENSION(491), INTENT(IN)          :: rep
      LOGICAL, DIMENSION(45, 45), INTENT(OUT)            :: logv
      TYPE(rotmat_type), POINTER                         :: ij_matrix
      REAL(KIND=dp), DIMENSION(45, 45), INTENT(OUT)      :: v
      LOGICAL, INTENT(IN)                                :: lgrad
      REAL(KIND=dp), DIMENSION(491), INTENT(IN), &
         OPTIONAL                                        :: rep_d
      REAL(KIND=dp), DIMENSION(3, 45, 45), INTENT(OUT), &
         OPTIONAL                                        :: v_d
      LOGICAL, DIMENSION(45, 45), INTENT(OUT), OPTIONAL  :: logv_d
      REAL(KIND=dp), DIMENSION(3), INTENT(IN), OPTIONAL  :: drij

      CHARACTER(len=*), PARAMETER :: routineN = 'rot_2el_2c_first', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, i1, ij, j1, k, k1, kl, l, l1, limkl, &
                                                            ll, mm, nd
      REAL(KIND=dp)                                      :: wrepp, wrepp_d(3)

      IF (lgrad) THEN
         IF(.NOT.(PRESENT(rep_d)))CALL cp__a("semi_empirical_int_utils.F",2101)
         IF(.NOT.(PRESENT(v_d)))CALL cp__a("semi_empirical_int_utils.F",2102)
         IF(.NOT.(PRESENT(logv_d)))CALL cp__a("semi_empirical_int_utils.F",2103)
         IF(.NOT.(PRESENT(drij)))CALL cp__a("semi_empirical_int_utils.F",2104)
      END IF
      limkl = indexb(kk, kk)
      DO k = 1, limkl
         DO i = 1, 45
            logv(i, k) = .FALSE.
            v(i, k) = 0.0_dp
         END DO
      END DO
      !
      DO i1 = 1, ii
         DO j1 = 1, i1
            ij = indexa(i1, j1)
            !
            DO k1 = 1, kk
               !
               DO l1 = 1, k1
                  kl = indexa(k1, l1)
                  nd = ijkl_ind(ij, kl)
                  IF (nd /= 0) THEN
                     !
                     wrepp = rep(nd)
                     ll = indexb(k1, l1)
                     mm = int2c_type(ll)
                     !
                     IF (mm == 1) THEN
                        v(ij, 1) = wrepp
                     ELSE IF (mm == 2) THEN
                        k = k1-1
                        v(ij, 2) = v(ij, 2)+ij_matrix%sp(k, 1)*wrepp
                        v(ij, 4) = v(ij, 4)+ij_matrix%sp(k, 2)*wrepp
                        v(ij, 7) = v(ij, 7)+ij_matrix%sp(k, 3)*wrepp
                     ELSE IF (mm == 3) THEN
                        k = k1-1
                        l = l1-1
                        v(ij, 3) = v(ij, 3)+ij_matrix%pp(1, k, l)*wrepp
                        v(ij, 6) = v(ij, 6)+ij_matrix%pp(2, k, l)*wrepp
                        v(ij, 10) = v(ij, 10)+ij_matrix%pp(3, k, l)*wrepp
                        v(ij, 5) = v(ij, 5)+ij_matrix%pp(4, k, l)*wrepp
                        v(ij, 8) = v(ij, 8)+ij_matrix%pp(5, k, l)*wrepp
                        v(ij, 9) = v(ij, 9)+ij_matrix%pp(6, k, l)*wrepp
                     ELSE IF (mm == 4) THEN
                        k = k1-4
                        v(ij, 11) = v(ij, 11)+ij_matrix%sd(k, 1)*wrepp
                        v(ij, 16) = v(ij, 16)+ij_matrix%sd(k, 2)*wrepp
                        v(ij, 22) = v(ij, 22)+ij_matrix%sd(k, 3)*wrepp
                        v(ij, 29) = v(ij, 29)+ij_matrix%sd(k, 4)*wrepp
                        v(ij, 37) = v(ij, 37)+ij_matrix%sd(k, 5)*wrepp
                     ELSE IF (mm == 5) THEN
                        k = k1-4
                        l = l1-1
                        v(ij, 12) = v(ij, 12)+ij_matrix%pd(1, k, l)*wrepp
                        v(ij, 13) = v(ij, 13)+ij_matrix%pd(2, k, l)*wrepp
                        v(ij, 14) = v(ij, 14)+ij_matrix%pd(3, k, l)*wrepp
                        v(ij, 17) = v(ij, 17)+ij_matrix%pd(4, k, l)*wrepp
                        v(ij, 18) = v(ij, 18)+ij_matrix%pd(5, k, l)*wrepp
                        v(ij, 19) = v(ij, 19)+ij_matrix%pd(6, k, l)*wrepp
                        v(ij, 23) = v(ij, 23)+ij_matrix%pd(7, k, l)*wrepp
                        v(ij, 24) = v(ij, 24)+ij_matrix%pd(8, k, l)*wrepp
                        v(ij, 25) = v(ij, 25)+ij_matrix%pd(9, k, l)*wrepp
                        v(ij, 30) = v(ij, 30)+ij_matrix%pd(10, k, l)*wrepp
                        v(ij, 31) = v(ij, 31)+ij_matrix%pd(11, k, l)*wrepp
                        v(ij, 32) = v(ij, 32)+ij_matrix%pd(12, k, l)*wrepp
                        v(ij, 38) = v(ij, 38)+ij_matrix%pd(13, k, l)*wrepp
                        v(ij, 39) = v(ij, 39)+ij_matrix%pd(14, k, l)*wrepp
                        v(ij, 40) = v(ij, 40)+ij_matrix%pd(15, k, l)*wrepp
                     ELSE IF (mm == 6) THEN
                        k = k1-4
                        l = l1-4
                        v(ij, 15) = v(ij, 15)+ij_matrix%dd(1, k, l)*wrepp
                        v(ij, 21) = v(ij, 21)+ij_matrix%dd(2, k, l)*wrepp
                        v(ij, 28) = v(ij, 28)+ij_matrix%dd(3, k, l)*wrepp
                        v(ij, 36) = v(ij, 36)+ij_matrix%dd(4, k, l)*wrepp
                        v(ij, 45) = v(ij, 45)+ij_matrix%dd(5, k, l)*wrepp
                        v(ij, 20) = v(ij, 20)+ij_matrix%dd(6, k, l)*wrepp
                        v(ij, 26) = v(ij, 26)+ij_matrix%dd(7, k, l)*wrepp
                        v(ij, 27) = v(ij, 27)+ij_matrix%dd(8, k, l)*wrepp
                        v(ij, 33) = v(ij, 33)+ij_matrix%dd(9, k, l)*wrepp
                        v(ij, 34) = v(ij, 34)+ij_matrix%dd(10, k, l)*wrepp
                        v(ij, 35) = v(ij, 35)+ij_matrix%dd(11, k, l)*wrepp
                        v(ij, 41) = v(ij, 41)+ij_matrix%dd(12, k, l)*wrepp
                        v(ij, 42) = v(ij, 42)+ij_matrix%dd(13, k, l)*wrepp
                        v(ij, 43) = v(ij, 43)+ij_matrix%dd(14, k, l)*wrepp
                        v(ij, 44) = v(ij, 44)+ij_matrix%dd(15, k, l)*wrepp
                     END IF
                  END IF
               END DO
            END DO
            DO kl = 1, limkl
               logv(ij, kl) = (ABS(v(ij, kl)) > 0.0_dp)
            END DO
         END DO
      END DO
      ! Gradients
      IF (lgrad) THEN
         DO k = 1, limkl
            DO i = 1, 45
               logv_d(i, k) = .FALSE.
               v_d(1, i, k) = 0.0_dp
               v_d(2, i, k) = 0.0_dp
               v_d(3, i, k) = 0.0_dp
            END DO
         END DO
         DO i1 = 1, ii
            DO j1 = 1, i1
               ij = indexa(i1, j1)
               !
               DO k1 = 1, kk
                  !
                  DO l1 = 1, k1
                     kl = indexa(k1, l1)
                     nd = ijkl_ind(ij, kl)
                     IF (nd /= 0) THEN
                        !
                        wrepp = rep(nd)
                        wrepp_d = rep_d(nd)*drij
                        ll = indexb(k1, l1)
                        mm = int2c_type(ll)
                        !
                        IF (mm == 1) THEN
                           v_d(1, ij, 1) = wrepp_d(1)
                           v_d(2, ij, 1) = wrepp_d(2)
                           v_d(3, ij, 1) = wrepp_d(3)
                        ELSE IF (mm == 2) THEN
                           k = k1-1
                           v_d(1, ij, 2) = v_d(1, ij, 2)+ij_matrix%sp_d(1, k, 1)*wrepp+ij_matrix%sp(k, 1)*wrepp_d(1)
                           v_d(1, ij, 4) = v_d(1, ij, 4)+ij_matrix%sp_d(1, k, 2)*wrepp+ij_matrix%sp(k, 2)*wrepp_d(1)
                           v_d(1, ij, 7) = v_d(1, ij, 7)+ij_matrix%sp_d(1, k, 3)*wrepp+ij_matrix%sp(k, 3)*wrepp_d(1)

                           v_d(2, ij, 2) = v_d(2, ij, 2)+ij_matrix%sp_d(2, k, 1)*wrepp+ij_matrix%sp(k, 1)*wrepp_d(2)
                           v_d(2, ij, 4) = v_d(2, ij, 4)+ij_matrix%sp_d(2, k, 2)*wrepp+ij_matrix%sp(k, 2)*wrepp_d(2)
                           v_d(2, ij, 7) = v_d(2, ij, 7)+ij_matrix%sp_d(2, k, 3)*wrepp+ij_matrix%sp(k, 3)*wrepp_d(2)

                           v_d(3, ij, 2) = v_d(3, ij, 2)+ij_matrix%sp_d(3, k, 1)*wrepp+ij_matrix%sp(k, 1)*wrepp_d(3)
                           v_d(3, ij, 4) = v_d(3, ij, 4)+ij_matrix%sp_d(3, k, 2)*wrepp+ij_matrix%sp(k, 2)*wrepp_d(3)
                           v_d(3, ij, 7) = v_d(3, ij, 7)+ij_matrix%sp_d(3, k, 3)*wrepp+ij_matrix%sp(k, 3)*wrepp_d(3)
                        ELSE IF (mm == 3) THEN
                           k = k1-1
                           l = l1-1
                           v_d(1, ij, 3) = v_d(1, ij, 3)+ij_matrix%pp_d(1, 1, k, l)*wrepp+ij_matrix%pp(1, k, l)*wrepp_d(1)
                           v_d(1, ij, 6) = v_d(1, ij, 6)+ij_matrix%pp_d(1, 2, k, l)*wrepp+ij_matrix%pp(2, k, l)*wrepp_d(1)
                           v_d(1, ij, 10) = v_d(1, ij, 10)+ij_matrix%pp_d(1, 3, k, l)*wrepp+ij_matrix%pp(3, k, l)*wrepp_d(1)
                           v_d(1, ij, 5) = v_d(1, ij, 5)+ij_matrix%pp_d(1, 4, k, l)*wrepp+ij_matrix%pp(4, k, l)*wrepp_d(1)
                           v_d(1, ij, 8) = v_d(1, ij, 8)+ij_matrix%pp_d(1, 5, k, l)*wrepp+ij_matrix%pp(5, k, l)*wrepp_d(1)
                           v_d(1, ij, 9) = v_d(1, ij, 9)+ij_matrix%pp_d(1, 6, k, l)*wrepp+ij_matrix%pp(6, k, l)*wrepp_d(1)

                           v_d(2, ij, 3) = v_d(2, ij, 3)+ij_matrix%pp_d(2, 1, k, l)*wrepp+ij_matrix%pp(1, k, l)*wrepp_d(2)
                           v_d(2, ij, 6) = v_d(2, ij, 6)+ij_matrix%pp_d(2, 2, k, l)*wrepp+ij_matrix%pp(2, k, l)*wrepp_d(2)
                           v_d(2, ij, 10) = v_d(2, ij, 10)+ij_matrix%pp_d(2, 3, k, l)*wrepp+ij_matrix%pp(3, k, l)*wrepp_d(2)
                           v_d(2, ij, 5) = v_d(2, ij, 5)+ij_matrix%pp_d(2, 4, k, l)*wrepp+ij_matrix%pp(4, k, l)*wrepp_d(2)
                           v_d(2, ij, 8) = v_d(2, ij, 8)+ij_matrix%pp_d(2, 5, k, l)*wrepp+ij_matrix%pp(5, k, l)*wrepp_d(2)
                           v_d(2, ij, 9) = v_d(2, ij, 9)+ij_matrix%pp_d(2, 6, k, l)*wrepp+ij_matrix%pp(6, k, l)*wrepp_d(2)

                           v_d(3, ij, 3) = v_d(3, ij, 3)+ij_matrix%pp_d(3, 1, k, l)*wrepp+ij_matrix%pp(1, k, l)*wrepp_d(3)
                           v_d(3, ij, 6) = v_d(3, ij, 6)+ij_matrix%pp_d(3, 2, k, l)*wrepp+ij_matrix%pp(2, k, l)*wrepp_d(3)
                           v_d(3, ij, 10) = v_d(3, ij, 10)+ij_matrix%pp_d(3, 3, k, l)*wrepp+ij_matrix%pp(3, k, l)*wrepp_d(3)
                           v_d(3, ij, 5) = v_d(3, ij, 5)+ij_matrix%pp_d(3, 4, k, l)*wrepp+ij_matrix%pp(4, k, l)*wrepp_d(3)
                           v_d(3, ij, 8) = v_d(3, ij, 8)+ij_matrix%pp_d(3, 5, k, l)*wrepp+ij_matrix%pp(5, k, l)*wrepp_d(3)
                           v_d(3, ij, 9) = v_d(3, ij, 9)+ij_matrix%pp_d(3, 6, k, l)*wrepp+ij_matrix%pp(6, k, l)*wrepp_d(3)
                        ELSE IF (mm == 4) THEN
                           k = k1-4
                           v_d(1, ij, 11) = v_d(1, ij, 11)+ij_matrix%sd_d(1, k, 1)*wrepp+ij_matrix%sd(k, 1)*wrepp_d(1)
                           v_d(1, ij, 16) = v_d(1, ij, 16)+ij_matrix%sd_d(1, k, 2)*wrepp+ij_matrix%sd(k, 2)*wrepp_d(1)
                           v_d(1, ij, 22) = v_d(1, ij, 22)+ij_matrix%sd_d(1, k, 3)*wrepp+ij_matrix%sd(k, 3)*wrepp_d(1)
                           v_d(1, ij, 29) = v_d(1, ij, 29)+ij_matrix%sd_d(1, k, 4)*wrepp+ij_matrix%sd(k, 4)*wrepp_d(1)
                           v_d(1, ij, 37) = v_d(1, ij, 37)+ij_matrix%sd_d(1, k, 5)*wrepp+ij_matrix%sd(k, 5)*wrepp_d(1)

                           v_d(2, ij, 11) = v_d(2, ij, 11)+ij_matrix%sd_d(2, k, 1)*wrepp+ij_matrix%sd(k, 1)*wrepp_d(2)
                           v_d(2, ij, 16) = v_d(2, ij, 16)+ij_matrix%sd_d(2, k, 2)*wrepp+ij_matrix%sd(k, 2)*wrepp_d(2)
                           v_d(2, ij, 22) = v_d(2, ij, 22)+ij_matrix%sd_d(2, k, 3)*wrepp+ij_matrix%sd(k, 3)*wrepp_d(2)
                           v_d(2, ij, 29) = v_d(2, ij, 29)+ij_matrix%sd_d(2, k, 4)*wrepp+ij_matrix%sd(k, 4)*wrepp_d(2)
                           v_d(2, ij, 37) = v_d(2, ij, 37)+ij_matrix%sd_d(2, k, 5)*wrepp+ij_matrix%sd(k, 5)*wrepp_d(2)

                           v_d(3, ij, 11) = v_d(3, ij, 11)+ij_matrix%sd_d(3, k, 1)*wrepp+ij_matrix%sd(k, 1)*wrepp_d(3)
                           v_d(3, ij, 16) = v_d(3, ij, 16)+ij_matrix%sd_d(3, k, 2)*wrepp+ij_matrix%sd(k, 2)*wrepp_d(3)
                           v_d(3, ij, 22) = v_d(3, ij, 22)+ij_matrix%sd_d(3, k, 3)*wrepp+ij_matrix%sd(k, 3)*wrepp_d(3)
                           v_d(3, ij, 29) = v_d(3, ij, 29)+ij_matrix%sd_d(3, k, 4)*wrepp+ij_matrix%sd(k, 4)*wrepp_d(3)
                           v_d(3, ij, 37) = v_d(3, ij, 37)+ij_matrix%sd_d(3, k, 5)*wrepp+ij_matrix%sd(k, 5)*wrepp_d(3)
                        ELSE IF (mm == 5) THEN
                           k = k1-4
                           l = l1-1
                           v_d(1, ij, 12) = v_d(1, ij, 12)+ij_matrix%pd_d(1, 1, k, l)*wrepp+ij_matrix%pd(1, k, l)*wrepp_d(1)
                           v_d(1, ij, 13) = v_d(1, ij, 13)+ij_matrix%pd_d(1, 2, k, l)*wrepp+ij_matrix%pd(2, k, l)*wrepp_d(1)
                           v_d(1, ij, 14) = v_d(1, ij, 14)+ij_matrix%pd_d(1, 3, k, l)*wrepp+ij_matrix%pd(3, k, l)*wrepp_d(1)
                           v_d(1, ij, 17) = v_d(1, ij, 17)+ij_matrix%pd_d(1, 4, k, l)*wrepp+ij_matrix%pd(4, k, l)*wrepp_d(1)
                           v_d(1, ij, 18) = v_d(1, ij, 18)+ij_matrix%pd_d(1, 5, k, l)*wrepp+ij_matrix%pd(5, k, l)*wrepp_d(1)
                           v_d(1, ij, 19) = v_d(1, ij, 19)+ij_matrix%pd_d(1, 6, k, l)*wrepp+ij_matrix%pd(6, k, l)*wrepp_d(1)
                           v_d(1, ij, 23) = v_d(1, ij, 23)+ij_matrix%pd_d(1, 7, k, l)*wrepp+ij_matrix%pd(7, k, l)*wrepp_d(1)
                           v_d(1, ij, 24) = v_d(1, ij, 24)+ij_matrix%pd_d(1, 8, k, l)*wrepp+ij_matrix%pd(8, k, l)*wrepp_d(1)
                           v_d(1, ij, 25) = v_d(1, ij, 25)+ij_matrix%pd_d(1, 9, k, l)*wrepp+ij_matrix%pd(9, k, l)*wrepp_d(1)
                           v_d(1, ij, 30) = v_d(1, ij, 30)+ij_matrix%pd_d(1, 10, k, l)*wrepp+ij_matrix%pd(10, k, l)*wrepp_d(1)
                           v_d(1, ij, 31) = v_d(1, ij, 31)+ij_matrix%pd_d(1, 11, k, l)*wrepp+ij_matrix%pd(11, k, l)*wrepp_d(1)
                           v_d(1, ij, 32) = v_d(1, ij, 32)+ij_matrix%pd_d(1, 12, k, l)*wrepp+ij_matrix%pd(12, k, l)*wrepp_d(1)
                           v_d(1, ij, 38) = v_d(1, ij, 38)+ij_matrix%pd_d(1, 13, k, l)*wrepp+ij_matrix%pd(13, k, l)*wrepp_d(1)
                           v_d(1, ij, 39) = v_d(1, ij, 39)+ij_matrix%pd_d(1, 14, k, l)*wrepp+ij_matrix%pd(14, k, l)*wrepp_d(1)
                           v_d(1, ij, 40) = v_d(1, ij, 40)+ij_matrix%pd_d(1, 15, k, l)*wrepp+ij_matrix%pd(15, k, l)*wrepp_d(1)

                           v_d(2, ij, 12) = v_d(2, ij, 12)+ij_matrix%pd_d(2, 1, k, l)*wrepp+ij_matrix%pd(1, k, l)*wrepp_d(2)
                           v_d(2, ij, 13) = v_d(2, ij, 13)+ij_matrix%pd_d(2, 2, k, l)*wrepp+ij_matrix%pd(2, k, l)*wrepp_d(2)
                           v_d(2, ij, 14) = v_d(2, ij, 14)+ij_matrix%pd_d(2, 3, k, l)*wrepp+ij_matrix%pd(3, k, l)*wrepp_d(2)
                           v_d(2, ij, 17) = v_d(2, ij, 17)+ij_matrix%pd_d(2, 4, k, l)*wrepp+ij_matrix%pd(4, k, l)*wrepp_d(2)
                           v_d(2, ij, 18) = v_d(2, ij, 18)+ij_matrix%pd_d(2, 5, k, l)*wrepp+ij_matrix%pd(5, k, l)*wrepp_d(2)
                           v_d(2, ij, 19) = v_d(2, ij, 19)+ij_matrix%pd_d(2, 6, k, l)*wrepp+ij_matrix%pd(6, k, l)*wrepp_d(2)
                           v_d(2, ij, 23) = v_d(2, ij, 23)+ij_matrix%pd_d(2, 7, k, l)*wrepp+ij_matrix%pd(7, k, l)*wrepp_d(2)
                           v_d(2, ij, 24) = v_d(2, ij, 24)+ij_matrix%pd_d(2, 8, k, l)*wrepp+ij_matrix%pd(8, k, l)*wrepp_d(2)
                           v_d(2, ij, 25) = v_d(2, ij, 25)+ij_matrix%pd_d(2, 9, k, l)*wrepp+ij_matrix%pd(9, k, l)*wrepp_d(2)
                           v_d(2, ij, 30) = v_d(2, ij, 30)+ij_matrix%pd_d(2, 10, k, l)*wrepp+ij_matrix%pd(10, k, l)*wrepp_d(2)
                           v_d(2, ij, 31) = v_d(2, ij, 31)+ij_matrix%pd_d(2, 11, k, l)*wrepp+ij_matrix%pd(11, k, l)*wrepp_d(2)
                           v_d(2, ij, 32) = v_d(2, ij, 32)+ij_matrix%pd_d(2, 12, k, l)*wrepp+ij_matrix%pd(12, k, l)*wrepp_d(2)
                           v_d(2, ij, 38) = v_d(2, ij, 38)+ij_matrix%pd_d(2, 13, k, l)*wrepp+ij_matrix%pd(13, k, l)*wrepp_d(2)
                           v_d(2, ij, 39) = v_d(2, ij, 39)+ij_matrix%pd_d(2, 14, k, l)*wrepp+ij_matrix%pd(14, k, l)*wrepp_d(2)
                           v_d(2, ij, 40) = v_d(2, ij, 40)+ij_matrix%pd_d(2, 15, k, l)*wrepp+ij_matrix%pd(15, k, l)*wrepp_d(2)

                           v_d(3, ij, 12) = v_d(3, ij, 12)+ij_matrix%pd_d(3, 1, k, l)*wrepp+ij_matrix%pd(1, k, l)*wrepp_d(3)
                           v_d(3, ij, 13) = v_d(3, ij, 13)+ij_matrix%pd_d(3, 2, k, l)*wrepp+ij_matrix%pd(2, k, l)*wrepp_d(3)
                           v_d(3, ij, 14) = v_d(3, ij, 14)+ij_matrix%pd_d(3, 3, k, l)*wrepp+ij_matrix%pd(3, k, l)*wrepp_d(3)
                           v_d(3, ij, 17) = v_d(3, ij, 17)+ij_matrix%pd_d(3, 4, k, l)*wrepp+ij_matrix%pd(4, k, l)*wrepp_d(3)
                           v_d(3, ij, 18) = v_d(3, ij, 18)+ij_matrix%pd_d(3, 5, k, l)*wrepp+ij_matrix%pd(5, k, l)*wrepp_d(3)
                           v_d(3, ij, 19) = v_d(3, ij, 19)+ij_matrix%pd_d(3, 6, k, l)*wrepp+ij_matrix%pd(6, k, l)*wrepp_d(3)
                           v_d(3, ij, 23) = v_d(3, ij, 23)+ij_matrix%pd_d(3, 7, k, l)*wrepp+ij_matrix%pd(7, k, l)*wrepp_d(3)
                           v_d(3, ij, 24) = v_d(3, ij, 24)+ij_matrix%pd_d(3, 8, k, l)*wrepp+ij_matrix%pd(8, k, l)*wrepp_d(3)
                           v_d(3, ij, 25) = v_d(3, ij, 25)+ij_matrix%pd_d(3, 9, k, l)*wrepp+ij_matrix%pd(9, k, l)*wrepp_d(3)
                           v_d(3, ij, 30) = v_d(3, ij, 30)+ij_matrix%pd_d(3, 10, k, l)*wrepp+ij_matrix%pd(10, k, l)*wrepp_d(3)
                           v_d(3, ij, 31) = v_d(3, ij, 31)+ij_matrix%pd_d(3, 11, k, l)*wrepp+ij_matrix%pd(11, k, l)*wrepp_d(3)
                           v_d(3, ij, 32) = v_d(3, ij, 32)+ij_matrix%pd_d(3, 12, k, l)*wrepp+ij_matrix%pd(12, k, l)*wrepp_d(3)
                           v_d(3, ij, 38) = v_d(3, ij, 38)+ij_matrix%pd_d(3, 13, k, l)*wrepp+ij_matrix%pd(13, k, l)*wrepp_d(3)
                           v_d(3, ij, 39) = v_d(3, ij, 39)+ij_matrix%pd_d(3, 14, k, l)*wrepp+ij_matrix%pd(14, k, l)*wrepp_d(3)
                           v_d(3, ij, 40) = v_d(3, ij, 40)+ij_matrix%pd_d(3, 15, k, l)*wrepp+ij_matrix%pd(15, k, l)*wrepp_d(3)
                        ELSE IF (mm == 6) THEN
                           k = k1-4
                           l = l1-4
                           v_d(1, ij, 15) = v_d(1, ij, 15)+ij_matrix%dd_d(1, 1, k, l)*wrepp+ij_matrix%dd(1, k, l)*wrepp_d(1)
                           v_d(1, ij, 21) = v_d(1, ij, 21)+ij_matrix%dd_d(1, 2, k, l)*wrepp+ij_matrix%dd(2, k, l)*wrepp_d(1)
                           v_d(1, ij, 28) = v_d(1, ij, 28)+ij_matrix%dd_d(1, 3, k, l)*wrepp+ij_matrix%dd(3, k, l)*wrepp_d(1)
                           v_d(1, ij, 36) = v_d(1, ij, 36)+ij_matrix%dd_d(1, 4, k, l)*wrepp+ij_matrix%dd(4, k, l)*wrepp_d(1)
                           v_d(1, ij, 45) = v_d(1, ij, 45)+ij_matrix%dd_d(1, 5, k, l)*wrepp+ij_matrix%dd(5, k, l)*wrepp_d(1)
                           v_d(1, ij, 20) = v_d(1, ij, 20)+ij_matrix%dd_d(1, 6, k, l)*wrepp+ij_matrix%dd(6, k, l)*wrepp_d(1)
                           v_d(1, ij, 26) = v_d(1, ij, 26)+ij_matrix%dd_d(1, 7, k, l)*wrepp+ij_matrix%dd(7, k, l)*wrepp_d(1)
                           v_d(1, ij, 27) = v_d(1, ij, 27)+ij_matrix%dd_d(1, 8, k, l)*wrepp+ij_matrix%dd(8, k, l)*wrepp_d(1)
                           v_d(1, ij, 33) = v_d(1, ij, 33)+ij_matrix%dd_d(1, 9, k, l)*wrepp+ij_matrix%dd(9, k, l)*wrepp_d(1)
                           v_d(1, ij, 34) = v_d(1, ij, 34)+ij_matrix%dd_d(1, 10, k, l)*wrepp+ij_matrix%dd(10, k, l)*wrepp_d(1)
                           v_d(1, ij, 35) = v_d(1, ij, 35)+ij_matrix%dd_d(1, 11, k, l)*wrepp+ij_matrix%dd(11, k, l)*wrepp_d(1)
                           v_d(1, ij, 41) = v_d(1, ij, 41)+ij_matrix%dd_d(1, 12, k, l)*wrepp+ij_matrix%dd(12, k, l)*wrepp_d(1)
                           v_d(1, ij, 42) = v_d(1, ij, 42)+ij_matrix%dd_d(1, 13, k, l)*wrepp+ij_matrix%dd(13, k, l)*wrepp_d(1)
                           v_d(1, ij, 43) = v_d(1, ij, 43)+ij_matrix%dd_d(1, 14, k, l)*wrepp+ij_matrix%dd(14, k, l)*wrepp_d(1)
                           v_d(1, ij, 44) = v_d(1, ij, 44)+ij_matrix%dd_d(1, 15, k, l)*wrepp+ij_matrix%dd(15, k, l)*wrepp_d(1)

                           v_d(2, ij, 15) = v_d(2, ij, 15)+ij_matrix%dd_d(2, 1, k, l)*wrepp+ij_matrix%dd(1, k, l)*wrepp_d(2)
                           v_d(2, ij, 21) = v_d(2, ij, 21)+ij_matrix%dd_d(2, 2, k, l)*wrepp+ij_matrix%dd(2, k, l)*wrepp_d(2)
                           v_d(2, ij, 28) = v_d(2, ij, 28)+ij_matrix%dd_d(2, 3, k, l)*wrepp+ij_matrix%dd(3, k, l)*wrepp_d(2)
                           v_d(2, ij, 36) = v_d(2, ij, 36)+ij_matrix%dd_d(2, 4, k, l)*wrepp+ij_matrix%dd(4, k, l)*wrepp_d(2)
                           v_d(2, ij, 45) = v_d(2, ij, 45)+ij_matrix%dd_d(2, 5, k, l)*wrepp+ij_matrix%dd(5, k, l)*wrepp_d(2)
                           v_d(2, ij, 20) = v_d(2, ij, 20)+ij_matrix%dd_d(2, 6, k, l)*wrepp+ij_matrix%dd(6, k, l)*wrepp_d(2)
                           v_d(2, ij, 26) = v_d(2, ij, 26)+ij_matrix%dd_d(2, 7, k, l)*wrepp+ij_matrix%dd(7, k, l)*wrepp_d(2)
                           v_d(2, ij, 27) = v_d(2, ij, 27)+ij_matrix%dd_d(2, 8, k, l)*wrepp+ij_matrix%dd(8, k, l)*wrepp_d(2)
                           v_d(2, ij, 33) = v_d(2, ij, 33)+ij_matrix%dd_d(2, 9, k, l)*wrepp+ij_matrix%dd(9, k, l)*wrepp_d(2)
                           v_d(2, ij, 34) = v_d(2, ij, 34)+ij_matrix%dd_d(2, 10, k, l)*wrepp+ij_matrix%dd(10, k, l)*wrepp_d(2)
                           v_d(2, ij, 35) = v_d(2, ij, 35)+ij_matrix%dd_d(2, 11, k, l)*wrepp+ij_matrix%dd(11, k, l)*wrepp_d(2)
                           v_d(2, ij, 41) = v_d(2, ij, 41)+ij_matrix%dd_d(2, 12, k, l)*wrepp+ij_matrix%dd(12, k, l)*wrepp_d(2)
                           v_d(2, ij, 42) = v_d(2, ij, 42)+ij_matrix%dd_d(2, 13, k, l)*wrepp+ij_matrix%dd(13, k, l)*wrepp_d(2)
                           v_d(2, ij, 43) = v_d(2, ij, 43)+ij_matrix%dd_d(2, 14, k, l)*wrepp+ij_matrix%dd(14, k, l)*wrepp_d(2)
                           v_d(2, ij, 44) = v_d(2, ij, 44)+ij_matrix%dd_d(2, 15, k, l)*wrepp+ij_matrix%dd(15, k, l)*wrepp_d(2)

                           v_d(3, ij, 15) = v_d(3, ij, 15)+ij_matrix%dd_d(3, 1, k, l)*wrepp+ij_matrix%dd(1, k, l)*wrepp_d(3)
                           v_d(3, ij, 21) = v_d(3, ij, 21)+ij_matrix%dd_d(3, 2, k, l)*wrepp+ij_matrix%dd(2, k, l)*wrepp_d(3)
                           v_d(3, ij, 28) = v_d(3, ij, 28)+ij_matrix%dd_d(3, 3, k, l)*wrepp+ij_matrix%dd(3, k, l)*wrepp_d(3)
                           v_d(3, ij, 36) = v_d(3, ij, 36)+ij_matrix%dd_d(3, 4, k, l)*wrepp+ij_matrix%dd(4, k, l)*wrepp_d(3)
                           v_d(3, ij, 45) = v_d(3, ij, 45)+ij_matrix%dd_d(3, 5, k, l)*wrepp+ij_matrix%dd(5, k, l)*wrepp_d(3)
                           v_d(3, ij, 20) = v_d(3, ij, 20)+ij_matrix%dd_d(3, 6, k, l)*wrepp+ij_matrix%dd(6, k, l)*wrepp_d(3)
                           v_d(3, ij, 26) = v_d(3, ij, 26)+ij_matrix%dd_d(3, 7, k, l)*wrepp+ij_matrix%dd(7, k, l)*wrepp_d(3)
                           v_d(3, ij, 27) = v_d(3, ij, 27)+ij_matrix%dd_d(3, 8, k, l)*wrepp+ij_matrix%dd(8, k, l)*wrepp_d(3)
                           v_d(3, ij, 33) = v_d(3, ij, 33)+ij_matrix%dd_d(3, 9, k, l)*wrepp+ij_matrix%dd(9, k, l)*wrepp_d(3)
                           v_d(3, ij, 34) = v_d(3, ij, 34)+ij_matrix%dd_d(3, 10, k, l)*wrepp+ij_matrix%dd(10, k, l)*wrepp_d(3)
                           v_d(3, ij, 35) = v_d(3, ij, 35)+ij_matrix%dd_d(3, 11, k, l)*wrepp+ij_matrix%dd(11, k, l)*wrepp_d(3)
                           v_d(3, ij, 41) = v_d(3, ij, 41)+ij_matrix%dd_d(3, 12, k, l)*wrepp+ij_matrix%dd(12, k, l)*wrepp_d(3)
                           v_d(3, ij, 42) = v_d(3, ij, 42)+ij_matrix%dd_d(3, 13, k, l)*wrepp+ij_matrix%dd(13, k, l)*wrepp_d(3)
                           v_d(3, ij, 43) = v_d(3, ij, 43)+ij_matrix%dd_d(3, 14, k, l)*wrepp+ij_matrix%dd(14, k, l)*wrepp_d(3)
                           v_d(3, ij, 44) = v_d(3, ij, 44)+ij_matrix%dd_d(3, 15, k, l)*wrepp+ij_matrix%dd(15, k, l)*wrepp_d(3)
                        END IF
                     END IF
                  END DO
               END DO
               DO kl = 1, limkl
                  logv_d(ij, kl) = (ANY(ABS(v_d(1:3, ij, kl)) > 0.0_dp))
               END DO
            END DO
         END DO
         IF (debug_this_module) THEN
            CALL rot_2el_2c_first_debug(sepi, sepj, rijv, se_int_control, se_taper, invert, ii, kk, v_d)
         END IF
      END IF
   END SUBROUTINE rot_2el_2c_first

! **************************************************************************************************
!> \brief Store the two-electron two-center integrals in diagonal form
!>
!> \param limij ...
!> \param limkl ...
!> \param ww ...
!> \param w ...
!> \param ww_dx ...
!> \param ww_dy ...
!> \param ww_dz ...
!> \param dw ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE store_2el_2c_diag(limij, limkl, ww, w, ww_dx, ww_dy, ww_dz, dw)

      INTEGER, INTENT(IN)                                :: limij, limkl
      REAL(KIND=dp), DIMENSION(limkl, limij), &
         INTENT(IN), OPTIONAL                            :: ww
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT), &
         OPTIONAL                                        :: w
      REAL(KIND=dp), DIMENSION(limkl, limij), &
         INTENT(IN), OPTIONAL                            :: ww_dx, ww_dy, ww_dz
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: dw

      CHARACTER(len=*), PARAMETER :: routineN = 'store_2el_2c_diag', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ij, kl, l

      l = 0
      IF (PRESENT(ww) .AND. PRESENT(w)) THEN
         DO ij = 1, limij
            DO kl = 1, limkl
               l = l+1
               w(l) = ww(kl, ij)
            END DO
         END DO
      ELSE IF (PRESENT(ww_dx) .AND. PRESENT(ww_dy) .AND. PRESENT(ww_dz) .AND. PRESENT(dw)) THEN
         DO ij = 1, limij
            DO kl = 1, limkl
               l = l+1
               dw(1, l) = ww_dx(kl, ij)
               dw(2, l) = ww_dy(kl, ij)
               dw(3, l) = ww_dz(kl, ij)
            END DO
         END DO
      ELSE
         CALL cp__b("semi_empirical_int_utils.F",2446,"")
      END IF

   END SUBROUTINE store_2el_2c_diag

END MODULE semi_empirical_int_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Integrals for semi-empiric methods
!> \par History
!>      JGH (27.10.2004) : separate routine for nuclear attraction integrals
!>      JGH (13.03.2006) : tapering function
!>      Teodoro Laino (03.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
!> \author JGH (11.10.2004)
! **************************************************************************************************
MODULE semi_empirical_int_num

   USE input_constants,                 ONLY: do_method_am1,&
                                              do_method_pchg,&
                                              do_method_pdg,&
                                              do_method_pm3,&
                                              do_method_pm6,&
                                              do_method_undef,&
                                              do_se_IS_kdso_d
   USE kinds,                           ONLY: dp
   USE multipole_types,                 ONLY: do_multipole_none
   USE physcon,                         ONLY: angstrom,&
                                              evolt
   USE semi_empirical_int_arrays,       ONLY: &
        ijkl_ind, ijkl_sym, inddd, inddp, indexa, indexb, indpp, int2c_type, l_index, rij_threshold
   USE semi_empirical_int_utils,        ONLY: ijkl_d,&
                                              ijkl_sp,&
                                              rot_2el_2c_first,&
                                              rotmat,&
                                              store_2el_2c_diag
   USE semi_empirical_types,            ONLY: rotmat_create,&
                                              rotmat_release,&
                                              rotmat_type,&
                                              se_int_control_type,&
                                              se_int_screen_type,&
                                              se_taper_type,&
                                              semi_empirical_type,&
                                              setup_se_int_control_type
   USE taper_types,                     ONLY: taper_eval
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_int_num'
   PUBLIC :: rotint_num, rotnuc_num, corecore_num, &
             drotint_num, drotnuc_num, dcorecore_num, &
             ssss_nucint_num, core_nucint_num, terep_num, &
             nucint_sp_num, terep_sp_num, terep_d_num, &
             nucint_d_num, corecore_el_num, dcorecore_el_num

CONTAINS

! **************************************************************************************************
!> \brief Computes the two particle interactions in the lab frame
!> \param sepi Atomic parameters of first atom
!> \param sepj Atomic parameters of second atom
!> \param rijv Coordinate vector i -> j
!> \param w Array of two-electron repulsion integrals.
!> \param se_int_control input parameters that control the calculation of SE
!>                           integrals (shortrange, R3 residual, screening type)
!> \param se_taper ...
!> \note routine adapted from mopac7 (rotate)
!>       written by Ernest R. Davidson, Indiana University.
!>       Teodoro Laino [tlaino] - University of Zurich 04.2008 : major rewriting
! **************************************************************************************************
   SUBROUTINE rotint_num(sepi, sepj, rijv, w, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rijv
      REAL(dp), DIMENSION(2025), INTENT(OUT)             :: w
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'rotint_num', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, i1, ii, ij, ij1, iminus, istep, &
                                                            iw_loc, j, j1, jj, k, kk, kl, l, &
                                                            limij, limkl, mm
      LOGICAL, DIMENSION(45, 45)                         :: logv
      REAL(dp)                                           :: rij
      REAL(KIND=dp)                                      :: cc, wrepp
      REAL(KIND=dp), DIMENSION(2025)                     :: ww
      REAL(KIND=dp), DIMENSION(45, 45)                   :: v
      REAL(KIND=dp), DIMENSION(491)                      :: rep
      TYPE(rotmat_type), POINTER                         :: ij_matrix

      NULLIFY (ij_matrix)
      rij = DOT_PRODUCT(rijv, rijv)
      IF (rij > rij_threshold) THEN
         ! The repulsion integrals over molecular frame (w) are stored in the
         ! order in which they will later be used.  ie.  (i,j/k,l) where
         ! j.le.i  and  l.le.k     and l varies most rapidly and i least
         ! rapidly.  (anti-normal computer storage)
         rij = SQRT(rij)

         ! Create the rotation matrix
         CALL rotmat_create(ij_matrix)
         CALL rotmat(sepi, sepj, rijv, rij, ij_matrix, do_derivatives=.FALSE.)

         ! Compute Integrals in Diatomic Frame
         CALL terep_num(sepi, sepj, rij, rep, se_taper=se_taper, se_int_control=se_int_control)

         ! Rotate Integrals
         ii = sepi%natorb
         kk = sepj%natorb
         IF (ii*kk > 0) THEN
            limij = sepi%atm_int_size
            limkl = sepj%atm_int_size
            istep = limkl*limij
            DO i1 = 1, istep
               ww(i1) = 0.0_dp
            END DO

            ! First step in rotation of integrals
            CALL rot_2el_2c_first(sepi, sepj, rijv, se_int_control, se_taper, .FALSE., ii, kk, rep, &
                                  logv, ij_matrix, v, lgrad=.FALSE.)

            ! Second step in rotation of integrals
            DO i1 = 1, ii
               DO j1 = 1, i1
                  ij = indexa(i1, j1)
                  jj = indexb(i1, j1)
                  mm = int2c_type(jj)
                  DO k = 1, kk
                     DO l = 1, k
                        kl = indexb(k, l)
                        IF (logv(ij, kl)) THEN
                           wrepp = v(ij, kl)
                           SELECT CASE (mm)
                           CASE (1)
                              ! (SS/)
                              i = 1
                              j = 1
                              iw_loc = (indexb(i, j)-1)*limkl+kl
                              ww(iw_loc) = wrepp
                           CASE (2)
                              ! (SP/)
                              j = 1
                              DO i = 1, 3
                                 iw_loc = (indexb(i+1, j)-1)*limkl+kl
                                 ww(iw_loc) = ww(iw_loc)+ij_matrix%sp(i1-1, i)*wrepp
                              END DO
                           CASE (3)
                              ! (PP/)
                              DO i = 1, 3
                                 cc = ij_matrix%pp(i, i1-1, j1-1)
                                 iw_loc = (indexb(i+1, i+1)-1)*limkl+kl
                                 ww(iw_loc) = ww(iw_loc)+cc*wrepp
                                 iminus = i-1
                                 IF (iminus /= 0) THEN
                                    DO j = 1, iminus
                                       cc = ij_matrix%pp(1+i+j, i1-1, j1-1)
                                       iw_loc = (indexb(i+1, j+1)-1)*limkl+kl
                                       ww(iw_loc) = ww(iw_loc)+cc*wrepp
                                    END DO
                                 END IF
                              END DO
                           CASE (4)
                              ! (SD/)
                              j = 1
                              DO i = 1, 5
                                 iw_loc = (indexb(i+4, j)-1)*limkl+kl
                                 ww(iw_loc) = ww(iw_loc)+ij_matrix%sd(i1-4, i)*wrepp
                              END DO
                           CASE (5)
                              ! (DP/)
                              DO i = 1, 5
                                 DO j = 1, 3
                                    iw_loc = (indexb(i+4, j+1)-1)*limkl+kl
                                    ij1 = 3*(i-1)+j
                                    ww(iw_loc) = ww(iw_loc)+ij_matrix%pd(ij1, i1-4, j1-1)*wrepp
                                 END DO
                              END DO
                           CASE (6)
                              ! (DD/)
                              DO i = 1, 5
                                 cc = ij_matrix%dd(i, i1-4, j1-4)
                                 iw_loc = (indexb(i+4, i+4)-1)*limkl+kl
                                 ww(iw_loc) = ww(iw_loc)+cc*wrepp
                                 iminus = i-1
                                 IF (iminus /= 0) THEN
                                    DO j = 1, iminus
                                       ij1 = inddd(i, j)
                                       cc = ij_matrix%dd(ij1, i1-4, j1-4)
                                       iw_loc = (indexb(i+4, j+4)-1)*limkl+kl
                                       ww(iw_loc) = ww(iw_loc)+cc*wrepp
                                    END DO
                                 END IF
                              END DO
                           END SELECT
                        END IF
                     END DO
                  END DO
               END DO
            END DO
         END IF
         CALL rotmat_release(ij_matrix)

         ! Store two electron integrals in the triangular format
         CALL store_2el_2c_diag(limij, limkl, ww, w)
      ENDIF
   END SUBROUTINE rotint_num

! **************************************************************************************************
!> \brief Calculates the derivative pf two-electron repulsion integrals and the
!>      nuclear attraction integrals w.r.t. |r|
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij interatomic distance
!> \param rep array of two-electron repulsion integrals
!> \param se_taper ...
!> \param se_int_control input parameters that control the calculation of SE
!>                         integrals (shortrange, R3 residual, screening type)
!> \par History
!>      03.2008 created [tlaino]
!> \author Teodoro Laino [tlaino] - Zurich University
! **************************************************************************************************
   SUBROUTINE terep_num(sepi, sepj, rij, rep, se_taper, se_int_control)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), DIMENSION(491), INTENT(OUT)              :: rep
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'terep_num', routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: ft
      TYPE(se_int_screen_type)                           :: se_int_screen

      ft = taper_eval(se_taper%taper, rij)
      ! In case of dumped integrals compute an additional taper term
      IF (se_int_control%integral_screening == do_se_IS_kdso_d) THEN
         se_int_screen%ft = taper_eval(se_taper%taper_add, rij)
      END IF

      ! Contribution from sp shells
      CALL terep_sp_num(sepi, sepj, rij, rep, se_int_control, se_int_screen, ft)

      IF (sepi%dorb .OR. sepj%dorb) THEN
         ! Compute the contribution from d shells
         CALL terep_d_num(sepi, sepj, rij, rep, se_int_control, se_int_screen, &
                          ft)
      END IF

   END SUBROUTINE terep_num

! **************************************************************************************************
!> \brief Calculates the  two-electron repulsion integrals - sp shell only
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij ...
!> \param rep array of two-electron repulsion integrals
!> \param se_int_control input parameters that control the calculation of SE
!>                         integrals (shortrange, R3 residual, screening type)
!> \param se_int_screen contains information for computing the screened
!>                         integrals KDSO-D
!> \param ft ...
!> \par History
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
! **************************************************************************************************
   SUBROUTINE terep_sp_num(sepi, sepj, rij, rep, se_int_control, se_int_screen, &
                           ft)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), DIMENSION(491), INTENT(OUT)              :: rep
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen
      REAL(dp), INTENT(IN)                               :: ft

      CHARACTER(len=*), PARAMETER :: routineN = 'terep_sp_num', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ij, j, k, kl, l, lasti, lastj, li, &
                                                            lj, lk, ll, nold, numb
      REAL(KIND=dp)                                      :: tmp

      lasti = sepi%natorb
      lastj = sepj%natorb
      DO i = 1, MIN(lasti, 4)
         li = l_index(i)
         DO j = 1, i
            lj = l_index(j)
            ij = indexa(i, j)
            DO k = 1, MIN(lastj, 4)
               lk = l_index(k)
               DO l = 1, k
                  ll = l_index(l)
                  kl = indexa(k, l)

                  numb = ijkl_ind(ij, kl)
                  IF (numb > 0) THEN
                     nold = ijkl_sym(numb)
                     IF (nold > 0) THEN
                        rep(numb) = rep(nold)
                     ELSE IF (nold < 0) THEN
                        rep(numb) = -rep(-nold)
                     ELSE IF (nold == 0) THEN
                        tmp = ijkl_sp(sepi, sepj, ij, kl, li, lj, lk, ll, 0, rij, &
                                      se_int_control, se_int_screen, do_method_undef) &
                              *ft
                        rep(numb) = tmp
                     END IF
                  END IF
               END DO
            END DO
         END DO
      END DO
   END SUBROUTINE terep_sp_num

! **************************************************************************************************
!> \brief Calculates the  two-electron repulsion integrals - d shell only
!> \param sepi ...
!> \param sepj ...
!> \param rij interatomic distance
!> \param rep array of two-electron repulsion integrals
!> \param se_int_control input parameters that control the calculation of SE
!>                         integrals (shortrange, R3 residual, screening type)
!> \param se_int_screen contains information for computing the screened
!>                         integrals KDSO-D
!> \param ft ...
!> \par History
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
! **************************************************************************************************
   SUBROUTINE terep_d_num(sepi, sepj, rij, rep, se_int_control, se_int_screen, &
                          ft)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), DIMENSION(491), INTENT(INOUT)            :: rep
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen
      REAL(dp), INTENT(IN)                               :: ft

      CHARACTER(len=*), PARAMETER :: routineN = 'terep_d_num', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ij, j, k, kl, l, lasti, lastj, li, &
                                                            lj, lk, ll, nold, numb
      REAL(KIND=dp)                                      :: tmp

      lasti = sepi%natorb
      lastj = sepj%natorb
      DO i = 1, lasti
         li = l_index(i)
         DO j = 1, i
            lj = l_index(j)
            ij = indexa(i, j)
            DO k = 1, lastj
               lk = l_index(k)
               DO l = 1, k
                  ll = l_index(l)
                  kl = indexa(k, l)

                  numb = ijkl_ind(ij, kl)
                  ! From 1 to 34 we store integrals involving sp shells
                  IF (numb > 34) THEN
                     nold = ijkl_sym(numb)
                     IF (nold > 34) THEN
                        rep(numb) = rep(nold)
                     ELSE IF (nold < -34) THEN
                        rep(numb) = -rep(-nold)
                     ELSE IF (nold == 0) THEN
                        tmp = ijkl_d(sepi, sepj, ij, kl, li, lj, lk, ll, 0, rij, &
                                     se_int_control, se_int_screen, do_method_undef) &
                              *ft
                        rep(numb) = tmp
                     END IF
                  END IF
               END DO
            END DO
         END DO
      END DO

   END SUBROUTINE terep_d_num

! **************************************************************************************************
!> \brief Computes the two-particle interactions.
!> \param sepi Atomic parameters of first atom
!> \param sepj Atomic parameters of second atom
!> \param rijv Coordinate vector i -> j
!> \param e1b Array of electron-nuclear attraction integrals,
!>                       e1b = Electron on atom ni attracting nucleus of nj.
!> \param e2a Array of electron-nuclear attraction integrals,
!>                       e2a = Electron on atom nj attracting nucleus of ni.
!> \param itype ...
!> \param se_int_control ...
!> \param se_taper ...
!> \note routine adapted from mopac7 (rotate)
!>       written by Ernest R. Davidson, Indiana University.
!>       Teodoro Laino [tlaino] - University of Zurich 04.2008 : major rewriting
!>       Teodoro Laino [tlaino] - University of Zurich 04.2008 : removed the core-core part
! **************************************************************************************************
   SUBROUTINE rotnuc_num(sepi, sepj, rijv, e1b, e2a, itype, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rijv
      REAL(dp), DIMENSION(45), INTENT(OUT), OPTIONAL     :: e1b, e2a
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'rotnuc_num', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, idd, idp, ind1, ind2, ipp, j, &
                                                            last_orbital(2), m, n
      LOGICAL                                            :: l_e1b, l_e2a, task(2)
      REAL(dp)                                           :: rij
      REAL(dp), DIMENSION(10, 2)                         :: core
      REAL(dp), DIMENSION(45)                            :: tmp
      TYPE(rotmat_type), POINTER                         :: ij_matrix

      NULLIFY (ij_matrix)
      l_e1b = PRESENT(e1b)
      l_e2a = PRESENT(e2a)
      rij = DOT_PRODUCT(rijv, rijv)

      IF (rij > rij_threshold) THEN
         rij = SQRT(rij)
         ! Create the rotation matrix
         CALL rotmat_create(ij_matrix)
         CALL rotmat(sepi, sepj, rijv, rij, ij_matrix, do_derivatives=.FALSE.)

         ! Compute Integrals in Diatomic Frame
         CALL core_nucint_num(sepi, sepj, rij, core=core, itype=itype, se_taper=se_taper, &
                              se_int_control=se_int_control)

         ! Copy parameters over to arrays for do loop.
         last_orbital(1) = sepi%natorb
         last_orbital(2) = sepj%natorb
         task(1) = l_e1b
         task(2) = l_e2a
         DO n = 1, 2
            IF (.NOT. task(n)) CYCLE
            DO i = 1, last_orbital(n)
               ind1 = i-1
               DO j = 1, i
                  ind2 = j-1
                  m = (i*(i-1))/2+j
                  ! Perform Rotations ...
                  IF (ind2 == 0) THEN
                     IF (ind1 == 0) THEN
                        ! Type of Integral (SS/)
                        tmp(m) = core(1, n)
                     ELSE IF (ind1 < 4) THEN
                        ! Type of Integral (SP/)
                        tmp(m) = ij_matrix%sp(1, ind1)*core(2, n)
                     ELSE
                        ! Type of Integral (SD/)
                        tmp(m) = ij_matrix%sd(1, ind1-3)*core(5, n)
                     END IF
                  ELSE IF (ind2 < 4) THEN
                     IF (ind1 < 4) THEN
                        ! Type of Integral (PP/)
                        ipp = indpp(ind1, ind2)
                        tmp(m) = core(3, n)*ij_matrix%pp(ipp, 1, 1)+ &
                                 core(4, n)*(ij_matrix%pp(ipp, 2, 2)+ij_matrix%pp(ipp, 3, 3))
                     ELSE
                        ! Type of Integral (PD/)
                        idp = inddp(ind1-3, ind2)
                        tmp(m) = core(6, n)*ij_matrix%pd(idp, 1, 1)+ &
                                 core(8, n)*(ij_matrix%pd(idp, 2, 2)+ij_matrix%pd(idp, 3, 3))
                     END IF
                  ELSE
                     ! Type of Integral (DD/)
                     idd = inddd(ind1-3, ind2-3)
                     tmp(m) = core(7, n)*ij_matrix%dd(idd, 1, 1)+ &
                              core(9, n)*(ij_matrix%dd(idd, 2, 2)+ij_matrix%dd(idd, 3, 3))+ &
                              core(10, n)*(ij_matrix%dd(idd, 4, 4)+ij_matrix%dd(idd, 5, 5))
                  END IF
               END DO
            END DO
            IF (n == 1) THEN
               DO i = 1, sepi%atm_int_size
                  e1b(i) = -tmp(i)
               END DO
            END IF
            IF (n == 2) THEN
               DO i = 1, sepj%atm_int_size
                  e2a(i) = -tmp(i)
               END DO
            END IF
         END DO
         CALL rotmat_release(ij_matrix)
      END IF
   END SUBROUTINE rotnuc_num

! **************************************************************************************************
!> \brief Computes the core-core interactions.
!> \param sepi Atomic parameters of first atom
!> \param sepj Atomic parameters of second atom
!> \param rijv Coordinate vector i -> j
!> \param enuc nuclear-nuclear repulsion term.
!> \param itype ...
!> \param se_int_control input parameters that control the calculation of SE
!>                           integrals (shortrange, R3 residual, screening type)
!> \param se_taper ...
!> \note routine adapted from mopac7 (rotate)
!>       written by Ernest R. Davidson, Indiana University.
!>       Teodoro Laino [tlaino] - University of Zurich 04.2008 : major rewriting
!>       Teodoro Laino [tlaino] - University of Zurich 04.2008 : splitted from rotnuc
! **************************************************************************************************
   SUBROUTINE corecore_num(sepi, sepj, rijv, enuc, itype, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rijv
      REAL(dp), INTENT(OUT)                              :: enuc
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'corecore_num', routineP = moduleN//':'//routineN

      INTEGER                                            :: ig, nt
      REAL(dp)                                           :: aab, alpi, alpj, apdg, ax, dai, daj, &
                                                            dbi, dbj, pai, paj, pbi, pbj, qcorr, &
                                                            rij, rija, scale, ssss, ssss_sr, tmp, &
                                                            xab, zaf, zbf, zz
      REAL(dp), DIMENSION(4)                             :: fni1, fni2, fni3, fnj1, fnj2, fnj3
      TYPE(se_int_control_type)                          :: se_int_control_off

      rij = DOT_PRODUCT(rijv, rijv)
      enuc = 0.0_dp
      IF (rij > rij_threshold) THEN
         rij = SQRT(rij)
         CALL setup_se_int_control_type(se_int_control_off, shortrange=.FALSE., do_ewald_r3=.FALSE., &
                                        do_ewald_gks=.FALSE., integral_screening=se_int_control%integral_screening, &
                                        max_multipole=do_multipole_none, pc_coulomb_int=.FALSE.)
         CALL ssss_nucint_num(sepi, sepj, rij, ssss=ssss, itype=itype, se_taper=se_taper, &
                              se_int_control=se_int_control_off)
         ! In case let's compute the short-range part of the (ss|ss) integral
         IF (se_int_control%shortrange) THEN
            CALL ssss_nucint_num(sepi, sepj, rij, ssss=ssss_sr, itype=itype, se_taper=se_taper, &
                                 se_int_control=se_int_control)
         ELSE
            ssss_sr = ssss
         END IF
         zz = sepi%zeff*sepj%zeff
         ! Nuclear-Nuclear electrostatic contribution
         enuc = zz*ssss_sr
         ! Method dependent correction terms
         IF (itype /= do_method_pm6) THEN
            alpi = sepi%alp
            alpj = sepj%alp
            scale = EXP(-alpi*rij)+EXP(-alpj*rij)

            nt = sepi%z+sepj%z
            IF (nt == 8 .OR. nt == 9) THEN
               IF (sepi%z == 7 .OR. sepi%z == 8) scale = scale+(angstrom*rij-1._dp)*EXP(-alpi*rij)
               IF (sepj%z == 7 .OR. sepj%z == 8) scale = scale+(angstrom*rij-1._dp)*EXP(-alpj*rij)
            ENDIF
            scale = ABS(scale*zz*ssss)
            zz = zz/rij
            IF (itype == do_method_am1 .OR. itype == do_method_pm3 .OR. itype == do_method_pdg) THEN
               IF (itype == do_method_am1 .AND. sepi%z == 5) THEN
                  !special case AM1 Boron
                  SELECT CASE (sepj%z)
                  CASE DEFAULT
                     nt = 1
                  CASE (1)
                     nt = 2
                  CASE (6)
                     nt = 3
                  CASE (9, 17, 35, 53)
                     nt = 4
                  END SELECT
                  fni1(:) = sepi%bfn1(:, nt)
                  fni2(:) = sepi%bfn2(:, nt)
                  fni3(:) = sepi%bfn3(:, nt)
               ELSE
                  fni1(:) = sepi%fn1(:)
                  fni2(:) = sepi%fn2(:)
                  fni3(:) = sepi%fn3(:)
               END IF
               IF (itype == do_method_am1 .AND. sepj%z == 5) THEN
                  !special case AM1 Boron
                  SELECT CASE (sepi%z)
                  CASE DEFAULT
                     nt = 1
                  CASE (1)
                     nt = 2
                  CASE (6)
                     nt = 3
                  CASE (9, 17, 35, 53)
                     nt = 4
                  END SELECT
                  fnj1(:) = sepj%bfn1(:, nt)
                  fnj2(:) = sepj%bfn2(:, nt)
                  fnj3(:) = sepj%bfn3(:, nt)
               ELSE
                  fnj1(:) = sepj%fn1(:)
                  fnj2(:) = sepj%fn2(:)
                  fnj3(:) = sepj%fn3(:)
               END IF
               ! AM1/PM3/PDG correction to nuclear repulsion
               DO ig = 1, SIZE(fni1)
                  IF (ABS(fni1(ig)) > 0._dp) THEN
                     ax = fni2(ig)*(rij-fni3(ig))**2
                     IF (ax <= 25._dp) THEN
                        scale = scale+zz*fni1(ig)*EXP(-ax)
                     ENDIF
                  ENDIF
                  IF (ABS(fnj1(ig)) > 0._dp) THEN
                     ax = fnj2(ig)*(rij-fnj3(ig))**2
                     IF (ax <= 25._dp) THEN
                        scale = scale+zz*fnj1(ig)*EXP(-ax)
                     ENDIF
                  ENDIF
               END DO
            ENDIF
            IF (itype == do_method_pdg) THEN
               ! PDDG function
               zaf = sepi%zeff/nt
               zbf = sepj%zeff/nt
               pai = sepi%pre(1)
               pbi = sepi%pre(2)
               paj = sepj%pre(1)
               pbj = sepj%pre(2)
               dai = sepi%d(1)
               dbi = sepi%d(2)
               daj = sepj%d(1)
               dbj = sepj%d(2)
               apdg = 10._dp*angstrom**2
               qcorr = &
                  (zaf*pai+zbf*paj)*EXP(-apdg*(rij-dai-daj)**2)+ &
                  (zaf*pai+zbf*pbj)*EXP(-apdg*(rij-dai-dbj)**2)+ &
                  (zaf*pbi+zbf*paj)*EXP(-apdg*(rij-dbi-daj)**2)+ &
                  (zaf*pbi+zbf*pbj)*EXP(-apdg*(rij-dbi-dbj)**2)
            ELSEIF (itype == do_method_pchg) THEN
               qcorr = 0.0_dp
               scale = 0.0_dp
            ELSE
               qcorr = 0.0_dp
            END IF
         ELSE
            rija = rij*angstrom
            scale = zz*ssss
            ! PM6 core-core terms
            xab = sepi%xab(sepj%z)
            aab = sepi%aab(sepj%z)
            IF ((sepi%z == 1 .AND. (sepj%z == 6 .OR. sepj%z == 7 .OR. sepj%z == 8)) .OR. &
                (sepj%z == 1 .AND. (sepi%z == 6 .OR. sepi%z == 7 .OR. sepi%z == 8))) THEN
               ! Special Case O-H or N-H or C-H
               scale = scale*(2._dp*xab*EXP(-aab*rija*rija))
            ELSEIF (sepi%z == 6 .AND. sepj%z == 6) THEN
               ! Special Case C-C
               scale = scale*(2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6))+9.28_dp*EXP(-5.98_dp*rija))
            ELSEIF ((sepi%z == 8 .AND. sepj%z == 14) .OR. &
                    (sepj%z == 8 .AND. sepi%z == 14)) THEN
               ! Special Case Si-O
               scale = scale*(2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6))-0.0007_dp*EXP(-(rija-2.9_dp)**2))
            ELSE
               ! General Case
               ! Factor of 2 found by experiment
               scale = scale*(2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6)))
            END IF
            ! General correction term a*exp(-b*(rij-c)^2)
            qcorr = (sepi%a*EXP(-sepi%b*(rij-sepi%c)**2))*sepi%zeff*sepj%zeff/rij+ &
                    (sepj%a*EXP(-sepj%b*(rij-sepj%c)**2))*sepi%zeff*sepj%zeff/rij
            ! Hard core repulsion
            tmp = (REAL(sepi%z, dp)**(1._dp/3._dp)+REAL(sepj%z, dp)**(1._dp/3._dp))
            qcorr = qcorr+1.e-8_dp/evolt*(tmp/rija)**12
         END IF
         enuc = enuc+scale+qcorr
      END IF
   END SUBROUTINE corecore_num

! **************************************************************************************************
!> \brief Computes the electrostatic core-core interactions only.
!> \param sepi Atomic parameters of first atom
!> \param sepj Atomic parameters of second atom
!> \param rijv Coordinate vector i -> j
!> \param enuc nuclear-nuclear repulsion term.
!> \param itype ...
!> \param se_int_control input parameters that control the calculation of SE
!>                           integrals (shortrange, R3 residual, screening type)
!> \param se_taper ...
!> \author Teodoro Laino [tlaino] - 05.2009
! **************************************************************************************************
   SUBROUTINE corecore_el_num(sepi, sepj, rijv, enuc, itype, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rijv
      REAL(dp), INTENT(OUT)                              :: enuc
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'corecore_el_num', &
         routineP = moduleN//':'//routineN

      REAL(dp)                                           :: rij, ssss, ssss_sr, zz
      TYPE(se_int_control_type)                          :: se_int_control_off

      rij = DOT_PRODUCT(rijv, rijv)
      enuc = 0.0_dp
      IF (rij > rij_threshold) THEN
         rij = SQRT(rij)
         CALL setup_se_int_control_type(se_int_control_off, shortrange=.FALSE., do_ewald_r3=.FALSE., &
                                        do_ewald_gks=.FALSE., integral_screening=se_int_control%integral_screening, &
                                        max_multipole=do_multipole_none, pc_coulomb_int=.FALSE.)
         CALL ssss_nucint_num(sepi, sepj, rij, ssss=ssss, itype=itype, se_taper=se_taper, &
                              se_int_control=se_int_control_off)
         ! In case let's compute the short-range part of the (ss|ss) integral
         IF (se_int_control%shortrange .OR. se_int_control%pc_coulomb_int) THEN
            CALL ssss_nucint_num(sepi, sepj, rij, ssss=ssss_sr, itype=itype, se_taper=se_taper, &
                                 se_int_control=se_int_control)
         ELSE
            ssss_sr = ssss
         END IF
         zz = sepi%zeff*sepj%zeff
         ! Nuclear-Nuclear electrostatic contribution
         enuc = zz*ssss_sr
      END IF
   END SUBROUTINE corecore_el_num

! **************************************************************************************************
!> \brief Calculates the SSSS integrals (main driver)
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij interatomic distance
!> \param ssss derivative of (ssss) integral
!>                          derivatives are intended w.r.t. rij
!> \param itype type of semi_empirical model
!>                          extension to the original routine to compute qm/mm integrals
!> \param se_taper ...
!> \param se_int_control input parameters that control the calculation of SE
!>                          integrals (shortrange, R3 residual, screening type)
!> \par History
!>      03.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! **************************************************************************************************
   SUBROUTINE ssss_nucint_num(sepi, sepj, rij, ssss, itype, se_taper, se_int_control)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), INTENT(OUT)                              :: ssss
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'ssss_nucint_num', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: ft
      TYPE(se_int_screen_type)                           :: se_int_screen

! Computing Tapering function

      ft = 1.0_dp
      IF (itype /= do_method_pchg) THEN
         ft = taper_eval(se_taper%taper, rij)
      END IF

      ! In case of dumped integrals compute an additional taper term
      IF (se_int_control%integral_screening == do_se_IS_kdso_d) THEN
         se_int_screen%ft = 1.0_dp
         IF (itype /= do_method_pchg) THEN
            se_int_screen%ft = taper_eval(se_taper%taper_add, rij)
         END IF
      END IF

      ! Contribution from the sp shells
      CALL nucint_sp_num(sepi, sepj, rij, ssss=ssss, itype=itype, &
                         se_int_control=se_int_control, se_int_screen=se_int_screen)

      ! Tapering the integrals
      ssss = ft*ssss

   END SUBROUTINE ssss_nucint_num

! **************************************************************************************************
!> \brief Calculates the nuclear attraction integrals CORE (main driver)
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij interatomic distance
!> \param core derivative of 4 X 2 array of electron-core attraction integrals
!>                          derivatives are intended w.r.t. rij
!> \param itype type of semi_empirical model
!>                          extension to the original routine to compute qm/mm integrals
!> \param se_taper ...
!> \param se_int_control input parameters that control the calculation of SE
!>                          integrals (shortrange, R3 residual, screening type)
!> \par History
!>      03.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! **************************************************************************************************
   SUBROUTINE core_nucint_num(sepi, sepj, rij, core, itype, se_taper, se_int_control)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), DIMENSION(10, 2), INTENT(OUT)            :: core
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'core_nucint_num', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: ft
      TYPE(se_int_screen_type)                           :: se_int_screen

! Computing the Tapering function

      ft = 1.0_dp
      IF (itype /= do_method_pchg) THEN
         ft = taper_eval(se_taper%taper, rij)
      END IF

      ! In case of dumped integrals compute an additional taper term
      IF (se_int_control%integral_screening == do_se_IS_kdso_d) THEN
         se_int_screen%ft = 1.0_dp
         IF (itype /= do_method_pchg) THEN
            se_int_screen%ft = taper_eval(se_taper%taper_add, rij)
         END IF
      END IF

      ! Contribution from the sp shells
      CALL nucint_sp_num(sepi, sepj, rij, core=core, itype=itype, &
                         se_int_control=se_int_control, se_int_screen=se_int_screen)

      IF (sepi%dorb .OR. sepj%dorb) THEN
         ! Compute the contribution from d shells
         CALL nucint_d_num(sepi, sepj, rij, core, itype, se_int_control, &
                           se_int_screen)
      END IF

      ! Tapering the integrals
      DO i = 1, sepi%core_size
         core(i, 1) = ft*core(i, 1)
      END DO
      DO i = 1, sepj%core_size
         core(i, 2) = ft*core(i, 2)
      END DO

   END SUBROUTINE core_nucint_num

! **************************************************************************************************
!> \brief ...
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param ssss ...
!> \param core ...
!> \param itype ...
!> \param se_int_control ...
!> \param se_int_screen ...
!> \par History
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : new driver
!>                    for computing integrals
!>      Teodoro Laino (04.2008) [tlaino] - Totally rewritten: nothing to do with
!>                    the old version
! **************************************************************************************************

   SUBROUTINE nucint_sp_num(sepi, sepj, rij, ssss, core, itype, se_int_control, &
                            se_int_screen)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), INTENT(INOUT), OPTIONAL                  :: ssss
      REAL(dp), DIMENSION(10, 2), INTENT(INOUT), &
         OPTIONAL                                        :: core
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen

      CHARACTER(len=*), PARAMETER :: routineN = 'nucint_sp_num', routineP = moduleN//':'//routineN

      INTEGER                                            :: ij, kl
      LOGICAL                                            :: l_core, l_ssss, si, sj

      l_core = PRESENT(core)
      l_ssss = PRESENT(ssss)
      IF (.NOT. (l_core .OR. l_ssss)) RETURN
      si = (sepi%natorb > 1)
      sj = (sepj%natorb > 1)

      ij = indexa(1, 1)
      IF (l_ssss) THEN
         ! Store the value for <S  S  | S  S  > (Used for computing the core-core interactions)
         ssss = ijkl_sp(sepi, sepj, ij, ij, 0, 0, 0, 0, -1, rij, se_int_control, se_int_screen, itype)
      END IF

      IF (l_core) THEN
         !     <S  S  | S  S  >
         kl = indexa(1, 1)
         core(1, 1) = ijkl_sp(sepi, sepj, kl, ij, 0, 0, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
         IF (si) THEN
            !  <S  P  | S  S  >
            kl = indexa(2, 1)
            core(2, 1) = ijkl_sp(sepi, sepj, kl, ij, 0, 1, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <P  P  | S  S  >
            kl = indexa(2, 2)
            core(3, 1) = ijkl_sp(sepi, sepj, kl, ij, 1, 1, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <P+ P+ | S  S  >
            kl = indexa(3, 3)
            core(4, 1) = ijkl_sp(sepi, sepj, kl, ij, 1, 1, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
         END IF

         !     <S  S  | S  S  >
         kl = indexa(1, 1)
         core(1, 2) = ijkl_sp(sepi, sepj, ij, kl, 0, 0, 0, 0, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
         IF (sj) THEN
            !  <S  S  | S  P  >
            kl = indexa(2, 1)
            core(2, 2) = ijkl_sp(sepi, sepj, ij, kl, 0, 0, 0, 1, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S  S  | P  P  >
            kl = indexa(2, 2)
            core(3, 2) = ijkl_sp(sepi, sepj, ij, kl, 0, 0, 1, 1, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S  S  | P+ P+ >
            kl = indexa(3, 3)
            core(4, 2) = ijkl_sp(sepi, sepj, ij, kl, 0, 0, 1, 1, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
         END IF
      END IF

   END SUBROUTINE nucint_sp_num

! **************************************************************************************************
!> \brief Calculates the nuclear attraction integrals involving d orbitals
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij interatomic distance
!> \param core 4 X 2 array of electron-core attraction integrals
!>
!>         The storage of the nuclear attraction integrals  core(kl/ij) iS
!>         (SS/)=1,   (SP/)=2,   (PP/)=3,   (P+P+/)=4,   (SD/)=5,
!>         (DP/)=6,   (DD/)=7,   (D+P+)=8,  (D+D+/)=9,   (D#D#)=10
!>
!>         where ij=1 if the orbitals centred on atom i,  =2 if on atom j.
!> \param itype type of semi_empirical model
!>                         extension to the original routine to compute qm/mm integrals
!> \param se_int_control input parameters that control the calculation of SE
!>                         integrals (shortrange, R3 residual, screening type)
!> \param se_int_screen contains information for computing the screened
!>                         integrals KDSO-D
!> \author
!>      Teodoro Laino (03.2008) [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE nucint_d_num(sepi, sepj, rij, core, itype, se_int_control, &
                           se_int_screen)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), DIMENSION(10, 2), INTENT(INOUT)          :: core
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen

      CHARACTER(len=*), PARAMETER :: routineN = 'nucint_d_num', routineP = moduleN//':'//routineN

      INTEGER                                            :: ij, kl

! Check if d-orbitals are present

      IF (sepi%dorb .OR. sepj%dorb) THEN
         ij = indexa(1, 1)
         IF (sepj%dorb) THEN
            !  <S S | D S>
            kl = indexa(5, 1)
            core(5, 2) = ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 0, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S S | D P >
            kl = indexa(5, 2)
            core(6, 2) = ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 1, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S S | D D >
            kl = indexa(5, 5)
            core(7, 2) = ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 2, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S S | D+P+>
            kl = indexa(6, 3)
            core(8, 2) = ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 1, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S S | D+D+>
            kl = indexa(6, 6)
            core(9, 2) = ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 2, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S S | D#D#>
            kl = indexa(8, 8)
            core(10, 2) = ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 2, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
         END IF
         IF (sepi%dorb) THEN
            !  <D S | S S>
            kl = indexa(5, 1)
            core(5, 1) = ijkl_d(sepi, sepj, kl, ij, 2, 0, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <D P | S S >
            kl = indexa(5, 2)
            core(6, 1) = ijkl_d(sepi, sepj, kl, ij, 2, 1, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <D D | S S >
            kl = indexa(5, 5)
            core(7, 1) = ijkl_d(sepi, sepj, kl, ij, 2, 2, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <D+P+| S S >
            kl = indexa(6, 3)
            core(8, 1) = ijkl_d(sepi, sepj, kl, ij, 2, 1, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <D+D+| S S >
            kl = indexa(6, 6)
            core(9, 1) = ijkl_d(sepi, sepj, kl, ij, 2, 2, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <D#D#| S S >
            kl = indexa(8, 8)
            core(10, 1) = ijkl_d(sepi, sepj, kl, ij, 2, 2, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
         END IF

      END IF
   END SUBROUTINE nucint_d_num

! **************************************************************************************************
!> \brief Numerical Derivatives for rotint
!> \param sepi ...
!> \param sepj ...
!> \param r ...
!> \param dw ...
!> \param delta ...
!> \param se_int_control ...
!> \param se_taper ...
! **************************************************************************************************
   SUBROUTINE drotint_num(sepi, sepj, r, dw, delta, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: r
      REAL(dp), DIMENSION(3, 2025), INTENT(OUT)          :: dw
      REAL(dp), INTENT(IN)                               :: delta
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'drotint_num', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, nsize
      REAL(dp)                                           :: od
      REAL(dp), DIMENSION(2025)                          :: wm, wp
      REAL(dp), DIMENSION(3)                             :: rr

      od = 0.5_dp/delta
      nsize = sepi%atm_int_size*sepj%atm_int_size
      DO i = 1, 3
         rr = r
         rr(i) = rr(i)+delta
         CALL rotint_num(sepi, sepj, rr, wp, se_int_control, se_taper=se_taper)
         rr(i) = rr(i)-2._dp*delta
         CALL rotint_num(sepi, sepj, rr, wm, se_int_control, se_taper=se_taper)
         DO j = 1, nsize
            dw(i, j) = od*(wp(j)-wm(j))
         END DO
      END DO

   END SUBROUTINE drotint_num

! **************************************************************************************************
!> \brief Numerical Derivatives for rotnuc
!> \param sepi ...
!> \param sepj ...
!> \param r ...
!> \param de1b ...
!> \param de2a ...
!> \param itype ...
!> \param delta ...
!> \param se_int_control ...
!> \param se_taper ...
! **************************************************************************************************
   SUBROUTINE drotnuc_num(sepi, sepj, r, de1b, de2a, itype, delta, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: r
      REAL(dp), DIMENSION(3, 45), INTENT(OUT), OPTIONAL  :: de1b, de2a
      INTEGER, INTENT(IN)                                :: itype
      REAL(dp), INTENT(IN)                               :: delta
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'drotnuc_num', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j
      LOGICAL                                            :: l_de1b, l_de2a
      REAL(dp)                                           :: od
      REAL(dp), DIMENSION(3)                             :: rr
      REAL(dp), DIMENSION(45)                            :: e1m, e1p, e2m, e2p

      l_de1b = PRESENT(de1b)
      l_de2a = PRESENT(de2a)
      IF (.NOT. (l_de1b .OR. l_de2a)) RETURN
      od = 0.5_dp/delta
      DO i = 1, 3
         rr = r
         rr(i) = rr(i)+delta
         CALL rotnuc_num(sepi, sepj, rr, e1p, e2p, itype, se_int_control, se_taper=se_taper)
         rr(i) = rr(i)-2._dp*delta
         CALL rotnuc_num(sepi, sepj, rr, e1m, e2m, itype, se_int_control, se_taper=se_taper)
         IF (l_de1b) THEN
            DO j = 1, sepi%atm_int_size
               de1b(i, j) = od*(e1p(j)-e1m(j))
            END DO
         END IF
         IF (l_de2a) THEN
            DO j = 1, sepj%atm_int_size
               de2a(i, j) = od*(e2p(j)-e2m(j))
            END DO
         END IF
      END DO
   END SUBROUTINE drotnuc_num

! **************************************************************************************************
!> \brief Numerical Derivatives for corecore
!> \param sepi ...
!> \param sepj ...
!> \param r ...
!> \param denuc ...
!> \param itype ...
!> \param delta ...
!> \param se_int_control ...
!> \param se_taper ...
! **************************************************************************************************
   SUBROUTINE dcorecore_num(sepi, sepj, r, denuc, itype, delta, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: r
      REAL(dp), DIMENSION(3), INTENT(OUT)                :: denuc
      INTEGER, INTENT(IN)                                :: itype
      REAL(dp), INTENT(IN)                               :: delta
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'dcorecore_num', routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(dp)                                           :: enucm, enucp, od
      REAL(dp), DIMENSION(3)                             :: rr

      od = 0.5_dp/delta
      DO i = 1, 3
         rr = r
         rr(i) = rr(i)+delta
         CALL corecore_num(sepi, sepj, rr, enucp, itype, se_int_control, se_taper=se_taper)
         rr(i) = rr(i)-2._dp*delta
         CALL corecore_num(sepi, sepj, rr, enucm, itype, se_int_control, se_taper=se_taper)
         denuc(i) = od*(enucp-enucm)
      END DO
   END SUBROUTINE dcorecore_num

! **************************************************************************************************
!> \brief Numerical Derivatives for corecore
!> \param sepi ...
!> \param sepj ...
!> \param r ...
!> \param denuc ...
!> \param itype ...
!> \param delta ...
!> \param se_int_control ...
!> \param se_taper ...
! **************************************************************************************************
   SUBROUTINE dcorecore_el_num(sepi, sepj, r, denuc, itype, delta, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: r
      REAL(dp), DIMENSION(3), INTENT(OUT)                :: denuc
      INTEGER, INTENT(IN)                                :: itype
      REAL(dp), INTENT(IN)                               :: delta
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'dcorecore_el_num', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(dp)                                           :: enucm, enucp, od
      REAL(dp), DIMENSION(3)                             :: rr

      od = 0.5_dp/delta
      DO i = 1, 3
         rr = r
         rr(i) = rr(i)+delta
         CALL corecore_el_num(sepi, sepj, rr, enucp, itype, se_int_control, se_taper=se_taper)
         rr(i) = rr(i)-2._dp*delta
         CALL corecore_el_num(sepi, sepj, rr, enucm, itype, se_int_control, se_taper=se_taper)
         denuc(i) = od*(enucp-enucm)
      END DO
   END SUBROUTINE dcorecore_el_num

END MODULE semi_empirical_int_num
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Analytical derivatives of Integrals for semi-empirical methods
!> \author Teodoro Laino - Zurich University 04.2007 [tlaino]
!> \par History
!>      23.11.2007 jhu   short range version of integrals
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
!>      Teodoro Laino (05.2008) [tlaino] - University of Zurich : analytical
!>                 derivatives for d-orbitals
! **************************************************************************************************
MODULE semi_empirical_int_ana

   USE input_constants,                 ONLY: do_method_am1,&
                                              do_method_pchg,&
                                              do_method_pdg,&
                                              do_method_pm3,&
                                              do_method_pm6,&
                                              do_method_undef,&
                                              do_se_IS_kdso_d
   USE kinds,                           ONLY: dp
   USE multipole_types,                 ONLY: do_multipole_none
   USE physcon,                         ONLY: angstrom,&
                                              evolt
   USE semi_empirical_int_arrays,       ONLY: &
        fac_x_to_z, ijkl_ind, ijkl_sym, inddd, inddp, indexa, indexb, indpp, int2c_type, l_index, &
        map_x_to_z, rij_threshold
   USE semi_empirical_int_num,          ONLY: nucint_d_num,&
                                              nucint_sp_num,&
                                              terep_d_num,&
                                              terep_sp_num
   USE semi_empirical_int_utils,        ONLY: d_ijkl_d,&
                                              d_ijkl_sp,&
                                              rot_2el_2c_first,&
                                              rotmat,&
                                              store_2el_2c_diag
   USE semi_empirical_types,            ONLY: rotmat_create,&
                                              rotmat_release,&
                                              rotmat_type,&
                                              se_int_control_type,&
                                              se_int_screen_type,&
                                              se_taper_type,&
                                              semi_empirical_type,&
                                              setup_se_int_control_type
   USE taper_types,                     ONLY: dtaper_eval,&
                                              taper_eval
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
!-----------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations         !
!   Copyright (C) 2000 - 2016  CP2K developers group                          !
!-----------------------------------------------------------------------------!

! *****************************************************************************
!> \brief Debug the derivatives of the the rotational matrices
!>
!> \author Teodoro Laino [tlaino] - University of Zurich
!> \date 04.2008 [tlaino]
! *****************************************************************************
INTERFACE 
   SUBROUTINE check_rotmat_der( sepi, sepj, rjiv, ij_matrix, do_invert)
     USE kinds,                           ONLY: dp
     USE semi_empirical_types,            ONLY: rotmat_type,&
                                                semi_empirical_type
     IMPLICIT NONE
     TYPE(semi_empirical_type), POINTER       :: sepi, sepj
     REAL(KIND=dp), DIMENSION(3), INTENT(IN)  :: rjiv
     TYPE(rotmat_type), POINTER               :: ij_matrix
     LOGICAL, INTENT(IN)                      :: do_invert
   END SUBROUTINE check_rotmat_der
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical NUCINT ssss
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE
  SUBROUTINE check_dssss_nucint_ana (sepi,sepj,r,dssss,itype,se_int_control,&
       se_taper)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), INTENT(IN)                     :: r
    REAL(dp), INTENT(IN)                     :: dssss
    INTEGER, INTENT(IN)                      :: itype
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
  END SUBROUTINE check_dssss_nucint_ana
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical NUCINT core
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
  SUBROUTINE check_dcore_nucint_ana (sepi,sepj,r,dcore,itype,se_int_control,&
       se_taper)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), INTENT(IN)                     :: r
    REAL(dp), DIMENSION(10, 2), INTENT(IN)   :: dcore
    INTEGER, INTENT(IN)                      :: itype
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
  END SUBROUTINE check_dcore_nucint_ana
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical ROTNUC
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
   SUBROUTINE check_drotnuc_ana(sepi, sepj, rijv, itype, se_int_control, se_taper,&
        e1b, e2a, de1b, de2a)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), DIMENSION(3), INTENT(IN)       :: rijv
    INTEGER, INTENT(IN)                      :: itype
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
    REAL(dp), DIMENSION(45), INTENT(IN), &
      OPTIONAL                               :: e1b, e2a
    REAL(dp), DIMENSION(45, 3), &
      INTENT(IN), OPTIONAL                   :: de1b, de2a
   END SUBROUTINE check_drotnuc_ana
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical CORECORE
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
  SUBROUTINE check_dcorecore_ana(sepi, sepj, rijv, itype,se_int_control,&
       se_taper, enuc, denuc)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), DIMENSION(3), INTENT(IN)       :: rijv
    INTEGER, INTENT(IN)                      :: itype
    REAL(dp), INTENT(IN), OPTIONAL           :: enuc
    REAL(dp), DIMENSION(3), INTENT(IN), &
         OPTIONAL                            :: denuc
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
  END SUBROUTINE check_dcorecore_ana

END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical rot_2el_2c_first
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
  SUBROUTINE rot_2el_2c_first_debug(sepi, sepj, rijv, se_int_control, se_taper,&
       invert, ii, kk, v_d)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(KIND=dp), DIMENSION(3), INTENT(IN)  :: rijv
    LOGICAL, INTENT(IN)                      :: invert
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
    INTEGER, INTENT(IN)                      :: ii, kk
    REAL(KIND=dp), DIMENSION(45, 45, 3), &
      INTENT(IN)                             :: v_d
  END SUBROUTINE rot_2el_2c_first_debug
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical check_dterep_ana
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
  SUBROUTINE check_dterep_ana (sepi,sepj,r,ri,dri,se_int_control,se_taper,lgrad)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), INTENT(IN)                     :: r
    REAL(dp), DIMENSION(491), INTENT(IN)     :: ri, dri
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    LOGICAL, INTENT(IN)                      :: lgrad
    TYPE(se_taper_type), POINTER             :: se_taper
  END SUBROUTINE check_dterep_ana
END INTERFACE 

! *****************************************************************************
!> \brief Check Numerical Vs Analytical check_rotint_ana
!> \note
!>      Debug routine
!> \par History
!>      04.2008 created [tlaino]
!> \author Teodoro Laino - Zurich University
! *****************************************************************************
INTERFACE 
  SUBROUTINE check_rotint_ana(sepi,sepj,rijv,w,dw,se_int_control,se_taper)
    USE kinds,                           ONLY: dp
    USE semi_empirical_types,            ONLY: semi_empirical_type,&
                                               se_int_control_type,&
                                               se_taper_type
    IMPLICIT NONE
    TYPE(semi_empirical_type), POINTER       :: sepi, sepj
    REAL(dp), DIMENSION(3), INTENT(IN)       :: rijv
    REAL(dp), DIMENSION(2025), INTENT(IN), &
      OPTIONAL                               :: w
    REAL(dp), DIMENSION(2025, 3), &
      INTENT(IN), OPTIONAL                   :: dw
    TYPE(se_int_control_type), INTENT(IN)    :: se_int_control
    TYPE(se_taper_type), POINTER             :: se_taper
  END SUBROUTINE check_rotint_ana
END INTERFACE 

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_int_ana'
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.
   PUBLIC :: rotnuc_ana, rotint_ana, corecore_ana, corecore_el_ana

CONTAINS

! **************************************************************************************************
!> \brief Computes analytical gradients for semiempirical integrals
!> \param sepi Atomic parameters of first atom
!> \param sepj Atomic parameters of second atom
!> \param rijv Coordinate vector i -> j
!> \param itype ...
!> \param e1b Array of electron-nuclear attraction integrals, Electron on atom ni attracting nucleus of nj.
!> \param e2a Array of electron-nuclear attraction integrals, Electron on atom nj attracting nucleus of ni.
!> \param de1b derivative of e1b term
!> \param de2a derivative of e2a term
!> \param se_int_control input parameters that control the calculation of SE
!>                           integrals (shortrange, R3 residual, screening type)
!> \param se_taper ...
!> \par History
!>      04.2007 created [tlaino]
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
!>      Teodoro Laino [tlaino] - University of Zurich 04.2008 : removed the core-core part
!> \author Teodoro Laino [tlaino] - Zurich University
!> \note
!>      Analytical version of the MOPAC rotnuc routine
! **************************************************************************************************
   RECURSIVE SUBROUTINE rotnuc_ana(sepi, sepj, rijv, itype, e1b, e2a, de1b, de2a, &
                                   se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rijv
      INTEGER, INTENT(IN)                                :: itype
      REAL(dp), DIMENSION(45), INTENT(OUT), OPTIONAL     :: e1b, e2a
      REAL(dp), DIMENSION(3, 45), INTENT(OUT), OPTIONAL  :: de1b, de2a
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'rotnuc_ana', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, idd, idp, ind1, ind2, ipp, j, &
                                                            last_orbital(2), m, n
      LOGICAL                                            :: invert, l_de1b, l_de2a, l_e1b, l_e2a, &
                                                            lgrad, task(2)
      REAL(KIND=dp)                                      :: rij, xtmp
      REAL(KIND=dp), DIMENSION(10, 2)                    :: core, dcore
      REAL(KIND=dp), DIMENSION(3)                        :: drij
      REAL(KIND=dp), DIMENSION(3, 45)                    :: tmp_d
      REAL(KIND=dp), DIMENSION(45)                       :: tmp
      TYPE(rotmat_type), POINTER                         :: ij_matrix

      NULLIFY (ij_matrix)
      rij = DOT_PRODUCT(rijv, rijv)
      ! Initialization
      l_e1b = PRESENT(e1b)
      l_e2a = PRESENT(e2a)
      l_de1b = PRESENT(de1b)
      l_de2a = PRESENT(de2a)
      lgrad = l_de1b .OR. l_de2a

      IF (rij > rij_threshold) THEN
         ! Compute Integrals in diatomic frame opportunely inverted
         rij = SQRT(rij)
         ! Create the rotation matrix
         CALL rotmat_create(ij_matrix)
         CALL rotmat(sepi, sepj, rijv, rij, ij_matrix, do_derivatives=lgrad, do_invert=invert)

         IF (lgrad) THEN
            drij(1) = rijv(1)/rij
            drij(2) = rijv(2)/rij
            drij(3) = rijv(3)/rij
            ! Possibly Invert Frame
            IF (invert) THEN
               xtmp = drij(3)
               drij(3) = drij(1)
               drij(1) = xtmp
            END IF
         END IF

         CALL dcore_nucint_ana(sepi, sepj, rij, core=core, dcore=dcore, itype=itype, se_taper=se_taper, &
                               se_int_control=se_int_control, lgrad=lgrad)

         ! Copy parameters over to arrays for do loop.
         last_orbital(1) = sepi%natorb
         last_orbital(2) = sepj%natorb
         task(1) = l_e1b
         task(2) = l_e2a
         DO n = 1, 2
            IF (.NOT. task(n)) CYCLE
            DO i = 1, last_orbital(n)
               ind1 = i-1
               DO j = 1, i
                  ind2 = j-1
                  m = (i*(i-1))/2+j
                  ! Perform Rotations ...
                  IF (ind2 == 0) THEN
                     IF (ind1 == 0) THEN
                        ! Type of Integral (SS/)
                        tmp(m) = core(1, n)
                     ELSE IF (ind1 < 4) THEN
                        ! Type of Integral (SP/)
                        tmp(m) = ij_matrix%sp(1, ind1)*core(2, n)
                     ELSE
                        ! Type of Integral (SD/)
                        tmp(m) = ij_matrix%sd(1, ind1-3)*core(5, n)
                     END IF
                  ELSE IF (ind2 < 4) THEN
                     IF (ind1 < 4) THEN
                        ! Type of Integral (PP/)
                        ipp = indpp(ind1, ind2)
                        tmp(m) = core(3, n)*ij_matrix%pp(ipp, 1, 1)+ &
                                 core(4, n)*(ij_matrix%pp(ipp, 2, 2)+ij_matrix%pp(ipp, 3, 3))
                     ELSE
                        ! Type of Integral (PD/)
                        idp = inddp(ind1-3, ind2)
                        tmp(m) = core(6, n)*ij_matrix%pd(idp, 1, 1)+ &
                                 core(8, n)*(ij_matrix%pd(idp, 2, 2)+ij_matrix%pd(idp, 3, 3))
                     END IF
                  ELSE
                     ! Type of Integral (DD/)
                     idd = inddd(ind1-3, ind2-3)
                     tmp(m) = core(7, n)*ij_matrix%dd(idd, 1, 1)+ &
                              core(9, n)*(ij_matrix%dd(idd, 2, 2)+ij_matrix%dd(idd, 3, 3))+ &
                              core(10, n)*(ij_matrix%dd(idd, 4, 4)+ij_matrix%dd(idd, 5, 5))
                  END IF
               END DO
            END DO
            IF (n == 1) THEN
               DO i = 1, sepi%atm_int_size
                  e1b(i) = -tmp(i)
               END DO
            END IF
            IF (n == 2) THEN
               DO i = 1, sepj%atm_int_size
                  e2a(i) = -tmp(i)
               END DO
            END IF
         END DO
         IF (invert .AND. l_e1b) CALL invert_integral(sepi, sepi, int1el=e1b)
         IF (invert .AND. l_e2a) CALL invert_integral(sepj, sepj, int1el=e2a)

         ! Possibly compute derivatives
         task(1) = l_de1b
         task(2) = l_de2a
         DO n = 1, 2
            IF (.NOT. task(n)) CYCLE
            DO i = 1, last_orbital(n)
               ind1 = i-1
               DO j = 1, i
                  ind2 = j-1
                  m = (i*(i-1))/2+j
                  ! Perform Rotations ...
                  IF (ind2 == 0) THEN
                     IF (ind1 == 0) THEN
                        ! Type of Integral (SS/)
                        tmp_d(1, m) = dcore(1, n)*drij(1)
                        tmp_d(2, m) = dcore(1, n)*drij(2)
                        tmp_d(3, m) = dcore(1, n)*drij(3)
                     ELSE IF (ind1 < 4) THEN
                        ! Type of Integral (SP/)
                        tmp_d(1, m) = ij_matrix%sp_d(1, 1, ind1)*core(2, n)+ &
                                      ij_matrix%sp(1, ind1)*dcore(2, n)*drij(1)

                        tmp_d(2, m) = ij_matrix%sp_d(2, 1, ind1)*core(2, n)+ &
                                      ij_matrix%sp(1, ind1)*dcore(2, n)*drij(2)

                        tmp_d(3, m) = ij_matrix%sp_d(3, 1, ind1)*core(2, n)+ &
                                      ij_matrix%sp(1, ind1)*dcore(2, n)*drij(3)
                     ELSE
                        ! Type of Integral (SD/)
                        tmp_d(1, m) = ij_matrix%sd_d(1, 1, ind1-3)*core(5, n)+ &
                                      ij_matrix%sd(1, ind1-3)*dcore(5, n)*drij(1)

                        tmp_d(2, m) = ij_matrix%sd_d(2, 1, ind1-3)*core(5, n)+ &
                                      ij_matrix%sd(1, ind1-3)*dcore(5, n)*drij(2)

                        tmp_d(3, m) = ij_matrix%sd_d(3, 1, ind1-3)*core(5, n)+ &
                                      ij_matrix%sd(1, ind1-3)*dcore(5, n)*drij(3)
                     END IF
                  ELSE IF (ind2 < 4) THEN
                     IF (ind1 < 4) THEN
                        ! Type of Integral (PP/)
                        ipp = indpp(ind1, ind2)
                        tmp_d(1, m) = dcore(3, n)*drij(1)*ij_matrix%pp(ipp, 1, 1)+ &
                                      core(3, n)*ij_matrix%pp_d(1, ipp, 1, 1)+ &
                                      dcore(4, n)*drij(1)*(ij_matrix%pp(ipp, 2, 2)+ij_matrix%pp(ipp, 3, 3))+ &
                                      core(4, n)*(ij_matrix%pp_d(1, ipp, 2, 2)+ij_matrix%pp_d(1, ipp, 3, 3))

                        tmp_d(2, m) = dcore(3, n)*drij(2)*ij_matrix%pp(ipp, 1, 1)+ &
                                      core(3, n)*ij_matrix%pp_d(2, ipp, 1, 1)+ &
                                      dcore(4, n)*drij(2)*(ij_matrix%pp(ipp, 2, 2)+ij_matrix%pp(ipp, 3, 3))+ &
                                      core(4, n)*(ij_matrix%pp_d(2, ipp, 2, 2)+ij_matrix%pp_d(2, ipp, 3, 3))

                        tmp_d(3, m) = dcore(3, n)*drij(3)*ij_matrix%pp(ipp, 1, 1)+ &
                                      core(3, n)*ij_matrix%pp_d(3, ipp, 1, 1)+ &
                                      dcore(4, n)*drij(3)*(ij_matrix%pp(ipp, 2, 2)+ij_matrix%pp(ipp, 3, 3))+ &
                                      core(4, n)*(ij_matrix%pp_d(3, ipp, 2, 2)+ij_matrix%pp_d(3, ipp, 3, 3))
                     ELSE
                        ! Type of Integral (PD/)
                        idp = inddp(ind1-3, ind2)
                        tmp_d(1, m) = dcore(6, n)*drij(1)*ij_matrix%pd(idp, 1, 1)+ &
                                      core(6, n)*ij_matrix%pd_d(1, idp, 1, 1)+ &
                                      dcore(8, n)*drij(1)*(ij_matrix%pd(idp, 2, 2)+ij_matrix%pd(idp, 3, 3))+ &
                                      core(8, n)*(ij_matrix%pd_d(1, idp, 2, 2)+ij_matrix%pd_d(1, idp, 3, 3))

                        tmp_d(2, m) = dcore(6, n)*drij(2)*ij_matrix%pd(idp, 1, 1)+ &
                                      core(6, n)*ij_matrix%pd_d(2, idp, 1, 1)+ &
                                      dcore(8, n)*drij(2)*(ij_matrix%pd(idp, 2, 2)+ij_matrix%pd(idp, 3, 3))+ &
                                      core(8, n)*(ij_matrix%pd_d(2, idp, 2, 2)+ij_matrix%pd_d(2, idp, 3, 3))

                        tmp_d(3, m) = dcore(6, n)*drij(3)*ij_matrix%pd(idp, 1, 1)+ &
                                      core(6, n)*ij_matrix%pd_d(3, idp, 1, 1)+ &
                                      dcore(8, n)*drij(3)*(ij_matrix%pd(idp, 2, 2)+ij_matrix%pd(idp, 3, 3))+ &
                                      core(8, n)*(ij_matrix%pd_d(3, idp, 2, 2)+ij_matrix%pd_d(3, idp, 3, 3))
                     END IF
                  ELSE
                     ! Type of Integral (DD/)
                     idd = inddd(ind1-3, ind2-3)
                     tmp_d(1, m) = dcore(7, n)*drij(1)*ij_matrix%dd(idd, 1, 1)+ &
                                   core(7, n)*ij_matrix%dd_d(1, idd, 1, 1)+ &
                                   dcore(9, n)*drij(1)*(ij_matrix%dd(idd, 2, 2)+ij_matrix%dd(idd, 3, 3))+ &
                                   core(9, n)*(ij_matrix%dd_d(1, idd, 2, 2)+ij_matrix%dd_d(1, idd, 3, 3))+ &
                                   dcore(10, n)*drij(1)*(ij_matrix%dd(idd, 4, 4)+ij_matrix%dd(idd, 5, 5))+ &
                                   core(10, n)*(ij_matrix%dd_d(1, idd, 4, 4)+ij_matrix%dd_d(1, idd, 5, 5))

                     tmp_d(2, m) = dcore(7, n)*drij(2)*ij_matrix%dd(idd, 1, 1)+ &
                                   core(7, n)*ij_matrix%dd_d(2, idd, 1, 1)+ &
                                   dcore(9, n)*drij(2)*(ij_matrix%dd(idd, 2, 2)+ij_matrix%dd(idd, 3, 3))+ &
                                   core(9, n)*(ij_matrix%dd_d(2, idd, 2, 2)+ij_matrix%dd_d(2, idd, 3, 3))+ &
                                   dcore(10, n)*drij(2)*(ij_matrix%dd(idd, 4, 4)+ij_matrix%dd(idd, 5, 5))+ &
                                   core(10, n)*(ij_matrix%dd_d(2, idd, 4, 4)+ij_matrix%dd_d(2, idd, 5, 5))

                     tmp_d(3, m) = dcore(7, n)*drij(3)*ij_matrix%dd(idd, 1, 1)+ &
                                   core(7, n)*ij_matrix%dd_d(3, idd, 1, 1)+ &
                                   dcore(9, n)*drij(3)*(ij_matrix%dd(idd, 2, 2)+ij_matrix%dd(idd, 3, 3))+ &
                                   core(9, n)*(ij_matrix%dd_d(3, idd, 2, 2)+ij_matrix%dd_d(3, idd, 3, 3))+ &
                                   dcore(10, n)*drij(3)*(ij_matrix%dd(idd, 4, 4)+ij_matrix%dd(idd, 5, 5))+ &
                                   core(10, n)*(ij_matrix%dd_d(3, idd, 4, 4)+ij_matrix%dd_d(3, idd, 5, 5))
                  END IF
               END DO
            END DO
            IF (n == 1) THEN
               DO i = 1, sepi%atm_int_size
                  de1b(1, i) = -tmp_d(1, i)
                  de1b(2, i) = -tmp_d(2, i)
                  de1b(3, i) = -tmp_d(3, i)
               END DO
            END IF
            IF (n == 2) THEN
               DO i = 1, sepj%atm_int_size
                  de2a(1, i) = -tmp_d(1, i)
                  de2a(2, i) = -tmp_d(2, i)
                  de2a(3, i) = -tmp_d(3, i)
               END DO
            END IF
         END DO
         IF (invert .AND. l_de1b) CALL invert_derivative(sepi, sepi, dint1el=de1b)
         IF (invert .AND. l_de2a) CALL invert_derivative(sepj, sepj, dint1el=de2a)
         CALL rotmat_release(ij_matrix)

         ! Possibly debug the analytical values versus the numerical ones
         IF (debug_this_module) THEN
            CALL check_drotnuc_ana(sepi, sepj, rijv, itype, se_int_control, se_taper, e1b, e2a, de1b, de2a)
         END IF
      END IF
   END SUBROUTINE rotnuc_ana

! **************************************************************************************************
!> \brief Computes analytical gradients for semiempirical core-core interaction.
!> \param sepi Atomic parameters of first atom
!> \param sepj Atomic parameters of second atom
!> \param rijv Coordinate vector i -> j
!> \param itype ...
!> \param enuc nuclear-nuclear repulsion term.
!> \param denuc derivative of nuclear-nuclear repulsion term.
!> \param se_int_control input parameters that control the calculation of SE
!>                           integrals (shortrange, R3 residual, screening type)
!> \param se_taper ...
!> \par History
!>      04.2007 created [tlaino]
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
!>       Teodoro Laino [tlaino] - University of Zurich 04.2008 : removed the
!>                 core-core part
!> \author Teodoro Laino [tlaino] - Zurich University
!> \note
!>      Analytical version of the MOPAC rotnuc routine
! **************************************************************************************************
   RECURSIVE SUBROUTINE corecore_ana(sepi, sepj, rijv, itype, enuc, denuc, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rijv
      INTEGER, INTENT(IN)                                :: itype
      REAL(dp), INTENT(OUT), OPTIONAL                    :: enuc
      REAL(dp), DIMENSION(3), INTENT(OUT), OPTIONAL      :: denuc
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'corecore_ana', routineP = moduleN//':'//routineN

      INTEGER                                            :: ig, nt
      LOGICAL                                            :: l_denuc, l_enuc
      REAL(dp) :: aab, alpi, alpj, apdg, ax, dai, daj, dax, dbi, dbj, denuc_loc, dqcorr, drija, &
         dscale, dssss, dssss_sr, dtmp, dzz, enuc_loc, pai, paj, pbi, pbj, qcorr, rij, rija, &
         scale, ssss, ssss_sr, tmp, xab, xtmp, zaf, zbf, zz
      REAL(dp), DIMENSION(3)                             :: drij
      REAL(dp), DIMENSION(4)                             :: fni1, fni2, fni3, fnj1, fnj2, fnj3
      TYPE(se_int_control_type)                          :: se_int_control_off

      rij = DOT_PRODUCT(rijv, rijv)
      ! Initialization
      l_enuc = PRESENT(enuc)
      l_denuc = PRESENT(denuc)
      IF ((rij > rij_threshold) .AND. (l_enuc .OR. l_denuc)) THEN
         ! Compute Integrals in diatomic frame
         rij = SQRT(rij)
         CALL setup_se_int_control_type(se_int_control_off, shortrange=.FALSE., do_ewald_r3=.FALSE., &
                                        do_ewald_gks=.FALSE., integral_screening=se_int_control%integral_screening, &
                                        max_multipole=do_multipole_none, pc_coulomb_int=.FALSE.)
         CALL dssss_nucint_ana(sepi, sepj, rij, ssss=ssss, dssss=dssss, itype=itype, se_taper=se_taper, &
                               se_int_control=se_int_control_off, lgrad=l_denuc)
         ! In case let's compute the short-range part of the (ss|ss) integral
         IF (se_int_control%shortrange) THEN
            CALL dssss_nucint_ana(sepi, sepj, rij, ssss=ssss_sr, dssss=dssss_sr, itype=itype, &
                                  se_taper=se_taper, se_int_control=se_int_control, lgrad=l_denuc)
         ELSE
            ssss_sr = ssss
            dssss_sr = dssss
         END IF
         ! Zeroing local method dependent core-core corrections
         enuc_loc = 0.0_dp
         denuc_loc = 0.0_dp
         qcorr = 0.0_dp
         scale = 0.0_dp
         dscale = 0.0_dp
         dqcorr = 0.0_dp
         zz = sepi%zeff*sepj%zeff
         ! Core Core electrostatic contribution
         IF (l_enuc) enuc_loc = zz*ssss_sr
         IF (l_denuc) denuc_loc = zz*dssss_sr
         ! Method dependent code
         tmp = zz*ssss
         IF (l_denuc) dtmp = zz*dssss
         IF (itype /= do_method_pm6) THEN
            alpi = sepi%alp
            alpj = sepj%alp
            scale = EXP(-alpi*rij)+EXP(-alpj*rij)
            IF (l_denuc) THEN
               dscale = -alpi*EXP(-alpi*rij)-alpj*EXP(-alpj*rij)
            END IF
            nt = sepi%z+sepj%z
            IF (nt == 8 .OR. nt == 9) THEN
               IF (sepi%z == 7 .OR. sepi%z == 8) THEN
                  scale = scale+(angstrom*rij-1._dp)*EXP(-alpi*rij)
                  IF (l_denuc) THEN
                     dscale = dscale+angstrom*EXP(-alpi*rij)-(angstrom*rij-1._dp)*alpi*EXP(-alpi*rij)
                  END IF
               END IF
               IF (sepj%z == 7 .OR. sepj%z == 8) THEN
                  scale = scale+(angstrom*rij-1._dp)*EXP(-alpj*rij)
                  IF (l_denuc) THEN
                     dscale = dscale+angstrom*EXP(-alpj*rij)-(angstrom*rij-1._dp)*alpj*EXP(-alpj*rij)
                  END IF
               END IF
            ENDIF
            IF (l_denuc) THEN
               dscale = SIGN(1.0_dp, scale*tmp)*(dscale*tmp+scale*dtmp)
               dzz = -zz/rij**2
            END IF
            scale = ABS(scale*tmp)
            zz = zz/rij
            IF (itype == do_method_am1 .OR. itype == do_method_pm3 .OR. itype == do_method_pdg) THEN
               IF (itype == do_method_am1 .AND. sepi%z == 5) THEN
                  !special case AM1 Boron
                  SELECT CASE (sepj%z)
                  CASE DEFAULT
                     nt = 1
                  CASE (1)
                     nt = 2
                  CASE (6)
                     nt = 3
                  CASE (9, 17, 35, 53)
                     nt = 4
                  END SELECT
                  fni1(1) = sepi%bfn1(1, nt)
                  fni1(2) = sepi%bfn1(2, nt)
                  fni1(3) = sepi%bfn1(3, nt)
                  fni1(4) = sepi%bfn1(4, nt)
                  fni2(1) = sepi%bfn2(1, nt)
                  fni2(2) = sepi%bfn2(2, nt)
                  fni2(3) = sepi%bfn2(3, nt)
                  fni2(4) = sepi%bfn2(4, nt)
                  fni3(1) = sepi%bfn3(1, nt)
                  fni3(2) = sepi%bfn3(2, nt)
                  fni3(3) = sepi%bfn3(3, nt)
                  fni3(4) = sepi%bfn3(4, nt)
               ELSE
                  fni1(1) = sepi%fn1(1)
                  fni1(2) = sepi%fn1(2)
                  fni1(3) = sepi%fn1(3)
                  fni1(4) = sepi%fn1(4)
                  fni2(1) = sepi%fn2(1)
                  fni2(2) = sepi%fn2(2)
                  fni2(3) = sepi%fn2(3)
                  fni2(4) = sepi%fn2(4)
                  fni3(1) = sepi%fn3(1)
                  fni3(2) = sepi%fn3(2)
                  fni3(3) = sepi%fn3(3)
                  fni3(4) = sepi%fn3(4)
               END IF
               IF (itype == do_method_am1 .AND. sepj%z == 5) THEN
                  !special case AM1 Boron
                  SELECT CASE (sepi%z)
                  CASE DEFAULT
                     nt = 1
                  CASE (1)
                     nt = 2
                  CASE (6)
                     nt = 3
                  CASE (9, 17, 35, 53)
                     nt = 4
                  END SELECT
                  fnj1(1) = sepj%bfn1(1, nt)
                  fnj1(2) = sepj%bfn1(2, nt)
                  fnj1(3) = sepj%bfn1(3, nt)
                  fnj1(4) = sepj%bfn1(4, nt)
                  fnj2(1) = sepj%bfn2(1, nt)
                  fnj2(2) = sepj%bfn2(2, nt)
                  fnj2(3) = sepj%bfn2(3, nt)
                  fnj2(4) = sepj%bfn2(4, nt)
                  fnj3(1) = sepj%bfn3(1, nt)
                  fnj3(2) = sepj%bfn3(2, nt)
                  fnj3(3) = sepj%bfn3(3, nt)
                  fnj3(4) = sepj%bfn3(4, nt)
               ELSE
                  fnj1(1) = sepj%fn1(1)
                  fnj1(2) = sepj%fn1(2)
                  fnj1(3) = sepj%fn1(3)
                  fnj1(4) = sepj%fn1(4)
                  fnj2(1) = sepj%fn2(1)
                  fnj2(2) = sepj%fn2(2)
                  fnj2(3) = sepj%fn2(3)
                  fnj2(4) = sepj%fn2(4)
                  fnj3(1) = sepj%fn3(1)
                  fnj3(2) = sepj%fn3(2)
                  fnj3(3) = sepj%fn3(3)
                  fnj3(4) = sepj%fn3(4)
               END IF
               ! AM1/PM3/PDG correction to nuclear repulsion
               DO ig = 1, SIZE(fni1)
                  IF (ABS(fni1(ig)) > 0._dp) THEN
                     ax = fni2(ig)*(rij-fni3(ig))**2
                     IF (ax <= 25._dp) THEN
                        scale = scale+zz*fni1(ig)*EXP(-ax)
                        IF (l_denuc) THEN
                           dax = fni2(ig)*2.0_dp*(rij-fni3(ig))
                           dscale = dscale+dzz*fni1(ig)*EXP(-ax)-dax*zz*fni1(ig)*EXP(-ax)
                        END IF
                     ENDIF
                  ENDIF
                  IF (ABS(fnj1(ig)) > 0._dp) THEN
                     ax = fnj2(ig)*(rij-fnj3(ig))**2
                     IF (ax <= 25._dp) THEN
                        scale = scale+zz*fnj1(ig)*EXP(-ax)
                        IF (l_denuc) THEN
                           dax = fnj2(ig)*2.0_dp*(rij-fnj3(ig))
                           dscale = dscale+dzz*fnj1(ig)*EXP(-ax)-dax*zz*fnj1(ig)*EXP(-ax)
                        END IF
                     ENDIF
                  ENDIF
               END DO
            ENDIF
            IF (itype == do_method_pdg) THEN
               ! PDDG function
               zaf = sepi%zeff/nt
               zbf = sepj%zeff/nt
               pai = sepi%pre(1)
               pbi = sepi%pre(2)
               paj = sepj%pre(1)
               pbj = sepj%pre(2)
               dai = sepi%d(1)
               dbi = sepi%d(2)
               daj = sepj%d(1)
               dbj = sepj%d(2)
               apdg = 10._dp*angstrom**2
               qcorr = (zaf*pai+zbf*paj)*EXP(-apdg*(rij-dai-daj)**2)+ &
                       (zaf*pai+zbf*pbj)*EXP(-apdg*(rij-dai-dbj)**2)+ &
                       (zaf*pbi+zbf*paj)*EXP(-apdg*(rij-dbi-daj)**2)+ &
                       (zaf*pbi+zbf*pbj)*EXP(-apdg*(rij-dbi-dbj)**2)
               IF (l_denuc) THEN
                  dqcorr = (zaf*pai+zbf*paj)*EXP(-apdg*(rij-dai-daj)**2)*(-2.0_dp*apdg*(rij-dai-daj))+ &
                           (zaf*pai+zbf*pbj)*EXP(-apdg*(rij-dai-dbj)**2)*(-2.0_dp*apdg*(rij-dai-dbj))+ &
                           (zaf*pbi+zbf*paj)*EXP(-apdg*(rij-dbi-daj)**2)*(-2.0_dp*apdg*(rij-dbi-daj))+ &
                           (zaf*pbi+zbf*pbj)*EXP(-apdg*(rij-dbi-dbj)**2)*(-2.0_dp*apdg*(rij-dbi-dbj))
               END IF
            ELSEIF (itype == do_method_pchg) THEN
               qcorr = 0.0_dp
               scale = 0.0_dp
               dscale = 0.0_dp
               dqcorr = 0.0_dp
            ELSE
               qcorr = 0.0_dp
               dqcorr = 0.0_dp
            END IF
         ELSE
            ! PM6 core-core terms
            scale = tmp
            IF (l_denuc) dscale = dtmp
            drija = angstrom
            rija = rij*drija
            xab = sepi%xab(sepj%z)
            aab = sepi%aab(sepj%z)
            IF ((sepi%z == 1 .AND. (sepj%z == 6 .OR. sepj%z == 7 .OR. sepj%z == 8)) .OR. &
                (sepj%z == 1 .AND. (sepi%z == 6 .OR. sepi%z == 7 .OR. sepi%z == 8))) THEN
               ! Special Case O-H or N-H or C-H
               IF (l_denuc) dscale = dscale*(2._dp*xab*EXP(-aab*rija*rija))- &
                                     scale*2._dp*xab*EXP(-aab*rija*rija)*(2.0_dp*aab*rija)*drija
               IF (l_enuc) scale = scale*(2._dp*xab*EXP(-aab*rija*rija))
            ELSEIF (sepi%z == 6 .AND. sepj%z == 6) THEN
               ! Special Case C-C
               IF (l_denuc) dscale = &
                  dscale*(2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6))+9.28_dp*EXP(-5.98_dp*rija)) &
                  -scale*2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6))*aab*(1.0_dp+6.0_dp*0.0003_dp*rija**5)*drija &
                  -scale*9.28_dp*EXP(-5.98_dp*rija)*5.98_dp*drija
               IF (l_enuc) scale = scale*(2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6))+9.28_dp*EXP(-5.98_dp*rija))
            ELSEIF ((sepi%z == 8 .AND. sepj%z == 14) .OR. &
                    (sepj%z == 8 .AND. sepi%z == 14)) THEN
               ! Special Case Si-O
               IF (l_denuc) dscale = &
                  dscale*(2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6))-0.0007_dp*EXP(-(rija-2.9_dp)**2)) &
                  -scale*2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6))*aab*(1.0_dp+6.0_dp*0.0003_dp*rija**5)*drija+ &
                  scale*0.0007_dp*EXP(-(rija-2.9_dp)**2)*(2.0_dp*(rija-2.9_dp)*drija)
               IF (l_enuc) scale = scale*(2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6))-0.0007_dp*EXP(-(rija-2.9_dp)**2))
            ELSE
               ! General Case
               ! Factor of 2 found by experiment
               IF (l_denuc) dscale = dscale*(2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6))) &
                                     -scale*2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6))*aab*(1.0_dp+6.0_dp*0.0003_dp*rija**5)*drija
               IF (l_enuc) scale = scale*(2._dp*xab*EXP(-aab*(rija+0.0003_dp*rija**6)))
            END IF
            ! General correction term a*exp(-b*(rij-c)^2)
            xtmp = 1.e-8_dp/evolt*((REAL(sepi%z, dp)**(1._dp/3._dp)+REAL(sepj%z, dp)**(1._dp/3._dp))/rija)**12
            IF (l_enuc) THEN
               qcorr = (sepi%a*EXP(-sepi%b*(rij-sepi%c)**2))*zz/rij+ &
                       (sepj%a*EXP(-sepj%b*(rij-sepj%c)**2))*zz/rij+ &
                       ! Hard core repulsion
                       xtmp
            END IF
            IF (l_denuc) THEN
               dqcorr = (sepi%a*EXP(-sepi%b*(rij-sepi%c)**2)*(-2.0_dp*sepi%b*(rij-sepi%c)))*zz/rij- &
                        (sepi%a*EXP(-sepi%b*(rij-sepi%c)**2))*zz/rij**2+ &
                        (sepj%a*EXP(-sepj%b*(rij-sepj%c)**2)*(-2.0_dp*sepj%b*(rij-sepj%c)))*zz/rij- &
                        (sepj%a*EXP(-sepj%b*(rij-sepj%c)**2))*zz/rij**2+ &
                        ! Hard core repulsion
                        (-12.0_dp*xtmp/rija*drija)
            END IF
         ENDIF

         ! Only at the very end let's sum-up the several contributions energy/derivatives
         ! This assignment should be method indipendent
         IF (l_enuc) THEN
            enuc = enuc_loc+scale+qcorr
         END IF
         IF (l_denuc) THEN
            drij(1) = rijv(1)/rij
            drij(2) = rijv(2)/rij
            drij(3) = rijv(3)/rij
            denuc = (denuc_loc+dscale+dqcorr)*drij
         END IF
         ! Debug statement
         IF (debug_this_module) THEN
            CALL check_dcorecore_ana(sepi, sepj, rijv, itype, se_int_control, se_taper, enuc, denuc)
         END IF
      ENDIF
   END SUBROUTINE corecore_ana

! **************************************************************************************************
!> \brief Computes analytical gradients for semiempirical core-core electrostatic
!>        interaction only.
!> \param sepi Atomic parameters of first atom
!> \param sepj Atomic parameters of second atom
!> \param rijv Coordinate vector i -> j
!> \param itype ...
!> \param enuc nuclear-nuclear electrostatic repulsion term.
!> \param denuc derivative of nuclear-nuclear electrostatic
!>                             repulsion term.
!> \param se_int_control input parameters that control the calculation of SE
!>                           integrals (shortrange, R3 residual, screening type)
!> \param se_taper ...
!> \par History
!>      04.2007 created [tlaino]
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
!>       Teodoro Laino [tlaino] - University of Zurich 04.2008 : removed the
!>                 core-core part
!> \author Teodoro Laino [tlaino] - Zurich University
!> \note
!>      Analytical version of the MOPAC rotnuc routine
! **************************************************************************************************
   RECURSIVE SUBROUTINE corecore_el_ana(sepi, sepj, rijv, itype, enuc, denuc, &
                                        se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rijv
      INTEGER, INTENT(IN)                                :: itype
      REAL(dp), INTENT(OUT), OPTIONAL                    :: enuc
      REAL(dp), DIMENSION(3), INTENT(OUT), OPTIONAL      :: denuc
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'corecore_el_ana', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: l_denuc, l_enuc
      REAL(dp)                                           :: drij(3), dssss, dssss_sr, rij, ssss, &
                                                            ssss_sr, tmp, zz
      TYPE(se_int_control_type)                          :: se_int_control_off

      rij = DOT_PRODUCT(rijv, rijv)
      ! Initialization
      l_enuc = PRESENT(enuc)
      l_denuc = PRESENT(denuc)
      IF ((rij > rij_threshold) .AND. (l_enuc .OR. l_denuc)) THEN
         ! Compute Integrals in diatomic frame
         rij = SQRT(rij)
         CALL setup_se_int_control_type(se_int_control_off, shortrange=.FALSE., do_ewald_r3=.FALSE., &
                                        do_ewald_gks=.FALSE., integral_screening=se_int_control%integral_screening, &
                                        max_multipole=do_multipole_none, pc_coulomb_int=.FALSE.)
         CALL dssss_nucint_ana(sepi, sepj, rij, ssss=ssss, dssss=dssss, itype=itype, se_taper=se_taper, &
                               se_int_control=se_int_control_off, lgrad=l_denuc)
         ! In case let's compute the short-range part of the (ss|ss) integral
         IF (se_int_control%shortrange .OR. se_int_control%pc_coulomb_int) THEN
            CALL dssss_nucint_ana(sepi, sepj, rij, ssss=ssss_sr, dssss=dssss_sr, itype=itype, &
                                  se_taper=se_taper, se_int_control=se_int_control, lgrad=l_denuc)
         ELSE
            ssss_sr = ssss
            dssss_sr = dssss
         END IF
         zz = sepi%zeff*sepj%zeff
         ! Core Core electrostatic contribution
         IF (l_enuc) enuc = zz*ssss_sr
         IF (l_denuc) THEN
            drij(1) = rijv(1)/rij
            drij(2) = rijv(2)/rij
            drij(3) = rijv(3)/rij
            tmp = zz*dssss_sr
            denuc = tmp*drij
         END IF
      END IF
   END SUBROUTINE corecore_el_ana

! **************************************************************************************************
!> \brief Exploits inversion symmetry to avoid divergence
!> \param sepi ...
!> \param sepj ...
!> \param int1el ...
!> \param int2el ...
!> \par History
!>      04.2007 created [tlaino]
!>      05.2008 New driver for integral invertion (supports d-orbitals)
!> \author Teodoro Laino - Zurich University
! **************************************************************************************************
   SUBROUTINE invert_integral(sepi, sepj, int1el, int2el)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(:), INTENT(INOUT), OPTIONAL    :: int1el, int2el

      CHARACTER(len=*), PARAMETER :: routineN = 'invert_integral', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: fdim, gind, gknd, i, imap, ind, j, jmap, &
                                                            jnd, k, kmap, knd, l, lmap, lnd, ndim, &
                                                            sdim, tdim, tind
      REAL(KIND=dp)                                      :: ifac, jfac, kfac, lfac
      REAL(KIND=dp), DIMENSION(2025)                     :: tmp2el
      REAL(KIND=dp), DIMENSION(45)                       :: tmp1el

! One-electron integral

      IF (PRESENT(int1el)) THEN
         fdim = sepi%atm_int_size
         ndim = 0
         DO i = 1, fdim
            tmp1el(i) = 0.0_dp
         END DO
         DO i = 1, sepi%natorb
         DO j = 1, i
            ndim = ndim+1

            ! Get the integral in the original frame (along z)
            DO ind = 1, 2
               imap = map_x_to_z(ind, i)
               IF (imap == 0) CYCLE
               ifac = fac_x_to_z(ind, i)
               DO jnd = 1, 2
                  jmap = map_x_to_z(jnd, j)
                  IF (jmap == 0) CYCLE
                  jfac = fac_x_to_z(jnd, j)
                  gind = indexb(imap, jmap)

                  tmp1el(ndim) = tmp1el(ndim)+ifac*jfac*int1el(gind)
               END DO
            END DO
         END DO
         END DO
         DO i = 1, fdim
            int1el(i) = tmp1el(i)
         END DO
      END IF

      ! Two electron integrals
      IF (PRESENT(int2el)) THEN
         sdim = sepi%atm_int_size
         tdim = sepj%atm_int_size
         fdim = sdim*tdim
         ndim = 0
         DO i = 1, fdim
            tmp2el(i) = 0.0_dp
         END DO
         DO i = 1, sepi%natorb
         DO j = 1, i
            DO k = 1, sepj%natorb
            DO l = 1, k
               ndim = ndim+1

               ! Get the integral in the original frame (along z)
               DO ind = 1, 2
                  imap = map_x_to_z(ind, i)
                  IF (imap == 0) CYCLE
                  ifac = fac_x_to_z(ind, i)
                  DO jnd = 1, 2
                     jmap = map_x_to_z(jnd, j)
                     IF (jmap == 0) CYCLE
                     jfac = fac_x_to_z(jnd, j)
                     gind = indexb(imap, jmap)

                     ! Get the integral in the original frame (along z)
                     DO knd = 1, 2
                        kmap = map_x_to_z(knd, k)
                        IF (kmap == 0) CYCLE
                        kfac = fac_x_to_z(knd, k)
                        DO lnd = 1, 2
                           lmap = map_x_to_z(lnd, l)
                           IF (lmap == 0) CYCLE
                           lfac = fac_x_to_z(lnd, l)
                           gknd = indexb(kmap, lmap)

                           tind = (gind-1)*tdim+gknd
                           tmp2el(ndim) = tmp2el(ndim)+ifac*jfac*lfac*kfac*int2el(tind)
                        END DO
                     END DO

                  END DO
               END DO

            END DO
            END DO
         END DO
         END DO
         DO i = 1, fdim
            int2el(i) = tmp2el(i)
         END DO
      END IF
   END SUBROUTINE invert_integral

! **************************************************************************************************
!> \brief Exploits inversion symmetry to avoid divergence
!> \param sepi ...
!> \param sepj ...
!> \param dint1el ...
!> \param dint2el ...
!> \par History
!>      04.2007 created [tlaino]
!> \author Teodoro Laino - Zurich University
! **************************************************************************************************
   SUBROUTINE invert_derivative(sepi, sepj, dint1el, dint2el)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT), &
         OPTIONAL                                        :: dint1el, dint2el

      CHARACTER(len=*), PARAMETER :: routineN = 'invert_derivative', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, m
      REAL(KIND=dp)                                      :: tmp

! Integral part

      DO i = 1, 3
         IF (PRESENT(dint1el)) THEN
            CALL invert_integral(sepi, sepj, int1el=dint1el(i, :))
         END IF
         IF (PRESENT(dint2el)) THEN
            CALL invert_integral(sepi, sepj, int2el=dint2el(i, :))
         END IF
      END DO

      ! Derivatives part
      IF (PRESENT(dint1el)) THEN
         DO m = 1, SIZE(dint1el, 2)
            tmp = dint1el(3, m)
            dint1el(3, m) = dint1el(1, m)
            dint1el(1, m) = tmp
         END DO
      END IF
      IF (PRESENT(dint2el)) THEN
         DO m = 1, SIZE(dint2el, 2)
            tmp = dint2el(3, m)
            dint2el(3, m) = dint2el(1, m)
            dint2el(1, m) = tmp
         END DO
      END IF
   END SUBROUTINE invert_derivative

! **************************************************************************************************
!> \brief Calculates the ssss integral and analytical derivatives (main driver)
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij interatomic distance
!> \param ssss ...
!> \param dssss derivative of (ssss) integral
!>                          derivatives are intended w.r.t. rij
!> \param itype ...
!> \param se_taper ...
!> \param se_int_control input parameters that control the calculation of SE
!>                          integrals (shortrange, R3 residual, screening type)
!> \param lgrad ...
!> \par History
!>      03.2007 created [tlaino]
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
!> \author Teodoro Laino - Zurich University
!> \note
!>      Analytical version - Analytical evaluation of gradients
!>      Teodoro Laino - Zurich University 04.2007
!>
! **************************************************************************************************
   SUBROUTINE dssss_nucint_ana(sepi, sepj, rij, ssss, dssss, itype, se_taper, se_int_control, &
                               lgrad)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), INTENT(OUT)                              :: ssss, dssss
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      LOGICAL, INTENT(IN)                                :: lgrad

      CHARACTER(len=*), PARAMETER :: routineN = 'dssss_nucint_ana', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: dft, ft
      TYPE(se_int_screen_type)                           :: se_int_screen

! Compute the Tapering function

      ft = 1.0_dp
      dft = 0.0_dp
      IF (itype /= do_method_pchg) THEN
         ft = taper_eval(se_taper%taper, rij)
         dft = dtaper_eval(se_taper%taper, rij)
      END IF
      ! Evaluate additional taper function for dumped integrals
      IF (se_int_control%integral_screening == do_se_IS_kdso_d) THEN
         se_int_screen%ft = 1.0_dp
         se_int_screen%dft = 0.0_dp
         IF (itype /= do_method_pchg) THEN
            se_int_screen%ft = taper_eval(se_taper%taper_add, rij)
            se_int_screen%dft = dtaper_eval(se_taper%taper_add, rij)
         END IF
      END IF

      ! Value of the integrals for sp shell
      CALL nucint_sp_num(sepi, sepj, rij, ssss=ssss, itype=itype, se_int_control=se_int_control, &
                         se_int_screen=se_int_screen)

      IF (lgrad) THEN
         ! Integrals derivatives for sp shell
         CALL dnucint_sp_ana(sepi, sepj, rij, dssss=dssss, itype=itype, se_int_control=se_int_control, &
                             se_int_screen=se_int_screen)
      END IF

      ! Tapering the value of the integrals
      IF (lgrad) THEN
         dssss = ft*dssss+dft*ssss
      END IF
      ssss = ft*ssss

      ! Debug Procedure.. Check valifity of analytical gradients of nucint
      IF (debug_this_module .AND. lgrad) THEN
         CALL check_dssss_nucint_ana(sepi, sepj, rij, dssss, itype, se_int_control, se_taper=se_taper)
      END IF
   END SUBROUTINE dssss_nucint_ana

! **************************************************************************************************
!> \brief Calculates the nuclear attraction integrals and analytical integrals (main driver)
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij interatomic distance
!> \param core ...
!> \param dcore derivative of 4 X 2 array of electron-core attraction integrals
!>                          derivatives are intended w.r.t. rij
!>         The storage of the nuclear attraction integrals  core(kl/ij) iS
!>         (SS/)=1,   (SO/)=2,   (OO/)=3,   (PP/)=4
!>         where ij=1 if the orbitals centred on atom i,  =2 if on atom j.
!> \param itype type of semi_empirical model
!>                          extension to the original routine to compute qm/mm
!>                          integrals
!> \param se_taper ...
!> \param se_int_control input parameters that control the calculation of SE
!>                          integrals (shortrange, R3 residual, screening type)
!> \param lgrad ...
!> \par History
!>      03.2007 created [tlaino]
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
!> \author Teodoro Laino - Zurich University
!> \note
!>      Analytical version - Analytical evaluation of gradients
!>      Teodoro Laino - Zurich University 04.2007
!>
! **************************************************************************************************
   SUBROUTINE dcore_nucint_ana(sepi, sepj, rij, core, dcore, itype, se_taper, &
                               se_int_control, lgrad)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), DIMENSION(10, 2), INTENT(OUT)            :: core, dcore
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      LOGICAL, INTENT(IN)                                :: lgrad

      CHARACTER(len=*), PARAMETER :: routineN = 'dcore_nucint_ana', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(KIND=dp)                                      :: dft, ft
      TYPE(se_int_screen_type)                           :: se_int_screen

! Compute the Tapering function

      ft = 1.0_dp
      dft = 0.0_dp
      IF (itype /= do_method_pchg) THEN
         ft = taper_eval(se_taper%taper, rij)
         dft = dtaper_eval(se_taper%taper, rij)
      END IF
      ! Evaluate additional taper function for dumped integrals
      IF (se_int_control%integral_screening == do_se_IS_kdso_d) THEN
         se_int_screen%ft = 1.0_dp
         se_int_screen%dft = 0.0_dp
         IF (itype /= do_method_pchg) THEN
            se_int_screen%ft = taper_eval(se_taper%taper_add, rij)
            se_int_screen%dft = dtaper_eval(se_taper%taper_add, rij)
         END IF
      END IF

      ! Value of the integrals for sp shell
      CALL nucint_sp_num(sepi, sepj, rij, core=core, itype=itype, &
                         se_int_control=se_int_control, se_int_screen=se_int_screen)

      IF (sepi%dorb .OR. sepj%dorb) THEN
         ! Compute the contribution from d-orbitals
         CALL nucint_d_num(sepi, sepj, rij, core, itype, &
                           se_int_control=se_int_control, se_int_screen=se_int_screen)
      END IF

      IF (lgrad) THEN
         ! Integrals derivatives for sp shell
         CALL dnucint_sp_ana(sepi, sepj, rij, dcore=dcore, itype=itype, &
                             se_int_control=se_int_control, se_int_screen=se_int_screen)

         IF (sepi%dorb .OR. sepj%dorb) THEN
            ! Integral derivatives involving d-orbitals
            CALL dnucint_d_ana(sepi, sepj, rij, dcore=dcore, itype=itype, &
                               se_int_control=se_int_control, se_int_screen=se_int_screen)
         END IF
      END IF

      ! Tapering the value of the integrals
      IF (lgrad) THEN
         DO i = 1, sepi%core_size
            dcore(i, 1) = ft*dcore(i, 1)+dft*core(i, 1)
         END DO
         DO i = 1, sepj%core_size
            dcore(i, 2) = ft*dcore(i, 2)+dft*core(i, 2)
         END DO
      END IF
      DO i = 1, sepi%core_size
         core(i, 1) = ft*core(i, 1)
      END DO
      DO i = 1, sepj%core_size
         core(i, 2) = ft*core(i, 2)
      END DO

      ! Debug Procedure.. Check valifity of analytical gradients of nucint
      IF (debug_this_module .AND. lgrad) THEN
         CALL check_dcore_nucint_ana(sepi, sepj, rij, dcore, itype, se_int_control, se_taper=se_taper)
      END IF
   END SUBROUTINE dcore_nucint_ana

! **************************************************************************************************
!> \brief Calculates the nuclear attraction integrals and derivatives for sp basis
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij interatomic distance
!> \param dssss derivative of (ssss) integral
!>                          derivatives are intended w.r.t. rij
!>         where ij=1 if the orbitals centred on atom i,  =2 if on atom j.
!> \param dcore derivative of 4 X 2 array of electron-core attraction integrals
!>         The storage of the nuclear attraction integrals  core(kl/ij) iS
!>         (SS/)=1,   (SP/)=2,   (PP/)=3,   (P+P+/)=4,   (SD/)=5,
!>         (DP/)=6,   (DD/)=7,   (D+P+)=8,  (D+D+/)=9,   (D#D#)=10
!> \param itype type of semi_empirical model
!>                          extension to the original routine to compute qm/mm
!>                          integrals
!> \param se_int_control input parameters that control the calculation of SE
!>                          integrals (shortrange, R3 residual, screening type)
!> \param se_int_screen ...
!> \par History
!>      04.2007 created [tlaino]
!>      Teodoro Laino (03.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
!>      05.2008 Teodoro Laino [tlaino] - University of Zurich: major rewriting
!> \author Teodoro Laino - Zurich University
!> \note
!>      Analytical version - Analytical evaluation of gradients
!>      Teodoro Laino - Zurich University 04.2007
!>      routine adapted from mopac7 (repp)
!>      vector version written by Ernest R. Davidson, Indiana University
! **************************************************************************************************
   SUBROUTINE dnucint_sp_ana(sepi, sepj, rij, dssss, dcore, itype, se_int_control, &
                             se_int_screen)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), INTENT(INOUT), OPTIONAL                  :: dssss
      REAL(dp), DIMENSION(10, 2), INTENT(INOUT), &
         OPTIONAL                                        :: dcore
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen

      CHARACTER(len=*), PARAMETER :: routineN = 'dnucint_sp_ana', routineP = moduleN//':'//routineN

      INTEGER                                            :: ij, kl
      LOGICAL                                            :: l_core, l_ssss, si, sj

      l_core = PRESENT(dcore)
      l_ssss = PRESENT(dssss)
      IF (.NOT. (l_core .OR. l_ssss)) RETURN

      si = (sepi%natorb > 1)
      sj = (sepj%natorb > 1)

      ij = indexa(1, 1)
      IF (l_ssss) THEN
         ! Store the value for the derivative of <S  S  | S  S  > (Used for computing the core-core interactions)
         dssss = d_ijkl_sp(sepi, sepj, ij, ij, 0, 0, 0, 0, -1, rij, se_int_control, se_int_screen, itype)
      END IF

      IF (l_core) THEN
         !     <S  S  | S  S  >
         kl = indexa(1, 1)
         dcore(1, 1) = d_ijkl_sp(sepi, sepj, kl, ij, 0, 0, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
         IF (si) THEN
            !  <S  P  | S  S  >
            kl = indexa(2, 1)
            dcore(2, 1) = d_ijkl_sp(sepi, sepj, kl, ij, 0, 1, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <P  P  | S  S  >
            kl = indexa(2, 2)
            dcore(3, 1) = d_ijkl_sp(sepi, sepj, kl, ij, 1, 1, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <P+ P+ | S  S  >
            kl = indexa(3, 3)
            dcore(4, 1) = d_ijkl_sp(sepi, sepj, kl, ij, 1, 1, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
         END IF

         !     <S  S  | S  S  >
         kl = indexa(1, 1)
         dcore(1, 2) = d_ijkl_sp(sepi, sepj, ij, kl, 0, 0, 0, 0, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
         IF (sj) THEN
            !  <S  S  | S  P  >
            kl = indexa(2, 1)
            dcore(2, 2) = d_ijkl_sp(sepi, sepj, ij, kl, 0, 0, 0, 1, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S  S  | P  P  >
            kl = indexa(2, 2)
            dcore(3, 2) = d_ijkl_sp(sepi, sepj, ij, kl, 0, 0, 1, 1, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S  S  | P+ P+ >
            kl = indexa(3, 3)
            dcore(4, 2) = d_ijkl_sp(sepi, sepj, ij, kl, 0, 0, 1, 1, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
         END IF
      END IF
   END SUBROUTINE dnucint_sp_ana

! **************************************************************************************************
!> \brief Calculates the analytical derivative of the nuclear attraction
!>        integrals involving d orbitals
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij interatomic distance
!> \param dcore 4 X 2 array of electron-core attraction integrals
!>         The storage of the nuclear attraction integrals  core(kl/ij) iS
!>         (SS/)=1,   (SP/)=2,   (PP/)=3,   (P+P+/)=4,   (SD/)=5,
!>         (DP/)=6,   (DD/)=7,   (D+P+)=8,  (D+D+/)=9,   (D#D#)=10
!>
!>         where ij=1 if the orbitals centred on atom i,  =2 if on atom j.
!> \param itype type of semi_empirical model
!>                         extension to the original routine to compute qm/mm
!>                         integrals
!> \param se_int_control input parameters that control the calculation of SE
!>                         integrals (shortrange, R3 residual, screening type)
!> \param se_int_screen ...
!> \author
!>      Teodoro Laino (05.2008) [tlaino] - University of Zurich: created
! **************************************************************************************************
   SUBROUTINE dnucint_d_ana(sepi, sepj, rij, dcore, itype, se_int_control, &
                            se_int_screen)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), DIMENSION(10, 2), INTENT(INOUT)          :: dcore
      INTEGER, INTENT(IN)                                :: itype
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen

      CHARACTER(len=*), PARAMETER :: routineN = 'dnucint_d_ana', routineP = moduleN//':'//routineN

      INTEGER                                            :: ij, kl

! Check if d-orbitals are present

      IF (sepi%dorb .OR. sepj%dorb) THEN
         ij = indexa(1, 1)
         IF (sepj%dorb) THEN
            !  <S S | D S>
            kl = indexa(5, 1)
            dcore(5, 2) = d_ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 0, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S S | D P >
            kl = indexa(5, 2)
            dcore(6, 2) = d_ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 1, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S S | D D >
            kl = indexa(5, 5)
            dcore(7, 2) = d_ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 2, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S S | D+P+>
            kl = indexa(6, 3)
            dcore(8, 2) = d_ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 1, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S S | D+D+>
            kl = indexa(6, 6)
            dcore(9, 2) = d_ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 2, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
            !  <S S | D#D#>
            kl = indexa(8, 8)
            dcore(10, 2) = d_ijkl_d(sepi, sepj, ij, kl, 0, 0, 2, 2, 1, rij, se_int_control, se_int_screen, itype)*sepi%zeff
         END IF
         IF (sepi%dorb) THEN
            !  <D S | S S>
            kl = indexa(5, 1)
            dcore(5, 1) = d_ijkl_d(sepi, sepj, kl, ij, 2, 0, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <D P | S S >
            kl = indexa(5, 2)
            dcore(6, 1) = d_ijkl_d(sepi, sepj, kl, ij, 2, 1, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <D D | S S >
            kl = indexa(5, 5)
            dcore(7, 1) = d_ijkl_d(sepi, sepj, kl, ij, 2, 2, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <D+P+| S S >
            kl = indexa(6, 3)
            dcore(8, 1) = d_ijkl_d(sepi, sepj, kl, ij, 2, 1, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <D+D+| S S >
            kl = indexa(6, 6)
            dcore(9, 1) = d_ijkl_d(sepi, sepj, kl, ij, 2, 2, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
            !  <D#D#| S S >
            kl = indexa(8, 8)
            dcore(10, 1) = d_ijkl_d(sepi, sepj, kl, ij, 2, 2, 0, 0, 2, rij, se_int_control, se_int_screen, itype)*sepj%zeff
         END IF
      END IF
   END SUBROUTINE dnucint_d_ana

! **************************************************************************************************
!> \brief calculates the derivative of the two-particle interactions
!> \param sepi Atomic parameters of first atom
!> \param sepj Atomic parameters of second atom
!> \param rijv Coordinate vector i -> j
!> \param w Array of two-electron repulsion integrals.
!> \param dw ...
!> \param se_int_control ...
!> \param se_taper ...
!> \par History
!>      04.2007 created [tlaino]
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
!> \author Teodoro Laino - Zurich University
!> \note
!>      Analytical version - Analytical evaluation of gradients
!>      Teodoro Laino - Zurich University 04.2007
!>      routine adapted from mopac7 (repp)
!>      vector version written by Ernest R. Davidson, Indiana University
! **************************************************************************************************
   RECURSIVE SUBROUTINE rotint_ana(sepi, sepj, rijv, w, dw, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rijv
      REAL(dp), DIMENSION(2025), INTENT(OUT), OPTIONAL   :: w
      REAL(dp), DIMENSION(3, 2025), INTENT(OUT), &
         OPTIONAL                                        :: dw
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'rotint_ana', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, i1, ii, ij, ij1, iminus, istep, &
                                                            iw_loc, j, j1, jj, k, kk, kl, l, &
                                                            limij, limkl, mm
      LOGICAL                                            :: invert, l_w, lgrad
      LOGICAL, DIMENSION(45, 45)                         :: logv, logv_d
      REAL(dp)                                           :: rij, xtmp
      REAL(dp), DIMENSION(3)                             :: drij
      REAL(KIND=dp)                                      :: cc, cc_d(3), wrepp, wrepp_d(3)
      REAL(KIND=dp), DIMENSION(2025)                     :: ww
      REAL(KIND=dp), DIMENSION(3, 2025)                  :: ww_d
      REAL(KIND=dp), DIMENSION(3, 45, 45)                :: v_d
      REAL(KIND=dp), DIMENSION(45, 45)                   :: v
      REAL(KIND=dp), DIMENSION(491)                      :: rep, rep_d
      TYPE(rotmat_type), POINTER                         :: ij_matrix

      NULLIFY (ij_matrix)
      l_w = PRESENT(w)
      lgrad = PRESENT(dw)
      IF (.NOT. (l_w .OR. lgrad)) RETURN

      rij = DOT_PRODUCT(rijv, rijv)
      IF (rij > rij_threshold) THEN
         ! The repulsion integrals over molecular frame (w) are stored in the
         ! order in which they will later be used.  ie.  (i,j/k,l) where
         ! j.le.i  and  l.le.k     and l varies most rapidly and i least
         ! rapidly.  (anti-normal computer storage)
         rij = SQRT(rij)

         ! Create the rotation matrix
         CALL rotmat_create(ij_matrix)
         CALL rotmat(sepi, sepj, rijv, rij, ij_matrix, do_derivatives=lgrad, do_invert=invert)

         ! Compute integrals in diatomic frame as well their derivatives (if requested)
         CALL dterep_ana(sepi, sepj, rij, rep, rep_d, se_taper, se_int_control, lgrad=lgrad)

         IF (lgrad) THEN
            drij(1) = rijv(1)/rij
            drij(2) = rijv(2)/rij
            drij(3) = rijv(3)/rij
            ! Possibly Invert Frame
            IF (invert) THEN
               xtmp = drij(3)
               drij(3) = drij(1)
               drij(1) = xtmp
            END IF
         END IF

         ii = sepi%natorb
         kk = sepj%natorb
         ! First step in rotation of integrals
         CALL rot_2el_2c_first(sepi, sepj, rijv, se_int_control, se_taper, invert, ii, kk, rep, logv, ij_matrix, &
                               v, lgrad, rep_d, v_d, logv_d, drij)

         ! Integrals if requested
         IF (l_w) THEN
            ! Rotate Integrals
            IF (ii*kk > 0) THEN
               limij = sepi%atm_int_size
               limkl = sepj%atm_int_size
               istep = limkl*limij
               DO i1 = 1, istep
                  ww(i1) = 0.0_dp
               END DO
               ! Second step in rotation of integrals
               DO i1 = 1, ii
                  DO j1 = 1, i1
                     ij = indexa(i1, j1)
                     jj = indexb(i1, j1)
                     mm = int2c_type(jj)
                     DO k = 1, kk
                        DO l = 1, k
                           kl = indexb(k, l)
                           IF (logv(ij, kl)) THEN
                              wrepp = v(ij, kl)
                              SELECT CASE (mm)
                              CASE (1)
                                 ! (SS/)
                                 i = 1
                                 j = 1
                                 iw_loc = (indexb(i, j)-1)*limkl+kl
                                 ww(iw_loc) = wrepp
                              CASE (2)
                                 ! (SP/)
                                 j = 1
                                 DO i = 1, 3
                                    iw_loc = (indexb(i+1, j)-1)*limkl+kl
                                    ww(iw_loc) = ww(iw_loc)+ij_matrix%sp(i1-1, i)*wrepp
                                 END DO
                              CASE (3)
                                 ! (PP/)
                                 DO i = 1, 3
                                    cc = ij_matrix%pp(i, i1-1, j1-1)
                                    iw_loc = (indexb(i+1, i+1)-1)*limkl+kl
                                    ww(iw_loc) = ww(iw_loc)+cc*wrepp
                                    iminus = i-1
                                    IF (iminus /= 0) THEN
                                       DO j = 1, iminus
                                          cc = ij_matrix%pp(1+i+j, i1-1, j1-1)
                                          iw_loc = (indexb(i+1, j+1)-1)*limkl+kl
                                          ww(iw_loc) = ww(iw_loc)+cc*wrepp
                                       END DO
                                    END IF
                                 END DO
                              CASE (4)
                                 ! (SD/)
                                 j = 1
                                 DO i = 1, 5
                                    iw_loc = (indexb(i+4, j)-1)*limkl+kl
                                    ww(iw_loc) = ww(iw_loc)+ij_matrix%sd(i1-4, i)*wrepp
                                 END DO
                              CASE (5)
                                 ! (DP/)
                                 DO i = 1, 5
                                    DO j = 1, 3
                                       iw_loc = (indexb(i+4, j+1)-1)*limkl+kl
                                       ij1 = 3*(i-1)+j
                                       ww(iw_loc) = ww(iw_loc)+ij_matrix%pd(ij1, i1-4, j1-1)*wrepp
                                    END DO
                                 END DO
                              CASE (6)
                                 ! (DD/)
                                 DO i = 1, 5
                                    cc = ij_matrix%dd(i, i1-4, j1-4)
                                    iw_loc = (indexb(i+4, i+4)-1)*limkl+kl
                                    ww(iw_loc) = ww(iw_loc)+cc*wrepp
                                    iminus = i-1
                                    IF (iminus /= 0) THEN
                                       DO j = 1, iminus
                                          ij1 = inddd(i, j)
                                          cc = ij_matrix%dd(ij1, i1-4, j1-4)
                                          iw_loc = (indexb(i+4, j+4)-1)*limkl+kl
                                          ww(iw_loc) = ww(iw_loc)+cc*wrepp
                                       END DO
                                    END IF
                                 END DO
                              END SELECT
                           END IF
                        END DO
                     END DO
                  END DO
               END DO
               ! Store two electron integrals in the triangular format
               CALL store_2el_2c_diag(limij, limkl, ww(1:istep), w)
               IF (invert) CALL invert_integral(sepi, sepj, int2el=w)
            END IF

            IF (debug_this_module) THEN
               ! Check value of integrals
               CALL check_rotint_ana(sepi, sepj, rijv, w, se_int_control=se_int_control, se_taper=se_taper)
            END IF
         END IF

         ! Gradients if requested
         IF (lgrad) THEN
            ! Rotate Integrals derivatives
            IF (ii*kk > 0) THEN
               limij = sepi%atm_int_size
               limkl = sepj%atm_int_size
               istep = limkl*limij
               DO i1 = 1, istep
                  ww_d(1, i1) = 0.0_dp
                  ww_d(2, i1) = 0.0_dp
                  ww_d(3, i1) = 0.0_dp
               END DO

               ! Second step in rotation of integrals
               DO i1 = 1, ii
                  DO j1 = 1, i1
                     ij = indexa(i1, j1)
                     jj = indexb(i1, j1)
                     mm = int2c_type(jj)
                     DO k = 1, kk
                        DO l = 1, k
                           kl = indexb(k, l)
                           IF (logv_d(ij, kl)) THEN
                              wrepp_d(1) = v_d(1, ij, kl)
                              wrepp_d(2) = v_d(2, ij, kl)
                              wrepp_d(3) = v_d(3, ij, kl)
                              wrepp = v(ij, kl)
                              SELECT CASE (mm)
                              CASE (1)
                                 ! (SS/)
                                 i = 1
                                 j = 1
                                 iw_loc = (indexb(i, j)-1)*limkl+kl
                                 ww_d(1, iw_loc) = wrepp_d(1)
                                 ww_d(2, iw_loc) = wrepp_d(2)
                                 ww_d(3, iw_loc) = wrepp_d(3)
                              CASE (2)
                                 ! (SP/)
                                 j = 1
                                 DO i = 1, 3
                                    iw_loc = (indexb(i+1, j)-1)*limkl+kl
                                    ww_d(1, iw_loc) = ww_d(1, iw_loc)+ij_matrix%sp_d(1, i1-1, i)*wrepp+ &
                                                      ij_matrix%sp(i1-1, i)*wrepp_d(1)

                                    ww_d(2, iw_loc) = ww_d(2, iw_loc)+ij_matrix%sp_d(2, i1-1, i)*wrepp+ &
                                                      ij_matrix%sp(i1-1, i)*wrepp_d(2)

                                    ww_d(3, iw_loc) = ww_d(3, iw_loc)+ij_matrix%sp_d(3, i1-1, i)*wrepp+ &
                                                      ij_matrix%sp(i1-1, i)*wrepp_d(3)
                                 END DO
                              CASE (3)
                                 ! (PP/)
                                 DO i = 1, 3
                                    cc = ij_matrix%pp(i, i1-1, j1-1)
                                    cc_d(1) = ij_matrix%pp_d(1, i, i1-1, j1-1)
                                    cc_d(2) = ij_matrix%pp_d(2, i, i1-1, j1-1)
                                    cc_d(3) = ij_matrix%pp_d(3, i, i1-1, j1-1)
                                    iw_loc = (indexb(i+1, i+1)-1)*limkl+kl
                                    ww_d(1, iw_loc) = ww_d(1, iw_loc)+cc_d(1)*wrepp+cc*wrepp_d(1)
                                    ww_d(2, iw_loc) = ww_d(2, iw_loc)+cc_d(2)*wrepp+cc*wrepp_d(2)
                                    ww_d(3, iw_loc) = ww_d(3, iw_loc)+cc_d(3)*wrepp+cc*wrepp_d(3)
                                    iminus = i-1
                                    IF (iminus /= 0) THEN
                                       DO j = 1, iminus
                                          cc = ij_matrix%pp(1+i+j, i1-1, j1-1)
                                          cc_d(1) = ij_matrix%pp_d(1, 1+i+j, i1-1, j1-1)
                                          cc_d(2) = ij_matrix%pp_d(2, 1+i+j, i1-1, j1-1)
                                          cc_d(3) = ij_matrix%pp_d(3, 1+i+j, i1-1, j1-1)
                                          iw_loc = (indexb(i+1, j+1)-1)*limkl+kl
                                          ww_d(1, iw_loc) = ww_d(1, iw_loc)+cc_d(1)*wrepp+cc*wrepp_d(1)
                                          ww_d(2, iw_loc) = ww_d(2, iw_loc)+cc_d(2)*wrepp+cc*wrepp_d(2)
                                          ww_d(3, iw_loc) = ww_d(3, iw_loc)+cc_d(3)*wrepp+cc*wrepp_d(3)
                                       END DO
                                    END IF
                                 END DO
                              CASE (4)
                                 ! (SD/)
                                 j = 1
                                 DO i = 1, 5
                                    iw_loc = (indexb(i+4, j)-1)*limkl+kl
                                    ww_d(1, iw_loc) = ww_d(1, iw_loc)+ij_matrix%sd_d(1, i1-4, i)*wrepp+ &
                                                      ij_matrix%sd(i1-4, i)*wrepp_d(1)

                                    ww_d(2, iw_loc) = ww_d(2, iw_loc)+ij_matrix%sd_d(2, i1-4, i)*wrepp+ &
                                                      ij_matrix%sd(i1-4, i)*wrepp_d(2)

                                    ww_d(3, iw_loc) = ww_d(3, iw_loc)+ij_matrix%sd_d(3, i1-4, i)*wrepp+ &
                                                      ij_matrix%sd(i1-4, i)*wrepp_d(3)
                                 END DO
                              CASE (5)
                                 ! (DP/)
                                 DO i = 1, 5
                                    DO j = 1, 3
                                       iw_loc = (indexb(i+4, j+1)-1)*limkl+kl
                                       ij1 = 3*(i-1)+j
                                       ww_d(1, iw_loc) = ww_d(1, iw_loc)+ij_matrix%pd_d(1, ij1, i1-4, j1-1)*wrepp+ &
                                                         ij_matrix%pd(ij1, i1-4, j1-1)*wrepp_d(1)

                                       ww_d(2, iw_loc) = ww_d(2, iw_loc)+ij_matrix%pd_d(2, ij1, i1-4, j1-1)*wrepp+ &
                                                         ij_matrix%pd(ij1, i1-4, j1-1)*wrepp_d(2)

                                       ww_d(3, iw_loc) = ww_d(3, iw_loc)+ij_matrix%pd_d(3, ij1, i1-4, j1-1)*wrepp+ &
                                                         ij_matrix%pd(ij1, i1-4, j1-1)*wrepp_d(3)
                                    END DO
                                 END DO
                              CASE (6)
                                 ! (DD/)
                                 DO i = 1, 5
                                    cc = ij_matrix%dd(i, i1-4, j1-4)
                                    cc_d = ij_matrix%dd_d(:, i, i1-4, j1-4)
                                    iw_loc = (indexb(i+4, i+4)-1)*limkl+kl
                                    ww_d(1, iw_loc) = ww_d(1, iw_loc)+cc_d(1)*wrepp+cc*wrepp_d(1)
                                    ww_d(2, iw_loc) = ww_d(2, iw_loc)+cc_d(2)*wrepp+cc*wrepp_d(2)
                                    ww_d(3, iw_loc) = ww_d(3, iw_loc)+cc_d(3)*wrepp+cc*wrepp_d(3)
                                    iminus = i-1
                                    IF (iminus /= 0) THEN
                                       DO j = 1, iminus
                                          ij1 = inddd(i, j)
                                          cc = ij_matrix%dd(ij1, i1-4, j1-4)
                                          cc_d(1) = ij_matrix%dd_d(1, ij1, i1-4, j1-4)
                                          cc_d(2) = ij_matrix%dd_d(2, ij1, i1-4, j1-4)
                                          cc_d(3) = ij_matrix%dd_d(3, ij1, i1-4, j1-4)
                                          iw_loc = (indexb(i+4, j+4)-1)*limkl+kl
                                          ww_d(1, iw_loc) = ww_d(1, iw_loc)+cc_d(1)*wrepp+cc*wrepp_d(1)
                                          ww_d(2, iw_loc) = ww_d(2, iw_loc)+cc_d(2)*wrepp+cc*wrepp_d(2)
                                          ww_d(3, iw_loc) = ww_d(3, iw_loc)+cc_d(3)*wrepp+cc*wrepp_d(3)
                                       END DO
                                    END IF
                                 END DO
                              END SELECT
                           END IF
                        END DO
                     END DO
                  END DO
               END DO
               ! Store two electron integrals in the triangular format
               CALL store_2el_2c_diag(limij, limkl, ww_dx=ww_d(1, 1:istep), ww_dy=ww_d(2, 1:istep), ww_dz=ww_d(3, 1:istep), &
                                      dw=dw)
               IF (invert) CALL invert_derivative(sepi, sepj, dint2el=dw)
            END IF

            IF (debug_this_module) THEN
               ! Check derivatives
               CALL check_rotint_ana(sepi, sepj, rijv, dw=dw, se_int_control=se_int_control, se_taper=se_taper)
            END IF
         END IF
         CALL rotmat_release(ij_matrix)
      ENDIF
   END SUBROUTINE rotint_ana

! **************************************************************************************************
!> \brief Calculates the derivative and the value of two-electron repulsion
!>      integrals and the nuclear attraction integrals w.r.t. |r|
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij interatomic distance
!> \param rep rray of two-electron repulsion integrals
!> \param rep_d array of two-electron repulsion integrals derivatives
!> \param se_taper ...
!> \param se_int_control input parameters that control the calculation of SE
!>                         integrals (shortrange, R3 residual, screening type)
!> \param lgrad ...
!> \par History
!>      03.2008 created [tlaino]
!> \author Teodoro Laino [tlaino] - Zurich University
! **************************************************************************************************
   RECURSIVE SUBROUTINE dterep_ana(sepi, sepj, rij, rep, rep_d, se_taper, &
                                   se_int_control, lgrad)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), INTENT(IN)                          :: rij
      REAL(KIND=dp), DIMENSION(491), INTENT(OUT)         :: rep, rep_d
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      LOGICAL, INTENT(IN)                                :: lgrad

      CHARACTER(len=*), PARAMETER :: routineN = 'dterep_ana', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ij, j, k, kl, l, lasti, lastj, li, &
                                                            lj, lk, ll, numb
      REAL(KIND=dp)                                      :: dft, ft, ft1
      TYPE(se_int_screen_type)                           :: se_int_screen

! Compute the tapering function and its derivatives

      ft = taper_eval(se_taper%taper, rij)
      dft = 0.0_dp
      ft1 = ft
      IF (lgrad) THEN
         ft1 = 1.0_dp
         dft = dtaper_eval(se_taper%taper, rij)
      END IF
      ! Evaluate additional taper function for dumped integrals
      IF (se_int_control%integral_screening == do_se_IS_kdso_d) THEN
         se_int_screen%ft = taper_eval(se_taper%taper_add, rij)
         IF (lgrad) &
            se_int_screen%dft = dtaper_eval(se_taper%taper_add, rij)
      END IF

      ! Integral Values for sp shells only
      CALL terep_sp_num(sepi, sepj, rij, rep, se_int_control=se_int_control, &
                        se_int_screen=se_int_screen, ft=ft1)

      IF (sepi%dorb .OR. sepj%dorb) THEN
         ! Compute the contribution from d-orbitals
         CALL terep_d_num(sepi, sepj, rij, rep, se_int_control=se_int_control, &
                          se_int_screen=se_int_screen, ft=ft1)
      END IF

      IF (lgrad) THEN
         ! Integral Derivatives
         CALL dterep_sp_ana(sepi, sepj, rij, rep_d, rep, se_int_control, &
                            se_int_screen, ft, dft)

         IF (sepi%dorb .OR. sepj%dorb) THEN
            ! Compute the derivatives from d-orbitals
            CALL dterep_d_ana(sepi, sepj, rij, rep_d, rep, se_int_control, &
                              se_int_screen, ft, dft)
         END IF

         ! Tapering Integral values
         lasti = sepi%natorb
         lastj = sepj%natorb
         DO i = 1, lasti
            li = l_index(i)
            DO j = 1, i
               lj = l_index(j)
               ij = indexa(i, j)
               DO k = 1, lastj
                  lk = l_index(k)
                  DO l = 1, k
                     ll = l_index(l)
                     kl = indexa(k, l)
                     numb = ijkl_ind(ij, kl)
                     IF (numb > 0) rep(numb) = rep(numb)*ft
                  END DO
               END DO
            END DO
         END DO
      END IF

      ! Possibly debug 2el 2cent integrals and derivatives
      IF (debug_this_module) THEN
         CALL check_dterep_ana(sepi, sepj, rij, rep, rep_d, se_int_control, se_taper=se_taper, &
                               lgrad=lgrad)
      END IF
   END SUBROUTINE dterep_ana

! **************************************************************************************************
!> \brief Calculates the derivative and the value of two-electron repulsion
!>      integrals and the nuclear attraction integrals w.r.t. |r| - sp shells only
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij interatomic distance
!> \param drep array of derivatives of two-electron repulsion integrals
!> \param rep array of two-electron repulsion integrals
!> \param se_int_control input parameters that control the calculation of SE
!>                         integrals (shortrange, R3 residual, screening type)
!> \param se_int_screen ...
!> \param ft ...
!> \param dft ...
!> \par History
!>      04.2007 created [tlaino]
!>      Teodoro Laino (03.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integrals
!>      05.2008 Teodoro Laino [tlaino] - University of Zurich: major rewriting
!> \author Teodoro Laino - Zurich University
!> \note
!>      Analytical version - Analytical evaluation of gradients
!>      Teodoro Laino - Zurich University 04.2007
!>      routine adapted from mopac7 (repp)
!>      vector version written by Ernest R. Davidson, Indiana University
! **************************************************************************************************
   SUBROUTINE dterep_sp_ana(sepi, sepj, rij, drep, rep, se_int_control, &
                            se_int_screen, ft, dft)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), DIMENSION(491), INTENT(OUT)              :: drep
      REAL(dp), DIMENSION(491), INTENT(IN)               :: rep
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen
      REAL(dp), INTENT(IN)                               :: ft, dft

      CHARACTER(len=*), PARAMETER :: routineN = 'dterep_sp_ana', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ij, j, k, kl, l, lasti, lastj, li, &
                                                            lj, lk, ll, nold, numb
      REAL(KIND=dp)                                      :: tmp

      lasti = sepi%natorb
      lastj = sepj%natorb
      DO i = 1, MIN(lasti, 4)
         li = l_index(i)
         DO j = 1, i
            lj = l_index(j)
            ij = indexa(i, j)
            DO k = 1, MIN(lastj, 4)
               lk = l_index(k)
               DO l = 1, k
                  ll = l_index(l)
                  kl = indexa(k, l)

                  numb = ijkl_ind(ij, kl)
                  IF (numb > 0) THEN
                     nold = ijkl_sym(numb)
                     IF (nold > 0) THEN
                        drep(numb) = drep(nold)
                     ELSE IF (nold < 0) THEN
                        drep(numb) = -drep(-nold)
                     ELSE IF (nold == 0) THEN
                        tmp = d_ijkl_sp(sepi, sepj, ij, kl, li, lj, lk, ll, 0, rij, &
                                        se_int_control, se_int_screen, do_method_undef)
                        drep(numb) = dft*rep(numb)+ft*tmp
                     END IF
                  END IF
               END DO
            END DO
         END DO
      END DO
   END SUBROUTINE dterep_sp_ana

! **************************************************************************************************
!> \brief Calculates the derivatives of the two-electron repulsion integrals - d shell only
!> \param sepi paramters of atom i
!> \param sepj paramters of atom j
!> \param rij interatomic distance
!> \param drep ...
!> \param rep array of two-electron repulsion integrals
!> \param se_int_control input parameters that control the calculation of
!>                         integrals (shortrange, R3 residual, screening type)
!> \param se_int_screen ...
!> \param ft ...
!> \param dft ...
!> \par History
!>      Teodoro Laino (05.2008) [tlaino] - University of Zurich : new driver
!>                 for computing integral derivatives for d-orbitals
! **************************************************************************************************
   SUBROUTINE dterep_d_ana(sepi, sepj, rij, drep, rep, se_int_control, &
                           se_int_screen, ft, dft)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), INTENT(IN)                               :: rij
      REAL(dp), DIMENSION(491), INTENT(INOUT)            :: drep
      REAL(dp), DIMENSION(491), INTENT(IN)               :: rep
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_int_screen_type), INTENT(IN)               :: se_int_screen
      REAL(dp), INTENT(IN)                               :: ft, dft

      CHARACTER(len=*), PARAMETER :: routineN = 'dterep_d_ana', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ij, j, k, kl, l, lasti, lastj, li, &
                                                            lj, lk, ll, nold, numb
      REAL(KIND=dp)                                      :: tmp

      lasti = sepi%natorb
      lastj = sepj%natorb
      DO i = 1, lasti
         li = l_index(i)
         DO j = 1, i
            lj = l_index(j)
            ij = indexa(i, j)
            DO k = 1, lastj
               lk = l_index(k)
               DO l = 1, k
                  ll = l_index(l)
                  kl = indexa(k, l)

                  numb = ijkl_ind(ij, kl)
                  ! From 1 to 34 we store integrals involving sp shells
                  IF (numb > 34) THEN
                     nold = ijkl_sym(numb)
                     IF (nold > 34) THEN
                        drep(numb) = drep(nold)
                     ELSE IF (nold < -34) THEN
                        drep(numb) = -drep(-nold)
                     ELSE IF (nold == 0) THEN
                        tmp = d_ijkl_d(sepi, sepj, ij, kl, li, lj, lk, ll, 0, rij, &
                                       se_int_control, se_int_screen, do_method_undef)
                        drep(numb) = dft*rep(numb)+ft*tmp
                     END IF
                  END IF
               END DO
            END DO
         END DO
      END DO
   END SUBROUTINE dterep_d_ana

END MODULE semi_empirical_int_ana
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Integral GKS scheme: The order of the integrals in makeCoul reflects
!>        the standard order by MOPAC
!> \par History
!>      Teodoro Laino [tlaino] - 04.2009 : Adapted size arrays to d-orbitals and
!>                               get rid of the alternative ordering. Using the
!>                               CP2K one.
!>      Teodoro Laino [tlaino] - 04.2009 : Skip nullification (speed-up)
!>      Teodoro Laino [tlaino] - 04.2009 : Speed-up due to fortran arrays order
!>                               optimization and collection of common pieces of
!>                               code
! **************************************************************************************************
MODULE semi_empirical_int_gks

   USE dg_rho0_types,                   ONLY: dg_rho0_type
   USE dg_types,                        ONLY: dg_get,&
                                              dg_type
   USE erf_fn,                          ONLY: erfc
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi,&
                                              oorootpi
   USE multipole_types,                 ONLY: do_multipole_none
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_pool_types,                   ONLY: pw_pool_type
   USE semi_empirical_int_arrays,       ONLY: indexb,&
                                              rij_threshold
   USE semi_empirical_mpole_types,      ONLY: semi_empirical_mpole_type
   USE semi_empirical_types,            ONLY: se_int_control_type,&
                                              semi_empirical_type,&
                                              setup_se_int_control_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_int_gks'
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.

   PUBLIC :: corecore_gks, rotnuc_gks, drotnuc_gks, rotint_gks, drotint_gks

CONTAINS

! **************************************************************************************************
!> \brief Computes the electron-nuclei integrals
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param e1b ...
!> \param e2a ...
!> \param se_int_control ...
! **************************************************************************************************
   SUBROUTINE rotnuc_gks(sepi, sepj, rij, e1b, e2a, se_int_control)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), DIMENSION(45), INTENT(OUT), OPTIONAL     :: e1b, e2a
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'rotnuc_gks', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, mu, nu
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      REAL(kind=dp), DIMENSION(45, 45)                   :: Coul

      rab = -rij

      IF (se_int_control%do_ewald_gks) THEN
         IF (DOT_PRODUCT(rij, rij) > rij_threshold) THEN
            CALL makeCoulE(rab, sepi, sepj, Coul, se_int_control)
         ELSE
            CALL makeCoulE0(sepi, Coul, se_int_control)
         END IF
      ELSE
         CALL makeCoul(rab, sepi, sepj, Coul, se_int_control)
      END IF

      i = 0
      DO mu = 1, sepi%natorb
         DO nu = 1, mu
            i = i+1
            e1b(i) = -Coul(i, 1)*sepj%zeff
         END DO
      END DO

      i = 0
      DO mu = 1, sepj%natorb
         DO nu = 1, mu
            i = i+1
            e2a(i) = -Coul(1, i)*sepi%zeff
         END DO
      END DO

   END SUBROUTINE rotnuc_gks

! **************************************************************************************************
!> \brief Computes the electron-electron integrals
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param w ...
!> \param se_int_control ...
! **************************************************************************************************
   SUBROUTINE rotint_gks(sepi, sepj, rij, w, se_int_control)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), DIMENSION(2025), INTENT(OUT), OPTIONAL   :: w
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'rotint_gks', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ind1, ind2, lam, mu, nu, sig
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      REAL(kind=dp), DIMENSION(45, 45)                   :: Coul

      rab = -rij

      IF (se_int_control%do_ewald_gks) THEN
         IF (DOT_PRODUCT(rij, rij) > rij_threshold) THEN
            CALL makeCoulE(rab, sepi, sepj, Coul, se_int_control)
         ELSE
            CALL makeCoulE0(sepi, Coul, se_int_control)
         END IF
      ELSE
         CALL makeCoul(rab, sepi, sepj, Coul, se_int_control)
      END IF

      i = 0
      ind1 = 0
      DO mu = 1, sepi%natorb
         DO nu = 1, mu
            ind1 = ind1+1
            ind2 = 0
            DO lam = 1, sepj%natorb
               DO sig = 1, lam
                  i = i+1
                  ind2 = ind2+1
                  w(i) = Coul(ind1, ind2)
               END DO
            END DO
         END DO
      END DO

   END SUBROUTINE rotint_gks

! **************************************************************************************************
!> \brief Computes the derivatives of the electron-nuclei integrals
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param de1b ...
!> \param de2a ...
!> \param se_int_control ...
! **************************************************************************************************
   SUBROUTINE drotnuc_gks(sepi, sepj, rij, de1b, de2a, se_int_control)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), DIMENSION(3, 45), INTENT(OUT), OPTIONAL  :: de1b, de2a
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'drotnuc_gks', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, mu, nu
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      REAL(kind=dp), DIMENSION(3, 45, 45)                :: dCoul

      rab = -rij

      IF (se_int_control%do_ewald_gks) THEN
         CALL makedCoulE(rab, sepi, sepj, dCoul, se_int_control)
      ELSE
         CALL makedCoul(rab, sepi, sepj, dCoul, se_int_control)
      END IF

      i = 0
      DO mu = 1, sepi%natorb
         DO nu = 1, mu
            i = i+1
            de1b(1, i) = dCoul(1, i, 1)*sepj%zeff
            de1b(2, i) = dCoul(2, i, 1)*sepj%zeff
            de1b(3, i) = dCoul(3, i, 1)*sepj%zeff
         END DO
      END DO

      i = 0
      DO mu = 1, sepj%natorb
         DO nu = 1, mu
            i = i+1
            de2a(1, i) = dCoul(1, 1, i)*sepi%zeff
            de2a(2, i) = dCoul(2, 1, i)*sepi%zeff
            de2a(3, i) = dCoul(3, 1, i)*sepi%zeff
         END DO
      END DO

   END SUBROUTINE drotnuc_gks

! **************************************************************************************************
!> \brief Computes the derivatives of the electron-electron integrals
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param dw ...
!> \param se_int_control ...
! **************************************************************************************************
   SUBROUTINE drotint_gks(sepi, sepj, rij, dw, se_int_control)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), DIMENSION(3, 2025), INTENT(OUT)          :: dw
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'drotint_gks', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ind1, ind2, lam, mu, nu, sig
      REAL(KIND=dp), DIMENSION(3)                        :: rab
      REAL(kind=dp), DIMENSION(3, 45, 45)                :: dCoul

      rab = -rij

      IF (se_int_control%do_ewald_gks) THEN
         CALL makedCoulE(rab, sepi, sepj, dCoul, se_int_control)
      ELSE
         CALL makedCoul(rab, sepi, sepj, dCoul, se_int_control)
      END IF

      i = 0
      ind1 = 0
      DO mu = 1, sepi%natorb
         DO nu = 1, mu
            ind1 = ind1+1
            ind2 = 0
            DO lam = 1, sepj%natorb
               DO sig = 1, lam
                  i = i+1
                  ind2 = ind2+1
                  dw(1, i) = -dCoul(1, ind1, ind2)
                  dw(2, i) = -dCoul(2, ind1, ind2)
                  dw(3, i) = -dCoul(3, ind1, ind2)
               END DO
            END DO
         END DO
      END DO

   END SUBROUTINE drotint_gks

! **************************************************************************************************
!> \brief Computes the primitives of the integrals (non-periodic case)
!> \param RAB ...
!> \param sepi ...
!> \param sepj ...
!> \param Coul ...
!> \param se_int_control ...
! **************************************************************************************************
   SUBROUTINE makeCoul(RAB, sepi, sepj, Coul, se_int_control)
      REAL(kind=dp), DIMENSION(3)                        :: RAB
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(kind=dp), DIMENSION(45, 45), INTENT(OUT)      :: Coul
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'makeCoul', routineP = moduleN//':'//routineN

      INTEGER                                            :: iA, iB, imA, imB, jA, jB, k1, k2, k3, k4
      LOGICAL                                            :: shortrange
      REAL(kind=dp)                                      :: a2, ACOULA, ACOULB, d1, d1f(3), d2, &
                                                            d2f(3, 3), d3, d3f(3, 3, 3), d4, &
                                                            d4f(3, 3, 3, 3), f, rr, w, w0, w1, w2, &
                                                            w3, w4, w5
      REAL(kind=dp), DIMENSION(3)                        :: v
      REAL(kind=dp), DIMENSION(3, 3, 45)                 :: M2A, M2B
      REAL(kind=dp), DIMENSION(3, 45)                    :: M1A, M1B
      REAL(kind=dp), DIMENSION(45)                       :: M0A, M0B

      shortrange = se_int_control%shortrange
      CALL get_se_slater_multipole(sepi, M0A, M1A, M2A, ACOULA)
      CALL get_se_slater_multipole(sepj, M0B, M1B, M2B, ACOULB)

      v(1) = RAB(1)
      v(2) = RAB(2)
      v(3) = RAB(3)
      rr = SQRT(DOT_PRODUCT(v, v))

      a2 = 0.5_dp*(1.0_dp/ACOULA+1.0_dp/ACOULB)
      w0 = a2*rr
      w = EXP(-w0)
      w1 = (1.0_dp+0.5_dp*w0)
      w2 = (w1+0.5_dp*w0+0.5_dp*w0**2)
      w3 = (w2+w0**3/6.0_dp)
      w4 = (w3+w0**4/30.0_dp)
      w5 = (w3+8.0_dp*w0**4/210.0_dp+w0**5/210.0_dp)

      IF (shortrange) THEN
         f = (-w*w1)/rr
         d1 = -1.0_dp*(-w*w2)/rr**3
         d2 = 3.0_dp*(-w*w3)/rr**5
         d3 = -15.0_dp*(-w*w4)/rr**7
         d4 = 105.0_dp*(-w*w5)/rr**9
      ELSE
         f = (1.0_dp-w*w1)/rr
         d1 = -1.0_dp*(1.0_dp-w*w2)/rr**3
         d2 = 3.0_dp*(1.0_dp-w*w3)/rr**5
         d3 = -15.0_dp*(1.0_dp-w*w4)/rr**7
         d4 = 105.0_dp*(1.0_dp-w*w5)/rr**9
      ENDIF

      CALL build_d_tensor_gks(d1f, d2f, d3f, d4f, v=v, d1=d1, d2=d2, d3=d3, d4=d4)

      imA = 0
      DO iA = 1, sepi%natorb
         DO jA = 1, iA
            imA = imA+1

            imB = 0
            DO iB = 1, sepj%natorb
               DO jB = 1, iB
                  imB = imB+1

                  w = M0A(imA)*M0B(imB)*f
                  DO k1 = 1, 3
                     w = w+(M1A(k1, imA)*M0B(imB)-M0A(imA)*M1B(k1, imB))*d1f(k1)
                  ENDDO
                  DO k2 = 1, 3
                     DO k1 = 1, 3
                        w = w+(M2A(k1, k2, imA)*M0B(imB)-M1A(k1, imA)*M1B(k2, imB)+M0A(imA)*M2B(k1, k2, imB))*d2f(k1, k2)
                     ENDDO
                  ENDDO
                  DO k3 = 1, 3
                     DO k2 = 1, 3
                        DO k1 = 1, 3
                           w = w+(-M2A(k1, k2, imA)*M1B(k3, imB)+M1A(k1, imA)*M2B(k2, k3, imB))*d3f(k1, k2, k3)
                        ENDDO
                     ENDDO
                  ENDDO

                  DO k4 = 1, 3
                     DO k3 = 1, 3
                        DO k2 = 1, 3
                           DO k1 = 1, 3
                              w = w+M2A(k1, k2, imA)*M2B(k3, k4, imB)*d4f(k1, k2, k3, k4)
                           ENDDO
                        ENDDO
                     ENDDO
                  ENDDO

                  Coul(imA, imB) = w
               ENDDO
            ENDDO
         ENDDO
      ENDDO

   END SUBROUTINE makeCoul

! **************************************************************************************************
!> \brief Computes the derivatives of the primitives of the integrals
!>        (non-periodic case)
!> \param RAB ...
!> \param sepi ...
!> \param sepj ...
!> \param dCoul ...
!> \param se_int_control ...
! **************************************************************************************************
   SUBROUTINE makedCoul(RAB, sepi, sepj, dCoul, se_int_control)
      REAL(kind=dp), DIMENSION(3)                        :: RAB
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(kind=dp), DIMENSION(3, 45, 45), INTENT(OUT)   :: dCoul
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'makedCoul', routineP = moduleN//':'//routineN

      INTEGER                                            :: iA, iB, imA, imB, jA, jB, k1, k2, k3, k4
      LOGICAL                                            :: shortrange
      REAL(kind=dp) :: a2, ACOULA, ACOULB, d1, d1f(3), d2, d2f(3, 3), d3, d3f(3, 3, 3), d4, &
         d4f(3, 3, 3, 3), d5, d5f(3, 3, 3, 3, 3), f, rr, tmp, w, w0, w1, w2, w3, w4, w5, w6
      REAL(kind=dp), DIMENSION(3)                        :: v, wv
      REAL(kind=dp), DIMENSION(3, 3, 45)                 :: M2A, M2B
      REAL(kind=dp), DIMENSION(3, 45)                    :: M1A, M1B
      REAL(kind=dp), DIMENSION(45)                       :: M0A, M0B

      shortrange = se_int_control%shortrange
      CALL get_se_slater_multipole(sepi, M0A, M1A, M2A, ACOULA)
      CALL get_se_slater_multipole(sepj, M0B, M1B, M2B, ACOULB)

      v(1) = RAB(1)
      v(2) = RAB(2)
      v(3) = RAB(3)
      rr = SQRT(DOT_PRODUCT(v, v))

      a2 = 0.5_dp*(1.0_dp/ACOULA+1.0_dp/ACOULB)
      w0 = a2*rr
      w = EXP(-w0)
      w1 = (1.0_dp+0.5_dp*w0)
      w2 = (w1+0.5_dp*w0+0.5_dp*w0**2)
      w3 = (w2+w0**3/6.0_dp)
      w4 = (w3+w0**4/30.0_dp)
      w5 = (w3+4.0_dp*w0**4/105.0_dp+w0**5/210.0_dp)
      w6 = (w3+15.0_dp*w0**4/378.0_dp+2.0_dp*w0**5/315.0_dp+w0**6/1890.0_dp)

      IF (shortrange) THEN
         f = (-w*w1)/rr
         d1 = -1.0_dp*(-w*w2)/rr**3
         d2 = 3.0_dp*(-w*w3)/rr**5
         d3 = -15.0_dp*(-w*w4)/rr**7
         d4 = 105.0_dp*(-w*w5)/rr**9
         d5 = -945.0_dp*(-w*w6)/rr**11
      ELSE
         f = (1.0_dp-w*w1)/rr
         d1 = -1.0_dp*(1.0_dp-w*w2)/rr**3
         d2 = 3.0_dp*(1.0_dp-w*w3)/rr**5
         d3 = -15.0_dp*(1.0_dp-w*w4)/rr**7
         d4 = 105.0_dp*(1.0_dp-w*w5)/rr**9
         d5 = -945.0_dp*(1.0_dp-w*w6)/rr**11
      ENDIF

      CALL build_d_tensor_gks(d1f, d2f, d3f, d4f, d5f, v, d1, d2, d3, d4, d5)

      imA = 0
      DO iA = 1, sepi%natorb
         DO jA = 1, iA
            imA = imA+1

            imB = 0
            DO iB = 1, sepj%natorb
               DO jB = 1, iB
                  imB = imB+1

                  tmp = M0A(imA)*M0B(imB)
                  wv(1) = tmp*d1f(1)
                  wv(2) = tmp*d1f(2)
                  wv(3) = tmp*d1f(3)
                  DO k1 = 1, 3
                     tmp = M1A(k1, imA)*M0B(imB)-M0A(imA)*M1B(k1, imB)
                     wv(1) = wv(1)+tmp*d2f(1, k1)
                     wv(2) = wv(2)+tmp*d2f(2, k1)
                     wv(3) = wv(3)+tmp*d2f(3, k1)
                  ENDDO
                  DO k2 = 1, 3
                     DO k1 = 1, 3
                        tmp = M2A(k1, k2, imA)*M0B(imB)-M1A(k1, imA)*M1B(k2, imB)+M0A(imA)*M2B(k1, k2, imB)
                        wv(1) = wv(1)+tmp*d3f(1, k1, k2)
                        wv(2) = wv(2)+tmp*d3f(2, k1, k2)
                        wv(3) = wv(3)+tmp*d3f(3, k1, k2)
                     ENDDO
                  ENDDO
                  DO k3 = 1, 3
                     DO k2 = 1, 3
                        DO k1 = 1, 3
                           tmp = -M2A(k1, k2, imA)*M1B(k3, imB)+M1A(k1, imA)*M2B(k2, k3, imB)
                           wv(1) = wv(1)+tmp*d4f(1, k1, k2, k3)
                           wv(2) = wv(2)+tmp*d4f(2, k1, k2, k3)
                           wv(3) = wv(3)+tmp*d4f(3, k1, k2, k3)
                        ENDDO
                     ENDDO
                  ENDDO

                  DO k4 = 1, 3
                     DO k3 = 1, 3
                        DO k2 = 1, 3
                           DO k1 = 1, 3
                              tmp = M2A(k1, k2, imA)*M2B(k3, k4, imB)
                              wv(1) = wv(1)+tmp*d5f(1, k1, k2, k3, k4)
                              wv(2) = wv(2)+tmp*d5f(2, k1, k2, k3, k4)
                              wv(3) = wv(3)+tmp*d5f(3, k1, k2, k3, k4)
                           ENDDO
                        ENDDO
                     ENDDO
                  ENDDO

                  dCoul(1, imA, imB) = wv(1)
                  dCoul(2, imA, imB) = wv(2)
                  dCoul(3, imA, imB) = wv(3)
               ENDDO
            ENDDO
         ENDDO
      ENDDO

   END SUBROUTINE makedCoul

! **************************************************************************************************
!> \brief Computes nuclei-nuclei interactions
!> \param sepi ...
!> \param sepj ...
!> \param rijv ...
!> \param enuc ...
!> \param denuc ...
!> \param se_int_control ...
! **************************************************************************************************
   SUBROUTINE corecore_gks(sepi, sepj, rijv, enuc, denuc, se_int_control)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rijv
      REAL(dp), INTENT(OUT), OPTIONAL                    :: enuc
      REAL(dp), DIMENSION(3), INTENT(OUT), OPTIONAL      :: denuc
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'corecore_gks', routineP = moduleN//':'//routineN

      LOGICAL                                            :: l_denuc, l_enuc
      REAL(dp)                                           :: alpi, alpj, dscale, rij, scale, zz
      REAL(kind=dp), DIMENSION(3, 45, 45)                :: dCoul, dCoulE
      REAL(kind=dp), DIMENSION(45, 45)                   :: Coul, CoulE
      TYPE(se_int_control_type)                          :: se_int_control_off

      rij = DOT_PRODUCT(rijv, rijv)

      l_enuc = PRESENT(enuc)
      l_denuc = PRESENT(denuc)
      IF ((rij > rij_threshold) .AND. (l_enuc .OR. l_denuc)) THEN

         rij = SQRT(rij)

         IF (se_int_control%shortrange) THEN
            CALL setup_se_int_control_type(se_int_control_off, shortrange=.FALSE., do_ewald_r3=.FALSE., &
                                           do_ewald_gks=.FALSE., integral_screening=se_int_control%integral_screening, &
                                           max_multipole=do_multipole_none, pc_coulomb_int=.FALSE.)
            CALL makeCoul(rijv, sepi, sepj, Coul, se_int_control_off)
            IF (l_denuc) CALL makedCoul(rijv, sepi, sepj, dCoul, se_int_control_off)
            IF (se_int_control%do_ewald_gks) THEN
               CALL makeCoulE(rijv, sepi, sepj, CoulE, se_int_control)
               IF (l_denuc) CALL makedCoulE(rijv, sepi, sepj, dCoulE, se_int_control)
            ELSE
               CALL makeCoul(rijv, sepi, sepj, CoulE, se_int_control)
               IF (l_denuc) CALL makedCoul(rijv, sepi, sepj, dCoulE, se_int_control)
            END IF
         ELSE
            CALL makeCoul(rijv, sepi, sepj, Coul, se_int_control)
            CoulE = Coul
            IF (l_denuc) CALL makedCoul(rijv, sepi, sepj, dCoul, se_int_control)
            IF (l_denuc) dCoulE = dCoul
         END IF

         scale = 0.0_dp
         dscale = 0.0_dp
         zz = sepi%zeff*sepj%zeff
         alpi = sepi%alp
         alpj = sepj%alp
         scale = EXP(-alpi*rij)+EXP(-alpj*rij)
         IF (l_enuc) THEN
            enuc = zz*CoulE(1, 1)+scale*zz*Coul(1, 1)
         END IF
         IF (l_denuc) THEN
            dscale = -alpi*EXP(-alpi*rij)-alpj*EXP(-alpj*rij)
            denuc(1) = zz*dCoulE(1, 1, 1)+dscale*(rijv(1)/rij)*zz*Coul(1, 1)+scale*zz*dCoul(1, 1, 1)
            denuc(2) = zz*dCoulE(2, 1, 1)+dscale*(rijv(2)/rij)*zz*Coul(1, 1)+scale*zz*dCoul(2, 1, 1)
            denuc(3) = zz*dCoulE(3, 1, 1)+dscale*(rijv(3)/rij)*zz*Coul(1, 1)+scale*zz*dCoul(3, 1, 1)
         END IF

      ELSE

         IF (se_int_control%do_ewald_gks) THEN
            zz = sepi%zeff*sepi%zeff
            CALL makeCoulE0(sepi, CoulE, se_int_control)
            IF (l_enuc) THEN
               enuc = zz*CoulE(1, 1)
            END IF
            IF (l_denuc) THEN
               denuc = 0._dp
            END IF
         END IF

      ENDIF
   END SUBROUTINE corecore_gks

! **************************************************************************************************
!> \brief Computes the primitives of the integrals (periodic case)
!> \param RAB ...
!> \param sepi ...
!> \param sepj ...
!> \param Coul ...
!> \param se_int_control ...
! **************************************************************************************************
   SUBROUTINE makeCoulE(RAB, sepi, sepj, Coul, se_int_control)
      REAL(KIND=dp), DIMENSION(3)                        :: RAB
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(45, 45), INTENT(OUT)      :: Coul
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'makeCoulE', routineP = moduleN//':'//routineN

      INTEGER                                            :: gpt, imA, imB, k1, k2, k3, k4, lp, mp, np
      INTEGER, DIMENSION(:, :), POINTER                  :: bds
      REAL(KIND=dp) :: a2, ACOULA, ACOULB, alpha, cc, d1, d1f(3), d2, d2f(3, 3), d3, d3f(3, 3, 3), &
         d4, d4f(3, 3, 3, 3), f, ff, kr, kr2, r1, r2, r3, r5, r7, r9, rr, ss, w, w0, w1, w2, w3, &
         w4, w5
      REAL(KIND=dp), DIMENSION(3)                        :: kk, v
      REAL(KIND=dp), DIMENSION(3, 3, 45)                 :: M2A, M2B
      REAL(KIND=dp), DIMENSION(3, 45)                    :: M1A, M1B
      REAL(KIND=dp), DIMENSION(45)                       :: M0A, M0B
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: rho0
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0
      TYPE(dg_type), POINTER                             :: dg
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_pool_type), POINTER                        :: pw_pool

      alpha = se_int_control%ewald_gks%alpha
      pw_pool => se_int_control%ewald_gks%pw_pool
      dg => se_int_control%ewald_gks%dg
      CALL dg_get(dg, dg_rho0=dg_rho0)
      rho0 => dg_rho0%density%pw%cr3d
      pw_grid => pw_pool%pw_grid
      bds => pw_grid%bounds

      CALL get_se_slater_multipole(sepi, M0A, M1A, M2A, ACOULA)
      CALL get_se_slater_multipole(sepj, M0B, M1B, M2B, ACOULB)

      v(1) = RAB(1)
      v(2) = RAB(2)
      v(3) = RAB(3)
      rr = SQRT(DOT_PRODUCT(v, v))

      r1 = 1.0_dp/rr
      r2 = r1*r1
      r3 = r2*r1
      r5 = r3*r2
      r7 = r5*r2
      r9 = r7*r2

      a2 = 0.5_dp*(1.0_dp/ACOULA+1.0_dp/ACOULB)

      w0 = a2*rr
      w = EXP(-w0)
      w1 = (1.0_dp+0.5_dp*w0)
      w2 = (w1+0.5_dp*w0+0.5_dp*w0**2)
      w3 = (w2+w0**3/6.0_dp)
      w4 = (w3+w0**4/30.0_dp)
      w5 = (w3+8.0_dp*w0**4/210.0_dp+w0**5/210.0_dp)

      f = (1.0_dp-w*w1)*r1
      d1 = -1.0_dp*(1.0_dp-w*w2)*r3
      d2 = 3.0_dp*(1.0_dp-w*w3)*r5
      d3 = -15.0_dp*(1.0_dp-w*w4)*r7
      d4 = 105.0_dp*(1.0_dp-w*w5)*r9

      kr = alpha*rr
      kr2 = kr*kr
      w0 = 1.0_dp-erfc(kr)
      w1 = 2.0_dp*oorootpi*EXP(-kr2)
      w2 = w1*kr

      f = f-w0*r1
      d1 = d1+(-w2+w0)*r3
      d2 = d2+(w2*(3.0_dp+kr2*2.0_dp)-3.0_dp*w0)*r5
      d3 = d3+(-w2*(15.0_dp+kr2*(10.0_dp+kr2*4.0_dp))+15.0_dp*w0)*r7
      d4 = d4+(w2*(105.0_dp+kr2*(70.0_dp+kr2*(28.0_dp+kr2*8.0_dp)))-105.0_dp*w0)*r9

      CALL build_d_tensor_gks(d1f, d2f, d3f, d4f, v=v, d1=d1, d2=d2, d3=d3, d4=d4)

      DO imA = 1, (sepi%natorb*(sepi%natorb+1))/2
         DO imB = 1, (sepj%natorb*(sepj%natorb+1))/2

            w = M0A(imA)*M0B(imB)*f
            DO k1 = 1, 3
               w = w+(M1A(k1, imA)*M0B(imB)-M0A(imA)*M1B(k1, imB))*d1f(k1)
            ENDDO
            DO k2 = 1, 3
               DO k1 = 1, 3
                  w = w+(M2A(k1, k2, imA)*M0B(imB)-M1A(k1, imA)*M1B(k2, imB)+M0A(imA)*M2B(k1, k2, imB))*d2f(k1, k2)
               ENDDO
            ENDDO
            DO k3 = 1, 3
               DO k2 = 1, 3
                  DO k1 = 1, 3
                     w = w+(-M2A(k1, k2, imA)*M1B(k3, imB)+M1A(k1, imA)*M2B(k2, k3, imB))*d3f(k1, k2, k3)
                  ENDDO
               ENDDO
            ENDDO

            DO k4 = 1, 3
               DO k3 = 1, 3
                  DO k2 = 1, 3
                     DO k1 = 1, 3
                        w = w+M2A(k1, k2, imA)*M2B(k3, k4, imB)*d4f(k1, k2, k3, k4)
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO

            Coul(imA, imB) = w
         ENDDO
      ENDDO

      v(1) = RAB(1)
      v(2) = RAB(2)
      v(3) = RAB(3)

      f = 0.0_dp
      d1f = 0.0_dp
      d2f = 0.0_dp
      d3f = 0.0_dp
      d4f = 0.0_dp

      DO gpt = 1, pw_grid%ngpts_cut_local
         lp = pw_grid%mapl%pos(pw_grid%g_hat(1, gpt))
         mp = pw_grid%mapm%pos(pw_grid%g_hat(2, gpt))
         np = pw_grid%mapn%pos(pw_grid%g_hat(3, gpt))

         lp = lp+bds(1, 1)
         mp = mp+bds(1, 2)
         np = np+bds(1, 3)

         IF (pw_grid%gsq(gpt) == 0.0_dp) CYCLE
         kk(:) = pw_grid%g(:, gpt)
         ff = 2.0_dp*fourpi*rho0(lp, mp, np)**2*pw_grid%vol/pw_grid%gsq(gpt)

         kr = DOT_PRODUCT(kk, v)
         cc = COS(kr)
         ss = SIN(kr)

         f = f+cc*ff
         DO k1 = 1, 3
            d1f(k1) = d1f(k1)-kk(k1)*ss*ff
         ENDDO
         DO k2 = 1, 3
            DO k1 = 1, 3
               d2f(k1, k2) = d2f(k1, k2)-kk(k1)*kk(k2)*cc*ff
            ENDDO
         ENDDO
         DO k3 = 1, 3
            DO k2 = 1, 3
               DO k1 = 1, 3
                  d3f(k1, k2, k3) = d3f(k1, k2, k3)+kk(k1)*kk(k2)*kk(k3)*ss*ff
               ENDDO
            ENDDO
         ENDDO
         DO k4 = 1, 3
            DO k3 = 1, 3
               DO k2 = 1, 3
                  DO k1 = 1, 3
                     d4f(k1, k2, k3, k4) = d4f(k1, k2, k3, k4)+kk(k1)*kk(k2)*kk(k3)*kk(k4)*cc*ff
                  ENDDO
               ENDDO
            ENDDO
         ENDDO

      ENDDO

      DO imA = 1, (sepi%natorb*(sepi%natorb+1))/2
         DO imB = 1, (sepj%natorb*(sepj%natorb+1))/2

            w = M0A(imA)*M0B(imB)*f
            DO k1 = 1, 3
               w = w+(M1A(k1, imA)*M0B(imB)-M0A(imA)*M1B(k1, imB))*d1f(k1)
            ENDDO
            DO k2 = 1, 3
               DO k1 = 1, 3
                  w = w+(M2A(k1, k2, imA)*M0B(imB)-M1A(k1, imA)*M1B(k2, imB)+M0A(imA)*M2B(k1, k2, imB))*d2f(k1, k2)
               ENDDO
            ENDDO
            DO k3 = 1, 3
               DO k2 = 1, 3
                  DO k1 = 1, 3
                     w = w+(-M2A(k1, k2, imA)*M1B(k3, imB)+M1A(k1, imA)*M2B(k2, k3, imB))*d3f(k1, k2, k3)
                  ENDDO
               ENDDO
            ENDDO

            DO k4 = 1, 3
               DO k3 = 1, 3
                  DO k2 = 1, 3
                     DO k1 = 1, 3
                        w = w+M2A(k1, k2, imA)*M2B(k3, k4, imB)*d4f(k1, k2, k3, k4)
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO

            Coul(imA, imB) = Coul(imA, imB)+w

         ENDDO
      ENDDO

      DO imA = 1, (sepi%natorb*(sepi%natorb+1))/2
         DO imB = 1, (sepj%natorb*(sepj%natorb+1))/2
            w = -M0A(imA)*M0B(imB)*0.25_dp*fourpi/(pw_grid%vol*alpha**2)
            Coul(imA, imB) = Coul(imA, imB)+w
         ENDDO
      ENDDO

   END SUBROUTINE makeCoulE

! **************************************************************************************************
!> \brief Computes the derivatives of the primitives of the integrals
!>        (periodic case)
!> \param RAB ...
!> \param sepi ...
!> \param sepj ...
!> \param dCoul ...
!> \param se_int_control ...
! **************************************************************************************************
   SUBROUTINE makedCoulE(RAB, sepi, sepj, dCoul, se_int_control)
      REAL(KIND=dp), DIMENSION(3)                        :: RAB
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(3, 45, 45), INTENT(OUT)   :: dCoul
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'makedCoulE', routineP = moduleN//':'//routineN

      INTEGER                                            :: gpt, imA, imB, k1, k2, k3, k4, k5, lp, &
                                                            mp, np
      INTEGER, DIMENSION(:, :), POINTER                  :: bds
      REAL(KIND=dp) :: a2, ACOULA, ACOULB, alpha, cc, d1, d1f(3), d2, d2f(3, 3), d3, d3f(3, 3, 3), &
         d4, d4f(3, 3, 3, 3), d5, d5f(3, 3, 3, 3, 3), f, ff, kr, kr2, r1, r11, r2, r3, r5, r7, r9, &
         rr, ss, tmp, w, w0, w1, w2, w3, w4, w5, w6
      REAL(KIND=dp), DIMENSION(3)                        :: kk, v, wv
      REAL(kind=dp), DIMENSION(3, 3, 45)                 :: M2A, M2B
      REAL(kind=dp), DIMENSION(3, 45)                    :: M1A, M1B
      REAL(kind=dp), DIMENSION(45)                       :: M0A, M0B
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: rho0
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0
      TYPE(dg_type), POINTER                             :: dg
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_pool_type), POINTER                        :: pw_pool

      alpha = se_int_control%ewald_gks%alpha
      pw_pool => se_int_control%ewald_gks%pw_pool
      dg => se_int_control%ewald_gks%dg
      CALL dg_get(dg, dg_rho0=dg_rho0)
      rho0 => dg_rho0%density%pw%cr3d
      pw_grid => pw_pool%pw_grid
      bds => pw_grid%bounds

      CALL get_se_slater_multipole(sepi, M0A, M1A, M2A, ACOULA)
      CALL get_se_slater_multipole(sepj, M0B, M1B, M2B, ACOULB)

      v(1) = RAB(1)
      v(2) = RAB(2)
      v(3) = RAB(3)
      rr = SQRT(DOT_PRODUCT(v, v))

      a2 = 0.5_dp*(1.0_dp/ACOULA+1.0_dp/ACOULB)

      r1 = 1.0_dp/rr
      r2 = r1*r1
      r3 = r2*r1
      r5 = r3*r2
      r7 = r5*r2
      r9 = r7*r2
      r11 = r9*r2

      w0 = a2*rr
      w = EXP(-w0)
      w1 = (1.0_dp+0.5_dp*w0)
      w2 = (w1+0.5_dp*w0+0.5_dp*w0**2)
      w3 = (w2+w0**3/6.0_dp)
      w4 = (w3+w0**4/30.0_dp)
      w5 = (w3+8.0_dp*w0**4/210.0_dp+w0**5/210.0_dp)
      w6 = (w3+5.0_dp*w0**4/126.0_dp+2.0_dp*w0**5/315.0_dp+w0**6/1890.0_dp)

      f = (1.0_dp-w*w1)*r1
      d1 = -1.0_dp*(1.0_dp-w*w2)*r3
      d2 = 3.0_dp*(1.0_dp-w*w3)*r5
      d3 = -15.0_dp*(1.0_dp-w*w4)*r7
      d4 = 105.0_dp*(1.0_dp-w*w5)*r9
      d5 = -945.0_dp*(1.0_dp-w*w6)*r11

      kr = alpha*rr
      kr2 = kr*kr
      w0 = 1.0_dp-erfc(kr)
      w1 = 2.0_dp*oorootpi*EXP(-kr2)
      w2 = w1*kr

      f = f-w0*r1
      d1 = d1+(-w2+w0)*r3
      d2 = d2+(w2*(3.0_dp+kr2*2.0_dp)-3.0_dp*w0)*r5
      d3 = d3+(-w2*(15.0_dp+kr2*(10.0_dp+kr2*4.0_dp))+15.0_dp*w0)*r7
      d4 = d4+(w2*(105.0_dp+kr2*(70.0_dp+kr2*(28.0_dp+kr2*8.0_dp)))-105.0_dp*w0)*r9
      d5 = d5+(-w2*(945.0_dp+kr2*(630.0_dp+kr2*(252.0_dp+kr2*(72.0_dp+kr2*16.0_dp))))+945.0_dp*w0)*r11

      CALL build_d_tensor_gks(d1f, d2f, d3f, d4f, d5f, v, d1, d2, d3, d4, d5)

      DO imA = 1, (sepi%natorb*(sepi%natorb+1))/2
         DO imB = 1, (sepj%natorb*(sepj%natorb+1))/2

            tmp = M0A(imA)*M0B(imB)
            wv(1) = tmp*d1f(1)
            wv(2) = tmp*d1f(2)
            wv(3) = tmp*d1f(3)

            DO k1 = 1, 3
               tmp = M1A(k1, imA)*M0B(imB)-M0A(imA)*M1B(k1, imB)
               wv(1) = wv(1)+tmp*d2f(1, k1)
               wv(2) = wv(2)+tmp*d2f(2, k1)
               wv(3) = wv(3)+tmp*d2f(3, k1)
            ENDDO
            DO k2 = 1, 3
               DO k1 = 1, 3
                  tmp = M2A(k1, k2, imA)*M0B(imB)-M1A(k1, imA)*M1B(k2, imB)+M0A(imA)*M2B(k1, k2, imB)
                  wv(1) = wv(1)+tmp*d3f(1, k1, k2)
                  wv(2) = wv(2)+tmp*d3f(2, k1, k2)
                  wv(3) = wv(3)+tmp*d3f(3, k1, k2)
               ENDDO
            ENDDO
            DO k3 = 1, 3
               DO k2 = 1, 3
                  DO k1 = 1, 3
                     tmp = -M2A(k1, k2, imA)*M1B(k3, imB)+M1A(k1, imA)*M2B(k2, k3, imB)
                     wv(1) = wv(1)+tmp*d4f(1, k1, k2, k3)
                     wv(2) = wv(2)+tmp*d4f(2, k1, k2, k3)
                     wv(3) = wv(3)+tmp*d4f(3, k1, k2, k3)
                  ENDDO
               ENDDO
            ENDDO

            DO k4 = 1, 3
               DO k3 = 1, 3
                  DO k2 = 1, 3
                     DO k1 = 1, 3
                        tmp = M2A(k1, k2, imA)*M2B(k3, k4, imB)
                        wv(1) = wv(1)+tmp*d5f(1, k1, k2, k3, k4)
                        wv(2) = wv(2)+tmp*d5f(2, k1, k2, k3, k4)
                        wv(3) = wv(3)+tmp*d5f(3, k1, k2, k3, k4)
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO

            dCoul(1, imA, imB) = wv(1)
            dCoul(2, imA, imB) = wv(2)
            dCoul(3, imA, imB) = wv(3)
         ENDDO
      ENDDO

      v(1) = RAB(1)
      v(2) = RAB(2)
      v(3) = RAB(3)

      f = 0.0_dp
      d1f = 0.0_dp
      d2f = 0.0_dp
      d3f = 0.0_dp
      d4f = 0.0_dp
      d5f = 0.0_dp

      DO gpt = 1, pw_grid%ngpts_cut_local
         lp = pw_grid%mapl%pos(pw_grid%g_hat(1, gpt))
         mp = pw_grid%mapm%pos(pw_grid%g_hat(2, gpt))
         np = pw_grid%mapn%pos(pw_grid%g_hat(3, gpt))

         lp = lp+bds(1, 1)
         mp = mp+bds(1, 2)
         np = np+bds(1, 3)

         IF (pw_grid%gsq(gpt) == 0.0_dp) CYCLE
         kk(:) = pw_grid%g(:, gpt)
         ff = 2.0_dp*fourpi*rho0(lp, mp, np)**2*pw_grid%vol/pw_grid%gsq(gpt)

         kr = DOT_PRODUCT(kk, v)
         cc = COS(kr)
         ss = SIN(kr)

         f = f+cc*ff
         DO k1 = 1, 3
            d1f(k1) = d1f(k1)-kk(k1)*ss*ff
         ENDDO
         DO k2 = 1, 3
            DO k1 = 1, 3
               d2f(k1, k2) = d2f(k1, k2)-kk(k1)*kk(k2)*cc*ff
            ENDDO
         ENDDO
         DO k3 = 1, 3
            DO k2 = 1, 3
               DO k1 = 1, 3
                  d3f(k1, k2, k3) = d3f(k1, k2, k3)+kk(k1)*kk(k2)*kk(k3)*ss*ff
               ENDDO
            ENDDO
         ENDDO
         DO k4 = 1, 3
            DO k3 = 1, 3
               DO k2 = 1, 3
                  DO k1 = 1, 3
                     d4f(k1, k2, k3, k4) = d4f(k1, k2, k3, k4)+kk(k1)*kk(k2)*kk(k3)*kk(k4)*cc*ff
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
         DO k5 = 1, 3
            DO k4 = 1, 3
               DO k3 = 1, 3
                  DO k2 = 1, 3
                     DO k1 = 1, 3
                        d5f(k1, k2, k3, k4, k5) = d5f(k1, k2, k3, k4, k5)-kk(k1)*kk(k2)*kk(k3)*kk(k4)*kk(k5)*ss*ff
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO
         ENDDO
      ENDDO

      DO imA = 1, (sepi%natorb*(sepi%natorb+1))/2
         DO imB = 1, (sepj%natorb*(sepj%natorb+1))/2
            tmp = M0A(imA)*M0B(imB)
            wv(1) = tmp*d1f(1)
            wv(2) = tmp*d1f(2)
            wv(3) = tmp*d1f(3)
            DO k1 = 1, 3
               tmp = M1A(k1, imA)*M0B(imB)-M0A(imA)*M1B(k1, imB)
               wv(1) = wv(1)+tmp*d2f(1, k1)
               wv(2) = wv(2)+tmp*d2f(2, k1)
               wv(3) = wv(3)+tmp*d2f(3, k1)
            ENDDO
            DO k2 = 1, 3
               DO k1 = 1, 3
                  tmp = M2A(k1, k2, imA)*M0B(imB)-M1A(k1, imA)*M1B(k2, imB)+M0A(imA)*M2B(k1, k2, imB)
                  wv(1) = wv(1)+tmp*d3f(1, k1, k2)
                  wv(2) = wv(2)+tmp*d3f(2, k1, k2)
                  wv(3) = wv(3)+tmp*d3f(3, k1, k2)
               ENDDO
            ENDDO
            DO k3 = 1, 3
               DO k2 = 1, 3
                  DO k1 = 1, 3
                     tmp = -M2A(k1, k2, imA)*M1B(k3, imB)+M1A(k1, imA)*M2B(k2, k3, imB)
                     wv(1) = wv(1)+tmp*d4f(1, k1, k2, k3)
                     wv(2) = wv(2)+tmp*d4f(2, k1, k2, k3)
                     wv(3) = wv(3)+tmp*d4f(3, k1, k2, k3)
                  ENDDO
               ENDDO
            ENDDO

            DO k4 = 1, 3
               DO k3 = 1, 3
                  DO k2 = 1, 3
                     DO k1 = 1, 3
                        tmp = M2A(k1, k2, imA)*M2B(k3, k4, imB)
                        wv(1) = wv(1)+tmp*d5f(1, k1, k2, k3, k4)
                        wv(2) = wv(2)+tmp*d5f(2, k1, k2, k3, k4)
                        wv(3) = wv(3)+tmp*d5f(3, k1, k2, k3, k4)
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO

            dCoul(1, imA, imB) = dCoul(1, imA, imB)+wv(1)
            dCoul(2, imA, imB) = dCoul(2, imA, imB)+wv(2)
            dCoul(3, imA, imB) = dCoul(3, imA, imB)+wv(3)
         ENDDO
      ENDDO

   END SUBROUTINE makedCoulE

! **************************************************************************************************
!> \brief Builds the tensor for the evaluation of the integrals with the
!>        cartesian multipoles
!> \param d1f ...
!> \param d2f ...
!> \param d3f ...
!> \param d4f ...
!> \param d5f ...
!> \param v ...
!> \param d1 ...
!> \param d2 ...
!> \param d3 ...
!> \param d4 ...
!> \param d5 ...
! **************************************************************************************************
   SUBROUTINE build_d_tensor_gks(d1f, d2f, d3f, d4f, d5f, v, d1, d2, d3, d4, d5)
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: d1f
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(OUT)        :: d2f
      REAL(KIND=dp), DIMENSION(3, 3, 3), INTENT(OUT)     :: d3f
      REAL(KIND=dp), DIMENSION(3, 3, 3, 3), INTENT(OUT)  :: d4f
      REAL(KIND=dp), DIMENSION(3, 3, 3, 3, 3), &
         INTENT(OUT), OPTIONAL                           :: d5f
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: v
      REAL(KIND=dp), INTENT(IN)                          :: d1, d2, d3, d4
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: d5

      INTEGER                                            :: k1, k2, k3, k4, k5
      REAL(KIND=dp)                                      :: w

      d1f = 0.0_dp
      d2f = 0.0_dp
      d3f = 0.0_dp
      d4f = 0.0_dp
      DO k1 = 1, 3
         d1f(k1) = d1f(k1)+v(k1)*d1
      ENDDO
      DO k1 = 1, 3
         DO k2 = 1, 3
            d2f(k2, k1) = d2f(k2, k1)+v(k1)*v(k2)*d2
         ENDDO
         d2f(k1, k1) = d2f(k1, k1)+d1
      ENDDO
      DO k1 = 1, 3
         DO k2 = 1, 3
            DO k3 = 1, 3
               d3f(k3, k2, k1) = d3f(k3, k2, k1)+v(k1)*v(k2)*v(k3)*d3
            ENDDO
            w = v(k1)*d2
            d3f(k1, k2, k2) = d3f(k1, k2, k2)+w
            d3f(k2, k1, k2) = d3f(k2, k1, k2)+w
            d3f(k2, k2, k1) = d3f(k2, k2, k1)+w
         ENDDO
      ENDDO
      DO k1 = 1, 3
         DO k2 = 1, 3
            DO k3 = 1, 3
               DO k4 = 1, 3
                  d4f(k4, k3, k2, k1) = d4f(k4, k3, k2, k1)+v(k1)*v(k2)*v(k3)*v(k4)*d4
               ENDDO
               w = v(k1)*v(k2)*d3
               d4f(k1, k2, k3, k3) = d4f(k1, k2, k3, k3)+w
               d4f(k1, k3, k2, k3) = d4f(k1, k3, k2, k3)+w
               d4f(k3, k1, k2, k3) = d4f(k3, k1, k2, k3)+w
               d4f(k1, k3, k3, k2) = d4f(k1, k3, k3, k2)+w
               d4f(k3, k1, k3, k2) = d4f(k3, k1, k3, k2)+w
               d4f(k3, k3, k1, k2) = d4f(k3, k3, k1, k2)+w
            ENDDO
            d4f(k1, k1, k2, k2) = d4f(k1, k1, k2, k2)+d2
            d4f(k1, k2, k1, k2) = d4f(k1, k2, k1, k2)+d2
            d4f(k1, k2, k2, k1) = d4f(k1, k2, k2, k1)+d2
         ENDDO
      ENDDO
      IF (PRESENT(d5f) .AND. PRESENT(d5)) THEN
         d5f = 0.0_dp

         DO k1 = 1, 3
            DO k2 = 1, 3
               DO k3 = 1, 3
                  DO k4 = 1, 3
                     DO k5 = 1, 3
                        d5f(k5, k4, k3, k2, k1) = d5f(k5, k4, k3, k2, k1)+v(k1)*v(k2)*v(k3)*v(k4)*v(k5)*d5
                     ENDDO
                     w = v(k1)*v(k2)*v(k3)*d4
                     d5f(k1, k2, k3, k4, k4) = d5f(k1, k2, k3, k4, k4)+w
                     d5f(k1, k2, k4, k3, k4) = d5f(k1, k2, k4, k3, k4)+w
                     d5f(k1, k4, k2, k3, k4) = d5f(k1, k4, k2, k3, k4)+w
                     d5f(k4, k1, k2, k3, k4) = d5f(k4, k1, k2, k3, k4)+w
                     d5f(k1, k2, k4, k4, k3) = d5f(k1, k2, k4, k4, k3)+w
                     d5f(k1, k4, k2, k4, k3) = d5f(k1, k4, k2, k4, k3)+w
                     d5f(k4, k1, k2, k4, k3) = d5f(k4, k1, k2, k4, k3)+w
                     d5f(k1, k4, k4, k2, k3) = d5f(k1, k4, k4, k2, k3)+w
                     d5f(k4, k1, k4, k2, k3) = d5f(k4, k1, k4, k2, k3)+w
                     d5f(k4, k4, k1, k2, k3) = d5f(k4, k4, k1, k2, k3)+w
                  ENDDO
                  w = v(k1)*d3
                  d5f(k1, k2, k2, k3, k3) = d5f(k1, k2, k2, k3, k3)+w
                  d5f(k1, k2, k3, k2, k3) = d5f(k1, k2, k3, k2, k3)+w
                  d5f(k1, k2, k3, k3, k2) = d5f(k1, k2, k3, k3, k2)+w
                  d5f(k2, k1, k2, k3, k3) = d5f(k2, k1, k2, k3, k3)+w
                  d5f(k2, k1, k3, k2, k3) = d5f(k2, k1, k3, k2, k3)+w
                  d5f(k2, k1, k3, k3, k2) = d5f(k2, k1, k3, k3, k2)+w
                  d5f(k2, k2, k1, k3, k3) = d5f(k2, k2, k1, k3, k3)+w
                  d5f(k2, k3, k1, k2, k3) = d5f(k2, k3, k1, k2, k3)+w
                  d5f(k2, k3, k1, k3, k2) = d5f(k2, k3, k1, k3, k2)+w
                  d5f(k2, k2, k3, k1, k3) = d5f(k2, k2, k3, k1, k3)+w
                  d5f(k2, k3, k2, k1, k3) = d5f(k2, k3, k2, k1, k3)+w
                  d5f(k2, k3, k3, k1, k2) = d5f(k2, k3, k3, k1, k2)+w
                  d5f(k2, k2, k3, k3, k1) = d5f(k2, k2, k3, k3, k1)+w
                  d5f(k2, k3, k2, k3, k1) = d5f(k2, k3, k2, k3, k1)+w
                  d5f(k2, k3, k3, k2, k1) = d5f(k2, k3, k3, k2, k1)+w
               ENDDO
            ENDDO
         ENDDO
      END IF
   END SUBROUTINE build_d_tensor_gks

! **************************************************************************************************
!> \brief ...
!> \param sepi ...
!> \param Coul ...
!> \param se_int_control ...
! **************************************************************************************************
   SUBROUTINE makeCoulE0(sepi, Coul, se_int_control)
      TYPE(semi_empirical_type), POINTER                 :: sepi
      REAL(KIND=dp), DIMENSION(45, 45), INTENT(OUT)      :: Coul
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control

      CHARACTER(len=*), PARAMETER :: routineN = 'makeCoulE0', routineP = moduleN//':'//routineN

      INTEGER                                            :: gpt, imA, imB, k1, k2, k3, k4, lp, mp, np
      INTEGER, DIMENSION(:, :), POINTER                  :: bds
      REAL(KIND=dp)                                      :: alpha, d2f(3, 3), d4f(3, 3, 3, 3), f, &
                                                            ff, w
      REAL(KIND=dp), DIMENSION(3)                        :: kk
      REAL(KIND=dp), DIMENSION(3, 3, 45)                 :: M2A
      REAL(KIND=dp), DIMENSION(3, 45)                    :: M1A
      REAL(KIND=dp), DIMENSION(45)                       :: M0A
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: rho0
      TYPE(dg_rho0_type), POINTER                        :: dg_rho0
      TYPE(dg_type), POINTER                             :: dg
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_pool_type), POINTER                        :: pw_pool

      alpha = se_int_control%ewald_gks%alpha
      pw_pool => se_int_control%ewald_gks%pw_pool
      dg => se_int_control%ewald_gks%dg
      CALL dg_get(dg, dg_rho0=dg_rho0)
      rho0 => dg_rho0%density%pw%cr3d
      pw_grid => pw_pool%pw_grid
      bds => pw_grid%bounds

      CALL get_se_slater_multipole(sepi, M0A, M1A, M2A)

      f = 0.0_dp
      d2f = 0.0_dp
      d4f = 0.0_dp

      DO gpt = 1, pw_grid%ngpts_cut_local
         lp = pw_grid%mapl%pos(pw_grid%g_hat(1, gpt))
         mp = pw_grid%mapm%pos(pw_grid%g_hat(2, gpt))
         np = pw_grid%mapn%pos(pw_grid%g_hat(3, gpt))

         lp = lp+bds(1, 1)
         mp = mp+bds(1, 2)
         np = np+bds(1, 3)

         IF (pw_grid%gsq(gpt) == 0.0_dp) CYCLE
         kk(:) = pw_grid%g(:, gpt)
         ff = 2.0_dp*fourpi*rho0(lp, mp, np)**2*pw_grid%vol/pw_grid%gsq(gpt)

         f = f+ff
         DO k2 = 1, 3
            DO k1 = 1, 3
               d2f(k1, k2) = d2f(k1, k2)-kk(k1)*kk(k2)*ff
            ENDDO
         ENDDO
         DO k4 = 1, 3
            DO k3 = 1, 3
               DO k2 = 1, 3
                  DO k1 = 1, 3
                     d4f(k1, k2, k3, k4) = d4f(k1, k2, k3, k4)+kk(k1)*kk(k2)*kk(k3)*kk(k4)*ff
                  ENDDO
               ENDDO
            ENDDO
         ENDDO

      ENDDO

      DO imA = 1, (sepi%natorb*(sepi%natorb+1))/2
         DO imB = 1, (sepi%natorb*(sepi%natorb+1))/2

            w = M0A(imA)*M0A(imB)*f
            DO k2 = 1, 3
               DO k1 = 1, 3
                  w = w+(M2A(k1, k2, imA)*M0A(imB)-M1A(k1, imA)*M1A(k2, imB)+M0A(imA)*M2A(k1, k2, imB))*d2f(k1, k2)
               ENDDO
            ENDDO

            DO k4 = 1, 3
               DO k3 = 1, 3
                  DO k2 = 1, 3
                     DO k1 = 1, 3
                        w = w+M2A(k1, k2, imA)*M2A(k3, k4, imB)*d4f(k1, k2, k3, k4)
                     ENDDO
                  ENDDO
               ENDDO
            ENDDO

            Coul(imA, imB) = w

         ENDDO
      ENDDO

      DO imA = 1, (sepi%natorb*(sepi%natorb+1))/2
         DO imB = 1, (sepi%natorb*(sepi%natorb+1))/2
            w = -M0A(imA)*M0A(imB)*0.25_dp*fourpi/(pw_grid%vol*alpha**2)
            Coul(imA, imB) = Coul(imA, imB)+w
         ENDDO
      ENDDO

      DO imA = 1, (sepi%natorb*(sepi%natorb+1))/2
         DO imB = 1, (sepi%natorb*(sepi%natorb+1))/2

            w = M0A(imA)*M0A(imB)
            Coul(imA, imB) = Coul(imA, imB)-2.0_dp*alpha*oorootpi*w

            w = 0.0_dp
            DO k1 = 1, 3
               w = w+M1A(k1, imA)*M1A(k1, imB)
               w = w-M0A(imA)*M2A(k1, k1, imB)
               w = w-M2A(k1, k1, imA)*M0A(imB)
            ENDDO
            Coul(imA, imB) = Coul(imA, imB)-4.0_dp*alpha**3*oorootpi*w/3.0_dp

            w = 0.0_dp
            DO k2 = 1, 3
               DO k1 = 1, 3
                  w = w+2.0_dp*M2A(k1, k2, imA)*M2A(k1, k2, imB)
                  w = w+M2A(k1, k1, imA)*M2A(k2, k2, imB)
               ENDDO
            ENDDO
            Coul(imA, imB) = Coul(imA, imB)-8.0_dp*alpha**5*oorootpi*w/5.0_dp
         ENDDO
      ENDDO
   END SUBROUTINE makeCoulE0

! **************************************************************************************************
!> \brief Retrieves the multipole for the Slater integral evaluation
!> \param sepi ...
!> \param M0 ...
!> \param M1 ...
!> \param M2 ...
!> \param ACOUL ...
! **************************************************************************************************
   SUBROUTINE get_se_slater_multipole(sepi, M0, M1, M2, ACOUL)
      TYPE(semi_empirical_type), POINTER                 :: sepi
      REAL(kind=dp), DIMENSION(45), INTENT(OUT)          :: M0
      REAL(kind=dp), DIMENSION(3, 45), INTENT(OUT)       :: M1
      REAL(kind=dp), DIMENSION(3, 3, 45), INTENT(OUT)    :: M2
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: ACOUL

      INTEGER                                            :: i, j, jint, size_1c_int
      TYPE(semi_empirical_mpole_type), POINTER           :: mpole

      NULLIFY (mpole)
      size_1c_int = SIZE(sepi%w_mpole)
      DO jint = 1, size_1c_int
         mpole => sepi%w_mpole(jint)%mpole
         i = mpole%indi
         j = mpole%indj
         M0(indexb(i, j)) = -mpole%cs

         M1(1, indexb(i, j)) = -mpole%ds(1)
         M1(2, indexb(i, j)) = -mpole%ds(2)
         M1(3, indexb(i, j)) = -mpole%ds(3)

         M2(1, 1, indexb(i, j)) = -mpole%qq(1, 1)/3.0_dp
         M2(2, 1, indexb(i, j)) = -mpole%qq(2, 1)/3.0_dp
         M2(3, 1, indexb(i, j)) = -mpole%qq(3, 1)/3.0_dp

         M2(1, 2, indexb(i, j)) = -mpole%qq(1, 2)/3.0_dp
         M2(2, 2, indexb(i, j)) = -mpole%qq(2, 2)/3.0_dp
         M2(3, 2, indexb(i, j)) = -mpole%qq(3, 2)/3.0_dp

         M2(1, 3, indexb(i, j)) = -mpole%qq(1, 3)/3.0_dp
         M2(2, 3, indexb(i, j)) = -mpole%qq(2, 3)/3.0_dp
         M2(3, 3, indexb(i, j)) = -mpole%qq(3, 3)/3.0_dp
      ENDDO
      IF (PRESENT(ACOUL)) ACOUL = sepi%acoul
   END SUBROUTINE get_se_slater_multipole

END MODULE semi_empirical_int_gks
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Set of wrappers for semi-empirical analytical/numerical Integrals
!>        routines
!> \author Teodoro Laino [tlaino] - University of Zurich
!> \date   04.2008
!> \par History
!>         05.2008 Teodoro Laino [tlaino] - University of Zurich - In core integrals
! **************************************************************************************************
MODULE semi_empirical_integrals

   USE hfx_compression_methods,         ONLY: hfx_add_mult_cache_elements,&
                                              hfx_get_mult_cache_elements
   USE input_constants,                 ONLY: do_se_IS_slater
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE memory_utilities,                ONLY: reallocate
   USE semi_empirical_int_ana,          ONLY: corecore_ana,&
                                              corecore_el_ana,&
                                              rotint_ana,&
                                              rotnuc_ana
   USE semi_empirical_int_gks,          ONLY: corecore_gks,&
                                              drotint_gks,&
                                              drotnuc_gks,&
                                              rotint_gks,&
                                              rotnuc_gks
   USE semi_empirical_int_num,          ONLY: corecore_el_num,&
                                              corecore_num,&
                                              dcorecore_el_num,&
                                              dcorecore_num,&
                                              drotint_num,&
                                              drotnuc_num,&
                                              rotint_num,&
                                              rotnuc_num
   USE semi_empirical_store_int_types,  ONLY: semi_empirical_si_type
   USE semi_empirical_types,            ONLY: se_int_control_type,&
                                              se_taper_type,&
                                              semi_empirical_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'semi_empirical_integrals'
   PUBLIC :: rotint, rotnuc, corecore, corecore_el, drotint, drotnuc, dcorecore, &
             dcorecore_el

CONTAINS

! **************************************************************************************************
!> \brief  wrapper for numerical/analytical 2 center 2 electrons integrals
!>         routines with possibility of incore storage/compression
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param w ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param store_int_env ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE rotint(sepi, sepj, rij, w, anag, se_int_control, se_taper, store_int_env)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), DIMENSION(2025), INTENT(OUT)             :: w
      LOGICAL                                            :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env

      CHARACTER(len=*), PARAMETER :: routineN = 'rotint', routineP = moduleN//':'//routineN

      INTEGER                                            :: buffer_left, buffer_size, buffer_start, &
                                                            cache_size, memory_usage, nbits, &
                                                            new_size, nints
      INTEGER(KIND=int_8)                                :: mem_compression_counter
      LOGICAL                                            :: buffer_overflow
      REAL(KIND=dp)                                      :: eps_storage

      w(:) = 0.0_dp
      IF (.NOT. store_int_env%memory_parameter%do_all_on_the_fly) THEN
         nints = (sepi%natorb*(sepi%natorb+1)/2)*(sepj%natorb*(sepj%natorb+1)/2)
         cache_size = store_int_env%memory_parameter%cache_size
         eps_storage = store_int_env%memory_parameter%eps_storage_scaling
         IF (store_int_env%filling_containers) THEN
            mem_compression_counter = store_int_env%memory_parameter%actual_memory_usage*cache_size
            IF (mem_compression_counter > store_int_env%memory_parameter%max_compression_counter) THEN
               buffer_overflow = .TRUE.
               store_int_env%memory_parameter%ram_counter = store_int_env%nbuffer
            ELSE
               store_int_env%nbuffer = store_int_env%nbuffer+1
               buffer_overflow = .FALSE.
            END IF
            ! Compute Integrals
            IF (se_int_control%integral_screening == do_se_IS_slater) THEN
               CALL rotint_gks(sepi, sepj, rij, w, se_int_control=se_int_control)
            ELSE
               IF (anag) THEN
                  CALL rotint_ana(sepi, sepj, rij, w, se_int_control=se_int_control, se_taper=se_taper)
               ELSE
                  CALL rotint_num(sepi, sepj, rij, w, se_int_control=se_int_control, se_taper=se_taper)
               END IF
            END IF
            ! Store integrals if we did not go overflow
            IF (.NOT. buffer_overflow) THEN
               IF (store_int_env%compress) THEN
                  ! Store integrals in the containers
                  IF (store_int_env%nbuffer > SIZE(store_int_env%max_val_buffer)) THEN
                     new_size = store_int_env%nbuffer+1000
                     CALL reallocate(store_int_env%max_val_buffer, 1, new_size)
                  END IF
                  store_int_env%max_val_buffer(store_int_env%nbuffer) = MAXVAL(ABS(w(1:nints)))

                  nbits = EXPONENT(store_int_env%max_val_buffer(store_int_env%nbuffer)/eps_storage)+1
                  buffer_left = nints
                  buffer_start = 1
                  DO WHILE (buffer_left > 0)
                     buffer_size = MIN(buffer_left, cache_size)
                     CALL hfx_add_mult_cache_elements(w(buffer_start:), &
                                                      buffer_size, nbits, &
                                                      store_int_env%integral_caches(nbits), &
                                                      store_int_env%integral_containers(nbits), &
                                                      eps_storage, 1.0_dp, &
                                                      store_int_env%memory_parameter%actual_memory_usage, &
                                                      .FALSE.)
                     buffer_left = buffer_left-buffer_size
                     buffer_start = buffer_start+buffer_size
                  END DO
               ELSE
                  ! Skip compression
                  memory_usage = store_int_env%memory_parameter%actual_memory_usage
                  IF(.NOT.((nints/1.2_dp) <= HUGE(0)-memory_usage))CALL cp__a("semi_empirical_integrals.F",139)
                  IF (memory_usage+nints > SIZE(store_int_env%uncompressed_container)) THEN
                     new_size = INT((memory_usage+nints)*1.2_dp)
                     CALL reallocate(store_int_env%uncompressed_container, 1, new_size)
                  END IF
                  store_int_env%uncompressed_container(memory_usage:memory_usage+nints-1) = w(1:nints)
                  store_int_env%memory_parameter%actual_memory_usage = memory_usage+nints
               END IF
            END IF
         ELSE
            ! Get integrals from the containers
            IF (store_int_env%memory_parameter%ram_counter == store_int_env%nbuffer) THEN
               buffer_overflow = .TRUE.
            ELSE
               store_int_env%nbuffer = store_int_env%nbuffer+1
               buffer_overflow = .FALSE.
            END IF
            ! Get integrals from cache unless we overflowed
            IF (.NOT. buffer_overflow) THEN
               IF (store_int_env%compress) THEN
                  ! Get Integrals from containers
                  nbits = EXPONENT(store_int_env%max_val_buffer(store_int_env%nbuffer)/eps_storage)+1
                  buffer_left = nints
                  buffer_start = 1
                  DO WHILE (buffer_left > 0)
                     buffer_size = MIN(buffer_left, cache_size)
                     CALL hfx_get_mult_cache_elements(w(buffer_start:), &
                                                      buffer_size, nbits, &
                                                      store_int_env%integral_caches(nbits), &
                                                      store_int_env%integral_containers(nbits), &
                                                      eps_storage, 1.0_dp, &
                                                      store_int_env%memory_parameter%actual_memory_usage, &
                                                      .FALSE.)
                     buffer_left = buffer_left-buffer_size
                     buffer_start = buffer_start+buffer_size
                  END DO
               ELSE
                  ! Skip compression
                  memory_usage = store_int_env%memory_parameter%actual_memory_usage
                  w(1:nints) = store_int_env%uncompressed_container(memory_usage:memory_usage+nints-1)
                  store_int_env%memory_parameter%actual_memory_usage = memory_usage+nints
               END IF
            ELSE
               IF (se_int_control%integral_screening == do_se_IS_slater) THEN
                  CALL rotint_gks(sepi, sepj, rij, w, se_int_control=se_int_control)
               ELSE
                  IF (anag) THEN
                     CALL rotint_ana(sepi, sepj, rij, w, se_int_control=se_int_control, se_taper=se_taper)
                  ELSE
                     CALL rotint_num(sepi, sepj, rij, w, se_int_control=se_int_control, se_taper=se_taper)
                  END IF
               END IF
            END IF
         END IF
      ELSE
         IF (se_int_control%integral_screening == do_se_IS_slater) THEN
            CALL rotint_gks(sepi, sepj, rij, w, se_int_control=se_int_control)
         ELSE
            IF (anag) THEN
               CALL rotint_ana(sepi, sepj, rij, w, se_int_control=se_int_control, se_taper=se_taper)
            ELSE
               CALL rotint_num(sepi, sepj, rij, w, se_int_control=se_int_control, se_taper=se_taper)
            END IF
         END IF
      END IF
   END SUBROUTINE rotint

! **************************************************************************************************
!> \brief wrapper for numerical/analytical 1 center 1 electron integrals
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param e1b ...
!> \param e2a ...
!> \param itype ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param store_int_env ...
!> \date   05.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE rotnuc(sepi, sepj, rij, e1b, e2a, itype, anag, se_int_control, se_taper, store_int_env)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), DIMENSION(45), INTENT(OUT), OPTIONAL     :: e1b, e2a
      INTEGER, INTENT(IN)                                :: itype
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_si_type), OPTIONAL, POINTER    :: store_int_env

      CHARACTER(len=*), PARAMETER :: routineN = 'rotnuc', routineP = moduleN//':'//routineN

      INTEGER                                            :: buffer_left, buffer_size, buffer_start, &
                                                            cache_size, memory_usage, nbits, &
                                                            new_size, nints, nints_1, nints_2
      INTEGER(KIND=int_8)                                :: mem_compression_counter
      LOGICAL                                            :: buffer_overflow, do_all_on_the_fly
      REAL(KIND=dp)                                      :: eps_storage, w(90)

      do_all_on_the_fly = .TRUE.
      IF (PRESENT(e1b)) e1b(:) = 0.0_dp
      IF (PRESENT(e2a)) e2a(:) = 0.0_dp
      IF (PRESENT(store_int_env)) do_all_on_the_fly = store_int_env%memory_parameter%do_all_on_the_fly
      IF (.NOT. do_all_on_the_fly) THEN
         nints_1 = 0
         nints_2 = 0
         IF (PRESENT(e1b)) nints_1 = (sepi%natorb*(sepi%natorb+1)/2)
         IF (PRESENT(e2a)) nints_2 = (sepj%natorb*(sepj%natorb+1)/2)
         nints = nints_1+nints_2
         ! This is the upper limit for an spd basis set
         IF(.NOT.(nints <= 90))CALL cp__a("semi_empirical_integrals.F",251)
         cache_size = store_int_env%memory_parameter%cache_size
         eps_storage = store_int_env%memory_parameter%eps_storage_scaling
         IF (store_int_env%filling_containers) THEN
            mem_compression_counter = store_int_env%memory_parameter%actual_memory_usage*cache_size
            IF (mem_compression_counter > store_int_env%memory_parameter%max_compression_counter) THEN
               buffer_overflow = .TRUE.
               store_int_env%memory_parameter%ram_counter = store_int_env%nbuffer
            ELSE
               store_int_env%nbuffer = store_int_env%nbuffer+1
               buffer_overflow = .FALSE.
            END IF
            ! Compute Integrals
            IF (se_int_control%integral_screening == do_se_IS_slater) THEN
               CALL rotnuc_gks(sepi, sepj, rij, e1b=e1b, e2a=e2a, &
                               se_int_control=se_int_control)
            ELSE
               IF (anag) THEN
                  CALL rotnuc_ana(sepi, sepj, rij, e1b=e1b, e2a=e2a, itype=itype, &
                                  se_int_control=se_int_control, se_taper=se_taper)
               ELSE
                  CALL rotnuc_num(sepi, sepj, rij, e1b=e1b, e2a=e2a, itype=itype, &
                                  se_int_control=se_int_control, se_taper=se_taper)
               END IF
            END IF
            ! Store integrals if we did not go overflow
            IF (.NOT. buffer_overflow) THEN
               IF (PRESENT(e1b)) w(1:nints_1) = e1b(1:nints_1)
               IF (PRESENT(e2a)) w(nints_1+1:nints) = e2a(1:nints_2)

               IF (store_int_env%compress) THEN
                  ! Store integrals in the containers
                  IF (store_int_env%nbuffer > SIZE(store_int_env%max_val_buffer)) THEN
                     new_size = store_int_env%nbuffer+1000
                     CALL reallocate(store_int_env%max_val_buffer, 1, new_size)
                  END IF
                  store_int_env%max_val_buffer(store_int_env%nbuffer) = MAXVAL(ABS(w(1:nints)))

                  nbits = EXPONENT(store_int_env%max_val_buffer(store_int_env%nbuffer)/eps_storage)+1
                  buffer_left = nints
                  buffer_start = 1
                  DO WHILE (buffer_left > 0)
                     buffer_size = MIN(buffer_left, cache_size)
                     CALL hfx_add_mult_cache_elements(w(buffer_start:), &
                                                      buffer_size, nbits, &
                                                      store_int_env%integral_caches(nbits), &
                                                      store_int_env%integral_containers(nbits), &
                                                      eps_storage, 1.0_dp, &
                                                      store_int_env%memory_parameter%actual_memory_usage, &
                                                      .FALSE.)
                     buffer_left = buffer_left-buffer_size
                     buffer_start = buffer_start+buffer_size
                  END DO
               ELSE
                  ! Skip compression
                  memory_usage = store_int_env%memory_parameter%actual_memory_usage
                  IF(.NOT.((nints/1.2_dp) <= HUGE(0)-memory_usage))CALL cp__a("semi_empirical_integrals.F",307)
                  IF (memory_usage+nints > SIZE(store_int_env%uncompressed_container)) THEN
                     new_size = INT((memory_usage+nints)*1.2_dp)
                     CALL reallocate(store_int_env%uncompressed_container, 1, new_size)
                  END IF
                  store_int_env%uncompressed_container(memory_usage:memory_usage+nints-1) = w(1:nints)
                  store_int_env%memory_parameter%actual_memory_usage = memory_usage+nints
               END IF
            END IF
         ELSE
            ! Get integrals from the containers
            IF (store_int_env%memory_parameter%ram_counter == store_int_env%nbuffer) THEN
               buffer_overflow = .TRUE.
            ELSE
               store_int_env%nbuffer = store_int_env%nbuffer+1
               buffer_overflow = .FALSE.
            END IF
            ! Get integrals from cache unless we overflowed
            IF (.NOT. buffer_overflow) THEN
               IF (store_int_env%compress) THEN
                  ! Get Integrals from containers
                  nbits = EXPONENT(store_int_env%max_val_buffer(store_int_env%nbuffer)/eps_storage)+1
                  buffer_left = nints
                  buffer_start = 1
                  DO WHILE (buffer_left > 0)
                     buffer_size = MIN(buffer_left, cache_size)
                     CALL hfx_get_mult_cache_elements(w(buffer_start:), &
                                                      buffer_size, nbits, &
                                                      store_int_env%integral_caches(nbits), &
                                                      store_int_env%integral_containers(nbits), &
                                                      eps_storage, 1.0_dp, &
                                                      store_int_env%memory_parameter%actual_memory_usage, &
                                                      .FALSE.)
                     buffer_left = buffer_left-buffer_size
                     buffer_start = buffer_start+buffer_size
                  END DO
               ELSE
                  ! Skip compression
                  memory_usage = store_int_env%memory_parameter%actual_memory_usage
                  w(1:nints) = store_int_env%uncompressed_container(memory_usage:memory_usage+nints-1)
                  store_int_env%memory_parameter%actual_memory_usage = memory_usage+nints
               END IF
               IF (PRESENT(e1b)) e1b(1:nints_1) = w(1:nints_1)
               IF (PRESENT(e2a)) e2a(1:nints_2) = w(nints_1+1:nints)
            ELSE
               IF (se_int_control%integral_screening == do_se_IS_slater) THEN
                  CALL rotnuc_gks(sepi, sepj, rij, e1b=e1b, e2a=e2a, &
                                  se_int_control=se_int_control)
               ELSE
                  IF (anag) THEN
                     CALL rotnuc_ana(sepi, sepj, rij, e1b=e1b, e2a=e2a, itype=itype, &
                                     se_int_control=se_int_control, se_taper=se_taper)
                  ELSE
                     CALL rotnuc_num(sepi, sepj, rij, e1b=e1b, e2a=e2a, itype=itype, &
                                     se_int_control=se_int_control, se_taper=se_taper)
                  END IF
               END IF
            END IF
         END IF
      ELSE
         IF (se_int_control%integral_screening == do_se_IS_slater) THEN
            CALL rotnuc_gks(sepi, sepj, rij, e1b=e1b, e2a=e2a, &
                            se_int_control=se_int_control)
         ELSE
            IF (anag) THEN
               CALL rotnuc_ana(sepi, sepj, rij, e1b=e1b, e2a=e2a, itype=itype, &
                               se_int_control=se_int_control, se_taper=se_taper)
            ELSE
               CALL rotnuc_num(sepi, sepj, rij, e1b=e1b, e2a=e2a, itype=itype, &
                               se_int_control=se_int_control, se_taper=se_taper)
            END IF
         END IF
      END IF

   END SUBROUTINE rotnuc

! **************************************************************************************************
!> \brief  wrapper for numerical/analytical routines
!>         core-core integrals, since are evaluated only once do not need to be
!>         stored.
!>
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param enuc ...
!> \param itype ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \date   04.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE corecore(sepi, sepj, rij, enuc, itype, anag, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), INTENT(OUT)                              :: enuc
      INTEGER, INTENT(IN)                                :: itype
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'corecore', routineP = moduleN//':'//routineN

      enuc = 0.0_dp
      IF (se_int_control%integral_screening == do_se_IS_slater) THEN
         CALL corecore_gks(sepi, sepj, rij, enuc=enuc, se_int_control=se_int_control)
      ELSE
         IF (anag) THEN
            CALL corecore_ana(sepi, sepj, rij, enuc=enuc, itype=itype, se_int_control=se_int_control, &
                              se_taper=se_taper)
         ELSE
            CALL corecore_num(sepi, sepj, rij, enuc=enuc, itype=itype, se_int_control=se_int_control, &
                              se_taper=se_taper)
         END IF
      END IF

   END SUBROUTINE corecore

! **************************************************************************************************
!> \brief  wrapper for numerical/analytical routines
!>         core-core electrostatic (only) integrals
!>
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param enuc ...
!> \param itype ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \date   05.2009
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE corecore_el(sepi, sepj, rij, enuc, itype, anag, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), INTENT(OUT)                              :: enuc
      INTEGER, INTENT(IN)                                :: itype
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'corecore_el', routineP = moduleN//':'//routineN

      enuc = 0.0_dp
      IF (anag) THEN
         CALL corecore_el_ana(sepi, sepj, rij, enuc=enuc, itype=itype, se_int_control=se_int_control, &
                              se_taper=se_taper)
      ELSE
         CALL corecore_el_num(sepi, sepj, rij, enuc=enuc, itype=itype, se_int_control=se_int_control, &
                              se_taper=se_taper)
      END IF

   END SUBROUTINE corecore_el

! **************************************************************************************************
!> \brief wrapper for numerical/analytical routines
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param dw ...
!> \param delta ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \date   04.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE drotint(sepi, sepj, rij, dw, delta, anag, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), DIMENSION(3, 2025), INTENT(OUT)          :: dw
      REAL(dp), INTENT(IN)                               :: delta
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'drotint', routineP = moduleN//':'//routineN

      dw(:, :) = 0.0_dp
      IF (se_int_control%integral_screening == do_se_IS_slater) THEN
         CALL drotint_gks(sepi, sepj, rij, dw=dw, se_int_control=se_int_control)
      ELSE
         IF (anag) THEN
            CALL rotint_ana(sepi, sepj, rij, dw=dw, se_int_control=se_int_control, se_taper=se_taper)
         ELSE
            CALL drotint_num(sepi, sepj, rij, dw, delta, se_int_control=se_int_control, se_taper=se_taper)
         END IF
      END IF

   END SUBROUTINE drotint

! **************************************************************************************************
!> \brief wrapper for numerical/analytical routines
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param de1b ...
!> \param de2a ...
!> \param itype ...
!> \param delta ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \date   04.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE drotnuc(sepi, sepj, rij, de1b, de2a, itype, delta, anag, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), DIMENSION(3, 45), INTENT(OUT), OPTIONAL  :: de1b, de2a
      INTEGER, INTENT(IN)                                :: itype
      REAL(dp), INTENT(IN)                               :: delta
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'drotnuc', routineP = moduleN//':'//routineN

      IF (PRESENT(de1b)) de1b(:, :) = 0.0_dp
      IF (PRESENT(de2a)) de2a(:, :) = 0.0_dp
      IF (se_int_control%integral_screening == do_se_IS_slater) THEN
         CALL drotnuc_gks(sepi, sepj, rij, de1b=de1b, de2a=de2a, &
                          se_int_control=se_int_control)
      ELSE
         IF (anag) THEN
            CALL rotnuc_ana(sepi, sepj, rij, de1b=de1b, de2a=de2a, itype=itype, &
                            se_int_control=se_int_control, se_taper=se_taper)
         ELSE
            CALL drotnuc_num(sepi, sepj, rij, de1b=de1b, de2a=de2a, itype=itype, &
                             delta=delta, se_int_control=se_int_control, se_taper=se_taper)
         END IF
      END IF

   END SUBROUTINE drotnuc

! **************************************************************************************************
!> \brief wrapper for numerical/analytical routines
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param denuc ...
!> \param itype ...
!> \param delta ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \date   04.2008
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE dcorecore(sepi, sepj, rij, denuc, itype, delta, anag, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), DIMENSION(3), INTENT(OUT)                :: denuc
      INTEGER, INTENT(IN)                                :: itype
      REAL(dp), INTENT(IN)                               :: delta
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'dcorecore', routineP = moduleN//':'//routineN

      denuc = 0.0_dp
      IF (se_int_control%integral_screening == do_se_IS_slater) THEN
         CALL corecore_gks(sepi, sepj, rij, denuc=denuc, se_int_control=se_int_control)
      ELSE
         IF (anag) THEN
            CALL corecore_ana(sepi, sepj, rij, denuc=denuc, itype=itype, se_int_control=se_int_control, &
                              se_taper=se_taper)
         ELSE
            CALL dcorecore_num(sepi, sepj, rij, denuc=denuc, delta=delta, itype=itype, &
                               se_int_control=se_int_control, se_taper=se_taper)
         END IF
      END IF

   END SUBROUTINE dcorecore

! **************************************************************************************************
!> \brief  wrapper for numerical/analytical routines
!>         core-core electrostatic (only) integrals derivatives
!>
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param denuc ...
!> \param itype ...
!> \param delta ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \date   05.2009
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE dcorecore_el(sepi, sepj, rij, denuc, itype, delta, anag, se_int_control, se_taper)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(dp), DIMENSION(3), INTENT(IN)                 :: rij
      REAL(dp), DIMENSION(3), INTENT(OUT)                :: denuc
      INTEGER, INTENT(IN)                                :: itype
      REAL(dp), INTENT(IN)                               :: delta
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper

      CHARACTER(len=*), PARAMETER :: routineN = 'dcorecore_el', routineP = moduleN//':'//routineN

      denuc = 0.0_dp
      IF (anag) THEN
         CALL corecore_el_ana(sepi, sepj, rij, denuc=denuc, itype=itype, se_int_control=se_int_control, &
                              se_taper=se_taper)
      ELSE
         CALL dcorecore_el_num(sepi, sepj, rij, denuc=denuc, delta=delta, itype=itype, &
                               se_int_control=se_int_control, se_taper=se_taper)
      END IF

   END SUBROUTINE dcorecore_el

END MODULE semi_empirical_integrals
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Provides the low level routines to build both the exchange and the
!>        Coulomb Fock matrices.. This routines support d-orbitals and should
!>        be changed only if one knows exactly what he is doing..
!> \author Teodoro Laino [tlaino] (05.2009) - Split and module reorganization
!> \par History
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : d-orbitals
!>      Teodoro Laino (09.2008) [tlaino] - University of Zurich : Speed-up
!>      Teodoro Laino (09.2008) [tlaino] - University of Zurich : Periodic SE
! **************************************************************************************************
MODULE se_fock_matrix_integrals

   USE kinds,                           ONLY: dp
   USE semi_empirical_int_arrays,       ONLY: se_orbital_pointer
   USE semi_empirical_integrals,        ONLY: drotint,&
                                              drotnuc,&
                                              rotint,&
                                              rotnuc
   USE semi_empirical_store_int_types,  ONLY: semi_empirical_si_type
   USE semi_empirical_types,            ONLY: se_int_control_type,&
                                              se_taper_type,&
                                              semi_empirical_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'se_fock_matrix_integrals'
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.

   PUBLIC :: fock2_1el, dfock2_1el, fock1_2el, fock2_1el_ew, fock2C_ew, &
             fock2C, dfock2C, fock2E, dfock2E, fock2_1el_r3, dfock2_1el_r3, &
             fock2C_r3, dfock2C_r3, se_coulomb_ij_interaction

CONTAINS

! **************************************************************************************************
!> \brief  Construction of 2-center 1-electron Fock Matrix
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param ksi_block DIMENSION(sepi%natorb, sepi%natorb)
!> \param ksj_block DIMENSION(sepi%natorb, sepi%natorb)
!> \param pi_block ...
!> \param pj_block ...
!> \param ecore ...
!> \param itype ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param store_int_env ...
!> \date   04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE fock2_1el(sepi, sepj, rij, ksi_block, ksj_block, pi_block, pj_block, &
                        ecore, itype, anag, se_int_control, se_taper, store_int_env)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rij
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ksi_block, ksj_block
      REAL(KIND=dp), &
         DIMENSION(sepi%natorb, sepi%natorb), INTENT(IN) :: pi_block
      REAL(KIND=dp), &
         DIMENSION(sepj%natorb, sepj%natorb), INTENT(IN) :: pj_block
      REAL(KIND=dp), DIMENSION(2), INTENT(INOUT)         :: ecore
      INTEGER, INTENT(IN)                                :: itype
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fock2_1el', routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i1L, i2, j1, j1L
      REAL(KIND=dp), DIMENSION(45)                       :: e1b, e2a

! Compute integrals

      CALL rotnuc(sepi, sepj, rij, e1b=e1b, e2a=e2a, itype=itype, anag=anag, &
                  se_int_control=se_int_control, se_taper=se_taper, store_int_env=store_int_env)
      !
      ! Add the electron-nuclear attraction term for atom sepi
      !
      i2 = 0
      DO i1L = 1, sepi%natorb
         i1 = se_orbital_pointer(i1L)
         DO j1L = 1, i1L-1
            j1 = se_orbital_pointer(j1L)
            i2 = i2+1
            ksi_block(i1, j1) = ksi_block(i1, j1)+e1b(i2)
            ksi_block(j1, i1) = ksi_block(i1, j1)
            ecore(1) = ecore(1)+2.0_dp*e1b(i2)*pi_block(i1, j1)
         END DO
         j1 = se_orbital_pointer(j1L)
         i2 = i2+1
         ksi_block(i1, j1) = ksi_block(i1, j1)+e1b(i2)
         ecore(1) = ecore(1)+e1b(i2)*pi_block(i1, j1)
      END DO
      !
      ! Add the electron-nuclear attraction term for atom sepj
      !
      i2 = 0
      DO i1L = 1, sepj%natorb
         i1 = se_orbital_pointer(i1L)
         DO j1L = 1, i1L-1
            j1 = se_orbital_pointer(j1L)
            i2 = i2+1
            ksj_block(i1, j1) = ksj_block(i1, j1)+e2a(i2)
            ksj_block(j1, i1) = ksj_block(i1, j1)
            ecore(2) = ecore(2)+2.0_dp*e2a(i2)*pj_block(i1, j1)
         END DO
         j1 = se_orbital_pointer(j1L)
         i2 = i2+1
         ksj_block(i1, j1) = ksj_block(i1, j1)+e2a(i2)
         ecore(2) = ecore(2)+e2a(i2)*pj_block(i1, j1)
      END DO

   END SUBROUTINE fock2_1el

! **************************************************************************************************
!> \brief Derivatives of 2-center 1-electron Fock Matrix
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param pi_block ...
!> \param pj_block ...
!> \param itype ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param force ...
!> \param delta ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE dfock2_1el(sepi, sepj, rij, pi_block, pj_block, itype, anag, &
                         se_int_control, se_taper, force, delta)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rij
      REAL(KIND=dp), &
         DIMENSION(sepi%natorb, sepi%natorb), INTENT(IN) :: pi_block
      REAL(KIND=dp), &
         DIMENSION(sepj%natorb, sepj%natorb), INTENT(IN) :: pj_block
      INTEGER, INTENT(IN)                                :: itype
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: force
      REAL(KIND=dp), INTENT(IN)                          :: delta

      CHARACTER(len=*), PARAMETER :: routineN = 'dfock2_1el', routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i1L, i2, j1, j1L
      REAL(KIND=dp)                                      :: tmp
      REAL(KIND=dp), DIMENSION(3, 45)                    :: de1b, de2a

! Compute integrals

      CALL drotnuc(sepi, sepj, rij, de1b=de1b, de2a=de2a, itype=itype, anag=anag, &
                   se_int_control=se_int_control, se_taper=se_taper, delta=delta)
      !
      ! Add the electron-nuclear attraction term for atom sepi
      !
      i2 = 0
      DO i1L = 1, sepi%natorb
         i1 = se_orbital_pointer(i1L)
         DO j1L = 1, i1L-1
            j1 = se_orbital_pointer(j1L)
            i2 = i2+1
            tmp = 2.0_dp*pi_block(i1, j1)
            force(1) = force(1)+de1b(1, i2)*tmp
            force(2) = force(2)+de1b(2, i2)*tmp
            force(3) = force(3)+de1b(3, i2)*tmp
         END DO
         j1 = se_orbital_pointer(j1L)
         i2 = i2+1
         force(1) = force(1)+de1b(1, i2)*pi_block(i1, j1)
         force(2) = force(2)+de1b(2, i2)*pi_block(i1, j1)
         force(3) = force(3)+de1b(3, i2)*pi_block(i1, j1)
      END DO
      !
      ! Add the electron-nuclear attraction term for atom sepj
      !
      i2 = 0
      DO i1L = 1, sepj%natorb
         i1 = se_orbital_pointer(i1L)
         DO j1L = 1, i1L-1
            j1 = se_orbital_pointer(j1L)
            i2 = i2+1
            tmp = 2.0_dp*pj_block(i1, j1)
            force(1) = force(1)+de2a(1, i2)*tmp
            force(2) = force(2)+de2a(2, i2)*tmp
            force(3) = force(3)+de2a(3, i2)*tmp
         END DO
         j1 = se_orbital_pointer(j1L)
         i2 = i2+1
         force(1) = force(1)+de2a(1, i2)*pj_block(i1, j1)
         force(2) = force(2)+de2a(2, i2)*pj_block(i1, j1)
         force(3) = force(3)+de2a(3, i2)*pj_block(i1, j1)
      END DO

   END SUBROUTINE dfock2_1el

! **************************************************************************************************
!> \brief Construction of 1-center 2-electron Fock Matrix
!> \param sep ...
!> \param p_tot ...
!> \param p_mat ...
!> \param f_mat DIMENSION(sep%natorb, sep%natorb)
!> \param factor ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE fock1_2el(sep, p_tot, p_mat, f_mat, factor)
      TYPE(semi_empirical_type), POINTER                 :: sep
      REAL(KIND=dp), DIMENSION(45, 45), INTENT(IN)       :: p_tot
      REAL(KIND=dp), DIMENSION(sep%natorb, sep%natorb), &
         INTENT(IN)                                      :: p_mat
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: f_mat
      REAL(KIND=dp), INTENT(IN)                          :: factor

      CHARACTER(len=*), PARAMETER :: routineN = 'fock1_2el', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ijw, ikw, iL, im, j, jL, jlw, jm, k, &
                                                            kL, klw, l, lL
      REAL(KIND=dp)                                      :: sum

!   One-center coulomb and exchange terms for semiempirical_type sep
!
!  F(i,j)=F(i,j)+sum(k,l)((PA(k,l)+PB(k,l))*<i,j|k,l>
!                        -(PA(k,l)        )*<i,k|j,l>), k,l on type sep.
!

      DO iL = 1, sep%natorb
         i = se_orbital_pointer(iL)
         DO jL = 1, iL
            j = se_orbital_pointer(jL)

            !    `J' Address IJ in W
            ijw = (iL*(iL-1))/2+jL
            sum = 0.0_dp
            DO kL = 1, sep%natorb
               k = se_orbital_pointer(kL)
               DO lL = 1, sep%natorb
                  l = se_orbital_pointer(lL)

                  !    `J' Address KL in W
                  im = MAX(kL, lL)
                  jm = MIN(kL, lL)
                  klw = (im*(im-1))/2+jm

                  !    `K' Address IK in W
                  im = MAX(kL, jL)
                  jm = MIN(kL, jL)
                  ikw = (im*(im-1))/2+jm

                  !    `K' Address JL in W
                  im = MAX(lL, iL)
                  jm = MIN(lL, iL)
                  jlw = (im*(im-1))/2+jm

                  sum = sum+p_tot(k, l)*sep%w(ijw, klw)-p_mat(k, l)*sep%w(ikw, jlw)
               END DO
            END DO
            f_mat(i, j) = f_mat(i, j)+factor*sum
            f_mat(j, i) = f_mat(i, j)
         END DO
      END DO
   END SUBROUTINE fock1_2el

! **************************************************************************************************
!> \brief Construction of 2-center 1-electron Fock Matrix (Ewald self term)
!> \param sep ...
!> \param rij ...
!> \param ks_block DIMENSION(sep%natorb, sep%natorb)
!> \param p_block ...
!> \param ecore ...
!> \param itype ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param store_int_env ...
!> \date 04.2009 [jgh]
!> \author jgh - University of Zurich
! **************************************************************************************************
   SUBROUTINE fock2_1el_ew(sep, rij, ks_block, p_block, ecore, itype, anag, &
                           se_int_control, se_taper, store_int_env)
      TYPE(semi_empirical_type), POINTER                 :: sep
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rij
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ks_block
      REAL(KIND=dp), DIMENSION(sep%natorb, sep%natorb), &
         INTENT(IN)                                      :: p_block
      REAL(KIND=dp), INTENT(INOUT)                       :: ecore
      INTEGER, INTENT(IN)                                :: itype
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fock2_1el_ew', routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i1L, i2, j1, j1L, n
      REAL(KIND=dp), DIMENSION(45)                       :: e1b, e2a

! Compute integrals

      CALL rotnuc(sep, sep, rij, e1b=e1b, e2a=e2a, itype=itype, anag=anag, &
                  se_int_control=se_int_control, se_taper=se_taper, store_int_env=store_int_env)
      !
      ! Add the electron-nuclear attraction term for atom sep
      ! e1b == e2a
      !
      n = (sep%natorb*(sep%natorb+1))/2
      i2 = 0
      DO i1L = 1, sep%natorb
         i1 = se_orbital_pointer(i1L)
         DO j1L = 1, i1L-1
            j1 = se_orbital_pointer(j1L)
            i2 = i2+1
            ks_block(i1, j1) = ks_block(i1, j1)+e1b(i2)
            ks_block(j1, i1) = ks_block(i1, j1)
            ecore = ecore+2._dp*e1b(i2)*p_block(i1, j1)
         END DO
         ! i1L == j1L
         i2 = i2+1
         ks_block(i1, i1) = ks_block(i1, i1)+e1b(i2)
         ecore = ecore+e1b(i2)*p_block(i1, i1)
      END DO

   END SUBROUTINE fock2_1el_ew

! **************************************************************************************************
!> \brief  Construction of 2-center Fock Matrix - Coulomb Self Terms (Ewald)
!> \param sep ...
!> \param rij ...
!> \param p_tot ...
!> \param f_mat DIMENSION(sep%natorb, sep%natorb)
!> \param factor ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param store_int_env ...
!> \date 04.2009 [jgh]
!> \author jgh - University of Zurich
! **************************************************************************************************
   SUBROUTINE fock2C_ew(sep, rij, p_tot, f_mat, factor, anag, se_int_control, &
                        se_taper, store_int_env)
      TYPE(semi_empirical_type), POINTER                 :: sep
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rij
      REAL(KIND=dp), DIMENSION(45, 45), INTENT(IN)       :: p_tot
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: f_mat
      REAL(KIND=dp), INTENT(IN)                          :: factor
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fock2C_ew', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iL, j, jL, k, kL, kr, l, lL, natorb
      REAL(KIND=dp)                                      :: a, aa, bb
      REAL(KIND=dp), DIMENSION(2025)                     :: w

! Evaluate integrals

      CALL rotint(sep, sep, rij, w, anag=anag, se_int_control=se_int_control, &
                  se_taper=se_taper, store_int_env=store_int_env)
      kr = 0
      natorb = sep%natorb
      DO iL = 1, natorb
         i = se_orbital_pointer(iL)
         aa = 2.0_dp
         DO jL = 1, iL
            j = se_orbital_pointer(jL)
            IF (i == j) THEN
               aa = 1.0_dp
            END IF
            DO kL = 1, natorb
               k = se_orbital_pointer(kL)
               bb = 2.0_dp
               DO lL = 1, kL
                  l = se_orbital_pointer(lL)
                  IF (k == l) THEN
                     bb = 1.0_dp
                  END IF
                  kr = kr+1
                  a = 0.5_dp*w(kr)*factor
                  ! Coulomb
                  f_mat(i, j) = f_mat(i, j)+bb*a*p_tot(k, l)
                  f_mat(k, l) = f_mat(k, l)+aa*a*p_tot(i, j)
                  f_mat(j, i) = f_mat(i, j)
                  f_mat(l, k) = f_mat(k, l)
               END DO
            END DO
         END DO
      END DO

   END SUBROUTINE fock2C_ew

! **************************************************************************************************
!> \brief  Construction of 2-center Fock Matrix - Coulomb Terms
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param switch ...
!> \param pi_tot ...
!> \param fi_mat DIMENSION(sepi%natorb, sepi%natorb)
!> \param pj_tot DIMENSION(sepj%natorb, sepj%natorb)
!> \param fj_mat ...
!> \param factor ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param store_int_env ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE fock2C(sepi, sepj, rij, switch, pi_tot, fi_mat, pj_tot, fj_mat, &
                     factor, anag, se_int_control, se_taper, store_int_env)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rij
      LOGICAL, INTENT(IN)                                :: switch
      REAL(KIND=dp), DIMENSION(45, 45), INTENT(IN)       :: pi_tot
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: fi_mat
      REAL(KIND=dp), DIMENSION(45, 45), INTENT(IN)       :: pj_tot
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: fj_mat
      REAL(KIND=dp), INTENT(IN)                          :: factor
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fock2C', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iL, j, jL, k, kL, kr, l, lL, natorb(2)
      REAL(KIND=dp)                                      :: a, aa, bb, irij(3)
      REAL(KIND=dp), DIMENSION(2025)                     :: w

! Evaluate integrals

      IF (.NOT. switch) THEN
         CALL rotint(sepi, sepj, rij, w, anag=anag, se_int_control=se_int_control, &
                     se_taper=se_taper, store_int_env=store_int_env)
      ELSE
         irij = -rij
         CALL rotint(sepj, sepi, irij, w, anag=anag, se_int_control=se_int_control, &
                     se_taper=se_taper, store_int_env=store_int_env)
      END IF
      kr = 0
      natorb(1) = sepi%natorb
      natorb(2) = sepj%natorb
      IF (switch) THEN
         natorb(1) = sepj%natorb
         natorb(2) = sepi%natorb
      END IF
      DO iL = 1, natorb(1)
         i = se_orbital_pointer(iL)
         aa = 2.0_dp
         DO jL = 1, iL
            j = se_orbital_pointer(jL)
            IF (i == j) THEN
               aa = 1.0_dp
            END IF
            DO kL = 1, natorb(2)
               k = se_orbital_pointer(kL)
               bb = 2.0_dp
               DO lL = 1, kL
                  l = se_orbital_pointer(lL)
                  IF (k == l) THEN
                     bb = 1.0_dp
                  END IF
                  kr = kr+1
                  a = w(kr)*factor
                  ! Coulomb
                  IF (.NOT. switch) THEN
                     fi_mat(i, j) = fi_mat(i, j)+bb*a*pj_tot(k, l)
                     fj_mat(k, l) = fj_mat(k, l)+aa*a*pi_tot(i, j)
                     fi_mat(j, i) = fi_mat(i, j)
                     fj_mat(l, k) = fj_mat(k, l)
                  ELSE
                     fj_mat(i, j) = fj_mat(i, j)+bb*a*pi_tot(k, l)
                     fi_mat(k, l) = fi_mat(k, l)+aa*a*pj_tot(i, j)
                     fj_mat(j, i) = fj_mat(i, j)
                     fi_mat(l, k) = fi_mat(k, l)
                  END IF
               END DO
            END DO
         END DO
      END DO

   END SUBROUTINE fock2C

! **************************************************************************************************
!> \brief Derivatives of 2-center Fock Matrix - Coulomb Terms
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param switch ...
!> \param pi_tot ...
!> \param pj_tot ...
!> \param factor ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param force ...
!> \param delta ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE dfock2C(sepi, sepj, rij, switch, pi_tot, pj_tot, factor, anag, &
                      se_int_control, se_taper, force, delta)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rij
      LOGICAL, INTENT(IN)                                :: switch
      REAL(KIND=dp), DIMENSION(45, 45), INTENT(IN)       :: pi_tot, pj_tot
      REAL(KIND=dp), INTENT(IN)                          :: factor
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: force
      REAL(KIND=dp), INTENT(IN)                          :: delta

      CHARACTER(len=*), PARAMETER :: routineN = 'dfock2C', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iL, j, jL, k, kL, kr, l, lL, natorb(2)
      REAL(KIND=dp)                                      :: aa, bb, tmp
      REAL(KIND=dp), DIMENSION(3)                        :: a, irij
      REAL(KIND=dp), DIMENSION(3, 2025)                  :: dw

! Evaluate integrals' derivatives

      IF (.NOT. switch) THEN
         CALL drotint(sepi, sepj, rij, dw, delta, anag=anag, se_int_control=se_int_control, &
                      se_taper=se_taper)
      ELSE
         irij = -rij
         CALL drotint(sepj, sepi, irij, dw, delta, anag=anag, se_int_control=se_int_control, &
                      se_taper=se_taper)
      END IF

      kr = 0
      natorb(1) = sepi%natorb
      natorb(2) = sepj%natorb
      IF (switch) THEN
         natorb(1) = sepj%natorb
         natorb(2) = sepi%natorb
      END IF
      DO iL = 1, natorb(1)
         i = se_orbital_pointer(iL)
         aa = 2.0_dp
         DO jL = 1, iL
            j = se_orbital_pointer(jL)
            IF (i == j) THEN
               aa = 1.0_dp
            END IF
            DO kL = 1, natorb(2)
               k = se_orbital_pointer(kL)
               bb = 2.0_dp
               DO lL = 1, kL
                  l = se_orbital_pointer(lL)
                  IF (k == l) THEN
                     bb = 1.0_dp
                  END IF
                  kr = kr+1
                  a(1) = dw(1, kr)*factor
                  a(2) = dw(2, kr)*factor
                  a(3) = dw(3, kr)*factor
                  ! Coulomb
                  IF (.NOT. switch) THEN
                     tmp = bb*aa*pj_tot(k, l)*pi_tot(i, j)
                  ELSE
                     tmp = bb*aa*pi_tot(k, l)*pj_tot(i, j)
                  END IF
                  force(1) = force(1)+a(1)*tmp
                  force(2) = force(2)+a(2)*tmp
                  force(3) = force(3)+a(3)*tmp
               END DO
            END DO
         END DO
      END DO
   END SUBROUTINE dfock2C

! **************************************************************************************************
!> \brief Construction of 2-center Fock Matrix - General Driver
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param switch ...
!> \param isize ...
!> \param pi_mat ...
!> \param fi_mat DIMENSION(isize(1), isize(2))
!> \param factor ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param store_int_env ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE fock2E(sepi, sepj, rij, switch, isize, pi_mat, fi_mat, factor, &
                     anag, se_int_control, se_taper, store_int_env)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rij
      LOGICAL, INTENT(IN)                                :: switch
      INTEGER, DIMENSION(2), INTENT(IN)                  :: isize
      REAL(KIND=dp), DIMENSION(isize(1), isize(2)), &
         INTENT(IN)                                      :: pi_mat
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: fi_mat
      REAL(KIND=dp), INTENT(IN)                          :: factor
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fock2E', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iL, j, jL, k, kL, kr, l, lL, natorb(2)
      REAL(KIND=dp)                                      :: a, aa, bb, irij(3)
      REAL(KIND=dp), DIMENSION(2025)                     :: w

! Evaluate integrals

      IF (.NOT. switch) THEN
         CALL rotint(sepi, sepj, rij, w, anag=anag, se_int_control=se_int_control, &
                     se_taper=se_taper, store_int_env=store_int_env)
      ELSE
         irij = -rij
         CALL rotint(sepj, sepi, irij, w, anag=anag, se_int_control=se_int_control, &
                     se_taper=se_taper, store_int_env=store_int_env)
      END IF
      kr = 0
      natorb(1) = sepi%natorb
      natorb(2) = sepj%natorb
      IF (switch) THEN
         natorb(1) = sepj%natorb
         natorb(2) = sepi%natorb
      END IF
      DO iL = 1, natorb(1)
         i = se_orbital_pointer(iL)
         aa = 2.0_dp
         DO jL = 1, iL
            j = se_orbital_pointer(jL)
            IF (i == j) THEN
               aa = 1.0_dp
            END IF
            DO kL = 1, natorb(2)
               k = se_orbital_pointer(kL)
               bb = 2.0_dp
               DO lL = 1, kL
                  l = se_orbital_pointer(lL)
                  IF (k == l) THEN
                     bb = 1.0_dp
                  END IF
                  kr = kr+1
                  a = w(kr)*factor
                  ! Exchange
                  a = a*aa*bb*0.25_dp
                  fi_mat(i, k) = fi_mat(i, k)-a*pi_mat(j, l)
                  fi_mat(i, l) = fi_mat(i, l)-a*pi_mat(j, k)
                  fi_mat(j, k) = fi_mat(j, k)-a*pi_mat(i, l)
                  fi_mat(j, l) = fi_mat(j, l)-a*pi_mat(i, k)
               END DO
            END DO
         END DO
      END DO

   END SUBROUTINE fock2E

! **************************************************************************************************
!> \brief Derivatives of 2-center Fock Matrix - General Driver
!> \param sepi ...
!> \param sepj ...
!> \param rij ...
!> \param switch ...
!> \param isize ...
!> \param pi_mat ...
!> \param factor ...
!> \param anag ...
!> \param se_int_control ...
!> \param se_taper ...
!> \param force ...
!> \param delta ...
!> \date 04.2008 [tlaino]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE dfock2E(sepi, sepj, rij, switch, isize, pi_mat, factor, anag, &
                      se_int_control, se_taper, force, delta)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rij
      LOGICAL, INTENT(IN)                                :: switch
      INTEGER, DIMENSION(2), INTENT(IN)                  :: isize
      REAL(KIND=dp), DIMENSION(isize(1), isize(2)), &
         INTENT(IN)                                      :: pi_mat
      REAL(KIND=dp), INTENT(IN)                          :: factor
      LOGICAL, INTENT(IN)                                :: anag
      TYPE(se_int_control_type), INTENT(IN)              :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: force
      REAL(KIND=dp), INTENT(IN)                          :: delta

      CHARACTER(len=*), PARAMETER :: routineN = 'dfock2E', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iL, j, jL, k, kL, kr, l, lL, natorb(2)
      REAL(KIND=dp)                                      :: aa, bb, tmp, tmp1, tmp2, tmp3, tmp4
      REAL(KIND=dp), DIMENSION(3)                        :: a, irij
      REAL(KIND=dp), DIMENSION(3, 2025)                  :: dw

! Evaluate integrals' derivatives

      IF (.NOT. switch) THEN
         CALL drotint(sepi, sepj, rij, dw, delta, anag=anag, se_int_control=se_int_control, &
                      se_taper=se_taper)
      ELSE
         irij = -rij
         CALL drotint(sepj, sepi, irij, dw, delta, anag=anag, se_int_control=se_int_control, &
                      se_taper=se_taper)
      END IF

      kr = 0
      natorb(1) = sepi%natorb
      natorb(2) = sepj%natorb
      IF (switch) THEN
         natorb(1) = sepj%natorb
         natorb(2) = sepi%natorb
      END IF
      DO iL = 1, natorb(1)
         i = se_orbital_pointer(iL)
         aa = 2.0_dp
         DO jL = 1, iL
            j = se_orbital_pointer(jL)
            IF (i == j) THEN
               aa = 1.0_dp
            END IF
            DO kL = 1, natorb(2)
               k = se_orbital_pointer(kL)
               bb = 2.0_dp
               DO lL = 1, kL
                  l = se_orbital_pointer(lL)
                  IF (k == l) THEN
                     bb = 1.0_dp
                  END IF
                  kr = kr+1
                  tmp = factor*aa*bb*0.25_dp
                  a(1) = dw(1, kr)*tmp
                  a(2) = dw(2, kr)*tmp
                  a(3) = dw(3, kr)*tmp
                  ! Exchange
                  tmp1 = pi_mat(j, l)*pi_mat(i, k)
                  tmp2 = pi_mat(j, k)*pi_mat(i, l)
                  tmp3 = pi_mat(i, l)*pi_mat(j, k)
                  tmp4 = pi_mat(i, k)*pi_mat(j, l)

                  force(1) = force(1)-a(1)*tmp1
                  force(1) = force(1)-a(1)*tmp2
                  force(1) = force(1)-a(1)*tmp3
                  force(1) = force(1)-a(1)*tmp4

                  force(2) = force(2)-a(2)*tmp1
                  force(2) = force(2)-a(2)*tmp2
                  force(2) = force(2)-a(2)*tmp3
                  force(2) = force(2)-a(2)*tmp4

                  force(3) = force(3)-a(3)*tmp1
                  force(3) = force(3)-a(3)*tmp2
                  force(3) = force(3)-a(3)*tmp3
                  force(3) = force(3)-a(3)*tmp4
               END DO
            END DO
         END DO
      END DO
   END SUBROUTINE dfock2E

! **************************************************************************************************
!> \brief  Construction of 2-center 1-electron Fock Matrix for the residual
!>         (1/R^3) integral part
!> \param sepi ...
!> \param sepj ...
!> \param ksi_block DIMENSION(sepi%natorb, sepi%natorb)
!> \param ksj_block DIMENSION(sepj%natorb, sepj%natorb)
!> \param pi_block ...
!> \param pj_block ...
!> \param e1b ...
!> \param e2a ...
!> \param ecore ...
!> \param rp ...
!> \date   12.2008 [tlaino]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE fock2_1el_r3(sepi, sepj, ksi_block, ksj_block, pi_block, pj_block, &
                           e1b, e2a, ecore, rp)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ksi_block, ksj_block
      REAL(KIND=dp), &
         DIMENSION(sepi%natorb, sepi%natorb), INTENT(IN) :: pi_block
      REAL(KIND=dp), &
         DIMENSION(sepj%natorb, sepj%natorb), INTENT(IN) :: pj_block
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: e1b, e2a
      REAL(KIND=dp), DIMENSION(2), INTENT(INOUT)         :: ecore
      REAL(KIND=dp), INTENT(IN)                          :: rp

      CHARACTER(len=*), PARAMETER :: routineN = 'fock2_1el_r3', routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i1L, i2

!
! Add the electron-nuclear residual attraction term for atom sepi
!

      i2 = 0
      DO i1L = 1, sepi%natorb
         i2 = i2+1
         i1 = se_orbital_pointer(i1L)
         ksi_block(i1, i1) = ksi_block(i1, i1)+e1b(i2)*rp
         ecore(1) = ecore(1)+e1b(i2)*rp*pi_block(i1, i1)
      END DO
      !
      ! Add the electron-nuclear residual attraction term for atom sepj
      !
      i2 = 0
      DO i1L = 1, sepj%natorb
         i2 = i2+1
         i1 = se_orbital_pointer(i1L)
         ksj_block(i1, i1) = ksj_block(i1, i1)+e2a(i2)*rp
         ecore(2) = ecore(2)+e2a(i2)*rp*pj_block(i1, i1)
      END DO

   END SUBROUTINE fock2_1el_r3

! **************************************************************************************************
!> \brief  Derivatives of 2-center 1-electron Fock Matrix residual (1/R^3)
!>         integral part
!> \param sepi ...
!> \param sepj ...
!> \param drp ...
!> \param pi_block ...
!> \param pj_block ...
!> \param force ...
!> \param e1b ...
!> \param e2a ...
!> \date   12.2008 [tlaino]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE dfock2_1el_r3(sepi, sepj, drp, pi_block, pj_block, force, e1b, e2a)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: drp
      REAL(KIND=dp), &
         DIMENSION(sepi%natorb, sepi%natorb), INTENT(IN) :: pi_block
      REAL(KIND=dp), &
         DIMENSION(sepj%natorb, sepj%natorb), INTENT(IN) :: pj_block
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: force
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: e1b, e2a

      CHARACTER(len=*), PARAMETER :: routineN = 'dfock2_1el_r3', routineP = moduleN//':'//routineN

      INTEGER                                            :: i1, i1L, i2
      REAL(KIND=dp)                                      :: tmp

!
! Add the electron-nuclear residual attraction term for atom sepi
!

      i2 = 0
      DO i1L = 1, sepi%natorb
         i1 = se_orbital_pointer(i1L)
         i2 = i2+1
         tmp = e1b(i2)*pi_block(i1, i1)
         force(1) = force(1)+tmp*drp(1)
         force(2) = force(2)+tmp*drp(2)
         force(3) = force(3)+tmp*drp(3)
      END DO
      !
      ! Add the electron-nuclear attraction term for atom sepj
      !
      i2 = 0
      DO i1L = 1, sepj%natorb
         i1 = se_orbital_pointer(i1L)
         i2 = i2+1
         tmp = e2a(i2)*pj_block(i1, i1)
         force(1) = force(1)+tmp*drp(1)
         force(2) = force(2)+tmp*drp(2)
         force(3) = force(3)+tmp*drp(3)
      END DO

   END SUBROUTINE dfock2_1el_r3

! **************************************************************************************************
!> \brief  Construction of 2-center Fock Matrix - Coulomb Terms for the residual
!>         (1/R^3) integral part
!> \param sepi ...
!> \param sepj ...
!> \param switch ...
!> \param pi_tot ...
!> \param fi_mat DIMENSION(sepi%natorb, sepi%natorb)
!> \param pj_tot ...
!> \param fj_mat DIMENSION(sepj%natorb, sepj%natorb)
!> \param factor ...
!> \param w ...
!> \param rp ...
!> \date   12.2008 [tlaino]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE fock2C_r3(sepi, sepj, switch, pi_tot, fi_mat, pj_tot, fj_mat, &
                        factor, w, rp)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      LOGICAL, INTENT(IN)                                :: switch
      REAL(KIND=dp), DIMENSION(45, 45), INTENT(IN)       :: pi_tot
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: fi_mat
      REAL(KIND=dp), DIMENSION(45, 45), INTENT(IN)       :: pj_tot
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: fj_mat
      REAL(KIND=dp), INTENT(IN)                          :: factor
      REAL(KIND=dp), DIMENSION(81), INTENT(IN)           :: w
      REAL(KIND=dp), INTENT(IN)                          :: rp

      CHARACTER(len=*), PARAMETER :: routineN = 'fock2C_r3', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iL, ind, j, k, kL, kr, natorb(2)
      REAL(KIND=dp)                                      :: a, w_l(81)

      natorb(1) = sepi%natorb
      natorb(2) = sepj%natorb
      IF (switch) THEN
         natorb(1) = sepj%natorb
         natorb(2) = sepi%natorb
         ! Reshuffle the integral array (natural storage order is sepi/sepj)
         kr = 0
         DO i = 1, sepj%natorb
            DO j = 1, sepi%natorb
               kr = kr+1
               ind = (j-1)*sepj%natorb+i
               w_l(kr) = w(ind)
            END DO
         END DO
      ELSE
         w_l = w
      END IF

      ! Modify the Fock Matrix
      kr = 0
      DO iL = 1, natorb(1)
         i = se_orbital_pointer(iL)
         DO kL = 1, natorb(2)
            k = se_orbital_pointer(kL)
            kr = kr+1
            a = w_l(kr)*factor*rp
            ! Coulomb
            IF (.NOT. switch) THEN
               fi_mat(i, i) = fi_mat(i, i)+a*pj_tot(k, k)
               fj_mat(k, k) = fj_mat(k, k)+a*pi_tot(i, i)
            ELSE
               fj_mat(i, i) = fj_mat(i, i)+a*pi_tot(k, k)
               fi_mat(k, k) = fi_mat(k, k)+a*pj_tot(i, i)
            END IF
         END DO
      END DO

   END SUBROUTINE fock2C_r3

! **************************************************************************************************
!> \brief  Derivatives of 2-center Fock Matrix - Coulomb Terms for the residual
!>         (1/R^3) integral part
!> \param sepi ...
!> \param sepj ...
!> \param switch ...
!> \param pi_tot ...
!> \param pj_tot ...
!> \param factor ...
!> \param w ...
!> \param drp ...
!> \param force ...
!> \date   12.2008 [tlaino]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE dfock2C_r3(sepi, sepj, switch, pi_tot, pj_tot, factor, w, drp, &
                         force)
      TYPE(semi_empirical_type), POINTER                 :: sepi, sepj
      LOGICAL, INTENT(IN)                                :: switch
      REAL(KIND=dp), DIMENSION(45, 45), INTENT(IN)       :: pi_tot, pj_tot
      REAL(KIND=dp), INTENT(IN)                          :: factor
      REAL(KIND=dp), DIMENSION(81), INTENT(IN)           :: w
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: drp
      REAL(KIND=dp), DIMENSION(3), INTENT(INOUT)         :: force

      CHARACTER(len=*), PARAMETER :: routineN = 'dfock2C_r3', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iL, ind, j, k, kL, kr, natorb(2)
      REAL(KIND=dp)                                      :: a(3), tmp, w_l(81)

      natorb(1) = sepi%natorb
      natorb(2) = sepj%natorb
      IF (switch) THEN
         natorb(1) = sepj%natorb
         natorb(2) = sepi%natorb
         ! Reshuffle the integral array (natural storage order is sepi/sepj)
         kr = 0
         DO i = 1, sepj%natorb
            DO j = 1, sepi%natorb
               kr = kr+1
               ind = (j-1)*sepj%natorb+i
               w_l(kr) = w(ind)
            END DO
         END DO
      ELSE
         w_l = w
      END IF

      ! Modify the Fock Matrix
      kr = 0
      DO iL = 1, natorb(1)
         i = se_orbital_pointer(iL)
         DO kL = 1, natorb(2)
            k = se_orbital_pointer(kL)
            kr = kr+1
            tmp = w_l(kr)*factor
            a(1) = tmp*drp(1)
            a(2) = tmp*drp(2)
            a(3) = tmp*drp(3)
            ! Coulomb
            IF (.NOT. switch) THEN
               tmp = pj_tot(k, k)*pi_tot(i, i)
            ELSE
               tmp = pi_tot(k, k)*pj_tot(i, i)
            END IF
            force(1) = force(1)+a(1)*tmp
            force(2) = force(2)+a(2)*tmp
            force(3) = force(3)+a(3)*tmp
         END DO
      END DO

   END SUBROUTINE dfock2C_r3

! **************************************************************************************************
!> \brief  Coulomb interaction multipolar correction
!> \param atom_a ...
!> \param atom_b ...
!> \param my_task ...
!> \param do_forces ...
!> \param do_efield ...
!> \param do_stress ...
!> \param charges ...
!> \param dipoles ...
!> \param quadrupoles ...
!> \param force_ab ...
!> \param efield0 ...
!> \param efield1 ...
!> \param efield2 ...
!> \param rab2 ...
!> \param rab ...
!> \param integral_value ...
!> \param ptens11 ...
!> \param ptens12 ...
!> \param ptens13 ...
!> \param ptens21 ...
!> \param ptens22 ...
!> \param ptens23 ...
!> \param ptens31 ...
!> \param ptens32 ...
!> \param ptens33 ...
!> \date   05.2009 [tlaino]
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE se_coulomb_ij_interaction(atom_a, atom_b, my_task, do_forces, do_efield, &
                                        do_stress, charges, dipoles, quadrupoles, force_ab, efield0, efield1, efield2, &
                                        rab2, rab, integral_value, ptens11, ptens12, ptens13, ptens21, ptens22, ptens23, &
                                        ptens31, ptens32, ptens33)
      INTEGER, INTENT(IN)                                :: atom_a, atom_b
      LOGICAL, DIMENSION(3)                              :: my_task
      LOGICAL, INTENT(IN)                                :: do_forces, do_efield, do_stress
      REAL(KIND=dp), DIMENSION(:), POINTER               :: charges
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: dipoles
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: quadrupoles
      REAL(KIND=dp), DIMENSION(3), INTENT(OUT)           :: force_ab
      REAL(KIND=dp), DIMENSION(:), POINTER               :: efield0
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: efield1, efield2
      REAL(KIND=dp), INTENT(IN)                          :: rab2
      REAL(KIND=dp), DIMENSION(3), INTENT(IN)            :: rab
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: integral_value
      REAL(KIND=dp), INTENT(INOUT)                       :: ptens11, ptens12, ptens13, ptens21, &
                                                            ptens22, ptens23, ptens31, ptens32, &
                                                            ptens33

      CHARACTER(len=*), PARAMETER :: routineN = 'se_coulomb_ij_interaction', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: a, b, c, d, e, i, j, k
      LOGICAL                                            :: do_efield0, do_efield1, do_efield2, &
                                                            force_eval
      LOGICAL, DIMENSION(3)                              :: do_task
      LOGICAL, DIMENSION(3, 3)                           :: task
      REAL(KIND=dp) :: ch_i, ch_j, ef0_i, ef0_j, eloc, energy, fac, fac_ij, ir, irab2, r, tij, &
         tmp, tmp1, tmp11, tmp12, tmp13, tmp2, tmp21, tmp22, tmp23, tmp31, tmp32, tmp33, tmp_ij, &
         tmp_ji
      REAL(KIND=dp), DIMENSION(0:5)                      :: f
      REAL(KIND=dp), DIMENSION(3)                        :: dp_i, dp_j, ef1_i, ef1_j, fr, tij_a
      REAL(KIND=dp), DIMENSION(3, 3)                     :: ef2_i, ef2_j, qp_i, qp_j, tij_ab
      REAL(KIND=dp), DIMENSION(3, 3, 3)                  :: tij_abc
      REAL(KIND=dp), DIMENSION(3, 3, 3, 3)               :: tij_abcd
      REAL(KIND=dp), DIMENSION(3, 3, 3, 3, 3)            :: tij_abcde

      do_task = my_task
      energy = 0.0_dp
      DO i = 1, 3
         IF (do_task(i)) THEN
            SELECT CASE (i)
            CASE (1)
               do_task(1) = (charges(atom_a) /= 0.0_dp) .OR. (charges(atom_b) /= 0.0_dp)
            CASE (2)
               do_task(2) = (ANY(dipoles(:, atom_a) /= 0.0_dp)) .OR. (ANY(dipoles(:, atom_b) /= 0.0_dp))
            CASE (3)
               do_task(3) = (ANY(quadrupoles(:, :, atom_a) /= 0.0_dp)) .OR. (ANY(quadrupoles(:, :, atom_b) /= 0.0_dp))
            END SELECT
         END IF
      END DO
      DO i = 1, 3
         DO j = i, 3
            task(j, i) = do_task(i) .AND. do_task(j)
            task(i, j) = task(j, i)
         END DO
      END DO
      do_efield0 = do_efield .AND. ASSOCIATED(efield0)
      do_efield1 = do_efield .AND. ASSOCIATED(efield1)
      do_efield2 = do_efield .AND. ASSOCIATED(efield2)

      fac_ij = 1.0_dp
      IF (atom_a == atom_b) fac_ij = 0.5_dp

                ! Compute the Short Range constribution according the task
                IF (debug_this_module) THEN
                   f         = HUGE(0.0_dp)
                   tij       = HUGE(0.0_dp)
                   tij_a     = HUGE(0.0_dp)
                   tij_ab    = HUGE(0.0_dp)
                   tij_abc   = HUGE(0.0_dp)
                   tij_abcd  = HUGE(0.0_dp)
                   tij_abcde = HUGE(0.0_dp)
                END IF
                r     = SQRT(rab2)
                irab2 = 1.0_dp/rab2
                ir    = 1.0_dp/r

                ! Compute the radial function
                ! code for point multipole without screening
                f(0)  = ir
                DO i = 1, 5
                   f(i) = irab2*f(i-1)
                END DO
                ! Compute the Tensor components
                force_eval = do_stress
                IF (task(1,1)) THEN
                   tij         = f(0)*fac_ij
                                                 force_eval = do_forces .OR.do_efield1
                END IF
                IF (task(2,2))                   force_eval = force_eval.OR.do_efield0
                IF (task(1,2).OR.force_eval) THEN
                   force_eval = do_stress
                   tij_a    = - rab*f(1)*fac_ij
                   IF (task(1,2))                force_eval = force_eval.OR. do_forces
                END IF
                IF (task(1,1))                   force_eval = force_eval.OR.do_efield2
                IF (task(3,3))                   force_eval = force_eval.OR.do_efield0
                IF (task(2,2).OR.task(3,1).OR.force_eval) THEN
                   force_eval = do_stress
                   DO b = 1,3
                      DO a = 1,3
                         tmp = rab(a)*rab(b)*fac_ij
                         tij_ab(a,b) = 3.0_dp*tmp*f(2)
                         IF (a==b) tij_ab(a,b) = tij_ab(a,b) - f(1)*fac_ij
                      END DO
                   END DO
                   IF (task(2,2).OR.task(3,1))   force_eval = force_eval.OR. do_forces
                END IF
                IF (task(2,2))                   force_eval = force_eval.OR.do_efield2
                IF (task(3,3))                   force_eval = force_eval.OR.do_efield1
                IF (task(3,2).OR.force_eval) THEN
                   force_eval = do_stress
                   DO c = 1, 3
                      DO b = 1, 3
                         DO a = 1, 3
                            tmp = rab(a)*rab(b)*rab(c)*fac_ij
                            tij_abc(a,b,c) = - 15.0_dp*tmp*f(3)
                            tmp = 3.0_dp*f(2)*fac_ij
                            IF (a==b) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(c)
                            IF (a==c) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(b)
                            IF (b==c) tij_abc(a,b,c) = tij_abc(a,b,c) + tmp*rab(a)
                         END DO
                      END DO
                   END DO
                   IF (task(3,2))                force_eval = force_eval.OR. do_forces
                END IF
                IF (task(3,3).OR.force_eval) THEN
                   force_eval = do_stress
                   DO d = 1, 3
                      DO c = 1, 3
                         DO b = 1, 3
                            DO a = 1, 3
                               tmp = rab(a)*rab(b)*rab(c)*rab(d)*fac_ij
                               tij_abcd(a,b,c,d) = 105.0_dp*tmp*f(4)
                               tmp1 = 15.0_dp*f(3)*fac_ij
                               tmp2 =  3.0_dp*f(2)*fac_ij
                               IF (a==b) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(c)*rab(d)
                                  IF (c==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (a==c) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(b)*rab(d)
                                  IF (b==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (a==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(b)*rab(c)
                               IF (b==c) THEN
                                            tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(d)
                                  IF (a==d) tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) + tmp2
                               END IF
                               IF (b==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(c)
                               IF (c==d)    tij_abcd(a,b,c,d) = tij_abcd(a,b,c,d) - tmp1*rab(a)*rab(b)
                            END DO
                         END DO
                      END DO
                   END DO
                   IF (task(3,3))                force_eval = force_eval.OR. do_forces
                END IF
                IF (force_eval) THEN
                   force_eval = do_stress
                   DO e = 1, 3
                      DO d = 1, 3
                         DO c = 1, 3
                            DO b = 1, 3
                               DO a = 1, 3
                                  tmp = rab(a)*rab(b)*rab(c)*rab(d)*rab(e)*fac_ij
                                  tij_abcde(a,b,c,d,e) = -945.0_dp*tmp*f(5)
                                  tmp1 = 105.0_dp*f(4)*fac_ij
                                  tmp2 =  15.0_dp*f(3)*fac_ij
                                  IF (a==b) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(c)*rab(d)*rab(e)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                  END IF
                                  IF (a==c) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(d)*rab(e)
                                     IF (b==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (b==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (a==d) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(c)*rab(e)
                                     IF (b==c) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(e)
                                     IF (b==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (a==e) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(b)*rab(c)*rab(d)
                                     IF (b==c) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(d)
                                     IF (b==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(c)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(b)
                                  END IF
                                  IF (b==c) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(d)*rab(e)
                                     IF (d==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (b==d) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(c)*rab(e)
                                     IF (c==e) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (b==e) THEN
                                               tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(c)*rab(d)
                                     IF (c==d) tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) - tmp2*rab(a)
                                  END IF
                                  IF (c==d)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(e)
                                  IF (c==e)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(d)
                                  IF (d==e)    tij_abcde(a,b,c,d,e) = tij_abcde(a,b,c,d,e) + tmp1*rab(a)*rab(b)*rab(c)
                               END DO
                            END DO
                         END DO
                      END DO
                   END DO
                END IF
                eloc  = 0.0_dp
                fr    = 0.0_dp
                ef0_i = 0.0_dp
                ef0_j = 0.0_dp
                ef1_j = 0.0_dp
                ef1_i = 0.0_dp
                ef2_j = 0.0_dp
                ef2_i = 0.0_dp


                ! Initialize the charge, dipole and quadrupole for atom A and B
                IF (debug_this_module) THEN
                   ch_j  = HUGE(0.0_dp)
                   ch_i  = HUGE(0.0_dp)
                   dp_j  = HUGE(0.0_dp)
                   dp_i  = HUGE(0.0_dp)
                   qp_j  = HUGE(0.0_dp)
                   qp_i  = HUGE(0.0_dp)
                END IF
                IF (ANY(task(1,:))) THEN
                   ch_j  = charges(atom_a)
                   ch_i  = charges(atom_b)
                END IF
                IF (ANY(task(2,:))) THEN
                   dp_j  = dipoles(:,atom_a)
                   dp_i  = dipoles(:,atom_b)
                END IF
                IF (ANY(task(3,:))) THEN
                   qp_j  = quadrupoles(:,:,atom_a)
                   qp_i  = quadrupoles(:,:,atom_b)
                END IF
                IF (task(1,1)) THEN
                   ! Charge - Charge
                   eloc = eloc + ch_i*tij*ch_j
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      fr(1) = fr(1) - ch_j * tij_a(1) * ch_i
                      fr(2) = fr(2) - ch_j * tij_a(2) * ch_i
                      fr(3) = fr(3) - ch_j * tij_a(3) * ch_i
                   END IF
                   ! Electric fields
                   IF (do_efield) THEN
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i + tij * ch_j

                         ef0_j = ef0_j + tij * ch_i
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) - tij_a(1) * ch_j
                         ef1_i(2) = ef1_i(2) - tij_a(2) * ch_j
                         ef1_i(3) = ef1_i(3) - tij_a(3) * ch_j

                         ef1_j(1) = ef1_j(1) + tij_a(1) * ch_i
                         ef1_j(2) = ef1_j(2) + tij_a(2) * ch_i
                         ef1_j(3) = ef1_j(3) + tij_a(3) * ch_i


                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         ef2_i(1,1) = ef2_i(1,1) - tij_ab(1,1) * ch_j
                         ef2_i(2,1) = ef2_i(2,1) - tij_ab(2,1) * ch_j
                         ef2_i(3,1) = ef2_i(3,1) - tij_ab(3,1) * ch_j
                         ef2_i(1,2) = ef2_i(1,2) - tij_ab(1,2) * ch_j
                         ef2_i(2,2) = ef2_i(2,2) - tij_ab(2,2) * ch_j
                         ef2_i(3,2) = ef2_i(3,2) - tij_ab(3,2) * ch_j
                         ef2_i(1,3) = ef2_i(1,3) - tij_ab(1,3) * ch_j
                         ef2_i(2,3) = ef2_i(2,3) - tij_ab(2,3) * ch_j
                         ef2_i(3,3) = ef2_i(3,3) - tij_ab(3,3) * ch_j

                         ef2_j(1,1) = ef2_j(1,1) - tij_ab(1,1) * ch_i
                         ef2_j(2,1) = ef2_j(2,1) - tij_ab(2,1) * ch_i
                         ef2_j(3,1) = ef2_j(3,1) - tij_ab(3,1) * ch_i
                         ef2_j(1,2) = ef2_j(1,2) - tij_ab(1,2) * ch_i
                         ef2_j(2,2) = ef2_j(2,2) - tij_ab(2,2) * ch_i
                         ef2_j(3,2) = ef2_j(3,2) - tij_ab(3,2) * ch_i
                         ef2_j(1,3) = ef2_j(1,3) - tij_ab(1,3) * ch_i
                         ef2_j(2,3) = ef2_j(2,3) - tij_ab(2,3) * ch_i
                         ef2_j(3,3) = ef2_j(3,3) - tij_ab(3,3) * ch_i
                      END IF
                   END IF
                END IF
                IF (task(2,2)) THEN
                   ! Dipole - Dipole
                   tmp= - (dp_i(1)*(tij_ab(1,1)*dp_j(1)+&
                                    tij_ab(2,1)*dp_j(2)+&
                                    tij_ab(3,1)*dp_j(3))+&
                           dp_i(2)*(tij_ab(1,2)*dp_j(1)+&
                                    tij_ab(2,2)*dp_j(2)+&
                                    tij_ab(3,2)*dp_j(3))+&
                           dp_i(3)*(tij_ab(1,3)*dp_j(1)+&
                                    tij_ab(2,3)*dp_j(2)+&
                                    tij_ab(3,3)*dp_j(3)))
                   eloc = eloc + tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         fr(k) = fr(k) +  dp_i(1)*(tij_abc(1,1,k)*dp_j(1)+&
                                                   tij_abc(2,1,k)*dp_j(2)+&
                                                   tij_abc(3,1,k)*dp_j(3))&
                                       +  dp_i(2)*(tij_abc(1,2,k)*dp_j(1)+&
                                                   tij_abc(2,2,k)*dp_j(2)+&
                                                   tij_abc(3,2,k)*dp_j(3))&
                                       +  dp_i(3)*(tij_abc(1,3,k)*dp_j(1)+&
                                                   tij_abc(2,3,k)*dp_j(2)+&
                                                   tij_abc(3,3,k)*dp_j(3))
                      END DO
                   END IF
                   ! Electric fields
                   IF (do_efield) THEN
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i - (tij_a(1)*dp_j(1)+&
                                          tij_a(2)*dp_j(2)+&
                                          tij_a(3)*dp_j(3))

                         ef0_j = ef0_j + (tij_a(1)*dp_i(1)+&
                                          tij_a(2)*dp_i(2)+&
                                          tij_a(3)*dp_i(3))
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) + (tij_ab(1,1)*dp_j(1)+&
                                                tij_ab(2,1)*dp_j(2)+&
                                                tij_ab(3,1)*dp_j(3))
                         ef1_i(2) = ef1_i(2) + (tij_ab(1,2)*dp_j(1)+&
                                                tij_ab(2,2)*dp_j(2)+&
                                                tij_ab(3,2)*dp_j(3))
                         ef1_i(3) = ef1_i(3) + (tij_ab(1,3)*dp_j(1)+&
                                                tij_ab(2,3)*dp_j(2)+&
                                                tij_ab(3,3)*dp_j(3))

                         ef1_j(1) = ef1_j(1) + (tij_ab(1,1)*dp_i(1)+&
                                                tij_ab(2,1)*dp_i(2)+&
                                                tij_ab(3,1)*dp_i(3))
                         ef1_j(2) = ef1_j(2) + (tij_ab(1,2)*dp_i(1)+&
                                                tij_ab(2,2)*dp_i(2)+&
                                                tij_ab(3,2)*dp_i(3))
                         ef1_j(3) = ef1_j(3) + (tij_ab(1,3)*dp_i(1)+&
                                                tij_ab(2,3)*dp_i(2)+&
                                                tij_ab(3,3)*dp_i(3))
                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         ef2_i(1,1) = ef2_i(1,1) + (tij_abc(1,1,1)*dp_j(1)+&
                                                    tij_abc(2,1,1)*dp_j(2)+&
                                                    tij_abc(3,1,1)*dp_j(3))
                         ef2_i(1,2) = ef2_i(1,2) + (tij_abc(1,1,2)*dp_j(1)+&
                                                    tij_abc(2,1,2)*dp_j(2)+&
                                                    tij_abc(3,1,2)*dp_j(3))
                         ef2_i(1,3) = ef2_i(1,3) + (tij_abc(1,1,3)*dp_j(1)+&
                                                    tij_abc(2,1,3)*dp_j(2)+&
                                                    tij_abc(3,1,3)*dp_j(3))
                         ef2_i(2,1) = ef2_i(2,1) + (tij_abc(1,2,1)*dp_j(1)+&
                                                    tij_abc(2,2,1)*dp_j(2)+&
                                                    tij_abc(3,2,1)*dp_j(3))
                         ef2_i(2,2) = ef2_i(2,2) + (tij_abc(1,2,2)*dp_j(1)+&
                                                    tij_abc(2,2,2)*dp_j(2)+&
                                                    tij_abc(3,2,2)*dp_j(3))
                         ef2_i(2,3) = ef2_i(2,3) + (tij_abc(1,2,3)*dp_j(1)+&
                                                    tij_abc(2,2,3)*dp_j(2)+&
                                                    tij_abc(3,2,3)*dp_j(3))
                         ef2_i(3,1) = ef2_i(3,1) + (tij_abc(1,3,1)*dp_j(1)+&
                                                    tij_abc(2,3,1)*dp_j(2)+&
                                                    tij_abc(3,3,1)*dp_j(3))
                         ef2_i(3,2) = ef2_i(3,2) + (tij_abc(1,3,2)*dp_j(1)+&
                                                    tij_abc(2,3,2)*dp_j(2)+&
                                                    tij_abc(3,3,2)*dp_j(3))
                         ef2_i(3,3) = ef2_i(3,3) + (tij_abc(1,3,3)*dp_j(1)+&
                                                    tij_abc(2,3,3)*dp_j(2)+&
                                                    tij_abc(3,3,3)*dp_j(3))

                         ef2_j(1,1) = ef2_j(1,1) - (tij_abc(1,1,1)*dp_i(1)+&
                                                    tij_abc(2,1,1)*dp_i(2)+&
                                                    tij_abc(3,1,1)*dp_i(3))
                         ef2_j(1,2) = ef2_j(1,2) - (tij_abc(1,1,2)*dp_i(1)+&
                                                    tij_abc(2,1,2)*dp_i(2)+&
                                                    tij_abc(3,1,2)*dp_i(3))
                         ef2_j(1,3) = ef2_j(1,3) - (tij_abc(1,1,3)*dp_i(1)+&
                                                    tij_abc(2,1,3)*dp_i(2)+&
                                                    tij_abc(3,1,3)*dp_i(3))
                         ef2_j(2,1) = ef2_j(2,1) - (tij_abc(1,2,1)*dp_i(1)+&
                                                    tij_abc(2,2,1)*dp_i(2)+&
                                                    tij_abc(3,2,1)*dp_i(3))
                         ef2_j(2,2) = ef2_j(2,2) - (tij_abc(1,2,2)*dp_i(1)+&
                                                    tij_abc(2,2,2)*dp_i(2)+&
                                                    tij_abc(3,2,2)*dp_i(3))
                         ef2_j(2,3) = ef2_j(2,3) - (tij_abc(1,2,3)*dp_i(1)+&
                                                    tij_abc(2,2,3)*dp_i(2)+&
                                                    tij_abc(3,2,3)*dp_i(3))
                         ef2_j(3,1) = ef2_j(3,1) - (tij_abc(1,3,1)*dp_i(1)+&
                                                    tij_abc(2,3,1)*dp_i(2)+&
                                                    tij_abc(3,3,1)*dp_i(3))
                         ef2_j(3,2) = ef2_j(3,2) - (tij_abc(1,3,2)*dp_i(1)+&
                                                    tij_abc(2,3,2)*dp_i(2)+&
                                                    tij_abc(3,3,2)*dp_i(3))
                         ef2_j(3,3) = ef2_j(3,3) - (tij_abc(1,3,3)*dp_i(1)+&
                                                    tij_abc(2,3,3)*dp_i(2)+&
                                                    tij_abc(3,3,3)*dp_i(3))
                      END IF
                   END IF
                END IF
                IF (task(2,1)) THEN
                   ! Dipole - Charge
                   tmp=   ch_j*(tij_a(1)*dp_i(1)+&
                                tij_a(2)*dp_i(2)+&
                                tij_a(3)*dp_i(3))&
                        - ch_i*(tij_a(1)*dp_j(1)+&
                                tij_a(2)*dp_j(2)+&
                                tij_a(3)*dp_j(3))
                   eloc = eloc + tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         fr(k) = fr(k) -  ch_j *(tij_ab(1,k)*dp_i(1)+&
                                                 tij_ab(2,k)*dp_i(2)+&
                                                 tij_ab(3,k)*dp_i(3))&
                                       +  ch_i *(tij_ab(1,k)*dp_j(1)+&
                                                 tij_ab(2,k)*dp_j(2)+&
                                                 tij_ab(3,k)*dp_j(3))
                      END DO
                   END IF
                END IF
                IF (task(3,3)) THEN
                   ! Quadrupole - Quadrupole
                   fac  = 1.0_dp/9.0_dp
                   tmp11 = qp_i(1,1)*(tij_abcd(1,1,1,1)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,1)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,1)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,1)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,1)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,1)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,1)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,1)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,1)*qp_j(3,3))
                   tmp21 = qp_i(2,1)*(tij_abcd(1,1,1,2)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,2)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,2)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,2)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,2)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,2)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,2)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,2)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,2)*qp_j(3,3))
                   tmp31 = qp_i(3,1)*(tij_abcd(1,1,1,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,1,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,1,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,1,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,1,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,1,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,1,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,1,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,1,3)*qp_j(3,3))
                   tmp22 = qp_i(2,2)*(tij_abcd(1,1,2,2)*qp_j(1,1)+&
                                      tij_abcd(2,1,2,2)*qp_j(2,1)+&
                                      tij_abcd(3,1,2,2)*qp_j(3,1)+&
                                      tij_abcd(1,2,2,2)*qp_j(1,2)+&
                                      tij_abcd(2,2,2,2)*qp_j(2,2)+&
                                      tij_abcd(3,2,2,2)*qp_j(3,2)+&
                                      tij_abcd(1,3,2,2)*qp_j(1,3)+&
                                      tij_abcd(2,3,2,2)*qp_j(2,3)+&
                                      tij_abcd(3,3,2,2)*qp_j(3,3))
                   tmp32 = qp_i(3,2)*(tij_abcd(1,1,2,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,2,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,2,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,2,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,2,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,2,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,2,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,2,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,2,3)*qp_j(3,3))
                   tmp33 = qp_i(3,3)*(tij_abcd(1,1,3,3)*qp_j(1,1)+&
                                      tij_abcd(2,1,3,3)*qp_j(2,1)+&
                                      tij_abcd(3,1,3,3)*qp_j(3,1)+&
                                      tij_abcd(1,2,3,3)*qp_j(1,2)+&
                                      tij_abcd(2,2,3,3)*qp_j(2,2)+&
                                      tij_abcd(3,2,3,3)*qp_j(3,2)+&
                                      tij_abcd(1,3,3,3)*qp_j(1,3)+&
                                      tij_abcd(2,3,3,3)*qp_j(2,3)+&
                                      tij_abcd(3,3,3,3)*qp_j(3,3))
                   tmp12 = tmp21
                   tmp13 = tmp31
                   tmp23 = tmp32
                   tmp   = tmp11 + tmp12 + tmp13 + &
                           tmp21 + tmp22 + tmp23 + &
                           tmp31 + tmp32 + tmp33

                   eloc = eloc + fac*tmp
                   ! Forces on particle i (locally b)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         tmp11 = qp_i(1,1)*(tij_abcde(1,1,1,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,1,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,1,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,1,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,1,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,1,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,1,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,1,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,1,1,k)*qp_j(3,3))
                         tmp21 = qp_i(2,1)*(tij_abcde(1,1,2,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,2,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,2,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,2,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,2,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,2,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,2,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,2,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,2,1,k)*qp_j(3,3))
                         tmp31 = qp_i(3,1)*(tij_abcde(1,1,3,1,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,1,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,1,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,1,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,1,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,1,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,1,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,1,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,1,k)*qp_j(3,3))
                         tmp22 = qp_i(2,2)*(tij_abcde(1,1,2,2,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,2,2,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,2,2,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,2,2,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,2,2,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,2,2,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,2,2,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,2,2,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,2,2,k)*qp_j(3,3))
                         tmp32 = qp_i(3,2)*(tij_abcde(1,1,3,2,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,2,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,2,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,2,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,2,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,2,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,2,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,2,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,2,k)*qp_j(3,3))
                         tmp33 = qp_i(3,3)*(tij_abcde(1,1,3,3,k)*qp_j(1,1)+&
                                            tij_abcde(2,1,3,3,k)*qp_j(2,1)+&
                                            tij_abcde(3,1,3,3,k)*qp_j(3,1)+&
                                            tij_abcde(1,2,3,3,k)*qp_j(1,2)+&
                                            tij_abcde(2,2,3,3,k)*qp_j(2,2)+&
                                            tij_abcde(3,2,3,3,k)*qp_j(3,2)+&
                                            tij_abcde(1,3,3,3,k)*qp_j(1,3)+&
                                            tij_abcde(2,3,3,3,k)*qp_j(2,3)+&
                                            tij_abcde(3,3,3,3,k)*qp_j(3,3))
                         tmp12 = tmp21
                         tmp13 = tmp31
                         tmp23 = tmp32
                         fr(k) = fr(k) - fac * ( tmp11 + tmp12 + tmp13 +&
                                                 tmp21 + tmp22 + tmp23 +&
                                                 tmp31 + tmp32 + tmp33  )
                      END DO
                   END IF
                   ! Electric field
                   IF (do_efield) THEN
                      fac = 1.0_dp/3.0_dp
                      ! Potential
                      IF (do_efield0) THEN
                         ef0_i = ef0_i + fac*(tij_ab(1,1)*qp_j(1,1)+&
                                              tij_ab(2,1)*qp_j(2,1)+&
                                              tij_ab(3,1)*qp_j(3,1)+&
                                              tij_ab(1,2)*qp_j(1,2)+&
                                              tij_ab(2,2)*qp_j(2,2)+&
                                              tij_ab(3,2)*qp_j(3,2)+&
                                              tij_ab(1,3)*qp_j(1,3)+&
                                              tij_ab(2,3)*qp_j(2,3)+&
                                              tij_ab(3,3)*qp_j(3,3))

                         ef0_j = ef0_j + fac*(tij_ab(1,1)*qp_i(1,1)+&
                                              tij_ab(2,1)*qp_i(2,1)+&
                                              tij_ab(3,1)*qp_i(3,1)+&
                                              tij_ab(1,2)*qp_i(1,2)+&
                                              tij_ab(2,2)*qp_i(2,2)+&
                                              tij_ab(3,2)*qp_i(3,2)+&
                                              tij_ab(1,3)*qp_i(1,3)+&
                                              tij_ab(2,3)*qp_i(2,3)+&
                                              tij_ab(3,3)*qp_i(3,3))
                      END IF
                      ! Electric field
                      IF (do_efield1) THEN
                         ef1_i(1) = ef1_i(1) - fac*(tij_abc(1,1,1)*qp_j(1,1)+&
                                                    tij_abc(2,1,1)*qp_j(2,1)+&
                                                    tij_abc(3,1,1)*qp_j(3,1)+&
                                                    tij_abc(1,2,1)*qp_j(1,2)+&
                                                    tij_abc(2,2,1)*qp_j(2,2)+&
                                                    tij_abc(3,2,1)*qp_j(3,2)+&
                                                    tij_abc(1,3,1)*qp_j(1,3)+&
                                                    tij_abc(2,3,1)*qp_j(2,3)+&
                                                    tij_abc(3,3,1)*qp_j(3,3))
                         ef1_i(2) = ef1_i(2) - fac*(tij_abc(1,1,2)*qp_j(1,1)+&
                                                    tij_abc(2,1,2)*qp_j(2,1)+&
                                                    tij_abc(3,1,2)*qp_j(3,1)+&
                                                    tij_abc(1,2,2)*qp_j(1,2)+&
                                                    tij_abc(2,2,2)*qp_j(2,2)+&
                                                    tij_abc(3,2,2)*qp_j(3,2)+&
                                                    tij_abc(1,3,2)*qp_j(1,3)+&
                                                    tij_abc(2,3,2)*qp_j(2,3)+&
                                                    tij_abc(3,3,2)*qp_j(3,3))
                         ef1_i(3) = ef1_i(3) - fac*(tij_abc(1,1,3)*qp_j(1,1)+&
                                                    tij_abc(2,1,3)*qp_j(2,1)+&
                                                    tij_abc(3,1,3)*qp_j(3,1)+&
                                                    tij_abc(1,2,3)*qp_j(1,2)+&
                                                    tij_abc(2,2,3)*qp_j(2,2)+&
                                                    tij_abc(3,2,3)*qp_j(3,2)+&
                                                    tij_abc(1,3,3)*qp_j(1,3)+&
                                                    tij_abc(2,3,3)*qp_j(2,3)+&
                                                    tij_abc(3,3,3)*qp_j(3,3))

                         ef1_j(1) = ef1_j(1) + fac*(tij_abc(1,1,1)*qp_i(1,1)+&
                                                    tij_abc(2,1,1)*qp_i(2,1)+&
                                                    tij_abc(3,1,1)*qp_i(3,1)+&
                                                    tij_abc(1,2,1)*qp_i(1,2)+&
                                                    tij_abc(2,2,1)*qp_i(2,2)+&
                                                    tij_abc(3,2,1)*qp_i(3,2)+&
                                                    tij_abc(1,3,1)*qp_i(1,3)+&
                                                    tij_abc(2,3,1)*qp_i(2,3)+&
                                                    tij_abc(3,3,1)*qp_i(3,3))
                         ef1_j(2) = ef1_j(2) + fac*(tij_abc(1,1,2)*qp_i(1,1)+&
                                                    tij_abc(2,1,2)*qp_i(2,1)+&
                                                    tij_abc(3,1,2)*qp_i(3,1)+&
                                                    tij_abc(1,2,2)*qp_i(1,2)+&
                                                    tij_abc(2,2,2)*qp_i(2,2)+&
                                                    tij_abc(3,2,2)*qp_i(3,2)+&
                                                    tij_abc(1,3,2)*qp_i(1,3)+&
                                                    tij_abc(2,3,2)*qp_i(2,3)+&
                                                    tij_abc(3,3,2)*qp_i(3,3))
                         ef1_j(3) = ef1_j(3) + fac*(tij_abc(1,1,3)*qp_i(1,1)+&
                                                    tij_abc(2,1,3)*qp_i(2,1)+&
                                                    tij_abc(3,1,3)*qp_i(3,1)+&
                                                    tij_abc(1,2,3)*qp_i(1,2)+&
                                                    tij_abc(2,2,3)*qp_i(2,2)+&
                                                    tij_abc(3,2,3)*qp_i(3,2)+&
                                                    tij_abc(1,3,3)*qp_i(1,3)+&
                                                    tij_abc(2,3,3)*qp_i(2,3)+&
                                                    tij_abc(3,3,3)*qp_i(3,3))
                      END IF
                      ! Electric field gradient
                      IF (do_efield2) THEN
                         tmp11 =   fac *(tij_abcd(1,1,1,1)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,1)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,1)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,1)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,1)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,1)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,1)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,1)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,1)*qp_j(3,3))
                         tmp12 =   fac *(tij_abcd(1,1,1,2)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,2)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,2)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,2)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,2)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,2)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,2)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,2)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,2)*qp_j(3,3))
                         tmp13 =   fac *(tij_abcd(1,1,1,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,1,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,1,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,1,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,1,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,1,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,1,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,1,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,1,3)*qp_j(3,3))
                         tmp22 =   fac *(tij_abcd(1,1,2,2)*qp_j(1,1)+&
                                         tij_abcd(2,1,2,2)*qp_j(2,1)+&
                                         tij_abcd(3,1,2,2)*qp_j(3,1)+&
                                         tij_abcd(1,2,2,2)*qp_j(1,2)+&
                                         tij_abcd(2,2,2,2)*qp_j(2,2)+&
                                         tij_abcd(3,2,2,2)*qp_j(3,2)+&
                                         tij_abcd(1,3,2,2)*qp_j(1,3)+&
                                         tij_abcd(2,3,2,2)*qp_j(2,3)+&
                                         tij_abcd(3,3,2,2)*qp_j(3,3))
                         tmp23 =   fac *(tij_abcd(1,1,2,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,2,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,2,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,2,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,2,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,2,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,2,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,2,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,2,3)*qp_j(3,3))
                         tmp33 =   fac *(tij_abcd(1,1,3,3)*qp_j(1,1)+&
                                         tij_abcd(2,1,3,3)*qp_j(2,1)+&
                                         tij_abcd(3,1,3,3)*qp_j(3,1)+&
                                         tij_abcd(1,2,3,3)*qp_j(1,2)+&
                                         tij_abcd(2,2,3,3)*qp_j(2,2)+&
                                         tij_abcd(3,2,3,3)*qp_j(3,2)+&
                                         tij_abcd(1,3,3,3)*qp_j(1,3)+&
                                         tij_abcd(2,3,3,3)*qp_j(2,3)+&
                                         tij_abcd(3,3,3,3)*qp_j(3,3))

                         ef2_i(1,1) = ef2_i(1,1) - tmp11
                         ef2_i(1,2) = ef2_i(1,2) - tmp12
                         ef2_i(1,3) = ef2_i(1,3) - tmp13
                         ef2_i(2,1) = ef2_i(2,1) - tmp12
                         ef2_i(2,2) = ef2_i(2,2) - tmp22
                         ef2_i(2,3) = ef2_i(2,3) - tmp23
                         ef2_i(3,1) = ef2_i(3,1) - tmp13
                         ef2_i(3,2) = ef2_i(3,2) - tmp23
                         ef2_i(3,3) = ef2_i(3,3) - tmp33

                         tmp11 =   fac *(tij_abcd(1,1,1,1)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,1)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,1)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,1)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,1)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,1)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,1)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,1)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,1)*qp_i(3,3))
                         tmp12 =   fac *(tij_abcd(1,1,1,2)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,2)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,2)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,2)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,2)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,2)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,2)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,2)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,2)*qp_i(3,3))
                         tmp13 =   fac *(tij_abcd(1,1,1,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,1,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,1,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,1,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,1,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,1,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,1,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,1,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,1,3)*qp_i(3,3))
                         tmp22 =   fac *(tij_abcd(1,1,2,2)*qp_i(1,1)+&
                                         tij_abcd(2,1,2,2)*qp_i(2,1)+&
                                         tij_abcd(3,1,2,2)*qp_i(3,1)+&
                                         tij_abcd(1,2,2,2)*qp_i(1,2)+&
                                         tij_abcd(2,2,2,2)*qp_i(2,2)+&
                                         tij_abcd(3,2,2,2)*qp_i(3,2)+&
                                         tij_abcd(1,3,2,2)*qp_i(1,3)+&
                                         tij_abcd(2,3,2,2)*qp_i(2,3)+&
                                         tij_abcd(3,3,2,2)*qp_i(3,3))
                         tmp23 =   fac *(tij_abcd(1,1,2,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,2,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,2,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,2,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,2,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,2,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,2,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,2,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,2,3)*qp_i(3,3))
                         tmp33 =   fac *(tij_abcd(1,1,3,3)*qp_i(1,1)+&
                                         tij_abcd(2,1,3,3)*qp_i(2,1)+&
                                         tij_abcd(3,1,3,3)*qp_i(3,1)+&
                                         tij_abcd(1,2,3,3)*qp_i(1,2)+&
                                         tij_abcd(2,2,3,3)*qp_i(2,2)+&
                                         tij_abcd(3,2,3,3)*qp_i(3,2)+&
                                         tij_abcd(1,3,3,3)*qp_i(1,3)+&
                                         tij_abcd(2,3,3,3)*qp_i(2,3)+&
                                         tij_abcd(3,3,3,3)*qp_i(3,3))

                         ef2_j(1,1) = ef2_j(1,1) - tmp11
                         ef2_j(1,2) = ef2_j(1,2) - tmp12
                         ef2_j(1,3) = ef2_j(1,3) - tmp13
                         ef2_j(2,1) = ef2_j(2,1) - tmp12
                         ef2_j(2,2) = ef2_j(2,2) - tmp22
                         ef2_j(2,3) = ef2_j(2,3) - tmp23
                         ef2_j(3,1) = ef2_j(3,1) - tmp13
                         ef2_j(3,2) = ef2_j(3,2) - tmp23
                         ef2_j(3,3) = ef2_j(3,3) - tmp33
                      END IF
                   END IF
                END IF
                IF (task(3,2)) THEN
                   ! Quadrupole - Dipole
                   fac = 1.0_dp/3.0_dp
                   ! Dipole i (locally B) - Quadrupole j (locally A)
                   tmp_ij = dp_i(1)*(tij_abc(1,1,1)*qp_j(1,1)+&
                                     tij_abc(2,1,1)*qp_j(2,1)+&
                                     tij_abc(3,1,1)*qp_j(3,1)+&
                                     tij_abc(1,2,1)*qp_j(1,2)+&
                                     tij_abc(2,2,1)*qp_j(2,2)+&
                                     tij_abc(3,2,1)*qp_j(3,2)+&
                                     tij_abc(1,3,1)*qp_j(1,3)+&
                                     tij_abc(2,3,1)*qp_j(2,3)+&
                                     tij_abc(3,3,1)*qp_j(3,3))+&
                            dp_i(2)*(tij_abc(1,1,2)*qp_j(1,1)+&
                                     tij_abc(2,1,2)*qp_j(2,1)+&
                                     tij_abc(3,1,2)*qp_j(3,1)+&
                                     tij_abc(1,2,2)*qp_j(1,2)+&
                                     tij_abc(2,2,2)*qp_j(2,2)+&
                                     tij_abc(3,2,2)*qp_j(3,2)+&
                                     tij_abc(1,3,2)*qp_j(1,3)+&
                                     tij_abc(2,3,2)*qp_j(2,3)+&
                                     tij_abc(3,3,2)*qp_j(3,3))+&
                            dp_i(3)*(tij_abc(1,1,3)*qp_j(1,1)+&
                                     tij_abc(2,1,3)*qp_j(2,1)+&
                                     tij_abc(3,1,3)*qp_j(3,1)+&
                                     tij_abc(1,2,3)*qp_j(1,2)+&
                                     tij_abc(2,2,3)*qp_j(2,2)+&
                                     tij_abc(3,2,3)*qp_j(3,2)+&
                                     tij_abc(1,3,3)*qp_j(1,3)+&
                                     tij_abc(2,3,3)*qp_j(2,3)+&
                                     tij_abc(3,3,3)*qp_j(3,3))

                   ! Dipole j (locally A) - Quadrupole i (locally B)
                   tmp_ji = dp_j(1)*(tij_abc(1,1,1)*qp_i(1,1)+&
                                     tij_abc(2,1,1)*qp_i(2,1)+&
                                     tij_abc(3,1,1)*qp_i(3,1)+&
                                     tij_abc(1,2,1)*qp_i(1,2)+&
                                     tij_abc(2,2,1)*qp_i(2,2)+&
                                     tij_abc(3,2,1)*qp_i(3,2)+&
                                     tij_abc(1,3,1)*qp_i(1,3)+&
                                     tij_abc(2,3,1)*qp_i(2,3)+&
                                     tij_abc(3,3,1)*qp_i(3,3))+&
                            dp_j(2)*(tij_abc(1,1,2)*qp_i(1,1)+&
                                     tij_abc(2,1,2)*qp_i(2,1)+&
                                     tij_abc(3,1,2)*qp_i(3,1)+&
                                     tij_abc(1,2,2)*qp_i(1,2)+&
                                     tij_abc(2,2,2)*qp_i(2,2)+&
                                     tij_abc(3,2,2)*qp_i(3,2)+&
                                     tij_abc(1,3,2)*qp_i(1,3)+&
                                     tij_abc(2,3,2)*qp_i(2,3)+&
                                     tij_abc(3,3,2)*qp_i(3,3))+&
                            dp_j(3)*(tij_abc(1,1,3)*qp_i(1,1)+&
                                     tij_abc(2,1,3)*qp_i(2,1)+&
                                     tij_abc(3,1,3)*qp_i(3,1)+&
                                     tij_abc(1,2,3)*qp_i(1,2)+&
                                     tij_abc(2,2,3)*qp_i(2,2)+&
                                     tij_abc(3,2,3)*qp_i(3,2)+&
                                     tij_abc(1,3,3)*qp_i(1,3)+&
                                     tij_abc(2,3,3)*qp_i(2,3)+&
                                     tij_abc(3,3,3)*qp_i(3,3))

                   tmp= fac * (tmp_ij - tmp_ji)
                   eloc = eloc + tmp
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         ! Dipole i (locally B) - Quadrupole j (locally A)
                         tmp_ij = dp_i(1)*(tij_abcd(1,1,1,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,1,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,1,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,1,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,1,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,1,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,1,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,1,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,1,k)*qp_j(3,3))+&
                                  dp_i(2)*(tij_abcd(1,1,2,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,2,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,2,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,2,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,2,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,2,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,2,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,2,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,2,k)*qp_j(3,3))+&
                                  dp_i(3)*(tij_abcd(1,1,3,k)*qp_j(1,1)+&
                                           tij_abcd(2,1,3,k)*qp_j(2,1)+&
                                           tij_abcd(3,1,3,k)*qp_j(3,1)+&
                                           tij_abcd(1,2,3,k)*qp_j(1,2)+&
                                           tij_abcd(2,2,3,k)*qp_j(2,2)+&
                                           tij_abcd(3,2,3,k)*qp_j(3,2)+&
                                           tij_abcd(1,3,3,k)*qp_j(1,3)+&
                                           tij_abcd(2,3,3,k)*qp_j(2,3)+&
                                           tij_abcd(3,3,3,k)*qp_j(3,3))

                         ! Dipole j (locally A) - Quadrupole i (locally B)
                         tmp_ji = dp_j(1)*(tij_abcd(1,1,1,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,1,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,1,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,1,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,1,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,1,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,1,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,1,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,1,k)*qp_i(3,3))+&
                                  dp_j(2)*(tij_abcd(1,1,2,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,2,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,2,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,2,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,2,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,2,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,2,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,2,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,2,k)*qp_i(3,3))+&
                                  dp_j(3)*(tij_abcd(1,1,3,k)*qp_i(1,1)+&
                                           tij_abcd(2,1,3,k)*qp_i(2,1)+&
                                           tij_abcd(3,1,3,k)*qp_i(3,1)+&
                                           tij_abcd(1,2,3,k)*qp_i(1,2)+&
                                           tij_abcd(2,2,3,k)*qp_i(2,2)+&
                                           tij_abcd(3,2,3,k)*qp_i(3,2)+&
                                           tij_abcd(1,3,3,k)*qp_i(1,3)+&
                                           tij_abcd(2,3,3,k)*qp_i(2,3)+&
                                           tij_abcd(3,3,3,k)*qp_i(3,3))

                         fr(k) = fr(k) - fac * (tmp_ij - tmp_ji)
                      END DO
                   END IF
                END IF
                IF (task(3,1)) THEN
                   ! Quadrupole - Charge
                   fac = 1.0_dp/3.0_dp

                   ! Quadrupole j (locally A) - Charge j (locally B)
                   tmp_ij = ch_i * (tij_ab(1,1)*qp_j(1,1)+&
                                    tij_ab(2,1)*qp_j(2,1)+&
                                    tij_ab(3,1)*qp_j(3,1)+&
                                    tij_ab(1,2)*qp_j(1,2)+&
                                    tij_ab(2,2)*qp_j(2,2)+&
                                    tij_ab(3,2)*qp_j(3,2)+&
                                    tij_ab(1,3)*qp_j(1,3)+&
                                    tij_ab(2,3)*qp_j(2,3)+&
                                    tij_ab(3,3)*qp_j(3,3))

                   ! Quadrupole i (locally B) - Charge j (locally A)
                   tmp_ji = ch_j * (tij_ab(1,1)*qp_i(1,1)+&
                                    tij_ab(2,1)*qp_i(2,1)+&
                                    tij_ab(3,1)*qp_i(3,1)+&
                                    tij_ab(1,2)*qp_i(1,2)+&
                                    tij_ab(2,2)*qp_i(2,2)+&
                                    tij_ab(3,2)*qp_i(3,2)+&
                                    tij_ab(1,3)*qp_i(1,3)+&
                                    tij_ab(2,3)*qp_i(2,3)+&
                                    tij_ab(3,3)*qp_i(3,3))

                   eloc = eloc + fac*(tmp_ij+tmp_ji)
                   IF (do_forces.OR.do_stress) THEN
                      DO k = 1, 3
                         ! Quadrupole j (locally A) - Charge i (locally B)
                         tmp_ij = ch_i * (tij_abc(1,1,k)*qp_j(1,1)+&
                                          tij_abc(2,1,k)*qp_j(2,1)+&
                                          tij_abc(3,1,k)*qp_j(3,1)+&
                                          tij_abc(1,2,k)*qp_j(1,2)+&
                                          tij_abc(2,2,k)*qp_j(2,2)+&
                                          tij_abc(3,2,k)*qp_j(3,2)+&
                                          tij_abc(1,3,k)*qp_j(1,3)+&
                                          tij_abc(2,3,k)*qp_j(2,3)+&
                                          tij_abc(3,3,k)*qp_j(3,3))

                         ! Quadrupole i (locally B) - Charge j (locally A)
                         tmp_ji = ch_j * (tij_abc(1,1,k)*qp_i(1,1)+&
                                          tij_abc(2,1,k)*qp_i(2,1)+&
                                          tij_abc(3,1,k)*qp_i(3,1)+&
                                          tij_abc(1,2,k)*qp_i(1,2)+&
                                          tij_abc(2,2,k)*qp_i(2,2)+&
                                          tij_abc(3,2,k)*qp_i(3,2)+&
                                          tij_abc(1,3,k)*qp_i(1,3)+&
                                          tij_abc(2,3,k)*qp_i(2,3)+&
                                          tij_abc(3,3,k)*qp_i(3,3))

                         fr(k) = fr(k) - fac *(tmp_ij + tmp_ji)
                      END DO
                   END IF
                END IF
                ! Electric fields
                IF (do_efield) THEN
                   ! Potential
                   IF (do_efield0) THEN
                      efield0(  atom_a) = efield0(  atom_a) + ef0_j

                      efield0(  atom_b) = efield0(  atom_b) + ef0_i
                   END IF
                   ! Electric field
                   IF (do_efield1) THEN
                      efield1(1,atom_a) = efield1(1,atom_a) + ef1_j(1)
                      efield1(2,atom_a) = efield1(2,atom_a) + ef1_j(2)
                      efield1(3,atom_a) = efield1(3,atom_a) + ef1_j(3)

                      efield1(1,atom_b) = efield1(1,atom_b) + ef1_i(1)
                      efield1(2,atom_b) = efield1(2,atom_b) + ef1_i(2)
                      efield1(3,atom_b) = efield1(3,atom_b) + ef1_i(3)
                   END IF
                   ! Electric field gradient
                   IF (do_efield2) THEN
                      efield2(1,atom_a) = efield2(1,atom_a) + ef2_j(1,1)
                      efield2(2,atom_a) = efield2(2,atom_a) + ef2_j(1,2)
                      efield2(3,atom_a) = efield2(3,atom_a) + ef2_j(1,3)
                      efield2(4,atom_a) = efield2(4,atom_a) + ef2_j(2,1)
                      efield2(5,atom_a) = efield2(5,atom_a) + ef2_j(2,2)
                      efield2(6,atom_a) = efield2(6,atom_a) + ef2_j(2,3)
                      efield2(7,atom_a) = efield2(7,atom_a) + ef2_j(3,1)
                      efield2(8,atom_a) = efield2(8,atom_a) + ef2_j(3,2)
                      efield2(9,atom_a) = efield2(9,atom_a) + ef2_j(3,3)

                      efield2(1,atom_b) = efield2(1,atom_b) + ef2_i(1,1)
                      efield2(2,atom_b) = efield2(2,atom_b) + ef2_i(1,2)
                      efield2(3,atom_b) = efield2(3,atom_b) + ef2_i(1,3)
                      efield2(4,atom_b) = efield2(4,atom_b) + ef2_i(2,1)
                      efield2(5,atom_b) = efield2(5,atom_b) + ef2_i(2,2)
                      efield2(6,atom_b) = efield2(6,atom_b) + ef2_i(2,3)
                      efield2(7,atom_b) = efield2(7,atom_b) + ef2_i(3,1)
                      efield2(8,atom_b) = efield2(8,atom_b) + ef2_i(3,2)
                      efield2(9,atom_b) = efield2(9,atom_b) + ef2_i(3,3)
                   END IF
                END IF
                IF (do_stress) THEN
                   ptens11 = ptens11 + rab(1) * fr(1)
                   ptens21 = ptens21 + rab(2) * fr(1)
                   ptens31 = ptens31 + rab(3) * fr(1)
                   ptens12 = ptens12 + rab(1) * fr(2)
                   ptens22 = ptens22 + rab(2) * fr(2)
                   ptens32 = ptens32 + rab(3) * fr(2)
                   ptens13 = ptens13 + rab(1) * fr(3)
                   ptens23 = ptens23 + rab(2) * fr(3)
                   ptens33 = ptens33 + rab(3) * fr(3)
                END IF

      IF (PRESENT(integral_value)) THEN
         integral_value = eloc
      END IF
      IF (do_forces) THEN
         force_ab = fr
      END IF
   END SUBROUTINE se_coulomb_ij_interaction

END MODULE se_fock_matrix_integrals

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Module that collects all Coulomb parts of the fock matrix construction
!>
!> \author Teodoro Laino (05.2009) [tlaino] - split and module reorganization
!> \par History
!>      Teodoro Laino (04.2008) [tlaino] - University of Zurich : d-orbitals
!>      Teodoro Laino (09.2008) [tlaino] - University of Zurich : Speed-up
!>      Teodoro Laino (09.2008) [tlaino] - University of Zurich : Periodic SE
!>      Teodoro Laino (05.2009) [tlaino] - Stress Tensor
!>
! **************************************************************************************************
MODULE se_fock_matrix_coulomb
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind_set
   USE atprop_types,                    ONLY: atprop_type
   USE cell_types,                      ONLY: cell_type
   USE cp_control_types,                ONLY: dft_control_type,&
                                              semi_empirical_control_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_allocate_matrix_set, cp_dbcsr_deallocate_matrix_set, &
        cp_dbcsr_distribute, cp_dbcsr_get_block_diag, cp_dbcsr_get_block_p, cp_dbcsr_init, &
        cp_dbcsr_p_type, cp_dbcsr_replicate_all, cp_dbcsr_set, cp_dbcsr_sum_replicated
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE ewald_environment_types,         ONLY: ewald_env_get,&
                                              ewald_environment_type
   USE ewald_pw_types,                  ONLY: ewald_pw_get,&
                                              ewald_pw_type
   USE ewalds_multipole,                ONLY: ewald_multipole_evaluate
   USE fist_neighbor_list_control,      ONLY: list_control
   USE fist_nonbond_env_types,          ONLY: fist_nonbond_env_type
   USE input_constants,                 ONLY: &
        do_method_am1, do_method_mndo, do_method_mndod, do_method_pdg, do_method_pm3, &
        do_method_pm6, do_method_pnnl, do_method_rm1, do_se_IS_slater
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE multipole_types,                 ONLY: do_multipole_charge,&
                                              do_multipole_dipole,&
                                              do_multipole_none,&
                                              do_multipole_quadrupole
   USE particle_types,                  ONLY: particle_type
   USE pw_poisson_types,                ONLY: do_ewald_ewald
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_environment_types,            ONLY: get_qs_env,&
                                              qs_environment_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_neighbor_list_types,          ONLY: get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE se_fock_matrix_integrals,        ONLY: &
        dfock2C, dfock2C_r3, dfock2_1el, dfock2_1el_r3, fock2C, fock2C_ew, fock2C_r3, fock2_1el, &
        fock2_1el_ew, fock2_1el_r3, se_coulomb_ij_interaction
   USE semi_empirical_int_arrays,       ONLY: rij_threshold,&
                                              se_orbital_pointer
   USE semi_empirical_integrals,        ONLY: corecore_el,&
                                              dcorecore_el
   USE semi_empirical_mpole_methods,    ONLY: quadrupole_sph_to_cart
   USE semi_empirical_mpole_types,      ONLY: nddo_mpole_type,&
                                              semi_empirical_mpole_type
   USE semi_empirical_store_int_types,  ONLY: semi_empirical_si_type
   USE semi_empirical_types,            ONLY: get_se_param,&
                                              se_int_control_type,&
                                              se_taper_type,&
                                              semi_empirical_p_type,&
                                              semi_empirical_type,&
                                              setup_se_int_control_type
   USE semi_empirical_utils,            ONLY: finalize_se_taper,&
                                              get_se_type,&
                                              initialize_se_taper
   USE virial_methods,                  ONLY: virial_pair_force
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'se_fock_matrix_coulomb'
   LOGICAL, PARAMETER, PRIVATE          :: debug_this_module = .FALSE.

   PUBLIC :: build_fock_matrix_coulomb, build_fock_matrix_coulomb_lr, &
             build_fock_matrix_coul_lr_r3

CONTAINS

! **************************************************************************************************
!> \brief Construction of the Coulomb part of the Fock matrix
!> \param qs_env ...
!> \param ks_matrix ...
!> \param matrix_p ...
!> \param energy ...
!> \param calculate_forces ...
!> \param store_int_env ...
!> \author JGH
! **************************************************************************************************
   SUBROUTINE build_fock_matrix_coulomb(qs_env, ks_matrix, matrix_p, energy, calculate_forces, &
                                        store_int_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_matrix, matrix_p
      TYPE(qs_energy_type), POINTER                      :: energy
      LOGICAL, INTENT(in)                                :: calculate_forces
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env

      CHARACTER(len=*), PARAMETER :: routineN = 'build_fock_matrix_coulomb', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, handle, iatom, ikind, inode, ispin, itype, jatom, jkind, natom, &
         natorb_a, natorb_a2, natorb_b, natorb_b2, nkind, nspins
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: se_natorb
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind
      LOGICAL                                            :: anag, atener, check, defined, found, &
                                                            switch, use_virial
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: se_defined
      REAL(KIND=dp)                                      :: delta, dr1, ecore2, ecores
      REAL(KIND=dp), DIMENSION(2)                        :: ecab
      REAL(KIND=dp), DIMENSION(2025)                     :: pa_a, pa_b, pb_a, pb_b
      REAL(KIND=dp), DIMENSION(3)                        :: force_ab, rij
      REAL(KIND=dp), DIMENSION(45, 45)                   :: p_block_tot_a, p_block_tot_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ksa_block_a, ksa_block_b, ksb_block_a, &
                                                            ksb_block_b, pa_block_a, pa_block_b, &
                                                            pb_block_a, pb_block_b
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: diagmat_ks, diagmat_p
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_se
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(se_int_control_type)                          :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_control_type), POINTER         :: se_control
      TYPE(semi_empirical_p_type), DIMENSION(:), POINTER :: se_kind_list
      TYPE(semi_empirical_type), POINTER                 :: se_kind_a, se_kind_b
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (dft_control, cell, force, particle_set, diagmat_ks, diagmat_p, &
               se_control, se_taper, virial, atprop)

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, cell=cell, se_taper=se_taper, &
                      para_env=para_env, sab_se=sab_se, atomic_kind_set=atomic_kind_set, atprop=atprop, &
                      qs_kind_set=qs_kind_set, particle_set=particle_set, virial=virial)

      ! Parameters
      CALL initialize_se_taper(se_taper, coulomb=.TRUE.)
      se_control => dft_control%qs_control%se_control
      anag = se_control%analytical_gradients
      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)
      atener = atprop%energy

      CALL setup_se_int_control_type(se_int_control, do_ewald_r3=se_control%do_ewald_r3, &
                                     do_ewald_gks=se_control%do_ewald_gks, integral_screening=se_control%integral_screening, &
                                     shortrange=(se_control%do_ewald .OR. se_control%do_ewald_gks), &
                                     max_multipole=se_control%max_multipole, pc_coulomb_int=.FALSE.)
      IF (se_control%do_ewald_gks) THEN
         CALL get_qs_env(qs_env=qs_env, ewald_env=ewald_env, ewald_pw=ewald_pw)
         CALL ewald_env_get(ewald_env, alpha=se_int_control%ewald_gks%alpha)
         CALL ewald_pw_get(ewald_pw, pw_big_pool=se_int_control%ewald_gks%pw_pool, &
                           dg=se_int_control%ewald_gks%dg)
      END IF

      nspins = dft_control%nspins
      IF(.NOT.(ASSOCIATED(matrix_p)))CALL cp__a("se_fock_matrix_coulomb.F",188)
      IF(.NOT.(SIZE(ks_matrix) > 0))CALL cp__a("se_fock_matrix_coulomb.F",189)

      nkind = SIZE(atomic_kind_set)
      IF (calculate_forces) THEN
         CALL get_qs_env(qs_env=qs_env, force=force)
         natom = SIZE(particle_set)
         delta = se_control%delta
         ! Allocate atom index for kind
         ALLOCATE (atom_of_kind(natom))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind)
      END IF

      CALL cp_dbcsr_allocate_matrix_set(diagmat_ks, nspins)
      CALL cp_dbcsr_allocate_matrix_set(diagmat_p, nspins)

      DO ispin = 1, nspins
         ! Allocate diagonal block matrices
         ALLOCATE (diagmat_p(ispin)%matrix, diagmat_ks(ispin)%matrix) !sm->dbcsr
         CALL cp_dbcsr_init(diagmat_p(ispin)%matrix)
         CALL cp_dbcsr_init(diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_get_block_diag(matrix_p(ispin)%matrix, diagmat_p(ispin)%matrix)
         CALL cp_dbcsr_get_block_diag(ks_matrix(ispin)%matrix, diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_set(diagmat_ks(ispin)%matrix, 0.0_dp)
         CALL cp_dbcsr_replicate_all(diagmat_p(ispin)%matrix)
         CALL cp_dbcsr_replicate_all(diagmat_ks(ispin)%matrix)
      END DO

      ecore2 = 0.0_dp
      itype = get_se_type(dft_control%qs_control%method_id)
      ALLOCATE (se_defined(nkind), se_kind_list(nkind), se_natorb(nkind))
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind_a)
         se_kind_list(ikind)%se_param => se_kind_a
         CALL get_se_param(se_kind_a, defined=defined, natorb=natorb_a)
         se_defined(ikind) = (defined .AND. natorb_a >= 1)
         se_natorb(ikind) = natorb_a
      END DO

      CALL neighbor_list_iterator_create(nl_iterator, sab_se)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, jatom=jatom, inode=inode, r=rij)
         IF (.NOT. se_defined(ikind)) CYCLE
         IF (.NOT. se_defined(jkind)) CYCLE
         se_kind_a => se_kind_list(ikind)%se_param
         se_kind_b => se_kind_list(jkind)%se_param
         natorb_a = se_natorb(ikind)
         natorb_b = se_natorb(jkind)
         natorb_a2 = natorb_a**2
         natorb_b2 = natorb_b**2

         IF (inode == 1) THEN
            CALL cp_dbcsr_get_block_p(matrix=diagmat_p(1)%matrix, &
                                      row=iatom, col=iatom, BLOCK=pa_block_a, found=found)
            IF(.NOT.(ASSOCIATED(pa_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",242)
            check = (SIZE(pa_block_a, 1) == natorb_a) .AND. (SIZE(pa_block_a, 2) == natorb_a)
            IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",244)
            CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(1)%matrix, &
                                      row=iatom, col=iatom, BLOCK=ksa_block_a, found=found)
            IF(.NOT.(ASSOCIATED(ksa_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",247)
            p_block_tot_a(1:natorb_a, 1:natorb_a) = 2.0_dp*pa_block_a
            pa_a(1:natorb_a2) = RESHAPE(pa_block_a, (/natorb_a2/))
            IF (nspins == 2) THEN
               CALL cp_dbcsr_get_block_p(matrix=diagmat_p(2)%matrix, &
                                         row=iatom, col=iatom, BLOCK=pa_block_b, found=found)
               IF(.NOT.(ASSOCIATED(pa_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",253)
               check = (SIZE(pa_block_b, 1) == natorb_a) .AND. (SIZE(pa_block_b, 2) == natorb_a)
               IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",255)
               CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(2)%matrix, &
                                         row=iatom, col=iatom, BLOCK=ksa_block_b, found=found)
               IF(.NOT.(ASSOCIATED(ksa_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",258)
               p_block_tot_a(1:natorb_a, 1:natorb_a) = pa_block_a+pa_block_b
               pa_b(1:natorb_a2) = RESHAPE(pa_block_b, (/natorb_a2/))
            END IF

         ENDIF

         dr1 = DOT_PRODUCT(rij, rij)
         IF (dr1 > rij_threshold) THEN
            ! Determine the order of the atoms, and in case switch them..
            IF (iatom <= jatom) THEN
               switch = .FALSE.
            ELSE
               switch = .TRUE.
            END IF
            ! Retrieve blocks for KS and P
            CALL cp_dbcsr_get_block_p(matrix=diagmat_p(1)%matrix, &
                                      row=jatom, col=jatom, BLOCK=pb_block_a, found=found)
            IF(.NOT.(ASSOCIATED(pb_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",276)
            check = (SIZE(pb_block_a, 1) == natorb_b) .AND. (SIZE(pb_block_a, 2) == natorb_b)
            IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",278)
            CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(1)%matrix, &
                                      row=jatom, col=jatom, BLOCK=ksb_block_a, found=found)
            IF(.NOT.(ASSOCIATED(ksb_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",281)
            p_block_tot_b(1:natorb_b, 1:natorb_b) = 2.0_dp*pb_block_a
            pb_a(1:natorb_b2) = RESHAPE(pb_block_a, (/natorb_b2/))
            ! Handle more than one configuration
            IF (nspins == 2) THEN
               CALL cp_dbcsr_get_block_p(matrix=diagmat_p(2)%matrix, &
                                         row=jatom, col=jatom, BLOCK=pb_block_b, found=found)
               IF(.NOT.(ASSOCIATED(pb_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",288)
               check = (SIZE(pb_block_b, 1) == natorb_b) .AND. (SIZE(pb_block_b, 2) == natorb_b)
               IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",290)
               CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(2)%matrix, &
                                         row=jatom, col=jatom, BLOCK=ksb_block_b, found=found)
               IF(.NOT.(ASSOCIATED(ksb_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",293)
               check = (SIZE(pb_block_a, 1) == SIZE(pb_block_b, 1)) .AND. (SIZE(pb_block_a, 2) == SIZE(pb_block_b, 2))
               IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",295)
               p_block_tot_b(1:natorb_b, 1:natorb_b) = pb_block_a+pb_block_b
               pb_b(1:natorb_b2) = RESHAPE(pb_block_b, (/natorb_b2/))
            END IF

            SELECT CASE (dft_control%qs_control%method_id)
            CASE (do_method_mndo, do_method_am1, do_method_pm3, do_method_pm6, do_method_pdg, &
                  do_method_rm1, do_method_mndod, do_method_pnnl)

               ! Two-centers One-electron terms
               IF (nspins == 1) THEN
                  ecab = 0._dp
                  CALL fock2_1el(se_kind_a, se_kind_b, rij, ksa_block_a, ksb_block_a, &
                                 pa_a, pb_a, ecore=ecab, itype=itype, anag=anag, se_int_control=se_int_control, &
                                 se_taper=se_taper, store_int_env=store_int_env)
                  ecore2 = ecore2+ecab(1)+ecab(2)
               ELSE IF (nspins == 2) THEN
                  ecab = 0._dp
                  CALL fock2_1el(se_kind_a, se_kind_b, rij, ksa_block_a, ksb_block_a, &
                                 pa_block_a, pb_block_a, ecore=ecab, itype=itype, anag=anag, &
                                 se_int_control=se_int_control, se_taper=se_taper, store_int_env=store_int_env)
                  CALL fock2_1el(se_kind_a, se_kind_b, rij, ksa_block_b, ksb_block_b, &
                                 pa_b, pb_b, ecore=ecab, itype=itype, anag=anag, se_int_control=se_int_control, &
                                 se_taper=se_taper, store_int_env=store_int_env)
                  ecore2 = ecore2+ecab(1)+ecab(2)
               END IF
               IF (atener) THEN
                  atprop%atecoul(iatom) = atprop%atecoul(iatom)+ecab(1)
                  atprop%atecoul(jatom) = atprop%atecoul(jatom)+ecab(2)
               END IF
               ! Coulomb Terms
               IF (nspins == 1) THEN
                  CALL fock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, ksa_block_a, p_block_tot_b, &
                              ksb_block_a, factor=0.5_dp, anag=anag, se_int_control=se_int_control, se_taper=se_taper, &
                              store_int_env=store_int_env)
               ELSE IF (nspins == 2) THEN
                  CALL fock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, ksa_block_a, p_block_tot_b, &
                              ksb_block_a, factor=1.0_dp, anag=anag, se_int_control=se_int_control, se_taper=se_taper, &
                              store_int_env=store_int_env)

                  CALL fock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, ksa_block_b, p_block_tot_b, &
                              ksb_block_b, factor=1.0_dp, anag=anag, se_int_control=se_int_control, se_taper=se_taper, &
                              store_int_env=store_int_env)
               END IF

               IF (calculate_forces) THEN
                  atom_a = atom_of_kind(iatom)
                  atom_b = atom_of_kind(jatom)

                  ! Derivatives of the Two-centre One-electron terms
                  force_ab = 0.0_dp
                  IF (nspins == 1) THEN
                     CALL dfock2_1el(se_kind_a, se_kind_b, rij, pa_a, pb_a, itype=itype, anag=anag, &
                                     se_int_control=se_int_control, se_taper=se_taper, force=force_ab, &
                                     delta=delta)
                  ELSE IF (nspins == 2) THEN
                     CALL dfock2_1el(se_kind_a, se_kind_b, rij, pa_block_a, pb_block_a, itype=itype, anag=anag, &
                                     se_int_control=se_int_control, se_taper=se_taper, force=force_ab, delta=delta)
                     CALL dfock2_1el(se_kind_a, se_kind_b, rij, pa_b, pb_b, itype=itype, anag=anag, &
                                     se_int_control=se_int_control, se_taper=se_taper, force=force_ab, delta=delta)
                  END IF
                  IF (use_virial) THEN
                     CALL virial_pair_force(virial%pv_virial, -1.0_dp, force_ab, rij)
                  END IF

                  ! Sum up force components
                  force(ikind)%all_potential(1, atom_a) = force(ikind)%all_potential(1, atom_a)-force_ab(1)
                  force(jkind)%all_potential(1, atom_b) = force(jkind)%all_potential(1, atom_b)+force_ab(1)

                  force(ikind)%all_potential(2, atom_a) = force(ikind)%all_potential(2, atom_a)-force_ab(2)
                  force(jkind)%all_potential(2, atom_b) = force(jkind)%all_potential(2, atom_b)+force_ab(2)

                  force(ikind)%all_potential(3, atom_a) = force(ikind)%all_potential(3, atom_a)-force_ab(3)
                  force(jkind)%all_potential(3, atom_b) = force(jkind)%all_potential(3, atom_b)+force_ab(3)

                  ! Derivatives of the Coulomb Terms
                  force_ab = 0._dp
                  IF (nspins == 1) THEN
                     CALL dfock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, p_block_tot_b, factor=0.25_dp, &
                                  anag=anag, se_int_control=se_int_control, se_taper=se_taper, force=force_ab, delta=delta)
                  ELSE IF (nspins == 2) THEN
                     CALL dfock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, p_block_tot_b, factor=0.50_dp, &
                                  anag=anag, se_int_control=se_int_control, se_taper=se_taper, force=force_ab, delta=delta)

                     CALL dfock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, p_block_tot_b, factor=0.50_dp, &
                                  anag=anag, se_int_control=se_int_control, se_taper=se_taper, force=force_ab, delta=delta)
                  END IF
                  IF (switch) THEN
                     force_ab(1) = -force_ab(1)
                     force_ab(2) = -force_ab(2)
                     force_ab(3) = -force_ab(3)
                  END IF
                  IF (use_virial) THEN
                     CALL virial_pair_force(virial%pv_virial, -1.0_dp, force_ab, rij)
                  END IF
                  ! Sum up force components
                  force(ikind)%rho_elec(1, atom_a) = force(ikind)%rho_elec(1, atom_a)-force_ab(1)
                  force(jkind)%rho_elec(1, atom_b) = force(jkind)%rho_elec(1, atom_b)+force_ab(1)

                  force(ikind)%rho_elec(2, atom_a) = force(ikind)%rho_elec(2, atom_a)-force_ab(2)
                  force(jkind)%rho_elec(2, atom_b) = force(jkind)%rho_elec(2, atom_b)+force_ab(2)

                  force(ikind)%rho_elec(3, atom_a) = force(ikind)%rho_elec(3, atom_a)-force_ab(3)
                  force(jkind)%rho_elec(3, atom_b) = force(jkind)%rho_elec(3, atom_b)+force_ab(3)
               END IF
            CASE DEFAULT
               CALL cp__b("se_fock_matrix_coulomb.F",401,"")
            END SELECT
         ELSE
            IF (se_int_control%do_ewald_gks) THEN
               IF(.NOT.(iatom == jatom))CALL cp__a("se_fock_matrix_coulomb.F",405)
               ! Two-centers One-electron terms
               ecores = 0._dp
               IF (nspins == 1) THEN
                  CALL fock2_1el_ew(se_kind_a, rij, ksa_block_a, pa_a, &
                                    ecore=ecores, itype=itype, anag=anag, se_int_control=se_int_control, &
                                    se_taper=se_taper, store_int_env=store_int_env)
               ELSE IF (nspins == 2) THEN
                  CALL fock2_1el_ew(se_kind_a, rij, ksa_block_a, pa_block_a, &
                                    ecore=ecores, itype=itype, anag=anag, se_int_control=se_int_control, &
                                    se_taper=se_taper, store_int_env=store_int_env)
                  CALL fock2_1el_ew(se_kind_a, rij, ksa_block_b, pa_b, &
                                    ecore=ecores, itype=itype, anag=anag, se_int_control=se_int_control, &
                                    se_taper=se_taper, store_int_env=store_int_env)
               END IF
               ecore2 = ecore2+ecores
               IF (atener) THEN
                  atprop%atecoul(iatom) = atprop%atecoul(iatom)+ecores
               END IF
               ! Coulomb Terms
               IF (nspins == 1) THEN
                  CALL fock2C_ew(se_kind_a, rij, p_block_tot_a, ksa_block_a, &
                                 factor=0.5_dp, anag=anag, se_int_control=se_int_control, se_taper=se_taper, &
                                 store_int_env=store_int_env)
               ELSE IF (nspins == 2) THEN
                  CALL fock2C_ew(se_kind_a, rij, p_block_tot_a, ksa_block_a, &
                                 factor=1.0_dp, anag=anag, se_int_control=se_int_control, se_taper=se_taper, &
                                 store_int_env=store_int_env)
                  CALL fock2C_ew(se_kind_a, rij, p_block_tot_a, ksa_block_b, &
                                 factor=1.0_dp, anag=anag, se_int_control=se_int_control, se_taper=se_taper, &
                                 store_int_env=store_int_env)
               END IF
            END IF
         END IF
      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DEALLOCATE (se_kind_list, se_defined, se_natorb)

      DO ispin = 1, nspins
         CALL cp_dbcsr_sum_replicated(diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_distribute(diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_add(ks_matrix(ispin)%matrix, diagmat_ks(ispin)%matrix, &
                           1.0_dp, 1.0_dp)
      END DO
      CALL cp_dbcsr_deallocate_matrix_set(diagmat_p)
      CALL cp_dbcsr_deallocate_matrix_set(diagmat_ks)

      IF (calculate_forces) THEN
         DEALLOCATE (atom_of_kind)
      END IF

      ! Two-centers one-electron terms
      CALL mp_sum(ecore2, para_env%group)
      energy%hartree = ecore2-energy%core
!      WRITE ( *, * ) 'IN SE_F_COUL', ecore2, energy%core

      CALL finalize_se_taper(se_taper)

      CALL timestop(handle)
   END SUBROUTINE build_fock_matrix_coulomb

! **************************************************************************************************
!> \brief  Long-Range part for SE Coulomb interactions
!> \param qs_env ...
!> \param ks_matrix ...
!> \param matrix_p ...
!> \param energy ...
!> \param calculate_forces ...
!> \param store_int_env ...
!> \date   08.2008 [created]
!> \author Teodoro Laino [tlaino] - University of Zurich
! **************************************************************************************************
   SUBROUTINE build_fock_matrix_coulomb_lr(qs_env, ks_matrix, matrix_p, energy, &
                                           calculate_forces, store_int_env)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_matrix, matrix_p
      TYPE(qs_energy_type), POINTER                      :: energy
      LOGICAL, INTENT(in)                                :: calculate_forces
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env

      CHARACTER(len=*), PARAMETER :: routineN = 'build_fock_matrix_coulomb_lr', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, ewald_type, forces_g_size, handle, iatom, ikind, ilist, indi, indj, &
         ispin, itype, iw, jint, natoms, natorb_a, nkind, nlocal_particles, node, nparticle_local, &
         nspins, size_1c_int
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind
      LOGICAL                                            :: anag, atener, defined, found, use_virial
      LOGICAL, DIMENSION(3)                              :: task
      REAL(KIND=dp)                                      :: e_neut, e_self, energy_glob, &
                                                            energy_local, enuc, fac, tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: forces_g, forces_r
      REAL(KIND=dp), DIMENSION(3)                        :: force_a
      REAL(KIND=dp), DIMENSION(3, 3)                     :: pv_glob, pv_local, qcart
      REAL(KIND=dp), DIMENSION(5)                        :: qsph
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ksa_block_a, pa_block_a
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: diagmat_ks, diagmat_p
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(fist_nonbond_env_type), POINTER               :: se_nonbond_env
      TYPE(nddo_mpole_type), POINTER                     :: se_nddo_mpole
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: se_section
      TYPE(semi_empirical_control_type), POINTER         :: se_control
      TYPE(semi_empirical_mpole_type), POINTER           :: mpole
      TYPE(semi_empirical_type), POINTER                 :: se_kind_a
      TYPE(virial_type), POINTER                         :: virial

      CALL timeset(routineN, handle)

      NULLIFY (dft_control, cell, force, particle_set, diagmat_ks, diagmat_p, local_particles, &
               se_control, ewald_env, ewald_pw, se_nddo_mpole, se_nonbond_env, se_section, mpole, &
               logger, virial, atprop)

      logger => cp_get_default_logger()
      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, cell=cell, para_env=para_env, &
                      atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set, particle_set=particle_set, &
                      ewald_env=ewald_env, &
                      local_particles=local_particles, ewald_pw=ewald_pw, se_nddo_mpole=se_nddo_mpole, &
                      se_nonbond_env=se_nonbond_env, virial=virial, atprop=atprop)

      nlocal_particles = SUM(local_particles%n_el(:))
      natoms = SIZE(particle_set)
      CALL ewald_env_get(ewald_env, ewald_type=ewald_type)
      SELECT CASE (ewald_type)
      CASE (do_ewald_ewald)
         forces_g_size = nlocal_particles
      CASE DEFAULT
         CALL cp__b("se_fock_matrix_coulomb.F",544,"Periodic SE implemented only for standard EWALD sums.")
      END SELECT

      ! Parameters
      se_section => section_vals_get_subs_vals(qs_env%input, "DFT%QS%SE")
      se_control => dft_control%qs_control%se_control
      anag = se_control%analytical_gradients
      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer) .AND. calculate_forces
      atener = atprop%energy

      nspins = dft_control%nspins
      IF(.NOT.(ASSOCIATED(matrix_p)))CALL cp__a("se_fock_matrix_coulomb.F",555)
      IF(.NOT.(SIZE(ks_matrix) > 0))CALL cp__a("se_fock_matrix_coulomb.F",556)

      CALL cp_dbcsr_allocate_matrix_set(diagmat_ks, nspins)
      CALL cp_dbcsr_allocate_matrix_set(diagmat_p, nspins)

      nkind = SIZE(atomic_kind_set)
      DO ispin = 1, nspins
         ! Allocate diagonal block matrices
         ALLOCATE (diagmat_p(ispin)%matrix, diagmat_ks(ispin)%matrix) !sm->dbcsr
         CALL cp_dbcsr_init(diagmat_p(ispin)%matrix)
         CALL cp_dbcsr_init(diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_get_block_diag(matrix_p(ispin)%matrix, diagmat_p(ispin)%matrix)
         CALL cp_dbcsr_get_block_diag(ks_matrix(ispin)%matrix, diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_set(diagmat_ks(ispin)%matrix, 0.0_dp)
         CALL cp_dbcsr_replicate_all(diagmat_p(ispin)%matrix)
         CALL cp_dbcsr_replicate_all(diagmat_ks(ispin)%matrix)
      END DO

      ! Check for implemented SE methods
      SELECT CASE (dft_control%qs_control%method_id)
      CASE (do_method_mndo, do_method_am1, do_method_pm3, do_method_pm6, do_method_pdg, &
            do_method_rm1, do_method_mndod, do_method_pnnl)
         itype = get_se_type(dft_control%qs_control%method_id)
      CASE DEFAULT
         CALL cp__b("se_fock_matrix_coulomb.F",580,"")
      END SELECT

      ! Zero arrays and possibly build neighbor lists
      energy_local = 0.0_dp
      energy_glob = 0.0_dp
      e_neut = 0.0_dp
      e_self = 0.0_dp
      task = .FALSE.
      SELECT CASE (se_control%max_multipole)
      CASE (do_multipole_none)
         ! Do Nothing
      CASE (do_multipole_charge)
         task(1) = .TRUE.
      CASE (do_multipole_dipole)
         task = .TRUE.
         task(3) = .FALSE.
      CASE (do_multipole_quadrupole)
         task = .TRUE.
      CASE DEFAULT
         CALL cp__b("se_fock_matrix_coulomb.F",600,"")
      END SELECT

      ! Build-up neighbor lists for real-space part of Ewald multipoles
      CALL list_control(atomic_kind_set, particle_set, local_particles, &
                        cell, se_nonbond_env, para_env, se_section)

      enuc = 0.0_dp
      energy%core_overlap = 0.0_dp
      se_nddo_mpole%charge = 0.0_dp
      se_nddo_mpole%dipole = 0.0_dp
      se_nddo_mpole%quadrupole = 0.0_dp

      DO ispin = 1, nspins
         ! Compute the NDDO mpole expansion
         DO ikind = 1, nkind
            CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind_a)
            CALL get_se_param(se_kind_a, defined=defined, natorb=natorb_a)

            IF (.NOT. defined .OR. natorb_a < 1) CYCLE

            nparticle_local = local_particles%n_el(ikind)
            DO ilist = 1, nparticle_local
               iatom = local_particles%list(ikind)%array(ilist)
               CALL cp_dbcsr_get_block_p(matrix=diagmat_p(ispin)%matrix, &
                                         row=iatom, col=iatom, BLOCK=pa_block_a, found=found)
               IF(.NOT.(ASSOCIATED(pa_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",626)
               ! Nuclei
               IF (task(1) .AND. ispin == 1) se_nddo_mpole%charge(iatom) = se_kind_a%zeff
               ! Electrons
               size_1c_int = SIZE(se_kind_a%w_mpole)
               DO jint = 1, size_1c_int
                  mpole => se_kind_a%w_mpole(jint)%mpole
                  indi = se_orbital_pointer(mpole%indi)
                  indj = se_orbital_pointer(mpole%indj)
                  fac = 1.0_dp
                  IF (indi /= indj) fac = 2.0_dp

                  ! Charge
                  IF (mpole%task(1) .AND. task(1)) THEN
                     se_nddo_mpole%charge(iatom) = se_nddo_mpole%charge(iatom)+ &
                                                   fac*pa_block_a(indi, indj)*mpole%c
                  END IF

                  ! Dipole
                  IF (mpole%task(2) .AND. task(2)) THEN
                     se_nddo_mpole%dipole(:, iatom) = se_nddo_mpole%dipole(:, iatom)+ &
                                                      fac*pa_block_a(indi, indj)*mpole%d(:)
                  END IF

                  ! Quadrupole
                  IF (mpole%task(3) .AND. task(3)) THEN
                     qsph = fac*mpole%qs*pa_block_a(indi, indj)
                     CALL quadrupole_sph_to_cart(qcart, qsph)
                     se_nddo_mpole%quadrupole(:, :, iatom) = se_nddo_mpole%quadrupole(:, :, iatom)+ &
                                                             qcart
                  END IF
               END DO
               ! Print some info about charge, dipole and quadrupole (debug purpose only)
               IF (debug_this_module) THEN
                  WRITE (*, '(I5,F12.6,5X,3F12.6,5X,9F12.6)') iatom, se_nddo_mpole%charge(iatom), &
                     se_nddo_mpole%dipole(:, iatom), se_nddo_mpole%quadrupole(:, :, iatom)
               END IF
            END DO
         END DO
      END DO
      CALL mp_sum(se_nddo_mpole%charge, para_env%group)
      CALL mp_sum(se_nddo_mpole%dipole, para_env%group)
      CALL mp_sum(se_nddo_mpole%quadrupole, para_env%group)

      ! Initialize for virial
      IF (use_virial) THEN
         pv_glob = 0.0_dp
         pv_local = 0.0_dp
      END IF

      ! Ewald Multipoles Sum
      iw = cp_print_key_unit_nr(logger, se_section, "PRINT%EWALD_INFO", extension=".seLog")
      IF (calculate_forces) THEN
         CALL get_qs_env(qs_env=qs_env, force=force)

         ! Allocate atom index for kind
         ALLOCATE (atom_of_kind(natoms))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind)

         ! Allocate and zeroing arrays
         ALLOCATE (forces_g(3, forces_g_size))
         ALLOCATE (forces_r(3, natoms))
         forces_g = 0.0_dp
         forces_r = 0.0_dp
         CALL ewald_multipole_evaluate( &
            ewald_env, ewald_pw, se_nonbond_env, cell, &
            particle_set, local_particles, energy_local, energy_glob, e_neut, e_self, task, &
            do_correction_bonded=.FALSE., do_forces=.TRUE., do_stress=use_virial, do_efield=.TRUE., &
            charges=se_nddo_mpole%charge, dipoles=se_nddo_mpole%dipole, quadrupoles=se_nddo_mpole%quadrupole, &
            forces_local=forces_g, forces_glob=forces_r, pv_glob=pv_glob, pv_local=pv_local, &
            efield0=se_nddo_mpole%efield0, efield1=se_nddo_mpole%efield1, efield2=se_nddo_mpole%efield2, iw=iw, &
            do_debug=.TRUE.)
         ! Only SR force have to be summed up.. the one in g-space are already fully local..
         CALL mp_sum(forces_r, para_env%group)
      ELSE
         CALL ewald_multipole_evaluate( &
            ewald_env, ewald_pw, se_nonbond_env, cell, &
            particle_set, local_particles, energy_local, energy_glob, e_neut, e_self, task, &
            do_correction_bonded=.FALSE., do_forces=.FALSE., do_stress=.FALSE., do_efield=.TRUE., &
            charges=se_nddo_mpole%charge, dipoles=se_nddo_mpole%dipole, quadrupoles=se_nddo_mpole%quadrupole, &
            efield0=se_nddo_mpole%efield0, efield1=se_nddo_mpole%efield1, efield2=se_nddo_mpole%efield2, &
            iw=iw, do_debug=.TRUE.)
      END IF
      CALL cp_print_key_finished_output(iw, logger, se_section, "PRINT%EWALD_INFO")

      ! Apply correction only when the Integral Scheme is different from Slater
      IF ((se_control%integral_screening /= do_se_IS_slater) .AND. (.NOT. debug_this_module)) THEN
         CALL build_fock_matrix_coul_lrc(qs_env, ks_matrix, matrix_p, energy, calculate_forces, &
                                         store_int_env, se_nddo_mpole, task, diagmat_p, diagmat_ks, virial, &
                                         pv_glob)
      END IF

      ! Virial for the long-range part and correction
      IF (use_virial) THEN
         ! Sum up contribution of pv_glob on each thread and keep only one copy of pv_local
         virial%pv_virial = virial%pv_virial+pv_glob
         IF (para_env%mepos == para_env%source) THEN
            virial%pv_virial = virial%pv_virial+pv_local
         END IF
      END IF

      ! Debug Statements
      IF (debug_this_module) THEN
         CALL mp_sum(energy_glob, para_env%group)
         WRITE (*, *) "TOTAL ENERGY AFTER EWALD:", energy_local+energy_glob+e_neut+e_self, &
            energy_local, energy_glob, e_neut, e_self
      END IF

      ! Modify the KS matrix and possibly compute derivatives
      node = 0
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind_a)
         CALL get_se_param(se_kind_a, defined=defined, natorb=natorb_a)

         IF (.NOT. defined .OR. natorb_a < 1) CYCLE

         nparticle_local = local_particles%n_el(ikind)
         DO ilist = 1, nparticle_local
            node = node+1
            iatom = local_particles%list(ikind)%array(ilist)
            DO ispin = 1, nspins
               CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(ispin)%matrix, &
                                         row=iatom, col=iatom, BLOCK=ksa_block_a, found=found)
               IF(.NOT.(ASSOCIATED(ksa_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",749)

               ! Modify Hamiltonian Matrix accordingly potential, field and electric field gradient
               size_1c_int = SIZE(se_kind_a%w_mpole)
               DO jint = 1, size_1c_int
                  tmp = 0.0_dp
                  mpole => se_kind_a%w_mpole(jint)%mpole
                  indi = se_orbital_pointer(mpole%indi)
                  indj = se_orbital_pointer(mpole%indj)

                  ! Charge
                  IF (mpole%task(1) .AND. task(1)) THEN
                     tmp = tmp+mpole%c*se_nddo_mpole%efield0(iatom)
                  END IF

                  ! Dipole
                  IF (mpole%task(2) .AND. task(2)) THEN
                     tmp = tmp-DOT_PRODUCT(mpole%d, se_nddo_mpole%efield1(:, iatom))
                  END IF

                  ! Quadrupole
                  IF (mpole%task(3) .AND. task(3)) THEN
                     tmp = tmp-(1.0_dp/3.0_dp)*SUM(mpole%qc*RESHAPE(se_nddo_mpole%efield2(:, iatom), (/3, 3/)))
                  END IF

                  ksa_block_a(indi, indj) = ksa_block_a(indi, indj)+tmp
                  ksa_block_a(indj, indi) = ksa_block_a(indi, indj)
               END DO
            END DO

            ! Nuclear term and forces
            IF (task(1)) enuc = enuc+se_kind_a%zeff*se_nddo_mpole%efield0(iatom)
            IF (atener) THEN
               atprop%atecoul(iatom) = atprop%atecoul(iatom)+ &
                                       0.5_dp*se_kind_a%zeff*se_nddo_mpole%efield0(iatom)
            END IF
            IF (calculate_forces) THEN
               atom_a = atom_of_kind(iatom)
               force_a = forces_r(1:3, iatom)+forces_g(1:3, node)
               ! Derivatives of the periodic Coulomb Terms
               force(ikind)%all_potential(:, atom_a) = force(ikind)%all_potential(:, atom_a)-force_a(:)
            END IF
         END DO
      END DO
      ! Sum nuclear energy contribution
      CALL mp_sum(enuc, para_env%group)
      energy%core_overlap = energy%core_overlap+energy%core_overlap0+0.5_dp*enuc

      ! Debug Statements
      IF (debug_this_module) THEN
         WRITE (*, *) "ENUC: ", enuc*0.5_dp
      END IF

      DO ispin = 1, nspins
         CALL cp_dbcsr_sum_replicated(diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_distribute(diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_add(ks_matrix(ispin)%matrix, diagmat_ks(ispin)%matrix, &
                           1.0_dp, 1.0_dp)
      END DO
      CALL cp_dbcsr_deallocate_matrix_set(diagmat_p)
      CALL cp_dbcsr_deallocate_matrix_set(diagmat_ks)

      ! Set the Fock matrix contribution to SCP
      IF (calculate_forces) THEN
         DEALLOCATE (atom_of_kind)
         DEALLOCATE (forces_g)
         DEALLOCATE (forces_r)
      END IF

      CALL timestop(handle)
   END SUBROUTINE build_fock_matrix_coulomb_lr

! **************************************************************************************************
!> \brief When doing long-range SE calculation, this module computes the correction
!>        between the mismatch of point-like multipoles and multipoles represented
!>        with charges
!> \param qs_env ...
!> \param ks_matrix ...
!> \param matrix_p ...
!> \param energy ...
!> \param calculate_forces ...
!> \param store_int_env ...
!> \param se_nddo_mpole ...
!> \param task ...
!> \param diagmat_p ...
!> \param diagmat_ks ...
!> \param virial ...
!> \param pv_glob ...
!> \author Teodoro Laino [tlaino] - 05.2009
! **************************************************************************************************
   SUBROUTINE build_fock_matrix_coul_lrc(qs_env, ks_matrix, matrix_p, energy, &
                                         calculate_forces, store_int_env, se_nddo_mpole, task, diagmat_p, diagmat_ks, &
                                         virial, pv_glob)

      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_matrix, matrix_p
      TYPE(qs_energy_type), POINTER                      :: energy
      LOGICAL, INTENT(in)                                :: calculate_forces
      TYPE(semi_empirical_si_type), POINTER              :: store_int_env
      TYPE(nddo_mpole_type), POINTER                     :: se_nddo_mpole
      LOGICAL, DIMENSION(3), INTENT(IN)                  :: task
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: diagmat_p, diagmat_ks
      TYPE(virial_type), POINTER                         :: virial
      REAL(KIND=dp), DIMENSION(3, 3), INTENT(INOUT)      :: pv_glob

      CHARACTER(len=*), PARAMETER :: routineN = 'build_fock_matrix_coul_lrc', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, handle, iatom, ikind, inode, itype, jatom, jkind, natom, &
         natorb_a, natorb_a2, natorb_b, natorb_b2, nkind, nspins, size1, size2
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: se_natorb
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind
      LOGICAL                                            :: anag, atener, check, defined, found, &
                                                            switch, use_virial
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: se_defined
      REAL(KIND=dp)                                      :: delta, dr1, ecore2, enuc, enuclear, &
                                                            ptens11, ptens12, ptens13, ptens21, &
                                                            ptens22, ptens23, ptens31, ptens32, &
                                                            ptens33
      REAL(KIND=dp), DIMENSION(2)                        :: ecab
      REAL(KIND=dp), DIMENSION(2025)                     :: pa_a, pa_b, pb_a, pb_b
      REAL(KIND=dp), DIMENSION(3)                        :: force_ab, force_ab0, rij
      REAL(KIND=dp), DIMENSION(45, 45)                   :: p_block_tot_a, p_block_tot_b
      REAL(KIND=dp), DIMENSION(:), POINTER               :: efield0
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: efield1, efield2, ksa_block_a, ksa_block_b, &
         ksb_block_a, ksb_block_b, pa_block_a, pa_block_b, pb_block_a, pb_block_b
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_lrc
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(se_int_control_type)                          :: se_int_control
      TYPE(se_taper_type), POINTER                       :: se_taper
      TYPE(semi_empirical_control_type), POINTER         :: se_control
      TYPE(semi_empirical_p_type), DIMENSION(:), POINTER :: se_kind_list
      TYPE(semi_empirical_type), POINTER                 :: se_kind_a, se_kind_b

      CALL timeset(routineN, handle)
      NULLIFY (dft_control, cell, force, particle_set, se_control, se_taper, &
               efield0, efield1, efield2, atprop)

      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, cell=cell, se_taper=se_taper, &
                      para_env=para_env, sab_lrc=sab_lrc, atomic_kind_set=atomic_kind_set, &
                      qs_kind_set=qs_kind_set, particle_set=particle_set, atprop=atprop)

      ! Parameters
      CALL initialize_se_taper(se_taper, lr_corr=.TRUE.)
      se_control => dft_control%qs_control%se_control
      anag = se_control%analytical_gradients
      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer) .AND. calculate_forces
      atener = atprop%energy

      CALL setup_se_int_control_type(se_int_control, do_ewald_r3=se_control%do_ewald_r3, &
                                     do_ewald_gks=.FALSE., integral_screening=se_control%integral_screening, &
                                     shortrange=.FALSE., max_multipole=se_control%max_multipole, &
                                     pc_coulomb_int=.TRUE.)

      nspins = dft_control%nspins
      IF(.NOT.(ASSOCIATED(matrix_p)))CALL cp__a("se_fock_matrix_coulomb.F",914)
      IF(.NOT.(SIZE(ks_matrix) > 0))CALL cp__a("se_fock_matrix_coulomb.F",915)
      IF(.NOT.(ASSOCIATED(diagmat_p)))CALL cp__a("se_fock_matrix_coulomb.F",916)
      IF(.NOT.(ASSOCIATED(diagmat_ks)))CALL cp__a("se_fock_matrix_coulomb.F",917)

      nkind = SIZE(atomic_kind_set)
      IF (calculate_forces) THEN
         CALL get_qs_env(qs_env=qs_env, force=force)
         natom = SIZE(particle_set)
         delta = se_control%delta
         ! Allocate atom index for kind
         ALLOCATE (atom_of_kind(natom))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind)
      END IF

      ! Allocate arrays for storing partial information on potential, field, field gradient
      size1 = SIZE(se_nddo_mpole%efield0)
      ALLOCATE (efield0(size1))
      efield0 = 0.0_dp
      size1 = SIZE(se_nddo_mpole%efield1, 1)
      size2 = SIZE(se_nddo_mpole%efield1, 2)
      ALLOCATE (efield1(size1, size2))
      efield1 = 0.0_dp
      size1 = SIZE(se_nddo_mpole%efield2, 1)
      size2 = SIZE(se_nddo_mpole%efield2, 2)
      ALLOCATE (efield2(size1, size2))
      efield2 = 0.0_dp

      ! Initialize if virial is requested
      IF (use_virial) THEN
         ptens11 = 0.0_dp; ptens12 = 0.0_dp; ptens13 = 0.0_dp
         ptens21 = 0.0_dp; ptens22 = 0.0_dp; ptens23 = 0.0_dp
         ptens31 = 0.0_dp; ptens32 = 0.0_dp; ptens33 = 0.0_dp
      END IF

      ! Start of the loop for the correction of the pair interactions
      ecore2 = 0.0_dp
      enuclear = 0.0_dp
      itype = get_se_type(dft_control%qs_control%method_id)

      ALLOCATE (se_defined(nkind), se_kind_list(nkind), se_natorb(nkind))
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind_a)
         se_kind_list(ikind)%se_param => se_kind_a
         CALL get_se_param(se_kind_a, defined=defined, natorb=natorb_a)
         se_defined(ikind) = (defined .AND. natorb_a >= 1)
         se_natorb(ikind) = natorb_a
      END DO

      CALL neighbor_list_iterator_create(nl_iterator, sab_lrc)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, jatom=jatom, inode=inode, r=rij)
         IF (.NOT. se_defined(ikind)) CYCLE
         IF (.NOT. se_defined(jkind)) CYCLE
         se_kind_a => se_kind_list(ikind)%se_param
         se_kind_b => se_kind_list(jkind)%se_param
         natorb_a = se_natorb(ikind)
         natorb_b = se_natorb(jkind)
         natorb_a2 = natorb_a**2
         natorb_b2 = natorb_b**2

         IF (inode == 1) THEN
            CALL cp_dbcsr_get_block_p(matrix=diagmat_p(1)%matrix, &
                                      row=iatom, col=iatom, BLOCK=pa_block_a, found=found)
            IF(.NOT.(ASSOCIATED(pa_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",978)
            check = (SIZE(pa_block_a, 1) == natorb_a) .AND. (SIZE(pa_block_a, 2) == natorb_a)
            IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",980)
            CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(1)%matrix, &
                                      row=iatom, col=iatom, BLOCK=ksa_block_a, found=found)
            IF(.NOT.(ASSOCIATED(ksa_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",983)
            p_block_tot_a(1:natorb_a, 1:natorb_a) = 2.0_dp*pa_block_a
            pa_a(1:natorb_a2) = RESHAPE(pa_block_a, (/natorb_a2/))
            IF (nspins == 2) THEN
               CALL cp_dbcsr_get_block_p(matrix=diagmat_p(2)%matrix, &
                                         row=iatom, col=iatom, BLOCK=pa_block_b, found=found)
               IF(.NOT.(ASSOCIATED(pa_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",989)
               check = (SIZE(pa_block_b, 1) == natorb_a) .AND. (SIZE(pa_block_b, 2) == natorb_a)
               IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",991)
               CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(2)%matrix, &
                                         row=iatom, col=iatom, BLOCK=ksa_block_b, found=found)
               IF(.NOT.(ASSOCIATED(ksa_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",994)
               p_block_tot_a(1:natorb_a, 1:natorb_a) = pa_block_a+pa_block_b
               pa_b(1:natorb_a2) = RESHAPE(pa_block_b, (/natorb_a2/))
            END IF

         ENDIF

         dr1 = DOT_PRODUCT(rij, rij)
         IF (dr1 > rij_threshold) THEN
            ! Determine the order of the atoms, and in case switch them..
            IF (iatom <= jatom) THEN
               switch = .FALSE.
            ELSE
               switch = .TRUE.
            END IF

            ! Point-like interaction corrections
            CALL se_coulomb_ij_interaction(iatom, jatom, task, do_forces=calculate_forces, &
                                           do_efield=.TRUE., do_stress=use_virial, charges=se_nddo_mpole%charge, &
                                           dipoles=se_nddo_mpole%dipole, quadrupoles=se_nddo_mpole%quadrupole, &
                                           force_ab=force_ab0, efield0=efield0, efield1=efield1, efield2=efield2, &
                                           rab2=dr1, rab=rij, ptens11=ptens11, ptens12=ptens12, ptens13=ptens13, &
                                           ptens21=ptens21, ptens22=ptens22, ptens23=ptens23, ptens31=ptens31, &
                                           ptens32=ptens32, ptens33=ptens33)

            ! Retrieve blocks for KS and P
            CALL cp_dbcsr_get_block_p(matrix=diagmat_p(1)%matrix, &
                                      row=jatom, col=jatom, BLOCK=pb_block_a, found=found)
            IF(.NOT.(ASSOCIATED(pb_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",1022)
            check = (SIZE(pb_block_a, 1) == natorb_b) .AND. (SIZE(pb_block_a, 2) == natorb_b)
            IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",1024)
            CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(1)%matrix, &
                                      row=jatom, col=jatom, BLOCK=ksb_block_a, found=found)
            IF(.NOT.(ASSOCIATED(ksb_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",1027)
            p_block_tot_b(1:natorb_b, 1:natorb_b) = 2.0_dp*pb_block_a
            pb_a(1:natorb_b2) = RESHAPE(pb_block_a, (/natorb_b2/))
            ! Handle more than one configuration
            IF (nspins == 2) THEN
               CALL cp_dbcsr_get_block_p(matrix=diagmat_p(2)%matrix, &
                                         row=jatom, col=jatom, BLOCK=pb_block_b, found=found)
               IF(.NOT.(ASSOCIATED(pb_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",1034)
               check = (SIZE(pb_block_b, 1) == natorb_b) .AND. (SIZE(pb_block_b, 2) == natorb_b)
               IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",1036)
               CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(2)%matrix, &
                                         row=jatom, col=jatom, BLOCK=ksb_block_b, found=found)
               IF(.NOT.(ASSOCIATED(ksb_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",1039)
               check = (SIZE(pb_block_a, 1) == SIZE(pb_block_b, 1)) .AND. (SIZE(pb_block_a, 2) == SIZE(pb_block_b, 2))
               IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",1041)
               p_block_tot_b(1:natorb_b, 1:natorb_b) = pb_block_a+pb_block_b
               pb_b(1:natorb_b2) = RESHAPE(pb_block_b, (/natorb_b2/))
            END IF

            SELECT CASE (dft_control%qs_control%method_id)
            CASE (do_method_mndo, do_method_am1, do_method_pm3, do_method_pm6, do_method_pdg, &
                  do_method_rm1, do_method_mndod)
               ! Evaluate nuclear contribution..
               CALL corecore_el(se_kind_a, se_kind_b, rij, enuc=enuc, itype=itype, anag=anag, &
                                se_int_control=se_int_control, se_taper=se_taper)
               enuclear = enuclear+enuc

               ! Two-centers One-electron terms
               IF (nspins == 1) THEN
                  ecab = 0._dp
                  CALL fock2_1el(se_kind_a, se_kind_b, rij, ksa_block_a, ksb_block_a, &
                                 pa_a, pb_a, ecore=ecab, itype=itype, anag=anag, se_int_control=se_int_control, &
                                 se_taper=se_taper, store_int_env=store_int_env)
                  ecore2 = ecore2+ecab(1)+ecab(2)
               ELSE IF (nspins == 2) THEN
                  ecab = 0._dp
                  CALL fock2_1el(se_kind_a, se_kind_b, rij, ksa_block_a, ksb_block_a, &
                                 pa_block_a, pb_block_a, ecore=ecab, itype=itype, anag=anag, &
                                 se_int_control=se_int_control, se_taper=se_taper, store_int_env=store_int_env)
                  CALL fock2_1el(se_kind_a, se_kind_b, rij, ksa_block_b, ksb_block_b, &
                                 pa_b, pb_b, ecore=ecab, itype=itype, anag=anag, se_int_control=se_int_control, &
                                 se_taper=se_taper, store_int_env=store_int_env)
                  ecore2 = ecore2+ecab(1)+ecab(2)
               END IF
               IF (atener) THEN
                  atprop%atecoul(iatom) = atprop%atecoul(iatom)+ecab(1)+0.5_dp*enuc
                  atprop%atecoul(jatom) = atprop%atecoul(jatom)+ecab(2)+0.5_dp*enuc
               END IF
               ! Coulomb Terms
               IF (nspins == 1) THEN
                  CALL fock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, ksa_block_a, p_block_tot_b, &
                              ksb_block_a, factor=0.5_dp, anag=anag, se_int_control=se_int_control, se_taper=se_taper, &
                              store_int_env=store_int_env)
               ELSE IF (nspins == 2) THEN
                  CALL fock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, ksa_block_a, p_block_tot_b, &
                              ksb_block_a, factor=1.0_dp, anag=anag, se_int_control=se_int_control, se_taper=se_taper, &
                              store_int_env=store_int_env)

                  CALL fock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, ksa_block_b, p_block_tot_b, &
                              ksb_block_b, factor=1.0_dp, anag=anag, se_int_control=se_int_control, se_taper=se_taper, &
                              store_int_env=store_int_env)
               END IF

               IF (calculate_forces) THEN
                  atom_a = atom_of_kind(iatom)
                  atom_b = atom_of_kind(jatom)

                  ! Evaluate nuclear contribution..
                  CALL dcorecore_el(se_kind_a, se_kind_b, rij, denuc=force_ab, itype=itype, delta=delta, &
                                    anag=anag, se_int_control=se_int_control, se_taper=se_taper)

                  ! Derivatives of the Two-centre One-electron terms
                  IF (nspins == 1) THEN
                     CALL dfock2_1el(se_kind_a, se_kind_b, rij, pa_a, pb_a, itype=itype, anag=anag, &
                                     se_int_control=se_int_control, se_taper=se_taper, force=force_ab, &
                                     delta=delta)
                  ELSE IF (nspins == 2) THEN
                     CALL dfock2_1el(se_kind_a, se_kind_b, rij, pa_block_a, pb_block_a, itype=itype, anag=anag, &
                                     se_int_control=se_int_control, se_taper=se_taper, force=force_ab, delta=delta)
                     CALL dfock2_1el(se_kind_a, se_kind_b, rij, pa_b, pb_b, itype=itype, anag=anag, &
                                     se_int_control=se_int_control, se_taper=se_taper, force=force_ab, delta=delta)
                  END IF
                  IF (use_virial) THEN
                     CALL virial_pair_force(virial%pv_virial, -1.0_dp, force_ab, rij)
                  END IF
                  force_ab = force_ab+force_ab0

                  ! Sum up force components
                  force(ikind)%all_potential(1, atom_a) = force(ikind)%all_potential(1, atom_a)-force_ab(1)
                  force(jkind)%all_potential(1, atom_b) = force(jkind)%all_potential(1, atom_b)+force_ab(1)

                  force(ikind)%all_potential(2, atom_a) = force(ikind)%all_potential(2, atom_a)-force_ab(2)
                  force(jkind)%all_potential(2, atom_b) = force(jkind)%all_potential(2, atom_b)+force_ab(2)

                  force(ikind)%all_potential(3, atom_a) = force(ikind)%all_potential(3, atom_a)-force_ab(3)
                  force(jkind)%all_potential(3, atom_b) = force(jkind)%all_potential(3, atom_b)+force_ab(3)

                  ! Derivatives of the Coulomb Terms
                  force_ab = 0._dp
                  IF (nspins == 1) THEN
                     CALL dfock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, p_block_tot_b, factor=0.25_dp, &
                                  anag=anag, se_int_control=se_int_control, se_taper=se_taper, force=force_ab, delta=delta)
                  ELSE IF (nspins == 2) THEN
                     CALL dfock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, p_block_tot_b, factor=0.50_dp, &
                                  anag=anag, se_int_control=se_int_control, se_taper=se_taper, force=force_ab, delta=delta)

                     CALL dfock2C(se_kind_a, se_kind_b, rij, switch, p_block_tot_a, p_block_tot_b, factor=0.50_dp, &
                                  anag=anag, se_int_control=se_int_control, se_taper=se_taper, force=force_ab, delta=delta)
                  END IF
                  IF (switch) THEN
                     force_ab(1) = -force_ab(1)
                     force_ab(2) = -force_ab(2)
                     force_ab(3) = -force_ab(3)
                  END IF
                  IF (use_virial) THEN
                     CALL virial_pair_force(virial%pv_virial, -1.0_dp, force_ab, rij)
                  END IF

                  ! Sum up force components
                  force(ikind)%rho_elec(1, atom_a) = force(ikind)%rho_elec(1, atom_a)-force_ab(1)
                  force(jkind)%rho_elec(1, atom_b) = force(jkind)%rho_elec(1, atom_b)+force_ab(1)

                  force(ikind)%rho_elec(2, atom_a) = force(ikind)%rho_elec(2, atom_a)-force_ab(2)
                  force(jkind)%rho_elec(2, atom_b) = force(jkind)%rho_elec(2, atom_b)+force_ab(2)

                  force(ikind)%rho_elec(3, atom_a) = force(ikind)%rho_elec(3, atom_a)-force_ab(3)
                  force(jkind)%rho_elec(3, atom_b) = force(jkind)%rho_elec(3, atom_b)+force_ab(3)
               END IF
            CASE DEFAULT
               CALL cp__b("se_fock_matrix_coulomb.F",1156,"")
            END SELECT
         END IF
      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DEALLOCATE (se_kind_list, se_defined, se_natorb)

      ! Sum-up Virial constribution (long-range correction)
      IF (use_virial) THEN
         pv_glob(1, 1) = pv_glob(1, 1)+ptens11
         pv_glob(1, 2) = pv_glob(1, 2)+(ptens12+ptens21)*0.5_dp
         pv_glob(1, 3) = pv_glob(1, 3)+(ptens13+ptens31)*0.5_dp
         pv_glob(2, 1) = pv_glob(1, 2)
         pv_glob(2, 2) = pv_glob(2, 2)+ptens22
         pv_glob(2, 3) = pv_glob(2, 3)+(ptens23+ptens32)*0.5_dp
         pv_glob(3, 1) = pv_glob(1, 3)
         pv_glob(3, 2) = pv_glob(2, 3)
         pv_glob(3, 3) = pv_glob(3, 3)+ptens33
      END IF

      IF (calculate_forces) THEN
         DEALLOCATE (atom_of_kind)
      END IF

      ! collect information on potential, field, field gradient
      CALL mp_sum(efield0, para_env%group)
      CALL mp_sum(efield1, para_env%group)
      CALL mp_sum(efield2, para_env%group)
      se_nddo_mpole%efield0 = se_nddo_mpole%efield0-efield0
      se_nddo_mpole%efield1 = se_nddo_mpole%efield1-efield1
      se_nddo_mpole%efield2 = se_nddo_mpole%efield2-efield2
      ! deallocate working arrays
      DEALLOCATE (efield0)
      DEALLOCATE (efield1)
      DEALLOCATE (efield2)

      ! Corrections for two-centers one-electron terms + nuclear terms
      CALL mp_sum(enuclear, para_env%group)
      CALL mp_sum(ecore2, para_env%group)
      energy%hartree = energy%hartree+ecore2
      energy%core_overlap = enuclear
      CALL finalize_se_taper(se_taper)
      CALL timestop(handle)
   END SUBROUTINE build_fock_matrix_coul_lrc

! **************************************************************************************************
!> \brief Construction of the residual part (1/R^3) of the Coulomb long-range
!>        term of the Fock matrix
!>        The 1/R^3 correction works in real-space strictly on the zero-cell,
!>        in order to avoid more parameters to be provided in the input..
!> \param qs_env ...
!> \param ks_matrix ...
!> \param matrix_p ...
!> \param energy ...
!> \param calculate_forces ...
!> \author Teodoro Laino [tlaino] - 12.2008
! **************************************************************************************************
   SUBROUTINE build_fock_matrix_coul_lr_r3(qs_env, ks_matrix, matrix_p, energy, calculate_forces)
      TYPE(qs_environment_type), POINTER                 :: qs_env
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ks_matrix, matrix_p
      TYPE(qs_energy_type), POINTER                      :: energy
      LOGICAL, INTENT(in)                                :: calculate_forces

      CHARACTER(len=*), PARAMETER :: routineN = 'build_fock_matrix_coul_lr_r3', &
         routineP = moduleN//':'//routineN

      INTEGER :: atom_a, atom_b, ewald_type, handle, iatom, ikind, inode, ispin, itype, jatom, &
         jkind, natoms, natorb_a, natorb_a2, natorb_b, natorb_b2, nkind, nlocal_particles, nspins
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: se_natorb
      INTEGER, DIMENSION(:), POINTER                     :: atom_of_kind
      LOGICAL                                            :: anag, atener, check, defined, found, &
                                                            switch
      LOGICAL, ALLOCATABLE, DIMENSION(:)                 :: se_defined
      REAL(KIND=dp)                                      :: dr1, ecore2, r2inv, r3inv, rinv
      REAL(KIND=dp), DIMENSION(2)                        :: ecab
      REAL(KIND=dp), DIMENSION(2025)                     :: pa_a, pa_b, pb_a, pb_b
      REAL(KIND=dp), DIMENSION(3)                        :: dr3inv, force_ab, rij
      REAL(KIND=dp), DIMENSION(45, 45)                   :: p_block_tot_a, p_block_tot_b
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: ksa_block_a, ksa_block_b, ksb_block_a, &
                                                            ksb_block_b, pa_block_a, pa_block_b, &
                                                            pb_block_a, pb_block_b
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(atprop_type), POINTER                         :: atprop
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: diagmat_ks, diagmat_p
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(distribution_1d_type), POINTER                :: local_particles
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(ewald_pw_type), POINTER                       :: ewald_pw
      TYPE(fist_nonbond_env_type), POINTER               :: se_nonbond_env
      TYPE(nddo_mpole_type), POINTER                     :: se_nddo_mpole
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab_orb
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      TYPE(qs_force_type), DIMENSION(:), POINTER         :: force
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(section_vals_type), POINTER                   :: se_section
      TYPE(semi_empirical_control_type), POINTER         :: se_control
      TYPE(semi_empirical_p_type), DIMENSION(:), POINTER :: se_kind_list
      TYPE(semi_empirical_type), POINTER                 :: se_kind_a, se_kind_b

      CALL timeset(routineN, handle)

      CALL get_qs_env(qs_env=qs_env) !sm->dbcsr

      NULLIFY (dft_control, cell, force, particle_set, diagmat_ks, &
               diagmat_p, local_particles, se_control, ewald_env, ewald_pw, &
               se_nddo_mpole, se_nonbond_env, se_section, sab_orb, logger, atprop)

      logger => cp_get_default_logger()
      CALL get_qs_env(qs_env=qs_env, dft_control=dft_control, cell=cell, para_env=para_env, &
                      atomic_kind_set=atomic_kind_set, qs_kind_set=qs_kind_set, particle_set=particle_set, &
                      ewald_env=ewald_env, atprop=atprop, &
                      local_particles=local_particles, ewald_pw=ewald_pw, se_nddo_mpole=se_nddo_mpole, &
                      se_nonbond_env=se_nonbond_env, sab_orb=sab_orb)

      nlocal_particles = SUM(local_particles%n_el(:))
      natoms = SIZE(particle_set)
      CALL ewald_env_get(ewald_env, ewald_type=ewald_type)
      SELECT CASE (ewald_type)
      CASE (do_ewald_ewald)
         ! Do Nothing
      CASE DEFAULT
         CALL cp__b("se_fock_matrix_coulomb.F",1284,"Periodic SE implemented only for standard EWALD sums.")
      END SELECT

      ! Parameters
      se_section => section_vals_get_subs_vals(qs_env%input, "DFT%QS%SE")
      se_control => dft_control%qs_control%se_control
      anag = se_control%analytical_gradients
      atener = atprop%energy

      nspins = dft_control%nspins
      IF(.NOT.(ASSOCIATED(matrix_p)))CALL cp__a("se_fock_matrix_coulomb.F",1294)
      IF(.NOT.(SIZE(ks_matrix) > 0))CALL cp__a("se_fock_matrix_coulomb.F",1295)

      CALL cp_dbcsr_allocate_matrix_set(diagmat_ks, nspins)
      CALL cp_dbcsr_allocate_matrix_set(diagmat_p, nspins)

      nkind = SIZE(atomic_kind_set)
      DO ispin = 1, nspins
         ! Allocate diagonal block matrices
         ALLOCATE (diagmat_p(ispin)%matrix, diagmat_ks(ispin)%matrix) !sm->dbcsr
         CALL cp_dbcsr_init(diagmat_p(ispin)%matrix)
         CALL cp_dbcsr_init(diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_get_block_diag(matrix_p(ispin)%matrix, diagmat_p(ispin)%matrix)
         CALL cp_dbcsr_get_block_diag(ks_matrix(ispin)%matrix, diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_set(diagmat_ks(ispin)%matrix, 0.0_dp)
         CALL cp_dbcsr_replicate_all(diagmat_p(ispin)%matrix)
         CALL cp_dbcsr_replicate_all(diagmat_ks(ispin)%matrix)
      END DO

      ! Possibly compute forces
      IF (calculate_forces) THEN
         CALL get_qs_env(qs_env=qs_env, force=force)
         ALLOCATE (atom_of_kind(natoms))
         CALL get_atomic_kind_set(atomic_kind_set=atomic_kind_set, atom_of_kind=atom_of_kind)
      END IF
      itype = get_se_type(dft_control%qs_control%method_id)

      ecore2 = 0.0_dp
      ! Real space part of the 1/R^3 sum
      ALLOCATE (se_defined(nkind), se_kind_list(nkind), se_natorb(nkind))
      DO ikind = 1, nkind
         CALL get_qs_kind(qs_kind_set(ikind), se_parameter=se_kind_a)
         se_kind_list(ikind)%se_param => se_kind_a
         CALL get_se_param(se_kind_a, defined=defined, natorb=natorb_a)
         se_defined(ikind) = (defined .AND. natorb_a >= 1)
         se_natorb(ikind) = natorb_a
      END DO

      CALL neighbor_list_iterator_create(nl_iterator, sab_orb)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, iatom=iatom, jatom=jatom, inode=inode, r=rij)
         IF (.NOT. se_defined(ikind)) CYCLE
         IF (.NOT. se_defined(jkind)) CYCLE
         se_kind_a => se_kind_list(ikind)%se_param
         se_kind_b => se_kind_list(jkind)%se_param
         natorb_a = se_natorb(ikind)
         natorb_b = se_natorb(jkind)
         natorb_a2 = natorb_a**2
         natorb_b2 = natorb_b**2

         IF (inode == 1) THEN
            CALL cp_dbcsr_get_block_p(matrix=diagmat_p(1)%matrix, &
                                      row=iatom, col=iatom, BLOCK=pa_block_a, found=found)
            IF(.NOT.(ASSOCIATED(pa_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",1347)
            check = (SIZE(pa_block_a, 1) == natorb_a) .AND. (SIZE(pa_block_a, 2) == natorb_a)
            IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",1349)
            CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(1)%matrix, &
                                      row=iatom, col=iatom, BLOCK=ksa_block_a, found=found)
            IF(.NOT.(ASSOCIATED(ksa_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",1352)
            p_block_tot_a(1:natorb_a, 1:natorb_a) = 2.0_dp*pa_block_a
            pa_a(1:natorb_a2) = RESHAPE(pa_block_a, (/natorb_a2/))
            IF (nspins == 2) THEN
               CALL cp_dbcsr_get_block_p(matrix=diagmat_p(2)%matrix, &
                                         row=iatom, col=iatom, BLOCK=pa_block_b, found=found)
               IF(.NOT.(ASSOCIATED(pa_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",1358)
               check = (SIZE(pa_block_b, 1) == natorb_a) .AND. (SIZE(pa_block_b, 2) == natorb_a)
               IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",1360)
               CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(2)%matrix, &
                                         row=iatom, col=iatom, BLOCK=ksa_block_b, found=found)
               IF(.NOT.(ASSOCIATED(ksa_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",1363)
               p_block_tot_a(1:natorb_a, 1:natorb_a) = pa_block_a+pa_block_b
               pa_b(1:natorb_a2) = RESHAPE(pa_block_b, (/natorb_a2/))
            END IF
         END IF

         dr1 = DOT_PRODUCT(rij, rij)
         IF (dr1 > rij_threshold) THEN
            ! Determine the order of the atoms, and in case switch them..
            IF (iatom <= jatom) THEN
               switch = .FALSE.
            ELSE
               switch = .TRUE.
            END IF
            ! Retrieve blocks for KS and P
            CALL cp_dbcsr_get_block_p(matrix=diagmat_p(1)%matrix, &
                                      row=jatom, col=jatom, BLOCK=pb_block_a, found=found)
            IF(.NOT.(ASSOCIATED(pb_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",1380)
            check = (SIZE(pb_block_a, 1) == natorb_b) .AND. (SIZE(pb_block_a, 2) == natorb_b)
            IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",1382)
            CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(1)%matrix, &
                                      row=jatom, col=jatom, BLOCK=ksb_block_a, found=found)
            IF(.NOT.(ASSOCIATED(ksb_block_a)))CALL cp__a("se_fock_matrix_coulomb.F",1385)
            p_block_tot_b(1:natorb_b, 1:natorb_b) = 2.0_dp*pb_block_a
            pb_a(1:natorb_b2) = RESHAPE(pb_block_a, (/natorb_b2/))
            ! Handle more than one configuration
            IF (nspins == 2) THEN
               CALL cp_dbcsr_get_block_p(matrix=diagmat_p(2)%matrix, &
                                         row=jatom, col=jatom, BLOCK=pb_block_b, found=found)
               IF(.NOT.(ASSOCIATED(pb_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",1392)
               check = (SIZE(pb_block_b, 1) == natorb_b) .AND. (SIZE(pb_block_b, 2) == natorb_b)
               IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",1394)
               CALL cp_dbcsr_get_block_p(matrix=diagmat_ks(2)%matrix, &
                                         row=jatom, col=jatom, BLOCK=ksb_block_b, found=found)
               IF(.NOT.(ASSOCIATED(ksb_block_b)))CALL cp__a("se_fock_matrix_coulomb.F",1397)
               check = (SIZE(pb_block_a, 1) == SIZE(pb_block_b, 1)) .AND. (SIZE(pb_block_a, 2) == SIZE(pb_block_b, 2))
               IF(.NOT.(check))CALL cp__a("se_fock_matrix_coulomb.F",1399)
               p_block_tot_b(1:natorb_b, 1:natorb_b) = pb_block_a+pb_block_b
               pb_b(1:natorb_b2) = RESHAPE(pb_block_b, (/natorb_b2/))
            END IF

            SELECT CASE (dft_control%qs_control%method_id)
            CASE (do_method_mndo, do_method_am1, do_method_pm3, do_method_pm6, do_method_pdg, &
                  do_method_rm1, do_method_mndod, do_method_pnnl)

               ! Pre-compute some quantities..
               r2inv = 1.0_dp/dr1
               rinv = SQRT(r2inv)
               r3inv = rinv**3

               ! Two-centers One-electron terms
               IF (nspins == 1) THEN
                  ecab = 0._dp
                  CALL fock2_1el_r3(se_kind_a, se_kind_b, ksa_block_a, ksb_block_a, pa_a, pb_a, &
                                    ecore=ecab, e1b=se_kind_a%expns3_int(jkind)%expns3%e1b, &
                                    e2a=se_kind_a%expns3_int(jkind)%expns3%e2a, rp=r3inv)
                  ecore2 = ecore2+ecab(1)+ecab(2)
               ELSE IF (nspins == 2) THEN
                  ecab = 0._dp
                  CALL fock2_1el_r3(se_kind_a, se_kind_b, ksa_block_a, ksb_block_a, pa_block_a, &
                                    pb_block_a, ecore=ecab, e1b=se_kind_a%expns3_int(jkind)%expns3%e1b, &
                                    e2a=se_kind_a%expns3_int(jkind)%expns3%e2a, rp=r3inv)

                  CALL fock2_1el_r3(se_kind_a, se_kind_b, ksa_block_b, ksb_block_b, pa_b, pb_b, &
                                    ecore=ecab, e1b=se_kind_a%expns3_int(jkind)%expns3%e1b, &
                                    e2a=se_kind_a%expns3_int(jkind)%expns3%e2a, rp=r3inv)
                  ecore2 = ecore2+ecab(1)+ecab(2)
               END IF
               IF (atener) THEN
                  atprop%atecoul(iatom) = atprop%atecoul(iatom)+ecab(1)
                  atprop%atecoul(jatom) = atprop%atecoul(jatom)+ecab(2)
               END IF
               ! Coulomb Terms
               IF (nspins == 1) THEN
                  CALL fock2C_r3(se_kind_a, se_kind_b, switch, p_block_tot_a, ksa_block_a, p_block_tot_b, &
                                 ksb_block_a, factor=0.5_dp, w=se_kind_a%expns3_int(jkind)%expns3%w, rp=r3inv)
               ELSE IF (nspins == 2) THEN
                  CALL fock2C_r3(se_kind_a, se_kind_b, switch, p_block_tot_a, ksa_block_a, p_block_tot_b, &
                                 ksb_block_a, factor=1.0_dp, w=se_kind_a%expns3_int(jkind)%expns3%w, rp=r3inv)

                  CALL fock2C_r3(se_kind_a, se_kind_b, switch, p_block_tot_a, ksa_block_b, p_block_tot_b, &
                                 ksb_block_b, factor=1.0_dp, w=se_kind_a%expns3_int(jkind)%expns3%w, rp=r3inv)
               END IF

               ! Compute forces if requested
               IF (calculate_forces) THEN
                  dr3inv = -3.0_dp*rij*r3inv*r2inv
                  atom_a = atom_of_kind(iatom)
                  atom_b = atom_of_kind(jatom)

                  force_ab = 0.0_dp
                  ! Derivatives of the One-centre One-electron terms
                  IF (nspins == 1) THEN
                     CALL dfock2_1el_r3(se_kind_a, se_kind_b, dr3inv, pa_a, pb_a, force_ab, &
                                        se_kind_a%expns3_int(jkind)%expns3%e1b, se_kind_a%expns3_int(jkind)%expns3%e2a)
                  ELSE IF (nspins == 2) THEN
                     CALL dfock2_1el_r3(se_kind_a, se_kind_b, dr3inv, pa_block_a, pb_block_a, force_ab, &
                                        se_kind_a%expns3_int(jkind)%expns3%e1b, se_kind_a%expns3_int(jkind)%expns3%e2a)

                     CALL dfock2_1el_r3(se_kind_a, se_kind_b, dr3inv, pa_b, pb_b, force_ab, &
                                        se_kind_a%expns3_int(jkind)%expns3%e1b, se_kind_a%expns3_int(jkind)%expns3%e2a)
                  END IF

                  ! Sum up force components
                  force(ikind)%all_potential(1, atom_a) = force(ikind)%all_potential(1, atom_a)-force_ab(1)
                  force(jkind)%all_potential(1, atom_b) = force(jkind)%all_potential(1, atom_b)+force_ab(1)

                  force(ikind)%all_potential(2, atom_a) = force(ikind)%all_potential(2, atom_a)-force_ab(2)
                  force(jkind)%all_potential(2, atom_b) = force(jkind)%all_potential(2, atom_b)+force_ab(2)

                  force(ikind)%all_potential(3, atom_a) = force(ikind)%all_potential(3, atom_a)-force_ab(3)
                  force(jkind)%all_potential(3, atom_b) = force(jkind)%all_potential(3, atom_b)+force_ab(3)

                  ! Derivatives of the Coulomb Terms
                  force_ab = 0.0_dp
                  IF (nspins == 1) THEN
                     CALL dfock2C_r3(se_kind_a, se_kind_b, switch, p_block_tot_a, p_block_tot_b, factor=0.25_dp, &
                                     w=se_kind_a%expns3_int(jkind)%expns3%w, drp=dr3inv, force=force_ab)
                  ELSE IF (nspins == 2) THEN
                     CALL dfock2C_r3(se_kind_a, se_kind_b, switch, p_block_tot_a, p_block_tot_b, factor=0.50_dp, &
                                     w=se_kind_a%expns3_int(jkind)%expns3%w, drp=dr3inv, force=force_ab)

                     CALL dfock2C_r3(se_kind_a, se_kind_b, switch, p_block_tot_a, p_block_tot_b, factor=0.50_dp, &
                                     w=se_kind_a%expns3_int(jkind)%expns3%w, drp=dr3inv, force=force_ab)
                  END IF

                  ! Sum up force components
                  force(ikind)%rho_elec(1, atom_a) = force(ikind)%rho_elec(1, atom_a)-force_ab(1)
                  force(jkind)%rho_elec(1, atom_b) = force(jkind)%rho_elec(1, atom_b)+force_ab(1)

                  force(ikind)%rho_elec(2, atom_a) = force(ikind)%rho_elec(2, atom_a)-force_ab(2)
                  force(jkind)%rho_elec(2, atom_b) = force(jkind)%rho_elec(2, atom_b)+force_ab(2)

                  force(ikind)%rho_elec(3, atom_a) = force(ikind)%rho_elec(3, atom_a)-force_ab(3)
                  force(jkind)%rho_elec(3, atom_b) = force(jkind)%rho_elec(3, atom_b)+force_ab(3)
               END IF
            CASE DEFAULT
               CALL cp__b("se_fock_matrix_coulomb.F",1500,"")
            END SELECT
         END IF
      END DO
      CALL neighbor_list_iterator_release(nl_iterator)

      DEALLOCATE (se_kind_list, se_defined, se_natorb)

      DO ispin = 1, nspins
         CALL cp_dbcsr_sum_replicated(diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_distribute(diagmat_ks(ispin)%matrix)
         CALL cp_dbcsr_add(ks_matrix(ispin)%matrix, diagmat_ks(ispin)%matrix, &
                           1.0_dp, 1.0_dp)
      END DO
      CALL cp_dbcsr_deallocate_matrix_set(diagmat_p)
      CALL cp_dbcsr_deallocate_matrix_set(diagmat_ks)

      IF (calculate_forces) THEN
         DEALLOCATE (atom_of_kind)
      END IF

      ! Two-centers one-electron terms
      CALL mp_sum(ecore2, para_env%group)
      energy%hartree = energy%hartree+ecore2

      CALL timestop(handle)
   END SUBROUTINE build_fock_matrix_coul_lr_r3

END MODULE se_fock_matrix_coulomb
