!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR data types
!> \author  Urban Borstnik
!> \date    2009-05-12
!> \version 0.95
!>
!> <b>Modification history:</b>
!> - Created 2009-05-12
! **************************************************************************************************
MODULE dbcsr_types
   USE array_types,                     ONLY: array_i1d_obj
   USE btree_i8_k_cp2d_v,               ONLY: btree_2d_c => btree
   USE btree_i8_k_dp2d_v,               ONLY: btree_2d_d => btree
   USE btree_i8_k_sp2d_v,               ONLY: btree_2d_s => btree
   USE btree_i8_k_zp2d_v,               ONLY: btree_2d_z => btree
   USE dbcsr_data_types,                ONLY: &
        dbcsr_data_area_type, dbcsr_data_obj, dbcsr_datatype_sizeof, dbcsr_memtype_default, &
        dbcsr_memtype_type, dbcsr_scalar_type, dbcsr_type_complex_4, dbcsr_type_complex_4_2d, &
        dbcsr_type_complex_8, dbcsr_type_complex_8_2d, dbcsr_type_complex_default, &
        dbcsr_type_int_4, dbcsr_type_real_4, dbcsr_type_real_4_2d, dbcsr_type_real_8, &
        dbcsr_type_real_8_2d, dbcsr_type_real_default
   USE kinds,                           ONLY: default_string_length,&
                                              int_8

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_types'

   PUBLIC :: dbcsr_obj, &
             dbcsr_type, &
             dbcsr_scalar_type, &
             dbcsr_data_obj, &
             dbcsr_data_area_type, &
             dbcsr_work_type, &
             dbcsr_1d_array_type, &
             dbcsr_2d_array_type, &
             dbcsr_mp_obj, &
             dbcsr_distribution_obj, &
             dbcsr_imagedistribution_type, &
             dbcsr_imagedistribution_obj, &
             dbcsr_iterator, &
             dbcsr_mutable_obj, &
             dbcsr_obj_type_p
   PUBLIC :: dbcsr_meta_size
   PUBLIC :: dbcsr_slot_size, &
             dbcsr_slot_row_p, &
             dbcsr_slot_col_i, &
             dbcsr_slot_blk_p, &
             dbcsr_slot_thr_c, &
             dbcsr_slot_coo_l, &
             dbcsr_slot_nblks, &
             dbcsr_slot_nze, &
             dbcsr_slot_dense, &
             dbcsr_slot_nblkrows_total, &
             dbcsr_slot_nblkcols_total, &
             dbcsr_slot_nfullrows_total, &
             dbcsr_slot_nfullcols_total, &
             dbcsr_slot_nblkrows_local, &
             dbcsr_slot_nblkcols_local, &
             dbcsr_slot_nfullrows_local, &
             dbcsr_slot_nfullcols_local, &
             dbcsr_slot_type, &
             dbcsr_slot_home_prow, &
             dbcsr_slot_home_pcol, &
             dbcsr_slot_home_rowi, &
             dbcsr_slot_home_coli, &
             dbcsr_slot_home_vprow, &
             dbcsr_slot_home_vpcol, &
             dbcsr_num_slots

   PUBLIC :: dbcsr_mpi_size_limits

   PUBLIC :: dbcsr_type_real_4, dbcsr_type_real_8, &
             dbcsr_type_complex_4, dbcsr_type_complex_8, &
             dbcsr_type_real_default, dbcsr_type_complex_default, &
             dbcsr_type_real_4_2d, dbcsr_type_real_8_2d, &
             dbcsr_type_complex_4_2d, dbcsr_type_complex_8_2d, &
             dbcsr_type_int_4
   PUBLIC :: dbcsr_datatype_sizeof
   PUBLIC :: dbcsr_memtype_type, &
             dbcsr_memtype_default

   PUBLIC :: dbcsr_type_invalid, dbcsr_type_no_symmetry, dbcsr_type_symmetric, &
             dbcsr_type_antisymmetric, dbcsr_type_hermitian, dbcsr_type_antihermitian
   PUBLIC :: dbcsr_no_transpose, dbcsr_transpose, dbcsr_conjugate_transpose
   PUBLIC :: dbcsr_repl_none, dbcsr_repl_row, dbcsr_repl_col, dbcsr_repl_full

   PUBLIC :: dbcsr_filter_frobenius
   PUBLIC :: dbcsr_norm_frobenius, dbcsr_norm_maxabsnorm, &
             dbcsr_norm_gershgorin, dbcsr_norm_column

   PUBLIC :: dbcsr_func_inverse, dbcsr_func_tanh, dbcsr_func_dtanh, &
             dbcsr_func_ddtanh, dbcsr_func_artanh, dbcsr_func_inverse_special, &
             dbcsr_func_spread_from_zero, &
             dbcsr_func_sin, &
             dbcsr_func_dsin, &
             dbcsr_func_ddsin, &
             dbcsr_func_asin, &
             dbcsr_func_truncate, &
             dbcsr_func_cos

   PUBLIC :: dbcsr_mpi_statistics_type

! **************************************************************************************************
!> \brief A processor (process) grid distribution
!> \var myproc                my processor (process) number
!> \var numproc               number of processors (processes)
!> \var myprow, mypcol        my process grid row and column
!> \var mp_group              message-passing group ID
!> \var pgrid                 processor grid
!> \var refcount              reference counter
!> \var subgroups_defined     whether the subgroups are defined
!> \var prow_group            per-process-row communicator
!> \var pcol_group            pre-process-column communicator
! **************************************************************************************************
   TYPE dbcsr_mp_type
      INTEGER                            :: mynode, numnodes, myprow, mypcol
      INTEGER                            :: mp_group
      INTEGER, DIMENSION(:, :), POINTER   :: pgrid => Null()
      INTEGER                            :: refcount
      LOGICAL                            :: subgroups_defined
      INTEGER                            :: prow_group, pcol_group
   END TYPE dbcsr_mp_type

! **************************************************************************************************
!> \brief A processor (process) grid distribution
!> \var myproc                my processor (process) number
!> \var numproc               number of processors (processes)
!> \var myprow, mypcol        my process grid row and column
!> \var pgrid                 processor grid
!> \var refcount              reference counter
! **************************************************************************************************
   TYPE dbcsr_mp_obj
      TYPE(dbcsr_mp_type), POINTER :: mp => Null()
   END TYPE dbcsr_mp_obj

! **************************************************************************************************
!> \brief Matrix distribution on the processor grid
!> \var row_dist, col_dist    standard row and column distributions of matrix
!>                            elements' rows and columns into processor grid
!>                            rows and columns
!> \var local_rows            list of rows local to the processor grid row
!> \var local_cols            list of columns local to the processor grid
!>                            column
!> \var other_l_rows          local rows for each process row
!> \var other_l_cols          local columns for each process column
!> \var has_other_l_rows      other_rows is defined
!> \var has_other_l_cols      other_cols is defined
!> \var global_row_map        mapping from rows to sequence in local rows
!>                            (global to local mapping)
!> \var global_col_map        mapping from rows to sequence in local columns
!>                            (global to local mapping)
!> \var has_global_row_map    whether other_row_map is defined
!> \var has_global_col_map    whether other_col_map is defined
!> \var row_map               distribution map for rows
!> \var col_map               distribution map for columns
!> \var thread_dist           thread distribution (of the rows)
!> \var numthreads            number of threads in the environment
!> \var mp_env                mulitprocessor environment on which the
!>                            distribution is based
!> \var refcount              reference counter
! **************************************************************************************************
   TYPE dbcsr_distribution_type
      TYPE(array_i1d_obj)               :: row_dist_block, col_dist_block
      TYPE(array_i1d_obj)               :: local_rows, local_cols
      TYPE(array_i1d_obj)               :: row_dist_cluster, col_dist_cluster
      INTEGER                           :: num_local_row_clusters, num_local_col_clusters
      TYPE(array_i1d_obj), DIMENSION(:), POINTER :: other_l_rows => Null(), &
                                                    other_l_cols => Null()
      LOGICAL                           :: has_other_l_rows, has_other_l_cols
      TYPE(array_i1d_obj)               :: global_row_map, global_col_map
      LOGICAL                           :: has_global_row_map, has_global_col_map
      TYPE(array_i1d_obj)               :: row_map, col_map
      LOGICAL                           :: has_thread_dist
      TYPE(array_i1d_obj)               :: thread_dist
      INTEGER                           :: num_threads
      TYPE(dbcsr_mp_obj)                :: mp_env
      INTEGER                           :: refcount
   END TYPE dbcsr_distribution_type
   TYPE dbcsr_distribution_obj
      TYPE(dbcsr_distribution_type), POINTER  :: d => Null()
   END TYPE dbcsr_distribution_obj

! **************************************************************************************************
!> \brief Image distributions are used to map incompatible processor row and
!>        column distributions.
!>
!> Used to ease storage or transfer between two different-sizes
!> sets. For example, if there are 4 real processor rows that are
!> mapped to 8 "virtual" processor rows, then there are two images for
!> every real processor row.
!> \var main                            the main distribution
!> \var row_image, col_image            distribution of matrix elements' rows
!>                                      and columns into image  rows and
!>                                      columns
!> \var row_decimation, col_decimation  Number of imaged rows/columns mapped
!>                                      to a real row/column
!> \var row_multiplicity, col_multiplicity   Number of real rows/columns mapped
!>                                      to a virtual row/column
!> \var refcount                        count of references
! **************************************************************************************************
   TYPE dbcsr_imagedistribution_type
      TYPE(dbcsr_distribution_obj)       :: main
      TYPE(array_i1d_obj)                :: row_image, col_image
      INTEGER                            :: row_decimation, col_decimation
      INTEGER                            :: row_multiplicity, col_multiplicity
      !
      TYPE(array_i1d_obj)                :: vrow_dist, vcol_dist
      !
      TYPE(array_i1d_obj), DIMENSION(:), POINTER :: other_vl_rows => Null(), &
                                                    other_vl_cols => Null()
      TYPE(array_i1d_obj)               :: global_vrow_map, global_vcol_map
      LOGICAL                           :: has_other_vl_rows, has_other_vl_cols, &
                                           has_global_vrow_map, has_global_vcol_map
      INTEGER :: id
      INTEGER :: refcount
   END TYPE dbcsr_imagedistribution_type

   TYPE dbcsr_imagedistribution_obj
      TYPE(dbcsr_imagedistribution_type), POINTER :: i => Null()
   END TYPE dbcsr_imagedistribution_obj

   ! Different method for dbcsr_filter
   INTEGER, PARAMETER            :: dbcsr_filter_frobenius = 1

   ! Different norm for dbcsr_norm
   INTEGER, PARAMETER            :: dbcsr_norm_frobenius = 1
   INTEGER, PARAMETER            :: dbcsr_norm_maxabsnorm = 2
   INTEGER, PARAMETER            :: dbcsr_norm_gershgorin = 3
   INTEGER, PARAMETER            :: dbcsr_norm_column = 4

! **************************************************************************************************
!> \brief Buffer for blocks
!> \var refcount        Reference counter
!> \var dirty           Whether any buffers are dirty
!> \var buffers         Buffers
!> \var rcb             Row and column and offset and dimensions of data in the
!>                      buffer
!> \var main            Main memory
!> \var backing         Backing memory (in lieu of main memory)
!> \var data_type       Data type used for the buffers
! **************************************************************************************************
   TYPE dbcsr_block_buffer_type
      INTEGER                                     :: refcount
      LOGICAL, DIMENSION(:), POINTER              :: dirty => Null()
      TYPE(dbcsr_data_obj), DIMENSION(:), POINTER :: buffers => Null()
      INTEGER, DIMENSION(:, :), POINTER            :: rcb => Null()
      TYPE(dbcsr_data_obj)                        :: main
      TYPE(dbcsr_data_obj), DIMENSION(:), POINTER :: backing => Null()
      INTEGER                                     :: data_type
   END TYPE dbcsr_block_buffer_type

! **************************************************************************************************
!> \brief Object for the buffer of blocks
!> \var t               Block buffer
! **************************************************************************************************
   TYPE dbcsr_block_buffer_obj
      TYPE(dbcsr_block_buffer_type), POINTER :: b => Null()
   END TYPE dbcsr_block_buffer_obj

! **************************************************************************************************
!> \brief An iterator over a DBCSR matrix.
!> \note This is briefly changed to allow being included in the dbcsr_type type
!>       What is written here is what the structure should be and not what it
!>       is.
!> \var matrix           the matrix
!> \var buffer_2d        Buffers for repointing 2d pointers (1 per thread)
!> \var row              Current row (per thread)
!> \var pos              Current position (per thread)
!> \var row_size         Size of current row
!> \var rbs, cbs         Pointer to row and column size array
!> \var roff, coff       Pointer to row and column offset array
!> \var local_indexing   The matrix has local indexing
!> \var contiguous_pointers   Whether pointers to data should be contiguous
!>                            in memory.
!> \var shared           Iterators share matrix
!> \var dynamic          Ignores the thread distribution (FCFS by block)
!> \var dynamic_byrows   Ignores the thread distribution (FCFS by row)
!> \var common_pos       Position when in mixed mode (row or block
!>                       depending in dynamic_byrows
!> \var local_rows       Mapping of local rows to global rows (if local
!>                       indexing is enabled)
!> \var global_rows      Mapping of global rows to local rows (if local
!>                       indexing is enabled)
! **************************************************************************************************
   TYPE dbcsr_iterator
      TYPE(dbcsr_obj), POINTER                       :: matrix => Null()
      TYPE(dbcsr_block_buffer_obj)                   :: buffer_2d
      INTEGER                                        :: pos, row, row_size, &
                                                        row_offset
      INTEGER, DIMENSION(:), POINTER :: rbs => Null(), &
                                        cbs => Null(), &
                                        roff => Null(), &
                                        coff => Null()
      LOGICAL            :: local_indexing
      LOGICAL            :: contiguous_pointers, transpose, read_only
      LOGICAL            :: shared, dynamic, dynamic_byrows
      INTEGER, POINTER   :: common_pos => Null()
      ! Copies from the matrix.
      INTEGER            :: nblks, nblkrows_total
      INTEGER, DIMENSION(:), POINTER :: row_p => Null(), &
                                        col_i => Null(), &
                                        blk_p => Null(), &
                                        tdist => Null(), &
                                        local_rows => Null(), &
                                        global_rows => Null()
      TYPE(dbcsr_data_obj) :: data_area
   END TYPE dbcsr_iterator

! **************************************************************************************************
!> \brief Data area with random acces reads, insertions, and deletions.
!> \var tree_s, tree_d, tree_c, tree_z  Data types for the various types
!> \var refcount                        Reference counter
!> \var data_type                       The data type that is stored
! **************************************************************************************************
   TYPE dbcsr_mutable_type
      TYPE(btree_2d_s)          :: btree_s
      TYPE(btree_2d_d)          :: btree_d
      TYPE(btree_2d_c)          :: btree_c
      TYPE(btree_2d_z)          :: btree_z
      INTEGER                   :: refcount
      INTEGER                   :: data_type
   END TYPE dbcsr_mutable_type

! **************************************************************************************************
!> \brief Object for the mutable data type
! **************************************************************************************************
   TYPE dbcsr_mutable_obj
      TYPE(dbcsr_mutable_type), POINTER :: m => Null()
   END TYPE dbcsr_mutable_obj

! **************************************************************************************************
!> \brief Used for assembling a real matrix.
!> \var m           the final bcsr matrix
!> \var data_area   holds actual values.
!> \var row_i       the row index of all of the blocks.
!> \var col_i       the column index of all of the blocks.
!> \var blk_p       the pointer into the data array of this block.
!> \var lastblk     index of the last block entered into the row_i, col_i, and
!>                  blk_p data structures
!> \var datasize    the actual size of data present in the data element
!> \var tree        tree used to index data blocks (alternative to the row_i,
!>                  col_i, and blk_p indices when index is scattered).
! **************************************************************************************************
   TYPE dbcsr_work_type
      TYPE(dbcsr_data_obj)                    :: data_area
      INTEGER, DIMENSION(:), POINTER          :: row_i => Null(), &
                                                 col_i => Null(), &
                                                 blk_p => Null()
      INTEGER        :: lastblk, datasize
      !TYPE(btree_2d_d)        :: tree
      TYPE(dbcsr_mutable_obj)                 :: mutable
   END TYPE dbcsr_work_type

! **************************************************************************************************
!> \brief The BCSR sparse matrix type.
!> \par arrays data and index hold the bulk of the data.
!> \var serial_number   a unique number of each created matrix
!> \var initialized whether the matrix is initialized (i.e., it is not random
!>                  garbage from memory)
!> \var valid       whether the matrix is valid (consistent)
!> \var name        name of the matrix
!> \var index(:)    agglomeration of the indices and offsets
!>                            of pointers into this array.
!> \par the pointers row_p, col_i, blk_p point into the index array.
!> \var row_p       points into the col_i and blk_p arrays, each element
!>                  (1:nblkrows_total+1) points to the previous row's last
!>                  element. So each rows has elements
!>                  row_p(row)+1:row_p(row+1).
!> \var coo_l       coordinate list (used for direct indexing)
!> \var col_i       the global blocked column number of this block.
!> \var blk_p       the pointer into the data array of this block.
!> \var thr_c       elements/thread for list index
!> \var row_blk_size          sizes (rows in a block) of blocked rows
!> \var col_blk_size          sizes (columns in a block) of blocked columns
!> \var row_blk_offset          row offset (size = nrow+1)
!> \var col_blk_offset          col offset (size = ncol+1)
!> \var max_rbs, max_cbs      maximal row and column sizes
!> \var dist        distribution used by this matrix
!> \var nblks       number of blocks locally present
!> \var nze         number of non-zero elements locally present
!> \var nblkrows_total, nblk_cols_total size of entire matrix in blocked rows
!>                  and blocked columns
!> \var nblkrows_local, nblk_cols_local size of local part of matrix in blocked
!>                  rows and blocked columns
!> \var nfullrows_total, nfull_cols_total size of entire matrix in full rows
!>                  and full columns
!> \var nfullrows_local, nfull_cols_local size of local part of matrix in full
!>                  rows and full columns
!> \var symmetry              matrix has symmetry
!> \var negate_real           symmetry is realized by negating the real part
!> \var negate_imaginary      symmetry is realized by negating complex part
!>                            (i.e., antisymmetric)
!> \var data_type             'r'/'R' for single/double precision real or
!>                            'c'/'C' for single/double precision complex data
!> \var replication_type      multi-process replication used in the matrix
!> \var transpose             matrix is logically transposed
!> \var bscs                  BCS Column instead of BCS Row
!> \var local_indexing        Local indexing of rows instead of global indexing.
!> \var local_rows            Map of local to global rows when local indexing
!>                            is enabled
!> \var local_rows            Map of global to local rows when local indexing
!>                            is enabled
!> \var index_memory_type     memory type for the index
!> \var data_memory_type      memory type for data
!> \var buffers               Block buffers
!> \var refcount              reference count
!> \var work_mutable          uses the mutable data for working and not the
!>                            append-only data
! **************************************************************************************************
   TYPE dbcsr_type
      INTEGER                            :: serial_number
      LOGICAL                            :: valid
      LOGICAL                            :: initialized = .FALSE.
      CHARACTER(LEN=default_string_length) :: name
      TYPE(dbcsr_data_obj)               :: data_area
      INTEGER, DIMENSION(:), POINTER     :: index => Null()
      INTEGER, DIMENSION(:), POINTER     :: row_p => Null(), &
                                            col_i => Null(), &
                                            blk_p => Null(), &
                                            thr_c => Null(), &
                                            coo_l => Null()
      TYPE(array_i1d_obj) row_blk_size, &
         col_blk_size, &
         row_blk_offset, &
         col_blk_offset, &
         local_rows, global_rows, &
         local_cols, global_cols
      LOGICAL has_local_rows, has_global_rows, &
         has_local_cols, has_global_cols
      INTEGER                            :: max_rbs, max_cbs

      INTEGER                             :: sparsity_id
      INTEGER                             :: id_nr ! use in sm_pool

      INTEGER                 :: nblks, nze
      INTEGER                 :: nblkrows_total, nblkcols_total
      INTEGER                 :: nfullrows_total, nfullcols_total
      INTEGER                 :: nblkrows_local, nblkcols_local
      INTEGER                 :: nfullrows_local, nfullcols_local
      INTEGER                 :: data_type
      CHARACTER               :: replication_type
      LOGICAL                 :: symmetry, negate_real, negate_imaginary
      LOGICAL                 :: bcsc, local_indexing, list_indexing
      TYPE(dbcsr_memtype_type):: data_memory_type, index_memory_type
      TYPE(dbcsr_block_buffer_obj) :: buffers
      TYPE(dbcsr_work_type), DIMENSION(:), POINTER :: wms => Null()
      TYPE(dbcsr_distribution_obj)       :: dist
      INTEGER                 :: refcount
      LOGICAL                 :: work_mutable
   END TYPE dbcsr_type

   CHARACTER, PARAMETER        :: dbcsr_type_invalid = '0'
   CHARACTER, PARAMETER        :: dbcsr_type_no_symmetry = 'N'
   CHARACTER, PARAMETER        :: dbcsr_type_symmetric = 'S'
   CHARACTER, PARAMETER        :: dbcsr_type_antisymmetric = 'A'
   CHARACTER, PARAMETER        :: dbcsr_type_hermitian = 'H'
   CHARACTER, PARAMETER        :: dbcsr_type_antihermitian = 'K'

   !
   ! multiply transpositions
   CHARACTER, PARAMETER        :: dbcsr_no_transpose = 'N'
   CHARACTER, PARAMETER        :: dbcsr_transpose = 'T'
   CHARACTER, PARAMETER        :: dbcsr_conjugate_transpose = 'C'

   CHARACTER, PARAMETER        :: dbcsr_repl_none = 'N'
   CHARACTER, PARAMETER        :: dbcsr_repl_row = 'R'
   CHARACTER, PARAMETER        :: dbcsr_repl_col = 'C'
   CHARACTER, PARAMETER        :: dbcsr_repl_full = 'A'

   !
   ! Function types
   INTEGER, PARAMETER          :: dbcsr_func_inverse = 0
   INTEGER, PARAMETER          :: dbcsr_func_tanh = 1
   INTEGER, PARAMETER          :: dbcsr_func_dtanh = 2
   INTEGER, PARAMETER          :: dbcsr_func_ddtanh = 3
   INTEGER, PARAMETER          :: dbcsr_func_artanh = 4
   INTEGER, PARAMETER          :: dbcsr_func_inverse_special = 5
   INTEGER, PARAMETER          :: dbcsr_func_spread_from_zero = 6
   INTEGER, PARAMETER          :: dbcsr_func_sin = 7
   INTEGER, PARAMETER          :: dbcsr_func_dsin = 8
   INTEGER, PARAMETER          :: dbcsr_func_ddsin = 9
   INTEGER, PARAMETER          :: dbcsr_func_asin = 10
   INTEGER, PARAMETER          :: dbcsr_func_cos = 11
   INTEGER, PARAMETER          :: dbcsr_func_truncate = 12

! **************************************************************************************************
!> These specify which array index in the index array is the start of the
!> specified variable. For example, row_p => index(cp_bcsr_slot_row_p).
! **************************************************************************************************
   INTEGER, PARAMETER          :: dbcsr_slot_size = 1 ! Size of the assigned
   ! values in the index
   ! array.
   INTEGER, PARAMETER          :: dbcsr_slot_nblks = 2
   INTEGER, PARAMETER          :: dbcsr_slot_nze = 3
   INTEGER, PARAMETER          :: dbcsr_slot_dense = 4
   INTEGER, PARAMETER          :: dbcsr_slot_nblkrows_total = 5
   INTEGER, PARAMETER          :: dbcsr_slot_nblkcols_total = 6
   INTEGER, PARAMETER          :: dbcsr_slot_nfullrows_total = 7
   INTEGER, PARAMETER          :: dbcsr_slot_nfullcols_total = 8
   INTEGER, PARAMETER          :: dbcsr_slot_nblkrows_local = 9
   INTEGER, PARAMETER          :: dbcsr_slot_nblkcols_local = 10
   INTEGER, PARAMETER          :: dbcsr_slot_nfullrows_local = 11
   INTEGER, PARAMETER          :: dbcsr_slot_nfullcols_local = 12
   INTEGER, PARAMETER          :: dbcsr_slot_type = 13
   INTEGER, PARAMETER          :: dbcsr_slot_home_prow = 14
   INTEGER, PARAMETER          :: dbcsr_slot_home_pcol = 15
   INTEGER, PARAMETER          :: dbcsr_slot_home_rowi = 16
   INTEGER, PARAMETER          :: dbcsr_slot_home_coli = 17
   INTEGER, PARAMETER          :: dbcsr_slot_home_vprow = 18
   INTEGER, PARAMETER          :: dbcsr_slot_home_vpcol = 19
   ! The number of meta fields.  Its value should be the index of the
   ! last slot listed above.
   INTEGER, PARAMETER          :: dbcsr_meta_size = 19
   INTEGER, PARAMETER          :: dbcsr_slot_row_p = dbcsr_meta_size+2
   INTEGER, PARAMETER          :: dbcsr_slot_col_i = dbcsr_meta_size+4
   INTEGER, PARAMETER          :: dbcsr_slot_blk_p = dbcsr_meta_size+6
   INTEGER, PARAMETER          :: dbcsr_slot_thr_c = dbcsr_meta_size+8
   INTEGER, PARAMETER          :: dbcsr_slot_coo_l = dbcsr_meta_size+10
   INTEGER, PARAMETER          :: dbcsr_num_slots = dbcsr_meta_size+11 ! previous + 1

   ! MPI message size limits (in bytes): 128, 8192, 32KB, 128KB, 4MB, 16MB
   INTEGER(KIND=int_8), DIMENSION(6), PARAMETER :: dbcsr_mpi_size_limits = &
                                                   (/2**7, 2**13, 2**15, 2**17, 2**22, 2**24/)

! **************************************************************************************************
!> \brief The BCSR sparse matrix object.
!> \var m  The DBCSR array.
! **************************************************************************************************
   TYPE dbcsr_obj
      TYPE(dbcsr_type) :: m
   END TYPE dbcsr_obj

! **************************************************************************************************
!> \brief Pointer to a work matrix.
!> \var w           the work matrix
! **************************************************************************************************
   TYPE dbcsr_work_type_p
      TYPE(dbcsr_work_type), POINTER     :: w => Null()
   END TYPE dbcsr_work_type_p

! **************************************************************************************************
!> \brief Pointer to a object.
!> \var matrix         the dbcsr_object
! **************************************************************************************************
   TYPE dbcsr_obj_type_p
      TYPE(dbcsr_obj), POINTER           :: matrix => Null()
   END TYPE dbcsr_obj_type_p

! **************************************************************************************************
!> \brief A 1-D array of DBCSR matrices
!> \var s           the array of matrices
!> \var refcount    reference counter
! **************************************************************************************************
   TYPE dbcsr_1d_array_obj
      TYPE(dbcsr_obj), DIMENSION(:), POINTER :: s => Null()
      INTEGER, POINTER                       :: refcount => Null()
   END TYPE dbcsr_1d_array_obj

! **************************************************************************************************
!> \brief A 2-D array of DBCSR matrices
!> \var s           the array of matrices
!> \var refcount    reference counter
! **************************************************************************************************
   TYPE dbcsr_2d_array_obj
      TYPE(dbcsr_obj), DIMENSION(:, :), POINTER :: s => Null()
      INTEGER, POINTER                         :: refcount => Null()
   END TYPE dbcsr_2d_array_obj

! **************************************************************************************************
!> \brief An array of DBCSR matrices
!> \var mats        the matrices
!> \var data_windows, index_windows     RMA windows
!> \var image_dist  image distribution
! **************************************************************************************************
   TYPE dbcsr_1d_array_type
      TYPE(dbcsr_obj), DIMENSION(:), POINTER :: mats => Null()
      TYPE(dbcsr_imagedistribution_obj) :: image_dist
   END TYPE dbcsr_1d_array_type

! **************************************************************************************************
!> \brief A 2-d array of DBCSR matrices
!> \var mats        the matrices
!> \var image_dist  image distribution
! **************************************************************************************************
   TYPE dbcsr_2d_array_type
      TYPE(dbcsr_obj), DIMENSION(:, :), POINTER :: mats => Null()
      TYPE(dbcsr_imagedistribution_obj) :: image_dist
   END TYPE dbcsr_2d_array_type

! **************************************************************************************************
!> \brief DBCSR MPI statistics
!> \var ...
!> \var ...
! **************************************************************************************************
   TYPE dbcsr_mpi_statistics_type
      INTEGER                               :: nimages
      INTEGER                               :: nexchanged
      INTEGER                               :: nfiltered
      ! rank 1: 1=right, 2=left
      ! rank 2: 1=total, 2=min, 3=max
      REAL, DIMENSION(2, 3)                :: data_size
      ! message size breakdown
      INTEGER(KIND=int_8), DIMENSION(SIZE(dbcsr_mpi_size_limits)+1, 2, 2) :: data_size_breakdown
   END TYPE dbcsr_mpi_statistics_type

END MODULE dbcsr_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Routine for sorting an array
!> \note
!>      CP2K:
!>      Please use the interface definded in util.F for calling sort().
!>      This could not go into cp_array__nametype1__utils.template because
!>      there some routines use the error-handling from cp_common_uses.f90,
!>      which in turn uses timings.F, which uses sort() :-(
!>
!>      DBCSR:
!>      Please use the interface definded in dbcsr_toollib.F for calling sort().
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify _lib__array__nametype1__sort.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************
MODULE dbcsr_array_sort_i8

   USE kinds,                           ONLY: int_8

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_array_sort_i8'

   PUBLIC :: dbcsr_1d_i8_sort

CONTAINS

! **************************************************************************************************
!> \brief Sorts an array inplace using a combination of merge- and bubble-sort.
!>        It also returns the indices, which the elements had before the sort.
!> \param arr the array to sort
!> \param n length of array
!> \param indices returns elements-indices before the sort
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_1d_i8_sort(arr, n, indices)
      INTEGER, INTENT(in)                                :: n
      INTEGER(kind=int_8), DIMENSION(1:n), INTENT(inout) :: arr
      INTEGER, DIMENSION(1:n), INTENT(out)               :: indices

      INTEGER                                            :: i
      INTEGER(kind=int_8), POINTER                       :: tmp_arr(:)
      INTEGER, POINTER                                   :: tmp_idx(:)

      IF (n == 0) RETURN ! for some reason this is a frequent case in cp2k

      ! scratch space used during the merge step
      ALLOCATE (tmp_arr((SIZE(arr)+1)/2), tmp_idx((SIZE(arr)+1)/2))
      indices = (/(i, i=1, SIZE(arr))/)

      CALL dbcsr_1d_i8_sort_low(arr(1:n), indices, tmp_arr, tmp_idx)

      DEALLOCATE (tmp_arr, tmp_idx)

   END SUBROUTINE dbcsr_1d_i8_sort

! **************************************************************************************************
!> \brief The actual sort routing.
!>            Only dbcsr_1d_i8_sort and itself should call this.
!> \param arr the array to sort
!> \param indices elements-indices before the sort
!> \param tmp_arr scratch space
!> \param tmp_idx scratch space
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   RECURSIVE SUBROUTINE dbcsr_1d_i8_sort_low(arr, indices, tmp_arr, tmp_idx)
      INTEGER(kind=int_8), DIMENSION(:), INTENT(inout)   :: arr
      INTEGER, DIMENSION(SIZE(arr)), INTENT(inout)       :: indices
      INTEGER(kind=int_8), DIMENSION((SIZE(arr)+1)/2), &
         INTENT(inout)                                   :: tmp_arr
      INTEGER, DIMENSION((SIZE(arr)+1)/2), INTENT(inout) :: tmp_idx

      INTEGER                                            :: i, j, k, m, t
      INTEGER(kind=int_8)                                :: a
      LOGICAL                                            :: swapped

      ! a,t:  used during swaping of elements in arr and indices

      swapped = .TRUE.

      ! If only a few elements are left we switch to bubble-sort for efficiency.
      IF (SIZE(arr) <= 7) THEN ! 7 seems to be a good choice for the moment
         DO j = SIZE(arr)-1, 1, -1
            swapped = .FALSE.
            DO i = 1, j
               IF (( arr(i+1) <  arr(i) )) THEN
                  ! swap arr(i) with arr(i+1)
                  a = arr(i)
                  arr(i) = arr(i+1)
                  arr(i+1) = a
                  ! swap indices(i) with indices(i+1)
                  t = indices(i)
                  indices(i) = indices(i+1)
                  indices(i+1) = t
                  swapped = .TRUE.
               END IF
            END DO
            IF (.NOT. swapped) EXIT
         END DO
         RETURN
      ENDIF

      ! split list in half and recusively sort both sublists
      m = (SIZE(arr)+1)/2 ! index where we going to divide the list in two
      CALL dbcsr_1d_i8_sort_low(arr(1:m), indices(1:m), tmp_arr, tmp_idx)
      CALL dbcsr_1d_i8_sort_low(arr(m+1:), indices(m+1:), tmp_arr, tmp_idx)

      ! Check for a special case: Can we just concate the two sorted sublists?
      ! This leads to O(n) scaling if the input is already sorted.
      IF (( arr(m+1) <  arr(m) )) THEN
         ! ...no - let's merge the two sorted sublists arr(:m) and arr(m+1:)
         ! Merge will be performed directly in arr. Need backup of first sublist.
         tmp_arr(1:m) = arr(1:m)
         tmp_idx(1:m) = indices(1:m)
         i = 1; ! number of elemens consumed from 1st sublist
         j = 1; ! number of elemens consumed from 2nd sublist
         k = 1; ! number of elemens already merged

         DO WHILE (i <= m .AND. j <= SIZE(arr)-m)
         IF (( arr(m+j) <  tmp_arr(i) )) THEN
            arr(k) = arr(m+j)
            indices(k) = indices(m+j)
            j = j+1
         ELSE
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
         ENDIF
         k = k+1
         ENDDO

         ! One of the two sublist is now empty.
         ! Copy possibly remaining tail of 1st sublist
         DO WHILE (i <= m)
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
            k = k+1
         ENDDO

         ! The possibly remaining tail of 2nd sublist is already at the right spot.

      ENDIF

   END SUBROUTINE dbcsr_1d_i8_sort_low

   ! template def put here so that line numbers in template and derived
   ! files are almost the same (multi-line use change it a bit)
   ! [template(type1,nametype1,USE,lessQ,defines)]
! ARGS:
!  USE = "USE kinds,       ONLY:int_8"
!  defines =
!    "#define CP_I_LESS_Q(el1,el2) ( el1 < el2 )
!     "
!  lessQ = "CP_I_LESS_Q"
!  nametype1 = "i8"
!  type1 = "INTEGER(kind=int_8)"

END MODULE dbcsr_array_sort_i8
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Routine for sorting an array
!> \note
!>      CP2K:
!>      Please use the interface definded in util.F for calling sort().
!>      This could not go into cp_array__nametype1__utils.template because
!>      there some routines use the error-handling from cp_common_uses.f90,
!>      which in turn uses timings.F, which uses sort() :-(
!>
!>      DBCSR:
!>      Please use the interface definded in dbcsr_toollib.F for calling sort().
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify _lib__array__nametype1__sort.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************
MODULE dbcsr_array_sort_s

   USE kinds,                           ONLY: real_4

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_array_sort_s'

   PUBLIC :: dbcsr_1d_s_sort

CONTAINS

! **************************************************************************************************
!> \brief Sorts an array inplace using a combination of merge- and bubble-sort.
!>        It also returns the indices, which the elements had before the sort.
!> \param arr the array to sort
!> \param n length of array
!> \param indices returns elements-indices before the sort
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_1d_s_sort(arr, n, indices)
      INTEGER, INTENT(in)                                :: n
      REAL(kind=real_4), DIMENSION(1:n), INTENT(inout)   :: arr
      INTEGER, DIMENSION(1:n), INTENT(out)               :: indices

      INTEGER                                            :: i
      INTEGER, POINTER                                   :: tmp_idx(:)
      REAL(kind=real_4), POINTER                         :: tmp_arr(:)

      IF (n == 0) RETURN ! for some reason this is a frequent case in cp2k

      ! scratch space used during the merge step
      ALLOCATE (tmp_arr((SIZE(arr)+1)/2), tmp_idx((SIZE(arr)+1)/2))
      indices = (/(i, i=1, SIZE(arr))/)

      CALL dbcsr_1d_s_sort_low(arr(1:n), indices, tmp_arr, tmp_idx)

      DEALLOCATE (tmp_arr, tmp_idx)

   END SUBROUTINE dbcsr_1d_s_sort

! **************************************************************************************************
!> \brief The actual sort routing.
!>            Only dbcsr_1d_s_sort and itself should call this.
!> \param arr the array to sort
!> \param indices elements-indices before the sort
!> \param tmp_arr scratch space
!> \param tmp_idx scratch space
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   RECURSIVE SUBROUTINE dbcsr_1d_s_sort_low(arr, indices, tmp_arr, tmp_idx)
      REAL(kind=real_4), DIMENSION(:), INTENT(inout)     :: arr
      INTEGER, DIMENSION(SIZE(arr)), INTENT(inout)       :: indices
      REAL(kind=real_4), DIMENSION((SIZE(arr)+1)/2), &
         INTENT(inout)                                   :: tmp_arr
      INTEGER, DIMENSION((SIZE(arr)+1)/2), INTENT(inout) :: tmp_idx

      INTEGER                                            :: i, j, k, m, t
      REAL(kind=real_4)                                  :: a
      LOGICAL                                            :: swapped

      ! a,t:  used during swaping of elements in arr and indices

      swapped = .TRUE.

      ! If only a few elements are left we switch to bubble-sort for efficiency.
      IF (SIZE(arr) <= 7) THEN ! 7 seems to be a good choice for the moment
         DO j = SIZE(arr)-1, 1, -1
            swapped = .FALSE.
            DO i = 1, j
               IF (( arr(i+1) <  arr(i) )) THEN
                  ! swap arr(i) with arr(i+1)
                  a = arr(i)
                  arr(i) = arr(i+1)
                  arr(i+1) = a
                  ! swap indices(i) with indices(i+1)
                  t = indices(i)
                  indices(i) = indices(i+1)
                  indices(i+1) = t
                  swapped = .TRUE.
               END IF
            END DO
            IF (.NOT. swapped) EXIT
         END DO
         RETURN
      ENDIF

      ! split list in half and recusively sort both sublists
      m = (SIZE(arr)+1)/2 ! index where we going to divide the list in two
      CALL dbcsr_1d_s_sort_low(arr(1:m), indices(1:m), tmp_arr, tmp_idx)
      CALL dbcsr_1d_s_sort_low(arr(m+1:), indices(m+1:), tmp_arr, tmp_idx)

      ! Check for a special case: Can we just concate the two sorted sublists?
      ! This leads to O(n) scaling if the input is already sorted.
      IF (( arr(m+1) <  arr(m) )) THEN
         ! ...no - let's merge the two sorted sublists arr(:m) and arr(m+1:)
         ! Merge will be performed directly in arr. Need backup of first sublist.
         tmp_arr(1:m) = arr(1:m)
         tmp_idx(1:m) = indices(1:m)
         i = 1; ! number of elemens consumed from 1st sublist
         j = 1; ! number of elemens consumed from 2nd sublist
         k = 1; ! number of elemens already merged

         DO WHILE (i <= m .AND. j <= SIZE(arr)-m)
         IF (( arr(m+j) <  tmp_arr(i) )) THEN
            arr(k) = arr(m+j)
            indices(k) = indices(m+j)
            j = j+1
         ELSE
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
         ENDIF
         k = k+1
         ENDDO

         ! One of the two sublist is now empty.
         ! Copy possibly remaining tail of 1st sublist
         DO WHILE (i <= m)
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
            k = k+1
         ENDDO

         ! The possibly remaining tail of 2nd sublist is already at the right spot.

      ENDIF

   END SUBROUTINE dbcsr_1d_s_sort_low

   ! template def put here so that line numbers in template and derived
   ! files are almost the same (multi-line use change it a bit)
   ! [template(type1,nametype1,USE,lessQ,defines)]
! ARGS:
!  USE = "USE kinds, ONLY: real_4"
!  defines =
!    "#define CP_R_LESS_Q(el1,el2) ( el1 < el2 )
!     "
!  lessQ = "CP_R_LESS_Q"
!  nametype1 = "s"
!  type1 = "REAL(kind=real_4)"

END MODULE dbcsr_array_sort_s
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Routine for sorting an array
!> \note
!>      CP2K:
!>      Please use the interface definded in util.F for calling sort().
!>      This could not go into cp_array__nametype1__utils.template because
!>      there some routines use the error-handling from cp_common_uses.f90,
!>      which in turn uses timings.F, which uses sort() :-(
!>
!>      DBCSR:
!>      Please use the interface definded in dbcsr_toollib.F for calling sort().
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify _lib__array__nametype1__sort.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************
MODULE dbcsr_array_sort_d

   USE kinds,                           ONLY: real_8

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_array_sort_d'

   PUBLIC :: dbcsr_1d_d_sort

CONTAINS

! **************************************************************************************************
!> \brief Sorts an array inplace using a combination of merge- and bubble-sort.
!>        It also returns the indices, which the elements had before the sort.
!> \param arr the array to sort
!> \param n length of array
!> \param indices returns elements-indices before the sort
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_1d_d_sort(arr, n, indices)
      INTEGER, INTENT(in)                                :: n
      REAL(kind=real_8), DIMENSION(1:n), INTENT(inout)   :: arr
      INTEGER, DIMENSION(1:n), INTENT(out)               :: indices

      INTEGER                                            :: i
      INTEGER, POINTER                                   :: tmp_idx(:)
      REAL(kind=real_8), POINTER                         :: tmp_arr(:)

      IF (n == 0) RETURN ! for some reason this is a frequent case in cp2k

      ! scratch space used during the merge step
      ALLOCATE (tmp_arr((SIZE(arr)+1)/2), tmp_idx((SIZE(arr)+1)/2))
      indices = (/(i, i=1, SIZE(arr))/)

      CALL dbcsr_1d_d_sort_low(arr(1:n), indices, tmp_arr, tmp_idx)

      DEALLOCATE (tmp_arr, tmp_idx)

   END SUBROUTINE dbcsr_1d_d_sort

! **************************************************************************************************
!> \brief The actual sort routing.
!>            Only dbcsr_1d_d_sort and itself should call this.
!> \param arr the array to sort
!> \param indices elements-indices before the sort
!> \param tmp_arr scratch space
!> \param tmp_idx scratch space
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   RECURSIVE SUBROUTINE dbcsr_1d_d_sort_low(arr, indices, tmp_arr, tmp_idx)
      REAL(kind=real_8), DIMENSION(:), INTENT(inout)     :: arr
      INTEGER, DIMENSION(SIZE(arr)), INTENT(inout)       :: indices
      REAL(kind=real_8), DIMENSION((SIZE(arr)+1)/2), &
         INTENT(inout)                                   :: tmp_arr
      INTEGER, DIMENSION((SIZE(arr)+1)/2), INTENT(inout) :: tmp_idx

      INTEGER                                            :: i, j, k, m, t
      REAL(kind=real_8)                                  :: a
      LOGICAL                                            :: swapped

      ! a,t:  used during swaping of elements in arr and indices

      swapped = .TRUE.

      ! If only a few elements are left we switch to bubble-sort for efficiency.
      IF (SIZE(arr) <= 7) THEN ! 7 seems to be a good choice for the moment
         DO j = SIZE(arr)-1, 1, -1
            swapped = .FALSE.
            DO i = 1, j
               IF (( arr(i+1) <  arr(i) )) THEN
                  ! swap arr(i) with arr(i+1)
                  a = arr(i)
                  arr(i) = arr(i+1)
                  arr(i+1) = a
                  ! swap indices(i) with indices(i+1)
                  t = indices(i)
                  indices(i) = indices(i+1)
                  indices(i+1) = t
                  swapped = .TRUE.
               END IF
            END DO
            IF (.NOT. swapped) EXIT
         END DO
         RETURN
      ENDIF

      ! split list in half and recusively sort both sublists
      m = (SIZE(arr)+1)/2 ! index where we going to divide the list in two
      CALL dbcsr_1d_d_sort_low(arr(1:m), indices(1:m), tmp_arr, tmp_idx)
      CALL dbcsr_1d_d_sort_low(arr(m+1:), indices(m+1:), tmp_arr, tmp_idx)

      ! Check for a special case: Can we just concate the two sorted sublists?
      ! This leads to O(n) scaling if the input is already sorted.
      IF (( arr(m+1) <  arr(m) )) THEN
         ! ...no - let's merge the two sorted sublists arr(:m) and arr(m+1:)
         ! Merge will be performed directly in arr. Need backup of first sublist.
         tmp_arr(1:m) = arr(1:m)
         tmp_idx(1:m) = indices(1:m)
         i = 1; ! number of elemens consumed from 1st sublist
         j = 1; ! number of elemens consumed from 2nd sublist
         k = 1; ! number of elemens already merged

         DO WHILE (i <= m .AND. j <= SIZE(arr)-m)
         IF (( arr(m+j) <  tmp_arr(i) )) THEN
            arr(k) = arr(m+j)
            indices(k) = indices(m+j)
            j = j+1
         ELSE
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
         ENDIF
         k = k+1
         ENDDO

         ! One of the two sublist is now empty.
         ! Copy possibly remaining tail of 1st sublist
         DO WHILE (i <= m)
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
            k = k+1
         ENDDO

         ! The possibly remaining tail of 2nd sublist is already at the right spot.

      ENDIF

   END SUBROUTINE dbcsr_1d_d_sort_low

   ! template def put here so that line numbers in template and derived
   ! files are almost the same (multi-line use change it a bit)
   ! [template(type1,nametype1,USE,lessQ,defines)]
! ARGS:
!  USE = "USE kinds, ONLY: real_8"
!  defines =
!    "#define CP_R_LESS_Q(el1,el2) ( el1 < el2 )
!     "
!  lessQ = "CP_R_LESS_Q"
!  nametype1 = "d"
!  type1 = "REAL(kind=real_8)"

END MODULE dbcsr_array_sort_d
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Routine for sorting an array
!> \note
!>      CP2K:
!>      Please use the interface definded in util.F for calling sort().
!>      This could not go into cp_array__nametype1__utils.template because
!>      there some routines use the error-handling from cp_common_uses.f90,
!>      which in turn uses timings.F, which uses sort() :-(
!>
!>      DBCSR:
!>      Please use the interface definded in dbcsr_toollib.F for calling sort().
!>
!>      **** DO NOT MODIFY THE .F FILES ****
!>      modify _lib__array__nametype1__sort.template instead
!>
!> \par History
!>      12.2012 first version [ole]
!> \author Ole Schuett
! **************************************************************************************************
MODULE dbcsr_array_sort_i4

   USE kinds,                           ONLY: int_4

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_array_sort_i4'

   PUBLIC :: dbcsr_1d_i4_sort

CONTAINS

! **************************************************************************************************
!> \brief Sorts an array inplace using a combination of merge- and bubble-sort.
!>        It also returns the indices, which the elements had before the sort.
!> \param arr the array to sort
!> \param n length of array
!> \param indices returns elements-indices before the sort
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_1d_i4_sort(arr, n, indices)
      INTEGER, INTENT(in)                                :: n
      INTEGER(kind=int_4), DIMENSION(1:n), INTENT(inout) :: arr
      INTEGER, DIMENSION(1:n), INTENT(out)               :: indices

      INTEGER                                            :: i
      INTEGER(kind=int_4), POINTER                       :: tmp_arr(:)
      INTEGER, POINTER                                   :: tmp_idx(:)

      IF (n == 0) RETURN ! for some reason this is a frequent case in cp2k

      ! scratch space used during the merge step
      ALLOCATE (tmp_arr((SIZE(arr)+1)/2), tmp_idx((SIZE(arr)+1)/2))
      indices = (/(i, i=1, SIZE(arr))/)

      CALL dbcsr_1d_i4_sort_low(arr(1:n), indices, tmp_arr, tmp_idx)

      DEALLOCATE (tmp_arr, tmp_idx)

   END SUBROUTINE dbcsr_1d_i4_sort

! **************************************************************************************************
!> \brief The actual sort routing.
!>            Only dbcsr_1d_i4_sort and itself should call this.
!> \param arr the array to sort
!> \param indices elements-indices before the sort
!> \param tmp_arr scratch space
!> \param tmp_idx scratch space
!> \par History
!>      12.2012 created [ole]
!> \author Ole Schuett
! **************************************************************************************************
   RECURSIVE SUBROUTINE dbcsr_1d_i4_sort_low(arr, indices, tmp_arr, tmp_idx)
      INTEGER(kind=int_4), DIMENSION(:), INTENT(inout)   :: arr
      INTEGER, DIMENSION(SIZE(arr)), INTENT(inout)       :: indices
      INTEGER(kind=int_4), DIMENSION((SIZE(arr)+1)/2), &
         INTENT(inout)                                   :: tmp_arr
      INTEGER, DIMENSION((SIZE(arr)+1)/2), INTENT(inout) :: tmp_idx

      INTEGER                                            :: i, j, k, m, t
      INTEGER(kind=int_4)                                :: a
      LOGICAL                                            :: swapped

      ! a,t:  used during swaping of elements in arr and indices

      swapped = .TRUE.

      ! If only a few elements are left we switch to bubble-sort for efficiency.
      IF (SIZE(arr) <= 7) THEN ! 7 seems to be a good choice for the moment
         DO j = SIZE(arr)-1, 1, -1
            swapped = .FALSE.
            DO i = 1, j
               IF (( arr(i+1) <  arr(i) )) THEN
                  ! swap arr(i) with arr(i+1)
                  a = arr(i)
                  arr(i) = arr(i+1)
                  arr(i+1) = a
                  ! swap indices(i) with indices(i+1)
                  t = indices(i)
                  indices(i) = indices(i+1)
                  indices(i+1) = t
                  swapped = .TRUE.
               END IF
            END DO
            IF (.NOT. swapped) EXIT
         END DO
         RETURN
      ENDIF

      ! split list in half and recusively sort both sublists
      m = (SIZE(arr)+1)/2 ! index where we going to divide the list in two
      CALL dbcsr_1d_i4_sort_low(arr(1:m), indices(1:m), tmp_arr, tmp_idx)
      CALL dbcsr_1d_i4_sort_low(arr(m+1:), indices(m+1:), tmp_arr, tmp_idx)

      ! Check for a special case: Can we just concate the two sorted sublists?
      ! This leads to O(n) scaling if the input is already sorted.
      IF (( arr(m+1) <  arr(m) )) THEN
         ! ...no - let's merge the two sorted sublists arr(:m) and arr(m+1:)
         ! Merge will be performed directly in arr. Need backup of first sublist.
         tmp_arr(1:m) = arr(1:m)
         tmp_idx(1:m) = indices(1:m)
         i = 1; ! number of elemens consumed from 1st sublist
         j = 1; ! number of elemens consumed from 2nd sublist
         k = 1; ! number of elemens already merged

         DO WHILE (i <= m .AND. j <= SIZE(arr)-m)
         IF (( arr(m+j) <  tmp_arr(i) )) THEN
            arr(k) = arr(m+j)
            indices(k) = indices(m+j)
            j = j+1
         ELSE
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
         ENDIF
         k = k+1
         ENDDO

         ! One of the two sublist is now empty.
         ! Copy possibly remaining tail of 1st sublist
         DO WHILE (i <= m)
            arr(k) = tmp_arr(i)
            indices(k) = tmp_idx(i)
            i = i+1
            k = k+1
         ENDDO

         ! The possibly remaining tail of 2nd sublist is already at the right spot.

      ENDIF

   END SUBROUTINE dbcsr_1d_i4_sort_low

   ! template def put here so that line numbers in template and derived
   ! files are almost the same (multi-line use change it a bit)
   ! [template(type1,nametype1,USE,lessQ,defines)]
! ARGS:
!  USE = "USE kinds,       ONLY:int_4"
!  defines =
!    "#define CP_I_LESS_Q(el1,el2) ( el1 < el2 )
!     "
!  lessQ = "CP_I_LESS_Q"
!  nametype1 = "i4"
!  type1 = "INTEGER(kind=int_4)"

END MODULE dbcsr_array_sort_i4
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Tools usually found in a standard library.
!> \author  Urban Borstnik
!> \date    2011-11-02
!> \version 1.0
!>
!> <b>Modification history:</b>
!> - Created 2011-11-02
! **************************************************************************************************
MODULE dbcsr_toollib

   USE dbcsr_array_sort_d,              ONLY: dbcsr_1d_d_sort
   USE dbcsr_array_sort_i4,             ONLY: dbcsr_1d_i4_sort
   USE dbcsr_array_sort_i8,             ONLY: dbcsr_1d_i8_sort
   USE dbcsr_array_sort_s,              ONLY: dbcsr_1d_s_sort
   USE kinds,                           ONLY: int_4,&
                                              int_8

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_toollib'

   ! Block sizes and arrays
   PUBLIC :: dbcsr_unpack_i8_2i4, make_coordinate_tuple
   PUBLIC :: swap
   ! string routines
   PUBLIC :: uppercase
   ! math routines
   PUBLIC :: gcd, lcm
   ! utility routines
   PUBLIC :: sort, joaat_hash
   PUBLIC :: ordered_search

   INTERFACE swap
      MODULE PROCEDURE iswap, bswap
   END INTERFACE

   INTERFACE sort
      MODULE PROCEDURE dbcsr_1d_i4_sort, dbcsr_1d_i8_sort
      MODULE PROCEDURE dbcsr_1d_s_sort, dbcsr_1d_d_sort
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param most ...
!> \param least ...
!> \retval tuple ...
! **************************************************************************************************
   ELEMENTAL FUNCTION make_coordinate_tuple(most, least) RESULT(tuple)
      INTEGER, INTENT(IN)                                :: most, least
      INTEGER(KIND=int_8)                                :: tuple

!tuple = IOR (ISHFT (most, 32), least)

      tuple = most
      tuple = IOR(ISHFT(tuple, 32), INT(least, int_8))
   END FUNCTION make_coordinate_tuple

! **************************************************************************************************
!> \brief Swaps two integers
!> \param a Integers to swap
!> \param b Integers to swap
! **************************************************************************************************
   ELEMENTAL SUBROUTINE iswap(a, b)
      INTEGER, INTENT(INOUT)                             :: a, b

      INTEGER                                            :: tmp

      tmp = a
      a = b
      b = tmp
   END SUBROUTINE iswap

! **************************************************************************************************
!> \brief Swaps two logicals
!> \param a Logicals to swap
!> \param b Logicals to swap
! **************************************************************************************************
   ELEMENTAL SUBROUTINE bswap(a, b)
      LOGICAL, INTENT(INOUT)                             :: a, b

      LOGICAL                                            :: tmp

      tmp = a
      a = b
      b = tmp
   END SUBROUTINE bswap

! **************************************************************************************************
!> \brief Splits an array of int8 values into two int4 arrays.
!> \param[in] merged          array of merged values
!> \param[out] array_upper    array to fill with the upper bytes of the merged
!>                            values
!> \param[out] array_lower    array to fill with the lower bytes of the merged
!>                            values
! **************************************************************************************************
   SUBROUTINE dbcsr_unpack_i8_2i4(merged, array_upper, array_lower)
      INTEGER(KIND=int_8), DIMENSION(:), INTENT(IN)      :: merged
      INTEGER(KIND=int_4), DIMENSION(:), INTENT(OUT)     :: array_upper, array_lower

      INTEGER(KIND=int_8), PARAMETER                     :: lmask8 = 4294967295_int_8

      INTEGER                                            :: i

!
!   ---------------------------------------------------------------------------
! Lmask is used to filter in the lower 4 bytes and so its lower 32 bits are
! set to 1: lmask8 = 2^32-1.
! Umask is used to filter in the higher 4 bytes and so its higher 32 bits
! are set to 1: umask8 = 2^32-1 << 32
!lmask8 = 4294967295 ! 2^32-1
!umask8 = 18446744069414584320 ! (2^32-1) * 2^32 = (2^64-1)-(2^32-1)

      DO i = 1, SIZE(merged)
         array_upper(i) = INT(ISHFT(merged(i), -32), KIND=int_4)
         array_lower(i) = INT(IAND(merged(i), lmask8), KIND=int_4)
      ENDDO
   END SUBROUTINE dbcsr_unpack_i8_2i4

! **************************************************************************************************
!> \brief ...
!> \param string ...
! **************************************************************************************************
   SUBROUTINE uppercase(string)
      CHARACTER(LEN=*), INTENT(INOUT)                    :: string

      INTEGER                                            :: i, iascii

      DO i = 1, LEN_TRIM(string)
         iascii = ICHAR(string(i:i))
         IF ((iascii >= 97) .AND. (iascii <= 122)) THEN
            string(i:i) = CHAR(iascii-32)
         END IF
      END DO

   END SUBROUTINE uppercase

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \retval gcd ...
! **************************************************************************************************
   ELEMENTAL FUNCTION gcd(a, b)
      INTEGER, INTENT(IN)                                :: a, b
      INTEGER                                            :: gcd

      INTEGER                                            :: aa, ab, l, rem, s

      aa = ABS(a)
      ab = ABS(b)
      IF (aa < ab) THEN
         s = aa
         l = ab
      ELSE
         s = ab
         l = aa
      ENDIF
      IF (s .NE. 0) THEN
         DO
            rem = MOD(l, s)
            IF (rem == 0) EXIT
            l = s
            s = rem
         ENDDO
         GCD = s
      ELSE
         GCD = l
      ENDIF
   END FUNCTION gcd

! **************************************************************************************************
!> \brief ...
!> \param a ...
!> \param b ...
!> \retval lcm ...
! **************************************************************************************************
   ELEMENTAL FUNCTION lcm(a, b)
      INTEGER, INTENT(IN)                                :: a, b
      INTEGER                                            :: lcm

      INTEGER                                            :: tmp

      tmp = gcd(a, b)
      IF (tmp == 0) THEN
         lcm = 0
      ELSE
         ! could still overflow if the true lcm is larger than maxint
         lcm = ABS((a/tmp)*b)
      END IF
   END FUNCTION lcm

! **************************************************************************************************
!> \brief generates the hash of a string and the index in the table
!> \param key a string of any length
!> \retval hash_index ...
!> \par History
!>       09.2006 created [Joost VandeVondele]
!>       2010-11-23 copied from CP2K and converted to take integers
!> \note
!>       http://en.wikipedia.org/wiki/Hash_table
!>       http://www.burtleburtle.net/bob/hash/doobs.html
!>       However, since fortran doesn't have an unsigned 4 byte int
!>       we compute it using an integer with the appropriate range
!>       we return already the index in the table as a final result
! **************************************************************************************************
   FUNCTION joaat_hash(key) RESULT(hash_index)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: key
      INTEGER                                            :: hash_index

      INTEGER(KIND=int_8), PARAMETER                     :: b32 = 2_int_8**32-1_int_8

      INTEGER                                            :: i, j
      INTEGER(KIND=int_8)                                :: byte, hash

      hash = 0_int_8
      DO i = 1, SIZE(key)
         DO j = 0, 3
            byte = IAND(ISHFT(key(i), -j*8), 255)
            hash = IAND(hash+byte, b32)
            hash = IAND(hash+IAND(ISHFT(hash, 10), b32), b32)
            hash = IAND(IEOR(hash, IAND(ISHFT(hash, -6), b32)), b32)
         ENDDO
      ENDDO
      hash = IAND(hash+IAND(ISHFT(hash, 3), b32), b32)
      hash = IAND(IEOR(hash, IAND(ISHFT(hash, -11), b32)), b32)
      hash = IAND(hash+IAND(ISHFT(hash, 15), b32), b32)
      ! In fortran 4-byte-integers have only 31 bits because they are signed
      ! In fortran the rightmost (least significant) bit is in position 0
      hash_index = INT(IBCLR(hash, 31))
   END FUNCTION joaat_hash

! **************************************************************************************************
!> \brief search a value in an ordered array of indices
!> \param array ...
!> \param key ...
!> \param loc ...
!> \param found ...
!> \param lb ...
!> \param ub ...
!> \retval hash_index ...
!> \par History
!>       04.2015 created [Alfio Lazzaro]
! **************************************************************************************************
   PURE SUBROUTINE ordered_search(array, key, loc, found, lb, ub)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: array
      INTEGER, INTENT(IN)                                :: key
      INTEGER, INTENT(OUT)                               :: loc
      LOGICAL, INTENT(OUT)                               :: found
      INTEGER, INTENT(IN), OPTIONAL                      :: lb, ub

      INTEGER                                            :: high, low, val

      found = .FALSE.
      IF (PRESENT(lb)) THEN
         low = lb
      ELSE
         low = LBOUND(array, 1)
      ENDIF
      IF (PRESENT(ub)) THEN
         high = ub
      ELSE
         high = UBOUND(array, 1)
      ENDIF
      loc = (low+high)/2
      DO WHILE (loc .GE. low .AND. loc .LE. high)
         val = array(loc)
         IF (val .EQ. key) THEN
            found = .TRUE.
            EXIT
         ELSEIF (val .LT. key) THEN
            low = loc+1
         ELSE
            high = loc-1
         ENDIF
         loc = (low+high)/2
      ENDDO
   END SUBROUTINE ordered_search

END MODULE dbcsr_toollib
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief
!> \author VW
!>
!> <b>Modification history:</b>
!> - Created Feb 2010
! **************************************************************************************************
MODULE dbcsr_error_handling

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   PRIVATE

   ! procedures
   PUBLIC :: dbcsr_assert
   ! parameters
   PUBLIC :: dbcsr_failure_level, dbcsr_fatal_level
   PUBLIC :: dbcsr_caller_error, dbcsr_wrong_args_error, &
             dbcsr_internal_error, &
             dbcsr_unimplemented_error_nr

   ! interfaces
   INTERFACE dbcsr_assert
      MODULE PROCEDURE dbcsr_int_assert
      MODULE PROCEDURE dbcsr_logical_assert
      MODULE PROCEDURE dbcsr_not_assert
   END INTERFACE

   INTEGER, PARAMETER :: dbcsr_error_stack_size = 30
   !! level of an error
   INTEGER, PARAMETER :: dbcsr_fatal_level = 3
   !! level of a failure
   INTEGER, PARAMETER :: dbcsr_failure_level = 2
   !! level of a note
   INTEGER, PARAMETER :: dbcsr_note_level = 0
   !! error number: no error
   INTEGER, PARAMETER :: dbcsr_no_error = 0
   !! error number: generic error on the side of the caller
   INTEGER, PARAMETER :: dbcsr_caller_error = 1
   !! error number: one or more arguments have and invalid value
   INTEGER, PARAMETER :: dbcsr_wrong_args_error = 100
   !! error number: precondition failed
   INTEGER, PARAMETER :: dbcsr_precondition_failed = 200
   !! error number: generic error inside the routine
   INTEGER, PARAMETER :: dbcsr_internal_error = -1
   !! error number: postcondition failed
   INTEGER, PARAMETER :: dbcsr_postcondition_failed = -200
   !! error number: invariant failed
   INTEGER, PARAMETER :: dbcsr_invariant_failed = -100
   !! error number: assertion failure
   INTEGER, PARAMETER :: dbcsr_assertion_failed = -300
   !! error number: not implemented
   INTEGER, PARAMETER :: dbcsr_unimplemented_error_nr = -1000

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_error_handling'

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param level ...
!> \param etype ...
!> \param routine ...
!> \param msg ...
!> \param line ...
!> \param[inout]
! **************************************************************************************************
   SUBROUTINE dbcsr_abort_low(level, etype, routine, msg, line)
      INTEGER, INTENT(in)                                :: level, etype
      CHARACTER(*), INTENT(in)                           :: routine, msg
      INTEGER, INTENT(in)                                :: line

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(level))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(etype))==-1) EXIT ;  END DO ; ENDIF
      CALL cp_abort(cp__l(routine, line), msg)

   END SUBROUTINE dbcsr_abort_low

! **************************************************************************************************
!> \brief Assertion
!> \param[in] left            left value
!> \param[in] rel             relation
!> \param[in] right           right value
!> \param[in] level           error level
!> \param[in] etype           error type
!> \param[in] routine         Routine name
!> \param[in] msg   Message to display if the assertion fails
!> \param line ...
! **************************************************************************************************
   SUBROUTINE dbcsr_int_assert(left, rel, right, level, etype, routine, msg, line)
      INTEGER, INTENT(IN)                                :: left
      CHARACTER(len=2), INTENT(IN)                       :: rel
      INTEGER, INTENT(IN)                                :: right, level, etype
      CHARACTER(len=*), INTENT(IN)                       :: routine, msg
      INTEGER, INTENT(IN)                                :: line

      LOGICAL                                            :: l

!   ---------------------------------------------------------------------------

      SELECT CASE (rel)
      CASE ("EQ")
         l = left .EQ. right
      CASE ("LT")
         l = left .LT. right
      CASE ("LE")
         l = left .LE. right
      CASE ("GT")
         l = left .GT. right
      CASE ("GE")
         l = left .GE. right
      CASE ("NE")
         l = left .NE. right
      CASE default
         CALL cp__b("dbcsr/error/dbcsr_error_handling.F",120,"Invalid relation specified: "//rel)
         l = .FALSE.
      END SELECT
      IF (.NOT. l) THEN
         WRITE (*, '(1X,A,1X,I9,A4,I9)') "ASSERTION FAILED:", &
            left, "."//rel//".", right
         CALL dbcsr_abort_low(level, etype, routine, msg, line)
      ENDIF
   END SUBROUTINE dbcsr_int_assert

! **************************************************************************************************
!> \brief Assertion
!> \param[in] left            left value
!> \param[in] rel             relation
!> \param[in] right           right value
!> \param[in] level           error level
!> \param[in] etype           error type
!> \param[in] routine         Routine name
!> \param[in] msg   Message to display if the assertion fails
!> \param line ...
! **************************************************************************************************
   SUBROUTINE dbcsr_logical_assert(left, rel, right, level, etype, routine, msg, line)
      LOGICAL, INTENT(IN)                                :: left
      CHARACTER(len=*), INTENT(IN)                       :: rel
      LOGICAL, INTENT(IN)                                :: right
      INTEGER, INTENT(IN)                                :: level, etype
      CHARACTER(len=*), INTENT(IN)                       :: routine, msg
      INTEGER, INTENT(IN)                                :: line

      LOGICAL                                            :: l

!   ---------------------------------------------------------------------------

      SELECT CASE (rel)
      CASE ("EQV")
         l = left .EQV. right
      CASE ("NEQV")
         l = left .NEQV. right
      CASE ("OR")
         l = left .OR. right
      CASE ("AND")
         l = left .AND. right
      CASE ("IMP")
         l = .NOT. left .OR. right
      CASE default
         CALL cp__b("dbcsr/error/dbcsr_error_handling.F",165,"Invalid relation specified: "//rel)
         l = .FALSE.
      END SELECT
      IF (.NOT. l) THEN
         WRITE (*, '(1X,A,1X,L1,A,L1)') "ASSERTION FAILED:", &
            left, "."//rel//".", right
         CALL dbcsr_abort_low(level, etype, routine, msg, line)
      ENDIF
   END SUBROUTINE dbcsr_logical_assert

! **************************************************************************************************
!> \brief Assertion
!> \param[in] rel             relation
!> \param[in] right           right value
!> \param[in] level           error level
!> \param[in] etype           error type
!> \param[in] routine         Routine name
!> \param[in] msg   Message to display if the assertion fails
!> \param line ...
! **************************************************************************************************
   SUBROUTINE dbcsr_not_assert(rel, right, level, etype, routine, msg, line)
      CHARACTER(len=3), INTENT(IN)                       :: rel
      LOGICAL, INTENT(IN)                                :: right
      INTEGER, INTENT(IN)                                :: level, etype
      CHARACTER(len=*), INTENT(IN)                       :: routine, msg
      INTEGER, INTENT(IN)                                :: line

      LOGICAL                                            :: l

!   ---------------------------------------------------------------------------

      SELECT CASE (rel)
      CASE ("NOT")
         l = .NOT. right
      CASE default
         CALL cp__b("dbcsr/error/dbcsr_error_handling.F",200,"Invalid relation specified: "//rel)
         l = .FALSE.
      END SELECT
      IF (.NOT. l) THEN
         WRITE (*, '(1X,A,1X,A,L1)') "ASSERTION FAILED:", &
            "."//rel//".", right
         CALL dbcsr_abort_low(level, etype, routine, msg, line)
      ENDIF
   END SUBROUTINE dbcsr_not_assert

END MODULE dbcsr_error_handling
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Configuration options for DBCSR
!> \author  Urban Borstnik
!> \date    2011-04-05
!> \version 1.0
!>
!> <b>Modification history:</b>
!> - Created 2011-04-05
! **************************************************************************************************
MODULE dbcsr_config
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_wrong_args_error
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_config'

   PUBLIC :: dbcsr_set_default_config

   PUBLIC :: dbcsr_set_conf_mm_stacksize, dbcsr_get_conf_mm_stacksize

   PUBLIC :: dbcsr_get_conf_mpi_mem

   PUBLIC :: dbcsr_set_conf_mm_driver, dbcsr_get_conf_mm_driver, dbcsr_get_conf_avg_elements_images

   PUBLIC :: dbcsr_set_conf_avg_elements_images, dbcsr_set_conf_nstacks, dbcsr_get_conf_nstacks

   PUBLIC :: dbcsr_set_conf_use_mpi_exp, dbcsr_get_conf_use_mpi_exp
   PUBLIC :: dbcsr_set_conf_use_mpi_filtering, dbcsr_get_conf_use_mpi_filtering
   PUBLIC :: dbcsr_set_conf_use_comm_thread, dbcsr_get_conf_use_comm_thread
   PUBLIC :: dbcsr_set_conf_comm_thread_load, dbcsr_get_conf_comm_thread_load
   PUBLIC :: dbcsr_get_conf_max_ele_block, dbcsr_set_conf_max_ele_block
   PUBLIC :: dbcsr_get_conf_randmat_seed, dbcsr_set_conf_randmat_seed

   ! These are for DBCSR internal use.
   !
   PUBLIC :: has_MPI
   PUBLIC :: has_acc
   PUBLIC :: mm_stack_size
   PUBLIC :: mm_driver, &
             mm_driver_auto, &
             mm_driver_blas, &
             mm_driver_matmul, &
             mm_driver_smm, &
             mm_driver_xsmm
   PUBLIC :: mm_name_auto, &
             mm_name_blas, &
             mm_name_matmul, &
             mm_name_smm, &
             mm_name_xsmm
   PUBLIC :: use_mpi_exp, use_mpi_filtering, use_comm_thread, comm_thread_load
   PUBLIC :: max_elements_per_block, avg_elements_images
   !
   PUBLIC :: is_initialized

   PUBLIC :: accdrv_avoid_after_busy, accdrv_min_flop_process, &
             accdrv_min_flop_sort, accdrv_do_inhomogenous, &
             accdrv_binning_nbins, accdrv_binning_binsize, &
             accdrv_posterior_streams, accdrv_posterior_buffers, &
             accdrv_priority_streams, accdrv_priority_buffers

   PUBLIC :: multrec_limit
   PUBLIC :: default_resize_factor
   PUBLIC :: randmat_counter

   ! First the constants are declared.
   REAL(KIND=dp), PARAMETER             :: default_resize_factor = 1.2_dp

   ! Possible drivers to use for matrix multiplications
   INTEGER, PARAMETER :: mm_driver_auto = 0
   INTEGER, PARAMETER :: mm_driver_matmul = 1
   INTEGER, PARAMETER :: mm_driver_blas = 2
   INTEGER, PARAMETER :: mm_driver_smm = 3
   INTEGER, PARAMETER :: mm_driver_xsmm = 4

   CHARACTER(len=*), PARAMETER :: mm_name_auto = "AUTO", &
                                  mm_name_blas = "BLAS", &
                                  mm_name_matmul = "MATMUL", &
                                  mm_name_smm = "SMM", &
                                  mm_name_xsmm = "XSMM"

   INTEGER, SAVE :: mm_driver = mm_driver_blas

   LOGICAL, PARAMETER :: has_smm = .FALSE.

   LOGICAL, PARAMETER :: has_smm_vec = .FALSE.

   LOGICAL, PARAMETER :: has_xsmm = .FALSE.

   LOGICAL, PARAMETER :: has_acc = .FALSE.

   LOGICAL, PARAMETER :: has_MPI = .FALSE.

   LOGICAL, PARAMETER :: has_omp_2_5 = .TRUE.

   ! Now the fallback defaults are defined.  The defaults should be set
   ! by calling the dbcsr_init_conf() subroutine.

   ! Use MPI-allocated memory.
   LOGICAL, SAVE :: use_MPI_memory = has_MPI

   ! Stack size to use for multiplication parameters
   INTEGER, SAVE :: mm_stack_size = 1000

   ! Default blocking parameter
   INTEGER, SAVE :: max_elements_per_block = 32

   ! Maximum number of elements for each image
   INTEGER, SAVE :: avg_elements_images = -1

   ! Number of stacks to use
   INTEGER, SAVE :: nm_stacks = 3
   INTEGER, SAVE :: nn_stacks = 3
   INTEGER, SAVE :: nk_stacks = 3

   ! RMA algorithm
   LOGICAL, SAVE :: use_mpi_exp = .FALSE.

   ! RMA MPI filtering
   LOGICAL, SAVE :: use_mpi_filtering = .TRUE.

   ! Configuration of an MPI progress thread
   LOGICAL, SAVE :: use_comm_thread = .TRUE.
   INTEGER, SAVE :: comm_thread_load = 100

   LOGICAL, SAVE :: is_initialized = .FALSE.

   INTEGER, SAVE :: multrec_limit = 512

   INTEGER, SAVE :: accdrv_priority_streams = 4
   INTEGER, SAVE :: accdrv_priority_buffers = 40
   INTEGER, SAVE :: accdrv_posterior_streams = 4
   INTEGER, SAVE :: accdrv_posterior_buffers = 80
   LOGICAL, SAVE :: accdrv_avoid_after_busy = .FALSE.
   INTEGER, SAVE :: accdrv_min_flop_process = 0
   INTEGER, SAVE :: accdrv_min_flop_sort = 4000
   LOGICAL, SAVE :: accdrv_do_inhomogenous = .TRUE.
   INTEGER, SAVE :: accdrv_binning_nbins = 4096
   INTEGER, SAVE :: accdrv_binning_binsize = 16

   ! a counter used to generate different random matrices from call to call
   INTEGER, SAVE :: randmat_counter = 12341313

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param driver ...
! **************************************************************************************************
   SUBROUTINE dbcsr_set_conf_mm_driver(driver)
      INTEGER, INTENT(IN)                                :: driver

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_conf_mm_driver', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!   ---------------------------------------------------------------------------
!defaults

      CALL timeset(routineN, error_handle)

      SELECT CASE (driver)
      CASE (mm_driver_auto)
         IF (has_xsmm) THEN
            mm_driver = mm_driver_xsmm
         ELSEIF (has_smm) THEN
            mm_driver = mm_driver_smm
         ELSE
            mm_driver = mm_driver_blas ! always available
         ENDIF
      CASE (mm_driver_matmul)
         mm_driver = mm_driver_matmul ! always available
      CASE (mm_driver_blas)
         mm_driver = mm_driver_blas ! always available
      CASE (mm_driver_smm)
         IF (.NOT. has_smm) CALL cp__b("dbcsr/base/dbcsr_config.F",213,"Support for libsmm not compiled in.")
         mm_driver = mm_driver_smm
      CASE (mm_driver_xsmm)
         IF (.NOT. has_xsmm) CALL cp__b("dbcsr/base/dbcsr_config.F",216,"Support for libxsmm not compiled in.")
         mm_driver = mm_driver_xsmm
      CASE DEFAULT
         CALL cp__b("dbcsr/base/dbcsr_config.F",219,"Unkown MM driver")
      END SELECT
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_set_conf_mm_driver

! **************************************************************************************************
!> \brief ...
!> \retval driver ...
! **************************************************************************************************
   FUNCTION dbcsr_get_conf_mm_driver() RESULT(driver)
      INTEGER                                            :: driver

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_conf_mm_driver', &
         routineP = moduleN//':'//routineN

      driver = mm_driver
   END FUNCTION dbcsr_get_conf_mm_driver

! **************************************************************************************************
!> \brief ...
!> \param use_mpi_mem ...
! **************************************************************************************************
   SUBROUTINE dbcsr_set_conf_mpi_mem(use_mpi_mem)
      LOGICAL, INTENT(IN)                                :: use_mpi_mem

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_conf_mpi_mem', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

      CALL timeset(routineN, error_handle)
      use_MPI_memory = use_mpi_mem
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_set_conf_mpi_mem

! **************************************************************************************************
!> \brief ...
!> \retval use_mpi_mem ...
! **************************************************************************************************
   FUNCTION dbcsr_get_conf_mpi_mem() RESULT(use_mpi_mem)
      LOGICAL                                            :: use_mpi_mem

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_conf_mpi_mem', &
         routineP = moduleN//':'//routineN

      use_mpi_mem = use_MPI_memory
   END FUNCTION dbcsr_get_conf_mpi_mem

! **************************************************************************************************
!> \brief ...
!> \param use_mm_stacksize ...
! **************************************************************************************************
   SUBROUTINE dbcsr_set_conf_mm_stacksize(use_mm_stacksize)
      INTEGER, INTENT(IN)                                :: use_mm_stacksize

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_conf_mm_stacksize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

      CALL timeset(routineN, error_handle)
      mm_stack_size = use_mm_stacksize
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_set_conf_mm_stacksize

! **************************************************************************************************
!> \brief ...
!> \retval use_mm_stacksize ...
! **************************************************************************************************
   FUNCTION dbcsr_get_conf_mm_stacksize() RESULT(use_mm_stacksize)
      INTEGER                                            :: use_mm_stacksize

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_conf_mm_stacksize', &
         routineP = moduleN//':'//routineN

      use_mm_stacksize = mm_stack_size
   END FUNCTION dbcsr_get_conf_mm_stacksize

! **************************************************************************************************
!> \brief ...
!> \param max_ele_block ...
! **************************************************************************************************
   SUBROUTINE dbcsr_set_conf_max_ele_block(max_ele_block)
      INTEGER, INTENT(IN)                                :: max_ele_block

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_conf_max_ele_block', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

      CALL timeset(routineN, error_handle)
      max_elements_per_block = max_ele_block
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_set_conf_max_ele_block

! **************************************************************************************************
!> \brief ...
!> \retval max_ele_block ...
! **************************************************************************************************
   FUNCTION dbcsr_get_conf_max_ele_block() RESULT(max_ele_block)
      INTEGER                                            :: max_ele_block

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_conf_max_ele_block', &
         routineP = moduleN//':'//routineN

      max_ele_block = max_elements_per_block
   END FUNCTION dbcsr_get_conf_max_ele_block

! **************************************************************************************************
!> \brief ...
!> \param randmat_seed ...
! **************************************************************************************************
   SUBROUTINE dbcsr_set_conf_randmat_seed(randmat_seed)
      INTEGER, INTENT(IN)                                :: randmat_seed

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_conf_randmat_seed', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

      CALL timeset(routineN, error_handle)
      randmat_counter = randmat_seed
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_set_conf_randmat_seed

! **************************************************************************************************
!> \brief ...
!> \retval max_ele_block ...
! **************************************************************************************************
   FUNCTION dbcsr_get_conf_randmat_seed() RESULT(randmat_seed)
      INTEGER                                            :: randmat_seed

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_conf_randmat_seed', &
         routineP = moduleN//':'//routineN

      randmat_seed = randmat_counter
   END FUNCTION dbcsr_get_conf_randmat_seed

! **************************************************************************************************
!> \brief ...
!> \param new_avg_elements_images ...
! **************************************************************************************************
   SUBROUTINE dbcsr_set_conf_avg_elements_images(new_avg_elements_images)
      INTEGER, INTENT(IN)                                :: new_avg_elements_images

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_conf_avg_elements_images', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

      CALL timeset(routineN, error_handle)
      avg_elements_images = new_avg_elements_images
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_set_conf_avg_elements_images

! **************************************************************************************************
!> \brief ...
!> \retval dbcsr_get_conf_avg_elements_images ...
! **************************************************************************************************
   FUNCTION dbcsr_get_conf_avg_elements_images()
      INTEGER :: dbcsr_get_conf_avg_elements_images

      dbcsr_get_conf_avg_elements_images = avg_elements_images
   END FUNCTION dbcsr_get_conf_avg_elements_images

! **************************************************************************************************
!> \brief ...
!> \param nstacks ...
! **************************************************************************************************
   SUBROUTINE dbcsr_set_conf_nstacks(nstacks)
      INTEGER, INTENT(IN)                                :: nstacks(:)

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_conf_nstacks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

      CALL timeset(routineN, error_handle)
      CALL dbcsr_assert(SIZE(nstacks), "EQ", 3, &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "ntacks size must be 3", 399)
      nm_stacks = nstacks(1)
      nn_stacks = nstacks(2)
      nk_stacks = nstacks(3)
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_set_conf_nstacks

! **************************************************************************************************
!> \brief ...
!> \param nstacks ...
! **************************************************************************************************
   SUBROUTINE dbcsr_get_conf_nstacks(nstacks)
      INTEGER, INTENT(OUT)                               :: nstacks(:)

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_conf_nstacks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

      CALL timeset(routineN, error_handle)
      CALL dbcsr_assert(SIZE(nstacks), "EQ", 3, &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "ntacks size must be 3", 421)
      nstacks(1) = nm_stacks
      nstacks(2) = nn_stacks
      nstacks(3) = nk_stacks
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_get_conf_nstacks

! **************************************************************************************************
!> \brief ...
!> \param mpi_exp ...
! **************************************************************************************************
   SUBROUTINE dbcsr_set_conf_use_mpi_exp(mpi_exp)
      LOGICAL, INTENT(IN)                                :: mpi_exp

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_conf_use_mpi_exp', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

      CALL timeset(routineN, error_handle)
      use_mpi_exp = mpi_exp
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_set_conf_use_mpi_exp

! **************************************************************************************************
!> \brief ...
!> \retval mpi_exp ...
! **************************************************************************************************
   FUNCTION dbcsr_get_conf_use_mpi_exp() RESULT(mpi_exp)
      LOGICAL                                            :: mpi_exp

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_conf_use_mpi_exp', &
         routineP = moduleN//':'//routineN

      mpi_exp = use_mpi_exp
   END FUNCTION dbcsr_get_conf_use_mpi_exp

! **************************************************************************************************
!> \brief ...
!> \param mpi_filtering ...
! **************************************************************************************************
   SUBROUTINE dbcsr_set_conf_use_mpi_filtering(mpi_filtering)
      LOGICAL, INTENT(IN)                                :: mpi_filtering

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_conf_use_mpi_filtering', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

      CALL timeset(routineN, error_handle)
      use_mpi_filtering = mpi_filtering
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_set_conf_use_mpi_filtering

! **************************************************************************************************
!> \brief ...
!> \retval mpi_filtering ...
! **************************************************************************************************
   FUNCTION dbcsr_get_conf_use_mpi_filtering() RESULT(mpi_filtering)
      LOGICAL                                            :: mpi_filtering

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_conf_use_mpi_filtering', &
         routineP = moduleN//':'//routineN

      mpi_filtering = use_mpi_filtering
   END FUNCTION dbcsr_get_conf_use_mpi_filtering

! **************************************************************************************************
!> \brief ...
!> \param comm_thread ...
! **************************************************************************************************
   SUBROUTINE dbcsr_set_conf_use_comm_thread(comm_thread)
      LOGICAL, INTENT(IN)                                :: comm_thread

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_conf_use_comm_thread', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

      CALL timeset(routineN, error_handle)
      use_comm_thread = comm_thread
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_set_conf_use_comm_thread

! **************************************************************************************************
!> \brief ...
!> \retval comm_thread ...
! **************************************************************************************************
   FUNCTION dbcsr_get_conf_use_comm_thread() RESULT(comm_thread)
      LOGICAL                                            :: comm_thread

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_conf_use_comm_thread', &
         routineP = moduleN//':'//routineN

      comm_thread = use_comm_thread
   END FUNCTION dbcsr_get_conf_use_comm_thread

! **************************************************************************************************
!> \brief ...
!> \param load ...
! **************************************************************************************************
   SUBROUTINE dbcsr_set_conf_comm_thread_load(load)
      INTEGER, INTENT(IN)                                :: load

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_conf_comm_thread_load', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, nthreads

      CALL timeset(routineN, error_handle)
      nthreads = 1
!$    nthreads = OMP_GET_MAX_THREADS()
      comm_thread_load = MAX(0, 90-(30*nthreads)/8)
      IF (load >= 0) &
         comm_thread_load = load
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_set_conf_comm_thread_load

! **************************************************************************************************
!> \brief ...
!> \retval load ...
! **************************************************************************************************
   FUNCTION dbcsr_get_conf_comm_thread_load() RESULT(load)
      INTEGER                                            :: load

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_conf_comm_thread_load', &
         routineP = moduleN//':'//routineN

      load = comm_thread_load
   END FUNCTION dbcsr_get_conf_comm_thread_load

! **************************************************************************************************
!> \brief Configure DBCSR driver with default values.
!>
!> Basically it reads the default values and reset them.
! **************************************************************************************************
   SUBROUTINE dbcsr_set_default_config()

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_default_config', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: comm_thread_load, handle, max_ele_block, &
                                                            mm_driver, mm_ss
      INTEGER, DIMENSION(3)                              :: nstacks
      LOGICAL                                            :: use_comm_thread, use_mpi_mem

      CALL timeset(routineN, handle)

      use_mpi_mem = dbcsr_get_conf_mpi_mem()
      mm_ss = -1
      mm_driver = dbcsr_get_conf_mm_driver()
      max_ele_block = dbcsr_get_conf_max_ele_block()
      nstacks(1:3) = -1
      use_comm_thread = dbcsr_get_conf_use_comm_thread()
      comm_thread_load = -1

      randmat_counter = 12341313

      CALL dbcsr_set_conf_mm_driver(mm_driver)
      CALL dbcsr_set_conf_mpi_mem(use_mpi_mem)
      CALL dbcsr_set_conf_max_ele_block(max_ele_block)
      IF (mm_ss > 0) CALL dbcsr_set_conf_mm_stacksize(mm_ss)
      IF (nstacks(1) > 0) CALL dbcsr_set_conf_nstacks(nstacks)

      CALL dbcsr_set_conf_use_comm_thread(use_comm_thread)
      CALL dbcsr_set_conf_comm_thread_load(comm_thread_load)

      CALL timestop(handle)

   END SUBROUTINE dbcsr_set_default_config

END MODULE dbcsr_config
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE min_heap

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: heap_t
   PUBLIC :: heap_get_first, heap_reset_first, heap_fill
   PUBLIC :: heap_new, heap_release

   ! Sets the types to the default integer types.
   INTEGER, PARAMETER :: keyt = KIND(0)
   INTEGER, PARAMETER :: valt = KIND(0)

   TYPE heap_node
      INTEGER(KIND=keyt) :: key
      INTEGER(KIND=valt) :: value
   END TYPE heap_node

   TYPE heap_node_e
      TYPE(heap_node) :: node
   END TYPE heap_node_e

   TYPE heap_t
      INTEGER :: n
      TYPE(heap_node_e), DIMENSION(:), POINTER :: nodes
   END TYPE heap_t

   INTERFACE node_set
      MODULE PROCEDURE node_set_node, node_set_heap
   END INTERFACE

   INTERFACE node_get
      MODULE PROCEDURE node_get_node, node_get_heap
   END INTERFACE

CONTAINS

   ! Lookup functions

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \retval parent ...
! **************************************************************************************************
   ELEMENTAL FUNCTION get_parent(n) RESULT(parent)
      INTEGER, INTENT(IN)                                :: n
      INTEGER                                            :: parent

      parent = INT(n/2)
   END FUNCTION get_parent

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \retval child ...
! **************************************************************************************************
   ELEMENTAL FUNCTION get_left_child(n) RESULT(child)
      INTEGER, INTENT(IN)                                :: n
      INTEGER                                            :: child

      child = 2*n
   END FUNCTION get_left_child

! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \retval child ...
! **************************************************************************************************
   ELEMENTAL FUNCTION get_right_child(n) RESULT(child)
      INTEGER, INTENT(IN)                                :: n
      INTEGER                                            :: child

      child = 2*n+1
   END FUNCTION get_right_child

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param n ...
!> \retval value ...
! **************************************************************************************************
   ELEMENTAL FUNCTION get_value(heap, n) RESULT(value)
      TYPE(heap_t), INTENT(IN)                           :: heap
      INTEGER, INTENT(IN)                                :: n
      INTEGER(KIND=valt)                                 :: value

      value = heap%nodes(n)%node%value
   END FUNCTION get_value

   ! Initialization functions

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE heap_new(heap, n)
      TYPE(heap_t), INTENT(OUT)                          :: heap
      INTEGER, INTENT(IN)                                :: n

      heap%n = n
      ALLOCATE (heap%nodes(n))
   END SUBROUTINE heap_new

! **************************************************************************************************
!> \brief ...
!> \param heap ...
! **************************************************************************************************
   SUBROUTINE heap_release(heap)
      TYPE(heap_t), INTENT(INOUT)                        :: heap

      DEALLOCATE (heap%nodes)
      heap%n = 0
   END SUBROUTINE heap_release

!> \brief Fill heap with given keys and values
!> \param[in,out] heap  Heap to fill.
!> \param[in] keys      Keys to store in heap
!> \param[in] values    Values (i.e., costs) of the keys.
!> \param[out] error    Error occurred.
! **************************************************************************************************
   SUBROUTINE heap_fill(heap, keys, values, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER(KIND=keyt), DIMENSION(:), INTENT(IN)       :: keys
      INTEGER(KIND=valt), DIMENSION(:), INTENT(IN)       :: values
      LOGICAL, INTENT(OUT)                               :: error

      INTEGER                                            :: first, i, n

!

      error = .FALSE.
      n = SIZE(keys)
      IF (SIZE(values) .NE. n .OR. heap%n .LT. n) THEN
         error = .TRUE.
      ELSE
         DO i = 1, n
            !   call node_set(heap, i, keys(i), values(i))
            heap%nodes(i)%node%key = keys(i)
            heap%nodes(i)%node%value = values(i)
         END DO
         ! Sort from the last full subtree
         first = get_parent(n)
         DO i = first, 1, -1
            CALL bubble_down(heap, i, error)
         ENDDO
      ENDIF
   END SUBROUTINE heap_fill

! **************************************************************************************************
!> \brief Returns the first heap element without removing it.
!> \param heap ...
!> \param key ...
!> \param value ...
!> \param found ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE heap_get_first(heap, key, value, found, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER(KIND=keyt), INTENT(OUT)                    :: key
      INTEGER(KIND=valt), INTENT(OUT)                    :: value
      LOGICAL, INTENT(OUT)                               :: found, error

!

      error = .FALSE.
      IF (heap%n .LT. 1) THEN
         found = .FALSE.
      ELSE
         found = .TRUE.
         CALL node_get(heap, 1, key, value)
      ENDIF
   END SUBROUTINE heap_get_first

! **************************************************************************************************
!> \brief Changes the value of the minimum heap element and
!>        rebalances the heap.
!> \param heap ...
!> \param value ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE heap_reset_first(heap, value, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER(KIND=valt), INTENT(IN)                     :: value
      LOGICAL, INTENT(OUT)                               :: error

!

      error = .FALSE.
      IF (heap%n .LT. 1) THEN
         error = .TRUE.
      ELSE
         heap%nodes(1)%node%value = value
         CALL bubble_down(heap, 1, error)
      ENDIF
   END SUBROUTINE heap_reset_first

! **************************************************************************************************
!> \brief Returns and removes the first heap element and rebalances
!>        the heap.
!> \param heap ...
!> \param key ...
!> \param value ...
!> \param found ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE heap_pop(heap, key, value, found, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER(KIND=keyt), INTENT(OUT)                    :: key
      INTEGER(KIND=valt), INTENT(OUT)                    :: value
      LOGICAL, INTENT(OUT)                               :: found, error

!

      CALL heap_get_first(heap, key, value, found, error)
      IF (.NOT. error .AND. found) THEN
         IF (heap%n .GT. 1) THEN
            CALL heap_copy_node(heap, 1, heap%n)
            heap%n = heap%n-1
            CALL bubble_down(heap, 1, error)
         ELSE
            heap%n = heap%n-1
         ENDIF
      ENDIF
   END SUBROUTINE heap_pop

! **************************************************************************************************
!> \brief Adds an element to the heap and rebalances it.
!> \param heap ...
!> \param key ...
!> \param value ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE heap_add(heap, key, value, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER(KIND=valt), INTENT(IN)                     :: value
      LOGICAL, INTENT(OUT)                               :: error

      INTEGER                                            :: new_n, new_pos

!

      error = .FALSE.
      new_n = heap%n+1
      CALL heap_reallocate(heap, new_n)
      heap%n = new_n
      CALL node_set(heap, new_n, key, value)
      CALL bubble_up(heap, new_n, new_pos, error)
      CALL bubble_down(heap, new_pos, error)
   END SUBROUTINE heap_add

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param e1 ...
!> \param e2 ...
! **************************************************************************************************
   PURE SUBROUTINE heap_swap(heap, e1, e2)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER, INTENT(IN)                                :: e1, e2

      TYPE(heap_node)                                    :: tmp_node

!

      tmp_node = heap%nodes(e1)%node
      heap%nodes(e1)%node = heap%nodes(e2)%node
      heap%nodes(e2)%node = tmp_node
   END SUBROUTINE heap_swap

! **************************************************************************************************
!> \brief Sets node e1 to e2
!> \param heap ...
!> \param e1 ...
!> \param e2 ...
! **************************************************************************************************
   PURE SUBROUTINE heap_copy_node(heap, e1, e2)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER, INTENT(IN)                                :: e1, e2

!

      heap%nodes(e1)%node = heap%nodes(e2)%node
   END SUBROUTINE heap_copy_node

! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   PURE SUBROUTINE node_get_node(node, key, value)
      TYPE(heap_node), INTENT(IN)                        :: node
      INTEGER(KIND=keyt), INTENT(OUT)                    :: key
      INTEGER(KIND=valt), INTENT(OUT)                    :: value

!

      key = node%key
      value = node%value
   END SUBROUTINE node_get_node
! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param n ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   PURE SUBROUTINE node_get_heap(heap, n, key, value)
      TYPE(heap_t), INTENT(IN)                           :: heap
      INTEGER, INTENT(IN)                                :: n
      INTEGER(KIND=keyt), INTENT(OUT)                    :: key
      INTEGER(KIND=valt), INTENT(OUT)                    :: value

!

      CALL node_get_node(heap%nodes(n)%node, key, value)
   END SUBROUTINE node_get_heap
! **************************************************************************************************
!> \brief ...
!> \param node ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   PURE SUBROUTINE node_set_node(node, key, value)
      TYPE(heap_node), INTENT(OUT)                       :: node
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER(KIND=valt), INTENT(IN)                     :: value

!

      node%key = key
      node%value = value
   END SUBROUTINE node_set_node
! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param n ...
!> \param key ...
!> \param value ...
! **************************************************************************************************
   PURE SUBROUTINE node_set_heap(heap, n, key, value)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER, INTENT(IN)                                :: n
      INTEGER(KIND=keyt), INTENT(IN)                     :: key
      INTEGER(KIND=valt), INTENT(IN)                     :: value

!

      CALL node_set_node(heap%nodes(n)%node, key, value)
   END SUBROUTINE node_set_heap

! **************************************************************************************************
!> \brief Balances a heap by bubbling down from the given element.
!> \param heap ...
!> \param first ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE bubble_down(heap, first, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER, INTENT(IN)                                :: first
      LOGICAL, INTENT(OUT)                               :: error

      INTEGER                                            :: e, left_child, right_child, smallest
      INTEGER(kind=valt)                                 :: left_child_value, min_value, &
                                                            right_child_value
      LOGICAL                                            :: all_done

!

      IF (first .GT. heap%n .OR. first .LT. 1) THEN
         error = .TRUE.
         RETURN
      ENDIF
      error = .FALSE.
      e = first
      all_done = .FALSE.
      ! Check whether we are finished, i.e,. whether the element to
      ! bubble down is childless.
      DO WHILE (e .LE. get_parent(heap%n) .AND. .NOT. all_done)
         ! Determines which node (current, left, or right child) has the
         ! smallest value.
         smallest = e
         min_value = get_value(heap, e)
         left_child = get_left_child(e)
         IF (left_child .LE. heap%n) THEN
            left_child_value = get_value(heap, left_child)
            IF (left_child_value .LT. min_value) THEN
               min_value = left_child_value
               smallest = left_child
            ENDIF
         ENDIF
         right_child = left_child+1
         IF (right_child .LE. heap%n) THEN
            right_child_value = get_value(heap, right_child)
            IF (right_child_value .LT. min_value) THEN
               min_value = right_child_value
               smallest = right_child
            ENDIF
         ENDIF
         !
         CALL heap_swap(heap, e, smallest)
         IF (smallest .EQ. e) THEN
            all_done = .TRUE.
         ELSE
            e = smallest
         ENDIF
      ENDDO
   END SUBROUTINE bubble_down

! **************************************************************************************************
!> \brief Balances a heap by bubbling up from the given element.
!> \param heap ...
!> \param first ...
!> \param new_pos ...
!> \param error ...
! **************************************************************************************************
   SUBROUTINE bubble_up(heap, first, new_pos, error)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER, INTENT(IN)                                :: first
      INTEGER, INTENT(OUT)                               :: new_pos
      LOGICAL, INTENT(OUT)                               :: error

      INTEGER                                            :: e, parent
      INTEGER(kind=valt)                                 :: my_value, parent_value
      LOGICAL                                            :: all_done

!

      IF (first .GT. heap%n .OR. first .LT. 1) THEN
         error = .TRUE.
         RETURN
      ENDIF
      error = .FALSE.
      e = first
      all_done = .FALSE.
      IF (e .GT. 1) THEN
         my_value = get_value(heap, e)
      ENDIF
      ! Check whether we are finished, i.e,. whether the element to
      ! bubble up is an orphan.
      new_pos = e
      DO WHILE (e .GT. 1 .AND. .NOT. all_done)
         ! Switches the parent and the current element if the current
         ! element's value is greater than the parent's value.
         parent = get_parent(e)
         parent_value = get_value(heap, parent)
         IF (my_value .LT. parent_value) THEN
            CALL heap_swap(heap, e, parent)
            e = parent
         ELSE
            all_done = .TRUE.
         ENDIF
      ENDDO
      new_pos = e
   END SUBROUTINE bubble_up

! **************************************************************************************************
!> \brief ...
!> \param heap ...
!> \param n ...
! **************************************************************************************************
   SUBROUTINE heap_reallocate(heap, n)
      TYPE(heap_t), INTENT(INOUT)                        :: heap
      INTEGER, INTENT(IN)                                :: n

      INTEGER, PARAMETER                                 :: min_size = 10
      REAL, PARAMETER                                    :: oversize = 1.4

      INTEGER                                            :: n_allocate
      TYPE(heap_node_e), DIMENSION(:), POINTER           :: new_nodes, tmp_nodes

      IF (n .GT. SIZE(heap%nodes)) THEN
         n_allocate = MAX(MAX(n, min_size), INT(n*oversize))
         ALLOCATE (new_nodes(n_allocate))
         new_nodes(1:heap%n) = heap%nodes(1:heap%n)
         tmp_nodes => heap%nodes
         heap%nodes => new_nodes
         DEALLOCATE (tmp_nodes)
      ENDIF
   END SUBROUTINE heap_reallocate

END MODULE min_heap
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Accelerator support
!> \author  Ole Schuett
!> \date    2013-04
! **************************************************************************************************
MODULE acc_hostmem
  USE acc_kinds,                       ONLY: int_4,&
                                             int_4_size,&
                                             int_8,&
                                             int_8_size,&
                                             real_4,&
                                             real_4_size,&
                                             real_8,&
                                             real_8_size
  USE acc_stream,                      ONLY: acc_stream_associated,&
                                             acc_stream_cptr,&
                                             acc_stream_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'acc_hostmem'

   LOGICAL, PARAMETER :: careful_mod = .TRUE.

   PUBLIC :: acc_hostmem_allocate, acc_hostmem_deallocate

   INTERFACE acc_hostmem_allocate
      MODULE PROCEDURE acc_hostmem_alloc_i, acc_hostmem_alloc_l
      MODULE PROCEDURE acc_hostmem_alloc_r, acc_hostmem_alloc_d
      MODULE PROCEDURE acc_hostmem_alloc_c, acc_hostmem_alloc_z
      MODULE PROCEDURE acc_hostmem_alloc_i_2D, acc_hostmem_alloc_l_2D
      MODULE PROCEDURE acc_hostmem_alloc_r_2D, acc_hostmem_alloc_d_2D
      MODULE PROCEDURE acc_hostmem_alloc_c_2D, acc_hostmem_alloc_z_2D
   END INTERFACE

   INTERFACE acc_hostmem_deallocate
      MODULE PROCEDURE acc_hostmem_dealloc_i, acc_hostmem_dealloc_l
      MODULE PROCEDURE acc_hostmem_dealloc_r, acc_hostmem_dealloc_d
      MODULE PROCEDURE acc_hostmem_dealloc_c, acc_hostmem_dealloc_z
      MODULE PROCEDURE acc_hostmem_dealloc_i_2D, acc_hostmem_dealloc_l_2D
      MODULE PROCEDURE acc_hostmem_dealloc_r_2D, acc_hostmem_dealloc_d_2D
      MODULE PROCEDURE acc_hostmem_dealloc_c_2D, acc_hostmem_dealloc_z_2D
   END INTERFACE


CONTAINS

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016 the CP2K developers group                                            !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Allocates 1D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n size given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_i (host_mem, n, stream)
    INTEGER(KIND=int_4), DIMENSION(:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",27,"acc_hostmem_alloc_i: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_i



! **************************************************************************************************
!> \brief Allocates 2D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n1 sizes given in terms of item-count (not bytes!)
!> \param n2 sizes given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_i_2D (host_mem, n1, n2, stream)
    INTEGER(KIND=int_4), DIMENSION(:,:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n1, n2
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n1))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n2))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",57,"acc_hostmem_alloc_i_2D: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_i_2D


! **************************************************************************************************
!> \brief Deallocates a 1D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_i (host_mem, stream)
    INTEGER(KIND=int_4), DIMENSION(:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_i', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",81,"acc_hostmem_dealloc_i: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_i


! **************************************************************************************************
!> \brief Deallocates a 2D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_i_2D (host_mem, stream)
    INTEGER(KIND=int_4), DIMENSION(:,:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_i_2D', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",105,"acc_hostmem_dealloc_i: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_i_2D
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016 the CP2K developers group                                            !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Allocates 1D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n size given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_l (host_mem, n, stream)
    INTEGER(KIND=int_8), DIMENSION(:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",27,"acc_hostmem_alloc_l: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_l



! **************************************************************************************************
!> \brief Allocates 2D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n1 sizes given in terms of item-count (not bytes!)
!> \param n2 sizes given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_l_2D (host_mem, n1, n2, stream)
    INTEGER(KIND=int_8), DIMENSION(:,:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n1, n2
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n1))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n2))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",57,"acc_hostmem_alloc_l_2D: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_l_2D


! **************************************************************************************************
!> \brief Deallocates a 1D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_l (host_mem, stream)
    INTEGER(KIND=int_8), DIMENSION(:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_l', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",81,"acc_hostmem_dealloc_l: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_l


! **************************************************************************************************
!> \brief Deallocates a 2D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_l_2D (host_mem, stream)
    INTEGER(KIND=int_8), DIMENSION(:,:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_l_2D', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",105,"acc_hostmem_dealloc_l: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_l_2D
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016 the CP2K developers group                                            !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Allocates 1D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n size given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_r (host_mem, n, stream)
    REAL(kind=real_4), DIMENSION(:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",27,"acc_hostmem_alloc_r: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_r



! **************************************************************************************************
!> \brief Allocates 2D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n1 sizes given in terms of item-count (not bytes!)
!> \param n2 sizes given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_r_2D (host_mem, n1, n2, stream)
    REAL(kind=real_4), DIMENSION(:,:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n1, n2
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n1))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n2))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",57,"acc_hostmem_alloc_r_2D: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_r_2D


! **************************************************************************************************
!> \brief Deallocates a 1D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_r (host_mem, stream)
    REAL(kind=real_4), DIMENSION(:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_r', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",81,"acc_hostmem_dealloc_r: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_r


! **************************************************************************************************
!> \brief Deallocates a 2D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_r_2D (host_mem, stream)
    REAL(kind=real_4), DIMENSION(:,:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_r_2D', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",105,"acc_hostmem_dealloc_r: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_r_2D
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016 the CP2K developers group                                            !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Allocates 1D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n size given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_d (host_mem, n, stream)
    REAL(kind=real_8), DIMENSION(:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",27,"acc_hostmem_alloc_d: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_d



! **************************************************************************************************
!> \brief Allocates 2D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n1 sizes given in terms of item-count (not bytes!)
!> \param n2 sizes given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_d_2D (host_mem, n1, n2, stream)
    REAL(kind=real_8), DIMENSION(:,:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n1, n2
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n1))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n2))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",57,"acc_hostmem_alloc_d_2D: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_d_2D


! **************************************************************************************************
!> \brief Deallocates a 1D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_d (host_mem, stream)
    REAL(kind=real_8), DIMENSION(:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_d', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",81,"acc_hostmem_dealloc_d: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_d


! **************************************************************************************************
!> \brief Deallocates a 2D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_d_2D (host_mem, stream)
    REAL(kind=real_8), DIMENSION(:,:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_d_2D', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",105,"acc_hostmem_dealloc_d: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_d_2D
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016 the CP2K developers group                                            !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Allocates 1D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n size given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_c (host_mem, n, stream)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",27,"acc_hostmem_alloc_c: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_c



! **************************************************************************************************
!> \brief Allocates 2D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n1 sizes given in terms of item-count (not bytes!)
!> \param n2 sizes given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_c_2D (host_mem, n1, n2, stream)
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n1, n2
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n1))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n2))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",57,"acc_hostmem_alloc_c_2D: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_c_2D


! **************************************************************************************************
!> \brief Deallocates a 1D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_c (host_mem, stream)
    COMPLEX(kind=real_4), DIMENSION(:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_c', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",81,"acc_hostmem_dealloc_c: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_c


! **************************************************************************************************
!> \brief Deallocates a 2D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_c_2D (host_mem, stream)
    COMPLEX(kind=real_4), DIMENSION(:,:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_c_2D', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",105,"acc_hostmem_dealloc_c: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_c_2D
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016 the CP2K developers group                                            !
!--------------------------------------------------------------------------------------------------!


! **************************************************************************************************
!> \brief Allocates 1D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n size given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_z (host_mem, n, stream)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",27,"acc_hostmem_alloc_z: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_z



! **************************************************************************************************
!> \brief Allocates 2D fortan-array as cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param n1 sizes given in terms of item-count (not bytes!)
!> \param n2 sizes given in terms of item-count (not bytes!)
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_alloc_z_2D (host_mem, n1, n2, stream)
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER           :: host_mem
    INTEGER, INTENT(IN)                      :: n1, n2
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n1))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n2))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",57,"acc_hostmem_alloc_z_2D: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_alloc_z_2D


! **************************************************************************************************
!> \brief Deallocates a 1D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_z (host_mem, stream)
    COMPLEX(kind=real_8), DIMENSION(:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_z', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",81,"acc_hostmem_dealloc_z: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_z


! **************************************************************************************************
!> \brief Deallocates a 2D fortan-array, which is cuda host-pinned memory.
!> \param host_mem pointer to array
!> \param stream ...
!> \author  Ole Schuett
! **************************************************************************************************
  SUBROUTINE acc_hostmem_dealloc_z_2D (host_mem, stream)
    COMPLEX(kind=real_8), DIMENSION(:,:), &
      POINTER                                :: host_mem
    TYPE(acc_stream_type), INTENT(IN)        :: stream
    CHARACTER(len=*), PARAMETER :: routineN = 'acc_hostmem_dealloc_z_2D', &
      routineP = moduleN//':'//routineN

    IF (SIZE (host_mem) == 0) RETURN
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(host_mem))==-1) EXIT ;  END DO ; ENDIF
    IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(stream))==-1) EXIT ;  END DO ; ENDIF
    CALL cp__b("acc/acc_hostmem.F",105,"acc_hostmem_dealloc_z: ACC not compiled in.")
  END SUBROUTINE acc_hostmem_dealloc_z_2D

! **************************************************************************************************
!> \brief Helper-routine performing allocation of host-pinned cuda memory.
!> \param host_mem_c_ptr pointer to allocated memory
!> \param n_bytes number of bytes to allocate
!> \param stream ...
! **************************************************************************************************


END MODULE acc_hostmem
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR pointer and unmanaged array utilities
!> \author  Urban Borstndik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2010-02-18 Moved from dbcsr_util
!> - 2010-06-18 Moved all pointer methods into here.
! **************************************************************************************************
MODULE dbcsr_ptr_util
   USE acc_hostmem,                     ONLY: acc_hostmem_allocate,&
                                              acc_hostmem_deallocate
   USE dbcsr_data_types,                ONLY: dbcsr_data_obj,&
                                              dbcsr_memtype_default,&
                                              dbcsr_memtype_type,&
                                              dbcsr_type_complex_4,&
                                              dbcsr_type_complex_8,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_unimplemented_error_nr,&
                                              dbcsr_wrong_args_error
   USE kinds,                           ONLY: dp,&
                                              int_4,&
                                              int_8,&
                                              real_4,&
                                              real_8
   USE message_passing,                 ONLY: mp_allocate,&
                                              mp_deallocate
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_ptr_util'

   LOGICAL, PARAMETER :: careful_mod = .FALSE.

   PUBLIC :: ensure_array_size
   PUBLIC :: memory_allocate, memory_deallocate
   PUBLIC :: memory_zero
   PUBLIC :: pointer_view
   PUBLIC :: pointer_rank_remap2
   PUBLIC :: memory_copy

   INTERFACE ensure_array_size
      MODULE PROCEDURE ensure_array_size_i, ensure_array_size_l
      MODULE PROCEDURE ensure_array_size_s, ensure_array_size_d, &
         ensure_array_size_c, ensure_array_size_z
   END INTERFACE

   INTERFACE pointer_view
      MODULE PROCEDURE pointer_view_s, pointer_view_d, &
         pointer_view_c, pointer_view_z
      MODULE PROCEDURE pointer_view_i, pointer_view_l
      MODULE PROCEDURE pointer_view_a
   END INTERFACE

   INTERFACE pointer_rank_remap2
      MODULE PROCEDURE pointer_s_rank_remap2, pointer_d_rank_remap2, &
         pointer_c_rank_remap2, pointer_z_rank_remap2, &
         pointer_l_rank_remap2, pointer_i_rank_remap2
   END INTERFACE

   INTERFACE memory_copy
      MODULE PROCEDURE mem_copy_i, mem_copy_l, &
         mem_copy_s, mem_copy_d, &
         mem_copy_c, mem_copy_z
   END INTERFACE

   INTERFACE memory_zero
      MODULE PROCEDURE mem_zero_i, mem_zero_l
      MODULE PROCEDURE mem_zero_s, mem_zero_d, mem_zero_c, mem_zero_z
   END INTERFACE

   INTERFACE memory_allocate
      MODULE PROCEDURE mem_alloc_i, mem_alloc_l, mem_alloc_s, mem_alloc_d, mem_alloc_c, mem_alloc_z
      MODULE PROCEDURE mem_alloc_i_2d, mem_alloc_l_2d, mem_alloc_s_2d, mem_alloc_d_2d, mem_alloc_c_2d, mem_alloc_z_2d
   END INTERFACE

   INTERFACE memory_deallocate
      MODULE PROCEDURE mem_dealloc_i, mem_dealloc_l, mem_dealloc_s, mem_dealloc_d, mem_dealloc_c, mem_dealloc_z
      MODULE PROCEDURE mem_dealloc_i_2d, mem_dealloc_l_2d, mem_dealloc_s_2d, mem_dealloc_d_2d, mem_dealloc_c_2d, mem_dealloc_z_2d
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief Repoints a pointer into a part of a data area
!> \param[in,out] new_area    repoints this encapsulated pointer
!> \param[in] area            area to point into
!> \param[in] offset          point to this offset in area
!> \param[in] len             (optional) length of data area to point to
!> \retval narea2 copy of new_area
! **************************************************************************************************
   FUNCTION pointer_view_a(new_area, area, offset, len) RESULT(narea2)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: new_area
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      INTEGER, INTENT(IN)                                :: offset
      INTEGER, INTENT(IN), OPTIONAL                      :: len
      TYPE(dbcsr_data_obj)                               :: narea2

      CHARACTER(len=*), PARAMETER :: routineN = 'pointer_view_a', routineP = moduleN//':'//routineN

      CALL dbcsr_assert(area%d%data_type, "EQ", new_area%d%data_type, &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "Incompatible data types.", 117)
      IF (PRESENT(len)) THEN
         SELECT CASE (area%d%data_type)
         CASE (dbcsr_type_real_4)
            new_area%d%r_sp => area%d%r_sp(offset:offset+len-1)
         CASE (dbcsr_type_real_8)
            new_area%d%r_dp => area%d%r_dp(offset:offset+len-1)
         CASE (dbcsr_type_complex_4)
            new_area%d%c_sp => area%d%c_sp(offset:offset+len-1)
         CASE (dbcsr_type_complex_8)
            new_area%d%c_dp => area%d%c_dp(offset:offset+len-1)
         CASE default
            CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",129,"Invalid data type.")
         END SELECT
      ELSE
         SELECT CASE (area%d%data_type)
         CASE (dbcsr_type_real_4)
            new_area%d%r_sp => area%d%r_sp(offset:)
         CASE (dbcsr_type_real_8)
            new_area%d%r_dp => area%d%r_dp(offset:)
         CASE (dbcsr_type_complex_4)
            new_area%d%c_sp => area%d%c_sp(offset:)
         CASE (dbcsr_type_complex_8)
            new_area%d%c_dp => area%d%c_dp(offset:)
         CASE default
            CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",142,"Invalid data type.")
         END SELECT
      ENDIF
      narea2 = new_area
   END FUNCTION pointer_view_a

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \retval view new pointer
! **************************************************************************************************
  FUNCTION pointer_view_i (original, lb, ub) RESULT (view)
    INTEGER(kind=int_4), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_i


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_i(array, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    INTEGER(kind=int_4), DIMENSION(:), POINTER                 :: array
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_i', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    INTEGER(kind=int_4), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       CALL dbcsr_assert (lb_new, "EQ", 1, &
            dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
            "Arrays must start at 1", 75)
       CALL mem_alloc_i (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_i (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       CALL mem_dealloc_i (array, mem_type=mem_type)
    ENDIF
    !
    ! Allocates the new array
    CALL dbcsr_assert (lb_new, "EQ", 1, &
         dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
         "Arrays must start at 1", 125)
    CALL mem_alloc_i (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",131,"Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_i (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_i (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_i (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       CALL mem_dealloc_i (array, mem_type=mem_type)
    ELSEIF (pad) THEN
       !newarray(:) = 0
       CALL mem_zero_i (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_i

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_i (dst, src, n)
    INTEGER, INTENT(IN) :: n
    INTEGER(kind=int_4), DIMENSION(1:n), INTENT(OUT) :: dst
    INTEGER(kind=int_4), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_i

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_i (dst, n)
    INTEGER, INTENT(IN) :: n
    INTEGER(kind=int_4), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = 0
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_i


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_i (mem, n, mem_type)
    INTEGER(kind=int_4), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_i', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_i


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_i_2d (mem, sizes, mem_type)
    INTEGER(kind=int_4), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_i_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",234,"Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",237,"MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_i_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_i (mem, mem_type)
    INTEGER(kind=int_4), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_i', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_i


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_i_2d (mem, mem_type)
    INTEGER(kind=int_4), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_i', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",294,"Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",297,"MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_i_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_i_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    INTEGER(kind=int_4), DIMENSION(:, :), &
      POINTER                                :: r2p
    INTEGER(kind=int_4), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_i_rank_remap2
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \retval view new pointer
! **************************************************************************************************
  FUNCTION pointer_view_l (original, lb, ub) RESULT (view)
    INTEGER(kind=int_8), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_l


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_l(array, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    INTEGER(kind=int_8), DIMENSION(:), POINTER                 :: array
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_l', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    INTEGER(kind=int_8), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       CALL dbcsr_assert (lb_new, "EQ", 1, &
            dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
            "Arrays must start at 1", 75)
       CALL mem_alloc_l (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_l (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       CALL mem_dealloc_l (array, mem_type=mem_type)
    ENDIF
    !
    ! Allocates the new array
    CALL dbcsr_assert (lb_new, "EQ", 1, &
         dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
         "Arrays must start at 1", 125)
    CALL mem_alloc_l (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",131,"Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_l (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_l (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_l (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       CALL mem_dealloc_l (array, mem_type=mem_type)
    ELSEIF (pad) THEN
       !newarray(:) = 0
       CALL mem_zero_l (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_l

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_l (dst, src, n)
    INTEGER, INTENT(IN) :: n
    INTEGER(kind=int_8), DIMENSION(1:n), INTENT(OUT) :: dst
    INTEGER(kind=int_8), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_l

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_l (dst, n)
    INTEGER, INTENT(IN) :: n
    INTEGER(kind=int_8), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = 0
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_l


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_l (mem, n, mem_type)
    INTEGER(kind=int_8), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_l', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_l


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_l_2d (mem, sizes, mem_type)
    INTEGER(kind=int_8), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_l_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",234,"Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",237,"MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_l_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_l (mem, mem_type)
    INTEGER(kind=int_8), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_l', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_l


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_l_2d (mem, mem_type)
    INTEGER(kind=int_8), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_l', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",294,"Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",297,"MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_l_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_l_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    INTEGER(kind=int_8), DIMENSION(:, :), &
      POINTER                                :: r2p
    INTEGER(kind=int_8), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_l_rank_remap2
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \retval view new pointer
! **************************************************************************************************
  FUNCTION pointer_view_d (original, lb, ub) RESULT (view)
    REAL(kind=real_8), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_d


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_d(array, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    REAL(kind=real_8), DIMENSION(:), POINTER                 :: array
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    REAL(kind=real_8), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       CALL dbcsr_assert (lb_new, "EQ", 1, &
            dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
            "Arrays must start at 1", 75)
       CALL mem_alloc_d (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_d (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       CALL mem_dealloc_d (array, mem_type=mem_type)
    ENDIF
    !
    ! Allocates the new array
    CALL dbcsr_assert (lb_new, "EQ", 1, &
         dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
         "Arrays must start at 1", 125)
    CALL mem_alloc_d (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",131,"Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_d (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_d (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_d (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       CALL mem_dealloc_d (array, mem_type=mem_type)
    ELSEIF (pad) THEN
       !newarray(:) = 0.0_real_8
       CALL mem_zero_d (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_d

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_d (dst, src, n)
    INTEGER, INTENT(IN) :: n
    REAL(kind=real_8), DIMENSION(1:n), INTENT(OUT) :: dst
    REAL(kind=real_8), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_d

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_d (dst, n)
    INTEGER, INTENT(IN) :: n
    REAL(kind=real_8), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = 0.0_real_8
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_d


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_d (mem, n, mem_type)
    REAL(kind=real_8), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_d


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_d_2d (mem, sizes, mem_type)
    REAL(kind=real_8), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_d_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",234,"Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",237,"MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_d_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_d (mem, mem_type)
    REAL(kind=real_8), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_d_2d (mem, mem_type)
    REAL(kind=real_8), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",294,"Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",297,"MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_d_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_d_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    REAL(kind=real_8), DIMENSION(:, :), &
      POINTER                                :: r2p
    REAL(kind=real_8), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_d_rank_remap2
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \retval view new pointer
! **************************************************************************************************
  FUNCTION pointer_view_z (original, lb, ub) RESULT (view)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_z


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_z(array, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER                 :: array
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       CALL dbcsr_assert (lb_new, "EQ", 1, &
            dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
            "Arrays must start at 1", 75)
       CALL mem_alloc_z (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_z (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       CALL mem_dealloc_z (array, mem_type=mem_type)
    ENDIF
    !
    ! Allocates the new array
    CALL dbcsr_assert (lb_new, "EQ", 1, &
         dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
         "Arrays must start at 1", 125)
    CALL mem_alloc_z (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",131,"Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_z (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_z (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_z (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       CALL mem_dealloc_z (array, mem_type=mem_type)
    ELSEIF (pad) THEN
       !newarray(:) = CMPLX(0.0, 0.0, real_8)
       CALL mem_zero_z (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_z

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_z (dst, src, n)
    INTEGER, INTENT(IN) :: n
    COMPLEX(kind=real_8), DIMENSION(1:n), INTENT(OUT) :: dst
    COMPLEX(kind=real_8), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_z

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_z (dst, n)
    INTEGER, INTENT(IN) :: n
    COMPLEX(kind=real_8), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = CMPLX(0.0, 0.0, real_8)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_z


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_z (mem, n, mem_type)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_z', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_z


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_z_2d (mem, sizes, mem_type)
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_z_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",234,"Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",237,"MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_z_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_z (mem, mem_type)
    COMPLEX(kind=real_8), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_z', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_z


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_z_2d (mem, mem_type)
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_z', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",294,"Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",297,"MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_z_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_z_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    COMPLEX(kind=real_8), DIMENSION(:, :), &
      POINTER                                :: r2p
    COMPLEX(kind=real_8), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_z_rank_remap2
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \retval view new pointer
! **************************************************************************************************
  FUNCTION pointer_view_s (original, lb, ub) RESULT (view)
    REAL(kind=real_4), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_s


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_s(array, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    REAL(kind=real_4), DIMENSION(:), POINTER                 :: array
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    REAL(kind=real_4), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       CALL dbcsr_assert (lb_new, "EQ", 1, &
            dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
            "Arrays must start at 1", 75)
       CALL mem_alloc_s (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_s (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       CALL mem_dealloc_s (array, mem_type=mem_type)
    ENDIF
    !
    ! Allocates the new array
    CALL dbcsr_assert (lb_new, "EQ", 1, &
         dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
         "Arrays must start at 1", 125)
    CALL mem_alloc_s (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",131,"Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_s (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_s (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_s (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       CALL mem_dealloc_s (array, mem_type=mem_type)
    ELSEIF (pad) THEN
       !newarray(:) = 0.0_real_4
       CALL mem_zero_s (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_s

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_s (dst, src, n)
    INTEGER, INTENT(IN) :: n
    REAL(kind=real_4), DIMENSION(1:n), INTENT(OUT) :: dst
    REAL(kind=real_4), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_s

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_s (dst, n)
    INTEGER, INTENT(IN) :: n
    REAL(kind=real_4), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = 0.0_real_4
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_s


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_s (mem, n, mem_type)
    REAL(kind=real_4), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_s', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_s


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_s_2d (mem, sizes, mem_type)
    REAL(kind=real_4), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_s_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",234,"Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",237,"MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_s_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_s (mem, mem_type)
    REAL(kind=real_4), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_s', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_s


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_s_2d (mem, mem_type)
    REAL(kind=real_4), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_s', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",294,"Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",297,"MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_s_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_s_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    REAL(kind=real_4), DIMENSION(:, :), &
      POINTER                                :: r2p
    REAL(kind=real_4), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_s_rank_remap2
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Returns a pointer with different bounds.
!> \param[in] original   original data pointer
!> \param[in] lb lower and upper bound for the new pointer view
!> \param[in] ub lower and upper bound for the new pointer view
!> \retval view new pointer
! **************************************************************************************************
  FUNCTION pointer_view_c (original, lb, ub) RESULT (view)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER :: original, view
    INTEGER, INTENT(IN)                  :: lb, ub
    view => original(lb:ub)
  END FUNCTION pointer_view_c


! **************************************************************************************************
!> \brief Ensures that an array is appropriately large.
!> \param[in,out] array       array to verify and possibly resize
!> \param[in] lb    (optional) desired array lower bound
!> \param[in] ub    desired array upper bound
!> \param[in] factor          (optional) factor by which to exagerrate
!>                            enlargements
!> \param[in] nocopy          (optional) copy array on enlargement; default
!>                            is to copy
!> \param[in] memory_type     (optional) use special memory
!> \param[in] zero_pad        (optional) zero new allocations; default is to
!>                            write nothing
! **************************************************************************************************
  SUBROUTINE ensure_array_size_c(array, lb, ub, factor,&
       nocopy, memory_type, zero_pad)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER                 :: array
    INTEGER, INTENT(IN), OPTIONAL                  :: lb
    INTEGER, INTENT(IN)                            :: ub
    REAL(KIND=dp), INTENT(IN), OPTIONAL            :: factor
    LOGICAL, INTENT(IN), OPTIONAL                  :: nocopy, zero_pad
    TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL :: memory_type

    CHARACTER(len=*), PARAMETER :: routineN = 'ensure_array_size_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: lb_new, lb_orig, &
                                                ub_new, ub_orig, old_size,&
                                                size_increase
    TYPE(dbcsr_memtype_type)                 :: mem_type
    LOGICAL                                  :: dbg, docopy, &
                                                pad
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: newarray

!   ---------------------------------------------------------------------------
    !CALL timeset(routineN, error_handler)
    dbg = .FALSE.

    IF (PRESENT (nocopy)) THEN
       docopy = .NOT. nocopy
    ELSE
       docopy = .TRUE.
    ENDIF
    IF (PRESENT (memory_type)) THEN
       mem_type = memory_type
    ELSE
       mem_type = dbcsr_memtype_default
    ENDIF
    lb_new = 1
    IF (PRESENT (lb)) lb_new = lb
    pad = .FALSE.
    IF (PRESENT (zero_pad)) pad = zero_pad
    !> Creates a new array if it doesn't yet exist.
    IF (.NOT.ASSOCIATED(array)) THEN
       CALL dbcsr_assert (lb_new, "EQ", 1, &
            dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
            "Arrays must start at 1", 75)
       CALL mem_alloc_c (array, ub, mem_type=mem_type)
       IF (pad .AND. ub .GT. 0) CALL mem_zero_c (array, ub)
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    lb_orig = LBOUND(array,1)
    ub_orig = UBOUND(array,1)
    old_size = ub_orig - lb_orig + 1
    ! The existing array is big enough.
    IF (lb_orig.LE.lb_new .AND. ub_orig.GE.ub) THEN
       !CALL timestop(error_handler)
       RETURN
    ENDIF
    ! A reallocation must be performed
    IF(dbg) WRITE(*,*)routineP//' Current bounds are',lb_orig,':',ub_orig,&
         '; special?' !,mem_type
    !CALL timeset(routineN,timing_handle)
    IF (lb_orig.GT.lb_new) THEN
       IF (PRESENT(factor)) THEN
          size_increase = lb_orig - lb_new
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          lb_new = MIN (lb_orig, lb_new - size_increase)
       ELSE
          lb_new = lb_orig
       ENDIF
    ENDIF
    IF (ub_orig.LT.ub) THEN
       IF (PRESENT(factor)) THEN
          size_increase = ub - ub_orig
          size_increase = MAX (NINT(size_increase*factor),&
                               NINT(old_size*(factor-1)),0)
          ub_new = MAX (ub_orig, ub + size_increase)
       ELSE
          ub_new = ub
       ENDIF
    ELSE
       ub_new = ub
    ENDIF
    IF(dbg) WRITE(*,*)routineP//' Resizing to bounds',lb_new,':',ub_new
    !
    ! Deallocates the old array if it's not needed to copy the old data.
    IF(.NOT.docopy) THEN
       CALL mem_dealloc_c (array, mem_type=mem_type)
    ENDIF
    !
    ! Allocates the new array
    CALL dbcsr_assert (lb_new, "EQ", 1, &
         dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN,&
         "Arrays must start at 1", 125)
    CALL mem_alloc_c (newarray, ub_new-lb_new+1, mem_type)
    !
    ! Now copy and/or zero pad.
    IF(docopy) THEN
       IF(dbg .AND. (lb_new.GT.lb_orig .OR. ub_new.LT.ub_orig))&
            CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",131,"Old extent exceeds the new one.")
       IF (ub_orig-lb_orig+1 .gt. 0) THEN
          !newarray(lb_orig:ub_orig) = array(lb_orig:ub_orig)
          CALL mem_copy_c (newarray(lb_orig:ub_orig),&
               array(lb_orig:ub_orig), ub_orig-lb_orig+1)
       ENDIF
       IF (pad) THEN
          !newarray(lb_new:lb_orig-1) = 0
          CALL mem_zero_c (newarray(lb_new:lb_orig-1), (lb_orig-1)-lb_new+1)
          !newarray(ub_orig+1:ub_new) = 0
          CALL mem_zero_c (newarray(ub_orig+1:ub_new), ub_new-(ub_orig+1)+1)
       ENDIF
       CALL mem_dealloc_c (array, mem_type=mem_type)
    ELSEIF (pad) THEN
       !newarray(:) = CMPLX(0.0, 0.0, real_4)
       CALL mem_zero_c (newarray, SIZE(newarray))
    ENDIF
    array => newarray
    IF (dbg) WRITE(*,*)routineP//' New array size', SIZE(array)
    !CALL timestop(error_handler)
  END SUBROUTINE ensure_array_size_c

! **************************************************************************************************
!> \brief Copies memory area
!> \param[out] dst   destination memory
!> \param[in] src    source memory
!> \param[in] n      length of copy
! **************************************************************************************************
  SUBROUTINE mem_copy_c (dst, src, n)
    INTEGER, INTENT(IN) :: n
    COMPLEX(kind=real_4), DIMENSION(1:n), INTENT(OUT) :: dst
    COMPLEX(kind=real_4), DIMENSION(1:n), INTENT(IN) :: src
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst,src)
    dst(:) = src(:)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_copy_c

! **************************************************************************************************
!> \brief Zeros memory area
!> \param[out] dst   destination memory
!> \param[in] n      length of elements to zero
! **************************************************************************************************
  SUBROUTINE mem_zero_c (dst, n)
    INTEGER, INTENT(IN) :: n
    COMPLEX(kind=real_4), DIMENSION(1:n), INTENT(OUT) :: dst
    !$OMP PARALLEL WORKSHARE DEFAULT(none) SHARED(dst)
    dst(:) = CMPLX(0.0, 0.0, real_4)
    !$OMP END PARALLEL WORKSHARE
  END SUBROUTINE mem_zero_c


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] n           length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_c (mem, n, mem_type)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER        :: mem
    INTEGER, INTENT(IN)                   :: n
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_c', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. n>1) THEN
       CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
!$omp critical(allocate)
       CALL mp_allocate(mem, n)
!$omp end critical(allocate)
    ELSE
       ALLOCATE(mem(n))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_c


! **************************************************************************************************
!> \brief Allocates memory
!> \param[out] mem        memory to allocate
!> \param[in] sizes length of elements to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_alloc_c_2d (mem, sizes, mem_type)
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER      :: mem
    INTEGER, DIMENSION(2), INTENT(IN)     :: sizes
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_alloc_c_2d', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",234,"Accelerator hostalloc not supported for 2D arrays.")
       !CALL acc_hostmem_allocate(mem, n, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",237,"MPI allocate not supported for 2D arrays.")
       !CALL mp_allocate(mem, n)
    ELSE
       ALLOCATE(mem(sizes(1), sizes(2)))
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_alloc_c_2d


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_c (mem, mem_type)
    COMPLEX(kind=real_4), DIMENSION(:), POINTER        :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_c', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc .AND. SIZE(mem)>1) THEN
       CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_c


! **************************************************************************************************
!> \brief Deallocates memory
!> \param[out] mem        memory to allocate
!> \param[in] mem_type    memory type
! **************************************************************************************************
  SUBROUTINE mem_dealloc_c_2d (mem, mem_type)
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER      :: mem
    TYPE(dbcsr_memtype_type), INTENT(IN)  :: mem_type
    CHARACTER(len=*), PARAMETER :: routineN = 'mem_dealloc_c', &
      routineP = moduleN//':'//routineN
    INTEGER                               :: error_handle
!   ---------------------------------------------------------------------------

    IF (careful_mod) &
       CALL timeset (routineN, error_handle)

    IF(mem_type%acc_hostalloc) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",294,"Accelerator host deallocate not supported for 2D arrays.")
       !CALL acc_hostmem_deallocate(mem, mem_type%acc_stream)
    ELSE IF(mem_type%mpi) THEN
       CALL cp__b("dbcsr/data/dbcsr_ptr_util.F",297,"MPI deallocate not supported for 2D arrays.")
       !CALL mp_deallocate(mem)
    ELSE
       DEALLOCATE(mem)
    ENDIF

    IF (careful_mod) &
       CALL timestop (error_handle)
  END SUBROUTINE mem_dealloc_c_2d


! **************************************************************************************************
!> \brief Sets a rank-2 pointer to rank-1 data using Fortran 2003 pointer
!>        rank remapping.
!> \param r2p ...
!> \param d1 ...
!> \param d2 ...
!> \param r1p ...
! **************************************************************************************************
  SUBROUTINE pointer_c_rank_remap2 (r2p, d1, d2, r1p)
    INTEGER, INTENT(IN)                      :: d1, d2
    COMPLEX(kind=real_4), DIMENSION(:, :), &
      POINTER                                :: r2p
    COMPLEX(kind=real_4), DIMENSION(:), &
      POINTER                                :: r1p

    r2p(1:d1,1:d2) => r1p(1:d1*d2)
  END SUBROUTINE pointer_c_rank_remap2

END MODULE dbcsr_ptr_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR data methods
!> \author  Urban Borstnik
!> \date    2010-06-15
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2010-02-18 Moved from dbcsr_methods
! **************************************************************************************************
MODULE dbcsr_data_methods_low

   USE acc_devmem,                      ONLY: acc_devmem_allocate_bytes,&
                                              acc_devmem_allocated,&
                                              acc_devmem_deallocate
   USE acc_event,                       ONLY: acc_event_create,&
                                              acc_event_destroy
   USE dbcsr_data_types,                ONLY: &
        dbcsr_data_area_type, dbcsr_data_obj, dbcsr_datatype_sizeof, dbcsr_memtype_type, &
        dbcsr_scalar_type, dbcsr_type_complex_4, dbcsr_type_complex_4_2d, dbcsr_type_complex_8, &
        dbcsr_type_complex_8_2d, dbcsr_type_int_4, dbcsr_type_int_8, dbcsr_type_real_4, &
        dbcsr_type_real_4_2d, dbcsr_type_real_8, dbcsr_type_real_8_2d
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_caller_error,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_ptr_util,                  ONLY: memory_allocate,&
                                              memory_deallocate,&
                                              memory_zero,&
                                              pointer_rank_remap2
   USE kinds,                           ONLY: dp,&
                                              real_4,&
                                              real_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_data_methods_low'

   PUBLIC :: dbcsr_type_is_2d, dbcsr_type_2d_to_1d, dbcsr_type_1d_to_2d
   PUBLIC :: dbcsr_scalar, dbcsr_scalar_one, dbcsr_scalar_zero, &
             dbcsr_scalar_are_equal, dbcsr_scalar_negative, &
             dbcsr_scalar_get_type, dbcsr_scalar_set_type, &
             dbcsr_scalar_fill_all, dbcsr_scalar_get_value, &
             dbcsr_data_valid
   PUBLIC :: dbcsr_data_init, dbcsr_data_hold, &
             dbcsr_data_get_size, dbcsr_data_get_type
   PUBLIC :: dbcsr_get_data, &
             dbcsr_data_set_pointer, &
             dbcsr_data_clear_pointer, &
             dbcsr_data_get_sizes, dbcsr_data_verify_bounds, &
             dbcsr_data_exists, dbcsr_data_get_memory_type
   PUBLIC :: dbcsr_data_zero
   PUBLIC :: dbcsr_data_set_size_referenced, dbcsr_data_get_size_referenced
   PUBLIC :: dbcsr_get_data_p, dbcsr_get_data_p_s, dbcsr_get_data_p_c, &
             dbcsr_get_data_p_d, dbcsr_get_data_p_z, &
             dbcsr_get_data_p_2d_s, dbcsr_get_data_p_2d_d, &
             dbcsr_get_data_p_2d_c, dbcsr_get_data_p_2d_z

   PUBLIC :: internal_data_allocate, internal_data_deallocate

!> \brief Encapsulates a scalar.
   INTERFACE dbcsr_scalar
      MODULE PROCEDURE dbcsr_scalar_s, dbcsr_scalar_d, &
         dbcsr_scalar_c, dbcsr_scalar_z
   END INTERFACE

!> \brief Encapsulates a scalar.
   INTERFACE dbcsr_scalar_get_value
      MODULE PROCEDURE dbcsr_scalar_get_value_s, dbcsr_scalar_get_value_d, &
         dbcsr_scalar_get_value_c, dbcsr_scalar_get_value_z
   END INTERFACE

   INTERFACE dbcsr_data_set_pointer
      MODULE PROCEDURE set_data_p_s, set_data_p_d, set_data_p_c, set_data_p_z
      MODULE PROCEDURE set_data_p_2d_s, set_data_p_2d_d, &
         set_data_p_2d_c, set_data_p_2d_z
      MODULE PROCEDURE set_data_area_area
   END INTERFACE

   INTERFACE dbcsr_get_data
      MODULE PROCEDURE get_data_s, get_data_d, get_data_c, get_data_z
      MODULE PROCEDURE get_data_2d_s, get_data_2d_d, get_data_2d_c, get_data_2d_z
   END INTERFACE

   INTERFACE dbcsr_get_data_p
      MODULE PROCEDURE dbcsr_get_data_c_s, dbcsr_get_data_c_c, &
         dbcsr_get_data_c_d, dbcsr_get_data_c_z
   END INTERFACE

   INTERFACE dbcsr_get_data_cptr
      MODULE PROCEDURE dbcsr_get_data_c_s, dbcsr_get_data_c_c, &
         dbcsr_get_data_c_d, dbcsr_get_data_c_z
   END INTERFACE

   INTERFACE dbcsr_data_get_sizes
      MODULE PROCEDURE dbcsr_data_get_sizes_any
      MODULE PROCEDURE dbcsr_data_get_sizes_1, dbcsr_data_get_sizes_2
   END INTERFACE

   LOGICAL, PARAMETER :: careful_mod = .FALSE.
   LOGICAL, PARAMETER :: debug_mod = .FALSE.

CONTAINS

! **************************************************************************************************
!> \brief Returns data type of a data area
!> \param[in] area         data area
!> \retval data_type       data type of the data area
! **************************************************************************************************
   PURE FUNCTION dbcsr_data_get_type(area) RESULT(data_type)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      INTEGER                                            :: data_type

      data_type = area%d%data_type
   END FUNCTION dbcsr_data_get_type

! **************************************************************************************************
!> \brief ...
!> \param area ...
!> \retval memory_type ...
! **************************************************************************************************
   FUNCTION dbcsr_data_get_memory_type(area) RESULT(memory_type)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      TYPE(dbcsr_memtype_type)                           :: memory_type

      memory_type = area%d%memory_type
   END FUNCTION dbcsr_data_get_memory_type

! **************************************************************************************************
!> \brief Initializes a data area
!> \param[inout] area         data area
! **************************************************************************************************
   SUBROUTINE dbcsr_data_init(area)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area

      NULLIFY (area%d)
   END SUBROUTINE dbcsr_data_init

! **************************************************************************************************
!> \brief Allocates pointers in the data type
!> \param[in,out] area        internal structure holding array pointers
!> \param[in] sizes           sizes to allocate to
! **************************************************************************************************
   SUBROUTINE internal_data_allocate(area, sizes)
      TYPE(dbcsr_data_area_type), INTENT(INOUT)          :: area
      INTEGER, DIMENSION(:), INTENT(IN)                  :: sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'internal_data_allocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      IF (debug_mod) &
         WRITE (*, *) routineN//" Setting to sizes", sizes
      IF (dbcsr_type_is_2d(area%data_type)) THEN
         CALL dbcsr_assert(SIZE(sizes), "EQ", 2, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "Sizes must have 2 elements for 2-D data", 171)
      ELSE
         CALL dbcsr_assert(SIZE(sizes), "EQ", 1, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "Sizes must have 1 elements for 1-D data", 175)
      ENDIF

      SELECT CASE (area%data_type)
      CASE (dbcsr_type_int_4)
         CALL memory_allocate(area%i4, n=sizes(1), mem_type=area%memory_type)
      CASE (dbcsr_type_int_8)
         CALL memory_allocate(area%i8, n=sizes(1), mem_type=area%memory_type)
      CASE (dbcsr_type_real_4)
         CALL memory_allocate(area%r_sp, n=sizes(1), mem_type=area%memory_type)
      CASE (dbcsr_type_real_8)
         CALL memory_allocate(area%r_dp, n=sizes(1), mem_type=area%memory_type)
      CASE (dbcsr_type_complex_4)
         CALL memory_allocate(area%c_sp, n=sizes(1), mem_type=area%memory_type)
      CASE (dbcsr_type_complex_8)
         CALL memory_allocate(area%c_dp, n=sizes(1), mem_type=area%memory_type)

      CASE (dbcsr_type_real_4_2d)
         CALL memory_allocate(area%r2_sp, sizes=sizes, mem_type=area%memory_type)
      CASE (dbcsr_type_real_8_2d)
         CALL memory_allocate(area%r2_dp, sizes=sizes, mem_type=area%memory_type)
      CASE (dbcsr_type_complex_4_2d)
         CALL memory_allocate(area%c2_sp, sizes=sizes, mem_type=area%memory_type)
      CASE (dbcsr_type_complex_8_2d)
         CALL memory_allocate(area%c2_dp, sizes=sizes, mem_type=area%memory_type)

      CASE default
         CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",202,"Invalid data type.")
      END SELECT

      IF (area%memory_type%acc_devalloc) THEN
         IF (sizes(1) > 1) &
            CALL acc_devmem_allocate_bytes(area%acc_devmem, dbcsr_datatype_sizeof(area%data_type)*sizes(1))
         CALL acc_event_create(area%acc_ready)
      ENDIF

      CALL timestop(error_handle)

   END SUBROUTINE internal_data_allocate

! **************************************************************************************************
!> \brief Allocates pointers in the data type
!> \param[in,out] area        internal structure holding array pointers
! **************************************************************************************************
   SUBROUTINE internal_data_deallocate(area)
      TYPE(dbcsr_data_area_type), INTENT(INOUT)          :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'internal_data_deallocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      SELECT CASE (area%data_type)
      CASE (dbcsr_type_int_4)
         CALL memory_deallocate(area%i4, mem_type=area%memory_type)
         NULLIFY (area%i4)
      CASE (dbcsr_type_int_8)
         CALL memory_deallocate(area%i8, mem_type=area%memory_type)
         NULLIFY (area%i8)
      CASE (dbcsr_type_real_4)
         CALL memory_deallocate(area%r_sp, mem_type=area%memory_type)
         NULLIFY (area%r_sp)
      CASE (dbcsr_type_real_8)
         CALL memory_deallocate(area%r_dp, mem_type=area%memory_type)
         NULLIFY (area%r_dp)
      CASE (dbcsr_type_complex_4)
         CALL memory_deallocate(area%c_sp, mem_type=area%memory_type)
         NULLIFY (area%c_sp)
      CASE (dbcsr_type_complex_8)
         CALL memory_deallocate(area%c_dp, mem_type=area%memory_type)
         NULLIFY (area%c_dp)

      CASE (dbcsr_type_real_4_2d)
         CALL memory_deallocate(area%r2_sp, mem_type=area%memory_type)
         NULLIFY (area%r2_sp)
      CASE (dbcsr_type_real_8_2d)
         CALL memory_deallocate(area%r2_dp, mem_type=area%memory_type)
         NULLIFY (area%r2_dp)
      CASE (dbcsr_type_complex_4_2d)
         CALL memory_deallocate(area%c2_sp, mem_type=area%memory_type)
         NULLIFY (area%c2_sp)
      CASE (dbcsr_type_complex_8_2d)
         CALL memory_deallocate(area%c2_dp, mem_type=area%memory_type)
         NULLIFY (area%c2_dp)

      CASE default
         CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",265,"Invalid data type.")
      END SELECT

      IF (area%memory_type%acc_devalloc) THEN
         IF (acc_devmem_allocated(area%acc_devmem)) &
            CALL acc_devmem_deallocate(area%acc_devmem)
         CALL acc_event_destroy(area%acc_ready)
      ENDIF

      CALL timestop(handle)
   END SUBROUTINE internal_data_deallocate

! **************************************************************************************************
!> \brief Clears pointers from the data area.
!> \param[inout] area         data area
! **************************************************************************************************
   SUBROUTINE dbcsr_data_clear_pointer(area)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_clear_pointer', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      IF (.NOT. ASSOCIATED(area%d)) THEN
         RETURN
      ENDIF

      IF (area%d%refcount .LE. 0) &
         CALL cp__w("dbcsr/data/dbcsr_data_methods_low.F",294,"Data seems to be unreferenced.")
      SELECT CASE (area%d%data_type)
      CASE (dbcsr_type_int_4)
         NULLIFY (area%d%i4)
      CASE (dbcsr_type_int_8)
         NULLIFY (area%d%i8)
      CASE (dbcsr_type_real_4)
         NULLIFY (area%d%r_sp)
      CASE (dbcsr_type_real_8)
         NULLIFY (area%d%r_dp)
      CASE (dbcsr_type_complex_4)
         NULLIFY (area%d%c_sp)
      CASE (dbcsr_type_complex_8)
         NULLIFY (area%d%c_dp)
      CASE (dbcsr_type_real_8_2d)
         NULLIFY (area%d%r2_dp)
      CASE (dbcsr_type_real_4_2d)
         NULLIFY (area%d%r2_sp)
      CASE (dbcsr_type_complex_8_2d)
         NULLIFY (area%d%c2_dp)
      CASE (dbcsr_type_complex_4_2d)
         NULLIFY (area%d%c2_sp)
      CASE default
         CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",317,"Invalid data type.")
      END SELECT
   END SUBROUTINE dbcsr_data_clear_pointer

! **************************************************************************************************
!> \brief Checks whether a data area is valid
!> \param[in] area         data area
!> \retval valid           whether the data area is valid
! **************************************************************************************************
   ELEMENTAL FUNCTION dbcsr_data_valid(area) RESULT(valid)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      LOGICAL                                            :: valid

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_valid', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      valid = ASSOCIATED(area%d)
   END FUNCTION dbcsr_data_valid

! **************************************************************************************************
!> \brief Checks whether a data pointer exists
!> \param[in] area         data area
!> \retval valid           whether the data pointer exists
! **************************************************************************************************
   FUNCTION dbcsr_data_exists(area) RESULT(valid)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      LOGICAL                                            :: valid

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_exists', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!   ---------------------------------------------------------------------------

      IF (careful_mod) THEN
         CALL timeset(routineN, error_handle)
      ENDIF
      !
      valid = dbcsr_data_valid(area)
      IF (.NOT. valid) &
         CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",360,"Data area is invalid.")

      SELECT CASE (area%d%data_type)
      CASE (dbcsr_type_int_4)
         valid = ASSOCIATED(area%d%i4)
      CASE (dbcsr_type_int_8)
         valid = ASSOCIATED(area%d%i8)
      CASE (dbcsr_type_real_4)
         valid = ASSOCIATED(area%d%r_sp)
      CASE (dbcsr_type_real_8)
         valid = ASSOCIATED(area%d%r_dp)
      CASE (dbcsr_type_complex_4)
         valid = ASSOCIATED(area%d%c_sp)
      CASE (dbcsr_type_complex_8)
         valid = ASSOCIATED(area%d%c_dp)
      CASE (dbcsr_type_real_4_2d)
         valid = ASSOCIATED(area%d%r2_sp)
      CASE (dbcsr_type_real_8_2d)
         valid = ASSOCIATED(area%d%r2_dp)
      CASE (dbcsr_type_complex_4_2d)
         valid = ASSOCIATED(area%d%c2_sp)
      CASE (dbcsr_type_complex_8_2d)
         valid = ASSOCIATED(area%d%c2_dp)
      CASE default
         CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",384,"Invalid data type.")
      END SELECT

      IF (careful_mod) THEN
         CALL timestop(error_handle)
      ENDIF
   END FUNCTION dbcsr_data_exists

! **************************************************************************************************
!> \brief Registers another use of the data area
!> \param[inout] area         data area
! **************************************************************************************************
   SUBROUTINE dbcsr_data_hold(area)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_hold', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      IF (careful_mod) THEN
         IF (.NOT. ASSOCIATED(area%d)) &
            CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",406,"Can't hold an empty data area.")
         CALL dbcsr_assert(area%d%refcount, "GT", 0, &
                           dbcsr_caller_error, dbcsr_fatal_level, &
                           routineN, "Should not hold an area with zero references.", &
                           410)
      ENDIF
      IF (.NOT. ASSOCIATED(area%d)) THEN
         RETURN
      ENDIF
!$OMP ATOMIC
      area%d%refcount = area%d%refcount+1
   END SUBROUTINE dbcsr_data_hold

! **************************************************************************************************
!> \brief Points data area data pointers to another data area
!>
!> Assumes that no memory will be lost when repointing the pointer in the data
!> area and that the area is initialized.
!> \param[in,out] area         data area to repoint
!> \param[in] rsize size of data area to point to
!> \param[in] csize size of data area to point to
!> \param[in] pointee          data area to point to
!> \param[in] source_lb        (optional) point to this offset in pointee
! **************************************************************************************************
   SUBROUTINE set_data_area_area(area, rsize, csize, pointee, source_lb)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area
      INTEGER, INTENT(IN)                                :: rsize, csize
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: pointee
      INTEGER, INTENT(IN), OPTIONAL                      :: source_lb

      CHARACTER(len=*), PARAMETER :: routineN = 'set_data_area_area', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=real_4), DIMENSION(:), POINTER        :: c_sp
      COMPLEX(KIND=real_8), DIMENSION(:), POINTER        :: c_dp
      INTEGER                                            :: bp, dt1, dt2, nze
      LOGICAL                                            :: compatible, pointee_is_2d, rmp
      REAL(KIND=real_4), DIMENSION(:), POINTER           :: r_sp
      REAL(KIND=real_8), DIMENSION(:), POINTER           :: r_dp

!   ---------------------------------------------------------------------------

      bp = 1; IF (PRESENT(source_lb)) bp = source_lb
      nze = rsize*csize
      dt1 = area%d%data_type
      dt2 = pointee%d%data_type
      IF (careful_mod) THEN
         compatible = dt1 .EQ. dt2 .OR. dt1 .EQ. dbcsr_type_1d_to_2d(dt2)
         IF (.NOT. compatible) &
            CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",455,"Can not point 1-d pointer to 2-d data")

      ENDIF
      pointee_is_2d = dbcsr_type_is_2d(dt2)
      IF (careful_mod) THEN
         IF (PRESENT(source_lb) .AND. pointee_is_2d) &
            CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",461,"Lower bound specification not possible with 2-d data")
         ! Check if size is OK.
         CALL dbcsr_assert(bp, "GE", 1, &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Attempt to point out of bounds", 465)
         CALL dbcsr_assert(bp+nze-1, "LE", dbcsr_data_get_size(pointee), &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Attempt to point out of bounds", 468)
      ENDIF
      ! There's a remap if the ranks are compatible but not equal.
      rmp = dt1 .NE. dt2
      SELECT CASE (dt2)
      CASE (dbcsr_type_real_4_2d)
         area%d%r2_sp => pointee%d%r2_sp(1:rsize, 1:csize)
      CASE (dbcsr_type_real_4)
         IF (rmp) THEN
            r_sp => dbcsr_get_data_p_s(pointee, bp, bp+nze-1)
            CALL pointer_rank_remap2(area%d%r2_sp, rsize, csize, &
                                     r_sp)
         ELSE
            area%d%r_sp => dbcsr_get_data_p_s(pointee, bp, bp+nze-1)
         ENDIF
      CASE (dbcsr_type_real_8_2d)
         area%d%r2_dp => pointee%d%r2_dp(1:rsize, 1:csize)
      CASE (dbcsr_type_real_8)
         IF (rmp) THEN
            r_dp => dbcsr_get_data_p_d(pointee, bp, bp+nze-1)
            CALL pointer_rank_remap2(area%d%r2_dp, rsize, csize, &
                                     r_dp)
         ELSE
            area%d%r_dp => dbcsr_get_data_p_d(pointee, bp, bp+nze-1)
         ENDIF
      CASE (dbcsr_type_complex_4_2d)
         area%d%c2_sp => pointee%d%c2_sp(1:rsize, 1:csize)
      CASE (dbcsr_type_complex_4)
         IF (rmp) THEN
            c_sp => dbcsr_get_data_p_c(pointee, bp, bp+nze-1)
            CALL pointer_rank_remap2(area%d%c2_sp, rsize, csize, &
                                     c_sp)
         ELSE
            area%d%c_sp => dbcsr_get_data_p_c(pointee, bp, bp+nze-1)
         ENDIF
      CASE (dbcsr_type_complex_8_2d)
         area%d%c2_dp => pointee%d%c2_dp(1:rsize, 1:csize)
      CASE (dbcsr_type_complex_8)
         IF (rmp) THEN
            c_dp => dbcsr_get_data_p_z(pointee, bp, bp+nze-1)
            CALL pointer_rank_remap2(area%d%c2_dp, rsize, csize, &
                                     c_dp)
         ELSE
            area%d%c_dp => dbcsr_get_data_p_z(pointee, bp, bp+nze-1)
         ENDIF
      CASE default
         CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",514,"Invalid data type")
      END SELECT
      CALL dbcsr_data_set_size_referenced(area, rsize*csize)
      IF (debug_mod) THEN
         CALL dbcsr_assert(dbcsr_data_get_size_referenced(area), "EQ", &
                           dbcsr_data_get_size(area), dbcsr_fatal_level, dbcsr_internal_error, &
                           routineN, "Size mismatch", 520)
      ENDIF
   END SUBROUTINE set_data_area_area

! **************************************************************************************************
!> \brief Returns the allocated data size
!> \param[in] area       data area
!> \retval data_size      size of data
! **************************************************************************************************
   FUNCTION dbcsr_data_get_size(area) RESULT(data_size)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      INTEGER                                            :: data_size

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_get_size', &
         routineP = moduleN//':'//routineN

      data_size = 0
      IF (ASSOCIATED(area%d)) THEN
         SELECT CASE (area%d%data_type)
         CASE (dbcsr_type_int_4)
            IF (ASSOCIATED(area%d%i4)) &
               data_size = SIZE(area%d%i4)
         CASE (dbcsr_type_int_8)
            IF (ASSOCIATED(area%d%i8)) &
               data_size = SIZE(area%d%i8)
         CASE (dbcsr_type_real_8)
            IF (ASSOCIATED(area%d%r_dp)) &
               data_size = SIZE(area%d%r_dp)
         CASE (dbcsr_type_real_4)
            IF (ASSOCIATED(area%d%r_sp)) &
               data_size = SIZE(area%d%r_sp)
         CASE (dbcsr_type_complex_8)
            IF (ASSOCIATED(area%d%c_dp)) &
               data_size = SIZE(area%d%c_dp)
         CASE (dbcsr_type_complex_4)
            IF (ASSOCIATED(area%d%c_sp)) &
               data_size = SIZE(area%d%c_sp)
         CASE (dbcsr_type_real_8_2d)
            IF (ASSOCIATED(area%d%r2_dp)) &
               data_size = SIZE(area%d%r2_dp)
         CASE (dbcsr_type_real_4_2d)
            IF (ASSOCIATED(area%d%r2_sp)) &
               data_size = SIZE(area%d%r2_sp)
         CASE (dbcsr_type_complex_8_2d)
            IF (ASSOCIATED(area%d%c2_dp)) &
               data_size = SIZE(area%d%c2_dp)
         CASE (dbcsr_type_complex_4_2d)
            IF (ASSOCIATED(area%d%c2_sp)) &
               data_size = SIZE(area%d%c2_sp)
         CASE default
            CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",570,"Incorrect data type")
         END SELECT
      ELSE
         CALL cp__w("dbcsr/data/dbcsr_data_methods_low.F",573,"Uninitialized data area")
         data_size = 0
      ENDIF
   END FUNCTION dbcsr_data_get_size

! **************************************************************************************************
!> \brief Verifies bounds of a data area
!> \param[in] area             Data area
!> \param[in] lb               lower bounds
!> \param[in] ub               upper bounds
! **************************************************************************************************
   SUBROUTINE dbcsr_data_verify_bounds(area, lb, ub)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      INTEGER, DIMENSION(:), INTENT(IN)                  :: lb, ub

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_verify_bounds', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_type, handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      data_type = dbcsr_data_get_type(area)
      IF (dbcsr_type_is_2d(data_type)) THEN
         CALL dbcsr_assert(SIZE(lb), "EQ", 2, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "size must be 2 for 2-d lb", 600)
         CALL dbcsr_assert(SIZE(ub), "EQ", 2, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "size must be 2 for 2-d ub", 603)
      ELSE
         CALL dbcsr_assert(SIZE(lb), "EQ", 1, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "size must be 1 for 1-d lb", 607)
         CALL dbcsr_assert(SIZE(ub), "EQ", 1, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "size must be 1 for 1-d ub", 610)
      ENDIF
      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         CALL dbcsr_assert(lb(1), "GE", LBOUND(area%d%r_sp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb r_sp", 616)
         CALL dbcsr_assert(ub(1), "LE", UBOUND(area%d%r_sp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub r_sp", 619)
      CASE (dbcsr_type_real_4_2d)
         CALL dbcsr_assert(lb(1), "GE", LBOUND(area%d%r2_sp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb r_sp 2d", 623)
         CALL dbcsr_assert(ub(1), "LE", UBOUND(area%d%r2_sp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub r_sp 2d", 626)
         CALL dbcsr_assert(lb(2), "GE", LBOUND(area%d%r2_sp, 2), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb r_sp 2d", 629)
         CALL dbcsr_assert(ub(2), "LE", UBOUND(area%d%r2_sp, 2), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub r_sp 2d", 632)
      CASE (dbcsr_type_real_8)
         CALL dbcsr_assert(lb(1), "GE", LBOUND(area%d%r_dp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb r_dp", 636)
         CALL dbcsr_assert(ub(1), "LE", UBOUND(area%d%r_dp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub r_dp", 639)
      CASE (dbcsr_type_real_8_2d)
         CALL dbcsr_assert(lb(1), "GE", LBOUND(area%d%r2_dp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb r_dp 2d", 643)
         CALL dbcsr_assert(ub(1), "LE", UBOUND(area%d%r2_dp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub r_dp 2d", 646)
         CALL dbcsr_assert(lb(2), "GE", LBOUND(area%d%r2_dp, 2), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb r_dp 2d", 649)
         CALL dbcsr_assert(ub(2), "LE", UBOUND(area%d%r2_dp, 2), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub r_dp 2d", 652)
      CASE (dbcsr_type_complex_4)
         CALL dbcsr_assert(lb(1), "GE", LBOUND(area%d%c_sp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb c_sp", 656)
         CALL dbcsr_assert(ub(1), "LE", UBOUND(area%d%c_sp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub c_sp", 659)
      CASE (dbcsr_type_complex_4_2d)
         CALL dbcsr_assert(lb(1), "GE", LBOUND(area%d%c2_sp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb c_sp 2d", 663)
         CALL dbcsr_assert(ub(1), "LE", UBOUND(area%d%c2_sp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub c_sp 2d", 666)
         CALL dbcsr_assert(lb(2), "GE", LBOUND(area%d%c2_sp, 2), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb c_sp 2d", 669)
         CALL dbcsr_assert(ub(2), "LE", UBOUND(area%d%c2_sp, 2), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub c_sp 2d", 672)
      CASE (dbcsr_type_complex_8)
         CALL dbcsr_assert(lb(1), "GE", LBOUND(area%d%c_dp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb c_dp", 676)
         CALL dbcsr_assert(ub(1), "LE", UBOUND(area%d%c_dp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub c_dp", 679)
      CASE (dbcsr_type_complex_8_2d)
         CALL dbcsr_assert(lb(1), "GE", LBOUND(area%d%c2_dp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb c_dp 2d", 683)
         CALL dbcsr_assert(ub(1), "LE", UBOUND(area%d%c2_dp, 1), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub c_dp 2d", 686)
         CALL dbcsr_assert(lb(2), "GE", LBOUND(area%d%c2_dp, 2), &
                           dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "lb c_dp 2d", 689)
         CALL dbcsr_assert(ub(2), "LE", UBOUND(area%d%c2_dp, 2), &
                           dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "ub c_dp 2d", 692)
      CASE default
         CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",694,"Invalid data type")
      END SELECT
      CALL timestop(handle)
   END SUBROUTINE dbcsr_data_verify_bounds

! **************************************************************************************************
!> \brief Clears a part of the data area
!> \param[in] area         data area
!> \param lb ...
!> \param ub ...
!> \note Optimized for clearing big 1-D data areas from all data types.
! **************************************************************************************************
   SUBROUTINE dbcsr_data_zero(area, lb, ub)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area
      INTEGER, DIMENSION(:), INTENT(in)                  :: lb, ub

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_zero', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!   ---------------------------------------------------------------------------

      IF (careful_mod) THEN
         CALL timeset(routineN, error_handle)
      ENDIF

      SELECT CASE (area%d%data_type)
      CASE (dbcsr_type_real_4)
         CALL memory_zero(area%d%r_sp(lb(1):), ub(1))
      CASE (dbcsr_type_real_8)
         CALL memory_zero(area%d%r_dp(lb(1):), ub(1))
      CASE (dbcsr_type_complex_4)
         CALL memory_zero(area%d%c_sp(lb(1):), ub(1))
      CASE (dbcsr_type_complex_8)
         CALL memory_zero(area%d%c_dp(lb(1):), ub(1))
      CASE (dbcsr_type_real_4_2d)
         area%d%r2_sp(lb(1):ub(1), lb(2):ub(2)) = 0.0_real_4
      CASE (dbcsr_type_real_8_2d)
         area%d%r2_dp(lb(1):ub(1), lb(2):ub(2)) = 0.0_real_8
      CASE (dbcsr_type_complex_4_2d)
         area%d%c2_sp(lb(1):ub(1), lb(2):ub(2)) = 0.0_real_4
      CASE (dbcsr_type_complex_8_2d)
         area%d%c2_dp(lb(1):ub(1), lb(2):ub(2)) = 0.0_real_8
      CASE default
         CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",739,"Invalid data type.")
      END SELECT

      IF (area%d%memory_type%acc_devalloc) &
         CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",743,"not yet supported for acc devmem")

      IF (careful_mod) THEN
         CALL timestop(error_handle)
      ENDIF
   END SUBROUTINE dbcsr_data_zero

! **************************************************************************************************
!> \brief Returns the allocated data size
!> \param[in] area       data area to query for size
!> \param[out] sizes     array with the data sizes
!> \param[out] valid     whether the data is actually allocated
! **************************************************************************************************
   SUBROUTINE dbcsr_data_get_sizes_any(area, sizes, valid)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: sizes
      LOGICAL, INTENT(OUT)                               :: valid

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_get_sizes_any', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

!   ---------------------------------------------------------------------------

      IF (careful_mod) &
         CALL timeset(routineN, handle)

      valid = .FALSE.
      sizes(:) = 0
      IF (ASSOCIATED(area%d)) THEN
         IF (careful_mod) THEN
            IF (dbcsr_type_is_2d(area%d%data_type)) THEN
               CALL dbcsr_assert(SIZE(sizes), "EQ", 2, dbcsr_fatal_level, &
                                 dbcsr_wrong_args_error, routineN, &
                                 "Sizes must have 2 elements for 2-D data", 778)
            ELSE
               CALL dbcsr_assert(SIZE(sizes), "EQ", 1, dbcsr_fatal_level, &
                                 dbcsr_wrong_args_error, routineN, &
                                 "Sizes must have 1 elements for 1-D data", 782)
            ENDIF
         ENDIF
         valid = dbcsr_data_exists(area)
         IF (valid) THEN
            SELECT CASE (area%d%data_type)
            CASE (dbcsr_type_real_8)
               sizes(1) = SIZE(area%d%r_dp)
            CASE (dbcsr_type_real_4)
               sizes(1) = SIZE(area%d%r_sp)
            CASE (dbcsr_type_complex_8)
               sizes(1) = SIZE(area%d%c_dp)
            CASE (dbcsr_type_complex_4)
               sizes(1) = SIZE(area%d%c_sp)
            CASE (dbcsr_type_real_8_2d)
               sizes(1) = SIZE(area%d%r2_dp, 1)
               sizes(2) = SIZE(area%d%r2_dp, 2)
            CASE (dbcsr_type_real_4_2d)
               sizes(1) = SIZE(area%d%r2_sp, 1)
               sizes(2) = SIZE(area%d%r2_sp, 2)
            CASE (dbcsr_type_complex_8_2d)
               sizes(1) = SIZE(area%d%c2_dp, 1)
               sizes(2) = SIZE(area%d%c2_dp, 2)
            CASE (dbcsr_type_complex_4_2d)
               sizes(1) = SIZE(area%d%c2_sp, 1)
               sizes(2) = SIZE(area%d%c2_sp, 2)
            CASE default
               CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",809,"Incorrect data type")
            END SELECT
         ENDIF
      ENDIF
      IF (careful_mod) &
         CALL timestop(handle)
   END SUBROUTINE dbcsr_data_get_sizes_any

! **************************************************************************************************
!> \brief Returns the allocated data size
!> \param[in] area       data area to query for size, should be 2-D
!> \param[out] row_size  row size
!> \param[out] col_size  column size
!> \param[out] valid     whether the data is actually allocated
! **************************************************************************************************
   SUBROUTINE dbcsr_data_get_sizes_2(area, row_size, col_size, valid)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      INTEGER, INTENT(OUT)                               :: row_size, col_size
      LOGICAL, INTENT(OUT)                               :: valid

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_get_sizes_2', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      INTEGER, DIMENSION(2)                              :: s

!   ---------------------------------------------------------------------------

      IF (careful_mod) &
         CALL timeset(routineN, handle)
      IF (ASSOCIATED(area%d)) THEN
         IF (careful_mod .AND. .NOT. dbcsr_type_is_2d(area%d%data_type)) &
            CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",841,"1-D data can not have column size")
         CALL dbcsr_data_get_sizes_any(area, s, valid)
         row_size = s(1)
         col_size = s(2)
      ELSE
         valid = .FALSE.
         row_size = 0
         col_size = 0
      ENDIF
      IF (careful_mod) &
         CALL timestop(handle)
   END SUBROUTINE dbcsr_data_get_sizes_2

! **************************************************************************************************
!> \brief Returns the allocated data size
!> \param[in] area         data area to query for size
!> \param[out] total_size  size of array
!> \param[out] valid       whether the data is actually allocated
! **************************************************************************************************
   SUBROUTINE dbcsr_data_get_sizes_1(area, total_size, valid)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      INTEGER, INTENT(OUT)                               :: total_size
      LOGICAL, INTENT(OUT)                               :: valid

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_get_sizes_1', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      INTEGER, DIMENSION(1)                              :: s

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(area%d)) THEN
         IF (careful_mod) &
            CALL dbcsr_assert("NOT", dbcsr_type_is_2d(area%d%data_type), &
                              dbcsr_fatal_level, &
                              dbcsr_wrong_args_error, routineN, &
                              "Should not use 2-D data", 880)
         CALL dbcsr_data_get_sizes_any(area, s, valid)
         total_size = s(1)
      ELSE
         valid = .FALSE.
         total_size = 0
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_data_get_sizes_1

! **************************************************************************************************
!> \brief Returns an encapsulated scalar "1"
!> \param[in] data_type       use the data type
!> \retval one                encapsulated value of one
! **************************************************************************************************
   ELEMENTAL FUNCTION dbcsr_scalar_one(data_type) RESULT(one)
      INTEGER, INTENT(IN)                                :: data_type
      TYPE(dbcsr_scalar_type)                            :: one

      one = dbcsr_scalar_zero(data_type)
      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         one%r_sp = 1.0_real_4
      CASE (dbcsr_type_real_8)
         one%r_dp = 1.0_real_8
      CASE (dbcsr_type_complex_4)
         one%c_sp = CMPLX(1.0, 0.0, real_4)
      CASE (dbcsr_type_complex_8)
         one%c_dp = CMPLX(1.0, 0.0, real_8)
      END SELECT
   END FUNCTION dbcsr_scalar_one

! **************************************************************************************************
!> \brief Returns an encapsulated scalar "sqrt(-1)"
!> \param[in] data_type       use the data type
!> \retval i                 encapsulated value of i
! **************************************************************************************************
   ELEMENTAL FUNCTION dbcsr_scalar_i(data_type) RESULT(i)
      INTEGER, INTENT(IN)                                :: data_type
      TYPE(dbcsr_scalar_type)                            :: i

      i = dbcsr_scalar_zero(data_type)
      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         i%r_sp = 0.0_real_4
      CASE (dbcsr_type_real_8)
         i%r_dp = 0.0_real_8
      CASE (dbcsr_type_complex_4)
         i%c_sp = CMPLX(0.0, 1.0, real_4)
      CASE (dbcsr_type_complex_8)
         i%c_dp = CMPLX(0.0, 1.0, real_8)
      END SELECT
   END FUNCTION dbcsr_scalar_i

! **************************************************************************************************
!> \brief Returns an encapsulated scalar "0"
!> \param[in] data_type       use the data type
!> \retval zero               encapsulated value of zero
! **************************************************************************************************
   ELEMENTAL FUNCTION dbcsr_scalar_zero(data_type) RESULT(zero)
      INTEGER, INTENT(IN)                                :: data_type
      TYPE(dbcsr_scalar_type)                            :: zero

      zero%data_type = data_type
      zero%r_sp = 0.0_real_4
      zero%r_dp = 0.0_real_8
      zero%c_sp = CMPLX(0.0, 0.0, real_4)
      zero%c_dp = CMPLX(0.0, 0.0, real_8)
   END FUNCTION dbcsr_scalar_zero

! **************************************************************************************************
!> \brief Returns whether an encapsulated scalar is equal to another value
!> \param[in] s1                    one value
!> \param[in] s2                    another value
!> \retval are_equal                whether values are equal
! **************************************************************************************************
   ELEMENTAL FUNCTION dbcsr_scalar_are_equal(s1, s2) RESULT(are_equal)
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: s1, s2
      LOGICAL                                            :: are_equal

      IF (s1%data_type .NE. s2%data_type) THEN
         are_equal = .FALSE.
      ELSE
         SELECT CASE (s1%data_type)
         CASE (dbcsr_type_real_4)
            are_equal = s1%r_sp .EQ. s2%r_sp
         CASE (dbcsr_type_real_8)
            are_equal = s1%r_dp .EQ. s2%r_dp
         CASE (dbcsr_type_complex_4)
            are_equal = s1%c_sp .EQ. s2%c_sp
         CASE (dbcsr_type_complex_8)
            are_equal = s1%c_dp .EQ. s2%c_dp
         CASE default
            are_equal = .FALSE.
         END SELECT
      ENDIF
   END FUNCTION dbcsr_scalar_are_equal

! **************************************************************************************************
!> \brief Returns an encapsulated scalar as a negation of the given
!> \param[in] s                    given value
!> \retval negated                 negated value
! **************************************************************************************************
   ELEMENTAL FUNCTION dbcsr_scalar_negative(s) RESULT(negated)
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: s
      TYPE(dbcsr_scalar_type)                            :: negated

      negated = dbcsr_scalar_zero(s%data_type)
      SELECT CASE (s%data_type)
      CASE (dbcsr_type_real_4)
         negated%r_sp = -s%r_sp
      CASE (dbcsr_type_real_8)
         negated%r_dp = -s%r_dp
      CASE (dbcsr_type_complex_4)
         negated%c_sp = -s%c_sp
      CASE (dbcsr_type_complex_8)
         negated%c_dp = -s%c_dp
      CASE default
         negated = dbcsr_scalar_zero(s%data_type)
      END SELECT
   END FUNCTION dbcsr_scalar_negative

! **************************************************************************************************
!> \brief Returns an encapsulated scalar as a negation of the given
!> \param s1 ...
!> \param s2 ...
!> \retval s_sum ...
! **************************************************************************************************
   ELEMENTAL FUNCTION dbcsr_scalar_add(s1, s2) RESULT(s_sum)
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: s1, s2
      TYPE(dbcsr_scalar_type)                            :: s_sum

      s_sum = dbcsr_scalar_zero(s1%data_type)
      SELECT CASE (s1%data_type)
      CASE (dbcsr_type_real_4)
         s_sum%r_sp = s1%r_sp+s2%r_sp
      CASE (dbcsr_type_real_8)
         s_sum%r_dp = s1%r_dp+s2%r_dp
      CASE (dbcsr_type_complex_4)
         s_sum%c_sp = s1%c_sp+s2%c_sp
      CASE (dbcsr_type_complex_8)
         s_sum%c_dp = s1%c_dp+s2%c_dp
      CASE default
         s_sum = dbcsr_scalar_zero(s1%data_type)
      END SELECT
   END FUNCTION dbcsr_scalar_add

! **************************************************************************************************
!> \brief ...
!> \param s1 ...
!> \param s2 ...
!> \retval s_product ...
! **************************************************************************************************
   ELEMENTAL FUNCTION dbcsr_scalar_multiply(s1, s2) RESULT(s_product)
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: s1, s2
      TYPE(dbcsr_scalar_type)                            :: s_product

      s_product = dbcsr_scalar_zero(s1%data_type)
      SELECT CASE (s1%data_type)
      CASE (dbcsr_type_real_4)
         s_product%r_sp = s1%r_sp*s2%r_sp
      CASE (dbcsr_type_real_8)
         s_product%r_dp = s1%r_dp*s2%r_dp
      CASE (dbcsr_type_complex_4)
         s_product%c_sp = s1%c_sp*s2%c_sp
      CASE (dbcsr_type_complex_8)
         s_product%c_dp = s1%c_dp*s2%c_dp
      CASE default
         s_product = dbcsr_scalar_zero(s1%data_type)
      END SELECT
   END FUNCTION dbcsr_scalar_multiply

! **************************************************************************************************
!> \brief Returns data type of a scalar
!> \param[in] scalar       scalar
!> \retval data_type       data type of the scalar
! **************************************************************************************************
   ELEMENTAL FUNCTION dbcsr_scalar_get_type(scalar) RESULT(data_type)
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: scalar
      INTEGER                                            :: data_type

      data_type = scalar%data_type
   END FUNCTION dbcsr_scalar_get_type

! **************************************************************************************************
!> \brief Sets data type of a scalar
!> \param[in] scalar       scalar
!> \param data_type ...
! **************************************************************************************************
   ELEMENTAL SUBROUTINE dbcsr_scalar_set_type(scalar, data_type)
      TYPE(dbcsr_scalar_type), INTENT(INOUT)             :: scalar
      INTEGER, INTENT(IN)                                :: data_type

      scalar%data_type = data_type
   END SUBROUTINE dbcsr_scalar_set_type

! **************************************************************************************************
!> \brief Fills all data and precision types from the set one
!> \param[in,out] scalar         data area
! **************************************************************************************************
   ELEMENTAL SUBROUTINE dbcsr_scalar_fill_all(scalar)
      TYPE(dbcsr_scalar_type), INTENT(INOUT)             :: scalar

      SELECT CASE (scalar%data_type)
      CASE (dbcsr_type_real_4)
         !scalar%r_sp = 0
         scalar%r_dp = REAL(scalar%r_sp, KIND=real_8)
         scalar%c_sp = CMPLX(scalar%r_sp, 0, KIND=real_4)
         scalar%c_dp = CMPLX(scalar%r_sp, 0, KIND=real_8)
      CASE (dbcsr_type_real_8)
         scalar%r_sp = REAL(scalar%r_dp, KIND=real_4)
         !scalar%r_dp = REAL(scalar%r_dp, KIND=real_8)
         scalar%c_sp = CMPLX(scalar%r_dp, 0, KIND=real_4)
         scalar%c_dp = CMPLX(scalar%r_dp, 0, KIND=real_8)
      CASE (dbcsr_type_complex_4)
         scalar%r_sp = REAL(scalar%c_sp, KIND=real_4)
         scalar%r_dp = REAL(scalar%c_sp, KIND=real_8)
         !scalar%c_sp = CMPLX(scalar%c_sp, KIND=real_4)
         scalar%c_dp = CMPLX(scalar%c_sp, KIND=real_8)
      CASE (dbcsr_type_complex_8)
         scalar%r_sp = REAL(scalar%c_dp, KIND=real_4)
         scalar%r_dp = REAL(scalar%c_dp, KIND=real_8)
         scalar%c_sp = CMPLX(scalar%c_dp, KIND=real_4)
         !scalar%c_dp = CMPLX(scalar%r_dp, KIND=real_8)
      END SELECT
   END SUBROUTINE dbcsr_scalar_fill_all

! **************************************************************************************************
!> \brief Checks whether the data type is 2-D.
!> \param data_type ...
!> \retval dbcsr_type_is_2d Data type is 2-D.
! **************************************************************************************************
   PURE FUNCTION dbcsr_type_is_2d(data_type)
      INTEGER, INTENT(IN)                                :: data_type
      LOGICAL                                            :: dbcsr_type_is_2d

      dbcsr_type_is_2d = data_type .EQ. dbcsr_type_real_4_2d .OR. &
                         data_type .EQ. dbcsr_type_real_8_2d .OR. &
                         data_type .EQ. dbcsr_type_complex_4_2d .OR. &
                         data_type .EQ. dbcsr_type_complex_8_2d
   END FUNCTION dbcsr_type_is_2d

! **************************************************************************************************
!> \brief Returns 1-d data type corresponding to the given 2-D one.
!> \param data_type ...
!> \retval dbcsr_type_2d_to_1d 1-D data type
! **************************************************************************************************
   PURE FUNCTION dbcsr_type_2d_to_1d(data_type)
      INTEGER, INTENT(IN)                                :: data_type
      INTEGER                                            :: dbcsr_type_2d_to_1d

      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4_2d)
         dbcsr_type_2d_to_1d = dbcsr_type_real_4
      CASE (dbcsr_type_real_8_2d)
         dbcsr_type_2d_to_1d = dbcsr_type_real_8
      CASE (dbcsr_type_complex_4_2d)
         dbcsr_type_2d_to_1d = dbcsr_type_complex_4
      CASE (dbcsr_type_complex_8_2d)
         dbcsr_type_2d_to_1d = dbcsr_type_complex_8
      CASE (dbcsr_type_real_4)
         dbcsr_type_2d_to_1d = dbcsr_type_real_4
      CASE (dbcsr_type_real_8)
         dbcsr_type_2d_to_1d = dbcsr_type_real_8
      CASE (dbcsr_type_complex_4)
         dbcsr_type_2d_to_1d = dbcsr_type_complex_4
      CASE (dbcsr_type_complex_8)
         dbcsr_type_2d_to_1d = dbcsr_type_complex_8
      CASE default
         dbcsr_type_2d_to_1d = -1
      END SELECT
   END FUNCTION dbcsr_type_2d_to_1d

! **************************************************************************************************
!> \brief Returns 2-D data type corresponding to the given 1-D one.
!> \param data_type ...
!> \retval dbcsr_type_1d_to_2d 2-D data type
! **************************************************************************************************
   PURE FUNCTION dbcsr_type_1d_to_2d(data_type)
      INTEGER, INTENT(IN)                                :: data_type
      INTEGER                                            :: dbcsr_type_1d_to_2d

      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4)
         dbcsr_type_1d_to_2d = dbcsr_type_real_4_2d
      CASE (dbcsr_type_real_8)
         dbcsr_type_1d_to_2d = dbcsr_type_real_8_2d
      CASE (dbcsr_type_complex_4)
         dbcsr_type_1d_to_2d = dbcsr_type_complex_4_2d
      CASE (dbcsr_type_complex_8)
         dbcsr_type_1d_to_2d = dbcsr_type_complex_8_2d
      CASE (dbcsr_type_real_4_2d)
         dbcsr_type_1d_to_2d = dbcsr_type_real_4_2d
      CASE (dbcsr_type_real_8_2d)
         dbcsr_type_1d_to_2d = dbcsr_type_real_8_2d
      CASE (dbcsr_type_complex_4_2d)
         dbcsr_type_1d_to_2d = dbcsr_type_complex_4_2d
      CASE (dbcsr_type_complex_8_2d)
         dbcsr_type_1d_to_2d = dbcsr_type_complex_8_2d
      CASE default
         dbcsr_type_1d_to_2d = -1
      END SELECT
   END FUNCTION dbcsr_type_1d_to_2d

! **************************************************************************************************
!> \brief Get actual data storage used for matrix
!> \param[in] area            Count data of this matrix
!> \retval data_size_referenced Data size used by matrix
! **************************************************************************************************
   PURE FUNCTION dbcsr_data_get_size_referenced(area) RESULT(data_size_referenced)
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: area
      INTEGER                                            :: data_size_referenced

      IF (ASSOCIATED(area%d)) THEN
         data_size_referenced = area%d%ref_size
      ELSE
         data_size_referenced = 0
      ENDIF
   END FUNCTION dbcsr_data_get_size_referenced

! **************************************************************************************************
!> \brief Sets the referenced size of the data area
!> \param[in,out] data_area  area for which to set referenced data size
!> \param[in] referenced_size     set referenced data size to this value
! **************************************************************************************************
   PURE SUBROUTINE dbcsr_data_set_size_referenced(data_area, referenced_size)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: data_area
      INTEGER, INTENT(IN)                                :: referenced_size

      data_area%d%ref_size = referenced_size
   END SUBROUTINE dbcsr_data_set_size_referenced

! **************************************************************************************************
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Sets a data pointer.
!> \param[inout] area     target data area
!> \param[in]    p        source data pointer
!> \par Assumptions
!>      Assumes that no memory will be lost when repointing the
!>      pointer in the data area and that the area is initialized.
! **************************************************************************************************
  SUBROUTINE set_data_p_d (area, p)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: area
    REAL(kind=real_8), DIMENSION(:), POINTER :: p

    CHARACTER(len=*), PARAMETER :: routineN = 'set_data_p_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    area%d%r_dp => p
  END SUBROUTINE set_data_p_d

! **************************************************************************************************
!> \brief Sets a data pointer.
!> \param[inout] area     target data area
!> \param[in]    p        source data pointer
!> \par Assumptions
!>      Assumes that no memory will be lost when repointing the
!>      pointer in the data area and that the area is initialized.
! **************************************************************************************************
  SUBROUTINE set_data_p_2d_d (area, p)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: area
    REAL(kind=real_8), DIMENSION(:,:), POINTER         :: p

    CHARACTER(len=*), PARAMETER :: routineN = 'set_data_p_2d_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    area%d%r2_dp => p
  END SUBROUTINE set_data_p_2d_d


! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[in] select_data_type   force datatype
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval data          pointer to data
!> \par Calling
!>      This routine is hidden behind the dbcsr_get_data interface, hence the
!>      need for the select_data_type argument.
!>      see dbcsr_get_data_p_d
! **************************************************************************************************
  FUNCTION dbcsr_get_data_c_d (area, select_data_type, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)         :: area
    REAL(kind=real_8), INTENT(IN)            :: select_data_type
    INTEGER, INTENT(IN), OPTIONAL  :: lb, ub
    REAL(kind=real_8), DIMENSION(:), POINTER :: DATA

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_c_d', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u

!   ---------------------------------------------------------------------------

    ! The select_data_type argument is needed to make this function unique
    ! enough to use in the interface.
    IF (KIND(select_data_type) .NE. KIND(DATA))&
       CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",75,"compiler borken")

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%r_dp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%r_dp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l .LT. LBOUND (area%d%r_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",85,"Out of bounds")
             IF(u .GT. UBOUND (area%d%r_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",87,"Out of bounds")
          ENDIF
          DATA => area%d%r_dp(l:u)
       ELSE
          DATA => area%d%r_dp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_c_d

! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \brief dbcsr_get_data_c_d
!> \param[in] area       data area
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval DATA pointer to data
!> \par Calling
!>      This routine can be called explicitly.
! **************************************************************************************************
  FUNCTION dbcsr_get_data_p_d (area, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)         :: area
    REAL(kind=real_8), DIMENSION(:), POINTER :: DATA
    INTEGER, INTENT(IN), OPTIONAL  :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_p_d', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%r_dp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%r_dp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l .LT. LBOUND (area%d%r_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",127,"Out of bounds")
             IF(u .GT. UBOUND (area%d%r_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",129,"Out of bounds")
          ENDIF
          DATA => area%d%r_dp(l:u)
       ELSE
          DATA => area%d%r_dp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_p_d

! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \brief dbcsr_get_data_c_d
!> \param[in] area       data area
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval DATA pointer to data
!> \par Calling
!>      This routine can be called explicitly.
! **************************************************************************************************
  FUNCTION dbcsr_get_data_p_2d_d (area, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)            :: area
    REAL(kind=real_8), DIMENSION(:,:), POINTER            :: DATA
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_p_2d_d', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(2)          :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%r2_dp)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%r2_dp)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l(1) .LT. LBOUND (area%d%r2_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",169,"Out of bounds")
             IF(l(2) .LT. LBOUND (area%d%r2_dp, 2))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",171,"Out of bounds")
             IF(u(1) .GT. UBOUND (area%d%r2_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",173,"Out of bounds")
             IF(u(2) .GT. UBOUND (area%d%r2_dp, 2))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",175,"Out of bounds")
          ENDIF
          DATA => area%d%r2_dp(l(1):u(1), l(2):u(2))
       ELSE
          DATA => area%d%r2_dp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_p_2d_d



! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[out] DATA pointer to data
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
! **************************************************************************************************
  SUBROUTINE get_data_d (area, DATA, lb, ub)
    TYPE(dbcsr_data_obj), INTENT(IN)  :: area
    REAL(kind=real_8), DIMENSION(:), POINTER    :: DATA
    INTEGER, INTENT(IN), OPTIONAL     :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'get_data_d', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%r_dp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%r_dp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             CALL dbcsr_assert (l, "GE", LBOUND (area%d%r_dp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",215)
             CALL dbcsr_assert (u, "LE", UBOUND (area%d%r_dp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",218)
          ENDIF
          DATA => area%d%r_dp(l:u)
       ELSE
          DATA => area%d%r_dp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END SUBROUTINE get_data_d


! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[out] DATA pointer to data
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
! **************************************************************************************************
  SUBROUTINE get_data_2d_d (area, DATA, lb, ub)
    TYPE(dbcsr_data_obj), INTENT(IN)            :: area
    REAL(kind=real_8), DIMENSION(:,:), POINTER            :: DATA
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'get_data_2d_d', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(2)          :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%r2_dp)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%r2_dp)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             CALL dbcsr_assert (l(1), "GE", LBOUND (area%d%r2_dp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",257)
             CALL dbcsr_assert (l(2), "GE", LBOUND (area%d%r2_dp, 2),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",260)
             CALL dbcsr_assert (u(1), "LE", UBOUND (area%d%r2_dp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",263)
             CALL dbcsr_assert (u(2), "LE", UBOUND (area%d%r2_dp, 2),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",266)
          ENDIF
          DATA => area%d%r2_dp(l(1):u(1), l(2):u(2))
       ELSE
          DATA => area%d%r2_dp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END SUBROUTINE get_data_2d_d

! **************************************************************************************************
!> \brief Sets a scalar in an encapsulated data structure
!> \param[in] scalar                    scalar to encapsulate
!> \retval encapsulated_scalar          encapsulated scalar 
! **************************************************************************************************
  ELEMENTAL FUNCTION dbcsr_scalar_d (scalar) RESULT (encapsulated_scalar)
    REAL(kind=real_8), INTENT(IN)       :: scalar
    TYPE(dbcsr_scalar_type)   :: encapsulated_scalar

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scalar_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    encapsulated_scalar = dbcsr_scalar_zero (dbcsr_type_real_8)
    encapsulated_scalar%r_dp = scalar
  END FUNCTION dbcsr_scalar_d

! **************************************************************************************************
!> \brief Sets a scalar in an encapsulated data structure
!> \param[in] encapsulated_scalar          encapsulated scalar
!> \param[out] value                       value of the scalar
! **************************************************************************************************
  ELEMENTAL SUBROUTINE dbcsr_scalar_get_value_d (encapsulated_scalar, value)
    TYPE(dbcsr_scalar_type), INTENT(IN) :: encapsulated_scalar
    REAL(kind=real_8), INTENT(OUT)                :: value

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scalar_get_value_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    value = encapsulated_scalar%r_dp
  END SUBROUTINE dbcsr_scalar_get_value_d
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Sets a data pointer.
!> \param[inout] area     target data area
!> \param[in]    p        source data pointer
!> \par Assumptions
!>      Assumes that no memory will be lost when repointing the
!>      pointer in the data area and that the area is initialized.
! **************************************************************************************************
  SUBROUTINE set_data_p_z (area, p)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: area
    COMPLEX(kind=real_8), DIMENSION(:), POINTER :: p

    CHARACTER(len=*), PARAMETER :: routineN = 'set_data_p_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    area%d%c_dp => p
  END SUBROUTINE set_data_p_z

! **************************************************************************************************
!> \brief Sets a data pointer.
!> \param[inout] area     target data area
!> \param[in]    p        source data pointer
!> \par Assumptions
!>      Assumes that no memory will be lost when repointing the
!>      pointer in the data area and that the area is initialized.
! **************************************************************************************************
  SUBROUTINE set_data_p_2d_z (area, p)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: area
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER         :: p

    CHARACTER(len=*), PARAMETER :: routineN = 'set_data_p_2d_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    area%d%c2_dp => p
  END SUBROUTINE set_data_p_2d_z


! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[in] select_data_type   force datatype
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval data          pointer to data
!> \par Calling
!>      This routine is hidden behind the dbcsr_get_data interface, hence the
!>      need for the select_data_type argument.
!>      see dbcsr_get_data_p_z
! **************************************************************************************************
  FUNCTION dbcsr_get_data_c_z (area, select_data_type, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)         :: area
    COMPLEX(kind=real_8), INTENT(IN)            :: select_data_type
    INTEGER, INTENT(IN), OPTIONAL  :: lb, ub
    COMPLEX(kind=real_8), DIMENSION(:), POINTER :: DATA

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_c_z', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u

!   ---------------------------------------------------------------------------

    ! The select_data_type argument is needed to make this function unique
    ! enough to use in the interface.
    IF (KIND(select_data_type) .NE. KIND(DATA))&
       CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",75,"compiler borken")

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%c_dp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%c_dp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l .LT. LBOUND (area%d%c_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",85,"Out of bounds")
             IF(u .GT. UBOUND (area%d%c_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",87,"Out of bounds")
          ENDIF
          DATA => area%d%c_dp(l:u)
       ELSE
          DATA => area%d%c_dp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_c_z

! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \brief dbcsr_get_data_c_z
!> \param[in] area       data area
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval DATA pointer to data
!> \par Calling
!>      This routine can be called explicitly.
! **************************************************************************************************
  FUNCTION dbcsr_get_data_p_z (area, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)         :: area
    COMPLEX(kind=real_8), DIMENSION(:), POINTER :: DATA
    INTEGER, INTENT(IN), OPTIONAL  :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_p_z', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%c_dp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%c_dp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l .LT. LBOUND (area%d%c_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",127,"Out of bounds")
             IF(u .GT. UBOUND (area%d%c_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",129,"Out of bounds")
          ENDIF
          DATA => area%d%c_dp(l:u)
       ELSE
          DATA => area%d%c_dp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_p_z

! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \brief dbcsr_get_data_c_z
!> \param[in] area       data area
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval DATA pointer to data
!> \par Calling
!>      This routine can be called explicitly.
! **************************************************************************************************
  FUNCTION dbcsr_get_data_p_2d_z (area, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)            :: area
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER            :: DATA
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_p_2d_z', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(2)          :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%c2_dp)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%c2_dp)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l(1) .LT. LBOUND (area%d%c2_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",169,"Out of bounds")
             IF(l(2) .LT. LBOUND (area%d%c2_dp, 2))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",171,"Out of bounds")
             IF(u(1) .GT. UBOUND (area%d%c2_dp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",173,"Out of bounds")
             IF(u(2) .GT. UBOUND (area%d%c2_dp, 2))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",175,"Out of bounds")
          ENDIF
          DATA => area%d%c2_dp(l(1):u(1), l(2):u(2))
       ELSE
          DATA => area%d%c2_dp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_p_2d_z



! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[out] DATA pointer to data
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
! **************************************************************************************************
  SUBROUTINE get_data_z (area, DATA, lb, ub)
    TYPE(dbcsr_data_obj), INTENT(IN)  :: area
    COMPLEX(kind=real_8), DIMENSION(:), POINTER    :: DATA
    INTEGER, INTENT(IN), OPTIONAL     :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'get_data_z', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%c_dp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%c_dp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             CALL dbcsr_assert (l, "GE", LBOUND (area%d%c_dp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",215)
             CALL dbcsr_assert (u, "LE", UBOUND (area%d%c_dp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",218)
          ENDIF
          DATA => area%d%c_dp(l:u)
       ELSE
          DATA => area%d%c_dp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END SUBROUTINE get_data_z


! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[out] DATA pointer to data
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
! **************************************************************************************************
  SUBROUTINE get_data_2d_z (area, DATA, lb, ub)
    TYPE(dbcsr_data_obj), INTENT(IN)            :: area
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER            :: DATA
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'get_data_2d_z', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(2)          :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%c2_dp)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%c2_dp)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             CALL dbcsr_assert (l(1), "GE", LBOUND (area%d%c2_dp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",257)
             CALL dbcsr_assert (l(2), "GE", LBOUND (area%d%c2_dp, 2),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",260)
             CALL dbcsr_assert (u(1), "LE", UBOUND (area%d%c2_dp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",263)
             CALL dbcsr_assert (u(2), "LE", UBOUND (area%d%c2_dp, 2),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",266)
          ENDIF
          DATA => area%d%c2_dp(l(1):u(1), l(2):u(2))
       ELSE
          DATA => area%d%c2_dp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END SUBROUTINE get_data_2d_z

! **************************************************************************************************
!> \brief Sets a scalar in an encapsulated data structure
!> \param[in] scalar                    scalar to encapsulate
!> \retval encapsulated_scalar          encapsulated scalar 
! **************************************************************************************************
  ELEMENTAL FUNCTION dbcsr_scalar_z (scalar) RESULT (encapsulated_scalar)
    COMPLEX(kind=real_8), INTENT(IN)       :: scalar
    TYPE(dbcsr_scalar_type)   :: encapsulated_scalar

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scalar_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    encapsulated_scalar = dbcsr_scalar_zero (dbcsr_type_complex_8)
    encapsulated_scalar%c_dp = scalar
  END FUNCTION dbcsr_scalar_z

! **************************************************************************************************
!> \brief Sets a scalar in an encapsulated data structure
!> \param[in] encapsulated_scalar          encapsulated scalar
!> \param[out] value                       value of the scalar
! **************************************************************************************************
  ELEMENTAL SUBROUTINE dbcsr_scalar_get_value_z (encapsulated_scalar, value)
    TYPE(dbcsr_scalar_type), INTENT(IN) :: encapsulated_scalar
    COMPLEX(kind=real_8), INTENT(OUT)                :: value

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scalar_get_value_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    value = encapsulated_scalar%c_dp
  END SUBROUTINE dbcsr_scalar_get_value_z
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Sets a data pointer.
!> \param[inout] area     target data area
!> \param[in]    p        source data pointer
!> \par Assumptions
!>      Assumes that no memory will be lost when repointing the
!>      pointer in the data area and that the area is initialized.
! **************************************************************************************************
  SUBROUTINE set_data_p_s (area, p)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: area
    REAL(kind=real_4), DIMENSION(:), POINTER :: p

    CHARACTER(len=*), PARAMETER :: routineN = 'set_data_p_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    area%d%r_sp => p
  END SUBROUTINE set_data_p_s

! **************************************************************************************************
!> \brief Sets a data pointer.
!> \param[inout] area     target data area
!> \param[in]    p        source data pointer
!> \par Assumptions
!>      Assumes that no memory will be lost when repointing the
!>      pointer in the data area and that the area is initialized.
! **************************************************************************************************
  SUBROUTINE set_data_p_2d_s (area, p)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: area
    REAL(kind=real_4), DIMENSION(:,:), POINTER         :: p

    CHARACTER(len=*), PARAMETER :: routineN = 'set_data_p_2d_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    area%d%r2_sp => p
  END SUBROUTINE set_data_p_2d_s


! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[in] select_data_type   force datatype
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval data          pointer to data
!> \par Calling
!>      This routine is hidden behind the dbcsr_get_data interface, hence the
!>      need for the select_data_type argument.
!>      see dbcsr_get_data_p_s
! **************************************************************************************************
  FUNCTION dbcsr_get_data_c_s (area, select_data_type, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)         :: area
    REAL(kind=real_4), INTENT(IN)            :: select_data_type
    INTEGER, INTENT(IN), OPTIONAL  :: lb, ub
    REAL(kind=real_4), DIMENSION(:), POINTER :: DATA

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_c_s', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u

!   ---------------------------------------------------------------------------

    ! The select_data_type argument is needed to make this function unique
    ! enough to use in the interface.
    IF (KIND(select_data_type) .NE. KIND(DATA))&
       CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",75,"compiler borken")

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%r_sp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%r_sp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l .LT. LBOUND (area%d%r_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",85,"Out of bounds")
             IF(u .GT. UBOUND (area%d%r_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",87,"Out of bounds")
          ENDIF
          DATA => area%d%r_sp(l:u)
       ELSE
          DATA => area%d%r_sp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_c_s

! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \brief dbcsr_get_data_c_s
!> \param[in] area       data area
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval DATA pointer to data
!> \par Calling
!>      This routine can be called explicitly.
! **************************************************************************************************
  FUNCTION dbcsr_get_data_p_s (area, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)         :: area
    REAL(kind=real_4), DIMENSION(:), POINTER :: DATA
    INTEGER, INTENT(IN), OPTIONAL  :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_p_s', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%r_sp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%r_sp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l .LT. LBOUND (area%d%r_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",127,"Out of bounds")
             IF(u .GT. UBOUND (area%d%r_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",129,"Out of bounds")
          ENDIF
          DATA => area%d%r_sp(l:u)
       ELSE
          DATA => area%d%r_sp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_p_s

! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \brief dbcsr_get_data_c_s
!> \param[in] area       data area
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval DATA pointer to data
!> \par Calling
!>      This routine can be called explicitly.
! **************************************************************************************************
  FUNCTION dbcsr_get_data_p_2d_s (area, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)            :: area
    REAL(kind=real_4), DIMENSION(:,:), POINTER            :: DATA
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_p_2d_s', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(2)          :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%r2_sp)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%r2_sp)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l(1) .LT. LBOUND (area%d%r2_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",169,"Out of bounds")
             IF(l(2) .LT. LBOUND (area%d%r2_sp, 2))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",171,"Out of bounds")
             IF(u(1) .GT. UBOUND (area%d%r2_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",173,"Out of bounds")
             IF(u(2) .GT. UBOUND (area%d%r2_sp, 2))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",175,"Out of bounds")
          ENDIF
          DATA => area%d%r2_sp(l(1):u(1), l(2):u(2))
       ELSE
          DATA => area%d%r2_sp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_p_2d_s



! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[out] DATA pointer to data
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
! **************************************************************************************************
  SUBROUTINE get_data_s (area, DATA, lb, ub)
    TYPE(dbcsr_data_obj), INTENT(IN)  :: area
    REAL(kind=real_4), DIMENSION(:), POINTER    :: DATA
    INTEGER, INTENT(IN), OPTIONAL     :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'get_data_s', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%r_sp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%r_sp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             CALL dbcsr_assert (l, "GE", LBOUND (area%d%r_sp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",215)
             CALL dbcsr_assert (u, "LE", UBOUND (area%d%r_sp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",218)
          ENDIF
          DATA => area%d%r_sp(l:u)
       ELSE
          DATA => area%d%r_sp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END SUBROUTINE get_data_s


! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[out] DATA pointer to data
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
! **************************************************************************************************
  SUBROUTINE get_data_2d_s (area, DATA, lb, ub)
    TYPE(dbcsr_data_obj), INTENT(IN)            :: area
    REAL(kind=real_4), DIMENSION(:,:), POINTER            :: DATA
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'get_data_2d_s', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(2)          :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%r2_sp)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%r2_sp)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             CALL dbcsr_assert (l(1), "GE", LBOUND (area%d%r2_sp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",257)
             CALL dbcsr_assert (l(2), "GE", LBOUND (area%d%r2_sp, 2),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",260)
             CALL dbcsr_assert (u(1), "LE", UBOUND (area%d%r2_sp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",263)
             CALL dbcsr_assert (u(2), "LE", UBOUND (area%d%r2_sp, 2),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",266)
          ENDIF
          DATA => area%d%r2_sp(l(1):u(1), l(2):u(2))
       ELSE
          DATA => area%d%r2_sp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END SUBROUTINE get_data_2d_s

! **************************************************************************************************
!> \brief Sets a scalar in an encapsulated data structure
!> \param[in] scalar                    scalar to encapsulate
!> \retval encapsulated_scalar          encapsulated scalar 
! **************************************************************************************************
  ELEMENTAL FUNCTION dbcsr_scalar_s (scalar) RESULT (encapsulated_scalar)
    REAL(kind=real_4), INTENT(IN)       :: scalar
    TYPE(dbcsr_scalar_type)   :: encapsulated_scalar

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scalar_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    encapsulated_scalar = dbcsr_scalar_zero (dbcsr_type_real_4)
    encapsulated_scalar%r_sp = scalar
  END FUNCTION dbcsr_scalar_s

! **************************************************************************************************
!> \brief Sets a scalar in an encapsulated data structure
!> \param[in] encapsulated_scalar          encapsulated scalar
!> \param[out] value                       value of the scalar
! **************************************************************************************************
  ELEMENTAL SUBROUTINE dbcsr_scalar_get_value_s (encapsulated_scalar, value)
    TYPE(dbcsr_scalar_type), INTENT(IN) :: encapsulated_scalar
    REAL(kind=real_4), INTENT(OUT)                :: value

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scalar_get_value_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    value = encapsulated_scalar%r_sp
  END SUBROUTINE dbcsr_scalar_get_value_s
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Sets a data pointer.
!> \param[inout] area     target data area
!> \param[in]    p        source data pointer
!> \par Assumptions
!>      Assumes that no memory will be lost when repointing the
!>      pointer in the data area and that the area is initialized.
! **************************************************************************************************
  SUBROUTINE set_data_p_c (area, p)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: area
    COMPLEX(kind=real_4), DIMENSION(:), POINTER :: p

    CHARACTER(len=*), PARAMETER :: routineN = 'set_data_p_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    area%d%c_sp => p
  END SUBROUTINE set_data_p_c

! **************************************************************************************************
!> \brief Sets a data pointer.
!> \param[inout] area     target data area
!> \param[in]    p        source data pointer
!> \par Assumptions
!>      Assumes that no memory will be lost when repointing the
!>      pointer in the data area and that the area is initialized.
! **************************************************************************************************
  SUBROUTINE set_data_p_2d_c (area, p)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: area
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER         :: p

    CHARACTER(len=*), PARAMETER :: routineN = 'set_data_p_2d_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    area%d%c2_sp => p
  END SUBROUTINE set_data_p_2d_c


! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[in] select_data_type   force datatype
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval data          pointer to data
!> \par Calling
!>      This routine is hidden behind the dbcsr_get_data interface, hence the
!>      need for the select_data_type argument.
!>      see dbcsr_get_data_p_c
! **************************************************************************************************
  FUNCTION dbcsr_get_data_c_c (area, select_data_type, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)         :: area
    COMPLEX(kind=real_4), INTENT(IN)            :: select_data_type
    INTEGER, INTENT(IN), OPTIONAL  :: lb, ub
    COMPLEX(kind=real_4), DIMENSION(:), POINTER :: DATA

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_c_c', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u

!   ---------------------------------------------------------------------------

    ! The select_data_type argument is needed to make this function unique
    ! enough to use in the interface.
    IF (KIND(select_data_type) .NE. KIND(DATA))&
       CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",75,"compiler borken")

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%c_sp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%c_sp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l .LT. LBOUND (area%d%c_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",85,"Out of bounds")
             IF(u .GT. UBOUND (area%d%c_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",87,"Out of bounds")
          ENDIF
          DATA => area%d%c_sp(l:u)
       ELSE
          DATA => area%d%c_sp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_c_c

! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \brief dbcsr_get_data_c_c
!> \param[in] area       data area
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval DATA pointer to data
!> \par Calling
!>      This routine can be called explicitly.
! **************************************************************************************************
  FUNCTION dbcsr_get_data_p_c (area, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)         :: area
    COMPLEX(kind=real_4), DIMENSION(:), POINTER :: DATA
    INTEGER, INTENT(IN), OPTIONAL  :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_p_c', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%c_sp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%c_sp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l .LT. LBOUND (area%d%c_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",127,"Out of bounds")
             IF(u .GT. UBOUND (area%d%c_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",129,"Out of bounds")
          ENDIF
          DATA => area%d%c_sp(l:u)
       ELSE
          DATA => area%d%c_sp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_p_c

! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \brief dbcsr_get_data_c_c
!> \param[in] area       data area
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
!> \retval DATA pointer to data
!> \par Calling
!>      This routine can be called explicitly.
! **************************************************************************************************
  FUNCTION dbcsr_get_data_p_2d_c (area, lb, ub) RESULT (DATA)
    TYPE(dbcsr_data_obj), INTENT(IN)            :: area
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER            :: DATA
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_p_2d_c', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(2)          :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%c2_sp)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%c2_sp)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             IF(l(1) .LT. LBOUND (area%d%c2_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",169,"Out of bounds")
             IF(l(2) .LT. LBOUND (area%d%c2_sp, 2))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",171,"Out of bounds")
             IF(u(1) .GT. UBOUND (area%d%c2_sp, 1))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",173,"Out of bounds")
             IF(u(2) .GT. UBOUND (area%d%c2_sp, 2))&
                CALL cp__b("dbcsr/data/dbcsr_data_methods_low.F",175,"Out of bounds")
          ENDIF
          DATA => area%d%c2_sp(l(1):u(1), l(2):u(2))
       ELSE
          DATA => area%d%c2_sp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END FUNCTION dbcsr_get_data_p_2d_c



! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[out] DATA pointer to data
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
! **************************************************************************************************
  SUBROUTINE get_data_c (area, DATA, lb, ub)
    TYPE(dbcsr_data_obj), INTENT(IN)  :: area
    COMPLEX(kind=real_4), DIMENSION(:), POINTER    :: DATA
    INTEGER, INTENT(IN), OPTIONAL     :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'get_data_c', &
      routineP = moduleN//':'//routineN

    INTEGER                        :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%c_sp, 1)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%c_sp, 1)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             CALL dbcsr_assert (l, "GE", LBOUND (area%d%c_sp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",215)
             CALL dbcsr_assert (u, "LE", UBOUND (area%d%c_sp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",218)
          ENDIF
          DATA => area%d%c_sp(l:u)
       ELSE
          DATA => area%d%c_sp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END SUBROUTINE get_data_c


! **************************************************************************************************
!> \brief Returns the single/double precision real/complex data
!> \param[in] area       data area
!> \param[out] DATA pointer to data
!> \param[in] lb         (optional) lower bound for pointer
!> \param[in] ub         (optional) upper bound for pointer
! **************************************************************************************************
  SUBROUTINE get_data_2d_c (area, DATA, lb, ub)
    TYPE(dbcsr_data_obj), INTENT(IN)            :: area
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER            :: DATA
    INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL :: lb, ub

    CHARACTER(len=*), PARAMETER :: routineN = 'get_data_2d_c', &
      routineP = moduleN//':'//routineN

    INTEGER, DIMENSION(2)          :: l, u
!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (area%d)) THEN
       IF (PRESENT (lb) .OR. PRESENT (ub)) THEN
          l = LBOUND (area%d%c2_sp)
          IF (PRESENT (lb)) l = lb
          u = UBOUND (area%d%c2_sp)
          IF (PRESENT (ub)) u = ub
          IF (debug_mod) THEN
             CALL dbcsr_assert (l(1), "GE", LBOUND (area%d%c2_sp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",257)
             CALL dbcsr_assert (l(2), "GE", LBOUND (area%d%c2_sp, 2),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",260)
             CALL dbcsr_assert (u(1), "LE", UBOUND (area%d%c2_sp, 1),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",263)
             CALL dbcsr_assert (u(2), "LE", UBOUND (area%d%c2_sp, 2),&
                  dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
                  "Out of bounds",266)
          ENDIF
          DATA => area%d%c2_sp(l(1):u(1), l(2):u(2))
       ELSE
          DATA => area%d%c2_sp
       ENDIF
    ELSE
       NULLIFY (DATA)
    ENDIF
  END SUBROUTINE get_data_2d_c

! **************************************************************************************************
!> \brief Sets a scalar in an encapsulated data structure
!> \param[in] scalar                    scalar to encapsulate
!> \retval encapsulated_scalar          encapsulated scalar 
! **************************************************************************************************
  ELEMENTAL FUNCTION dbcsr_scalar_c (scalar) RESULT (encapsulated_scalar)
    COMPLEX(kind=real_4), INTENT(IN)       :: scalar
    TYPE(dbcsr_scalar_type)   :: encapsulated_scalar

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scalar_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    encapsulated_scalar = dbcsr_scalar_zero (dbcsr_type_complex_4)
    encapsulated_scalar%c_sp = scalar
  END FUNCTION dbcsr_scalar_c

! **************************************************************************************************
!> \brief Sets a scalar in an encapsulated data structure
!> \param[in] encapsulated_scalar          encapsulated scalar
!> \param[out] value                       value of the scalar
! **************************************************************************************************
  ELEMENTAL SUBROUTINE dbcsr_scalar_get_value_c (encapsulated_scalar, value)
    TYPE(dbcsr_scalar_type), INTENT(IN) :: encapsulated_scalar
    COMPLEX(kind=real_4), INTENT(OUT)                :: value

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scalar_get_value_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    value = encapsulated_scalar%c_sp
  END SUBROUTINE dbcsr_scalar_get_value_c
END MODULE dbcsr_data_methods_low
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR Memory Pool to avoid slow allocations of accelerator memory
!> \author  Ole Schuett
! **************************************************************************************************
MODULE dbcsr_mem_methods

   USE acc_stream,                      ONLY: acc_stream_associated,&
                                              acc_stream_equal,&
                                              acc_stream_type
   USE dbcsr_data_methods_low,          ONLY: dbcsr_data_exists,&
                                              dbcsr_data_get_size,&
                                              internal_data_deallocate
   USE dbcsr_data_types,                ONLY: dbcsr_data_obj,&
                                              dbcsr_mempool_entry_type,&
                                              dbcsr_mempool_type,&
                                              dbcsr_memtype_type
   USE kinds,                           ONLY: dp

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads,&
!$                    omp_set_lock, omp_unset_lock, omp_init_lock, omp_lock_kind, omp_destroy_lock

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_mem_methods'

   PUBLIC :: dbcsr_mempool_get, dbcsr_mempool_add, dbcsr_mempool_ensure_capacity
   PUBLIC :: dbcsr_mempool_destruct, dbcsr_mempool_clear
   PUBLIC :: dbcsr_memtype_setup, dbcsr_memtype_equal

   LOGICAL, PARAMETER :: debug = .FALSE.

CONTAINS

! **************************************************************************************************
!> \brief Creates a memory pool.
!> \param pool ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mempool_create(pool)
      TYPE(dbcsr_mempool_type), POINTER                  :: pool

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_mempool_create', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(pool)) CALL cp__b("dbcsr/data/dbcsr_mem_methods.F",54,"pool already allocated")
      ALLOCATE (pool)
!$    CALL OMP_INIT_LOCK(pool%lock)
      ALLOCATE (pool%root) !root allways allocated, but unsused. Simplifies looping.
   END SUBROUTINE dbcsr_mempool_create

! **************************************************************************************************
!> \brief Ensures that mempool has at least the given capacity.
!> \param pool ...
!> \param capacity ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mempool_ensure_capacity(pool, capacity)
      TYPE(dbcsr_mempool_type), POINTER                  :: pool
      INTEGER, INTENT(IN)                                :: capacity

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_mempool_ensure_capacity', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(pool)) CALL cp__b("dbcsr/data/dbcsr_mem_methods.F",73,"pool not allocated")
!$    CALL OMP_SET_LOCK(pool%lock)
      pool%capacity = MAX(pool%capacity, capacity)
!$    CALL OMP_UNSET_LOCK(pool%lock)

   END SUBROUTINE dbcsr_mempool_ensure_capacity

! **************************************************************************************************
!> \brief Picks a suitable data_area from mempool, returns Null() if none found.
!> \param memtype ...
!> \param datatype ...
!> \param datasize ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dbcsr_mempool_get(memtype, datatype, datasize) RESULT(res)
      TYPE(dbcsr_memtype_type)                           :: memtype
      INTEGER, INTENT(IN)                                :: datatype, datasize
      TYPE(dbcsr_data_obj)                               :: res

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_mempool_get', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: best_size, s
      TYPE(dbcsr_mempool_entry_type), POINTER            :: best_cur, best_prev, cur, prev
      TYPE(dbcsr_mempool_type), POINTER                  :: pool

      pool => memtype%pool
      IF (.NOT. ASSOCIATED(pool)) CALL cp__b("dbcsr/data/dbcsr_mem_methods.F",101,"pool not allocated")

!$    CALL OMP_SET_LOCK(pool%lock)
      res%d => Null()
      best_cur => Null()
      best_prev => Null()
      best_size = HUGE(1)
      prev => Null()
      cur => pool%root
      DO WHILE (ASSOCIATED(cur%next))
         prev => cur
         cur => cur%next
         s = dbcsr_data_get_size(cur%area)
         IF (s < datasize) CYCLE
         IF (.NOT. dbcsr_memtype_equal(cur%area%d%memory_type, memtype)) CYCLE
         IF (cur%area%d%data_type /= datatype) CYCLE
         !we found a match
         IF (s < best_size) THEN
            best_cur => cur
            best_prev => prev
            best_size = s
         ENDIF
      ENDDO

      IF (ASSOCIATED(best_cur)) THEN
         IF (best_cur%area%d%refcount /= 0) CALL cp__b("dbcsr/data/dbcsr_mem_methods.F",126,"refcount /= 0")
         best_cur%area%d%refcount = 1
         best_prev%next => best_cur%next
         res = best_cur%area
         DEALLOCATE (best_cur)
      END IF
!$    CALL OMP_UNSET_LOCK(pool%lock)

      IF (.NOT. ASSOCIATED(res%d)) &
         CALL mempool_collect_garbage(pool)
   END FUNCTION dbcsr_mempool_get

! **************************************************************************************************
!> \brief Adds an unused (refcount==0) data_area to the pool.
!> \param area ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mempool_add(area)
      TYPE(dbcsr_data_obj)                               :: area

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_mempool_add', &
         routineP = moduleN//':'//routineN

      TYPE(dbcsr_mempool_entry_type), POINTER            :: new_entry
      TYPE(dbcsr_mempool_type), POINTER                  :: pool

      pool => area%d%memory_type%pool
      IF (.NOT. ASSOCIATED(pool)) CALL cp__b("dbcsr/data/dbcsr_mem_methods.F",153,"pool not allocated")
      IF (.NOT. dbcsr_data_exists(area)) CALL cp__b("dbcsr/data/dbcsr_mem_methods.F",154,"area not allocated")
      IF (area%d%refcount /= 0) CALL cp__b("dbcsr/data/dbcsr_mem_methods.F",155,"refcount /= 0")

      CALL mempool_collect_garbage(pool)

!$    CALL OMP_SET_LOCK(pool%lock)
      ALLOCATE (new_entry)
      new_entry%area = area
      new_entry%next => pool%root%next
      pool%root%next => new_entry
!$    CALL OMP_UNSET_LOCK(pool%lock)
   END SUBROUTINE dbcsr_mempool_add

! **************************************************************************************************
!> \brief Ensures that pool_size < max_size, e.g. that there is a free slot.
!> \param pool ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE mempool_collect_garbage(pool)
      TYPE(dbcsr_mempool_type), POINTER                  :: pool

      CHARACTER(LEN=*), PARAMETER :: routineN = 'mempool_collect_garbage', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: n
      TYPE(dbcsr_mempool_entry_type), POINTER            :: cur, prev

      IF (.NOT. ASSOCIATED(pool)) CALL cp__b("dbcsr/data/dbcsr_mem_methods.F",181,"pool not allocated")

!$    CALL OMP_SET_LOCK(pool%lock)
      prev => pool%root
      cur => pool%root%next
      n = 0
      DO WHILE (ASSOCIATED(cur))
         n = n+1
         IF (n >= pool%capacity) THEN
            CALL internal_data_deallocate(cur%area%d)
            DEALLOCATE (cur%area%d)
            prev%next => cur%next
            DEALLOCATE (cur)
            cur => prev%next
         ELSE
            prev => cur
            cur => cur%next
         ENDIF
      ENDDO
!$    CALL OMP_UNSET_LOCK(pool%lock)
   END SUBROUTINE mempool_collect_garbage

! **************************************************************************************************
!> \brief Finalizes mempool, includes deallocation of all contained data_areas.
!> \param pool ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mempool_destruct(pool)
      TYPE(dbcsr_mempool_type), POINTER                  :: pool

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_mempool_destruct', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(pool)) CALL cp__b("dbcsr/data/dbcsr_mem_methods.F",214,"pool not allocated")

      CALL dbcsr_mempool_clear(pool)

!$    CALL OMP_DESTROY_LOCK(pool%lock)
      DEALLOCATE (pool%root)
      DEALLOCATE (pool)

   END SUBROUTINE dbcsr_mempool_destruct

! **************************************************************************************************
!> \brief Deallocates all data_areas contained in given mempool.
!> \param pool ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_mempool_clear(pool)
      TYPE(dbcsr_mempool_type), POINTER                  :: pool

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_mempool_clear', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(dbcsr_mempool_entry_type), POINTER            :: cur, prev

      IF (.NOT. ASSOCIATED(pool)) CALL cp__b("dbcsr/data/dbcsr_mem_methods.F",238,"pool not allocated")

      CALL timeset(routineN, handle)

!$    CALL OMP_SET_LOCK(pool%lock)
      cur => pool%root%next
      DO WHILE (ASSOCIATED(cur))
         CALL internal_data_deallocate(cur%area%d)
         DEALLOCATE (cur%area%d)
         prev => cur
         cur => cur%next
         DEALLOCATE (prev)
      ENDDO
      NULLIFY (pool%root%next)
!$    CALL OMP_UNSET_LOCK(pool%lock)

      CALL timestop(handle)
   END SUBROUTINE dbcsr_mempool_clear

! **************************************************************************************************
!> \brief Ensures that given memtype has requested settings.
!> \param memtype ...
!> \param acc_hostalloc ...
!> \param acc_devalloc ...
!> \param mpi ...
!> \param acc_stream ...
!> \param oversize_factor ...
!> \param has_pool ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_memtype_setup(memtype, acc_hostalloc, acc_devalloc, mpi, &
                                  acc_stream, oversize_factor, has_pool)
      TYPE(dbcsr_memtype_type), INTENT(INOUT)            :: memtype
      LOGICAL, INTENT(IN), OPTIONAL                      :: acc_hostalloc, acc_devalloc, mpi
      TYPE(acc_stream_type), OPTIONAL                    :: acc_stream
      REAL(KIND=dp), OPTIONAL                            :: oversize_factor
      LOGICAL, INTENT(IN), OPTIONAL                      :: has_pool

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_memtype_setup', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: is_ok, my_has_pool
      TYPE(dbcsr_memtype_type)                           :: aim

! variable aim is initialized with default values from type definition

      my_has_pool = .FALSE.
      IF (PRESENT(has_pool)) my_has_pool = has_pool
      IF (PRESENT(acc_hostalloc)) aim%acc_hostalloc = acc_hostalloc
      IF (PRESENT(acc_devalloc)) aim%acc_devalloc = acc_devalloc
      IF (PRESENT(mpi)) aim%mpi = mpi
      IF (PRESENT(acc_stream)) aim%acc_stream = acc_stream
      IF (PRESENT(oversize_factor)) aim%oversize_factor = oversize_factor

      IF (.NOT. aim%acc_devalloc .EQV. acc_stream_associated(aim%acc_stream)) &
         CALL cp__b("dbcsr/data/dbcsr_mem_methods.F",293,"acc_stream missing")

      is_ok = .TRUE.
      is_ok = is_ok .AND. (memtype%acc_hostalloc .EQV. aim%acc_hostalloc)
      is_ok = is_ok .AND. (memtype%acc_devalloc .EQV. aim%acc_devalloc)
      is_ok = is_ok .AND. (memtype%mpi .EQV. aim%mpi)
      is_ok = is_ok .AND. acc_stream_equal(memtype%acc_stream, aim%acc_stream)
      is_ok = is_ok .AND. (memtype%oversize_factor == aim%oversize_factor)
      is_ok = is_ok .AND. (ASSOCIATED(memtype%pool) .EQV. my_has_pool)

      IF (.NOT. is_ok) THEN
         IF (ASSOCIATED(memtype%pool)) &
            CALL dbcsr_mempool_destruct(memtype%pool)

         memtype%acc_hostalloc = aim%acc_hostalloc
         memtype%acc_devalloc = aim%acc_devalloc
         memtype%mpi = aim%mpi
         memtype%acc_stream = aim%acc_stream
         memtype%oversize_factor = aim%oversize_factor
         IF (my_has_pool) &
            CALL dbcsr_mempool_create(memtype%pool)
      ENDIF
   END SUBROUTINE dbcsr_memtype_setup

! **************************************************************************************************
!> \brief Test if two memtypes are equal
!> \param mt1 ...
!> \param mt2 ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION dbcsr_memtype_equal(mt1, mt2) RESULT(res)
      TYPE(dbcsr_memtype_type), INTENT(in)               :: mt1, mt2
      LOGICAL                                            :: res

      res = (mt1%mpi .EQV. mt2%mpi) .AND. &
            (mt1%acc_hostalloc .EQV. mt2%acc_hostalloc) .AND. &
            (mt1%acc_devalloc .EQV. mt2%acc_devalloc) .AND. &
            (ASSOCIATED(mt1%pool) .EQV. ASSOCIATED(mt2%pool)) .AND. &
            (.NOT. ASSOCIATED(mt1%pool) .OR. ASSOCIATED(mt1%pool, mt2%pool))
   END FUNCTION dbcsr_memtype_equal

END MODULE dbcsr_mem_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR data methods
!> \author  Urban Borstnik
!> \date    2010-06-15
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2010-02-18 Moved from dbcsr_methods
! **************************************************************************************************
MODULE dbcsr_data_methods
   USE acc_devmem,                      ONLY: acc_devmem_allocate_bytes,&
                                              acc_devmem_allocated,&
                                              acc_devmem_dev2host,&
                                              acc_devmem_ensure_size_bytes,&
                                              acc_devmem_host2dev,&
                                              acc_devmem_setzero_bytes,&
                                              acc_devmem_size_in_bytes
   USE acc_event,                       ONLY: acc_event_record
   USE dbcsr_data_methods_low,          ONLY: &
        dbcsr_data_clear_pointer, dbcsr_data_exists, dbcsr_data_get_memory_type, &
        dbcsr_data_get_size, dbcsr_data_get_size_referenced, dbcsr_data_get_sizes, &
        dbcsr_data_get_type, dbcsr_data_hold, dbcsr_data_init, dbcsr_data_set_pointer, &
        dbcsr_data_set_size_referenced, dbcsr_data_valid, dbcsr_data_zero, dbcsr_get_data, &
        dbcsr_get_data_p, dbcsr_get_data_p_c, dbcsr_get_data_p_d, dbcsr_get_data_p_s, &
        dbcsr_get_data_p_z, dbcsr_scalar, dbcsr_scalar_are_equal, dbcsr_scalar_fill_all, &
        dbcsr_scalar_get_type, dbcsr_scalar_get_value, dbcsr_scalar_negative, dbcsr_scalar_one, &
        dbcsr_scalar_set_type, dbcsr_scalar_zero, dbcsr_type_1d_to_2d, dbcsr_type_2d_to_1d, &
        dbcsr_type_is_2d, internal_data_allocate, internal_data_deallocate
   USE dbcsr_data_types,                ONLY: &
        dbcsr_data_obj, dbcsr_datatype_sizeof, dbcsr_memtype_default, dbcsr_memtype_type, &
        dbcsr_type_complex_4, dbcsr_type_complex_8, dbcsr_type_int_4, dbcsr_type_int_8, &
        dbcsr_type_real_4, dbcsr_type_real_8
   USE dbcsr_mem_methods,               ONLY: dbcsr_mempool_add,&
                                              dbcsr_mempool_get
   USE dbcsr_ptr_util,                  ONLY: ensure_array_size
   USE kinds,                           ONLY: dp,&
                                              int_4,&
                                              int_8,&
                                              real_4,&
                                              real_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_data_methods'
   LOGICAL, PARAMETER :: careful_mod = .FALSE.

   INTEGER, SAVE                        :: id = 0

   PUBLIC :: dbcsr_type_2d_to_1d, dbcsr_type_1d_to_2d
   PUBLIC :: dbcsr_scalar, dbcsr_scalar_one, dbcsr_scalar_zero, &
             dbcsr_scalar_are_equal, dbcsr_scalar_negative, &
             dbcsr_scalar_get_type, dbcsr_scalar_set_type, &
             dbcsr_scalar_fill_all, dbcsr_scalar_get_value, &
             dbcsr_data_valid
   PUBLIC :: dbcsr_data_init, dbcsr_data_new, dbcsr_data_hold, &
             dbcsr_data_release, dbcsr_data_get_size, dbcsr_data_get_type
   PUBLIC :: dbcsr_get_data, &
             dbcsr_data_set_pointer, &
             dbcsr_data_clear_pointer, &
             dbcsr_data_ensure_size, &
             dbcsr_data_get_sizes, &
             dbcsr_data_get_memory_type
   PUBLIC :: dbcsr_data_set_size_referenced, dbcsr_data_get_size_referenced
   PUBLIC :: dbcsr_get_data_p, dbcsr_get_data_p_s, dbcsr_get_data_p_c, &
             dbcsr_get_data_p_d, dbcsr_get_data_p_z
   PUBLIC :: dbcsr_data_host2dev, dbcsr_data_dev2host

CONTAINS

! **************************************************************************************************
!> \brief Transfers data from host- to device-buffer, asynchronously.
!> \param area ...
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_data_host2dev(area)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_host2dev', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=real_4), DIMENSION(:), POINTER        :: c_sp
      COMPLEX(KIND=real_8), DIMENSION(:), POINTER        :: c_dp
      INTEGER(KIND=int_4), DIMENSION(:), POINTER         :: i4
      INTEGER(KIND=int_8), DIMENSION(:), POINTER         :: i8
      REAL(KIND=real_4), DIMENSION(:), POINTER           :: r_sp
      REAL(KIND=real_8), DIMENSION(:), POINTER           :: r_dp

      IF (.NOT. acc_devmem_allocated(area%d%acc_devmem)) RETURN !nothing to do
      IF (area%d%ref_size == 0) RETURN !nothing to do

      SELECT CASE (area%d%data_type)
      CASE (dbcsr_type_int_4)
         i4 => area%d%i4(:area%d%ref_size)
         CALL acc_devmem_host2dev(area%d%acc_devmem, hostmem=i4, stream=area%d%memory_type%acc_stream)
      CASE (dbcsr_type_int_8)
         i8 => area%d%i8(:area%d%ref_size)
         CALL acc_devmem_host2dev(area%d%acc_devmem, hostmem=i8, stream=area%d%memory_type%acc_stream)
      CASE (dbcsr_type_real_4)
         r_sp => area%d%r_sp(:area%d%ref_size)
         CALL acc_devmem_host2dev(area%d%acc_devmem, hostmem=r_sp, stream=area%d%memory_type%acc_stream)
      CASE (dbcsr_type_real_8)
         r_dp => area%d%r_dp(:area%d%ref_size)
         CALL acc_devmem_host2dev(area%d%acc_devmem, hostmem=r_dp, stream=area%d%memory_type%acc_stream)
      CASE (dbcsr_type_complex_4)
         c_sp => area%d%c_sp(:area%d%ref_size)
         CALL acc_devmem_host2dev(area%d%acc_devmem, hostmem=c_sp, stream=area%d%memory_type%acc_stream)
      CASE (dbcsr_type_complex_8)
         c_dp => area%d%c_dp(:area%d%ref_size)
         CALL acc_devmem_host2dev(area%d%acc_devmem, hostmem=c_dp, stream=area%d%memory_type%acc_stream)
      CASE default
         CALL cp__b("dbcsr/data/dbcsr_data_methods.F",119,"Invalid data type.")
      END SELECT

      CALL acc_event_record(area%d%acc_ready, area%d%memory_type%acc_stream)
   END SUBROUTINE dbcsr_data_host2dev

! **************************************************************************************************
!> \brief Transfers data from device- to host-buffer, asynchronously.
!> \param area ...
!> \author  Ole Schuett
! **************************************************************************************************
   SUBROUTINE dbcsr_data_dev2host(area)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_dev2host', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=real_4), DIMENSION(:), POINTER        :: c_sp
      COMPLEX(KIND=real_8), DIMENSION(:), POINTER        :: c_dp
      REAL(KIND=real_4), DIMENSION(:), POINTER           :: r_sp
      REAL(KIND=real_8), DIMENSION(:), POINTER           :: r_dp

      IF (area%d%ref_size == 0) RETURN !nothing to do

      SELECT CASE (area%d%data_type)
      CASE (dbcsr_type_real_4)
         r_sp => area%d%r_sp(:area%d%ref_size)
         CALL acc_devmem_dev2host(area%d%acc_devmem, hostmem=r_sp, stream=area%d%memory_type%acc_stream)
      CASE (dbcsr_type_real_8)
         r_dp => area%d%r_dp(:area%d%ref_size)
         CALL acc_devmem_dev2host(area%d%acc_devmem, hostmem=r_dp, stream=area%d%memory_type%acc_stream)
      CASE (dbcsr_type_complex_4)
         c_sp => area%d%c_sp(:area%d%ref_size)
         CALL acc_devmem_dev2host(area%d%acc_devmem, hostmem=c_sp, stream=area%d%memory_type%acc_stream)
      CASE (dbcsr_type_complex_8)
         c_dp => area%d%c_dp(:area%d%ref_size)
         CALL acc_devmem_dev2host(area%d%acc_devmem, hostmem=c_dp, stream=area%d%memory_type%acc_stream)
      CASE default
         CALL cp__b("dbcsr/data/dbcsr_data_methods.F",157,"Invalid data type.")
      END SELECT

   END SUBROUTINE dbcsr_data_dev2host

! **************************************************************************************************
!> \brief Initializes a data area and all the actual data pointers
!> \param[inout] area         data area
!> \param[in] data_type       select data type to use
!> \param[in] data_size       (optional) allocate this much data
!> \param[in] data_size2      (optional) second dimension data size
!> \param[in] memory_type     (optional) type of memory to use
! **************************************************************************************************
   SUBROUTINE dbcsr_data_new(area, data_type, data_size, data_size2, &
                             memory_type)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area
      INTEGER, INTENT(IN)                                :: data_type
      INTEGER, INTENT(IN), OPTIONAL                      :: data_size, data_size2
      TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL     :: memory_type

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_new', routineP = moduleN//':'//routineN

      INTEGER                                            :: d, handle, total_size_oversized, &
                                                            total_size_requested
      INTEGER, DIMENSION(2)                              :: sizes_oversized, sizes_requested
      TYPE(dbcsr_memtype_type)                           :: my_memory_type

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(area%d)) &
         CALL cp__b("dbcsr/data/dbcsr_data_methods.F",189,"area already associcated")

      my_memory_type = dbcsr_memtype_default
      IF (PRESENT(memory_type)) my_memory_type = memory_type

      sizes_requested(:) = 0; d = 1
      IF (PRESENT(data_size)) sizes_requested(1) = data_size

      IF (dbcsr_type_is_2d(data_type)) THEN
         d = 2
         IF (PRESENT(data_size2)) sizes_requested(2) = data_size2

         IF (PRESENT(data_size) .NEQV. PRESENT(data_size2)) &
            CALL cp__b("dbcsr/data/dbcsr_data_methods.F",202,"Must specify 2 sizes for 2-D data")
      ENDIF

      sizes_oversized = INT(sizes_requested*my_memory_type%oversize_factor)
      total_size_requested = PRODUCT(sizes_requested(1:d))
      total_size_oversized = PRODUCT(sizes_oversized(1:d))

      IF (total_size_requested > 1 .AND. ASSOCIATED(my_memory_type%pool)) THEN
         area = dbcsr_mempool_get(my_memory_type, data_type, total_size_requested)
      ENDIF

      IF (.NOT. ASSOCIATED(area%d)) THEN
         ALLOCATE (area%d)
!$OMP        CRITICAL (crit_area_id)
         id = id+1
         area%d%id = id
!$OMP        END CRITICAL (crit_area_id)
         area%d%refcount = 1
         area%d%memory_type = my_memory_type
         area%d%data_type = data_type
         area%d%ref_size = 0
         IF (PRESENT(data_size)) THEN
            area%d%ref_size = total_size_oversized
            CALL internal_data_allocate(area%d, sizes_oversized(1:d))
         END IF
      ENDIF

      CALL timestop(handle)
   END SUBROUTINE dbcsr_data_new

! **************************************************************************************************
!> \brief Ensures a minimum size of a previously-setup data area.
!>
!> The data area must have been previously setup with dbcsr_data_new.
!> \param[inout] area         data area
!> \param[in] data_size       allocate this much data
!> \param[in] nocopy          (optional) do not keep potentially existing data,
!>                            default is to keep it
!> \param[in] zero_pad        (optional) pad new data with zeros
!> \param[in] factor          (optional) increase size by this factor
! **************************************************************************************************
   SUBROUTINE dbcsr_data_ensure_size(area, data_size, nocopy, zero_pad, factor)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area
      INTEGER, INTENT(IN)                                :: data_size
      LOGICAL, INTENT(IN), OPTIONAL                      :: nocopy, zero_pad
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: factor

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_ensure_size', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: current_size, handle, wanted_size
      LOGICAL                                            :: nocp, pad
      TYPE(dbcsr_data_obj)                               :: area_tmp

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, handle)
      IF (.NOT. ASSOCIATED(area%d)) &
         CALL cp__b("dbcsr/data/dbcsr_data_methods.F",260,"Data area must be setup.")
      current_size = dbcsr_data_get_size(area)

      ! allocate some more as padding for libsmm kernels which read over the end.
      IF (data_size .GT. 0) THEN
         wanted_size = data_size+10
      ELSE
         wanted_size = data_size
      ENDIF

      !IF(area%d%memory_type%acc_devalloc) THEN
      !    IF(current_size==acc_devmem_size(area%d%acc_devmem)) &
      !      WRITE (*,*) "dbcsr_data_ensure_size: Host and device buffer differ in size."
      !END IF
      !   CALL dbcsr_assert(current_size==acc_devmem_size(area%d%acc_devmem),&
      !          dbcsr_fatal_level, dbcsr_caller_error,&
      !          routineN, "Host and device buffer differ in size.",276)

      CALL dbcsr_data_set_size_referenced(area, data_size)
      IF (current_size .GT. 1 .AND. current_size .GE. wanted_size) THEN
         IF (careful_mod) CALL timestop(handle)
         RETURN
      ENDIF
      !
      nocp = .FALSE.
      IF (PRESENT(nocopy)) nocp = nocopy
      pad = .FALSE.
      IF (PRESENT(zero_pad)) pad = zero_pad

      IF (dbcsr_data_exists(area)) THEN
         IF (nocp .AND. dbcsr_data_get_size(area) <= 1) &
            CALL internal_data_deallocate(area%d)
      END IF

      IF (.NOT. dbcsr_data_exists(area)) THEN
         IF (ASSOCIATED(area%d%memory_type%pool)) THEN
            area_tmp = dbcsr_mempool_get(area%d%memory_type, area%d%data_type, wanted_size)
            IF (ASSOCIATED(area_tmp%d)) THEN
               area_tmp%d%ref_size = wanted_size
               area_tmp%d%refcount = area%d%refcount
               DEALLOCATE (area%d)
               area = area_tmp
            END IF
         END IF

         IF (.NOT. dbcsr_data_exists(area)) &
            CALL internal_data_allocate(area%d, (/wanted_size/))

         IF (pad) CALL dbcsr_data_zero(area, (/1/), (/wanted_size/))
      ELSE
         SELECT CASE (area%d%data_type)
         CASE (dbcsr_type_int_8)
            CALL ensure_array_size(area%d%i8, ub=wanted_size, &
                                   memory_type=area%d%memory_type, &
                                   nocopy=nocp, zero_pad=zero_pad, &
                                   factor=factor)
         CASE (dbcsr_type_int_4)
            CALL ensure_array_size(area%d%i4, ub=wanted_size, &
                                   memory_type=area%d%memory_type, &
                                   nocopy=nocp, zero_pad=zero_pad, &
                                   factor=factor)
         CASE (dbcsr_type_real_8)
            CALL ensure_array_size(area%d%r_dp, ub=wanted_size, &
                                   memory_type=area%d%memory_type, &
                                   nocopy=nocp, zero_pad=zero_pad, &
                                   factor=factor)
         CASE (dbcsr_type_real_4)
            CALL ensure_array_size(area%d%r_sp, ub=wanted_size, &
                                   memory_type=area%d%memory_type, &
                                   nocopy=nocp, zero_pad=zero_pad, &
                                   factor=factor)
         CASE (dbcsr_type_complex_8)
            CALL ensure_array_size(area%d%c_dp, ub=wanted_size, &
                                   memory_type=area%d%memory_type, &
                                   nocopy=nocp, zero_pad=zero_pad, &
                                   factor=factor)
         CASE (dbcsr_type_complex_4)
            CALL ensure_array_size(area%d%c_sp, ub=wanted_size, &
                                   memory_type=area%d%memory_type, &
                                   nocopy=nocp, zero_pad=zero_pad, &
                                   factor=factor)
         CASE default
            CALL cp__b("dbcsr/data/dbcsr_data_methods.F",342,"Invalid data type are supported")
         END SELECT

         IF (area%d%memory_type%acc_devalloc) THEN
            IF (.NOT. acc_devmem_allocated(area%d%acc_devmem)) THEN
               CALL acc_devmem_allocate_bytes(area%d%acc_devmem, &
                                              dbcsr_datatype_sizeof(area%d%data_type)*dbcsr_data_get_size(area))
               IF (pad) CALL acc_devmem_setzero_bytes(area%d%acc_devmem, stream=area%d%memory_type%acc_stream)
            ELSE
               CALL acc_devmem_ensure_size_bytes(area%d%acc_devmem, &
                                                 area%d%memory_type%acc_stream, &
                                                 dbcsr_datatype_sizeof(area%d%data_type)*dbcsr_data_get_size(area), &
                                                 nocopy, zero_pad)
            END IF
            CALL acc_event_record(area%d%acc_ready, area%d%memory_type%acc_stream)
            IF (dbcsr_datatype_sizeof(area%d%data_type)*dbcsr_data_get_size(area) &
                /= acc_devmem_size_in_bytes(area%d%acc_devmem)) &
               CALL cp__b("dbcsr/data/dbcsr_data_methods.F",359,"Host and device buffer differ in size.")
         ENDIF

      ENDIF
      IF (careful_mod) CALL timestop(handle)
   END SUBROUTINE dbcsr_data_ensure_size

! **************************************************************************************************
!> \brief Removes a reference and/or clears the data area.
!> \param[inout] area         data area
! **************************************************************************************************
   SUBROUTINE dbcsr_data_release(area)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF (.NOT. ASSOCIATED(area%d)) &
         CALL cp__w("dbcsr/data/dbcsr_data_methods.F",383,"Data seems to be unreferenced.")
      IF (ASSOCIATED(area%d)) THEN
         !
         IF (careful_mod) THEN
            IF (area%d%refcount .LE. 0) &
               CALL cp__w("dbcsr/data/dbcsr_data_methods.F",388,"Data seems to be unreferenced.")
         ENDIF
         !
         area%d%refcount = area%d%refcount-1
         ! If we're releasing the last reference, then free the memory.
         IF (area%d%refcount .EQ. 0) THEN
            IF (.NOT. dbcsr_data_exists(area)) THEN
               DEALLOCATE (area%d)
            ELSE IF (dbcsr_data_get_size(area) > 1 .AND. ASSOCIATED(area%d%memory_type%pool)) THEN
               area%d%ref_size = 0
               CALL dbcsr_mempool_add(area)
            ELSE
               CALL internal_data_deallocate(area%d)
               DEALLOCATE (area%d)
            ENDIF
            NULLIFY (area%d)
         ENDIF
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE dbcsr_data_release

END MODULE dbcsr_data_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Base methods on DBCSR data structures
!> \author  Urban Borstnik
!> \date    2009-05-12
!> \version 0.95
!>
!> <b>Modification history:</b>
!> - Created
! **************************************************************************************************
MODULE dbcsr_methods
   USE array_types,                     ONLY: array_data,&
                                              array_nullify,&
                                              array_release
   USE btree_I8_k_cp2d_v,               ONLY: btree_destroy_c => btree_delete,&
                                              btree_new_c => btree_new
   USE btree_I8_k_dp2d_v,               ONLY: btree_destroy_d => btree_delete,&
                                              btree_new_d => btree_new
   USE btree_I8_k_sp2d_v,               ONLY: btree_destroy_s => btree_delete,&
                                              btree_new_s => btree_new
   USE btree_I8_k_zp2d_v,               ONLY: btree_destroy_z => btree_delete,&
                                              btree_new_z => btree_new
   USE dbcsr_data_methods,              ONLY: dbcsr_data_init,&
                                              dbcsr_data_release
   USE dbcsr_ptr_util,                  ONLY: memory_deallocate
   USE dbcsr_types,                     ONLY: &
        dbcsr_1d_array_type, dbcsr_2d_array_type, dbcsr_data_obj, dbcsr_distribution_obj, &
        dbcsr_imagedistribution_obj, dbcsr_imagedistribution_type, dbcsr_memtype_type, &
        dbcsr_mp_obj, dbcsr_mutable_obj, dbcsr_obj, dbcsr_type, dbcsr_type_antihermitian, &
        dbcsr_type_antisymmetric, dbcsr_type_complex_4, dbcsr_type_complex_8, &
        dbcsr_type_hermitian, dbcsr_type_invalid, dbcsr_type_no_symmetry, dbcsr_type_real_4, &
        dbcsr_type_real_8, dbcsr_type_symmetric, dbcsr_work_type
   USE kinds,                           ONLY: default_string_length
   USE message_passing,                 ONLY: mp_comm_free
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads
   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_methods'

   INTEGER, PUBLIC, SAVE :: dbcsr_matrix_counter = 111111

   PUBLIC :: dbcsr_init
   PUBLIC :: dbcsr_release
   PUBLIC :: dbcsr_valid_index, dbcsr_is_initialized
   PUBLIC :: dbcsr_release_locals
   PUBLIC :: dbcsr_distribution, &
             dbcsr_get_matrix_type, dbcsr_get_data_type, dbcsr_get_replication_type, &
             dbcsr_row_block_sizes, dbcsr_col_block_sizes, &
             dbcsr_nblkrows_total, dbcsr_nblkcols_total, dbcsr_nfullrows_total, &
             dbcsr_nfullcols_total, dbcsr_nblkcols_local, dbcsr_nblkrows_local, &
             dbcsr_max_row_size, dbcsr_max_col_size, &
             dbcsr_get_index_memory_type, dbcsr_get_data_memory_type, &
             dbcsr_name, dbcsr_setname, dbcsr_get_data_size, &
             dbcsr_use_mutable, dbcsr_wm_use_mutable, dbcsr_has_symmetry, &
             dbcsr_get_nze, dbcsr_nfullrows_local
   PUBLIC :: dbcsr_get_data_size_used
   PUBLIC :: dbcsr_col_block_offsets, dbcsr_row_block_offsets
   PUBLIC :: dbcsr_data_area
   PUBLIC :: dbcsr_get_num_blocks

   PUBLIC :: dbcsr_blk_row_size, dbcsr_blk_column_size, &
             dbcsr_blk_row_offset, dbcsr_blk_col_offset

   PUBLIC :: dbcsr_destroy_array
   PUBLIC :: dbcsr_image_dist_init, dbcsr_image_dist_hold, dbcsr_image_dist_release

   PUBLIC :: dbcsr_mutable_init, dbcsr_mutable_new, dbcsr_mutable_destroy, &
             dbcsr_mutable_release, &
             dbcsr_mutable_instantiated

   PUBLIC :: dbcsr_distribution_release, dbcsr_distribution_init
   PUBLIC :: dbcsr_mp_release, dbcsr_mp_grid_remove

   INTERFACE dbcsr_init
      MODULE PROCEDURE dbcsr_init_type, dbcsr_init_obj
   END INTERFACE

   INTERFACE dbcsr_valid_index
      MODULE PROCEDURE dbcsr_valid_index_type, dbcsr_valid_index_obj
   END INTERFACE

   INTERFACE dbcsr_is_initialized
      MODULE PROCEDURE dbcsr_is_initialized_type, dbcsr_is_initialized_obj
   END INTERFACE

   INTERFACE dbcsr_blk_row_size
      MODULE PROCEDURE dbcsr_blk_row_size_type, dbcsr_blk_row_size_obj
   END INTERFACE

   INTERFACE dbcsr_blk_column_size
      MODULE PROCEDURE dbcsr_blk_column_size_type, dbcsr_blk_column_size_obj
   END INTERFACE

   ! For the 1-D and 2-D arrays

   INTERFACE dbcsr_destroy_array
      MODULE PROCEDURE dbcsr_destroy_1d_array, dbcsr_destroy_2d_array
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief Initializes a DBCSR matrix but does not create it.
!> \param[out] matrix         uninitialized matrix
! **************************************************************************************************
   SUBROUTINE dbcsr_init_type(matrix)
      TYPE(dbcsr_type), INTENT(OUT)                      :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_init_type', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      matrix%initialized = .TRUE.
      matrix%valid = .FALSE.
      matrix%refcount = 0
      ! Nullifies all pointers.
      NULLIFY (matrix%index, matrix%row_p, matrix%col_i, &
               matrix%blk_p, matrix%thr_c, matrix%coo_l)
      CALL dbcsr_data_init(matrix%data_area)
      CALL dbcsr_distribution_init(matrix%dist)
      CALL array_nullify(matrix%row_blk_size)
      CALL array_nullify(matrix%col_blk_size)
      CALL array_nullify(matrix%row_blk_offset)
      CALL array_nullify(matrix%col_blk_offset)
      CALL array_nullify(matrix%local_rows)
      CALL array_nullify(matrix%global_rows)
      CALL array_nullify(matrix%local_cols)
      CALL array_nullify(matrix%global_cols)
      NULLIFY (matrix%wms)
   END SUBROUTINE dbcsr_init_type

! **************************************************************************************************
!> \brief ...
!> \param dist ...
! **************************************************************************************************
   SUBROUTINE dbcsr_distribution_init(dist)
      TYPE(dbcsr_distribution_obj), INTENT(OUT)          :: dist

      NULLIFY (dist%d)
   END SUBROUTINE dbcsr_distribution_init

! **************************************************************************************************
!> \brief Initializes a DBCSR matrix but does not allocate any memory.
!> \param[out] matrix         uninitialized matrix
! **************************************************************************************************
   SUBROUTINE dbcsr_init_obj(matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_init_obj', routineP = moduleN//':'//routineN

      CALL dbcsr_init(matrix%m)
   END SUBROUTINE dbcsr_init_obj

! **************************************************************************************************
!> \brief Returns whether the index structure of the matrix is valid.
!> \param[in] matrix          verify index validity of this matrix
!> \retval valid_index        index validity
! **************************************************************************************************
   PURE FUNCTION dbcsr_valid_index_type(matrix) RESULT(valid_index)
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      LOGICAL                                            :: valid_index

      valid_index = matrix%valid
   END FUNCTION dbcsr_valid_index_type

! **************************************************************************************************
!> \brief Returns whether the index structure of the matrix is valid.
!> \param[in] matrix          verify index validity of this matrix
!> \retval valid_index        index validity
! **************************************************************************************************
   PURE FUNCTION dbcsr_valid_index_obj(matrix) RESULT(valid_index)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      LOGICAL                                            :: valid_index

      valid_index = dbcsr_valid_index_type(matrix%m)
   END FUNCTION dbcsr_valid_index_obj

! **************************************************************************************************
!> \brief Returns whether the index structure of the matrix is valid.
!> \param[in] matrix          verify index validity of this matrix
!> \retval initialized        the matrix is initialized
! **************************************************************************************************
   PURE FUNCTION dbcsr_is_initialized_type(matrix) RESULT(initialized)
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      LOGICAL                                            :: initialized

      initialized = matrix%initialized
   END FUNCTION dbcsr_is_initialized_type

! **************************************************************************************************
!> \brief Returns whether the index structure of the matrix is valid.
!> \param[in] matrix          verify index validity of this matrix
!> \retval initialized        the matrix is initialized
! **************************************************************************************************
   PURE FUNCTION dbcsr_is_initialized_obj(matrix) RESULT(initialized)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      LOGICAL                                            :: initialized

      initialized = dbcsr_is_initialized_type(matrix%m)
   END FUNCTION dbcsr_is_initialized_obj

! **************************************************************************************************
!> \brief Releases a reference for a DBCSR matrix
!>
!> If there are no references left, the matrix is destroyed.
!> \param[in,out] matrix DBCSR matrix
! **************************************************************************************************
   RECURSIVE SUBROUTINE dbcsr_release(matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_release', routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      IF (.NOT. dbcsr_is_initialized(matrix)) &
         CALL cp__w("dbcsr/base/dbcsr_methods.F",225,"Matrix not initialized")
      IF (matrix%m%initialized) THEN
         matrix%m%refcount = matrix%m%refcount-1
         IF (matrix%m%refcount .EQ. 0) THEN
            CALL dbcsr_destroy(matrix)
         ENDIF
      ENDIF
   END SUBROUTINE dbcsr_release

! **************************************************************************************************
!> \brief Deallocates and destroys a matrix.
!> \param[in,out] matrix      matrix
!> \param[in] force           (optional) force deallocation
! **************************************************************************************************
   RECURSIVE SUBROUTINE dbcsr_destroy(matrix, force)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: force

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_destroy', routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle
      LOGICAL                                            :: force_all

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      IF (.NOT. dbcsr_is_initialized(matrix)) &
         CALL cp__w("dbcsr/base/dbcsr_methods.F",252,"Can not destroy uninitialized matrix object.")
      force_all = .FALSE.
      IF (PRESENT(force)) force_all = force
      IF (dbcsr_is_initialized(matrix)) THEN
         IF (.NOT. force_all .AND. matrix%m%refcount .NE. 0) &
            CALL cp__w("dbcsr/base/dbcsr_methods.F",257,"You should not destroy referenced matrix.")
         IF (force_all .AND. matrix%m%refcount .GT. 1) &
            CALL cp__w("dbcsr/base/dbcsr_methods.F",259,"You should not destroy referenced matrix.")
         IF (force_all .OR. matrix%m%refcount .EQ. 0) THEN
            IF (ASSOCIATED(matrix%m%wms)) &
               CALL cp__w("dbcsr/base/dbcsr_methods.F",262,"Destroying unfinalized matrix")
            IF (ASSOCIATED(matrix%m%index)) THEN
               CALL memory_deallocate(matrix%m%index, matrix%m%index_memory_type)
            ENDIF
            CALL dbcsr_data_release(matrix%m%data_area)
            CALL array_release(matrix%m%row_blk_size)
            CALL array_release(matrix%m%col_blk_size)
            CALL array_release(matrix%m%row_blk_offset)
            CALL array_release(matrix%m%col_blk_offset)
            CALL dbcsr_distribution_release(matrix%m%dist)
            CALL dbcsr_release_locals(matrix)
            matrix%m%valid = .FALSE.
            CALL dbcsr_init(matrix%m)
         ENDIF
      ENDIF
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_destroy

! **************************************************************************************************
!> \brief Releases and potentially destrops a distribution
!> \param dist ...
! **************************************************************************************************
   SUBROUTINE dbcsr_distribution_release(dist)
      TYPE(dbcsr_distribution_obj), INTENT(INOUT)        :: dist

!   ---------------------------------------------------------------------------

      IF (ASSOCIATED(dist%d)) THEN
         dist%d%refcount = dist%d%refcount-1
         IF (dist%d%refcount .EQ. 0) THEN
            CALL array_release(dist%d%row_dist_block)
            CALL array_release(dist%d%col_dist_block)
            CALL array_release(dist%d%local_rows)
            CALL array_release(dist%d%local_cols)
            CALL array_release(dist%d%row_dist_cluster)
            CALL array_release(dist%d%col_dist_cluster)
            CALL dbcsr_mp_release(dist%d%mp_env)
            IF (dist%d%has_thread_dist) &
               CALL array_release(dist%d%thread_dist)
            CALL array_release(dist%d%row_map)
            CALL array_release(dist%d%col_map)
            CALL dbcsr_dist_release_locals(dist)
            DEALLOCATE (dist%d)
            CALL dbcsr_distribution_init(dist)
         ENDIF
      ENDIF
   END SUBROUTINE dbcsr_distribution_release

! **************************************************************************************************
!> \brief ...
!> \param dist ...
! **************************************************************************************************
   SUBROUTINE dbcsr_dist_release_locals(dist)
      TYPE(dbcsr_distribution_obj), INTENT(INOUT)        :: dist

      INTEGER                                            :: i

      IF (dist%d%has_other_l_rows) THEN
         DO i = LBOUND(dist%d%other_l_rows, 1), UBOUND(dist%d%other_l_rows, 1)
            CALL array_release(dist%d%other_l_rows(i))
         END DO
         DEALLOCATE (dist%d%other_l_rows)
      ENDIF
      IF (dist%d%has_other_l_cols) THEN
         DO i = LBOUND(dist%d%other_l_cols, 1), UBOUND(dist%d%other_l_cols, 1)
            CALL array_release(dist%d%other_l_cols(i))
         END DO
         DEALLOCATE (dist%d%other_l_cols)
      ENDIF
      IF (dist%d%has_global_row_map) THEN
         CALL array_release(dist%d%global_row_map)
      ENDIF
      IF (dist%d%has_global_col_map) THEN
         CALL array_release(dist%d%global_col_map)
      ENDIF
      dist%d%has_other_l_rows = .FALSE.
      dist%d%has_other_l_cols = .FALSE.
      dist%d%has_global_row_map = .FALSE.
      dist%d%has_global_col_map = .FALSE.
   END SUBROUTINE dbcsr_dist_release_locals

! **************************************************************************************************
!> \brief Releases and potentially destrops an mp_env
!> \param[in,out] mp_env         multiprocessor environment
! **************************************************************************************************
   SUBROUTINE dbcsr_mp_release(mp_env)
      TYPE(dbcsr_mp_obj), INTENT(INOUT)                  :: mp_env

!   ---------------------------------------------------------------------------

      IF (ASSOCIATED(mp_env%mp)) THEN
         mp_env%mp%refcount = mp_env%mp%refcount-1
         IF (mp_env%mp%refcount .LE. 0) THEN
            CALL dbcsr_mp_grid_remove(mp_env)
            ! KG workaround
            !CALL mp_comm_free (mp_env%mp%mp_group)
            DEALLOCATE (mp_env%mp%pgrid)
            DEALLOCATE (mp_env%mp)
         ENDIF
      ENDIF
   END SUBROUTINE dbcsr_mp_release

! **************************************************************************************************
!> \brief Removes an MPI cartesian process grid
!> \param[in,out] mp_env      multiprocessor environment
! **************************************************************************************************
   SUBROUTINE dbcsr_mp_grid_remove(mp_env)
      TYPE(dbcsr_mp_obj), INTENT(INOUT)                  :: mp_env

      IF (mp_env%mp%subgroups_defined) THEN
         CALL mp_comm_free(mp_env%mp%prow_group)
         CALL mp_comm_free(mp_env%mp%pcol_group)
      ENDIF
   END SUBROUTINE dbcsr_mp_grid_remove

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
! **************************************************************************************************
   SUBROUTINE dbcsr_release_locals(matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      IF (matrix%m%has_local_rows) &
         CALL array_release(matrix%m%local_rows)
      IF (matrix%m%has_global_rows) &
         CALL array_release(matrix%m%global_rows)
      IF (matrix%m%has_local_cols) &
         CALL array_release(matrix%m%local_cols)
      IF (matrix%m%has_global_cols) &
         CALL array_release(matrix%m%global_cols)
      matrix%m%has_local_rows = .FALSE.
      matrix%m%has_global_rows = .FALSE.
      matrix%m%has_local_cols = .FALSE.
      matrix%m%has_global_cols = .FALSE.
   END SUBROUTINE dbcsr_release_locals

!  SUBROUTINE dbcsr_release_vlocals (matrix)
!    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
!
!    IF (matrix%m%has_local_vrows) &
!         CALL array_release (matrix%m%local_vrows)
!    IF (matrix%m%has_global_vrows) &
!         CALL array_release (matrix%m%global_vrows)
!    IF (matrix%m%has_local_vcols) &
!         CALL array_release (matrix%m%local_vcols)
!    IF (matrix%m%has_global_vcols) &
!         CALL array_release (matrix%m%global_vcols)
!    matrix%m%has_local_vrows  = .FALSE.
!    matrix%m%has_global_vrows = .FALSE.
!    matrix%m%has_local_vcols  = .FALSE.
!    matrix%m%has_global_vcols = .FALSE.
!  END SUBROUTINE dbcsr_release_vlocals
!

! Pertaining to the dbcsr matrix.

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nblkrows_total ...
! **************************************************************************************************
   FUNCTION dbcsr_nblkrows_total(matrix) RESULT(nblkrows_total)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: nblkrows_total

      nblkrows_total = matrix%m%nblkrows_total
   END FUNCTION dbcsr_nblkrows_total

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nblkcols_total ...
! **************************************************************************************************
   FUNCTION dbcsr_nblkcols_total(matrix) RESULT(nblkcols_total)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: nblkcols_total

      nblkcols_total = matrix%m%nblkcols_total
   END FUNCTION dbcsr_nblkcols_total
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nfullrows_total ...
! **************************************************************************************************
   FUNCTION dbcsr_nfullrows_total(matrix) RESULT(nfullrows_total)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: nfullrows_total

      nfullrows_total = matrix%m%nfullrows_total
   END FUNCTION dbcsr_nfullrows_total
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nfullcols_total ...
! **************************************************************************************************
   FUNCTION dbcsr_nfullcols_total(matrix) RESULT(nfullcols_total)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: nfullcols_total

      nfullcols_total = matrix%m%nfullcols_total
   END FUNCTION dbcsr_nfullcols_total
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nblkrows_local ...
! **************************************************************************************************
   FUNCTION dbcsr_nblkrows_local(matrix) RESULT(nblkrows_local)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: nblkrows_local

      nblkrows_local = matrix%m%nblkrows_local
   END FUNCTION dbcsr_nblkrows_local
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nblkcols_local ...
! **************************************************************************************************
   FUNCTION dbcsr_nblkcols_local(matrix) RESULT(nblkcols_local)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: nblkcols_local

      nblkcols_local = matrix%m%nblkcols_local
   END FUNCTION dbcsr_nblkcols_local
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nfullrows_local ...
! **************************************************************************************************
   FUNCTION dbcsr_nfullrows_local(matrix) RESULT(nfullrows_local)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: nfullrows_local

      nfullrows_local = matrix%m%nfullrows_local
   END FUNCTION dbcsr_nfullrows_local
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval nfullcols_local ...
! **************************************************************************************************
   FUNCTION dbcsr_nfullcols_local(matrix) RESULT(nfullcols_local)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: nfullcols_local

      nfullcols_local = matrix%m%nfullcols_local
   END FUNCTION dbcsr_nfullcols_local
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval max_row_size ...
! **************************************************************************************************
   FUNCTION dbcsr_max_row_size(matrix) RESULT(max_row_size)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: max_row_size

      max_row_size = matrix%m%max_rbs
   END FUNCTION dbcsr_max_row_size
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval max_col_size ...
! **************************************************************************************************
   FUNCTION dbcsr_max_col_size(matrix) RESULT(max_col_size)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: max_col_size

      max_col_size = matrix%m%max_cbs
   END FUNCTION dbcsr_max_col_size

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval distribution ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution(matrix) RESULT(distribution)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_distribution_obj)                       :: distribution

      distribution = matrix%m%dist
   END FUNCTION dbcsr_distribution

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval name ...
! **************************************************************************************************
   FUNCTION dbcsr_name(matrix) RESULT(name)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      CHARACTER(len=default_string_length)               :: name

      name = matrix%m%name
   END FUNCTION dbcsr_name

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param newname ...
!> \retval name ...
! **************************************************************************************************

   SUBROUTINE dbcsr_setname(matrix, newname)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      CHARACTER(len=*), INTENT(IN)                       :: newname

      matrix%m%name = newname
   END SUBROUTINE dbcsr_setname

! **************************************************************************************************
!> \brief Returns whether this work matrix uses the mutable type
!> \param[in] wm              work matrix
!> \retval use_mutable        use the mutable and not append-only working
!>                            structures
! **************************************************************************************************
   PURE FUNCTION dbcsr_wm_use_mutable(wm) RESULT(use_mutable)
      TYPE(dbcsr_work_type), INTENT(IN)                  :: wm
      LOGICAL                                            :: use_mutable

!   ---------------------------------------------------------------------------

      use_mutable = dbcsr_mutable_instantiated(wm%mutable)
   END FUNCTION dbcsr_wm_use_mutable

! **************************************************************************************************
!> \brief Returns whether work matrices should use the mutable data type
!> \param[in] matrix          matrix
!> \retval use_mutable        use the mutable and not append-only working
!>                            structures
! **************************************************************************************************
   PURE FUNCTION dbcsr_use_mutable(matrix) RESULT(use_mutable)
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      LOGICAL                                            :: use_mutable

!   ---------------------------------------------------------------------------

      use_mutable = matrix%work_mutable
   END FUNCTION dbcsr_use_mutable

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval row_blk_sizes ...
! **************************************************************************************************
   FUNCTION dbcsr_row_block_sizes(matrix) RESULT(row_blk_sizes)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, DIMENSION(:), POINTER                     :: row_blk_sizes

      row_blk_sizes => array_data(matrix%m%row_blk_size)
   END FUNCTION dbcsr_row_block_sizes

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval col_blk_sizes ...
! **************************************************************************************************
   FUNCTION dbcsr_col_block_sizes(matrix) RESULT(col_blk_sizes)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_sizes

      col_blk_sizes => array_data(matrix%m%col_blk_size)
   END FUNCTION dbcsr_col_block_sizes

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval col_blk_offsets ...
! **************************************************************************************************
   FUNCTION dbcsr_col_block_offsets(matrix) RESULT(col_blk_offsets)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_offsets

      col_blk_offsets => array_data(matrix%m%col_blk_offset)
   END FUNCTION dbcsr_col_block_offsets

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval row_blk_offsets ...
! **************************************************************************************************
   FUNCTION dbcsr_row_block_offsets(matrix) RESULT(row_blk_offsets)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, DIMENSION(:), POINTER                     :: row_blk_offsets

      row_blk_offsets => array_data(matrix%m%row_blk_offset)
   END FUNCTION dbcsr_row_block_offsets

! **************************************************************************************************
!> \brief Returns the blocked row size of a row
!>
!> This routine is optimized for speed and no checks are performed.
!> \param[in] matrix          DBCSR matrix
!> \param[in] row             row number
!> \retval row_size           blocked row size
! **************************************************************************************************
   PURE FUNCTION dbcsr_blk_row_size_type(matrix, row) RESULT(row_size)
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      INTEGER, INTENT(IN)                                :: row
      INTEGER                                            :: row_size

      row_size = matrix%row_blk_size%low%data(row)
   END FUNCTION dbcsr_blk_row_size_type
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \retval row_size ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_blk_row_size_obj(matrix, row) RESULT(row_size)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, INTENT(IN)                                :: row
      INTEGER                                            :: row_size

      row_size = matrix%m%row_blk_size%low%data(row)
   END FUNCTION dbcsr_blk_row_size_obj
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param row ...
!> \retval row_offset ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_blk_row_offset(matrix, row) RESULT(row_offset)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, INTENT(IN)                                :: row
      INTEGER                                            :: row_offset

      row_offset = matrix%m%row_blk_offset%low%data(row)
   END FUNCTION dbcsr_blk_row_offset

! **************************************************************************************************
!> \brief Returns the blocked column size of a column
!>
!> This routine is optimized for speed and no checks are performed.
!> \param[in] matrix          DBCSR matrix
!> \param[in] column          column number
!> \retval column_size        blocked row size
! **************************************************************************************************
   PURE FUNCTION dbcsr_blk_column_size_type(matrix, column) RESULT(column_size)
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      INTEGER, INTENT(IN)                                :: column
      INTEGER                                            :: column_size

      column_size = matrix%col_blk_size%low%data(column)
   END FUNCTION dbcsr_blk_column_size_type
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param column ...
!> \retval column_size ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_blk_column_size_obj(matrix, column) RESULT(column_size)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, INTENT(IN)                                :: column
      INTEGER                                            :: column_size

      column_size = matrix%m%col_blk_size%low%data(column)
   END FUNCTION dbcsr_blk_column_size_obj
! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param col ...
!> \retval col_offset ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_blk_col_offset(matrix, col) RESULT(col_offset)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, INTENT(IN)                                :: col
      INTEGER                                            :: col_offset

      col_offset = matrix%m%col_blk_offset%low%data(col)
   END FUNCTION dbcsr_blk_col_offset

! **************************************************************************************************
!> \brief Returns the data area
!> \param matrix     matrix from which to get data
!> \retval data_area data area
! **************************************************************************************************
   FUNCTION dbcsr_data_area(matrix) RESULT(data_area)
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      TYPE(dbcsr_data_obj)                               :: data_area

      data_area = matrix%data_area
   END FUNCTION dbcsr_data_area

! **************************************************************************************************
!> \brief Returns the matrix type
!> \param matrix              query this matrix
!> \retval matrix_type        matrix_type (see dbcsr_types.F for possible
!>                            values)
! **************************************************************************************************
   PURE FUNCTION dbcsr_get_matrix_type(matrix) RESULT(matrix_type)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      CHARACTER                                          :: matrix_type

      matrix_type = dbcsr_type_invalid
      IF (matrix%m%symmetry) THEN
         IF ((.NOT. matrix%m%negate_real) .AND. matrix%m%negate_imaginary) THEN
            matrix_type = dbcsr_type_hermitian
         ELSEIF (matrix%m%negate_real .AND. (.NOT. matrix%m%negate_imaginary)) THEN
            matrix_type = dbcsr_type_antihermitian
         ELSEIF (matrix%m%negate_real .AND. matrix%m%negate_imaginary) THEN
            matrix_type = dbcsr_type_antisymmetric
         ELSEIF ((.NOT. matrix%m%negate_real) .AND. (.NOT. matrix%m%negate_imaginary)) THEN
            matrix_type = dbcsr_type_symmetric
         ENDIF
      ELSE
         matrix_type = dbcsr_type_no_symmetry
      ENDIF
   END FUNCTION dbcsr_get_matrix_type

! **************************************************************************************************
!> \brief Whether matrix has symmetry
!> \param matrix              query this matrix
!> \retval has_symmetry       matrix has symmetry
! **************************************************************************************************
   PURE FUNCTION dbcsr_has_symmetry(matrix) RESULT(has_symmetry)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      LOGICAL                                            :: has_symmetry

      has_symmetry = matrix%m%symmetry
   END FUNCTION dbcsr_has_symmetry

! **************************************************************************************************
!> \brief Returns the data type stored in the matrix
!> \param matrix              query this matrix
!> \retval repl_type repl_type (see dbcsr_types.F for possible
!>                            values)
! **************************************************************************************************

   PURE FUNCTION dbcsr_get_replication_type(matrix) RESULT(repl_type)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      CHARACTER                                          :: repl_type

      repl_type = matrix%m%replication_type
   END FUNCTION dbcsr_get_replication_type

! **************************************************************************************************
!> \brief Returns the data type stored in the matrix
!> \param matrix              query this matrix
!> \retval data_type          data_type (see dbcsr_types.F for possible
!>                            values)
! **************************************************************************************************
   PURE FUNCTION dbcsr_get_data_type(matrix) RESULT(data_type)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: data_type

      data_type = matrix%m%data_type
   END FUNCTION dbcsr_get_data_type

! **************************************************************************************************
!> \brief Returns the type of memory used for data in the matrix
!> \param[in] matrix           query this matrix
!> \retval memory_type         memory type used for data
!> \note It returns the declared data type, not the actually used type
! **************************************************************************************************
   FUNCTION dbcsr_get_data_memory_type(matrix) &
      RESULT(memory_type)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_memtype_type)                           :: memory_type

      memory_type = matrix%m%data_memory_type
   END FUNCTION dbcsr_get_data_memory_type

! **************************************************************************************************
!> \brief Returns the type of memory used for the index in the matrix
!> \param[in] matrix          query this matrix
!> \retval memory_type        memory type used for the index
! **************************************************************************************************
   FUNCTION dbcsr_get_index_memory_type(matrix) RESULT(memory_type)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_memtype_type)                           :: memory_type

      memory_type = matrix%m%index_memory_type
   END FUNCTION dbcsr_get_index_memory_type

!! **************************************************************************************************
!!> \brief Returns whether the matrix uses specially-allocated memory
!!> \param[in] matrix          query this matrix
!!> \param[out] uses_special   whether the matrix uses specially allocated
!!>                            memory
!! **************************************************************************************************
!  PURE FUNCTION uses_special_memory_matrix (matrix) RESULT (uses_special)
!    TYPE(dbcsr_obj), INTENT(IN)              :: matrix
!    LOGICAL                                  :: uses_special
!
!    uses_special = matrix%m%data_memory_type .NE. dbcsr_memory_default
!  END FUNCTION uses_special_memory_matrix
!
!
!! **************************************************************************************************
!!> \brief Returns whether the data area uses speciall-allocated memory
!!> \param[in] data_area       query this data area
!!> \param[out] uses_special   whether the data area uses specially allocated
!!>                            memory
!! **************************************************************************************************
!  PURE FUNCTION uses_special_memory_area (area) RESULT (uses_special)
!    TYPE(dbcsr_data_obj), INTENT(IN)         :: area
!    LOGICAL                                  :: uses_special
!
!    IF (ASSOCIATED (area%d)) THEN
!       uses_special = area%d%memory_type .NE. dbcsr_memory_default
!    ELSE
!       uses_special = .FALSE.
!    ENDIF
!  END FUNCTION uses_special_memory_area
!

! **************************************************************************************************
!> \brief Returns the allocated data size of a DBCSR matrix
!> \param[in] matrix      matrix
!> \retval data_size      size of data
! **************************************************************************************************
   FUNCTION dbcsr_get_data_size(matrix) RESULT(data_size)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: data_size

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_size', &
         routineP = moduleN//':'//routineN

      data_size = 0
      IF (ASSOCIATED(matrix%m%data_area%d)) THEN
         SELECT CASE (matrix%m%data_area%d%data_type)
         CASE (dbcsr_type_real_8)
            IF (ASSOCIATED(matrix%m%data_area%d%r_dp)) &
               data_size = SIZE(matrix%m%data_area%d%r_dp)
         CASE (dbcsr_type_real_4)
            IF (ASSOCIATED(matrix%m%data_area%d%r_sp)) &
               data_size = SIZE(matrix%m%data_area%d%r_sp)
         CASE (dbcsr_type_complex_8)
            IF (ASSOCIATED(matrix%m%data_area%d%c_dp)) &
               data_size = SIZE(matrix%m%data_area%d%c_dp)
         CASE (dbcsr_type_complex_4)
            IF (ASSOCIATED(matrix%m%data_area%d%c_sp)) &
               data_size = SIZE(matrix%m%data_area%d%c_sp)
         CASE default
            CALL cp__b("dbcsr/base/dbcsr_methods.F",893,"Incorrect data type")
         END SELECT
      ELSE
         CALL cp__w("dbcsr/base/dbcsr_methods.F",896,"Uninitialized data area")
         data_size = 0
      ENDIF
   END FUNCTION dbcsr_get_data_size

! **************************************************************************************************
!> \brief Count actual data storage used for matrix data.
!> \param[in] matrix          Count data of this matrix
!> \retval data_size          Data size used by matrix
! **************************************************************************************************
   FUNCTION dbcsr_get_data_size_used(matrix) RESULT(data_size)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: data_size

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_data_size_used', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, error_handle, nze, row
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_sizes, row_blk_sizes

!type(dbcsr_iterator_type) :: iter
!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      row_blk_sizes => dbcsr_row_block_sizes(matrix)
      col_blk_sizes => dbcsr_col_block_sizes(matrix)
      data_size = 0
!$OMP     DO
      DO row = 1, matrix%m%nblkrows_total
         DO blk = matrix%m%row_p(row)+1, matrix%m%row_p(row+1)
            col = matrix%m%col_i(blk)
            IF (matrix%m%blk_p(blk) .NE. 0) THEN
               nze = row_blk_sizes(row)*col_blk_sizes(col)
               data_size = data_size+nze
            ENDIF
         ENDDO
      ENDDO
!$OMP     END DO
      CALL timestop(error_handle)
   END FUNCTION dbcsr_get_data_size_used

! **************************************************************************************************
!> \brief Returns the number of blocks in the matrix
!> \param matrix     matrix from which to get data
!> \retval num_blocks ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_get_num_blocks(matrix) RESULT(num_blocks)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: num_blocks

      num_blocks = matrix%m%nblks
   END FUNCTION dbcsr_get_num_blocks

! **************************************************************************************************
!> \brief Returns the number of non-zero elements in the matrix
!> \param matrix     matrix from which to get data
!> \retval num_nze ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_get_nze(matrix) RESULT(num_nze)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER                                            :: num_nze

      num_nze = matrix%m%nze
   END FUNCTION dbcsr_get_nze

! **************************************************************************************************
! Arrays
! **************************************************************************************************

! **************************************************************************************************
!> \brief Releases all matrices in a 1-d arrray.
!> \param[in,out] marray      matrix array
! **************************************************************************************************
   SUBROUTINE dbcsr_destroy_1d_array(marray)
      TYPE(dbcsr_1d_array_type), INTENT(INOUT)           :: marray

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_destroy_1d_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

!   ---------------------------------------------------------------------------

      DO i = LBOUND(marray%mats, 1), UBOUND(marray%mats, 1)
         CALL dbcsr_destroy(marray%mats(i), force=.TRUE.)
      ENDDO
      CALL dbcsr_image_dist_release(marray%image_dist)
      DEALLOCATE (marray%mats)
   END SUBROUTINE dbcsr_destroy_1d_array

! **************************************************************************************************
!> \brief Releases all matrices in 2-d arrray.
!> \param[in,out] marray      matrix array
! **************************************************************************************************
   SUBROUTINE dbcsr_destroy_2d_array(marray)
      TYPE(dbcsr_2d_array_type), INTENT(INOUT)           :: marray

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_destroy_2d_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col, row

!   ---------------------------------------------------------------------------

      DO row = LBOUND(marray%mats, 1), UBOUND(marray%mats, 1)
         DO col = LBOUND(marray%mats, 2), UBOUND(marray%mats, 2)
            CALL dbcsr_destroy(marray%mats(row, col), force=.TRUE.)
         ENDDO
      ENDDO
      CALL dbcsr_image_dist_release(marray%image_dist)
      DEALLOCATE (marray%mats)
   END SUBROUTINE dbcsr_destroy_2d_array

! **************************************************************************************************
!> \brief Releases a reference to and possible deallocates an image
!>         distribution
!> \param imgdist ...
! **************************************************************************************************
   SUBROUTINE dbcsr_image_dist_release(imgdist)
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist

      IF (ASSOCIATED(imgdist%i)) THEN
         imgdist%i%refcount = imgdist%i%refcount-1
         IF (imgdist%i%refcount .EQ. 0) THEN
            CALL dbcsr_destroy_image_dist(imgdist%i)
            DEALLOCATE (imgdist%i)
         ENDIF
      ENDIF
   END SUBROUTINE dbcsr_image_dist_release

! **************************************************************************************************
!> \brief Retains a reference to an image distribution
!> \param imgdist ...
! **************************************************************************************************
   SUBROUTINE dbcsr_image_dist_hold(imgdist)
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist

      imgdist%i%refcount = imgdist%i%refcount+1
   END SUBROUTINE dbcsr_image_dist_hold

! **************************************************************************************************
!> \brief Initialized an image distribution
!> \param imgdist ...
!> \par Akin to nullify.
! **************************************************************************************************
   SUBROUTINE dbcsr_image_dist_init(imgdist)
      TYPE(dbcsr_imagedistribution_obj), INTENT(OUT)     :: imgdist

      NULLIFY (imgdist%i)
   END SUBROUTINE dbcsr_image_dist_init

! **************************************************************************************************
!> \brief Destroys a DBCSR distribution for a matrix multiplication based on
!>        the right matrix
!> \param imgdist product distribution repetition
! **************************************************************************************************
   SUBROUTINE dbcsr_destroy_image_dist(imgdist)
      TYPE(dbcsr_imagedistribution_type), INTENT(INOUT)  :: imgdist

      INTEGER                                            :: i

!   ---------------------------------------------------------------------------

      CALL array_release(imgdist%row_image)
      CALL array_release(imgdist%col_image)
      CALL dbcsr_distribution_release(imgdist%main)
      !
      CALL array_release(imgdist%vrow_dist)
      CALL array_release(imgdist%vcol_dist)
      !
      IF (imgdist%has_other_vl_rows) THEN
         DO i = LBOUND(imgdist%other_vl_rows, 1), UBOUND(imgdist%other_vl_rows, 1)
            CALL array_release(imgdist%other_vl_rows(i))
         END DO
         DEALLOCATE (imgdist%other_vl_rows)
         imgdist%has_other_vl_rows = .FALSE.
      ENDIF
      !
      IF (imgdist%has_other_vl_cols) THEN
         DO i = LBOUND(imgdist%other_vl_cols, 1), UBOUND(imgdist%other_vl_cols, 1)
            CALL array_release(imgdist%other_vl_cols(i))
         END DO
         DEALLOCATE (imgdist%other_vl_cols)
         imgdist%has_other_vl_cols = .FALSE.
      ENDIF
      !
      IF (imgdist%has_global_vrow_map) THEN
         CALL array_release(imgdist%global_vrow_map)
      ENDIF
      IF (imgdist%has_global_vcol_map) THEN
         CALL array_release(imgdist%global_vcol_map)
      ENDIF
   END SUBROUTINE dbcsr_destroy_image_dist

! **************************************************************************************************
! Mutable data
! **************************************************************************************************

! **************************************************************************************************
!> \brief Initializes a mutable data type
!> \param[out] mutable        mutable data
! **************************************************************************************************
   SUBROUTINE dbcsr_mutable_init(mutable)
      TYPE(dbcsr_mutable_obj), INTENT(OUT)               :: mutable

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mutable_init', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      NULLIFY (mutable%m)
   END SUBROUTINE dbcsr_mutable_init

! **************************************************************************************************
!> \brief Destroys a mutable data type
!> \param[in,out] mutable     mutable data
! **************************************************************************************************
   SUBROUTINE dbcsr_mutable_destroy(mutable)
      TYPE(dbcsr_mutable_obj), INTENT(INOUT)             :: mutable

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mutable_destroy', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      IF (ASSOCIATED(mutable%m)) THEN
         CALL btree_destroy_s(mutable%m%btree_s)
         CALL btree_destroy_d(mutable%m%btree_d)
         CALL btree_destroy_c(mutable%m%btree_c)
         CALL btree_destroy_z(mutable%m%btree_z)
         DEALLOCATE (mutable%m)
      ENDIF
      NULLIFY (mutable%m)
   END SUBROUTINE dbcsr_mutable_destroy

! **************************************************************************************************
!> \brief Deregisters a reference to the mutable data type
!>
!> The object is destroy when there is no reference to it left.
!> \param[in,out] mutable     mutable data
! **************************************************************************************************
   SUBROUTINE dbcsr_mutable_release(mutable)
      TYPE(dbcsr_mutable_obj), INTENT(INOUT)             :: mutable

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mutable_release', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      IF (.NOT. ASSOCIATED(mutable%m)) &
         CALL cp__b("dbcsr/base/dbcsr_methods.F",1146,"Mutable data area not instantiated")
      mutable%m%refcount = mutable%m%refcount-1
      IF (mutable%m%refcount .EQ. 0) THEN
         CALL dbcsr_mutable_destroy(mutable)
      ENDIF
   END SUBROUTINE dbcsr_mutable_release

! **************************************************************************************************
!> \brief Creates a new mutable instance.
!>
!> \param[in,out] mutable     mutable data
!> \param[in] data_type       data type to be stored here (see dbcsr_types for
!>                            possibilities)
! **************************************************************************************************
   SUBROUTINE dbcsr_mutable_new(mutable, data_type)
      TYPE(dbcsr_mutable_obj), INTENT(INOUT)             :: mutable
      INTEGER, INTENT(IN)                                :: data_type

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_mutable_new', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      IF (ASSOCIATED(mutable%m)) &
         CALL cp__b("dbcsr/base/dbcsr_methods.F",1170,"Mutable data area already instantiated")
      IF (data_type .NE. dbcsr_type_real_4 &
          .AND. data_type .NE. dbcsr_type_real_8 &
          .AND. data_type .NE. dbcsr_type_complex_4 &
          .AND. data_type .NE. dbcsr_type_complex_8) &
         CALL cp__b("dbcsr/base/dbcsr_methods.F",1175,"Invalid data type")
      ALLOCATE (mutable%m)
      mutable%m%refcount = 1
      mutable%m%data_type = data_type
      CALL btree_new_s(mutable%m%btree_s)
      CALL btree_new_d(mutable%m%btree_d)
      CALL btree_new_c(mutable%m%btree_c)
      CALL btree_new_z(mutable%m%btree_z)
   END SUBROUTINE dbcsr_mutable_new

! **************************************************************************************************
!> \brief Deregisters a reference to the mutable data type
!>
!> The object is destroy when there is no reference to it left.
!> \param[in] mutable         mutable data
!> \retval instantiated       whether the object is instantiated
! **************************************************************************************************
   PURE FUNCTION dbcsr_mutable_instantiated(mutable) RESULT(instantiated)
      TYPE(dbcsr_mutable_obj), INTENT(IN)                :: mutable
      LOGICAL                                            :: instantiated

!   ---------------------------------------------------------------------------

      instantiated = ASSOCIATED(mutable%m)
   END FUNCTION dbcsr_mutable_instantiated

END MODULE dbcsr_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Routines related to message passing
!> \author  Ole Schuett
!> \date    10-2014
! **************************************************************************************************
MODULE dbcsr_mp_methods
   USE dbcsr_methods,                   ONLY: dbcsr_mp_grid_remove,&
                                              dbcsr_mp_release
   USE dbcsr_types,                     ONLY: dbcsr_mp_obj
   USE message_passing,                 ONLY: mp_cart_create,&
                                              mp_cart_sub,&
                                              mp_comm_free

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_mp_methods'

   PUBLIC :: dbcsr_mp_new, dbcsr_mp_hold, dbcsr_mp_release, &
             dbcsr_mp_pgrid, dbcsr_mp_numnodes, dbcsr_mp_mynode, dbcsr_mp_group, &
             dbcsr_mp_new_transposed, dbcsr_mp_nprows, dbcsr_mp_npcols, &
             dbcsr_mp_myprow, dbcsr_mp_mypcol, &
             dbcsr_mp_my_row_group, dbcsr_mp_my_col_group, &
             dbcsr_mp_has_subgroups, dbcsr_mp_get_process, &
             dbcsr_mp_grid_setup, dbcsr_mp_grid_remove, &
             dbcsr_mp_init, dbcsr_mp_active

CONTAINS

! **************************************************************************************************
!> \brief Initializes a new process grid
!> \param mp_env ...
! **************************************************************************************************
   SUBROUTINE dbcsr_mp_init(mp_env)
      TYPE(dbcsr_mp_obj), INTENT(OUT)                    :: mp_env

      NULLIFY (mp_env%mp)
   END SUBROUTINE dbcsr_mp_init

! **************************************************************************************************
!> \brief Checks whether this process is part of the message passing environment
!> \param mp_env ...
!> \retval active ...
! **************************************************************************************************
   FUNCTION dbcsr_mp_active(mp_env) RESULT(active)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      LOGICAL                                            :: active

      active = ASSOCIATED(mp_env%mp)
   END FUNCTION dbcsr_mp_active

! **************************************************************************************************
!> \brief Creates new process grid
!> \param[out] mp_env         multiprocessor environment
!> \param[in] pgrid           process grid
!> \param mp_group ...
!> \param[in] mynode          my processor number
!> \param[in] numnodes        total number of processors (processes)
!> \param myprow ...
!> \param mypcol ...
!> \par History
!>      UB (2010-02-04)       Duplicates own communicator and sets up
!>                            cartesian grid
! **************************************************************************************************
   SUBROUTINE dbcsr_mp_new(mp_env, pgrid, mp_group, mynode, numnodes, myprow, &
                           mypcol)
      TYPE(dbcsr_mp_obj), INTENT(OUT)                    :: mp_env
      INTEGER, DIMENSION(0:, 0:), INTENT(IN)             :: pgrid
      INTEGER, INTENT(IN)                                :: mp_group, mynode
      INTEGER, INTENT(IN), OPTIONAL                      :: numnodes, myprow, mypcol

      INTEGER                                            :: pcol, prow

!   ---------------------------------------------------------------------------

      ALLOCATE (mp_env%mp)
      mp_env%mp%refcount = 1
      ALLOCATE (mp_env%mp%pgrid(0:SIZE(pgrid, 1)-1, 0:SIZE(pgrid, 2)-1))
      mp_env%mp%pgrid(:, :) = pgrid(:, :)
      mp_env%mp%mynode = mynode
      mp_env%mp%mp_group = mp_group
      IF (PRESENT(numnodes)) THEN
         mp_env%mp%numnodes = numnodes
      ELSE
         mp_env%mp%numnodes = SIZE(pgrid)
      ENDIF
      IF (PRESENT(myprow) .AND. PRESENT(mypcol)) THEN
         mp_env%mp%myprow = myprow
         mp_env%mp%mypcol = mypcol
      ELSE
         mp_env%mp%myprow = -33777
         mp_env%mp%mypcol = -33777
         column_loop: DO pcol = LBOUND(pgrid, 2), UBOUND(pgrid, 2)
            row_loop: DO prow = LBOUND(pgrid, 1), UBOUND(pgrid, 1)
               test_position: IF (pgrid(prow, pcol) .EQ. mynode) THEN
                  mp_env%mp%myprow = prow
                  mp_env%mp%mypcol = pcol
                  EXIT column_loop
               ENDIF test_position
            ENDDO row_loop
         ENDDO column_loop
      ENDIF
      mp_env%mp%subgroups_defined = .FALSE.
!!!! KG workaround in place.
!!!! This will be the replacement:
!    ALLOCATE(mp_env%mp)
!    mp_env%mp%refcount = 1
!    ndims = 2
!    dims(1:2) = (/ SIZE (pgrid, 1), SIZE (pgrid, 2) /)
!    CALL mp_cart_create (mp_group, ndims,&
!         dims, my_pos,&
!         mp_env%mp%mp_group)
!    CALL mp_environ (mp_env%mp%numnodes, mp_env%mp%mynode, mp_env%mp%mp_group)
!    mp_env%mp%myprow = my_pos(1)
!    mp_env%mp%mypcol = my_pos(2)
!    ALLOCATE (mp_env%mp%pgrid (0:SIZE(pgrid, 1)-1, 0:SIZE(pgrid, 2)-1 ))
!    column_loop: DO pcol = 0, SIZE (mp_env%mp%pgrid, 2)-1
!       row_loop: DO prow = 0, SIZE (mp_env%mp%pgrid, 1)-1
!          CALL mp_cart_rank (mp_env%mp%mp_group, (/ prow, pcol /),&
!               mp_env%mp%pgrid (prow, pcol))
!       ENDDO row_loop
!    ENDDO column_loop
!    mp_env%mp%subgroups_defined = .FALSE.
   END SUBROUTINE dbcsr_mp_new

! **************************************************************************************************
!> \brief Sets up MPI cartesian process grid
!> \param[in,out] mp_env      multiprocessor environment
! **************************************************************************************************
   SUBROUTINE dbcsr_mp_grid_setup(mp_env)
      TYPE(dbcsr_mp_obj), INTENT(INOUT)                  :: mp_env

      INTEGER                                            :: ndims, tmp_group
      INTEGER, DIMENSION(2)                              :: dims, my_pos
      LOGICAL, DIMENSION(2)                              :: remain

!   ---------------------------------------------------------------------------

      IF (.NOT. mp_env%mp%subgroups_defined) THEN
         ! KG workaround.
         ! This will be deleted (replaced by code in mp_new).
         ndims = 2
         dims(1:2) = (/SIZE(mp_env%mp%pgrid, 1), SIZE(mp_env%mp%pgrid, 2)/)
         CALL mp_cart_create(mp_env%mp%mp_group, ndims, &
                             dims, my_pos, &
                             tmp_group)
         IF (my_pos(1) .NE. mp_env%mp%myprow) &
            CALL cp__b("dbcsr/dist/dbcsr_mp_methods.F",158,"Got different MPI process grid")
         IF (my_pos(2) .NE. mp_env%mp%mypcol) &
            CALL cp__b("dbcsr/dist/dbcsr_mp_methods.F",160,"Got different MPI process grid")
         !
         remain = (/.FALSE., .TRUE./)
         CALL mp_cart_sub(tmp_group, remain, mp_env%mp%prow_group)
         remain = (/.TRUE., .FALSE./)
         CALL mp_cart_sub(tmp_group, remain, mp_env%mp%pcol_group)
         CALL mp_comm_free(tmp_group)
         mp_env%mp%subgroups_defined = .TRUE.
      ENDIF
   END SUBROUTINE dbcsr_mp_grid_setup

! **************************************************************************************************
!> \brief Marks another use of the mp_env
!> \param[in,out] mp_env      multiprocessor environment
! **************************************************************************************************
   PURE SUBROUTINE dbcsr_mp_hold(mp_env)
      TYPE(dbcsr_mp_obj), INTENT(INOUT)                  :: mp_env

!   ---------------------------------------------------------------------------

      mp_env%mp%refcount = mp_env%mp%refcount+1
   END SUBROUTINE dbcsr_mp_hold

! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \param prow ...
!> \param pcol ...
!> \retval process ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_mp_get_process(mp_env, prow, pcol) RESULT(process)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, INTENT(IN)                                :: prow, pcol
      INTEGER                                            :: process

      process = mp_env%mp%pgrid(prow, pcol)
   END FUNCTION dbcsr_mp_get_process

! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \retval pgrid ...
! **************************************************************************************************
   FUNCTION dbcsr_mp_pgrid(mp_env) RESULT(pgrid)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, DIMENSION(:, :), POINTER                  :: pgrid

      pgrid => mp_env%mp%pgrid
   END FUNCTION dbcsr_mp_pgrid
! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \retval numnodes ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_mp_numnodes(mp_env) RESULT(numnodes)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER                                            :: numnodes

      numnodes = mp_env%mp%numnodes
   END FUNCTION dbcsr_mp_numnodes
! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \retval mynode ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_mp_mynode(mp_env) RESULT(mynode)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER                                            :: mynode

      mynode = mp_env%mp%mynode
   END FUNCTION dbcsr_mp_mynode
! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \retval mp_group ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_mp_group(mp_env) RESULT(mp_group)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER                                            :: mp_group

      mp_group = mp_env%mp%mp_group
   END FUNCTION dbcsr_mp_group
! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \retval nprows ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_mp_nprows(mp_env) RESULT(nprows)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER                                            :: nprows

      nprows = SIZE(mp_env%mp%pgrid, 1)
   END FUNCTION dbcsr_mp_nprows
! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \retval npcols ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_mp_npcols(mp_env) RESULT(npcols)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER                                            :: npcols

      npcols = SIZE(mp_env%mp%pgrid, 2)
   END FUNCTION dbcsr_mp_npcols
! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \retval myprow ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_mp_myprow(mp_env) RESULT(myprow)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER                                            :: myprow

      myprow = mp_env%mp%myprow
   END FUNCTION dbcsr_mp_myprow
! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \retval mypcol ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_mp_mypcol(mp_env) RESULT(mypcol)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER                                            :: mypcol

      mypcol = mp_env%mp%mypcol
   END FUNCTION dbcsr_mp_mypcol
! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \retval has_subgroups ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_mp_has_subgroups(mp_env) RESULT(has_subgroups)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      LOGICAL                                            :: has_subgroups

      has_subgroups = mp_env%mp%subgroups_defined
   END FUNCTION dbcsr_mp_has_subgroups
! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \retval row_group ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_mp_my_row_group(mp_env) RESULT(row_group)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER                                            :: row_group

      row_group = mp_env%mp%prow_group
   END FUNCTION dbcsr_mp_my_row_group
! **************************************************************************************************
!> \brief ...
!> \param mp_env ...
!> \retval col_group ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_mp_my_col_group(mp_env) RESULT(col_group)
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER                                            :: col_group

      col_group = mp_env%mp%pcol_group
   END FUNCTION dbcsr_mp_my_col_group

! **************************************************************************************************
!> \brief Transposes a multiprocessor environment
!> \param[out] mp_t           transposed multiprocessor environment
!> \param[in] mp              original multiprocessor environment
! **************************************************************************************************
   SUBROUTINE dbcsr_mp_new_transposed(mp_t, mp)
      TYPE(dbcsr_mp_obj), INTENT(OUT)                    :: mp_t
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp

!   ---------------------------------------------------------------------------

      CALL dbcsr_mp_new(mp_t, TRANSPOSE(dbcsr_mp_pgrid(mp)), &
                        dbcsr_mp_group(mp), &
                        dbcsr_mp_mynode(mp), dbcsr_mp_numnodes(mp), &
                        dbcsr_mp_mypcol(mp), dbcsr_mp_myprow(mp))
   END SUBROUTINE dbcsr_mp_new_transposed

END MODULE dbcsr_mp_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Routines related to DBCSR distributions
!> \author  Ole Schuett
!> \date    10-2014
! **************************************************************************************************
MODULE dbcsr_dist_methods
   USE array_types,                     ONLY: array_data,&
                                              array_exists,&
                                              array_i1d_obj,&
                                              array_new,&
                                              array_nullify,&
                                              array_release,&
                                              array_size
   USE dbcsr_config,                    ONLY: avg_elements_images,&
                                              comm_thread_load
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_methods,                   ONLY: dbcsr_distribution_init,&
                                              dbcsr_distribution_release
   USE dbcsr_mp_methods,                ONLY: dbcsr_mp_hold,&
                                              dbcsr_mp_mypcol,&
                                              dbcsr_mp_myprow,&
                                              dbcsr_mp_npcols,&
                                              dbcsr_mp_nprows
   USE dbcsr_toollib,                   ONLY: lcm,&
                                              sort
   USE dbcsr_types,                     ONLY: dbcsr_distribution_obj,&
                                              dbcsr_mp_obj
   USE kinds,                           ONLY: sp
   USE min_heap,                        ONLY: heap_fill,&
                                              heap_get_first,&
                                              heap_new,&
                                              heap_release,&
                                              heap_reset_first,&
                                              heap_t
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads, OMP_IN_PARALLEL
   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_dist_methods'

   PUBLIC :: dbcsr_distribution_new, dbcsr_distribution_hold, &
             dbcsr_distribution_release, dbcsr_distribution_init, &
             dbcsr_distribution_mp, dbcsr_distribution_processor, &
             dbcsr_distribution_nrows, dbcsr_distribution_ncols, &
             dbcsr_distribution_row_dist, dbcsr_distribution_col_dist, &
             dbcsr_distribution_nlocal_rows, dbcsr_distribution_nlocal_cols, &
             dbcsr_distribution_local_rows, dbcsr_distribution_local_cols, &
             dbcsr_distribution_local_rows_obj, dbcsr_distribution_local_cols_obj, &
             dbcsr_distribution_thread_dist, dbcsr_distribution_has_threads, &
             dbcsr_distribution_make_threads, dbcsr_distribution_no_threads, &
             dbcsr_distribution_num_threads, &
             dbcsr_distribution_get_num_images_1d, &
             dbcsr_distribution_has_row_clusters, dbcsr_distribution_has_col_clusters, &
             dbcsr_distribution_num_local_row_clusters, dbcsr_distribution_num_local_col_clusters, &
             dbcsr_distribution_row_clusters, dbcsr_distribution_col_clusters

CONTAINS

! **************************************************************************************************
!> \brief Creates new distribution
!> \param[out] dist           distribution
!> \param[in] mp_env          multiprocessing environment
!> \param row_dist_block ...
!> \param col_dist_block ...
!> \param local_rows ...
!> \param local_cols ...
!> \param row_dist_cluster ...
!> \param col_dist_cluster ...
!> \param reuse_arrays ...
! **************************************************************************************************
   SUBROUTINE dbcsr_distribution_new(dist, mp_env, row_dist_block, col_dist_block, &
                                     local_rows, local_cols, &
                                     row_dist_cluster, col_dist_cluster, &
                                     reuse_arrays)
      TYPE(dbcsr_distribution_obj), INTENT(OUT)          :: dist
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: row_dist_block, col_dist_block
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL, &
         POINTER                                         :: local_rows, local_cols, &
                                                            row_dist_cluster, col_dist_cluster
      LOGICAL, OPTIONAL                                  :: reuse_arrays

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_distribution_new', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, lcmv, mypcoor, nclusters, &
                                                            npcols, nprows, seq
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: cluster_dist
      INTEGER, DIMENSION(:), POINTER                     :: col_dist_tmp, row_dist_tmp

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      nprows = dbcsr_mp_nprows(mp_env)
      npcols = dbcsr_mp_npcols(mp_env)
      lcmv = lcm(nprows, npcols)

      ALLOCATE (dist%d)
      dist%d%refcount = 1
      row_dist_tmp => row_dist_block
      col_dist_tmp => col_dist_block
      CALL array_new(dist%d%row_dist_block, row_dist_tmp, gift=reuse_arrays)
      CALL array_new(dist%d%col_dist_block, col_dist_tmp, gift=reuse_arrays)

      nclusters = 0
      CALL array_nullify(dist%d%row_dist_cluster)
      IF (PRESENT(row_dist_cluster)) THEN
         IF (ASSOCIATED(row_dist_cluster)) THEN
            ! check against the case where there are more clusters than blocks
            IF (SIZE(row_dist_cluster) .GE. lcmv) THEN
               row_dist_tmp => row_dist_cluster
               CALL array_new(dist%d%row_dist_cluster, row_dist_tmp, gift=reuse_arrays)
               CALL dbcsr_assert(SIZE(row_dist_cluster), "EQ", SIZE(row_dist_block), &
                                 dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                                 "Blocks and clusters size are different for row grid", &
                                 128)
               nclusters = MAXVAL(row_dist_cluster)
            ENDIF
         ENDIF
      ENDIF
      CALL array_nullify(dist%d%col_dist_cluster)
      IF (PRESENT(col_dist_cluster)) THEN
         IF (ASSOCIATED(col_dist_cluster)) THEN
            ! check against the case where there are more clusters than blocks
            IF (SIZE(col_dist_cluster) .GE. lcmv) THEN
               col_dist_tmp => col_dist_cluster
               CALL array_new(dist%d%col_dist_cluster, col_dist_tmp, gift=reuse_arrays)
               CALL dbcsr_assert(SIZE(col_dist_cluster), "EQ", SIZE(col_dist_block), &
                                 dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                                 "Blocks and clusters size are different for column grid", &
                                 143)
               nclusters = MAX(nclusters, MAXVAL(col_dist_cluster))
            ENDIF
         ENDIF
      ENDIF
      IF (nclusters .GT. 0) THEN
         nclusters = CEILING(DBLE(nclusters)/lcmv)*lcmv
      ENDIF

      dist%d%mp_env = mp_env
      CALL dbcsr_mp_hold(dist%d%mp_env)
      ! Verify given process row distribution.
      CALL dbcsr_assert(MAXVAL(row_dist_block), "LT", nprows, &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "A process row is too big for process grid", &
                        158)
      ! Verify given process column distribution.
      CALL dbcsr_assert(MAXVAL(col_dist_block), "LT", npcols, &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "A process column is too big for process grid", &
                        163)
      IF (PRESENT(local_rows)) THEN
         CALL array_new(dist%d%local_rows, local_rows, gift=reuse_arrays)
      ELSE
         mypcoor = dbcsr_mp_myprow(mp_env)
         i = COUNT(row_dist_block .EQ. mypcoor)
         ALLOCATE (row_dist_tmp(i))
         seq = 1
         DO i = 1, SIZE(row_dist_block)
            IF (row_dist_block(i) .EQ. mypcoor) THEN
               row_dist_tmp(seq) = i
               seq = seq+1
            ENDIF
         ENDDO
         CALL array_new(dist%d%local_rows, row_dist_tmp, gift=.TRUE.)
      ENDIF
      IF (PRESENT(local_cols)) THEN
         CALL array_new(dist%d%local_cols, local_cols, gift=reuse_arrays)
      ELSE
         mypcoor = dbcsr_mp_mypcol(mp_env)
         i = COUNT(col_dist_block .EQ. mypcoor)
         ALLOCATE (col_dist_tmp(i))
         seq = 1
         DO i = 1, SIZE(col_dist_block)
            IF (col_dist_block(i) .EQ. mypcoor) THEN
               col_dist_tmp(seq) = i
               seq = seq+1
            ENDIF
         ENDDO
         CALL array_new(dist%d%local_cols, col_dist_tmp, gift=.TRUE.)
      ENDIF

      ! check cluster distributions -- rows
      dist%d%num_local_row_clusters = 1
      IF (array_exists(dist%d%row_dist_cluster)) THEN
         ! cluster/proc distribution,
         ! map of the local clusters,
         ! # blocks per cluster
         ALLOCATE (cluster_dist(nclusters))
         cluster_dist(:) = -1
         seq = 0
         mypcoor = dbcsr_mp_myprow(mp_env)
         DO i = 1, SIZE(row_dist_block)
            IF (cluster_dist(row_dist_cluster(i)) .EQ. -1) THEN
               cluster_dist(row_dist_cluster(i)) = row_dist_block(i)
               IF (row_dist_block(i) .EQ. mypcoor) seq = seq+1
            ELSE
               CALL dbcsr_assert(cluster_dist(row_dist_cluster(i)), "EQ", row_dist_block(i), &
                                 dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                                 "Wrong cluster distribution", &
                                 213)
            END IF
         END DO
         CALL dbcsr_assert(seq*nprows, "EQ", nclusters, &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Clusters distribution is not uniform across processors", &
                           219)
         dist%d%num_local_row_clusters = seq
         DEALLOCATE (cluster_dist)
      END IF

      ! check cluster distributions -- cols
      dist%d%num_local_col_clusters = 1
      IF (array_exists(dist%d%col_dist_cluster)) THEN
         ! cluster/proc distribution,
         ! map of the local clusters,
         ! # blocks per cluster
         ALLOCATE (cluster_dist(nclusters))
         cluster_dist(:) = -1
         seq = 0
         mypcoor = dbcsr_mp_mypcol(mp_env)
         DO i = 1, SIZE(col_dist_block)
            IF (cluster_dist(col_dist_cluster(i)) .EQ. -1) THEN
               cluster_dist(col_dist_cluster(i)) = col_dist_block(i)
               IF (col_dist_block(i) .EQ. mypcoor) seq = seq+1
            ELSE
               CALL dbcsr_assert(cluster_dist(col_dist_cluster(i)), "EQ", col_dist_block(i), &
                                 dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                                 "Wrong cluster distribution", &
                                 242)
            END IF
         END DO
         CALL dbcsr_assert(seq*npcols, "EQ", nclusters, &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Clusters distribution is not uniform across processors", &
                           248)
         dist%d%num_local_col_clusters = seq
         DEALLOCATE (cluster_dist)
      END IF

      dist%d%num_threads = 1
!$    dist%d%num_threads = OMP_GET_MAX_THREADS()
      dist%d%has_thread_dist = .FALSE.
      CALL array_nullify(dist%d%thread_dist)
      CALL array_nullify(dist%d%row_map)
      CALL array_nullify(dist%d%col_map)
      NULLIFY (dist%d%other_l_rows)
      NULLIFY (dist%d%other_l_cols)
      dist%d%has_other_l_rows = .FALSE.
      dist%d%has_other_l_cols = .FALSE.
      CALL array_nullify(dist%d%global_row_map)
      CALL array_nullify(dist%d%global_col_map)
      dist%d%has_global_row_map = .FALSE.
      dist%d%has_global_col_map = .FALSE.

      CALL timestop(handle)

   END SUBROUTINE dbcsr_distribution_new

! **************************************************************************************************
!> \brief Marks another use of the distribution
!> \param dist ...
! **************************************************************************************************
   SUBROUTINE dbcsr_distribution_hold(dist)
      TYPE(dbcsr_distribution_obj), INTENT(INOUT)        :: dist

!   ---------------------------------------------------------------------------

!$OMP ATOMIC
      dist%d%refcount = dist%d%refcount+1
   END SUBROUTINE dbcsr_distribution_hold

! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval mp_env ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_mp(dist) RESULT(mp_env)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      TYPE(dbcsr_mp_obj)                                 :: mp_env

!   ---------------------------------------------------------------------------

      mp_env = dist%d%mp_env
   END FUNCTION dbcsr_distribution_mp
! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval nrows ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_distribution_nrows(dist) RESULT(nrows)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER                                            :: nrows

      nrows = array_size(dist%d%row_dist_block)
   END FUNCTION dbcsr_distribution_nrows
! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval ncols ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_distribution_ncols(dist) RESULT(ncols)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER                                            :: ncols

      ncols = array_size(dist%d%col_dist_block)
   END FUNCTION dbcsr_distribution_ncols
! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval row_dist ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_row_dist(dist) RESULT(row_dist)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER, DIMENSION(:), POINTER                     :: row_dist

!   ---------------------------------------------------------------------------

      row_dist => array_data(dist%d%row_dist_block)
   END FUNCTION dbcsr_distribution_row_dist

! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval col_dist ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_col_dist(dist) RESULT(col_dist)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER, DIMENSION(:), POINTER                     :: col_dist

!   ---------------------------------------------------------------------------

      col_dist => array_data(dist%d%col_dist_block)
   END FUNCTION dbcsr_distribution_col_dist

! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval nlocalrows ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_distribution_nlocal_rows(dist) RESULT(nlocalrows)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER                                            :: nlocalrows

      nlocalrows = array_size(dist%d%local_rows)
   END FUNCTION dbcsr_distribution_nlocal_rows
! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval nlocalcols ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_distribution_nlocal_cols(dist) RESULT(nlocalcols)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER                                            :: nlocalcols

      nlocalcols = array_size(dist%d%local_cols)
   END FUNCTION dbcsr_distribution_nlocal_cols
! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval local_rows ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_local_rows(dist) RESULT(local_rows)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER, DIMENSION(:), POINTER                     :: local_rows

      local_rows => array_data(dist%d%local_rows)
   END FUNCTION dbcsr_distribution_local_rows
! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval local_rows ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_local_rows_obj(dist) RESULT(local_rows)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      TYPE(array_i1d_obj)                                :: local_rows

      local_rows = dist%d%local_rows
   END FUNCTION dbcsr_distribution_local_rows_obj
! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval local_cols ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_local_cols(dist) RESULT(local_cols)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER, DIMENSION(:), POINTER                     :: local_cols

      local_cols => array_data(dist%d%local_cols)
   END FUNCTION dbcsr_distribution_local_cols
! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval local_cols ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_local_cols_obj(dist) RESULT(local_cols)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      TYPE(array_i1d_obj)                                :: local_cols

      local_cols = dist%d%local_cols
   END FUNCTION dbcsr_distribution_local_cols_obj
   !
! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \param row ...
!> \param col ...
!> \retval processor ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_distribution_processor(dist, row, col) &
      RESULT(processor)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER, INTENT(IN)                                :: row, col
      INTEGER                                            :: processor

      INTEGER                                            :: c, r

      IF (ASSOCIATED(dist%d%row_map%low)) THEN ! instead of array_exists
         r = dist%d%row_map%low%data(row)
      ELSE
         r = row
      ENDIF
      IF (ASSOCIATED(dist%d%col_map%low)) THEN ! instead of array_exists
         c = dist%d%col_map%low%data(col)
      ELSE
         c = col
      ENDIF
      processor = dist%d%mp_env%mp%pgrid(dist%d%row_dist_block%low%data(r), &
                                         dist%d%col_dist_block%low%data(c))
   END FUNCTION dbcsr_distribution_processor

! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval thread_dist ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_thread_dist(dist) RESULT(thread_dist)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      TYPE(array_i1d_obj)                                :: thread_dist

!   ---------------------------------------------------------------------------

      thread_dist = dist%d%thread_dist
   END FUNCTION dbcsr_distribution_thread_dist

! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval has_thread_dist ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_distribution_has_threads(dist) RESULT(has_thread_dist)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      LOGICAL                                            :: has_thread_dist

!   ---------------------------------------------------------------------------

      has_thread_dist = dist%d%has_thread_dist
   END FUNCTION dbcsr_distribution_has_threads

! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval num_threads ...
! **************************************************************************************************
   PURE FUNCTION dbcsr_distribution_num_threads(dist) RESULT(num_threads)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER                                            :: num_threads

!   ---------------------------------------------------------------------------

      num_threads = dist%d%num_threads
   END FUNCTION dbcsr_distribution_num_threads

! **************************************************************************************************
!> \brief Count number of images in the product matrix
!> \param[in] matrix_dense_size_1d  1D size of the (equivalent) dense matrix
!> \param nblocks ...
!> \param[in] nprows                Number of row processors
!> \param[in] npcols                Number of column processors
!> \retval num_images               Number of images
! **************************************************************************************************
   FUNCTION dbcsr_distribution_get_num_images_1d(matrix_dense_size_1d, nblocks, &
                                                 nprows, npcols) RESULT(num_images_1d)
      INTEGER, INTENT(IN)                                :: matrix_dense_size_1d, nblocks, nprows, &
                                                            npcols
      INTEGER                                            :: num_images_1d

      INTEGER                                            :: lcmv

      lcmv = lcm(nprows, npcols)
      num_images_1d = lcmv
      IF (matrix_dense_size_1d .EQ. 0) RETURN

      IF (avg_elements_images .GT. 0) THEN
         num_images_1d = num_images_1d* &
                         CEILING((DBLE(matrix_dense_size_1d)/num_images_1d)/ &
                                 SQRT(DBLE(avg_elements_images)))
      ENDIF
      ! limiting # clusters to be close to # atoms
      IF (num_images_1d .GT. nblocks .AND. nblocks .GT. 0) THEN
         num_images_1d = CEILING(DBLE(nblocks)/lcmv)*lcmv
      ENDIF
   END FUNCTION dbcsr_distribution_get_num_images_1d

! **************************************************************************************************
!> \brief ...
!> \param[in] dist  ...
!> \retval has_row_clusters ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_has_row_clusters(dist) RESULT(has_row_clusters)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      LOGICAL                                            :: has_row_clusters

      has_row_clusters = array_exists(dist%d%row_dist_cluster)
   END FUNCTION dbcsr_distribution_has_row_clusters

! **************************************************************************************************
!> \brief ...
!> \param[in] dist  ...
!> \retval has_col_clusters ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_has_col_clusters(dist) RESULT(has_col_clusters)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      LOGICAL                                            :: has_col_clusters

      has_col_clusters = array_exists(dist%d%col_dist_cluster)
   END FUNCTION dbcsr_distribution_has_col_clusters

! **************************************************************************************************
!> \brief ...
!> \param[in] dist  ...
!> \retval num_local_row_clusters ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_num_local_row_clusters(dist) RESULT(num_local_row_clusters)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER                                            :: num_local_row_clusters

      num_local_row_clusters = dist%d%num_local_row_clusters
   END FUNCTION dbcsr_distribution_num_local_row_clusters

! **************************************************************************************************
!> \brief ...
!> \param[in] dist  ...
!> \retval num_local_col_clusters ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_num_local_col_clusters(dist) RESULT(num_local_col_clusters)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER                                            :: num_local_col_clusters

      num_local_col_clusters = dist%d%num_local_col_clusters
   END FUNCTION dbcsr_distribution_num_local_col_clusters

! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval row_dist_cluster ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_row_clusters(dist) RESULT(row_dist_cluster)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER, DIMENSION(:), POINTER                     :: row_dist_cluster

      row_dist_cluster => array_data(dist%d%row_dist_cluster)
   END FUNCTION dbcsr_distribution_row_clusters

! **************************************************************************************************
!> \brief ...
!> \param dist ...
!> \retval col_dist_cluster ...
! **************************************************************************************************
   FUNCTION dbcsr_distribution_col_clusters(dist) RESULT(col_dist_cluster)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER, DIMENSION(:), POINTER                     :: col_dist_cluster

      col_dist_cluster => array_data(dist%d%col_dist_cluster)
   END FUNCTION dbcsr_distribution_col_clusters

! **************************************************************************************************
!> \brief Creates a distribution for threads
!> \param[in,out] dist   Add thread distribution to this distribution
!> \param[in] row_sizes  (optional) row block sizes
! **************************************************************************************************
   SUBROUTINE dbcsr_distribution_make_threads(dist, row_sizes)
      TYPE(dbcsr_distribution_obj), INTENT(INOUT), &
         TARGET                                          :: dist
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: row_sizes

      TYPE(dbcsr_distribution_obj), POINTER              :: dist_p

!   ---------------------------------------------------------------------------

      dist_p => dist
!$    IF (.NOT. OMP_IN_PARALLEL()) THEN
!$OMP        PARALLEL DEFAULT(NONE) &
!$OMP                 SHARED(dist_p,row_sizes)
!$       CALL make_threads(dist_p, row_sizes=row_sizes)
!$OMP        END PARALLEL
!$    ELSE
         CALL make_threads(dist_p, row_sizes=row_sizes)
!$OMP        BARRIER
!$    ENDIF
   END SUBROUTINE dbcsr_distribution_make_threads

! **************************************************************************************************
!> \brief Creates a distribution for threads
!> \param[in,out] dist   Add thread distribution to this distribution
!> \param[in] row_sizes  (optional) row block sizes
!> \par Presence of row_sizes
!>      When row_sizes is present then the thread distribution
!>      attempts to distribute rows to threads such that the sum of
!>      delegated row sizes is approximately matched for all rows.
!>
!>      When row_sizes is not present then a random distribution is chosen.
! **************************************************************************************************
   SUBROUTINE make_threads(dist, row_sizes)
      TYPE(dbcsr_distribution_obj), POINTER              :: dist
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: row_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'make_threads', routineP = moduleN//':'//routineN

      INTEGER                                            :: block_size, block_size0, cur_block, &
                                                            group_size, i, last_row, nlrows, &
                                                            nrows, nthreads, row, t, t_cnt
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: itemp1, itemp2, reorder, sorted_row_sizes
      INTEGER, DIMENSION(:), POINTER                     :: lrows, td
      LOGICAL                                            :: assigned, found, heap_error
      REAL(kind=sp)                                      :: load_fraction, rn, soft_thr
      TYPE(heap_t)                                       :: t_heap

!   ---------------------------------------------------------------------------

      nthreads = 1
!$    nthreads = OMP_GET_NUM_THREADS(); 
!$    CALL dbcsr_assert(dist%d%num_threads, "EQ", nthreads, &
!$                      dbcsr_fatal_level, dbcsr_internal_error, routineN, &
!$                      "Thread number has changed", 647)
      nrows = dbcsr_distribution_nrows(dist)
      nlrows = dbcsr_distribution_nlocal_rows(dist)
      lrows => dbcsr_distribution_local_rows(dist)

!$OMP     BARRIER
!$OMP     MASTER

      load_fraction = REAL(comm_thread_load)/100.0
      IF (nthreads == 1) load_fraction = 1.0

      IF (.NOT. dist%d%has_thread_dist) THEN
         dist%d%num_threads = nthreads
         group_size = 0; cur_block = 0

         ALLOCATE (td(nrows))
         dist%d%has_thread_dist = .TRUE.
         CALL array_new(dist%d%thread_dist, td, gift=.TRUE.)
         td => array_data(dist%d%thread_dist)

         IF (PRESENT(row_sizes)) THEN
            ! The goal is to distribute rows to threads as equally as
            ! possible. The row sizes are first sorted. Each group of
            ! equally sized rows (group_size rows of size cur_block) is
            ! distributed to threads (keeping consecutive rows
            ! together). The group is divided into equally-sized blocks
            ! (block_size0, block_size).  Leftover rows (those that can
            ! not be equally distributed to threads) are then assigned
            ! to threads so that each thread's commulative load attempts
            ! to be equal. This distribution is achieved using a heap.
            !
            ! The heap is used to distribute "leftover"rows to threads.
            ! Lefotver rows are those of the same size that can not be
            ! evenly distributed among all threads.
            CALL heap_new(t_heap, nthreads-1)
            ! We do not want thread 0 to be in the heap.
            ALLOCATE (itemp1(1:nthreads-1))
            ALLOCATE (itemp2(1:nthreads-1))
            DO i = 1, nthreads-1
               itemp1(i) = i
               itemp2(i) = 0
            ENDDO
            CALL heap_fill(t_heap, itemp1, itemp2, heap_error)
            DEALLOCATE (itemp1, itemp2)
            ALLOCATE (sorted_row_sizes(nrows))
            ALLOCATE (reorder(nrows))
            sorted_row_sizes(:) = row_sizes(:)
            CALL sort(sorted_row_sizes, nrows, reorder)

            row = 1
            DO WHILE (row .LE. nrows)
               cur_block = sorted_row_sizes(nrows-row+1)
               assigned = .FALSE.
               group_size = 0

               last_row = nrows-row+1
               DO i = last_row, 1, -1
                  IF (cur_block == sorted_row_sizes(i)) THEN
                     group_size = group_size+1
                     row = row+1
                  ELSE
                     EXIT
                  END IF
               END DO

               soft_thr = load_fraction+nthreads-1
               block_size0 = INT(load_fraction*(group_size/soft_thr))
               block_size = INT(group_size/soft_thr)

               !blocks for master thread
               IF (block_size0 > 0) &
                  td(reorder(last_row:last_row-block_size0+1:-1)) = 0

               !Other threads
               IF (block_size > 0) THEN
                  DO t = 1, nthreads-1
                     td(reorder(last_row-block_size0-(t-1)*block_size: &
                                last_row-block_size0-(t)*block_size+1:-1)) = t
                  END DO
               END IF

               !Leftover bocks
               DO i = last_row-block_size0-(nthreads-1)*block_size, last_row+1-group_size, -1
                  CALL heap_get_first(t_heap, t, t_cnt, found, heap_error)
                  t_cnt = t_cnt+cur_block
                  CALL heap_reset_first(t_heap, t_cnt, heap_error)
                  td(reorder(i)) = t
               END DO

            END DO
            CALL heap_release(t_heap)
            DEALLOCATE (sorted_row_sizes)
            DEALLOCATE (reorder)
         ELSE
            DO t = 1, nrows
               IF (.FALSE.) THEN
                  td(t) = MOD(t-1, nthreads)
               ELSE
                  CALL RANDOM_NUMBER(rn)
                  ! Makes sure the numbers are in the proper integer range.
                  td(t) = MOD(INT(rn*REAL(nthreads)), nthreads)
               ENDIF
            END DO
         ENDIF
      ENDIF
!$OMP     END MASTER
   END SUBROUTINE make_threads

!> \brief Removes the thread distribution from a distribution
!> \param dist ...
! **************************************************************************************************
   SUBROUTINE dbcsr_distribution_no_threads(dist)
      TYPE(dbcsr_distribution_obj), INTENT(INOUT)        :: dist

!$OMP MASTER
      CALL array_release(dist%d%thread_dist)
      dist%d%has_thread_dist = .FALSE.
!$OMP END MASTER
   END SUBROUTINE dbcsr_distribution_no_threads

END MODULE dbcsr_dist_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR operations on distributions
!> \author  Urban Borstnik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2010-02-18 Moved from dbcsr_util
! **************************************************************************************************
MODULE dbcsr_dist_operations

   USE array_types,                     ONLY: &
        array_data, array_equality, array_exists, array_hold, array_i1d_obj, array_new, &
        array_nullify, array_release, array_size
   USE dbcsr_config,                    ONLY: max_elements_per_block
   USE dbcsr_dist_methods,              ONLY: &
        dbcsr_distribution_col_clusters, dbcsr_distribution_col_dist, &
        dbcsr_distribution_has_col_clusters, dbcsr_distribution_has_row_clusters, &
        dbcsr_distribution_has_threads, dbcsr_distribution_hold, dbcsr_distribution_init, &
        dbcsr_distribution_local_cols, dbcsr_distribution_local_rows, &
        dbcsr_distribution_make_threads, dbcsr_distribution_mp, dbcsr_distribution_ncols, &
        dbcsr_distribution_new, dbcsr_distribution_no_threads, dbcsr_distribution_nrows, &
        dbcsr_distribution_num_local_col_clusters, dbcsr_distribution_num_local_row_clusters, &
        dbcsr_distribution_processor, dbcsr_distribution_release, dbcsr_distribution_row_clusters, &
        dbcsr_distribution_row_dist, dbcsr_distribution_thread_dist
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_methods,                   ONLY: dbcsr_distribution,&
                                              dbcsr_is_initialized,&
                                              dbcsr_release_locals
   USE dbcsr_mp_methods,                ONLY: dbcsr_mp_mypcol,&
                                              dbcsr_mp_myprow,&
                                              dbcsr_mp_new_transposed,&
                                              dbcsr_mp_npcols,&
                                              dbcsr_mp_nprows,&
                                              dbcsr_mp_release
   USE dbcsr_toollib,                   ONLY: gcd,&
                                              lcm,&
                                              ordered_search,&
                                              sort,&
                                              swap
   USE dbcsr_types,                     ONLY: &
        dbcsr_distribution_obj, dbcsr_imagedistribution_obj, dbcsr_mp_obj, dbcsr_obj, &
        dbcsr_slot_home_pcol, dbcsr_slot_home_prow, dbcsr_slot_home_vpcol, dbcsr_slot_home_vprow, &
        dbcsr_slot_nblkcols_local, dbcsr_slot_nblkrows_local, dbcsr_type
   USE kinds,                           ONLY: sp
   USE min_heap,                        ONLY: heap_fill,&
                                              heap_get_first,&
                                              heap_new,&
                                              heap_release,&
                                              heap_reset_first,&
                                              heap_t
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  !$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

  IMPLICIT NONE

  PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_dist_operations'

  INTEGER :: idid = 0


  ! Queries
  PUBLIC :: dbcsr_get_block_index, checker_square_proc,&
            dbcsr_get_stored_coordinates, dbcsr_get_stored_block_info,&
            checker_tr, get_stored_canonical, dbcsr_find_column

  ! New/transformed distributions
  PUBLIC :: dbcsr_transpose_distribution,&
            dbcsr_transpose_dims,&
            dbcsr_reblocking_targets,&
            create_bl_distribution,&
            dbcsr_create_dist_r_unrot,&
            dbcsr_create_dist_block_cyclic

  PUBLIC :: dbcsr_create_image_dist

  ! Helper routines
  PUBLIC :: dbcsr_dist_bin,&
            make_sizes_dense,&
            image_calculator

  PUBLIC :: dbcsr_get_local_rows, dbcsr_get_local_cols,&
            dbcsr_get_local_vrows, dbcsr_get_local_vcols,&
            dbcsr_reset_locals, dbcsr_reset_vlocals

  ! For dense distributions
  PUBLIC :: dbcsr_make_dists_dense


  INTERFACE dbcsr_get_stored_block_info
     MODULE PROCEDURE get_stored_block_info_type, get_stored_block_info_obj
  END INTERFACE

  INTERFACE dbcsr_get_stored_coordinates
     MODULE PROCEDURE get_stored_coordinates_type, get_stored_coordinates_obj
  END INTERFACE

  INTERFACE dbcsr_get_block_index
     MODULE PROCEDURE get_block_index_type, get_block_index_obj
  END INTERFACE



  LOGICAL, PARAMETER :: careful_mod = .FALSE.
  LOGICAL, PARAMETER :: debug_mod = .FALSE.

CONTAINS



! **************************************************************************************************
!> \brief Determines whether a transpose must be applied
!> \param[in] row   The absolute matrix row.
!> \param[in] column          The absolute matrix column.
!> \retval transpose ...
! **************************************************************************************************
  ELEMENTAL FUNCTION checker_tr(row, column) RESULT(transpose)
      INTEGER, INTENT(IN)                                :: row, column
      LOGICAL                                            :: transpose

    transpose = BTEST(column+row, 0) .EQV. column.GE.row

  END FUNCTION checker_tr

! **************************************************************************************************
!> \brief Determines the home process for a given logical matrix element.
!> \param[in] row             logical row
!> \param[in] col             logical column
!> \param[in] pgrid           process grid
!> \param[in] row_dist        row distribution
!> \param[in] col_dist        column distribution
!> \retval process            home process of the given element
!> \note This routine is a more low-level version of
!>       dbcsr_get_stored_coordinate without the setting the row and column
!>       to the stored position.
!> \note It assumes a square matrix.
! **************************************************************************************************
  PURE FUNCTION checker_square_proc(row, col, pgrid,&
       row_dist, col_dist) RESULT(process)
      INTEGER, INTENT(IN)                                :: row, col
      INTEGER, DIMENSION(0:, 0:), INTENT(IN)             :: pgrid
      INTEGER, DIMENSION(:), INTENT(IN)                  :: row_dist, col_dist
      INTEGER                                            :: process

    IF (.NOT. checker_tr (row, col)) THEN
       process = pgrid (row_dist (row), col_dist(col))
    ELSE
       process = pgrid (row_dist (col), col_dist(row))
    ENDIF
  END FUNCTION checker_square_proc


! **************************************************************************************************
!> \brief Sets the correct source matrix, row, column and possible data
!>        transposition for a given matrix and matrix logical row and
!>        column.
!> \param[in] matrix          DBCSR matrix
!> \param[in,out] row         input is logical row, output is lookup row
!> \param[in,out] column      input is logical column, output is lookup
!>                            column
!> \param[out] processor      (optional) returns the processor on which this
!>                            block resides
! **************************************************************************************************
  PURE SUBROUTINE get_stored_coordinates_type(matrix, row, column,processor)
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      INTEGER, INTENT(INOUT)                             :: row, column
      INTEGER, INTENT(OUT), OPTIONAL                     :: processor

!   ---------------------------------------------------------------------------
! SM-compatible way

    IF (PRESENT (processor)) THEN
       IF (matrix%symmetry .AND. checker_tr(row, column)) THEN
          ! The transposed way.
          processor = dbcsr_distribution_processor (matrix%dist, column, row)
       ELSE
          ! The default way.
          processor = dbcsr_distribution_processor (matrix%dist, row, column)
       ENDIF
    ENDIF
  END SUBROUTINE get_stored_coordinates_type

! **************************************************************************************************
!> \brief Canonical logic
!> \param matrix ...
!> \param row ...
!> \param column ...
!> \param transpose ...
!> \param processor ...
! **************************************************************************************************
  PURE SUBROUTINE get_stored_canonical(matrix, row, column,&
       transpose, processor)
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      INTEGER, INTENT(INOUT)                             :: row, column
      LOGICAL, INTENT(INOUT)                             :: transpose
      INTEGER, INTENT(OUT), OPTIONAL                     :: processor

      INTEGER                                            :: tmp
      LOGICAL                                            :: straight

! The old way
!straight = matrix%transpose .OR. matrix%symmetry

    straight = matrix%symmetry
    straight = (.NOT. matrix%symmetry)&
         .OR.&
         (straight .EQV. .NOT.checker_tr(row, column))
    !transpose = .NOT. straight! .NEQV. transpose
    transpose =  straight .EQV. transpose
    IF (.NOT. straight) THEN
       tmp = row
       row = column
       column = tmp
    ENDIF
    IF (PRESENT (processor)) THEN
       processor = dbcsr_distribution_processor (matrix%dist, row, column)
    END IF
  END SUBROUTINE get_stored_canonical

! **************************************************************************************************
!> \brief Sets the correct source matrix, row, column and possible data
!>        transposition for a given matrix and matrix logical row and
!>        column.
!> \param[in] matrix          DBCSR matrix
!> \param[in,out] row         input is logical row, output is lookup row
!> \param[in,out] column      input is logical column, output is lookup
!>                            column
!> \param[out] processor      (optional) returns the processor on which this
!>                            block resides
! **************************************************************************************************
  SUBROUTINE get_stored_coordinates_obj(matrix, row, column, processor)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, INTENT(INOUT)                             :: row, column
      INTEGER, INTENT(OUT), OPTIONAL                     :: processor

      CHARACTER(len=*), PARAMETER :: routineN = 'get_stored_coordinates_obj', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    IF (debug_mod .AND. .NOT.dbcsr_is_initialized(matrix))&
       CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",253,"matrix%m not associated")
    IF (PRESENT (processor)) THEN
       CALL dbcsr_get_stored_coordinates (matrix%m, row, column, processor)
    ELSE
       CALL dbcsr_get_stored_coordinates (matrix%m, row, column)
    ENDIF
  END SUBROUTINE get_stored_coordinates_obj


! **************************************************************************************************
!> \brief Looks up a block's index given logical coordinates.
!> \param[in] matrix          DBCSR matrix
!> \param[in] row             logical row
!> \param[in] column          logical column
!> \param[out] stored_row     row where block is actually stored
!> \param[out] stored_column  column where block is actually stored
!> \param[out] transpose      whether the data must be transposed
!> \param[out] found          whether the block was found
!> \param[out] block_number   returns the block number of the row
!>                            and column
!> \param[out] data_offset    (optional) data offset for the block; 0 if
!>                            nonexistent
! **************************************************************************************************
  PURE SUBROUTINE get_block_index_type(matrix, row, column, stored_row,&
       stored_column, transpose, found, block_number, data_offset)
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      INTEGER, INTENT(IN)                                :: row, column
      INTEGER, INTENT(OUT)                               :: stored_row, stored_column
      LOGICAL, INTENT(OUT)                               :: transpose, found
      INTEGER, INTENT(OUT)                               :: block_number
      INTEGER, INTENT(OUT), OPTIONAL                     :: data_offset

!   ---------------------------------------------------------------------------

    stored_row = row
    stored_column = column
    transpose = .FALSE.
    CALL dbcsr_get_stored_coordinates (matrix, stored_row, stored_column)
    CALL get_stored_block_info_type(matrix, stored_row, stored_column,&
         found, block_number, data_offset=data_offset, transposed=transpose)
  END SUBROUTINE get_block_index_type

! **************************************************************************************************
!> \brief Sets the correct source matrix, row, column and possible data
!>        transposition for a given matrix and matrix logical row and
!>        column.
!> \param[in] matrix          DBCSR matrix
!> \param[in] row             logical row
!> \param[in] column          logical column
!> \param[out] stored_row     row where block is actually stored
!> \param[out] stored_column  column where block is actually stored
!> \param[out] transpose      whether the data must be transposed
!> \param[out] found          whether the block was found
!> \param[out] block_number   returns the block number of the row
!>                            and column
!> \param[out] data_offset    (optional) data offset for the block; 0 if
!>                            nonexistent
! **************************************************************************************************
  PURE SUBROUTINE get_block_index_obj(matrix, row, column, stored_row,&
       stored_column, transpose, found, block_number, data_offset)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, INTENT(IN)                                :: row, column
      INTEGER, INTENT(OUT)                               :: stored_row, stored_column
      LOGICAL, INTENT(OUT)                               :: transpose, found
      INTEGER, INTENT(OUT)                               :: block_number
      INTEGER, INTENT(OUT), OPTIONAL                     :: data_offset

!   ---------------------------------------------------------------------------

    IF (PRESENT (data_offset)) THEN
       CALL dbcsr_get_block_index (matrix%m, row, column, stored_row,&
            stored_column, transpose, found, block_number,&
            data_offset=data_offset)
    ELSE
       CALL dbcsr_get_block_index (matrix%m, row, column, stored_row,&
            stored_column, transpose, found, block_number)
    ENDIF
  END SUBROUTINE get_block_index_obj

! **************************************************************************************************
!> \brief Returns the index to a queried block, given a real (stored) row and
!>        column
!> \param[in] matrix          bcsr matrix
!> \param[in] row             input is logical row, output is lookup row
!> \param[in] column          input is logical column, output is lookup
!>                            column
!> \param[out] found          whether the block was found
!> \param[out] block_number   returns the block number of the row
!>                            and column
!> \param lb_row_col ...
!> \param[out] data_offset    (optional) data offset for the block; 0 if
!>                            nonexistent
!> \param[out] transposed     (optional) whether the block is stored transposed
!>                            according to its position
! **************************************************************************************************
  PURE SUBROUTINE get_stored_block_info_type(matrix, row, column,&
       found, block_number, lb_row_col, data_offset, transposed)
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      INTEGER, INTENT(IN)                                :: row, column
      LOGICAL, INTENT(OUT)                               :: found
      INTEGER, INTENT(OUT)                               :: block_number
      INTEGER, DIMENSION(2), INTENT(INOUT), OPTIONAL     :: lb_row_col
      INTEGER, INTENT(OUT), OPTIONAL                     :: data_offset
      LOGICAL, INTENT(OUT), OPTIONAL                     :: transposed

      INTEGER                                            :: blk_last, blk_offset, offset

!   ---------------------------------------------------------------------------

    IF (ASSOCIATED (matrix%row_p)) THEN
       blk_last = matrix%row_p(row+1)
       blk_offset = 0
       IF (blk_last .GT. 0) THEN
          IF (PRESENT(lb_row_col)) THEN
             IF (lb_row_col(1).EQ.row) THEN
                blk_offset = lb_row_col(2)
             ENDIF
          ENDIF
          CALL dbcsr_find_column(column, matrix%row_p(row)+blk_offset+1, blk_last,&
               matrix%col_i, matrix%blk_p,&
               block_number, found)
          blk_offset = block_number-matrix%row_p(row)
       ELSE
          found = .FALSE.
       ENDIF
       IF (PRESENT(lb_row_col)) THEN
          lb_row_col(1) = row
          lb_row_col(2) = blk_offset
       ENDIF
    ELSE
       found = .FALSE.
    ENDIF
    IF (found) THEN
       IF (PRESENT (data_offset) .OR. PRESENT (transposed)) THEN
          offset = matrix%blk_p(block_number)
       ENDIF
       IF (PRESENT (data_offset)) THEN
          data_offset = ABS (offset)
       ENDIF
       IF (PRESENT (transposed)) THEN
          transposed = offset .LT. 0
       ENDIF
    ELSE
       IF (PRESENT (data_offset)) THEN
          data_offset = 0
       ENDIF
       IF (PRESENT (transposed)) THEN
          transposed = .FALSE.
       ENDIF
    ENDIF
  END SUBROUTINE get_stored_block_info_type

! **************************************************************************************************
!> \brief Returns the index to a queried block, given a real (stored) row and
!>        column
!> \param[in] matrix          bcsr matrix
!> \param[in] row             input is logical row, output is lookup row
!> \param[in] column          input is logical column, output is lookup
!>                            column
!> \param[out] found          whether the block was found
!> \param[out] block_number   returns the block number of the row
!>                            and column; 0 if nonexistent
!> \param lb_row_col ...
!> \param[out] data_offset    (optional) data offset for the block; 0 if
!>                            nonexistent
!> \param[out] transposed     (optional) whether the block is stored transposed
!>                            according to its position
! **************************************************************************************************
  PURE SUBROUTINE get_stored_block_info_obj(matrix, row, column,&
       found, block_number, lb_row_col, data_offset, transposed)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, INTENT(IN)                                :: row, column
      LOGICAL, INTENT(OUT)                               :: found
      INTEGER, INTENT(OUT)                               :: block_number
      INTEGER, DIMENSION(2), INTENT(INOUT), OPTIONAL     :: lb_row_col
      INTEGER, INTENT(OUT), OPTIONAL                     :: data_offset
      LOGICAL, INTENT(OUT), OPTIONAL                     :: transposed

!   ---------------------------------------------------------------------------

    IF (PRESENT (data_offset)) THEN
       IF (PRESENT (transposed)) THEN
          CALL dbcsr_get_stored_block_info (matrix%m, row, column, found,&
               block_number, lb_row_col, data_offset, transposed)
       ELSE
          CALL dbcsr_get_stored_block_info (matrix%m, row, column, found,&
               block_number, lb_row_col, data_offset)
       ENDIF
    ELSE
       IF (PRESENT (transposed)) THEN
          CALL dbcsr_get_stored_block_info (matrix%m, row, column, found,&
               block_number, lb_row_col, transposed=transposed)
       ELSE
          CALL dbcsr_get_stored_block_info (matrix%m, row, column, found,&
               block_number, lb_row_col)
       ENDIF
    ENDIF
  END SUBROUTINE get_stored_block_info_obj

! **************************************************************************************************
!> \brief Transposes a distribution
!> \param[out] dist_tr        transposed distribution
!> \param[in] dist_normal current distribution
! **************************************************************************************************
  SUBROUTINE dbcsr_transpose_distribution(dist_tr, dist_normal)
      TYPE(dbcsr_distribution_obj), INTENT(OUT)          :: dist_tr
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist_normal

      TYPE(dbcsr_mp_obj)                                 :: mp_env_tr

!   ---------------------------------------------------------------------------

    CALL dbcsr_mp_new_transposed (mp_env_tr, dbcsr_distribution_mp (&
         dist_normal))
    CALL dbcsr_distribution_new(dist_tr, mp_env_tr, &
         dbcsr_distribution_col_dist (dist_normal), &
         dbcsr_distribution_row_dist (dist_normal), &
         dbcsr_distribution_local_cols (dist_normal), &
         dbcsr_distribution_local_rows (dist_normal), &
         dbcsr_distribution_col_clusters (dist_normal), &
         dbcsr_distribution_row_clusters (dist_normal))
    CALL dbcsr_mp_release (mp_env_tr)
  END SUBROUTINE dbcsr_transpose_distribution

! **************************************************************************************************
!> \brief Transposes a distribution but keeps the same mp_env
!> \param[out] dist_tr        transposed distribution
!> \param[in] dist_normal current distribution
! **************************************************************************************************
  SUBROUTINE dbcsr_transpose_dims(dist_tr, dist_normal)
      TYPE(dbcsr_distribution_obj), INTENT(OUT)          :: dist_tr
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist_normal

      INTEGER                                            :: i, j, nclusters, ncols_tr, npcols_tr, &
                                                            nprows_tr, nrows_tr, vgcd, vlcm
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: img_data, reorder
      INTEGER, DIMENSION(:), POINTER                     :: col_dist_data_tr, col_img_data_tr, &
                                                            row_dist_data_tr, row_img_data_tr
      TYPE(dbcsr_mp_obj)                                 :: mp_env

!   ---------------------------------------------------------------------------

    mp_env = dbcsr_distribution_mp (dist_normal)
    !
    ! transpose the rows/cols
    nrows_tr = dbcsr_distribution_ncols (dist_normal)
    ncols_tr = dbcsr_distribution_nrows (dist_normal)
    ! procs are not transposed
    nprows_tr = dbcsr_mp_nprows (mp_env)
    npcols_tr = dbcsr_mp_npcols (mp_env)
    vgcd = gcd(nprows_tr,npcols_tr)
    vlcm = lcm(nprows_tr,npcols_tr)
    !
    ALLOCATE (row_dist_data_tr (nrows_tr))
    ALLOCATE (row_img_data_tr (nrows_tr))
    IF (dbcsr_distribution_has_col_clusters(dist_normal)) THEN
       nclusters = npcols_tr*dbcsr_distribution_num_local_col_clusters (dist_normal)
       ALLOCATE(img_data(nrows_tr))
       img_data(:) = MOD(dbcsr_distribution_col_clusters (dist_normal)-1,&
            dbcsr_distribution_num_local_col_clusters (dist_normal))
       CALL rebin_imaged_distribution_clusters (row_dist_data_tr, row_img_data_tr,&
            dbcsr_distribution_col_dist (dist_normal),&
            img_data,&
            nprows_tr,&
            nclusters/gcd(nclusters,npcols_tr),&
            nclusters)
       ! Sort clusters for convenience
       ALLOCATE(reorder(SIZE(row_img_data_tr)))
       CALL sort(row_img_data_tr,SIZE(row_img_data_tr),reorder)
       ! Realign blocks distribution
       DO i = 1, SIZE(reorder)
          j = reorder(i)
          IF (i.GE.j) CYCLE
          CALL swap(reorder(i),reorder(j))
          CALL swap(row_dist_data_tr(i),row_dist_data_tr(j))
       ENDDO
       DEALLOCATE(img_data,reorder)
    ELSE
       CALL rebin_distribution (row_dist_data_tr, row_img_data_tr,&
            dbcsr_distribution_col_dist (dist_normal),&
            nprows_tr,&
            nprows_tr/vgcd,vlcm/nprows_tr)
       !
       ! discard image distribution, it will be build later in the multiplication
       DEALLOCATE(row_img_data_tr)
       NULLIFY(row_img_data_tr)
    ENDIF
    !
    ALLOCATE (col_dist_data_tr (ncols_tr))
    ALLOCATE (col_img_data_tr (ncols_tr))
    IF (dbcsr_distribution_has_row_clusters(dist_normal)) THEN
       nclusters = nprows_tr*dbcsr_distribution_num_local_row_clusters (dist_normal)
       ALLOCATE(img_data(ncols_tr))
       img_data(:) = MOD(dbcsr_distribution_row_clusters (dist_normal)-1,&
            dbcsr_distribution_num_local_row_clusters (dist_normal))
       CALL rebin_imaged_distribution_clusters (col_dist_data_tr, col_img_data_tr,&
            dbcsr_distribution_row_dist (dist_normal),&
            img_data,&
            npcols_tr,&
            nclusters/gcd(nclusters,nprows_tr),&
            nclusters)
       ! Sort clusters for convenience
       ALLOCATE(reorder(SIZE(col_img_data_tr)))
       CALL sort(col_img_data_tr,SIZE(col_img_data_tr),reorder)
       ! Realign blocks distribution
       DO i = 1, SIZE(reorder)
          j = reorder(i)
          IF (i.GE.j) CYCLE
          CALL swap(reorder(i),reorder(j))
          CALL swap(col_dist_data_tr(i),col_dist_data_tr(j))
       ENDDO
       DEALLOCATE(img_data,reorder)
    ELSE
       CALL rebin_distribution (col_dist_data_tr, col_img_data_tr,&
            dbcsr_distribution_row_dist (dist_normal),&
            npcols_tr,&
            npcols_tr/vgcd,vlcm/npcols_tr)
       !
       ! discard image distribution, it will be build later in the multiplication
       DEALLOCATE(col_img_data_tr)
       NULLIFY(col_img_data_tr)
    ENDIF
    !
    CALL dbcsr_distribution_new(dist_tr, mp_env, &
         row_dist_data_tr, col_dist_data_tr, &
         row_dist_cluster=row_img_data_tr, col_dist_cluster=col_img_data_tr, &
         reuse_arrays=.TRUE.)

  END SUBROUTINE dbcsr_transpose_dims


! **************************************************************************************************
!> \brief Evenly distributes elements into bins
!> \param[out] array          elements to bin
!> \param[in] nbins           number of bins into which elemnts are distributed
! **************************************************************************************************
  PURE SUBROUTINE evenly_bin (array, nbins)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: array
      INTEGER, INTENT(IN)                                :: nbins

      INTEGER                                            :: bin, first, last, maxbinsize

!   ---------------------------------------------------------------------------

    maxbinsize = CEILING (REAL(SIZE (array)) / REAL(nbins))
    DO bin = 1, nbins
       first = 1 + (bin-1)*maxbinsize
       last = MIN (1 + (bin)*maxbinsize-1, SIZE (array))
       array (first : last) = bin - 1
    ENDDO
  END SUBROUTINE evenly_bin


! **************************************************************************************************
!> \brief Makes new distribution with decimation and multiplicity
!> \param[out] new_bins      new real distribution
!> \param[out] images        new image distribution
!> \param[in] source_bins    Basis for the new distribution and images
!> \param[in] nbins          number of bins in the new real distribution
!> \param[in] multiplicity   multiplicity
!> \param[in] nimages        number of images in the new distribution
!> \par Definition of multiplicity and nimages
!>      Multiplicity and decimation (number of images) are used to
!>      match process grid coordinates on non-square process
!>      grids. Given source_nbins and target_nbins, their relation is
!>                source_nbins * target_multiplicity
!>              = target_nbins * target_nimages.
!>      It is best when both multiplicity and nimages are small. To
!>      get these two factors, then, one can use the following formulas:
!>          nimages      = lcm(source_nbins, target_nbins) / target_nbins
!>          multiplicity = target_nbins / gcd(source_nbins, target_nbins)
!>      from the target's point of view (nimages = target_nimages).
!> \par Mapping
!>      The new distribution comprises of real bins and images within
!>      bins. These can be view as target_nbins*nimages virtual
!>      columns. These same virtual columns are also
!>      source_nbins*multiplicity in number. Therefore these virtual
!>      columns are mapped from source_nbins*multiplicity onto
!>      target_bins*nimages (each target bin has nimages images):
!>      Source 4: |1 2 3|4 5 6|7 8 9|A B C| (4*3)
!>      Target 6: |1 2|3 4|5 6|7 8|9 A|B C| (6*2)
!>      multiplicity=3, nimages=2, 12 virtual columns (1-C).
!>      Source bin elements are evenly mapped into one of multiplicity
!>      virtual columns. Other (non-even, block-size aware) mappings
!>      could be better.
! **************************************************************************************************
  SUBROUTINE rebin_distribution (new_bins, images, source_bins,&
       nbins, multiplicity, nimages)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: new_bins, images
      INTEGER, DIMENSION(:), INTENT(IN)                  :: source_bins
      INTEGER, INTENT(IN)                                :: nbins, multiplicity, nimages

      CHARACTER(len=*), PARAMETER :: routineN = 'rebin_distribution', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bin, i, old_nbins, virtual_bin
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: bin_multiplier

!   ---------------------------------------------------------------------------

    IF(MOD(nbins*nimages,multiplicity).NE.0)&
       CALL cp__w("dbcsr/dist/dbcsr_dist_operations.F",654,"mulitplicity is not divisor of new process grid coordinate")
    old_nbins = (nbins * nimages) / multiplicity
    ALLOCATE (bin_multiplier(0:old_nbins-1))
    bin_multiplier(:) = 0
    DO i = 1, SIZE (new_bins)
       IF (i .LE. SIZE (source_bins)) THEN
          bin = source_bins (i)
       ELSE
          ! Fill remainder with a cyclic distribution
          bin = MOD (i, old_nbins)
       ENDIF
       virtual_bin = bin*multiplicity+bin_multiplier(bin)
       new_bins(i) = virtual_bin / nimages
       images(i) = 1 + MOD (virtual_bin, nimages)
       bin_multiplier(bin) = bin_multiplier(bin)+1
       IF (bin_multiplier(bin) .GE. multiplicity) THEN
          bin_multiplier(bin) = 0
       ENDIF
    ENDDO
  END SUBROUTINE rebin_distribution

! **************************************************************************************************
!> \brief Makes new distribution with decimation and multiplicity
!>        Multiplicity is beeing ignored, maybe this is a bug
!> \param[out] images        new image distribution
!> \param[in] my_bins        Basis for the new images
!> \param[in] nbins          number of bins in the new real distribution
!> \param[in] nimages        number of images in the new distribution
!> \par Definition of multiplicity and nimages
!>      Multiplicity and decimation (number of images) are used to
!>      match process grid coordinates on non-square process
!>      grids. Given source_nbins and target_nbins, their relation is
!>                source_nbins * target_multiplicity
!>              = target_nbins * target_nimages.
!>      It is best when both multiplicity and nimages are small. To
!>      get these two factors, then, one can use the following formulas:
!>          nimages      = lcm(source_nbins, target_nbins) / target_nbins
!>          multiplicity = target_nbins / gcd(source_nbins, target_nbins)
!>      from the target's point of view (nimages = target_nimages).
!> \par Mapping
!>      The new distribution comprises of real bins and images within
!>      bins. These can be view as target_nbins*nimages virtual
!>      columns. These same virtual columns are also
!>      source_nbins*multiplicity in number. Therefore these virtual
!>      columns are mapped from source_nbins*multiplicity onto
!>      target_bins*nimages (each target bin has nimages images):
!>      Source 4: |1 2 3|4 5 6|7 8 9|A B C| (4*3)
!>      Target 6: |1 2|3 4|5 6|7 8|9 A|B C| (6*2)
!>      multiplicity=3, nimages=2, 12 virtual columns (1-C).
!>      Source bin elements are evenly mapped into one of multiplicity
!>      virtual columns. Other (non-even, block-size aware) mappings
!>      could be better.
! **************************************************************************************************
  SUBROUTINE reimage_distribution (images, my_bins,&
       nbins, nimages)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: images
      INTEGER, DIMENSION(:), INTENT(IN)                  :: my_bins
      INTEGER, INTENT(IN)                                :: nbins, nimages

      CHARACTER(len=*), PARAMETER :: routineN = 'reimage_distribution', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bin, i
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: bin_multiplier

!   ---------------------------------------------------------------------------

    ALLOCATE (bin_multiplier(0:nbins-1))
    bin_multiplier(:) = 0
    DO i = 1, SIZE (my_bins)
       bin = my_bins (i)
       images(i) = 1 + bin_multiplier(bin)
       bin_multiplier(bin) = bin_multiplier(bin)+1
       IF (bin_multiplier(bin) .GE. nimages) THEN
          bin_multiplier(bin) = 0
       ENDIF
    ENDDO
  END SUBROUTINE reimage_distribution

! **************************************************************************************************
!> \brief Makes new virtual distribution of rows/columns.
!> \param[in] nelements      number of elements
!> \param[out] vbins         virtual bins
!> \param[in] bins           bins to which elements belong
!> \param[in] decimation     matching between bins
!> \param[in] images         images to which element belong
! **************************************************************************************************
  PURE SUBROUTINE make_vdistribution (nelements, vbins, bins, decimation, images)
      INTEGER, INTENT(IN)                                :: nelements
      INTEGER, DIMENSION(nelements), INTENT(OUT)         :: vbins
      INTEGER, DIMENSION(nelements), INTENT(IN)          :: bins
      INTEGER, INTENT(IN)                                :: decimation
      INTEGER, DIMENSION(nelements), INTENT(IN)          :: images

      CHARACTER(len=*), PARAMETER :: routineN = 'make_vdistribution', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: el

!   ---------------------------------------------------------------------------

    DO el = 1 , nelements
       vbins(el) = bins(el) * decimation + images(el)-1
    END DO
  END SUBROUTINE make_vdistribution


! **************************************************************************************************
!> \brief Makes new distribution with multiplicity
!> \param[out] new_bins      new real distribution
!> \param[out] images        new image distribution
!> \param[in] source_bins    Basis for the new distribution and images
!> \param[in] source_images  Basis for the new distribution and images
!> \param[in] nbins          number of bins in the new real distribution
!> \param[in] multiplicity   multiplicity
!> \param[in] nimages        number of images in the new distribution
!> \par Definition of multiplicity and nimages
!>      Multiplicity and number of images are used to match process
!>      grid coordinates on non-square process grids. Given
!>      source_nbins and target_nbins, their relation is
!>          source_nbins * multiplicity = target_nbins * nimages.
!>      It is best when both multiplicity and nimages are small. To
!>      get these two factors, then, one can use the following formulas:
!>          nimages      = lcm(source_nbins, target_nbins) / target_nbins
!>          multiplicity = target_nbins / gcd(source_nbins, target_nbins)
!> \par Mapping
!>      The new distribution comprises of real bins and images within
!>      bins. These can be view as target_nbins*nimages virtual
!>      columns. These same virtual columns are also
!>      source_nbins*multiplicity in number. Therefore these virtual
!>      columns are mapped from source_nbins*multiplicity onto
!>      target_bins*nimages (each target bin has nimages images):
!>      Source 4: |1 2 3|4 5 6|7 8 9|A B C| (4*3)
!>      Target 6: |1 2|3 4|5 6|7 8|9 A|B C| (6*2)
!>      multiplicity=3, nimages=2, 12 virtual columns (1-C).
!>      Source bin elements are evenly mapped into one of multiplicity
!>      virtual columns. Other (non-even, block-size aware) mappings
!>      could be better.
! **************************************************************************************************
  SUBROUTINE rebin_imaged_distribution (new_bins, images,&
       source_bins, source_images, nbins, multiplicity, nimages)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: new_bins, images
      INTEGER, DIMENSION(:), INTENT(IN)                  :: source_bins, source_images
      INTEGER, INTENT(IN)                                :: nbins, multiplicity, nimages

      CHARACTER(len=*), PARAMETER :: routineN = 'rebin_imaged_distribution', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, virtual_bin

!   ---------------------------------------------------------------------------

    DO i = 1, SIZE (new_bins)
       IF (i .LE. SIZE (source_bins)) THEN
          virtual_bin = source_bins(i)*multiplicity + source_images(i)-1
       ELSE
          ! Fill remainder with a cyclic distribution
          virtual_bin = MOD (i, nbins*nimages)
       ENDIF
       new_bins(i) = virtual_bin / nimages
       images(i) = 1 + MOD (virtual_bin, nimages)
       IF(new_bins(i).ge.nbins)&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",816,"Wrong bin calculation")
       IF(images(i).gt.nimages)&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",818,"Wrong image calculation")
    ENDDO
  END SUBROUTINE rebin_imaged_distribution

! **************************************************************************************************
!> \brief ...
!> \param new_bins ...
!> \param new_images ...
!> \param source_bins ...
!> \param source_images ...
!> \param nbins ...
!> \param multiplicity ...
!> \param nclusters ...
! **************************************************************************************************
  SUBROUTINE rebin_imaged_distribution_clusters (new_bins, new_images,&
       source_bins, source_images, nbins, multiplicity, nclusters)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: new_bins, new_images
      INTEGER, DIMENSION(:), INTENT(IN)                  :: source_bins, source_images
      INTEGER, INTENT(IN)                                :: nbins, multiplicity, nclusters

      CHARACTER(len=*), PARAMETER :: routineN = 'rebin_imaged_distribution_clusters', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dim, i, nimages, virtual_bin

!   ---------------------------------------------------------------------------

    dim = SIZE (new_bins)
    nimages = nclusters/nbins
    IF(dim.NE.SIZE(source_bins))&
       CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",848,"Wrong bins array dimension")
    IF(dim.NE.SIZE(new_images).OR.dim.NE.SIZE(source_images))&
       CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",850,"Wrong images array dimensions")
    DO i = 1, dim
       virtual_bin = source_bins(i)*multiplicity + source_images(i)
       new_bins(i) = virtual_bin / nimages
       new_images(i) = 1 + virtual_bin
       IF(new_bins(i).ge.nbins)&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",856,"Wrong bin calculation")
       IF(new_images(i).gt.nclusters)&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",858,"Wrong image calculation")
    ENDDO
  END SUBROUTINE rebin_imaged_distribution_clusters


! **************************************************************************************************
!> \brief ...
!> \param imgdist ...
!> \param dist ...
!> \param template ...
! **************************************************************************************************
  SUBROUTINE dbcsr_new_image_dist (imgdist, dist,&
       template)
      TYPE(dbcsr_imagedistribution_obj), INTENT(OUT)     :: imgdist
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      TYPE(dbcsr_imagedistribution_obj), INTENT(IN)      :: template

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_new_image_dist', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    idid = idid+1
    ALLOCATE (imgdist%i)
    imgdist%i%refcount = 1
    imgdist%i%id = idid
    imgdist%i%row_decimation   = template%i%row_decimation
    imgdist%i%row_multiplicity = template%i%row_multiplicity
    imgdist%i%col_decimation   = template%i%col_decimation
    imgdist%i%col_multiplicity = template%i%col_multiplicity
    !
    NULLIFY (imgdist%i%other_vl_rows)
    NULLIFY (imgdist%i%other_vl_cols)
    CALL array_nullify (imgdist%i%global_vrow_map)
    CALL array_nullify (imgdist%i%global_vcol_map)
    imgdist%i%has_other_vl_rows = .FALSE.
    imgdist%i%has_other_vl_cols = .FALSE.
    imgdist%i%has_global_vrow_map = .FALSE.
    imgdist%i%has_global_vcol_map = .FALSE.
    !
    imgdist%i%main = dist
    CALL dbcsr_distribution_hold (imgdist%i%main)
    !
  END SUBROUTINE dbcsr_new_image_dist

! **************************************************************************************************
!> \brief Creates an image distribution given the other compatibility images
!> \param[out] imgdist                  distribution repetition
!> \param[in] dist                      distribution for which to form the
!>                                      image distribution
!> \param[in] match_row_pdist           (optional) match the new row
!>                                      distribution to this row distribution
!> \param[in] match_row_idist           (optional) match the row
!>                                      distribution to these row images
!> \param[in] match_row_nbins           number of bins in the distribution to
!>                                      match the local rows
!> \param[in] match_col_pdist           (optional) match the new column
!>                                      distribution to this column
!>                                      distribution
!> \param[in] match_col_idist           (optional) match the new column
!>                                      distribution to these column images
!> \param[in] match_col_nbins           number of bins in the distribution to
!>                                      match the local columns
!> \param nimages_rows ...
!> \param nimages_cols ...
! **************************************************************************************************
  SUBROUTINE dbcsr_create_image_dist(imgdist, dist,&
       match_row_pdist, match_row_idist, match_row_nbins,&
       match_col_pdist, match_col_idist, match_col_nbins,&
       nimages_rows, nimages_cols)
      TYPE(dbcsr_imagedistribution_obj), INTENT(OUT)     :: imgdist
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: match_row_pdist, match_row_idist
      INTEGER, INTENT(IN)                                :: match_row_nbins
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: match_col_pdist, match_col_idist
      INTEGER, INTENT(IN)                                :: match_col_nbins, nimages_rows, &
                                                            nimages_cols

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_create_image_dist', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ncols, npcols, nprows, nrows
      INTEGER, DIMENSION(:), POINTER                     :: col_dist_data, col_img_data, col_vdist, &
                                                            row_dist_data, row_img_data, row_vdist
      LOGICAL                                            :: new_col_dist, new_row_dist
      TYPE(dbcsr_distribution_obj)                       :: new_dist
      TYPE(dbcsr_mp_obj)                                 :: mp_env

!   ---------------------------------------------------------------------------

    idid = idid+1
    ALLOCATE (imgdist%i)
    imgdist%i%refcount = 1
    imgdist%i%id = idid
    CALL dbcsr_distribution_init (new_dist)
    CALL dbcsr_distribution_init (imgdist%i%main)
    mp_env = dbcsr_distribution_mp (dist)
    ! Determine the factors.
    nrows = dbcsr_distribution_nrows (dist)
    ncols = dbcsr_distribution_ncols (dist)
    nprows = dbcsr_mp_nprows (mp_env)
    npcols = dbcsr_mp_npcols (mp_env)
    IF (debug_mod) WRITE (*,'(1X,A,I5,"x",I5)')routineN//"pgrid",&
         nprows, npcols
    !
    ! Create the new row distribution and row image distribution
    imgdist%i%row_decimation = nimages_rows / nprows
    imgdist%i%row_multiplicity = nimages_rows / gcd (nimages_rows, match_row_nbins)
    new_row_dist = .FALSE.
    !
    IF (debug_mod) WRITE (*,*)routineN//'row decimation, multiplicity',&
         imgdist%i%row_decimation, imgdist%i%row_multiplicity
    IF (debug_mod) WRITE(*,*)routineN//" nprows, match prows", nprows, match_row_nbins
    ALLOCATE (row_img_data (nrows))
    ALLOCATE (row_vdist (nrows))
    !
    IF (dbcsr_distribution_has_row_clusters(dist)) THEN
       IF (PRESENT(match_row_pdist).AND.PRESENT(match_row_idist).AND.&
            imgdist%i%row_decimation.NE.imgdist%i%row_multiplicity) THEN
          ALLOCATE (row_dist_data (nrows))
          new_row_dist = .TRUE.
          CALL rebin_imaged_distribution (row_dist_data, row_img_data,&
               match_row_pdist, match_row_idist,&
               nprows,&
               imgdist%i%row_multiplicity, imgdist%i%row_decimation)
       ELSE
          row_dist_data => dbcsr_distribution_row_dist (dist)
          IF (imgdist%i%row_decimation.EQ.1) THEN
             row_img_data(:) = 1
          ELSE
             row_img_data = MOD(dbcsr_distribution_row_clusters(dist),imgdist%i%row_decimation)+1
          ENDIF
       ENDIF
    ELSE
       IF (imgdist%i%row_decimation.EQ.1 .AND. imgdist%i%row_multiplicity.EQ.1 .AND. &
            .NOT. PRESENT (match_row_pdist)) THEN
          row_dist_data => dbcsr_distribution_row_dist (dist)
          row_img_data(:) = 1
       ELSE
          IF (PRESENT (match_row_pdist)) THEN
             ALLOCATE (row_dist_data (nrows))
             new_row_dist = .TRUE.
             IF (PRESENT (match_row_idist)) THEN
                CALL rebin_imaged_distribution (row_dist_data, row_img_data,&
                     match_row_pdist, match_row_idist,&
                     nprows,&
                     imgdist%i%row_multiplicity, imgdist%i%row_decimation)
             ELSE
                CALL rebin_distribution (row_dist_data, row_img_data,&
                     match_row_pdist, &
                     nprows,&
                     imgdist%i%row_multiplicity, imgdist%i%row_decimation)
             ENDIF
          ELSE
             row_dist_data => dbcsr_distribution_row_dist (dist)
             CALL reimage_distribution (row_img_data,&
                  row_dist_data,nprows,imgdist%i%row_decimation)
          ENDIF
       ENDIF
    ENDIF
    CALL make_vdistribution (nrows, row_vdist, row_dist_data,&
         imgdist%i%row_decimation, row_img_data)
    CALL array_new (imgdist%i%vrow_dist, row_vdist, gift=.TRUE.)
    !
    ! Create the new column distribution and column image distribution
    imgdist%i%col_decimation = nimages_cols / npcols
    imgdist%i%col_multiplicity = nimages_cols / gcd (nimages_cols, match_col_nbins)
    new_col_dist = .FALSE.
    !
    IF (debug_mod) WRITE (*,*)routineN//'col decimation, multiplicity',&
         imgdist%i%col_decimation, imgdist%i%col_multiplicity
    IF (debug_mod) WRITE(*,*)routineN//" npcols, match pcols", npcols, match_col_nbins
    ALLOCATE (col_img_data (ncols))
    ALLOCATE (col_vdist (ncols))
    !
    IF (dbcsr_distribution_has_col_clusters(dist)) THEN
       IF (PRESENT(match_col_pdist).AND.PRESENT(match_col_idist).AND.&
            imgdist%i%col_decimation.NE.imgdist%i%col_multiplicity) THEN
          ALLOCATE (col_dist_data (ncols))
          new_col_dist = .TRUE.
          CALL rebin_imaged_distribution (col_dist_data, col_img_data,&
               match_col_pdist, match_col_idist,&
               npcols,&
               imgdist%i%col_multiplicity, imgdist%i%col_decimation)
       ELSE
          col_dist_data => dbcsr_distribution_col_dist (dist)
          IF (imgdist%i%col_decimation.EQ.1) THEN
             col_img_data(:) = 1
          ELSE
             col_img_data = MOD(dbcsr_distribution_col_clusters(dist),imgdist%i%col_decimation)+1
          ENDIF
       ENDIF
    ELSE
       IF (imgdist%i%col_decimation.EQ.1 .AND. imgdist%i%col_multiplicity.EQ.1 .AND. &
            .NOT. PRESENT (match_col_pdist)) THEN
          col_dist_data => dbcsr_distribution_col_dist (dist)
          col_img_data(:) = 1
       ELSE
          IF (PRESENT (match_col_pdist)) THEN
             ALLOCATE (col_dist_data (ncols))
             new_col_dist = .TRUE.
             IF (PRESENT (match_col_idist)) THEN
                CALL rebin_imaged_distribution (col_dist_data, col_img_data,&
                     match_col_pdist, match_col_idist,&
                     npcols,&
                     imgdist%i%col_multiplicity, imgdist%i%col_decimation)
             ELSE
                CALL rebin_distribution (col_dist_data, col_img_data,&
                     match_col_pdist,&
                     npcols,&
                     imgdist%i%col_multiplicity, imgdist%i%col_decimation)
             ENDIF
          ELSE
             col_dist_data => dbcsr_distribution_col_dist (dist)
             CALL reimage_distribution (col_img_data,&
                  col_dist_data,&
                  npcols,imgdist%i%col_decimation)
          ENDIF
       ENDIF
    ENDIF
    CALL make_vdistribution (ncols, col_vdist, col_dist_data,&
         imgdist%i%col_decimation, col_img_data)
    CALL array_new (imgdist%i%vcol_dist, col_vdist, gift=.TRUE.)
    !
    ! Copy the row & column distribution from old distribution
    IF (new_row_dist.AND.new_col_dist) THEN
       CALL dbcsr_distribution_new(new_dist,&
            mp_env,&
            row_dist_data, col_dist_data,&
            reuse_arrays=.TRUE.)
    ELSE
       CALL dbcsr_distribution_new(new_dist,&
            mp_env,&
            row_dist_data, col_dist_data)
       IF (new_row_dist) DEALLOCATE(row_dist_data)
       IF (new_col_dist) DEALLOCATE(col_dist_data)
    ENDIF
    ! Now finish the distribution image.
    imgdist%i%main = new_dist
    CALL array_new (imgdist%i%col_image, col_img_data, gift=.TRUE.)
    CALL array_new (imgdist%i%row_image, row_img_data, gift=.TRUE.)
    !
    imgdist%i%has_other_vl_rows = .FALSE.
    imgdist%i%has_other_vl_cols = .FALSE.
    imgdist%i%has_global_vrow_map = .FALSE.
    imgdist%i%has_global_vcol_map = .FALSE.
    !
!$  IF (dbcsr_distribution_has_threads(dist)) THEN
!$     imgdist%i%main%d%num_threads = dist%d%num_threads
!$     imgdist%i%main%d%has_thread_dist = .TRUE.
!$     imgdist%i%main%d%thread_dist = dist%d%thread_dist
!$     CALL array_hold(imgdist%i%main%d%thread_dist)
!$  ENDIF
  END SUBROUTINE dbcsr_create_image_dist



! **************************************************************************************************
!> \brief Calculates the intersections of blocks
!> \param[out] ints           intersections of blocks
!> \param[in] numints        maximum number of expected intersections 
!> \param[out] n_src_dsts     offset and number intersections belonging
!>                            to source blocks
!> \param[in] src_sizes       sizes of source blocks
!> \param[in] dst_sizes       sizes of target blocks
!> \par ints output format
!>      The ints array should be up to twice as large as the number of
!>      intersecting blocks. Each entry is comprised of the target
!>      block and the common length along with the offsets of the
!>      intersection in the old and new blocks.
!> \par n_src_dsts format
!>      This arrays stored the number of intersecting blocks in common
!>      (position 2) and the offset of the first common intersecting
!>      block (position 1).
!> \par Interpretation (Mapping old blocks into new blocks)
!>      The old element belongs to block B. Lookup row B in
!>      n_src_dsts.  The first element F tells you which is the first
!>      new block to map into and the second element tells you into
!>      how many new blocks N you have to map.  You then look up rows
!>      F to F+N-1 in ints. The first block tells you into which block
!>      it is mapped and the secend element tells you how many
!>      elements they have in common. The third element specifies the
!>      offset of the intesection in the old block while the fourth
!>      specifies the offset of the intersection in the new block.
!> \par History
!>      UB [2009] Created
!>      UB [2010-03-22] Offsets of the intersections in the old/new blocks
!>                      are calculated & saved. Zero-length blocks are
!>                      skipped.
!> \note This routine is used in the counting and sending loops in
!>       dbcsr_complete_redistribute
! **************************************************************************************************
  SUBROUTINE dbcsr_reblocking_targets (ints, numints, n_src_dsts,&
       src_sizes, dst_sizes)
      INTEGER, INTENT(INOUT)                             :: numints
      INTEGER, DIMENSION(4, numints), INTENT(OUT)        :: ints
      INTEGER, DIMENSION(:, :), INTENT(OUT)              :: n_src_dsts
      INTEGER, DIMENSION(:), INTENT(IN)                  :: src_sizes, dst_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_reblocking_targets', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: common_extent, current_dst, current_int, &
                                                            current_src, dst_off, n_dst, n_src, &
                                                            s_dst, s_src, src_off

!   ---------------------------------------------------------------------------

    n_src = SIZE(src_sizes)
    n_dst = SIZE(dst_sizes)
    current_int = 0
    current_src = 0
    s_src = 0 ! HUGE(0)
    DO WHILE (s_src .EQ. 0 .AND. current_src .LE. n_src)
       current_src = current_src + 1
       src_off = 1
       IF (current_src .LE. n_src) THEN
          s_src = src_sizes (current_src)
          n_src_dsts(:, current_src) = (/ current_int+1, 0 /)
       ENDIF
    END DO
    current_dst = 0
    s_dst = 0 ! HUGE(0)
    DO WHILE (s_dst .EQ. 0 .AND. current_dst .LE. n_dst)
       current_dst = current_dst + 1
       dst_off = 1
       IF (current_dst .LE. n_dst) s_dst = dst_sizes (current_dst)
    END DO
    current_int = current_int + 1
    DO WHILE (current_src .LE. n_src .AND. current_dst .LE. n_dst)
       CALL dbcsr_assert (current_int, "LE", numints, &
            dbcsr_fatal_level, dbcsr_internal_error, routineN,&
            "Ran out of space.", 1190)
       ! Calculate how many elements do the current blocks have in
       ! common and record these as going to the current target block.
       common_extent = MIN (s_src, s_dst)
       ints(1, current_int) = current_dst ! target block
       ints(2, current_int) = common_extent
       ints(3, current_int) = src_off
       ints(4, current_int) = dst_off
       ! We've used up the common extents.
       s_src = s_src - common_extent
       s_dst = s_dst - common_extent
       src_off = src_off + common_extent
       dst_off = dst_off + common_extent
       ! We've used up another block.
       n_src_dsts(2,current_src) = n_src_dsts(2,current_src) + 1
       ! Check if we've used up the whole source block.
       DO WHILE (s_src .EQ. 0 .AND. current_src .LE. n_src)
          current_src = current_src + 1
          src_off = 1
          IF (current_src .LE. n_src) THEN
             s_src = src_sizes (current_src)
             n_src_dsts(:, current_src) = (/ current_int+1, 0 /)
          ENDIF
       END DO
       DO WHILE (s_dst .EQ. 0 .AND. current_dst .LE. n_dst)
          current_dst = current_dst + 1
          dst_off = 1
          IF (current_dst .LE. n_dst) s_dst = dst_sizes (current_dst)
       END DO
       current_int = current_int + 1
    ENDDO
    IF (current_src .LT. n_src) &
         n_src_dsts(:, current_src+1:n_src) = -7
    numints = current_int - 1
  END SUBROUTINE dbcsr_reblocking_targets



! **************************************************************************************************
!> \brief Distributes elements into blocks and into bins
!>
!> \param[out] block_distribution       block distribution to bins
!> \param[out] block_size       sizes of blocks
!> \param[in] nelements number of elements to bin
!> \param[in] nbins             number of bins
!> \par Term clarification
!>      An example: blocks are atom blocks and bins are process rows/colums.
! **************************************************************************************************
  SUBROUTINE create_bl_distribution (block_distribution,&
       block_size, nelements, nbins)
      INTEGER, DIMENSION(:), INTENT(OUT), POINTER        :: block_distribution, block_size
      INTEGER, INTENT(IN)                                :: nelements, nbins

      CHARACTER(len=*), PARAMETER :: routineN = 'create_bl_distribution', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bin, blk_layer, element_stack, els, &
                                                            estimated_blocks, max_blocks_per_bin, &
                                                            nblks, nblocks, stat
      INTEGER, DIMENSION(:), POINTER                     :: blk_dist, blk_sizes

!   ---------------------------------------------------------------------------

    NULLIFY (block_distribution)
    NULLIFY (block_size)
    ! Define the sizes on which we build the distribution.
    IF (nelements .GT. 0) THEN

       nblocks = CEILING(REAL(nelements)/REAL(max_elements_per_block))
       max_blocks_per_bin = CEILING(REAL(nblocks)/REAL(nbins))

       IF (debug_mod) THEN
          WRITE(*,'(1X,A,1X,A,I7,A,I7,A)')routineP,"For",nelements,&
               " elements and",nbins," bins"
          WRITE(*,'(1X,A,1X,A,I7,A)')routineP,"There are",&
               max_elements_per_block," max elements per block"
          WRITE(*,'(1X,A,1X,A,I7,A)')routineP,"There are",&
               nblocks," blocks"
          WRITE(*,'(1X,A,1X,A,I7,A)')routineP,"There are",&
               max_blocks_per_bin," max blocks/bin"
       ENDIF

       estimated_blocks = max_blocks_per_bin * nbins
       ALLOCATE (blk_dist (estimated_blocks), stat=stat)
       IF(stat /= 0)&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1275,"blk_dist")
       ALLOCATE (blk_sizes (estimated_blocks), stat=stat)
       IF(stat /= 0)&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1278,"blk_sizes")
       element_stack = 0
       nblks = 0
       DO blk_layer = 1, max_blocks_per_bin
          DO bin = 0, nbins-1
             els = MIN (max_elements_per_block, nelements-element_stack)
             IF (els .GT. 0) THEN
                element_stack = element_stack + els
                nblks = nblks + 1
                blk_dist(nblks) = bin
                blk_sizes(nblks) = els
                IF (debug_mod) WRITE(*,'(1X,A,I5,A,I5,A,I5)')routineP//" Assigning",&
                     els," elements as block",nblks," to bin",bin
             ENDIF
          ENDDO
       ENDDO
       ! Create the output arrays.
       IF (nblks .EQ. estimated_blocks) THEN
          block_distribution => blk_dist
          block_size => blk_sizes
       ELSE
          ALLOCATE (block_distribution (nblks), stat=stat)
          IF(stat /= 0)&
             CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1301,"blk_dist")
          block_distribution(:)=blk_dist(1:nblks)
          DEALLOCATE (blk_dist)
          ALLOCATE (block_size (nblks), stat=stat)
          IF(stat /= 0)&
             CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1306,"blk_sizes")
          block_size(:)=blk_sizes(1:nblks)
          DEALLOCATE(blk_sizes)
       ENDIF
    ELSE
       ALLOCATE (block_distribution (0), stat=stat)
       IF(stat /= 0)&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1313,"blk_dist")
       ALLOCATE (block_size (0), stat=stat)
       IF(stat /= 0)&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1316,"blk_sizes")
    ENDIF
1579 FORMAT (I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5)
    IF (debug_mod) THEN
       WRITE(*,'(1X,A,A)')routineP//" Distribution"
       WRITE(*,1579) block_distribution(:)
       WRITE(*,'(1X,A,A)')routineP//" Sizes"
       WRITE(*,1579) block_size(:)
    ENDIF
  END SUBROUTINE create_bl_distribution

! **************************************************************************************************
!> \brief Creates a new distribution for the right matrix in a matrix
!>        multiplication with unrotated grid.
!> \param[out] dist_right     new distribution for the right matrix
!> \param[in] dist_left       the distribution of the left matrix
!> \param[in] ncolumns        number of columns in right matrix
!> \param[out] right_col_blk_sizes      sizes of blocks in the created column
!> \par The new row distribution for the right matrix is the same as the row
!>      distribution of the left matrix, while the column distribution is
!>      created so that it is appropriate to the parallel environment.
! **************************************************************************************************
  SUBROUTINE dbcsr_create_dist_r_unrot(dist_right, dist_left, ncolumns,&
       right_col_blk_sizes)
      TYPE(dbcsr_distribution_obj), INTENT(OUT)          :: dist_right
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist_left
      INTEGER, INTENT(IN)                                :: ncolumns
      INTEGER, DIMENSION(:), INTENT(OUT), POINTER        :: right_col_blk_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_create_dist_r_unrot', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: multiplicity, new_npcols, new_nprows, &
                                                            nimages
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: tmp_images
      INTEGER, DIMENSION(:), POINTER                     :: right_col_dist, right_row_dist
      TYPE(dbcsr_mp_obj)                                 :: right_mp_env

!   ---------------------------------------------------------------------------

    right_mp_env = dbcsr_distribution_mp (dist_left)
    ! Create the column distribution
    CALL create_bl_distribution (right_col_dist,&
         right_col_blk_sizes, ncolumns, dbcsr_mp_npcols (right_mp_env))
    ! Create an even row distribution.
    ALLOCATE (right_row_dist(dbcsr_distribution_ncols (dist_left)))
    ALLOCATE (tmp_images(dbcsr_distribution_ncols (dist_left)))
    new_nprows = dbcsr_mp_nprows (right_mp_env)
    new_npcols = dbcsr_mp_npcols (right_mp_env)
    nimages = lcm (new_nprows, new_npcols) / new_nprows
    multiplicity = new_nprows / gcd (new_nprows, new_npcols)
    CALL rebin_distribution (right_row_dist, tmp_images,& 
         dbcsr_distribution_col_dist (dist_left),&
         new_nprows, multiplicity, nimages)
    CALL dbcsr_distribution_new (dist_right, right_mp_env,&
         right_row_dist, right_col_dist, &
         reuse_arrays=.TRUE.)
    DEALLOCATE(tmp_images)
  END SUBROUTINE dbcsr_create_dist_r_unrot

! **************************************************************************************************
!> \brief Creates a block-cyclic compatible distribution
!>
!>        All blocks in a dimension, except for possibly the last
!>        block, have the same size.
!> \param[out] dist           the elemental distribution
!> \param[in] nrows           number of full rows
!> \param[in] ncolumns        number of full columns
!> \param[in] nrow_block      size of row blocks
!> \param[in] ncol_block      size of column blocks
!> \param[in] mp_env          multiprocess environment
!> \param[out] row_blk_sizes  row block sizes
!> \param[out] col_blk_sizes  column block sizes
! **************************************************************************************************
  SUBROUTINE dbcsr_create_dist_block_cyclic (dist, nrows, ncolumns,&
       nrow_block, ncol_block, mp_env, row_blk_sizes, col_blk_sizes)
      TYPE(dbcsr_distribution_obj), INTENT(OUT)          :: dist
      INTEGER, INTENT(IN)                                :: nrows, ncolumns, nrow_block, ncol_block
      TYPE(dbcsr_mp_obj), INTENT(IN)                     :: mp_env
      INTEGER, DIMENSION(:), INTENT(OUT), POINTER        :: row_blk_sizes, col_blk_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_create_dist_block_cyclic', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nblkcols, nblkrows, npcols, nprows, &
                                                            pdim, sz
      INTEGER, DIMENSION(:), POINTER                     :: cd_data, rd_data

!   ---------------------------------------------------------------------------
! Row sizes

    IF(nrow_block.EQ.0) THEN
       nblkrows = 0
       sz = 0
    ELSE
       nblkrows = nrows / nrow_block
       sz = MOD(nrows, nrow_block)
    ENDIF
    IF (sz .GT. 0) nblkrows = nblkrows + 1
    ALLOCATE (row_blk_sizes(nblkrows), rd_data(nblkrows))
    row_blk_sizes = nrow_block
    IF (sz .NE. 0) row_blk_sizes(nblkrows) = sz

    ! Column sizes
    IF(ncol_block.EQ.0) THEN
       nblkcols = 0
       sz = 0
    ELSE
       nblkcols = ncolumns / ncol_block
       sz = MOD(ncolumns, ncol_block)
    ENDIF
    IF (sz .GT. 0) nblkcols = nblkcols + 1
    ALLOCATE (col_blk_sizes(nblkcols), cd_data(nblkcols))
    col_blk_sizes = ncol_block
    IF (sz .NE. 0) col_blk_sizes(nblkcols) = sz
    !
    IF(debug_mod) THEN
       WRITE(*,*)routineN//" nrows,nrow_block,nblkrows=",&
            nrows,nrow_block,nblkrows
       WRITE(*,*)routineN//" ncols,ncol_block,nblkcols=",&
            ncolumns,ncol_block,nblkcols
    ENDIF
    ! Calculate process row distribution
    nprows = dbcsr_mp_nprows (mp_env)
    DO pdim = 0 , MIN(nprows-1,nblkrows-1)
       rd_data(1+pdim:nblkrows:nprows) = pdim
    END DO
    ! Calculate process column distribution
    npcols = dbcsr_mp_npcols (mp_env)
    DO pdim = 0 , MIN(npcols-1, nblkcols-1)
       cd_data(1+pdim:nblkcols:npcols) = pdim
    END DO
    !
    IF(debug_mod) THEN
       WRITE(*,*)routineN//" row_dist",&
            rd_data
       WRITE(*,*)routineN//" col_dist",&
            cd_data
    ENDIF
    !
    CALL dbcsr_distribution_new (dist, mp_env,&
         rd_data, cd_data, reuse_arrays=.TRUE.)
    !
  END SUBROUTINE dbcsr_create_dist_block_cyclic

! **************************************************************************************************
!> \brief Matches row/block sizes and offsets to a given distribution
!> \param old_sizes ...
!> \param mapping ...
!> \param nel_new ...
!> \param new_sizes ...
!> \note Used for making matrices dense/undense
! **************************************************************************************************
  SUBROUTINE make_sizes_dense (old_sizes, mapping, nel_new, new_sizes)
      TYPE(array_i1d_obj), INTENT(IN)                    :: old_sizes, mapping
      INTEGER, INTENT(IN)                                :: nel_new
      TYPE(array_i1d_obj), INTENT(OUT)                   :: new_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'make_sizes_dense', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: el, nel_old
      INTEGER, DIMENSION(:), POINTER                     :: map, new_s, old_s

!   ---------------------------------------------------------------------------

    map => array_data (mapping)
    old_s => array_data (old_sizes)
    nel_old = array_size (old_sizes)
    ALLOCATE (new_s (nel_new))
    new_s(:) = 0
    DO el = 1, nel_old
       new_s(map(el)) = new_s(map(el)) + old_s(el)
    ENDDO
    CALL array_new (new_sizes, new_s, gift=.TRUE.)
  END SUBROUTINE make_sizes_dense


! **************************************************************************************************
!> \brief Makes a 1-D distribution dense.
!> \param[in] nbins Number of bins in the main distribution
!> \param[in] old_dist         Current distribution
!> \param[out] dense_dist      Dense distribution
!> \param[out] dist_map        Map from current to dense distribution
!> \param[in] nsubdist         Number of bins in the subdistribution
!> \param[in] subdist          Subdistribution
! **************************************************************************************************
  SUBROUTINE dbcsr_make_1dist_dense (nbins, old_dist, dense_dist, dist_map, nsubdist, subdist)
      INTEGER, INTENT(IN)                                :: nbins
      TYPE(array_i1d_obj), INTENT(IN)                    :: old_dist
      TYPE(array_i1d_obj), INTENT(OUT)                   :: dense_dist, dist_map
      INTEGER, INTENT(IN)                                :: nsubdist
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: subdist

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_1dist_dense', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: b, i, n_new_bins
      INTEGER, DIMENSION(:), POINTER                     :: dense, map, old_d

!   ---------------------------------------------------------------------------

    IF (nsubdist .EQ. 0) THEN
       n_new_bins = nbins
    ELSE
       n_new_bins = nbins * nsubdist
    ENDIF
    old_d => array_data (old_dist)
    ALLOCATE (dense (n_new_bins))
    ALLOCATE (map (array_size (old_dist)))
    !
    IF (nsubdist .EQ. 0) THEN
       dense(:) = (/ (b, b = 0, n_new_bins-1) /)
       map(:) = old_d(:)+1
    ELSE
       DO i = 1, nbins
          dense((i-1)*nsubdist+1 : (i)*nsubdist) = i-1
       ENDDO
       map(:) = old_d(:)*nsubdist + subdist(:) + 1
    ENDIF
    !
    CALL array_new (dense_dist, dense, gift=.TRUE.)
    CALL array_new (dist_map, map, gift=.TRUE.)
  END SUBROUTINE dbcsr_make_1dist_dense


! **************************************************************************************************
!> \brief Prepares distributions for making dense matrices.
!> \param product_dist ...
!> \param left_rdist ...
!> \param right_rdist ...
!> \param dense_product_dist ...
!> \param dense_left_rdist ...
!> \param dense_right_rdist ...
!> \param partial ...
!> \param m_map ...
!> \param k_vmap ...
!> \param n_map ...
!> \param old_m_sizes ...
! **************************************************************************************************
  SUBROUTINE dbcsr_make_dists_dense (product_dist, left_rdist, right_rdist,&
       dense_product_dist, dense_left_rdist, dense_right_rdist,&
       partial,&
       m_map, k_vmap, n_map,&
       old_m_sizes)
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: product_dist
      TYPE(dbcsr_imagedistribution_obj), INTENT(IN)      :: left_rdist, right_rdist
      TYPE(dbcsr_distribution_obj), INTENT(OUT)          :: dense_product_dist
      TYPE(dbcsr_imagedistribution_obj), INTENT(OUT)     :: dense_left_rdist, dense_right_rdist
      LOGICAL, INTENT(IN)                                :: partial
      TYPE(array_i1d_obj), INTENT(OUT)                   :: m_map, k_vmap, n_map
      TYPE(array_i1d_obj), INTENT(IN)                    :: old_m_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_dists_dense', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, i, j, k_nbins, m_nbins, &
                                                            n_nbins, nthreads
      INTEGER, DIMENSION(:), POINTER                     :: tdist
      TYPE(array_i1d_obj)                                :: new_k_idist, new_k_pdist, new_k_vdist, &
                                                            new_m_dist, new_m_sizes, new_n_dist, &
                                                            old_k_vdist, old_m_dist, old_n_dist
      TYPE(dbcsr_distribution_obj)                       :: dense_left_dist, dense_right_dist

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, error_handle)
    !
    IF(.NOT.dbcsr_distribution_has_threads (product_dist))&
       CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1585,"Product distribution must have threads.")
    tdist => array_data (dbcsr_distribution_thread_dist (product_dist))
    old_m_dist = product_dist%d%row_dist_block
    old_n_dist = product_dist%d%col_dist_block
    old_k_vdist = right_rdist%i%vrow_dist
    m_nbins = dbcsr_mp_nprows (product_dist%d%mp_env)
    n_nbins = dbcsr_mp_npcols (product_dist%d%mp_env)
    k_nbins = dbcsr_mp_nprows (right_rdist%i%main%d%mp_env) * right_rdist%i%row_decimation
    IF(.NOT.array_equality (old_k_vdist, left_rdist%i%vcol_dist))&
       CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1594,"k distribution mismtach")
    nthreads = product_dist%d%num_threads
    !
    IF (partial) THEN
       new_m_dist = old_m_dist
       CALL array_hold (new_m_dist)
       new_n_dist = old_n_dist
       CALL array_hold (new_n_dist)
       dense_product_dist = product_dist
       CALL dbcsr_distribution_hold (dense_product_dist)
       CALL array_new (m_map, (/ (i, i=1, array_size(new_m_dist)) /), lb=1)
       CALL array_new (n_map, (/ (i, i=1, array_size(new_n_dist)) /), lb=1)
    ELSE
       CALL dbcsr_make_1dist_dense (m_nbins, old_m_dist, new_m_dist, m_map,&
            nthreads, tdist)
       CALL dbcsr_make_1dist_dense (n_nbins, old_n_dist, new_n_dist, n_map, 0)
       CALL dbcsr_distribution_new (dense_product_dist, product_dist%d%mp_env,&
            array_data(new_m_dist), array_data(new_n_dist))
       CALL make_sizes_dense (old_m_sizes, m_map, array_size(new_m_dist), new_m_sizes)
       CALL dbcsr_distribution_make_threads (dense_product_dist,&
            array_data(new_m_sizes))
       CALL array_release (new_m_sizes)
       tdist => array_data (dbcsr_distribution_thread_dist (dense_product_dist))
       ! Resets the thread distribution to be in-order.
       DO i = 1, m_nbins
          tdist((i-1)*nthreads+1 : (i)*nthreads) = (/ (j, j=0,nthreads-1) /)
       ENDDO
    ENDIF
    !
    CALL dbcsr_make_1dist_dense (k_nbins, old_k_vdist, new_k_vdist, k_vmap, 0)
    CALL v_to_p_i_dist_o (new_k_vdist,&
         left_rdist%i%col_decimation, new_k_pdist, new_k_idist)
    ! Left
    CALL dbcsr_distribution_new (dense_left_dist, left_rdist%i%main%d%mp_env,&
         array_data(new_m_dist), array_data(new_k_pdist))
    CALL dbcsr_distribution_no_threads (dense_left_dist)
    dense_left_dist%d%thread_dist = dbcsr_distribution_thread_dist (dense_product_dist)
    CALL array_hold (dense_left_dist%d%thread_dist)
    dense_left_dist%d%has_thread_dist = .TRUE.
    CALL dbcsr_new_image_dist (dense_left_rdist, dense_left_dist, left_rdist)
    CALL dbcsr_distribution_release (dense_left_dist)
    CALL array_new (dense_left_rdist%i%row_image,&
         (/ (1, i=1, array_size (new_m_dist)) /), lb=1)
    dense_left_rdist%i%col_image = new_k_idist
    CALL array_hold (new_k_idist)
    dense_left_rdist%i%vrow_dist = new_m_dist
    CALL array_hold (new_m_dist)
    dense_left_rdist%i%vcol_dist = new_k_vdist
    CALL array_hold (new_k_vdist)
    !
    CALL array_release (new_k_pdist)
    CALL array_release (new_k_idist)
    ! Right
    CALL v_to_p_i_dist_o (new_k_vdist,&
         right_rdist%i%row_decimation, new_k_pdist, new_k_idist)
    CALL dbcsr_distribution_new (dense_right_dist, right_rdist%i%main%d%mp_env,&
         array_data(new_k_pdist), array_data(new_n_dist))
    CALL dbcsr_new_image_dist (dense_right_rdist, dense_right_dist, right_rdist)
    CALL dbcsr_distribution_release (dense_right_dist)
    CALL array_new (dense_right_rdist%i%col_image,&
         (/ (1, i=1, array_size (new_n_dist)) /), lb=1)
    dense_right_rdist%i%row_image = new_k_idist
    CALL array_hold (new_k_idist)
    dense_right_rdist%i%vrow_dist = new_k_vdist
    CALL array_hold (new_k_vdist)
    dense_right_rdist%i%vcol_dist = new_n_dist
    CALL array_hold (new_n_dist)
    !
    CALL array_release (new_k_idist)
    CALL array_release (new_k_pdist)
    CALL array_release (new_m_dist)
    CALL array_release (new_n_dist)
    CALL array_release (new_k_vdist)
    !
    CALL timestop(error_handle)
  END SUBROUTINE dbcsr_make_dists_dense


! **************************************************************************************************
!> \brief Finds the block that has the given column.
!>
!>        If the block having the queried column is found, the blk parameter
!>        is set to this block number and the found parameter is true.
!>        Otherwise found is false and the block number is invalid.
!> \param[in] find_col        column to find
!> \param[in] frst_blk        first block number in row
!> \param[in] last_blk        last block number in row
!> \param[in] col_i           col indices
!> \param[in] blk_p           block pointers
!> \param[out] blk            block number with searched-for column
!> \param[out] found          flag specified whether a block that has the
!>                            correct column was found
!> \par Index validity
!>      The blk_p array of block pointers is a required parameter to enable
!>      the detection of deleted blocks.
! **************************************************************************************************
  PURE SUBROUTINE dbcsr_find_column(find_col,frst_blk,last_blk,col_i,blk_p,&
       blk,found)
      INTEGER, INTENT(IN)                                :: find_col, frst_blk, last_blk
      INTEGER, DIMENSION(:), INTENT(IN)                  :: col_i, blk_p
      INTEGER, INTENT(OUT)                               :: blk
      LOGICAL, INTENT(OUT)                               :: found

      CALL ordered_search(col_i,find_col,blk,found,frst_blk,last_blk)
      IF (found) THEN
         found =  blk_p(blk) .NE. 0
      ENDIF

  END SUBROUTINE dbcsr_find_column

! **************************************************************************************************
!> \brief Finds the local virtual elements
!>
!> All elements are mapped at once.  Therefore an entry in the
!> resulting array can be used as a lookup index for any of the local
!> element arrays.  The distribution itself tells into which array to
!> look.
!> \param[out] local_elements local virtual elements
!> \param[in] bin_distribution           distribution of elements to bins
!> \param[in] nbins                      number of bins
! **************************************************************************************************
  SUBROUTINE find_all_local_elements (local_elements, &
       bin_distribution, nbins)
      INTEGER, INTENT(IN)                                :: nbins
      INTEGER, DIMENSION(:), INTENT(IN)                  :: bin_distribution
      TYPE(array_i1d_obj), DIMENSION(0:nbins-1), &
         INTENT(INOUT)                                   :: local_elements

      INTEGER                                            :: bin, el
      INTEGER, DIMENSION(0:nbins-1)                      :: nlve

    nlve(:) = 0
    DO el = 1, SIZE (bin_distribution)
       bin = bin_distribution (el)
       nlve(bin) = nlve(bin) + 1
       local_elements(bin)%low%data(nlve(bin)) = el
    ENDDO
  END SUBROUTINE find_all_local_elements

! **************************************************************************************************
!> \brief Points to local virtual elements.
!>
!> All elements are mapped at once.  Therefore an entry in the
!> resulting array can be used as a lookup index for any of the local
!> element arrays.  The distribution itself tells into which array to
!> look.
!> \param[out] global_elements   enumerated local elements
!> \param[in] bin_distribution           distribution of elements to bins
!> \param[in] nbins                      number of bins
! **************************************************************************************************
  PURE SUBROUTINE map_all_local_elements (global_elements, &
       bin_distribution, nbins)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: global_elements
      INTEGER, DIMENSION(:), INTENT(IN)                  :: bin_distribution
      INTEGER, INTENT(IN)                                :: nbins

      INTEGER                                            :: bin, el
      INTEGER, DIMENSION(0:nbins-1)                      :: nlve

    nlve(:) = 0
    DO el = 1, SIZE (bin_distribution)
       bin = bin_distribution (el)
       nlve(bin) = nlve(bin) + 1
       global_elements(el) = nlve(bin)
    ENDDO
  END SUBROUTINE map_all_local_elements


! **************************************************************************************************
!> \brief Determines mapping from local to global rows
!> \param[in] dist              mapping for this distribution
!> \param[in] local_rows        local elements for specified row
!> \param[in] local_prow find local elemnts for this local row
! **************************************************************************************************
  SUBROUTINE dbcsr_get_local_rows (dist, local_rows, local_prow)
      TYPE(dbcsr_distribution_obj), INTENT(INOUT)        :: dist
      TYPE(array_i1d_obj), INTENT(OUT)                   :: local_rows
      INTEGER, INTENT(IN)                                :: local_prow

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_local_rows', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: el, error_handle, nprows, prow
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: nle
      INTEGER, DIMENSION(:), POINTER                     :: itmp, row_dist

    CALL timeset(routineN, error_handle)
    ! If the current local row mappings do not exist, create them.
    IF (.NOT. dist%d%has_other_l_rows) THEN
       dist%d%has_other_l_rows = .TRUE.
       nprows = dbcsr_mp_nprows (dbcsr_distribution_mp (dist))
       ALLOCATE (dist%d%other_l_rows(0:dbcsr_mp_nprows (dist%d%mp_env)-1))
       ALLOCATE (nle (0:nprows-1))
       row_dist => dbcsr_distribution_row_dist (dist)
       ! Count the number of local elements per row.
       nle(:) = 0
       DO el = 1, SIZE (row_dist)
          prow = row_dist(el)
          nle(prow) = nle(prow) + 1
       ENDDO
       DO prow = 0, nprows-1
          ALLOCATE(itmp(nle(prow)))
          itmp=0
          CALL array_new (dist%d%other_l_rows(prow),&
               itmp, lb=1)
          DEALLOCATE(itmp)
       ENDDO
       DEALLOCATE (nle)
       CALL find_all_local_elements (dist%d%other_l_rows, row_dist, nprows)
    ELSE
       IF (careful_mod .AND. .NOT.ASSOCIATED (dist%d%other_l_rows))&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1805,"Local rows mapping does not exist.")
    ENDIF
    local_rows = dist%d%other_l_rows(local_prow)
    CALL timestop(error_handle)
  END SUBROUTINE dbcsr_get_local_rows

! **************************************************************************************************
!> \brief Determines mapping from local to global columnss
!> \param[in] dist              mapping for this distribution
!> \param[in] local_cols        local elements for specified column
!> \param[in] local_pcol         find local elemnts for this local column 
! **************************************************************************************************
  SUBROUTINE dbcsr_get_local_cols (dist, local_cols, local_pcol)
      TYPE(dbcsr_distribution_obj), INTENT(INOUT)        :: dist
      TYPE(array_i1d_obj), INTENT(OUT)                   :: local_cols
      INTEGER, INTENT(IN)                                :: local_pcol

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_local_cols', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: el, error_handle, npcols, pcol
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: nle
      INTEGER, DIMENSION(:), POINTER                     :: col_dist, itmp

    CALL timeset(routineN, error_handle)
    ! If the current local col mappings do not exist, create them.
    IF (.NOT. dist%d%has_other_l_cols) THEN
       dist%d%has_other_l_cols = .TRUE.
       npcols = dbcsr_mp_npcols (dbcsr_distribution_mp (dist))
       ALLOCATE (dist%d%other_l_cols(0:dbcsr_mp_npcols (dist%d%mp_env)-1))
       ALLOCATE (nle (0:npcols-1))
       col_dist => dbcsr_distribution_col_dist (dist)
       ! Count the number of local elements per col.
       nle(:) = 0
       DO el = 1, SIZE (col_dist)
          pcol = col_dist(el)
          nle(pcol) = nle(pcol) + 1
       ENDDO
       DO pcol = 0, npcols-1
          ALLOCATE(itmp(nle(pcol)))
          itmp=0
          CALL array_new (dist%d%other_l_cols(pcol),&
               itmp, lb=1)
          DEALLOCATE(itmp)
       ENDDO
       DEALLOCATE (nle)
       CALL find_all_local_elements (dist%d%other_l_cols, col_dist, npcols)
    ELSE
       IF (careful_mod.AND..NOT.ASSOCIATED (dist%d%other_l_cols))&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1854,"Local columns mapping does not exist.")
    ENDIF
    local_cols = dist%d%other_l_cols(local_pcol)
    CALL timestop(error_handle)
  END SUBROUTINE dbcsr_get_local_cols


! **************************************************************************************************
!> \brief Determines mapping from global to local rows
!> \param[in] dist              mapping for this distribution
!> \param[in] row_map           mapping to local rows
! **************************************************************************************************
  SUBROUTINE dbcsr_get_global_row_map (dist, row_map)
      TYPE(dbcsr_distribution_obj), INTENT(INOUT)        :: dist
      TYPE(array_i1d_obj), INTENT(OUT)                   :: row_map

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_global_row_map', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, nprows
      INTEGER, DIMENSION(:), POINTER                     :: rmap, row_dist

    CALL timeset(routineN, error_handle)
    ! If the current local row mappings do not exist, create them.
    IF (.NOT. dist%d%has_global_row_map) THEN
       row_dist => dbcsr_distribution_row_dist (dist)
       ALLOCATE (rmap (SIZE (row_dist)))
       nprows = dbcsr_mp_nprows (dbcsr_distribution_mp (dist))
       CALL map_all_local_elements (rmap, row_dist, nprows)
       CALL array_new (dist%d%global_row_map, rmap, gift=.TRUE.)
       dist%d%has_global_row_map = .TRUE.
    ELSE
       IF (careful_mod.AND..NOT.array_exists (dist%d%global_row_map))&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1887,"Row map does not exist.")
    ENDIF
    row_map = dist%d%global_row_map
    CALL timestop(error_handle)
  END SUBROUTINE dbcsr_get_global_row_map


! **************************************************************************************************
!> \brief Determines mapping from global to local columns
!> \param[in] dist              mapping for this distribution
!> \param[in] col_map           mapping to local columns
! **************************************************************************************************
  SUBROUTINE dbcsr_get_global_col_map (dist, col_map)
      TYPE(dbcsr_distribution_obj), INTENT(INOUT)        :: dist
      TYPE(array_i1d_obj), INTENT(OUT)                   :: col_map

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_global_col_map', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, npcols
      INTEGER, DIMENSION(:), POINTER                     :: cmap, col_dist

    CALL timeset(routineN, error_handle)
    ! If the current local col mappings do not exist, create them.
    IF (.NOT. dist%d%has_global_col_map) THEN
       col_dist => dbcsr_distribution_col_dist (dist)
       ALLOCATE (cmap (SIZE (col_dist)))
       npcols = dbcsr_mp_npcols (dbcsr_distribution_mp (dist))
       CALL map_all_local_elements (cmap, col_dist, npcols)
       CALL array_new (dist%d%global_col_map, cmap, gift=.TRUE.)
       dist%d%has_global_col_map = .TRUE.
    ELSE
       IF (careful_mod.AND..NOT.array_exists (dist%d%global_col_map))&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",1920,"Column map does not exist.")
    ENDIF
    col_map = dist%d%global_col_map
    CALL timestop(error_handle)
  END SUBROUTINE dbcsr_get_global_col_map


! **************************************************************************************************
!> \brief Resets local rows, columns to the correct arrays and values.
!> \param matrix ...
! **************************************************************************************************
  SUBROUTINE dbcsr_reset_locals (matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      TYPE(dbcsr_distribution_obj)                       :: dist

!   ---------------------------------------------------------------------------

    dist = dbcsr_distribution (matrix)
    CALL dbcsr_release_locals (matrix)
    ! Rows
    IF (dbg) &
         WRITE(*,*)"reset local rows for ", TRIM(matrix%m%name),&
         matrix%m%nblkrows_local, "prow", matrix%m%index(dbcsr_slot_home_prow),&
         dbcsr_mp_myprow (dbcsr_distribution_mp (matrix%m%dist))
    CALL dbcsr_get_local_rows (dist, matrix%m%local_rows,&
         matrix%m%index(dbcsr_slot_home_prow))
    CALL array_hold (matrix%m%local_rows)
    IF (dbg) WRITE(*,*)"local rows", matrix%m%local_rows%low%data
    matrix%m%nblkrows_local = array_size (matrix%m%local_rows)
    CALL dbcsr_get_global_row_map (dist, matrix%m%global_rows)
    CALL array_hold (matrix%m%global_rows)
    matrix%m%has_local_rows = .TRUE.
    matrix%m%has_global_rows = .TRUE.
    ! Columns
    IF (dbg) &
         WRITE(*,*)"reset local cols for ", TRIM(matrix%m%name),&
         matrix%m%nblkcols_local, "pcol", matrix%m%index(dbcsr_slot_home_pcol),&
         dbcsr_mp_mypcol (dbcsr_distribution_mp (matrix%m%dist))
    CALL dbcsr_get_local_cols (dist, matrix%m%local_cols,&
         matrix%m%index(dbcsr_slot_home_pcol))
    CALL array_hold (matrix%m%local_cols)
    IF (dbg) WRITE(*,*)"local cols", matrix%m%local_cols%low%data
    matrix%m%nblkcols_local = array_size (matrix%m%local_cols)
    CALL dbcsr_get_global_col_map (dist, matrix%m%global_cols)
    CALL array_hold (matrix%m%global_cols)
    matrix%m%has_local_cols = .TRUE.
    matrix%m%has_global_cols = .TRUE.
    !
  END SUBROUTINE dbcsr_reset_locals

! **************************************************************************************************
!> \brief Determines mapping from local to global virtual process rows
!> \param[in] imgdist           image distribution
!> \param[out] local_vrows       local rows 
!> \param[in] local_vprow       the local virtual process row
! **************************************************************************************************
  SUBROUTINE dbcsr_get_local_vrows (imgdist, local_vrows, local_vprow)
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist
      TYPE(array_i1d_obj), INTENT(OUT)                   :: local_vrows
      INTEGER, INTENT(IN)                                :: local_vprow

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_local_vrows', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: el, error_handle, nvprows, vprow
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: itmp, nle
      INTEGER, DIMENSION(:), POINTER                     :: vrow_dist

    IF (careful_mod) CALL timeset(routineN, error_handle)
    ! If the current local row mappings do not exist, create them.
    IF (.NOT. imgdist%i%has_other_vl_rows) THEN
       imgdist%i%has_other_vl_rows = .TRUE.
       nvprows = dbcsr_mp_nprows (dbcsr_distribution_mp (imgdist%i%main))&
               * imgdist%i%row_decimation
       ALLOCATE (imgdist%i%other_vl_rows(0:nvprows-1))
       ALLOCATE (nle (0:nvprows-1))
       vrow_dist => array_data (imgdist%i%vrow_dist)
       ! Count the number of local elements per row.
       nle(:) = 0
       DO el = 1, SIZE (vrow_dist)
          vprow = vrow_dist(el)
          nle(vprow) = nle(vprow) + 1
       ENDDO
       DO vprow = 0, nvprows-1
          ALLOCATE(itmp(nle(vprow)))
          itmp=0
          CALL array_new (imgdist%i%other_vl_rows(vprow),&
               itmp, lb=1)
          DEALLOCATE(itmp)
       ENDDO
       DEALLOCATE (nle)
       CALL find_all_local_elements (imgdist%i%other_vl_rows, vrow_dist, nvprows)
    ELSE
       IF (careful_mod.AND. .NOT.ASSOCIATED (imgdist%i%other_vl_rows))&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",2017,"Local rows mapping does not exist.")
    ENDIF
    local_vrows = imgdist%i%other_vl_rows(local_vprow)
    IF (careful_mod) CALL timestop(error_handle)
  END SUBROUTINE dbcsr_get_local_vrows

! **************************************************************************************************
!> \brief Determines mapping from local to global virtual process columnss
!> \param[in] imgdist           image distribution
!> \param[out] local_vcols       local columnss 
!> \param[in] local_vpcol       the local virtual process column
! **************************************************************************************************
  SUBROUTINE dbcsr_get_local_vcols (imgdist, local_vcols, local_vpcol)
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist
      TYPE(array_i1d_obj), INTENT(OUT)                   :: local_vcols
      INTEGER, INTENT(IN)                                :: local_vpcol

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_local_vcols', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: el, error_handle, nvpcols, vpcol
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: nle
      INTEGER, DIMENSION(:), POINTER                     :: itmp, vcol_dist

    IF (careful_mod) CALL timeset(routineN, error_handle)
    ! If the current local col mappings do not exist, create them.
    IF (.NOT. imgdist%i%has_other_vl_cols) THEN
       imgdist%i%has_other_vl_cols = .TRUE.
       nvpcols = dbcsr_mp_npcols (dbcsr_distribution_mp (imgdist%i%main))&
               * imgdist%i%col_decimation
       ALLOCATE (imgdist%i%other_vl_cols(0:nvpcols-1))
       ALLOCATE (nle (0:nvpcols-1))
       vcol_dist => array_data (imgdist%i%vcol_dist)
       ! Count the number of local elements per col.
       nle(:) = 0
       DO el = 1, SIZE (vcol_dist)
          vpcol = vcol_dist(el)
          nle(vpcol) = nle(vpcol) + 1
       ENDDO
       DO vpcol = 0, nvpcols-1
          ALLOCATE(itmp(nle(vpcol)))
          itmp=0
          CALL array_new (imgdist%i%other_vl_cols(vpcol),&
               itmp, lb=1)
          DEALLOCATE(itmp)
       ENDDO
       DEALLOCATE (nle)
       CALL find_all_local_elements (imgdist%i%other_vl_cols, vcol_dist, nvpcols)
    ELSE
       IF (careful_mod .AND. .NOT.ASSOCIATED (imgdist%i%other_vl_cols))&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",2067,"Local cols mapping does not exist.")
    ENDIF
    local_vcols = imgdist%i%other_vl_cols(local_vpcol)
    IF (careful_mod) CALL timestop(error_handle)
  END SUBROUTINE dbcsr_get_local_vcols

! **************************************************************************************************
!> \brief Determines mapping from global to virtual local rows
!> \param[in] imgdist           mapping for this image distribution
!> \param[in] vrow_map           mapping to local rows
! **************************************************************************************************
  SUBROUTINE dbcsr_get_global_vrow_map (imgdist, vrow_map)
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist
      TYPE(array_i1d_obj), INTENT(OUT)                   :: vrow_map

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_global_vrow_map', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, nvprows
      INTEGER, DIMENSION(:), POINTER                     :: rmap, vrow_dist

    IF (careful_mod) CALL timeset(routineN, error_handle)
    ! If the current local row mappings do not exist, create them.
    IF (.NOT. imgdist%i%has_global_vrow_map) THEN
       vrow_dist => array_data (imgdist%i%vrow_dist)
       ALLOCATE (rmap (SIZE (vrow_dist)))
       nvprows = dbcsr_mp_nprows (dbcsr_distribution_mp (imgdist%i%main))&
               * imgdist%i%row_decimation
       CALL map_all_local_elements (rmap, vrow_dist, nvprows)
       CALL array_new (imgdist%i%global_vrow_map, rmap, gift=.TRUE.)
       imgdist%i%has_global_vrow_map = .TRUE.
    ELSE
       IF (careful_mod.AND..NOT.array_exists (imgdist%i%global_vrow_map))&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",2100,"Row map does not exist.")
    ENDIF
    vrow_map = imgdist%i%global_vrow_map
    IF (careful_mod) CALL timestop(error_handle)
  END SUBROUTINE dbcsr_get_global_vrow_map

! **************************************************************************************************
!> \brief Determines mapping from global to virtual local columnss
!> \param[in] imgdist           mapping for this image distribution
!> \param[in] vcol_map           mapping to local columnss 
! **************************************************************************************************
  SUBROUTINE dbcsr_get_global_vcol_map (imgdist, vcol_map)
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist
      TYPE(array_i1d_obj), INTENT(OUT)                   :: vcol_map

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_global_vcol_map', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, nvpcols
      INTEGER, DIMENSION(:), POINTER                     :: rmap, vcol_dist

    IF (careful_mod) CALL timeset(routineN, error_handle)
    ! If the current local col mappings do not exist, create them.
    IF (.NOT. imgdist%i%has_global_vcol_map) THEN
       vcol_dist => array_data (imgdist%i%vcol_dist)
       ALLOCATE (rmap (SIZE (vcol_dist)))
       nvpcols = dbcsr_mp_npcols (dbcsr_distribution_mp (imgdist%i%main))&
               * imgdist%i%col_decimation
       CALL map_all_local_elements (rmap, vcol_dist, nvpcols)
       CALL array_new (imgdist%i%global_vcol_map, rmap, gift=.TRUE.)
       imgdist%i%has_global_vcol_map = .TRUE.
    ELSE
       IF (careful_mod.AND..NOT.array_exists (imgdist%i%global_vcol_map))&
          CALL cp__b("dbcsr/dist/dbcsr_dist_operations.F",2133,"Col map does not exist.")
    ENDIF
    vcol_map = imgdist%i%global_vcol_map
    IF (careful_mod) CALL timestop(error_handle)
  END SUBROUTINE dbcsr_get_global_vcol_map



! **************************************************************************************************
!> \brief Resets local rows, columns to the correct arrays and values
!>         for images.
!> \param matrix ...
!> \param imgdist ...
!> \param do_rows ...
! **************************************************************************************************
  SUBROUTINE dbcsr_reset_vlocals (matrix, imgdist, do_rows)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      TYPE(dbcsr_imagedistribution_obj), INTENT(INOUT)   :: imgdist
      LOGICAL, INTENT(IN), OPTIONAL                      :: do_rows

      LOGICAL                                            :: my_do_rows

!   ---------------------------------------------------------------------------

    CALL dbcsr_release_locals (matrix)
    my_do_rows = .TRUE.
    IF (PRESENT(do_rows)) my_do_rows = do_rows
    ! Rows
    IF (.NOT.PRESENT(do_rows) .OR. my_do_rows) THEN
       CALL dbcsr_get_local_vrows (imgdist, matrix%m%local_rows,&
            matrix%m%index(dbcsr_slot_home_vprow))
    ELSE
       matrix%m%local_rows = imgdist%i%main%d%local_rows
    ENDIF
    CALL array_hold (matrix%m%local_rows)
    matrix%m%has_local_rows = .TRUE.
    matrix%m%nblkrows_local = array_size (matrix%m%local_rows)
    matrix%m%index(dbcsr_slot_nblkrows_local) = array_size (matrix%m%local_rows)
    CALL dbcsr_get_global_vrow_map (imgdist, matrix%m%global_rows)
    CALL array_hold (matrix%m%global_rows)
    matrix%m%has_global_rows = .TRUE.
    ! Columns
    IF (.NOT.PRESENT(do_rows) .OR. .NOT.my_do_rows) THEN
       CALL dbcsr_get_local_vcols (imgdist, matrix%m%local_cols,&
            matrix%m%index(dbcsr_slot_home_vpcol))
    ELSE
       matrix%m%local_cols = imgdist%i%main%d%local_cols
    ENDIF
    CALL array_hold (matrix%m%local_cols)
    matrix%m%has_local_cols = .TRUE.
    matrix%m%nblkcols_local = array_size (matrix%m%local_cols)
    matrix%m%index(dbcsr_slot_nblkcols_local) = array_size (matrix%m%local_cols)
    CALL dbcsr_get_global_vcol_map (imgdist, matrix%m%global_cols)
    CALL array_hold (matrix%m%global_cols)
    matrix%m%has_global_cols = .TRUE.
  END SUBROUTINE dbcsr_reset_vlocals

! **************************************************************************************************
!> \brief Converts virtual 1-D distribution to process and image
!> \param nel ...
!> \param vdist ...
!> \param nim ...
!> \param pdist ...
!> \param idist ...
! **************************************************************************************************
  pure SUBROUTINE v_to_p_i_dist (nel, vdist, nim, pdist, idist)
      INTEGER, INTENT(in)                                :: nel
      INTEGER, DIMENSION(1:nel), INTENT(in)              :: vdist
      INTEGER, INTENT(in)                                :: nim
      INTEGER, DIMENSION(1:nel), INTENT(out)             :: pdist, idist

      INTEGER                                            :: i

    DO i = 1 , nel
       pdist(i) = vdist(i) / nim
       idist(i) = MOD (vdist(i), nim) + 1
    END DO
  END SUBROUTINE v_to_p_i_dist
! **************************************************************************************************
!> \brief ...
!> \param vdist ...
!> \param nim ...
!> \param pdist ...
!> \param idist ...
! **************************************************************************************************
  SUBROUTINE v_to_p_i_dist_o (vdist, nim, pdist, idist)
      TYPE(array_i1d_obj), INTENT(in)                    :: vdist
      INTEGER, INTENT(in)                                :: nim
      TYPE(array_i1d_obj), INTENT(out)                   :: pdist, idist

      INTEGER                                            :: nel
      INTEGER, DIMENSION(:), POINTER                     :: id, pd, vd

    nel = array_size (vdist)
    vd => array_data (vdist)
    ALLOCATE (pd(nel), id(nel))
    CALL v_to_p_i_dist (nel, vd, nim, pd, id)
    CALL array_new (pdist, pd, gift=.TRUE.)
    CALL array_new (idist, id, gift=.TRUE.)
  END SUBROUTINE v_to_p_i_dist_o


! **************************************************************************************************
!> \brief Determines whether most blocks are stored transposed in normally.
!> \param[in] blk_p           Pointers to blocks
!> \retval mostly_non_transposed ...
!> \note Tries to be quick and not necessarily accurate.
! **************************************************************************************************
  PURE FUNCTION mostly_non_transposed (blk_p)
    INTEGER, DIMENSION(:), INTENT(IN) :: blk_p
    LOGICAL                           :: mostly_non_transposed

    INTEGER            :: n, str, sntr
    INTEGER, PARAMETER :: allcheck_cutoff = 8

    n = SIZE (blk_p)
    IF (n .EQ. 0) THEN
       mostly_non_transposed = .TRUE.
       RETURN
    ENDIF
    str = 0
    sntr = 0
    CALL check_range (blk_p, 1, allcheck_cutoff, sntr, str)
    IF (n .GT. 4*allcheck_cutoff) THEN
       CALL check_range (blk_p, (n-allcheck_cutoff)/2, (n+allcheck_cutoff)/2,&
            sntr, str)
       CALL check_range (blk_p, n-allcheck_cutoff, n, sntr, str)
    ENDIF
    IF (str .EQ. 0) THEN
       mostly_non_transposed = .TRUE.
       RETURN
    ELSE
       ! Bias towards .TRUE.
       mostly_non_transposed = ((2*str) / (1+str+sntr)) .EQ. 0
    ENDIF
    RETURN
  CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param blk_p ...
!> \param lb ...
!> \param ub ...
!> \param sntr ...
!> \param str ...
! **************************************************************************************************
    PURE SUBROUTINE check_range (blk_p, lb, ub, sntr, str)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: blk_p
      INTEGER, INTENT(IN)                                :: lb, ub
      INTEGER, INTENT(INOUT)                             :: sntr, str

      INTEGER                                            :: b1, b2

      b1 = MAX (1, lb)
      b2 = MIN (SIZE(blk_p), ub)
      sntr = sntr + COUNT (blk_p(b1:b2) .GT. 0)
      str = str + COUNT (blk_p(b1:b2) .LT. 0)
    END SUBROUTINE check_range
  END FUNCTION mostly_non_transposed

! **************************************************************************************************
!> \brief Transform between virtual process rows/columns and actual process rows/columns and images therein.
!> \param image_dist ...
!> \param prow ...
!> \param rowi ...
!> \param pcol ...
!> \param coli ...
!> \param vprow ...
!> \param vpcol ...
!> \param myprow ...
!> \param mypcol ...
!> \param myrowi ...
!> \param mycoli ...
!> \param myvprow ...
!> \param myvpcol ...
!> \param vprow_shift ...
!> \param vpcol_shift ...
!> \param shifting ...
!> \par Shifting
!>  (L)eft and (R)ight shifting are "shifts from", (l)eft and (r)ight
!>  are "shifts to".  A caller (or the my* specifications) would use
!>  L/R to see which data he has (i.e., from where his data was
!>  shifted).  To see where the caller's data goes to, use l/r.
! **************************************************************************************************
  SUBROUTINE image_calculator (image_dist,&
       prow, rowi, pcol, coli, vprow, vpcol,&
       myprow, mypcol, myrowi, mycoli, myvprow, myvpcol,&
       vprow_shift, vpcol_shift,&
       shifting)
      TYPE(dbcsr_imagedistribution_obj), INTENT(IN)      :: image_dist
      INTEGER, INTENT(OUT), OPTIONAL                     :: prow, rowi, pcol, coli, vprow, vpcol
      INTEGER, INTENT(IN), OPTIONAL                      :: myprow, mypcol, myrowi, mycoli, myvprow, &
                                                            myvpcol, vprow_shift, vpcol_shift
      CHARACTER, INTENT(IN), OPTIONAL                    :: shifting

      CHARACTER(len=*), PARAMETER :: routineN = 'image_calculator', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col_mult, my_pcol, my_prow, ncol_images, &
                                                            npcols, nprows, nrow_images, nvpcols, &
                                                            nvprows, row_mult, vcol, vrow
      TYPE(dbcsr_mp_obj)                                 :: mp

!   ---------------------------------------------------------------------------

    IF (careful_mod) THEN
       CALL dbcsr_assert (.not.PRESENT (myvprow), "IMP",&
            PRESENT(mycoli),&
            dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
            "Must specify either (process row and row image) or (virtual process row)",&
            2342)
       CALL dbcsr_assert (.not.PRESENT (myvpcol), "IMP",&
            PRESENT(mycoli),&
            dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
            "Must specify either (process col and col image) or (virtual process col)",&
            2347)
    ENDIF
    !
    mp = image_dist%i%main%d%mp_env
    nprows = SIZE (mp%mp%pgrid, 1)
    npcols = SIZE (mp%mp%pgrid, 2)
    nrow_images = image_dist%i%row_decimation
    ncol_images = image_dist%i%col_decimation
    row_mult = image_dist%i%row_multiplicity
    col_mult = image_dist%i%col_multiplicity
    nvprows = nprows * nrow_images
    nvpcols = npcols * ncol_images
    !
    IF (PRESENT (myprow)) THEN
       my_prow = myprow
    ELSE
       my_prow = mp%mp%myprow
    ENDIF
    IF (PRESENT (mypcol)) THEN
       my_pcol = mypcol
    ELSE
       my_pcol = mp%mp%mypcol
    ENDIF
    !
    IF (.NOT. PRESENT (myvprow)) THEN
       vrow = my_prow*nrow_images+myrowi-1
    ELSE
       vrow = myvprow
    ENDIF
    IF (.NOT.PRESENT (myvpcol)) THEN
       vcol = my_pcol*ncol_images+mycoli-1
    ELSE
       vcol = myvpcol
    ENDIF
    !
    IF (PRESENT (vprow_shift)) vrow = vrow + vprow_shift
    IF (PRESENT (vpcol_shift)) vcol = vcol + vpcol_shift
    IF (PRESENT (shifting)) THEN
       SELECT CASE (shifting)
       CASE ('R')
          vrow = vrow + my_pcol*row_mult
       CASE ('L')
          vcol = vcol + my_prow*col_mult
       CASE ('r')
          vrow = vrow - my_pcol*row_mult
       CASE ('l')
          vcol = vcol - my_prow*col_mult
       END SELECT
    ENDIF
    vrow = MODULO (vrow, nvprows)
    vcol = MODULO (vcol, nvpcols)
    IF (PRESENT (prow)) prow = vrow / nrow_images
    IF (PRESENT (rowi)) rowi = MODULO (vrow, nrow_images)+1
    IF (PRESENT (pcol)) pcol = vcol / ncol_images
    IF (PRESENT (coli)) coli = MODULO (vcol, ncol_images)+1
    IF (PRESENT (vprow)) vprow = vrow
    IF (PRESENT (vpcol)) vpcol = vcol
  END SUBROUTINE image_calculator

! **************************************************************************************************
!> \brief Creates a sane 1-D distribution
!> \param[out] bin_dist    Distribution of elements to bins
!> \param[in] nelements    Number of elements
!> \param[in] nbins        Number of bins
!> \param[in] element_sizes  (optional) sizes of elements
! **************************************************************************************************
  SUBROUTINE dbcsr_dist_bin (bin_dist, nelements, nbins, element_sizes)
      INTEGER, DIMENSION(:), INTENT(OUT), POINTER        :: bin_dist
      INTEGER, INTENT(IN)                                :: nelements, nbins
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: element_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_dist_bin', routineP = moduleN//':'//routineN

      INTEGER                                            :: bin, bin_cnt, error_handle, i
      LOGICAL                                            :: found, heap_error
      REAL(kind=sp)                                      :: rn
      TYPE(heap_t)                                       :: bin_heap

!   ---------------------------------------------------------------------------

    CALL timeset (routineN, error_handle)
    ALLOCATE (bin_dist (nelements))
    IF (PRESENT (element_sizes)) THEN
       CALL dbcsr_assert (SIZE(element_sizes), "EQ", nelements,&
            dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
            "Array of element sizes does not match the number of elements.",&
            2433)
       CALL heap_new (bin_heap, nbins)
       CALL heap_fill (bin_heap, (/ (bin, bin = 0, nbins-1) /),&
                                 (/ (0,   bin = 0, nbins-1) /), heap_error)
       DO i = 1, nelements
          CALL heap_get_first (bin_heap, bin, bin_cnt, found, heap_error)
          bin_dist(i) = bin
          bin_cnt = bin_cnt + element_sizes(i)
          CALL heap_reset_first (bin_heap, bin_cnt, heap_error)
       ENDDO
       CALL heap_release (bin_heap)
    ELSE
       DO i = 1, nelements
          CALL RANDOM_NUMBER(rn)
          bin_dist(i) = MOD (INT (rn*REAL (nbins, kind=sp)), nbins)
       ENDDO
    ENDIF
    CALL timestop (error_handle)
  END SUBROUTINE dbcsr_dist_bin

END MODULE dbcsr_dist_operations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Operations on the DBCSR index
!> \author  Urban Borstnik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - Created 2010-02-18
! **************************************************************************************************
MODULE dbcsr_index_operations

   USE array_types,                     ONLY: array_data,&
                                              array_size
   USE dbcsr_config,                    ONLY: default_resize_factor
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_local_cols,&
                                              dbcsr_distribution_local_rows,&
                                              dbcsr_distribution_ncols,&
                                              dbcsr_distribution_nlocal_cols,&
                                              dbcsr_distribution_nlocal_rows,&
                                              dbcsr_distribution_nrows,&
                                              dbcsr_distribution_thread_dist
   USE dbcsr_dist_operations,           ONLY: get_stored_canonical
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_caller_error,&
                                              dbcsr_failure_level,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_unimplemented_error_nr,&
                                              dbcsr_wrong_args_error
   USE dbcsr_methods,                   ONLY: dbcsr_distribution,&
                                              dbcsr_get_index_memory_type
   USE dbcsr_ptr_util,                  ONLY: ensure_array_size,&
                                              memory_allocate,&
                                              memory_deallocate
   USE dbcsr_toollib,                   ONLY: joaat_hash,&
                                              sort,&
                                              swap
   USE dbcsr_types,                     ONLY: &
        dbcsr_distribution_obj, dbcsr_meta_size, dbcsr_num_slots, dbcsr_obj, dbcsr_slot_blk_p, &
        dbcsr_slot_col_i, dbcsr_slot_coo_l, dbcsr_slot_home_prow, dbcsr_slot_home_vprow, &
        dbcsr_slot_nblkcols_local, dbcsr_slot_nblkcols_total, dbcsr_slot_nblkrows_local, &
        dbcsr_slot_nblkrows_total, dbcsr_slot_nblks, dbcsr_slot_nfullcols_local, dbcsr_slot_nze, &
        dbcsr_slot_row_p, dbcsr_slot_size, dbcsr_slot_thr_c, dbcsr_type
   USE kinds,                           ONLY: int_4,&
                                              int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_index_operations'

   LOGICAL, PARAMETER :: careful_mod = .FALSE.
   LOGICAL, PARAMETER :: debug_mod = .FALSE.

   ! Index transformations
   PUBLIC :: dbcsr_make_index_canonical, &
             transpose_index_local
   PUBLIC :: dbcsr_make_index_local_row, dbcsr_has_local_row_index
   PUBLIC :: dbcsr_make_index_list
   ! Dense/Sparse
   PUBLIC :: make_dense_index, make_undense_index
   ! Working with DBCSR and linear indices
   PUBLIC :: dbcsr_make_dbcsr_index, dbcsr_sort_indices, &
             merge_index_arrays, &
             dbcsr_expand_row_index, &
             dbcsr_count_row_index, dbcsr_build_row_index, &
             dbcsr_index_prune_deleted, &
             dbcsr_index_compact
   PUBLIC :: dbcsr_index_checksum
   ! Index array manipulation
   PUBLIC :: dbcsr_addto_index_array, dbcsr_clearfrom_index_array, &
             dbcsr_repoint_index, dbcsr_make_index_exist

   INTERFACE dbcsr_count_row_index
      MODULE PROCEDURE dbcsr_count_row_index_copy, &
         dbcsr_count_row_index_inplace
   END INTERFACE

   INTERFACE dbcsr_build_row_index
      MODULE PROCEDURE dbcsr_build_row_index_copy, &
         dbcsr_build_row_index_inplace
   END INTERFACE

CONTAINS

! **************************************************************************************************
!> \brief Makes a canonical index given the index arrays
!>
!> \param new_row_p ...
!> \param new_col_i ...
!> \param new_blk_p ...
!> \param old_row_p ...
!> \param old_col_i ...
!> \param old_blk_p ...
!> \param matrix ...
!> \par Description of canonical ordering
!> A non-(anti)symmetric matrix is left as is. Otherwise, the row and column
!> are stored in the position prescribed by the distribution.
!> \note
!> This routine uses hard-coded logic as to what constitutes a
!> canonical ordering
! **************************************************************************************************
   SUBROUTINE make_index_canonical(new_row_p, new_col_i, new_blk_p, &
                                   old_row_p, old_col_i, old_blk_p, matrix)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: new_row_p, new_col_i, new_blk_p
      INTEGER, DIMENSION(:), INTENT(IN)                  :: old_row_p, old_col_i, old_blk_p
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'make_index_canonical', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, nblks, row, stored_col, &
                                                            stored_row
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: row_i
      LOGICAL                                            :: tr

!   ---------------------------------------------------------------------------

      nblks = SIZE(old_blk_p)
      ALLOCATE (row_i(nblks))
      IF (debug_mod) THEN
         WRITE (*, *) "old row_p", old_row_p
         WRITE (*, *) "old col_i", old_col_i
         WRITE (*, *) "old blk_p", old_blk_p
      ENDIF
      DO row = 1, SIZE(old_row_p)-1
         DO blk = old_row_p(row)+1, old_row_p(row+1)
            col = old_col_i(blk)
            stored_row = row
            stored_col = col
            tr = .FALSE.
            CALL get_stored_canonical(matrix, stored_row, stored_col, tr)
            IF (debug_mod) &
               WRITE (*, '(A,2(1X,I5),A,2(1X,I5),";",I7,1X,L1)') &
               routineN//" X->", row, col, "->", &
               stored_row, stored_col, blk, tr
            row_i(blk) = stored_row
            new_col_i(blk) = stored_col
            IF (.NOT. tr) THEN
               new_blk_p(blk) = old_blk_p(blk)
            ELSE
               new_blk_p(blk) = -old_blk_p(blk)
            ENDIF
         ENDDO
      ENDDO
      CALL dbcsr_sort_indices(nblks, row_i, new_col_i, blk_p=new_blk_p)
      ! Re-create the index
      CALL dbcsr_make_dbcsr_index(new_row_p, row_i, SIZE(new_row_p)-1, nblks)
      IF (debug_mod) THEN
         WRITE (*, *) "new row_p", new_row_p
         WRITE (*, *) "new row_i", row_i
         WRITE (*, *) "new col_i", new_col_i
         WRITE (*, *) "new blk_p", new_blk_p
      ENDIF
   END SUBROUTINE make_index_canonical

! **************************************************************************************************
!> \brief Makes a CP2K triangular index given the index arrays
!>
!> \param new_row_p ...
!> \param new_col_i ...
!> \param new_blk_p ...
!> \param old_row_p ...
!> \param old_col_i ...
!> \param old_blk_p ...
!> \param matrix ...
!> \par Description of canonical ordering
!> A non-(anti)symmetric matrix is left as is. Otherwise, the row and column
!> are stored in the position prescribed by the distribution.
!> \note
!> This routine uses hard-coded logic as to what constitutes a
!> canonical ordering
! **************************************************************************************************
   SUBROUTINE make_index_triangular(new_row_p, new_col_i, new_blk_p, &
                                    old_row_p, old_col_i, old_blk_p, matrix)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: new_row_p, new_col_i, new_blk_p
      INTEGER, DIMENSION(:), INTENT(IN)                  :: old_row_p, old_col_i, old_blk_p
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'make_index_triangular', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, nblks, row, stored_col, &
                                                            stored_row
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: row_i
      LOGICAL                                            :: tr

!   ---------------------------------------------------------------------------

      nblks = SIZE(old_blk_p)
      ALLOCATE (row_i(nblks))
      IF (debug_mod) THEN
         WRITE (*, *) "old row_p", old_row_p
         WRITE (*, *) "old col_i", old_col_i
         WRITE (*, *) "old blk_p", old_blk_p
      ENDIF
      DO row = 1, SIZE(old_row_p)-1
         DO blk = old_row_p(row)+1, old_row_p(row+1)
            col = old_col_i(blk)
            stored_row = row
            stored_col = col
            tr = .FALSE.
            CALL get_stored_canonical(matrix, stored_row, stored_col, tr)
            IF (stored_row .GT. stored_col) THEN
               CALL swap(stored_row, stored_col)
               tr = .NOT. tr
            ENDIF
            IF (debug_mod) &
               WRITE (*, '(A,2(1X,I5),A,2(1X,I5),";",I7,1X,L1)') &
               routineN//" X->", row, col, "->", &
               stored_row, stored_col, blk, tr
            row_i(blk) = stored_row
            new_col_i(blk) = stored_col
            IF (.NOT. tr) THEN
               new_blk_p(blk) = old_blk_p(blk)
            ELSE
               new_blk_p(blk) = -old_blk_p(blk)
            ENDIF
         ENDDO
      ENDDO
      CALL dbcsr_sort_indices(nblks, row_i, new_col_i, blk_p=new_blk_p)
      ! Re-create the index
      CALL dbcsr_make_dbcsr_index(new_row_p, row_i, SIZE(new_row_p)-1, nblks)
      IF (debug_mod) THEN
         WRITE (*, *) "new row_p", new_row_p
         WRITE (*, *) "new row_i", row_i
         WRITE (*, *) "new col_i", new_col_i
         WRITE (*, *) "new blk_p", new_blk_p
      ENDIF
   END SUBROUTINE make_index_triangular

! **************************************************************************************************
!> \brief Collapses a row_p index
!> \param row_p ...
!> \param row_i ...
!> \param nrows ...
!> \param nblks ...
! **************************************************************************************************
   SUBROUTINE dbcsr_make_dbcsr_index(row_p, row_i, nrows, nblks)
      INTEGER, INTENT(in)                                :: nrows, nblks
      INTEGER, DIMENSION(1:nrows+1), INTENT(out)         :: row_p
      INTEGER, DIMENSION(1:nblks), INTENT(in)            :: row_i

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_dbcsr_index', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, error_handle, row

      CALL timeset(routineN, error_handle)

      row_p(1) = 0
      row_p(nrows+1) = nblks
      row = 1
      blk = 1
      DO WHILE (row .LE. nrows)
         IF (blk .LE. nblks) THEN
            DO WHILE (row_i(blk) .EQ. row)
               blk = blk+1
               IF (blk .GT. nblks) THEN
                  row_p(row+1) = nblks-1
                  EXIT
               ENDIF
            ENDDO
         ENDIF
         row_p(row+1) = blk-1
         row = row+1
      ENDDO

      CALL timestop(error_handle)

   END SUBROUTINE dbcsr_make_dbcsr_index

! **************************************************************************************************
!> \brief Expands a row_p index
!> \param row_p ...
!> \param row_i ...
!> \param nrows ...
!> \param nblks ...
! **************************************************************************************************
   PURE SUBROUTINE dbcsr_expand_row_index(row_p, row_i, nrows, nblks)
      INTEGER, INTENT(in)                                :: nrows, nblks
      INTEGER, DIMENSION(1:nrows+1), INTENT(IN)          :: row_p
      INTEGER, DIMENSION(1:nblks), INTENT(out)           :: row_i

      INTEGER                                            :: row

      DO row = 1, nrows
         row_i(row_p(row)+1:row_p(row+1)) = row
      ENDDO
   END SUBROUTINE dbcsr_expand_row_index

! **************************************************************************************************
!> \brief Counts columns-per-row count from row index array, in-place.
!> \param[in,out] rows        the row_p index (input); the count of the number
!>                            of columns per row (output)
!> \param[in] nrows           number of rows
! **************************************************************************************************
   PURE SUBROUTINE dbcsr_count_row_index_inplace(rows, nrows)
      INTEGER, INTENT(IN)                                :: nrows
      INTEGER, DIMENSION(1:nrows+1), INTENT(INOUT)       :: rows

      INTEGER                                            :: row

      DO row = 1, nrows
         rows(row) = rows(row+1)-rows(row)
      ENDDO
      rows(nrows+1) = 0
   END SUBROUTINE dbcsr_count_row_index_inplace

! **************************************************************************************************
!> \brief Counts columns-per-row count from row index array.
!> \param[in] rows            the row_p index (input)
!> \param[out] counts         the count of the number of columns per row
!> \param[in] nrows           number of rows
! **************************************************************************************************
   PURE SUBROUTINE dbcsr_count_row_index_copy(rows, counts, nrows)
      INTEGER, INTENT(IN)                                :: nrows
      INTEGER, DIMENSION(1:nrows), INTENT(OUT)           :: counts
      INTEGER, DIMENSION(1:nrows+1), INTENT(IN)          :: rows

      INTEGER                                            :: row

      DO row = 1, nrows
         counts(row) = rows(row+1)-rows(row)
      END DO
   END SUBROUTINE dbcsr_count_row_index_copy

! **************************************************************************************************
!> \brief Builds row index array from a columns-per-row count, in-place.
!> \param[in,out] rows        count of the number of colums per row (input);
!>                            the row_p index (output)
!> \param[in] nrows           number of rows
! **************************************************************************************************
   PURE SUBROUTINE dbcsr_build_row_index_inplace(rows, nrows)
      INTEGER, INTENT(IN)                                :: nrows
      INTEGER, DIMENSION(1:nrows+1), INTENT(INOUT)       :: rows

      INTEGER                                            :: o, old_count, row

      old_count = rows(1)
      rows(1) = 0
      IF (nrows .GE. 1) THEN
         DO row = 2, nrows+1
            o = rows(row)
            rows(row) = rows(row-1)+old_count
            old_count = o
         ENDDO
      ENDIF
   END SUBROUTINE dbcsr_build_row_index_inplace

! **************************************************************************************************
!> \brief Builds row index array from a columns-per-row count.
!> \param[in] counts          count of the number of colums per row
!> \param[out] rows           count of the number of colums per row (input);
!>                            the row_p index (output)
!> \param[in] nrows           number of rows
! **************************************************************************************************
   PURE SUBROUTINE dbcsr_build_row_index_copy(counts, rows, nrows)
      INTEGER, INTENT(IN)                                :: nrows
      INTEGER, DIMENSION(1:nrows+1), INTENT(OUT)         :: rows
      INTEGER, DIMENSION(1:nrows), INTENT(IN)            :: counts

!WTF?!rows(1) = 0
!WTF?!IF (nrows .GE. 1) THEN
!WTF?!   DO row = 2, nrows+1
!WTF?!      rows(row) = rows(row-1) + counts(rows-1)
!WTF?!   ENDDO
!WTF?!ENDIF

      rows(1:nrows) = counts(1:nrows)
      CALL dbcsr_build_row_index_inplace(rows, nrows)
   END SUBROUTINE dbcsr_build_row_index_copy

! **************************************************************************************************
!> \brief Adds data to the index. Incleases the index size when neccessary.
!> \param[in,out] matrix      bcsr matrix
!> \param[in] slot  which index array to add (e.g., dbcsr_slot_row_blk_sizes)
!> \param[in] DATA  (optional) array holding the index data to add to the index
!>                  array
!> \param[in] reservation     (optional) only reserve space for subsequent
!>                            array
!> \param[in] extra (optional) reserve extra space for later additions
! **************************************************************************************************
   SUBROUTINE dbcsr_addto_index_array(matrix, slot, DATA, reservation, extra)
      TYPE(dbcsr_type), INTENT(INOUT)                    :: matrix
      INTEGER, INTENT(IN)                                :: slot
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: DATA
      INTEGER, INTENT(IN), OPTIONAL                      :: reservation, extra

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_addto_index_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: deplus, space, ub, ub_new

!   ---------------------------------------------------------------------------

      IF (debug_mod) THEN
         IF (.NOT. ASSOCIATED(matrix%index)) &
            CALL cp__b("dbcsr/block/dbcsr_index_operations.F",408,"Index must be preallocated.")
         CALL dbcsr_assert(UBOUND(matrix%index, 1), 'GE', dbcsr_num_slots, &
                           dbcsr_failure_level, dbcsr_internal_error, routineP, &
                           "Actual index size less than declared size", 411)
         CALL dbcsr_assert(PRESENT(DATA), 'OR', PRESENT(reservation), dbcsr_fatal_level, &
                           dbcsr_caller_error, routineP, &
                           'Either an array or its size must be specified.', 414)
         WRITE (*, *) routineP//' index', matrix%index(:dbcsr_num_slots)
      ENDIF
      IF (PRESENT(reservation)) THEN
         space = reservation
      ELSE
         space = SIZE(DATA)
      ENDIF
      IF (PRESENT(extra)) THEN
         deplus = extra
      ELSE
         deplus = 0
      ENDIF
      ub = UBOUND(matrix%index, 1)
      !> The data area was not defined or the new area is greater than the old.
      IF (matrix%index(slot) .EQ. 0 .OR. &
          space .GT. matrix%index(slot+1)-matrix%index(slot)+1) THEN
         IF (debug_mod) WRITE (*, *) routineP//' Slot', slot, 'not filled, adding at', &
            matrix%index(dbcsr_slot_size)+1, 'sized', space
         matrix%index(slot) = matrix%index(dbcsr_slot_size)+1
         matrix%index(slot+1) = matrix%index(slot)+space-1
         matrix%index(dbcsr_slot_size) = matrix%index(slot+1)
      ENDIF
      ! Shorten an index entry.
      IF (space .LT. matrix%index(slot+1)-matrix%index(slot)+1) THEN
         IF (debug_mod) WRITE (*, *) routineP//' Shortening index'
         matrix%index(slot+1) = matrix%index(slot)+space-1
         CALL dbcsr_repoint_index(matrix, slot)
      ENDIF
      ub_new = matrix%index(slot+1)+deplus
      IF (debug_mod) WRITE (*, *) routineP//' need', space, 'at', matrix%index(slot), &
         'to', matrix%index(slot+1), '(', ub_new, ')', 'have', ub
      IF (ub_new .GT. ub) THEN
         IF (debug_mod) WRITE (*, *) routineP//' Reallocating index to ubound', ub_new
         !CALL reallocate(matrix%index, 1, ub_new)
         CALL ensure_array_size(matrix%index, lb=1, ub=ub_new, &
                                factor=default_resize_factor, &
                                nocopy=.FALSE., &
                                memory_type=matrix%index_memory_type)
         CALL dbcsr_repoint_index(matrix)
      ENDIF
      IF (debug_mod) WRITE (*, *) routineP//' Adding slot', slot, 'at', &
         matrix%index(slot), 'sized', space
      CALL dbcsr_repoint_index(matrix, slot)
      IF (PRESENT(DATA)) &
         matrix%index(matrix%index(slot):matrix%index(slot+1)) = DATA(:)
   END SUBROUTINE dbcsr_addto_index_array

! **************************************************************************************************
!> \brief Removes data from the index.
!> \param[in,out] matrix      bcsr matrix
!> \param[in] slot  which index array to remove (e.g., dbcsr_slot_row_blk_sizes)
! **************************************************************************************************
   SUBROUTINE dbcsr_clearfrom_index_array(matrix, slot)
      TYPE(dbcsr_type), INTENT(INOUT)                    :: matrix
      INTEGER, INTENT(IN)                                :: slot

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_clearfrom_index_array', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: space
      INTEGER, DIMENSION(5)                              :: max_extents

!   ---------------------------------------------------------------------------

      IF (.NOT. ASSOCIATED(matrix%index)) &
         CALL cp__b("dbcsr/block/dbcsr_index_operations.F",480,"Index must be preallocated.")
      CALL dbcsr_assert(UBOUND(matrix%index, 1), 'GE', dbcsr_num_slots, &
                        dbcsr_failure_level, dbcsr_internal_error, routineP, &
                        "Actual index size less than declared size", 483)
      IF (debug_mod) WRITE (*, *) routineP//' index', &
         matrix%index(:dbcsr_num_slots)
      ! Clear index entry pointer
      matrix%index(slot) = 1
      matrix%index(slot+1) = 0
      CALL dbcsr_repoint_index(matrix, slot)
      ! Update the declared index size
      max_extents = (/ &
                    matrix%index(dbcsr_slot_row_p+1), &
                    matrix%index(dbcsr_slot_col_i+1), &
                    matrix%index(dbcsr_slot_blk_p+1), &
                    matrix%index(dbcsr_slot_thr_c+1), &
                    matrix%index(dbcsr_slot_coo_l+1)/)
      space = MAX(MAXVAL(max_extents), dbcsr_num_slots)
      matrix%index(dbcsr_slot_size) = space
   END SUBROUTINE dbcsr_clearfrom_index_array

! **************************************************************************************************
!> \brief Updates the index pointers of a bcsr matrix
!> \param[in,out] m           matrix for which index pointers are updated
!> \param[in] slot            (optional) only repoint this index
! **************************************************************************************************
   SUBROUTINE dbcsr_repoint_index(m, slot)
      TYPE(dbcsr_type), INTENT(INOUT)                    :: m
      INTEGER, INTENT(IN), OPTIONAL                      :: slot

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_repoint_index', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: s
      LOGICAL                                            :: all

!   ---------------------------------------------------------------------------

      IF (m%nblks .NE. m%index(dbcsr_slot_nblks)) THEN
         m%nblks = m%index(dbcsr_slot_nblks)
         m%nze = m%index(dbcsr_slot_nze)
      ENDIF
      all = .TRUE.
      IF (PRESENT(slot)) THEN
         all = .FALSE.
         s = slot
      ELSE
         s = 0
      ENDIF

      IF (m%index(dbcsr_slot_row_p) .GT. 0 .AND. all .OR. &
          s .EQ. dbcsr_slot_row_p) THEN
         IF (m%index(dbcsr_slot_row_p) .GT. 0) THEN
            m%row_p => m%index(m%index(dbcsr_slot_row_p): &
                               m%index(dbcsr_slot_row_p+1))
         ELSE
            NULLIFY (m%row_p)
         ENDIF
      ENDIF
      IF (m%index(dbcsr_slot_col_i) .GT. 0 .AND. all .OR. &
          s .EQ. dbcsr_slot_col_i) THEN
         IF (m%index(dbcsr_slot_col_i) .GT. 0) THEN
            m%col_i => m%index(m%index(dbcsr_slot_col_i): &
                               m%index(dbcsr_slot_col_i+1))
         ELSE
            NULLIFY (m%col_i)
         ENDIF
      ENDIF
      IF (m%index(dbcsr_slot_blk_p) .GT. 0 .AND. all .OR. &
          s .EQ. dbcsr_slot_blk_p) THEN
         IF (m%index(dbcsr_slot_blk_p) .GT. 0) THEN
            m%blk_p => m%index(m%index(dbcsr_slot_blk_p): &
                               m%index(dbcsr_slot_blk_p+1))
         ELSE
            NULLIFY (m%blk_p)
         ENDIF
      ENDIF
      IF (m%index(dbcsr_slot_thr_c) .GT. 0 .AND. all .OR. &
          s .EQ. dbcsr_slot_thr_c) THEN
         IF (m%index(dbcsr_slot_thr_c) .GT. 0) THEN
            m%thr_c => m%index(m%index(dbcsr_slot_thr_c): &
                               m%index(dbcsr_slot_thr_c+1))
         ELSE
            NULLIFY (m%thr_c)
         ENDIF
      ENDIF
      IF (m%index(dbcsr_slot_coo_l) .GT. 0 .AND. all .OR. &
          s .EQ. dbcsr_slot_coo_l) THEN
         IF (m%index(dbcsr_slot_coo_l) .GT. 0) THEN
            m%coo_l => m%index(m%index(dbcsr_slot_coo_l): &
                               m%index(dbcsr_slot_coo_l+1))
         ELSE
            NULLIFY (m%coo_l)
         ENDIF
      ENDIF
      IF (all) THEN
         m%index(dbcsr_slot_nblks) = m%nblks
         m%index(dbcsr_slot_nze) = m%nze
      ENDIF
   END SUBROUTINE dbcsr_repoint_index

! **************************************************************************************************
!> \brief ...
!> \param[in,out] m           Create index for this matrix
! **************************************************************************************************

   SUBROUTINE dbcsr_make_index_exist(m)
      TYPE(dbcsr_type), INTENT(INOUT)                    :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_index_exist', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      IF (.NOT. ASSOCIATED(m%index)) &
         CALL cp__b("dbcsr/block/dbcsr_index_operations.F",598,"Index array does not yet exist.")
      IF (.NOT. ASSOCIATED(m%row_p)) THEN
         CALL dbcsr_addto_index_array(m, dbcsr_slot_row_p, &
                                      reservation=m%nblkrows_total+1)
         m%row_p(:) = 0
      ENDIF
      IF (.NOT. ASSOCIATED(m%col_i)) THEN
         CALL dbcsr_addto_index_array(m, dbcsr_slot_col_i, &
                                      reservation=0)
      ENDIF
      IF (.NOT. ASSOCIATED(m%blk_p)) THEN
         CALL dbcsr_addto_index_array(m, dbcsr_slot_blk_p, &
                                      reservation=0)
      ENDIF
      CALL dbcsr_repoint_index(m)
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_make_index_exist

! **************************************************************************************************
!> \brief Sorts the rows & columns of a work matrix
!> \param[in] n               number of blocks (elements) to sort
!> \param[in,out] row_i       row indices
!> \param[in,out] col_i       column indices
!> \param[in,out] blk_p       (optional) block pointers
!> \param[in,out] blk_d       (optional) data storage
!> \par Description
!>      Sorts the row and column indices so that the rows monotonically
!>      increase and the columns monotonically increase within each row.
!>      Passing the blk_p array rearranges the block pointers accordingly.
!>      This must be done if they are pointing to valid data, otherwise
!>      they become invalid.
! **************************************************************************************************
   SUBROUTINE dbcsr_sort_indices(n, row_i, col_i, blk_p, blk_d)
      INTEGER, INTENT(IN)                                :: n
      INTEGER, DIMENSION(1:), INTENT(INOUT)              :: row_i, col_i
      INTEGER, DIMENSION(1:), INTENT(INOUT), OPTIONAL    :: blk_p, blk_d

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_sort_indices', &
         routineP = moduleN//':'//routineN
      INTEGER(KIND=int_8), PARAMETER                     :: lmask8 = 4294967295_int_8

      INTEGER                                            :: error_handle, i, stat
      INTEGER(KIND=int_8), ALLOCATABLE, DIMENSION(:)     :: sort_keys
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: buf, buf_d

!   ---------------------------------------------------------------------------

      IF (SIZE(row_i) .EQ. 0) RETURN

      CALL timeset(routineN, error_handle)

      CALL dbcsr_assert(SIZE(row_i), 'GE', n, dbcsr_failure_level, &
                        dbcsr_caller_error, routineP, 'row_i too small', 650)
      CALL dbcsr_assert(SIZE(col_i), 'GE', n, dbcsr_failure_level, &
                        dbcsr_caller_error, routineP, 'col_i too small', 652)
      IF (PRESENT(blk_p)) CALL dbcsr_assert(SIZE(blk_p), 'GE', n, dbcsr_failure_level, &
                                            dbcsr_caller_error, routineP, 'blk_p too small', 654)
      IF (PRESENT(blk_p)) THEN
         ALLOCATE (buf(n), stat=stat)
         IF (stat /= 0) CALL cp__b("dbcsr/block/dbcsr_index_operations.F",657,'buf')
         buf(1:n) = blk_p(1:n)
      ENDIF
      IF (PRESENT(blk_d)) THEN
         ALLOCATE (buf_d(n), stat=stat)
         IF (stat /= 0) CALL cp__b("dbcsr/block/dbcsr_index_operations.F",662,'buf_d')
         buf_d(1:n) = blk_d(1:n)
      ENDIF
      !> Create an ordering for both rows and columns. If the blk_p must
      !> be rearranged, then the col_i array will be used as a
      !> permutation vector.
      ALLOCATE (sort_keys(n), stat=stat)
      CALL dbcsr_assert(stat, "EQ", 0, dbcsr_fatal_level, dbcsr_internal_error, &
                        routineN, "Could not allocate memory for sorting buffer.", 670)
      sort_keys(:) = IOR(ISHFT(INT(row_i(1:n), int_8), 32), INT(col_i(1:n), int_8))
      IF (PRESENT(blk_p)) col_i(1:n) = (/(i, i=1, n)/)
      !> Now do a nice quicksort.
      CALL sort(sort_keys, n, col_i)
      ! Since blk_d is usually not present we can have two loops that
      ! are essentially the same.
      IF (PRESENT(blk_p)) THEN
         DO i = 1, n
            blk_p(i) = buf(col_i(i))
         ENDDO
         DEALLOCATE (buf)
      END IF
      IF (PRESENT(blk_d)) THEN
         DO i = 1, n
            blk_d(i) = buf_d(col_i(i))
         ENDDO
         DEALLOCATE (buf_d)
      ENDIF
      DO i = 1, n
         col_i(i) = INT(IAND(sort_keys(i), lmask8), int_4)
         row_i(i) = INT(ISHFT(sort_keys(i), -32), int_4)
      ENDDO
      DEALLOCATE (sort_keys)
      IF (debug_mod .AND. PRESENT(blk_p)) &
         WRITE (*, *) routineP//' sort, blk_p =', blk_p
      CALL timestop(error_handle)

   END SUBROUTINE dbcsr_sort_indices

! **************************************************************************************************
!> \brief Removes the deleted blocks from the index.
!> \param[in,out] matrix      Prune the index of this matrix.
!> \par Description
! **************************************************************************************************
   SUBROUTINE dbcsr_index_prune_deleted(matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_index_prune_deleted', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, nblks_max, new_blk, nrows, &
                                                            old_blk, row, stat
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: new_blk_p, new_col_i, new_row_count
      INTEGER, DIMENSION(:), POINTER                     :: old_blk_p, old_col_i, old_row_p

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      !
      old_row_p => matrix%m%row_p
      old_col_i => matrix%m%col_i
      old_blk_p => matrix%m%blk_p
      !
      nrows = matrix%m%nblkrows_total
      nblks_max = old_row_p(nrows+1)
      ALLOCATE (new_row_count(nrows), stat=stat)
      CALL dbcsr_assert(stat, "EQ", 0, dbcsr_fatal_level, dbcsr_internal_error, &
                        routineN, "Can not allocate memory.", 728)
      ALLOCATE (new_col_i(nblks_max), stat=stat)
      CALL dbcsr_assert(stat, "EQ", 0, dbcsr_fatal_level, dbcsr_internal_error, &
                        routineN, "Can not allocate memory.", 731)
      ALLOCATE (new_blk_p(nblks_max), stat=stat)
      CALL dbcsr_assert(stat, "EQ", 0, dbcsr_fatal_level, dbcsr_internal_error, &
                        routineN, "Can not allocate memory.", 734)
      !
      ! Build up the new index from all non-deleted blocks in the
      ! existing index.
      new_blk = 0
      DO row = 1, nrows
         new_row_count(row) = 0
         DO old_blk = old_row_p(row)+1, old_row_p(row+1)
            IF (old_blk_p(old_blk) .GT. 0) THEN
               new_blk = new_blk+1
               new_row_count(row) = new_row_count(row)+1
               new_col_i(new_blk) = old_col_i(old_blk)
               new_blk_p(new_blk) = old_blk_p(old_blk)
            ENDIF
         ENDDO
      ENDDO
      !
      CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_row_p)
      CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_col_i)
      CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_blk_p)
      CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_row_p, &
                                   reservation=nrows+1, extra=2*new_blk)
      old_row_p => matrix%m%row_p
      CALL dbcsr_build_row_index(counts=new_row_count, rows=old_row_p, &
                                 nrows=nrows)
      CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_col_i, DATA=new_col_i(1:new_blk))
      CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_blk_p, DATA=new_blk_p(1:new_blk))
      matrix%m%nblks = new_blk
      matrix%m%index(dbcsr_slot_nblks) = new_blk
      !
      DEALLOCATE (new_col_i, new_blk_p, new_row_count)
      !
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_index_prune_deleted

! **************************************************************************************************
!> \brief Re-indexes row_p and blk_i according to columns.
!> \param[out] new_col_p      new column pointer
!> \param[out] new_row_i      new row index
!> \param[in] old_row_p       old row pointer
!> \param[in] old_col_i       old column index
!> \param[out] new_blk_p      (optional) new block pointer
!> \param[in] old_blk_p       (optional) old block pointer
!> \par The re-indexing is equivalent to a local-only transpose.
! **************************************************************************************************
   SUBROUTINE transpose_index_local(new_col_p, new_row_i, old_row_p, &
                                    old_col_i, new_blk_p, old_blk_p)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: new_col_p, new_row_i
      INTEGER, DIMENSION(:), INTENT(IN)                  :: old_row_p, old_col_i
      INTEGER, DIMENSION(:), INTENT(OUT), OPTIONAL       :: new_blk_p
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: old_blk_p

      CHARACTER(len=*), PARAMETER :: routineN = 'transpose_index_local', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, nblks, ncols_new, nrows_old
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: new_col_i
      LOGICAL                                            :: blks

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      blks = PRESENT(new_blk_p) .AND. PRESENT(old_blk_p)
      nblks = SIZE(old_col_i)
      nrows_old = SIZE(old_row_p)-1
      ncols_new = SIZE(new_col_p)-1
      IF (blks) new_blk_p(:) = old_blk_p(:)
      ALLOCATE (new_col_i(nblks))
      CALL dbcsr_expand_row_index(old_row_p, new_row_i, nrows_old, nblks)
      new_col_i(:) = old_col_i(:)
      CALL dbcsr_sort_indices(nblks, new_col_i, new_row_i, new_blk_p)
      CALL dbcsr_make_dbcsr_index(new_col_p, new_col_i, ncols_new, nblks)
      DEALLOCATE (new_col_i)
      CALL timestop(error_handle)
   END SUBROUTINE transpose_index_local

! **************************************************************************************************
!> \brief Makes a canonical index to the distribution.
!> \param[inout] matrix       matrix for which to make canonical index
!> \param[in] cp2k            (optional) make CP2K triangular index
!>                            from canonical; default is false
!> \par Canonical means that it respects the distribution.
! **************************************************************************************************
   SUBROUTINE dbcsr_make_index_canonical(matrix, cp2k)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: cp2k

      INTEGER                                            :: nb, nc, nr
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: new_blk_p, new_col_i, new_row_p
      LOGICAL                                            :: rev

!   ---------------------------------------------------------------------------

      rev = .FALSE.
      IF (PRESENT(cp2k)) rev = cp2k
      nr = SIZE(matrix%m%row_p)
      ALLOCATE (new_row_p(nr))
      nc = SIZE(matrix%m%col_i)
      ALLOCATE (new_col_i(nc))
      nb = SIZE(matrix%m%blk_p)
      ALLOCATE (new_blk_p(nb))
      IF (rev) THEN
         CALL make_index_triangular(new_row_p, new_col_i, new_blk_p, &
                                    matrix%m%row_p, matrix%m%col_i, matrix%m%blk_p, matrix%m)
      ELSE
         CALL make_index_canonical(new_row_p, new_col_i, new_blk_p, &
                                   matrix%m%row_p, matrix%m%col_i, matrix%m%blk_p, matrix%m)
      ENDIF
      matrix%m%row_p(:) = new_row_p
      matrix%m%col_i(:) = new_col_i
      matrix%m%blk_p(:) = new_blk_p
   END SUBROUTINE dbcsr_make_index_canonical

! **************************************************************************************************
!> \brief Makes the index for a dense matrix
!> \param[out] row_p Storage for new index
!> \param[out] col_i Storage for new index
!> \param[out] blk_p Storage for new index
!> \param[in] nblkrows_total                  Total blocked rows
!> \param[in] nblkcols_total                  Total blocked columns
!> \param[in] myblkrows                       List of blocked rows in my
!>                                            process row
!> \param[in] myblkcols                       List of blocked columns in my
!>                                            process column
!> \param row_blk_offsets ...
!> \param col_blk_offsets ...
!> \param[in,out] meta                        Metadata updates for new index
!> \param[in] make_tr                         Dense blocks are transposed
!> \note Used for making matrices dense/undense
! **************************************************************************************************
   SUBROUTINE make_dense_index(row_p, col_i, blk_p, &
                               nblkrows_total, nblkcols_total, myblkrows, myblkcols, &
                               row_blk_offsets, col_blk_offsets, meta, make_tr)

      !INTEGER, DIMENSION(:), INTENT(OUT)       :: row_p, col_i, blk_p
      INTEGER, INTENT(IN)                                :: nblkrows_total
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: blk_p, col_i
      INTEGER, DIMENSION(1:nblkrows_total+1), &
         INTENT(OUT)                                     :: row_p
      INTEGER, INTENT(IN)                                :: nblkcols_total
      INTEGER, DIMENSION(:), INTENT(IN)                  :: myblkrows, myblkcols, row_blk_offsets, &
                                                            col_blk_offsets
      INTEGER, DIMENSION(dbcsr_meta_size), INTENT(INOUT) :: meta
      LOGICAL, INTENT(IN), OPTIONAL                      :: make_tr

      CHARACTER(len=*), PARAMETER :: routineN = 'make_dense_index', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, c, col_l, mynblkcols, mynblkrows, &
                                                            nblks, nze, prev_row, row, row_l, &
                                                            sign_carrier, sz

!   ---------------------------------------------------------------------------

      sign_carrier = 1
      IF (PRESENT(make_tr)) THEN
         IF (make_tr) sign_carrier = -1
      ENDIF
      mynblkrows = SIZE(myblkrows)
      mynblkcols = SIZE(myblkcols)
      meta(dbcsr_slot_nblkrows_local) = mynblkrows
      meta(dbcsr_slot_nblkcols_local) = mynblkcols
      nblks = mynblkrows*mynblkcols
      nze = 1
      IF (nblks .EQ. 0) THEN
         row_p(1:) = 0
      ELSE
         row_p(1) = 0
         !row_p(nrows+1) = nblks
         prev_row = 1
         blk = 0
         DO row_l = 1, mynblkrows
            row = myblkrows(row_l)
            row_p(prev_row+1:row) = blk
            DO col_l = 1, mynblkcols
               c = myblkcols(col_l)
               col_i(blk+col_l) = c
               sz = (row_blk_offsets(row+1)-row_blk_offsets(row))* &
                    (col_blk_offsets(c+1)-col_blk_offsets(c))
               IF (sz .GT. 0) THEN
                  blk_p(blk+col_l) = SIGN(nze, sign_carrier)
                  nze = nze+sz
               ELSE
                  blk_p(blk+col_l) = 0
               ENDIF
            ENDDO
            prev_row = row
            blk = blk+mynblkcols
         END DO
         CALL dbcsr_assert(blk, "EQ", nblks, dbcsr_fatal_level, &
                           dbcsr_internal_error, routineN, "Block mismatch", 924)
         row_p(prev_row+1:nblkrows_total+1) = nblks
      ENDIF
      IF (debug_mod) THEN
         WRITE (*, *) routineN//" new index"
         WRITE (*, *) "row_p=", row_p
         WRITE (*, *) "col_i=", col_i
         WRITE (*, *) "blk_p=", blk_p
      ENDIF
      meta(dbcsr_slot_nblkrows_total) = nblkrows_total
      meta(dbcsr_slot_nblkcols_total) = nblkcols_total
   END SUBROUTINE make_dense_index

! **************************************************************************************************
!> \brief Makes a blocked index from a dense matrix
!> \param[out] row_p Storage for new index
!> \param[out] col_i Storage for new index
!> \param[out] blk_p Storage for new index
!> \param[in] distribution     Blocked distribution
!> \param local_row_offsets ...
!> \param local_col_offsets ...
!> \param[in,out] meta         Metadata updates for new index
!> \note Used for making matrices dense/undense
! **************************************************************************************************
   SUBROUTINE make_undense_index( &
      row_p, col_i, blk_p, &
      distribution, local_row_offsets, local_col_offsets, &
      meta)
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: row_p, col_i, blk_p
      TYPE(dbcsr_distribution_obj)                       :: distribution
      INTEGER, DIMENSION(:), INTENT(IN)                  :: local_row_offsets, local_col_offsets
      INTEGER, DIMENSION(dbcsr_meta_size), INTENT(INOUT) :: meta

      CHARACTER(len=*), PARAMETER :: routineN = 'make_undense_index', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col, lr, lrow, nblkcols_local, &
                                                            nblkrows_local, nblkrows_total, &
                                                            nfullcols_local, prev_row, row
      INTEGER, DIMENSION(:), POINTER                     :: local_cols, local_rows

!   ---------------------------------------------------------------------------

      local_cols => dbcsr_distribution_local_cols(distribution)
      local_rows => dbcsr_distribution_local_rows(distribution)
      meta(dbcsr_slot_nblkrows_total) = dbcsr_distribution_nrows(distribution)
      meta(dbcsr_slot_nblkcols_total) = dbcsr_distribution_ncols(distribution)
      meta(dbcsr_slot_nblkrows_local) = dbcsr_distribution_nlocal_rows(distribution)
      meta(dbcsr_slot_nblkcols_local) = dbcsr_distribution_nlocal_cols(distribution)
      nblkrows_total = meta(dbcsr_slot_nblkrows_total)
      nblkcols_local = meta(dbcsr_slot_nblkcols_local)
      nblkrows_local = meta(dbcsr_slot_nblkrows_local)
      nfullcols_local = meta(dbcsr_slot_nfullcols_local)
      ! Fill the row_p array.
      lr = 0
      row_p(1) = 0
      prev_row = 1
      DO lrow = 1, nblkrows_local
         row = local_rows(lrow)
         row_p(prev_row+1:row) = lr
         lr = lr+nblkcols_local
         row_p(row+1) = lr
         prev_row = row
      ENDDO
      row_p(prev_row+1:nblkrows_total+1) = lr
      !
      DO row = 1, nblkrows_local
         DO col = 1, nblkcols_local
            col_i(nblkcols_local*(row-1)+col) = local_cols(col)
            blk_p(nblkcols_local*(row-1)+col) = 1+ &
                                                (local_row_offsets(row)-1)*nfullcols_local &
                                                +(local_col_offsets(col)-1)* &
                                                (local_row_offsets(row+1)-local_row_offsets(row))
         END DO
      END DO
   END SUBROUTINE make_undense_index

! **************************************************************************************************
!> \brief Merges two indices
!> \param[in,out] new_row_i merged result
!> \param[in,out] new_col_i merged result
!> \param[in,out] new_blk_p merged result
!> \param[in] new_size                  size of merged index
!> \param[in,out] old_row_i current index
!> \param[in,out] old_col_i current index
!> \param[in,out] old_blk_p current index
!> \param[in] old_size                  size of current index
!> \param[in] add_ip                    index to add into the current index
!> \param[in] add_size                  size of index to add into the current
!>                                      index
!> \param new_blk_d ...
!> \param old_blk_d ...
!> \param[in] added_size_offset         (optional) specify base of added sizes
!> \param[in] added_sizes               (optional) specify sizes of added
!>                                      blocks
!> \param[out] added_size               (optional) counts number of sizes of
!>                                      added blocks
!> \param[out] added_nblks              (optional) actual number of new
!>                                      elements
!> \par Added sizes
!>      added_size_offset and added_sizes can be optionally
!>      specified. This is meant for cases where the added blocks may
!>      be duplicates of existing blocks. In this way it is possible
!>      to recalculate new block pointers to avoid wasted space.
!> \note Used in local multiply
!>
!> Assumes they are both pre-sorted
! **************************************************************************************************
   SUBROUTINE merge_index_arrays(new_row_i, new_col_i, new_blk_p, new_size, &
                                 old_row_i, old_col_i, old_blk_p, old_size, &
                                 add_ip, add_size, new_blk_d, old_blk_d, &
                                 added_size_offset, added_sizes, added_size, added_nblks)
      INTEGER, INTENT(IN)                                :: new_size
      INTEGER, DIMENSION(new_size), INTENT(OUT)          :: new_blk_p, new_col_i, new_row_i
      INTEGER, INTENT(IN)                                :: old_size
      INTEGER, DIMENSION(old_size), INTENT(IN)           :: old_blk_p, old_col_i, old_row_i
      INTEGER, INTENT(IN)                                :: add_size
      INTEGER, DIMENSION(3, add_size), INTENT(IN)        :: add_ip
      INTEGER, DIMENSION(new_size), INTENT(OUT), &
         OPTIONAL                                        :: new_blk_d
      INTEGER, DIMENSION(old_size), INTENT(IN), OPTIONAL :: old_blk_d
      INTEGER, INTENT(IN), OPTIONAL                      :: added_size_offset
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: added_sizes
      INTEGER, INTENT(OUT), OPTIONAL                     :: added_size, added_nblks

      CHARACTER(len=*), PARAMETER :: routineN = 'merge_index_arrays', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: add_blk, bp, i, merge_from_whom, &
                                                            new_blk, old_blk
      LOGICAL                                            :: multidata

!   ---------------------------------------------------------------------------

      multidata = PRESENT(old_blk_d) .AND. PRESENT(new_blk_d)
      IF (old_size+add_size .NE. new_size) &
         CALL cp__w("dbcsr/block/dbcsr_index_operations.F",1060,"Mismatch of new and old size")
      CALL dbcsr_assert(PRESENT(added_size_offset), "EQV", &
                        PRESENT(added_sizes), dbcsr_fatal_level, dbcsr_wrong_args_error, &
                        routineN, "Must specify a set of arguments", 1063)
      CALL dbcsr_assert(PRESENT(added_sizes), "EQV", PRESENT(added_size), &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, &
                        routineN, "Must specify a set of arguments", 1066)
      IF (debug_mod) THEN
         WRITE (*, *) " Old array", old_size
         DO i = 1, old_size
            WRITE (*, '(I7,2X,I7,2X,I7)') old_row_i(i), old_col_i(i), old_blk_p(i)
         ENDDO
         WRITE (*, *) " Add array", add_size
         DO i = 1, add_size
            WRITE (*, '(I7,2X,I7,2X,I7)') add_ip(1:3, i)
         ENDDO
      ENDIF
      IF (PRESENT(added_nblks)) added_nblks = 0
      IF (PRESENT(added_size)) THEN
         added_size = 0
         bp = added_size_offset
      ENDIF
      IF (add_size .GT. 0) THEN
         old_blk = 1
         add_blk = 1
         new_blk = 1
         IF (old_size .EQ. 0) THEN
            new_row_i(1:add_size) = add_ip(1, 1:add_size)
            new_col_i(1:add_size) = add_ip(2, 1:add_size)
            new_blk_p(1:add_size) = add_ip(3, 1:add_size)
            !IF (multidata) new_blk_d(1:add_size) = add_ip(4, 1:add_size)
            IF (PRESENT(added_nblks)) added_nblks = add_size
            IF (PRESENT(added_size)) added_size = SUM(added_sizes)
         ELSE
            DO WHILE (new_blk .LE. new_size)
               merge_from_whom = 0
               IF (old_blk .LE. old_size .AND. add_blk .LE. add_size) THEN
                  IF (add_ip(1, add_blk) .EQ. old_row_i(old_blk) &
                      .AND. add_ip(2, add_blk) .EQ. old_col_i(old_blk)) THEN
                     IF (debug_mod) THEN
                        WRITE (*, *) "Duplicate block! addblk", &
                           add_blk, "oldblk", old_blk
                     ENDIF
                  ENDIF
                  ! Rows come first
                  IF (add_ip(1, add_blk) .LT. old_row_i(old_blk)) THEN
                     merge_from_whom = 2
                  ELSEIF (add_ip(1, add_blk) .GT. old_row_i(old_blk)) THEN
                     merge_from_whom = 1
                  ELSE ! Same rows, so now come the columns
                     IF (add_ip(2, add_blk) .LT. old_col_i(old_blk)) THEN
                        ! Merges from the add array
                        merge_from_whom = 2
                     ELSEIF (add_ip(2, add_blk) .GT. old_col_i(old_blk)) THEN
                        ! Merges from the old array
                        merge_from_whom = 1
                     ELSE
                        ! Merge from old array and skip one in the new array
                        IF (debug_mod) THEN
                           WRITE (*, *) "Duplicate, keeping old", &
                              add_ip(1, add_blk), add_ip(2, add_blk)
                        ENDIF
                        merge_from_whom = 1
                        add_blk = add_blk+1
                     ENDIF
                  ENDIF
               ELSE
                  IF (add_blk .LE. add_size) THEN
                     ! Merges from the add array
                     merge_from_whom = 2
                  ELSEIF (old_blk .LE. old_size) THEN
                     ! Merges from the old array
                     merge_from_whom = 1
                  ELSE
                     ! Hmmm, nothing to merge...
                     merge_from_whom = 0
                     !WRITE(*,*)"Ran out of data to merge"
                  ENDIF
               ENDIF
               SELECT CASE (merge_from_whom)
               CASE (2)
                  ! Merges from the add array
                  new_row_i(new_blk) = add_ip(1, add_blk)
                  new_col_i(new_blk) = add_ip(2, add_blk)
                  new_blk_p(new_blk) = add_ip(3, add_blk)
                  !IF (multidata) new_blk_d(new_blk) = add_ip(4, add_blk)
                  IF (PRESENT(added_nblks)) added_nblks = added_nblks+1
                  IF (PRESENT(added_sizes)) THEN
                     new_blk_p(new_blk) = bp
                     bp = bp+added_sizes(add_blk)
                     added_size = added_size+added_sizes(add_blk)
                  ENDIF
                  add_blk = add_blk+1
               CASE (1)
                  ! Merges from the old array
                  new_row_i(new_blk) = old_row_i(old_blk)
                  new_col_i(new_blk) = old_col_i(old_blk)
                  new_blk_p(new_blk) = old_blk_p(old_blk)
                  IF (multidata) new_blk_p(new_blk) = old_blk_d(old_blk)
                  old_blk = old_blk+1
               CASE DEFAULT
                  !WRITE(*,*)"Nothing to merge"
               END SELECT
               new_blk = new_blk+1
            ENDDO
         ENDIF
      ELSE
         new_row_i(1:old_size) = old_row_i(1:old_size)
         new_col_i(1:old_size) = old_col_i(1:old_size)
         new_blk_p(1:old_size) = old_blk_p(1:old_size)
         IF (multidata) new_blk_d(1:old_size) = old_blk_d(1:old_size)
      ENDIF
      IF (debug_mod) THEN
         WRITE (*, *) " New array"
         DO i = 1, new_size
            WRITE (*, '(4(2X,I7))') new_row_i(i), new_col_i(i), new_blk_p(i)
         ENDDO
      ENDIF
   END SUBROUTINE merge_index_arrays

! **************************************************************************************************
!> \brief Converts BCSR global row index to local row index.
!> \param[in,out] matrix       matrix for which to make canonical index
! **************************************************************************************************
   SUBROUTINE dbcsr_make_index_local_row(matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_index_local_row', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, lrow, nlocal_rows, &
                                                            ntotal_rows, prow, stat
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: local_row_p
      INTEGER, DIMENSION(:), POINTER                     :: local_rows

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      IF (.NOT. ASSOCIATED(matrix%m%row_p)) &
         CALL cp__b("dbcsr/block/dbcsr_index_operations.F",1199,"The row index must be initialized.")
      CALL dbcsr_assert("NOT", matrix%m%bcsc, &
                        dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN, &
                        "Not support for BCSC yet.", 1202)
      !
      prow = matrix%m%index(dbcsr_slot_home_vprow)
      IF (prow .LT. 0) THEN
         prow = matrix%m%index(dbcsr_slot_home_prow)
      ENDIF
      nlocal_rows = matrix%m%nblkrows_local
      ALLOCATE (local_row_p(nlocal_rows+1), stat=stat)
      CALL dbcsr_assert(stat, "EQ", 0, &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Could not allocate memory.", 1212)
      !> The existing row_p is converted from an indexing array into a
      !> counting array.  Because it is later discarded, the counting is
      !> done in-place.
      ntotal_rows = matrix%m%nblkrows_total
      CALL dbcsr_count_row_index(matrix%m%row_p, ntotal_rows)
      ! We first have to find the local rows for the given prow.
      local_rows => array_data(matrix%m%local_rows)
      CALL dbcsr_assert(SIZE(local_rows), "EQ", nlocal_rows, &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Mismatch in the number of local rows.", 1222)
      !> The counts are mapped to local rows,
      DO lrow = 1, nlocal_rows
         local_row_p(lrow) = matrix%m%row_p(local_rows(lrow))
      ENDDO
      CALL dbcsr_assert(SUM(matrix%m%row_p(1:ntotal_rows)), "EQ", &
                        SUM(local_row_p(1:nlocal_rows)), &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Inconsistent row counts. Perhaps non-local rows contain data?.", &
                        1231)
      !> then converted into an index.
      CALL dbcsr_build_row_index(local_row_p, nlocal_rows)
      !> The local row index replaces the global one.
      CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_row_p)
      CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_row_p, DATA=local_row_p)
      !> Finally the matrix is designated as having a local-based index.
      matrix%m%local_indexing = .TRUE.
      IF (careful_mod) THEN
         CALL dbcsr_assert(array_size(matrix%m%local_rows), "EQ", nlocal_rows, &
                           dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                           "Inconsistent local row counts.", &
                           1243)
         CALL dbcsr_assert(array_size(matrix%m%global_rows), "EQ", ntotal_rows, &
                           dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                           "Inconsistent global row counts.", &
                           1247)
         IF (array_size(matrix%m%global_rows) .EQ. 0) THEN
            CALL dbcsr_assert(nlocal_rows, "EQ", 0, &
                              dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                              "Invalid number of local or global rows.", &
                              1252)
         ENDIF
      ENDIF
      DEALLOCATE (local_row_p)
      !
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_make_index_local_row

! **************************************************************************************************
!> \brief Converts BCSR index into list indices (similar to work matrices)
!> \param[in,out] matrix       matrix for which to make canonical index
!> \param[in] thread_redist    make a thread subdistribution
! **************************************************************************************************
   SUBROUTINE dbcsr_make_index_list(matrix, thread_redist)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      LOGICAL, INTENT(IN)                                :: thread_redist

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_make_index_list', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: blk, error_handle, ithread, my_cnt, &
                                                            nblks, nrows, nthreads
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: blki
      INTEGER, DIMENSION(0)                              :: zero_len_array
      INTEGER, DIMENSION(:), POINTER                     :: global_cols, local_rows, td, thr_c

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      IF (.NOT. ASSOCIATED(matrix%m%row_p)) &
         CALL cp__b("dbcsr/block/dbcsr_index_operations.F",1283,"The row index must be initialized.")
      CALL dbcsr_assert("NOT", matrix%m%list_indexing, &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "List index already exists?", 1286)
      CALL dbcsr_assert("NOT", matrix%m%bcsc, &
                        dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN, &
                        "Not support for BCSC yet.", 1289)
      CALL dbcsr_assert(matrix%m%nblks, "EQ", SIZE(matrix%m%col_i), &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Block count mismatch.", 1292)
      CALL dbcsr_assert(matrix%m%nblks, "EQ", SIZE(matrix%m%blk_p), &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Block count mismatch", 1295)
      !
      IF (matrix%m%local_indexing) THEN
         CALL dbcsr_assert(SIZE(matrix%m%row_p)-1, "EQ", matrix%m%nblkrows_local, &
                           dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                           "Local row index incorrectly sized.", 1300)
      ELSE
         CALL dbcsr_assert(SIZE(matrix%m%row_p)-1, "EQ", matrix%m%nblkrows_total, &
                           dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                           "Global row index incorrectly sized", 1304)
      ENDIF
      !
      matrix%m%list_indexing = .TRUE.
      !
      IF (matrix%m%local_indexing) THEN
         global_cols => array_data(matrix%m%global_cols)
         nrows = matrix%m%nblkrows_local
      ELSE
         nrows = matrix%m%nblkrows_total
      ENDIF
      !
      nblks = matrix%m%nblks
      ALLOCATE (blki(3, nblks))
      CALL dbcsr_expand_row_index(matrix%m%row_p, blki(1, :), nrows, nblks)
      IF (matrix%m%local_indexing) THEN
         ! If local indexing is enabled, then the rows but not the
         ! columns are already localized
         IF (dbg) THEN
            WRITE (*, *) routineN//" Making local columns"
            WRITE (*, '(10(1X,i7))') global_cols
            WRITE (*, *) 'local'
            WRITE (*, '(10(1X,i7))') array_data(matrix%m%local_cols)
         ENDIF
         DO blk = 1, nblks
            blki(2, blk) = global_cols(matrix%m%col_i(blk))
            blki(3, blk) = matrix%m%blk_p(blk)
         ENDDO
      ELSE
         IF (dbg) WRITE (*, *) routineN//" Not making local columns"
         DO blk = 1, nblks
            blki(2, blk) = matrix%m%col_i(blk)
            blki(3, blk) = matrix%m%blk_p(blk)
         END DO
      ENDIF
      CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_row_p)
      CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_col_i)
      CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_blk_p)
      nthreads = 0
!$    nthreads = OMP_GET_MAX_THREADS()
      !
      ! Reshuffle according to threads
      IF (nthreads .GT. 0 .AND. thread_redist) THEN
         td => array_data(dbcsr_distribution_thread_dist(dbcsr_distribution(matrix)))
         IF (matrix%m%local_indexing) THEN
            local_rows => array_data(matrix%m%local_rows)
         ENDIF
!$OMP        PARALLEL DEFAULT (NONE) &
!$OMP        PRIVATE (my_cnt, ithread, blk) &
!$OMP        SHARED (td, blki, nthreads, thr_c, nblks, matrix,local_rows)
         !
         ithread = 0
!$       ithread = OMP_GET_THREAD_NUM()
!$OMP        MASTER
!$       nthreads = OMP_GET_NUM_THREADS()
         CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_thr_c, &
                                      reservation=nthreads+1, extra=3*nblks)
         thr_c => matrix%m%thr_c
         CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_coo_l, &
                                      reservation=3*nblks)
!$OMP        END MASTER
!$OMP        BARRIER
         my_cnt = 0
         IF (matrix%m%local_indexing) THEN
            my_cnt = COUNT(td(local_rows(blki(1, :))) .EQ. ithread)
         ELSE
            my_cnt = COUNT(td(blki(1, :)) .EQ. ithread)
         ENDIF
         !DO blk = 1, nblks
         !   IF (td(blki(1, blk)) .EQ. ithread) my_cnt = my_cnt+1
         !ENDDO
         thr_c(ithread+1) = my_cnt
!$OMP        BARRIER
!$OMP        MASTER
         CALL dbcsr_build_row_index_inplace(thr_c, nthreads)
!$OMP        END MASTER
!$OMP        BARRIER
         my_cnt = (thr_c(ithread+1)+1)*3-2
         IF (matrix%m%local_indexing) THEN
            DO blk = 1, nblks
               IF (td(local_rows(blki(1, blk))) .EQ. ithread) THEN
                  matrix%m%coo_l(my_cnt:my_cnt+2) = blki(1:3, blk)
                  my_cnt = my_cnt+3
               ENDIF
            ENDDO
         ELSE
            DO blk = 1, nblks
               IF (td(blki(1, blk)) .EQ. ithread) THEN
                  matrix%m%coo_l(my_cnt:my_cnt+2) = blki(1:3, blk)
                  my_cnt = my_cnt+3
               ENDIF
            ENDDO
         ENDIF
!$OMP        END PARALLEL
      ELSE
         ! Small price to pay for avoiding infinite recursions.
         DO blk = 2, nblks
            IF (blki(1, blk) .EQ. blki(1, blk-1) &
                .AND. blki(2, blk) .EQ. blki(2, blk-1)) &
               ! Weird assertion to give some idea of the two blocks.
               CALL dbcsr_assert(-blki(1, blk), "EQ", blki(2, blk), &
                                 dbcsr_fatal_level, dbcsr_caller_error, routineN, &
                                 "Should not have duplicate matrix blocks. (-row, col) is duplicated.", &
                                 1407)
         ENDDO
         !
         IF (nblks > 0) THEN
            CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_coo_l, &
                                         DATA=RESHAPE(blki, (/3*nblks/)))
         ELSE
            CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_coo_l, &
                                         DATA=zero_len_array)
         ENDIF
      ENDIF

      DEALLOCATE (blki)
      !
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_make_index_list

! **************************************************************************************************
!> \brief Compacts an index.
!> \param[in,out] matrix       matrix for which to make canonical index
! **************************************************************************************************
   SUBROUTINE dbcsr_index_compact(matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_index_compact', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, new_size, size_blk_p, &
                                                            size_col_i, size_coo_l, size_row_p, &
                                                            size_thr_c
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: blk_p, col_i, coo_l, meta, row_p, thr_c
      LOGICAL                                            :: compact, has_blk_p, has_col_i, &
                                                            has_coo_l, has_row_p, has_thr_c

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      !> Ensures the index pointers are set.
      CALL dbcsr_repoint_index(matrix%m)
      ! Check that compaction is even needed.
      has_row_p = ASSOCIATED(matrix%m%row_p)
      IF (has_row_p) THEN
         size_row_p = SIZE(matrix%m%row_p)
      ELSE
         size_row_p = 0
      ENDIF
      has_col_i = ASSOCIATED(matrix%m%col_i)
      IF (has_col_i) THEN
         size_col_i = SIZE(matrix%m%col_i)
      ELSE
         size_col_i = 0
      ENDIF
      has_blk_p = ASSOCIATED(matrix%m%blk_p)
      IF (has_blk_p) THEN
         size_blk_p = SIZE(matrix%m%blk_p)
      ELSE
         size_blk_p = 0
      ENDIF
      has_thr_c = ASSOCIATED(matrix%m%thr_c)
      IF (has_thr_c) THEN
         size_thr_c = SIZE(matrix%m%thr_c)
      ELSE
         size_thr_c = 0
      ENDIF
      has_coo_l = ASSOCIATED(matrix%m%coo_l)
      IF (has_coo_l) THEN
         size_coo_l = SIZE(matrix%m%coo_l)
      ELSE
         size_coo_l = 0
      ENDIF
      !
      new_size = dbcsr_num_slots+ &
                 size_row_p+size_col_i+size_blk_p+size_thr_c+size_coo_l
      compact = new_size .LT. SIZE(matrix%m%index)
      IF (compact) THEN
         ! Store old index arrays.
         IF (has_row_p) THEN
            ALLOCATE (row_p(size_row_p))
            row_p(:) = matrix%m%row_p(:)
         ENDIF
         IF (has_col_i) THEN
            ALLOCATE (col_i(size_col_i))
            col_i(:) = matrix%m%col_i(:)
         ENDIF
         IF (has_blk_p) THEN
            ALLOCATE (blk_p(size_blk_p))
            blk_p(:) = matrix%m%blk_p(:)
         ENDIF
         IF (has_thr_c) THEN
            ALLOCATE (thr_c(size_thr_c))
            thr_c(:) = matrix%m%thr_c(:)
         ENDIF
         IF (has_coo_l) THEN
            ALLOCATE (coo_l(size_coo_l))
            coo_l(:) = matrix%m%coo_l(:)
         ENDIF
         ALLOCATE (meta(dbcsr_num_slots))
         meta(:) = matrix%m%index(1:dbcsr_num_slots)
         ! Clear the index.
         CALL memory_deallocate(matrix%m%index, &
                                dbcsr_get_index_memory_type(matrix))
         NULLIFY (matrix%m%index)
         CALL memory_allocate(matrix%m%index, new_size, &
                              dbcsr_get_index_memory_type(matrix))
         !
         ! Now copy the old index arrays into the index. We must not
         ! copy the positions of the old pointers.
         matrix%m%index(1:dbcsr_meta_size) = meta(1:dbcsr_meta_size)
         matrix%m%index(dbcsr_meta_size+1:) = 0
         matrix%m%index(dbcsr_slot_size) = dbcsr_num_slots
         IF (has_thr_c) THEN
            CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_thr_c, thr_c)
            DEALLOCATE (thr_c)
         ENDIF
         IF (has_row_p) THEN
            CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_row_p, row_p)
            DEALLOCATE (row_p)
         ENDIF
         IF (has_col_i) THEN
            CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_col_i, col_i)
            DEALLOCATE (col_i)
         ENDIF
         IF (has_blk_p) THEN
            CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_blk_p, blk_p)
            DEALLOCATE (blk_p)
         ENDIF
         IF (has_coo_l) THEN
            CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_coo_l, coo_l)
            DEALLOCATE (coo_l)
         ENDIF
         DEALLOCATE (meta)
         IF (careful_mod) THEN
            ! This is pretty strong but it should be true.
            CALL dbcsr_assert(matrix%m%index(dbcsr_slot_size), "EQ", new_size, &
                              dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                              "Unexpected index size.", 1542)
            CALL dbcsr_assert(SIZE(matrix%m%index), "EQ", new_size, &
                              dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                              "Unexpected index size.", 1545)
         ENDIF
         CALL dbcsr_repoint_index(matrix%m)
      ENDIF
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_index_compact

! **************************************************************************************************
!> \brief Calculates the checksum of an index.
!> \param[in] matrix           matrix for which to make canonical index
!> \param checksum ...
! **************************************************************************************************
   SUBROUTINE dbcsr_index_checksum(matrix, checksum)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      INTEGER, INTENT(OUT)                               :: checksum

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_index_checksum', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      !
      checksum = joaat_hash((/joaat_hash(matrix%m%row_p), &
                              joaat_hash(matrix%m%col_i), &
                              joaat_hash(matrix%m%blk_p)/))
      !
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_index_checksum

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \retval local_indexing ...
! **************************************************************************************************
   FUNCTION dbcsr_has_local_row_index(matrix) RESULT(local_indexing)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      LOGICAL                                            :: local_indexing

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_has_local_row_index', &
         routineP = moduleN//':'//routineN

      local_indexing = matrix%m%local_indexing
   END FUNCTION dbcsr_has_local_row_index

END MODULE dbcsr_index_operations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR sparse matrix utility routines
!> \author  Urban Borstnik
!> \date    2009-05-12
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - Created 2009-05-12
! **************************************************************************************************
MODULE dbcsr_util

   USE array_types,                     ONLY: array_data
   USE dbcsr_data_methods,              ONLY: dbcsr_data_get_size,&
                                              dbcsr_data_get_size_referenced,&
                                              dbcsr_get_data
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_local_cols,&
                                              dbcsr_distribution_local_rows,&
                                              dbcsr_distribution_mp,&
                                              dbcsr_distribution_ncols,&
                                              dbcsr_distribution_nlocal_cols,&
                                              dbcsr_distribution_nlocal_rows,&
                                              dbcsr_distribution_nrows
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_methods,                   ONLY: dbcsr_blk_col_offset,&
                                              dbcsr_blk_row_offset,&
                                              dbcsr_has_symmetry,&
                                              dbcsr_is_initialized,&
                                              dbcsr_valid_index
   USE dbcsr_mp_methods,                ONLY: dbcsr_mp_group,&
                                              dbcsr_mp_mypcol,&
                                              dbcsr_mp_myprow
   USE dbcsr_toollib,                   ONLY: sort,&
                                              swap
   USE dbcsr_types,                     ONLY: &
        dbcsr_distribution_obj, dbcsr_meta_size, dbcsr_num_slots, dbcsr_obj, dbcsr_slot_blk_p, &
        dbcsr_slot_col_i, dbcsr_slot_dense, dbcsr_slot_home_coli, dbcsr_slot_home_pcol, &
        dbcsr_slot_home_prow, dbcsr_slot_home_rowi, dbcsr_slot_home_vpcol, dbcsr_slot_home_vprow, &
        dbcsr_slot_nblkcols_local, dbcsr_slot_nblkcols_total, dbcsr_slot_nblkrows_local, &
        dbcsr_slot_nblkrows_total, dbcsr_slot_nblks, dbcsr_slot_nfullcols_local, &
        dbcsr_slot_nfullcols_total, dbcsr_slot_nfullrows_local, dbcsr_slot_nfullrows_total, &
        dbcsr_slot_nze, dbcsr_slot_row_p, dbcsr_slot_type, dbcsr_type, dbcsr_type_complex_4, &
        dbcsr_type_complex_8, dbcsr_type_real_4, dbcsr_type_real_8
   USE kinds,                           ONLY: dp,&
                                              int_4,&
                                              int_8,&
                                              real_4,&
                                              real_8
   USE message_passing,                 ONLY: mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

  !$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads
  IMPLICIT NONE
  PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_util'

  ! Main
  PUBLIC :: dbcsr_checksum, dbcsr_verify_matrix,&
            dbcsr_pack_meta, dbcsr_unpack_meta, meta_from_dist
  ! Block sizes and arrays
  PUBLIC :: convert_sizes_to_offsets, convert_offsets_to_sizes,&
            global_offsets_to_local,&
            nfull_elements,&
            dbcsr_calc_block_sizes,&
            find_block_of_element,&
            get_internal_offsets,&
            map_most_common
  ! utility routines
  PUBLIC :: count_bins, sgn


  LOGICAL, PARAMETER :: bcsr_info =    .FALSE.
  LOGICAL, PARAMETER :: bcsr_verbose = .FALSE.

CONTAINS


! **************************************************************************************************
!> \brief Finds block to which a full element belongs.
!> \param[in] full            full element
!> \param[out] block          block to which full belongs
!> \param nblocks ...
!> \param[in] block_offsets   starting full elements of blocks
!> \param[in] hint            where to start looking; ignored if 0
!> \par Assumptions
!>        It is assumed that block_start and block_end are sorted and
!>        that hint is in the range [0, nblocks].
! **************************************************************************************************
  SUBROUTINE find_block_of_element(full, block, nblocks,&
       block_offsets, hint)
      INTEGER, INTENT(in)                                :: full
      INTEGER, INTENT(out)                               :: block
      INTEGER, INTENT(in)                                :: nblocks
      INTEGER, DIMENSION(1:nblocks+1), INTENT(in)        :: block_offsets
      INTEGER, INTENT(in)                                :: hint

      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: count

    IF (hint .NE. 0) THEN
       block = hint
    ELSE
       block = MAX(1,(nblocks+1)/2)
    ENDIF
    count = 0
    DO WHILE (block_offsets(block) .GT. full .OR. block_offsets(block+1)-1 .LT. full)
       IF (block_offsets(block) .GT. full) THEN
          block = block-1
       ELSEIF (block_offsets(block+1)-1 .LT. full) THEN
          block = block+1
       ENDIF
       count = count+1
       IF (dbg) THEN
          IF (count .GT. nblocks .OR. block .LT. 1 .OR. block .GT. nblocks) THEN
             WRITE(*,'(1X,A,I9,A,I9,A)')"Want to find block",&
                  block," of",nblocks," blocks"
             IF(count .GT. nblocks)&
                CALL cp__b("dbcsr/dist/dbcsr_util.F",127,"Too many searches")
          ENDIF
       ENDIF
    ENDDO
  END SUBROUTINE find_block_of_element



! **************************************************************************************************
!> \brief The sum of a subset of rows/columns
!> \param[in] all_offsets     ordered offsets of all the elements
!> \param[in] local_elements  enumerated local elements
!> \retval nfull_elements     sum of sizes of local elemetns 
!> \note Used for making matrices dense/undense
! **************************************************************************************************
  PURE FUNCTION nfull_elements (all_offsets, local_elements)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: all_offsets, local_elements
      INTEGER                                            :: nfull_elements

      INTEGER                                            :: el, lel

    nfull_elements = 0
    DO lel = 1, SIZE(local_elements)
       el = local_elements(lel)
       nfull_elements = nfull_elements + all_offsets(el+1) - all_offsets(el)
    ENDDO
  END FUNCTION nfull_elements



! **************************************************************************************************
!> \brief Converts sizes to offsets
!>
!> \param[in] sizes           array with sizes
!> \param[out] offsets_start  offsets of starts
!> \param[out] offsets_stop   (optional) offsets of ends
! **************************************************************************************************

  PURE SUBROUTINE convert_sizes_to_offsets (sizes,&
       offsets_start, offsets_stop)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: sizes
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: offsets_start
      INTEGER, DIMENSION(:), INTENT(OUT), OPTIONAL       :: offsets_stop

      CHARACTER(len=*), PARAMETER :: routineN = 'convert_sizes_to_offsets', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n

!   ---------------------------------------------------------------------------

    n = SIZE (sizes)
    IF (n .GT. 0) THEN
       offsets_start(1) = 1
       IF (PRESENT (offsets_stop)) offsets_stop(1) = sizes(1)
       IF (.NOT. PRESENT (offsets_stop)) THEN
          DO i = 2, n
             offsets_start(i) = offsets_start(i-1) + sizes(i-1)
          ENDDO
          IF(SIZE(offsets_start).GT.n) &
               offsets_start(n+1) = offsets_start(n) + sizes(n)
       ELSE
          DO i = 2, n
             offsets_start(i) = offsets_start(i-1) + sizes(i-1)
             offsets_stop(i) = offsets_stop(i-1) + sizes(i)
          ENDDO
          IF(SIZE(offsets_start).GT.n) &
               offsets_start(n+1) = offsets_start(n) + sizes(n)
       ENDIF
    ELSE
       IF (.NOT. PRESENT (offsets_stop)) THEN
          offsets_start(1) = 0
       ENDIF
    ENDIF
  END SUBROUTINE convert_sizes_to_offsets

! **************************************************************************************************
!> \brief Converts offsets to sizes
!>
!> If the offsets of ends are not given, then the array of sizes is assumed
!> to be one greater than the desired sizes.
!>
!> \param[in] offsets_start  offsets of starts
!> \param[out] sizes         array with sizes
!> \param[in] offsets_stop   (optional) offsets of ends
! **************************************************************************************************

  PURE SUBROUTINE convert_offsets_to_sizes (offsets_start, sizes, offsets_stop)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: offsets_start
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: sizes
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: offsets_stop

      CHARACTER(len=*), PARAMETER :: routineN = 'convert_offsets_to_sizes', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, n

!   ---------------------------------------------------------------------------

    n = SIZE (offsets_start)
    IF (PRESENT (offsets_stop)) THEN
       sizes(:) = offsets_stop(:) - offsets_start(:) + 1
    ELSE
       IF (n .GT. 1) THEN
          DO i = 1, n-1
             sizes(i) = sizes(i+1) - sizes(i)
          ENDDO
       ENDIF
    ENDIF
  END SUBROUTINE convert_offsets_to_sizes

! **************************************************************************************************
!> \brief Converts global offsets to local
!> \param[in] global_offsets   Offsets of elements in the global grid
!> \param[in] local_elements   Which elements are local
!> \param[out] local_offsets   Offsets of local elements.
!> \par Global vs. Local Indexing
!>      local_offsets may be sized according to the
!>      local index (|local_elements+|1) or the
!>      global index (|global_offsets|).
! **************************************************************************************************
  SUBROUTINE global_offsets_to_local (global_offsets,&
       local_elements, local_offsets)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: global_offsets, local_elements
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: local_offsets

      CHARACTER(len=*), PARAMETER :: routineN = 'global_offsets_to_local', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: acc, el, lel, nglobal, nlo, nlocal, &
                                                            prev_el, sz
      LOGICAL                                            :: local

!   ---------------------------------------------------------------------------

    nglobal = SIZE (global_offsets)-1
    nlocal = SIZE (local_elements)
    nlo = SIZE (local_offsets)-1
    local = .NOT. (nglobal .EQ. nlo)
    IF (local) THEN
       CALL dbcsr_assert (nlocal, "EQ", nlo,&
            dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
            "Invalid size for local offsets",269)
    ENDIF
    IF (local) THEN
       acc = 1
       DO lel = 1, nlocal
          local_offsets(lel) = acc
          el = local_elements(lel)
          sz = global_offsets(el+1)-global_offsets(el)
          acc = acc + sz
       ENDDO
       local_offsets(nlocal+1) = acc
    ELSE
       acc = 1
       prev_el=0
       DO lel = 1, nlocal
          el = local_elements(lel)
          local_offsets(prev_el+1:el) = acc
          sz = global_offsets(el+1)-global_offsets(el)
          acc = acc + sz
          prev_el = el
       ENDDO
       local_offsets(prev_el+1:nglobal+1) = acc
    ENDIF
  END SUBROUTINE global_offsets_to_local

! **************************************************************************************************
!> \brief Finds internal offsets
!>
!> For all local blocks in blk_local_els, it calculates its offset in
!> the dense block to which it belongs.
!> \param blk_local_els ...
!> \param el_map ...
!> \param blk_el_offsets ...
!> \param dense_el_offsets ...
!> \param internal_offsets ...
! **************************************************************************************************
  SUBROUTINE get_internal_offsets(blk_local_els, el_map, blk_el_offsets,&
       dense_el_offsets, internal_offsets)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: blk_local_els, el_map, blk_el_offsets, &
                                                            dense_el_offsets
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: internal_offsets

      CHARACTER(len=*), PARAMETER :: routineN = 'get_internal_offsets', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk_el, d_el, i, ndense, nlblk
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: off_acc

!   ---------------------------------------------------------------------------

    nlblk = SIZE(blk_local_els)
    ndense = SIZE(dense_el_offsets)
    ALLOCATE (off_acc (ndense))
    off_acc(:) = 0
    internal_offsets(:) = 0
    DO i = 1, nlblk
       blk_el = blk_local_els(i)
       d_el = el_map(blk_el)
       internal_offsets(blk_el) = off_acc(d_el)
       off_acc(d_el) = off_acc(d_el) + blk_el_offsets(blk_el+1) - blk_el_offsets(blk_el)
    ENDDO
    DEALLOCATE (off_acc)
  END SUBROUTINE get_internal_offsets

! **************************************************************************************************
!> \brief Calculates explicit sizes for all data blocks.
!> \param[out] sizes          sizes of all data blocks
!> \param[in] row_p index structure
!> \param[in] col_i index structure
!> \param[in] rbs row block sizes
!> \param[in] cbs column block sizes
! **************************************************************************************************
  SUBROUTINE dbcsr_calc_block_sizes(sizes, row_p, col_i, rbs, cbs)
      INTEGER, DIMENSION(*), INTENT(OUT)                 :: sizes
      INTEGER, DIMENSION(:), INTENT(IN)                  :: row_p
      INTEGER, DIMENSION(*), INTENT(IN)                  :: col_i, rbs, cbs

      INTEGER                                            :: blk, nrows, row, row_size

    nrows = SIZE(row_p)-1
    !$OMP DO
    DO row = 1, nrows
       row_size = rbs(row)
       DO blk = row_p(row)+1, row_p(row+1)
          sizes(blk) = row_size * cbs(col_i(blk))
       END DO
    ENDDO
    !$OMP END DO
  END SUBROUTINE dbcsr_calc_block_sizes


! **************************************************************************************************
!> \brief ...
!> \param n ...
!> \param oldsign ...
!> \param x ...
!> \retval val ...
! **************************************************************************************************
  ELEMENTAL FUNCTION sgn (n, oldsign, x) RESULT (val)
      INTEGER, INTENT(IN)                                :: n, oldsign
      LOGICAL, INTENT(IN)                                :: x
      INTEGER                                            :: val

    IF (.NOT.x) THEN
       val = SIGN (n, oldsign)
    ELSE
       val = -SIGN (n, oldsign)
    ENDIF
  END FUNCTION sgn


! **************************************************************************************************
!> \brief Fills meta information from a given distribution_2d
!> \param[out] meta           meta information array to fill
!> \param dist                processor distribution
!> \param[in] row_blk_size    row block sizes
!> \param[in] col_blk_size    column block sizes
! **************************************************************************************************
  SUBROUTINE meta_from_dist (meta, dist, row_blk_size, col_blk_size)
      INTEGER, DIMENSION(dbcsr_meta_size), INTENT(OUT)   :: meta
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: row_blk_size, col_blk_size

      CHARACTER(len=*), PARAMETER :: routineN = 'meta_from_dist', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, nfullcols_local, nfullcols_total, &
                                                            nfullrows_local, nfullrows_total
      INTEGER, DIMENSION(:), POINTER                     :: blkcols_local, blkrows_local

!   ---------------------------------------------------------------------------

    blkrows_local => dbcsr_distribution_local_rows (dist)
    blkcols_local => dbcsr_distribution_local_cols (dist)
    nfullrows_total = SUM (row_blk_size)
    nfullcols_total = SUM (col_blk_size)
    nfullrows_local = 0
    nfullcols_local = 0
    DO i = 1, dbcsr_distribution_nlocal_rows (dist)
       nfullrows_local = nfullrows_local + row_blk_size(blkrows_local(i))
    ENDDO
    DO i = 1, dbcsr_distribution_nlocal_cols (dist)
       nfullcols_local = nfullcols_local + col_blk_size(blkcols_local(i))
    ENDDO
    meta(:) = 0
    meta(5)  = dbcsr_distribution_nrows (dist)
    meta(6)  = dbcsr_distribution_ncols (dist)
    meta(7)  = nfullrows_total
    meta(8)  = nfullcols_total
    meta(9)  = dbcsr_distribution_nlocal_rows (dist)
    meta(10) = dbcsr_distribution_nlocal_cols (dist)
    meta(11) = nfullrows_local
    meta(12) = nfullcols_local
    meta(dbcsr_slot_home_prow) = dbcsr_mp_myprow (dbcsr_distribution_mp (dist))
    meta(dbcsr_slot_home_rowi) = 1
    meta(dbcsr_slot_home_pcol) = dbcsr_mp_mypcol (dbcsr_distribution_mp (dist))
    meta(dbcsr_slot_home_coli) = 1
    meta(dbcsr_slot_home_vprow) = -1
    meta(dbcsr_slot_home_vpcol) = -1
  END SUBROUTINE meta_from_dist


! **************************************************************************************************
!> \brief Copies metadata into an array.
!> \param[in] matrix      Matrix
!> \param[out] meta       Metadata elements
! **************************************************************************************************
  SUBROUTINE dbcsr_pack_meta(matrix, meta)
      TYPE(dbcsr_type), INTENT(IN)                       :: matrix
      INTEGER, DIMENSION(dbcsr_meta_size), INTENT(OUT)   :: meta

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_pack_meta', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    meta(dbcsr_slot_nblks)  = matrix%nblks
    meta(dbcsr_slot_nze)    = matrix%nze
    meta(dbcsr_slot_nblkrows_total ) = matrix%nblkrows_total
    meta(dbcsr_slot_nblkcols_total ) = matrix%nblkcols_total
    meta(dbcsr_slot_nfullrows_total) = matrix%nfullrows_total
    meta(dbcsr_slot_nfullcols_total) = matrix%nfullcols_total
    meta(dbcsr_slot_nblkrows_local ) = matrix%nblkrows_local
    meta(dbcsr_slot_nblkcols_local ) = matrix%nblkcols_local
    meta(dbcsr_slot_nfullrows_local) = matrix%nfullrows_local
    meta(dbcsr_slot_nfullcols_local) = matrix%nfullcols_local
    meta(dbcsr_slot_dense) = 0
    meta(dbcsr_slot_type) = 0
    !IF (matrix%transpose)&
    !     meta(dbcsr_slot_type) = IBSET (meta(dbcsr_slot_type), 0)
    IF (matrix%symmetry)&
         meta(dbcsr_slot_type) = IBSET (meta(dbcsr_slot_type), 1)
    IF (matrix%negate_real)&
         meta(dbcsr_slot_type) = IBSET (meta(dbcsr_slot_type), 2)
    IF (matrix%negate_imaginary)&
         meta(dbcsr_slot_type) = IBSET (meta(dbcsr_slot_type), 3)
  END SUBROUTINE dbcsr_pack_meta

! **************************************************************************************************
!> \brief Sets metadata form an array.
!> \param[in,out] matrix  Matrix
!> \param[in] meta        Metadata elements
! **************************************************************************************************
  SUBROUTINE dbcsr_unpack_meta(matrix, meta)
      TYPE(dbcsr_type), INTENT(INOUT)                    :: matrix
      INTEGER, DIMENSION(dbcsr_meta_size), INTENT(IN)    :: meta

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_unpack_meta', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    matrix%nblks = meta(dbcsr_slot_nblks)
    matrix%nze   = meta(dbcsr_slot_nze)
    matrix%nblkrows_total  = meta(dbcsr_slot_nblkrows_total )
    matrix%nblkcols_total  = meta(dbcsr_slot_nblkcols_total )
    matrix%nfullrows_total = meta(dbcsr_slot_nfullrows_total)
    matrix%nfullcols_total = meta(dbcsr_slot_nfullcols_total)
    matrix%nblkrows_local  = meta(dbcsr_slot_nblkrows_local )
    matrix%nblkcols_local  = meta(dbcsr_slot_nblkcols_local )
    matrix%nfullrows_local = meta(dbcsr_slot_nfullrows_local)
    matrix%nfullcols_local = meta(dbcsr_slot_nfullcols_local)
    matrix%index(dbcsr_slot_dense) = 0
    !matrix%transpose = BTEST (meta(dbcsr_slot_type), 0)
    matrix%symmetry = BTEST (meta(dbcsr_slot_type), 1)
    matrix%negate_real = BTEST (meta(dbcsr_slot_type), 2)
    matrix%negate_imaginary = BTEST (meta(dbcsr_slot_type), 3)
  END SUBROUTINE dbcsr_unpack_meta

! **************************************************************************************************
!> \brief Calculates the checksum of a DBCSR matrix.
!> \param[in] matrix          matrix
!> \param[in] local           (optional) no global communication
!> \param[in] pos             (optional) position-dependent checksum
!> \retval checksum calculated checksum
! **************************************************************************************************
  FUNCTION dbcsr_checksum(matrix, local, pos) RESULT(checksum)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: local, pos
      REAL(KIND=dp)                                      :: checksum

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_checksum', routineP = moduleN//':'//routineN

      COMPLEX(KIND=real_4), DIMENSION(:), POINTER        :: c_sp
      COMPLEX(KIND=real_8), DIMENSION(:), POINTER        :: c_dp
      INTEGER                                            :: bc, blk, blk_p, br, co, handle, m, mn, &
                                                            n, ro
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, row_blk_size
      LOGICAL                                            :: nocomm, pd, tr
      REAL(KIND=dp)                                      :: blk_cs, local_cs, local_cs_row
      REAL(KIND=real_4), DIMENSION(:), POINTER           :: r_sp
      REAL(KIND=real_8), DIMENSION(:), POINTER           :: r_dp

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, handle)
   IF(.NOT.dbcsr_valid_index(matrix))&
      CALL cp__b("dbcsr/dist/dbcsr_util.F",525,"Invalid matrix.")
    nocomm = .FALSE.
    IF (PRESENT (local)) nocomm = local
    IF (PRESENT (pos)) THEN
       pd = pos
    ELSE
       pd = .FALSE.
    ENDIF
    row_blk_size => array_data (matrix%m%row_blk_size)
    col_blk_size => array_data (matrix%m%col_blk_size)
    local_cs = 0.0_dp
    SELECT CASE (matrix%m%data_type)
    CASE (dbcsr_type_real_8)
       CALL dbcsr_get_data (matrix%m%data_area, r_dp)
    CASE (dbcsr_type_real_4)
       CALL dbcsr_get_data (matrix%m%data_area, r_sp)
    CASE (dbcsr_type_complex_8)
       CALL dbcsr_get_data (matrix%m%data_area, c_dp)
    CASE (dbcsr_type_complex_4)
       CALL dbcsr_get_data (matrix%m%data_area, c_sp)
    END SELECT
    DO br = 1, matrix%m%nblkrows_total
       m = row_blk_size(br)
       ro = dbcsr_blk_row_offset (matrix, br)
       local_cs_row=0
       !$OMP PARALLEL DO DEFAULT(NONE) &
       !$OMP          PRIVATE(bc,m,n,mn,blk_p,blk_cs,tr,co) &
       !$OMP          SHARED(pd,br,matrix,ro,row_blk_size,col_blk_size,r_dp, r_sp, c_dp,c_sp) &
       !$OMP           REDUCTION(+:local_cs_row)
       DO blk = matrix%m%row_p(br)+1, matrix%m%row_p(br+1)
          bc = matrix%m%col_i(blk)
          m = row_blk_size(br)
          n = col_blk_size(bc)
          mn = m*n
          blk_p = ABS(matrix%m%blk_p(blk))
          tr = matrix%m%blk_p(blk) .LT. 0
          IF (blk_p .NE. 0) THEN
             IF (mn .GT. 0) THEN
                IF (tr) CALL swap (m, n)
                co = dbcsr_blk_col_offset (matrix, bc)
                ! Calculate DDOT
                SELECT CASE (matrix%m%data_type)
                CASE (dbcsr_type_real_8)
                   IF (pd) THEN
                      blk_cs = pd_blk_cs (m, n, r_dp(blk_p:blk_p+mn-1),&
                           tr, ro, co)
                   ELSE
                      blk_cs = REAL(DOT_PRODUCT(r_dp(blk_p:blk_p+mn-1),&
                                                r_dp(blk_p:blk_p+mn-1)), KIND=dp)
                   ENDIF
                CASE (dbcsr_type_real_4)
                   IF (pd) THEN
                      blk_cs = pd_blk_cs (m,n, REAL(r_sp(blk_p:blk_p+mn-1), KIND=dp),&
                           tr, ro, co)
                   ELSE
                      blk_cs = REAL(DOT_PRODUCT(r_sp(blk_p:blk_p+mn-1),&
                                                r_sp(blk_p:blk_p+mn-1)), KIND=dp)
                   ENDIF
                CASE (dbcsr_type_complex_8)
                   IF (pd) THEN
                      blk_cs = pd_blk_cs (m, n, REAL(c_dp(blk_p:blk_p+mn-1), KIND=dp),&
                           tr, ro, co)
                   ELSE
                      blk_cs = REAL(DOT_PRODUCT(c_dp(blk_p:blk_p+mn-1),&
                                                c_dp(blk_p:blk_p+mn-1)), KIND=dp)
                   ENDIF
                CASE (dbcsr_type_complex_4)
                   IF (pd) THEN
                      blk_cs = pd_blk_cs (m, n, REAL(c_sp(blk_p:blk_p+mn-1), KIND=dp),&
                           tr, ro, co)
                   ELSE
                      blk_cs = REAL(DOT_PRODUCT(c_sp(blk_p:blk_p+mn-1),&
                                                c_sp(blk_p:blk_p+mn-1)), KIND=dp)
                   ENDIF
                CASE default
                   blk_cs = 0.0_dp
                END SELECT
             ELSE
                blk_cs = 0.0_dp
             ENDIF
             local_cs_row = local_cs_row + blk_cs
          ENDIF
       ENDDO
       local_cs=local_cs+local_cs_row
    ENDDO
    checksum = local_cs
    IF (.NOT. nocomm) THEN
       CALL mp_sum(local_cs, dbcsr_mp_group (dbcsr_distribution_mp (&
            matrix%m%dist)))
       checksum = local_cs
    ENDIF
    CALL timestop(handle)
  END FUNCTION dbcsr_checksum

! **************************************************************************************************
!> \brief ...
!> \param ld ...
!> \param od ...
!> \param DATA ...
!> \param tr ...
!> \param ro ...
!> \param co ...
!> \retval pd_cs ...
! **************************************************************************************************
  PURE FUNCTION pd_blk_cs (ld, od, DATA, tr, ro, co) RESULT (pd_cs)
      INTEGER, INTENT(IN)                                :: ld, od
      REAL(KIND=dp), DIMENSION(ld, od), INTENT(IN)       :: DATA
      LOGICAL, INTENT(IN)                                :: tr
      INTEGER, INTENT(IN)                                :: ro, co
      REAL(KIND=dp)                                      :: pd_cs

      INTEGER                                            :: c, cs, r, rs

    pd_cs = 0.0_dp
    rs = ld ; cs = od
    IF (tr) THEN
       CALL swap (rs, cs)
       DO r = 1, rs
          DO c = 1, cs
             pd_cs = pd_cs + DATA(c, r)*LOG(ABS(REAL((ro+r-1),KIND=dp)*REAL((co+c-1), KIND=dp)))
          ENDDO
       ENDDO
    ELSE
       DO c = 1, cs
          DO r = 1, rs
             pd_cs = pd_cs + DATA(r, c)*LOG(ABS(REAL((ro+r-1),KIND=dp)*REAL((co+c-1), KIND=dp)))
          ENDDO
       ENDDO
    ENDIF
  END FUNCTION pd_blk_cs


! **************************************************************************************************
!> \brief Verify the correctness of a BCSR matrix.
!> \param[in] m     bcsr matrix
!> \param[in] verbosity     how detailed errors are; 0=nothing; 1=summary at
!>                          end if matrix not consistent; 2=also individual
!>                          errors; 3=always print info about matrix; >3=even
!>                          more info
!> \param[in] local         (optional) no global communication
! **************************************************************************************************
  SUBROUTINE dbcsr_verify_matrix(m, verbosity, local)
      TYPE(dbcsr_obj), INTENT(IN)                        :: m
      INTEGER, INTENT(IN), OPTIONAL                      :: verbosity
      LOGICAL, INTENT(IN), OPTIONAL                      :: local

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_verify_matrix', r = moduleN//':'//routineN

      INTEGER                                            :: bc, blk, blk_p, br, &
                                                            data_size_referenced, dbg, handle, i, &
                                                            mb, mn, n, n_have_blocks_local, &
                                                            n_have_blocks_total, prev_br
      INTEGER(KIND=int_8)                                :: n_full_blocks_total
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, row_blk_size
      LOGICAL                                            :: nocomm
      REAL(KIND=dp)                                      :: sparsity_total

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, handle)
    IF(.NOT.dbcsr_is_initialized(m))&
       CALL cp__b("dbcsr/dist/dbcsr_util.F",686,"Can not verify non-existing matrix object.")
    dbg = 2
    nocomm = .FALSE.
    IF (PRESENT (local)) nocomm = local
    IF (PRESENT (verbosity)) dbg = verbosity
    IF (dbg.GE.3) WRITE(*,'(1X,A,A,A,3(L1))')r//'Matrix name: ', m%m%name,&
         " of types ", m%m%symmetry, m%m%negate_real,&
         m%m%negate_imaginary
    IF (dbg.GE.3) THEN
       WRITE(*,'(1X,A,I5,"x",I5,A,I5,"x",I5)')r//' Size blocked',&
            m%m%nblkrows_total, m%m%nblkcols_total, ", full ",&
            m%m%nfullrows_total, m%m%nfullcols_total
    ENDIF
    row_blk_size => array_data (m%m%row_blk_size)
    col_blk_size => array_data (m%m%col_blk_size)
    !
    IF(.NOT.dbcsr_has_symmetry (m)) THEN
       n_full_blocks_total = INT(m%m%nblkrows_total,KIND=int_8) * INT(m%m%nblkcols_total,KIND=int_8)
    ELSE
       CALL dbcsr_assert(m%m%nblkrows_total, "EQ", m%m%nblkcols_total,&
            dbcsr_fatal_level, dbcsr_internal_error, r,&
            'Symmetric matrix is not square',707)
       n_full_blocks_total = INT(m%m%nblkrows_total,KIND=int_8) * (m%m%nblkrows_total+1)/2
    ENDIF
    n_have_blocks_local = m%m%nblks
2045 FORMAT (I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5,1X,I5)
2047 FORMAT (I7,1X,I7,1X,I7,1X,I7,1X,I7,1X,I7,1X,I7,1X,I7,1X,I7,1X,I7)
    IF (dbg.GE.4) THEN
       WRITE(*,'(1X,A)')r//' index='
       WRITE(*,2045)m%m%index(:dbcsr_num_slots)
    ENDIF
    CALL dbcsr_assert(m%m%index(1), "GT", 0,&
         dbcsr_fatal_level, dbcsr_internal_error, r,&
         'Index size 0',719)
    DO i = dbcsr_slot_row_p, dbcsr_num_slots
       !CALL dbcsr_assert(m%m%index(i), "GT", 0,&
       !     dbcsr_fatal_level, dbcsr_internal_error, r,&
       !     'Index member is 0',723)
       IF (.NOT. (i.EQ.dbcsr_slot_col_i.OR.i.EQ.dbcsr_slot_blk_p)) THEN
            CALL dbcsr_assert(m%m%index(i), "LE", m%m%index(1),&
                 dbcsr_fatal_level, dbcsr_internal_error, r,&
                 'Index member is greater than size',727)
         ENDIF
    ENDDO
    !
    IF(dbg.GE.4) WRITE(*,*)r//' row_p extents',m%m%index(dbcsr_slot_row_p+1),&
         m%m%index(dbcsr_slot_row_p), SIZE(m%m%row_p)
    CALL dbcsr_assert(m%m%index(dbcsr_slot_row_p+1)-m%m%index(dbcsr_slot_row_p)+1,&
         "EQ", m%m%nblkrows_total+1,&
         dbcsr_fatal_level, dbcsr_internal_error, r,&
         'Size of row_p index inconsistent with number of rows',736)
    CALL dbcsr_assert(SIZE(m%m%row_p), "EQ", m%m%nblkrows_total+1,&
         dbcsr_fatal_level, dbcsr_internal_error, r,&
         'Size of row_p inconsistent with number of rows',739)
    !
    IF(dbg.GE.4) WRITE(*,*)r//' col_i extents',m%m%index(dbcsr_slot_col_i+1),&
         m%m%index(dbcsr_slot_col_i), SIZE(m%m%col_i)
    CALL dbcsr_assert(m%m%index(dbcsr_slot_col_i+1)-m%m%index(dbcsr_slot_col_i)+1,&
         "EQ", m%m%nblks,&
         dbcsr_fatal_level, dbcsr_internal_error, r,&
         'Size of col_i index inconsistent with number of blocks',746)
    CALL dbcsr_assert(SIZE(m%m%col_i), "EQ", m%m%nblks,&
         dbcsr_fatal_level, dbcsr_internal_error, r,&
         'Size of col inconsistent with number of blocks',749)
    !
    IF(dbg.GE.4) WRITE(*,*)r//' blk_p extents',m%m%index(dbcsr_slot_blk_p+1),&
         m%m%index(dbcsr_slot_blk_p), SIZE(m%m%blk_p)
    CALL dbcsr_assert(m%m%index(dbcsr_slot_blk_p+1)-m%m%index(dbcsr_slot_blk_p)+1,&
         "EQ", m%m%nblks,&
         dbcsr_fatal_level, dbcsr_internal_error, r,&
         'Size of blk_p index inconsistent with number of blocks',756)
    CALL dbcsr_assert(SIZE(m%m%col_i), "EQ", m%m%nblks,&
         dbcsr_fatal_level, dbcsr_internal_error, r,&
         'Size of blk_p inconsistent with number of blocks',759)
    !
    CALL dbcsr_assert(SIZE(row_blk_size), "EQ", m%m%nblkrows_total,&
         dbcsr_fatal_level, dbcsr_internal_error, r,&
         'Row block size array inconsistent with number of blocked rows',763)
    CALL dbcsr_assert(SIZE(col_blk_size), "EQ", m%m%nblkcols_total,&
         dbcsr_fatal_level, dbcsr_internal_error, r,&
         'Column block size array inconsistent with number of blocked columns',766)
    !
    IF(dbg.GE.4) THEN
       WRITE(*,'(1X,A,I7,A,I7)')r//' nze=',m%m%nze,'data size',&
            dbcsr_data_get_size(m%m%data_area)
    ENDIF
    data_size_referenced = dbcsr_data_get_size_referenced(m%m%data_area)
    !This tends to be too verbose and usually untrue for symmetric
    !matrices.
    !CALL dbcsr_assert(dbcsr_get_data_size(m%m%data_area), "GE", m%m%nze,&
    !     dbcsr_fatal_level, dbcsr_internal_error, r,&
    !     'Data storage may be too small.',777)
    IF (dbg.GE.5) THEN
       WRITE (*,'(1X,A,I7,A)') r//' size=',SIZE(m%m%row_p),' row_p='
       WRITE (*,2047) m%m%row_p(1:m%m%nblkrows_total+1)
       WRITE (*,'(1X,A)') r//' col_i='
       WRITE (*,2047) m%m%col_i(1:m%m%nblks)
       WRITE (*,'(1X,A)') r//' blk_p='
       WRITE (*,2047) m%m%blk_p(1:m%m%nblks)
    ENDIF
    prev_br = 0
    DO br = 1, m%m%nblkrows_total
       CALL dbcsr_assert(m%m%row_p(br), "GE", 0,&
            dbcsr_fatal_level, dbcsr_internal_error, r,&
            'row_p less than zero',790)
       IF(br.GT.1) CALL dbcsr_assert(m%m%row_p(br), "GE", m%m%row_p(prev_br),&
            dbcsr_fatal_level, dbcsr_internal_error, r,&
            'row_p decreases',793)
       mb = row_blk_size(br)
       CALL dbcsr_assert(mb, "GE", 0,&
            dbcsr_fatal_level, dbcsr_internal_error, r,&
            'Row blocked size is negative',797)
       DO blk = m%m%row_p(br)+1, m%m%row_p(br+1)
          CALL dbcsr_assert(blk, "GT", 0,&
               dbcsr_fatal_level, dbcsr_internal_error, r,&
               'Block number is zero',801)
          CALL dbcsr_assert(blk, "LE", m%m%nblks,&
               dbcsr_fatal_level, dbcsr_internal_error, r,&
               'Block number too high',804)
          bc = m%m%col_i(blk)
          IF (dbg.GE.5) THEN
             WRITE(*,'(1X,A,I7,"(",I5,",",I5,")")')r//' block',blk,br,bc
          ENDIF
          CALL dbcsr_assert(bc, "GT", 0,&
               dbcsr_fatal_level, dbcsr_internal_error, r,&
               'col_i is zero',811)
          CALL dbcsr_assert(bc, "LE", m%m%nblkcols_total,&
               dbcsr_fatal_level, dbcsr_internal_error, r,&
               'col_i too high',814)
          n = col_blk_size(bc)
          CALL dbcsr_assert(n, "GE", 0,&
               dbcsr_fatal_level, dbcsr_internal_error, r,&
               'Column blocked size is negative',818)
          blk_p = m%m%blk_p(blk)
          mn = mb*n
          !CALL dbcsr_assert(dbg,blk_p.GT.0, r, 'Block pointer is negative')
          !CALL dbcsr_assert(blk_p, "LE", m%m%nze,&
          !     dbcsr_fatal_level, dbcsr_internal_error, r,&
          !     'Block pointer too large',824)
          !CALL dbcsr_assert(blk_p+mn-1, "LE", m%m%nze,&
          !     dbcsr_fatal_level, dbcsr_internal_error, r,&
          !     'Block extends too far',827)
          IF (mn .GT. 0) THEN
             CALL dbcsr_assert (ABS(blk_p), "LE", data_size_referenced,&
                  dbcsr_fatal_level, dbcsr_internal_error, r,&
                  "Block pointer pointso outside of declared referenced area",&
                  832)
          ENDIF
          CALL dbcsr_assert (ABS(blk_p)+mn-1, "LE", data_size_referenced,&
               dbcsr_fatal_level, dbcsr_internal_error, r,&
               "Block extends outside of declared referenced area",&
               837)
       ENDDO
       prev_br = br
    ENDDO
    IF (dbg.GE.3.AND..NOT.nocomm) THEN
       CALL mp_sum(n_have_blocks_local, dbcsr_mp_group (dbcsr_distribution_mp (&
            m%m%dist)))
       n_have_blocks_total = n_have_blocks_local
       sparsity_total = REAL(n_have_blocks_total)&
            / REAL(n_full_blocks_total)*100.0
       !WRITE(*,FMT='(30A,F5.1,A)')r//' Sparsity: ', sparsity_total,'%'
       WRITE(*,FMT='(1X,A,F5.1,A)')r//' Non-sparsity: ',&
            sparsity_total,'%'
    ENDIF

    CALL timestop(handle)
  END SUBROUTINE dbcsr_verify_matrix


! **************************************************************************************************
!> \brief ...
!> \param nelements ...
!> \param bins ...
!> \param nbins ...
!> \param bin_counts ...
! **************************************************************************************************
  PURE SUBROUTINE count_bins(nelements, bins, nbins, bin_counts)
      INTEGER, INTENT(IN)                                :: nelements
      INTEGER, DIMENSION(1:nelements), INTENT(IN)        :: bins
      INTEGER, INTENT(IN)                                :: nbins
      INTEGER, DIMENSION(1:nbins), INTENT(OUT)           :: bin_counts

      INTEGER                                            :: el

    bin_counts(:) = 0
    DO el = 1, nelements
       bin_counts(bins(el)) = bin_counts(bins(el)) + 1
    ENDDO
  END SUBROUTINE count_bins

! **************************************************************************************************
!> \brief Makes a lookup table from the most common elements.
!> \param[in] array                  Array for which to find the most common
!>                                   elements.
!> \param[out] most_common_map       Ranking of the most common elements
!>                                   in array
!> \param[in] nmost_common           The number of most common elements
!> \param[out] most_common_elements  The most common elements in array
!> \param[in] size_limit             Limit maximum size to this value
!> \param max_val ...
!> \par Lookup table
!>      The lookup table is indexed by the most common array values
!>      (i.e., block sizes).  Its values are the order of their frequency.
! **************************************************************************************************
  SUBROUTINE map_most_common(array, most_common_map, nmost_common,&
       most_common_elements, size_limit, max_val)
      INTEGER, DIMENSION(:), INTENT(IN)                  :: array
      INTEGER(KIND=int_4), DIMENSION(:), POINTER         :: most_common_map
      INTEGER, INTENT(IN)                                :: nmost_common
      INTEGER, DIMENSION(:), INTENT(OUT)                 :: most_common_elements
      INTEGER, INTENT(IN)                                :: size_limit
      INTEGER, INTENT(OUT)                               :: max_val

      CHARACTER(LEN=*), PARAMETER :: routineN = 'map_most_common', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, max_val_l, nmc
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: permutation, size_counts

!   ---------------------------------------------------------------------------

    IF (SIZE (array) .GT. 0) THEN
       max_val = MAXVAL (array)
       max_val_l = MIN(MIN (size_limit, max_val), INT(HUGE(most_common_map)))
    ELSE
       max_val = 0
       max_val_l = 0
    ENDIF
    ! Count the frequency of all block sizes up to max_val_l.
    ALLOCATE (size_counts (0:max_val_l))
    ALLOCATE (permutation (0:max_val_l))
    size_counts = 0
    permutation = 0
    DO i = 1, SIZE(array)
       ! Counts are decreased to easily get a reverse sort order.
       IF (array(i) .LE. max_val_l) &
            size_counts(array(i)) = size_counts(array(i)) - 1
    END DO
    IF (SIZE(array) .GT. 0) THEN
       CALL sort (size_counts, max_val_l+1, permutation)
    ENDIF
    ! Limiting nmc to max_val_l prevents out-of-bounds.
    nmc = MIN(nmost_common, max_val_l)
    ! Determine the biggest block size and allocate the map.
    ALLOCATE (most_common_map(0:max_val_l))
    ! Create the mapping from block size to order.
    most_common_map = nmost_common+1
    DO i = 1, nmc
       most_common_map(permutation(i-1)-1) = i
    END DO
    ! Copy the most common elements
    most_common_elements(:) = 0
    most_common_elements(1:nmc) = permutation(0:nmc-1)-1
    DEALLOCATE (size_counts)
    DEALLOCATE (permutation)
  END SUBROUTINE map_most_common


END MODULE dbcsr_util
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Routines for basic block transformations.
!> \author  Urban Borstnik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2009-05-12 moved from dbcsr_util
! **************************************************************************************************
MODULE dbcsr_block_operations

   USE acc_devmem,                      ONLY: acc_devmem_allocated,&
                                              acc_devmem_setzero_bytes
   USE dbcsr_data_methods_low,          ONLY: &
        dbcsr_data_exists, dbcsr_data_get_size, dbcsr_data_get_size_referenced, &
        dbcsr_data_get_type, dbcsr_data_verify_bounds, dbcsr_get_data_p_2d_c, &
        dbcsr_get_data_p_2d_d, dbcsr_get_data_p_2d_s, dbcsr_get_data_p_2d_z, dbcsr_get_data_p_c, &
        dbcsr_get_data_p_d, dbcsr_get_data_p_s, dbcsr_get_data_p_z, dbcsr_type_2d_to_1d, &
        dbcsr_type_is_2d
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_caller_error,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_ptr_util,                  ONLY: memory_copy
   USE dbcsr_types,                     ONLY: &
        dbcsr_data_obj, dbcsr_datatype_sizeof, dbcsr_scalar_type, dbcsr_type_complex_4, &
        dbcsr_type_complex_4_2d, dbcsr_type_complex_8, dbcsr_type_complex_8_2d, dbcsr_type_real_4, &
        dbcsr_type_real_4_2d, dbcsr_type_real_8, dbcsr_type_real_8_2d
   USE kinds,                           ONLY: dp,&
                                              real_4,&
                                              real_8,&
                                              sp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads
   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_block_operations'

   PUBLIC :: dbcsr_block_transpose, dbcsr_block_transpose_aa, dbcsr_data_set
   PUBLIC :: dbcsr_block_copy_aa, dbcsr_block_partial_copy
   PUBLIC :: dbcsr_data_clear
   PUBLIC :: set_block2d_diagonal, get_block2d_diagonal
   PUBLIC :: dbcsr_block_scale, dbcsr_block_conjg, dbcsr_block_real_neg

   PUBLIC :: dbcsr_data_copy
   PUBLIC :: block_add_on_diag

   PUBLIC :: block_add

   ! For quick access
   PUBLIC :: block_copy_s, block_copy_d, &
             block_copy_c, block_copy_z

   INTERFACE dbcsr_block_transpose
      MODULE PROCEDURE block_transpose_inplace_s, block_transpose_inplace_d, &
         block_transpose_inplace_c, block_transpose_inplace_z
      MODULE PROCEDURE block_transpose_copy_d, block_transpose_copy_s, &
         block_transpose_copy_z, block_transpose_copy_c
      MODULE PROCEDURE block_transpose_copy_2d1d_d, &
         block_transpose_copy_2d1d_s, &
         block_transpose_copy_2d1d_z, &
         block_transpose_copy_2d1d_c
      MODULE PROCEDURE block_transpose_copy_1d2d_d, &
         block_transpose_copy_1d2d_s, &
         block_transpose_copy_1d2d_z, &
         block_transpose_copy_1d2d_c
      MODULE PROCEDURE dbcsr_block_transpose_aa, dbcsr_block_transpose_a
   END INTERFACE

   INTERFACE dbcsr_block_copy
      MODULE PROCEDURE block_copy_2d1d_s, block_copy_2d1d_d, &
         block_copy_2d1d_c, block_copy_2d1d_z
      MODULE PROCEDURE block_copy_1d2d_s, block_copy_1d2d_d, &
         block_copy_1d2d_c, block_copy_1d2d_z
      MODULE PROCEDURE block_copy_1d1d_s, block_copy_1d1d_d, &
         block_copy_1d1d_c, block_copy_1d1d_z
      MODULE PROCEDURE block_copy_2d2d_s, block_copy_2d2d_d, &
         block_copy_2d2d_c, block_copy_2d2d_z
   END INTERFACE

   INTERFACE dbcsr_data_clear
      MODULE PROCEDURE dbcsr_data_clear_nt
      MODULE PROCEDURE dbcsr_data_clear0
   END INTERFACE

   ! Supports copy between two data areas, or to a data area from a
   ! given explicit array.
   INTERFACE dbcsr_data_set
      MODULE PROCEDURE dbcsr_data_copy_aa, dbcsr_data_set_as, &
         dbcsr_data_set_ad, dbcsr_data_set_ac, dbcsr_data_set_az
   END INTERFACE

   INTERFACE dbcsr_data_copy
      MODULE PROCEDURE dbcsr_data_copy_aa2, dbcsr_data_set_as, &
         dbcsr_data_set_ad, dbcsr_data_set_ac, dbcsr_data_set_az
   END INTERFACE

   INTERFACE get_block2d_diagonal
      MODULE PROCEDURE get_block2d_diagonal_a
      MODULE PROCEDURE get_block2d_diagonal_s, get_block2d_diagonal_d, &
         get_block2d_diagonal_c, get_block2d_diagonal_z
   END INTERFACE

   INTERFACE set_block2d_diagonal
      MODULE PROCEDURE set_block2d_diagonal_a
      MODULE PROCEDURE set_block2d_diagonal_s, set_block2d_diagonal_d, &
         set_block2d_diagonal_c, set_block2d_diagonal_z
   END INTERFACE

   INTERFACE block_add_on_diag
      MODULE PROCEDURE block_add_on_diag_anytype
      MODULE PROCEDURE block_2d_add_on_diag_d, block_2d_add_on_diag_s, &
         block_2d_add_on_diag_z, block_2d_add_on_diag_c
      MODULE PROCEDURE block_add_on_diag_d, block_add_on_diag_s, &
         block_add_on_diag_z, block_add_on_diag_c
   END INTERFACE

   INTERFACE block_add
      MODULE PROCEDURE block_add_anytype
      MODULE PROCEDURE block_add_s, block_add_d, block_add_c, block_add_z
   END INTERFACE block_add

   LOGICAL, PARAMETER :: debug_mod = .FALSE.
   LOGICAL, PARAMETER :: careful_mod = .FALSE.

CONTAINS

! **************************************************************************************************
!> \brief Copy data from one data area to another.
!>
!> There are no checks done for correctness!
!> \param[in] dst        destination data area
!> \param[in] src        source data area
!> \param[in] row_size   row size of existing block
!> \param[in] col_size   column size of existing block
!> \param[in] lb         (optional) lower bound for destination (and source if
!>                       not given explicity)
!> \param[in] source_lb  (optional) lower bound of source
!> \param[in] scale      (optional) scale data
!> \param[in] lb2        (optional) lower bound of 2nd dimension for target
!> \param[in] source_lb2 (optional) lower bound of 2nd dimension for source
! **************************************************************************************************
   SUBROUTINE dbcsr_block_transpose_aa(dst, src, &
                                       row_size, col_size, lb, source_lb, scale, lb2, source_lb2)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: dst
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: src
      INTEGER, INTENT(IN)                                :: row_size, col_size
      INTEGER, INTENT(IN), OPTIONAL                      :: lb, source_lb
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: scale
      INTEGER, INTENT(IN), OPTIONAL                      :: lb2, source_lb2

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_block_transpose_aa', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_size, lb2_s, lb2_t, lb_s, lb_t, &
                                                            ub_s, ub_t

!   ---------------------------------------------------------------------------

      IF (debug_mod) THEN
         CALL dbcsr_assert(ASSOCIATED(dst%d), 'AND', ASSOCIATED(src%d), &
                           dbcsr_fatal_level, dbcsr_caller_error, routineN, &
                           "Data areas must be setup.", 171)
         CALL dbcsr_assert(dst%d%data_type, 'EQ', src%d%data_type, dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "Data type must be the same.", 173)
         IF (dst%d%data_type .NE. dbcsr_type_real_8 &
             .AND. dst%d%data_type .NE. dbcsr_type_real_4 &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_8 &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_4 &
             .AND. dst%d%data_type .NE. dbcsr_type_real_8_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_real_4_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_8_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_4_2d) &
            CALL cp__w("dbcsr/block/dbcsr_block_operations.F",182,"Incorrect data type.")
         IF (PRESENT(scale)) THEN
            IF (dbcsr_type_is_2d(src%d%data_type)) THEN
               CALL dbcsr_assert(scale%data_type, &
                                 'EQ', dbcsr_type_2d_to_1d(src%d%data_type), &
                                 dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                                 "Incompatible data types", 188)
            ELSE
               CALL dbcsr_assert(scale%data_type, 'EQ', src%d%data_type, dbcsr_fatal_level, &
                                 dbcsr_wrong_args_error, routineN, "Incompatible data types", 191)
            ENDIF
         ENDIF
      ENDIF
      data_size = row_size*col_size
      lb_t = 1
      IF (PRESENT(lb)) lb_t = lb
      ub_t = lb_t+data_size-1
      IF (PRESENT(source_lb)) THEN
         lb_s = source_lb
         ub_s = source_lb+data_size-1
      ELSE
         lb_s = lb_t
         ub_s = ub_t
      ENDIF
      lb2_t = 1
      IF (PRESENT(lb2)) lb2_t = lb2
      IF (PRESENT(source_lb2)) THEN
         lb2_s = source_lb2
      ELSE
         lb2_s = lb2_t
      ENDIF
      SELECT CASE (src%d%data_type)
      CASE (dbcsr_type_real_8)
         IF (PRESENT(scale)) THEN
            CALL dbcsr_block_transpose(dst%d%r_dp(lb_t:ub_t), &
                                       src%d%r_dp(lb_s:ub_s)*scale%r_dp, &
                                       row_size, col_size)
         ELSE
            CALL dbcsr_block_transpose(dst%d%r_dp(lb_t:ub_t), &
                                       src%d%r_dp(lb_s:ub_s), &
                                       row_size, col_size)
         ENDIF
      CASE (dbcsr_type_real_4)
         IF (PRESENT(scale)) THEN
            CALL dbcsr_block_transpose(dst%d%r_sp(lb_t:ub_t), &
                                       src%d%r_sp(lb_s:ub_s)*scale%r_sp, &
                                       row_size, col_size)
         ELSE
            CALL dbcsr_block_transpose(dst%d%r_sp(lb_t:ub_t), &
                                       src%d%r_sp(lb_s:ub_s), &
                                       row_size, col_size)
         ENDIF
      CASE (dbcsr_type_complex_8)
         IF (PRESENT(scale)) THEN
            CALL dbcsr_block_transpose(dst%d%c_dp(lb_t:ub_t), &
                                       src%d%c_dp(lb_s:ub_s)*scale%c_dp, &
                                       row_size, col_size)
         ELSE
            CALL dbcsr_block_transpose(dst%d%c_dp(lb_t:ub_t), &
                                       src%d%c_dp(lb_s:ub_s), &
                                       row_size, col_size)
         ENDIF
      CASE (dbcsr_type_complex_4)
         IF (PRESENT(scale)) THEN
            CALL dbcsr_block_transpose(dst%d%c_sp(lb_t:ub_t), &
                                       src%d%c_sp(lb_s:ub_s)*scale%c_sp, &
                                       row_size, col_size)
         ELSE
            CALL dbcsr_block_transpose(dst%d%c_sp(lb_t:ub_t), &
                                       src%d%c_sp(lb_s:ub_s), &
                                       row_size, col_size)
         ENDIF
      CASE (dbcsr_type_real_8_2d)
         IF (PRESENT(scale)) THEN
            dst%d%r2_dp(lb_t:lb_t+col_size-1, lb2_t:lb2_t+row_size-1) = &
               TRANSPOSE( &
               src%d%r2_dp(lb_s:lb_s+row_size-1, lb2_s:lb2_s+col_size-1) &
               *scale%r_dp)
         ELSE
            dst%d%r2_dp(lb_t:lb_t+col_size-1, lb2_t:lb2_t+row_size-1) = &
               TRANSPOSE( &
               src%d%r2_dp(lb_s:lb_s+row_size-1, lb2_s:lb2_s+col_size-1))
         ENDIF
      CASE (dbcsr_type_real_4_2d)
         IF (PRESENT(scale)) THEN
            dst%d%r2_sp(lb_t:lb_t+col_size-1, lb2_t:lb2_t+row_size-1) = &
               TRANSPOSE( &
               src%d%r2_sp(lb_s:lb_s+row_size-1, lb2_s:lb2_s+col_size-1) &
               *scale%r_sp)
         ELSE
            dst%d%r2_sp(lb_t:lb_t+col_size-1, lb2_t:lb2_t+row_size-1) = &
               TRANSPOSE( &
               src%d%r2_sp(lb_s:lb_s+row_size-1, lb2_s:lb2_s+col_size-1))
         ENDIF
      CASE (dbcsr_type_complex_8_2d)
         IF (PRESENT(scale)) THEN
            dst%d%c2_dp(lb_t:lb_t+col_size-1, lb2_t:lb2_t+row_size-1) = &
               TRANSPOSE( &
               src%d%c2_dp(lb_s:lb_s+row_size-1, lb2_s:lb2_s+col_size-1) &
               *scale%c_dp)
         ELSE
            dst%d%c2_dp(lb_t:lb_t+col_size-1, lb2_t:lb2_t+row_size-1) = &
               TRANSPOSE( &
               src%d%c2_dp(lb_s:lb_s+row_size-1, lb2_s:lb2_s+col_size-1))
         ENDIF
      CASE (dbcsr_type_complex_4_2d)
         IF (PRESENT(scale)) THEN
            dst%d%c2_sp(lb_t:lb_t+col_size-1, lb2_t:lb2_t+row_size-1) = &
               TRANSPOSE( &
               src%d%c2_sp(lb_s:lb_s+row_size-1, lb2_s:lb2_s+col_size-1) &
               *scale%c_sp)
         ELSE
            dst%d%c2_sp(lb_t:lb_t+col_size-1, lb2_t:lb2_t+row_size-1) = &
               TRANSPOSE( &
               src%d%c2_sp(lb_s:lb_s+row_size-1, lb2_s:lb2_s+col_size-1))
         ENDIF
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",299,"Incorrect data type.")
      END SELECT
   END SUBROUTINE dbcsr_block_transpose_aa

! **************************************************************************************************
!> \brief Copy data from one data area to another.
!>
!> There are no checks done for correctness!
!> \param[in] dst        destination data area
!> \param[in] src        source data area
!> \param[in] row_size   row size of existing block
!> \param[in] col_size   column size of existing block
!> \param[in] lb         (optional) lower bound for destination (and source if
!>                       not given explicity)
!> \param[in] source_lb  (optional) lower bound of source
!> \param[in] scale      (optional) scale data
! **************************************************************************************************
   SUBROUTINE dbcsr_block_copy_aa(dst, src, &
                                  row_size, col_size, lb, source_lb, scale)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: dst
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: src
      INTEGER, INTENT(IN)                                :: row_size, col_size
      INTEGER, INTENT(IN), OPTIONAL                      :: lb, source_lb
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_block_copy_aa', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_size, lb_s, lb_t, ub_s, ub_t

!   ---------------------------------------------------------------------------

      IF (debug_mod) THEN
         CALL dbcsr_assert(ASSOCIATED(dst%d), 'AND', ASSOCIATED(src%d), &
                           dbcsr_fatal_level, dbcsr_caller_error, routineN, &
                           "Data areas must be setup.", 334)
         CALL dbcsr_assert(dst%d%data_type, 'EQ', src%d%data_type, dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, "Data type must be the same.", 336)
         IF (dst%d%data_type .NE. dbcsr_type_real_8 &
             .AND. dst%d%data_type .NE. dbcsr_type_real_4 &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_8 &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_4) &
            CALL cp__w("dbcsr/block/dbcsr_block_operations.F",341,"Incorrect data type.")
         IF (PRESENT(scale)) THEN
            IF (dbcsr_type_is_2d(src%d%data_type)) THEN
               CALL dbcsr_assert(scale%data_type, &
                                 'EQ', dbcsr_type_2d_to_1d(src%d%data_type), &
                                 dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                                 "Incompatible data types", 347)
            ELSE
               CALL dbcsr_assert(scale%data_type, 'EQ', src%d%data_type, dbcsr_fatal_level, &
                                 dbcsr_wrong_args_error, routineN, "Incompatible data types", 350)
            ENDIF
         ENDIF
      ENDIF
      data_size = row_size*col_size
      lb_t = 1
      IF (PRESENT(lb)) lb_t = lb
      ub_t = lb_t+data_size-1
      IF (PRESENT(source_lb)) THEN
         lb_s = source_lb
         ub_s = source_lb+data_size-1
      ELSE
         lb_s = lb_t
         ub_s = ub_t
      ENDIF
      SELECT CASE (src%d%data_type)
      CASE (dbcsr_type_real_8)
         IF (PRESENT(scale)) THEN
            CALL dbcsr_block_copy(dst%d%r_dp(lb_t:ub_t), &
                                  src%d%r_dp(lb_s:ub_s)*scale%r_dp, &
                                  row_size, col_size)
         ELSE
            CALL dbcsr_block_copy(dst%d%r_dp(lb_t:ub_t), &
                                  src%d%r_dp(lb_s:ub_s), &
                                  row_size, col_size)
         ENDIF
      CASE (dbcsr_type_real_4)
         IF (PRESENT(scale)) THEN
            CALL dbcsr_block_copy(dst%d%r_sp(lb_t:ub_t), &
                                  src%d%r_sp(lb_s:ub_s)*scale%r_sp, &
                                  row_size, col_size)
         ELSE
            CALL dbcsr_block_copy(dst%d%r_sp(lb_t:ub_t), &
                                  src%d%r_sp(lb_s:ub_s), &
                                  row_size, col_size)
         ENDIF
      CASE (dbcsr_type_complex_8)
         IF (PRESENT(scale)) THEN
            CALL dbcsr_block_copy(dst%d%c_dp(lb_t:ub_t), &
                                  src%d%c_dp(lb_s:ub_s)*scale%c_dp, &
                                  row_size, col_size)
         ELSE
            CALL dbcsr_block_copy(dst%d%c_dp(lb_t:ub_t), &
                                  src%d%c_dp(lb_s:ub_s), &
                                  row_size, col_size)
         ENDIF
      CASE (dbcsr_type_complex_4)
         IF (PRESENT(scale)) THEN
            CALL dbcsr_block_copy(dst%d%c_sp(lb_t:ub_t), &
                                  src%d%c_sp(lb_s:ub_s)*scale%c_sp, &
                                  row_size, col_size)
         ELSE
            CALL dbcsr_block_copy(dst%d%c_sp(lb_t:ub_t), &
                                  src%d%c_sp(lb_s:ub_s), &
                                  row_size, col_size)
         ENDIF
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",407,"Incorrect data type.")
      END SELECT
   END SUBROUTINE dbcsr_block_copy_aa

! **************************************************************************************************
!> \brief Scale a data area.
!>
!> There are no checks done for correctness!
!> \param dst ...
!> \param[in] scale      (optional) scale data
!> \param[in] row_size   row size of existing block
!> \param[in] col_size   column size of existing block
!> \param[in] lb         (optional) lower bound for destination (and source if
!>                       not given explicity)
!> \param[in] lb2        (optional) lower bound of 2nd dimension for target
!> \par History
!> - 2010-09 [??] Copied from block_transpose?
!> - 2010-09-20 [UB] Swaps/corrects row/column definitions for 2-D bounds
! **************************************************************************************************
   SUBROUTINE dbcsr_block_scale(dst, scale, &
                                row_size, col_size, lb, lb2)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: dst
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: scale
      INTEGER, INTENT(IN)                                :: row_size, col_size
      INTEGER, INTENT(IN), OPTIONAL                      :: lb, lb2

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_block_scale', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_size, handle, lb2_t, lb_t, ub_t

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, handle)
      IF (debug_mod) THEN
         IF (.NOT. ASSOCIATED(dst%d)) &
            CALL cp__b("dbcsr/block/dbcsr_block_operations.F",443,"Data area must be setup.")
         IF (dst%d%data_type .NE. dbcsr_type_real_8 &
             .AND. dst%d%data_type .NE. dbcsr_type_real_4 &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_8 &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_4 &
             .AND. dst%d%data_type .NE. dbcsr_type_real_8_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_real_4_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_8_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_4_2d) &
            CALL cp__w("dbcsr/block/dbcsr_block_operations.F",452,"Incorrect data type.")
      ENDIF
      CALL dbcsr_assert(scale%data_type, &
                        'EQ', dbcsr_type_2d_to_1d(dst%d%data_type), &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "Incompatible data types", 457)

      data_size = row_size*col_size
      lb_t = 1
      IF (PRESENT(lb)) lb_t = lb
      ub_t = lb_t+data_size-1
      lb2_t = 1
      IF (PRESENT(lb2)) lb2_t = lb2
      SELECT CASE (dst%d%data_type)
      CASE (dbcsr_type_real_8)
         dst%d%r_dp(lb_t:ub_t) = dst%d%r_dp(lb_t:ub_t)*scale%r_dp
      CASE (dbcsr_type_real_4)
         dst%d%r_sp(lb_t:ub_t) = dst%d%r_sp(lb_t:ub_t)*scale%r_sp
      CASE (dbcsr_type_complex_8)
         dst%d%c_dp(lb_t:ub_t) = dst%d%c_dp(lb_t:ub_t)*scale%c_dp
      CASE (dbcsr_type_complex_4)
         dst%d%c_sp(lb_t:ub_t) = dst%d%c_sp(lb_t:ub_t)*scale%c_sp
      CASE (dbcsr_type_real_8_2d)
         dst%d%r2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            dst%d%r2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1)*scale%r_dp
      CASE (dbcsr_type_real_4_2d)
         dst%d%r2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            dst%d%r2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1)*scale%r_sp
      CASE (dbcsr_type_complex_8_2d)
         dst%d%c2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            dst%d%c2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1)*scale%c_dp
      CASE (dbcsr_type_complex_4_2d)
         dst%d%c2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            dst%d%c2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1)*scale%c_sp
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",487,"Incorrect data type.")
      END SELECT
      IF (careful_mod) CALL timestop(handle)
   END SUBROUTINE dbcsr_block_scale

! **************************************************************************************************
!> \brief Negates the real part of a block
!>
!> There are no checks done for correctness!
!> \param dst ...
!> \param[in] row_size   row size of existing block
!> \param[in] col_size   column size of existing block
!> \param[in] lb         (optional) lower bound for destination (and source if
!>                       not given explicity)
!> \param[in] lb2        (optional) lower bound of 2nd dimension for target
!> \par History
!> - 2010-09 [??] Copied from block_transpose?
!> - 2010-09-20 [UB] Swaps/corrects row/column definitions for 2-D bounds
! **************************************************************************************************
   SUBROUTINE dbcsr_block_real_neg(dst, &
                                   row_size, col_size, lb, lb2)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: dst
      INTEGER, INTENT(IN)                                :: row_size, col_size
      INTEGER, INTENT(IN), OPTIONAL                      :: lb, lb2

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_block_real_neg', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_size, handle, lb2_t, lb_t, ub_t

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, handle)
      IF (debug_mod) THEN
         IF (.NOT. ASSOCIATED(dst%d)) &
            CALL cp__b("dbcsr/block/dbcsr_block_operations.F",522,"Data area must be setup.")
         IF (dst%d%data_type .NE. dbcsr_type_real_8 &
             .AND. dst%d%data_type .NE. dbcsr_type_real_4 &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_8 &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_4 &
             .AND. dst%d%data_type .NE. dbcsr_type_real_8_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_real_4_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_8_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_4_2d) &
            CALL cp__w("dbcsr/block/dbcsr_block_operations.F",531,"Incorrect data type.")
      ENDIF

      data_size = row_size*col_size
      lb_t = 1
      IF (PRESENT(lb)) lb_t = lb
      ub_t = lb_t+data_size-1
      lb2_t = 1
      IF (PRESENT(lb2)) lb2_t = lb2
      SELECT CASE (dst%d%data_type)
      CASE (dbcsr_type_real_8)
         dst%d%r_dp(lb_t:ub_t) = -dst%d%r_dp(lb_t:ub_t)
      CASE (dbcsr_type_real_4)
         dst%d%r_sp(lb_t:ub_t) = -dst%d%r_sp(lb_t:ub_t)
      CASE (dbcsr_type_complex_8)
         dst%d%c_dp(lb_t:ub_t) = CMPLX( &
                                 REAL(dst%d%c_dp(lb_t:ub_t), KIND=real_8), &
                                 AIMAG(dst%d%c_dp(lb_t:ub_t)), &
                                 KIND=real_8)
      CASE (dbcsr_type_complex_4)
         dst%d%c_sp(lb_t:ub_t) = CMPLX( &
                                 REAL(dst%d%c_sp(lb_t:ub_t), KIND=real_4), &
                                 AIMAG(dst%d%c_sp(lb_t:ub_t)), &
                                 KIND=real_4)
      CASE (dbcsr_type_real_8_2d)
         dst%d%r2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            -dst%d%r2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1)
      CASE (dbcsr_type_real_4_2d)
         dst%d%r2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            -dst%d%r2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1)
      CASE (dbcsr_type_complex_8_2d)
         dst%d%c2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            CMPLX( &
            REAL(dst%d%c2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1), KIND=real_8), &
            AIMAG(dst%d%c2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1)), &
            KIND=real_8)
      CASE (dbcsr_type_complex_4_2d)
         dst%d%c2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            CMPLX( &
            REAL(dst%d%c2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1), KIND=real_4), &
            AIMAG(dst%d%c2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1)), &
            KIND=real_4)
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",574,"Incorrect data type.")
      END SELECT
      IF (careful_mod) CALL timestop(handle)
   END SUBROUTINE dbcsr_block_real_neg

! **************************************************************************************************
!> \brief Conjugate a data area.
!>
!> There are no checks done for correctness!
!> \param dst ...
!> \param[in] row_size   row size of existing block
!> \param[in] col_size   column size of existing block
!> \param[in] lb         (optional) lower bound for destination (and source if
!>                       not given explicity)
!> \param[in] lb2        (optional) lower bound of 2nd dimension for target
!> \par History
!> - 2010-09 [??] Copied from block_transpose?
!> - 2010-09-20 [UB] Swaps/corrects row/column definitions for 2-D bounds
! **************************************************************************************************
   SUBROUTINE dbcsr_block_conjg(dst, &
                                row_size, col_size, lb, lb2)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: dst
      INTEGER, INTENT(IN)                                :: row_size, col_size
      INTEGER, INTENT(IN), OPTIONAL                      :: lb, lb2

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_block_conjg', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_size, handle, lb2_t, lb_t, ub_t

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, handle)
      IF (debug_mod) THEN
         IF (.NOT. ASSOCIATED(dst%d)) &
            CALL cp__b("dbcsr/block/dbcsr_block_operations.F",609,"Data area must be setup.")
         IF (dst%d%data_type .NE. dbcsr_type_real_8 &
             .AND. dst%d%data_type .NE. dbcsr_type_real_4 &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_8 &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_4 &
             .AND. dst%d%data_type .NE. dbcsr_type_real_8_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_real_4_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_8_2d &
             .AND. dst%d%data_type .NE. dbcsr_type_complex_4_2d) &
            CALL cp__w("dbcsr/block/dbcsr_block_operations.F",618,"Incorrect data type.")
      ENDIF

      data_size = row_size*col_size
      lb_t = 1
      IF (PRESENT(lb)) lb_t = lb
      ub_t = lb_t+data_size-1
      lb2_t = 1
      IF (PRESENT(lb2)) lb2_t = lb2
      SELECT CASE (dst%d%data_type)
      CASE (dbcsr_type_real_8)
         dst%d%r_dp(lb_t:ub_t) = dst%d%r_dp(lb_t:ub_t)
      CASE (dbcsr_type_real_4)
         dst%d%r_sp(lb_t:ub_t) = dst%d%r_sp(lb_t:ub_t)
      CASE (dbcsr_type_complex_8)
         dst%d%c_dp(lb_t:ub_t) = CONJG(dst%d%c_dp(lb_t:ub_t))
      CASE (dbcsr_type_complex_4)
         dst%d%c_sp(lb_t:ub_t) = CONJG(dst%d%c_sp(lb_t:ub_t))
      CASE (dbcsr_type_real_8_2d)
         dst%d%r2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            dst%d%r2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1)
      CASE (dbcsr_type_real_4_2d)
         dst%d%r2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            dst%d%r2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1)
      CASE (dbcsr_type_complex_8_2d)
         dst%d%c2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            CONJG(dst%d%c2_dp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1))
      CASE (dbcsr_type_complex_4_2d)
         dst%d%c2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1) = &
            CONJG(dst%d%c2_sp(lb_t:lb_t+row_size-1, lb2_t:lb2_t+col_size-1))
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",649,"Incorrect data type.")
      END SELECT
      IF (careful_mod) CALL timestop(handle)
   END SUBROUTINE dbcsr_block_conjg

! **************************************************************************************************
!> \brief In-place transpose of encapsulated data
!>
!> There are no checks done for correctness!
!> \param[in] area       encapsulated data area
!> \param[in] row_size   number of rows in existing block
!> \param[in] col_size   number of columns in existing block
! **************************************************************************************************
   SUBROUTINE dbcsr_block_transpose_a(area, row_size, col_size)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area
      INTEGER, INTENT(IN)                                :: row_size, col_size

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_block_transpose_a', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

!   ---------------------------------------------------------------------------

      IF (careful_mod) &
         CALL timeset(routineN, handle)
      SELECT CASE (area%d%data_type)
      CASE (dbcsr_type_real_8)
         CALL dbcsr_block_transpose(area%d%r_dp, &
                                    row_size, col_size)
      CASE (dbcsr_type_real_4)
         CALL dbcsr_block_transpose(area%d%r_sp, &
                                    row_size, col_size)
      CASE (dbcsr_type_complex_8)
         CALL dbcsr_block_transpose(area%d%c_dp, &
                                    row_size, col_size)
      CASE (dbcsr_type_complex_4)
         CALL dbcsr_block_transpose(area%d%c_sp, &
                                    row_size, col_size)
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",689,"Invalid data type")
      END SELECT
      IF (careful_mod) &
         CALL timestop(handle)
   END SUBROUTINE dbcsr_block_transpose_a

! **************************************************************************************************
!> \brief Copy data from one data area to another.
!>
!> There are no checks done for correctness!
!> \param[in] dst        destination data area
!> \param[in] lb         lower bound for destination (and source if
!>                       not given explicity)
!> \param[in] data_size  number of elements to copy
!> \param[in] src        source data area
!> \param[in] source_lb  (optional) lower bound of source
!> \param[in] scale      (optional) scale by this factor
!> \param[in] lb2        (optional) 2nd dimension lower bound
!> \param[in] data_size2 (optional) 2nd dimension data size
!> \param[in] source_lb2 (optional) 2nd dimension lower bound for source
! **************************************************************************************************
   SUBROUTINE dbcsr_data_copy_aa(dst, lb, data_size, src, source_lb, scale, &
                                 lb2, data_size2, source_lb2)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: dst
      INTEGER, INTENT(IN)                                :: lb, data_size
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: src
      INTEGER, INTENT(IN), OPTIONAL                      :: source_lb
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: scale
      INTEGER, INTENT(IN), OPTIONAL                      :: lb2, data_size2, source_lb2

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_copy_aa', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: lb2_s, lb_s, ub, ub2, ub2_s, ub_s

!   ---------------------------------------------------------------------------

      IF (debug_mod) THEN
         IF (.NOT. ASSOCIATED(dst%d) .OR. .NOT. ASSOCIATED(src%d)) &
            CALL cp__b("dbcsr/block/dbcsr_block_operations.F",728,"Data areas must be setup.")
         IF (dst%d%data_type .NE. src%d%data_type) &
            CALL cp__b("dbcsr/block/dbcsr_block_operations.F",730,"Data type must be the same.")
      ENDIF
      IF (PRESENT(scale) .AND. careful_mod) THEN
         IF (dbcsr_type_is_2d(src%d%data_type)) THEN
            IF (scale%data_type .NE. dbcsr_type_2d_to_1d(src%d%data_type)) &
               CALL cp__b("dbcsr/block/dbcsr_block_operations.F",735,"Incomptable data types")
         ELSE
            IF (scale%data_type .NE. src%d%data_type) &
               CALL cp__b("dbcsr/block/dbcsr_block_operations.F",738,"Incomptable data types")
         ENDIF
      ENDIF
      ub = lb+data_size-1
      IF (PRESENT(source_lb)) THEN
         lb_s = source_lb
         ub_s = source_lb+data_size-1
      ELSE
         lb_s = lb
         ub_s = ub
      ENDIF
      IF (careful_mod) THEN
         IF (dbcsr_type_is_2d(src%d%data_type) .AND. .NOT. PRESENT(lb2)) &
            CALL cp__b("dbcsr/block/dbcsr_block_operations.F",751,"Must specify lb2 for 2-D data area")
         IF (dbcsr_type_is_2d(src%d%data_type) .AND. .NOT. PRESENT(data_size2)) &
            CALL cp__b("dbcsr/block/dbcsr_block_operations.F",753,"Must specify data_size2 for 2-D data area")
      ENDIF
      IF (PRESENT(lb2)) THEN
         IF (careful_mod) THEN
            IF (.NOT. dbcsr_type_is_2d(src%d%data_type)) &
               CALL cp__w("dbcsr/block/dbcsr_block_operations.F",758,"Should not specify lb2 for 1-D data")
         ENDIF
         ub2 = lb2+data_size2-1
         IF (PRESENT(source_lb2)) THEN
            lb2_s = source_lb2
            ub2_s = source_lb2+data_size2-1
         ELSE
            lb2_s = lb2
            ub2_s = ub2
         ENDIF
         !write(*,*)routineN//" lb,ub2D <- S",lb2, ub2, lb2_s, ub2_s
      ENDIF
      SELECT CASE (src%d%data_type)
      CASE (dbcsr_type_real_4)
         IF (debug_mod) THEN
            IF (.NOT. ASSOCIATED(dst%d%r_sp)) &
               CALL cp__b("dbcsr/block/dbcsr_block_operations.F",774,"associated(dst%d%r_sp)")
            IF (.NOT. ASSOCIATED(src%d%r_sp)) &
               CALL cp__b("dbcsr/block/dbcsr_block_operations.F",776,"associated(src%d%r_sp)")
            CALL dbcsr_assert(lb, "GE", LBOUND(dst%d%r_sp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lb dst%d%r_sp", 778)
            CALL dbcsr_assert(ub, "LE", UBOUND(dst%d%r_sp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "ub dst%d%r_sp", 780)
            CALL dbcsr_assert(lb_s, "GE", LBOUND(src%d%r_sp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lb src%d%r_sp", 782)
            CALL dbcsr_assert(ub_s, "LE", UBOUND(src%d%r_sp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "ub src%d%r_sp", 784)
         ENDIF
         IF (PRESENT(scale)) THEN
            dst%d%r_sp(lb:ub) = scale%r_sp*src%d%r_sp(lb_s:ub_s)
         ELSE
            dst%d%r_sp(lb:ub) = src%d%r_sp(lb_s:ub_s)
         ENDIF
      CASE (dbcsr_type_real_8)
         IF (debug_mod) THEN
            IF (.NOT. ASSOCIATED(dst%d%r_dp)) &
               CALL cp__b("dbcsr/block/dbcsr_block_operations.F",794,"associated(dst%d%r_dp)")
            IF (.NOT. ASSOCIATED(src%d%r_dp)) &
               CALL cp__b("dbcsr/block/dbcsr_block_operations.F",796,"associated(src%d%r_dp)")
            CALL dbcsr_assert(lb, "GE", LBOUND(dst%d%r_dp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lb dst%d%r_dp", 798)
            CALL dbcsr_assert(ub, "LE", UBOUND(dst%d%r_dp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "ub dst%d%r_dp", 800)
            CALL dbcsr_assert(lb_s, "GE", LBOUND(src%d%r_dp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lb src%d%r_dp", 802)
            CALL dbcsr_assert(ub_s, "LE", UBOUND(src%d%r_dp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "ub src%d%r_dp", 804)
         ENDIF
         IF (PRESENT(scale)) THEN
            dst%d%r_dp(lb:ub) = scale%r_dp*src%d%r_dp(lb_s:ub_s)
         ELSE
            dst%d%r_dp(lb:ub) = src%d%r_dp(lb_s:ub_s)
         ENDIF
      CASE (dbcsr_type_complex_4)
         IF (debug_mod) THEN
            IF (.NOT. ASSOCIATED(dst%d%c_sp)) &
               CALL cp__b("dbcsr/block/dbcsr_block_operations.F",814,"associated(dst%d%c_sp)")
            IF (.NOT. ASSOCIATED(src%d%c_sp)) &
               CALL cp__b("dbcsr/block/dbcsr_block_operations.F",816,"associated(src%d%c_sp)")
            CALL dbcsr_assert(lb, "GE", LBOUND(dst%d%c_sp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lb dst%d%c_sp", 818)
            CALL dbcsr_assert(ub, "LE", UBOUND(dst%d%c_sp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "ub dst%d%c_sp", 820)
            CALL dbcsr_assert(lb_s, "GE", LBOUND(src%d%c_sp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lb src%d%c_sp", 822)
            CALL dbcsr_assert(ub_s, "LE", UBOUND(src%d%c_sp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "ub src%d%c_sp", 824)
         ENDIF
         IF (PRESENT(scale)) THEN
            dst%d%c_sp(lb:ub) = scale%c_sp*src%d%c_sp(lb_s:ub_s)
         ELSE
            dst%d%c_sp(lb:ub) = src%d%c_sp(lb_s:ub_s)
         ENDIF
      CASE (dbcsr_type_complex_8)
         IF (debug_mod) THEN
            IF (.NOT. ASSOCIATED(dst%d%c_dp)) &
               CALL cp__b("dbcsr/block/dbcsr_block_operations.F",834,"associated(dst%d%c_dp)")
            IF (.NOT. ASSOCIATED(src%d%c_dp)) &
               CALL cp__b("dbcsr/block/dbcsr_block_operations.F",836,"associated(src%d%c_dp)")
            CALL dbcsr_assert(lb, "GE", LBOUND(dst%d%c_dp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lb dst%d%c_dp", 838)
            CALL dbcsr_assert(ub, "LE", UBOUND(dst%d%c_dp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "ub dst%d%c_dp", 840)
            CALL dbcsr_assert(lb_s, "GE", LBOUND(src%d%c_dp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lb src%d%c_dp", 842)
            CALL dbcsr_assert(ub_s, "LE", UBOUND(src%d%c_dp, 1), dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "ub src%d%c_dp", 844)
         ENDIF
         IF (PRESENT(scale)) THEN
            dst%d%c_dp(lb:ub) = scale%c_dp*src%d%c_dp(lb_s:ub_s)
         ELSE
            dst%d%c_dp(lb:ub) = src%d%c_dp(lb_s:ub_s)
         ENDIF
      CASE (dbcsr_type_real_4_2d)
         IF (PRESENT(scale)) THEN
            dst%d%r2_sp(lb:ub, lb2:ub2) = &
               scale%r_sp*src%d%r2_sp(lb_s:ub_s, lb2_s:ub2_s)
         ELSE
            dst%d%r2_sp(lb:ub, lb2:ub2) = src%d%r2_sp(lb_s:ub_s, lb2_s:ub2_s)
         ENDIF
      CASE (dbcsr_type_real_8_2d)
         IF (PRESENT(scale)) THEN
            dst%d%r2_dp(lb:ub, lb2:ub2) = &
               scale%r_dp*src%d%r2_dp(lb_s:ub_s, lb2_s:ub2_s)
         ELSE
            dst%d%r2_dp(lb:ub, lb2:ub2) = src%d%r2_dp(lb_s:ub_s, lb2_s:ub2_s)
         ENDIF
      CASE (dbcsr_type_complex_4_2d)
         IF (PRESENT(scale)) THEN
            dst%d%c2_sp(lb:ub, lb2:ub2) = &
               scale%c_sp*src%d%c2_sp(lb_s:ub_s, lb2_s:ub2_s)
         ELSE
            dst%d%c2_sp(lb:ub, lb2:ub2) = src%d%c2_sp(lb_s:ub_s, lb2_s:ub2_s)
         ENDIF
      CASE (dbcsr_type_complex_8_2d)
         IF (PRESENT(scale)) THEN
            dst%d%c2_dp(lb:ub, lb2:ub2) = &
               scale%c_dp*src%d%c2_dp(lb_s:ub_s, lb2_s:ub2_s)
         ELSE
            dst%d%c2_dp(lb:ub, lb2:ub2) = src%d%c2_dp(lb_s:ub_s, lb2_s:ub2_s)
         ENDIF
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",880,"Invalid data type")
      END SELECT
   END SUBROUTINE dbcsr_data_copy_aa

! **************************************************************************************************
!> \brief Copy data from one data area to another, the most basic form.
!>
!> There are no checks done for correctness!
!> \param[in,out] dst    destination data area
!> \param[in] dst_lb     lower bounds for destination
!> \param[in] dst_sizes  sizes for destination
!> \param[in] src        source data area
!> \param[in] src_lb     lower bounds for source
!> \param[in] src_sizes  sizes for source
! **************************************************************************************************
   SUBROUTINE dbcsr_data_copy_aa2(dst, dst_lb, dst_sizes, &
                                  src, src_lb, src_sizes)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: dst
      INTEGER, DIMENSION(:), INTENT(IN)                  :: dst_lb, dst_sizes
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: src
      INTEGER, DIMENSION(:), INTENT(IN)                  :: src_lb, src_sizes

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_copy_aa2', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: dst_d, dst_dt, handle, src_d, src_dt
      INTEGER, DIMENSION(2)                              :: dst_ub, src_ub

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      !
      src_dt = dbcsr_data_get_type(src)
      dst_dt = dbcsr_data_get_type(dst)
      IF (dbcsr_type_is_2d(src_dt)) THEN
         src_d = 2
      ELSE
         src_d = 1
      ENDIF
      IF (dbcsr_type_is_2d(dst_dt)) THEN
         dst_d = 2
      ELSE
         dst_d = 1
      ENDIF
      src_ub(1:src_d) = src_lb(1:src_d)+src_sizes(1:src_d)-1
      dst_ub(1:dst_d) = dst_lb(1:dst_d)+dst_sizes(1:dst_d)-1
      IF (careful_mod) THEN
         IF (.NOT. dbcsr_data_exists(dst)) &
            CALL cp__b("dbcsr/block/dbcsr_block_operations.F",928,"Invalid target data area")
         IF (.NOT. dbcsr_data_exists(src)) &
            CALL cp__b("dbcsr/block/dbcsr_block_operations.F",930,"Invalid source data area")
         CALL dbcsr_assert(dbcsr_type_2d_to_1d(src_dt), "EQ", &
                           dbcsr_type_2d_to_1d(dst_dt), dbcsr_fatal_level, &
                           dbcsr_caller_error, routineN, &
                           "Data types must be comptable: ", 934)
         IF (dbcsr_type_is_2d(dst_dt)) THEN
            CALL dbcsr_assert(SIZE(dst_lb), "EQ", 2, dbcsr_fatal_level, &
                              dbcsr_wrong_args_error, routineN, &
                              "size must be 2 for 2-d dst_lb", 938)
            CALL dbcsr_assert(SIZE(dst_sizes), "EQ", 2, dbcsr_fatal_level, &
                              dbcsr_wrong_args_error, routineN, &
                              "size must be 2 for 2-d dst_sizes", 941)
         ELSE
            CALL dbcsr_assert(SIZE(dst_lb), "EQ", 1, dbcsr_fatal_level, &
                              dbcsr_wrong_args_error, routineN, &
                              "size must be 1 for 1-d dst_lb", 945)
            CALL dbcsr_assert(SIZE(dst_sizes), "EQ", 1, dbcsr_fatal_level, &
                              dbcsr_wrong_args_error, routineN, &
                              "size must be 1 for 1-d dst_sizes", 948)
         ENDIF
         IF (dbcsr_type_is_2d(src_dt)) THEN
            CALL dbcsr_assert(SIZE(src_lb), "EQ", 2, dbcsr_fatal_level, &
                              dbcsr_wrong_args_error, routineN, &
                              "size must be 2 for 2-d src_lb", 953)
            CALL dbcsr_assert(SIZE(src_sizes), "EQ", 2, dbcsr_fatal_level, &
                              dbcsr_wrong_args_error, routineN, &
                              "size must be 2 for 2-d src_sizes", 956)
         ELSE
            CALL dbcsr_assert(SIZE(src_lb), "EQ", 1, dbcsr_fatal_level, &
                              dbcsr_wrong_args_error, routineN, &
                              "size must be 1 for 1-d src_lb", 960)
            CALL dbcsr_assert(SIZE(src_sizes), "EQ", 1, dbcsr_fatal_level, &
                              dbcsr_wrong_args_error, routineN, &
                              "size must be 1 for 1-d src_sizes", 963)
         ENDIF
         IF (debug_mod) THEN
            CALL dbcsr_data_verify_bounds(dst, dst_lb(1:dst_d), dst_ub(1:dst_d))
            CALL dbcsr_data_verify_bounds(src, src_lb(1:src_d), src_ub(1:src_d))
         ENDIF
      ENDIF
      !
      SELECT CASE (src_dt)
      CASE (dbcsr_type_real_4)
         IF (dbcsr_type_is_2d(dst_dt)) THEN
            CALL dbcsr_block_copy(dst%d%r2_sp(dst_lb(1):dst_ub(1), &
                                              dst_lb(2):dst_ub(2)), &
                                  src%d%r_sp(src_lb(1):src_ub(1)), &
                                  src_sizes(1), 1)
         ELSE
            CALL dbcsr_block_copy(dst%d%r_sp(dst_lb(1):dst_ub(1)), &
                                  src%d%r_sp(src_lb(1):src_ub(1)), &
                                  src_sizes(1), 1)
         ENDIF
      CASE (dbcsr_type_real_8)
         IF (dbcsr_type_is_2d(dst_dt)) THEN
            CALL dbcsr_block_copy(dst%d%r2_dp(dst_lb(1):dst_ub(1), &
                                              dst_lb(2):dst_ub(2)), &
                                  src%d%r_dp(src_lb(1):src_ub(1)), &
                                  src_sizes(1), 1)
         ELSE
            CALL dbcsr_block_copy(dst%d%r_dp(dst_lb(1):dst_ub(1)), &
                                  src%d%r_dp(src_lb(1):src_ub(1)), &
                                  src_sizes(1), 1)
         ENDIF
      CASE (dbcsr_type_complex_4)
         IF (dbcsr_type_is_2d(dst_dt)) THEN
            CALL dbcsr_block_copy(dst%d%c2_sp(dst_lb(1):dst_ub(1), &
                                              dst_lb(2):dst_ub(2)), &
                                  src%d%c_sp(src_lb(1):src_ub(1)), &
                                  src_sizes(1), 1)
         ELSE
            CALL dbcsr_block_copy(dst%d%c_sp(dst_lb(1):dst_ub(1)), &
                                  src%d%c_sp(src_lb(1):src_ub(1)), &
                                  src_sizes(1), 1)
         ENDIF
      CASE (dbcsr_type_complex_8)
         IF (dbcsr_type_is_2d(dst_dt)) THEN
            CALL dbcsr_block_copy(dst%d%c2_dp(dst_lb(1):dst_ub(1), &
                                              dst_lb(2):dst_ub(2)), &
                                  src%d%c_dp(src_lb(1):src_ub(1)), &
                                  src_sizes(1), 1)
         ELSE
            CALL dbcsr_block_copy(dst%d%c_dp(dst_lb(1):dst_ub(1)), &
                                  src%d%c_dp(src_lb(1):src_ub(1)), &
                                  src_sizes(1), 1)
         ENDIF
      CASE (dbcsr_type_real_4_2d)
         IF (dbcsr_type_is_2d(dst_dt)) THEN
            CALL dbcsr_block_copy(dst%d%r2_sp(dst_lb(1):dst_ub(1), &
                                              dst_lb(2):dst_ub(2)), &
                                  src%d%r2_sp(src_lb(1):src_ub(1), &
                                              src_lb(2):src_ub(2)), &
                                  dst_sizes(1), dst_sizes(2))
         ELSE
            CALL dbcsr_block_copy(dst%d%r_sp(dst_lb(1):dst_ub(1)), &
                                  src%d%r2_sp(src_lb(1):src_ub(1), &
                                              src_lb(2):src_ub(2)), &
                                  dst_sizes(1), dst_sizes(2))
         ENDIF
      CASE (dbcsr_type_real_8_2d)
         IF (dbcsr_type_is_2d(dst_dt)) THEN
            CALL dbcsr_block_copy(dst%d%r2_dp(dst_lb(1):dst_ub(1), &
                                              dst_lb(2):dst_ub(2)), &
                                  src%d%r2_dp(src_lb(1):src_ub(1), &
                                              src_lb(2):src_ub(2)), &
                                  dst_sizes(1), dst_sizes(2))
         ELSE
            CALL dbcsr_block_copy(dst%d%r_dp(dst_lb(1):dst_ub(1)), &
                                  src%d%r2_dp(src_lb(1):src_ub(1), &
                                              src_lb(2):src_ub(2)), &
                                  dst_sizes(1), dst_sizes(2))
         ENDIF
      CASE (dbcsr_type_complex_4_2d)
         IF (dbcsr_type_is_2d(dst_dt)) THEN
            CALL dbcsr_block_copy(dst%d%c2_sp(dst_lb(1):dst_ub(1), &
                                              dst_lb(2):dst_ub(2)), &
                                  src%d%c2_sp(src_lb(1):src_ub(1), &
                                              src_lb(2):src_ub(2)), &
                                  dst_sizes(1), dst_sizes(2))
         ELSE
            CALL dbcsr_block_copy(dst%d%c_sp(dst_lb(1):dst_ub(1)), &
                                  src%d%c2_sp(src_lb(1):src_ub(1), &
                                              src_lb(2):src_ub(2)), &
                                  dst_sizes(1), dst_sizes(2))
         ENDIF
      CASE (dbcsr_type_complex_8_2d)
         IF (dbcsr_type_is_2d(dst_dt)) THEN
            CALL dbcsr_block_copy(dst%d%c2_dp(dst_lb(1):dst_ub(1), &
                                              dst_lb(2):dst_ub(2)), &
                                  src%d%c2_dp(src_lb(1):src_ub(1), &
                                              src_lb(2):src_ub(2)), &
                                  dst_sizes(1), dst_sizes(2))
         ELSE
            CALL dbcsr_block_copy(dst%d%c_dp(dst_lb(1):dst_ub(1)), &
                                  src%d%c2_dp(src_lb(1):src_ub(1), &
                                              src_lb(2):src_ub(2)), &
                                  dst_sizes(1), dst_sizes(2))
         ENDIF
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1069,"Invalid data type")
      END SELECT
      CALL timestop(handle)
   END SUBROUTINE dbcsr_data_copy_aa2

! **************************************************************************************************
!> \brief Clears a data area, possibly transposed.
!> \param area ...
!> \param lb ...
!> \param ub ...
!> \param value ...
!> \param lb2 ...
!> \param ub2 ...
!> \param tr ...
! **************************************************************************************************
   SUBROUTINE dbcsr_data_clear_nt(area, lb, ub, value, lb2, ub2, tr)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area
      INTEGER, INTENT(IN), OPTIONAL                      :: lb, ub
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: value
      INTEGER, INTENT(IN), OPTIONAL                      :: lb2, ub2
      LOGICAL, INTENT(in)                                :: tr

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_clear_nt', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      IF (tr) THEN
         CALL dbcsr_data_clear0(area, lb=lb2, ub=ub2, value=value, lb2=lb, ub2=ub)
      ELSE
         CALL dbcsr_data_clear0(area, lb=lb, ub=ub, value=value, lb2=lb2, ub2=ub2)
      ENDIF
   END SUBROUTINE dbcsr_data_clear_nt

! **************************************************************************************************
!> \brief Clears a data area
!> \param[in,out] area   area with encapsulated data
!> \param[in] lb (optional) lower bound for clearing
!> \param[in] ub (optional) lower bound for clearing
!> \param[in] value      (optional) value to use for clearing
!> \param[in] lb2 (optional) upper bound for clearing
!> \param[in] ub2 (optional) upper bound for clearing
! **************************************************************************************************
   SUBROUTINE dbcsr_data_clear0(area, lb, ub, value, lb2, ub2)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: area
      INTEGER, INTENT(IN), OPTIONAL                      :: lb, ub
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: value
      INTEGER, INTENT(IN), OPTIONAL                      :: lb2, ub2

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_clear0', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: l, l2, s, u, u2

!   ---------------------------------------------------------------------------
! CALL timeset(routineN, handle)

      IF (.NOT. ASSOCIATED(area%d)) &
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1127,"Data area must be setup.")
      IF (PRESENT(value)) THEN
         IF (area%d%data_type .NE. value%data_type) &
            CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1130,"Incompatible data types")
      ENDIF

      IF (acc_devmem_allocated(area%d%acc_devmem)) THEN
         IF (PRESENT(value)) &
            CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1135,"dbcsr_data_clear0 with value not implemented for acc_devmem")
         s = dbcsr_datatype_sizeof(area%d%data_type)
         CALL acc_devmem_setzero_bytes(area%d%acc_devmem, s*lb, s*ub, area%d%memory_type%acc_stream)
      END IF

      SELECT CASE (area%d%data_type)
      CASE (dbcsr_type_real_4)
         l = LBOUND(area%d%r_sp, 1)
         u = UBOUND(area%d%r_sp, 1)
         IF (PRESENT(lb)) THEN
            CALL dbcsr_assert(lb, "GE", l, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower bound too low", 1146)
            l = lb
         ENDIF
         IF (PRESENT(ub)) THEN
            CALL dbcsr_assert(ub, "LE", u, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper bound too high", 1151)
            u = ub
         ENDIF
         IF (PRESENT(value)) THEN
            area%d%r_sp(l:u) = value%r_sp
         ELSE
            area%d%r_sp(l:u) = 0.0_real_4
         ENDIF
      CASE (dbcsr_type_real_8)
         l = LBOUND(area%d%r_dp, 1)
         u = UBOUND(area%d%r_dp, 1)
         IF (PRESENT(lb)) THEN
            CALL dbcsr_assert(lb, "GE", l, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower bound too low", 1164)
            l = lb
         ENDIF
         IF (PRESENT(ub)) THEN
            CALL dbcsr_assert(ub, "LE", u, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper bound too high", 1169)
            u = ub
         ENDIF
         IF (PRESENT(value)) THEN
            area%d%r_dp(l:u) = value%r_dp
         ELSE
            area%d%r_dp(l:u) = 0.0_real_8
         ENDIF
      CASE (dbcsr_type_complex_4)
         l = LBOUND(area%d%c_sp, 1)
         u = UBOUND(area%d%c_sp, 1)
         IF (PRESENT(lb)) THEN
            CALL dbcsr_assert(lb, "GE", l, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower bound too low", 1182)
            l = lb
         ENDIF
         IF (PRESENT(ub)) THEN
            CALL dbcsr_assert(ub, "LE", u, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper bound too high", 1187)
            u = ub
         ENDIF
         IF (PRESENT(value)) THEN
            area%d%c_sp(l:u) = value%c_sp
         ELSE
            area%d%c_sp(l:u) = CMPLX(0.0, 0.0, real_4)
         ENDIF
      CASE (dbcsr_type_complex_8)
         l = LBOUND(area%d%c_dp, 1)
         u = UBOUND(area%d%c_dp, 1)
         IF (PRESENT(lb)) THEN
            CALL dbcsr_assert(lb, "GE", l, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower bound too low", 1200)
            l = lb
         ENDIF
         IF (PRESENT(ub)) THEN
            CALL dbcsr_assert(ub, "LE", u, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper bound too high", 1205)
            u = ub
         ENDIF
         IF (PRESENT(value)) THEN
            area%d%c_dp(l:u) = value%c_dp
         ELSE
            area%d%c_dp(l:u) = CMPLX(0.0, 0.0, real_8)
         ENDIF
      CASE (dbcsr_type_real_4_2d)
         l = LBOUND(area%d%r2_sp, 1)
         u = UBOUND(area%d%r2_sp, 1)
         l2 = LBOUND(area%d%r2_sp, 2)
         u2 = UBOUND(area%d%r2_sp, 2)
         IF (PRESENT(lb)) THEN
            CALL dbcsr_assert(lb, "GE", l, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower bound too low", 1220)
            l = lb
         ENDIF
         IF (PRESENT(ub)) THEN
            CALL dbcsr_assert(ub, "LE", u, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper bound too high", 1225)
            u = ub
         ENDIF
         IF (PRESENT(lb2)) THEN
            CALL dbcsr_assert(lb2, "GE", l2, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower2 bound too low", 1230)
            l2 = lb2
         ENDIF
         IF (PRESENT(ub2)) THEN
            CALL dbcsr_assert(ub2, "LE", u2, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper2 bound too high", 1235)
            u2 = ub2
         ENDIF
         IF (PRESENT(value)) THEN
            area%d%r2_sp(l:u, l2:u2) = value%r_sp
         ELSE
            area%d%r2_sp(l:u, l2:u2) = 0.0_real_4
         ENDIF
      CASE (dbcsr_type_real_8_2d)
         l = LBOUND(area%d%r2_dp, 1)
         u = UBOUND(area%d%r2_dp, 1)
         l2 = LBOUND(area%d%r2_dp, 2)
         u2 = UBOUND(area%d%r2_dp, 2)
         IF (PRESENT(lb)) THEN
            CALL dbcsr_assert(lb, "GE", l, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower bound too low", 1250)
            l = lb
         ENDIF
         IF (PRESENT(ub)) THEN
            CALL dbcsr_assert(ub, "LE", u, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper bound too high", 1255)
            u = ub
         ENDIF
         IF (PRESENT(lb2)) THEN
            CALL dbcsr_assert(lb2, "GE", l2, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower2 bound too low", 1260)
            l2 = lb2
         ENDIF
         IF (PRESENT(ub2)) THEN
            CALL dbcsr_assert(ub2, "LE", u2, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper2 bound too high", 1265)
            u2 = ub2
         ENDIF
         IF (PRESENT(value)) THEN
            area%d%r2_dp(l:u, l2:u2) = value%r_dp
         ELSE
            area%d%r2_dp(l:u, l2:u2) = 0.0_real_8
         ENDIF
      CASE (dbcsr_type_complex_4_2d)
         l = LBOUND(area%d%c2_sp, 1)
         u = UBOUND(area%d%c2_sp, 1)
         l2 = LBOUND(area%d%c2_sp, 2)
         u2 = UBOUND(area%d%c2_sp, 2)
         IF (PRESENT(lb)) THEN
            CALL dbcsr_assert(lb, "GE", l, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower bound too low", 1280)
            l = lb
         ENDIF
         IF (PRESENT(ub)) THEN
            CALL dbcsr_assert(ub, "LE", u, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper bound too high", 1285)
            u = ub
         ENDIF
         IF (PRESENT(lb2)) THEN
            CALL dbcsr_assert(lb2, "GE", l2, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower2 bound too low", 1290)
            l2 = lb2
         ENDIF
         IF (PRESENT(ub2)) THEN
            CALL dbcsr_assert(ub2, "LE", u2, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper2 bound too high", 1295)
            u2 = ub2
         ENDIF
         IF (PRESENT(value)) THEN
            area%d%c2_sp(l:u, l2:u2) = value%c_sp
         ELSE
            area%d%c2_sp(l:u, l2:u2) = CMPLX(0.0, 0.0, real_4)
         ENDIF
      CASE (dbcsr_type_complex_8_2d)
         l = LBOUND(area%d%c2_dp, 1)
         u = UBOUND(area%d%c2_dp, 1)
         l2 = LBOUND(area%d%c2_dp, 2)
         u2 = UBOUND(area%d%c2_dp, 2)
         IF (PRESENT(lb)) THEN
            CALL dbcsr_assert(lb, "GE", l, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower bound too low", 1310)
            l = lb
         ENDIF
         IF (PRESENT(ub)) THEN
            CALL dbcsr_assert(ub, "LE", u, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper bound too high", 1315)
            u = ub
         ENDIF
         IF (PRESENT(lb2)) THEN
            CALL dbcsr_assert(lb2, "GE", l2, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "lower2 bound too low", 1320)
            l2 = lb2
         ENDIF
         IF (PRESENT(ub2)) THEN
            CALL dbcsr_assert(ub2, "LE", u2, dbcsr_fatal_level, &
                              dbcsr_caller_error, routineN, "upper2 bound too high", 1325)
            u2 = ub2
         ENDIF
         IF (PRESENT(value)) THEN
            area%d%c2_dp(l:u, l2:u2) = value%c_dp
         ELSE
            area%d%c2_dp(l:u, l2:u2) = CMPLX(0.0, 0.0, real_8)
         ENDIF
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1334,"Invalid or unsupported data type.")
      END SELECT

      ! CALL timestop(handle)

   END SUBROUTINE dbcsr_data_clear0

! **************************************************************************************************
!> \brief Copies a block subset
!> \param[in,out] dst  target data area
!> \param[in] dst_rs   target block row size (logical)
!> \param[in] dst_cs   target block column size (logical)
!> \param[in] dst_tr   whether target block is transposed
!> \param[in] src      source data area
!> \param[in] src_rs   source block row size (logical)
!> \param[in] src_cs   source block column size (logical)
!> \param[in] src_tr   whether source block is transposed
!> \param[in] dst_r_lb   first row in target
!> \param[in] dst_c_lb   first column in target
!> \param[in] src_r_lb   first_row in source
!> \param[in] src_c_lb   first column in target
!> \param[in] nrow       number of rows to copy
!> \param[in] ncol       number of columns to copy
!> \param[in] dst_offset offset in target
!> \param[in] src_offset offset in source
! **************************************************************************************************
   SUBROUTINE dbcsr_block_partial_copy(dst, dst_rs, dst_cs, dst_tr, &
                                       src, src_rs, src_cs, src_tr, &
                                       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                       dst_offset, src_offset)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: dst
      INTEGER, INTENT(IN)                                :: dst_rs, dst_cs
      LOGICAL                                            :: dst_tr
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: src
      INTEGER, INTENT(IN)                                :: src_rs, src_cs
      LOGICAL                                            :: src_tr
      INTEGER, INTENT(IN)                                :: dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, &
                                                            nrow, ncol
      INTEGER, INTENT(IN), OPTIONAL                      :: dst_offset, src_offset

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_block_partial_copy', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: verification = careful_mod

      INTEGER                                            :: dst_o, src_o
      LOGICAL                                            :: src_is_2d

!   ---------------------------------------------------------------------------

      IF (careful_mod) &
         CALL dbcsr_assert(dbcsr_type_2d_to_1d(dbcsr_data_get_type(dst)), &
                           "EQ", dbcsr_type_2d_to_1d(dbcsr_data_get_type(src)), &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Incompatible data types.", 1387)
      dst_o = 0; src_o = 0
      IF (PRESENT(dst_offset)) dst_o = dst_offset
      IF (PRESENT(src_offset)) src_o = src_offset
      IF (verification) THEN
         CALL dbcsr_assert(dst_r_lb+nrow-1, "LE", dst_rs, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, "Incompatible dst row sizes", 1393)
         CALL dbcsr_assert(dst_c_lb+ncol-1, "LE", dst_cs, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, "Incompatible dst col sizes", 1395)
         CALL dbcsr_assert(src_r_lb+nrow-1, "LE", src_rs, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, "Incompatible src row sizes", 1397)
         CALL dbcsr_assert(src_c_lb+ncol-1, "LE", src_cs, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, "Incompatible src col sizes", 1399)
      ENDIF
      !
      src_is_2d = dbcsr_type_is_2d(dbcsr_data_get_type(src))
      SELECT CASE (dbcsr_data_get_type (dst))
      CASE (dbcsr_type_real_4)
         IF (src_is_2d) THEN
            CALL block_partial_copy_1d2d_s(dst%d%r_sp, dst_rs, dst_cs, dst_tr, &
                                           src%d%r2_sp, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                           dst_offset=dst_o)
         ELSE
            CALL block_partial_copy_s(dst%d%r_sp, dst_rs, dst_cs, dst_tr, &
                                      src%d%r_sp, src_rs, src_cs, src_tr, &
                                      dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                      dst_offset=dst_o, src_offset=src_o)
         ENDIF
      CASE (dbcsr_type_real_8)
         IF (src_is_2d) THEN
            CALL block_partial_copy_1d2d_d(dst%d%r_dp, dst_rs, dst_cs, dst_tr, &
                                           src%d%r2_dp, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                           dst_offset=dst_o)
         ELSE
            CALL block_partial_copy_d(dst%d%r_dp, dst_rs, dst_cs, dst_tr, &
                                      src%d%r_dp, src_rs, src_cs, src_tr, &
                                      dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                      dst_offset=dst_o, src_offset=src_o)
         ENDIF
      CASE (dbcsr_type_complex_4)
         IF (src_is_2d) THEN
            CALL block_partial_copy_1d2d_c(dst%d%c_sp, dst_rs, dst_cs, dst_tr, &
                                           src%d%c2_sp, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                           dst_offset=dst_o)
         ELSE
            CALL block_partial_copy_c(dst%d%c_sp, dst_rs, dst_cs, dst_tr, &
                                      src%d%c_sp, src_rs, src_cs, src_tr, &
                                      dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                      dst_offset=dst_o, src_offset=src_o)
         ENDIF
      CASE (dbcsr_type_complex_8)
         IF (src_is_2d) THEN
            CALL block_partial_copy_1d2d_z(dst%d%c_dp, dst_rs, dst_cs, dst_tr, &
                                           src%d%c2_dp, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                           dst_offset=dst_o)
         ELSE
            CALL block_partial_copy_z(dst%d%c_dp, dst_rs, dst_cs, dst_tr, &
                                      src%d%c_dp, src_rs, src_cs, src_tr, &
                                      dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                      dst_offset=dst_o, src_offset=src_o)
         ENDIF
      CASE (dbcsr_type_real_4_2d)
         IF (src_is_2d) THEN
            CALL block_partial_copy_2d2d_s(dst%d%r2_sp, dst_tr, &
                                           src%d%r2_sp, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol)
         ELSE
            CALL block_partial_copy_2d1d_s(dst%d%r2_sp, dst_tr, &
                                           src%d%r_sp, src_rs, src_cs, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                           src_offset=src_o)
         ENDIF
      CASE (dbcsr_type_real_8_2d)
         IF (src_is_2d) THEN
            CALL block_partial_copy_2d2d_d(dst%d%r2_dp, dst_tr, &
                                           src%d%r2_dp, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol)
         ELSE
            CALL block_partial_copy_2d1d_d(dst%d%r2_dp, dst_tr, &
                                           src%d%r_dp, src_rs, src_cs, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                           src_offset=src_o)
         ENDIF
      CASE (dbcsr_type_complex_4_2d)
         IF (src_is_2d) THEN
            CALL block_partial_copy_2d2d_c(dst%d%c2_sp, dst_tr, &
                                           src%d%c2_sp, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol)
         ELSE
            CALL block_partial_copy_2d1d_c(dst%d%c2_sp, dst_tr, &
                                           src%d%c_sp, src_rs, src_cs, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                           src_offset=src_o)
         ENDIF
      CASE (dbcsr_type_complex_8_2d)
         IF (src_is_2d) THEN
            CALL block_partial_copy_2d2d_z(dst%d%c2_dp, dst_tr, &
                                           src%d%c2_dp, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol)
         ELSE
            CALL block_partial_copy_2d1d_z(dst%d%c2_dp, dst_tr, &
                                           src%d%c_dp, src_rs, src_cs, src_tr, &
                                           dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol, &
                                           src_offset=src_o)
         ENDIF
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1497,"Invalid data type.")
      END SELECT
   END SUBROUTINE dbcsr_block_partial_copy

! **************************************************************************************************
!> \brief Sets the diagonal of a square data block
!> \param[in,out] block_data  sets the diagonal of this data block
!> \param[in] diagonal        set diagonal of block_data to these values
!> \param[in] d               block size
!> \par Off-diagonal values
!>      Other values are untouched.
! **************************************************************************************************
   SUBROUTINE set_block2d_diagonal_a(block_data, diagonal, d)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: block_data
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: diagonal
      INTEGER, INTENT(IN)                                :: d

      CHARACTER(len=*), PARAMETER :: routineN = 'set_block2d_diagonal_a', &
         routineP = moduleN//':'//routineN

      COMPLEX(kind=dp), DIMENSION(:, :), POINTER         :: c2_dp
      COMPLEX(kind=sp), DIMENSION(:, :), POINTER         :: c2_sp
      INTEGER                                            :: dt_block, dt_diag
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: r2_dp
      REAL(kind=sp), DIMENSION(:, :), POINTER            :: r2_sp

!   ---------------------------------------------------------------------------

      dt_block = dbcsr_data_get_type(block_data)
      dt_diag = dbcsr_data_get_type(diagonal)
      IF (.NOT. dbcsr_type_is_2d(dt_block)) &
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1528,"Block must be 2-d")
      CALL dbcsr_assert("NOT", dbcsr_type_is_2d(dt_diag), dbcsr_fatal_level, &
                        dbcsr_internal_error, routineN, "Diagonal can not be 2-d", 1530)
      SELECT CASE (dt_diag)
      CASE (dbcsr_type_real_4)
         r2_sp => dbcsr_get_data_p_2d_s(block_data)
         CALL set_block2d_diagonal(r2_sp, &
                                   dbcsr_get_data_p_s(diagonal), d)
      CASE (dbcsr_type_real_8)
         r2_dp => dbcsr_get_data_p_2d_d(block_data)
         CALL set_block2d_diagonal(r2_dp, &
                                   dbcsr_get_data_p_d(diagonal), d)
      CASE (dbcsr_type_complex_4)
         c2_sp => dbcsr_get_data_p_2d_c(block_data)
         CALL set_block2d_diagonal(c2_sp, &
                                   dbcsr_get_data_p_c(diagonal), d)
      CASE (dbcsr_type_complex_8)
         c2_dp => dbcsr_get_data_p_2d_z(block_data)
         CALL set_block2d_diagonal(c2_dp, &
                                   dbcsr_get_data_p_z(diagonal), d)
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1549,"Invalid data type!")
      END SELECT
   END SUBROUTINE set_block2d_diagonal_a

! **************************************************************************************************
!> \brief Gets the diagonal of a square data block
!> \param[in] block_data      gets the diagonal of this data block
!> \param[out] diagonal       values of the diagonal elements
!> \param[in] d               dimension of block
! **************************************************************************************************
   SUBROUTINE get_block2d_diagonal_a(block_data, diagonal, d)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: block_data, diagonal
      INTEGER, INTENT(IN)                                :: d

      CHARACTER(len=*), PARAMETER :: routineN = 'get_block2d_diagonal_a', &
         routineP = moduleN//':'//routineN

      COMPLEX(kind=dp), DIMENSION(:), POINTER            :: c_dp
      COMPLEX(kind=dp), DIMENSION(:, :), POINTER         :: c2_dp
      COMPLEX(kind=sp), DIMENSION(:), POINTER            :: c_sp
      COMPLEX(kind=sp), DIMENSION(:, :), POINTER         :: c2_sp
      INTEGER                                            :: dt_block, dt_diag
      REAL(kind=dp), DIMENSION(:), POINTER               :: r_dp
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: r2_dp
      REAL(kind=sp), DIMENSION(:), POINTER               :: r_sp
      REAL(kind=sp), DIMENSION(:, :), POINTER            :: r2_sp

!   ---------------------------------------------------------------------------

      dt_block = dbcsr_data_get_type(block_data)
      dt_diag = dbcsr_data_get_type(diagonal)
      IF (.NOT. dbcsr_type_is_2d(dt_block)) &
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1581,"Block must be 2-d")
      CALL dbcsr_assert("NOT", dbcsr_type_is_2d(dt_diag), dbcsr_fatal_level, &
                        dbcsr_internal_error, routineN, "Diagonal can not be 2-d", 1583)
      SELECT CASE (dt_diag)
      CASE (dbcsr_type_real_4)
         r2_sp => dbcsr_get_data_p_2d_s(block_data)
         r_sp => dbcsr_get_data_p_s(diagonal)
         CALL get_block2d_diagonal_s(r2_sp, r_sp, d)
      CASE (dbcsr_type_real_8)
         r2_dp => dbcsr_get_data_p_2d_d(block_data)
         r_dp => dbcsr_get_data_p_d(diagonal)
         CALL get_block2d_diagonal_d(r2_dp, r_dp, d)
      CASE (dbcsr_type_complex_4)
         c2_sp => dbcsr_get_data_p_2d_c(block_data)
         c_sp => dbcsr_get_data_p_c(diagonal)
         CALL get_block2d_diagonal_c(c2_sp, c_sp, d)
      CASE (dbcsr_type_complex_8)
         c2_dp => dbcsr_get_data_p_2d_z(block_data)
         c_dp => dbcsr_get_data_p_z(diagonal)
         CALL get_block2d_diagonal_z(c2_dp, c_dp, d)
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1602,"Invalid data type!")
      END SELECT
   END SUBROUTINE get_block2d_diagonal_a

! **************************************************************************************************
!> \brief Adds a value to the diagonal of the block
!> \param[in,out] block_a      Sets the diagonal of this block
!> \param[in] alpha_scalar     Adds this value to the diagonal elements of the
!>                             block
!> \param[in] block_dimension  Dimension of the block
!> \param imin ...
!> \param imax ...
! **************************************************************************************************
   SUBROUTINE block_add_on_diag_anytype(block_a, alpha_scalar, block_dimension, &
                                        imin, imax)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: block_a
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: alpha_scalar
      INTEGER, INTENT(IN)                                :: block_dimension
      INTEGER, INTENT(in), OPTIONAL                      :: imin, imax

      CHARACTER(len=*), PARAMETER :: routineN = 'block_add_on_diag_anytype', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      CALL dbcsr_assert(dbcsr_type_2d_to_1d(dbcsr_data_get_type(block_a)), "EQ", &
                        alpha_scalar%data_type, dbcsr_fatal_level, dbcsr_wrong_args_error, &
                        routineN, "Mismatched data types.", 1629)
      SELECT CASE (dbcsr_data_get_type (block_a))
      CASE (dbcsr_type_real_4)
         CALL block_add_on_diag( &
            block_dimension, block_a%d%r_sp, alpha_scalar%r_sp, imin, imax)
      CASE (dbcsr_type_real_8)
         CALL block_add_on_diag( &
            block_dimension, block_a%d%r_dp, alpha_scalar%r_dp, imin, imax)
      CASE (dbcsr_type_complex_4)
         CALL block_add_on_diag( &
            block_dimension, block_a%d%c_sp, alpha_scalar%c_sp, imin, imax)
      CASE (dbcsr_type_complex_8)
         CALL block_add_on_diag( &
            block_dimension, block_a%d%c_dp, alpha_scalar%c_dp, imin, imax)
      CASE (dbcsr_type_real_4_2d)
         CALL block_add_on_diag( &
            block_dimension, block_a%d%r2_sp, alpha_scalar%r_sp, imin, imax)
      CASE (dbcsr_type_real_8_2d)
         CALL block_add_on_diag( &
            block_dimension, block_a%d%r2_dp, alpha_scalar%r_dp, imin, imax)
      CASE (dbcsr_type_complex_4_2d)
         CALL block_add_on_diag( &
            block_dimension, block_a%d%c2_sp, alpha_scalar%c_sp, imin, imax)
      CASE (dbcsr_type_complex_8_2d)
         CALL block_add_on_diag( &
            block_dimension, block_a%d%c2_dp, alpha_scalar%c_dp, imin, imax)
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1656,"Invalid data type!")
      END SELECT
   END SUBROUTINE block_add_on_diag_anytype

! **************************************************************************************************
!> \brief Adds two blocks
!> \param[in,out] block_a      Block to add to
!> \param[in] block_b          Block to add to block_a
!> \param len ...
! **************************************************************************************************
   SUBROUTINE block_add_anytype(block_a, block_b, len)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: block_a
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: block_b
      INTEGER, INTENT(IN), OPTIONAL                      :: len

      CHARACTER(len=*), PARAMETER :: routineN = 'block_add_anytype', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: n

!   ---------------------------------------------------------------------------

      IF (careful_mod) &
         CALL dbcsr_assert(dbcsr_data_get_type(block_a), "EQ", &
                           dbcsr_data_get_type(block_a), &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Mismatched data types.", 1682)
      IF (PRESENT(len)) THEN
         n = len
         CALL dbcsr_assert(dbcsr_data_get_size(block_b), "GE", n, &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Block B too small.", 1687)
      ELSE
         n = dbcsr_data_get_size_referenced(block_b)
      ENDIF
      CALL dbcsr_assert(dbcsr_data_get_size(block_a), "GE", n, &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "Block A too small.", 1693)
      SELECT CASE (dbcsr_data_get_type (block_a))
      CASE (dbcsr_type_real_4)
         CALL block_add_s(block_a%d%r_sp, block_b%d%r_sp, n)
      CASE (dbcsr_type_real_8)
         CALL block_add_d(block_a%d%r_dp, block_b%d%r_dp, n)
      CASE (dbcsr_type_complex_4)
         CALL block_add_c(block_a%d%c_sp, block_b%d%c_sp, n)
      CASE (dbcsr_type_complex_8)
         CALL block_add_z(block_a%d%c_dp, block_b%d%c_dp, n)
      CASE (dbcsr_type_real_4_2d)
         CALL block_add_s(block_a%d%r2_sp, block_b%d%r2_sp, n)
      CASE (dbcsr_type_real_8_2d)
         CALL block_add_d(block_a%d%r2_dp, block_b%d%r2_dp, n)
      CASE (dbcsr_type_complex_4_2d)
         CALL block_add_c(block_a%d%c2_sp, block_b%d%c2_sp, n)
      CASE (dbcsr_type_complex_8_2d)
         CALL block_add_z(block_a%d%c2_dp, block_b%d%c2_dp, n)
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_operations.F",1712,"Invalid data type!")
      END SELECT
   END SUBROUTINE block_add_anytype

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Sets the diagonal of a square data block
!> \param[out] block_data     sets the diagonal of this data block
!> \param[in] diagonal        set diagonal of block_data to these values
!> \param[in] d               dimension of block
!> \par Off-diagonal values
!>      Other values are untouched.
! **************************************************************************************************
  PURE SUBROUTINE set_block2d_diagonal_d (block_data, diagonal, d)
    INTEGER, INTENT(IN)                    :: d
    REAL(kind=real_8), DIMENSION(d,d), INTENT(INOUT) :: block_data
    REAL(kind=real_8), DIMENSION(d), INTENT(IN)      :: diagonal

    INTEGER                                :: i

!   ---------------------------------------------------------------------------

    DO i = 1 , d
       block_data(i,i) = diagonal(i)
    END DO
  END SUBROUTINE set_block2d_diagonal_d


! **************************************************************************************************
!> \brief Gets the diagonal of a square data block
!> \param[in] block_data      gets the diagonal of this data block
!> \param[out] diagonal       values of the diagonal elements
!> \param[in] d               dimension of block
! **************************************************************************************************
  PURE SUBROUTINE get_block2d_diagonal_d (block_data, diagonal, d)
    INTEGER, INTENT(IN)                 :: d
    REAL(kind=real_8), DIMENSION(d,d), INTENT(IN) :: block_data
    REAL(kind=real_8), DIMENSION(d), INTENT(OUT)  :: diagonal

    INTEGER                             :: i

!   ---------------------------------------------------------------------------

    DO i = 1 , d
       diagonal(i) = block_data(i, i)
    END DO
  END SUBROUTINE get_block2d_diagonal_d


! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_rs ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_rs ...
!> \param src_cs ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param dst_offset ...
!> \param src_offset ...
!> \note see block_partial_copy_a 
! **************************************************************************************************
  SUBROUTINE block_partial_copy_d(dst, dst_rs, dst_cs, dst_tr,&
       src, src_rs, src_cs, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       dst_offset, src_offset)
    REAL(kind=real_8), DIMENSION(:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: dst_rs, dst_cs
    INTEGER, INTENT(IN)                      :: src_offset, dst_offset
    LOGICAL                                  :: dst_tr
    REAL(kind=real_8), DIMENSION(:), &
      INTENT(IN)                             :: src
    INTEGER, INTENT(IN)                      :: src_rs, src_cs
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop.
!  rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
                = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs) &
              = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
             = src(src_offset+src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs)&
             = src(src_offset + src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_d

! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_rs ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param dst_offset ...
!> \note see block_partial_copy_a 
! **************************************************************************************************
  SUBROUTINE block_partial_copy_1d2d_d(dst, dst_rs, dst_cs, dst_tr,&
       src, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       dst_offset)
    REAL(kind=real_8), DIMENSION(:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: dst_rs, dst_cs
    INTEGER, INTENT(IN)                      :: dst_offset
    LOGICAL                                  :: dst_tr
    REAL(kind=real_8), DIMENSION(:,:), &
      INTENT(IN)                             :: src
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_1d2d_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
                = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs) &
              = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs)&
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_1d2d_d
! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_rs ...
!> \param src_cs ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param src_offset ...
!> \note see block_partial_copy_a
! **************************************************************************************************
  SUBROUTINE block_partial_copy_2d1d_d(dst, dst_tr,&
       src, src_rs, src_cs, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       src_offset)
    REAL(kind=real_8), DIMENSION(:,:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: src_offset
    LOGICAL                                  :: dst_tr
    REAL(kind=real_8), DIMENSION(:), &
      INTENT(IN)                             :: src
    INTEGER, INTENT(IN)                      :: src_rs, src_cs
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_2d1d_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
                = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row) &
              = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
             = src(src_offset+src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row)&
             = src(src_offset + src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_2d1d_d
! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_tr ...
!> \param src ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \note see block_partial_copy_a
! **************************************************************************************************
  SUBROUTINE block_partial_copy_2d2d_d(dst, dst_tr,&
       src, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol)
    REAL(kind=real_8), DIMENSION(:,:), &
      INTENT(INOUT)                          :: dst
    LOGICAL                                  :: dst_tr
    REAL(kind=real_8), DIMENSION(:,:), &
      INTENT(IN)                             :: src
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_2d2d_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
                = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row) &
              = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row)&
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_2d2d_d


! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output data
!> \param[in] extent_in       input data
!> \param[in] n               number of elements to copy
!> \param[in] out_fe          first element of output
!> \param[in] in_fe           first element of input
! **************************************************************************************************
  PURE SUBROUTINE block_copy_d(extent_out, extent_in, n, out_fe, in_fe)
    INTEGER, INTENT(IN)                           :: n, out_fe, in_fe
    REAL(kind=real_8), DIMENSION(*), INTENT(OUT)  :: extent_out
    REAL(kind=real_8), DIMENSION(*), INTENT(IN)   :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(out_fe : out_fe+n-1) = extent_in(in_fe : in_fe+n-1)
  END SUBROUTINE block_copy_d


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_d(extent_out, extent_in,&
       rows, columns)
    REAL(kind=real_8), DIMENSION(:), INTENT(OUT) :: extent_out
    REAL(kind=real_8), DIMENSION(:), INTENT(IN)  :: extent_in
    INTEGER, INTENT(IN)                :: rows, columns

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(1:rows*columns) = RESHAPE(TRANSPOSE(&
         RESHAPE(extent_in(1:rows*columns), (/rows, columns/))), (/rows*columns/))
  END SUBROUTINE block_transpose_copy_d

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_2d1d_d(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_8), DIMENSION(rows,columns), INTENT(OUT) :: extent_out
    REAL(kind=real_8), DIMENSION(:), INTENT(IN)             :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d1d_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(extent_in, (/rows, columns/))
  END SUBROUTINE block_copy_2d1d_d

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_1d1d_d(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_8), DIMENSION(rows*columns), INTENT(OUT) :: extent_out
    REAL(kind=real_8), DIMENSION(rows*columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_1d1d_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(:) = extent_in(:)
  END SUBROUTINE block_copy_1d1d_d

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_2d2d_d(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_8), DIMENSION(rows,columns), INTENT(OUT) :: extent_out
    REAL(kind=real_8), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d2d_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(:,:) = extent_in(:,:)
  END SUBROUTINE block_copy_2d2d_d


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_2d1d_d(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_8), DIMENSION(columns,rows), INTENT(OUT) :: extent_out
    REAL(kind=real_8), DIMENSION(:), INTENT(IN)             :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d1d_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = TRANSPOSE(RESHAPE(extent_in, (/rows, columns/)))
  END SUBROUTINE block_transpose_copy_2d1d_d

! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_1d2d_d(extent_out, extent_in,&
       rows, columns)
    REAL(kind=real_8), DIMENSION(:), INTENT(OUT)            :: extent_out
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_8), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_copy_1d2d_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(extent_in, (/rows*columns/))
  END SUBROUTINE block_copy_1d2d_d


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_1d2d_d(extent_out, extent_in,&
       rows, columns)
    REAL(kind=real_8), DIMENSION(:), INTENT(OUT)            :: extent_out
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_8), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_1d2d_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(TRANSPOSE(extent_in), (/rows*columns/))
  END SUBROUTINE block_transpose_copy_1d2d_d


! **************************************************************************************************
!> \brief In-place block transpose.
!> \param[in,out] extent      Matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_inplace_d(extent, rows, columns)
    INTEGER, INTENT(IN)                      :: rows, columns
    REAL(kind=real_8), DIMENSION(rows*columns), &
      INTENT(INOUT)                          :: extent
    REAL(kind=real_8), DIMENSION(rows*columns)         :: extent_tr

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_inplace_d', &
      routineP = moduleN//':'//routineN

    INTEGER :: r, c
!   ---------------------------------------------------------------------------

    DO r = 1 , columns
      DO c = 1 , rows
       extent_tr(r + (c-1)*columns) = extent(c + (r-1)*rows)
      END DO
    END DO
    DO r = 1 , columns
      DO c = 1 , rows
       extent(r + (c-1)*columns) = extent_tr(r + (c-1)*columns)
      END DO
    END DO
  END SUBROUTINE block_transpose_inplace_d


! **************************************************************************************************
!> \brief Copy data from a double real array to a data area
!>
!> There are no checks done for correctness!
!> \param[in] dst        destination data area
!> \param[in] lb         lower bound for destination (and source if
!>                       not given explicity)
!> \param[in] data_size  number of elements to copy
!> \param[in] src        source data array
!> \param[in] source_lb  (optional) lower bound of source
! **************************************************************************************************
  SUBROUTINE dbcsr_data_set_ad (dst, lb, data_size, src, source_lb)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: dst
    INTEGER, INTENT(IN)                      :: lb, data_size
    REAL(kind=real_8), DIMENSION(:), INTENT(IN)        :: src
    INTEGER, INTENT(IN), OPTIONAL            :: source_lb
    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_set_ad', &
         routineP = moduleN//':'//routineN
    INTEGER                                  :: lb_s, ub, ub_s
!   ---------------------------------------------------------------------------
    IF (debug_mod) THEN
       IF(.NOT.ASSOCIATED (dst%d))&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",555,"Target data area must be setup.")
       IF(SIZE(src) .LT. data_size)&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",557,"Not enough source data.")
       IF(dst%d%data_type .NE. dbcsr_type_real_8)&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",559,"Data type mismatch.")
    ENDIF
    ub = lb + data_size - 1
    IF (PRESENT (source_lb)) THEN
       lb_s = source_lb
       ub_s = source_lb + data_size-1
    ELSE
       lb_s = lb
       ub_s = ub
    ENDIF
    CALL memory_copy(dst%d%r_dp(lb:ub),src(lb_s:ub_s),data_size)
  END SUBROUTINE dbcsr_data_set_ad

! **************************************************************************************************
!> \brief ...
!> \param m ...
!> \param blk ...
!> \param alpha ...
!> \param imin ...
!> \param imax ...
! **************************************************************************************************
  PURE SUBROUTINE block_2d_add_on_diag_d(m, blk, alpha, imin, imax)
    INTEGER, INTENT(IN)                      :: m
    REAL(kind=real_8), INTENT(INOUT), DIMENSION(m,m)   :: blk
    REAL(kind=real_8), INTENT(IN)                      :: alpha
    INTEGER, INTENT(IN), OPTIONAL            :: imin, imax

    CHARACTER(len=*), PARAMETER :: routineN = 'block_2d_add_on_diag_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i

!   ---------------------------------------------------------------------------

    IF(PRESENT(imin).AND.PRESENT(imax)) THEN
       DO i = MAX(1,imin),MIN(m,imax)
         blk(i,i) = blk(i,i) + alpha
       ENDDO
    ELSE
       DO i = 1,m
          blk(i,i) = blk(i,i) + alpha
       END DO
    ENDIF
  END SUBROUTINE block_2d_add_on_diag_d

! **************************************************************************************************
!> \brief ...
!> \param m ...
!> \param blk ...
!> \param alpha ...
!> \param imin ...
!> \param imax ...
! **************************************************************************************************
  PURE SUBROUTINE block_add_on_diag_d(m, blk, alpha, imin, imax)
    INTEGER, INTENT(IN)                      :: m
    REAL(kind=real_8), INTENT(INOUT), DIMENSION(m*m)   :: blk
    REAL(kind=real_8), INTENT(IN)                      :: alpha
    INTEGER, INTENT(IN), OPTIONAL            :: imin, imax

    CHARACTER(len=*), PARAMETER :: routineN = 'block_add_on_diag_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    IF(PRESENT(imin).AND.PRESENT(imax)) THEN
       CALL block_2d_add_on_diag_d(m, blk, alpha, imin, imax)
    ELSE
       CALL block_2d_add_on_diag_d(m, blk, alpha, 1, m)
    ENDIF
  END SUBROUTINE block_add_on_diag_d

! **************************************************************************************************
!> \brief ...
!> \param block_a ...
!> \param block_b ...
!> \param len ...
! **************************************************************************************************
  PURE SUBROUTINE block_add_d(block_a, block_b, len)
    INTEGER, INTENT(IN)                    :: len
    REAL(kind=real_8), DIMENSION(len), INTENT(INOUT) :: block_a
    REAL(kind=real_8), DIMENSION(len), INTENT(IN)    :: block_b
    block_a(1:len) = block_a(1:len) + block_b(1:len)
  END SUBROUTINE block_add_d
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Sets the diagonal of a square data block
!> \param[out] block_data     sets the diagonal of this data block
!> \param[in] diagonal        set diagonal of block_data to these values
!> \param[in] d               dimension of block
!> \par Off-diagonal values
!>      Other values are untouched.
! **************************************************************************************************
  PURE SUBROUTINE set_block2d_diagonal_z (block_data, diagonal, d)
    INTEGER, INTENT(IN)                    :: d
    COMPLEX(kind=real_8), DIMENSION(d,d), INTENT(INOUT) :: block_data
    COMPLEX(kind=real_8), DIMENSION(d), INTENT(IN)      :: diagonal

    INTEGER                                :: i

!   ---------------------------------------------------------------------------

    DO i = 1 , d
       block_data(i,i) = diagonal(i)
    END DO
  END SUBROUTINE set_block2d_diagonal_z


! **************************************************************************************************
!> \brief Gets the diagonal of a square data block
!> \param[in] block_data      gets the diagonal of this data block
!> \param[out] diagonal       values of the diagonal elements
!> \param[in] d               dimension of block
! **************************************************************************************************
  PURE SUBROUTINE get_block2d_diagonal_z (block_data, diagonal, d)
    INTEGER, INTENT(IN)                 :: d
    COMPLEX(kind=real_8), DIMENSION(d,d), INTENT(IN) :: block_data
    COMPLEX(kind=real_8), DIMENSION(d), INTENT(OUT)  :: diagonal

    INTEGER                             :: i

!   ---------------------------------------------------------------------------

    DO i = 1 , d
       diagonal(i) = block_data(i, i)
    END DO
  END SUBROUTINE get_block2d_diagonal_z


! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_rs ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_rs ...
!> \param src_cs ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param dst_offset ...
!> \param src_offset ...
!> \note see block_partial_copy_a 
! **************************************************************************************************
  SUBROUTINE block_partial_copy_z(dst, dst_rs, dst_cs, dst_tr,&
       src, src_rs, src_cs, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       dst_offset, src_offset)
    COMPLEX(kind=real_8), DIMENSION(:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: dst_rs, dst_cs
    INTEGER, INTENT(IN)                      :: src_offset, dst_offset
    LOGICAL                                  :: dst_tr
    COMPLEX(kind=real_8), DIMENSION(:), &
      INTENT(IN)                             :: src
    INTEGER, INTENT(IN)                      :: src_rs, src_cs
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop.
!  rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
                = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs) &
              = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
             = src(src_offset+src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs)&
             = src(src_offset + src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_z

! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_rs ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param dst_offset ...
!> \note see block_partial_copy_a 
! **************************************************************************************************
  SUBROUTINE block_partial_copy_1d2d_z(dst, dst_rs, dst_cs, dst_tr,&
       src, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       dst_offset)
    COMPLEX(kind=real_8), DIMENSION(:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: dst_rs, dst_cs
    INTEGER, INTENT(IN)                      :: dst_offset
    LOGICAL                                  :: dst_tr
    COMPLEX(kind=real_8), DIMENSION(:,:), &
      INTENT(IN)                             :: src
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_1d2d_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
                = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs) &
              = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs)&
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_1d2d_z
! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_rs ...
!> \param src_cs ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param src_offset ...
!> \note see block_partial_copy_a
! **************************************************************************************************
  SUBROUTINE block_partial_copy_2d1d_z(dst, dst_tr,&
       src, src_rs, src_cs, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       src_offset)
    COMPLEX(kind=real_8), DIMENSION(:,:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: src_offset
    LOGICAL                                  :: dst_tr
    COMPLEX(kind=real_8), DIMENSION(:), &
      INTENT(IN)                             :: src
    INTEGER, INTENT(IN)                      :: src_rs, src_cs
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_2d1d_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
                = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row) &
              = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
             = src(src_offset+src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row)&
             = src(src_offset + src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_2d1d_z
! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_tr ...
!> \param src ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \note see block_partial_copy_a
! **************************************************************************************************
  SUBROUTINE block_partial_copy_2d2d_z(dst, dst_tr,&
       src, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol)
    COMPLEX(kind=real_8), DIMENSION(:,:), &
      INTENT(INOUT)                          :: dst
    LOGICAL                                  :: dst_tr
    COMPLEX(kind=real_8), DIMENSION(:,:), &
      INTENT(IN)                             :: src
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_2d2d_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
                = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row) &
              = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row)&
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_2d2d_z


! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output data
!> \param[in] extent_in       input data
!> \param[in] n               number of elements to copy
!> \param[in] out_fe          first element of output
!> \param[in] in_fe           first element of input
! **************************************************************************************************
  PURE SUBROUTINE block_copy_z(extent_out, extent_in, n, out_fe, in_fe)
    INTEGER, INTENT(IN)                           :: n, out_fe, in_fe
    COMPLEX(kind=real_8), DIMENSION(*), INTENT(OUT)  :: extent_out
    COMPLEX(kind=real_8), DIMENSION(*), INTENT(IN)   :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(out_fe : out_fe+n-1) = extent_in(in_fe : in_fe+n-1)
  END SUBROUTINE block_copy_z


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_z(extent_out, extent_in,&
       rows, columns)
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(OUT) :: extent_out
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN)  :: extent_in
    INTEGER, INTENT(IN)                :: rows, columns

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(1:rows*columns) = RESHAPE(TRANSPOSE(&
         RESHAPE(extent_in(1:rows*columns), (/rows, columns/))), (/rows*columns/))
  END SUBROUTINE block_transpose_copy_z

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_2d1d_z(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_8), DIMENSION(rows,columns), INTENT(OUT) :: extent_out
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN)             :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d1d_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(extent_in, (/rows, columns/))
  END SUBROUTINE block_copy_2d1d_z

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_1d1d_z(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_8), DIMENSION(rows*columns), INTENT(OUT) :: extent_out
    COMPLEX(kind=real_8), DIMENSION(rows*columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_1d1d_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(:) = extent_in(:)
  END SUBROUTINE block_copy_1d1d_z

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_2d2d_z(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_8), DIMENSION(rows,columns), INTENT(OUT) :: extent_out
    COMPLEX(kind=real_8), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d2d_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(:,:) = extent_in(:,:)
  END SUBROUTINE block_copy_2d2d_z


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_2d1d_z(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_8), DIMENSION(columns,rows), INTENT(OUT) :: extent_out
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN)             :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d1d_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = TRANSPOSE(RESHAPE(extent_in, (/rows, columns/)))
  END SUBROUTINE block_transpose_copy_2d1d_z

! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_1d2d_z(extent_out, extent_in,&
       rows, columns)
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(OUT)            :: extent_out
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_8), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_copy_1d2d_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(extent_in, (/rows*columns/))
  END SUBROUTINE block_copy_1d2d_z


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_1d2d_z(extent_out, extent_in,&
       rows, columns)
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(OUT)            :: extent_out
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_8), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_1d2d_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(TRANSPOSE(extent_in), (/rows*columns/))
  END SUBROUTINE block_transpose_copy_1d2d_z


! **************************************************************************************************
!> \brief In-place block transpose.
!> \param[in,out] extent      Matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_inplace_z(extent, rows, columns)
    INTEGER, INTENT(IN)                      :: rows, columns
    COMPLEX(kind=real_8), DIMENSION(rows*columns), &
      INTENT(INOUT)                          :: extent
    COMPLEX(kind=real_8), DIMENSION(rows*columns)         :: extent_tr

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_inplace_z', &
      routineP = moduleN//':'//routineN

    INTEGER :: r, c
!   ---------------------------------------------------------------------------

    DO r = 1 , columns
      DO c = 1 , rows
       extent_tr(r + (c-1)*columns) = extent(c + (r-1)*rows)
      END DO
    END DO
    DO r = 1 , columns
      DO c = 1 , rows
       extent(r + (c-1)*columns) = extent_tr(r + (c-1)*columns)
      END DO
    END DO
  END SUBROUTINE block_transpose_inplace_z


! **************************************************************************************************
!> \brief Copy data from a double real array to a data area
!>
!> There are no checks done for correctness!
!> \param[in] dst        destination data area
!> \param[in] lb         lower bound for destination (and source if
!>                       not given explicity)
!> \param[in] data_size  number of elements to copy
!> \param[in] src        source data array
!> \param[in] source_lb  (optional) lower bound of source
! **************************************************************************************************
  SUBROUTINE dbcsr_data_set_az (dst, lb, data_size, src, source_lb)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: dst
    INTEGER, INTENT(IN)                      :: lb, data_size
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN)        :: src
    INTEGER, INTENT(IN), OPTIONAL            :: source_lb
    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_set_az', &
         routineP = moduleN//':'//routineN
    INTEGER                                  :: lb_s, ub, ub_s
!   ---------------------------------------------------------------------------
    IF (debug_mod) THEN
       IF(.NOT.ASSOCIATED (dst%d))&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",555,"Target data area must be setup.")
       IF(SIZE(src) .LT. data_size)&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",557,"Not enough source data.")
       IF(dst%d%data_type .NE. dbcsr_type_complex_8)&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",559,"Data type mismatch.")
    ENDIF
    ub = lb + data_size - 1
    IF (PRESENT (source_lb)) THEN
       lb_s = source_lb
       ub_s = source_lb + data_size-1
    ELSE
       lb_s = lb
       ub_s = ub
    ENDIF
    CALL memory_copy(dst%d%c_dp(lb:ub),src(lb_s:ub_s),data_size)
  END SUBROUTINE dbcsr_data_set_az

! **************************************************************************************************
!> \brief ...
!> \param m ...
!> \param blk ...
!> \param alpha ...
!> \param imin ...
!> \param imax ...
! **************************************************************************************************
  PURE SUBROUTINE block_2d_add_on_diag_z(m, blk, alpha, imin, imax)
    INTEGER, INTENT(IN)                      :: m
    COMPLEX(kind=real_8), INTENT(INOUT), DIMENSION(m,m)   :: blk
    COMPLEX(kind=real_8), INTENT(IN)                      :: alpha
    INTEGER, INTENT(IN), OPTIONAL            :: imin, imax

    CHARACTER(len=*), PARAMETER :: routineN = 'block_2d_add_on_diag_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i

!   ---------------------------------------------------------------------------

    IF(PRESENT(imin).AND.PRESENT(imax)) THEN
       DO i = MAX(1,imin),MIN(m,imax)
         blk(i,i) = blk(i,i) + alpha
       ENDDO
    ELSE
       DO i = 1,m
          blk(i,i) = blk(i,i) + alpha
       END DO
    ENDIF
  END SUBROUTINE block_2d_add_on_diag_z

! **************************************************************************************************
!> \brief ...
!> \param m ...
!> \param blk ...
!> \param alpha ...
!> \param imin ...
!> \param imax ...
! **************************************************************************************************
  PURE SUBROUTINE block_add_on_diag_z(m, blk, alpha, imin, imax)
    INTEGER, INTENT(IN)                      :: m
    COMPLEX(kind=real_8), INTENT(INOUT), DIMENSION(m*m)   :: blk
    COMPLEX(kind=real_8), INTENT(IN)                      :: alpha
    INTEGER, INTENT(IN), OPTIONAL            :: imin, imax

    CHARACTER(len=*), PARAMETER :: routineN = 'block_add_on_diag_z', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    IF(PRESENT(imin).AND.PRESENT(imax)) THEN
       CALL block_2d_add_on_diag_z(m, blk, alpha, imin, imax)
    ELSE
       CALL block_2d_add_on_diag_z(m, blk, alpha, 1, m)
    ENDIF
  END SUBROUTINE block_add_on_diag_z

! **************************************************************************************************
!> \brief ...
!> \param block_a ...
!> \param block_b ...
!> \param len ...
! **************************************************************************************************
  PURE SUBROUTINE block_add_z(block_a, block_b, len)
    INTEGER, INTENT(IN)                    :: len
    COMPLEX(kind=real_8), DIMENSION(len), INTENT(INOUT) :: block_a
    COMPLEX(kind=real_8), DIMENSION(len), INTENT(IN)    :: block_b
    block_a(1:len) = block_a(1:len) + block_b(1:len)
  END SUBROUTINE block_add_z
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Sets the diagonal of a square data block
!> \param[out] block_data     sets the diagonal of this data block
!> \param[in] diagonal        set diagonal of block_data to these values
!> \param[in] d               dimension of block
!> \par Off-diagonal values
!>      Other values are untouched.
! **************************************************************************************************
  PURE SUBROUTINE set_block2d_diagonal_s (block_data, diagonal, d)
    INTEGER, INTENT(IN)                    :: d
    REAL(kind=real_4), DIMENSION(d,d), INTENT(INOUT) :: block_data
    REAL(kind=real_4), DIMENSION(d), INTENT(IN)      :: diagonal

    INTEGER                                :: i

!   ---------------------------------------------------------------------------

    DO i = 1 , d
       block_data(i,i) = diagonal(i)
    END DO
  END SUBROUTINE set_block2d_diagonal_s


! **************************************************************************************************
!> \brief Gets the diagonal of a square data block
!> \param[in] block_data      gets the diagonal of this data block
!> \param[out] diagonal       values of the diagonal elements
!> \param[in] d               dimension of block
! **************************************************************************************************
  PURE SUBROUTINE get_block2d_diagonal_s (block_data, diagonal, d)
    INTEGER, INTENT(IN)                 :: d
    REAL(kind=real_4), DIMENSION(d,d), INTENT(IN) :: block_data
    REAL(kind=real_4), DIMENSION(d), INTENT(OUT)  :: diagonal

    INTEGER                             :: i

!   ---------------------------------------------------------------------------

    DO i = 1 , d
       diagonal(i) = block_data(i, i)
    END DO
  END SUBROUTINE get_block2d_diagonal_s


! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_rs ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_rs ...
!> \param src_cs ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param dst_offset ...
!> \param src_offset ...
!> \note see block_partial_copy_a 
! **************************************************************************************************
  SUBROUTINE block_partial_copy_s(dst, dst_rs, dst_cs, dst_tr,&
       src, src_rs, src_cs, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       dst_offset, src_offset)
    REAL(kind=real_4), DIMENSION(:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: dst_rs, dst_cs
    INTEGER, INTENT(IN)                      :: src_offset, dst_offset
    LOGICAL                                  :: dst_tr
    REAL(kind=real_4), DIMENSION(:), &
      INTENT(IN)                             :: src
    INTEGER, INTENT(IN)                      :: src_rs, src_cs
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop.
!  rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
                = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs) &
              = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
             = src(src_offset+src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs)&
             = src(src_offset + src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_s

! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_rs ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param dst_offset ...
!> \note see block_partial_copy_a 
! **************************************************************************************************
  SUBROUTINE block_partial_copy_1d2d_s(dst, dst_rs, dst_cs, dst_tr,&
       src, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       dst_offset)
    REAL(kind=real_4), DIMENSION(:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: dst_rs, dst_cs
    INTEGER, INTENT(IN)                      :: dst_offset
    LOGICAL                                  :: dst_tr
    REAL(kind=real_4), DIMENSION(:,:), &
      INTENT(IN)                             :: src
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_1d2d_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
                = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs) &
              = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs)&
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_1d2d_s
! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_rs ...
!> \param src_cs ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param src_offset ...
!> \note see block_partial_copy_a
! **************************************************************************************************
  SUBROUTINE block_partial_copy_2d1d_s(dst, dst_tr,&
       src, src_rs, src_cs, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       src_offset)
    REAL(kind=real_4), DIMENSION(:,:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: src_offset
    LOGICAL                                  :: dst_tr
    REAL(kind=real_4), DIMENSION(:), &
      INTENT(IN)                             :: src
    INTEGER, INTENT(IN)                      :: src_rs, src_cs
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_2d1d_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
                = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row) &
              = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
             = src(src_offset+src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row)&
             = src(src_offset + src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_2d1d_s
! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_tr ...
!> \param src ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \note see block_partial_copy_a
! **************************************************************************************************
  SUBROUTINE block_partial_copy_2d2d_s(dst, dst_tr,&
       src, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol)
    REAL(kind=real_4), DIMENSION(:,:), &
      INTENT(INOUT)                          :: dst
    LOGICAL                                  :: dst_tr
    REAL(kind=real_4), DIMENSION(:,:), &
      INTENT(IN)                             :: src
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_2d2d_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
                = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row) &
              = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row)&
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_2d2d_s


! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output data
!> \param[in] extent_in       input data
!> \param[in] n               number of elements to copy
!> \param[in] out_fe          first element of output
!> \param[in] in_fe           first element of input
! **************************************************************************************************
  PURE SUBROUTINE block_copy_s(extent_out, extent_in, n, out_fe, in_fe)
    INTEGER, INTENT(IN)                           :: n, out_fe, in_fe
    REAL(kind=real_4), DIMENSION(*), INTENT(OUT)  :: extent_out
    REAL(kind=real_4), DIMENSION(*), INTENT(IN)   :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(out_fe : out_fe+n-1) = extent_in(in_fe : in_fe+n-1)
  END SUBROUTINE block_copy_s


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_s(extent_out, extent_in,&
       rows, columns)
    REAL(kind=real_4), DIMENSION(:), INTENT(OUT) :: extent_out
    REAL(kind=real_4), DIMENSION(:), INTENT(IN)  :: extent_in
    INTEGER, INTENT(IN)                :: rows, columns

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(1:rows*columns) = RESHAPE(TRANSPOSE(&
         RESHAPE(extent_in(1:rows*columns), (/rows, columns/))), (/rows*columns/))
  END SUBROUTINE block_transpose_copy_s

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_2d1d_s(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_4), DIMENSION(rows,columns), INTENT(OUT) :: extent_out
    REAL(kind=real_4), DIMENSION(:), INTENT(IN)             :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d1d_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(extent_in, (/rows, columns/))
  END SUBROUTINE block_copy_2d1d_s

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_1d1d_s(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_4), DIMENSION(rows*columns), INTENT(OUT) :: extent_out
    REAL(kind=real_4), DIMENSION(rows*columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_1d1d_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(:) = extent_in(:)
  END SUBROUTINE block_copy_1d1d_s

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_2d2d_s(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_4), DIMENSION(rows,columns), INTENT(OUT) :: extent_out
    REAL(kind=real_4), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d2d_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(:,:) = extent_in(:,:)
  END SUBROUTINE block_copy_2d2d_s


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_2d1d_s(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_4), DIMENSION(columns,rows), INTENT(OUT) :: extent_out
    REAL(kind=real_4), DIMENSION(:), INTENT(IN)             :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d1d_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = TRANSPOSE(RESHAPE(extent_in, (/rows, columns/)))
  END SUBROUTINE block_transpose_copy_2d1d_s

! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_1d2d_s(extent_out, extent_in,&
       rows, columns)
    REAL(kind=real_4), DIMENSION(:), INTENT(OUT)            :: extent_out
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_4), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_copy_1d2d_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(extent_in, (/rows*columns/))
  END SUBROUTINE block_copy_1d2d_s


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_1d2d_s(extent_out, extent_in,&
       rows, columns)
    REAL(kind=real_4), DIMENSION(:), INTENT(OUT)            :: extent_out
    INTEGER, INTENT(IN)                           :: rows, columns
    REAL(kind=real_4), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_1d2d_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(TRANSPOSE(extent_in), (/rows*columns/))
  END SUBROUTINE block_transpose_copy_1d2d_s


! **************************************************************************************************
!> \brief In-place block transpose.
!> \param[in,out] extent      Matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_inplace_s(extent, rows, columns)
    INTEGER, INTENT(IN)                      :: rows, columns
    REAL(kind=real_4), DIMENSION(rows*columns), &
      INTENT(INOUT)                          :: extent
    REAL(kind=real_4), DIMENSION(rows*columns)         :: extent_tr

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_inplace_s', &
      routineP = moduleN//':'//routineN

    INTEGER :: r, c
!   ---------------------------------------------------------------------------

    DO r = 1 , columns
      DO c = 1 , rows
       extent_tr(r + (c-1)*columns) = extent(c + (r-1)*rows)
      END DO
    END DO
    DO r = 1 , columns
      DO c = 1 , rows
       extent(r + (c-1)*columns) = extent_tr(r + (c-1)*columns)
      END DO
    END DO
  END SUBROUTINE block_transpose_inplace_s


! **************************************************************************************************
!> \brief Copy data from a double real array to a data area
!>
!> There are no checks done for correctness!
!> \param[in] dst        destination data area
!> \param[in] lb         lower bound for destination (and source if
!>                       not given explicity)
!> \param[in] data_size  number of elements to copy
!> \param[in] src        source data array
!> \param[in] source_lb  (optional) lower bound of source
! **************************************************************************************************
  SUBROUTINE dbcsr_data_set_as (dst, lb, data_size, src, source_lb)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: dst
    INTEGER, INTENT(IN)                      :: lb, data_size
    REAL(kind=real_4), DIMENSION(:), INTENT(IN)        :: src
    INTEGER, INTENT(IN), OPTIONAL            :: source_lb
    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_set_as', &
         routineP = moduleN//':'//routineN
    INTEGER                                  :: lb_s, ub, ub_s
!   ---------------------------------------------------------------------------
    IF (debug_mod) THEN
       IF(.NOT.ASSOCIATED (dst%d))&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",555,"Target data area must be setup.")
       IF(SIZE(src) .LT. data_size)&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",557,"Not enough source data.")
       IF(dst%d%data_type .NE. dbcsr_type_real_4)&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",559,"Data type mismatch.")
    ENDIF
    ub = lb + data_size - 1
    IF (PRESENT (source_lb)) THEN
       lb_s = source_lb
       ub_s = source_lb + data_size-1
    ELSE
       lb_s = lb
       ub_s = ub
    ENDIF
    CALL memory_copy(dst%d%r_sp(lb:ub),src(lb_s:ub_s),data_size)
  END SUBROUTINE dbcsr_data_set_as

! **************************************************************************************************
!> \brief ...
!> \param m ...
!> \param blk ...
!> \param alpha ...
!> \param imin ...
!> \param imax ...
! **************************************************************************************************
  PURE SUBROUTINE block_2d_add_on_diag_s(m, blk, alpha, imin, imax)
    INTEGER, INTENT(IN)                      :: m
    REAL(kind=real_4), INTENT(INOUT), DIMENSION(m,m)   :: blk
    REAL(kind=real_4), INTENT(IN)                      :: alpha
    INTEGER, INTENT(IN), OPTIONAL            :: imin, imax

    CHARACTER(len=*), PARAMETER :: routineN = 'block_2d_add_on_diag_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i

!   ---------------------------------------------------------------------------

    IF(PRESENT(imin).AND.PRESENT(imax)) THEN
       DO i = MAX(1,imin),MIN(m,imax)
         blk(i,i) = blk(i,i) + alpha
       ENDDO
    ELSE
       DO i = 1,m
          blk(i,i) = blk(i,i) + alpha
       END DO
    ENDIF
  END SUBROUTINE block_2d_add_on_diag_s

! **************************************************************************************************
!> \brief ...
!> \param m ...
!> \param blk ...
!> \param alpha ...
!> \param imin ...
!> \param imax ...
! **************************************************************************************************
  PURE SUBROUTINE block_add_on_diag_s(m, blk, alpha, imin, imax)
    INTEGER, INTENT(IN)                      :: m
    REAL(kind=real_4), INTENT(INOUT), DIMENSION(m*m)   :: blk
    REAL(kind=real_4), INTENT(IN)                      :: alpha
    INTEGER, INTENT(IN), OPTIONAL            :: imin, imax

    CHARACTER(len=*), PARAMETER :: routineN = 'block_add_on_diag_s', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    IF(PRESENT(imin).AND.PRESENT(imax)) THEN
       CALL block_2d_add_on_diag_s(m, blk, alpha, imin, imax)
    ELSE
       CALL block_2d_add_on_diag_s(m, blk, alpha, 1, m)
    ENDIF
  END SUBROUTINE block_add_on_diag_s

! **************************************************************************************************
!> \brief ...
!> \param block_a ...
!> \param block_b ...
!> \param len ...
! **************************************************************************************************
  PURE SUBROUTINE block_add_s(block_a, block_b, len)
    INTEGER, INTENT(IN)                    :: len
    REAL(kind=real_4), DIMENSION(len), INTENT(INOUT) :: block_a
    REAL(kind=real_4), DIMENSION(len), INTENT(IN)    :: block_b
    block_a(1:len) = block_a(1:len) + block_b(1:len)
  END SUBROUTINE block_add_s
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Sets the diagonal of a square data block
!> \param[out] block_data     sets the diagonal of this data block
!> \param[in] diagonal        set diagonal of block_data to these values
!> \param[in] d               dimension of block
!> \par Off-diagonal values
!>      Other values are untouched.
! **************************************************************************************************
  PURE SUBROUTINE set_block2d_diagonal_c (block_data, diagonal, d)
    INTEGER, INTENT(IN)                    :: d
    COMPLEX(kind=real_4), DIMENSION(d,d), INTENT(INOUT) :: block_data
    COMPLEX(kind=real_4), DIMENSION(d), INTENT(IN)      :: diagonal

    INTEGER                                :: i

!   ---------------------------------------------------------------------------

    DO i = 1 , d
       block_data(i,i) = diagonal(i)
    END DO
  END SUBROUTINE set_block2d_diagonal_c


! **************************************************************************************************
!> \brief Gets the diagonal of a square data block
!> \param[in] block_data      gets the diagonal of this data block
!> \param[out] diagonal       values of the diagonal elements
!> \param[in] d               dimension of block
! **************************************************************************************************
  PURE SUBROUTINE get_block2d_diagonal_c (block_data, diagonal, d)
    INTEGER, INTENT(IN)                 :: d
    COMPLEX(kind=real_4), DIMENSION(d,d), INTENT(IN) :: block_data
    COMPLEX(kind=real_4), DIMENSION(d), INTENT(OUT)  :: diagonal

    INTEGER                             :: i

!   ---------------------------------------------------------------------------

    DO i = 1 , d
       diagonal(i) = block_data(i, i)
    END DO
  END SUBROUTINE get_block2d_diagonal_c


! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_rs ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_rs ...
!> \param src_cs ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param dst_offset ...
!> \param src_offset ...
!> \note see block_partial_copy_a 
! **************************************************************************************************
  SUBROUTINE block_partial_copy_c(dst, dst_rs, dst_cs, dst_tr,&
       src, src_rs, src_cs, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       dst_offset, src_offset)
    COMPLEX(kind=real_4), DIMENSION(:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: dst_rs, dst_cs
    INTEGER, INTENT(IN)                      :: src_offset, dst_offset
    LOGICAL                                  :: dst_tr
    COMPLEX(kind=real_4), DIMENSION(:), &
      INTENT(IN)                             :: src
    INTEGER, INTENT(IN)                      :: src_rs, src_cs
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop.
!  rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
                = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs) &
              = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
             = src(src_offset+src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs)&
             = src(src_offset + src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_c

! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_rs ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param dst_offset ...
!> \note see block_partial_copy_a 
! **************************************************************************************************
  SUBROUTINE block_partial_copy_1d2d_c(dst, dst_rs, dst_cs, dst_tr,&
       src, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       dst_offset)
    COMPLEX(kind=real_4), DIMENSION(:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: dst_rs, dst_cs
    INTEGER, INTENT(IN)                      :: dst_offset
    LOGICAL                                  :: dst_tr
    COMPLEX(kind=real_4), DIMENSION(:,:), &
      INTENT(IN)                             :: src
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_1d2d_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
                = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs) &
              = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_r_lb+row + (dst_c_lb+col-1)*dst_rs) &
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_offset+dst_c_lb+col + (dst_r_lb+row-1)*dst_cs)&
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_1d2d_c
! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_cs ...
!> \param dst_tr ...
!> \param src ...
!> \param src_rs ...
!> \param src_cs ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \param src_offset ...
!> \note see block_partial_copy_a
! **************************************************************************************************
  SUBROUTINE block_partial_copy_2d1d_c(dst, dst_tr,&
       src, src_rs, src_cs, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol,&
       src_offset)
    COMPLEX(kind=real_4), DIMENSION(:,:), &
      INTENT(INOUT)                          :: dst
    INTEGER, INTENT(IN)                      :: src_offset
    LOGICAL                                  :: dst_tr
    COMPLEX(kind=real_4), DIMENSION(:), &
      INTENT(IN)                             :: src
    INTEGER, INTENT(IN)                      :: src_rs, src_cs
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_2d1d_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
                = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row) &
              = src(src_offset+src_r_lb+row+(src_c_lb+col-1)*src_rs)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
             = src(src_offset+src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row)&
             = src(src_offset + src_c_lb+col+(src_r_lb+row-1)*src_cs)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_2d1d_c
! **************************************************************************************************
!> \brief Copies a block subset
!> \param dst ...
!> \param dst_tr ...
!> \param src ...
!> \param src_tr ...
!> \param dst_r_lb ...
!> \param dst_c_lb ...
!> \param src_r_lb ...
!> \param src_c_lb ...
!> \param nrow ...
!> \param ncol ...
!> \note see block_partial_copy_a
! **************************************************************************************************
  SUBROUTINE block_partial_copy_2d2d_c(dst, dst_tr,&
       src, src_tr,&
       dst_r_lb, dst_c_lb, src_r_lb, src_c_lb, nrow, ncol)
    COMPLEX(kind=real_4), DIMENSION(:,:), &
      INTENT(INOUT)                          :: dst
    LOGICAL                                  :: dst_tr
    COMPLEX(kind=real_4), DIMENSION(:,:), &
      INTENT(IN)                             :: src
    LOGICAL                                  :: src_tr
    INTEGER, INTENT(IN)                      :: dst_r_lb, dst_c_lb, src_r_lb, &
                                                src_c_lb, nrow, ncol

    CHARACTER(len=*), PARAMETER :: routineN = 'block_partial_copy_2d2d_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, row

!   ---------------------------------------------------------------------------
! Factors out the 4 combinations to remove branches from the inner loop. rs is the logical row size so it always remains the leading dimension.
    IF (.NOT. dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
                = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (dst_tr .AND. .NOT. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row) &
              = src(src_r_lb+row, src_c_lb+col)
         END DO
       END DO
    ELSEIF (.NOT. dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_r_lb+row, dst_c_lb+col) &
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ELSEIF (dst_tr .AND. src_tr) THEN
       DO col = 0,ncol-1
         DO row = 0,nrow-1
          dst(dst_c_lb+col, dst_r_lb+row)&
             = src(src_c_lb+col, src_r_lb+row)
         END DO
       END DO
    ENDIF
  END SUBROUTINE block_partial_copy_2d2d_c


! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output data
!> \param[in] extent_in       input data
!> \param[in] n               number of elements to copy
!> \param[in] out_fe          first element of output
!> \param[in] in_fe           first element of input
! **************************************************************************************************
  PURE SUBROUTINE block_copy_c(extent_out, extent_in, n, out_fe, in_fe)
    INTEGER, INTENT(IN)                           :: n, out_fe, in_fe
    COMPLEX(kind=real_4), DIMENSION(*), INTENT(OUT)  :: extent_out
    COMPLEX(kind=real_4), DIMENSION(*), INTENT(IN)   :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_d', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(out_fe : out_fe+n-1) = extent_in(in_fe : in_fe+n-1)
  END SUBROUTINE block_copy_c


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_c(extent_out, extent_in,&
       rows, columns)
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(OUT) :: extent_out
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN)  :: extent_in
    INTEGER, INTENT(IN)                :: rows, columns

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(1:rows*columns) = RESHAPE(TRANSPOSE(&
         RESHAPE(extent_in(1:rows*columns), (/rows, columns/))), (/rows*columns/))
  END SUBROUTINE block_transpose_copy_c

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_2d1d_c(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_4), DIMENSION(rows,columns), INTENT(OUT) :: extent_out
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN)             :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d1d_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(extent_in, (/rows, columns/))
  END SUBROUTINE block_copy_2d1d_c

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_1d1d_c(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_4), DIMENSION(rows*columns), INTENT(OUT) :: extent_out
    COMPLEX(kind=real_4), DIMENSION(rows*columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_1d1d_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(:) = extent_in(:)
  END SUBROUTINE block_copy_1d1d_c

! **************************************************************************************************
!> \brief Copy a block
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_2d2d_c(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_4), DIMENSION(rows,columns), INTENT(OUT) :: extent_out
    COMPLEX(kind=real_4), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d2d_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out(:,:) = extent_in(:,:)
  END SUBROUTINE block_copy_2d2d_c


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 2-d array
!> \param[in] extent_in       input matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_2d1d_c(extent_out, extent_in,&
       rows, columns)
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_4), DIMENSION(columns,rows), INTENT(OUT) :: extent_out
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN)             :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_2d1d_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = TRANSPOSE(RESHAPE(extent_in, (/rows, columns/)))
  END SUBROUTINE block_transpose_copy_2d1d_c

! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_copy_1d2d_c(extent_out, extent_in,&
       rows, columns)
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(OUT)            :: extent_out
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_4), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_copy_1d2d_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(extent_in, (/rows*columns/))
  END SUBROUTINE block_copy_1d2d_c


! **************************************************************************************************
!> \brief Copy and transpose block.
!> \param[out] extent_out     output matrix in the form of a 1-d array
!> \param[in] extent_in       input matrix in the form of a 2-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_copy_1d2d_c(extent_out, extent_in,&
       rows, columns)
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(OUT)            :: extent_out
    INTEGER, INTENT(IN)                           :: rows, columns
    COMPLEX(kind=real_4), DIMENSION(rows,columns), INTENT(IN)  :: extent_in

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_copy_1d2d_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    extent_out = RESHAPE(TRANSPOSE(extent_in), (/rows*columns/))
  END SUBROUTINE block_transpose_copy_1d2d_c


! **************************************************************************************************
!> \brief In-place block transpose.
!> \param[in,out] extent      Matrix in the form of a 1-d array
!> \param[in] rows input matrix size
!> \param[in] columns input matrix size
! **************************************************************************************************
  PURE SUBROUTINE block_transpose_inplace_c(extent, rows, columns)
    INTEGER, INTENT(IN)                      :: rows, columns
    COMPLEX(kind=real_4), DIMENSION(rows*columns), &
      INTENT(INOUT)                          :: extent
    COMPLEX(kind=real_4), DIMENSION(rows*columns)         :: extent_tr

    CHARACTER(len=*), PARAMETER :: routineN = 'block_transpose_inplace_c', &
      routineP = moduleN//':'//routineN

    INTEGER :: r, c
!   ---------------------------------------------------------------------------

    DO r = 1 , columns
      DO c = 1 , rows
       extent_tr(r + (c-1)*columns) = extent(c + (r-1)*rows)
      END DO
    END DO
    DO r = 1 , columns
      DO c = 1 , rows
       extent(r + (c-1)*columns) = extent_tr(r + (c-1)*columns)
      END DO
    END DO
  END SUBROUTINE block_transpose_inplace_c


! **************************************************************************************************
!> \brief Copy data from a double real array to a data area
!>
!> There are no checks done for correctness!
!> \param[in] dst        destination data area
!> \param[in] lb         lower bound for destination (and source if
!>                       not given explicity)
!> \param[in] data_size  number of elements to copy
!> \param[in] src        source data array
!> \param[in] source_lb  (optional) lower bound of source
! **************************************************************************************************
  SUBROUTINE dbcsr_data_set_ac (dst, lb, data_size, src, source_lb)
    TYPE(dbcsr_data_obj), INTENT(INOUT)      :: dst
    INTEGER, INTENT(IN)                      :: lb, data_size
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN)        :: src
    INTEGER, INTENT(IN), OPTIONAL            :: source_lb
    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_set_ac', &
         routineP = moduleN//':'//routineN
    INTEGER                                  :: lb_s, ub, ub_s
!   ---------------------------------------------------------------------------
    IF (debug_mod) THEN
       IF(.NOT.ASSOCIATED (dst%d))&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",555,"Target data area must be setup.")
       IF(SIZE(src) .LT. data_size)&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",557,"Not enough source data.")
       IF(dst%d%data_type .NE. dbcsr_type_complex_4)&
          CALL cp__b("dbcsr/block/dbcsr_block_operations.F",559,"Data type mismatch.")
    ENDIF
    ub = lb + data_size - 1
    IF (PRESENT (source_lb)) THEN
       lb_s = source_lb
       ub_s = source_lb + data_size-1
    ELSE
       lb_s = lb
       ub_s = ub
    ENDIF
    CALL memory_copy(dst%d%c_sp(lb:ub),src(lb_s:ub_s),data_size)
  END SUBROUTINE dbcsr_data_set_ac

! **************************************************************************************************
!> \brief ...
!> \param m ...
!> \param blk ...
!> \param alpha ...
!> \param imin ...
!> \param imax ...
! **************************************************************************************************
  PURE SUBROUTINE block_2d_add_on_diag_c(m, blk, alpha, imin, imax)
    INTEGER, INTENT(IN)                      :: m
    COMPLEX(kind=real_4), INTENT(INOUT), DIMENSION(m,m)   :: blk
    COMPLEX(kind=real_4), INTENT(IN)                      :: alpha
    INTEGER, INTENT(IN), OPTIONAL            :: imin, imax

    CHARACTER(len=*), PARAMETER :: routineN = 'block_2d_add_on_diag_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: i

!   ---------------------------------------------------------------------------

    IF(PRESENT(imin).AND.PRESENT(imax)) THEN
       DO i = MAX(1,imin),MIN(m,imax)
         blk(i,i) = blk(i,i) + alpha
       ENDDO
    ELSE
       DO i = 1,m
          blk(i,i) = blk(i,i) + alpha
       END DO
    ENDIF
  END SUBROUTINE block_2d_add_on_diag_c

! **************************************************************************************************
!> \brief ...
!> \param m ...
!> \param blk ...
!> \param alpha ...
!> \param imin ...
!> \param imax ...
! **************************************************************************************************
  PURE SUBROUTINE block_add_on_diag_c(m, blk, alpha, imin, imax)
    INTEGER, INTENT(IN)                      :: m
    COMPLEX(kind=real_4), INTENT(INOUT), DIMENSION(m*m)   :: blk
    COMPLEX(kind=real_4), INTENT(IN)                      :: alpha
    INTEGER, INTENT(IN), OPTIONAL            :: imin, imax

    CHARACTER(len=*), PARAMETER :: routineN = 'block_add_on_diag_c', &
      routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

    IF(PRESENT(imin).AND.PRESENT(imax)) THEN
       CALL block_2d_add_on_diag_c(m, blk, alpha, imin, imax)
    ELSE
       CALL block_2d_add_on_diag_c(m, blk, alpha, 1, m)
    ENDIF
  END SUBROUTINE block_add_on_diag_c

! **************************************************************************************************
!> \brief ...
!> \param block_a ...
!> \param block_b ...
!> \param len ...
! **************************************************************************************************
  PURE SUBROUTINE block_add_c(block_a, block_b, len)
    INTEGER, INTENT(IN)                    :: len
    COMPLEX(kind=real_4), DIMENSION(len), INTENT(INOUT) :: block_a
    COMPLEX(kind=real_4), DIMENSION(len), INTENT(IN)    :: block_b
    block_a(1:len) = block_a(1:len) + block_b(1:len)
  END SUBROUTINE block_add_c

END MODULE dbcsr_block_operations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR data operations
!> \author  Urban Borstnik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2010-02-18 Moved from dbcsr_util
! **************************************************************************************************
MODULE dbcsr_data_operations

   USE dbcsr_block_operations,          ONLY: dbcsr_block_transpose,&
                                              dbcsr_data_copy,&
                                              dbcsr_data_set
   USE dbcsr_data_methods,              ONLY: dbcsr_data_get_size,&
                                              dbcsr_data_get_size_referenced,&
                                              dbcsr_data_hold,&
                                              dbcsr_data_release,&
                                              dbcsr_data_set_size_referenced,&
                                              dbcsr_get_data
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_failure_level,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_types,                     ONLY: dbcsr_data_obj,&
                                              dbcsr_obj,&
                                              dbcsr_type_complex_4,&
                                              dbcsr_type_complex_8,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8
   USE dbcsr_util,                      ONLY: sgn
   USE kinds,                           ONLY: real_4,&
                                              real_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_data_operations'

   PUBLIC :: dbcsr_data_copyall, dbcsr_data_convert, &
             dbcsr_copy_sort_data, &
             dbcsr_sort_data
   PUBLIC :: dbcsr_switch_data_area

CONTAINS

! **************************************************************************************************
!> \brief Sets the data area of a matrix
!> \param[in,out] matrix     matrix for which to set the data area
!> \param[in] data_area      data area to set
!> \param[out] previous_data_area  (optional) previous data area
! **************************************************************************************************
   SUBROUTINE dbcsr_switch_data_area(matrix, data_area, previous_data_area)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: data_area
      TYPE(dbcsr_data_obj), INTENT(OUT), OPTIONAL        :: previous_data_area

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_switch_data_area', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (PRESENT(previous_data_area)) THEN
         previous_data_area = matrix%m%data_area
      ELSE
         CALL dbcsr_data_release(matrix%m%data_area)
      ENDIF
      matrix%m%data_area = data_area
      CALL dbcsr_data_hold(matrix%m%data_area)
      CALL timestop(handle)
   END SUBROUTINE dbcsr_switch_data_area

! **************************************************************************************************
!> \brief Copies a data area, deep by default.
!> \param[in,out] target_area     target data area
!> \param[in]     source_area     source data area
!> \param[in]     shallow         shallow copy (default is deep)
! **************************************************************************************************
   SUBROUTINE dbcsr_data_copyall(target_area, source_area, shallow)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: target_area
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: source_area
      LOGICAL, INTENT(IN), OPTIONAL                      :: shallow

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_copyall', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, i, n
      LOGICAL                                            :: shallow_copy

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      IF (.NOT. ASSOCIATED(source_area%d)) &
         CALL cp__b("dbcsr/block/dbcsr_data_operations.F",107,"Attempt to copy unassigned data")
      IF (source_area%d%refcount .LE. 0) &
         CALL cp__w("dbcsr/block/dbcsr_data_operations.F",109,"Attempt to copy unheld data")
      shallow_copy = .FALSE.
      IF (PRESENT(shallow)) shallow_copy = shallow
      IF (shallow_copy) THEN
         target_area = source_area
         CALL dbcsr_data_hold(target_area)
      ELSE
         IF (.NOT. ASSOCIATED(target_area%d)) &
            CALL cp__b("dbcsr/block/dbcsr_data_operations.F",117,"Target area does not exist.")
         CALL dbcsr_data_set_size_referenced(target_area, &
                                             dbcsr_data_get_size_referenced(source_area))
         n = dbcsr_data_get_size_referenced(source_area)
         SELECT CASE (target_area%d%data_type)
         CASE (dbcsr_type_real_4)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(I) SHARED(target_area,source_area,n)
            DO i = 1, n
               target_area%d%r_sp(i) = source_area%d%r_sp(i)
            ENDDO
         CASE (dbcsr_type_real_8)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(I) SHARED(target_area,source_area,n)
            DO i = 1, n
               target_area%d%r_dp(i) = source_area%d%r_dp(i)
            ENDDO
         CASE (dbcsr_type_complex_4)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(I) SHARED(target_area,source_area,n)
            DO i = 1, n
               target_area%d%c_sp(i) = source_area%d%c_sp(i)
            ENDDO
         CASE (dbcsr_type_complex_8)
!$OMP           PARALLEL DO DEFAULT(NONE) PRIVATE(I) SHARED(target_area,source_area,n)
            DO i = 1, n
               target_area%d%c_dp(i) = source_area%d%c_dp(i)
            ENDDO
         CASE default
            CALL cp__b("dbcsr/block/dbcsr_data_operations.F",143,"Invalid data type")
         END SELECT
      ENDIF
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_data_copyall

! **************************************************************************************************
!> \brief Copies a data area, converting data type
!> \param[inout] target_area     target data area
!> \param[in]    source_area     source data area
!> \param[in]    drop_real       (optional) drops real part of complex
!>                               numbers instead of the imaginary part; default
!>                               is false
!> \param[in]    multiply_by_i   (optional) converts real to complex by placing
!>                               into imaginary instead of real part
! **************************************************************************************************
   SUBROUTINE dbcsr_data_convert(target_area, source_area, drop_real, &
                                 multiply_by_i)
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: target_area
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: source_area
      LOGICAL, INTENT(IN), OPTIONAL                      :: drop_real, multiply_by_i

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_data_convert', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=real_4), DIMENSION(:), POINTER        :: s_data_c, t_data_c
      COMPLEX(KIND=real_8), DIMENSION(:), POINTER        :: s_data_z, t_data_z
      INTEGER                                            :: n, ns, nt
      LOGICAL                                            :: keep_real, noimult
      REAL(KIND=real_4), DIMENSION(:), POINTER           :: s_data_r, t_data_r
      REAL(KIND=real_8), DIMENSION(:), POINTER           :: s_data_d, t_data_d

!   ---------------------------------------------------------------------------

      IF (.NOT. ASSOCIATED(source_area%d)) &
         CALL cp__w("dbcsr/block/dbcsr_data_operations.F",178,"Attempt to copy unassigned data")
      IF (source_area%d%refcount .LE. 0) &
         CALL cp__w("dbcsr/block/dbcsr_data_operations.F",180,"Attempt to copy unheld data")
      IF (.NOT. ASSOCIATED(source_area%d)) THEN
         RETURN
      ENDIF
      keep_real = .TRUE.
      IF (PRESENT(drop_real)) keep_real = .NOT. drop_real
      noimult = .TRUE.
      IF (PRESENT(multiply_by_i)) noimult = .NOT. multiply_by_i
      ns = dbcsr_data_get_size_referenced(source_area)
      nt = dbcsr_data_get_size_referenced(target_area)
      n = MIN(ns, nt)
      IF (n .GT. 0) THEN
         SELECT CASE (source_area%d%data_type)
         CASE (dbcsr_type_real_8)
            CALL dbcsr_get_data(source_area, s_data_d)
            SELECT CASE (target_area%d%data_type)
            CASE (dbcsr_type_real_8)
               CALL dbcsr_get_data(target_area, t_data_d)
               t_data_d(1:n) = s_data_d(1:n)
            CASE (dbcsr_type_real_4)
               CALL dbcsr_get_data(target_area, t_data_r)
               t_data_r(1:n) = REAL(s_data_d(1:n), KIND=real_4)
            CASE (dbcsr_type_complex_8)
               CALL dbcsr_get_data(target_area, t_data_z)
               IF (noimult) THEN
                  t_data_z(1:n) = CMPLX(s_data_d(1:n), KIND=real_8)
               ELSE
                  t_data_z(1:n) = CMPLX(0.0, s_data_d(1:n), KIND=real_8)
               ENDIF
            CASE (dbcsr_type_complex_4)
               CALL dbcsr_get_data(target_area, t_data_c)
               IF (noimult) THEN
                  t_data_c(1:n) = CMPLX(s_data_d(1:n), KIND=real_4)
               ELSE
                  t_data_c(1:n) = CMPLX(0.0, s_data_d(1:n), KIND=real_4)
               ENDIF
            CASE default
               CALL cp__b("dbcsr/block/dbcsr_data_operations.F",217,"Invalid data type")
            END SELECT
         CASE (dbcsr_type_real_4)
            CALL dbcsr_get_data(source_area, s_data_r)
            SELECT CASE (target_area%d%data_type)
            CASE (dbcsr_type_real_8)
               CALL dbcsr_get_data(target_area, t_data_d)
               t_data_d(1:n) = REAL(s_data_r(1:n), KIND=real_8)
            CASE (dbcsr_type_real_4)
               CALL dbcsr_get_data(target_area, t_data_r)
               t_data_r(1:n) = s_data_r(1:n)
            CASE (dbcsr_type_complex_8)
               CALL dbcsr_get_data(target_area, t_data_z)
               IF (noimult) THEN
                  t_data_z(1:n) = CMPLX(s_data_r(1:n), KIND=real_8)
               ELSE
                  t_data_z(1:n) = CMPLX(0.0, s_data_r(1:n), KIND=real_8)
               ENDIF
            CASE (dbcsr_type_complex_4)
               CALL dbcsr_get_data(target_area, t_data_c)
               IF (noimult) THEN
                  t_data_c(1:n) = CMPLX(s_data_r(1:n), KIND=real_4)
               ELSE
                  t_data_c(1:n) = CMPLX(0.0, s_data_r(1:n), KIND=real_4)
               ENDIF
            CASE default
               CALL cp__b("dbcsr/block/dbcsr_data_operations.F",243,"Invalid data type")
            END SELECT
         CASE (dbcsr_type_complex_8)
            CALL dbcsr_get_data(source_area, s_data_z)
            SELECT CASE (target_area%d%data_type)
            CASE (dbcsr_type_real_8)
               CALL dbcsr_get_data(target_area, t_data_d)
               IF (keep_real) THEN
                  t_data_d(1:n) = REAL(s_data_z(1:n), KIND=real_8)
               ELSE
                  t_data_d(1:n) = AIMAG(s_data_z(1:n))
               ENDIF
            CASE (dbcsr_type_real_4)
               CALL dbcsr_get_data(target_area, t_data_r)
               IF (keep_real) THEN
                  t_data_r(1:n) = REAL(s_data_z(1:n), KIND=real_4)
               ELSE
                  t_data_r(1:n) = REAL(AIMAG(s_data_z(1:n)), KIND=real_4)
               ENDIF
            CASE (dbcsr_type_complex_8)
               CALL dbcsr_get_data(target_area, t_data_z)
               t_data_z(1:n) = s_data_z(1:n)
            CASE (dbcsr_type_complex_4)
               CALL dbcsr_get_data(target_area, t_data_c)
               t_data_c(1:n) = CMPLX(s_data_z(1:n), KIND=real_4)
            CASE default
               CALL cp__b("dbcsr/block/dbcsr_data_operations.F",269,"Invalid data type")
            END SELECT
         CASE (dbcsr_type_complex_4)
            CALL dbcsr_get_data(source_area, s_data_c)
            SELECT CASE (target_area%d%data_type)
            CASE (dbcsr_type_real_8)
               CALL dbcsr_get_data(target_area, t_data_d)
               IF (keep_real) THEN
                  t_data_d(1:n) = REAL(s_data_c(1:n), KIND=real_8)
               ELSE
                  t_data_d(1:n) = REAL(AIMAG(s_data_c(1:n)), KIND=real_8)
               ENDIF
            CASE (dbcsr_type_real_4)
               CALL dbcsr_get_data(target_area, t_data_r)
               IF (keep_real) THEN
                  t_data_r(1:n) = REAL(s_data_c(1:n), KIND=real_4)
               ELSE
                  t_data_r(1:n) = AIMAG(s_data_c(1:n))
               ENDIF
            CASE (dbcsr_type_complex_8)
               CALL dbcsr_get_data(target_area, t_data_z)
               t_data_z(1:n) = CMPLX(s_data_c(1:n), KIND=real_8)
            CASE (dbcsr_type_complex_4)
               CALL dbcsr_get_data(target_area, t_data_c)
               t_data_c(1:n) = s_data_c(1:n)
            CASE default
               CALL cp__b("dbcsr/block/dbcsr_data_operations.F",295,"Invalid data type")
            END SELECT
         CASE default
            CALL cp__b("dbcsr/block/dbcsr_data_operations.F",298,"Invalid data type")
         END SELECT
      ENDIF
   END SUBROUTINE dbcsr_data_convert

! **************************************************************************************************
!> \brief Sorts the data in a matrix so that the data blocks follow
!>        sequentially and does various transposing options.
!>
!> As opposed to dbcsr_sort_data, this routine calculates block sizes
!> \param[out] blk_p          re-arranged block pointers reflecting the new
!>                            data order
!> \param[in] old_blk_p       current block pointers
!> \param[in] row_p index
!> \param[in] col_i index
!> \param[in] rbs sizes of the blocked rows
!> \param[in] cbs sizes of the blocked columns
!> \param[out] dst            sorted data
!> \param[in] src             existing unordered data
!> \param[in] mark_transposed           mark data as transposed by negating the
!>                                      blk_p index entries
!> \param[in] transpose_blocks          transpose data blocks
! **************************************************************************************************
   SUBROUTINE dbcsr_copy_sort_data(blk_p, old_blk_p, row_p, col_i, rbs, cbs, &
                                   dst, src, mark_transposed, transpose_blocks)
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: blk_p
      INTEGER, DIMENSION(:), INTENT(IN)                  :: old_blk_p, row_p, col_i, rbs, cbs
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: dst
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: src
      LOGICAL, INTENT(IN), OPTIONAL                      :: mark_transposed, transpose_blocks

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_copy_sort_data', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col_size, nblks, nrows, nze, &
                                                            nze_prev, row, row_size
      LOGICAL                                            :: mark, trb

!   ---------------------------------------------------------------------------
! Analyze parameters

      mark = .FALSE.
      IF (PRESENT(mark_transposed)) mark = mark_transposed
      trb = .FALSE.
      IF (PRESENT(transpose_blocks)) trb = transpose_blocks
      !
      nblks = SIZE(old_blk_p)
      nrows = SIZE(row_p)-1
      CALL dbcsr_assert(SIZE(blk_p), 'GE', nblks, dbcsr_failure_level, &
                        dbcsr_wrong_args_error, routineN, 'Destination blk_p too small.', 347)
      IF (nblks .GE. 1) &
         blk_p(1) = SGN(1, old_blk_p(1), mark)
      nze_prev = 0
      DO row = 1, nrows
         row_size = rbs(row)
         DO blk = row_p(row)+1, row_p(row+1)
            IF (old_blk_p(blk) .NE. 0) THEN
               col_size = cbs(col_i(blk))
               nze = row_size*col_size
               IF (blk .GT. 1) THEN
                  blk_p(blk) = SGN(ABS(blk_p(blk-1))+nze_prev, old_blk_p(blk), &
                                   mark)
               ENDIF
               CALL dbcsr_assert( &
                  ABS(blk_p(blk))+nze-1, 'LE', dbcsr_data_get_size(dst), &
                  dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                  'Destination data space is too small.', 364)
               IF (.NOT. trb) THEN
                  CALL dbcsr_data_copy(dst=dst, dst_lb=(/ABS(blk_p(blk))/), &
                                       dst_sizes=(/nze/), &
                                       src=src, src_lb=(/ABS(old_blk_p(blk))/), &
                                       src_sizes=(/nze/))
                  !CALL dbcsr_data_set (dst, ABS(blk_p(blk)), nze,&
                  !     src, source_lb=ABS(old_blk_p(blk)))
               ELSE
                  CALL dbcsr_block_transpose(dst, src, &
                                             col_size, row_size, &
                                             lb=ABS(blk_p(blk)), source_lb=ABS(old_blk_p(blk)))
               ENDIF
               nze_prev = nze
            ENDIF ! blk exists
         ENDDO ! blk
      ENDDO ! row
   END SUBROUTINE dbcsr_copy_sort_data

! **************************************************************************************************
!> \brief Sorts the data in a matrix so that the data blocks follow
!>        sequentially.
!> \param[out] blk_p          re-arranged block pointers reflecting the new
!>                            data order
!> \param[in] old_blk_p       current block pointers
!> \param[in] sizes           sizes of the data blocks
!> \param[out] dsts           sorted data
!> \param[in] src             existing unordered data
!> \param[in] srcs            (optional) multiple source areas
!> \param old_blk_d ...
! **************************************************************************************************
   SUBROUTINE dbcsr_sort_data(blk_p, old_blk_p, sizes, dsts, src, &
                              srcs, old_blk_d)
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: blk_p
      INTEGER, DIMENSION(:), INTENT(IN)                  :: old_blk_p, sizes
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: dsts
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: src
      TYPE(dbcsr_data_obj), DIMENSION(:), INTENT(IN), &
         OPTIONAL                                        :: srcs
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: old_blk_d

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_sort_data', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, nblks
      LOGICAL                                            :: multidata

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      multidata = PRESENT(srcs) .AND. PRESENT(old_blk_d)
      nblks = SIZE(old_blk_p)
      IF (nblks .GT. 0) THEN
!$OMP        BARRIER
!$OMP        MASTER
         blk_p(1) = SIGN(1, old_blk_p(1))
         DO i = 2, nblks
            blk_p(i) = SIGN(ABS(blk_p(i-1))+sizes(i-1), old_blk_p(i))
         ENDDO
         CALL dbcsr_data_set_size_referenced(dsts, &
                                             ABS(blk_p(nblks))+sizes(nblks)-1)
!$OMP        END MASTER
!$OMP        BARRIER
!$OMP        DO
         DO i = 1, nblks
            IF (old_blk_p(i) .NE. 0) THEN
               IF (.NOT. multidata) THEN
                  CALL dbcsr_data_set(dsts, &
                                      ABS(blk_p(i)), sizes(i), &
                                      src, source_lb=ABS(old_blk_p(i)))
                  !dst(ABS(blk_p(i)):ABS(blk_p(i))+sizes(i)-1) =&
                  !     src(ABS(old_blk_p(i)):ABS(old_blk_p(i))+sizes(i)-1)
               ELSE
                  CALL dbcsr_data_set(dsts, &
                                      ABS(blk_p(i)), sizes(i), &
                                      srcs(old_blk_d(i)), source_lb=ABS(old_blk_p(i)))
                  !dst(ABS(blk_p(i)):ABS(blk_p(i))+sizes(i)-1) =&
                  !     srcs(old_blk_d(i))%d&
                  !     %r_dp(ABS(old_blk_p(i)):ABS(old_blk_p(i))+sizes(i)-1)
               ENDIF
            ENDIF
         ENDDO
!$OMP        END DO NOWAIT
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_sort_data

END MODULE dbcsr_data_operations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR iterator operations
!> \author  Urban Borstnik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!  - 2010-02-18 Moved from dbcsr_operations and dbcsr_methods
! **************************************************************************************************
MODULE dbcsr_iterator_operations
   USE array_types,                     ONLY: array_data,&
                                              array_exists
   USE dbcsr_data_methods,              ONLY: dbcsr_data_hold,&
                                              dbcsr_data_release,&
                                              dbcsr_data_set_pointer,&
                                              dbcsr_get_data
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_has_threads,&
                                              dbcsr_distribution_num_threads,&
                                              dbcsr_distribution_thread_dist
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_methods,                   ONLY: dbcsr_distribution
   USE dbcsr_ptr_util,                  ONLY: pointer_rank_remap2
   USE dbcsr_toollib,                   ONLY: swap
   USE dbcsr_types,                     ONLY: dbcsr_data_obj,&
                                              dbcsr_distribution_obj,&
                                              dbcsr_iterator,&
                                              dbcsr_obj,&
                                              dbcsr_scalar_type
   USE kinds,                           ONLY: real_4,&
                                              real_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads, omp_in_parallel

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_iterator_operations'

   INTEGER, PRIVATE, POINTER, SAVE      :: common_int_pointer

   PUBLIC :: dbcsr_iterator_start, dbcsr_iterator_stop

   PUBLIC :: dbcsr_iterator_blocks_left, dbcsr_iterator_next_block

   INTERFACE dbcsr_iterator_next_block
      MODULE PROCEDURE iterator_next_block_index, &
         iterator_next_area_block
      MODULE PROCEDURE iterator_next_2d_block_d, &
         iterator_next_2d_block_s, &
         iterator_next_2d_block_c, &
         iterator_next_2d_block_z, &
         iterator_next_1d_block_d, &
         iterator_next_1d_block_s, &
         iterator_next_1d_block_c, &
         iterator_next_1d_block_z
   END INTERFACE

   LOGICAL, PARAMETER :: careful_mod = .FALSE.

   INTEGER, PARAMETER, PRIVATE :: rpslot_owner = 1
   INTEGER, PARAMETER, PRIVATE :: rpslot_addblks = 2
   INTEGER, PARAMETER, PRIVATE :: rpslot_addoffset = 3
   INTEGER, PARAMETER, PRIVATE :: rpslot_oldblks = 4
   INTEGER, PARAMETER, PRIVATE :: rpslot_oldoffset = 5
   INTEGER, PARAMETER, PRIVATE :: rpslot_totaloffset = 6
   INTEGER, PARAMETER, PRIVATE :: rpnslots = 6

   LOGICAL, PARAMETER, PRIVATE :: detailed_timing = .FALSE.

   TYPE block_parameters
      LOGICAL :: tr
      INTEGER :: logical_rows, logical_cols
      INTEGER :: offset, nze
   END TYPE block_parameters

   TYPE dgemm_join
      INTEGER :: p_a, p_b, p_c
      INTEGER :: last_k, last_n
      TYPE(dbcsr_scalar_type) :: alpha, beta
   END TYPE dgemm_join

CONTAINS

! **************************************************************************************************
! Iterator functions
! **************************************************************************************************

! **************************************************************************************************
!> \brief Sets up an iterator
!> \param[out] iterator   the iterator
!> \param[in] matrix      DBCSR matrix
!> \param[in] shared      (optional) The matrix is shared between several
!>                        iterators. Default is .TRUE.
!> \param[in] dynamic     (optional) Threads are given blocks regardless of
!>                        the thread distribution; default is .FALSE.
!> \param[in] dynamic_byrows  (optional) Threads are given blocks regardless
!>                            of the thread distribution, but still grouped
!>                            by rows; default is .FALSE.
!> \param[in] contiguous_pointers  (optional) Whether returned pointers need
!>                                 to be contiguous; default is FALSE.
!> \param[in] read_only            (optional) User promises not to change
!>                                 returned data; default is FALSE
!>
!> \par Contiguous pointers
!> Contiguous pointers may incur reallocation penalties but enable quick
!> passing of arrays to routines with unspecified interfaces (i.e., direct
!> calls to BLACS or MPI).
!> \par Threading
!>      The TYPE(dbcsr_iterator) variable should be thread-private.
!> \par
!>      The iterator has several modes of operation when used with
!>      OpenMP. Two options can be set to influence the behavior.
!> \par Threading: shared vs. non-shared
!>      The "shared" flag specifies that several threads will be
!>      iterating through the same matrix.
!>      - Sharing is the default when called from an active parallel
!>        region. In the shared mode no two threads will receive the
!>        same block; i.e., the work is split among the threads.
!>      - If each (or one) thread needs to iterator through all blocks
!>        then shared should be set to .FALSE.. (E.g., when called
!>        from an enclosing MASTER region or when each thread has its
!>        own matrix.)
!>      - It is safe to use an iterator in non-shared mode with only
!>        one thread.  No thread synchronization constructs are used
!>        in this case)
!> \par Threading in shared mode
!>      When in shared mode there are three possibilities to select
!>      how the blocks are distributed to the threads.
!>      <DL>
!>      <DT>Thread distribution</DT>
!>      <DD>The default is to use the thread distribution. The thread
!>          distribution statically maps rows to threads and should be
!>          used whenever retaining a consistent mapping among
!>          subsequent iterations is important.</DD>
!>      <DT>Dynamic scheduling</DT>
!>      <DD>If the dynamic flag is .TRUE., then blocks are given to
!>          threads dynamically. By default the assignment is grouped
!>          by rows (to minimize synchronization); however, if the
!>          dynamic_byrows flag is .FALSE. then every block is
!>          assigned dynamically.</DD></DL>
! **************************************************************************************************
   SUBROUTINE dbcsr_iterator_start(iterator, matrix, shared, dynamic, &
                                   dynamic_byrows, contiguous_pointers, read_only)
      TYPE(dbcsr_iterator), INTENT(OUT)                  :: iterator
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: shared, dynamic, dynamic_byrows, &
                                                            contiguous_pointers, read_only

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_iterator_start', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle
      TYPE(dbcsr_distribution_obj)                       :: dist

!   ---------------------------------------------------------------------------

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(dynamic))==-1) EXIT ;  END DO ; ENDIF ! only used with OMP

      CALL timeset(routineN, error_handle)
      iterator%shared = .TRUE.
!$    iterator%shared = omp_in_parallel()
      IF (PRESENT(shared)) iterator%shared = shared
      iterator%dynamic = .TRUE.
!$    iterator%dynamic = .FALSE.
!$    IF (PRESENT(dynamic)) iterator%dynamic = dynamic
      IF (PRESENT(dynamic_byrows)) THEN
         iterator%dynamic_byrows = dynamic_byrows
         IF (iterator%dynamic_byrows) iterator%dynamic = .TRUE.
      ELSE
         iterator%dynamic_byrows = iterator%dynamic
!$       iterator%dynamic_byrows = iterator%dynamic
      ENDIF
!$    IF (.NOT. iterator%shared) THEN
!$       iterator%dynamic = .FALSE.
!$    ENDIF
      dist = dbcsr_distribution(matrix)
!$    IF (.NOT. dbcsr_distribution_has_threads(dist)) &
!$       CALL cp__w("dbcsr/block/dbcsr_iterator_operations.F",188,"Thread distribution should be defined for OpenMP.")
      CALL dbcsr_assert(.NOT. iterator%dynamic, "IMP", &
                        dbcsr_distribution_has_threads(dist), &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "Thread distribution must be defined for non-dynamic iterator.", &
                        193)
!$    IF (omp_in_parallel()) &
!$       CALL dbcsr_assert(omp_get_num_threads(), "EQ", &
!$                         dbcsr_distribution_num_threads(dist), &
!$                         dbcsr_fatal_level, dbcsr_internal_error, routineN, &
!$                         "Number of threads has changed!", 198)
      !Synchronize the positions
      NULLIFY (iterator%common_pos)
      IF (iterator%dynamic) THEN
         ! All threads point into the master thread's data space
         ! (temporarily using the common_int_pointer variable). This is
         ! not the nicest OpenMP way of doing this but it is also not
         ! explicity forbidden.
         !
!$OMP        BARRIER
!$OMP        MASTER
         ALLOCATE (iterator%common_pos)
         common_int_pointer => iterator%common_pos
         common_int_pointer = 0
!$OMP        FLUSH (common_int_pointer)
!$OMP        END MASTER
!$OMP        BARRIER
         IF (.NOT. ASSOCIATED(iterator%common_pos)) THEN
            iterator%common_pos => common_int_pointer
         ENDIF
!$OMP        BARRIER
      ENDIF
      !
      IF (PRESENT(contiguous_pointers)) THEN
         iterator%contiguous_pointers = contiguous_pointers
      ELSE
         iterator%contiguous_pointers = .TRUE.
      ENDIF
      IF (PRESENT(read_only)) THEN
         iterator%read_only = read_only
      ELSE
         iterator%read_only = .FALSE.
      ENDIF
      iterator%row = 0
      iterator%pos = 0
      iterator%rbs => array_data(matrix%m%row_blk_size)
      iterator%cbs => array_data(matrix%m%col_blk_size)
      iterator%roff => array_data(matrix%m%row_blk_offset)
      iterator%coff => array_data(matrix%m%col_blk_offset)

      iterator%local_indexing = matrix%m%local_indexing
      !CALL dbcsr_assert (iterator%local_indexing, "IMP",&
      !     iterator%dynamic,&
      !     dbcsr_fatal_level, dbcsr_caller_error, routineN,&
      !     "Locally-indexed matrices can only have a dynamic iterator.",&
      !     243)
      CALL dbcsr_assert(iterator%local_indexing, "IMP", &
                        array_exists(matrix%m%local_rows), &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Local rows mapping array should exist when local indexing is used.", &
                        248)
      CALL dbcsr_assert(iterator%local_indexing, "IMP", &
                        array_exists(matrix%m%global_rows), &
                        dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                        "Global rows mapping array should exist when local indexing is used.", &
                        253)
      iterator%global_rows => array_data(matrix%m%global_rows)
      iterator%local_rows => array_data(matrix%m%local_rows)

      iterator%transpose = .FALSE. !matrix%m%transpose
      iterator%nblks = matrix%m%nblks
      IF (iterator%transpose) THEN
         iterator%nblkrows_total = matrix%m%nblkcols_total
      ELSE
         iterator%nblkrows_total = matrix%m%nblkrows_total
      ENDIF

      iterator%row_p => matrix%m%row_p
      iterator%col_i => matrix%m%col_i
      iterator%blk_p => matrix%m%blk_p
!$OMP     CRITICAL (crit_data)
      iterator%data_area = matrix%m%data_area
      CALL dbcsr_data_hold(iterator%data_area)
!$OMP     END CRITICAL (crit_data)
      iterator%row_size = 0
      IF (.NOT. iterator%dynamic) THEN
         iterator%tdist => array_data(dbcsr_distribution_thread_dist(dist))
      ELSE
         NULLIFY (iterator%tdist)
      ENDIF
!$    IF (iterator%dynamic) THEN
!$OMP           SINGLE
!$       IF (iterator%dynamic_byrows) THEN
!$          iterator%common_pos = omp_get_num_threads()
!$       ENDIF
!$OMP           END SINGLE
!$       CALL dbcsr_iterator_seek(iterator, omp_get_thread_num()+1)
!$    ELSE
         CALL dbcsr_iterator_seek(iterator, 1)
!$    ENDIF
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_iterator_start

! **************************************************************************************************
!> \brief Stops up an iterator
!> \param[out] iterator   the iterator
! **************************************************************************************************
   SUBROUTINE dbcsr_iterator_stop(iterator)
      TYPE(dbcsr_iterator), INTENT(INOUT)                :: iterator

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_iterator_stop', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      iterator%row = 0
      iterator%pos = 0

      NULLIFY (iterator%tdist)
!$OMP     CRITICAL (crit_data)
      CALL dbcsr_data_release(iterator%data_area)
!$OMP     END CRITICAL (crit_data)
      IF (iterator%dynamic) THEN
!$OMP        BARRIER
!$OMP        MASTER
         common_int_pointer => iterator%common_pos
         DEALLOCATE (common_int_pointer)
!$OMP        FLUSH (common_int_pointer)
!$OMP        END MASTER
         NULLIFY (iterator%common_pos)
!$OMP        BARRIER
      ENDIF
      IF (iterator%local_indexing) THEN
         NULLIFY (iterator%local_rows)
         NULLIFY (iterator%global_rows)
      ENDIF
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_iterator_stop

! **************************************************************************************************
!> \brief Finds the first valid block, inclusive from the current position.
!>
!> If there is no valid block, pos is set to 0
!> \param[in,out] pos        input: current position; output: next valid
!>                           position or 0
!> \param[in] maxpos         maximal allowed position
!> \param[in] blk_p          block pointers, used to check validity
! **************************************************************************************************
   PURE SUBROUTINE find_first_valid_block(pos, maxpos, blk_p)
      INTEGER, INTENT(INOUT)                             :: pos
      INTEGER, INTENT(IN)                                :: maxpos
      INTEGER, DIMENSION(:), INTENT(IN)                  :: blk_p

      CHARACTER(len=*), PARAMETER :: routineN = 'find_first_valid_block', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------
!IF (pos .LT. 1) pos = 1

      DO WHILE (pos .LE. maxpos)
         IF (blk_p(pos) .EQ. 0) THEN
            !WRITE(*,*)routineN//" Skipping deleted block."
            pos = pos+1
         ELSE
            EXIT
         ENDIF
      ENDDO
      IF (pos .GT. maxpos) pos = 0
   END SUBROUTINE find_first_valid_block

! **************************************************************************************************
!> \brief Finds the row to which the current block belongs
!>
!> If there is no valid block, pos is set to 0
!> \param[in] pos            current position
!> \param[in,out] row        input: current row; output: the row corresponding
!>                           to the position
!> \param[in] maxrows        maxmimum row
!> \param[in] row_p          row pointers
! **************************************************************************************************
   PURE SUBROUTINE find_proper_row(pos, row, maxrows, row_p)
      INTEGER, INTENT(IN)                                :: pos
      INTEGER, INTENT(INOUT)                             :: row
      INTEGER, INTENT(IN)                                :: maxrows
      INTEGER, DIMENSION(:), INTENT(IN)                  :: row_p

      CHARACTER(len=*), PARAMETER :: routineN = 'find_proper_row', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      IF (pos .GT. 0) THEN
         IF (row .LT. 1) THEN
            row = 1
         ELSEIF (row .GT. maxrows) THEN
            row = maxrows
         ENDIF
         DO WHILE (row_p(row+1) .LT. pos)
            row = row+1
            IF (row .GT. maxrows) THEN
               row = 0
               EXIT
            ENDIF
         ENDDO
      ELSE
         row = 0
      ENDIF
   END SUBROUTINE find_proper_row

! **************************************************************************************************
!> \brief Finds the next proper position accounting for threads
!>
!> First time: pos and row are set to 0.
!> If there is no valid block, pos is set to 0
!> \param[inout] pos         current position and updated position
!> \param[in,out] row        input: current row; output: the row corresponding
!>                           to the next proper position
!> \param[in] maxpos         maximum allowable position
!> \param[in] maxrows        maxmimum row
!> \param[in] blk_p          block pointercs
!> \param[in] row_p          row pointers
!> \param[in] tdist          (optional) thread distribution
!> \param[in] tid            my thread number
!> \param local2global ...
! **************************************************************************************************
   PURE SUBROUTINE find_proper_position(pos, row, maxpos, maxrows, &
                                        blk_p, row_p, tdist, tid, local2global)
      INTEGER, INTENT(INOUT)                             :: pos, row
      INTEGER, INTENT(IN)                                :: maxpos, maxrows
      INTEGER, DIMENSION(:), INTENT(IN)                  :: blk_p, row_p
      INTEGER, DIMENSION(1:maxrows), INTENT(IN), &
         OPTIONAL                                        :: tdist
      INTEGER, INTENT(IN), OPTIONAL                      :: tid
      INTEGER, DIMENSION(1:*), INTENT(IN), OPTIONAL      :: local2global

      CHARACTER(len=*), PARAMETER :: routineN = 'find_proper_position', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: local, row_inrange, row_ok

!   ---------------------------------------------------------------------------

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tdist))==-1) EXIT ;  END DO ; ENDIF ! only used with OMP
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(tid))==-1) EXIT ;  END DO ; ENDIF ! only used with OMP

      local = PRESENT(local2global)
      IF (maxpos .GE. 1) THEN
         !IF (pos.EQ.0) pos = 1
         CALL find_first_valid_block(pos, maxpos, blk_p)
         CALL find_proper_row(pos, row, maxrows, row_p)
         row_inrange = row .NE. 0 .AND. row .LE. maxrows
         row_ok = row_inrange
!$       IF (present(tdist) .AND. PRESENT(tid) .AND. row_inrange) THEN
!$          IF (.NOT. local) THEN
!$             row_ok = tdist(row) .EQ. tid
!$          ELSE
!$             row_ok = tdist(local2global(row)) .EQ. tid
!$          ENDIF
!$       ENDIF
         DO WHILE (row_inrange .AND. .NOT. row_ok)
            row = row+1
            pos = row_p(row)+1
            IF (row .GT. maxrows) THEN
               row = 0
               EXIT
            ENDIF
            CALL find_first_valid_block(pos, maxpos, blk_p)
            CALL find_proper_row(pos, row, maxrows, row_p)
            row_inrange = row .NE. 0
            row_ok = row_inrange
!$          IF (present(tdist) .AND. PRESENT(tid) .AND. row_inrange) THEN
!$             IF (.NOT. local) THEN
!$                row_ok = tdist(row) .EQ. tid
!$             ELSE
!$                row_ok = tdist(local2global(row)) .EQ. tid
!$             ENDIF
!$          ENDIF
         ENDDO
         IF (row .EQ. 0) pos = 0
      ELSE
         pos = 0
         row = 0
      ENDIF
   END SUBROUTINE find_proper_position

! **************************************************************************************************
!> \brief Advances to the next block
!> \param[in,out] iterator     the iterator
! **************************************************************************************************
   SUBROUTINE iterator_advance(iterator)
      TYPE(dbcsr_iterator), INTENT(INOUT)                :: iterator

      INTEGER                                            :: ithread, my_old_row, next_row, p
      LOGICAL                                            :: advance, jumped_row

!   ---------------------------------------------------------------------------

      ithread = 0
!$    ithread = omp_get_thread_num()
      IF (iterator%dynamic .AND. iterator%shared) THEN
         IF (iterator%dynamic_byrows) THEN
            ! In this case common_row holds the last assigned row.
            !
            ! The idea is to advance in this thread's row. If it gets bumped
            ! into the next row, then we have to find the correct one.
            advance = .TRUE.
            DO WHILE (advance)
               iterator%pos = iterator%pos+1
               my_old_row = iterator%row
               CALL find_proper_position_caller(iterator)
               jumped_row = iterator%row .GT. my_old_row
               advance = jumped_row
               IF (jumped_row) THEN
!$OMP                 CRITICAL (crit_common_pos)
                  ! Set the common_pos to the next available row.
                  next_row = MAX(iterator%common_pos+1, iterator%row)
                  next_row = next_row-iterator%common_pos
                  iterator%common_pos = iterator%common_pos+next_row
                  iterator%row = iterator%common_pos
!$OMP                 END CRITICAL (crit_common_pos)
                  IF (iterator%row .GT. iterator%nblkrows_total) THEN
                     iterator%pos = 0
                     iterator%row = 0
                     advance = .FALSE.
                  ELSE
                     ! To be incremented in the next loop.
                     IF (.NOT. iterator%local_indexing) THEN
                        iterator%pos = iterator%row_p(iterator%row)
                     ELSE
                        iterator%pos = iterator%row_p(iterator%global_rows(iterator%row))
                     ENDIF
                  ENDIF
               ENDIF
            ENDDO
         ELSE
            ! In this case common_pos holds the last-assigned block.
            !
!$OMP           CRITICAL (crit_common_pos)
            iterator%common_pos = iterator%common_pos+1
            iterator%pos = iterator%common_pos
            CALL find_proper_position_caller(iterator)
            p = iterator%pos
            iterator%common_pos = MAX(iterator%common_pos, p)
!$OMP           END CRITICAL (crit_common_pos)
         ENDIF
      ELSEIF (iterator%shared) THEN
         iterator%pos = iterator%pos+1
         ithread = 0
!$       ithread = OMP_GET_THREAD_NUM()
         CALL find_proper_position_caller(iterator, use_ithread=ithread)
      ELSE
         iterator%pos = iterator%pos+1
         CALL find_proper_position_caller(iterator)
      ENDIF
   END SUBROUTINE iterator_advance

! **************************************************************************************************
!> \brief ...
!> \param iterator ...
!> \param use_ithread ...
! **************************************************************************************************
   SUBROUTINE find_proper_position_caller(iterator, use_ithread)
      TYPE(dbcsr_iterator), INTENT(INOUT)                :: iterator
      INTEGER, INTENT(in), OPTIONAL                      :: use_ithread

      INTEGER                                            :: post_pos, post_row, pre_pos, pre_row, r

      pre_row = iterator%row
      pre_pos = iterator%pos
      IF (.NOT. iterator%local_indexing) THEN
         IF (PRESENT(use_ithread)) THEN
            CALL find_proper_position(iterator%pos, &
                                      iterator%row, iterator%nblks, iterator%nblkrows_total, &
                                      iterator%blk_p, iterator%row_p, &
                                      tdist=iterator%tdist, tid=use_ithread)
         ELSE
            CALL find_proper_position(iterator%pos, &
                                      iterator%row, iterator%nblks, iterator%nblkrows_total, &
                                      iterator%blk_p, iterator%row_p)
         ENDIF
      ELSE
         IF (iterator%row .GT. 0) THEN
            r = iterator%global_rows(iterator%row)
         ELSE
            r = 0
         ENDIF
         IF (PRESENT(use_ithread)) THEN
            CALL find_proper_position(iterator%pos, &
                                      r, iterator%nblks, SIZE(iterator%local_rows), &
                                      iterator%blk_p, iterator%row_p, &
                                      local2global=iterator%local_rows, &
                                      tdist=iterator%tdist, tid=use_ithread)
         ELSE
            CALL find_proper_position(iterator%pos, &
                                      r, iterator%nblks, SIZE(iterator%local_rows), &
                                      iterator%blk_p, iterator%row_p, &
                                      local2global=iterator%local_rows)
         ENDIF
         IF (r .GT. 0) THEN
            iterator%row = iterator%local_rows(r)
         ELSE
            iterator%row = 0
         ENDIF
      ENDIF
      post_row = iterator%row
      post_pos = iterator%pos
   END SUBROUTINE find_proper_position_caller

! **************************************************************************************************
!> \brief Updates the row info stored in the iterator
!> \param[in,out] iterator     the iterator
!> \note Added to handle the complexity introduced with the transpose
! **************************************************************************************************
   PURE SUBROUTINE update_row_info(iterator)
      TYPE(dbcsr_iterator), INTENT(INOUT)                :: iterator

      IF (iterator%row .GT. 0) THEN
         IF (iterator%transpose) THEN
            iterator%row_size = iterator%cbs(iterator%row)
            iterator%row_offset = iterator%coff(iterator%row)
         ELSE
            iterator%row_size = iterator%rbs(iterator%row)
            iterator%row_offset = iterator%roff(iterator%row)
         ENDIF
      ENDIF
   END SUBROUTINE update_row_info

! **************************************************************************************************
!> \brief Places the iterator to the desired row.
!> \param[in,out] iterator     the iterator
!> \param[in] row              seek to this row
! **************************************************************************************************
   SUBROUTINE dbcsr_iterator_seek(iterator, row)
      TYPE(dbcsr_iterator), INTENT(INOUT)                :: iterator
      INTEGER, INTENT(in)                                :: row

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_iterator_seek', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      IF (iterator%nblks .GT. 0 .AND. row .LE. iterator%nblkrows_total) THEN
         iterator%row = row
         ! This line is replaced because iterator_advance increments the block
         ! number
         !iterator%pos = iterator%row_p(row)+1
         iterator%pos = iterator%row_p(row) ! +1-1
         CALL iterator_advance(iterator)
         CALL update_row_info(iterator)
      ELSE
         iterator%row = 0
         iterator%pos = 0
      ENDIF
      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_iterator_seek

! **************************************************************************************************
!> \brief Gets the index information of the next block, no data.
!> \param[in,out] iterator   the iterator
!> \param[out] row           row of the data block
!> \param[out] column        column of the data block
!> \param[out] blk           block number
!> \param[out] transposed    (optional) whether block is transposed
!> \param[out] blk_p         (optional) index into block data array
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
!> \param[out] row_offset    (optional) logical row offset of block
!> \param[out] col_offset    (optional) logical column offset of block
! **************************************************************************************************
   SUBROUTINE iterator_next_block_index(iterator, row, column, blk, &
                                        transposed, blk_p, row_size, col_size, row_offset, col_offset)
      TYPE(dbcsr_iterator), INTENT(INOUT)                :: iterator
      INTEGER, INTENT(OUT)                               :: row, column, blk
      LOGICAL, INTENT(OUT), OPTIONAL                     :: transposed
      INTEGER, INTENT(OUT), OPTIONAL                     :: blk_p, row_size, col_size, row_offset, &
                                                            col_offset

      CHARACTER(len=*), PARAMETER :: routineN = 'iterator_next_block_index', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bp, error_handle

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, error_handle)
      IF (iterator%pos .LE. iterator%nblks &
          .AND. iterator%pos .NE. 0) THEN
         row = iterator%row
         column = iterator%col_i(iterator%pos)
         IF (iterator%transpose) CALL swap(row, column)
         blk = iterator%pos
         IF (PRESENT(row_size)) row_size = iterator%row_size
         IF (PRESENT(col_size)) col_size = iterator%cbs(column)
         IF (PRESENT(row_offset)) row_offset = iterator%row_offset
         IF (PRESENT(col_offset)) col_offset = iterator%coff(column)
         IF (PRESENT(blk_p) .OR. PRESENT(transposed)) THEN
            bp = iterator%blk_p(iterator%pos)
            IF (PRESENT(blk_p)) blk_p = bp
            IF (PRESENT(transposed)) transposed = bp .LT. 0
         ENDIF
         CALL iterator_advance(iterator)
         CALL update_row_info(iterator)
      ELSE
         row = 0
         column = 0
      ENDIF
      IF (careful_mod) CALL timestop(error_handle)
   END SUBROUTINE iterator_next_block_index

! **************************************************************************************************
!> \brief Gets the next data block encapsulated in an object.
!> \param[in,out] iterator   the iterator
!> \param[out] row           row of the data block
!> \param[out] column        column of the data block
!> \param[out] block         encapsulated data
!> \param[out] transposed    whether the block data is transposed
!> \param[out] block_number  (optional) block number
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
!> \param[out] row_offset    (optional) logical row offset of block
!> \param[out] col_offset    (optional) logical column offset of block
! **************************************************************************************************
   SUBROUTINE iterator_next_area_block(iterator, row, column, block, &
                                       transposed, block_number, row_size, col_size, row_offset, col_offset)
      TYPE(dbcsr_iterator), INTENT(INOUT)                :: iterator
      INTEGER, INTENT(OUT)                               :: row, column
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: block
      LOGICAL, INTENT(OUT)                               :: transposed
      INTEGER, INTENT(OUT), OPTIONAL                     :: block_number, row_size, col_size, &
                                                            row_offset, col_offset

      CHARACTER(len=*), PARAMETER :: routineN = 'iterator_next_area_block', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk_p, block_col_size, block_row_size, &
                                                            bp, csize, error_handle, nze, rsize

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, error_handle)

      IF (iterator%pos .LE. iterator%nblks &
          .AND. iterator%pos .NE. 0) THEN
         row = iterator%row
         column = iterator%col_i(iterator%pos)
         IF (iterator%transpose) CALL swap(row, column)
         blk_p = iterator%blk_p(iterator%pos)
         transposed = blk_p .LT. 0
         bp = ABS(blk_p)
         rsize = iterator%row_size
         csize = iterator%cbs(column)
         block_row_size = rsize
         block_col_size = csize
         nze = rsize*csize
         IF (PRESENT(row_size)) row_size = rsize
         IF (PRESENT(col_size)) col_size = csize
         IF (PRESENT(row_offset)) row_offset = iterator%row_offset
         IF (PRESENT(col_offset)) col_offset = iterator%coff(column)
         ! Redirect the encapsulated pointer to the correct pointer here.
         IF (transposed) CALL swap(rsize, csize)
         CALL dbcsr_data_set_pointer(block, rsize, csize, iterator%data_area, &
                                     source_lb=bp)
         IF (PRESENT(block_number)) block_number = iterator%pos
         ! Move to the next non-deleted position.
         CALL iterator_advance(iterator)
         CALL update_row_info(iterator)
      ELSE
         row = 0
         column = 0
         IF (PRESENT(block_number)) block_number = 0
      ENDIF

      IF (careful_mod) CALL timestop(error_handle)

   END SUBROUTINE iterator_next_area_block

! **************************************************************************************************
!> \brief Returns whether there any blocks left in the iterator.
!> \param[in] iterator       the iterator
!> \retval blocks_left ...
!> \param[in]
! **************************************************************************************************
   PURE FUNCTION dbcsr_iterator_blocks_left(iterator) RESULT(blocks_left)
      TYPE(dbcsr_iterator), INTENT(IN)                   :: iterator
      LOGICAL                                            :: blocks_left

      blocks_left = iterator%pos .NE. 0
   END FUNCTION dbcsr_iterator_blocks_left

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Gets the next data block, single/double precision real/complex
!> \param[in,out] iterator   the iterator
!> \param[out] row           row of the data block
!> \param[out] column        column of the data block
!> \param[out] block         pointer to the data block
!> \param[out] transposed    whether the block data is transposed
!> \param[out] block_number  (optional) block number
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE iterator_next_1d_block_d (iterator, row, column, block,&
       transposed, block_number, row_size, col_size, row_offset, col_offset)
    TYPE(dbcsr_iterator), INTENT(INOUT)      :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    REAL(kind=real_8), DIMENSION(:), POINTER :: block
    LOGICAL, INTENT(OUT)                     :: transposed
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size,&
         row_offset, col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'iterator_next_1d_block_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk_p, bp, csize, nze, rsize

!   ---------------------------------------------------------------------------
! If we're pointing to a valid block, return that block.


    IF (iterator%pos .LE. iterator%nblks&
         .AND. iterator%pos .NE. 0) THEN
       row = iterator%row
       column = iterator%col_i(iterator%pos)
       IF (iterator%transpose) CALL swap (row, column)
       blk_p = iterator%blk_p(iterator%pos)
       transposed = blk_p .LT. 0
       bp = ABS (blk_p)
       rsize = iterator%row_size
       csize = iterator%cbs(column)
       nze = rsize * csize
       IF (PRESENT (row_size)) row_size = rsize
       IF (PRESENT (col_size)) col_size = csize
       IF (PRESENT (row_offset)) row_offset = iterator%row_offset
       IF (PRESENT (col_offset)) col_offset = iterator%coff(column)
       CALL dbcsr_get_data (iterator%data_area, block,&
            lb=bp, ub=bp+nze-1)
       IF (PRESENT (block_number)) block_number = iterator%pos
       ! Move to the next non-deleted position.
       CALL iterator_advance (iterator)
       CALL update_row_info (iterator)
    ELSE
       row = 0
       column = 0
       NULLIFY (block)
       IF (PRESENT (block_number)) block_number = 0
    ENDIF
  END SUBROUTINE iterator_next_1d_block_d



! **************************************************************************************************
!> \brief Gets the next data block, single/double precision real/complex
!> \param[in,out] iterator   the iterator
!> \param[out] row           row of the data block
!> \param[out] column        column of the data block
!> \param[out] block         pointer to the data block
!> \param[out] transposed    whether the block data is transposed
!> \param[out] block_number  (optional) block number
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE iterator_next_2d_block_d (iterator, row, column,&
       block, transposed,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(dbcsr_iterator), INTENT(INOUT)      :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    REAL(kind=real_8), DIMENSION(:, :), &
      POINTER                                :: block
    LOGICAL, INTENT(OUT)                     :: transposed
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size, row_offset, col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'iterator_next_2d_block_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk_p, bp, csize, nze, rsize, &
                                                block_row_size, block_col_size
    REAL(kind=real_8), DIMENSION(:), POINTER           :: lin_blk_p
    INTEGER                                  :: error_handle

!   ---------------------------------------------------------------------------
! If we're pointing to a valid block, return that block.

    IF (careful_mod) CALL timeset (routineN, error_handle)
    IF (iterator%pos .LE. iterator%nblks&
         .AND. iterator%pos .NE. 0) THEN
       row = iterator%row
       column = iterator%col_i(iterator%pos)
       IF (iterator%transpose) CALL swap (row, column)
       blk_p = iterator%blk_p(iterator%pos)
       transposed = blk_p .LT. 0
       bp = ABS (blk_p)
       rsize = iterator%row_size
       csize = iterator%cbs(column)
       block_row_size = rsize
       block_col_size = csize
       IF (PRESENT (row_size)) row_size = rsize
       IF (PRESENT (col_size)) col_size = csize
       IF (PRESENT (row_offset)) row_offset = iterator%row_offset
       IF (PRESENT (col_offset)) col_offset = iterator%coff(column)
       nze = rsize * csize
       IF (transposed) CALL swap (rsize, csize)
       CALL dbcsr_get_data (iterator%data_area, lin_blk_p,&
            lb=bp, ub=bp+nze-1)
       CALL pointer_rank_remap2 (block, rsize, csize, lin_blk_p)
       IF (PRESENT (block_number)) block_number = iterator%pos
       ! Move to the next non-deleted position.
       CALL iterator_advance (iterator)
       CALL update_row_info (iterator)
    ELSE
       row = 0
       column = 0
       NULLIFY (block)
       IF (PRESENT (block_number)) block_number = 0
    ENDIF
    IF (careful_mod) CALL timestop (error_handle)
  END SUBROUTINE iterator_next_2d_block_d
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Gets the next data block, single/double precision real/complex
!> \param[in,out] iterator   the iterator
!> \param[out] row           row of the data block
!> \param[out] column        column of the data block
!> \param[out] block         pointer to the data block
!> \param[out] transposed    whether the block data is transposed
!> \param[out] block_number  (optional) block number
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE iterator_next_1d_block_z (iterator, row, column, block,&
       transposed, block_number, row_size, col_size, row_offset, col_offset)
    TYPE(dbcsr_iterator), INTENT(INOUT)      :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    COMPLEX(kind=real_8), DIMENSION(:), POINTER :: block
    LOGICAL, INTENT(OUT)                     :: transposed
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size,&
         row_offset, col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'iterator_next_1d_block_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk_p, bp, csize, nze, rsize

!   ---------------------------------------------------------------------------
! If we're pointing to a valid block, return that block.


    IF (iterator%pos .LE. iterator%nblks&
         .AND. iterator%pos .NE. 0) THEN
       row = iterator%row
       column = iterator%col_i(iterator%pos)
       IF (iterator%transpose) CALL swap (row, column)
       blk_p = iterator%blk_p(iterator%pos)
       transposed = blk_p .LT. 0
       bp = ABS (blk_p)
       rsize = iterator%row_size
       csize = iterator%cbs(column)
       nze = rsize * csize
       IF (PRESENT (row_size)) row_size = rsize
       IF (PRESENT (col_size)) col_size = csize
       IF (PRESENT (row_offset)) row_offset = iterator%row_offset
       IF (PRESENT (col_offset)) col_offset = iterator%coff(column)
       CALL dbcsr_get_data (iterator%data_area, block,&
            lb=bp, ub=bp+nze-1)
       IF (PRESENT (block_number)) block_number = iterator%pos
       ! Move to the next non-deleted position.
       CALL iterator_advance (iterator)
       CALL update_row_info (iterator)
    ELSE
       row = 0
       column = 0
       NULLIFY (block)
       IF (PRESENT (block_number)) block_number = 0
    ENDIF
  END SUBROUTINE iterator_next_1d_block_z



! **************************************************************************************************
!> \brief Gets the next data block, single/double precision real/complex
!> \param[in,out] iterator   the iterator
!> \param[out] row           row of the data block
!> \param[out] column        column of the data block
!> \param[out] block         pointer to the data block
!> \param[out] transposed    whether the block data is transposed
!> \param[out] block_number  (optional) block number
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE iterator_next_2d_block_z (iterator, row, column,&
       block, transposed,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(dbcsr_iterator), INTENT(INOUT)      :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    COMPLEX(kind=real_8), DIMENSION(:, :), &
      POINTER                                :: block
    LOGICAL, INTENT(OUT)                     :: transposed
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size, row_offset, col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'iterator_next_2d_block_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk_p, bp, csize, nze, rsize, &
                                                block_row_size, block_col_size
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: lin_blk_p
    INTEGER                                  :: error_handle

!   ---------------------------------------------------------------------------
! If we're pointing to a valid block, return that block.

    IF (careful_mod) CALL timeset (routineN, error_handle)
    IF (iterator%pos .LE. iterator%nblks&
         .AND. iterator%pos .NE. 0) THEN
       row = iterator%row
       column = iterator%col_i(iterator%pos)
       IF (iterator%transpose) CALL swap (row, column)
       blk_p = iterator%blk_p(iterator%pos)
       transposed = blk_p .LT. 0
       bp = ABS (blk_p)
       rsize = iterator%row_size
       csize = iterator%cbs(column)
       block_row_size = rsize
       block_col_size = csize
       IF (PRESENT (row_size)) row_size = rsize
       IF (PRESENT (col_size)) col_size = csize
       IF (PRESENT (row_offset)) row_offset = iterator%row_offset
       IF (PRESENT (col_offset)) col_offset = iterator%coff(column)
       nze = rsize * csize
       IF (transposed) CALL swap (rsize, csize)
       CALL dbcsr_get_data (iterator%data_area, lin_blk_p,&
            lb=bp, ub=bp+nze-1)
       CALL pointer_rank_remap2 (block, rsize, csize, lin_blk_p)
       IF (PRESENT (block_number)) block_number = iterator%pos
       ! Move to the next non-deleted position.
       CALL iterator_advance (iterator)
       CALL update_row_info (iterator)
    ELSE
       row = 0
       column = 0
       NULLIFY (block)
       IF (PRESENT (block_number)) block_number = 0
    ENDIF
    IF (careful_mod) CALL timestop (error_handle)
  END SUBROUTINE iterator_next_2d_block_z
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Gets the next data block, single/double precision real/complex
!> \param[in,out] iterator   the iterator
!> \param[out] row           row of the data block
!> \param[out] column        column of the data block
!> \param[out] block         pointer to the data block
!> \param[out] transposed    whether the block data is transposed
!> \param[out] block_number  (optional) block number
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE iterator_next_1d_block_s (iterator, row, column, block,&
       transposed, block_number, row_size, col_size, row_offset, col_offset)
    TYPE(dbcsr_iterator), INTENT(INOUT)      :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    REAL(kind=real_4), DIMENSION(:), POINTER :: block
    LOGICAL, INTENT(OUT)                     :: transposed
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size,&
         row_offset, col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'iterator_next_1d_block_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk_p, bp, csize, nze, rsize

!   ---------------------------------------------------------------------------
! If we're pointing to a valid block, return that block.


    IF (iterator%pos .LE. iterator%nblks&
         .AND. iterator%pos .NE. 0) THEN
       row = iterator%row
       column = iterator%col_i(iterator%pos)
       IF (iterator%transpose) CALL swap (row, column)
       blk_p = iterator%blk_p(iterator%pos)
       transposed = blk_p .LT. 0
       bp = ABS (blk_p)
       rsize = iterator%row_size
       csize = iterator%cbs(column)
       nze = rsize * csize
       IF (PRESENT (row_size)) row_size = rsize
       IF (PRESENT (col_size)) col_size = csize
       IF (PRESENT (row_offset)) row_offset = iterator%row_offset
       IF (PRESENT (col_offset)) col_offset = iterator%coff(column)
       CALL dbcsr_get_data (iterator%data_area, block,&
            lb=bp, ub=bp+nze-1)
       IF (PRESENT (block_number)) block_number = iterator%pos
       ! Move to the next non-deleted position.
       CALL iterator_advance (iterator)
       CALL update_row_info (iterator)
    ELSE
       row = 0
       column = 0
       NULLIFY (block)
       IF (PRESENT (block_number)) block_number = 0
    ENDIF
  END SUBROUTINE iterator_next_1d_block_s



! **************************************************************************************************
!> \brief Gets the next data block, single/double precision real/complex
!> \param[in,out] iterator   the iterator
!> \param[out] row           row of the data block
!> \param[out] column        column of the data block
!> \param[out] block         pointer to the data block
!> \param[out] transposed    whether the block data is transposed
!> \param[out] block_number  (optional) block number
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE iterator_next_2d_block_s (iterator, row, column,&
       block, transposed,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(dbcsr_iterator), INTENT(INOUT)      :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    REAL(kind=real_4), DIMENSION(:, :), &
      POINTER                                :: block
    LOGICAL, INTENT(OUT)                     :: transposed
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size, row_offset, col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'iterator_next_2d_block_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk_p, bp, csize, nze, rsize, &
                                                block_row_size, block_col_size
    REAL(kind=real_4), DIMENSION(:), POINTER           :: lin_blk_p
    INTEGER                                  :: error_handle

!   ---------------------------------------------------------------------------
! If we're pointing to a valid block, return that block.

    IF (careful_mod) CALL timeset (routineN, error_handle)
    IF (iterator%pos .LE. iterator%nblks&
         .AND. iterator%pos .NE. 0) THEN
       row = iterator%row
       column = iterator%col_i(iterator%pos)
       IF (iterator%transpose) CALL swap (row, column)
       blk_p = iterator%blk_p(iterator%pos)
       transposed = blk_p .LT. 0
       bp = ABS (blk_p)
       rsize = iterator%row_size
       csize = iterator%cbs(column)
       block_row_size = rsize
       block_col_size = csize
       IF (PRESENT (row_size)) row_size = rsize
       IF (PRESENT (col_size)) col_size = csize
       IF (PRESENT (row_offset)) row_offset = iterator%row_offset
       IF (PRESENT (col_offset)) col_offset = iterator%coff(column)
       nze = rsize * csize
       IF (transposed) CALL swap (rsize, csize)
       CALL dbcsr_get_data (iterator%data_area, lin_blk_p,&
            lb=bp, ub=bp+nze-1)
       CALL pointer_rank_remap2 (block, rsize, csize, lin_blk_p)
       IF (PRESENT (block_number)) block_number = iterator%pos
       ! Move to the next non-deleted position.
       CALL iterator_advance (iterator)
       CALL update_row_info (iterator)
    ELSE
       row = 0
       column = 0
       NULLIFY (block)
       IF (PRESENT (block_number)) block_number = 0
    ENDIF
    IF (careful_mod) CALL timestop (error_handle)
  END SUBROUTINE iterator_next_2d_block_s
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Gets the next data block, single/double precision real/complex
!> \param[in,out] iterator   the iterator
!> \param[out] row           row of the data block
!> \param[out] column        column of the data block
!> \param[out] block         pointer to the data block
!> \param[out] transposed    whether the block data is transposed
!> \param[out] block_number  (optional) block number
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE iterator_next_1d_block_c (iterator, row, column, block,&
       transposed, block_number, row_size, col_size, row_offset, col_offset)
    TYPE(dbcsr_iterator), INTENT(INOUT)      :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    COMPLEX(kind=real_4), DIMENSION(:), POINTER :: block
    LOGICAL, INTENT(OUT)                     :: transposed
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size,&
         row_offset, col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'iterator_next_1d_block_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk_p, bp, csize, nze, rsize

!   ---------------------------------------------------------------------------
! If we're pointing to a valid block, return that block.


    IF (iterator%pos .LE. iterator%nblks&
         .AND. iterator%pos .NE. 0) THEN
       row = iterator%row
       column = iterator%col_i(iterator%pos)
       IF (iterator%transpose) CALL swap (row, column)
       blk_p = iterator%blk_p(iterator%pos)
       transposed = blk_p .LT. 0
       bp = ABS (blk_p)
       rsize = iterator%row_size
       csize = iterator%cbs(column)
       nze = rsize * csize
       IF (PRESENT (row_size)) row_size = rsize
       IF (PRESENT (col_size)) col_size = csize
       IF (PRESENT (row_offset)) row_offset = iterator%row_offset
       IF (PRESENT (col_offset)) col_offset = iterator%coff(column)
       CALL dbcsr_get_data (iterator%data_area, block,&
            lb=bp, ub=bp+nze-1)
       IF (PRESENT (block_number)) block_number = iterator%pos
       ! Move to the next non-deleted position.
       CALL iterator_advance (iterator)
       CALL update_row_info (iterator)
    ELSE
       row = 0
       column = 0
       NULLIFY (block)
       IF (PRESENT (block_number)) block_number = 0
    ENDIF
  END SUBROUTINE iterator_next_1d_block_c



! **************************************************************************************************
!> \brief Gets the next data block, single/double precision real/complex
!> \param[in,out] iterator   the iterator
!> \param[out] row           row of the data block
!> \param[out] column        column of the data block
!> \param[out] block         pointer to the data block
!> \param[out] transposed    whether the block data is transposed
!> \param[out] block_number  (optional) block number
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
!> \param row_offset ...
!> \param col_offset ...
! **************************************************************************************************
  SUBROUTINE iterator_next_2d_block_c (iterator, row, column,&
       block, transposed,&
       block_number, row_size, col_size, row_offset, col_offset)
    TYPE(dbcsr_iterator), INTENT(INOUT)      :: iterator
    INTEGER, INTENT(OUT)                     :: row, column
    COMPLEX(kind=real_4), DIMENSION(:, :), &
      POINTER                                :: block
    LOGICAL, INTENT(OUT)                     :: transposed
    INTEGER, INTENT(OUT), OPTIONAL           :: block_number
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size, row_offset, col_offset

    CHARACTER(len=*), PARAMETER :: routineN = 'iterator_next_2d_block_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk_p, bp, csize, nze, rsize, &
                                                block_row_size, block_col_size
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: lin_blk_p
    INTEGER                                  :: error_handle

!   ---------------------------------------------------------------------------
! If we're pointing to a valid block, return that block.

    IF (careful_mod) CALL timeset (routineN, error_handle)
    IF (iterator%pos .LE. iterator%nblks&
         .AND. iterator%pos .NE. 0) THEN
       row = iterator%row
       column = iterator%col_i(iterator%pos)
       IF (iterator%transpose) CALL swap (row, column)
       blk_p = iterator%blk_p(iterator%pos)
       transposed = blk_p .LT. 0
       bp = ABS (blk_p)
       rsize = iterator%row_size
       csize = iterator%cbs(column)
       block_row_size = rsize
       block_col_size = csize
       IF (PRESENT (row_size)) row_size = rsize
       IF (PRESENT (col_size)) col_size = csize
       IF (PRESENT (row_offset)) row_offset = iterator%row_offset
       IF (PRESENT (col_offset)) col_offset = iterator%coff(column)
       nze = rsize * csize
       IF (transposed) CALL swap (rsize, csize)
       CALL dbcsr_get_data (iterator%data_area, lin_blk_p,&
            lb=bp, ub=bp+nze-1)
       CALL pointer_rank_remap2 (block, rsize, csize, lin_blk_p)
       IF (PRESENT (block_number)) block_number = iterator%pos
       ! Move to the next non-deleted position.
       CALL iterator_advance (iterator)
       CALL update_row_info (iterator)
    ELSE
       row = 0
       column = 0
       NULLIFY (block)
       IF (PRESENT (block_number)) block_number = 0
    ENDIF
    IF (careful_mod) CALL timestop (error_handle)
  END SUBROUTINE iterator_next_2d_block_c

END MODULE dbcsr_iterator_operations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR work matrix utilities
!> \author  Urban Borstnik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2010-02-18 Copied from dbcsr_util
! **************************************************************************************************
MODULE dbcsr_work_operations

   USE array_types,                     ONLY: array_data,&
                                              array_hold,&
                                              array_i1d_obj,&
                                              array_new,&
                                              array_nullify,&
                                              array_release,&
                                              array_size
   USE btree_I8_k_cp2d_v,               ONLY: btree_2d_data_c => cp2d
   USE btree_I8_k_dp2d_v,               ONLY: btree_2d_data_d => dp2d
   USE btree_I8_k_sp2d_v,               ONLY: btree_2d_data_s => sp2d
   USE btree_I8_k_zp2d_v,               ONLY: btree_2d_data_z => zp2d
   USE dbcsr_block_operations,          ONLY: block_copy_c,&
                                              block_copy_d,&
                                              block_copy_s,&
                                              block_copy_z,&
                                              dbcsr_data_copy,&
                                              dbcsr_data_set
   USE dbcsr_config,                    ONLY: default_resize_factor
   USE dbcsr_data_methods,              ONLY: &
        dbcsr_data_clear_pointer, dbcsr_data_ensure_size, dbcsr_data_get_memory_type, &
        dbcsr_data_get_size, dbcsr_data_get_size_referenced, dbcsr_data_get_type, dbcsr_data_hold, &
        dbcsr_data_init, dbcsr_data_new, dbcsr_data_release, dbcsr_data_set_size_referenced, &
        dbcsr_data_valid, dbcsr_get_data_p_c, dbcsr_get_data_p_d, dbcsr_get_data_p_s, &
        dbcsr_get_data_p_z
   USE dbcsr_data_operations,           ONLY: dbcsr_data_copyall,&
                                              dbcsr_sort_data,&
                                              dbcsr_switch_data_area
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_has_threads,&
                                              dbcsr_distribution_hold,&
                                              dbcsr_distribution_make_threads,&
                                              dbcsr_distribution_ncols,&
                                              dbcsr_distribution_nrows,&
                                              dbcsr_distribution_release
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_caller_error,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_unimplemented_error_nr,&
                                              dbcsr_wrong_args_error
   USE dbcsr_index_operations,          ONLY: dbcsr_addto_index_array,&
                                              dbcsr_build_row_index,&
                                              dbcsr_clearfrom_index_array,&
                                              dbcsr_index_prune_deleted,&
                                              dbcsr_make_dbcsr_index,&
                                              dbcsr_make_index_exist,&
                                              dbcsr_repoint_index,&
                                              dbcsr_sort_indices
   USE dbcsr_iterator_operations,       ONLY: dbcsr_iterator_blocks_left,&
                                              dbcsr_iterator_next_block,&
                                              dbcsr_iterator_start,&
                                              dbcsr_iterator_stop
   USE dbcsr_mem_methods,               ONLY: dbcsr_memtype_equal
   USE dbcsr_methods,                   ONLY: &
        dbcsr_col_block_sizes, dbcsr_distribution, dbcsr_get_data_memory_type, &
        dbcsr_get_data_size_used, dbcsr_get_data_type, dbcsr_get_index_memory_type, &
        dbcsr_get_matrix_type, dbcsr_get_replication_type, dbcsr_init, dbcsr_is_initialized, &
        dbcsr_matrix_counter, dbcsr_max_col_size, dbcsr_max_row_size, dbcsr_mutable_destroy, &
        dbcsr_mutable_init, dbcsr_mutable_instantiated, dbcsr_mutable_new, dbcsr_mutable_release, &
        dbcsr_name, dbcsr_row_block_sizes, dbcsr_use_mutable, dbcsr_valid_index, &
        dbcsr_wm_use_mutable
   USE dbcsr_ptr_util,                  ONLY: ensure_array_size
   USE dbcsr_toollib,                   ONLY: dbcsr_unpack_i8_2i4,&
                                              sort,&
                                              uppercase
   USE dbcsr_types,                     ONLY: &
        dbcsr_data_obj, dbcsr_distribution_obj, dbcsr_iterator, dbcsr_memtype_default, &
        dbcsr_memtype_type, dbcsr_meta_size, dbcsr_num_slots, dbcsr_obj, dbcsr_repl_col, &
        dbcsr_repl_full, dbcsr_repl_none, dbcsr_repl_row, dbcsr_slot_blk_p, dbcsr_slot_col_i, &
        dbcsr_slot_home_coli, dbcsr_slot_home_pcol, dbcsr_slot_home_prow, dbcsr_slot_home_rowi, &
        dbcsr_slot_home_vpcol, dbcsr_slot_home_vprow, dbcsr_slot_nblkcols_local, &
        dbcsr_slot_nblkcols_total, dbcsr_slot_nblkrows_local, dbcsr_slot_nblkrows_total, &
        dbcsr_slot_nblks, dbcsr_slot_nfullcols_local, dbcsr_slot_nfullcols_total, &
        dbcsr_slot_nfullrows_local, dbcsr_slot_nfullrows_total, dbcsr_slot_nze, dbcsr_slot_row_p, &
        dbcsr_slot_size, dbcsr_type, dbcsr_type_antihermitian, dbcsr_type_antisymmetric, &
        dbcsr_type_complex_4, dbcsr_type_complex_8, dbcsr_type_hermitian, dbcsr_type_no_symmetry, &
        dbcsr_type_real_4, dbcsr_type_real_8, dbcsr_type_real_default, dbcsr_type_symmetric, &
        dbcsr_work_type
   USE dbcsr_util,                      ONLY: convert_sizes_to_offsets,&
                                              dbcsr_calc_block_sizes,&
                                              dbcsr_verify_matrix,&
                                              meta_from_dist
   USE kinds,                           ONLY: default_string_length,&
                                              int_8,&
                                              real_4,&
                                              real_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads, omp_in_parallel

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_work_operations'

   LOGICAL, PARAMETER :: careful_mod = .FALSE.

   PUBLIC :: dbcsr_create, dbcsr_work_create, dbcsr_finalize, dbcsr_special_finalize
   PUBLIC :: dbcsr_add_wm_from_matrix, &
             add_work_coordinate, &
             dbcsr_work_destroy_all

   INTERFACE dbcsr_create
      MODULE PROCEDURE dbcsr_create_new, dbcsr_create_template
   END INTERFACE

   TYPE i_array_p
      INTEGER, DIMENSION(:), POINTER :: p
   END TYPE i_array_p

CONTAINS

! **************************************************************************************************
!> \brief Creates a matrix, allocating the essentials.
!> \param[in,out] matrix      new matrix
!> \param name ...
!> \param[in] dist            distribution_2d distribution
!> \param[in] matrix_type     'N' for normal, 'T' for transposed, 'S' for
!>                            symmetric, and 'A' for antisymmetric
!> \param row_blk_size ...
!> \param col_blk_size ...
!> \param row_blk_size_obj ...
!> \param col_blk_size_obj ...
!> \param[in] nze             (optional) number of elements
!> \param[in] data_type       type of data from 'rRcC' for single/double
!>                            precision real/complex, default is 'R'
!> \param data_buffer ...
!> \param[in] data_memory_type     (optional) allocate indices and data using
!>                            special memory
!> \param[in] index_memory_type     (optional) allocate indices using
!>                            special memory
!> \param max_rbs ...
!> \param max_cbs ...
!> \param row_blk_offset ...
!> \param col_blk_offset ...
!> \param thread_dist ...
!> \param[in] reuse           (optional) reuses an existing matrix, default
!>                            is to create a fresh one
!> \param reuse_arrays ...
!> \param[in] mutable_work    uses the mutable data for working and not the
!>                            append-only data; default is append-only
!> \param make_index ...
!> \param[in] replication_type     replication to be used for this matrix;
!>                                 default is dbcsr_repl_none
!> \par The matrix itself is allocated, as well as the essential parts of
!>      the index. When passed the nze argument, the data is also allocated
!>      to that size.
!>                                 see dbcsr_types.F
! **************************************************************************************************
   SUBROUTINE dbcsr_create_new(matrix, name, dist, matrix_type, &
                               row_blk_size, col_blk_size, row_blk_size_obj, col_blk_size_obj, &
                               nze, data_type, data_buffer, &
                               data_memory_type, index_memory_type, &
                               max_rbs, max_cbs, &
                               row_blk_offset, col_blk_offset, &
                               thread_dist, &
                               reuse, reuse_arrays, mutable_work, make_index, replication_type)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      CHARACTER(len=*), INTENT(IN)                       :: name
      TYPE(dbcsr_distribution_obj), INTENT(IN)           :: dist
      CHARACTER, INTENT(IN)                              :: matrix_type
      INTEGER, DIMENSION(:), INTENT(IN), POINTER         :: row_blk_size, col_blk_size
      TYPE(array_i1d_obj), INTENT(IN), OPTIONAL          :: row_blk_size_obj, col_blk_size_obj
      INTEGER, INTENT(IN), OPTIONAL                      :: nze, data_type
      TYPE(dbcsr_data_obj), INTENT(IN), OPTIONAL         :: data_buffer
      TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL     :: data_memory_type, index_memory_type
      INTEGER, INTENT(IN), OPTIONAL                      :: max_rbs, max_cbs
      TYPE(array_i1d_obj), INTENT(IN), OPTIONAL          :: row_blk_offset, col_blk_offset
      TYPE(dbcsr_distribution_obj), INTENT(IN), OPTIONAL :: thread_dist
      LOGICAL, INTENT(IN), OPTIONAL                      :: reuse, reuse_arrays, mutable_work, &
                                                            make_index
      CHARACTER, INTENT(IN), OPTIONAL                    :: replication_type

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_create_new', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: matrix_type_l
      INTEGER                                            :: handle, my_nze
      INTEGER, DIMENSION(:), POINTER                     :: vec_col_blk_offset, vec_row_blk_offset
      INTEGER, DIMENSION(dbcsr_meta_size)                :: new_meta
      LOGICAL                                            :: hijack, my_make_index

!   ---------------------------------------------------------------------------

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(thread_dist))==-1) EXIT ;  END DO ; ENDIF ! only used with OMP

      CALL timeset(routineN, handle)
      IF (.NOT. dbcsr_is_initialized(matrix)) &
         CALL cp__b("dbcsr/block/dbcsr_work_operations.F",204,"Matrix may not have been initialized with dbcsr_init")

      ! Reuse matrix only if has actually been allocated.
      hijack = ASSOCIATED(matrix%m%index)
      IF (PRESENT(reuse)) hijack = reuse

      my_make_index = .TRUE.
      IF (PRESENT(make_index)) my_make_index = make_index

      IF (.NOT. hijack) THEN
         CALL dbcsr_init(matrix%m)
         matrix%m%refcount = 1
      ENDIF
!$OMP     CRITICAL
      matrix%m%serial_number = dbcsr_matrix_counter
      dbcsr_matrix_counter = dbcsr_matrix_counter+1
!$OMP     END CRITICAL
      ! Mark matrix index as having an invalid index.
      matrix%m%valid = .FALSE.
      matrix%m%name = name
      ! Sets the type of matrix building/modifying work structures.
      IF (PRESENT(mutable_work)) THEN
         matrix%m%work_mutable = mutable_work
      ELSE
         matrix%m%work_mutable = .FALSE.
      ENDIF
      ! Sets the correct data type.
      IF (PRESENT(data_type)) THEN
         SELECT CASE (data_type)
         CASE (dbcsr_type_real_4)
            matrix%m%data_type = dbcsr_type_real_4
         CASE (dbcsr_type_real_8)
            matrix%m%data_type = dbcsr_type_real_8
         CASE (dbcsr_type_complex_4)
            matrix%m%data_type = dbcsr_type_complex_4
         CASE (dbcsr_type_complex_8)
            matrix%m%data_type = dbcsr_type_complex_8
         CASE DEFAULT
            CALL cp__b("dbcsr/block/dbcsr_work_operations.F",242,"Invalid matrix type")
         END SELECT
      ELSE
         matrix%m%data_type = dbcsr_type_real_default
      ENDIF

      matrix%m%data_memory_type = dbcsr_memtype_default
      IF (PRESENT(data_memory_type)) &
         matrix%m%data_memory_type = data_memory_type

      matrix%m%index_memory_type = dbcsr_memtype_default
      IF (PRESENT(index_memory_type)) &
         matrix%m%index_memory_type = index_memory_type

      IF (hijack) THEN
         ! Release/deallocate elements that are replaced or not needed
         ! by the new matrix. This is similar to what dbcsr_destroy
         ! does, except that it keeps the index and data.
         CALL array_release(matrix%m%row_blk_size)
         CALL array_release(matrix%m%col_blk_size)
         CALL array_release(matrix%m%row_blk_offset)
         CALL array_release(matrix%m%col_blk_offset)
         IF (matrix%m%has_local_rows) &
            CALL array_release(matrix%m%local_rows)
         IF (matrix%m%has_global_rows) &
            CALL array_release(matrix%m%global_rows)
         IF (matrix%m%has_local_cols) &
            CALL array_release(matrix%m%local_cols)
         IF (matrix%m%has_global_cols) &
            CALL array_release(matrix%m%global_cols)
         CALL dbcsr_distribution_release(matrix%m%dist)
         IF (ASSOCIATED(matrix%m%wms)) THEN
            CALL dbcsr_work_destroy_all(matrix%m)
         ENDIF
         CALL array_nullify(matrix%m%local_rows)
         CALL array_nullify(matrix%m%global_rows)
         CALL array_nullify(matrix%m%local_cols)
         CALL array_nullify(matrix%m%global_cols)
      ELSE
         ! Invalidate index
         NULLIFY (matrix%m%index)
         ! Invalidate data
         IF (PRESENT(data_buffer)) THEN
            IF (.NOT. dbcsr_data_valid(data_buffer)) &
               CALL cp__b("dbcsr/block/dbcsr_work_operations.F",286,"Input data buffer not valid.")
            CALL dbcsr_assert(matrix%m%data_type, "EQ", &
                              data_buffer%d%data_type, &
                              dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                              "Input buffer data type different by matrix data type.", &
                              291)
            matrix%m%data_memory_type = data_buffer%d%memory_type
            matrix%m%data_area = data_buffer
            CALL dbcsr_data_hold(matrix%m%data_area)
         ELSE
            CALL dbcsr_data_init(matrix%m%data_area)
         ENDIF
      ENDIF
      ! These are always invalidated.
      NULLIFY (matrix%m%row_p, matrix%m%col_i, matrix%m%blk_p, matrix%m%thr_c, &
               matrix%m%coo_l)
      IF (PRESENT(row_blk_size_obj)) THEN
         matrix%m%row_blk_size = row_blk_size_obj
         CALL array_hold(matrix%m%row_blk_size)
      ELSE
         CALL array_new(matrix%m%row_blk_size, row_blk_size, gift=reuse_arrays)
      ENDIF
      IF (PRESENT(max_rbs)) THEN
         matrix%m%max_rbs = max_rbs
      ELSE IF (array_size(matrix%m%row_blk_size) .GT. 0) THEN
         matrix%m%max_rbs = MAXVAL(array_data(matrix%m%row_blk_size))
      ELSE
         matrix%m%max_rbs = 0
      ENDIF
      IF (PRESENT(col_blk_size_obj)) THEN
         matrix%m%col_blk_size = col_blk_size_obj
         CALL array_hold(matrix%m%col_blk_size)
      ELSE
         CALL array_new(matrix%m%col_blk_size, col_blk_size, gift=reuse_arrays)
      ENDIF

      IF (PRESENT(max_cbs)) THEN
         matrix%m%max_cbs = max_cbs
      ELSE IF (array_size(matrix%m%col_blk_size) .GT. 0) THEN
         matrix%m%max_cbs = MAXVAL(array_data(matrix%m%col_blk_size))
      ELSE
         matrix%m%max_cbs = 0
      ENDIF
      !
      CALL dbcsr_assert(array_size(matrix%m%row_blk_size), "EQ", &
                        dbcsr_distribution_nrows(dist), &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "Number of blocked rows does match blocked row distribution.", &
                        334)
      CALL dbcsr_assert(array_size(matrix%m%col_blk_size), "EQ", &
                        dbcsr_distribution_ncols(dist), &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "Number of blocked columns does match blocked column distribution.", &
                        339)
      ! initialize row/col offsets
      IF (PRESENT(row_blk_offset)) THEN
         CALL dbcsr_assert(dbcsr_distribution_nrows(dist)+1, "EQ", &
                           array_size(row_blk_offset), &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Number of blocked offset rows does match blocked row distribution.", &
                           346)
         matrix%m%row_blk_offset = row_blk_offset
         CALL array_hold(matrix%m%row_blk_offset)
      ELSE
         ALLOCATE (vec_row_blk_offset(array_size(matrix%m%row_blk_size)+1))
         CALL convert_sizes_to_offsets(array_data(matrix%m%row_blk_size), vec_row_blk_offset)
         CALL array_new(matrix%m%row_blk_offset, vec_row_blk_offset, gift=.TRUE.)
      ENDIF

      IF (PRESENT(col_blk_offset)) THEN
         CALL dbcsr_assert(dbcsr_distribution_ncols(dist)+1, "EQ", &
                           array_size(col_blk_offset), &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Number of blocked offset columns does match blocked colum distribution.", &
                           360)
         matrix%m%col_blk_offset = col_blk_offset
         CALL array_hold(matrix%m%col_blk_offset)
      ELSE
         ALLOCATE (vec_col_blk_offset(array_size(matrix%m%col_blk_size)+1))
         CALL convert_sizes_to_offsets(array_data(matrix%m%col_blk_size), vec_col_blk_offset)
         CALL array_new(matrix%m%col_blk_offset, vec_col_blk_offset, gift=.TRUE.)
      ENDIF

      matrix%m%dist = dist
      CALL dbcsr_distribution_hold(matrix%m%dist)
!$    IF (.NOT. dbcsr_distribution_has_threads(matrix%m%dist) .AND. PRESENT(thread_dist)) THEN
!$       IF (dbcsr_distribution_has_threads(thread_dist)) THEN
!$          matrix%m%dist%d%num_threads = thread_dist%d%num_threads
!$          matrix%m%dist%d%has_thread_dist = .TRUE.
!$          matrix%m%dist%d%thread_dist = thread_dist%d%thread_dist
!$          CALL array_hold(matrix%m%dist%d%thread_dist)
!$       ENDIF
!$    ENDIF
!$    IF (.NOT. dbcsr_distribution_has_threads(matrix%m%dist)) THEN
!$       CALL dbcsr_distribution_make_threads(matrix%m%dist, &
!$                                            array_data(matrix%m%row_blk_size))
!$    ENDIF
      ! Set up some data.
      IF (my_make_index) THEN
         CALL meta_from_dist(new_meta, dist, array_data(matrix%m%row_blk_size), &
                             array_data(matrix%m%col_blk_size))
         matrix%m%nblkrows_total = new_meta(dbcsr_slot_nblkrows_total)
         matrix%m%nblkcols_total = new_meta(dbcsr_slot_nblkcols_total)
         matrix%m%nfullrows_total = new_meta(dbcsr_slot_nfullrows_total)
         matrix%m%nfullcols_total = new_meta(dbcsr_slot_nfullcols_total)
         matrix%m%nblkrows_local = new_meta(dbcsr_slot_nblkrows_local)
         matrix%m%nblkcols_local = new_meta(dbcsr_slot_nblkcols_local)
         matrix%m%nfullrows_local = new_meta(dbcsr_slot_nfullrows_local)
         matrix%m%nfullcols_local = new_meta(dbcsr_slot_nfullcols_local)
      ENDIF
      my_nze = 0; IF (PRESENT(nze)) my_nze = nze
      matrix%m%nblks = 0
      matrix%m%nze = 0

      IF (PRESENT(replication_type)) THEN
         IF (replication_type .NE. dbcsr_repl_none &
             .AND. replication_type .NE. dbcsr_repl_full &
             .AND. replication_type .NE. dbcsr_repl_row &
             .AND. replication_type .NE. dbcsr_repl_col) &
            CALL cp__b("dbcsr/block/dbcsr_work_operations.F",405,"Invalid replication type '"//replication_type//"'")
         IF (replication_type .EQ. dbcsr_repl_row .OR. replication_type .EQ. dbcsr_repl_col) &
            CALL cp__w("dbcsr/block/dbcsr_work_operations.F",407,"Row and column replication not fully supported")
         matrix%m%replication_type = replication_type
      ELSE
         matrix%m%replication_type = dbcsr_repl_none
      ENDIF
      !
      ! Setup a matrix from scratch
      IF (.NOT. hijack) THEN
         IF (.NOT. PRESENT(data_buffer)) THEN
            CALL dbcsr_data_new(matrix%m%data_area, matrix%m%data_type, my_nze, &
                                memory_type=matrix%m%data_memory_type)
            CALL dbcsr_data_set_size_referenced(matrix%m%data_area, 0)
         ENDIF
         !
         IF (my_make_index) THEN
            NULLIFY (matrix%m%index)
            CALL ensure_array_size(matrix%m%index, lb=1, ub=dbcsr_num_slots, &
                                   zero_pad=.TRUE., memory_type=matrix%m%index_memory_type)
         ENDIF
      ENDIF
      IF (my_make_index) THEN
         IF (LBOUND(matrix%m%index, 1) .GT. 1 &
             .OR. UBOUND(matrix%m%index, 1) .LT. dbcsr_num_slots) &
            CALL cp__b("dbcsr/block/dbcsr_work_operations.F",430,"Index is not large enough")
         matrix%m%index(1:dbcsr_num_slots) = 0
         matrix%m%index(1:dbcsr_meta_size) = new_meta(1:dbcsr_meta_size)
         matrix%m%index(dbcsr_slot_size) = dbcsr_num_slots
      ENDIF
      !
      matrix%m%symmetry = .FALSE.
      matrix%m%negate_real = .FALSE.
      matrix%m%negate_imaginary = .FALSE.
      !matrix%m%transpose = .FALSE.
      matrix_type_l = matrix_type
      CALL uppercase(matrix_type_l)
      SELECT CASE (matrix_type_l)
      CASE (dbcsr_type_no_symmetry)
      CASE (dbcsr_type_symmetric)
         matrix%m%symmetry = .TRUE.
      CASE (dbcsr_type_antisymmetric)
         matrix%m%symmetry = .TRUE.
         matrix%m%negate_real = .TRUE.
         matrix%m%negate_imaginary = .TRUE.
      CASE (dbcsr_type_hermitian)
         matrix%m%symmetry = .TRUE.
         matrix%m%negate_imaginary = .TRUE.
      CASE (dbcsr_type_antihermitian)
         matrix%m%symmetry = .TRUE.
         matrix%m%negate_real = .TRUE.
      CASE DEFAULT
         CALL cp__b("dbcsr/block/dbcsr_work_operations.F",457,"Invalid matrix type.")
      END SELECT
      matrix%m%bcsc = .FALSE.
      matrix%m%local_indexing = .FALSE.
      matrix%m%list_indexing = .FALSE.
      CALL array_nullify(matrix%m%local_rows)
      CALL array_nullify(matrix%m%global_rows)
      CALL array_nullify(matrix%m%local_cols)
      CALL array_nullify(matrix%m%global_cols)
      matrix%m%has_local_rows = .FALSE.
      matrix%m%has_global_rows = .FALSE.
      matrix%m%has_local_cols = .FALSE.
      matrix%m%has_global_cols = .FALSE.
      IF (my_make_index) THEN
         CALL dbcsr_make_index_exist(matrix%m)
      ENDIF
      matrix%m%valid = .TRUE.
      CALL timestop(handle)
   END SUBROUTINE dbcsr_create_new

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param template ...
!> \param name ...
!> \param dist ...
!> \param matrix_type ...
!> \param row_blk_size ...
!> \param col_blk_size ...
!> \param nze ...
!> \param data_type ...
!> \param data_buffer ...
!> \param data_memory_type ...
!> \param index_memory_type ...
!> \param max_rbs ...
!> \param max_cbs ...
!> \param row_blk_offset ...
!> \param col_blk_offset ...
!> \param reuse_arrays ...
!> \param mutable_work ...
!> \param make_index ...
!> \param replication_type ...
! **************************************************************************************************
   SUBROUTINE dbcsr_create_template(matrix, template, name, dist, matrix_type, &
                                    row_blk_size, col_blk_size, nze, data_type, &
                                    data_buffer, data_memory_type, index_memory_type, &
                                    max_rbs, max_cbs, &
                                    row_blk_offset, col_blk_offset, &
                                    reuse_arrays, mutable_work, make_index, replication_type)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      TYPE(dbcsr_obj), INTENT(IN)                        :: template
      CHARACTER(len=*), INTENT(IN), OPTIONAL             :: name
      TYPE(dbcsr_distribution_obj), INTENT(IN), OPTIONAL :: dist
      CHARACTER, INTENT(IN), OPTIONAL                    :: matrix_type
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL, &
         POINTER                                         :: row_blk_size, col_blk_size
      INTEGER, INTENT(IN), OPTIONAL                      :: nze, data_type
      TYPE(dbcsr_data_obj), INTENT(IN), OPTIONAL         :: data_buffer
      TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL     :: data_memory_type, index_memory_type
      INTEGER, INTENT(IN), OPTIONAL                      :: max_rbs, max_cbs
      TYPE(array_i1d_obj), INTENT(IN), OPTIONAL          :: row_blk_offset, col_blk_offset
      LOGICAL, INTENT(IN), OPTIONAL                      :: reuse_arrays, mutable_work, make_index
      CHARACTER, INTENT(IN), OPTIONAL                    :: replication_type

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_create_template', &
         routineP = moduleN//':'//routineN

      CHARACTER                                          :: new_matrix_type, new_replication_type
      CHARACTER(len=default_string_length)               :: new_name
      INTEGER                                            :: new_data_type, new_max_cbs, new_max_rbs
      INTEGER, DIMENSION(:), POINTER                     :: new_col_blk_size, new_row_blk_size, &
                                                            vec_col_blk_offset, vec_row_blk_offset
      LOGICAL                                            :: my_make_index, new_mutable_work
      TYPE(array_i1d_obj)                                :: new_col_blk_offset, new_row_blk_offset
      TYPE(dbcsr_distribution_obj)                       :: new_dist
      TYPE(dbcsr_memtype_type)                           :: new_data_memory_type, &
                                                            new_index_memory_type

!   ---------------------------------------------------------------------------

      IF (PRESENT(name)) THEN
         new_name = TRIM(name)
      ELSE
         new_name = TRIM(dbcsr_name(template))
      ENDIF
      IF (PRESENT(dist)) THEN
         new_dist = dist
      ELSE
         new_dist = dbcsr_distribution(template)
      ENDIF
      IF (PRESENT(matrix_type)) THEN
         new_matrix_type = matrix_type
      ELSE
         new_matrix_type = dbcsr_get_matrix_type(template)
      ENDIF
      IF (PRESENT(max_rbs)) new_max_rbs = max_rbs
      IF (PRESENT(row_blk_offset)) new_row_blk_offset = row_blk_offset

      IF (PRESENT(row_blk_size)) THEN
         new_row_blk_size => row_blk_size
         IF (.NOT. PRESENT(max_rbs)) &
            new_max_rbs = MAXVAL(new_row_blk_size)
         IF (.NOT. PRESENT(row_blk_offset)) THEN
            ALLOCATE (vec_row_blk_offset(SIZE(new_row_blk_size)+1))
            CALL convert_sizes_to_offsets(new_row_blk_size, vec_row_blk_offset)
            CALL array_new(new_row_blk_offset, vec_row_blk_offset, gift=.TRUE.)
         ENDIF
      ELSE
         new_row_blk_size => dbcsr_row_block_sizes(template)
         IF (.NOT. PRESENT(max_rbs)) &
            new_max_rbs = dbcsr_max_row_size(template)
         IF (.NOT. PRESENT(row_blk_offset)) &
            new_row_blk_offset = template%m%row_blk_offset
      ENDIF
      IF (PRESENT(max_cbs)) new_max_cbs = max_cbs
      IF (PRESENT(col_blk_offset)) new_col_blk_offset = col_blk_offset
      IF (PRESENT(col_blk_size)) THEN
         new_col_blk_size => col_blk_size
         IF (.NOT. PRESENT(max_cbs)) &
            new_max_cbs = MAXVAL(new_col_blk_size)
         IF (.NOT. PRESENT(col_blk_offset)) THEN
            ALLOCATE (vec_col_blk_offset(SIZE(new_col_blk_size)+1))
            CALL convert_sizes_to_offsets(new_col_blk_size, vec_col_blk_offset)
            CALL array_new(new_col_blk_offset, vec_col_blk_offset, gift=.TRUE.)
         ENDIF
      ELSE
         new_col_blk_size => dbcsr_col_block_sizes(template)
         IF (.NOT. PRESENT(max_cbs)) &
            new_max_cbs = dbcsr_max_col_size(template)
         IF (.NOT. PRESENT(col_blk_offset)) &
            new_col_blk_offset = template%m%col_blk_offset
      ENDIF
      IF (PRESENT(data_type)) THEN
         new_data_type = data_type
      ELSE
         new_data_type = dbcsr_get_data_type(template)
      ENDIF
      IF (PRESENT(data_memory_type)) THEN
         new_data_memory_type = data_memory_type
      ELSE
         new_data_memory_type = dbcsr_get_data_memory_type(template)
      ENDIF
      IF (PRESENT(index_memory_type)) THEN
         new_index_memory_type = index_memory_type
      ELSE
         new_index_memory_type = dbcsr_get_index_memory_type(template)
      ENDIF
      IF (PRESENT(replication_type)) THEN
         new_replication_type = replication_type
      ELSE
         new_replication_type = dbcsr_get_replication_type(template)
      ENDIF
      IF (PRESENT(mutable_work)) THEN
         new_mutable_work = mutable_work
      ELSE
         new_mutable_work = dbcsr_use_mutable(template%m)
      ENDIF
      IF (PRESENT(row_blk_size)) THEN
         CALL dbcsr_create(matrix, name=new_name, dist=new_dist, &
                           matrix_type=new_matrix_type, &
                           row_blk_size=new_row_blk_size, &
                           col_blk_size=new_col_blk_size, &
                           nze=nze, &
                           data_type=new_data_type, &
                           data_buffer=data_buffer, &
                           data_memory_type=new_data_memory_type, &
                           index_memory_type=new_index_memory_type, &
                           max_rbs=new_max_rbs, max_cbs=new_max_cbs, &
                           row_blk_offset=new_row_blk_offset, col_blk_offset=new_col_blk_offset, &
                           reuse_arrays=reuse_arrays, &
                           mutable_work=new_mutable_work, &
                           make_index=make_index, &
                           replication_type=new_replication_type)
      ELSE
         CALL dbcsr_create(matrix, name=new_name, dist=new_dist, &
                           matrix_type=new_matrix_type, &
                           row_blk_size=new_row_blk_size, &
                           col_blk_size=new_col_blk_size, &
                           row_blk_size_obj=template%m%row_blk_size, &
                           col_blk_size_obj=template%m%col_blk_size, &
                           nze=nze, &
                           data_type=new_data_type, &
                           data_buffer=data_buffer, &
                           data_memory_type=new_data_memory_type, &
                           index_memory_type=new_index_memory_type, &
                           max_rbs=new_max_rbs, max_cbs=new_max_cbs, &
                           row_blk_offset=new_row_blk_offset, col_blk_offset=new_col_blk_offset, &
                           thread_dist=dbcsr_distribution(template), &
                           reuse_arrays=reuse_arrays, &
                           mutable_work=new_mutable_work, &
                           make_index=make_index, &
                           replication_type=new_replication_type)
      ENDIF
      ! Copy stuff from the meta-array.  These are not normally needed,
      ! but have to be here for creating matrices from "image" matrices.
      my_make_index = .TRUE.
      IF (PRESENT(make_index)) my_make_index = make_index
      IF (my_make_index) THEN
         matrix%m%index(dbcsr_slot_home_prow) = template%m%index(dbcsr_slot_home_prow)
         matrix%m%index(dbcsr_slot_home_rowi) = template%m%index(dbcsr_slot_home_rowi)
         matrix%m%index(dbcsr_slot_home_pcol) = template%m%index(dbcsr_slot_home_pcol)
         matrix%m%index(dbcsr_slot_home_coli) = template%m%index(dbcsr_slot_home_coli)
         matrix%m%index(dbcsr_slot_home_vprow) = template%m%index(dbcsr_slot_home_vprow)
         matrix%m%index(dbcsr_slot_home_vpcol) = template%m%index(dbcsr_slot_home_vpcol)
      ENDIF
      IF (PRESENT(row_blk_size) .AND. .NOT. PRESENT(row_blk_offset)) THEN
         CALL array_release(new_row_blk_offset)
      ENDIF
      IF (PRESENT(col_blk_size) .AND. .NOT. PRESENT(col_blk_offset)) THEN
         CALL array_release(new_col_blk_offset)
      ENDIF

   END SUBROUTINE dbcsr_create_template

! **************************************************************************************************
!> \brief Initializes one work matrix
!> \param[out] wm             initialized work matrix
!> \param data_type ...
!> \param[in] nblks_guess     (optional) estimated number of blocks
!> \param[in] sizedata_guess  (optional) estimated size of data
!> \param memory_type ...
! **************************************************************************************************
   SUBROUTINE dbcsr_init_wm(wm, data_type, nblks_guess, sizedata_guess, memory_type)
      TYPE(dbcsr_work_type), INTENT(OUT)                 :: wm
      INTEGER, INTENT(IN)                                :: data_type
      INTEGER, INTENT(IN), OPTIONAL                      :: nblks_guess, sizedata_guess
      TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL     :: memory_type

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_init_wm', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nblks, stat

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, handle)
      wm%lastblk = 0
      wm%datasize = 0
      ! Index
      IF (PRESENT(nblks_guess)) THEN
         nblks = nblks_guess
         CALL dbcsr_assert(nblks_guess, "GE", 0, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "Can not have negative block count.", 699)
         ALLOCATE (wm%row_i(nblks), stat=stat)
         IF (stat /= 0) CALL cp__b("dbcsr/block/dbcsr_work_operations.F",701,"wm%row_i")
         ALLOCATE (wm%col_i(nblks), stat=stat)
         IF (stat /= 0) CALL cp__b("dbcsr/block/dbcsr_work_operations.F",703,"wm%col_i")
         ALLOCATE (wm%blk_p(nblks), stat=stat)
         IF (stat /= 0) CALL cp__b("dbcsr/block/dbcsr_work_operations.F",705,"wm%blk_p")
      ELSE
         NULLIFY (wm%row_i, wm%col_i, wm%blk_p)
         !nblks = CEILING (REAL (matrix%nblkrows_local * matrix%nblkcols_local)&
         !     / REAL (dbcsr_mp_numnodes (dbcsr_distribution_mp (matrix%dist))))
      ENDIF
      ! Data
      CALL dbcsr_data_init(wm%data_area)
      IF (PRESENT(sizedata_guess)) THEN
         CALL dbcsr_assert(sizedata_guess, "GE", 0, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "Can not have negative data size.", 716)
         CALL dbcsr_data_new(wm%data_area, data_type, &
                             data_size=sizedata_guess, memory_type=memory_type)
      ELSE
         CALL dbcsr_data_new(wm%data_area, data_type, memory_type=memory_type)
      ENDIF
      CALL dbcsr_mutable_init(wm%mutable)
      IF (careful_mod) CALL timestop(handle)
   END SUBROUTINE dbcsr_init_wm

! **************************************************************************************************
!> \brief Creates a the working matrix(es) for a DBCSR matrix.
!> \param[out] matrix         new matrix
!> \param[in] nblks_guess     (optional) estimated number of blocks
!> \param[in] sizedata_guess  (optional) estimated size of data
!> \param[in] n               (optional) number work matrices to create,
!>                            default is 1
!> \param[in] work_mutable    (optional) use mutable work type, default is
!>                            what was specified in create
!> \param memory_type ...
! **************************************************************************************************
   SUBROUTINE dbcsr_work_create(matrix, nblks_guess, sizedata_guess, n, &
                                work_mutable, memory_type)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      INTEGER, INTENT(IN), OPTIONAL                      :: nblks_guess, sizedata_guess, n
      LOGICAL, INTENT(in), OPTIONAL                      :: work_mutable
      TYPE(dbcsr_memtype_type), INTENT(IN), OPTIONAL     :: memory_type

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_work_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iw, nw, ow
      LOGICAL                                            :: wms_new, wms_realloc
      TYPE(dbcsr_work_type), DIMENSION(:), POINTER       :: wms

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (.NOT. dbcsr_is_initialized(matrix)) &
         CALL cp__b("dbcsr/block/dbcsr_work_operations.F",755,"Can not create work matrix for nonexisting matrix object.")
      IF (PRESENT(n)) THEN
         nw = n
      ELSE
         nw = 1
!$       IF (omp_in_parallel()) THEN
!$          nw = omp_get_num_threads()
!$       ELSE
!$          nw = omp_get_max_threads()
!$       ENDIF
      ENDIF
!$OMP     MASTER
      wms_new = .NOT. ASSOCIATED(matrix%m%wms)
      wms_realloc = .FALSE.
      IF (ASSOCIATED(matrix%m%wms)) THEN
         ow = SIZE(matrix%m%wms)
         IF (ow .LT. nw) &
            CALL cp__w("dbcsr/block/dbcsr_work_operations.F",772,"Number of work matrices less than threads.")
         IF (ow .LT. nw) wms_realloc = .TRUE.
      ENDIF
      IF (PRESENT(work_mutable)) THEN
         matrix%m%work_mutable = work_mutable
      ENDIF
      IF (wms_realloc) THEN
         ALLOCATE (wms(nw))
         wms(1:ow) = matrix%m%wms(1:ow)
         DEALLOCATE (matrix%m%wms)
         matrix%m%wms => wms
         DO iw = ow+1, nw
            CALL dbcsr_init_wm(matrix%m%wms(iw), matrix%m%data_type, &
                               nblks_guess=nblks_guess, sizedata_guess=sizedata_guess, &
                               memory_type=memory_type)
            IF (matrix%m%work_mutable) &
               CALL dbcsr_mutable_new(matrix%m%wms(iw)%mutable, &
                                      dbcsr_get_data_type(matrix))
         END DO
      ENDIF
      IF (wms_new) THEN
         ALLOCATE (matrix%m%wms(nw))
         DO iw = 1, nw
            CALL dbcsr_init_wm(matrix%m%wms(iw), matrix%m%data_type, &
                               nblks_guess=nblks_guess, sizedata_guess=sizedata_guess, &
                               memory_type=memory_type)
            IF (matrix%m%work_mutable) &
               CALL dbcsr_mutable_new(matrix%m%wms(iw)%mutable, &
                                      dbcsr_get_data_type(matrix))
         END DO
      ENDIF
      matrix%m%valid = .FALSE.
!$OMP     END MASTER
      CALL timestop(handle)
   END SUBROUTINE dbcsr_work_create

! **************************************************************************************************
!> \brief Creates the final dbcsr_type matrix from the working matrix.
!>
!> Work matrices (array or tree-based) are merged into the base DBCSR matrix.
!>
!> If a matrix is marked as having a valid index, then nothing is done.
!>
!> Deleted blocks are pruned from the index.
!> \param[in,out] matrix      final matrix
!> \param[in] reshuffle       whether the data should be reshuffled,
!>                            default is false
! **************************************************************************************************
   SUBROUTINE dbcsr_finalize(matrix, reshuffle)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: reshuffle

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_finalize', routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: handle, i, nblks, nwms, start_offset
      INTEGER, ALLOCATABLE, DIMENSION(:), SAVE, TARGET   :: empty_row_p
      INTEGER, DIMENSION(:), POINTER, SAVE               :: old_blk_p, old_col_i, old_row_p
      LOGICAL                                            :: can_quick, fake_row_p, sort_data, spawn

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

!$OMP     BARRIER
      IF (.NOT. dbcsr_is_initialized(matrix)) &
         CALL cp__b("dbcsr/block/dbcsr_work_operations.F",838,"Can not finalize uninitialized matrix.")

      ! If the matrix is not marked as dirty then skip the work.
      IF (dbcsr_valid_index(matrix)) THEN
         !"No need to finalize a valid matrix, skipping."
         !
         ! A matrix with a valid index should not have associated work
         ! arrays.  This may happen when this routine is called on a
         ! matrix that was not changed.
!$OMP        BARRIER
!$OMP        MASTER
         IF (ASSOCIATED(matrix%m%wms)) &
            CALL dbcsr_work_destroy_all(matrix%m)
         matrix%m%valid = .TRUE.
!$OMP        END MASTER
!$OMP        BARRIER
         CALL timestop(handle)
         RETURN
      ENDIF
      !
      ! If possible, data copying is avoided.
      IF (PRESENT(reshuffle)) THEN
         sort_data = reshuffle
      ELSE
         sort_data = .FALSE.
      ENDIF
      !
      ! Now make sure that a valid row_p exists. Also clear the row_p if
      ! the matrix is declared to have 0 blocks.
!$OMP     MASTER
      fake_row_p = .NOT. ASSOCIATED(matrix%m%row_p)
      IF (ASSOCIATED(matrix%m%row_p)) THEN
         fake_row_p = SIZE(matrix%m%row_p) .LE. 1
      ENDIF
      fake_row_p = fake_row_p .OR. matrix%m%nblks .EQ. 0
!$OMP     END MASTER
      !
      ! See where data will be appended in the main data
      ! area. Alternatively, set to the start if the matrix is declared
      ! to have no data. (This value is ignored if reshuffle is true
      ! because the main data area is always new.)
      start_offset = matrix%m%nze
      i = dbcsr_get_data_size_used(matrix)
!$OMP     MASTER
      matrix%m%nze = 0
!$OMP     END MASTER
!$OMP     BARRIER
!$OMP     ATOMIC
      matrix%m%nze = matrix%m%nze+i
!$OMP     BARRIER
      IF (dbg) THEN
         WRITE (*, *) routineN//" sizes", matrix%m%nze, i, &
            dbcsr_data_get_size_referenced(matrix%m%data_area), &
            dbcsr_data_get_size(matrix%m%data_area)
      ENDIF
      IF (.FALSE. .AND. dbcsr_data_get_size_referenced(matrix%m%data_area) .NE. &
          matrix%m%nze) THEN
         IF (matrix%m%nze .NE. dbcsr_data_get_size_referenced(matrix%m%data_area)) &
            CALL cp__w("dbcsr/block/dbcsr_work_operations.F",896,"Should reshuffle.")
         IF (ASSOCIATED(matrix%m%wms)) THEN
            sort_data = .NOT. dbcsr_wm_use_mutable(matrix%m%wms(1))
         ENDIF
      ENDIF
      IF (sort_data .AND. matrix%m%nze .GT. 0) THEN
         CALL dbcsr_add_wm_from_matrix(matrix)
         matrix%m%nze = 0
!$OMP        MASTER
         fake_row_p = .TRUE.
!$OMP        END MASTER
      ENDIF
      start_offset = dbcsr_data_get_size_referenced(matrix%m%data_area)+1
      IF (matrix%m%nze .EQ. 0) start_offset = 1
!$OMP     MASTER
      matrix%m%index(dbcsr_slot_nze) = matrix%m%nze
      IF (fake_row_p) THEN
         ALLOCATE (empty_row_p(matrix%m%nblkrows_total+1))
         empty_row_p(:) = 0
         CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_row_p, &
                                      DATA=empty_row_p, extra=0)
         CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_col_i, &
                                      reservation=0)
         CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_blk_p, &
                                      reservation=0)
         CALL dbcsr_repoint_index(matrix%m)
      ENDIF
!$OMP     END MASTER
      !
!$OMP     BARRIER
      can_quick = can_quickly_finalize(matrix)
!$OMP     BARRIER
      ! If the matrix, work matricies, and environment fit several
      ! criteria, then a quick O(1) finalization is performed.
      IF (can_quick .AND. .NOT. sort_data) THEN
         CALL quick_finalize(matrix)
      ELSE
         !
!$OMP        MASTER
         !
         ! Create work matrices if not yet existing
         IF (.NOT. ASSOCIATED(matrix%m%wms)) THEN
            nwms = 1
!$          nwms = omp_get_num_threads()
            CALL dbcsr_work_create(matrix, n=nwms)
         ENDIF
!$OMP        END MASTER
!$OMP        BARRIER
         !
         ! Ensure index arrays at least exist.
!$OMP        DO SCHEDULE (STATIC, 1)
         DO i = 1, SIZE(matrix%m%wms)
            IF (.NOT. ASSOCIATED(matrix%m%wms(i)%row_i)) THEN
               CALL ensure_array_size(matrix%m%wms(i)%row_i, ub=0)
            ENDIF
            IF (.NOT. ASSOCIATED(matrix%m%wms(i)%col_i)) THEN
               CALL ensure_array_size(matrix%m%wms(i)%col_i, ub=0)
            ENDIF
            IF (.NOT. ASSOCIATED(matrix%m%wms(i)%blk_p)) THEN
               CALL ensure_array_size(matrix%m%wms(i)%blk_p, ub=0)
            ENDIF
         ENDDO
!$OMP        ENDDO
         !
         ! Check for deleted blocks
!$OMP        MASTER
         nblks = matrix%m%row_p(matrix%m%nblkrows_total+1)
         IF (ANY(matrix%m%blk_p(1:nblks) .EQ. 0)) THEN
            CALL dbcsr_index_prune_deleted(matrix)
         ENDIF
         old_row_p => matrix%m%row_p
         old_col_i => matrix%m%col_i
         old_blk_p => matrix%m%blk_p
!$OMP        END MASTER
         !
!$OMP        BARRIER
         ! Check to see if we will need to create a parallel environment
         ! (needed when there are multiple work matrices but we are not
         ! in an OpenMP parallel section.)
         !
         ! A parallel section is created is used when the matrix has
         ! more work matrices. It's a shortcut when the finalize is
         ! called from a non-parallel environment whereas the matrix was
         ! built/modified in a parallel environment
         nwms = SIZE(matrix%m%wms)
         spawn = .FALSE.
!$       IF (.NOT. OMP_IN_PARALLEL()) THEN
!$          IF (nwms .GT. 1) spawn = .TRUE.
!$       ENDIF
         IF (spawn) THEN
!$OMP           PARALLEL IF (spawn) &
!$OMP                    DEFAULT (NONE) &
!$OMP                    SHARED (matrix, old_row_p, old_col_i, old_blk_p,&
!$OMP                            start_offset, sort_data)
            CALL dbcsr_merge_all(matrix%m, &
                                 old_row_p, old_col_i, old_blk_p, &
                                 sort_data=sort_data)
!$OMP           END PARALLEL
         ELSE
            CALL dbcsr_merge_all(matrix%m, &
                                 old_row_p, old_col_i, old_blk_p, &
                                 sort_data=sort_data)
         ENDIF
      ENDIF
!$OMP BARRIER
!$OMP MASTER
      ! Clean up.
      IF (ASSOCIATED(matrix%m%wms)) THEN
         CALL dbcsr_work_destroy_all(matrix%m)
      ENDIF
      matrix%m%valid = .TRUE.
!$OMP END MASTER
!$OMP BARRIER
      IF (dbg) THEN
!$OMP        SINGLE
         CALL dbcsr_verify_matrix(matrix)
!$OMP        END SINGLE
      ENDIF
!$OMP MASTER
      IF (fake_row_p) THEN
         DEALLOCATE (empty_row_p)
      ENDIF
!$OMP END MASTER
!$OMP BARRIER
      CALL timestop(handle)
   END SUBROUTINE dbcsr_finalize

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param reshuffle ...
! **************************************************************************************************
   SUBROUTINE dbcsr_special_finalize(matrix, reshuffle)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      LOGICAL, INTENT(IN), OPTIONAL                      :: reshuffle

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_special_finalize', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: handle
      LOGICAL                                            :: can_quick, sort_data

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF (.NOT. dbcsr_is_initialized(matrix)) &
         CALL cp__b("dbcsr/block/dbcsr_work_operations.F",1044,"Can not finalize uninitialized matrix.")
      IF (matrix%m%nblks .NE. 0) &
         CALL cp__b("dbcsr/block/dbcsr_work_operations.F",1046,"Optimized finalize requires empty matrix.")
      IF (dbcsr_valid_index(matrix)) &
         CALL cp__b("dbcsr/block/dbcsr_work_operations.F",1048,"Optimized finalize requires invalid matrix.")
      IF (.NOT. ASSOCIATED(matrix%m%wms)) &
         CALL cp__b("dbcsr/block/dbcsr_work_operations.F",1050,"Optimized finalize requires work matrices exist.")
      IF (SIZE(matrix%m%wms) .NE. 1) &
         CALL cp__b("dbcsr/block/dbcsr_work_operations.F",1052,"Optimized finalize requires single work matrix")

      !
      ! If possible, data copying is avoided.
      IF (PRESENT(reshuffle)) THEN
         sort_data = reshuffle
      ELSE
         sort_data = .FALSE.
      ENDIF
!$OMP     BARRIER
      can_quick = can_quickly_finalize(matrix)
!$OMP     BARRIER

      ! If the matrix, work matricies, and environment fit several
      ! criteria, then a quick O(1) finalization is performed.
      IF (can_quick .AND. .NOT. sort_data) THEN
         CALL quick_finalize(matrix)
      ELSE
         IF (.NOT. sort_data) &
            CALL cp__b("dbcsr/block/dbcsr_work_operations.F",1071,"merge_single_wm only supports data sorting")
         !
         ! Ensure needed index arrays at least exist.
!$OMP        MASTER
         !
         IF (.NOT. ASSOCIATED(matrix%m%wms(1)%row_i)) THEN
            CALL ensure_array_size(matrix%m%wms(1)%row_i, ub=0)
         ENDIF
         IF (.NOT. ASSOCIATED(matrix%m%wms(1)%col_i)) THEN
            CALL ensure_array_size(matrix%m%wms(1)%col_i, ub=0)
         ENDIF
         IF (.NOT. ASSOCIATED(matrix%m%wms(1)%blk_p)) THEN
            CALL ensure_array_size(matrix%m%wms(1)%blk_p, ub=0)
         ENDIF
         !
!$OMP        END MASTER
!$OMP        BARRIER
         !
!$OMP        PARALLEL DEFAULT (NONE), SHARED(matrix)
         CALL dbcsr_merge_single_wm(matrix%m)
!$OMP        END PARALLEL

      ENDIF

!$OMP     MASTER
      ! Clean up.
      IF (ASSOCIATED(matrix%m%wms)) THEN
         CALL dbcsr_work_destroy_all(matrix%m)
      ENDIF
      matrix%m%valid = .TRUE.
!$OMP     END MASTER
!$OMP     BARRIER
      IF (dbg) THEN
!$OMP        SINGLE
         CALL dbcsr_verify_matrix(matrix)
!$OMP        END SINGLE
      ENDIF
!$OMP     BARRIER
      CALL timestop(handle)
   END SUBROUTINE dbcsr_special_finalize

! **************************************************************************************************
!> \brief Checks whether the matrix can be finalized with minimal copying.
!> \param[in] matrix          matrix to check
!> \retval quick              whether matrix can be quickly finalized
! **************************************************************************************************
   FUNCTION can_quickly_finalize(matrix) RESULT(quick)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      LOGICAL                                            :: quick

!   ---------------------------------------------------------------------------

      IF (ASSOCIATED(matrix%m%wms)) THEN
         quick = matrix%m%nblks .EQ. 0
         quick = quick .AND. SIZE(matrix%m%wms) .EQ. 1 .AND. &
                 .NOT. dbcsr_wm_use_mutable(matrix%m%wms(1))
         IF (quick) THEN
            quick = quick .AND. &
                    dbcsr_memtype_equal( &
                    dbcsr_data_get_memory_type(matrix%m%wms(1)%data_area), &
                    dbcsr_data_get_memory_type(matrix%m%data_area))
            quick = quick .AND. &
                    ASSOCIATED(matrix%m%wms(1)%row_i)
         ENDIF
      ELSE
         quick = .FALSE.
      ENDIF
   END FUNCTION can_quickly_finalize

! **************************************************************************************************
!> \brief Performs quick finalization of matrix
!>
!> The data area from the work matrix is accepted as the new matrix's data
!> area and the index is built from the work matrix.
!> \param[in,out] matrix      matrix to finalize
! **************************************************************************************************
   SUBROUTINE quick_finalize(matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'quick_finalize', routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle, nblks, nrows

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
!$OMP     SECTIONS
!$OMP     SECTION
      nblks = matrix%m%wms(1)%lastblk
      nrows = matrix%m%nblkrows_total
      CALL dbcsr_sort_indices(nblks, &
                              matrix%m%wms(1)%row_i, &
                              matrix%m%wms(1)%col_i, &
                              matrix%m%wms(1)%blk_p)
      CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_row_p)
      CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_col_i)
      CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_blk_p)
      CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_row_p, &
                                   reservation=nrows+1, extra=2*nblks)
      CALL dbcsr_make_dbcsr_index(matrix%m%row_p, matrix%m%wms(1)%row_i, &
                                  nrows, nblks)
      CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_col_i, &
                                   DATA=matrix%m%wms(1)%col_i(1:nblks))
      CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_blk_p, &
                                   DATA=matrix%m%wms(1)%blk_p(1:nblks))
      matrix%m%nblks = nblks
      matrix%m%nze = matrix%m%wms(1)%datasize
      matrix%m%index(dbcsr_slot_nblks) = nblks
      matrix%m%index(dbcsr_slot_nze) = matrix%m%wms(1)%datasize
      CALL dbcsr_repoint_index(matrix%m)
!$OMP     SECTION
      CALL dbcsr_switch_data_area(matrix, matrix%m%wms(1)%data_area)
!$OMP     END SECTIONS
      CALL timestop(error_handle)
   END SUBROUTINE quick_finalize

! **************************************************************************************************
!> \brief Creates a work matrix from the data present in a finalized matrix.
!> \param[in,out] matrix      DBCSR matrix
!> \param[in] limits          (optional) the limits to use for copying
! **************************************************************************************************
   SUBROUTINE dbcsr_add_wm_from_matrix(matrix, limits)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      INTEGER, DIMENSION(4), INTENT(IN), OPTIONAL        :: limits

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_add_wm_from_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ithread, nthreads, nwms, &
                                                            old_nwms, size_used
      LOGICAL                                            :: preexists

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
!$OMP     BARRIER
      preexists = ASSOCIATED(matrix%m%wms)
      IF (preexists) THEN
         old_nwms = SIZE(matrix%m%wms)
         IF (old_nwms .EQ. 0) &
            CALL cp__w("dbcsr/block/dbcsr_work_operations.F",1211,"Nonexisting work matrices?!")
      ELSE
         old_nwms = 0
      ENDIF
      nthreads = 1; ithread = 0
!$    nthreads = OMP_GET_NUM_THREADS(); ithread = OMP_GET_THREAD_NUM()
      IF (nthreads .GT. 1) THEN
         CALL dbcsr_assert(old_nwms .EQ. nthreads, "OR", old_nwms .EQ. 0, &
                           dbcsr_fatal_level, dbcsr_caller_error, routineN, &
                           "Number of work matrices and threads do not match", &
                           1221)
      ENDIF
      nwms = MAX(1, old_nwms)
!$    nwms = MAX(nwms, nthreads)
!$OMP     BARRIER
!$OMP     MASTER
      IF (.NOT. ASSOCIATED(matrix%m%wms)) THEN
         CALL dbcsr_work_create(matrix, &
                                INT(matrix%m%nblks*default_resize_factor/nwms), &
                                INT(matrix%m%nze*default_resize_factor/nwms), &
                                n=nwms, work_mutable=.FALSE.)
      ENDIF
!$OMP     END MASTER
!$OMP     BARRIER
      size_used = matrix%m%nze
      CALL dbcsr_fill_wm_from_matrix(matrix%m%wms, matrix, size_used, &
                                     limits=limits)
!$OMP     BARRIER
      CALL timestop(handle)
   END SUBROUTINE dbcsr_add_wm_from_matrix

! **************************************************************************************************
!> \brief Fills index and data of the work matrix from the
!>        previously-finalized one.
!> \param[in,out] wm          the work matrix to fill
!> \param[in,out] matrix      DBCSR matrix
!> \param size_used ...
!> \param[in] limits          (optional) only fills blocks within this range
!> \par limits
!> The limits is a 4-tuple
!> (lower_row, higher_row, lower_column, higher_column).
! **************************************************************************************************
   SUBROUTINE dbcsr_fill_wm_from_matrix(wm, matrix, size_used, limits)
      TYPE(dbcsr_work_type), DIMENSION(:), INTENT(INOUT) :: wm
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      INTEGER, INTENT(IN)                                :: size_used
      INTEGER, DIMENSION(4), INTENT(IN), OPTIONAL        :: limits

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_fill_wm_from_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, blk_p, col, handle, ithread, &
                                                            nthreads, nwms, nze, row, wblk_p, &
                                                            which_wm, wm_first, wm_last
      LOGICAL                                            :: careful, limit, mt, tr
      LOGICAL, SAVE                                      :: mutable
      TYPE(dbcsr_data_obj)                               :: data_block
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      nwms = SIZE(matrix%m%wms)
      mt = .FALSE.
!$    IF (nwms .GT. 1) mt = omp_get_num_threads() .GT. 1
      ithread = 0; nthreads = 1
!$    ithread = omp_get_thread_num()
!$    nthreads = omp_get_num_threads()
      limit = PRESENT(limits)
      careful = size_used+size_used/8 &
                .LT. dbcsr_data_get_size_referenced(matrix%m%data_area)
      CALL dbcsr_data_init(data_block)
      CALL dbcsr_data_new(data_block, dbcsr_data_get_type(matrix%m%data_area))
      IF (mt) THEN
         wm_first = ithread+1
         wm_last = ithread+1
      ELSE
         wm_first = 1
         wm_last = nwms
      ENDIF
      ! Prepares the work matrices to accept the main data.
!$OMP     MASTER
      mutable = .FALSE.
      DO which_wm = 1, nwms
         mutable = mutable .OR. dbcsr_wm_use_mutable(wm(which_wm))
      ENDDO
!$OMP     END MASTER
!$OMP     BARRIER
      DO which_wm = wm_first, wm_last
         CALL dbcsr_assert("NOT", dbcsr_wm_use_mutable(wm(which_wm)), &
                           dbcsr_fatal_level, dbcsr_unimplemented_error_nr, routineN, &
                           "Adding main matrix into mutable not supported.", 1302)
         IF (mutable) THEN
            IF (.NOT. dbcsr_mutable_instantiated(wm(which_wm)%mutable)) THEN
               CALL dbcsr_mutable_new(wm(which_wm)%mutable, matrix%m%data_type)
            ENDIF
         ELSE
            ! We don't know how much data we'll get so we have to be generous.
            CALL dbcsr_data_ensure_size(wm(which_wm)%data_area, &
                                        size_used/nwms)
            CALL dbcsr_data_set_size_referenced(wm(which_wm)%data_area, 0)
         ENDIF
      ENDDO
      ! Now copy the data
      CALL dbcsr_iterator_start(iter, matrix, shared=mt, &
                                contiguous_pointers=.FALSE., read_only=.TRUE.)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, data_block, &
                                        transposed=tr, block_number=blk)
         IF (limit) THEN
            IF (.NOT. within_limits(row, col, limits)) CYCLE
         ENDIF
         blk_p = matrix%m%blk_p(blk)
         which_wm = ithread+1
         wblk_p = SIGN(wm(which_wm)%datasize+1, blk_p)
         nze = dbcsr_data_get_size(data_block)
         IF (mt .OR. limit .OR. careful .OR. mutable) THEN
            ! The data gets copied block by block so the block pointers
            ! are ordered accordingly.
            IF (.NOT. mutable) THEN
               CALL add_work_coordinate(wm(which_wm), row, col, wblk_p)
               CALL dbcsr_data_ensure_size(wm(which_wm)%data_area, &
                                           ABS(wblk_p)+nze-1, factor=default_resize_factor)
               CALL dbcsr_data_set_size_referenced(wm(which_wm)%data_area, &
                                                   ABS(wblk_p)+nze-1)
               CALL dbcsr_data_set(wm(which_wm)%data_area, &
                                   lb=ABS(wblk_p), &
                                   data_size=nze, &
                                   src=data_block, source_lb=1)
            ENDIF
         ELSE
            ! The data gets copied all at once so the block pointers
            ! should remain the same as they were.
            CALL add_work_coordinate(wm(which_wm), row, col, blk_p)
         ENDIF
         IF (.NOT. mutable) &
            wm(which_wm)%datasize = wm(which_wm)%datasize+nze
      ENDDO
      CALL dbcsr_iterator_stop(iter)
      CALL dbcsr_data_clear_pointer(data_block)
      CALL dbcsr_data_release(data_block)
      ! Copy all blocks at once
      IF (.NOT. mt .AND. .NOT. limit .AND. .NOT. careful .AND. .NOT. mutable) THEN
         DO which_wm = 1, nwms
            CALL dbcsr_data_ensure_size(wm(which_wm)%data_area, &
                                        dbcsr_data_get_size_referenced(matrix%m%data_area))
            CALL dbcsr_data_copyall(wm(which_wm)%data_area, matrix%m%data_area)
            wm(which_wm)%datasize = dbcsr_data_get_size_referenced(wm(which_wm)%data_area)
         ENDDO
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_fill_wm_from_matrix

! **************************************************************************************************
!> \brief Checks whether a point is within bounds
!> \param[in] row point to check
!> \param[in] column point to check
!> \param[in] limits         limits (low_row, high_row, low_col, high_col)
!> \retval within_limits     whether the point is within the bounds
! **************************************************************************************************

   PURE FUNCTION within_limits(row, column, limits)
      INTEGER, INTENT(IN)                                :: row, column
      INTEGER, DIMENSION(4), INTENT(IN)                  :: limits
      LOGICAL                                            :: within_limits

      within_limits = row .GE. limits(1) .AND. row .LE. limits(2) .AND. &
                      column .GE. limits(3) .AND. column .LE. limits(4)
   END FUNCTION within_limits

! **************************************************************************************************
!> \brief Deallocates and destroys a work matrix.
!> \param[in,out] wm          work matrix
! **************************************************************************************************
   SUBROUTINE dbcsr_work_destroy(wm)
      TYPE(dbcsr_work_type), INTENT(INOUT)               :: wm

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_work_destroy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, handle)

      IF (ASSOCIATED(wm%row_i)) THEN
         DEALLOCATE (wm%row_i)
      ENDIF
      IF (ASSOCIATED(wm%col_i)) THEN
         DEALLOCATE (wm%col_i)
      ENDIF
      IF (ASSOCIATED(wm%blk_p)) THEN
         DEALLOCATE (wm%blk_p)
      ENDIF
      CALL dbcsr_data_release(wm%data_area)
      CALL dbcsr_mutable_destroy(wm%mutable)

      IF (careful_mod) CALL timestop(handle)

   END SUBROUTINE dbcsr_work_destroy

! **************************************************************************************************
!> \brief Deallocates and destroys a work matrix.
!> \param m ...
! **************************************************************************************************
   SUBROUTINE dbcsr_work_destroy_all(m)
      TYPE(dbcsr_type), INTENT(INOUT)                    :: m

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_work_destroy_all', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF (.NOT. ASSOCIATED(m%wms)) &
         CALL cp__w("dbcsr/block/dbcsr_work_operations.F",1430,"Want to destroy nonexisting work matrices.")
      IF (ASSOCIATED(m%wms)) THEN
         DO i = 1, SIZE(m%wms)
            CALL dbcsr_work_destroy(m%wms(i))
         ENDDO
         DEALLOCATE (m%wms)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE dbcsr_work_destroy_all

! **************************************************************************************************
!> \brief Adds a coordinate (or other data) into a work matrix's row_i and
!>        col_i arrays and returns its position.
!> \param[in,out] matrix      work matrix
!> \param[in] row row data to add
!> \param[in] col col data to add
!> \param[in] blk   (optional) block pointer to add
!> \param[out] index          (optional) saved position
!> \note  Uses the matrix%lastblk to keep track of the current position.
! **************************************************************************************************
   SUBROUTINE add_work_coordinate(matrix, row, col, blk, index)
      TYPE(dbcsr_work_type), INTENT(INOUT)               :: matrix
      INTEGER, INTENT(IN)                                :: row, col
      INTEGER, INTENT(IN), OPTIONAL                      :: blk
      INTEGER, INTENT(OUT), OPTIONAL                     :: index

      CHARACTER(len=*), PARAMETER :: routineN = 'add_work_coordinate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, handle)
      matrix%lastblk = matrix%lastblk+1
      CALL ensure_array_size(matrix%row_i, ub=matrix%lastblk, &
                             factor=default_resize_factor)
      CALL ensure_array_size(matrix%col_i, ub=matrix%lastblk, &
                             factor=default_resize_factor)
      matrix%row_i(matrix%lastblk) = row
      matrix%col_i(matrix%lastblk) = col
      IF (PRESENT(blk)) THEN
         CALL ensure_array_size(matrix%blk_p, ub=matrix%lastblk, &
                                factor=default_resize_factor)
         matrix%blk_p(matrix%lastblk) = blk
      ENDIF
      IF (PRESENT(index)) index = matrix%lastblk
      IF (careful_mod) CALL timestop(handle)
   END SUBROUTINE add_work_coordinate

! **************************************************************************************************
!> \brief Merge data from matrix and work matrices into the final matrix.
!>
!> \param[in,out] matrix      matrix to work on
!> \param old_row_p ...
!> \param old_col_i ...
!> \param old_blk_p ...
!> \param[in] sort_data       whether data will be fully sorted
! **************************************************************************************************
   SUBROUTINE dbcsr_merge_all(matrix, old_row_p, old_col_i, old_blk_p, &
                              sort_data)
      TYPE(dbcsr_type), INTENT(INOUT)                    :: matrix
      INTEGER, DIMENSION(*), INTENT(IN)                  :: old_row_p, old_col_i, old_blk_p
      LOGICAL, INTENT(IN)                                :: sort_data

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_merge_all', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: handle, my_row_count, nblks, &
                                                            new_data_size, nrows, nwms, row, t, &
                                                            which_wm, wm_datasize
      INTEGER, ALLOCATABLE, DIMENSION(:), SAVE           :: all_data_offsets, all_data_sizes, &
                                                            new_blk_p_sorted, new_blk_sizes, &
                                                            new_row_p
      INTEGER, ALLOCATABLE, DIMENSION(:), SAVE, TARGET   :: blk_d, new_blk_p, new_col_i
      INTEGER, DIMENSION(:), POINTER                     :: my_row_p
      INTEGER, DIMENSION(:), POINTER, SAVE               :: cbs, rbs
      INTEGER, SAVE                                      :: max_row_count, nblks_to_add = 0, &
                                                            new_nblks = 0, new_nze = 0
      TYPE(dbcsr_data_obj), ALLOCATABLE, DIMENSION(:), &
         SAVE                                            :: all_data_areas
      TYPE(dbcsr_work_type), POINTER                     :: wm
      TYPE(i_array_p), DIMENSION(:), POINTER, SAVE       :: all_blk_p, all_col_i, all_row_p

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      !> \par Outline

      !> Each thread has a work matrix.  These must be merged and made
      !> into a new index.  If sort_data is False, then the data areas
      !> are simply appended.  This is probably quicker but the data
      !> blocks are not in order and the data size may expand beyond
      !> what is needed.  If sort_data is True, then data blocks are
      !> sorted in order.

      IF (dbg) WRITE (*, *) routineN//" starting, sort?", sort_data
!$OMP     BARRIER
      nrows = matrix%nblkrows_total
      nwms = SIZE(matrix%wms)
!$    call dbcsr_assert(nwms, "EQ", OMP_GET_NUM_THREADS(), &
!$                      dbcsr_fatal_level, dbcsr_caller_error, routineN, &
!$                      "Number of threads does not match number of work matrices.", &
!$                      1536)
      which_wm = 1
!$    which_wm = OMP_GET_THREAD_NUM()+1
      wm => matrix%wms(which_wm)
      !
      ! Currently B-tree-based work matrices are converted to array form.
      IF (dbcsr_wm_use_mutable(wm)) THEN
         SELECT CASE (wm%mutable%m%data_type)
         CASE (dbcsr_type_real_4)
            CALL tree_to_linear_s(wm)
         CASE (dbcsr_type_real_8)
            CALL tree_to_linear_d(wm)
         CASE (dbcsr_type_complex_4)
            CALL tree_to_linear_c(wm)
         CASE (dbcsr_type_complex_8)
            CALL tree_to_linear_z(wm)
         CASE default
            CALL cp__b("dbcsr/block/dbcsr_work_operations.F",1553,"Invalid data type")
         END SELECT
      ENDIF
      !
      ! Initializations and some counts from the threads are summed.
      !
!$OMP     MASTER
      nblks_to_add = 0
      new_nblks = old_row_p(nrows+1)
      new_nze = matrix%nze
      ALLOCATE (all_row_p(nwms))
      ALLOCATE (all_col_i(nwms))
      ALLOCATE (all_blk_p(nwms))
      ALLOCATE (all_data_sizes(0:nwms))
      ALLOCATE (all_data_offsets(nwms))
      IF (sort_data) ALLOCATE (all_data_areas(0:nwms))
      IF (sort_data) THEN
         CALL dbcsr_data_init(all_data_areas(0))
         all_data_areas(0) = matrix%data_area
!$OMP        CRITICAL (crit_data)
         CALL dbcsr_data_hold(all_data_areas(0))
!$OMP        END CRITICAL (crit_data)
         !
         ! The following is valid because the data actually referenced
         ! by blocks is explicitly calculated in dbcsr_finalize.
         all_data_sizes(0) = matrix%nze
      ELSE
         all_data_sizes(0) = dbcsr_data_get_size_referenced(matrix%data_area)
      ENDIF
      rbs => array_data(matrix%row_blk_size)
      cbs => array_data(matrix%col_blk_size)
!$OMP     END MASTER
      !
!$OMP     BARRIER
      wm_datasize = wm%datasize
      nblks = wm%lastblk
!$OMP     ATOMIC
      nblks_to_add = nblks_to_add+nblks
!$OMP     ATOMIC
      new_nblks = new_nblks+nblks
!$OMP     ATOMIC
      new_nze = new_nze+wm_datasize
!$OMP     BARRIER
      !
!$OMP     MASTER
      ALLOCATE (new_row_p(nrows+1))
      ALLOCATE (new_col_i(new_nblks))
      ALLOCATE (new_blk_p(new_nblks))
      IF (sort_data) THEN
         ALLOCATE (blk_d(new_nblks))
      ELSE
         ALLOCATE (blk_d(1))
      ENDIF
!$OMP     END MASTER
      !
      ! Each thread creates a row_p index for its new blocks.  This gives the
      ! number of new blocks per thread per row.
      CALL dbcsr_sort_indices(nblks, wm%row_i, wm%col_i, wm%blk_p)
      ALLOCATE (my_row_p(nrows+1))
      CALL dbcsr_make_dbcsr_index(my_row_p, wm%row_i, nrows, nblks)
      !
      ! The threads must share their row_p, col_i, blk_p, and data areas
      ! among themselves.
      all_row_p(which_wm)%p => my_row_p
      IF (sort_data) THEN
         all_data_sizes(which_wm) = wm%datasize
      ELSE
         all_data_sizes(which_wm) = wm%datasize
         CALL dbcsr_assert(wm%datasize, "LE", dbcsr_data_get_size_referenced(wm%data_area), &
                           dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                           "Data size mismatch.", 1623)
      ENDIF
      all_col_i(which_wm)%p => wm%col_i
      all_blk_p(which_wm)%p => wm%blk_p
!$OMP     BARRIER
      IF (sort_data) THEN
!$OMP        MASTER
         all_data_offsets(:) = 0
!$OMP        END MASTER
         CALL dbcsr_data_init(all_data_areas(which_wm))
         all_data_areas(which_wm) = wm%data_area
!$OMP        CRITICAL (crit_data)
         CALL dbcsr_data_hold(all_data_areas(which_wm))
!$OMP        END CRITICAL (crit_data)
      ELSE
!$OMP        MASTER
         all_data_offsets(1) = all_data_sizes(0)
         DO t = 2, nwms
            all_data_offsets(t) = all_data_offsets(t-1)+all_data_sizes(t-1)
         ENDDO
!$OMP        END MASTER
      ENDIF
      !
      ! The new row counts are created, then the new row_p index is created.
      !
!$OMP     DO
      DO row = 1, nrows
         my_row_count = old_row_p(row+1)-old_row_p(row)
         DO t = 1, nwms
            my_row_count = my_row_count &
                 &       +all_row_p(t)%p(row+1)-all_row_p(t)%p(row)
         ENDDO
         new_row_p(row) = my_row_count
      ENDDO
!$OMP     END DO
!$OMP     MASTER
      max_row_count = MAXVAL(new_row_p(1:nrows))
      CALL dbcsr_build_row_index(new_row_p, nrows)
!$OMP     END MASTER
!$OMP     BARRIER
      !
      ! The new index is built
      CALL merge_index(new_row_p, new_col_i, new_blk_p, &
                       blk_d, &
                       old_row_p, old_col_i, old_blk_p, &
                       all_row_p, all_col_i, all_blk_p, &
                       all_data_offsets, nwms, nrows, max_row_count, sort_data)
      !
      ! The data is then merged in one of two ways.
      !
!$OMP     MASTER
      new_data_size = SUM(all_data_sizes)
      IF (sort_data) THEN
         ! The matrix gets a fresh data area.  Blocks from the work
         ! matrices will be copied into it in order.
         !
!$OMP        CRITICAL (crit_data)
         CALL dbcsr_data_release(matrix%data_area)
         CALL dbcsr_data_init(matrix%data_area)
!$OMP        END CRITICAL (crit_data)
         CALL dbcsr_data_new(matrix%data_area, &
                             data_size=new_data_size, &
                             data_type=dbcsr_data_get_type(all_data_areas(0)), &
                             memory_type=dbcsr_data_get_memory_type(all_data_areas(0)))
         ALLOCATE (new_blk_p_sorted(new_nblks))
         ALLOCATE (new_blk_sizes(new_nblks))
      ELSE
         ! Data stored in the work matrices will be just appended to the
         ! current data area.
         CALL dbcsr_data_ensure_size(matrix%data_area, new_data_size)
      ENDIF
!$OMP     END MASTER
!$OMP     BARRIER
      IF (sort_data) THEN
         CALL dbcsr_calc_block_sizes(new_blk_sizes, &
                                     new_row_p, new_col_i, rbs, cbs)
         CALL dbcsr_sort_data(new_blk_p_sorted, new_blk_p, &
                              new_blk_sizes, dsts=matrix%data_area, &
                              src=all_data_areas(0), &
                              srcs=all_data_areas, old_blk_d=blk_d)
      ELSE
         IF (all_data_sizes(which_wm) .GT. 0) THEN
            CALL dbcsr_data_copy(dst=matrix%data_area, &
                                 dst_lb=(/all_data_offsets(which_wm)+1/), &
                                 dst_sizes=(/all_data_sizes(which_wm)/), &
                                 src=wm%data_area, &
                                 src_lb=(/1/), &
                                 src_sizes=(/all_data_sizes(which_wm)/))
         ENDIF
      ENDIF
      !
      ! Creates a new index array.
      !
!$OMP     BARRIER
!$OMP     MASTER
      CALL dbcsr_clearfrom_index_array(matrix, dbcsr_slot_row_p)
      CALL dbcsr_clearfrom_index_array(matrix, dbcsr_slot_col_i)
      CALL dbcsr_clearfrom_index_array(matrix, dbcsr_slot_blk_p)
      CALL dbcsr_addto_index_array(matrix, dbcsr_slot_row_p, &
                                   DATA=new_row_p(1:nrows+1), extra=new_nblks*2)
      CALL dbcsr_addto_index_array(matrix, dbcsr_slot_col_i, &
                                   DATA=new_col_i(1:new_nblks))
      IF (sort_data) THEN
         CALL dbcsr_addto_index_array(matrix, dbcsr_slot_blk_p, &
                                      DATA=new_blk_p_sorted)
      ELSE
         CALL dbcsr_addto_index_array(matrix, dbcsr_slot_blk_p, &
                                      DATA=new_blk_p)
      ENDIF
      matrix%nblks = new_nblks
      matrix%nze = new_data_size
      matrix%index(dbcsr_slot_nblks) = matrix%nblks
      matrix%index(dbcsr_slot_nze) = matrix%nze
      CALL dbcsr_repoint_index(matrix)
!$OMP     END MASTER
      !
!$OMP     BARRIER
      !
      DEALLOCATE (my_row_p)
      IF (sort_data) THEN
!$OMP        MASTER
!$OMP        CRITICAL (crit_data)
         CALL dbcsr_data_release(all_data_areas(0))
!$OMP        END CRITICAL (crit_data)
         DO which_wm = 1, nwms
!$OMP           CRITICAL (crit_data)
            CALL dbcsr_data_release(all_data_areas(which_wm))
!$OMP           END CRITICAL (crit_data)
         ENDDO
!$OMP        END MASTER
      ENDIF
!$OMP     BARRIER
!$OMP     MASTER
      DEALLOCATE (all_row_p)
      DEALLOCATE (all_col_i)
      DEALLOCATE (all_blk_p)
      DEALLOCATE (new_row_p)
      DEALLOCATE (new_col_i)
      DEALLOCATE (new_blk_p)
      DEALLOCATE (all_data_sizes)
      DEALLOCATE (all_data_offsets)
      IF (sort_data) THEN
         DEALLOCATE (all_data_areas)
         DEALLOCATE (new_blk_sizes)
         DEALLOCATE (new_blk_p_sorted)
      ENDIF
      DEALLOCATE (blk_d)
!$OMP     END MASTER
!$OMP     BARRIER
      IF (dbg) WRITE (*, *) routineN//" stopped"
      CALL timestop(handle)
   END SUBROUTINE dbcsr_merge_all

! **************************************************************************************************
!> \brief Sort data from the WM into the final matrix, based closely on dbcsr_merge_all
!>
!> \param[in,out] matrix      matrix to work on
! **************************************************************************************************
   SUBROUTINE dbcsr_merge_single_wm(matrix)
      TYPE(dbcsr_type), INTENT(INOUT)                    :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_merge_single_wm', &
         routineP = moduleN//':'//routineN
      LOGICAL, PARAMETER                                 :: dbg = .FALSE.

      INTEGER                                            :: handle, nblks, nrows
      INTEGER, ALLOCATABLE, DIMENSION(:), SAVE           :: new_blk_p_sorted, new_blk_sizes, &
                                                            new_row_p
      INTEGER, ALLOCATABLE, DIMENSION(:), SAVE, TARGET   :: blk_d
      INTEGER, DIMENSION(:), POINTER, SAVE               :: cbs, rbs
      TYPE(dbcsr_work_type), POINTER                     :: wm

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      !> \par Outline

      !> There is a single work matrix.  Data blocks are sorted and copied
      !> into the matrix data area (which is empty).  The index is made consistent

      nrows = matrix%nblkrows_total
      wm => matrix%wms(1)
      nblks = wm%lastblk
      IF (dbcsr_wm_use_mutable(wm)) &
         CALL cp__b("dbcsr/block/dbcsr_work_operations.F",1807,"Number of threads does not match number of work matrices.")
      !
      ! The following is valid because the data actually referenced
      ! by blocks is explicitly calculated in dbcsr_finalize.
!$OMP     MASTER
      rbs => array_data(matrix%row_blk_size)
      cbs => array_data(matrix%col_blk_size)
      !
      ! Initializations
      !
      ALLOCATE (new_row_p(nrows+1))
      ALLOCATE (blk_d(nblks))
      ALLOCATE (new_blk_p_sorted(nblks))
      ALLOCATE (new_blk_sizes(nblks))
      !
      ! Master thread creates a row_p index for the (sorted) blocks.
      CALL dbcsr_sort_indices(nblks, wm%row_i, wm%col_i, wm%blk_p)
      CALL dbcsr_make_dbcsr_index(new_row_p, wm%row_i, nrows, nblks)
      !
      !
      ! The matrix data area is resized.  Blocks from the work
      ! matrices will be copied into it in order.
      !
      CALL dbcsr_data_ensure_size(matrix%data_area, wm%datasize, nocopy=.TRUE.)
!$OMP     END MASTER
!$OMP     BARRIER
      CALL dbcsr_calc_block_sizes(new_blk_sizes, &
                                  new_row_p, wm%col_i, rbs, cbs)
      CALL dbcsr_sort_data(new_blk_p_sorted, wm%blk_p, &
                           new_blk_sizes, dsts=matrix%data_area, &
                           src=wm%data_area)
      !
      ! Creates a new index array.
      !
!$OMP     BARRIER
!$OMP     MASTER
      CALL dbcsr_clearfrom_index_array(matrix, dbcsr_slot_row_p)
      CALL dbcsr_clearfrom_index_array(matrix, dbcsr_slot_col_i)
      CALL dbcsr_clearfrom_index_array(matrix, dbcsr_slot_blk_p)
      CALL dbcsr_addto_index_array(matrix, dbcsr_slot_row_p, &
                                   DATA=new_row_p(1:nrows+1), extra=nblks*2)
      CALL dbcsr_addto_index_array(matrix, dbcsr_slot_col_i, &
                                   DATA=wm%col_i(1:nblks))
      CALL dbcsr_addto_index_array(matrix, dbcsr_slot_blk_p, &
                                   DATA=new_blk_p_sorted)
      matrix%nblks = nblks
      matrix%nze = wm%datasize
      matrix%index(dbcsr_slot_nblks) = matrix%nblks
      matrix%index(dbcsr_slot_nze) = matrix%nze
      CALL dbcsr_repoint_index(matrix)
      DEALLOCATE (new_row_p)
      DEALLOCATE (new_blk_sizes)
      DEALLOCATE (new_blk_p_sorted)
      DEALLOCATE (blk_d)
!$OMP     END MASTER
      IF (dbg) WRITE (*, *) routineN//" stopped"
      CALL timestop(handle)
   END SUBROUTINE dbcsr_merge_single_wm

! **************************************************************************************************
!> \brief Builds a new index from several work matrices.
!>
!> \param new_row_p ...
!> \param new_col_i ...
!> \param new_blk_p ...
!> \param blk_d ...
!> \param old_row_p ...
!> \param old_col_i ...
!> \param old_blk_p ...
!> \param all_row_p ...
!> \param all_col_i ...
!> \param all_blk_p ...
!> \param all_data_offsets ...
!> \param nwms ...
!> \param nrows ...
!> \param max_row_count ...
!> \param sort_data ...
! **************************************************************************************************
   SUBROUTINE merge_index(new_row_p, new_col_i, new_blk_p, &
                          blk_d, old_row_p, old_col_i, old_blk_p, &
                          all_row_p, all_col_i, all_blk_p, &
                          all_data_offsets, nwms, nrows, max_row_count, sort_data)
      INTEGER, DIMENSION(*), INTENT(IN)                  :: new_row_p
      INTEGER, DIMENSION(*), INTENT(OUT), TARGET         :: new_col_i, new_blk_p
      INTEGER, DIMENSION(*), INTENT(IN), TARGET          :: blk_d
      INTEGER, DIMENSION(*), INTENT(IN)                  :: old_row_p, old_col_i, old_blk_p
      TYPE(i_array_p), DIMENSION(*), INTENT(IN)          :: all_row_p, all_col_i, all_blk_p
      INTEGER, DIMENSION(*), INTENT(IN)                  :: all_data_offsets
      INTEGER, INTENT(IN)                                :: nwms, nrows, max_row_count
      LOGICAL, INTENT(IN)                                :: sort_data

      CHARACTER(len=*), PARAMETER :: routineN = 'merge_index', routineP = moduleN//':'//routineN

      INTEGER                                            :: blk1, blk2, error_handle, first_row_blk, &
                                                            last_row_blk, row, src_blk_1, &
                                                            src_blk_2, t
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: blk_p_buff, tmp_arr
      INTEGER, DIMENSION(:), POINTER                     :: blk_span, col_span, d_span

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      !
      ALLOCATE (tmp_arr(max_row_count))
      ALLOCATE (blk_p_buff(max_row_count))
      !
!$OMP     DO
      DO row = 1, nrows
         first_row_blk = new_row_p(row)+1
         last_row_blk = new_row_p(row+1)
         col_span => new_col_i(first_row_blk:last_row_blk)
         blk_span => new_blk_p(first_row_blk:last_row_blk)
         IF (sort_data) d_span => blk_d(first_row_blk:last_row_blk)
         src_blk_1 = old_row_p(row)+1
         src_blk_2 = old_row_p(row+1)
         blk1 = 1
         blk2 = blk1+(src_blk_2-src_blk_1+1)-1
         col_span(blk1:blk2) = old_col_i(src_blk_1:src_blk_2)
         blk_span(blk1:blk2) = old_blk_p(src_blk_1:src_blk_2)
         IF (sort_data) THEN
            d_span(blk1:blk2) = 1
            DO t = 1, nwms
               src_blk_1 = all_row_p(t)%p(row)+1
               src_blk_2 = all_row_p(t)%p(row+1)
               blk1 = blk2+1
               blk2 = blk1+(src_blk_2-src_blk_1+1)-1
               col_span(blk1:blk2) = all_col_i(t)%p(src_blk_1:src_blk_2)
               blk_span(blk1:blk2) = all_blk_p(t)%p(src_blk_1:src_blk_2)
               d_span(blk1:blk2) = t+1
            ENDDO
         ELSE
            DO t = 1, nwms
               src_blk_1 = all_row_p(t)%p(row)+1
               src_blk_2 = all_row_p(t)%p(row+1)
               blk1 = blk2+1
               blk2 = blk1+(src_blk_2-src_blk_1+1)-1
               col_span(blk1:blk2) = all_col_i(t)%p(src_blk_1:src_blk_2)
               blk_span(blk1:blk2) = all_blk_p(t)%p(src_blk_1:src_blk_2) &
                    &              +all_data_offsets(t)
            ENDDO
         ENDIF
         CALL sort(col_span, SIZE(col_span), tmp_arr)
         blk_p_buff(1:SIZE(blk_span)) = blk_span(:)
         blk_span(1:SIZE(blk_span)) = blk_p_buff(tmp_arr(1:SIZE(blk_span)))
      ENDDO
!$OMP     END DO
      !
      DEALLOCATE (tmp_arr)
      DEALLOCATE (blk_p_buff)
      !
      CALL timestop(error_handle)
   END SUBROUTINE merge_index

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Converts mutable data to linear (array) type.
!>
!> \param[in,out] wm      work matrix to convert
! **************************************************************************************************
  SUBROUTINE tree_to_linear_d(wm)
    USE btree_I8_k_dp2d_v,&
        ONLY: btree_2d_data_d => dp2d,&
              btree_destroy_d => btree_delete,&
              btree_size_d => btree_get_entries
    TYPE(dbcsr_work_type), INTENT(INOUT)     :: wm

    CHARACTER(len=*), PARAMETER :: routineN = 'tree_to_linear_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk, blk_p, treesize, &
                                                error_handler, needed_size
    INTEGER(KIND=int_8), ALLOCATABLE, &
      DIMENSION(:)                           :: keys
    REAL(kind=real_8), DIMENSION(:), POINTER           :: target_data
    REAL(kind=real_8), DIMENSION(:, :), POINTER        :: block_2d
    TYPE(btree_2d_data_d), ALLOCATABLE, &
      DIMENSION(:)                           :: values

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, error_handler)
    ! srt = .TRUE. ! Not needed because of the copy
    treesize = btree_size_d(wm%mutable%m%btree_d)
    IF(wm%lastblk .NE. treesize)&
       CALL cp__b("dbcsr/block/dbcsr_work_operations.F",36,"Mismatch in number of blocks")
    ALLOCATE (keys(treesize), values(treesize))
    CALL btree_destroy_d (wm%mutable%m%btree_d, keys, values)
    CALL ensure_array_size (wm%row_i, ub=treesize)
    CALL ensure_array_size (wm%col_i, ub=treesize)
    CALL dbcsr_unpack_i8_2i4 (keys, wm%row_i,&
         wm%col_i)
    ! For now we also fill the data, sloooowly, but this should
    ! be avoided and the data should be copied directly from the
    ! source in the subroutine's main loop.
    CALL ensure_array_size (wm%blk_p, ub=treesize)
    needed_size=0
    DO blk= 1, treesize
       block_2d => values(blk)%p
       needed_size=needed_size+SIZE(block_2d)
    ENDDO
    wm%datasize=needed_size
    CALL dbcsr_data_ensure_size (wm%data_area,&
         wm%datasize)
    target_data => dbcsr_get_data_p_d (wm%data_area)
    blk_p = 1
    DO blk = 1, treesize
       block_2d => values(blk)%p
       IF (.NOT. values(blk)%tr) THEN
          wm%blk_p(blk) = blk_p
       ELSE
          wm%blk_p(blk) = -blk_p
       ENDIF
       CALL block_copy_d (target_data, block_2d,&
            SIZE (block_2d), blk_p, 1)
       blk_p = blk_p + SIZE(block_2d)
       DEALLOCATE (block_2d)
    ENDDO
    DEALLOCATE (keys, values)
    CALL dbcsr_mutable_release (wm%mutable)
    CALL timestop(error_handler)
  END SUBROUTINE tree_to_linear_d

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Converts mutable data to linear (array) type.
!>
!> \param[in,out] wm      work matrix to convert
! **************************************************************************************************
  SUBROUTINE tree_to_linear_z(wm)
    USE btree_I8_k_zp2d_v,&
        ONLY: btree_2d_data_z => zp2d,&
              btree_destroy_z => btree_delete,&
              btree_size_z => btree_get_entries
    TYPE(dbcsr_work_type), INTENT(INOUT)     :: wm

    CHARACTER(len=*), PARAMETER :: routineN = 'tree_to_linear_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk, blk_p, treesize, &
                                                error_handler, needed_size
    INTEGER(KIND=int_8), ALLOCATABLE, &
      DIMENSION(:)                           :: keys
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: target_data
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER        :: block_2d
    TYPE(btree_2d_data_z), ALLOCATABLE, &
      DIMENSION(:)                           :: values

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, error_handler)
    ! srt = .TRUE. ! Not needed because of the copy
    treesize = btree_size_z(wm%mutable%m%btree_z)
    IF(wm%lastblk .NE. treesize)&
       CALL cp__b("dbcsr/block/dbcsr_work_operations.F",36,"Mismatch in number of blocks")
    ALLOCATE (keys(treesize), values(treesize))
    CALL btree_destroy_z (wm%mutable%m%btree_z, keys, values)
    CALL ensure_array_size (wm%row_i, ub=treesize)
    CALL ensure_array_size (wm%col_i, ub=treesize)
    CALL dbcsr_unpack_i8_2i4 (keys, wm%row_i,&
         wm%col_i)
    ! For now we also fill the data, sloooowly, but this should
    ! be avoided and the data should be copied directly from the
    ! source in the subroutine's main loop.
    CALL ensure_array_size (wm%blk_p, ub=treesize)
    needed_size=0
    DO blk= 1, treesize
       block_2d => values(blk)%p
       needed_size=needed_size+SIZE(block_2d)
    ENDDO
    wm%datasize=needed_size
    CALL dbcsr_data_ensure_size (wm%data_area,&
         wm%datasize)
    target_data => dbcsr_get_data_p_z (wm%data_area)
    blk_p = 1
    DO blk = 1, treesize
       block_2d => values(blk)%p
       IF (.NOT. values(blk)%tr) THEN
          wm%blk_p(blk) = blk_p
       ELSE
          wm%blk_p(blk) = -blk_p
       ENDIF
       CALL block_copy_z (target_data, block_2d,&
            SIZE (block_2d), blk_p, 1)
       blk_p = blk_p + SIZE(block_2d)
       DEALLOCATE (block_2d)
    ENDDO
    DEALLOCATE (keys, values)
    CALL dbcsr_mutable_release (wm%mutable)
    CALL timestop(error_handler)
  END SUBROUTINE tree_to_linear_z

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Converts mutable data to linear (array) type.
!>
!> \param[in,out] wm      work matrix to convert
! **************************************************************************************************
  SUBROUTINE tree_to_linear_s(wm)
    USE btree_I8_k_sp2d_v,&
        ONLY: btree_2d_data_s => sp2d,&
              btree_destroy_s => btree_delete,&
              btree_size_s => btree_get_entries
    TYPE(dbcsr_work_type), INTENT(INOUT)     :: wm

    CHARACTER(len=*), PARAMETER :: routineN = 'tree_to_linear_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk, blk_p, treesize, &
                                                error_handler, needed_size
    INTEGER(KIND=int_8), ALLOCATABLE, &
      DIMENSION(:)                           :: keys
    REAL(kind=real_4), DIMENSION(:), POINTER           :: target_data
    REAL(kind=real_4), DIMENSION(:, :), POINTER        :: block_2d
    TYPE(btree_2d_data_s), ALLOCATABLE, &
      DIMENSION(:)                           :: values

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, error_handler)
    ! srt = .TRUE. ! Not needed because of the copy
    treesize = btree_size_s(wm%mutable%m%btree_s)
    IF(wm%lastblk .NE. treesize)&
       CALL cp__b("dbcsr/block/dbcsr_work_operations.F",36,"Mismatch in number of blocks")
    ALLOCATE (keys(treesize), values(treesize))
    CALL btree_destroy_s (wm%mutable%m%btree_s, keys, values)
    CALL ensure_array_size (wm%row_i, ub=treesize)
    CALL ensure_array_size (wm%col_i, ub=treesize)
    CALL dbcsr_unpack_i8_2i4 (keys, wm%row_i,&
         wm%col_i)
    ! For now we also fill the data, sloooowly, but this should
    ! be avoided and the data should be copied directly from the
    ! source in the subroutine's main loop.
    CALL ensure_array_size (wm%blk_p, ub=treesize)
    needed_size=0
    DO blk= 1, treesize
       block_2d => values(blk)%p
       needed_size=needed_size+SIZE(block_2d)
    ENDDO
    wm%datasize=needed_size
    CALL dbcsr_data_ensure_size (wm%data_area,&
         wm%datasize)
    target_data => dbcsr_get_data_p_s (wm%data_area)
    blk_p = 1
    DO blk = 1, treesize
       block_2d => values(blk)%p
       IF (.NOT. values(blk)%tr) THEN
          wm%blk_p(blk) = blk_p
       ELSE
          wm%blk_p(blk) = -blk_p
       ENDIF
       CALL block_copy_s (target_data, block_2d,&
            SIZE (block_2d), blk_p, 1)
       blk_p = blk_p + SIZE(block_2d)
       DEALLOCATE (block_2d)
    ENDDO
    DEALLOCATE (keys, values)
    CALL dbcsr_mutable_release (wm%mutable)
    CALL timestop(error_handler)
  END SUBROUTINE tree_to_linear_s

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Converts mutable data to linear (array) type.
!>
!> \param[in,out] wm      work matrix to convert
! **************************************************************************************************
  SUBROUTINE tree_to_linear_c(wm)
    USE btree_I8_k_cp2d_v,&
        ONLY: btree_2d_data_c => cp2d,&
              btree_destroy_c => btree_delete,&
              btree_size_c => btree_get_entries
    TYPE(dbcsr_work_type), INTENT(INOUT)     :: wm

    CHARACTER(len=*), PARAMETER :: routineN = 'tree_to_linear_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk, blk_p, treesize, &
                                                error_handler, needed_size
    INTEGER(KIND=int_8), ALLOCATABLE, &
      DIMENSION(:)                           :: keys
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: target_data
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER        :: block_2d
    TYPE(btree_2d_data_c), ALLOCATABLE, &
      DIMENSION(:)                           :: values

!   ---------------------------------------------------------------------------

    CALL timeset(routineN, error_handler)
    ! srt = .TRUE. ! Not needed because of the copy
    treesize = btree_size_c(wm%mutable%m%btree_c)
    IF(wm%lastblk .NE. treesize)&
       CALL cp__b("dbcsr/block/dbcsr_work_operations.F",36,"Mismatch in number of blocks")
    ALLOCATE (keys(treesize), values(treesize))
    CALL btree_destroy_c (wm%mutable%m%btree_c, keys, values)
    CALL ensure_array_size (wm%row_i, ub=treesize)
    CALL ensure_array_size (wm%col_i, ub=treesize)
    CALL dbcsr_unpack_i8_2i4 (keys, wm%row_i,&
         wm%col_i)
    ! For now we also fill the data, sloooowly, but this should
    ! be avoided and the data should be copied directly from the
    ! source in the subroutine's main loop.
    CALL ensure_array_size (wm%blk_p, ub=treesize)
    needed_size=0
    DO blk= 1, treesize
       block_2d => values(blk)%p
       needed_size=needed_size+SIZE(block_2d)
    ENDDO
    wm%datasize=needed_size
    CALL dbcsr_data_ensure_size (wm%data_area,&
         wm%datasize)
    target_data => dbcsr_get_data_p_c (wm%data_area)
    blk_p = 1
    DO blk = 1, treesize
       block_2d => values(blk)%p
       IF (.NOT. values(blk)%tr) THEN
          wm%blk_p(blk) = blk_p
       ELSE
          wm%blk_p(blk) = -blk_p
       ENDIF
       CALL block_copy_c (target_data, block_2d,&
            SIZE (block_2d), blk_p, 1)
       blk_p = blk_p + SIZE(block_2d)
       DEALLOCATE (block_2d)
    ENDDO
    DEALLOCATE (keys, values)
    CALL dbcsr_mutable_release (wm%mutable)
    CALL timestop(error_handler)
  END SUBROUTINE tree_to_linear_c


END MODULE dbcsr_work_operations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   DBCSR block access
!> \author  Urban Borstnik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!  - 2010-02-18 Moved from dbcsr_operations and dbcsr_methods
!  - 2010-04-22 Added block buffer operations
! **************************************************************************************************
MODULE dbcsr_block_access
   USE array_types,                     ONLY: array_data
   USE btree_I8_k_cp2d_v,               ONLY: btree_2d_data_c => cp2d,&
                                              btree_add_c => btree_add,&
                                              btree_get_c => btree_find
   USE btree_I8_k_dp2d_v,               ONLY: btree_2d_data_d => dp2d,&
                                              btree_add_d => btree_add,&
                                              btree_get_d => btree_find
   USE btree_I8_k_sp2d_v,               ONLY: btree_2d_data_s => sp2d,&
                                              btree_add_s => btree_add,&
                                              btree_get_s => btree_find
   USE btree_I8_k_zp2d_v,               ONLY: btree_2d_data_z => zp2d,&
                                              btree_add_z => btree_add,&
                                              btree_get_z => btree_find
   USE dbcsr_block_operations,          ONLY: dbcsr_data_clear,&
                                              dbcsr_data_set
   USE dbcsr_config,                    ONLY: default_resize_factor
   USE dbcsr_data_methods,              ONLY: dbcsr_data_clear_pointer,&
                                              dbcsr_data_ensure_size,&
                                              dbcsr_data_get_size_referenced,&
                                              dbcsr_data_set_pointer,&
                                              dbcsr_get_data,&
                                              dbcsr_get_data_p
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_local_cols,&
                                              dbcsr_distribution_local_rows,&
                                              dbcsr_distribution_mp
   USE dbcsr_dist_operations,           ONLY: dbcsr_get_block_index,&
                                              dbcsr_get_stored_block_info,&
                                              dbcsr_get_stored_coordinates
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_caller_error,&
                                              dbcsr_failure_level,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_index_operations,          ONLY: dbcsr_addto_index_array,&
                                              dbcsr_clearfrom_index_array,&
                                              dbcsr_expand_row_index,&
                                              dbcsr_make_dbcsr_index,&
                                              dbcsr_sort_indices,&
                                              merge_index_arrays
   USE dbcsr_methods,                   ONLY: &
        dbcsr_blk_column_size, dbcsr_blk_row_size, dbcsr_distribution, dbcsr_get_data_type, &
        dbcsr_get_num_blocks, dbcsr_mutable_instantiated, dbcsr_mutable_new, dbcsr_nblkrows_total, &
        dbcsr_use_mutable, dbcsr_wm_use_mutable
   USE dbcsr_mp_methods,                ONLY: dbcsr_mp_mynode
   USE dbcsr_ptr_util,                  ONLY: pointer_rank_remap2,&
                                              pointer_view
   USE dbcsr_toollib,                   ONLY: make_coordinate_tuple,&
                                              swap
   USE dbcsr_types,                     ONLY: &
        dbcsr_data_obj, dbcsr_obj, dbcsr_scalar_type, dbcsr_slot_blk_p, dbcsr_slot_col_i, &
        dbcsr_slot_nblks, dbcsr_slot_nze, dbcsr_type_complex_4, dbcsr_type_complex_4_2d, &
        dbcsr_type_complex_8, dbcsr_type_complex_8_2d, dbcsr_type_real_4, dbcsr_type_real_4_2d, &
        dbcsr_type_real_8, dbcsr_type_real_8_2d
   USE dbcsr_work_operations,           ONLY: add_work_coordinate,&
                                              dbcsr_work_create
   USE kinds,                           ONLY: dp,&
                                              int_8,&
                                              real_4,&
                                              real_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_block_access'

   PUBLIC :: dbcsr_get_block_p, &
             dbcsr_put_block, dbcsr_remove_block

   PUBLIC :: dbcsr_reserve_block2d, &
             dbcsr_reserve_blocks, dbcsr_reserve_all_blocks, dbcsr_reserve_diag_blocks

   INTERFACE dbcsr_get_block_p
      MODULE PROCEDURE dbcsr_get_block_p_d, dbcsr_get_block_p_s, &
         dbcsr_get_block_p_z, dbcsr_get_block_p_c
      MODULE PROCEDURE dbcsr_get_2d_block_p_d, dbcsr_get_2d_block_p_s, &
         dbcsr_get_2d_block_p_z, dbcsr_get_2d_block_p_c
      MODULE PROCEDURE dbcsr_get_block_p_area
   END INTERFACE

   INTERFACE dbcsr_put_block
      MODULE PROCEDURE dbcsr_put_block_area
      MODULE PROCEDURE dbcsr_put_block_d, dbcsr_put_block_s, &
         dbcsr_put_block_z, dbcsr_put_block_c
      MODULE PROCEDURE dbcsr_put_block2d_d, dbcsr_put_block2d_s, &
         dbcsr_put_block2d_z, dbcsr_put_block2d_c
   END INTERFACE

   INTERFACE dbcsr_reserve_block2d
      MODULE PROCEDURE dbcsr_reserve_block2d_s, dbcsr_reserve_block2d_d, &
         dbcsr_reserve_block2d_c, dbcsr_reserve_block2d_z
   END INTERFACE

   INTERFACE dbcsr_set_block_pointer
      MODULE PROCEDURE dbcsr_set_block_pointer_any
      MODULE PROCEDURE dbcsr_set_block_pointer_2d_s, &
         dbcsr_set_block_pointer_2d_d, &
         dbcsr_set_block_pointer_2d_c, &
         dbcsr_set_block_pointer_2d_z
   END INTERFACE

   LOGICAL, PARAMETER :: careful_mod = .FALSE.
   LOGICAL, PARAMETER :: debug_mod = .FALSE.

   INTEGER, PARAMETER, PRIVATE :: rpslot_owner = 1
   INTEGER, PARAMETER, PRIVATE :: rpslot_addblks = 2
   INTEGER, PARAMETER, PRIVATE :: rpslot_addoffset = 3
   INTEGER, PARAMETER, PRIVATE :: rpslot_oldblks = 4
   INTEGER, PARAMETER, PRIVATE :: rpslot_oldoffset = 5
   INTEGER, PARAMETER, PRIVATE :: rpslot_totaloffset = 6
   INTEGER, PARAMETER, PRIVATE :: rpnslots = 6

   LOGICAL, PARAMETER, PRIVATE :: detailed_timing = .FALSE.

   TYPE block_parameters
      LOGICAL :: tr
      INTEGER :: logical_rows, logical_cols
      INTEGER :: offset, nze
   END TYPE block_parameters

   TYPE dgemm_join
      INTEGER :: p_a, p_b, p_c
      INTEGER :: last_k, last_n
      TYPE(dbcsr_scalar_type) :: alpha, beta
   END TYPE dgemm_join

CONTAINS

! **************************************************************************************************
!> \brief Marks a block for removal from a DBCSR matrix. Handles
!>        symmetric matrices.
!> \param[in]  matrix         DBCSR matrix
!> \param[in]  row            row of block to remove
!> \param[in]  col            column of block to remove
!> \param block_nze ...
!> \param[in]  block_number   (optional) the block number, if it is known
! **************************************************************************************************
   SUBROUTINE dbcsr_remove_block(matrix, row, col, block_nze, block_number)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      INTEGER, INTENT(IN)                                :: row, col, block_nze
      INTEGER, INTENT(IN), OPTIONAL                      :: block_number

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_remove_block', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: b, c, error_handle, r
      LOGICAL                                            :: found, tr

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, error_handle)
      IF (PRESENT(block_number)) THEN
         b = block_number
         IF (block_number .GT. matrix%m%nblks) &
            CALL cp__b("dbcsr/block/dbcsr_block_access.F",175,"Block number too big.")
         found = .TRUE.
      ELSE
         CALL dbcsr_get_block_index(matrix, row, col, r, c, tr, found, b)
      ENDIF
      b = ABS(b)
      IF (found .AND. b .GT. 0) THEN
         ! Mark the block for deletion.
         matrix%m%blk_p(b) = 0
         matrix%m%valid = .FALSE.
         ! update nze accordingly
         matrix%m%nze = matrix%m%nze-block_nze
         IF (debug_mod) THEN
            CALL dbcsr_assert(matrix%m%nze, "GE", 0, dbcsr_failure_level, &
                              dbcsr_caller_error, routineN, "nze < 0!", 189)
         ENDIF
      ELSE
         IF (debug_mod) THEN
            IF (b .EQ. 0) &
               CALL cp__w("dbcsr/block/dbcsr_block_access.F",194,"Block does not exist or is already deleted.")
         ENDIF
      ENDIF
      IF (careful_mod) CALL timestop(error_handle)
   END SUBROUTINE dbcsr_remove_block

! **************************************************************************************************
!> \brief Gets a block from a dbcsr matrix as a data area
!> \param[in]  matrix DBCSR matrix
!> \param[in]  row    the row
!> \param[in]  col    the column
!> \param[out] block  the block to get
!> \param[in] tr      whether the data is transposed
!> \param[out] found  whether the block exists in the matrix
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
!> \par Data area
!>      The pointer encapsulated in the data area points to data stored in the
!>      matrix. It must be 2-dimensional.
! **************************************************************************************************
   SUBROUTINE dbcsr_get_block_p_area(matrix, row, col, block, tr, found, &
                                     row_size, col_size)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, INTENT(IN)                                :: row, col
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: block
      LOGICAL, INTENT(OUT)                               :: tr, found
      INTEGER, INTENT(OUT), OPTIONAL                     :: row_size, col_size

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_block_p_area', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, csize, error_handle, iw, offset, &
                                                            rsize, stored_col, stored_row
      LOGICAL                                            :: stored_tr
      TYPE(btree_2d_data_c)                              :: data_block_c
      TYPE(btree_2d_data_d)                              :: data_block_d
      TYPE(btree_2d_data_s)                              :: data_block_s
      TYPE(btree_2d_data_z)                              :: data_block_z

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, error_handle)
      CALL dbcsr_get_block_index(matrix, row, col, stored_row, stored_col, &
                                 stored_tr, found, blk, offset)

      tr = stored_tr

      rsize = dbcsr_blk_row_size(matrix%m, stored_row)
      csize = dbcsr_blk_column_size(matrix%m, stored_col)
      IF (PRESENT(row_size)) row_size = rsize
      IF (PRESENT(col_size)) col_size = csize

      CALL dbcsr_data_clear_pointer(block)
      IF (found) THEN
         CALL dbcsr_set_block_pointer(matrix, block, rsize, csize, stored_tr, offset)
      ELSEIF (ASSOCIATED(matrix%m%wms)) THEN
         iw = 1
!$       iw = omp_get_thread_num()+1
         IF (.NOT. dbcsr_use_mutable(matrix%m)) &
            CALL cp__b("dbcsr/block/dbcsr_block_access.F",253,"Can not retrieve blocks from non-mutable work matrices.")
         IF (dbcsr_mutable_instantiated(matrix%m%wms(iw)%mutable)) THEN
            SELECT CASE (block%d%data_type)
            CASE (dbcsr_type_real_4_2d)
               CALL btree_get_s( &
                  matrix%m%wms(iw)%mutable%m%btree_s, &
                  make_coordinate_tuple(stored_row, stored_col), &
                  data_block_s, found)
               IF (found) THEN
                  CALL dbcsr_data_set_pointer(block, data_block_s%p)
               ENDIF
            CASE (dbcsr_type_real_8_2d)
               CALL btree_get_d( &
                  matrix%m%wms(iw)%mutable%m%btree_d, &
                  make_coordinate_tuple(stored_row, stored_col), &
                  data_block_d, found)
               IF (found) THEN
                  CALL dbcsr_data_set_pointer(block, data_block_d%p)
               ENDIF
            CASE (dbcsr_type_complex_4_2d)
               CALL btree_get_c( &
                  matrix%m%wms(iw)%mutable%m%btree_c, &
                  make_coordinate_tuple(stored_row, stored_col), &
                  data_block_c, found)
               IF (found) THEN
                  CALL dbcsr_data_set_pointer(block, data_block_c%p)
               ENDIF
            CASE (dbcsr_type_complex_8_2d)
               CALL btree_get_z( &
                  matrix%m%wms(iw)%mutable%m%btree_z, &
                  make_coordinate_tuple(stored_row, stored_col), &
                  data_block_z, found)
               IF (found) THEN
                  CALL dbcsr_data_set_pointer(block, data_block_z%p)
               ENDIF
            CASE default
               CALL cp__b("dbcsr/block/dbcsr_block_access.F",289,"Only 2-D data for block pointers!")
            END SELECT
         ENDIF
      ENDIF
      IF (careful_mod) CALL timestop(error_handle)
   END SUBROUTINE dbcsr_get_block_p_area

! **************************************************************************************************
!> \brief
!>          We allow :
!>                  matrix(dp) [+]= [scale(dp)] * block(dp)
!>                  matrix(dp) [+]= [scale(dp)] * block(sp)
!>                  matrix(sp) [+]= [scale(dp)] * block(sp)
!> \param matrix ...
!> \param row ...
!> \param col ...
!> \param block ...
!> \param lb_row_col ...
!> \param transposed ...
!> \param summation ...
!> \param flop ...
!> \param scale ...
!> \param[in]
!> \param[out]
!>
! **************************************************************************************************
   SUBROUTINE dbcsr_put_block_area(matrix, row, col, block, lb_row_col, transposed, &
                                   summation, flop, scale)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      INTEGER, INTENT(IN)                                :: row, col
      TYPE(dbcsr_data_obj)                               :: block
      INTEGER, DIMENSION(2), INTENT(INOUT), OPTIONAL     :: lb_row_col
      LOGICAL, INTENT(IN), OPTIONAL                      :: transposed, summation
      INTEGER(KIND=int_8), INTENT(INOUT), OPTIONAL       :: flop
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_put_block_area', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: data_type_m, error_handle
      LOGICAL                                            :: do_scale

!   ---------------------------------------------------------------------------

      IF (careful_mod) CALL timeset(routineN, error_handle)
      data_type_m = dbcsr_get_data_type(matrix)
      do_scale = PRESENT(scale)
      IF (do_scale) THEN
         !CALL dbcsr_assert (data_type_m .EQ. scale%data_type, dbcsr_fatal_level,&
         !     dbcsr_wrong_args_error, routineN, "Incompatible data types matrix="//&
         !     data_type_m//" scale="//scale%data_type)
      ENDIF
      IF (.NOT. ASSOCIATED(block%d)) &
         CALL cp__b("dbcsr/block/dbcsr_block_access.F",342,"Can only add valid data block!")
      SELECT CASE (block%d%data_type)
      CASE (dbcsr_type_real_4)
         IF (do_scale) THEN
            IF (data_type_m .EQ. dbcsr_type_real_4) THEN
               CALL dbcsr_put_block(matrix, row, col, block%d%r_sp, lb_row_col, transposed, &
                                    summation, flop, scale=scale%r_sp)
            ELSEIF (data_type_m .EQ. dbcsr_type_real_8) THEN
               CALL dbcsr_put_block(matrix, row, col, REAL(block%d%r_sp, real_8), lb_row_col, transposed, &
                                    summation, flop, scale=REAL(scale%r_sp, real_8))
            ENDIF
         ELSE
            IF (data_type_m .EQ. dbcsr_type_real_4) THEN
               CALL dbcsr_put_block(matrix, row, col, block%d%r_sp, lb_row_col, transposed, &
                                    summation, flop)
            ELSEIF (data_type_m .EQ. dbcsr_type_real_8) THEN
               CALL dbcsr_put_block(matrix, row, col, REAL(block%d%r_sp, real_8), lb_row_col, transposed, &
                                    summation, flop)
            ENDIF
         ENDIF
      CASE (dbcsr_type_real_8)
         IF (do_scale) THEN
            CALL dbcsr_put_block(matrix, row, col, block%d%r_dp, lb_row_col, transposed, &
                                 summation, flop, scale=scale%r_dp)
         ELSE
            CALL dbcsr_put_block(matrix, row, col, block%d%r_dp, lb_row_col, transposed, &
                                 summation, flop)
         ENDIF
      CASE (dbcsr_type_complex_4)
         IF (do_scale) THEN
            CALL dbcsr_put_block(matrix, row, col, block%d%c_sp, lb_row_col, transposed, &
                                 summation, flop, scale=scale%c_sp)
         ELSE
            CALL dbcsr_put_block(matrix, row, col, block%d%c_sp, lb_row_col, transposed, &
                                 summation, flop)
         ENDIF
      CASE (dbcsr_type_complex_8)
         IF (do_scale) THEN
            CALL dbcsr_put_block(matrix, row, col, block%d%c_dp, lb_row_col, transposed, &
                                 summation, flop, scale=scale%c_dp)
         ELSE
            CALL dbcsr_put_block(matrix, row, col, block%d%c_dp, lb_row_col, transposed, &
                                 summation, flop)
         ENDIF
      CASE (dbcsr_type_real_4_2d)
         IF (do_scale) THEN
            CALL dbcsr_put_block(matrix, row, col, block%d%r2_sp, lb_row_col, transposed, &
                                 summation, flop, scale=scale%r_sp)
         ELSE
            CALL dbcsr_put_block(matrix, row, col, block%d%r2_sp, lb_row_col, transposed, &
                                 summation, flop)
         ENDIF
      CASE (dbcsr_type_real_8_2d)
         IF (do_scale) THEN
            CALL dbcsr_put_block(matrix, row, col, block%d%r2_dp, lb_row_col, transposed, &
                                 summation, flop, scale=scale%r_dp)
         ELSE
            CALL dbcsr_put_block(matrix, row, col, block%d%r2_dp, lb_row_col, transposed, &
                                 summation, flop)
         ENDIF
      CASE (dbcsr_type_complex_4_2d)
         IF (do_scale) THEN
            CALL dbcsr_put_block(matrix, row, col, block%d%c2_sp, lb_row_col, transposed, &
                                 summation, flop, scale=scale%c_sp)
         ELSE
            CALL dbcsr_put_block(matrix, row, col, block%d%c2_sp, lb_row_col, transposed, &
                                 summation, flop)
         ENDIF
      CASE (dbcsr_type_complex_8_2d)
         IF (do_scale) THEN
            CALL dbcsr_put_block(matrix, row, col, block%d%c2_dp, lb_row_col, transposed, &
                                 summation, flop, scale=scale%c_dp)
         ELSE
            CALL dbcsr_put_block(matrix, row, col, block%d%c2_dp, lb_row_col, transposed, &
                                 summation, flop)
         ENDIF
      CASE default
         CALL cp__b("dbcsr/block/dbcsr_block_access.F",419,"Invalid data type")
      END SELECT
      IF (careful_mod) CALL timestop(error_handle)
   END SUBROUTINE dbcsr_put_block_area

! **************************************************************************************************
!> \brief Inserts all blocks of a dbcsr matrix to make it a full matrix.
!>        Thus obviously not linear scaling.
!> \param[in,out] matrix      Matrix into which blocks should be added.
! **************************************************************************************************
   SUBROUTINE dbcsr_reserve_all_blocks(matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_reserve_all_blocks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk_count, col, col_local, col_s, &
                                                            error_handle, myrank, rank, row, &
                                                            row_local, row_s
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: columns, rows
      INTEGER, DIMENSION(:), POINTER                     :: local_cols, local_rows
      LOGICAL                                            :: tr

      CALL timeset(routineN, error_handle)

      myrank = dbcsr_mp_mynode(dbcsr_distribution_mp(dbcsr_distribution(matrix)))
      local_rows => dbcsr_distribution_local_rows(dbcsr_distribution(matrix))
      local_cols => dbcsr_distribution_local_cols(dbcsr_distribution(matrix))

      blk_count = 0
      ! should be possible to loop only over the local blockrows/blockcols
      DO row_local = 1, SIZE(local_rows)
         DO col_local = 1, SIZE(local_cols)
            tr = .FALSE.
            row = local_rows(row_local)
            col = local_cols(col_local)
            row_s = row; col_s = col
            CALL dbcsr_get_stored_coordinates(matrix, row_s, col_s, rank)
            ! is that the correct condition for symmetric matrices ?
            IF (rank .EQ. myrank .AND. row_s .EQ. row .AND. col_s .EQ. col) blk_count = blk_count+1
         ENDDO
      ENDDO

      ALLOCATE (rows(blk_count), columns(blk_count))

      blk_count = 0
      DO row_local = 1, SIZE(local_rows)
         DO col_local = 1, SIZE(local_cols)
            tr = .FALSE.
            row = local_rows(row_local)
            col = local_cols(col_local)
            row_s = row; col_s = col
            CALL dbcsr_get_stored_coordinates(matrix, row_s, col_s, rank)
            IF (rank .EQ. myrank .AND. row_s .EQ. row .AND. col_s .EQ. col) THEN
               blk_count = blk_count+1
               rows(blk_count) = row
               columns(blk_count) = col
            ENDIF
         ENDDO
      ENDDO

      CALL dbcsr_reserve_blocks(matrix, rows, columns)

      CALL timestop(error_handle)

   END SUBROUTINE dbcsr_reserve_all_blocks

! **************************************************************************************************
!> \brief Inserts diagonal blocks of a dbcsr matrix to make it a matrix with at least all diagonal blocks present
!> \param[in,out] matrix      Matrix into which blocks should be added.
! **************************************************************************************************
   SUBROUTINE dbcsr_reserve_diag_blocks(matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_reserve_diag_blocks', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk_count, col, col_s, myrank, rank, &
                                                            row, row_s
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: columns, rows
      LOGICAL                                            :: tr

      myrank = dbcsr_mp_mynode(dbcsr_distribution_mp(dbcsr_distribution(matrix)))

      blk_count = 0
      ! should be possible to loop only over the local blockrows/blockcols
      DO row = 1, dbcsr_nblkrows_total(matrix)
         col = row
         tr = .FALSE.
         row_s = row; col_s = col
         CALL dbcsr_get_stored_coordinates(matrix, row_s, col_s, rank)
         IF (rank .EQ. myrank .AND. row_s .EQ. row .AND. col_s .EQ. col) blk_count = blk_count+1
      ENDDO

      ALLOCATE (rows(blk_count), columns(blk_count))

      blk_count = 0
      DO row = 1, dbcsr_nblkrows_total(matrix)
         col = row
         tr = .FALSE.
         row_s = row; col_s = col
         CALL dbcsr_get_stored_coordinates(matrix, row_s, col_s, rank)
         IF (rank .EQ. myrank .AND. row_s .EQ. row .AND. col_s .EQ. col) THEN
            blk_count = blk_count+1
            rows(blk_count) = row
            columns(blk_count) = col
         ENDIF
      ENDDO

      CALL dbcsr_reserve_blocks(matrix, rows, columns)

   END SUBROUTINE dbcsr_reserve_diag_blocks

! **************************************************************************************************
!> \brief Inserts block reservations into a matrix, avoiding the work matrix.
!> \param[in,out] matrix      Matrix into which blocks should be added.
!> \param[in] rows            Rows of the blocks to add
!> \param[in] columns         Columns of the blocks to add
!> \param[in] blk_pointers    (optional) block pointers to use for new blocks
!> \par Data
!>      No data can be specified; instead, space is reserved and zeroed. To
!>      add data, call dbcsr_put_block afterwards.
!> \par Reserving existing blocks
!>      Duplicates are not added, but allocations may be greater than
!>      the minimum necessary.
!> \par blk_pointers
!>      When blk_pointers is passed, the newly added blocks use these pointers.
!>      No data is cleared in this case
! **************************************************************************************************
   SUBROUTINE dbcsr_reserve_blocks(matrix, rows, columns, blk_pointers)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      INTEGER, DIMENSION(:), INTENT(IN)                  :: rows, columns
      INTEGER, DIMENSION(:), INTENT(IN), OPTIONAL        :: blk_pointers

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_reserve_blocks', &
         routineP = moduleN//':'//routineN

      INTEGER :: blk, blk_p, data_size_new, data_size_old, handle, nblkrows, nblks_actual_added, &
         nblks_added, nblks_new, nblks_old, new_data_sizes, nze
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: add_blkp, add_cols, add_rows, &
                                                            added_sizes, new_blk_p, new_col_i, &
                                                            new_row_i, old_row_i
      INTEGER, ALLOCATABLE, DIMENSION(:, :)              :: added_blk_info

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      CALL dbcsr_assert(SIZE(rows), "EQ", SIZE(columns), dbcsr_fatal_level, &
                        dbcsr_wrong_args_error, routineN, &
                        "Size of rows and columns array must match.", 568)
      IF (PRESENT(blk_pointers)) THEN
         CALL dbcsr_assert(SIZE(rows), "EQ", SIZE(blk_pointers), &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Size of rows and block pointecs arrays must match.", &
                           573)
         data_size_old = 0
      ELSE
         ! Get current data size
         data_size_old = dbcsr_data_get_size_referenced(matrix%m%data_area)
      ENDIF
      !> Ensures that the rows and columns are sorted.
      nblks_added = SIZE(rows)
      ALLOCATE (add_rows(nblks_added))
      add_rows(:) = rows(:)
      ALLOCATE (add_cols(nblks_added))
      add_cols(:) = columns(:)
      IF (PRESENT(blk_pointers)) THEN
         ALLOCATE (add_blkp(nblks_added))
         add_blkp(:) = blk_pointers(:)
         CALL dbcsr_sort_indices(nblks_added, add_rows, add_cols, &
                                 blk_p=add_blkp)
      ELSE
         CALL dbcsr_sort_indices(nblks_added, add_rows, add_cols)
      ENDIF
      nblks_old = dbcsr_get_num_blocks(matrix)
      nblkrows = dbcsr_nblkrows_total(matrix)
      CALL dbcsr_assert(SIZE(rows) .GT. 0, "IMP", &
                        nblkrows .GT. 0, dbcsr_fatal_level, &
                        dbcsr_internal_error, routineN, &
                        "Can not add blocks to matrix with no rows.", 598)
      ! Adjust the index.
      ! Get the old row indices
      ALLOCATE (old_row_i(nblks_old))
      CALL dbcsr_expand_row_index(matrix%m%row_p, old_row_i, &
                                  nblkrows, nblks_old)
      ! Calculate new block pointers. Possibly high estimates.
      new_data_sizes = 0
      blk_p = data_size_old+1 ! New blocks start at the end of the old
      ALLOCATE (added_blk_info(3, nblks_added))
      ALLOCATE (added_sizes(nblks_added))
      DO blk = 1, nblks_added
         IF (PRESENT(blk_pointers)) THEN
            blk_p = add_blkp(blk)
         ENDIF
         added_blk_info(1:3, blk) = (/add_rows(blk), add_cols(blk), blk_p/)
         nze = dbcsr_blk_row_size(matrix, add_rows(blk)) &
               *dbcsr_blk_column_size(matrix, add_cols(blk))
         added_sizes(blk) = nze
         blk_p = blk_p+nze
      ENDDO
      DEALLOCATE (add_rows)
      DEALLOCATE (add_cols)
      IF (PRESENT(blk_pointers)) DEALLOCATE (add_blkp)
      !
      nblks_new = nblks_old+nblks_added ! Possibly high estimate
      ALLOCATE (new_row_i(nblks_new))
      ALLOCATE (new_col_i(nblks_new))
      ALLOCATE (new_blk_p(nblks_new))
      ! Merge the two indices
      IF (PRESENT(blk_pointers)) THEN
         CALL merge_index_arrays(new_row_i, new_col_i, new_blk_p, nblks_new, &
                                 old_row_i, matrix%m%col_i, matrix%m%blk_p, nblks_old, &
                                 added_blk_info, nblks_added, added_nblks=nblks_actual_added)
         data_size_new = 0
      ELSE
         CALL merge_index_arrays(new_row_i, new_col_i, new_blk_p, nblks_new, &
                                 old_row_i, matrix%m%col_i, matrix%m%blk_p, nblks_old, &
                                 added_blk_info, nblks_added, added_nblks=nblks_actual_added, &
                                 added_sizes=added_sizes, added_size_offset=data_size_old+1, &
                                 added_size=data_size_new)
      ENDIF
      nblks_new = nblks_actual_added+nblks_old
      ! Free some memory
      DEALLOCATE (added_blk_info)
      DEALLOCATE (added_sizes)
      DEALLOCATE (old_row_i)
      ! We can skip this if no block was actually added.
      IF (nblks_actual_added .GT. 0) THEN
         ! Write the new index
         matrix%m%nblks = nblks_new
         matrix%m%nze = matrix%m%nze+data_size_new
         matrix%m%index(dbcsr_slot_nblks) = matrix%m%nblks
         matrix%m%index(dbcsr_slot_nze) = matrix%m%index(dbcsr_slot_nze)
         CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_col_i)
         CALL dbcsr_clearfrom_index_array(matrix%m, dbcsr_slot_blk_p)
         CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_col_i, &
                                      new_col_i(1:nblks_new), &
                                      extra=nblks_new)
         CALL dbcsr_addto_index_array(matrix%m, dbcsr_slot_blk_p, &
                                      new_blk_p(1:nblks_new))
         CALL dbcsr_make_dbcsr_index(matrix%m%row_p, new_row_i(1:nblks_new), &
                                     nblkrows, nblks_new)
         IF (.NOT. PRESENT(blk_pointers)) THEN
            ! Resize data area to fit the new blocks.
            CALL dbcsr_data_ensure_size(matrix%m%data_area, &
                                        data_size=matrix%m%nze)
            ! Zero the new data blocks.
            CALL dbcsr_data_clear(matrix%m%data_area, &
                                  lb=data_size_old+1, ub=matrix%m%nze)
         ENDIF
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_reserve_blocks

! **************************************************************************************************
!> \brief Sets a pointer, possibly using the buffers.
!> \param[in] matrix           Matrix to use
!> \param[in,out] pointer_any The pointer to set
!> \param[in] rsize Row sizes of block to point to
!> \param[in] csize Column sizes of block to point to
!> \param[in] main_tr          Whether block is transposed in the matrix
!> \param[in] base_offset      The block pointer
! **************************************************************************************************
   SUBROUTINE dbcsr_set_block_pointer_any(matrix, pointer_any, &
                                          rsize, csize, main_tr, base_offset)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: pointer_any
      INTEGER, INTENT(IN)                                :: rsize, csize
      LOGICAL, INTENT(IN)                                :: main_tr
      INTEGER, INTENT(IN)                                :: base_offset

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_block_pointer_any', &
         routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      IF (main_tr) THEN
         CALL dbcsr_data_set_pointer(pointer_any, csize, rsize, &
                                     matrix%m%data_area, source_lb=base_offset)
      ELSE
         CALL dbcsr_data_set_pointer(pointer_any, rsize, csize, &
                                     matrix%m%data_area, source_lb=base_offset)
      ENDIF
   END SUBROUTINE dbcsr_set_block_pointer_any

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Gets a 2-d block from a dbcsr matrix
!> \param[in]  matrix DBCSR matrix
!> \param[in]  row    the row
!> \param[in]  col    the column
!> \param[out] block  the block to get (rank-2 array)
!> \param[out] tr     whether the data is transposed
!> \param[out] found  whether the block exists in the matrix
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
! **************************************************************************************************
  SUBROUTINE dbcsr_get_2d_block_p_d(matrix,row,col,block,tr,found,&
       row_size, col_size)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_8), DIMENSION(:,:), POINTER         :: block
    LOGICAL, INTENT(OUT)                     :: tr
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_2d_block_p_d', &
      routineP = moduleN//':'//routineN

    REAL(kind=real_8), DIMENSION(:), POINTER           :: block_1d
    INTEGER                                  :: rsize, csize,&
                                                blk, nze, offset,&
                                                stored_row,&
                                                stored_col, iw, nwms
    INTEGER                                  :: error_handle
    TYPE(btree_2d_data_d)          :: data_block
    LOGICAL                                  :: stored_tr
    REAL(kind=real_8), DIMENSION(1,1), TARGET, SAVE    :: block0
!   ---------------------------------------------------------------------------
    IF (careful_mod) CALL timeset (routineN, error_handle)
    IF (debug_mod) THEN
       CALL dbcsr_assert (matrix%m%data_type, "EQ", dbcsr_type_real_8,&
            dbcsr_fatal_level, dbcsr_caller_error,&
            routineN, "Data type mismatch for requested block.",43)
    ENDIF

    CALL dbcsr_get_block_index (matrix, row, col, stored_row, stored_col,&
         stored_tr, found, blk, offset)
    tr = stored_tr

    rsize = dbcsr_blk_row_size (matrix%m, stored_row)
    csize = dbcsr_blk_column_size (matrix%m, stored_col)
    IF (PRESENT (row_size)) row_size = rsize
    IF (PRESENT (col_size)) col_size = csize

    NULLIFY (block)
    IF(found) THEN
       nze = rsize*csize
       IF(nze.eq.0) THEN
          found = .TRUE.
          block => block0(1:0, 1:0)
       ELSE
          block_1d => pointer_view (dbcsr_get_data_p (&
               matrix%m%data_area, 0.0_real_8), offset, offset+nze-1)
          CALL dbcsr_set_block_pointer (matrix, block, rsize, csize, offset)
       ENDIF
    ELSEIF (ASSOCIATED (matrix%m%wms)) THEN
       nwms = SIZE(matrix%m%wms)
       iw = 1
!$     CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$        dbcsr_fatal_level, dbcsr_internal_error,&
!$        routineN, "Number of work matrices not equal to number of threads", 71)
!$     iw = omp_get_thread_num () + 1
       IF(.NOT.dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",74,"Can not retrieve blocks from non-mutable work matrices.")
       IF (dbcsr_use_mutable (matrix%m)) THEN
          IF (.NOT. dbcsr_mutable_instantiated(matrix%m%wms(iw)%mutable)) THEN
             CALL dbcsr_mutable_new(matrix%m%wms(iw)%mutable,&
                  dbcsr_get_data_type(matrix))
          ENDIF
          CALL btree_get_d (&
               matrix%m%wms(iw)%mutable%m%btree_d,&
               make_coordinate_tuple(stored_row, stored_col),&
               data_block, found)
          IF (found) THEN
             block => data_block%p
          ENDIF
       ENDIF
    ENDIF
    IF (careful_mod) CALL timestop (error_handle)
  END SUBROUTINE dbcsr_get_2d_block_p_d


! **************************************************************************************************
!> \brief Gets a 1-d block from a dbcsr matrix
!> \param[in]  matrix DBCSR matrix
!> \param[in]  row    the row
!> \param[in]  col    the column
!> \param[out] block  the block to get (rank-1 array)
!> \param[out] tr     whether the data is transposed
!> \param[out] found  whether the block exists in the matrix
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
! **************************************************************************************************
  SUBROUTINE dbcsr_get_block_p_d(matrix,row,col,block,tr,found,&
       row_size, col_size)
    TYPE(dbcsr_obj), INTENT(IN)              :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_8), DIMENSION(:), POINTER           :: block
    LOGICAL, INTENT(OUT)                     :: tr
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_block_p_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk, csize, &
                                                nze, offset, &
                                                rsize, stored_row,&
                                                stored_col
    LOGICAL                                  :: stored_tr

!   ---------------------------------------------------------------------------

    IF (debug_mod) THEN
       CALL dbcsr_assert (matrix%m%data_type, "EQ", dbcsr_type_real_8,&
            dbcsr_fatal_level, dbcsr_caller_error,&
            routineN, "Data type mismatch for requested block.",127)
    ENDIF

    CALL dbcsr_get_block_index (matrix, row, col, stored_row, stored_col,&
         stored_tr, found, blk, offset)
    tr = stored_tr

    rsize = dbcsr_blk_row_size (matrix%m, stored_row)
    csize = dbcsr_blk_column_size (matrix%m, stored_col)
    IF (PRESENT (row_size)) row_size = rsize
    IF (PRESENT (col_size)) col_size = csize

    NULLIFY (block)
    IF(found) THEN
       nze = rsize*csize
       !
       block => pointer_view (&
            dbcsr_get_data_p (matrix%m%data_area, 0.0_real_8), offset, offset+nze-1&
            )
    ELSEIF (ASSOCIATED (matrix%m%wms)) THEN
       IF(.NOT.dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",148,"Can not retrieve blocks from non-mutable work matrices.")
       IF(dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",150,"Can not retrieve rank-1 block pointers from mutable work matrices.")
    ENDIF
  END SUBROUTINE dbcsr_get_block_p_d


! **************************************************************************************************
!> \brief Put a 2-D block in a DBCSR matrix using the btree
!> \param[in.out] matrix      DBCSR matrix
!> \param[in]  row            the row
!> \param[in]  col            the column
!> \param[in]  block          the block to reserve; added if not NULL
!> \param[in] transposed      the block holds transposed data
!> \param[out] existed        (optional) block already existed
! **************************************************************************************************
  SUBROUTINE dbcsr_reserve_block2d_d(matrix, row, col, block,&
       transposed, existed)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_8), DIMENSION(:,:), POINTER         :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed
    LOGICAL, INTENT(OUT), OPTIONAL           :: existed

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_reserve_block2d_d', &
      routineP = moduleN//':'//routineN

    TYPE(btree_2d_data_d)          :: data_block, data_block2
    INTEGER                                  :: col_size, row_size, &
                                                stored_row, stored_col, &
                                                iw, nwms
    INTEGER, DIMENSION(:), POINTER           :: col_blk_size, row_blk_size
    LOGICAL                                  :: found, gift, tr, sym_tr
    REAL(kind=real_8), DIMENSION(:,:), POINTER         :: original_block

!   ---------------------------------------------------------------------------

    gift = ASSOCIATED (block)
    IF (gift) THEN
       original_block => block
    ELSE
       NULLIFY (original_block)
    ENDIF
    row_blk_size => array_data (matrix%m%row_blk_size)
    col_blk_size => array_data (matrix%m%col_blk_size)
    row_size = row_blk_size(row)
    col_size = col_blk_size(col)

    stored_row = row ; stored_col = col
    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    sym_tr = .FALSE.
    CALL dbcsr_get_stored_coordinates (matrix, stored_row, stored_col)
    IF (.NOT.ASSOCIATED (matrix%m%wms)) THEN
       CALL dbcsr_work_create (matrix, work_mutable=.TRUE.)
       !$OMP MASTER
       matrix%m%valid = .FALSE.
       !$OMP END MASTER
       !$OMP BARRIER
    ENDIF

    NULLIFY (data_block%p)
    IF (.NOT. gift) THEN
       ALLOCATE (data_block%p (row_size, col_size))
       block => data_block%p
    ELSE
       data_block%p => block
    ENDIF
    data_block%tr = tr

    nwms = SIZE(matrix%m%wms)
    iw = 1
!$  CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$     dbcsr_fatal_level, dbcsr_internal_error,&
!$     routineN, "Number of work matrices not equal to number of threads", &
!$     226)
!$  iw = omp_get_thread_num () + 1
    CALL btree_add_d (matrix%m%wms(iw)%mutable%m%btree_d,&
         make_coordinate_tuple(stored_row, stored_col),&
         data_block, found, data_block2)

    IF (.NOT. found) THEN
!$OMP CRITICAL (critical_reserve_block2d)
       matrix%m%valid = .FALSE.
!$OMP END CRITICAL (critical_reserve_block2d)
       matrix%m%wms(iw)%lastblk = matrix%m%wms(iw)%lastblk + 1
       matrix%m%wms(iw)%datasize = matrix%m%wms(iw)%datasize + row_size*col_size
    ELSE
       IF (.NOT. gift) THEN
          DEALLOCATE (data_block%p)
       ELSE
          DEALLOCATE (original_block)
       ENDIF
       block => data_block2%p
    ENDIF
    IF (PRESENT (existed)) existed = found
  END SUBROUTINE dbcsr_reserve_block2d_d

! **************************************************************************************************
!> \brief Put a 2-D block in a DBCSR matrix
!> \param[in.out] matrix      DBCSR matrix
!> \param[in]  row            the row
!> \param[in]  col            the column
!> \param[in]  block          the block to put
!> \param[in]  transposed     the block is transposed
!> \param[in]  summation      (optional) if block exists, then sum the new
!>                            block to the old one instead of replacing it
!> \param[in]  scale          (optional) scale the block being added
! **************************************************************************************************
  SUBROUTINE dbcsr_put_block2d_d(matrix, row, col, block, lb_row_col, transposed,&
       summation, flop, scale)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_8), DIMENSION(:,:), INTENT(IN)      :: block
    INTEGER, DIMENSION(2), OPTIONAL, INTENT(INOUT) :: lb_row_col
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed, summation
    INTEGER(KIND=int_8), INTENT(INOUT), OPTIONAL :: flop
    REAL(kind=real_8), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_put_block2d_d', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: tr, do_sum

    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    IF (PRESENT (summation)) THEN
       do_sum = summation
    ELSE
       do_sum = .FALSE.
    ENDIF
    IF (PRESENT (scale)) THEN
       CALL dbcsr_put_block (matrix, row, col,&
            RESHAPE (block, (/SIZE(block)/)), lb_row_col, tr, do_sum, flop, scale)
    ELSE
       CALL dbcsr_put_block (matrix, row, col,&
            RESHAPE (block, (/SIZE(block)/)), lb_row_col, tr, do_sum, flop)
    ENDIF
  END SUBROUTINE dbcsr_put_block2d_d

! **************************************************************************************************
!> \brief Inserts a block in a dbcsr matrix.
!>
!> If the block exists, the current data is overwritten.
!> \param[in]  matrix         DBCSR matrix
!> \param[in]  row            the logical row
!> \param[in]  col            the logical column
!> \param[in]  block          the block to put
!> \param[in]  transposed     (optional) the block is transposed
!> \param[in]  summation      (optional) if block exists, then sum the new
!>                            block to the old one instead of replacing it
!> \param[in]  scale          (optional) scale the OBblock being added
! **************************************************************************************************
  SUBROUTINE dbcsr_put_block_d(matrix, row, col, block, lb_row_col, transposed,&
       summation, flop, scale)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_8), DIMENSION(:), INTENT(IN)        :: block
    INTEGER, DIMENSION(2), OPTIONAL, INTENT(INOUT) :: lb_row_col
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed, summation
    INTEGER(KIND=int_8), INTENT(INOUT), OPTIONAL :: flop
    REAL(kind=real_8), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_put_block_d', &
      routineP = moduleN//':'//routineN

    TYPE(btree_2d_data_d)          :: data_block, data_block2
    INTEGER                                  :: blk, col_size, &
                                                nze, offset, &
                                                row_size, blk_p,&
                                                stored_row, stored_col,&
                                                iw, nwms
    LOGICAL                                  :: found, tr, do_sum, tr_diff,&
                                                sym_tr
    REAL(kind=real_8), DIMENSION(:), POINTER           :: block_1d
    INTEGER(KIND=int_8)                      :: my_flop

!   ---------------------------------------------------------------------------
    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    IF (PRESENT (summation)) THEN
       do_sum = summation
    ELSE
       do_sum = .FALSE.
    ENDIF
    my_flop = 0
    row_size = dbcsr_blk_row_size(matrix, row)
    col_size = dbcsr_blk_column_size(matrix, col)
    IF (tr) CALL swap (row_size, col_size)

    stored_row = row ; stored_col = col; sym_tr = .FALSE.
    CALL dbcsr_get_stored_coordinates (matrix%m, stored_row, stored_col)
    nze = row_size*col_size
    !
    IF (debug_mod) THEN
       CALL dbcsr_assert (SIZE(block), "GE", nze, dbcsr_fatal_level,&
            dbcsr_caller_error, routineN, "Invalid block dimensions",353)
    ENDIF
    CALL dbcsr_get_stored_block_info (matrix%m, stored_row, stored_col,&
         found, blk, lb_row_col, offset)
    IF(found) THEN
       ! let's copy the block
       offset = ABS (offset)
       ! Fix the index if the new block's transpose flag is different
       ! from the old one.
       tr_diff = .FALSE.
       IF (matrix%m%blk_p(blk).LT.0 .NEQV. tr) THEN
          tr_diff = .TRUE.
          matrix%m%blk_p(blk) = -matrix%m%blk_p(blk)
       ENDIF
       block_1d => pointer_view (dbcsr_get_data_p (&
            matrix%m%data_area, 0.0_real_8), offset, offset+nze-1)
       IF (nze .GT. 0) THEN
          IF (do_sum) THEN
             IF(tr_diff) &
                  block_1d = RESHAPE(TRANSPOSE(RESHAPE(block_1d,(/col_size,row_size/))),(/nze/))
             IF (PRESENT (scale)) THEN
                CALL daxpy (nze, scale, block(1:nze), 1,&
                     block_1d, 1)
             ELSE
                CALL daxpy (nze, 1.0_real_8, block(1:nze), 1,&
                     block_1d, 1)
             ENDIF
             my_flop = my_flop + nze * 2
          ELSE
             IF (PRESENT (scale)) THEN
                CALL dcopy (nze, scale*block(1:nze), 1,&
                     block_1d, 1)
             ELSE
                CALL dcopy (nze, block(1:nze), 1,&
                     block_1d, 1)
             ENDIF
          ENDIF
       ENDIF
    ELSE
       !!@@@
       !call cp_assert (associated (matrix%m%wms), cp_fatal_level,&
       !     cp_caller_error, routineN, "Work matrices not prepared")
       IF (.NOT.ASSOCIATED (matrix%m%wms)) THEN
          CALL dbcsr_work_create (matrix, nblks_guess=1,&
               sizedata_guess=SIZE(block))
       ENDIF
       nwms = SIZE(matrix%m%wms)
       iw = 1
!$     IF (debug_mod) THEN
!$     CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$        dbcsr_fatal_level, dbcsr_internal_error,&
!$        routineN, "Number of work matrices not equal to number of threads", 404)
!$     ENDIF
!$     iw = omp_get_thread_num () + 1
       blk_p = matrix%m%wms(iw)%datasize + 1
       IF (.NOT.dbcsr_wm_use_mutable (matrix%m%wms(iw))) THEN
          IF (tr) blk_p = -blk_p
          CALL add_work_coordinate (matrix%m%wms(iw), row, col, blk_p)
          CALL dbcsr_data_ensure_size (matrix%m%wms(iw)%data_area,&
               matrix%m%wms(iw)%datasize+SIZE(block),&
               factor=default_resize_factor)
          IF (PRESENT (scale)) THEN
             CALL dbcsr_data_set (matrix%m%wms(iw)%data_area, ABS(blk_p),&
                  data_size=SIZE(block), src=scale*block, source_lb=1)
          ELSE
             CALL dbcsr_data_set (matrix%m%wms(iw)%data_area, ABS(blk_p),&
                  data_size=SIZE(block), src=block, source_lb=1)
          ENDIF
       ELSE
          ALLOCATE (data_block%p (row_size, col_size))
          IF (PRESENT (scale)) THEN
             data_block%p(:,:) = scale*RESHAPE (block, (/row_size, col_size/))
          ELSE
             data_block%p(:,:) = RESHAPE (block, (/row_size, col_size/))
          ENDIF
          data_block%tr = tr
          IF (.NOT. dbcsr_mutable_instantiated(matrix%m%wms(iw)%mutable)) THEN
             CALL dbcsr_mutable_new(matrix%m%wms(iw)%mutable,&
                  dbcsr_get_data_type(matrix))
          ENDIF
          IF (.NOT. do_sum) THEN
             CALL btree_add_d (&
                  matrix%m%wms(iw)%mutable%m%btree_d,&
                  make_coordinate_tuple(stored_row, stored_col),&
                  data_block, found, data_block2, replace=.TRUE.)
             IF (found) THEN
                IF(.NOT.ASSOCIATED(data_block2%p))&
                   CALL cp__w("dbcsr/block/dbcsr_block_access.F",440,"Data was not present in block")
                IF (ASSOCIATED (data_block2%p)) DEALLOCATE (data_block2%p)
             ENDIF
          ELSE
             CALL btree_add_d (&
                  matrix%m%wms(iw)%mutable%m%btree_d,&
                  make_coordinate_tuple(stored_row, stored_col),&
                  data_block, found, data_block2, replace=.FALSE.)
             IF (found) THEN
                IF(nze > 0) &
                   CALL daxpy (nze, 1.0_real_8, block(1), 1,&
                        data_block2%p(1,1), 1)
                IF(.NOT.ASSOCIATED(data_block%p))&
                   CALL cp__w("dbcsr/block/dbcsr_block_access.F",453,"Data was not present in block")
                IF (ASSOCIATED (data_block%p)) DEALLOCATE (data_block%p)
             ENDIF
          ENDIF
          IF (.NOT. found) THEN
             matrix%m%wms(iw)%lastblk = matrix%m%wms(iw)%lastblk + 1
          ENDIF
       ENDIF
       IF (.NOT. found) THEN
          matrix%m%wms(iw)%datasize = matrix%m%wms(iw)%datasize + SIZE (block)
       ELSE
       ENDIF
!$OMP CRITICAL (dbcsr_put_block_critical)
       matrix%m%valid = .FALSE.
!$OMP END CRITICAL (dbcsr_put_block_critical)
    ENDIF
    IF (PRESENT(flop)) flop = flop + my_flop
  END SUBROUTINE dbcsr_put_block_d


! **************************************************************************************************
!> \brief Sets a pointer, possibly using the buffers.
!> \param[in] matrix           Matrix to use
!> \param pointer_any The pointer to set
!> \param rsize Row size of block to point to
!> \param csize Column size of block to point to
!> \param[in] base_offset      The block pointer
! **************************************************************************************************
  SUBROUTINE dbcsr_set_block_pointer_2d_d (&
       matrix, pointer_any, rsize, csize, base_offset)
    TYPE(dbcsr_obj), INTENT(IN)              :: matrix
    REAL(kind=real_8), DIMENSION(:,:), POINTER         :: pointer_any
    INTEGER, INTENT(IN)                      :: rsize, csize
    INTEGER, INTENT(IN)                      :: base_offset

    CHARACTER(len=*), PARAMETER :: &
      routineN = 'dbcsr_set_block_pointer_2d_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: error_handler
    REAL(kind=real_8), DIMENSION(:), POINTER           :: lin_blk_p

!   ---------------------------------------------------------------------------

    IF (careful_mod) CALL timeset (routineN, error_handler)
    CALL dbcsr_get_data (matrix%m%data_area, lin_blk_p,&
         lb=base_offset, ub=base_offset+rsize*csize-1)
    CALL pointer_rank_remap2 (pointer_any, rsize, csize,&
         lin_blk_p)
    IF (careful_mod) CALL timestop (error_handler)
  END SUBROUTINE dbcsr_set_block_pointer_2d_d
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Gets a 2-d block from a dbcsr matrix
!> \param[in]  matrix DBCSR matrix
!> \param[in]  row    the row
!> \param[in]  col    the column
!> \param[out] block  the block to get (rank-2 array)
!> \param[out] tr     whether the data is transposed
!> \param[out] found  whether the block exists in the matrix
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
! **************************************************************************************************
  SUBROUTINE dbcsr_get_2d_block_p_z(matrix,row,col,block,tr,found,&
       row_size, col_size)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER         :: block
    LOGICAL, INTENT(OUT)                     :: tr
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_2d_block_p_z', &
      routineP = moduleN//':'//routineN

    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: block_1d
    INTEGER                                  :: rsize, csize,&
                                                blk, nze, offset,&
                                                stored_row,&
                                                stored_col, iw, nwms
    INTEGER                                  :: error_handle
    TYPE(btree_2d_data_z)          :: data_block
    LOGICAL                                  :: stored_tr
    COMPLEX(kind=real_8), DIMENSION(1,1), TARGET, SAVE    :: block0
!   ---------------------------------------------------------------------------
    IF (careful_mod) CALL timeset (routineN, error_handle)
    IF (debug_mod) THEN
       CALL dbcsr_assert (matrix%m%data_type, "EQ", dbcsr_type_complex_8,&
            dbcsr_fatal_level, dbcsr_caller_error,&
            routineN, "Data type mismatch for requested block.",43)
    ENDIF

    CALL dbcsr_get_block_index (matrix, row, col, stored_row, stored_col,&
         stored_tr, found, blk, offset)
    tr = stored_tr

    rsize = dbcsr_blk_row_size (matrix%m, stored_row)
    csize = dbcsr_blk_column_size (matrix%m, stored_col)
    IF (PRESENT (row_size)) row_size = rsize
    IF (PRESENT (col_size)) col_size = csize

    NULLIFY (block)
    IF(found) THEN
       nze = rsize*csize
       IF(nze.eq.0) THEN
          found = .TRUE.
          block => block0(1:0, 1:0)
       ELSE
          block_1d => pointer_view (dbcsr_get_data_p (&
               matrix%m%data_area, CMPLX(0.0, 0.0, real_8)), offset, offset+nze-1)
          CALL dbcsr_set_block_pointer (matrix, block, rsize, csize, offset)
       ENDIF
    ELSEIF (ASSOCIATED (matrix%m%wms)) THEN
       nwms = SIZE(matrix%m%wms)
       iw = 1
!$     CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$        dbcsr_fatal_level, dbcsr_internal_error,&
!$        routineN, "Number of work matrices not equal to number of threads", 71)
!$     iw = omp_get_thread_num () + 1
       IF(.NOT.dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",74,"Can not retrieve blocks from non-mutable work matrices.")
       IF (dbcsr_use_mutable (matrix%m)) THEN
          IF (.NOT. dbcsr_mutable_instantiated(matrix%m%wms(iw)%mutable)) THEN
             CALL dbcsr_mutable_new(matrix%m%wms(iw)%mutable,&
                  dbcsr_get_data_type(matrix))
          ENDIF
          CALL btree_get_z (&
               matrix%m%wms(iw)%mutable%m%btree_z,&
               make_coordinate_tuple(stored_row, stored_col),&
               data_block, found)
          IF (found) THEN
             block => data_block%p
          ENDIF
       ENDIF
    ENDIF
    IF (careful_mod) CALL timestop (error_handle)
  END SUBROUTINE dbcsr_get_2d_block_p_z


! **************************************************************************************************
!> \brief Gets a 1-d block from a dbcsr matrix
!> \param[in]  matrix DBCSR matrix
!> \param[in]  row    the row
!> \param[in]  col    the column
!> \param[out] block  the block to get (rank-1 array)
!> \param[out] tr     whether the data is transposed
!> \param[out] found  whether the block exists in the matrix
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
! **************************************************************************************************
  SUBROUTINE dbcsr_get_block_p_z(matrix,row,col,block,tr,found,&
       row_size, col_size)
    TYPE(dbcsr_obj), INTENT(IN)              :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: block
    LOGICAL, INTENT(OUT)                     :: tr
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_block_p_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk, csize, &
                                                nze, offset, &
                                                rsize, stored_row,&
                                                stored_col
    LOGICAL                                  :: stored_tr

!   ---------------------------------------------------------------------------

    IF (debug_mod) THEN
       CALL dbcsr_assert (matrix%m%data_type, "EQ", dbcsr_type_complex_8,&
            dbcsr_fatal_level, dbcsr_caller_error,&
            routineN, "Data type mismatch for requested block.",127)
    ENDIF

    CALL dbcsr_get_block_index (matrix, row, col, stored_row, stored_col,&
         stored_tr, found, blk, offset)
    tr = stored_tr

    rsize = dbcsr_blk_row_size (matrix%m, stored_row)
    csize = dbcsr_blk_column_size (matrix%m, stored_col)
    IF (PRESENT (row_size)) row_size = rsize
    IF (PRESENT (col_size)) col_size = csize

    NULLIFY (block)
    IF(found) THEN
       nze = rsize*csize
       !
       block => pointer_view (&
            dbcsr_get_data_p (matrix%m%data_area, CMPLX(0.0, 0.0, real_8)), offset, offset+nze-1&
            )
    ELSEIF (ASSOCIATED (matrix%m%wms)) THEN
       IF(.NOT.dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",148,"Can not retrieve blocks from non-mutable work matrices.")
       IF(dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",150,"Can not retrieve rank-1 block pointers from mutable work matrices.")
    ENDIF
  END SUBROUTINE dbcsr_get_block_p_z


! **************************************************************************************************
!> \brief Put a 2-D block in a DBCSR matrix using the btree
!> \param[in.out] matrix      DBCSR matrix
!> \param[in]  row            the row
!> \param[in]  col            the column
!> \param[in]  block          the block to reserve; added if not NULL
!> \param[in] transposed      the block holds transposed data
!> \param[out] existed        (optional) block already existed
! **************************************************************************************************
  SUBROUTINE dbcsr_reserve_block2d_z(matrix, row, col, block,&
       transposed, existed)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER         :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed
    LOGICAL, INTENT(OUT), OPTIONAL           :: existed

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_reserve_block2d_z', &
      routineP = moduleN//':'//routineN

    TYPE(btree_2d_data_z)          :: data_block, data_block2
    INTEGER                                  :: col_size, row_size, &
                                                stored_row, stored_col, &
                                                iw, nwms
    INTEGER, DIMENSION(:), POINTER           :: col_blk_size, row_blk_size
    LOGICAL                                  :: found, gift, tr, sym_tr
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER         :: original_block

!   ---------------------------------------------------------------------------

    gift = ASSOCIATED (block)
    IF (gift) THEN
       original_block => block
    ELSE
       NULLIFY (original_block)
    ENDIF
    row_blk_size => array_data (matrix%m%row_blk_size)
    col_blk_size => array_data (matrix%m%col_blk_size)
    row_size = row_blk_size(row)
    col_size = col_blk_size(col)

    stored_row = row ; stored_col = col
    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    sym_tr = .FALSE.
    CALL dbcsr_get_stored_coordinates (matrix, stored_row, stored_col)
    IF (.NOT.ASSOCIATED (matrix%m%wms)) THEN
       CALL dbcsr_work_create (matrix, work_mutable=.TRUE.)
       !$OMP MASTER
       matrix%m%valid = .FALSE.
       !$OMP END MASTER
       !$OMP BARRIER
    ENDIF

    NULLIFY (data_block%p)
    IF (.NOT. gift) THEN
       ALLOCATE (data_block%p (row_size, col_size))
       block => data_block%p
    ELSE
       data_block%p => block
    ENDIF
    data_block%tr = tr

    nwms = SIZE(matrix%m%wms)
    iw = 1
!$  CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$     dbcsr_fatal_level, dbcsr_internal_error,&
!$     routineN, "Number of work matrices not equal to number of threads", &
!$     226)
!$  iw = omp_get_thread_num () + 1
    CALL btree_add_z (matrix%m%wms(iw)%mutable%m%btree_z,&
         make_coordinate_tuple(stored_row, stored_col),&
         data_block, found, data_block2)

    IF (.NOT. found) THEN
!$OMP CRITICAL (critical_reserve_block2d)
       matrix%m%valid = .FALSE.
!$OMP END CRITICAL (critical_reserve_block2d)
       matrix%m%wms(iw)%lastblk = matrix%m%wms(iw)%lastblk + 1
       matrix%m%wms(iw)%datasize = matrix%m%wms(iw)%datasize + row_size*col_size
    ELSE
       IF (.NOT. gift) THEN
          DEALLOCATE (data_block%p)
       ELSE
          DEALLOCATE (original_block)
       ENDIF
       block => data_block2%p
    ENDIF
    IF (PRESENT (existed)) existed = found
  END SUBROUTINE dbcsr_reserve_block2d_z

! **************************************************************************************************
!> \brief Put a 2-D block in a DBCSR matrix
!> \param[in.out] matrix      DBCSR matrix
!> \param[in]  row            the row
!> \param[in]  col            the column
!> \param[in]  block          the block to put
!> \param[in]  transposed     the block is transposed
!> \param[in]  summation      (optional) if block exists, then sum the new
!>                            block to the old one instead of replacing it
!> \param[in]  scale          (optional) scale the block being added
! **************************************************************************************************
  SUBROUTINE dbcsr_put_block2d_z(matrix, row, col, block, lb_row_col, transposed,&
       summation, flop, scale)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_8), DIMENSION(:,:), INTENT(IN)      :: block
    INTEGER, DIMENSION(2), OPTIONAL, INTENT(INOUT) :: lb_row_col
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed, summation
    INTEGER(KIND=int_8), INTENT(INOUT), OPTIONAL :: flop
    COMPLEX(kind=real_8), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_put_block2d_z', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: tr, do_sum

    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    IF (PRESENT (summation)) THEN
       do_sum = summation
    ELSE
       do_sum = .FALSE.
    ENDIF
    IF (PRESENT (scale)) THEN
       CALL dbcsr_put_block (matrix, row, col,&
            RESHAPE (block, (/SIZE(block)/)), lb_row_col, tr, do_sum, flop, scale)
    ELSE
       CALL dbcsr_put_block (matrix, row, col,&
            RESHAPE (block, (/SIZE(block)/)), lb_row_col, tr, do_sum, flop)
    ENDIF
  END SUBROUTINE dbcsr_put_block2d_z

! **************************************************************************************************
!> \brief Inserts a block in a dbcsr matrix.
!>
!> If the block exists, the current data is overwritten.
!> \param[in]  matrix         DBCSR matrix
!> \param[in]  row            the logical row
!> \param[in]  col            the logical column
!> \param[in]  block          the block to put
!> \param[in]  transposed     (optional) the block is transposed
!> \param[in]  summation      (optional) if block exists, then sum the new
!>                            block to the old one instead of replacing it
!> \param[in]  scale          (optional) scale the OBblock being added
! **************************************************************************************************
  SUBROUTINE dbcsr_put_block_z(matrix, row, col, block, lb_row_col, transposed,&
       summation, flop, scale)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN)        :: block
    INTEGER, DIMENSION(2), OPTIONAL, INTENT(INOUT) :: lb_row_col
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed, summation
    INTEGER(KIND=int_8), INTENT(INOUT), OPTIONAL :: flop
    COMPLEX(kind=real_8), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_put_block_z', &
      routineP = moduleN//':'//routineN

    TYPE(btree_2d_data_z)          :: data_block, data_block2
    INTEGER                                  :: blk, col_size, &
                                                nze, offset, &
                                                row_size, blk_p,&
                                                stored_row, stored_col,&
                                                iw, nwms
    LOGICAL                                  :: found, tr, do_sum, tr_diff,&
                                                sym_tr
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: block_1d
    INTEGER(KIND=int_8)                      :: my_flop

!   ---------------------------------------------------------------------------
    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    IF (PRESENT (summation)) THEN
       do_sum = summation
    ELSE
       do_sum = .FALSE.
    ENDIF
    my_flop = 0
    row_size = dbcsr_blk_row_size(matrix, row)
    col_size = dbcsr_blk_column_size(matrix, col)
    IF (tr) CALL swap (row_size, col_size)

    stored_row = row ; stored_col = col; sym_tr = .FALSE.
    CALL dbcsr_get_stored_coordinates (matrix%m, stored_row, stored_col)
    nze = row_size*col_size
    !
    IF (debug_mod) THEN
       CALL dbcsr_assert (SIZE(block), "GE", nze, dbcsr_fatal_level,&
            dbcsr_caller_error, routineN, "Invalid block dimensions",353)
    ENDIF
    CALL dbcsr_get_stored_block_info (matrix%m, stored_row, stored_col,&
         found, blk, lb_row_col, offset)
    IF(found) THEN
       ! let's copy the block
       offset = ABS (offset)
       ! Fix the index if the new block's transpose flag is different
       ! from the old one.
       tr_diff = .FALSE.
       IF (matrix%m%blk_p(blk).LT.0 .NEQV. tr) THEN
          tr_diff = .TRUE.
          matrix%m%blk_p(blk) = -matrix%m%blk_p(blk)
       ENDIF
       block_1d => pointer_view (dbcsr_get_data_p (&
            matrix%m%data_area, CMPLX(0.0, 0.0, real_8)), offset, offset+nze-1)
       IF (nze .GT. 0) THEN
          IF (do_sum) THEN
             IF(tr_diff) &
                  block_1d = RESHAPE(TRANSPOSE(RESHAPE(block_1d,(/col_size,row_size/))),(/nze/))
             IF (PRESENT (scale)) THEN
                CALL zaxpy (nze, scale, block(1:nze), 1,&
                     block_1d, 1)
             ELSE
                CALL zaxpy (nze, CMPLX(1.0, 0.0, real_8), block(1:nze), 1,&
                     block_1d, 1)
             ENDIF
             my_flop = my_flop + nze * 2
          ELSE
             IF (PRESENT (scale)) THEN
                CALL zcopy (nze, scale*block(1:nze), 1,&
                     block_1d, 1)
             ELSE
                CALL zcopy (nze, block(1:nze), 1,&
                     block_1d, 1)
             ENDIF
          ENDIF
       ENDIF
    ELSE
       !!@@@
       !call cp_assert (associated (matrix%m%wms), cp_fatal_level,&
       !     cp_caller_error, routineN, "Work matrices not prepared")
       IF (.NOT.ASSOCIATED (matrix%m%wms)) THEN
          CALL dbcsr_work_create (matrix, nblks_guess=1,&
               sizedata_guess=SIZE(block))
       ENDIF
       nwms = SIZE(matrix%m%wms)
       iw = 1
!$     IF (debug_mod) THEN
!$     CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$        dbcsr_fatal_level, dbcsr_internal_error,&
!$        routineN, "Number of work matrices not equal to number of threads", 404)
!$     ENDIF
!$     iw = omp_get_thread_num () + 1
       blk_p = matrix%m%wms(iw)%datasize + 1
       IF (.NOT.dbcsr_wm_use_mutable (matrix%m%wms(iw))) THEN
          IF (tr) blk_p = -blk_p
          CALL add_work_coordinate (matrix%m%wms(iw), row, col, blk_p)
          CALL dbcsr_data_ensure_size (matrix%m%wms(iw)%data_area,&
               matrix%m%wms(iw)%datasize+SIZE(block),&
               factor=default_resize_factor)
          IF (PRESENT (scale)) THEN
             CALL dbcsr_data_set (matrix%m%wms(iw)%data_area, ABS(blk_p),&
                  data_size=SIZE(block), src=scale*block, source_lb=1)
          ELSE
             CALL dbcsr_data_set (matrix%m%wms(iw)%data_area, ABS(blk_p),&
                  data_size=SIZE(block), src=block, source_lb=1)
          ENDIF
       ELSE
          ALLOCATE (data_block%p (row_size, col_size))
          IF (PRESENT (scale)) THEN
             data_block%p(:,:) = scale*RESHAPE (block, (/row_size, col_size/))
          ELSE
             data_block%p(:,:) = RESHAPE (block, (/row_size, col_size/))
          ENDIF
          data_block%tr = tr
          IF (.NOT. dbcsr_mutable_instantiated(matrix%m%wms(iw)%mutable)) THEN
             CALL dbcsr_mutable_new(matrix%m%wms(iw)%mutable,&
                  dbcsr_get_data_type(matrix))
          ENDIF
          IF (.NOT. do_sum) THEN
             CALL btree_add_z (&
                  matrix%m%wms(iw)%mutable%m%btree_z,&
                  make_coordinate_tuple(stored_row, stored_col),&
                  data_block, found, data_block2, replace=.TRUE.)
             IF (found) THEN
                IF(.NOT.ASSOCIATED(data_block2%p))&
                   CALL cp__w("dbcsr/block/dbcsr_block_access.F",440,"Data was not present in block")
                IF (ASSOCIATED (data_block2%p)) DEALLOCATE (data_block2%p)
             ENDIF
          ELSE
             CALL btree_add_z (&
                  matrix%m%wms(iw)%mutable%m%btree_z,&
                  make_coordinate_tuple(stored_row, stored_col),&
                  data_block, found, data_block2, replace=.FALSE.)
             IF (found) THEN
                IF(nze > 0) &
                   CALL zaxpy (nze, CMPLX(1.0, 0.0, real_8), block(1), 1,&
                        data_block2%p(1,1), 1)
                IF(.NOT.ASSOCIATED(data_block%p))&
                   CALL cp__w("dbcsr/block/dbcsr_block_access.F",453,"Data was not present in block")
                IF (ASSOCIATED (data_block%p)) DEALLOCATE (data_block%p)
             ENDIF
          ENDIF
          IF (.NOT. found) THEN
             matrix%m%wms(iw)%lastblk = matrix%m%wms(iw)%lastblk + 1
          ENDIF
       ENDIF
       IF (.NOT. found) THEN
          matrix%m%wms(iw)%datasize = matrix%m%wms(iw)%datasize + SIZE (block)
       ELSE
       ENDIF
!$OMP CRITICAL (dbcsr_put_block_critical)
       matrix%m%valid = .FALSE.
!$OMP END CRITICAL (dbcsr_put_block_critical)
    ENDIF
    IF (PRESENT(flop)) flop = flop + my_flop
  END SUBROUTINE dbcsr_put_block_z


! **************************************************************************************************
!> \brief Sets a pointer, possibly using the buffers.
!> \param[in] matrix           Matrix to use
!> \param pointer_any The pointer to set
!> \param rsize Row size of block to point to
!> \param csize Column size of block to point to
!> \param[in] base_offset      The block pointer
! **************************************************************************************************
  SUBROUTINE dbcsr_set_block_pointer_2d_z (&
       matrix, pointer_any, rsize, csize, base_offset)
    TYPE(dbcsr_obj), INTENT(IN)              :: matrix
    COMPLEX(kind=real_8), DIMENSION(:,:), POINTER         :: pointer_any
    INTEGER, INTENT(IN)                      :: rsize, csize
    INTEGER, INTENT(IN)                      :: base_offset

    CHARACTER(len=*), PARAMETER :: &
      routineN = 'dbcsr_set_block_pointer_2d_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: error_handler
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: lin_blk_p

!   ---------------------------------------------------------------------------

    IF (careful_mod) CALL timeset (routineN, error_handler)
    CALL dbcsr_get_data (matrix%m%data_area, lin_blk_p,&
         lb=base_offset, ub=base_offset+rsize*csize-1)
    CALL pointer_rank_remap2 (pointer_any, rsize, csize,&
         lin_blk_p)
    IF (careful_mod) CALL timestop (error_handler)
  END SUBROUTINE dbcsr_set_block_pointer_2d_z
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Gets a 2-d block from a dbcsr matrix
!> \param[in]  matrix DBCSR matrix
!> \param[in]  row    the row
!> \param[in]  col    the column
!> \param[out] block  the block to get (rank-2 array)
!> \param[out] tr     whether the data is transposed
!> \param[out] found  whether the block exists in the matrix
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
! **************************************************************************************************
  SUBROUTINE dbcsr_get_2d_block_p_s(matrix,row,col,block,tr,found,&
       row_size, col_size)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_4), DIMENSION(:,:), POINTER         :: block
    LOGICAL, INTENT(OUT)                     :: tr
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_2d_block_p_s', &
      routineP = moduleN//':'//routineN

    REAL(kind=real_4), DIMENSION(:), POINTER           :: block_1d
    INTEGER                                  :: rsize, csize,&
                                                blk, nze, offset,&
                                                stored_row,&
                                                stored_col, iw, nwms
    INTEGER                                  :: error_handle
    TYPE(btree_2d_data_s)          :: data_block
    LOGICAL                                  :: stored_tr
    REAL(kind=real_4), DIMENSION(1,1), TARGET, SAVE    :: block0
!   ---------------------------------------------------------------------------
    IF (careful_mod) CALL timeset (routineN, error_handle)
    IF (debug_mod) THEN
       CALL dbcsr_assert (matrix%m%data_type, "EQ", dbcsr_type_real_4,&
            dbcsr_fatal_level, dbcsr_caller_error,&
            routineN, "Data type mismatch for requested block.",43)
    ENDIF

    CALL dbcsr_get_block_index (matrix, row, col, stored_row, stored_col,&
         stored_tr, found, blk, offset)
    tr = stored_tr

    rsize = dbcsr_blk_row_size (matrix%m, stored_row)
    csize = dbcsr_blk_column_size (matrix%m, stored_col)
    IF (PRESENT (row_size)) row_size = rsize
    IF (PRESENT (col_size)) col_size = csize

    NULLIFY (block)
    IF(found) THEN
       nze = rsize*csize
       IF(nze.eq.0) THEN
          found = .TRUE.
          block => block0(1:0, 1:0)
       ELSE
          block_1d => pointer_view (dbcsr_get_data_p (&
               matrix%m%data_area, 0.0_real_4), offset, offset+nze-1)
          CALL dbcsr_set_block_pointer (matrix, block, rsize, csize, offset)
       ENDIF
    ELSEIF (ASSOCIATED (matrix%m%wms)) THEN
       nwms = SIZE(matrix%m%wms)
       iw = 1
!$     CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$        dbcsr_fatal_level, dbcsr_internal_error,&
!$        routineN, "Number of work matrices not equal to number of threads", 71)
!$     iw = omp_get_thread_num () + 1
       IF(.NOT.dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",74,"Can not retrieve blocks from non-mutable work matrices.")
       IF (dbcsr_use_mutable (matrix%m)) THEN
          IF (.NOT. dbcsr_mutable_instantiated(matrix%m%wms(iw)%mutable)) THEN
             CALL dbcsr_mutable_new(matrix%m%wms(iw)%mutable,&
                  dbcsr_get_data_type(matrix))
          ENDIF
          CALL btree_get_s (&
               matrix%m%wms(iw)%mutable%m%btree_s,&
               make_coordinate_tuple(stored_row, stored_col),&
               data_block, found)
          IF (found) THEN
             block => data_block%p
          ENDIF
       ENDIF
    ENDIF
    IF (careful_mod) CALL timestop (error_handle)
  END SUBROUTINE dbcsr_get_2d_block_p_s


! **************************************************************************************************
!> \brief Gets a 1-d block from a dbcsr matrix
!> \param[in]  matrix DBCSR matrix
!> \param[in]  row    the row
!> \param[in]  col    the column
!> \param[out] block  the block to get (rank-1 array)
!> \param[out] tr     whether the data is transposed
!> \param[out] found  whether the block exists in the matrix
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
! **************************************************************************************************
  SUBROUTINE dbcsr_get_block_p_s(matrix,row,col,block,tr,found,&
       row_size, col_size)
    TYPE(dbcsr_obj), INTENT(IN)              :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_4), DIMENSION(:), POINTER           :: block
    LOGICAL, INTENT(OUT)                     :: tr
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_block_p_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk, csize, &
                                                nze, offset, &
                                                rsize, stored_row,&
                                                stored_col
    LOGICAL                                  :: stored_tr

!   ---------------------------------------------------------------------------

    IF (debug_mod) THEN
       CALL dbcsr_assert (matrix%m%data_type, "EQ", dbcsr_type_real_4,&
            dbcsr_fatal_level, dbcsr_caller_error,&
            routineN, "Data type mismatch for requested block.",127)
    ENDIF

    CALL dbcsr_get_block_index (matrix, row, col, stored_row, stored_col,&
         stored_tr, found, blk, offset)
    tr = stored_tr

    rsize = dbcsr_blk_row_size (matrix%m, stored_row)
    csize = dbcsr_blk_column_size (matrix%m, stored_col)
    IF (PRESENT (row_size)) row_size = rsize
    IF (PRESENT (col_size)) col_size = csize

    NULLIFY (block)
    IF(found) THEN
       nze = rsize*csize
       !
       block => pointer_view (&
            dbcsr_get_data_p (matrix%m%data_area, 0.0_real_4), offset, offset+nze-1&
            )
    ELSEIF (ASSOCIATED (matrix%m%wms)) THEN
       IF(.NOT.dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",148,"Can not retrieve blocks from non-mutable work matrices.")
       IF(dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",150,"Can not retrieve rank-1 block pointers from mutable work matrices.")
    ENDIF
  END SUBROUTINE dbcsr_get_block_p_s


! **************************************************************************************************
!> \brief Put a 2-D block in a DBCSR matrix using the btree
!> \param[in.out] matrix      DBCSR matrix
!> \param[in]  row            the row
!> \param[in]  col            the column
!> \param[in]  block          the block to reserve; added if not NULL
!> \param[in] transposed      the block holds transposed data
!> \param[out] existed        (optional) block already existed
! **************************************************************************************************
  SUBROUTINE dbcsr_reserve_block2d_s(matrix, row, col, block,&
       transposed, existed)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_4), DIMENSION(:,:), POINTER         :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed
    LOGICAL, INTENT(OUT), OPTIONAL           :: existed

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_reserve_block2d_s', &
      routineP = moduleN//':'//routineN

    TYPE(btree_2d_data_s)          :: data_block, data_block2
    INTEGER                                  :: col_size, row_size, &
                                                stored_row, stored_col, &
                                                iw, nwms
    INTEGER, DIMENSION(:), POINTER           :: col_blk_size, row_blk_size
    LOGICAL                                  :: found, gift, tr, sym_tr
    REAL(kind=real_4), DIMENSION(:,:), POINTER         :: original_block

!   ---------------------------------------------------------------------------

    gift = ASSOCIATED (block)
    IF (gift) THEN
       original_block => block
    ELSE
       NULLIFY (original_block)
    ENDIF
    row_blk_size => array_data (matrix%m%row_blk_size)
    col_blk_size => array_data (matrix%m%col_blk_size)
    row_size = row_blk_size(row)
    col_size = col_blk_size(col)

    stored_row = row ; stored_col = col
    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    sym_tr = .FALSE.
    CALL dbcsr_get_stored_coordinates (matrix, stored_row, stored_col)
    IF (.NOT.ASSOCIATED (matrix%m%wms)) THEN
       CALL dbcsr_work_create (matrix, work_mutable=.TRUE.)
       !$OMP MASTER
       matrix%m%valid = .FALSE.
       !$OMP END MASTER
       !$OMP BARRIER
    ENDIF

    NULLIFY (data_block%p)
    IF (.NOT. gift) THEN
       ALLOCATE (data_block%p (row_size, col_size))
       block => data_block%p
    ELSE
       data_block%p => block
    ENDIF
    data_block%tr = tr

    nwms = SIZE(matrix%m%wms)
    iw = 1
!$  CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$     dbcsr_fatal_level, dbcsr_internal_error,&
!$     routineN, "Number of work matrices not equal to number of threads", &
!$     226)
!$  iw = omp_get_thread_num () + 1
    CALL btree_add_s (matrix%m%wms(iw)%mutable%m%btree_s,&
         make_coordinate_tuple(stored_row, stored_col),&
         data_block, found, data_block2)

    IF (.NOT. found) THEN
!$OMP CRITICAL (critical_reserve_block2d)
       matrix%m%valid = .FALSE.
!$OMP END CRITICAL (critical_reserve_block2d)
       matrix%m%wms(iw)%lastblk = matrix%m%wms(iw)%lastblk + 1
       matrix%m%wms(iw)%datasize = matrix%m%wms(iw)%datasize + row_size*col_size
    ELSE
       IF (.NOT. gift) THEN
          DEALLOCATE (data_block%p)
       ELSE
          DEALLOCATE (original_block)
       ENDIF
       block => data_block2%p
    ENDIF
    IF (PRESENT (existed)) existed = found
  END SUBROUTINE dbcsr_reserve_block2d_s

! **************************************************************************************************
!> \brief Put a 2-D block in a DBCSR matrix
!> \param[in.out] matrix      DBCSR matrix
!> \param[in]  row            the row
!> \param[in]  col            the column
!> \param[in]  block          the block to put
!> \param[in]  transposed     the block is transposed
!> \param[in]  summation      (optional) if block exists, then sum the new
!>                            block to the old one instead of replacing it
!> \param[in]  scale          (optional) scale the block being added
! **************************************************************************************************
  SUBROUTINE dbcsr_put_block2d_s(matrix, row, col, block, lb_row_col, transposed,&
       summation, flop, scale)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_4), DIMENSION(:,:), INTENT(IN)      :: block
    INTEGER, DIMENSION(2), OPTIONAL, INTENT(INOUT) :: lb_row_col
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed, summation
    INTEGER(KIND=int_8), INTENT(INOUT), OPTIONAL :: flop
    REAL(kind=real_4), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_put_block2d_s', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: tr, do_sum

    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    IF (PRESENT (summation)) THEN
       do_sum = summation
    ELSE
       do_sum = .FALSE.
    ENDIF
    IF (PRESENT (scale)) THEN
       CALL dbcsr_put_block (matrix, row, col,&
            RESHAPE (block, (/SIZE(block)/)), lb_row_col, tr, do_sum, flop, scale)
    ELSE
       CALL dbcsr_put_block (matrix, row, col,&
            RESHAPE (block, (/SIZE(block)/)), lb_row_col, tr, do_sum, flop)
    ENDIF
  END SUBROUTINE dbcsr_put_block2d_s

! **************************************************************************************************
!> \brief Inserts a block in a dbcsr matrix.
!>
!> If the block exists, the current data is overwritten.
!> \param[in]  matrix         DBCSR matrix
!> \param[in]  row            the logical row
!> \param[in]  col            the logical column
!> \param[in]  block          the block to put
!> \param[in]  transposed     (optional) the block is transposed
!> \param[in]  summation      (optional) if block exists, then sum the new
!>                            block to the old one instead of replacing it
!> \param[in]  scale          (optional) scale the OBblock being added
! **************************************************************************************************
  SUBROUTINE dbcsr_put_block_s(matrix, row, col, block, lb_row_col, transposed,&
       summation, flop, scale)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    REAL(kind=real_4), DIMENSION(:), INTENT(IN)        :: block
    INTEGER, DIMENSION(2), OPTIONAL, INTENT(INOUT) :: lb_row_col
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed, summation
    INTEGER(KIND=int_8), INTENT(INOUT), OPTIONAL :: flop
    REAL(kind=real_4), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_put_block_s', &
      routineP = moduleN//':'//routineN

    TYPE(btree_2d_data_s)          :: data_block, data_block2
    INTEGER                                  :: blk, col_size, &
                                                nze, offset, &
                                                row_size, blk_p,&
                                                stored_row, stored_col,&
                                                iw, nwms
    LOGICAL                                  :: found, tr, do_sum, tr_diff,&
                                                sym_tr
    REAL(kind=real_4), DIMENSION(:), POINTER           :: block_1d
    INTEGER(KIND=int_8)                      :: my_flop

!   ---------------------------------------------------------------------------
    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    IF (PRESENT (summation)) THEN
       do_sum = summation
    ELSE
       do_sum = .FALSE.
    ENDIF
    my_flop = 0
    row_size = dbcsr_blk_row_size(matrix, row)
    col_size = dbcsr_blk_column_size(matrix, col)
    IF (tr) CALL swap (row_size, col_size)

    stored_row = row ; stored_col = col; sym_tr = .FALSE.
    CALL dbcsr_get_stored_coordinates (matrix%m, stored_row, stored_col)
    nze = row_size*col_size
    !
    IF (debug_mod) THEN
       CALL dbcsr_assert (SIZE(block), "GE", nze, dbcsr_fatal_level,&
            dbcsr_caller_error, routineN, "Invalid block dimensions",353)
    ENDIF
    CALL dbcsr_get_stored_block_info (matrix%m, stored_row, stored_col,&
         found, blk, lb_row_col, offset)
    IF(found) THEN
       ! let's copy the block
       offset = ABS (offset)
       ! Fix the index if the new block's transpose flag is different
       ! from the old one.
       tr_diff = .FALSE.
       IF (matrix%m%blk_p(blk).LT.0 .NEQV. tr) THEN
          tr_diff = .TRUE.
          matrix%m%blk_p(blk) = -matrix%m%blk_p(blk)
       ENDIF
       block_1d => pointer_view (dbcsr_get_data_p (&
            matrix%m%data_area, 0.0_real_4), offset, offset+nze-1)
       IF (nze .GT. 0) THEN
          IF (do_sum) THEN
             IF(tr_diff) &
                  block_1d = RESHAPE(TRANSPOSE(RESHAPE(block_1d,(/col_size,row_size/))),(/nze/))
             IF (PRESENT (scale)) THEN
                CALL saxpy (nze, scale, block(1:nze), 1,&
                     block_1d, 1)
             ELSE
                CALL saxpy (nze, 1.0_real_4, block(1:nze), 1,&
                     block_1d, 1)
             ENDIF
             my_flop = my_flop + nze * 2
          ELSE
             IF (PRESENT (scale)) THEN
                CALL scopy (nze, scale*block(1:nze), 1,&
                     block_1d, 1)
             ELSE
                CALL scopy (nze, block(1:nze), 1,&
                     block_1d, 1)
             ENDIF
          ENDIF
       ENDIF
    ELSE
       !!@@@
       !call cp_assert (associated (matrix%m%wms), cp_fatal_level,&
       !     cp_caller_error, routineN, "Work matrices not prepared")
       IF (.NOT.ASSOCIATED (matrix%m%wms)) THEN
          CALL dbcsr_work_create (matrix, nblks_guess=1,&
               sizedata_guess=SIZE(block))
       ENDIF
       nwms = SIZE(matrix%m%wms)
       iw = 1
!$     IF (debug_mod) THEN
!$     CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$        dbcsr_fatal_level, dbcsr_internal_error,&
!$        routineN, "Number of work matrices not equal to number of threads", 404)
!$     ENDIF
!$     iw = omp_get_thread_num () + 1
       blk_p = matrix%m%wms(iw)%datasize + 1
       IF (.NOT.dbcsr_wm_use_mutable (matrix%m%wms(iw))) THEN
          IF (tr) blk_p = -blk_p
          CALL add_work_coordinate (matrix%m%wms(iw), row, col, blk_p)
          CALL dbcsr_data_ensure_size (matrix%m%wms(iw)%data_area,&
               matrix%m%wms(iw)%datasize+SIZE(block),&
               factor=default_resize_factor)
          IF (PRESENT (scale)) THEN
             CALL dbcsr_data_set (matrix%m%wms(iw)%data_area, ABS(blk_p),&
                  data_size=SIZE(block), src=scale*block, source_lb=1)
          ELSE
             CALL dbcsr_data_set (matrix%m%wms(iw)%data_area, ABS(blk_p),&
                  data_size=SIZE(block), src=block, source_lb=1)
          ENDIF
       ELSE
          ALLOCATE (data_block%p (row_size, col_size))
          IF (PRESENT (scale)) THEN
             data_block%p(:,:) = scale*RESHAPE (block, (/row_size, col_size/))
          ELSE
             data_block%p(:,:) = RESHAPE (block, (/row_size, col_size/))
          ENDIF
          data_block%tr = tr
          IF (.NOT. dbcsr_mutable_instantiated(matrix%m%wms(iw)%mutable)) THEN
             CALL dbcsr_mutable_new(matrix%m%wms(iw)%mutable,&
                  dbcsr_get_data_type(matrix))
          ENDIF
          IF (.NOT. do_sum) THEN
             CALL btree_add_s (&
                  matrix%m%wms(iw)%mutable%m%btree_s,&
                  make_coordinate_tuple(stored_row, stored_col),&
                  data_block, found, data_block2, replace=.TRUE.)
             IF (found) THEN
                IF(.NOT.ASSOCIATED(data_block2%p))&
                   CALL cp__w("dbcsr/block/dbcsr_block_access.F",440,"Data was not present in block")
                IF (ASSOCIATED (data_block2%p)) DEALLOCATE (data_block2%p)
             ENDIF
          ELSE
             CALL btree_add_s (&
                  matrix%m%wms(iw)%mutable%m%btree_s,&
                  make_coordinate_tuple(stored_row, stored_col),&
                  data_block, found, data_block2, replace=.FALSE.)
             IF (found) THEN
                IF(nze > 0) &
                   CALL saxpy (nze, 1.0_real_4, block(1), 1,&
                        data_block2%p(1,1), 1)
                IF(.NOT.ASSOCIATED(data_block%p))&
                   CALL cp__w("dbcsr/block/dbcsr_block_access.F",453,"Data was not present in block")
                IF (ASSOCIATED (data_block%p)) DEALLOCATE (data_block%p)
             ENDIF
          ENDIF
          IF (.NOT. found) THEN
             matrix%m%wms(iw)%lastblk = matrix%m%wms(iw)%lastblk + 1
          ENDIF
       ENDIF
       IF (.NOT. found) THEN
          matrix%m%wms(iw)%datasize = matrix%m%wms(iw)%datasize + SIZE (block)
       ELSE
       ENDIF
!$OMP CRITICAL (dbcsr_put_block_critical)
       matrix%m%valid = .FALSE.
!$OMP END CRITICAL (dbcsr_put_block_critical)
    ENDIF
    IF (PRESENT(flop)) flop = flop + my_flop
  END SUBROUTINE dbcsr_put_block_s


! **************************************************************************************************
!> \brief Sets a pointer, possibly using the buffers.
!> \param[in] matrix           Matrix to use
!> \param pointer_any The pointer to set
!> \param rsize Row size of block to point to
!> \param csize Column size of block to point to
!> \param[in] base_offset      The block pointer
! **************************************************************************************************
  SUBROUTINE dbcsr_set_block_pointer_2d_s (&
       matrix, pointer_any, rsize, csize, base_offset)
    TYPE(dbcsr_obj), INTENT(IN)              :: matrix
    REAL(kind=real_4), DIMENSION(:,:), POINTER         :: pointer_any
    INTEGER, INTENT(IN)                      :: rsize, csize
    INTEGER, INTENT(IN)                      :: base_offset

    CHARACTER(len=*), PARAMETER :: &
      routineN = 'dbcsr_set_block_pointer_2d_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: error_handler
    REAL(kind=real_4), DIMENSION(:), POINTER           :: lin_blk_p

!   ---------------------------------------------------------------------------

    IF (careful_mod) CALL timeset (routineN, error_handler)
    CALL dbcsr_get_data (matrix%m%data_area, lin_blk_p,&
         lb=base_offset, ub=base_offset+rsize*csize-1)
    CALL pointer_rank_remap2 (pointer_any, rsize, csize,&
         lin_blk_p)
    IF (careful_mod) CALL timestop (error_handler)
  END SUBROUTINE dbcsr_set_block_pointer_2d_s
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Gets a 2-d block from a dbcsr matrix
!> \param[in]  matrix DBCSR matrix
!> \param[in]  row    the row
!> \param[in]  col    the column
!> \param[out] block  the block to get (rank-2 array)
!> \param[out] tr     whether the data is transposed
!> \param[out] found  whether the block exists in the matrix
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
! **************************************************************************************************
  SUBROUTINE dbcsr_get_2d_block_p_c(matrix,row,col,block,tr,found,&
       row_size, col_size)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER         :: block
    LOGICAL, INTENT(OUT)                     :: tr
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_2d_block_p_c', &
      routineP = moduleN//':'//routineN

    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: block_1d
    INTEGER                                  :: rsize, csize,&
                                                blk, nze, offset,&
                                                stored_row,&
                                                stored_col, iw, nwms
    INTEGER                                  :: error_handle
    TYPE(btree_2d_data_c)          :: data_block
    LOGICAL                                  :: stored_tr
    COMPLEX(kind=real_4), DIMENSION(1,1), TARGET, SAVE    :: block0
!   ---------------------------------------------------------------------------
    IF (careful_mod) CALL timeset (routineN, error_handle)
    IF (debug_mod) THEN
       CALL dbcsr_assert (matrix%m%data_type, "EQ", dbcsr_type_complex_4,&
            dbcsr_fatal_level, dbcsr_caller_error,&
            routineN, "Data type mismatch for requested block.",43)
    ENDIF

    CALL dbcsr_get_block_index (matrix, row, col, stored_row, stored_col,&
         stored_tr, found, blk, offset)
    tr = stored_tr

    rsize = dbcsr_blk_row_size (matrix%m, stored_row)
    csize = dbcsr_blk_column_size (matrix%m, stored_col)
    IF (PRESENT (row_size)) row_size = rsize
    IF (PRESENT (col_size)) col_size = csize

    NULLIFY (block)
    IF(found) THEN
       nze = rsize*csize
       IF(nze.eq.0) THEN
          found = .TRUE.
          block => block0(1:0, 1:0)
       ELSE
          block_1d => pointer_view (dbcsr_get_data_p (&
               matrix%m%data_area, CMPLX(0.0, 0.0, real_4)), offset, offset+nze-1)
          CALL dbcsr_set_block_pointer (matrix, block, rsize, csize, offset)
       ENDIF
    ELSEIF (ASSOCIATED (matrix%m%wms)) THEN
       nwms = SIZE(matrix%m%wms)
       iw = 1
!$     CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$        dbcsr_fatal_level, dbcsr_internal_error,&
!$        routineN, "Number of work matrices not equal to number of threads", 71)
!$     iw = omp_get_thread_num () + 1
       IF(.NOT.dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",74,"Can not retrieve blocks from non-mutable work matrices.")
       IF (dbcsr_use_mutable (matrix%m)) THEN
          IF (.NOT. dbcsr_mutable_instantiated(matrix%m%wms(iw)%mutable)) THEN
             CALL dbcsr_mutable_new(matrix%m%wms(iw)%mutable,&
                  dbcsr_get_data_type(matrix))
          ENDIF
          CALL btree_get_c (&
               matrix%m%wms(iw)%mutable%m%btree_c,&
               make_coordinate_tuple(stored_row, stored_col),&
               data_block, found)
          IF (found) THEN
             block => data_block%p
          ENDIF
       ENDIF
    ENDIF
    IF (careful_mod) CALL timestop (error_handle)
  END SUBROUTINE dbcsr_get_2d_block_p_c


! **************************************************************************************************
!> \brief Gets a 1-d block from a dbcsr matrix
!> \param[in]  matrix DBCSR matrix
!> \param[in]  row    the row
!> \param[in]  col    the column
!> \param[out] block  the block to get (rank-1 array)
!> \param[out] tr     whether the data is transposed
!> \param[out] found  whether the block exists in the matrix
!> \param[out] row_size      (optional) logical row size of block
!> \param[out] col_size      (optional) logical column size of block
! **************************************************************************************************
  SUBROUTINE dbcsr_get_block_p_c(matrix,row,col,block,tr,found,&
       row_size, col_size)
    TYPE(dbcsr_obj), INTENT(IN)              :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: block
    LOGICAL, INTENT(OUT)                     :: tr
    LOGICAL, INTENT(OUT)                     :: found
    INTEGER, INTENT(OUT), OPTIONAL           :: row_size, col_size

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_block_p_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: blk, csize, &
                                                nze, offset, &
                                                rsize, stored_row,&
                                                stored_col
    LOGICAL                                  :: stored_tr

!   ---------------------------------------------------------------------------

    IF (debug_mod) THEN
       CALL dbcsr_assert (matrix%m%data_type, "EQ", dbcsr_type_complex_4,&
            dbcsr_fatal_level, dbcsr_caller_error,&
            routineN, "Data type mismatch for requested block.",127)
    ENDIF

    CALL dbcsr_get_block_index (matrix, row, col, stored_row, stored_col,&
         stored_tr, found, blk, offset)
    tr = stored_tr

    rsize = dbcsr_blk_row_size (matrix%m, stored_row)
    csize = dbcsr_blk_column_size (matrix%m, stored_col)
    IF (PRESENT (row_size)) row_size = rsize
    IF (PRESENT (col_size)) col_size = csize

    NULLIFY (block)
    IF(found) THEN
       nze = rsize*csize
       !
       block => pointer_view (&
            dbcsr_get_data_p (matrix%m%data_area, CMPLX(0.0, 0.0, real_4)), offset, offset+nze-1&
            )
    ELSEIF (ASSOCIATED (matrix%m%wms)) THEN
       IF(.NOT.dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",148,"Can not retrieve blocks from non-mutable work matrices.")
       IF(dbcsr_use_mutable (matrix%m))&
          CALL cp__b("dbcsr/block/dbcsr_block_access.F",150,"Can not retrieve rank-1 block pointers from mutable work matrices.")
    ENDIF
  END SUBROUTINE dbcsr_get_block_p_c


! **************************************************************************************************
!> \brief Put a 2-D block in a DBCSR matrix using the btree
!> \param[in.out] matrix      DBCSR matrix
!> \param[in]  row            the row
!> \param[in]  col            the column
!> \param[in]  block          the block to reserve; added if not NULL
!> \param[in] transposed      the block holds transposed data
!> \param[out] existed        (optional) block already existed
! **************************************************************************************************
  SUBROUTINE dbcsr_reserve_block2d_c(matrix, row, col, block,&
       transposed, existed)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER         :: block
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed
    LOGICAL, INTENT(OUT), OPTIONAL           :: existed

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_reserve_block2d_c', &
      routineP = moduleN//':'//routineN

    TYPE(btree_2d_data_c)          :: data_block, data_block2
    INTEGER                                  :: col_size, row_size, &
                                                stored_row, stored_col, &
                                                iw, nwms
    INTEGER, DIMENSION(:), POINTER           :: col_blk_size, row_blk_size
    LOGICAL                                  :: found, gift, tr, sym_tr
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER         :: original_block

!   ---------------------------------------------------------------------------

    gift = ASSOCIATED (block)
    IF (gift) THEN
       original_block => block
    ELSE
       NULLIFY (original_block)
    ENDIF
    row_blk_size => array_data (matrix%m%row_blk_size)
    col_blk_size => array_data (matrix%m%col_blk_size)
    row_size = row_blk_size(row)
    col_size = col_blk_size(col)

    stored_row = row ; stored_col = col
    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    sym_tr = .FALSE.
    CALL dbcsr_get_stored_coordinates (matrix, stored_row, stored_col)
    IF (.NOT.ASSOCIATED (matrix%m%wms)) THEN
       CALL dbcsr_work_create (matrix, work_mutable=.TRUE.)
       !$OMP MASTER
       matrix%m%valid = .FALSE.
       !$OMP END MASTER
       !$OMP BARRIER
    ENDIF

    NULLIFY (data_block%p)
    IF (.NOT. gift) THEN
       ALLOCATE (data_block%p (row_size, col_size))
       block => data_block%p
    ELSE
       data_block%p => block
    ENDIF
    data_block%tr = tr

    nwms = SIZE(matrix%m%wms)
    iw = 1
!$  CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$     dbcsr_fatal_level, dbcsr_internal_error,&
!$     routineN, "Number of work matrices not equal to number of threads", &
!$     226)
!$  iw = omp_get_thread_num () + 1
    CALL btree_add_c (matrix%m%wms(iw)%mutable%m%btree_c,&
         make_coordinate_tuple(stored_row, stored_col),&
         data_block, found, data_block2)

    IF (.NOT. found) THEN
!$OMP CRITICAL (critical_reserve_block2d)
       matrix%m%valid = .FALSE.
!$OMP END CRITICAL (critical_reserve_block2d)
       matrix%m%wms(iw)%lastblk = matrix%m%wms(iw)%lastblk + 1
       matrix%m%wms(iw)%datasize = matrix%m%wms(iw)%datasize + row_size*col_size
    ELSE
       IF (.NOT. gift) THEN
          DEALLOCATE (data_block%p)
       ELSE
          DEALLOCATE (original_block)
       ENDIF
       block => data_block2%p
    ENDIF
    IF (PRESENT (existed)) existed = found
  END SUBROUTINE dbcsr_reserve_block2d_c

! **************************************************************************************************
!> \brief Put a 2-D block in a DBCSR matrix
!> \param[in.out] matrix      DBCSR matrix
!> \param[in]  row            the row
!> \param[in]  col            the column
!> \param[in]  block          the block to put
!> \param[in]  transposed     the block is transposed
!> \param[in]  summation      (optional) if block exists, then sum the new
!>                            block to the old one instead of replacing it
!> \param[in]  scale          (optional) scale the block being added
! **************************************************************************************************
  SUBROUTINE dbcsr_put_block2d_c(matrix, row, col, block, lb_row_col, transposed,&
       summation, flop, scale)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_4), DIMENSION(:,:), INTENT(IN)      :: block
    INTEGER, DIMENSION(2), OPTIONAL, INTENT(INOUT) :: lb_row_col
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed, summation
    INTEGER(KIND=int_8), INTENT(INOUT), OPTIONAL :: flop
    COMPLEX(kind=real_4), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_put_block2d_c', &
      routineP = moduleN//':'//routineN

    LOGICAL                                  :: tr, do_sum

    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    IF (PRESENT (summation)) THEN
       do_sum = summation
    ELSE
       do_sum = .FALSE.
    ENDIF
    IF (PRESENT (scale)) THEN
       CALL dbcsr_put_block (matrix, row, col,&
            RESHAPE (block, (/SIZE(block)/)), lb_row_col, tr, do_sum, flop, scale)
    ELSE
       CALL dbcsr_put_block (matrix, row, col,&
            RESHAPE (block, (/SIZE(block)/)), lb_row_col, tr, do_sum, flop)
    ENDIF
  END SUBROUTINE dbcsr_put_block2d_c

! **************************************************************************************************
!> \brief Inserts a block in a dbcsr matrix.
!>
!> If the block exists, the current data is overwritten.
!> \param[in]  matrix         DBCSR matrix
!> \param[in]  row            the logical row
!> \param[in]  col            the logical column
!> \param[in]  block          the block to put
!> \param[in]  transposed     (optional) the block is transposed
!> \param[in]  summation      (optional) if block exists, then sum the new
!>                            block to the old one instead of replacing it
!> \param[in]  scale          (optional) scale the OBblock being added
! **************************************************************************************************
  SUBROUTINE dbcsr_put_block_c(matrix, row, col, block, lb_row_col, transposed,&
       summation, flop, scale)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    INTEGER, INTENT(IN)                      :: row, col
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN)        :: block
    INTEGER, DIMENSION(2), OPTIONAL, INTENT(INOUT) :: lb_row_col
    LOGICAL, INTENT(IN), OPTIONAL            :: transposed, summation
    INTEGER(KIND=int_8), INTENT(INOUT), OPTIONAL :: flop
    COMPLEX(kind=real_4), INTENT(IN), OPTIONAL            :: scale

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_put_block_c', &
      routineP = moduleN//':'//routineN

    TYPE(btree_2d_data_c)          :: data_block, data_block2
    INTEGER                                  :: blk, col_size, &
                                                nze, offset, &
                                                row_size, blk_p,&
                                                stored_row, stored_col,&
                                                iw, nwms
    LOGICAL                                  :: found, tr, do_sum, tr_diff,&
                                                sym_tr
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: block_1d
    INTEGER(KIND=int_8)                      :: my_flop

!   ---------------------------------------------------------------------------
    IF (PRESENT (transposed)) THEN
       tr = transposed
    ELSE
       tr = .FALSE.
    ENDIF
    IF (PRESENT (summation)) THEN
       do_sum = summation
    ELSE
       do_sum = .FALSE.
    ENDIF
    my_flop = 0
    row_size = dbcsr_blk_row_size(matrix, row)
    col_size = dbcsr_blk_column_size(matrix, col)
    IF (tr) CALL swap (row_size, col_size)

    stored_row = row ; stored_col = col; sym_tr = .FALSE.
    CALL dbcsr_get_stored_coordinates (matrix%m, stored_row, stored_col)
    nze = row_size*col_size
    !
    IF (debug_mod) THEN
       CALL dbcsr_assert (SIZE(block), "GE", nze, dbcsr_fatal_level,&
            dbcsr_caller_error, routineN, "Invalid block dimensions",353)
    ENDIF
    CALL dbcsr_get_stored_block_info (matrix%m, stored_row, stored_col,&
         found, blk, lb_row_col, offset)
    IF(found) THEN
       ! let's copy the block
       offset = ABS (offset)
       ! Fix the index if the new block's transpose flag is different
       ! from the old one.
       tr_diff = .FALSE.
       IF (matrix%m%blk_p(blk).LT.0 .NEQV. tr) THEN
          tr_diff = .TRUE.
          matrix%m%blk_p(blk) = -matrix%m%blk_p(blk)
       ENDIF
       block_1d => pointer_view (dbcsr_get_data_p (&
            matrix%m%data_area, CMPLX(0.0, 0.0, real_4)), offset, offset+nze-1)
       IF (nze .GT. 0) THEN
          IF (do_sum) THEN
             IF(tr_diff) &
                  block_1d = RESHAPE(TRANSPOSE(RESHAPE(block_1d,(/col_size,row_size/))),(/nze/))
             IF (PRESENT (scale)) THEN
                CALL caxpy (nze, scale, block(1:nze), 1,&
                     block_1d, 1)
             ELSE
                CALL caxpy (nze, CMPLX(1.0, 0.0, real_4), block(1:nze), 1,&
                     block_1d, 1)
             ENDIF
             my_flop = my_flop + nze * 2
          ELSE
             IF (PRESENT (scale)) THEN
                CALL ccopy (nze, scale*block(1:nze), 1,&
                     block_1d, 1)
             ELSE
                CALL ccopy (nze, block(1:nze), 1,&
                     block_1d, 1)
             ENDIF
          ENDIF
       ENDIF
    ELSE
       !!@@@
       !call cp_assert (associated (matrix%m%wms), cp_fatal_level,&
       !     cp_caller_error, routineN, "Work matrices not prepared")
       IF (.NOT.ASSOCIATED (matrix%m%wms)) THEN
          CALL dbcsr_work_create (matrix, nblks_guess=1,&
               sizedata_guess=SIZE(block))
       ENDIF
       nwms = SIZE(matrix%m%wms)
       iw = 1
!$     IF (debug_mod) THEN
!$     CALL dbcsr_assert (nwms, "GE", omp_get_num_threads(),&
!$        dbcsr_fatal_level, dbcsr_internal_error,&
!$        routineN, "Number of work matrices not equal to number of threads", 404)
!$     ENDIF
!$     iw = omp_get_thread_num () + 1
       blk_p = matrix%m%wms(iw)%datasize + 1
       IF (.NOT.dbcsr_wm_use_mutable (matrix%m%wms(iw))) THEN
          IF (tr) blk_p = -blk_p
          CALL add_work_coordinate (matrix%m%wms(iw), row, col, blk_p)
          CALL dbcsr_data_ensure_size (matrix%m%wms(iw)%data_area,&
               matrix%m%wms(iw)%datasize+SIZE(block),&
               factor=default_resize_factor)
          IF (PRESENT (scale)) THEN
             CALL dbcsr_data_set (matrix%m%wms(iw)%data_area, ABS(blk_p),&
                  data_size=SIZE(block), src=scale*block, source_lb=1)
          ELSE
             CALL dbcsr_data_set (matrix%m%wms(iw)%data_area, ABS(blk_p),&
                  data_size=SIZE(block), src=block, source_lb=1)
          ENDIF
       ELSE
          ALLOCATE (data_block%p (row_size, col_size))
          IF (PRESENT (scale)) THEN
             data_block%p(:,:) = scale*RESHAPE (block, (/row_size, col_size/))
          ELSE
             data_block%p(:,:) = RESHAPE (block, (/row_size, col_size/))
          ENDIF
          data_block%tr = tr
          IF (.NOT. dbcsr_mutable_instantiated(matrix%m%wms(iw)%mutable)) THEN
             CALL dbcsr_mutable_new(matrix%m%wms(iw)%mutable,&
                  dbcsr_get_data_type(matrix))
          ENDIF
          IF (.NOT. do_sum) THEN
             CALL btree_add_c (&
                  matrix%m%wms(iw)%mutable%m%btree_c,&
                  make_coordinate_tuple(stored_row, stored_col),&
                  data_block, found, data_block2, replace=.TRUE.)
             IF (found) THEN
                IF(.NOT.ASSOCIATED(data_block2%p))&
                   CALL cp__w("dbcsr/block/dbcsr_block_access.F",440,"Data was not present in block")
                IF (ASSOCIATED (data_block2%p)) DEALLOCATE (data_block2%p)
             ENDIF
          ELSE
             CALL btree_add_c (&
                  matrix%m%wms(iw)%mutable%m%btree_c,&
                  make_coordinate_tuple(stored_row, stored_col),&
                  data_block, found, data_block2, replace=.FALSE.)
             IF (found) THEN
                IF(nze > 0) &
                   CALL caxpy (nze, CMPLX(1.0, 0.0, real_4), block(1), 1,&
                        data_block2%p(1,1), 1)
                IF(.NOT.ASSOCIATED(data_block%p))&
                   CALL cp__w("dbcsr/block/dbcsr_block_access.F",453,"Data was not present in block")
                IF (ASSOCIATED (data_block%p)) DEALLOCATE (data_block%p)
             ENDIF
          ENDIF
          IF (.NOT. found) THEN
             matrix%m%wms(iw)%lastblk = matrix%m%wms(iw)%lastblk + 1
          ENDIF
       ENDIF
       IF (.NOT. found) THEN
          matrix%m%wms(iw)%datasize = matrix%m%wms(iw)%datasize + SIZE (block)
       ELSE
       ENDIF
!$OMP CRITICAL (dbcsr_put_block_critical)
       matrix%m%valid = .FALSE.
!$OMP END CRITICAL (dbcsr_put_block_critical)
    ENDIF
    IF (PRESENT(flop)) flop = flop + my_flop
  END SUBROUTINE dbcsr_put_block_c


! **************************************************************************************************
!> \brief Sets a pointer, possibly using the buffers.
!> \param[in] matrix           Matrix to use
!> \param pointer_any The pointer to set
!> \param rsize Row size of block to point to
!> \param csize Column size of block to point to
!> \param[in] base_offset      The block pointer
! **************************************************************************************************
  SUBROUTINE dbcsr_set_block_pointer_2d_c (&
       matrix, pointer_any, rsize, csize, base_offset)
    TYPE(dbcsr_obj), INTENT(IN)              :: matrix
    COMPLEX(kind=real_4), DIMENSION(:,:), POINTER         :: pointer_any
    INTEGER, INTENT(IN)                      :: rsize, csize
    INTEGER, INTENT(IN)                      :: base_offset

    CHARACTER(len=*), PARAMETER :: &
      routineN = 'dbcsr_set_block_pointer_2d_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: error_handler
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: lin_blk_p

!   ---------------------------------------------------------------------------

    IF (careful_mod) CALL timeset (routineN, error_handler)
    CALL dbcsr_get_data (matrix%m%data_area, lin_blk_p,&
         lb=base_offset, ub=base_offset+rsize*csize-1)
    CALL pointer_rank_remap2 (pointer_any, rsize, csize,&
         lin_blk_p)
    IF (careful_mod) CALL timestop (error_handler)
  END SUBROUTINE dbcsr_set_block_pointer_2d_c

END MODULE dbcsr_block_access
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief collection of types used in dbcsr_arnoldi
!> \par History
!>       2014.09 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

MODULE dbcsr_arnoldi_types
   USE dbcsr_types,                     ONLY: dbcsr_obj
   USE kinds,                           ONLY: real_4,&
                                              real_8

   IMPLICIT NONE

! Type that gets created during the arnoldi procedure and contains basically everything
! As it is not quite clear what the user will request, this is the most general way to satisfy all needs:
! Give him everything we have and create some easy to use routines to post process externally
   TYPE arnoldi_control
      LOGICAL                                 :: local_comp, converged
      INTEGER                                 :: myproc, mp_group, pcol_group, prow_group
      INTEGER                                 :: max_iter ! Maximum number of iterations
      INTEGER                                 :: current_step ! In case subspace converged early contains last iteration
      INTEGER                                 :: nval_req
      INTEGER                                 :: selection_crit
      INTEGER                                 :: nval_out
      INTEGER                                 :: nrestart
      REAL(real_8)                            :: threshold
      LOGICAL                                 :: symmetric
      LOGICAL                                 :: generalized_ev
      LOGICAL                                 :: iram
      LOGICAL                                 :: has_initial_vector
      INTEGER, DIMENSION(:), POINTER          :: selected_ind ! list of indicies matching the selection criterium
   END TYPE arnoldi_control

   TYPE arnoldi_data_d
      REAL(kind=real_8), POINTER, DIMENSION(:)         :: f_vec ! the local parts of the residual vector
      REAL(kind=real_8), POINTER, DIMENSION(:, :)      :: Hessenberg ! the Hessenberg matrix
      REAL(kind=real_8), POINTER, DIMENSION(:, :)      :: local_history ! the complete set of orthonormal vectors (local part)
      COMPLEX(real_8), POINTER, DIMENSION(:)           :: evals ! the real part of the eigenvalues (if complex both)
      COMPLEX(real_8), POINTER, DIMENSION(:, :)        :: revec ! the right eigenvectors
      REAL(kind=real_8)                                :: rho_scale ! scling factor for general eig arnoldi
      REAL(kind=real_8), POINTER, DIMENSION(:)         :: x_vec ! eigenvector for genreal eig arnoldi
   END TYPE arnoldi_data_d

   TYPE arnoldi_data_s
      REAL(kind=real_4), POINTER, DIMENSION(:)         :: f_vec ! the local parts of the residual vector
      REAL(kind=real_4), POINTER, DIMENSION(:, :)      :: Hessenberg ! the Hessenberg matrix
      REAL(kind=real_4), POINTER, DIMENSION(:, :)      :: local_history ! the complete set of orthonormal vectors (local part)
      COMPLEX(real_4), POINTER, DIMENSION(:)           :: evals ! the real part of the eigenvalues (if complex both)
      COMPLEX(real_4), POINTER, DIMENSION(:, :)        :: revec ! the right eigenvectors
      REAL(kind=real_4)                                :: rho_scale ! scling factor for general eig arnoldi
      REAL(kind=real_4), POINTER, DIMENSION(:)         :: x_vec ! eigenvector for genreal eig arnoldi
   END TYPE arnoldi_data_s

   TYPE arnoldi_data_z
      COMPLEX(kind=real_8), POINTER, DIMENSION(:)      :: f_vec ! the local parts of the residual vector
      COMPLEX(kind=real_8), POINTER, DIMENSION(:, :)   :: Hessenberg ! the Hessenberg matrix
      COMPLEX(kind=real_8), POINTER, DIMENSION(:, :)   :: local_history ! the complete set of orthonormal vectors (local part)
      COMPLEX(real_8), POINTER, DIMENSION(:)           :: evals ! the real part of the eigenvalues (if complex both)
      COMPLEX(real_8), POINTER, DIMENSION(:, :)        :: revec ! the right eigenvectors
      COMPLEX(kind=real_8)                             :: rho_scale ! scling factor for general eig arnoldi
      COMPLEX(kind=real_8), POINTER, DIMENSION(:)      :: x_vec ! eigenvector for genreal eig arnoldi
   END TYPE arnoldi_data_z

   TYPE arnoldi_data_c
      COMPLEX(kind=real_4), POINTER, DIMENSION(:)      :: f_vec ! the local parts of the residual vector
      COMPLEX(kind=real_4), POINTER, DIMENSION(:, :)   :: Hessenberg ! the Hessenberg matrix
      COMPLEX(kind=real_4), POINTER, DIMENSION(:, :)   :: local_history ! the complete set of orthonormal vectors (local part)
      COMPLEX(real_4), POINTER, DIMENSION(:)           :: evals ! the real part of the eigenvalues (if complex both)
      COMPLEX(real_4), POINTER, DIMENSION(:, :)        :: revec ! the right eigenvectors
      COMPLEX(kind=real_4)                             :: rho_scale ! scling factor for general eig arnoldi
      COMPLEX(kind=real_4), POINTER, DIMENSION(:)      :: x_vec ! eigenvector for genreal eig arnoldi
   END TYPE arnoldi_data_c

   TYPE dbcsr_arnoldi_data
      TYPE(arnoldi_data_s), POINTER, PRIVATE              :: data_s => NULL()
      TYPE(arnoldi_data_d), POINTER, PRIVATE              :: data_d => NULL()
      TYPE(arnoldi_data_c), POINTER, PRIVATE              :: data_c => NULL()
      TYPE(arnoldi_data_z), POINTER, PRIVATE              :: data_z => NULL()
      TYPE(arnoldi_control), POINTER, PRIVATE             :: control
   END TYPE dbcsr_arnoldi_data

   TYPE m_x_v_vectors
      TYPE(dbcsr_obj)                          :: input_vec
      TYPE(dbcsr_obj)                          :: result_vec
      TYPE(dbcsr_obj)                          :: rep_col_vec
      TYPE(dbcsr_obj)                          :: rep_row_vec
   END TYPE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_arnoldi_types'

   PUBLIC :: dbcsr_arnoldi_data, m_x_v_vectors, get_data_d, get_data_s, get_sel_ind, &
             get_data_z, get_data_c, get_control, has_s_real, has_d_real, arnoldi_control, &
             has_s_cmplx, has_d_cmplx, arnoldi_data_d, arnoldi_data_s, arnoldi_data_z, arnoldi_data_c, &
             get_evals_d, get_evals_c, get_evals_z, get_evals_s, set_control, set_data_d, set_data_s, &
             set_data_z, set_data_c
CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \param control ...
! **************************************************************************************************
   SUBROUTINE set_control(ar_data, control)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      TYPE(arnoldi_control), POINTER                     :: control

      ar_data%control => control
   END SUBROUTINE set_control

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval selected_ind ...
! **************************************************************************************************
   FUNCTION get_sel_ind(ar_data) RESULT(selected_ind)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      INTEGER, DIMENSION(:), POINTER                     :: selected_ind

      selected_ind => ar_data%control%selected_ind

   END FUNCTION get_sel_ind

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval data_d ...
! **************************************************************************************************
   FUNCTION get_data_d(ar_data) RESULT(data_d)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      TYPE(arnoldi_data_d), POINTER                      :: data_d

      data_d => ar_data%data_d

   END FUNCTION get_data_d

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval data_s ...
! **************************************************************************************************
   FUNCTION get_data_s(ar_data) RESULT(data_s)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      TYPE(arnoldi_data_s), POINTER                      :: data_s

      data_s => ar_data%data_s

   END FUNCTION get_data_s

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval data_z ...
! **************************************************************************************************
   FUNCTION get_data_z(ar_data) RESULT(data_z)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      TYPE(arnoldi_data_z), POINTER                      :: data_z

      data_z => ar_data%data_z

   END FUNCTION get_data_z

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval data_c ...
! **************************************************************************************************
   FUNCTION get_data_c(ar_data) RESULT(data_c)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      TYPE(arnoldi_data_c), POINTER                      :: data_c

      data_c => ar_data%data_c

   END FUNCTION get_data_c

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \param data_d ...
! **************************************************************************************************
   SUBROUTINE set_data_d(ar_data, data_d)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      TYPE(arnoldi_data_d), POINTER                      :: data_d

      ar_data%data_d => data_d

   END SUBROUTINE set_data_d

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \param data_s ...
! **************************************************************************************************
   SUBROUTINE set_data_s(ar_data, data_s)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      TYPE(arnoldi_data_s), POINTER                      :: data_s

      ar_data%data_s => data_s

   END SUBROUTINE set_data_s

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \param data_c ...
! **************************************************************************************************
   SUBROUTINE set_data_c(ar_data, data_c)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      TYPE(arnoldi_data_c), POINTER                      :: data_c

      ar_data%data_c => data_c

   END SUBROUTINE set_data_c

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \param data_z ...
! **************************************************************************************************
   SUBROUTINE set_data_z(ar_data, data_z)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      TYPE(arnoldi_data_z), POINTER                      :: data_z

      ar_data%data_z => data_z

   END SUBROUTINE set_data_z

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval control ...
! **************************************************************************************************
   FUNCTION get_control(ar_data) RESULT(control)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      TYPE(arnoldi_control), POINTER                     :: control

      control => ar_data%control

   END FUNCTION get_control

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval is_present ...
! **************************************************************************************************
   FUNCTION has_d_real(ar_data) RESULT(is_present)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      LOGICAL                                            :: is_present

      is_present = ASSOCIATED(ar_data%data_d)

   END FUNCTION has_d_real

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval is_present ...
! **************************************************************************************************
   FUNCTION has_s_real(ar_data) RESULT(is_present)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      LOGICAL                                            :: is_present

      is_present = ASSOCIATED(ar_data%data_s)

   END FUNCTION has_s_real

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval is_present ...
! **************************************************************************************************
   FUNCTION has_d_cmplx(ar_data) RESULT(is_present)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      LOGICAL                                            :: is_present

      is_present = ASSOCIATED(ar_data%data_z)

   END FUNCTION has_d_cmplx

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval is_present ...
! **************************************************************************************************
   FUNCTION has_s_cmplx(ar_data) RESULT(is_present)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      LOGICAL                                            :: is_present

      is_present = ASSOCIATED(ar_data%data_c)

   END FUNCTION has_s_cmplx

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval evals ...
! **************************************************************************************************
   FUNCTION get_evals_d(ar_data) RESULT(evals)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      COMPLEX(real_8), DIMENSION(:), POINTER             :: evals

      evals => ar_data%data_d%evals

   END FUNCTION get_evals_d

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval evals ...
! **************************************************************************************************
   FUNCTION get_evals_s(ar_data) RESULT(evals)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      COMPLEX(real_4), DIMENSION(:), POINTER             :: evals

      evals => ar_data%data_s%evals

   END FUNCTION get_evals_s

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval evals ...
! **************************************************************************************************
   FUNCTION get_evals_z(ar_data) RESULT(evals)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      COMPLEX(real_8), DIMENSION(:), POINTER             :: evals

      evals => ar_data%data_z%evals

   END FUNCTION get_evals_z

! **************************************************************************************************
!> \brief ...
!> \param ar_data ...
!> \retval evals ...
! **************************************************************************************************
   FUNCTION get_evals_c(ar_data) RESULT(evals)
      TYPE(dbcsr_arnoldi_data)                           :: ar_data
      COMPLEX(real_4), DIMENSION(:), POINTER             :: evals

      evals => ar_data%data_c%evals

   END FUNCTION get_evals_c

END MODULE dbcsr_arnoldi_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Wrappers to BLAS calls.
!> \author  Urban Borstnik
!> \date    2010-02-18
!> \version 0.9
!>
!> <b>Modification history:</b>
!> - 2010-02-18 Moved from dbcsr_util
! **************************************************************************************************
MODULE dbcsr_blas_operations
   USE dbcsr_types,                     ONLY: dbcsr_data_obj,&
                                              dbcsr_type_complex_4,&
                                              dbcsr_type_complex_8,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8
   USE kinds,                           ONLY: int_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_blas_operations'

   PUBLIC :: dbcsr_lapack_larnv, set_larnv_seed

CONTAINS

! **************************************************************************************************
!> \brief generate a seed respecting the lapack constraints,
!>        - values between 0..4095 (2**12-1)
!>        - iseed(4) odd
!>        also try to avoid iseed that are zero.
!>        Have but with a unique 2D mapping (irow,icol), and a 'mini-seed' ival
!> \param irow 1..nrow
!> \param nrow nrow
!> \param icol 1..ncol
!> \param ncol ncol
!> \param ival mini-seed
!> \param iseed a lapack compatible seed
! **************************************************************************************************
   SUBROUTINE set_larnv_seed(irow, nrow, icol, ncol, ival, iseed)
      INTEGER, INTENT(IN)                                :: irow, nrow, icol, ncol, ival
      INTEGER, INTENT(OUT)                               :: iseed(4)

      CHARACTER(len=*), PARAMETER :: routineN = 'set_larnv_seed', routineP = moduleN//':'//routineN

      INTEGER(KIND=int_8)                                :: map

      map = ((irow-1+icol*INT(nrow, int_8))*(1+MODULO(ival, 2**16)))*2+1+0*ncol ! ncol used
      iseed(4) = INT(MODULO(map, 2_int_8**12)); map = map/2_int_8**12; ! keep odd
      iseed(3) = INT(MODULO(IEOR(map, 3541_int_8), 2_int_8**12)); map = map/2_int_8**12
      iseed(2) = INT(MODULO(IEOR(map, 1153_int_8), 2_int_8**12)); map = map/2_int_8**12
      iseed(1) = INT(MODULO(IEOR(map, 2029_int_8), 2_int_8**12)); map = map/2_int_8**12
   END SUBROUTINE set_larnv_seed
! **************************************************************************************************
!> \brief fils an array with random numbers
!> \param idist type of distribution (1..3, see lapack docs)
!> \param iseed requires each int to be in the range 0..2**12, and the iseed(4) odd
!> \param n ...
!> \param x ...
! **************************************************************************************************
   SUBROUTINE dbcsr_lapack_larnv(idist, iseed, n, x)
      INTEGER, INTENT(IN)                                :: idist
      INTEGER, DIMENSION(:), INTENT(INOUT)               :: iseed
      INTEGER, INTENT(IN)                                :: n
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: x

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_lapack_larnv', &
         routineP = moduleN//':'//routineN

      SELECT CASE (x%d%data_type)
      CASE (dbcsr_type_real_4)
         CALL slarnv(idist, iseed, n, x%d%r_sp)
      CASE (dbcsr_type_real_8)
         CALL dlarnv(idist, iseed, n, x%d%r_dp)
      CASE (dbcsr_type_complex_4)
         CALL clarnv(idist, iseed, n, x%d%c_sp)
      CASE (dbcsr_type_complex_8)
         CALL zlarnv(idist, iseed, n, x%d%c_dp)
      CASE default
         CALL cp__b("dbcsr/ops/dbcsr_blas_operations.F",86,"Invalid data type")
      END SELECT

   END SUBROUTINE dbcsr_lapack_larnv

END MODULE dbcsr_blas_operations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief   Higher-level operations on DBCSR matrices.
!> \author  Urban Borstnik
!> \date    2009-05-12
!> \version 0.9
!>
!> <b>Modification history:</b>
!  - Created 2009-05-12
! **************************************************************************************************

MODULE dbcsr_operations
   USE array_types,                     ONLY: array_data,&
                                              array_get
   USE dbcsr_blas_operations,           ONLY: set_larnv_seed
   USE dbcsr_block_access,              ONLY: dbcsr_get_block_p,&
                                              dbcsr_put_block,&
                                              dbcsr_remove_block,&
                                              dbcsr_reserve_blocks
   USE dbcsr_block_operations,          ONLY: block_add_on_diag,&
                                              dbcsr_block_conjg,&
                                              dbcsr_block_partial_copy,&
                                              dbcsr_block_real_neg,&
                                              dbcsr_block_scale,&
                                              dbcsr_data_clear,&
                                              get_block2d_diagonal,&
                                              set_block2d_diagonal
   USE dbcsr_data_methods,              ONLY: &
        dbcsr_data_clear_pointer, dbcsr_data_ensure_size, dbcsr_data_get_size, &
        dbcsr_data_get_size_referenced, dbcsr_data_get_type, dbcsr_data_init, dbcsr_data_new, &
        dbcsr_data_release, dbcsr_data_set_pointer, dbcsr_get_data, dbcsr_scalar, &
        dbcsr_scalar_are_equal, dbcsr_scalar_fill_all, dbcsr_scalar_get_type, dbcsr_scalar_one, &
        dbcsr_scalar_zero, dbcsr_type_1d_to_2d
   USE dbcsr_data_operations,           ONLY: dbcsr_data_convert,&
                                              dbcsr_data_copyall,&
                                              dbcsr_switch_data_area
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_col_dist,&
                                              dbcsr_distribution_local_cols,&
                                              dbcsr_distribution_local_rows,&
                                              dbcsr_distribution_mp,&
                                              dbcsr_distribution_row_dist
   USE dbcsr_dist_operations,           ONLY: checker_square_proc,&
                                              checker_tr,&
                                              dbcsr_find_column,&
                                              dbcsr_get_stored_coordinates
   USE dbcsr_error_handling,            ONLY: dbcsr_assert,&
                                              dbcsr_fatal_level,&
                                              dbcsr_internal_error,&
                                              dbcsr_wrong_args_error
   USE dbcsr_index_operations,          ONLY: dbcsr_index_checksum,&
                                              dbcsr_index_compact,&
                                              dbcsr_repoint_index
   USE dbcsr_iterator_operations,       ONLY: dbcsr_iterator_blocks_left,&
                                              dbcsr_iterator_next_block,&
                                              dbcsr_iterator_start,&
                                              dbcsr_iterator_stop
   USE dbcsr_methods,                   ONLY: &
        dbcsr_col_block_offsets, dbcsr_distribution, dbcsr_get_data_size, dbcsr_get_data_type, &
        dbcsr_get_index_memory_type, dbcsr_get_matrix_type, dbcsr_get_num_blocks, &
        dbcsr_get_replication_type, dbcsr_has_symmetry, dbcsr_is_initialized, dbcsr_max_col_size, &
        dbcsr_max_row_size, dbcsr_name, dbcsr_nblkcols_total, dbcsr_nblkrows_total, &
        dbcsr_nfullcols_total, dbcsr_nfullrows_total, dbcsr_row_block_offsets, dbcsr_valid_index
   USE dbcsr_mp_methods,                ONLY: dbcsr_mp_group,&
                                              dbcsr_mp_mynode,&
                                              dbcsr_mp_mypcol,&
                                              dbcsr_mp_myprow,&
                                              dbcsr_mp_numnodes,&
                                              dbcsr_mp_pgrid
   USE dbcsr_ptr_util,                  ONLY: ensure_array_size,&
                                              pointer_view
   USE dbcsr_toollib,                   ONLY: swap
   USE dbcsr_types,                     ONLY: &
        dbcsr_data_obj, dbcsr_distribution_obj, dbcsr_filter_frobenius, dbcsr_func_artanh, &
        dbcsr_func_asin, dbcsr_func_cos, dbcsr_func_ddsin, dbcsr_func_ddtanh, dbcsr_func_dsin, &
        dbcsr_func_dtanh, dbcsr_func_inverse, dbcsr_func_inverse_special, dbcsr_func_sin, &
        dbcsr_func_spread_from_zero, dbcsr_func_tanh, dbcsr_func_truncate, dbcsr_iterator, &
        dbcsr_mp_obj, dbcsr_norm_column, dbcsr_norm_frobenius, dbcsr_norm_gershgorin, &
        dbcsr_norm_maxabsnorm, dbcsr_obj, dbcsr_repl_full, dbcsr_repl_none, dbcsr_scalar_type, &
        dbcsr_type_antihermitian, dbcsr_type_antisymmetric, dbcsr_type_complex_4, &
        dbcsr_type_complex_8, dbcsr_type_hermitian, dbcsr_type_invalid, dbcsr_type_no_symmetry, &
        dbcsr_type_real_4, dbcsr_type_real_8, dbcsr_type_symmetric
   USE dbcsr_util,                      ONLY: find_block_of_element
   USE dbcsr_work_operations,           ONLY: dbcsr_create,&
                                              dbcsr_finalize,&
                                              dbcsr_work_create
   USE kinds,                           ONLY: dp,&
                                              int_8,&
                                              real_4,&
                                              real_8,&
                                              sp
   USE message_passing,                 ONLY: mp_allgather,&
                                              mp_max,&
                                              mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_operations'

   ! prettify protection
   CHARACTER, PARAMETER :: xa = dbcsr_type_hermitian, xb = dbcsr_type_antihermitian, &
                           xc = dbcsr_type_no_symmetry

   PUBLIC :: dbcsr_trace, dbcsr_add_on_diag, &
             dbcsr_set, dbcsr_scale, dbcsr_add, dbcsr_copy, &
             dbcsr_copy_into_existing, &
             dbcsr_get_diag, dbcsr_set_diag, &
             dbcsr_get_block_diag, dbcsr_hadamard_product, &
             dbcsr_filter, dbcsr_scale_by_vector, &
             dbcsr_function_of_elements, &
             dbcsr_triu, &
             dbcsr_init_random
   PUBLIC :: dbcsr_sum_replicated
   PUBLIC :: dbcsr_norm, dbcsr_conjg, &
             dbcsr_gershgorin_norm, dbcsr_maxabs, dbcsr_frobenius_norm

   PUBLIC :: dbcsr_crop_matrix
   PUBLIC :: dbcsr_get_info, dbcsr_may_be_dense, dbcsr_get_occupation

! The interfaces for the generic routines found in the generated
! generic files.

   INTERFACE dbcsr_trace
      MODULE PROCEDURE dbcsr_trace_a_any
      MODULE PROCEDURE dbcsr_trace_a_s, dbcsr_trace_a_d, &
         dbcsr_trace_a_c, dbcsr_trace_a_z
      MODULE PROCEDURE dbcsr_trace_ab_s, dbcsr_trace_a_b_d, &
         dbcsr_trace_ab_c, dbcsr_trace_ab_z
   END INTERFACE

   INTERFACE dbcsr_scale
      MODULE PROCEDURE dbcsr_scale_anytype
      MODULE PROCEDURE dbcsr_scale_s, dbcsr_scale_d, &
         dbcsr_scale_c, dbcsr_scale_z
   END INTERFACE

   INTERFACE dbcsr_scale_by_vector
      MODULE PROCEDURE dbcsr_scale_by_vector_anytype
      MODULE PROCEDURE dbcsr_scale_by_vector_s, dbcsr_scale_by_vector_d, &
         dbcsr_scale_by_vector_c, dbcsr_scale_by_vector_z
   END INTERFACE

   INTERFACE dbcsr_set
      MODULE PROCEDURE dbcsr_set_anytype
      MODULE PROCEDURE dbcsr_set_s, dbcsr_set_d, dbcsr_set_c, dbcsr_set_z
   END INTERFACE

   INTERFACE dbcsr_add
      MODULE PROCEDURE dbcsr_add_anytype
      MODULE PROCEDURE dbcsr_add_s, dbcsr_add_d, &
         dbcsr_add_c, dbcsr_add_z
   END INTERFACE

   INTERFACE dbcsr_add_on_diag
      MODULE PROCEDURE dbcsr_add_on_diag_anytype
      MODULE PROCEDURE dbcsr_add_on_diag_s, dbcsr_add_on_diag_d, &
         dbcsr_add_on_diag_c, dbcsr_add_on_diag_z
   END INTERFACE

   INTERFACE dbcsr_filter
      MODULE PROCEDURE dbcsr_filter_anytype
      MODULE PROCEDURE dbcsr_filter_s, dbcsr_filter_d, &
         dbcsr_filter_c, dbcsr_filter_z
   END INTERFACE

   INTERFACE dbcsr_get_diag
      MODULE PROCEDURE dbcsr_get_diag_anytype
      MODULE PROCEDURE dbcsr_get_diag_s, dbcsr_get_diag_d, &
         dbcsr_get_diag_c, dbcsr_get_diag_z
   END INTERFACE

   INTERFACE dbcsr_set_diag
      MODULE PROCEDURE dbcsr_set_diag_anytype
      MODULE PROCEDURE dbcsr_set_diag_s, dbcsr_set_diag_d, &
         dbcsr_set_diag_c, dbcsr_set_diag_z
   END INTERFACE

   INTERFACE dbcsr_norm
      MODULE PROCEDURE dbcsr_norm_anytype
      MODULE PROCEDURE dbcsr_norm_r4_scal
      MODULE PROCEDURE dbcsr_norm_r4_vec, dbcsr_norm_r8_vec
   END INTERFACE
   INTERFACE dbcsr_gershgorin_norm
      MODULE PROCEDURE dbcsr_gershgorin_norm_r8
   END INTERFACE
   INTERFACE dbcsr_maxabs
      MODULE PROCEDURE dbcsr_maxabs_r8
   END INTERFACE
   INTERFACE dbcsr_frobenius_norm
      MODULE PROCEDURE dbcsr_frobenius_norm_r8
   END INTERFACE

   LOGICAL, PARAMETER :: debug_mod = .FALSE.
   LOGICAL, PARAMETER :: careful_mod = .FALSE.


   INTEGER, PARAMETER, PRIVATE :: rpslot_owner = 1
   INTEGER, PARAMETER, PRIVATE :: rpslot_addblks = 2
   INTEGER, PARAMETER, PRIVATE :: rpslot_addoffset = 3
   INTEGER, PARAMETER, PRIVATE :: rpslot_oldblks = 4
   INTEGER, PARAMETER, PRIVATE :: rpslot_oldoffset = 5
   INTEGER, PARAMETER, PRIVATE :: rpslot_totaloffset = 6
   INTEGER, PARAMETER, PRIVATE :: rpnslots = 6

CONTAINS

! **************************************************************************************************
!> \brief Conjugate a DBCSR matrix
!> \param[inout] matrix       DBCSR matrix
!>
! **************************************************************************************************
   SUBROUTINE dbcsr_conjg(matrix)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_conjg', routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, data_type, handle, row
      LOGICAL                                            :: tr
      TYPE(dbcsr_data_obj)                               :: data_any
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------
!

      CALL timeset(routineN, handle)
      data_type = dbcsr_get_data_type(matrix)
      CALL dbcsr_data_init(data_any)
      CALL dbcsr_data_new(data_any, data_type)
      CALL dbcsr_iterator_start(iter, matrix)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, data_any, tr, blk)
         SELECT CASE (data_type)
         CASE (dbcsr_type_complex_4)
            data_any%d%c_sp = CONJG(data_any%d%c_sp)
         CASE (dbcsr_type_complex_8)
            data_any%d%c_dp = CONJG(data_any%d%c_dp)
         CASE DEFAULT
            ! needed for g95
         END SELECT
      ENDDO
      CALL dbcsr_iterator_stop(iter)
      CALL dbcsr_data_clear_pointer(data_any)
      CALL dbcsr_data_release(data_any)
      CALL timestop(handle)
   END SUBROUTINE dbcsr_conjg

! **************************************************************************************************
!> \brief fill a dbcsr matrix with zeros
!> \param matrix_a ...
! **************************************************************************************************
   SUBROUTINE dbcsr_zero(matrix_a)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_zero', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      SELECT CASE (dbcsr_get_data_type (matrix_a))
      CASE (dbcsr_type_complex_4)
!$OMP       PARALLEL WORKSHARE DEFAULT(NONE), SHARED(matrix_a)
         matrix_a%m%data_area%d%c_sp = (0.0, 0.0)
!$OMP       END PARALLEL WORKSHARE
      CASE (dbcsr_type_complex_8)
!$OMP       PARALLEL WORKSHARE DEFAULT(NONE), SHARED(matrix_a)
         matrix_a%m%data_area%d%c_dp = (0.0_dp, 0.0_dp)
!$OMP       END PARALLEL WORKSHARE
      CASE (dbcsr_type_real_4)
!$OMP       PARALLEL WORKSHARE DEFAULT(NONE), SHARED(matrix_a)
         matrix_a%m%data_area%d%r_sp = 0.0
!$OMP       END PARALLEL WORKSHARE
      CASE (dbcsr_type_real_8)
!$OMP       PARALLEL WORKSHARE DEFAULT(NONE), SHARED(matrix_a)
         matrix_a%m%data_area%d%r_dp = 0.0_dp
!$OMP       END PARALLEL WORKSHARE
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE dbcsr_zero

! **************************************************************************************************
!> \brief Scales a DBCSR matrix by alpha
!> \param[inout] matrix_a       DBCSR matrix
!> \param[in] alpha_scalar      (optional) a scalar
!> \param[in] limits            (optional) Scale only a subbox
!> \par Limits
!> A 4-tuple describing (first_row, last_row, first_column, last_column). Set
!> to 0 to avoid limiting.
! **************************************************************************************************
   SUBROUTINE dbcsr_scale_anytype(matrix_a, alpha_scalar, limits)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: alpha_scalar
      INTEGER, DIMENSION(4), INTENT(IN), OPTIONAL        :: limits

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scale_anytype', &
         routineP = moduleN//':'//routineN
      INTEGER, PARAMETER                                 :: first_col_i = 3, first_row_i = 1, &
                                                            last_col_i = 4, last_row_i = 2

      INTEGER :: a_col, a_col_size, a_row, a_row_size, col_offset, handle, row_offset, &
         scale_col_offset, scale_col_size, scale_row_offset, scale_row_size
      INTEGER, DIMENSION(4)                              :: my_limits
      LOGICAL                                            :: do_scale, has_limits, tr
      TYPE(dbcsr_data_obj)                               :: data_any
      TYPE(dbcsr_iterator)                               :: iter
      TYPE(dbcsr_scalar_type)                            :: one

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      ! Limits are only honored if the argument is present and any are
      ! non-zero.
      IF (PRESENT(limits)) THEN
         has_limits = ANY(limits(:) .NE. 0)
      ELSE
         has_limits = .FALSE.
      ENDIF
      my_limits(first_row_i) = 1
      my_limits(last_row_i) = dbcsr_nfullrows_total(matrix_a)
      my_limits(first_col_i) = 1
      my_limits(last_col_i) = dbcsr_nfullcols_total(matrix_a)
      IF (has_limits) THEN
         IF (limits(last_col_i) .NE. 0) THEN
            IF (debug_mod) THEN
               CALL dbcsr_assert(limits(last_col_i) .GE. 0, "AND", &
                                 limits(last_col_i) .LE. dbcsr_nfullcols_total(matrix_a), &
                                 dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                                 "Specified last column is out of bounds.", 340)
            ENDIF
            my_limits(last_col_i) = limits(last_col_i)
         ENDIF
         IF (limits(first_col_i) .NE. 0) THEN
            IF (debug_mod) THEN
               CALL dbcsr_assert(limits(first_col_i) .GE. 0, "AND", &
                                 limits(first_col_i) .LE. dbcsr_nfullcols_total(matrix_a), &
                                 dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                                 "Specified first column is out of bounds.", 349)
            ENDIF
            my_limits(first_col_i) = limits(first_col_i)
         ENDIF
         IF (limits(last_row_i) .NE. 0) THEN
            IF (debug_mod) THEN
               CALL dbcsr_assert(limits(last_row_i) .GE. 0, "AND", &
                                 limits(last_row_i) .LE. dbcsr_nfullrows_total(matrix_a), &
                                 dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                                 "Specified last row is out of bounds.", 358)
            ENDIF
            my_limits(last_row_i) = limits(last_row_i)
         ENDIF
         IF (limits(first_row_i) .NE. 0) THEN
            IF (debug_mod) THEN
               CALL dbcsr_assert(limits(first_row_i) .GE. 0, "AND", &
                                 limits(first_row_i) .LE. dbcsr_nfullrows_total(matrix_a), &
                                 dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                                 "Specified first row is out of bounds.", 367)
            ENDIF
            my_limits(first_row_i) = limits(first_row_i)
         ENDIF
      ENDIF
      !
      ! quick return if possible
      one = dbcsr_scalar_one(dbcsr_scalar_get_type(alpha_scalar))
      do_scale = .NOT. dbcsr_scalar_are_equal(alpha_scalar, one)
      !
      ! let's go
      IF (do_scale) THEN
!$OMP        PARALLEL DEFAULT (NONE) &
!$OMP                 PRIVATE (iter, data_any) &
!$OMP                 PRIVATE (a_row, a_col, tr, a_row_size, a_col_size, &
!$OMP                          row_offset, col_offset) &
!$OMP                 PRIVATE (scale_row_size, scale_col_size,&
!$OMP                          scale_row_offset, scale_col_offset) &
!$OMP                 SHARED (matrix_a, my_limits,alpha_scalar)
         CALL dbcsr_data_init(data_any)
         CALL dbcsr_data_new(data_any, dbcsr_type_1d_to_2d(dbcsr_get_data_type(matrix_a)))
         CALL dbcsr_iterator_start(iter, matrix_a, read_only=.FALSE., &
                                   contiguous_pointers=.FALSE., dynamic=.TRUE., &
                                   dynamic_byrows=.TRUE., shared=.TRUE.)
         iterations: DO WHILE (dbcsr_iterator_blocks_left(iter))
            CALL dbcsr_iterator_next_block(iter, a_row, a_col, data_any, tr, &
                                           row_size=a_row_size, col_size=a_col_size, &
                                           row_offset=row_offset, col_offset=col_offset)
            IF (a_row_size .GT. 0 .AND. a_col_size .GT. 0) THEN
               CALL frame_block_limit(a_row_size, row_offset, &
                                      my_limits(first_row_i), my_limits(last_row_i), &
                                      scale_row_size, scale_row_offset)
               CALL frame_block_limit(a_col_size, col_offset, &
                                      my_limits(first_col_i), my_limits(last_col_i), &
                                      scale_col_size, scale_col_offset)
               IF (tr) THEN
                  CALL swap(scale_row_size, scale_col_size)
                  CALL swap(scale_row_offset, scale_col_offset)
               ENDIF
               CALL dbcsr_block_scale(data_any, scale=alpha_scalar, &
                                      row_size=scale_row_size, col_size=scale_col_size, &
                                      lb=scale_row_offset, lb2=scale_col_offset)
            ENDIF
         ENDDO iterations
         CALL dbcsr_iterator_stop(iter)
         CALL dbcsr_data_clear_pointer(data_any)
         CALL dbcsr_data_release(data_any)
!$OMP        END PARALLEL
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_scale_anytype

! **************************************************************************************************
!> \brief Determines the effect of limits on a block
!> \param[in] block_size     size of block
!> \param[in] block_offset   global offset of block
!> \param[in] first_limit    lower limit
!> \param[in] last_limit     upper limit
!> \param[out] frame_size    size of block region within the limits
!> \param[out] frame_offset  starting position of the block region
!>                           that is within the limits
! **************************************************************************************************
   ELEMENTAL SUBROUTINE frame_block_limit(block_size, block_offset, &
                                          first_limit, last_limit, &
                                          frame_size, frame_offset)
      INTEGER, INTENT(IN)                                :: block_size, block_offset, first_limit, &
                                                            last_limit
      INTEGER, INTENT(OUT)                               :: frame_size, frame_offset

      INTEGER                                            :: f, l

      f = MAX(block_offset, first_limit)
      l = MIN(block_offset+block_size-1, last_limit)
      frame_size = MAX(l-f+1, 0)
      frame_offset = MIN(f-block_offset+1, block_size)
   END SUBROUTINE frame_block_limit

! **************************************************************************************************
!> \brief Scales a DBCSR matrix by alpha
!> \param[inout] matrix_a       DBCSR matrix
!> \param[in] alpha             the scaling vector
!> \param[in] side              apply the scaling from the side
! **************************************************************************************************
   SUBROUTINE dbcsr_scale_by_vector_anytype(matrix_a, alpha, side)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a
      TYPE(dbcsr_data_obj), INTENT(IN), OPTIONAL         :: alpha
      CHARACTER(LEN=*), INTENT(IN)                       :: side

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scale_by_vector_anytype', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: a_blk, a_col, a_col_size, a_nze, a_row, &
                                                            a_row_size, col_offset, data_type, &
                                                            handle, i, icol, irow, row_offset
      LOGICAL                                            :: right, tr
      TYPE(dbcsr_data_obj)                               :: data_any
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

! check that alpha and matrix have the same data type
      IF (dbcsr_get_data_type(matrix_a) .NE. alpha%d%data_type) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",471,"wrong data type matrix_a")
      !
      ! set vars
      right = .TRUE.
      SELECT CASE (side)
      CASE ('right'); right = .TRUE.
      CASE ('left'); right = .FALSE.
      CASE DEFAULT
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",479,"wrong side="//side)
      END SELECT
      !
      ! let's go
      data_type = dbcsr_get_data_type(matrix_a)
      CALL dbcsr_data_init(data_any)
      CALL dbcsr_data_new(data_any, dbcsr_get_data_type(matrix_a))
      CALL dbcsr_iterator_start(iter, matrix_a)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, a_row, a_col, data_any, tr, &
                                        block_number=a_blk, &
                                        row_size=a_row_size, col_size=a_col_size, &
                                        row_offset=row_offset, col_offset=col_offset)
         a_nze = a_row_size*a_col_size
         IF (a_nze .EQ. 0) CYCLE ! Skip empty blocks
         !
         ! let's scale
         IF (right) THEN
            SELECT CASE (data_type)
            CASE (dbcsr_type_real_4)
               DO i = 1, a_col_size
                  DO icol = (i-1)*a_row_size+1, (i-1)*a_row_size+a_row_size
                     data_any%d%r_sp(icol) = data_any%d%r_sp(icol)*alpha%d%r_sp(col_offset+i-1)
                  END DO
               END DO
            CASE (dbcsr_type_real_8)
               DO i = 1, a_col_size
                  DO icol = (i-1)*a_row_size+1, (i-1)*a_row_size+a_row_size
                     data_any%d%r_dp(icol) = data_any%d%r_dp(icol)*alpha%d%r_dp(col_offset+i-1)
                  END DO
               END DO
            CASE (dbcsr_type_complex_4)
               DO i = 1, a_col_size
                  DO icol = (i-1)*a_row_size+1, (i-1)*a_row_size+a_row_size
                     data_any%d%c_sp(icol) = data_any%d%c_sp(icol)*alpha%d%c_sp(col_offset+i-1)
                  END DO
               END DO
            CASE (dbcsr_type_complex_8)
               DO i = 1, a_col_size
                  DO icol = (i-1)*a_row_size+1, (i-1)*a_row_size+a_row_size
                     data_any%d%c_dp(icol) = data_any%d%c_dp(icol)*alpha%d%c_dp(col_offset+i-1)
                  END DO
               END DO
            END SELECT
         ELSE
            SELECT CASE (data_type)
            CASE (dbcsr_type_real_4)
               DO i = 1, a_row_size
                  DO irow = i, i+a_col_size**2-1, a_col_size
                     data_any%d%r_sp(irow) = data_any%d%r_sp(irow)*alpha%d%r_sp(row_offset+i-1)
                  END DO
               ENDDO
            CASE (dbcsr_type_real_8)
               DO i = 1, a_row_size
                  DO irow = i, i+a_col_size**2-1, a_col_size
                     data_any%d%r_dp(irow) = data_any%d%r_dp(irow)*alpha%d%r_dp(row_offset+i-1)
                  END DO
               ENDDO
            CASE (dbcsr_type_complex_4)
               DO i = 1, a_row_size
                  DO irow = i, i+a_col_size**2-1, a_col_size
                     data_any%d%c_sp(irow) = data_any%d%c_sp(irow)*alpha%d%c_sp(row_offset+i-1)
                  END DO
               ENDDO
            CASE (dbcsr_type_complex_8)
               DO i = 1, a_row_size
                  DO irow = i, i+a_col_size**2-1, a_col_size
                     data_any%d%c_dp(irow) = data_any%d%c_dp(irow)*alpha%d%c_dp(row_offset+i-1)
                  END DO
               ENDDO
            END SELECT
         ENDIF
      ENDDO
      CALL dbcsr_iterator_stop(iter)
      CALL dbcsr_data_clear_pointer(data_any)
      CALL dbcsr_data_release(data_any)
      CALL timestop(handle)

   END SUBROUTINE dbcsr_scale_by_vector_anytype

! **************************************************************************************************
!> \brief Set a DBCSR matrix by alpha
!> \param[inout] matrix       DBCSR matrix
!> \param[in] alpha           a scalar
!>
! **************************************************************************************************
   SUBROUTINE dbcsr_set_anytype(matrix, alpha)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: alpha

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_anytype', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, handle, row
      LOGICAL                                            :: tr
      TYPE(dbcsr_data_obj)                               :: data_block
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------
!

      CALL timeset(routineN, handle)
      CALL dbcsr_data_init(data_block)
      CALL dbcsr_data_new(data_block, dbcsr_get_data_type(matrix))
      CALL dbcsr_iterator_start(iter, matrix)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, data_block, tr, blk)
         CALL dbcsr_data_clear(data_block, value=alpha)
      ENDDO
      CALL dbcsr_iterator_stop(iter)
      CALL dbcsr_data_clear_pointer(data_block)
      CALL dbcsr_data_release(data_block)
      CALL timestop(handle)
      !
   END SUBROUTINE dbcsr_set_anytype

! **************************************************************************************************
!> \brief add and scale matrices
!>    A = alpha*A + beta*B or
!> \param[in,out] matrix_a   DBCSR matrix
!> \param[in] matrix_b       DBCSR matrix
!> \param[in] alpha_scalar   (optional)
!> \param[in] beta_scalar    (optional)
!>
!> \param flop ...
! **************************************************************************************************
   SUBROUTINE dbcsr_add_anytype(matrix_a, matrix_b, alpha_scalar, beta_scalar, flop)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_b
      TYPE(dbcsr_scalar_type), INTENT(IN), OPTIONAL      :: alpha_scalar, beta_scalar
      INTEGER(KIND=int_8), INTENT(INOUT), OPTIONAL       :: flop

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_add_anytype', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, data_type_a, data_type_b, &
                                                            handle, row, size_a, size_b
      INTEGER(KIND=int_8)                                :: my_flop
      INTEGER, DIMENSION(2)                              :: lb_row_col
      LOGICAL                                            :: do_scale, tr
      TYPE(dbcsr_data_obj)                               :: data_block
      TYPE(dbcsr_iterator)                               :: iter
      TYPE(dbcsr_scalar_type)                            :: my_alpha_scalar, my_beta_scalar

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (.NOT. dbcsr_valid_index(matrix_a)) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",627,"Invalid matrix")

      IF ((dbcsr_get_matrix_type(matrix_a) .EQ. dbcsr_type_symmetric .OR. &
           dbcsr_get_matrix_type(matrix_a) .EQ. dbcsr_type_antisymmetric) .NEQV. &
          (dbcsr_get_matrix_type(matrix_b) .EQ. dbcsr_type_symmetric .OR. &
           dbcsr_get_matrix_type(matrix_b) .EQ. dbcsr_type_antisymmetric)) THEN
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",633,"Summing general with symmetric matrix NYI")
      ENDIF

      data_type_a = dbcsr_get_data_type(matrix_a)
      data_type_b = dbcsr_get_data_type(matrix_b)
      !
      my_alpha_scalar = dbcsr_scalar_one(data_type_a)
      IF (PRESENT(alpha_scalar)) my_alpha_scalar = alpha_scalar
      my_beta_scalar = dbcsr_scalar_one(data_type_b)
      IF (PRESENT(beta_scalar)) my_beta_scalar = beta_scalar
      !
      ! let's go
      IF (dbcsr_nblkrows_total(matrix_a) .NE. dbcsr_nblkrows_total(matrix_b)) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",646,"matrices not consistent")

      do_scale = dbcsr_scalar_are_equal( &
                 my_beta_scalar, dbcsr_scalar_one(data_type_b))

      CALL dbcsr_scale(matrix_a, alpha_scalar=my_alpha_scalar)

      ! Pre-size work arrays of matrix_a to avoid continuous reallocation.
      size_a = dbcsr_data_get_size_referenced(matrix_a%m%data_area)
      size_b = dbcsr_data_get_size_referenced(matrix_b%m%data_area)
      IF (.NOT. dbcsr_scalar_are_equal(my_beta_scalar, &
                                       dbcsr_scalar_zero(data_type_b))) THEN
         my_flop = 0
!$OMP        PARALLEL DEFAULT (NONE) &
!$OMP                 PRIVATE (iter, data_block) &
!$OMP                 PRIVATE (row, col, tr, blk, lb_row_col) &
!$OMP                 SHARED (matrix_a, matrix_b, data_type_b, size_b, size_a) &
!$OMP                 SHARED (do_scale, my_beta_scalar) &
!$OMP                 REDUCTION (+ : my_flop)
         lb_row_col(:) = 0
         IF (size_b .GT. size_a .AND. matrix_b%m%nblks .GT. matrix_a%m%nblks) THEN
            CALL dbcsr_work_create(matrix_a, &
                                   nblks_guess=matrix_b%m%nblks-matrix_a%m%nblks, &
                                   sizedata_guess=size_b-size_a, &
                                   work_mutable=.FALSE.)
         ELSE
            CALL dbcsr_work_create(matrix_a, &
                                   work_mutable=.FALSE.)
         ENDIF
!$OMP        BARRIER
         CALL dbcsr_data_init(data_block)
         CALL dbcsr_data_new(data_block, data_type_b)
         CALL dbcsr_iterator_start(iter, matrix_b, &
                                   shared=.TRUE., read_only=.TRUE., contiguous_pointers=.FALSE., &
                                   dynamic=.TRUE., dynamic_byrows=.TRUE.)

         DO WHILE (dbcsr_iterator_blocks_left(iter))

            CALL dbcsr_iterator_next_block(iter, row, col, data_block, tr, blk)

            IF (do_scale) THEN
               CALL dbcsr_put_block(matrix_a, row, col, data_block, lb_row_col, tr, &
                                    summation=.TRUE., flop=my_flop)
            ELSE
               CALL dbcsr_put_block(matrix_a, row, col, data_block, lb_row_col, tr, &
                                    summation=.TRUE., flop=my_flop, scale=my_beta_scalar)
            ENDIF

         ENDDO

         CALL dbcsr_iterator_stop(iter)
         CALL dbcsr_finalize(matrix_a)
         CALL dbcsr_data_clear_pointer(data_block)
         CALL dbcsr_data_release(data_block)
!$OMP        END PARALLEL
         IF (PRESENT(flop)) flop = flop+my_flop
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_add_anytype

!> \brief Interface for dbcsr_add
!> \param matrix_a ...
!> \param matrix_b ...
!> \param alpha_scalar ...
!> \param beta_scalar ...
! **************************************************************************************************
   SUBROUTINE dbcsr_add_d(matrix_a, matrix_b, alpha_scalar, beta_scalar)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_b
      REAL(real_8), INTENT(IN)                           :: alpha_scalar, beta_scalar

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_add_d', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      IF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_real_8 .AND. &
          dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_real_8) THEN
         CALL dbcsr_add_anytype(matrix_a, matrix_b, &
                                alpha_scalar=dbcsr_scalar(alpha_scalar), &
                                beta_scalar=dbcsr_scalar(beta_scalar))
      ELSEIF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_real_4 .AND. &
              dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_real_4) THEN
         CALL dbcsr_add_anytype(matrix_a, matrix_b, &
                                alpha_scalar=dbcsr_scalar(REAL(alpha_scalar, real_4)), &
                                beta_scalar=dbcsr_scalar(REAL(beta_scalar, real_4)))
      ELSEIF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_complex_4 .AND. &
              dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_complex_4) THEN
         CALL dbcsr_add_anytype(matrix_a, matrix_b, &
                                alpha_scalar=dbcsr_scalar(CMPLX(alpha_scalar, 0, real_4)), &
                                beta_scalar=dbcsr_scalar(CMPLX(beta_scalar, 0, real_4)))
      ELSEIF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_complex_8 .AND. &
              dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_complex_8) THEN
         CALL dbcsr_add_anytype(matrix_a, matrix_b, &
                                alpha_scalar=dbcsr_scalar(CMPLX(alpha_scalar, 0, real_8)), &
                                beta_scalar=dbcsr_scalar(CMPLX(beta_scalar, 0, real_8)))
      ELSEIF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_real_8 .AND. &
              dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_real_4) THEN
         CALL dbcsr_add_anytype(matrix_a, matrix_b, &
                                alpha_scalar=dbcsr_scalar(alpha_scalar), &
                                beta_scalar=dbcsr_scalar(REAL(beta_scalar, real_4)))
      ELSEIF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_real_4 .AND. &
              dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_real_8) THEN
         CALL dbcsr_add_anytype(matrix_a, matrix_b, &
                                alpha_scalar=dbcsr_scalar(REAL(alpha_scalar, real_4)), &
                                beta_scalar=dbcsr_scalar(beta_scalar))
      ELSE
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",753,"Invalid combination of data type, NYI")
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_add_d

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param alpha_scalar ...
!> \param beta_scalar ...
! **************************************************************************************************
   SUBROUTINE dbcsr_add_s(matrix_a, matrix_b, alpha_scalar, beta_scalar)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_b
      REAL(real_4), INTENT(IN)                           :: alpha_scalar, beta_scalar

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_add_s', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      IF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_real_4 .AND. &
          dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_real_4) THEN
         CALL dbcsr_add_anytype(matrix_a, matrix_b, &
                                alpha_scalar=dbcsr_scalar(alpha_scalar), &
                                beta_scalar=dbcsr_scalar(beta_scalar))
      ELSE
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",781,"Invalid combination of data type, NYI")
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_add_s

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param alpha_scalar ...
!> \param beta_scalar ...
! **************************************************************************************************
   SUBROUTINE dbcsr_add_z(matrix_a, matrix_b, alpha_scalar, beta_scalar)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_b
      COMPLEX(real_8), INTENT(IN)                        :: alpha_scalar, beta_scalar

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_add_z', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      IF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_complex_8 .AND. &
          dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_complex_8) THEN
         CALL dbcsr_add_anytype(matrix_a, matrix_b, &
                                alpha_scalar=dbcsr_scalar(alpha_scalar), &
                                beta_scalar=dbcsr_scalar(beta_scalar))
      ELSEIF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_complex_4 .AND. &
              dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_complex_4) THEN
         CALL dbcsr_add_anytype(matrix_a, matrix_b, &
                                alpha_scalar=dbcsr_scalar(CMPLX(alpha_scalar, KIND=real_4)), &
                                beta_scalar=dbcsr_scalar(CMPLX(beta_scalar, KIND=real_4)))
      ELSE
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",814,"Invalid combination of data type, NYI")
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_add_z

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param alpha_scalar ...
!> \param beta_scalar ...
! **************************************************************************************************
   SUBROUTINE dbcsr_add_c(matrix_a, matrix_b, alpha_scalar, beta_scalar)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_b
      COMPLEX(real_4), INTENT(IN)                        :: alpha_scalar, beta_scalar

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_add_c', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      IF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_complex_4 .AND. &
          dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_complex_4) THEN
         CALL dbcsr_add_anytype(matrix_a, matrix_b, &
                                alpha_scalar=dbcsr_scalar(alpha_scalar), &
                                beta_scalar=dbcsr_scalar(beta_scalar))
      ELSE
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",842,"Invalid combination of data type, NYI")
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_add_c

! **************************************************************************************************
!> \brief Computes varios functions (defined by func) of matrix elements
!>        NOTE: sign(A,B) returns the value of A with the sign of B
!>        dbcsr_func_inverse:   1/(a1*x+a0)
!>           fails if the inversion produces infinite numbers
!>        dbcsr_func_inverse_special: 1/(x+sign(a0,x))
!>           safe inverse: if a0>0 then the denominator is never zero
!>        dbcsr_func_tanh:    tanh(a1*x+a0)
!>        dbcsr_func_dtanh:   d(tanh(a1*x+a0)) / dx
!>        dbcsr_func_ddtanh:  d2(tanh(a1*x+a0)) / dx2
!>        dbcsr_func_artanh:  artanh(a1*x+a0)=ln[(1+(a1*x+a0))/(1-(a1*x+a0))]/2
!>           fails if |a1*x+a0| >= 1
!>        dbcsr_func_sread_from_zero:  if |x|<|a0| then x=sign(a0,x)
!>        dbcsr_func_truncate:  if |x|>|a0| then x=sign(a0,x)
!>        dbcsr_func_sin:     sin(a1*x+a0)
!>        dbcsr_func_cos:     cos(a1*x+a0)
!>        dbcsr_func_dsin:    d(sin(a1*x+a0)) / dx = a1*cos(a1*x+a0)
!>        dbcsr_func_ddsin:   d2(sin(a1*x+a0)) / dx2 = -a1*a1*sin(a1*x+a0)
!>        dbcsr_func_asin:    asin(a1*x+a0)
!>           fails if |a1*x+a0| > 1
!> \param[in] matrix_a        DBCSR matrix
!> \param func ...
!> \param a0 ...
!> \param a1 ...
!> \param a2 ...
! **************************************************************************************************
   SUBROUTINE dbcsr_function_of_elements(matrix_a, func, a0, a1, a2)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a
      INTEGER, INTENT(IN)                                :: func
      REAL(kind=dp), INTENT(IN), OPTIONAL                :: a0, a1, a2

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_function_of_elements', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, col_size, data_type, handle, &
                                                            ielem, nze, row, row_size
      LOGICAL                                            :: tr_a
      REAL(kind=dp)                                      :: p0, p1, p2
      TYPE(dbcsr_data_obj)                               :: a_data
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      IF (PRESENT(a0)) THEN
         p0 = a0
      ELSE
         p0 = 0.0_dp
      ENDIF
      IF (PRESENT(a1)) THEN
         p1 = a1
      ELSE
         p1 = 1.0_dp
      ENDIF
      IF (PRESENT(a2)) THEN
         p2 = a2
      ELSE
         p2 = 0.0_dp
      ENDIF

      data_type = dbcsr_get_data_type(matrix_a)
      CALL dbcsr_data_init(a_data)
      CALL dbcsr_data_new(a_data, data_type)
      CALL dbcsr_iterator_start(iter, matrix_a)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, a_data, tr_a, blk, &
                                        row_size=row_size, col_size=col_size)
         nze = row_size*col_size
         SELECT CASE (data_type)
            !CASE (dbcsr_type_real_4)
            !   a_data%d%r_sp(1:nze) = 1.0_real_4/a_data%d%r_sp(1:nze)
            !   CALL dbcsr_assert(MAXVAL(ABS(a_data%d%r_sp)).LT.HUGE(0.0_real_4),&
            !           dbcsr_fatal_level,dbcsr_internal_error, routineN,&
            !           "Division by zero", 921)
         CASE (dbcsr_type_real_8)
            SELECT CASE (func)
            CASE (dbcsr_func_spread_from_zero)
               ! if |x|<|a0| then x=|a0|*sign(x)
               DO ielem = 1, nze
                  IF (ABS(a_data%d%r_dp(ielem)) .LT. ABS(p0)) THEN
                     a_data%d%r_dp(ielem) = SIGN(p0, a_data%d%r_dp(ielem))
                  ENDIF
               ENDDO
            CASE (dbcsr_func_truncate)
               ! if |x|>|a0| then x=|a0|*sign(x)
               DO ielem = 1, nze
                  IF (ABS(a_data%d%r_dp(ielem)) .GT. ABS(p0)) THEN
                     a_data%d%r_dp(ielem) = SIGN(p0, a_data%d%r_dp(ielem))
                  ENDIF
               ENDDO
            CASE (dbcsr_func_inverse_special)
               !IF (MINVAL(ABS(a_data%d%r_dp)).le.ABS(p2)) THEN
               !   ! there is at least one near-zero element,
               !   ! invert element-by-element
               !   DO ielem=1,nze
               !     IF (a_data%d%r_dp(ielem).le.ABS(p2)) THEN
               !        a_data%d%r_dp(ielem) = 0.0_real_8
               !     ELSE
               !        a_data%d%r_dp(ielem) = &
               !           1.0_real_8/(p1*a_data%d%r_dp(ielem)+p0)
               !     ENDIF
               !   ENDDO
               !ELSE
               !   a_data%d%r_dp(1:nze) = 1.0_real_8/(p1*a_data%d%r_dp(1:nze)+p0)
               !ENDIF
               a_data%d%r_dp(1:nze) = 1.0_real_8/(a_data%d%r_dp(1:nze)+SIGN(p0, a_data%d%r_dp(1:nze)))
            CASE (dbcsr_func_inverse)
               a_data%d%r_dp(1:nze) = 1.0_real_8/(p1*a_data%d%r_dp(1:nze)+p0)
               IF (MAXVAL(ABS(a_data%d%r_dp)) .GE. HUGE(0.0_real_8)) &
                  CALL cp__b("dbcsr/ops/dbcsr_operations.F",957,"Division by zero")
            CASE (dbcsr_func_tanh)
               a_data%d%r_dp(1:nze) = TANH(p1*a_data%d%r_dp(1:nze)+p0)
            CASE (dbcsr_func_dtanh)
               a_data%d%r_dp(1:nze) = TANH(p1*a_data%d%r_dp(1:nze)+p0)
               a_data%d%r_dp(1:nze) = a_data%d%r_dp(1:nze)**2
               a_data%d%r_dp(1:nze) = p1*(1.0_real_8-a_data%d%r_dp(1:nze))
            CASE (dbcsr_func_ddtanh)
               a_data%d%r_dp(1:nze) = TANH(p1*a_data%d%r_dp(1:nze)+p0)
               a_data%d%r_dp(1:nze) = a_data%d%r_dp(1:nze)**3-a_data%d%r_dp(1:nze)
               a_data%d%r_dp(1:nze) = 2.0_real_8*(p1**2)*a_data%d%r_dp(1:nze)
            CASE (dbcsr_func_artanh)
               a_data%d%r_dp(1:nze) = p1*a_data%d%r_dp(1:nze)+p0
               IF (MAXVAL(ABS(a_data%d%r_dp)) .GE. 1.0_real_8) &
                  CALL cp__b("dbcsr/ops/dbcsr_operations.F",971,"ARTANH is undefined for |x|>=1")
               a_data%d%r_dp(1:nze) = (1.0_real_8+a_data%d%r_dp(1:nze)) &
                                      /(1.0_real_8-a_data%d%r_dp(1:nze))
               a_data%d%r_dp(1:nze) = 0.5_real_8*LOG(a_data%d%r_dp(1:nze))
            CASE (dbcsr_func_sin)
               a_data%d%r_dp(1:nze) = SIN(p1*a_data%d%r_dp(1:nze)+p0)
            CASE (dbcsr_func_cos)
               a_data%d%r_dp(1:nze) = COS(p1*a_data%d%r_dp(1:nze)+p0)
            CASE (dbcsr_func_dsin)
               a_data%d%r_dp(1:nze) = p1*COS(p1*a_data%d%r_dp(1:nze)+p0)
            CASE (dbcsr_func_ddsin)
               a_data%d%r_dp(1:nze) = -p1*p1*SIN(p1*a_data%d%r_dp(1:nze)+p0)
            CASE (dbcsr_func_asin)
               a_data%d%r_dp(1:nze) = p1*a_data%d%r_dp(1:nze)+p0
               IF (MAXVAL(ABS(a_data%d%r_dp)) .GT. 1.0_real_8) &
                  CALL cp__b("dbcsr/ops/dbcsr_operations.F",986,"ASIN is undefined for |x|>1")
               a_data%d%r_dp(1:nze) = ASIN(a_data%d%r_dp(1:nze))
            CASE DEFAULT
               CALL cp__b("dbcsr/ops/dbcsr_operations.F",989,"Unknown function of matrix elements")
            END SELECT
            !CASE (dbcsr_type_complex_4)
            !CASE (dbcsr_type_complex_8)
         CASE DEFAULT
            CALL cp__b("dbcsr/ops/dbcsr_operations.F",994,"Operation is implemented only for dp real values")
         END SELECT
      ENDDO
      CALL dbcsr_iterator_stop(iter)
      CALL dbcsr_data_clear_pointer(a_data)
      CALL dbcsr_data_release(a_data)
      CALL timestop(handle)

   END SUBROUTINE dbcsr_function_of_elements

! **************************************************************************************************
!> \brief Hadamard product
!>    C = A . B (C needs to be different from A and B)
!> \param[in] matrix_a DBCSR matrix
!> \param[in] matrix_b DBCSR matrix
!> \param[in,out] matrix_c        DBCSR matrix
!> \param b_assume_value ...
! **************************************************************************************************
   SUBROUTINE dbcsr_hadamard_product(matrix_a, matrix_b, matrix_c, &
                                     b_assume_value)

      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a, matrix_b
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_c
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: b_assume_value

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_hadamard_product', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, col_size, data_type, handle, &
                                                            nze, row, row_size
      LOGICAL                                            :: assume_blocks_in_b, found, tr_a, tr_b
      REAL(KIND=dp)                                      :: assumed_b_value
      TYPE(dbcsr_data_obj)                               :: a_data, b_data, c_data
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      IF (PRESENT(b_assume_value)) THEN
         assume_blocks_in_b = .TRUE.
         assumed_b_value = b_assume_value
      ELSE
         assume_blocks_in_b = .FALSE.
         assumed_b_value = 0.0_dp
      ENDIF

      CALL timeset(routineN, handle)
      IF (dbcsr_get_data_type(matrix_a) .NE. dbcsr_get_data_type(matrix_b) .OR. &
          dbcsr_get_data_type(matrix_a) .NE. dbcsr_get_data_type(matrix_c)) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",1042,"data types not consistent, need to fix that")

      IF (dbcsr_nblkrows_total(matrix_a) .NE. dbcsr_nblkrows_total(matrix_b) .OR. &
          dbcsr_nblkrows_total(matrix_c) .NE. dbcsr_nblkrows_total(matrix_a)) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",1046,"matrices not consistent")

      data_type = dbcsr_get_data_type(matrix_a)
      CALL dbcsr_data_init(c_data)
      CALL dbcsr_data_new(c_data, data_type, &
                          data_size=dbcsr_max_row_size(matrix_a)*dbcsr_max_col_size(matrix_a))
      CALL dbcsr_set(matrix_c, dbcsr_scalar_zero(data_type))
      CALL dbcsr_data_init(a_data)
      CALL dbcsr_data_new(a_data, data_type)
      CALL dbcsr_data_init(b_data)
      CALL dbcsr_data_new(b_data, data_type)
      CALL dbcsr_iterator_start(iter, matrix_a)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         SELECT CASE (dbcsr_get_data_type (matrix_a))
            !CASE (dbcsr_type_real_4)
         CASE (dbcsr_type_real_8)
            CALL dbcsr_iterator_next_block(iter, row, col, a_data, tr_a, blk, &
                                           row_size=row_size, col_size=col_size)
            nze = row_size*col_size
            CALL dbcsr_get_block_p(matrix_b, row, col, b_data, tr_b, found)
            IF (tr_a .NEQV. tr_b) &
               CALL cp__b("dbcsr/ops/dbcsr_operations.F",1067,"tr not consistent, need to fix that")
            IF (found) THEN
               SELECT CASE (data_type)
               CASE (dbcsr_type_real_4)
                  c_data%d%r_sp(1:nze) = a_data%d%r_sp(1:nze)*b_data%d%r_sp(1:nze)
               CASE (dbcsr_type_real_8)
                  c_data%d%r_dp(1:nze) = a_data%d%r_dp(1:nze)*b_data%d%r_dp(1:nze)
               CASE (dbcsr_type_complex_4)
                  c_data%d%c_sp(1:nze) = a_data%d%c_sp(1:nze)*b_data%d%c_sp(1:nze)
               CASE (dbcsr_type_complex_8)
                  c_data%d%c_dp(1:nze) = a_data%d%c_dp(1:nze)*b_data%d%c_dp(1:nze)
               END SELECT
               CALL dbcsr_put_block(matrix_c, row, col, c_data, transposed=tr_a, &
                                    summation=.FALSE.)
            ELSE
               IF (assume_blocks_in_b) THEN ! this makes not too much sense, to delete ?
                  SELECT CASE (data_type)
                  CASE (dbcsr_type_real_4)
                     c_data%d%r_sp(1:nze) = a_data%d%r_sp(1:nze)*REAL(assumed_b_value, KIND=sp)
                  CASE (dbcsr_type_real_8)
                     c_data%d%r_dp(1:nze) = a_data%d%r_dp(1:nze)*assumed_b_value
                  CASE (dbcsr_type_complex_4)
                     c_data%d%c_sp(1:nze) = a_data%d%c_sp(1:nze)*REAL(assumed_b_value, KIND=sp)
                  CASE (dbcsr_type_complex_8)
                     c_data%d%c_dp(1:nze) = a_data%d%c_dp(1:nze)*assumed_b_value
                  END SELECT
                  CALL dbcsr_put_block(matrix_c, row, col, c_data, transposed=tr_a, &
                                       summation=.FALSE.)
               ENDIF
            ENDIF
            !CASE (dbcsr_type_complex_4)
            !CASE (dbcsr_type_complex_8)
         CASE DEFAULT
            CALL cp__b("dbcsr/ops/dbcsr_operations.F",1100,"Only real double precision")
         END SELECT
      ENDDO
      CALL dbcsr_iterator_stop(iter)
      CALL dbcsr_finalize(matrix_c)
      CALL dbcsr_data_clear_pointer(a_data)
      CALL dbcsr_data_clear_pointer(b_data)
      CALL dbcsr_data_release(c_data)
      CALL dbcsr_data_release(a_data)
      CALL dbcsr_data_release(b_data)
      CALL timestop(handle)
   END SUBROUTINE dbcsr_hadamard_product

! **************************************************************************************************
!> \brief add a constant to the diagonal of a matrix
!> \param[inout] matrix       DBCSR matrix
!> \param[in]    alpha_scalar scalar
!> \param first_row ...
!> \param last_row ...
! **************************************************************************************************
   SUBROUTINE dbcsr_add_on_diag_anytype(matrix, alpha_scalar, first_row, last_row)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: alpha_scalar
      INTEGER, INTENT(in), OPTIONAL                      :: first_row, last_row

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_add_on_diag_anytype', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, hold, imax, imin, my_first_row, &
                                                            my_last_row, mynode, offset_beg, &
                                                            offset_end, row, row_size, stored_row
      INTEGER, DIMENSION(:), POINTER                     :: row_blk_offsets
      LOGICAL                                            :: found, tr
      TYPE(dbcsr_data_obj)                               :: buff, data_a, small_buff

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (dbcsr_nblkrows_total(matrix) .NE. dbcsr_nblkcols_total(matrix) .OR. &
          dbcsr_nfullrows_total(matrix) .NE. dbcsr_nfullrows_total(matrix)) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",1140,"matrices not consistent")

      my_first_row = 1
      my_last_row = dbcsr_nfullrows_total(matrix)
      IF (PRESENT(first_row)) my_first_row = first_row
      IF (PRESENT(last_row)) my_last_row = last_row

      mynode = dbcsr_mp_mynode(dbcsr_distribution_mp(dbcsr_distribution(matrix)))
      row_blk_offsets => dbcsr_col_block_offsets(matrix)

      CALL dbcsr_work_create(matrix, work_mutable=.TRUE.)
      CALL dbcsr_data_init(buff)
      CALL dbcsr_data_init(data_a)
      CALL dbcsr_data_init(small_buff)
      CALL dbcsr_data_new(data_a, &
                          dbcsr_type_1d_to_2d(dbcsr_get_data_type(matrix)))
      CALL dbcsr_data_new(buff, &
                          dbcsr_type_1d_to_2d(dbcsr_get_data_type(matrix)), &
                          dbcsr_max_row_size(matrix), dbcsr_max_col_size(matrix))
      CALL dbcsr_data_new(small_buff, &
                          dbcsr_type_1d_to_2d(dbcsr_get_data_type(matrix)))

      DO row = 1, dbcsr_nblkrows_total(matrix)
         tr = .FALSE.
         stored_row = row
         CALL dbcsr_get_stored_coordinates(matrix, stored_row, stored_row, hold)
         IF (hold .EQ. mynode) THEN
            CALL dbcsr_get_block_p(matrix, stored_row, stored_row, data_a, tr, &
                                   found, row_size=row_size)
            offset_beg = row_blk_offsets(row)
            offset_end = row_blk_offsets(row+1)-1
            IF (my_first_row .GT. offset_end .OR. my_last_row .LT. offset_beg) CYCLE
            imin = 1
            IF (my_first_row .GT. offset_beg) THEN
               imin = my_first_row-offset_beg+1
            ENDIF
            imax = row_size
            IF (my_last_row .LT. offset_end) THEN
               imax = my_last_row-offset_end+row_size
            ENDIF
            IF (found) THEN
               CALL block_add_on_diag(data_a, alpha_scalar, row_size, &
                                      imin=imin, imax=imax)
            ELSE
               CALL dbcsr_data_set_pointer(small_buff, row_size, row_size, &
                                           buff)
               CALL dbcsr_data_clear(small_buff)
               CALL block_add_on_diag(small_buff, alpha_scalar, row_size, &
                                      imin=imin, imax=imax)
               CALL dbcsr_put_block(matrix, stored_row, stored_row, small_buff)
            ENDIF
         ENDIF
      ENDDO

      CALL dbcsr_data_clear_pointer(data_a)
      CALL dbcsr_data_clear_pointer(small_buff)
      CALL dbcsr_data_release(small_buff)
      CALL dbcsr_data_release(buff)
      CALL dbcsr_data_release(data_a)
      CALL dbcsr_finalize(matrix)
      CALL timestop(handle)
   END SUBROUTINE dbcsr_add_on_diag_anytype

! **************************************************************************************************
!> \brief ... TODO : unify with other version which is generic in the data_type
!> \param matrix ...
!> \param keep_sparsity ...
! **************************************************************************************************
   SUBROUTINE dbcsr_init_random(matrix, keep_sparsity)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      LOGICAL, OPTIONAL                                  :: keep_sparsity

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_init_random', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col, col_size, handle, hold, iseed(4), &
                                                            mynode, ncol, nrow, row, row_size, &
                                                            stored_col, stored_row
      INTEGER, DIMENSION(:), POINTER                     :: col_blk_size, row_blk_size
      LOGICAL                                            :: found, my_keep_sparsity, tr
      REAL(real_8), ALLOCATABLE, DIMENSION(:)            :: rnd
      REAL(real_8), DIMENSION(:, :), POINTER             :: buff, data_d

!   ---------------------------------------------------------------------------

      my_keep_sparsity = .FALSE.
      IF (PRESENT(keep_sparsity)) my_keep_sparsity = keep_sparsity

      CALL timeset(routineN, handle)

      row_blk_size => array_data(matrix%m%row_blk_size)
      col_blk_size => array_data(matrix%m%col_blk_size)
      mynode = dbcsr_mp_mynode(dbcsr_distribution_mp(dbcsr_distribution(matrix)))
      CALL dbcsr_work_create(matrix, work_mutable=.TRUE.)

      ALLOCATE (rnd(MAXVAL(row_blk_size)*MAXVAL(col_blk_size)))
      nrow = dbcsr_nblkrows_total(matrix)
      ncol = dbcsr_nblkcols_total(matrix)
      DO row = 1, nrow
      DO col = 1, ncol
         row_size = row_blk_size(row)
         col_size = col_blk_size(col)
         tr = .FALSE.
         stored_row = row
         stored_col = col
         CALL dbcsr_get_stored_coordinates(matrix, stored_row, stored_col, hold)
         IF (hold .EQ. mynode) THEN
            CALL dbcsr_get_block_p(matrix, stored_row, stored_col, data_d, tr, found)
            IF (found .OR. (.NOT. my_keep_sparsity)) THEN
               ! set the seed for dlarnv, is here to guarantee same value of the random numbers
               ! for all layouts (and block distributions)
               CALL set_larnv_seed(row, nrow, col, ncol, 1, iseed)
               CALL dlarnv(1, iseed, row_size*col_size, rnd(1))
            ENDIF
            IF (found) THEN
               CALL dcopy(row_size*col_size, rnd, 1, data_d, 1)
            ELSE
               IF (.NOT. my_keep_sparsity) THEN
                  ALLOCATE (buff(row_size, col_size))
                  CALL dcopy(row_size*col_size, rnd, 1, buff, 1)
                  CALL dbcsr_put_block(matrix, stored_row, stored_col, buff)
                  DEALLOCATE (buff)
               ENDIF
            ENDIF
         ENDIF
      ENDDO
      ENDDO
      DEALLOCATE (rnd)

      CALL dbcsr_finalize(matrix)
      CALL timestop(handle)

   END SUBROUTINE dbcsr_init_random

! **************************************************************************************************
!> \brief get the diagonal of a dbcsr matrix
!> \param[in] matrix    the matrix
!> \param[inout] diag   the diagonal
!>
! **************************************************************************************************
   SUBROUTINE dbcsr_get_block_diag(matrix, diag)

      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: diag

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_block_diag', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, handle, row
      LOGICAL                                            :: tr
      TYPE(dbcsr_data_obj)                               :: data_a
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      CALL dbcsr_create(diag, name='diag of '//TRIM(matrix%m%name), &
                        template=matrix)

      CALL dbcsr_data_init(data_a)
      CALL dbcsr_data_new(data_a, dbcsr_get_data_type(matrix))
      CALL dbcsr_iterator_start(iter, matrix)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, data_a, tr, blk)
         IF (row .EQ. col) CALL dbcsr_put_block(diag, row, col, data_a, transposed=tr)
      ENDDO
      CALL dbcsr_iterator_stop(iter)
      CALL dbcsr_data_clear_pointer(data_a)
      CALL dbcsr_data_release(data_a)
      CALL dbcsr_finalize(diag)
      CALL timestop(handle)
   END SUBROUTINE dbcsr_get_block_diag

! **************************************************************************************************
!> \brief get the diagonal of a dbcsr matrix
!> \param[in] matrix    the matrix
!> \param[inout] diag   diagonal
!>
! **************************************************************************************************
   SUBROUTINE dbcsr_get_diag_anytype(matrix, diag)

      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      TYPE(dbcsr_data_obj), INTENT(INOUT)                :: diag

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_diag_anytype', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, data_type, handle, row, &
                                                            row_offset, row_size
      LOGICAL                                            :: tr
      TYPE(dbcsr_data_obj)                               :: data_a, diag_a
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      CALL dbcsr_assert(dbcsr_nfullrows_total(matrix), "LE", dbcsr_data_get_size(diag), &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "Diagonal is too small", 1339)

      CALL dbcsr_data_clear(diag)
      data_type = dbcsr_get_data_type(matrix)

      CALL dbcsr_data_init(data_a)
      CALL dbcsr_data_new(data_a, &
                          dbcsr_type_1d_to_2d(data_type))
      CALL dbcsr_data_init(diag_a)
      CALL dbcsr_data_new(diag_a, data_type)

      CALL dbcsr_iterator_start(iter, matrix)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, data_a, tr, blk, &
                                        row_offset=row_offset, row_size=row_size)
         IF (row .EQ. col) THEN
            diag_a = pointer_view(diag_a, diag, offset=row_offset, len=row_size)
            CALL get_block2d_diagonal(data_a, diag_a, row_size)
         ENDIF
      ENDDO
      CALL dbcsr_iterator_stop(iter)

      CALL dbcsr_data_clear_pointer(diag_a)
      CALL dbcsr_data_release(diag_a)
      CALL dbcsr_data_clear_pointer(data_a)
      CALL dbcsr_data_release(data_a)
      CALL timestop(handle)
   END SUBROUTINE dbcsr_get_diag_anytype

! **************************************************************************************************
!> \brief set the diagonal of a dbcsr matrix
!> \param[in] matrix    the matrix
!> \param[inout] diag   diagonal
!>
! **************************************************************************************************
   SUBROUTINE dbcsr_set_diag_anytype(matrix, diag)

      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      TYPE(dbcsr_data_obj), INTENT(IN)                   :: diag

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_diag_anytype', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, dt, handle, row, row_offset, &
                                                            row_size
      LOGICAL                                            :: tr
      TYPE(dbcsr_data_obj)                               :: data_a, diag_a
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      CALL dbcsr_assert(dbcsr_nfullrows_total(matrix), "LE", dbcsr_data_get_size(diag), &
                        dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                        "Diagonal too small", 1394)

      dt = dbcsr_get_data_type(matrix)
      CALL dbcsr_data_init(data_a)
      CALL dbcsr_data_new(data_a, dbcsr_type_1d_to_2d(dt))
      CALL dbcsr_data_init(diag_a)
      CALL dbcsr_data_new(diag_a, dt)

      CALL dbcsr_iterator_start(iter, matrix)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, data_a, tr, blk, &
                                        row_offset=row_offset, row_size=row_size)
         IF (row .EQ. col) THEN
            diag_a = pointer_view(diag_a, diag, offset=row_offset, len=row_size)
            CALL set_block2d_diagonal(data_a, diag_a, row_size)
         ENDIF
      ENDDO
      CALL dbcsr_iterator_stop(iter)

      CALL dbcsr_data_clear_pointer(diag_a)
      CALL dbcsr_data_release(diag_a)
      CALL dbcsr_data_clear_pointer(data_a)
      CALL dbcsr_data_release(data_a)
      CALL timestop(handle)
   END SUBROUTINE dbcsr_set_diag_anytype

! **************************************************************************************************
!> \brief checks if matrix symmetry and data_type are consistent
!> \brief note: does not check the symmetry of the data itself
!> \param matrix_type ...
!> \param data_type ...
!> \retval symmetry_consistent ...
! **************************************************************************************************
   LOGICAL FUNCTION symmetry_consistent(matrix_type, data_type)
      CHARACTER, INTENT(IN)                    :: matrix_type
      INTEGER, INTENT(IN)                      :: data_type

      CHARACTER(len=*), PARAMETER :: routineN = 'symmetry_consistent', &
                                     routineP = moduleN//':'//routineN

      symmetry_consistent = .FALSE.

      SELECT CASE (data_type)
      CASE (dbcsr_type_real_4, dbcsr_type_real_8)
         SELECT CASE (matrix_type)
         CASE (dbcsr_type_no_symmetry, dbcsr_type_symmetric, dbcsr_type_antisymmetric)
            symmetry_consistent = .TRUE.
         END SELECT
      CASE (dbcsr_type_complex_4, dbcsr_type_complex_8)
         SELECT CASE (matrix_type)
         CASE (dbcsr_type_no_symmetry, dbcsr_type_hermitian, dbcsr_type_antihermitian)
            symmetry_consistent = .TRUE.
         END SELECT
      CASE DEFAULT
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",1448,"Invalid data type.")
      END SELECT

   END FUNCTION symmetry_consistent

! **************************************************************************************************
!> \brief checks if symmetries of two matrices are compatible for copying
!> \brief data from matrix_a(source) to matrix_b(target)
!> \param matrix_type_a ...
!> \param matrix_type_b ...
!> \retval symmetry_compatible ...
! **************************************************************************************************
   LOGICAL FUNCTION symmetry_compatible(matrix_type_a, matrix_type_b)
      CHARACTER, INTENT(IN)                    :: matrix_type_a, matrix_type_b

      CHARACTER(len=*), PARAMETER :: routineN = 'symmetry_compatible', &
                                     routineP = moduleN//':'//routineN

      symmetry_compatible = .FALSE.

      SELECT CASE (matrix_type_a)
      CASE (dbcsr_type_no_symmetry)
         SELECT CASE (matrix_type_b)
         CASE (dbcsr_type_no_symmetry)
            symmetry_compatible = .TRUE.
         END SELECT
      CASE (dbcsr_type_symmetric, dbcsr_type_hermitian)
         SELECT CASE (matrix_type_b)
         CASE (dbcsr_type_symmetric, dbcsr_type_hermitian)
            symmetry_compatible = .TRUE.
         END SELECT
      CASE (dbcsr_type_antisymmetric, dbcsr_type_antihermitian)
         SELECT CASE (matrix_type_b)
         CASE (dbcsr_type_antisymmetric, dbcsr_type_antihermitian)
            symmetry_compatible = .TRUE.
         END SELECT
      CASE DEFAULT
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",1485,"Invalid matrix type.")
      END SELECT

   END FUNCTION symmetry_compatible

! **************************************************************************************************
!> \brief copy a matrix
!> \param[inout] matrix_b       target DBCSR matrix
!> \param[in]    matrix_a       source DBCSR matrix
!> \param[in]    name           (optional) name of the new matrix
!> \param[in]    keep_sparsity  (optional) keep the target matrix sparsity;
!>                              default is False.
!> \param[in]    shallow_data   (optional) shallow data copy
!> \param[in]    keep_imaginary  (optional) when copy from complex to real,&
!>                               the default is to keep only the real part; if
!>                               this flag is set, the imaginary part is used
!> \param[in]    matrix_type     'N' for normal, 'T' for transposed, 'S' for
!>                               symmetric, and 'A' for antisymmetric
! **************************************************************************************************
   SUBROUTINE dbcsr_copy(matrix_b, matrix_a, name, keep_sparsity, &
                         shallow_data, keep_imaginary, matrix_type)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_b
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: name
      LOGICAL, INTENT(IN), OPTIONAL                      :: keep_sparsity, shallow_data, &
                                                            keep_imaginary
      CHARACTER, INTENT(IN), OPTIONAL                    :: matrix_type

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_copy', routineP = moduleN//':'//routineN

      CHARACTER                                          :: new_matrix_type, repl_type
      INTEGER                                            :: handle, new_type
      LOGICAL                                            :: keep_sparse, shallow

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (.NOT. symmetry_consistent(dbcsr_get_matrix_type(matrix_a), dbcsr_get_data_type(matrix_a))) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",1523,"Source matrix symmetry not consistent with its data type.")
      shallow = .FALSE.; IF (PRESENT(shallow_data)) shallow = shallow_data
      keep_sparse = .FALSE.
      IF (PRESENT(keep_sparsity)) keep_sparse = keep_sparsity
      IF (.NOT. dbcsr_is_initialized(matrix_b)) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",1528,"Target matrix must be initialized")
      IF (keep_sparse .AND. .NOT. dbcsr_valid_index(matrix_b)) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",1530,"Target matrix must be valid to keep its sparsity")
      IF (keep_sparse .AND. shallow) &
         CALL cp__w("dbcsr/ops/dbcsr_operations.F",1532,"Shallow copy not compatibly with sparsity retainment")
      IF (keep_sparse) THEN
         IF (PRESENT(name)) matrix_b%m%name = name
         CALL dbcsr_copy_into_existing(matrix_b, matrix_a)
      ELSE
         IF (dbcsr_valid_index(matrix_b)) THEN
            new_type = dbcsr_get_data_type(matrix_b)
            repl_type = dbcsr_get_replication_type(matrix_b)
         ELSE
            new_type = dbcsr_get_data_type(matrix_a)
            repl_type = dbcsr_get_replication_type(matrix_a)
         ENDIF
         new_matrix_type = dbcsr_get_matrix_type(matrix_a)
         IF (PRESENT(matrix_type)) THEN
            IF (.NOT. symmetry_compatible(dbcsr_get_matrix_type(matrix_a), matrix_type)) &
               CALL cp_abort(cp__l("dbcsr/ops/dbcsr_operations.F",1547), "Specified target matrix symmetry "//matrix_type// &
                             " not compatible with source matrix type "//dbcsr_get_matrix_type(matrix_a))
            new_matrix_type = matrix_type
         END IF
         IF (.NOT. symmetry_consistent(new_matrix_type, new_type)) &
            CALL cp_abort(cp__l("dbcsr/ops/dbcsr_operations.F",1552), "Target matrix symmetry "// &
                          new_matrix_type//" not consistent with its data type.")
         IF (PRESENT(name)) THEN
            CALL dbcsr_create(matrix_b, name=TRIM(name), &
                              template=matrix_a, &
                              matrix_type=new_matrix_type, &
                              data_type=new_type)
         ELSE
            CALL dbcsr_create(matrix_b, &
                              name='copy of '//TRIM(dbcsr_name(matrix_a)), &
                              data_type=new_type, &
                              matrix_type=new_matrix_type, &
                              template=matrix_a)
         ENDIF
         CALL ensure_array_size(matrix_b%m%index, ub=SIZE(matrix_a%m%index), &
                                memory_type=dbcsr_get_index_memory_type(matrix_b))
         !
         ! copy index and data
         matrix_b%m%index(1:SIZE(matrix_a%m%index)) = matrix_a%m%index(:)
         IF (.NOT. shallow) THEN
            CALL dbcsr_assert(matrix_a%m%nze, "LE", dbcsr_get_data_size(matrix_a), &
                              dbcsr_fatal_level, dbcsr_internal_error, routineN, &
                              "Source matrix sizes not consistent!", 1574)
            CALL dbcsr_data_ensure_size(matrix_b%m%data_area, &
                                        dbcsr_get_data_size(matrix_a))
            IF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_get_data_type(matrix_b)) &
               THEN
               CALL dbcsr_data_copyall(matrix_b%m%data_area, &
                                       matrix_a%m%data_area)
            ELSE
               CALL dbcsr_data_convert(matrix_b%m%data_area, &
                                       matrix_a%m%data_area, drop_real=keep_imaginary)
            ENDIF
         ELSE
            IF (dbcsr_get_data_type(matrix_a) .NE. dbcsr_get_data_type(matrix_b)) &
               CALL cp__b("dbcsr/ops/dbcsr_operations.F",1587,"Shallow copy only possible when retaining data type.")
            CALL dbcsr_switch_data_area(matrix_b, matrix_a%m%data_area)
         ENDIF
         !
         ! the row_p, col_i and blk_p ...
         CALL dbcsr_repoint_index(matrix_b%m)
         matrix_b%m%nze = matrix_a%m%nze
         matrix_b%m%nblks = matrix_b%m%nblks
         matrix_b%m%valid = .TRUE.

         matrix_b%m%sparsity_id = matrix_a%m%sparsity_id
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_copy

! **************************************************************************************************
!> \brief copy a matrix, retaining current sparsity
!> \param[inout] matrix_b       target DBCSR matrix
!> \param[in]    matrix_a       source DBCSR matrix
! **************************************************************************************************
   SUBROUTINE dbcsr_copy_into_existing(matrix_b, matrix_a)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_b
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_copy_into_existing', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col_size, data_type, dst_col, dst_row, &
                                                            handle, rel, row_size, src_col, &
                                                            src_cs, src_row, src_rs
      LOGICAL                                            :: dst_tr, making_symmetric, neg_imag, &
                                                            neg_real, src_tr
      TYPE(dbcsr_data_obj)                               :: dst_data, src_data
      TYPE(dbcsr_iterator)                               :: dst_iter, src_iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (dbcsr_get_data_type(matrix_b) .NE. dbcsr_get_data_type(matrix_a)) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",1626,"Matrices have different data types.")
      data_type = dbcsr_get_data_type(matrix_b)
      neg_real = matrix_b%m%negate_real
      neg_imag = matrix_b%m%negate_imaginary
      making_symmetric = dbcsr_has_symmetry(matrix_b) &
                         .AND. .NOT. dbcsr_has_symmetry(matrix_a)
      IF (making_symmetric) THEN
         CALL dbcsr_copy_into_existing_sym(matrix_b, matrix_a)
         CALL timestop(handle)
         RETURN
      ENDIF
      CALL dbcsr_data_init(src_data)
      CALL dbcsr_data_init(dst_data)
      CALL dbcsr_data_new(src_data, data_type)
      CALL dbcsr_data_new(dst_data, data_type)
      CALL dbcsr_iterator_start(src_iter, matrix_a)
      CALL dbcsr_iterator_start(dst_iter, matrix_b)
      ! Iterate through the blocks of the source and destination
      ! matrix. There are three possibilites: 1. copy the data for
      ! blocks present in both; 2 skip source blocks not present in the
      ! target; 3 zero blocks not present in the source.
      IF (dbcsr_iterator_blocks_left(src_iter)) THEN
         CALL dbcsr_iterator_next_block(src_iter, src_row, src_col, src_data, &
                                        src_tr)
      ELSE
         src_row = 0; src_col = 0
      ENDIF
      DO WHILE (dbcsr_iterator_blocks_left(dst_iter))
         CALL dbcsr_iterator_next_block(dst_iter, dst_row, dst_col, dst_data, &
                                        dst_tr, row_size=row_size, col_size=col_size)
         ! Now find the source position that is greater or equal to the
         ! target one. I.e, skip blocks that the target doesn't have.
         rel = pos_relation(dst_row, dst_col, src_row, src_col)
         DO WHILE (rel .EQ. 1 .AND. dbcsr_iterator_blocks_left(src_iter))
            CALL dbcsr_iterator_next_block(src_iter, src_row, src_col, &
                                           src_data, src_tr, row_size=src_rs, col_size=src_cs)
            rel = pos_relation(dst_row, dst_col, src_row, src_col)
         ENDDO
         SELECT CASE (rel)
         CASE (-1, 1)
            ! Target lags source or ran out of source
            CALL dbcsr_data_clear(dst_data)
         CASE (0)
            ! Copy the data
            IF (dbcsr_data_get_size(src_data) .NE. dbcsr_data_get_size(dst_data)) &
               CALL cp__b("dbcsr/ops/dbcsr_operations.F",1671,"Block sizes not equal!")
            IF (src_tr .EQV. dst_tr) THEN
               CALL dbcsr_data_copyall(dst_data, src_data)
            ELSE
               CALL dbcsr_block_partial_copy(dst=dst_data, dst_tr=dst_tr, &
                                             dst_rs=row_size, dst_cs=col_size, &
                                             dst_r_lb=1, dst_c_lb=1, &
                                             src=src_data, src_tr=src_tr, &
                                             src_rs=src_rs, src_cs=src_cs, &
                                             src_r_lb=1, src_c_lb=1, &
                                             nrow=row_size, ncol=col_size)
               IF (neg_real) THEN
                  CALL dbcsr_block_real_neg(dst_data, row_size, col_size)
               ENDIF
               IF (neg_imag) THEN
                  CALL dbcsr_block_conjg(dst_data, row_size, col_size)
               ENDIF
            ENDIF
         CASE default
            CALL cp__b("dbcsr/ops/dbcsr_operations.F",1690,"Trouble syncing iterators")
         END SELECT
      END DO
      CALL dbcsr_iterator_stop(src_iter)
      CALL dbcsr_iterator_stop(dst_iter)
      CALL dbcsr_data_clear_pointer(src_data)
      CALL dbcsr_data_clear_pointer(dst_data)
      CALL dbcsr_data_release(src_data)
      CALL dbcsr_data_release(dst_data)
      CALL timestop(handle)

   END SUBROUTINE dbcsr_copy_into_existing

! **************************************************************************************************
!> \brief copy a matrix, retaining current sparsity
!> \param[inout] matrix_b       target DBCSR matrix
!> \param[in]    matrix_a       source DBCSR matrix
! **************************************************************************************************
   SUBROUTINE dbcsr_copy_into_existing_sym(matrix_b, matrix_a)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_b
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_copy_into_existing_sym', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col_size, data_type, dst_col, dst_row, &
                                                            handle, row_size, src_col, src_cs, &
                                                            src_row, src_rs
      LOGICAL                                            :: dst_tr, found, neg_imag, neg_real, src_tr
      TYPE(dbcsr_data_obj)                               :: dst_data, src_data
      TYPE(dbcsr_iterator)                               :: dst_iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (dbcsr_get_data_type(matrix_b) .NE. dbcsr_get_data_type(matrix_a)) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",1726,"Matrices have different data types.")
      data_type = dbcsr_get_data_type(matrix_b)
      IF (.NOT. dbcsr_has_symmetry(matrix_b) .OR. dbcsr_has_symmetry(matrix_a)) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",1729,"Must copy from non-symmetric to symmetric matrix.")
      neg_real = matrix_b%m%negate_real
      neg_imag = matrix_b%m%negate_imaginary

      CALL dbcsr_data_init(src_data)
      CALL dbcsr_data_init(dst_data)
      CALL dbcsr_data_new(src_data, data_type)
      CALL dbcsr_data_new(dst_data, data_type)
      CALL dbcsr_iterator_start(dst_iter, matrix_b)
      ! Iterate through the blocks of the destination matrix.  For each
      ! one, try to find an appropriate source matrix block and copy it
      ! into the destination matrix.
      DO WHILE (dbcsr_iterator_blocks_left(dst_iter))
         CALL dbcsr_iterator_next_block(dst_iter, dst_row, dst_col, dst_data, &
                                        dst_tr, row_size=row_size, col_size=col_size)
         src_row = dst_row
         src_col = dst_col
         IF (checker_tr(dst_row, dst_col)) &
            CALL swap(src_row, src_col)
         CALL dbcsr_get_block_p(matrix_a, src_row, src_col, src_data, src_tr, &
                                found=found, row_size=src_rs, col_size=src_cs)
         IF (.NOT. found) THEN
            CALL dbcsr_data_clear(dst_data)
         ELSE
            IF (checker_tr(dst_row, dst_col)) THEN
               src_tr = .NOT. src_tr
               CALL swap(src_rs, src_cs)
            ENDIF
            CALL dbcsr_block_partial_copy(dst=dst_data, dst_tr=dst_tr, &
                                          dst_rs=row_size, dst_cs=col_size, &
                                          dst_r_lb=1, dst_c_lb=1, &
                                          src=src_data, src_tr=src_tr, &
                                          src_rs=src_rs, src_cs=src_cs, &
                                          src_r_lb=1, src_c_lb=1, &
                                          nrow=row_size, ncol=col_size)
            IF (neg_real .AND. checker_tr(dst_row, dst_col)) THEN
               CALL dbcsr_block_real_neg(dst_data, row_size, col_size)
            ENDIF
            IF (neg_imag .AND. checker_tr(dst_row, dst_col)) THEN
               CALL dbcsr_block_conjg(dst_data, row_size, col_size)
            ENDIF
         ENDIF
      END DO
      CALL dbcsr_iterator_stop(dst_iter)
      CALL dbcsr_data_clear_pointer(src_data)
      CALL dbcsr_data_clear_pointer(dst_data)
      CALL dbcsr_data_release(src_data)
      CALL dbcsr_data_release(dst_data)
      CALL timestop(handle)

   END SUBROUTINE dbcsr_copy_into_existing_sym

! **************************************************************************************************
!> \brief Determines the relation between two matrix positions.
!> \param row1 ...
!> \param col1 ...
!> \param row2 ...
!> \param col2 ...
!> \retval relation  Relation between positions 1 and 2.
!>                    0: same
!>                   -1: pos1 < pos2
!>                    1: pos1 > pos2
! **************************************************************************************************
   ELEMENTAL FUNCTION pos_relation(row1, col1, row2, col2) RESULT(relation)
      INTEGER, INTENT(IN)                                :: row1, col1, row2, col2
      INTEGER                                            :: relation

      IF (row1 .LT. row2) THEN
         relation = -1
      ELSEIF (row1 .GT. row2) THEN
         relation = 1
      ELSE ! rows are equal, check column
         IF (col1 .LT. col2) THEN
            relation = -1
         ELSEIF (col1 .GT. col2) THEN
            relation = 1
         ELSE
            relation = 0
         ENDIF
      ENDIF
   END FUNCTION pos_relation

! **************************************************************************************************
!> \brief Copy a submatrix.
!> \param[inout] matrix_b       target DBCSR matrix
!> \param[in]    matrix_a       source DBCSR matrix
!> \param[in]    name           (optional) name of the new matrix
!> \param[in]    block_row_bounds     (optional) rows to extract (array of
!>                                    size 2 holding the lower and upper
!>                                    inclusive bounds)
!> \param[in]    block_column_bounds  (optional) columns to extract (array of
!>                                    size 2 holding the lower and upper
!>                                    inclusive bounds)
!> \param[in]    shallow_data   (optional) shallow data copy
! **************************************************************************************************
   SUBROUTINE dbcsr_copy_submatrix(matrix_b, matrix_a, name, &
                                   block_row_bounds, block_column_bounds, &
                                   shallow_data)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_b
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: name
      INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL        :: block_row_bounds, block_column_bounds
      LOGICAL, INTENT(IN), OPTIONAL                      :: shallow_data

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_copy_submatrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk_p, col, handle, nblocks, new_blk, &
                                                            old_blk, row
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: blkp_list, col_list, row_list
      LOGICAL                                            :: shallow, tr
      TYPE(dbcsr_data_obj)                               :: data_block
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (PRESENT(shallow_data)) THEN
         shallow = shallow_data
      ELSE
         shallow = .FALSE.
      ENDIF
      ! Verify assumptions.
      IF (PRESENT(block_row_bounds)) THEN
         CALL dbcsr_assert(SIZE(block_row_bounds), "EQ", 2, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "Size of bounds specifier must be 2", 1855)
      ENDIF
      IF (PRESENT(block_column_bounds)) THEN
         CALL dbcsr_assert(SIZE(block_column_bounds), "EQ", 2, &
                           dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "Size of bounds specifier must be 2", 1861)
      ENDIF
      ! Setup target matrix
      CALL dbcsr_create(matrix_b, name=name, template=matrix_a)
      CALL dbcsr_finalize(matrix_b)
      IF (.NOT. shallow) THEN
         ! Non-shallow copy uses the standard iterator on the source and
         ! block put on the target.
!
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP          PRIVATE (data_block, iter, row, col, tr) &
!$OMP          SHARED (matrix_a, matrix_b,&
!$OMP                  block_row_bounds, block_column_bounds)
         CALL dbcsr_work_create(matrix_b, work_mutable=.FALSE.)
         CALL dbcsr_data_init(data_block)
         CALL dbcsr_data_new(data_block, dbcsr_get_data_type(matrix_a))
         CALL dbcsr_iterator_start(iter, matrix_a, dynamic=.TRUE., &
                                   dynamic_byrows=.TRUE.)
         DO WHILE (dbcsr_iterator_blocks_left(iter))
            CALL dbcsr_iterator_next_block(iter, row, col, data_block, tr)
            ! Only keep the block if they are within the specified bounds.
            IF (PRESENT(block_row_bounds)) THEN
               IF (row .LT. block_row_bounds(1)) CYCLE
               IF (row .GT. block_row_bounds(2)) CYCLE
            ENDIF
            IF (PRESENT(block_column_bounds)) THEN
               IF (col .LT. block_column_bounds(1)) CYCLE
               IF (col .GT. block_column_bounds(2)) CYCLE
            ENDIF
            CALL dbcsr_put_block(matrix_b, row, col, data_block, transposed=tr)
         END DO
         CALL dbcsr_iterator_stop(iter)
         CALL dbcsr_data_clear_pointer(data_block)
         CALL dbcsr_data_release(data_block)
         CALL dbcsr_finalize(matrix_b)
!$OMP END PARALLEL
      ELSE
         ! For the shallow copy the source matrix data is referenced.
         CALL dbcsr_switch_data_area(matrix_b, matrix_a%m%data_area)
         nblocks = dbcsr_get_num_blocks(matrix_a) ! High estimate.
         ! Shallow copy goes through source's data blocks and inserts
         ! the only the ones corresponding to the submatrix specifier
         ! into the target. Block pointers must remain the same as in
         ! the source.
         ALLOCATE (row_list(nblocks), col_list(nblocks), blkp_list(nblocks))
         !
         CALL dbcsr_iterator_start(iter, matrix_a)
         new_blk = 1
         DO WHILE (dbcsr_iterator_blocks_left(iter))
            CALL dbcsr_iterator_next_block(iter, row, col, &
                                           blk=old_blk, blk_p=blk_p)
            ! Only keep the block if they are within the specified bounds.
            IF (PRESENT(block_row_bounds)) THEN
               IF (row .LT. block_row_bounds(1)) CYCLE
               IF (row .GT. block_row_bounds(2)) CYCLE
            ENDIF
            IF (PRESENT(block_column_bounds)) THEN
               IF (col .LT. block_column_bounds(1)) CYCLE
               IF (col .GT. block_column_bounds(2)) CYCLE
            ENDIF
            row_list(new_blk) = row
            col_list(new_blk) = col
            blkp_list(new_blk) = blk_p
            new_blk = new_blk+1
         ENDDO
         new_blk = new_blk-1
         CALL dbcsr_iterator_stop(iter)
         CALL dbcsr_reserve_blocks(matrix_b, row_list(1:new_blk), &
                                   col_list(1:new_blk), blkp_list(1:new_blk))
      ENDIF
      !
      CALL timestop(handle)
   END SUBROUTINE dbcsr_copy_submatrix

! **************************************************************************************************
!> \brief Crop and copies a matrix.
!> \param[inout] matrix_b       target DBCSR matrix
!> \param[in]    matrix_a       source DBCSR matrix
!> \param[in]    full_row_bounds      (optional) rows to extract (array of
!>                                    size 2 holding the lower and upper
!>                                    inclusive bounds)
!> \param[in]    full_column_bounds   (optional) columns to extract (array of
!>                                    size 2 holding the lower and upper
!>                                    inclusive bounds)
!> \param shallow_data ...
! **************************************************************************************************
   SUBROUTINE dbcsr_crop_matrix(matrix_b, matrix_a, &
                                full_row_bounds, full_column_bounds, &
                                shallow_data)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_b
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix_a
      INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL        :: full_row_bounds, full_column_bounds
      LOGICAL, INTENT(IN), OPTIONAL                      :: shallow_data

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_crop_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: col, f_col_f, f_row_f, handle, l_col_l, &
                                                            l_row_l, row
      INTEGER, DIMENSION(2)                              :: block_col_bounds, block_row_bounds
      LOGICAL                                            :: part_col, part_f_col, part_f_row, &
                                                            part_l_col, part_l_row, part_row, &
                                                            shallow, tr
      TYPE(dbcsr_data_obj)                               :: data_block
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      IF (PRESENT(shallow_data)) THEN
         shallow = shallow_data
      ELSE
         shallow = .FALSE.
      ENDIF
      block_row_bounds = 0
      block_col_bounds = 0
      part_col = .FALSE.
      part_row = .FALSE.
      !
      ! If row bounds are present, they must be converted to block
      ! addressing.
      IF (PRESENT(full_row_bounds)) THEN
         CALL dbcsr_assert(SIZE(full_row_bounds), "EQ", 2, dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "Size of bounds specifier must be 2", 1985)
         CALL dbcsr_assert(full_row_bounds(1), "GE", 0, &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Invalid first row bound.", 1988)
         CALL dbcsr_assert(full_row_bounds(2), "LE", dbcsr_nfullrows_total(matrix_a), &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Invalid last row bound.", 1991)
         part_f_row = .FALSE.
         IF (full_row_bounds(1) .EQ. 0) THEN
            block_row_bounds(1) = 1
         ELSE
            CALL find_block_of_element(full_row_bounds(1), block_row_bounds(1), &
                                       dbcsr_nblkrows_total(matrix_a), &
                                       dbcsr_row_block_offsets(matrix_a), &
                                       hint=0)
            part_f_row = array_get(dbcsr_row_block_offsets(matrix_a), block_row_bounds(1)) &
                         .NE. full_row_bounds(1)
         ENDIF
         f_row_f = -7
         IF (part_f_row) THEN
            ! Block offset of last cleared row
            f_row_f = full_row_bounds(1)- &
                      array_get(dbcsr_row_block_offsets(matrix_a), block_row_bounds(1))
         ENDIF
         part_l_row = .FALSE.
         IF (full_row_bounds(2) .EQ. 0) THEN
            block_row_bounds(2) = dbcsr_nblkrows_total(matrix_a)
         ELSE
            CALL find_block_of_element(full_row_bounds(2), block_row_bounds(2), &
                                       dbcsr_nblkrows_total(matrix_a), &
                                       dbcsr_row_block_offsets(matrix_a), &
                                       hint=0)
            part_l_row = array_get(dbcsr_row_block_offsets(matrix_a), block_row_bounds(2)+1)-1 &
                         .NE. full_row_bounds(2)
         ENDIF
         ! Block offset of first cleared row
         l_row_l = -7
         IF (part_l_row) THEN
            l_row_l = 2+full_row_bounds(2)- &
                      array_get(dbcsr_row_block_offsets(matrix_a), block_row_bounds(2))
         ENDIF
         part_row = part_f_row .OR. part_l_row
      ENDIF
      !
      ! If column bounds are present, they must be converted to block
      ! addressing.
      IF (PRESENT(full_column_bounds)) THEN
         CALL dbcsr_assert(SIZE(full_column_bounds), "EQ", 2, &
                           dbcsr_fatal_level, &
                           dbcsr_wrong_args_error, routineN, &
                           "Size of bounds specifier must be 2", 2035)
         CALL dbcsr_assert(full_column_bounds(1), "GE", 0, &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Invalid first column bound.", 2038)
         CALL dbcsr_assert(full_column_bounds(2), "LE", dbcsr_nfullcols_total(matrix_a), &
                           dbcsr_fatal_level, dbcsr_wrong_args_error, routineN, &
                           "Invalid last column bound.", 2041)
         part_f_col = .FALSE.
         IF (full_column_bounds(1) .EQ. 0) THEN
            block_col_bounds(1) = 1
         ELSE
            CALL find_block_of_element(full_column_bounds(1), block_col_bounds(1), &
                                       dbcsr_nblkcols_total(matrix_a), &
                                       dbcsr_col_block_offsets(matrix_a), &
                                       hint=0)
            part_f_col = array_get(dbcsr_col_block_offsets(matrix_a), block_col_bounds(1)) &
                         .NE. full_column_bounds(1)
         ENDIF
         f_col_f = -7
         IF (part_f_col) THEN
            ! Block offset of last cleared column
            f_col_f = full_column_bounds(1)- &
                      array_get(dbcsr_col_block_offsets(matrix_a), block_col_bounds(1))
         ENDIF
         part_l_col = .FALSE.
         IF (full_column_bounds(2) .EQ. 0) THEN
            block_col_bounds(2) = dbcsr_nblkcols_total(matrix_a)
         ELSE
            CALL find_block_of_element(full_column_bounds(2), block_col_bounds(2), &
                                       dbcsr_nblkcols_total(matrix_a), &
                                       dbcsr_col_block_offsets(matrix_a), &
                                       hint=0)
            part_l_col = array_get(dbcsr_col_block_offsets(matrix_a), block_col_bounds(2)+1)-1 &
                         .NE. full_column_bounds(2)
         ENDIF
         l_col_l = -7
         IF (part_l_col) THEN
            ! Block offset of first cleared column
            l_col_l = 2+full_column_bounds(2)- &
                      array_get(dbcsr_col_block_offsets(matrix_a), block_col_bounds(2))
         ENDIF
         part_col = part_f_col .OR. part_l_col
      ENDIF
      !
      ! First copy the blocks then perform the intra-block zeroing.
      CALL dbcsr_copy_submatrix(matrix_b, matrix_a, &
                                block_row_bounds=block_row_bounds, &
                                block_column_bounds=block_col_bounds, &
                                shallow_data=shallow)
      IF (part_row .OR. part_col) THEN
!$OMP PARALLEL DEFAULT (NONE) &
!$OMP          PRIVATE (data_block, iter, row, col, tr) &
!$OMP          SHARED (matrix_b,&
!$OMP                  part_row, part_f_row, part_l_row, f_row_f, l_row_l, &
!$OMP                  part_col, part_f_col, part_l_col, f_col_f, l_col_l,&
!$OMP                  block_row_bounds, block_col_bounds)
         CALL dbcsr_data_init(data_block)
         CALL dbcsr_data_new(data_block, dbcsr_type_1d_to_2d(dbcsr_get_data_type(matrix_b)))
         CALL dbcsr_iterator_start(iter, matrix_b, &
                                   dynamic=.TRUE., dynamic_byrows=.TRUE.)
         DO WHILE (dbcsr_iterator_blocks_left(iter))
            CALL dbcsr_iterator_next_block(iter, row, col, data_block, tr)
            IF (part_row) THEN
               IF (row .LT. block_row_bounds(1)) CYCLE
               IF (row .GT. block_row_bounds(2)) CYCLE
            ENDIF
            IF (part_col) THEN
               IF (col .LT. block_col_bounds(1)) CYCLE
               IF (col .GT. block_col_bounds(2)) CYCLE
            ENDIF
            IF (part_row) THEN
               IF (part_f_row .AND. row .EQ. block_row_bounds(1)) THEN
                  CALL dbcsr_data_clear(data_block, ub=f_row_f, tr=tr)
               ENDIF
               IF (part_l_row .AND. row .EQ. block_row_bounds(2)) THEN
                  CALL dbcsr_data_clear(data_block, lb=l_row_l, tr=tr)
               ENDIF
            ENDIF
            IF (part_col) THEN
               IF (part_f_col .AND. col .EQ. block_col_bounds(1)) THEN
                  CALL dbcsr_data_clear(data_block, ub2=f_col_f, tr=tr)
               ENDIF
               IF (part_l_col .AND. col .EQ. block_col_bounds(2)) THEN
                  CALL dbcsr_data_clear(data_block, lb2=l_col_l, tr=tr)
               ENDIF
            ENDIF
         END DO
         CALL dbcsr_iterator_stop(iter)
         CALL dbcsr_data_clear_pointer(data_block)
         CALL dbcsr_data_release(data_block)
         CALL dbcsr_finalize(matrix_b)
!$OMP END PARALLEL
      ENDIF
      !
      CALL timestop(handle)
   END SUBROUTINE dbcsr_crop_matrix

! **************************************************************************************************
!> \brief  triu of a dbcsr matrix
!> \param[inout] matrix_a  the matrix
!>
! **************************************************************************************************
   SUBROUTINE dbcsr_triu(matrix_a)

      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_triu', routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, blk_nze, col, col_size, handle, i, &
                                                            j, row, row_size
      LOGICAL                                            :: tr
      REAL(dp), DIMENSION(:, :), POINTER                 :: DATA
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      CALL dbcsr_iterator_start(iter, matrix_a)

      DO WHILE (dbcsr_iterator_blocks_left(iter))
         CALL dbcsr_iterator_next_block(iter, row, col, DATA, tr, &
                                        block_number=blk, row_size=row_size, col_size=col_size)
         blk_nze = row_size*col_size
         IF (row .GT. col) CALL dbcsr_remove_block(matrix_a, row, col, blk_nze, blk)
         IF (row .EQ. col) THEN
            DO j = 1, col_size
            DO i = j+1, row_size
               DATA(i, j) = 0.0_dp
            ENDDO
            ENDDO
         ENDIF
      ENDDO

      CALL dbcsr_iterator_stop(iter)

      CALL dbcsr_finalize(matrix_a)
      CALL timestop(handle)
   END SUBROUTINE dbcsr_triu

! **************************************************************************************************
!> \brief filter a dbcsr matrix
!> \param[inout] matrix the matrix
!> \param[in] eps          the threshold
!> \param[in] method       how the matrix is filtered
!> \param[in] use_absolute ?
!> \param[in] filter_diag  NYI
!> \param[in] quick        (optional) filter just the index (no data copying)
! **************************************************************************************************
   SUBROUTINE dbcsr_filter_anytype(matrix, eps, method, &
                                   use_absolute, filter_diag, quick)

      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      TYPE(dbcsr_scalar_type), INTENT(IN)                :: eps
      INTEGER, INTENT(IN), OPTIONAL                      :: method
      LOGICAL, INTENT(in), OPTIONAL                      :: use_absolute, filter_diag, quick

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_filter_anytype', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=real_4), DIMENSION(:), POINTER        :: data_c
      COMPLEX(KIND=real_8), DIMENSION(:), POINTER        :: data_z
      INTEGER                                            :: blk, blk_nze, col, col_size, handle, &
                                                            my_method, row, row_size
      LOGICAL                                            :: gt0, my_filter_diag, reshuffle, tr
      REAL(KIND=real_4)                                  :: nrm_s, sdot
      REAL(KIND=real_4), DIMENSION(:), POINTER           :: data_s
      REAL(KIND=real_8)                                  :: ddot, my_absolute, nrm_d
      REAL(KIND=real_8), DIMENSION(:), POINTER           :: data_d
      REAL(KIND=real_8), EXTERNAL                        :: DZNRM2, SCNRM2
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      my_method = dbcsr_filter_frobenius
      IF (PRESENT(method)) my_method = method
      my_absolute = 1.0_dp
      IF (PRESENT(use_absolute)) my_absolute = dbcsr_maxabs(matrix)
      my_filter_diag = .TRUE.
      IF (PRESENT(filter_diag)) my_filter_diag = filter_diag
      reshuffle = .TRUE.
      IF (PRESENT(quick)) reshuffle = .NOT. quick

      SELECT CASE (eps%data_type)
      CASE (dbcsr_type_real_4)
         gt0 = eps%r_sp .GT. 0.0_real_4
      CASE (dbcsr_type_real_8)
         gt0 = eps%r_dp .GT. 0.0_real_8
      CASE (dbcsr_type_complex_4)
         gt0 = ABS(eps%c_sp) .GT. 0.0_real_4
      CASE (dbcsr_type_complex_8)
         gt0 = ABS(eps%c_dp) .GT. 0.0_real_8
      CASE default
         gt0 = .FALSE.
      END SELECT

      IF (gt0) THEN

!$OMP        PARALLEL DEFAULT(NONE) PRIVATE(iter,row,col,data_s,data_d,data_c,data_z,tr, &
!$OMP                                       blk,row_size,col_size,blk_nze,nrm_d,nrm_s) &
!$OMP                               SHARED(my_method,reshuffle,my_absolute,eps,matrix)

         CALL dbcsr_iterator_start(iter, matrix, contiguous_pointers=.TRUE.)
         DO WHILE (dbcsr_iterator_blocks_left(iter))
            SELECT CASE (dbcsr_get_data_type (matrix))

            CASE (dbcsr_type_real_4)
               CALL dbcsr_iterator_next_block(iter, row, col, data_s, tr, blk, &
                                              row_size, col_size)
               blk_nze = row_size*col_size
               IF (blk_nze .EQ. 0) CYCLE ! Skip empty blocks
               SELECT CASE (my_method)
               CASE (dbcsr_filter_frobenius)
                  !
                  ! Frobenius based
                  nrm_s = SQRT(SDOT(SIZE(data_s), data_s(1), 1, data_s(1), 1))
                  IF (nrm_s .LT. my_absolute*eps%r_sp) &
                     CALL dbcsr_remove_block(matrix, row, col, blk_nze, blk)
               CASE DEFAULT
                  CALL cp__b("dbcsr/ops/dbcsr_operations.F",2254,"Only Frobenius based filtering")
               END SELECT

            CASE (dbcsr_type_real_8)
               CALL dbcsr_iterator_next_block(iter, row, col, data_d, tr, blk, &
                                              row_size, col_size)
               blk_nze = row_size*col_size
               IF (blk_nze .EQ. 0) CYCLE ! Skip empty blocks
               SELECT CASE (my_method)
               CASE (dbcsr_filter_frobenius)
                  !
                  ! Frobenius based
                  nrm_d = SQRT(DDOT(SIZE(data_d), data_d(1), 1, data_d(1), 1))
                  IF (nrm_d .LT. my_absolute*eps%r_dp) &
                     CALL dbcsr_remove_block(matrix, row, col, blk_nze, blk)
               CASE DEFAULT
                  CALL cp__b("dbcsr/ops/dbcsr_operations.F",2270,"Only Frobenius based filtering")
               END SELECT

            CASE (dbcsr_type_complex_4)
               CALL dbcsr_iterator_next_block(iter, row, col, data_c, tr, blk, &
                                              row_size, col_size)
               blk_nze = row_size*col_size
               IF (blk_nze .EQ. 0) CYCLE ! Skip empty blocks
               SELECT CASE (my_method)
               CASE (dbcsr_filter_frobenius)
                  !
                  ! Frobenius based
                  nrm_d = SCNRM2(SIZE(data_c), data_c(1), 1)
                  IF (nrm_d .LT. my_absolute*eps%r_dp) &
                     CALL dbcsr_remove_block(matrix, row, col, blk_nze, blk)
               CASE DEFAULT
                  CALL cp__b("dbcsr/ops/dbcsr_operations.F",2286,"Only Frobenius based filtering")
               END SELECT

            CASE (dbcsr_type_complex_8)
               CALL dbcsr_iterator_next_block(iter, row, col, data_z, tr, blk, &
                                              row_size, col_size)
               blk_nze = row_size*col_size
               IF (blk_nze .EQ. 0) CYCLE ! Skip empty blocks
               SELECT CASE (my_method)
               CASE (dbcsr_filter_frobenius)
                  !
                  ! Frobenius based
                  nrm_d = DZNRM2(SIZE(data_z), data_z(1), 1)
                  IF (nrm_d .LT. my_absolute*eps%r_dp) &
                     CALL dbcsr_remove_block(matrix, row, col, blk_nze, blk)
               CASE DEFAULT
                  CALL cp__b("dbcsr/ops/dbcsr_operations.F",2302,"Only Frobenius based filtering")
               END SELECT

            CASE DEFAULT
               CALL cp__b("dbcsr/ops/dbcsr_operations.F",2306,"Wrong data type")
            END SELECT
         ENDDO
         CALL dbcsr_iterator_stop(iter)
         CALL dbcsr_finalize(matrix, reshuffle=reshuffle)
!$OMP        END PARALLEL

         IF (reshuffle) CALL dbcsr_index_compact(matrix)

      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_filter_anytype

! **************************************************************************************************
!> \brief compute a norm of a dbcsr matrix
!> \param[in] matrix  the matrix
!> \param[in] which_norm ...
!>
!> \param norm_scalar ...
!> \param norm_vector ...
! **************************************************************************************************
   SUBROUTINE dbcsr_norm_anytype(matrix, which_norm, norm_scalar, norm_vector)

      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      INTEGER, INTENT(IN)                                :: which_norm
      REAL(KIND=real_8), INTENT(OUT), OPTIONAL           :: norm_scalar
      TYPE(dbcsr_data_obj), INTENT(INOUT), OPTIONAL      :: norm_vector

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_norm_anytype', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: blk, col, col_offset, handle, i, j, row, &
                                                            row_offset
      LOGICAL                                            :: tr
      TYPE(dbcsr_data_obj)                               :: data_a
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      SELECT CASE (which_norm)

      CASE (dbcsr_norm_frobenius)

         IF (PRESENT(norm_scalar)) norm_scalar = dbcsr_frobenius_norm(matrix)

      CASE (dbcsr_norm_maxabsnorm)

         IF (PRESENT(norm_scalar)) norm_scalar = dbcsr_maxabs(matrix)

      CASE (dbcsr_norm_gershgorin)

         IF (PRESENT(norm_scalar)) norm_scalar = dbcsr_gershgorin_norm(matrix)

      CASE (dbcsr_norm_column)

         IF (PRESENT(norm_vector)) THEN
            CALL dbcsr_assert(dbcsr_data_get_type(norm_vector), "EQ", &
                              dbcsr_get_data_type(matrix), dbcsr_fatal_level, &
                              dbcsr_wrong_args_error, routineN, &
                              "Mismatched vector/matrix data types", 2367)
            IF (dbcsr_has_symmetry(matrix)) THEN
               CALL dbcsr_assert(dbcsr_data_get_size(norm_vector), "GE", &
                                 dbcsr_nfullrows_total(matrix), dbcsr_fatal_level, &
                                 dbcsr_wrong_args_error, routineN, "Passed vector too small", &
                                 2372)
            END IF
            CALL dbcsr_assert(dbcsr_data_get_size(norm_vector), "GE", &
                              dbcsr_nfullcols_total(matrix), dbcsr_fatal_level, &
                              dbcsr_wrong_args_error, routineN, "Passed vector too small", &
                              2377)
            CALL dbcsr_data_init(data_a)
            CALL dbcsr_data_new(data_a, dbcsr_type_1d_to_2d(dbcsr_get_data_type(matrix)))
            CALL dbcsr_data_clear(norm_vector)
            CALL dbcsr_iterator_start(iter, matrix)
            DO WHILE (dbcsr_iterator_blocks_left(iter))
               CALL dbcsr_iterator_next_block(iter, row, col, data_a, tr, &
                                              blk, row_offset=row_offset, col_offset=col_offset)
               SELECT CASE (dbcsr_get_data_type (matrix))
               CASE (dbcsr_type_real_4)
                  IF (dbcsr_has_symmetry(matrix) .AND. row .NE. col) THEN
                     DO j = 1, SIZE(data_a%d%r2_sp, 2)
                     DO i = 1, SIZE(data_a%d%r2_sp, 1)
                        norm_vector%d%r_sp(col_offset+j-1) &
                           = norm_vector%d%r_sp(col_offset+j-1) &
                             +data_a%d%r2_sp(i, j)**2
                        norm_vector%d%r_sp(row_offset+i-1) &
                           = norm_vector%d%r_sp(row_offset+i-1) &
                             +data_a%d%r2_sp(i, j)**2
                     ENDDO
                     ENDDO
                  ELSE
                     DO j = 1, SIZE(data_a%d%r2_sp, 2)
                     DO i = 1, SIZE(data_a%d%r2_sp, 1)
                        norm_vector%d%r_sp(col_offset+j-1) &
                           = norm_vector%d%r_sp(col_offset+j-1) &
                             +data_a%d%r2_sp(i, j)*data_a%d%r2_sp(i, j)
                     ENDDO
                     ENDDO
                  ENDIF
               CASE (dbcsr_type_real_8)
                  IF (dbcsr_has_symmetry(matrix) .AND. row .NE. col) THEN
                     DO j = 1, SIZE(data_a%d%r2_dp, 2)
                     DO i = 1, SIZE(data_a%d%r2_dp, 1)
                        norm_vector%d%r_dp(col_offset+j-1) &
                           = norm_vector%d%r_dp(col_offset+j-1) &
                             +data_a%d%r2_dp(i, j)**2
                        norm_vector%d%r_dp(row_offset+i-1) &
                           = norm_vector%d%r_dp(row_offset+i-1) &
                             +data_a%d%r2_dp(i, j)**2
                     ENDDO
                     ENDDO
                  ELSE
                     DO j = 1, SIZE(data_a%d%r2_dp, 2)
                     DO i = 1, SIZE(data_a%d%r2_dp, 1)
                        norm_vector%d%r_dp(col_offset+j-1) &
                           = norm_vector%d%r_dp(col_offset+j-1) &
                             +data_a%d%r2_dp(i, j)*data_a%d%r2_dp(i, j)
                     ENDDO
                     ENDDO
                  ENDIF
                  !CASE (dbcsr_type_complex_4)
                  !CASE (dbcsr_type_complex_8)
               CASE DEFAULT
                  CALL cp__b("dbcsr/ops/dbcsr_operations.F",2431,"Only real values")
               END SELECT
            ENDDO
            CALL dbcsr_iterator_stop(iter)
            CALL dbcsr_data_clear_pointer(data_a)
            CALL dbcsr_data_release(data_a)
            SELECT CASE (dbcsr_get_data_type (matrix))
            CASE (dbcsr_type_real_4)
               CALL mp_sum(norm_vector%d%r_sp, &
                           dbcsr_mp_group(dbcsr_distribution_mp(matrix%m%dist)))
               norm_vector%d%r_sp = SQRT(norm_vector%d%r_sp)
            CASE (dbcsr_type_real_8)
               CALL mp_sum(norm_vector%d%r_dp, &
                           dbcsr_mp_group(dbcsr_distribution_mp(matrix%m%dist)))
               norm_vector%d%r_dp = SQRT(norm_vector%d%r_dp)
            END SELECT
         ENDIF

      CASE DEFAULT

         CALL cp__b("dbcsr/ops/dbcsr_operations.F",2451,"this norm is NYI")

      END SELECT
      CALL timestop(handle)
   END SUBROUTINE dbcsr_norm_anytype

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param which_norm ...
!> \param norm_vector ...
! **************************************************************************************************
   SUBROUTINE dbcsr_norm_r8_vec(matrix, which_norm, norm_vector)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      INTEGER, INTENT(IN)                                :: which_norm
      REAL(KIND=real_8), DIMENSION(:), INTENT(OUT), &
         TARGET                                          :: norm_vector

      REAL(KIND=real_8), DIMENSION(:), POINTER           :: v_p
      TYPE(dbcsr_data_obj)                               :: norm_vector_a

      CALL dbcsr_data_init(norm_vector_a)
      CALL dbcsr_data_new(norm_vector_a, dbcsr_type_real_8)
      v_p => norm_vector
      CALL dbcsr_data_set_pointer(norm_vector_a, v_p)
      CALL dbcsr_norm_anytype(matrix, which_norm, norm_vector=norm_vector_a)
      CALL dbcsr_data_clear_pointer(norm_vector_a)
      CALL dbcsr_data_release(norm_vector_a)
   END SUBROUTINE dbcsr_norm_r8_vec

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param which_norm ...
!> \param norm_scalar ...
! **************************************************************************************************
   SUBROUTINE dbcsr_norm_r4_scal(matrix, which_norm, norm_scalar)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      INTEGER, INTENT(IN)                                :: which_norm
      REAL(KIND=real_4), INTENT(OUT)                     :: norm_scalar

      REAL(KIND=real_8)                                  :: norm_r8

      CALL dbcsr_norm_anytype(matrix, which_norm, norm_scalar=norm_r8)
      norm_scalar = REAL(norm_r8, KIND=real_4)
   END SUBROUTINE dbcsr_norm_r4_scal

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param which_norm ...
!> \param norm_vector ...
! **************************************************************************************************
   SUBROUTINE dbcsr_norm_r4_vec(matrix, which_norm, norm_vector)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      INTEGER, INTENT(IN)                                :: which_norm
      REAL(KIND=real_4), DIMENSION(:), INTENT(OUT), &
         TARGET                                          :: norm_vector

      REAL(KIND=real_4), DIMENSION(:), POINTER           :: v_p
      TYPE(dbcsr_data_obj)                               :: norm_vector_a

      CALL dbcsr_data_init(norm_vector_a)
      CALL dbcsr_data_new(norm_vector_a, dbcsr_type_real_4)
      v_p => norm_vector
      CALL dbcsr_data_set_pointer(norm_vector_a, v_p)
      CALL dbcsr_norm_anytype(matrix, which_norm, norm_vector=norm_vector_a)
      CALL dbcsr_data_clear_pointer(norm_vector_a)
      CALL dbcsr_data_release(norm_vector_a)
   END SUBROUTINE dbcsr_norm_r4_vec

! **************************************************************************************************
!> \brief compute a norm of a dbcsr matrix
!> \param[in] matrix  the matrix
!> \retval norm ...
! **************************************************************************************************
   FUNCTION dbcsr_gershgorin_norm_r8(matrix) RESULT(norm)

      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      REAL(KIND=real_8)                                  :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_gershgorin_norm_r8', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=real_4), DIMENSION(:, :), POINTER     :: data_c
      COMPLEX(KIND=real_8), DIMENSION(:, :), POINTER     :: data_z
      INTEGER                                            :: blk, col, col_offset, i, j, nc, nr, row, &
                                                            row_offset
      LOGICAL                                            :: any_sym, tr
      REAL(KIND=real_4), DIMENSION(:, :), POINTER        :: data_r
      REAL(KIND=real_8), DIMENSION(:, :), POINTER        :: data_d
      REAL(real_8), ALLOCATABLE, DIMENSION(:)            :: buff_d
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      nr = dbcsr_nfullrows_total(matrix)
      nc = dbcsr_nfullcols_total(matrix)

      any_sym = dbcsr_get_matrix_type(matrix) .EQ. dbcsr_type_symmetric .OR. &
                dbcsr_get_matrix_type(matrix) .EQ. dbcsr_type_antisymmetric

      IF (nr .NE. nc) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",2554,"not a square matrix")

      norm = 0.0_dp
      ALLOCATE (buff_d(nr))
      buff_d = 0.0_dp
      CALL dbcsr_iterator_start(iter, matrix)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         SELECT CASE (dbcsr_get_data_type (matrix))
         CASE (dbcsr_type_real_4)
            CALL dbcsr_iterator_next_block(iter, row, col, data_r, tr, blk, &
                                           row_offset=row_offset, col_offset=col_offset)
            DO j = 1, SIZE(data_r, 2)
            DO i = 1, SIZE(data_r, 1)
               buff_d(row_offset+i-1) = buff_d(row_offset+i-1)+ABS(data_r(i, j))
               IF (any_sym .AND. row .NE. col) &
                  buff_d(col_offset+j-1) = buff_d(col_offset+j-1)+ABS(data_r(i, j))
            ENDDO
            ENDDO
         CASE (dbcsr_type_real_8)
            CALL dbcsr_iterator_next_block(iter, row, col, data_d, tr, blk, &
                                           row_offset=row_offset, col_offset=col_offset)
            DO j = 1, SIZE(data_d, 2)
            DO i = 1, SIZE(data_d, 1)
               buff_d(row_offset+i-1) = buff_d(row_offset+i-1)+ABS(data_d(i, j))
               IF (any_sym .AND. row .NE. col) &
                  buff_d(col_offset+j-1) = buff_d(col_offset+j-1)+ABS(data_d(i, j))
            ENDDO
            ENDDO
         CASE (dbcsr_type_complex_4)
            CALL dbcsr_iterator_next_block(iter, row, col, data_c, tr, blk, &
                                           row_offset=row_offset, col_offset=col_offset)
            DO j = 1, SIZE(data_c, 2)
            DO i = 1, SIZE(data_c, 1)
               buff_d(row_offset+i-1) = buff_d(row_offset+i-1)+ABS(data_c(i, j))
               IF (any_sym .AND. row .NE. col) &
                  CALL cp__b("dbcsr/ops/dbcsr_operations.F",2589,"Only nonsymmetric matrix so far")
               !     buff_d(col_offset+j-1) = buff_d(col_offset+j-1) + ABS(data_c(i,j))
            ENDDO
            ENDDO
         CASE (dbcsr_type_complex_8)
            CALL dbcsr_iterator_next_block(iter, row, col, data_z, tr, blk, &
                                           row_offset=row_offset, col_offset=col_offset)
            DO j = 1, SIZE(data_z, 2)
            DO i = 1, SIZE(data_z, 1)
               buff_d(row_offset+i-1) = buff_d(row_offset+i-1)+ABS(data_z(i, j))
               IF (any_sym .AND. row .NE. col) &
                  CALL cp__b("dbcsr/ops/dbcsr_operations.F",2600,"Only nonsymmetric matrix so far")
               !     buff_d(col_offset+j-1) = buff_d(col_offset+j-1) + ABS(data_z(i,j))
            ENDDO
            ENDDO
         CASE DEFAULT
            CALL cp__b("dbcsr/ops/dbcsr_operations.F",2605,"Wrong data type")
         END SELECT
      ENDDO
      CALL dbcsr_iterator_stop(iter)
      CALL mp_sum(buff_d, dbcsr_mp_group(dbcsr_distribution_mp(matrix%m%dist)))
      norm = MAXVAL(buff_d)
      DEALLOCATE (buff_d)

   END FUNCTION dbcsr_gershgorin_norm_r8

! **************************************************************************************************
!> \brief compute a norm of a dbcsr matrix
!> \param[in] matrix  the matrix
!> \retval norm ...
! **************************************************************************************************
   FUNCTION dbcsr_maxabs_r8(matrix) RESULT(norm)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      REAL(real_8)                                       :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_maxabs_r8', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=real_4), DIMENSION(:, :), POINTER     :: data_c
      COMPLEX(KIND=real_8), DIMENSION(:, :), POINTER     :: data_z
      INTEGER                                            :: blk, col, row
      LOGICAL                                            :: tr
      REAL(KIND=real_4), DIMENSION(:, :), POINTER        :: data_r
      REAL(KIND=real_8), DIMENSION(:, :), POINTER        :: data_d
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      norm = 0.0_dp
      CALL dbcsr_iterator_start(iter, matrix)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         SELECT CASE (dbcsr_get_data_type (matrix))
         CASE (dbcsr_type_real_4)
            CALL dbcsr_iterator_next_block(iter, row, col, data_r, tr, blk)
            norm = MAX(norm, REAL(MAXVAL(ABS(data_r)), dp))
         CASE (dbcsr_type_real_8)
            CALL dbcsr_iterator_next_block(iter, row, col, data_d, tr, blk)
            norm = MAX(norm, MAXVAL(ABS(data_d)))
         CASE (dbcsr_type_complex_4)
            CALL dbcsr_iterator_next_block(iter, row, col, data_c, tr, blk)
            norm = MAX(norm, REAL(MAXVAL(ABS(data_c)), dp))
         CASE (dbcsr_type_complex_8)
            CALL dbcsr_iterator_next_block(iter, row, col, data_z, tr, blk)
            norm = MAX(norm, MAXVAL(ABS(data_z)))
         CASE DEFAULT
            CALL cp__b("dbcsr/ops/dbcsr_operations.F",2654,"Wrong data type")
         END SELECT
      ENDDO
      CALL dbcsr_iterator_stop(iter)

      CALL mp_max(norm, dbcsr_mp_group(dbcsr_distribution_mp(matrix%m%dist)))

   END FUNCTION dbcsr_maxabs_r8

! **************************************************************************************************
!> \brief compute a norm of a dbcsr matrix
!> \param[in] matrix  the matrix
!> \param local ...
!> \retval norm ...
! **************************************************************************************************
   FUNCTION dbcsr_frobenius_norm_r8(matrix, local) RESULT(norm)

      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix
      LOGICAL, INTENT(in), OPTIONAL                      :: local
      REAL(KIND=real_8)                                  :: norm

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_frobenius_norm_r8', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=real_4), DIMENSION(:, :), POINTER     :: data_c
      COMPLEX(KIND=real_8), DIMENSION(:, :), POINTER     :: data_z
      INTEGER                                            :: blk, col, handle, row
      LOGICAL                                            :: any_sym, my_local, tr
      REAL(KIND=real_4), DIMENSION(:, :), POINTER        :: data_r
      REAL(KIND=real_8), DIMENSION(:, :), POINTER        :: data_d
      REAL(real_8)                                       :: fac
      TYPE(dbcsr_iterator)                               :: iter

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)

      my_local = .FALSE.
      IF (PRESENT(local)) my_local = local

      any_sym = dbcsr_get_matrix_type(matrix) .EQ. dbcsr_type_symmetric .OR. &
                dbcsr_get_matrix_type(matrix) .EQ. dbcsr_type_antisymmetric

      norm = 0.0_dp
      CALL dbcsr_iterator_start(iter, matrix)
      DO WHILE (dbcsr_iterator_blocks_left(iter))
         SELECT CASE (dbcsr_get_data_type (matrix))
         CASE (dbcsr_type_real_4)
            CALL dbcsr_iterator_next_block(iter, row, col, data_r, tr, blk)
            fac = 1.0_dp
            IF (any_sym .AND. row .NE. col) fac = 2.0_dp
            norm = norm+fac*SUM(data_r**2)
         CASE (dbcsr_type_real_8)
            CALL dbcsr_iterator_next_block(iter, row, col, data_d, tr, blk)
            fac = 1.0_dp
            IF (any_sym .AND. row .NE. col) fac = 2.0_dp
            norm = norm+fac*SUM(data_d**2)
         CASE (dbcsr_type_complex_4)
            CALL dbcsr_iterator_next_block(iter, row, col, data_c, tr, blk)
            fac = 1.0_dp
            IF (any_sym .AND. row .NE. col) &
               CALL cp__b("dbcsr/ops/dbcsr_operations.F",2715,"Only nonsymmetric matrix so far")
            norm = norm+fac*REAL(SUM(CONJG(data_c)*data_c), KIND=real_8)
         CASE (dbcsr_type_complex_8)
            CALL dbcsr_iterator_next_block(iter, row, col, data_z, tr, blk)
            fac = 1.0_dp
            IF (any_sym .AND. row .NE. col) &
               CALL cp__b("dbcsr/ops/dbcsr_operations.F",2721,"Only nonsymmetric matrix so far")
            norm = norm+fac*REAL(SUM(CONJG(data_z)*data_z), KIND=real_8)
         CASE DEFAULT
            CALL cp__b("dbcsr/ops/dbcsr_operations.F",2724,"Wrong data type")
         END SELECT
      ENDDO
      CALL dbcsr_iterator_stop(iter)
      IF (.NOT. my_local) CALL mp_sum(norm, dbcsr_mp_group(dbcsr_distribution_mp(matrix%m%dist)))
      norm = SQRT(norm)

      CALL timestop(handle)

   END FUNCTION dbcsr_frobenius_norm_r8

! **************************************************************************************************
!> \brief Sums blocks in a replicated dbcsr matrix, which has the same structure on all ranks.
!> \param[in,out] matrix      dbcsr matrix to operate on
! **************************************************************************************************
   SUBROUTINE dbcsr_sum_replicated(matrix)
      TYPE(dbcsr_obj), INTENT(inout)                     :: matrix

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_sum_replicated', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: comm, handle, index_checksum, mynode, &
                                                            numnodes
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: all_checksums
      TYPE(dbcsr_mp_obj)                                 :: mp

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, handle)
      mp = dbcsr_distribution_mp(dbcsr_distribution(matrix))
      comm = dbcsr_mp_group(mp)
      numnodes = dbcsr_mp_numnodes(mp)
      mynode = dbcsr_mp_mynode(mp)
      !
      ALLOCATE (all_checksums(numnodes))
      CALL dbcsr_index_checksum(matrix, index_checksum)
      CALL mp_allgather(index_checksum, all_checksums, comm)
      !
      IF (.NOT. ALL(all_checksums .EQ. index_checksum)) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",2763,"Replicated matrices do not all have the same index structure.")
      !
      SELECT CASE (dbcsr_data_get_type (matrix%m%data_area))
      CASE (dbcsr_type_real_4)
         CALL mp_sum(matrix%m%data_area%d%r_sp, comm)
      CASE (dbcsr_type_real_8)
         CALL mp_sum(matrix%m%data_area%d%r_dp, comm)
      CASE (dbcsr_type_complex_4)
         CALL mp_sum(matrix%m%data_area%d%c_sp, comm)
      CASE (dbcsr_type_complex_8)
         CALL mp_sum(matrix%m%data_area%d%c_dp, comm)
      CASE default
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",2775,"Incorrect data type")
      END SELECT
      !
      CALL timestop(handle)
   END SUBROUTINE dbcsr_sum_replicated

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param matrix_b ...
!> \param trace ...
!> \param trans_a ...
!> \param trans_b ...
!> \param local_sum ...
! **************************************************************************************************
   SUBROUTINE dbcsr_trace_a_b_d(matrix_a, matrix_b, trace, trans_a, trans_b, local_sum)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a, matrix_b
      REAL(kind=real_8), INTENT(INOUT)                   :: trace
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: trans_a, trans_b
      LOGICAL, INTENT(IN), OPTIONAL                      :: local_sum

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_trace_a_b_d', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(kind=real_4)                                  :: trace_4

      CALL timeset(routineN, handle)
      IF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_real_8 .AND. &
          dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_real_8 .OR. &
          dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_real_4 .AND. &
          dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_real_8 .OR. &
          dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_real_8 .AND. &
          dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_real_4) THEN
         CALL dbcsr_trace_ab_d(matrix_a, matrix_b, trace, trans_a, trans_b, local_sum)
      ELSEIF (dbcsr_get_data_type(matrix_a) .EQ. dbcsr_type_real_4 .AND. &
              dbcsr_get_data_type(matrix_b) .EQ. dbcsr_type_real_4) THEN
         trace_4 = 0.0_real_4
         CALL dbcsr_trace_ab_s(matrix_a, matrix_b, trace_4, trans_a, trans_b, local_sum)
         trace = REAL(trace_4, real_8)
      ELSE
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",2816,"Invalid combination of data type, NYI")
      ENDIF
      CALL timestop(handle)
   END SUBROUTINE dbcsr_trace_a_b_d

! **************************************************************************************************
!> \brief ...
!> \param matrix_a ...
!> \param trace ...
! **************************************************************************************************
   SUBROUTINE dbcsr_trace_a_any(matrix_a, trace)
      TYPE(dbcsr_obj), INTENT(INOUT)                     :: matrix_a
      TYPE(dbcsr_scalar_type), INTENT(INOUT)             :: trace

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_trace_a_any', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: error_handle

!   ---------------------------------------------------------------------------

      CALL timeset(routineN, error_handle)
      SELECT CASE (dbcsr_scalar_get_type (trace))
      CASE (dbcsr_type_real_4)
         CALL dbcsr_trace(matrix_a, trace%r_sp)
      CASE (dbcsr_type_real_8)
         CALL dbcsr_trace(matrix_a, trace%r_dp)
      CASE (dbcsr_type_complex_4)
         CALL dbcsr_trace(matrix_a, trace%c_sp)
      CASE (dbcsr_type_complex_8)
         CALL dbcsr_trace(matrix_a, trace%c_dp)
      CASE default
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",2848,"Invalid data type.")
      END SELECT

      CALL timestop(error_handle)
   END SUBROUTINE dbcsr_trace_a_any

! **************************************************************************************************
!> \brief check if a block is not in the limits
!> \param[in] row ...
!> \param[in] col ...
!> \param[in] block_row_limits ...
!> \param[in] block_column_limits ...
!> \retval dbcsr_block_in_limits ...
! **************************************************************************************************
   FUNCTION dbcsr_block_in_limits(row, col, block_row_limits, block_column_limits)
      INTEGER, INTENT(in)                                :: row, col
      INTEGER, DIMENSION(2), INTENT(in), OPTIONAL        :: block_row_limits, block_column_limits
      LOGICAL                                            :: dbcsr_block_in_limits

      dbcsr_block_in_limits = .TRUE.
      IF (PRESENT(block_row_limits)) THEN
         IF (row .LT. block_row_limits(1)) dbcsr_block_in_limits = .FALSE.
         IF (row .GT. block_row_limits(2)) dbcsr_block_in_limits = .FALSE.
      ENDIF
      IF (PRESENT(block_column_limits)) THEN
         IF (col .LT. block_column_limits(1)) dbcsr_block_in_limits = .FALSE.
         IF (col .GT. block_column_limits(2)) dbcsr_block_in_limits = .FALSE.
      ENDIF
   END FUNCTION dbcsr_block_in_limits

! **************************************************************************************************
!> \brief Gets information about a matrix
!> \param[in] matrix          matrix to query
!> \param[out] nblkrows_total ...
!> \param[out] nblkcols_total ...
!> \param[out] nfullrows_total ...
!> \param[out] nfullcols_total ...
!> \param[out] nblkrows_local ...
!> \param[out] nblkcols_local ...
!> \param[out] nfullrows_local ...
!> \param[out] nfullcols_local ...
!> \param[out] my_prow ...
!> \param[out] my_pcol ...
!> \param[out] local_rows ...
!> \param[out] local_cols ...
!> \param[out] proc_row_dist ...
!> \param[out] proc_col_dist ...
!> \param[out] row_blk_size ...
!> \param[out] col_blk_size ...
!> \param[out] row_blk_offset ...
!> \param[out] col_blk_offset ...
!> \param[out] distribution   the data distribution of the matrix
!> \param[out] name           matrix name
!> \param[out] data_area      data_area
!> \param[out] matrix_type    matrix type (regular, symmetric, see
!>                            dbcsr_types.F for values)
!> \param[out] data_type      data type (single/double precision real/complex)
! **************************************************************************************************
   SUBROUTINE dbcsr_get_info(matrix, nblkrows_total, nblkcols_total, &
                             nfullrows_total, nfullcols_total, &
                             nblkrows_local, nblkcols_local, &
                             nfullrows_local, nfullcols_local, &
                             my_prow, my_pcol, &
                             local_rows, local_cols, proc_row_dist, proc_col_dist, &
                             row_blk_size, col_blk_size, row_blk_offset, col_blk_offset, distribution, name, data_area, &
                             matrix_type, data_type)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      INTEGER, INTENT(OUT), OPTIONAL :: nblkrows_total, nblkcols_total, nfullrows_total, &
         nfullcols_total, nblkrows_local, nblkcols_local, nfullrows_local, nfullcols_local, &
         my_prow, my_pcol
      INTEGER, DIMENSION(:), OPTIONAL, POINTER :: local_rows, local_cols, proc_row_dist, &
         proc_col_dist, row_blk_size, col_blk_size, row_blk_offset, col_blk_offset
      TYPE(dbcsr_distribution_obj), INTENT(OUT), &
         OPTIONAL                                        :: distribution
      CHARACTER(len=*), INTENT(OUT), OPTIONAL            :: name
      TYPE(dbcsr_data_obj), INTENT(OUT), OPTIONAL        :: data_area
      CHARACTER, OPTIONAL                                :: matrix_type
      INTEGER, OPTIONAL                                  :: data_type

      CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_info', routineP = moduleN//':'//routineN

!   ---------------------------------------------------------------------------

      IF (.NOT. matrix%m%initialized) &
         CALL cp__b("dbcsr/ops/dbcsr_operations.F",2932,"Matrix not initialized")
      !vw avoid massive priting of warnings
      !CALL cp__w("dbcsr/ops/dbcsr_operations.F",2934,"Invalid matrix")
      IF (PRESENT(nblkrows_total)) nblkrows_total = matrix%m%nblkrows_total
      IF (PRESENT(nblkcols_total)) nblkcols_total = matrix%m%nblkcols_total
      IF (PRESENT(nfullrows_total)) nfullrows_total = matrix%m%nfullrows_total
      IF (PRESENT(nfullcols_total)) nfullcols_total = matrix%m%nfullcols_total
      IF (PRESENT(nblkrows_local)) nblkrows_local = matrix%m%nblkrows_local
      IF (PRESENT(nblkcols_local)) nblkcols_local = matrix%m%nblkcols_local
      IF (PRESENT(nfullrows_local)) nfullrows_local = matrix%m%nfullrows_local
      IF (PRESENT(nfullcols_local)) nfullcols_local = matrix%m%nfullcols_local
      IF (PRESENT(row_blk_size)) row_blk_size => array_data(matrix%m%row_blk_size)
      IF (PRESENT(col_blk_size)) col_blk_size => array_data(matrix%m%col_blk_size)
      IF (PRESENT(row_blk_offset)) row_blk_offset => array_data(matrix%m%row_blk_offset)
      IF (PRESENT(col_blk_offset)) col_blk_offset => array_data(matrix%m%col_blk_offset)
      IF (PRESENT(distribution)) distribution = matrix%m%dist
      IF (PRESENT(name)) name = matrix%m%name
      IF (PRESENT(data_area)) data_area = matrix%m%data_area
      IF (PRESENT(matrix_type)) THEN
         matrix_type = dbcsr_get_matrix_type(matrix)
         IF (matrix_type .EQ. dbcsr_type_invalid) &
            CALL cp__b("dbcsr/ops/dbcsr_operations.F",2953,"Incorrect symmetry")
      ENDIF
      IF (PRESENT(data_type)) data_type = matrix%m%data_type
      IF (PRESENT(local_rows)) &
         local_rows => dbcsr_distribution_local_rows(matrix%m%dist)
      IF (PRESENT(local_cols)) &
         local_cols => dbcsr_distribution_local_cols(matrix%m%dist)
      IF (PRESENT(proc_row_dist)) &
         proc_row_dist => dbcsr_distribution_row_dist(matrix%m%dist)
      IF (PRESENT(proc_col_dist)) &
         proc_col_dist => dbcsr_distribution_col_dist(matrix%m%dist)
      IF (PRESENT(my_prow)) &
         my_prow = dbcsr_mp_myprow(dbcsr_distribution_mp(matrix%m%dist))
      IF (PRESENT(my_pcol)) &
         my_pcol = dbcsr_mp_mypcol(dbcsr_distribution_mp(matrix%m%dist))
   END SUBROUTINE dbcsr_get_info

! **************************************************************************************************
!> \brief Returns whether the matrix could be represeneted in a dense form
!> \param[in] matrix          matrix
!> \param occ_thresh ...
!> \retval may_be_dense    use the mutable and not append-only working
!>                            structures
! **************************************************************************************************
   FUNCTION dbcsr_may_be_dense(matrix, occ_thresh) RESULT(may_be_dense)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      REAL(real_8), INTENT(in)                           :: occ_thresh
      LOGICAL                                            :: may_be_dense

      REAL(real_8)                                       :: occ

!   ---------------------------------------------------------------------------

      occ = dbcsr_get_occupation(matrix)
      may_be_dense = .NOT. (occ .LT. occ_thresh)
      ! make sure every proc sees the same
      CALL mp_sum(may_be_dense, dbcsr_mp_group(dbcsr_distribution_mp(matrix%m%dist)))
   END FUNCTION dbcsr_may_be_dense

! **************************************************************************************************
!> \brief Returns the occupation of the matrix
!> \param matrix     matrix from which to get the occupation
!> \retval occupation ...
! **************************************************************************************************
   FUNCTION dbcsr_get_occupation(matrix) RESULT(occupation)
      TYPE(dbcsr_obj), INTENT(IN)                        :: matrix
      REAL(KIND=real_8)                                  :: occupation

      INTEGER                                            :: nfullcols, nfullrows
      INTEGER(KIND=int_8)                                :: nze_global
      INTEGER, DIMENSION(:), POINTER                     :: row_blk_size

      nze_global = matrix%m%nze
      CALL mp_sum(nze_global, dbcsr_mp_group(dbcsr_distribution_mp(matrix%m%dist)))

      nfullrows = dbcsr_nfullrows_total(matrix)
      nfullcols = dbcsr_nfullcols_total(matrix)

      row_blk_size => array_data(matrix%m%row_blk_size)

      IF (nfullrows .NE. 0 .AND. nfullcols .NE. 0) THEN
         IF (dbcsr_has_symmetry(matrix)) THEN
            IF (2*nze_global .EQ. &
                (INT(nfullrows, KIND=int_8)*INT(nfullrows+1, KIND=int_8)+SUM(row_blk_size*(row_blk_size-1)))) THEN
               occupation = 1.0_real_8
            ELSE
               occupation = 2.0_real_8*REAL(nze_global, real_8)/ &
                            (REAL(nfullrows, real_8)*REAL(nfullrows+1, real_8)+ &
                             SUM(REAL(row_blk_size, real_8)*REAL(row_blk_size-1, real_8)))
            ENDIF
         ELSE
            IF (nze_global .EQ. INT(nfullrows, KIND=int_8)*INT(nfullcols, KIND=int_8)) THEN
               occupation = 1.0_real_8
            ELSE
               occupation = REAL(nze_global, real_8)/(REAL(nfullrows, real_8)*REAL(nfullcols, real_8))
            ENDIF
         ENDIF
      ELSE
         occupation = 0.0_real_8
      ENDIF
   END FUNCTION dbcsr_get_occupation

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief traces a DBCSR matrix
!> \param[in] matrix_a       DBCSR matrix
!> \param[out] trace         the trace of the matrix
!>
! **************************************************************************************************
  SUBROUTINE dbcsr_trace_a_d(matrix_a, trace)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a
    REAL(kind=real_8), INTENT(INOUT)                   :: trace

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_trace_a_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: a_blk, a_col, a_col_size, &
                                                a_nze, a_row, a_row_size, i, &
                                                mynode, error_handle
    INTEGER, DIMENSION(:), POINTER           :: col_blk_size, row_blk_size,&
                                                row_dist, col_dist
    REAL(kind=real_8), DIMENSION(:), POINTER           :: a_data, data_p
    INTEGER, DIMENSION(:,:), POINTER         :: pgrid
    TYPE(dbcsr_distribution_obj)             :: dist

!   ---------------------------------------------------------------------------
    CALL timeset(routineN, error_handle)

    row_blk_size => array_data (matrix_a%m%row_blk_size)
    col_blk_size => array_data (matrix_a%m%col_blk_size)
    CALL dbcsr_assert (dbcsr_get_data_type (matrix_a), "EQ", dbcsr_type_real_8,&
         dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
         "Incompatible data types", 35)
    CALL dbcsr_get_data (matrix_a%m%data_area, data_p)
    dist = dbcsr_distribution (matrix_a)
    mynode = dbcsr_mp_mynode (dbcsr_distribution_mp (dist))
    pgrid => dbcsr_mp_pgrid (dbcsr_distribution_mp (dist))
    row_dist => dbcsr_distribution_row_dist (dist)
    col_dist => dbcsr_distribution_col_dist (dist)
    !
    ! let's go
    trace = REAL(0.0,real_8)
    DO a_row = 1, matrix_a%m%nblkrows_total
       a_row_size = row_blk_size(a_row)
       DO a_blk = matrix_a%m%row_p(a_row)+1,matrix_a%m%row_p(a_row+1)
          IF (a_blk .EQ. 0) CYCLE
          a_col = matrix_a%m%col_i(a_blk)
          IF(a_col.ne.a_row) CYCLE
          ! We must skip non-local blocks in a replicated matrix.
          IF(matrix_a%m%replication_type .NE. dbcsr_repl_full) THEN
             IF (mynode .NE. checker_square_proc (a_row, a_col, pgrid,&
                  row_dist, col_dist)) CYCLE
          ENDIF
          a_col_size = col_blk_size(a_col)
          IF(a_row_size.NE.a_col_size)&
             CALL cp__b("dbcsr/ops/dbcsr_operations.F",58,"is that a square matrix?")
          a_nze = a_row_size**2
          a_data => pointer_view (data_p, ABS(matrix_a%m%blk_p(a_blk)),&
               ABS(matrix_a%m%blk_p(a_blk))+a_nze-1)
          !data_a => matrix_a%m%data(ABS(matrix_a%m%blk_p(a_blk)):ABS(matrix_a%m%blk_p(a_blk))+a_nze-1)
          !
          ! let's trace the block
          DO i = 1,a_row_size
             trace = trace + a_data((i-1)*a_row_size+i)
          ENDDO
       ENDDO ! a_col
    ENDDO ! a_row
    !
    ! summe
    CALL mp_sum(trace,dbcsr_mp_group(dbcsr_distribution_mp(matrix_a%m%dist)))

    CALL timestop(error_handle)
  END SUBROUTINE dbcsr_trace_a_d

! **************************************************************************************************
!> \brief traces a product of DBCSR matrices
!> \param[in] matrix_a DBCSR matrices
!> \param[in] matrix_b DBCSR matrices
!> \param[out] trace             the trace of the product of the matrices
!> \param[in] trans_a            (optional) is matrix_a transposed or not?
!> \param[in] trans_b            (optional) is matrix_b transposed or not?
!> \param local_sum ...
! **************************************************************************************************
  SUBROUTINE dbcsr_trace_ab_d(matrix_a, matrix_b, trace, trans_a, trans_b, local_sum)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a, matrix_b
    REAL(kind=real_8), INTENT(INOUT)                   :: trace
    CHARACTER(LEN=*), INTENT(IN), OPTIONAL   :: trans_a, trans_b
    LOGICAL, INTENT(IN), OPTIONAL            :: local_sum

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_trace_ab_d', &
      routineP = moduleN//':'//routineN

    CHARACTER(LEN=1)                         :: my_trans_a, my_trans_b
    INTEGER :: a_blk, a_col, a_col_size, a_row_size, b_blk, b_col_size, &
      b_frst_blk, b_last_blk, b_row_size, nze, row, a_beg, a_end, b_beg, b_end
    INTEGER, DIMENSION(:), POINTER           :: a_col_blk_size, &
                                                a_row_blk_size, &
                                                b_col_blk_size, b_row_blk_size
    REAL(kind=real_8)                                  :: sym_fac, fac
    LOGICAL                                  :: found, my_local_sum
    REAL(real_4), EXTERNAL                   :: SDOT
    REAL(real_8), EXTERNAL                   :: DDOT
    COMPLEX(real_4), EXTERNAL                :: CDOTU
    COMPLEX(real_8), EXTERNAL                :: ZDOTU
    REAL(real_4), DIMENSION(:), POINTER      :: a_data_r, b_data_r
    REAL(real_8), DIMENSION(:), POINTER      :: a_data_d, b_data_d
    COMPLEX(real_4), DIMENSION(:), POINTER   :: a_data_c, b_data_c
    COMPLEX(real_8), DIMENSION(:), POINTER   :: a_data_z, b_data_z

!   ---------------------------------------------------------------------------

    my_local_sum = .FALSE.
    IF(PRESENT(local_sum)) my_local_sum=local_sum
    IF(.NOT.my_local_sum) THEN!fixme
       IF(matrix_a%m%replication_type .NE. dbcsr_repl_none&
            .OR. matrix_b%m%replication_type .NE. dbcsr_repl_none)&
            CALL cp__b("dbcsr/ops/dbcsr_operations.F",119,"Trace of product of replicated matrices not yet possible.")
    ENDIF

    sym_fac = REAL(1.0,real_8)
    IF((dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_antisymmetric).AND.&
       (dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_antisymmetric)) sym_fac = REAL(2.0,real_8)

    ! tracing a symmetric with a general matrix is not implemented, as it would require communication of blocks
    IF((dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_antisymmetric).NEQV. &
       (dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_antisymmetric)) THEN
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",133,"Tracing general with symmetric matrix NYI")
    ENDIF

    a_row_blk_size => array_data (matrix_a%m%row_blk_size)
    a_col_blk_size => array_data (matrix_a%m%col_blk_size)
    b_row_blk_size => array_data (matrix_b%m%row_blk_size)
    b_col_blk_size => array_data (matrix_b%m%col_blk_size)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_r)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_r)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_d)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_d)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_c)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_c)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_z)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_z)

    my_trans_a = 'T'
    IF(PRESENT(trans_a)) my_trans_a = trans_a
    my_trans_b = 'N'
    IF(PRESENT(trans_b)) my_trans_b = trans_b
    IF(my_trans_a.NE.'T'.OR.my_trans_b.NE.'N')&
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",154,"this combination of transpose is NYI")
    !
    ! let's go
    trace = REAL(0.0,real_8)
    IF(matrix_a%m%nblkrows_total.NE.matrix_b%m%nblkrows_total)&
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",159,"this combination of transpose is NYI")
    DO row = 1, matrix_a%m%nblkrows_total
       a_row_size = a_row_blk_size(row)
       b_row_size = b_row_blk_size(row)
       IF(a_row_size.NE.b_row_size) THEN
          WRITE(*,*) 'a_row_size',a_row_size
          WRITE(*,*) 'b_row_size',b_row_size
          CALL cp__b("dbcsr/ops/dbcsr_operations.F",166,"matrices not consistent")
       ENDIF
       b_blk = matrix_b%m%row_p(row)+1
       b_frst_blk = matrix_b%m%row_p(row)+1
       b_last_blk = matrix_b%m%row_p(row+1)
       DO a_blk = matrix_a%m%row_p(row)+1,matrix_a%m%row_p(row+1)
          IF (matrix_a%m%blk_p(a_blk) .EQ. 0) CYCLE ! Deleted block
          a_col = matrix_a%m%col_i(a_blk)
          a_col_size = a_col_blk_size(a_col)
          !
          ! find the b_blk we assume here that the colums are ordered !
          CALL dbcsr_find_column(a_col,b_frst_blk,b_last_blk,matrix_b%m%col_i,&
               matrix_b%m%blk_p,b_blk,found)
          IF(found) THEN
             b_col_size = b_col_blk_size(a_col)
             IF(a_col_size.NE.b_col_size)  THEN
                WRITE(*,*) 'a_col_size',a_col_size
                WRITE(*,*) 'b_col_size',b_col_size
                CALL cp__b("dbcsr/ops/dbcsr_operations.F",184,"matrices not consistent")
             ENDIF
             !
             nze = a_row_size*a_col_size
             !
             IF(nze.GT.0) THEN
                !
                ! let's trace the blocks
                a_beg = ABS(matrix_a%m%blk_p(a_blk))
                a_end = a_beg + nze - 1
                b_beg = ABS(matrix_b%m%blk_p(b_blk))
                b_end = b_beg + nze - 1
                fac = REAL(1.0,real_8)
                IF(row.NE.a_col) fac = sym_fac
                ! this is a mess, no need to support that..
                IF(    matrix_a%m%data_type.EQ.dbcsr_type_real_4.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_4) THEN
                   trace = trace +REAL(fac * SDOT (nze,&
                        a_data_r(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_r(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_8)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_4.AND.&
                     matrix_b%m%data_type.EQ.dbcsr_type_real_8) THEN
                   trace = trace + &
                        REAL(fac * SUM ( a_data_r(a_beg:a_end) * b_data_d(b_beg:b_end) ),kind=real_8)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_4) THEN
                   trace = trace + &
                        REAL(fac * SUM ( a_data_d(a_beg:a_end) * b_data_r(b_beg:b_end) ),kind=real_8)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_8) THEN
                   trace = trace + REAL(fac * DDOT (nze,&
                        a_data_d(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_d(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_8)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_complex_4.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_complex_4) THEN
                   trace = trace + REAL(fac * CDOTU (nze,&
                        a_data_c(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_c(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_8)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_complex_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_complex_8) THEN
                   trace = trace + REAL(fac * ZDOTU (nze,&
                        a_data_z(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_z(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_8)
                ELSE
                   CALL cp__b("dbcsr/ops/dbcsr_operations.F",228,"combination of types NYI")
                ENDIF
             ENDIF
          ENDIF
       ENDDO ! a_col
    ENDDO ! a_row
    !
    ! summe
    IF(.NOT.my_local_sum) &
           CALL mp_sum(trace,dbcsr_mp_group(dbcsr_distribution_mp(matrix_a%m%dist)))

  END SUBROUTINE dbcsr_trace_ab_d


! **************************************************************************************************
!> \brief Interface for matrix scaling by a scalar
!> \param matrix_a ...
!> \param alpha_scalar ...
!> \param last_column ...
! **************************************************************************************************
  SUBROUTINE dbcsr_scale_d(matrix_a, alpha_scalar, last_column)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a
    REAL(kind=real_8), INTENT(IN)                      :: alpha_scalar
    INTEGER, INTENT(IN), OPTIONAL            :: last_column

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scale_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: error_handler
    TYPE(dbcsr_scalar_type)                  :: sc

    sc = dbcsr_scalar (alpha_scalar)
    CALL dbcsr_scalar_fill_all (sc)
    sc%data_type = dbcsr_get_data_type (matrix_a)
    CALL timeset(routineN, error_handler)
    IF (PRESENT (last_column)) THEN
       CALL dbcsr_scale_anytype(matrix_a,&
            alpha_scalar=sc,&
            limits=(/0,0,0,last_column/))
    ELSE
       CALL dbcsr_scale_anytype(matrix_a, alpha_scalar=sc)
    ENDIF
    CALL timestop(error_handler)
  END SUBROUTINE dbcsr_scale_d

! **************************************************************************************************
!> \brief Interface for matrix scaling by a vector
!> \param matrix_a ...
!> \param alpha ...
!> \param side ...
! **************************************************************************************************
  SUBROUTINE dbcsr_scale_by_vector_d(matrix_a, alpha, side)
    TYPE(dbcsr_obj), INTENT(INOUT)            :: matrix_a
    REAL(kind=real_8), DIMENSION(:), INTENT(IN), TARGET :: alpha
    CHARACTER(LEN=*), INTENT(IN)              :: side
    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scale_by_vector_d', &
      routineP = moduleN//':'//routineN
    REAL(kind=real_8), DIMENSION(:), POINTER            :: tmp_p
    TYPE(dbcsr_data_obj)                      :: enc_alpha_vec

    CALL dbcsr_data_init (enc_alpha_vec)
    CALL dbcsr_data_new (enc_alpha_vec, dbcsr_type_real_8)
    tmp_p => alpha
    CALL dbcsr_data_set_pointer (enc_alpha_vec, tmp_p)
    CALL dbcsr_scale_by_vector_anytype(matrix_a, enc_alpha_vec, side)
    CALL dbcsr_data_clear_pointer (enc_alpha_vec)
    CALL dbcsr_data_release (enc_alpha_vec)
  END SUBROUTINE dbcsr_scale_by_vector_d


! **************************************************************************************************
!> \brief Interface for dbcsr_set
!> \param matrix ...
!> \param alpha ...
! **************************************************************************************************
  SUBROUTINE dbcsr_set_d(matrix, alpha)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    REAL(kind=real_8), INTENT(IN)                      :: alpha
    IF (alpha==0.0_real_8) THEN
      CALL dbcsr_zero(matrix)
    ELSE
      CALL dbcsr_set_anytype(matrix, dbcsr_scalar(alpha))
    ENDIF
  END SUBROUTINE dbcsr_set_d

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param eps ...
!> \param method ...
!> \param use_absolute ...
!> \param filter_diag ...
!> \param quick ...
! **************************************************************************************************
  SUBROUTINE dbcsr_filter_d (matrix, eps, method, use_absolute, &
       filter_diag, quick)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    REAL(kind=real_8), INTENT(IN)                      :: eps
    INTEGER, INTENT(IN), OPTIONAL            :: method
    LOGICAL, INTENT(in), OPTIONAL            :: use_absolute, filter_diag, quick
    CALL dbcsr_filter_anytype (matrix, dbcsr_scalar(eps), method, &
         use_absolute, filter_diag, quick)
  END SUBROUTINE dbcsr_filter_d

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_set_diag_d(matrix, diag)
    TYPE(dbcsr_obj), INTENT(INOUT)            :: matrix
    REAL(kind=real_8), DIMENSION(:), INTENT(IN), TARGET :: diag

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_diag_d', &
      routineP = moduleN//':'//routineN

    REAL(kind=real_8), DIMENSION(:), POINTER           :: diag_p
    TYPE(dbcsr_data_obj)                     :: diag_a

    diag_p => diag
    CALL dbcsr_data_init (diag_a)
    CALL dbcsr_data_new (diag_a, dbcsr_get_data_type(matrix))
    CALL dbcsr_data_set_pointer (diag_a, diag_p)
    CALL dbcsr_set_diag(matrix, diag_a)
    CALL dbcsr_data_clear_pointer (diag_a)
    CALL dbcsr_data_release (diag_a)
  END SUBROUTINE dbcsr_set_diag_d

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_get_diag_d(matrix, diag)

    TYPE(dbcsr_obj), INTENT(IN)                    :: matrix
    REAL(kind=real_8), DIMENSION(:), INTENT(INOUT), TARGET   :: diag

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_diag_d', &
      routineP = moduleN//':'//routineN

    REAL(kind=real_8), DIMENSION(:), POINTER           :: diag_p
    TYPE(dbcsr_data_obj)                     :: diag_a

    diag_p => diag
    CALL dbcsr_data_init (diag_a)
    CALL dbcsr_data_new (diag_a, dbcsr_get_data_type(matrix))
    CALL dbcsr_data_set_pointer (diag_a, diag_p)
    CALL dbcsr_get_diag(matrix, diag_a)
    CALL dbcsr_data_clear_pointer (diag_a)
    CALL dbcsr_data_release (diag_a)
  END SUBROUTINE dbcsr_get_diag_d


! **************************************************************************************************
!> \brief add a constant to the diagonal of a matrix
!> \param[inout] matrix       DBCSR matrix
!> \param[in]    alpha_scalar scalar
!> \param first_row ...
!> \param last_row ...
! **************************************************************************************************
  SUBROUTINE dbcsr_add_on_diag_d(matrix, alpha_scalar, first_row, last_row)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    REAL(kind=real_8), INTENT(IN)                      :: alpha_scalar
    integer, intent(in), optional            :: first_row, last_row

    CALL dbcsr_add_on_diag(matrix, dbcsr_scalar(alpha_scalar), first_row, last_row)
  END SUBROUTINE dbcsr_add_on_diag_d
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief traces a DBCSR matrix
!> \param[in] matrix_a       DBCSR matrix
!> \param[out] trace         the trace of the matrix
!>
! **************************************************************************************************
  SUBROUTINE dbcsr_trace_a_z(matrix_a, trace)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: trace

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_trace_a_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: a_blk, a_col, a_col_size, &
                                                a_nze, a_row, a_row_size, i, &
                                                mynode, error_handle
    INTEGER, DIMENSION(:), POINTER           :: col_blk_size, row_blk_size,&
                                                row_dist, col_dist
    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: a_data, data_p
    INTEGER, DIMENSION(:,:), POINTER         :: pgrid
    TYPE(dbcsr_distribution_obj)             :: dist

!   ---------------------------------------------------------------------------
    CALL timeset(routineN, error_handle)

    row_blk_size => array_data (matrix_a%m%row_blk_size)
    col_blk_size => array_data (matrix_a%m%col_blk_size)
    CALL dbcsr_assert (dbcsr_get_data_type (matrix_a), "EQ", dbcsr_type_complex_8,&
         dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
         "Incompatible data types", 35)
    CALL dbcsr_get_data (matrix_a%m%data_area, data_p)
    dist = dbcsr_distribution (matrix_a)
    mynode = dbcsr_mp_mynode (dbcsr_distribution_mp (dist))
    pgrid => dbcsr_mp_pgrid (dbcsr_distribution_mp (dist))
    row_dist => dbcsr_distribution_row_dist (dist)
    col_dist => dbcsr_distribution_col_dist (dist)
    !
    ! let's go
    trace = REAL(0.0,real_8)
    DO a_row = 1, matrix_a%m%nblkrows_total
       a_row_size = row_blk_size(a_row)
       DO a_blk = matrix_a%m%row_p(a_row)+1,matrix_a%m%row_p(a_row+1)
          IF (a_blk .EQ. 0) CYCLE
          a_col = matrix_a%m%col_i(a_blk)
          IF(a_col.ne.a_row) CYCLE
          ! We must skip non-local blocks in a replicated matrix.
          IF(matrix_a%m%replication_type .NE. dbcsr_repl_full) THEN
             IF (mynode .NE. checker_square_proc (a_row, a_col, pgrid,&
                  row_dist, col_dist)) CYCLE
          ENDIF
          a_col_size = col_blk_size(a_col)
          IF(a_row_size.NE.a_col_size)&
             CALL cp__b("dbcsr/ops/dbcsr_operations.F",58,"is that a square matrix?")
          a_nze = a_row_size**2
          a_data => pointer_view (data_p, ABS(matrix_a%m%blk_p(a_blk)),&
               ABS(matrix_a%m%blk_p(a_blk))+a_nze-1)
          !data_a => matrix_a%m%data(ABS(matrix_a%m%blk_p(a_blk)):ABS(matrix_a%m%blk_p(a_blk))+a_nze-1)
          !
          ! let's trace the block
          DO i = 1,a_row_size
             trace = trace + a_data((i-1)*a_row_size+i)
          ENDDO
       ENDDO ! a_col
    ENDDO ! a_row
    !
    ! summe
    CALL mp_sum(trace,dbcsr_mp_group(dbcsr_distribution_mp(matrix_a%m%dist)))

    CALL timestop(error_handle)
  END SUBROUTINE dbcsr_trace_a_z

! **************************************************************************************************
!> \brief traces a product of DBCSR matrices
!> \param[in] matrix_a DBCSR matrices
!> \param[in] matrix_b DBCSR matrices
!> \param[out] trace             the trace of the product of the matrices
!> \param[in] trans_a            (optional) is matrix_a transposed or not?
!> \param[in] trans_b            (optional) is matrix_b transposed or not?
!> \param local_sum ...
! **************************************************************************************************
  SUBROUTINE dbcsr_trace_ab_z(matrix_a, matrix_b, trace, trans_a, trans_b, local_sum)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a, matrix_b
    COMPLEX(kind=real_8), INTENT(INOUT)                   :: trace
    CHARACTER(LEN=*), INTENT(IN), OPTIONAL   :: trans_a, trans_b
    LOGICAL, INTENT(IN), OPTIONAL            :: local_sum

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_trace_ab_z', &
      routineP = moduleN//':'//routineN

    CHARACTER(LEN=1)                         :: my_trans_a, my_trans_b
    INTEGER :: a_blk, a_col, a_col_size, a_row_size, b_blk, b_col_size, &
      b_frst_blk, b_last_blk, b_row_size, nze, row, a_beg, a_end, b_beg, b_end
    INTEGER, DIMENSION(:), POINTER           :: a_col_blk_size, &
                                                a_row_blk_size, &
                                                b_col_blk_size, b_row_blk_size
    COMPLEX(kind=real_8)                                  :: sym_fac, fac
    LOGICAL                                  :: found, my_local_sum
    REAL(real_4), EXTERNAL                   :: SDOT
    REAL(real_8), EXTERNAL                   :: DDOT
    COMPLEX(real_4), EXTERNAL                :: CDOTU
    COMPLEX(real_8), EXTERNAL                :: ZDOTU
    REAL(real_4), DIMENSION(:), POINTER      :: a_data_r, b_data_r
    REAL(real_8), DIMENSION(:), POINTER      :: a_data_d, b_data_d
    COMPLEX(real_4), DIMENSION(:), POINTER   :: a_data_c, b_data_c
    COMPLEX(real_8), DIMENSION(:), POINTER   :: a_data_z, b_data_z

!   ---------------------------------------------------------------------------

    my_local_sum = .FALSE.
    IF(PRESENT(local_sum)) my_local_sum=local_sum
    IF(.NOT.my_local_sum) THEN!fixme
       IF(matrix_a%m%replication_type .NE. dbcsr_repl_none&
            .OR. matrix_b%m%replication_type .NE. dbcsr_repl_none)&
            CALL cp__b("dbcsr/ops/dbcsr_operations.F",119,"Trace of product of replicated matrices not yet possible.")
    ENDIF

    sym_fac = REAL(1.0,real_8)
    IF((dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_antisymmetric).AND.&
       (dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_antisymmetric)) sym_fac = REAL(2.0,real_8)

    ! tracing a symmetric with a general matrix is not implemented, as it would require communication of blocks
    IF((dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_antisymmetric).NEQV. &
       (dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_antisymmetric)) THEN
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",133,"Tracing general with symmetric matrix NYI")
    ENDIF

    a_row_blk_size => array_data (matrix_a%m%row_blk_size)
    a_col_blk_size => array_data (matrix_a%m%col_blk_size)
    b_row_blk_size => array_data (matrix_b%m%row_blk_size)
    b_col_blk_size => array_data (matrix_b%m%col_blk_size)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_r)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_r)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_d)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_d)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_c)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_c)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_z)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_z)

    my_trans_a = 'T'
    IF(PRESENT(trans_a)) my_trans_a = trans_a
    my_trans_b = 'N'
    IF(PRESENT(trans_b)) my_trans_b = trans_b
    IF(my_trans_a.NE.'T'.OR.my_trans_b.NE.'N')&
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",154,"this combination of transpose is NYI")
    !
    ! let's go
    trace = REAL(0.0,real_8)
    IF(matrix_a%m%nblkrows_total.NE.matrix_b%m%nblkrows_total)&
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",159,"this combination of transpose is NYI")
    DO row = 1, matrix_a%m%nblkrows_total
       a_row_size = a_row_blk_size(row)
       b_row_size = b_row_blk_size(row)
       IF(a_row_size.NE.b_row_size) THEN
          WRITE(*,*) 'a_row_size',a_row_size
          WRITE(*,*) 'b_row_size',b_row_size
          CALL cp__b("dbcsr/ops/dbcsr_operations.F",166,"matrices not consistent")
       ENDIF
       b_blk = matrix_b%m%row_p(row)+1
       b_frst_blk = matrix_b%m%row_p(row)+1
       b_last_blk = matrix_b%m%row_p(row+1)
       DO a_blk = matrix_a%m%row_p(row)+1,matrix_a%m%row_p(row+1)
          IF (matrix_a%m%blk_p(a_blk) .EQ. 0) CYCLE ! Deleted block
          a_col = matrix_a%m%col_i(a_blk)
          a_col_size = a_col_blk_size(a_col)
          !
          ! find the b_blk we assume here that the colums are ordered !
          CALL dbcsr_find_column(a_col,b_frst_blk,b_last_blk,matrix_b%m%col_i,&
               matrix_b%m%blk_p,b_blk,found)
          IF(found) THEN
             b_col_size = b_col_blk_size(a_col)
             IF(a_col_size.NE.b_col_size)  THEN
                WRITE(*,*) 'a_col_size',a_col_size
                WRITE(*,*) 'b_col_size',b_col_size
                CALL cp__b("dbcsr/ops/dbcsr_operations.F",184,"matrices not consistent")
             ENDIF
             !
             nze = a_row_size*a_col_size
             !
             IF(nze.GT.0) THEN
                !
                ! let's trace the blocks
                a_beg = ABS(matrix_a%m%blk_p(a_blk))
                a_end = a_beg + nze - 1
                b_beg = ABS(matrix_b%m%blk_p(b_blk))
                b_end = b_beg + nze - 1
                fac = REAL(1.0,real_8)
                IF(row.NE.a_col) fac = sym_fac
                ! this is a mess, no need to support that..
                IF(    matrix_a%m%data_type.EQ.dbcsr_type_real_4.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_4) THEN
                   trace = trace +CMPLX(fac * SDOT (nze,&
                        a_data_r(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_r(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_8)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_4.AND.&
                     matrix_b%m%data_type.EQ.dbcsr_type_real_8) THEN
                   trace = trace + &
                        CMPLX(fac * SUM ( a_data_r(a_beg:a_end) * b_data_d(b_beg:b_end) ),kind=real_8)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_4) THEN
                   trace = trace + &
                        CMPLX(fac * SUM ( a_data_d(a_beg:a_end) * b_data_r(b_beg:b_end) ),kind=real_8)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_8) THEN
                   trace = trace + CMPLX(fac * DDOT (nze,&
                        a_data_d(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_d(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_8)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_complex_4.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_complex_4) THEN
                   trace = trace + CMPLX(fac * CDOTU (nze,&
                        a_data_c(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_c(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_8)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_complex_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_complex_8) THEN
                   trace = trace + CMPLX(fac * ZDOTU (nze,&
                        a_data_z(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_z(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_8)
                ELSE
                   CALL cp__b("dbcsr/ops/dbcsr_operations.F",228,"combination of types NYI")
                ENDIF
             ENDIF
          ENDIF
       ENDDO ! a_col
    ENDDO ! a_row
    !
    ! summe
    IF(.NOT.my_local_sum) &
           CALL mp_sum(trace,dbcsr_mp_group(dbcsr_distribution_mp(matrix_a%m%dist)))

  END SUBROUTINE dbcsr_trace_ab_z


! **************************************************************************************************
!> \brief Interface for matrix scaling by a scalar
!> \param matrix_a ...
!> \param alpha_scalar ...
!> \param last_column ...
! **************************************************************************************************
  SUBROUTINE dbcsr_scale_z(matrix_a, alpha_scalar, last_column)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a
    COMPLEX(kind=real_8), INTENT(IN)                      :: alpha_scalar
    INTEGER, INTENT(IN), OPTIONAL            :: last_column

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scale_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: error_handler
    TYPE(dbcsr_scalar_type)                  :: sc

    sc = dbcsr_scalar (alpha_scalar)
    CALL dbcsr_scalar_fill_all (sc)
    sc%data_type = dbcsr_get_data_type (matrix_a)
    CALL timeset(routineN, error_handler)
    IF (PRESENT (last_column)) THEN
       CALL dbcsr_scale_anytype(matrix_a,&
            alpha_scalar=sc,&
            limits=(/0,0,0,last_column/))
    ELSE
       CALL dbcsr_scale_anytype(matrix_a, alpha_scalar=sc)
    ENDIF
    CALL timestop(error_handler)
  END SUBROUTINE dbcsr_scale_z

! **************************************************************************************************
!> \brief Interface for matrix scaling by a vector
!> \param matrix_a ...
!> \param alpha ...
!> \param side ...
! **************************************************************************************************
  SUBROUTINE dbcsr_scale_by_vector_z(matrix_a, alpha, side)
    TYPE(dbcsr_obj), INTENT(INOUT)            :: matrix_a
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN), TARGET :: alpha
    CHARACTER(LEN=*), INTENT(IN)              :: side
    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scale_by_vector_z', &
      routineP = moduleN//':'//routineN
    COMPLEX(kind=real_8), DIMENSION(:), POINTER            :: tmp_p
    TYPE(dbcsr_data_obj)                      :: enc_alpha_vec

    CALL dbcsr_data_init (enc_alpha_vec)
    CALL dbcsr_data_new (enc_alpha_vec, dbcsr_type_complex_8)
    tmp_p => alpha
    CALL dbcsr_data_set_pointer (enc_alpha_vec, tmp_p)
    CALL dbcsr_scale_by_vector_anytype(matrix_a, enc_alpha_vec, side)
    CALL dbcsr_data_clear_pointer (enc_alpha_vec)
    CALL dbcsr_data_release (enc_alpha_vec)
  END SUBROUTINE dbcsr_scale_by_vector_z


! **************************************************************************************************
!> \brief Interface for dbcsr_set
!> \param matrix ...
!> \param alpha ...
! **************************************************************************************************
  SUBROUTINE dbcsr_set_z(matrix, alpha)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    COMPLEX(kind=real_8), INTENT(IN)                      :: alpha
    IF (alpha==CMPLX(0.0, 0.0, real_8)) THEN
      CALL dbcsr_zero(matrix)
    ELSE
      CALL dbcsr_set_anytype(matrix, dbcsr_scalar(alpha))
    ENDIF
  END SUBROUTINE dbcsr_set_z

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param eps ...
!> \param method ...
!> \param use_absolute ...
!> \param filter_diag ...
!> \param quick ...
! **************************************************************************************************
  SUBROUTINE dbcsr_filter_z (matrix, eps, method, use_absolute, &
       filter_diag, quick)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    COMPLEX(kind=real_8), INTENT(IN)                      :: eps
    INTEGER, INTENT(IN), OPTIONAL            :: method
    LOGICAL, INTENT(in), OPTIONAL            :: use_absolute, filter_diag, quick
    CALL dbcsr_filter_anytype (matrix, dbcsr_scalar(eps), method, &
         use_absolute, filter_diag, quick)
  END SUBROUTINE dbcsr_filter_z

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_set_diag_z(matrix, diag)
    TYPE(dbcsr_obj), INTENT(INOUT)            :: matrix
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(IN), TARGET :: diag

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_diag_z', &
      routineP = moduleN//':'//routineN

    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: diag_p
    TYPE(dbcsr_data_obj)                     :: diag_a

    diag_p => diag
    CALL dbcsr_data_init (diag_a)
    CALL dbcsr_data_new (diag_a, dbcsr_get_data_type(matrix))
    CALL dbcsr_data_set_pointer (diag_a, diag_p)
    CALL dbcsr_set_diag(matrix, diag_a)
    CALL dbcsr_data_clear_pointer (diag_a)
    CALL dbcsr_data_release (diag_a)
  END SUBROUTINE dbcsr_set_diag_z

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_get_diag_z(matrix, diag)

    TYPE(dbcsr_obj), INTENT(IN)                    :: matrix
    COMPLEX(kind=real_8), DIMENSION(:), INTENT(INOUT), TARGET   :: diag

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_diag_z', &
      routineP = moduleN//':'//routineN

    COMPLEX(kind=real_8), DIMENSION(:), POINTER           :: diag_p
    TYPE(dbcsr_data_obj)                     :: diag_a

    diag_p => diag
    CALL dbcsr_data_init (diag_a)
    CALL dbcsr_data_new (diag_a, dbcsr_get_data_type(matrix))
    CALL dbcsr_data_set_pointer (diag_a, diag_p)
    CALL dbcsr_get_diag(matrix, diag_a)
    CALL dbcsr_data_clear_pointer (diag_a)
    CALL dbcsr_data_release (diag_a)
  END SUBROUTINE dbcsr_get_diag_z


! **************************************************************************************************
!> \brief add a constant to the diagonal of a matrix
!> \param[inout] matrix       DBCSR matrix
!> \param[in]    alpha_scalar scalar
!> \param first_row ...
!> \param last_row ...
! **************************************************************************************************
  SUBROUTINE dbcsr_add_on_diag_z(matrix, alpha_scalar, first_row, last_row)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    COMPLEX(kind=real_8), INTENT(IN)                      :: alpha_scalar
    integer, intent(in), optional            :: first_row, last_row

    CALL dbcsr_add_on_diag(matrix, dbcsr_scalar(alpha_scalar), first_row, last_row)
  END SUBROUTINE dbcsr_add_on_diag_z
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief traces a DBCSR matrix
!> \param[in] matrix_a       DBCSR matrix
!> \param[out] trace         the trace of the matrix
!>
! **************************************************************************************************
  SUBROUTINE dbcsr_trace_a_s(matrix_a, trace)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a
    REAL(kind=real_4), INTENT(INOUT)                   :: trace

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_trace_a_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: a_blk, a_col, a_col_size, &
                                                a_nze, a_row, a_row_size, i, &
                                                mynode, error_handle
    INTEGER, DIMENSION(:), POINTER           :: col_blk_size, row_blk_size,&
                                                row_dist, col_dist
    REAL(kind=real_4), DIMENSION(:), POINTER           :: a_data, data_p
    INTEGER, DIMENSION(:,:), POINTER         :: pgrid
    TYPE(dbcsr_distribution_obj)             :: dist

!   ---------------------------------------------------------------------------
    CALL timeset(routineN, error_handle)

    row_blk_size => array_data (matrix_a%m%row_blk_size)
    col_blk_size => array_data (matrix_a%m%col_blk_size)
    CALL dbcsr_assert (dbcsr_get_data_type (matrix_a), "EQ", dbcsr_type_real_4,&
         dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
         "Incompatible data types", 35)
    CALL dbcsr_get_data (matrix_a%m%data_area, data_p)
    dist = dbcsr_distribution (matrix_a)
    mynode = dbcsr_mp_mynode (dbcsr_distribution_mp (dist))
    pgrid => dbcsr_mp_pgrid (dbcsr_distribution_mp (dist))
    row_dist => dbcsr_distribution_row_dist (dist)
    col_dist => dbcsr_distribution_col_dist (dist)
    !
    ! let's go
    trace = REAL(0.0,real_4)
    DO a_row = 1, matrix_a%m%nblkrows_total
       a_row_size = row_blk_size(a_row)
       DO a_blk = matrix_a%m%row_p(a_row)+1,matrix_a%m%row_p(a_row+1)
          IF (a_blk .EQ. 0) CYCLE
          a_col = matrix_a%m%col_i(a_blk)
          IF(a_col.ne.a_row) CYCLE
          ! We must skip non-local blocks in a replicated matrix.
          IF(matrix_a%m%replication_type .NE. dbcsr_repl_full) THEN
             IF (mynode .NE. checker_square_proc (a_row, a_col, pgrid,&
                  row_dist, col_dist)) CYCLE
          ENDIF
          a_col_size = col_blk_size(a_col)
          IF(a_row_size.NE.a_col_size)&
             CALL cp__b("dbcsr/ops/dbcsr_operations.F",58,"is that a square matrix?")
          a_nze = a_row_size**2
          a_data => pointer_view (data_p, ABS(matrix_a%m%blk_p(a_blk)),&
               ABS(matrix_a%m%blk_p(a_blk))+a_nze-1)
          !data_a => matrix_a%m%data(ABS(matrix_a%m%blk_p(a_blk)):ABS(matrix_a%m%blk_p(a_blk))+a_nze-1)
          !
          ! let's trace the block
          DO i = 1,a_row_size
             trace = trace + a_data((i-1)*a_row_size+i)
          ENDDO
       ENDDO ! a_col
    ENDDO ! a_row
    !
    ! summe
    CALL mp_sum(trace,dbcsr_mp_group(dbcsr_distribution_mp(matrix_a%m%dist)))

    CALL timestop(error_handle)
  END SUBROUTINE dbcsr_trace_a_s

! **************************************************************************************************
!> \brief traces a product of DBCSR matrices
!> \param[in] matrix_a DBCSR matrices
!> \param[in] matrix_b DBCSR matrices
!> \param[out] trace             the trace of the product of the matrices
!> \param[in] trans_a            (optional) is matrix_a transposed or not?
!> \param[in] trans_b            (optional) is matrix_b transposed or not?
!> \param local_sum ...
! **************************************************************************************************
  SUBROUTINE dbcsr_trace_ab_s(matrix_a, matrix_b, trace, trans_a, trans_b, local_sum)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a, matrix_b
    REAL(kind=real_4), INTENT(INOUT)                   :: trace
    CHARACTER(LEN=*), INTENT(IN), OPTIONAL   :: trans_a, trans_b
    LOGICAL, INTENT(IN), OPTIONAL            :: local_sum

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_trace_ab_s', &
      routineP = moduleN//':'//routineN

    CHARACTER(LEN=1)                         :: my_trans_a, my_trans_b
    INTEGER :: a_blk, a_col, a_col_size, a_row_size, b_blk, b_col_size, &
      b_frst_blk, b_last_blk, b_row_size, nze, row, a_beg, a_end, b_beg, b_end
    INTEGER, DIMENSION(:), POINTER           :: a_col_blk_size, &
                                                a_row_blk_size, &
                                                b_col_blk_size, b_row_blk_size
    REAL(kind=real_4)                                  :: sym_fac, fac
    LOGICAL                                  :: found, my_local_sum
    REAL(real_4), EXTERNAL                   :: SDOT
    REAL(real_8), EXTERNAL                   :: DDOT
    COMPLEX(real_4), EXTERNAL                :: CDOTU
    COMPLEX(real_8), EXTERNAL                :: ZDOTU
    REAL(real_4), DIMENSION(:), POINTER      :: a_data_r, b_data_r
    REAL(real_8), DIMENSION(:), POINTER      :: a_data_d, b_data_d
    COMPLEX(real_4), DIMENSION(:), POINTER   :: a_data_c, b_data_c
    COMPLEX(real_8), DIMENSION(:), POINTER   :: a_data_z, b_data_z

!   ---------------------------------------------------------------------------

    my_local_sum = .FALSE.
    IF(PRESENT(local_sum)) my_local_sum=local_sum
    IF(.NOT.my_local_sum) THEN!fixme
       IF(matrix_a%m%replication_type .NE. dbcsr_repl_none&
            .OR. matrix_b%m%replication_type .NE. dbcsr_repl_none)&
            CALL cp__b("dbcsr/ops/dbcsr_operations.F",119,"Trace of product of replicated matrices not yet possible.")
    ENDIF

    sym_fac = REAL(1.0,real_4)
    IF((dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_antisymmetric).AND.&
       (dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_antisymmetric)) sym_fac = REAL(2.0,real_4)

    ! tracing a symmetric with a general matrix is not implemented, as it would require communication of blocks
    IF((dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_antisymmetric).NEQV. &
       (dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_antisymmetric)) THEN
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",133,"Tracing general with symmetric matrix NYI")
    ENDIF

    a_row_blk_size => array_data (matrix_a%m%row_blk_size)
    a_col_blk_size => array_data (matrix_a%m%col_blk_size)
    b_row_blk_size => array_data (matrix_b%m%row_blk_size)
    b_col_blk_size => array_data (matrix_b%m%col_blk_size)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_r)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_r)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_d)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_d)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_c)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_c)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_z)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_z)

    my_trans_a = 'T'
    IF(PRESENT(trans_a)) my_trans_a = trans_a
    my_trans_b = 'N'
    IF(PRESENT(trans_b)) my_trans_b = trans_b
    IF(my_trans_a.NE.'T'.OR.my_trans_b.NE.'N')&
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",154,"this combination of transpose is NYI")
    !
    ! let's go
    trace = REAL(0.0,real_4)
    IF(matrix_a%m%nblkrows_total.NE.matrix_b%m%nblkrows_total)&
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",159,"this combination of transpose is NYI")
    DO row = 1, matrix_a%m%nblkrows_total
       a_row_size = a_row_blk_size(row)
       b_row_size = b_row_blk_size(row)
       IF(a_row_size.NE.b_row_size) THEN
          WRITE(*,*) 'a_row_size',a_row_size
          WRITE(*,*) 'b_row_size',b_row_size
          CALL cp__b("dbcsr/ops/dbcsr_operations.F",166,"matrices not consistent")
       ENDIF
       b_blk = matrix_b%m%row_p(row)+1
       b_frst_blk = matrix_b%m%row_p(row)+1
       b_last_blk = matrix_b%m%row_p(row+1)
       DO a_blk = matrix_a%m%row_p(row)+1,matrix_a%m%row_p(row+1)
          IF (matrix_a%m%blk_p(a_blk) .EQ. 0) CYCLE ! Deleted block
          a_col = matrix_a%m%col_i(a_blk)
          a_col_size = a_col_blk_size(a_col)
          !
          ! find the b_blk we assume here that the colums are ordered !
          CALL dbcsr_find_column(a_col,b_frst_blk,b_last_blk,matrix_b%m%col_i,&
               matrix_b%m%blk_p,b_blk,found)
          IF(found) THEN
             b_col_size = b_col_blk_size(a_col)
             IF(a_col_size.NE.b_col_size)  THEN
                WRITE(*,*) 'a_col_size',a_col_size
                WRITE(*,*) 'b_col_size',b_col_size
                CALL cp__b("dbcsr/ops/dbcsr_operations.F",184,"matrices not consistent")
             ENDIF
             !
             nze = a_row_size*a_col_size
             !
             IF(nze.GT.0) THEN
                !
                ! let's trace the blocks
                a_beg = ABS(matrix_a%m%blk_p(a_blk))
                a_end = a_beg + nze - 1
                b_beg = ABS(matrix_b%m%blk_p(b_blk))
                b_end = b_beg + nze - 1
                fac = REAL(1.0,real_4)
                IF(row.NE.a_col) fac = sym_fac
                ! this is a mess, no need to support that..
                IF(    matrix_a%m%data_type.EQ.dbcsr_type_real_4.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_4) THEN
                   trace = trace +REAL(fac * SDOT (nze,&
                        a_data_r(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_r(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_4)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_4.AND.&
                     matrix_b%m%data_type.EQ.dbcsr_type_real_8) THEN
                   trace = trace + &
                        REAL(fac * SUM ( a_data_r(a_beg:a_end) * b_data_d(b_beg:b_end) ),kind=real_4)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_4) THEN
                   trace = trace + &
                        REAL(fac * SUM ( a_data_d(a_beg:a_end) * b_data_r(b_beg:b_end) ),kind=real_4)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_8) THEN
                   trace = trace + REAL(fac * DDOT (nze,&
                        a_data_d(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_d(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_4)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_complex_4.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_complex_4) THEN
                   trace = trace + REAL(fac * CDOTU (nze,&
                        a_data_c(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_c(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_4)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_complex_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_complex_8) THEN
                   trace = trace + REAL(fac * ZDOTU (nze,&
                        a_data_z(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_z(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_4)
                ELSE
                   CALL cp__b("dbcsr/ops/dbcsr_operations.F",228,"combination of types NYI")
                ENDIF
             ENDIF
          ENDIF
       ENDDO ! a_col
    ENDDO ! a_row
    !
    ! summe
    IF(.NOT.my_local_sum) &
           CALL mp_sum(trace,dbcsr_mp_group(dbcsr_distribution_mp(matrix_a%m%dist)))

  END SUBROUTINE dbcsr_trace_ab_s


! **************************************************************************************************
!> \brief Interface for matrix scaling by a scalar
!> \param matrix_a ...
!> \param alpha_scalar ...
!> \param last_column ...
! **************************************************************************************************
  SUBROUTINE dbcsr_scale_s(matrix_a, alpha_scalar, last_column)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a
    REAL(kind=real_4), INTENT(IN)                      :: alpha_scalar
    INTEGER, INTENT(IN), OPTIONAL            :: last_column

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scale_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: error_handler
    TYPE(dbcsr_scalar_type)                  :: sc

    sc = dbcsr_scalar (alpha_scalar)
    CALL dbcsr_scalar_fill_all (sc)
    sc%data_type = dbcsr_get_data_type (matrix_a)
    CALL timeset(routineN, error_handler)
    IF (PRESENT (last_column)) THEN
       CALL dbcsr_scale_anytype(matrix_a,&
            alpha_scalar=sc,&
            limits=(/0,0,0,last_column/))
    ELSE
       CALL dbcsr_scale_anytype(matrix_a, alpha_scalar=sc)
    ENDIF
    CALL timestop(error_handler)
  END SUBROUTINE dbcsr_scale_s

! **************************************************************************************************
!> \brief Interface for matrix scaling by a vector
!> \param matrix_a ...
!> \param alpha ...
!> \param side ...
! **************************************************************************************************
  SUBROUTINE dbcsr_scale_by_vector_s(matrix_a, alpha, side)
    TYPE(dbcsr_obj), INTENT(INOUT)            :: matrix_a
    REAL(kind=real_4), DIMENSION(:), INTENT(IN), TARGET :: alpha
    CHARACTER(LEN=*), INTENT(IN)              :: side
    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scale_by_vector_s', &
      routineP = moduleN//':'//routineN
    REAL(kind=real_4), DIMENSION(:), POINTER            :: tmp_p
    TYPE(dbcsr_data_obj)                      :: enc_alpha_vec

    CALL dbcsr_data_init (enc_alpha_vec)
    CALL dbcsr_data_new (enc_alpha_vec, dbcsr_type_real_4)
    tmp_p => alpha
    CALL dbcsr_data_set_pointer (enc_alpha_vec, tmp_p)
    CALL dbcsr_scale_by_vector_anytype(matrix_a, enc_alpha_vec, side)
    CALL dbcsr_data_clear_pointer (enc_alpha_vec)
    CALL dbcsr_data_release (enc_alpha_vec)
  END SUBROUTINE dbcsr_scale_by_vector_s


! **************************************************************************************************
!> \brief Interface for dbcsr_set
!> \param matrix ...
!> \param alpha ...
! **************************************************************************************************
  SUBROUTINE dbcsr_set_s(matrix, alpha)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    REAL(kind=real_4), INTENT(IN)                      :: alpha
    IF (alpha==0.0_real_4) THEN
      CALL dbcsr_zero(matrix)
    ELSE
      CALL dbcsr_set_anytype(matrix, dbcsr_scalar(alpha))
    ENDIF
  END SUBROUTINE dbcsr_set_s

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param eps ...
!> \param method ...
!> \param use_absolute ...
!> \param filter_diag ...
!> \param quick ...
! **************************************************************************************************
  SUBROUTINE dbcsr_filter_s (matrix, eps, method, use_absolute, &
       filter_diag, quick)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    REAL(kind=real_4), INTENT(IN)                      :: eps
    INTEGER, INTENT(IN), OPTIONAL            :: method
    LOGICAL, INTENT(in), OPTIONAL            :: use_absolute, filter_diag, quick
    CALL dbcsr_filter_anytype (matrix, dbcsr_scalar(eps), method, &
         use_absolute, filter_diag, quick)
  END SUBROUTINE dbcsr_filter_s

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_set_diag_s(matrix, diag)
    TYPE(dbcsr_obj), INTENT(INOUT)            :: matrix
    REAL(kind=real_4), DIMENSION(:), INTENT(IN), TARGET :: diag

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_diag_s', &
      routineP = moduleN//':'//routineN

    REAL(kind=real_4), DIMENSION(:), POINTER           :: diag_p
    TYPE(dbcsr_data_obj)                     :: diag_a

    diag_p => diag
    CALL dbcsr_data_init (diag_a)
    CALL dbcsr_data_new (diag_a, dbcsr_get_data_type(matrix))
    CALL dbcsr_data_set_pointer (diag_a, diag_p)
    CALL dbcsr_set_diag(matrix, diag_a)
    CALL dbcsr_data_clear_pointer (diag_a)
    CALL dbcsr_data_release (diag_a)
  END SUBROUTINE dbcsr_set_diag_s

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_get_diag_s(matrix, diag)

    TYPE(dbcsr_obj), INTENT(IN)                    :: matrix
    REAL(kind=real_4), DIMENSION(:), INTENT(INOUT), TARGET   :: diag

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_diag_s', &
      routineP = moduleN//':'//routineN

    REAL(kind=real_4), DIMENSION(:), POINTER           :: diag_p
    TYPE(dbcsr_data_obj)                     :: diag_a

    diag_p => diag
    CALL dbcsr_data_init (diag_a)
    CALL dbcsr_data_new (diag_a, dbcsr_get_data_type(matrix))
    CALL dbcsr_data_set_pointer (diag_a, diag_p)
    CALL dbcsr_get_diag(matrix, diag_a)
    CALL dbcsr_data_clear_pointer (diag_a)
    CALL dbcsr_data_release (diag_a)
  END SUBROUTINE dbcsr_get_diag_s


! **************************************************************************************************
!> \brief add a constant to the diagonal of a matrix
!> \param[inout] matrix       DBCSR matrix
!> \param[in]    alpha_scalar scalar
!> \param first_row ...
!> \param last_row ...
! **************************************************************************************************
  SUBROUTINE dbcsr_add_on_diag_s(matrix, alpha_scalar, first_row, last_row)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    REAL(kind=real_4), INTENT(IN)                      :: alpha_scalar
    integer, intent(in), optional            :: first_row, last_row

    CALL dbcsr_add_on_diag(matrix, dbcsr_scalar(alpha_scalar), first_row, last_row)
  END SUBROUTINE dbcsr_add_on_diag_s
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief traces a DBCSR matrix
!> \param[in] matrix_a       DBCSR matrix
!> \param[out] trace         the trace of the matrix
!>
! **************************************************************************************************
  SUBROUTINE dbcsr_trace_a_c(matrix_a, trace)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: trace

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_trace_a_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: a_blk, a_col, a_col_size, &
                                                a_nze, a_row, a_row_size, i, &
                                                mynode, error_handle
    INTEGER, DIMENSION(:), POINTER           :: col_blk_size, row_blk_size,&
                                                row_dist, col_dist
    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: a_data, data_p
    INTEGER, DIMENSION(:,:), POINTER         :: pgrid
    TYPE(dbcsr_distribution_obj)             :: dist

!   ---------------------------------------------------------------------------
    CALL timeset(routineN, error_handle)

    row_blk_size => array_data (matrix_a%m%row_blk_size)
    col_blk_size => array_data (matrix_a%m%col_blk_size)
    CALL dbcsr_assert (dbcsr_get_data_type (matrix_a), "EQ", dbcsr_type_complex_4,&
         dbcsr_fatal_level, dbcsr_wrong_args_error, routineN,&
         "Incompatible data types", 35)
    CALL dbcsr_get_data (matrix_a%m%data_area, data_p)
    dist = dbcsr_distribution (matrix_a)
    mynode = dbcsr_mp_mynode (dbcsr_distribution_mp (dist))
    pgrid => dbcsr_mp_pgrid (dbcsr_distribution_mp (dist))
    row_dist => dbcsr_distribution_row_dist (dist)
    col_dist => dbcsr_distribution_col_dist (dist)
    !
    ! let's go
    trace = REAL(0.0,real_4)
    DO a_row = 1, matrix_a%m%nblkrows_total
       a_row_size = row_blk_size(a_row)
       DO a_blk = matrix_a%m%row_p(a_row)+1,matrix_a%m%row_p(a_row+1)
          IF (a_blk .EQ. 0) CYCLE
          a_col = matrix_a%m%col_i(a_blk)
          IF(a_col.ne.a_row) CYCLE
          ! We must skip non-local blocks in a replicated matrix.
          IF(matrix_a%m%replication_type .NE. dbcsr_repl_full) THEN
             IF (mynode .NE. checker_square_proc (a_row, a_col, pgrid,&
                  row_dist, col_dist)) CYCLE
          ENDIF
          a_col_size = col_blk_size(a_col)
          IF(a_row_size.NE.a_col_size)&
             CALL cp__b("dbcsr/ops/dbcsr_operations.F",58,"is that a square matrix?")
          a_nze = a_row_size**2
          a_data => pointer_view (data_p, ABS(matrix_a%m%blk_p(a_blk)),&
               ABS(matrix_a%m%blk_p(a_blk))+a_nze-1)
          !data_a => matrix_a%m%data(ABS(matrix_a%m%blk_p(a_blk)):ABS(matrix_a%m%blk_p(a_blk))+a_nze-1)
          !
          ! let's trace the block
          DO i = 1,a_row_size
             trace = trace + a_data((i-1)*a_row_size+i)
          ENDDO
       ENDDO ! a_col
    ENDDO ! a_row
    !
    ! summe
    CALL mp_sum(trace,dbcsr_mp_group(dbcsr_distribution_mp(matrix_a%m%dist)))

    CALL timestop(error_handle)
  END SUBROUTINE dbcsr_trace_a_c

! **************************************************************************************************
!> \brief traces a product of DBCSR matrices
!> \param[in] matrix_a DBCSR matrices
!> \param[in] matrix_b DBCSR matrices
!> \param[out] trace             the trace of the product of the matrices
!> \param[in] trans_a            (optional) is matrix_a transposed or not?
!> \param[in] trans_b            (optional) is matrix_b transposed or not?
!> \param local_sum ...
! **************************************************************************************************
  SUBROUTINE dbcsr_trace_ab_c(matrix_a, matrix_b, trace, trans_a, trans_b, local_sum)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a, matrix_b
    COMPLEX(kind=real_4), INTENT(INOUT)                   :: trace
    CHARACTER(LEN=*), INTENT(IN), OPTIONAL   :: trans_a, trans_b
    LOGICAL, INTENT(IN), OPTIONAL            :: local_sum

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_trace_ab_c', &
      routineP = moduleN//':'//routineN

    CHARACTER(LEN=1)                         :: my_trans_a, my_trans_b
    INTEGER :: a_blk, a_col, a_col_size, a_row_size, b_blk, b_col_size, &
      b_frst_blk, b_last_blk, b_row_size, nze, row, a_beg, a_end, b_beg, b_end
    INTEGER, DIMENSION(:), POINTER           :: a_col_blk_size, &
                                                a_row_blk_size, &
                                                b_col_blk_size, b_row_blk_size
    COMPLEX(kind=real_4)                                  :: sym_fac, fac
    LOGICAL                                  :: found, my_local_sum
    REAL(real_4), EXTERNAL                   :: SDOT
    REAL(real_8), EXTERNAL                   :: DDOT
    COMPLEX(real_4), EXTERNAL                :: CDOTU
    COMPLEX(real_8), EXTERNAL                :: ZDOTU
    REAL(real_4), DIMENSION(:), POINTER      :: a_data_r, b_data_r
    REAL(real_8), DIMENSION(:), POINTER      :: a_data_d, b_data_d
    COMPLEX(real_4), DIMENSION(:), POINTER   :: a_data_c, b_data_c
    COMPLEX(real_8), DIMENSION(:), POINTER   :: a_data_z, b_data_z

!   ---------------------------------------------------------------------------

    my_local_sum = .FALSE.
    IF(PRESENT(local_sum)) my_local_sum=local_sum
    IF(.NOT.my_local_sum) THEN!fixme
       IF(matrix_a%m%replication_type .NE. dbcsr_repl_none&
            .OR. matrix_b%m%replication_type .NE. dbcsr_repl_none)&
            CALL cp__b("dbcsr/ops/dbcsr_operations.F",119,"Trace of product of replicated matrices not yet possible.")
    ENDIF

    sym_fac = REAL(1.0,real_4)
    IF((dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_antisymmetric).AND.&
       (dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_antisymmetric)) sym_fac = REAL(2.0,real_4)

    ! tracing a symmetric with a general matrix is not implemented, as it would require communication of blocks
    IF((dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_a).EQ.dbcsr_type_antisymmetric).NEQV. &
       (dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_symmetric.OR.&
        dbcsr_get_matrix_type(matrix_b).EQ.dbcsr_type_antisymmetric)) THEN
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",133,"Tracing general with symmetric matrix NYI")
    ENDIF

    a_row_blk_size => array_data (matrix_a%m%row_blk_size)
    a_col_blk_size => array_data (matrix_a%m%col_blk_size)
    b_row_blk_size => array_data (matrix_b%m%row_blk_size)
    b_col_blk_size => array_data (matrix_b%m%col_blk_size)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_r)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_r)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_d)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_d)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_c)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_c)
    CALL dbcsr_get_data (matrix_a%m%data_area, a_data_z)
    CALL dbcsr_get_data (matrix_b%m%data_area, b_data_z)

    my_trans_a = 'T'
    IF(PRESENT(trans_a)) my_trans_a = trans_a
    my_trans_b = 'N'
    IF(PRESENT(trans_b)) my_trans_b = trans_b
    IF(my_trans_a.NE.'T'.OR.my_trans_b.NE.'N')&
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",154,"this combination of transpose is NYI")
    !
    ! let's go
    trace = REAL(0.0,real_4)
    IF(matrix_a%m%nblkrows_total.NE.matrix_b%m%nblkrows_total)&
       CALL cp__b("dbcsr/ops/dbcsr_operations.F",159,"this combination of transpose is NYI")
    DO row = 1, matrix_a%m%nblkrows_total
       a_row_size = a_row_blk_size(row)
       b_row_size = b_row_blk_size(row)
       IF(a_row_size.NE.b_row_size) THEN
          WRITE(*,*) 'a_row_size',a_row_size
          WRITE(*,*) 'b_row_size',b_row_size
          CALL cp__b("dbcsr/ops/dbcsr_operations.F",166,"matrices not consistent")
       ENDIF
       b_blk = matrix_b%m%row_p(row)+1
       b_frst_blk = matrix_b%m%row_p(row)+1
       b_last_blk = matrix_b%m%row_p(row+1)
       DO a_blk = matrix_a%m%row_p(row)+1,matrix_a%m%row_p(row+1)
          IF (matrix_a%m%blk_p(a_blk) .EQ. 0) CYCLE ! Deleted block
          a_col = matrix_a%m%col_i(a_blk)
          a_col_size = a_col_blk_size(a_col)
          !
          ! find the b_blk we assume here that the colums are ordered !
          CALL dbcsr_find_column(a_col,b_frst_blk,b_last_blk,matrix_b%m%col_i,&
               matrix_b%m%blk_p,b_blk,found)
          IF(found) THEN
             b_col_size = b_col_blk_size(a_col)
             IF(a_col_size.NE.b_col_size)  THEN
                WRITE(*,*) 'a_col_size',a_col_size
                WRITE(*,*) 'b_col_size',b_col_size
                CALL cp__b("dbcsr/ops/dbcsr_operations.F",184,"matrices not consistent")
             ENDIF
             !
             nze = a_row_size*a_col_size
             !
             IF(nze.GT.0) THEN
                !
                ! let's trace the blocks
                a_beg = ABS(matrix_a%m%blk_p(a_blk))
                a_end = a_beg + nze - 1
                b_beg = ABS(matrix_b%m%blk_p(b_blk))
                b_end = b_beg + nze - 1
                fac = REAL(1.0,real_4)
                IF(row.NE.a_col) fac = sym_fac
                ! this is a mess, no need to support that..
                IF(    matrix_a%m%data_type.EQ.dbcsr_type_real_4.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_4) THEN
                   trace = trace +CMPLX(fac * SDOT (nze,&
                        a_data_r(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_r(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_4)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_4.AND.&
                     matrix_b%m%data_type.EQ.dbcsr_type_real_8) THEN
                   trace = trace + &
                        CMPLX(fac * SUM ( a_data_r(a_beg:a_end) * b_data_d(b_beg:b_end) ),kind=real_4)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_4) THEN
                   trace = trace + &
                        CMPLX(fac * SUM ( a_data_d(a_beg:a_end) * b_data_r(b_beg:b_end) ),kind=real_4)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_real_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_real_8) THEN
                   trace = trace + CMPLX(fac * DDOT (nze,&
                        a_data_d(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_d(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_4)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_complex_4.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_complex_4) THEN
                   trace = trace + CMPLX(fac * CDOTU (nze,&
                        a_data_c(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_c(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_4)
                ELSEIF(matrix_a%m%data_type.EQ.dbcsr_type_complex_8.AND.&
                       matrix_b%m%data_type.EQ.dbcsr_type_complex_8) THEN
                   trace = trace + CMPLX(fac * ZDOTU (nze,&
                        a_data_z(ABS(matrix_a%m%blk_p(a_blk))),1,&
                        b_data_z(ABS(matrix_b%m%blk_p(b_blk))),1),kind=real_4)
                ELSE
                   CALL cp__b("dbcsr/ops/dbcsr_operations.F",228,"combination of types NYI")
                ENDIF
             ENDIF
          ENDIF
       ENDDO ! a_col
    ENDDO ! a_row
    !
    ! summe
    IF(.NOT.my_local_sum) &
           CALL mp_sum(trace,dbcsr_mp_group(dbcsr_distribution_mp(matrix_a%m%dist)))

  END SUBROUTINE dbcsr_trace_ab_c


! **************************************************************************************************
!> \brief Interface for matrix scaling by a scalar
!> \param matrix_a ...
!> \param alpha_scalar ...
!> \param last_column ...
! **************************************************************************************************
  SUBROUTINE dbcsr_scale_c(matrix_a, alpha_scalar, last_column)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix_a
    COMPLEX(kind=real_4), INTENT(IN)                      :: alpha_scalar
    INTEGER, INTENT(IN), OPTIONAL            :: last_column

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scale_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: error_handler
    TYPE(dbcsr_scalar_type)                  :: sc

    sc = dbcsr_scalar (alpha_scalar)
    CALL dbcsr_scalar_fill_all (sc)
    sc%data_type = dbcsr_get_data_type (matrix_a)
    CALL timeset(routineN, error_handler)
    IF (PRESENT (last_column)) THEN
       CALL dbcsr_scale_anytype(matrix_a,&
            alpha_scalar=sc,&
            limits=(/0,0,0,last_column/))
    ELSE
       CALL dbcsr_scale_anytype(matrix_a, alpha_scalar=sc)
    ENDIF
    CALL timestop(error_handler)
  END SUBROUTINE dbcsr_scale_c

! **************************************************************************************************
!> \brief Interface for matrix scaling by a vector
!> \param matrix_a ...
!> \param alpha ...
!> \param side ...
! **************************************************************************************************
  SUBROUTINE dbcsr_scale_by_vector_c(matrix_a, alpha, side)
    TYPE(dbcsr_obj), INTENT(INOUT)            :: matrix_a
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN), TARGET :: alpha
    CHARACTER(LEN=*), INTENT(IN)              :: side
    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_scale_by_vector_c', &
      routineP = moduleN//':'//routineN
    COMPLEX(kind=real_4), DIMENSION(:), POINTER            :: tmp_p
    TYPE(dbcsr_data_obj)                      :: enc_alpha_vec

    CALL dbcsr_data_init (enc_alpha_vec)
    CALL dbcsr_data_new (enc_alpha_vec, dbcsr_type_complex_4)
    tmp_p => alpha
    CALL dbcsr_data_set_pointer (enc_alpha_vec, tmp_p)
    CALL dbcsr_scale_by_vector_anytype(matrix_a, enc_alpha_vec, side)
    CALL dbcsr_data_clear_pointer (enc_alpha_vec)
    CALL dbcsr_data_release (enc_alpha_vec)
  END SUBROUTINE dbcsr_scale_by_vector_c


! **************************************************************************************************
!> \brief Interface for dbcsr_set
!> \param matrix ...
!> \param alpha ...
! **************************************************************************************************
  SUBROUTINE dbcsr_set_c(matrix, alpha)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    COMPLEX(kind=real_4), INTENT(IN)                      :: alpha
    IF (alpha==CMPLX(0.0, 0.0, real_4)) THEN
      CALL dbcsr_zero(matrix)
    ELSE
      CALL dbcsr_set_anytype(matrix, dbcsr_scalar(alpha))
    ENDIF
  END SUBROUTINE dbcsr_set_c

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param eps ...
!> \param method ...
!> \param use_absolute ...
!> \param filter_diag ...
!> \param quick ...
! **************************************************************************************************
  SUBROUTINE dbcsr_filter_c (matrix, eps, method, use_absolute, &
       filter_diag, quick)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    COMPLEX(kind=real_4), INTENT(IN)                      :: eps
    INTEGER, INTENT(IN), OPTIONAL            :: method
    LOGICAL, INTENT(in), OPTIONAL            :: use_absolute, filter_diag, quick
    CALL dbcsr_filter_anytype (matrix, dbcsr_scalar(eps), method, &
         use_absolute, filter_diag, quick)
  END SUBROUTINE dbcsr_filter_c

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_set_diag_c(matrix, diag)
    TYPE(dbcsr_obj), INTENT(INOUT)            :: matrix
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(IN), TARGET :: diag

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_set_diag_c', &
      routineP = moduleN//':'//routineN

    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: diag_p
    TYPE(dbcsr_data_obj)                     :: diag_a

    diag_p => diag
    CALL dbcsr_data_init (diag_a)
    CALL dbcsr_data_new (diag_a, dbcsr_get_data_type(matrix))
    CALL dbcsr_data_set_pointer (diag_a, diag_p)
    CALL dbcsr_set_diag(matrix, diag_a)
    CALL dbcsr_data_clear_pointer (diag_a)
    CALL dbcsr_data_release (diag_a)
  END SUBROUTINE dbcsr_set_diag_c

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_get_diag_c(matrix, diag)

    TYPE(dbcsr_obj), INTENT(IN)                    :: matrix
    COMPLEX(kind=real_4), DIMENSION(:), INTENT(INOUT), TARGET   :: diag

    CHARACTER(len=*), PARAMETER :: routineN = 'dbcsr_get_diag_c', &
      routineP = moduleN//':'//routineN

    COMPLEX(kind=real_4), DIMENSION(:), POINTER           :: diag_p
    TYPE(dbcsr_data_obj)                     :: diag_a

    diag_p => diag
    CALL dbcsr_data_init (diag_a)
    CALL dbcsr_data_new (diag_a, dbcsr_get_data_type(matrix))
    CALL dbcsr_data_set_pointer (diag_a, diag_p)
    CALL dbcsr_get_diag(matrix, diag_a)
    CALL dbcsr_data_clear_pointer (diag_a)
    CALL dbcsr_data_release (diag_a)
  END SUBROUTINE dbcsr_get_diag_c


! **************************************************************************************************
!> \brief add a constant to the diagonal of a matrix
!> \param[inout] matrix       DBCSR matrix
!> \param[in]    alpha_scalar scalar
!> \param first_row ...
!> \param last_row ...
! **************************************************************************************************
  SUBROUTINE dbcsr_add_on_diag_c(matrix, alpha_scalar, first_row, last_row)
    TYPE(dbcsr_obj), INTENT(INOUT)           :: matrix
    COMPLEX(kind=real_4), INTENT(IN)                      :: alpha_scalar
    integer, intent(in), optional            :: first_row, last_row

    CALL dbcsr_add_on_diag(matrix, dbcsr_scalar(alpha_scalar), first_row, last_row)
  END SUBROUTINE dbcsr_add_on_diag_c

END MODULE dbcsr_operations
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief operations for skinny matrices/vectors expressed in dbcsr form
!> \par History
!>       2014.10 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

MODULE dbcsr_vector_operations_low
   USE dbcsr_block_access,              ONLY: dbcsr_reserve_all_blocks
   USE dbcsr_data_methods,              ONLY: dbcsr_get_data_p
   USE dbcsr_dist_methods,              ONLY: dbcsr_distribution_col_clusters,&
                                              dbcsr_distribution_col_dist,&
                                              dbcsr_distribution_mp,&
                                              dbcsr_distribution_new,&
                                              dbcsr_distribution_release,&
                                              dbcsr_distribution_row_clusters,&
                                              dbcsr_distribution_row_dist
   USE dbcsr_iterator_operations,       ONLY: dbcsr_iterator_blocks_left,&
                                              dbcsr_iterator_next_block,&
                                              dbcsr_iterator_start,&
                                              dbcsr_iterator_stop
   USE dbcsr_methods,                   ONLY: dbcsr_col_block_sizes,&
                                              dbcsr_distribution,&
                                              dbcsr_get_matrix_type,&
                                              dbcsr_init,&
                                              dbcsr_release,&
                                              dbcsr_row_block_sizes
   USE dbcsr_operations,                ONLY: dbcsr_copy,&
                                              dbcsr_get_info,&
                                              dbcsr_set
   USE dbcsr_types,                     ONLY: &
        dbcsr_distribution_obj, dbcsr_iterator, dbcsr_obj, dbcsr_scalar_type, &
        dbcsr_type_antisymmetric, dbcsr_type_complex_4, dbcsr_type_complex_8, &
        dbcsr_type_no_symmetry, dbcsr_type_real_4, dbcsr_type_real_8, dbcsr_type_symmetric
   USE dbcsr_work_operations,           ONLY: dbcsr_create
   USE kinds,                           ONLY: real_4,&
                                              real_8
   USE message_passing,                 ONLY: mp_bcast,&
                                              mp_sum
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

   IMPLICIT NONE

! the following types provide fast access to distributed dbcsr vectors
! *****************************************************************************
!> \brief Types needed for the hashtable.
! *****************************************************************************
  TYPE ele_type
     INTEGER :: c=0
     INTEGER :: p=0
  END TYPE ele_type

  TYPE hash_table_type
     TYPE(ele_type), DIMENSION(:), POINTER :: table
     INTEGER :: nele=0
     INTEGER :: nmax=0
     INTEGER :: prime=0
  END TYPE hash_table_type

   TYPE block_ptr_d
      REAL(real_8), DIMENSION(:, :), POINTER          :: ptr => NULL()
      INTEGER                                         :: assigned_thread
   END TYPE
   TYPE block_ptr_s
      REAL(real_4), DIMENSION(:, :), POINTER          :: ptr => NULL()
      INTEGER                                         :: assigned_thread
   END TYPE
   TYPE block_ptr_c
      COMPLEX(real_4), DIMENSION(:, :), POINTER       :: ptr => NULL()
      INTEGER                                         :: assigned_thread
   END TYPE
   TYPE block_ptr_z
      COMPLEX(real_8), DIMENSION(:, :), POINTER       :: ptr => NULL()
      INTEGER                                         :: assigned_thread
   END TYPE

   TYPE fast_vec_access_type
      TYPE(hash_table_type) :: hash_table
      TYPE(block_ptr_d), DIMENSION(:), ALLOCATABLE :: blk_map_d
      TYPE(block_ptr_s), DIMENSION(:), ALLOCATABLE :: blk_map_s
      TYPE(block_ptr_c), DIMENSION(:), ALLOCATABLE :: blk_map_c
      TYPE(block_ptr_z), DIMENSION(:), ALLOCATABLE :: blk_map_z
   END TYPE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_vector_operations_low'

   PUBLIC :: dbcsr_matrix_colvec_multiply_low, &
             create_col_vec_from_matrix, &
             create_row_vec_from_matrix, &
             create_replicated_col_vec_from_matrix, &
             create_replicated_row_vec_from_matrix

   INTERFACE dbcsr_matrix_colvec_multiply_low
      MODULE PROCEDURE dbcsr_matrix_colvec_multiply_low_any
      MODULE PROCEDURE dbcsr_matrix_colvec_multiply_low_d, dbcsr_matrix_colvec_multiply_low_s, &
         dbcsr_matrix_colvec_multiply_low_c, dbcsr_matrix_colvec_multiply_low_z
   END INTERFACE

CONTAINS

! finally we are here so we can make a final decision which argument types to pass, ... long way to here

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
   SUBROUTINE dbcsr_matrix_colvec_multiply_low_any(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
      TYPE(dbcsr_obj)                                    :: matrix, vec_in, vec_out
      TYPE(dbcsr_scalar_type)                            :: alpha, beta
      TYPE(dbcsr_obj)                                    :: work_row, work_col

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_colvec_multiply_low_any', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, matrix_type

      CALL timeset(routineN, handle)

      matrix_type = matrix%m%data_type
      SELECT CASE (matrix_type)
      CASE (dbcsr_type_real_8)
         CALL dbcsr_matrix_colvec_multiply_low(matrix, vec_in, vec_out, alpha%r_dp, beta%r_dp, work_row, work_col)
      CASE (dbcsr_type_real_4)
         CALL dbcsr_matrix_colvec_multiply_low(matrix, vec_in, vec_out, alpha%r_sp, beta%r_sp, work_row, work_col)
      CASE (dbcsr_type_complex_8)
         CALL dbcsr_matrix_colvec_multiply_low(matrix, vec_in, vec_out, alpha%c_dp, beta%c_dp, work_row, work_col)
      CASE (dbcsr_type_complex_4)
         CALL dbcsr_matrix_colvec_multiply_low(matrix, vec_in, vec_out, alpha%c_sp, beta%c_sp, work_row, work_col)
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE dbcsr_matrix_colvec_multiply_low_any

! **************************************************************************************************
!> \brief creates a dbcsr col vector like object which lives on proc_col 0
!>        and has the same row dist as the template matrix
!>        the returned matrix is fully allocated and all blocks are set to 0
!>        this is not a sparse object (and must never be)
!> \param dbcsr_vec  the vector object to create must be allocated but not initialized
!> \param matrix a dbcsr matrix used as template
!> \param ncol number of vectors in the dbcsr_object (1 for vector, n for skinny matrix)
! **************************************************************************************************
   SUBROUTINE create_col_vec_from_matrix(dbcsr_vec, matrix, ncol)
      TYPE(dbcsr_obj)                                    :: dbcsr_vec, matrix
      INTEGER                                            :: ncol

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_col_vec_from_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, npcol
      INTEGER, DIMENSION(:), POINTER                     :: col_sizes, dist_col
      TYPE(dbcsr_distribution_obj)                       :: dist_col_vec, distri

      CALL timeset(routineN, handle)

      CALL dbcsr_get_info(matrix=matrix, distribution=distri)
      npcol = SIZE(distri%d%mp_env%mp%pgrid, 2)
      ALLOCATE (dist_col(1)); ALLOCATE (col_sizes(1))
      dist_col = 0; col_sizes = ncol

      CALL dbcsr_distribution_new(dist_col_vec, &
                                  dbcsr_distribution_mp(dbcsr_distribution(matrix)), &
                                  dbcsr_distribution_row_dist(dbcsr_distribution(matrix)), &
                                  dist_col, &
                                  row_dist_cluster=dbcsr_distribution_row_clusters(dbcsr_distribution(matrix)))
      CALL dbcsr_init(dbcsr_vec)
      CALL dbcsr_create(dbcsr_vec, "D", dist_col_vec, dbcsr_type_no_symmetry, &
                        dbcsr_row_block_sizes(matrix), col_sizes, &
                        nze=0, data_type=matrix%m%data_type)

      CALL dbcsr_reserve_all_blocks(dbcsr_vec)

      CALL dbcsr_distribution_release(dist_col_vec)
      DEALLOCATE (dist_col, col_sizes)

      CALL timestop(handle)

   END SUBROUTINE create_col_vec_from_matrix

! **************************************************************************************************
!> \brief creates a dbcsr row vector like object which lives on proc_row 0
!>        and has the same row dist as the template matrix
!>        the returned matrix is fully allocated and all blocks are set to 0
!>        this is not a sparse object (and must never be)
!> \param dbcsr_vec ...
!> \param matrix a dbcsr matrix used as template
!> \param nrow number of vectors in the dbcsr_object (1 for vector, n for skinny matrix)
! **************************************************************************************************
   SUBROUTINE create_row_vec_from_matrix(dbcsr_vec, matrix, nrow)
      TYPE(dbcsr_obj)                                    :: dbcsr_vec, matrix
      INTEGER                                            :: nrow

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_row_vec_from_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nprow
      INTEGER, DIMENSION(:), POINTER                     :: dist_row, row_sizes
      TYPE(dbcsr_distribution_obj)                       :: dist_row_vec, distri

      CALL timeset(routineN, handle)

      CALL dbcsr_get_info(matrix=matrix, distribution=distri)
      nprow = SIZE(distri%d%mp_env%mp%pgrid, 1)
      ALLOCATE (dist_row(1)); ALLOCATE (row_sizes(1))
      dist_row = 0; row_sizes = nrow

      CALL dbcsr_distribution_new(dist_row_vec, &
                                  dbcsr_distribution_mp(dbcsr_distribution(matrix)), &
                                  dist_row, &
                                  dbcsr_distribution_col_dist(dbcsr_distribution(matrix)), &
                                  col_dist_cluster=dbcsr_distribution_col_clusters(dbcsr_distribution(matrix)))

      CALL dbcsr_init(dbcsr_vec)
      CALL dbcsr_create(dbcsr_vec, "D", dist_row_vec, dbcsr_type_no_symmetry, &
                        row_sizes, dbcsr_col_block_sizes(matrix), &
                        nze=0, data_type=matrix%m%data_type)

      CALL dbcsr_reserve_all_blocks(dbcsr_vec)

      CALL dbcsr_distribution_release(dist_row_vec)
      DEALLOCATE (dist_row, row_sizes)

      CALL timestop(handle)

   END SUBROUTINE create_row_vec_from_matrix

! **************************************************************************************************
!> \brief creates a col vector like object whose blocks can be replicated
!>        along the processor row and has the same row dist as the template matrix
!>        the returned matrix is fully allocated and all blocks are set to 0
!>        this is not a sparse object (and must never be)
!> \param dbcsr_vec the vector object to create must be allocated but not initialized
!> \param matrix a dbcsr matrix used as template
!> \param ncol number of vectors in the dbcsr_object (1 for vector, n for skinny matrix)
! **************************************************************************************************
   SUBROUTINE create_replicated_col_vec_from_matrix(dbcsr_vec, matrix, ncol)
      TYPE(dbcsr_obj)                                    :: dbcsr_vec, matrix
      INTEGER                                            :: ncol

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_replicated_col_vec_from_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, npcol
      INTEGER, DIMENSION(:), POINTER                     :: col_sizes, dist_col
      TYPE(dbcsr_distribution_obj)                       :: dist_col_vec, distri

      CALL timeset(routineN, handle)

      CALL dbcsr_get_info(matrix=matrix, distribution=distri)
      npcol = SIZE(distri%d%mp_env%mp%pgrid, 2)
      ALLOCATE (dist_col(npcol)); ALLOCATE (col_sizes(npcol))
      col_sizes = ncol
      DO i = 0, npcol-1
         dist_col(i+1) = i
      END DO

      CALL dbcsr_distribution_new(dist_col_vec, &
                                  dbcsr_distribution_mp(dbcsr_distribution(matrix)), &
                                  dbcsr_distribution_row_dist(dbcsr_distribution(matrix)), &
                                  dist_col, &
                                  row_dist_cluster=dbcsr_distribution_row_clusters(dbcsr_distribution(matrix)))

      CALL dbcsr_init(dbcsr_vec)
      CALL dbcsr_create(dbcsr_vec, "D", dist_col_vec, dbcsr_type_no_symmetry, &
                        dbcsr_row_block_sizes(matrix), col_sizes, &
                        nze=0, data_type=matrix%m%data_type)

      CALL dbcsr_reserve_all_blocks(dbcsr_vec)

      CALL dbcsr_distribution_release(dist_col_vec)
      DEALLOCATE (dist_col, col_sizes)

      CALL timestop(handle)

   END SUBROUTINE create_replicated_col_vec_from_matrix

! **************************************************************************************************
!> \brief creates a row vector like object whose blocks can be replicated
!>        along the processor col and has the same col dist as the template matrix
!>        the returned matrix is fully allocated and all blocks are set to 0
!>        this is not a sparse object (and must never be)
!> \param dbcsr_vec the vector object to create must be allocated but not initialized
!> \param matrix a dbcsr matrix used as template
!> \param nrow number of vectors in the dbcsr_object (1 for vector, n for skinny matrix)
! **************************************************************************************************
   SUBROUTINE create_replicated_row_vec_from_matrix(dbcsr_vec, matrix, nrow)
      TYPE(dbcsr_obj)                                    :: dbcsr_vec, matrix
      INTEGER                                            :: nrow

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_replicated_row_vec_from_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, nprow
      INTEGER, DIMENSION(:), POINTER                     :: dist_row, row_sizes
      TYPE(dbcsr_distribution_obj)                       :: dist_row_vec, distri

      CALL timeset(routineN, handle)

      CALL dbcsr_get_info(matrix=matrix, distribution=distri)
      nprow = SIZE(distri%d%mp_env%mp%pgrid, 1)
      ALLOCATE (dist_row(nprow)); ALLOCATE (row_sizes(nprow))
      row_sizes = nrow
      DO i = 0, nprow-1
         dist_row(i+1) = i
      END DO

      CALL dbcsr_distribution_new(dist_row_vec, &
                                  dbcsr_distribution_mp(dbcsr_distribution(matrix)), &
                                  dist_row, &
                                  dbcsr_distribution_col_dist(dbcsr_distribution(matrix)), &
                                  col_dist_cluster=dbcsr_distribution_col_clusters(dbcsr_distribution(matrix)))

      CALL dbcsr_init(dbcsr_vec)
      CALL dbcsr_create(dbcsr_vec, "D", dist_row_vec, dbcsr_type_no_symmetry, &
                        row_sizes, dbcsr_col_block_sizes(matrix), &
                        nze=0, data_type=matrix%m%data_type)

      CALL dbcsr_reserve_all_blocks(dbcsr_vec)

      CALL dbcsr_distribution_release(dist_row_vec)
      DEALLOCATE (dist_row, row_sizes)

      CALL timestop(handle)

   END SUBROUTINE create_replicated_row_vec_from_matrix

! **************************************************************************************************
!> \brief given a column vector, prepare the fast_vec_access container
!> \param vec ...
!> \param fast_vec_access ...
! **************************************************************************************************
   SUBROUTINE create_fast_col_vec_access(vec, fast_vec_access)
      TYPE(dbcsr_obj)                                    :: vec
      TYPE(fast_vec_access_type)                         :: fast_vec_access

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_fast_col_vec_access', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      SELECT CASE (vec%m%data_type)
      CASE (dbcsr_type_real_8)
         CALL create_fast_col_vec_access_d(vec, fast_vec_access)
      CASE (dbcsr_type_real_4)
         CALL create_fast_col_vec_access_s(vec, fast_vec_access)
      CASE (dbcsr_type_complex_8)
         CALL create_fast_col_vec_access_z(vec, fast_vec_access)
      CASE (dbcsr_type_complex_4)
         CALL create_fast_col_vec_access_c(vec, fast_vec_access)
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE create_fast_col_vec_access

! **************************************************************************************************
!> \brief given a row vector, prepare the fast_vec_access_container
!> \param vec ...
!> \param fast_vec_access ...
! **************************************************************************************************
   SUBROUTINE create_fast_row_vec_access(vec, fast_vec_access)
      TYPE(dbcsr_obj)                                    :: vec
      TYPE(fast_vec_access_type)                         :: fast_vec_access

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_fast_row_vec_access', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      SELECT CASE (vec%m%data_type)
      CASE (dbcsr_type_real_8)
         CALL create_fast_row_vec_access_d(vec, fast_vec_access)
      CASE (dbcsr_type_real_4)
         CALL create_fast_row_vec_access_s(vec, fast_vec_access)
      CASE (dbcsr_type_complex_8)
         CALL create_fast_row_vec_access_c(vec, fast_vec_access)
      CASE (dbcsr_type_complex_4)
         CALL create_fast_row_vec_access_z(vec, fast_vec_access)
      END SELECT

      CALL timestop(handle)

   END SUBROUTINE create_fast_row_vec_access

! **************************************************************************************************
!> \brief release all memory associated with the fast_vec_access type
!> \param fast_vec_access ...
! **************************************************************************************************
   SUBROUTINE release_fast_vec_access(fast_vec_access)
      TYPE(fast_vec_access_type)                         :: fast_vec_access

      CHARACTER(LEN=*), PARAMETER :: routineN = 'release_fast_vec_access', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      CALL hash_table_release(fast_vec_access%hash_table)
      IF (ALLOCATED(fast_vec_access%blk_map_d)) DEALLOCATE (fast_vec_access%blk_map_d)
      IF (ALLOCATED(fast_vec_access%blk_map_s)) DEALLOCATE (fast_vec_access%blk_map_s)
      IF (ALLOCATED(fast_vec_access%blk_map_c)) DEALLOCATE (fast_vec_access%blk_map_c)
      IF (ALLOCATED(fast_vec_access%blk_map_z)) DEALLOCATE (fast_vec_access%blk_map_z)

      CALL timestop(handle)

   END SUBROUTINE release_fast_vec_access

! -----------------------------------------------------------------------------
! Beginning of hashtable.
! this file can be 'INCLUDE'ed verbatim in various place, where it needs to be
! part of the module to guarantee inlining
! hashes (c,p) pairs, where p is assumed to be >0
! on return (0 is used as a flag for not present)
!
!
! *****************************************************************************
!> \brief finds a prime equal or larger than i, needed at creation
!> \param i ...
!> \retval res ...
! *****************************************************************************
  FUNCTION matching_prime(i) RESULT(res)
    INTEGER, INTENT(IN)                      :: i
    INTEGER                                  :: res

    INTEGER                                  :: j

    res=i
    j=0
    DO WHILE (j<res)
      DO j=2,res-1
         IF (MOD(res,j)==0) THEN
            res=res+1
            EXIT
         ENDIF
      ENDDO
    ENDDO
  END FUNCTION

! *****************************************************************************
!> \brief create a hash_table of given initial size. 
!>        the hash table will expand as needed (but this requires rehashing)
!> \param hash_table ...
!> \param table_size ...
! *****************************************************************************
  SUBROUTINE hash_table_create(hash_table,table_size)
    TYPE(hash_table_type)                    :: hash_table
    INTEGER, INTENT(IN)                      :: table_size

    INTEGER                                  :: j

    ! guarantee a minimal hash table size (8), so that expansion works

    j=3
    DO WHILE(2**j-1<table_size)
       j=j+1
    ENDDO
    hash_table%nmax=2**j-1
    hash_table%prime=matching_prime(hash_table%nmax)
    hash_table%nele=0
    ALLOCATE(hash_table%table(0:hash_table%nmax))
  END SUBROUTINE hash_table_create

! *****************************************************************************
!> \brief ...
!> \param hash_table ...
! *****************************************************************************
  SUBROUTINE hash_table_release(hash_table)
    TYPE(hash_table_type)                    :: hash_table

    hash_table%nmax=0
    hash_table%nele=0
    DEALLOCATE(hash_table%table)

  END SUBROUTINE hash_table_release

! *****************************************************************************
!> \brief add a pair (c,p) to the hash table
!> \param hash_table ...
!> \param c this value is being hashed
!> \param p this is being stored
! *****************************************************************************
  RECURSIVE SUBROUTINE hash_table_add(hash_table,c,p)
    TYPE(hash_table_type), INTENT(INOUT)     :: hash_table
    INTEGER, INTENT(IN)                      :: c, p

    REAL(KIND=real_8), PARAMETER :: hash_table_expand = 1.5_real_8, &
      inv_hash_table_fill = 2.5_real_8

    INTEGER                                  :: i, j
    TYPE(ele_type), ALLOCATABLE, &
      DIMENSION(:)                           :: tmp_hash

! if too small, make a copy and rehash in a larger table

    IF (hash_table%nele*inv_hash_table_fill>hash_table%nmax) THEN
       ALLOCATE(tmp_hash(LBOUND(hash_table%table,1):UBOUND(hash_table%table,1)))
       tmp_hash(:)=hash_table%table
       CALL hash_table_release(hash_table)
       CALL hash_table_create(hash_table,INT((UBOUND(tmp_hash,1)+8)*hash_table_expand))
       DO i=LBOUND(tmp_hash,1),UBOUND(tmp_hash,1)
          IF (tmp_hash(i)%c.NE.0) THEN
             CALL hash_table_add(hash_table,tmp_hash(i)%c,tmp_hash(i)%p)
          ENDIF
       ENDDO
       DEALLOCATE(tmp_hash)
    ENDIF

    hash_table%nele=hash_table%nele+1
    i=IAND(c*hash_table%prime,hash_table%nmax)

    DO j=i,hash_table%nmax
       IF (hash_table%table(j)%c==0 .OR. hash_table%table(j)%c==c) THEN
          hash_table%table(j)%c=c
          hash_table%table(j)%p=p
          RETURN
       ENDIF
    ENDDO
    DO j=0,i-1
       IF (hash_table%table(j)%c==0 .OR. hash_table%table(j)%c==c) THEN
          hash_table%table(j)%c=c
          hash_table%table(j)%p=p
          RETURN
       ENDIF
    ENDDO

  END SUBROUTINE hash_table_add

! *****************************************************************************
!> \brief ...
!> \param hash_table ...
!> \param c ...
!> \retval p ...
! *****************************************************************************
  PURE FUNCTION hash_table_get(hash_table,c) RESULT(p)
    TYPE(hash_table_type), INTENT(IN)        :: hash_table
    INTEGER, INTENT(IN)                      :: c
    INTEGER                                  :: p

    INTEGER                                  :: i, j

    i=IAND(c*hash_table%prime,hash_table%nmax)

    ! catch the likely case first
    IF (hash_table%table(i)%c==c) THEN
       p=hash_table%table(i)%p
       RETURN
    ENDIF

    DO j=i,hash_table%nmax
       IF (hash_table%table(j)%c==0 .OR. hash_table%table(j)%c==c) THEN
          p=hash_table%table(j)%p
          RETURN
       ENDIF
    ENDDO
    DO j=0,i-1
       IF (hash_table%table(j)%c==0 .OR. hash_table%table(j)%c==c) THEN
          p=hash_table%table(j)%p
          RETURN
       ENDIF
    ENDDO

    ! we should never reach this point.
    p=HUGE(p)

  END FUNCTION hash_table_get

! End of hashtable
! -----------------------------------------------------------------------------

! **************************************************************************************************
!> \brief the real driver routine for the multiply, not all symmetries implemented yet
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrix_colvec_multiply_low_d(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    REAL(kind=real_8)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_colvec_multiply_low', &
      routineP = moduleN//':'//routineN

    CHARACTER                                :: matrix_type

    matrix_type=dbcsr_get_matrix_type(matrix)
    SELECT CASE(matrix_type)
    CASE(dbcsr_type_no_symmetry)
       CALL dbcsr_matrix_vector_mult_d(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    CASE(dbcsr_type_symmetric)
       CALL dbcsr_sym_matrix_vector_mult_d(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    CASE(dbcsr_type_antisymmetric)
        ! Not yet implemented, should mainly be some prefactor magic, but who knows how antisymmetric matrices are stored???
       CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",29,"NYI, antisymmetric matrix not permitted")
    CASE DEFAULT
       CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",31,"Unknown matrix type, ...")
    END SELECT

  END SUBROUTINE dbcsr_matrix_colvec_multiply_low_d

! **************************************************************************************************
!> \brief low level routines for matrix vector multiplies
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrix_vector_mult_d(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    REAL(kind=real_8)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_vector_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, &
                                                myprow, &
                                                ncols, pcol_group, nrows, &
                                                prow_group, row, &
                                                handle, handle1, ithread 
    LOGICAL                                  :: transposed
    REAL(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    REAL(kind=real_8), DIMENSION(:, :), POINTER       :: data_d, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter
    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col
    INTEGER                                  :: prow, pcol

    CALL timeset(routineN, handle)
    ithread=0

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group 
    pcol_group=distri%d%mp_env%mp%pcol_group 
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)

! Transfer the correct parts of the input vector to the correct locations so we can do a local multiply
    CALL dbcsr_col_vec_to_rep_row_d(vec_in, work_col, work_row, fast_vec_col)

! Set the work vector for the results to 0
    CALL dbcsr_set(work_col, 0.0_real_8)

! Perform the local multiply. Here we exploit, that we have the blocks replicated on the mpi processes
! It is important to note, that the input and result vector are distributed differently (row wise, col wise respectively)
    CALL timeset(routineN//"_local_mm", handle1)

!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,transposed,ithread,pcol,prow) &
!$OMP          SHARED(matrix,fast_vec_col,fast_vec_row)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(fast_vec_col%blk_map_d(prow)%assigned_thread .NE. ithread ) CYCLE
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       fast_vec_col%blk_map_d(prow)%ptr=fast_vec_col%blk_map_d(prow)%ptr+&
            MATMUL(data_d,TRANSPOSE(fast_vec_row%blk_map_d(pcol)%ptr))
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

! sum all the data onto the first processor col where the original vector is stored
    data_vec => dbcsr_get_data_p (work_col%m%data_area, select_data_type=0.0_real_8)
    CALL dbcsr_get_info(matrix=work_col, nfullrows_local=nrows, nfullcols_local=ncols)
    CALL mp_sum(data_vec(1:nrows*ncols), prow_group)

! Local copy on the first mpi col (as this is the localtion of the vec_res blocks) of the result vector
! from the replicated to the original vector. Let's play it safe and use the iterator
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_d(prow)%ptr)) THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*fast_vec_col%blk_map_d(prow)%ptr(:,:)
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL release_fast_vec_access(fast_vec_row)
    CALL release_fast_vec_access(fast_vec_col)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_matrix_vector_mult_d

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
!> \param skip_diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrixT_vector_mult_d(matrix, vec_in, vec_out, alpha, beta, work_row, work_col, skip_diag)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    REAL(kind=real_8)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col
    LOGICAL                                  :: skip_diag

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrixT_vector_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, col_size, mypcol, &
                                                myprow, &
                                                ncols, pcol_group, nrows, &
                                                prow_group, row, row_size, &
                                                handle, handle1, ithread
    LOGICAL                                  :: transposed
    REAL(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    REAL(kind=real_8), DIMENSION(:, :), POINTER       :: data_d, vec_bl, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col
    INTEGER                                  :: prow, pcol

    CALL timeset(routineN, handle)
    ithread=0

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group; pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol; myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)

! Set the work vector for the results to 0
    CALL dbcsr_set(work_row, 0.0_real_8)

! Transfer the correct parts of the input vector to the replicated vector on proc_col 0
    CALL dbcsr_iterator_start(iter, vec_in)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed, row_size=row_size, col_size=col_size)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       fast_vec_col%blk_map_d(prow)%ptr(1:row_size, 1:col_size)= vec_bl(1:row_size, 1:col_size)
    END DO
    CALL dbcsr_iterator_stop(iter)
! Replicate the data on all processore in the row
    data_vec => dbcsr_get_data_p (work_col%m%data_area, select_data_type=0.0_real_8)
    CALL mp_bcast(data_vec, 0, prow_group)

! Perform the local multiply. Here it is obvious why the vectors are replicated on the mpi rows and cols
    CALL timeset(routineN//"local_mm", handle1)
    CALL dbcsr_get_info(matrix=work_col, nfullcols_local=ncols)
!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,row_size,col_size,transposed,ithread,prow,pcol) &
!$OMP          SHARED(matrix,fast_vec_row,fast_vec_col,skip_diag,ncols)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed, row_size=row_size, col_size=col_size)
       IF(skip_diag.AND.col==row)CYCLE
       prow=hash_table_get(fast_vec_col%hash_table,row)
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       IF ( ASSOCIATED(fast_vec_row%blk_map_d(pcol)%ptr) .AND. &
            ASSOCIATED(fast_vec_col%blk_map_d(prow)%ptr) )THEN
          IF(fast_vec_row%blk_map_d(pcol)%assigned_thread .NE. ithread ) CYCLE
          fast_vec_row%blk_map_d(pcol)%ptr=fast_vec_row%blk_map_d(pcol)%ptr+&
               MATMUL(TRANSPOSE(fast_vec_col%blk_map_d(prow)%ptr),data_d)
       ELSE
          prow=hash_table_get(fast_vec_row%hash_table,row)
          pcol=hash_table_get(fast_vec_col%hash_table,col)
          IF(fast_vec_row%blk_map_d(prow)%assigned_thread .NE. ithread ) CYCLE
          fast_vec_row%blk_map_d(prow)%ptr=fast_vec_row%blk_map_d(prow)%ptr+&
             MATMUL(TRANSPOSE(fast_vec_col%blk_map_d(pcol)%ptr),TRANSPOSE(data_d))
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

! sum all the data within a processor column to obtain the replicated result
    data_vec => dbcsr_get_data_p (work_row%m%data_area, select_data_type=0.0_real_8)
! we use the replicated vector but the final answer is only summed to proc_col 0 for efficiency
    CALL dbcsr_get_info(matrix=work_row, nfullrows_local=nrows, nfullcols_local=ncols)
    CALL mp_sum(data_vec(1:nrows*ncols), pcol_group)

! Convert the result to a column wise distribution
    CALL dbcsr_rep_row_to_rep_col_vec_d(work_col, work_row, fast_vec_row)
    
! Create_the final vector by summing it to the result vector which lives on proc_col 0
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed, row_size=row_size)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_d(prow)%ptr)) THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*fast_vec_col%blk_map_d(prow)%ptr(:,:)
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_matrixT_vector_mult_d 

! **************************************************************************************************
!> \brief ...
!> \param vec_in ...
!> \param rep_col_vec ...
!> \param rep_row_vec ...
!> \param fast_vec_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_col_vec_to_rep_row_d(vec_in, rep_col_vec, rep_row_vec, fast_vec_col)
    TYPE(dbcsr_obj)                          :: vec_in, rep_col_vec, &
                                                rep_row_vec
    TYPE(fast_vec_access_type), INTENT(IN)   :: fast_vec_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_col_vec_to_rep_row', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, myprow, ncols, &
                                                nrows, pcol_group, &
                                                prow_group, row, handle
    INTEGER, DIMENSION(:), POINTER           :: local_cols, row_dist
    LOGICAL                                  :: transposed
    REAL(kind=real_8), DIMENSION(:), POINTER          :: data_vec, data_vec_rep
    REAL(kind=real_8), DIMENSION(:, :), POINTER       :: vec_row
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

! get information about the parallel environment
    CALL dbcsr_get_info(matrix=vec_in, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group
    pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow

! Get the vector which tells us which blocks are local to which processor row in the col vec
    row_dist=> dbcsr_distribution_row_dist (dbcsr_distribution(rep_col_vec))

! Copy the local vector to the replicated on the first processor column (this is where vec_in lives)
    CALL dbcsr_get_info(matrix=rep_col_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_col_vec%m%data_area, select_data_type=0.0_real_8)
    data_vec => dbcsr_get_data_p (vec_in%m%data_area, select_data_type=0.0_real_8)
    IF(mypcol==0)data_vec_rep(1:nrows*ncols)=data_vec(1:nrows*ncols)
! Replicate the data along the row
    CALL mp_bcast(data_vec_rep(1:nrows*ncols), 0, prow_group)

! Here it gets a bit tricky as we are dealing with two different parallel layouts:
! The rep_col_vec contains all blocks local to the row distribution of the vector. 
! The rep_row_vec only needs the fraction which is local to the col distribution.
! However in most cases this won't the complete set of block which can be obtained from col_vector p_row i
! Anyway, as the blocks don't repeat in the col_vec, a different fraction of the row vec will be available
! on every replica in the processor column, by summing along the column we end up with the complete vector everywhere
! Hope this clarifies the idea
    CALL dbcsr_set(rep_row_vec, 0.0_real_8)
    CALL dbcsr_get_info(matrix=rep_row_vec, nfullrows_local=nrows, local_cols=local_cols, nfullcols_local=ncols)
    CALL dbcsr_iterator_start(iter, rep_row_vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_row, transposed)
       IF(row_dist(col)==myprow)THEN
          vec_row=TRANSPOSE(fast_vec_col%blk_map_d(hash_table_get(fast_vec_col%hash_table,col))%ptr)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)
    CALL dbcsr_get_info(matrix=rep_row_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_row_vec%m%data_area, select_data_type=0.0_real_8)
    CALL mp_sum(data_vec_rep(1:ncols*nrows), pcol_group)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_col_vec_to_rep_row_d    
       
! **************************************************************************************************
!> \brief ...
!> \param rep_col_vec ...
!> \param rep_row_vec ...
!> \param fast_vec_row ...
!> \param fast_vec_col_add ...
! **************************************************************************************************
  SUBROUTINE dbcsr_rep_row_to_rep_col_vec_d(rep_col_vec, rep_row_vec, fast_vec_row, fast_vec_col_add)
    TYPE(dbcsr_obj)                          :: rep_col_vec, rep_row_vec
    TYPE(fast_vec_access_type), OPTIONAL     :: fast_vec_col_add
    TYPE(fast_vec_access_type)               :: fast_vec_row

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_rep_row_to_rep_col_vec', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, myprow, ncols, &
                                                nrows, pcol_group, &
                                                prow_group, row, handle
    INTEGER, DIMENSION(:), POINTER           :: col_dist
    LOGICAL                                  :: transposed
    REAL(kind=real_8), DIMENSION(:), POINTER          :: data_vec_rep
    REAL(kind=real_8), DIMENSION(:, :), POINTER       :: vec_col
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

! get information about the parallel environment
    CALL dbcsr_get_info(matrix=rep_col_vec, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group
    pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow
! Get the vector which tells us which blocks are local to which processor col in the row vec
    col_dist=> dbcsr_distribution_col_dist (dbcsr_distribution(rep_row_vec))

! The same trick as described above with opposite direction
    CALL dbcsr_set(rep_col_vec, 0.0_real_8)
    CALL dbcsr_iterator_start(iter, rep_col_vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_col, transposed)
       IF(col_dist(row)==mypcol)THEN
          vec_col=TRANSPOSE(fast_vec_row%blk_map_d(hash_table_get(fast_vec_row%hash_table,row))%ptr)
       END IF
       ! this one is special and allows to add the elements of a not yet summed replicated 
       ! column vector as it appears in M*V(row_rep) as result. Save an mp_sum in the symmetric case
       IF(PRESENT(fast_vec_col_add))vec_col=vec_col+&
            fast_vec_col_add%blk_map_d(hash_table_get(fast_vec_col_add%hash_table,row))%ptr(:,:)
    END DO
    CALL dbcsr_iterator_stop(iter)
    CALL dbcsr_get_info(matrix=rep_col_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_col_vec%m%data_area, select_data_type=0.0_real_8)
    CALL mp_sum(data_vec_rep(1:nrows*ncols), prow_group)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_rep_row_to_rep_col_vec_d


! **************************************************************************************************
!> \brief given a column vector, prepare the fast_vec_access container
!> \param vec ...
!> \param fast_vec_access ...
! **************************************************************************************************
  SUBROUTINE create_fast_col_vec_access_d(vec, fast_vec_access)
    TYPE(dbcsr_obj)                          :: vec
    TYPE(fast_vec_access_type)               :: fast_vec_access

    CHARACTER(LEN=*), PARAMETER :: routineN = 'create_fast_col_vec_access_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, nblk_local
    INTEGER                                  :: col, row, iblock, nthreads
    LOGICAL                                  :: transposed
    REAL(kind=real_8), DIMENSION(:, :), POINTER       :: vec_bl
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

    ! figure out the number of threads
    nthreads = 1
!$OMP PARALLEL DEFAULT(NONE) SHARED(nthreads)
!$OMP MASTER
    !$ nthreads = OMP_GET_NUM_THREADS()
!$OMP END MASTER
!$OMP END PARALLEL

    CALL dbcsr_get_info(matrix=vec, nblkrows_local=nblk_local)
    ! 4 times makes sure the table is big enough to limit collisions.
    CALL hash_table_create(fast_vec_access%hash_table,4*nblk_local)
    ! include zero for effective dealing with values not in the hash table (will return 0)
    ALLOCATE(fast_vec_access%blk_map_d(0:nblk_local))

    CALL dbcsr_get_info(matrix=vec, nblkcols_local=col)
    IF (col.GT.1) CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",413,"BUG")

    ! go through the blocks of the vector
    iblock=0
    CALL dbcsr_iterator_start(iter, vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed)
       iblock=iblock+1
       CALL hash_table_add(fast_vec_access%hash_table,row,iblock)
       fast_vec_access%blk_map_d(iblock)%ptr=>vec_bl
       fast_vec_access%blk_map_d(iblock)%assigned_thread=MOD(iblock,nthreads)
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE create_fast_col_vec_access_d

! **************************************************************************************************
!> \brief given a row vector, prepare the fast_vec_access_container
!> \param vec ...
!> \param fast_vec_access ...
! **************************************************************************************************
  SUBROUTINE create_fast_row_vec_access_d(vec, fast_vec_access)
    TYPE(dbcsr_obj)                          :: vec
    TYPE(fast_vec_access_type)               :: fast_vec_access

    CHARACTER(LEN=*), PARAMETER :: routineN = 'create_fast_row_vec_access_d', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, nblk_local
    INTEGER                                  :: col, row, iblock, nthreads
    LOGICAL                                  :: transposed
    REAL(kind=real_8), DIMENSION(:, :), POINTER       :: vec_bl
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

    ! figure out the number of threads
    nthreads = 1
!$OMP PARALLEL DEFAULT(NONE) SHARED(nthreads)
!$OMP MASTER
    !$ nthreads = OMP_GET_NUM_THREADS()
!$OMP END MASTER
!$OMP END PARALLEL

    CALL dbcsr_get_info(matrix=vec, nblkcols_local=nblk_local)
    ! 4 times makes sure the table is big enough to limit collisions.
    CALL hash_table_create(fast_vec_access%hash_table,4*nblk_local)
    ! include zero for effective dealing with values not in the hash table (will return 0)
    ALLOCATE(fast_vec_access%blk_map_d(0:nblk_local))

    ! sanity check
    CALL dbcsr_get_info(matrix=vec, nblkrows_local=row)
    IF (row.GT.1) CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",467,"BUG")

    ! go through the blocks of the vector
    iblock=0
    CALL dbcsr_iterator_start(iter, vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed)
       iblock=iblock+1
       CALL hash_table_add(fast_vec_access%hash_table,col,iblock)
       fast_vec_access%blk_map_d(iblock)%ptr=>vec_bl
       fast_vec_access%blk_map_d(iblock)%assigned_thread=MOD(iblock,nthreads)
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE create_fast_row_vec_access_d

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_sym_matrix_vector_mult_d(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    REAL(kind=real_8)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_sym_m_v_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, &
                                                myprow, &
                                                pcol_group, nrows, ncols,&
                                                prow_group, row, &
                                                handle, handle1, ithread, vec_dim
    LOGICAL                                  :: transposed
    REAL(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    REAL(kind=real_8), DIMENSION(:, :), POINTER       :: data_d, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter
    TYPE(dbcsr_obj)                          :: result_row, result_col

    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col, res_fast_vec_row, res_fast_vec_col
    INTEGER                                  :: prow, pcol, rprow, rpcol

    CALL timeset(routineN, handle)
    ithread=0
! We need some work matrices as we try to exploit operations on the replicated vectors which are duplicated otherwise
    CALL dbcsr_get_info(matrix=vec_in,nfullcols_total=vec_dim)
! This is a performance hack as the new creation of a replicated vector is a fair bit more expensive
    CALL dbcsr_init(result_col)
    CALL dbcsr_set(work_col, 0.0_real_8)
    CALL dbcsr_copy(result_col, work_col)
    CALL dbcsr_init(result_row)
    CALL dbcsr_set(work_row, 0.0_real_8)
    CALL dbcsr_copy(result_row, work_row)

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group; pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol; myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)
    CALL create_fast_row_vec_access(result_row, res_fast_vec_row)
    CALL create_fast_col_vec_access(result_col, res_fast_vec_col)

! Transfer the correct parts of the input vector to the correct locations so we can do a local multiply
    CALL dbcsr_col_vec_to_rep_row_d(vec_in, work_col, work_row, fast_vec_col)

! Probably I should rename the routine above as it delivers both the replicated row and column vector

! Perform the local multiply. Here we exploit, that we have the blocks replicated on the mpi processes
! It is important to note, that the input and result vector are distributed differently (row wise, col wise respectively)
    CALL timeset(routineN//"_local_mm", handle1)

!------ perform the multiplication, we have to take car to take the correct blocks ----------

!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,transposed,ithread,pcol,prow,rpcol,rprow) &
!$OMP          SHARED(matrix,fast_vec_row,res_fast_vec_col,res_fast_vec_row,fast_vec_col)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed)
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       rprow=hash_table_get(res_fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_row%blk_map_d(pcol)%ptr) .AND.&
          ASSOCIATED(res_fast_vec_col%blk_map_d(rprow)%ptr))THEN
          IF(res_fast_vec_col%blk_map_d(rprow)%assigned_thread .EQ. ithread ) THEN
             res_fast_vec_col%blk_map_d(rprow)%ptr=res_fast_vec_col%blk_map_d(rprow)%ptr+&
               MATMUL(data_d,TRANSPOSE(fast_vec_row%blk_map_d(pcol)%ptr))
          END IF
          prow=hash_table_get(fast_vec_col%hash_table,row)
          rpcol=hash_table_get(res_fast_vec_row%hash_table,col)
          IF(res_fast_vec_row%blk_map_d(rpcol)%assigned_thread .EQ. ithread .AND. row .NE. col) THEN
             res_fast_vec_row%blk_map_d(rpcol)%ptr=res_fast_vec_row%blk_map_d(rpcol)%ptr+&
                MATMUL(TRANSPOSE(fast_vec_col%blk_map_d(prow)%ptr),data_d)
          END IF
       ELSE
          rpcol=hash_table_get(res_fast_vec_col%hash_table,col)
          prow=hash_table_get(fast_vec_row%hash_table,row)
          IF(res_fast_vec_col%blk_map_d(rpcol)%assigned_thread .EQ. ithread ) THEN
             res_fast_vec_col%blk_map_d(rpcol)%ptr=res_fast_vec_col%blk_map_d(rpcol)%ptr+&
                TRANSPOSE(MATMUL(fast_vec_row%blk_map_d(prow)%ptr,data_d))
          END IF
          rprow=hash_table_get(res_fast_vec_row%hash_table,row)
          pcol=hash_table_get(fast_vec_col%hash_table,col)
          IF(res_fast_vec_row%blk_map_d(rprow)%assigned_thread .EQ. ithread  .AND. row .NE. col ) THEN
             res_fast_vec_row%blk_map_d(rprow)%ptr=res_fast_vec_row%blk_map_d(rprow)%ptr+&
                TRANSPOSE(MATMUL(data_d,fast_vec_col%blk_map_d(pcol)%ptr))
          END IF
       END IF 
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

    ! sum all the data within a processor column to obtain the replicated result from lower
    data_vec => dbcsr_get_data_p (result_row%m%data_area, select_data_type=0.0_real_8)
    CALL dbcsr_get_info(matrix=result_row, nfullrows_local=nrows, nfullcols_local=ncols)

    CALL mp_sum(data_vec(1:nrows*ncols), pcol_group)
!
!! Convert the results to a column wise distribution, this is a bit involved as the result_row is fully replicated
!! While the result_col still has the partial results in parallel. The routine below takes care of that and saves an
!! mp_sum. Of the res_row vectors are created only taking the approriate element (0 otherwise) while the res_col
!! parallel bits are locally added. The mp_sum magically creates the correct vector
    CALL dbcsr_rep_row_to_rep_col_vec_d(work_col, result_row, res_fast_vec_row, res_fast_vec_col)

!    ! Create_the final vector by summing it to the result vector which lives on proc_col 0 lower
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_d(prow)%ptr))THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*(fast_vec_col%blk_map_d(prow)%ptr(:, :)) 
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL release_fast_vec_access(fast_vec_row)
    CALL release_fast_vec_access(fast_vec_col)
    CALL release_fast_vec_access(res_fast_vec_row)
    CALL release_fast_vec_access(res_fast_vec_col)

    CALL dbcsr_release(result_row); CALL dbcsr_release(result_col)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_sym_matrix_vector_mult_d

! **************************************************************************************************
!> \brief the real driver routine for the multiply, not all symmetries implemented yet
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrix_colvec_multiply_low_z(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    COMPLEX(kind=real_8)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_colvec_multiply_low', &
      routineP = moduleN//':'//routineN

    CHARACTER                                :: matrix_type

    matrix_type=dbcsr_get_matrix_type(matrix)
    SELECT CASE(matrix_type)
    CASE(dbcsr_type_no_symmetry)
       CALL dbcsr_matrix_vector_mult_z(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    CASE(dbcsr_type_symmetric)
       CALL dbcsr_sym_matrix_vector_mult_z(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    CASE(dbcsr_type_antisymmetric)
        ! Not yet implemented, should mainly be some prefactor magic, but who knows how antisymmetric matrices are stored???
       CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",29,"NYI, antisymmetric matrix not permitted")
    CASE DEFAULT
       CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",31,"Unknown matrix type, ...")
    END SELECT

  END SUBROUTINE dbcsr_matrix_colvec_multiply_low_z

! **************************************************************************************************
!> \brief low level routines for matrix vector multiplies
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrix_vector_mult_z(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    COMPLEX(kind=real_8)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_vector_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, &
                                                myprow, &
                                                ncols, pcol_group, nrows, &
                                                prow_group, row, &
                                                handle, handle1, ithread 
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER       :: data_d, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter
    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col
    INTEGER                                  :: prow, pcol

    CALL timeset(routineN, handle)
    ithread=0

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group 
    pcol_group=distri%d%mp_env%mp%pcol_group 
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)

! Transfer the correct parts of the input vector to the correct locations so we can do a local multiply
    CALL dbcsr_col_vec_to_rep_row_z(vec_in, work_col, work_row, fast_vec_col)

! Set the work vector for the results to 0
    CALL dbcsr_set(work_col, CMPLX(0.0, 0.0, real_8))

! Perform the local multiply. Here we exploit, that we have the blocks replicated on the mpi processes
! It is important to note, that the input and result vector are distributed differently (row wise, col wise respectively)
    CALL timeset(routineN//"_local_mm", handle1)

!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,transposed,ithread,pcol,prow) &
!$OMP          SHARED(matrix,fast_vec_col,fast_vec_row)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(fast_vec_col%blk_map_z(prow)%assigned_thread .NE. ithread ) CYCLE
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       fast_vec_col%blk_map_z(prow)%ptr=fast_vec_col%blk_map_z(prow)%ptr+&
            MATMUL(data_d,TRANSPOSE(fast_vec_row%blk_map_z(pcol)%ptr))
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

! sum all the data onto the first processor col where the original vector is stored
    data_vec => dbcsr_get_data_p (work_col%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
    CALL dbcsr_get_info(matrix=work_col, nfullrows_local=nrows, nfullcols_local=ncols)
    CALL mp_sum(data_vec(1:nrows*ncols), prow_group)

! Local copy on the first mpi col (as this is the localtion of the vec_res blocks) of the result vector
! from the replicated to the original vector. Let's play it safe and use the iterator
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_z(prow)%ptr)) THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*fast_vec_col%blk_map_z(prow)%ptr(:,:)
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL release_fast_vec_access(fast_vec_row)
    CALL release_fast_vec_access(fast_vec_col)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_matrix_vector_mult_z

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
!> \param skip_diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrixT_vector_mult_z(matrix, vec_in, vec_out, alpha, beta, work_row, work_col, skip_diag)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    COMPLEX(kind=real_8)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col
    LOGICAL                                  :: skip_diag

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrixT_vector_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, col_size, mypcol, &
                                                myprow, &
                                                ncols, pcol_group, nrows, &
                                                prow_group, row, row_size, &
                                                handle, handle1, ithread
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER       :: data_d, vec_bl, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col
    INTEGER                                  :: prow, pcol

    CALL timeset(routineN, handle)
    ithread=0

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group; pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol; myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)

! Set the work vector for the results to 0
    CALL dbcsr_set(work_row, CMPLX(0.0, 0.0, real_8))

! Transfer the correct parts of the input vector to the replicated vector on proc_col 0
    CALL dbcsr_iterator_start(iter, vec_in)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed, row_size=row_size, col_size=col_size)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       fast_vec_col%blk_map_z(prow)%ptr(1:row_size, 1:col_size)= vec_bl(1:row_size, 1:col_size)
    END DO
    CALL dbcsr_iterator_stop(iter)
! Replicate the data on all processore in the row
    data_vec => dbcsr_get_data_p (work_col%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
    CALL mp_bcast(data_vec, 0, prow_group)

! Perform the local multiply. Here it is obvious why the vectors are replicated on the mpi rows and cols
    CALL timeset(routineN//"local_mm", handle1)
    CALL dbcsr_get_info(matrix=work_col, nfullcols_local=ncols)
!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,row_size,col_size,transposed,ithread,prow,pcol) &
!$OMP          SHARED(matrix,fast_vec_row,fast_vec_col,skip_diag,ncols)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed, row_size=row_size, col_size=col_size)
       IF(skip_diag.AND.col==row)CYCLE
       prow=hash_table_get(fast_vec_col%hash_table,row)
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       IF ( ASSOCIATED(fast_vec_row%blk_map_z(pcol)%ptr) .AND. &
            ASSOCIATED(fast_vec_col%blk_map_z(prow)%ptr) )THEN
          IF(fast_vec_row%blk_map_z(pcol)%assigned_thread .NE. ithread ) CYCLE
          fast_vec_row%blk_map_z(pcol)%ptr=fast_vec_row%blk_map_z(pcol)%ptr+&
               MATMUL(TRANSPOSE(fast_vec_col%blk_map_z(prow)%ptr),data_d)
       ELSE
          prow=hash_table_get(fast_vec_row%hash_table,row)
          pcol=hash_table_get(fast_vec_col%hash_table,col)
          IF(fast_vec_row%blk_map_z(prow)%assigned_thread .NE. ithread ) CYCLE
          fast_vec_row%blk_map_z(prow)%ptr=fast_vec_row%blk_map_z(prow)%ptr+&
             MATMUL(TRANSPOSE(fast_vec_col%blk_map_z(pcol)%ptr),TRANSPOSE(data_d))
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

! sum all the data within a processor column to obtain the replicated result
    data_vec => dbcsr_get_data_p (work_row%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
! we use the replicated vector but the final answer is only summed to proc_col 0 for efficiency
    CALL dbcsr_get_info(matrix=work_row, nfullrows_local=nrows, nfullcols_local=ncols)
    CALL mp_sum(data_vec(1:nrows*ncols), pcol_group)

! Convert the result to a column wise distribution
    CALL dbcsr_rep_row_to_rep_col_vec_z(work_col, work_row, fast_vec_row)
    
! Create_the final vector by summing it to the result vector which lives on proc_col 0
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed, row_size=row_size)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_z(prow)%ptr)) THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*fast_vec_col%blk_map_z(prow)%ptr(:,:)
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_matrixT_vector_mult_z 

! **************************************************************************************************
!> \brief ...
!> \param vec_in ...
!> \param rep_col_vec ...
!> \param rep_row_vec ...
!> \param fast_vec_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_col_vec_to_rep_row_z(vec_in, rep_col_vec, rep_row_vec, fast_vec_col)
    TYPE(dbcsr_obj)                          :: vec_in, rep_col_vec, &
                                                rep_row_vec
    TYPE(fast_vec_access_type), INTENT(IN)   :: fast_vec_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_col_vec_to_rep_row', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, myprow, ncols, &
                                                nrows, pcol_group, &
                                                prow_group, row, handle
    INTEGER, DIMENSION(:), POINTER           :: local_cols, row_dist
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_8), DIMENSION(:), POINTER          :: data_vec, data_vec_rep
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER       :: vec_row
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

! get information about the parallel environment
    CALL dbcsr_get_info(matrix=vec_in, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group
    pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow

! Get the vector which tells us which blocks are local to which processor row in the col vec
    row_dist=> dbcsr_distribution_row_dist (dbcsr_distribution(rep_col_vec))

! Copy the local vector to the replicated on the first processor column (this is where vec_in lives)
    CALL dbcsr_get_info(matrix=rep_col_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_col_vec%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
    data_vec => dbcsr_get_data_p (vec_in%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
    IF(mypcol==0)data_vec_rep(1:nrows*ncols)=data_vec(1:nrows*ncols)
! Replicate the data along the row
    CALL mp_bcast(data_vec_rep(1:nrows*ncols), 0, prow_group)

! Here it gets a bit tricky as we are dealing with two different parallel layouts:
! The rep_col_vec contains all blocks local to the row distribution of the vector. 
! The rep_row_vec only needs the fraction which is local to the col distribution.
! However in most cases this won't the complete set of block which can be obtained from col_vector p_row i
! Anyway, as the blocks don't repeat in the col_vec, a different fraction of the row vec will be available
! on every replica in the processor column, by summing along the column we end up with the complete vector everywhere
! Hope this clarifies the idea
    CALL dbcsr_set(rep_row_vec, CMPLX(0.0, 0.0, real_8))
    CALL dbcsr_get_info(matrix=rep_row_vec, nfullrows_local=nrows, local_cols=local_cols, nfullcols_local=ncols)
    CALL dbcsr_iterator_start(iter, rep_row_vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_row, transposed)
       IF(row_dist(col)==myprow)THEN
          vec_row=TRANSPOSE(fast_vec_col%blk_map_z(hash_table_get(fast_vec_col%hash_table,col))%ptr)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)
    CALL dbcsr_get_info(matrix=rep_row_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_row_vec%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
    CALL mp_sum(data_vec_rep(1:ncols*nrows), pcol_group)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_col_vec_to_rep_row_z    
       
! **************************************************************************************************
!> \brief ...
!> \param rep_col_vec ...
!> \param rep_row_vec ...
!> \param fast_vec_row ...
!> \param fast_vec_col_add ...
! **************************************************************************************************
  SUBROUTINE dbcsr_rep_row_to_rep_col_vec_z(rep_col_vec, rep_row_vec, fast_vec_row, fast_vec_col_add)
    TYPE(dbcsr_obj)                          :: rep_col_vec, rep_row_vec
    TYPE(fast_vec_access_type), OPTIONAL     :: fast_vec_col_add
    TYPE(fast_vec_access_type)               :: fast_vec_row

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_rep_row_to_rep_col_vec', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, myprow, ncols, &
                                                nrows, pcol_group, &
                                                prow_group, row, handle
    INTEGER, DIMENSION(:), POINTER           :: col_dist
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_8), DIMENSION(:), POINTER          :: data_vec_rep
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER       :: vec_col
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

! get information about the parallel environment
    CALL dbcsr_get_info(matrix=rep_col_vec, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group
    pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow
! Get the vector which tells us which blocks are local to which processor col in the row vec
    col_dist=> dbcsr_distribution_col_dist (dbcsr_distribution(rep_row_vec))

! The same trick as described above with opposite direction
    CALL dbcsr_set(rep_col_vec, CMPLX(0.0, 0.0, real_8))
    CALL dbcsr_iterator_start(iter, rep_col_vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_col, transposed)
       IF(col_dist(row)==mypcol)THEN
          vec_col=TRANSPOSE(fast_vec_row%blk_map_z(hash_table_get(fast_vec_row%hash_table,row))%ptr)
       END IF
       ! this one is special and allows to add the elements of a not yet summed replicated 
       ! column vector as it appears in M*V(row_rep) as result. Save an mp_sum in the symmetric case
       IF(PRESENT(fast_vec_col_add))vec_col=vec_col+&
            fast_vec_col_add%blk_map_z(hash_table_get(fast_vec_col_add%hash_table,row))%ptr(:,:)
    END DO
    CALL dbcsr_iterator_stop(iter)
    CALL dbcsr_get_info(matrix=rep_col_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_col_vec%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
    CALL mp_sum(data_vec_rep(1:nrows*ncols), prow_group)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_rep_row_to_rep_col_vec_z


! **************************************************************************************************
!> \brief given a column vector, prepare the fast_vec_access container
!> \param vec ...
!> \param fast_vec_access ...
! **************************************************************************************************
  SUBROUTINE create_fast_col_vec_access_z(vec, fast_vec_access)
    TYPE(dbcsr_obj)                          :: vec
    TYPE(fast_vec_access_type)               :: fast_vec_access

    CHARACTER(LEN=*), PARAMETER :: routineN = 'create_fast_col_vec_access_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, nblk_local
    INTEGER                                  :: col, row, iblock, nthreads
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER       :: vec_bl
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

    ! figure out the number of threads
    nthreads = 1
!$OMP PARALLEL DEFAULT(NONE) SHARED(nthreads)
!$OMP MASTER
    !$ nthreads = OMP_GET_NUM_THREADS()
!$OMP END MASTER
!$OMP END PARALLEL

    CALL dbcsr_get_info(matrix=vec, nblkrows_local=nblk_local)
    ! 4 times makes sure the table is big enough to limit collisions.
    CALL hash_table_create(fast_vec_access%hash_table,4*nblk_local)
    ! include zero for effective dealing with values not in the hash table (will return 0)
    ALLOCATE(fast_vec_access%blk_map_z(0:nblk_local))

    CALL dbcsr_get_info(matrix=vec, nblkcols_local=col)
    IF (col.GT.1) CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",413,"BUG")

    ! go through the blocks of the vector
    iblock=0
    CALL dbcsr_iterator_start(iter, vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed)
       iblock=iblock+1
       CALL hash_table_add(fast_vec_access%hash_table,row,iblock)
       fast_vec_access%blk_map_z(iblock)%ptr=>vec_bl
       fast_vec_access%blk_map_z(iblock)%assigned_thread=MOD(iblock,nthreads)
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE create_fast_col_vec_access_z

! **************************************************************************************************
!> \brief given a row vector, prepare the fast_vec_access_container
!> \param vec ...
!> \param fast_vec_access ...
! **************************************************************************************************
  SUBROUTINE create_fast_row_vec_access_z(vec, fast_vec_access)
    TYPE(dbcsr_obj)                          :: vec
    TYPE(fast_vec_access_type)               :: fast_vec_access

    CHARACTER(LEN=*), PARAMETER :: routineN = 'create_fast_row_vec_access_z', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, nblk_local
    INTEGER                                  :: col, row, iblock, nthreads
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER       :: vec_bl
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

    ! figure out the number of threads
    nthreads = 1
!$OMP PARALLEL DEFAULT(NONE) SHARED(nthreads)
!$OMP MASTER
    !$ nthreads = OMP_GET_NUM_THREADS()
!$OMP END MASTER
!$OMP END PARALLEL

    CALL dbcsr_get_info(matrix=vec, nblkcols_local=nblk_local)
    ! 4 times makes sure the table is big enough to limit collisions.
    CALL hash_table_create(fast_vec_access%hash_table,4*nblk_local)
    ! include zero for effective dealing with values not in the hash table (will return 0)
    ALLOCATE(fast_vec_access%blk_map_z(0:nblk_local))

    ! sanity check
    CALL dbcsr_get_info(matrix=vec, nblkrows_local=row)
    IF (row.GT.1) CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",467,"BUG")

    ! go through the blocks of the vector
    iblock=0
    CALL dbcsr_iterator_start(iter, vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed)
       iblock=iblock+1
       CALL hash_table_add(fast_vec_access%hash_table,col,iblock)
       fast_vec_access%blk_map_z(iblock)%ptr=>vec_bl
       fast_vec_access%blk_map_z(iblock)%assigned_thread=MOD(iblock,nthreads)
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE create_fast_row_vec_access_z

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_sym_matrix_vector_mult_z(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    COMPLEX(kind=real_8)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_sym_m_v_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, &
                                                myprow, &
                                                pcol_group, nrows, ncols,&
                                                prow_group, row, &
                                                handle, handle1, ithread, vec_dim
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_8), DIMENSION(:), POINTER          :: data_vec
    COMPLEX(kind=real_8), DIMENSION(:, :), POINTER       :: data_d, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter
    TYPE(dbcsr_obj)                          :: result_row, result_col

    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col, res_fast_vec_row, res_fast_vec_col
    INTEGER                                  :: prow, pcol, rprow, rpcol

    CALL timeset(routineN, handle)
    ithread=0
! We need some work matrices as we try to exploit operations on the replicated vectors which are duplicated otherwise
    CALL dbcsr_get_info(matrix=vec_in,nfullcols_total=vec_dim)
! This is a performance hack as the new creation of a replicated vector is a fair bit more expensive
    CALL dbcsr_init(result_col)
    CALL dbcsr_set(work_col, CMPLX(0.0, 0.0, real_8))
    CALL dbcsr_copy(result_col, work_col)
    CALL dbcsr_init(result_row)
    CALL dbcsr_set(work_row, CMPLX(0.0, 0.0, real_8))
    CALL dbcsr_copy(result_row, work_row)

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group; pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol; myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)
    CALL create_fast_row_vec_access(result_row, res_fast_vec_row)
    CALL create_fast_col_vec_access(result_col, res_fast_vec_col)

! Transfer the correct parts of the input vector to the correct locations so we can do a local multiply
    CALL dbcsr_col_vec_to_rep_row_z(vec_in, work_col, work_row, fast_vec_col)

! Probably I should rename the routine above as it delivers both the replicated row and column vector

! Perform the local multiply. Here we exploit, that we have the blocks replicated on the mpi processes
! It is important to note, that the input and result vector are distributed differently (row wise, col wise respectively)
    CALL timeset(routineN//"_local_mm", handle1)

!------ perform the multiplication, we have to take car to take the correct blocks ----------

!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,transposed,ithread,pcol,prow,rpcol,rprow) &
!$OMP          SHARED(matrix,fast_vec_row,res_fast_vec_col,res_fast_vec_row,fast_vec_col)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed)
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       rprow=hash_table_get(res_fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_row%blk_map_z(pcol)%ptr) .AND.&
          ASSOCIATED(res_fast_vec_col%blk_map_z(rprow)%ptr))THEN
          IF(res_fast_vec_col%blk_map_z(rprow)%assigned_thread .EQ. ithread ) THEN
             res_fast_vec_col%blk_map_z(rprow)%ptr=res_fast_vec_col%blk_map_z(rprow)%ptr+&
               MATMUL(data_d,TRANSPOSE(fast_vec_row%blk_map_z(pcol)%ptr))
          END IF
          prow=hash_table_get(fast_vec_col%hash_table,row)
          rpcol=hash_table_get(res_fast_vec_row%hash_table,col)
          IF(res_fast_vec_row%blk_map_z(rpcol)%assigned_thread .EQ. ithread .AND. row .NE. col) THEN
             res_fast_vec_row%blk_map_z(rpcol)%ptr=res_fast_vec_row%blk_map_z(rpcol)%ptr+&
                MATMUL(TRANSPOSE(fast_vec_col%blk_map_z(prow)%ptr),data_d)
          END IF
       ELSE
          rpcol=hash_table_get(res_fast_vec_col%hash_table,col)
          prow=hash_table_get(fast_vec_row%hash_table,row)
          IF(res_fast_vec_col%blk_map_z(rpcol)%assigned_thread .EQ. ithread ) THEN
             res_fast_vec_col%blk_map_z(rpcol)%ptr=res_fast_vec_col%blk_map_z(rpcol)%ptr+&
                TRANSPOSE(MATMUL(fast_vec_row%blk_map_z(prow)%ptr,data_d))
          END IF
          rprow=hash_table_get(res_fast_vec_row%hash_table,row)
          pcol=hash_table_get(fast_vec_col%hash_table,col)
          IF(res_fast_vec_row%blk_map_z(rprow)%assigned_thread .EQ. ithread  .AND. row .NE. col ) THEN
             res_fast_vec_row%blk_map_z(rprow)%ptr=res_fast_vec_row%blk_map_z(rprow)%ptr+&
                TRANSPOSE(MATMUL(data_d,fast_vec_col%blk_map_z(pcol)%ptr))
          END IF
       END IF 
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

    ! sum all the data within a processor column to obtain the replicated result from lower
    data_vec => dbcsr_get_data_p (result_row%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_8))
    CALL dbcsr_get_info(matrix=result_row, nfullrows_local=nrows, nfullcols_local=ncols)

    CALL mp_sum(data_vec(1:nrows*ncols), pcol_group)
!
!! Convert the results to a column wise distribution, this is a bit involved as the result_row is fully replicated
!! While the result_col still has the partial results in parallel. The routine below takes care of that and saves an
!! mp_sum. Of the res_row vectors are created only taking the approriate element (0 otherwise) while the res_col
!! parallel bits are locally added. The mp_sum magically creates the correct vector
    CALL dbcsr_rep_row_to_rep_col_vec_z(work_col, result_row, res_fast_vec_row, res_fast_vec_col)

!    ! Create_the final vector by summing it to the result vector which lives on proc_col 0 lower
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_z(prow)%ptr))THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*(fast_vec_col%blk_map_z(prow)%ptr(:, :)) 
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL release_fast_vec_access(fast_vec_row)
    CALL release_fast_vec_access(fast_vec_col)
    CALL release_fast_vec_access(res_fast_vec_row)
    CALL release_fast_vec_access(res_fast_vec_col)

    CALL dbcsr_release(result_row); CALL dbcsr_release(result_col)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_sym_matrix_vector_mult_z

! **************************************************************************************************
!> \brief the real driver routine for the multiply, not all symmetries implemented yet
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrix_colvec_multiply_low_s(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    REAL(kind=real_4)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_colvec_multiply_low', &
      routineP = moduleN//':'//routineN

    CHARACTER                                :: matrix_type

    matrix_type=dbcsr_get_matrix_type(matrix)
    SELECT CASE(matrix_type)
    CASE(dbcsr_type_no_symmetry)
       CALL dbcsr_matrix_vector_mult_s(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    CASE(dbcsr_type_symmetric)
       CALL dbcsr_sym_matrix_vector_mult_s(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    CASE(dbcsr_type_antisymmetric)
        ! Not yet implemented, should mainly be some prefactor magic, but who knows how antisymmetric matrices are stored???
       CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",29,"NYI, antisymmetric matrix not permitted")
    CASE DEFAULT
       CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",31,"Unknown matrix type, ...")
    END SELECT

  END SUBROUTINE dbcsr_matrix_colvec_multiply_low_s

! **************************************************************************************************
!> \brief low level routines for matrix vector multiplies
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrix_vector_mult_s(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    REAL(kind=real_4)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_vector_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, &
                                                myprow, &
                                                ncols, pcol_group, nrows, &
                                                prow_group, row, &
                                                handle, handle1, ithread 
    LOGICAL                                  :: transposed
    REAL(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    REAL(kind=real_4), DIMENSION(:, :), POINTER       :: data_d, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter
    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col
    INTEGER                                  :: prow, pcol

    CALL timeset(routineN, handle)
    ithread=0

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group 
    pcol_group=distri%d%mp_env%mp%pcol_group 
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)

! Transfer the correct parts of the input vector to the correct locations so we can do a local multiply
    CALL dbcsr_col_vec_to_rep_row_s(vec_in, work_col, work_row, fast_vec_col)

! Set the work vector for the results to 0
    CALL dbcsr_set(work_col, 0.0_real_4)

! Perform the local multiply. Here we exploit, that we have the blocks replicated on the mpi processes
! It is important to note, that the input and result vector are distributed differently (row wise, col wise respectively)
    CALL timeset(routineN//"_local_mm", handle1)

!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,transposed,ithread,pcol,prow) &
!$OMP          SHARED(matrix,fast_vec_col,fast_vec_row)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(fast_vec_col%blk_map_s(prow)%assigned_thread .NE. ithread ) CYCLE
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       fast_vec_col%blk_map_s(prow)%ptr=fast_vec_col%blk_map_s(prow)%ptr+&
            MATMUL(data_d,TRANSPOSE(fast_vec_row%blk_map_s(pcol)%ptr))
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

! sum all the data onto the first processor col where the original vector is stored
    data_vec => dbcsr_get_data_p (work_col%m%data_area, select_data_type=0.0_real_4)
    CALL dbcsr_get_info(matrix=work_col, nfullrows_local=nrows, nfullcols_local=ncols)
    CALL mp_sum(data_vec(1:nrows*ncols), prow_group)

! Local copy on the first mpi col (as this is the localtion of the vec_res blocks) of the result vector
! from the replicated to the original vector. Let's play it safe and use the iterator
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_s(prow)%ptr)) THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*fast_vec_col%blk_map_s(prow)%ptr(:,:)
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL release_fast_vec_access(fast_vec_row)
    CALL release_fast_vec_access(fast_vec_col)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_matrix_vector_mult_s

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
!> \param skip_diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrixT_vector_mult_s(matrix, vec_in, vec_out, alpha, beta, work_row, work_col, skip_diag)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    REAL(kind=real_4)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col
    LOGICAL                                  :: skip_diag

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrixT_vector_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, col_size, mypcol, &
                                                myprow, &
                                                ncols, pcol_group, nrows, &
                                                prow_group, row, row_size, &
                                                handle, handle1, ithread
    LOGICAL                                  :: transposed
    REAL(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    REAL(kind=real_4), DIMENSION(:, :), POINTER       :: data_d, vec_bl, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col
    INTEGER                                  :: prow, pcol

    CALL timeset(routineN, handle)
    ithread=0

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group; pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol; myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)

! Set the work vector for the results to 0
    CALL dbcsr_set(work_row, 0.0_real_4)

! Transfer the correct parts of the input vector to the replicated vector on proc_col 0
    CALL dbcsr_iterator_start(iter, vec_in)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed, row_size=row_size, col_size=col_size)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       fast_vec_col%blk_map_s(prow)%ptr(1:row_size, 1:col_size)= vec_bl(1:row_size, 1:col_size)
    END DO
    CALL dbcsr_iterator_stop(iter)
! Replicate the data on all processore in the row
    data_vec => dbcsr_get_data_p (work_col%m%data_area, select_data_type=0.0_real_4)
    CALL mp_bcast(data_vec, 0, prow_group)

! Perform the local multiply. Here it is obvious why the vectors are replicated on the mpi rows and cols
    CALL timeset(routineN//"local_mm", handle1)
    CALL dbcsr_get_info(matrix=work_col, nfullcols_local=ncols)
!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,row_size,col_size,transposed,ithread,prow,pcol) &
!$OMP          SHARED(matrix,fast_vec_row,fast_vec_col,skip_diag,ncols)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed, row_size=row_size, col_size=col_size)
       IF(skip_diag.AND.col==row)CYCLE
       prow=hash_table_get(fast_vec_col%hash_table,row)
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       IF ( ASSOCIATED(fast_vec_row%blk_map_s(pcol)%ptr) .AND. &
            ASSOCIATED(fast_vec_col%blk_map_s(prow)%ptr) )THEN
          IF(fast_vec_row%blk_map_s(pcol)%assigned_thread .NE. ithread ) CYCLE
          fast_vec_row%blk_map_s(pcol)%ptr=fast_vec_row%blk_map_s(pcol)%ptr+&
               MATMUL(TRANSPOSE(fast_vec_col%blk_map_s(prow)%ptr),data_d)
       ELSE
          prow=hash_table_get(fast_vec_row%hash_table,row)
          pcol=hash_table_get(fast_vec_col%hash_table,col)
          IF(fast_vec_row%blk_map_s(prow)%assigned_thread .NE. ithread ) CYCLE
          fast_vec_row%blk_map_s(prow)%ptr=fast_vec_row%blk_map_s(prow)%ptr+&
             MATMUL(TRANSPOSE(fast_vec_col%blk_map_s(pcol)%ptr),TRANSPOSE(data_d))
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

! sum all the data within a processor column to obtain the replicated result
    data_vec => dbcsr_get_data_p (work_row%m%data_area, select_data_type=0.0_real_4)
! we use the replicated vector but the final answer is only summed to proc_col 0 for efficiency
    CALL dbcsr_get_info(matrix=work_row, nfullrows_local=nrows, nfullcols_local=ncols)
    CALL mp_sum(data_vec(1:nrows*ncols), pcol_group)

! Convert the result to a column wise distribution
    CALL dbcsr_rep_row_to_rep_col_vec_s(work_col, work_row, fast_vec_row)
    
! Create_the final vector by summing it to the result vector which lives on proc_col 0
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed, row_size=row_size)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_s(prow)%ptr)) THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*fast_vec_col%blk_map_s(prow)%ptr(:,:)
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_matrixT_vector_mult_s 

! **************************************************************************************************
!> \brief ...
!> \param vec_in ...
!> \param rep_col_vec ...
!> \param rep_row_vec ...
!> \param fast_vec_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_col_vec_to_rep_row_s(vec_in, rep_col_vec, rep_row_vec, fast_vec_col)
    TYPE(dbcsr_obj)                          :: vec_in, rep_col_vec, &
                                                rep_row_vec
    TYPE(fast_vec_access_type), INTENT(IN)   :: fast_vec_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_col_vec_to_rep_row', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, myprow, ncols, &
                                                nrows, pcol_group, &
                                                prow_group, row, handle
    INTEGER, DIMENSION(:), POINTER           :: local_cols, row_dist
    LOGICAL                                  :: transposed
    REAL(kind=real_4), DIMENSION(:), POINTER          :: data_vec, data_vec_rep
    REAL(kind=real_4), DIMENSION(:, :), POINTER       :: vec_row
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

! get information about the parallel environment
    CALL dbcsr_get_info(matrix=vec_in, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group
    pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow

! Get the vector which tells us which blocks are local to which processor row in the col vec
    row_dist=> dbcsr_distribution_row_dist (dbcsr_distribution(rep_col_vec))

! Copy the local vector to the replicated on the first processor column (this is where vec_in lives)
    CALL dbcsr_get_info(matrix=rep_col_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_col_vec%m%data_area, select_data_type=0.0_real_4)
    data_vec => dbcsr_get_data_p (vec_in%m%data_area, select_data_type=0.0_real_4)
    IF(mypcol==0)data_vec_rep(1:nrows*ncols)=data_vec(1:nrows*ncols)
! Replicate the data along the row
    CALL mp_bcast(data_vec_rep(1:nrows*ncols), 0, prow_group)

! Here it gets a bit tricky as we are dealing with two different parallel layouts:
! The rep_col_vec contains all blocks local to the row distribution of the vector. 
! The rep_row_vec only needs the fraction which is local to the col distribution.
! However in most cases this won't the complete set of block which can be obtained from col_vector p_row i
! Anyway, as the blocks don't repeat in the col_vec, a different fraction of the row vec will be available
! on every replica in the processor column, by summing along the column we end up with the complete vector everywhere
! Hope this clarifies the idea
    CALL dbcsr_set(rep_row_vec, 0.0_real_4)
    CALL dbcsr_get_info(matrix=rep_row_vec, nfullrows_local=nrows, local_cols=local_cols, nfullcols_local=ncols)
    CALL dbcsr_iterator_start(iter, rep_row_vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_row, transposed)
       IF(row_dist(col)==myprow)THEN
          vec_row=TRANSPOSE(fast_vec_col%blk_map_s(hash_table_get(fast_vec_col%hash_table,col))%ptr)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)
    CALL dbcsr_get_info(matrix=rep_row_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_row_vec%m%data_area, select_data_type=0.0_real_4)
    CALL mp_sum(data_vec_rep(1:ncols*nrows), pcol_group)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_col_vec_to_rep_row_s    
       
! **************************************************************************************************
!> \brief ...
!> \param rep_col_vec ...
!> \param rep_row_vec ...
!> \param fast_vec_row ...
!> \param fast_vec_col_add ...
! **************************************************************************************************
  SUBROUTINE dbcsr_rep_row_to_rep_col_vec_s(rep_col_vec, rep_row_vec, fast_vec_row, fast_vec_col_add)
    TYPE(dbcsr_obj)                          :: rep_col_vec, rep_row_vec
    TYPE(fast_vec_access_type), OPTIONAL     :: fast_vec_col_add
    TYPE(fast_vec_access_type)               :: fast_vec_row

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_rep_row_to_rep_col_vec', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, myprow, ncols, &
                                                nrows, pcol_group, &
                                                prow_group, row, handle
    INTEGER, DIMENSION(:), POINTER           :: col_dist
    LOGICAL                                  :: transposed
    REAL(kind=real_4), DIMENSION(:), POINTER          :: data_vec_rep
    REAL(kind=real_4), DIMENSION(:, :), POINTER       :: vec_col
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

! get information about the parallel environment
    CALL dbcsr_get_info(matrix=rep_col_vec, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group
    pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow
! Get the vector which tells us which blocks are local to which processor col in the row vec
    col_dist=> dbcsr_distribution_col_dist (dbcsr_distribution(rep_row_vec))

! The same trick as described above with opposite direction
    CALL dbcsr_set(rep_col_vec, 0.0_real_4)
    CALL dbcsr_iterator_start(iter, rep_col_vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_col, transposed)
       IF(col_dist(row)==mypcol)THEN
          vec_col=TRANSPOSE(fast_vec_row%blk_map_s(hash_table_get(fast_vec_row%hash_table,row))%ptr)
       END IF
       ! this one is special and allows to add the elements of a not yet summed replicated 
       ! column vector as it appears in M*V(row_rep) as result. Save an mp_sum in the symmetric case
       IF(PRESENT(fast_vec_col_add))vec_col=vec_col+&
            fast_vec_col_add%blk_map_s(hash_table_get(fast_vec_col_add%hash_table,row))%ptr(:,:)
    END DO
    CALL dbcsr_iterator_stop(iter)
    CALL dbcsr_get_info(matrix=rep_col_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_col_vec%m%data_area, select_data_type=0.0_real_4)
    CALL mp_sum(data_vec_rep(1:nrows*ncols), prow_group)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_rep_row_to_rep_col_vec_s


! **************************************************************************************************
!> \brief given a column vector, prepare the fast_vec_access container
!> \param vec ...
!> \param fast_vec_access ...
! **************************************************************************************************
  SUBROUTINE create_fast_col_vec_access_s(vec, fast_vec_access)
    TYPE(dbcsr_obj)                          :: vec
    TYPE(fast_vec_access_type)               :: fast_vec_access

    CHARACTER(LEN=*), PARAMETER :: routineN = 'create_fast_col_vec_access_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, nblk_local
    INTEGER                                  :: col, row, iblock, nthreads
    LOGICAL                                  :: transposed
    REAL(kind=real_4), DIMENSION(:, :), POINTER       :: vec_bl
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

    ! figure out the number of threads
    nthreads = 1
!$OMP PARALLEL DEFAULT(NONE) SHARED(nthreads)
!$OMP MASTER
    !$ nthreads = OMP_GET_NUM_THREADS()
!$OMP END MASTER
!$OMP END PARALLEL

    CALL dbcsr_get_info(matrix=vec, nblkrows_local=nblk_local)
    ! 4 times makes sure the table is big enough to limit collisions.
    CALL hash_table_create(fast_vec_access%hash_table,4*nblk_local)
    ! include zero for effective dealing with values not in the hash table (will return 0)
    ALLOCATE(fast_vec_access%blk_map_s(0:nblk_local))

    CALL dbcsr_get_info(matrix=vec, nblkcols_local=col)
    IF (col.GT.1) CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",413,"BUG")

    ! go through the blocks of the vector
    iblock=0
    CALL dbcsr_iterator_start(iter, vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed)
       iblock=iblock+1
       CALL hash_table_add(fast_vec_access%hash_table,row,iblock)
       fast_vec_access%blk_map_s(iblock)%ptr=>vec_bl
       fast_vec_access%blk_map_s(iblock)%assigned_thread=MOD(iblock,nthreads)
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE create_fast_col_vec_access_s

! **************************************************************************************************
!> \brief given a row vector, prepare the fast_vec_access_container
!> \param vec ...
!> \param fast_vec_access ...
! **************************************************************************************************
  SUBROUTINE create_fast_row_vec_access_s(vec, fast_vec_access)
    TYPE(dbcsr_obj)                          :: vec
    TYPE(fast_vec_access_type)               :: fast_vec_access

    CHARACTER(LEN=*), PARAMETER :: routineN = 'create_fast_row_vec_access_s', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, nblk_local
    INTEGER                                  :: col, row, iblock, nthreads
    LOGICAL                                  :: transposed
    REAL(kind=real_4), DIMENSION(:, :), POINTER       :: vec_bl
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

    ! figure out the number of threads
    nthreads = 1
!$OMP PARALLEL DEFAULT(NONE) SHARED(nthreads)
!$OMP MASTER
    !$ nthreads = OMP_GET_NUM_THREADS()
!$OMP END MASTER
!$OMP END PARALLEL

    CALL dbcsr_get_info(matrix=vec, nblkcols_local=nblk_local)
    ! 4 times makes sure the table is big enough to limit collisions.
    CALL hash_table_create(fast_vec_access%hash_table,4*nblk_local)
    ! include zero for effective dealing with values not in the hash table (will return 0)
    ALLOCATE(fast_vec_access%blk_map_s(0:nblk_local))

    ! sanity check
    CALL dbcsr_get_info(matrix=vec, nblkrows_local=row)
    IF (row.GT.1) CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",467,"BUG")

    ! go through the blocks of the vector
    iblock=0
    CALL dbcsr_iterator_start(iter, vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed)
       iblock=iblock+1
       CALL hash_table_add(fast_vec_access%hash_table,col,iblock)
       fast_vec_access%blk_map_s(iblock)%ptr=>vec_bl
       fast_vec_access%blk_map_s(iblock)%assigned_thread=MOD(iblock,nthreads)
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE create_fast_row_vec_access_s

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_sym_matrix_vector_mult_s(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    REAL(kind=real_4)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_sym_m_v_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, &
                                                myprow, &
                                                pcol_group, nrows, ncols,&
                                                prow_group, row, &
                                                handle, handle1, ithread, vec_dim
    LOGICAL                                  :: transposed
    REAL(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    REAL(kind=real_4), DIMENSION(:, :), POINTER       :: data_d, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter
    TYPE(dbcsr_obj)                          :: result_row, result_col

    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col, res_fast_vec_row, res_fast_vec_col
    INTEGER                                  :: prow, pcol, rprow, rpcol

    CALL timeset(routineN, handle)
    ithread=0
! We need some work matrices as we try to exploit operations on the replicated vectors which are duplicated otherwise
    CALL dbcsr_get_info(matrix=vec_in,nfullcols_total=vec_dim)
! This is a performance hack as the new creation of a replicated vector is a fair bit more expensive
    CALL dbcsr_init(result_col)
    CALL dbcsr_set(work_col, 0.0_real_4)
    CALL dbcsr_copy(result_col, work_col)
    CALL dbcsr_init(result_row)
    CALL dbcsr_set(work_row, 0.0_real_4)
    CALL dbcsr_copy(result_row, work_row)

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group; pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol; myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)
    CALL create_fast_row_vec_access(result_row, res_fast_vec_row)
    CALL create_fast_col_vec_access(result_col, res_fast_vec_col)

! Transfer the correct parts of the input vector to the correct locations so we can do a local multiply
    CALL dbcsr_col_vec_to_rep_row_s(vec_in, work_col, work_row, fast_vec_col)

! Probably I should rename the routine above as it delivers both the replicated row and column vector

! Perform the local multiply. Here we exploit, that we have the blocks replicated on the mpi processes
! It is important to note, that the input and result vector are distributed differently (row wise, col wise respectively)
    CALL timeset(routineN//"_local_mm", handle1)

!------ perform the multiplication, we have to take car to take the correct blocks ----------

!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,transposed,ithread,pcol,prow,rpcol,rprow) &
!$OMP          SHARED(matrix,fast_vec_row,res_fast_vec_col,res_fast_vec_row,fast_vec_col)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed)
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       rprow=hash_table_get(res_fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_row%blk_map_s(pcol)%ptr) .AND.&
          ASSOCIATED(res_fast_vec_col%blk_map_s(rprow)%ptr))THEN
          IF(res_fast_vec_col%blk_map_s(rprow)%assigned_thread .EQ. ithread ) THEN
             res_fast_vec_col%blk_map_s(rprow)%ptr=res_fast_vec_col%blk_map_s(rprow)%ptr+&
               MATMUL(data_d,TRANSPOSE(fast_vec_row%blk_map_s(pcol)%ptr))
          END IF
          prow=hash_table_get(fast_vec_col%hash_table,row)
          rpcol=hash_table_get(res_fast_vec_row%hash_table,col)
          IF(res_fast_vec_row%blk_map_s(rpcol)%assigned_thread .EQ. ithread .AND. row .NE. col) THEN
             res_fast_vec_row%blk_map_s(rpcol)%ptr=res_fast_vec_row%blk_map_s(rpcol)%ptr+&
                MATMUL(TRANSPOSE(fast_vec_col%blk_map_s(prow)%ptr),data_d)
          END IF
       ELSE
          rpcol=hash_table_get(res_fast_vec_col%hash_table,col)
          prow=hash_table_get(fast_vec_row%hash_table,row)
          IF(res_fast_vec_col%blk_map_s(rpcol)%assigned_thread .EQ. ithread ) THEN
             res_fast_vec_col%blk_map_s(rpcol)%ptr=res_fast_vec_col%blk_map_s(rpcol)%ptr+&
                TRANSPOSE(MATMUL(fast_vec_row%blk_map_s(prow)%ptr,data_d))
          END IF
          rprow=hash_table_get(res_fast_vec_row%hash_table,row)
          pcol=hash_table_get(fast_vec_col%hash_table,col)
          IF(res_fast_vec_row%blk_map_s(rprow)%assigned_thread .EQ. ithread  .AND. row .NE. col ) THEN
             res_fast_vec_row%blk_map_s(rprow)%ptr=res_fast_vec_row%blk_map_s(rprow)%ptr+&
                TRANSPOSE(MATMUL(data_d,fast_vec_col%blk_map_s(pcol)%ptr))
          END IF
       END IF 
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

    ! sum all the data within a processor column to obtain the replicated result from lower
    data_vec => dbcsr_get_data_p (result_row%m%data_area, select_data_type=0.0_real_4)
    CALL dbcsr_get_info(matrix=result_row, nfullrows_local=nrows, nfullcols_local=ncols)

    CALL mp_sum(data_vec(1:nrows*ncols), pcol_group)
!
!! Convert the results to a column wise distribution, this is a bit involved as the result_row is fully replicated
!! While the result_col still has the partial results in parallel. The routine below takes care of that and saves an
!! mp_sum. Of the res_row vectors are created only taking the approriate element (0 otherwise) while the res_col
!! parallel bits are locally added. The mp_sum magically creates the correct vector
    CALL dbcsr_rep_row_to_rep_col_vec_s(work_col, result_row, res_fast_vec_row, res_fast_vec_col)

!    ! Create_the final vector by summing it to the result vector which lives on proc_col 0 lower
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_s(prow)%ptr))THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*(fast_vec_col%blk_map_s(prow)%ptr(:, :)) 
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL release_fast_vec_access(fast_vec_row)
    CALL release_fast_vec_access(fast_vec_col)
    CALL release_fast_vec_access(res_fast_vec_row)
    CALL release_fast_vec_access(res_fast_vec_col)

    CALL dbcsr_release(result_row); CALL dbcsr_release(result_col)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_sym_matrix_vector_mult_s

! **************************************************************************************************
!> \brief the real driver routine for the multiply, not all symmetries implemented yet
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrix_colvec_multiply_low_c(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    COMPLEX(kind=real_4)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_colvec_multiply_low', &
      routineP = moduleN//':'//routineN

    CHARACTER                                :: matrix_type

    matrix_type=dbcsr_get_matrix_type(matrix)
    SELECT CASE(matrix_type)
    CASE(dbcsr_type_no_symmetry)
       CALL dbcsr_matrix_vector_mult_c(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    CASE(dbcsr_type_symmetric)
       CALL dbcsr_sym_matrix_vector_mult_c(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    CASE(dbcsr_type_antisymmetric)
        ! Not yet implemented, should mainly be some prefactor magic, but who knows how antisymmetric matrices are stored???
       CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",29,"NYI, antisymmetric matrix not permitted")
    CASE DEFAULT
       CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",31,"Unknown matrix type, ...")
    END SELECT

  END SUBROUTINE dbcsr_matrix_colvec_multiply_low_c

! **************************************************************************************************
!> \brief low level routines for matrix vector multiplies
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrix_vector_mult_c(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    COMPLEX(kind=real_4)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_vector_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, &
                                                myprow, &
                                                ncols, pcol_group, nrows, &
                                                prow_group, row, &
                                                handle, handle1, ithread 
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER       :: data_d, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter
    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col
    INTEGER                                  :: prow, pcol

    CALL timeset(routineN, handle)
    ithread=0

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group 
    pcol_group=distri%d%mp_env%mp%pcol_group 
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)

! Transfer the correct parts of the input vector to the correct locations so we can do a local multiply
    CALL dbcsr_col_vec_to_rep_row_c(vec_in, work_col, work_row, fast_vec_col)

! Set the work vector for the results to 0
    CALL dbcsr_set(work_col, CMPLX(0.0, 0.0, real_4))

! Perform the local multiply. Here we exploit, that we have the blocks replicated on the mpi processes
! It is important to note, that the input and result vector are distributed differently (row wise, col wise respectively)
    CALL timeset(routineN//"_local_mm", handle1)

!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,transposed,ithread,pcol,prow) &
!$OMP          SHARED(matrix,fast_vec_col,fast_vec_row)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(fast_vec_col%blk_map_c(prow)%assigned_thread .NE. ithread ) CYCLE
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       fast_vec_col%blk_map_c(prow)%ptr=fast_vec_col%blk_map_c(prow)%ptr+&
            MATMUL(data_d,TRANSPOSE(fast_vec_row%blk_map_c(pcol)%ptr))
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

! sum all the data onto the first processor col where the original vector is stored
    data_vec => dbcsr_get_data_p (work_col%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
    CALL dbcsr_get_info(matrix=work_col, nfullrows_local=nrows, nfullcols_local=ncols)
    CALL mp_sum(data_vec(1:nrows*ncols), prow_group)

! Local copy on the first mpi col (as this is the localtion of the vec_res blocks) of the result vector
! from the replicated to the original vector. Let's play it safe and use the iterator
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_c(prow)%ptr)) THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*fast_vec_col%blk_map_c(prow)%ptr(:,:)
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL release_fast_vec_access(fast_vec_row)
    CALL release_fast_vec_access(fast_vec_col)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_matrix_vector_mult_c

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
!> \param skip_diag ...
! **************************************************************************************************
  SUBROUTINE dbcsr_matrixT_vector_mult_c(matrix, vec_in, vec_out, alpha, beta, work_row, work_col, skip_diag)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    COMPLEX(kind=real_4)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col
    LOGICAL                                  :: skip_diag

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrixT_vector_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, col_size, mypcol, &
                                                myprow, &
                                                ncols, pcol_group, nrows, &
                                                prow_group, row, row_size, &
                                                handle, handle1, ithread
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER       :: data_d, vec_bl, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col
    INTEGER                                  :: prow, pcol

    CALL timeset(routineN, handle)
    ithread=0

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group; pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol; myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)

! Set the work vector for the results to 0
    CALL dbcsr_set(work_row, CMPLX(0.0, 0.0, real_4))

! Transfer the correct parts of the input vector to the replicated vector on proc_col 0
    CALL dbcsr_iterator_start(iter, vec_in)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed, row_size=row_size, col_size=col_size)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       fast_vec_col%blk_map_c(prow)%ptr(1:row_size, 1:col_size)= vec_bl(1:row_size, 1:col_size)
    END DO
    CALL dbcsr_iterator_stop(iter)
! Replicate the data on all processore in the row
    data_vec => dbcsr_get_data_p (work_col%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
    CALL mp_bcast(data_vec, 0, prow_group)

! Perform the local multiply. Here it is obvious why the vectors are replicated on the mpi rows and cols
    CALL timeset(routineN//"local_mm", handle1)
    CALL dbcsr_get_info(matrix=work_col, nfullcols_local=ncols)
!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,row_size,col_size,transposed,ithread,prow,pcol) &
!$OMP          SHARED(matrix,fast_vec_row,fast_vec_col,skip_diag,ncols)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed, row_size=row_size, col_size=col_size)
       IF(skip_diag.AND.col==row)CYCLE
       prow=hash_table_get(fast_vec_col%hash_table,row)
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       IF ( ASSOCIATED(fast_vec_row%blk_map_c(pcol)%ptr) .AND. &
            ASSOCIATED(fast_vec_col%blk_map_c(prow)%ptr) )THEN
          IF(fast_vec_row%blk_map_c(pcol)%assigned_thread .NE. ithread ) CYCLE
          fast_vec_row%blk_map_c(pcol)%ptr=fast_vec_row%blk_map_c(pcol)%ptr+&
               MATMUL(TRANSPOSE(fast_vec_col%blk_map_c(prow)%ptr),data_d)
       ELSE
          prow=hash_table_get(fast_vec_row%hash_table,row)
          pcol=hash_table_get(fast_vec_col%hash_table,col)
          IF(fast_vec_row%blk_map_c(prow)%assigned_thread .NE. ithread ) CYCLE
          fast_vec_row%blk_map_c(prow)%ptr=fast_vec_row%blk_map_c(prow)%ptr+&
             MATMUL(TRANSPOSE(fast_vec_col%blk_map_c(pcol)%ptr),TRANSPOSE(data_d))
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

! sum all the data within a processor column to obtain the replicated result
    data_vec => dbcsr_get_data_p (work_row%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
! we use the replicated vector but the final answer is only summed to proc_col 0 for efficiency
    CALL dbcsr_get_info(matrix=work_row, nfullrows_local=nrows, nfullcols_local=ncols)
    CALL mp_sum(data_vec(1:nrows*ncols), pcol_group)

! Convert the result to a column wise distribution
    CALL dbcsr_rep_row_to_rep_col_vec_c(work_col, work_row, fast_vec_row)
    
! Create_the final vector by summing it to the result vector which lives on proc_col 0
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed, row_size=row_size)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_c(prow)%ptr)) THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*fast_vec_col%blk_map_c(prow)%ptr(:,:)
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_matrixT_vector_mult_c 

! **************************************************************************************************
!> \brief ...
!> \param vec_in ...
!> \param rep_col_vec ...
!> \param rep_row_vec ...
!> \param fast_vec_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_col_vec_to_rep_row_c(vec_in, rep_col_vec, rep_row_vec, fast_vec_col)
    TYPE(dbcsr_obj)                          :: vec_in, rep_col_vec, &
                                                rep_row_vec
    TYPE(fast_vec_access_type), INTENT(IN)   :: fast_vec_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_col_vec_to_rep_row', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, myprow, ncols, &
                                                nrows, pcol_group, &
                                                prow_group, row, handle
    INTEGER, DIMENSION(:), POINTER           :: local_cols, row_dist
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_4), DIMENSION(:), POINTER          :: data_vec, data_vec_rep
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER       :: vec_row
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

! get information about the parallel environment
    CALL dbcsr_get_info(matrix=vec_in, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group
    pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow

! Get the vector which tells us which blocks are local to which processor row in the col vec
    row_dist=> dbcsr_distribution_row_dist (dbcsr_distribution(rep_col_vec))

! Copy the local vector to the replicated on the first processor column (this is where vec_in lives)
    CALL dbcsr_get_info(matrix=rep_col_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_col_vec%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
    data_vec => dbcsr_get_data_p (vec_in%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
    IF(mypcol==0)data_vec_rep(1:nrows*ncols)=data_vec(1:nrows*ncols)
! Replicate the data along the row
    CALL mp_bcast(data_vec_rep(1:nrows*ncols), 0, prow_group)

! Here it gets a bit tricky as we are dealing with two different parallel layouts:
! The rep_col_vec contains all blocks local to the row distribution of the vector. 
! The rep_row_vec only needs the fraction which is local to the col distribution.
! However in most cases this won't the complete set of block which can be obtained from col_vector p_row i
! Anyway, as the blocks don't repeat in the col_vec, a different fraction of the row vec will be available
! on every replica in the processor column, by summing along the column we end up with the complete vector everywhere
! Hope this clarifies the idea
    CALL dbcsr_set(rep_row_vec, CMPLX(0.0, 0.0, real_4))
    CALL dbcsr_get_info(matrix=rep_row_vec, nfullrows_local=nrows, local_cols=local_cols, nfullcols_local=ncols)
    CALL dbcsr_iterator_start(iter, rep_row_vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_row, transposed)
       IF(row_dist(col)==myprow)THEN
          vec_row=TRANSPOSE(fast_vec_col%blk_map_c(hash_table_get(fast_vec_col%hash_table,col))%ptr)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)
    CALL dbcsr_get_info(matrix=rep_row_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_row_vec%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
    CALL mp_sum(data_vec_rep(1:ncols*nrows), pcol_group)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_col_vec_to_rep_row_c    
       
! **************************************************************************************************
!> \brief ...
!> \param rep_col_vec ...
!> \param rep_row_vec ...
!> \param fast_vec_row ...
!> \param fast_vec_col_add ...
! **************************************************************************************************
  SUBROUTINE dbcsr_rep_row_to_rep_col_vec_c(rep_col_vec, rep_row_vec, fast_vec_row, fast_vec_col_add)
    TYPE(dbcsr_obj)                          :: rep_col_vec, rep_row_vec
    TYPE(fast_vec_access_type), OPTIONAL     :: fast_vec_col_add
    TYPE(fast_vec_access_type)               :: fast_vec_row

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_rep_row_to_rep_col_vec', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, myprow, ncols, &
                                                nrows, pcol_group, &
                                                prow_group, row, handle
    INTEGER, DIMENSION(:), POINTER           :: col_dist
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_4), DIMENSION(:), POINTER          :: data_vec_rep
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER       :: vec_col
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

! get information about the parallel environment
    CALL dbcsr_get_info(matrix=rep_col_vec, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group
    pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol
    myprow=distri%d%mp_env%mp%myprow
! Get the vector which tells us which blocks are local to which processor col in the row vec
    col_dist=> dbcsr_distribution_col_dist (dbcsr_distribution(rep_row_vec))

! The same trick as described above with opposite direction
    CALL dbcsr_set(rep_col_vec, CMPLX(0.0, 0.0, real_4))
    CALL dbcsr_iterator_start(iter, rep_col_vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_col, transposed)
       IF(col_dist(row)==mypcol)THEN
          vec_col=TRANSPOSE(fast_vec_row%blk_map_c(hash_table_get(fast_vec_row%hash_table,row))%ptr)
       END IF
       ! this one is special and allows to add the elements of a not yet summed replicated 
       ! column vector as it appears in M*V(row_rep) as result. Save an mp_sum in the symmetric case
       IF(PRESENT(fast_vec_col_add))vec_col=vec_col+&
            fast_vec_col_add%blk_map_c(hash_table_get(fast_vec_col_add%hash_table,row))%ptr(:,:)
    END DO
    CALL dbcsr_iterator_stop(iter)
    CALL dbcsr_get_info(matrix=rep_col_vec, nfullrows_local=nrows, nfullcols_local=ncols)
    data_vec_rep => dbcsr_get_data_p (rep_col_vec%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
    CALL mp_sum(data_vec_rep(1:nrows*ncols), prow_group)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_rep_row_to_rep_col_vec_c


! **************************************************************************************************
!> \brief given a column vector, prepare the fast_vec_access container
!> \param vec ...
!> \param fast_vec_access ...
! **************************************************************************************************
  SUBROUTINE create_fast_col_vec_access_c(vec, fast_vec_access)
    TYPE(dbcsr_obj)                          :: vec
    TYPE(fast_vec_access_type)               :: fast_vec_access

    CHARACTER(LEN=*), PARAMETER :: routineN = 'create_fast_col_vec_access_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, nblk_local
    INTEGER                                  :: col, row, iblock, nthreads
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER       :: vec_bl
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

    ! figure out the number of threads
    nthreads = 1
!$OMP PARALLEL DEFAULT(NONE) SHARED(nthreads)
!$OMP MASTER
    !$ nthreads = OMP_GET_NUM_THREADS()
!$OMP END MASTER
!$OMP END PARALLEL

    CALL dbcsr_get_info(matrix=vec, nblkrows_local=nblk_local)
    ! 4 times makes sure the table is big enough to limit collisions.
    CALL hash_table_create(fast_vec_access%hash_table,4*nblk_local)
    ! include zero for effective dealing with values not in the hash table (will return 0)
    ALLOCATE(fast_vec_access%blk_map_c(0:nblk_local))

    CALL dbcsr_get_info(matrix=vec, nblkcols_local=col)
    IF (col.GT.1) CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",413,"BUG")

    ! go through the blocks of the vector
    iblock=0
    CALL dbcsr_iterator_start(iter, vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed)
       iblock=iblock+1
       CALL hash_table_add(fast_vec_access%hash_table,row,iblock)
       fast_vec_access%blk_map_c(iblock)%ptr=>vec_bl
       fast_vec_access%blk_map_c(iblock)%assigned_thread=MOD(iblock,nthreads)
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE create_fast_col_vec_access_c

! **************************************************************************************************
!> \brief given a row vector, prepare the fast_vec_access_container
!> \param vec ...
!> \param fast_vec_access ...
! **************************************************************************************************
  SUBROUTINE create_fast_row_vec_access_c(vec, fast_vec_access)
    TYPE(dbcsr_obj)                          :: vec
    TYPE(fast_vec_access_type)               :: fast_vec_access

    CHARACTER(LEN=*), PARAMETER :: routineN = 'create_fast_row_vec_access_c', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: handle, nblk_local
    INTEGER                                  :: col, row, iblock, nthreads
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER       :: vec_bl
    TYPE(dbcsr_iterator)                     :: iter

    CALL timeset(routineN, handle)

    ! figure out the number of threads
    nthreads = 1
!$OMP PARALLEL DEFAULT(NONE) SHARED(nthreads)
!$OMP MASTER
    !$ nthreads = OMP_GET_NUM_THREADS()
!$OMP END MASTER
!$OMP END PARALLEL

    CALL dbcsr_get_info(matrix=vec, nblkcols_local=nblk_local)
    ! 4 times makes sure the table is big enough to limit collisions.
    CALL hash_table_create(fast_vec_access%hash_table,4*nblk_local)
    ! include zero for effective dealing with values not in the hash table (will return 0)
    ALLOCATE(fast_vec_access%blk_map_c(0:nblk_local))

    ! sanity check
    CALL dbcsr_get_info(matrix=vec, nblkrows_local=row)
    IF (row.GT.1) CALL cp__b("dbcsr/vec/dbcsr_vector_operations_low.F",467,"BUG")

    ! go through the blocks of the vector
    iblock=0
    CALL dbcsr_iterator_start(iter, vec)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_bl, transposed)
       iblock=iblock+1
       CALL hash_table_add(fast_vec_access%hash_table,col,iblock)
       fast_vec_access%blk_map_c(iblock)%ptr=>vec_bl
       fast_vec_access%blk_map_c(iblock)%assigned_thread=MOD(iblock,nthreads)
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL timestop(handle)

  END SUBROUTINE create_fast_row_vec_access_c

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
  SUBROUTINE dbcsr_sym_matrix_vector_mult_c(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
    TYPE(dbcsr_obj)                          :: matrix, vec_in, vec_out
    COMPLEX(kind=real_4)                          :: alpha, beta
    TYPE(dbcsr_obj)                          :: work_row, work_col

    CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_sym_m_v_mult', &
      routineP = moduleN//':'//routineN

    INTEGER                                  :: col, mypcol, &
                                                myprow, &
                                                pcol_group, nrows, ncols,&
                                                prow_group, row, &
                                                handle, handle1, ithread, vec_dim
    LOGICAL                                  :: transposed
    COMPLEX(kind=real_4), DIMENSION(:), POINTER          :: data_vec
    COMPLEX(kind=real_4), DIMENSION(:, :), POINTER       :: data_d, vec_res
    TYPE(dbcsr_distribution_obj)             :: distri
    TYPE(dbcsr_iterator)                     :: iter
    TYPE(dbcsr_obj)                          :: result_row, result_col

    TYPE(fast_vec_access_type)               :: fast_vec_row, fast_vec_col, res_fast_vec_row, res_fast_vec_col
    INTEGER                                  :: prow, pcol, rprow, rpcol

    CALL timeset(routineN, handle)
    ithread=0
! We need some work matrices as we try to exploit operations on the replicated vectors which are duplicated otherwise
    CALL dbcsr_get_info(matrix=vec_in,nfullcols_total=vec_dim)
! This is a performance hack as the new creation of a replicated vector is a fair bit more expensive
    CALL dbcsr_init(result_col)
    CALL dbcsr_set(work_col, CMPLX(0.0, 0.0, real_4))
    CALL dbcsr_copy(result_col, work_col)
    CALL dbcsr_init(result_row)
    CALL dbcsr_set(work_row, CMPLX(0.0, 0.0, real_4))
    CALL dbcsr_copy(result_row, work_row)

! Collect some data about the parallel environment. We will use them later to move the vector around
    CALL dbcsr_get_info(matrix=matrix, distribution=distri)
    prow_group=distri%d%mp_env%mp%prow_group; pcol_group=distri%d%mp_env%mp%pcol_group
    mypcol=distri%d%mp_env%mp%mypcol; myprow=distri%d%mp_env%mp%myprow

    CALL create_fast_row_vec_access(work_row, fast_vec_row)
    CALL create_fast_col_vec_access(work_col, fast_vec_col)
    CALL create_fast_row_vec_access(result_row, res_fast_vec_row)
    CALL create_fast_col_vec_access(result_col, res_fast_vec_col)

! Transfer the correct parts of the input vector to the correct locations so we can do a local multiply
    CALL dbcsr_col_vec_to_rep_row_c(vec_in, work_col, work_row, fast_vec_col)

! Probably I should rename the routine above as it delivers both the replicated row and column vector

! Perform the local multiply. Here we exploit, that we have the blocks replicated on the mpi processes
! It is important to note, that the input and result vector are distributed differently (row wise, col wise respectively)
    CALL timeset(routineN//"_local_mm", handle1)

!------ perform the multiplication, we have to take car to take the correct blocks ----------

!$OMP PARALLEL DEFAULT(NONE) PRIVATE(row,col,iter,data_d,transposed,ithread,pcol,prow,rpcol,rprow) &
!$OMP          SHARED(matrix,fast_vec_row,res_fast_vec_col,res_fast_vec_row,fast_vec_col)
    !$ ithread = omp_get_thread_num ()
    CALL dbcsr_iterator_start(iter, matrix, shared=.FALSE.)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, data_d, transposed)
       pcol=hash_table_get(fast_vec_row%hash_table,col)
       rprow=hash_table_get(res_fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_row%blk_map_c(pcol)%ptr) .AND.&
          ASSOCIATED(res_fast_vec_col%blk_map_c(rprow)%ptr))THEN
          IF(res_fast_vec_col%blk_map_c(rprow)%assigned_thread .EQ. ithread ) THEN
             res_fast_vec_col%blk_map_c(rprow)%ptr=res_fast_vec_col%blk_map_c(rprow)%ptr+&
               MATMUL(data_d,TRANSPOSE(fast_vec_row%blk_map_c(pcol)%ptr))
          END IF
          prow=hash_table_get(fast_vec_col%hash_table,row)
          rpcol=hash_table_get(res_fast_vec_row%hash_table,col)
          IF(res_fast_vec_row%blk_map_c(rpcol)%assigned_thread .EQ. ithread .AND. row .NE. col) THEN
             res_fast_vec_row%blk_map_c(rpcol)%ptr=res_fast_vec_row%blk_map_c(rpcol)%ptr+&
                MATMUL(TRANSPOSE(fast_vec_col%blk_map_c(prow)%ptr),data_d)
          END IF
       ELSE
          rpcol=hash_table_get(res_fast_vec_col%hash_table,col)
          prow=hash_table_get(fast_vec_row%hash_table,row)
          IF(res_fast_vec_col%blk_map_c(rpcol)%assigned_thread .EQ. ithread ) THEN
             res_fast_vec_col%blk_map_c(rpcol)%ptr=res_fast_vec_col%blk_map_c(rpcol)%ptr+&
                TRANSPOSE(MATMUL(fast_vec_row%blk_map_c(prow)%ptr,data_d))
          END IF
          rprow=hash_table_get(res_fast_vec_row%hash_table,row)
          pcol=hash_table_get(fast_vec_col%hash_table,col)
          IF(res_fast_vec_row%blk_map_c(rprow)%assigned_thread .EQ. ithread  .AND. row .NE. col ) THEN
             res_fast_vec_row%blk_map_c(rprow)%ptr=res_fast_vec_row%blk_map_c(rprow)%ptr+&
                TRANSPOSE(MATMUL(data_d,fast_vec_col%blk_map_c(pcol)%ptr))
          END IF
       END IF 
    END DO
    CALL dbcsr_iterator_stop(iter)
!$OMP END PARALLEL

    CALL timestop(handle1)

    ! sum all the data within a processor column to obtain the replicated result from lower
    data_vec => dbcsr_get_data_p (result_row%m%data_area, select_data_type=CMPLX(0.0, 0.0, real_4))
    CALL dbcsr_get_info(matrix=result_row, nfullrows_local=nrows, nfullcols_local=ncols)

    CALL mp_sum(data_vec(1:nrows*ncols), pcol_group)
!
!! Convert the results to a column wise distribution, this is a bit involved as the result_row is fully replicated
!! While the result_col still has the partial results in parallel. The routine below takes care of that and saves an
!! mp_sum. Of the res_row vectors are created only taking the approriate element (0 otherwise) while the res_col
!! parallel bits are locally added. The mp_sum magically creates the correct vector
    CALL dbcsr_rep_row_to_rep_col_vec_c(work_col, result_row, res_fast_vec_row, res_fast_vec_col)

!    ! Create_the final vector by summing it to the result vector which lives on proc_col 0 lower
    CALL dbcsr_iterator_start(iter, vec_out)
    DO WHILE (dbcsr_iterator_blocks_left(iter))
       CALL dbcsr_iterator_next_block(iter, row, col, vec_res, transposed)
       prow=hash_table_get(fast_vec_col%hash_table,row)
       IF(ASSOCIATED(fast_vec_col%blk_map_c(prow)%ptr))THEN
          vec_res(:, :)= beta*vec_res(:, :)+alpha*(fast_vec_col%blk_map_c(prow)%ptr(:, :)) 
       ELSE
          vec_res(:, :)= beta*vec_res(:, :)
       END IF
    END DO
    CALL dbcsr_iterator_stop(iter)

    CALL release_fast_vec_access(fast_vec_row)
    CALL release_fast_vec_access(fast_vec_col)
    CALL release_fast_vec_access(res_fast_vec_row)
    CALL release_fast_vec_access(res_fast_vec_col)

    CALL dbcsr_release(result_row); CALL dbcsr_release(result_col)

    CALL timestop(handle)

  END SUBROUTINE dbcsr_sym_matrix_vector_mult_c


END MODULE dbcsr_vector_operations_low
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief operations for skinny matrices/vectors expressed in dbcsr form
!> \par History
!>       2014.10 created [Florian Schiffmann]
!> \author Florian Schiffmann
! **************************************************************************************************

MODULE dbcsr_vector_operations
   USE dbcsr_data_methods,              ONLY: dbcsr_scalar,&
                                              dbcsr_scalar_fill_all
   USE dbcsr_types,                     ONLY: dbcsr_obj,&
                                              dbcsr_scalar_type,&
                                              dbcsr_type_complex_4,&
                                              dbcsr_type_complex_8,&
                                              dbcsr_type_real_4,&
                                              dbcsr_type_real_8
   USE dbcsr_vector_operations_low,     ONLY: create_col_vec_from_matrix,&
                                              create_replicated_col_vec_from_matrix,&
                                              create_replicated_row_vec_from_matrix,&
                                              create_row_vec_from_matrix,&
                                              dbcsr_matrix_colvec_multiply_low
   USE kinds,                           ONLY: real_4,&
                                              real_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dbcsr_vector_operations'

   PUBLIC :: dbcsr_matrix_colvec_multiply, &
             create_col_vec_from_matrix, &
             create_replicated_col_vec_from_matrix, &
             create_replicated_row_vec_from_matrix

   INTERFACE dbcsr_matrix_colvec_multiply
      MODULE PROCEDURE dbcsr_matrix_colvec_multiply_d, dbcsr_matrix_colvec_multiply_s, &
         dbcsr_matrix_colvec_multiply_z, dbcsr_matrix_colvec_multiply_c
   END INTERFACE
CONTAINS

! **************************************************************************************************
!> \brief multiplies a dbcsr matrix with a column vector like dbcsr matrix.
!>        v_out=beta*v_out+alpha*M*V
!>        IMPORTANT: vector have to be created via the vec create routines:
!>                   create_col_vec_from_matrix,
!>                   create_row_vec_from_matrix,
!>                   create_rep_col_vec_from_matrix,
!>                   create_rep_row_vec_from_matrix
!>        WARNING:   Do not filter the vectors as they are assumed to be non
!>                   sparse in the underlying routines. If your vector is
!>                   sparse, fill it!!!
!> \param matrix a dbcsr matrix
!> \param vec_in the vector to be multiplied (only available on proc_col 0)
!> \param vec_out the result vector (only available on proc_col 0)
!> \param alpha  as described in formula
!> \param beta  as described in formula
!> \param work_row a work row vector replicated on all proc_cols.
!> \param work_col a work col vector replicated on all proc_rows.
! **************************************************************************************************

   SUBROUTINE dbcsr_matrix_colvec_multiply_d(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
      TYPE(dbcsr_obj)                                    :: matrix, vec_in, vec_out
      REAL(real_8)                                       :: alpha, beta
      TYPE(dbcsr_obj)                                    :: work_row, work_col

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_colvec_multiply_d', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: matrix_type
      TYPE(dbcsr_scalar_type)                            :: a, b

      matrix_type = matrix%m%data_type
      SELECT CASE (matrix_type)
      CASE (dbcsr_type_real_8, dbcsr_type_real_4, dbcsr_type_complex_4, dbcsr_type_complex_8)
         ! We can convert the prefactors if necessary
      CASE DEFAULT
         CALL cp__b("dbcsr/vec/dbcsr_vector_operations.F",84,"Unknown data type in matrix")
      END SELECT

      a = dbcsr_scalar(alpha); b = dbcsr_scalar(beta)
      CALL dbcsr_scalar_fill_all(a); CALL dbcsr_scalar_fill_all(b)

      CALL dbcsr_matrix_colvec_multiply_low(matrix, vec_in, vec_out, a, b, work_row, work_col)

   END SUBROUTINE dbcsr_matrix_colvec_multiply_d

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
   SUBROUTINE dbcsr_matrix_colvec_multiply_s(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
      TYPE(dbcsr_obj)                                    :: matrix, vec_in, vec_out
      REAL(real_4)                                       :: alpha, beta
      TYPE(dbcsr_obj)                                    :: work_row, work_col

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_colvec_multiply_s', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: matrix_type
      TYPE(dbcsr_scalar_type)                            :: a, b

      matrix_type = matrix%m%data_type
      SELECT CASE (matrix_type)
      CASE (dbcsr_type_real_8, dbcsr_type_real_4, dbcsr_type_complex_4, dbcsr_type_complex_8)
         ! We can convert the prefactors if necessary
      CASE DEFAULT
         CALL cp__b("dbcsr/vec/dbcsr_vector_operations.F",120,"Unknown data type in matrix")
      END SELECT

      a = dbcsr_scalar(alpha); b = dbcsr_scalar(beta)
      CALL dbcsr_scalar_fill_all(a); CALL dbcsr_scalar_fill_all(b)

      CALL dbcsr_matrix_colvec_multiply_low(matrix, vec_in, vec_out, a, b, work_row, work_col)

   END SUBROUTINE dbcsr_matrix_colvec_multiply_s

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
   SUBROUTINE dbcsr_matrix_colvec_multiply_c(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
      TYPE(dbcsr_obj)                                    :: matrix, vec_in, vec_out
      COMPLEX(real_4)                                    :: alpha, beta
      TYPE(dbcsr_obj)                                    :: work_row, work_col

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_colvec_multiply_c', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: matrix_type
      TYPE(dbcsr_scalar_type)                            :: a, b

      matrix_type = matrix%m%data_type
      SELECT CASE (matrix_type)
      CASE (dbcsr_type_real_8, dbcsr_type_real_4)
         CALL cp__b("dbcsr/vec/dbcsr_vector_operations.F",154,"Complex prefactors dissallowed for real matrices")
      CASE (dbcsr_type_complex_4, dbcsr_type_complex_8)
         ! We can convert the prefactors if necessary
      CASE DEFAULT
         CALL cp__b("dbcsr/vec/dbcsr_vector_operations.F",158,"Unknown data type in matrix")
      END SELECT

      a = dbcsr_scalar(alpha); b = dbcsr_scalar(beta)
      CALL dbcsr_scalar_fill_all(a); CALL dbcsr_scalar_fill_all(b)

      CALL dbcsr_matrix_colvec_multiply_low(matrix, vec_in, vec_out, a, b, work_row, work_col)

   END SUBROUTINE dbcsr_matrix_colvec_multiply_c

! **************************************************************************************************
!> \brief ...
!> \param matrix ...
!> \param vec_in ...
!> \param vec_out ...
!> \param alpha ...
!> \param beta ...
!> \param work_row ...
!> \param work_col ...
! **************************************************************************************************
   SUBROUTINE dbcsr_matrix_colvec_multiply_z(matrix, vec_in, vec_out, alpha, beta, work_row, work_col)
      TYPE(dbcsr_obj)                                    :: matrix, vec_in, vec_out
      COMPLEX(real_8)                                    :: alpha, beta
      TYPE(dbcsr_obj)                                    :: work_row, work_col

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dbcsr_matrix_colvec_multiply_z', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: matrix_type
      TYPE(dbcsr_scalar_type)                            :: a, b

      matrix_type = matrix%m%data_type
      SELECT CASE (matrix_type)
      CASE (dbcsr_type_real_8, dbcsr_type_real_4)
         CALL cp__b("dbcsr/vec/dbcsr_vector_operations.F",192,"Complex prefactors dissallowed for real matrices")
      CASE (dbcsr_type_complex_4, dbcsr_type_complex_8)
         ! We can convert the prefactors if necessary
      CASE DEFAULT
         CALL cp__b("dbcsr/vec/dbcsr_vector_operations.F",196,"Unknown data type in matrix")
      END SELECT

      a = dbcsr_scalar(alpha); b = dbcsr_scalar(beta)
      CALL dbcsr_scalar_fill_all(a); CALL dbcsr_scalar_fill_all(b)

      CALL dbcsr_matrix_colvec_multiply_low(matrix, vec_in, vec_out, a, b, work_row, work_col)

   END SUBROUTINE dbcsr_matrix_colvec_multiply_z

END MODULE dbcsr_vector_operations