!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief wrapper for the pools of matrixes
!> \par History
!>      05.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE qs_matrix_pools
   USE cp_blacs_env,                    ONLY: cp_blacs_env_type
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type,&
                                              cp_fm_pool_type,&
                                              fm_pool_create,&
                                              fm_pool_get_el_struct,&
                                              fm_pool_release,&
                                              fm_pool_retain,&
                                              fm_pools_dealloc
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_get,&
                                              cp_fm_struct_get_ncol_block,&
                                              cp_fm_struct_get_nrow_block,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_matrix_pools'
   INTEGER, SAVE, PRIVATE :: last_mpools_id = 0

   PUBLIC :: qs_matrix_pools_type
   PUBLIC :: mpools_retain, mpools_release, mpools_get, &
             mpools_create, mpools_rebuild_fm_pools

! **************************************************************************************************
!> \brief container for the pools of matrixes used by qs
!> \param id_nr identification number
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param ao_mo_fm_pools pools with (ao x mo) full matrixes (same order as
!>        c).
!> \param ao_ao_fm_pools pools with (ao x ao) full matrixes (same order as
!>        c).
!> \param mo_mo_fm_pools pools with (mo x mo) full matrixes (same
!>        order as c).
!> \param ao_mosub_fm_pools pools with (ao x mosub) full matrixes, where mosub
!>        are a subset of the mos
!> \param mosub_mosub_fm_pools pools with (mosub x mosub) full matrixes, where mosub
!>        are a subset of the mos
!>
!> \param maxao_maxao_fm_pools pool of matrixes big enough to accomodate any
!>        aoxao matrix (useful for temp matrixes)
!> \param maxao_maxmo_fm_pools pool of matrixes big enough to accomodate any
!>        aoxmo matrix (useful for temp matrixes)
!> \param maxmo_maxmo_fm_pools pool of matrixes big enough to accomodate any
!>        moxmo matrix (useful for temp matrixes)
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE qs_matrix_pools_type
      INTEGER :: id_nr, ref_count
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER        :: ao_mo_fm_pools, &
                                                               ao_ao_fm_pools, mo_mo_fm_pools
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER        :: ao_mosub_fm_pools, &
                                                               mosub_mosub_fm_pools
   END TYPE qs_matrix_pools_type

CONTAINS

! **************************************************************************************************
!> \brief retains the given qs_matrix_pools_type
!> \param mpools the matrix pools type to retain
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE mpools_retain(mpools)
      TYPE(qs_matrix_pools_type), POINTER                :: mpools

      CHARACTER(len=*), PARAMETER :: routineN = 'mpools_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(mpools)))CALL cp__a("qs_matrix_pools.F",90)
      IF(.NOT.(mpools%ref_count > 0))CALL cp__a("qs_matrix_pools.F",91)
      mpools%ref_count = mpools%ref_count+1
   END SUBROUTINE mpools_retain

! **************************************************************************************************
!> \brief releases the given mpools
!> \param mpools the matrix pools type to retain
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE mpools_release(mpools)
      TYPE(qs_matrix_pools_type), POINTER                :: mpools

      CHARACTER(len=*), PARAMETER :: routineN = 'mpools_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(mpools)) THEN
         IF(.NOT.(mpools%ref_count > 0))CALL cp__a("qs_matrix_pools.F",108)
         mpools%ref_count = mpools%ref_count-1
         IF (mpools%ref_count == 0) THEN
            CALL fm_pools_dealloc(mpools%ao_mo_fm_pools)
            CALL fm_pools_dealloc(mpools%ao_ao_fm_pools)
            CALL fm_pools_dealloc(mpools%mo_mo_fm_pools)
            IF (ASSOCIATED(mpools%ao_mosub_fm_pools)) THEN
               CALL fm_pools_dealloc(mpools%ao_mosub_fm_pools)
            END IF
            IF (ASSOCIATED(mpools%mosub_mosub_fm_pools)) THEN
               CALL fm_pools_dealloc(mpools%mosub_mosub_fm_pools)
            END IF
            DEALLOCATE (mpools)
         END IF
      END IF
      NULLIFY (mpools)
   END SUBROUTINE mpools_release

! **************************************************************************************************
!> \brief returns various attributes of the mpools (notably the pools
!>      contained in it)
!> \param mpools the matrix pools object you want info about
!> \param ao_mo_fm_pools ...
!> \param ao_ao_fm_pools ...
!> \param mo_mo_fm_pools ...
!> \param ao_mosub_fm_pools ...
!> \param mosub_mosub_fm_pools ...
!> \param maxao_maxmo_fm_pool ...
!> \param maxao_maxao_fm_pool ...
!> \param maxmo_maxmo_fm_pool ...
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE mpools_get(mpools, ao_mo_fm_pools, ao_ao_fm_pools, &
                         mo_mo_fm_pools, ao_mosub_fm_pools, mosub_mosub_fm_pools, &
                         maxao_maxmo_fm_pool, maxao_maxao_fm_pool, maxmo_maxmo_fm_pool)
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(cp_fm_pool_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: ao_mo_fm_pools, ao_ao_fm_pools, &
                                                            mo_mo_fm_pools, ao_mosub_fm_pools, &
                                                            mosub_mosub_fm_pools
      TYPE(cp_fm_pool_type), OPTIONAL, POINTER           :: maxao_maxmo_fm_pool, &
                                                            maxao_maxao_fm_pool, &
                                                            maxmo_maxmo_fm_pool

      CHARACTER(len=*), PARAMETER :: routineN = 'mpools_get', routineP = moduleN//':'//routineN

      IF (PRESENT(ao_mo_fm_pools)) ao_mo_fm_pools => mpools%ao_mo_fm_pools
      IF (PRESENT(maxao_maxmo_fm_pool)) THEN
         IF (ASSOCIATED(mpools%ao_mo_fm_pools)) THEN
            maxao_maxmo_fm_pool => mpools%ao_mo_fm_pools(1)%pool
         ELSE
            NULLIFY (maxao_maxmo_fm_pool) ! raise an error?
         END IF
      END IF
      IF (PRESENT(ao_ao_fm_pools)) ao_ao_fm_pools => mpools%ao_ao_fm_pools
      IF (PRESENT(maxao_maxao_fm_pool)) THEN
         IF (ASSOCIATED(mpools%ao_ao_fm_pools)) THEN
            maxao_maxao_fm_pool => mpools%ao_ao_fm_pools(1)%pool
         ELSE
            NULLIFY (maxao_maxao_fm_pool) ! raise an error?
         END IF
      END IF
      IF (PRESENT(mo_mo_fm_pools)) mo_mo_fm_pools => mpools%mo_mo_fm_pools
      IF (PRESENT(maxmo_maxmo_fm_pool)) THEN
         IF (ASSOCIATED(mpools%mo_mo_fm_pools)) THEN
            maxmo_maxmo_fm_pool => mpools%mo_mo_fm_pools(1)%pool
         ELSE
            NULLIFY (maxmo_maxmo_fm_pool) ! raise an error?
         END IF
      END IF
      IF (PRESENT(ao_mosub_fm_pools)) ao_mosub_fm_pools => mpools%ao_mosub_fm_pools
      IF (PRESENT(mosub_mosub_fm_pools)) mosub_mosub_fm_pools => mpools%mosub_mosub_fm_pools
   END SUBROUTINE mpools_get

! **************************************************************************************************
!> \brief creates a mpools
!> \param mpools the mpools to create
!> \par History
!>      04.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE mpools_create(mpools)
      TYPE(qs_matrix_pools_type), POINTER                :: mpools

      CHARACTER(len=*), PARAMETER :: routineN = 'mpools_create', routineP = moduleN//':'//routineN

      ALLOCATE (mpools)
      NULLIFY (mpools%ao_ao_fm_pools, mpools%ao_mo_fm_pools, &
               mpools%mo_mo_fm_pools, mpools%ao_mosub_fm_pools, &
               mpools%mosub_mosub_fm_pools)
      mpools%ref_count = 1
      last_mpools_id = last_mpools_id+1
      mpools%id_nr = last_mpools_id
   END SUBROUTINE mpools_create

! **************************************************************************************************
!> \brief rebuilds the pools of the (ao x mo, ao x ao , mo x mo) full matrixes
!> \param mpools the environement where the pools should be rebuilt
!> \param mos the molecular orbitals (qs_env%c), must contain up to
!>        date nmo and nao
!> \param blacs_env the blacs environment of the full matrixes
!> \param para_env the parallel environment of the matrixes
!> \param nmosub number of the orbitals for the creation
!>        of the pools containing only a subset of mos (OPTIONAL)
!> \par History
!>      08.2002 created [fawzi]
!>      04.2005 added pools for a subset of mos [MI]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE mpools_rebuild_fm_pools(mpools, mos, blacs_env, para_env, &
                                      nmosub)
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, DIMENSION(2), INTENT(IN), OPTIONAL        :: nmosub

      CHARACTER(len=*), PARAMETER :: routineN = 'mpools_rebuild_fm_pools', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, max_nmo, min_nmo, nao, &
                                                            ncg, ncol_block, nmo, nrg, nrow_block, &
                                                            nspins
      LOGICAL                                            :: prepare_subset, should_rebuild
      TYPE(cp_fm_pool_type), POINTER                     :: p_att
      TYPE(cp_fm_struct_type), POINTER                   :: fmstruct

      CALL timeset(routineN, handle)

      nrow_block = cp_fm_struct_get_nrow_block()
      ncol_block = cp_fm_struct_get_ncol_block()

      NULLIFY (fmstruct, p_att)
      prepare_subset = .FALSE.
      IF (PRESENT(nmosub)) THEN
         IF (nmosub(1) > 0) prepare_subset = .TRUE.
      END IF

      IF(.NOT.(ASSOCIATED(mos)))CALL cp__a("qs_matrix_pools.F",248)
      IF (.NOT. ASSOCIATED(mpools)) THEN
         CALL mpools_create(mpools)
      END IF
      nspins = SIZE(mos)

      IF (ASSOCIATED(mpools%ao_mo_fm_pools)) THEN
         IF (nspins /= SIZE(mpools%ao_mo_fm_pools)) THEN
            CALL fm_pools_dealloc(mpools%ao_mo_fm_pools)
         END IF
      END IF
      IF (.NOT. ASSOCIATED(mpools%ao_mo_fm_pools)) THEN
         ALLOCATE (mpools%ao_mo_fm_pools(nspins))
         DO ispin = 1, nspins
            NULLIFY (mpools%ao_mo_fm_pools(ispin)%pool)
         END DO
      END IF

      IF (ASSOCIATED(mpools%ao_ao_fm_pools)) THEN
         IF (nspins /= SIZE(mpools%ao_ao_fm_pools)) THEN
            CALL fm_pools_dealloc(mpools%ao_ao_fm_pools)
         END IF
      END IF
      IF (.NOT. ASSOCIATED(mpools%ao_ao_fm_pools)) THEN
         ALLOCATE (mpools%ao_ao_fm_pools(nspins))
         DO ispin = 1, nspins
            NULLIFY (mpools%ao_ao_fm_pools(ispin)%pool)
         END DO
      END IF

      IF (ASSOCIATED(mpools%mo_mo_fm_pools)) THEN
         IF (nspins /= SIZE(mpools%mo_mo_fm_pools)) THEN
            CALL fm_pools_dealloc(mpools%mo_mo_fm_pools)
         END IF
      END IF
      IF (.NOT. ASSOCIATED(mpools%mo_mo_fm_pools)) THEN
         ALLOCATE (mpools%mo_mo_fm_pools(nspins))
         DO ispin = 1, nspins
            NULLIFY (mpools%mo_mo_fm_pools(ispin)%pool)
         END DO
      END IF

      IF (prepare_subset) THEN

         IF (ASSOCIATED(mpools%ao_mosub_fm_pools)) THEN
            IF (nspins /= SIZE(mpools%ao_mosub_fm_pools)) THEN
               CALL fm_pools_dealloc(mpools%ao_mosub_fm_pools)
            END IF
         END IF
         IF (.NOT. ASSOCIATED(mpools%ao_mosub_fm_pools)) THEN
            ALLOCATE (mpools%ao_mosub_fm_pools(nspins))
            DO ispin = 1, nspins
               NULLIFY (mpools%ao_mosub_fm_pools(ispin)%pool)
            END DO
         END IF

         IF (ASSOCIATED(mpools%mosub_mosub_fm_pools)) THEN
            IF (nspins /= SIZE(mpools%mosub_mosub_fm_pools)) THEN
               CALL fm_pools_dealloc(mpools%mosub_mosub_fm_pools)
            END IF
         END IF
         IF (.NOT. ASSOCIATED(mpools%mosub_mosub_fm_pools)) THEN
            ALLOCATE (mpools%mosub_mosub_fm_pools(nspins))
            DO ispin = 1, nspins
               NULLIFY (mpools%mosub_mosub_fm_pools(ispin)%pool)
            END DO
         END IF

      END IF ! prepare_subset

      CALL get_mo_set(mos(1)%mo_set, nao=nao, nmo=min_nmo)
      max_nmo = min_nmo
      DO ispin = 2, SIZE(mos)
         CALL get_mo_set(mos(ispin)%mo_set, nmo=nmo)
         IF (max_nmo < nmo) THEN
            CALL cp__b("qs_matrix_pools.F",323,"the mo with the most orbitals must be the first ")
         END IF
         min_nmo = MIN(min_nmo, nmo)
      END DO

      ! aoao pools
      should_rebuild = .FALSE.
      DO ispin = 1, nspins
         p_att => mpools%ao_ao_fm_pools(ispin)%pool
         should_rebuild = (should_rebuild .OR. (.NOT. ASSOCIATED(p_att)))
         IF (.NOT. should_rebuild) THEN
            fmstruct => fm_pool_get_el_struct(mpools%ao_ao_fm_pools(ispin)%pool)
            CALL cp_fm_struct_get(fmstruct, nrow_global=nrg, &
                                  ncol_global=ncg)
            CALL get_mo_set(mos(1)%mo_set, nao=nao, nmo=nmo)
            should_rebuild = nao /= nrg .OR. nao /= ncg
         END IF
      END DO
      IF (should_rebuild) THEN
         DO ispin = 1, nspins
            CALL fm_pool_release(mpools%ao_ao_fm_pools(ispin)%pool)
         END DO

         CALL cp_fm_struct_create(fmstruct, nrow_global=nao, &
                                  ncol_global=nao, para_env=para_env, &
                                  context=blacs_env, &
                                  nrow_block=nrow_block, &
                                  ncol_block=ncol_block)
         CALL fm_pool_create(mpools%ao_ao_fm_pools(1)%pool, fmstruct)
         CALL cp_fm_struct_release(fmstruct)
         DO ispin = 2, SIZE(mos)
            mpools%ao_ao_fm_pools(ispin)%pool => mpools%ao_ao_fm_pools(1)%pool
            CALL fm_pool_retain(mpools%ao_ao_fm_pools(1)%pool)
         END DO
      END IF

      ! aomo pools
      should_rebuild = .FALSE.
      DO ispin = 1, nspins
         p_att => mpools%ao_mo_fm_pools(ispin)%pool
         should_rebuild = (should_rebuild .OR. (.NOT. ASSOCIATED(p_att)))
         IF (.NOT. should_rebuild) THEN
            fmstruct => fm_pool_get_el_struct(mpools%ao_mo_fm_pools(ispin) &
                                              %pool)
            CALL cp_fm_struct_get(fmstruct, nrow_global=nrg, &
                                  ncol_global=ncg)
            CALL get_mo_set(mos(1)%mo_set, nao=nao, nmo=nmo)
            should_rebuild = nao /= nrg .OR. nmo /= ncg
         END IF
      END DO
      IF (should_rebuild) THEN
         DO ispin = 1, nspins
            CALL fm_pool_release(mpools%ao_mo_fm_pools(ispin)%pool)
         END DO

         IF (max_nmo == min_nmo) THEN
            CALL cp_fm_struct_create(fmstruct, nrow_global=nao, &
                                     ncol_global=max_nmo, para_env=para_env, &
                                     context=blacs_env, &
                                     nrow_block=nrow_block, &
                                     ncol_block=ncol_block)
            CALL fm_pool_create(mpools%ao_mo_fm_pools(1)%pool, fmstruct)
            CALL cp_fm_struct_release(fmstruct)
            DO ispin = 2, SIZE(mos)
               mpools%ao_mo_fm_pools(ispin)%pool => mpools%ao_mo_fm_pools(1)%pool
               CALL fm_pool_retain(mpools%ao_mo_fm_pools(1)%pool)
            END DO
         ELSE
            DO ispin = 1, SIZE(mos)
               CALL get_mo_set(mos(ispin)%mo_set, nmo=nmo, nao=nao)
               CALL cp_fm_struct_create(fmstruct, nrow_global=nao, &
                                        ncol_global=nmo, para_env=para_env, &
                                        context=blacs_env, &
                                        nrow_block=nrow_block, &
                                        ncol_block=ncol_block)
               CALL fm_pool_create(mpools%ao_mo_fm_pools(ispin)%pool, &
                                   fmstruct)
               CALL cp_fm_struct_release(fmstruct)
            END DO
         END IF
      END IF

      ! momo pools
      should_rebuild = .FALSE.
      DO ispin = 1, nspins
         p_att => mpools%mo_mo_fm_pools(ispin)%pool
         should_rebuild = (should_rebuild .OR. (.NOT. ASSOCIATED(p_att)))
         IF (.NOT. should_rebuild) THEN
            fmstruct => fm_pool_get_el_struct(p_att)
            CALL cp_fm_struct_get(fmstruct, nrow_global=nrg, &
                                  ncol_global=ncg)
            CALL get_mo_set(mos(1)%mo_set, nao=nao, nmo=nmo)
            should_rebuild = nmo /= nrg .OR. nmo /= ncg
         END IF
      END DO
      IF (should_rebuild) THEN
         DO ispin = 1, nspins
            CALL fm_pool_release(mpools%mo_mo_fm_pools(ispin)%pool)
         END DO

         IF (max_nmo == min_nmo) THEN
            CALL cp_fm_struct_create(fmstruct, nrow_global=max_nmo, &
                                     ncol_global=max_nmo, para_env=para_env, &
                                     context=blacs_env, &
                                     nrow_block=nrow_block, &
                                     ncol_block=ncol_block)
            CALL fm_pool_create(mpools%mo_mo_fm_pools(1)%pool, &
                                fmstruct)
            CALL cp_fm_struct_release(fmstruct)
            DO ispin = 2, SIZE(mos)
               mpools%mo_mo_fm_pools(ispin)%pool => mpools%mo_mo_fm_pools(1)%pool
               CALL fm_pool_retain(mpools%mo_mo_fm_pools(1)%pool)
            END DO
         ELSE
            DO ispin = 1, SIZE(mos)
               NULLIFY (mpools%mo_mo_fm_pools(ispin)%pool)
               CALL get_mo_set(mos(ispin)%mo_set, nmo=nmo, nao=nao)
               CALL cp_fm_struct_create(fmstruct, nrow_global=nmo, &
                                        ncol_global=nmo, para_env=para_env, &
                                        context=blacs_env, &
                                        nrow_block=nrow_block, &
                                        ncol_block=ncol_block)
               CALL fm_pool_create(mpools%mo_mo_fm_pools(ispin)%pool, &
                                   fmstruct)
               CALL cp_fm_struct_release(fmstruct)
            END DO
         END IF
      END IF

      IF (prepare_subset) THEN
         ! aomosub pools
         should_rebuild = .FALSE.
         DO ispin = 1, nspins
            p_att => mpools%ao_mosub_fm_pools(ispin)%pool
            should_rebuild = (should_rebuild .OR. (.NOT. ASSOCIATED(p_att)))
            IF (.NOT. should_rebuild) THEN
               fmstruct => fm_pool_get_el_struct(mpools%ao_mosub_fm_pools(ispin) &
                                                 %pool)
               CALL cp_fm_struct_get(fmstruct, nrow_global=nrg, &
                                     ncol_global=ncg)
               CALL get_mo_set(mos(1)%mo_set, nao=nao)
               should_rebuild = nao /= nrg .OR. nmosub(ispin) /= ncg
            END IF
         END DO
         IF (should_rebuild) THEN
            DO ispin = 1, nspins
               CALL fm_pool_release(mpools%ao_mosub_fm_pools(ispin)%pool)
            END DO

            IF (nspins == 1 .OR. nmosub(1) == nmosub(2)) THEN
               CALL cp_fm_struct_create(fmstruct, nrow_global=nao, &
                                        ncol_global=nmosub(1), para_env=para_env, &
                                        context=blacs_env, &
                                        nrow_block=nrow_block, &
                                        ncol_block=ncol_block)
               CALL fm_pool_create(mpools%ao_mosub_fm_pools(1)%pool, fmstruct)
               CALL cp_fm_struct_release(fmstruct)
               DO ispin = 2, SIZE(mos)
                  mpools%ao_mosub_fm_pools(ispin)%pool => mpools%ao_mosub_fm_pools(1)%pool
                  CALL fm_pool_retain(mpools%ao_mosub_fm_pools(1)%pool)
               END DO
            ELSE
               DO ispin = 1, SIZE(mos)
                  CALL get_mo_set(mos(ispin)%mo_set, nao=nao)
                  CALL cp_fm_struct_create(fmstruct, nrow_global=nao, &
                                           ncol_global=nmosub(1), para_env=para_env, &
                                           context=blacs_env, &
                                           nrow_block=nrow_block, &
                                           ncol_block=ncol_block)
                  CALL fm_pool_create(mpools%ao_mosub_fm_pools(ispin)%pool, &
                                      fmstruct)
                  CALL cp_fm_struct_release(fmstruct)
               END DO
            END IF
         END IF ! should_rebuild

         ! mosubmosub pools
         should_rebuild = .FALSE.
         DO ispin = 1, nspins
            p_att => mpools%mosub_mosub_fm_pools(ispin)%pool
            should_rebuild = (should_rebuild .OR. (.NOT. ASSOCIATED(p_att)))
            IF (.NOT. should_rebuild) THEN
               fmstruct => fm_pool_get_el_struct(p_att)
               CALL cp_fm_struct_get(fmstruct, nrow_global=nrg, &
                                     ncol_global=ncg)
               should_rebuild = nmosub(ispin) /= nrg .OR. nmosub(ispin) /= ncg
            END IF
         END DO
         IF (should_rebuild) THEN
            DO ispin = 1, nspins
               CALL fm_pool_release(mpools%mosub_mosub_fm_pools(ispin)%pool)
            END DO

            IF (nspins == 1 .OR. nmosub(1) == nmosub(2)) THEN
               CALL cp_fm_struct_create(fmstruct, nrow_global=nmosub(1), &
                                        ncol_global=nmosub(1), para_env=para_env, &
                                        context=blacs_env, &
                                        nrow_block=nrow_block, &
                                        ncol_block=ncol_block)
               CALL fm_pool_create(mpools%mosub_mosub_fm_pools(1)%pool, &
                                   fmstruct)
               CALL cp_fm_struct_release(fmstruct)
               DO ispin = 2, SIZE(mos)
                  mpools%mosub_mosub_fm_pools(ispin)%pool => mpools%mosub_mosub_fm_pools(1)%pool
                  CALL fm_pool_retain(mpools%mosub_mosub_fm_pools(1)%pool)
               END DO
            ELSE
               DO ispin = 1, SIZE(mos)
                  NULLIFY (mpools%mosub_mosub_fm_pools(ispin)%pool)
                  CALL cp_fm_struct_create(fmstruct, nrow_global=nmosub(ispin), &
                                           ncol_global=nmosub(ispin), para_env=para_env, &
                                           context=blacs_env, &
                                           nrow_block=nrow_block, &
                                           ncol_block=ncol_block)
                  CALL fm_pool_create(mpools%mosub_mosub_fm_pools(ispin)%pool, &
                                      fmstruct)
                  CALL cp_fm_struct_release(fmstruct)
               END DO
            END IF
         END IF ! should_rebuild
      END IF ! prepare_subset

      CALL timestop(handle)
   END SUBROUTINE mpools_rebuild_fm_pools

! **************************************************************************************************

END MODULE qs_matrix_pools
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief function that build the kpoints section of the input
!> \par History
!>      init [07.2014]
!> \author JGH
! **************************************************************************************************
MODULE input_cp2k_kpoints
   USE bibliography,                    ONLY: MacDonald1978,&
                                              Monkhorst1976
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_create,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_kpoints'

   INTEGER, PARAMETER                      :: use_real_wfn = 101, &
                                              use_complex_wfn = 100

   PUBLIC :: create_kpoints_section
   PUBLIC :: use_real_wfn, use_complex_wfn

CONTAINS

! **************************************************************************************************
!> \brief Creates the Kpoints section
!>  SECTION: &kpoint... &end
!>      SCHEME [None, Gamma, Monkhorst-Pack, MacDonald, General]
!>        { nx ny nz }
!>        { nx ny nz  sx sy sz }
!>      KPOINT           x1 y1 z1 w1
!>      SYMMETRY         [on, off]
!>      WAVEFUNCTION     [real, complex]
!>      FULL_GRID        [on, off]
!>      VERBOSE          [on, off]
!>      EPS_GEO          value
!>      PARALLEL_GROUP_SIZE   [-1,0,n]
!>
!> \param section the section to create
!> \author JGH
! **************************************************************************************************
   SUBROUTINE create_kpoints_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_kpoints_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_kpoints.F",65)
      CALL section_create(section, name="KPOINTS", &
                          description="Sets up the kpoints.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="SCHEME", &
                          description="Kpoint scheme to be used.<br>"// &
                          "Available options: NONE, GAMMA, MONKHORST-PACK, MACDONALD, and GENERAL.<br>"// &
                          "For MONKHORST-PACK and MACDONALD the number of k points in all "// &
                          "3 dimensions has to be supplied along with the keyword. "// &
                          "E.g. <pre>MONKHORST-PACK 12 12 8.</pre>", &
                          usage="SCHEME {KPMETHOD} {integer} {integer} ..", &
                          citations=(/Monkhorst1976, MacDonald1978/), &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KPOINT", &
                          description="Specify kpoint coordinates and weight. ", &
                          usage="KPOINT x  y  z  w", repeats=.TRUE., &
                          n_var=4, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SYMMETRY", &
                          description="Use symmetry to reduce the number of kpoints.", &
                          usage="SYMMETRY <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FULL_GRID", &
                          description="Use full non-reduced kpoint grid.", &
                          usage="FULL_GRID <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VERBOSE", &
                          description="Verbose output information.", &
                          usage="VERBOSE <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_GEO", &
                          description="Accuracy in symmetry determination.", &
                          usage="EPS_GEO <real>", &
                          default_r_val=1.0e-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARALLEL_GROUP_SIZE", &
                          description="Number of processors to be used for a single kpoint. "// &
                          " This number must divide the total number of processes."// &
                          " The number of groups must divide the total number of kpoints."// &
                          " Value=-1 (smallest possible number of processes per group, satisfying the constraints)."// &
                          " Value=0 (all processes)."// &
                          " Value=n (exactly n processes).", &
                          usage="PARALLEL_GROUP_SIZE <integer>", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WAVEFUNCTIONS", &
                          description="Use real/complex wavefunctions if possible.", &
                          usage="WAVEFUNCTIONS REAL", &
                          default_i_val=use_complex_wfn, &
                          enum_c_vals=s2a("REAL", "COMPLEX"), &
                          enum_desc=s2a("Use real wavefunctions (if possible by kpoints specified).", &
                                        "Use complex wavefunctions (default)."), &
                          enum_i_vals=(/use_real_wfn, use_complex_wfn/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_kpoints_section

END MODULE input_cp2k_kpoints
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types and basic routines needed for a kpoint calculation
!> \par History
!>       2014.07 created [JGH]
!>       2014.11 unified k-point and gamma-point code [Ole Schuett]
!> \author JGH
! **************************************************************************************************
MODULE kpoint_types
   USE cp_blacs_env,                    ONLY: cp_blacs_env_release,&
                                              cp_blacs_env_type
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_env,                     ONLY: cp_cart_release,&
                                              cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_cart_type,&
                                              cp_para_env_type
   USE input_cp2k_kpoints,              ONLY: use_complex_wfn,&
                                              use_real_wfn
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE qs_matrix_pools,                 ONLY: mpools_release,&
                                              qs_matrix_pools_type
   USE qs_mo_types,                     ONLY: deallocate_mo_set,&
                                              mo_set_p_type
   USE qs_neighbor_list_types,          ONLY: neighbor_list_set_p_type
   USE string_utilities,                ONLY: uppercase
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'kpoint_types'

   PUBLIC :: kpoint_type
   PUBLIC :: kpoint_create, kpoint_release, get_kpoint_info
   PUBLIC :: read_kpoint_section, write_kpoint_info
   PUBLIC :: kpoint_env_type, kpoint_env_p_type
   PUBLIC :: kpoint_env_create

! **************************************************************************************************
!> \brief Keeps information about a specific k-point
!> \param nkpoint   the kpoint index
!> \param wkp       weight of this kpoint
!> \param xkp       kpoint coordinates
!> \param is_local  if this kpoint is calculated on a single thread
!> \param mos       associated MOs (r/i,spin)
!> \param pmat      associated density matrix (r/i,spin)
!> \param wmat      associated energy weighted density matrix (r/i,spin)
!> \author JGH
! **************************************************************************************************
   TYPE kpoint_env_type
      INTEGER                                           :: nkpoint
      REAL(KIND=dp)                                     :: wkp
      REAL(KIND=dp), DIMENSION(3)                       :: xkp
      LOGICAL                                           :: is_local
      TYPE(mo_set_p_type), DIMENSION(:, :), POINTER      :: mos
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: pmat
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER       :: wmat
   END TYPE kpoint_env_type

   TYPE kpoint_env_p_type
      TYPE(kpoint_env_type), POINTER                    :: kpoint_env
   END TYPE kpoint_env_p_type

! **************************************************************************************************
!> \brief Contains information about kpoints
!> \par History
!>       2014.07 created [JGH]
!> \param kp_scheme           [input] Type of kpoint grid
!> \param nkp_grid            [input] Grid points
!> \param kp_shift            [input] Shift of the grid
!> \param use_real_wfn        [input] real/complex wfn
!> \param symmetry            [input] use symmetry (atoms) to reduce kpoints
!> \param full_grid           [input] don't reduce kpoints at all
!> \param verbose             [input] more output information
!> \param eps_geo             [input] accuracy of atom symmetry detection
!> \param parallel_group_size [input] kpoint group size
!> \param nkp     number of kpoints
!> \param xkp     kpoint coordinates
!> \param wkp     kpoint weights
!> \param para_env 'global' parallel environment
!> \param cart 2d distribution of the processors for the kpoints,
!>        a column work together on the same kpoint
!> \param para_env_full the global para env that contains all the kpoints,
!>        this is just the cart as para_env
!> \param para_env_kp parallel environment of the kpoint calculation
!> \param para_env_inter_kp parallel environment between kpoints
!> \param iogrp  this kpoint group has the IO processor
!> \param nkp_groups   number of kpoint groups
!> \param kp_dist      kpoints distribution on groups
!> \param kp_range     kpoints distribution for local processor
!> \param blacs_env    BLACS env for the kpoint group
!> \param opmats       Operator matrices
!> \param kp_env       Information for each kpoint
!> \param mpools       FM matrix pools for kpoint groups
!> \author JGH
! **************************************************************************************************
   TYPE kpoint_type
      CHARACTER(LEN=default_string_length)   :: kp_scheme
      INTEGER, DIMENSION(3)                   :: nkp_grid
      REAL(KIND=dp), DIMENSION(3)             :: kp_shift
      LOGICAL                                 :: use_real_wfn
      LOGICAL                                 :: symmetry
      LOGICAL                                 :: full_grid
      LOGICAL                                 :: verbose
      REAL(KIND=dp)                           :: eps_geo
      INTEGER                                 :: parallel_group_size
      INTEGER                                 :: nkp
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: xkp => Null()
      REAL(KIND=dp), DIMENSION(:), POINTER    :: wkp => Null()
      ! parallel environment
      TYPE(cp_para_env_type), POINTER         :: para_env => Null()
      TYPE(cp_blacs_env_type), POINTER        :: blacs_env_all => Null()
      TYPE(cp_para_cart_type), POINTER        :: cart => Null()
      TYPE(cp_para_env_type), POINTER         :: para_env_full => Null(), &
                                                 para_env_kp => Null(), &
                                                 para_env_inter_kp => Null()
      LOGICAL                                 :: iogrp
      INTEGER                                 :: nkp_groups
      INTEGER, DIMENSION(:, :), POINTER        :: kp_dist => Null()
      INTEGER, DIMENSION(2)                   :: kp_range
      TYPE(cp_blacs_env_type), POINTER        :: blacs_env => Null()
      INTEGER, DIMENSION(:, :, :), POINTER      :: cell_to_index => Null()
      TYPE(neighbor_list_set_p_type), &
         DIMENSION(:), POINTER                 :: sab_nl => Null()
      ! environment
      TYPE(kpoint_env_p_type), DIMENSION(:), &
         POINTER                               :: kp_env => Null()
      TYPE(qs_matrix_pools_type), POINTER     :: mpools => Null()
   END TYPE kpoint_type

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief Create a kpoint environment
!> \param kpoint  All the kpoint information
!> \author JGH
! **************************************************************************************************
   SUBROUTINE kpoint_create(kpoint)
      TYPE(kpoint_type), POINTER                         :: kpoint

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kpoint_create', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(kpoint)))CALL cp__a("kpoint_types.F",159)

      ALLOCATE (kpoint)

      kpoint%kp_scheme = ""
      kpoint%nkp_grid = 0
      kpoint%kp_shift = 0.0_dp
      kpoint%symmetry = .FALSE.
      kpoint%verbose = .FALSE.
      kpoint%full_grid = .FALSE.
      kpoint%use_real_wfn = .FALSE.
      kpoint%eps_geo = 1.0e-6_dp
      kpoint%parallel_group_size = -1

      kpoint%nkp = 0

      NULLIFY (kpoint%xkp, kpoint%wkp)
      NULLIFY (kpoint%kp_dist)

      NULLIFY (kpoint%para_env)
      NULLIFY (kpoint%blacs_env_all)
      NULLIFY (kpoint%cart)
      NULLIFY (kpoint%para_env_full, kpoint%para_env_kp, kpoint%para_env_inter_kp)
      NULLIFY (kpoint%blacs_env)
      kpoint%nkp_groups = 0
      kpoint%iogrp = .FALSE.
      kpoint%kp_range = 0

      NULLIFY (kpoint%kp_env)
      NULLIFY (kpoint%mpools)

      ALLOCATE (kpoint%cell_to_index(0:0, 0:0, 0:0))
      kpoint%cell_to_index(:, :, :) = 1
   END SUBROUTINE kpoint_create

! **************************************************************************************************
!> \brief  Release a kpoint environment, deallocate all data
!> \param kpoint  The kpoint environment
!> \author JGH
! **************************************************************************************************
   SUBROUTINE kpoint_release(kpoint)
      TYPE(kpoint_type), POINTER                         :: kpoint

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kpoint_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: ik

      IF (ASSOCIATED(kpoint)) THEN

         IF (ASSOCIATED(kpoint%xkp)) THEN
            DEALLOCATE (kpoint%xkp)
         END IF
         IF (ASSOCIATED(kpoint%wkp)) THEN
            DEALLOCATE (kpoint%wkp)
         END IF
         IF (ASSOCIATED(kpoint%kp_dist)) THEN
            DEALLOCATE (kpoint%kp_dist)
         ENDIF

         CALL mpools_release(kpoint%mpools)
         CALL cp_blacs_env_release(kpoint%blacs_env)
         CALL cp_blacs_env_release(kpoint%blacs_env_all)

         CALL cp_cart_release(kpoint%cart)
         CALL cp_para_env_release(kpoint%para_env)
         CALL cp_para_env_release(kpoint%para_env_full)
         CALL cp_para_env_release(kpoint%para_env_kp)
         CALL cp_para_env_release(kpoint%para_env_inter_kp)

         IF (ASSOCIATED(kpoint%cell_to_index)) DEALLOCATE (kpoint%cell_to_index)

         IF (ASSOCIATED(kpoint%kp_env)) THEN
            DO ik = 1, SIZE(kpoint%kp_env)
               CALL kpoint_env_release(kpoint%kp_env(ik)%kpoint_env)
            END DO
            DEALLOCATE (kpoint%kp_env)
         END IF

         DEALLOCATE (kpoint)

      END IF

   END SUBROUTINE kpoint_release

! **************************************************************************************************
!> \brief Retrieve information from a kpoint environment
!> \param kpoint        The kpoint environment
!> \param kp_scheme     Type of kpoint grid
!> \param nkp_grid      Grid points
!> \param kp_shift      Shift of the grid
!> \param symmetry      use symmetry (atoms) to reduce kpoints
!> \param verbose       more output information
!> \param full_grid     don't reduce kpoints at all
!> \param use_real_wfn  real/complex wfn
!> \param eps_geo       accuracy of atom symmetry detection
!> \param parallel_group_size kpoint group size
!> \param kp_range      kpoints distribution for local processor
!> \param nkp           number of kpoints
!> \param xkp           kpoint coordinates
!> \param wkp           kpoint weights
!> \param para_env      'global' parallel environment
!> \param blacs_env_all BLACS env for the total environment
!> \param cart          2d distribution of the processors for the kpoints,
!>        a column work together on the same kpoint
!> \param para_env_full the global para env that contains all the kpoints,
!>        this is just the cart as para_env
!> \param para_env_kp   parallel environment of the kpoint calculation
!> \param para_env_inter_kp   parallel environment between kpoints
!> \param blacs_env     BLACS env for the kpoint group
!> \param kp_env        Information for each kpoint
!> \param mpools        FM matrix pools for kpoint groups
!> \param iogrp         this kpoint group has the IO processor
!> \param nkp_groups    number of kpoint groups
!> \param kp_dist       kpoints distribution on groups
!> \param cell_to_index given a cell triple, returns the real space index
!> \param sab_nl        neighbourlist that defines real space matrices
!> \author JGH
! **************************************************************************************************
   SUBROUTINE get_kpoint_info(kpoint, kp_scheme, nkp_grid, kp_shift, symmetry, verbose, &
                              full_grid, use_real_wfn, eps_geo, parallel_group_size, kp_range, nkp, xkp, wkp, &
                              para_env, blacs_env_all, cart, para_env_full, para_env_kp, para_env_inter_kp, blacs_env, &
                              kp_env, mpools, iogrp, nkp_groups, kp_dist, cell_to_index, sab_nl)
      TYPE(kpoint_type), POINTER                         :: kpoint
      CHARACTER(LEN=*), OPTIONAL                         :: kp_scheme
      INTEGER, DIMENSION(3), OPTIONAL                    :: nkp_grid
      REAL(KIND=dp), DIMENSION(3), OPTIONAL              :: kp_shift
      LOGICAL, OPTIONAL                                  :: symmetry, verbose, full_grid, &
                                                            use_real_wfn
      REAL(KIND=dp), OPTIONAL                            :: eps_geo
      INTEGER, OPTIONAL                                  :: parallel_group_size
      INTEGER, DIMENSION(2), OPTIONAL                    :: kp_range
      INTEGER, OPTIONAL                                  :: nkp
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: xkp
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: wkp
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(cp_blacs_env_type), OPTIONAL, POINTER         :: blacs_env_all
      TYPE(cp_para_cart_type), OPTIONAL, POINTER         :: cart
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env_full, para_env_kp, &
                                                            para_env_inter_kp
      TYPE(cp_blacs_env_type), OPTIONAL, POINTER         :: blacs_env
      TYPE(kpoint_env_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: kp_env
      TYPE(qs_matrix_pools_type), OPTIONAL, POINTER      :: mpools
      LOGICAL, OPTIONAL                                  :: iogrp
      INTEGER, OPTIONAL                                  :: nkp_groups
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: kp_dist
      INTEGER, DIMENSION(:, :, :), OPTIONAL, POINTER     :: cell_to_index
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: sab_nl

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_kpoint_info', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(kpoint)))CALL cp__a("kpoint_types.F",312)

      IF (PRESENT(kp_scheme)) kp_scheme = kpoint%kp_scheme
      IF (PRESENT(nkp_grid)) nkp_grid = kpoint%nkp_grid
      IF (PRESENT(kp_shift)) kp_shift = kpoint%kp_shift
      IF (PRESENT(symmetry)) symmetry = kpoint%symmetry
      IF (PRESENT(verbose)) verbose = kpoint%verbose
      IF (PRESENT(full_grid)) full_grid = kpoint%full_grid
      IF (PRESENT(use_real_wfn)) use_real_wfn = kpoint%use_real_wfn
      IF (PRESENT(eps_geo)) eps_geo = kpoint%eps_geo
      IF (PRESENT(parallel_group_size)) parallel_group_size = kpoint%parallel_group_size

      IF (PRESENT(nkp)) nkp = kpoint%nkp
      IF (PRESENT(wkp)) wkp => kpoint%wkp
      IF (PRESENT(xkp)) xkp => kpoint%xkp

      IF (PRESENT(para_env)) para_env => kpoint%para_env
      IF (PRESENT(cart)) cart => kpoint%cart
      IF (PRESENT(para_env_full)) para_env_full => kpoint%para_env_full
      IF (PRESENT(para_env_kp)) para_env_kp => kpoint%para_env_kp
      IF (PRESENT(para_env_inter_kp)) para_env_inter_kp => kpoint%para_env_inter_kp
      IF (PRESENT(blacs_env_all)) blacs_env_all => kpoint%blacs_env_all
      IF (PRESENT(blacs_env)) blacs_env => kpoint%blacs_env

      IF (PRESENT(iogrp)) iogrp = kpoint%iogrp
      IF (PRESENT(kp_range)) kp_range = kpoint%kp_range
      IF (PRESENT(nkp_groups)) nkp_groups = kpoint%nkp_groups
      IF (PRESENT(kp_dist)) kp_dist => kpoint%kp_dist

      IF (PRESENT(kp_env)) kp_env => kpoint%kp_env
      IF (PRESENT(mpools)) mpools => kpoint%mpools

      IF (PRESENT(cell_to_index)) cell_to_index => kpoint%cell_to_index
      IF (PRESENT(sab_nl)) sab_nl => kpoint%sab_nl

   END SUBROUTINE get_kpoint_info

! **************************************************************************************************
!> \brief Set information in a kpoint environment
!> \param kpoint        The kpoint environment
!> \param kp_scheme     Type of kpoint grid
!> \param nkp_grid      Grid points
!> \param kp_shift      Shift of the grid
!> \param symmetry      use symmetry (atoms) to reduce kpoints
!> \param verbose       more output information
!> \param full_grid     don't reduce kpoints at all
!> \param use_real_wfn  real/complex wfn
!> \param eps_geo       accuracy of atom symmetry detection
!> \param parallel_group_size kpoint group size
!> \param kp_range      kpoints distribution for local processor
!> \param nkp           number of kpoints
!> \param xkp           kpoint coordinates
!> \param wkp           kpoint weights
!> \param para_env      'global' parallel environment
!> \param blacs_env_all BLACS env for the total environment
!> \param cart          2d distribution of the processors for the kpoints,
!>        a column work together on the same kpoint
!> \param para_env_full the global para env that contains all the kpoints,
!>        this is just the cart as para_env
!> \param para_env_kp   parallel environment of the kpoint calculation
!> \param para_env_inter_kp   parallel environment between kpoints
!> \param blacs_env     BLACS env for the kpoint group
!> \param kp_env        Information for each kpoint
!> \param mpools        FM matrix pools for kpoint groups
!> \param iogrp         this kpoint group has the IO processor
!> \param nkp_groups    number of kpoint groups
!> \param kp_dist       kpoints distribution on groups
!> \param cell_to_index given a cell triple, returns the real space index
!> \param sab_nl        neighbourlist that defines real space matrices
!> \author JGH
! **************************************************************************************************
   SUBROUTINE set_kpoint_info(kpoint, kp_scheme, nkp_grid, kp_shift, symmetry, verbose, &
                              full_grid, use_real_wfn, eps_geo, parallel_group_size, kp_range, nkp, xkp, wkp, &
                              para_env, blacs_env_all, cart, para_env_full, para_env_kp, para_env_inter_kp, blacs_env, &
                              kp_env, mpools, iogrp, nkp_groups, kp_dist, cell_to_index, sab_nl)
      TYPE(kpoint_type), POINTER                         :: kpoint
      CHARACTER(LEN=*), OPTIONAL                         :: kp_scheme
      INTEGER, DIMENSION(3), OPTIONAL                    :: nkp_grid
      REAL(KIND=dp), DIMENSION(3), OPTIONAL              :: kp_shift
      LOGICAL, OPTIONAL                                  :: symmetry, verbose, full_grid, &
                                                            use_real_wfn
      REAL(KIND=dp), OPTIONAL                            :: eps_geo
      INTEGER, OPTIONAL                                  :: parallel_group_size
      INTEGER, DIMENSION(2), OPTIONAL                    :: kp_range
      INTEGER, OPTIONAL                                  :: nkp
      REAL(KIND=dp), DIMENSION(:, :), OPTIONAL, POINTER  :: xkp
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: wkp
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(cp_blacs_env_type), OPTIONAL, POINTER         :: blacs_env_all
      TYPE(cp_para_cart_type), OPTIONAL, POINTER         :: cart
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env_full, para_env_kp, &
                                                            para_env_inter_kp
      TYPE(cp_blacs_env_type), OPTIONAL, POINTER         :: blacs_env
      TYPE(kpoint_env_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: kp_env
      TYPE(qs_matrix_pools_type), OPTIONAL, POINTER      :: mpools
      LOGICAL, OPTIONAL                                  :: iogrp
      INTEGER, OPTIONAL                                  :: nkp_groups
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: kp_dist
      INTEGER, DIMENSION(:, :, :), OPTIONAL, POINTER     :: cell_to_index
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: sab_nl

      CHARACTER(LEN=*), PARAMETER :: routineN = 'set_kpoint_info', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(kpoint)))CALL cp__a("kpoint_types.F",418)

      IF (PRESENT(kp_scheme)) kpoint%kp_scheme = kp_scheme
      IF (PRESENT(nkp_grid)) kpoint%nkp_grid = nkp_grid
      IF (PRESENT(kp_shift)) kpoint%kp_shift = kp_shift
      IF (PRESENT(symmetry)) kpoint%symmetry = symmetry
      IF (PRESENT(verbose)) kpoint%verbose = verbose
      IF (PRESENT(full_grid)) kpoint%full_grid = full_grid
      IF (PRESENT(use_real_wfn)) kpoint%use_real_wfn = use_real_wfn
      IF (PRESENT(eps_geo)) kpoint%eps_geo = eps_geo
      IF (PRESENT(parallel_group_size)) kpoint%parallel_group_size = parallel_group_size

      IF (PRESENT(nkp)) kpoint%nkp = nkp
      IF (PRESENT(wkp)) kpoint%wkp => wkp
      IF (PRESENT(xkp)) kpoint%xkp => xkp

      IF (PRESENT(para_env)) kpoint%para_env => para_env
      IF (PRESENT(cart)) kpoint%cart => cart
      IF (PRESENT(para_env_full)) kpoint%para_env_full => para_env_full
      IF (PRESENT(para_env_kp)) kpoint%para_env_kp => para_env_kp
      IF (PRESENT(para_env_inter_kp)) kpoint%para_env_inter_kp => para_env_inter_kp
      IF (PRESENT(blacs_env_all)) kpoint%blacs_env_all => blacs_env_all
      IF (PRESENT(blacs_env)) kpoint%blacs_env => blacs_env

      IF (PRESENT(iogrp)) kpoint%iogrp = iogrp
      IF (PRESENT(kp_range)) kpoint%kp_range = kp_range
      IF (PRESENT(nkp_groups)) kpoint%nkp_groups = nkp_groups
      IF (PRESENT(kp_dist)) kpoint%kp_dist => kp_dist

      IF (PRESENT(kp_env)) kpoint%kp_env => kp_env
      IF (PRESENT(mpools)) kpoint%mpools => mpools
      IF (PRESENT(sab_nl)) kpoint%sab_nl => sab_nl

      IF (PRESENT(cell_to_index)) THEN
         IF (ASSOCIATED(kpoint%cell_to_index)) DEALLOCATE (kpoint%cell_to_index)
         kpoint%cell_to_index => cell_to_index
      ENDIF

   END SUBROUTINE set_kpoint_info

! **************************************************************************************************
!> \brief Read the kpoint input section
!> \param kpoint  The kpoint environment
!> \param kpoint_section The input section
!> \author JGH
! **************************************************************************************************
   SUBROUTINE read_kpoint_section(kpoint, kpoint_section)
      TYPE(kpoint_type), POINTER                         :: kpoint
      TYPE(section_vals_type), POINTER                   :: kpoint_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_kpoint_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length), &
         DIMENSION(:), POINTER                           :: tmpstringlist
      INTEGER                                            :: i, n_rep, nval, wfntype
      LOGICAL                                            :: available
      REAL(KIND=dp), DIMENSION(:), POINTER               :: reallist

      IF(.NOT.(ASSOCIATED(kpoint)))CALL cp__a("kpoint_types.F",477)

      CALL section_vals_get(kpoint_section, explicit=available)

      IF (available) THEN
         CALL section_vals_val_get(kpoint_section, "SCHEME", c_vals=tmpstringlist)
         nval = SIZE(tmpstringlist)
         IF(.NOT.(nval >= 1))CALL cp__a("kpoint_types.F",484)
         kpoint%kp_scheme = tmpstringlist(1)
         CALL uppercase(kpoint%kp_scheme)

         ! SCHEME [None, Gamma, Monkhorst-Pack, MacDonald, General]
         SELECT CASE (kpoint%kp_scheme)
         CASE ("NONE")
            ! do nothing
         CASE ("GAMMA")
            ! do nothing
         CASE ("MONKHORST-PACK")
            IF(.NOT.(nval >= 4))CALL cp__a("kpoint_types.F",495)
            DO i = 2, 4
               READ (tmpstringlist(i), *) kpoint%nkp_grid(i-1)
            END DO
         CASE ("MACDONALD")
            IF(.NOT.(nval >= 7))CALL cp__a("kpoint_types.F",500)
            DO i = 2, 4
               READ (tmpstringlist(i), *) kpoint%nkp_grid(i-1)
            END DO
            DO i = 5, 7
               READ (tmpstringlist(i), *) kpoint%kp_shift(i-4)
            END DO
         CASE ("GENERAL")
            CALL section_vals_val_get(kpoint_section, "KPOINT", n_rep_val=n_rep)
            kpoint%nkp = n_rep
            ALLOCATE (kpoint%xkp(3, n_rep), kpoint%wkp(n_rep))
            DO i = 1, n_rep
               CALL section_vals_val_get(kpoint_section, "KPOINT", i_rep_val=i, &
                                         r_vals=reallist)
               nval = SIZE(reallist)
               IF(.NOT.(nval >= 4))CALL cp__a("kpoint_types.F",515)
               kpoint%xkp(1:3, i) = reallist(1:3)
               kpoint%wkp(i) = reallist(4)
            END DO
         CASE DEFAULT
            CALL cp__b("kpoint_types.F",520,"")
         END SELECT

         CALL section_vals_val_get(kpoint_section, "SYMMETRY", l_val=kpoint%symmetry)
         CALL section_vals_val_get(kpoint_section, "WAVEFUNCTIONS", i_val=wfntype)
         CALL section_vals_val_get(kpoint_section, "VERBOSE", l_val=kpoint%verbose)
         CALL section_vals_val_get(kpoint_section, "FULL_GRID", l_val=kpoint%full_grid)
         CALL section_vals_val_get(kpoint_section, "EPS_GEO", r_val=kpoint%eps_geo)
         CALL section_vals_val_get(kpoint_section, "PARALLEL_GROUP_SIZE", &
                                   i_val=kpoint%parallel_group_size)
         SELECT CASE (wfntype)
         CASE (use_real_wfn)
            kpoint%use_real_wfn = .TRUE.
         CASE (use_complex_wfn)
            kpoint%use_real_wfn = .FALSE.
         CASE DEFAULT
            CALL cp__b("kpoint_types.F",536,"")
         END SELECT

      ELSE
         kpoint%kp_scheme = "NONE"
      END IF

   END SUBROUTINE read_kpoint_section

! **************************************************************************************************
!> \brief Write information on the kpoints to output
!> \param kpoint  The kpoint environment
!> \param dft_section  DFT section information
!> \author JGH
! **************************************************************************************************
   SUBROUTINE write_kpoint_info(kpoint, dft_section)
      TYPE(kpoint_type), POINTER                         :: kpoint
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(len=*), PARAMETER :: routineN = 'write_kpoint_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, punit
      TYPE(cp_logger_type), POINTER                      :: logger

      IF(.NOT.(ASSOCIATED(kpoint)))CALL cp__a("kpoint_types.F",561)

      NULLIFY (logger)
      logger => cp_get_default_logger()

      punit = cp_print_key_unit_nr(logger, dft_section, "PRINT%KPOINTS", extension=".Log")
      IF (punit > 0) THEN

         IF (kpoint%kp_scheme /= "NONE") THEN
            WRITE (punit, '(/," ",79("*"),/,T37,A,/," ",79("*"))') "Kpoints"
         END IF
         SELECT CASE (kpoint%kp_scheme)
         CASE ("NONE")
            ! be silent
         CASE ("GAMMA")
            WRITE (punit, '(A,T57,A)') ' BRILLOUIN|', ' Gamma-point calculation'
         CASE ("MONKHORST-PACK")
            WRITE (punit, '(A,T61,A20)') ' BRILLOUIN| K-point scheme ', '      Monkhorst-Pack'
            WRITE (punit, '(A,T66,3I5)') ' BRILLOUIN| K-Point grid', kpoint%nkp_grid
            WRITE (punit, '(A,T66,G15.6)') ' BRILLOUIN| Accuracy in Symmetry determination', kpoint%eps_geo
         CASE ("MACDONALD")
            WRITE (punit, '(A,T71,A10)') ' BRILLOUIN| K-point scheme ', ' MacDonald'
            WRITE (punit, '(A,T66,3I5)') ' BRILLOUIN| K-Point grid', kpoint%nkp_grid
            WRITE (punit, '(A,T51,3F10.4)') ' BRILLOUIN| K-Point shift', kpoint%kp_shift
            WRITE (punit, '(A,T66,G15.6)') ' BRILLOUIN| Accuracy in Symmetry determination', kpoint%eps_geo
         CASE ("GENERAL")
            WRITE (punit, '(A,T71,A10)') ' BRILLOUIN| K-point scheme ', '   General'
         CASE DEFAULT
            CALL cp__b("kpoint_types.F",589,"")
         END SELECT
         IF (kpoint%kp_scheme /= "NONE") THEN
            IF (kpoint%symmetry) THEN
               WRITE (punit, '(A,T76,A)') ' BRILLOUIN| K-Point point group symmetrization', '   ON'
            ELSE
               WRITE (punit, '(A,T76,A)') ' BRILLOUIN| K-Point point group symmetrization', '  OFF'
            END IF
            IF (kpoint%use_real_wfn) THEN
               WRITE (punit, '(A,T76,A)') ' BRILLOUIN| Wavefunction type', ' REAL'
            ELSE
               WRITE (punit, '(A,T73,A)') ' BRILLOUIN| Wavefunction type', ' COMPLEX'
            END IF
            IF (kpoint%full_grid) THEN
               WRITE (punit, '(A,T76,A)') ' BRILLOUIN| Use full k-point grid     '
            END IF
            IF (kpoint%kp_scheme /= "GAMMA") THEN
               WRITE (punit, '(A,T71,I10)') ' BRILLOUIN| List of Kpoints', kpoint%nkp
               WRITE (punit, '(A,T30,A,T48,A,T63,A,T78,A)') &
                  ' BRILLOUIN| Number ', 'Weight', 'X', 'Y', 'Z'
               DO i = 1, kpoint%nkp
                  WRITE (punit, '(A,I5,3X,4F15.5)') ' BRILLOUIN| ', i, kpoint%wkp(i), &
                     kpoint%xkp(1, i), kpoint%xkp(2, i), kpoint%xkp(3, i)
               END DO
            END IF
            WRITE (punit, '(" ",79("*"))')
         END IF

      END IF
      CALL cp_print_key_finished_output(punit, logger, dft_section, "PRINT%KPOINTS")

   END SUBROUTINE write_kpoint_info

! **************************************************************************************************
!> \brief Create a single kpoint environment
!> \param kp_env  Single kpoint environment
!> \author JGH
! **************************************************************************************************
   SUBROUTINE kpoint_env_create(kp_env)
      TYPE(kpoint_env_type), POINTER                     :: kp_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kpoint_env_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(kp_env)))CALL cp__a("kpoint_types.F",633)

      ALLOCATE (kp_env)

      kp_env%nkpoint = 0
      kp_env%wkp = 0.0_dp
      kp_env%xkp = 0.0_dp
      kp_env%is_local = .FALSE.

      NULLIFY (kp_env%mos)
      NULLIFY (kp_env%pmat)
      NULLIFY (kp_env%wmat)

   END SUBROUTINE kpoint_env_create

! **************************************************************************************************
!> \brief Release a single kpoint environment
!> \param kp_env  Single kpoint environment
!> \author JGH
! **************************************************************************************************
   SUBROUTINE kpoint_env_release(kp_env)
      TYPE(kpoint_env_type), POINTER                     :: kp_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'kpoint_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ic, is

      IF (ASSOCIATED(kp_env)) THEN

         IF (ASSOCIATED(kp_env%mos)) THEN
            DO is = 1, SIZE(kp_env%mos, 2)
               DO ic = 1, SIZE(kp_env%mos, 1)
                  CALL deallocate_mo_set(kp_env%mos(ic, is)%mo_set)
               END DO
            END DO
            DEALLOCATE (kp_env%mos)
         END IF

         IF (ASSOCIATED(kp_env%pmat)) THEN
            DO is = 1, SIZE(kp_env%pmat, 2)
               DO ic = 1, SIZE(kp_env%pmat, 1)
                  CALL cp_fm_release(kp_env%pmat(ic, is)%matrix)
               END DO
            END DO
            DEALLOCATE (kp_env%pmat)
         END IF

         IF (ASSOCIATED(kp_env%wmat)) THEN
            DO is = 1, SIZE(kp_env%wmat, 2)
               DO ic = 1, SIZE(kp_env%wmat, 1)
                  CALL cp_fm_release(kp_env%wmat(ic, is)%matrix)
               END DO
            END DO
            DEALLOCATE (kp_env%wmat)
         END IF

         DEALLOCATE (kp_env)

      END IF

   END SUBROUTINE kpoint_env_release

! **************************************************************************************************
!> \brief Get information from a single kpoint environment
!> \param kpoint_env Single kpoint environment
!> \param nkpoint    Index of kpoint
!> \param wkp        Weight of kpoint
!> \param xkp        Coordinates of kpoint
!> \param is_local   Is this kpoint local (single cpu group)
!> \param mos        MOs of this kpoint
!> \author JGH
! **************************************************************************************************
   SUBROUTINE get_kpoint_env(kpoint_env, nkpoint, wkp, xkp, is_local, mos)
      TYPE(kpoint_env_type), POINTER                     :: kpoint_env
      INTEGER, OPTIONAL                                  :: nkpoint
      REAL(KIND=dp), OPTIONAL                            :: wkp
      REAL(KIND=dp), DIMENSION(3), OPTIONAL              :: xkp
      LOGICAL, OPTIONAL                                  :: is_local
      TYPE(mo_set_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: mos

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_kpoint_env', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(kpoint_env)))CALL cp__a("kpoint_types.F",717)

      IF (PRESENT(nkpoint)) nkpoint = kpoint_env%nkpoint
      IF (PRESENT(wkp)) wkp = kpoint_env%wkp
      IF (PRESENT(xkp)) xkp = kpoint_env%xkp
      IF (PRESENT(is_local)) is_local = kpoint_env%is_local
      IF (PRESENT(mos)) mos => kpoint_env%mos

   END SUBROUTINE get_kpoint_env

! **************************************************************************************************

END MODULE kpoint_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Datatype to translate between k-points (2d) and gamma-point (1d) code.
!> \note  In principle storing just the 2d pointer would be sufficient.
!>        However due to a bug in ifort with the deallocation of
!>        bounds-remapped pointers, we also have to store the original
!>        1d pointer used for allocation.
!>
!> \par History
!>      11.2014 created [Ole Schuett]
!> \author Ole Schuett
! **************************************************************************************************
MODULE kpoint_transitional
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   PUBLIC :: kpoint_transitional_type, kpoint_transitional_release
   PUBLIC :: get_1d_pointer, get_2d_pointer, set_1d_pointer, set_2d_pointer

   TYPE kpoint_transitional_type
      PRIVATE
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER          :: ptr_1d => Null()
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER        :: ptr_2d => Null()
      LOGICAL                                               :: set_as_1d = .FALSE.
   END TYPE kpoint_transitional_type

CONTAINS

! **************************************************************************************************
!> \brief Smart getter, raises an error when called during a k-point calculation
!> \param this ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION get_1d_pointer(this) RESULT(res)
      TYPE(kpoint_transitional_type)                     :: this
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: res

      IF (ASSOCIATED(this%ptr_1d)) THEN
         IF (SIZE(this%ptr_2d, 2) /= 1) &
            CALL cp__b("kpoint_transitional.F",49,"Method not implemented for k-points")
      ENDIF

      res => this%ptr_1d
   END FUNCTION get_1d_pointer

! **************************************************************************************************
!> \brief Simple getter, neede because of PRIVATE
!> \param this ...
!> \retval res ...
!> \author Ole Schuett
! **************************************************************************************************
   FUNCTION get_2d_pointer(this) RESULT(res)
      TYPE(kpoint_transitional_type)                     :: this
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: res

      res => this%ptr_2d
   END FUNCTION get_2d_pointer

! **************************************************************************************************
!> \brief Assigns a 1D pointer
!> \param this ...
!> \param ptr_1d ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE set_1d_pointer(this, ptr_1d)
      TYPE(kpoint_transitional_type)                     :: this
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER       :: ptr_1d

      INTEGER                                            :: n

      IF (ASSOCIATED(ptr_1d)) THEN
         n = SIZE(ptr_1d)
         this%ptr_1d => ptr_1d
         this%ptr_2d(1:n, 1:1) => ptr_1d
         this%set_as_1d = .TRUE.
      ELSE
         this%ptr_1d => Null()
         this%ptr_2d => Null()
      ENDIF
   END SUBROUTINE set_1d_pointer

! **************************************************************************************************
!> \brief Assigns a 2D pointer
!> \param this ...
!> \param ptr_2d ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE set_2d_pointer(this, ptr_2d)
      TYPE(kpoint_transitional_type)                     :: this
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER    :: ptr_2d

      IF (ASSOCIATED(ptr_2d)) THEN
         this%ptr_1d => ptr_2d(:, 1)
         this%ptr_2d => ptr_2d
         this%set_as_1d = .FALSE.
      ELSE
         this%ptr_1d => Null()
         this%ptr_2d => Null()
      ENDIF
   END SUBROUTINE set_2d_pointer

! **************************************************************************************************
!> \brief Release the matrix set, using the right pointer
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE kpoint_transitional_release(this)
      TYPE(kpoint_transitional_type)                     :: this

      IF (ASSOCIATED(this%ptr_1d)) THEN
         IF (this%set_as_1d) THEN
            CALL cp_dbcsr_deallocate_matrix_set(this%ptr_1d)
         ELSE
            CALL cp_dbcsr_deallocate_matrix_set(this%ptr_2d)
         ENDIF
      ENDIF
      NULLIFY (this%ptr_1d, this%ptr_2d)
   END SUBROUTINE kpoint_transitional_release

END MODULE kpoint_transitional
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief superstucture that hold various representations of the density and
!>      keeps track of which ones are valid
!> \par History
!>      08.2002 created [fawzi]
!>      08.2014 kpoints [JGH]
!>      11.2014 make qs_rho_type PRIVATE [Ole Schuett]
!>      11.2014 unified k-point and gamma-point code [Ole Schuett]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qs_rho_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_p_type
   USE kinds,                           ONLY: dp
   USE kpoint_transitional,             ONLY: get_1d_pointer,&
                                              get_2d_pointer,&
                                              kpoint_transitional_release,&
                                              kpoint_transitional_type,&
                                              set_1d_pointer,&
                                              set_2d_pointer
   USE pw_types,                        ONLY: pw_p_type,&
                                              pw_release
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_rho_types'

   PUBLIC :: qs_rho_p_type, qs_rho_type
   PUBLIC :: qs_rho_retain, qs_rho_release, &
             qs_rho_get, qs_rho_set, qs_rho_clear, qs_rho_create

! **************************************************************************************************
!> \brief keeps the density in various representations, keeping track of
!>      which ones are valid.
!> \param most attributes are array with either lda or lsd_alpha,lsd_beta.
!> \param rho_ao the filtered rho in the localized atom basis (to have rho(r)
!>        the filtred matrix is enough, but rho(r,r') is lost).
!> \param rho_ao_kp the filtered rho in the localized atom basis (to have rho(r)
!>        the filtered matrix is enough, but rho(r,r') is lost).
!>        for kpoints, in real space index form
!> \param rho_r grids with rho in the real space
!> \param tau_r grids with the kinetic energy density in real space
!> \param rho_g grids with rho in the g space
!> \param tau_g grids with the kinetic energy density in g space
!> \param rho_g_valid , rho_r_valid, tau_r_valid, tau_g_valid: if the
!>        corresponding component is valid
!> \param ref_count the reference count, when it becomes 0 the type
!>        is deallocated.
!> \param rebuild_each how often a rebuild should be done by default
!> \param tot_rho_r the total charge in r space (valid only if rho_r is)
!> \note
!>      If pw_p_type would implement retain/release it would be nice to
!>      store also the core charge and the qs_charges in this structure...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE qs_rho_type
      PRIVATE
      TYPE(kpoint_transitional_type)                 :: rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER   :: rho_ao_im => Null()
      TYPE(pw_p_type), DIMENSION(:), POINTER         :: rho_g => Null(), &
                                                        rho_r => Null(), &
                                                        drho_g => Null(), &
                                                        drho_r => Null(), &
                                                        tau_g => Null(), &
                                                        tau_r => Null()
      ! Final rho_iter of last SCCS cycle (r-space)
      TYPE(pw_p_type), POINTER                       :: rho_r_sccs => Null()
      LOGICAL                                        :: rho_g_valid = .FALSE., &
                                                        rho_r_valid = .FALSE., &
                                                        drho_r_valid = .FALSE., &
                                                        drho_g_valid = .FALSE., &
                                                        tau_r_valid = .FALSE., &
                                                        tau_g_valid = .FALSE., &
                                                        soft_valid = .FALSE.
      INTEGER                                        :: ref_count = -1, &
                                                        id_nr = -1, &
                                                        rebuild_each = -1
      REAL(KIND=dp), DIMENSION(:), POINTER ::         tot_rho_r => Null(), &
                                              tot_rho_g => Null()
   END TYPE qs_rho_type

! **************************************************************************************************
   TYPE qs_rho_p_type
      TYPE(qs_rho_type), POINTER         :: rho
   END TYPE qs_rho_p_type

   INTEGER, PRIVATE, SAVE :: last_rho_id_nr = 0

CONTAINS

! **************************************************************************************************
!> \brief Allocates a new instance of rho.
!> \param rho ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE qs_rho_create(rho)
      TYPE(qs_rho_type), POINTER                         :: rho

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_rho_create', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(rho)) CALL cp__b("qs_rho_types.F",111,"rho already associated")

      ALLOCATE (rho)
      last_rho_id_nr = last_rho_id_nr+1
      rho%id_nr = last_rho_id_nr
      rho%rebuild_each = 5
      rho%ref_count = 1
   END SUBROUTINE qs_rho_create

! **************************************************************************************************
!> \brief retains a rho_struct by increasing the reference count by one
!>      (to be called when you want to keep a shared copy)
!> \param rho_struct the structure to retain
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_rho_retain(rho_struct)
      TYPE(qs_rho_type), POINTER                         :: rho_struct

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_rho_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(rho_struct)))CALL cp__a("qs_rho_types.F",133)
      IF(.NOT.(rho_struct%ref_count > 0))CALL cp__a("qs_rho_types.F",134)
      rho_struct%ref_count = rho_struct%ref_count+1
   END SUBROUTINE qs_rho_retain

! **************************************************************************************************
!> \brief releases a rho_struct by decreasing the reference count by one
!>      and deallocating if it reaches 0 (to be called when you don't want
!>      anymore a shared copy)
!> \param rho_struct the structure to retain
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_rho_release(rho_struct)
      TYPE(qs_rho_type), POINTER                         :: rho_struct

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_rho_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(rho_struct)) THEN
         IF(.NOT.(rho_struct%ref_count > 0))CALL cp__a("qs_rho_types.F",153)
         rho_struct%ref_count = rho_struct%ref_count-1
         IF (rho_struct%ref_count < 1) THEN
            CALL qs_rho_clear(rho_struct)
            DEALLOCATE (rho_struct)
         END IF
      END IF

      NULLIFY (rho_struct)

   END SUBROUTINE qs_rho_release

! **************************************************************************************************
!> \brief Deallocates all components, whithout deallocating rho_struct itself.
!> \param rho_struct ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE qs_rho_clear(rho_struct)
      TYPE(qs_rho_type), POINTER                         :: rho_struct

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_rho_clear', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(rho_struct%rho_r)) THEN
         DO i = 1, SIZE(rho_struct%rho_r)
            CALL pw_release(rho_struct%rho_r(i)%pw)
         END DO
         DEALLOCATE (rho_struct%rho_r)
      END IF
      IF (ASSOCIATED(rho_struct%drho_r)) THEN
         DO i = 1, SIZE(rho_struct%drho_r)
            CALL pw_release(rho_struct%drho_r(i)%pw)
         END DO
         DEALLOCATE (rho_struct%drho_r)
      END IF
      IF (ASSOCIATED(rho_struct%drho_g)) THEN
         DO i = 1, SIZE(rho_struct%drho_g)
            CALL pw_release(rho_struct%drho_g(i)%pw)
         END DO
         DEALLOCATE (rho_struct%drho_g)
      END IF
      IF (ASSOCIATED(rho_struct%tau_r)) THEN
         DO i = 1, SIZE(rho_struct%tau_r)
            CALL pw_release(rho_struct%tau_r(i)%pw)
         END DO
         DEALLOCATE (rho_struct%tau_r)
      END IF
      IF (ASSOCIATED(rho_struct%rho_g)) THEN
         DO i = 1, SIZE(rho_struct%rho_g)
            CALL pw_release(rho_struct%rho_g(i)%pw)
         END DO
         DEALLOCATE (rho_struct%rho_g)
      END IF
      IF (ASSOCIATED(rho_struct%tau_g)) THEN
         DO i = 1, SIZE(rho_struct%tau_g)
            CALL pw_release(rho_struct%tau_g(i)%pw)
         END DO
         DEALLOCATE (rho_struct%tau_g)
      END IF
      IF (ASSOCIATED(rho_struct%rho_r_sccs)) THEN
         CALL pw_release(rho_struct%rho_r_sccs%pw)
         DEALLOCATE (rho_struct%rho_r_sccs)
      END IF

      CALL kpoint_transitional_release(rho_struct%rho_ao)

      IF (ASSOCIATED(rho_struct%rho_ao_im)) &
         CALL cp_dbcsr_deallocate_matrix_set(rho_struct%rho_ao_im)
      IF (ASSOCIATED(rho_struct%tot_rho_r)) &
         DEALLOCATE (rho_struct%tot_rho_r)
      IF (ASSOCIATED(rho_struct%tot_rho_g)) &
         DEALLOCATE (rho_struct%tot_rho_g)

   END SUBROUTINE qs_rho_clear

! **************************************************************************************************
!> \brief returns info about the density described by this object.
!>      If some representation is not available an error is issued
!> \param rho_struct ...
!> \param rho_ao ...
!> \param rho_ao_im ...
!> \param rho_ao_kp ...
!> \param rho_r ...
!> \param drho_r ...
!> \param rho_g ...
!> \param drho_g ...
!> \param tau_r ...
!> \param tau_g ...
!> \param rho_r_valid ...
!> \param drho_r_valid ...
!> \param rho_g_valid ...
!> \param drho_g_valid ...
!> \param tau_r_valid ...
!> \param tau_g_valid ...
!> \param rebuild_each ...
!> \param tot_rho_r ...
!> \param tot_rho_g ...
!> \param rho_r_sccs ...
!> \param soft_valid ...
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_rho_get(rho_struct, rho_ao, rho_ao_im, rho_ao_kp, rho_r, drho_r, &
                         rho_g, drho_g, tau_r, tau_g, rho_r_valid, drho_r_valid, rho_g_valid, &
                         drho_g_valid, tau_r_valid, tau_g_valid, rebuild_each, tot_rho_r, tot_rho_g, &
                         rho_r_sccs, soft_valid)
      TYPE(qs_rho_type), POINTER                         :: rho_struct
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho_ao, rho_ao_im
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: rho_ao_kp
      TYPE(pw_p_type), DIMENSION(:), OPTIONAL, POINTER   :: rho_r, drho_r, rho_g, drho_g, tau_r, &
                                                            tau_g
      LOGICAL, INTENT(out), OPTIONAL                     :: rho_r_valid, drho_r_valid, rho_g_valid, &
                                                            drho_g_valid, tau_r_valid, tau_g_valid
      INTEGER, INTENT(out), OPTIONAL                     :: rebuild_each
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: tot_rho_r, tot_rho_g
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: rho_r_sccs
      LOGICAL, INTENT(out), OPTIONAL                     :: soft_valid

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_rho_get', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(rho_struct)))CALL cp__a("qs_rho_types.F",277)
      IF(.NOT.(rho_struct%ref_count > 0))CALL cp__a("qs_rho_types.F",278)

      IF (PRESENT(rho_ao)) rho_ao => get_1d_pointer(rho_struct%rho_ao)
      IF (PRESENT(rho_ao_kp)) rho_ao_kp => get_2d_pointer(rho_struct%rho_ao)

      IF (PRESENT(rho_ao_im)) rho_ao_im => rho_struct%rho_ao_im
      IF (PRESENT(rho_r)) rho_r => rho_struct%rho_r
      IF (PRESENT(drho_r)) drho_r => rho_struct%drho_r
      IF (PRESENT(rho_g)) rho_g => rho_struct%rho_g
      IF (PRESENT(drho_g)) drho_g => rho_struct%drho_g
      IF (PRESENT(tau_r)) tau_r => rho_struct%tau_r
      IF (PRESENT(tau_g)) tau_g => rho_struct%tau_g
      IF (PRESENT(rho_r_valid)) rho_r_valid = rho_struct%rho_r_valid
      IF (PRESENT(rho_g_valid)) rho_g_valid = rho_struct%rho_g_valid
      IF (PRESENT(drho_r_valid)) drho_r_valid = rho_struct%drho_r_valid
      IF (PRESENT(drho_g_valid)) drho_g_valid = rho_struct%drho_g_valid
      IF (PRESENT(tau_r_valid)) tau_r_valid = rho_struct%tau_r_valid
      IF (PRESENT(tau_g_valid)) tau_g_valid = rho_struct%tau_g_valid
      IF (PRESENT(soft_valid)) soft_valid = rho_struct%soft_valid
      IF (PRESENT(rebuild_each)) rebuild_each = rho_struct%rebuild_each
      IF (PRESENT(tot_rho_r)) tot_rho_r => rho_struct%tot_rho_r
      IF (PRESENT(tot_rho_g)) tot_rho_g => rho_struct%tot_rho_g
      IF (PRESENT(rho_r_sccs)) rho_r_sccs => rho_struct%rho_r_sccs

   END SUBROUTINE qs_rho_get

! **************************************************************************************************
!> \brief ...
!> \param rho_struct ...
!> \param rho_ao ...
!> \param rho_ao_im ...
!> \param rho_ao_kp ...
!> \param rho_r ...
!> \param drho_r ...
!> \param rho_g ...
!> \param drho_g ...
!> \param tau_r ...
!> \param tau_g ...
!> \param rho_r_valid ...
!> \param drho_r_valid ...
!> \param rho_g_valid ...
!> \param drho_g_valid ...
!> \param tau_r_valid ...
!> \param tau_g_valid ...
!> \param rebuild_each ...
!> \param tot_rho_r ...
!> \param tot_rho_g ...
!> \param rho_r_sccs ...
!> \param soft_valid ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE qs_rho_set(rho_struct, rho_ao, rho_ao_im, rho_ao_kp, rho_r, drho_r, &
                         rho_g, drho_g, tau_r, tau_g, rho_r_valid, drho_r_valid, rho_g_valid, &
                         drho_g_valid, tau_r_valid, tau_g_valid, rebuild_each, tot_rho_r, tot_rho_g, &
                         rho_r_sccs, soft_valid)
      TYPE(qs_rho_type), POINTER                         :: rho_struct
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rho_ao, rho_ao_im
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: rho_ao_kp
      TYPE(pw_p_type), DIMENSION(:), OPTIONAL, POINTER   :: rho_r, drho_r, rho_g, drho_g, tau_r, &
                                                            tau_g
      LOGICAL, INTENT(in), OPTIONAL                      :: rho_r_valid, drho_r_valid, rho_g_valid, &
                                                            drho_g_valid, tau_r_valid, tau_g_valid
      INTEGER, INTENT(in), OPTIONAL                      :: rebuild_each
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: tot_rho_r, tot_rho_g
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: rho_r_sccs
      LOGICAL, INTENT(in), OPTIONAL                      :: soft_valid

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_rho_set', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(rho_struct)))CALL cp__a("qs_rho_types.F",349)
      IF(.NOT.(rho_struct%ref_count > 0))CALL cp__a("qs_rho_types.F",350)

      IF (PRESENT(rho_ao)) CALL set_1d_pointer(rho_struct%rho_ao, rho_ao)
      IF (PRESENT(rho_ao_kp)) CALL set_2d_pointer(rho_struct%rho_ao, rho_ao_kp)

      IF (PRESENT(rho_ao_im)) rho_struct%rho_ao_im => rho_ao_im
      IF (PRESENT(rho_r)) rho_struct%rho_r => rho_r
      IF (PRESENT(rho_g)) rho_struct%rho_g => rho_g
      IF (PRESENT(drho_r)) rho_struct%drho_r => drho_r
      IF (PRESENT(drho_g)) rho_struct%drho_g => drho_g
      IF (PRESENT(tau_r)) rho_struct%tau_r => tau_r
      IF (PRESENT(tau_g)) rho_struct%tau_g => tau_g
      IF (PRESENT(rho_r_valid)) rho_struct%rho_r_valid = rho_r_valid
      IF (PRESENT(rho_g_valid)) rho_struct%rho_g_valid = rho_g_valid
      IF (PRESENT(drho_r_valid)) rho_struct%drho_r_valid = drho_r_valid
      IF (PRESENT(drho_g_valid)) rho_struct%drho_g_valid = drho_g_valid
      IF (PRESENT(tau_r_valid)) rho_struct%tau_r_valid = tau_r_valid
      IF (PRESENT(tau_g_valid)) rho_struct%tau_g_valid = tau_g_valid
      IF (PRESENT(soft_valid)) rho_struct%soft_valid = soft_valid
      IF (PRESENT(rebuild_each)) rho_struct%rebuild_each = rebuild_each
      IF (PRESENT(tot_rho_r)) rho_struct%tot_rho_r => tot_rho_r
      IF (PRESENT(tot_rho_g)) rho_struct%tot_rho_g => tot_rho_g
      IF (PRESENT(rho_r_sccs)) rho_struct%rho_r_sccs => rho_r_sccs

   END SUBROUTINE qs_rho_set

END MODULE qs_rho_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      11.2002 created [fawzi]
!>      10.2014 Moved many components from qs_env to here [Ole Schuett]
!>      11.2014 unified k-point and gamma-point code [Ole Schuett]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE qs_ks_types
   USE admm_dm_types,                   ONLY: admm_dm_release,&
                                              admm_dm_type
   USE atomic_kind_types,               ONLY: atomic_kind_type
   USE atprop_types,                    ONLY: atprop_type
   USE cell_types,                      ONLY: cell_type
   USE cp_blacs_env,                    ONLY: cp_blacs_env_release,&
                                              cp_blacs_env_retain,&
                                              cp_blacs_env_type
   USE cp_control_types,                ONLY: dft_control_release,&
                                              dft_control_retain,&
                                              dft_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_distribution_release,&
                                              cp_dbcsr_p_type,&
                                              dbcsr_distribution_obj
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_result_types,                 ONLY: cp_result_type
   USE cp_subsys_types,                 ONLY: cp_subsys_type
   USE distribution_1d_types,           ONLY: distribution_1d_type
   USE distribution_2d_types,           ONLY: distribution_2d_release,&
                                              distribution_2d_retain,&
                                              distribution_2d_type
   USE kpoint_transitional,             ONLY: get_1d_pointer,&
                                              get_2d_pointer,&
                                              kpoint_transitional_release,&
                                              kpoint_transitional_type,&
                                              set_1d_pointer,&
                                              set_2d_pointer
   USE kpoint_types,                    ONLY: kpoint_create,&
                                              kpoint_release,&
                                              kpoint_type
   USE molecule_kind_types,             ONLY: molecule_kind_type
   USE molecule_types_new,              ONLY: molecule_type
   USE particle_types,                  ONLY: particle_type
   USE pw_env_types,                    ONLY: pw_env_release,&
                                              pw_env_retain,&
                                              pw_env_type
   USE pw_types,                        ONLY: pw_p_type,&
                                              pw_release,&
                                              pw_type
   USE qs_energy_types,                 ONLY: qs_energy_type
   USE qs_force_types,                  ONLY: qs_force_type
   USE qs_kind_types,                   ONLY: qs_kind_type
   USE qs_neighbor_list_types,          ONLY: deallocate_neighbor_list_set,&
                                              neighbor_list_set_p_type
   USE qs_rho_types,                    ONLY: qs_rho_create,&
                                              qs_rho_release,&
                                              qs_rho_type
   USE qs_subsys_types,                 ONLY: qs_subsys_get,&
                                              qs_subsys_release,&
                                              qs_subsys_retain,&
                                              qs_subsys_type
   USE task_list_types,                 ONLY: deallocate_task_list,&
                                              task_list_type
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_ks_types'

   PUBLIC :: qs_ks_env_type, qs_ks_env_create, qs_ks_did_change
   PUBLIC :: qs_ks_release, qs_ks_retain, get_ks_env, set_ks_env

! **************************************************************************************************
!> \brief calculation environement to calculate the ks matrix,
!>      holds all the needed vars.
!>      assumes that the core hamiltonian and energy are up to date.
!> \param s_mstruct_changed if the sparsity structure of s in qs_env where this
!>        ks_env lives has changed
!> \param rho_changed if the rho stored in the qs_env has changed
!> \param forces_up_to_date if the forces are up to date
!> \param potential_changed if some external potentian changed (such as due to the constraint)
!>
!> \param matrix_h core hamiltonian (matrix_h(1)), if needed by the calculation also
!>        its derivatives wrt. x,y, and z (matrix_h(2:4))
!> \param matrix_ks Khon-Sham matrix (filtred by the structure of S).
!>        with LSD an array with the differents orbitals
!> \param matrix_s the overlap matrix (matrix_s(1)), if needed by the calculation also
!>        its derivatives wrt. x,y, and z (matrix_s(2:4))
!> \param kinetic kinetic part of h
!> \param gamma_matrix gamma (Coulomb) matrix in DFTB, if needed by the calculation also
!>        its derivatives wrt. x,y, and z
!>
!> \param rho the (old) density, in various representations (ao+grid)
!> \param rho_xc the (old) soft density, used for the GAPW_XC method only
!> \param rho_core a grid with the charges of the cores of the atoms in the
!>        reciprocal (g) space
!> \param vppl a realspace grid with the local pseudopotential
!> \param vee a realspace grid with the external electrostatic potential
!>
!> \param neighbor_list_id the current version of the neighbor_list
!> \param sab_orb ,sac_ae, sac_ppl, sap_ppnl:
!>        neighbor lists for the calculation of the core Hamiltonian matrix
!> \param sap_oce: neighbor lists for the calculation of the expansion
!>        coefficients of the local atomic densities rho1_hard and rho1_soft
!> \param sab_se: neighbor lists for the calculation of the 2 centers
!>        hartree term in semi-empirical methods
!> \param sac_tbe: neighbor lists for the calculation of the tight binding
!>        Ewald terms (DFTB)
!> \param sab_core neighbor lists for the calculation of the core interactions
!> \param sab_all neighbor lists for the calculation of the  matrix element of
!>        non-symmetric operators
!> \param sab_vdw: neighbor lists for the calculation of dispersion interactions
!> \param sab_scp: neighbor lists for the calculation of self-consistent polarization
!> \param sab_almo: neighbor lists to create ALMO delocalization template
!> \param sab_kp: neighbor lists to create kp image cell lists
!>
!> \param kpoints information on the kpoints used
!> \param subsys the particles, molecules,... of this environment
!> \param dft_control object that contains the values of various parameters
!>        relevant to a dft calculation
!> \param distribution_2d: distribution of the atom pairs between the processors
!>
!> \par History
!>      04.2002 created [fawzi]
!>      10.2002 introduced pools, uses updated rho as input,
!>              removed most temporary variables, renamed may vars,
!>              began conversion to LSD [fawzi]
!>      10.2014 Moved many components from qs_env here [Ole Schuett]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE qs_ks_env_type
      PRIVATE
      LOGICAL                                               :: s_mstruct_changed = .TRUE., &
                                                               rho_changed = .TRUE., &
                                                               potential_changed = .TRUE., &
                                                               forces_up_to_date = .FALSE.
      INTEGER                                               :: n_evals = 0, &
                                                               id_nr = -1, &
                                                               ref_count = 1

      ! hartree is supposed to contain the hartree potential (for cube ouput)
      ! ugly to keep it always around only for a cube output...
      TYPE(pw_type), POINTER                                :: v_hartree_rspace => Null()

      TYPE(kpoint_transitional_type)                        :: matrix_ks, &
                                                               matrix_s, &
                                                               kinetic, &
                                                               matrix_h, &
                                                               matrix_w, &
                                                               matrix_vxc

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER          :: matrix_ks_im => Null(), &
                                                               matrix_w_mp2 => Null(), &
                                                               matrix_p_mp2 => Null(), &
                                                               gamma_matrix => Null(), &
                                                               matrix_s_aux_fit => Null(), &
                                                               matrix_s_aux_fit_vs_orb => Null(), &
                                                               matrix_ks_aux_fit => Null(), &
                                                               matrix_ks_aux_fit_im => Null(), &
                                                               matrix_ks_aux_fit_dft => Null(), &
                                                               matrix_ks_aux_fit_hfx => Null()

      TYPE(qs_rho_type), POINTER                            :: rho => Null(), &
                                                               rho_buffer => Null(), &
                                                               rho_xc => Null(), &
                                                               rho_aux_fit => Null(), &
                                                               rho_aux_fit_buffer => Null()

      TYPE(pw_p_type), POINTER                              :: rho_core => Null(), &
                                                               vppl => Null(), &
                                                               rho_nlcc => Null(), &
                                                               rho_nlcc_g => Null(), &
                                                               vee => Null()

      INTEGER                                               :: neighbor_list_id = -1
      TYPE(neighbor_list_set_p_type), DIMENSION(:), POINTER :: sab_orb => Null(), &
                                                               sac_ae => Null(), &
                                                               sac_ppl => Null(), &
                                                               sap_ppnl => Null(), &
                                                               sap_oce => Null(), &
                                                               sab_se => Null(), &
                                                               sab_vdw => Null(), &
                                                               sab_scp => Null(), &
                                                               sab_tbe => Null(), &
                                                               sab_core => Null(), &
                                                               sab_all => Null(), &
                                                               sab_aux_fit => Null(), &
                                                               sab_aux_fit_vs_orb => Null(), &
                                                               sab_aux_fit_asymm => Null(), &
                                                               sab_lrc => Null(), &
                                                               sab_almo => Null(), &
                                                               sab_kp => Null()

      TYPE(task_list_type), POINTER                         :: task_list => Null()
      TYPE(task_list_type), POINTER                         :: task_list_aux_fit => Null()
      TYPE(task_list_type), POINTER                         :: task_list_soft => Null()

      TYPE(kpoint_type), POINTER                            :: kpoints => Null()
      TYPE(qs_subsys_type), POINTER                         :: subsys => Null()
      TYPE(dft_control_type), POINTER                       :: dft_control => Null()
      TYPE(dbcsr_distribution_obj), POINTER                 :: dbcsr_dist => Null()
      TYPE(distribution_2d_type), POINTER                   :: distribution_2d => Null()
      TYPE(pw_env_type), POINTER                            :: pw_env => Null()
      TYPE(cp_para_env_type), POINTER                       :: para_env => Null()
      TYPE(cp_blacs_env_type), POINTER                      :: blacs_env => Null()
      TYPE(admm_dm_type), POINTER                           :: admm_dm => Null()
   END TYPE qs_ks_env_type

! **************************************************************************************************
!> \brief type to build arrays of pointers
!> \param ks_env the ks_env pointer
!> \par History
!>      11.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE qs_ks_env_p_type
      TYPE(qs_ks_env_type), POINTER :: ks_env
   END TYPE qs_ks_env_p_type

CONTAINS

! **************************************************************************************************
!> \brief Allocates a new instance of ks_env.
!> \param ks_env ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE qs_ks_env_create(ks_env)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'qs_ks_env_create', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(ks_env)) CALL cp__b("qs_ks_types.F",242,"ks_env already associated")
      ALLOCATE (ks_env)
      CALL qs_rho_create(ks_env%rho)
      CALL qs_rho_create(ks_env%rho_aux_fit)
      CALL qs_rho_create(ks_env%rho_aux_fit_buffer)
      CALL qs_rho_create(ks_env%rho_xc)
      CALL qs_rho_create(ks_env%rho_buffer)
      CALL kpoint_create(ks_env%kpoints)
   END SUBROUTINE qs_ks_env_create

! **************************************************************************************************
!> \brief ...
!> \param ks_env ...
!> \param v_hartree_rspace ...
!> \param s_mstruct_changed ...
!> \param rho_changed ...
!> \param potential_changed ...
!> \param forces_up_to_date ...
!> \param matrix_h ...
!> \param matrix_ks ...
!> \param matrix_ks_im ...
!> \param matrix_vxc ...
!> \param matrix_ks_aux_fit ...
!> \param matrix_ks_aux_fit_im ...
!> \param matrix_ks_aux_fit_dft ...
!> \param matrix_ks_aux_fit_hfx ...
!> \param kinetic ...
!> \param matrix_s ...
!> \param matrix_s_aux_fit ...
!> \param matrix_s_aux_fit_vs_orb ...
!> \param matrix_w ...
!> \param matrix_w_mp2 ...
!> \param matrix_p_mp2 ...
!> \param gamma_matrix ...
!> \param matrix_h_kp ...
!> \param matrix_ks_kp ...
!> \param matrix_vxc_kp ...
!> \param kinetic_kp ...
!> \param matrix_s_kp ...
!> \param matrix_w_kp ...
!> \param rho ...
!> \param rho_aux_fit ...
!> \param rho_aux_fit_buffer ...
!> \param rho_buffer ...
!> \param rho_xc ...
!> \param vppl ...
!> \param rho_core ...
!> \param rho_nlcc ...
!> \param rho_nlcc_g ...
!> \param vee ...
!> \param neighbor_list_id ...
!> \param sab_orb ...
!> \param sab_aux_fit ...
!> \param sab_aux_fit_asymm ...
!> \param sab_aux_fit_vs_orb ...
!> \param sab_all ...
!> \param sac_ae ...
!> \param sac_ppl ...
!> \param sap_ppnl ...
!> \param sap_oce ...
!> \param sab_lrc ...
!> \param sab_se ...
!> \param sab_tbe ...
!> \param sab_core ...
!> \param sab_vdw ...
!> \param sab_scp ...
!> \param sab_almo ...
!> \param sab_kp ...
!> \param task_list ...
!> \param task_list_aux_fit ...
!> \param task_list_soft ...
!> \param kpoints ...
!> \param do_kpoints ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param cell ...
!> \param cell_ref ...
!> \param use_ref_cell ...
!> \param particle_set ...
!> \param energy ...
!> \param force ...
!> \param local_particles ...
!> \param local_molecules ...
!> \param molecule_kind_set ...
!> \param molecule_set ...
!> \param subsys ...
!> \param cp_subsys ...
!> \param virial ...
!> \param results ...
!> \param atprop ...
!> \param nkind ...
!> \param natom ...
!> \param dft_control ...
!> \param dbcsr_dist ...
!> \param distribution_2d ...
!> \param pw_env ...
!> \param para_env ...
!> \param blacs_env ...
!> \param nelectron_total ...
!> \param nelectron_spin ...
!> \param admm_dm ...
! **************************************************************************************************
   SUBROUTINE get_ks_env(ks_env, v_hartree_rspace, &
                         s_mstruct_changed, rho_changed, &
                         potential_changed, forces_up_to_date, &
                         matrix_h, matrix_ks, matrix_ks_im, matrix_vxc, matrix_ks_aux_fit, &
                         matrix_ks_aux_fit_im, matrix_ks_aux_fit_dft, matrix_ks_aux_fit_hfx, &
                         kinetic, matrix_s, matrix_s_aux_fit, matrix_s_aux_fit_vs_orb, &
                         matrix_w, matrix_w_mp2, matrix_p_mp2, gamma_matrix, &
                         matrix_h_kp, matrix_ks_kp, matrix_vxc_kp, kinetic_kp, matrix_s_kp, matrix_w_kp, &
                         rho, rho_aux_fit, rho_aux_fit_buffer, &
                         rho_buffer, rho_xc, &
                         vppl, rho_core, rho_nlcc, rho_nlcc_g, vee, &
                         neighbor_list_id, &
                         sab_orb, sab_aux_fit, sab_aux_fit_asymm, &
                         sab_aux_fit_vs_orb, sab_all, sac_ae, sac_ppl, sap_ppnl, sap_oce, sab_lrc, &
                         sab_se, sab_tbe, sab_core, sab_vdw, sab_scp, sab_almo, sab_kp, &
                         task_list, task_list_aux_fit, task_list_soft, &
                         kpoints, do_kpoints, &
                         atomic_kind_set, qs_kind_set, cell, cell_ref, use_ref_cell, &
                         particle_set, energy, force, local_particles, local_molecules, &
                         molecule_kind_set, molecule_set, subsys, cp_subsys, virial, results, atprop, &
                         nkind, natom, dft_control, dbcsr_dist, distribution_2d, pw_env, &
                         para_env, blacs_env, nelectron_total, nelectron_spin, admm_dm)

      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(pw_type), OPTIONAL, POINTER                   :: v_hartree_rspace
      LOGICAL, OPTIONAL                                  :: s_mstruct_changed, rho_changed, &
                                                            potential_changed, forces_up_to_date
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, POINTER :: matrix_h, matrix_ks, matrix_ks_im, &
         matrix_vxc, matrix_ks_aux_fit, matrix_ks_aux_fit_im, matrix_ks_aux_fit_dft, &
         matrix_ks_aux_fit_hfx, kinetic, matrix_s, matrix_s_aux_fit, matrix_s_aux_fit_vs_orb, &
         matrix_w, matrix_w_mp2, matrix_p_mp2, gamma_matrix
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: matrix_h_kp, matrix_ks_kp, &
                                                            matrix_vxc_kp, kinetic_kp, &
                                                            matrix_s_kp, matrix_w_kp
      TYPE(qs_rho_type), OPTIONAL, POINTER               :: rho, rho_aux_fit, rho_aux_fit_buffer, &
                                                            rho_buffer, rho_xc
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: vppl, rho_core, rho_nlcc, rho_nlcc_g, vee
      INTEGER, OPTIONAL                                  :: neighbor_list_id
      TYPE(neighbor_list_set_p_type), DIMENSION(:), OPTIONAL, POINTER :: sab_orb, sab_aux_fit, &
         sab_aux_fit_asymm, sab_aux_fit_vs_orb, sab_all, sac_ae, sac_ppl, sap_ppnl, sap_oce, &
         sab_lrc, sab_se, sab_tbe, sab_core, sab_vdw, sab_scp, sab_almo, sab_kp
      TYPE(task_list_type), OPTIONAL, POINTER            :: task_list, task_list_aux_fit, &
                                                            task_list_soft
      TYPE(kpoint_type), OPTIONAL, POINTER               :: kpoints
      LOGICAL, OPTIONAL                                  :: do_kpoints
      TYPE(atomic_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: qs_kind_set
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell, cell_ref
      LOGICAL, OPTIONAL                                  :: use_ref_cell
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      TYPE(qs_energy_type), OPTIONAL, POINTER            :: energy
      TYPE(qs_force_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: force
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_particles, local_molecules
      TYPE(molecule_kind_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_kind_set
      TYPE(molecule_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: molecule_set
      TYPE(qs_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(cp_subsys_type), OPTIONAL, POINTER            :: cp_subsys
      TYPE(virial_type), OPTIONAL, POINTER               :: virial
      TYPE(cp_result_type), OPTIONAL, POINTER            :: results
      TYPE(atprop_type), OPTIONAL, POINTER               :: atprop
      INTEGER, OPTIONAL                                  :: nkind, natom
      TYPE(dft_control_type), OPTIONAL, POINTER          :: dft_control
      TYPE(dbcsr_distribution_obj), OPTIONAL, POINTER    :: dbcsr_dist
      TYPE(distribution_2d_type), OPTIONAL, POINTER      :: distribution_2d
      TYPE(pw_env_type), OPTIONAL, POINTER               :: pw_env
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(cp_blacs_env_type), OPTIONAL, POINTER         :: blacs_env
      INTEGER, OPTIONAL                                  :: nelectron_total
      INTEGER, DIMENSION(2), OPTIONAL                    :: nelectron_spin
      TYPE(admm_dm_type), OPTIONAL, POINTER              :: admm_dm

      IF (.NOT. ASSOCIATED(ks_env)) CALL cp__b("qs_ks_types.F",422,"get_ks_env: not associated")
      IF (ks_env%ref_count < 1) CALL cp__b("qs_ks_types.F",423,"get_ks_env: ks_env%ref_count<1")
      IF (.NOT. ASSOCIATED(ks_env%subsys)) CALL cp__b("qs_ks_types.F",424,"get_ks_env: subsys not associated")

      IF (PRESENT(s_mstruct_changed)) s_mstruct_changed = ks_env%s_mstruct_changed
      IF (PRESENT(rho_changed)) rho_changed = ks_env%rho_changed
      IF (PRESENT(potential_changed)) potential_changed = ks_env%potential_changed
      IF (PRESENT(forces_up_to_date)) forces_up_to_date = ks_env%forces_up_to_date
      IF (PRESENT(v_hartree_rspace)) v_hartree_rspace => ks_env%v_hartree_rspace

      IF (PRESENT(matrix_ks)) matrix_ks => get_1d_pointer(ks_env%matrix_ks)
      IF (PRESENT(matrix_s)) matrix_s => get_1d_pointer(ks_env%matrix_s)
      IF (PRESENT(kinetic)) kinetic => get_1d_pointer(ks_env%kinetic)
      IF (PRESENT(matrix_h)) matrix_h => get_1d_pointer(ks_env%matrix_h)
      IF (PRESENT(matrix_w)) matrix_w => get_1d_pointer(ks_env%matrix_w)
      IF (PRESENT(matrix_vxc)) matrix_vxc => get_1d_pointer(ks_env%matrix_vxc)

      IF (PRESENT(matrix_ks_kp)) matrix_ks_kp => get_2d_pointer(ks_env%matrix_ks)
      IF (PRESENT(matrix_s_kp)) matrix_s_kp => get_2d_pointer(ks_env%matrix_s)
      IF (PRESENT(matrix_w_kp)) matrix_w_kp => get_2d_pointer(ks_env%matrix_w)
      IF (PRESENT(kinetic_kp)) kinetic_kp => get_2d_pointer(ks_env%kinetic)
      IF (PRESENT(matrix_h_kp)) matrix_h_kp => get_2d_pointer(ks_env%matrix_h)
      IF (PRESENT(matrix_vxc_kp)) matrix_vxc_kp => get_2d_pointer(ks_env%matrix_vxc)

      IF (PRESENT(matrix_ks_im)) matrix_ks_im => ks_env%matrix_ks_im
      IF (PRESENT(matrix_ks_aux_fit)) matrix_ks_aux_fit => ks_env%matrix_ks_aux_fit
      IF (PRESENT(matrix_ks_aux_fit_im)) matrix_ks_aux_fit_im => ks_env%matrix_ks_aux_fit_im
      IF (PRESENT(matrix_ks_aux_fit_dft)) matrix_ks_aux_fit_dft => ks_env%matrix_ks_aux_fit_dft
      IF (PRESENT(matrix_ks_aux_fit_hfx)) matrix_ks_aux_fit_hfx => ks_env%matrix_ks_aux_fit_hfx
      IF (PRESENT(matrix_s_aux_fit)) matrix_s_aux_fit => ks_env%matrix_s_aux_fit
      IF (PRESENT(matrix_s_aux_fit_vs_orb)) matrix_s_aux_fit_vs_orb => ks_env%matrix_s_aux_fit_vs_orb
      IF (PRESENT(matrix_w_mp2)) matrix_w_mp2 => ks_env%matrix_w_mp2
      IF (PRESENT(matrix_p_mp2)) matrix_p_mp2 => ks_env%matrix_p_mp2
      IF (PRESENT(gamma_matrix)) gamma_matrix => ks_env%gamma_matrix
      IF (PRESENT(rho)) rho => ks_env%rho
      IF (PRESENT(rho_aux_fit)) rho_aux_fit => ks_env%rho_aux_fit
      IF (PRESENT(rho_aux_fit_buffer)) rho_aux_fit_buffer => ks_env%rho_aux_fit_buffer
      IF (PRESENT(rho_buffer)) rho_buffer => ks_env%rho_buffer
      IF (PRESENT(rho_xc)) rho_xc => ks_env%rho_xc
      IF (PRESENT(rho_core)) rho_core => ks_env%rho_core
      IF (PRESENT(rho_nlcc)) rho_nlcc => ks_env%rho_nlcc
      IF (PRESENT(rho_nlcc_g)) rho_nlcc_g => ks_env%rho_nlcc_g
      IF (PRESENT(vppl)) vppl => ks_env%vppl
      IF (PRESENT(vee)) vee => ks_env%vee

      IF (PRESENT(neighbor_list_id)) neighbor_list_id = ks_env%neighbor_list_id
      IF (PRESENT(sab_orb)) sab_orb => ks_env%sab_orb
      IF (PRESENT(sab_aux_fit)) sab_aux_fit => ks_env%sab_aux_fit
      IF (PRESENT(sab_aux_fit_asymm)) sab_aux_fit_asymm => ks_env%sab_aux_fit_asymm
      IF (PRESENT(sab_aux_fit_vs_orb)) sab_aux_fit_vs_orb => ks_env%sab_aux_fit_vs_orb
      IF (PRESENT(sab_all)) sab_all => ks_env%sab_all
      IF (PRESENT(sab_vdw)) sab_vdw => ks_env%sab_vdw
      IF (PRESENT(sab_scp)) sab_scp => ks_env%sab_scp
      IF (PRESENT(sac_ae)) sac_ae => ks_env%sac_ae
      IF (PRESENT(sac_ppl)) sac_ppl => ks_env%sac_ppl
      IF (PRESENT(sap_ppnl)) sap_ppnl => ks_env%sap_ppnl
      IF (PRESENT(sap_oce)) sap_oce => ks_env%sap_oce
      IF (PRESENT(sab_se)) sab_se => ks_env%sab_se
      IF (PRESENT(sab_lrc)) sab_lrc => ks_env%sab_lrc
      IF (PRESENT(sab_tbe)) sab_tbe => ks_env%sab_tbe
      IF (PRESENT(sab_core)) sab_core => ks_env%sab_core
      IF (PRESENT(sab_almo)) sab_almo => ks_env%sab_almo
      IF (PRESENT(sab_kp)) sab_kp => ks_env%sab_kp
      IF (PRESENT(dft_control)) dft_control => ks_env%dft_control
      IF (PRESENT(dbcsr_dist)) dbcsr_dist => ks_env%dbcsr_dist
      IF (PRESENT(distribution_2d)) distribution_2d => ks_env%distribution_2d
      IF (PRESENT(pw_env)) pw_env => ks_env%pw_env
      IF (PRESENT(para_env)) para_env => ks_env%para_env
      IF (PRESENT(blacs_env)) blacs_env => ks_env%blacs_env
      IF (PRESENT(admm_dm)) admm_dm => ks_env%admm_dm

      IF (PRESENT(task_list)) task_list => ks_env%task_list
      IF (PRESENT(task_list_aux_fit)) task_list_aux_fit => ks_env%task_list_aux_fit
      IF (PRESENT(task_list_soft)) task_list_soft => ks_env%task_list_soft

      IF (PRESENT(kpoints)) kpoints => ks_env%kpoints
      IF (PRESENT(do_kpoints)) THEN
         do_kpoints = (ks_env%kpoints%nkp > 0)
      END IF

      IF (PRESENT(subsys)) subsys => ks_env%subsys
      CALL qs_subsys_get(ks_env%subsys, &
                         local_molecules_new=local_molecules, &
                         local_particles=local_particles, &
                         atprop=atprop, &
                         virial=virial, &
                         results=results, &
                         cell=cell, &
                         cell_ref=cell_ref, &
                         use_ref_cell=use_ref_cell, &
                         energy=energy, &
                         force=force, &
                         qs_kind_set=qs_kind_set, &
                         cp_subsys=cp_subsys, &
                         atomic_kind_set=atomic_kind_set, &
                         particle_set=particle_set, &
                         molecule_kind_set=molecule_kind_set, &
                         molecule_set=molecule_set, &
                         natom=natom, &
                         nkind=nkind, &
                         nelectron_total=nelectron_total, &
                         nelectron_spin=nelectron_spin)

   END SUBROUTINE get_ks_env

! **************************************************************************************************
!> \brief ...
!> \param ks_env ...
!> \param v_hartree_rspace ...
!> \param s_mstruct_changed ...
!> \param rho_changed ...
!> \param potential_changed ...
!> \param forces_up_to_date ...
!> \param matrix_h ...
!> \param matrix_ks ...
!> \param matrix_ks_im ...
!> \param matrix_vxc ...
!> \param matrix_ks_aux_fit ...
!> \param matrix_ks_aux_fit_im ...
!> \param matrix_ks_aux_fit_dft ...
!> \param matrix_ks_aux_fit_hfx ...
!> \param kinetic ...
!> \param matrix_s ...
!> \param matrix_s_aux_fit ...
!> \param matrix_s_aux_fit_vs_orb ...
!> \param matrix_w ...
!> \param matrix_w_mp2 ...
!> \param matrix_p_mp2 ...
!> \param gamma_matrix ...
!> \param matrix_h_kp ...
!> \param matrix_ks_kp ...
!> \param matrix_vxc_kp ...
!> \param kinetic_kp ...
!> \param matrix_s_kp ...
!> \param matrix_w_kp ...
!> \param vppl ...
!> \param rho_core ...
!> \param rho_nlcc ...
!> \param rho_nlcc_g ...
!> \param vee ...
!> \param neighbor_list_id ...
!> \param sab_orb ...
!> \param sab_aux_fit ...
!> \param sab_aux_fit_asymm ...
!> \param sab_aux_fit_vs_orb ...
!> \param sab_all ...
!> \param sac_ae ...
!> \param sac_ppl ...
!> \param sap_ppnl ...
!> \param sap_oce ...
!> \param sab_lrc ...
!> \param sab_se ...
!> \param sab_tbe ...
!> \param sab_core ...
!> \param sab_vdw ...
!> \param sab_scp ...
!> \param sab_almo ...
!> \param sab_kp ...
!> \param task_list ...
!> \param task_list_aux_fit ...
!> \param task_list_soft ...
!> \param subsys ...
!> \param dft_control ...
!> \param dbcsr_dist ...
!> \param distribution_2d ...
!> \param pw_env ...
!> \param para_env ...
!> \param blacs_env ...
!> \param admm_dm ...
! **************************************************************************************************
   SUBROUTINE set_ks_env(ks_env, v_hartree_rspace, &
                         s_mstruct_changed, rho_changed, &
                         potential_changed, forces_up_to_date, &
                         matrix_h, matrix_ks, matrix_ks_im, matrix_vxc, matrix_ks_aux_fit, matrix_ks_aux_fit_im, &
                         matrix_ks_aux_fit_dft, matrix_ks_aux_fit_hfx, kinetic, matrix_s, matrix_s_aux_fit, &
                         matrix_s_aux_fit_vs_orb, matrix_w, matrix_w_mp2, matrix_p_mp2, gamma_matrix, &
                         matrix_h_kp, matrix_ks_kp, matrix_vxc_kp, kinetic_kp, matrix_s_kp, matrix_w_kp, &
                         vppl, rho_core, rho_nlcc, rho_nlcc_g, vee, &
                         neighbor_list_id, &
                         sab_orb, sab_aux_fit, sab_aux_fit_asymm, &
                         sab_aux_fit_vs_orb, sab_all, sac_ae, sac_ppl, sap_ppnl, sap_oce, sab_lrc, &
                         sab_se, sab_tbe, sab_core, sab_vdw, sab_scp, sab_almo, sab_kp, &
                         task_list, task_list_aux_fit, task_list_soft, &
                         subsys, dft_control, dbcsr_dist, distribution_2d, pw_env, &
                         para_env, blacs_env, admm_dm)

      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      TYPE(pw_type), OPTIONAL, POINTER                   :: v_hartree_rspace
      LOGICAL, OPTIONAL                                  :: s_mstruct_changed, rho_changed, &
                                                            potential_changed, forces_up_to_date
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, POINTER :: matrix_h, matrix_ks, matrix_ks_im, &
         matrix_vxc, matrix_ks_aux_fit, matrix_ks_aux_fit_im, matrix_ks_aux_fit_dft, &
         matrix_ks_aux_fit_hfx, kinetic, matrix_s, matrix_s_aux_fit, matrix_s_aux_fit_vs_orb, &
         matrix_w, matrix_w_mp2, matrix_p_mp2, gamma_matrix
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: matrix_h_kp, matrix_ks_kp, &
                                                            matrix_vxc_kp, kinetic_kp, &
                                                            matrix_s_kp, matrix_w_kp
      TYPE(pw_p_type), OPTIONAL, POINTER                 :: vppl, rho_core, rho_nlcc, rho_nlcc_g, vee
      INTEGER, OPTIONAL                                  :: neighbor_list_id
      TYPE(neighbor_list_set_p_type), DIMENSION(:), OPTIONAL, POINTER :: sab_orb, sab_aux_fit, &
         sab_aux_fit_asymm, sab_aux_fit_vs_orb, sab_all, sac_ae, sac_ppl, sap_ppnl, sap_oce, &
         sab_lrc, sab_se, sab_tbe, sab_core, sab_vdw, sab_scp, sab_almo, sab_kp
      TYPE(task_list_type), OPTIONAL, POINTER            :: task_list, task_list_aux_fit, &
                                                            task_list_soft
      TYPE(qs_subsys_type), OPTIONAL, POINTER            :: subsys
      TYPE(dft_control_type), OPTIONAL, POINTER          :: dft_control
      TYPE(dbcsr_distribution_obj), OPTIONAL, POINTER    :: dbcsr_dist
      TYPE(distribution_2d_type), OPTIONAL, POINTER      :: distribution_2d
      TYPE(pw_env_type), OPTIONAL, POINTER               :: pw_env
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(cp_blacs_env_type), OPTIONAL, POINTER         :: blacs_env
      TYPE(admm_dm_type), OPTIONAL, POINTER              :: admm_dm

      IF (.NOT. ASSOCIATED(ks_env)) CALL cp__b("qs_ks_types.F",636,"set_ks_env: not associated")
      IF (ks_env%ref_count < 1) CALL cp__b("qs_ks_types.F",637,"set_ks_env: ks_env%ref_count<1")

      IF (PRESENT(s_mstruct_changed)) ks_env%s_mstruct_changed = s_mstruct_changed
      IF (PRESENT(rho_changed)) ks_env%rho_changed = rho_changed
      IF (PRESENT(potential_changed)) ks_env%potential_changed = potential_changed
      IF (PRESENT(forces_up_to_date)) ks_env%forces_up_to_date = forces_up_to_date
      IF (ks_env%s_mstruct_changed .OR. ks_env%potential_changed .OR. ks_env%rho_changed) &
         ks_env%forces_up_to_date = .FALSE.
      IF (PRESENT(v_hartree_rspace)) ks_env%v_hartree_rspace => v_hartree_rspace

      IF (PRESENT(matrix_h)) CALL set_1d_pointer(ks_env%matrix_h, matrix_h)
      IF (PRESENT(matrix_h_kp)) CALL set_2d_pointer(ks_env%matrix_h, matrix_h_kp)

      IF (PRESENT(matrix_ks)) CALL set_1d_pointer(ks_env%matrix_ks, matrix_ks)
      IF (PRESENT(matrix_ks_kp)) CALL set_2d_pointer(ks_env%matrix_ks, matrix_ks_kp)

      IF (PRESENT(matrix_vxc)) CALL set_1d_pointer(ks_env%matrix_vxc, matrix_vxc)
      IF (PRESENT(matrix_vxc_kp)) CALL set_2d_pointer(ks_env%matrix_vxc, matrix_vxc_kp)

      IF (PRESENT(matrix_s)) CALL set_1d_pointer(ks_env%matrix_s, matrix_s)
      IF (PRESENT(matrix_s_kp)) CALL set_2d_pointer(ks_env%matrix_s, matrix_s_kp)

      IF (PRESENT(kinetic)) CALL set_1d_pointer(ks_env%kinetic, kinetic)
      IF (PRESENT(kinetic_kp)) CALL set_2d_pointer(ks_env%kinetic, kinetic_kp)

      IF (PRESENT(matrix_w)) CALL set_1d_pointer(ks_env%matrix_w, matrix_w)
      IF (PRESENT(matrix_w_kp)) CALL set_2d_pointer(ks_env%matrix_w, matrix_w_kp)

      IF (PRESENT(matrix_ks_im)) ks_env%matrix_ks_im => matrix_ks_im
      IF (PRESENT(matrix_ks_aux_fit)) ks_env%matrix_ks_aux_fit => matrix_ks_aux_fit
      IF (PRESENT(matrix_ks_aux_fit_im)) ks_env%matrix_ks_aux_fit_im => matrix_ks_aux_fit_im
      IF (PRESENT(matrix_ks_aux_fit_dft)) ks_env%matrix_ks_aux_fit_dft => matrix_ks_aux_fit_dft
      IF (PRESENT(matrix_ks_aux_fit_hfx)) ks_env%matrix_ks_aux_fit_hfx => matrix_ks_aux_fit_hfx
      IF (PRESENT(matrix_s_aux_fit)) ks_env%matrix_s_aux_fit => matrix_s_aux_fit
      IF (PRESENT(matrix_s_aux_fit_vs_orb)) ks_env%matrix_s_aux_fit_vs_orb => matrix_s_aux_fit_vs_orb
      IF (PRESENT(matrix_w_mp2)) ks_env%matrix_w_mp2 => matrix_w_mp2
      IF (PRESENT(matrix_p_mp2)) ks_env%matrix_p_mp2 => matrix_p_mp2
      IF (PRESENT(gamma_matrix)) ks_env%gamma_matrix => gamma_matrix
      IF (PRESENT(rho_core)) ks_env%rho_core => rho_core
      IF (PRESENT(rho_nlcc)) ks_env%rho_nlcc => rho_nlcc
      IF (PRESENT(rho_nlcc_g)) ks_env%rho_nlcc_g => rho_nlcc_g
      IF (PRESENT(vppl)) ks_env%vppl => vppl
      IF (PRESENT(vee)) ks_env%vee => vee

      IF (PRESENT(neighbor_list_id)) ks_env%neighbor_list_id = neighbor_list_id
      IF (PRESENT(sab_orb)) ks_env%sab_orb => sab_orb
      IF (PRESENT(sab_aux_fit)) ks_env%sab_aux_fit => sab_aux_fit
      IF (PRESENT(sab_aux_fit_asymm)) ks_env%sab_aux_fit_asymm => sab_aux_fit_asymm
      IF (PRESENT(sab_aux_fit_vs_orb)) ks_env%sab_aux_fit_vs_orb => sab_aux_fit_vs_orb
      IF (PRESENT(sab_vdw)) ks_env%sab_vdw => sab_vdw
      IF (PRESENT(sab_scp)) ks_env%sab_scp => sab_scp
      IF (PRESENT(sab_all)) ks_env%sab_all => sab_all
      IF (PRESENT(sac_ae)) ks_env%sac_ae => sac_ae
      IF (PRESENT(sac_ppl)) ks_env%sac_ppl => sac_ppl
      IF (PRESENT(sap_ppnl)) ks_env%sap_ppnl => sap_ppnl
      IF (PRESENT(sap_oce)) ks_env%sap_oce => sap_oce
      IF (PRESENT(sab_se)) ks_env%sab_se => sab_se
      IF (PRESENT(sab_lrc)) ks_env%sab_lrc => sab_lrc
      IF (PRESENT(sab_tbe)) ks_env%sab_tbe => sab_tbe
      IF (PRESENT(sab_core)) ks_env%sab_core => sab_core
      IF (PRESENT(sab_almo)) ks_env%sab_almo => sab_almo
      IF (PRESENT(sab_kp)) ks_env%sab_kp => sab_kp
      IF (PRESENT(admm_dm)) ks_env%admm_dm => admm_dm

      IF (PRESENT(task_list)) ks_env%task_list => task_list
      IF (PRESENT(task_list_aux_fit)) ks_env%task_list_aux_fit => task_list_aux_fit
      IF (PRESENT(task_list_soft)) ks_env%task_list_soft => task_list_soft

      IF (PRESENT(subsys)) THEN
         CALL qs_subsys_release(ks_env%subsys)
         CALL qs_subsys_retain(subsys)
         ks_env%subsys => subsys
      END IF
      IF (PRESENT(dft_control)) THEN
         CALL dft_control_retain(dft_control)
         CALL dft_control_release(ks_env%dft_control)
         ks_env%dft_control => dft_control
      END IF
      IF (PRESENT(dbcsr_dist)) THEN
         IF (ASSOCIATED(ks_env%dbcsr_dist)) THEN
            CALL cp_dbcsr_distribution_release(ks_env%dbcsr_dist)
            DEALLOCATE (ks_env%dbcsr_dist)
         ENDIF
         ks_env%dbcsr_dist => dbcsr_dist
      ENDIF
      IF (PRESENT(distribution_2d)) THEN
         CALL distribution_2d_retain(distribution_2d)
         CALL distribution_2d_release(ks_env%distribution_2d)
         ks_env%distribution_2d => distribution_2d
      ENDIF
      IF (PRESENT(pw_env)) THEN
         CALL pw_env_retain(pw_env)
         CALL pw_env_release(ks_env%pw_env)
         ks_env%pw_env => pw_env
      END IF
      IF (PRESENT(para_env)) THEN
         CALL cp_para_env_retain(para_env)
         CALL cp_para_env_release(ks_env%para_env)
         ks_env%para_env => para_env
      ENDIF
      IF (PRESENT(blacs_env)) THEN
         CALL cp_blacs_env_retain(blacs_env)
         CALL cp_blacs_env_release(ks_env%blacs_env)
         ks_env%blacs_env => blacs_env
      ENDIF
   END SUBROUTINE set_ks_env

! **************************************************************************************************
!> \brief releases the ks_env (see doc/ReferenceCounting.html)
!> \param ks_env the ks_env to be released
!> \par History
!>      05.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_ks_release(ks_env)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ks_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(ks_env)) THEN
         IF(.NOT.(ks_env%ref_count > 0))CALL cp__a("qs_ks_types.F",757)
         ks_env%ref_count = ks_env%ref_count-1

         IF (ks_env%ref_count < 1) THEN
            IF (ASSOCIATED(ks_env%v_hartree_rspace)) &
               CALL pw_release(ks_env%v_hartree_rspace)
            IF (ASSOCIATED(ks_env%matrix_ks_im)) &
               CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_ks_im)

            CALL kpoint_transitional_release(ks_env%matrix_ks)
            CALL kpoint_transitional_release(ks_env%matrix_h)
            CALL kpoint_transitional_release(ks_env%matrix_vxc)
            CALL kpoint_transitional_release(ks_env%matrix_s)
            CALL kpoint_transitional_release(ks_env%matrix_w)
            CALL kpoint_transitional_release(ks_env%kinetic)

            IF (ASSOCIATED(ks_env%matrix_ks_aux_fit)) &
               CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_ks_aux_fit)
            IF (ASSOCIATED(ks_env%matrix_ks_aux_fit_im)) &
               CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_ks_aux_fit_im)
            IF (ASSOCIATED(ks_env%matrix_ks_aux_fit_dft)) &
               CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_ks_aux_fit_dft)
            IF (ASSOCIATED(ks_env%matrix_ks_aux_fit_hfx)) &
               CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_ks_aux_fit_hfx)
            IF (ASSOCIATED(ks_env%matrix_s_aux_fit)) &
               CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_s_aux_fit)
            IF (ASSOCIATED(ks_env%matrix_s_aux_fit_vs_orb)) &
               CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_s_aux_fit_vs_orb)
            IF (ASSOCIATED(ks_env%matrix_w_mp2)) &
               CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_w_mp2)
            IF (ASSOCIATED(ks_env%matrix_p_mp2)) &
               CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_p_mp2)
            IF (ASSOCIATED(ks_env%gamma_matrix)) &
               CALL cp_dbcsr_deallocate_matrix_set(ks_env%gamma_matrix)
            IF (ASSOCIATED(ks_env%rho)) &
               CALL qs_rho_release(ks_env%rho)
            IF (ASSOCIATED(ks_env%rho_buffer)) &
               CALL qs_rho_release(ks_env%rho_buffer)
            IF (ASSOCIATED(ks_env%rho_xc)) &
               CALL qs_rho_release(ks_env%rho_xc)
            IF (ASSOCIATED(ks_env%rho_aux_fit)) &
               CALL qs_rho_release(ks_env%rho_aux_fit)
            IF (ASSOCIATED(ks_env%rho_aux_fit_buffer)) &
               CALL qs_rho_release(ks_env%rho_aux_fit_buffer)
            IF (ASSOCIATED(ks_env%distribution_2d)) &
               CALL distribution_2d_release(ks_env%distribution_2d)
            IF (ASSOCIATED(ks_env%task_list)) &
               CALL deallocate_task_list(ks_env%task_list)
            IF (ASSOCIATED(ks_env%task_list_aux_fit)) &
               CALL deallocate_task_list(ks_env%task_list_aux_fit)
            IF (ASSOCIATED(ks_env%task_list_soft)) &
               CALL deallocate_task_list(ks_env%task_list_soft)

            IF (ASSOCIATED(ks_env%rho_nlcc_g)) THEN
               CALL pw_release(ks_env%rho_nlcc_g%pw)
               DEALLOCATE (ks_env%rho_nlcc_g)
            ENDIF
            IF (ASSOCIATED(ks_env%rho_nlcc)) THEN
               CALL pw_release(ks_env%rho_nlcc%pw)
               DEALLOCATE (ks_env%rho_nlcc)
            ENDIF
            IF (ASSOCIATED(ks_env%rho_core)) THEN
               CALL pw_release(ks_env%rho_core%pw)
               DEALLOCATE (ks_env%rho_core)
            ENDIF
            IF (ASSOCIATED(ks_env%vppl)) THEN
               CALL pw_release(ks_env%vppl%pw)
               DEALLOCATE (ks_env%vppl)
            ENDIF
            IF (ASSOCIATED(ks_env%vee)) THEN
               CALL pw_release(ks_env%vee%pw)
               DEALLOCATE (ks_env%vee)
            ENDIF
            IF (ASSOCIATED(ks_env%dbcsr_dist)) THEN
               CALL cp_dbcsr_distribution_release(ks_env%dbcsr_dist)
               DEALLOCATE (ks_env%dbcsr_dist)
            ENDIF

            CALL release_sab(ks_env%sab_orb)
            CALL release_sab(ks_env%sac_ae)
            CALL release_sab(ks_env%sac_ppl)
            CALL release_sab(ks_env%sap_ppnl)
            CALL release_sab(ks_env%sap_oce)
            CALL release_sab(ks_env%sab_se)
            CALL release_sab(ks_env%sab_vdw)
            CALL release_sab(ks_env%sab_scp)
            CALL release_sab(ks_env%sab_tbe)
            CALL release_sab(ks_env%sab_core)
            CALL release_sab(ks_env%sab_all)
            CALL release_sab(ks_env%sab_aux_fit)
            CALL release_sab(ks_env%sab_aux_fit_vs_orb)
            CALL release_sab(ks_env%sab_aux_fit_asymm)
            CALL release_sab(ks_env%sab_lrc)
            CALL release_sab(ks_env%sab_almo)
            CALL release_sab(ks_env%sab_kp)
            CALL dft_control_release(ks_env%dft_control)
            CALL kpoint_release(ks_env%kpoints)
            CALL qs_subsys_release(ks_env%subsys)
            CALL pw_env_release(ks_env%pw_env)
            CALL cp_para_env_release(ks_env%para_env)
            CALL cp_blacs_env_release(ks_env%blacs_env)
            CALL admm_dm_release(ks_env%admm_dm)

            DEALLOCATE (ks_env)
         END IF
      END IF
      NULLIFY (ks_env)
   END SUBROUTINE qs_ks_release

! **************************************************************************************************
!> \brief releases an array of neighbor_list_sets
!> \param sab ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE release_sab(sab)
      TYPE(neighbor_list_set_p_type), DIMENSION(:), &
         POINTER                                         :: sab

      INTEGER                                            :: i

      IF (ASSOCIATED(sab)) THEN
         DO i = 1, SIZE(sab)
            CALL deallocate_neighbor_list_set(sab(i)%neighbor_list_set)
         END DO
         DEALLOCATE (sab)
      END IF
   END SUBROUTINE release_sab

! **************************************************************************************************
!> \brief retains the given ks_environment
!> \param ks_env the KohnSham environment to retain
!> \par History
!>      11.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_ks_retain(ks_env)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ks_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ks_env)))CALL cp__a("qs_ks_types.F",897)
      IF(.NOT.(ks_env%ref_count > 0))CALL cp__a("qs_ks_types.F",898)
      ks_env%ref_count = ks_env%ref_count+1
   END SUBROUTINE qs_ks_retain

! **************************************************************************************************
!> \brief tells that some of the things relevant to the ks calculation
!>      did change. has to be called when changes happen otherwise
!>      the calculation will give wrong results.
!> \param ks_env the environement that is informed about the changes
!> \param s_mstruct_changed if true it means that the structure of the
!>        overlap matrix has changed
!>        (atoms have moved)
!> \param rho_changed if true it means that the density has changed
!> \param potential_changed ...
!> \param full_reset if true everything has changed
!> \par History
!>      4.2002 created [fawzi]
!>     12.2014 moved from qs_ks_methods, added deallocation of KS-matrices [Ole Schuett]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE qs_ks_did_change(ks_env, s_mstruct_changed, rho_changed, &
                               potential_changed, full_reset)
      TYPE(qs_ks_env_type), POINTER                      :: ks_env
      LOGICAL, INTENT(in), OPTIONAL                      :: s_mstruct_changed, rho_changed, &
                                                            potential_changed, full_reset

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ks_did_change', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      LOGICAL                                            :: my_mstruct_chg

      CALL timeset(routineN, handle)
      my_mstruct_chg = .FALSE.

      IF(.NOT.(ASSOCIATED(ks_env)))CALL cp__a("qs_ks_types.F",933)

      IF (PRESENT(rho_changed)) THEN
         IF (rho_changed) ks_env%rho_changed = .TRUE.
      END IF

      IF (PRESENT(potential_changed)) THEN
         IF (potential_changed) ks_env%potential_changed = .TRUE.
      END IF

      IF (PRESENT(s_mstruct_changed)) THEN
         IF (s_mstruct_changed) my_mstruct_chg = .TRUE.
      END IF

      IF (PRESENT(full_reset)) THEN
         IF (full_reset) THEN
            ks_env%potential_changed = .TRUE.
            my_mstruct_chg = .TRUE.
         ENDIF
      END IF

      IF (my_mstruct_chg) THEN
         ks_env%s_mstruct_changed = .TRUE.
         ! *** deallocate matrices that will have the wrong structure ***
         CALL kpoint_transitional_release(ks_env%matrix_ks)
         CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_ks_aux_fit)
         !TODO: deallocate imaginary parts as well
         !CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_ks_im)
         !CALL cp_dbcsr_deallocate_matrix_set(ks_env%matrix_ks_aux_fit_im)
      ENDIF

      CALL timestop(handle)

   END SUBROUTINE qs_ks_did_change

END MODULE qs_ks_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Provides types that are used to interface with the libint library
!> \par History
!>      10.2009 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
MODULE hfx_libint_wrapper_types

   USE ISO_C_BINDING,                   ONLY: C_DOUBLE,&
                                              C_PTR

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: lib_int, &
             lib_deriv, &
             prim_data, &
             prim_data_f_size, &
             build_eri_size, &
             build_deriv1_eri_size, &
             libint_max_am, &
             libderiv_max_am1

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'hfx_libint_wrapper_types'

! *** Set default libint parameters if not specified via arch file


   INTEGER, PARAMETER :: libint_max_am = 5
   INTEGER, PARAMETER :: libderiv_max_am1 = 4
   INTEGER, PARAMETER :: prim_data_f_size = 4*(libint_max_am-1)+1
   INTEGER, PARAMETER :: libint_vrr_classes_size = 2*(libint_max_am-1)+1
   INTEGER, PARAMETER :: libint_dvrr_classes_size = 2*(libderiv_max_am1-1)+1
   INTEGER, PARAMETER :: build_eri_size = libint_max_am-1
   INTEGER, PARAMETER :: build_deriv1_eri_size = libderiv_max_am1-1

   TYPE, BIND(C) :: prim_data
      REAL(C_DOUBLE) :: F(prim_data_f_size)
      REAL(C_DOUBLE) :: U(3, 6)
      REAL(C_DOUBLE) :: twozeta_a
      REAL(C_DOUBLE) :: twozeta_b
      REAL(C_DOUBLE) :: twozeta_c
      REAL(C_DOUBLE) :: twozeta_d
      REAL(C_DOUBLE) :: oo2z
      REAL(C_DOUBLE) :: oo2n
      REAL(C_DOUBLE) :: oo2zn
      REAL(C_DOUBLE) :: poz
      REAL(C_DOUBLE) :: pon
      REAL(C_DOUBLE) :: oo2p
      REAL(C_DOUBLE) :: ss_r12_ss
   END TYPE prim_data

   TYPE, BIND(C) :: lib_int
      TYPE(C_PTR)     :: int_stack
      TYPE(C_PTR)     :: PrimQuartet
      REAL(C_DOUBLE)  :: AB(3)
      REAL(C_DOUBLE)  :: CD(3)
      TYPE(C_PTR)     :: vrr_classes(libint_vrr_classes_size, libint_vrr_classes_size)
      TYPE(C_PTR)     :: vrr_stack
   END TYPE lib_int

   TYPE, BIND(C) :: lib_deriv
      TYPE(C_PTR)     :: int_stack
      TYPE(C_PTR)     :: PrimQuartet
      TYPE(C_PTR)     :: zero_stack
      TYPE(C_PTR)     :: ABCD(156)
      REAL(C_DOUBLE)  :: AB(3)
      REAL(C_DOUBLE)  :: CD(3)
      TYPE(C_PTR)     :: deriv_classes(12, libint_dvrr_classes_size, libint_dvrr_classes_size)
      TYPE(C_PTR)     :: deriv2_classes(144, libint_dvrr_classes_size, libint_dvrr_classes_size)
      TYPE(C_PTR)     :: dvrr_classes(libint_dvrr_classes_size, libint_dvrr_classes_size)
      TYPE(C_PTR)     :: dvtt_stack
   END TYPE lib_deriv

END MODULE hfx_libint_wrapper_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Interface to the Libint-Library or a c++ wrapper.
!> \par History
!>      11.2007 created [Manuel Guidon]
!>      10.2009 refactored [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
MODULE hfx_libint_wrapper

   USE ISO_C_BINDING,                   ONLY: C_DOUBLE,&
                                              C_F_POINTER,&
                                              C_F_PROCPOINTER,&
                                              C_INT,&
                                              C_LOC,&
                                              C_PTR,&
                                              c_funptr
   USE hfx_libint_wrapper_types,        ONLY: build_deriv1_eri_size,&
                                              build_eri_size,&
                                              lib_deriv,&
                                              lib_int,&
                                              libderiv_max_am1,&
                                              libint_max_am,&
                                              prim_data
   USE kinds,                           ONLY: dp
   USE orbital_pointers,                ONLY: nco
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: initialize_libint, terminate_libint, &
             initialize_libderiv, get_eris, get_derivs, terminate_libderiv

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'hfx_libint_wrapper'


!****************************************************************************!
!****************************************************************************!
!***                                                                      ***!
!***  WHAT FOLLOWS IS CODE TO PROVIDE STUB ROUTINES IN ABSENCE OF __LIBINT **!
!***                                                                      ***!
!****************************************************************************!
!****************************************************************************!

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param lib ...
!> \param max_am ...
! **************************************************************************************************
   SUBROUTINE initialize_libint(lib, max_am)
      TYPE(lib_int)                                      :: lib
      INTEGER                                            :: max_am

      CHARACTER(LEN=*), PARAMETER :: routineN = 'initialize_libint', &
         routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(lib))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(max_am))==-1) EXIT ;  END DO ; ENDIF

   END SUBROUTINE initialize_libint

! **************************************************************************************************
!> \brief ...
!> \param deriv ...
!> \param max_am ...
! **************************************************************************************************
   SUBROUTINE initialize_libderiv(deriv, max_am)
      TYPE(lib_deriv)                                    :: deriv
      INTEGER                                            :: max_am

      CHARACTER(LEN=*), PARAMETER :: routineN = 'initialize_libderiv', &
         routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(deriv))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(max_am))==-1) EXIT ;  END DO ; ENDIF

   END SUBROUTINE initialize_libderiv

! **************************************************************************************************
!> \brief ...
!> \param lib ...
! **************************************************************************************************
   SUBROUTINE terminate_libint(lib)
      TYPE(lib_int)                                      :: lib

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(lib))==-1) EXIT ;  END DO ; ENDIF

   END SUBROUTINE terminate_libint

! **************************************************************************************************
!> \brief ...
!> \param deriv ...
! **************************************************************************************************
   SUBROUTINE terminate_libderiv(deriv)
      TYPE(lib_deriv)                                    :: deriv

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(deriv))==-1) EXIT ;  END DO ; ENDIF

   END SUBROUTINE terminate_libderiv

! **************************************************************************************************
!> \brief ...
!> \param n_d ...
!> \param n_c ...
!> \param n_b ...
!> \param n_a ...
!> \param lib ...
!> \param prim ...
!> \param p_work ...
!> \param a_mysize ...
! **************************************************************************************************
   SUBROUTINE get_eris(n_d, n_c, n_b, n_a, lib, prim, p_work, a_mysize)
      INTEGER, INTENT(IN)                                :: n_d, n_c, n_b, n_a
      TYPE(lib_int)                                      :: lib
      TYPE(prim_data), TARGET                            :: prim
      REAL(dp), DIMENSION(:), POINTER                    :: p_work
      INTEGER                                            :: a_mysize(1)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n_a))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n_b))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n_c))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n_d))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(lib))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(prim))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(p_work))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(a_mysize))==-1) EXIT ;  END DO ; ENDIF

   END SUBROUTINE get_eris

! **************************************************************************************************
!> \brief ...
!> \param n_d ...
!> \param n_c ...
!> \param n_b ...
!> \param n_a ...
!> \param deriv ...
!> \param prim ...
!> \param work_forces ...
!> \param a_mysize ...
! **************************************************************************************************
   SUBROUTINE get_derivs(n_d, n_c, n_b, n_a, deriv, prim, work_forces, a_mysize)
      INTEGER, INTENT(IN)                                :: n_d, n_c, n_b, n_a
      TYPE(lib_deriv)                                    :: deriv
      TYPE(prim_data), TARGET                            :: prim
      REAL(dp), DIMENSION(nco(n_a)*nco(n_b)*nco(n_c)*nco&
         (n_d), 12)                                      :: work_forces
      INTEGER                                            :: a_mysize(1)

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n_a))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n_b))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n_c))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(n_d))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(deriv))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(prim))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(work_forces))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(a_mysize))==-1) EXIT ;  END DO ; ENDIF

   END SUBROUTINE get_derivs

END MODULE hfx_libint_wrapper
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Some auxiliary functions and subroutines needed for HFX calculations
!> \par History
!>      04.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
MODULE hfx_helpers
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE
   PRIVATE

   PUBLIC :: count_cells_perd, &
             next_image_cell_perd

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'hfx_helpers'

!***

CONTAINS

! **************************************************************************************************
!> \brief - Auxiliary function for creating periodic neighbor cells
!> \param shell number of shells in each coordinate direction
!> \param perd ...
!> \retval count_cells_perd ...
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   FUNCTION count_cells_perd(shell, perd)
      INTEGER, INTENT(IN)                                :: shell, perd(3)
      INTEGER                                            :: count_cells_perd

      INTEGER                                            :: i, j, k

      count_cells_perd = 0
      DO i = -shell*perd(1), shell*perd(1)
         DO j = -shell*perd(2), shell*perd(2)
            DO k = -shell*perd(3), shell*perd(3)
               IF ((i**2+j**2+k**2 == shell)) count_cells_perd = count_cells_perd+1
            ENDDO
         ENDDO
      ENDDO
   END FUNCTION count_cells_perd

! **************************************************************************************************
!> \brief - Auxiliary function for creating periodic neighbor cells
!> \param m ...
!> \param perd ...
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE next_image_cell_perd(m, perd)
      INTEGER                                            :: m(3)
      INTEGER, INTENT(IN)                                :: perd(3)

      INTEGER                                            :: i, j, k, shell
      LOGICAL                                            :: found

      found = .FALSE.
      shell = SUM(m**2)
      outer: DO
         DO i = -shell*perd(1), shell*perd(1)
         DO j = -shell*perd(2), shell*perd(2)
            inner: DO k = -shell*perd(3), shell*perd(3)
               IF (.NOT. (i**2+j**2+k**2 == shell)) CYCLE inner
               IF (found) THEN
                  m = (/i, j, k/)
                  EXIT outer
               ENDIF
               IF (ALL(M .EQ. (/i, j, k/))) found = .TRUE.
            ENDDO inner
         ENDDO
         ENDDO
         shell = shell+1
      ENDDO outer
   END SUBROUTINE next_image_cell_perd

! **************************************************************************************************

END MODULE hfx_helpers
!
! * Copyright (c) 2008, 2009, Joost VandeVondele and Manuel Guidon
! * All rights reserved.
! *
! * Redistribution and use in source and binary forms, with or without
! * modification, are permitted provided that the following conditions are met:
! *     * Redistributions of source code must retain the above copyright
! *       notice, this list of conditions and the following disclaimer.
! *     * Redistributions in binary form must reproduce the above copyright
! *       notice, this list of conditions and the following disclaimer in the
! *       documentation and/or other materials provided with the distribution.
! *
! * THIS SOFTWARE IS PROVIDED BY Joost VandeVondele and Manuel Guidon AS IS AND ANY
! * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
! * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
! * DISCLAIMED. IN NO EVENT SHALL Joost VandeVondele or Manuel Guidon BE LIABLE FOR ANY
! * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
! * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
! * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
! * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
! * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
! * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
!
!
!   This module computes the basic integrals for the
!   truncated coulomb operator
!
!   res(1) =G_0(R,T)= ((2*erf(sqrt(t))+erf(R-sqrt(t))-erf(R+sqrt(t)))/sqrt(t))
!
!   and up to 21 derivatives with respect to T
!
!   res(n+1)=(-1)**n d^n/dT^n G_0(R,T)
!
!
!   The function is only computed for values of R,T which fulfil
!
!   R**2 - 11.0_dp*R + 0.0_dp < T < R**2 + 11.0_dp*R + 50.0_dp
!   where R>=0 T>=0
!
!   for T larger than the upper bound, 0 is returned
!   (which is accurate at least up to 1.0E-16)
!   while for T smaller than the lower bound, the caller is instructed
!   to use the conventional gamma function instead
!   (i.e. the limit of above expression for R to Infinity)
!
!   Joost VandeVondele and Manuel Guidon, Nov 2008.
! instead of a module one could use: INTEGER, PARAMETER ::  dp=KIND(0.0D0)

MODULE t_c_g0
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_bcast
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: t_c_g0_n, init, free_C0
   REAL(KIND=dp), DIMENSION(:, :), ALLOCATABLE, SAVE :: C0
   INTEGER, PARAMETER :: degree = 13
   REAL(KIND=dp), PARAMETER :: target_error = 0.100000E-08
   INTEGER, PARAMETER :: nderiv_max = 21
   INTEGER, SAVE      :: nderiv_init = -1
   INTEGER, SAVE      :: patches = -1
CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param RES ...
!> \param use_gamma ...
!> \param R ...
!> \param T ...
!> \param NDERIV ...
! **************************************************************************************************
   SUBROUTINE t_c_g0_n(RES, use_gamma, R, T, NDERIV)
      REAL(KIND=dp), INTENT(OUT)                         :: RES(*)
      LOGICAL, INTENT(OUT)                               :: use_gamma
      REAL(KIND=dp), INTENT(IN)                          :: R, T
      INTEGER, INTENT(IN)                                :: NDERIV

      REAL(KIND=dp)                                      :: lower, TG1, TG2, upper, X1, X2

      use_gamma = .FALSE.
      upper = R**2+11.0_dp*R+50.0_dp
      lower = R**2-11.0_dp*R+0.0_dp
      IF (T > upper) THEN
         RES(1:NDERIV+1) = 0.0_dp
         RETURN
      ENDIF
      IF (R <= 11.0_dp) THEN
         X2 = R/11.0_dp
         upper = R**2+11.0_dp*R+50.0_dp
         lower = 0.0_dp
         X1 = (T-lower)/(upper-lower)
         IF (X1 <= 0.500000000000000000E+00_dp) THEN
            IF (X2 <= 0.500000000000000000E+00_dp) THEN
               IF (X2 <= 0.250000000000000000E+00_dp) THEN
                  IF (X2 <= 0.125000000000000000E+00_dp) THEN
                     IF (X1 <= 0.250000000000000000E+00_dp) THEN
                        IF (X2 <= 0.625000000000000000E-01_dp) THEN
                           IF (X1 <= 0.125000000000000000E+00_dp) THEN
                              IF (X2 <= 0.312500000000000000E-01_dp) THEN
                                 IF (X1 <= 0.625000000000000000E-01_dp) THEN
                                    IF (X2 <= 0.156250000000000000E-01_dp) THEN
                                       IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.156250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 1))
                                       ELSE
                                          TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.156250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 2))
                                       ENDIF
                                    ELSE
                                       IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.468750000000000000E-01_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 3))
                                       ELSE
                                          TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.468750000000000000E-01_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 4))
                                       ENDIF
                                    ENDIF
                                 ELSE
                                    IF (X2 <= 0.156250000000000000E-01_dp) THEN
                                       TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       TG2 = (2*X2-0.156250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 5))
                                    ELSE
                                       TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       TG2 = (2*X2-0.468750000000000000E-01_dp)*0.640000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 6))
                                    ENDIF
                                 ENDIF
                              ELSE
                                 IF (X1 <= 0.625000000000000000E-01_dp) THEN
                                    IF (X2 <= 0.468750000000000000E-01_dp) THEN
                                       IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.781250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 7))
                                       ELSE
                                          TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.781250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 8))
                                       ENDIF
                                    ELSE
                                       IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.109375000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 9))
                                       ELSE
                                          TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.109375000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 10))
                                       ENDIF
                                    ENDIF
                                 ELSE
                                    IF (X2 <= 0.468750000000000000E-01_dp) THEN
                                       TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       TG2 = (2*X2-0.781250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 11))
                                    ELSE
                                       TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       TG2 = (2*X2-0.109375000000000000E+00_dp)*0.640000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 12))
                                    ENDIF
                                 ENDIF
                              ENDIF
                           ELSE
                              IF (X2 <= 0.312500000000000000E-01_dp) THEN
                                 IF (X1 <= 0.187500000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    TG2 = (2*X2-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 13))
                                 ELSE
                                    TG1 = (2*X1-0.437500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    TG2 = (2*X2-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 14))
                                 ENDIF
                              ELSE
                                 IF (X1 <= 0.187500000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    TG2 = (2*X2-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 15))
                                 ELSE
                                    TG1 = (2*X1-0.437500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    TG2 = (2*X2-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 16))
                                 ENDIF
                              ENDIF
                           ENDIF
                        ELSE
                           IF (X1 <= 0.125000000000000000E+00_dp) THEN
                              IF (X2 <= 0.937500000000000000E-01_dp) THEN
                                 IF (X1 <= 0.625000000000000000E-01_dp) THEN
                                    IF (X2 <= 0.781250000000000000E-01_dp) THEN
                                       IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.140625000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 17))
                                       ELSE
                                          TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.140625000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 18))
                                       ENDIF
                                    ELSE
                                       IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.171875000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 19))
                                       ELSE
                                          TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.171875000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 20))
                                       ENDIF
                                    ENDIF
                                 ELSE
                                    IF (X2 <= 0.781250000000000000E-01_dp) THEN
                                       TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       TG2 = (2*X2-0.140625000000000000E+00_dp)*0.640000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 21))
                                    ELSE
                                       TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       TG2 = (2*X2-0.171875000000000000E+00_dp)*0.640000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 22))
                                    ENDIF
                                 ENDIF
                              ELSE
                                 IF (X1 <= 0.625000000000000000E-01_dp) THEN
                                    IF (X2 <= 0.109375000000000000E+00_dp) THEN
                                       IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.203125000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 23))
                                       ELSE
                                          TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.203125000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 24))
                                       ENDIF
                                    ELSE
                                       IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.234375000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 25))
                                       ELSE
                                          TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.234375000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 26))
                                       ENDIF
                                    ENDIF
                                 ELSE
                                    IF (X2 <= 0.109375000000000000E+00_dp) THEN
                                       TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       TG2 = (2*X2-0.203125000000000000E+00_dp)*0.640000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 27))
                                    ELSE
                                       TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       TG2 = (2*X2-0.234375000000000000E+00_dp)*0.640000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 28))
                                    ENDIF
                                 ENDIF
                              ENDIF
                           ELSE
                              IF (X1 <= 0.187500000000000000E+00_dp) THEN
                                 IF (X2 <= 0.937500000000000000E-01_dp) THEN
                                    TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    TG2 = (2*X2-0.156250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 29))
                                 ELSE
                                    TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    TG2 = (2*X2-0.218750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 30))
                                 ENDIF
                              ELSE
                                 TG1 = (2*X1-0.437500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 TG2 = (2*X2-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 31))
                              ENDIF
                           ENDIF
                        ENDIF
                     ELSE
                        IF (X1 <= 0.375000000000000000E+00_dp) THEN
                           TG1 = (2*X1-0.625000000000000000E+00_dp)*0.800000000000000000E+01_dp
                           TG2 = (2*X2-0.125000000000000000E+00_dp)*0.800000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 32))
                        ELSE
                           TG1 = (2*X1-0.875000000000000000E+00_dp)*0.800000000000000000E+01_dp
                           TG2 = (2*X2-0.125000000000000000E+00_dp)*0.800000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 33))
                        ENDIF
                     ENDIF
                  ELSE
                     IF (X1 <= 0.250000000000000000E+00_dp) THEN
                        IF (X2 <= 0.187500000000000000E+00_dp) THEN
                           IF (X1 <= 0.125000000000000000E+00_dp) THEN
                              IF (X2 <= 0.156250000000000000E+00_dp) THEN
                                 IF (X1 <= 0.625000000000000000E-01_dp) THEN
                                    IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                       IF (X2 <= 0.140625000000000000E+00_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.265625000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 34))
                                       ELSE
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.296875000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 35))
                                       ENDIF
                                    ELSE
                                       TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.281250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 36))
                                    ENDIF
                                 ELSE
                                    IF (X1 <= 0.937500000000000000E-01_dp) THEN
                                       TG1 = (2*X1-0.156250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.281250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 37))
                                    ELSE
                                       TG1 = (2*X1-0.218750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.281250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 38))
                                    ENDIF
                                 ENDIF
                              ELSE
                                 IF (X1 <= 0.625000000000000000E-01_dp) THEN
                                    IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                       IF (X2 <= 0.171875000000000000E+00_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.328125000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 39))
                                       ELSE
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.359375000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 40))
                                       ENDIF
                                    ELSE
                                       TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.343750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 41))
                                    ENDIF
                                 ELSE
                                    TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    TG2 = (2*X2-0.343750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 42))
                                 ENDIF
                              ENDIF
                           ELSE
                              IF (X1 <= 0.187500000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 TG2 = (2*X2-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 43))
                              ELSE
                                 TG1 = (2*X1-0.437500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 TG2 = (2*X2-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 44))
                              ENDIF
                           ENDIF
                        ELSE
                           IF (X1 <= 0.125000000000000000E+00_dp) THEN
                              IF (X1 <= 0.625000000000000000E-01_dp) THEN
                                 IF (X2 <= 0.218750000000000000E+00_dp) THEN
                                    IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                       IF (X2 <= 0.203125000000000000E+00_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.390625000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 45))
                                       ELSE
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.421875000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 46))
                                       ENDIF
                                    ELSE
                                       TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.406250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 47))
                                    ENDIF
                                 ELSE
                                    IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                       IF (X2 <= 0.234375000000000000E+00_dp) THEN
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.453125000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 48))
                                       ELSE
                                          TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                          TG2 = (2*X2-0.484375000000000000E+00_dp)*0.640000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 49))
                                       ENDIF
                                    ELSE
                                       TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.468750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 50))
                                    ENDIF
                                 ENDIF
                              ELSE
                                 IF (X2 <= 0.218750000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    TG2 = (2*X2-0.406250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 51))
                                 ELSE
                                    TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    TG2 = (2*X2-0.468750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 52))
                                 ENDIF
                              ENDIF
                           ELSE
                              IF (X1 <= 0.187500000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 TG2 = (2*X2-0.437500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 53))
                              ELSE
                                 TG1 = (2*X1-0.437500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 TG2 = (2*X2-0.437500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 54))
                              ENDIF
                           ENDIF
                        ENDIF
                     ELSE
                        IF (X1 <= 0.375000000000000000E+00_dp) THEN
                           IF (X1 <= 0.312500000000000000E+00_dp) THEN
                              TG1 = (2*X1-0.562500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.375000000000000000E+00_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 55))
                           ELSE
                              TG1 = (2*X1-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.375000000000000000E+00_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 56))
                           ENDIF
                        ELSE
                           TG1 = (2*X1-0.875000000000000000E+00_dp)*0.800000000000000000E+01_dp
                           TG2 = (2*X2-0.375000000000000000E+00_dp)*0.800000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 57))
                        ENDIF
                     ENDIF
                  ENDIF
               ELSE
                  IF (X1 <= 0.250000000000000000E+00_dp) THEN
                     IF (X1 <= 0.125000000000000000E+00_dp) THEN
                        IF (X1 <= 0.625000000000000000E-01_dp) THEN
                           IF (X2 <= 0.375000000000000000E+00_dp) THEN
                              IF (X2 <= 0.312500000000000000E+00_dp) THEN
                                 IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                    IF (X2 <= 0.281250000000000000E+00_dp) THEN
                                       TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.531250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 58))
                                    ELSE
                                       TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.593750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 59))
                                    ENDIF
                                 ELSE
                                    IF (X2 <= 0.281250000000000000E+00_dp) THEN
                                       TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.531250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 60))
                                    ELSE
                                       TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.593750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 61))
                                    ENDIF
                                 ENDIF
                              ELSE
                                 IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                    IF (X2 <= 0.343750000000000000E+00_dp) THEN
                                       TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.656250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 62))
                                    ELSE
                                       TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                       TG2 = (2*X2-0.718750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 63))
                                    ENDIF
                                 ELSE
                                    TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 64))
                                 ENDIF
                              ENDIF
                           ELSE
                              IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                 IF (X2 <= 0.437500000000000000E+00_dp) THEN
                                    IF (X1 <= 0.156250000000000000E-01_dp) THEN
                                       TG1 = (2*X1-0.156250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                       TG2 = (2*X2-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 65))
                                    ELSE
                                       TG1 = (2*X1-0.468750000000000000E-01_dp)*0.640000000000000000E+02_dp
                                       TG2 = (2*X2-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 66))
                                    ENDIF
                                 ELSE
                                    IF (X1 <= 0.156250000000000000E-01_dp) THEN
                                       TG1 = (2*X1-0.156250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                       TG2 = (2*X2-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 67))
                                    ELSE
                                       TG1 = (2*X1-0.468750000000000000E-01_dp)*0.640000000000000000E+02_dp
                                       TG2 = (2*X2-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 68))
                                    ENDIF
                                 ENDIF
                              ELSE
                                 IF (X2 <= 0.437500000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 69))
                                 ELSE
                                    TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 70))
                                 ENDIF
                              ENDIF
                           ENDIF
                        ELSE
                           IF (X2 <= 0.375000000000000000E+00_dp) THEN
                              IF (X2 <= 0.312500000000000000E+00_dp) THEN
                                 IF (X1 <= 0.937500000000000000E-01_dp) THEN
                                    TG1 = (2*X1-0.156250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.562500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 71))
                                 ELSE
                                    TG1 = (2*X1-0.218750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.562500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 72))
                                 ENDIF
                              ELSE
                                 IF (X1 <= 0.937500000000000000E-01_dp) THEN
                                    TG1 = (2*X1-0.156250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 73))
                                 ELSE
                                    TG1 = (2*X1-0.218750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 74))
                                 ENDIF
                              ENDIF
                           ELSE
                              IF (X1 <= 0.937500000000000000E-01_dp) THEN
                                 IF (X2 <= 0.437500000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.156250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 75))
                                 ELSE
                                    TG1 = (2*X1-0.156250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 76))
                                 ENDIF
                              ELSE
                                 IF (X2 <= 0.437500000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.218750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 77))
                                 ELSE
                                    TG1 = (2*X1-0.218750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 78))
                                 ENDIF
                              ENDIF
                           ENDIF
                        ENDIF
                     ELSE
                        IF (X2 <= 0.375000000000000000E+00_dp) THEN
                           IF (X1 <= 0.187500000000000000E+00_dp) THEN
                              IF (X2 <= 0.312500000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 TG2 = (2*X2-0.562500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 79))
                              ELSE
                                 IF (X1 <= 0.156250000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.281250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 80))
                                 ELSE
                                    TG1 = (2*X1-0.343750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 81))
                                 ENDIF
                              ENDIF
                           ELSE
                              IF (X2 <= 0.312500000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.437500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 TG2 = (2*X2-0.562500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 82))
                              ELSE
                                 TG1 = (2*X1-0.437500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 TG2 = (2*X2-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 83))
                              ENDIF
                           ENDIF
                        ELSE
                           IF (X1 <= 0.187500000000000000E+00_dp) THEN
                              IF (X2 <= 0.437500000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 TG2 = (2*X2-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 84))
                              ELSE
                                 IF (X1 <= 0.156250000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.281250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 85))
                                 ELSE
                                    TG1 = (2*X1-0.343750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 86))
                                 ENDIF
                              ENDIF
                           ELSE
                              IF (X1 <= 0.218750000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.406250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.875000000000000000E+00_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 87))
                              ELSE
                                 TG1 = (2*X1-0.468750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.875000000000000000E+00_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 88))
                              ENDIF
                           ENDIF
                        ENDIF
                     ENDIF
                  ELSE
                     IF (X1 <= 0.375000000000000000E+00_dp) THEN
                        IF (X2 <= 0.375000000000000000E+00_dp) THEN
                           IF (X1 <= 0.312500000000000000E+00_dp) THEN
                              TG1 = (2*X1-0.562500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.625000000000000000E+00_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 89))
                           ELSE
                              TG1 = (2*X1-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.625000000000000000E+00_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 90))
                           ENDIF
                        ELSE
                           IF (X1 <= 0.312500000000000000E+00_dp) THEN
                              IF (X1 <= 0.281250000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.531250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.875000000000000000E+00_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 91))
                              ELSE
                                 TG1 = (2*X1-0.593750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.875000000000000000E+00_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 92))
                              ENDIF
                           ELSE
                              TG1 = (2*X1-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.875000000000000000E+00_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 93))
                           ENDIF
                        ENDIF
                     ELSE
                        IF (X1 <= 0.437500000000000000E+00_dp) THEN
                           TG1 = (2*X1-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                           TG2 = (2*X2-0.750000000000000000E+00_dp)*0.400000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 94))
                        ELSE
                           TG1 = (2*X1-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                           TG2 = (2*X2-0.750000000000000000E+00_dp)*0.400000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 95))
                        ENDIF
                     ENDIF
                  ENDIF
               ENDIF
            ELSE
               IF (X1 <= 0.250000000000000000E+00_dp) THEN
                  IF (X1 <= 0.125000000000000000E+00_dp) THEN
                     IF (X1 <= 0.625000000000000000E-01_dp) THEN
                        IF (X1 <= 0.312500000000000000E-01_dp) THEN
                           IF (X1 <= 0.156250000000000000E-01_dp) THEN
                              IF (X1 <= 0.781250000000000000E-02_dp) THEN
                                 IF (X2 <= 0.750000000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.781250000000000000E-02_dp)*0.128000000000000000E+03_dp
                                    TG2 = (2*X2-0.125000000000000000E+01_dp)*0.400000000000000000E+01_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 96))
                                 ELSE
                                    TG1 = (2*X1-0.781250000000000000E-02_dp)*0.128000000000000000E+03_dp
                                    TG2 = (2*X2-0.175000000000000000E+01_dp)*0.400000000000000000E+01_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 97))
                                 ENDIF
                              ELSE
                                 IF (X2 <= 0.750000000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.234375000000000000E-01_dp)*0.128000000000000000E+03_dp
                                    TG2 = (2*X2-0.125000000000000000E+01_dp)*0.400000000000000000E+01_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 98))
                                 ELSE
                                    TG1 = (2*X1-0.234375000000000000E-01_dp)*0.128000000000000000E+03_dp
                                    TG2 = (2*X2-0.175000000000000000E+01_dp)*0.400000000000000000E+01_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 99))
                                 ENDIF
                              ENDIF
                           ELSE
                              IF (X2 <= 0.750000000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.468750000000000000E-01_dp)*0.640000000000000000E+02_dp
                                 TG2 = (2*X2-0.125000000000000000E+01_dp)*0.400000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 100))
                              ELSE
                                 TG1 = (2*X1-0.468750000000000000E-01_dp)*0.640000000000000000E+02_dp
                                 TG2 = (2*X2-0.175000000000000000E+01_dp)*0.400000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 101))
                              ENDIF
                           ENDIF
                        ELSE
                           IF (X2 <= 0.750000000000000000E+00_dp) THEN
                              IF (X2 <= 0.625000000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 102))
                              ELSE
                                 TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 103))
                              ENDIF
                           ELSE
                              IF (X1 <= 0.468750000000000000E-01_dp) THEN
                                 TG1 = (2*X1-0.781250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                 TG2 = (2*X2-0.175000000000000000E+01_dp)*0.400000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 104))
                              ELSE
                                 TG1 = (2*X1-0.109375000000000000E+00_dp)*0.640000000000000000E+02_dp
                                 TG2 = (2*X2-0.175000000000000000E+01_dp)*0.400000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 105))
                              ENDIF
                           ENDIF
                        ENDIF
                     ELSE
                        IF (X2 <= 0.750000000000000000E+00_dp) THEN
                           IF (X2 <= 0.625000000000000000E+00_dp) THEN
                              IF (X1 <= 0.937500000000000000E-01_dp) THEN
                                 TG1 = (2*X1-0.156250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 106))
                              ELSE
                                 TG1 = (2*X1-0.218750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 107))
                              ENDIF
                           ELSE
                              IF (X1 <= 0.937500000000000000E-01_dp) THEN
                                 TG1 = (2*X1-0.156250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 108))
                              ELSE
                                 TG1 = (2*X1-0.218750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 109))
                              ENDIF
                           ENDIF
                        ELSE
                           IF (X1 <= 0.937500000000000000E-01_dp) THEN
                              TG1 = (2*X1-0.156250000000000000E+00_dp)*0.320000000000000000E+02_dp
                              TG2 = (2*X2-0.175000000000000000E+01_dp)*0.400000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 110))
                           ELSE
                              TG1 = (2*X1-0.218750000000000000E+00_dp)*0.320000000000000000E+02_dp
                              TG2 = (2*X2-0.175000000000000000E+01_dp)*0.400000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 111))
                           ENDIF
                        ENDIF
                     ENDIF
                  ELSE
                     IF (X2 <= 0.750000000000000000E+00_dp) THEN
                        IF (X2 <= 0.625000000000000000E+00_dp) THEN
                           IF (X1 <= 0.187500000000000000E+00_dp) THEN
                              IF (X1 <= 0.156250000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.281250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 112))
                              ELSE
                                 TG1 = (2*X1-0.343750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 113))
                              ENDIF
                           ELSE
                              IF (X1 <= 0.218750000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.406250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 114))
                              ELSE
                                 TG1 = (2*X1-0.468750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 115))
                              ENDIF
                           ENDIF
                        ELSE
                           IF (X1 <= 0.187500000000000000E+00_dp) THEN
                              IF (X1 <= 0.156250000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.281250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 116))
                              ELSE
                                 TG1 = (2*X1-0.343750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 117))
                              ENDIF
                           ELSE
                              IF (X1 <= 0.218750000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.406250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 118))
                              ELSE
                                 TG1 = (2*X1-0.468750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 119))
                              ENDIF
                           ENDIF
                        ENDIF
                     ELSE
                        IF (X1 <= 0.187500000000000000E+00_dp) THEN
                           IF (X2 <= 0.875000000000000000E+00_dp) THEN
                              TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 120))
                           ELSE
                              TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.187500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 121))
                           ENDIF
                        ELSE
                           IF (X2 <= 0.875000000000000000E+00_dp) THEN
                              IF (X1 <= 0.218750000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.406250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 122))
                              ELSE
                                 TG1 = (2*X1-0.468750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 123))
                              ENDIF
                           ELSE
                              TG1 = (2*X1-0.437500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.187500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 124))
                           ENDIF
                        ENDIF
                     ENDIF
                  ENDIF
               ELSE
                  IF (X1 <= 0.375000000000000000E+00_dp) THEN
                     IF (X2 <= 0.750000000000000000E+00_dp) THEN
                        IF (X1 <= 0.312500000000000000E+00_dp) THEN
                           IF (X2 <= 0.625000000000000000E+00_dp) THEN
                              IF (X1 <= 0.281250000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.531250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 125))
                              ELSE
                                 TG1 = (2*X1-0.593750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 126))
                              ENDIF
                           ELSE
                              IF (X1 <= 0.281250000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.531250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 127))
                              ELSE
                                 TG1 = (2*X1-0.593750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 128))
                              ENDIF
                           ENDIF
                        ELSE
                           IF (X2 <= 0.625000000000000000E+00_dp) THEN
                              TG1 = (2*X1-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 129))
                           ELSE
                              IF (X1 <= 0.343750000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.656250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 130))
                              ELSE
                                 TG1 = (2*X1-0.718750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 131))
                              ENDIF
                           ENDIF
                        ENDIF
                     ELSE
                        IF (X1 <= 0.312500000000000000E+00_dp) THEN
                           IF (X2 <= 0.875000000000000000E+00_dp) THEN
                              IF (X1 <= 0.281250000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.531250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 132))
                              ELSE
                                 TG1 = (2*X1-0.593750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 133))
                              ENDIF
                           ELSE
                              IF (X1 <= 0.281250000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.531250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.187500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 134))
                              ELSE
                                 TG1 = (2*X1-0.593750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.187500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 135))
                              ENDIF
                           ENDIF
                        ELSE
                           IF (X2 <= 0.875000000000000000E+00_dp) THEN
                              IF (X1 <= 0.343750000000000000E+00_dp) THEN
                                 TG1 = (2*X1-0.656250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 136))
                              ELSE
                                 TG1 = (2*X1-0.718750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 137))
                              ENDIF
                           ELSE
                              TG1 = (2*X1-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.187500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 138))
                           ENDIF
                        ENDIF
                     ENDIF
                  ELSE
                     IF (X2 <= 0.750000000000000000E+00_dp) THEN
                        IF (X1 <= 0.437500000000000000E+00_dp) THEN
                           IF (X2 <= 0.625000000000000000E+00_dp) THEN
                              TG1 = (2*X1-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 139))
                           ELSE
                              TG1 = (2*X1-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 140))
                           ENDIF
                        ELSE
                           TG1 = (2*X1-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                           TG2 = (2*X2-0.125000000000000000E+01_dp)*0.400000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 141))
                        ENDIF
                     ELSE
                        IF (X1 <= 0.437500000000000000E+00_dp) THEN
                           IF (X2 <= 0.875000000000000000E+00_dp) THEN
                              TG1 = (2*X1-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 142))
                           ELSE
                              TG1 = (2*X1-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.187500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 143))
                           ENDIF
                        ELSE
                           IF (X2 <= 0.875000000000000000E+00_dp) THEN
                              TG1 = (2*X1-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 144))
                           ELSE
                              TG1 = (2*X1-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.187500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 145))
                           ENDIF
                        ENDIF
                     ENDIF
                  ENDIF
               ENDIF
            ENDIF
         ELSE
            IF (X1 <= 0.750000000000000000E+00_dp) THEN
               IF (X2 <= 0.500000000000000000E+00_dp) THEN
                  IF (X1 <= 0.625000000000000000E+00_dp) THEN
                     IF (X2 <= 0.250000000000000000E+00_dp) THEN
                        TG1 = (2*X1-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                        TG2 = (2*X2-0.250000000000000000E+00_dp)*0.400000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 146))
                     ELSE
                        TG1 = (2*X1-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                        TG2 = (2*X2-0.750000000000000000E+00_dp)*0.400000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 147))
                     ENDIF
                  ELSE
                     TG1 = (2*X1-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                     TG2 = (2*X2-0.500000000000000000E+00_dp)*0.200000000000000000E+01_dp
                     CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 148))
                  ENDIF
               ELSE
                  IF (X1 <= 0.625000000000000000E+00_dp) THEN
                     IF (X2 <= 0.750000000000000000E+00_dp) THEN
                        IF (X1 <= 0.562500000000000000E+00_dp) THEN
                           TG1 = (2*X1-0.106250000000000000E+01_dp)*0.160000000000000000E+02_dp
                           TG2 = (2*X2-0.125000000000000000E+01_dp)*0.400000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 149))
                        ELSE
                           TG1 = (2*X1-0.118750000000000000E+01_dp)*0.160000000000000000E+02_dp
                           TG2 = (2*X2-0.125000000000000000E+01_dp)*0.400000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 150))
                        ENDIF
                     ELSE
                        IF (X1 <= 0.562500000000000000E+00_dp) THEN
                           TG1 = (2*X1-0.106250000000000000E+01_dp)*0.160000000000000000E+02_dp
                           TG2 = (2*X2-0.175000000000000000E+01_dp)*0.400000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 151))
                        ELSE
                           TG1 = (2*X1-0.118750000000000000E+01_dp)*0.160000000000000000E+02_dp
                           TG2 = (2*X2-0.175000000000000000E+01_dp)*0.400000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 152))
                        ENDIF
                     ENDIF
                  ELSE
                     IF (X1 <= 0.687500000000000000E+00_dp) THEN
                        TG1 = (2*X1-0.131250000000000000E+01_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.150000000000000000E+01_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 153))
                     ELSE
                        TG1 = (2*X1-0.143750000000000000E+01_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.150000000000000000E+01_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 154))
                     ENDIF
                  ENDIF
               ENDIF
            ELSE
               TG1 = (2*X1-0.175000000000000000E+01_dp)*0.400000000000000000E+01_dp
               TG2 = (2*X2-0.100000000000000000E+01_dp)*0.100000000000000000E+01_dp
               CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 155))
            ENDIF
         ENDIF
      ELSE
         IF (T < lower) THEN
            use_gamma = .TRUE.
            RETURN
         ENDIF
         X2 = 11.0_dp/R
         X1 = (T-lower)/(upper-lower)
         IF (X1 <= 0.500000000000000000E+00_dp) THEN
            IF (X1 <= 0.250000000000000000E+00_dp) THEN
               IF (X2 <= 0.500000000000000000E+00_dp) THEN
                  IF (X1 <= 0.125000000000000000E+00_dp) THEN
                     IF (X2 <= 0.250000000000000000E+00_dp) THEN
                        TG1 = (2*X1-0.125000000000000000E+00_dp)*0.800000000000000000E+01_dp
                        TG2 = (2*X2-0.250000000000000000E+00_dp)*0.400000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 156))
                     ELSE
                        TG1 = (2*X1-0.125000000000000000E+00_dp)*0.800000000000000000E+01_dp
                        TG2 = (2*X2-0.750000000000000000E+00_dp)*0.400000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 157))
                     ENDIF
                  ELSE
                     TG1 = (2*X1-0.375000000000000000E+00_dp)*0.800000000000000000E+01_dp
                     TG2 = (2*X2-0.500000000000000000E+00_dp)*0.200000000000000000E+01_dp
                     CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 158))
                  ENDIF
               ELSE
                  IF (X1 <= 0.125000000000000000E+00_dp) THEN
                     IF (X2 <= 0.750000000000000000E+00_dp) THEN
                        IF (X2 <= 0.625000000000000000E+00_dp) THEN
                           TG1 = (2*X1-0.125000000000000000E+00_dp)*0.800000000000000000E+01_dp
                           TG2 = (2*X2-0.112500000000000000E+01_dp)*0.800000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 159))
                        ELSE
                           IF (X1 <= 0.625000000000000000E-01_dp) THEN
                              TG1 = (2*X1-0.625000000000000000E-01_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 160))
                           ELSE
                              TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 161))
                           ENDIF
                        ENDIF
                     ELSE
                        IF (X1 <= 0.625000000000000000E-01_dp) THEN
                           IF (X2 <= 0.875000000000000000E+00_dp) THEN
                              IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                 IF (X2 <= 0.812500000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.156250000000000000E+01_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 162))
                                 ELSE
                                    TG1 = (2*X1-0.312500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.168750000000000000E+01_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 163))
                                 ENDIF
                              ELSE
                                 TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 164))
                              ENDIF
                           ELSE
                              IF (X1 <= 0.312500000000000000E-01_dp) THEN
                                 IF (X2 <= 0.937500000000000000E+00_dp) THEN
                                    IF (X1 <= 0.156250000000000000E-01_dp) THEN
                                       IF (X2 <= 0.906250000000000000E+00_dp) THEN
                                          TG1 = (2*X1-0.156250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                          TG2 = (2*X2-0.178125000000000000E+01_dp)*0.320000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 165))
                                       ELSE
                                          TG1 = (2*X1-0.156250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                          TG2 = (2*X2-0.184375000000000000E+01_dp)*0.320000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 166))
                                       ENDIF
                                    ELSE
                                       TG1 = (2*X1-0.468750000000000000E-01_dp)*0.640000000000000000E+02_dp
                                       TG2 = (2*X2-0.181250000000000000E+01_dp)*0.160000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 167))
                                    ENDIF
                                 ELSE
                                    IF (X1 <= 0.156250000000000000E-01_dp) THEN
                                       IF (X2 <= 0.968750000000000000E+00_dp) THEN
                                          IF (X1 <= 0.781250000000000000E-02_dp) THEN
                                             IF (X2 <= 0.953125000000000000E+00_dp) THEN
                                                TG1 = (2*X1-0.781250000000000000E-02_dp)*0.128000000000000000E+03_dp
                                                TG2 = (2*X2-0.189062500000000000E+01_dp)*0.640000000000000000E+02_dp
                                                CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 168))
                                             ELSE
                                                TG1 = (2*X1-0.781250000000000000E-02_dp)*0.128000000000000000E+03_dp
                                                TG2 = (2*X2-0.192187500000000000E+01_dp)*0.640000000000000000E+02_dp
                                                CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 169))
                                             ENDIF
                                          ELSE
                                             TG1 = (2*X1-0.234375000000000000E-01_dp)*0.128000000000000000E+03_dp
                                             TG2 = (2*X2-0.190625000000000000E+01_dp)*0.320000000000000000E+02_dp
                                             CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 170))
                                          ENDIF
                                       ELSE
                                          IF (X1 <= 0.781250000000000000E-02_dp) THEN
                                             IF (X2 <= 0.984375000000000000E+00_dp) THEN
                                                TG1 = (2*X1-0.781250000000000000E-02_dp)*0.128000000000000000E+03_dp
                                                TG2 = (2*X2-0.195312500000000000E+01_dp)*0.640000000000000000E+02_dp
                                                CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 171))
                                             ELSE
                                                TG1 = (2*X1-0.781250000000000000E-02_dp)*0.128000000000000000E+03_dp
                                                TG2 = (2*X2-0.198437500000000000E+01_dp)*0.640000000000000000E+02_dp
                                                CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 172))
                                             ENDIF
                                          ELSE
                                             IF (X2 <= 0.984375000000000000E+00_dp) THEN
                                                TG1 = (2*X1-0.234375000000000000E-01_dp)*0.128000000000000000E+03_dp
                                                TG2 = (2*X2-0.195312500000000000E+01_dp)*0.640000000000000000E+02_dp
                                                CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 173))
                                             ELSE
                                                TG1 = (2*X1-0.234375000000000000E-01_dp)*0.128000000000000000E+03_dp
                                                TG2 = (2*X2-0.198437500000000000E+01_dp)*0.640000000000000000E+02_dp
                                                CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 174))
                                             ENDIF
                                          ENDIF
                                       ENDIF
                                    ELSE
                                       IF (X2 <= 0.968750000000000000E+00_dp) THEN
                                          TG1 = (2*X1-0.468750000000000000E-01_dp)*0.640000000000000000E+02_dp
                                          TG2 = (2*X2-0.190625000000000000E+01_dp)*0.320000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 175))
                                       ELSE
                                          IF (X1 <= 0.234375000000000000E-01_dp) THEN
                                             TG1 = (2*X1-0.390625000000000000E-01_dp)*0.128000000000000000E+03_dp
                                             TG2 = (2*X2-0.196875000000000000E+01_dp)*0.320000000000000000E+02_dp
                                             CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 176))
                                          ELSE
                                             TG1 = (2*X1-0.546875000000000000E-01_dp)*0.128000000000000000E+03_dp
                                             TG2 = (2*X2-0.196875000000000000E+01_dp)*0.320000000000000000E+02_dp
                                             CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 177))
                                          ENDIF
                                       ENDIF
                                    ENDIF
                                 ENDIF
                              ELSE
                                 IF (X2 <= 0.937500000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.937500000000000000E-01_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.181250000000000000E+01_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 178))
                                 ELSE
                                    IF (X1 <= 0.468750000000000000E-01_dp) THEN
                                       IF (X2 <= 0.968750000000000000E+00_dp) THEN
                                          TG1 = (2*X1-0.781250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                          TG2 = (2*X2-0.190625000000000000E+01_dp)*0.320000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 179))
                                       ELSE
                                          TG1 = (2*X1-0.781250000000000000E-01_dp)*0.640000000000000000E+02_dp
                                          TG2 = (2*X2-0.196875000000000000E+01_dp)*0.320000000000000000E+02_dp
                                          CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 180))
                                       ENDIF
                                    ELSE
                                       TG1 = (2*X1-0.109375000000000000E+00_dp)*0.640000000000000000E+02_dp
                                       TG2 = (2*X2-0.193750000000000000E+01_dp)*0.160000000000000000E+02_dp
                                       CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 181))
                                    ENDIF
                                 ENDIF
                              ENDIF
                           ENDIF
                        ELSE
                           IF (X2 <= 0.875000000000000000E+00_dp) THEN
                              TG1 = (2*X1-0.187500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 182))
                           ELSE
                              IF (X1 <= 0.937500000000000000E-01_dp) THEN
                                 IF (X2 <= 0.937500000000000000E+00_dp) THEN
                                    TG1 = (2*X1-0.156250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.181250000000000000E+01_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 183))
                                 ELSE
                                    TG1 = (2*X1-0.156250000000000000E+00_dp)*0.320000000000000000E+02_dp
                                    TG2 = (2*X2-0.193750000000000000E+01_dp)*0.160000000000000000E+02_dp
                                    CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 184))
                                 ENDIF
                              ELSE
                                 TG1 = (2*X1-0.218750000000000000E+00_dp)*0.320000000000000000E+02_dp
                                 TG2 = (2*X2-0.187500000000000000E+01_dp)*0.800000000000000000E+01_dp
                                 CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 185))
                              ENDIF
                           ENDIF
                        ENDIF
                     ENDIF
                  ELSE
                     IF (X2 <= 0.750000000000000000E+00_dp) THEN
                        TG1 = (2*X1-0.375000000000000000E+00_dp)*0.800000000000000000E+01_dp
                        TG2 = (2*X2-0.125000000000000000E+01_dp)*0.400000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 186))
                     ELSE
                        IF (X1 <= 0.187500000000000000E+00_dp) THEN
                           IF (X2 <= 0.875000000000000000E+00_dp) THEN
                              TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 187))
                           ELSE
                              TG1 = (2*X1-0.312500000000000000E+00_dp)*0.160000000000000000E+02_dp
                              TG2 = (2*X2-0.187500000000000000E+01_dp)*0.800000000000000000E+01_dp
                              CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 188))
                           ENDIF
                        ELSE
                           TG1 = (2*X1-0.437500000000000000E+00_dp)*0.160000000000000000E+02_dp
                           TG2 = (2*X2-0.175000000000000000E+01_dp)*0.400000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 189))
                        ENDIF
                     ENDIF
                  ENDIF
               ENDIF
            ELSE
               IF (X1 <= 0.375000000000000000E+00_dp) THEN
                  IF (X1 <= 0.312500000000000000E+00_dp) THEN
                     IF (X1 <= 0.281250000000000000E+00_dp) THEN
                        TG1 = (2*X1-0.531250000000000000E+00_dp)*0.320000000000000000E+02_dp
                        TG2 = (2*X2-0.100000000000000000E+01_dp)*0.100000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 190))
                     ELSE
                        TG1 = (2*X1-0.593750000000000000E+00_dp)*0.320000000000000000E+02_dp
                        TG2 = (2*X2-0.100000000000000000E+01_dp)*0.100000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 191))
                     ENDIF
                  ELSE
                     IF (X2 <= 0.500000000000000000E+00_dp) THEN
                        TG1 = (2*X1-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.500000000000000000E+00_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 192))
                     ELSE
                        TG1 = (2*X1-0.687500000000000000E+00_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.150000000000000000E+01_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 193))
                     ENDIF
                  ENDIF
               ELSE
                  IF (X1 <= 0.437500000000000000E+00_dp) THEN
                     IF (X2 <= 0.500000000000000000E+00_dp) THEN
                        TG1 = (2*X1-0.812500000000000000E+00_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.500000000000000000E+00_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 194))
                     ELSE
                        IF (X1 <= 0.406250000000000000E+00_dp) THEN
                           TG1 = (2*X1-0.781250000000000000E+00_dp)*0.320000000000000000E+02_dp
                           TG2 = (2*X2-0.150000000000000000E+01_dp)*0.200000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 195))
                        ELSE
                           TG1 = (2*X1-0.843750000000000000E+00_dp)*0.320000000000000000E+02_dp
                           TG2 = (2*X2-0.150000000000000000E+01_dp)*0.200000000000000000E+01_dp
                           CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 196))
                        ENDIF
                     ENDIF
                  ELSE
                     IF (X2 <= 0.500000000000000000E+00_dp) THEN
                        TG1 = (2*X1-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.500000000000000000E+00_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 197))
                     ELSE
                        TG1 = (2*X1-0.937500000000000000E+00_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.150000000000000000E+01_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 198))
                     ENDIF
                  ENDIF
               ENDIF
            ENDIF
         ELSE
            IF (X1 <= 0.750000000000000000E+00_dp) THEN
               IF (X1 <= 0.625000000000000000E+00_dp) THEN
                  IF (X2 <= 0.500000000000000000E+00_dp) THEN
                     IF (X1 <= 0.562500000000000000E+00_dp) THEN
                        TG1 = (2*X1-0.106250000000000000E+01_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.500000000000000000E+00_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 199))
                     ELSE
                        TG1 = (2*X1-0.118750000000000000E+01_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.500000000000000000E+00_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 200))
                     ENDIF
                  ELSE
                     IF (X1 <= 0.562500000000000000E+00_dp) THEN
                        TG1 = (2*X1-0.106250000000000000E+01_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.150000000000000000E+01_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 201))
                     ELSE
                        TG1 = (2*X1-0.118750000000000000E+01_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.150000000000000000E+01_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 202))
                     ENDIF
                  ENDIF
               ELSE
                  IF (X2 <= 0.500000000000000000E+00_dp) THEN
                     IF (X1 <= 0.687500000000000000E+00_dp) THEN
                        TG1 = (2*X1-0.131250000000000000E+01_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.500000000000000000E+00_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 203))
                     ELSE
                        TG1 = (2*X1-0.143750000000000000E+01_dp)*0.160000000000000000E+02_dp
                        TG2 = (2*X2-0.500000000000000000E+00_dp)*0.200000000000000000E+01_dp
                        CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 204))
                     ENDIF
                  ELSE
                     TG1 = (2*X1-0.137500000000000000E+01_dp)*0.800000000000000000E+01_dp
                     TG2 = (2*X2-0.150000000000000000E+01_dp)*0.200000000000000000E+01_dp
                     CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 205))
                  ENDIF
               ENDIF
            ELSE
               IF (X1 <= 0.875000000000000000E+00_dp) THEN
                  TG1 = (2*X1-0.162500000000000000E+01_dp)*0.800000000000000000E+01_dp
                  TG2 = (2*X2-0.100000000000000000E+01_dp)*0.100000000000000000E+01_dp
                  CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 206))
               ELSE
                  TG1 = (2*X1-0.187500000000000000E+01_dp)*0.800000000000000000E+01_dp
                  TG2 = (2*X2-0.100000000000000000E+01_dp)*0.100000000000000000E+01_dp
                  CALL PD2VAL(RES, NDERIV, TG1, TG2, C0(1, 207))
               ENDIF
            ENDIF
         ENDIF
      ENDIF
   END SUBROUTINE t_c_g0_n
! iunit contains the data file to initialize the table
! Nder is the number of derivatives that will actually be used
! **************************************************************************************************
!> \brief ...
!> \param Nder ...
!> \param iunit ...
!> \param mepos ...
!> \param group ...
! **************************************************************************************************
   SUBROUTINE init(Nder, iunit, mepos, group)
      INTEGER, INTENT(IN)                                :: Nder, iunit, mepos, group

      INTEGER                                            :: I
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: chunk

      patches = 207
      IF (Nder > nderiv_max) CALL cp__b("t_c_g0.F",1353,"T_C_G0 init failed")
      nderiv_init = Nder
      IF (ALLOCATED(C0)) DEALLOCATE (C0)
      ! round up to a multiple of 32 to give some generous alignment for each C0
      ALLOCATE (C0(32*((31+(Nder+1)*(degree+1)*(degree+2)/2)/32), patches))
      ! init mpi'ed buffers to silence warnings under valgrind
      C0 = 1.0E99_dp
      IF (mepos == 0) THEN
         ALLOCATE (chunk((nderiv_max+1)*(degree+1)*(degree+2)/2))
         DO I = 1, patches
            READ (iunit, *) chunk
            C0(1:(Nder+1)*(degree+1)*(degree+2)/2, I) = chunk(1:(Nder+1)*(degree+1)*(degree+2)/2)
         ENDDO
         DEALLOCATE (chunk)
      END IF
      CALL mp_bcast(C0, 0, group)

   END SUBROUTINE init
! **************************************************************************************************
!> \brief ...
! **************************************************************************************************
   SUBROUTINE free_C0()
      IF (ALLOCATED(C0)) DEALLOCATE (C0)
      nderiv_init = -1
   END SUBROUTINE free_C0
! **************************************************************************************************
!> \brief ...
!> \param RES ...
!> \param NDERIV ...
!> \param TG1 ...
!> \param TG2 ...
!> \param C0 ...
! **************************************************************************************************
   SUBROUTINE PD2VAL(RES, NDERIV, TG1, TG2, C0)
      REAL(KIND=dp), INTENT(OUT)                         :: res(*)
      INTEGER, INTENT(IN)                                :: NDERIV
      REAL(KIND=dp), INTENT(IN)                          :: TG1, TG2, C0(105, *)

      REAL(KIND=dp), PARAMETER :: SQRT2 = 1.4142135623730950488016887242096980785696718753_dp

      INTEGER                                            :: K
      REAL(KIND=dp)                                      :: T1(0:13), T2(0:13)

      T1(0) = 1.0_dp
      T2(0) = 1.0_dp
      T1(1) = SQRT2*TG1
      T2(1) = SQRT2*TG2
      T1(2) = 2*TG1*T1(1)-SQRT2
      T2(2) = 2*TG2*T2(1)-SQRT2
      T1(3) = 2*TG1*T1(2)-T1(1)
      T2(3) = 2*TG2*T2(2)-T2(1)
      T1(4) = 2*TG1*T1(3)-T1(2)
      T2(4) = 2*TG2*T2(3)-T2(2)
      T1(5) = 2*TG1*T1(4)-T1(3)
      T2(5) = 2*TG2*T2(4)-T2(3)
      T1(6) = 2*TG1*T1(5)-T1(4)
      T2(6) = 2*TG2*T2(5)-T2(4)
      T1(7) = 2*TG1*T1(6)-T1(5)
      T2(7) = 2*TG2*T2(6)-T2(5)
      T1(8) = 2*TG1*T1(7)-T1(6)
      T2(8) = 2*TG2*T2(7)-T2(6)
      T1(9) = 2*TG1*T1(8)-T1(7)
      T2(9) = 2*TG2*T2(8)-T2(7)
      T1(10) = 2*TG1*T1(9)-T1(8)
      T2(10) = 2*TG2*T2(9)-T2(8)
      T1(11) = 2*TG1*T1(10)-T1(9)
      T2(11) = 2*TG2*T2(10)-T2(9)
      T1(12) = 2*TG1*T1(11)-T1(10)
      T2(12) = 2*TG2*T2(11)-T2(10)
      T1(13) = 2*TG1*T1(12)-T1(11)
      T2(13) = 2*TG2*T2(12)-T2(11)
      DO K = 1, NDERIV+1
         RES(K) = 0.0_dp
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:13), C0(1:14, K))*T2(0)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:12), C0(15:27, K))*T2(1)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:11), C0(28:39, K))*T2(2)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:10), C0(40:50, K))*T2(3)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:9), C0(51:60, K))*T2(4)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:8), C0(61:69, K))*T2(5)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:7), C0(70:77, K))*T2(6)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:6), C0(78:84, K))*T2(7)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:5), C0(85:90, K))*T2(8)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:4), C0(91:95, K))*T2(9)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:3), C0(96:99, K))*T2(10)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:2), C0(100:102, K))*T2(11)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:1), C0(103:104, K))*T2(12)
         RES(K) = RES(K)+DOT_PRODUCT(T1(0:0), C0(105:105, K))*T2(13)
      ENDDO
   END SUBROUTINE PD2VAL
END MODULE t_c_g0
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types and set/get functions for HFX
!> \par History
!>      04.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
MODULE hfx_types
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE bibliography,                    ONLY: cite_reference,&
                                              guidon2008,&
                                              guidon2009
   USE cell_types,                      ONLY: cell_type,&
                                              get_cell,&
                                              plane_distance,&
                                              scaled_to_real
   USE cp_control_types,                ONLY: dft_control_type
   USE cp_files,                        ONLY: close_file,&
                                              file_exists,&
                                              open_file
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE erf_fn,                          ONLY: erfc
   USE hfx_helpers,                     ONLY: count_cells_perd,&
                                              next_image_cell_perd
   USE hfx_libint_wrapper,              ONLY: initialize_libderiv,&
                                              initialize_libint,&
                                              terminate_libderiv,&
                                              terminate_libint
   USE hfx_libint_wrapper_types,        ONLY: lib_deriv,&
                                              lib_int,&
                                              prim_data_f_size
   USE input_constants,                 ONLY: &
        do_hfx_auto_shells, do_hfx_potential_coulomb, do_hfx_potential_gaussian, &
        do_hfx_potential_id, do_hfx_potential_long, do_hfx_potential_mix_cl, &
        do_hfx_potential_mix_cl_trunc, do_hfx_potential_mix_lg, do_hfx_potential_short, &
        do_hfx_potential_truncated, hfx_do_eval_energy, hfx_do_eval_forces
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: default_path_length,&
                                              dp,&
                                              int_8
   USE machine,                         ONLY: m_chdir,&
                                              m_getcwd
   USE mathconstants,                   ONLY: pi
   USE orbital_pointers,                ONLY: nco,&
                                              ncoset,&
                                              nso
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              get_qs_kind_set,&
                                              qs_kind_type
   USE string_utilities,                ONLY: compress
   USE t_c_g0,                          ONLY: free_C0

!$ USE OMP_LIB, ONLY: omp_get_max_threads, omp_get_thread_num, omp_get_num_threads

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: hfx_type, hfx_create, hfx_release, &
             hfx_set_distr_energy, &
             hfx_set_distr_forces, &
             hfx_cell_type, hfx_distribution, &
             hfx_potential_type, hfx_screening_type, &
             hfx_memory_type, hfx_load_balance_type, hfx_general_type, &
             hfx_container_type, hfx_cache_type, &
             hfx_basis_type, parse_memory_section, &
             hfx_init_container, &
             hfx_basis_info_type, hfx_screen_coeff_type, &
             hfx_reset_memory_usage_counter, pair_list_type, pair_list_element_type, &
             pair_set_list_type, hfx_p_kind, hfx_2D_map, hfx_pgf_list, &
             hfx_pgf_product_list, hfx_block_range_type, &
             alloc_containers, dealloc_containers, hfx_task_list_type, init_t_c_g0_lmax, &
             hfx_create_neighbor_cells, hfx_create_basis_types, hfx_release_basis_types


   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'hfx_types'
   INTEGER, PARAMETER, PUBLIC                 :: max_atom_block = 32
   INTEGER, PARAMETER, PUBLIC                 :: max_images = 27
   REAL(dp), PARAMETER, PUBLIC                :: log_zero = -1000.0_dp
   REAL(dp), PARAMETER, PUBLIC                :: powell_min_log = -20.0_dp
   REAL(KIND=dp), DIMENSION(0:10), &
      PARAMETER, PUBLIC                       :: mul_fact = (/1.0_dp, &
                                                              1.1781_dp, &
                                                              1.3333_dp, &
                                                              1.4726_dp, &
                                                              1.6000_dp, &
                                                              1.7181_dp, &
                                                              1.8286_dp, &
                                                              1.9328_dp, &
                                                              2.0317_dp, &
                                                              2.1261_dp, &
                                                              2.2165_dp/)

   INTEGER, SAVE                                         :: init_t_c_g0_lmax = -1

!***

! **************************************************************************************************
   TYPE hfx_potential_type
      INTEGER                                  :: potential_type !! 1/r/ erfc(wr)/r ...
      REAL(dp)                                 :: omega !! w
      REAL(dp)                                 :: scale_coulomb !! scaling factor for mixed potential
      REAL(dp)                                 :: scale_longrange !! scaling factor for mixed potential
      REAL(dp)                                 :: scale_gaussian !! scaling factor for mixed potential
      REAL(dp)                                 :: cutoff_radius !! cutoff radius if cutoff potential in use
      CHARACTER(default_path_length)           :: filename
   END TYPE

! **************************************************************************************************
   TYPE hfx_screening_type
      REAL(dp)                                 :: eps_schwarz !! threshold
      REAL(dp)                                 :: eps_schwarz_forces !! threshold
      LOGICAL                                  :: do_p_screening_forces !! screen on P^2 ?
      LOGICAL                                  :: do_initial_p_screening !! screen on inital guess?
   END TYPE

! **************************************************************************************************
   TYPE hfx_memory_type
      INTEGER                                  :: max_memory !! user def max memory MiB
      INTEGER(int_8)                           :: max_compression_counter !! corresponding number of reals
      INTEGER(int_8)                           :: final_comp_counter_energy
      LOGICAL                                  :: do_all_on_the_fly !! max mem == 0 ?
      REAL(dp)                                 :: eps_storage_scaling
      INTEGER                                  :: cache_size
      INTEGER                                  :: bits_max_val
      INTEGER                                  :: actual_memory_usage
      INTEGER                                  :: actual_memory_usage_disk
      INTEGER(int_8)                           :: max_compression_counter_disk
      LOGICAL                                  :: do_disk_storage
      CHARACTER(len=default_path_length)       :: storage_location
      INTEGER(int_8)                           :: ram_counter
      INTEGER(int_8)                           :: ram_counter_forces
      INTEGER(int_8)                           :: size_p_screen
      LOGICAL                                  :: treat_forces_in_core
      LOGICAL                                  :: recalc_forces
   END TYPE

! **************************************************************************************************
   TYPE hfx_periodic_type
      INTEGER                                  :: number_of_shells !! number of periodic image cells
      LOGICAL                                  :: do_periodic !! periodic ?
      INTEGER                                  :: perd(3) !! x,xy,xyz,...
      INTEGER                                  :: mode
      REAL(dp)                                 :: R_max_stress
      INTEGER                                  :: number_of_shells_from_input
   END TYPE

! **************************************************************************************************
   TYPE hfx_load_balance_type
      INTEGER                                  :: nbins
      INTEGER                                  :: block_size
      INTEGER                                  :: nblocks
      LOGICAL                                  :: rtp_redistribute
      LOGICAL                                  :: blocks_initialized
      LOGICAL                                  :: do_randomize
   END TYPE

! **************************************************************************************************
   TYPE hfx_general_type
      REAL(dp)                                 :: fraction !! for hybrids
      LOGICAL                                  :: treat_lsd_in_core
   END TYPE

! **************************************************************************************************
   TYPE hfx_cell_type
      REAL(dp)                                 :: cell(3)
      REAL(dp)                                 :: cell_r(3)
   END TYPE

! **************************************************************************************************
   TYPE hfx_distribution
      INTEGER(int_8)                           :: istart
      INTEGER(int_8)                           :: number_of_atom_quartets
      INTEGER(int_8)                           :: cost
      REAL(KIND=dp)                            :: time_first_scf
      REAL(KIND=dp)                            :: time_other_scf
      REAL(KIND=dp)                            :: time_forces
      INTEGER(int_8)                           :: ram_counter
   END TYPE

! **************************************************************************************************
   TYPE pair_list_element_type
      INTEGER, DIMENSION(2) :: pair
      INTEGER, DIMENSION(2) :: set_bounds
      INTEGER, DIMENSION(2) :: kind_pair
      REAL(KIND=dp)         :: r1(3), r2(3)
      REAL(KIND=dp)         :: dist2
   END TYPE

   ! **************************************************************************************************
   TYPE pair_set_list_type
      INTEGER, DIMENSION(2) :: pair
   END TYPE

! **************************************************************************************************
   TYPE pair_list_type
      TYPE(pair_list_element_type), DIMENSION(max_atom_block**2) :: elements
      INTEGER :: n_element
   END TYPE pair_list_type

! **************************************************************************************************
   TYPE hfx_compression_type
      INTEGER(int_8), DIMENSION(:), POINTER    :: compressed_data
      INTEGER(int_8)                           :: element_pointer
      INTEGER                                  :: bit_pointer
   END TYPE

! **************************************************************************************************
   TYPE hfx_cache_type
      INTEGER(int_8), DIMENSION(1024)    :: DATA
      INTEGER                                  :: element_counter
   END TYPE

! **************************************************************************************************
   TYPE hfx_container_node
      TYPE(hfx_container_node), POINTER        :: next, prev
      INTEGER(int_8), DIMENSION(1024)    :: DATA
   END TYPE

! **************************************************************************************************
   TYPE hfx_container_type
      TYPE(hfx_container_node), POINTER        :: first, current
      INTEGER                                  :: element_counter
      INTEGER(int_8)                           :: file_counter
      CHARACTER(LEN=5)                         :: desc
      INTEGER                                  :: unit
      CHARACTER(default_path_length)           :: filename
   END TYPE

! **************************************************************************************************
   TYPE hfx_basis_type
      INTEGER, DIMENSION(:), POINTER           :: lmax
      INTEGER, DIMENSION(:), POINTER           :: lmin
      INTEGER, DIMENSION(:), POINTER           :: npgf
      INTEGER                                  :: nset
      REAL(dp), DIMENSION(:, :), POINTER        :: zet
      INTEGER, DIMENSION(:), POINTER           :: nsgf
      INTEGER, DIMENSION(:, :), POINTER         :: first_sgf
      REAL(dp), DIMENSION(:, :), POINTER        :: sphi
      INTEGER                                  :: nsgf_total
      INTEGER, DIMENSION(:, :), POINTER         :: nl
      INTEGER, DIMENSION(:, :), POINTER         :: nsgfl
      INTEGER, DIMENSION(:), POINTER           :: nshell
      REAL(dp), DIMENSION(:, :, :, :), POINTER &
         :: sphi_ext
      REAL(dp), DIMENSION(:), POINTER          :: set_radius
      REAL(dp), DIMENSION(:, :), POINTER        :: pgf_radius
      REAL(dp)                                 :: kind_radius
   END TYPE

! **************************************************************************************************
   TYPE hfx_basis_info_type
      INTEGER                                  :: max_set
      INTEGER                                  :: max_sgf
      INTEGER                                  :: max_am
   END TYPE

! **************************************************************************************************
   TYPE hfx_screen_coeff_type
      REAL(dp)                                 :: x(2)
   END TYPE

! **************************************************************************************************
   TYPE hfx_p_kind
      REAL(dp), DIMENSION(:, :, :, :), POINTER    :: p_kind
   END TYPE

! **************************************************************************************************
   TYPE hfx_2D_map
      INTEGER, DIMENSION(:), POINTER           :: iatom_list
      INTEGER, DIMENSION(:), POINTER           :: jatom_list
   END TYPE

! **************************************************************************************************
   TYPE hfx_pgf_image
      REAL(dp)                                 :: ra(3), rb(3)
      REAL(dp)                                 :: rab2
      REAL(dp)                                 :: S1234
      REAL(dp)                                 :: P(3)
      REAL(dp)                                 :: R
      REAL(dp)                                 :: pgf_max
      REAL(dp), DIMENSION(3)                   :: bcell
   END TYPE

! **************************************************************************************************
   TYPE hfx_pgf_list
      TYPE(hfx_pgf_image), DIMENSION(:), POINTER &
         :: image_list
      INTEGER                                  :: nimages
      REAL(dp)                                 :: zetapzetb
      REAL(dp)                                 :: ZetaInv
      REAL(dp)                                 :: zeta, zetb
      INTEGER                                  :: ipgf, jpgf
   END TYPE

! **************************************************************************************************
   TYPE hfx_pgf_product_list
      REAL(dp)                                 :: ra(3), rb(3), rc(3), rd(3)
      REAL(dp)                                 :: ZetapEtaInv
      REAL(dp)                                 :: Rho, RhoInv
      REAL(dp)                                 :: P(3), Q(3), W(3)
      REAL(dp)                                 :: AB(3), CD(3)
      REAL(dp)                                 :: Fm(prim_data_f_size)
   END TYPE

! **************************************************************************************************
   TYPE hfx_block_range_type
      INTEGER        :: istart, iend
      INTEGER(int_8) :: cost
   END TYPE

! **************************************************************************************************
   TYPE hfx_task_list_type
      INTEGER                                  :: thread_id
      INTEGER                                  :: bin_id
      INTEGER(int_8)                           :: cost
   END TYPE

! **************************************************************************************************
!> \brief stores some data used in construction of Kohn-Sham matrix
!> \param potential_parameter stores information on the potential (1/r, erfc(wr)/r
!> \param screening_parameter stores screening infos such as epsilon
!> \param memory_parameter stores infos on memory used for in-core calculations
!> \param periodic_parameter stores information on how to apply pbc
!> \param load_balance_parameter contains infos for Monte Carlo simulated annealing
!> \param general_paramter at the moment stores the fraction of HF amount to be included
!> \param maxval_container stores the maxvals in compressed form
!> \param maxval_cache cache for maxvals in decompressed form
!> \param integral_containers 64 containers for compressed integrals
!> \param integral_caches 64 caches for decompressed integrals
!> \param neighbor_cells manages handling of periodic cells
!> \param distribution_energy stores information on parallelization of energy
!> \param distribution_forces stores information on parallelization of forces
!> \param initial_p stores the initial guess if requested
!> \param is_assoc_atomic_block reflects KS sparsity
!> \param number_of_p_entries Size of P matrix
!> \param n_rep_hf Number of HFX replicas
!> \param b_first_load_balance_x flag to indicate if it is enough just to update
!>        the distribution of the integrals
!> \param full_ks_x full ks matrices
!> \param lib libint type for eris
!> \param basis_info contains information for basis sets
!> \param screen_funct_coeffs_pgf pgf based near field screening coefficients
!> \param pair_dist_radii_pgf pgf based radii coefficients of pair distributions
!> \param screen_funct_coeffs_set set based near field screening coefficients
!> \param screen_funct_coeffs_kind kind based near field screening coefficients
!> \param screen_funct_is_initialized flag that indicates if the coefficients
!>        have already been fitted
!> \par History
!>      11.2006 created [Manuel Guidon]
!>      02.2009 completely rewritten due to new screening
!> \author Manuel Guidon
! **************************************************************************************************
   TYPE hfx_type
      TYPE(hfx_potential_type)                 :: potential_parameter
      TYPE(hfx_screening_type)                 :: screening_parameter
      TYPE(hfx_memory_type)                    :: memory_parameter
      TYPE(hfx_periodic_type)                  :: periodic_parameter
      TYPE(hfx_load_balance_type)              :: load_balance_parameter
      TYPE(hfx_general_type)                   :: general_parameter
      TYPE(hfx_container_type), DIMENSION(:), &
         POINTER        :: maxval_container
      TYPE(hfx_cache_type), DIMENSION(:), &
         POINTER            :: maxval_cache
      TYPE(hfx_container_type), DIMENSION(:, :), &
         POINTER        :: integral_containers
      TYPE(hfx_cache_type), DIMENSION(:, :), &
         POINTER            :: integral_caches
      TYPE(hfx_container_type), DIMENSION(:), &
         POINTER        :: maxval_container_forces
      TYPE(hfx_cache_type), DIMENSION(:), &
         POINTER            :: maxval_cache_forces
      TYPE(hfx_container_type), DIMENSION(:, :), &
         POINTER        :: integral_containers_forces
      TYPE(hfx_cache_type), DIMENSION(:, :), &
         POINTER            :: integral_caches_forces

      TYPE(hfx_container_type), POINTER        :: maxval_container_disk
      TYPE(hfx_cache_type)                     :: maxval_cache_disk
      TYPE(hfx_container_type), POINTER, &
         DIMENSION(:)  :: integral_containers_disk
      TYPE(hfx_cache_type)                     :: integral_caches_disk(64)

      TYPE(hfx_cell_type), DIMENSION(:), &
         POINTER                       :: neighbor_cells
      TYPE(hfx_distribution), DIMENSION(:), &
         POINTER         :: distribution_energy
      TYPE(hfx_distribution), DIMENSION(:), &
         POINTER         :: distribution_forces
      INTEGER, DIMENSION(:, :), POINTER         :: is_assoc_atomic_block
      INTEGER                                  :: number_of_p_entries
      TYPE(hfx_basis_type), DIMENSION(:), &
         POINTER           :: basis_parameter
      INTEGER                                  :: n_rep_hf
      LOGICAL                                  :: b_first_load_balance_energy, &
                                                  b_first_load_balance_forces
      REAL(dp), DIMENSION(:, :), POINTER        :: full_ks_alpha
      REAL(dp), DIMENSION(:, :), POINTER        :: full_ks_beta
      TYPE(lib_int)                            :: lib
      TYPE(hfx_basis_info_type)                :: basis_info
      TYPE(hfx_screen_coeff_type), &
         DIMENSION(:, :, :, :, :, :), POINTER     :: screen_funct_coeffs_pgf, &
                                                     pair_dist_radii_pgf
      TYPE(hfx_screen_coeff_type), &
         DIMENSION(:, :, :, :), POINTER         :: screen_funct_coeffs_set
      TYPE(hfx_screen_coeff_type), &
         DIMENSION(:, :), POINTER             :: screen_funct_coeffs_kind
      LOGICAL                                  :: screen_funct_is_initialized
      TYPE(hfx_p_kind), DIMENSION(:), POINTER  :: initial_p
      TYPE(hfx_p_kind), DIMENSION(:), POINTER  :: initial_p_forces
      INTEGER, DIMENSION(:), POINTER           :: map_atom_to_kind_atom
      TYPE(hfx_2D_map), DIMENSION(:), POINTER  :: map_atoms_to_cpus
      INTEGER, DIMENSION(:, :), POINTER         :: atomic_block_offset
      INTEGER, DIMENSION(:, :, :, :), POINTER     :: set_offset
      INTEGER, DIMENSION(:), POINTER           :: block_offset
      TYPE(hfx_block_range_type), DIMENSION(:), &
         POINTER      :: blocks
      TYPE(hfx_task_list_type), DIMENSION(:), &
         POINTER        :: task_list
      REAL(dp), DIMENSION(:, :), POINTER        :: pmax_atom, pmax_atom_forces
      TYPE(lib_int)                            :: lib_int
      TYPE(lib_deriv)                          :: lib_deriv
      REAL(dp), DIMENSION(:, :), POINTER        :: pmax_block
      LOGICAL, DIMENSION(:, :), POINTER         :: atomic_pair_list
      LOGICAL, DIMENSION(:, :), POINTER         :: atomic_pair_list_forces
   END TYPE hfx_type

CONTAINS

! **************************************************************************************************
!> \brief - This routine allocates and initializes all types in hfx_data
!> \param x_data contains all relevant data structures for hfx runs
!> \param para_env ...
!> \param hfx_section input section
!> \param natom , max_set: quantities that define some array sizes
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param dft_control ...
!> \param cell ...
!> \param do_exx ...
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
!> \note
!>      - All POINTERS and ALLOCATABLES are allocated, even if their size is
!>        unknown at invocation time
! **************************************************************************************************
   SUBROUTINE hfx_create(x_data, para_env, hfx_section, natom, atomic_kind_set, qs_kind_set, &
                         dft_control, cell, do_exx)
      TYPE(hfx_type), DIMENSION(:, :), POINTER           :: x_data
      TYPE(cp_para_env_type)                             :: para_env
      TYPE(section_vals_type), POINTER                   :: hfx_section
      INTEGER, INTENT(IN)                                :: natom
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(cell_type), POINTER                           :: cell
      LOGICAL, OPTIONAL                                  :: do_exx

      CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_create', routineP = moduleN//':'//routineN

      CHARACTER(LEN=512)                                 :: error_msg
      CHARACTER(LEN=default_path_length)                 :: char_val
      INTEGER :: handle, i, i_thread, iatom, ikind, int_val, irep, jkind, max_set, n_rep_hf, &
         n_threads, natom_a, natom_b, nkind, nseta, nsetb, pbc_shells, storage_id
      INTEGER, DIMENSION(:), POINTER                     :: atom2kind, kind_of
      LOGICAL                                            :: logic_val, my_do_exx
      REAL(dp)                                           :: real_val
      TYPE(hfx_type), POINTER                            :: actual_x_data
      TYPE(section_vals_type), POINTER                   :: hf_pbc_section, hf_sub_section

      CALL timeset(routineN, handle)

      CALL cite_reference(Guidon2008)
      CALL cite_reference(Guidon2009)
      !! There might be 2 hf sections
      CALL section_vals_get(hfx_section, n_repetition=n_rep_hf)
      n_threads = 1
!$    n_threads = omp_get_max_threads()
      ALLOCATE (x_data(n_rep_hf, n_threads))
      DO i_thread = 1, n_threads
         DO irep = 1, n_rep_hf
            actual_x_data => x_data(irep, i_thread)
            !! Get data from input file
            !!
            !! GENERAL params
            CALL section_vals_val_get(hfx_section, "FRACTION", r_val=real_val, i_rep_section=irep)
            actual_x_data%general_parameter%fraction = real_val
            actual_x_data%n_rep_hf = n_rep_hf

            NULLIFY (actual_x_data%map_atoms_to_cpus)

            CALL section_vals_val_get(hfx_section, "TREAT_LSD_IN_CORE", l_val=logic_val, i_rep_section=irep)
            actual_x_data%general_parameter%treat_lsd_in_core = logic_val

            !! MEMORY section
            hf_sub_section => section_vals_get_subs_vals(hfx_section, "MEMORY", i_rep_section=irep)
            CALL parse_memory_section(actual_x_data%memory_parameter, hf_sub_section, storage_id, i_thread, &
                                      n_threads, para_env, irep, skip_disk=.FALSE., skip_in_core_forces=.FALSE.)

            !! PERIODIC section
            hf_sub_section => section_vals_get_subs_vals(hfx_section, "PERIODIC", i_rep_section=irep)
            CALL section_vals_val_get(hf_sub_section, "NUMBER_OF_SHELLS", i_val=int_val)
            actual_x_data%periodic_parameter%number_of_shells = int_val
            actual_x_data%periodic_parameter%mode = int_val
            CALL get_cell(cell=cell, periodic=actual_x_data%periodic_parameter%perd)
            IF (SUM(actual_x_data%periodic_parameter%perd) == 0) THEN
               actual_x_data%periodic_parameter%do_periodic = .FALSE.
            ELSE
               actual_x_data%periodic_parameter%do_periodic = .TRUE.
            END IF

            !! SCREENING section
            hf_sub_section => section_vals_get_subs_vals(hfx_section, "SCREENING", i_rep_section=irep)
            CALL section_vals_val_get(hf_sub_section, "EPS_SCHWARZ", r_val=real_val)
            actual_x_data%screening_parameter%eps_schwarz = real_val
            CALL section_vals_val_get(hf_sub_section, "EPS_SCHWARZ_FORCES", r_val=real_val)
            actual_x_data%screening_parameter%eps_schwarz_forces = real_val
            CALL section_vals_val_get(hf_sub_section, "SCREEN_P_FORCES", l_val=logic_val)
            actual_x_data%screening_parameter%do_p_screening_forces = logic_val
            CALL section_vals_val_get(hf_sub_section, "SCREEN_ON_INITIAL_P", l_val=logic_val)
            actual_x_data%screening_parameter%do_initial_p_screening = logic_val
            actual_x_data%screen_funct_is_initialized = .FALSE.

            !! INTERACTION_POTENTIAL section
            hf_sub_section => section_vals_get_subs_vals(hfx_section, "INTERACTION_POTENTIAL", i_rep_section=irep)
            CALL section_vals_val_get(hf_sub_section, "POTENTIAL_TYPE", i_val=int_val)
            actual_x_data%potential_parameter%potential_type = int_val
            CALL section_vals_val_get(hf_sub_section, "OMEGA", r_val=real_val)
            actual_x_data%potential_parameter%omega = real_val
            CALL section_vals_val_get(hf_sub_section, "SCALE_COULOMB", r_val=real_val)
            actual_x_data%potential_parameter%scale_coulomb = real_val
            CALL section_vals_val_get(hf_sub_section, "SCALE_LONGRANGE", r_val=real_val)
            actual_x_data%potential_parameter%scale_longrange = real_val
            CALL section_vals_val_get(hf_sub_section, "SCALE_GAUSSIAN", r_val=real_val)
            actual_x_data%potential_parameter%scale_gaussian = real_val
            IF (actual_x_data%potential_parameter%potential_type == do_hfx_potential_truncated .OR. &
                actual_x_data%potential_parameter%potential_type == do_hfx_potential_mix_cl_trunc) THEN
               CALL section_vals_val_get(hf_sub_section, "CUTOFF_RADIUS", r_val=real_val)
               actual_x_data%potential_parameter%cutoff_radius = real_val
               CALL section_vals_val_get(hf_sub_section, "T_C_G_DATA", c_val=char_val)
               CALL compress(char_val, .TRUE.)
               ! ** Check if file is there
               IF (.NOT. file_exists(char_val)) THEN
                  WRITE (error_msg, '(A,A,A)') "Truncated hfx calculation requested. The file containing "// &
                     "the data could not be found at ", TRIM(char_val), " Please check T_C_G_DATA "// &
                     "in the INTERCATION_POTENTIAL section"
                  CALL cp__b("hfx_types.F",566,error_msg)
               ELSE
                  actual_x_data%potential_parameter%filename = char_val
               END IF
            END IF
            IF (actual_x_data%potential_parameter%potential_type == do_hfx_potential_short) THEN
               CALL erfc_cutoff(actual_x_data%screening_parameter%eps_schwarz, &
                                actual_x_data%potential_parameter%omega, &
                                actual_x_data%potential_parameter%cutoff_radius)
            END IF

            !! LOAD_BALANCE section
            hf_sub_section => section_vals_get_subs_vals(hfx_section, "LOAD_BALANCE", i_rep_section=irep)
            CALL section_vals_val_get(hf_sub_section, "NBINS", i_val=int_val)
            actual_x_data%load_balance_parameter%nbins = MAX(1, int_val)
            actual_x_data%load_balance_parameter%blocks_initialized = .FALSE.

            CALL section_vals_val_get(hf_sub_section, "RANDOMIZE", l_val=logic_val)
            actual_x_data%load_balance_parameter%do_randomize = logic_val

            actual_x_data%load_balance_parameter%rtp_redistribute = .FALSE.
            IF (ASSOCIATED(dft_control%rtp_control)) &
               actual_x_data%load_balance_parameter%rtp_redistribute = dft_control%rtp_control%hfx_redistribute

            CALL section_vals_val_get(hf_sub_section, "BLOCK_SIZE", i_val=int_val)
            ! negative values ask for a computed default
            IF (int_val <= 0) THEN
               ! this gives a reasonable number of blocks for binning, yet typically results in blocking.
               int_val = CEILING(0.1_dp*natom/ &
                                 REAL(actual_x_data%load_balance_parameter%nbins*n_threads*para_env%num_pe, KIND=dp)**(0.25_dp))
            ENDIF
            ! at least 1 atom per block, and avoid overly large blocks
            actual_x_data%load_balance_parameter%block_size = MIN(max_atom_block, MAX(1, int_val))

            my_do_exx = .FALSE.
            IF (PRESENT(do_exx)) THEN
               my_do_exx = do_exx
            END IF

            ! in the case of RPA exchange (postprocessing after SCF), we do not use ADMM
            IF (my_do_exx) THEN
               CALL hfx_create_basis_types(actual_x_data%basis_parameter, actual_x_data%basis_info, qs_kind_set, &
                                           .FALSE.)
            ELSE
               CALL hfx_create_basis_types(actual_x_data%basis_parameter, actual_x_data%basis_info, qs_kind_set, &
                                           dft_control%do_admm)
            END IF

!!**************************************************************************************************
!! **        !! ** This code writes the contraction routines
!! **        !! ** Very UGLY: BASIS_SET has to be 1 primitive and lmin=lmax=l. For g-functions
!! **        !! **
!! **        !! ** 1  4  4  1  1
!! **        !! **    1.0  1.0
!! **        !! **
!! **        k = max_am - 1
!! **        write(filename,'(A,I0,A)') "sphi",k+1,"a"
!! **        OPEN(UNIT=31415,FILE=filename)
!! **        DO i=ncoset(k)+1,SIZE(sphi_a,1)
!! **          DO j=1,SIZE(sphi_a,2)
!! **            IF( sphi_a(i,j) /= 0.0_dp) THEN
!! **              write(31415,'(A,I0,A,I0,A,I0,A,I0,A,I0,A)') "buffer1(i+imax*(",&
!! **                          j,&
!! **                          "-1)) = buffer1(i+imax*(",&
!! **                          j,&
!! **                          "-1)) + work(",&
!! **                          i-ncoset(k),&
!! **                          "+(i-1)*kmax) * sphi_a(",&
!! **                          i-ncoset(k),&
!! **                          ",",&
!! **                          j,&
!! **                          "+s_offset_a1)"
!! **            END IF
!! **          END DO
!! **        END DO
!! **        CLOSE(UNIT=31415)
!! **        write(filename,'(A,I0,A)') "sphi",k+1,"b"
!! **        OPEN(UNIT=31415,FILE=filename)
!! **        DO i=ncoset(k)+1,SIZE(sphi_a,1)
!! **          DO j=1,SIZE(sphi_a,2)
!! **            IF( sphi_a(i,j) /= 0.0_dp) THEN
!! **               write(31415,'(A,I0,A,I0,A,I0,A,I0,A,I0,A)') "buffer2(i+imax*(",&
!! **                          j,&
!! **                          "-1)) = buffer2(i+imax*(",&
!! **                          j,&
!! **                          "-1)) + buffer1(",&
!! **                          i-ncoset(k),&
!! **                          "+(i-1)*kmax) * sphi_b(",&
!! **                          i-ncoset(k),&
!! **                          ",",&
!! **                          j,&
!! **                          "+s_offset_b1)"
!! **
!! **            END IF
!! **          END DO
!! **        END DO
!! **        CLOSE(UNIT=31415)
!! **        write(filename,'(A,I0,A)') "sphi",k+1,"c"
!! **        OPEN(UNIT=31415,FILE=filename)
!! **        DO i=ncoset(k)+1,SIZE(sphi_a,1)
!! **          DO j=1,SIZE(sphi_a,2)
!! **            IF( sphi_a(i,j) /= 0.0_dp) THEN
!! **               write(31415,'(A,I0,A,I0,A,I0,A,I0,A,I0,A)') "buffer1(i+imax*(",&
!! **                          j,&
!! **                          "-1)) = buffer1(i+imax*(",&
!! **                          j,&
!! **                          "-1)) + buffer2(",&
!! **                          i-ncoset(k),&
!! **                          "+(i-1)*kmax) * sphi_c(",&
!! **                          i-ncoset(k),&
!! **                          ",",&
!! **                          j,&
!! **                          "+s_offset_c1)"
!! **
!! **            END IF
!! **          END DO
!! **        END DO
!! **        CLOSE(UNIT=31415)
!! **        write(filename,'(A,I0,A)') "sphi",k+1,"d"
!! **        OPEN(UNIT=31415,FILE=filename)
!! **        DO i=ncoset(k)+1,SIZE(sphi_a,1)
!! **          DO j=1,SIZE(sphi_a,2)
!! **            IF( sphi_a(i,j) /= 0.0_dp) THEN
!! **
!! **
!! **               write(31415,'(A,I0,A)') "primitives(s_offset_a1+i3, s_offset_b1+i2, s_offset_c1+i1, s_offset_d1+",&
!! **                           j,")= &"
!! **               write(31415,'(A,I0,A)') "primitives(s_offset_a1+i3, s_offset_b1+i2, s_offset_c1+i1, s_offset_d1+",&
!! **                           j,")+ &"
!! **               write(31415,'(A,I0,A,I0,A,I0,A)') "buffer1(",&
!! **                          i-ncoset(k),&
!! **                          "+(i-1)*kmax) * sphi_d(",&
!! **                          i-ncoset(k),&
!! **                          ",",&
!! **                          j,&
!! **                          "+s_offset_d1)"
!! **
!! **
!! **            END IF
!! **          END DO
!! **        END DO
!! **        CLOSE(UNIT=31415)
!! **        stop
!! *************************************************************************************************************************

            IF (actual_x_data%periodic_parameter%do_periodic) THEN
               hf_pbc_section => section_vals_get_subs_vals(hfx_section, "PERIODIC", i_rep_section=irep)
               CALL section_vals_val_get(hf_pbc_section, "NUMBER_OF_SHELLS", i_val=pbc_shells)
               actual_x_data%periodic_parameter%number_of_shells_from_input = pbc_shells
               ALLOCATE (actual_x_data%neighbor_cells(1))
               CALL hfx_create_neighbor_cells(actual_x_data, pbc_shells, cell, i_thread)
            ELSE
               ALLOCATE (actual_x_data%neighbor_cells(1))
               ! ** Initialize this guy to enable non periodic stress regtests
               actual_x_data%periodic_parameter%R_max_stress = 1.0_dp
            END IF

            nkind = SIZE(qs_kind_set, 1)
            max_set = actual_x_data%basis_info%max_set

            !! ** This guy is allocated on the master thread only
            IF (i_thread == 1) THEN
               ALLOCATE (actual_x_data%is_assoc_atomic_block(natom, natom))
               ALLOCATE (actual_x_data%atomic_block_offset(natom, natom))
               ALLOCATE (actual_x_data%set_offset(max_set, max_set, nkind, nkind))
               ALLOCATE (actual_x_data%block_offset(para_env%num_pe+1))
            END IF

            ALLOCATE (actual_x_data%distribution_forces(1))
            ALLOCATE (actual_x_data%distribution_energy(1))

            actual_x_data%memory_parameter%size_p_screen = 0_int_8
            IF (i_thread == 1) THEN
               ALLOCATE (actual_x_data%atomic_pair_list(natom, natom))
               ALLOCATE (actual_x_data%atomic_pair_list_forces(natom, natom))
            END IF

            IF (actual_x_data%screening_parameter%do_initial_p_screening .OR. &
                actual_x_data%screening_parameter%do_p_screening_forces) THEN
               !! ** This guy is allocated on the master thread only
               IF (i_thread == 1) THEN
                  ALLOCATE (actual_x_data%pmax_atom(natom, natom))
                  ALLOCATE (actual_x_data%initial_p(nkind*(nkind+1)/2))
                  i = 1
                  DO ikind = 1, nkind
                     CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom_a)
                     nseta = actual_x_data%basis_parameter(ikind)%nset
                     DO jkind = ikind, nkind
                        CALL get_atomic_kind(atomic_kind_set(jkind), natom=natom_b)
                        nsetb = actual_x_data%basis_parameter(jkind)%nset
                        ALLOCATE (actual_x_data%initial_p(i)%p_kind(nseta, nsetb, natom_a, natom_b))
                        actual_x_data%memory_parameter%size_p_screen = &
                           actual_x_data%memory_parameter%size_p_screen+nseta*nsetb*natom_a*natom_b
                        i = i+1
                     END DO
                  END DO
                  IF (actual_x_data%memory_parameter%treat_forces_in_core) THEN
                     ALLOCATE (actual_x_data%pmax_atom_forces(natom, natom))

                     ALLOCATE (actual_x_data%initial_p_forces(nkind*(nkind+1)/2))
                     i = 1
                     DO ikind = 1, nkind
                        CALL get_atomic_kind(atomic_kind_set(ikind), natom=natom_a)
                        nseta = actual_x_data%basis_parameter(ikind)%nset
                        DO jkind = ikind, nkind
                           CALL get_atomic_kind(atomic_kind_set(jkind), natom=natom_b)
                           nsetb = actual_x_data%basis_parameter(jkind)%nset
                           ALLOCATE (actual_x_data%initial_p_forces(i)%p_kind(nseta, nsetb, natom_a, natom_b))
                           actual_x_data%memory_parameter%size_p_screen = &
                              actual_x_data%memory_parameter%size_p_screen+nseta*nsetb*natom_a*natom_b
                           i = i+1
                        END DO
                     END DO
                  END IF
               END IF
               ALLOCATE (actual_x_data%map_atom_to_kind_atom(natom))
               ALLOCATE (kind_of(natom))
               ALLOCATE (atom2kind(nkind))
               CALL get_atomic_kind_set(atomic_kind_set, kind_of=kind_of)

               atom2kind = 0
               DO iatom = 1, natom
                  ikind = kind_of(iatom)
                  atom2kind(ikind) = atom2kind(ikind)+1
                  actual_x_data%map_atom_to_kind_atom(iatom) = atom2kind(ikind)
               END DO
               DEALLOCATE (kind_of, atom2kind)
            END IF

            ! ** Initialize libint type
            CALL initialize_libint(actual_x_data%lib, actual_x_data%basis_info%max_am)
            CALL initialize_libderiv(actual_x_data%lib_deriv, actual_x_data%basis_info%max_am)

            CALL alloc_containers(actual_x_data, 1, hfx_do_eval_energy)
            CALL alloc_containers(actual_x_data, 1, hfx_do_eval_forces)

            actual_x_data%maxval_cache_disk%element_counter = 1
            ALLOCATE (actual_x_data%maxval_container_disk)
            ALLOCATE (actual_x_data%maxval_container_disk%first)
            actual_x_data%maxval_container_disk%first%prev => NULL()
            actual_x_data%maxval_container_disk%first%next => NULL()
            actual_x_data%maxval_container_disk%current => actual_x_data%maxval_container_disk%first
            actual_x_data%maxval_container_disk%current%data = 0
            actual_x_data%maxval_container_disk%element_counter = 1
            actual_x_data%maxval_container_disk%file_counter = 1
            actual_x_data%maxval_container_disk%desc = 'Max_'
            actual_x_data%maxval_container_disk%unit = -1
            WRITE (actual_x_data%maxval_container_disk%filename, '(A,I0,A,A,A)') &
               TRIM(actual_x_data%memory_parameter%storage_location), &
               storage_id, "_", actual_x_data%maxval_container_disk%desc, "6"
            CALL compress(actual_x_data%maxval_container_disk%filename, .TRUE.)
            ALLOCATE (actual_x_data%integral_containers_disk(64))
            DO i = 1, 64
               actual_x_data%integral_caches_disk(i)%element_counter = 1
               actual_x_data%integral_caches_disk(i)%data = 0
               ALLOCATE (actual_x_data%integral_containers_disk(i)%first)
               actual_x_data%integral_containers_disk(i)%first%prev => NULL()
               actual_x_data%integral_containers_disk(i)%first%next => NULL()
               actual_x_data%integral_containers_disk(i)%current => actual_x_data%integral_containers_disk(i)%first
               actual_x_data%integral_containers_disk(i)%current%data = 0
               actual_x_data%integral_containers_disk(i)%element_counter = 1
               actual_x_data%integral_containers_disk(i)%file_counter = 1
               actual_x_data%integral_containers_disk(i)%desc = 'Int_'
               actual_x_data%integral_containers_disk(i)%unit = -1
               WRITE (actual_x_data%integral_containers_disk(i)%filename, '(A,I0,A,A,I0)') &
                  TRIM(actual_x_data%memory_parameter%storage_location), &
                  storage_id, "_", actual_x_data%integral_containers_disk(i)%desc, i
               CALL compress(actual_x_data%integral_containers_disk(i)%filename, .TRUE.)
            END DO

            actual_x_data%b_first_load_balance_energy = .TRUE.
            actual_x_data%b_first_load_balance_forces = .TRUE.
         END DO
      END DO

      DO irep = 1, n_rep_hf
         actual_x_data => x_data(irep, 1)
         CALL hfx_print_info(actual_x_data, hfx_section, irep)
      END DO
      CALL timestop(handle)
   END SUBROUTINE hfx_create

! **************************************************************************************************
!> \brief - compute a truncation radius for the shortrange operator
!> \param eps target accuracy
!> \param omg screening parameter
!> \param r_cutoff cutoff radius
!> \par History
!>      10.2012 created [Hossein Banihashemian]
!> \author Hossein Banihashemian
! **************************************************************************************************
   SUBROUTINE erfc_cutoff(eps, omg, r_cutoff)
      IMPLICIT NONE

      REAL(dp), INTENT(in)  :: eps, omg
      REAL(dp), INTENT(out) :: r_cutoff

      CHARACTER(LEN=*), PARAMETER :: routineN = 'erfc_cutoff', &
                                     routineP = moduleN//':'//routineN

      REAL(dp), PARAMETER :: abstol = 1E-10_dp, soltol = 1E-16_dp
      REAL(dp) :: r0, f0, fprime0, delta_r
      INTEGER :: iter, handle
      INTEGER, PARAMETER :: iterMAX = 1000

      CALL timeset(routineN, handle)

      ! initial guess assuming that we are in the asymptotic regime of the erf, and the solution is about 10.
      r0 = SQRT(-LOG(eps*omg*10**2))/omg
      CALL eval_transc_func(r0, eps, omg, f0, fprime0)

      DO iter = 1, iterMAX
         delta_r = f0/fprime0
         r0 = r0-delta_r
         CALL eval_transc_func(r0, eps, omg, f0, fprime0)
         IF (ABS(delta_r) .LT. abstol .OR. ABS(f0) .LT. soltol) EXIT
      END DO
      IF(.NOT.(iter <= itermax))CALL cp__a("hfx_types.F",883)
      r_cutoff = r0

      CALL timestop(handle)
   CONTAINS
! **************************************************************************************************
!> \brief ...
!> \param r ...
!> \param eps ...
!> \param omega ...
!> \param fn ...
!> \param df ...
! **************************************************************************************************
      SUBROUTINE eval_transc_func(r, eps, omega, fn, df)
      REAL(dp), INTENT(in)                               :: r, eps, omega
      REAL(dp), INTENT(out)                              :: fn, df

         fn = erfc(omega*r)-r*eps
         df = -omega*2*EXP(-omega**2*r**2)/SQRT(pi)-eps
      END SUBROUTINE eval_transc_func
   END SUBROUTINE erfc_cutoff
! **************************************************************************************************
!> \brief - This routine allocates and initializes the basis_info and basis_parameter types
!> \param basis_parameter ...
!> \param basis_info ...
!> \param qs_kind_set ...
!> \param do_admm ...
!> \par History
!>      07.2011 refactored
! **************************************************************************************************
   SUBROUTINE hfx_create_basis_types(basis_parameter, basis_info, qs_kind_set, do_admm)
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_parameter
      TYPE(hfx_basis_info_type)                          :: basis_info
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      LOGICAL                                            :: do_admm

      CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_create_basis_types', &
         routineP = moduleN//':'//routineN

      INTEGER :: co_counter, handle, i, ikind, ipgf, iset, j, k, la, max_am_kind, max_coeff, &
         max_nsgfl, max_pgf, max_pgf_kind, max_set, nkind, nl_count, nset, nseta, offset_a, &
         offset_a1, s_offset_nl_a, sgfa, so_counter
      INTEGER, DIMENSION(:), POINTER                     :: la_max, la_min, npgfa, nshell
      INTEGER, DIMENSION(:, :), POINTER                  :: first_sgfa, nl_a
      REAL(dp), DIMENSION(:, :), POINTER                 :: sphi_a
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_a

      CALL timeset(routineN, handle)

      !! BASIS parameter
      nkind = SIZE(qs_kind_set, 1)
      ALLOCATE (basis_parameter(nkind))
      max_set = 0
      DO ikind = 1, nkind
         IF (.NOT. do_admm) THEN
            CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_a, basis_type="ORB")
            CALL get_qs_kind_set(qs_kind_set, &
                                 maxsgf=basis_info%max_sgf, &
                                 maxnset=basis_info%max_set, &
                                 maxlgto=basis_info%max_am)

         ELSE
            CALL get_qs_kind(qs_kind_set(ikind), basis_set=orb_basis_a, basis_type="AUX_FIT")
            CALL get_qs_kind_set(qs_kind_set, &
                                 maxsgf=basis_info%max_sgf, &
                                 maxnset=basis_info%max_set, &
                                 maxlgto=basis_info%max_am, &
                                 basis_type="AUX_FIT")
         END IF
         IF (basis_info%max_set < max_set) CALL cp__b("hfx_types.F",952,"UNEXPECTED MAX_SET")
         max_set = MAX(max_set, basis_info%max_set)
         CALL get_gto_basis_set(gto_basis_set=orb_basis_a, &
                                lmax=basis_parameter(ikind)%lmax, &
                                lmin=basis_parameter(ikind)%lmin, &
                                npgf=basis_parameter(ikind)%npgf, &
                                nset=basis_parameter(ikind)%nset, &
                                zet=basis_parameter(ikind)%zet, &
                                nsgf_set=basis_parameter(ikind)%nsgf, &
                                first_sgf=basis_parameter(ikind)%first_sgf, &
                                sphi=basis_parameter(ikind)%sphi, &
                                nsgf=basis_parameter(ikind)%nsgf_total, &
                                l=basis_parameter(ikind)%nl, &
                                nshell=basis_parameter(ikind)%nshell, &
                                set_radius=basis_parameter(ikind)%set_radius, &
                                pgf_radius=basis_parameter(ikind)%pgf_radius, &
                                kind_radius=basis_parameter(ikind)%kind_radius)
      END DO
      DO ikind = 1, nkind
         ALLOCATE (basis_parameter(ikind)%nsgfl(0:basis_info%max_am, max_set))
         basis_parameter(ikind)%nsgfl = 0
         nset = basis_parameter(ikind)%nset
         nshell => basis_parameter(ikind)%nshell
         DO iset = 1, nset
            DO i = 0, basis_info%max_am
               nl_count = 0
               DO j = 1, nshell(iset)
                  IF (basis_parameter(ikind)%nl(j, iset) == i) nl_count = nl_count+1
               END DO
               basis_parameter(ikind)%nsgfl(i, iset) = nl_count
            END DO
         END DO
      END DO

      max_nsgfl = 0
      max_pgf = 0
      DO ikind = 1, nkind
         max_coeff = 0
         max_am_kind = 0
         max_pgf_kind = 0
         npgfa => basis_parameter(ikind)%npgf
         nseta = basis_parameter(ikind)%nset
         nl_a => basis_parameter(ikind)%nsgfl
         la_max => basis_parameter(ikind)%lmax
         la_min => basis_parameter(ikind)%lmin
         DO iset = 1, nseta
            max_pgf_kind = MAX(max_pgf_kind, npgfa(iset))
            max_pgf = MAX(max_pgf, npgfa(iset))
            DO la = la_min(iset), la_max(iset)
               max_nsgfl = MAX(max_nsgfl, nl_a(la, iset))
               max_coeff = MAX(max_coeff, nso(la)*nl_a(la, iset)*nco(la))
               max_am_kind = MAX(max_am_kind, la)
            END DO
         END DO
         ALLOCATE (basis_parameter(ikind)%sphi_ext(max_coeff, 0:max_am_kind, max_pgf_kind, nseta))
         basis_parameter(ikind)%sphi_ext = 0.0_dp
      END DO

      DO ikind = 1, nkind
         sphi_a => basis_parameter(ikind)%sphi
         nseta = basis_parameter(ikind)%nset
         la_max => basis_parameter(ikind)%lmax
         la_min => basis_parameter(ikind)%lmin
         npgfa => basis_parameter(ikind)%npgf
         first_sgfa => basis_parameter(ikind)%first_sgf
         nl_a => basis_parameter(ikind)%nsgfl
         DO iset = 1, nseta
            sgfa = first_sgfa(1, iset)
            DO ipgf = 1, npgfa(iset)
               offset_a1 = (ipgf-1)*ncoset(la_max(iset))
               s_offset_nl_a = 0
               DO la = la_min(iset), la_max(iset)
                  offset_a = offset_a1+ncoset(la-1)
                  co_counter = 0
                  co_counter = co_counter+1
                  so_counter = 0
                  DO k = sgfa+s_offset_nl_a, sgfa+s_offset_nl_a+nso(la)*nl_a(la, iset)-1
                     DO i = offset_a+1, offset_a+nco(la)
                        so_counter = so_counter+1
                        basis_parameter(ikind)%sphi_ext(so_counter, la, ipgf, iset) = sphi_a(i, k)
                     END DO
                  END DO
                  s_offset_nl_a = s_offset_nl_a+nso(la)*(nl_a(la, iset))
               END DO
            END DO
         END DO
      END DO

      CALL timestop(handle)

   END SUBROUTINE hfx_create_basis_types

! **************************************************************************************************
!> \brief ...
!> \param basis_parameter ...
! **************************************************************************************************
   SUBROUTINE hfx_release_basis_types(basis_parameter)
      TYPE(hfx_basis_type), DIMENSION(:), POINTER        :: basis_parameter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_release_basis_types', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)

      !! BASIS parameter
      DO i = 1, SIZE(basis_parameter)
         DEALLOCATE (basis_parameter(i)%nsgfl)
         DEALLOCATE (basis_parameter(i)%sphi_ext)
      END DO
      DEALLOCATE (basis_parameter)
      CALL timestop(handle)

   END SUBROUTINE hfx_release_basis_types

! **************************************************************************************************
!> \brief - Parses the memory section
!> \param memory_parameter ...
!> \param hf_sub_section ...
!> \param storage_id ...
!> \param i_thread ...
!> \param n_threads ...
!> \param para_env ...
!> \param irep ...
!> \param skip_disk ...
!> \param skip_in_core_forces ...
! **************************************************************************************************
   SUBROUTINE parse_memory_section(memory_parameter, hf_sub_section, storage_id, &
                                   i_thread, n_threads, para_env, irep, skip_disk, skip_in_core_forces)
      TYPE(hfx_memory_type)                              :: memory_parameter
      TYPE(section_vals_type), POINTER                   :: hf_sub_section
      INTEGER, INTENT(OUT), OPTIONAL                     :: storage_id
      INTEGER, INTENT(IN), OPTIONAL                      :: i_thread, n_threads
      TYPE(cp_para_env_type), OPTIONAL                   :: para_env
      INTEGER, INTENT(IN), OPTIONAL                      :: irep
      LOGICAL, INTENT(IN)                                :: skip_disk, skip_in_core_forces

      CHARACTER(LEN=*), PARAMETER :: routineN = 'parse_memory_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=512)                                 :: error_msg
      CHARACTER(LEN=default_path_length)                 :: char_val, filename, orig_wd
      INTEGER                                            :: int_val, stat
      LOGICAL                                            :: check, logic_val
      REAL(dp)                                           :: real_val

      check = (PRESENT(storage_id) .EQV. PRESENT(i_thread)) .AND. &
              (PRESENT(storage_id) .EQV. PRESENT(n_threads)) .AND. &
              (PRESENT(storage_id) .EQV. PRESENT(para_env)) .AND. &
              (PRESENT(storage_id) .EQV. PRESENT(irep))
      IF(.NOT.(check))CALL cp__a("hfx_types.F",1103)

      ! Memory Storage
      CALL section_vals_val_get(hf_sub_section, "MAX_MEMORY", i_val=int_val)
      memory_parameter%max_memory = int_val
      memory_parameter%max_compression_counter = int_val*1024_int_8*128_int_8
      CALL section_vals_val_get(hf_sub_section, "EPS_STORAGE", r_val=real_val)
      memory_parameter%eps_storage_scaling = real_val
      IF (int_val == 0) THEN
         memory_parameter%do_all_on_the_fly = .TRUE.
      ELSE
         memory_parameter%do_all_on_the_fly = .FALSE.
      END IF
      memory_parameter%cache_size = 1024
      memory_parameter%bits_max_val = 6
      memory_parameter%actual_memory_usage = 1
      IF (.NOT. skip_in_core_forces) THEN
         CALL section_vals_val_get(hf_sub_section, "TREAT_FORCES_IN_CORE", l_val=logic_val)
         memory_parameter%treat_forces_in_core = logic_val
      END IF

      ! ** IF MAX_MEM == 0 overwrite this flag to false
      IF (memory_parameter%do_all_on_the_fly) memory_parameter%treat_forces_in_core = .FALSE.

      ! Disk Storage
      IF (.NOT. skip_disk) THEN
         memory_parameter%actual_memory_usage_disk = 1
         CALL section_vals_val_get(hf_sub_section, "MAX_DISK_SPACE", i_val=int_val)
         memory_parameter%max_compression_counter_disk = int_val*1024_int_8*128_int_8
         IF (int_val == 0) THEN
            memory_parameter%do_disk_storage = .FALSE.
         ELSE
            memory_parameter%do_disk_storage = .TRUE.
         END IF
         CALL section_vals_val_get(hf_sub_section, "STORAGE_LOCATION", c_val=char_val)
         CALL compress(char_val, .TRUE.)
         !! Add ending / if necessary

         IF (SCAN(char_val, "/", .TRUE.) /= LEN_TRIM(char_val)) THEN
            WRITE (filename, '(A,A)') TRIM(char_val), "/"
            CALL compress(filename)
         ELSE
            filename = TRIM(char_val)
         END IF
         CALL compress(filename, .TRUE.)

         !! quickly check if we can write on storage_location
         CALL m_getcwd(orig_wd)
         CALL m_chdir(TRIM(filename), stat)
         IF (stat /= 0) THEN
            WRITE (error_msg, '(A,A,A)') "Request for disk storage failed due to unknown error while writing to ", &
               TRIM(filename), ". Please check STORAGE_LOCATION"
            CALL cp__b("hfx_types.F",1155,error_msg)
         END IF
         CALL m_chdir(orig_wd, stat)

         memory_parameter%storage_location = filename
         CALL compress(memory_parameter%storage_location, .TRUE.)
      ELSE
         memory_parameter%do_disk_storage = .FALSE.
      END IF
      IF (PRESENT(storage_id)) THEN
         storage_id = (irep-1)*para_env%num_pe*n_threads+para_env%mepos*n_threads+i_thread-1
      END IF
   END SUBROUTINE parse_memory_section

! **************************************************************************************************
!> \brief - This routine deallocates all data structures
!> \param x_data contains all relevant data structures for hfx runs
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_release(x_data)
      TYPE(hfx_type), DIMENSION(:, :), POINTER           :: x_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, i_thread, irep, n_rep_hf, n_threads
      TYPE(hfx_type), POINTER                            :: actual_x_data

!! There might be 2 hf sections

      n_rep_hf = x_data(1, 1)%n_rep_hf
      n_threads = 1
!$    n_threads = omp_get_max_threads()

      IF (x_data(1, 1)%potential_parameter%potential_type == do_hfx_potential_truncated .OR. &
          x_data(1, 1)%potential_parameter%potential_type == do_hfx_potential_mix_cl_trunc) THEN
         init_t_c_g0_lmax = -1
         CALL free_C0()
      END IF
      DO i_thread = 1, n_threads
         DO irep = 1, n_rep_hf
            actual_x_data => x_data(irep, i_thread)
            DEALLOCATE (actual_x_data%neighbor_cells)
            DEALLOCATE (actual_x_data%distribution_energy)
            DEALLOCATE (actual_x_data%distribution_forces)

            IF (actual_x_data%load_balance_parameter%blocks_initialized) THEN
               DEALLOCATE (actual_x_data%blocks)
               IF (i_thread == 1) THEN
                  DEALLOCATE (actual_x_data%pmax_block)
               END IF
            END IF

            IF (i_thread == 1) THEN
               DEALLOCATE (actual_x_data%atomic_pair_list)
               DEALLOCATE (actual_x_data%atomic_pair_list_forces)
            END IF

            IF (actual_x_data%screening_parameter%do_initial_p_screening .OR. &
                actual_x_data%screening_parameter%do_p_screening_forces) THEN
               IF (i_thread == 1) THEN
                  DEALLOCATE (actual_x_data%pmax_atom)
                  DO i = 1, SIZE(actual_x_data%initial_p)
                     DEALLOCATE (actual_x_data%initial_p(i)%p_kind)
                  END DO
                  DEALLOCATE (actual_x_data%initial_p)
                  IF (actual_x_data%memory_parameter%treat_forces_in_core) THEN
                     DEALLOCATE (actual_x_data%pmax_atom_forces)
                     DO i = 1, SIZE(actual_x_data%initial_p_forces)
                        DEALLOCATE (actual_x_data%initial_p_forces(i)%p_kind)
                     END DO
                     DEALLOCATE (actual_x_data%initial_p_forces)
                  END IF
               END IF
               DEALLOCATE (actual_x_data%map_atom_to_kind_atom)
            END IF
            IF (i_thread == 1) THEN
               DEALLOCATE (actual_x_data%is_assoc_atomic_block)
               DEALLOCATE (actual_x_data%atomic_block_offset)
               DEALLOCATE (actual_x_data%set_offset)
               DEALLOCATE (actual_x_data%block_offset)
            END IF

            !! BASIS parameter
            CALL hfx_release_basis_types(actual_x_data%basis_parameter)

            !MK Release libint and libderiv data structure
            CALL terminate_libint(actual_x_data%lib)
            CALL terminate_libderiv(actual_x_data%lib_deriv)

            !! Deallocate containers
            CALL dealloc_containers(actual_x_data, hfx_do_eval_energy)
            CALL dealloc_containers(actual_x_data, hfx_do_eval_forces)

            !! Deallocate containers
            CALL hfx_init_container(actual_x_data%maxval_container_disk, &
                                    actual_x_data%memory_parameter%actual_memory_usage_disk, &
                                    .FALSE.)
            IF (actual_x_data%memory_parameter%do_disk_storage) THEN
               CALL close_file(unit_number=actual_x_data%maxval_container_disk%unit, file_status="DELETE")
            END IF
            DEALLOCATE (actual_x_data%maxval_container_disk%first)
            DEALLOCATE (actual_x_data%maxval_container_disk)

            DO i = 1, 64
               CALL hfx_init_container(actual_x_data%integral_containers_disk(i), &
                                       actual_x_data%memory_parameter%actual_memory_usage_disk, &
                                       .FALSE.)
               IF (actual_x_data%memory_parameter%do_disk_storage) THEN
                  CALL close_file(unit_number=actual_x_data%integral_containers_disk(i)%unit, file_status="DELETE")
               END IF
               DEALLOCATE (actual_x_data%integral_containers_disk(i)%first)
            END DO
            DEALLOCATE (actual_x_data%integral_containers_disk)

            ! ** screening functions
            IF (actual_x_data%screen_funct_is_initialized) THEN
               DEALLOCATE (actual_x_data%screen_funct_coeffs_set)
               DEALLOCATE (actual_x_data%screen_funct_coeffs_kind)
               DEALLOCATE (actual_x_data%pair_dist_radii_pgf)
               DEALLOCATE (actual_x_data%screen_funct_coeffs_pgf)
               actual_x_data%screen_funct_is_initialized = .FALSE.
            END IF

            ! ** maps
            IF (ASSOCIATED(actual_x_data%map_atoms_to_cpus)) THEN
               DO i = 1, SIZE(actual_x_data%map_atoms_to_cpus)
                  DEALLOCATE (actual_x_data%map_atoms_to_cpus(i)%iatom_list)
                  DEALLOCATE (actual_x_data%map_atoms_to_cpus(i)%jatom_list)
               END DO
               DEALLOCATE (actual_x_data%map_atoms_to_cpus)
            END IF
         END DO

      END DO
      DEALLOCATE (x_data)
   END SUBROUTINE hfx_release

! **************************************************************************************************
!> \brief - This routine computes the neighbor cells that are taken into account
!>        in periodic runs
!> \param x_data contains all relevant data structures for hfx runs
!> \param pbc_shells number of shells taken into account
!> \param cell cell
!> \param i_thread current thread ID
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_create_neighbor_cells(x_data, pbc_shells, cell, i_thread)
      TYPE(hfx_type), POINTER                            :: x_data
      INTEGER, INTENT(INOUT)                             :: pbc_shells
      TYPE(cell_type), POINTER                           :: cell
      INTEGER, INTENT(IN)                                :: i_thread

      CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_create_neighbor_cells', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=512)                                 :: error_msg
      CHARACTER(LEN=64)                                  :: char_nshells
      INTEGER :: i, idx, ikind, ipgf, iset, ishell, j, jkind, jpgf, jset, jshell, k, kshell, l, &
         m(3), max_shell, nseta, nsetb, total_number_of_cells, ub, ub_max
      INTEGER, DIMENSION(:), POINTER                     :: la_max, lb_max, npgfa, npgfb
      LOGICAL                                            :: image_cell_found, nothing_more_to_add
      REAL(dp) :: cross_product(3), dist_min, distance(14), l_min, normal(3, 6), P(3, 14), &
         plane_vector(3, 2), point_in_plane(3), r(3), R1, R_max, R_max_stress, s(3), x, y, z, Zeta1
      REAL(dp), DIMENSION(:, :), POINTER                 :: zeta, zetb
      TYPE(hfx_cell_type), ALLOCATABLE, DIMENSION(:)     :: tmp_neighbor_cells

      total_number_of_cells = 0

      ! ** Check some settings
      IF (i_thread == 1) THEN
         IF (x_data%potential_parameter%potential_type /= do_hfx_potential_truncated .AND. &
             x_data%potential_parameter%potential_type /= do_hfx_potential_short .AND. &
             x_data%potential_parameter%potential_type /= do_hfx_potential_mix_cl_trunc) THEN
            CALL cp_warn(cp__l("hfx_types.F",1332), &
                         "Periodic Hartree Fock calculation requested without use "// &
                         "of a truncated or shortrange potential. This may lead to unphysical total energies. "// &
                         "Use a truncated  potential to avoid possible problems.")
         ELSE
            l_min = MIN(plane_distance(1, 0, 0, cell), &
                        plane_distance(0, 1, 0, cell), &
                        plane_distance(0, 0, 1, cell))
            l_min = 0.5_dp*l_min
            IF (x_data%potential_parameter%cutoff_radius >= l_min) THEN
               CALL cp_warn(cp__l("hfx_types.F",1342), &
                            "Periodic Hartree Fock calculation requested with use "// &
                            "of a truncated or shortrange potential. The cutoff radius is larger than half "// &
                            "the minimal cell dimension. This may lead to unphysical "// &
                            "total energies. Reduce the cutoff radius in order to avoid "// &
                            "possible problems.")
            END IF
         END IF
      END IF

      SELECT CASE (x_data%potential_parameter%potential_type)
      CASE (do_hfx_potential_truncated, do_hfx_potential_mix_cl_trunc, do_hfx_potential_short)
         R_max = 0.0_dp
         DO ikind = 1, SIZE(x_data%basis_parameter)
            la_max => x_data%basis_parameter(ikind)%lmax
            zeta => x_data%basis_parameter(ikind)%zet
            nseta = x_data%basis_parameter(ikind)%nset
            npgfa => x_data%basis_parameter(ikind)%npgf
            DO jkind = 1, SIZE(x_data%basis_parameter)
               lb_max => x_data%basis_parameter(jkind)%lmax
               zetb => x_data%basis_parameter(jkind)%zet
               nsetb = x_data%basis_parameter(jkind)%nset
               npgfb => x_data%basis_parameter(jkind)%npgf
               DO iset = 1, nseta
                  DO jset = 1, nsetb
                     DO ipgf = 1, npgfa(iset)
                        DO jpgf = 1, npgfb(jset)
                           Zeta1 = zeta(ipgf, iset)+zetb(jpgf, jset)
                           R1 = 1.0_dp/SQRT(Zeta1)*mul_fact(la_max(iset)+lb_max(jset))* &
                                SQRT(-LOG(x_data%screening_parameter%eps_schwarz))
                           R_max = MAX(R1, R_max)
                        END DO
                     END DO
                  END DO
               END DO
            END DO
         END DO

         R_max = 2.0_dp*R_max+x_data%potential_parameter%cutoff_radius
         nothing_more_to_add = .FALSE.
         max_shell = 0
         total_number_of_cells = 0
         ub = 1
         DEALLOCATE (x_data%neighbor_cells)
         ALLOCATE (x_data%neighbor_cells(1))
         x_data%neighbor_cells(1)%cell = 0.0_dp
         x_data%neighbor_cells(1)%cell_r = 0.0_dp

         ! ** What follows is kind of a ray tracing algorithm
         ! ** Given a image cell (ishell, jshell, kshell) we try to figure out the
         ! ** shortest distance of this image cell to the basic unit cell (0,0,0), i.e. the point
         ! ** (0.0, 0.0, 0.0)
         ! ** This is achieved by checking the 8 Corners of the cell, and, in addition, the shortest distance
         ! ** to all 6 faces. The faces are only taken into account if the penetration point of the normal
         ! ** to the plane defined by a face lies within this face.
         ! ** This is very fast, because no trigonometric functions are being used
         ! ** The points are defined as follows
         ! **
         ! **
         ! **               _________________________
         ! **              /P4____________________P8/|
         ! **             / / ___________________/ / |
         ! **            / / /| |               / /  |       z
         ! **           / / / | |              / / . |      /|\  _ y
         ! **          / / /| | |             / / /| |       |   /|
         ! **         / / / | | |            / / / | |       |  /
         ! **        / / /  | | |           / / /| | |       | /
         ! **       / /_/___| | |__________/ / / | | |       |/
         ! **      /P2______| | |_________P6/ /  | | |       ----------> x
         ! **      | _______| | |_________| | |  | | |
         ! **      | | |    | | |________________| | |
         ! **      | | |    |P3___________________P7 |
         ! **      | | |   / / _________________  / /
         ! **      | | |  / / /           | | |/ / /
         ! **      | | | / / /            | | | / /
         ! **      | | |/ / /             | | |/ /
         ! **      | | | / /              | | ' /
         ! **      | | |/_/_______________| |  /
         ! **      | |____________________| | /
         ! **      |P1_____________________P5/
         ! **
         ! **

         DO WHILE (.NOT. nothing_more_to_add)
            ! Calculate distances to the eight points P1 to P8
            image_cell_found = .FALSE.
            ALLOCATE (tmp_neighbor_cells(1:ub))
            DO i = 1, ub-1
               tmp_neighbor_cells(i) = x_data%neighbor_cells(i)
            END DO
            ub_max = (2*max_shell+1)**3
            DEALLOCATE (x_data%neighbor_cells)
            ALLOCATE (x_data%neighbor_cells(1:ub_max))
            DO i = 1, ub-1
               x_data%neighbor_cells(i) = tmp_neighbor_cells(i)
            END DO
            DO i = ub, ub_max
               x_data%neighbor_cells(i)%cell = 0.0_dp
               x_data%neighbor_cells(i)%cell_r = 0.0_dp
            END DO

            DEALLOCATE (tmp_neighbor_cells)

            DO ishell = -max_shell, max_shell
            DO jshell = -max_shell, max_shell
            DO kshell = -max_shell, max_shell
               IF (MAX(ABS(ishell), ABS(jshell), ABS(kshell)) /= max_shell) CYCLE
               idx = 0
               DO j = 0, 1
                  x = -1.0_dp/2.0_dp+j*1.0_dp
                  DO k = 0, 1
                     y = -1.0_dp/2.0_dp+k*1.0_dp
                     DO l = 0, 1
                        z = -1.0_dp/2.0_dp+l*1.0_dp
                        idx = idx+1
                        P(1, idx) = x+ishell
                        P(2, idx) = y+jshell
                        P(3, idx) = z+kshell
                        CALL scaled_to_real(r, P(:, idx), cell)
                        distance(idx) = SQRT(SUM(r**2))
                        P(1:3, idx) = r
                     END DO
                  END DO
               END DO
               ! Now check distance to Faces and only take them into account if the base point lies within quadrilateral

               ! Face A (1342) 1 is the reference
               idx = idx+1
               plane_vector(:, 1) = P(:, 3)-P(:, 1)
               plane_vector(:, 2) = P(:, 2)-P(:, 1)
               cross_product(1) = plane_vector(2, 1)*plane_vector(3, 2)-plane_vector(3, 1)*plane_vector(2, 2)
               cross_product(2) = plane_vector(3, 1)*plane_vector(1, 2)-plane_vector(1, 1)*plane_vector(3, 2)
               cross_product(3) = plane_vector(1, 1)*plane_vector(2, 2)-plane_vector(2, 1)*plane_vector(1, 2)
               normal(:, 1) = cross_product/SQRT(SUM(cross_product**2))
               point_in_plane = -normal(:, 1)*(normal(1, 1)*P(1, 1)+normal(2, 1)*P(2, 1)+normal(3, 1)*P(3, 1))

               IF (point_is_in_quadrilateral(P(:, 1), P(:, 3), P(:, 4), P(:, 2), point_in_plane)) THEN
                  distance(idx) = ABS(normal(1, 1)*P(1, 1)+normal(2, 1)*P(2, 1)+normal(3, 1)*P(3, 1))
               ELSE
                  distance(idx) = HUGE(distance(idx))
               END IF

               ! Face B (1562) 1 is the reference
               idx = idx+1
               plane_vector(:, 1) = P(:, 2)-P(:, 1)
               plane_vector(:, 2) = P(:, 5)-P(:, 1)
               cross_product(1) = plane_vector(2, 1)*plane_vector(3, 2)-plane_vector(3, 1)*plane_vector(2, 2)
               cross_product(2) = plane_vector(3, 1)*plane_vector(1, 2)-plane_vector(1, 1)*plane_vector(3, 2)
               cross_product(3) = plane_vector(1, 1)*plane_vector(2, 2)-plane_vector(2, 1)*plane_vector(1, 2)
               normal(:, 1) = cross_product/SQRT(SUM(cross_product**2))
               point_in_plane = -normal(:, 1)*(normal(1, 1)*P(1, 1)+normal(2, 1)*P(2, 1)+normal(3, 1)*P(3, 1))

               IF (point_is_in_quadrilateral(P(:, 1), P(:, 5), P(:, 6), P(:, 2), point_in_plane)) THEN
                  distance(idx) = ABS(normal(1, 1)*P(1, 1)+normal(2, 1)*P(2, 1)+normal(3, 1)*P(3, 1))
               ELSE
                  distance(idx) = HUGE(distance(idx))
               END IF

               ! Face C (5786) 5 is the reference
               idx = idx+1
               plane_vector(:, 1) = P(:, 7)-P(:, 5)
               plane_vector(:, 2) = P(:, 6)-P(:, 5)
               cross_product(1) = plane_vector(2, 1)*plane_vector(3, 2)-plane_vector(3, 1)*plane_vector(2, 2)
               cross_product(2) = plane_vector(3, 1)*plane_vector(1, 2)-plane_vector(1, 1)*plane_vector(3, 2)
               cross_product(3) = plane_vector(1, 1)*plane_vector(2, 2)-plane_vector(2, 1)*plane_vector(1, 2)
               normal(:, 1) = cross_product/SQRT(SUM(cross_product**2))
               point_in_plane = -normal(:, 1)*(normal(1, 1)*P(1, 5)+normal(2, 1)*P(2, 5)+normal(3, 1)*P(3, 5))

               IF (point_is_in_quadrilateral(P(:, 5), P(:, 7), P(:, 8), P(:, 6), point_in_plane)) THEN
                  distance(idx) = ABS(normal(1, 1)*P(1, 5)+normal(2, 1)*P(2, 5)+normal(3, 1)*P(3, 5))
               ELSE
                  distance(idx) = HUGE(distance(idx))
               END IF

               ! Face D (3784) 3 is the reference
               idx = idx+1
               plane_vector(:, 1) = P(:, 7)-P(:, 3)
               plane_vector(:, 2) = P(:, 4)-P(:, 3)
               cross_product(1) = plane_vector(2, 1)*plane_vector(3, 2)-plane_vector(3, 1)*plane_vector(2, 2)
               cross_product(2) = plane_vector(3, 1)*plane_vector(1, 2)-plane_vector(1, 1)*plane_vector(3, 2)
               cross_product(3) = plane_vector(1, 1)*plane_vector(2, 2)-plane_vector(2, 1)*plane_vector(1, 2)
               normal(:, 1) = cross_product/SQRT(SUM(cross_product**2))
               point_in_plane = -normal(:, 1)*(normal(1, 1)*P(1, 3)+normal(2, 1)*P(2, 3)+normal(3, 1)*P(3, 3))

               IF (point_is_in_quadrilateral(P(:, 3), P(:, 7), P(:, 8), P(:, 4), point_in_plane)) THEN
                  distance(idx) = ABS(normal(1, 1)*P(1, 3)+normal(2, 1)*P(2, 3)+normal(3, 1)*P(3, 3))
               ELSE
                  distance(idx) = HUGE(distance(idx))
               END IF

               ! Face E (2684) 2 is the reference
               idx = idx+1
               plane_vector(:, 1) = P(:, 6)-P(:, 2)
               plane_vector(:, 2) = P(:, 4)-P(:, 2)
               cross_product(1) = plane_vector(2, 1)*plane_vector(3, 2)-plane_vector(3, 1)*plane_vector(2, 2)
               cross_product(2) = plane_vector(3, 1)*plane_vector(1, 2)-plane_vector(1, 1)*plane_vector(3, 2)
               cross_product(3) = plane_vector(1, 1)*plane_vector(2, 2)-plane_vector(2, 1)*plane_vector(1, 2)
               normal(:, 1) = cross_product/SQRT(SUM(cross_product**2))
               point_in_plane = -normal(:, 1)*(normal(1, 1)*P(1, 2)+normal(2, 1)*P(2, 2)+normal(3, 1)*P(3, 2))

               IF (point_is_in_quadrilateral(P(:, 2), P(:, 6), P(:, 8), P(:, 4), point_in_plane)) THEN
                  distance(idx) = ABS(normal(1, 1)*P(1, 2)+normal(2, 1)*P(2, 2)+normal(3, 1)*P(3, 2))
               ELSE
                  distance(idx) = HUGE(distance(idx))
               END IF

               ! Face F (1573) 1 is the reference
               idx = idx+1
               plane_vector(:, 1) = P(:, 5)-P(:, 1)
               plane_vector(:, 2) = P(:, 3)-P(:, 1)
               cross_product(1) = plane_vector(2, 1)*plane_vector(3, 2)-plane_vector(3, 1)*plane_vector(2, 2)
               cross_product(2) = plane_vector(3, 1)*plane_vector(1, 2)-plane_vector(1, 1)*plane_vector(3, 2)
               cross_product(3) = plane_vector(1, 1)*plane_vector(2, 2)-plane_vector(2, 1)*plane_vector(1, 2)
               normal(:, 1) = cross_product/SQRT(SUM(cross_product**2))
               point_in_plane = -normal(:, 1)*(normal(1, 1)*P(1, 1)+normal(2, 1)*P(2, 1)+normal(3, 1)*P(3, 1))

               IF (point_is_in_quadrilateral(P(:, 1), P(:, 5), P(:, 7), P(:, 3), point_in_plane)) THEN
                  distance(idx) = ABS(normal(1, 1)*P(1, 1)+normal(2, 1)*P(2, 1)+normal(3, 1)*P(3, 1))
               ELSE
                  distance(idx) = HUGE(distance(idx))
               END IF

               dist_min = MINVAL(distance)
               IF (max_shell == 0) THEN
                  image_cell_found = .TRUE.
               END IF
               IF (dist_min < R_max) THEN
                  total_number_of_cells = total_number_of_cells+1
                  x_data%neighbor_cells(ub)%cell = REAL((/ishell, jshell, kshell/), dp)
                  ub = ub+1
                  image_cell_found = .TRUE.
               END IF

            END DO
            END DO
            END DO
            IF (image_cell_found) THEN
               max_shell = max_shell+1
            ELSE
               nothing_more_to_add = .TRUE.
            END IF
         END DO
         ! now remove what is not needed
         ALLOCATE (tmp_neighbor_cells(total_number_of_cells))
         DO i = 1, ub-1
            tmp_neighbor_cells(i) = x_data%neighbor_cells(i)
         END DO
         DEALLOCATE (x_data%neighbor_cells)
         ! If we only need the supercell, total_number_of_cells is still 0, repair
         IF (total_number_of_cells == 0) THEN
            total_number_of_cells = 1
            ALLOCATE (x_data%neighbor_cells(total_number_of_cells))
            DO i = 1, total_number_of_cells
               x_data%neighbor_cells(i)%cell = 0.0_dp
               x_data%neighbor_cells(i)%cell_r = 0.0_dp
            END DO
         ELSE
            ALLOCATE (x_data%neighbor_cells(total_number_of_cells))
            DO i = 1, total_number_of_cells
               x_data%neighbor_cells(i) = tmp_neighbor_cells(i)
            END DO
         END IF
         DEALLOCATE (tmp_neighbor_cells)

         IF (x_data%periodic_parameter%number_of_shells == do_hfx_auto_shells) THEN
            ! Do nothing
         ELSE
            total_number_of_cells = 0
            DO i = 0, x_data%periodic_parameter%number_of_shells
               total_number_of_cells = total_number_of_cells+count_cells_perd(i, x_data%periodic_parameter%perd)
            END DO
            IF (total_number_of_cells < SIZE(x_data%neighbor_cells)) THEN
               IF (i_thread == 1) THEN

                  WRITE (char_nshells, '(I3)') SIZE(x_data%neighbor_cells)
                  WRITE (error_msg, '(A,A,A)') "Periodic Hartree Fock calculation requested with use "// &
                     "of a truncated potential. The number of shells to be considered "// &
                     "might be too small. CP2K conservatively estimates to need "//TRIM(char_nshells)//" periodic images "// &
                     "Please carefully check if you get converged results."
                  CALL cp__w("hfx_types.F",1621,error_msg)
               END IF
            END IF
            total_number_of_cells = 0
            DO i = 0, x_data%periodic_parameter%number_of_shells
               total_number_of_cells = total_number_of_cells+count_cells_perd(i, x_data%periodic_parameter%perd)
            END DO
            DEALLOCATE (x_data%neighbor_cells)

            ALLOCATE (x_data%neighbor_cells(total_number_of_cells))
            m = 0
            i = 1
            DO WHILE (SUM(m**2) <= x_data%periodic_parameter%number_of_shells)
               x_data%neighbor_cells(i)%cell = REAL(m, dp)
               CALL next_image_cell_perd(m, x_data%periodic_parameter%perd)
               i = i+1
            ENDDO
         END IF
      CASE DEFAULT
         total_number_of_cells = 0
         IF (pbc_shells == -1) pbc_shells = 0
         DO i = 0, pbc_shells
            total_number_of_cells = total_number_of_cells+count_cells_perd(i, x_data%periodic_parameter%perd)
         END DO
         DEALLOCATE (x_data%neighbor_cells)

         ALLOCATE (x_data%neighbor_cells(total_number_of_cells))

         m = 0
         i = 1
         DO WHILE (SUM(m**2) <= pbc_shells)
            x_data%neighbor_cells(i)%cell = REAL(m, dp)
            CALL next_image_cell_perd(m, x_data%periodic_parameter%perd)
            i = i+1
         ENDDO
      END SELECT

      ! ** Transform into real coord
      DO i = 1, SIZE(x_data%neighbor_cells)
         r = 0.0_dp
         x_data%neighbor_cells(i)%cell_r(:) = 0.0_dp
         s = x_data%neighbor_cells(i)%cell(:)
         CALL scaled_to_real(x_data%neighbor_cells(i)%cell_r, s, cell)
      END DO
      x_data%periodic_parameter%number_of_shells = pbc_shells

      R_max_stress = 0.0_dp
      DO i = 1, SIZE(x_data%neighbor_cells)
         R_max_stress = MAX(R_max_stress, MAXVAL(ABS(x_data%neighbor_cells(i)%cell_r(:))))
      END DO
      R_max_stress = R_max_stress+ABS(MAXVAL(cell%hmat(:, :)))
      x_data%periodic_parameter%R_max_stress = R_max_stress

   END SUBROUTINE hfx_create_neighbor_cells

   ! performs a fuzzy check of being in a quadrilateral
! **************************************************************************************************
!> \brief ...
!> \param A ...
!> \param B ...
!> \param C ...
!> \param D ...
!> \param P ...
!> \retval point_is_in_quadrilateral ...
! **************************************************************************************************
   FUNCTION point_is_in_quadrilateral(A, B, C, D, P)
      REAL(dp)                                           :: A(3), B(3), C(3), D(3), P(3)
      LOGICAL                                            :: point_is_in_quadrilateral

      REAL(dp), PARAMETER :: fuzzy = 1000.0_dp*EPSILON(1.0_dp)

      REAL(dp)                                           :: dot00, dot01, dot02, dot11, dot12, &
                                                            invDenom, u, v, v0(3), v1(3), v2(3)

      point_is_in_quadrilateral = .FALSE.

      ! ** Check for both triangles ABC and ACD
      ! **
      ! **     D -------------- C
      ! **    /                /
      ! **   /                /
      ! **  A----------------B
      ! **
      ! **
      ! **

      ! ** ABC

      v0 = D-A
      v1 = C-A
      v2 = P-A

      ! ** Compute dot products
      dot00 = DOT_PRODUCT(v0, v0)
      dot01 = DOT_PRODUCT(v0, v1)
      dot02 = DOT_PRODUCT(v0, v2)
      dot11 = DOT_PRODUCT(v1, v1)
      dot12 = DOT_PRODUCT(v1, v2)

      ! ** Compute barycentric coordinates
      invDenom = 1/(dot00*dot11-dot01*dot01)
      u = (dot11*dot02-dot01*dot12)*invDenom
      v = (dot00*dot12-dot01*dot02)*invDenom
      ! ** Check if point is in triangle
      IF ((u >= 0-fuzzy) .AND. (v >= 0-fuzzy) .AND. (u+v <= 1+fuzzy)) THEN
         point_is_in_quadrilateral = .TRUE.
         RETURN
      END IF
      v0 = C-A
      v1 = B-A
      v2 = P-A

      ! ** Compute dot products
      dot00 = DOT_PRODUCT(v0, v0)
      dot01 = DOT_PRODUCT(v0, v1)
      dot02 = DOT_PRODUCT(v0, v2)
      dot11 = DOT_PRODUCT(v1, v1)
      dot12 = DOT_PRODUCT(v1, v2)

      ! ** Compute barycentric coordinates
      invDenom = 1/(dot00*dot11-dot01*dot01)
      u = (dot11*dot02-dot01*dot12)*invDenom
      v = (dot00*dot12-dot01*dot02)*invDenom

      ! ** Check if point is in triangle
      IF ((u >= 0-fuzzy) .AND. (v >= 0-fuzzy) .AND. (u+v <= 1+fuzzy)) THEN
         point_is_in_quadrilateral = .TRUE.
         RETURN
      END IF

   END FUNCTION point_is_in_quadrilateral

! **************************************************************************************************
!> \brief - This routine deletes all list entries in a container in order to
!>        deallocate the memory.
!> \param container container that contains the compressed elements
!> \param memory_usage ...
!> \param do_disk_storage ...
!> \par History
!>      10.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_init_container(container, memory_usage, do_disk_storage)
      TYPE(hfx_container_type)                           :: container
      INTEGER                                            :: memory_usage
      LOGICAL                                            :: do_disk_storage

      CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_init_container', &
         routineP = moduleN//':'//routineN

      TYPE(hfx_container_node), POINTER                  :: current, next

!! DEALLOCATE memory

      current => container%first
      DO WHILE (ASSOCIATED(current))
         next => current%next
         DEALLOCATE (current)
         current => next
      END DO

      !! Allocate first list entry, init members
      ALLOCATE (container%first)
      container%first%prev => NULL()
      container%first%next => NULL()
      container%current => container%first
      container%current%data = 0
      container%element_counter = 1
      memory_usage = 1

      IF (do_disk_storage) THEN
         !! close the file, if this is no the first time
         IF (container%unit /= -1) THEN
            CALL close_file(unit_number=container%unit)
         END IF
         CALL open_file(file_name=TRIM(container%filename), file_status="UNKNOWN", file_form="UNFORMATTED", file_action="WRITE", &
                        unit_number=container%unit)
      END IF

   END SUBROUTINE hfx_init_container

! **************************************************************************************************
!> \brief - This routine stores the data obtained from the load balance routine
!>        for the energy
!> \param ptr_to_distr contains data to store
!> \param x_data contains all relevant data structures for hfx runs
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_set_distr_energy(ptr_to_distr, x_data)
      TYPE(hfx_distribution), DIMENSION(:), POINTER      :: ptr_to_distr
      TYPE(hfx_type), POINTER                            :: x_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_set_distr_energy', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (x_data%distribution_energy)

      ALLOCATE (x_data%distribution_energy(SIZE(ptr_to_distr)))
      x_data%distribution_energy = ptr_to_distr

   END SUBROUTINE hfx_set_distr_energy

! **************************************************************************************************
!> \brief - This routine stores the data obtained from the load balance routine
!>        for the forces
!> \param ptr_to_distr contains data to store
!> \param x_data contains all relevant data structures for hfx runs
!> \par History
!>      09.2007 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_set_distr_forces(ptr_to_distr, x_data)
      TYPE(hfx_distribution), DIMENSION(:), POINTER      :: ptr_to_distr
      TYPE(hfx_type), POINTER                            :: x_data

      CHARACTER(LEN=*), PARAMETER :: routineN = 'hfx_set_distr_forces', &
         routineP = moduleN//':'//routineN

      DEALLOCATE (x_data%distribution_forces)

      ALLOCATE (x_data%distribution_forces(SIZE(ptr_to_distr)))
      x_data%distribution_forces = ptr_to_distr

   END SUBROUTINE hfx_set_distr_forces

! **************************************************************************************************
!> \brief - resets the maximum memory usage for a HFX calculation subtracting
!>          all relevant buffers from the input MAX_MEM value and add 10% of
!>          safety margin
!> \param memory_parameter Memory information
!> \param subtr_size_mb size of buffers in MiB
!> \par History
!>      02.2009 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_reset_memory_usage_counter(memory_parameter, subtr_size_mb)

      TYPE(hfx_memory_type)                              :: memory_parameter
      INTEGER(int_8), INTENT(IN)                         :: subtr_size_mb

      INTEGER(int_8)                                     :: max_memory

      max_memory = memory_parameter%max_memory
      max_memory = max_memory-subtr_size_mb
      IF (max_memory <= 0) THEN
         memory_parameter%do_all_on_the_fly = .TRUE.
         memory_parameter%max_compression_counter = 0
      ELSE
         memory_parameter%do_all_on_the_fly = .FALSE.
         memory_parameter%max_compression_counter = max_memory*1024_int_8*128_int_8
      END IF
   END SUBROUTINE hfx_reset_memory_usage_counter

! **************************************************************************************************
!> \brief - This routine prints some information on HFX
!> \param x_data contains all relevant data structures for hfx runs
!> \param hfx_section HFX input section
!> \param i_rep current replica ID
!> \par History
!>      03.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE hfx_print_info(x_data, hfx_section, i_rep)
      TYPE(hfx_type), POINTER                            :: x_data
      TYPE(section_vals_type), POINTER                   :: hfx_section
      INTEGER, INTENT(IN)                                :: i_rep

      INTEGER                                            :: iw
      REAL(dp)                                           :: rc_ang
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()

      iw = cp_print_key_unit_nr(logger, hfx_section, "HF_INFO", &
                                extension=".scfLog")
      IF (iw > 0) THEN
         WRITE (UNIT=iw, FMT="(/,(T3,A,T61,I20))") &
            "HFX_INFO| Replica ID:     ", i_rep
         WRITE (UNIT=iw, FMT="((T3,A,T73,ES8.1))") &
            "HFX_INFO| EPS_SCHWARZ:     ", x_data%screening_parameter%eps_schwarz
         WRITE (UNIT=iw, FMT="((T3,A,T73,ES8.1))") &
            "HFX_INFO| EPS_SCHWARZ_FORCES     ", x_data%screening_parameter%eps_schwarz_forces
         WRITE (UNIT=iw, FMT="((T3,A,T73,ES8.1))") &
            "HFX_INFO| EPS_STORAGE_SCALING:     ", x_data%memory_parameter%eps_storage_scaling
         WRITE (UNIT=iw, FMT="((T3,A,T61,I20))") &
            "HFX_INFO| NBINS:     ", x_data%load_balance_parameter%nbins
         WRITE (UNIT=iw, FMT="((T3,A,T61,I20))") &
            "HFX_INFO| BLOCK_SIZE:     ", x_data%load_balance_parameter%block_size
         WRITE (iw, '(T3,A,T61,F20.10)') &
            "HFX_INFO| FRACTION:     ", x_data%general_parameter%fraction
         SELECT CASE (x_data%potential_parameter%potential_type)
         CASE (do_hfx_potential_coulomb)
            WRITE (UNIT=iw, FMT="((T3,A,T74,A))") &
               "HFX_INFO| Interaction Potential:     ", "COULOMB"
         CASE (do_hfx_potential_short)
            WRITE (UNIT=iw, FMT="((T3,A,T71,A))") &
               "HFX_INFO| Interaction Potential:    ", "SHORTRANGE"
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| Omega:     ", x_data%potential_parameter%omega
            rc_ang = cp_unit_from_cp2k(x_data%potential_parameter%cutoff_radius, "angstrom")
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| Cutoff Radius [angstrom]:     ", rc_ang
         CASE (do_hfx_potential_long)
            WRITE (UNIT=iw, FMT="((T3,A,T72,A))") &
               "HFX_INFO| Interaction Potential:     ", "LONGRANGE"
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| Omega:     ", x_data%potential_parameter%omega
         CASE (do_hfx_potential_mix_cl)
            WRITE (UNIT=iw, FMT="((T3,A,T75,A))") &
               "HFX_INFO| Interaction Potential:     ", "MIX_CL"
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| Omega:     ", x_data%potential_parameter%omega
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| SCALE_COULOMB:     ", x_data%potential_parameter%scale_coulomb
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| SCALE_LONGRANGE:     ", x_data%potential_parameter%scale_longrange
         CASE (do_hfx_potential_gaussian)
            WRITE (UNIT=iw, FMT="((T3,A,T73,A))") &
               "HFX_INFO| Interaction Potential:     ", "GAUSSIAN"
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| Omega:     ", x_data%potential_parameter%omega
         CASE (do_hfx_potential_mix_lg)
            WRITE (UNIT=iw, FMT="((T3,A,T75,A))") &
               "HFX_INFO| Interaction Potential:    ", "MIX_LG"
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| Omega:     ", x_data%potential_parameter%omega
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| SCALE_LONGRANGE:     ", x_data%potential_parameter%scale_longrange
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| SCALE_GAUSSIAN:    ", x_data%potential_parameter%scale_gaussian
         CASE (do_hfx_potential_id)
            WRITE (UNIT=iw, FMT="((T3,A,T73,A))") &
               "HFX_INFO| Interaction Potential:    ", "IDENTITY"
         CASE (do_hfx_potential_truncated)
            WRITE (UNIT=iw, FMT="((T3,A,T72,A))") &
               "HFX_INFO| Interaction Potential:    ", "TRUNCATED"
            rc_ang = cp_unit_from_cp2k(x_data%potential_parameter%cutoff_radius, "angstrom")
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| Cutoff Radius [angstrom]:     ", rc_ang
         CASE (do_hfx_potential_mix_cl_trunc)
            WRITE (UNIT=iw, FMT="((T3,A,T65,A))") &
               "HFX_INFO| Interaction Potential:    ", "TRUNCATED MIX_CL"
            rc_ang = cp_unit_from_cp2k(x_data%potential_parameter%cutoff_radius, "angstrom")
            WRITE (iw, '(T3,A,T61,F20.10)') &
               "HFX_INFO| Cutoff Radius [angstrom]:     ", rc_ang
         END SELECT
         IF (x_data%periodic_parameter%do_periodic) THEN
            IF (x_data%periodic_parameter%mode == -1) THEN
               WRITE (UNIT=iw, FMT="((T3,A,T77,A))") &
                  "HFX_INFO| NUMBER_OF_SHELLS:     ", "AUTO"
            ELSE
               WRITE (UNIT=iw, FMT="((T3,A,T61,I20))") &
                  "HFX_INFO| NUMBER_OF_SHELLS:     ", x_data%periodic_parameter%mode
            END IF
            WRITE (UNIT=iw, FMT="((T3,A,T61,I20))") &
               "HFX_INFO| Number of periodic shells considered:     ", x_data%periodic_parameter%number_of_shells
            WRITE (UNIT=iw, FMT="((T3,A,T61,I20),/)") &
               "HFX_INFO| Number of periodic cells considered:     ", SIZE(x_data%neighbor_cells)
         ELSE
            WRITE (UNIT=iw, FMT="((T3,A,T77,A))") &
               "HFX_INFO| Number of periodic shells considered:     ", "NONE"
            WRITE (UNIT=iw, FMT="((T3,A,T77,A),/)") &
               "HFX_INFO| Number of periodic cells considered:     ", "NONE"
         END IF
      END IF
      CALL cp_print_key_finished_output(iw, logger, hfx_section, &
                                        "HF_INFO")
   END SUBROUTINE hfx_print_info

! **************************************************************************************************
!> \brief ...
!> \param x_data ...
!> \param eval_type ...
! **************************************************************************************************
   SUBROUTINE dealloc_containers(x_data, eval_type)
      TYPE(hfx_type), POINTER                            :: x_data
      INTEGER, INTENT(IN)                                :: eval_type

      CHARACTER(LEN=*), PARAMETER :: routineN = 'dealloc_containers', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bin, i

      SELECT CASE (eval_type)
      CASE (hfx_do_eval_energy)
         DO bin = 1, SIZE(x_data%maxval_container)
            CALL hfx_init_container(x_data%maxval_container(bin), x_data%memory_parameter%actual_memory_usage, &
                                    .FALSE.)
            DEALLOCATE (x_data%maxval_container(bin)%first)
         END DO
         DEALLOCATE (x_data%maxval_container)
         DEALLOCATE (x_data%maxval_cache)

         DO bin = 1, SIZE(x_data%integral_containers, 2)
            DO i = 1, 64
               CALL hfx_init_container(x_data%integral_containers(i, bin), x_data%memory_parameter%actual_memory_usage, &
                                       .FALSE.)
               DEALLOCATE (x_data%integral_containers(i, bin)%first)
            END DO
         END DO
         DEALLOCATE (x_data%integral_containers)

         DEALLOCATE (x_data%integral_caches)
      CASE (hfx_do_eval_forces)
         DO bin = 1, SIZE(x_data%maxval_container_forces)
            CALL hfx_init_container(x_data%maxval_container_forces(bin), x_data%memory_parameter%actual_memory_usage, &
                                    .FALSE.)
            DEALLOCATE (x_data%maxval_container_forces(bin)%first)
         END DO
         DEALLOCATE (x_data%maxval_container_forces)
         DEALLOCATE (x_data%maxval_cache_forces)

         DO bin = 1, SIZE(x_data%integral_containers_forces, 2)
            DO i = 1, 64
               CALL hfx_init_container(x_data%integral_containers_forces(i, bin), x_data%memory_parameter%actual_memory_usage, &
                                       .FALSE.)
               DEALLOCATE (x_data%integral_containers_forces(i, bin)%first)
            END DO
         END DO
         DEALLOCATE (x_data%integral_containers_forces)

         DEALLOCATE (x_data%integral_caches_forces)

      END SELECT

   END SUBROUTINE dealloc_containers

! **************************************************************************************************
!> \brief ...
!> \param x_data ...
!> \param bin_size ...
!> \param eval_type ...
! **************************************************************************************************
   SUBROUTINE alloc_containers(x_data, bin_size, eval_type)
      TYPE(hfx_type), POINTER                            :: x_data
      INTEGER, INTENT(IN)                                :: bin_size, eval_type

      CHARACTER(LEN=*), PARAMETER :: routineN = 'alloc_containers', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bin, i

      SELECT CASE (eval_type)
      CASE (hfx_do_eval_energy)
         ALLOCATE (x_data%maxval_cache(bin_size))
         DO bin = 1, bin_size
            x_data%maxval_cache(bin)%element_counter = 1
         END DO
         ALLOCATE (x_data%maxval_container(bin_size))
         DO bin = 1, bin_size
            ALLOCATE (x_data%maxval_container(bin)%first)
            x_data%maxval_container(bin)%first%prev => NULL()
            x_data%maxval_container(bin)%first%next => NULL()
            x_data%maxval_container(bin)%current => x_data%maxval_container(bin)%first
            x_data%maxval_container(bin)%current%data = 0
            x_data%maxval_container(bin)%element_counter = 1
         END DO

         ALLOCATE (x_data%integral_containers(64, bin_size))
         ALLOCATE (x_data%integral_caches(64, bin_size))

         DO bin = 1, bin_size
            DO i = 1, 64
               x_data%integral_caches(i, bin)%element_counter = 1
               x_data%integral_caches(i, bin)%data = 0
               ALLOCATE (x_data%integral_containers(i, bin)%first)
               x_data%integral_containers(i, bin)%first%prev => NULL()
               x_data%integral_containers(i, bin)%first%next => NULL()
               x_data%integral_containers(i, bin)%current => x_data%integral_containers(i, bin)%first
               x_data%integral_containers(i, bin)%current%data = 0
               x_data%integral_containers(i, bin)%element_counter = 1
            END DO
         END DO
      CASE (hfx_do_eval_forces)
         ALLOCATE (x_data%maxval_cache_forces(bin_size))
         DO bin = 1, bin_size
            x_data%maxval_cache_forces(bin)%element_counter = 1
         END DO
         ALLOCATE (x_data%maxval_container_forces(bin_size))
         DO bin = 1, bin_size
            ALLOCATE (x_data%maxval_container_forces(bin)%first)
            x_data%maxval_container_forces(bin)%first%prev => NULL()
            x_data%maxval_container_forces(bin)%first%next => NULL()
            x_data%maxval_container_forces(bin)%current => x_data%maxval_container_forces(bin)%first
            x_data%maxval_container_forces(bin)%current%data = 0
            x_data%maxval_container_forces(bin)%element_counter = 1
         END DO

         ALLOCATE (x_data%integral_containers_forces(64, bin_size))
         ALLOCATE (x_data%integral_caches_forces(64, bin_size))

         DO bin = 1, bin_size
            DO i = 1, 64
               x_data%integral_caches_forces(i, bin)%element_counter = 1
               x_data%integral_caches_forces(i, bin)%data = 0
               ALLOCATE (x_data%integral_containers_forces(i, bin)%first)
               x_data%integral_containers_forces(i, bin)%first%prev => NULL()
               x_data%integral_containers_forces(i, bin)%first%next => NULL()
               x_data%integral_containers_forces(i, bin)%current => x_data%integral_containers_forces(i, bin)%first
               x_data%integral_containers_forces(i, bin)%current%data = 0
               x_data%integral_containers_forces(i, bin)%element_counter = 1
            END DO
         END DO

      END SELECT

   END SUBROUTINE alloc_containers

END MODULE hfx_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief types of  preconditioners
!> \par History
!>      Separate types from construction and application
!> \author Joost VandeVondele (09.2002)
! **************************************************************************************************
MODULE preconditioner_types
   USE cp_blacs_env,                    ONLY: cp_blacs_env_release,&
                                              cp_blacs_env_retain,&
                                              cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_release_p,&
                                              cp_dbcsr_type
   USE cp_fm_types,                     ONLY: cp_fm_release,&
                                              cp_fm_type
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: cholesky_reduce,&
                                              ot_precond_solver_default
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'preconditioner_types'

   PUBLIC  :: preconditioner_type, preconditioner_p_type
   PUBLIC  :: init_preconditioner
   PUBLIC  :: preconditioner_in_use
   PUBLIC  :: destroy_preconditioner

! **************************************************************************************************
   TYPE preconditioner_type
!    PRIVATE
      TYPE(cp_dbcsr_type), POINTER :: sparse_matrix
      TYPE(cp_fm_type), POINTER :: fm
      TYPE(cp_dbcsr_type), POINTER           :: dbcsr_matrix
      TYPE(cp_dbcsr_type), POINTER           :: max_ev_vector
      TYPE(cp_dbcsr_type), POINTER           :: min_ev_vector
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:) :: inverse_history
      TYPE(cp_para_env_type), POINTER   :: para_env
      TYPE(cp_blacs_env_type), POINTER   :: ctxt
      INTEGER :: in_use, solver, ihistory, cholesky_use
      REAL(KIND=dp), DIMENSION(:), POINTER :: occ_evals, full_evals
      REAL(KIND=dp) :: energy_gap
      REAL(KIND=dp) :: condition_num
   END TYPE preconditioner_type

! **************************************************************************************************
   TYPE preconditioner_p_type
      TYPE(preconditioner_type), POINTER :: preconditioner
   END TYPE preconditioner_p_type

CONTAINS

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param preconditioner ...
!> \retval preconditioner_in_use ...
! **************************************************************************************************
   FUNCTION preconditioner_in_use(preconditioner)
      TYPE(preconditioner_type)                          :: preconditioner
      LOGICAL                                            :: preconditioner_in_use

      preconditioner_in_use = .NOT. (preconditioner%in_use .EQ. 0)
   END FUNCTION

! **************************************************************************************************
!> \brief ...
!> \param preconditioner_env ...
!> \param para_env ...
!> \param blacs_env ...
! **************************************************************************************************
   SUBROUTINE init_preconditioner(preconditioner_env, para_env, blacs_env)

      TYPE(preconditioner_type)                          :: preconditioner_env
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(cp_blacs_env_type), POINTER                   :: blacs_env

      NULLIFY (preconditioner_env%sparse_matrix)
      NULLIFY (preconditioner_env%fm)
      NULLIFY (preconditioner_env%dbcsr_matrix)
      NULLIFY (preconditioner_env%occ_evals)
      NULLIFY (preconditioner_env%full_evals)
      NULLIFY (preconditioner_env%inverse_history)
      NULLIFY (preconditioner_env%max_ev_vector)
      NULLIFY (preconditioner_env%min_ev_vector)
      preconditioner_env%solver = ot_precond_solver_default
      preconditioner_env%para_env => para_env
      preconditioner_env%ctxt => blacs_env
      !inverse is used for filtering in update set it to something huge to
      ! avoid filtering if the information is not available
      preconditioner_env%condition_num = -1.0_dp
      preconditioner_env%ihistory = 0

      CALL cp_para_env_retain(preconditioner_env%para_env)
      CALL cp_blacs_env_retain(preconditioner_env%ctxt)

   END SUBROUTINE init_preconditioner

! **************************************************************************************************
!> \brief ...
!> \param preconditioner_env ...
! **************************************************************************************************
   SUBROUTINE destroy_preconditioner(preconditioner_env)

      TYPE(preconditioner_type)                          :: preconditioner_env

      CHARACTER(len=*), PARAMETER :: routineN = 'destroy_preconditioner', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(preconditioner_env%sparse_matrix)) THEN
         CALL cp_dbcsr_deallocate_matrix(preconditioner_env%sparse_matrix)
         NULLIFY (preconditioner_env%sparse_matrix)
      ENDIF

      IF (ASSOCIATED(preconditioner_env%fm)) THEN
         CALL cp_fm_release(preconditioner_env%fm)
      ENDIF
      IF (ASSOCIATED(preconditioner_env%dbcsr_matrix)) THEN
         CALL cp_dbcsr_release_p(preconditioner_env%dbcsr_matrix)
      ENDIF
      IF (ASSOCIATED(preconditioner_env%max_ev_vector)) THEN
         CALL cp_dbcsr_release_p(preconditioner_env%max_ev_vector)
      ENDIF
      IF (ASSOCIATED(preconditioner_env%min_ev_vector)) THEN
         CALL cp_dbcsr_release_p(preconditioner_env%min_ev_vector)
      ENDIF
      IF (ASSOCIATED(preconditioner_env%occ_evals)) THEN
         DEALLOCATE (preconditioner_env%occ_evals)
      ENDIF
      IF (ASSOCIATED(preconditioner_env%full_evals)) THEN
         DEALLOCATE (preconditioner_env%full_evals)
      ENDIF
      IF (ASSOCIATED(preconditioner_env%inverse_history)) THEN
         DO i = 1, SIZE(preconditioner_env%inverse_history)
            CALL cp_dbcsr_release_p(preconditioner_env%inverse_history(i)%matrix)
         END DO
         DEALLOCATE (preconditioner_env%inverse_history)
      ENDIF
      CALL cp_para_env_release(preconditioner_env%para_env)
      CALL cp_blacs_env_release(preconditioner_env%ctxt)

      preconditioner_env%in_use = 0
      preconditioner_env%cholesky_use = cholesky_reduce

      CALL timestop(handle)

   END SUBROUTINE destroy_preconditioner

END MODULE preconditioner_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief orbital transformations
!> \par History
!>      Added Taylor expansion based computation of the matrix functions (01.2004)
!>      added additional rotation variables for non-equivalent occupied orbs (08.2004)
!> \author Joost VandeVondele (06.2002)
! **************************************************************************************************
MODULE qs_ot_types
   USE bibliography,                    ONLY: VandeVondele2003,&
                                              Weber2008,&
                                              cite_reference
   USE cp_blacs_env,                    ONLY: cp_blacs_env_release,&
                                              cp_blacs_env_retain,&
                                              cp_blacs_env_type
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_allocate_matrix_set, cp_dbcsr_deallocate_matrix_set, cp_dbcsr_init_p, &
        cp_dbcsr_p_type, cp_dbcsr_release_p, cp_dbcsr_set, cp_dbcsr_type, &
        dbcsr_type_complex_default, dbcsr_type_no_symmetry, dbcsr_type_real_4, &
        dbcsr_type_real_default
   USE cp_dbcsr_operations,             ONLY: cp_dbcsr_m_by_n_from_row_template,&
                                              cp_dbcsr_m_by_n_from_template
   USE cp_fm_struct,                    ONLY: cp_fm_struct_get,&
                                              cp_fm_struct_type
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: &
        cholesky_reduce, ls_2pnt, ls_3pnt, ls_gold, ls_none, ot_algo_irac, ot_algo_taylor_or_diag, &
        ot_chol_irac, ot_lwdn_irac, ot_mini_broyden, ot_mini_cg, ot_mini_diis, ot_mini_sd, &
        ot_poly_irac, ot_precond_full_all, ot_precond_full_kinetic, ot_precond_full_single, &
        ot_precond_full_single_inverse, ot_precond_none, ot_precond_s_inverse, &
        ot_precond_solver_default, ot_precond_solver_direct, ot_precond_solver_inv_chol, &
        ot_precond_solver_update
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE preconditioner_types,            ONLY: preconditioner_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   PUBLIC  :: qs_ot_type
   PUBLIC  :: qs_ot_settings_type
   PUBLIC  :: qs_ot_destroy
   PUBLIC  :: qs_ot_allocate
   PUBLIC  :: qs_ot_init
   PUBLIC  :: qs_ot_settings_init
   PUBLIC  :: ot_readwrite_input

   ! notice, this variable needs to be copyable !
   ! needed for spins as e.g. in qs_ot_scf      !
! **************************************************************************************************
   TYPE qs_ot_settings_type
      LOGICAL           :: do_rotation, do_ener
      LOGICAL           :: ks
      CHARACTER(LEN=4)  :: ot_method
      CHARACTER(LEN=3)  :: ot_algorithm
      CHARACTER(LEN=4)  :: line_search_method
      CHARACTER(LEN=20) :: preconditioner_name
      INTEGER           :: preconditioner_type
      INTEGER           :: cholesky_type
      CHARACTER(LEN=20) :: precond_solver_name
      INTEGER           :: precond_solver_type
      LOGICAL           :: safer_diis
      REAL(KIND=dp)   :: ds_min
      REAL(KIND=dp)   :: energy_gap
      INTEGER           :: diis_m
      REAL(KIND=dp)   :: gold_target
      REAL(KIND=dp)   :: eps_taylor ! minimum accuracy of the taylor expansion
      INTEGER           :: max_taylor ! maximum order of the taylor expansion before switching to diagonalization
      INTEGER           :: irac_degree ! this is used to control the refinement polynomial degree
      INTEGER           :: max_irac ! maximum number of iteration for the refinement
      REAL(KIND=dp)   :: eps_irac ! target accuracy for the refinement
      REAL(KIND=dp)   :: eps_irac_quick_exit
      REAL(dp)          :: eps_irac_filter_matrix
      REAL(KIND=dp)   :: eps_irac_switch
      LOGICAL           :: on_the_fly_loc
      CHARACTER(LEN=4)  :: ortho_irac
      LOGICAL           :: mixed_precision
      LOGICAL           :: occupation_preconditioner, add_nondiag_energy
      REAL(KIND=dp)   :: nondiag_energy_strength
      REAL(KIND=dp)   :: broyden_beta, broyden_gamma, broyden_sigma
      REAL(KIND=dp)   :: broyden_eta, broyden_omega, broyden_sigma_decrease
      REAL(KIND=dp)   :: broyden_sigma_min
      LOGICAL           :: broyden_forget_history, broyden_adaptive_sigma
      LOGICAL           :: broyden_enable_flip
   END TYPE qs_ot_settings_type

! **************************************************************************************************
   TYPE qs_ot_type
      ! this sets the method to be used
      TYPE(qs_ot_settings_type) :: settings
      LOGICAL                   :: restricted

      ! first part of the variables, for occupied subspace invariant optimisation

      ! add a preconditioner matrix. should be symmetric and positive definite
      ! the type of this matrix might change in the future
      TYPE(preconditioner_type), POINTER :: preconditioner

      ! these will/might change during iterations

      ! OT / TOD
      TYPE(cp_dbcsr_type), POINTER :: matrix_p
      TYPE(cp_dbcsr_type), POINTER :: matrix_r
      TYPE(cp_dbcsr_type), POINTER :: matrix_sinp
      TYPE(cp_dbcsr_type), POINTER :: matrix_cosp
      TYPE(cp_dbcsr_type), POINTER :: matrix_sinp_b
      TYPE(cp_dbcsr_type), POINTER :: matrix_cosp_b
      TYPE(cp_dbcsr_type), POINTER :: matrix_buf1
      TYPE(cp_dbcsr_type), POINTER :: matrix_buf2
      TYPE(cp_dbcsr_type), POINTER :: matrix_buf3
      TYPE(cp_dbcsr_type), POINTER :: matrix_buf4
      TYPE(cp_dbcsr_type), POINTER :: matrix_os
      TYPE(cp_dbcsr_type), POINTER :: matrix_buf1_ortho
      TYPE(cp_dbcsr_type), POINTER :: matrix_buf2_ortho

      REAL(KIND=dp), DIMENSION(:), POINTER :: evals
      REAL(KIND=dp), DIMENSION(:), POINTER :: dum

      ! matrix os valid
      LOGICAL os_valid

      ! for efficient/parallel writing to the blacs_matrix
      TYPE(cp_para_env_type), POINTER :: para_env
      TYPE(cp_blacs_env_type), POINTER :: blacs_env

      ! mo-like vectors
      TYPE(cp_dbcsr_type), POINTER :: matrix_c0, matrix_sc0, matrix_psc0

      ! OT / IR
      TYPE(cp_dbcsr_type), POINTER :: buf1_k_k_nosym, buf2_k_k_nosym, &
                                      buf3_k_k_nosym, buf4_k_k_nosym, &
                                      buf1_k_k_sym, buf2_k_k_sym, buf3_k_k_sym, buf4_k_k_sym, &
                                      p_k_k_sym, buf1_n_k, buf1_n_k_dp

      ! only here for the ease of programming. These will have to be supplied
      ! explicitly at all times
      TYPE(cp_dbcsr_type), POINTER :: matrix_x, matrix_sx, matrix_gx
      TYPE(cp_dbcsr_type), POINTER :: matrix_dx, matrix_gx_old

      LOGICAL :: use_gx_old, use_dx

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: matrix_h_e, matrix_h_x

      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: ls_diis
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: lss_diis
      REAL(KIND=dp), DIMENSION(:), POINTER  :: c_diis
      REAL(KIND=dp), DIMENSION(:), POINTER  :: c_broy
      REAL(KIND=dp), DIMENSION(:), POINTER  :: energy_h
      INTEGER, DIMENSION(:), POINTER         :: ipivot

      REAL(KIND=dp)  :: ot_pos(53), ot_energy(53), ot_grad(53) ! HARD LIMIT FOR THE LS
      INTEGER          :: line_search_left, line_search_right, line_search_mid
      INTEGER          :: line_search_count
      LOGICAL          :: line_search_might_be_done
      REAL(KIND=dp)  :: delta, gnorm, gnorm_old, etotal, gradient
      LOGICAL          :: energy_only
      INTEGER          :: diis_iter
      CHARACTER(LEN=8) :: OT_METHOD_FULL
      INTEGER          :: OT_count
      REAL(KIND=dp)  :: ds_min
      REAL(KIND=dp)  :: broyden_adaptive_sigma

      LOGICAL          :: do_taylor
      INTEGER          :: taylor_order
      REAL(KIND=dp)  :: largest_eval_upper_bound

      ! second part of the variables, if an explicit rotation is required as well
      TYPE(cp_dbcsr_type), POINTER :: rot_mat_u ! rotation matrix
      TYPE(cp_dbcsr_type), POINTER :: rot_mat_x ! antisymmetric matrix that parametrises rot_matrix_u
      TYPE(cp_dbcsr_type), POINTER :: rot_mat_dedu ! derivative of the total energy wrt to u
      TYPE(cp_dbcsr_type), POINTER :: rot_mat_chc ! for convencience, the matrix c^T H c

      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: rot_mat_h_e
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: rot_mat_h_x
      TYPE(cp_dbcsr_type), POINTER :: rot_mat_gx
      TYPE(cp_dbcsr_type), POINTER :: rot_mat_gx_old
      TYPE(cp_dbcsr_type), POINTER :: rot_mat_dx

      REAL(KIND=dp), DIMENSION(:), POINTER :: rot_mat_evals
      TYPE(cp_dbcsr_type), POINTER :: rot_mat_evec

      ! third part of the variables, if we need to optimize orbital energies
      REAL(KIND=dp), POINTER, DIMENSION(:) :: ener_x
      REAL(KIND=dp), POINTER, DIMENSION(:) :: ener_dx
      REAL(KIND=dp), POINTER, DIMENSION(:) :: ener_gx
      REAL(KIND=dp), POINTER, DIMENSION(:) :: ener_gx_old
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: ener_h_e
      REAL(KIND=dp), POINTER, DIMENSION(:, :) :: ener_h_x
   END TYPE qs_ot_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_ot_types'

CONTAINS

! **************************************************************************************************
!> \brief sets default values for the settings type
!> \param settings ...
!> \par History
!>      10.2004 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE qs_ot_settings_init(settings)
      TYPE(qs_ot_settings_type)                          :: settings

      settings%ot_method = "CG"
      settings%ot_algorithm = "TOD"
      settings%diis_m = 7
      settings%preconditioner_name = "FULL_KINETIC"
      settings%preconditioner_type = ot_precond_full_kinetic
      settings%cholesky_type = cholesky_reduce
      settings%precond_solver_name = "CHOLESKY_INVERSE"
      settings%precond_solver_type = ot_precond_solver_inv_chol
      settings%line_search_method = "2PNT"
      settings%ds_min = 0.15_dp
      settings%safer_diis = .TRUE.
      settings%energy_gap = 0.2_dp
      settings%eps_taylor = 1.0E-16_dp
      settings%max_taylor = 4
      settings%gold_target = 0.01_dp
      settings%do_rotation = .FALSE.
      settings%do_ener = .FALSE.
      settings%mixed_precision = .FALSE.
      settings%irac_degree = 4
      settings%max_irac = 50
      settings%eps_irac = 1.0E-10_dp
      settings%eps_irac_quick_exit = 1.0E-5_dp
      settings%eps_irac_switch = 1.0E-2
      settings%eps_irac_filter_matrix = 0.0_dp
      settings%on_the_fly_loc = .FALSE.
      settings%ortho_irac = "CHOL"
      settings%ks = .TRUE.
      settings%occupation_preconditioner = .FALSE.
      settings%add_nondiag_energy = .FALSE.
      settings%nondiag_energy_strength = 0.0_dp
   END SUBROUTINE qs_ot_settings_init

   ! init matrices, needs c0 and sc0 so that c0*sc0=1
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE qs_ot_init(qs_ot_env)
      TYPE(qs_ot_type)                                   :: qs_ot_env

      qs_ot_env%OT_energy(:) = 0.0_dp
      qs_ot_env%OT_pos(:) = 0.0_dp
      qs_ot_env%OT_grad(:) = 0.0_dp
      qs_ot_env%line_search_count = 0

      qs_ot_env%energy_only = .FALSE.
      qs_ot_env%gnorm_old = 1.0_dp
      qs_ot_env%diis_iter = 0
      qs_ot_env%ds_min = qs_ot_env%settings%ds_min
      qs_ot_env%os_valid = .FALSE.

      CALL cp_dbcsr_set(qs_ot_env%matrix_gx, 0.0_dp)

      IF (qs_ot_env%use_dx) &
         CALL cp_dbcsr_set(qs_ot_env%matrix_dx, 0.0_dp)

      IF (qs_ot_env%use_gx_old) &
         CALL cp_dbcsr_set(qs_ot_env%matrix_gx_old, 0.0_dp)

      IF (qs_ot_env%settings%do_rotation) THEN
         CALL cp_dbcsr_set(qs_ot_env%rot_mat_gx, 0.0_dp)
         IF (qs_ot_env%use_dx) &
            CALL cp_dbcsr_set(qs_ot_env%rot_mat_dx, 0.0_dp)
         IF (qs_ot_env%use_gx_old) &
            CALL cp_dbcsr_set(qs_ot_env%rot_mat_gx_old, 0.0_dp)
      ENDIF
      IF (qs_ot_env%settings%do_ener) THEN
         qs_ot_env%ener_gx(:) = 0.0_dp
         IF (qs_ot_env%use_dx) &
            qs_ot_env%ener_dx(:) = 0.0_dp
         IF (qs_ot_env%use_gx_old) &
            qs_ot_env%ener_gx_old(:) = 0.0_dp
      ENDIF

   END SUBROUTINE qs_ot_init

   ! allocates the data in qs_ot_env, for a calculation with fm_struct_ref
   ! ortho_k allows for specifying an additional orthogonal subspace (i.e. c will
   ! be kept orthogonal provided c0 was, used in qs_ot_eigensolver)
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
!> \param matrix_s ...
!> \param fm_struct_ref ...
!> \param ortho_k ...
! **************************************************************************************************
   SUBROUTINE qs_ot_allocate(qs_ot_env, matrix_s, fm_struct_ref, ortho_k)
      TYPE(qs_ot_type)                                   :: qs_ot_env
      TYPE(cp_dbcsr_type), POINTER                       :: matrix_s
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_ref
      INTEGER, OPTIONAL                                  :: ortho_k

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_allocate', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, k, m_diis, my_ortho_k, n, ncoef, &
                                                            special_data_type
      LOGICAL                                            :: mixed_precision
      TYPE(cp_blacs_env_type), POINTER                   :: context
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CALL cite_reference(VandeVondele2003)

      NULLIFY (qs_ot_env%preconditioner)
      NULLIFY (qs_ot_env%matrix_psc0)
      NULLIFY (qs_ot_env%para_env)
      NULLIFY (qs_ot_env%blacs_env)

      CALL cp_fm_struct_get(fm_struct_ref, nrow_global=n, ncol_global=k, &
                            para_env=para_env, context=context)

      qs_ot_env%para_env => para_env
      qs_ot_env%blacs_env => context
      CALL cp_para_env_retain(para_env)
      CALL cp_blacs_env_retain(context)

      IF (PRESENT(ortho_k)) THEN
         my_ortho_k = ortho_k
      ELSE
         my_ortho_k = k
      ENDIF

      m_diis = qs_ot_env%settings%diis_m
      mixed_precision = qs_ot_env%settings%mixed_precision

      special_data_type = dbcsr_type_real_default
      IF (mixed_precision) special_data_type = dbcsr_type_real_4

      qs_ot_env%use_gx_old = .FALSE.
      qs_ot_env%use_dx = .FALSE.

      SELECT CASE (qs_ot_env%settings%ot_method)
      CASE ("SD")
         ! nothing
      CASE ("CG")
         qs_ot_env%use_gx_old = .TRUE.
         qs_ot_env%use_dx = .TRUE.
      CASE ("DIIS", "BROY")
         IF (m_diis .LT. 1) CALL cp__b("qs_ot_types.F",350,"m_diis less than one")
      CASE DEFAULT
         CALL cp__b("qs_ot_types.F",352,"Unknown option")
      END SELECT

      IF (qs_ot_env%settings%ot_method .EQ. "DIIS" .OR. &
          qs_ot_env%settings%ot_method .EQ. "BROY") THEN
         ALLOCATE (qs_ot_env%ls_diis(m_diis+1, m_diis+1))
         qs_ot_env%ls_diis = 0.0_dp
         ALLOCATE (qs_ot_env%lss_diis(m_diis+1, m_diis+1))
         ALLOCATE (qs_ot_env%c_diis(m_diis+1))
         ALLOCATE (qs_ot_env%c_broy(m_diis))
         ALLOCATE (qs_ot_env%energy_h(m_diis))
         ALLOCATE (qs_ot_env%ipivot(m_diis+1))
      ENDIF

      ALLOCATE (qs_ot_env%evals(k))
      ALLOCATE (qs_ot_env%dum(k))

      NULLIFY (qs_ot_env%matrix_os)
      NULLIFY (qs_ot_env%matrix_buf1_ortho)
      NULLIFY (qs_ot_env%matrix_buf2_ortho)
      NULLIFY (qs_ot_env%matrix_p)
      NULLIFY (qs_ot_env%matrix_r)
      NULLIFY (qs_ot_env%matrix_sinp)
      NULLIFY (qs_ot_env%matrix_cosp)
      NULLIFY (qs_ot_env%matrix_sinp_b)
      NULLIFY (qs_ot_env%matrix_cosp_b)
      NULLIFY (qs_ot_env%matrix_buf1)
      NULLIFY (qs_ot_env%matrix_buf2)
      NULLIFY (qs_ot_env%matrix_buf3)
      NULLIFY (qs_ot_env%matrix_buf4)
      NULLIFY (qs_ot_env%matrix_c0)
      NULLIFY (qs_ot_env%matrix_sc0)
      NULLIFY (qs_ot_env%matrix_x)
      NULLIFY (qs_ot_env%matrix_sx)
      NULLIFY (qs_ot_env%matrix_gx)
      NULLIFY (qs_ot_env%matrix_gx_old)
      NULLIFY (qs_ot_env%matrix_dx)
      NULLIFY (qs_ot_env%buf1_k_k_nosym)
      NULLIFY (qs_ot_env%buf2_k_k_nosym)
      NULLIFY (qs_ot_env%buf3_k_k_nosym)
      NULLIFY (qs_ot_env%buf4_k_k_nosym)
      NULLIFY (qs_ot_env%buf1_k_k_sym)
      NULLIFY (qs_ot_env%buf2_k_k_sym)
      NULLIFY (qs_ot_env%buf3_k_k_sym)
      NULLIFY (qs_ot_env%buf4_k_k_sym)
      NULLIFY (qs_ot_env%buf1_n_k)
      NULLIFY (qs_ot_env%buf1_n_k_dp)
      NULLIFY (qs_ot_env%p_k_k_sym)

      ! COMMON MATRICES
      CALL cp_dbcsr_init_p(qs_ot_env%matrix_c0)
      CALL cp_dbcsr_m_by_n_from_row_template(qs_ot_env%matrix_c0, template=matrix_s, n=k, &
                                             sym=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_init_p(qs_ot_env%matrix_sc0)
      CALL cp_dbcsr_m_by_n_from_row_template(qs_ot_env%matrix_sc0, template=matrix_s, n=my_ortho_k, &
                                             sym=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_init_p(qs_ot_env%matrix_x)
      CALL cp_dbcsr_m_by_n_from_row_template(qs_ot_env%matrix_x, template=matrix_s, n=k, &
                                             sym=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_init_p(qs_ot_env%matrix_sx)
      CALL cp_dbcsr_m_by_n_from_row_template(qs_ot_env%matrix_sx, template=matrix_s, n=k, &
                                             sym=dbcsr_type_no_symmetry)

      CALL cp_dbcsr_init_p(qs_ot_env%matrix_gx)
      CALL cp_dbcsr_m_by_n_from_row_template(qs_ot_env%matrix_gx, template=matrix_s, n=k, &
                                             sym=dbcsr_type_no_symmetry, data_type=special_data_type)

      IF (qs_ot_env%use_dx) THEN
         CALL cp_dbcsr_init_p(qs_ot_env%matrix_dx)
         CALL cp_dbcsr_m_by_n_from_row_template(qs_ot_env%matrix_dx, template=matrix_s, n=k, &
                                                sym=dbcsr_type_no_symmetry, data_type=special_data_type)
      ENDIF

      IF (qs_ot_env%use_gx_old) THEN
         CALL cp_dbcsr_init_p(qs_ot_env%matrix_gx_old)
         CALL cp_dbcsr_m_by_n_from_row_template(qs_ot_env%matrix_gx_old, template=matrix_s, n=k, &
                                                sym=dbcsr_type_no_symmetry, data_type=special_data_type)
      ENDIF

      SELECT CASE (qs_ot_env%settings%ot_algorithm)
      CASE ("TOD")
         CALL cp_dbcsr_init_p(qs_ot_env%matrix_p)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_p, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_r)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_r, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_sinp)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_sinp, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_cosp)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_cosp, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_sinp_b)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_sinp_b, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_cosp_b)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_cosp_b, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_buf1)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_buf1, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_buf2)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_buf2, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_buf3)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_buf3, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_buf4)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_buf4, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_os)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_os, template=matrix_s, m=my_ortho_k, n=my_ortho_k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_buf1_ortho)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_buf1_ortho, template=matrix_s, m=my_ortho_k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%matrix_buf2_ortho)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%matrix_buf2_ortho, template=matrix_s, m=my_ortho_k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

      CASE ("REF")
         CALL cp_dbcsr_init_p(qs_ot_env%buf1_k_k_nosym)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%buf1_k_k_nosym, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%buf2_k_k_nosym)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%buf2_k_k_nosym, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%buf3_k_k_nosym)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%buf3_k_k_nosym, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%buf4_k_k_nosym)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%buf4_k_k_nosym, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

! It claims to be symmetric but to avoid dbcsr conusion nonsymmetric is kept
         CALL cp_dbcsr_init_p(qs_ot_env%buf1_k_k_sym)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%buf1_k_k_sym, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%buf2_k_k_sym)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%buf2_k_k_sym, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%buf3_k_k_sym)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%buf3_k_k_sym, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)
         !
         CALL cp_dbcsr_init_p(qs_ot_env%buf4_k_k_sym)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%buf4_k_k_sym, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)
         !
         CALL cp_dbcsr_init_p(qs_ot_env%p_k_k_sym)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%p_k_k_sym, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)
         !
         CALL cp_dbcsr_init_p(qs_ot_env%buf1_n_k)
         CALL cp_dbcsr_m_by_n_from_row_template(qs_ot_env%buf1_n_k, template=matrix_s, n=k, &
                                                sym=dbcsr_type_no_symmetry)
         !
         IF (mixed_precision) THEN
            CALL cp_dbcsr_init_p(qs_ot_env%buf1_n_k_dp)
            CALL cp_dbcsr_m_by_n_from_row_template(qs_ot_env%buf1_n_k_dp, template=matrix_s, n=k, &
                                                   sym=dbcsr_type_no_symmetry)
         ENDIF
         !

      END SELECT

      IF (qs_ot_env%settings%ot_method .EQ. "DIIS" .OR. &
          qs_ot_env%settings%ot_method .EQ. "BROY") THEN
         NULLIFY (qs_ot_env%matrix_h_e)
         NULLIFY (qs_ot_env%matrix_h_x)
         CALL cp_dbcsr_allocate_matrix_set(qs_ot_env%matrix_h_e, m_diis)
         CALL cp_dbcsr_allocate_matrix_set(qs_ot_env%matrix_h_x, m_diis)
         DO i = 1, m_diis
            CALL cp_dbcsr_init_p(qs_ot_env%matrix_h_x(i)%matrix)
            CALL cp_dbcsr_m_by_n_from_row_template(qs_ot_env%matrix_h_x(i)%matrix, template=matrix_s, n=k, &
                                                   sym=dbcsr_type_no_symmetry, data_type=special_data_type)

            CALL cp_dbcsr_init_p(qs_ot_env%matrix_h_e(i)%matrix)
            CALL cp_dbcsr_m_by_n_from_row_template(qs_ot_env%matrix_h_e(i)%matrix, template=matrix_s, n=k, &
                                                   sym=dbcsr_type_no_symmetry, data_type=special_data_type)
         ENDDO
      ENDIF

      NULLIFY (qs_ot_env%rot_mat_u, qs_ot_env%rot_mat_x, qs_ot_env%rot_mat_h_e, qs_ot_env%rot_mat_h_x, &
               qs_ot_env%rot_mat_gx, qs_ot_env%rot_mat_gx_old, qs_ot_env%rot_mat_dx, &
               qs_ot_env%rot_mat_evals, qs_ot_env%rot_mat_evec, qs_ot_env%rot_mat_dedu, qs_ot_env%rot_mat_chc)

      IF (qs_ot_env%settings%do_rotation) THEN
         CALL cp_dbcsr_init_p(qs_ot_env%rot_mat_u)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%rot_mat_u, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%rot_mat_x)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%rot_mat_x, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%rot_mat_dedu)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%rot_mat_dedu, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         CALL cp_dbcsr_init_p(qs_ot_env%rot_mat_chc)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%rot_mat_chc, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         IF (qs_ot_env%settings%ot_method .EQ. "DIIS") THEN
            CALL cp_dbcsr_allocate_matrix_set(qs_ot_env%rot_mat_h_e, m_diis)
            CALL cp_dbcsr_allocate_matrix_set(qs_ot_env%rot_mat_h_x, m_diis)
            DO i = 1, m_diis
               CALL cp_dbcsr_init_p(qs_ot_env%rot_mat_h_e(i)%matrix)
               CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%rot_mat_h_e(i)%matrix, template=matrix_s, m=k, n=k, &
                                                  sym=dbcsr_type_no_symmetry)

               CALL cp_dbcsr_init_p(qs_ot_env%rot_mat_h_x(i)%matrix)
               CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%rot_mat_h_x(i)%matrix, template=matrix_s, m=k, n=k, &
                                                  sym=dbcsr_type_no_symmetry)
            ENDDO
         ENDIF

         ALLOCATE (qs_ot_env%rot_mat_evals(k))

         CALL cp_dbcsr_init_p(qs_ot_env%rot_mat_evec)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%rot_mat_evec, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry, data_type=dbcsr_type_complex_default)

         CALL cp_dbcsr_init_p(qs_ot_env%rot_mat_gx)
         CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%rot_mat_gx, template=matrix_s, m=k, n=k, &
                                            sym=dbcsr_type_no_symmetry)

         IF (qs_ot_env%use_gx_old) THEN
            CALL cp_dbcsr_init_p(qs_ot_env%rot_mat_gx_old)
            CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%rot_mat_gx_old, template=matrix_s, m=k, n=k, &
                                               sym=dbcsr_type_no_symmetry)
         ENDIF

         IF (qs_ot_env%use_dx) THEN
            CALL cp_dbcsr_init_p(qs_ot_env%rot_mat_dx)
            CALL cp_dbcsr_m_by_n_from_template(qs_ot_env%rot_mat_dx, template=matrix_s, m=k, n=k, &
                                               sym=dbcsr_type_no_symmetry)
         ENDIF

      ENDIF

      IF (qs_ot_env%settings%do_ener) THEN
         ncoef = k
         ALLOCATE (qs_ot_env%ener_x(ncoef))

         IF (qs_ot_env%settings%ot_method .EQ. "DIIS") THEN
            ALLOCATE (qs_ot_env%ener_h_e(m_diis, ncoef))
            ALLOCATE (qs_ot_env%ener_h_x(m_diis, ncoef))
         ENDIF

         ALLOCATE (qs_ot_env%ener_gx(ncoef))

         IF (qs_ot_env%use_gx_old) THEN
            ALLOCATE (qs_ot_env%ener_gx_old(ncoef))
         ENDIF

         IF (qs_ot_env%use_dx) THEN
            ALLOCATE (qs_ot_env%ener_dx(ncoef))
            qs_ot_env%ener_dx = 0.0_dp
         ENDIF
      ENDIF

   END SUBROUTINE qs_ot_allocate

   ! deallocates data
! **************************************************************************************************
!> \brief ...
!> \param qs_ot_env ...
! **************************************************************************************************
   SUBROUTINE qs_ot_destroy(qs_ot_env)
      TYPE(qs_ot_type)                                   :: qs_ot_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_ot_destroy', routineP = moduleN//':'//routineN

      CALL cp_para_env_release(qs_ot_env%para_env)
      CALL cp_blacs_env_release(qs_ot_env%blacs_env)

      DEALLOCATE (qs_ot_env%evals)
      DEALLOCATE (qs_ot_env%dum)

      IF (ASSOCIATED(qs_ot_env%matrix_os)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_os)
      IF (ASSOCIATED(qs_ot_env%matrix_p)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_p)
      IF (ASSOCIATED(qs_ot_env%matrix_cosp)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_cosp)
      IF (ASSOCIATED(qs_ot_env%matrix_sinp)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_sinp)
      IF (ASSOCIATED(qs_ot_env%matrix_r)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_r)
      IF (ASSOCIATED(qs_ot_env%matrix_cosp_b)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_cosp_b)
      IF (ASSOCIATED(qs_ot_env%matrix_sinp_b)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_sinp_b)
      IF (ASSOCIATED(qs_ot_env%matrix_buf1)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_buf1)
      IF (ASSOCIATED(qs_ot_env%matrix_buf2)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_buf2)
      IF (ASSOCIATED(qs_ot_env%matrix_buf3)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_buf3)
      IF (ASSOCIATED(qs_ot_env%matrix_buf4)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_buf4)
      IF (ASSOCIATED(qs_ot_env%matrix_buf1_ortho)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_buf1_ortho)
      IF (ASSOCIATED(qs_ot_env%matrix_buf2_ortho)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_buf2_ortho)
      IF (ASSOCIATED(qs_ot_env%matrix_c0)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_c0)
      IF (ASSOCIATED(qs_ot_env%matrix_sc0)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_sc0)
      IF (ASSOCIATED(qs_ot_env%matrix_psc0)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_psc0)
      IF (ASSOCIATED(qs_ot_env%matrix_x)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_x)
      IF (ASSOCIATED(qs_ot_env%matrix_sx)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_sx)
      IF (ASSOCIATED(qs_ot_env%matrix_gx)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_gx)
      IF (ASSOCIATED(qs_ot_env%matrix_dx)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_dx)
      IF (ASSOCIATED(qs_ot_env%matrix_gx_old)) CALL cp_dbcsr_release_p(qs_ot_env%matrix_gx_old)
      IF (ASSOCIATED(qs_ot_env%buf1_k_k_nosym)) CALL cp_dbcsr_release_p(qs_ot_env%buf1_k_k_nosym)
      IF (ASSOCIATED(qs_ot_env%buf2_k_k_nosym)) CALL cp_dbcsr_release_p(qs_ot_env%buf2_k_k_nosym)
      IF (ASSOCIATED(qs_ot_env%buf3_k_k_nosym)) CALL cp_dbcsr_release_p(qs_ot_env%buf3_k_k_nosym)
      IF (ASSOCIATED(qs_ot_env%buf4_k_k_nosym)) CALL cp_dbcsr_release_p(qs_ot_env%buf4_k_k_nosym)
      IF (ASSOCIATED(qs_ot_env%p_k_k_sym)) CALL cp_dbcsr_release_p(qs_ot_env%p_k_k_sym)
      IF (ASSOCIATED(qs_ot_env%buf1_k_k_sym)) CALL cp_dbcsr_release_p(qs_ot_env%buf1_k_k_sym)
      IF (ASSOCIATED(qs_ot_env%buf2_k_k_sym)) CALL cp_dbcsr_release_p(qs_ot_env%buf2_k_k_sym)
      IF (ASSOCIATED(qs_ot_env%buf3_k_k_sym)) CALL cp_dbcsr_release_p(qs_ot_env%buf3_k_k_sym)
      IF (ASSOCIATED(qs_ot_env%buf4_k_k_sym)) CALL cp_dbcsr_release_p(qs_ot_env%buf4_k_k_sym)
      IF (ASSOCIATED(qs_ot_env%buf1_n_k)) CALL cp_dbcsr_release_p(qs_ot_env%buf1_n_k)
      IF (ASSOCIATED(qs_ot_env%buf1_n_k_dp)) CALL cp_dbcsr_release_p(qs_ot_env%buf1_n_k_dp)

      IF (qs_ot_env%settings%ot_method .EQ. "DIIS" .OR. &
          qs_ot_env%settings%ot_method .EQ. "BROY") THEN
         CALL cp_dbcsr_deallocate_matrix_set(qs_ot_env%matrix_h_x)
         CALL cp_dbcsr_deallocate_matrix_set(qs_ot_env%matrix_h_e)
         DEALLOCATE (qs_ot_env%ls_diis)
         DEALLOCATE (qs_ot_env%lss_diis)
         DEALLOCATE (qs_ot_env%c_diis)
         DEALLOCATE (qs_ot_env%c_broy)
         DEALLOCATE (qs_ot_env%energy_h)
         DEALLOCATE (qs_ot_env%ipivot)
      ENDIF

      IF (qs_ot_env%settings%do_rotation) THEN

         IF (ASSOCIATED(qs_ot_env%rot_mat_u)) CALL cp_dbcsr_release_p(qs_ot_env%rot_mat_u)
         IF (ASSOCIATED(qs_ot_env%rot_mat_x)) CALL cp_dbcsr_release_p(qs_ot_env%rot_mat_x)
         IF (ASSOCIATED(qs_ot_env%rot_mat_dedu)) CALL cp_dbcsr_release_p(qs_ot_env%rot_mat_dedu)
         IF (ASSOCIATED(qs_ot_env%rot_mat_chc)) CALL cp_dbcsr_release_p(qs_ot_env%rot_mat_chc)

         IF (qs_ot_env%settings%ot_method .EQ. "DIIS") THEN
            CALL cp_dbcsr_deallocate_matrix_set(qs_ot_env%rot_mat_h_x)
            CALL cp_dbcsr_deallocate_matrix_set(qs_ot_env%rot_mat_h_e)
         ENDIF

         DEALLOCATE (qs_ot_env%rot_mat_evals)

         IF (ASSOCIATED(qs_ot_env%rot_mat_evec)) CALL cp_dbcsr_release_p(qs_ot_env%rot_mat_evec)
         IF (ASSOCIATED(qs_ot_env%rot_mat_gx)) CALL cp_dbcsr_release_p(qs_ot_env%rot_mat_gx)
         IF (ASSOCIATED(qs_ot_env%rot_mat_gx_old)) CALL cp_dbcsr_release_p(qs_ot_env%rot_mat_gx_old)
         IF (ASSOCIATED(qs_ot_env%rot_mat_dx)) CALL cp_dbcsr_release_p(qs_ot_env%rot_mat_dx)
      ENDIF

      IF (qs_ot_env%settings%do_ener) THEN
         DEALLOCATE (qs_ot_env%ener_x)
         DEALLOCATE (qs_ot_env%ener_gx)
         IF (qs_ot_env%settings%ot_method .EQ. "DIIS") THEN
            DEALLOCATE (qs_ot_env%ener_h_x)
            DEALLOCATE (qs_ot_env%ener_h_e)
         ENDIF
         IF (qs_ot_env%use_dx) THEN
            DEALLOCATE (qs_ot_env%ener_dx)
         ENDIF
         IF (qs_ot_env%use_gx_old) THEN
            DEALLOCATE (qs_ot_env%ener_gx_old)
         ENDIF
      END IF

   END SUBROUTINE qs_ot_destroy

! **************************************************************************************************
!> \brief ...
!> \param settings ...
!> \param ot_section ...
!> \param output_unit ...
! **************************************************************************************************
   SUBROUTINE ot_readwrite_input(settings, ot_section, output_unit)
      TYPE(qs_ot_settings_type)                          :: settings
      TYPE(section_vals_type), POINTER                   :: ot_section
      INTEGER, INTENT(IN)                                :: output_unit

      CHARACTER(len=*), PARAMETER :: routineN = 'ot_readwrite_input', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ls_method, ot_algorithm, &
                                                            ot_method, ot_ortho_irac

      CALL timeset(routineN, handle)

      ! choose algorithm
      CALL section_vals_val_get(ot_section, "ALGORITHM", i_val=ot_algorithm)
      SELECT CASE (ot_algorithm)
      CASE (ot_algo_taylor_or_diag)
         settings%ot_algorithm = "TOD"
      CASE (ot_algo_irac)
         CALL cite_reference(Weber2008)
         settings%ot_algorithm = "REF"
      CASE DEFAULT
         CALL cp__b("qs_ot_types.F",764,"Value unknown")
      END SELECT

      ! irac input
      CALL section_vals_val_get(ot_section, "IRAC_DEGREE", i_val=settings%irac_degree)
      IF (settings%irac_degree < 2 .OR. settings%irac_degree > 4) THEN
         CALL cp__b("qs_ot_types.F",770,"READ OT IRAC_DEGREE: Value unknown")
      ENDIF
      CALL section_vals_val_get(ot_section, "MAX_IRAC", i_val=settings%max_irac)
      IF (settings%max_irac < 1) THEN
         CALL cp__b("qs_ot_types.F",774,"READ OT MAX_IRAC: VALUE MUST BE GREATER THAN ZERO")
      ENDIF
      CALL section_vals_val_get(ot_section, "EPS_IRAC_FILTER_MATRIX", r_val=settings%eps_irac_filter_matrix)
      CALL section_vals_val_get(ot_section, "EPS_IRAC", r_val=settings%eps_irac)
      IF (settings%eps_irac < 0.0_dp) THEN
         CALL cp__b("qs_ot_types.F",779,"READ OT EPS_IRAC: VALUE MUST BE GREATER THAN ZERO")
      ENDIF
      CALL section_vals_val_get(ot_section, "EPS_IRAC_QUICK_EXIT", r_val=settings%eps_irac_quick_exit)
      IF (settings%eps_irac_quick_exit < 0.0_dp) THEN
         CALL cp__b("qs_ot_types.F",783,"READ OT EPS_IRAC_QUICK_EXIT: VALUE MUST BE GREATER THAN ZERO")
      ENDIF

      CALL section_vals_val_get(ot_section, "EPS_IRAC_SWITCH", r_val=settings%eps_irac_switch)
      IF (settings%eps_irac_switch < 0.0_dp) THEN
         CALL cp__b("qs_ot_types.F",788,"READ OT EPS_IRAC_SWITCH: VALUE MUST BE GREATER THAN ZERO")
      ENDIF

      CALL section_vals_val_get(ot_section, "MIXED_PRECISION", l_val=settings%mixed_precision)
      IF (settings%mixed_precision .AND. ot_algorithm .NE. ot_algo_irac) THEN
         CALL cp__b("qs_ot_types.F",793,"MIXED_PRECISION implemented only for IRAC so far")
      ENDIF

      CALL section_vals_val_get(ot_section, "ORTHO_IRAC", i_val=ot_ortho_irac)
      SELECT CASE (ot_ortho_irac)
      CASE (ot_chol_irac)
         settings%ortho_irac = "CHOL"
      CASE (ot_poly_irac)
         settings%ortho_irac = "POLY"
      CASE (ot_lwdn_irac)
         settings%ortho_irac = "LWDN"
      CASE DEFAULT
         CALL cp__b("qs_ot_types.F",805,"READ OT ORTHO_IRAC: Value unknown")
      END SELECT

      CALL section_vals_val_get(ot_section, "ON_THE_FLY_LOC", l_val=settings%on_the_fly_loc)

      CALL section_vals_val_get(ot_section, "MINIMIZER", i_val=ot_method)
      ! compatibility
      SELECT CASE (ot_method)
      CASE (ot_mini_sd)
         settings%ot_method = "SD"
      CASE (ot_mini_cg)
         settings%ot_method = "CG"
      CASE (ot_mini_diis)
         settings%ot_method = "DIIS"
         CALL section_vals_val_get(ot_section, "N_HISTORY_VEC", i_val=settings%diis_m)
      CASE (ot_mini_broyden)
         CALL section_vals_val_get(ot_section, "N_HISTORY_VEC", i_val=settings%diis_m)
         CALL section_vals_val_get(ot_section, "BROYDEN_BETA", r_val=settings%broyden_beta)
         CALL section_vals_val_get(ot_section, "BROYDEN_GAMMA", r_val=settings%broyden_gamma)
         CALL section_vals_val_get(ot_section, "BROYDEN_SIGMA", r_val=settings%broyden_sigma)
         CALL section_vals_val_get(ot_section, "BROYDEN_ETA", r_val=settings%broyden_eta)
         CALL section_vals_val_get(ot_section, "BROYDEN_OMEGA", r_val=settings%broyden_omega)
         CALL section_vals_val_get(ot_section, "BROYDEN_SIGMA_DECREASE", r_val=settings%broyden_sigma_decrease)
         CALL section_vals_val_get(ot_section, "BROYDEN_SIGMA_MIN", r_val=settings%broyden_sigma_min)
         CALL section_vals_val_get(ot_section, "BROYDEN_FORGET_HISTORY", l_val=settings%broyden_forget_history)
         CALL section_vals_val_get(ot_section, "BROYDEN_ADAPTIVE_SIGMA", l_val=settings%broyden_adaptive_sigma)
         CALL section_vals_val_get(ot_section, "BROYDEN_ENABLE_FLIP", l_val=settings%broyden_enable_flip)
         settings%ot_method = "BROY"
      CASE DEFAULT
         CALL cp__b("qs_ot_types.F",834,"READ OTSCF MINIMIZER: Value unknown")
      END SELECT
      CALL section_vals_val_get(ot_section, "SAFER_DIIS", l_val=settings%safer_diis)
      CALL section_vals_val_get(ot_section, "LINESEARCH", i_val=ls_method)
      SELECT CASE (ls_method)
      CASE (ls_none)
         settings%line_search_method = "NONE"
      CASE (ls_2pnt)
         settings%line_search_method = "2PNT"
      CASE (ls_3pnt)
         settings%line_search_method = "3PNT"
      CASE (ls_gold)
         settings%line_search_method = "GOLD"
         CALL section_vals_val_get(ot_section, "GOLD_TARGET", r_val=settings%gold_target)
      CASE DEFAULT
         CALL cp__b("qs_ot_types.F",849,"READ OTSCF LS: Value unknown")
      END SELECT

      CALL section_vals_val_get(ot_section, "PRECOND_SOLVER", i_val=settings%precond_solver_type)
      SELECT CASE (settings%precond_solver_type)
      CASE (ot_precond_solver_default)
         settings%precond_solver_name = "DEFAULT"
      CASE (ot_precond_solver_inv_chol)
         settings%precond_solver_name = "INVERSE_CHOLESKY"
      CASE (ot_precond_solver_direct)
         settings%precond_solver_name = "DIRECT"
      CASE (ot_precond_solver_update)
         settings%precond_solver_name = "INVERSE_UPDATE"
      CASE DEFAULT
         CALL cp__b("qs_ot_types.F",863,"READ OTSCF SOLVER: Value unknown")
      END SELECT

      !If these values are negative we will set them "optimal" for a given precondtioner below
      CALL section_vals_val_get(ot_section, "STEPSIZE", r_val=settings%ds_min)
      CALL section_vals_val_get(ot_section, "ENERGY_GAP", r_val=settings%energy_gap)

      CALL section_vals_val_get(ot_section, "PRECONDITIONER", i_val=settings%preconditioner_type)
      SELECT CASE (settings%preconditioner_type)
      CASE (ot_precond_none)
         settings%preconditioner_name = "NONE"
         IF (settings%ds_min < 0.0_dp) settings%ds_min = 0.15_dp
         IF (settings%energy_gap < 0.0_dp) settings%energy_gap = 0.2_dp
      CASE (ot_precond_full_single)
         settings%preconditioner_name = "FULL_SINGLE"
         IF (settings%ds_min < 0.0_dp) settings%ds_min = 0.15_dp
         IF (settings%energy_gap < 0.0_dp) settings%energy_gap = 0.2_dp
      CASE (ot_precond_full_single_inverse)
         settings%preconditioner_name = "FULL_SINGLE_INVERSE"
         IF (settings%ds_min < 0.0_dp) settings%ds_min = 0.08_dp
         IF (settings%energy_gap < 0.0_dp) settings%energy_gap = 0.08_dp
      CASE (ot_precond_full_all)
         settings%preconditioner_name = "FULL_ALL"
         IF (settings%ds_min < 0.0_dp) settings%ds_min = 0.15_dp
         IF (settings%energy_gap < 0.0_dp) settings%energy_gap = 0.08_dp
      CASE (ot_precond_full_kinetic)
         settings%preconditioner_name = "FULL_KINETIC"
         IF (settings%ds_min < 0.0_dp) settings%ds_min = 0.15_dp
         IF (settings%energy_gap < 0.0_dp) settings%energy_gap = 0.2_dp
      CASE (ot_precond_s_inverse)
         settings%preconditioner_name = "FULL_S_INVERSE"
         IF (settings%ds_min < 0.0_dp) settings%ds_min = 0.15_dp
         IF (settings%energy_gap < 0.0_dp) settings%energy_gap = 0.2_dp
      CASE DEFAULT
         CALL cp__b("qs_ot_types.F",897,"READ OTSCF PRECONDITIONER: Value unknown")
      END SELECT
      CALL section_vals_val_get(ot_section, "CHOLESKY", i_val=settings%cholesky_type)
      CALL section_vals_val_get(ot_section, "EPS_TAYLOR", r_val=settings%eps_taylor)
      CALL section_vals_val_get(ot_section, "MAX_TAYLOR", i_val=settings%max_taylor)
      CALL section_vals_val_get(ot_section, "ROTATION", l_val=settings%do_rotation)
      CALL section_vals_val_get(ot_section, "ENERGIES", l_val=settings%do_ener)
      CALL section_vals_val_get(ot_section, "OCCUPATION_PRECONDITIONER", &
                                l_val=settings%occupation_preconditioner)
      CALL section_vals_val_get(ot_section, "NONDIAG_ENERGY", l_val=settings%add_nondiag_energy)
      CALL section_vals_val_get(ot_section, "NONDIAG_ENERGY_STRENGTH", &
                                r_val=settings%nondiag_energy_strength)
      ! not yet fully implemented
      IF(.NOT.(.NOT. settings%do_ener))CALL cp__a("qs_ot_types.F",910)

      ! write OT output

      IF (output_unit > 0) THEN
         WRITE (output_unit, '(/,A)') "  ----------------------------------- OT ---------------------------------------"
         IF (settings%do_rotation) THEN
            WRITE (output_unit, '(A)') "  Allowing for rotations "
         END IF
         IF (settings%do_ener) THEN
            WRITE (output_unit, '(A,L2)') "  Optimizing orbital energies "
         END IF
         SELECT CASE (settings%OT_METHOD)
         CASE ("SD")
            WRITE (output_unit, '(A)') "  Minimizer      : SD                  : steepest descent"
         CASE ("CG")
            WRITE (output_unit, '(A)') "  Minimizer      : CG                  : conjugate gradient"
         CASE ("DIIS")
            WRITE (output_unit, '(A)') "  Minimizer      : DIIS                : direct inversion"
            WRITE (output_unit, '(A)') "                                         in the iterative subspace"
            WRITE (output_unit, '(A,I3,A)') "                                         using ", settings%diis_m, " DIIS vectors"
            IF (settings%safer_diis) THEN
               WRITE (output_unit, '(A,I3,A)') "                                         safer DIIS on"
            ELSE
               WRITE (output_unit, '(A,I3,A)') "                                         safer DIIS off"
            ENDIF
         CASE ("BROY")
            WRITE (output_unit, '(A)') "  Minimizer      : BROYDEN             : Broyden "
            WRITE (output_unit, '(A,F16.8)') "                   BETA                : ", settings%broyden_beta
            WRITE (output_unit, '(A,F16.8)') "                   GAMMA               : ", settings%broyden_gamma
            WRITE (output_unit, '(A,F16.8)') "                   SIGMA               : ", settings%broyden_sigma
            WRITE (output_unit, '(A,I3,A)') "                            using      : - ", &
               settings%diis_m, " BROYDEN vectors"
         CASE DEFAULT
            WRITE (output_unit, '(3A)') "  Minimizer      :      ", settings%OT_METHOD, " : UNKNOWN"
         END SELECT
         SELECT CASE (settings%preconditioner_name)
         CASE ("FULL_SINGLE")
            WRITE (output_unit, '(A)') "  Preconditioner : FULL_SINGLE         : diagonalization based"
         CASE ("FULL_SINGLE_INVERSE")
            WRITE (output_unit, '(A,/,A)') "  Preconditioner : FULL_SINGLE_INVERSE : inversion of ", &
               "                                         H + eS - 2*(Sc)(c^T*H*c+const)(Sc)^T"
         CASE ("FULL_ALL")
            WRITE (output_unit, '(A)') "  Preconditioner : FULL_ALL            : diagonalization, state selective"
         CASE ("FULL_KINETIC")
            WRITE (output_unit, '(A)') "  Preconditioner : FULL_KINETIC        : inversion of T + eS"
         CASE ("FULL_S_INVERSE")
            WRITE (output_unit, '(A)') "  Preconditioner : FULL_S_INVERSE      : cholesky inversion of S"
         CASE ("SPARSE_DIAG")
            WRITE (output_unit, '(A)') &
               "  Preconditioner : SPARSE_DIAG    : diagonal atomic block diagonalization"
         CASE ("SPARSE_KINETIC")
            WRITE (output_unit, '(A)') "  Preconditioner : SPARSE_KINETIC      : sparse linear solver for T + eS"
         CASE ("NONE")
            WRITE (output_unit, '(A)') "  Preconditioner : NONE"
         CASE DEFAULT
            WRITE (output_unit, '(3A)') "  Preconditioner : ", settings%preconditioner_name, " : UNKNOWN"
         END SELECT

         WRITE (output_unit, '(A)') "  Precond_solver : "//TRIM(settings%precond_solver_name)

         IF (settings%OT_METHOD .EQ. "SD" .OR. settings%OT_METHOD .EQ. "CG") THEN
            SELECT CASE (settings%line_search_method)
            CASE ("2PNT")
               WRITE (output_unit, '(A)') "  Line search    : 2PNT                : 2 energies, one gradient"
            CASE ("3PNT")
               WRITE (output_unit, '(A)') "  Line search    : 3PNT                : 3 energies"
            CASE ("GOLD")
               WRITE (output_unit, '(A)') "  Line search    : GOLD                : bracketing and golden section search"
               WRITE (output_unit, '(A,F14.8)') "                   target rel accuracy : ", settings%gold_target
            CASE ("NONE")
               WRITE (output_unit, '(A)') "  Line search    : NONE"
            CASE DEFAULT
               WRITE (output_unit, '(3A)') "  Line search : ", settings%line_search_method, " : UNKNOWN"
            END SELECT
         ENDIF
         WRITE (output_unit, '(A,F14.8,T49,A,F14.8)') "  stepsize       :", settings%ds_min, &
            "  energy_gap     :", settings%energy_gap
         IF (settings%ot_algorithm .EQ. 'TOD') THEN
            WRITE (output_unit, '(A,E14.5,T49,A,I14)') "  eps_taylor     :", settings%eps_taylor, &
               "  max_taylor     :", settings%max_taylor
         ENDIF
         IF (settings%ot_algorithm .EQ. 'REF') THEN
            WRITE (output_unit, '(A,1X,A,T49,A,I14)') "  ortho_irac         :", settings%ortho_irac, &
               "  irac_degree        :", settings%irac_degree
            WRITE (output_unit, '(A,I14,T49,A,E14.5)') "  max_irac           :", settings%max_irac, &
               "  eps_irac           :", settings%eps_irac
            WRITE (output_unit, '(A,E14.5,T49,A,E14.5)') "  eps_irac_switch    :", settings%eps_irac_switch, &
               "  eps_irac_quick_exit:", settings%eps_irac_quick_exit
            WRITE (output_unit, '(A,L2)') "  on_the_fly_loc     :", settings%on_the_fly_loc
         END IF
         IF (settings%mixed_precision) THEN
            WRITE (output_unit, '(A)') "  Preconditioner arithmetic in mixed_precision    "
         END IF
         WRITE (output_unit, '(A)') "  ----------------------------------- OT ---------------------------------------"
         WRITE (UNIT=output_unit, &
                FMT="(/,T3,A,T12,A,T31,A,T39,A,T59,A,T75,A,/,T3,A)") &
            "Step", "Update method", "Time", "Convergence", "Total energy", "Change", &
            REPEAT("-", 78)
      END IF

      CALL timestop(handle)

   END SUBROUTINE ot_readwrite_input
! **************************************************************************************************

END MODULE qs_ot_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief manage control variables for the maximum overlap method
! **************************************************************************************************
MODULE qs_mom_types
   USE bibliography,                    ONLY: Gilbert2008
   USE input_constants,                 ONLY: momproj_norm,&
                                              momproj_sum
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_create,&
                                              section_type
   USE input_val_types,                 ONLY: integer_t
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_mom_types'

   PUBLIC :: create_mom_section

CONTAINS

! **************************************************************************************************
!> \brief Create CP2K input section for variable occupancy using the Maximum
!>        Overlap Method. Only with diagonalization methods, i.e. not with OT
!> \param section section to create
!> \date       20.06.2013
!> \par History
!>      06.2013 created [MattW]
!>      01.2016 (DE)OCC_ALPHA and (DE)OCC_BETA keywords accept a list of
!>              molecular orbitals. Added two extra keywords: START_ITER
!>              and PROJ_FORMULA [Sergey Chulkov]
!> \author     MattW
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE create_mom_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_mom_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("qs_mom_types.F",54)

      CALL section_create(section, &
                          name="MOM", &
                          description="Define type and parameters for the maximum overlap method (MOM) "// &
                          "to determine orbital occupancies. "// &
                          "The MOM procedures activated by this section are only active for diagonalization "// &
                          "methods, i.e. not with minimization methods based on OT. "// &
                          "Incompatible with core-level excitation spectra (XAS).", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE., &
                          citations=(/Gilbert2008/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, &
                          name="_SECTION_PARAMETERS_", &
                          description="Controls the activation of the MOM procedure", &
                          usage="MOM ON", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="START_ITER", &
                          description="SCF iteration cycle to start the MOM procedure. "// &
                          "Could be used for ground state calculations only "// &
                          "in order to stabilise oscillating SCF iterations", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=integer_t, &
                          default_i_val=0, &
                          usage="START_ITER 2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="DEOCC_ALPHA", &
                          description="Alpha orbitals to be deoccupied", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          type_of_var=integer_t, &
                          default_i_val=0, &
                          usage="DEOCC_ALPHA 10 11 ...")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="DEOCC_BETA", &
                          description="Beta orbitals to be deoccupied", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          type_of_var=integer_t, &
                          default_i_val=0, &
                          usage="DEOCC_BETA 10 11 ...")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="OCC_ALPHA", &
                          description="Alpha orbitals to be occupied", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          type_of_var=integer_t, &
                          default_i_val=0, &
                          usage="OCC_ALPHA 12 15 ...")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="OCC_BETA", &
                          description="Beta orbitals to be occupied", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          type_of_var=integer_t, &
                          default_i_val=0, &
                          usage="OCC_BETA 12 15 ...")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PROJ_FORMULA", &
                          description="Projection formula to be used", &
                          usage="SCF_GUESS RESTART", default_i_val=momproj_norm, &
                          enum_c_vals=s2a("NORM", "SUM"), &
                          enum_desc=s2a( &
                          "The one which ignores the phase of molecular orbitals: proj_j = \sqrt(\sum_i overlap(i, j)^2)", &
                          "The one proposed in the original paper: proj_j = |\sum_i overlap(i, j)|"), &
                          enum_i_vals=(/momproj_norm, momproj_sum/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_mom_section

END MODULE qs_mom_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
MODULE qs_rho_atom_types

   USE kinds,                           ONLY: dp,&
                                              int_size
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_rho_atom_types'

   TYPE rho_atom_coeff
      REAL(dp), DIMENSION(:, :), POINTER :: r_coef
   END TYPE rho_atom_coeff

   TYPE rho_atom_type
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER  :: cpc_h, cpc_s
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER  :: rho_rad_h, &
                     rho_rad_s, &
                     vrho_rad_h, &
                     vrho_rad_s
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER  :: drho_rad_h, &
                     drho_rad_s
      TYPE(rho_atom_coeff), DIMENSION(:, :), &
         POINTER  :: rho_rad_h_d, &
                     rho_rad_s_d
      TYPE(rho_atom_coeff), DIMENSION(:, :), &
         POINTER  :: trho_rad_h, &
                     trho_rad_s

      INTEGER                             :: rhoa_of_atom
      REAL(dp)                            :: exc_h, &
                                             exc_s
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER   :: ga_Vlocal_gb_h, &
                      ga_Vlocal_gb_s
   END TYPE rho_atom_type

   TYPE rho_atom_p_type
      TYPE(rho_atom_type), POINTER       :: rho_atom
   END TYPE rho_atom_p_type

   PUBLIC :: allocate_rho_atom_set, &
             deallocate_rho_atom_set, &
             get_rho_atom

   PUBLIC :: rho_atom_coeff, rho_atom_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param rho_atom_set ...
!> \param natom ...
! **************************************************************************************************
   SUBROUTINE allocate_rho_atom_set(rho_atom_set, natom)

      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_rho_atom_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iat, istat

      IF (ASSOCIATED(rho_atom_set)) THEN
         CALL deallocate_rho_atom_set(rho_atom_set)
      END IF

      ALLOCATE (rho_atom_set(natom), STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 81, &
                                       "allocate_rho_atom_set", natom*int_size)

      DO iat = 1, natom

         NULLIFY (rho_atom_set(iat)%cpc_h)
         NULLIFY (rho_atom_set(iat)%cpc_s)
         NULLIFY (rho_atom_set(iat)%rho_rad_h)
         NULLIFY (rho_atom_set(iat)%rho_rad_s)
         NULLIFY (rho_atom_set(iat)%drho_rad_h)
         NULLIFY (rho_atom_set(iat)%drho_rad_s)
         NULLIFY (rho_atom_set(iat)%rho_rad_h_d)
         NULLIFY (rho_atom_set(iat)%rho_rad_s_d)
         NULLIFY (rho_atom_set(iat)%vrho_rad_h)
         NULLIFY (rho_atom_set(iat)%vrho_rad_s)
         NULLIFY (rho_atom_set(iat)%trho_rad_h)
         NULLIFY (rho_atom_set(iat)%trho_rad_s)
         NULLIFY (rho_atom_set(iat)%ga_Vlocal_gb_h)
         NULLIFY (rho_atom_set(iat)%ga_Vlocal_gb_s)

      ENDDO

   END SUBROUTINE allocate_rho_atom_set

! **************************************************************************************************
!> \brief ...
!> \param rho_atom_set ...
! **************************************************************************************************
   SUBROUTINE deallocate_rho_atom_set(rho_atom_set)

      TYPE(rho_atom_type), DIMENSION(:), POINTER         :: rho_atom_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_rho_atom_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iat, j, n, natom

      IF (ASSOCIATED(rho_atom_set)) THEN

         natom = SIZE(rho_atom_set)

         DO iat = 1, natom
            IF (ASSOCIATED(rho_atom_set(iat)%cpc_h)) THEN
               IF (ASSOCIATED(rho_atom_set(iat)%cpc_h(1)%r_coef)) THEN
                  n = SIZE(rho_atom_set(iat)%cpc_h, 1)
                  DO i = 1, n
                     DEALLOCATE (rho_atom_set(iat)%cpc_h(i)%r_coef)
                     DEALLOCATE (rho_atom_set(iat)%cpc_s(i)%r_coef)
                  END DO
               END IF
               DEALLOCATE (rho_atom_set(iat)%cpc_h)
               DEALLOCATE (rho_atom_set(iat)%cpc_s)
            END IF
            IF (ASSOCIATED(rho_atom_set(iat)%ga_Vlocal_gb_h)) THEN
               IF (ASSOCIATED(rho_atom_set(iat)%ga_Vlocal_gb_h(1)%r_coef)) THEN
                  n = SIZE(rho_atom_set(iat)%ga_Vlocal_gb_h, 1)
                  DO i = 1, n
                     DEALLOCATE (rho_atom_set(iat)%ga_Vlocal_gb_h(i)%r_coef)
                     DEALLOCATE (rho_atom_set(iat)%ga_Vlocal_gb_s(i)%r_coef)
                  END DO
               END IF
               DEALLOCATE (rho_atom_set(iat)%ga_Vlocal_gb_h)
               DEALLOCATE (rho_atom_set(iat)%ga_Vlocal_gb_s)
            END IF

            IF (ASSOCIATED(rho_atom_set(iat)%drho_rad_h)) THEN
               IF (ASSOCIATED(rho_atom_set(iat)%drho_rad_h(1)%r_coef)) THEN
                  n = SIZE(rho_atom_set(iat)%drho_rad_h, 1)
                  DO i = 1, n
                     DEALLOCATE (rho_atom_set(iat)%drho_rad_h(i)%r_coef)
                     DEALLOCATE (rho_atom_set(iat)%drho_rad_s(i)%r_coef)
                     DO j = 1, 3
                        DEALLOCATE (rho_atom_set(iat)%rho_rad_h_d(j, i)%r_coef)
                        DEALLOCATE (rho_atom_set(iat)%rho_rad_s_d(j, i)%r_coef)
                     END DO
                  ENDDO
               END IF
               DEALLOCATE (rho_atom_set(iat)%drho_rad_h)
               DEALLOCATE (rho_atom_set(iat)%drho_rad_s)
               DEALLOCATE (rho_atom_set(iat)%rho_rad_h_d)
               DEALLOCATE (rho_atom_set(iat)%rho_rad_s_d)
            ENDIF

            IF (ASSOCIATED(rho_atom_set(iat)%rho_rad_h)) THEN
               IF (ASSOCIATED(rho_atom_set(iat)%rho_rad_h(1)%r_coef)) THEN
                  n = SIZE(rho_atom_set(iat)%rho_rad_h)
                  DO i = 1, n
                     DEALLOCATE (rho_atom_set(iat)%rho_rad_h(i)%r_coef)
                  ENDDO
               ENDIF
               DEALLOCATE (rho_atom_set(iat)%rho_rad_h)
            ENDIF

            IF (ASSOCIATED(rho_atom_set(iat)%rho_rad_s)) THEN
               IF (ASSOCIATED(rho_atom_set(iat)%rho_rad_s(1)%r_coef)) THEN
                  n = SIZE(rho_atom_set(iat)%rho_rad_s)
                  DO i = 1, n
                     DEALLOCATE (rho_atom_set(iat)%rho_rad_s(i)%r_coef)
                  ENDDO
               ENDIF
               DEALLOCATE (rho_atom_set(iat)%rho_rad_s)
            ENDIF

            IF (ASSOCIATED(rho_atom_set(iat)%vrho_rad_h)) THEN
               IF (ASSOCIATED(rho_atom_set(iat)%vrho_rad_h(1)%r_coef)) THEN
                  n = SIZE(rho_atom_set(iat)%vrho_rad_h)
                  DO i = 1, n
                     DEALLOCATE (rho_atom_set(iat)%vrho_rad_h(i)%r_coef)
                  ENDDO
               END IF
               DEALLOCATE (rho_atom_set(iat)%vrho_rad_h)
            ENDIF

            IF (ASSOCIATED(rho_atom_set(iat)%vrho_rad_s)) THEN
               IF (ASSOCIATED(rho_atom_set(iat)%vrho_rad_s(1)%r_coef)) THEN
                  n = SIZE(rho_atom_set(iat)%vrho_rad_s)
                  DO i = 1, n
                     DEALLOCATE (rho_atom_set(iat)%vrho_rad_s(i)%r_coef)
                  ENDDO
               END IF
               DEALLOCATE (rho_atom_set(iat)%vrho_rad_s)
            ENDIF

            IF (ASSOCIATED(rho_atom_set(iat)%trho_rad_h)) THEN
               IF (ASSOCIATED(rho_atom_set(iat)%trho_rad_h(1, 1)%r_coef)) THEN
                  n = SIZE(rho_atom_set(iat)%trho_rad_h, 2)
                  DO i = 1, n
                     DEALLOCATE (rho_atom_set(iat)%trho_rad_h(1, i)%r_coef)
                     DEALLOCATE (rho_atom_set(iat)%trho_rad_h(2, i)%r_coef)
                     DEALLOCATE (rho_atom_set(iat)%trho_rad_h(3, i)%r_coef)
                  ENDDO
               ENDIF
               DEALLOCATE (rho_atom_set(iat)%trho_rad_h)
            ENDIF

            IF (ASSOCIATED(rho_atom_set(iat)%trho_rad_s)) THEN
               IF (ASSOCIATED(rho_atom_set(iat)%trho_rad_s(1, 1)%r_coef)) THEN
                  n = SIZE(rho_atom_set(iat)%trho_rad_s, 2)
                  DO i = 1, n
                     DEALLOCATE (rho_atom_set(iat)%trho_rad_s(1, i)%r_coef)
                     DEALLOCATE (rho_atom_set(iat)%trho_rad_s(2, i)%r_coef)
                     DEALLOCATE (rho_atom_set(iat)%trho_rad_s(3, i)%r_coef)
                  ENDDO
               ENDIF
               DEALLOCATE (rho_atom_set(iat)%trho_rad_s)
            ENDIF

         ENDDO

         DEALLOCATE (rho_atom_set)

      ELSE

         CALL cp_abort(cp__l("qs_rho_atom_types.F",234), &
                       "The pointer rho_atom_set is not associated and "// &
                       "cannot be deallocated")

      END IF

   END SUBROUTINE deallocate_rho_atom_set

! **************************************************************************************************
!> \brief ...
!> \param rho_atom ...
!> \param cpc_h ...
!> \param cpc_s ...
!> \param rho_rad_h ...
!> \param rho_rad_s ...
!> \param drho_rad_h ...
!> \param drho_rad_s ...
!> \param vrho_rad_h ...
!> \param vrho_rad_s ...
!> \param rho_rad_h_d ...
!> \param rho_rad_s_d ...
!> \param trho_rad_h ...
!> \param trho_rad_s ...
!> \param ga_Vlocal_gb_h ...
!> \param ga_Vlocal_gb_s ...
! **************************************************************************************************
   SUBROUTINE get_rho_atom(rho_atom, cpc_h, cpc_s, rho_rad_h, rho_rad_s, &
                           drho_rad_h, drho_rad_s, vrho_rad_h, vrho_rad_s, &
                           rho_rad_h_d, rho_rad_s_d, trho_rad_h, trho_rad_s, &
                           ga_Vlocal_gb_h, ga_Vlocal_gb_s)

      TYPE(rho_atom_type), INTENT(IN), POINTER           :: rho_atom
      TYPE(rho_atom_coeff), DIMENSION(:), INTENT(OUT), &
         OPTIONAL, POINTER                               :: cpc_h, cpc_s, rho_rad_h, rho_rad_s, &
                                                            drho_rad_h, drho_rad_s, vrho_rad_h, &
                                                            vrho_rad_s
      TYPE(rho_atom_coeff), DIMENSION(:, :), &
         INTENT(OUT), OPTIONAL, POINTER                  :: rho_rad_h_d, rho_rad_s_d, trho_rad_h, &
                                                            trho_rad_s
      TYPE(rho_atom_coeff), DIMENSION(:), INTENT(OUT), &
         OPTIONAL, POINTER                               :: ga_Vlocal_gb_h, ga_Vlocal_gb_s

      CHARACTER(len=*), PARAMETER :: routineN = 'get_rho_atom', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(rho_atom)) THEN
         IF (PRESENT(cpc_h)) cpc_h => rho_atom%cpc_h
         IF (PRESENT(cpc_s)) cpc_s => rho_atom%cpc_s
         IF (PRESENT(rho_rad_h)) rho_rad_h => rho_atom%rho_rad_h
         IF (PRESENT(rho_rad_s)) rho_rad_s => rho_atom%rho_rad_s
         IF (PRESENT(drho_rad_h)) drho_rad_h => rho_atom%drho_rad_h
         IF (PRESENT(drho_rad_s)) drho_rad_s => rho_atom%drho_rad_s
         IF (PRESENT(rho_rad_h_d)) rho_rad_h_d => rho_atom%rho_rad_h_d
         IF (PRESENT(rho_rad_s_d)) rho_rad_s_d => rho_atom%rho_rad_s_d
         IF (PRESENT(vrho_rad_h)) vrho_rad_h => rho_atom%vrho_rad_h
         IF (PRESENT(vrho_rad_s)) vrho_rad_s => rho_atom%vrho_rad_s
         IF (PRESENT(trho_rad_h)) trho_rad_h => rho_atom%trho_rad_h
         IF (PRESENT(trho_rad_s)) trho_rad_s => rho_atom%trho_rad_s
         IF (PRESENT(ga_Vlocal_gb_h)) ga_Vlocal_gb_h => rho_atom%ga_Vlocal_gb_h
         IF (PRESENT(ga_Vlocal_gb_s)) ga_Vlocal_gb_s => rho_atom%ga_Vlocal_gb_s
      ELSE
         CALL cp__b("qs_rho_atom_types.F",294,"The pointer rho_atom is not associated")
      END IF

   END SUBROUTINE get_rho_atom

END MODULE qs_rho_atom_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module that contains the definitions of the scf types
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE qs_density_mixing_types
   USE ao_util,                         ONLY: exp_radius
   USE input_constants,                 ONLY: broy_mix,&
                                              broy_mix_new,&
                                              direct_p_mix,&
                                              gaussian,&
                                              kerker_mix,&
                                              multisec_mix,&
                                              no_mix,&
                                              pulay_mix
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_create,&
                                              section_type,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE input_val_types,                 ONLY: real_t
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE qs_rho_atom_types,               ONLY: rho_atom_coeff
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_density_mixing_types'

   INTEGER, PARAMETER, PUBLIC :: no_mixing_nr = 0, direct_mixing_nr = 1, &
                                 gspace_mixing_nr = 2, pulay_mixing_nr = 3, &
                                 broyden_mixing_nr = 4, &
                                 broyden_mixing_new_nr = 5, &
                                 multisecant_mixing_nr = 6
   PUBLIC :: cp_1d_z_p_type, mixing_storage_create, mixing_storage_type, mixing_storage_release, create_mixing_section

   TYPE cp_1d_z_p_type
      COMPLEX(dp), DIMENSION(:), POINTER :: cc
   END TYPE cp_1d_z_p_type

   TYPE mixing_storage_type
      INTEGER :: ref_count, ig_max, ncall, nbuffer, n_simple_mix, nskip_mixing, p_metric_method
      INTEGER, POINTER, DIMENSION(:) :: ig_global_index
      LOGICAL, POINTER, DIMENSION(:) :: paw
      CHARACTER(len=15) :: iter_method
      REAL(dp) :: alpha, bconst, beta, broy_w0, max_g2, max_gvec_exp, pulay_alpha, pulay_beta, r_step, &
                  reg_par, sigma_max, wc, wmax
      REAL(dp), DIMENSION(:), POINTER :: p_metric
      REAL(KIND=dp), DIMENSION(:), POINTER :: kerker_factor
      REAL(KIND=dp), DIMENSION(:), POINTER :: special_metric
      REAL(dp), DIMENSION(:, :), POINTER :: weight
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: norm_res_buffer, pulay_matrix
      REAL(dp), DIMENSION(:, :, :), POINTER :: fmat, gmat, smat
      TYPE(cp_1d_z_p_type), DIMENSION(:), POINTER :: last_res, rhoin, rhoin_old
      TYPE(cp_1d_z_p_type), DIMENSION(:, :), POINTER :: delta_res, u_vec, z_vec
      TYPE(cp_1d_z_p_type), DIMENSION(:, :), POINTER ::drho_buffer, rhoin_buffer, &
                                                        res_buffer
      TYPE(rho_atom_coeff), DIMENSION(:, :), POINTER :: cpc_h_lastres, cpc_s_lastres
      TYPE(rho_atom_coeff), DIMENSION(:, :), POINTER :: cpc_h_in, cpc_s_in
      TYPE(rho_atom_coeff), DIMENSION(:, :), POINTER :: cpc_h_old, cpc_s_old
      TYPE(rho_atom_coeff), DIMENSION(:, :, :), POINTER :: dcpc_h_in, dcpc_s_in
   END TYPE mixing_storage_type

CONTAINS

! **************************************************************************************************
!> \brief creates a mixing_storage
!> \param mixing_store ...
!> \param mixing_section ...
!> \param mixing_method ...
!> \param ecut ...
!> \par History
!>      05.2009 created [MI]
!> \author [MI]
! **************************************************************************************************
   SUBROUTINE mixing_storage_create(mixing_store, mixing_section, mixing_method, ecut)
      TYPE(mixing_storage_type), POINTER                 :: mixing_store
      TYPE(section_vals_type), POINTER                   :: mixing_section
      INTEGER, INTENT(IN)                                :: mixing_method
      REAL(dp), INTENT(IN)                               :: ecut

      CHARACTER(len=*), PARAMETER :: routineN = 'mixing_storage_create', &
         routineP = moduleN//':'//routineN

      REAL(dp)                                           :: alpha, eps, gcut

      IF(.NOT.(.NOT. ASSOCIATED(mixing_store)))CALL cp__a("qs_density_mixing_types.F",101)
      ALLOCATE (mixing_store)

      mixing_store%ref_count = 1
      mixing_store%nbuffer = 0
      mixing_store%n_simple_mix = 0
      mixing_store%ncall = 0
      mixing_store%alpha = 1.0_dp
      mixing_store%pulay_beta = 1.0_dp
      mixing_store%beta = 1.0_dp
      mixing_store%iter_method = "NoMix"
      mixing_store%max_g2 = 2._dp*ecut

      NULLIFY (mixing_store%p_metric)
      NULLIFY (mixing_store%kerker_factor)
      NULLIFY (mixing_store%special_metric)
      NULLIFY (mixing_store%pulay_matrix)
      NULLIFY (mixing_store%weight)
      NULLIFY (mixing_store%fmat)
      NULLIFY (mixing_store%gmat)
      NULLIFY (mixing_store%smat)
      NULLIFY (mixing_store%last_res)
      NULLIFY (mixing_store%rhoin)
      NULLIFY (mixing_store%rhoin_old)
      NULLIFY (mixing_store%delta_res)
      NULLIFY (mixing_store%u_vec)
      NULLIFY (mixing_store%z_vec)
      NULLIFY (mixing_store%drho_buffer)
      NULLIFY (mixing_store%rhoin_buffer)
      NULLIFY (mixing_store%res_buffer)
      NULLIFY (mixing_store%norm_res_buffer)
      NULLIFY (mixing_store%ig_global_index)
      NULLIFY (mixing_store%paw)
      NULLIFY (mixing_store%cpc_h_in)
      NULLIFY (mixing_store%cpc_s_in)
      NULLIFY (mixing_store%cpc_h_old)
      NULLIFY (mixing_store%cpc_s_old)
      NULLIFY (mixing_store%dcpc_h_in)
      NULLIFY (mixing_store%dcpc_s_in)
      NULLIFY (mixing_store%cpc_h_lastres)
      NULLIFY (mixing_store%cpc_s_lastres)

      CALL section_vals_val_get(mixing_section, "ALPHA", r_val=mixing_store%alpha)
      CALL section_vals_val_get(mixing_section, "BETA", r_val=mixing_store%beta)
      CALL section_vals_val_get(mixing_section, "N_SIMPLE_MIX", i_val=mixing_store%n_simple_mix)
      CALL section_vals_val_get(mixing_section, "NBUFFER", i_val=mixing_store%nbuffer)
      CALL section_vals_val_get(mixing_section, "NSKIP", i_val=mixing_store%nskip_mixing)
      CALL section_vals_val_get(mixing_section, "MAX_GVEC_EXP", r_val=mixing_store%max_gvec_exp)

      IF (mixing_store%max_gvec_exp > 0._dp) THEN
         alpha = 0.25_dp/mixing_store%max_gvec_exp
         eps = 1.e-4_dp
         gcut = exp_radius(3, alpha, eps, 1.0_dp)
         mixing_store%max_g2 = gcut*gcut
      END IF

      SELECT CASE (mixing_method)
      CASE (gspace_mixing_nr)
         mixing_store%nbuffer = 1
      CASE (pulay_mixing_nr)
         CALL section_vals_val_get(mixing_section, "PULAY_ALPHA", r_val=mixing_store%pulay_alpha)
         CALL section_vals_val_get(mixing_section, "PULAY_BETA", r_val=mixing_store%pulay_beta)
      CASE (broyden_mixing_nr)
         CALL section_vals_val_get(mixing_section, "BROY_W0", r_val=mixing_store%broy_w0)
         mixing_store%bconst = 20.0_dp
      CASE (broyden_mixing_new_nr)
         CALL section_vals_val_get(mixing_section, "BROY_WREF", r_val=mixing_store%wc)
         CALL section_vals_val_get(mixing_section, "BROY_WMAX", r_val=mixing_store%wmax)
         mixing_store%bconst = 20.0_dp
         mixing_store%p_metric_method = 1
      CASE (multisecant_mixing_nr)
         CALL section_vals_val_get(mixing_section, "REGULARIZATION", r_val=mixing_store%reg_par)
         CALL section_vals_val_get(mixing_section, "MAX_STEP", r_val=mixing_store%sigma_max)
         CALL section_vals_val_get(mixing_section, "R_FACTOR", r_val=mixing_store%r_step)
      END SELECT

   END SUBROUTINE mixing_storage_create

! **************************************************************************************************
!> \brief releases a mixing_storage
!> \param mixing_store ...
!> \par History
!>      05.2009 created [MI]
!> \author [MI]
! **************************************************************************************************
   SUBROUTINE mixing_storage_release(mixing_store)
      TYPE(mixing_storage_type), POINTER                 :: mixing_store

      CHARACTER(len=*), PARAMETER :: routineN = 'mixing_storage_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k

      IF (ASSOCIATED(mixing_store)) THEN
         IF(.NOT.(mixing_store%ref_count > 0))CALL cp__a("qs_density_mixing_types.F",195)
         mixing_store%ref_count = mixing_store%ref_count-1
         IF (mixing_store%ref_count == 0) THEN

            IF (ASSOCIATED(mixing_store%kerker_factor)) THEN
               DEALLOCATE (mixing_store%kerker_factor)
            ENDIF

            IF (ASSOCIATED(mixing_store%special_metric)) THEN
               DEALLOCATE (mixing_store%special_metric)
            ENDIF

            IF (ASSOCIATED(mixing_store%pulay_matrix)) THEN
               DEALLOCATE (mixing_store%pulay_matrix)
            ENDIF

            IF (ASSOCIATED(mixing_store%rhoin_buffer)) THEN
               DO i = 1, SIZE(mixing_store%rhoin_buffer, 2)
                  DO j = 1, SIZE(mixing_store%rhoin_buffer, 1)
                     DEALLOCATE (mixing_store%rhoin_buffer(j, i)%cc)
                  END DO
               END DO
               DEALLOCATE (mixing_store%rhoin_buffer)
            ENDIF

            IF (ASSOCIATED(mixing_store%paw)) THEN
               DEALLOCATE (mixing_store%paw)
            END IF
            IF (ASSOCIATED(mixing_store%cpc_h_in)) THEN
               DO j = 1, SIZE(mixing_store%cpc_h_in, 2)
               DO k = 1, SIZE(mixing_store%cpc_h_in, 1)
                  IF (ASSOCIATED(mixing_store%cpc_h_in(k, j)%r_coef)) THEN
                     DEALLOCATE (mixing_store%cpc_h_in(k, j)%r_coef)
                     DEALLOCATE (mixing_store%cpc_s_in(k, j)%r_coef)
                  END IF
               END DO
               END DO
               DEALLOCATE (mixing_store%cpc_h_in)
               DEALLOCATE (mixing_store%cpc_s_in)
            END IF
            IF (ASSOCIATED(mixing_store%cpc_h_old)) THEN
               DO j = 1, SIZE(mixing_store%cpc_h_old, 2)
               DO k = 1, SIZE(mixing_store%cpc_h_old, 1)
                  IF (ASSOCIATED(mixing_store%cpc_h_old(k, j)%r_coef)) THEN
                     DEALLOCATE (mixing_store%cpc_h_old(k, j)%r_coef)
                     DEALLOCATE (mixing_store%cpc_s_old(k, j)%r_coef)
                  END IF
               END DO
               END DO
               DEALLOCATE (mixing_store%cpc_h_old)
               DEALLOCATE (mixing_store%cpc_s_old)
            END IF
            IF (ASSOCIATED(mixing_store%dcpc_h_in)) THEN
               DO i = 1, SIZE(mixing_store%dcpc_h_in, 3)
               DO j = 1, SIZE(mixing_store%dcpc_h_in, 2)
               DO k = 1, SIZE(mixing_store%dcpc_h_in, 1)
                  IF (ASSOCIATED(mixing_store%dcpc_h_in(k, j, i)%r_coef)) THEN
                     DEALLOCATE (mixing_store%dcpc_h_in(k, j, i)%r_coef)
                     DEALLOCATE (mixing_store%dcpc_s_in(k, j, i)%r_coef)
                  END IF
               END DO
               END DO
               END DO
               DEALLOCATE (mixing_store%dcpc_h_in)
               DEALLOCATE (mixing_store%dcpc_s_in)
            END IF
            IF (ASSOCIATED(mixing_store%cpc_h_lastres)) THEN
               DO j = 1, SIZE(mixing_store%cpc_h_lastres, 2)
               DO k = 1, SIZE(mixing_store%cpc_h_lastres, 1)
                  IF (ASSOCIATED(mixing_store%cpc_h_lastres(k, j)%r_coef)) THEN
                     DEALLOCATE (mixing_store%cpc_h_lastres(k, j)%r_coef)
                     DEALLOCATE (mixing_store%cpc_s_lastres(k, j)%r_coef)
                  END IF
               END DO
               END DO
               DEALLOCATE (mixing_store%cpc_h_lastres)
               DEALLOCATE (mixing_store%cpc_s_lastres)
            ENDIF

            IF (ASSOCIATED(mixing_store%res_buffer)) THEN
               DO i = 1, SIZE(mixing_store%res_buffer, 2)
                  DO j = 1, SIZE(mixing_store%res_buffer, 1)
                     DEALLOCATE (mixing_store%res_buffer(j, i)%cc)
                  END DO
               END DO
               DEALLOCATE (mixing_store%res_buffer)
            ENDIF

            IF (ASSOCIATED(mixing_store%norm_res_buffer)) THEN
               DEALLOCATE (mixing_store%norm_res_buffer)
            ENDIF

            IF (ASSOCIATED(mixing_store%ig_global_index)) THEN
               DEALLOCATE (mixing_store%ig_global_index)
            ENDIF

            IF (ASSOCIATED(mixing_store%drho_buffer)) THEN
               DO i = 1, SIZE(mixing_store%drho_buffer, 2)
                  DO j = 1, SIZE(mixing_store%drho_buffer, 1)
                     DEALLOCATE (mixing_store%drho_buffer(j, i)%cc)
                  END DO
               END DO
               DEALLOCATE (mixing_store%drho_buffer)
            ENDIF

            IF (ASSOCIATED(mixing_store%last_res)) THEN
               DO i = 1, SIZE(mixing_store%last_res)
                  DEALLOCATE (mixing_store%last_res(i)%cc)
               END DO
               DEALLOCATE (mixing_store%last_res)
            END IF

            IF (ASSOCIATED(mixing_store%rhoin)) THEN
               DO i = 1, SIZE(mixing_store%rhoin)
                  DEALLOCATE (mixing_store%rhoin(i)%cc)
               END DO
               DEALLOCATE (mixing_store%rhoin)
            END IF

            IF (ASSOCIATED(mixing_store%rhoin_old)) THEN
               DO i = 1, SIZE(mixing_store%rhoin_old)
                  DEALLOCATE (mixing_store%rhoin_old(i)%cc)
               END DO
               DEALLOCATE (mixing_store%rhoin_old)
            END IF

            IF (ASSOCIATED(mixing_store%p_metric)) THEN
               DEALLOCATE (mixing_store%p_metric)
            END IF

            IF (ASSOCIATED(mixing_store%weight)) THEN
               DEALLOCATE (mixing_store%weight)
            END IF

            IF (ASSOCIATED(mixing_store%fmat)) THEN
               DEALLOCATE (mixing_store%fmat)
            END IF

            IF (ASSOCIATED(mixing_store%delta_res)) THEN
               DO i = 1, SIZE(mixing_store%delta_res, 2)
                  DO j = 1, SIZE(mixing_store%delta_res, 1)
                     DEALLOCATE (mixing_store%delta_res(j, i)%cc)
                  END DO
               END DO
               DEALLOCATE (mixing_store%delta_res)
            END IF

            IF (ASSOCIATED(mixing_store%u_vec)) THEN
               DO i = 1, SIZE(mixing_store%u_vec, 2)
                  DO j = 1, SIZE(mixing_store%u_vec, 1)
                     DEALLOCATE (mixing_store%u_vec(j, i)%cc)
                  END DO
               END DO
               DEALLOCATE (mixing_store%u_vec)
            END IF

            IF (ASSOCIATED(mixing_store%z_vec)) THEN
               DO i = 1, SIZE(mixing_store%z_vec, 2)
                  DO j = 1, SIZE(mixing_store%z_vec, 1)
                     DEALLOCATE (mixing_store%z_vec(j, i)%cc)
                  END DO
               END DO
               DEALLOCATE (mixing_store%z_vec)
            END IF

            DEALLOCATE (mixing_store)
         END IF
         NULLIFY (mixing_store)
      END IF

   END SUBROUTINE mixing_storage_release

! **************************************************************************************************
!> \brief      Create CP2K input section for the mixing of the density matrix to
!>             be used only with diagonalization methods, i.e. not with OT
!> \param section ...
!> \param ls_scf ...
!> \date       20.02.2009
!> \par History
!>      02.2015 moved here from input_cp2k_dft.F, modified for use in LS SCF
!>              [Patrick Seewald]
!> \author     MI
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE create_mixing_section(section, ls_scf)

      TYPE(section_type), POINTER                        :: section
      LOGICAL, INTENT(IN), OPTIONAL                      :: ls_scf

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_mixing_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: section_name
      INTEGER                                            :: default_mix
      LOGICAL                                            :: ls
      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("qs_density_mixing_types.F",392)

      IF (PRESENT(ls_scf)) THEN
         IF (ls_scf) THEN
            ls = .TRUE.
         ELSE
            ls = .FALSE.
         ENDIF
      ELSE
         ls = .FALSE.
      ENDIF

      IF (ls) THEN
         section_name = "RHO_MIXING"
      ELSE
         section_name = "MIXING"
      ENDIF

      CALL section_create(section, &
                          name=section_name, &
                          description="Define type and parameters for mixing "// &
                          "procedures to be applied to the density matrix. Normally, "// &
                          "only one type of mixing method should be accepted. The mixing "// &
                          "procedures activated by this section are only active for diagonalization "// &
                          "methods and linear scaling SCF, i.e. not with minimization methods based "// &
                          "on OT.", &
                          n_keywords=16, &
                          n_subsections=0, &
                          repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, &
                          name="_SECTION_PARAMETERS_", &
                          description="Controls the activation of the mixing procedure", &
                          usage="&MIXING ON", &
                          default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      IF (.NOT. ls) THEN
         default_mix = direct_p_mix
      ELSE
         default_mix = broy_mix
      ENDIF

      CALL keyword_create(keyword, &
                          name="METHOD", &
                          description="Mixing method to be applied", &
                          repeats=.FALSE., &
                          usage="METHOD KERKER_MIXING", &
                          default_i_val=default_mix, &
                          enum_c_vals=s2a("NONE", &
                                          "DIRECT_P_MIXING", &
                                          "KERKER_MIXING", &
                                          "PULAY_MIXING", &
                                          "BROYDEN_MIXING", &
                                          "BROYDEN_MIXING_NEW", &
                                          "MULTISECANT_MIXING"), &
                          enum_i_vals=(/no_mix, direct_p_mix, kerker_mix, pulay_mix, broy_mix, &
                                        broy_mix_new, multisec_mix/), &
                          enum_desc=s2a("No mixing is applied", &
                                        "Direct mixing of new and old density matrices", &
                                        "Mixing of the potential in reciprocal space using the Kerker damping", &
                                        "Pulay mixing", "Broyden mixing", "Broyden mixing second version", &
                                        "Multisecant scheme for mixing"))

      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="ALPHA", &
                          description="Fraction of new density to be included", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.4_dp, &
                          usage="ALPHA 0.2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="BETA", &
                          description="Denominator parameter in Kerker damping "// &
                          "introduced to suppress charge sloshing: rho_mix(g) ="// &
                          "rho_in(g) + alpha*g^2/(g^2 + beta^2)*(rho_out(g)-"// &
                          "rho_in(g))", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.5_dp, &
                          unit_str="bohr^-1", &
                          usage="BETA 1.5")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="PULAY_ALPHA", &
                          description="Fraction of new density to be added to the Pulay expansion", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.0_dp, &
                          usage="PULAY_ALPHA 0.2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="PULAY_BETA", &
                          description="Fraction of residual contribution to be added to Pulay expansion", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=1.0_dp, &
                          usage="PULAY_BETA 0.2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NMIXING", &
                          description="Minimal number of density mixing (should be greater than 0),"// &
                          "before starting DIIS", &
                          usage="NMIXING 1", default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NBUFFER", &
                          variants=s2a("NPULAY", "NBROYDEN", "NMULTISECANT"), &
                          description="Number of previous steps stored for the actual mixing scheme", &
                          usage="NBUFFER 2", default_i_val=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="BROY_W0", &
                          description=" w0 parameter used in Broyden mixing", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.01_dp, &
                          usage="BROY_W0 0.03")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="BROY_WREF", &
                          description="", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=100.0_dp, &
                          usage="BROY_WREF 0.2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="BROY_WMAX", &
                          description="", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=30.0_dp, &
                          usage="BROY_WMAX 10.0")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="REGULARIZATION", &
                          description="Regularization parameter to stabilize "// &
                          "the inversion of the residual matrix {Yn^t Yn} in the "// &
                          "multisecant mixing scheme (noise)", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.00001_dp, &
                          usage="REGULARIZATION 0.000001")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="MAX_STEP", &
                          description="Upper bound for the magnitude of the "// &
                          "unpredicted step size in the update by the "// &
                          "multisecant mixing scheme", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.1_dp, &
                          usage="MAX_STEP .2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="R_FACTOR", &
                          description="Control factor for the magnitude of the "// &
                          "unpredicted step size in the update by the "// &
                          "multisecant mixing scheme", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.05_dp, &
                          usage="R_FACTOR .12")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NSKIP", &
                          variants=(/"NSKIP_MIXING"/), &
                          description="Number of initial iteration for which the mixing is skipped", &
                          usage="NSKIP 10", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_SIMPLE_MIX", &
                          variants=(/"NSIMPLEMIX"/), &
                          description="Number of kerker damping iterations before starting other mixing procedures", &
                          usage="NSIMPLEMIX", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_GVEC_EXP", &
                          description="Restricts the G-space mixing to lower part of G-vector spectrum,"// &
                          " up to a G0, by assigning the exponent of the Gaussian that can be "// &
                          "represented by vectors smaller than G0 within a certain accuracy. ", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=-1._dp, &
                          usage="MAX_GVEC_EXP 3.")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_mixing_section
END MODULE qs_density_mixing_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief different utils that are useful to manipulate splines on the regular
!>      grid of a pw
!> \par History
!>      05.2003 created [fawzi]
!>      08.2004 removed spline evaluation method using more than 2 read streams
!>              (pw_compose_stripe_rs3), added linear solver based spline
!>              inversion [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE pw_spline_utils
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type,&
                                              cp_to_string
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: twopi
   USE message_passing,                 ONLY: mp_alltoall,&
                                              mp_comm_compare,&
                                              mp_sendrecv,&
                                              mp_sum
   USE pw_grid_types,                   ONLY: FULLSPACE,&
                                              PW_MODE_LOCAL
   USE pw_methods,                      ONLY: pw_axpy,&
                                              pw_copy,&
                                              pw_integral_ab,&
                                              pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_release,&
                                              pw_pool_retain,&
                                              pw_pool_type
   USE pw_types,                        ONLY: COMPLEXDATA1D,&
                                              REALDATA3D,&
                                              REALSPACE,&
                                              RECIPROCALSPACE,&
                                              pw_p_type,&
                                              pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pw_spline_utils'

   INTEGER, PARAMETER, PUBLIC               :: no_precond = 0, &
                                               precond_spl3_aint = 1, &
                                               precond_spl3_1 = 2, &
                                               precond_spl3_aint2 = 3, &
                                               precond_spl3_2 = 4, &
                                               precond_spl3_3 = 5

   REAL(KIND=dp), PUBLIC, PARAMETER, DIMENSION(4) :: nn10_coeffs = &
                                                     (/125._dp/216._dp, 25._dp/432._dp, 5._dp/864._dp, 1._dp/1728._dp/), &
                                                     spline3_coeffs = &
                                                     (/8._dp/(27._dp), 2._dp/(27._dp), 1._dp/(27._dp*2._dp), &
                                                       1._dp/(27._dp*8._dp)/), &
                                                     spline2_coeffs = &
                                                     (/27._dp/(64._dp), 9._dp/(64._dp*2_dp), 3._dp/(64._dp*4._dp), &
                                                       1._dp/(64._dp*8._dp)/), &
                                                     nn50_coeffs = &
                                                     (/15625._dp/17576._dp, 625._dp/35152._dp, 25._dp/70304._dp, &
                                                       1._dp/140608._dp/), &
                                                     spl3_aint_coeff = &
                                                     (/46._dp/27._dp, -2._dp/(27._dp), -1._dp/(27._dp*2._dp), &
                                                       -1._dp/(27._dp*8._dp)/), &
                                                     spl3_precond1_coeff = &
                                                     (/64._dp/3._dp, -8._dp/3._dp, -1._dp/3._dp, -1._dp/24._dp/), &
                                                     spl3_1d_transf_coeffs = &
                                                     (/2._dp/3._dp, 23._dp/48._dp, 1._dp/6._dp, 1._dp/48._dp/)

   REAL(KIND=dp), PUBLIC, PARAMETER, DIMENSION(3) :: spline3_deriv_coeffs = &
                                                     (/2.0_dp/9.0_dp, 1.0_dp/18.0_dp, 1.0_dp/72.0_dp/), &
                                                     spline2_deriv_coeffs = &
                                                     (/9.0_dp/32.0_dp, 3.0_dp/64.0_dp, 1.0_dp/128.0_dp/), &
                                                     nn10_deriv_coeffs = &
                                                     (/25._dp/72._dp, 5._dp/144, 1._dp/288._dp/), &
                                                     nn50_deriv_coeffs = &
                                                     (/625._dp/1352._dp, 25._dp/2704._dp, 1._dp/5408._dp/), &
                                                     spl3_1d_coeffs0 = &
                                                     (/1._dp/6_dp, 2._dp/3._dp, 1._dp/6._dp/), &
                                                     spl3_1d_transf_border1 = &
                                                     (/0.517977704_dp, 0.464044595_dp, 0.17977701e-1_dp/)

   INTEGER, SAVE, PRIVATE :: last_precond_id = 0

   PUBLIC :: pw_spline3_interpolate_values_g, &
             pw_spline3_deriv_g
   PUBLIC :: pw_spline_scale_deriv
   PUBLIC :: pw_spline2_interpolate_values_g, &
             pw_spline2_deriv_g
   PUBLIC :: pw_nn_smear_r, pw_nn_deriv_r, &
             spl3_nopbc, spl3_pbc, spl3_nopbct
   PUBLIC :: add_fine2coarse, add_coarse2fine
   PUBLIC :: pw_spline_precond_create, &
             pw_spline_do_precond, &
             pw_spline_precond_set_kind, &
             find_coeffs, &
             pw_spline_precond_release, &
             pw_spline_precond_type, &
             Eval_Interp_Spl3_pbc, &
             Eval_d_Interp_Spl3_pbc

!***

! **************************************************************************************************
!> \brief stores information for the preconditioner used to calculate the
!>      coeffs of splines
!> \author fawzi
! **************************************************************************************************
   TYPE pw_spline_precond_type
      INTEGER :: ref_count, id_nr, kind
      REAL(kind=dp), DIMENSION(4) :: coeffs
      REAL(kind=dp), DIMENSION(3) :: coeffs_1d
      LOGICAL :: sharpen, normalize, pbc, transpose
      TYPE(pw_pool_type), POINTER :: pool
   END TYPE pw_spline_precond_type

CONTAINS

! **************************************************************************************************
!> \brief calculates the FFT of the coefficents of the quadratic spline that
!>      interpolates the given values
!> \param spline_g on entry the FFT of the values to interpolate as cc,
!>        will contain the FFT of the coefficents of the spline
!> \par History
!>      06.2003 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      does not work with spherical cutoff
! **************************************************************************************************
   SUBROUTINE pw_spline2_interpolate_values_g(spline_g)
      TYPE(pw_type), POINTER                             :: spline_g

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_spline2_interpolate_values_g', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ii, j, k
      INTEGER, DIMENSION(2, 3)                           :: gbo
      INTEGER, DIMENSION(3)                              :: n_tot
      REAL(KIND=dp)                                      :: c23, coeff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cosIVals, cosJVals, cosKVals

      CALL timeset(routineN, handle)

      n_tot(1:3) = spline_g%pw_grid%npts(1:3)
      gbo = spline_g%pw_grid%bounds

      IF(.NOT.(ASSOCIATED(spline_g)))CALL cp__a("pw/pw_spline_utils.F",154)
      IF(.NOT.(spline_g%in_use == COMPLEXDATA1D))CALL cp__a("pw/pw_spline_utils.F",155)
      IF(.NOT.(spline_g%in_space == RECIPROCALSPACE))CALL cp__a("pw/pw_spline_utils.F",156)
      IF(.NOT.(.NOT. spline_g%pw_grid%spherical))CALL cp__a("pw/pw_spline_utils.F",157)
      IF(.NOT.(spline_g%pw_grid%grid_span == FULLSPACE))CALL cp__a("pw/pw_spline_utils.F",158)

      ALLOCATE (cosIVals(gbo(1, 1):gbo(2, 1)), cosJVals(gbo(1, 2):gbo(2, 2)), &
                cosKVals(gbo(1, 3):gbo(2, 3)))

      coeff = twopi/n_tot(1)
!$OMP     PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(cosIVals,coeff,gbo)
      DO i = gbo(1, 1), gbo(2, 1)
         cosIVals(i) = COS(coeff*REAL(i, dp))
      END DO
      coeff = twopi/n_tot(2)
!$OMP     PARALLEL DO PRIVATE(j) DEFAULT(NONE) SHARED(cosJVals,coeff,gbo)
      DO j = gbo(1, 2), gbo(2, 2)
         cosJVals(j) = COS(coeff*REAL(j, dp))
      END DO
      coeff = twopi/n_tot(3)
!$OMP     PARALLEL DO PRIVATE(k) DEFAULT(NONE) SHARED(cosKVals,coeff,gbo)
      DO k = gbo(1, 3), gbo(2, 3)
         cosKVals(k) = COS(coeff*REAL(k, dp))
      END DO

!$OMP     PARALLEL DO PRIVATE(i,j,k,ii,coeff,c23) DEFAULT(NONE) SHARED(spline_g,cosIVals,cosJVals,cosKVals)
      DO ii = 1, SIZE(spline_g%cc)
         i = spline_g%pw_grid%g_hat(1, ii)
         j = spline_g%pw_grid%g_hat(2, ii)
         k = spline_g%pw_grid%g_hat(3, ii)

         c23 = cosJVals(j)*cosKVals(k)
         coeff = 64.0_dp/(cosIVals(i)*c23+ &
                          (cosIVals(i)*cosJVals(j)+cosIVals(i)*cosKVals(k)+c23)*3.0_dp+ &
                          (cosIVals(i)+cosJVals(j)+cosKVals(k))*9.0_dp+ &
                          27.0_dp)

         spline_g%cc(ii) = spline_g%cc(ii)*coeff

      END DO
      DEALLOCATE (cosIVals, cosJVals, cosKVals)

      CALL timestop(handle)
   END SUBROUTINE pw_spline2_interpolate_values_g

! **************************************************************************************************
!> \brief calculates the FFT of the coefficents of the2 cubic spline that
!>      interpolates the given values
!> \param spline_g on entry the FFT of the values to interpolate as cc,
!>        will contain the FFT of the coefficents of the spline
!> \par History
!>      06.2003 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      does not work with spherical cutoff
!>      stupid distribution for cos calculation, it should calculate only the
!>      needed cos, and avoid the mp_sum
! **************************************************************************************************
   SUBROUTINE pw_spline3_interpolate_values_g(spline_g)
      TYPE(pw_type), POINTER                             :: spline_g

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_spline3_interpolate_values_g', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, ii, j, k
      INTEGER, DIMENSION(2, 3)                           :: gbo
      INTEGER, DIMENSION(3)                              :: n_tot
      REAL(KIND=dp)                                      :: c23, coeff
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: cosIVals, cosJVals, cosKVals

      CALL timeset(routineN, handle)

      n_tot(1:3) = spline_g%pw_grid%npts(1:3)
      gbo = spline_g%pw_grid%bounds

      IF(.NOT.(ASSOCIATED(spline_g)))CALL cp__a("pw/pw_spline_utils.F",229)
      IF(.NOT.(spline_g%in_use == COMPLEXDATA1D))CALL cp__a("pw/pw_spline_utils.F",230)
      IF(.NOT.(spline_g%in_space == RECIPROCALSPACE))CALL cp__a("pw/pw_spline_utils.F",231)
      IF(.NOT.(.NOT. spline_g%pw_grid%spherical))CALL cp__a("pw/pw_spline_utils.F",232)
      IF(.NOT.(spline_g%pw_grid%grid_span == FULLSPACE))CALL cp__a("pw/pw_spline_utils.F",233)

      ALLOCATE (cosIVals(gbo(1, 1):gbo(2, 1)), &
                cosJVals(gbo(1, 2):gbo(2, 2)), &
                cosKVals(gbo(1, 3):gbo(2, 3)))

      coeff = twopi/n_tot(1)
!$OMP     PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(cosIVals,coeff,gbo)
      DO i = gbo(1, 1), gbo(2, 1)
         cosIVals(i) = COS(coeff*REAL(i, dp))
      END DO
      coeff = twopi/n_tot(2)
!$OMP     PARALLEL DO PRIVATE(j) DEFAULT(NONE) SHARED(cosJVals,coeff,gbo)
      DO j = gbo(1, 2), gbo(2, 2)
         cosJVals(j) = COS(coeff*REAL(j, dp))
      END DO
      coeff = twopi/n_tot(3)
!$OMP     PARALLEL DO PRIVATE(k) DEFAULT(NONE) SHARED(cosKVals,coeff,gbo)
      DO k = gbo(1, 3), gbo(2, 3)
         cosKVals(k) = COS(coeff*REAL(k, dp))
      END DO

!$OMP     PARALLEL DO DEFAULT(NONE) PRIVATE(i,j,k,ii,coeff,c23) SHARED(spline_g,cosIVals,cosJVals,cosKVals)
      DO ii = 1, SIZE(spline_g%cc)
         i = spline_g%pw_grid%g_hat(1, ii)
         j = spline_g%pw_grid%g_hat(2, ii)
         k = spline_g%pw_grid%g_hat(3, ii)
         ! no opt
!FM                coeff=1.0/((cosVal(1)*cosVal(2)*cosVal(3))/27.0_dp+&
!FM                     (cosVal(1)*cosVal(2)+cosVal(1)*cosVal(3)+&
!FM                     cosVal(2)*cosVal(3))*2.0_dp/27.0_dp+&
!FM                     (cosVal(1)+cosVal(2)+cosVal(3))*4.0_dp/27.0_dp+&
!FM                     8.0_dp/27.0_dp)
         ! opt
         c23 = cosJVals(j)*cosKVals(k)
         coeff = 27.0_dp/(cosIVals(i)*c23+ &
                          (cosIVals(i)*cosJVals(j)+cosIVals(i)*cosKVals(k)+c23)*2.0_dp+ &
                          (cosIVals(i)+cosJVals(j)+cosKVals(k))*4.0_dp+ &
                          8.0_dp)

         spline_g%cc(ii) = spline_g%cc(ii)*coeff

      END DO
      DEALLOCATE (cosIVals, cosJVals, cosKVals)

      CALL timestop(handle)
   END SUBROUTINE pw_spline3_interpolate_values_g

! **************************************************************************************************
!> \brief rescales the derivatives from gridspacing=1 to the real derivatives
!> \param deriv_vals_r an array of x,y,z derivatives
!> \param transpose if true applies the transpose of the map (defaults to
!>        false)
!> \param scale a scaling factor (defaults to 1.0)
!> \par History
!>      06.2003 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE pw_spline_scale_deriv(deriv_vals_r, transpose, scale)
      TYPE(pw_p_type), DIMENSION(3)                      :: deriv_vals_r
      LOGICAL, INTENT(in), OPTIONAL                      :: transpose
      REAL(KIND=dp), INTENT(in), OPTIONAL                :: scale

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_spline_scale_deriv', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, idir, j, k
      INTEGER, DIMENSION(2, 3)                           :: bo
      INTEGER, DIMENSION(3)                              :: n_tot
      LOGICAL                                            :: diag, my_transpose
      REAL(KIND=dp)                                      :: dVal1, dVal2, dVal3, my_scale, scalef
      REAL(KIND=dp), DIMENSION(3, 3)                     :: dh_inv, h_grid
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: ddata, ddata2, ddata3

      CALL timeset(routineN, handle)

      my_transpose = .FALSE.
      IF (PRESENT(transpose)) my_transpose = transpose
      my_scale = 1.0_dp
      IF (PRESENT(scale)) my_scale = scale
      n_tot(1:3) = deriv_vals_r(1)%pw%pw_grid%npts(1:3)
      bo = deriv_vals_r(1)%pw%pw_grid%bounds_local
      dh_inv = deriv_vals_r(1)%pw%pw_grid%dh_inv

      ! map grid to real derivative
      diag = .TRUE.
      IF (my_transpose) THEN
         DO j = 1, 3
            DO i = 1, 3
               h_grid(j, i) = my_scale*dh_inv(i, j) ! REAL(n_tot(i),dp)*cell_h_inv(i,j)
               IF (i /= j .AND. h_grid(j, i) /= 0.0_dp) diag = .FALSE.
            END DO
         END DO
      ELSE
         DO j = 1, 3
            DO i = 1, 3
               h_grid(i, j) = my_scale*dh_inv(i, j) ! REAL(n_tot(i),dp)*cell_h_inv(i,j)
               IF (i /= j .AND. h_grid(i, j) /= 0.0_dp) diag = .FALSE.
            END DO
         END DO
      END IF

      IF (diag) THEN
         DO idir = 1, 3
            ddata => deriv_vals_r(idir)%pw%cr3d
            scalef = h_grid(idir, idir)
            CALL dscal((bo(2, 1)-bo(1, 1)+1)*(bo(2, 2)-bo(1, 2)+1)*(bo(2, 3)-bo(1, 3)+1), &
                       scalef, ddata, 1)
         END DO
      ELSE
         ddata => deriv_vals_r(1)%pw%cr3d
         ddata2 => deriv_vals_r(2)%pw%cr3d
         ddata3 => deriv_vals_r(3)%pw%cr3d
!$OMP        PARALLEL DO DEFAULT(NONE) PRIVATE(k,j,i,dVal1,dVal2,dVal3) &
!$OMP                 SHARED(ddata,ddata2,ddata3,h_grid,bo)
         DO k = bo(1, 3), bo(2, 3)
            DO j = bo(1, 2), bo(2, 2)
               DO i = bo(1, 1), bo(2, 1)

                  dVal1 = ddata(i, j, k)
                  dVal2 = ddata2(i, j, k)
                  dVal3 = ddata3(i, j, k)

                  ddata(i, j, k) = h_grid(1, 1)*dVal1+ &
                                   h_grid(2, 1)*dVal2+h_grid(3, 1)*dVal3
                  ddata2(i, j, k) = h_grid(1, 2)*dVal1+ &
                                    h_grid(2, 2)*dVal2+h_grid(3, 2)*dVal3
                  ddata3(i, j, k) = h_grid(1, 3)*dVal1+ &
                                    h_grid(2, 3)*dVal2+h_grid(3, 3)*dVal3

               END DO
            END DO
         END DO
      END IF

      CALL timestop(handle)
   END SUBROUTINE pw_spline_scale_deriv

! **************************************************************************************************
!> \brief calculates the FFT of the values of the x,y,z (idir=1,2,3)
!>      derivative of the cubic spline
!> \param spline_g on entry the FFT of the coefficents of the spline
!>        will contain the FFT of the derivative
!> \param idir direction of the derivative
!> \par History
!>      06.2003 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      the distance between gridpoints is assumed to be 1
! **************************************************************************************************
   SUBROUTINE pw_spline3_deriv_g(spline_g, idir)
      TYPE(pw_type), POINTER                             :: spline_g
      INTEGER, INTENT(in)                                :: idir

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_spline3_deriv_g', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: inv9 = 1.0_dp/9.0_dp

      INTEGER                                            :: handle, i, ii, j, k
      INTEGER, DIMENSION(2, 3)                           :: bo, gbo
      INTEGER, DIMENSION(3)                              :: n, n_tot
      REAL(KIND=dp)                                      :: coeff, tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: csIVals, csJVals, csKVals

      CALL timeset(routineN, handle)

      n(1:3) = spline_g%pw_grid%npts_local(1:3)
      n_tot(1:3) = spline_g%pw_grid%npts(1:3)
      bo = spline_g%pw_grid%bounds_local
      gbo = spline_g%pw_grid%bounds

      IF(.NOT.(ASSOCIATED(spline_g)))CALL cp__a("pw/pw_spline_utils.F",404)
      IF(.NOT.(spline_g%in_use == COMPLEXDATA1D))CALL cp__a("pw/pw_spline_utils.F",405)
      IF(.NOT.(spline_g%in_space == RECIPROCALSPACE))CALL cp__a("pw/pw_spline_utils.F",406)
      IF(.NOT.(.NOT. spline_g%pw_grid%spherical))CALL cp__a("pw/pw_spline_utils.F",407)
      IF(.NOT.(spline_g%pw_grid%grid_span == FULLSPACE))CALL cp__a("pw/pw_spline_utils.F",408)

      ALLOCATE (csIVals(gbo(1, 1):gbo(2, 1)), &
                csJVals(gbo(1, 2):gbo(2, 2)), &
                csKVals(gbo(1, 3):gbo(2, 3)))

      coeff = twopi/n_tot(1)
      IF (idir == 1) THEN
!$OMP        PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(gbo,csIVals,coeff)
         DO i = gbo(1, 1), gbo(2, 1)
            csIVals(i) = SIN(coeff*REAL(i, dp))
         END DO
      ELSE
!$OMP        PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(gbo,csIVals,coeff)
         DO i = gbo(1, 1), gbo(2, 1)
            csIVals(i) = COS(coeff*REAL(i, dp))
         END DO
      END IF
      coeff = twopi/n_tot(2)
      IF (idir == 2) THEN
!$OMP        PARALLEL DO PRIVATE(j) DEFAULT(NONE) SHARED(gbo,csJVals,coeff)
         DO j = gbo(1, 2), gbo(2, 2)
            csJVals(j) = SIN(coeff*REAL(j, dp))
         END DO
      ELSE
!$OMP        PARALLEL DO PRIVATE(j) DEFAULT(NONE) SHARED(gbo,csJVals,coeff)
         DO j = gbo(1, 2), gbo(2, 2)
            csJVals(j) = COS(coeff*REAL(j, dp))
         END DO
      END IF
      coeff = twopi/n_tot(3)
      IF (idir == 3) THEN
!$OMP        PARALLEL DO PRIVATE(k) DEFAULT(NONE) SHARED(gbo,csKVals,coeff)
         DO k = gbo(1, 3), gbo(2, 3)
            csKVals(k) = SIN(coeff*REAL(k, dp))
         END DO
      ELSE
!$OMP        PARALLEL DO PRIVATE(k) DEFAULT(NONE) SHARED(gbo,csKVals,coeff)
         DO k = gbo(1, 3), gbo(2, 3)
            csKVals(k) = COS(coeff*REAL(k, dp))
         END DO
      END IF

      SELECT CASE (idir)
      CASE (1)
         ! x deriv
!$OMP        PARALLEL DO PRIVATE(ii,k,j,i,coeff,tmp) DEFAULT(NONE) SHARED(spline_g,csIVals,csJVals,csKVals)
         DO ii = 1, SIZE(spline_g%cc)
            i = spline_g%pw_grid%g_hat(1, ii)
            j = spline_g%pw_grid%g_hat(2, ii)
            k = spline_g%pw_grid%g_hat(3, ii)
!FM                ! formula
!FM                coeff=(sinVal(1)*cosVal(2)*cosVal(3))/9.0_dp+&
!FM                     (sinVal(1)*cosVal(2)+sinVal(1)*cosVal(3))*2.0_dp/9.0_dp+&
!FM                     sinVal(1)*4.0_dp/9.0_dp
            tmp = csIVals(i)*csJVals(j)
            coeff = (tmp*csKVals(k)+ &
                     (tmp+csIVals(i)*csKVals(k))*2.0_dp+ &
                     csIVals(i)*4.0_dp)*inv9

            spline_g%cc(ii) = spline_g%cc(ii)* &
                              CMPLX(0.0_dp, coeff, dp)
         END DO
      CASE (2)
         ! y deriv
!$OMP        PARALLEL DO PRIVATE(ii,k,j,i,coeff,tmp) DEFAULT(NONE) SHARED(spline_g,csIVals,csJVals,csKVals)
         DO ii = 1, SIZE(spline_g%cc)
            i = spline_g%pw_grid%g_hat(1, ii)
            j = spline_g%pw_grid%g_hat(2, ii)
            k = spline_g%pw_grid%g_hat(3, ii)

            tmp = csIVals(i)*csJVals(j)
            coeff = (tmp*csKVals(k)+ &
                     (tmp+csJVals(j)*csKVals(k))*2.0_dp+ &
                     csJVals(j)*4.0_dp)*inv9

            spline_g%cc(ii) = spline_g%cc(ii)* &
                              CMPLX(0.0_dp, coeff, dp)
         END DO
      CASE (3)
         ! z deriv
!$OMP        PARALLEL DO PRIVATE(ii,k,j,i,coeff,tmp) DEFAULT(NONE) SHARED(spline_g,csIVals,csJVals,csKVals)
         DO ii = 1, SIZE(spline_g%cc)
            i = spline_g%pw_grid%g_hat(1, ii)
            j = spline_g%pw_grid%g_hat(2, ii)
            k = spline_g%pw_grid%g_hat(3, ii)

            tmp = csIVals(i)*csKVals(k)
            coeff = (tmp*csJVals(j)+ &
                     (tmp+csJVals(j)*csKVals(k))*2.0_dp+ &
                     csKVals(k)*4.0_dp)*inv9

            spline_g%cc(ii) = spline_g%cc(ii)* &
                              CMPLX(0.0_dp, coeff, dp)
         END DO
      END SELECT

      DEALLOCATE (csIVals, csJVals, csKVals)

      CALL timestop(handle)
   END SUBROUTINE pw_spline3_deriv_g

! **************************************************************************************************
!> \brief calculates the FFT of the values of the x,y,z (idir=1,2,3)
!>      derivative of the quadratic spline
!> \param spline_g on entry the FFT of the coefficents of the spline
!>        will contain the FFT of the derivative
!> \param idir direction of the derivative
!> \par History
!>      06.2003 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      the distance between gridpoints is assumed to be 1
! **************************************************************************************************
   SUBROUTINE pw_spline2_deriv_g(spline_g, idir)
      TYPE(pw_type), POINTER                             :: spline_g
      INTEGER, INTENT(in)                                :: idir

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_spline2_deriv_g', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: inv16 = 1.0_dp/16.0_dp

      INTEGER                                            :: handle, i, ii, j, k
      INTEGER, DIMENSION(2, 3)                           :: bo
      INTEGER, DIMENSION(3)                              :: n, n_tot
      REAL(KIND=dp)                                      :: coeff, tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: csIVals, csJVals, csKVals

      CALL timeset(routineN, handle)

      n(1:3) = spline_g%pw_grid%npts_local(1:3)
      n_tot(1:3) = spline_g%pw_grid%npts(1:3)
      bo = spline_g%pw_grid%bounds

      IF(.NOT.(ASSOCIATED(spline_g)))CALL cp__a("pw/pw_spline_utils.F",542)
      IF(.NOT.(spline_g%in_use == COMPLEXDATA1D))CALL cp__a("pw/pw_spline_utils.F",543)
      IF(.NOT.(spline_g%in_space == RECIPROCALSPACE))CALL cp__a("pw/pw_spline_utils.F",544)
      IF(.NOT.(.NOT. spline_g%pw_grid%spherical))CALL cp__a("pw/pw_spline_utils.F",545)
      IF(.NOT.(spline_g%pw_grid%grid_span == FULLSPACE))CALL cp__a("pw/pw_spline_utils.F",546)

      ALLOCATE (csIVals(bo(1, 1):bo(2, 1)), csJVals(bo(1, 2):bo(2, 2)), &
                csKVals(bo(1, 3):bo(2, 3)))

      coeff = twopi/n_tot(1)
      IF (idir == 1) THEN
!$OMP        PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(bo,coeff,csIVals)
         DO i = bo(1, 1), bo(2, 1)
            csIVals(i) = SIN(coeff*REAL(i, dp))
         END DO
      ELSE
!$OMP        PARALLEL DO PRIVATE(i) DEFAULT(NONE) SHARED(bo,coeff,csIVals)
         DO i = bo(1, 1), bo(2, 1)
            csIVals(i) = COS(coeff*REAL(i, dp))
         END DO
      END IF
      coeff = twopi/n_tot(2)
      IF (idir == 2) THEN
!$OMP        PARALLEL DO PRIVATE(j) DEFAULT(NONE) SHARED(bo,coeff,csJVals)
         DO j = bo(1, 2), bo(2, 2)
            csJVals(j) = SIN(coeff*REAL(j, dp))
         END DO
      ELSE
!$OMP        PARALLEL DO PRIVATE(j) DEFAULT(NONE) SHARED(bo,coeff,csJVals)
         DO j = bo(1, 2), bo(2, 2)
            csJVals(j) = COS(coeff*REAL(j, dp))
         END DO
      END IF
      coeff = twopi/n_tot(3)
      IF (idir == 3) THEN
!$OMP        PARALLEL DO PRIVATE(k) DEFAULT(NONE) SHARED(bo,coeff,csKVals)
         DO k = bo(1, 3), bo(2, 3)
            csKVals(k) = SIN(coeff*REAL(k, dp))
         END DO
      ELSE
!$OMP        PARALLEL DO PRIVATE(k) DEFAULT(NONE) SHARED(bo,coeff,csKVals)
         DO k = bo(1, 3), bo(2, 3)
            csKVals(k) = COS(coeff*REAL(k, dp))
         END DO
      END IF

      SELECT CASE (idir)
      CASE (1)
         ! x deriv
!$OMP        PARALLEL DO PRIVATE(ii,k,j,i,coeff,tmp) SHARED(spline_g,csIVals,csJVals,csKVals) DEFAULT(NONE)
         DO ii = 1, SIZE(spline_g%cc)
            i = spline_g%pw_grid%g_hat(1, ii)
            j = spline_g%pw_grid%g_hat(2, ii)
            k = spline_g%pw_grid%g_hat(3, ii)
!FM                ! formula
!FM                coeff=(sinVal(1)*cosVal(2)*cosVal(3))/16.0_dp+&
!FM                     (sinVal(1)*cosVal(2)+sinVal(1)*cosVal(3))*3.0_dp/16.0_dp+&
!FM                     sinVal(1)*9.0_dp/16.0_dp
            tmp = csIVals(i)*csJVals(j)
            coeff = (tmp*csKVals(k)+ &
                     (tmp+csIVals(i)*csKVals(k))*3.0_dp+ &
                     csIVals(i)*9.0_dp)*inv16

            spline_g%cc(ii) = spline_g%cc(ii)* &
                              CMPLX(0.0_dp, coeff, dp)
         END DO
      CASE (2)
         ! y deriv
!$OMP        PARALLEL DO PRIVATE(ii,k,j,i,coeff,tmp) DEFAULT(NONE) SHARED(spline_g,csIVals,csJVals,csKVals)
         DO ii = 1, SIZE(spline_g%cc)
            i = spline_g%pw_grid%g_hat(1, ii)
            j = spline_g%pw_grid%g_hat(2, ii)
            k = spline_g%pw_grid%g_hat(3, ii)

            tmp = csIVals(i)*csJVals(j)
            coeff = (tmp*csKVals(k)+ &
                     (tmp+csJVals(j)*csKVals(k))*3.0_dp+ &
                     csJVals(j)*9.0_dp)*inv16

            spline_g%cc(ii) = spline_g%cc(ii)* &
                              CMPLX(0.0_dp, coeff, dp)
         END DO
      CASE (3)
         ! z deriv
!$OMP        PARALLEL DO PRIVATE(ii,k,j,i,coeff,tmp) DEFAULT(NONE) SHARED(spline_g,csIVals,csJVals,csKVals)
         DO ii = 1, SIZE(spline_g%cc)
            i = spline_g%pw_grid%g_hat(1, ii)
            j = spline_g%pw_grid%g_hat(2, ii)
            k = spline_g%pw_grid%g_hat(3, ii)

            tmp = csIVals(i)*csKVals(k)
            coeff = (tmp*csJVals(j)+ &
                     (tmp+csJVals(j)*csKVals(k))*3.0_dp+ &
                     csKVals(k)*9.0_dp)*inv16

            spline_g%cc(ii) = spline_g%cc(ii)* &
                              CMPLX(0.0_dp, coeff, dp)
         END DO
      END SELECT

      DEALLOCATE (csIVals, csJVals, csKVals)

      CALL timestop(handle)
   END SUBROUTINE pw_spline2_deriv_g

! **************************************************************************************************
!> \brief applies a nearest neighbor linear operator to a stripe in x direction:
!>      out_val(i)=sum(weight(j)*in_val(i+j-1),j=0..2)
!> \param weights the weights of the linear operator
!> \param in_val the argument to the operator
!> \param in_val_first the first argument (needed to calculate out_val(1))
!> \param in_val_last the last argument (needed to calculate out_val(n_el))
!> \param out_val the place where the result is accumulated
!> \param n_el the number of elements in in_v and out_v
!> \par History
!>      04.2004 created [fawzi]
!> \author fawzi
!> \note
!>      uses 2 read streams and 1 write stream
! **************************************************************************************************
   SUBROUTINE pw_compose_stripe(weights, in_val, in_val_first, in_val_last, &
                                out_val, n_el)
      REAL(kind=dp), DIMENSION(0:2), INTENT(in)          :: weights
      REAL(kind=dp), DIMENSION(*), INTENT(in)            :: in_val
      REAL(kind=dp), INTENT(in)                          :: in_val_first, in_val_last
      REAL(kind=dp), DIMENSION(*), INTENT(inout)         :: out_val
      INTEGER                                            :: n_el

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_compose_stripe', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i
      REAL(kind=dp)                                      :: v0, v1, v2

!1:n_el), &
!1:n_el), &

      IF (n_el < 1) RETURN
      v0 = in_val_first
      v1 = in_val(1)
      IF (weights(1) == 0.0_dp) THEN
         ! optimized version for x deriv
         DO i = 1, n_el-3, 3
            v2 = in_val(i+1)
            out_val(i) = out_val(i)+ &
                         weights(0)*v0+ &
                         weights(2)*v2
            v0 = in_val(i+2)
            out_val(i+1) = out_val(i+1)+ &
                           weights(0)*v1+ &
                           weights(2)*v0
            v1 = in_val(i+3)
            out_val(i+2) = out_val(i+2)+ &
                           weights(0)*v2+ &
                           weights(2)*v1
         END DO
      ELSE
         ! generic version
         DO i = 1, n_el-3, 3
            v2 = in_val(i+1)
            out_val(i) = out_val(i)+ &
                         weights(0)*v0+ &
                         weights(1)*v1+ &
                         weights(2)*v2
            v0 = in_val(i+2)
            out_val(i+1) = out_val(i+1)+ &
                           weights(0)*v1+ &
                           weights(1)*v2+ &
                           weights(2)*v0
            v1 = in_val(i+3)
            out_val(i+2) = out_val(i+2)+ &
                           weights(0)*v2+ &
                           weights(1)*v0+ &
                           weights(2)*v1
         END DO
      END IF
      SELECT CASE (MODULO (n_el-1, 3))
      CASE (0)
         v2 = in_val_last
         out_val(n_el) = out_val(n_el)+ &
                         weights(0)*v0+ &
                         weights(1)*v1+ &
                         weights(2)*v2
      CASE (1)
         v2 = in_val(n_el)
         out_val(n_el-1) = out_val(n_el-1)+ &
                           weights(0)*v0+ &
                           weights(1)*v1+ &
                           weights(2)*v2
         v0 = in_val_last
         out_val(n_el) = out_val(n_el)+ &
                         weights(0)*v1+ &
                         weights(1)*v2+ &
                         weights(2)*v0
      CASE (2)
         v2 = in_val(n_el-1)
         out_val(n_el-2) = out_val(n_el-2)+ &
                           weights(0)*v0+ &
                           weights(1)*v1+ &
                           weights(2)*v2
         v0 = in_val(n_el)
         out_val(n_el-1) = out_val(n_el-1)+ &
                           weights(0)*v1+ &
                           weights(1)*v2+ &
                           weights(2)*v0
         v1 = in_val_last
         out_val(n_el) = out_val(n_el)+ &
                         weights(0)*v2+ &
                         weights(1)*v0+ &
                         weights(2)*v1
      END SELECT

   END SUBROUTINE pw_compose_stripe

! **************************************************************************************************
!> \brief private routine that computes pw_nn_compose_r (it seems that without
!>      passing arrays in this way either some compiler do a copyin/out (xlf)
!>      or by inlining suboptimal code is produced (nag))
!> \param weights a 3x3x3 array with the linear operator
!> \param in_val the argument for the linear operator
!> \param out_val place where the value of the linear oprator should be added
!> \param pw_in pw to be able to get the needed meta data about in_val and
!>        out_val
!> \param bo boundaries of in_val and out_val
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_nn_compose_r_work(weights, in_val, out_val, pw_in, bo)
      REAL(kind=dp), DIMENSION(0:2, 0:2, 0:2)            :: weights
      INTEGER, DIMENSION(2, 3)                           :: bo
      TYPE(pw_type), POINTER                             :: pw_in
      REAL(kind=dp), DIMENSION(bo(1, 1):bo(2, 1), bo(1, &
         2):bo(2, 2), bo(1, 3):bo(2, 3)), INTENT(inout)  :: out_val
      REAL(kind=dp), DIMENSION(bo(1, 1):bo(2, 1), bo(1, &
         2):bo(2, 2), bo(1, 3):bo(2, 3)), INTENT(in)     :: in_val

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_nn_compose_r_work', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, jw, k, kw, myj, myk
      INTEGER, DIMENSION(2, 3)                           :: gbo
      INTEGER, DIMENSION(3)                              :: s
      LOGICAL                                            :: has_boundary, yderiv, zderiv
      REAL(kind=dp)                                      :: in_val_f, in_val_l
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: l_boundary, tmp, u_boundary

      zderiv = ALL(weights(:, :, 1) == 0.0_dp)
      yderiv = ALL(weights(:, 1, :) == 0.0_dp)
      bo = pw_in%pw_grid%bounds_local
      gbo = pw_in%pw_grid%bounds
      DO i = 1, 3
         s(i) = bo(2, i)-bo(1, i)+1
      END DO
      IF (ANY(s < 1)) RETURN
      has_boundary = ANY(pw_in%pw_grid%bounds_local(:, 1) /= &
                         pw_in%pw_grid%bounds(:, 1))
      IF (has_boundary) THEN
         ALLOCATE (l_boundary(bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)), &
                   u_boundary(bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)), &
                   tmp(bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
         tmp(:, :) = pw_in%cr3d(bo(2, 1), :, :)
         CALL mp_sendrecv(tmp, pw_in%pw_grid%para%pos_of_x( &
                          gbo(1, 1)+MODULO(bo(2, 1)+1-gbo(1, 1), gbo(2, 1)-gbo(1, 1)+1)), &
                          l_boundary, pw_in%pw_grid%para%pos_of_x( &
                          gbo(1, 1)+MODULO(bo(1, 1)-1-gbo(1, 1), gbo(2, 1)-gbo(1, 1)+1)), &
                          pw_in%pw_grid%para%group)
         tmp(:, :) = pw_in%cr3d(bo(1, 1), :, :)
         CALL mp_sendrecv(tmp, pw_in%pw_grid%para%pos_of_x( &
                          gbo(1, 1)+MODULO(bo(1, 1)-1-gbo(1, 1), gbo(2, 1)-gbo(1, 1)+1)), &
                          u_boundary, pw_in%pw_grid%para%pos_of_x( &
                          gbo(1, 1)+MODULO(bo(2, 1)+1-gbo(1, 1), gbo(2, 1)-gbo(1, 1)+1)), &
                          pw_in%pw_grid%para%group)
         DEALLOCATE (tmp)
      END IF

!$OMP   PARALLEL DO DEFAULT(NONE) PRIVATE(k,kw,myk,j,jw,myj,in_val_f,&
!$OMP       in_val_l) SHARED(zderiv,yderiv,bo,in_val,out_val,s,l_boundary,&
!$OMP       u_boundary,weights,has_boundary)
      DO k = 0, s(3)-1
         DO kw = 0, 2
            myk = bo(1, 3)+MODULO(k+kw-1, s(3))
            IF (zderiv .AND. kw == 1) CYCLE
            DO j = 0, s(2)-1
               DO jw = 0, 2
                  myj = bo(1, 2)+MODULO(j+jw-1, s(2))
                  IF (yderiv .AND. jw == 1) CYCLE
                  IF (has_boundary) THEN
                     in_val_f = l_boundary(myj, myk)
                     in_val_l = u_boundary(myj, myk)
                  ELSE
                     in_val_f = in_val(bo(2, 1), myj, myk)
                     in_val_l = in_val(bo(1, 1), myj, myk)
                  END IF
                  CALL pw_compose_stripe(weights=weights(:, jw, kw), &
                                         in_val=in_val(:, myj, myk), &
                                         in_val_first=in_val_f, in_val_last=in_val_l, &
                                         out_val=out_val(:, bo(1, 2)+j, bo(1, 3)+k), n_el=s(1))
               END DO
            END DO
         END DO
      END DO
      IF (has_boundary) THEN
         DEALLOCATE (l_boundary, u_boundary)
      END IF
   END SUBROUTINE pw_nn_compose_r_work

! **************************************************************************************************
!> \brief applies a nearest neighbor linear operator to a pw in real space
!> \param weights a 3x3x3 array with the linear operator
!> \param pw_in the argument for the linear operator
!> \param pw_out place where the value of the linear oprator should be added
!> \author fawzi
!> \note
!>      has specialized versions for derivative operator (with central values==0)
! **************************************************************************************************
   SUBROUTINE pw_nn_compose_r(weights, pw_in, pw_out)
      REAL(kind=dp), DIMENSION(0:2, 0:2, 0:2)            :: weights
      TYPE(pw_type), POINTER                             :: pw_in, pw_out

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_nn_compose_r', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(pw_in)))CALL cp__a("pw/pw_spline_utils.F",866)
      IF(.NOT.(pw_in%in_space == REALSPACE))CALL cp__a("pw/pw_spline_utils.F",867)
      IF(.NOT.(pw_in%in_use == REALDATA3D))CALL cp__a("pw/pw_spline_utils.F",868)
      IF(.NOT.(ASSOCIATED(pw_out)))CALL cp__a("pw/pw_spline_utils.F",869)
      IF(.NOT.(pw_out%in_space == REALSPACE))CALL cp__a("pw/pw_spline_utils.F",870)
      IF(.NOT.(pw_out%in_use == REALDATA3D))CALL cp__a("pw/pw_spline_utils.F",871)
      IF (.NOT. ALL(pw_in%pw_grid%bounds_local(:, 2:3) == pw_in%pw_grid%bounds(:, 2:3))) THEN
         CALL cp__b("pw/pw_spline_utils.F",873,"wrong pw distribution")
      END IF
      CALL pw_nn_compose_r_work(weights=weights, in_val=pw_in%cr3d, &
                                out_val=pw_out%cr3d, pw_in=pw_in, bo=pw_in%pw_grid%bounds_local)
      CALL timestop(handle)
   END SUBROUTINE pw_nn_compose_r

! **************************************************************************************************
!> \brief calculates the values of a nearest neighbor smearing
!> \param pw_in the argument for the linear operator
!> \param pw_out place where the smeared values should be added
!> \param coeffs array with the coefficent of the smearing, ordered with
!>        the distance from the center: coeffs(1) the coeff of the central
!>        element, coeffs(2) the coeff of the 6 element with distance 1,
!>        coeff(3) the coeff of the 12 elements at distance sqrt(2),
!>        coeff(4) the coeff of the 8 elements at distance sqrt(3).
!> \author Fawzi Mohamed
!> \note
!>      does not normalize the smear to 1.
!>      with coeff=(/ 8._dp/27._dp, 2._dp/27._dp, 1._dp/54._dp, 1._dp/216._dp /)
!>      is equivalent to pw_spline3_evaluate_values_g, with
!>      coeff=(/ 27._dp/64._dp, 9._dp/128._dp, 3._dp/256._dp, 1._dp/512._dp /)
! **************************************************************************************************
   SUBROUTINE pw_nn_smear_r(pw_in, pw_out, coeffs)
      TYPE(pw_type), POINTER                             :: pw_in, pw_out
      REAL(KIND=dp), DIMENSION(4), INTENT(in)            :: coeffs

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_nn_smear_r', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k
      REAL(kind=dp), DIMENSION(-1:1, -1:1, -1:1)         :: weights

      DO k = -1, 1
         DO j = -1, 1
            DO i = -1, 1
               weights(i, j, k) = coeffs(ABS(i)+ABS(j)+ABS(k)+1)
            END DO
         END DO
      END DO

      CALL pw_nn_compose_r(weights=weights, pw_in=pw_in, pw_out=pw_out)
   END SUBROUTINE pw_nn_smear_r

! **************************************************************************************************
!> \brief calculates a nearest neighbor central derivative.
!>      for the x dir:
!>      pw_out%cr3d(i,j,k)=( pw_in(i+1,j,k)-pw_in(i-1,j,k) )*coeff(1)+
!>             ( pw_in(i+1,j(+-)1,k)-pw_in(i-1,j(+-)1,k)+
!>               pw_in(i+1,j,k(+-)1)-pw_in(i-1,j,k(+-)1) )*coeff(2)+
!>             ( pw_in(i+1,j(+-)1,k(+-)1)-pw_in(i-1,j(+-)1,k(+-)1)+
!>               pw_in(i+1,j(+-)1,k(-+)1)-pw_in(i-1,j(+-)1,k(-+)1) )*coeff(3)
!>      periodic boundary conditions are applied
!> \param pw_in the argument for the linear operator
!> \param pw_out place where the smeared values should be added
!> \param coeffs array with the coefficent of the front (positive) plane
!>        of the central derivative, ordered with
!>        the distance from the center: coeffs(1) the coeff of the central
!>        element, coeffs(2) the coeff of the 4 element with distance 1,
!>        coeff(3) the coeff of the 4 elements at distance sqrt(2)
!> \param idir ...
!> \author Fawzi Mohamed
!> \note
!>      with coeff=(/ 2.0_dp/9.0_dp,  1.0_dp/18.0_dp, 1.0_dp/72.0_dp /)
!>      is equivalent to pw_spline3_deriv_r, with
!>      coeff=(/ 9.0_dp/32.0_dp, 3.0_dp/64.0_dp, 1.0_dp/128.0_dp /)
!>      to pw_spline2_deriv_r
!>      coeff=(/ 25._dp/72._dp, 5._dp/144, 1._dp/288._dp /)
! **************************************************************************************************
   SUBROUTINE pw_nn_deriv_r(pw_in, pw_out, coeffs, idir)
      TYPE(pw_type), POINTER                             :: pw_in, pw_out
      REAL(KIND=dp), DIMENSION(3), INTENT(in)            :: coeffs
      INTEGER                                            :: idir

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_nn_deriv_r', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, idirVal, j, k
      REAL(kind=dp), DIMENSION(-1:1, -1:1, -1:1)         :: weights

      DO k = -1, 1
         DO j = -1, 1
            DO i = -1, 1
               SELECT CASE (idir)
               CASE (1)
                  idirVal = i
               CASE (2)
                  idirVal = j
               CASE (3)
                  idirVal = k
               CASE default
                  CALL cp__b("pw/pw_spline_utils.F",962,"invalid idir ("//TRIM(cp_to_string(idir))//")")
               END SELECT
               IF (idirVal == 0) THEN
                  weights(i, j, k) = 0.0_dp
               ELSE
                  weights(i, j, k) = REAL(idirVal, dp)*coeffs(ABS(i)+ABS(j)+ABS(k))
               END IF
            END DO
         END DO
      END DO

      CALL pw_nn_compose_r(weights=weights, pw_in=pw_in, pw_out=pw_out)
   END SUBROUTINE pw_nn_deriv_r

! **************************************************************************************************
!> \brief low level function that adds a coarse grid
!>      to a fine grid.
!>      If pbc is true periodic boundary conditions are applied
!>
!>      It will add to
!>
!>        fine_values(2*coarse_bounds(1,1):2*coarse_bounds(2,1),
!>                    2*coarse_bounds(1,2):2*coarse_bounds(2,2),
!>                    2*coarse_bounds(1,3):2*coarse_bounds(2,3))
!>
!>      using
!>
!>        coarse_coeffs(coarse_bounds(1,1):coarse_bounds(2,1),
!>                      coarse_bounds(1,2):coarse_bounds(2,2),
!>                      coarse_bounds(1,3):coarse_bounds(2,3))
!>
!>      composed with the weights obtained by the direct product of the
!>      1d coefficents weights:
!>
!>      for i,j,k in -3..3
!>         w(i,j,k)=weights_1d(abs(i)+1)*weights_1d(abs(j)+1)*
!>                  weights_1d(abs(k)+1)
!> \param coarse_coeffs_pw the values of the coefficients
!> \param fine_values_pw where to add the values due to the
!>        coarse coeffs
!> \param weights_1d the weights of the 1d smearing
!> \param w_border0 the 1d weight at the border (when pbc is false)
!> \param w_border1 the 1d weights for a point one off the border
!>        (w_border1(1) is the weight of the coefficent at the border)
!>        (used if pbc is false)
!> \param pbc if periodic boundary conditions should be applied
!> \param safe_computation ...
!> \author fawzi
!> \note
!>      coarse looping is continuos, I did not check if keeping the fine looping
!>      contiguos is better.
!>      And I ask myself yet again why, why we use x-slice distribution,
!>      z-slice distribution would be much better performancewise
!>      (and would semplify this code enormously).
!>      fine2coarse has much more understandable parallel part (build up of
!>      send/rcv sizes,... but worse if you have really a lot of processors,
!>      probabily irrelevant because it is not critical) [fawzi].
! **************************************************************************************************
   SUBROUTINE add_coarse2fine(coarse_coeffs_pw, fine_values_pw, &
                              weights_1d, w_border0, w_border1, pbc, safe_computation)
      TYPE(pw_type), POINTER                             :: coarse_coeffs_pw, fine_values_pw
      REAL(kind=dp), DIMENSION(4), INTENT(in)            :: weights_1d
      REAL(kind=dp), INTENT(in)                          :: w_border0
      REAL(kind=dp), DIMENSION(3), INTENT(in)            :: w_border1
      LOGICAL, INTENT(in)                                :: pbc
      LOGICAL, INTENT(in), OPTIONAL                      :: safe_computation

      CHARACTER(len=*), PARAMETER :: routineN = 'add_coarse2fine', &
         routineP = moduleN//':'//routineN

      INTEGER :: coarse_slice_size, f_shift(3), fi, fi_lb, fi_ub, fj, fk, handle, handle2, i, ii, &
         ij, ik, ip, j, k, my_lb, my_ub, n_procs, p, p_lb, p_old, p_ub, rcv_tot_size, rest_b, &
         s(3), send_tot_size, sf, shift, ss, x, x_att, xx
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: rcv_offset, rcv_size, real_rcv_size, &
                                                            send_offset, send_size, sent_size
      INTEGER, DIMENSION(2, 3)                           :: coarse_bo, coarse_gbo, fine_bo, &
                                                            fine_gbo, my_coarse_bo
      INTEGER, DIMENSION(:), POINTER                     :: pos_of_x
      LOGICAL                                            :: has_i_lbound, has_i_ubound, is_split, &
                                                            safe_calc
      REAL(kind=dp)                                      :: v0, v1, v2, v3, wi, wj, wk
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: rcv_buf, send_buf
      REAL(kind=dp), DIMENSION(3)                        :: w_0, ww0
      REAL(kind=dp), DIMENSION(4)                        :: w_1, ww1
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: coarse_coeffs, fine_values

      CALL timeset(routineN, handle)
!    CALL timeset(routineN//"_pre",handle2)
      safe_calc = .FALSE.
      IF (PRESENT(safe_computation)) safe_calc = safe_computation
      CALL mp_comm_compare(coarse_coeffs_pw%pw_grid%para%group, &
                           fine_values_pw%pw_grid%para%group, ii)
      IF (ii > 1) THEN
         CALL cp__b("pw/pw_spline_utils.F",1055,"")
      END IF
      my_coarse_bo = coarse_coeffs_pw%pw_grid%bounds_local
      coarse_gbo = coarse_coeffs_pw%pw_grid%bounds
      fine_bo = fine_values_pw%pw_grid%bounds_local
      fine_gbo = fine_values_pw%pw_grid%bounds
      f_shift = fine_gbo(1, :)-2*coarse_gbo(1, :)
      DO j = 2, 3
         DO i = 1, 2
            coarse_bo(i, j) = FLOOR((fine_bo(i, j)-f_shift(j))/2.)
         END DO
      END DO
      IF (fine_bo(1, 1) <= fine_bo(2, 1)) THEN
         coarse_bo(1, 1) = FLOOR((fine_bo(1, 1)-2-f_shift(1))/2.)
         coarse_bo(2, 1) = FLOOR((fine_bo(2, 1)+3-f_shift(1))/2.)
      ELSE
         coarse_bo(1, 1) = coarse_gbo(2, 1)
         coarse_bo(2, 1) = coarse_gbo(2, 1)-1
      END IF
      is_split = ANY(coarse_gbo(:, 1) /= my_coarse_bo(:, 1))
      IF (.NOT. is_split .OR. .NOT. pbc) THEN
         coarse_bo(1, 1) = MAX(coarse_gbo(1, 1), coarse_bo(1, 1))
         coarse_bo(2, 1) = MIN(coarse_gbo(2, 1), coarse_bo(2, 1))
      END IF
      has_i_ubound = (fine_gbo(2, 1) /= fine_bo(2, 1)) .OR. pbc .AND. is_split
      has_i_lbound = (fine_gbo(1, 1) /= fine_bo(1, 1)) .OR. pbc .AND. is_split

      IF (pbc) THEN
         IF(.NOT.(ALL(fine_gbo(1, :) == 2*coarse_gbo(1, :)+f_shift)))CALL cp__a("pw/pw_spline_utils.F",1083)
         IF(.NOT.(ALL(fine_gbo(2, :) == 2*coarse_gbo(2, :)+1+f_shift)))CALL cp__a("pw/pw_spline_utils.F",1084)
      ELSE
         IF(.NOT.(ALL(fine_gbo(2, :) == 2*coarse_gbo(2, :)+f_shift)))CALL cp__a("pw/pw_spline_utils.F",1086)
         IF(.NOT.(ALL(fine_gbo(1, :) == 2*coarse_gbo(1, :)+f_shift)))CALL cp__a("pw/pw_spline_utils.F",1087)
      END IF

      coarse_coeffs => coarse_coeffs_pw%cr3d
      DO i = 1, 3
         s(i) = coarse_gbo(2, i)-coarse_gbo(1, i)+1
      END DO
!       CALL timestop(handle2)
      ! *** parallel case
      IF (is_split) THEN
         CALL timeset(routineN//"_comm", handle2)
         coarse_slice_size = (coarse_bo(2, 2)-coarse_bo(1, 2)+1)* &
                             (coarse_bo(2, 3)-coarse_bo(1, 3)+1)
         n_procs = coarse_coeffs_pw%pw_grid%para%group_size
         ALLOCATE (send_size(0:n_procs-1), send_offset(0:n_procs-1), &
                   sent_size(0:n_procs-1), rcv_size(0:n_procs-1), &
                   rcv_offset(0:n_procs-1), real_rcv_size(0:n_procs-1))

         ! ** rcv size count

         pos_of_x => coarse_coeffs_pw%pw_grid%para%pos_of_x
         p_old = pos_of_x(coarse_gbo(1, 1) &
                          +MODULO(coarse_bo(1, 1)-coarse_gbo(1, 1), s(1)))
         rcv_size = 0
         DO x = coarse_bo(1, 1), coarse_bo(2, 1)
            p = pos_of_x(coarse_gbo(1, 1)+MODULO(x-coarse_gbo(1, 1), s(1)))
            rcv_size(p) = rcv_size(p)+coarse_slice_size
         END DO

         ! ** send size count

         pos_of_x => fine_values_pw%pw_grid%para%pos_of_x
         sf = fine_gbo(2, 1)-fine_gbo(1, 1)+1
         fi_lb = 2*my_coarse_bo(1, 1)-3+f_shift(1)
         fi_ub = 2*my_coarse_bo(2, 1)+3+f_shift(1)
         IF (.NOT. pbc) THEN
            fi_lb = MAX(fi_lb, fine_gbo(1, 1))
            fi_ub = MIN(fi_ub, fine_gbo(2, 1))
         ELSE
            fi_ub = MIN(fi_ub, fi_lb+sf-1)
         END IF
         p_old = pos_of_x(fine_gbo(1, 1)+MODULO(fi_lb-fine_gbo(1, 1), sf))
         p_lb = FLOOR((fi_lb-2-f_shift(1))/2.)
         send_size = 0
         DO x = fi_lb, fi_ub
            p = pos_of_x(fine_gbo(1, 1)+MODULO(x-fine_gbo(1, 1), sf))
            IF (p /= p_old) THEN
               p_ub = FLOOR((x-1+3-f_shift(1))/2.)

               send_size(p_old) = send_size(p_old)+(MIN(p_ub, my_coarse_bo(2, 1)) &
                                                    -MAX(p_lb, my_coarse_bo(1, 1))+1)*coarse_slice_size

               IF (pbc) THEN
                  DO xx = p_lb, coarse_gbo(1, 1)-1
                     x_att = coarse_gbo(1, 1)+MODULO(xx-coarse_gbo(1, 1), s(1))
                     IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                        send_size(p_old) = send_size(p_old)+coarse_slice_size
                     END IF
                  END DO
                  DO xx = coarse_gbo(2, 1)+1, p_ub
                     x_att = coarse_gbo(1, 1)+MODULO(xx-coarse_gbo(1, 1), s(1))
                     IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                        send_size(p_old) = send_size(p_old)+coarse_slice_size
                     END IF
                  END DO
               END IF

               p_old = p
               p_lb = FLOOR((x-2-f_shift(1))/2.)
            END IF
         END DO
         p_ub = FLOOR((fi_ub+3-f_shift(1))/2.)

         send_size(p_old) = send_size(p_old)+(MIN(p_ub, my_coarse_bo(2, 1)) &
                                              -MAX(p_lb, my_coarse_bo(1, 1))+1)*coarse_slice_size

         IF (pbc) THEN
            DO xx = p_lb, coarse_gbo(1, 1)-1
               x_att = coarse_gbo(1, 1)+MODULO(xx-coarse_gbo(1, 1), s(1))
               IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                  send_size(p_old) = send_size(p_old)+coarse_slice_size
               END IF
            END DO
            DO xx = coarse_gbo(2, 1)+1, p_ub
               x_att = coarse_gbo(1, 1)+MODULO(xx-coarse_gbo(1, 1), s(1))
               IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                  send_size(p_old) = send_size(p_old)+coarse_slice_size
               END IF
            END DO
         END IF
         ! ** offsets & alloc send-rcv

         send_tot_size = 0
         DO ip = 0, n_procs-1
            send_offset(ip) = send_tot_size
            send_tot_size = send_tot_size+send_size(ip)
         END DO
         ALLOCATE (send_buf(0:send_tot_size-1))

         rcv_tot_size = 0
         DO ip = 0, n_procs-1
            rcv_offset(ip) = rcv_tot_size
            rcv_tot_size = rcv_tot_size+rcv_size(ip)
         END DO
         IF (.NOT. rcv_tot_size == (coarse_bo(2, 1)-coarse_bo(1, 1)+1)*coarse_slice_size) THEN
            CALL cp__b("pw/pw_spline_utils.F",1192,"Error calculating rcv_tot_size ")
         END IF
         ALLOCATE (rcv_buf(0:rcv_tot_size-1))

         ! ** fill send buffer

         p_old = pos_of_x(fine_gbo(1, 1)+MODULO(fi_lb-fine_gbo(1, 1), sf))
         p_lb = FLOOR((fi_lb-2-f_shift(1))/2.)
         sent_size(:) = send_offset
         ss = my_coarse_bo(2, 1)-my_coarse_bo(1, 1)+1
         DO x = fi_lb, fi_ub
            p = pos_of_x(fine_gbo(1, 1)+MODULO(x-fine_gbo(1, 1), sf))
            IF (p /= p_old) THEN
               shift = FLOOR((fine_gbo(1, 1)+MODULO(x-1-fine_gbo(1, 1), sf)-f_shift(1))/2._dp)- &
                       FLOOR((x-1-f_shift(1))/2._dp)
               p_ub = FLOOR((x-1+3-f_shift(1))/2._dp)

               IF (pbc) THEN
                  DO xx = p_lb+shift, coarse_gbo(1, 1)-1
                     x_att = coarse_gbo(1, 1)+MODULO(xx-coarse_gbo(1, 1), sf)
                     IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                        CALL dcopy(coarse_slice_size, &
                                   coarse_coeffs(x_att, my_coarse_bo(1, 2), &
                                                 my_coarse_bo(1, 3)), ss, send_buf(sent_size(p_old)), 1)
                        sent_size(p_old) = sent_size(p_old)+coarse_slice_size
                     END IF
                  END DO
               END IF

               ii = sent_size(p_old)
               DO k = coarse_bo(1, 3), coarse_bo(2, 3)
                  DO j = coarse_bo(1, 2), coarse_bo(2, 2)
                     DO i = MAX(p_lb+shift, my_coarse_bo(1, 1)), MIN(p_ub+shift, my_coarse_bo(2, 1))
                        send_buf(ii) = coarse_coeffs(i, j, k)
                        ii = ii+1
                     END DO
                  END DO
               END DO
               sent_size(p_old) = ii

               IF (pbc) THEN
                  DO xx = coarse_gbo(2, 1)+1, p_ub+shift
                     x_att = coarse_gbo(1, 1)+MODULO(xx-coarse_gbo(1, 1), s(1))
                     IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                        CALL dcopy(coarse_slice_size, &
                                   coarse_coeffs(x_att, my_coarse_bo(1, 2), &
                                                 my_coarse_bo(1, 3)), ss, &
                                   send_buf(sent_size(p_old)), 1)
                        sent_size(p_old) = sent_size(p_old)+coarse_slice_size
                     END IF
                  END DO
               END IF

               p_old = p
               p_lb = FLOOR((x-2-f_shift(1))/2.)
            END IF
         END DO
         shift = FLOOR((fine_gbo(1, 1)+MODULO(x-1-fine_gbo(1, 1), sf)-f_shift(1))/2._dp)- &
                 FLOOR((x-1-f_shift(1))/2._dp)
         p_ub = FLOOR((fi_ub+3-f_shift(1))/2.)

         IF (pbc) THEN
            DO xx = p_lb+shift, coarse_gbo(1, 1)-1
               x_att = coarse_gbo(1, 1)+MODULO(xx-coarse_gbo(1, 1), s(1))
               IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                  CALL dcopy(coarse_slice_size, &
                             coarse_coeffs(x_att, my_coarse_bo(1, 2), &
                                           my_coarse_bo(1, 3)), ss, send_buf(sent_size(p_old)), 1)
                  sent_size(p_old) = sent_size(p_old)+coarse_slice_size
               END IF
            END DO
         END IF

         ii = sent_size(p_old)
         DO k = coarse_bo(1, 3), coarse_bo(2, 3)
            DO j = coarse_bo(1, 2), coarse_bo(2, 2)
               DO i = MAX(p_lb+shift, my_coarse_bo(1, 1)), MIN(p_ub+shift, my_coarse_bo(2, 1))
                  send_buf(ii) = coarse_coeffs(i, j, k)
                  ii = ii+1
               END DO
            END DO
         END DO
         sent_size(p_old) = ii

         IF (pbc) THEN
            DO xx = coarse_gbo(2, 1)+1, p_ub+shift
               x_att = coarse_gbo(1, 1)+MODULO(xx-coarse_gbo(1, 1), s(1))
               IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                  CALL dcopy(coarse_slice_size, &
                             coarse_coeffs(x_att, my_coarse_bo(1, 2), &
                                           my_coarse_bo(1, 3)), ss, send_buf(sent_size(p_old)), 1)
                  sent_size(p_old) = sent_size(p_old)+coarse_slice_size
               END IF
            END DO
         END IF

         IF(.NOT.(ALL(sent_size(:n_procs-2) == send_offset(1:))))CALL cp__a("pw/pw_spline_utils.F",1288)
         IF(.NOT.(sent_size(n_procs-1) == send_tot_size))CALL cp__a("pw/pw_spline_utils.F",1289)
         ! test send/rcv sizes
         CALL mp_alltoall(send_size, real_rcv_size, 1, coarse_coeffs_pw%pw_grid%para%group)
         IF(.NOT.(ALL(real_rcv_size == rcv_size)))CALL cp__a("pw/pw_spline_utils.F",1292)
         ! all2all
         CALL mp_alltoall(sb=send_buf, scount=send_size, sdispl=send_offset, &
                          rb=rcv_buf, rcount=rcv_size, rdispl=rcv_offset, &
                          group=coarse_coeffs_pw%pw_grid%para%group)

         ! ** reorder rcv buffer
         ! (actually reordering should be needed only with pbc)

         ALLOCATE (coarse_coeffs(coarse_bo(1, 1):coarse_bo(2, 1), &
                                 coarse_bo(1, 2):coarse_bo(2, 2), &
                                 coarse_bo(1, 3):coarse_bo(2, 3)))

         my_lb = MAX(coarse_gbo(1, 1), coarse_bo(1, 1))
         my_ub = MIN(coarse_gbo(2, 1), coarse_bo(2, 1))
         pos_of_x => coarse_coeffs_pw%pw_grid%para%pos_of_x
         sent_size(:) = rcv_offset
         ss = coarse_bo(2, 1)-coarse_bo(1, 1)+1
         DO x = my_ub+1, coarse_bo(2, 1)
            p_old = pos_of_x(coarse_gbo(1, 1)+MODULO(x-coarse_gbo(1, 1), s(1)))
            CALL dcopy(coarse_slice_size, &
                       rcv_buf(sent_size(p_old)), 1, &
                       coarse_coeffs(x, coarse_bo(1, 2), &
                                     coarse_bo(1, 3)), ss)
            sent_size(p_old) = sent_size(p_old)+coarse_slice_size
         END DO
         p_old = pos_of_x(coarse_gbo(1, 1) &
                          +MODULO(my_lb-coarse_gbo(1, 1), s(1)))
         p_lb = my_lb
         DO x = my_lb, my_ub
            p = pos_of_x(coarse_gbo(1, 1)+MODULO(x-coarse_gbo(1, 1), s(1)))
            IF (p /= p_old) THEN
               p_ub = x-1

               ii = sent_size(p_old)
               DO k = coarse_bo(1, 3), coarse_bo(2, 3)
                  DO j = coarse_bo(1, 2), coarse_bo(2, 2)
                     DO i = p_lb, p_ub
                        coarse_coeffs(i, j, k) = rcv_buf(ii)
                        ii = ii+1
                     END DO
                  END DO
               END DO
               sent_size(p_old) = ii

               p_lb = x
               p_old = p
            END IF
            rcv_size(p) = rcv_size(p)+coarse_slice_size
         END DO
         p_ub = my_ub
         ii = sent_size(p_old)
         DO k = coarse_bo(1, 3), coarse_bo(2, 3)
            DO j = coarse_bo(1, 2), coarse_bo(2, 2)
               DO i = p_lb, p_ub
                  coarse_coeffs(i, j, k) = rcv_buf(ii)
                  ii = ii+1
               END DO
            END DO
         END DO
         sent_size(p_old) = ii
         DO x = coarse_bo(1, 1), my_lb-1
            p_old = pos_of_x(coarse_gbo(1, 1)+MODULO(x-coarse_gbo(1, 1), s(1)))
            CALL dcopy(coarse_slice_size, &
                       rcv_buf(sent_size(p_old)), 1, &
                       coarse_coeffs(x, coarse_bo(1, 2), &
                                     coarse_bo(1, 3)), ss)
            sent_size(p_old) = sent_size(p_old)+coarse_slice_size
         END DO

         IF(.NOT.(ALL(sent_size(0:n_procs-2) == rcv_offset(1:))))CALL cp__a("pw/pw_spline_utils.F",1362)
         IF(.NOT.(sent_size(n_procs-1) == rcv_tot_size))CALL cp__a("pw/pw_spline_utils.F",1363)

         ! dealloc
         DEALLOCATE (send_size, send_offset, rcv_size, rcv_offset)
         DEALLOCATE (send_buf, rcv_buf, real_rcv_size)
         CALL timestop(handle2)

      END IF
      fine_values => fine_values_pw%cr3d
      w_0 = (/weights_1d(3), weights_1d(1), weights_1d(3)/)
      w_1 = (/weights_1d(4), weights_1d(2), weights_1d(2), weights_1d(4)/)

      DO k = coarse_bo(1, 3), coarse_bo(2, 3)
         DO ik = -3, 3
            IF (pbc) THEN
               wk = weights_1d(ABS(ik)+1)
               fk = fine_gbo(1, 3)+MODULO(2*k+ik-fine_gbo(1, 3)+f_shift(3), 2*s(3))
            ELSE
               fk = 2*k+ik+f_shift(3)
               IF (fk <= fine_bo(1, 3)+1 .OR. fk >= fine_bo(2, 3)-1) THEN
                  IF (fk < fine_bo(1, 3) .OR. fk > fine_bo(2, 3)) CYCLE
                  IF (fk == fine_bo(1, 3) .OR. fk == fine_bo(2, 3)) THEN
                     IF (ik /= 0) CYCLE
                     wk = w_border0
                  ELSE IF (fk == 2*coarse_bo(1, 3)+1+f_shift(3)) THEN
                     SELECT CASE (ik)
                     CASE (1)
                        wk = w_border1(1)
                     CASE (-1)
                        wk = w_border1(2)
                     CASE (-3)
                        wk = w_border1(3)
                     CASE default
                        CALL cp__b("pw/pw_spline_utils.F",1396,"")
                        CYCLE
                     END SELECT
                  ELSE
                     SELECT CASE (ik)
                     CASE (3)
                        wk = w_border1(3)
                     CASE (1)
                        wk = w_border1(2)
                     CASE (-1)
                        wk = w_border1(1)
                     CASE default
                        CALL cp__b("pw/pw_spline_utils.F",1408,"")
                        CYCLE
                     END SELECT
                  END IF
               ELSE
                  wk = weights_1d(ABS(ik)+1)
               END IF
            END IF
            DO j = coarse_bo(1, 2), coarse_bo(2, 2)
               DO ij = -3, 3
                  IF (pbc) THEN
                     wj = weights_1d(ABS(ij)+1)*wk
                     fj = fine_gbo(1, 2)+MODULO(2*j+ij-fine_gbo(1, 2)+f_shift(2), 2*s(2))
                  ELSE
                     fj = 2*j+ij+f_shift(2)
                     IF (fj <= fine_bo(1, 2)+1 .OR. fj >= fine_bo(2, 2)-1) THEN
                        IF (fj < fine_bo(1, 2) .OR. fj > fine_bo(2, 2)) CYCLE
                        IF (fj == fine_bo(1, 2) .OR. fj == fine_bo(2, 2)) THEN
                           IF (ij /= 0) CYCLE
                           wj = w_border0*wk
                        ELSE IF (fj == 2*coarse_bo(1, 2)+1+f_shift(2)) THEN
                           SELECT CASE (ij)
                           CASE (1)
                              wj = w_border1(1)*wk
                           CASE (-1)
                              wj = w_border1(2)*wk
                           CASE (-3)
                              wj = w_border1(3)*wk
                           CASE default
                              CYCLE
                           END SELECT
                        ELSE
                           SELECT CASE (ij)
                           CASE (-1)
                              wj = w_border1(1)*wk
                           CASE (1)
                              wj = w_border1(2)*wk
                           CASE (3)
                              wj = w_border1(3)*wk
                           CASE default
                              CYCLE
                           END SELECT
                        END IF
                     ELSE
                        wj = weights_1d(ABS(ij)+1)*wk
                     END IF
                  END IF

                  IF (fine_bo(2, 1)-fine_bo(1, 1) < 7 .OR. safe_calc) THEN
!                      CALL timeset(routineN//"_safe",handle2)
                     DO i = coarse_bo(1, 1), coarse_bo(2, 1)
                        DO ii = -3, 3
                           IF (pbc .AND. .NOT. is_split) THEN
                              wi = weights_1d(ABS(ii)+1)*wj
                              fi = fine_gbo(1, 1)+MODULO(2*i+ii-fine_gbo(1, 1)+f_shift(1), 2*s(1))
                           ELSE
                              fi = 2*i+ii+f_shift(1)
                              IF (fi < fine_bo(1, 1) .OR. fi > fine_bo(2, 1)) CYCLE
                              IF (.NOT. pbc .AND. (fi <= fine_gbo(1, 1)+1 .OR. &
                                                   fi >= fine_gbo(2, 1)-1)) THEN
                                 IF (fi == fine_gbo(1, 1) .OR. fi == fine_gbo(2, 1)) THEN
                                    IF (ii /= 0) CYCLE
                                    wi = w_border0*wj
                                 ELSE IF (fi == fine_gbo(1, 1)+1) THEN
                                    SELECT CASE (ii)
                                    CASE (1)
                                       wi = w_border1(1)*wj
                                    CASE (-1)
                                       wi = w_border1(2)*wj
                                    CASE (-3)
                                       wi = w_border1(3)*wj
                                    CASE default
                                       CYCLE
                                    END SELECT
                                 ELSE
                                    SELECT CASE (ii)
                                    CASE (-1)
                                       wi = w_border1(1)*wj
                                    CASE (1)
                                       wi = w_border1(2)*wj
                                    CASE (3)
                                       wi = w_border1(3)*wj
                                    CASE default
                                       CYCLE
                                    END SELECT
                                 END IF
                              ELSE
                                 wi = weights_1d(ABS(ii)+1)*wj
                              END IF
                           END IF
                           fine_values(fi, fj, fk) = &
                              fine_values(fi, fj, fk)+ &
                              wi*coarse_coeffs(i, j, k)
                        END DO
                     END DO
!                      CALL timestop(handle2)
                  ELSE
!                      CALL timeset(routineN//"_core1",handle2)
                     ww0 = wj*w_0
                     ww1 = wj*w_1
                     IF (pbc .AND. .NOT. is_split) THEN
                        v3 = coarse_coeffs(coarse_bo(2, 1), j, k)
                        i = coarse_bo(1, 1)
                        fi = 2*i+f_shift(1)
                        v0 = coarse_coeffs(i, j, k)
                        v1 = coarse_coeffs(i+1, j, k)
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  ww0(1)*v3+ww0(2)*v0+ww0(3)*v1
                        v2 = coarse_coeffs(i+2, j, k)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  ww1(1)*v3+ww1(2)*v0+ww1(3)*v1+ww1(4)*v2
                     ELSE IF (.NOT. has_i_lbound) THEN
                        i = coarse_bo(1, 1)
                        fi = 2*i+f_shift(1)
                        v0 = coarse_coeffs(i, j, k)
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  w_border0*wj*v0
                        v1 = coarse_coeffs(i+1, j, k)
                        v2 = coarse_coeffs(i+2, j, k)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  wj*(w_border1(1)*v0+w_border1(2)*v1+ &
                                                      w_border1(3)*v2)
                     ELSE
                        i = coarse_bo(1, 1)
                        v0 = coarse_coeffs(i, j, k)
                        v1 = coarse_coeffs(i+1, j, k)
                        v2 = coarse_coeffs(i+2, j, k)
                        fi = 2*i+f_shift(1)+1
                        IF (.NOT. (fi+1 == fine_bo(1, 1) .OR. &
                                   fi+2 == fine_bo(1, 1))) THEN
                           CALL cp_abort(cp__l("pw/pw_spline_utils.F",1540), &
                                         "unexpected start index "// &
                                         TRIM(cp_to_string(coarse_bo(1, 1)))//" "// &
                                         TRIM(cp_to_string(fi)))
                        END IF
                     END IF
                     fi = fi+1
                     IF (fi >= fine_bo(1, 1)) THEN
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  ww0(1)*v0+ww0(2)*v1+ &
                                                  ww0(3)*v2
                     ELSE
                        IF(.NOT.(fi+1 == fine_bo(1, 1)))CALL cp__a("pw/pw_spline_utils.F",1552)
                     END IF
!                      CALL timestop(handle2)
!                      CALL timeset(routineN//"_core",handle2)
                     DO i = coarse_bo(1, 1)+3, FLOOR((fine_bo(2, 1)-f_shift(1))/2.)-3, 4
                        v3 = coarse_coeffs(i, j, k)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  (ww1(1)*v0+ww1(2)*v1+ &
                                                   ww1(3)*v2+ww1(4)*v3)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  (ww0(1)*v1+ww0(2)*v2+ &
                                                   ww0(3)*v3)
                        v0 = coarse_coeffs(i+1, j, k)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  (ww1(4)*v0+ww1(1)*v1+ &
                                                   ww1(2)*v2+ww1(3)*v3)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  (ww0(1)*v2+ww0(2)*v3+ &
                                                   ww0(3)*v0)
                        v1 = coarse_coeffs(i+2, j, k)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  (ww1(3)*v0+ww1(4)*v1+ &
                                                   ww1(1)*v2+ww1(2)*v3)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  (ww0(1)*v3+ww0(2)*v0+ &
                                                   ww0(3)*v1)
                        v2 = coarse_coeffs(i+3, j, k)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  (ww1(2)*v0+ww1(3)*v1+ &
                                                   ww1(4)*v2+ww1(1)*v3)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  (ww0(1)*v0+ww0(2)*v1+ &
                                                   ww0(3)*v2)
                     END DO
!                      CALL timestop(handle2)
!                      CALL timeset(routineN//"_clean",handle2)
                     rest_b = MODULO(FLOOR((fine_bo(2, 1)-f_shift(1))/2.)-coarse_bo(1, 1)-3+1, 4)
                     IF (rest_b > 0) THEN
                        i = FLOOR((fine_bo(2, 1)-f_shift(1))/2.)-rest_b+1
                        v3 = coarse_coeffs(i, j, k)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  (ww1(1)*v0+ww1(2)*v1+ &
                                                   ww1(3)*v2+ww1(4)*v3)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  (ww0(1)*v1+ww0(2)*v2+ &
                                                   ww0(3)*v3)
                        IF (rest_b > 1) THEN
                           v0 = coarse_coeffs(i+1, j, k)
                           fi = fi+1
                           fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                     (ww1(4)*v0+ww1(1)*v1+ &
                                                      ww1(2)*v2+ww1(3)*v3)
                           fi = fi+1
                           fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                     (ww0(1)*v2+ww0(2)*v3+ &
                                                      ww0(3)*v0)
                           IF (rest_b > 2) THEN
                              v1 = coarse_coeffs(i+2, j, k)
                              fi = fi+1
                              fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                        (ww1(3)*v0+ww1(4)*v1+ &
                                                         ww1(1)*v2+ww1(2)*v3)
                              fi = fi+1
                              fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                        (ww0(1)*v3+ww0(2)*v0+ &
                                                         ww0(3)*v1)
                              IF (pbc .AND. .NOT. is_split) THEN
                                 v2 = coarse_coeffs(coarse_bo(1, 1), j, k)
                                 fi = fi+1
                                 fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                           ww1(1)*v3+ww1(2)*v0+ww1(3)*v1+ww1(4)*v2
                                 fi = fi+1
                                 fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                           ww0(1)*v0+ww0(2)*v1+ww0(3)*v2
                                 v3 = coarse_coeffs(coarse_bo(1, 1)+1, j, k)
                                 fi = fi+1
                                 fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                           ww1(1)*v0+ww1(2)*v1+ww1(3)*v2+ww1(4)*v3
                              ELSE IF (has_i_ubound) THEN
                                 v2 = coarse_coeffs(i+3, j, k)
                                 fi = fi+1
                                 fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                           ww1(1)*v3+ww1(2)*v0+ww1(3)*v1+ww1(4)*v2
                                 fi = fi+1
                                 fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                           ww0(1)*v0+ww0(2)*v1+ww0(3)*v2
                                 IF (fi+1 == fine_bo(2, 1)) THEN
                                    v3 = coarse_coeffs(i+4, j, k)
                                    fi = fi+1
                                    fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                              ww1(1)*v0+ww1(2)*v1+ww1(3)*v2+ww1(4)*v3
                                 END IF
                              ELSE
                                 fi = fi+1
                                 fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                           wj*(w_border1(3)*v3+w_border1(2)*v0+ &
                                                               w_border1(1)*v1)
                                 fi = fi+1
                                 fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                           w_border0*wj*v1
                              END IF
                           ELSE IF (pbc .AND. .NOT. is_split) THEN
                              v1 = coarse_coeffs(coarse_bo(1, 1), j, k)
                              fi = fi+1
                              fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                        ww1(1)*v2+ww1(2)*v3+ww1(3)*v0+ww1(4)*v1
                              fi = fi+1
                              fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                        ww0(1)*v3+ww0(2)*v0+ww0(3)*v1
                              v2 = coarse_coeffs(coarse_bo(1, 1)+1, j, k)
                              fi = fi+1
                              fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                        ww1(1)*v3+ww1(2)*v0+ww1(3)*v1+ww1(4)*v2
                           ELSE IF (has_i_ubound) THEN
                              v1 = coarse_coeffs(i+2, j, k)
                              fi = fi+1
                              fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                        ww1(1)*v2+ww1(2)*v3+ww1(3)*v0+ww1(4)*v1
                              fi = fi+1
                              fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                        ww0(1)*v3+ww0(2)*v0+ww0(3)*v1
                              IF (fi+1 == fine_bo(2, 1)) THEN
                                 v2 = coarse_coeffs(i+3, j, k)
                                 fi = fi+1
                                 fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                           ww1(1)*v3+ww1(2)*v0+ww1(3)*v1+ww1(4)*v2
                              END IF
                           ELSE
                              fi = fi+1
                              fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                        wj*(w_border1(3)*v2+w_border1(2)*v3+ &
                                                            w_border1(1)*v0)
                              fi = fi+1
                              fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                        w_border0*wj*v0
                           END IF
                        ELSE IF (pbc .AND. .NOT. is_split) THEN
                           v0 = coarse_coeffs(coarse_bo(1, 1), j, k)
                           fi = fi+1
                           fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                     ww1(1)*v1+ww1(2)*v2+ww1(3)*v3+ww1(4)*v0
                           fi = fi+1
                           fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                     ww0(1)*v2+ww0(2)*v3+ww0(3)*v0
                           v1 = coarse_coeffs(coarse_bo(1, 1)+1, j, k)
                           fi = fi+1
                           fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                     ww1(1)*v2+ww1(2)*v3+ww1(3)*v0+ww1(4)*v1
                        ELSE IF (has_i_ubound) THEN
                           v0 = coarse_coeffs(i+1, j, k)
                           fi = fi+1
                           fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                     ww1(1)*v1+ww1(2)*v2+ww1(3)*v3+ww1(4)*v0
                           fi = fi+1
                           fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                     ww0(1)*v2+ww0(2)*v3+ww0(3)*v0
                           IF (fi+1 == fine_bo(2, 1)) THEN
                              v1 = coarse_coeffs(i+2, j, k)
                              fi = fi+1
                              fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                        ww1(1)*v2+ww1(2)*v3+ww1(3)*v0+ww1(4)*v1
                           END IF
                        ELSE
                           fi = fi+1
                           fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                     wj*(w_border1(3)*v1+w_border1(2)*v2+ &
                                                         w_border1(1)*v3)
                           fi = fi+1
                           fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                     w_border0*wj*v3
                        END IF
                     ELSE IF (pbc .AND. .NOT. is_split) THEN
                        v3 = coarse_coeffs(coarse_bo(1, 1), j, k)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  ww1(1)*v0+ww1(2)*v1+ww1(3)*v2+ww1(4)*v3
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  ww0(1)*v1+ww0(2)*v2+ww0(3)*v3
                        v0 = coarse_coeffs(coarse_bo(1, 1)+1, j, k)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  ww1(1)*v1+ww1(2)*v2+ww1(3)*v3+ww1(4)*v0
                     ELSE IF (has_i_ubound) THEN
                        v3 = coarse_coeffs(i, j, k)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  ww1(1)*v0+ww1(2)*v1+ww1(3)*v2+ww1(4)*v3
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  ww0(1)*v1+ww0(2)*v2+ww0(3)*v3
                        IF (fi+1 == fine_bo(2, 1)) THEN
                           v0 = coarse_coeffs(i+1, j, k)
                           fi = fi+1
                           fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                     ww1(1)*v1+ww1(2)*v2+ww1(3)*v3+ww1(4)*v0
                        END IF
                     ELSE
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  wj*(w_border1(3)*v0+w_border1(2)*v1+ &
                                                      w_border1(1)*v2)
                        fi = fi+1
                        fine_values(fi, fj, fk) = fine_values(fi, fj, fk)+ &
                                                  w_border0*wj*v2
                     END IF
                     IF(.NOT.(fi == fine_bo(2, 1)))CALL cp__a("pw/pw_spline_utils.F",1768)
                  END IF
!                   CALL timestop(handle2)
               END DO
            END DO
         END DO
      END DO

      IF (is_split) THEN
         DEALLOCATE (coarse_coeffs)
      END IF
      CALL timestop(handle)
   END SUBROUTINE add_coarse2fine

! **************************************************************************************************
!> \brief low level function that adds a coarse grid (without boundary)
!>      to a fine grid.
!>
!>      It will add to
!>
!>        coarse_coeffs(coarse_bounds(1,1):coarse_bounds(2,1),
!>                      coarse_bounds(1,2):coarse_bounds(2,2),
!>                      coarse_bounds(1,3):coarse_bounds(2,3))
!>
!>      using
!>
!>        fine_values(2*coarse_bounds(1,1):2*coarse_bounds(2,1),
!>                    2*coarse_bounds(1,2):2*coarse_bounds(2,2),
!>                    2*coarse_bounds(1,3):2*coarse_bounds(2,3))
!>
!>      composed with the weights obtained by the direct product of the
!>      1d coefficents weights:
!>
!>      for i,j,k in -3..3
!>         w(i,j,k)=weights_1d(abs(i)+1)*weights_1d(abs(j)+1)*
!>                  weights_1d(abs(k)+1)
!> \param fine_values_pw 3d array where to add the values due to the
!>        coarse coeffs
!> \param coarse_coeffs_pw 3d array with boundary of size 1 with the values of the
!>        coefficients
!> \param weights_1d the weights of the 1d smearing
!> \param w_border0 the 1d weight at the border
!> \param w_border1 the 1d weights for a point one off the border
!>        (w_border1(1) is the weight of the coefficent at the border)
!> \param pbc ...
!> \param safe_computation ...
!> \author fawzi
!> \note
!>      see coarse2fine for some relevant notes
! **************************************************************************************************
   SUBROUTINE add_fine2coarse(fine_values_pw, coarse_coeffs_pw, &
                              weights_1d, w_border0, w_border1, pbc, safe_computation)
      TYPE(pw_type), POINTER                             :: fine_values_pw, coarse_coeffs_pw
      REAL(kind=dp), DIMENSION(4), INTENT(in)            :: weights_1d
      REAL(kind=dp), INTENT(in)                          :: w_border0
      REAL(kind=dp), DIMENSION(3), INTENT(in)            :: w_border1
      LOGICAL, INTENT(in)                                :: pbc
      LOGICAL, INTENT(in), OPTIONAL                      :: safe_computation

      CHARACTER(len=*), PARAMETER :: routineN = 'add_fine2coarse', &
         routineP = moduleN//':'//routineN

      INTEGER :: coarse_slice_size, f_shift(3), fi, fj, fk, handle, handle2, i, ii, ij, ik, ip, j, &
         k, n_procs, p, p_old, rcv_tot_size, rest_b, s(3), send_tot_size, ss, x, x_att
      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: pp_lb, pp_ub, rcv_offset, rcv_size, &
                                                            real_rcv_size, send_offset, send_size, &
                                                            sent_size
      INTEGER, DIMENSION(2, 3)                           :: coarse_bo, coarse_gbo, fine_bo, &
                                                            fine_gbo, my_coarse_bo
      INTEGER, DIMENSION(:), POINTER                     :: pos_of_x
      LOGICAL                                            :: has_i_lbound, has_i_ubound, is_split, &
                                                            local_data, safe_calc
      REAL(kind=dp)                                      :: vv0, vv1, vv2, vv3, vv4, vv5, vv6, vv7, &
                                                            wi, wj, wk
      REAL(kind=dp), ALLOCATABLE, DIMENSION(:)           :: rcv_buf, send_buf
      REAL(kind=dp), DIMENSION(3)                        :: w_0, ww0
      REAL(kind=dp), DIMENSION(4)                        :: w_1, ww1
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: coarse_coeffs, fine_values

      CALL timeset(routineN, handle)

      safe_calc = .FALSE.
      IF (PRESENT(safe_computation)) safe_calc = safe_computation

      my_coarse_bo = coarse_coeffs_pw%pw_grid%bounds_local
      coarse_gbo = coarse_coeffs_pw%pw_grid%bounds
      fine_bo = fine_values_pw%pw_grid%bounds_local
      fine_gbo = fine_values_pw%pw_grid%bounds
      f_shift = fine_gbo(1, :)-2*coarse_gbo(1, :)
      is_split = ANY(coarse_gbo(:, 1) /= my_coarse_bo(:, 1))
      coarse_bo = my_coarse_bo
      IF (fine_bo(1, 1) <= fine_bo(2, 1)) THEN
         coarse_bo(1, 1) = FLOOR(REAL(fine_bo(1, 1)-f_shift(1), dp)/2._dp)-1
         coarse_bo(2, 1) = FLOOR(REAL(fine_bo(2, 1)+1-f_shift(1), dp)/2._dp)+1
      ELSE
         coarse_bo(1, 1) = coarse_gbo(2, 1)
         coarse_bo(2, 1) = coarse_gbo(2, 1)-1
      END IF
      IF (.NOT. is_split .OR. .NOT. pbc) THEN
         coarse_bo(1, 1) = MAX(coarse_gbo(1, 1), coarse_bo(1, 1))
         coarse_bo(2, 1) = MIN(coarse_gbo(2, 1), coarse_bo(2, 1))
      END IF
      has_i_ubound = (fine_gbo(2, 1) /= fine_bo(2, 1)) .OR. pbc .AND. is_split
      has_i_lbound = (fine_gbo(1, 1) /= fine_bo(1, 1)) .OR. pbc .AND. is_split

      IF (pbc) THEN
         IF(.NOT.(ALL(fine_gbo(1, :) == 2*coarse_gbo(1, :)+f_shift)))CALL cp__a("pw/pw_spline_utils.F",1874)
         IF(.NOT.(ALL(fine_gbo(2, :) == 2*coarse_gbo(2, :)+f_shift+1)))CALL cp__a("pw/pw_spline_utils.F",1875)
      ELSE
         IF(.NOT.(ALL(fine_gbo(2, :) == 2*coarse_gbo(2, :)+f_shift)))CALL cp__a("pw/pw_spline_utils.F",1877)
         IF(.NOT.(ALL(fine_gbo(1, :) == 2*coarse_gbo(1, :)+f_shift)))CALL cp__a("pw/pw_spline_utils.F",1878)
      END IF
      IF(.NOT.(coarse_gbo(2, 1)-coarse_gbo(1, 2) > 1))CALL cp__a("pw/pw_spline_utils.F",1880)
      local_data = is_split ! ANY(coarse_bo/=my_coarse_bo)
      IF (local_data) THEN
         ALLOCATE (coarse_coeffs(coarse_bo(1, 1):coarse_bo(2, 1), &
                                 coarse_bo(1, 2):coarse_bo(2, 2), &
                                 coarse_bo(1, 3):coarse_bo(2, 3)))
         coarse_coeffs = 0._dp
      ELSE
         coarse_coeffs => coarse_coeffs_pw%cr3d
      END IF

      fine_values => fine_values_pw%cr3d
      w_0 = (/weights_1d(3), weights_1d(1), weights_1d(3)/)
      w_1 = (/weights_1d(4), weights_1d(2), weights_1d(2), weights_1d(4)/)

      DO i = 1, 3
         s(i) = coarse_gbo(2, i)-coarse_gbo(1, i)+1
      END DO
      IF (ANY(s < 1)) RETURN

      DO k = coarse_bo(1, 3), coarse_bo(2, 3)
         DO ik = -3, 3
            IF (pbc) THEN
               wk = weights_1d(ABS(ik)+1)
               fk = fine_gbo(1, 3)+MODULO(2*k+ik-fine_gbo(1, 3)+f_shift(3), 2*s(3))
            ELSE
               fk = 2*k+ik+f_shift(3)
               IF (fk <= fine_bo(1, 3)+1 .OR. fk >= fine_bo(2, 3)-1) THEN
                  IF (fk < fine_bo(1, 3) .OR. fk > fine_bo(2, 3)) CYCLE
                  IF (fk == fine_bo(1, 3) .OR. fk == fine_bo(2, 3)) THEN
                     IF (ik /= 0) CYCLE
                     wk = w_border0
                  ELSE IF (fk == fine_bo(1, 3)+1) THEN
                     SELECT CASE (ik)
                     CASE (1)
                        wk = w_border1(1)
                     CASE (-1)
                        wk = w_border1(2)
                     CASE (-3)
                        wk = w_border1(3)
                     CASE default
                        CALL cp__b("pw/pw_spline_utils.F",1921,"")
                        CYCLE
                     END SELECT
                  ELSE
                     SELECT CASE (ik)
                     CASE (3)
                        wk = w_border1(3)
                     CASE (1)
                        wk = w_border1(2)
                     CASE (-1)
                        wk = w_border1(1)
                     CASE default
                        CALL cp__b("pw/pw_spline_utils.F",1933,"")
                        CYCLE
                     END SELECT
                  END IF
               ELSE
                  wk = weights_1d(ABS(ik)+1)
               END IF
            END IF
            DO j = coarse_bo(1, 2), coarse_bo(2, 2)
               DO ij = -3, 3
                  IF (pbc) THEN
                     fj = fine_gbo(1, 2)+MODULO(2*j+ij-fine_gbo(1, 2)+f_shift(2), &
                                                2*s(2))
                     wj = weights_1d(ABS(ij)+1)*wk
                  ELSE
                     fj = 2*j+ij+f_shift(2)
                     IF (fj <= fine_bo(1, 2)+1 .OR. fj >= fine_bo(2, 2)-1) THEN
                        IF (fj < fine_bo(1, 2) .OR. fj > fine_bo(2, 2)) CYCLE
                        IF (fj == fine_bo(1, 2) .OR. fj == fine_bo(2, 2)) THEN
                           IF (ij /= 0) CYCLE
                           wj = w_border0*wk
                        ELSE IF (fj == fine_bo(1, 2)+1) THEN
                           SELECT CASE (ij)
                           CASE (1)
                              wj = w_border1(1)*wk
                           CASE (-1)
                              wj = w_border1(2)*wk
                           CASE (-3)
                              wj = w_border1(3)*wk
                           CASE default
                              CALL cp__b("pw/pw_spline_utils.F",1963,"")
                              CYCLE
                           END SELECT
                        ELSE
                           SELECT CASE (ij)
                           CASE (-1)
                              wj = w_border1(1)*wk
                           CASE (1)
                              wj = w_border1(2)*wk
                           CASE (3)
                              wj = w_border1(3)*wk
                           CASE default
                              CALL cp__b("pw/pw_spline_utils.F",1975,"")
                              CYCLE
                           END SELECT
                        END IF
                     ELSE
                        wj = weights_1d(ABS(ij)+1)*wk
                     END IF
                  END IF

                  IF (coarse_bo(2, 1)-coarse_bo(1, 1) < 7 .OR. safe_calc) THEN
                     DO i = coarse_bo(1, 1), coarse_bo(2, 1)
                        DO ii = -3, 3
                           IF (pbc .AND. .NOT. is_split) THEN
                              wi = weights_1d(ABS(ii)+1)*wj
                              fi = fine_gbo(1, 1)+MODULO(2*i+ii-fine_gbo(1, 1)+f_shift(1), 2*s(1))
                           ELSE
                              fi = 2*i+ii+f_shift(1)
                              IF (fi < fine_bo(1, 1) .OR. fi > fine_bo(2, 1)) CYCLE
                              IF (((.NOT. pbc) .AND. fi <= fine_gbo(1, 1)+1) .OR. &
                                  ((.NOT. pbc) .AND. fi >= fine_gbo(2, 1)-1)) THEN
                                 IF (fi == fine_gbo(1, 1) .OR. fi == fine_gbo(2, 1)) THEN
                                    IF (ii /= 0) CYCLE
                                    wi = w_border0*wj
                                 ELSE IF (fi == fine_gbo(1, 1)+1) THEN
                                    SELECT CASE (ii)
                                    CASE (1)
                                       wi = w_border1(1)*wj
                                    CASE (-1)
                                       wi = w_border1(2)*wj
                                    CASE (-3)
                                       wi = w_border1(3)*wj
                                    CASE default
                                       CYCLE
                                    END SELECT
                                 ELSE
                                    SELECT CASE (ii)
                                    CASE (-1)
                                       wi = w_border1(1)*wj
                                    CASE (1)
                                       wi = w_border1(2)*wj
                                    CASE (3)
                                       wi = w_border1(3)*wj
                                    CASE default
                                       CYCLE
                                    END SELECT
                                 END IF
                              ELSE
                                 wi = weights_1d(ABS(ii)+1)*wj
                              END IF
                           END IF
                           coarse_coeffs(i, j, k) = &
                              coarse_coeffs(i, j, k)+ &
                              wi*fine_values(fi, fj, fk)
                        END DO
                     END DO
                  ELSE
                     ww0 = wj*w_0
                     ww1 = wj*w_1
                     IF (pbc .AND. .NOT. is_split) THEN
                        i = coarse_bo(1, 1)-1
                        vv2 = fine_values(fine_bo(2, 1)-2, fj, fk)
                        vv3 = fine_values(fine_bo(2, 1)-1, fj, fk)
                        vv4 = fine_values(fine_bo(2, 1), fj, fk)
                        fi = fine_bo(1, 1)
                        vv5 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv6 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv7 = fine_values(fi, fj, fk)
                        coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k) &
                                                   +ww1(4)*vv2+ww0(3)*vv3+ww1(3)*vv4+ww0(2)*vv5+ww1(2)*vv6+ww0(1)*vv7
                        coarse_coeffs(i+2, j, k) = coarse_coeffs(i+2, j, k) &
                                                   +ww1(4)*vv4+ww0(3)*vv5+ww1(3)*vv6+ww0(2)*vv7
                        coarse_coeffs(i+3, j, k) = coarse_coeffs(i+3, j, k) &
                                                   +ww1(4)*vv6+ww0(3)*vv7
                     ELSE IF (has_i_lbound) THEN
                        i = coarse_bo(1, 1)
                        fi = fine_bo(1, 1)-1
                        IF (i+1 == FLOOR((fine_bo(1, 1)+1-f_shift(1))/2._dp)) THEN
                           fi = fi+1
                           vv0 = fine_values(fi, fj, fk)
                           coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k)+ &
                                                    vv0*ww0(3)
                           coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k)+ &
                                                      vv0*ww0(2)
                           coarse_coeffs(i+2, j, k) = coarse_coeffs(i+2, j, k)+ &
                                                      vv0*ww0(1)
                        END IF
                     ELSE
                        i = coarse_bo(1, 1)
                        fi = 2*i+f_shift(1)
                        vv0 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv1 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv2 = fine_values(fi, fj, fk)
                        coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k)+ &
                                                 (vv0*w_border0+vv1*w_border1(1))*wj+vv2*ww0(1)
                        coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k)+ &
                                                   wj*w_border1(2)*vv1+ww0(2)*vv2
                        coarse_coeffs(i+2, j, k) = coarse_coeffs(i+2, j, k)+ &
                                                   wj*w_border1(3)*vv1+ww0(3)*vv2
                     END IF
                     DO i = coarse_bo(1, 1)+3, FLOOR((fine_bo(2, 1)-f_shift(1))/2._dp)-3, 4
                        fi = fi+1
                        vv0 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv1 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv2 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv3 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv4 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv5 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv6 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv7 = fine_values(fi, fj, fk)
                        coarse_coeffs(i-3, j, k) = coarse_coeffs(i-3, j, k) &
                                                   +ww1(1)*vv0
                        coarse_coeffs(i-2, j, k) = coarse_coeffs(i-2, j, k) &
                                                   +ww1(2)*vv0+ww0(1)*vv1+ww1(1)*vv2
                        coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                   +ww1(3)*vv0+ww0(2)*vv1+ww1(2)*vv2+ww0(1)*vv3+ww1(1)*vv4
                        coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                 +ww1(4)*vv0+ww0(3)*vv1+ww1(3)*vv2+ww0(2)*vv3+ww1(2)*vv4+ww0(1)*vv5+ww1(1)*vv6
                        coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k) &
                                                   +ww1(4)*vv2+ww0(3)*vv3+ww1(3)*vv4+ww0(2)*vv5+ww1(2)*vv6+ww0(1)*vv7
                        coarse_coeffs(i+2, j, k) = coarse_coeffs(i+2, j, k) &
                                                   +ww1(4)*vv4+ww0(3)*vv5+ww1(3)*vv6+ww0(2)*vv7
                        coarse_coeffs(i+3, j, k) = coarse_coeffs(i+3, j, k) &
                                                   +ww1(4)*vv6+ww0(3)*vv7
                     END DO
                     IF (.NOT. FLOOR((fine_bo(2, 1)-f_shift(1))/2._dp)-coarse_bo(1, 1) >= 4) THEN
                        CALL cp__b("pw/pw_spline_utils.F",2111,"FLOOR((fine_bo(2,1)-f_shift(1))/2._dp)-coarse_bo(1,1)>=4")
                     END IF
                     rest_b = MODULO(FLOOR((fine_bo(2, 1)-f_shift(1))/2._dp)-coarse_bo(1, 1)-6, 4)
                     i = FLOOR((fine_bo(2, 1)-f_shift(1))/2._dp)-3-rest_b+4
                     IF(.NOT.(fi == (i-2)*2+f_shift(1)))CALL cp__a("pw/pw_spline_utils.F",2115)
                     IF (rest_b > 0) THEN
                        fi = fi+1
                        vv0 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv1 = fine_values(fi, fj, fk)
                        coarse_coeffs(i-3, j, k) = coarse_coeffs(i-3, j, k) &
                                                   +ww1(1)*vv0
                        coarse_coeffs(i-2, j, k) = coarse_coeffs(i-2, j, k) &
                                                   +ww1(2)*vv0+ww0(1)*vv1
                        coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                   +ww1(3)*vv0+ww0(2)*vv1
                        coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                 +ww1(4)*vv0+ww0(3)*vv1
                        IF (rest_b > 1) THEN
                           fi = fi+1
                           vv2 = fine_values(fi, fj, fk)
                           fi = fi+1
                           vv3 = fine_values(fi, fj, fk)
                           coarse_coeffs(i-2, j, k) = coarse_coeffs(i-2, j, k) &
                                                      +ww1(1)*vv2
                           coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                      +ww1(2)*vv2+ww0(1)*vv3
                           coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                    +ww1(3)*vv2+ww0(2)*vv3
                           coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k) &
                                                      +ww1(4)*vv2+ww0(3)*vv3
                           IF (rest_b > 2) THEN
                              fi = fi+1
                              vv4 = fine_values(fi, fj, fk)
                              fi = fi+1
                              vv5 = fine_values(fi, fj, fk)
                              fi = fi+1
                              vv6 = fine_values(fi, fj, fk)
                              fi = fi+1
                              vv7 = fine_values(fi, fj, fk)
                              coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                         +ww1(1)*vv4
                              IF (has_i_ubound) THEN
                                 IF (coarse_bo(2, 1)-2 == FLOOR((fine_bo(2, 1)-f_shift(1))/2._dp)) THEN
                                    fi = fi+1
                                    vv0 = fine_values(fi, fj, fk)
                                    coarse_coeffs(i+4, j, k) = coarse_coeffs(i+4, j, k) &
                                                               +vv0*ww1(4)
                                 ELSE
                                    vv0 = 0._dp
                                 END IF
                                 coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                          +ww1(2)*vv4+ww0(1)*vv5+ww1(1)*vv6
                                 coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k) &
                                                            +ww1(3)*vv4+ww0(2)*vv5+ww1(2)*vv6+ww0(1)*vv7+vv0*ww1(1)
                                 coarse_coeffs(i+2, j, k) = coarse_coeffs(i+2, j, k) &
                                                            +ww1(4)*vv4+ww0(3)*vv5+ww1(3)*vv6+ww0(2)*vv7+vv0*ww1(2)
                                 coarse_coeffs(i+3, j, k) = coarse_coeffs(i+3, j, k) &
                                                            +ww1(4)*vv6+ww0(3)*vv7+vv0*ww1(3)
                              ELSEIF (pbc .AND. .NOT. is_split) THEN
                                 fi = fi+1
                                 vv0 = fine_values(fi, fj, fk)
                                 vv1 = fine_values(fine_bo(1, 1), fj, fk)
                                 vv2 = fine_values(fine_bo(1, 1)+1, fj, fk)
                                 coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                          +ww1(2)*vv4+ww0(1)*vv5+ww1(1)*vv6
                                 coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k) &
                                                            +ww1(3)*vv4+ww0(2)*vv5+ww1(2)*vv6+ww0(1)*vv7+vv0*ww1(1)
                                 coarse_coeffs(i+2, j, k) = coarse_coeffs(i+2, j, k) &
                                                            +ww1(4)*vv4+ww0(3)*vv5+ww1(3)*vv6+ww0(2)*vv7+vv0*ww1(2) &
                                                            +vv1*ww0(1)+vv2*ww1(1)
                              ELSE
                                 coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                          +ww1(2)*vv4+ww0(1)*vv5+wj*w_border1(3)*vv6
                                 coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k) &
                                                            +ww1(3)*vv4+ww0(2)*vv5+wj*w_border1(2)*vv6
                                 coarse_coeffs(i+2, j, k) = coarse_coeffs(i+2, j, k) &
                                                            +ww1(4)*vv4+ww0(3)*vv5+wj*w_border1(1)*vv6+w_border0*wj*vv7
                              END IF
                           ELSE
                              fi = fi+1
                              vv4 = fine_values(fi, fj, fk)
                              fi = fi+1
                              vv5 = fine_values(fi, fj, fk)
                              IF (has_i_ubound) THEN
                                 IF (coarse_bo(2, 1)-2 == FLOOR((fine_bo(2, 1)-f_shift(1))/2._dp)) THEN
                                    fi = fi+1
                                    vv6 = fine_values(fi, fj, fk)
                                    coarse_coeffs(i+3, j, k) = coarse_coeffs(i+3, j, k) &
                                                               +ww1(4)*vv6
                                 ELSE
                                    vv6 = 0._dp
                                 END IF
                                 coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                            +ww1(1)*vv4
                                 coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                          +ww1(2)*vv4+ww0(1)*vv5+ww1(1)*vv6
                                 coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k) &
                                                            +ww1(3)*vv4+ww0(2)*vv5+ww1(2)*vv6
                                 coarse_coeffs(i+2, j, k) = coarse_coeffs(i+2, j, k) &
                                                            +ww1(4)*vv4+ww0(3)*vv5+ww1(3)*vv6
                              ELSEIF (pbc .AND. .NOT. is_split) THEN
                                 fi = fi+1
                                 vv6 = fine_values(fi, fj, fk)
                                 vv7 = fine_values(fine_bo(1, 1), fj, fk)
                                 vv0 = fine_values(fine_bo(1, 1)+1, fj, fk)
                                 coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                            +ww1(1)*vv4
                                 coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                          +ww1(4)*vv0+ww0(3)*vv1+ww1(3)*vv2+ww0(2)*vv3+ &
                                                          ww1(2)*vv4+ww0(1)*vv5+ww1(1)*vv6
                                 coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k) &
                                                            +ww1(4)*vv2+ww0(3)*vv3+ww1(3)*vv4+ww0(2)*vv5+ww1(2)*vv6 &
                                                            +ww0(1)*vv7+ww1(1)*vv0
                              ELSE
                                 coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                            +wj*w_border1(3)*vv4
                                 coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                          +wj*w_border1(2)*vv4
                                 coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k) &
                                                            +wj*(w_border1(1)*vv4+w_border0*vv5)
                              END IF
                           END IF
                        ELSE
                           fi = fi+1
                           vv2 = fine_values(fi, fj, fk)
                           fi = fi+1
                           vv3 = fine_values(fi, fj, fk)
                           IF (has_i_ubound) THEN
                              IF (coarse_bo(2, 1)-2 == FLOOR((fine_bo(2, 1)-f_shift(1))/2._dp)) THEN
                                 fi = fi+1
                                 vv4 = fine_values(fi, fj, fk)
                                 coarse_coeffs(i+2, j, k) = coarse_coeffs(i+2, j, k) &
                                                            +ww1(4)*vv4
                              ELSE
                                 vv4 = 0._dp
                              END IF
                              coarse_coeffs(i-2, j, k) = coarse_coeffs(i-2, j, k) &
                                                         +ww1(1)*vv2
                              coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                         +ww1(2)*vv2+ww0(1)*vv3+ww1(1)*vv4
                              coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                       +ww1(3)*vv2+ww0(2)*vv3+ww1(2)*vv4
                              coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k) &
                                                         +ww1(4)*vv2+ww0(3)*vv3+ww1(3)*vv4
                           ELSEIF (pbc .AND. .NOT. is_split) THEN
                              fi = fi+1
                              vv4 = fine_values(fi, fj, fk)
                              vv5 = fine_values(fine_bo(1, 1), fj, fk)
                              vv6 = fine_values(fine_bo(1, 1)+1, fj, fk)
                              coarse_coeffs(i-2, j, k) = coarse_coeffs(i-2, j, k) &
                                                         +ww1(1)*vv2
                              coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                         +ww1(2)*vv2+ww0(1)*vv3+ww1(1)*vv4
                              coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                       +ww1(3)*vv2+ww0(2)*vv3+ww1(2)*vv4+vv5*ww0(1)+ww1(1)*vv6
                           ELSE
                              coarse_coeffs(i-2, j, k) = coarse_coeffs(i-2, j, k) &
                                                         +wj*w_border1(3)*vv2
                              coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                         +wj*w_border1(2)*vv2
                              coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                       +wj*(w_border1(1)*vv2+w_border0*vv3)
                           END IF
                        END IF
                     ELSE
                        fi = fi+1
                        vv0 = fine_values(fi, fj, fk)
                        fi = fi+1
                        vv1 = fine_values(fi, fj, fk)
                        IF (has_i_ubound) THEN
                           IF (coarse_bo(2, 1)-2 == FLOOR((fine_bo(2, 1)-f_shift(1))/2._dp)) THEN
                              fi = fi+1
                              vv2 = fine_values(fi, fj, fk)
                              coarse_coeffs(i+1, j, k) = coarse_coeffs(i+1, j, k) &
                                                         +ww1(4)*vv2
                           ELSE
                              vv2 = 0._dp
                           END IF
                           coarse_coeffs(i-3, j, k) = coarse_coeffs(i-3, j, k) &
                                                      +ww1(1)*vv0
                           coarse_coeffs(i-2, j, k) = coarse_coeffs(i-2, j, k) &
                                                      +ww1(2)*vv0+ww0(1)*vv1+ww1(1)*vv2
                           coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                      +ww1(3)*vv0+ww0(2)*vv1+ww1(2)*vv2
                           coarse_coeffs(i, j, k) = coarse_coeffs(i, j, k) &
                                                    +ww1(4)*vv0+ww0(3)*vv1+ww1(3)*vv2
                        ELSEIF (pbc .AND. .NOT. is_split) THEN
                           fi = fi+1
                           vv2 = fine_values(fi, fj, fk)
                           vv3 = fine_values(fine_bo(1, 1), fk, fk)
                           vv4 = fine_values(fine_bo(1, 1)+1, fk, fk)
                           coarse_coeffs(i-3, j, k) = coarse_coeffs(i-3, j, k) &
                                                      +ww1(1)*vv0
                           coarse_coeffs(i-2, j, k) = coarse_coeffs(i-2, j, k) &
                                                      +ww1(2)*vv0+ww0(1)*vv1+ww1(1)*vv2
                           coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                      +ww1(3)*vv0+ww0(2)*vv1+ww1(2)*vv2+ww0(1)*vv3+ww1(1)*vv4
                        ELSE
                           coarse_coeffs(i-3, j, k) = coarse_coeffs(i-3, j, k) &
                                                      +wj*w_border1(3)*vv0
                           coarse_coeffs(i-2, j, k) = coarse_coeffs(i-2, j, k) &
                                                      +wj*w_border1(2)*vv0
                           coarse_coeffs(i-1, j, k) = coarse_coeffs(i-1, j, k) &
                                                      +wj*(w_border1(1)*vv0+w_border0*vv1)
                        END IF
                     END IF
                     IF(.NOT.(fi == fine_bo(2, 1)))CALL cp__a("pw/pw_spline_utils.F",2318)
                  END IF
               END DO
            END DO
         END DO
      END DO

      ! *** parallel case
      IF (is_split) THEN
         CALL timeset(routineN//"_comm", handle2)
         coarse_slice_size = (coarse_bo(2, 2)-coarse_bo(1, 2)+1)* &
                             (coarse_bo(2, 3)-coarse_bo(1, 3)+1)
         n_procs = coarse_coeffs_pw%pw_grid%para%group_size
         ALLOCATE (send_size(0:n_procs-1), send_offset(0:n_procs-1), &
                   sent_size(0:n_procs-1), rcv_size(0:n_procs-1), &
                   rcv_offset(0:n_procs-1), pp_lb(0:n_procs-1), &
                   pp_ub(0:n_procs-1), real_rcv_size(0:n_procs-1))

         ! ** send size count

         pos_of_x => coarse_coeffs_pw%pw_grid%para%pos_of_x
         send_size = 0
         DO x = coarse_bo(1, 1), coarse_bo(2, 1)
            p = pos_of_x(coarse_gbo(1, 1)+MODULO(x-coarse_gbo(1, 1), s(1)))
            send_size(p) = send_size(p)+coarse_slice_size
         END DO

         ! ** rcv size count

         pos_of_x => fine_values_pw%pw_grid%para%pos_of_x
         p_old = pos_of_x(fine_gbo(1, 1))
         pp_lb = fine_gbo(2, 1)
         pp_ub = fine_gbo(2, 1)-1
         pp_lb(p_old) = fine_gbo(1, 1)
         DO x = fine_gbo(1, 1), fine_gbo(2, 1)
            p = pos_of_x(x)
            IF (p /= p_old) THEN
               pp_ub(p_old) = x-1
               pp_lb(p) = x
               p_old = p
            END IF
         END DO
         pp_ub(p_old) = fine_gbo(2, 1)

         DO ip = 0, n_procs-1
            IF (pp_lb(ip) <= pp_ub(ip)) THEN
               pp_lb(ip) = FLOOR(REAL(pp_lb(ip)-f_shift(1), dp)/2._dp)-1
               pp_ub(ip) = FLOOR(REAL(pp_ub(ip)+1-f_shift(1), dp)/2._dp)+1
            ELSE
               pp_lb(ip) = coarse_gbo(2, 1)
               pp_ub(ip) = coarse_gbo(2, 1)-1
            END IF
            IF (.NOT. is_split .OR. .NOT. pbc) THEN
               pp_lb(ip) = MAX(pp_lb(ip), coarse_gbo(1, 1))
               pp_ub(ip) = MIN(pp_ub(ip), coarse_gbo(2, 1))
            END IF
         END DO

         rcv_size = 0
         DO ip = 0, n_procs-1
            DO x = pp_lb(ip), coarse_gbo(1, 1)-1
               x_att = coarse_gbo(1, 1)+MODULO(x-coarse_gbo(1, 1), s(1))
               IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                  rcv_size(ip) = rcv_size(ip)+coarse_slice_size
               END IF
            END DO
            rcv_size(ip) = rcv_size(ip)+coarse_slice_size* &
                           MAX(0, &
                               MIN(pp_ub(ip), my_coarse_bo(2, 1))-MAX(pp_lb(ip), my_coarse_bo(1, 1))+1)
            DO x = coarse_gbo(2, 1)+1, pp_ub(ip)
               x_att = coarse_gbo(1, 1)+MODULO(x-coarse_gbo(1, 1), s(1))
               IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                  rcv_size(ip) = rcv_size(ip)+coarse_slice_size
               END IF
            END DO
         END DO

         ! ** offsets & alloc send-rcv

         send_tot_size = 0
         DO ip = 0, n_procs-1
            send_offset(ip) = send_tot_size
            send_tot_size = send_tot_size+send_size(ip)
         END DO
         IF (send_tot_size /= (coarse_bo(2, 1)-coarse_bo(1, 1)+1)*coarse_slice_size) &
            CALL cp__b("pw/pw_spline_utils.F",2403,"Error calculating send_tot_size")
         ALLOCATE (send_buf(0:send_tot_size-1))

         rcv_tot_size = 0
         DO ip = 0, n_procs-1
            rcv_offset(ip) = rcv_tot_size
            rcv_tot_size = rcv_tot_size+rcv_size(ip)
         END DO
         ALLOCATE (rcv_buf(0:rcv_tot_size-1))

         ! ** fill send buffer

         pos_of_x => coarse_coeffs_pw%pw_grid%para%pos_of_x
         p_old = pos_of_x(coarse_gbo(1, 1) &
                          +MODULO(coarse_bo(1, 1)-coarse_gbo(1, 1), s(1)))
         sent_size(:) = send_offset
         ss = coarse_bo(2, 1)-coarse_bo(1, 1)+1
         DO x = coarse_bo(1, 1), coarse_bo(2, 1)
            p = pos_of_x(coarse_gbo(1, 1)+MODULO(x-coarse_gbo(1, 1), s(1)))
            CALL dcopy(coarse_slice_size, &
                       coarse_coeffs(x, coarse_bo(1, 2), &
                                     coarse_bo(1, 3)), ss, send_buf(sent_size(p)), 1)
            sent_size(p) = sent_size(p)+coarse_slice_size
         END DO

         IF (ANY(sent_size(0:n_procs-2) /= send_offset(1:n_procs-1))) &
            CALL cp__b("pw/pw_spline_utils.F",2429,"error 1 filling send buffer")
         IF (sent_size(n_procs-1) /= send_tot_size) &
            CALL cp__b("pw/pw_spline_utils.F",2431,"error 2 filling send buffer")

         IF (local_data) THEN
            DEALLOCATE (coarse_coeffs)
         ELSE
            NULLIFY (coarse_coeffs)
         END IF

         IF(.NOT.(ALL(sent_size(:n_procs-2) == send_offset(1:))))CALL cp__a("pw/pw_spline_utils.F",2439)
         IF(.NOT.(sent_size(n_procs-1) == send_tot_size))CALL cp__a("pw/pw_spline_utils.F",2440)
         ! test send/rcv sizes
         CALL mp_alltoall(send_size, real_rcv_size, 1, coarse_coeffs_pw%pw_grid%para%group)

         IF(.NOT.(ALL(real_rcv_size == rcv_size)))CALL cp__a("pw/pw_spline_utils.F",2444)
         ! all2all
         CALL mp_alltoall(sb=send_buf, scount=send_size, sdispl=send_offset, &
                          rb=rcv_buf, rcount=rcv_size, rdispl=rcv_offset, &
                          group=coarse_coeffs_pw%pw_grid%para%group)

         ! ** sum & reorder rcv buffer

         sent_size(:) = rcv_offset
         DO ip = 0, n_procs-1

            DO x = pp_lb(ip), coarse_gbo(1, 1)-1
               x_att = coarse_gbo(1, 1)+MODULO(x-coarse_gbo(1, 1), s(1))
               IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                  ii = sent_size(ip)
                  DO k = coarse_bo(1, 3), coarse_bo(2, 3)
                     DO j = coarse_bo(1, 2), coarse_bo(2, 2)
                        coarse_coeffs_pw%cr3d(x_att, j, k) = coarse_coeffs_pw%cr3d(x_att, j, k)+rcv_buf(ii)
                        ii = ii+1
                     END DO
                  END DO
                  sent_size(ip) = ii
               END IF
            END DO

            ii = sent_size(ip)
            DO x_att = MAX(pp_lb(ip), my_coarse_bo(1, 1)), MIN(pp_ub(ip), my_coarse_bo(2, 1))
               DO k = coarse_bo(1, 3), coarse_bo(2, 3)
                  DO j = coarse_bo(1, 2), coarse_bo(2, 2)
                     coarse_coeffs_pw%cr3d(x_att, j, k) = coarse_coeffs_pw%cr3d(x_att, j, k)+rcv_buf(ii)
                     ii = ii+1
                  END DO
               END DO
            END DO
            sent_size(ip) = ii

            DO x = coarse_gbo(2, 1)+1, pp_ub(ip)
               x_att = coarse_gbo(1, 1)+MODULO(x-coarse_gbo(1, 1), s(1))
               IF (x_att >= my_coarse_bo(1, 1) .AND. x_att <= my_coarse_bo(2, 1)) THEN
                  ii = sent_size(ip)
                  DO k = coarse_bo(1, 3), coarse_bo(2, 3)
                     DO j = coarse_bo(1, 2), coarse_bo(2, 2)
                        coarse_coeffs_pw%cr3d(x_att, j, k) = coarse_coeffs_pw%cr3d(x_att, j, k)+rcv_buf(ii)
                        ii = ii+1
                     END DO
                  END DO
                  sent_size(ip) = ii
               END IF
            END DO

         END DO

         IF (ANY(sent_size(0:n_procs-2) /= rcv_offset(1:n_procs-1))) &
            CALL cp__b("pw/pw_spline_utils.F",2497,"error 1 handling the rcv buffer")
         IF (sent_size(n_procs-1) /= rcv_tot_size) &
            CALL cp__b("pw/pw_spline_utils.F",2499,"error 2 handling the rcv buffer")

         ! dealloc
         DEALLOCATE (send_size, send_offset, rcv_size, rcv_offset)
         DEALLOCATE (send_buf, rcv_buf, real_rcv_size)
         DEALLOCATE (pp_ub, pp_lb)
         CALL timestop(handle2)
      ELSE
         IF(.NOT.(.NOT. local_data))CALL cp__a("pw/pw_spline_utils.F",2507)
      END IF

      CALL timestop(handle)
   END SUBROUTINE add_fine2coarse

! **************************************************************************************************
!> \brief ...
!> \param preconditioner the preconditioner to create
!> \param precond_kind the kind of preconditioner to use
!> \param pool a pool with grids of the same type as the elements to
!>        precondition
!> \param pbc if periodic boundary conditions should be applied
!> \param transpose ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_spline_precond_create(preconditioner, precond_kind, &
                                       pool, pbc, transpose)
      TYPE(pw_spline_precond_type), POINTER              :: preconditioner
      INTEGER, INTENT(in)                                :: precond_kind
      TYPE(pw_pool_type), POINTER                        :: pool
      LOGICAL, INTENT(in)                                :: pbc, transpose

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_spline_precond_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (preconditioner)
      last_precond_id = last_precond_id+1
      preconditioner%id_nr = last_precond_id
      preconditioner%ref_count = 1
      preconditioner%kind = no_precond
      preconditioner%pool => pool
      preconditioner%pbc = pbc
      preconditioner%transpose = transpose
      CALL pw_pool_retain(pool)
      CALL pw_spline_precond_set_kind(preconditioner, precond_kind)
   END SUBROUTINE pw_spline_precond_create

! **************************************************************************************************
!> \brief switches the types of precoditioner to use
!> \param preconditioner the preconditioner to be changed
!> \param precond_kind the new kind of preconditioner to use
!> \param pbc ...
!> \param transpose ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_spline_precond_set_kind(preconditioner, precond_kind, pbc, &
                                         transpose)
      TYPE(pw_spline_precond_type), POINTER              :: preconditioner
      INTEGER, INTENT(in)                                :: precond_kind
      LOGICAL, INTENT(in), OPTIONAL                      :: pbc, transpose

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_spline_precond_set_kind', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: do_3d_coeff
      REAL(kind=dp)                                      :: s

      IF(.NOT.(ASSOCIATED(preconditioner)))CALL cp__a("pw/pw_spline_utils.F",2565)
      IF(.NOT.(preconditioner%ref_count > 0))CALL cp__a("pw/pw_spline_utils.F",2566)
      IF (PRESENT(transpose)) preconditioner%transpose = transpose
      do_3d_coeff = .FALSE.
      preconditioner%kind = precond_kind
      IF (PRESENT(pbc)) preconditioner%pbc = pbc
      SELECT CASE (precond_kind)
      CASE (no_precond)
      CASE (precond_spl3_aint2)
         preconditioner%coeffs_1d = (/-1.66_dp*0.25_dp, 1.66_dp, -1.66_dp*0.25_dp/)
         preconditioner%sharpen = .FALSE.
         preconditioner%normalize = .FALSE.
         do_3d_coeff = .TRUE.
      CASE (precond_spl3_3)
         preconditioner%coeffs_1d(1) = -0.25_dp*1.6_dp
         preconditioner%coeffs_1d(2) = 1.6_dp
         preconditioner%coeffs_1d(3) = -0.25_dp*1.6_dp
         preconditioner%sharpen = .FALSE.
         preconditioner%normalize = .FALSE.
         do_3d_coeff = .TRUE.
      CASE (precond_spl3_2)
         preconditioner%coeffs_1d(1) = -0.26_dp*1.76_dp
         preconditioner%coeffs_1d(2) = 1.76_dp
         preconditioner%coeffs_1d(3) = -0.26_dp*1.76_dp
         preconditioner%sharpen = .FALSE.
         preconditioner%normalize = .FALSE.
         do_3d_coeff = .TRUE.
      CASE (precond_spl3_aint)
         preconditioner%coeffs_1d = spl3_1d_coeffs0
         preconditioner%sharpen = .TRUE.
         preconditioner%normalize = .TRUE.
         do_3d_coeff = .TRUE.
      CASE (precond_spl3_1)
         preconditioner%coeffs_1d(1) = 0.5_dp/3._dp**(1._dp/3._dp)
         preconditioner%coeffs_1d(2) = 4._dp/3._dp**(1._dp/3._dp)
         preconditioner%coeffs_1d(3) = 0.5_dp/3._dp**(1._dp/3._dp)
         preconditioner%sharpen = .TRUE.
         preconditioner%normalize = .FALSE.
         do_3d_coeff = .TRUE.
      CASE default
         CALL cp__b("pw/pw_spline_utils.F",2605,"")
      END SELECT
      IF (do_3d_coeff) THEN
         s = 1._dp
         IF (preconditioner%sharpen) s = -1._dp
         preconditioner%coeffs(1) = &
            s*preconditioner%coeffs_1d(2)* &
            preconditioner%coeffs_1d(2)* &
            preconditioner%coeffs_1d(2)
         preconditioner%coeffs(2) = &
            s*preconditioner%coeffs_1d(1)* &
            preconditioner%coeffs_1d(2)* &
            preconditioner%coeffs_1d(2)
         preconditioner%coeffs(3) = &
            s*preconditioner%coeffs_1d(1)* &
            preconditioner%coeffs_1d(1)* &
            preconditioner%coeffs_1d(2)
         preconditioner%coeffs(4) = &
            s*preconditioner%coeffs_1d(1)* &
            preconditioner%coeffs_1d(1)* &
            preconditioner%coeffs_1d(1)
         IF (preconditioner%sharpen) THEN
            IF (preconditioner%normalize) THEN
               preconditioner%coeffs(1) = 2._dp+ &
                                          preconditioner%coeffs(1)
            ELSE
               preconditioner%coeffs(1) = -preconditioner%coeffs(1)
            END IF
         END IF
      END IF
   END SUBROUTINE pw_spline_precond_set_kind

! **************************************************************************************************
!> \brief retains the preconditioner
!> \param preconditioner the preconditioner to retain
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_spline_precond_retain(preconditioner)
      TYPE(pw_spline_precond_type), POINTER              :: preconditioner

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_spline_precond_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(preconditioner)))CALL cp__a("pw/pw_spline_utils.F",2648)
      IF(.NOT.(preconditioner%ref_count > 1))CALL cp__a("pw/pw_spline_utils.F",2649)
      preconditioner%ref_count = preconditioner%ref_count+1
   END SUBROUTINE pw_spline_precond_retain

! **************************************************************************************************
!> \brief releases the preconditioner
!> \param preconditioner the preconditioner to release
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_spline_precond_release(preconditioner)
      TYPE(pw_spline_precond_type), POINTER              :: preconditioner

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_spline_precond_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(preconditioner)) THEN
         IF(.NOT.(preconditioner%ref_count > 0))CALL cp__a("pw/pw_spline_utils.F",2665)
         preconditioner%ref_count = preconditioner%ref_count-1
         IF (preconditioner%ref_count == 0) THEN
            CALL pw_pool_release(preconditioner%pool)
            DEALLOCATE (preconditioner)
         END IF
      END IF
   END SUBROUTINE pw_spline_precond_release

! **************************************************************************************************
!> \brief applies the preconditioner to the system of equations to find the
!>      coefficents of the spline
!> \param preconditioner the preconditioner to apply
!> \param in_v the grid on which the preconditioner should be applied
!> \param out_v place to store the preconditioner applied on v_out
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_spline_do_precond(preconditioner, in_v, out_v)
      TYPE(pw_spline_precond_type), POINTER              :: preconditioner
      TYPE(pw_type), POINTER                             :: in_v, out_v

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_spline_do_precond', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(preconditioner)))CALL cp__a("pw/pw_spline_utils.F",2689)
      IF(.NOT.(preconditioner%ref_count > 0))CALL cp__a("pw/pw_spline_utils.F",2690)
      SELECT CASE (preconditioner%kind)
      CASE (no_precond)
         CALL pw_copy(in_v, out_v)
      CASE (precond_spl3_aint, precond_spl3_1)
         CALL pw_zero(out_v)
         IF (preconditioner%pbc) THEN
            CALL pw_nn_smear_r(pw_in=in_v, pw_out=out_v, &
                               coeffs=preconditioner%coeffs)
         ELSE
            CALL pw_nn_compose_r_no_pbc(weights_1d=preconditioner%coeffs_1d, &
                                        pw_in=in_v, pw_out=out_v, sharpen=preconditioner%sharpen, &
                                        normalize=preconditioner%normalize, &
                                        transpose=preconditioner%transpose)
         END IF
      CASE (precond_spl3_3, precond_spl3_2, precond_spl3_aint2)
         CALL pw_zero(out_v)
         IF (preconditioner%pbc) THEN
            CALL pw_nn_smear_r(pw_in=in_v, pw_out=out_v, &
                               coeffs=preconditioner%coeffs)
         ELSE
            CALL pw_nn_compose_r_no_pbc(weights_1d=preconditioner%coeffs_1d, &
                                        pw_in=in_v, pw_out=out_v, sharpen=preconditioner%sharpen, &
                                        normalize=preconditioner%normalize, smooth_boundary=.TRUE., &
                                        transpose=preconditioner%transpose)
         END IF
      CASE default
         CALL cp__b("pw/pw_spline_utils.F",2717,"")
      END SELECT
   END SUBROUTINE pw_spline_do_precond

! **************************************************************************************************
!> \brief solves iteratively (CG) a systmes of linear equations
!>           linOp(coeffs)=values
!>      (for example those needed to find the coefficents of a spline)
!>      Returns true if the it succeded to acheive the requested accuracy
!> \param values the right hand side of the system
!> \param coeffs will contain the solution of the system (and on entry
!>        it contains the starting point)
!> \param linOp the linear operator to be inverted
!> \param preconditioner the preconditioner to apply
!> \param pool a pool of grids (for the temporary objects)
!> \param eps_r the requested precision on the residual
!> \param eps_x the requested precision on the solution
!> \param max_iter maximum number of iteration allowed
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION find_coeffs(values, coeffs, linOp, preconditioner, pool, &
                        eps_r, eps_x, max_iter) RESULT(res)
      TYPE(pw_type), POINTER                             :: values, coeffs
      INTERFACE
! **************************************************************************************************
         SUBROUTINE linOp(pw_in, pw_out)
         USE pw_types, ONLY: pw_type
            TYPE(pw_type), POINTER :: pw_in, pw_out
         END SUBROUTINE linOp
      END INTERFACE
      TYPE(pw_spline_precond_type), POINTER              :: preconditioner
      TYPE(pw_pool_type), POINTER                        :: pool
      REAL(kind=dp), INTENT(in)                          :: eps_r, eps_x
      INTEGER, INTENT(in)                                :: max_iter
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'find_coeffs', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iiter, iter, j, k
      INTEGER, DIMENSION(2, 3)                           :: bo
      LOGICAL                                            :: last
      REAL(kind=dp)                                      :: alpha, beta, eps_r_att, eps_x_att, r_z, &
                                                            r_z_new
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(pw_type), POINTER                             :: Ap, p, r, z

      last = .FALSE.

      res = .FALSE.
      logger => cp_get_default_logger()
      CALL pw_pool_create_pw(pool, r, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pool, z, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pool, p, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pool, Ap, use_data=REALDATA3D, in_space=REALSPACE)

      !CALL cp_add_iter_level(logger%iter_info,level_name="SPLINE_FIND_COEFFS")
      ext_do: DO iiter = 1, max_iter, 10
         CALL pw_zero(r)
         CALL linOp(pw_in=coeffs, pw_out=r)
         r%cr3d = -r%cr3d
         CALL pw_axpy(values, r)
         CALL pw_spline_do_precond(preconditioner, in_v=r, out_v=z)
         CALL pw_copy(z, p)
         r_z = pw_integral_ab(r, z)

         DO iter = iiter, MIN(iiter+9, max_iter)
            eps_r_att = SQRT(pw_integral_ab(r, r))
            IF (eps_r_att == 0._dp) THEN
               eps_x_att = 0._dp
               last = .TRUE.
            ELSE
               CALL pw_zero(Ap)
               CALL linOp(pw_in=p, pw_out=Ap)
               alpha = r_z/pw_integral_ab(Ap, p)

               CALL pw_axpy(p, coeffs, alpha=alpha)

               eps_x_att = alpha*SQRT(pw_integral_ab(p, p)) ! try to spare if unneded?
               IF (eps_r_att < eps_r .AND. eps_x_att < eps_x) last = .TRUE.
            END IF
            !CALL cp_iterate(logger%iter_info,last=last)
            IF (last) THEN
               res = .TRUE.
               EXIT ext_do
            END IF

            CALL pw_axpy(Ap, r, alpha=-alpha)

            CALL pw_spline_do_precond(preconditioner, in_v=r, out_v=z)

            r_z_new = pw_integral_ab(r, z)
            beta = r_z_new/r_z
            r_z = r_z_new

            bo = p%pw_grid%bounds_local
            DO k = bo(1, 3), bo(2, 3)
               DO j = bo(1, 2), bo(2, 2)
                  DO i = bo(1, 1), bo(2, 1)
                     p%cr3d(i, j, k) = z%cr3d(i, j, k)+beta*p%cr3d(i, j, k)
                  END DO
               END DO
            END DO

         END DO
      END DO ext_do
      !CALL cp_rm_iter_level(logger%iter_info,level_name="SPLINE_FIND_COEFFS")

      CALL pw_pool_give_back_pw(pool, r)
      CALL pw_pool_give_back_pw(pool, z)
      CALL pw_pool_give_back_pw(pool, p)
      CALL pw_pool_give_back_pw(pool, Ap)
   END FUNCTION find_coeffs

! **************************************************************************************************
!> \brief adds to pw_out pw_in composed with the weights
!>      pw_out%cr3d(i,j,k)=pw_out%cr3d(i,j,k)+sum(pw_in%cr3d(i+l,j+m,k+n)*
!>         weights_1d(abs(l)+1)*weights_1d(abs(m)+1)*weights_1d(abs(n)+1),
!>         l=-1..1,m=-1..1,n=-1..1)
!> \param weights_1d ...
!> \param pw_in ...
!> \param pw_out ...
!> \param sharpen ...
!> \param normalize ...
!> \param transpose ...
!> \param smooth_boundary ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE pw_nn_compose_r_no_pbc(weights_1d, pw_in, pw_out, &
                                     sharpen, normalize, transpose, smooth_boundary)
      REAL(kind=dp), DIMENSION(-1:1)                     :: weights_1d
      TYPE(pw_type), POINTER                             :: pw_in, pw_out
      LOGICAL, INTENT(in), OPTIONAL                      :: sharpen, normalize, transpose, &
                                                            smooth_boundary

      CHARACTER(len=*), PARAMETER :: routineN = 'pw_nn_compose_r_no_pbc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: first_index, i, j, jw, k, kw, &
                                                            last_index, myj, myk, n_els
      INTEGER, DIMENSION(2, 3)                           :: bo, gbo
      INTEGER, DIMENSION(3)                              :: s
      LOGICAL                                            :: has_l_boundary, has_u_boundary, &
                                                            is_split, my_normalize, my_sharpen, &
                                                            my_smooth_boundary, my_transpose
      REAL(kind=dp)                                      :: in_val_f, in_val_l, in_val_tmp, w_j, w_k
      REAL(kind=dp), DIMENSION(-1:1)                     :: w
      REAL(kind=dp), DIMENSION(:, :), POINTER            :: l_boundary, tmp, u_boundary
      REAL(kind=dp), DIMENSION(:, :, :), POINTER         :: in_val, out_val

      bo = pw_in%pw_grid%bounds_local
      gbo = pw_in%pw_grid%bounds
      in_val => pw_in%cr3d
      out_val => pw_out%cr3d
      my_sharpen = .FALSE.
      IF (PRESENT(sharpen)) my_sharpen = sharpen
      my_normalize = .FALSE.
      IF (PRESENT(normalize)) my_normalize = normalize
      my_transpose = .FALSE.
      IF (PRESENT(transpose)) my_transpose = transpose
      my_smooth_boundary = .FALSE.
      IF (PRESENT(smooth_boundary)) my_smooth_boundary = smooth_boundary
      IF(.NOT.(.NOT. my_normalize .OR. my_sharpen))CALL cp__a("pw/pw_spline_utils.F",2879)
      IF(.NOT.(.NOT. my_smooth_boundary .OR. .NOT. my_sharpen))CALL cp__a("pw/pw_spline_utils.F",2880)
      DO i = 1, 3
         s(i) = bo(2, i)-bo(1, i)+1
      END DO
      IF (ANY(s < 1)) RETURN
      is_split = ANY(pw_in%pw_grid%bounds_local(:, 1) /= &
                     pw_in%pw_grid%bounds(:, 1))
      has_l_boundary = (gbo(1, 1) == bo(1, 1))
      has_u_boundary = (gbo(2, 1) == bo(2, 1))
      IF (is_split) THEN
         ALLOCATE (l_boundary(bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)), &
                   u_boundary(bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)), &
                   tmp(bo(1, 2):bo(2, 2), bo(1, 3):bo(2, 3)))
         tmp(:, :) = pw_in%cr3d(bo(2, 1), :, :)
         CALL mp_sendrecv(tmp, pw_in%pw_grid%para%pos_of_x( &
                          gbo(1, 1)+MODULO(bo(2, 1)+1-gbo(1, 1), gbo(2, 1)-gbo(1, 1)+1)), &
                          l_boundary, pw_in%pw_grid%para%pos_of_x( &
                          gbo(1, 1)+MODULO(bo(1, 1)-1-gbo(1, 1), gbo(2, 1)-gbo(1, 1)+1)), &
                          pw_in%pw_grid%para%group)
         tmp(:, :) = pw_in%cr3d(bo(1, 1), :, :)
         CALL mp_sendrecv(tmp, pw_in%pw_grid%para%pos_of_x( &
                          gbo(1, 1)+MODULO(bo(1, 1)-1-gbo(1, 1), gbo(2, 1)-gbo(1, 1)+1)), &
                          u_boundary, pw_in%pw_grid%para%pos_of_x( &
                          gbo(1, 1)+MODULO(bo(2, 1)+1-gbo(1, 1), gbo(2, 1)-gbo(1, 1)+1)), &
                          pw_in%pw_grid%para%group)
         DEALLOCATE (tmp)
      END IF

      n_els = s(1)
      IF (has_l_boundary) THEN
         n_els = n_els-1
         first_index = bo(1, 1)+1
      ELSE
         first_index = bo(1, 1)
      END IF
      IF (has_u_boundary) THEN
         n_els = n_els-1
         last_index = bo(2, 1)-1
      ELSE
         last_index = bo(2, 1)
      END IF
      !#omp parallel do default(none) private(k,kw,myk,j,jw,myj,in_val_f,&
      !#omp     in_val_l) shared(zderiv,yderiv,bo,in_val,out_val,s,l_boundary,&
      !#omp     u_boundary,weights,is_split)
      DO k = bo(1, 3), bo(2, 3)
         DO kw = -1, 1
            myk = k+kw
            IF (my_transpose) THEN
               IF (k >= gbo(2, 3)-1 .OR. k <= gbo(1, 3)+1) THEN
                  IF (k == gbo(2, 3) .OR. k == gbo(1, 3)) THEN
                     IF (myk < gbo(2, 3) .AND. myk > gbo(1, 3)) THEN
                        w_k = weights_1d(kw)
                        IF (my_smooth_boundary) THEN
                           w_k = weights_1d(kw)/weights_1d(0)
                        END IF
                     ELSE IF (kw == 0) THEN
                        w_k = 1._dp
                     ELSE
                        CYCLE
                     END IF
                  ELSE
                     IF (myk == gbo(2, 3) .OR. myk == gbo(1, 3)) CYCLE
                     w_k = weights_1d(kw)
                  END IF
               ELSE
                  w_k = weights_1d(kw)
               END IF
            ELSE
               IF (k >= gbo(2, 3)-1 .OR. k <= gbo(1, 3)+1) THEN
                  IF (k == gbo(2, 3) .OR. k == gbo(1, 3)) THEN
                     IF (kw /= 0) CYCLE
                     w_k = 1._dp
                  ELSE
                     IF (my_smooth_boundary .AND. ((k == gbo(1, 3)+1 .AND. myk == gbo(1, 3)) .OR. &
                                                   (k == gbo(2, 3)-1 .AND. myk == gbo(2, 3)))) THEN
                        w_k = weights_1d(kw)/weights_1d(0)
                     ELSE
                        w_k = weights_1d(kw)
                     END IF
                  END IF
               ELSE
                  w_k = weights_1d(kw)
               END IF
            END IF
            DO j = bo(1, 2), bo(2, 2)
               DO jw = -1, 1
                  myj = j+jw
                  IF (j < gbo(2, 2)-1 .AND. j > gbo(1, 2)+1) THEN
                     w_j = w_k*weights_1d(jw)
                  ELSE
                     IF (my_transpose) THEN
                        IF (j == gbo(2, 2) .OR. j == gbo(1, 2)) THEN
                           IF (myj < gbo(2, 2) .AND. myj > gbo(1, 2)) THEN
                              w_j = weights_1d(jw)*w_k
                              IF (my_smooth_boundary) THEN
                                 w_j = weights_1d(jw)/weights_1d(0)*w_k
                              END IF
                           ELSE IF (jw == 0) THEN
                              w_j = w_k
                           ELSE
                              CYCLE
                           END IF
                        ELSE
                           IF (myj == gbo(2, 2) .OR. myj == gbo(1, 2)) CYCLE
                           w_j = w_k*weights_1d(jw)
                        END IF
                     ELSE
                        IF (j == gbo(2, 2) .OR. j == gbo(1, 2)) THEN
                           IF (jw /= 0) CYCLE
                           w_j = w_k
                        ELSE IF (my_smooth_boundary .AND. ((j == gbo(1, 2)+1 .AND. myj == gbo(1, 2)) .OR. &
                                                           (j == gbo(2, 2)-1 .AND. myj == gbo(2, 2)))) THEN
                           w_j = w_k*weights_1d(jw)/weights_1d(0)
                        ELSE
                           w_j = w_k*weights_1d(jw)
                        END IF
                     END IF
                  END IF

                  IF (has_l_boundary) THEN
                     IF (my_transpose) THEN
                        IF (s(1) == 1) THEN
                           IF(.NOT.(.NOT. has_u_boundary))CALL cp__a("pw/pw_spline_utils.F",3002)
                           in_val_tmp = u_boundary(myj, myk)
                        ELSE
                           in_val_tmp = in_val(bo(1, 1)+1, myj, myk)
                        END IF
                        IF (my_sharpen) THEN
                           IF (kw == 0 .AND. jw == 0) THEN
                              IF (my_normalize) THEN
                                 out_val(bo(1, 1), j, k) = out_val(bo(1, 1), j, k)+ &
                                                           (2.0_dp-w_j)*in_val(bo(1, 1), myj, myk)- &
                                                           in_val_tmp*weights_1d(1)*w_j
                              ELSE
                                 out_val(bo(1, 1), j, k) = out_val(bo(1, 1), j, k)+ &
                                                           in_val(bo(1, 1), myj, myk)*w_j- &
                                                           in_val_tmp*weights_1d(1)*w_j
                              END IF
                           ELSE
                              out_val(bo(1, 1), j, k) = out_val(bo(1, 1), j, k)- &
                                                        in_val(bo(1, 1), myj, myk)*w_j- &
                                                        in_val_tmp*weights_1d(1)*w_j
                           END IF
                        ELSE IF (my_smooth_boundary) THEN
                           out_val(bo(1, 1), j, k) = out_val(bo(1, 1), j, k)+ &
                                                     w_j*(in_val(bo(1, 1), myj, myk)+ &
                                                          in_val_tmp*weights_1d(1)/weights_1d(0))
                        ELSE
                           out_val(bo(1, 1), j, k) = out_val(bo(1, 1), j, k)+ &
                                                     w_j*(in_val(bo(1, 1), myj, myk)+ &
                                                          in_val_tmp*weights_1d(1))
                        END IF
                        in_val_f = 0.0_dp
                     ELSE
                        in_val_f = in_val(bo(1, 1), myj, myk)
                        IF (my_sharpen) THEN
                           IF (kw == 0 .AND. jw == 0) THEN
                              IF (my_normalize) THEN
                                 out_val(bo(1, 1), j, k) = out_val(bo(1, 1), j, k)+ &
                                                           (2.0_dp-w_j)*in_val_f
                              ELSE
                                 out_val(bo(1, 1), j, k) = out_val(bo(1, 1), j, k)+ &
                                                           in_val_f*w_j
                              END IF
                           ELSE
                              out_val(bo(1, 1), j, k) = out_val(bo(1, 1), j, k)- &
                                                        in_val_f*w_j
                           END IF
                        ELSE
                           out_val(bo(1, 1), j, k) = out_val(bo(1, 1), j, k)+ &
                                                     in_val_f*w_j
                        END IF
                     END IF
                  ELSE
                     in_val_f = l_boundary(myj, myk)
                  END IF
                  IF (has_u_boundary) THEN
                     IF (my_transpose) THEN
                        in_val_l = in_val(bo(2, 1), myj, myk)
                        IF (s(1) == 1) THEN
                           IF(.NOT.(.NOT. has_l_boundary))CALL cp__a("pw/pw_spline_utils.F",3060)
                           in_val_tmp = l_boundary(myj, myk)
                        ELSE
                           in_val_tmp = in_val(bo(2, 1)-1, myj, myk)
                        END IF
                        IF (my_sharpen) THEN
                           IF (kw == 0 .AND. jw == 0) THEN
                              IF (my_normalize) THEN
                                 out_val(bo(2, 1), j, k) = out_val(bo(2, 1), j, k)+ &
                                                           in_val_l*(2._dp-w_j)- &
                                                           in_val_tmp*weights_1d(1)*w_j
                              ELSE
                                 out_val(bo(2, 1), j, k) = out_val(bo(2, 1), j, k)+ &
                                                           in_val_l*w_j- &
                                                           in_val_tmp*weights_1d(1)*w_j
                              END IF
                           ELSE
                              out_val(bo(2, 1), j, k) = out_val(bo(2, 1), j, k)- &
                                                        w_j*in_val_l- &
                                                        in_val_tmp*weights_1d(1)*w_j
                           END IF
                        ELSE IF (my_smooth_boundary) THEN
                           out_val(bo(2, 1), j, k) = out_val(bo(2, 1), j, k)+ &
                                                     w_j*(in_val_l+in_val_tmp*weights_1d(1)/weights_1d(0))
                        ELSE
                           out_val(bo(2, 1), j, k) = out_val(bo(2, 1), j, k)+ &
                                                     w_j*(in_val_l+in_val_tmp*weights_1d(1))
                        END IF
                        in_val_l = 0._dp
                     ELSE
                        in_val_l = in_val(bo(2, 1), myj, myk)
                        IF (my_sharpen) THEN
                           IF (kw == 0 .AND. jw == 0) THEN
                              IF (my_normalize) THEN
                                 out_val(bo(2, 1), j, k) = out_val(bo(2, 1), j, k)+ &
                                                           in_val_l*(2._dp-w_j)
                              ELSE
                                 out_val(bo(2, 1), j, k) = out_val(bo(2, 1), j, k)+ &
                                                           in_val_l*w_j
                              END IF
                           ELSE
                              out_val(bo(2, 1), j, k) = out_val(bo(2, 1), j, k)- &
                                                        w_j*in_val_l
                           END IF
                        ELSE
                           out_val(bo(2, 1), j, k) = out_val(bo(2, 1), j, k)+ &
                                                     w_j*in_val_l
                        END IF
                     END IF
                  ELSE
                     in_val_l = u_boundary(myj, myk)
                  END IF
                  IF (last_index >= first_index) THEN
                     IF (my_transpose) THEN
                        IF (bo(1, 1)-1 == gbo(1, 1)) THEN
                           in_val_f = 0._dp
                        ELSE IF (bo(2, 1)+1 == gbo(2, 1)) THEN
                           in_val_l = 0._dp
                        END IF
                     END IF
                     IF (my_sharpen) THEN
                        w = -weights_1d*w_j
                        IF (kw == 0 .AND. jw == 0) THEN
                           IF (my_normalize) THEN
                              w(0) = w(0)+2._dp
                           ELSE
                              w(0) = -w(0)
                           END IF
                        END IF
                     ELSE
                        w = weights_1d*w_j
                     END IF
                     IF (my_smooth_boundary .AND. (.NOT. my_transpose)) THEN
                        IF (gbo(1, 1)+1 >= bo(1, 1) .AND. &
                            gbo(1, 1)+1 <= bo(2, 1) .AND. gbo(2, 1)-gbo(1, 1) > 2) THEN
                           IF (gbo(1, 1) >= bo(1, 1)) THEN
                              out_val(gbo(1, 1)+1, j, k) = out_val(gbo(1, 1)+1, j, k)+ &
                                                           in_val(gbo(1, 1), myj, myk)*w_j*weights_1d(-1)* &
                                                           (1._dp/weights_1d(0)-1._dp)
                           ELSE
                              out_val(gbo(1, 1)+1, j, k) = out_val(gbo(1, 1)+1, j, k)+ &
                                                           l_boundary(myj, myk)*w_j*weights_1d(-1)* &
                                                           (1._dp/weights_1d(0)-1._dp)
                           END IF
                        END IF
                     END IF
                     CALL pw_compose_stripe(weights=w, &
                                            in_val=in_val(first_index:last_index, myj, myk), &
                                            in_val_first=in_val_f, in_val_last=in_val_l, &
                                            out_val=out_val(first_index:last_index, j, k), &
                                            n_el=n_els)
!FM                   call pw_compose_stripe2(weights=w,&
!FM                        in_val=in_val,&
!FM                        in_val_first=in_val_f,in_val_last=in_val_l,&
!FM                        out_val=out_val,&
!FM                        first_val=first_index,last_val=last_index,&
!FM                        myj=myj,myk=myk,j=j,k=k)
                     IF (my_smooth_boundary .AND. (.NOT. my_transpose)) THEN
                        IF (gbo(2, 1)-1 >= bo(1, 1) .AND. &
                            gbo(2, 1)-1 <= bo(2, 1) .AND. gbo(2, 1)-gbo(1, 1) > 2) THEN
                           IF (gbo(2, 1) <= bo(2, 1)) THEN
                              out_val(gbo(2, 1)-1, j, k) = out_val(gbo(2, 1)-1, j, k)+ &
                                                           in_val(gbo(2, 1), myj, myk)*w_j*weights_1d(1)* &
                                                           (1._dp/weights_1d(0)-1._dp)
                           ELSE
                              out_val(gbo(2, 1)-1, j, k) = out_val(gbo(2, 1)-1, j, k)+ &
                                                           u_boundary(myj, myk)*w_j*weights_1d(1)* &
                                                           (1._dp/weights_1d(0)-1._dp)
                           END IF
                        END IF
                     END IF

                  END IF
               END DO
            END DO
         END DO
      END DO

      IF (is_split) THEN
         DEALLOCATE (l_boundary, u_boundary)
      END IF
   END SUBROUTINE pw_nn_compose_r_no_pbc

! **************************************************************************************************
!> \brief ...
!> \param pw_in ...
!> \param pw_out ...
! **************************************************************************************************
   SUBROUTINE spl3_nopbc(pw_in, pw_out)
      TYPE(pw_type), POINTER                             :: pw_in, pw_out

      CALL pw_zero(pw_out)
      CALL pw_nn_compose_r_no_pbc(weights_1d=spl3_1d_coeffs0, pw_in=pw_in, &
                                  pw_out=pw_out, sharpen=.FALSE., normalize=.FALSE.)
   END SUBROUTINE spl3_nopbc

! **************************************************************************************************
!> \brief ...
!> \param pw_in ...
!> \param pw_out ...
! **************************************************************************************************
   SUBROUTINE spl3_nopbct(pw_in, pw_out)
      TYPE(pw_type), POINTER                             :: pw_in, pw_out

      CALL pw_zero(pw_out)
      CALL pw_nn_compose_r_no_pbc(weights_1d=spl3_1d_coeffs0, pw_in=pw_in, &
                                  pw_out=pw_out, sharpen=.FALSE., normalize=.FALSE., transpose=.TRUE.)
   END SUBROUTINE spl3_nopbct

! **************************************************************************************************
!> \brief ...
!> \param pw_in ...
!> \param pw_out ...
! **************************************************************************************************
   SUBROUTINE spl3_pbc(pw_in, pw_out)
      TYPE(pw_type), POINTER                             :: pw_in, pw_out

      CALL pw_zero(pw_out)
      CALL pw_nn_smear_r(pw_in, pw_out, coeffs=spline3_coeffs)
   END SUBROUTINE spl3_pbc

! **************************************************************************************************
!> \brief Evaluates the PBC interpolated Spline (pw) function on the generic
!>      input vector (vec)
!> \param vec ...
!> \param pw ...
!> \retval val ...
!> \par History
!>      12.2007 Adapted for use with distributed grids [rdeclerck]
!> \author Teodoro Laino 12/2005 [tlaino]
!> \note
!>      Requires the Spline coefficients to be computed with PBC
! **************************************************************************************************
   FUNCTION Eval_Interp_Spl3_pbc(vec, pw) RESULT(val)
      REAL(KIND=dp), DIMENSION(3), INTENT(in)            :: vec
      TYPE(pw_type), POINTER                             :: pw
      REAL(KIND=dp)                                      :: val

      CHARACTER(len=*), PARAMETER :: routineN = 'Eval_Interp_Spl3_pbc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ivec(3), j, k, npts(3)
      INTEGER, DIMENSION(2, 3)                           :: bo, bo_l
      INTEGER, DIMENSION(4)                              :: ii, ij, ik
      LOGICAL                                            :: my_mpsum
      REAL(KIND=dp) :: a1, a2, a3, b1, b2, b3, c1, c2, c3, d1, d2, d3, dr1, dr2, dr3, e1, e2, e3, &
         f1, f2, f3, g1, g2, g3, h1, h2, h3, p1, p2, p3, q1, q2, q3, r1, r2, r3, s1, s2, s3, s4, &
         t1, t2, t3, t4, u1, u2, u3, v1, v2, v3, v4, xd1, xd2, xd3
      REAL(KIND=dp), DIMENSION(4, 4, 4)                  :: box
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: grid

      NULLIFY (grid)
      my_mpsum = (pw%pw_grid%para%mode /= PW_MODE_LOCAL)
      npts = pw%pw_grid%npts
      ivec = FLOOR(vec/pw%pw_grid%dr)
      dr1 = pw%pw_grid%dr(1)
      dr2 = pw%pw_grid%dr(2)
      dr3 = pw%pw_grid%dr(3)

      xd1 = (vec(1)/dr1)-REAL(ivec(1), kind=dp)
      xd2 = (vec(2)/dr2)-REAL(ivec(2), kind=dp)
      xd3 = (vec(3)/dr3)-REAL(ivec(3), kind=dp)
      grid => pw%cr3d(:, :, :)
      bo = pw%pw_grid%bounds
      bo_l = pw%pw_grid%bounds_local

      ik(1) = MODULO(ivec(3)-1, npts(3))+bo(1, 3)
      ik(2) = MODULO(ivec(3), npts(3))+bo(1, 3)
      ik(3) = MODULO(ivec(3)+1, npts(3))+bo(1, 3)
      ik(4) = MODULO(ivec(3)+2, npts(3))+bo(1, 3)

      ij(1) = MODULO(ivec(2)-1, npts(2))+bo(1, 2)
      ij(2) = MODULO(ivec(2), npts(2))+bo(1, 2)
      ij(3) = MODULO(ivec(2)+1, npts(2))+bo(1, 2)
      ij(4) = MODULO(ivec(2)+2, npts(2))+bo(1, 2)

      ii(1) = MODULO(ivec(1)-1, npts(1))+bo(1, 1)
      ii(2) = MODULO(ivec(1), npts(1))+bo(1, 1)
      ii(3) = MODULO(ivec(1)+1, npts(1))+bo(1, 1)
      ii(4) = MODULO(ivec(1)+2, npts(1))+bo(1, 1)

      DO k = 1, 4
         DO j = 1, 4
            DO i = 1, 4
               IF ( &
                  ii(i) >= bo_l(1, 1) .AND. &
                  ii(i) <= bo_l(2, 1) .AND. &
                  ij(j) >= bo_l(1, 2) .AND. &
                  ij(j) <= bo_l(2, 2) .AND. &
                  ik(k) >= bo_l(1, 3) .AND. &
                  ik(k) <= bo_l(2, 3) &
                  ) THEN
                  box(i, j, k) = grid(ii(i)+1-bo_l(1, 1), &
                                      ij(j)+1-bo_l(1, 2), &
                                      ik(k)+1-bo_l(1, 3))
               ELSE
                  box(i, j, k) = 0.0_dp
               END IF
            END DO
         END DO
      END DO

      a1 = 3.0_dp+xd1
      a2 = a1*a1
      a3 = a2*a1
      b1 = 2.0_dp+xd1
      b2 = b1*b1
      b3 = b2*b1
      c1 = 1.0_dp+xd1
      c2 = c1*c1
      c3 = c2*c1
      d1 = xd1
      d2 = d1*d1
      d3 = d2*d1
      e1 = 3.0_dp+xd2
      e2 = e1*e1
      e3 = e2*e1
      f1 = 2.0_dp+xd2
      f2 = f1*f1
      f3 = f2*f1
      g1 = 1.0_dp+xd2
      g2 = g1*g1
      g3 = g2*g1
      h1 = xd2
      h2 = h1*h1
      h3 = h2*h1
      p1 = 3.0_dp+xd3
      p2 = p1*p1
      p3 = p2*p1
      q1 = 2.0_dp+xd3
      q2 = q1*q1
      q3 = q2*q1
      r1 = 1.0_dp+xd3
      r2 = r1*r1
      r3 = r2*r1
      u1 = xd3
      u2 = u1*u1
      u3 = u2*u1

      t1 = 1.0_dp/6.0_dp*(64.0_dp-48.0_dp*a1+12.0_dp*a2-a3)
      t2 = -22.0_dp/3.0_dp+10.0_dp*b1-4.0_dp*b2+0.5_dp*b3
      t3 = 2.0_dp/3.0_dp-2.0_dp*c1+2.0_dp*c2-0.5_dp*c3
      t4 = 1.0_dp/6.0_dp*d3
      s1 = 1.0_dp/6.0_dp*(64.0_dp-48.0_dp*e1+12.0_dp*e2-e3)
      s2 = -22.0_dp/3.0_dp+10.0_dp*f1-4.0_dp*f2+0.5_dp*f3
      s3 = 2.0_dp/3.0_dp-2.0_dp*g1+2.0_dp*g2-0.5_dp*g3
      s4 = 1.0_dp/6.0_dp*h3
      v1 = 1.0_dp/6.0_dp*(64.0_dp-48.0_dp*p1+12.0_dp*p2-p3)
      v2 = -22.0_dp/3.0_dp+10.0_dp*q1-4.0_dp*q2+0.5_dp*q3
      v3 = 2.0_dp/3.0_dp-2.0_dp*r1+2.0_dp*r2-0.5_dp*r3
      v4 = 1.0_dp/6.0_dp*u3

      val = ((box(1, 1, 1)*t1+box(2, 1, 1)*t2+box(3, 1, 1)*t3+box(4, 1, 1)*t4)*s1+ &
             (box(1, 2, 1)*t1+box(2, 2, 1)*t2+box(3, 2, 1)*t3+box(4, 2, 1)*t4)*s2+ &
             (box(1, 3, 1)*t1+box(2, 3, 1)*t2+box(3, 3, 1)*t3+box(4, 3, 1)*t4)*s3+ &
             (box(1, 4, 1)*t1+box(2, 4, 1)*t2+box(3, 4, 1)*t3+box(4, 4, 1)*t4)*s4)*v1+ &
            ((box(1, 1, 2)*t1+box(2, 1, 2)*t2+box(3, 1, 2)*t3+box(4, 1, 2)*t4)*s1+ &
             (box(1, 2, 2)*t1+box(2, 2, 2)*t2+box(3, 2, 2)*t3+box(4, 2, 2)*t4)*s2+ &
             (box(1, 3, 2)*t1+box(2, 3, 2)*t2+box(3, 3, 2)*t3+box(4, 3, 2)*t4)*s3+ &
             (box(1, 4, 2)*t1+box(2, 4, 2)*t2+box(3, 4, 2)*t3+box(4, 4, 2)*t4)*s4)*v2+ &
            ((box(1, 1, 3)*t1+box(2, 1, 3)*t2+box(3, 1, 3)*t3+box(4, 1, 3)*t4)*s1+ &
             (box(1, 2, 3)*t1+box(2, 2, 3)*t2+box(3, 2, 3)*t3+box(4, 2, 3)*t4)*s2+ &
             (box(1, 3, 3)*t1+box(2, 3, 3)*t2+box(3, 3, 3)*t3+box(4, 3, 3)*t4)*s3+ &
             (box(1, 4, 3)*t1+box(2, 4, 3)*t2+box(3, 4, 3)*t3+box(4, 4, 3)*t4)*s4)*v3+ &
            ((box(1, 1, 4)*t1+box(2, 1, 4)*t2+box(3, 1, 4)*t3+box(4, 1, 4)*t4)*s1+ &
             (box(1, 2, 4)*t1+box(2, 2, 4)*t2+box(3, 2, 4)*t3+box(4, 2, 4)*t4)*s2+ &
             (box(1, 3, 4)*t1+box(2, 3, 4)*t2+box(3, 3, 4)*t3+box(4, 3, 4)*t4)*s3+ &
             (box(1, 4, 4)*t1+box(2, 4, 4)*t2+box(3, 4, 4)*t3+box(4, 4, 4)*t4)*s4)*v4

      IF (my_mpsum) CALL mp_sum(val, pw%pw_grid%para%group)

   END FUNCTION Eval_Interp_Spl3_pbc

! **************************************************************************************************
!> \brief Evaluates the derivatives of the PBC interpolated Spline (pw)
!>      function on the generic input vector (vec)
!> \param vec ...
!> \param pw ...
!> \retval val ...
!> \par History
!>      12.2007 Adapted for use with distributed grids [rdeclerck]
!> \author Teodoro Laino 12/2005 [tlaino]
!> \note
!>      Requires the Spline coefficients to be computed with PBC
! **************************************************************************************************
   FUNCTION Eval_d_Interp_Spl3_pbc(vec, pw) RESULT(val)
      REAL(KIND=dp), DIMENSION(3), INTENT(in)            :: vec
      TYPE(pw_type), POINTER                             :: pw
      REAL(KIND=dp)                                      :: val(3)

      CHARACTER(len=*), PARAMETER :: routineN = 'Eval_d_Interp_Spl3_pbc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ivec(3), j, k, npts(3)
      INTEGER, DIMENSION(2, 3)                           :: bo, bo_l
      INTEGER, DIMENSION(4)                              :: ii, ij, ik
      LOGICAL                                            :: my_mpsum
      REAL(KIND=dp) :: a1, a2, a3, b1, b2, b3, c1, c2, c3, d1, d2, d3, dr1, dr1i, dr2, dr2i, dr3, &
         dr3i, e1, e2, e3, f1, f2, f3, g1, g2, g3, h1, h2, h3, p1, p2, p3, q1, q2, q3, r1, r2, r3, &
         s1, s1d, s1o, s2, s2d, s2o, s3, s3d, s3o, s4, s4d, s4o, t1, t1d, t1o, t2, t2d, t2o, t3, &
         t3d, t3o, t4, t4d, t4o, u1, u2, u3, v1, v1d, v1o, v2, v2d, v2o, v3, v3d, v3o, v4, v4d, &
         v4o, xd1, xd2, xd3
      REAL(KIND=dp), DIMENSION(4, 4, 4)                  :: box
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER         :: grid

      NULLIFY (grid)
      my_mpsum = (pw%pw_grid%para%mode /= PW_MODE_LOCAL)
      npts = pw%pw_grid%npts
      ivec = FLOOR(vec/pw%pw_grid%dr)
      dr1 = pw%pw_grid%dr(1)
      dr2 = pw%pw_grid%dr(2)
      dr3 = pw%pw_grid%dr(3)
      dr1i = 1.0_dp/dr1
      dr2i = 1.0_dp/dr2
      dr3i = 1.0_dp/dr3
      xd1 = (vec(1)/dr1)-REAL(ivec(1), kind=dp)
      xd2 = (vec(2)/dr2)-REAL(ivec(2), kind=dp)
      xd3 = (vec(3)/dr3)-REAL(ivec(3), kind=dp)
      grid => pw%cr3d(:, :, :)
      bo = pw%pw_grid%bounds
      bo_l = pw%pw_grid%bounds_local

      ik(1) = MODULO(ivec(3)-1, npts(3))+bo(1, 3)
      ik(2) = MODULO(ivec(3), npts(3))+bo(1, 3)
      ik(3) = MODULO(ivec(3)+1, npts(3))+bo(1, 3)
      ik(4) = MODULO(ivec(3)+2, npts(3))+bo(1, 3)

      ij(1) = MODULO(ivec(2)-1, npts(2))+bo(1, 2)
      ij(2) = MODULO(ivec(2), npts(2))+bo(1, 2)
      ij(3) = MODULO(ivec(2)+1, npts(2))+bo(1, 2)
      ij(4) = MODULO(ivec(2)+2, npts(2))+bo(1, 2)

      ii(1) = MODULO(ivec(1)-1, npts(1))+bo(1, 1)
      ii(2) = MODULO(ivec(1), npts(1))+bo(1, 1)
      ii(3) = MODULO(ivec(1)+1, npts(1))+bo(1, 1)
      ii(4) = MODULO(ivec(1)+2, npts(1))+bo(1, 1)

      DO k = 1, 4
         DO j = 1, 4
            DO i = 1, 4
               IF ( &
                  ii(i) >= bo_l(1, 1) .AND. &
                  ii(i) <= bo_l(2, 1) .AND. &
                  ij(j) >= bo_l(1, 2) .AND. &
                  ij(j) <= bo_l(2, 2) .AND. &
                  ik(k) >= bo_l(1, 3) .AND. &
                  ik(k) <= bo_l(2, 3) &
                  ) THEN
                  box(i, j, k) = grid(ii(i)+1-bo_l(1, 1), &
                                      ij(j)+1-bo_l(1, 2), &
                                      ik(k)+1-bo_l(1, 3))
               ELSE
                  box(i, j, k) = 0.0_dp
               END IF
            END DO
         END DO
      END DO

      a1 = 3.0_dp+xd1
      a2 = a1*a1
      a3 = a2*a1
      b1 = 2.0_dp+xd1
      b2 = b1*b1
      b3 = b2*b1
      c1 = 1.0_dp+xd1
      c2 = c1*c1
      c3 = c2*c1
      d1 = xd1
      d2 = d1*d1
      d3 = d2*d1
      e1 = 3.0_dp+xd2
      e2 = e1*e1
      e3 = e2*e1
      f1 = 2.0_dp+xd2
      f2 = f1*f1
      f3 = f2*f1
      g1 = 1.0_dp+xd2
      g2 = g1*g1
      g3 = g2*g1
      h1 = xd2
      h2 = h1*h1
      h3 = h2*h1
      p1 = 3.0_dp+xd3
      p2 = p1*p1
      p3 = p2*p1
      q1 = 2.0_dp+xd3
      q2 = q1*q1
      q3 = q2*q1
      r1 = 1.0_dp+xd3
      r2 = r1*r1
      r3 = r2*r1
      u1 = xd3
      u2 = u1*u1
      u3 = u2*u1

      t1o = 1.0_dp/6.0_dp*(64.0_dp-48.0_dp*a1+12.0_dp*a2-a3)
      t2o = -22.0_dp/3.0_dp+10.0_dp*b1-4.0_dp*b2+0.5_dp*b3
      t3o = 2.0_dp/3.0_dp-2.0_dp*c1+2.0_dp*c2-0.5_dp*c3
      t4o = 1.0_dp/6.0_dp*d3
      s1o = 1.0_dp/6.0_dp*(64.0_dp-48.0_dp*e1+12.0_dp*e2-e3)
      s2o = -22.0_dp/3.0_dp+10.0_dp*f1-4.0_dp*f2+0.5_dp*f3
      s3o = 2.0_dp/3.0_dp-2.0_dp*g1+2.0_dp*g2-0.5_dp*g3
      s4o = 1.0_dp/6.0_dp*h3
      v1o = 1.0_dp/6.0_dp*(64.0_dp-48.0_dp*p1+12.0_dp*p2-p3)
      v2o = -22.0_dp/3.0_dp+10.0_dp*q1-4.0_dp*q2+0.5_dp*q3
      v3o = 2.0_dp/3.0_dp-2.0_dp*r1+2.0_dp*r2-0.5_dp*r3
      v4o = 1.0_dp/6.0_dp*u3

      t1d = -8.0_dp+4.0_dp*a1-0.5_dp*a2
      t2d = 10.0_dp-8.0_dp*b1+1.5_dp*b2
      t3d = -2.0_dp+4.0_dp*c1-1.5_dp*c2
      t4d = 0.5_dp*d2
      s1d = -8.0_dp+4.0_dp*e1-0.5_dp*e2
      s2d = 10.0_dp-8.0_dp*f1+1.5_dp*f2
      s3d = -2.0_dp+4.0_dp*g1-1.5_dp*g2
      s4d = 0.5_dp*h2
      v1d = -8.0_dp+4.0_dp*p1-0.5_dp*p2
      v2d = 10.0_dp-8.0_dp*q1+1.5_dp*q2
      v3d = -2.0_dp+4.0_dp*r1-1.5_dp*r2
      v4d = 0.5_dp*u2

      t1 = t1d*dr1i
      t2 = t2d*dr1i
      t3 = t3d*dr1i
      t4 = t4d*dr1i
      s1 = s1o
      s2 = s2o
      s3 = s3o
      s4 = s4o
      v1 = v1o
      v2 = v2o
      v3 = v3o
      v4 = v4o
      val(1) = ((box(1, 1, 1)*t1+box(2, 1, 1)*t2+box(3, 1, 1)*t3+box(4, 1, 1)*t4)*s1+ &
                (box(1, 2, 1)*t1+box(2, 2, 1)*t2+box(3, 2, 1)*t3+box(4, 2, 1)*t4)*s2+ &
                (box(1, 3, 1)*t1+box(2, 3, 1)*t2+box(3, 3, 1)*t3+box(4, 3, 1)*t4)*s3+ &
                (box(1, 4, 1)*t1+box(2, 4, 1)*t2+box(3, 4, 1)*t3+box(4, 4, 1)*t4)*s4)*v1+ &
               ((box(1, 1, 2)*t1+box(2, 1, 2)*t2+box(3, 1, 2)*t3+box(4, 1, 2)*t4)*s1+ &
                (box(1, 2, 2)*t1+box(2, 2, 2)*t2+box(3, 2, 2)*t3+box(4, 2, 2)*t4)*s2+ &
                (box(1, 3, 2)*t1+box(2, 3, 2)*t2+box(3, 3, 2)*t3+box(4, 3, 2)*t4)*s3+ &
                (box(1, 4, 2)*t1+box(2, 4, 2)*t2+box(3, 4, 2)*t3+box(4, 4, 2)*t4)*s4)*v2+ &
               ((box(1, 1, 3)*t1+box(2, 1, 3)*t2+box(3, 1, 3)*t3+box(4, 1, 3)*t4)*s1+ &
                (box(1, 2, 3)*t1+box(2, 2, 3)*t2+box(3, 2, 3)*t3+box(4, 2, 3)*t4)*s2+ &
                (box(1, 3, 3)*t1+box(2, 3, 3)*t2+box(3, 3, 3)*t3+box(4, 3, 3)*t4)*s3+ &
                (box(1, 4, 3)*t1+box(2, 4, 3)*t2+box(3, 4, 3)*t3+box(4, 4, 3)*t4)*s4)*v3+ &
               ((box(1, 1, 4)*t1+box(2, 1, 4)*t2+box(3, 1, 4)*t3+box(4, 1, 4)*t4)*s1+ &
                (box(1, 2, 4)*t1+box(2, 2, 4)*t2+box(3, 2, 4)*t3+box(4, 2, 4)*t4)*s2+ &
                (box(1, 3, 4)*t1+box(2, 3, 4)*t2+box(3, 3, 4)*t3+box(4, 3, 4)*t4)*s3+ &
                (box(1, 4, 4)*t1+box(2, 4, 4)*t2+box(3, 4, 4)*t3+box(4, 4, 4)*t4)*s4)*v4

      t1 = t1o
      t2 = t2o
      t3 = t3o
      t4 = t4o
      s1 = s1d*dr2i
      s2 = s2d*dr2i
      s3 = s3d*dr2i
      s4 = s4d*dr2i
      v1 = v1o
      v2 = v2o
      v3 = v3o
      v4 = v4o
      val(2) = ((box(1, 1, 1)*t1+box(2, 1, 1)*t2+box(3, 1, 1)*t3+box(4, 1, 1)*t4)*s1+ &
                (box(1, 2, 1)*t1+box(2, 2, 1)*t2+box(3, 2, 1)*t3+box(4, 2, 1)*t4)*s2+ &
                (box(1, 3, 1)*t1+box(2, 3, 1)*t2+box(3, 3, 1)*t3+box(4, 3, 1)*t4)*s3+ &
                (box(1, 4, 1)*t1+box(2, 4, 1)*t2+box(3, 4, 1)*t3+box(4, 4, 1)*t4)*s4)*v1+ &
               ((box(1, 1, 2)*t1+box(2, 1, 2)*t2+box(3, 1, 2)*t3+box(4, 1, 2)*t4)*s1+ &
                (box(1, 2, 2)*t1+box(2, 2, 2)*t2+box(3, 2, 2)*t3+box(4, 2, 2)*t4)*s2+ &
                (box(1, 3, 2)*t1+box(2, 3, 2)*t2+box(3, 3, 2)*t3+box(4, 3, 2)*t4)*s3+ &
                (box(1, 4, 2)*t1+box(2, 4, 2)*t2+box(3, 4, 2)*t3+box(4, 4, 2)*t4)*s4)*v2+ &
               ((box(1, 1, 3)*t1+box(2, 1, 3)*t2+box(3, 1, 3)*t3+box(4, 1, 3)*t4)*s1+ &
                (box(1, 2, 3)*t1+box(2, 2, 3)*t2+box(3, 2, 3)*t3+box(4, 2, 3)*t4)*s2+ &
                (box(1, 3, 3)*t1+box(2, 3, 3)*t2+box(3, 3, 3)*t3+box(4, 3, 3)*t4)*s3+ &
                (box(1, 4, 3)*t1+box(2, 4, 3)*t2+box(3, 4, 3)*t3+box(4, 4, 3)*t4)*s4)*v3+ &
               ((box(1, 1, 4)*t1+box(2, 1, 4)*t2+box(3, 1, 4)*t3+box(4, 1, 4)*t4)*s1+ &
                (box(1, 2, 4)*t1+box(2, 2, 4)*t2+box(3, 2, 4)*t3+box(4, 2, 4)*t4)*s2+ &
                (box(1, 3, 4)*t1+box(2, 3, 4)*t2+box(3, 3, 4)*t3+box(4, 3, 4)*t4)*s3+ &
                (box(1, 4, 4)*t1+box(2, 4, 4)*t2+box(3, 4, 4)*t3+box(4, 4, 4)*t4)*s4)*v4

      t1 = t1o
      t2 = t2o
      t3 = t3o
      t4 = t4o
      s1 = s1o
      s2 = s2o
      s3 = s3o
      s4 = s4o
      v1 = v1d*dr3i
      v2 = v2d*dr3i
      v3 = v3d*dr3i
      v4 = v4d*dr3i
      val(3) = ((box(1, 1, 1)*t1+box(2, 1, 1)*t2+box(3, 1, 1)*t3+box(4, 1, 1)*t4)*s1+ &
                (box(1, 2, 1)*t1+box(2, 2, 1)*t2+box(3, 2, 1)*t3+box(4, 2, 1)*t4)*s2+ &
                (box(1, 3, 1)*t1+box(2, 3, 1)*t2+box(3, 3, 1)*t3+box(4, 3, 1)*t4)*s3+ &
                (box(1, 4, 1)*t1+box(2, 4, 1)*t2+box(3, 4, 1)*t3+box(4, 4, 1)*t4)*s4)*v1+ &
               ((box(1, 1, 2)*t1+box(2, 1, 2)*t2+box(3, 1, 2)*t3+box(4, 1, 2)*t4)*s1+ &
                (box(1, 2, 2)*t1+box(2, 2, 2)*t2+box(3, 2, 2)*t3+box(4, 2, 2)*t4)*s2+ &
                (box(1, 3, 2)*t1+box(2, 3, 2)*t2+box(3, 3, 2)*t3+box(4, 3, 2)*t4)*s3+ &
                (box(1, 4, 2)*t1+box(2, 4, 2)*t2+box(3, 4, 2)*t3+box(4, 4, 2)*t4)*s4)*v2+ &
               ((box(1, 1, 3)*t1+box(2, 1, 3)*t2+box(3, 1, 3)*t3+box(4, 1, 3)*t4)*s1+ &
                (box(1, 2, 3)*t1+box(2, 2, 3)*t2+box(3, 2, 3)*t3+box(4, 2, 3)*t4)*s2+ &
                (box(1, 3, 3)*t1+box(2, 3, 3)*t2+box(3, 3, 3)*t3+box(4, 3, 3)*t4)*s3+ &
                (box(1, 4, 3)*t1+box(2, 4, 3)*t2+box(3, 4, 3)*t3+box(4, 4, 3)*t4)*s4)*v3+ &
               ((box(1, 1, 4)*t1+box(2, 1, 4)*t2+box(3, 1, 4)*t3+box(4, 1, 4)*t4)*s1+ &
                (box(1, 2, 4)*t1+box(2, 2, 4)*t2+box(3, 2, 4)*t3+box(4, 2, 4)*t4)*s2+ &
                (box(1, 3, 4)*t1+box(2, 3, 4)*t2+box(3, 3, 4)*t3+box(4, 3, 4)*t4)*s3+ &
                (box(1, 4, 4)*t1+box(2, 4, 4)*t2+box(3, 4, 4)*t3+box(4, 4, 4)*t4)*s4)*v4

      IF (my_mpsum) CALL mp_sum(val, pw%pw_grid%para%group)

   END FUNCTION Eval_d_Interp_Spl3_pbc

END MODULE pw_spline_utils
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      01.2008 Created
!> \author Joost
! **************************************************************************************************
MODULE input_cp2k_rsgrid

   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_create,&
                                              section_type
   USE kinds,                           ONLY: dp
   USE realspace_grid_types,            ONLY: rsgrid_automatic,&
                                              rsgrid_distributed,&
                                              rsgrid_replicated
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_rsgrid'

   PUBLIC :: create_rsgrid_section

!***
CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param section ...
!> \author Joost
! **************************************************************************************************
   SUBROUTINE create_rsgrid_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_rsgrid_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_rsgrid.F",49)
      CALL section_create( &
         section, name="RS_GRID", &
         description="Set options that influence how the realspace grids are being distributed in parallel runs.", &
         n_keywords=5, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="DISTRIBUTION_TYPE", &
                          description="Parallelization strategy.", &
                          usage="DISTRIBUTION_TYPE DISTRIBUTED", &
                          enum_c_vals=s2a("AUTOMATIC", "DISTRIBUTED", "REPLICATED"), &
                          enum_i_vals=(/rsgrid_automatic, rsgrid_distributed, rsgrid_replicated/), &
                          enum_desc=s2a("Use heuristic rules to decide between distributed and replicated", &
                                        "Force a distributed setup if possible", &
                                        "Force a replicated setup"), &
                          default_i_val=rsgrid_automatic)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISTRIBUTION_LAYOUT", &
                          description="Specifies the number of slices in the x, y and z directions."// &
                          "-1 specifies that any number of slices is OK."// &
                          "If a given distribution can not be satisfied, a replicated grid will result."// &
                          "Also see LOCK_DISTRIBUTION.", &
                          usage="DISTRIBUTION_LAYOUT", &
                          repeats=.FALSE., n_var=3, &
                          default_i_vals=(/-1, -1, -1/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_DISTRIBUTED_LEVEL", &
                          description="If the multigrid-level of a grid is larger than the parameter,"// &
                          " it will not be distributed in the automatic scheme.", &
                          usage="MAX_DISTRIBUTED_LEVEL 1", &
                          default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LOCK_DISTRIBUTION", &
                          description="Expert use only, only basic QS deals correctly with a non-default value."// &
                          "If the distribution is locked, a grid will have the same distribution as"// &
                          "the next finer multigrid (provided it is distributed)."// &
                          "If unlocked, all grids can be distributed freely.", &
                          usage="LOCK_DISTRIBUTION TRUE", &
                          default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MEMORY_FACTOR", &
                          description="A grid will only be distributed if the memory usage for that grid (including halo) "// &
                          "is smaller than a replicated grid by this parameter.", &
                          usage="MEMORY_FACTOR 4.0", &
                          default_r_val=2.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="HALO_REDUCTION_FACTOR", &
                          description="Can be used to reduce the halo of the distributed grid (experimental features).", &
                          usage="HALO_REDUCTION_FACTOR 0.5", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_rsgrid_section

END MODULE input_cp2k_rsgrid
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief function that build the poisson section of the input
!> \par History
!>      03.2006 fusing of poisson_dft and poisson_mm
!> \author fawzi
! **************************************************************************************************
MODULE input_cp2k_poisson
   USE bibliography,                    ONLY: &
        Aguado2003, BaniHashemian2016, Blochl1995, Darden1993, Essmann1995, Ewald1921, &
        Genovese2006, Genovese2007, Laino2008, Martyna1999, Toukmaji1996
   USE cell_types,                      ONLY: use_perd_none,&
                                              use_perd_x,&
                                              use_perd_xy,&
                                              use_perd_xyz,&
                                              use_perd_xz,&
                                              use_perd_y,&
                                              use_perd_yz,&
                                              use_perd_z
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              low_print_level,&
                                              medium_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE dct,                             ONLY: neumannX,&
                                              neumannXY,&
                                              neumannXYZ,&
                                              neumannXZ,&
                                              neumannY,&
                                              neumannYZ,&
                                              neumannZ
   USE dielectric_types,                ONLY: &
        derivative_cd3, derivative_cd5, derivative_cd7, derivative_fft, derivative_fft_use_deps, &
        derivative_fft_use_drho, rho_dependent, spatially_dependent, spatially_rho_dependent
   USE dirichlet_bc_types,              ONLY: CIRCUMSCRIBED,&
                                              INSCRIBED,&
                                              xy_plane,&
                                              xz_plane,&
                                              yz_plane
   USE input_constants,                 ONLY: do_fist_pol_cg,&
                                              do_fist_pol_none,&
                                              do_fist_pol_sc
   USE input_cp2k_rsgrid,               ONLY: create_rsgrid_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: enum_t,&
                                              integer_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE multipole_types,                 ONLY: do_multipole_charge,&
                                              do_multipole_dipole,&
                                              do_multipole_none,&
                                              do_multipole_quadrupole
   USE ps_implicit_types,               ONLY: MIXED_BC,&
                                              MIXED_PERIODIC_BC,&
                                              NEUMANN_BC,&
                                              PERIODIC_BC
   USE pw_poisson_types,                ONLY: &
        do_ewald_ewald, do_ewald_none, do_ewald_pme, do_ewald_spme, pw_poisson_analytic, &
        pw_poisson_implicit, pw_poisson_mt, pw_poisson_multipole, pw_poisson_periodic, &
        pw_poisson_wavelet
   USE pw_spline_utils,                 ONLY: no_precond,&
                                              precond_spl3_1,&
                                              precond_spl3_2,&
                                              precond_spl3_3,&
                                              precond_spl3_aint,&
                                              precond_spl3_aint2
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_poisson'

   PUBLIC :: create_poisson_section, &
             create_gspace_interp_section, &
             create_ewald_section
!***
CONTAINS

! **************************************************************************************************
!> \brief Creates the Poisson section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_poisson_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_poisson_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",107)
      CALL section_create(section, name="poisson", &
                          description="Sets up the poisson resolutor.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)
      CALL keyword_create(keyword, name="POISSON_SOLVER", &
                          variants=(/"POISSON", "PSOLVER"/), &
                          description="Specify which kind of solver to use to solve the Poisson equation.", &
                          usage="POISSON_SOLVER char", &
                          enum_c_vals=s2a("PERIODIC", "ANALYTIC", "MT", "MULTIPOLE", "WAVELET", "IMPLICIT"), &
                          enum_i_vals=(/pw_poisson_periodic, pw_poisson_analytic, pw_poisson_mt, pw_poisson_multipole, &
                                        pw_poisson_wavelet, pw_poisson_implicit/), &
                          citations=(/Blochl1995, Martyna1999, Genovese2006, Genovese2007/), &
                          default_i_val=pw_poisson_periodic)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PERIODIC", &
                          description="Specify the directions on wich apply PBC. Important notice, "// &
                          " this only applies to the electrostatics."// &
                          " See the CELL section to specify the periodicity used for e.g. the pair lists."// &
                          " Typically the settings should be the same.", &
                          usage="PERIODIC (x|y|z|xy|xz|yz|xyz|none)", &
                          enum_c_vals=s2a("x", "y", "z", "xy", "xz", "yz", "xyz", "none"), &
                          enum_i_vals=(/use_perd_x, use_perd_y, use_perd_z, &
                                        use_perd_xy, use_perd_xz, use_perd_yz, &
                                        use_perd_xyz, use_perd_none/), &
                          default_i_val=use_perd_xyz)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_mt_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_wavelet_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_multipole_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_ewald_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_implicit_ps_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_poisson_section

! **************************************************************************************************
!> \brief Section to set-up parameters for decoupling using the Bloechl scheme
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_multipole_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_multipole_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",174)

      CALL section_create(section, name="MULTIPOLE", &
                          description="This section is used to set up the decoupling of QM periodic images with "// &
                          "the use of density derived atomic point charges.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)
      CALL keyword_create(keyword, name="RCUT", &
                          description="Real space cutoff for the Ewald sum.", &
                          usage="RCUT {real}", n_var=1, type_of_var=real_t, &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EWALD_PRECISION", &
                          description="Precision achieved in the Ewald sum.", &
                          usage="EWALD_PRECISION {real}", n_var=1, type_of_var=real_t, &
                          unit_str="hartree", default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ANALYTICAL_GTERM", &
                          description="Evaluates the Gterm in the Ewald Scheme analytically instead of using Splines.", &
                          usage="ANALYTICAL_GTERM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NGRIDS", &
                          description="Specifies the number of grid points used for the Interpolation of the G-space term", &
                          usage="NGRIDS <integer> <iteger> <integer> ", n_var=3, default_i_vals=(/50, 50, 50/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_gspace_interp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL cp_print_key_section_create(subsection, "check_spline", &
                                       description="Controls the checking of the G-space term Spline Interpolation.", &
                                       print_level=medium_print_level, filename="GSpace-SplInterp")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL cp_print_key_section_create(subsection, "program_run_info", &
                                       description="Controls the printing of basic information during the run", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_multipole_section

! **************************************************************************************************
!> \brief Creates the Martyna-Tuckerman section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_mt_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mt_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",240)
      CALL section_create(section, name="mt", &
                          description="Sets up parameters of  Martyna-Tuckerman poisson solver. "// &
                          "Note that exact results are only guaranteed if the unit cell is "// &
                          "twice as large as charge density (and serious artefacts can result "// &
                          "if the cell is much smaller).", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/Martyna1999/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ALPHA", &
                          description="Convergence parameter ALPHA*RMIN. Default value 7.0", &
                          usage="ALPHA real", &
                          n_var=1, default_r_val=7.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REL_CUTOFF", &
                          description="Specify the multiplicative factor for the CUTOFF keyword in MULTI_GRID "// &
                          " section. The result gives the cutoff at which the 1/r non-periodic FFT3D is evaluated."// &
                          "Default is 2.0", &
                          usage="REL_CUTOFF real", &
                          n_var=1, default_r_val=2.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_mt_section

! **************************************************************************************************
!> \brief ...
!> \param section will contain the ewald section
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_ewald_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ewald_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",283)
      CALL section_create(section, name="ewald", &
                          description="Ewald parameters controlling electrostatic only for CLASSICAL MM.", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE., &
                          citations=(/Ewald1921, Darden1993, Essmann1995, Toukmaji1996, Laino2008/))

      NULLIFY (keyword, print_key, subsection)
      CALL keyword_create( &
         keyword, name="EWALD_TYPE", &
         description="The type of ewald you want to perform.", &
         citations=(/Ewald1921, Essmann1995, Darden1993/), &
         usage="EWALD_TYPE (NONE|EWALD|PME|SPME)", &
         default_i_val=do_ewald_ewald, &
         enum_c_vals=(/"none     ", &
                       "ewald    ", &
                       "pme      ", &
                       "spme     "/), &
         enum_i_vals=(/do_ewald_none, &
                       do_ewald_ewald, &
                       do_ewald_pme, &
                       do_ewald_spme/), &
         enum_desc=s2a("NONE standard real-space coulomb potential is computed together with the non-bonded contributions", &
                       "EWALD is the standard non-fft based ewald", &
                       "PME is the particle mesh using fft interpolation", &
                       "SPME is the smooth particle mesh using beta-Euler splines (recommended)"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EWALD_ACCURACY", &
                          description="Expected accuracy in the Ewald sum. This number affects only the calculation of "// &
                          "the cutoff for the real-space term of the ewald summation (EWALD|PME|SPME) as well as the "// &
                          "construction of the neighbor lists (if the cutoff for non-bonded terms is smaller than the "// &
                          "value employed to compute the EWALD real-space term). This keyword has no "// &
                          "effect on the reciprocal space term (which can be tuned independently).", &
                          usage="EWALD_ACCURACY {real}", n_var=1, type_of_var=real_t, &
                          unit_str="hartree", default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Explicitly provide the real-space cutoff of the ewald summation (EWALD|PME|SPME). "// &
                          "If present, overwrites the estimate of EWALD_ACCURACY and may affect the "// &
                          "construction of the neighbor lists for non-bonded terms (in FIST), if the value "// &
                          "specified is larger than the cutoff for non-bonded interactions.", &
                          usage="RCUT 5.0", n_var=1, type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="alpha", &
                          description="alpha parameter associated with Ewald (EWALD|PME|SPME). "// &
                          "Recommended for small systems is is alpha = 3.5 / r_cut. "// &
                          "Tuning alpha, r_cut and gmax is needed to obtain O(N**1.5) scaling for ewald.", &
                          usage="alpha .30", &
                          default_r_val=cp_unit_to_cp2k(value=0.35_dp, unit_str="angstrom^-1"), &
                          unit_str='angstrom^-1')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="gmax", &
                          description="number of grid points (SPME and EWALD). If a single number is specified,"// &
                          "the same number of points is used for all three directions on the grid."// &
                          "If three numbers are given, each direction can have a different number of points."// &
                          "The number of points needs to be FFTable (which depends on the library used) and odd for EWALD."// &
                          "The optimal number depends e.g. on alpha and the size of the cell. 1 point per Angstrom is common.", &
                          usage="gmax 25 25 25", n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ns_max", &
                          description="number of grid points on small mesh (PME only), should be odd.", &
                          usage="ns_max 11", default_i_val=11)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="o_spline", &
                          description="order of the beta-Euler spline (SPME only)", &
                          usage="o_spline 6", default_i_val=6)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="epsilon", &
                          description="tolerance of gaussians for fft interpolation (PME only)", &
                          usage="epsilon 1e-6", default_r_val=1.e-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL create_rsgrid_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL section_create(subsection, name="MULTIPOLES", &
                          description="Enables the use of multipoles in the treatment of the electrostatics.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE., &
                          citations=(/Aguado2003, Laino2008/))

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Controls the activation of the Multipoles", &
                          usage="&MULTIPOLES T", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_MULTIPOLE_EXPANSION", &
                          description="Specify the maximum level of multipoles expansion used "// &
                          " for the electrostatics.", &
                          usage="MAX_MULTIPOLE_EXPANSION DIPOLE", &
                          enum_c_vals=s2a("NONE", "CHARGE", "DIPOLE", "QUADRUPOLE"), &
                          enum_desc=s2a("No multipolar terms! Check the codes providing a zero contribution.", &
                                        "Use up to the Charge term", &
                                        "Use up to the Dipole term", &
                                        "Use up to the Quadrupole term"), &
                          enum_i_vals=(/do_multipole_none, do_multipole_charge, do_multipole_dipole, &
                                        do_multipole_quadrupole/), type_of_var=enum_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POL_SCF", &
                          description="Specify the method to obtain self consistent induced "// &
                          "multipole moments.", &
                          usage="POL_SCF CONJUGATE_GRADIENT", &
                          enum_c_vals=s2a("NONE", "SELF_CONSISTENT", "CONJUGATE_GRADIENT"), &
                          enum_desc=s2a("No inducible multipoles.", &
                                        "Conventional self-consistent iteration.", &
                                        "Linear conjugate-gradient optimization of the sum "// &
                                        "of the electrostatic and induction energy. This "// &
                                        "method does not support non-linear polarization "// &
                                        "but is sometimes faster."), &
                          enum_i_vals=(/do_fist_pol_none, do_fist_pol_sc, do_fist_pol_cg/), &
                          type_of_var=enum_t, default_i_val=do_fist_pol_none)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_IPOL_ITER", &
                          description="Specify the maximum number of iterations for induced "// &
                          "dipoles", &
                          usage="MAX_IPOL_ITER {int}", type_of_var=integer_t, &
                          n_var=1, default_i_val=0)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_POL", &
                          description="Specify the rmsd threshold for the derivatives "// &
                          "of the energy towards the Cartesian dipoles components", &
                          usage="EPS_POL {real}", type_of_var=real_t, &
                          n_var=1, default_r_val=0.5e-07_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL section_create(subsection, name="PRINT", &
                          description="Controls printing of Ewald properties", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)
      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="controls the printing of ewald setup", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_ewald_section

! **************************************************************************************************
!> \brief creates the interpolation section for the periodic QM/MM
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_gspace_interp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_gspace_interp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",464)
      CALL section_create(section, name="interpolator", &
                          description="controls the interpolation for the G-space term", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, print_key)

      CALL keyword_create(keyword, name="aint_precond", &
                          description="the approximate inverse to use to get the starting point"// &
                          " for the linear solver of the spline3 methods", &
                          usage="kind spline3", &
                          default_i_val=precond_spl3_aint, &
                          enum_c_vals=s2a("copy", "spl3_nopbc_aint1", "spl3_nopbc_precond1", &
                                          "spl3_nopbc_aint2", "spl3_nopbc_precond2", "spl3_nopbc_precond3"), &
                          enum_i_vals=(/no_precond, precond_spl3_aint, precond_spl3_1, &
                                        precond_spl3_aint2, precond_spl3_2, precond_spl3_3/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="precond", &
                          description="The preconditioner used"// &
                          " for the linear solver of the spline3 methods", &
                          usage="kind spline3", &
                          default_i_val=precond_spl3_3, &
                          enum_c_vals=s2a("copy", "spl3_nopbc_aint1", "spl3_nopbc_precond1", &
                                          "spl3_nopbc_aint2", "spl3_nopbc_precond2", "spl3_nopbc_precond3"), &
                          enum_i_vals=(/no_precond, precond_spl3_aint, precond_spl3_1, &
                                        precond_spl3_aint2, precond_spl3_2, precond_spl3_3/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="eps_x", &
                          description="accuracy on the solution for spline3 the interpolators", &
                          usage="eps_x 1.e-15", default_r_val=1.e-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="eps_r", &
                          description="accuracy on the residual for spline3 the interpolators", &
                          usage="eps_r 1.e-15", default_r_val=1.e-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="max_iter", &
                          variants=(/'maxiter'/), &
                          description="the maximum number of iterations", &
                          usage="max_iter 200", default_i_val=100)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "conv_info", &
                                       description="if convergence information about the linear solver"// &
                                       " of the spline methods should be printed", &
                                       print_level=medium_print_level, each_iter_names=s2a("SPLINE_FIND_COEFFS"), &
                                       each_iter_values=(/10/), filename="__STD_OUT__", &
                                       add_last=add_last_numeric)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_gspace_interp_section

! **************************************************************************************************
!> \brief Creates the wavelet section
!> \param section the section to create
!> \author fschiff
!> \note
!>      this approach is based on the development of T. Deutsch and S. Goedecker
! **************************************************************************************************
   SUBROUTINE create_wavelet_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_wavelet_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",541)
      CALL section_create( &
         section, name="wavelet", &
         description="Sets up parameters of  wavelet based poisson solver."// &
         "This solver allows for non-periodic (PERIODIC NONE) boundary conditions and slab-boundary conditions "// &
         "(but only PERIODIC XZ)."// &
         "It does not require very large unit cells, only that the density goes to zero on the faces of the cell."// &
         "The use of PREFERRED_FFT_LIBRARY FFTSG is required", &
         n_keywords=1, n_subsections=0, repeats=.FALSE., &
         citations=(/Genovese2006, Genovese2007/))

      NULLIFY (keyword)

      CALL keyword_create( &
         keyword, name="SCF_TYPE", &
         description="Type of scaling function used in the wavelet approach, the total energy depends on this choice,"// &
         "and the convergence with respect to cutoff depends on the selected scaling functions."// &
         "Possible values are 8,14,16,20,24,30,40,50,60,100 ", &
         usage="SCF_TYPE integer", &
         n_var=1, default_i_val=40)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_wavelet_section

! **************************************************************************************************
!> \brief Creates the section for the implicit (generalized) poisson solver
!> \param section the section to be created
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE create_implicit_ps_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_implicit_ps_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",580)
      CALL section_create(section, name="IMPLICIT", &
                          description="Parameters for the implicit (generalized) Poisson solver.", &
                          citations=(/BaniHashemian2016/), &
                          n_keywords=6, n_subsections=2, repeats=.FALSE.)

      NULLIFY (subsection, keyword)

      CALL create_dielectric_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_dbc_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create( &
         keyword, name="BOUNDARY_CONDITIONS", &
         enum_c_vals=s2a('PERIODIC', 'MIXED', 'MIXED_PERIODIC', 'NEUMANN'), &
         enum_desc=s2a('periodic boundary conditions', 'Dirichlet + homogeneous Neumann boundary conditions', &
                       'Dirichlet + periodic boundary conditions', 'homogeneous Neumann BC (zero-average solution)'), &
         enum_i_vals=(/PERIODIC_BC, MIXED_BC, MIXED_PERIODIC_BC, NEUMANN_BC/), &
         description="Specifies the type of boundary conditions. Dirichlet=fixed value, Neumann=zero normal deriv. "// &
         "Mixed and Neumann boundaries essentially requires FFTW3 so that all grid sizes are FFT-able.", &
         usage="BOUNDARY_CONDITIONS <bc_type>", default_i_val=PERIODIC_BC)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ZERO_INITIAL_GUESS", &
                          description="Whether or not to use zero potential as initial guess.", &
                          usage="ZERO_INITIAL_GUESS <logical>", default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="max_iter", &
                          description="Maximum number of iterations.", &
                          usage="max_iter <integer>", default_i_val=30)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="tol", &
                          description="Stopping tolerance.", &
                          usage="tol <real>", default_r_val=1.0E-8_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OR_PARAMETER", variants=s2a('omega'), &
                          description="Over-relaxation parameter (large epsilon requires smaller omega ~0.1).", &
                          usage="OR_PARAMETER <real>", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="NEUMANN_DIRECTIONS", &
         enum_c_vals=s2a('XYZ', 'XY', 'XZ', 'YZ', 'X', 'Y', 'Z'), &
         enum_i_vals=(/neumannXYZ, neumannXY, neumannXZ, neumannYZ, neumannX, neumannY, neumannZ/), &
         description="Directions in which homogeneous Neumann conditions are imposed. In the remaining directions "// &
         "periodic conditions will be enforced. Having specified MIXED or NEUMANN as BOUNDARY_CONDITIONS, "// &
         "the keyword is meant to be used to combine periodic and homogeneous Neumann conditions at the "// &
         "boundaries of the simulation cell.", &
         usage="NEUMANN_DIRECTIONS <direction>", default_i_val=neumannXYZ)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_implicit_ps_section

! **************************************************************************************************
!> \brief Creates the dielectric constant section.
!>  The dielectric constant is defined as a function of electronic density.
!>  [see O. Andreussi, I. Dabo, and N. Marzari, J. Chem. Phys., 136, 064102(2012)]
!> \param section the section to be created
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE create_dielectric_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_dielectric_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",662)
      CALL section_create(section, name="DIELECTRIC", &
                          description="Parameters for the dielectric constant function.", &
                          n_keywords=6, n_subsections=2, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="DIELECTRIC_CORE_CORRECTION", &
                          description="Avoid spurious values of the dielectric constant at the ionic core for pseudopotentials "// &
                          "where the electron density goes to zero at the core (e.g. GTH). "// &
                          "The correction is based on rho_core.", &
                          usage="DIELECTRIC_CORE_CORRECTION <logical>", default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="DIELECTRIC_FUNCTION_TYPE", &
         enum_c_vals=s2a('density_dependent', 'spatially_dependent', 'spatially_rho_dependent'), &
         enum_i_vals=(/rho_dependent, spatially_dependent, spatially_rho_dependent/), &
         enum_desc=s2a("Dielectric constant as a function of the electron density "// &
                       "as e.g. proposed within the SCCS model.", &
                       "Various regions with different dielectric constants.", &
                       "Various regions with different dielectric constants. The dielectric constant decays to 1.0, "// &
                       "wherever the electron density is present."), &
         description="Preferred type for the dielectric constant function.", &
         usage="DIELECTRIC_FUNCTION_TYPE  <method>", default_i_val=rho_dependent)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="dielectric_constant", variants=s2a('epsilon'), &
                          description="Dielectric constant in the bulk of the solvent.", &
                          usage="dielectric_constant <real>", default_r_val=80.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="rho_min", &
                          description="Lower density threshold.", &
                          usage="rho_min <real>", default_r_val=1.0E-4_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="rho_max", &
                          description="Upper density threshold.", &
                          usage="rho_max <real>", default_r_val=1.0E-3_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="DERIVATIVE_METHOD", &
         enum_c_vals=s2a('fft', 'fft_use_deps', 'fft_use_drho', 'cd3', 'cd5', 'cd7'), &
         enum_i_vals=(/derivative_fft, derivative_fft_use_deps, derivative_fft_use_drho, &
                       derivative_cd3, derivative_cd5, derivative_cd7/), &
         enum_desc=s2a("FFT based deriv of epsilon, without correction (high cutoff needed).", &
                       "FFT based deriv of epsilon, with correction using gradient of epsilon (high cutoff needed).", &
                       "FFT based deriv of epsilon, with correction using gradient of rho (high cutoff needed).", &
                       "3-point central difference derivative.", &
                       "5-point central difference derivative.", &
                       "7-point central difference derivative (recommended)."), &
         description="Preferred method for evaluating the gradient of ln(eps).", &
         usage="DERIVATIVE_METHOD  <method>", default_i_val=derivative_cd7)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_dielec_aa_cuboidal_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_dielec_xaa_annular_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_dielectric_section

! **************************************************************************************************
!> \brief Creates the section for creating axis-aligned cuboidal dielectric region.
!> \param section the section to be created
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE create_dielec_aa_cuboidal_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_dielec_aa_cuboidal_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",748)
      CALL section_create(section, name="DIELEC_AA_CUBOIDAL", &
                          description="Parameters for creating axis-aligned cuboidal dielectric region.", &
                          n_keywords=5, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="dielectric_constant", variants=s2a('epsilon'), &
                          description="value of the dielectric constant inside the region.", &
                          usage="dielectric_constant <real>", default_r_val=80.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="X_xtnt", &
                          description="The X extents of the cuboid.", &
                          usage="X_xtnt <xmin(real)> <xmax(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Y_xtnt", &
                          description="The Y extents of the cuboid.", &
                          usage="Y_xtnt <ymin(real)> <ymax(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Z_xtnt", &
                          description="The Z extents of the cuboid.", &
                          usage="Z_xtnt <zmin(real)> <zmax(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SMOOTHING_WIDTH", variants=s2a('zeta'), &
                          description="The width of the standard mollifier.", &
                          usage="SMOOTHING_WIDTH <real>", unit_str="angstrom", default_r_val=0.2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_dielec_aa_cuboidal_section

! **************************************************************************************************
!> \brief Creates the section for creating x-axis-aligned annular dielectric region.
!> \param section the section to be created
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE create_dielec_xaa_annular_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_dielec_xaa_annular_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",803)
      CALL section_create(section, name="DIELEC_XAA_ANNULAR", &
                          description="Parameters for creating x-axis-aligned annular dielectric region.", &
                          n_keywords=5, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="dielectric_constant", variants=s2a('epsilon'), &
                          description="value of the dielectric constant inside the region.", &
                          usage="dielectric_constant <real>", default_r_val=80.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="X_xtnt", &
                          description="The X extents of the annulus.", &
                          usage="X_xtnt <xmin(real)> <xmax(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="base_center", &
                          description="The y and z coordinates of the annulus' base center.", &
                          usage="base_center <y(real)> <z(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="base_radii", &
                          description="The base radius of the annulus.", &
                          usage="base_radius <r1(real)> <r2(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="smoothing_width", variants=s2a('zeta'), &
                          description="The width of the standard mollifier.", &
                          usage="smoothing_width <real>", unit_str="angstrom", default_r_val=0.2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_dielec_xaa_annular_section

! **************************************************************************************************
!> \brief Creates the section for Dirichlet boundary conditions
!> \param section the section to be created
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE create_dbc_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_dbc_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",859)
      CALL section_create(section, name="DIRICHLET_BC", &
                          description="Parameters for creating Dirichlet type boundary conditions.", &
                          n_keywords=1, n_subsections=4, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="VERBOSE_OUTPUT", &
                          description="Print out the coordinates of the vertices defining Dirichlet regions and their "// &
                          "tessellations (in Angstrom), the values of the electrostatic potential at the regions (in a.u.), "// &
                          "and their corresponding evaluated Lagrange multipliers.", &
                          usage="VERBOSE_OUTPUT <logical>", default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)

      CALL create_aa_planar_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_planar_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_xaa_cylindrical_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_aa_cuboidal_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_dbc_section

! **************************************************************************************************
!> \brief Creates the section for creating axis-aligned planar Dirichlet BC.
!> \param section the section to be created
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE create_aa_planar_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_aa_planar_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",907)
      CALL section_create(section, name="AA_PLANAR", &
                          description="Parameters for creating axis-aligned planar (rectangular) Dirichlet boundary regions.", &
                          n_keywords=9, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="v_D", &
                          description="The value of the fixed potential to be imposed at the axis-aligned Dirichlet boundary.", &
                          usage="v_D <real>", unit_str="volt", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARALLEL_PLANE", &
                          enum_c_vals=s2a('XY', 'YZ', 'XZ'), &
                          enum_i_vals=(/xy_plane, yz_plane, xz_plane/), &
                          description="The coordinate plane that the region is parallel to.", &
                          usage="PARALLEL_PLANE <plane>", &
                          type_of_var=enum_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INTERCEPT", &
                          description="The intercept of the rectangle's plane.", &
                          usage="INTERCEPT <real>", unit_str="angstrom", &
                          type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="X_xtnt", &
                          description="The X extents of the rectangle.", &
                          usage="X_xtnt <xmin(real)> <xmax(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Y_xtnt", &
                          description="The Y extents of the rectangle.", &
                          usage="Y_xtnt <ymin(real)> <ymax(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Z_xtnt", &
                          description="The Z extents of the rectangle.", &
                          usage="Z_xtnt <zmin(real)> <zmax(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_PRTN", &
                          description="The number of partitions in the directions of the unit vectors generating the "// &
                          "corresponding PARALLEL_PLANE (e1, e2 or e3) for tiling the rectangluar region.", &
                          usage="N_PRTN <integer> <integer>", &
                          n_var=2, default_i_vals=(/1, 1/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="THICKNESS", &
                          description="The thickness of the planar Dirichlet region.", &
                          usage="THICKNESS <real>", default_r_val=0.75_dp, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SMOOTHING_WIDTH", variants=s2a('SIGMA'), &
                          description="The width of the transition region between the Dirichlet subdomain and its surrounding.", &
                          usage="SMOOTHING_WIDTH <real>", default_r_val=0.2_dp, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_aa_planar_section

! **************************************************************************************************
!> \brief Creates the section for creating axis-aligned planar Dirichlet BC.
!> \param section the section to be created
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE create_planar_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_planar_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",992)
      CALL section_create( &
         section, name="PLANAR", &
         description="Parameters for creating planar (rectangular) Dirichlet boundary regions with given vertices.", &
         n_keywords=7, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="v_D", &
                          description="The value of the fixed potential to be imposed at the planar Dirichlet boundary.", &
                          usage="v_D <real>", unit_str="volt", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="A", &
                          description="Coordinates of the vertex A.", &
                          usage="A <x(real)> <y(real)> <z(real)>", unit_str="angstrom", &
                          n_var=3, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="B", &
                          description="Coordinates of the vertex B.", &
                          usage="B <x(real)> <y(real)> <z(real)>", unit_str="angstrom", &
                          n_var=3, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="C", &
                          description="Coordinates of the vertex C.", &
                          usage="C <x(real)> <y(real)> <z(real)>", unit_str="angstrom", &
                          n_var=3, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="N_PRTN", &
         description="The number of partitions along the edges for tiling the rectangular region. If the edges "// &
         "have different lengths, from the two given values, the larger one will be assigned to the longer edge.", &
         usage="N_PRTN <integer> <integer>", &
         n_var=2, default_i_vals=(/1, 1/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="THICKNESS", &
                          description="The thickness of the planar Dirichlet region.", &
                          usage="THICKNESS <real>", default_r_val=0.75_dp, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SMOOTHING_WIDTH", variants=s2a('SIGMA'), &
                          description="The width of the transition region between the Dirichlet subdomain and its surrounding.", &
                          usage="SMOOTHING_WIDTH <real>", default_r_val=0.2_dp, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_planar_section

! **************************************************************************************************
!> \brief Creates the section for creating x-axis-aligned cylindrical Dirichlet BC.
!> \param section the section to be created
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE create_xaa_cylindrical_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_xaa_cylindrical_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",1063)
      CALL section_create(section, name="XAA_CYLINDRICAL", &
                          description="Parameters for creating x-axis-aligned cylindrical Dirichlet boundary regions.", &
                          n_keywords=10, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="v_D", &
                          description="The value of the fixed potential to be imposed at the cylindrical Dirichlet boundary.", &
                          usage="v_D <real>", unit_str="volt", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="X_xtnt", &
                          description="The X extents of the cylinder.", &
                          usage="X_xtnt <xmin(real)> <xmax(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BASE_CENTER", &
                          description="The y and z coordinates of the cylinder's base center.", &
                          usage="BASE_CENTER <y(real)> <z(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BASE_RADIUS", &
                          description="The base radius of the cylinder.", &
                          usage="BASE_RADIUS <real>", default_r_val=1.0_dp, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_SIDES", &
                          description="The number of sides (faces) of the n-gonal prism approximating the cylinder.", &
                          usage="N_SIDES <integer>", default_i_val=5)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="APX_TYPE", &
                          enum_c_vals=s2a('CIRCUMSCRIBED', 'INSCRIBED'), &
                          enum_i_vals=(/CIRCUMSCRIBED, INSCRIBED/), &
                          description="Specifies the type of the n-gonal prism approximating the cylinder.", &
                          usage="APX_TYPE <apx_type>", default_i_val=CIRCUMSCRIBED)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="N_PRTN", &
         description="The number of partitions along the face edges of the prism for tiling. If the edges "// &
         "have different lengths, from the two given values, the larger one will be assigned to the longer edge.", &
         usage="N_PRTN <integer> <integer>", &
         n_var=2, default_i_vals=(/1, 1/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="THICKNESS", &
                          description="The thickness of the cylinder.", &
                          usage="THICKNESS <real>", default_r_val=0.75_dp, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SMOOTHING_WIDTH", variants=s2a('SIGMA'), &
                          description="The width of the transition region between the Dirichlet subdomain and its surrounding.", &
                          usage="SMOOTHING_WIDTH <real>", default_r_val=0.2_dp, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="delta_alpha", &
         description="A central angle specifying the gap between the faces of the n-gonal prism. To avoide overlap "// &
         "between the cuboids (of the given thickness) built on top of the faces, a larger value is required if the"// &
         " number of faces (N_SIDES) is quite few and/or the base radius is fairly small.", &
         usage="delta_alpha <real>", default_r_val=0.05_dp, unit_str="rad")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_xaa_cylindrical_section

! **************************************************************************************************
!> \brief Creates the section for creating axis-aligned cuboidal Dirichlet region.
!> \param section the section to be created
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE create_aa_cuboidal_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_aa_cuboidal_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_poisson.F",1155)
      CALL section_create(section, name="AA_CUBOIDAL", &
                          description="Parameters for creating axis-aligned cuboidal Dirichlet regions.", &
                          n_keywords=6, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="v_D", &
                          description="The value of the fixed potential to be imposed at the region.", &
                          usage="v_D <real>", unit_str="volt", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="X_xtnt", &
                          description="The X extents of the cuboid.", &
                          usage="X_xtnt <xmin(real)> <xmax(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Y_xtnt", &
                          description="The Y extents of the cuboid.", &
                          usage="Y_xtnt <ymin(real)> <ymax(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Z_xtnt", &
                          description="The Z extents of the cuboid.", &
                          usage="Z_xtnt <zmin(real)> <zmax(real)>", unit_str="angstrom", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_PRTN", &
                          description="The number of partitions in the x, y and z directions for partitioning the cuboid.", &
                          usage="N_PRTN <integer> <integer> <integer>", &
                          n_var=3, default_i_vals=(/1, 1, 1/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SMOOTHING_WIDTH", variants=s2a('SIGMA'), &
                          description="The width of the transition region between the Dirichlet subdomain and its surrounding.", &
                          usage="SMOOTHING_WIDTH <real>", default_r_val=0.2_dp, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_aa_cuboidal_section

END MODULE input_cp2k_poisson
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief A generic framework to calculate step lengths for 1D line search
!> \author Ole Schuett
! **************************************************************************************************
MODULE linesearch
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              cp_print_key_unit_nr,&
                                              low_print_level
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'linesearch'

   PUBLIC :: linesearch_type

   INTEGER, PARAMETER :: linesearch_method_adapt = 1
   INTEGER, PARAMETER :: linesearch_method_2pnt = 2
   INTEGER, PARAMETER :: linesearch_method_3pnt = 3
   INTEGER, PARAMETER :: linesearch_method_gold = 4
   INTEGER, PARAMETER :: linesearch_method_none = 5

   TYPE linesearch_2pnt_type
      REAL(KIND=dp), DIMENSION(2)                    :: energies = 0.0
      REAL(KIND=dp)                                  :: scan_step = 0.0
      REAL(KIND=dp)                                  :: last_step_size = 0.0
      REAL(KIND=dp)                                  :: max_step_size = 0.0
      INTEGER                                        :: count = 1
   END TYPE linesearch_2pnt_type

   TYPE linesearch_3pnt_type
      REAL(KIND=dp), DIMENSION(3)                    :: energies = 0.0
      REAL(KIND=dp), DIMENSION(3)                    :: scan_steps = 0.0
      REAL(KIND=dp)                                  :: last_step_size = 0.0
      REAL(KIND=dp)                                  :: max_step_size = 0.0
      REAL(KIND=dp)                                  :: tiny_step_size = 0.0
      INTEGER                                        :: count = 1
   END TYPE linesearch_3pnt_type

   TYPE linesearch_adapt_type
      REAL(KIND=dp)                                  :: last_step_size = 0.0
      REAL(KIND=dp)                                  :: left_x = 0.0
      REAL(KIND=dp)                                  :: middle_x = 0.0
      REAL(KIND=dp)                                  :: right_x = 0.0
      REAL(KIND=dp)                                  :: left_e = 0.0
      REAL(KIND=dp)                                  :: middle_e = 0.0
      REAL(KIND=dp)                                  :: right_e = 0.0
      LOGICAL                                        :: have_left = .FALSE.
      LOGICAL                                        :: have_middle = .FALSE.
      LOGICAL                                        :: have_right = .FALSE.
      INTEGER                                        :: count = 0
   END TYPE linesearch_adapt_type

   TYPE linesearch_gold_type
      REAL(KIND=dp)                                  :: scan_steps = 0.0
      REAL(KIND=dp)                                  :: last_step_size = 0.0
      REAL(KIND=dp)                                  :: eps_step_size = 0.0
      REAL(KIND=dp)                                  :: left_x = 0.0
      REAL(KIND=dp)                                  :: middle_x = 0.0
      REAL(KIND=dp)                                  :: right_x = 0.0
      REAL(KIND=dp)                                  :: left_e = 0.0
      REAL(KIND=dp)                                  :: middle_e = 0.0
      REAL(KIND=dp)                                  :: right_e = 0.0
      LOGICAL                                        :: have_left = .FALSE.
      LOGICAL                                        :: have_middle = .FALSE.
      LOGICAL                                        :: have_right = .FALSE.
      LOGICAL                                        :: gave_up = .FALSE.
   END TYPE linesearch_gold_type

   TYPE linesearch_type
      PRIVATE
      REAL(KIND=dp), PUBLIC                          :: step_size = 0.0_dp
      LOGICAL, PUBLIC                                :: starts = .FALSE.
      TYPE(linesearch_adapt_type), POINTER           :: state_adapt => Null()
      TYPE(linesearch_2pnt_type), POINTER            :: state_2pnt => Null()
      TYPE(linesearch_3pnt_type), POINTER            :: state_3pnt => Null()
      TYPE(linesearch_gold_type), POINTER            :: state_gold => Null()
      INTEGER                                        :: iw = -1
      INTEGER                                        :: method = -1
      CHARACTER(LEN=10)                              :: label = ""
      REAL(KIND=dp)                                  :: init_step_size = 0.0_dp
      REAL(dp)                                       :: eps_step_size = 0.0_dp
      REAL(dp)                                       :: max_step_size = 0.0_dp
      REAL(dp)                                       :: tiny_step_size = 0.0_dp
   END TYPE linesearch_type

   PUBLIC :: linesearch_create_section, linesearch_init, linesearch_finalize
   PUBLIC :: linesearch_step, linesearch_reset

CONTAINS

! **************************************************************************************************
!> \brief Declare the line search input section.
!> \param section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE linesearch_create_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'linesearch_create_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_section, printkey

      NULLIFY (keyword, print_section, printkey)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("linesearch.F",130)
      CALL section_create(section, "LINE_SEARCH", repeats=.FALSE., &
                          description="Detail settings or linesearch method.")

      CALL keyword_create( &
         keyword, name="METHOD", &
         description="Linesearch method.", &
         default_i_val=linesearch_method_adapt, &
         enum_c_vals=s2a("ADAPT", "3PNT", "2PNT", "GOLD", "NONE"), &
         enum_desc=s2a("extrapolates usually based on 3 points, uses additional points on demand, very robust.", &
                       "extrapolate based on 3 points", &
                       "extrapolate based on 2 points and the slope (super fast, but might get stuck at saddle points)", &
                       "perform 1D golden section search of the minimum (very expensive)", &
                       "always take steps of fixed INITIAL_STEP_SIZE"), &
         enum_i_vals=(/linesearch_method_adapt, linesearch_method_3pnt, linesearch_method_2pnt, &
                       linesearch_method_gold, linesearch_method_none/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INITIAL_STEP_SIZE", &
                          description="Initial step length", &
                          default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_STEP_SIZE", &
                          description="Maximum step length", &
                          default_r_val=3.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TINY_STEP_SIZE", &
                          description="Step length taken if negative step is suggested.", &
                          default_r_val=0.002_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_STEP_SIZE", &
                          description="Convergence criterion of GOLD method.", &
                          default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(print_section, name="PRINT", &
                          description="Print section", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)

      CALL cp_print_key_section_create(printkey, "RUN_INFO", &
                                       description="General run information", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")

      CALL section_add_subsection(print_section, printkey)
      CALL section_release(printkey)
      CALL section_add_subsection(section, print_section)
      CALL section_release(print_section)

   END SUBROUTINE linesearch_create_section

! **************************************************************************************************
!> \brief Initialize linesearch from given input section
!> \param this ...
!> \param section ...
!> \param label ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE linesearch_init(this, section, label)
      TYPE(linesearch_type), INTENT(INOUT)               :: this
      TYPE(section_vals_type), POINTER                   :: section
      CHARACTER(LEN=*)                                   :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'linesearch_init', &
         routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: logger

      CALL section_vals_val_get(section, "METHOD", i_val=this%method)
      CALL section_vals_val_get(section, "INITIAL_STEP_SIZE", r_val=this%init_step_size)
      CALL section_vals_val_get(section, "MAX_STEP_SIZE", r_val=this%max_step_size)
      CALL section_vals_val_get(section, "TINY_STEP_SIZE", r_val=this%tiny_step_size)
      CALL section_vals_val_get(section, "EPS_STEP_SIZE", r_val=this%eps_step_size)

      IF(.NOT.(LEN_TRIM(label) <= 10))CALL cp__a("linesearch.F",211)
      this%label = label
      logger => cp_get_default_logger()
      this%iw = cp_print_key_unit_nr(logger, section, "PRINT%RUN_INFO", &
                                     extension=".linesearchlog")

      CALL linesearch_init_low(this)

   END SUBROUTINE linesearch_init

! **************************************************************************************************
!> \brief Helper routine to (re)-initialize line search machinery
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE linesearch_init_low(this)
      TYPE(linesearch_type), INTENT(INOUT)               :: this

      this%step_size = 0.0_dp
      this%starts = .TRUE.

      SELECT CASE (this%method)
      CASE (linesearch_method_adapt)
         ALLOCATE (this%state_adapt)
         this%state_adapt%last_step_size = this%init_step_size
      CASE (linesearch_method_2pnt)
         ALLOCATE (this%state_2pnt)
         this%state_2pnt%max_step_size = this%max_step_size
         this%state_2pnt%last_step_size = this%init_step_size
      CASE (linesearch_method_3pnt)
         ALLOCATE (this%state_3pnt)
         this%state_3pnt%last_step_size = this%init_step_size
         this%state_3pnt%max_step_size = this%max_step_size
         this%state_3pnt%tiny_step_size = this%tiny_step_size
      CASE (linesearch_method_gold)
         ALLOCATE (this%state_gold)
         this%state_gold%last_step_size = this%init_step_size
         this%state_gold%eps_step_size = this%eps_step_size
      CASE (linesearch_method_none)
         ! nothing todo
      CASE DEFAULT
         CALL cp__b("linesearch.F",252,"unkown method")
      END SELECT

   END SUBROUTINE linesearch_init_low

! **************************************************************************************************
!> \brief Finzalize line search machinery
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE linesearch_finalize(this)
      TYPE(linesearch_type), INTENT(INOUT)               :: this

      CHARACTER(len=*), PARAMETER :: routineN = 'linesearch_finalize', &
         routineP = moduleN//':'//routineN

      SELECT CASE (this%method)
      CASE (linesearch_method_adapt)
         DEALLOCATE (this%state_adapt)
      CASE (linesearch_method_2pnt)
         DEALLOCATE (this%state_2pnt)
      CASE (linesearch_method_3pnt)
         DEALLOCATE (this%state_3pnt)
      CASE (linesearch_method_gold)
         DEALLOCATE (this%state_gold)
      CASE (linesearch_method_none)
         ! nothing todo
      CASE DEFAULT
         CALL cp__b("linesearch.F",280,"unkown method")
      END SELECT

      !TODO: should finish printkey, but don't have the section here
   END SUBROUTINE linesearch_finalize

! **************************************************************************************************
!> \brief Reset line search to initial state
!> \param this ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE linesearch_reset(this)
      TYPE(linesearch_type), INTENT(INOUT)               :: this

      CALL linesearch_finalize(this)
      CALL linesearch_init_low(this)
   END SUBROUTINE linesearch_reset

! **************************************************************************************************
!> \brief Calculate step length of next line search step.
!> \param this ...
!> \param energy ...
!> \param slope ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE linesearch_step(this, energy, slope)
      TYPE(linesearch_type), INTENT(INOUT)               :: this
      REAL(KIND=dp), INTENT(IN)                          :: energy, slope

      CHARACTER(len=*), PARAMETER :: routineN = 'linesearch_step', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: is_done
      REAL(KIND=dp)                                      :: step_size

      SELECT CASE (this%method)
      CASE (linesearch_method_adapt)
         CALL linesearch_adapt(this%state_adapt, energy, step_size, is_done, this%iw, TRIM(this%label))
      CASE (linesearch_method_2pnt)
         CALL linesearch_2pnt(this%state_2pnt, energy, slope, step_size, is_done, this%iw, TRIM(this%label))
      CASE (linesearch_method_3pnt)
         CALL linesearch_3pnt(this%state_3pnt, energy, step_size, is_done, this%iw, TRIM(this%label))
      CASE (linesearch_method_gold)
         CALL linesearch_gold(this%state_gold, energy, step_size, is_done, this%iw, TRIM(this%label))
      CASE (linesearch_method_none)
         step_size = this%init_step_size ! take steps of fixed length
         is_done = .TRUE.
      CASE DEFAULT
         CALL cp__b("linesearch.F",328,"unkown method")
      END SELECT

      this%step_size = step_size
      this%starts = is_done
   END SUBROUTINE linesearch_step

! **************************************************************************************************
!> \brief Perform a 2pnt linesearch
!> \param this ...
!> \param energy ...
!> \param slope ...
!> \param step_size ...
!> \param is_done ...
!> \param unit_nr ...
!> \param label ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE linesearch_2pnt(this, energy, slope, step_size, is_done, unit_nr, label)
      TYPE(linesearch_2pnt_type), INTENT(INOUT)          :: this
      REAL(KIND=dp), INTENT(IN)                          :: energy, slope
      REAL(KIND=dp), INTENT(OUT)                         :: step_size
      LOGICAL, INTENT(OUT)                               :: is_done
      INTEGER, INTENT(IN)                                :: unit_nr
      CHARACTER(len=*), INTENT(IN)                       :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'linesearch_2pnt', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: a, b, c, pred_energy, x2

      this%energies(this%count) = energy
      is_done = .FALSE.

      SELECT CASE (this%count)
      CASE (1)
         step_size = 2.0_dp*this%last_step_size
         this%scan_step = step_size
         this%count = 2
      CASE (2)
         c = this%energies(1)
         b = -slope
         x2 = this%scan_step
         a = (this%energies(2)-b*x2-c)/(x2**2)

         IF (a < 0.0_dp) THEN
            IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| had to quench curvature"
            a = 1.0E-15_dp
         ENDIF

         step_size = -b/(2.0_dp*a)
         pred_energy = a*step_size**2+b*step_size+c

         IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| 2pnt suggested step_size: ", step_size
         IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| 2pnt predicted energy", pred_energy

         IF (pred_energy > this%energies(1) .OR. pred_energy > this%energies(2)) THEN
            CALL cp__b("linesearch.F",385,label//"LS| predicted energy not below test points")
         ENDIF

         IF (step_size > this%max_step_size) THEN
            step_size = this%max_step_size
            IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| limiting step_size to MAX_STEP_SIZE"
         ENDIF

         this%last_step_size = step_size
         this%count = 1
         is_done = .TRUE.
      CASE DEFAULT
         CALL cp__b("linesearch.F",397,"this should not happen")
      END SELECT

   END SUBROUTINE linesearch_2pnt

! **************************************************************************************************
!> \brief Perform a 3pnt linesearch
!> \param this ...
!> \param energy ...
!> \param step_size ...
!> \param is_done ...
!> \param unit_nr ...
!> \param label ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE linesearch_3pnt(this, energy, step_size, is_done, unit_nr, label)
      TYPE(linesearch_3pnt_type), INTENT(INOUT)          :: this
      REAL(KIND=dp), INTENT(IN)                          :: energy
      REAL(KIND=dp), INTENT(OUT)                         :: step_size
      LOGICAL, INTENT(OUT)                               :: is_done
      INTEGER, INTENT(IN)                                :: unit_nr
      CHARACTER(len=*), INTENT(IN)                       :: label

      REAL(KIND=dp)                                      :: a, b, c, denom, pred_energy, x1, x2, x3, &
                                                            y1, y2, y3

      this%energies(this%count) = energy
      is_done = .FALSE.

      SELECT CASE (this%count)
      CASE (1)
         step_size = (2.0_dp/3.0_dp)*this%last_step_size
         IF (step_size < this%tiny_step_size) THEN
            IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| initial step size too small, using TINY_STEP_SIZE"
            step_size = this%tiny_step_size
         ENDIF
         this%scan_steps(1) = 0.0_dp
         this%scan_steps(2) = step_size
         this%count = 2
      CASE (2)
         IF (this%energies(1) > this%energies(2)) THEN
            step_size = 2.0_dp*this%scan_steps(2)
         ELSE
            step_size = 0.5_dp*this%scan_steps(2)
         END IF
         this%scan_steps(3) = step_size
         this%count = 3
      CASE (3)
         ! fitting y = a*x^2 + b*x + c
         y1 = this%energies(1); y2 = this%energies(2); y3 = this%energies(3)
         x1 = this%scan_steps(1); x2 = this%scan_steps(2); x3 = this%scan_steps(3)

         IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| 3pnt scan_steps: ", this%scan_steps
         IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| 3pnt energies: ", this%energies

         ! Cramer's Rule
         denom = (x1-x2)*(x1-x3)*(x2-x3)
         a = (x3*(y2-y1)+x2*(y1-y3)+x1*(y3-y2))/denom
         b = (x3**2*(y1-y2)+x2**2*(y3-y1)+x1**2*(y2-y3))/denom
         c = (x2*x3*(x2-x3)*y1+x3*x1*(x3-x1)*y2+x1*x2*(x1-x2)*y3)/denom

         step_size = -b/(2.0_dp*a)
         pred_energy = a*step_size**2+b*step_size+c
         IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| 3pnt suggested step_size: ", step_size
         IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| 3pnt predicted energy", pred_energy

         IF (a < 0) THEN
            step_size = -2.0_dp*step_size
            IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| inverting step size"
         END IF

         IF (step_size < 0) THEN
            step_size = this%tiny_step_size
            IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| makeing a step of size TINY_STEP_SIZE"
         ENDIF

         IF (step_size > this%max_step_size) THEN
            step_size = this%max_step_size
            IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| limiting step_size to MAX_STEP_SIZE"
         ENDIF

         this%last_step_size = step_size
         this%count = 1
         is_done = .TRUE.
      CASE DEFAULT
         CALL cp__b("linesearch.F",482,"this should not happen")
      END SELECT

   END SUBROUTINE linesearch_3pnt

! **************************************************************************************************
!> \brief Perform an adaptive linesearch
!> \param this ...
!> \param energy ...
!> \param step_size ...
!> \param is_done ...
!> \param unit_nr ...
!> \param label ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE linesearch_adapt(this, energy, step_size, is_done, unit_nr, label)
      TYPE(linesearch_adapt_type), INTENT(INOUT)         :: this
      REAL(KIND=dp), INTENT(IN)                          :: energy
      REAL(KIND=dp), INTENT(OUT)                         :: step_size
      LOGICAL, INTENT(OUT)                               :: is_done
      INTEGER, INTENT(IN)                                :: unit_nr
      CHARACTER(len=*), INTENT(IN)                       :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'linesearch_adapt', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER                           :: grow_factor = 2.0_dp, &
                                                            shrink_factor = 0.5_dp

      REAL(KIND=dp)                                      :: a, b, c, denom, pred_energy, x1, x2, x3, &
                                                            y1, y2, y3

      is_done = .FALSE.
      this%count = this%count+1

      IF (.NOT. this%have_left) THEN
         this%left_x = 0.0_dp
         this%left_e = energy
         this%have_left = .TRUE.
         step_size = this%last_step_size

      ELSE IF (.NOT. (this%have_middle .OR. this%have_right)) THEN
         IF (energy < this%left_e) THEN
            this%middle_e = energy
            this%middle_x = this%last_step_size
            this%have_middle = .TRUE.
            step_size = this%middle_x*grow_factor
         ELSE
            this%right_e = energy
            this%right_x = this%last_step_size
            this%have_right = .TRUE.
            step_size = this%right_x*shrink_factor
         ENDIF

      ELSE IF (.NOT. this%have_right) THEN
         IF (energy < this%middle_e) THEN
            this%middle_e = energy
            this%middle_x = this%last_step_size
            step_size = this%middle_x*grow_factor
         ELSE
            this%right_e = energy
            this%right_x = this%last_step_size
            this%have_right = .TRUE.
         ENDIF

      ELSE IF (.NOT. this%have_middle) THEN
         IF (energy > this%left_e) THEN
            this%right_e = energy
            this%right_x = this%last_step_size
            step_size = this%right_x*shrink_factor
         ELSE
            this%middle_e = energy
            this%middle_x = this%last_step_size
            this%have_middle = .TRUE.
         ENDIF
      ENDIF

      IF (this%count > 3) THEN
         IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| Need extra step"
      ENDIF
      IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| adapt: ", this%have_left, this%have_middle, this%have_right
      IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| adapt: scan_steps: ", this%left_x, this%middle_x, this%right_x
      IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| adapt: energies: ", this%left_e, this%middle_e, this%right_e

      IF (this%have_left .AND. this%have_middle .AND. this%have_right) THEN
         ! fitting y = a*x^2 + b*x + c
         y1 = this%left_e; y2 = this%middle_e; y3 = this%right_e
         x1 = this%left_x; x2 = this%middle_x; x3 = this%right_x

         ! Cramer's rule
         denom = (x1-x2)*(x1-x3)*(x2-x3)
         a = (x3*(y2-y1)+x2*(y1-y3)+x1*(y3-y2))/denom
         b = (x3**2*(y1-y2)+x2**2*(y3-y1)+x1**2*(y2-y3))/denom
         c = (x2*x3*(x2-x3)*y1+x3*x1*(x3-x1)*y2+x1*x2*(x1-x2)*y3)/denom

         IF (ABS(a) /= 0.0_dp) THEN
            step_size = -b/(2.0_dp*a)
         ELSE
            step_size = 0.0_dp
         ENDIF

         IF(.NOT.(step_size >= 0.0_dp))CALL cp__a("linesearch.F",582)
         pred_energy = a*step_size**2+b*step_size+c
         IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| adapt: suggested step_size: ", step_size
         IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| adapt: predicted energy", pred_energy

         ! reset
         is_done = .TRUE.
         this%count = 0
         this%have_left = .FALSE.
         this%have_middle = .FALSE.
         this%have_right = .FALSE.
         this%left_e = 0.0
         this%middle_e = 0.0
         this%right_e = 0.0
         this%left_x = 0.0
         this%middle_x = 0.0
         this%right_x = 0.0
      END IF

      this%last_step_size = step_size
   END SUBROUTINE linesearch_adapt

! **************************************************************************************************
!> \brief Perform a gold linesearch
!> \param this ...
!> \param energy ...
!> \param step_size ...
!> \param is_done ...
!> \param unit_nr ...
!> \param label ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE linesearch_gold(this, energy, step_size, is_done, unit_nr, label)
      TYPE(linesearch_gold_type), INTENT(INOUT)          :: this
      REAL(KIND=dp), INTENT(IN)                          :: energy
      REAL(KIND=dp), INTENT(OUT)                         :: step_size
      LOGICAL, INTENT(OUT)                               :: is_done
      INTEGER, INTENT(IN)                                :: unit_nr
      CHARACTER(len=*), INTENT(IN)                       :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'linesearch_gold', &
         routineP = moduleN//':'//routineN
      REAL(KIND=dp), PARAMETER :: phi = (1.0_dp+SQRT(5.0_dp))/2.0_dp

      REAL(KIND=dp)                                      :: a, b, d

      is_done = .FALSE.

      IF (this%gave_up) &
         CALL cp__b("linesearch.F",631,"had to give up, should not be called again")

      IF (.NOT. this%have_left) THEN
         this%left_x = 0.0_dp
         this%left_e = energy
         this%have_left = .TRUE.
         step_size = this%last_step_size

      ELSE IF (.NOT. (this%have_middle .OR. this%have_right)) THEN
         IF (energy < this%left_e) THEN
            this%middle_e = energy
            this%middle_x = this%scan_steps
            this%have_middle = .TRUE.
            step_size = this%middle_x*phi
         ELSE
            this%right_e = energy
            this%right_x = this%scan_steps
            this%have_right = .TRUE.
            step_size = this%right_x/phi
         ENDIF

      ELSE IF (.NOT. this%have_right) THEN
         IF (energy < this%middle_e) THEN
            this%middle_e = energy
            this%middle_x = this%scan_steps
            step_size = this%middle_x*phi
         ELSE
            this%right_e = energy
            this%right_x = this%scan_steps
            this%have_right = .TRUE.
         ENDIF

      ELSE IF (.NOT. this%have_middle) THEN
         IF (energy > this%left_e) THEN
            this%right_e = energy
            this%right_x = this%scan_steps
            step_size = this%right_x/phi
         ELSE
            this%middle_e = energy
            this%middle_x = this%scan_steps
            this%have_middle = .TRUE.
         ENDIF

      ELSE !up and running
         a = this%middle_x-this%left_x
         b = this%right_x-this%middle_x
         IF (energy < this%middle_e) THEN
            IF (a < b) THEN
               this%left_e = this%middle_e
               this%left_x = this%middle_x
            ELSE
               this%right_e = this%middle_e
               this%right_x = this%middle_x
            ENDIF
            this%middle_e = energy
            this%middle_x = this%scan_steps
         ELSE
            IF (a < b) THEN
               this%right_e = energy
               this%right_x = this%scan_steps
            ELSE
               this%left_e = energy
               this%left_x = this%scan_steps
            ENDIF
         ENDIF
      ENDIF

      IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| gold: ", this%have_left, this%have_middle, this%have_right
      IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| gold: ", this%left_x, this%middle_x, this%right_x
      IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| gold: ", this%left_e, this%middle_e, this%right_e

      IF (this%have_left .AND. this%have_middle .AND. this%have_right) THEN
         a = this%middle_x-this%left_x
         b = this%right_x-this%middle_x
         IF (ABS(MIN(a, b)*phi-MAX(a, b)) > 1.0E-10) &
            CALL cp__b("linesearch.F",706,"golden-ratio gone")

         IF (a < b) THEN
            step_size = this%middle_x+a/phi
         ELSE
            step_size = this%middle_x-b/phi
         END IF

         d = ABS(this%right_x-this%left_x)/(ABS(this%middle_x)+ABS(step_size))
         IF (d < this%eps_step_size) THEN
            step_size = this%middle_x
            this%last_step_size = step_size
            is_done = .TRUE.

            IF (unit_nr > 0) WRITE (unit_nr, *) label, "LS| gold done, step-size: ", step_size

            this%have_left = .FALSE.
            this%have_middle = .FALSE.
            this%have_right = .FALSE.
            this%left_e = 0.0
            this%middle_e = 0.0
            this%right_e = 0.0
            this%left_x = 0.0
            this%middle_x = 0.0
            this%right_x = 0.0
         ENDIF

      END IF

      IF (step_size < 1E-10) CALL cp__b("linesearch.F",735,"linesearch failed / done")

      this%scan_steps = step_size
   END SUBROUTINE linesearch_gold

END MODULE linesearch
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types used by the PAO machinery
!> \author Ole Schuett
! **************************************************************************************************
MODULE pao_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_type
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE linesearch,                      ONLY: linesearch_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_types'

   PUBLIC :: pao_env_type, training_matrix_type

   TYPE pao_env_type
      ! input values
      REAL(KIND=dp)                    :: eps_pao = 0.0_dp
      REAL(KIND=dp)                    :: cg_reset_limit = 0.1_dp
      REAL(KIND=dp)                    :: mixing = 0.0_dp
      REAL(KIND=dp)                    :: penalty_dist = 0.0_dp
      REAL(KIND=dp)                    :: penalty_strength = 0.0_dp
      REAL(KIND=dp)                    :: check_unitary_tol = 0.0_dp
      REAL(KIND=dp)                    :: check_grad_tol = 0.0_dp
      REAL(KIND=dp)                    :: num_grad_eps = 0.0_dp
      INTEGER                          :: num_grad_order = -1
      INTEGER                          :: max_pao = -1
      INTEGER                          :: max_cycles = -1
      INTEGER                          :: write_cycles = -1
      INTEGER                          :: parameterization = -1
      INTEGER                          :: optimizer = -1
      INTEGER                          :: cg_init_steps = -1
      CHARACTER(LEN=default_path_length) :: preopt_dm_file = ""
      CHARACTER(LEN=default_path_length) :: restart_file = ""
      CHARACTER(LEN=default_path_length) :: ml_training_file = ""

      INTEGER                          :: ml_method = -1
      REAL(KIND=dp)                    :: ml_tolerance = 0.0_dp
      REAL(KIND=dp)                    :: gp_noise_var = 0.0_dp
      REAL(KIND=dp)                    :: gp_scale = 0.0_dp

      ! output units
      INTEGER                          :: iw = -1
      INTEGER                          :: iw_opt = -1

      ! state variable
      INTEGER                          :: istep = -1
      REAL(KIND=dp)                    :: energy_prev = 0.0_dp
      REAL(KIND=dp)                    :: step_start_time = 0.0_dp
      TYPE(linesearch_type)            :: linesearch
      LOGICAL                          :: need_initial_scf = .FALSE.
      LOGICAL                          :: precondition = .FALSE.

      ! matrices
      TYPE(cp_dbcsr_type)              :: matrix_X
      TYPE(cp_dbcsr_type)              :: matrix_U
      TYPE(cp_dbcsr_type)              :: matrix_U0
      TYPE(cp_dbcsr_type)              :: matrix_H0
      TYPE(cp_dbcsr_type)              :: matrix_Y
      TYPE(cp_dbcsr_type)              :: matrix_N
      TYPE(cp_dbcsr_type)              :: matrix_N_inv
      TYPE(cp_dbcsr_type)              :: matrix_X_orig
      TYPE(cp_dbcsr_type)              :: matrix_G
      TYPE(cp_dbcsr_type)              :: matrix_G_prev
      TYPE(cp_dbcsr_type)              :: matrix_D
      TYPE(cp_dbcsr_type)              :: matrix_D_preconed
      TYPE(cp_dbcsr_type)              :: matrix_G_preconed
      TYPE(cp_dbcsr_type)              :: matrix_V_terms
      TYPE(cp_dbcsr_type)              :: matrix_BFGS
      TYPE(cp_dbcsr_type)              :: matrix_precon
      TYPE(cp_dbcsr_type)              :: matrix_precon_inv

   END TYPE

   TYPE training_matrix_type
      CHARACTER(LEN=default_string_length)      :: kindname = ""
      REAL(dp), DIMENSION(:, :), ALLOCATABLE    :: inputs
      REAL(dp), DIMENSION(:, :), ALLOCATABLE    :: outputs
      REAL(dp), DIMENSION(:, :, :), ALLOCATABLE :: NN ! Neural Network
      REAL(dp), DIMENSION(:, :), ALLOCATABLE    :: GP ! Gaussian Process
   END TYPE training_matrix_type

END MODULE pao_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE pao_input
   USE bibliography,                    ONLY: Berghold2011
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              cp_print_key_unit_nr,&
                                              high_print_level,&
                                              low_print_level
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE linesearch,                      ONLY: linesearch_create_section
   USE pao_types,                       ONLY: pao_env_type
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pao_input'

   PUBLIC :: create_pao_section, parse_pao_section, id2str

   INTEGER, PARAMETER, PUBLIC               :: pao_rotinv_param = 101, &
                                               pao_fock_param = 102, &
                                               pao_exp_param = 103, &
                                               pao_gth_param = 104, &
                                               pao_gthexp_param = 105, &
                                               pao_opt_cg = 301, &
                                               pao_opt_bfgs = 302, &
                                               pao_ml_gp = 401, &
                                               pao_ml_nn = 402

CONTAINS

! **************************************************************************************************
!> \brief Declare the PAO input section
!> \param pao ...
!> \param input ...
! **************************************************************************************************
   SUBROUTINE parse_pao_section(pao, input)
      TYPE(pao_env_type), POINTER                        :: pao
      TYPE(section_vals_type), POINTER                   :: input

      CHARACTER(len=*), PARAMETER :: routineN = 'parse_pao_section', &
         routineP = moduleN//':'//routineN

      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)

      ! open main logger
      logger => cp_get_default_logger()
      pao%iw = cp_print_key_unit_nr(logger, input, "DFT%LS_SCF%PAO%PRINT%RUN_INFO", extension=".paolog")
      pao%iw_opt = cp_print_key_unit_nr(logger, input, "DFT%LS_SCF%PAO%PRINT%OPT_INFO", extension=".paooptlog")

      IF (pao%iw > 0) WRITE (pao%iw, *) "" ! an empty separator line

      ! parse input and print
      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%EPS_PAO", r_val=pao%eps_pao)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,E10.1)") " PAO|", "EPS_PAO", pao%eps_pao

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%MIXING", r_val=pao%mixing)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,E10.1)") " PAO|", "MIXING", pao%mixing

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%MAX_PAO", i_val=pao%max_pao)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,I10)") " PAO|", "MAX_PAO", pao%max_pao

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%MAX_CYCLES", i_val=pao%max_cycles)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,I10)") " PAO|", "MAX_CYCLES", pao%max_cycles

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%PARAMETERIZATION", i_val=pao%parameterization)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,A10)") " PAO|", "PARAMETERIZATION", id2str(pao%parameterization)

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%PRECONDITION", l_val=pao%precondition)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,L10)") " PAO|", "PRECONDITION", pao%precondition

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%PENALTY_DISTANCE", r_val=pao%penalty_dist)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,E10.1)") " PAO|", "PENALTY_DISTANCE", pao%penalty_dist
      IF (pao%penalty_dist < 0.0_dp) CALL cp__b("pao_input.F",94,"PAO: PENALTY_DISTANCE < 0")

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%PENALTY_STRENGTH", r_val=pao%penalty_strength)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,E10.1)") " PAO|", "PENALTY_STRENGTH", pao%penalty_strength
      IF (pao%penalty_strength < 0.0_dp) CALL cp__b("pao_input.F",98,"PAO: PENALTY_STRENGTH < 0")

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%OPTIMIZER", i_val=pao%optimizer)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,A10)") " PAO|", "OPTIMIZER", id2str(pao%optimizer)

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%CG_INIT_STEPS", i_val=pao%cg_init_steps)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,I10)") " PAO|", "CG_INIT_STEPS", pao%cg_init_steps
      IF (pao%cg_init_steps < 1) CALL cp__b("pao_input.F",105,"PAO: CG_INIT_STEPS < 1")

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%CG_RESET_LIMIT", r_val=pao%cg_reset_limit)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,E10.1)") " PAO|", "CG_RESET_LIMIT", pao%cg_reset_limit
      IF (pao%cg_reset_limit < 0.0_dp) CALL cp__b("pao_input.F",109,"PAO: CG_RESET_LIMIT < 0")

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%CHECK_UNITARY_TOL", r_val=pao%check_unitary_tol)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,E10.1)") " PAO|", "CHECK_UNITARY_TOL", pao%check_unitary_tol

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%CHECK_GRADIENT_TOL", r_val=pao%check_grad_tol)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,E10.1)") " PAO|", "CHECK_GRADIENT_TOL", pao%check_grad_tol

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%NUM_GRADIENT_ORDER", i_val=pao%num_grad_order)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,I10)") " PAO|", "NUM_GRADIENT_ORDER", pao%num_grad_order

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%NUM_GRADIENT_EPS", r_val=pao%num_grad_eps)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,E10.1)") " PAO|", "NUM_GRADIENT_EPS", pao%num_grad_eps
      IF (pao%num_grad_eps < 0.0_dp) CALL cp__b("pao_input.F",122,"PAO: NUM_GRADIENT_EPS < 0")

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%PRINT%RESTART%WRITE_CYCLES", i_val=pao%write_cycles)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,I10)") " PAO|", "PRINT%RESTART%WRITE_CYCLES", pao%write_cycles

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%RESTART_FILE", c_val=pao%restart_file)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,A)") " PAO|", "RESTART_FILE ", TRIM(pao%restart_file)

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%PREOPT_DM_FILE", c_val=pao%preopt_dm_file)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,A)") " PAO|", "PREOPT_DM_FILE ", TRIM(pao%preopt_dm_file)

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%MACHINE_LEARNING%TRAINING_FILE", c_val=pao%ml_training_file)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,A)") " PAO|", "MACHINE_LEARNING%TRAINING_FILE ", TRIM(pao%ml_training_file)

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%MACHINE_LEARNING%METHOD", i_val=pao%ml_method)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,A10)") " PAO|", "MACHINE_LEARNING%METHOD", id2str(pao%ml_method)

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%MACHINE_LEARNING%TOLERANCE", r_val=pao%ml_tolerance)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,E10.1)") " PAO|", "MACHINE_LEARNING%TOLERANCE", pao%ml_tolerance

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%MACHINE_LEARNING%GP_NOISE_VAR", r_val=pao%gp_noise_var)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,E10.1)") " PAO|", "MACHINE_LEARNING%GP_NOISE_VAR", pao%gp_noise_var

      CALL section_vals_val_get(input, "DFT%LS_SCF%PAO%MACHINE_LEARNING%GP_SCALE", r_val=pao%gp_scale)
      IF (pao%iw > 0) WRITE (pao%iw, "(A,T40,A,T71,E10.1)") " PAO|", "MACHINE_LEARNING%GP_SCALE", pao%gp_scale

      IF (pao%iw > 0) WRITE (pao%iw, *) "" ! an empty separator line

   END SUBROUTINE parse_pao_section

! **************************************************************************************************
!> \brief Helper routine
!> \param id ...
!> \retval s ...
! **************************************************************************************************
   FUNCTION id2str(id) RESULT(s)
      INTEGER                                            :: id
      CHARACTER(LEN=10)                                  :: s

      CHARACTER(len=*), PARAMETER :: routineN = 'id2str', routineP = moduleN//':'//routineN

      SELECT CASE (id)
      CASE (pao_gthexp_param)
         s = "GTHEXP"
      CASE (pao_gth_param)
         s = "GTH"
      CASE (pao_rotinv_param)
         s = "ROTINV"
      CASE (pao_fock_param)
         s = "FOCK"
      CASE (pao_exp_param)
         s = "EXP"
      CASE (pao_opt_cg)
         s = "CG"
      CASE (pao_opt_bfgs)
         s = "BFGS"
      CASE (pao_ml_gp)
         s = "GAUSSIAN_PROCESS"
      CASE (pao_ml_nn)
         s = "NEURAL_NETWORK"
      CASE DEFAULT
         CALL cp__b("pao_input.F",183,"PAO: unknown id")
      END SELECT
      s = ADJUSTR(s)
   END FUNCTION id2str

! **************************************************************************************************
!> \brief Creates the PAO subsection of the linear scaling section.
!> \param section ...
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE create_pao_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_pao_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: printkey, subsection

      NULLIFY (keyword, subsection, printkey)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("pao_input.F",204)
      CALL section_create(section, "PAO", repeats=.FALSE., &
                          description="Polarized Atomic Orbital Method", &
                          citations=(/Berghold2011/))

      ! Convergence Criteria *****************************************************
      CALL keyword_create(keyword, name="EPS_PAO", &
                          description="Convergence criteria for PAO optimization.", &
                          default_r_val=1.e-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIXING", &
                          description="Mixing fraction of new and old optimizied solutions.", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_PAO", &
                          description="Maximum number of PAO basis optimization steps.", &
                          default_i_val=1000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_CYCLES", &
                          description="Maximum number of PAO line search cycles for a given hamiltonian.", &
                          default_i_val=1000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Parametrization **********************************************************
      CALL keyword_create(keyword, name="PARAMETERIZATION", &
                          description="Parametrization of the mapping between the primary and the PAO basis.", &
                          enum_c_vals=s2a("GTH", "GTHEXP", "ROTINV", "FOCK", "EXP"), &
                          enum_i_vals=(/pao_gth_param, pao_gthexp_param, pao_rotinv_param, &
                                        pao_fock_param, pao_exp_param/), &
                          enum_desc=s2a("Parametrization based on GTH pseudo potentials", &
                                        "Parametrization based on GTH pseudo potentials and matrix exponential", &
                                        "Rotational invariant parametrization", &
                                        "Fock matrix parametrization", &
                                        "Original matrix exponential parametrization"), &
                          default_i_val=pao_rotinv_param)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PENALTY_DISTANCE", &
                          description="Distance at which approaching eigenvalues are penalized to prevent degeneration.", &
                          default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PENALTY_STRENGTH", &
                          description="Strength of the penalty term which prevents degenerate eigenvalues.", &
                          default_r_val=0.005_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRECONDITION", &
                          description="Apply a preconditioner to the parametrization.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Preopt  ******************************************************************
      CALL keyword_create(keyword, name="PREOPT_DM_FILE", &
                          description="Read pre-optimized density matrix from given file.", &
                          repeats=.FALSE., default_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Misc ********************************************************************
      CALL keyword_create(keyword, name="RESTART_FILE", &
                          description="Reads given files as restart for PAO basis", &
                          repeats=.FALSE., default_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHECK_GRADIENT_TOL", &
                          description="Tolerance for check of full analytic gradient against the numeric one."// &
                          " Negative values mean don't check at all.", &
                          default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUM_GRADIENT_EPS", &
                          description="Step length used for the numeric derivative when checking the gradient.", &
                          default_r_val=1e-8_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUM_GRADIENT_ORDER", &
                          description="Order of the numeric derivative when checking the gradient. "// &
                          "Possible values are 2, 4, and 6.", &
                          default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHECK_UNITARY_TOL", &
                          description="Check if rotation matrix is unitary."// &
                          " Negative values mean don't check at all.", &
                          default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Machine Learning *********************************************************
      CALL section_create(subsection, name="MACHINE_LEARNING", description="Machine learning section")

      CALL keyword_create(keyword, name="METHOD", &
                          description="Machine learning scheme used to predict PAO basis sets.", &
                          enum_c_vals=s2a("GAUSSIAN_PROCESS", "NEURAL_NETWORK"), &
                          enum_i_vals=(/pao_ml_gp, pao_ml_nn/), &
                          enum_desc=s2a("Gaussian Process", "Neural Network"), &
                          default_i_val=pao_ml_gp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TOLERANCE", &
                          description="Maximum variance tolerated when making predictions.", &
                          default_r_val=1.0E-2_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TRAINING_FILE", &
                          description="File which constains a lists PAO-restart file used for training.", &
                          repeats=.FALSE., default_c_val="")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GP_NOISE_VAR", &
                          description="Variance of noise used for Gaussian Process machine learning.", &
                          default_r_val=0.1_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GP_SCALE", &
                          description="Length scale used for Gaussian Process machine learning.", &
                          default_r_val=0.05_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! Output *******************************************************************
      CALL section_create(subsection, name="PRINT", &
                          description="Printkey section", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)

      CALL cp_print_key_section_create(printkey, "RUN_INFO", &
                                       description="Normal output by PAO", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, printkey)
      CALL section_release(printkey)

      CALL cp_print_key_section_create(printkey, "OPT_INFO", &
                                       description="Output by the optimizer", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, printkey)
      CALL section_release(printkey)

      CALL cp_print_key_section_create(printkey, "RESTART", &
                                       description="Restart file of PAO basis", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")

      CALL keyword_create(keyword, name="BACKUP_COPIES", &
                          description="Specifies the maximum number of backup copies.", &
                          usage="BACKUP_COPIES {int}", &
                          default_i_val=1)
      CALL section_add_keyword(printkey, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WRITE_CYCLES", &
                          description="Maximum number of PAO line search cycles until a restart is written.", &
                          default_i_val=100)
      CALL section_add_keyword(printkey, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, printkey)
      CALL section_release(printkey)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! OPT stuff ****************************************************************
      CALL keyword_create(keyword, name="OPTIMIZER", &
                          description="Optimizer used to find PAO basis.", &
                          enum_c_vals=s2a("CG", "BFGS"), &
                          enum_i_vals=(/pao_opt_cg, pao_opt_bfgs/), &
                          enum_desc=s2a("Conjugate gradient algorithm", &
                                        "Broyden-Fletcher-Goldfarb-Shanno algorithm"), &
                          default_i_val=pao_opt_cg)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CG_INIT_STEPS", &
                          description="Number of steepest descent steps before starting the"// &
                          " conjugate gradients optimization.", &
                          default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CG_RESET_LIMIT", &
                          description="The CG is reseted if the cosine of the angle between the last "// &
                          "search direction and the new gradient is larger that the limit.", &
                          default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL linesearch_create_section(subsection)
      CALL section_add_subsection(section, subsection)

      CALL section_release(subsection)
   END SUBROUTINE create_pao_section

END MODULE pao_input
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief input for the linear scaling (LS) section
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE input_cp2k_ls
   USE bibliography,                    ONLY: Lin2009,&
                                              Lin2013,&
                                              Niklasson2003,&
                                              Shao2003,&
                                              VandeVondele2012
   USE cp_output_handling,              ONLY: cp_print_key_section_create,&
                                              high_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        ls_cluster_atomic, ls_cluster_molecular, ls_s_inversion_hotelling, &
        ls_s_inversion_sign_sqrt, ls_s_preconditioner_atomic, ls_s_preconditioner_molecular, &
        ls_s_preconditioner_none, ls_scf_line_search_3point, ls_scf_line_search_3point_2d, &
        ls_scf_ns, ls_scf_pexsi, ls_scf_tc2, ls_scf_trs4
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: integer_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE pao_input,                       ONLY: create_pao_section
   USE qs_density_mixing_types,         ONLY: create_mixing_section
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_ls'

   PUBLIC :: create_ls_scf_section

CONTAINS
! **************************************************************************************************
!> \brief creates the linear scaling scf section
!> \param section ...
!> \author Joost VandeVondele [2010-10]
! **************************************************************************************************
   SUBROUTINE create_ls_scf_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ls_scf_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_ls.F",62)
      CALL section_create(section, "LS_SCF", &
                          description="Specifies the parameters of the linear scaling SCF routines", &
                          n_keywords=24, n_subsections=3, repeats=.FALSE., &
                          citations=(/VandeVondele2012/))

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="LS_DIIS", &
                          description="Perform DIIS within linear scaling SCF", &
                          usage="LS_DIIS", lone_keyword_l_val=.TRUE., &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INI_DIIS", &
                          description="Iteration cycle to start DIIS Kohn-Sham matrix update", &
                          usage="INI_DIIS 2", default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_DIIS", &
                          description="Size of LS_DIIS buffer", &
                          usage="MAX_DIIS 4", default_i_val=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NMIXING", &
                          description="Minimal number of density mixing before start DIIS", &
                          usage="NMIXING 2", default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_DIIS", &
                          description="Threshold on the convergence to start using DIIS", &
                          usage="EPS_DIIS 1.e-1", default_r_val=1.e-1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_SCF", &
                          description="Maximum number of SCF iteration to be performed for one optimization", &
                          usage="MAX_SCF 200", default_i_val=20)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EPS_SCF", &
         description="Target accuracy for the SCF convergence in terms of change of the total energy per electron.", &
         usage="EPS_SCF 1.e-6", default_r_val=1.e-7_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIXING_FRACTION", &
                          description="Mixing density matrices uses the specified fraction in the SCF procedure.", &
                          usage="MIXING_FRACTION 0.4", default_r_val=0.45_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_FILTER", &
                          description="Threshold used for filtering matrix operations.", &
                          usage="EPS_FILTER 1.0E-7", default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_LANCZOS", &
                          description="Threshold used for lanczos estimates.", &
                          usage="EPS_LANCZOS 1.0E-4", default_r_val=1.0E-3_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ITER_LANCZOS", &
                          description="Maximum number of lanczos iterations.", &
                          usage="MAX_ITER_LANCZOS ", default_i_val=128)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MU", &
                          description="Value (or initial guess) for the chemical potential,"// &
                          " i.e. some suitable energy between HOMO and LUMO energy.", &
                          usage="MU 0.0", default_r_val=-0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FIXED_MU", &
                          description="Should the calculation be performed at fixed chemical potential,"// &
                          " or should it be found fixing the number of electrons", &
                          usage="FIXED_MU .TRUE.", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXTRAPOLATION_ORDER", &
                          description="Number of previous matrices used for the ASPC extrapolation of the initial guess. "// &
                          "0 implies that an atomic guess is used at each step. "// &
                          "low (1-2) will result in a drift of the constant of motion during MD. "// &
                          "high (>5) might be somewhat unstable, leading to more SCF iterations.", &
                          usage="EXTRAPOLATION_ORDER 3", default_i_val=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="S_PRECONDITIONER", &
                          description="Preconditions S with some appropriate form.", &
                          usage="S_PRECONDITIONER MOLECULAR", &
                          default_i_val=ls_s_preconditioner_atomic, &
                          enum_c_vals=s2a("NONE", "ATOMIC", "MOLECULAR"), &
                          enum_desc=s2a("No preconditioner", &
                                        "Using atomic blocks", &
                                        "Using molecular sub-blocks. Recommended if molecules are defined and not too large."), &
                          enum_i_vals=(/ls_s_preconditioner_none, ls_s_preconditioner_atomic, ls_s_preconditioner_molecular/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PURIFICATION_METHOD", &
                          description="Scheme used to purify the Kohn-Sham matrix into the density matrix.", &
                          usage="PURIFICATION_METHOD TRS4", &
                          default_i_val=ls_scf_ns, &
                          citations=(/VandeVondele2012, Niklasson2003/), &
                          enum_c_vals=s2a("SIGN_MATRIX", "TRS4", "TC2", "PEXSI"), &
                          enum_desc=s2a("Newton-Schulz sign matrix iteration.", &
                                        "Trace resetting 4th order scheme", &
                                        "Trace conserving 2nd order scheme", &
                                        "PEXSI method"), &
                          enum_i_vals=(/ls_scf_ns, ls_scf_trs4, ls_scf_tc2, ls_scf_pexsi/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DYNAMIC_THRESHOLD", &
                          description="Should the threshold for the purification be chosen dynamically", &
                          usage="DYNAMIC_THRESHOLD .TRUE.", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NON_MONOTONIC", &
                          description="Should the purification be performed non-monotonically. Relevant for TC2 only.", &
                          usage="NON_MONOTONIC .TRUE.", default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="MATRIX_CLUSTER_TYPE", &
         description="Specify how atomic blocks should be clustered in the used matrices, in order to improve flop rate, "// &
         "and possibly speedup the matrix multiply. Note that the atomic s_preconditioner can not be used."// &
         "Furthermore, since screening is on matrix blocks, "// &
         "slightly more accurate results can be expected with molecular.", &
         usage="MATRIX_CLUSTER_TYPE MOLECULAR", &
         default_i_val=ls_cluster_atomic, &
         enum_c_vals=s2a("ATOMIC", "MOLECULAR"), &
         enum_desc=s2a("Using atomic blocks", &
                       "Using molecular blocks."), &
         enum_i_vals=(/ls_cluster_atomic, ls_cluster_molecular/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SINGLE_PRECISION_MATRICES", &
                          description="Matrices used within the LS code can be either double or single precision.", &
                          usage="SINGLE_PRECISION_MATRICES", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="RESTART_WRITE", &
         description="Write the density matrix at the end of the SCF (currently requires EXTRAPOLATION_ORDER>0). "// &
         "Files might be rather large.", &
         usage="RESTART_READ", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART_READ", &
                          description="Read the density matrix before the (first) SCF.", &
                          usage="RESTART_READ", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="S_INVERSION", &
                          description="Method used to compute the inverse of S.", &
                          usage="S_PRECONDITIONER MOLECULAR", &
                          default_i_val=ls_s_inversion_sign_sqrt, &
                          enum_c_vals=s2a("SIGN_SQRT", "HOTELLING"), &
                          enum_desc=s2a("Using the inverse sqrt as obtained from sign function iterations.", &
                                        "Using the Hotellign iteration."), &
                          enum_i_vals=(/ls_s_inversion_sign_sqrt, ls_s_inversion_hotelling/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SIGN_SQRT_ORDER", &
                          description="Order of the sqrt iteration, should be 2..5, 3 or 5 recommended", &
                          usage="SIGN_SQRT_ORDER 5", &
                          default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REPORT_ALL_SPARSITIES", &
                          description="Run the sparsity report at the end of the SCF", &
                          usage="REPORT_ALL_SPARSITIES", default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PERFORM_MU_SCAN", &
                          description="Do a scan of the chemical potential after the SCF", &
                          usage="PERFORM_MU_SCAN", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHECK_S_INV", &
                          description="Perform an accuracy check on the inverse/sqrt of the s matrix.", &
                          usage="CHECK_S_INV", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_ls_curvy_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_chebyshev_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mixing_section(subsection, ls_scf=.TRUE.)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_pexsi_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_pao_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_ls_scf_section

! **************************************************************************************************
!> \brief creates the DOS section
!> \param section ...
!> \author Joost VandeVondele, Jinwoong Cha [2012-10]
! **************************************************************************************************
   SUBROUTINE create_chebyshev_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_chebyshev_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_ls.F",306)

      CALL section_create(section, "CHEBYSHEV", &
                          description="Specifies the parameters needed for the chebyshev expansion based properties.", &
                          n_keywords=24, n_subsections=3, repeats=.FALSE.)

      NULLIFY (keyword)
      NULLIFY (print_key)

      CALL keyword_create(keyword, name="N_CHEBYSHEV", &
                          description="Order of the polynomial expansion.", &
                          usage="N_CHEBYSHEV 2000", default_i_val=500)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! A DOS print key
      CALL cp_print_key_section_create(print_key, "DOS", &
                                       description="Controls the printing of the Density of States (DOS).", &
                                       print_level=high_print_level, filename="")
      CALL keyword_create(keyword, name="N_GRIDPOINTS", &
                          description="Number of points in the computed DOS", &
                          usage="N_GRIDPOINTS 10000", default_i_val=2000)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      ! Energy specific electron density cubes
      CALL cp_print_key_section_create( &
         print_key, &
         name="PRINT_SPECIFIC_E_DENSITY_CUBE", &
         description="Controls the printing of cube files with "// &
         "the electronic density (states) "// &
         "contributing to the density of states within "// &
         "the specific energy range "// &
         "(MIN_ENERGY &le; E &le; MAX_ENERGY). MIN_ENERGY and MAX_ENERGY need to be specified explicitly.", &
         print_level=high_print_level, filename="")

      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIN_ENERGY", &
                          description="Lower bounds of the energy ranges of interest.", &
                          usage="MIN_ENERGY -1.01 -0.62 0.10 .. ", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ENERGY", &
                          description="Upper bounds of the energy ranges of interest.", &
                          usage="MAX_ENERGY -0.81 -0.43 0.22 .. ", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_chebyshev_section

! **************************************************************************************************
!> \brief creates the curvy_steps section in linear scaling scf
!> \param section ...
!> \author Florian Schiffmann [2012-10]
! **************************************************************************************************
   SUBROUTINE create_ls_curvy_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ls_curvy_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_ls.F",384)
      CALL section_create(section, "CURVY_STEPS", &
                          description="Specifies the parameters of the linear scaling SCF routines", &
                          n_keywords=24, n_subsections=3, repeats=.FALSE., &
                          citations=(/Shao2003/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="LINE_SEARCH", &
                          description="Line serch type used in the curvy_setp optimization.", &
                          usage="LINE Search 3POINT", default_i_val=ls_scf_line_search_3point, &
                          enum_c_vals=s2a("3POINT", "3POINT_2D"), &
                          enum_desc=s2a("Performs a three point line search", &
                                        "Only for spin unrestricted calcualtions. Separate step sizes for alpha and beta spin"// &
                                        " using a fit to a 2D parabolic function"), &
                          enum_i_vals=(/ls_scf_line_search_3point, ls_scf_line_search_3point_2d/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_BCH_HISTORY", &
                          description="Number of stored matrices in the Baker-Campbell-Hausdorff series. "// &
                          "Reduces the BCH evaluation during line search but can be memory intense. ", &
                          usage="N_BCH_HISTORY 5", &
                          default_i_val=7)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIN_HESSIAN_SHIFT", &
                          description="Minimal eigenvalue shift for the Hessian in the Newton iteration."// &
                          " Useful for small band gap systems (0.5-1.0 recommended). ", &
                          usage="MIN_HESSIAN_SHIFT 0.0", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FILTER_FACTOR", &
                          description="Allows to set a seperate EPS_FILTER in the newton iterations."// &
                          " The new EPS is EPS_FILTER*FILTER_FACTOR.", &
                          usage="FILTER_FACTOR 10.0", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FILTER_FACTOR_SCALE", &
                          description="Allows for dynamic EPS_FILTER. Updates the filter factor every SCF-Newton "// &
                          "step by FILTER_FACTOR=FILTER_FACTOR*FILTER_FACTOR_SCALE", &
                          usage="FILTER_FACTOR_SCALE 0.5", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIN_FILTER", &
                          description="Lowest EPS_FILTER in dynamic filtering. Given as multiple of EPS_FILTER:"// &
                          " EPS_FILTER_MIN=EPS_FILTER*MIN_FILTER", &
                          usage="FILTER_FACTOR 1.0", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_ls_curvy_section

! **************************************************************************************************
!> \brief creates the PEXSI library subsection of the linear scaling section.
!> \param section ...
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE create_pexsi_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_pexsi_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_ls.F",456)

      CALL section_create(section, "PEXSI", &
                          description="Specifies the parameters of the PEXSI library. The density matrix is calculated "// &
                          "with PEXSI if PURIFICATION_METHOD (in LS_SCF section) is set to PEXSI. "// &
                          "The computational cost of PEXSI is at most quadratically scaling w.r.t. the system size "// &
                          "and PEXSI is applicable to insulating and metallic systems. The value of EPS_PGF_ORB "// &
                          "(in QS input section) defines the sparsity of the matrices sent to PEXSI and EPS_FILTER "// &
                          "is overwritten with 0.", &
                          n_keywords=17, repeats=.FALSE., citations=(/Lin2009, Lin2013/))
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="TEMPERATURE", &
                          description="Electronic temperature", &
                          default_r_val=cp_unit_to_cp2k(value=300.0_dp, unit_str="K"), &
                          unit_str="K")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GAP", &
                          description="Spectral gap. Note: This can be set to be 0 in most cases.", &
                          default_r_val=0.0_dp, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUM_POLE", &
                          description="Number of terms in the pole expansion (should be even).", &
                          default_i_val=64)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="IS_INERTIA_COUNT", &
                          description="Whether inertia counting is used each time the DFT driver "// &
                          "of PEXSI is invoked. If FALSE, inertia counting is still used in the "// &
                          "first SCF iteration.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_PEXSI_ITER", &
                          description="Maximum number of PEXSI iterations after each inertia counting procedure.", &
                          default_i_val=5)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MU_MIN_0", &
                          description="Initial guess of lower bound for mu.", &
                          default_r_val=-5.0_dp, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MU_MAX_0", &
                          description="Initial guess of upper bound for mu.", &
                          default_r_val=5.0_dp, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MU_INERTIA_TOLERANCE", &
                          description="Stopping criterion in terms of the chemical potential for the "// &
                          "inertia counting procedure.", &
                          default_r_val=0.01_dp, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MU_INERTIA_EXPANSION", &
                          description="If the chemical potential is not in the initial interval, "// &
                          "the interval is expanded by MU_INERTIA_EXPANSION.", &
                          default_r_val=0.15_dp, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MU_PEXSI_SAFE_GUARD", &
                          description="Safe guard criterion in terms of the chemical potential to "// &
                          "reinvoke the inertia counting procedure.", &
                          default_r_val=0.01_dp, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUM_ELECTRON_PEXSI_TOLERANCE", &
                          description="Stopping criterion of the PEXSI iteration in terms of "// &
                          "The number of electrons compared to the exact number of electrons. "// &
                          "This threshold is the target tolerance applied at convergence of SCF.", &
                          default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUM_ELECTRON_INITIAL_TOLERANCE", &
                          description="The same as NUM_ELECTRON_PEXSI_TOLERANCE but applied in the first SCF steps. "// &
                          "If set to a value smaller than NUM_ELECTRON_PEXSI_TOLERANCE, it is overwritten with "// &
                          "NUM_ELECTRON_PEXSI_TOLERANCE (default). If set to a value larger than "// &
                          "NUM_ELECTRON_PEXSI_TOLERANCE, the PEXSI tolerance in number of electrons is set adaptively "// &
                          "according to the SCF convergence error of the previous SCF step. This reduces the number "// &
                          "of PEXSI iterations in the first SCF steps but leads to at least one more SCF step.", &
                          default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ORDERING", &
                          description="Ordering strategy for factorization and selected inversion.", &
                          enum_c_vals=s2a("PARALLEL", "SEQUENTIAL", "MULTIPLE_MINIMUM_DEGREE"), &
                          enum_desc=s2a("Parallel ordering using ParMETIS/PT-SCOTCH (PARMETIS option in SuperLU_DIST)", &
                                        "Sequential ordering using METIS (METIS_AT_PLUS_A option in SuperLU_DIST)", &
                                        "Multiple minimum degree ordering (MMD_AT_PLUS_A option in SuperLU_DIST)"), &
                          enum_i_vals=(/0, 1, 2/), default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NP_SYMB_FACT", &
                          description="Number of processors for PARMETIS/PT-SCOTCH. Only used if ORDERING is set to PARALLEL. "// &
                          "If 0, the number of processors for PARMETIS/PT-SCOTCH will be set equal to the number of "// &
                          "MPI ranks per pole. Note: if more than one processor is used, a segmentation fault may occur in the "// &
                          "symbolic factorization phase.", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VERBOSITY", &
                          description="The level of output information.", &
                          enum_c_vals=s2a("SILENT", "BASIC", "DETAILED"), &
                          enum_i_vals=(/0, 1, 2/), default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIN_RANKS_PER_POLE", &
                          description="The minimum number of processors used for each pole. The real "// &
                          "number of processors per pole is the smallest number greater or equal to "// &
                          "MIN_RANKS_PER_POLE that divides MPI size without remainder. For efficiency, MIN_RANKS_PER_POLE "// &
                          "should be a small numbers (limited by the available memory).", &
                          default_i_val=64)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CSR_SCREENING", &
                          description="Whether distance screening should be applied to improve sparsity of CSR matrices.", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_pexsi_section

END MODULE input_cp2k_ls
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief input for the ALMO SCF section
!> \author Rustam Khaliullin
! **************************************************************************************************
MODULE input_cp2k_almo
   USE bibliography,                    ONLY: Khaliullin2007,&
                                              Khaliullin2008,&
                                              Khaliullin2013
   USE input_constants,                 ONLY: &
        almo_deloc_none, almo_deloc_scf, almo_deloc_x, almo_deloc_x_then_scf, &
        almo_deloc_xalmo_1diag, almo_deloc_xalmo_scf, almo_deloc_xalmo_x, almo_frz_crystal, &
        almo_frz_none, almo_scf_diag, almo_scf_pcg, atomic_guess, cg_dai_yuan, cg_fletcher, &
        cg_fletcher_reeves, cg_hager_zhang, cg_hestenes_stiefel, cg_liu_storey, cg_polak_ribiere, &
        cg_zero, molecular_guess, optimizer_diis, optimizer_pcg, prec_default, prec_zero
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_almo'

   INTEGER, PARAMETER, PRIVATE :: optimizer_block_diagonal_diis = 1
   INTEGER, PARAMETER, PRIVATE :: optimizer_block_diagonal_pcg = 2
   INTEGER, PARAMETER, PRIVATE :: optimizer_xalmo_pcg = 3

   PUBLIC :: create_almo_scf_section

CONTAINS

! **************************************************************************************************
!> \brief create the almo scf section
!> \param section ...
!> \par History
!>       2011.05 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE create_almo_scf_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_almo_scf_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_almo.F",61)
      CALL section_create(section, "ALMO_SCF", &
                          description="Settings for a class of efficient linear scaling methods based "// &
                          "on absolutely localized orbitals"// &
                          " (ALMOs). ALMO methods are currently restricted to closed-shell molecular systems.", &
                          n_keywords=4, n_subsections=3, repeats=.FALSE., &
                          citations=(/Khaliullin2013/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="EPS_FILTER", &
                          description="Threshold for the matrix sparsity filter", &
                          usage="EPS_FILTER 1.e-6", default_r_val=1.e-7_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALMO_SCF_GUESS", &
                          description="The method to generate initial ALMOs.", &
                          usage="ALMO_SCF_GUESS MOLECULAR", &
                          default_i_val=atomic_guess, &
                          enum_c_vals=s2a("MOLECULAR", "ATOMIC"), &
                          enum_desc=s2a("SCF calculations on single molecules controled by the regular SCF "// &
                                        "keywords outside ALMO options. This kind of calculation is expensive "// &
                                        "and only recommended if ALMO SCF does not converge from the ATOMIC guess.", &
                                        "Superpoisiton of atomic densities."), &
                          enum_i_vals=(/molecular_guess, atomic_guess/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALMO_ALGORITHM", &
                          description="Specifies the algorithm to update block-diagonal ALMOs during the SCF procedure.", &
                          usage="ALMO_ALGORITHM DIAG", &
                          default_i_val=almo_scf_diag, &
                          !enum_c_vals=s2a("DIAG", "DM_SIGN","PCG"),&
                          enum_c_vals=s2a("DIAG", "PCG"), &
                          enum_desc=s2a("DIIS-accelerated diagonalization controlled by ALMO_OPTIMIZER_DIIS. "// &
                                        "Recommended for large systems containing small fragments.", &
                                        !"Update the density matrix using linear scaling routines. "//&
                                        !"Recommended if large fragments are present.",&
                                        "Energy minimization with a PCG algorithm controlled by ALMO_OPTIMIZER_PCG."), &
                          !enum_i_vals=(/almo_scf_diag,almo_scf_dm_sign,almo_scf_pcg/),&
                          enum_i_vals=(/almo_scf_diag, almo_scf_pcg/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="DELOCALIZE_METHOD", &
         description="Methods to reintroduce electron delocalization, which is excluded "// &
         "with the block-diagonal ALMO reference. Electron delocalization can "// &
         "be computed using either fully delocalized MOs or spatially restricted "// &
         "ALMOs (called extended ALMOs or XALMOs). All methods below use a PCG "// &
         "optimizer controlled by XALMO_OPTIMIZER_PCG. The only exception is "// &
         "the non-iterative XALMO_1DIAG.", &
         usage="DELOCALIZE_METHOD XALMO_X", &
         default_i_val=almo_deloc_xalmo_x, &
         enum_c_vals=s2a("NONE", "XALMO_1DIAG", "XALMO_X", "XALMO_SCF", "FULL_X", "FULL_SCF", "FULL_X_THEN_SCF"), &
         enum_desc=s2a("Neglect electron delocalization", &
                       "Correction based on one diagonalization of the spatially projected Hamiltonian (XALMO)", &
                       "Single excitation correction (no Hamiltonian re-build) with spatial restrictions (XALMO)", &
                       "Self-consistent treatment of delocalization with spatial restrictions (XALMO)", &
                       "Single excitation correction (no Hamiltonian re-build) without spatial restrictions", &
                       "Self-consistent treatment of delocalization without spatial restrictions", &
                       "Single excitation correction followed by full SCF procedure, both without spatial restrictions"), &
         enum_i_vals=(/almo_deloc_none, almo_deloc_xalmo_1diag, almo_deloc_xalmo_x, almo_deloc_xalmo_scf, &
                       almo_deloc_x, almo_deloc_scf, almo_deloc_x_then_scf/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="XALMO_R_CUTOFF_FACTOR", &
                          description="Controls the localization radius of XALMOs: "// &
                          !"r0 = r0_factor*(radius(at1)+radius(at2)) + r0_shift",&
                          "R_cutoff = XALMO_R_CUTOFF_FACTOR*(radius(at1)+radius(at2))", &
                          usage="XALMO_R_CUTOFF_FACTOR 1.6", default_r_val=1.50_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOCALIZE_EPS_ITER",&
      !     description="Obsolete and to be deleted: use EPS_ERROR in XALMO_OPTIMIZER_PCG",&
      !     usage="DELOCALIZE_EPS_ITER 1.e-5", default_r_val=1.e-5_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOCALIZE_EPS_LIN_SEARCH",&
      !     description="Obsolete and to be deleted: use EPS_ERROR_LIN_SEARCH in XALMO_OPTIMIZER_PCG",&
      !     usage="DELOCALIZE_EPS_LIN_SEARCH 1.e-6", default_r_val=1.e-7_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="BLOCKED_MAX_ITER",&
      !     description="Obsolete and to be deleted: use MAX_ITER in ALMO_OPTIMIZER_DIIS or ALMO_OPTIMIZER_PCG",&
      !     usage="BLOCKED_MAX_ITER 200", default_i_val=100)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="BLOCKED_EPS_ITER",&
      !     description="Obsolete and to be deleted: use EPS_ERROR in ALMO_OPTIMIZER_DIIS or ALMO_OPTIMIZER_PCG",&
      !     usage="BLOCKED_EPS_ITER 1.e-5", default_r_val=1.e-5_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOCALIZE_MAX_ITER",&
      !     description="Obsolete and to be deleted: use MAX_ITER in XALMO_OPTIMIZER_PCG",&
      !     usage="DELOCALIZE_MAX_ITER 200", default_i_val=100)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DOMAIN_LAYOUT_MOS",&
      !     description="Each electron in the system is constrained to its own delocalization domain."//&
      !     " This keyword creates groups of electrons that share the same domain.",&
      !     usage="DOMAIN_LAYOUT_MOS MOLECULAR",&
      !     default_i_val=almo_domain_layout_molecular,&
      !     enum_c_vals=s2a( "ORBITAL", "ATOMIC", "MOLECULAR"),&
      !     enum_desc=s2a("Each electron can have its own delocalization domain",&
      !     "All electrons of an atom are delocalized over the same domain",&
      !     "All electrons of a molecule are delocalized over the same domain."//&
      !     " This is the only implemented option"),&
      !     enum_i_vals=(/almo_domain_layout_orbital,almo_domain_layout_atomic,almo_domain_layout_molecular/))
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DOMAIN_LAYOUT_AOS",&
      !     description="Atomic orbitals or groups of atomic orbitals represent domains over which electrons "//&
      !     "can be delocalized. This keyword imposes constraints on the structure of domains",&
      !     usage="DOMAIN_LAYOUT_MOS MOLECULAR",&
      !     default_i_val=almo_domain_layout_molecular,&
      !     enum_c_vals=s2a("ATOMIC", "MOLECULAR"),&
      !     enum_desc=s2a("Atomic blocks represent domains. That is, if a basis function on an atom is in"//&
      !     " domain A then all basis functions on this atom are in domain A.",&
      !     "Molecular subsets represent domains. That is, if a basis function on a molecule is"//&
      !     " in domain A then all basis functions on this molecule are in domain A. "//&
      !     "This is the only implemented option"),&
      !     enum_i_vals=(/almo_domain_layout_atomic,almo_domain_layout_molecular/))
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="MATRIX_CLUSTERING_MOS",&
      !     description="Blocks of matrices in the MO basis set are distributed for parallel computations. "//&
      !     "This keywords specifies the type of matrix blocks.",&
      !     usage="MATRIX_CLUSTERING_MOS MOLECULAR",&
      !     default_i_val=almo_mat_distr_molecular,&
      !     enum_c_vals=s2a("ATOMIC", "MOLECULAR"),&
      !     enum_desc=s2a("Not recommended. ZZZ Maybe used for the PAO-based methods in the future",&
      !                   "All molecular orbitals of a molecule belong to the same block."),&
      !     enum_i_vals=(/almo_mat_distr_atomic,almo_mat_distr_molecular/))
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="MATRIX_CLUSTERING_AOS",&
      !     description="Blocks of matrices in the AO basis set are distributed for parallel computations."//&
      !     " This keywords specifies the type of matrix blocks.",&
      !     usage="MATRIX_CLUSTERING_AOS MOLECULAR",&
      !     default_i_val=almo_mat_distr_molecular,&
      !     enum_c_vals=s2a("ATOMIC", "MOLECULAR"),&
      !     enum_desc=s2a("All atomic orbitals of an atom belong to the "//&
      !     "same block. Use only if there are very large molecules in the system. "//&
      !     "ZZZ Maybe used for the PAO-based methods in the future",&
      !     "All atomic orbitals of a molecule belong to the same block."),&
      !     enum_i_vals=(/almo_mat_distr_atomic,almo_mat_distr_molecular/))
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="CONSTRAINT_TYPE",&
      !     description="Determines the type of ALMO constraints",&
      !     usage="CONSTRAINT_TYPE BLOCK_DIAGONAL",&
      !     default_i_val=almo_constraint_block_diagonal,&
      !     enum_c_vals=s2a("BLOCK_DIAGONAL", "DISTANCE", "AO_OVERLAP"),&
      !     enum_desc=s2a("MO coefficient matrix is block-diagonal",&
      !     "MO coefficients are quenched according to the distance criterion",&
      !     "MO coefficients are quenched according to the AO overlap criterion"),&
      !     enum_i_vals=(/almo_constraint_block_diagonal,almo_constraint_distance,&
      !     almo_constraint_ao_overlap/))
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !!CALL keyword_create(keyword, name="OUTER_MAX_SCF_Q",&
      !!     description="Maximum number of the outer loop SCF iterations for optimization of quenched ALMOs",&
      !!     usage="OUTER_MAX_SCF_Q 10", default_i_val=0)
      !!CALL section_add_keyword(section,keyword)
      !!CALL keyword_release(keyword)

      !!CALL keyword_create(keyword, name="EPS_USE_PREV_AS_GUESS",&
      !!     description="SCF convergence below which quantities from previous iterations"//&
      !!                 " can be used as initial guess for the current iteration.",&
      !!     usage="EPS_USE_PREV_AS_GUESS 0.01", default_r_val=0.001_dp)
      !!CALL section_add_keyword(section,keyword)
      !!CALL keyword_release(keyword)

      !!CALL keyword_create(keyword, name="MIXING_FRACTION",&
      !!     description="Weight of the new KS matrix in the mixing during the SCF procedure.",&
      !!     usage="MIXING_FRACTION 0.45", default_r_val=0.45_dp)
      !!CALL section_add_keyword(section,keyword)
      !!CALL keyword_release(keyword)

      !!CALL keyword_create(keyword, name="FIXED_MU",&
      !!     description="Fix chemical potential or optimize it to get "//&
      !!                 "the correct number of electrons",&
      !!     usage="FIXED_MU .TRUE.", default_l_val=.FALSE.,&
      !!     lone_keyword_l_val=.TRUE.)
      !!CALL section_add_keyword(section,keyword)
      !!CALL keyword_release(keyword)

      !!CALL keyword_create(keyword, name="MU",&
      !!     description="Value (or initial guess) for the chemical potential."//&
      !!                 " Provide energy between HOMO and LUMO energy.",&
      !!     usage="MU 0.0", default_r_val=-0.1_dp)
      !!CALL section_add_keyword(section,keyword)
      !!CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="XALMO_ALGORITHM",&
      !     description="Specifies the algorithm to update ALMOs on eXtended domains (XALMOs).",&
      !     usage="XALMO_ALGORITHM PCG",&
      !     default_i_val=almo_scf_diag,&
      !     enum_c_vals=s2a("DDIAG", "PCG"),&
      !     enum_desc=s2a("Domain diagonalization",&
      !                   "Preconditioned conjugate gradient"),&
      !     enum_i_vals=(/almo_scf_diag,almo_scf_pcg/))
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="QUENCHER_RADIUS_TYPE",&
      !     description="Determines the type of atomic radii used for imposing the ALMO constraints",&
      !     usage="QUENCHER_RADIUS_TYPE VDW",&
      !     default_i_val=do_bondparm_vdw,&
      !     enum_c_vals=s2a("COVALENT", "VDW"),&
      !     enum_desc=s2a("Covalent atomic radii",&
      !     "Van der Waals atomic radii"),&
      !     enum_i_vals=(/do_bondparm_covalent,do_bondparm_vdw/))
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="QUENCHER_R0_FACTOR",&
      !     description="Parameter to calculate the inner soft cutoff radius: "//&
      !     !"r0 = r0_factor*(radius(at1)+radius(at2)) + r0_shift",&
      !     "r0 = r0_factor*(radius(at1)+radius(at2))",&
      !     usage="QUENCHER_R0_FACTOR 1.05", default_r_val=1.05_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !!CALL keyword_create(keyword, name="QUENCHER_R0_SHIFT",&
      !!     description="Parameter to calculate the inner soft cutoff radius (in Angstrom): "//&
      !!     "r0 = r0_factor*(radius(at1)+radius(at2)) + r0_shift",&
      !!     usage="QUENCHER_R0_SHIFT 0.0", default_r_val=0.0_dp)
      !!
      !!CALL section_add_keyword(section,keyword)
      !!CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="QUENCHER_R1_FACTOR",&
      !     description="Parameter to calculate the outer soft cutoff radius: "//&
      !     !"r1 = r1_factor*(radius(at1)+radius(at2)) + r1_shift",&
      !     "r1 = r1_factor*(radius(at1)+radius(at2))",&
      !     usage="QUENCHER_R1_FACTOR 1.55", default_r_val=1.55_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !!CALL keyword_create(keyword, name="QUENCHER_R1_SHIFT",&
      !!     description="Parameter to calculate the outer soft cutoff radius (in Angstrom): "//&
      !!     "r1 = r1_factor*(radius(at1)+radius(at2)) + r1_shift",&
      !!     usage="QUENCHER_R1_SHIFT 0.0", default_r_val=0.0_dp)
      !!
      !!CALL section_add_keyword(section,keyword)
      !!CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="QUENCHER_AO_OVERLAP_0",&
      !     description="Overlap value of the inner soft cutoff",&
      !     usage="QUENCHER_AO_OVERLAP_0 1.0E-4", default_r_val=1.0E-4_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="QUENCHER_AO_OVERLAP_1",&
      !     description="Overlap value of the outer soft cutoff",&
      !     usage="QUENCHER_AO_OVERLAP_1 1.0E-6", default_r_val=1.0E-6_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="ENVELOPE_AMPLITUDE",&
      !     description="Defines an upper bound on the maximum norm of the MO coefficients",&
      !     usage="ENVELOPE_AMPLITUDE 1.0", default_r_val=1.0_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOC_CAYLEY_TENSOR_TYPE",&
      !     description="Tensor properties of occupied and virtual indices",&
      !     usage="DELOC_CAYLEY_TENSOR_TYPE ORTHOGONAL",&
      !     default_i_val=tensor_orthogonal,&
      !     enum_c_vals=s2a("ORTHOGONAL", "BIORTHOGONAL"),&
      !     enum_desc=s2a("Orthogonalize both occupied and virtual orbitals",&
      !     "Contravariant virtual (MOs or AOs) and covariant occupied orbitals"),&
      !     enum_i_vals=(/tensor_orthogonal,tensor_up_down/))
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOC_CAYLEY_CONJUGATOR",&
      !     description="Various methods to compute step directions in the CG algorithm",&
      !     usage="DELOC_CAYLEY_CONJUGATOR POLAK_RIBIERE",&
      !     default_i_val=cg_hager_zhang,&
      !     enum_c_vals=s2a("ZERO", "POLAK_RIBIERE", "FLETCHER_REEVES",&
      !     "HESTENES_STIEFEL", "FLETCHER", "LIU_STOREY", "DAI_YUAN","HAGER_ZHANG"),&
      !     enum_desc=s2a("Steepest descent","Polak and Ribiere",&
      !     "Fletcher and Reeves","Hestenes and Stiefel",&
      !     "Fletcher (Conjugate descent)","Liu and Storey",&
      !     "Dai and Yuan","Hager and Zhang"),&
      !     enum_i_vals=(/cg_zero,cg_polak_ribiere,cg_fletcher_reeves,&
      !                   cg_hestenes_stiefel,cg_fletcher,cg_liu_storey,&
      !                   cg_dai_yuan,cg_hager_zhang/))
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOC_CAYLEY_MAX_ITER",&
      !     description="Maximum number of CG iterations to solve Ricatti equations",&
      !     usage="DELOC_CAYLEY_MAX_ITER 100",default_i_val=50)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOC_CAYLEY_EPS_CONVERGENCE",&
      !     description="Convergence criterion of the CG algorithm",&
      !     usage="DELOC_CAYLEY_EPS_CONVERGENCE 1.e-6", default_r_val=1.e-7_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOC_CAYLEY_VIR_PRECOND",&
      !     description="Use preconditioner for the virtual subspace",&
      !     usage="DELOC_CAYLEY_VIR_PRECOND .TRUE.", default_l_val=.TRUE.,&
      !     lone_keyword_l_val=.TRUE.)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOC_CAYLEY_OCC_PRECOND",&
      !     description="Use preconditioner for the occupied subspace",&
      !     usage="DELOC_CAYLEY_OCC_PRECOND .TRUE.", default_l_val=.TRUE.,&
      !     lone_keyword_l_val=.TRUE.)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOC_TRUNCATE_VIRTUALS",&
      !     description="Truncation of the virtual subspace",&
      !     usage="DELOC_TRUNCATE_VIRTUALS MINIMAL",&
      !     default_i_val=virt_full,&
      !     enum_c_vals=s2a("FULL", "MINIMAL","OCC_SIZE", "EXACT_NUMBER_PER_DOMAIN"),&
      !     enum_desc=s2a("Keep all virtual orbitals","Retained virtuals "//&
      !     "complement occupied orbitals to form the minimal basis set",&
      !     "Number of virtuals is equal to the number of occupied orbitals",&
      !     "Specify exact number of virtuals per domain with DELOC_VIRT_PER_DOMAIN"),&
      !     enum_i_vals=(/virt_full,virt_minimal,virt_occ_size,&
      !                   virt_number/))
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOC_VIRT_PER_DOMAIN",&
      !     description="Number of virtual orbitals (per domain, atom or molecule) "//&
      !     "retained to obtain the delocalization correction",&
      !     usage="DELOC_VIRT_PER_DOMAIN",default_i_val=-1)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOC_USE_OCC_ORBS",&
      !     description="Use occupied orbitals (as opposed to density matrix) "//&
      !     "to calculate correction for electron delocalization",&
      !     usage="DELOC_USE_OCC_ORBS .TRUE.", default_l_val=.TRUE.,&
      !     lone_keyword_l_val=.TRUE.)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOC_CAYLEY_USE_VIRT_ORBS",&
      !     description="Use virtual orbitals (as opposed to the 1-P projector) "//&
      !     "to calculate correction for electron delocalization. Works only if "//&
      !     "DELOC_USE_OCC_ORBS is set to TRUE",&
      !     usage="DELOC_CAYLEY_USE_VIRT_ORBS .TRUE.", default_l_val=.FALSE.,&
      !     lone_keyword_l_val=.TRUE.)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="DELOC_CAYLEY_LINEAR",&
      !     description="Neglect the quadratic term in the Riccati equations. "//&
      !     "Equivalent to the first order correction to the occupied orbitals "//&
      !     "(second order correction to the energy)",&
      !     usage="DELOC_CAYLEY_LINEAR .FALSE.", default_l_val=.FALSE.,&
      !     lone_keyword_l_val=.TRUE.)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="OPT_K_OUTER_MAX_ITER",&
      !     description="Maximum number of outer loop iterations to optimize retained virtual orbitals",&
      !     usage="OPT_K_OUTER_MAX_ITER 10",default_i_val=1)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="OPT_K_MAX_ITER",&
      !     description="Maximum number of iterations to optimize retained virtual orbitals",&
      !     usage="OPT_K_MAX_ITER 100",default_i_val=100)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="OPT_K_EPS_CONVERGENCE",&
      !     description="Convergence criterion of the optimization algorithm",&
      !     usage="OPT_K_EPS_CONVERGENCE 1.e-5", default_r_val=1.e-5_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="OPT_K_TRIAL_STEP_SIZE",&
      !     description="Size of the trial step along the gradient",&
      !     usage="OPT_K_TRIAL_STEP_SIZE 0.05", default_r_val=0.05_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="OPT_K_TRIAL_STEP_SIZE_MULTIPLIER",&
      !     description="The trial step size is obtained by multiplying the optimal step size "//&
      !     "from the previous iteration",&
      !     usage="OPT_K_TRIAL_STEP_SIZE_multiplier 1.0", default_r_val=1.4_dp)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="OPT_K_CONJ_ITER_START",&
      !     description="Iteration for switching from the steepest descent algorithm "//&
      !     "to conjugate gradient",&
      !     usage="OPT_K_CONJ_ITER_START 5",default_i_val=0)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="OPT_K_CONJ_ITER_FREQ_RESET",&
      !     description="Reset frequency of the conjugate gradient direction",&
      !     usage="OPT_K_CONJ_ITER_FREQ_RESET 20",default_i_val=1000000)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="OPT_K_CONJUGATOR",&
      !     description="Various methods to compute step directions in the CG algorithm",&
      !     usage="OPT_K_CONJUGATOR POLAK_RIBIERE",&
      !     default_i_val=cg_hager_zhang,&
      !     enum_c_vals=s2a("ZERO", "POLAK_RIBIERE", "FLETCHER_REEVES",&
      !     "HESTENES_STIEFEL", "FLETCHER", "LIU_STOREY", "DAI_YUAN","HAGER_ZHANG"),&
      !     enum_desc=s2a("Steepest descent","Polak and Ribiere",&
      !     "Fletcher and Reeves","Hestenes and Stiefel",&
      !     "Fletcher (Conjugate descent)","Liu and Storey",&
      !     "Dai and Yuan","Hager and Zhang"),&
      !     enum_i_vals=(/cg_zero,cg_polak_ribiere,cg_fletcher_reeves,&
      !                   cg_hestenes_stiefel,cg_fletcher,cg_liu_storey,&
      !                   cg_dai_yuan,cg_hager_zhang/))
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="OPT_K_PREC_ITER_START",&
      !     description="Start using the preconditioner (approximate preconditioners "//&
      !     "might not be valid on early iterations)",&
      !     usage="OPT_K_PREC_ITER_START 2",default_i_val=0)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      !CALL keyword_create(keyword, name="OPT_K_PREC_ITER_FREQ_UPDATE",&
      !     description="Frequency for updating the preconditioner",&
      !     usage="OPT_K_PREC_ITER_FREQ_UPDATE 10",default_i_val=1)
      !CALL section_add_keyword(section,keyword)
      !CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL create_optimizer_section(subsection, optimizer_block_diagonal_diis)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_optimizer_section(subsection, optimizer_block_diagonal_pcg)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_optimizer_section(subsection, optimizer_xalmo_pcg)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      !NULLIFY(subsection)
      !CALL create_almo_eda_section(subsection)
      !CALL section_add_subsection(section, subsection)
      !CALL section_release(subsection)

   END SUBROUTINE create_almo_scf_section

! **************************************************************************************************
!> \brief The ALMO EDA section controls decomposition analysis based on ALMOs
!> \param section ...
!> \par History
!>       2014.10 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE create_almo_eda_section(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_almo_eda_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_almo.F",552)

      CALL section_create(section, "ALMO_DA", &
                          description="Controls decomposition analysis based on ALMOs and XALMOs. "// &
                          "Not yet fully implemented.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/Khaliullin2007, Khaliullin2008/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="FRZ_TERM", &
                          description="Perform calculations on single molecules to compute the frozen density term", &
                          usage="FRZ_TERM ISOLATED", default_i_val=0, &
                          !enum_c_vals=s2a("SKIP", "ISOLATED", "CRYSTAL"),&
                          enum_c_vals=s2a("SKIP", "CRYSTAL"), &
                          enum_desc=s2a("Do not compute the frozen energy term.", &
                                        !"Use isolated molecules as the reference.",&
                                        "Use molecules in their positions in the crystal cell as the reference. "// &
                                        !"Substantially more expensive than ISOLATED. "//&
                                        "Be careful interpreting this term for systems with charged fragmetns."), &
                          enum_i_vals=(/almo_frz_none, &
                                        !              almo_frz_isolated,&
                                        almo_frz_crystal/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_almo_eda_section

! **************************************************************************************************
!> \brief The optimizer section is a collection of keywords that are similar
!>        to all optimization methods (e.g. target error, number of iterations)
!> \param section ...
!> \param optimizer_id - allows to adapt the standard section for specific needs
!> \par History
!>       2012.03 created [Rustam Z Khaliullin]
!>       2014.10 fully integrated [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE create_optimizer_section(section, optimizer_id)

      TYPE(section_type), POINTER                        :: section
      INTEGER, INTENT(IN)                                :: optimizer_id

      CHARACTER(len=*), PARAMETER :: routineN = 'create_optimizer_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: optimizer_type
      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_almo.F",601)

      ! choose the name of the section
      SELECT CASE (optimizer_id)
      CASE (optimizer_block_diagonal_diis)
         CALL section_create(section, "ALMO_OPTIMIZER_DIIS", &
                             description="Controls the iterative DIIS-accelerated optimization of block-diagonal ALMOs.", &
                             n_keywords=3, n_subsections=0, repeats=.FALSE.)
         optimizer_type = optimizer_diis
      CASE (optimizer_block_diagonal_pcg)
         CALL section_create(section, "ALMO_OPTIMIZER_PCG", &
                             description="Controls the PCG optimization of block-diagonal ALMOs.", &
                             n_keywords=6, n_subsections=0, repeats=.FALSE.)
         optimizer_type = optimizer_pcg
      CASE (optimizer_xalmo_pcg)
         CALL section_create(section, "XALMO_OPTIMIZER_PCG", &
                             description="Controls the PCG optimization of extended ALMOs.", &
                             n_keywords=6, n_subsections=0, repeats=.FALSE.)
         optimizer_type = optimizer_pcg
      CASE DEFAULT
         CALL cp__b("input_cp2k_almo.F",621,"No default values allowed")
      END SELECT

      NULLIFY (keyword)

      ! add common keywords
      CALL keyword_create(keyword, name="MAX_ITER", &
                          description="Maximum number of iterations", &
                          usage="MAX_ITER 100", default_i_val=20)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_ERROR", &
                          description="Target value of the MAX norm of the error", &
                          usage="EPS_ERROR 1.E-6", default_r_val=1.0E-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! add keywords specific to each type
      SELECT CASE (optimizer_type)
      CASE (optimizer_diis)

         CALL keyword_create(keyword, name="N_DIIS", &
                             description="Number of error vectors to be used in the DIIS "// &
                             "optimization procedure", &
                             usage="N_DIIS 5", default_i_val=6)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

      CASE (optimizer_pcg)

         CALL keyword_create(keyword, name="LIN_SEARCH_EPS_ERROR", &
                             description="Target value of the gradient norm during the linear search", &
                             usage="LIN_SEARCH_EPS_ERROR 1.E-2", default_r_val=1.0E-3_dp)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL keyword_create(keyword, name="LIN_SEARCH_STEP_SIZE_GUESS", &
                             description="The size of the first step in the linear search", &
                             usage="LIN_SEARCH_STEP_SIZE_GUESS 0.1", default_r_val=1.0_dp)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL keyword_create(keyword, name="MAX_ITER_OUTER_LOOP", &
                             description="Maximum number of iterations in the outer loop. "// &
                             "Use the outer loop to update the preconditioner and reset the conjugator. "// &
                             "This can speed up convergence significantly.", &
                             usage="MAX_ITER 10", default_i_val=0)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL keyword_create(keyword, name="PRECONDITIONER", &
                             description="Select a preconditioner for the conjugate gradient optimization", &
                             usage="PRECONDITIONER NONE", &
                             default_i_val=-1, &
                             enum_c_vals=s2a("DEFAULT", "NONE"), &
                             enum_desc=s2a("Default preconditioner", "Do not use preconditioner"), &
                             enum_i_vals=(/prec_default, prec_zero/))
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL keyword_create(keyword, name="CONJUGATOR", &
                             description="Various methods to compute step directions in the PCG optimization", &
                             usage="CONJUGATOR POLAK_RIBIERE", &
                             default_i_val=cg_hager_zhang, &
                             enum_c_vals=s2a("ZERO", "POLAK_RIBIERE", "FLETCHER_REEVES", &
                                             "HESTENES_STIEFEL", "FLETCHER", "LIU_STOREY", "DAI_YUAN", "HAGER_ZHANG"), &
                             enum_desc=s2a("Steepest descent", "Polak and Ribiere", &
                                           "Fletcher and Reeves", "Hestenes and Stiefel", &
                                           "Fletcher (Conjugate descent)", "Liu and Storey", &
                                           "Dai and Yuan", "Hager and Zhang"), &
                             enum_i_vals=(/cg_zero, cg_polak_ribiere, cg_fletcher_reeves, &
                                           cg_hestenes_stiefel, cg_fletcher, cg_liu_storey, &
                                           cg_dai_yuan, cg_hager_zhang/))
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

      END SELECT

   END SUBROUTINE create_optimizer_section

!!! **************************************************************************************************
!!!> \brief The developer section allows uninterrupted development of the code
!!!>        by keeping untested keywords neatly separated from those publicly
!!!>        available
!!!> \param section ...
!!!> \param error ...
!!!> \par History
!!!>       2014.10 created [Rustam Z Khaliullin]
!!!> \author Rustam Z Khaliullin
!!! **************************************************************************************************
!!  SUBROUTINE create_developer_section(section)
!!
!!    TYPE(section_type), POINTER              :: section
!!
!!    CHARACTER(len=*), PARAMETER :: routineN = 'create_developer_section', &
!!      routineP = moduleN//':'//routineN
!!
!!    TYPE(keyword_type), POINTER              :: keyword
!!
!!       CALL section_create(section,"DEVELOPER",&
!!            description="Developer section for all ALMO-based methods. "//&
!!                        "Allows uninterrupted development of the code "//&
!!                        "by keeping untested keywords neatly separated from those publicly available",&
!!            n_keywords=3, n_subsections=0, repeats=.FALSE.)
!!
!!
!!       NULLIFY (keyword)
!!
!!  END SUBROUTINE create_developer_section

END MODULE input_cp2k_almo
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief creates the mm section of the input
!> \note
!>      moved out of input_cp2k
!> \par History
!>      04.2004 created
!> \author fawzi
! **************************************************************************************************
MODULE input_cp2k_mm
   USE bibliography,                    ONLY: &
        Devynck2012, Dick1958, Foiles1986, Mitchell1993, QUIP_ref, Siepmann1995, Tersoff1988, &
        Tosi1964a, Tosi1964b, Yamada2000
   USE cp_output_handling,              ONLY: cp_print_key_section_create,&
                                              debug_print_level,&
                                              high_print_level,&
                                              low_print_level,&
                                              medium_print_level,&
                                              silent_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE force_field_kind_types,          ONLY: &
        do_ff_amber, do_ff_charmm, do_ff_cubic, do_ff_fues, do_ff_g87, do_ff_g96, do_ff_harmonic, &
        do_ff_legendre, do_ff_mixed_bend_stretch, do_ff_mm2, do_ff_mm3, do_ff_mm4, do_ff_morse, &
        do_ff_opls, do_ff_quartic, do_ff_undef
   USE input_constants,                 ONLY: use_mom_ref_coac,&
                                              use_mom_ref_com,&
                                              use_mom_ref_user,&
                                              use_mom_ref_zero
   USE input_cp2k_poisson,              ONLY: create_poisson_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              lchar_t,&
                                              real_t
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_mm'

   PUBLIC :: create_mm_section, create_dipoles_section
   PUBLIC :: create_NONBONDED14_section, create_LJ_section, create_Williams_section, &
             create_Goodwin_section, &
             create_GENPOT_section, create_neighbor_lists_section
   PUBLIC :: create_CHARGE_section
!***
CONTAINS

! **************************************************************************************************
!> \brief Create the input section for FIST.. Come on.. Let's get woohooo
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_mm_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mm_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",78)
      CALL section_create(section, name="mm", &
                          description="This section contains all information to run a MM calculation.", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (subsection)

      CALL create_forcefield_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_neighbor_lists_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_poisson_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_print_mm_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_mm_section

! **************************************************************************************************
!> \brief Create the print mm section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_print_mm_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_print_mm_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",117)
      CALL section_create(section, name="print", &
                          description="Section of possible print options in MM code.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      NULLIFY (print_key, keyword)

      CALL cp_print_key_section_create(print_key, "DERIVATIVES", &
                                       description="Controls the printing of derivatives.", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "EWALD_INFO", &
                                       description="Controls the printing of Ewald energy components during the "// &
                                       "evaluation of the electrostatics.", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL create_dipoles_section(print_key, "DIPOLE", medium_print_level)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "NEIGHBOR_LISTS", &
                                       description="Activates the printing of the neighbor lists.", &
                                       print_level=high_print_level, filename="", unit_str="angstrom")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "ITER_INFO", &
                                       description="Activates the printing of iteration info during the self-consistent "// &
                                       "calculation of a polarizable forcefield.", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "SUBCELL", &
                                       description="Activates the printing of the subcells used for the"// &
                                       "generation of neighbor lists.", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "PROGRAM_BANNER", &
                                       description="Controls the printing of the banner of the MM program", &
                                       print_level=silent_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing of information regarding the run.", &
                                       print_level=low_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "FF_PARAMETER_FILE", description= &
                                       "Controls the printing of Force Field parameter file", &
                                       print_level=debug_print_level+1, filename="", common_iter_levels=2)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "FF_INFO", description= &
                                       "Controls the printing of information in the forcefield settings", &
                                       print_level=high_print_level, filename="__STD_OUT__")

      CALL keyword_create(keyword, "spline_info", &
                          description="if the printkey is active prints information regarding the splines"// &
                          " used in the nonbonded interactions", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, "spline_data", &
                          description="if the printkey is active prints on separated files the splined function"// &
                          " together with the reference one. Useful to check the spline behavior.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_print_mm_section

! **************************************************************************************************
!> \brief Create the forcefield section. This section is useful to set up the
!>      proper force_field for FIST calculations
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_forcefield_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_forcefield_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",217)
      CALL section_create(section, name="FORCEFIELD", &
                          description="Section specifying information regarding how to set up properly"// &
                          " a force_field for the classical calculations.", &
                          n_keywords=2, n_subsections=2, repeats=.FALSE.)

      NULLIFY (subsection, keyword)

      CALL keyword_create( &
         keyword, name="PARMTYPE", &
         description="Define the kind of torsion potential", &
         usage="PARMTYPE {OFF,CHM,G87,G96}", &
         enum_c_vals=s2a("OFF", "CHM", "G87", "G96", "AMBER"), &
         enum_desc=s2a("Provides force field parameters through the input file", &
                       "Provides force field parameters through an external file with CHARMM format", &
                       "Provides force field parameters through an external file with GROMOS 87 format", &
                       "Provides force field parameters through an external file with GROMOS 96 format", &
                       "Provides force field parameters through an external file with AMBER format (from v.8 on)"), &
         enum_i_vals=(/do_ff_undef, &
                       do_ff_charmm, &
                       do_ff_g87, &
                       do_ff_g96, &
                       do_ff_amber/), &
         default_i_val=do_ff_undef)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARM_FILE_NAME", &
                          description="Specifies the filename that contains the parameters of the FF.", &
                          usage="PARM_FILE_NAME {FILENAME}", type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VDW_SCALE14", &
                          description="Scaling factor for the VDW 1-4 ", &
                          usage="VDW_SCALE14 1.0", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EI_SCALE14", &
                          description="Scaling factor for the electrostatics 1-4 ", &
                          usage="EI_SCALE14 1.0", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SHIFT_CUTOFF", &
                          description="Add a constant energy shift to the real-space "// &
                          "non-bonding interactions (both Van der Waals and "// &
                          "electrostatic) such that the energy at the cutoff radius is "// &
                          "zero. This makes the non-bonding interactions continuous at "// &
                          "the cutoff.", &
                          usage="SHIFT_CUTOFF <LOGICAL>", default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DO_NONBONDED", &
                          description="Controls the computation of all the real-space "// &
                          "(short-range) nonbonded interactions. This also "// &
                          "includes the real-space corrections for excluded "// &
                          "or scaled 1-2, 1-3 and 1-4 interactions. When set "// &
                          "to F, the neighborlists are not created and all "// &
                          "interactions that depend on them are not computed.", &
                          usage="DO_NONBONDED T", default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="IGNORE_MISSING_CRITICAL_PARAMS", &
                          description="Do not abort when critical force-field parameters "// &
                          "are missing. CP2K will run as if the terms containing the "// &
                          "missing parameters are zero.", &
                          usage="IGNORE_MISSING_BOND_PARAMS T", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MULTIPLE_POTENTIAL", &
                          description="Enables the possibility to define NONBONDED and NONBONDED14 as a"// &
                          " sum of different kinds of potential. Useful for piecewise defined potentials.", &
                          usage="MULTIPLE_POTENTIAL T", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      !Universal scattering potential at very short distances
      CALL keyword_create(keyword, name="ZBL_SCATTERING", &
                          description="A short range repulsive potential is added, to simulate "// &
                          "collisions and scattering.", &
                          usage="ZBL_SCATTERING T", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      !
      ! subsections
      !
      CALL create_SPLINE_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_NONBONDED_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_NONBONDED14_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_CHARGE_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_CHARGES_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_SHELL_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_BOND_section(subsection, "BOND")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_BEND_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_TORSION_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_IMPROPER_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_OPBEND_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_DIPOLE_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_QUADRUPOLE_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_forcefield_section

! **************************************************************************************************
!> \brief This section specifies the parameters for the splines
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_SPLINE_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_SPLINE_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",376)
      CALL section_create(section, name="SPLINE", &
                          description="specifies parameters to set up the splines used in the"// &
                          " nonboned interactions (both pair body potential and many body potential)", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="R0_NB", &
                          description="Specify the minimum value of the distance interval "// &
                          " that brackets the value of emax_spline.", &
                          usage="R0_NB <REAL>", default_r_val=cp_unit_to_cp2k(value=0.9_dp, &
                                                                              unit_str="bohr"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT_NB", &
                          description="Cutoff radius for nonbonded interactions. This value overrides "// &
                          " the value specified in the potential definition and is global for all potentials.", &
                          usage="RCUT_NB {real}", default_r_val=cp_unit_to_cp2k(value=-1.0_dp, &
                                                                                unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EMAX_SPLINE", &
                          description="Specify the maximum value of the potential up to which"// &
                          " splines will be constructed", &
                          usage="EMAX_SPLINE <REAL>", &
                          default_r_val=0.5_dp, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EMAX_ACCURACY", &
                          description="Specify the maximum value of energy used to check the accuracy"// &
                          " requested through EPS_SPLINE. Energy values larger than EMAX_ACCURACY"// &
                          " generally do not  satisfy the requested accuracy", &
                          usage="EMAX_ACCURACY <REAL>", default_r_val=0.02_dp, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_SPLINE", &
                          description="Specify the threshold for the choice of the number of"// &
                          " points used in the splines (comparing the splined value with the "// &
                          " analytically evaluated one)", &
                          usage="EPS_SPLINE <REAL>", default_r_val=1.0E-7_dp, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="NPOINTS", &
         description="Override the default search for an accurate spline by specifying a fixed number of spline points.", &
         usage="NPOINTS 1024", default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="UNIQUE_SPLINE", &
                          description="For few potentials (Lennard-Jones) one global optimal spline is generated instead"// &
                          " of different optimal splines for each kind of potential", &
                          usage="UNIQUE_SPLINE <LOGICAL>", lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_SPLINE_section

! **************************************************************************************************
!> \brief This section specifies the torsion of the MM atoms
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_TORSION_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_TORSION_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",455)
      CALL section_create(section, name="TORSION", &
                          description="Specifies the torsion potential of the MM system.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kinds involved in the tors.", &
                          usage="ATOMS {KIND1} {KIND2} {KIND3} {KIND4}", type_of_var=char_t, &
                          n_var=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="KIND", &
                          description="Define the kind of torsion potential", &
                          usage="KIND CHARMM", &
                          enum_c_vals=s2a("CHARMM", "G87", "G96", "AMBER", "OPLS"), &
                          enum_desc=s2a("Functional Form (CHARMM|G87|G96|AMBER): K * [ 1 +  cos[M*PHI - PHI0]]", &
                                        "Functional Form (CHARMM|G87|G96|AMBER): K * [ 1 +  cos[M*PHI - PHI0]]", &
                                        "Functional Form (CHARMM|G87|G96|AMBER): K * [ 1 +  cos[M*PHI - PHI0]]", &
                                        "Functional Form (CHARMM|G87|G96|AMBER): K * [ 1 +  cos[M*PHI - PHI0]]", &
                                        "Functional Form: K / 2 * [ 1 + (-1)^(M-1) * cos[M*PHI]]"), &
                          enum_i_vals=(/do_ff_charmm, &
                                        do_ff_g87, &
                                        do_ff_g96, &
                                        do_ff_amber, &
                                        do_ff_opls/), &
                          default_i_val=do_ff_charmm)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K", &
                          description="Defines the force constant of the potential", &
                          usage="K {real}", type_of_var=real_t, &
                          n_var=1, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PHI0", &
                          description="Defines the phase of the potential.", &
                          usage="PHI0 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="rad", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="M", &
                          description="Defines the multiplicity of the potential.", &
                          usage="M {integer}", type_of_var=integer_t, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_TORSION_section

! **************************************************************************************************
!> \brief This section specifies the improper torsion of the MM atoms
!> \param section the section to create
!> \author louis vanduyfhuys
! **************************************************************************************************
   SUBROUTINE create_IMPROPER_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_IMPROPER_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",522)
      CALL section_create(section, name="IMPROPER", &
                          description="Specifies the improper torsion potential of the MM system.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kinds involved in the improper tors.", &
                          usage="ATOMS {KIND1} {KIND2} {KIND3} {KIND4}", type_of_var=char_t, &
                          n_var=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="KIND", &
                          description="Define the kind of improper torsion potential", &
                          usage="KIND CHARMM", &
                          enum_c_vals=s2a("CHARMM", "G87", "G96", "HARMONIC"), &
                          enum_desc=s2a("Functional Form (CHARMM): K * [ PHI - PHI0 ]**2", &
                                        "Functional Form (G87|G96|HARMONIC): 0.5 * K * [ PHI - PHI0 ]**2", &
                                        "Functional Form (G87|G96|HARMONIC): 0.5 * K * [ PHI - PHI0 ]**2", &
                                        "Functional Form (G87|G96|HARMONIC): 0.5 * K * [ PHI - PHI0 ]**2"), &
                          enum_i_vals=(/do_ff_charmm, &
                                        do_ff_g87, &
                                        do_ff_g96, &
                                        do_ff_harmonic/), &
                          default_i_val=do_ff_charmm)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K", &
                          description="Defines the force constant of the potential", &
                          usage="K {real}", type_of_var=real_t, &
                          n_var=1, unit_str="hartree*rad^-2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PHI0", &
                          description="Defines the phase of the potential.", &
                          usage="PHI0 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="rad")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_IMPROPER_section

! **************************************************************************************************
!> \brief This section specifies the out of plane bend of the MM atoms
!> \param section the section to create
!> \author louis vanduyfhuys
! **************************************************************************************************
   SUBROUTINE create_OPBEND_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_OPBEND_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",580)
      CALL section_create(section, name="OPBEND", &
                          description="Specifies the out of plane bend potential of the MM system."// &
                          "(Only defined for atom quadruples which are also defined as an improper"// &
                          " pattern in the topology.)", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kinds involved in the opbend.", &
                          usage="ATOMS {KIND1} {KIND2} {KIND3} {KIND4}", type_of_var=char_t, &
                          n_var=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="KIND", &
                          description="Define the kind of out of plane bend potential", &
                          usage="KIND HARMONIC", &
                          enum_c_vals=s2a("HARMONIC", "MM2", "MM3", "MM4"), &
                          enum_desc=s2a("Functional Form (HARMONIC): 0.5 * K * [ PHI - PHI0 ]**2", &
                                        "Functional Form (MM2|MM3|MM4): K * [ PHI - PHI0 ]**2", &
                                        "Functional Form (MM2|MM3|MM4): K * [ PHI - PHI0 ]**2", &
                                        "Functional Form (MM2|MM3|MM4): K * [ PHI - PHI0 ]**2"), &
                          enum_i_vals=(/do_ff_harmonic, &
                                        do_ff_mm2, &
                                        do_ff_mm3, &
                                        do_ff_mm4/), &
                          default_i_val=do_ff_harmonic)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K", &
                          description="Defines the force constant of the potential", &
                          usage="K {real}", type_of_var=real_t, &
                          n_var=1, unit_str="hartree*rad^-2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PHI0", &
                          description="Defines the phase of the potential.", &
                          usage="PHI0 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="rad")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_OPBEND_section

! **************************************************************************************************
!> \brief This section specifies the bend of the MM atoms
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_BEND_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_BEND_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",641)
      CALL section_create(section, name="BEND", &
                          description="Specifies the bend potential of the MM system.", &
                          n_keywords=11, n_subsections=1, repeats=.TRUE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kinds involved in the bend.", &
                          usage="ATOMS {KIND1} {KIND2} {KIND3}", type_of_var=char_t, &
                          n_var=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, name="KIND", &
         description="Define the kind of bend potential", &
         usage="KIND HARMONIC", &
         enum_c_vals=s2a("HARMONIC", "CHARMM", "AMBER", "G87", "G96", "CUBIC", "MIXED_BEND_STRETCH", "MM3", &
                         "LEGENDRE"), &
         enum_desc=s2a("Functional Form (HARMONIC|G87): 1/2*K*(THETA-THETA0)^2", &
                       "Functional Form (CHARMM|AMBER): K*(THETA-THETA0)^2", &
                       "Functional Form (CHARMM|AMBER): K*(THETA-THETA0)^2", &
                       "Functional Form (HARMONIC|G87): 1/2*K*(THETA-THETA0)^2", &
                       "Functional Form (G96): 1/2*K*(COS(THETA)-THETA0)^2", &
                       "Functional Form (CUBIC): K*(THETA-THETA0)**2*(1+CB*(THETA-THETA0))", &
                       "Functional Form (MIXED_BEND_STRETCH): K*(THETA-THETA0)**2*(1+CB*(THETA-THETA0))+"// &
                       " KSS*(R12-R012)*(R32-R032)+KBS12*(R12-R012)*(THETA-THETA0)+KBS32*(R32-R032)*(THETA-THETA0)", &
                       "Functional Form (MM3): 1/2*K*(THETA-THETA0)**2*(1-0.014*(THETA-THETA0)+5.6E-5*(THETA-THETA0)**2"// &
                       " -7.0E-7*(THETA-THETA0)**3+9.0E-10*(THETA-THETA0)**4)+KBS12*(R12-R012)*(THETA-THETA0)+"// &
                       " KBS32*(R32-R032)*(THETA-THETA0)", &
                       "Functional Form (LEGENDRE): sum_{i=0}^N c_i*P_i(COS(THETA)) "), &
         enum_i_vals=(/do_ff_harmonic, &
                       do_ff_charmm, &
                       do_ff_amber, &
                       do_ff_g87, &
                       do_ff_g96, &
                       do_ff_cubic, &
                       do_ff_mixed_bend_stretch, &
                       do_ff_mm3, &
                       do_ff_legendre/), &
         default_i_val=do_ff_charmm)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K", &
                          description="Defines the force constant of the potential", &
                          usage="K {real}", type_of_var=real_t, &
                          n_var=1, unit_str="hartree*rad^-2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CB", &
                          description="Defines the the cubic force constant of the bend", &
                          usage="CB {real}", default_r_val=0.0_dp, type_of_var=real_t, &
                          n_var=1, unit_str="rad^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="R012", &
                          description="Mixed bend stretch parameter", &
                          usage="R012 {real}", default_r_val=0.0_dp, type_of_var=real_t, &
                          n_var=1, unit_str="bohr")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="R032", &
                          description="Mixed bend stretch parameter", &
                          usage="R032 {real}", default_r_val=0.0_dp, type_of_var=real_t, &
                          n_var=1, unit_str="bohr")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="KBS12", &
                          description="Mixed bend stretch parameter", &
                          usage="KBS12 {real}", default_r_val=0.0_dp, type_of_var=real_t, &
                          n_var=1, unit_str="hartree*bohr^-1*rad^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="KBS32", &
                          description="Mixed bend stretch parameter", &
                          usage="KBS32 {real}", default_r_val=0.0_dp, type_of_var=real_t, &
                          n_var=1, unit_str="hartree*bohr^-1*rad^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="KSS", &
                          description="Mixed bend stretch parameter", &
                          usage="KSS {real}", default_r_val=0.0_dp, type_of_var=real_t, &
                          n_var=1, unit_str="hartree*bohr^-2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="THETA0", &
                          description="Defines the equilibrium angle.", &
                          usage="THETA0 {real}", type_of_var=real_t, &
                          n_var=1, unit_str='rad')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LEGENDRE", &
                          description="Specifies the coefficients for the legendre"// &
                          " expansion of the bending potential."// &
                          "'THETA0' and 'K' are not used, but need to be specified."// &
                          "Use an arbitrary value.", usage="LEGENDRE {REAL} {REAL} ...", &
                          default_r_val=0.0d0, type_of_var=real_t, &
                          n_var=-1, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Create the Urey-Bradley section
      CALL create_BOND_section(subsection, "UB")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_BEND_section

! **************************************************************************************************
!> \brief This section specifies the bond of the MM atoms
!> \param section the section to create
!> \param label ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_BOND_section(section, label)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*), INTENT(IN)                       :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'create_BOND_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: tag
      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",771)
      NULLIFY (keyword)

      IF (TRIM(label) == "UB") THEN
         tag = " Urey-Bradley "
         CALL section_create(section, name=TRIM(label), &
                             description="Specifies the Urey-Bradley potential between the external atoms"// &
                             " defining the angle", &
                             n_keywords=1, n_subsections=0, repeats=.FALSE.)

      ELSE
         tag = " Bond "
         CALL section_create(section, name=TRIM(label), &
                             description="Specifies the bond potential", &
                             n_keywords=1, n_subsections=0, repeats=.TRUE.)

         CALL keyword_create(keyword, name="ATOMS", &
                             description="Defines the atomic kinds involved in the bond.", &
                             usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                             n_var=2)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)
      END IF

      CALL keyword_create(keyword=keyword, name="KIND", &
                          description="Define the kind of"//TRIM(tag)//"potential.", &
                          usage="KIND HARMONIC", &
                          enum_c_vals=s2a("HARMONIC", "CHARMM", "AMBER", "G87", "G96", "QUARTIC", &
                                          "MORSE", "CUBIC", "FUES"), &
                          enum_desc=s2a("Functional Form (HARMONIC|G87): 1/2*K*(R-R0)^2", &
                                        "Functional Form (CHARMM|AMBER): K*(R-R0)^2", &
                                        "Functional Form (CHARMM|AMBER): K*(R-R0)^2", &
                                        "Functional Form (HARMONIC|G87): 1/2*K*(R-R0)^2", &
                                        "Functional Form (G96): 1/4*K*(R^2-R0^2)^2", &
                                        "Functional Form (QUARTIC): (1/2*K1+[1/3*K2+1/4*K3*|R-R0|]*|R-R0|)(R-R0)^2", &
                                        "Functional Form (MORSE): K1*[(1-exp(-K2*(R-R0)))^2-1])", &
                                        "Functional Form (CUBIC): K*(R-R0)^2*(1+cs*(R-R0)+7/12*(cs^2*(R-R0)^2))", &
                                        "Functional Form (FUES): 1/2*K*R0^2*(1+R0/R*(R0/R-2))"), &
                          enum_i_vals=(/do_ff_harmonic, &
                                        do_ff_charmm, &
                                        do_ff_amber, &
                                        do_ff_g87, &
                                        do_ff_g96, &
                                        do_ff_quartic, &
                                        do_ff_morse, &
                                        do_ff_cubic, &
                                        do_ff_fues/), &
                          default_i_val=do_ff_charmm)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K", &
                          description="Defines the force constant of the potential. "// &
                          "For MORSE potentials 2 numbers are expected. "// &
                          "For QUARTIC potentials 3 numbers are expected.", &
                          usage="K {real}", type_of_var=real_t, &
                          n_var=-1, unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CS", &
                          description="Defines the cubic stretch term.", &
                          usage="CS {real}", default_r_val=0.0_dp, type_of_var=real_t, &
                          n_var=1, unit_str="bohr^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="R0", &
                          description="Defines the equilibrium distance.", &
                          usage="R0 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="bohr")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_BOND_section

! **************************************************************************************************
!> \brief This section specifies the charge of the MM atoms
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_charges_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_charges_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",860)
      CALL section_create(section, name="charges", &
                          description="Allow to specify an array of classical charges, thus avoiding the"// &
                          " packing and permitting the usage of different charges for same atomic types.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Value of the charge for the individual atom. Order MUST reflect"// &
                          " the one specified for the geometry.", repeats=.TRUE., usage="{Real}", &
                          type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_charges_section

! **************************************************************************************************
!> \brief This section specifies the charge of the MM atoms
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_charge_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_charge_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",889)
      CALL section_create(section, name="charge", &
                          description="This section specifies the charge of the MM atoms", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOM", &
                          description="Defines the atomic kind of the charge.", &
                          usage="ATOM {KIND1}", type_of_var=char_t, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHARGE", &
                          description="Defines the charge of the MM atom in electron charge unit.", &
                          usage="CHARGE {real}", type_of_var=real_t, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_charge_section

! **************************************************************************************************
!> \brief This section specifies the isotropic polarizability of the MM atoms
!> \param section the section to create
!> \author Marcel Baer
! **************************************************************************************************
   SUBROUTINE create_quadrupole_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_quadrupole_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",925)
      CALL section_create( &
         section, name="QUADRUPOLE", &
         description="This section specifies that we will perform an SCF quadrupole calculation of the MM atoms. "// &
         "Needs KEYWORD POL_SCF in POISSON secton", &
         n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOM", &
                          description="Defines the atomic kind of the SCF quadrupole.", &
                          usage="ATOM {KIND1}", type_of_var=char_t, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CPOL", &
                          description="Defines the isotropic polarizability of the MM atom.", &
                          usage="CPOL {real}", type_of_var=real_t, &
                          n_var=1, unit_str='internal_cp2k')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_quadrupole_section

! **************************************************************************************************
!> \brief This section specifies the isotropic polarizability of the MM atoms
!> \param section the section to create
!> \author Marcel Baer
! **************************************************************************************************
   SUBROUTINE create_dipole_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_dipole_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",964)
      CALL section_create(section, name="DIPOLE", &
                          description="This section specifies that we will perform an SCF dipole calculation of the MM atoms. "// &
                          "Needs KEYWORD POL_SCF in POISSON secton", &
                          n_keywords=1, n_subsections=1, repeats=.TRUE.)

      NULLIFY (subsection, keyword)

      CALL keyword_create(keyword, name="ATOM", &
                          description="Defines the atomic kind of the SCF dipole.", &
                          usage="ATOM {KIND1}", type_of_var=char_t, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="APOL", &
                          description="Defines the isotropic polarizability of the MM atom.", &
                          usage="APOL {real}", type_of_var=real_t, &
                          n_var=1, unit_str='angstrom^3')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_DAMPING_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_dipole_section

! **************************************************************************************************
!> \brief This section specifies the idamping parameters for polarizable atoms
!> \param section the section to create
!> \author Rodolphe Vuilleumier
! **************************************************************************************************
   SUBROUTINE create_damping_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_damping_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1004)
      CALL section_create(section, name="DAMPING", &
                          description="This section specifies optional electric field damping for the polarizable atoms. ", &
                          n_keywords=4, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOM", &
                          description="Defines the atomic kind for this damping function.", &
                          usage="ATOM {KIND1}", type_of_var=char_t, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TYPE", &
                          description="Defines the damping type.", &
                          usage="TYPE {string}", type_of_var=char_t, &
                          n_var=1, default_c_val="TANG-TOENNIES")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ORDER", &
                          description="Defines the order for this damping.", &
                          usage="ORDER {integer}", type_of_var=integer_t, &
                          n_var=1, default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BIJ", &
                          description="Defines the BIJ parameter for this damping.", &
                          usage="BIJ {real}", type_of_var=real_t, &
                          n_var=1, unit_str='angstrom^-1')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CIJ", &
                          description="Defines the CIJ parameter for this damping.", &
                          usage="CIJ {real}", type_of_var=real_t, &
                          n_var=1, unit_str='')
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_damping_section

! **************************************************************************************************
!> \brief This section specifies the charge of the MM atoms
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_shell_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_shell_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1061)
      CALL section_create(section, name="SHELL", &
                          description="This section specifies the parameters for shell-model potentials", &
                          n_keywords=6, n_subsections=0, repeats=.TRUE., &
                          citations=(/Dick1958, Mitchell1993, Devynck2012/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="The kind for which the shell potential parameters are given ", &
                          usage="H", default_c_val="DEFAULT")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CORE_CHARGE", &
                          variants=(/"CORE"/), &
                          description="Partial charge assigned to the core (electron charge units)", &
                          usage="CORE_CHARGE {real}", &
                          default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SHELL_CHARGE", &
                          variants=(/"SHELL"/), &
                          description="Partial charge assigned to the shell (electron charge units)", &
                          usage="SHELL_CHARGE {real}", &
                          default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MASS_FRACTION", &
                          variants=(/"MASS"/), &
                          description="Fraction of the mass of the atom to be assigned to the shell", &
                          usage="MASS_FRACTION {real}", &
                          default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K2_SPRING", &
                          variants=s2a("K2", "SPRING"), &
                          description="Force constant k2 of the spring potential 1/2*k2*r^2 + 1/24*k4*r^4 "// &
                          "binding a core-shell pair when a core-shell potential is employed.", &
                          repeats=.FALSE., &
                          usage="K2_SPRING {real}", &
                          default_r_val=-1.0_dp, &
                          unit_str="hartree*bohr^-2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="K4_SPRING", &
                          variants=s2a("K4"), &
                          description="Force constant k4 of the spring potential 1/2*k2*r^2 + 1/24*k4*r^4 "// &
                          "binding a core-shell pair when a core-shell potential is employed. "// &
                          "By default a harmonic spring potential is used, i.e. k4 is zero.", &
                          repeats=.FALSE., &
                          usage="K4_SPRING {real}", &
                          default_r_val=0.0_dp, &
                          unit_str="hartree*bohr^-4")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_DISTANCE", &
                          description="Assign a maximum elongation of the spring, "// &
                          "if negative no limit is imposed", &
                          usage="MAX_DISTANCE 0.0", &
                          default_r_val=-1.0_dp, &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SHELL_CUTOFF", &
                          description="Define a screening function to exclude some neighbors "// &
                          " of the shell when electrostatic interaction are considered, "// &
                          "if negative no screening is operated", &
                          usage="SHELL_CUTOFF -1.0", &
                          default_r_val=-1.0_dp, &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_shell_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for 1-4 NON-BONDED
!>      Interactions
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_NONBONDED14_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_NONBONDED14_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1157)
      CALL section_create(section, name="nonbonded14", &
                          description="This section specifies the input parameters for 1-4 NON-BONDED interactions.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (subsection)
      CALL create_LJ_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_Williams_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_Goodwin_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_GENPOT_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_NONBONDED14_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for 1-4 NON-BONDED
!>      Interactions
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_NONBONDED_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_NONBONDED_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1195)
      CALL section_create(section, name="nonbonded", &
                          description="This section specifies the input parameters for NON-BONDED interactions.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (subsection)
      CALL create_LJ_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_Williams_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_EAM_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_QUIP_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_Goodwin_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_IPBV_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_BMHFT_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_BMHFTD_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_Buck4r_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_Buckmorse_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_GENPOT_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_Tersoff_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_Siepmann_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_NONBONDED_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for generation of
!>      neighbor lists
!> \param section the section to create
!> \author teo [07.2007] - Zurich University
! **************************************************************************************************
   SUBROUTINE create_neighbor_lists_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_neighbor_lists_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1270)
      CALL section_create(section, name="neighbor_lists", &
                          description="This section specifies the input parameters for the construction of"// &
                          " neighbor lists.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="VERLET_SKIN", &
                          description="Defines the Verlet Skin for the generation of the neighbor lists", &
                          usage="VERLET_SKIN {real}", default_r_val=cp_unit_to_cp2k(value=1.0_dp, &
                                                                                    unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="neighbor_lists_from_scratch", &
                          description="This keyword enables the building of the neighbouring list from scratch.", &
                          usage="neighbor_lists_from_scratch logical", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GEO_CHECK", &
                          description="This keyword enables the check that two atoms are never below the minimum"// &
                          " value used to construct the splines during the construction of the neighbouring list."// &
                          " Disabling this keyword avoids CP2K to abort in case two atoms are below the minimum "// &
                          " value of the radius used to generate the splines.", &
                          usage="GEO_CHECK", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_neighbor_lists_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for a generic potential form
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_GENPOT_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_GENPOT_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1316)
      CALL section_create(section, name="GENPOT", &
                          description="This section specifies the input parameters for a generic potential type."// &
                          "A functional form is specified. Mathematical Operators recognized are +, -, *, /, ** "// &
                          "or alternatively ^, whereas symbols for brackets must be (). "// &
                          "The function parser recognizes the (single argument) Fortran 90 intrinsic functions "// &
                          "abs, exp, log10, log, sqrt, sinh, cosh, tanh, sin, cos, tan, asin, acos, atan. "// &
                          "Parsing for INTRINSIC functions is CASE INsensitive.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the generic potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FUNCTION", &
                          description="Specifies the functional form in mathematical notation.", &
                          usage="FUNCTION  a*EXP(-b*x^2)/x+D*log10(x)", type_of_var=lchar_t, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VARIABLES", &
                          description="Defines the variable of the functional form.", &
                          usage="VARIABLES x", type_of_var=char_t, &
                          n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARAMETERS", &
                          description="Defines the parameters of the functional form", &
                          usage="PARAMETERS a b D", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VALUES", &
                          description="Defines the values of parameter of the functional form", &
                          usage="VALUES ", type_of_var=real_t, &
                          n_var=-1, repeats=.TRUE., unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="UNITS", &
                          description="Optionally, allows to define valid CP2K unit strings for each parameter value. "// &
                          "It is assumed that the corresponding parameter value is specified in this unit.", &
                          usage="UNITS angstrom eV*angstrom^-1 angstrom^1 K", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Defines the cutoff parameter of the generic potential", &
                          usage="RCUT {real}", default_r_val=cp_unit_to_cp2k(value=10.0_dp, &
                                                                             unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMIN", &
                          description="Defines the lower bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMIN {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMAX", &
                          description="Defines the upper bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMAX {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_GENPOT_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for EAM  potential type
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_EAM_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_EAM_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1408)
      CALL section_create(section, name="EAM", &
                          description="This section specifies the input parameters for EAM potential type.", &
                          citations=(/Foiles1986/), n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the nonbond potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARM_FILE_NAME", &
                          variants=(/"PARMFILE"/), &
                          description="Specifies the filename that contains the tabulated EAM potential. "// &
                          "File structure: the first line of the potential file contains a title. "// &
                          "The second line contains: atomic number, mass and lattice constant. "// &
                          "These information are parsed but not used in CP2K. The third line contains: "// &
                          "dr: increment of r for the tabulated values of density and phi (assuming r starts in 0) [angstrom]; "// &
                          "drho: increment of density for the tabulated values of the embedding function (assuming rho starts "// &
                          "in 0) [au_c]; cutoff: cutoff of the EAM potential; npoints: number of points in tabulated. Follow "// &
                          "in order npoints lines for rho [au_c] and its derivative [au_c*angstrom^-1]; npoints lines for "// &
                          "PHI [ev] and its derivative [ev*angstrom^-1] and npoint lines for the embedded function [ev] "// &
                          "and its derivative [ev*au_c^-1].", &
                          usage="PARM_FILE_NAME {FILENAME}", default_lc_val=" ")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_EAM_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for QUIP  potential type
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_QUIP_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_QUIP_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1453)
      CALL section_create(section, name="QUIP", &
                          description="This section specifies the input parameters for QUIP potential type. "// &
                          "Mainly intended for things like GAP corrections to DFT "// &
                          "to achieve correlated-wavefunction-like accuracy. "// &
                          "Requires linking with quip library from <a href=""http://www.libatoms.org"" "// &
                          "target=""_blank"">http://www.libatoms.org</a> .", &
                          citations=(/QUIP_ref/), n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kinds involved in the QUIP potential. "// &
                          "For more than 2 elements, &QUIP section must be repeated until each element "// &
                          "has been mentioned at least once. Set IGNORE_MISSING_CRITICAL_PARAMS to T "// &
                          "in enclosing &FORCEFIELD section to avoid having to list every pair of elements separately.", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARM_FILE_NAME", &
                          variants=(/"PARMFILE"/), &
                          description="Specifies the filename that contains the QUIP potential.", &
                          usage="PARM_FILE_NAME {FILENAME}", default_lc_val="quip_params.xml")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INIT_ARGS", &
                          description="Specifies the potential initialization arguments for the QUIP potential. "// &
                          "If blank (default) first potential defined in QUIP parameter file will be used.", &
                          usage="INIT_ARGS", default_c_vals=(/""/), &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CALC_ARGS", &
                          description="Specifies the potential calculation arguments for the QUIP potential.", &
                          usage="CALC_ARGS", default_c_vals=(/""/), &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_QUIP_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for Lennard-Jones potential type
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_LJ_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_LJ_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1511)
      CALL section_create(section, name="lennard-jones", &
                          description="This section specifies the input parameters for LENNARD-JONES potential type."// &
                          "Functional form: V(r) = 4.0 * EPSILON * [(SIGMA/r)^12-(SIGMA/r)^6].", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the nonbond potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPSILON", &
                          description="Defines the EPSILON parameter of the LJ potential", &
                          usage="EPSILON {real}", type_of_var=real_t, &
                          n_var=1, unit_str="K_e")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SIGMA", &
                          description="Defines the SIGMA parameter of the LJ potential", &
                          usage="SIGMA {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Defines the cutoff parameter of the LJ potential", &
                          usage="RCUT {real}", default_r_val=cp_unit_to_cp2k(value=10.0_dp, &
                                                                             unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMIN", &
                          description="Defines the lower bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMIN {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMAX", &
                          description="Defines the upper bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMAX {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_LJ_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for Williams potential type
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_Williams_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_Williams_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1577)
      CALL section_create(section, name="williams", &
                          description="This section specifies the input parameters for WILLIAMS potential type."// &
                          "Functional form: V(r) = A*EXP(-B*r) - C / r^6 .", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the nonbond potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="A", &
                          description="Defines the A parameter of the Williams potential", &
                          usage="A {real}", type_of_var=real_t, &
                          n_var=1, unit_str="K_e")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="B", &
                          description="Defines the B parameter of the Williams potential", &
                          usage="B {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="C", &
                          description="Defines the C parameter of the Williams potential", &
                          usage="C {real}", type_of_var=real_t, &
                          n_var=1, unit_str="K_e*angstrom^6")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Defines the cutoff parameter of the Williams potential", &
                          usage="RCUT {real}", default_r_val=cp_unit_to_cp2k(value=10.0_dp, &
                                                                             unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMIN", &
                          description="Defines the lower bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMIN {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMAX", &
                          description="Defines the upper bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMAX {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_Williams_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for Goodwin potential type
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_Goodwin_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_Goodwin_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1650)
      CALL section_create(section, name="goodwin", &
                          description="This section specifies the input parameters for GOODWIN potential type."// &
                          "Functional form: V(r) = EXP(M*(-(r/DC)**MC+(D/DC)**MC))*VR0*(D/r)**M.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the nonbond potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VR0", &
                          description="Defines the VR0 parameter of the Goodwin potential", &
                          usage="VR0 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="K_e")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="D", &
                          description="Defines the D parameter of the Goodwin potential", &
                          usage="D {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DC", &
                          description="Defines the DC parameter of the Goodwin potential", &
                          usage="DC {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="M", &
                          description="Defines the M parameter of the Goodwin potential", &
                          usage="M {real}", type_of_var=integer_t, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MC", &
                          description="Defines the MC parameter of the Goodwin potential", &
                          usage="MC {real}", type_of_var=integer_t, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Defines the cutoff parameter of the Goodwin potential", &
                          usage="RCUT {real}", default_r_val=cp_unit_to_cp2k(value=10.0_dp, &
                                                                             unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMIN", &
                          description="Defines the lower bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMIN {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMAX", &
                          description="Defines the upper bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMAX {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_Goodwin_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for IPBV potential type
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_ipbv_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ipbv_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1736)
      CALL section_create(section, name="ipbv", &
                          description="This section specifies the input parameters for IPBV potential type."// &
                          "Functional form: Implicit table function.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the IPBV nonbond potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Defines the cutoff parameter of the IPBV potential", &
                          usage="RCUT {real}", default_r_val=cp_unit_to_cp2k(value=10.0_dp, &
                                                                             unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMIN", &
                          description="Defines the lower bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMIN {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMAX", &
                          description="Defines the upper bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMAX {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_ipbv_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for BMHFT potential type
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_BMHFT_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_BMHFT_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1788)
      CALL section_create(section, name="BMHFT", &
                          description="This section specifies the input parameters for BMHFT potential type."// &
                          "Functional form: V(r) = A * EXP(-B*r) - C/r^6 - D/r^8."// &
                          "Values available inside cp2k only for the Na/Cl pair.", &
                          citations=(/Tosi1964a, Tosi1964b/), n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the BMHFT nonbond potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAP_ATOMS", &
                          description="Defines the kinds for which internally is defined the BMHFT nonbond potential"// &
                          " at the moment only Na and Cl.", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Defines the cutoff parameter of the BMHFT potential", &
                          usage="RCUT {real}", default_r_val=7.8_dp, &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="A", &
                          description="Defines the A parameter of the Fumi-Tosi Potential", &
                          usage="A {real}", type_of_var=real_t, &
                          n_var=1, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="B", &
                          description="Defines the B parameter of the Fumi-Tosi Potential", &
                          usage="B {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="C", &
                          description="Defines the C parameter of the Fumi-Tosi Potential", &
                          usage="C {real}", type_of_var=real_t, &
                          n_var=1, unit_str="hartree*angstrom^6")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="D", &
                          description="Defines the D parameter of the Fumi-Tosi Potential", &
                          usage="D {real}", type_of_var=real_t, &
                          n_var=1, unit_str="hartree*angstrom^8")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMIN", &
                          description="Defines the lower bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMIN {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMAX", &
                          description="Defines the upper bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMAX {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_BMHFT_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for BMHFTD potential type
!> \param section the section to create
!> \author Mathieu Salanne 05.2010
! **************************************************************************************************
   SUBROUTINE create_BMHFTD_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_BMHFTD_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1876)
      CALL section_create(section, name="BMHFTD", &
                          description="This section specifies the input parameters for BMHFTD potential type."// &
                          "Functional form: V(r) = A*exp(-B*r) - f_6*(r)C/r^6 - f_8(r)*D/r^8."// &
                          "where f_order(r)=1-exp(-BD * r) * \sum_{k=0}^order (BD * r)^k / k! ."// &
                          "(Tang-Toennies damping function)"// &
                          "No values available inside cp2k.", &
                          citations=(/Tosi1964a, Tosi1964b/), n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the BMHFTD nonbond potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAP_ATOMS", &
                          description="Defines the kinds for which internally is defined the BMHFTD nonbond potential"// &
                          " at the moment no species included.", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Defines the cutoff parameter of the BMHFTD potential", &
                          usage="RCUT {real}", default_r_val=7.8_dp, &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="A", &
                          description="Defines the A parameter of the dispersion-damped Fumi-Tosi Potential", &
                          usage="A {real}", type_of_var=real_t, &
                          n_var=1, unit_str="hartree")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="B", &
                          description="Defines the B parameter of the dispersion-damped Fumi-Tosi Potential", &
                          usage="B {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="C", &
                          description="Defines the C parameter of the dispersion-damped Fumi-Tosi Potential", &
                          usage="C {real}", type_of_var=real_t, &
                          n_var=1, unit_str="hartree*angstrom^6")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="D", &
                          description="Defines the D parameter of the dispersion-damped Fumi-Tosi Potential", &
                          usage="D {real}", type_of_var=real_t, &
                          n_var=1, unit_str="hartree*angstrom^8")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BD", &
                          description="Defines the BD parameter of the dispersion-damped Fumi-Tosi Potential", &
                          usage="D {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ORDER", &
                          description="Defines the order for this damping.", &
                          usage="ORDER {integer}", type_of_var=integer_t, &
                          n_var=1, default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMIN", &
                          description="Defines the lower bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMIN {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMAX", &
                          description="Defines the upper bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMAX {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_BMHFTD_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for Buckingham 4 ranges potential type
!> \param section the section to create
!> \author MI
! **************************************************************************************************
   SUBROUTINE create_Buck4r_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_Buck4r_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",1980)
      CALL section_create(section, name="BUCK4RANGES", &
                          description="This section specifies the input parameters for the Buckingham 4-ranges"// &
                          " potential type."// &
                          " Functional form:<ul>"// &
                          "<li>V(r) = A*EXP(-B*r) for r &lt; r<sub>1</sub></li>"// &
                          "<li>V(r) = Sum_n POLY1(n)*r<sup>n</sup> for r<sub>1</sub> &le; r &lt; r<sub>2</sub></li>"// &
                          "<li>V(r) = Sum_n POLY2(n)*r<sup>n</sup> for r<sub>2</sub> &le; r &lt; r<sub>3</sub></li>"// &
                          "<li>V(r) = -C/r<sup>6</sup> for r &ge; r<sub>3</sub></li></ul>", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the nonbond potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="A", &
                          description="Defines the A parameter of the Buckingham potential", &
                          usage="A {real}", type_of_var=real_t, &
                          n_var=1, unit_str="K_e")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="B", &
                          description="Defines the B parameter of the Buckingham potential", &
                          usage="B {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="C", &
                          description="Defines the C parameter of the Buckingham  potential", &
                          usage="C {real}", type_of_var=real_t, &
                          n_var=1, unit_str="K_e*angstrom^6")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="R1", &
                          description="Defines the upper bound of the first range ", &
                          usage="R1 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="R2", &
                          description="Defines the upper bound of the second range ", &
                          usage="R2 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="R3", &
                          description="Defines the upper bound of the third range ", &
                          usage="R3 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POLY1", &
                          description="Coefficients of the polynomial used in the second range"// &
                          "This keyword can be repeated several times.", &
                          usage="POLY1 C1 C2 C3 ..", &
                          n_var=-1, unit_str="K_e", type_of_var=real_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POLY2", &
                          description="Coefficients of the polynomial used in the third range"// &
                          "This keyword can be repeated several times.", &
                          usage="POLY1 C1 C2 C3 ..", &
                          n_var=-1, unit_str="K_e", type_of_var=real_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Defines the cutoff parameter of the Buckingham potential", &
                          usage="RCUT {real}", default_r_val=cp_unit_to_cp2k(value=10.0_dp, &
                                                                             unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMIN", &
                          description="Defines the lower bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMIN {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMAX", &
                          description="Defines the upper bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMAX {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_Buck4r_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for Buckingham + Morse potential type
!> \param section the section to create
!> \author MI
! **************************************************************************************************
   SUBROUTINE create_Buckmorse_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_Buckmorse_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",2095)
      CALL section_create( &
         section, name="BUCKMORSE", &
         description="This section specifies the input parameters for"// &
         " Buckingham plus Morse potential type "// &
         " Functional Form: V(r) = F0*(B1+B2)*EXP([A1+A2-r]/[B1+B2])-C/r^6+D*{EXP[-2*beta*(r-R0)]-2*EXP[-beta*(r-R0)]}.", &
         citations=(/Yamada2000/), n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the nonbond potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="F0", &
                          description="Defines the f0 parameter of Buckingham+Morse potential", &
                          usage="F0 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="K_e*angstrom^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="A1", &
                          description="Defines the A1 parameter of Buckingham+Morse potential", &
                          usage="A1 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="A2", &
                          description="Defines the A2 parameter of Buckingham+Morse potential", &
                          usage="A2 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="B1", &
                          description="Defines the B1 parameter of Buckingham+Morse potential", &
                          usage="B1 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="B2", &
                          description="Defines the B2 parameter of Buckingham+Morse potential", &
                          usage="B2 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="C", &
                          description="Defines the C parameter of Buckingham+Morse  potential", &
                          usage="C {real}", type_of_var=real_t, &
                          n_var=1, unit_str="K_e*angstrom^6")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="D", &
                          description="Defines the amplitude for the Morse part ", &
                          usage="D {real}", type_of_var=real_t, &
                          n_var=1, unit_str="K_e")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="R0", &
                          description="Defines the equilibrium distance for the Morse part ", &
                          usage="R0 {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="Beta", &
                          description="Defines the width for the Morse part ", &
                          usage="Beta {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Defines the cutoff parameter of the Buckingham potential", &
                          usage="RCUT {real}", default_r_val=cp_unit_to_cp2k(value=10.0_dp, &
                                                                             unit_str="angstrom"), &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMIN", &
                          description="Defines the lower bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMIN {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMAX", &
                          description="Defines the upper bound of the potential. If not set the range is the"// &
                          " full range generate by the spline", usage="RMAX {real}", &
                          type_of_var=real_t, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_Buckmorse_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for Tersoff potential type
!>      (Tersoff, J. PRB 39(8), 5566, 1989)
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_Tersoff_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_Tersoff_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",2212)
      CALL section_create(section, name="TERSOFF", &
                          description="This section specifies the input parameters for Tersoff potential type.", &
                          citations=(/Tersoff1988/), n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the nonbond potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="A", &
                          description="Defines the A parameter of Tersoff potential", &
                          usage="A {real}", type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=1.8308E3_dp, &
                                                        unit_str="eV"), &
                          n_var=1, unit_str="eV")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="B", &
                          description="Defines the B parameter of Tersoff potential", &
                          usage="B {real}", type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=4.7118E2_dp, &
                                                        unit_str="eV"), &
                          n_var=1, unit_str="eV")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="lambda1", &
                          description="Defines the lambda1 parameter of Tersoff potential", &
                          usage="lambda1 {real}", type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=2.4799_dp, &
                                                        unit_str="angstrom^-1"), &
                          n_var=1, unit_str="angstrom^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="lambda2", &
                          description="Defines the lambda2 parameter of Tersoff potential", &
                          usage="lambda2 {real}", type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=1.7322_dp, &
                                                        unit_str="angstrom^-1"), &
                          n_var=1, unit_str="angstrom^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="alpha", &
                          description="Defines the alpha parameter of Tersoff potential", &
                          usage="alpha {real}", type_of_var=real_t, &
                          default_r_val=0.0_dp, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="beta", &
                          description="Defines the beta parameter of Tersoff potential", &
                          usage="beta {real}", type_of_var=real_t, &
                          default_r_val=1.0999E-6_dp, &
                          n_var=1, unit_str="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="n", &
                          description="Defines the n parameter of Tersoff potential", &
                          usage="n {real}", type_of_var=real_t, &
                          default_r_val=7.8734E-1_dp, &
                          n_var=1, unit_str="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="c", &
                          description="Defines the c parameter of Tersoff potential", &
                          usage="c {real}", type_of_var=real_t, &
                          default_r_val=1.0039E5_dp, &
                          n_var=1, unit_str="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="d", &
                          description="Defines the d parameter of Tersoff potential", &
                          usage="d {real}", type_of_var=real_t, &
                          default_r_val=1.6218E1_dp, &
                          n_var=1, unit_str="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="h", &
                          description="Defines the h parameter of Tersoff potential", &
                          usage="h {real}", type_of_var=real_t, &
                          default_r_val=-5.9826E-1_dp, &
                          n_var=1, unit_str="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="lambda3", &
                          description="Defines the lambda3 parameter of Tersoff potential", &
                          usage="lambda3 {real}", type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=1.7322_dp, &
                                                        unit_str="angstrom^-1"), &
                          n_var=1, unit_str="angstrom^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="bigR", &
                          description="Defines the bigR parameter of Tersoff potential", &
                          usage="bigR {real}", type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=2.85_dp, &
                                                        unit_str="angstrom"), &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="bigD", &
                          description="Defines the D parameter of Tersoff potential", &
                          usage="bigD {real}", type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=0.15_dp, &
                                                        unit_str="angstrom"), &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Defines the cutoff parameter of the tersoff potential. "// &
                          " This parameter is in principle already defined by the values of "// &
                          " bigD and bigR. But it is necessary to define it when using the tersoff "// &
                          " in conjuction with other potentials (for the same atomic pair) in order to have"// &
                          " the same consistent definition of RCUT for all potentials.", &
                          usage="RCUT {real}", type_of_var=real_t, &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_Tersoff_section

! **************************************************************************************************
!> \brief This section specifies the input parameters for Siepmann-Sprik
!>        potential type
!>      (Siepmann and Sprik, J. Chem. Phys. 102(1) 511, 1995)
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_Siepmann_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_Siepmann_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mm.F",2364)
      CALL section_create(section, name="SIEPMANN", &
                          description="This section specifies the input parameters for the"// &
                          " Siepmann-Sprik potential type. Consists of 4 terms:"// &
                          " T1+T2+T3+T4. The terms T1=A/rij^alpha and T2=-C/rij^6"// &
                          " have to be given via the GENPOT section. The terms T3+T4"// &
                          " are obtained from the SIEPMANN section. The Siepmann-Sprik"// &
                          " potential is designed for water-metal chemisorption.", &
                          citations=(/Siepmann1995/), n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Defines the atomic kind involved in the nonbond potential", &
                          usage="ATOMS {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="B", &
                          description="Defines the B parameter of Siepmann potential", &
                          usage="B {real}", type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=0.6_dp, &
                                                        unit_str="angstrom"), &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="D", &
                          description="Defines the D parameter of Siepmann potential", &
                          usage="D {real}", type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=3.688388_dp, &
                                                        unit_str="internal_cp2k"), &
                          n_var=1, unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="E", &
                          description="Defines the E parameter of Siepmann potential", &
                          usage="E {real}", type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=9.069025_dp, &
                                                        unit_str="internal_cp2k"), &
                          n_var=1, unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="F", &
                          description="Defines the F parameter of Siepmann potential", &
                          usage="B {real}", type_of_var=real_t, &
                          default_r_val=13.3_dp, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
!
      CALL keyword_create(keyword, name="beta", &
                          description="Defines the beta parameter of Siepmann potential", &
                          usage="beta {real}", type_of_var=real_t, &
                          default_r_val=10.0_dp, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
!
      CALL keyword_create(keyword, name="RCUT", &
                          description="Defines the cutoff parameter of Siepmann potential", &
                          usage="RCUT {real}", type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=3.2_dp, &
                                                        unit_str="angstrom"), &
                          n_var=1, unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
!
      CALL keyword_create(keyword, name="ALLOW_OH_FORMATION", &
                          description=" The Siepmann-Sprik potential is actually designed for intact"// &
                          " water molecules only. If water is treated at the QM level,"// &
                          " water molecules can potentially dissociate, i.e."// &
                          " some O-H bonds might be stretched leading temporarily"// &
                          " to the formation of OH- ions. This keyword allows the"// &
                          " the formation of such ions. The T3 term (dipole term)"// &
                          " is then switched off for evaluating the interaction"// &
                          " between the OH- ion and the metal.", &
                          usage="ALLOW_OH_FORMATION TRUE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALLOW_H3O_FORMATION", &
                          description=" The Siepmann-Sprik potential is designed for intact water"// &
                          " molecules only. If water is treated at the QM level"// &
                          " and an acid is present, hydronium ions might occur."// &
                          " This keyword allows the formation of hydronium ions."// &
                          " The T3 term (dipole term) is switched off for evaluating"// &
                          " the interaction between hydronium and the metal.", &
                          usage="ALLOW_H3O_FORMATION TRUE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_Siepmann_section

! **************************************************************************************************
!> \brief creates the input section for the qs part
!> \param print_key ...
!> \param label ...
!> \param print_level ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_dipoles_section(print_key, label, print_level)
      TYPE(section_type), POINTER                        :: print_key
      CHARACTER(LEN=*), INTENT(IN)                       :: label
      INTEGER, INTENT(IN)                                :: print_level

      CHARACTER(len=*), PARAMETER :: routineN = 'create_dipoles_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(print_key)))CALL cp__a("input_cp2k_mm.F",2478)
      CALL cp_print_key_section_create(print_key_section=print_key, name=TRIM(label), &
                                       description="Section controlling the calculation of "//TRIM(label)//"."// &
                                       " Note that the result in the periodic case might be defined modulo a certain period,"// &
                                       " determined by the lattice vectors. During MD, this can lead to jumps.", &
                                       print_level=print_level, filename="__STD_OUT__")

      NULLIFY (keyword)
      CALL keyword_create(keyword=keyword, &
                          name="PERIODIC", &
                          description="Use Berry phase formula (PERIODIC=T) or simple operator (PERIODIC=F). "// &
                          "The latter normally requires that the CELL is periodic NONE.", &
                          usage="PERIODIC {logical}", &
                          repeats=.FALSE., &
                          n_var=1, &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="REFERENCE", &
                          variants=s2a("REF"), &
                          description="Define the reference point for the calculation of the electrostatic moment.", &
                          usage="REFERENCE COM", &
                          enum_c_vals=s2a("COM", "COAC", "USER_DEFINED", "ZERO"), &
                          enum_desc=s2a("Use Center of Mass", &
                                        "Use Center of Atomic Charges", &
                                        "Use User Defined Point (Keyword:REF_POINT)", &
                                        "Use Origin of Coordinate System"), &
                          enum_i_vals=(/use_mom_ref_com, &
                                        use_mom_ref_coac, &
                                        use_mom_ref_user, &
                                        use_mom_ref_zero/), &
                          default_i_val=use_mom_ref_zero)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="REFERENCE_POINT", &
                          variants=s2a("REF_POINT"), &
                          description="Fixed reference point for the calculations of the electrostatic moment.", &
                          usage="REFERENCE_POINT x y z", &
                          repeats=.FALSE., &
                          n_var=3, default_r_vals=(/0._dp, 0._dp, 0._dp/), &
                          type_of_var=real_t, &
                          unit_str='bohr')
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_dipoles_section

END MODULE input_cp2k_mm
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      - taken out of input_cp2k_motion
!> \author teo & fawzi
! **************************************************************************************************
MODULE input_cp2k_colvar
   USE bibliography,                    ONLY: Branduardi2007
   USE colvar_types,                    ONLY: &
        do_clv_fix_point, do_clv_geo_center, do_clv_x, do_clv_xy, do_clv_xyz, do_clv_xz, do_clv_y, &
        do_clv_yz, do_clv_z, plane_def_atoms, plane_def_vec
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              high_print_level,&
                                              low_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: gaussian,&
                                              numerical,&
                                              rmsd_all,&
                                              rmsd_list,&
                                              rmsd_weightlist
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              lchar_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_colvar'

   PUBLIC :: create_colvar_section, &
             create_colvar_xyz_d_section, &
             create_colvar_xyz_od_section

CONTAINS

! **************************************************************************************************
!> \brief creates the colvar section
!> \param section the section to be created
!> \param skip_recursive_colvar ...
!> \author teo
! **************************************************************************************************
   RECURSIVE SUBROUTINE create_colvar_section(section, skip_recursive_colvar)
      TYPE(section_type), POINTER                        :: section
      LOGICAL, OPTIONAL                                  :: skip_recursive_colvar

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_section', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: skip
      TYPE(section_type), POINTER                        :: print_key, subsection

      skip = .FALSE.
      IF (PRESENT(skip_recursive_colvar)) skip = skip_recursive_colvar
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",72)
      CALL section_create(section, name="COLVAR", &
                          description="This section specifies the nature of the collective variables.", &
                          n_keywords=1, n_subsections=1, repeats=.TRUE.)
      NULLIFY (subsection, print_key)

      CALL create_colvar_var_section(subsection=subsection, &
                                     section=section, skip_recursive_colvar=skip)

      CALL section_create(subsection, name="PRINT", &
                          description="Controls the printing of the colvar specifications", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)
      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing of basic information during colvar setup.", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_clv_info_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_section

! **************************************************************************************************
!> \brief Create the restart section for colvar restraints
!>      This section will be only used for restraint restarts.
!>      Constraints are handled automatically
!> \param section the section to create
!> \author Teodoro Laino 08.2006
! **************************************************************************************************
   SUBROUTINE create_clv_info_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_clv_info_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",114)
      NULLIFY (keyword)
      CALL section_create(section, name="COLVAR_FUNC_INFO", &
                          description="Specify further data possibly used by colvars, depending "// &
                          "on the starting geometry, for computing the functions value.", &
                          n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Colvar function data."// &
                          " The order is an internal order. So if you decide to edit/modify/add these values by hand"// &
                          " you should know very well what you are doing.!", repeats=.TRUE., &
                          usage="{Real} ...", type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_clv_info_section

! **************************************************************************************************
!> \brief creates the collective variables for the colvar section
!> \param subsection ...
!> \param section the section to be created
!> \param skip_recursive_colvar ...
!> \author teo
! **************************************************************************************************
   RECURSIVE SUBROUTINE create_colvar_var_section(subsection, section, skip_recursive_colvar)
      TYPE(section_type), POINTER                        :: subsection, section
      LOGICAL, INTENT(IN)                                :: skip_recursive_colvar

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_var_section', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(subsection)))CALL cp__a("input_cp2k_colvar.F",145)
      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",146)

      CALL create_colvar_dist_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_angle_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_torsion_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_coord_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_pop_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_gyr_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_d_pl_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_a_pl_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_rot_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_dfunct_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_qparm_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_hydronium_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_acid_hyd_dist_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_acid_hyd_shell_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_rmsd_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_xyz_d_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_xyz_od_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_u_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_wc_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_hbp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_ring_puckering_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_cond_dist_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      IF (.NOT. skip_recursive_colvar) THEN
         CALL create_colvar_rpath_section(subsection)
         CALL section_add_subsection(section, subsection)
         CALL section_release(subsection)

         CALL create_colvar_dpath_section(subsection)
         CALL section_add_subsection(section, subsection)
         CALL section_release(subsection)

         CALL create_colvar_comb_section(subsection)
         CALL section_add_subsection(section, subsection)
         CALL section_release(subsection)
      ENDIF

   END SUBROUTINE create_colvar_var_section

! **************************************************************************************************
!> \brief collective variables specifying coordination
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_colvar_coord_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_coord_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",266)
      CALL section_create(section, name="coordination", &
                          description="Section to define the coordination number as a collective variable.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (subsection, keyword)

      CALL keyword_create(keyword, name="ATOMS_FROM", &
                          variants=(/"POINTS_FROM"/), &
                          description="Specify indexes of atoms/points building the coordination variable. ", &
                          usage="ATOMS_FROM {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS_TO", &
                          variants=(/"POINTS_TO"/), &
                          description="Specify indexes of atoms/points building the coordination variable. ", &
                          usage="ATOMS_TO {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS_TO_B", &
                          variants=(/"POINTS_TO_B"/), &
                          description="For the CV given by the multiplication of two coorination numbers,"// &
                          " here specify indexes of the third set of atoms/points. ", &
                          usage="ATOMS_TO_B {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KINDS_FROM", &
                          description="Specify alternatively kinds of atoms building the coordination variable.", &
                          usage="KINDS_FROM {CHAR} {CHAR} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KINDS_TO", &
                          description="Specify alternatively kinds of atoms building the coordination variable.", &
                          usage="KINDS_TO {CHAR} {CHAR} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KINDS_TO_B", &
                          description="For the CV given by the multiplication of two coorination numbers,"// &
                          " here specify alternatively kinds of atoms building the coordination variable.", &
                          usage="KINDS_TO_B {CHAR} {CHAR} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, name="R0", &
                          variants=(/"R_0"/), &
                          description="Specify the R0 parameter in the coordination function.", &
                          usage="R0 {real}", default_r_val=3.0_dp, &
                          unit_str="bohr", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NN", &
                          variants=(/"EXPON_NUMERATOR"/), &
                          description="Sets the value of the numerator of the exponential factor"// &
                          "in the coordination FUNCTION.", &
                          usage="NN {integer}", default_i_val=6, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ND", &
                          variants=(/"EXPON_DENOMINATOR"/), &
                          description="Sets the value of the denominator of the exponential factor"// &
                          "in the coordination FUNCTION.", &
                          usage="ND {integer}", default_i_val=12, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="R0_B", &
                          variants=(/"R_0_B"/), &
                          description="For the CV given by the multiplication of two coorination numbers,"// &
                          " specify the R0 parameter in the second coordination function.", &
                          usage="R0_B {real}", default_r_val=3.0_dp, &
                          unit_str="bohr", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NN_B", &
                          variants=(/"EXPON_NUMERATOR_B"/), &
                          description="For the CV given by the multiplication of two coorination numbers,"// &
                          "Sets the value of the numerator of the exponential factor"// &
                          "in the coordination FUNCTION.", &
                          usage="NN_B {integer}", default_i_val=6, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ND_B", &
                          variants=(/"EXPON_DENOMINATOR_B"/), &
                          description="For the CV given by the multiplication of two coorination numbers,"// &
                          "Sets the value of the denominator of the exponential factor"// &
                          "in the coordination FUNCTION.", &
                          usage="ND_B {integer}", default_i_val=12, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_colvar_coord_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_colvar_cond_dist_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_cond_dist_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",395)
      CALL section_create(section, name="CONDITIONED_DISTANCE", &
                          description="Section to define the conditioned distance as a collective variable.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (subsection, keyword)

      CALL keyword_create(keyword, name="ATOMS_DISTANCE", &
                          description="Specify indexes of atoms/points from which the distance is computed. ", &
                          usage="ATOMS_DISTANCE {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS_FROM", &
                          variants=(/"POINTS_FROM"/), &
                          description="Specify indexes of atoms/points building the coordination variable. ", &
                          usage="ATOMS_FROM {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS_TO", &
                          variants=(/"POINTS_TO"/), &
                          description="Specify indexes of atoms/points building the coordination variable. ", &
                          usage="ATOMS_TO {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KINDS_FROM", &
                          description="Specify alternatively kinds of atoms building the coordination variable.", &
                          usage="KINDS_FROM {CHAR} {CHAR} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KINDS_TO", &
                          description="Specify alternatively kinds of atoms building the coordination variable.", &
                          usage="KINDS_TO {CHAR} {CHAR} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, name="R0", &
                          variants=(/"R_0"/), &
                          description="Specify the R0 parameter in the coordination function.", &
                          usage="R0 {real}", default_r_val=3.0_dp, &
                          unit_str="bohr", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NN", &
                          variants=(/"EXPON_NUMERATOR"/), &
                          description="Sets the value of the numerator of the exponential factor"// &
                          "in the coordination FUNCTION.", &
                          usage="NN {integer}", default_i_val=6, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ND", &
                          variants=(/"EXPON_DENOMINATOR"/), &
                          description="Sets the value of the denominator of the exponential factor"// &
                          "in the coordination FUNCTION.", &
                          usage="ND {integer}", default_i_val=12, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LAMBDA", &
                          description="Specify the lambda parameter at the exponent of the conditioned distance function.", &
                          usage="R0 {real}", default_r_val=3.0_dp, &
                          unit_str="bohr", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_colvar_cond_dist_section

! **************************************************************************************************
!> \brief collective variables specifying population of a specie based on coordination
!> \param section the section to be created
!> \date  01.2009
!> \author Fabio Sterpone
! **************************************************************************************************
   SUBROUTINE create_colvar_pop_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_pop_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",494)
      CALL section_create(section, name="population", &
                          description="Section to define the population of specie as a collective variable.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (subsection, keyword)

      CALL keyword_create(keyword, name="ATOMS_FROM", &
                          variants=(/"POINTS_FROM"/), &
                          description="Specify indexes of atoms/points building the coordination variable. ", &
                          usage="ATOMS_FROM {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS_TO", &
                          variants=(/"POINTS_TO"/), &
                          description="Specify indexes of atoms/points building the coordination variable. ", &
                          usage="ATOMS_TO {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KINDS_FROM", &
                          description="Specify alternatively kinds of atoms building the coordination variable.", &
                          usage="KINDS_FROM {CHAR} {CHAR} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KINDS_TO", &
                          description="Specify alternatively kinds of atoms building the coordination variable.", &
                          usage="KINDS_TO {CHAR} {CHAR} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, name="R0", &
                          variants=(/"R_0"/), &
                          description="Specify the R0 parameter in the coordination function.", &
                          usage="R0 {real}", default_r_val=3.0_dp, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NN", &
                          variants=(/"EXPON_NUMERATOR"/), &
                          description="Sets the value of the numerator of the exponential factor"// &
                          "in the coordination FUNCTION.", &
                          usage="NN {integer}", default_i_val=6, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ND", &
                          variants=(/"EXPON_DENOMINATOR"/), &
                          description="Sets the value of the denominator of the exponential factor"// &
                          "in the coordination FUNCTION.", &
                          usage="ND {integer}", default_i_val=12, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="n0", &
                          variants=(/"n_0"/), &
                          description="Specify the n0 parameter that sets the coordination of the species.", &
                          usage="n0 {integer}", default_i_val=4, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SIGMA", &
                          description="Specify the gaussian width of used to build the population istogram.", &
                          usage="SIGMA {real}", default_r_val=0.5_dp, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_colvar_pop_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_colvar_gyr_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_gyr_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",592)
      CALL section_create(section, name="GYRATION_RADIUS", &
                          description="Section to define the gyration radius as a collective variable.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (subsection, keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          variants=(/"POINTS"/), &
                          description="Specify indexes of atoms/points defyining the gyration radius variable. ", &
                          usage="ATOMS {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KINDS", &
                          description="Specify alternatively kinds of atoms defining the gyration radius.", &
                          usage="KINDS {CHAR} {CHAR} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_gyr_section

! **************************************************************************************************
!> \brief collective variables specifying torsion
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_colvar_dfunct_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_dfunct_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",635)
      CALL section_create(section, name="DISTANCE_FUNCTION", &
                          description="Section to define functions between two distances as collective variables."// &
                          " The function is defined as d1+coeff*d2", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="ATOMS", &
                          variants=(/"POINTS"/), &
                          description="Specifies the indexes of atoms/points for the two bonds d1=(1-2) d2=(3-4).", &
                          usage="ATOMS {integer} {integer} {integer} {integer}", &
                          n_var=4, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COEFFICIENT", &
                          description="Specifies the coefficient in the function for the constraint."// &
                          " -1.0 has to be used for distance difference, 1.0 for distance addition", &
                          usage="COEFFICIENT {real}", &
                          type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PBC", &
                          description="Whether periodic boundary conditions should be applied on the "// &
                          "atomic position before computing the colvar or not.", &
                          usage="PBC", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_dfunct_section

! **************************************************************************************************
!> \brief collective variables specifying torsion
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_colvar_torsion_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_torsion_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",688)
      CALL section_create(section, name="torsion", &
                          description="Section to define the torsion as a collective variables.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="ATOMS", &
                          variants=(/"POINTS"/), &
                          description="Specifies the indexes of atoms/points defining the torsion.", &
                          usage="ATOMS {integer} {integer} {integer} {integer}", &
                          n_var=4, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_torsion_section

! **************************************************************************************************
!> \brief collective variables specifying torsion
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_colvar_rot_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_rot_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",724)
      CALL section_create(section, name="bond_rotation", &
                          description="Section to define the rotation of a bond/line with respect to"// &
                          "another bond/line", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="P1_BOND1", &
                          description="Specifies the index of atom/point defining the first point"// &
                          "of the first bond/line.", &
                          usage="P1_BOND1 {integer}", &
                          n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="P2_BOND1", &
                          description="Specifies the index of atom/point defining the second point"// &
                          "of the first bond/line.", &
                          usage="P2_BOND1 {integer}", &
                          n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="P1_BOND2", &
                          description="Specifies the index of atom/point defining the first point"// &
                          "of the second bond/line.", &
                          usage="P1_BOND2 {integer}", &
                          n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="P2_BOND2", &
                          description="Specifies the index of atom/point defining the second point"// &
                          "of the second bond/line.", &
                          usage="P2_BOND2 {integer}", &
                          n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_rot_section

! **************************************************************************************************
!> \brief collective variables specifying angles
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_colvar_angle_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_angle_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",785)
      CALL section_create(section, name="angle", &
                          description="Section to define the angle as a collective variables.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="ATOMS", &
                          variants=(/"POINTS"/), &
                          description="Specifies the indexes of atoms/points defining the angle.", &
                          usage="ATOMS {integer} {integer} {integer}", &
                          n_var=3, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_angle_section

! **************************************************************************************************
!> \brief creates the colvar section regarded to the collective variables dist
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_colvar_dist_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_dist_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",820)
      CALL section_create(section, name="distance", &
                          description="Section to define the distance as a collective variables.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="ATOMS", &
                          variants=(/"POINTS"/), &
                          description="Specifies the indexes of atoms/points defining the distance.", &
                          usage="ATOMS {integer} {integer}", &
                          n_var=2, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="AXIS", &
                          description="Define the axes along which the colvar should be evaluated", &
                          usage="AXIS (XYZ | X | Y | Z | XY| XZ | YZ)", &
                          enum_c_vals=s2a("XYZ", "X", "Y", "Z", "XY", "XZ", "YZ"), &
                          enum_i_vals=(/do_clv_xyz, do_clv_x, do_clv_y, do_clv_z, do_clv_xy, do_clv_xz, do_clv_yz/), &
                          default_i_val=do_clv_xyz)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_dist_section

! **************************************************************************************************
!> \brief creates the colvar section regarded to the collective variables dist
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_colvar_xyz_d_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_xyz_d_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",863)
      CALL section_create(section, name="XYZ_DIAG", &
                          description="Section to define the distance of an atom from its starting "// &
                          "position ((X-X(0))^2+(Y-Y(0))^2+(Z-Z(0))^2) or part of its components as a collective variable."// &
                          "If absolute_position is specified, instead the CV is represented by the "// &
                          "instantaneous position of the atom (only available for X, Y or Z components).", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="ATOM", &
                          variants=(/"POINT"/), &
                          description="Specifies the index of the atom/point.", &
                          usage="ATOM {integer}", &
                          n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COMPONENT", &
                          description="Define the component of the position vector which will be used "// &
                          "as a colvar.", &
                          usage="AXIS (XYZ | X | Y | Z | XY| XZ | YZ)", &
                          enum_c_vals=s2a("XYZ", "X", "Y", "Z", "XY", "XZ", "YZ"), &
                          enum_i_vals=(/do_clv_xyz, do_clv_x, do_clv_y, do_clv_z, do_clv_xy, do_clv_xz, do_clv_yz/), &
                          default_i_val=do_clv_xyz)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PBC", &
                          description="Whether periodic boundary conditions should be applied on the "// &
                          "atomic position before computing the colvar or not.", &
                          usage="PBC", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ABSOLUTE_POSITION", &
                          description="If enabled, the absolute position of the atoms will be used. ", &
                          usage="ABSOLUTE_POSITION", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_xyz_d_section

! **************************************************************************************************
!> \brief creates the colvar section regarded to the collective variables dist
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_colvar_xyz_od_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_xyz_od_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",926)
      CALL section_create(section, name="XYZ_OUTERDIAG", &
                          description="Section to define the cross term (XA-XA(0))*(XB-XB(0))+(XA-XA(0))*(YB-YB(0))"// &
                          " or part of its components as a collective variable. The final term is given by the product "// &
                          " of the components of A with the components of B.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="ATOMS", &
                          variants=(/"POINTS"/), &
                          description="Specifies the index of the atoms/points A and B.", &
                          usage="ATOMS {integer} {integer}", &
                          n_var=2, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COMPONENT_A", &
                          description="Define the component of the position vector which will be used "// &
                          "as a colvar for atom A.", &
                          usage="AXIS (XYZ | X | Y | Z | XY| XZ | YZ)", &
                          enum_c_vals=s2a("XYZ", "X", "Y", "Z", "XY", "XZ", "YZ"), &
                          enum_i_vals=(/do_clv_xyz, do_clv_x, do_clv_y, do_clv_z, do_clv_xy, do_clv_xz, do_clv_yz/), &
                          default_i_val=do_clv_xyz)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COMPONENT_B", &
                          description="Define the component of the position vector which will be used "// &
                          "as a colvar for atom B.", &
                          usage="AXIS (XYZ | X | Y | Z | XY| XZ | YZ)", &
                          enum_c_vals=s2a("XYZ", "X", "Y", "Z", "XY", "XZ", "YZ"), &
                          enum_i_vals=(/do_clv_xyz, do_clv_x, do_clv_y, do_clv_z, do_clv_xy, do_clv_xz, do_clv_yz/), &
                          default_i_val=do_clv_xyz)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PBC", &
                          description="Whether periodic boundary conditions should be applied on the "// &
                          "atomic position before computing the colvar or not.", &
                          usage="PBC", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_xyz_od_section

! **************************************************************************************************
!> \brief energy as collective variable
!> \param section the section to be created
!> \author Sebastiano Caravati
! **************************************************************************************************
   SUBROUTINE create_colvar_u_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_u_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",991)
      CALL section_create(section, name="u", &
                          description="Section to define the energy as a generalized collective variable.", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)

      NULLIFY (subsection, keyword)
      CALL section_create(subsection, name="MIXED", &
                          description="This section allows to use any function of the energy subsystems "// &
                          " in a mixed_env calculation as a collective variable.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="ENERGY_FUNCTION", &
                          description="Specifies the functional form of the collective variable in mathematical notation.", &
                          usage="ENERGY_FUNCTION (E1+E2-LOG(E1/E2))", type_of_var=lchar_t, &
                          n_var=1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VARIABLES", &
                          description="Defines the variables of the functional form. To allow an efficient"// &
                          " mapping the order of the energy variables will be considered identical to the"// &
                          " order of the force_eval in the force_eval_order list.", &
                          usage="VARIABLES x", type_of_var=char_t, &
                          n_var=-1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARAMETERS", &
                          description="Defines the parameters of the functional form", &
                          usage="PARAMETERS a b D", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VALUES", &
                          description="Defines the values of parameter of the functional form", &
                          usage="VALUES ", type_of_var=real_t, &
                          n_var=-1, repeats=.TRUE., unit_str="internal_cp2k")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="UNITS", &
                          description="Optionally, allows to define valid CP2K unit strings for each parameter value. "// &
                          "It is assumed that the corresponding parameter value is specified in this unit.", &
                          usage="UNITS angstrom eV*angstrom^-1 angstrom^1 K", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DX", &
                          description="Parameter used for computing the derivative with the Ridders' method.", &
                          usage="DX <REAL>", default_r_val=0.1_dp, unit_str="bohr")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ERROR_LIMIT", &
                          description="Checks that the error in computing the derivative is not larger than "// &
                          "the value set; in case error is larger a warning message is printed.", &
                          usage="ERROR_LIMIT <REAL>", default_r_val=1.0E-12_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_u_section

! **************************************************************************************************
!> \brief creates the colvar section regarded to the collective variables distance
!>      of a point from a plane
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_colvar_d_pl_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_d_pl_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",1073)
      CALL section_create(section, name="distance_point_plane", &
                          description="Section to define the distance of a point from a plane "// &
                          "as a collective variables.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="PBC", &
                          description="Whether periodic boundary conditions should be applied on the "// &
                          "atomic position before computing the colvar or not.", &
                          usage="PBC", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS_PLANE", &
                          variants=(/"POINTS_PLANE"/), &
                          description="Specifies the indexes of atoms/points defining the plane.", &
                          usage="ATOMS_PLANE <INTEGER> <INTEGER> <INTEGER>", &
                          n_var=3, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOM_POINT", &
                          variants=(/"POINT_POINT"/), &
                          description="Specifies the atom/point index defining the point.", &
                          usage="ATOM_POINT <INTEGER>", &
                          n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_d_pl_section

! **************************************************************************************************
!> \brief creates the colvar section regarded to the collective variables
!>      angles betweem two planes
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_colvar_a_pl_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_a_pl_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",1126)
      CALL section_create(section, name="angle_plane_plane", &
                          description="This section defines the angle between two planes "// &
                          "as a collective variables.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL section_create(subsection, name="PLANE", &
                          description="This section defines the plane. When using this colvar, "// &
                          "two plane section must be defined!", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="DEF_TYPE", &
                          description="Specify how the plane is defined: either by 3 atoms or by a fixed normal "// &
                          "vector. At least one plane must be defined through atoms.", &
                          usage="DEF_TYPE ATOMS", &
                          default_i_val=plane_def_atoms, &
                          enum_c_vals=s2a("ATOMS", "VECTOR"), &
                          enum_desc=s2a("Plane defined by the position of 3 atoms", &
                                        "Plane defined by a fixed normal vector"), &
                          enum_i_vals=(/plane_def_atoms, plane_def_vec/))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specifies the indexes of 3 atoms/points defining the plane.", &
                          usage="ATOMS <INTEGER> <INTEGER> <INTEGER>", &
                          n_var=3, type_of_var=integer_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NORMAL_VECTOR", &
                          description="Alternatively to 3 atoms/points one can define one of the two, "// &
                          "planes by defining its NORMAL vector.", &
                          usage="NORMAL_VECTOR 0.0 1.0 0.0", &
                          n_var=3, type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_colvar_a_pl_section

! **************************************************************************************************
!> \brief create a geometrical point as a function of several atom coordinates
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_point_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_point_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",1186)
      CALL section_create(section, name="POINT", &
                          description="Enables the possibility to use geometrical centers instead of single atoms"// &
                          " to define colvars", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="TYPE", &
                          description="Chooses the type of geometrical point", &
                          usage="type (GEO_CENTER|FIX_POINT)", &
                          enum_c_vals=s2a("GEO_CENTER", "FIX_POINT"), &
                          enum_desc=s2a("Conmputes the geometrical center of the listed atoms", &
                                        "Defines a fixed point in space"), &
                          enum_i_vals=(/do_clv_geo_center, do_clv_fix_point/), &
                          default_i_val=do_clv_geo_center)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specifies the indexes of atoms defining the geometrical center", &
                          usage="ATOMS {integer} {integer} {integer} {integer}", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="WEIGHTS", &
         description="Specifies the weights for a weighted geometrical center. Default is 1/natoms for every atom", &
         usage="WEIGHTS {real} {real} {real} {real}", &
         n_var=-1, type_of_var=real_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="XYZ", &
                          description="Specifies the xyz of the fixed point (if the case)", &
                          usage="XYZ {real} {real} {real}", &
                          n_var=3, type_of_var=real_t, unit_str="bohr", &
                          repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_point_section

! **************************************************************************************************
!> \brief collective variables specifying torsion
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_colvar_qparm_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_qparm_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",1244)
      CALL section_create(section, name="qparm", &
                          description="Section to define the Q parameter (crystalline order parameter) as a collective variable.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="ATOMS_FROM", &
                          variants=(/"POINTS_FROM"/), &
                          description="Specify indexes of atoms/points building the coordination variable. ", &
                          usage="ATOMS_FROM {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS_TO", &
                          variants=(/"POINTS_TO"/), &
                          description="Specify indexes of atoms/points building the coordination variable. ", &
                          usage="ATOMS_TO {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Specifies the distance cutoff for neighbors.", &
                          usage="RCUT {real}", &
                          n_var=1, unit_str="angstrom", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="L", &
                          description="Specifies the L spherical harmonics from Ylm.", &
                          usage="L {integer}", &
                          n_var=1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALPHA", &
                          description="Specifies the width of the Fermi-Dirac style smearing around RCUT.", &
                          usage="ALPHA {real}", unit_str="angstrom^-1", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_qparm_section

! **************************************************************************************************
!> \brief collective variables specifying hydronium solvation
!> \param section the section to be created
!> \author Marcel Baer
! **************************************************************************************************
   SUBROUTINE create_colvar_hydronium_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_hydronium_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",1307)
      CALL section_create(section, name="hydronium", &
                          description="Section to define the formation of a hydronium as a"// &
                          " collective variable. Number of oxygens in the 1st shell of the"// &
                          " hydronium. Adapted from Equation (3) in Supplementary Info of"// &
                          " J. Am. Chem. Soc.,128, 2006, 11318, i.e. omitting the cutoff function"// &
                          " and summing only over the oxygens of water.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="OXYGENS", &
                          description="Specifies indexes of atoms building the coordination variable."// &
                          " Oxygens of the water molecules.", &
                          usage="OXYGENS {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="HYDROGENS", &
                          description="Specifies indexes of atoms building the coordination variable."// &
                          " Hydrogens of the water molecules.", &
                          usage="HYDROGENS {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ROO", &
                          description="Specifies the rc parameter in the coordination function:"// &
                          " number of oxygens per water oxygen.", &
                          usage="ROO {real}", default_r_val=cp_unit_to_cp2k(value=3.0_dp, &
                                                                            unit_str="bohr"), unit_str="bohr", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pOO", &
                          variants=(/"EXPON_NUMERATORA"/), &
                          description="Sets the value of the numerator of the exponential factor"// &
                          " in the coordination function: number of oxygens per water oxygen.", &
                          usage="pOO {integer}", default_i_val=6, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="qOO", &
                          variants=(/"EXPON_DENOMINATORA"/), &
                          description="Sets the value of the denominator of the exponential factor"// &
                          " in the coordination function: number of oxygens per water oxygen.", &
                          usage="qOO {integer}", default_i_val=12, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ROH", &
                          description="Specifies the rc parameter in the coordination function: "// &
                          " number of hydrogens per water molecule.", &
                          usage="ROH {real}", default_r_val=cp_unit_to_cp2k(value=3.0_dp, &
                                                                            unit_str="bohr"), unit_str="bohr", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pOH", &
                          variants=(/"EXPON_NUMERATORB"/), &
                          description="Sets the value of the numerator of the exponential factor"// &
                          " in the coordination function: number of hydrogens per water molecule.", &
                          usage="pOH {integer}", default_i_val=6, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="qOH", &
                          variants=(/"EXPON_DENOMINATORB"/), &
                          description="Sets the value of the denominator of the exponential factor"// &
                          " in the coordination function: number of hydrogens per water molecule.", &
                          usage="qOH {integer}", default_i_val=12, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NH", &
                          description="Specifies the NH parameter in the M function.", &
                          usage="NH {real}", default_r_val=3.0_dp, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pM", &
                          variants=(/"EXPON_NUMERATOR"/), &
                          description="Sets the value of the numerator of the exponential factor"// &
                          " in the M function.", &
                          usage="pM {integer}", default_i_val=8, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="qM", &
                          variants=(/"EXPON_DENOMINATOR"/), &
                          description="Sets the value of the denominator of the exponential factor"// &
                          " in the M function.", &
                          usage="qM {integer}", default_i_val=16, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LAMBDA", &
                          description="Specify the LAMBDA parameter in the hydronium function.", &
                          usage="LAMBDA {real}", default_r_val=10.0_dp, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_colvar_hydronium_section

! **************************************************************************************************
!> \brief collective variables specifying the solvation of carboxylic acid;
!>        distance between hydronium ion and acetate ion; Equation (2) in
!>        Supplementary Information of J. Am. Chem. Soc.,128,  2006, 11318
!> \param section the section to be created
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE create_colvar_acid_hyd_dist_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_acid_hyd_dist_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",1435)
      CALL section_create(section, name="ACID_HYDRONIUM_DISTANCE", &
                          description="Section to define the dissociation of a carboxylic acid in"// &
                          " water. Distance between hydronium ion and acetate ion. Equation (2)"// &
                          " in Supplementary Info of J. Am. Chem. Soc.,128,  2006, 11318.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="OXYGENS_WATER", &
                          description="Specifies indexes of atoms building the coordination variable."// &
                          " Oxygens of the water molecules. ", &
                          usage="OXYGENS {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OXYGENS_ACID", &
                          description="Specifies indexes of atoms building the coordination variable."// &
                          " Oxygens of the carboxyl groups.", &
                          usage="OXYGENS {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="HYDROGENS", &
                          description="Specifies indexes of atoms building the coordination variable."// &
                          " Hydrogens of the water molecules and of the carboxyl groups.", &
                          usage="HYDROGENS {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pWOH", &
                          description="Sets the value of the numerator of the exponential factor"// &
                          " in the coordination function: number of hydrogens per water molecule.", &
                          usage="pWOH {integer}", default_i_val=8, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="qWOH", &
                          description="Sets the value of the denominator of the exponential factor"// &
                          " in the coordination function: number of hydrogens per water molecule.", &
                          usage="qWOH {integer}", default_i_val=16, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RWOH", &
                          description="Specify the rc parameter in the coordination function:"// &
                          " number of hydrogens per water molecule.", &
                          usage="RWOH {real}", default_r_val=cp_unit_to_cp2k(value=2.4_dp, &
                                                                             unit_str="bohr"), unit_str="bohr", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pAOH", &
                          description="Sets the value of the numerator of the exponential factor"// &
                          " in the coordination function: number of hydrogens per carboxyl group.", &
                          usage="pAOH {integer}", default_i_val=6, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="qAOH", &
                          description="Sets the value of the denominator of the exponential factor"// &
                          " in the coordination function: number of hydrogens per carboxyl group.", &
                          usage="qAOH {integer}", default_i_val=14, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RAOH", &
                          description="Specify the rc parameter in the coordination function: "// &
                          " number of hydrogens per carboxyl group.", &
                          usage="RAOH {real}", default_r_val=cp_unit_to_cp2k(value=2.4_dp, &
                                                                             unit_str="bohr"), unit_str="bohr", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pCUT", &
                          description="Sets the value of the numerator of the exponential factor"// &
                          " in the cutoff function.", &
                          usage="pCUT {integer}", default_i_val=6, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="qCUT", &
                          description="Sets the value of the denominator of the exponential factor"// &
                          " in the cutoff function.", &
                          usage="qCUT {integer}", default_i_val=12, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NC", &
                          description="Specifies the NC parameter in the cutoff function.", &
                          usage="NC {real}", default_r_val=0.56_dp, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LAMBDA", &
                          variants=(/"LAMBDA"/), &
                          description="Specifies the LAMBDA parameter carboxylic acid function.", &
                          usage="LAMBDA {real}", default_r_val=20.0_dp, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_colvar_acid_hyd_dist_section

! **************************************************************************************************
!> \brief collective variables specifying the solvation of carboxylic acid;
!>        number of oxygens in the 1st shell of the hydronium; Equation (3) in
!>        Supplementary Information of J. Am. Chem. Soc.,128,  2006, 11318
!> \param section the section to be created
!> \author Dorothea Golze
! **************************************************************************************************
   SUBROUTINE create_colvar_acid_hyd_shell_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_acid_hyd_shell_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",1564)
      CALL section_create(section, name="ACID_HYDRONIUM_SHELL", &
                          description="Section to define the dissociation of a carboxylic acid in"// &
                          " water. Number of oxygens in the 1st shell of the hydronium. Equation (3)"// &
                          " in Supplementary Info of J. Am. Chem. Soc.,128,  2006, 11318. Similar to"// &
                          " the HYDRONIUM colvar, but with modification for the acid.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="OXYGENS_WATER", &
                          description="Specifies indexes of atoms building the coordination variable."// &
                          " Oxygens of the water molecules. ", &
                          usage="OXYGENS {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OXYGENS_ACID", &
                          description="Specifies indexes of atoms building the coordination variable."// &
                          " Oxygens of the carboxyl groups.", &
                          usage="OXYGENS {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="HYDROGENS", &
                          description="Specifies indexes of atoms building the coordination variable. "// &
                          " Hydrogens of the water molecules and of the carboxyl groups.", &
                          usage="HYDROGENS {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pWOH", &
                          description="Sets the value of the numerator of the exponential factor"// &
                          " in the coordination function: number of hydrogens per water molecule.", &
                          usage="pWOH {integer}", default_i_val=8, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="qWOH", &
                          description="Sets the value of the denominator of the exponential factor"// &
                          " in the coordination function: number of hydrogens per water molecule.", &
                          usage="qWOH {integer}", default_i_val=16, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RWOH", &
                          description="Specifies the rc parameter in the coordination function: "// &
                          " number of hydrogens per water molecule.", &
                          usage="RWOH {real}", default_r_val=cp_unit_to_cp2k(value=2.4_dp, &
                                                                             unit_str="bohr"), unit_str="bohr", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pAOH", &
                          description="Sets the value of the numerator of the exponential factor"// &
                          " in the coordination function: number of hydrogens per carboxyl group.", &
                          usage="pAOH {integer}", default_i_val=6, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="qAOH", &
                          description="Sets the value of the denominator of the exponential factor"// &
                          " in the coordination function: number of hydrogens per carboxyl group.", &
                          usage="qAOH {integer}", default_i_val=14, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RAOH", &
                          description="Specifies the rc parameter in the coordination function:"// &
                          " number of hydrogens per carboxyl group.", &
                          usage="RAOH {real}", default_r_val=cp_unit_to_cp2k(value=2.4_dp, &
                                                                             unit_str="bohr"), unit_str="bohr", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pOO", &
                          description="Sets the value of the numerator of the exponential factor"// &
                          " in the coordination function: number of oxygens per water oxygen.", &
                          usage="pOO {integer}", default_i_val=6, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="qOO", &
                          description="Sets the value of the denominator of the exponential factor"// &
                          " in the coordination function: number of oxygens per water oxygen.", &
                          usage="qOO {integer}", default_i_val=12, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ROO", &
                          description="Specifies the rc parameter in the coordination function:"// &
                          " number of oxygens per water oxygen.", &
                          usage="ROO {real}", default_r_val=cp_unit_to_cp2k(value=5.5_dp, &
                                                                            unit_str="bohr"), unit_str="bohr", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pM", &
                          description="Sets the value of the numerator of the exponential factor"// &
                          " in the M function.", &
                          usage="pM {integer}", default_i_val=8, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="qM", &
                          description="Sets the value of the denominator of the exponential factor"// &
                          " in the M function.", &
                          usage="qM {integer}", default_i_val=16, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NH", &
                          description="Specifies the NH parameter in the M function.", &
                          usage="NH {real}", default_r_val=2.2_dp, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pCUT", &
                          description="Sets the value of the numerator of the exponential factor"// &
                          "in the cutoff function.", &
                          usage="pCUT {integer}", default_i_val=6, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="qCUT", &
                          description="Sets the value of the denominator of the exponential factor"// &
                          "in the cutoff function.", &
                          usage="qCUT {integer}", default_i_val=12, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NC", &
                          description="Specifies the NC parameter in the cutoff function.", &
                          usage="NC {real}", default_r_val=0.9_dp, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LAMBDA", &
                          variants=(/"LAMBDA"/), &
                          description="Specifies the LAMBDA parameter carboxylic acid function.", &
                          usage="LAMBDA {real}", default_r_val=10.0_dp, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_colvar_acid_hyd_shell_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_colvar_rmsd_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_rmsd_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection, subsubsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",1739)
      CALL section_create(section, name="rmsd", &
                          description="Section to define a CV as function of RMSD computed with respect to"// &
                          " given reference configurations. For 2 configurations the colvar is equal to:"// &
                          " ss = (RMSDA-RMSDB)/(RMSDA+RMSDB), while if only 1 configuration is given, then the"// &
                          " colvar is just the RMSD from that frame.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection, subsubsection)
      CALL keyword_create(keyword, name="SUBSET_TYPE", &
                          description="Define the subsytem used to compute the RMSD", &
                          usage="SUBSET_TYPE ALL", &
                          enum_c_vals=s2a("ALL", "LIST", "WEIGHT_LIST"), &
                          enum_i_vals=(/rmsd_all, rmsd_list, rmsd_weightlist/), &
                          default_i_val=rmsd_all)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALIGN_FRAMES", &
                          description="Whether the reference frames should be aligned to minimize the RMSD", &
                          usage="ALIGN_FRAME", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specify indexes of atoms building the subset. ", &
                          usage="ATOMS {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHTS", &
                          description="Specify weights of atoms building the subset. ", &
                          usage="weightS {real} {real} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="FRAME", &
                          description="Specify coordinates of the frame (number of frames can be either 1 or 2)", &
                          repeats=.TRUE.)

      CALL keyword_create(keyword, name="COORD_FILE_NAME", &
                          description="Name of the xyz file with coordinates (alternative to &COORD section)", &
                          usage="COORD_FILE_NAME <CHAR>", &
                          default_lc_val="")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL create_coord_section_cv(subsubsection, "RMSD")
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_rmsd_section

! **************************************************************************************************
!> \brief collective variables specifying the space orthogonal to the reaction path
!>        in the space spanned by the involved collective coordinates
!> \param section the section to be created
!> \author fschiff
! **************************************************************************************************
   SUBROUTINE create_colvar_rpath_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_rpath_section', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",1810)
      CALL section_create(section, name="REACTION_PATH", &
                          description="Section defining a one dimensional reaction path in an Q-dimensional space of colvars. "// &
                          "Constraining this colvar, allows to sample the space orthogonal to the reaction path, "// &
                          "both in the Q-dimensional colvar and 3N-Q remaining coordinates. "// &
                          "For the details of the function see cited literature.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/Branduardi2007/))

      CALL keywords_colvar_path(section)
   END SUBROUTINE create_colvar_rpath_section

! **************************************************************************************************
!> \brief Distance from reaction path
!> \param section the section to be created
!> \author 01.2010
! **************************************************************************************************
   SUBROUTINE create_colvar_dpath_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_dpath_section', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",1833)
      CALL section_create(section, name="DISTANCE_FROM_PATH", &
                          description="Section defining the distance from a one dimensional reaction "// &
                          "path in an Q-dimensional space of colvars. "// &
                          "Constraining this colvar, allows to sample the space equidistant to the reaction path, "// &
                          "both in the Q-dimensional colvar and 3N-Q remaining coordinates. "// &
                          "For the details of the function see cited literature.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/Branduardi2007/))

      CALL keywords_colvar_path(section)
   END SUBROUTINE create_colvar_dpath_section

! **************************************************************************************************
!> \brief Section describinf keywords for both reaction path and distance from reaction path
!> \param section the section to be created
!> \author 01.2010
! **************************************************************************************************
   SUBROUTINE keywords_colvar_path(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'keywords_colvar_path', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection, subsubsection

      NULLIFY (keyword, subsection, subsubsection, print_key)
      CALL create_colvar_section(subsection, skip_recursive_colvar=.TRUE.)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, name="DISTANCES_RMSD", &
                          description=" ", &
                          usage="DISTANCES_RMSD T", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RMSD", &
                          description=" ", &
                          usage="RMSD T", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SUBSET_TYPE", &
                          description="Define the subsytem used to compute the RMSD", &
                          usage="SUBSET_TYPE ALL", &
                          enum_c_vals=s2a("ALL", "LIST"), &
                          enum_i_vals=(/rmsd_all, rmsd_list/), &
                          default_i_val=rmsd_all)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALIGN_FRAMES", &
                          description="Whether the reference frames should be aligned to minimize the RMSD", &
                          usage="ALIGN_FRAME", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specify indexes of atoms building the subset. ", &
                          usage="ATOMS {integer} {integer} ..", repeats=.TRUE., &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="FRAME", &
                          description="Specify coordinates of the frame", &
                          repeats=.TRUE.)

      CALL keyword_create(keyword, name="COORD_FILE_NAME", &
                          description="Name of the xyz file with coordinates (alternative to &COORD section)", &
                          usage="COORD_FILE_NAME <CHAR>", &
                          default_lc_val="")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL create_coord_section_cv(subsubsection, "RMSD")
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, name="FUNCTION", &
                          description="Specifies the ith element of the vector valued function that defines the reaction path. "// &
                          "This keyword needs to repeat exactly Q times, and the order must match the order of the colvars. "// &
                          "The VARIABLE (e.g. T) which parametrises the curve can be used as the target of a constraint.", &
                          usage="FUNCTION (sin(T+2)+2*T)", type_of_var=lchar_t, &
                          n_var=1, default_lc_val="0", repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VARIABLE", &
                          description="Specifies the name of the variable that parametrises the FUNCTION "// &
                          "defining the reaction path.", &
                          usage="VARIABLE T", type_of_var=char_t, &
                          n_var=1, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="LAMBDA", &
         description="Specifies the exponent of the Gaussian used in the integral representation of the colvar."// &
         "The shape of the space orthogonal to the reaction path is defined by this choice. "// &
         "In the limit of large values, it is given by the plane orthogonal to the path."// &
         "In practice, modest values are required for stable numerical integration.", &
         usage="LAMBDA {real}", &
         type_of_var=real_t, default_r_val=5.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STEP_SIZE", &
                          description="Step size in the numerical integration, "// &
                          "a few thousand points are common, and the proper number also depends on LAMBDA.", &
                          usage="STEP_SIZE {real}", &
                          type_of_var=real_t, default_r_val=0.01_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RANGE", &
                          description="The range of VARIABLE used for the parametrisation.", &
                          usage="RANGE <REAL> <REAL>", &
                          n_var=2, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create( &
         print_key, name="MAP", &
         description="Activating this print key will print once a file with the values of the FUNCTION on a grid "// &
         "of COLVAR values in a specified range. "// &
         "GRID_SPACING and RANGE for every COLVAR has to be specified again in the same order as they are in the input.", &
         print_level=high_print_level, filename="PATH")

      CALL keyword_create(keyword, name="RANGE", &
                          description="The range of of the grid of the COLVAR.", &
                          usage="RANGE <REAL> <REAL>", &
                          n_var=2, type_of_var=real_t, repeats=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GRID_SPACING", &
                          description="Distance between two gridpoints for the grid on the COLVAR", &
                          usage="STEP_SIZE {real}", repeats=.TRUE., &
                          type_of_var=real_t, default_r_val=0.01_dp)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE keywords_colvar_path

! **************************************************************************************************
!> \brief Colvar allowing a combination of COLVARS
!> \param section the section to be created
!> \author Teodoro Laino [tlaino] - 12.2008
! **************************************************************************************************
   SUBROUTINE create_colvar_comb_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_comb_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",2004)
      CALL section_create(section, name="COMBINE_COLVAR", &
                          description="Allows the possibility to combine several COLVARs into one COLVAR "// &
                          "with a generic function.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)
      CALL create_colvar_section(subsection, skip_recursive_colvar=.TRUE.)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, name="FUNCTION", &
                          description="Specifies the function used to combine different COLVARs into one.", &
                          ! **************************************************************************************************
                          !> \brief ...
                          !> \param CV1^2 ...
                          !> \param CV2^2 ...
                          !> \param " ...
                          !> \param type_of_var=lchar_t ...
                          !> \param n_var=1 ...
                          !> \param error=error ...
                          ! **************************************************************************************************
                          usage="FUNCTION SQRT(CV1^2+CV2^2)", type_of_var=lchar_t, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VARIABLES", &
                          description="Specifies the name of the variable that parametrises the FUNCTION "// &
                          "defining how COLVARS should be combined. The matching follows the same order of the "// &
                          "COLVARS definition in the input file.", &
                          usage="VARIABLE CV1 CV2 CV3", type_of_var=char_t, n_var=-1, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARAMETERS", &
                          description="Defines the parameters of the functional form", &
                          usage="PARAMETERS a b D", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VALUES", &
                          description="Defines the values of  parameter of the functional form", &
                          usage="VALUES ", type_of_var=real_t, &
                          n_var=-1, repeats=.TRUE., unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DX", &
                          description="Parameter used for computing the derivative of the combination "// &
                          "of COLVARs with the Ridders' method.", &
                          usage="DX <REAL>", default_r_val=0.1_dp, unit_str="bohr")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ERROR_LIMIT", &
                          description="Checks that the error in computing the derivative is not larger than "// &
                          "the value set; in case error is larger a warning message is printed.", &
                          usage="ERROR_LIMIT <REAL>", default_r_val=1.0E-12_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_colvar_comb_section

! **************************************************************************************************
!> \brief Creates the coord section
!> \param section the section to create
!> \param name ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_coord_section_cv(section, name)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*), INTENT(IN)                       :: name

      CHARACTER(len=*), PARAMETER :: routineN = 'create_coord_section_cv', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",2084)
      CALL section_create(section, name="coord", &
                          description="The positions for "//TRIM(name)//" used for restart", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify positions of the system", repeats=.TRUE., &
                          usage="{Real} ...", type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_coord_section_cv

! **************************************************************************************************
!> \brief collective variables specifying h bonds
!> \param section the section to be created
!> \author alin m elena
! **************************************************************************************************
   SUBROUTINE create_colvar_wc_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_wc_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",2112)
      CALL section_create(section, name="wc", &
                          description="Section to define the hbond wannier centre as a collective variables.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Parameter used for computing the cutoff radius for searching "// &
                          "the wannier centres around an atom", &
                          usage="RCUT <REAL>", default_r_val=0.529177208590000_dp, unit_str="angstrom", &
                          type_of_var=real_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          variants=(/"POINTS"/), &
                          description="Specifies the indexes of atoms/points defining the bond (Od, H, Oa).", &
                          usage="ATOMS {integer} {integer} {integer}", &
                          n_var=3, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_wc_section

   ! **************************************************************************************************
!> \brief collective variables specifying h bonds= wire
!> \param section the section to be created
!> \author alin m elena
! **************************************************************************************************
   SUBROUTINE create_colvar_hbp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_hbp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",2155)
      CALL section_create(section, name="hbp", &
                          description="Section to define the hbond wannier centre as a collective variables.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="RCUT", &
                          description="Parameter used for computing the cutoff radius for searching "// &
                          "the wannier centres around an atom", &
                          usage="RCUT <REAL>", default_r_val=0.529177208590000_dp, unit_str="angstrom", &
                          type_of_var=real_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SHIFT", &
                          description="Parameter used for shifting each term in the sum ", &
                          usage="SHIFT <REAL>", default_r_val=0.5_dp, &
                          type_of_var=real_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NPOINTS", &
                          description="The number of points in the path", &
                          usage="NPOINTS {integer}", default_i_val=-1, &
                          n_var=1, type_of_var=integer_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          variants=(/"POINTS"/), &
                          description="Specifies the indexes of atoms/points defining the bond (Od, H, Oa).", &
                          usage="ATOMS {integer} {integer} {integer}", &
                          n_var=3, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_hbp_section

! **************************************************************************************************
!> \brief collective variables specifying ring puckering
!> \brief D. Cremer and J.A. Pople, JACS 97 1354 (1975)
!> \param section the section to be created
!> \author Marcel Baer
! **************************************************************************************************
   SUBROUTINE create_colvar_ring_puckering_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_colvar_ring_puckering_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_colvar.F",2213)
      CALL section_create(section, name="RING_PUCKERING", &
                          description="Section to define general ring puckering collective variables.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="ATOMS", &
                          variants=(/"POINTS"/), &
                          description="Specifies the indexes of atoms/points defining the ring."// &
                          "At least 4 Atoms are needed.", &
                          usage="ATOMS {integer} {integer} {integer} ..", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COORDINATE", &
                          description="Indicate the coordinate to be used. Follow the Cremer-Pople definition for a N ring."// &
                          "0 is the total puckering variable Q,"// &
                          "2..[N/2] are puckering coordinates."// &
                          "-2..-[N/2-1] are puckering angles.", &
                          usage="COORDINATE {integer}", default_i_val=0, &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Must be present in each colvar and handled properly
      CALL create_point_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_colvar_ring_puckering_section

! **************************************************************************************************

END MODULE input_cp2k_colvar
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief builds the subsystem section of the input
!> \par History
!>      10.2005 split input_cp2k [fawzi]
!> \author teo & fawzi
! **************************************************************************************************
MODULE input_cp2k_subsys
   USE bibliography,                    ONLY: Goedecker1996,&
                                              Guidon2010,&
                                              Hartwigsen1998,&
                                              Krack2005,&
                                              VandeVondele2005a,&
                                              VandeVondele2007
   USE cell_types,                      ONLY: &
        cell_sym_cubic, cell_sym_hexagonal, cell_sym_monoclinic, cell_sym_none, &
        cell_sym_orthorhombic, cell_sym_rhombohedral, cell_sym_tetragonal_ab, &
        cell_sym_tetragonal_ac, cell_sym_tetragonal_bc, cell_sym_triclinic, use_perd_none, &
        use_perd_x, use_perd_xy, use_perd_xyz, use_perd_xz, use_perd_y, use_perd_yz, use_perd_z
   USE cp_output_handling,              ONLY: cp_print_key_section_create,&
                                              debug_print_level,&
                                              high_print_level,&
                                              medium_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        do_add, do_bondparm_covalent, do_bondparm_vdw, do_cell_cp2k, do_cell_xsc, do_conn_amb7, &
        do_conn_g87, do_conn_g96, do_conn_generate, do_conn_mol_set, do_conn_off, do_conn_psf, &
        do_conn_psf_u, do_conn_user, do_coord_cif, do_coord_cp2k, do_coord_crd, do_coord_g96, &
        do_coord_off, do_coord_pdb, do_coord_xtl, do_coord_xyz, do_remove, do_skip_11, do_skip_12, &
        do_skip_13, do_skip_14, dump_pdb, gaussian
   USE input_cp2k_colvar,               ONLY: create_colvar_section
   USE input_cp2k_mm,                   ONLY: create_neighbor_lists_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              lchar_t,&
                                              real_t
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE physcon,                         ONLY: bohr
   USE string_utilities,                ONLY: newline,&
                                              s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_subsys'

   PUBLIC :: create_subsys_section, &
             create_cell_section, &
             create_structure_data_section, &
             create_rng_section
!***
CONTAINS

! **************************************************************************************************
!> \brief creates the cell section
!> \param section ...
!> \param periodic ...
!> \param label ...
!> \author teo
! **************************************************************************************************
   RECURSIVE SUBROUTINE create_cell_section(section, periodic, label)
      TYPE(section_type), POINTER                        :: section
      INTEGER, INTENT(IN), OPTIONAL                      :: periodic
      CHARACTER(LEN=*), OPTIONAL                         :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'create_cell_section', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_label
      INTEGER                                            :: my_periodic
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      my_periodic = use_perd_xyz
      my_label = "CELL"
      IF (PRESENT(periodic)) my_periodic = periodic
      IF (PRESENT(label)) my_label = label
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",93)
      CALL section_create(section, TRIM(my_label), &
                          description="Input parameters needed to set up the "//TRIM(my_label)//".", &
                          n_keywords=6, n_subsections=1, repeats=.FALSE.)

      NULLIFY (keyword, subsection)
      CALL keyword_create(keyword, name="A", &
                          description="Specify the Cartesian components for the cell vector A. "// &
                          "This defines the first column of the h matrix.", &
                          usage="A  10.000  0.000  0.000", unit_str="angstrom", &
                          n_var=3, type_of_var=real_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="B", &
                          description="Specify the Cartesian components for the cell vector B. "// &
                          "This defines the second column of the h matrix.", &
                          usage="B   0.000 10.000  0.000", unit_str="angstrom", &
                          n_var=3, type_of_var=real_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="C", &
                          description="Specify the Cartesian components for the cell vector C. "// &
                          "This defines the third column of the h matrix.", &
                          usage="C   0.000  0.000 10.000", unit_str="angstrom", &
                          n_var=3, type_of_var=real_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ABC", &
                          description="Specify the lengths of the cell vectors A, B, and C, which"// &
                          " defines the diagonal elements of h matrix for an orthorhombic cell."// &
                          " For non-orthorhombic cells it is possible either to specify the angles "// &
                          "ALPHA, BETA, GAMMA via ALPHA_BETA_GAMMA keyword or alternatively use the keywords "// &
                          "A, B, and C. The convention is that A lies along the X-axis, B is in the XY plane.", &
                          usage="ABC 10.000 10.000 10.000", unit_str="angstrom", &
                          n_var=3, type_of_var=real_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALPHA_BETA_GAMMA", &
                          variants=(/"ANGLES"/), &
                          description="Specify the angles between the vectors A, B and C when using the ABC keyword. "// &
                          "The convention is that A lies along the X-axis, B is in the XY plane. "// &
                          "ALPHA is the angle between B and C, BETA is the angle between A and C and "// &
                          "GAMMA the angle between A and B.", &
                          usage="ALPHA_BETA_GAMMA [deg] 90.0 90.0 120.0", unit_str="deg", &
                          n_var=3, default_r_vals=(/cp_unit_to_cp2k(value=90.0_dp, unit_str="deg"), &
                                                    cp_unit_to_cp2k(value=90.0_dp, unit_str="deg"), &
                                                    cp_unit_to_cp2k(value=90.0_dp, unit_str="deg")/), &
                          repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CELL_FILE_NAME", &
                          description="Possibility to read the cell from an external file ", &
                          repeats=.FALSE., usage="CELL_FILE_NAME <CHARACTER>", &
                          type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CELL_FILE_FORMAT", &
                          description="Specify the format of the cell file (if used)", &
                          usage="CELL_FILE_FORMAT (CP2K|XSC)", &
                          enum_c_vals=s2a("CP2K", "XSC"), &
                          enum_i_vals=(/do_cell_cp2k, do_cell_xsc/), &
                          enum_desc=s2a("Cell info in the CP2K native format.", &
                                        "Cell info in the XSC format (NAMD)"), &
                          default_i_val=do_cell_cp2k)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PERIODIC", &
                          description="Specify the directions for which periodic boundary conditions (PBC) will be applied. "// &
                          "Important notice: This applies to the generation of the pair lists as well as to the "// &
                          "application of the PBCs to positions. "// &
                          "See the POISSON section to specify the periodicity used for the electrostatics. "// &
                          "Typically the settings should be the same.", &
                          usage="PERIODIC (x|y|z|xy|xz|yz|xyz|none)", &
                          enum_c_vals=s2a("x", "y", "z", "xy", "xz", "yz", "xyz", "none"), &
                          enum_i_vals=(/use_perd_x, use_perd_y, use_perd_z, &
                                        use_perd_xy, use_perd_xz, use_perd_yz, &
                                        use_perd_xyz, use_perd_none/), &
                          default_i_val=my_periodic)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MULTIPLE_UNIT_CELL", &
                          description="Specifies the numbers of repetition in space (X, Y, Z) of the defined cell, "// &
                          "assuming it as a unit cell. This keyword affects only the CELL specification. The same keyword "// &
                          "in SUBSYS%TOPOLOGY%MULTIPLE_UNIT_CELL should be modified in order to affect the coordinates "// &
                          "specification.", usage="MULTIPLE_UNIT_CELL 1 1 1", &
                          n_var=3, default_i_vals=(/1, 1, 1/), repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="SYMMETRY", &
         description="Imposes an initial cell symmetry.", &
         usage="SYMMETRY monoclinic", &
         enum_desc=s2a("No cell symmetry", &
                       "Triclinic (a &ne; b &ne; c &ne; a, &alpha; &ne; &beta; &ne; &gamma; &ne; &alpha; &ne; 90&deg;)", &
                       "Monoclinic (a &ne; b &ne; c &ne; a, &alpha; = &gamma; = 90&deg;, &beta; &ne; 90&deg;)", &
                       "Orthorhombic (a &ne; b &ne; c, &alpha; = &beta; = &gamma; = 90&deg;)", &
                       "Tetragonal (a = b &ne; c, &alpha; = &beta; = &gamma; = 90&deg;)", &
                       "Tetragonal (a = c &ne; b, &alpha; = &beta; = &gamma; = 90&deg;)", &
                       "Tetragonal (a &ne; b = c, &alpha; = &beta; = &gamma; = 90&deg;)", &
                       "Tetragonal (alias for TETRAGONAL_AB)", &
                       "Rhombohedral (a = b = c, &alpha; = &beta; = &gamma; &ne; 90&deg;)", &
                       "Hexagonal (a = b &ne; c, &alpha; = &beta; = 90&deg;, &gamma; = 60&deg;)", &
                       "Cubic (a = b = c, &alpha; = &beta; = &gamma; = 90&deg;)"), &
         enum_c_vals=s2a("NONE", "TRICLINIC", "MONOCLINIC", "ORTHORHOMBIC", "TETRAGONAL_AB", "TETRAGONAL_AC", &
                         "TETRAGONAL_BC", "TETRAGONAL", "RHOMBOHEDRAL", "HEXAGONAL", "CUBIC"), &
         enum_i_vals=(/cell_sym_none, cell_sym_triclinic, cell_sym_monoclinic, cell_sym_orthorhombic, &
                       cell_sym_tetragonal_ab, cell_sym_tetragonal_ac, cell_sym_tetragonal_bc, &
                       cell_sym_tetragonal_ab, cell_sym_rhombohedral, cell_sym_hexagonal, cell_sym_cubic/), &
         default_i_val=cell_sym_none)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      IF (TRIM(my_label) == "CELL") THEN
         CALL create_cell_section(subsection, periodic, "CELL_REF")
         CALL section_add_subsection(section, subsection)
         CALL section_release(subsection)
      END IF
   END SUBROUTINE create_cell_section

! **************************************************************************************************
!> \brief Creates the random number restart section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_rng_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_rng_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",234)
      CALL section_create(section, name="RNG_INIT", &
                          description="Information to initialize the parallel random number generator streams", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Specify an initial RNG stream record", repeats=.TRUE., &
                          usage="{RNG record string}", type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_rng_section

! **************************************************************************************************
!> \brief creates the structure of a subsys, i.e. a full set of
!>      atoms+mol+bounds+cell
!> \param section the section to create
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_subsys_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_subsys_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",262)
      CALL section_create(section, name="subsys", &
                          description="a subsystem: coordinates, topology, molecules and cell", &
                          n_keywords=0, n_subsections=9, repeats=.FALSE.)

      NULLIFY (subsection)

      CALL create_rng_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_cell_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_coord_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_velocity_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_kind_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_topology_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_colvar_section(section=subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_multipole_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_shell_coord_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_shell_vel_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_core_coord_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_core_vel_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_subsys_print_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_subsys_section

! **************************************************************************************************
!> \brief Creates the subsys print section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_subsys_print_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_subsys_print_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      NULLIFY (print_key, keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",338)
      CALL section_create(section, name="print", &
                          description="Controls printings related to the subsys", &
                          n_keywords=0, n_subsections=9, repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "atomic_coordinates", &
                                       description="controls the output of the atomic coordinates when setting up the"// &
                                       "force environment. For printing coordinates during MD or GEO refer to the keyword"// &
                                       " trajectory.", unit_str="angstrom", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL create_structure_data_section(print_key)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "interatomic_distances", &
                                       description="controls the output of the interatomic distances when setting up the"// &
                                       "force environment", unit_str="angstrom", &
                                       print_level=debug_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "topology_info", description= &
                                       "controls the printing of information in the topology settings", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword, "xtl_info", &
                          description="Prints information when parsing XTL files.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "cif_info", &
                          description="Prints information when parsing CIF files.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "pdb_info", &
                          description="Prints information when parsing PDB files.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "xyz_info", &
                          description="Prints information when parsing XYZ files.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "psf_info", &
                          description="Prints information when parsing PSF files.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "amber_info", &
                          description="Prints information when parsing ABER topology files.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "g96_info", &
                          description="Prints information when parsing G96 files.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "crd_info", &
                          description="Prints information when parsing CRD files.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "gtop_info", &
                          description="Prints information when parsing GROMOS topology files.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "util_info", &
                          description="Prints information regarding topology utilities", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "generate_info", &
                          description="Prints information regarding topology generation", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "cell", &
                                       description="controls the output of the cell parameters", &
                                       print_level=medium_print_level, filename="__STD_OUT__", &
                                       unit_str="angstrom")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "kinds", &
                                       description="controls the output of information on the kinds", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="potential", &
                          description="If the printkey is activated controls the printing of the"// &
                          " fist_potential, gth_potential or all electron"// &
                          " potential information", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="basis_set", &
                          description="If the printkey is activated controls the printing of basis set information", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="se_parameters", &
                          description="If the printkey is activated controls the printing of the semi-empirical parameters.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "SYMMETRY", &
                                       description="controls the output of symmetry information", &
                                       print_level=debug_print_level+1, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="MOLECULE", &
                          description="Assume the system is an isolated molecule", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="EPS_GEO", &
                          description="Accuracy required for symmetry detection", &
                          default_r_val=1.e-4_dp)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="STANDARD_ORIENTATION", &
                          description="Print molecular coordinates in standard orientation", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="INERTIA", &
                          description="Print molecular inertia tensor", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="SYMMETRY_ELEMENTS", &
                          description="Print symmetry elements", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="ALL", &
                          description="Print all symmetry information", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="ROTATION_MATRICES", &
                          description="All the rotation matrices of the point group", &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="CHECK_SYMMETRY", &
                          description="Check if calculated symmetry has expected value."// &
                          " Use either Schoenfliess or Hermann-Maugin symbols", &
                          default_c_val="NONE")
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "molecules", &
                                       description="controls the output of information on the molecules", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "radii", &
                                       description="controls the output of radii information", unit_str="angstrom", &
                                       print_level=high_print_level, filename="__STD_OUT__")

      CALL keyword_create(keyword, name="core_charges_radii", &
                          description="If the printkey is activated controls the printing of the radii of the core charges", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="pgf_radii", &
                          description="If the printkey is activated controls the printing of the core gaussian radii", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="set_radii", &
                          description="If the printkey is activated controls the printing of the set_radii", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="kind_radii", &
                          description="If the printkey is activated controls the printing of the kind_radii", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="core_charge_radii", &
                          description="If the printkey is activated controls the printing of the core_charge_radii", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="gth_ppl_radii", &
                          description="If the printkey is activated controls the printing of the "// &
                          "gth pseudo potential local radii", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="gth_ppnl_radii", &
                          description="If the printkey is activated controls the printing of the "// &
                          "gth pseudo potential non local radii", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="gapw_prj_radii", &
                          description="If the printkey is activated controls the printing of the gapw projector radii", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_subsys_print_section

! **************************************************************************************************
!> \brief Creates the multipole section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_multipole_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_multipole_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",579)
      CALL section_create(section, name="multipoles", &
                          description="Specifies the dipoles and quadrupoles for particles.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)
      CALL section_create(subsection, name="dipoles", &
                          description="Specifies the dipoles of the particles.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="The dipole components for each atom in the format:"// &
                          "<p><tt><big>D<sub>x</sub> D<sub>y</sub> D<sub>z</sub></big></tt></p>", &
                          repeats=.TRUE., usage="{Real} {Real} {Real}", &
                          type_of_var=real_t, n_var=3)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="quadrupoles", &
                          description="Specifies the quadrupoles of the particles.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="The quadrupole components for each atom in the format:"// &
                          "<p><big><tt>Q<sub>xx</sub> Q<sub>xy</sub> Q<sub>xz</sub> Q<sub>yy</sub> "// &
                          "Q<sub>yz</sub> Q<sub>zz</sub></big></tt></p>", &
                          repeats=.TRUE., usage="{Real} {Real} {Real} {Real} {Real} {Real}", &
                          type_of_var=real_t, n_var=6)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_multipole_section

! **************************************************************************************************
!> \brief creates structure data section for output.. both subsys  (for initialization)
!>      and motion section..
!> \param print_key ...
! **************************************************************************************************
   SUBROUTINE create_structure_data_section(print_key)
      TYPE(section_type), POINTER                        :: print_key

      CHARACTER(len=*), PARAMETER :: routineN = 'create_structure_data_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(print_key)))CALL cp__a("input_cp2k_subsys.F",627)

      NULLIFY (keyword)

      CALL cp_print_key_section_create(print_key, name="STRUCTURE_DATA", &
                                       description="Request the printing of special structure data during a structure "// &
                                       "optimization (in MOTION%PRINT) or when setting up a subsys (in SUBSYS%PRINT).", &
                                       print_level=high_print_level, filename="__STD_OUT__", unit_str="angstrom")

      CALL keyword_create(keyword, name="POSITION", variants=(/"POS"/), &
                          description="Print the position vectors in Cartesian coordinates of the atoms specified "// &
                          "by a list of their indices", &
                          usage="POSITION {integer} {integer} {integer}..{integer}", n_var=-1, repeats=.TRUE., &
                          type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POSITION_SCALED", variants=(/"POS_SCALED"/), &
                          description="Print the position vectors in scaled coordinates of the atoms specified "// &
                          "by a list of their indices", &
                          usage="POSITION_SCALED {integer} {integer} {integer}..{integer}", n_var=-1, repeats=.TRUE., &
                          type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISTANCE", variants=(/"DIS"/), &
                          description="Print the distance between the atoms a and b specified by their indices", &
                          usage="DISTANCE {integer} {integer}", n_var=2, repeats=.TRUE., &
                          type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ANGLE", variants=(/"ANG"/), &
                          description="Print the angle formed by the atoms specified by their indices", &
                          usage="ANGLE {integer} {integer} {integer}", n_var=3, repeats=.TRUE., &
                          type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DIHEDRAL_ANGLE", variants=s2a("DIHEDRAL", "DIH"), &
                          description="Print the dihedral angle between the planes defined by the atoms (a,b,c) and "// &
                          "the atoms (b,c,d) specified by their indices", &
                          usage="DIHEDRAL_ANGLE {integer}  {integer} {integer} {integer}", n_var=4, &
                          repeats=.TRUE., type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_structure_data_section

! **************************************************************************************************
!> \brief Creates the velocity section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_velocity_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_velocity_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",689)
      CALL section_create(section, name="velocity", &
                          description="The velocities for simple systems or "// &
                          "the centroid mode in PI runs, xyz format by default", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="PINT_UNIT", &
                          description="Specify the units of measurement for the velocities "// &
                          "(currently works only for the path integral code). "// &
                          "All available CP2K units can be used.", &
                          usage="UNIT angstrom*au_t^-1", &
                          default_c_val="bohr*au_t^-1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="The atomic velocities in the format:"// &
                          "<p><tt><big>v<sub>x</sub> v<sub>y</sub> v<sub>z</sub></big></tt></p>"// &
                          "The same order as for the atomic coordinates is assumed.", &
                          repeats=.TRUE., usage="{Real} {Real} {Real}", &
                          type_of_var=real_t, n_var=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_velocity_section

! **************************************************************************************************
!> \brief Creates the shell velocity section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_shell_vel_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_shell_vel_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",728)
      CALL section_create(section, name="shell_velocity", &
                          description="The velocities of shells for shell-model potentials, "// &
                          "in xyz format  ", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="The shell particle velocities in the format:"// &
                          "<p><tt><big>v<sub>x</sub> v<sub>y</sub> v<sub>z</sub></big></tt></p>"// &
                          "The same order as for the shell particle coordinates is assumed.", &
                          repeats=.TRUE., usage="{Real} {Real} {Real}", &
                          type_of_var=real_t, n_var=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_shell_vel_section

! **************************************************************************************************
!> \brief Creates the shell velocity section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_core_vel_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_core_vel_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",759)
      CALL section_create(section, name="core_velocity", &
                          description="The velocities of cores for shell-model potentials, "// &
                          "in xyz format  ", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="The core particle velocities in the format:"// &
                          "<p><tt><big>v<sub>x</sub> v<sub>y</sub> v<sub>z</sub></big></tt></p>"// &
                          "The same order as for the core particle coordinates is assumed.", &
                          repeats=.TRUE., usage="{Real} {Real} {Real}", &
                          type_of_var=real_t, n_var=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_core_vel_section

! **************************************************************************************************
!> \brief Creates the &POTENTIAL section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_potential_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_potential_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      CALL section_create(section, name="potential", &
                          description="Section used to specify Potentials.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="CP2K Pseudo Potential Standard Format (GTH, ALL)", &
                          repeats=.TRUE., type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_potential_section

! **************************************************************************************************
!> \brief Creates the &KG_POTENTIAL section
!> \param section the section to create
!> \author JGH
! **************************************************************************************************
   SUBROUTINE create_kgpot_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_kgpot_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      CALL section_create(section, name="kg_potential", &
                          description="Section used to specify KG Potentials.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="CP2K KG TNADD Potential Standard Format (TNADD)", &
                          repeats=.TRUE., type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_kgpot_section

! **************************************************************************************************
!> \brief Creates the &BASIS section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_basis_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_basis_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      CALL section_create(section, name="BASIS", &
                          description="Section used to specify a general basis set for QM calculations.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="The type of basis set defined in this section.", &
                          lone_keyword_c_val="Orbital", &
                          usage="Orbital", default_c_val="Orbital")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="_DEFAULT_KEYWORD_", &
         repeats=.TRUE., type_of_var=lchar_t, &
         description="<u>CP2K Basis Set Standard Format</u>"//newline// &
         "<pre>"//newline// &
         "Element symbol  Name of the basis set  Alias names"//newline// &
         "nset (repeat the following block of lines nset times)"//newline// &
         "n lmin lmax nexp nshell(lmin) nshell(lmin+1) ... nshell(lmax-1) nshell(lmax)"//newline// &
         "a(1)      c(1,l,1)      c(1,l,2) ...      c(1,l,nshell(l)-1)      c(1,l,nshell(l)), l=lmin,lmax"//newline// &
         "a(2)      c(2,l,1)      c(2,l,2) ...      c(2,l,nshell(l)-1)      c(2,l,nshell(l)), l=lmin,lmax"//newline// &
         " .         .             .                 .                       ."//newline// &
         " .         .             .                 .                       ."//newline// &
         " .         .             .                 .                       ."//newline// &
         "a(nexp-1) c(nexp-1,l,1) c(nexp-1,l,2) ... c(nexp-1,l,nshell(l)-1) c(nexp-1,l,nshell(l)), l=lmin,lmax"//newline// &
         "a(nexp)   c(nexp,l,1)   c(nexp,l,2)   ... c(nexp,l,nshell(l)-1)   c(nexp,l,nshell(l)), l=lmin,lmax"//newline// &
         newline// &
         newline// &
         "nset     : Number of exponent sets"//newline// &
         "n        : Principle quantum number (only for orbital label printing)"//newline// &
         "lmax     : Maximum angular momentum quantum number l"//newline// &
         "lmin     : Minimum angular momentum quantum number l"//newline// &
         "nshell(l): Number of shells for angular momentum quantum number l"//newline// &
         "a        : Exponent"//newline// &
         "c        : Contraction coefficient"//newline// &
         "</pre>"//newline// &
         "Source: ftp://ftp.aip.org/epaps/journ_chem_phys/E-JCPSA6-127-308733/BASIS_MOLOPT_JCP.txt")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_basis_section

! **************************************************************************************************
!> \brief Creates the &COORD section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_coord_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_coord_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",897)
      CALL section_create(section, name="coord", &
                          description="The coordinates for simple systems (like small QM cells)"// &
                          " are specified here by default using explicit XYZ coordinates.  "// &
                          " More complex systems should be given via an external coordinate "// &
                          " file in the SUBSYS%TOPOLOGY section.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="UNIT", &
                          description='Specify the unit of measurement for the coordinates in input'// &
                          "All available CP2K units can be used.", &
                          usage="UNIT angstrom", default_c_val="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SCALED", &
                          description='Specify if the coordinateds in input are scaled.', &
                          usage="SCALED F", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="The atomic coordinates in the format:"// &
                          "<p><tt>ATOMIC_KIND  X Y Z  MOLNAME</tt></p>"// &
                          "The <tt>MOLNAME</tt> is optional. If not provided the molecule name "// &
                          "is internally created. All other fields after <tt>MOLNAME</tt> are simply ignored.", &
                          repeats=.TRUE., usage="{{String} {Real} {Real} {Real} {String}}", &
                          type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_coord_section

! **************************************************************************************************
!> \brief Creates the &SHELL_COORD section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_shell_coord_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_shell_coord_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",943)
      CALL section_create(section, name="shell_coord", &
                          description="The shell coordinates for the shell-model potentials"// &
                          " xyz format with an additional column for the index of the corresponding particle", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="UNIT", &
                          description='Specify the unit of measurement for the coordinates in input'// &
                          "All available CP2K units can be used.", &
                          usage="UNIT angstrom", default_c_val="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SCALED", &
                          description='Specify if the coordinateds in input are scaled.', &
                          usage="SCALED F", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="The shell particle coordinates in the format:"// &
                          "<p><tt>ATOMIC_KIND  X Y Z  ATOMIC_INDEX</tt></p>"// &
                          "The <tt>ATOMIC_INDEX</tt> refers to the atom the shell particle belongs to.", &
                          repeats=.TRUE., usage="{{String} {Real} {Real} {Real} {Integer}}", &
                          type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_shell_coord_section

! **************************************************************************************************
!> \brief Creates the &core_COORD section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_core_coord_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_core_coord_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",987)
      CALL section_create(section, name="core_coord", &
                          description="The core coordinates for the shell-model potentials"// &
                          " xyz format with an additional column for the index of the corresponding particle", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="UNIT", &
                          description='Specify the unit of measurement for the coordinates in input'// &
                          "All available CP2K units can be used.", &
                          usage="UNIT angstrom", default_c_val="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SCALED", &
                          description='Specify if the coordinateds in input are scaled.', &
                          usage="SCALED F", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="The core particle coordinates in the format:"// &
                          "<p><tt>ATOMIC_KIND  X Y Z  ATOMIC_INDEX</tt></p>"// &
                          "The <tt>ATOMIC_INDEX</tt> refers to the atom the core particle belongs to.", &
                          repeats=.TRUE., usage="{{String} {Real} {Real} {Real} {Integer}}", &
                          type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_core_coord_section

! **************************************************************************************************
!> \brief Creates the QM/MM section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_kind_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_kind_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",1032)

      CALL section_create(section, name="KIND", &
                          description="The description of the kind of the atoms (mostly for QM)", &
                          n_keywords=19, n_subsections=1, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="The name of the kind described in this section.", &
                          usage="H", default_c_val="DEFAULT")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BASIS_SET", &
                          description="The primary Gaussian basis set (NONE implies no basis used, meaningful with GHOST)", &
                          usage="BASIS_SET [type] DZVP", type_of_var=char_t, default_c_vals=(/" ", " "/), &
                          citations=(/VandeVondele2005a, VandeVondele2007/), &
                          repeats=.TRUE., n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! old type basis set input keywords
      ! kept for back compatability
      CALL keyword_create(keyword, name="AUX_BASIS_SET", &
                          variants=s2a("AUXILIARY_BASIS_SET", "AUX_BASIS"), &
                          description="DEPRECATED (use BASIS_SET): The auxliliary basis set (GTO type)", &
                          usage="AUX_BASIS_SET DZVP", default_c_val=" ", &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RI_AUX_BASIS_SET", &
                          variants=s2a("RI_MP2_BASIS_SET", "RI_RPA_BASIS_SET", "RI_AUX_BASIS"), &
                          description="DEPRECATED (use BASIS_SET): The RI auxliliary basis set used in WF_CORRELATION (GTO type)", &
                          usage="RI_AUX_BASIS_SET DZVP", default_c_val=" ", &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LRI_BASIS_SET", &
                          variants=s2a("LRI_BASIS"), &
                          description="DEPRECATED (use BASIS_SET): The local resolution of identity basis set (GTO type)", &
                          usage="LRI_BASIS_SET DZVP", default_c_val=" ", &
                          n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="AUX_FIT_BASIS_SET", &
         variants=s2a("AUXILIARY_FIT_BASIS_SET", "AUX_FIT_BASIS"), &
         description="DEPRECATED (use BASIS_SET): The auxliliary basis set (GTO type) for auxiliary density matrix method", &
         usage="AUX_FIT_BASIS_SET DZVP", default_c_val=" ", &
         citations=(/Guidon2010/), &
         n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      ! end of old basis set keywords

      CALL keyword_create(keyword, name="ELEC_CONF", &
                          description="Specifies the electronic configration used in construction the "// &
                          "atomic initial guess (see the pseudo potential file for the default values.", &
                          usage="ELEC_COND n_elec(s)  n_elec(p)  n_elec(d)  ... ", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CORE_CORRECTION", &
                          description="Corrects the effective nuclear charge", &
                          usage="CORE_CORRECTION 1.0", n_var=1, &
                          default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ELEMENT", &
                          variants=(/"ELEMENT_SYMBOL"/), &
                          description="The element of the actual kind "// &
                          "(if not given it is inferred from the kind name)", &
                          usage="ELEMENT O", type_of_var=char_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MASS", &
                          variants=s2a("ATOMIC_MASS", "ATOMIC_WEIGHT", "WEIGHT"), &
                          description="The mass of the atom "// &
                          "(if negative or non present it is inferred from the element symbol)", &
                          usage="MASS 2.0", type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POTENTIAL", &
                          variants=(/"POT"/), &
                          description="The name of the pseudopotential for the defined kind.", &
                          usage="POTENTIAL <PSEUDO-POTENTIAL-NAME>", default_c_val="GTH", n_var=1, &
                          citations=(/Goedecker1996, Hartwigsen1998, Krack2005/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KG_POTENTIAL", &
                          variants=(/"KG_POT"/), &
                          description="The name of the non-additive atomic kinetic energy potential.", &
                          usage="KG_POTENTIAL <TNADD-POTENTIAL-NAME>", default_c_val="NONE", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="HARD_EXP_RADIUS", &
                          description="The region where the hard density is supposed to be confined"// &
                          "(GAPW)(in Bohr, default is 1.2 for H and 1.512 otherwise)", &
                          usage="HARD_EXP_RADIUS 0.9", type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_RAD_LOCAL", &
                          description="Max radius for the basis functions used to"// &
                          " generate the local projectors in GAPW [Bohr]", &
                          usage="MAX_RAD_LOCAL 15.0", default_r_val=13.0_dp*bohr)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RHO0_EXP_RADIUS", &
                          description="the radius which defines the atomic region where "// &
                          "the hard compensation density is confined."// &
                          "should be less than HARD_EXP_RADIUS (GAPW)(Bohr, default equals HARD_EXP_RADIUS)", &
                          usage="RHO_EXP_RADIUS 0.9", type_of_var=real_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LEBEDEV_GRID", &
                          description="The number of points for the angular part of "// &
                          "the local grid (GAPW)", &
                          usage="LEBEDEV_GRID 40", default_i_val=50)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RADIAL_GRID", &
                          description="The number of points for the radial part of "// &
                          "the local grid (GAPW)", &
                          usage="RADIAL_GRID 70", default_i_val=50)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MM_RADIUS", &
                          description="Defines the radius of the electrostatic multipole "// &
                          "of the atom in Fist. This radius applies to the charge, the "// &
                          "dipole and the quadrupole. When zero, the atom is treated as "// &
                          "a point multipole, otherwise it is treated as a Gaussian "// &
                          "charge distribution with the given radius: "// &
                          "p(x,y,z)*N*exp(-(x**2+y**2+z**2)/(2*MM_RADIUS**2)), where N is "// &
                          "a normalization constant. In the core-shell model, only the "// &
                          "shell is treated as a Gaussian and the core is always a point "// &
                          "charge.", &
                          usage="MM_RADIUS {real}", default_r_val=0.0_dp, type_of_var=real_t, &
                          unit_str="angstrom", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DFTB3_PARAM", &
                          description="The third order parameter (derivative of hardness) used in "// &
                          "diagonal DFTB3 correction.", &
                          usage="DFTB3_PARAM 0.2", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LMAX_DFTB", &
                          description="The maximum l-quantum number of the DFTB basis for this kind.", &
                          usage="LMAX_DFTB 1", default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAO", &
                          description="The number of MAOs (Modified Atomic Orbitals) for this kind.", &
                          usage="MAO 4", default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Logicals
      CALL keyword_create(keyword, name="SE_P_ORBITALS_ON_H", &
                          description="Forces the usage of p-orbitals on H for SEMI-EMPIRICAL calculations. "// &
                          " This keyword applies only when the KIND is specifying an Hydrogen element. In all "// &
                          " other cases is simply ignored. ", &
                          usage="SE_P_ORBITALS_ON_H", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GPW_TYPE", &
                          description="Force one type to be treated by the GPW scheme,"// &
                          " whatever are its primitives, even if the GAPW method is used", &
                          usage="GPW_TYPE", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="GHOST", &
                          description="This keyword makes all atoms of this kind "// &
                          "ghost atoms, i.e. without pseudo or nuclear charge."// &
                          "Useful to just have the basis set at that position (BSSE calculations),"// &
                          "or to have a non-interacting particle with BASIS_SET NONE", &
                          usage="GHOST", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="NO_OPTIMIZE", &
                          description="Skip optimization of this type (used in specific basis set or"// &
                          " potential optimization schemes", &
                          usage="NO_OPTIMIZE", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PAO_BASIS_SIZE", &
                          description="The block size used for the polarized atomic orbital basis.", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL create_pao_potential_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_basis_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_potential_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_kgpot_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_dft_plus_u_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_bs_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_kind_section

! **************************************************************************************************
!> \brief Creates the PAO_POTENTIAL section
!> \param section the section to create
!> \author Ole Schuett
! **************************************************************************************************
   SUBROUTINE create_pao_potential_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_pao_potential_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",1291)
      NULLIFY (keyword)

      CALL section_create(section, name="PAO_POTENTIAL", &
                          description="Settings of the PAO potentials, which are atomic kind specific.")

      CALL keyword_create(keyword, name="MAXL", &
                          description="Maximum angular moment of the potential.", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BETA", &
                          description="Scaling factor of the potential's Gaussian Hull functions. "// &
                          "The exponent is given by pot_exp = 0.5/(R_vdw*pot_beta)**2.", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHT", &
                          description="Weight of Gaussian potential term.", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_PROJECTOR", &
                          description="Maximum angular moment of the potential's projectors.", &
                          default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NEIGHBORS", &
                          description="Number of neigbors used for spanning the potential.", &
                          default_i_val=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_pao_potential_section

! **************************************************************************************************
!> \brief      Create CP2K input section for BS method: imposing atomic orbital occupation
!>             different from default in initialization of the density  matrix
!>             it works only with GUESS ATOMIC
!> \param section ...
!> \date       05.08.2009
!> \author     MI
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE create_bs_section(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_bs_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",1349)

      CALL section_create(section, &
                          name="BS", &
                          description="Define the required atomic orbital occupation "// &
                          "assigned in initialization of the density matrix, by adding or "// &
                          "subtracting electrons from specific angular momentum channels. "// &
                          "It works only with GUESS ATOMIC.", &
                          n_keywords=0, &
                          n_subsections=2, &
                          repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, &
                          name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the BS section", &
                          usage="&BS ON", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="ALPHA", description="alpha spin", &
                          n_keywords=3, &
                          n_subsections=0, &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, &
                          name="NEL", &
                          description="Orbital ccupation change per angular momentum quantum number."// &
                          "In unrestricted calculations applied to spin alpha.", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          default_i_val=-1, &
                          usage="NEL 2")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="L", &
                          variants=(/"L"/), &
                          description="Angular momentum quantum number of the "// &
                          "orbitals whose occupation is changed", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          default_i_val=-1, &
                          usage="L 2")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="N", &
                          variants=(/"N"/), &
                          description="Principal quantum number of the"// &
                          "orbitals whose occupation is changed. "// &
                          "Default is the first not occupied", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          default_i_val=0, &
                          usage="N 2")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="BETA", description="beta spin", &
                          n_keywords=3, &
                          n_subsections=0, &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, &
                          name="NEL", &
                          description="Orbital ccupation change per angular momentum quantum number."// &
                          "Applied to spin beta and active only in unrestricted calculations.", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          default_i_val=-1, &
                          usage="NEL 2")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="L", &
                          description="Angular momentum quantum number of the"// &
                          "orbitals of beta spin whose occupation is changed."// &
                          "Active only for unrestricted calculations", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          default_i_val=-1, &
                          usage="L 2")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="N", &
                          description="Principal quantum number of the"// &
                          "orbitals of beta spin whose occupation is changed. "// &
                          "Default is the first not occupied."// &
                          "Active only for unrestricted calculations", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          default_i_val=0, &
                          usage="N 2")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_bs_section

! **************************************************************************************************
!> \brief Create the topology section for FIST.. and the base is running running...
!>      Contains all information regarding topology to be read in input file..
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_topology_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_topology_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",1476)
      CALL section_create(section, name="TOPOLOGY", &
                          description="Section specifying information regarding how to handle the topology"// &
                          " for classical runs.", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, print_key)
      ! Logical
      CALL keyword_create(keyword, name="CHARGE_OCCUP", &
                          variants=(/"CHARGE_O"/), &
                          description="Read MM charges from the OCCUP field of PDB file.", &
                          usage="CHARGE_OCCUP logical", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHARGE_BETA", &
                          variants=(/"CHARGE_B"/), &
                          description="Read MM charges from the BETA field of PDB file.", &
                          usage="CHARGE_BETA logical", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHARGE_EXTENDED", &
                          description="Read MM charges from the very last field of PDB file (starting from column 81)."// &
                          " No limitations of number of digits.", &
                          usage="CHARGE_EXTENDED logical", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARA_RES", &
                          description="For a protein, each residue is now considered a molecule", &
                          usage="PARA_RES logical", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MOL_CHECK", &
                          description="Check molecules have the same number of atom and names.", &
                          usage="MOL_CHECK logical", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="USE_G96_VELOCITY", &
                          description="Use the velocities in the G96 coordinate files as the starting velocity", &
                          usage="USE_G96_VELOCITY logical", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Character
      CALL keyword_create(keyword, name="COORD_FILE_NAME", &
                          variants=s2a("COORD_FILE"), &
                          description="Specifies the filename that contains coordinates.", &
                          usage="COORD_FILE_NAME <FILENAME>", type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COORD_FILE_FORMAT", &
                          variants=s2a("COORDINATE"), &
                          description="Set up the way in which coordinates will be read.", &
                          usage="COORD_FILE_FORMAT (OFF|PDB|XYZ|G96|CRD|CIF|XTL|CP2K)", &
                          enum_c_vals=s2a("OFF", "PDB", "XYZ", "G96", "CRD", "CIF", "XTL", "CP2K"), &
                          enum_i_vals=(/do_coord_off, do_coord_pdb, do_coord_xyz, do_coord_g96, do_coord_crd, &
                                        do_coord_cif, do_coord_xtl, do_coord_cp2k/), &
                          enum_desc=s2a( &
                          "Coordinates read in the &COORD section of the input file", &
                          "Coordinates provided through a PDB file format", &
                          "Coordinates provided through an XYZ file format", &
                          "Coordinates provided through a GROMOS96 file format", &
                          "Coordinates provided through an AMBER file format", &
                          "Coordinates provided through a CIF (Crystallographic Information File) file format", &
                          "Coordinates provided through a XTL (MSI native) file format", &
                          "Read the coordinates in CP2K &COORD section format from an external file. "// &
                          "NOTE: This file will be overwritten with the latest coordinates."), &
                          default_i_val=do_coord_off)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUMBER_OF_ATOMS", &
                          variants=s2a("NATOMS", "NATOM"), &
                          description="Optionally define the number of atoms read from an external file "// &
                          "(see COORD_FILE_NAME) if the COORD_FILE_FORMAT CP2K is used", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=integer_t, &
                          default_i_val=-1, &
                          usage="NATOMS 768000")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL connectivity_framework(section, do_conn_generate)

      CALL keyword_create(keyword, name="DISABLE_EXCLUSION_LISTS", &
                          description="Do not build any exclusion lists.", &
                          usage="DISABLE_EXCLUSION_LISTS", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCLUDE_VDW", &
                          description="Specifies which kind of Van der Waals interaction to skip.", &
                          usage="EXCLUDE_VDW (1-1||1-2||1-3||1-4)", &
                          enum_c_vals=s2a("1-1", "1-2", "1-3", "1-4"), &
                          enum_i_vals=(/do_skip_11, do_skip_12, do_skip_13, do_skip_14/), &
                          default_i_val=do_skip_13)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXCLUDE_EI", &
                          description="Specifies which kind of Electrostatic interaction to skip.", &
                          usage="EXCLUDE_EI (1-1||1-2||1-3||1-4)", &
                          enum_c_vals=s2a("1-1", "1-2", "1-3", "1-4"), &
                          enum_i_vals=(/do_skip_11, do_skip_12, do_skip_13, do_skip_14/), &
                          default_i_val=do_skip_13)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="AUTOGEN_EXCLUDE_LISTS", &
                          description="When True, the exclude lists are solely based on"// &
                          " the bond data in the topology. The (minimal)"// &
                          " number of bonds between two atoms is used to"// &
                          " determine if the atom pair is added to an"// &
                          " exclusion list. When False, 1-2 exclusion is based"// &
                          " on bonds in the topology, 1-3 exclusion is based"// &
                          " on bonds and bends in the topology, 1-4 exclusion"// &
                          " is based on bonds, bends and dihedrals in the"// &
                          " topology. This implies that a missing dihedral in"// &
                          " the topology will cause the corresponding 1-4 pair"// &
                          " not to be in the exclusion list, in case 1-4"// &
                          " exclusion is requested for VDW or EI interactions.", &
                          usage="AUTOGEN_EXCLUDE_LISTS logical", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="MULTIPLE_UNIT_CELL", &
         description="Specifies the numbers of repetition in space (X, Y, Z) of the defined cell, "// &
         "assuming it as a unit cell. This keyword affects only the coordinates specification. The same keyword "// &
         "in SUBSYS%CELL%MULTIPLE_UNIT_CELL should be modified in order to affect the cell "// &
         "specification.", usage="MULTIPLE_UNIT_CELL 1 1 1", &
         n_var=3, default_i_vals=(/1, 1, 1/), repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MEMORY_PROGRESSION_FACTOR", &
                          description="This keyword is quite technical and should normally not be changed by the user. It "// &
                          "affects the memory allocation during the construction of the topology. It does NOT affect the "// &
                          "memory used once the topology is built.", &
                          n_var=1, default_r_val=1.2_dp, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "DUMP_PDB", &
                                       description="controls the dumping of the PDB at the starting geometry", &
                                       print_level=debug_print_level, filename="dump")
      CALL section_add_subsection(section, print_key)

      CALL keyword_create(keyword, name="CHARGE_OCCUP", &
                          variants=(/"CHARGE_O"/), &
                          description="Write the MM charges to the OCCUP field of the PDB file", &
                          usage="CHARGE_OCCUP logical", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHARGE_BETA", &
                          variants=(/"CHARGE_B"/), &
                          description="Write the MM charges to the BETA field of the PDB file", &
                          usage="CHARGE_BETA logical", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHARGE_EXTENDED", &
                          description="Write the MM charges to the very last field of the PDB file (starting from column 81)", &
                          usage="CHARGE_EXTENDED logical", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "DUMP_PSF", &
                                       description="controls the dumping of the PSF connectivity", &
                                       print_level=debug_print_level, filename="dump")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      NULLIFY (subsection)
      CALL create_exclude_list_section(subsection, "EXCLUDE_VDW_LIST")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_exclude_list_section(subsection, "EXCLUDE_EI_LIST")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_center_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_generate_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_molset_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_topology_section

! **************************************************************************************************
!> \brief Setup a list of fine exclusion elements
!> \param section the section to create
!> \param header ...
!> \author Teodoro Laino [tlaino] - 12.2009
! **************************************************************************************************
   SUBROUTINE create_exclude_list_section(section, header)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(LEN=*), INTENT(IN)                       :: header

      CHARACTER(len=*), PARAMETER :: routineN = 'create_exclude_list_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",1707)
      NULLIFY (keyword)
      CALL section_create(section, TRIM(header), &
                          description="Speficy bonds (via atom kinds) for fine tuning of 1-2 "// &
                          "exclusion lists. If this section is not present the 1-2 exclusion is "// &
                          "applied to all bond kinds. When this section is present the 1-2 exclusion "// &
                          "is applied ONLY to the bonds defined herein. This section allows ONLY fine tuning of 1-2 "// &
                          "interactions. ", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="BOND", &
                          description="Specify the atom kinds involved in the bond for which 1-2 exclusion holds.", &
                          usage="BOND {KIND1} {KIND2}", type_of_var=char_t, &
                          n_var=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_exclude_list_section

! **************************************************************************************************
!> \brief Specify keywords used to center molecule in the box
!> \param section the section to create
!> \author Teodoro Laino [tlaino] - University of Zurich - 06.2009
! **************************************************************************************************
   SUBROUTINE create_center_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_center_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",1738)
      NULLIFY (keyword)
      CALL section_create(section, "CENTER_COORDINATES", &
                          description="Allows centering the coordinates of the system in the box. "// &
                          "The centering point can be defined by the user.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="Controls the activation of the centering method", &
                          usage="&CENTER_COORDINATES T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CENTER_POINT", &
                          description="Specify the point used for centering the coordinates. Default is to "// &
                          "center the system in cell/2. ", type_of_var=real_t, n_var=3, &
                          repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_center_section

! **************************************************************************************************
!> \brief Specify keywords used to setup several molecules with few connectivity files
!> \param section the section to create
!> \author Teodoro Laino [tlaino] - University of Zurich - 08.2008
! **************************************************************************************************
   SUBROUTINE create_molset_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_molset_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection, subsubsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",1775)
      NULLIFY (keyword, subsection, subsubsection)
      CALL section_create(section, name="MOL_SET", &
                          description="Specify the connectivity of a full system specifying the connectivity"// &
                          " of the fragments of the system.", &
                          n_keywords=2, n_subsections=0, repeats=.FALSE.)

      ! MOLECULES
      CALL section_create(subsection, name="MOLECULE", &
                          description="Specify information about the connectivity of single molecules", &
                          n_keywords=2, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="NMOL", &
                          description="number of molecules ", &
                          usage="NMOL {integer}", default_i_val=1)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL connectivity_framework(subsection, do_conn_psf)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! MERGE MOLECULES
      CALL section_create(subsection, name="MERGE_MOLECULES", &
                          description="Enables the creation of connecting bridges (bonds, angles, torsions, impropers)"// &
                          " between the two or more molecules defined with independent connectivity.", &
                          n_keywords=2, n_subsections=0, repeats=.FALSE.)

      CALL section_create(subsubsection, name="bonds", &
                          description="Defines new bonds", n_keywords=2, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Two integer indexes per line defining the new bond."// &
                          " Indexes must be relative to the full system and not to the single molecules", &
                          repeats=.TRUE., &
                          usage="{Integer} {Integer}", type_of_var=integer_t, n_var=2)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)

      CALL section_create(subsubsection, name="angles", &
                          description="Defines new angles", n_keywords=2, n_subsections=0, &
                          repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Three integer indexes per line defining the new angle"// &
                          " Indexes must be relative to the full system and not to the single molecules", repeats=.TRUE., &
                          usage="{Integer} {Integer} {Integer}", type_of_var=integer_t, n_var=3)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)

      CALL section_create(subsubsection, name="torsions", &
                          description="Defines new torsions", n_keywords=2, n_subsections=0, &
                          repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Four integer indexes per line defining the new torsion"// &
                          " Indexes must be relative to the full system and not to the single molecules", repeats=.TRUE., &
                          usage="{Integer} {Integer} {Integer} {Integer}", type_of_var=integer_t, n_var=4)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)

      CALL section_create(subsubsection, name="impropers", &
                          description="Defines new impropers", n_keywords=2, n_subsections=0, &
                          repeats=.FALSE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="Four integer indexes per line defining the new improper"// &
                          " Indexes must be relative to the full system and not to the single molecules", repeats=.TRUE., &
                          usage="{Integer} {Integer} {Integer} {Integer}", type_of_var=integer_t, n_var=4)
      CALL section_add_keyword(subsubsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, subsubsection)
      CALL section_release(subsubsection)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_molset_section

! **************************************************************************************************
!> \brief Specify keywords used to generate connectivity
!> \param section the section to create
!> \author Teodoro Laino [tlaino] - University of Zurich - 08.2008
! **************************************************************************************************
   SUBROUTINE create_generate_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_generate_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",1870)
      NULLIFY (keyword, subsection)
      CALL section_create(section, name="GENERATE", &
                          description="Setup of keywords controlling the generation of the connectivity", &
                          n_keywords=2, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="REORDER", &
                          description="Reorder a list of atomic coordinates into order so it can be packed correctly.", &
                          usage="REORDER <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CREATE_MOLECULES", &
                          description="Create molecules names and definition. Can be used to override the "// &
                          " molecules specifications of a possible input connectivity or to create molecules"// &
                          " specifications for file types as XYZ, missing of molecules definitions.", &
                          usage="CREATE_MOLECULES <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BONDPARM", &
                          description="Used in conjunction with BONDPARM_FACTOR to "// &
                          "help determine wheather there is bonding "// &
                          "between two atoms based on a distance criteria. "// &
                          "Can use covalent radii information or VDW radii information", &
                          usage="BONDPARM (COVALENT||VDW)", &
                          enum_c_vals=s2a("COVALENT", "VDW"), &
                          enum_i_vals=(/do_bondparm_covalent, do_bondparm_vdw/), &
                          default_i_val=do_bondparm_covalent)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BONDPARM_FACTOR", &
                          description="Used in conjunction with BONDPARM to help "// &
                          "determine wheather there is bonding between "// &
                          "two atoms based on a distance criteria.", &
                          usage="bondparm_factor {real}", default_r_val=1.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="BONDLENGTH_MAX", &
                          description="Maximum distance to generate neighbor lists to build connectivity", &
                          usage="BONDLENGTH_MAX <real>", default_r_val=cp_unit_to_cp2k(value=3.0_dp, &
                                                                                       unit_str="angstrom"), unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="BONDLENGTH_MIN", &
                          description="Minimum distance to generate neighbor lists to build connectivity", &
                          usage="BONDLENGTH_MIN <real>", default_r_val=cp_unit_to_cp2k(value=0.01_dp, &
                                                                                       unit_str="angstrom"), unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! BONDS
      CALL section_create(subsection, name="BOND", &
                          description="Section used to add/remove  bonds in the connectivity."// &
                          " Useful for systems with a complex connectivity, difficult to find out automatically.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the bond", &
                          usage="&BOND (ADD|REMOVE)", &
                          enum_c_vals=s2a("ADD", "REMOVE"), &
                          enum_i_vals=(/do_add, do_remove/), &
                          default_i_val=do_add)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specifies two atomic index united by a covalent bond", &
                          usage="ATOMS {integer} {integer}", type_of_var=integer_t, n_var=2, &
                          repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! ANGLES
      CALL section_create(subsection, name="ANGLE", &
                          description="Section used to add/remove angles in the connectivity."// &
                          " Useful for systems with a complex connectivity, difficult to find out automatically.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the bond", &
                          usage="&ANGLE (ADD|REMOVE)", &
                          enum_c_vals=s2a("ADD", "REMOVE"), &
                          enum_i_vals=(/do_add, do_remove/), &
                          default_i_val=do_add)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specifies two atomic index united by a covalent bond", &
                          usage="ATOMS {integer} {integer} {integer} ", type_of_var=integer_t, n_var=3, &
                          repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! TORSIONS
      CALL section_create(subsection, name="TORSION", &
                          description="Section used to add/remove torsion in the connectivity."// &
                          " Useful for systems with a complex connectivity, difficult to find out automatically.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the bond", &
                          usage="&TORSION (ADD|REMOVE)", &
                          enum_c_vals=s2a("ADD", "REMOVE"), &
                          enum_i_vals=(/do_add, do_remove/), &
                          default_i_val=do_add)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specifies two atomic index united by a covalent bond", &
                          usage="ATOMS {integer} {integer} {integer} {integer} ", type_of_var=integer_t, n_var=4, &
                          repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! IMPROPERS
      CALL section_create(subsection, name="IMPROPER", &
                          description="Section used to add/remove improper in the connectivity."// &
                          " Useful for systems with a complex connectivity, difficult to find out automatically.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the bond", &
                          usage="&IMPROPER (ADD|REMOVE)", &
                          enum_c_vals=s2a("ADD", "REMOVE"), &
                          enum_i_vals=(/do_add, do_remove/), &
                          default_i_val=do_add)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specifies two atomic index united by a covalent bond", &
                          usage="ATOMS {integer} {integer} {integer} {integer} ", type_of_var=integer_t, n_var=4, &
                          repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! ISOLATED ATOMS
      CALL section_create(subsection, name="ISOLATED_ATOMS", &
                          description=" This section specifies the  atoms that one considers isolated. Useful when present "// &
                          " ions in solution.", n_keywords=1, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="LIST", &
                          description="Specifies a list of atomic indexes of the isolated ion", &
                          usage="LIST {integer}", type_of_var=integer_t, n_var=-1, &
                          repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! Neighbor lists keys and printing handling the construction of NL for the connectivity
      CALL create_neighbor_lists_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_gen_print_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_generate_section

! **************************************************************************************************
!> \brief Create the print gen section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_gen_print_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_gen_print_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",2064)
      CALL section_create(section, name="print", &
                          description="Section of possible print options in GENERATE code.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "NEIGHBOR_LISTS", &
                                       description="Activates the printing of the neighbor lists used"// &
                                       " for generating the connectivity.", print_level=high_print_level, &
                                       filename="", unit_str="angstrom")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "SUBCELL", &
                                       description="Activates the printing of the subcells used for the"// &
                                       "generation of neighbor lists for connectivity.", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_gen_print_section

! **************************************************************************************************
!> \brief Specify keywords used to define connectivity
!> \param section the section to create
!> \param default ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE connectivity_framework(section, default)
      TYPE(section_type), POINTER                        :: section
      INTEGER, INTENT(IN)                                :: default

      CHARACTER(len=*), PARAMETER :: routineN = 'connectivity_framework', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",2101)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="CONN_FILE_NAME", &
                          variants=(/"CONN_FILE"/), &
                          description="Specifies the filename that contains the molecular connectivity.", &
                          usage="CONN_FILE_NAME <FILENAME>", type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CONN_FILE_FORMAT", &
                          variants=(/"CONNECTIVITY"/), &
                          description="Ways to determine and generate a molecules. "// &
                          "Default is to use GENERATE", &
                          usage="CONN_FILE_FORMAT (PSF|UPSF|MOL_SET|GENERATE|OFF|G87|G96|AMBER|USER)", &
                          enum_c_vals=s2a("PSF", "UPSF", "MOL_SET", "GENERATE", "OFF", "G87", "G96", "AMBER", "USER"), &
                          enum_i_vals=(/do_conn_psf, &
                                        do_conn_psf_u, &
                                        do_conn_mol_set, &
                                        do_conn_generate, &
                                        do_conn_off, &
                                        do_conn_g87, &
                                        do_conn_g96, &
                                        do_conn_amb7, &
                                        do_conn_user/), &
                          enum_desc=s2a("Use  a PSF file to determine the connectivity."// &
                                        " (support standard CHARMM/XPLOR and EXT CHARMM)", &
                                        "Read a PSF file in an unformatted way (useful for not so standard PSF).", &
                                        "Use multiple PSF (for now...) files to generate the whole sytem.", &
                                        "Use a simple distance criteria. (Look at keyword BONDPARM)", &
                                        "Do not generate molecules. (e.g. for QS or ill defined systems)", &
                                        "Use GROMOS G87 topology file.", &
                                        "Use GROMOS G96 topology file.", &
                                        "Use AMBER topology file for reading connectivity (compatible starting from AMBER V.7)", &
                                        "Allows the definition of molecules and residues based on the 5th and 6th column of "// &
                                        "the COORD section. This option can be handy for the definition of molecules with QS "// &
                                        "or to save memory in the case of very large systems (use PARA_RES off)."), &
                          default_i_val=default)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE connectivity_framework

! **************************************************************************************************
!> \brief      Create CP2K input section for the DFT+U method parameters
!> \param section ...
!> \date       01.11.2007
!> \author     Matthias Krack (MK)
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE create_dft_plus_u_section(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_dft_plus_u_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_subsys.F",2159)

      CALL section_create(section, &
                          name="DFT_PLUS_U", &
                          description="Define the parameters for a DFT+U run", &
                          n_keywords=3, &
                          n_subsections=1, &
                          repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, &
                          name="_SECTION_PARAMETERS_", &
                          description="Controls the activation of the DFT+U section", &
                          usage="&DFT_PLUS_U ON", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="L", &
                          description="Angular momentum quantum number of the"// &
                          "orbitals to which the correction is applied", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=integer_t, &
                          default_i_val=-1, &
                          usage="L 2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="U_MINUS_J", &
                          description="Effective parameter U(eff) = U - J", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.0_dp, &
                          unit_str="au_e", &
                          usage="U_MINUS_J [eV] 1.4")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="U_RAMPING", &
                          description="Increase the effective U parameter stepwise using the specified "// &
                          "increment until the target value given by U_MINUS_J is reached.", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.0_dp, &
                          unit_str="au_e", &
                          usage="U_RAMPING [eV] 0.1")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="EPS_U_RAMPING", &
                          description="Threshold value (SCF convergence) for incrementing the effective "// &
                          "U value when U ramping is active.", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=1.0E-5_dp, &
                          usage="EPS_U_RAMPING 1.0E-6")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="INIT_U_RAMPING_EACH_SCF", &
                          description="Set the initial U ramping value to zero before each wavefunction optimisation. "// &
                          "The default is to apply U ramping only for the initial wavefunction optimisation.", &
                          repeats=.FALSE., &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE., &
                          usage="INIT_U_RAMPING_EACH_SCF on")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)

      CALL section_create(subsection, &
                          name="ENFORCE_OCCUPATION", &
                          description="Enforce and control a special (initial) orbital occupation. "// &
                          "Note, this feature works only for the methods MULLIKEN and LOWDIN. "// &
                          "It should only be used to prepare an initial configuration. An "// &
                          "inadequate parameter choice can easily inhibit SCF convergence.", &
                          n_keywords=5, &
                          n_subsections=0, &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, &
                          name="_SECTION_PARAMETERS_", &
                          description="Controls the activation of the ENFORCE_OCCUPATION section", &
                          usage="&ENFORCE_OCCUPATION ON", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="ORBITALS", &
                          variants=(/"M"/), &
                          description="Select orbitals and occupation order. An input of 1 to 2*L+1 integer values in "// &
                          "the range -L to L defining the M values of the spherical orbitals is expected.", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          type_of_var=integer_t, &
                          default_i_val=0, &
                          usage="ORBITALS 0 +1 -1")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="EPS_SCF", &
                          description="The occupation constraint is enforced until this threshold value "// &
                          "for the SCF convergence criterion is reached", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=1.0E30_dp, &
                          usage="EPS_SCF 0.001")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="MAX_SCF", &
                          description="The occupation constraint is applied for this number of initial SCF iterations", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=integer_t, &
                          default_i_val=-1, &
                          usage="MAX_SCF 5")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="SMEAR", &
                          description="The occupation constraint is applied with smearing", &
                          repeats=.FALSE., &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE., &
                          usage="SMEAR ON")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_dft_plus_u_section

END MODULE input_cp2k_subsys
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      - taken out of input_cp2k_motion
!> \author Ole Schuett
! **************************************************************************************************
MODULE input_cp2k_motion_print
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              high_print_level,&
                                              low_print_level,&
                                              silent_print_level
   USE input_constants,                 ONLY: dump_atomic,&
                                              dump_dcd,&
                                              dump_dcd_aligned_cell,&
                                              dump_pdb,&
                                              dump_xmol
   USE input_cp2k_subsys,               ONLY: create_structure_data_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE string_utilities,                ONLY: newline,&
                                              s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_motion_print'

   PUBLIC :: create_motion_print_section, add_format_keyword

CONTAINS

! **************************************************************************************************
!> \brief creates the motion%print section
!> \param section the section to be created
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_motion_print_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_motion_print_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      NULLIFY (keyword, section, print_key)

      CALL section_create(section, name="print", &
                          description="Controls the printing properties during an MD run", &
                          n_keywords=0, n_subsections=1, repeats=.TRUE.)

      CALL cp_print_key_section_create(print_key, "TRAJECTORY", &
                                       description="Controls the output of the trajectory", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="", unit_str="angstrom")
      CALL add_format_keyword(keyword, print_key, pos=.TRUE., &
                              description="Specifies the format of the output file for the trajectory.")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key, "SHELL_TRAJECTORY", &
         description="Controls the output of the trajectory of shells when the shell-model is used ", &
         print_level=high_print_level, common_iter_levels=1, &
         filename="", unit_str="angstrom")
      CALL add_format_keyword(keyword, print_key, pos=.TRUE., &
                              description="Specifies the format of the output file for the trajectory of shells.")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CORE_TRAJECTORY", &
                                       description="Controls the output of the trajectory of cores when the shell-model is used ", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="", unit_str="angstrom")
      CALL add_format_keyword(keyword, print_key, pos=.TRUE., &
                              description="Specifies the format of the output file for the trajectory of cores.")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CELL", &
                                       description="Controls the output of the simulation cell. "// &
                                       "For later analysis of the trajectory it is recommendable that the "// &
                                       "frequency of printing is the same as the one used for the trajectory file.", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "VELOCITIES", &
                                       description="Controls the output of the velocities."//newline// &
                                       "The default unit for velocities is bohr/au_time. The au_time is derived from the "// &
                                       "hbar value (1.054e-34 J*sec) and the value of the hartree unit of energy "// &
                                       "(27.21 eV or 4.359e-18 J) as hbar/Ehartree = 2.42e-17 sec = 0.02042 fs. "// &
                                       "Having an atom with a mass m in AMU the kinetic energy 1/2mv^2 will be obtained "// &
                                       "in Hartree (i.e. au) multiplying by 911.447 .", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="", unit_str="bohr*au_t^-1")
      CALL add_format_keyword(keyword, print_key, pos=.FALSE., &
                              description="Specifies the format of the output file for the velocities.")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "SHELL_VELOCITIES", &
                                       description="Controls the output of the velocities of shells when the shell model is used", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="", unit_str="bohr*au_t^-1")
      CALL add_format_keyword(keyword, print_key, pos=.FALSE., &
                              description="Specifies the format of the output file for the velocities of shells.")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CORE_VELOCITIES", &
                                       description="controls the output of the velocities of cores when the shell model is used", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="", unit_str="bohr*au_t^-1")
      CALL add_format_keyword(keyword, print_key, pos=.FALSE., &
                              description="Specifies the format of the output file for the velocities of cores.")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL create_structure_data_section(print_key)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key, "FORCE_MIXING_LABELS", &
         description="Controls the output of the force mixing (FORCE_EVAL&QMMM&FORCE_MIXING) labels", &
         print_level=high_print_level, common_iter_levels=1, &
         filename="")
      CALL add_format_keyword(keyword, print_key, pos=.FALSE., &
                              description="Specifies the format of the output file for the force mixing labels.")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "FORCES", &
                                       description="Controls the output of the forces", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="", unit_str="hartree*bohr^-1")
      CALL add_format_keyword(keyword, print_key, pos=.FALSE., &
                              description="Specifies the format of the output file for the forces.")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "SHELL_FORCES", &
                                       description="controls the output of the forces on shells when shell-model is used", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="", unit_str="hartree*bohr^-1")
      CALL add_format_keyword(keyword, print_key, pos=.FALSE., &
                              description="Specifies the format of the output file for the forces on shells.")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CORE_FORCES", &
                                       description="controls the output of the forces on cores when shell-model is used", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="", unit_str="hartree*bohr^-1")
      CALL add_format_keyword(keyword, print_key, pos=.FALSE., &
                              description="Specifies the format of the output file for the forces on cores.")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "MIXED_ENERGIES", &
                                       description="Controls the output of the energies of the two"// &
                                       "regular FORCE_EVALS in the MIXED method"// &
                                       "printed is step,time,Etot,E_F1,E_F2,CONS_QNT", &
                                       print_level=low_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "STRESS", &
                                       description="Controls the output of the stress tensor", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RESTART", &
                                       description="Controls the dumping of the restart file during runs. "// &
                                       "By default keeps a short history of three restarts. See also RESTART_HISTORY", &
                                       each_iter_names=s2a("MD"), each_iter_values=(/20/), &
                                       print_level=silent_print_level, common_iter_levels=1, &
                                       add_last=add_last_numeric, filename="")

      CALL keyword_create(keyword, name="BACKUP_COPIES", &
                          description="Specifies the maximum number of backup copies.", &
                          usage="BACKUP_COPIES {int}", &
                          default_i_val=3)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SPLIT_RESTART_FILE", &
                          description="If specified selected input sections, which are growing with the "// &
                          "number of atoms in the system, are written to another restart file "// &
                          "in binary format instead of the default restart file in human "// &
                          "readable ASCII format. This split of the restart file may "// &
                          "provide significant memory savings and an accelerated I/O for "// &
                          "systems with a very large number of atoms", &
                          usage="SPLIT_RESTART_FILE yes", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RESTART_HISTORY", &
                                       description="Dumps unique restart files during the run keeping all of them."// &
                                       "Most useful if recovery is needed at a later point.", &
                                       print_level=low_print_level, common_iter_levels=0, &
                                       each_iter_names=s2a("MD", "GEO_OPT", "ROT_OPT"), each_iter_values=(/500, 500, 500/), &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "TRANSLATION_VECTOR", &
                                       description="Dumps the translation vector applied along an MD (if any). Useful"// &
                                       " for postprocessing of QMMM trajectories in which the QM fragment is continuously"// &
                                       " centered in the QM box", &
                                       print_level=high_print_level, common_iter_levels=1, &
                                       filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_motion_print_section

! **************************************************************************************************
!> \brief creates the FORMAT keyword
!> \param keyword ...
!> \param section will contain the pint section
!> \param pos ...
!> \param description ...
!> \author Teodoro Laino 10.2008 [tlaino]
! **************************************************************************************************
   SUBROUTINE add_format_keyword(keyword, section, pos, description)
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      LOGICAL, INTENT(IN)                                :: pos
      CHARACTER(LEN=*), INTENT(IN)                       :: description

      CHARACTER(len=*), PARAMETER :: routineN = 'add_format_keyword', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input_cp2k_motion_print.F",258)
      IF(.NOT.(.NOT. ASSOCIATED(keyword)))CALL cp__a("input_cp2k_motion_print.F",259)

      IF (pos) THEN

         CALL keyword_create( &
            keyword, name="FORMAT", &
            description=description, usage="FORMAT (ATOMIC|DCD|PDB|XMOL|XYZ)", &
            default_i_val=dump_xmol, &
            enum_c_vals=s2a("ATOMIC", "DCD", "DCD_ALIGNED_CELL", "PDB", "XMOL", "XYZ"), &
            enum_i_vals=(/dump_atomic, dump_dcd, dump_dcd_aligned_cell, dump_pdb, dump_xmol, dump_xmol/), &
            enum_desc=s2a("Write only the coordinates X,Y,Z without element symbols to a formatted file", &
                          "Write the coordinates (no element labels) and the cell information to a binary file", &
                          "Like DCD, but the dumped coordinates refer to an aligned cell following the common convention: "// &
                          "the cell vector <b>a</b> is aligned with the <i>x</i> axis and the cell vector <b>b</b> lies in "// &
                          "the <i>xy</i> plane. This allows the reconstruction of scaled coordinates from the DCD data only.", &
                          "Write the atomic information in PDB format to a formatted file", &
                          "Mostly known as XYZ format, provides in a formatted file: element_symbol X Y Z", &
                          "Alias name for XMOL"))
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL keyword_create(keyword, name="CHARGE_OCCUP", &
                             variants=(/"CHARGE_O"/), &
                             description="Write the MM charges to the OCCUP field of the PDB file", &
                             usage="CHARGE_OCCUP logical", &
                             default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL keyword_create(keyword, name="CHARGE_BETA", &
                             variants=(/"CHARGE_B"/), &
                             description="Write the MM charges to the BETA field of the PDB file", &
                             usage="CHARGE_BETA logical", &
                             default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

         CALL keyword_create(keyword, name="CHARGE_EXTENDED", &
                             description="Write the MM charges to the very last field of the PDB file (starting from column 81)", &
                             usage="CHARGE_EXTENDED logical", &
                             default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

      ELSE

         CALL keyword_create(keyword, name="FORMAT", &
                             description=description, usage="FORMAT (ATOMIC|DCD|XMOL|XYZ)", &
                             default_i_val=dump_xmol, &
                             enum_c_vals=s2a("ATOMIC", "DCD", "XMOL", "XYZ"), &
                             enum_i_vals=(/dump_atomic, dump_dcd, dump_xmol, dump_xmol/), &
                             enum_desc=s2a("Write only the coordinates X,Y,Z without element symbols to a formatted file", &
                                           "Write the coordinates (no element labels) and the cell information to a binary file", &
                                           "Mostly known as XYZ format, provides in a formatted file: element_symbol X Y Z", &
                                           "Alias name for XMOL"))
         CALL section_add_keyword(section, keyword)
         CALL keyword_release(keyword)

      END IF

   END SUBROUTINE add_format_keyword

END MODULE input_cp2k_motion_print
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief input section for NEGF based quantum transport calculations
!>        (integration with the quantum transport code OMEN)
!>
!> \par History
!>       07.2013 created [Hossein Bani-Hashemian]
!> \author Hossein Bani-Hashemian
! **************************************************************************************************
MODULE input_cp2k_transport
   USE bibliography,                    ONLY: Bruck2014
   USE input_constants,                 ONLY: &
        injmethod_beyn, injmethod_evp, injsign_negative, injsign_positive, linsolver_banded, &
        linsolver_full, linsolver_mumps, linsolver_splitsolve, linsolver_superlu, &
        rlaxisint_GaussChebyshev, rlaxisint_readfromfile, rlaxisint_trapezoidal, &
        transport_diagscalapack, transport_localscf, transport_negf, transport_transmission, &
        transport_writeout
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_transport'

   PUBLIC :: create_transport_section

CONTAINS

! **************************************************************************************************
!> \brief creates the TRANSPORT section
!> \param[inout] section the section to be created
! **************************************************************************************************
   SUBROUTINE create_transport_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_transport_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_transport.F",56)
      CALL section_create(section, "TRANSPORT", &
                          description="Specifies the parameters for transport, sets parameters for the OMEN code, "// &
                          "see also http://www.nano-tcad.ethz.ch/ ", &
                          citations=(/Bruck2014/), &
                          n_keywords=22, n_subsections=2, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create( &
         keyword, name="TRANSPORT_METHOD", &
         description="Preferred method for transport calculations.", &
         usage="TRANSPORT_METHOD <method>", &
         default_i_val=transport_negf, &
         enum_c_vals=s2a("DIAG", "TRANSPORT", "LOCAL_SCF", "TRANSMISSION", "WRITEOUT"), &
         enum_desc=s2a("standard diagonalization using ScaLapack routines", &
                       "self-consistent CP2K and OMEN transport calculations", &
                       "CP2K valence Hamiltonian + OMEN self-consistent calculations on conduction electrons", &
                       "self-consistent transmission calculations without applied bias voltage", &
                       "write out the Hamiltonian and overlap matrices"), &
         enum_i_vals=(/transport_diagscalapack, transport_negf, transport_localscf, &
                       transport_transmission, transport_writeout/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_ABSCISSAE", &
                          description="The number of abscissae per integration interval on the real axis.", &
                          usage="N_ABSCISSAE <integer>", default_i_val=64)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_KPOINTS", &
                          description="The number of k points for determination of the singularities.", &
                          usage="N_KPOINTS <integer>", default_i_val=64)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUM_INTERVAL", &
                          description="Max number of energy points per small interval.", &
                          usage="NUM_INTERVAL <integer>", default_i_val=10)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TASKS_PER_POINT", &
                          description="Minimum number of tasks per energy point.", &
                          usage="TASKS_PER_POINT <integer>", default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GPUS_PER_POINT", &
                          description="Number of GPUs per energy point for SplitSolve. Needs to be a power of two", &
                          usage="GPUS_PER_POINT <integer>", default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COLZERO_THRESHOLD", &
                          description="The smallest number that is not zero in the full diagonalization part.", &
                          usage="COLZERO_THRESHOLD <real>", default_r_val=1.0E-12_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_LIMIT", &
                          description="The smallest eigenvalue that is kept.", &
                          usage="EPS_LIMIT <real>", default_r_val=1.0E-4_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_LIMIT_CC", &
                          description="The smallest eigenvalue that is kept on the complex contour.", &
                          usage="EPS_LIMIT_CC <real>", default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_DECAY", &
                          description="The smallest imaginary part that a decaying eigenvalue may have not to be "// &
                          "considered as propagating.", &
                          usage="EPS_DECAY <real>", default_r_val=1.0E-4_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_SINGULARITY_CURVATURES", &
                          description="Filter for degenerate bands in the bandstructure.", &
                          usage="EPS_SINGULARITY_CURVATURES <real>", default_r_val=1.0E-12_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_MU", &
                          description="Accuracy to which the Fermi level should be determined.", &
                          usage="EPS_MU <real>", default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_EIGVAL_DEGEN", &
                          description="Filter for degenerate bands in the injection vector.", &
                          usage="EPS_EIGVAL_DEGEN <real>", default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_FERMI", &
                          description="Cutoff for the tail of the Fermi function.", &
                          usage="EPS_FERMI <real>", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ENERGY_INTERVAL", &
                          description="Average distance for big intervals in energy vector.", &
                          usage="ENERGY_INTERVAL <real>", default_r_val=1.0E-3_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIN_INTERVAL", &
                          description="Smallest enery distance in energy vector.", &
                          usage="MIN_INTERVAL <real>", default_r_val=1.0E-4_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TEMPERATURE", &
                          description="Temperature.", &
                          usage="TEMPERATURE [K] 300.0", default_r_val=300.0_dp, unit_str="K")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CSR_SCREENING", &
                          description="Whether distance screening should be applied to improve sparsity of CSR matrices.", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="LINEAR_SOLVER", &
         description="Solver to solve the linear system of equations.", &
         usage="LINEAR_SOLVER <solver>", &
         default_i_val=linsolver_superlu, &
         enum_c_vals=s2a("SplitSolve", "SuperLU", "MUMPS", "Full", "Banded"), &
         enum_i_vals=(/linsolver_splitsolve, linsolver_superlu, linsolver_mumps, linsolver_full, linsolver_banded/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INJECTION_METHOD", &
                          description="Method to solve the eigenvalue problem for the open boundary conditions.", &
                          usage="INJECTION_METHOD <method>", &
                          default_i_val=injmethod_beyn, &
                          enum_c_vals=s2a("EVP", "BEYN"), &
                          enum_desc=s2a("Full eigenvalue solver.", &
                                        "Beyn eigenvalue solver."), &
                          enum_i_vals=(/injmethod_evp, injmethod_beyn/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="CUTOUT", &
         description="Degrees of freedom at the beginning and the end of the structure (in the number of atoms) "// &
         "where the density should not be changed.", &
         usage="CUTOUT <integer> <integer>", &
         n_var=2, default_i_vals=(/0, 0/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REAL_AXIS_INTEGRATION_METHOD", &
                          description="Integration method for the real axis.", &
                          usage="REAL_AXIS_INTEGRATION_METHOD <method>", &
                          default_i_val=rlaxisint_GaussChebyshev, &
                          enum_c_vals=s2a("Gauss_Chebyshev", "Trapezoidal_rule", "Read"), &
                          enum_desc=s2a("Gauss-Chebyshev integration between singularity points.", &
                                        "Trapezoidal rule on the total range.", &
                                        "Read integration points from a file (named E.dat)."), &
                          enum_i_vals=(/rlaxisint_GaussChebyshev, rlaxisint_trapezoidal, rlaxisint_readfromfile/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)

      CALL create_contact_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_beyn_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_transport_section

! **************************************************************************************************
!> \brief Creates the section for creating contacts.
!> \param[inout] section the section to be created
! **************************************************************************************************
   SUBROUTINE create_contact_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_contact_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_transport.F",250)
      CALL section_create(section, name="CONTACT", &
                          description="Parameters for defining device contacts.", &
                          n_keywords=5, n_subsections=0, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="BANDWIDTH", &
                          description="The number of neighboring unit cells that one unit cell interacts with.", &
                          usage="BANDWIDTH <integer>", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="START", &
                          description="Index of the first atom in the contact unit cell. Set to 0 to define the contact "// &
                          "unit cell as the first/last N_ATOMS of the structure (after cutout)", &
                          usage="START <integer>", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_ATOMS", &
                          description="Number of atoms in the contact unit cell.", &
                          usage="N_ATOMS <integer>", default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INJECTION_SIGN", &
                          description="Contact unit cell interacts with unit cells to the right (positive) or "// &
                          "to the left (negative).", &
                          usage="INJECTION_SIGN <integer>", &
                          default_i_val=injsign_positive, &
                          enum_c_vals=s2a("POSITIVE", "NEGATIVE"), &
                          enum_desc=s2a("When the contact unit cell is at the upper left corner of the Hamiltonian.", &
                                        "When the contact unit cell is at the lower right corner of the Hamiltonian."), &
                          enum_i_vals=(/injsign_positive, injsign_negative/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INJECTING_CONTACT", &
                          description="whether or not the contact can inject electrons.", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_contact_section

! **************************************************************************************************
!> \brief Creates the section for the Beyn eigenvalue solver.
!> \param[inout] section the section to be created
! **************************************************************************************************
   SUBROUTINE create_beyn_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_beyn_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_transport.F",308)
      CALL section_create(section, name="BEYN", &
                          description="Parameters for the Beyn eigenvalue solver.", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="N_RAND", &
                          description="Number of random vectors as a fraction of the size of the unit cell.", &
                          usage="N_RAND <real>", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_RAND_CC", &
                          description="Number of random vectors as a fraction of the size of the unit cell "// &
                          "for the complex contour.", &
                          usage="N_RAND_CC <real>", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SVD_CUTOFF", &
                          description="Cutoff for the singular values in the Beyn solver.", &
                          usage="SVD_CUTOFF <real>", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_POINTS_BEYN", &
                          description="Number of integration points per circle in the Beyn solver.", &
                          usage="N_POINTS_BEYN <integer>", default_i_val=32)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ONE_CIRCLE", &
                          description="Set to .TRUE. if only one circle instead of two should be used in the Beyn solver.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_beyn_section

END MODULE input_cp2k_transport
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief function that builds the distribution section of the input
!> \par History
!>      04.2007 created
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE input_cp2k_distribution

   USE input_constants,                 ONLY: model_block_count,&
                                              model_block_lmax,&
                                              model_block_surface
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_create,&
                                              section_type
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_distribution'

   PUBLIC :: create_distribution_section

CONTAINS

! **************************************************************************************************
!> \brief Creates the distribution section
!> \param section the section to create
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE create_distribution_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_distribution_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_distribution.F",49)
      CALL section_create(section, name="DISTRIBUTION", &
                          description="can be used used to tune the parallel distribution of the data", &
                          n_keywords=2, n_subsections=2, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="COST_MODEL", &
                          description="The cost model that needs to be minimized ", &
                          usage="COST_MODEL BLOCK_COUNT", &
                          enum_c_vals=s2a("BLOCK_COUNT", "BLOCK_SURFACE", "BLOCK_LMAX"), &
                          enum_i_vals=(/model_block_count, model_block_surface, model_block_lmax/), &
                          enum_desc=s2a("the number of blocks", &
                                        "the number of blocks weighted by the number elements per block", &
                                        "the number of blocks weighted by the sum of the lmax"), &
                          default_i_val=model_block_count)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="2D_MOLECULAR_DISTRIBUTION", &
                          description="Distribute the atoms so that atoms belonging to a given molecule"// &
                          " are on the same CPU for the 2D distribution. This might give rise to a"// &
                          " worse distribution but reduces memory needs of finding the optimal distribution.", &
                          usage="2D_MOLECULAR_DISTRIBUTION TRUE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SKIP_OPTIMIZATION", &
                          description="Do not optimize the distribution, go for something very simple."// &
                          " Might be useful if the optimization, which scales quadratically in system size, is too expensive.", &
                          usage="SKIP_OPTIMIZATION TRUE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BASIC_OPTIMIZATION", &
                          description="Creates a distribution based on a few heuristics using only minimal memory "// &
                          "and CPU time.", &
                          usage="BASIC_OPTIMIZATION TRUE", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BASIC_SPATIAL_OPTIMIZATION", &
                          description="Creates a distribution with spatial info, using only minimal memory "// &
                          "and CPU time.", &
                          usage="BASIC_SPATIAL_OPTIMIZATION TRUE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BASIC_CLUSTER_OPTIMIZATION", &
                          description="Creates a distribution with spatial info, using recursively KMEANS clustering. ", &
                          usage="BASIC_CLUSTER_OPTIMIZATION TRUE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SYMMETRIC", &
                          description="Take the symmetry of the distribution_2d into account.", &
                          usage="SYMMETRIC TRUE", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_distribution_section

END MODULE input_cp2k_distribution
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_fb_input
   USE bibliography,                    ONLY: Rayson2009
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_create,&
                                              section_type
   USE input_val_types,                 ONLY: logical_t,&
                                              real_t
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_input'

   PUBLIC :: create_filtermatrix_section

CONTAINS

! **************************************************************************************************
!> \brief Input section for filter matrix diagonalisation method
!> \param section : section to be created
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE create_filtermatrix_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_filtermatrix_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("qs_fb_input.F",42)

      CALL section_create(section, "FILTER_MATRIX", &
                          description=" ", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, &
                          name="FILTER_TEMPERATURE", &
                          description="Temperature used for the filter function used "// &
                          "to construct the filter matrix.", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=10000.0_dp, &
                                                        unit_str="K"), &
                          unit_str="K", &
                          usage="FILTER_TEMPERATURE [K] 10000", &
                          citations=(/Rayson2009/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="AUTO_CUTOFF_SCALE", &
                          description="Scalar constant multiplied to maximum orbital "// &
                          "size of each atom, used for automatically "// &
                          "creating cutoff radii for atomic matrices", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.5_dp, &
                          usage="AUTO_CUTOFF_SCALE 0.5_dp", &
                          citations=(/Rayson2009/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="EPS_FB", &
                          description="Default tolerance used in generating the filter "// &
                          "matrix. Anything less than EPS_FB will be "// &
                          "regarded as zero", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=1.e-12_dp, &
                          usage="EPS_FB 1.e-12", &
                          citations=(/Rayson2009/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="COLLECTIVE_COMMUNICATION", &
                          description="If set to TRUE, then all MPI communications "// &
                          "required for the construction of the "// &
                          "filter matrix is done at the start and end "// &
                          "of each filter matrix calculation. This "// &
                          "makes communications more efficient, at "// &
                          "the expense of using more memory. If you "// &
                          "find the fb_fltrmat_add_blkcol_mpi times "// &
                          "at the end of CP2K output is high, then "// &
                          "run again with this option set to .TRUE.", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=logical_t, &
                          default_l_val=.FALSE., &
                          usage="COLLECTIVE_COMMUNICATION T")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_filtermatrix_section

END MODULE qs_fb_input
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief input constants for xc
! **************************************************************************************************
MODULE xc_input_constants
   IMPLICIT NONE
   PRIVATE

   INTEGER, PARAMETER, PUBLIC               :: xc_deriv_pw = 0, &
                                               xc_deriv_spline2 = 1, &
                                               xc_deriv_spline3 = 2, &
                                               xc_deriv_spline2_smooth = 3, &
                                               xc_deriv_spline3_smooth = 4, &
                                               xc_deriv_nn10_smooth = 5, &
                                               xc_deriv_nn50_smooth = 6, &
                                               xc_deriv_collocate = 7

   INTEGER, PARAMETER, PUBLIC               :: xc_rho_no_smooth = 0, &
                                               xc_rho_spline2_smooth = 1, &
                                               xc_rho_spline3_smooth = 2, &
                                               xc_rho_nn10 = 3, &
                                               xc_rho_nn50 = 4

   !VWN xc functional flavours
   INTEGER, PARAMETER, PUBLIC               :: do_vwn5 = 1, &
                                               do_vwn3 = 2

   INTEGER, PARAMETER, PUBLIC               :: xc_pbe_orig = 11, &
                                               xc_pbe_rev = 12, &
                                               xc_pbe_sol = 13

   INTEGER, PARAMETER, PUBLIC               :: xc_old_f_routine = 0, &
                                               xc_new_f_routine = 1, &
                                               xc_test_lsd_f_routine = 3, &
                                               xc_debug_new_routine = 4

   INTEGER, PARAMETER, PUBLIC               :: xgga_b88 = 101, &
                                               xgga_b88x = 102, &
                                               xgga_pw86 = 103, &
                                               xgga_pw91 = 104, &
                                               xgga_pbex = 105, &
                                               xgga_revpbe = 106, &
                                               xgga_optx = 107, &
                                               xgga_hcth = 108, &
                                               xgga_tpss = 109, &
                                               xalpha = 110, &
                                               xgga_pbe = 111, &
                                               xgga_opt = 112, &
                                               xgga_ev93 = 113

   INTEGER, PARAMETER, PUBLIC               :: cgga_lyp = 201, &
                                               c_pade = 202, &
                                               cgga_hcth93 = 203, &
                                               cgga_hcth120 = 204, &
                                               cgga_hcth147 = 205, &
                                               cgga_hcth407 = 206, &
                                               cgga_cs1 = 207, &
                                               cgga_pw86 = 208, &
                                               c_pw92 = 209, &
                                               c_pw92dmc = 210, &
                                               c_pw92vmc = 211, &
                                               c_pz = 212, &
                                               c_pzdmc = 213, &
                                               c_pzvmc = 214, &
                                               c_vwn = 215, &
                                               cgga_pbe = 216

   INTEGER, PARAMETER, PUBLIC               :: ke_ol1 = 301, &
                                               ke_ol2 = 302, &
                                               ke_llp = 303, &
                                               ke_pw86 = 304, &
                                               ke_pw91 = 305, &
                                               ke_lc = 306, &
                                               ke_t92 = 307, &
                                               ke_pbe = 308, &
                                               ke_revpbe = 309, &
                                               ke_tf = 310, &
                                               ke_tfw = 311

   INTEGER, PARAMETER, PUBLIC               :: pw_orig = 209, &
                                               pw_dmc = 210, &
                                               pw_vmc = 211

   INTEGER, PARAMETER, PUBLIC               :: pz_orig = 212, &
                                               pz_dmc = 213, &
                                               pz_vmc = 214

   INTEGER, PARAMETER, PUBLIC               :: xc_b97_orig = 401, &
                                               xc_b97_grimme = 402, &
                                               xc_b97_mardirossian = 403

END MODULE xc_input_constants
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types and initialization routines for Minimax-Ewald method for electron
!>        repulsion integrals.
!> \par History
!>       2015 09 created
!> \author Patrick Seewald
! **************************************************************************************************

MODULE eri_mme_types

   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_section_create,&
                                              cp_print_key_unit_nr,&
                                              medium_print_level
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'eri_mme_types'

   REAL(KIND=dp), PARAMETER, PUBLIC :: lattice_sum_prec = 1.0E-20_dp ! truncation threshold for summing PGFs
   INTEGER, PARAMETER, PUBLIC :: dynamic_lattice_space = 0, rec_lattice_space = 1, dir_lattice_space = 2
   INTEGER, PARAMETER, PUBLIC :: n_minimax_max = 53

   PUBLIC :: eri_mme_param, &
             eri_mme_init_read_input, &
             eri_mme_release, &
             create_eri_mme_section

   TYPE eri_mme_param
      TYPE(section_vals_type), &
         POINTER                    :: mme_section => NULL()
      INTEGER                          :: n_minimax
      REAL(KIND=dp), POINTER, &
         DIMENSION(:)               :: minimax_aw => NULL()
      REAL(KIND=dp), DIMENSION(3, 3) :: hmat, h_inv
      REAL(KIND=dp)                  :: vol
      LOGICAL                          :: is_ortho
      REAL(KIND=dp)                  :: cutoff
      LOGICAL                          :: do_calib_cutoff
      LOGICAL                          :: print_calib
      REAL(KIND=dp)                  :: cutoff_min, cutoff_max, cutoff_delta, &
                                        cutoff_eps, error_tot
      INTEGER                          :: G_count, R_count
      REAL(KIND=dp)                  :: G_min, R_min
      LOGICAL                          :: is_valid
      INTEGER                          :: l_max
   END TYPE eri_mme_param

CONTAINS

! **************************************************************************************************
!> \brief Create main input section
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_eri_mme_section(section)
      TYPE(section_type), INTENT(INOUT), POINTER         :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_eri_mme_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      NULLIFY (keyword, print_key, subsection)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("eri_mme_types.F",89)
      CALL section_create(section, "ERI_MME", &
                          description="Parameters for the calculation of electron repulsion "// &
                          "integrals (ERI) using the Minimax-Ewald (MME) method. "// &
                          "Note: N_MINIMAX is the only parameter to be tuned for accuracy, "// &
                          "all other parameters can be left to default. "// &
                          "Method may not yet be sufficiently accurate since the number "// &
                          "of terms in minimax approximation is limited to 15 in the current "// &
                          "implementation.", &
                          n_keywords=5, n_subsections=1)

      CALL keyword_create(keyword, &
                          name="N_MINIMAX", &
                          description="Number of terms in minimax approximation of "// &
                          "reciprocal space potential. ", &
                          default_i_val=20)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="CUTOFF", &
                          description="User-defined energy cutoff to be used only if "// &
                          "DO_CALIBRATE_CUTOFF is set to .FALSE. ", &
                          default_r_val=300.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="DO_CALIBRATE_CUTOFF", &
                          description="Whether the energy cutoff shall be calibrated to "// &
                          "minimize upper bound error estimate. ", &
                          default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "ERI_MME_INFO", &
                                       description="Controls the printing info.", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL keyword_create(keyword, &
                          name="PRINT_CALIB", &
                          description="Print detailed info on calibration. ", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, "CUTOFF_CALIB", &
                          description="Parameters for the calibration of the energy cutoff by "// &
                          "minimizing the errors due to finite cutoff and minimax approximation. "// &
                          "Implemented as bisection of error(minimax) - error(cutoff). Not "// &
                          "implemented for non-orthorhombic cells. ", &
                          n_keywords=5, n_subsections=0)

      CALL keyword_create(keyword, &
                          name="MIN", &
                          description="Initial guess of lower bound for cutoff. ", &
                          default_r_val=10.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="MAX", &
                          description="Initial guess of upper bound for cutoff. ", &
                          default_r_val=10000.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="DELTA", &
                          description="Relative widening of cutoff interval in case starting "// &
                          "values are not valid. ", &
                          default_r_val=0.9_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="EPS", &
                          description="Relative cutoff precision required to stop calibration. ", &
                          default_r_val=0.01_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_eri_mme_section

! **************************************************************************************************
!> \brief Read input and initialize parameter type
!> \param mme_section ...
!> \param param ...
! **************************************************************************************************
   SUBROUTINE eri_mme_init_read_input(mme_section, param)
      TYPE(section_vals_type), INTENT(INOUT), POINTER    :: mme_section
      TYPE(eri_mme_param), INTENT(INOUT)                 :: param

      CHARACTER(len=*), PARAMETER :: routineN = 'eri_mme_init_read_input', &
         routineP = moduleN//':'//routineN

      CHARACTER(len=2)                                   :: string
      INTEGER                                            :: unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: subsection

      logger => cp_get_default_logger()
      unit_nr = cp_print_key_unit_nr(logger, mme_section, "ERI_MME_INFO", &
                                     extension=".eri_mme")

      NULLIFY (subsection)
      CALL section_vals_val_get(mme_section, "N_MINIMAX", i_val=param%n_minimax)
      WRITE (string, '(I2)') n_minimax_max
      IF (param%n_minimax .GT. n_minimax_max) &
         CALL cp__b("eri_mme_types.F",205,"The maximum allowed number of minimax points N_MINIMAX is "//TRIM(string))
      CALL section_vals_val_get(mme_section, "CUTOFF", r_val=param%cutoff)
      CALL section_vals_val_get(mme_section, "DO_CALIBRATE_CUTOFF", l_val=param%do_calib_cutoff)
      CALL section_vals_val_get(mme_section, "PRINT_CALIB", l_val=param%print_calib)
      subsection => section_vals_get_subs_vals(mme_section, "CUTOFF_CALIB")
      CALL section_vals_val_get(subsection, "MIN", r_val=param%cutoff_min)
      CALL section_vals_val_get(subsection, "MAX", r_val=param%cutoff_max)
      CALL section_vals_val_get(subsection, "EPS", r_val=param%cutoff_eps)
      CALL section_vals_val_get(subsection, "DELTA", r_val=param%cutoff_delta)
      param%mme_section => mme_section

      ALLOCATE (param%minimax_aw(2*param%n_minimax))

      param%is_valid = .FALSE.

      IF (unit_nr > 0) THEN
         WRITE (unit_nr, '(/T2, A, T78, I3/)') "ERI_MME| Number of terms in minimax approximation:", param%n_minimax
      ENDIF

      CALL cp_print_key_finished_output(unit_nr, logger, param%mme_section, "ERI_MME_INFO")

   END SUBROUTINE eri_mme_init_read_input

! **************************************************************************************************
!> \brief ...
!> \param param ...
! **************************************************************************************************
   SUBROUTINE eri_mme_release(param)
      TYPE(eri_mme_param), INTENT(INOUT)                 :: param

      CHARACTER(LEN=*), PARAMETER :: routineN = 'eri_mme_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(param%minimax_aw)) DEALLOCATE (param%minimax_aw)
   END SUBROUTINE eri_mme_release

END MODULE eri_mme_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief function that builds the hartree fock exchange section of the input
!> \par History
!>      09.2007 created
!> \author Manuel Guidon
! **************************************************************************************************
MODULE input_cp2k_hfx
   USE bibliography,                    ONLY: Guidon2008,&
                                              Guidon2009
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              medium_print_level
   USE input_constants,                 ONLY: &
        do_hfx_potential_coulomb, do_hfx_potential_gaussian, do_hfx_potential_id, &
        do_hfx_potential_long, do_hfx_potential_mix_cl, do_hfx_potential_mix_cl_trunc, &
        do_hfx_potential_mix_lg, do_hfx_potential_short, do_hfx_potential_truncated, ehrenfest, &
        gaussian
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_hfx'

   PUBLIC :: create_hfx_section

CONTAINS

! **************************************************************************************************
!> \brief creates the input section for the hf part
!> \param section the section to create
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE create_hfx_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_hfx_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_hfx.F",60)
      CALL section_create(section, "HF", &
                          description="Sets up the Hartree-Fock parameters if requested ", &
                          n_keywords=5, n_subsections=2, repeats=.TRUE., &
                          citations=(/Guidon2008, Guidon2009/))

      NULLIFY (keyword, print_key, subsection)

      CALL keyword_create(keyword, name="FRACTION", &
                          description="The fraction of Hartree-Fock to add to the total energy. "// &
                          "1.0 implies standard Hartree-Fock if used with XC_FUNCTIONAL NONE. "// &
                          "NOTE: In a mixed potential calculation this should be set to 1.0, otherwise "// &
                          "all parts are multiplied with this factor. ", &
                          usage="FRACTION 1.0", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TREAT_LSD_IN_CORE", &
                          description="Determines how spin denisities are taken into account. "// &
                          "If true, the beta spin density is included via a second in core call. "// &
                          "If false, alpha and beta spins are done in one shot ", &
                          usage="TREAT_LSD_IN_CORE TRUE", default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PW_HFX", &
                          description="Compute the Hartree-Fock energy also in the plane wave basis."// &
                          "The value is ignored, and intended for debugging only.", &
                          usage="PW_HFX FALSE", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PW_HFX_BLOCKSIZE", &
                          description="Improve the performance of pw_hfx at the cost of some additional memory "// &
                          "by storing the realspace representation of PW_HFX_BLOCKSIZE states.", &
                          usage="PW_HFX_BLOCKSIZE 20", default_i_val=20)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "HF_INFO", &
                                       description="Controls the printing basic info about hf method", &
                                       print_level=medium_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL create_hf_pbc_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_hf_screening_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_hf_potential_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_hf_load_balance_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_hf_memory_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_hfx_section

! **************************************************************************************************
!> \brief !****f* input_cp2k_dft/create_hf_load_balance_section [1.0] *
!>
!>      creates the input section for the hf potential part
!> \param section the section to create
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE create_hf_load_balance_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_hf_load_balance_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_hfx.F",144)
      CALL section_create(section, "LOAD_BALANCE", &
                          description="Parameters influencing the load balancing of the HF", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/guidon2008/))

      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="NBINS", &
         description="Number of bins per process used to group atom quartets.", &
         usage="NBINS 32", &
         default_i_val=64)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="BLOCK_SIZE", &
         description="Determines the blocking used for the atomic quartet loops. "// &
         "A proper choice can speedup the calculation. The default (-1) is automatic.", &
         usage="BLOCK_SIZE 4", &
         default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="RANDOMIZE", &
         description="This flag controls the randomization of the bin assignment to processes. "// &
         "For highly ordered input structures with a bad load balance, setting "// &
         "this flag to TRUE might improve.", &
         usage="RANDOMIZE TRUE", &
         default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "PRINT", &
                                       description="Controls the printing of info about load balance", &
                                       print_level=medium_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)

      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, &
                          name="LOAD_BALANCE_INFO", &
                          description="Activates the printing of load balance information ", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_release(print_key)

   END SUBROUTINE create_hf_load_balance_section

! **************************************************************************************************
!> \brief !****f* input_cp2k_dft/create_hf_potential_section [1.0] *
!>
!>      creates the input section for the hf potential part
!> \param section the section to create
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE create_hf_potential_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_hf_potential_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_hfx.F",215)
      CALL section_create(section, "INTERACTION_POTENTIAL", &
                          description="Sets up interaction potential if requested ", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/guidon2008, guidon2009/))

      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="POTENTIAL_TYPE", &
         description="Which interaction potential should be used "// &
         "(Coulomb, longrange or shortrange).", &
         usage="POTENTIAL_TYPE SHORTRANGE", &
         enum_c_vals=s2a("COULOMB", "SHORTRANGE", "LONGRANGE", "MIX_CL", "GAUSSIAN", &
                         "MIX_LG", "IDENTITY", "TRUNCATED", "MIX_CL_TRUNC"), &
         enum_i_vals=(/do_hfx_potential_coulomb, do_hfx_potential_short, do_hfx_potential_long, &
                       do_hfx_potential_mix_cl, do_hfx_potential_gaussian, do_hfx_potential_mix_lg, &
                       do_hfx_potential_id, do_hfx_potential_truncated, do_hfx_potential_mix_cl_trunc/), &
         enum_desc=s2a("Coulomb potential: 1/r", &
                       "Shortrange potential: erfc(omega*r)/r", &
                       "Longrange potential: erf(omega*r)/r", &
                       "Mix coulomb and longrange potential: 1/r + erf(omega*r)/r", &
                       "Damped Gaussian potential: exp(-omega^2*r^2)", &
                       "Mix Gaussian and longrange potential: erf(omega*r)/r + exp(-omega^2*r^2)", &
                       "Overlap", &
                       "Truncated coulomb potential: if (r &lt; R_c) 1/r else 0", &
                       "Truncated Mix coulomb and longrange potential, assumes/requires that the erf has fully decayed at R_c"), &
         default_i_val=do_hfx_potential_coulomb)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="OMEGA", &
         description="Parameter for short/longrange interaction", &
         usage="OMEGA 0.5", &
         default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SCALE_COULOMB", &
                          description="Scales Hartree-Fock contribution arising from a coulomb potential. "// &
                          "Only valid when doing a mixed potential calculation", &
                          usage="SCALE_COULOMB 1.0", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SCALE_LONGRANGE", &
                          description="Scales Hartree-Fock contribution arising from a longrange potential. "// &
                          "Only valid when doing a mixed potential calculation", &
                          usage="SCALE_LONGRANGE 1.0", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SCALE_GAUSSIAN", &
                          description="Scales Hartree-Fock contribution arising from a gaussian potential. "// &
                          "Only valid when doing a mixed potential calculation", &
                          usage="SCALE_GAUSSIAN 1.0", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUTOFF_RADIUS", &
                          description="Determines cutoff radius for the truncated 1/r potential. "// &
                          "Only valid when doing truncated calculation", &
                          usage="CUTOFF_RADIUS 10.0", type_of_var=real_t, & ! default_r_val=10.0_dp,&
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="T_C_G_DATA", &
         description="Location of the file t_c_g.dat that contains the data for the "// &
         "evaluation of the truncated gamma function ", &
         usage="T_C_G_DATA /data/t_c_g.dat", &
         default_c_val="t_c_g.dat")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_hf_potential_section

!****f* input_cp2k_dft/create_hf_screening_section [1.0] *

! **************************************************************************************************
!> \brief creates the input section for the hf screening part
!> \param section the section to create
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE create_hf_screening_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_hf_screening_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_hfx.F",312)
      CALL section_create(section, "SCREENING", &
                          description="Sets up screening parameters if requested ", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/guidon2008, guidon2009/))

      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="EPS_SCHWARZ", &
         description="Screens the near field part of the electronic repulsion "// &
         "integrals using the Schwarz inequality for the given "// &
         "threshold.", &
         usage="EPS_SCHWARZ 1.0E-6", &
         default_r_val=1.0E-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="EPS_SCHWARZ_FORCES", &
         description="Screens the near field part of the electronic repulsion "// &
         "integrals using the Schwarz inequality for the given "// &
         "threshold. This will be approximately the accuracy of the forces, "// &
         " and should normally be similar to EPS_SCF", &
         usage="EPS_SCHWARZ_FORCES 1.0E-5", &
         default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="SCREEN_P_FORCES", &
         description="Screens the electronic repulsion integrals for the forces "// &
         "using the density matrix. "// &
         "This results in a significant speedup for large systems, "// &
         "but might require a somewhat tigher EPS_SCHWARZ_FORCES.", &
         usage="SCREEN_P_FORCES TRUE", &
         default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="SCREEN_ON_INITIAL_P", &
                          description="Screen on an initial density matrix. For the first MD step"// &
                          " this matrix must be provided by a Restart File.", &
                          usage="SCREEN_ON_INITIAL_P  TRUE", default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="P_SCREEN_CORRECTION_FACTOR", &
                          description="Recalculates integrals on the fly if the actual density matrix is"// &
                          " larger by a given factor than the initial one. If the factor is set"// &
                          " to 0.0_dp, this feature is disabled.", &
                          usage="P_SCREEN_CORRECTION_FACTOR  0.0_dp", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_hf_screening_section

! **************************************************************************************************
!> \brief creates the input section for the hf-pbc part
!> \param section the section to create
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE create_hf_pbc_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_hf_pbc_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_hfx.F",388)
      CALL section_create(section, "PERIODIC", &
                          description="Sets up periodic boundary condition parameters if requested ", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/guidon2008, guidon2009/))
      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="NUMBER_OF_SHELLS", &
         description="Number of shells taken into account for periodicity. "// &
         "By default, cp2k tries to automatically evaluate this number. "// &
         "This algorithm might be to conservative, resulting in some overhead. "// &
         "You can try to adjust this number in order to make a calculation cheaper. ", &
         usage="NUMBER_OF_SHELLS 2", &
         default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_hf_pbc_section

! **************************************************************************************************
!> \brief creates the input section for the hf-memory part
!> \param section the section to create
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE create_hf_memory_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_hf_memory_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_hfx.F",421)
      CALL section_create(section, "MEMORY", &
                          description="Sets up memory parameters for the storage of the ERI's if requested ", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE., &
                          citations=(/guidon2008/))
      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="EPS_STORAGE_SCALING", &
         variants=(/"EPS_STORAGE"/), &
         description="Scaling factor to scale eps_schwarz. Storage threshold for compression "// &
         "will be EPS_SCHWARZ*EPS_STORAGE_SCALING.", &
         usage="EPS_STORAGE 1.0E-2", &
         default_r_val=1.0E0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="MAX_MEMORY", &
         description="Defines the maximum amount of memory [MiB] to be consumed by the full HFX module. "// &
         "All temporary buffers and helper arrays are subtracted from this number. "// &
         "What remains will be used for storage of integrals. NOTE: This number "// &
         "is assumed to represent the memory available to one MPI process. "// &
         "When running a threaded version, cp2k automatically takes care of "// &
         "distributing the memory among all the threads within a process.", &
         usage="MAX_MEMORY 256", &
         default_i_val=512)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="STORAGE_LOCATION", &
         description="Loaction where ERI's are stored if MAX_DISK_SPACE /=0 "// &
         "Expects a path to a directory. ", &
         usage="STORAGE_LOCATION /data/scratch", &
         default_c_val=".")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="MAX_DISK_SPACE", &
         description="Defines the maximum amount of disk space [MiB] used to store precomputed "// &
         "compressed four-center integrals. If 0, nothing is stored to disk", &
         usage="MAX_DISK_SPACE 256", &
         default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TREAT_FORCES_IN_CORE", &
                          description="Determines whether the derivative ERI's should be stored to RAM or not. "// &
                          "Only meaningful when performing Ehrenfest MD. "// &
                          "Memory usage is defined via MAX_MEMORY, i.e. the memory is shared wit the energy ERI's.", &
                          usage="TREAT_FORCES_IN_CORE TRUE", default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_hf_memory_section

END MODULE input_cp2k_hfx
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief input section for MP2
!> \par History
!>      05.2011 created
!> \author MDB
! **************************************************************************************************
MODULE input_cp2k_mp2
   USE bibliography,                    ONLY: DelBen2012,&
                                              DelBen2013,&
                                              DelBen2015,&
                                              DelBen2015b
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              debug_print_level,&
                                              high_print_level,&
                                              low_print_level,&
                                              medium_print_level,&
                                              silent_print_level
   USE eri_mme_types,                   ONLY: create_eri_mme_section
   USE input_constants,                 ONLY: &
        do_eri_gpw, do_eri_mme, do_hfx_potential_coulomb, do_mp2_potential_tshpsc, gaussian, &
        mp2_method_direct, mp2_method_gpw, mp2_method_none, mp2_ri_optimize_basis, numerical, &
        ri_coulomb, ri_mp2_laplace, ri_mp2_method_gpw, ri_overlap, ri_rpa_g0w0_crossing_newton, &
        ri_rpa_g0w0_crossing_none, ri_rpa_g0w0_crossing_z_shot, ri_rpa_method_gpw, &
        wfc_mm_style_gemm, wfc_mm_style_syrk
   USE input_cp2k_hfx,                  ONLY: create_hfx_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: integer_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_mp2'

   PUBLIC :: create_mp2_section

CONTAINS

! **************************************************************************************************
!> \brief creates the input section for the mp2 part
!> \param section the section to create
!> \author MDB
! **************************************************************************************************
   SUBROUTINE create_mp2_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mp2_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mp2.F",69)
      CALL section_create(section, "WF_CORRELATION", &
                          description="Sets up the Wavefunction-based Correlation parameters if requested ", &
                          n_keywords=8, n_subsections=6, repeats=.TRUE., &
                          citations=(/DelBen2012, DelBen2013, DelBen2015, DelBen2015b/))

      NULLIFY (keyword, print_key, subsection)

      CALL keyword_create( &
         keyword=keyword, &
         name="METHOD", &
         citations=(/DelBen2012, DelBen2013/), &
         description="Which method should be used to compute the MP2 energy", &
         usage="METHOD MP2_GPW", &
         enum_c_vals=s2a("NONE", "DIRECT_CANONICAL", "MP2_GPW", "RI_MP2_GPW", "RI_RPA_GPW", "RI_SOS_LAPLACE", &
                         "OPTIMIZE_RI_BASIS"), &
         enum_i_vals=(/mp2_method_none, mp2_method_direct, mp2_method_gpw, ri_mp2_method_gpw, &
                       ri_rpa_method_gpw, ri_mp2_laplace, mp2_ri_optimize_basis/), &
         enum_desc=s2a("Skip MP2 calculation", &
                       "Use the direct mp2 canonical approach", &
                       "Use the GPW approach to MP2", &
                       "Use the GPW approach to RI-MP2", &
                       "Use the GPW approach to RI-RPA", &
                       "Use the GPW approach to RI-SOS-Laplace-MP2", &
                       "Optimize RIMP2 basis set"), &
         default_i_val=mp2_method_direct)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="MEMORY", &
         description="Maximum allowed total memory usage during MP2 methods [MiB].", &
         usage="MEMORY 1500 ", &
         default_r_val=1.024E+3_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="SCALE_S", &
         description="Scaling factor of the singlet energy component (opposite spin, OS). ", &
         usage="SCALE_S  1.0", &
         default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="SCALE_T", &
         description="Scaling factor of the triplet energy component (same spin, SS).", &
         usage="SCALE_T  1.0", &
         default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="GROUP_SIZE", &
         variants=(/"NUMBER_PROC"/), &
         description="Group size used in the computation of the integrals. Default is to use all processors (GROUP_SIZE=-1)."// &
         "A smaller group size (for example the node size), might a better choice if the actual MP2 time is large "// &
         "compared to integral computation time. This is usually the case if the total number of processors is not too large.", &
         usage="GROUP_SIZE 32", &
         default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="ROW_BLOCK", &
         variants=(/"ROW_BLOCK_SIZE"/), &
         description="Size of the row block used in the SCALAPACK block cyclic data distribution."// &
         "Default is (ROW_BLOCK=-1) is automatic. "// &
         "A proper choice can speedup the parallel matrix multiplication in the case of RI-RPA and RI-SOS-MP2-Laplace.", &
         usage="ROW_BLOCK 512", &
         default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="COL_BLOCK", &
         variants=(/"COL_BLOCK_SIZE"/), &
         description="Size of the column block used in the SCALAPACK block cyclic data distribution."// &
         "Default is (COL_BLOCK=-1) is automatic. "// &
         "A proper choice can speedup the parallel matrix multiplication in the case of RI-RPA and RI-SOS-MP2-Laplace.", &
         usage="COL_BLOCK 512", &
         default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="CALC_COND_NUM", &
         variants=(/"CALC_CONDITION_NUMBER"/), &
         description="Calculate the condition number of the (P|Q) matrix for the RI methods.", &
         usage="CALC_COND_NUM", &
         default_l_val=.FALSE., &
         lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="RI_METRIC", &
         variants=(/"RI"/), &
         description="Define which RI metric should be used for calculating B^P_ia.", &
         usage="RI_METRIC COULOMB", &
         enum_c_vals=s2a("COULOMB", "OVERLAP"), &
         enum_i_vals=(/ri_coulomb, ri_overlap/), &
         enum_desc=s2a("Use Coulomb metric.", &
                       "Use overlap metric, recommended in combination with imaginary time for O(N^3) RPA."), &
         default_i_val=ri_coulomb)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ERI_METHOD", &
                          description="Method for calculating periodic electron repulsion integrals "// &
                          "(MME method is faster but experimental, in the current implementation only 2-center ERIs with "// &
                          "Coulomb metric are supported).", &
                          usage="ERI_METHOD MME", &
                          enum_c_vals=s2a("GPW", "MME"), &
                          enum_i_vals=(/do_eri_gpw, do_eri_mme/), &
                          enum_desc=s2a("Uses Gaussian Plane Wave method [DelBen2013]", &
                                        "Uses MiniMax-Ewald method (ERI_MME subsection)"), &
                          default_i_val=do_eri_gpw)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "MP2_INFO", &
                                       description="Controls the printing basic info about MP2 method", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL create_mp2_direct(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_wfc_gpw(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_ri_mp2(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_opt_ri_basis(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_ri_rpa(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_ri_laplace(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_cphf(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mp2_potential(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_eri_mme_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_mp2_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_mp2_direct(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mp2_direct', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mp2.F",255)
      CALL section_create(section, "DIRECT_CANONICAL", &
                          description="Parameters influencing the direct canonical method", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="BIG_SEND", &
         description="Send big messages between processes (useful for >48 processors).", &
         usage="BIG_SEND", &
         default_l_val=.TRUE., &
         lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_mp2_direct

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_ri_mp2(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ri_mp2', routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mp2.F",285)
      CALL section_create(section, "RI_MP2", &
                          description="Parameters influencing the RI MP2 method", &
                          n_keywords=3, n_subsections=0, repeats=.FALSE., &
                          citations=(/DelBen2013/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="BLOCK_SIZE", &
                          variants=(/"MESSAGE_SIZE"/), &
                          description="Determines the blocking used for communication in RI-MP2. Larger BLOCK_SIZE "// &
                          "reduces communication but requires more memory. The default (-1) is automatic.", &
                          usage="BLOCK_SIZE 2", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_CANONICAL", &
                          description="Threshold for discriminate if a given ij pairs "// &
                          "of the unrelaxed MP2 density matrix has to be "// &
                          "calculated with a canonical reformulation based "// &
                          "on the occupied eigenvalues differences.", &
                          usage="EPS_CANONICAL 1.0E-8", type_of_var=real_t, &
                          default_r_val=1.0E-7_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="FREE_HFX_BUFFER", &
         description="Free the buffer containing the 4 center integrals used in the Hartree-Fock exchange calculation. "// &
         "This will be effective only for gradients calculations, since for the energy only "// &
         "case, the buffers are released by default. (Right now debugging only).", &
         usage="FREE_HFX_BUFFER", &
         default_l_val=.TRUE., &
         lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_ri_mp2

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_opt_ri_basis(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_opt_ri_basis', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mp2.F",338)
      CALL section_create(section, "OPT_RI_BASIS", &
                          description="Parameters influencing the optimization of the RI MP2 basis. "// &
                          "Only exponents of non-contracted auxiliary basis can be optimized. "// &
                          "An initial RI auxiliary basis has to be specified.", &
                          n_keywords=6, n_subsections=0, repeats=.FALSE., &
                          citations=(/DelBen2013/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="DELTA_I_REL", &
                          variants=(/"DI_REL"/), &
                          description="Target accuracy in the relative deviation of the amplitudes calculated with "// &
                          "and without RI approximation, (more details in Chem.Phys.Lett.294(1998)143).", &
                          usage="DELTA_I_REL  1.0E-6_dp", &
                          default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DELTA_RI", &
                          variants=(/"DRI"/), &
                          description="Target accuracy in the absolute difference between the RI-MP2 "// &
                          "and the exact MP2 energy, DRI=ABS(E_MP2-E_RI-MP2).", &
                          usage="DELTA_RI  1.0E-6_dp", &
                          default_r_val=5.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_DERIV", &
                          variants=(/"EPS_NUM_DERIV"/), &
                          description="The derivatives of the MP2 energy with respect to the "// &
                          "exponents of the basis are calculated numerically. "// &
                          "The change in the exponent a_i employed for the numerical evaluation "// &
                          "is defined as h_i=EPS_DERIV*a_i.", &
                          usage="EPS_DERIV  1.0E-3_dp", &
                          default_r_val=1.0E-3_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ITER", &
                          variants=(/"MAX_NUM_ITER"/), &
                          description="Specifies the maximum number of steps in the RI basis optimization.", &
                          usage="MAX_ITER 100", &
                          default_i_val=50)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUM_FUNC", &
                          description="Specifies the number of function, for each angular momentum (s, p, d ...), "// &
                          "employed in the automatically generated initial guess. "// &
                          "This will be effective only if RI_AUX_BASIS_SET in the KIND section is not specified.", &
                          usage="NUM_FUNC {number of s func.} {number of p func.} ...", &
                          n_var=-1, default_i_vals=(/-1/), type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="BASIS_SIZE", &
                          description="Specifies the size of the auxiliary basis set automatically "// &
                          "generated as initial guess. This will be effective only if RI_AUX_BASIS_SET "// &
                          "in the KIND section and NUM_FUNC are not specified.", &
                          usage="BASIS_SIZE  (MEDIUM|LARGE|VERY_LARGE)", &
                          enum_c_vals=s2a("MEDIUM", "LARGE", "VERY_LARGE"), &
                          enum_i_vals=(/0, 1, 2/), &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_opt_ri_basis

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_ri_laplace(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ri_laplace', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mp2.F",419)
      CALL section_create(section, "RI_LAPLACE", &
                          description="Parameters influencing the RI-SOS-MP2-Laplace method", &
                          n_keywords=2, n_subsections=0, repeats=.FALSE., &
                          citations=(/DelBen2013/))

      NULLIFY (keyword)

      CALL keyword_create( &
         keyword, name="QUADRATURE_POINTS", &
         variants=(/"LAPLACE_NUM_QUAD_POINTS"/), &
         description="Number of quadrature points for the numerical integration in the RI-SOS-MP2-Laplace method.", &
         usage="QUADRATURE_POINTS 6", &
         default_i_val=5)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="SIZE_INTEG_GROUP", &
         variants=(/"LAPLACE_GROUP_SIZE"/), &
         description="Group size for the integration in the Laplace method, that is the number of processes involved in "// &
         "the computation of each integration point. SIZE_INTEG_GROUP has to be a multiple "// &
         "of GROUP_SIZE in the WF_CORRELATION section. The default (-1) "// &
         "is automatic.", &
         usage="SIZE_INTEG_GROUP 16", &
         default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_ri_laplace

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_ri_rpa(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ri_rpa', routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mp2.F",462)
      CALL section_create(section, "RI_RPA", &
                          description="Parameters influencing the RI RPA method", &
                          n_keywords=2, n_subsections=1, repeats=.FALSE., &
                          citations=(/DelBen2013, DelBen2015/))

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="QUADRATURE_POINTS", &
                          variants=(/"RPA_NUM_QUAD_POINTS"/), &
                          description="Number of quadrature points for the numerical integration in the RI-RPA method.", &
                          usage="QUADRATURE_POINTS 60", &
                          default_i_val=40)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SIZE_FREQ_INTEG_GROUP", &
                          variants=(/"RPA_GROUP_SIZE"/), &
                          description="Group size for frequency integration, that is the number of processes involved in "// &
                          "the computation of each integration point. SIZE_FREQ_INTEG_GROUP has to be a multiple "// &
                          "of GROUP_SIZE in the WF_CORRELATION section. The default (-1) "// &
                          "is automatic.", &
                          usage="SIZE_FREQ_INTEG_GROUP 16", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="MM_STYLE", &
                          description="Matrix multiplication style for the Q matrix.", &
                          usage="MM_STYLE GEMM", &
                          enum_c_vals=s2a("GEMM", "SYRK"), &
                          enum_i_vals=(/wfc_mm_style_gemm, wfc_mm_style_syrk/), &
                          enum_desc=s2a("Use pdgemm: more flops, maybe faster.", &
                                        "Use pdysrk: fewer flops, maybe slower."), &
                          default_i_val=wfc_mm_style_gemm)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="MINIMAX_QUADRATURE", &
         variants=(/"MINIMAX"/), &
         description="Use the Minimax quadrature scheme for performing the numerical integration. "// &
         "Maximum number of quadrature point limited to 20.", &
         usage="MINIMAX_QUADRATURE", &
         default_l_val=.FALSE., &
         lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="RI_G0W0", &
         variants=(/"GW"/), &
         description="Decide whether to perform an RI_G0W0 calculation on top of RI_ RPA.", &
         usage="RI_G0W0", &
         default_l_val=.FALSE., &
         lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="ADMM", &
         description="Decide whether to perform ADMM in the exact exchange calc. for RPA and/or GW. "// &
         "Not really recommended. In most cases, the Hartree-Fock exchange is not expensive "// &
         "and there is no need for ADMM. Only in case of diffuse functions for periodic systems "// &
         "(only reasonable for GW bandgap calculations) exact exchange can be expensive and then, "// &
         "we recommend using RI for exchange, see G0W0 section.", &
         usage="ADMM", &
         default_l_val=.FALSE., &
         lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="IM_TIME", &
         variants=(/"IMAG_TIME"/), &
         description="Decide whether to calculate Q-matrix for imaginary times.", &
         usage="IM_TIME", &
         default_l_val=.FALSE., &
         lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! here we generate a hfx subsection to use in the case EXX has to be computed after RPA
      CALL create_hfx_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! here we generate a G0W0 subsection to use if G0W0 is desired
      CALL create_ri_g0w0(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_ri_rpa

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_ri_g0w0(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ri_g0w0', routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mp2.F",572)
      CALL section_create(section, "RI_G0W0", &
                          description="Parameters influencing the RI-G0W0 method", &
                          n_keywords=12, n_subsections=0, repeats=.FALSE., &
                          citations=(/DelBen2013/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="CORR_MOS_OCC", &
                          variants=(/"CORR_OCC"/), &
                          description="Number of occupied MOs whose energies are corrected by RI-G0W0. "// &
                          "Counting beginning from HOMO, e.g. 3 corrected occ. MOs correspond "// &
                          "to correction of HOMO, HOMO-1 and HOMO-2. Numerical effort and "// &
                          "storage of RI-G0W0 increase linearly with this number. In case you "// &
                          "want to correct all occ. MOs, insert a number larger than the number "// &
                          "of occ. MOs.", &
                          usage="CORR_OCC 3", &
                          default_i_val=10)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CORR_MOS_VIRT", &
                          variants=(/"CORR_VIRT"/), &
                          description="Number of virtual MOs whose energies are corrected by RI-G0W0. "// &
                          "Counting beginning from LUMO, e.g. 3 corrected occ. MOs correspond "// &
                          "to correction of LUMO, LUMO+1 and LUMO+2. Numerical effort and "// &
                          "storage of RI-G0W0 increase linearly with this number. In case you "// &
                          "want to correct all virt. MOs, insert a number larger than the number "// &
                          "of virt. MOs.", &
                          usage="CORR_VIRT 3", &
                          default_i_val=10)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SCALING", &
                          variants=(/"A_SCALING"/), &
                          description="Set the scaling factor a for the Clenshaw-Curtis grid by hand. "// &
                          "If SCALING < 0, a is calculated as for RI-RPA. Until now, the best "// &
                          "choice seems to be 0.2. The grid points of the frequency grid are "// &
                          "just multiplied by SCALING: omega_j = omega_j*Scaling.", &
                          usage="SCALING 0.2", &
                          default_r_val=0.2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NUMB_POLES", &
                          description="Number of poles for the fitting. Usually, two poles are sufficient. ", &
                          usage="NUMB_POLES 2", &
                          default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OMEGA_MAX_FIT", &
                          description="Determines fitting range for the self-energy on the imaginary axis: "// &
                          "[0, OMEGA_MAX_FIT] for virt orbitals, [-OMEGA_MAX_FIT,0] for occ orbitals. "// &
                          "Unit: Hartree. Default: 0.734996 H = 20 eV. ", &
                          usage="OMEGA_MAX_FIT 0.5", &
                          default_r_val=0.734996_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STOP_CRIT", &
                          variants=(/"STOP_CRIT_1"/), &
                          description="Convergence criterion for the fit. The fitting iteration terminates, if "// &
                          "chi^2(n+1)/chi^2(n)-1 < STOP_CRIT.", &
                          usage="STOP_CRIT 1.0E-7", &
                          default_r_val=1E-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRINT_FIT_ERROR", &
                          variants=(/"FIT_ERROR"/), &
                          description="Print the statistical error due to the fitting. This error in most cases "// &
                          "is very pessimistic (e.g. can be a factor of 10 too large).", &
                          usage="FIT_ERROR", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ITER_FIT", &
                          description="Maximum number of iterations for the Levenberg-Marquard fit.", &
                          usage="MAX_ITER_FIT 10000", &
                          default_i_val=10000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHECK_FIT", &
                          description="If true, the self-energy on the imaginary axis and the fit are printed"// &
                          "the file self_energy_of_MO_<level>_for_imaginary_frequency.", &
                          usage="CHECK_FIT", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CROSSING_SEARCH", &
                          description="Determines, how the self_energy is evaluated on the real axis.", &
                          usage="CROSSING_SEARCH Z_SHOT", &
                          enum_c_vals=s2a("NONE", "Z_SHOT", "NEWTON"), &
                          enum_i_vals=(/ri_rpa_g0w0_crossing_none, ri_rpa_g0w0_crossing_z_shot, ri_rpa_g0w0_crossing_newton/), &
                          enum_desc=s2a("Evaluate the correlation self-energy at the energy eigenvalue of SCF.", &
                                        "Calculate the derivative of Sigma and out of it Z. Then extrapolate using Z.", &
                                        "Make a Newton-Raphson fix point iteration."), &
                          default_i_val=ri_rpa_g0w0_crossing_z_shot)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FERMI_LEVEL_OFFSET", &
                          description="Fermi level for occ. orbitals: e_HOMO + FERMI_LEVEL_OFFSET; "// &
                          "Fermi level for virt. orbitals: e_LUMO - FERMI_LEVEL_OFFSET. "// &
                          "In case e_homo + FERMI_LEVEL_OFFSET < e_lumo - FERMI_LEVEL_OFFSET,"// &
                          "we set Fermi level = (e_HOMO+e_LUMO)/2", &
                          usage="FERMI_LEVEL_OFFSET 1.0E-2", &
                          default_r_val=2.0E-2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EV_SC_ITER", &
                          description="Maximum number of iterations for eigenvalue self-consistency cycle. The "// &
                          "computational effort of RPA and GW scales linearly with this number.", &
                          usage="EV_SC_ITER 3", &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_EV_SC_ITER", &
                          description="Target accuracy for the eigenvalue self-consistency. "// &
                          "If the G0W0 HOMO-LUMO gap differs by less than the "// &
                          "target accuracy during the iteration, the eigenvalue "// &
                          "self-consistency cycle stops. Unit: Hartree.", &
                          usage="EPS_EV_SC_ITER 0.00005", &
                          default_r_val=0.00005_dp, &
                          unit_str="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="HF_LIKE_EV_START", &
                          description="If true, take as input for GW/RPA corrected HF-like eigenvalues according "// &
                          "to PRB 83, 115103 (2011), Sec. IV.", &
                          usage="HF_LIKE_EV_START", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EV_SC_GW_REMOVE_NEG_VIRT_ENERGIES", &
                          variants=(/"REMOVE_NEG"/), &
                          description="Set all GW energies of virtual orbitals which are below the HOMO to the "// &
                          "value of the last eigenvalue-selfconsistency cycle.", &
                          usage="EV_SC_GW_REMOVE_NEG_VIRT_ENERGIES", &
                          default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRINT_GW_DETAILS", &
                          description="If true, prints additional information on the quasiparticle energies.", &
                          usage="PRINT_GW_DETAILS", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RI_SIGMA_X", &
                          description="If true, the exchange self-energy is calculated approximatively with RI. "// &
                          "This is only recommended in case exact exchange is very costly, e.g. when "// &
                          "using diffuse basis functions (seems not to work for periodic systems).", &
                          usage="RI_SIGMA_X", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="RI_METRIC", &
         variants=(/"RI"/), &
         description="Define which RI metric should be used for calculating B^P_nm needed for GW.", &
         usage="RI_METRIC COULOMB", &
         enum_c_vals=s2a("COULOMB", "OVERLAP"), &
         enum_i_vals=(/ri_coulomb, ri_overlap/), &
         enum_desc=s2a("Use Coulomb metric.", &
                       "Use overlap metric."), &
         default_i_val=ri_coulomb)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIX_EXCHANGE", &
                          description="If true, mix CP2K-HFX (EXX) based on the truncated Coulomb operator and. "// &
                          "RI-HFX. Parameter can be used to speed up the convergence of the GW HOMO-LUMO gap of  "// &
                          "periodic systems with the cell size.", &
                          usage="MIX_EXCHANGE", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="FRACTION_EXX", &
         variants=(/"ALPHA"/), &
         description="Mixing parameter between official CP2K-HFX (EXX) based on the truncated Coulomb operator "// &
         "(fraction = alpha) and the RI-HFX (fraction = 1-alpha). To be valid, set MIX_HFX TRUE. "// &
         "Parameter can be used to speed up the convergence of the GW HOMO-LUMO gap of periodic systems "// &
         "with the cell size.", &
         usage="FRACTION_EXX 0.21", &
         default_r_val=0.21_dp, &
         unit_str="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CONTOUR_DEF_START", &
                          variants=(/"CD_START"/), &
                          description="Number of the lowest MO for which contour deformation (CD) is enabled and "// &
                          "analytic continuation is disabled. If negative, CD is disabled.", &
                          usage="CD_START 1", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CONTOUR_DEF_END", &
                          variants=(/"CD_END"/), &
                          description="Number of the highest MO for which contour deformation (CD) is enabled and "// &
                          "analytic continuation is disabled. If negative, CD is disabled.", &
                          usage="CD_END 100", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CONTOUR_DEF_OFFSET", &
                          variants=(/"CD_OFFSET"/), &
                          description="Offset for computing the Z-factor in the contour deformation technique.", &
                          usage="CONTOUR_DEF_OFFSET 1.0E-2", &
                          default_r_val=2.0E-2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_ri_g0w0

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_wfc_gpw(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_wfc_gpw', routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mp2.F",822)
      CALL section_create(section, "WFC_GPW", &
                          description="Parameters for the GPW approach in Wavefunction-based Correlation methods", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="EPS_GRID", &
                          description="Determines a threshold for the GPW based integration", &
                          usage="EPS_GRID 1.0E-9 ", type_of_var=real_t, &
                          default_r_val=1.0E-8_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EPS_FILTER", &
         description="Determines a threshold for the DBCSR based multiply (usually 10 times smaller than EPS_GRID).", &
         usage="EPS_FILTER 1.0E-10 ", type_of_var=real_t, &
         default_r_val=1.0E-9_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUTOFF", &
                          description="The cutoff of the finest grid level in the MP2 gpw integration.", &
                          usage="CUTOFF 300", type_of_var=real_t, &
                          default_r_val=300.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REL_CUTOFF", &
                          variants=(/"RELATIVE_CUTOFF"/), &
                          description="Determines the grid at which a Gaussian is mapped.", &
                          usage="REL_CUTOFF 50", type_of_var=real_t, &
                          default_r_val=50.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MULTIPOLE_TWO_CENT_INT", &
                          description="Use fast algorithm to calculate two-center RI integrals based on multipole expansion", &
                          usage="MULTIPOLE_TWO_CENT_INT", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRINT_LEVEL", &
                          variants=(/"IOLEVEL"/), &
                          description="How much output is written by the individual groups.", &
                          usage="PRINT_LEVEL HIGH", &
                          default_i_val=silent_print_level, enum_c_vals= &
                          s2a("SILENT", "LOW", "MEDIUM", "HIGH", "DEBUG"), &
                          enum_desc=s2a("Almost no output", &
                                        "Little output", "Quite some output", "Lots of output", &
                                        "Everything is written out, useful for debugging purposes only"), &
                          enum_i_vals=(/silent_print_level, low_print_level, medium_print_level, &
                                        high_print_level, debug_print_level/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_wfc_gpw

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_cphf(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_cphf', routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mp2.F",893)
      CALL section_create( &
         section, "CPHF", &
         description="Parameters influencing the solution of the Z-vector equations in MP2 gradients calculations.", &
         n_keywords=2, n_subsections=0, repeats=.FALSE., &
         citations=(/DelBen2013/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="MAX_ITER", &
                          variants=(/"MAX_NUM_ITER"/), &
                          description="Maximum number of iterations allowed for the solution of the Z-vector equations.", &
                          usage="MAX_ITER  50", &
                          default_i_val=30)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_CONV", &
                          description="Convergence threshold for the solution of the Z-vector equations. "// &
                          "The Z-vector equations have the form of a linear system of equations Ax=b, "// &
                          "convergence is achieved when |Ax-b|<=EPS_CONV.", &
                          usage="EPS_CONV 1.0E-6", type_of_var=real_t, &
                          default_r_val=1.0E-4_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_cphf

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_mp2_potential(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mp2_potential', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_mp2.F",933)
      CALL section_create(section, "INTERACTION_POTENTIAL", &
                          description="Parameters the interaction potential in computing the biel integrals", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="POTENTIAL_TYPE", &
         description="Which interaction potential should be used "// &
         "(Coulomb, TShPSC operator).", &
         usage="POTENTIAL_TYPE TSHPSC", &
         enum_c_vals=s2a("COULOMB", "TShPSC"), &
         enum_i_vals=(/do_hfx_potential_coulomb, &
                       do_mp2_potential_TShPSC/), &
         enum_desc=s2a("Coulomb potential: 1/r", &
                       "TShPSC:<ul><li>1/x - s/Rc for x &le; Rc</li>"// &
                       "<li>(1 - s)/Rc - (x - Rc)/Rc^2 + (x - Rc)^2/Rc^3 - "// &
                       "(2*n^2 - 7*n + 9 - 4*s)*(x - Rc)^3/(Rc^4*(n^2 - 2*n + 1)*(n - 1)) + "// &
                       "(6-3*s - 4*n + n^2)*(x - Rc)^4/(Rc^5*(n^4 - 4*n^3 + 6*n^2 - 4*n + 1)) "// &
                       "for Rc &lt; x &le; n*Rc (4th order polynomial)</li>"// &
                       "<li>0 for x &gt; n*Rc</li></ul>"), &
         default_i_val=do_hfx_potential_coulomb)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TRUNCATION_RADIUS", &
                          description="Determines truncation radius for the truncated TShPSC potential. "// &
                          "Only valid when doing truncated calculation", &
                          usage="TRUNCATION_RADIUS 10.0", type_of_var=real_t, &
                          default_r_val=10.0_dp, &
                          unit_str="angstrom")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="TShPSC_DATA", &
         description="Location of the file TShPSC.dat that contains the data for the "// &
         "evaluation of the TShPSC G0 ", &
         usage="TShPSC_DATA t_sh_p_s_c.dat", &
         default_c_val="t_sh_p_s_c.dat")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_mp2_potential

END MODULE input_cp2k_mp2

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief function that build the xc section of the input
!> \par History
!>      10.2009 moved out of input_cp2k_dft [jgh]
!> \author fawzi
! **************************************************************************************************
MODULE input_cp2k_xc
   USE bibliography,                    ONLY: &
        Becke1988, Becke1997, BeckeRoussel1989, Goedecker1996, Grimme2006, Grimme2010, Grimme2011, &
        Heyd2004, Lee1988, Marques2012, Ortiz1994, Perdew1981, Perdew1996, Perdew2008, &
        Proynov2007, Tao2003, Tran2013, Vosko1980, Wellendorff2012, Zhang1998
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              high_print_level
   USE input_constants,                 ONLY: &
        do_adiabatic_hybrid_mcy3, do_adiabatic_model_pade, gaussian, slater, vdw_nl_drsll, &
        vdw_nl_lmkll, vdw_nl_rvv10, vdw_pairpot_dftd2, vdw_pairpot_dftd3, vdw_pairpot_dftd3bj, &
        xc_funct_b3lyp, xc_funct_beefvdw, xc_funct_blyp, xc_funct_bp, xc_funct_hcth120, &
        xc_funct_no_shortcut, xc_funct_olyp, xc_funct_pade, xc_funct_pbe, xc_funct_pbe0, &
        xc_funct_tpss, xc_none, xc_pot_energy_none, xc_pot_energy_sum_eigenvalues, &
        xc_pot_energy_xc_functional, xc_vdw_fun_none, xc_vdw_fun_nonloc, xc_vdw_fun_pairpot
   USE input_cp2k_hfx,                  ONLY: create_hfx_section
   USE input_cp2k_mp2,                  ONLY: create_mp2_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE string_utilities,                ONLY: s2a
   USE xc_input_constants,              ONLY: &
        c_pw92, c_pw92dmc, c_pw92vmc, c_pz, c_pzdmc, c_pzvmc, do_vwn3, do_vwn5, ke_lc, ke_llp, &
        ke_ol1, ke_ol2, ke_pbe, ke_pw86, ke_pw91, ke_t92, pz_orig, xalpha, xc_b97_grimme, &
        xc_b97_mardirossian, xc_b97_orig, xc_debug_new_routine, xc_deriv_collocate, &
        xc_deriv_nn10_smooth, xc_deriv_nn50_smooth, xc_deriv_pw, xc_deriv_spline2, &
        xc_deriv_spline2_smooth, xc_deriv_spline3, xc_deriv_spline3_smooth, xc_new_f_routine, &
        xc_old_f_routine, xc_pbe_orig, xc_pbe_rev, xc_pbe_sol, xc_rho_nn10, xc_rho_nn50, &
        xc_rho_no_smooth, xc_rho_spline2_smooth, xc_rho_spline3_smooth, xc_test_lsd_f_routine, &
        xgga_b88x, xgga_ev93, xgga_opt, xgga_pbex, xgga_pw86, xgga_pw91, xgga_revpbe
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_xc'

   PUBLIC :: create_xc_section

CONTAINS

! **************************************************************************************************
!> \brief creates the structure of the section needed to select the xc functional
!> \param section the section that will be created
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_xc_fun_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_xc_fun_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_xc.F",76)
      CALL section_create(section, name="xc_functional", &
                          description="The xc functional to use", &
                          n_keywords=0, n_subsections=4, repeats=.FALSE., &
                          citations=(/Ortiz1994, Becke1988, Perdew1996, Zhang1998, Lee1988, &
                                      Heyd2004, Vosko1980, Goedecker1996, Perdew1981, &
                                      Tao2003, Wellendorff2012/))

      NULLIFY (subsection, keyword)
      CALL keyword_create( &
         keyword, name="_SECTION_PARAMETERS_", &
         description="Shortcut for the most common functional combinations.", &
         usage="&xc_functional BLYP", &
         enum_c_vals=s2a("B3LYP", "PBE0", "BLYP", "BP", "PADE", "LDA", "PBE", &
                         "TPSS", "HCTH120", "OLYP", "BEEFVDW", "NO_SHORTCUT", "NONE"), &
         enum_i_vals=(/xc_funct_b3lyp, xc_funct_pbe0, xc_funct_blyp, xc_funct_bp, xc_funct_pade, xc_funct_pade, xc_funct_pbe, &
                       xc_funct_tpss, xc_funct_hcth120, xc_funct_olyp, xc_funct_beefvdw, xc_funct_no_shortcut, xc_none/), &
         enum_desc=s2a("B3LYP", &
                       "PBE0. See note in section XC/XC_FUNCTIONAL/PBE.", &
                       "BLYP", "BP", "PADE", "Alias for PADE", &
                       "PBE. See note in section XC/XC_FUNCTIONAL/PBE.", &
                       "TPSS", "HCTH120", "OLYP", &
                       "BEEFVDW", "NO_SHORTCUT", "NONE"), &
         default_i_val=xc_funct_no_shortcut, &
         lone_keyword_i_val=xc_funct_no_shortcut)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="BECKE88", &
                          description="Uses the Becke 88 exchange functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Becke1988/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_x", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="LYP_ADIABATIC", &
                          description="Uses the LYP correlation functional in an adiabatic fashion", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Lee1988/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "LAMBDA", &
                          description="Defines the parameter of the adiabatic curve.", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="BECKE88_LR_ADIABATIC", &
                          description="Uses the Becke 88 longrange exchange functional in an adiabatic fashion", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Becke1988/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_x", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "OMEGA", &
                          description="Potential parameter in erf(omega*r)/r", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "LAMBDA", &
                          description="Defines the parameter of the adiabatic curve", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="BECKE88_LR", &
                          description="Uses the Becke 88 longrange exchange functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Becke1988/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_x", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "OMEGA", &
                          description="Potential parameter in erf(omega*r)/r", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="LYP", &
                          description="Uses the LYP functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Lee1988/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_c", &
                          description="scales the correlation part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="PADE", &
                          description="Uses the PADE functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Goedecker1996/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="HCTH", &
                          description="Uses the HCTH class of functionals", &
                          n_keywords=0, n_subsections=0, repeats=.TRUE.)
      CALL keyword_create(keyword, name="PARAMETER_SET", &
                          description="Which version of the parameters should be used", &
                          usage="PARAMETER_SET 407", &
                          enum_c_vals=(/"93 ", "120", "147", "407"/), &
                          enum_i_vals=(/93, 120, 147, 407/), &
                          default_i_val=120)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="OPTX", &
                          description="Uses the OPTX functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_x", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_libxc_section(subsection, "LIBXC", &
                                "Uses functionals from LIBXC, see also "// &
                                "http://www.tddft.org/programs/octopus/wiki/index.php/Libxc_functionals ", &
                                "FUNCTIONAL GGA_X_PBE GGA_C_PBE")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_libxc_section(subsection, "KE_LIBXC", &
                                "To be used for KG runs. Uses kinetic energy functionals from LIBXC, "// &
                                "see also http://www.tddft.org/programs/octopus/wiki/index.php/Libxc_functionals ", &
                                "FUNCTIONAL GGA_K_LLP")
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="CS1", &
                          description="Uses the CS1 functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="XGGA", &
                          description="Uses one of the XGGA functionals (optimized versions of "// &
                          "some of these functionals might be available outside this section).", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="FUNCTIONAL", &
                          description="Which one of the XGGA functionals should be used", &
                          usage="FUNCTIONAL PW86X", &
                          enum_c_vals=(/ &
                          "BECKE88X", &
                          "PW86X   ", &
                          "PW91X   ", &
                          "PBEX    ", &
                          "REV_PBEX", &
                          "OPTX    ", &
                          "EV93    "/), &
                          enum_i_vals=(/xgga_b88x, xgga_pw86, xgga_pw91, xgga_pbex, xgga_revpbe, xgga_opt, xgga_ev93/), &
                          default_i_val=xgga_b88x)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="KE_GGA", &
                          description="Uses one of the KE_GGA functionals (optimized versions of "// &
                          "some of these functionals might be available outside this section). "// &
                          "These functionals are needed for the computation of the kinetic "// &
                          "energy in the Kim-Gordon method.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="FUNCTIONAL", &
                          description="Which one of the KE_GGA functionals should be used", &
                          usage="FUNCTIONAL (OL1|OL2|LLP|PW86|PW91|LC|T92|PBE)", &
                          enum_c_vals=(/"OL1 ", "OL2 ", "LLP ", "PW86", "PW91", "LC  ", "T92 ", "PBE "/), &
                          enum_i_vals=(/ke_ol1, ke_ol2, ke_llp, ke_pw86, ke_pw91, ke_lc, ke_t92, ke_pbe/), &
                          enum_desc=s2a("Uses first Ou-Yang and Levy functional, currently not producing correct results", &
                                        "Uses second Ou-Yang and Levy functional, currently not producing correct results", &
                                        "Uses Lee, Lee, and Parr functional", &
                                        "Uses Perdew and Wang's 1986 functional", &
                                        "Uses Perdew and Wang's 1991 functional", &
                                        "Uses Lembarki and Chermette functional", &
                                        "Uses Thakkar functional", &
                                        "Uses the 1996 functional of Perdew, Burke and Ernzerhof"), &
                          default_i_val=ke_llp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="P86C", &
                          description="Uses the P86C functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_c", &
                          description="scales the correlation part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="PW92", &
                          description="Uses the PerdewWang correlation functional.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "SCALE", &
                          description="Scaling of the energy functional", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="PARAMETRIZATION", &
                          description="Which one of parametrizations should be used", &
                          usage="PARAMETRIZATION DMC", &
                          enum_c_vals=(/ &
                          "ORIGINAL", &
                          "DMC     ", &
                          "VMC     "/), &
                          enum_i_vals=(/c_pw92, c_pw92dmc, c_pw92vmc/), &
                          default_i_val=c_pw92)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="PZ81", &
                          description="Uses the PZ functional.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE., &
                          citations=(/Perdew1981, Ortiz1994/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="PARAMETRIZATION", &
                          description="Which one of parametrizations should be used", &
                          usage="PARAMETRIZATION DMC", &
                          enum_c_vals=(/ &
                          "ORIGINAL", &
                          "DMC     ", &
                          "VMC     "/), &
                          enum_i_vals=(/c_pz, c_pzdmc, c_pzvmc/), &
                          default_i_val=pz_orig)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_c", &
                          description="scales the correlation part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="TFW", &
                          description="Uses the TFW functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="TF", &
                          description="Uses the TF functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="VWN", &
                          description="Uses the VWN functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Vosko1980/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_c", &
                          description="scales the correlation part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FUNCTIONAL_TYPE", &
                          description="Which version of the VWN functional should be used", &
                          usage="FUNCTIONAL_TYPE VWN5", &
                          enum_c_vals=s2a("VWN5", "VWN3"), &
                          enum_i_vals=(/do_vwn5, do_vwn3/), &
                          enum_desc=s2a("This is the recommended (correct) version of the VWN functional", &
                                        "This version is the default in Gaussian, but not recommended."// &
                                        "Notice that it is also employed in Gaussian's default version of B3LYP"), &
                          default_i_val=do_vwn5)

      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="XALPHA", &
                          description="Uses the XALPHA (SLATER) functional.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="XA", &
                          description="Value of the xa parameter (this does not change the exponent, "// &
                          "just the mixing)", &
                          usage="XA 0.7", default_r_val=2._dp/3._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_x", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="TPSS", &
                          description="Uses the TPSS functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Tao2003/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_x", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_c", &
                          description="scales the correlation part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="PBE", &
                          description="Uses the PBE functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Perdew1996, Zhang1998, Perdew2008/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "parametrization", &
                          description="switches between the different "// &
                          "parametrizations of the functional. "// &
                          "Note: Beta parameters used have only 5 significant digits, "// &
                          "as published. For higher precision and program comparison "// &
                          "use section XC/XC_FUNCTIONAL/LIBXC.", &
                          enum_i_vals=(/xc_pbe_orig, xc_pbe_rev, xc_pbe_sol/), &
                          enum_c_vals=(/"ORIG  ", "revPBE", "PBEsol"/), &
                          enum_desc=(/"original PBE                        ", &
                                      "revised PBE (revPBE)                ", &
                                      "PBE for solids and surfaces (PBEsol)"/), &
                          default_i_val=xc_pbe_orig)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_x", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_c", &
                          description="scales the correlation part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="XWPBE", &
                          description="Uses the short range PBE functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Heyd2004/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_x", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_x0", &
                          description="scales the exchange part of the original hole PBE-functional", &
                          default_r_val=0.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "omega", &
                          description="screening parameter", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="BECKE97", &
                          description="Uses the Becke 97 exchange correlation functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Becke1997, Grimme2006/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create( &
         keyword, "scale_x", &
         description="scales the exchange part of the functional, if -1 the default for the given parametrization is used", &
         default_r_val=-1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_c", &
                          description="scales the correlation part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "parametrization", &
                          description="switches between the B97 and Grimme parametrization ", &
                          enum_i_vals=(/xc_b97_orig, xc_b97_grimme, xc_b97_grimme, xc_b97_mardirossian/), &
                          enum_c_vals=(/"ORIG      ", "B97GRIMME ", "B97_GRIMME", "wB97X-V   "/), &
                          default_i_val=xc_b97_orig)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="BECKE_ROUSSEL", &
                          description="Becke Roussel exchange hole model. Can be used"// &
                          "as long range correction with a truncated coulomb potential", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/BeckeRoussel1989, Proynov2007/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_x", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "CUTOFF_RADIUS", &
                          description="Defines the cutoff radius for the truncation. "// &
                          "If put to zero, the standard full range potential will be used", &
                          usage="CUTOFF_RADIUS 2.0", default_r_val=0.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "GAMMA", &
                          description="Parameter in the exchange hole. "// &
                          "Usually this is put to 1.0 or 0.8", &
                          usage="GAMMA 0.8", default_r_val=1.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="LDA_HOLE_T_C_LR", &
                          description="LDA exchange hole model in truncated coulomb potential", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "SCALE_X", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "CUTOFF_RADIUS", &
                          description="Defines cutoff for lower integration boundary", &
                          default_r_val=0.0_dp, unit_str="angstrom")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="PBE_HOLE_T_C_LR", &
                          description="PBE exchange hole model in trucanted coulomb potential", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "SCALE_X", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "CUTOFF_RADIUS", &
                          description="Defines cutoff for lower integration boundary", &
                          default_r_val=1.0_dp, unit_str="angstrom")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="GV09", &
                          description="Combination of three different exchange hole models", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "SCALE_X", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "CUTOFF_RADIUS", &
                          description="Defines cutoff for lower integration boundary", &
                          default_r_val=0.0_dp, unit_str="angstrom")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "GAMMA", &
                          description="Parameter for Becke Roussel hole", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="BEEF", & !rk: BEEF Exchange
                          description="Uses the BEEFvdW exchange functional", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE., &
                          citations=(/Wellendorff2012/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, "scale_x", &
                          description="scales the exchange part of the functional", &
                          default_r_val=1._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_xc_fun_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
!> \param name ...
!> \param description ...
!> \param usage ...
! **************************************************************************************************
   SUBROUTINE create_libxc_section(section, name, description, usage)
      TYPE(section_type), POINTER                        :: section
      CHARACTER(len=*), INTENT(in)                       :: name, description, usage

      CHARACTER(len=*), PARAMETER :: routineN = 'create_libxc_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_xc.F",730)
      IF(.NOT.(name == "LIBXC" .OR. name == "KE_LIBXC"))CALL cp__a("input_cp2k_xc.F",731)
      NULLIFY (keyword)
      CALL section_create(section, name, description, &
                          n_keywords=3, n_subsections=0, repeats=.FALSE., &
                          citations=(/Marques2012/))
      CALL keyword_create(keyword, "_SECTION_PARAMETERS_", &
                          description="activates the functional", &
                          lone_keyword_l_val=.TRUE., default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="FUNCTIONAL", &
                          description="names of the functionals, see also "// &
                          "http://www.tddft.org/programs/octopus/wiki/index.php/Libxc_functionals ."// &
                          "The precise list of available functionals depends on "// &
                          "the version of libxc interfaced (currently 2.2.2).", &
                          usage=usage, type_of_var=char_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="SCALE", &
                          description="scaling factors of the functionals", &
                          usage="SCALE 1.0 1.0", type_of_var=real_t, &
                          default_r_vals=(/1.0_dp/), n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="PARAMETERS", &
                          description="parameters of the functionals", &
                          type_of_var=real_t, default_r_vals=(/1e20_dp/), n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE

! **************************************************************************************************
!> \brief creates the structure of the section needed to select an xc potential
!> \param section the section that will be created
!> \author thomas chassaing
! **************************************************************************************************
   SUBROUTINE create_xc_potential_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_xc_potential_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_xc.F",776)
      CALL section_create(section, name="xc_potential", &
                          description="The xc potential to use (CAREFUL: xc potential here refers "// &
                          "to potentials that are not derived from an xc functional, but rather are "// &
                          "modelled directly. Therefore there is no consistent xc energy available. "// &
                          "To still get an energy expression, see ENERGY below", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)

      NULLIFY (subsection, keyword)
      CALL section_create(subsection, name="SAOP", &
                          description="Uses the SAOP potential", &
                          n_keywords=3, n_subsections=0, repeats=.TRUE.)
      CALL keyword_create(keyword, name="ALPHA", &
                          description="Value of the alpha parameter (default = 1.19).", &
                          usage="ALPHA 1.19", default_r_val=1.19_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="BETA", &
                          description="Value of the beta parameter (default = 0.01).", &
                          usage="BETA 0.01", default_r_val=0.01_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="K_RHO", &
                          description="Value of the K_rho parameter (default = 0.42).", &
                          usage="ALPHA 0.42", default_r_val=0.42_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, name="ENERGY", &
                          description="How to determine the total energy.", &
                          usage="ENERGY [NONE,XC_FUNCTIONAL,SUM_EIGENVALUES", &
                          enum_c_vals=s2a("NONE", "XC_FUNCTIONAL", "FUNCTIONAL", "SUM_EIGENVALUES", "SOE"), &
                          enum_i_vals=(/xc_pot_energy_none, &
                                        xc_pot_energy_xc_functional, &
                                        xc_pot_energy_xc_functional, &
                                        xc_pot_energy_sum_eigenvalues, &
                                        xc_pot_energy_sum_eigenvalues/), &
                          default_i_val=xc_pot_energy_none)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_xc_potential_section

! **************************************************************************************************
!> \brief creates the structure of the section needed for vdW potentials
!> \param section the section that will be created
!> \author jgh
! **************************************************************************************************
   SUBROUTINE create_vdw_potential_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_vdw_potential_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_xc.F",835)
      CALL section_create(section, name="vdw_potential", &
                          description="This section combines all possible additional dispersion "// &
                          "corrections to the normal XC functionals. This can be more functionals "// &
                          "or simple empirical pair potentials. ", &
                          citations=(/grimme2006, Tran2013/), &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)

      NULLIFY (subsection, keyword)
      CALL keyword_create(keyword, name="POTENTIAL_TYPE", &
                          variants=s2a("DISPERSION_FUNCTIONAL"), &
                          description="Type of dispersion/vdW functional or potential to use", &
                          usage="POTENTIAL_TYPE (NONE|PAIR_POTENTIAL|NON_LOCAL)", &
                          enum_c_vals=s2a("NONE", "PAIR_POTENTIAL", "NON_LOCAL"), &
                          enum_i_vals=(/xc_vdw_fun_none, xc_vdw_fun_pairpot, xc_vdw_fun_nonloc/), &
                          enum_desc=s2a("No dispersion/van der Waals functional", &
                                        "Pair potential van der Waals density functional", &
                                        "Nonlocal van der Waals density functional"), &
                          default_i_val=xc_vdw_fun_none)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="PAIR_POTENTIAL", &
                          description="Information on the pair potential to calculate dispersion", &
                          n_keywords=5, n_subsections=0, repeats=.TRUE.)
      CALL keyword_create(keyword, name="R_CUTOFF", &
                          description="Range of potential. The cutoff will be 2 times this value", &
                          usage="R_CUTOFF 24.0", default_r_val=20.0_dp, &
                          unit_str="angstrom")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="TYPE", &
                          description="Type of potential", &
                          citations=(/grimme2006, grimme2010, grimme2011/), &
                          usage="TYPE (DFTD2|DFTD3|DFTD3(BJ))", &
                          enum_c_vals=s2a("DFTD2", "DFTD3", "DFTD3(BJ)"), &
                          enum_i_vals=(/vdw_pairpot_dftd2, vdw_pairpot_dftd3, vdw_pairpot_dftd3bj/), &
                          enum_desc=s2a("Grimme D2 method", &
                                        "Grimme D3 method (zero damping)", &
                                        "Grimme D3 method (Becke-Johnson damping)"), &
                          default_i_val=vdw_pairpot_dftd3)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="PARAMETER_FILE_NAME", &
                          description="Name of the parameter file, may include a path", &
                          usage="PARAMETER_FILE_NAME <FILENAME>", &
                          default_lc_val="DISPERSION_PARAMETERS")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="REFERENCE_FUNCTIONAL", &
                          description="Use parameters for this specific density functional. " &
                          //"For available D3 and D3(BJ) parameters see: " &
                          //"http://www.thch.uni-bonn.de/tc/downloads/DFT-D3/functionals.html, " &
                          //"http://www.thch.uni-bonn.de/tc/downloads/DFT-D3/functionalsbj.html", &
                          usage="REFERENCE_FUNCTIONAL <functional>", &
                          type_of_var=char_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="SCALING", &
                          description="XC Functional dependent scaling parameter, if set to zero CP2K attempts"// &
                          " to guess the xc functional that is in use and sets the associated scaling parameter.", &
                          usage="SCALING 0.2", default_r_val=0._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="EXP_PRE", &
                          description="Prefactor in exponential damping factor (DFT-D2 potential)", &
                          usage="EXP_PRE 20.", default_r_val=20._dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="EPS_CN", &
                          description="Cutoff value for coordination number function (DFT-D3 method)", &
                          usage="EPS_CN 1.e-6_dp", default_r_val=1.e-6_dp)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="D3_SCALING", &
                          description="XC Functional dependent scaling parameters (s6,sr6,s8) for the DFT-D3 method,"// &
                          " if set to zero CP2K attempts"// &
                          " to guess the xc functional from REFERENCE_FUNCTIONAL and sets the associated scaling parameter.", &
                          usage="D3_SCALING 1.0 1.0 1.0", n_var=3, &
                          default_r_vals=(/0.0_dp, 0.0_dp, 0.0_dp/))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="D3BJ_SCALING", &
                          description="XC Functional dependent scaling parameters (s6,a1,s8,a2) for the DFT-D3(BJ) method,"// &
                          " if set to zero CP2K attempts"// &
                          " to guess the xc functional from REFERENCE_FUNCTIONAL and sets the associated scaling parameter.", &
                          usage="D3BJ_SCALING 1.0 1.0 1.0 1.0", n_var=4, &
                          default_r_vals=(/0.0_dp, 0.0_dp, 0.0_dp, 0.0_dp/))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="CALCULATE_C9_TERM", &
                          description="Calculate C9 terms in DFT-D3 model", &
                          usage="CALCULATE_C9_TERM", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="REFERENCE_C9_TERM", &
                          description="Calculate C9 terms in DFT-D3 model using reference coordination numbers", &
                          usage="REFERENCE_C9_TERM", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="LONG_RANGE_CORRECTION", &
                          description="Calculate a long range correction to the DFT-D3 model", &
                          usage="LONG_RANGE_CORRECTION", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="VERBOSE_OUTPUT", &
                          description="Extensive output for the DFT-D2 and DFT-D3 models", &
                          usage="VERBOSE_OUTPUT", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      ! Set coordination numbers by atom kinds
      CALL keyword_create(keyword, name="KIND_COORDINATION_NUMBERS", &
                          description="Specifies the coordination number for a kind for the C9 term in DFT-D3.", &
                          usage="KIND_COORDINATION_NUMBERS CN kind ", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      ! Set coordination numbers by atom numbers
      CALL keyword_create(keyword, name="ATOM_COORDINATION_NUMBERS", &
                          description="Specifies the coordination number of a set of atoms for the C9 term in DFT-D3.", &
                          usage="ATOM_COORDINATION_NUMBERS CN atom1 atom2 ... ", repeats=.TRUE., &
                          n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      ! parameter specification atom by atom
      CALL keyword_create(keyword, name="ATOMPARM", &
                          description="Specifies parameters for atom types (in atomic units). If "// &
                          "not provided default parameters are used (DFT-D2).", &
                          usage="ATOMPARM <ELEMENT> <C6_parameter> <vdw_radii>", &
                          repeats=.TRUE., n_var=-1, type_of_var=char_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "PRINT_DFTD", &
                                       description="Controls the printing of some info about DFTD contributions", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! nonlocal section
      NULLIFY (subsection, keyword)
      CALL section_create(subsection, name="NON_LOCAL", &
                          description="Information on the non local dispersion functional", &
                          n_keywords=0, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="TYPE", &
                          description="Type of functional (the corresponding kernel data file should be selected)."// &
                          "Allows for common forms such as vdW-DF, vdW-DF2, optB88-vdW, rVV10.", &
                          usage="TYPE DRSLL", &
                          enum_c_vals=s2a("DRSLL", "LMKLL", "RVV10"), &
                          enum_i_vals=(/vdw_nl_DRSLL, vdw_nl_LMKLL, vdw_nl_RVV10/), &
                          enum_desc=s2a("Dion-Rydberg-Schroeder-Langreth-Lundqvist nonlocal van der Waals density functional", &
                                        "Lee-Murray-Kong-Lundqvist-Langreth nonlocal van der Waals density functional", &
                                        "Revised Vydrov-van Voorhis nonlocal van der Waals density functional"), &
                          citations=(/Tran2013/), &
                          default_i_val=vdw_nl_DRSLL)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="VERBOSE_OUTPUT", &
                          description="Extensive output for non local functionals", &
                          usage="VERBOSE_OUTPUT", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="KERNEL_FILE_NAME", &
                          description="Name of the kernel data file, may include a path."// &
                          "vdW_kernel_table.dat is for DRSLL and LMKLL and"// &
                          "rVV10_kernel_table.dat is for rVV10.", &
                          usage="KERNEL_FILE_NAME <FILENAME>", &
                          default_lc_val="vdW_kernel_table.dat")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="CUTOFF", &
                          description="The cutoff of the FFT grid used in the calculation "// &
                          "of the nonlocal vdW functional [Ry].", &
                          usage="CUTOFF 300", &
                          default_r_val=-1._dp, unit_str="Ry")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="PARAMETERS", &
                          description="Parameters b and C of the rVV10 functional", &
                          usage="PARAMETERS 6.3 0.0093", &
                          type_of_var=real_t, default_r_vals=(/6.3_dp, 0.0093_dp/), n_var=2)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_vdw_potential_section

! **************************************************************************************************
!> \brief creates the input section for the xc part
!> \param section the section to create
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_xc_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_xc_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_xc.F",1050)
      CALL section_create(section, "xc", &
                          description="Parameters needed for the calculation of the eXchange and Correlation potential", &
                          n_keywords=5, n_subsections=2, repeats=.FALSE.)

      NULLIFY (subsection, keyword)

      CALL keyword_create(keyword, name="density_cutoff", &
                          description="The cutoff on the density used by the xc calculation", &
                          usage="density_cutoff 1.e-11", default_r_val=1.0e-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="gradient_cutoff", &
                          description="The cutoff on the gradient of the density used by the "// &
                          "xc calculation", &
                          usage="gradient_cutoff 1.e-11", default_r_val=1.0e-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DENSITY_SMOOTH_CUTOFF_RANGE", &
                          description="Parameter for the smoothing procedure in"// &
                          "xc calculation", &
                          usage="gradient_cutoff {real}", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="tau_cutoff", &
                          description="The cutoff on tau used by the xc calculation", &
                          usage="tau_cutoff 1.e-11", default_r_val=1.0e-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="FUNCTIONAL_ROUTINE", &
         description="Select the code for xc calculation", &
         usage="FUNCTIONAL_ROUTINE (NEW|OLD|TEST_LSD|DEBUG)", &
         default_i_val=xc_new_f_routine, &
         enum_c_vals=s2a("NEW", "OLD", "TEST_LSD", "DEBUG"), &
         enum_i_vals=(/xc_new_f_routine, xc_old_f_routine, &
                       xc_test_lsd_f_routine, xc_debug_new_routine/), &
         enum_desc=s2a("Use new code for exchange-correlation functional calculation", &
                       "Use old code for exchange-correlation functional calculation", &
                       "Use test local-spin-density approximation code for exchange-correlation functional calculation", &
                       "Use debug new code for exchange-correlation functional calculation"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="xc_grid", & !FM to do
                          description="The xc parameters used when calculating the xc on the grid", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="xc_smooth_rho", &
                          description="The density smoothing used for the xc calculation", &
                          usage="xc_smooth_rho nn10", default_i_val=xc_rho_no_smooth, &
                          enum_c_vals=s2a("NONE", "NN50", "NN10", "SPLINE2", "NN6", "SPLINE3", "NN4"), &
                          enum_i_vals=(/xc_rho_no_smooth, xc_rho_nn50, xc_rho_nn10, &
                                        xc_rho_spline2_smooth, xc_rho_spline2_smooth, xc_rho_spline3_smooth, &
                                        xc_rho_spline3_smooth/))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="xc_deriv", &
                          description="The method used to compute the derivatives", &
                          usage="xc_deriv NN10_SMOOTH", default_i_val=xc_deriv_pw, &
                          enum_c_vals=s2a("PW", "SPLINE3", "SPLINE2", "NN50_SMOOTH", "NN10_SMOOTH", &
                                          "SPLINE2_SMOOTH", "NN6_SMOOTH", "SPLINE3_SMOOTH", "NN4_SMOOTH", "COLLOCATE"), &
                          enum_i_vals=(/xc_deriv_pw, xc_deriv_spline3, xc_deriv_spline2, &
                                        xc_deriv_nn50_smooth, xc_deriv_nn10_smooth, xc_deriv_spline2_smooth, &
                                        xc_deriv_spline2_smooth, xc_deriv_spline3_smooth, xc_deriv_spline3_smooth, &
                                        xc_deriv_collocate/))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="use_finer_grid", &
                          description="Uses a finer grid only to calculate the xc", &
                          usage="use_finer_grid", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_xc_fun_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_hfx_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mp2_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_adiabatic_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_xc_potential_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_vdw_potential_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_xc_section

! **************************************************************************************************
!> \brief creates the section for adiabatic hybrid functionals
!> \param section ...
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE create_adiabatic_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_adiabatic_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_xc.F",1172)
      CALL section_create(section, "ADIABATIC_RESCALING", &
                          description="Parameters for self interation corrected hybrid functionals", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="FUNCTIONAL_TYPE", &
         description="Which Hybrid functional should be used. "// &
         "(Has to be consistent with the definitions in XC and HF).", &
         usage="FUNCTIONAL_TYPE MCY3", &
         enum_c_vals=s2a("MCY3"), &
         enum_i_vals=(/do_adiabatic_hybrid_mcy3/), &
         enum_desc=s2a("Use MCY3 hybrid functional"), &
         default_i_val=do_adiabatic_hybrid_mcy3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="LAMBDA", &
         description="The point to be used along the adiabatic curve (0 &lt; &lambda; &lt; 1)", &
         usage="LAMBDA 0.71", &
         default_r_val=0.71_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="OMEGA", &
         description="Long-range parameter", &
         usage="OMEGA 0.2", &
         default_r_val=0.2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="FUNCTIONAL_MODEL", &
         description="Which model for the coupling constant integration should be used. ", &
         usage="FUNCTIONAL_MODEL PADE", &
         enum_c_vals=s2a("PADE"), &
         enum_i_vals=(/do_adiabatic_model_pade/), &
         enum_desc=s2a("Use pade model: W(lambda)=a+(b*lambda)/(1+c*lambda)"), &
         default_i_val=do_adiabatic_model_pade)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_adiabatic_section

END MODULE input_cp2k_xc
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief function that build the dft section of the input
!> \par History
!>      10.2005 moved out of input_cp2k [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE input_cp2k_dft
   USE bibliography,                    ONLY: &
        Andreussi2012, Avezac2005, BaniHashemian2016, Bengtsson1999, Blochl1995, Brelaz1979, &
        Dewar1977, Dewar1985, Dudarev1997, Dudarev1998, Ehrhardt1985, Elstner1998, Fattebert2002, &
        Guidon2010, Heinzmann1976, Hu2007, Hunt2003, Iannuzzi2005, Iannuzzi2006, Iannuzzi2007, &
        Kolafa2004, Krack2000, Krack2002, Kunert2003, Lippert1997, Lippert1999, Perdew1981, &
        Porezag1995, Repasky2002, Rocha2006, Schenter2008, Schiffmann2015, Seifert1996, Souza2002, &
        Stengel2009, Stewart1982, Stewart1989, Stewart2007, Thiel1992, Tozer1996, Umari2002, &
        VandeVondele2003, VandeVondele2005a, VandeVondele2005b, VandeVondele2006, Weber2008, &
        Zhao1994, Zhechkov2005
   USE cp_output_handling,              ONLY: add_last_numeric,&
                                              cp_print_key_section_create,&
                                              debug_print_level,&
                                              high_print_level,&
                                              low_print_level,&
                                              medium_print_level,&
                                              silent_print_level
   USE cp_units,                        ONLY: cp_unit_to_cp2k
   USE input_constants,                 ONLY: &
        atomic_guess, casci_canonical, cholesky_dbcsr, cholesky_inverse, cholesky_off, &
        cholesky_reduce, cholesky_restore, constant_env, core_guess, custom_env, &
        diag_block_davidson, diag_block_krylov, diag_filter_matrix, diag_ot, diag_standard, &
        dispersion_d3, dispersion_uff, do_admm_aux_exch_func_bee, do_admm_aux_exch_func_default, &
        do_admm_aux_exch_func_none, do_admm_aux_exch_func_opt, do_admm_aux_exch_func_pbex, &
        do_admm_basis_projection, do_admm_blocked_projection, do_admm_blocking_purify_full, &
        do_admm_charge_constrained_projection, do_admm_exch_scaling_merlot, &
        do_admm_exch_scaling_none, do_admm_purify_cauchy, do_admm_purify_cauchy_subspace, &
        do_admm_purify_mcweeny, do_admm_purify_mo_diag, do_admm_purify_mo_no_diag, &
        do_admm_purify_none, do_admm_purify_none_dm, do_arnoldi, do_bch, do_cn, &
        do_ddapc_constraint, do_ddapc_restraint, do_em, do_etrs, do_full_density, do_gapw_gcs, &
        do_gapw_gct, do_gapw_log, do_loc_both, do_loc_crazy, do_loc_direct, do_loc_homo, &
        do_loc_jacobi, do_loc_l1_norm_sd, do_loc_lumo, do_loc_max, do_loc_min, do_loc_none, &
        do_lri_inv, do_lri_inv_auto, do_lri_opt_all, do_lri_opt_coeff, do_lri_opt_exps, &
        do_lri_pseudoinv_diag, do_lri_pseudoinv_svd, do_method_am1, do_method_dftb, &
        do_method_gapw, do_method_gapw_xc, do_method_gpw, do_method_lrigpw, do_method_mndo, &
        do_method_mndod, do_method_ofgpw, do_method_pdg, do_method_pm3, do_method_pm6, &
        do_method_pnnl, do_method_rm1, do_method_scptb, do_pade, do_ppl_analytic, do_ppl_grid, &
        do_pwgrid_ns_fullspace, do_pwgrid_ns_halfspace, do_pwgrid_spherical, do_s2_constraint, &
        do_s2_restraint, do_se_is_kdso, do_se_is_kdso_d, do_se_is_slater, do_se_lr_ewald, &
        do_se_lr_ewald_gks, do_se_lr_ewald_r3, do_se_lr_none, do_spin_density, do_taylor, &
        ehrenfest, eri_method_full_gpw, eri_method_gpw_ht, eri_operator_coulomb, eri_operator_erf, &
        eri_operator_erfc, eri_operator_gaussian, eri_operator_yukawa, gaussian, gaussian_env, &
        general_roks, high_spin_roks, history_guess, kg_color_dsatur, kg_color_greedy, &
        kg_tnadd_atomic, kg_tnadd_embed, ls_2pnt, ls_3pnt, ls_gold, ls_none, mao_basis_ext, &
        mao_basis_orb, mao_basis_prim, mao_projection, mopac_guess, no_excitations, no_guess, &
        numerical, oe_gllb, oe_lb, oe_none, oe_saop, oe_sic, op_loc_berry, op_loc_boys, &
        op_loc_pipek, orb_dx2, orb_dxy, orb_dy2, orb_dyz, orb_dz2, orb_dzx, orb_px, orb_py, &
        orb_pz, orb_s, ot_algo_irac, ot_algo_taylor_or_diag, ot_chol_irac, ot_lwdn_irac, &
        ot_mini_broyden, ot_mini_cg, ot_mini_diis, ot_mini_sd, ot_poly_irac, ot_precond_full_all, &
        ot_precond_full_kinetic, ot_precond_full_single, ot_precond_full_single_inverse, &
        ot_precond_none, ot_precond_s_inverse, ot_precond_solver_default, &
        ot_precond_solver_direct, ot_precond_solver_inv_chol, ot_precond_solver_update, &
        outer_scf_becke_constraint, outer_scf_ddapc_constraint, outer_scf_none, &
        outer_scf_optimizer_bisect, outer_scf_optimizer_diis, outer_scf_optimizer_none, &
        outer_scf_optimizer_sd, outer_scf_s2_constraint, plus_u_lowdin, plus_u_mulliken, &
        plus_u_mulliken_charges, pw_interp, ramp_env, random_guess, real_time_propagation, &
        ref_charge_atomic, ref_charge_mulliken, rel_dkh, rel_none, rel_pot_erfc, rel_pot_full, &
        rel_sczora_mp, rel_trans_atom, rel_trans_full, rel_trans_molecule, rel_zora, &
        rel_zora_full, rel_zora_mp, restart_guess, sccs_andreussi, sccs_derivative_cd3, &
        sccs_derivative_cd5, sccs_derivative_cd7, sccs_derivative_fft, sccs_fattebert_gygi, &
        shape_function_density, shape_function_gaussian, sic_ad, sic_eo, sic_list_all, &
        sic_list_unpaired, sic_mauri_spz, sic_mauri_us, sic_none, slater, smear_energy_window, &
        smear_fermi_dirac, smear_list, sparse_guess, spline3_nopbc_interp, spline3_pbc_interp, &
        tddfpt_davidson, tddfpt_excitations, tddfpt_lanczos, tddfpt_singlet, tddfpt_triplet, &
        use_coulomb, use_diff, use_no, use_restart_wfn, use_rt_restart, use_scf_wfn, &
        wannier_projection, weight_type_mass, weight_type_unit, wfi_aspc_nr, wfi_frozen_method_nr, &
        wfi_linear_p_method_nr, wfi_linear_ps_method_nr, wfi_linear_wf_method_nr, &
        wfi_ps_method_nr, wfi_use_guess_method_nr, wfi_use_prev_p_method_nr, &
        wfi_use_prev_rho_r_method_nr, wfi_use_prev_wf_method_nr, xas_1s_type, xas_2p_type, &
        xas_2s_type, xas_dip_len2, xas_dip_vel, xas_dscf, xas_none, xas_tp_fh, xas_tp_flex, &
        xas_tp_hh, xas_tp_xfh, xas_tp_xhh, xes_tp_val
   USE input_cp2k_almo,                 ONLY: create_almo_scf_section
   USE input_cp2k_distribution,         ONLY: create_distribution_section
   USE input_cp2k_kpoints,              ONLY: create_kpoints_section
   USE input_cp2k_ls,                   ONLY: create_ls_scf_section
   USE input_cp2k_mm,                   ONLY: create_dipoles_section,&
                                              create_neighbor_lists_section
   USE input_cp2k_motion_print,         ONLY: add_format_keyword
   USE input_cp2k_poisson,              ONLY: create_poisson_section
   USE input_cp2k_rsgrid,               ONLY: create_rsgrid_section
   USE input_cp2k_transport,            ONLY: create_transport_section
   USE input_cp2k_xc,                   ONLY: create_xc_section
   USE input_keyword_types,             ONLY: keyword_create,&
                                              keyword_release,&
                                              keyword_type
   USE input_section_types,             ONLY: section_add_keyword,&
                                              section_add_subsection,&
                                              section_create,&
                                              section_release,&
                                              section_type
   USE input_val_types,                 ONLY: char_t,&
                                              integer_t,&
                                              lchar_t,&
                                              logical_t,&
                                              real_t
   USE kinds,                           ONLY: dp
   USE pw_grids,                        ONLY: do_pw_grid_blocked_false,&
                                              do_pw_grid_blocked_free,&
                                              do_pw_grid_blocked_true
   USE pw_spline_utils,                 ONLY: no_precond,&
                                              precond_spl3_1,&
                                              precond_spl3_2,&
                                              precond_spl3_3,&
                                              precond_spl3_aint,&
                                              precond_spl3_aint2
   USE qs_density_mixing_types,         ONLY: create_mixing_section
   USE qs_fb_input,                     ONLY: create_filtermatrix_section
   USE qs_mom_types,                    ONLY: create_mom_section
   USE string_utilities,                ONLY: newline,&
                                              s2a
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'input_cp2k_dft'

   PUBLIC :: create_dft_section
   PUBLIC :: create_bsse_section, create_qs_section
   PUBLIC :: create_scf_section
   PUBLIC :: create_interp_section, create_localize_section

CONTAINS

! **************************************************************************************************
!> \brief creates the dft section
!> \param section the section to be created
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_dft_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_dft_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",150)
      CALL section_create(section, name="dft", &
                          description="parameter needed by dft programs", &
                          n_keywords=3, n_subsections=4, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="BASIS_SET_FILE_NAME", &
                          description="Name of the basis set file, may include a path", &
                          usage="BASIS_SET_FILE_NAME <FILENAME>", &
                          type_of_var=lchar_t, repeats=.TRUE., &
                          default_lc_val="BASIS_SET", n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POTENTIAL_FILE_NAME", &
                          description="Name of the pseudo potential file, may include a path", &
                          usage="POTENTIAL_FILE_NAME <FILENAME>", &
                          default_lc_val="POTENTIAL")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WFN_RESTART_FILE_NAME", &
                          variants=(/"RESTART_FILE_NAME"/), &
                          description="Name of the wavefunction restart file, may include a path."// &
                          " If no file is specified, the default is to open the file as generated by the wfn restart print key.", &
                          usage="WFN_RESTART_FILE_NAME <FILENAME>", &
                          type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="UKS", &
                          variants=s2a("UNRESTRICTED_KOHN_SHAM", &
                                       "LSD", &
                                       "SPIN_POLARIZED"), &
                          description="Requests a spin-polarized calculation using alpha "// &
                          "and beta orbitals, i.e. no spin restriction is applied", &
                          usage="LSD", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, &
                          name="ROKS", &
                          variants=(/"RESTRICTED_OPEN_KOHN_SHAM"/), &
                          description="Requests a restricted open Kohn-Sham calculation", &
                          usage="ROKS", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, &
                          name="MULTIPLICITY", &
                          variants=(/"MULTIP"/), &
                          description="Two times the total spin plus one. "// &
                          "Specify 3 for a triplet, 4 for a quartet, "// &
                          "and so on. Default is 1 (singlet) for an "// &
                          "even number and 2 (doublet) for an odd number "// &
                          "of electrons.", &
                          usage="MULTIPLICITY 3", &
                          default_i_val=0) ! this default value is just a flag to get the above
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="CHARGE", &
                          description="The total charge of the system", &
                          usage="CHARGE -1", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="EXCITATIONS", &
                          description="If excitations should be calculated", &
                          usage="EXCITATIONS", &
                          enum_c_vals=s2a("NONE", "TDLR", "TDDFPT"), &
                          enum_i_vals=(/no_excitations, tddfpt_excitations, &
                                        tddfpt_excitations/), &
                          default_i_val=no_excitations)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="PLUS_U_METHOD", &
                          description="Method employed for the calculation of the DFT+U contribution", &
                          repeats=.FALSE., &
                          enum_c_vals=s2a("LOWDIN", "MULLIKEN", "MULLIKEN_CHARGES"), &
                          enum_i_vals=(/plus_u_lowdin, plus_u_mulliken, plus_u_mulliken_charges/), &
                          enum_desc=s2a("Method based on Lowdin population analysis "// &
                                        "(computationally expensive, since the diagonalization of the "// &
                                        "overlap matrix is required, but possibly more robust than Mulliken)", &
                                        "Method based on Mulliken population analysis using the net AO and "// &
                                        "overlap populations (computationally cheap method)", &
                                        "Method based on Mulliken gross orbital populations (GOP)"), &
                          n_var=1, &
                          default_i_val=plus_u_mulliken, &
                          usage="METHOD Lowdin")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="RELAX_MULTIPLICITY", &
                          variants=(/"RELAX_MULTIP"/), &
                          description="Do not enforce the occupation of alpha and beta MOs due to the initially "// &
                          "defined multiplicity, but rather follow an Aufbau principle. "// &
                          "A threshold value greater than zero activates this option. "// &
                          "Larger threshold values increase the probability for a spin flip. "// &
                          "This option is only valid for unrestricted (i.e. spin polarised) "// &
                          "Kohn-Sham (UKS) calculations.", &
                          usage="RELAX_MULTIPLICITY 0.00001", &
                          repeats=.FALSE., &
                          default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SUBCELLS", &
                          description="Read the grid size for subcell generation in the construction of "// &
                          "neighbor lists.", usage="SUBCELLS 1.5", &
                          n_var=1, default_r_val=2.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="SURFACE_DIPOLE_CORRECTION", &
                          variants=s2a("SURFACE_DIPOLE", &
                                       "SURF_DIP"), &
                          description="For slab calculations with asymmetric geometries, activate the correction of "// &
                          " the electrostatic potential with "// &
                          "by compensating for the surface dipole. Implemented only for slabs with normal "// &
                          " parallel to one Cartesian axis. The normal direction is given by the keyword SURF_DIP_DIR ", &
                          usage="SURF_DIP", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE., &
                          citations=(/Bengtsson1999/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="SURF_DIP_DIR", &
                          description="Cartesian axis parallel to surface normal.", &
                          enum_c_vals=s2a("X", "Y", "Z"), &
                          enum_i_vals=(/1, 2, 3/), &
                          enum_desc=s2a("Along x", "Along y", "Along z"), &
                          n_var=1, &
                          default_i_val=3, &
                          usage="SURF_DIP_DIR Z")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      NULLIFY (subsection)
      CALL create_scf_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_ls_scf_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_almo_scf_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_kg_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_admm_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_qs_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_tddfpt_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mgrid_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_xc_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_relativistic_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_sic_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_low_spin_roks_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_efield_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_per_efield_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_ext_pot_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_transport_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! ZMP sections to include the external density or v_xc potential
      CALL create_ext_den_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_ext_vxc_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_poisson_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_kpoints_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_implicit_solv_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_density_fitting_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_xas_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_localize_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_rtp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_print_dft_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_sccs_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_dft_section

! **************************************************************************************************
!> \brief Implicit Solvation Model
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_implicit_solv_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_implicit_solv_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      NULLIFY (keyword, subsection, print_key)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",422)
      CALL section_create(section, name="SCRF", &
                          description="Adds an implicit solvation model to the DFT calculation."// &
                          " Know also as Self Consistent Reaction Field.", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="EPS_OUT", &
                          description="Value of the dielectric constant outside the sphere", &
                          usage="EPS_OUT <REAL>", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LMAX", &
                          description="Maximum value of L used in the multipole expansion", &
                          usage="LMAX <INTEGER>", &
                          default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_sphere_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing basic info about the method", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_implicit_solv_section

! **************************************************************************************************
!> \brief Create Sphere cavity
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_sphere_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_sphere_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (keyword, subsection)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",469)
      CALL section_create(section, name="SPHERE", &
                          description="Treats the implicit solvent environment like a sphere", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="RADIUS", &
                          description="Value of the spherical cavity in the dielectric medium", &
                          usage="RADIUS <REAL>", &
                          unit_str="angstrom", &
                          type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_center_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_sphere_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_center_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_center_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",502)
      CALL section_create(section, name="CENTER", &
                          description="Defines the center of the sphere.", &
                          n_keywords=0, n_subsections=0, repeats=.FALSE.)
      CALL keyword_create(keyword, name="XYZ", &
                          description="Coordinates of the center of the sphere", &
                          usage="XYZ <REAL> <REAL> <REAL>", &
                          unit_str="angstrom", &
                          type_of_var=real_t, n_var=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOM_LIST", &
                          description="Defines a list of atoms to define the center of the sphere", &
                          usage="ATOM_LIST <INTEGER> .. <INTEGER>", &
                          type_of_var=integer_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WEIGHT_TYPE", &
                          description="Defines the weight used to define the center of the sphere"// &
                          " (if ATOM_LIST is provided)", &
                          usage="WEIGHT (UNIT|MASS)", &
                          enum_c_vals=(/"UNIT", "MASS"/), &
                          enum_i_vals=(/weight_type_unit, weight_type_mass/), &
                          default_i_val=weight_type_unit)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FIXED", &
                          description="Specify if the center of the sphere should be fixed or"// &
                          " allowed to move", &
                          usage="FIXED <LOGICAL>", &
                          default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_center_section

! **************************************************************************************************
!> \brief parameters fo the localization of wavefunctions
!> \param section ...
!> \par History
!>      03.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE create_localize_section(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_localize_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, print_section, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",556)

      NULLIFY (keyword, print_key)
      CALL section_create(section, name="LOCALIZE", &
                          description="Use one of the available methods to define the localization "// &
                          " and possibly to optimize it to a minimum or a maximum.", &
                          n_keywords=8, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the MOS localization procedure", &
                          usage="&LOCALIZE T", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ITER", &
                          description="Maximum number of iterations used for localization methods", &
                          usage="MAX_ITER 2000", default_i_val=10000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="MAX_CRAZY_ANGLE", &
         description="Largest allowed angle for the crazy rotations algorithm (smaller is slower but more stable).", &
         usage="MAX_CRAZY_ANGLE 0.1", unit_str="rad", default_r_val=0.2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CRAZY_SCALE", &
                          description="scale angles", &
                          usage="CRAZY_SCALE 0.9", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CRAZY_USE_DIAG", &
                          description="Use diagonalization (slow) or pade based calculation of matrix exponentials.", &
                          usage="CRAZY_USE_DIAG ", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="USE_HISTORY", &
         description="Generate an improved initial guess based on a history of results, which is useful during MD."// &
         "Will only work if the number of states to be localized remains constant.", &
         usage="USE_HISTORY ", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EPS_OCCUPATION", &
         description="Tolerance in the occupation number to select only fully occupied orbitals for the rotation", &
         usage="EPS_OCCUPATION 1.E-5", default_r_val=1.0E-8_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OUT_ITER_EACH", &
                          description="Every how many iterations of the localization algorithm"// &
                          "(Jacobi) the tolerance value is printed out", &
                          usage="OUT_ITER_EACH 100", default_i_val=100)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_LOCALIZATION", &
                          description="Tolerance used in the convergence criterium of the localization methods.", &
                          usage="EPS_LOCALIZATION 1.0E-2", default_r_val=1.0E-4_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIN_OR_MAX", &
                          description="Requires the maximization of the spread of the wfn", &
                          usage="MIN_OR_MAX (SPREADMIN|SPREADMAX)", &
                          enum_c_vals=(/"SPREADMIN", "SPREADMAX"/), &
                          enum_i_vals=(/do_loc_min, do_loc_max/), &
                          default_i_val=do_loc_min)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="METHOD", &
         description="Method of optimization if any", &
         usage="METHOD (JACOBI|CRAZY|DIRECT|L1SD|NONE)", &
         enum_c_vals=s2a("NONE", "JACOBI", "CRAZY", "L1SD", "DIRECT"), &
         enum_i_vals=(/do_loc_none, &
                       do_loc_jacobi, &
                       do_loc_crazy, &
                       do_loc_l1_norm_sd, &
                       do_loc_direct/), &
         enum_desc=s2a("No localization is applied", &
                       "Using 2 x 2 rotations of the orbitals, slow but robust", &
                       "A new fast method is applied, might be slightly less robust than jacobi, but usually much faster", &
                       "Steepest descent minimization of an approximate l1 norm", &
                       "Using a direct minimisation approach"), &
         default_i_val=do_loc_jacobi)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="JACOBI_FALLBACK", &
                          description="Use Jacobi method in case no convergence was achieved"// &
                          " by using the crazy rotations method.", &
                          usage="JACOBI_FALLBACK", default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART", &
                          description="Restart the localization from a set of orbitals"// &
                          " read from a localization restart file.", &
                          usage="RESTART", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LOCHOMO_RESTART_FILE_NAME", &
                          description="File name where to read the MOS from"// &
                          "which to restart the localization procedure for occupied states", &
                          usage="LOCHOMO_RESTART_FILE_NAME <FILENAME>", &
                          type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LOCLUMO_RESTART_FILE_NAME", &
                          description="File name where to read the MOS from"// &
                          "which to restart the localization procedure for unoccupied states", &
                          usage="LOCLUMO_RESTART_FILE_NAME <FILENAME>", &
                          type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OPERATOR", &
                          description="Type of opertator which defines the spread functional", &
                          usage="OPERATOR (BERRY|BOYS|PIPEK)", &
                          enum_c_vals=s2a("BERRY", "BOYS", "PIPEK"), &
                          enum_i_vals=(/op_loc_berry, op_loc_boys, op_loc_pipek/), &
                          default_i_val=op_loc_berry)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LIST", &
                          description="Indexes of the occupied wfn to be localized"// &
                          "This keyword can be repeated several times"// &
                          "(useful if you have to specify many indexes).", &
                          usage="LIST 1 2", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LIST_UNOCCUPIED", &
                          description="Indexes of the unoccupied states to be localized, "// &
                          "up to now only valid in combination with GPW. "// &
                          "This keyword has to be present if unoccupied states should be localized. "// &
                          "This keyword can be repeated several times"// &
                          "(useful if you have to specify many indexes).", &
                          usage="LIST 1 2", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STATES", &
                          description="Which states to localize, LUMO up to now only available in GPW", &
                          usage="STATES (HOMO|LUMO|ALL)", &
                          enum_c_vals=s2a("OCCUPIED", "UNOCCUPIED", "ALL"), &
                          enum_i_vals=(/do_loc_homo, do_loc_lumo, do_loc_both/), &
                          default_i_val=do_loc_homo)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="ENERGY_RANGE", &
         description="Select the orbitals to be localized within the given energy range."// &
         "This type of selection cannot be added on top of the selection through a LIST. It reads to reals that are"// &
         " lower and higher boundaries of the energy range.", &
         usage=" ENERGY_RANGE lower_bound {real}, higher_bound {real}", &
         repeats=.FALSE., &
         n_var=2, default_r_vals=(/0._dp, 0._dp/), unit_str='eV', &
         type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (print_section)
      CALL section_create(print_section, name="PRINT", &
                          description="Collects all printing options related to the Wannier centers and "// &
                          "properties computed with Wannier centers.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)
      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing basic info about the method", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(print_section, print_key)
      CALL section_release(print_key)
      ! Add printing of wannier infos
      CALL print_wanniers(print_section)
      NULLIFY (subsection)
      ! Total Dipoles with wannier
      CALL create_dipoles_section(subsection, "TOTAL_DIPOLE", debug_print_level+1)
      CALL section_add_subsection(print_section, subsection)
      CALL section_release(subsection)
      ! Molecular Dipoles with wannier
      CALL create_dipoles_section(subsection, "MOLECULAR_DIPOLES", debug_print_level+1)
      CALL section_add_subsection(print_section, subsection)
      CALL section_release(subsection)
      ! Molecular States with wannier
      CALL create_molecular_states_section(subsection)
      CALL section_add_subsection(print_section, subsection)
      CALL section_release(subsection)
      ! Wannier States with wannier
      CALL create_wannier_states_section(subsection)
      CALL section_add_subsection(print_section, subsection)
      CALL section_release(subsection)
      CALL section_add_subsection(section, print_section)
      CALL section_release(print_section)

   END SUBROUTINE create_localize_section

! **************************************************************************************************
!> \brief Controls the printing of the basic info coming from the LOCALIZE
!>        section
!> \param section ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE print_wanniers(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'print_wanniers', routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",783)
      NULLIFY (print_key, keyword)
      CALL cp_print_key_section_create(print_key, "WANNIER_CUBES", &
                                       description="Controls the printing of the wannier functions ", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUBES_LU_BOUNDS", &
                          variants=(/"CUBES_LU"/), &
                          description="The lower and upper index of the states to be printed as cube", &
                          usage="CUBES_LU_BOUNDS integer integer", &
                          n_var=2, default_i_vals=(/0, -2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUBES_LIST", &
                          description="Indexes of the states to be printed as cube files"// &
                          "This keyword can be repeated several times"// &
                          "(useful if you have to specify many indexes).", &
                          usage="CUBES_LIST 1 2", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "WANNIER_CENTERS", &
                                       description="Controls the printing of the wannier functions", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="", &
                                       unit_str="angstrom")

      CALL keyword_create(keyword, name="IONS+CENTERS", &
                          description="prints out the wannier centers together with the particles", &
                          usage="IONS+CENTERS", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL add_format_keyword(keyword, print_key, pos=.TRUE., &
                              description="Specifies the format of the output file when IONS+CENTERS is enabled.")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "WANNIER_SPREADS", &
                                       description="Controls the printing of the wannier functions", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "LOC_RESTART", &
                                       description="Controls the printing of restart file for localized MOS", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE print_wanniers

! **************************************************************************************************
!> \brief Create the print dft section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_print_dft_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_print_dft_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, sub_print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",869)
      CALL section_create(section, name="PRINT", &
                          description="Section of possible print options in DFT code.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      NULLIFY (print_key, keyword, subsection)

      CALL cp_print_key_section_create(print_key, "PROGRAM_BANNER", &
                                       description="Controls the printing of the banner of the MM program", &
                                       print_level=silent_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "KINETIC_ENERGY", &
                                       description="Controls the printing of the kinetic energy", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "DERIVATIVES", &
                                       description="Print all derivatives after the DFT calculation", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword=keyword, &
                          name="ndigits", &
                          description="Specify the number of digits used to print derivatives", &
                          default_i_val=6)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key_section=print_key, &
                                       name="neighbor_lists", &
                                       description="Controls the printing of the neighbor lists", &
                                       print_level=debug_print_level, filename="", unit_str="angstrom")
      CALL keyword_create(keyword=keyword, &
                          name="sab_orb", &
                          description="Activates the printing of the orbital "// &
                          "orbital neighbor lists, "// &
                          "i.e. the overlap neighbor lists", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sab_aux_fit", &
                          description="Activates the printing of the orbital "// &
                          "orbital neighbor lists wavefunction fitting basis, "// &
                          "i.e. the overlap neighbor lists", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sab_aux_fit_vs_orb", &
                          description="Activates the printing of the orbital "// &
                          "orbital mixed neighbor lists of wavefunction fitting basis, "// &
                          "and the orbital basis, i.e. the overlap neighbor lists", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sab_scp", &
                          description="Activates the printing of the vdW SCP "// &
                          "neighbor lists ", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sab_vdw", &
                          description="Activates the printing of the vdW "// &
                          "neighbor lists (from DFT, DFTB, SE), "// &
                          "i.e. the dispersion neighbor lists", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sab_cn", &
                          description="Activates the printing of the "// &
                          "neighbor lists used for coordination numbers in vdW DFT-D3", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sac_ae", &
                          description="Activates the printing of the orbital "// &
                          "nuclear attraction neighbor lists (erfc potential)", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sac_ppl", &
                          description="Activates the printing of the orbital "// &
                          "GTH-PPL neighbor lists (local part of the "// &
                          "Goedecker-Teter-Hutter pseudo potentials)", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sap_ppnl", &
                          description="Activates the printing of the orbital "// &
                          "GTH-PPNL neighbor lists (non-local part of the"// &
                          "Goedecker-Teter-Hutter pseudo potentials)", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sap_oce", &
                          description="Activates the printing of the orbital "// &
                          "PAW-projector neighbor lists (only GAPW)", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sab_se", &
                          description="Activates the printing of the two-center "// &
                          "neighbor lists for Coulomb type interactions in NDDO ", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sab_lrc", &
                          description="Activates the printing of the long-range SE correction "// &
                          "neighbor lists (only when doing long-range SE with integral scheme KDSO and KDSO-d)", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sab_tbe", &
                          description="Activates the printing of the DFTB Ewald "// &
                          "neighbor lists ", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sab_core", &
                          description="Activates the printing of core interaction "// &
                          "neighbor lists ", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="soo_list", &
                          description="Activates the printing of RI orbital-orbital "// &
                          "neighbor lists ", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="sip_list", &
                          description="Activates the printing of RI basis-projector interaction "// &
                          "neighbor lists ", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "SUBCELL", &
                                       description="Activates the printing of the subcells used for the"// &
                                       "generation of neighbor lists.", unit_str="angstrom", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "AO_MATRICES", &
                                       description="Controls the printing of the ao (i.e. contracted gaussian) matrices (debug).", &
                                       print_level=debug_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword=keyword, name="OMIT_HEADERS", &
                          description="Print only the matrix data, not the row and column headers", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="NDIGITS", &
                          description="Specify the number of digits used to print the AO matrices", &
                          default_i_val=6)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="CORE_HAMILTONIAN", &
                          description="If the printkey is activated controls the printing of the hamiltonian matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="DENSITY", &
                          description="If the printkey is activated controls the printing of the density (P) matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KINETIC_ENERGY", &
                          description="If the printkey is activated controls the printing of the kinetic energy matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="KOHN_SHAM_MATRIX", &
                          description="If the printkey is activated controls the printing of the kohn-sham matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create( &
         keyword, name="MATRIX_VXC", &
         description="If the printkey is activated compute and print the matrix of the exchange and correlation potential."// &
         "Only the GGA part for GPW is printed", &
         default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="ORTHO", &
                          description="If the printkey is activated controls the printing of the orthogonalization matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="OVERLAP", &
                          description="If the printkey is activated controls the printing of the overlap matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FERMI_CONTACT", &
                          description="If the printkey is activated controls the printing of the Fermi contact matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create( &
         keyword, name="PSO", &
         description="If the printkey is activated controls the printing of the paramagnetic spin-orbit matrices", &
         default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create( &
         keyword, name="EFG", &
         description="If the printkey is activated controls the printing of the electric field gradient matrices", &
         default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="POTENTIAL_ENERGY", &
                          description="If the printkey is activated controls the printing of the potential energy matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="OCE_HARD", &
                          description="If the printkey is activated controls the printing of the OCE HARD matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="OCE_SOFT", &
                          description="If the printkey is activated controls the printing of the OCE SOFT matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="W_MATRIX", &
                          description="If the printkey is activated controls the printing of the w matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="W_MATRIX_AUX_FIT", &
                          description="If the printkey is activated controls the printing of the w matrix", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="DERIVATIVES", &
                          description="If the printkey is activated controls the printing "// &
                          "of derivatives (for the matrixes that support this)", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key, "MO", &
         description="Controls the printing of the molecular orbitals. "// &
         "Note that this is only functional with diagonalization based methods, in particular not with OT (see MO_CUBES)", &
         print_level=high_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="Cartesian", &
                          description="If the printkey is activated controls the printing of the mo in the cartesian basis", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="EIGENVALUES", variants=s2a("EIGVALS"), &
                          description="If the printkey is activated controls the printing of the eigenvalues of the mos", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="EIGENVECTORS", variants=s2a("EIGVECS"), &
                          description="If the printkey is activated controls the printing of the eigenvectors of the mos", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="OCCUPATION_NUMBERS", variants=s2a("OCCNUMS"), &
                          description="If the printkey is activated controls the printing of the occupation numbers of the mos", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="NDIGITS", &
                          description="Specify the number of digits used to print the MO eigenvalues and occupation numbers", &
                          default_i_val=6)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, &
                          name="MO_INDEX_RANGE", &
                          variants=s2a("MO_RANGE", "RANGE"), &
                          description="Allows to print only a subset of the MO eigenvectors or eigenvalues. "// &
                          "The indices of the first and the last MO have to be specified", &
                          repeats=.FALSE., &
                          n_var=2, &
                          type_of_var=integer_t, &
                          default_i_vals=(/0, 0/), &
                          usage="MO_INDEX_RANGE 10 15")
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL create_mo_cubes_section(print_key)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL create_stm_section(print_key)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL create_wfn_mix_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_active_space_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="GAPW", &
                          description="Controls the printing of some gapw related information (debug).", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)
      CALL cp_print_key_section_create(print_key, "projectors", &
                                       description="If the printkey is activated controls if information on"// &
                                       " the projectors is printed.", &
                                       print_level=debug_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      CALL cp_print_key_section_create(print_key, "rho0_information", &
                                       description="If the printkey is activated controls if information on rho0 is printed.", &
                                       print_level=debug_print_level, filename="__STD_OUT__", unit_str="angstrom")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL cp_print_key_section_create(print_key, "dft_control_parameters", &
                                       description="Controls the printing of dft control parameters.", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "KPOINTS", &
                                       description="Controls the printing of kpoint information.", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "OVERLAP_CONDITION", &
                                       description="Controls the the checking and printing of and estimate "// &
                                       "of the overlap matrix condition number", &
                                       print_level=medium_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, &
                                       name="E_DENSITY_CUBE", &
                                       description="Controls the printing of cube files with "// &
                                       "the electronic density and, for LSD "// &
                                       "calculations, the spin density", &
                                       print_level=high_print_level, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, &
                          name="TOTAL_DENSITY", &
                          description="Print the total electronic density in the case "// &
                          "of a GAPW run. This keyword has only an effect, "// &
                          "if PAW atoms are present. The default is to print "// &
                          "only the soft part of the electronic density and "// &
                          "to ignore the hard part. NOTE: The total density in "// &
                          "real space might exhibit unphysical features like "// &
                          "spikes due to the finite and thus truncated g vector "// &
                          "expansion", &
                          usage="TOTAL_DENSITY {logical}", &
                          repeats=.FALSE., &
                          n_var=1, &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="XRD_INTERFACE", &
                          description="It activates the print out of exponents and coefficients for the "// &
                          " Gaussian expansion of the core densities, based on atom calculations for each kind."// &
                          " The resulting core dansities are needed to compute the form factors."// &
                          " If GAPW the local densities are also given in terms of a Gaussian expansion,"// &
                          " by fitting the difference between local-fhard and local-soft density for each atom."// &
                          " In this case the keyword TOTAL_DENSITY is set to FALSE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NGAUSS", &
                          description="Number of Gaussian functions used in the expansion of atomic (core) density", &
                          usage="NGAUSS 10", n_var=1, default_i_val=12, type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "tot_density_cube", &
                                       description="Controls printing of cube files with "// &
                                       "the total density (electrons+atomic core). Note that "// &
                                       "the value of the total density is positive where the "// &
                                       "electron density dominates and negative where the core is.", &
                                       print_level=high_print_level, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "v_hartree_cube", &
                                       description="Controls the printing of a cube file with eletrostatic "// &
                                       " potential generated by the total density (electrons+ions). It is "// &
                                       " valid only for QS with GPW formalism."// &
                                       " Note that by convention the potential has opposite sign than the expected physical one.", &
                                       print_level=high_print_level, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "external_potential_cube", &
                                       description="Controls the printing of a cube file with external "// &
                                       " potential from the DFT%EXTERNAL_POTENTIAL section only.", &
                                       print_level=high_print_level, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

! cube files for data generated by the implicit (generalized) Poisson solver
      CALL create_implicit_psolver_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

! ZMP adding the print section for  the v_xc cube
      CALL cp_print_key_section_create(print_key, "v_xc_cube", &
                                       description="Controls the printing of a cube file with xc "// &
                                       " potential generated by the ZMP method (for the moment). It is "// &
                                       " valid only for QS with GPW formalism .", &
                                       print_level=high_print_level, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "efield_cube", &
                                       description="Controls the printing of cube files with electric "// &
                                       " field generated by the total density (electrons+ions). It is "// &
                                       " valid only for QS with GPW formalism .", &
                                       print_level=high_print_level, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "ELF_CUBE", &
                                       description="Controls printing of cube files with "// &
                                       "the electron localization function (ELF). Note that "// &
                                       "the value of ELF is defined between 0 and 1:  Pauli kinetic energy density normalized "// &
                                       " by the kinetic energy density of a uniform el. gas of same density.", &
                                       print_level=high_print_level, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="density_cutoff", &
                          description=" ", &
                          usage="density_cutoff 0.0001", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=1.0e-10_dp)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL create_pdos_section(print_key)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

!
      !Printing of Moments
      CALL create_dipoles_section(print_key, "MOMENTS", high_print_level)
      CALL keyword_create( &
         keyword=keyword, &
         name="MAX_MOMENT", &
         description="Maximum moment to be calculated. Values higher than 1 not implemented under periodic boundaries.", &
         usage="MAX_MOMENT {integer}", &
         repeats=.FALSE., &
         n_var=1, &
         type_of_var=integer_t, &
         default_i_val=1)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, &
                          name="MAGNETIC", &
                          description="Calculate also magnetic moments, only implemented without periodic boundaries", &
                          usage="MAGNETIC yes", &
                          repeats=.FALSE., &
                          n_var=1, &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      ! Mulliken population analysis
      CALL cp_print_key_section_create(print_key, "MULLIKEN", &
                                       description="Controls the printing of the Mulliken (spin) population analysis", &
                                       print_level=medium_print_level, filename="__STD_OUT__", &
                                       common_iter_levels=1)
      CALL keyword_create( &
         keyword=keyword, &
         name="PRINT_GOP", &
         description="Print the gross orbital populations (GOP) in addition to the gross atomic populations (GAP) "// &
         "and net charges", &
         usage="PRINT_GOP yes", &
         repeats=.FALSE., &
         n_var=1, &
         default_l_val=.FALSE., &
         lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="PRINT_ALL", &
         description="Print all information including the full net AO and overlap population matrix", &
         usage="PRINT_ALL yes", &
         repeats=.FALSE., &
         n_var=1, &
         default_l_val=.FALSE., &
         lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      ! Lowdin population analysis (fairly expensive to compute, so only at high)
      CALL cp_print_key_section_create(print_key, "LOWDIN", &
                                       description="Controls the printing of the Lowdin (spin) population analysis", &
                                       print_level=high_print_level, filename="__STD_OUT__", &
                                       common_iter_levels=1)
      CALL keyword_create( &
         keyword=keyword, &
         name="PRINT_GOP", &
         description="Print the orbital populations in addition to the atomic populations and net charges", &
         usage="PRINT_GOP yes", &
         repeats=.FALSE., &
         n_var=1, &
         default_l_val=.FALSE., &
         lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="PRINT_ALL", &
         description="Print all information including the full symmetrically orthogonalised density matrix", &
         usage="PRINT_ALL yes", &
         repeats=.FALSE., &
         n_var=1, &
         default_l_val=.FALSE., &
         lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      ! Hirshfeld population analysis
      CALL cp_print_key_section_create(print_key, "HIRSHFELD", &
                                       description="Controls the printing of the Hirshfeld (spin) population analysis", &
                                       print_level=medium_print_level, filename="__STD_OUT__", &
                                       common_iter_levels=1)
      CALL keyword_create(keyword=keyword, name="SELF_CONSISTENT", &
                          description="Calculate charges from the Hirscheld-I (self_consistent) method."// &
                          " This scales only the full shape function, not the added charge as in the original scheme.", &
                          usage="SELF_CONSISTENT yes", repeats=.FALSE., n_var=1, &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="SHAPE_FUNCTION", &
                          description="Type of shape function used for Hirshfeld partitioning.", &
                          usage="SHAPE_FUNCTION {Gaussian,Density}", repeats=.FALSE., n_var=1, &
                          default_i_val=shape_function_gaussian, &
                          enum_c_vals=s2a("GAUSSIAN", "DENSITY"), &
                          enum_desc=s2a("Single Gaussian with Colvalent radius", &
                                        "Atomic density expanded in multiple Gaussians"), &
                          enum_i_vals=(/shape_function_gaussian, shape_function_density/))
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="REFERENCE_CHARGE", &
                          description="Charge of atomic partitioning function for Hirshfeld method.", &
                          usage="REFERENCE_CHARGE {Atomic,Mulliken}", repeats=.FALSE., n_var=1, &
                          default_i_val=ref_charge_atomic, &
                          enum_c_vals=s2a("ATOMIC", "MULLIKEN"), &
                          enum_desc=s2a("Use atomic core charges", "Calculate Mulliken charges"), &
                          enum_i_vals=(/ref_charge_atomic, ref_charge_mulliken/))
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      ! MAO (modified atomic orbital) analysis
      CALL cp_print_key_section_create(print_key, "MAO_ANALYSIS", &
                                       description="Controls the printing of the MAO (modified atomic orbital) analysis", &
                                       print_level=debug_print_level, filename="__STD_OUT__", &
                                       common_iter_levels=1, &
                                       citations=(/Heinzmann1976, Ehrhardt1985/))
      CALL keyword_create(keyword=keyword, name="EPS_FILTER", &
                          description="Threshold for matrix elements in MAO determination.", &
                          usage="EPS_FILTER reps", repeats=.FALSE., n_var=1, &
                          default_r_val=1.e-8_dp, type_of_var=real_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="REFERENCE_BASIS", &
                          description="Basis set used to construct MAO's.", &
                          usage="REFERENCE_BASIS {ORBITAL,PRIMITIVE,EXTERNAL}", repeats=.FALSE., n_var=1, &
                          default_i_val=mao_basis_orb, &
                          enum_c_vals=s2a("ORBITAL", "PRIMITIVE", "EXTERNAL"), &
                          enum_desc=s2a("Use standard orbital basis set", "Construct basis from primitives of the orbital basis", &
                                        "Read external basis (MAO)"), &
                          enum_i_vals=(/mao_basis_orb, mao_basis_prim, mao_basis_ext/))
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="PRINT_BASIS", &
                          description="Print out MAO reference basis.", &
                          usage="PRINT_BASIS {logical}", repeats=.FALSE., n_var=1, &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE., type_of_var=logical_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="EPS_GRAD", &
                          description="Threshold for gradient in MAO optimization.", &
                          usage="EPS_GRAD reps", repeats=.FALSE., n_var=1, &
                          default_r_val=1.e-4_dp, type_of_var=real_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="EPS_FUNCTION", &
                          description="Threshold for electron defect in MAO optimization.", &
                          usage="EPS_FUNCTION feps", repeats=.FALSE., n_var=1, &
                          default_r_val=1.e-3_dp, type_of_var=real_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="MAX_ITER", &
                          description="Maximum allowed iterations for MAO optimization.", &
                          usage="MAX_ITER iter", repeats=.FALSE., n_var=1, &
                          default_i_val=0, type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="NEGLECT_ABC", &
                          description="Neglect 3 atom terms in MAO analysis.", &
                          usage="NEGLECT_ABC {logical}", repeats=.FALSE., n_var=1, &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE., type_of_var=logical_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="AB_THRESHOLD", &
                          description="Threshold for printing of AB shared electron numbers.", &
                          usage="AB_THRESHOLD thr", repeats=.FALSE., n_var=1, &
                          default_r_val=1.e-2_dp, type_of_var=real_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="ABC_THRESHOLD", &
                          description="Threshold for printing of ABC shared electron numbers.", &
                          usage="ABC_THRESHOLD thr", repeats=.FALSE., n_var=1, &
                          default_r_val=1.e-5_dp, type_of_var=real_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword=keyword, name="ANALYZE_UNASSIGNED_CHARGE", &
                          description="Calculate atomic contributions to the unassigned charge.", &
                          usage="ANALYZE_UNASSIGNED_CHARGE {logical}", repeats=.FALSE., n_var=1, &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE., type_of_var=logical_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      ! Xray diffraction
      CALL cp_print_key_section_create( &
         print_key_section=print_key, &
         name="XRAY_DIFFRACTION_SPECTRUM", &
         description="Calculate and print the coherent X-ray "// &
         "diffraction spectrum", &
         print_level=debug_print_level, &
         filename="", &
         citations=(/Krack2000, Krack2002/))
      CALL keyword_create( &
         keyword=keyword, &
         name="Q_MAX", &
         variants=(/"Q_MAXIMUM"/), &
         description="Maximum Q value calculated for the spectrum", &
         usage="Q_MAX {real}", &
         repeats=.FALSE., &
         n_var=1, &
         type_of_var=real_t, &
         default_r_val=cp_unit_to_cp2k(value=20.0_dp, &
                                       unit_str="angstrom^-1"), &
         unit_str="angstrom^-1")
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key_section=print_key, &
                                       name="ELECTRIC_FIELD_GRADIENT", &
                                       description="Calculate and print the electric field gradients "// &
                                       "at atomic positions", &
                                       print_level=debug_print_level, &
                                       filename="__STD_OUT__")

      CALL keyword_create(keyword=keyword, &
                          name="INTERPOLATION", &
                          description="Use interpolation method from real space grid", &
                          usage="INTERPOLATION {logical}", &
                          repeats=.FALSE., &
                          n_var=1, &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="GSPACE_SMOOTHING", &
                          description="Use a G-space smoothing function", &
                          usage="GSPACE_SMOOTHING cutoff {real}, width {real}", &
                          repeats=.FALSE., &
                          n_var=2, default_r_vals=(/-1._dp, -1._dp/), &
                          type_of_var=real_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="DEBUG", &
                          description="Print additional debug output", &
                          usage="DEBUG {logical}", &
                          repeats=.FALSE., &
                          n_var=1, &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL create_gspace_interp_section(subsection)
      CALL section_add_subsection(print_key, subsection)
      CALL section_release(subsection)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key_section=print_key, &
                                       name="BASIS_MOLOPT_QUANTITIES", &
                                       description="Print the two quantities needed in the basis molopt generation:"// &
                                       " total energy and condition number of the overlap matrix (S matrix)", &
                                       print_level=debug_print_level, &
                                       filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key_section=print_key, &
                                       name="HYPERFINE_COUPLING_TENSOR", &
                                       description="Calculate and print the EPR hyperfine coupling tensor"// &
                                       " at atomic positions", &
                                       print_level=debug_print_level, &
                                       filename="__STD_OUT__")

      CALL keyword_create(keyword=keyword, &
                          name="INTERACTION_RADIUS", &
                          description="Radius of interaction for EPR hyperfine tensor calculation", &
                          usage="INTERACTION_RADIUS radius {real}", &
                          repeats=.FALSE., &
                          n_var=1, default_r_val=10._dp, &
                          type_of_var=real_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key_section=print_key, &
                                       name="OPTIMIZE_LRI_BASIS", &
                                       description="Optimize the exponents of the LRI basis set", &
                                       print_level=low_print_level, &
                                       filename="OPTIMIZED_LRI_BASIS")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key_section=print_key, &
         name="PLUS_U", &
         description="Controls the printing for the DFT+U methods", &
         print_level=high_print_level, &
         filename="__STD_OUT__", &
         each_iter_names=s2a("QS_SCF"), &
         each_iter_values=(/0/), &
         citations=(/Dudarev1997, Dudarev1998/))
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key_section=print_key, &
         name="SCCS", &
         description="Controls the printing for the SCCS models", &
         print_level=high_print_level, &
         filename="__STD_OUT__", &
         each_iter_names=s2a("QS_SCF"), &
         each_iter_values=(/0/), &
         citations=(/Fattebert2002, Andreussi2012/))

      NULLIFY (sub_print_key)

      CALL cp_print_key_section_create( &
         print_key_section=sub_print_key, &
         name="DENSITY_GRADIENT", &
         description="Controls the printing of the cube files with "// &
         "the norm of the density gradient |&nabla;&rho;| "// &
         "used by the SCCS model.", &
         print_level=debug_print_level, &
         filename="", &
         each_iter_names=s2a("QS_SCF"), &
         each_iter_values=(/0/))
      CALL keyword_create(keyword, name="STRIDE", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 "// &
                          "numbers (for X,Y,Z) or 1 number valid for all components.", &
                          n_var=-1, &
                          default_i_vals=(/2, 2, 2/), &
                          type_of_var=integer_t, &
                          repeats=.FALSE.)
      CALL section_add_keyword(sub_print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="Append the cube files when they already exist", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE., &
                          repeats=.FALSE.)
      CALL section_add_keyword(sub_print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(print_key, sub_print_key)
      CALL section_release(sub_print_key)

      CALL cp_print_key_section_create( &
         print_key_section=sub_print_key, &
         name="DIELECTRIC_FUNCTION", &
         description="Controls the printing of the cube files with "// &
         "the dielectric function used by the SCCS model. "// &
         "This function determines the cavity formed by a solute in "// &
         "a solvent and thus it can be used for the visualisaton of the cavity.", &
         print_level=debug_print_level, &
         filename="", &
         each_iter_names=s2a("QS_SCF"), &
         each_iter_values=(/0/), &
         citations=(/Fattebert2002, Andreussi2012/))
      CALL keyword_create(keyword, name="STRIDE", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 "// &
                          "numbers (for X,Y,Z) or 1 number valid for all components.", &
                          n_var=-1, &
                          default_i_vals=(/2, 2, 2/), &
                          type_of_var=integer_t, &
                          repeats=.FALSE.)
      CALL section_add_keyword(sub_print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="Append the cube files when they already exist", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE., &
                          repeats=.FALSE.)
      CALL section_add_keyword(sub_print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(print_key, sub_print_key)
      CALL section_release(sub_print_key)

      CALL cp_print_key_section_create( &
         print_key_section=sub_print_key, &
         name="POLARISATION_POTENTIAL", &
         description="Controls the printing of the cube files with the "// &
         "polarisation potential &phi;<sup>pol</sup> used by the SCCS model with "// &
         "total potential &phi;<sup>tot</sup> = &phi;<sup>sol</sup> + &phi;<sup>pol</sup>", &
         print_level=debug_print_level, &
         filename="", &
         each_iter_names=s2a("QS_SCF"), &
         each_iter_values=(/0/), &
         citations=(/Fattebert2002, Andreussi2012/))
      CALL keyword_create(keyword, name="STRIDE", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 "// &
                          "numbers (for X,Y,Z) or 1 number valid for all components.", &
                          n_var=-1, &
                          default_i_vals=(/2, 2, 2/), &
                          type_of_var=integer_t, &
                          repeats=.FALSE.)
      CALL section_add_keyword(sub_print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="Append the cube files when they already exist", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE., &
                          repeats=.FALSE.)
      CALL section_add_keyword(sub_print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(print_key, sub_print_key)
      CALL section_release(sub_print_key)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_print_dft_section

! **************************************************************************************************
!> \brief creates the input section for dealing with homo lumos, including dumping cubes
!> \param print_key ...
! **************************************************************************************************
   SUBROUTINE create_mo_cubes_section(print_key)
      TYPE(section_type), POINTER                        :: print_key

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mo_cubes_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)

      CALL cp_print_key_section_create(print_key, "MO_CUBES", &
                                       description="Controls the printing of cubes of the molecular orbitals.", &
                                       print_level=high_print_level, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="write_cube", &
                          description="If the MO cube file should be written. If false, the eigenvalues are still computed."// &
                          " Can also be useful in combination with STM calculations", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="nlumo", &
                          description="If the printkey is activated controls the number of lumos"// &
                          " that are printed and dumped as a cube (-1=all)", &
                          default_i_val=0)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create( &
         keyword, name="nhomo", &
         description="If the printkey is activated controls the number of homos that dumped as a cube (-1=all),"// &
         " eigenvalues are always all dumped", &
         default_i_val=1)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_mo_cubes_section

! **************************************************************************************************
!> \brief ...
!> \param print_key ...
! **************************************************************************************************
   SUBROUTINE create_pdos_section(print_key)

      TYPE(section_type), POINTER                        :: print_key

      CHARACTER(len=*), PARAMETER :: routineN = 'create_pdos_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (subsection)
      NULLIFY (keyword)

      CALL cp_print_key_section_create(print_key, "PDOS", &
                                       description="Print out the DOS projected per kind and per angular momentum  ", &
                                       print_level=debug_print_level, common_iter_levels=1, filename="")
      CALL keyword_create(keyword, name="COMPONENTS", &
                          description="Print out pdos distinguishing all angular momentum components.", &
                          usage="COMPONENTS", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="Append the pdos obtained at different iterations to the pdos  output file."// &
                          "By defaut the file is overwritten", &
                          usage="APPEND", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create( &
         keyword, name="NLUMO", &
         description="Number of virtual orbitals to be added to the MO set (-1=all)."//newline// &
         "CAUTION: Setting this value to be higher than the number of states present may cause a Cholesky error.", &
         usage="NLUMO integer", default_i_val=0)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="OUT_EACH_MO", &
                          description="Output on the status of the calculation every OUT_EACH_MO states. If -1 no output", &
                          usage="OUT_EACH_MO integer", default_i_val=-1)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
!
      CALL section_create(subsection, name="LDOS", &
                          description="Controls the printing of local PDOS, projected on subsets"// &
                          " of atoms given through lists", &
                          n_keywords=4, n_subsections=0, repeats=.TRUE.)
      CALL keyword_create(keyword, name="COMPONENTS", &
                          description="Print out pdos distinguishing all angular momentum components.", &
                          usage="COMPONENTS", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LIST", &
                          description="Specifies a list of indexes of atoms where to project the DOS  ", &
                          usage="LIST {integer}  {integer} ..  {integer} ", type_of_var=integer_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(print_key, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="R_LDOS", &
                          description="Controls the printing of local PDOS, projected on 3D volume in real space,"// &
                          " the volume is defined in terms of position with respect to atoms in the lists", &
                          n_keywords=4, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="LIST", &
                          description="Specifies a list of indexes of atoms used to define the real space volume  ", &
                          usage="LIST {integer}  {integer} ..  {integer} ", type_of_var=integer_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="XRANGE", &
                          description="range of positions in Cartesian direction x: all grid points within "// &
                          " this range from at least one atom of the list are considered", &
                          usage="XRANGE -10.0 10.0", unit_str="angstrom", n_var=2, type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="YRANGE", &
                          description="range of positions in Cartesian direction y: all grid points within "// &
                          " this range from at least one atom of the list are considered", &
                          usage="YRANGE -10.0 10.0", unit_str="angstrom", n_var=2, type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="ZRANGE", &
                          description="range of positions in Cartesian direction z: all grid points within "// &
                          " this range from at least one atom of the list are considered", &
                          usage="ZRANGE -10.0 10.0", unit_str="angstrom", n_var=2, type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ERANGE", &
                          description="only project states with the eigenvalues in the given interval. "// &
                          "Default is all states.", &
                          usage="ERANGE -1.0 1.0", unit_str="hartree", n_var=2, type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(print_key, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_pdos_section

! **************************************************************************************************
!> \brief ...
!> \param print_key ...
! **************************************************************************************************
   SUBROUTINE create_stm_section(print_key)
      TYPE(section_type), POINTER                        :: print_key

      CHARACTER(len=*), PARAMETER :: routineN = 'create_stm_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)

      CALL cp_print_key_section_create(print_key, "STM", &
                                       description="Controls the printing of cubes for the generation of STM images.", &
                                       print_level=debug_print_level, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="nlumo", &
                          description="If the printkey is activated controls the number of additional lumos"// &
                          " that are computed to be able to reproduce STM images obtained"// &
                          "  from positive bias (imaging unoccupied states)", &
                          default_i_val=0)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BIAS", &
                          description="Bias energy for scanning tunneling microscopy (STM) image generation."// &
                          "Orbital densities are summed according to the bias energy."// &
                          "For negative values, states in the range ]EF+bias,EF] are summed,"// &
                          "While positive values sum states in the range [EF,EF+bias[."// &
                          "If postive biases are used, sufficiently many unoccupied stated"// &
                          " (see ADDED_MOS and NLUMO ) should be computed.", &
                          n_var=-1, type_of_var=real_t, default_r_vals=(/0.0_dp/), unit_str='eV')
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TH_TORB", &
                          description="Tip orbital symmetry in Tersoff-Hamann approximation to compute STM images", &
                          repeats=.TRUE., &
                          default_i_val=orb_s, &
                          usage="TH_TORB s dz2", &
                          enum_c_vals=s2a("S", "PX", "PY", "PZ", "DXY", "DYZ", "DZX", "DX2", "DY2", "DZ2"), &
                          enum_i_vals=(/orb_s, orb_px, orb_py, orb_pz, orb_dxy, orb_dyz, orb_dzx, orb_dx2, orb_dy2, orb_dz2/), &
                          enum_desc=s2a("s orbital", "px orbital", "py orbital", "pz orbital", &
                                        "dxy orbital", "dyz orbital", "dzx orbital", "x^2 orbital", "y^2 orbital", "z^2 orbital"))
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REF_ENERGY", &
                          description="By default the reference energy is the Fermi energy. In order to compare"// &
                          " with STS experiments, where specific energy ranges are addressed, here"// &
                          "  one can set a different reference energy."// &
                          " The energy range is anyway controlled by the BIAS", &
                          type_of_var=real_t, default_r_val=0.0_dp, unit_str='eV')
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_stm_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_wfn_mix_section(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_wfn_mix_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (subsection)
      NULLIFY (keyword)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",2147)

      CALL section_create(section, name="WFN_MIX", &
                          description="A section that allows manipulation of the MO coeffs,"// &
                          " e.g. for changing a ground state into an excited state."// &
                          "Starting from a copy of the original MOs, changes can be made"// &
                          "by adding linear combinations of HOMO/LUMO of the original MOs to the result MOs", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="OVERWRITE_MOS", &
                          description="If the keyword is active molecular orbitals in memory will be replaced by the mixed wfn."// &
                          " In combination with RTP or EMD no restart will be required to use the mixed wfn.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="UPDATE", &
                          description="update a result MO with with a linear combination of of original MOs."// &
                          " This section can be repeated to build arbitrary linear combinations using repeatedly y=a*y+b*x.", &
                          n_keywords=1, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="RESULT_MO_INDEX", &
                          description="Index of the MO (y) to be modified. Counting down in energy with HOMO=1", &
                          usage="RESULT_MO_INDEX 1", type_of_var=integer_t, default_i_val=0)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESULT_MARKED_STATE", &
                          description="Specifies the MO according to "// &
                          "the marks set in MOLECULAR_STATES. The value corresponds to the repetition "// &
                          " of MARK_STATES in MOLECULAR_STATES", &
                          usage="ORIG_MARKED_STATE 1", type_of_var=integer_t, default_i_val=0)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESULT_SPIN_INDEX", &
                          description="Spin of the MO (y) to be modified.", &
                          enum_c_vals=s2a("Alpha", "Beta"), &
                          enum_i_vals=(/1, 2/), & ! direct index in array
                          default_i_val=1, &
                          enum_desc=s2a("Majority spin", "Minority spin"))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESULT_SCALE", &
                          description="Scaling factor of the result variable (a).", &
                          usage="RESULT_SCALE 0.0", type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ORIG_MO_INDEX", &
                          description="Index of the original MO (x). "// &
                          "Counting down in energy with HOMO=1 or up from LUMO=1, depending on ORIG_IS_VIRTUAL.", &
                          usage="ORIG_MO_INDEX 1", type_of_var=integer_t, default_i_val=0)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ORIG_MARKED_STATE", &
                          description="Specifies the MO according to "// &
                          "the marks set in MOLECULAR_STATES. The value corresponds to the repetition "// &
                          " of MARK_STATES in MOLECULAR_STATES", &
                          usage="ORIG_MARKED_STATE 1", type_of_var=integer_t, default_i_val=0)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ORIG_SPIN_INDEX", &
                          description="Spin of the MO (x) to be modified.", &
                          enum_c_vals=s2a("Alpha", "Beta"), &
                          enum_i_vals=(/1, 2/), & ! direct index in array
                          default_i_val=1, &
                          enum_desc=s2a("Majority spin", "Minority spin"))
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ORIG_SCALE", &
                          description="Scaling factor of the result variable (b).", &
                          usage="ORIG_SCALE 0.0", type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ORIG_IS_VIRTUAL", &
                          description="The original MO (x) is a LUMO.", &
                          usage="ORIG_IS_VIRTUAL", type_of_var=logical_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_wfn_mix_section

! **************************************************************************************************
!> \brief creates the input section for the molecular states
!> \param print_key ...
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_molecular_states_section(print_key)
      TYPE(section_type), POINTER                        :: print_key

      CHARACTER(len=*), PARAMETER :: routineN = 'create_molecular_states_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key2

      IF(.NOT.(.NOT. ASSOCIATED(print_key)))CALL cp__a("input_cp2k_dft.F",2252)
      NULLIFY (print_key2, keyword)
      CALL cp_print_key_section_create(print_key, "MOLECULAR_STATES", &
                                       description="Controls printing of molecular states ", &
                                       print_level=high_print_level, filename=" ", citations=(/Hunt2003/))

      CALL keyword_create( &
         keyword, name="CUBE_EVAL_RANGE", &
         description="only write cubes if the eigenvalues of the corresponding molecular states lie in the given interval. "// &
         "Default is all states.", &
         usage="CUBE_EVAL_RANGE -1.0 1.0", unit_str="hartree", n_var=2, type_of_var=real_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MARK_STATES", &
                          description="Can be used to mark given molecular states."// &
                          " Sets a mark to both, occupied and unoccupied states. "// &
                          "Occupied states are counted beginning with HOMO=1, "// &
                          "unoccupied states are counted beginning with LUMO=1, "// &
                          "This is only meaningful in combination with WFN_MIX. "// &
                          "First integer specifies the molecule, second integer specifies the state.", &
                          usage="MARK_STATES integer integer", &
                          n_var=2, default_i_vals=(/-1, -1/), type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key2, "cubes", &
                                       description="Controls the printing of cube files", &
                                       print_level=high_print_level, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key2, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(print_key, print_key2)
      CALL section_release(print_key2)
   END SUBROUTINE create_molecular_states_section

! **************************************************************************************************
!> \brief ...
!> \param print_key ...
! **************************************************************************************************
   SUBROUTINE create_wannier_states_section(print_key)
      TYPE(section_type), POINTER                        :: print_key

      CHARACTER(len=*), PARAMETER :: routineN = 'create_wannier_states_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key2

      IF(.NOT.(.NOT. ASSOCIATED(print_key)))CALL cp__a("input_cp2k_dft.F",2305)
      NULLIFY (print_key2, keyword)
      CALL cp_print_key_section_create(print_key, "WANNIER_STATES", &
                                       description="Controls printing of molecular states ", &
                                       print_level=high_print_level, filename=" ")

      CALL keyword_create( &
         keyword, name="CUBE_EVAL_RANGE", &
         description="only write cubes if the eigenvalues of the corresponding molecular states lie in the given interval. "// &
         "Default is all states.", &
         usage="CUBE_EVAL_RANGE -1.0 1.0", unit_str="hartree", n_var=2, type_of_var=real_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MARK_STATES", &
                          description="Can be used to mark given molecular states."// &
                          " Sets a mark to both, occupied and unoccupied states. "// &
                          "Occupied states are counted beginning with HOMO=1, "// &
                          "unoccupied states are counted beginning with LUMO=1, "// &
                          "This is only meaningful in combination with WFN_MIX. "// &
                          "First integer specifies the molecule, second integer specifies the state.", &
                          usage="MARK_STATES integer integer", &
                          n_var=2, default_i_vals=(/-1, -1/), type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key2, "cubes", &
                                       description="Controls the printing of cube files", &
                                       print_level=high_print_level, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key2, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(print_key, print_key2)
      CALL section_release(print_key2)
   END SUBROUTINE create_wannier_states_section

! **************************************************************************************************
!> \brief creates the input section for the qs part
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_qs_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_qs_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",2359)
      CALL section_create(section, "qs", &
                          description="parameters needed to set up the Quickstep framework", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      ! Reals
      CALL keyword_create(keyword, name="EPS_DEFAULT", &
                          description="Try setting all EPS_xxx to values leading to an energy correct up to EPS_DEFAULT", &
                          usage="EPS_DEFAULT real", default_r_val=1.0E-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_CORE_CHARGE", &
                          description="Precision for mapping the core charges.Overrides EPS_DEFAULT/100.0 value", &
                          usage="EPS_CORE_CHARGE real", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EPS_GVG_RSPACE", &
         variants=(/"EPS_GVG"/), &
         description="Sets precision of the realspace KS matrix element integration. Overrides SQRT(EPS_DEFAULT) value", &
         usage="EPS_GVG_RSPACE real", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_PGF_ORB", &
                          description="Sets precision of the overlap matrix elements. Overrides SQRT(EPS_DEFAULT) value", &
                          usage="EPS_PGF_ORB real", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EPS_KG_ORB", &
         description="Sets precision used in coloring the subsets for the Kim-Gordon method. Overrides SQRT(EPS_DEFAULT) value", &
         usage="EPS_KG_ORB 1.0E-8", &
         type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_PPL", &
                          description="Adjusts the precision for the local part of the pseudo potential. ", &
                          usage="EPS_PPL real", type_of_var=real_t, default_r_val=1.0E-2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EPS_PPNL", &
         description="Sets precision of the non-local part of the pseudo potential. Overrides sqrt(EPS_DEFAULT) value", &
         usage="EPS_PPNL real", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_CPC", &
                          description="Sets precision of the GAPW projection. Overrides EPS_DEFAULT value", &
                          usage="EPS_CPC real", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_RHO", &
                          description="Sets precision of the density mapping on the grids.Overrides EPS_DEFAULT value", &
                          usage="EPS_RHO real", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_RHO_RSPACE", &
                          description="Sets precision of the density mapping in rspace.Overrides EPS_DEFAULT value."// &
                          ".Overrides EPS_RHO value", &
                          usage="EPS_RHO_RSPACE real", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_RHO_GSPACE", &
                          description="Sets precision of the density mapping in gspace.Overrides EPS_DEFAULT value."// &
                          ".Overrides EPS_RHO value", &
                          usage="EPS_RHO_GSPACE real", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_FILTER_MATRIX", &
                          description="Sets the threshold for filtering matrix elements.", &
                          usage="EPS_FILTER_MATRIX 1.0E-6", type_of_var=real_t, default_r_val=0.0E0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPSFIT", &
                          variants=(/"EPS_FIT"/), &
                          description="GAPW: precision to give the extention of a hard gaussian ", &
                          usage="EPSFIT real", default_r_val=1.0E-4_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPSISO", &
                          variants=(/"EPS_ISO"/), &
                          description="GAPW: precision to determine an isolated projector", &
                          usage="EPSISO real", default_r_val=1.0E-12_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPSSVD", &
                          variants=(/"EPS_SVD"/), &
                          description="GAPW: tolerance used in the singular value decomposition of the projector matrix", &
                          usage="EPS_SVD real", default_r_val=1.0E-8_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPSRHO0", &
                          variants=s2a("EPSVRHO0", "EPS_VRHO0"), &
                          description="GAPW : precision to determine the range of V(rho0-rho0soft)", &
                          usage="EPSRHO0 real", default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALPHA0_HARD", &
                          variants=s2a("ALPHA0_H", "ALPHA0"), &
                          description="GAPW: Exponent for hard compensation charge", &
                          usage="ALPHA0_HARD real", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="FORCE_PAW", &
         description="Use the GAPW scheme also for atoms with soft basis sets, i.e. "// &
         " the local densities are computed even if hard and soft should be equal. "// &
         "If this keyword is not set to true, those atoms with soft basis sets are treated by a GPW scheme, i.e. "// &
         "the corresponding density contribution goes on the global grid and is expanded in PW. "// &
         " This option nullifies the effect of the GPW_TYPE in the atomic KIND", &
         usage="FORCE_PAW", &
         default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_RAD_LOCAL", &
                          description="GAPW : maximum radius of gaussian functions"// &
                          " included in the generation of projectors", &
                          usage="MAX_RAD_LOCAL real", default_r_val=25.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Logicals
      CALL keyword_create(keyword, name="LS_SCF", &
                          description="Perform a linear scaling SCF", &
                          usage="LS_SCF", lone_keyword_l_val=.TRUE., &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALMO_SCF", &
                          description="Perform ALMO SCF", &
                          usage="ALMO_SCF", lone_keyword_l_val=.TRUE., &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TRANSPORT", &
                          description="Perform transport calculations (coupling CP2K and OMEN)", &
                          usage="TRANSPORT", lone_keyword_l_val=.TRUE., &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="KG_METHOD", &
                          description="Use a Kim-Gordon-like scheme.", &
                          usage="KG_METHOD", lone_keyword_l_val=.TRUE., &
                          default_l_val=.FALSE., citations=(/Iannuzzi2006, Brelaz1979/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAP_CONSISTENT", &
                          description="Compute the exact derivative (Hks) of the energy with respect to the density matrix. "// &
                          "This is slightly more expensive than using an approximate computation, "// &
                          "but consistent mapping can improve the stability of the SCF procedure, "// &
                          "especially for a tight EPS_SCF and a less tight EPS_DEFAULT.", &
                          usage="MAP_CONSISTENT FALSE", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Integers
      CALL keyword_create(keyword, name="LMAXN1", &
                          variants=(/"LMAXRHO1"/), &
                          description="GAPW : max L number for expansion of the atomic densities in spherical gaussians", &
                          usage="LMAXN1 integer", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LMAXN0", &
                          variants=(/"LMAXRHO0"/), &
                          description="GAPW : max L number for the expansion compensation densities in spherical gaussians", &
                          usage="LMAXN0 integer", &
                          default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LADDN0", &
                          description="GAPW : integer added to the max L of the basis set, used to determine the "// &
                          "maximum value of L for the compensation charge density.", &
                          usage="LADDN0 integer", &
                          default_i_val=99)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Characters
      CALL keyword_create(keyword, name="QUADRATURE", &
                          description="GAPW: algorithm to construct the atomic radial grids", &
                          usage="QUADRATURE GC_SIMPLE", &
                          enum_c_vals=s2a("GC_SIMPLE", "GC_TRANSFORMED", "GC_LOG"), &
                          enum_i_vals=(/do_gapw_gcs, do_gapw_gct, do_gapw_log/), &
                          enum_desc=s2a("Gauss-Chebyshev quadrature", &
                                        "Transformed Gauss-Chebyshev quadrature", &
                                        "Logarithmic transformed Gauss-Chebyshev quadrature"), &
                          default_i_val=do_gapw_log)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PW_GRID", &
                          description="What kind of PW_GRID should be employed", &
                          usage="PW_GRID NS-FULLSPACE", &
                          enum_c_vals=s2a("SPHERICAL", "NS-FULLSPACE", "NS-HALFSPACE"), &
                          enum_desc=s2a("- not tested", " tested", " - not tested"), &
                          enum_i_vals=(/do_pwgrid_spherical, do_pwgrid_ns_fullspace, do_pwgrid_ns_halfspace/), &
                          default_i_val=do_pwgrid_ns_fullspace)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PW_GRID_LAYOUT", &
                          description="Force a particular real-space layout for the plane waves grids. "// &
                          "Numbers &le; 0 mean that this dimension is free, incorrect layouts will be ignored. "// &
                          "The default (/-1,-1/) causes CP2K to select a good value, "// &
                          "i.e. plane distributed for large grids, more general distribution for small grids.", &
                          usage="PW_GRID_LAYOUT 4 16", &
                          repeats=.FALSE., n_var=2, &
                          default_i_vals=(/-1, -1/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PW_GRID_BLOCKED", &
                          description="Can be used to set the distribution in g-space for the pw grids and their FFT.", &
                          usage="PW_GRID_BLOCKED FREE", &
                          enum_c_vals=s2a("FREE", "TRUE", "FALSE"), &
                          enum_desc=s2a("CP2K will select an appropriate value", "blocked", "not blocked"), &
                          enum_i_vals=(/do_pw_grid_blocked_free, do_pw_grid_blocked_true, do_pw_grid_blocked_false/), &
                          default_i_val=do_pw_grid_blocked_free)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EXTRAPOLATION", &
         variants=s2a("INTERPOLATION", "WF_INTERPOLATION"), &
         description="Extrapolation strategy for the wavefunction during e.g. MD."// &
         "PS and ASPC are recommended, see also EXTRAPOLATION_ORDER.", &
         citations=(/Kolafa2004, VandeVondele2005a/), &
         usage="EXTRAPOLATION PS", &
         enum_c_vals=s2a("USE_GUESS", "USE_PREV_P", "USE_PREV_RHO_R", "LINEAR_WF", &
                         "LINEAR_P", "LINEAR_PS", "USE_PREV_WF", "PS", "FROZEN", "ASPC"), &
         enum_desc=s2a( &
         "Use the method specified with SCF_GUESS, i.e. no extrapolation", &
         "Use the previous density matrix", &
         "Use the previous density in real space", &
         "Linear extrapolation of the wavefunction", &
         "Linear extrapolation of the density matrix", &
         "Linear extrapolation of the density matrix times the overlap matrix", &
         "Use the previous wavefunction", &
         "Higher order extrapolation of the density matrix times the overlap matrix", &
         "Frozen ...", &
         "Always stable predictor corrector, similar to PS, but going for MD stability instead of intial guess accuracy."), &
         enum_i_vals=(/ &
         wfi_use_guess_method_nr, &
         wfi_use_prev_p_method_nr, &
         wfi_use_prev_rho_r_method_nr, &
         wfi_linear_wf_method_nr, &
         wfi_linear_p_method_nr, &
         wfi_linear_ps_method_nr, &
         wfi_use_prev_wf_method_nr, &
         wfi_ps_method_nr, &
         wfi_frozen_method_nr, &
         wfi_aspc_nr/), &
         default_i_val=wfi_aspc_nr)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXTRAPOLATION_ORDER", &
                          description="Order for the PS or ASPC extrapolation (typically 2-4). "// &
                          "Higher order might bring more accuracy, but comes, "// &
                          "for large systems, also at some cost. "// &
                          "In some cases, a high order extrapolation is not stable,"// &
                          " and the order needs to be reduced.", &
                          usage="EXTRAPOLATION_ORDER {integer}", default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="METHOD", &
                          description="Specifies the electronic structure method that should be employed", &
                          usage="METHOD GAPW", &
                          enum_c_vals=s2a("GAPW", "GAPW_XC", "GPW", "LRIGPW", &
                                          "MNDO", "MNDOD", "AM1", "PM3", "PM6", "PDG", "RM1", "PNNL", "DFTB", "OFGPW", "SCPTB"), &
                          enum_desc=s2a("Gaussian and augmented plane waves method", &
                                        "Gaussian and augmented plane waves method only for XC", &
                                        "Gaussian and plane waves method", &
                                        "Local resolution of identity method", &
                                        "MNDO semiempirical", "MNDO-d semiempirical", "AM1 semiempirical", &
                                        "PM3 semiempirical", "PM6 semiempirical", "PDG semiempirical", &
                                        "RM1 semiempirical", &
                                        "PNNL semiempirical", &
                                        "DFTB Density Functional based Tight-Binding", &
                                        "OFGPW Orbital-free GPW method", &
                                        "SCPTB Self-Consistent-Polarization Tight-Binding"), &
                          enum_i_vals=(/do_method_gapw, do_method_gapw_xc, do_method_gpw, do_method_lrigpw, &
                                        do_method_mndo, do_method_mndod, do_method_am1, do_method_pm3, &
                                        do_method_pm6, do_method_pdg, do_method_rm1, do_method_pnnl, do_method_dftb, &
                                        do_method_ofgpw, do_method_scptb/), &
                          citations=(/Lippert1997, Lippert1999, Krack2000, VandeVondele2005a, &
                                      VandeVondele2006, Dewar1977, Dewar1985, Rocha2006, Stewart1989, Thiel1992, &
                                      Repasky2002, Stewart2007, Schenter2008/), &
                          default_i_val=do_method_gpw)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CORE_PPL", &
                          description="Specifies the method used to calculate the local pseudopotential contribution.", &
                          usage="CORE_PPL ANALYTIC", &
                          enum_c_vals=s2a("ANALYTIC", "GRID"), &
                          enum_desc=s2a("Analytic integration of integrals", &
                                        "Numerical integration on real space grid. Lumped together with core charge"), &
                          enum_i_vals=(/do_ppl_analytic, do_ppl_grid/), &
                          default_i_val=do_ppl_analytic)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_distribution_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_dftb_control_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_scptb_control_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_se_control_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mulliken_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_ddapc_restraint_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_becke_restraint_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_s2_restraint_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_lrigpw_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_optimize_lri_basis_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_qs_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_dftb_control_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_dftb_control_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",2746)
      CALL section_create(section, "DFTB", &
                          description="Parameters needed to set up the DFTB methods", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE., &
                          citations=(/Porezag1995, Seifert1996, Elstner1998, Zhechkov2005/))

      NULLIFY (subsection)
      CALL create_dftb_parameter_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="self_consistent", &
                          description="Use self-consistent method", &
                          citations=(/Elstner1998/), &
                          usage="SELF_CONSISTENT", default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="orthogonal_basis", &
                          description="Assume orthogonal basis set", &
                          usage="ORTHOGONAL_BASIS", default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="do_ewald", &
                          description="Use Ewald type method instead of direct sum for Coulomb interaction", &
                          usage="DO_EWALD", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="dispersion", &
                          description="Use dispersion correction", &
                          citations=(/Zhechkov2005/), lone_keyword_l_val=.TRUE., &
                          usage="DISPERSION", default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DIAGONAL_DFTB3", &
                          description="Use a diagonal version of the 3rd order energy correction (DFTB3) ", &
                          lone_keyword_l_val=.TRUE., &
                          usage="DIAGONAL_DFTB3", default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="HB_SR_GAMMA", &
                          description="Uses a modified version for the GAMMA within the SCC-DFTB scheme, "// &
                          "specifically tuned for hydrogen bonds.", &
                          citations=(/Hu2007/), lone_keyword_l_val=.TRUE., &
                          usage="HB_SR_GAMMA", default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="eps_disp", &
                          description="Define accuracy of dispersion interaction", &
                          usage="EPS_DISP", default_r_val=0.0001_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_dftb_control_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_scptb_control_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_scptb_control_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",2819)
      CALL section_create(section, "SCPTB", &
                          description="Parameters needed to set up the SCPTB methods", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="PARAMETER_FILE_NAME", &
                          description="Specify file that contains the atomic parameters", &
                          usage="PARAMETER_FILE_NAME filename", &
                          n_var=1, type_of_var=char_t, default_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPERSION_PARAMETER_FILE", &
                          description="Specify file that contains the atomic dispersion parameters", &
                          usage="DISPERSION_PARAMETER_FILE filename", &
                          n_var=1, type_of_var=char_t, default_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPERSION", &
                          description="Use dispersion correction", &
                          lone_keyword_l_val=.TRUE., &
                          usage="DISPERSION", default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPERSION_RADIUS", &
                          description="Define radius of dispersion interaction", &
                          usage="DISPERSION_RADIUS", default_r_val=15._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COORDINATION_CUTOFF", &
                          description="Define cutoff for coordination number calculation", &
                          usage="COORDINATION_CUTOFF", default_r_val=1.e-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="D3_SCALING", &
                          description="Scaling parameters (s6,sr6,s8) for the D3 dispersion method,", &
                          usage="D3_SCALING 1.0 1.0 1.0", n_var=3, default_r_vals=(/1.0_dp, 1.0_dp, 1.0_dp/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STO_NG", &
                          description="Provides the order of the Slater orbital expansion of Gaussian-Type Orbitals.", &
                          usage="STO_NG", default_i_val=6)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PAIR_CUTOFF", &
                          description="Define cutoff for pair potential calculation", &
                          usage="PAIR_CUTOFF", default_r_val=1.e-8_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="do_ewald", &
                          description="Use Ewald type method instead of direct sum for Coulomb interaction", &
                          usage="DO_EWALD", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="do_scc", &
                          description="Use self consistent charge method. Can be used together with DO_SCP to get TB method", &
                          usage="DO_SCC", default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="do_scp", &
                          description="Use SCP method. Can be used to switch off SCP to get a SCC-DFTB method", &
                          usage="DO_SCP", default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_scptb_control_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_admm_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_admm_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",2909)
      CALL section_create(section, "AUXILIARY_DENSITY_MATRIX_METHOD", &
                          description="Parameters needed for the ADMM method.", &
                          n_keywords=1, n_subsections=1, repeats=.FALSE., &
                          citations=(/Guidon2010/))

      CALL keyword_create( &
         keyword=keyword, &
         name="ADMM_PURIFICATION_METHOD", &
         description="Method that shall be used for wavefunction fitting. Use MO_DIAG for MD.", &
         enum_c_vals=s2a("NONE", "CAUCHY", "CAUCHY_SUBSPACE", "MO_DIAG", "MO_NO_DIAG", "MCWEENY", "NONE_DM"), &
         enum_i_vals=(/do_admm_purify_none, do_admm_purify_cauchy, do_admm_purify_cauchy_subspace, &
                       do_admm_purify_mo_diag, do_admm_purify_mo_no_diag, &
                       do_admm_purify_mcweeny, do_admm_purify_none_dm/), &
         enum_desc=s2a("Do not apply any purification", &
                       "Perform purification via general Cauchy representation", &
                       "Perform purification via Cauchy representation in occupied subspace", &
                       "Calculate MO derivatives via Cauchy representation by diagonalization", &
                       "Calculate MO derivatives via Cauchy representation by inversion", &
                       "Perform original McWeeny purification via matrix multiplications", &
                       "Do not apply any purification, works directly with density matrix"), &
         default_i_val=do_admm_purify_mo_diag)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="METHOD", &
         description="Method that shall be used for wavefunction fitting. Use BASIS_PROJECTION for MD.", &
         enum_c_vals=s2a("BASIS_PROJECTION", "BLOCKED_PROJECTION_PURIFY_FULL", "BLOCKED_PROJECTION", &
                         "CHARGE_CONSTRAINED_PROJECTION"), &
         enum_i_vals=(/do_admm_basis_projection, do_admm_blocking_purify_full, do_admm_blocked_projection, &
                       do_admm_charge_constrained_projection/), &
         enum_desc=s2a("Construct auxiliary density matrix from auxiliary basis.", &
                       "Construct auxiliary density from a blocked Fock matrix,"// &
                       " but use the original matrix for purification.", &
                       "Construct auxiliary density from a blocked Fock matrix.", &
                       "Construct auxiliary density from auxiliary basis enforcing charge constrain."), &
         default_i_val=do_admm_basis_projection)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="EXCH_SCALING_MODEL", &
         description="Scaling of the exchange correction calculated by the auxiliary density matrix.", &
         enum_c_vals=s2a("NONE", "MERLOT"), &
         enum_i_vals=(/do_admm_exch_scaling_none, do_admm_exch_scaling_merlot/), &
         enum_desc=s2a("No scaling is enabled, refers to methods ADMM1, ADMM2 or ADMMQ.", &
                       "Exchange scaling according to Merlot (2014)"), &
         default_i_val=do_admm_exch_scaling_none)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="EXCH_CORRECTION_FUNC", &
         description="Exchange functional which is used for the ADMM correction.", &
         enum_c_vals=s2a("DEFAULT", "PBEX", "NONE", "OPTX", "BECKE88X"), &
         enum_i_vals=(/do_admm_aux_exch_func_default, do_admm_aux_exch_func_pbex, &
                       do_admm_aux_exch_func_none, do_admm_aux_exch_func_opt, do_admm_aux_exch_func_bee/), &
         enum_desc=s2a("Use PBEX functional for exchange correction, also for truncated and"// &
                       " short Coulomb interaction.", &
                       "Use PBEX functional for exchange correction but always normal form.", &
                       "No correction: X(D)-x(d)-> 0.", &
                       "Use OPTX functional for exchange correction.", &
                       "Use Becke88X functional for exchange correction."), &
         default_i_val=do_admm_aux_exch_func_default)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BLOCK_LIST", &
                          description="Specifies a list of atoms.", &
                          usage="LIST {integer} {integer} .. {integer}", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_FILTER", &
                          description="Define accuracy of DBCSR operations", &
                          usage="EPS_FILTER", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_admm_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_dftb_parameter_section(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_dftb_parameter_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3008)

      CALL section_create(section, name="PARAMETER", &
                          description="Information on where to find DFTB parameters", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="SK_FILE", &
                          description="Define parameter file for atom pair", &
                          usage="SK_FILE a1 a2 filename", &
                          n_var=3, type_of_var=char_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARAM_FILE_PATH", &
                          description="Specify the directory with the DFTB parameter files. "// &
                          "Used in combination with the filenames specified in the file "// &
                          "given in PARAM_FILE_NAME.", usage="PARAM_FILE_PATH pathname", &
                          n_var=1, type_of_var=char_t, default_c_val="./")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARAM_FILE_NAME", &
                          description="Specify file that contains the names of "// &
                          "Slater-Koster tables: A plain text file, each line has the "// &
                          'format "ATOM1 ATOM2 filename.spl".', &
                          usage="PARAM_FILE_NAME filename", &
                          n_var=1, type_of_var=char_t, default_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPERSION_TYPE", &
                          description="Use dispersion correction of the specified type."// &
                          " Dispersion correction has to be switched on in the DFTB section.", &
                          usage="DISPERSION_TYPE (UFF|D3)", &
                          enum_c_vals=s2a("UFF", "D3"), &
                          enum_i_vals=(/dispersion_uff, dispersion_d3/), &
                          enum_desc=s2a("Uses the UFF force field for a pair potential dispersion correction.", &
                                        "Uses the Grimme D3 method (simplified) for a pair potential dispersion correction."), &
                          default_i_val=dispersion_uff)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="UFF_FORCE_FIELD", &
                          description="Name of file with UFF parameters that will be used "// &
                          "for the dispersion correction. Needs to be specified when "// &
                          "DISPERSION==.TRUE., otherwise cp2k crashes with a Segmentation "// &
                          "Fault.", usage="UFF_FORCE_FIELD filename", &
                          n_var=1, type_of_var=char_t, default_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPERSION_PARAMETER_FILE", &
                          description="Specify file that contains the atomic dispersion "// &
                          "parameters for the D3 method", &
                          usage="DISPERSION_PARAMETER_FILE filename", &
                          n_var=1, type_of_var=char_t, default_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPERSION_RADIUS", &
                          description="Define radius of dispersion interaction", &
                          usage="DISPERSION_RADIUS", default_r_val=15._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COORDINATION_CUTOFF", &
                          description="Define cutoff for coordination number calculation", &
                          usage="COORDINATION_CUTOFF", default_r_val=1.e-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="D3_SCALING", &
                          description="Scaling parameters (s6,sr6,s8) for the D3 dispersion method,", &
                          usage="D3_SCALING 1.0 1.0 1.0", n_var=3, default_r_vals=(/0.0_dp, 0.0_dp, 0.0_dp/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="HB_SR_PARAM", &
                          description="Uses a modified version for the GAMMA within the SCC-DFTB scheme, "// &
                          "specifically tuned for hydrogen bonds. Specify the exponent used in the exponential.", &
                          usage="HB_SR_PARAM {real}", default_r_val=4.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_dftb_parameter_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_se_control_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_se_control_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3109)
      CALL section_create(section, "SE", &
                          description="Parameters needed to set up the Semi-empirical methods", &
                          n_keywords=8, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="ORTHOGONAL_BASIS", &
                          description="Assume orthogonal basis set. This flag is overwritten by "// &
                          "methods with fixed orthogonal/non-orthogonal basis set.", &
                          usage="ORTHOGONAL_BASIS", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STO_NG", &
                          description="Provides the order of the Slater orbital expansion of Gaussian-Type Orbitals.", &
                          usage="STO_NG", default_i_val=6)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ANALYTICAL_GRADIENTS", &
                          description="Nuclear Gradients are computed analytically or numerically", &
                          usage="ANALYTICAL_GRADIENTS", default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DELTA", &
                          description="Step size in finite difference force calculation", &
                          usage="DELTA {real} ", default_r_val=1.e-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INTEGRAL_SCREENING", &
                          description="Specifies the functional form for the ", &
                          usage="INTEGRAL_SCREENING (KDSO|KDSO-D|SLATER)", &
                          enum_c_vals=s2a("KDSO", "KDSO-D", "SLATER"), &
                          enum_i_vals=(/do_se_IS_kdso, do_se_IS_kdso_d, do_se_IS_slater/), &
                          enum_desc=s2a("Uses the standard NDDO Klopman-Dewar-Sabelli-Ohno equation "// &
                                        "for the screening of the Coulomb interactions.", &
                                        "Uses a modified Klopman-Dewar-Sabelli-Ohno equation, dumping the screening "// &
                                        "parameter for the Coulomb interactions.", &
                                        "Uses an exponential Slater-type function for modelling the Coulomb interactions."), &
                          default_i_val=do_se_IS_kdso)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PERIODIC", &
                          description="Specifies the type of treatment for the electrostatic long-range part "// &
                          "in semi-empirical calculations.", &
                          usage="PERIODIC (NONE|EWALD|EWALD_R3|EWALD_GKS)", &
                          enum_c_vals=s2a("NONE", "EWALD", "EWALD_R3", "EWALD_GKS"), &
                          enum_i_vals=(/do_se_lr_none, do_se_lr_ewald, do_se_lr_ewald_r3, do_se_lr_ewald_gks/), &
                          enum_desc=s2a("The long-range part is not explicitly treaten. The interaction "// &
                                        "depends uniquely on the Cutoffs used for the calculation.", &
                                        "Enables the usage of Multipoles Ewald summation schemes. The short-range part "// &
                                        "is tapered according the value of RC_COULOMB.", &
                                        "Enables the usage of Multipoles Ewald summation schemes together with a long-range "// &
                                        "treatment for the 1/R^3 term, deriving from the short-range component. This option "// &
                                        "is active only for K-DSO integral screening type.", &
                                        "Use Ewald directly in Coulomb integral evaluation, works only with Slater screening"), &
                          default_i_val=do_se_lr_none)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FORCE_KDSO-D_EXCHANGE", &
                          description="This keywords forces the usage of the KDSO-D integral screening "// &
                          "for the Exchange integrals (default is to apply the screening only to the "// &
                          "Coulomb integrals.", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPERSION", &
                          description="Use dispersion correction", &
                          lone_keyword_l_val=.TRUE., &
                          usage="DISPERSION", default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPERSION_PARAMETER_FILE", &
                          description="Specify file that contains the atomic dispersion parameters", &
                          usage="DISPERSION_PARAMETER_FILE filename", &
                          n_var=1, type_of_var=char_t, default_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPERSION_RADIUS", &
                          description="Define radius of dispersion interaction", &
                          usage="DISPERSION_RADIUS", default_r_val=15._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COORDINATION_CUTOFF", &
                          description="Define cutoff for coordination number calculation", &
                          usage="COORDINATION_CUTOFF", default_r_val=1.e-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="D3_SCALING", &
                          description="Scaling parameters (s6,sr6,s8) for the D3 dispersion method,", &
                          usage="D3_SCALING 1.0 1.0 1.0", n_var=3, default_r_vals=(/0.0_dp, 0.0_dp, 0.0_dp/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL create_coulomb_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_exchange_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_screening_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_lr_corr_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_neighbor_lists_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_se_memory_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_se_print_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_se_ga_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_se_control_section

! **************************************************************************************************
!> \brief Create the COULOMB se section
!> \param section the section to create
!> \date  03.2009
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE create_lr_corr_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_lr_corr_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3259)
      CALL section_create(section, name="LR_CORRECTION", &
                          description="Setup parameters for the evaluation of the long-range correction term in SE "// &
                          "calculations.", n_keywords=0, n_subsections=1, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="CUTOFF", &
                          description="Atomic Cutoff Radius Cutoff for the evaluation of the long-ranbe correction integrals. ", &
                          usage="CUTOFF {real} ", unit_str="angstrom", &
                          default_r_val=cp_unit_to_cp2k(value=6.0_dp, unit_str="angstrom"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RC_TAPER", &
                          description="Atomic Cutoff Radius Cutoff for Tapering the long-range correction integrals. "// &
                          "If not specified it assumes the same value specified for the CUTOFF.", &
                          usage="RC_TAPER {real} ", unit_str="angstrom", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RC_RANGE", &
                          description="Range of cutoff switch function (tapering): 0.5*(1-TANH((r-r0)/RC_RANGE)), "// &
                          "where r0=2.0*RC_TAPER-20.0*RC_RANGE.", &
                          usage="RC_RANGE {real} ", unit_str="angstrom", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_lr_corr_section

! **************************************************************************************************
!> \brief Create the COULOMB se section
!> \param section the section to create
!> \date  03.2009
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE create_coulomb_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_coulomb_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3302)
      CALL section_create(section, name="COULOMB", &
                          description="Setup parameters for the evaluation of the COULOMB term in SE "// &
                          "calculations.", n_keywords=0, n_subsections=1, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create( &
         keyword, name="CUTOFF", &
         description="Atomic Cutoff Radius Cutoff for the evaluation of the  Coulomb integrals. "// &
         "For non-periodic calculation the default value is exactly the full cell dimension, in order "// &
         "to evaluate all pair interactions. Instead, for periodic calculations the default numerical value is used.", &
         usage="CUTOFF {real} ", unit_str="angstrom", &
         default_r_val=cp_unit_to_cp2k(value=12.0_dp, unit_str="angstrom"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RC_TAPER", &
                          description="Atomic Cutoff Radius Cutoff for Tapering Coulomb integrals. "// &
                          "If not specified it assumes the same value specified for the CUTOFF.", &
                          usage="RC_TAPER {real} ", unit_str="angstrom", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RC_RANGE", &
                          description="Range of cutoff switch function (tapering): 0.5*(1-TANH((r-r0)/RC_RANGE)), "// &
                          "where r0=2.0*RC_TAPER-20.0*RC_RANGE.", &
                          usage="RC_RANGE {real} ", unit_str="angstrom", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_coulomb_section

! **************************************************************************************************
!> \brief Create the EXCHANGE se section
!> \param section the section to create
!> \date  03.2009
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE create_exchange_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_exchange_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3348)
      CALL section_create(section, name="EXCHANGE", &
                          description="Setup parameters for the evaluation of the EXCHANGE and "// &
                          " core Hamiltonian terms in SE calculations.", n_keywords=0, n_subsections=1, &
                          repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="CUTOFF", &
                          description="Atomic Cutoff Radius Cutoff for the evaluation of the Exchange integrals. "// &
                          "For non-periodic calculation the default value is exactly the full cell dimension, in order "// &
                          "to evaluate all pair interactions. Instead, for periodic calculations the default is the "// &
                          "minimum value between 1/4 of the cell dimension and the value specified in input (either"// &
                          " explicitly defined or the default numerical value).", &
                          usage="CUTOFF {real} ", unit_str="angstrom", &
                          default_r_val=cp_unit_to_cp2k(value=12.0_dp, unit_str="angstrom"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RC_TAPER", &
                          description="Atomic Cutoff Radius Cutoff for Tapering Exchange integrals. "// &
                          "If not specified it assumes the same value specified for the CUTOFF.", &
                          usage="RC_TAPER {real} ", unit_str="angstrom", type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RC_RANGE", &
                          description="Range of cutoff switch function (tapering): 0.5*(1-TANH((r-r0)/RC_RANGE)), "// &
                          "where r0=2.0*RC_TAPER-20.0*RC_RANGE.", &
                          usage="RC_RANGE {real} ", unit_str="angstrom", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_exchange_section

! **************************************************************************************************
!> \brief Create the SCREENING se section
!> \param section the section to create
!> \date  03.2009
!> \author Teodoro Laino [tlaino]
! **************************************************************************************************
   SUBROUTINE create_screening_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_screening_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3396)
      CALL section_create(section, name="SCREENING", &
                          description="Setup parameters for the tapering of the Coulomb/Exchange Screening in "// &
                          "KDSO-D integral scheme,", n_keywords=0, n_subsections=1, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="RC_TAPER", &
                          description="Atomic Cutoff Radius Cutoff for Tapering the screening term. ", &
                          usage="RC_TAPER {real} ", unit_str="angstrom", &
                          default_r_val=cp_unit_to_cp2k(value=12.0_dp, unit_str="angstrom"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RC_RANGE", &
                          description="Range of cutoff switch function (tapering): 0.5*(1-TANH((r-r0)/RC_RANGE)), "// &
                          "where r0=2*RC_TAPER-20*RC_RANGE.", &
                          usage="RC_RANGE {real} ", unit_str="angstrom", default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_screening_section

! **************************************************************************************************
!> \brief Create the print se section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_se_print_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_se_print_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3431)
      CALL section_create(section, name="print", &
                          description="Section of possible print options in SE code.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "NEIGHBOR_LISTS", &
                                       description="Activates the printing of the neighbor lists used"// &
                                       " for the periodic SE calculations.", &
                                       print_level=high_print_level, filename="", unit_str="angstrom")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "SUBCELL", &
                                       description="Activates the printing of the subcells used for the"// &
                                       "generation of neighbor lists for periodic SE.", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "EWALD_INFO", &
                                       description="Activates the printing of the information for "// &
                                       "Ewald multipole summation in periodic SE.", &
                                       print_level=high_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_se_print_section

! **************************************************************************************************
!> \brief creates the input section for use with the GA part of the code
!> \param section the section to create
!> \author Teodoro Laino [tlaino] - University of Zurich - 05.2008
! **************************************************************************************************
   SUBROUTINE create_se_ga_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_se_ga_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3473)
      CALL section_create(section, "GA", &
                          description="Sets up memory parameters for the storage of the integrals", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="NCELLS", &
         description="Defines the number of linked cells for the neighbor list. "// &
         "Default value is number of processors", &
         usage="NCELLS 10", &
         default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_se_ga_section

! **************************************************************************************************
!> \brief creates the input section for the se-memory part integral storage
!> \param section the section to create
!> \author Teodoro Laino [tlaino] - University of Zurich - 05.2008
! **************************************************************************************************
   SUBROUTINE create_se_memory_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_se_memory_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3503)
      CALL section_create(section, "MEMORY", &
                          description="Sets up memory parameters for the storage of the integrals", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create( &
         keyword=keyword, &
         name="EPS_STORAGE", &
         description="Storage threshold for compression is EPS_STORAGE", &
         usage="EPS_STORAGE 1.0E-10", &
         default_r_val=1.0E-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword=keyword, &
         name="MAX_MEMORY", &
         description="Defines the maximum amount of memory [MB] used to store precomputed "// &
         "(possibly compressed) two-electron two-center integrals", &
         usage="MAX_MEMORY 256", &
         default_i_val=50)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COMPRESS", &
                          description="Enables the compression of the integrals in memory.", &
                          usage="COMPRESS <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_se_memory_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_mulliken_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mulliken_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3549)
      CALL section_create(section, "MULLIKEN_RESTRAINT", &
                          description="Use mulliken charges in a restraint (check code for details)", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="STRENGTH", &
                          description="force constant of the restraint", &
                          usage="STRENGTH {real} ", default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET", &
                          description="target value of the restraint", &
                          usage="TARGET {real} ", default_r_val=1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specifies the list of atoms that is summed in the restraint", &
                          usage="ATOMS {integer} {integer} .. {integer}", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_mulliken_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_density_fitting_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_density_fitting_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      NULLIFY (keyword, print_key)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3589)
      CALL section_create(section, "DENSITY_FITTING", &
                          description="Setup parameters for density fitting (Bloechl charges or density derived "// &
                          " atomic point charges (DDAPC) charges)", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE., &
                          citations=(/Blochl1995/))

      CALL keyword_create(keyword, name="NUM_GAUSS", &
                          description="Specifies the numbers of gaussian used to fit the QM density for each atomic site.", &
                          usage="NUM_GAUSS {integer}", &
                          n_var=1, type_of_var=integer_t, default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PFACTOR", &
                          description="Specifies the progression factor for the gaussian exponent for each atomic site.", &
                          usage="PFACTOR {real}", &
                          n_var=1, type_of_var=real_t, default_r_val=1.5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIN_RADIUS", &
                          description="Specifies the smallest radius of the gaussian used in the fit. All other radius are"// &
                          " obtained with the progression factor.", &
                          usage="MIN_RADIUS {real}", &
                          unit_str="angstrom", n_var=1, type_of_var=real_t, default_r_val=0.5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RADII", &
                          description="Specifies all the radius of the gaussian used in the fit for each atomic site. The use"// &
                          " of this keyword disables all other keywords of this section.", &
                          usage="RADII {real} {real} .. {real}", &
                          unit_str="angstrom", n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GCUT", &
                          description="Cutoff for charge fit in G-space.", &
                          usage="GCUT {real}", &
                          n_var=1, type_of_var=real_t, default_r_val=SQRT(6.0_dp))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing of basic information during the run", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="__STD_OUT__")

      CALL keyword_create(keyword, name="CONDITION_NUMBER", &
                          description="Prints information regarding the condition numbers of the A matrix (to be inverted)", &
                          usage="ANALYTICAL_GTERM <LOGICAL>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_density_fitting_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_ddapc_restraint_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ddapc_restraint_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      NULLIFY (keyword, print_key)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3663)
      CALL section_create(section, "DDAPC_RESTRAINT", &
                          description="Use DDAPC charges in a restraint (check code for details)", &
                          n_keywords=7, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="TYPE_OF_DENSITY", &
                          description="Specifies the type of density used for the fitting", &
                          usage="TYPE_OF_DENSITY (FULL|SPIN)", &
                          enum_c_vals=s2a("FULL", "SPIN"), &
                          enum_i_vals=(/do_full_density, do_spin_density/), &
                          enum_desc=s2a("Full density", "Spin density"), &
                          default_i_val=do_full_density)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STRENGTH", &
                          description="force constant of the restraint", &
                          usage="STRENGTH {real} ", default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET", &
                          description="target value of the restraint", &
                          usage="TARGET {real} ", default_r_val=1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specifies the list of atoms that is summed in the restraint", &
                          usage="ATOMS {integer} {integer} .. {integer}", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COEFF", &
                          description="Defines the the coefficient of the atom in the atom list (default is one) ", &
                          usage="COEFF 1.0 -1.0", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FUNCTIONAL_FORM", &
                          description="Specifies the functional form of the term added", &
                          usage="FUNCTIONAL_FORM RESTRAINT", &
                          enum_c_vals=s2a("RESTRAINT", "CONSTRAINT"), &
                          enum_i_vals=(/do_ddapc_restraint, do_ddapc_constraint/), &
                          enum_desc=s2a("Harmonic potential: s*(q-t)**2", "Constraint form: s*(q-t)"), &
                          default_i_val=do_ddapc_restraint)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing basic info about the method", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_ddapc_restraint_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_becke_restraint_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_becke_restraint_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      NULLIFY (keyword, print_key)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3736)
      CALL section_create(section, "BECKE_RESTRAINT", &
                          description="Use Becke weight population in a restraint/constraint ", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="STRENGTH", &
                          description="force constant of the restraint", &
                          usage="STRENGTH {real} ", default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET", &
                          description="target value of the restraint", &
                          usage="TARGET {real} ", default_r_val=1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS", &
                          description="Specifies the list of atoms that is summed in the restraint", &
                          usage="ATOMS {integer} {integer} .. {integer}", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="COEFF", &
                          description="Defines the the coefficient of the atom in the atom list (default is one)", &
                          usage="COEFF 1.0 -1.0", &
                          type_of_var=real_t, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FUNCTIONAL_FORM", &
                          description="Specifies the functional form of the term added", &
                          usage="FUNCTIONAL_FORM RESTRAINT", &
                          enum_c_vals=s2a("RESTRAINT", "CONSTRAINT"), &
                          enum_i_vals=(/do_ddapc_restraint, do_ddapc_constraint/), &
                          enum_desc=s2a("Harmonic potential: s*(q-t)**2", "Constraint form: s*(q-t)"), &
                          default_i_val=do_ddapc_restraint)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TYPE_OF_DENSITY", &
                          description="Specifies the type of density used for the fitting", &
                          usage="TYPE_OF_DENSITY (FULL|SPIN)", &
                          enum_c_vals=s2a("FULL", "SPIN"), &
                          enum_i_vals=(/do_full_density, do_spin_density/), &
                          enum_desc=s2a("Full density", "Spin density"), &
                          default_i_val=do_full_density)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing basic info about the method", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_becke_restraint_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_s2_restraint_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_s2_restraint_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3808)

      CALL section_create(section, "S2_RESTRAINT", &
                          description="Use S2 in a re/constraint (OT only)", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE.)

      CALL keyword_create(keyword, name="STRENGTH", &
                          description="force constant of the restraint", &
                          usage="STRENGTH {real} ", default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TARGET", &
                          description="target value of the restraint", &
                          usage="TARGET {real} ", default_r_val=1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FUNCTIONAL_FORM", &
                          description="Specifies the functional form of the term added", &
                          usage="FUNCTIONAL_FORM RESTRAINT", &
                          enum_c_vals=s2a("RESTRAINT", "CONSTRAINT"), &
                          enum_i_vals=(/do_s2_restraint, do_s2_constraint/), &
                          enum_desc=s2a("Harmonic potential: s*(q-t)**2", "Constraint form: s*(q-t)"), &
                          default_i_val=do_s2_restraint)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_s2_restraint_section

! **************************************************************************************************
!> \brief creates the input section for the tddfpt part
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_tddfpt_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_tddfpt_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",3852)
      CALL section_create(section, "tddfpt", &
                          description="parameters needed to set up the Time Dependent Density Functional PT", &
                          n_keywords=5, n_subsections=1, repeats=.FALSE., &
                          citations=(/Iannuzzi2005/))

      NULLIFY (subsection, keyword)

      ! Integer
      CALL keyword_create(keyword, name="MAX_KV", &
                          variants=s2a("MAX_VECTORS"), &
                          description=" maximal number of Krylov space vectors", &
                          usage="MAX_KV someInteger>0", &
                          n_var=1, type_of_var=integer_t, &
                          default_i_val=60)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTARTS", &
                          variants=s2a("N_RESTARTS"), &
                          description=" maximal number subspace search restarts", &
                          usage="RESTARTS someInteger>0", &
                          n_var=1, type_of_var=integer_t, &
                          default_i_val=5)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NEV", &
                          variants=s2a("N_EV", "EV"), &
                          description=" number of excitations to calculate", &
                          usage="NEV someInteger>0", &
                          n_var=1, type_of_var=integer_t, &
                          default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NLUMO", &
                          description=" number of additional unoccupied orbitals ", &
                          usage="NLUMO 10", &
                          n_var=1, type_of_var=integer_t, &
                          default_i_val=5)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NREORTHO", &
                          variants=s2a("N_REORTHO", "REORTHO", "REORTHOGONALITAZIONS"), &
                          description=" number of reorthogonalization steps", &
                          usage="NREORTHO someInteger>0", &
                          n_var=1, type_of_var=integer_t, &
                          default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Logical
      CALL keyword_create(keyword, name="KERNEL", &
                          variants=s2a("DO_KERNEL"), &
                          description="compute the kernel (debug purpose only)", &
                          usage="KERNEL logical_value", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LSD_SINGLETS", &
                          description="compute singlets using lsd vxc kernel", &
                          usage="LSD_SINGLETS logical_value", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INVERT_S", &
                          variants=s2a("INVERT_OVERLAP"), &
                          description="use the inverse of the overlap matrix", &
                          usage="INVERT_S logical_value", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRECONDITIONER", &
                          variants=s2a("PRECOND"), &
                          description="use the preconditioner (only for Davidson)", &
                          usage="PRECONDITIONER logical_value", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Character
      CALL keyword_create(keyword, name="RES_ETYPE", &
                          variants=s2a("RESTRICTED_EXCITATIONS_TYPE", "RES_E_TYPE"), &
                          description="(singlets/triplets) for restricted calculation", &
                          usage="RES_ETYPE T", &
                          enum_c_vals=s2a("S", "SINGLET", "SINGLETS", "T", "TRIPLET", "TRIPLETS"), &
                          enum_i_vals=(/tddfpt_singlet, tddfpt_singlet, tddfpt_singlet, &
                                        tddfpt_triplet, tddfpt_triplet, tddfpt_triplet/), &
                          default_i_val=tddfpt_singlet)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DIAG_METHOD", &
                          variants=s2a("DIAGONALIZATION_METHOD", "METHOD"), &
                          description="Diagonalization method used in tddfpt", &
                          usage="DIAG_METHOD DAVIDSON", &
                          enum_c_vals=s2a("DAVIDSON", "LANCZOS"), &
                          enum_i_vals=(/tddfpt_davidson, tddfpt_lanczos/), &
                          default_i_val=tddfpt_davidson)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OE_CORR", &
                          variants=s2a("ORBITAL_EIGENVALUES_CORRECTION"), &
                          description="Which type of orbital eigenvalue correction to use\n"// &
                          "(to yield better HOMO-LUMO energies)", &
                          usage="OE_CORR SAOP", &
                          enum_c_vals=s2a("NONE", "LB", "LB_ALPHA", "LB94", "GLLB", "GLB", "SAOP", "SIC"), &
                          enum_i_vals=(/oe_none, oe_lb, oe_lb, oe_lb, oe_gllb, oe_gllb, oe_saop, oe_sic/), &
                          default_i_val=oe_none)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      ! Real
      CALL keyword_create(keyword, name="CONVERGENCE", &
                          variants=s2a("CONV"), &
                          description="The convergence of the eigenvalues", &
                          usage="CONVERGENCE 1.0E-6 ", &
                          n_var=1, type_of_var=real_t, &
                          default_r_val=1.0e-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_xc_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_sic_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_tddfpt_section

! **************************************************************************************************
!> \brief creates the input section for the relativistic part
!> \param section the section to create
!> \author jens
! **************************************************************************************************
   SUBROUTINE create_relativistic_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_relativistic_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",4003)
      CALL section_create(section, "relativistic", &
                          description="parameters needed and setup for relativistic calculations", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="method", &
                          description="type of relativistic correction used", &
                          usage="method (NONE|DKH|ZORA)", default_i_val=rel_none, &
                          enum_c_vals=s2a("NONE", "DKH", "ZORA"), &
                          enum_i_vals=(/rel_none, rel_dkh, rel_zora/), &
                          enum_desc=s2a("Use no relativistic correction", &
                                        "Use Douglas-Kroll-Hess method", &
                                        "Use ZORA method"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DKH_order", &
                          description="The order of the DKH transformation ", &
                          usage="DKH_order 2", default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ZORA_type", &
                          description="Type of ZORA method to be used", &
                          usage="ZORA_type scMP", default_i_val=rel_zora_full, &
                          enum_c_vals=s2a("FULL", "MP", "scMP"), &
                          enum_desc=s2a("Full ZORA method (not implemented)", &
                                        "ZORA with atomic model potential", &
                                        "Scaled ZORA with atomic model potential"), &
                          enum_i_vals=(/rel_zora_full, rel_zora_mp, rel_sczora_mp/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="transformation", &
                          description="Type of DKH transformation", &
                          usage="transformation (FULL|MOLECULE|ATOM)", default_i_val=rel_trans_full, &
                          enum_c_vals=s2a("FULL", "MOLECULE", "ATOM"), &
                          enum_i_vals=(/rel_trans_full, rel_trans_molecule, rel_trans_atom/), &
                          enum_desc=s2a("Use full matrix transformation", &
                                        "Use transformation blocked by molecule", &
                                        "Use atomic blocks"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="z_cutoff", &
                          description="The minimal atomic number considered for atom transformation", &
                          usage="z_cutoff 50", default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="potential", &
                          description="External potential used in DKH transformation, full 1/r or erfc(r)/r", &
                          usage="POTENTIAL {FULL,ERFC}", default_i_val=rel_pot_erfc, &
                          enum_c_vals=s2a("FULL", "ERFC"), &
                          enum_i_vals=(/rel_pot_full, rel_pot_erfc/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_relativistic_section

! **************************************************************************************************
!> \brief creates the structure of the section with the DFT SCF parameters
!> \param section will contain the SCF section
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_scf_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_scf_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      NULLIFY (print_key)

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",4081)
      CALL section_create(section, "scf", &
                          description="Parameters needed to perform an SCF run.", &
                          n_keywords=18, n_subsections=7, repeats=.FALSE.)

      NULLIFY (subsection)

      CALL create_ot_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_diagonalization_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_outer_scf_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_smear_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mixing_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_mom_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="MAX_ITER_LUMO", &
                          variants=(/"MAX_ITER_LUMOS"/), &
                          description="The maximum number of iteration for the lumo computation", &
                          usage="MAX_ITER_LUMO 100", default_i_val=299)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_LUMO", &
                          variants=(/"EPS_LUMOS"/), &
                          description="target accuracy of the computation of the lumo energy", &
                          usage="EPS_LUMO 1.e-6", default_r_val=1.0e-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_SCF", &
                          description="Maximum number of SCF iteration to be performed for one optimization", &
                          usage="MAX_SCF 200", default_i_val=50)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_SCF_HISTORY", variants=(/"MAX_SCF_HIST"/), &
                          description="Maximum number of SCF iterations after the history pipeline is filled", &
                          usage="MAX_SCF_HISTORY 1", default_i_val=0, lone_keyword_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_DIIS", &
                          variants=(/"MAX_DIIS_BUFFER_SIZE"/), &
                          description="Maximum number of DIIS vectors to be used", &
                          usage="MAX_DIIS 3", default_i_val=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LEVEL_SHIFT", &
                          variants=(/"LSHIFT"/), &
                          description="Use level shifting to improve convergence", &
                          usage="LEVEL_SHIFT 0.1", default_r_val=0._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_SCF", &
                          description="Target accuracy for the SCF convergence.", &
                          usage="EPS_SCF 1.e-6", default_r_val=1.e-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_SCF_HISTORY", variants=(/"EPS_SCF_HIST"/), &
                          description="Target accuracy for the SCF convergence after the history pipeline is filled.", &
                          usage="EPS_SCF_HISTORY 1.e-5", default_r_val=0.0_dp, lone_keyword_r_val=1.0e-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHOLESKY", &
                          description="If the cholesky method should be used for computing "// &
                          "the inverse of S, and in this case calling which Lapack routines", &
                          usage="CHOLESKY REDUCE", default_i_val=cholesky_restore, &
                          enum_c_vals=s2a("OFF", "REDUCE", "RESTORE", "INVERSE", "INVERSE_DBCSR"), &
                          enum_desc=s2a("The cholesky algorithm is not used", "Reduce is called", &
                                        "Reduce is replaced by two restore", &
                                        "Restore uses operator multiply by inverse of the triangular matrix", &
                                        "Like inverse, but matrix stored as dbcsr, sparce matrix algebra used when possible"), &
                          enum_i_vals=(/cholesky_off, cholesky_reduce, cholesky_restore, cholesky_inverse, cholesky_dbcsr/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_EIGVAL", &
                          description="Throw away linear combinations of basis functions with a small eigenvalue in S", &
                          usage="EPS_EIGVAL 1.0", default_r_val=1.0e-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_DIIS", &
                          description="Threshold on the convergence to start using DIAG/DIIS", &
                          usage="EPS_DIIS 5.0e-2", default_r_val=0.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="SCF_GUESS", &
         description="Change the initial guess for the wavefunction.", &
         usage="SCF_GUESS RESTART", default_i_val=atomic_guess, &
         enum_c_vals=s2a("ATOMIC", "RESTART", "RANDOM", "CORE", &
                         "HISTORY_RESTART", "MOPAC", "SPARSE", "NONE"), &
         enum_desc=s2a("Generate an atomic density using the atomic code", &
                       "Use the RESTART file as an initial guess (and ATOMIC if not present).", &
                       "Use random wavefunction coefficients.", &
                       "Diagonalize the core hamiltonian for an initial guess.", &
                       "Extrapolated from previous RESTART files.", &
                       "Use same guess as MOPAC for semi-empirical methods or a simple diagonal density matrix for other methods", &
                       "Generate a sparse wavefunction using the atomic code (for OT based methods)", &
                       "Skip initial guess (only for NON-SCC DFTB)."), &
         enum_i_vals=(/atomic_guess, restart_guess, random_guess, core_guess, &
                       history_guess, mopac_guess, sparse_guess, no_guess/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NROW_BLOCK", &
                          description="sets the number of rows in a scalapack block", &
                          usage="NROW_BLOCK 31", default_i_val=32)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NCOL_BLOCK", &
                          description="Sets the number of columns in a scalapack block", &
                          usage="NCOL_BLOCK 31", default_i_val=32)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ADDED_MOS", &
                          description="Number of additional MOS added for each spin", &
                          usage="ADDED_MOS", default_i_val=0, n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="ROKS_SCHEME", &
                          description="Selects the ROKS scheme when ROKS is applied.", &
                          usage="ROKS_SCHEME HIGH-SPIN", &
                          repeats=.FALSE., &
                          n_var=1, &
                          enum_c_vals=s2a("GENERAL", "HIGH-SPIN"), &
                          enum_i_vals=(/general_roks, high_spin_roks/), &
                          default_i_val=high_spin_roks)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="ROKS_F", &
                          variants=(/"F_ROKS"/), &
                          description="Allows to define the parameter f for the "// &
                          "general ROKS scheme.", &
                          usage="ROKS_PARAMETER 1/2", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="ROKS_PARAMETERS", &
                          variants=(/"ROKS_PARAMETER"/), &
                          description="Allows to define all parameters for the high-spin "// &
                          "ROKS scheme explicitly. "// &
                          "The full set of 6 parameters has to be specified "// &
                          "in the order acc, bcc, aoo, boo, avv, bvv", &
                          usage="ROKS_PARAMETERS 1/2 1/2 1/2 1/2 1/2 1/2", &
                          repeats=.FALSE., &
                          n_var=6, &
                          type_of_var=real_t, &
                          default_r_vals=(/-0.5_dp, 1.5_dp, 0.5_dp, 0.5_dp, 1.5_dp, -0.5_dp/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, "PRINT", "Printing of information during the SCF.", &
                          repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "RESTART", &
                                       description="Controls the dumping of the MO restart file during SCF."// &
                                       "By default keeps a short history of three restarts."// &
                                       "See also RESTART_HISTORY", &
                                       print_level=low_print_level, common_iter_levels=3, &
                                       each_iter_names=s2a("QS_SCF"), each_iter_values=(/20/), &
                                       add_last=add_last_numeric, filename="RESTART")
      CALL keyword_create(keyword, name="BACKUP_COPIES", &
                          description="Specifies the maximum index of backup copies.", &
                          usage="BACKUP_COPIES {int}", &
                          default_i_val=3)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key, "RESTART_HISTORY", &
         description="Dumps unique MO restart files during the run keeping all of them.", &
         print_level=low_print_level, common_iter_levels=0, &
         each_iter_names=s2a("__ROOT__", "MD", "GEO_OPT", "ROT_OPT", "NEB", "METADYNAMICS", "QS_SCF"), &
         each_iter_values=(/500, 500, 500, 500, 500, 500, 500/), &
         filename="RESTART")
      CALL keyword_create(keyword, name="BACKUP_COPIES", &
                          description="Specifies the maximum index of backup copies.", &
                          usage="BACKUP_COPIES {int}", &
                          default_i_val=3)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "iteration_info", &
                                       description="Controls the printing of basic iteration information during the SCF.", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="time_cumul", &
                          description="If the printkey is activated switches the printing of timings"// &
                          " to cumulative (over the SCF).", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing of basic information during the SCF.", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "MO_ORTHONORMALITY", &
                                       description="Controls the printing relative to the orthonormality of MOs (CT S C).", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "MO_MAGNITUDE", &
                                       description="Prints the min/max eigenvalues of the overlap of the MOs without S (CT C).", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "detailed_energy", &
                                       description="Controls the printing of detailed energy information.", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "diis_info", &
                                       description="Controls the printing of DIIS information.", &
                                       print_level=high_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "total_densities", &
                                       description="Controls the printing of total densities.", &
                                       print_level=medium_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "Lanczos", &
                                       description="Controls the printing of information on Lanczos refinement iterations.", &
                                       print_level=medium_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key, "DIAG_SUB_SCF", &
         description="Controls the printing of information on subspace diagonalization internal loop. ", &
         print_level=medium_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "Davidson", &
                                       description="Controls the printing of information on Davidson iterations.", &
                                       print_level=medium_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "FILTER_MATRIX", &
                                       description="Controls the printing of information on Filter Matrix method.", &
                                       print_level=medium_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL keyword_create(keyword, name="DM_RESTART_WRITE", &
                          description="Write the density matrix into a binary file at the end of the SCF.", &
                          usage="DM_RESTART_WRITE", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_scf_section

! **************************************************************************************************
!> \brief creates the KG section
!> \param section ...
!> \author Martin Haeufel [2012.07]
! **************************************************************************************************
   SUBROUTINE create_kg_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_kg_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",4401)
      CALL section_create(section, "KG_METHOD", &
                          description="Specifies the parameters for a Kim-Gordon-like partitioning"// &
                          " into molecular subunits", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE., &
                          citations=(/Iannuzzi2006, Brelaz1979/))

      NULLIFY (keyword, subsection, print_key)

      CALL keyword_create(keyword, name="COLORING_METHOD", &
                          description="Which algorithm to use for coloring.", &
                          usage="COLORING_METHOD GREEDY", &
                          default_i_val=kg_color_dsatur, &
                          enum_c_vals=s2a("DSATUR", "GREEDY"), &
                          enum_desc=s2a("Maximum degree of saturation, relatively accurate", &
                                        "Greedy, fast coloring, less accurate"), &
                          enum_i_vals=(/kg_color_dsatur, kg_color_greedy/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TNADD_METHOD", &
                          description="Algorithm to use for the calculation of the nonadditive kinetic energy.", &
                          usage="TNADD_METHOD ATOMIC", &
                          default_i_val=kg_tnadd_embed, &
                          enum_c_vals=s2a("EMBEDDING", "ATOMIC"), &
                          enum_desc=s2a("Use full embedding potential (see Iannuzzi et al)", &
                                        "Use sum of atomic model potentials"), &
                          enum_i_vals=(/kg_tnadd_embed, kg_tnadd_atomic/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL section_create(subsection, name="PRINT", &
                          description="Print section", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      CALL cp_print_key_section_create(print_key, "NEIGHBOR_LISTS", &
                                       description="Controls the printing of the neighbor lists.", &
                                       print_level=low_print_level, filename="__STD_OUT__", unit_str="angstrom")

      CALL keyword_create(keyword=keyword, &
                          name="SAB_ORB_FULL", &
                          description="Activates the printing of the full orbital "// &
                          "orbital neighbor lists.", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="SAB_ORB_MOLECULAR", &
                          description="Activates the printing of the orbital "// &
                          "orbital neighbor lists for molecular subsets.", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="SAC_KIN", &
                          description="Activates the printing of the orbital "// &
                          "atomic potential neighbor list.", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_kg_section

! **************************************************************************************************
!> \brief creates the structure of the section with SCF parameters
!>      controlling an other loop
!> \param section will contain the SCF section
!> \author Joost VandeVondele [2006.03]
! **************************************************************************************************
   SUBROUTINE create_outer_scf_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_outer_scf_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",4489)
      CALL section_create(section, "OUTER_SCF", &
                          description="parameters controlling the outer SCF loop", &
                          n_keywords=9, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the outer SCF loop", &
                          usage="&OUTER_SCF ON", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TYPE", &
                          description="Specifies which kind of outer SCF should be employed", &
                          usage="TYPE DDAPC_CONSTRAINT ", &
                          default_i_val=outer_scf_none, &
                          enum_c_vals=s2a("DDAPC_CONSTRAINT", "S2_CONSTRAINT", "BECKE_CONSTRAINT", "NONE"), &
                          enum_desc=s2a("Enforce a constraint on the DDAPC, requires the corresponding section", &
                                        "Enforce a constraint on the S2, requires the corresponding section", &
                                        "Enforce a constraint on the Becke weight population,requires the corresponding section", &
                                        "Do nothing in the outer loop, useful for resetting the inner loop,"), &
                          enum_i_vals=(/outer_scf_ddapc_constraint, outer_scf_s2_constraint, &
                                        outer_scf_becke_constraint, outer_scf_none/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="OPTIMIZER", &
         description="Method used to bring the outer loop to a stationary point", &
         usage="OPTIMIZER SD", &
         default_i_val=outer_scf_optimizer_none, &
         enum_c_vals=s2a("SD", "DIIS", "NONE", "BISECT"), &
         enum_desc=s2a("Takes steps in the direction of the gradient, multiplied by step_size", &
                       "Uses a Direct Inversion in the Iterative Subspace method", &
                       "Do nothing, useful only with the none type", &
                       "Bisection on the gradient, useful for difficult one dimensional cases"), &
         enum_i_vals=(/outer_scf_optimizer_sd, outer_scf_optimizer_diis, outer_scf_optimizer_none, outer_scf_optimizer_bisect/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BISECT_TRUST_COUNT", &
                          description="Maximum number of times the same point will be used in bisection,"// &
                          " a small number guards against the effect of wrongly converged states.", &
                          usage="BISECT_TRUST_COUNT 5", default_i_val=10)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_SCF", &
                          description="The target gradient of the outer SCF variables. "// &
                          "Notice that the EPS_SCF of the inner loop also determines "// &
                          "the value that can be reached in the outer loop, "// &
                          "typically EPS_SCF of the outer loop must be smaller "// &
                          "than EPS_SCF of the inner loop.", &
                          usage="EPS_SCF 1.0E-6 ", default_r_val=1.0E-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DIIS_BUFFER_LENGTH", &
                          description="Maximum number of DIIS vectors used ", &
                          usage="DIIS_BUFFER_LENGTH 5", default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXTRAPOLATION_ORDER", &
                          description="Number of past states used in the extrapolation of the variables during e.g. MD", &
                          usage="EXTRAPOLATION_ORDER 5", default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_SCF", &
                          description="The maximum number of outer loops ", &
                          usage="MAX_SCF 20", default_i_val=50)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STEP_SIZE", &
                          description="The initial step_size used in the optimizer (currently steepest descent)."// &
                          "Note that in cases where a sadle point is sought for (DDAPC_CONSTRAINT),"// &
                          " this can be negative", &
                          usage="STEP_SIZE -1.0", default_r_val=0.5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_outer_scf_section

! **************************************************************************************************
!> \brief Create the BSSE section for counterpoise correction
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_bsse_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_bsse_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",4589)
      CALL section_create(section, name="BSSE", &
                          description="This section is used to set up the BSSE calculation. "// &
                          "It also requires that for each atomic kind X a kind X_ghost is present, "// &
                          "with the GHOST keyword specified, in addition to the other required fields.", &
                          n_keywords=3, n_subsections=1, repeats=.FALSE.)

      NULLIFY (keyword, subsection)
      ! FRAGMENT SECTION
      CALL section_create(subsection, name="FRAGMENT", &
                          description="Specify the atom number belonging to this fragment.", &
                          n_keywords=2, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="LIST", &
                          description="Specifies a list of atoms.", &
                          usage="LIST {integer} {integer} .. {integer}", &
                          repeats=.TRUE., n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      ! CONFIGURATION SECTION
      CALL section_create(subsection, name="CONFIGURATION", &
                          description="Specify additional parameters for the combinatorial configurations.", &
                          n_keywords=2, n_subsections=0, repeats=.TRUE.)

      CALL keyword_create(keyword, name="GLB_CONF", &
                          description="Specifies the global configuration using 1 or 0.", &
                          usage="GLB_CONF {integer} {integer} .. {integer}", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SUB_CONF", &
                          description="Specifies the subconfiguration using 1 or 0 belonging to the global configuration.", &
                          usage="SUB_CONF {integer} {integer} .. {integer}", &
                          n_var=-1, type_of_var=integer_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, &
                          name="MULTIPLICITY", &
                          variants=(/"MULTIP"/), &
                          description="Specify for each fragment the multiplicity. Two times the total spin plus one. "// &
                          "Specify 3 for a triplet, 4 for a quartet,and so on. Default is 1 (singlet) for an "// &
                          "even number and 2 (doublet) for an odd number of electrons.", &
                          usage="MULTIPLICITY 3", &
                          default_i_val=0) ! this default value is just a flag to get the above
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHARGE", &
                          description="The total charge for each fragment.", &
                          usage="CHARGE -1", &
                          default_i_val=0)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, name="FRAGMENT_ENERGIES", &
                          description="This section contains the energies of the fragments already"// &
                          " computed. It is useful as a summary and specifically for restarting BSSE runs.", &
                          n_keywords=2, n_subsections=0, repeats=.TRUE.)
      CALL keyword_create(keyword, name="_DEFAULT_KEYWORD_", &
                          description="The energy computed for each fragment", repeats=.TRUE., &
                          usage="{REAL}", type_of_var=real_t)
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_print_bsse_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_bsse_section

! **************************************************************************************************
!> \brief Create the print bsse section
!> \param section the section to create
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_print_bsse_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_print_bsse_section', &
         routineP = moduleN//':'//routineN

      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",4682)
      CALL section_create(section, name="print", &
                          description="Section of possible print options in BSSE code.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing of information regarding the run.", &
                                       print_level=low_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RESTART", &
                                       description="Controls the dumping of the restart file during BSSE runs."// &
                                       "By default the restart is updated after each configuration calculation is "// &
                                       " completed.", &
                                       print_level=silent_print_level, common_iter_levels=0, &
                                       add_last=add_last_numeric, filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_print_bsse_section

! **************************************************************************************************
!> \brief creates the interpolation section for the periodic QM/MM
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_gspace_interp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_gspace_interp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",4719)
      CALL section_create(section, name="interpolator", &
                          description="controls the interpolation for the G-space term", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, print_key)

      CALL keyword_create(keyword, name="aint_precond", &
                          description="the approximate inverse to use to get the starting point"// &
                          " for the linear solver of the spline3 methods", &
                          usage="kind spline3", &
                          default_i_val=precond_spl3_aint, &
                          enum_c_vals=s2a("copy", "spl3_nopbc_aint1", "spl3_nopbc_precond1", &
                                          "spl3_nopbc_aint2", "spl3_nopbc_precond2", "spl3_nopbc_precond3"), &
                          enum_i_vals=(/no_precond, precond_spl3_aint, precond_spl3_1, &
                                        precond_spl3_aint2, precond_spl3_2, precond_spl3_3/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="precond", &
                          description="The preconditioner used"// &
                          " for the linear solver of the spline3 methods", &
                          usage="kind spline3", &
                          default_i_val=precond_spl3_3, &
                          enum_c_vals=s2a("copy", "spl3_nopbc_aint1", "spl3_nopbc_precond1", &
                                          "spl3_nopbc_aint2", "spl3_nopbc_precond2", "spl3_nopbc_precond3"), &
                          enum_i_vals=(/no_precond, precond_spl3_aint, precond_spl3_1, &
                                        precond_spl3_aint2, precond_spl3_2, precond_spl3_3/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="eps_x", &
                          description="accuracy on the solution for spline3 the interpolators", &
                          usage="eps_x 1.e-15", default_r_val=1.e-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="eps_r", &
                          description="accuracy on the residual for spline3 the interpolators", &
                          usage="eps_r 1.e-15", default_r_val=1.e-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="max_iter", &
                          variants=(/'maxiter'/), &
                          description="the maximum number of iterations", &
                          usage="max_iter 200", default_i_val=100)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "conv_info", &
                                       description="if convergence information about the linear solver"// &
                                       " of the spline methods should be printed", &
                                       print_level=medium_print_level, each_iter_names=s2a("SPLINE_FIND_COEFFS"), &
                                       each_iter_values=(/10/), filename="__STD_OUT__", &
                                       add_last=add_last_numeric)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_gspace_interp_section

! **************************************************************************************************
!> \brief input section for optional parameters for LRIGPW
!>        LRI: local resolution of identity
!> \param section the section to create
!> \author Dorothea Golze [02.2015]
! **************************************************************************************************
   SUBROUTINE create_lrigpw_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_lrigpw_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",4795)
      CALL section_create(section, name="LRIGPW", &
                          description="This section specifies optional parameters for LRIGPW.", &
                          n_keywords=3, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="LRI_OVERLAP_MATRIX", &
                          description="Specifies whether to calculate the inverse or the "// &
                          "pseudoinverse of the overlap matrix of the auxiliary "// &
                          "basis set. Calculating the pseudoinverse is necessary "// &
                          "for very large auxiliary basis sets, but more expensive. "// &
                          "Using the pseudoinverse, consistent forces are not "// &
                          "guaranteed yet.", &
                          usage="LRI_OVERLAP_MATRIX INVERSE", &
                          enum_c_vals=s2a("INVERSE", "PSEUDO_INVERSE_SVD", "PSEUDO_INVERSE_DIAG", &
                                          "AUTOSELECT"), &
                          enum_desc=s2a("Calculate inverse of the overlap matrix.", &
                                        "Calculate the pseuodinverse of the overlap matrix "// &
                                        "using singular value decomposition.", &
                                        "Calculate the pseudoinverse of the overlap matrix "// &
                                        "by prior diagonalization.", &
                                        "Choose automatically for each pair whether to "// &
                                        "calculate the inverse or pseudoinverse based on the "// &
                                        "condition number of the overlap matrix for each pair. "// &
                                        "Calculating the pseudoinverse is much more expensive."), &
                          enum_i_vals=(/do_lri_inv, do_lri_pseudoinv_svd, &
                                        do_lri_pseudoinv_diag, do_lri_inv_auto/), &
                          default_i_val=do_lri_inv)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_CONDITION_NUM", &
                          description="If AUTOSELECT is chosen for LRI_OVERLAP_MATRIX, this "// &
                          "keyword specifies that the pseudoinverse is calculated "// &
                          "only if the LOG of the condition number of the lri "// &
                          "overlap matrix is larger than the given value.", &
                          usage="MAX_CONDITION_NUM 20.0", default_r_val=20.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DEBUG_LRI_INTEGRALS", &
                          description="Debug the integrals needed for LRIGPW.", &
                          usage="DEBUG_LRI_INTEGRALS TRUE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SHG_LRI_INTEGRALS", &
                          description="Uses the SHG (solid harmonic Gaussian) integral "// &
                          "scheme instead of Obara-Saika", &
                          usage="SHG_LRI_INTEGRALS TRUE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_lrigpw_section

! **************************************************************************************************
!> \brief input section for optimization of the auxililary basis for LRIGPW
!> \param section the section to create
!> \author Dorothea Golze [05.2014]
! **************************************************************************************************
   SUBROUTINE create_optimize_lri_basis_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_optimize_lri_basis_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",4867)
      CALL section_create(section, name="OPTIMIZE_LRI_BASIS", &
                          description="This section specifies the parameters for optimizing "// &
                          "the lri auxiliary basis sets for LRIGPW. The Powell optimizer is used.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="ACCURACY", &
                          description="Target accuracy for the objective function (RHOEND)", &
                          usage="ACCURACY 5.0E-4", default_r_val=1.0E-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_FUN", &
                          description="Maximum number of function evaluations", &
                          usage="MAX_FUN 200", default_i_val=4000)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STEP_SIZE", &
                          description="Initial step size for search algorithm (RHOBEG)", &
                          usage="STEP_SIZE 1.0E-1", default_r_val=5.0E-2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CONDITION_WEIGHT", &
                          description="This keyword allows to give different weight "// &
                          "factors to the condition number (LOG(cond) is used).", &
                          usage="CONDITION_WEIGHT 1.0E-4", default_r_val=1.0E-6_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="USE_CONDITION_NUMBER", &
                          description="Determines whether condition number should be part "// &
                          "of optimization or not", &
                          usage="USE_CONDITION_NUMBER", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GEOMETRIC_SEQUENCE", &
                          description="Exponents are assumed to be a geometric squence. "// &
                          "Only the minimal and maximal exponents of one set are optimized and "// &
                          "the other exponents are obtained by geometric progression.", &
                          usage="GEOMETRIC_SEQUENCE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DEGREES_OF_FREEDOM", &
                          description="Specifies the degrees of freedom in the basis "// &
                          "optimization.", &
                          usage="DEGREES_OF_FREEDOM ALL", &
                          enum_c_vals=s2a("ALL", "COEFFICIENTS", "EXPONENTS"), &
                          enum_desc=s2a("Set all parameters in the basis to be variable.", &
                                        "Set all coefficients in the basis set to be variable.", &
                                        "Set all exponents in the basis to be variable."), &
                          enum_i_vals=(/do_lri_opt_all, do_lri_opt_coeff, do_lri_opt_exps/), &
                          default_i_val=do_lri_opt_exps)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_constrain_exponents_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_optimize_lri_basis_section

! **************************************************************************************************
!> \brief input section for constraints for auxiliary basis set optimization
!> \param section the section to create
!> \author Dorothea Golze [11.2014]
! **************************************************************************************************
   SUBROUTINE create_constrain_exponents_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_constrain_exponents_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      CALL section_create(section, name="CONSTRAIN_EXPONENTS", &
                          description="specifies constraints for the exponents of the "// &
                          "lri auxiliary basis sets in the optimization.", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="SCALE", &
                          description="Defines the upper and lower boundaries as "// &
                          "(1+scale)*exp and (1-scale)*exp. Fermi-like constraint "// &
                          "function", &
                          usage="SCALE 0.3", default_r_val=0.3_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FERMI_EXP", &
                          description="Exponent in the fermi-like constraint function. ", &
                          usage="FERMI_EXP 2.63", default_r_val=2.63391_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_constrain_exponents_section

! **************************************************************************************************
!> \brief creates the multigrid
!> \param section ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_mgrid_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_mgrid_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",4986)
      CALL section_create(section, name="mgrid", &
                          description="multigrid information", &
                          n_keywords=5, n_subsections=1, repeats=.FALSE.)
      NULLIFY (keyword)
      CALL keyword_create(keyword, name="NGRIDS", &
                          description="The number of multigrids to use", &
                          usage="ngrids 1", default_i_val=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword=keyword, name="cutoff", &
                          description="The cutoff of the finest grid level. Default value for "// &
                          "SE or DFTB calculation is 1.0 [Ry].", &
                          usage="cutoff 300", default_r_val=cp_unit_to_cp2k(value=280.0_dp, &
                                                                            unit_str="Ry"), n_var=1, unit_str="Ry")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="progression_factor", &
                          description="Factor used to find the cutoff of the multigrids that"// &
                          " where not given explicitly", &
                          usage="progression_factor <integer>", default_r_val=3._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="commensurate", &
                          description="If the grids should be commensurate. If true overrides "// &
                          "the progression factor and the cutoffs of the sub grids", &
                          usage="commensurate", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="realspace", &
                          description="If both rho and rho_gspace are needed ", &
                          usage="realspace", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REL_CUTOFF", &
                          variants=(/"RELATIVE_CUTOFF"/), &
                          description="Determines the grid at which a Gaussian is mapped,"// &
                          " giving the cutoff used for a gaussian with alpha=1."// &
                          " A value 50+-10Ry might be required for highly accurate results, "// &
                          " Or for simulations with a variable cell."// &
                          " Versions prior to 2.3 used a default of 30Ry.", &
                          usage="RELATIVE_CUTOFF real", default_r_val=20.0_dp, &
                          unit_str="Ry")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MULTIGRID_SET", &
                          description="Activate a manual setting of the multigrids", &
                          usage="MULTIGRID_SET", default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword &
                          , name="SKIP_LOAD_BALANCE_DISTRIBUTED" &
                          , description="Skips load balancing on distributed multigrids.  " &
                          //"Memory usage is O(p) so may be used " &
                          //"for all but the very largest runs." &
                          , usage="SKIP_LOAD_BALANCE_DISTRIBUTED" &
                          , default_l_val=.FALSE. &
                          , lone_keyword_l_val=.TRUE. &
                          )
!    CALL keyword_create(keyword, name="SKIP_LOAD_BALANCE_DISTRIBUTED",&
!         description="Skip load balancing on distributed multigrids, which might be memory intensive."//&
!                     "If not explicitly specified, runs using more than 1024 MPI tasks will default to .TRUE.",&
!         usage="SKIP_LOAD_BALANCE_DISTRIBUTED", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)

      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MULTIGRID_CUTOFF", &
                          variants=(/"CUTOFF_LIST"/), &
                          description="List of cutoff values to set up multigrids manually", &
                          usage="MULTIGRID_CUTOFF 200.0 100.0 ", n_var=-1, type_of_var=real_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL create_rsgrid_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_interp_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_mgrid_section

! **************************************************************************************************
!> \brief creates the interpolation section
!> \param section ...
!> \author tlaino
! **************************************************************************************************
   SUBROUTINE create_interp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_interp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",5094)
      CALL section_create(section, name="interpolator", &
                          description="kind of interpolation used between the multigrids", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, print_key)

      CALL keyword_create(keyword, name="kind", &
                          description="the interpolator to use", &
                          usage="kind spline3", &
                          default_i_val=pw_interp, &
                          enum_c_vals=s2a("pw", "spline3_nopbc", "spline3"), &
                          enum_i_vals=(/pw_interp, &
                                        spline3_nopbc_interp, spline3_pbc_interp/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="safe_computation", &
                          description="if a non unrolled calculation is to be performed in parallel", &
                          usage="safe_computation OFF", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="aint_precond", &
                          description="the approximate inverse to use to get the starting point"// &
                          " for the linear solver of the spline3 methods", &
                          usage="aint_precond copy", &
                          default_i_val=precond_spl3_aint, &
                          enum_c_vals=s2a("copy", "spl3_nopbc_aint1", "spl3_nopbc_aint2", &
                                          "spl3_nopbc_precond1", "spl3_nopbc_precond2", "spl3_nopbc_precond3"), &
                          enum_i_vals=(/no_precond, precond_spl3_aint, precond_spl3_aint2, &
                                        precond_spl3_1, precond_spl3_2, precond_spl3_3/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="precond", &
                          description="The preconditioner used"// &
                          " for the linear solver of the spline3 methods", &
                          usage="PRECOND copy", &
                          default_i_val=precond_spl3_3, &
                          enum_c_vals=s2a("copy", "spl3_nopbc_aint1", "spl3_nopbc_aint2", &
                                          "spl3_nopbc_precond1", "spl3_nopbc_precond2", "spl3_nopbc_precond3"), &
                          enum_i_vals=(/no_precond, precond_spl3_aint, precond_spl3_aint2, &
                                        precond_spl3_1, precond_spl3_2, precond_spl3_3/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="eps_x", &
                          description="accuracy on the solution for spline3 the interpolators", &
                          usage="eps_x 1.e-15", default_r_val=1.e-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="eps_r", &
                          description="accuracy on the residual for spline3 the interpolators", &
                          usage="eps_r 1.e-15", default_r_val=1.e-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="max_iter", &
                          variants=(/'maxiter'/), &
                          description="the maximum number of iterations", &
                          usage="max_iter 200", default_i_val=100)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "conv_info", &
                                       description="if convergence information about the linear solver"// &
                                       " of the spline methods should be printed", &
                                       print_level=medium_print_level, each_iter_names=s2a("SPLINE_FIND_COEFFS"), &
                                       each_iter_values=(/10/), filename="__STD_OUT__", &
                                       add_last=add_last_numeric)
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_interp_section

! **************************************************************************************************
!> \brief creates the sic (self interaction correction) section
!> \param section ...
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE create_sic_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_sic_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",5187)
      CALL section_create(section, "sic", &
                          description="parameters for the self interaction correction", &
                          n_keywords=6, n_subsections=0, repeats=.FALSE., &
                          citations=(/VandeVondele2005b, Perdew1981, Avezac2005/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="SIC_SCALING_A", &
                          description="Scaling of the coulomb term in sic [experimental]", &
                          usage="SIC_SCALING_A 0.5", &
                          citations=(/VandeVondele2005b/), &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SIC_SCALING_B", &
                          description="Scaling of the xc term in sic [experimental]", &
                          usage="SIC_SCALING_B 0.5", &
                          citations=(/VandeVondele2005b/), &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SIC_METHOD", &
                          description="Method used to remove the self interaction", &
                          usage="SIC_METHOD MAURI_US", &
                          default_i_val=sic_none, &
                          enum_c_vals=s2a("NONE", "MAURI_US", "MAURI_SPZ", "AD", "EXPLICIT_ORBITALS"), &
                          enum_i_vals=(/sic_none, sic_mauri_us, sic_mauri_spz, sic_ad, sic_eo/), &
                          enum_desc=s2a("Do not apply a sic correction", &
                                        "Employ a (scaled) correction proposed by Mauri and co-workers"// &
                                        " on the spin density / doublet unpaired orbital", &
                                        "Employ a (scaled) Perdew-Zunger expression"// &
                                        " on the spin density / doublet unpaired orbital", &
                                        "The average density correction", &
                                        "(scaled) Perdew-Zunger correction explicitly on a set of orbitals."), &
                          citations=(/VandeVondele2005b, Perdew1981, Avezac2005/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ORBITAL_SET", &
                          description="Type of orbitals treated with the SIC", &
                          usage="ORBITAL_SET ALL", &
                          default_i_val=sic_list_unpaired, &
                          enum_c_vals=s2a("UNPAIRED", "ALL"), &
                          enum_desc=s2a("correction for the unpaired orbitals only, requires a restricted open shell calculation", &
                                        "correction for all orbitals, requires a LSD or ROKS calculation"), &
                          enum_i_vals=(/sic_list_unpaired, sic_list_all/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_sic_section

! **************************************************************************************************
!> \brief creates the low spin roks section
!> \param section ...
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE create_low_spin_roks_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_low_spin_roks_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",5254)
      CALL section_create(section, "LOW_SPIN_ROKS", &
                          description="Specify the details of the low spin ROKS method."// &
                          "In particular, one can specify various terms added to the energy of the high spin roks configuration"// &
                          " with a energy scaling factor, and a prescription of the spin state.", &
                          n_keywords=6, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="ENERGY_SCALING", &
                          description="The scaling factors for each term added to the total energy."// &
                          "This list should contain one number for each term added to the total energy.", &
                          usage="ENERGY_SCALING 1.0 -1.0 ", &
                          n_var=-1, type_of_var=real_t, repeats=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create( &
         keyword, name="SPIN_CONFIGURATION", &
         description="for each singly occupied orbital, specify if this should be an alpha (=1) or a beta (=2) orbital"// &
         "This keyword should be repeated, each repetition corresponding to an additional term.", &
         usage="SPIN_CONFIGURATION 1 2", &
         n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_low_spin_roks_section

! **************************************************************************************************
!> \brief Creates the section for applying an electrostatic external potential
!> \param section ...
!> \date 12.2009
!> \author teo
! **************************************************************************************************
   SUBROUTINE create_ext_pot_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ext_pot_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",5294)
      CALL section_create(section, name="EXTERNAL_POTENTIAL", &
                          description="Section controlling the presence of an electrostatic "// &
                          "external potential dependent on the atomic positions (X,Y,Z). "// &
                          "As the external potential is currently applied via a grid, "// &
                          "it only works with DFT based methods (GPW/GAPW) that already use "// &
                          "a grid based approach to solve the Poisson equation.", &
                          n_keywords=7, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="FUNCTION", &
                          description="Specifies the functional form in mathematical notation. Variables must be the atomic "// &
                          "coordinates (X,Y,Z) of the grid.", usage="FUNCTION  X^2+Y^2+Z^2+LOG(ABS(X+Y))", &
                          type_of_var=lchar_t, n_var=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PARAMETERS", &
                          description="Defines the parameters of the functional form", &
                          usage="PARAMETERS a b D", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="VALUES", &
                          description="Defines the values of parameter of the functional form", &
                          usage="VALUES ", type_of_var=real_t, &
                          n_var=-1, repeats=.TRUE., unit_str="internal_cp2k")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="UNITS", &
                          description="Optionally, allows to define valid CP2K unit strings for each parameter value. "// &
                          "It is assumed that the corresponding parameter value is specified in this unit.", &
                          usage="UNITS angstrom eV*angstrom^-1 angstrom^1 K", type_of_var=char_t, &
                          n_var=-1, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STATIC", &
                          description="Specifies the external potential as STATIC or time dependent. At the moment "// &
                          "only static potentials are implemented.", &
                          usage="STATIC T", default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DX", &
                          description="Parameter used for computing the derivative with the Ridders' method.", &
                          usage="DX <REAL>", default_r_val=0.1_dp, unit_str="bohr")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ERROR_LIMIT", &
                          description="Checks that the error in computing the derivative is not larger than "// &
                          "the value set; in case error is larger a warning message is printed.", &
                          usage="ERROR_LIMIT <REAL>", default_r_val=1.0E-12_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      !keyword for reading the external potential from cube file
      CALL keyword_create(keyword, name="READ_FROM_CUBE", &
                          description="Switch for reading the external potential from file pot.cube. The values "// &
                          "of the potential must be on the grid points of the realspace grid.", &
                          usage="READ_FROM_CUBE T", default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      !keyword for scaling the external potential that is read from file by a constant factor
      CALL keyword_create(keyword, name="SCALING_FACTOR", &
                          description="A factor for scaling the the external potential that is read from file."// &
                          "The value of the potential at each grid point is multiplied by this factor.", &
                          usage="SCALING_FACTOR <REAL>", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_ext_pot_section

! **************************************************************************************************
!> \brief ZMP Creates the section for reading user supplied external density
!> \param section ...
!> \date 03.2011
!> \author D. Varsano [daniele.varsano@nano.cnr.it]
! **************************************************************************************************
   SUBROUTINE create_ext_den_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ext_den_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",5384)
      CALL section_create(section, name="EXTERNAL_DENSITY", &
                          description="Section for the use of the ZMP technique on external densities.", &
                          n_keywords=4, n_subsections=0, repeats=.FALSE., &
                          citations=(/Zhao1994, Tozer1996/))
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="FILE_DENSITY", &
                          description="Specifies the filename containing the target density in *.cube format."// &
                          "In the MGRID section it must be imposed NGRID 1, as it works with only"// &
                          "one grid. The number of points in each direction, and the spacing must"// &
                          "be previously defined choosing the plane waves cut-off in section MGRID"// &
                          "keyword CUTOFF, and the cube dimention in section SUBSYS / CELL / keyword ABC", &
                          usage="DENSITY_FILE_NAME <FILENAME>", &
                          type_of_var=char_t, default_c_val="RHO_O.dat", n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LAMBDA", &
                          description="Lagrange multiplier defined in the constraint ZMP method. When starting, use"// &
                          "small values when starting from scratch (around 5,10). Then gradually increase"// &
                          "the values depending, restarting from the previous calculation with the smaller"// &
                          "value. To choose the progressive values of LAMBDA look at the convergence of the"// &
                          " eigenvalues.", &
                          usage="DX <REAL>", default_r_val=10.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ZMP_CONSTRAINT", &
                          description="Specify which kind of constraint to solve the ZMP equation. The COULOMB default"// &
                          "option is more stable.", &
                          usage="ZMP_CONSTRAINT <CHAR>", &
                          enum_c_vals=s2a("COULOMB", "DIFF", "NONE"), &
                          enum_i_vals=(/use_coulomb, use_diff, use_no/), &
                          enum_desc=s2a("Coulomb constraint, integral of [rho_0(r)-rho(r)]/|r-r'|", &
                                        "Simple constraint, [rho_0(r)-rho(r)]", &
                                        "No constrain imposed"), &
                          default_i_val=use_coulomb)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FERMI_AMALDI", &
                          description="Add the Fermi-Amaldi contribution to the Hartree potential."// &
                          "It leads to a more stable convergence.", &
                          usage="FERMI_AMALDI <LOGICAL>", &
                          repeats=.FALSE., &
                          n_var=1, &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_ext_den_section

! **************************************************************************************************
!> \brief ZMP Creates the section for creating the external v_xc
!> \param section ...
!> \date 03.2011
!> \author D. Varsano [daniele.varsano@nano.cnr.it]
! **************************************************************************************************
   SUBROUTINE create_ext_vxc_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ext_vxc_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",5451)
      CALL section_create(section, name="EXTERNAL_VXC", &
                          description="SCF convergence with external v_xc calculated through previous ZMP"// &
                          "calculation", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)
      NULLIFY (keyword)

      CALL keyword_create(keyword, name="FILE_VXC", &
                          description="The *.cube filename containing the v_xc potential. This works only"// &
                          "with NGRID 1 imposed in the MGRID section. The number of points in each"// &
                          "direction, and the spacing must equal to those previously used in the ZMP"// &
                          "calculation and defined through the plane wave cut-off and the cube dimension"// &
                          "respectively set in section MGRID / keyword CUTOFF, and in section SUBSYS /"// &
                          "CELL / keyword ABC", &
                          usage="FILE_VXC <FILENAME>", &
                          type_of_var=char_t, default_c_val="VXC_O.dat", n_var=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)
   END SUBROUTINE create_ext_vxc_section

! **************************************************************************************************
!> \brief creates the section for static periodic fields
!> \param section ...
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE create_per_efield_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_per_efield_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",5484)
      CALL section_create(section, "PERIODIC_EFIELD", &
                          description="parameters for finite periodic electric field computed using "// &
                          "the Berry phase approach. IMPORTANT: Can only be used in combination "// &
                          " with OT. Can not be used in combination with RTP or EMD.", &
                          citations=(/Souza2002, Umari2002/), &
                          n_keywords=6, n_subsections=1, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="INTENSITY", &
                          description="Intensity of the electric field in a.u", &
                          usage="INTENSITY  0.001", &
                          default_r_val=0._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POLARISATION", &
                          description="Polarisation vector of electric field", &
                          usage="POLARISIATION  0.0 0.0 1.0", &
                          repeats=.FALSE., n_var=3, &
                          type_of_var=real_t, default_r_vals=(/0.0_dp, 0.0_dp, 1.0_dp/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DISPLACEMENT_FIELD", &
                          description="Use the displacement field formulation.", &
                          usage="DISPLACEMENT_FIELD T", &
                          citations=(/Stengel2009/), &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_per_efield_section
! **************************************************************************************************
!> \brief creates the section for time dependent nonperiodic fields
!> \param section ...
!> \author Florian Schiffmann
! **************************************************************************************************
   SUBROUTINE create_efield_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_efield_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",5533)
      CALL section_create(section, "EFIELD", &
                          description="parameters for finite, time dependent, nonperiodic electric fields", &
                          n_keywords=6, n_subsections=1, repeats=.TRUE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="INTENSITY", &
                          description="Intensity of the electric field. For real-time propagation (RTP) units are "// &
                          "in W*cm-2 which corresponds "// &
                          "to a maximal amplitude in a.u. of sqrt(I/(3.50944*10^16)). "// &
                          "For a constant local field in isolated system calclulations, units are in a.u..", &
                          usage="INTENSITY  0.001", &
                          default_r_val=0._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="POLARISATION", &
                          description="Polarisation vector of electric field", &
                          usage="POLARISATION  0.0 0.0 1.0", &
                          repeats=.FALSE., n_var=3, &
                          type_of_var=real_t, default_r_vals=(/0.0_dp, 0.0_dp, 1.0_dp/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WAVELENGTH", &
                          description="Wavelength of efield field for real-time propagation (RTP) calculations.", &
                          usage="Wavelength  1.E0", &
                          default_r_val=0._dp, unit_str="nm")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PHASE", &
                          description="Phase offset of the cosine given in multiples of pi. "// &
                          "Used in real-time propagation (RTP) calculations.", &
                          usage="Phase  1.E0", &
                          default_r_val=0._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ENVELOP", &
                          description="Shape of the efield pulse used in real-time propagation (RTP) calculations.", &
                          usage="ENVELOP CONSTANT", &
                          default_i_val=constant_env, &
                          enum_c_vals=s2a("CONSTANT", "GAUSSIAN", "RAMP", "CUSTOM"), &
                          enum_desc=s2a("No envelop function is applied to the strength", &
                                        "A Gaussian function is used as envelop ", &
                                        "Linear tune in/out of the field", &
                                        "A custom field read from a file"), &
                          enum_i_vals=(/constant_env, gaussian_env, ramp_env, custom_env/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_constant_env_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_gaussian_env_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_ramp_env_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_custom_env_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_efield_section

! **************************************************************************************************
!> \brief makes the orbital transformation section
!> \param section ...
!> \par History
!>      11.2004 created [Joost VandeVondele]
! **************************************************************************************************
   SUBROUTINE create_ot_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ot_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",5618)
      CALL section_create(section, "OT", &
                          description="Sets the various options for the orbital transformation (OT) method. "// &
                          "Default settings already provide an efficient, yet robust method. "// &
                          "Most systems benefit from using the FULL_ALL preconditioner "// &
                          "combined with a small value (0.001) of ENERGY_GAP. "// &
                          "Well-behaved systems might benefit from using a DIIS minimizer.", &
                          n_keywords=27, n_subsections=0, repeats=.FALSE., &
                          citations=(/VandeVondele2003, Weber2008/))

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the ot method", &
                          usage="&OT T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALGORITHM", &
                          description="Algorithm to be used for OT", &
                          usage="ALGORITHM STRICT", &
                          default_i_val=ot_algo_taylor_or_diag, &
                          enum_c_vals=s2a("STRICT", "IRAC"), &
                          enum_desc=s2a("Strict orthogonality: Taylor or diagonalization based algorithm.", &
                                        "Orbital Transformation based Iterative Refinement "// &
                                        "of the Approximative Congruence transformation (OT/IR)."), &
                          enum_i_vals=(/ot_algo_taylor_or_diag, ot_algo_irac/), &
                          citations=(/VandeVondele2003, VandeVondele2005a, Weber2008/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="IRAC_DEGREE", &
                          description="The refinement polynomial degree (2, 3 or 4).", &
                          usage="IRAC_DEGREE 4", &
                          default_i_val=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_IRAC", &
                          description="Maximum allowed refinement iteration.", &
                          usage="MAX_IRAC 5", &
                          default_i_val=50)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MIXED_PRECISION", &
                          description="Uses a mixed precision algorithm."// &
                          "With a well behaved basis set (i.e. condition number less than 1/eps_sp)"// &
                          "it provides double precision accuracy results and up to a 2 fold speedup for building and "// &
                          "applying the preconditioner.", &
                          usage="MIXED_PRECISION T", &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ORTHO_IRAC", &
                          description="The orthogonality method.", &
                          usage="ORTHO_IRAC POLY", &
                          default_i_val=ot_chol_irac, &
                          enum_c_vals=s2a("CHOL", "POLY", "LWDN"), &
                          enum_desc=s2a("Cholesky.", "Polynomial.", "Loewdin."), &
                          enum_i_vals=(/ot_chol_irac, ot_poly_irac, ot_lwdn_irac/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_IRAC_FILTER_MATRIX", &
                          description="Sets the threshold for filtering the matrices.", &
                          usage="EPS_IRAC_FILTER_MATRIX 1.0E-5", &
                          default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_IRAC", &
                          description="Targeted accuracy during the refinement iteration.", &
                          usage="EPS_IRAC 1.0E-5", &
                          default_r_val=1.0E-10_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_IRAC_QUICK_EXIT", &
                          description="Only one extra refinement iteration is "// &
                          "done when the norm is below this value.", &
                          usage="EPS_IRAC_QUICK_EXIT 1.0E-2", &
                          default_r_val=1.0E-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_IRAC_SWITCH", &
                          description="The algorithm switches to the polynomial "// &
                          "refinement when the norm is below this value.", &
                          usage="EPS_IRAC_SWITCH 1.0E-3", &
                          default_r_val=1.0E-2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ON_THE_FLY_LOC", &
                          description="On the fly localization of the molecular orbitals. "// &
                          "Can only be used with OT/IRAC.", &
                          usage="ON_THE_FLY_LOC T", &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="MINIMIZER", &
         description="Minimizer to be used with the OT method", &
         usage="MINIMIZER DIIS", &
         default_i_val=ot_mini_cg, &
         enum_c_vals=s2a("SD", "CG", "DIIS", "BROYDEN"), &
         enum_desc=s2a("Steepest descent: not recommended", "Conjugate Gradients: most reliable, use for difficult systems."// &
                       " The total energy should decrease at every OT CG step if the line search is appropriate.", &
                       "Direct inversion in the iterative subspace: less reliable than CG, but sometimes about 50% faster", &
                       "Broyden mixing approximating the inverse Hessian"), &
         enum_i_vals=(/ot_mini_sd, ot_mini_cg, ot_mini_diis, ot_mini_broyden/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SAFE_DIIS", &
                          variants=(/"SAFER_DIIS"/), &
                          description="Reject DIIS steps if they point away from the"// &
                          " minimum, do SD in that case.", &
                          usage="SAFE_DIIS ON", default_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="N_HISTORY_VEC", &
                          variants=s2a("NDIIS", "N_DIIS", "N_BROYDEN"), &
                          description="Number of history vectors to be used with DIIS or BROYDEN", &
                          usage="N_DIIS 4", &
                          default_i_val=7)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BROYDEN_BETA", &
                          description="Underrelaxation for the broyden mixer", &
                          usage="BROYDEN_BETA 0.9", &
                          default_r_val=0.9_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BROYDEN_GAMMA", &
                          description="Backtracking parameter", &
                          usage="BROYDEN_GAMMA 0.5", &
                          default_r_val=0.5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BROYDEN_SIGMA", &
                          description="Curvature of energy functional.", &
                          usage="BROYDEN_SIGMA 0.25", &
                          default_r_val=0.25_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BROYDEN_ETA", &
                          description="Dampening of estimated energy curvature.", &
                          usage="BROYDEN_ETA 0.7", &
                          default_r_val=0.7_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BROYDEN_OMEGA", &
                          description="Growth limit of curvature.", &
                          usage="BROYDEN_OMEGA 1.1", &
                          default_r_val=1.1_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BROYDEN_SIGMA_DECREASE", &
                          description="Reduction of curvature on bad approximation.", &
                          usage="BROYDEN_SIGMA_DECREASE 0.7", &
                          default_r_val=0.7_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BROYDEN_SIGMA_MIN", &
                          description="Minimum adaptive curvature.", &
                          usage="BROYDEN_SIGMA_MIN 0.05", &
                          default_r_val=0.05_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BROYDEN_FORGET_HISTORY", &
                          description="Forget history on bad approximation", &
                          usage="BROYDEN_FORGET_HISTORY OFF", default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BROYDEN_ADAPTIVE_SIGMA", &
                          description="Enable adaptive curvature estimation", &
                          usage="BROYDEN_ADAPTIVE_SIGMA ON", default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BROYDEN_ENABLE_FLIP", &
                          description="Ensure positive definite update", &
                          usage="BROYDEN_ENABLE_FLIP ON", default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="LINESEARCH", &
                          variants=(/"LINE_SEARCH"/), &
                          description="1D line search algorithm to be used with the OT minimizer,"// &
                          " in increasing order of robustness and cost. MINIMIZER CG combined with"// &
                          " LINESEARCH GOLD should always find an electronic minimum. "// &
                          " Whereas the 2PNT minimizer is almost always OK, 3PNT might be needed for systems"// &
                          " in which successive OT CG steps do not decrease the total energy.", &
                          usage="LINESEARCH GOLD", &
                          default_i_val=ls_2pnt, &
                          enum_c_vals=s2a("NONE", "2PNT", "3PNT", "GOLD"), &
                          enum_desc=s2a("take fixed lenght steps", "extrapolate based on 2 points", &
                                        "... or on 3 points", "perform 1D golden section search of the minimum (very expensive)"), &
                          enum_i_vals=(/ls_none, ls_2pnt, ls_3pnt, ls_gold/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="STEPSIZE", &
         description="Initial stepsize used for the line search, sometimes this parameter can be reduced to stablize DIIS"// &
         " or to improve the CG behavior in the first few steps."// &
         " The optimal value depends on the quality of the preconditioner."// &
         " A negative values leaves the choice to CP2K depending on the preconditioner.", &
         usage="STEPSIZE 0.4", &
         default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="GOLD_TARGET", &
                          description="Target relative uncertainty in the location of the minimum for LINESEARCH GOLD", &
                          usage="GOLD_TARGET 0.1", &
                          default_r_val=0.01_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="PRECONDITIONER", &
         description="Type of preconditioner to be used with all minimization schemes. "// &
         "They differ in effectiveness, cost of construction, cost of application. "// &
         "Properly preconditioned minimization can be orders of magnitude faster than doing nothing.", &
         usage="PRECONDITIONER FULL_ALL", &
         default_i_val=ot_precond_full_kinetic, &
         enum_c_vals=s2a("FULL_ALL", "FULL_SINGLE_INVERSE", "FULL_SINGLE", "FULL_KINETIC", "FULL_S_INVERSE", &
                         "NONE"), &
         enum_desc=s2a("Most effective state selective preconditioner based on diagonalization, "// &
                       "requires the ENERGY_GAP parameter to be an underestimate of the HOMO-LUMO gap. "// &
                       "This preconditioner is recommended for almost all systems, except very large systems where "// &
                       "make_preconditioner would dominate the total computational cost.", &
                       "Based on H-eS cholesky inversion, similar to FULL_SINGLE in preconditioning efficiency "// &
                       "but cheaper to construct, "// &
                       "might be somewhat less robust. Recommended for large systems.", &
                       "Based on H-eS diagonalisation, not as good as FULL_ALL, but somewhat cheaper to apply. ", &
                       "Cholesky inversion of S and T, fast construction, robust, and relatively good, "// &
                       "use for very large systems.", &
                       "Cholesky inversion of S, not as good as FULL_KINETIC, yet equally expensive.", &
                       "skip preconditioning"), &
         enum_i_vals=(/ot_precond_full_all, ot_precond_full_single_inverse, ot_precond_full_single, &
                       ot_precond_full_kinetic, ot_precond_s_inverse, ot_precond_none/), &
         citations=(/VandeVondele2003, Weber2008, Schiffmann2015/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHOLESKY", &
                          description="If FULL_ALL the cholesky decomposition of the S matrix is used. "// &
                          "Options on the algorithm to be used.", &
                          usage="CHOLESKY REDUCE", default_i_val=cholesky_reduce, &
                          enum_c_vals=s2a("OFF", "REDUCE", "RESTORE", "INVERSE", "INVERSE_DBCSR"), &
                          enum_desc=s2a("The cholesky algorithm is not used", "Reduce is called", &
                                        "Reduce is replaced by two restore", &
                                        "Restore uses operator multiply by inverse of the triangular matrix", &
                                        "Like inverse, but matrix stored as dbcsr, sparce matrix algebra used when possible"), &
                          enum_i_vals=(/cholesky_off, cholesky_reduce, cholesky_restore, cholesky_inverse, cholesky_dbcsr/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="PRECOND_SOLVER", &
         description="How the preconditioner is applied to the residual.", &
         usage="PRECOND_SOLVER DIRECT", &
         default_i_val=ot_precond_solver_default, &
         enum_c_vals=s2a("DEFAULT", "DIRECT", "INVERSE_CHOLESKY", "INVERSE_UPDATE"), &
         enum_desc=s2a("the default", "Cholesky decomposition followed by triangular solve "// &
                       "(works for FULL_KINETIC/SINGLE_INVERSE/S_INVERSE)", &
                       "Cholesky decomposition followed by explicit inversion "// &
                       "(works for FULL_KINETIC/SINGLE_INVERSE/S_INVERSE)", &
                       "Performs a Hotelling update of the inverse if a previous preconditioner is present. "// &
                       "Mainly useful for GPU accelerated systems (works for FULL_KINETIC/SINGLE_INVERSE/S_INVERSE)"), &
         enum_i_vals=(/ot_precond_solver_default, &
                       ot_precond_solver_direct, &
                       ot_precond_solver_inv_chol, &
                       ot_precond_solver_update/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="ENERGY_GAP", &
         description="Should be an estimate for the energy gap [a.u.] (HOMO-LUMO) and is used in preconditioning, "// &
         "especially effective with the FULL_ALL preconditioner, in which case it should be an underestimate "// &
         "of the gap (can be a small number, e.g. 0.002)."// &
         " FULL_SINGLE_INVERSE takes it as lower bound (values below 0.05 can cause stability issues)."// &
         " In general, heigher values will tame the preconditioner in case of poor initial guesses."// &
         " A negative value will leave the choice to CP2K depending on type of preconditioner.", &
         usage="ENERGY_GAP 0.001", &
         default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EPS_TAYLOR", &
         variants=(/"EPSTAYLOR"/), &
         description="Target accuracy of the taylor expansion for the matrix functions, should normally be kept as is.", &
         usage="EPS_TAYLOR 1.0E-15", &
         default_r_val=1.0E-16_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="MAX_TAYLOR", &
         description="Maximum order of the Taylor expansion before diagonalisation is prefered, for large parallel runs"// &
         " a slightly higher order could sometimes result in a small speedup.", &
         usage="MAX_TAYLOR 5", &
         default_i_val=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ROTATION", &
                          description="Introduce additional variables so that rotations of the occupied"// &
                          " subspace are allowed as well, only needed for cases where the energy is not invariant under "// &
                          " a rotation of the occupied subspace such as non-singlet restricted calculations "// &
                          " or fractional occupations.", &
                          usage="ROTATION", lone_keyword_l_val=.TRUE., &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ENERGIES", &
                          description="Optimize orbital energies for use in Fermi-Dirac smearing "// &
                          "(requires ROTATION and FD smearing to be active).", &
                          usage="ENERGIES", lone_keyword_l_val=.TRUE., &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OCCUPATION_PRECONDITIONER", &
                          description="Preconditioner with the occupation numbers (FD smearing)", &
                          usage="OCCUPATION_PRECONDITIONER", lone_keyword_l_val=.TRUE., &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NONDIAG_ENERGY", &
                          description="Add a non-diagonal energy penalty (FD smearing)", &
                          usage="NONDIAG_ENERGY", lone_keyword_l_val=.TRUE., &
                          default_l_val=.FALSE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NONDIAG_ENERGY_STRENGTH", &
                          description="The prefactor for the non-diagonal energy penalty (FD smearing)", &
                          usage="NONDIAG_ENERGY_STRENGTH", default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_ot_section

! **************************************************************************************************
!> \brief creates the diagonalization section
!> \param section ...
!> \par History
!>      10.2008 created [JGH]
! **************************************************************************************************
   SUBROUTINE create_diagonalization_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_diagonalization_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",6002)
      CALL section_create(section, "DIAGONALIZATION", &
                          description="Set up type and parameters for Kohn-Sham matrix diagonalization.", &
                          n_keywords=0, n_subsections=1, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of the diagonalization method", &
                          usage="&DIAGONALIZATION T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALGORITHM", &
                          description="Algorithm to be used for diagonalization", &
                          usage="ALGORITHM STANDARD", &
                          default_i_val=diag_standard, &
                          enum_c_vals=s2a("STANDARD", "OT", "LANCZOS", "DAVIDSON", "FILTER_MATRIX"), &
                          enum_desc=s2a("Standard diagonalization: LAPACK methods or Jacobi.", &
                                        "Iterative diagonalization using OT method", &
                                        "Block Krylov-space approach to self-consistent diagonalisation", &
                                        "Preconditioned blocked Davidson", &
                                        "Filter matrix diagonalization"), &
                          enum_i_vals=(/diag_standard, diag_ot, diag_block_krylov, diag_block_davidson, &
                                        diag_filter_matrix/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="JACOBI_THRESHOLD", &
                          description="Controls the accuracy of the pseudo-diagonalization method using Jacobi rotations", &
                          usage="JACOBI_THRESHOLD 1.0E-6", &
                          default_r_val=1.0E-7_dp, &
                          citations=(/Stewart1982/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_JACOBI", &
                          description="Below this threshold value for the SCF convergence the pseudo-diagonalization "// &
                          "method using Jacobi rotations is activated. This method is much faster than a "// &
                          "real diagonalization and it is even speeding up while achieving full convergence."// &
                          "However, it needs a pre-converged wavefunction obtained by at least one real "// &
                          "diagonalization which is further optimized while keeping the original eigenvalue "// &
                          "spectrum. The MO eigenvalues are NOT updated. The method might be useful to speed "// &
                          "up calculations for large systems e.g. using a semi-empirical method.", &
                          usage="EPS_JACOBI 1.0E-5", &
                          default_r_val=0.0_dp, &
                          citations=(/Stewart1982/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_ADAPT", &
                          description="Required accuracy in iterative diagonalization as compared to current SCF convergence", &
                          usage="EPS_ADAPT 0.01", &
                          default_r_val=0._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ITER", &
                          description="Maximum number of iterations in iterative diagonalization", &
                          usage="MAX_ITER 20", &
                          default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_ITER", &
                          description="Required accuracy in iterative diagonalization", &
                          usage="EPS_ITER 1.e-8", &
                          default_r_val=1.e-8_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)
      CALL create_ot_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_krylov_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_diag_subspace_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_davidson_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      NULLIFY (subsection)
      CALL create_filtermatrix_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_diagonalization_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_davidson_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_davidson_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",6114)
      CALL section_create(section, "DAVIDSON", &
                          description=" ", &
                          n_keywords=2, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create( &
         keyword, name="PRECONDITIONER", &
         description="Type of preconditioner to be used with all minimization schemes. ", &
         usage="PRECONDITIONER FULL_ALL", &
         default_i_val=ot_precond_full_all, &
         enum_c_vals=s2a("FULL_ALL", "FULL_SINGLE_INVERSE", "NONE"), &
         enum_desc=s2a("Most effective state selective preconditioner based on diagonalization ", &
                       "Based on H-eS cholesky inversion, similar to FULL_SINGLE in preconditioning efficiency "// &
                       "but cheaper to construct, might be somewhat less robust. Recommended for large systems.", &
                       "skip preconditioning"), &
         enum_i_vals=(/ot_precond_full_all, ot_precond_full_single_inverse, ot_precond_none/), &
         citations=(/VandeVondele2003/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRECOND_SOLVER", &
                          description="How the preconditioner is applied to the residual.", &
                          usage="PRECOND_SOLVER DIRECT", &
                          default_i_val=ot_precond_solver_default, &
                          enum_c_vals=s2a("DEFAULT", "DIRECT", "INVERSE_CHOLESKY"), &
                          enum_desc=s2a("the default", "Cholesky decomposition followed by triangular solve "// &
                                        "(works for FULL_KINETIC/SINGLE_INVERSE/S_INVERSE)", &
                                        "Cholesky decomposition followed by explicit inversion "// &
                                        "(works for FULL_KINETIC/SINGLE_INVERSE/S_INVERSE)"), &
                          enum_i_vals=(/ot_precond_solver_default, &
                                        ot_precond_solver_direct, &
                                        ot_precond_solver_inv_chol/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="ENERGY_GAP", &
         description="Should be an estimate for the energy gap [a.u.] (HOMO-LUMO) and is used in preconditioning, "// &
         "especially effective with the FULL_ALL preconditioner, in which case it should be an underestimate "// &
         "of the gap (0.001 doing normally fine). For the other preconditioners, making this value larger (0.2)"// &
         " will tame the preconditioner in case of poor initial guesses.", &
         usage="ENERGY_GAP 0.001", &
         default_r_val=0.2_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NEW_PREC_EACH", &
                          description="Number of SCF iterations after which a new Preconditioner is computed", &
                          usage="NEW_PREC_EACH 10", default_i_val=20)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FIRST_PREC", &
                          description="First SCF iteration at which a Preconditioner is employed", &
                          usage="FIRST_PREC 1", default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CONV_MOS_PERCENT", &
                          description="Minimal percent of MOS that have to converge within the Davidson loop"// &
                          " before the SCF iteration is completed and a new Hamiltonian is computed", &
                          usage="CONV_MOS_PERCENT 0.8", default_r_val=0.5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SPARSE_MOS", &
                          description="Use MOS as sparse matrix and avoid as much as possible multiplications with full matrices", &
                          usage="SPARSE_MOS", default_l_val=.TRUE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_davidson_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_krylov_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_krylov_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",6202)
      CALL section_create(section, "KRYLOV", &
                          description=" ", &
                          n_keywords=2, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="NKRYLOV", &
                          description="Dimension of the Krylov space used for the Lanczos refinement", &
                          usage="NKRYLOV 20", &
                          default_i_val=4)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NBLOCK", &
                          description="Size of the block of vectors refined simultaneously by the Lanczos procedure", &
                          usage="NBLOCK 1", &
                          default_i_val=32)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_KRYLOV", &
                          description="Convergence criterion for the MOs", &
                          usage="EPS_KRYLOV 0.00001", &
                          default_r_val=0.0000001_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_STD_DIAG", &
                          description="Level of convergence to be reached before starting the Lanczos procedure."// &
                          " Above this threshold a standard diagonalization method is used. "// &
                          " If negative Lanczos is started at the first iteration", &
                          usage="EPS_STD_DIAG 0.001", &
                          default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CHECK_MOS_CONV", &
                          description="This requires to check the convergence of MOS also when standard "// &
                          "diagonalization steps are performed, if the block krylov approach is active.", &
                          usage="CHECK_MOS_CONV T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_krylov_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_diag_subspace_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_diag_subspace_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",6263)
      CALL section_create(section, "DIAG_SUB_SCF", &
                          description="Activation of self-consistenf subspace refinement by diagonalization "// &
                          "of H by adjusting the occupation but keeping the MOS unchanged.", &
                          n_keywords=2, n_subsections=1, repeats=.FALSE.)

      NULLIFY (keyword, subsection)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of inner SCF loop to refine occupations in MOS subspace", &
                          usage="&DIAG_SUB_SCF T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ITER", &
                          description="Maximum number of iterations for the SCF inner loop", &
                          usage="MAX_ITER 20", &
                          default_i_val=2)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_ENE", &
                          description="Required energy accuracy for convergence of subspace diagonalization", &
                          usage="EPS_ENE 1.e-8", &
                          default_r_val=1.e-4_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_ADAPT_SCF", &
                          description="Required density matrix accuracy as compared to current SCF convergence", &
                          usage="EPS_ADAPT_SCF 1.e-1", &
                          default_r_val=1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EPS_SKIP_SUB_DIAG", &
         description="Level of convergence to be reached before starting the internal loop of subspace rotations."// &
         " Above this threshold only the outer diagonalization method is used. "// &
         " If negative the subspace rotation is started at the first iteration", &
         usage="EPS_SKIP_SUB_DIAG 0.001", &
         default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_mixing_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)
   END SUBROUTINE create_diag_subspace_section

! **************************************************************************************************
!> \brief makes the input section for core-level spectroscopy simulations
!> \param section ...
!> \par History
!>      03.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE create_xas_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_xas_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",6330)
      CALL section_create(section, "xas", &
                          description="Sets the method of choice to calculate core-level excitation spectra. "// &
                          "The occupied states from  which we calculate the "// &
                          "excitation should be specified. "// &
                          "Localization of the orbitals may be useful.", &
                          n_keywords=10, n_subsections=1, repeats=.FALSE., &
                          citations=(/Iannuzzi2007/))

      NULLIFY (keyword, subsection, print_key)

      CALL keyword_create(keyword, name="_SECTION_PARAMETERS_", &
                          description="controls the activation of core-level spectroscopy simulations", &
                          usage="&XAS T", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="METHOD", &
                          variants=(/"XAS_METHOD"/), &
                          description="Method to be used to calculate core-level excitation spectra", &
                          usage="METHOD TP_HH", &
                          default_i_val=xas_none, &
                          enum_c_vals=s2a("NONE", "TP_HH", "TP_FH", "TP_VAL", "TP_XHH", "TP_XFH", "DSCF", "TP_FLEX"), &
                          enum_desc=s2a( &
                          "No core electron spectroscopy", "Transition potential half-hole", &
                          "Transition potential full-hole", "Hole in homo for X-ray emission only ", &
                          "Transition potential excited half-hole", &
                          "Transition potential excited full-hole ", &
                          "DSCF calculations to compute the first (core)excited state", &
                          "Transition potential with generalized core occupation and total number of electrons"), &
                          enum_i_vals=(/xas_none, xas_tp_hh, xas_tp_fh, xes_tp_val, xas_tp_xhh, &
                                        xas_tp_xfh, xas_dscf, xas_tp_flex/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="XAS_CORE", &
                          description="Occupation of the core state in XAS calculation by TP_FLEX.", &
                          usage="XAS_CORE 0.5", &
                          default_r_val=0.5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="XAS_TOT_EL", &
                          description="Total number of electrons for spin channel alpha, in XAS calculation by TP_FLEX."// &
                          "If it is a negative value, the number of electrons is set to GS number of electrons "// &
                          "minus the amount subtracted from the core state", &
                          usage="XAS_TOT_EL 10", &
                          default_r_val=-1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="XES_CORE", &
                          description="Occupation of the core state in XES calculation by TP_VAL. "// &
                          "The HOMO is emptied by the same amount.", &
                          usage="XES_CORE 0.5", &
                          default_r_val=1._dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="XES_EMPTY_HOMO", &
                          description="Set the occupation of the HOMO in XES calculation by TP_VAL. "// &
                          "The HOMO can be emptied or not, if the core is still full.", &
                          usage="XES_EMPTY_HOMO", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DIPOLE_FORM", &
                          variants=(/"DIP_FORM"/), &
                          description="Type of integral to get the oscillator strengths "// &
                          "in the diipole approximation", &
                          usage="DIPOLE_FORM string", &
                          default_i_val=xas_dip_vel, &
                          enum_c_vals=s2a("LENGTH", "VELOCITY"), &
                          enum_desc=s2a("Length form &lang; i | e r | j &rang;", &
                                        "Velocity form &lang; i | d/dr | j &rang;"), &
                          enum_i_vals=(/xas_dip_len2, xas_dip_vel/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

!   replace the specialized keyword with standard scf section
!    scf_env is added to xas_env

      NULLIFY (subsection)
      CALL create_scf_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL keyword_create(keyword, name="STATE_TYPE", &
                          variants=(/"TYPE"/), &
                          description="Type of the orbitas that are excited for the xas spectra calculation", &
                          usage="STATE_TYPE 1S", &
                          default_i_val=xas_1s_type, &
                          enum_c_vals=s2a("1S", "2S", "2P"), &
                          enum_desc=s2a("1s orbitals", "2s orbitals", "2p orbitals"), &
                          enum_i_vals=(/xas_1s_type, xas_2s_type, xas_2p_type/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STATE_SEARCH", &
                          description="# of states where to look for the one to be excited", &
                          usage="STATE_SEARCH 1", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ATOMS_LIST", &
                          variants=(/"AT_LIST"/), &
                          description="Indexes of the atoms to be excited. "// &
                          "This keyword can be repeated several times"// &
                          "(useful if you have to specify many indexes).", &
                          usage="ATOMS_LIST {integer}  {integer} ..  {integer} ", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ADDED_MOS", &
                          description="Number of additional MOS added spin up only", &
                          usage="ADDED_MOS {integer}", default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAX_ITER_ADDED", &
                          description="maximum number of iteration in calculation of added orbitals", &
                          usage="MAX_ITER_ADDED 100", default_i_val=2999)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_ADDED", &
                          description="target accuracy incalculation of the added orbitals", &
                          usage="EPS_ADDED 1.e-6", default_r_val=1.0e-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="NGAUSS", &
                          description="Number of gto's for the expansion of the STO "// &
                          "of the type given by STATE_TYPE", &
                          usage="NGAUSS {integer}", default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="RESTART", &
                          description="Restart the excited state if the restart file exists", &
                          usage="RESTART", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="WFN_RESTART_FILE_NAME", &
                          variants=(/"RESTART_FILE_NAME"/), &
                          description="Root of the file names where to read the MOS from "// &
                          "which to restart the calculation of the core level excited states", &
                          usage="WFN_RESTART_FILE_NAME <FILENAME>", &
                          type_of_var=lchar_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_localize_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, "PRINT", &
                          "printing of information during the core-level spectroscopy simulation", &
                          repeats=.FALSE.)

      ! Add printing of wannier infos
      CALL print_wanniers(subsection)

      CALL cp_print_key_section_create(print_key, "iteration_info", &
                                       description="Controls the printing of basic iteration information during the XAS SCF.", &
                                       print_level=low_print_level, filename="__STD_OUT__")
      CALL keyword_create(keyword, name="time_cumul", &
                          description="If the printkey is activated switches the printing of timings"// &
                          " to cumulative (over the SCF).", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "program_run_info", &
                                       description="Controls the printing of basic iteration information in CLS", &
                                       print_level=low_print_level, add_last=add_last_numeric, filename="__STD_OUT__")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "XES_SPECTRUM", &
                                       description="Controls the dumping of the CLS output files containing the emission spectra", &
                                       print_level=low_print_level, common_iter_levels=3, filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create( &
         print_key, "XAS_SPECTRUM", &
         description="Controls the dumping of the CLS output files containing the absorption spectra", &
         print_level=low_print_level, common_iter_levels=3, filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL create_pdos_section(print_key)
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RESTART", &
                                       description="Controls the dumping of MO restart file during the SCF."// &
                                       "of a Core-Level-Spectroscopy calculation. For each new excited atom,"// &
                                       "one different restart file is dumped. These restart files should be"// &
                                       "employed only to restart the same type of CLS calculation, "// &
                                       "i.e. with the same core potential.", &
                                       print_level=low_print_level, common_iter_levels=3, each_iter_names=s2a("XAS_SCF"), &
                                       add_last=add_last_numeric, each_iter_values=(/3/), filename="")
      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CLS_FUNCTION_CUBES", &
                                       description="Controls the printing of the relaxed orbitals ", &
                                       print_level=high_print_level, common_iter_levels=3, add_last=add_last_numeric, filename="")
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUBES_LU_BOUNDS", &
                          variants=(/"CUBES_LU"/), &
                          description="The lower and upper index of the states to be printed as cube", &
                          usage="CUBES_LU_BOUNDS integer integer", &
                          n_var=2, default_i_vals=(/0, -2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUBES_LIST", &
                          description="Indexes of the states to be printed as cube files"// &
                          "This keyword can be repeated several times"// &
                          "(useful if you have to specify many indexes).", &
                          usage="CUBES_LIST 1 2", &
                          n_var=-1, type_of_var=integer_t, repeats=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(subsection, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_xas_section

! **************************************************************************************************
!> \brief      Create CP2K input section for the smearing of occupation numbers
!> \param section ...
!> \date       27.08.2008
!> \author     Matthias Krack (MK)
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE create_smear_section(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_smear_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",6604)

      CALL section_create(section, &
                          name="SMEAR", &
                          description="Define the smearing of the MO occupation numbers", &
                          n_keywords=6, &
                          n_subsections=0, &
                          repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, &
                          name="_SECTION_PARAMETERS_", &
                          description="Controls the activation of smearing", &
                          usage="&SMEAR ON", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="METHOD", &
                          description="Smearing method to be applied", &
                          usage="METHOD Fermi_Dirac", &
                          default_i_val=smear_energy_window, &
                          enum_c_vals=s2a("FERMI_DIRAC", "ENERGY_WINDOW", "LIST"), &
                          enum_i_vals=(/smear_fermi_dirac, smear_energy_window, smear_list/), &
                          enum_desc=s2a("Fermi-Dirac distribution defined by the keyword ELECTRONIC_TEMPERATURE", &
                                        "Energy window defined by the keyword WINDOW_SIZE", &
                                        "Use a fixed list of occupations"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="LIST", &
                          description="A list of fractional occupations to use. Must match the number of states "// &
                          "and sum up to the correct number of electrons", &
                          repeats=.FALSE., &
                          n_var=-1, &
                          type_of_var=real_t, &
                          usage="LIST 2.0 0.6666 0.6666 0.66666 0.0 0.0")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="ELECTRONIC_TEMPERATURE", &
                          variants=s2a("ELEC_TEMP", "TELEC"), &
                          description="Electronic temperature in the case of Fermi-Dirac smearing", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=cp_unit_to_cp2k(value=300.0_dp, unit_str="K"), &
                          unit_str="K", &
                          usage="ELECTRONIC_TEMPERATURE [K] 300")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="EPS_FERMI_DIRAC", &
                          description="Accuracy checks on occupation numbers use this as a tolerance", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=1.0E-10_dp, &
                          usage="EPS_FERMI_DIRAC 1.0E-6")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="WINDOW_SIZE", &
                          description="Size of the energy window centred at the Fermi level", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.0_dp, &
                          unit_str="au_e", &
                          usage="WINDOW_SIZE [eV] 0.3")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="FIXED_MAGNETIC_MOMENT", &
                          description="Imposed difference between the numbers of electrons of spin up "// &
                          "and spin down: m = n(up) - n(down). A negative value (default) allows "// &
                          "for a change of the magnetic moment. -1 specifically keeps an integer "// &
                          "number of spin up and spin down electrons.", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=-100.0_dp, &
                          usage="FIXED_MAGNETIC_MOMENT 1.5")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_smear_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_rtp_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_rtp_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, print_section

      NULLIFY (keyword)
      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",6713)
      CALL section_create(section, "REAL_TIME_PROPAGATION", &
                          description="Parameters needed to set up the real time propagation"// &
                          " for the electron dynamics", &
                          n_keywords=4, n_subsections=4, repeats=.FALSE., &
                          citations=(/Kunert2003/))

      CALL keyword_create(keyword, name="MAX_ITER", &
                          description="Maximal number of iterations for the self consistent propagator loop.", &
                          usage="MAX_ITER 10", &
                          default_i_val=10)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EPS_ITER", &
                          description="Convergence criterium for the self consistent propagator loop.", &
                          usage="EPS_ITER 1.0E-5", &
                          default_r_val=1.0E-7_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ASPC_ORDER", &
                          description="Speciefies how many steps will be used for extrapolation. "// &
                          "One will be always used which is means X(t+dt)=X(t)", &
                          usage="ASPC_ORDER 3", &
                          default_i_val=3)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MAT_EXP", &
                          description="Which method should be used to calculate the exponential"// &
                          " in the propagator. For Ehrenfest MD only the Taylor method works, "// &
                          "for real time propagation diagonalization works as well.", &
                          usage="MAT_EXP TAYLOR", default_i_val=do_arnoldi, &
                          enum_c_vals=s2a("TAYLOR", "PADE", "ARNOLDI", "BCH"), &
                          enum_i_vals=(/do_taylor, do_pade, do_arnoldi, do_bch/), &
                          enum_desc=s2a("exponential is evaluated using scaling and squaring in combination"// &
                                        " with a taylor expansion of the exponential.", &
                                        "uses scaling and squaring together with the pade approximation", &
                                        "uses arnoldi subspace algorithm to compute exp(H)*MO directly, can't be used in "// &
                                        "combination with Crank Nicholson or density propagation", &
                                        "Uses a Baker-Campbell-Hausdorff expansion to propagate the density matrix,"// &
                                        " only works for density propagation"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DENSITY_PROPAGATION", &
                          description="The density matrix is propagated instead of the molecular orbitals. "// &
                          "This allows for a linear scaling simulation", &
                          usage="DENSITY_PROPAGATION .TRUE.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SC_CHECK_START", &
                          description="Speciefies how many iteration steps will be done without "// &
                          "a check for self consistency. Can save some time in big calculations.", &
                          usage="SC_CHECK_START 3", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="EXP_ACCURACY", &
                          description="Accuracy for the taylor and pade approximation. "// &
                          "This is only an upper bound bound since the norm used for the guess "// &
                          "is an upper bound for the needed one.", &
                          usage="EXP_ACCURACY 1.0E-6", &
                          default_r_val=1.0E-9_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PROPAGATOR", &
                          description="Which propagator should be used for the orbitals", &
                          usage="PROPAGATOR ETRS", default_i_val=do_etrs, &
                          enum_c_vals=s2a("ETRS", "CN", "EM"), &
                          enum_i_vals=(/do_etrs, do_cn, do_em/), &
                          enum_desc=s2a("enforced time reversible symmetry", &
                                        "Crank Nicholson propagator", &
                                        "Exponential midpoint propagator"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INITIAL_WFN", &
                          description="Controls the initial WFN used for propagation.", &
                          usage="INITIAL_WFN SCF_WFN", default_i_val=use_scf_wfn, &
                          enum_c_vals=s2a("SCF_WFN", "RESTART_WFN", "RT_RESTART"), &
                          enum_i_vals=(/use_scf_wfn, use_restart_wfn, use_rt_restart/), &
                          enum_desc=s2a("An SCF run is performed to get the initial state.", &
                                        "A wavefunction from a previous SCF is propagated. Especially useful,"// &
                                        " if electronic constraints or restraints are used in the previous calculation, "// &
                                        "since these do not work in the rtp scheme.", &
                                        "use the wavefunction of a real time propagation/ehrenfest run"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="APPLY_DELTA_PULSE", &
                          description="Applies a delta kick to the initial wfn (only RTP for now - the EMD "// &
                          " case is not yet implemented).", &
                          usage="APPLY_DELTA_PULSE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PERIODIC", &
                          description="Apply a delta-kick that is compatible with periodic boundary conditions"// &
                          " for any value of DELTA_PULSE_SCALE. Uses perturbation theory for the preparation of"// &
                          " the initial wfn. Note that the pulse is only applied when INITIAL_WFN is set to SCF_WFN,"// &
                          " and not for restarts (RT_RESTART).", &
                          usage="PERIODIC", &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DELTA_PULSE_DIRECTION", &
                          description="Direction of the applied electric field. The k vector is given as"// &
                          " 2*Pi*[i,j,k]*inv(h_mat), which for PERIODIC .FALSE. yields exp(ikr) periodic with"// &
                          " the unit cell, only if DELTA_PULSE_SCALE is set to unity. For an orthorhombic cell"// &
                          " [1,0,0] yields [2*Pi/L_x,0,0]. For small cells, this results in a very large kick.", &
                          usage="DELTA_PULSE_DIRECTION 1 1 1", n_var=3, default_i_vals=(/1, 0, 0/), &
                          type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="DELTA_PULSE_SCALE", &
                          description="Scale the k vector, which for PERIODIC .FALSE. results in exp(ikr) no"// &
                          " longer being periodic with the unit cell. The norm of k is the strength of the"// &
                          " applied electric field in atomic units.", &
                          usage="DELTA_PULSE_SCALE 0.01 ", n_var=1, default_r_val=0.001_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="HFX_BALANCE_IN_CORE", &
                          description="If HFX is used, this keyword forces a redistribution/recalculation"// &
                          " of the integrals, balanced with respect to the in core steps.", &
                          usage="HFX_BALANCE_IN_CORE", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="ORTHONORMAL", &
         description="Performs rtp in the orthonormal basis. This keyword is only intended for development purposes."// &
         "It should not be used outside code development", &
         usage="ORTHONORMAL", &
         default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MCWEENY_MAX_ITER", &
                          description="Determines the maximum amount of McWeeny steps used after each converged"// &
                          " step in density propagation", &
                          usage="MCWEENY_MAX_ITER 2", default_i_val=1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="ACCURACY_REFINEMENT", &
         description="If using density propagation some parts should be calculated with a higher accuracy than the rest"// &
         " to reduce numerical noise. This factor determines by how much the filtering threshold is"// &
         " reduced for these calculations.", &
         usage="ACCURACY_REFINEMENT", default_i_val=100)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="MCWEENY_EPS", &
                          description="Threshold after which McWeeny is terminated", &
                          usage="MCWEENY_EPS 0.00001", &
                          default_r_val=0.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (print_section)
      CALL section_create(print_section, name="PRINT", &
                          description="Section of possible print options for an RTP runs", &
                          repeats=.FALSE.)

      NULLIFY (print_key)
      CALL cp_print_key_section_create(print_key, "PROGRAM_RUN_INFO", &
                                       description="Controls the printing within real time propagation and Eherenfest dynamics", &
                                       print_level=low_print_level, filename="__STD_OUT__")
      CALL section_add_subsection(print_section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RESTART", &
                                       description="Controls the dumping of the MO restart file during rtp. "// &
                                       "By default keeps a short history of three restarts."// &
                                       "See also RESTART_HISTORY. In density propagation this controls the printing of P.", &
                                       print_level=low_print_level, common_iter_levels=3, &
                                       each_iter_names=s2a("MD"), each_iter_values=(/20/), &
                                       add_last=add_last_numeric, filename="RESTART")
      CALL keyword_create(keyword, name="BACKUP_COPIES", &
                          description="Specifies the maximum index of backup copies.", &
                          usage="BACKUP_COPIES {int}", &
                          default_i_val=3)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(print_section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "RESTART_HISTORY", &
                                       description="Dumps unique MO restart files during the run keeping all of them. "// &
                                       "In density propagation it dumps the density matrix instead", &
                                       print_level=low_print_level, common_iter_levels=0, &
                                       each_iter_names=s2a("MD"), &
                                       each_iter_values=(/500/), &
                                       filename="RESTART")
      CALL keyword_create(keyword, name="BACKUP_COPIES", &
                          description="Specifies the maximum index of backup copies.", &
                          usage="BACKUP_COPIES {int}", &
                          default_i_val=3)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL section_add_subsection(print_section, print_key)
      CALL section_release(print_key)

      CALL cp_print_key_section_create(print_key, "CURRENT", &
                                       description="Print the current during an EMD simulation to cube files.", &
                                       print_level=high_print_level, common_iter_levels=0, &
                                       each_iter_names=s2a("MD"), &
                                       each_iter_values=(/20/), &
                                       filename="current")
      CALL keyword_create(keyword, name="BACKUP_COPIES", &
                          description="Specifies the maximum index of backup copies.", &
                          usage="BACKUP_COPIES {int}", &
                          default_i_val=1)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="STRIDE", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(print_section, print_key)
      CALL section_release(print_key)

      CALL section_add_subsection(section, print_section)
      CALL section_release(print_section)

   END SUBROUTINE create_rtp_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_constant_env_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_constant_env_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",6968)
      CALL section_create(section, "CONSTANT_ENV", &
                          description="parameters for a constant envelop", &
                          n_keywords=6, n_subsections=1, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="START_STEP", &
                          description="First step the field is applied ", &
                          usage="START_STEP 0", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="END_STEP", &
                          description="Last step the field is applied", &
                          usage="END_STEP 2", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_constant_env_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_gaussian_env_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_gaussian_env_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",7003)
      CALL section_create(section, "GAUSSIAN_ENV", &
                          description="parameters for a gaussian envelop", &
                          n_keywords=6, n_subsections=1, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="T0", &
                          description="Center of the gaussian envelop (maximum of the gaussian)", &
                          usage="T0 2.0E0", &
                          default_r_val=0.0E0_dp, &
                          unit_str="fs")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SIGMA", &
                          description="Width of the gaussian ", &
                          usage="SIGMA 2.0E0", &
                          default_r_val=-1.0E0_dp, &
                          unit_str="fs")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_gaussian_env_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_ramp_env_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_ramp_env_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",7040)
      CALL section_create(section, "RAMP_ENV", &
                          description="Parameters for an trapeziodal envelop ", &
                          n_keywords=6, n_subsections=1, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="START_STEP_IN", &
                          description="Step when the electric field starts to be applied ", &
                          usage="START_STEP_IN 0", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="END_STEP_IN", &
                          description="Step when the field reaches the full strength", &
                          usage="END_STEP_IN 2", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="START_STEP_OUT", &
                          description="Step when the field starts to vanish ", &
                          usage="START_STEP 0", &
                          default_i_val=0)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="END_STEP_OUT", &
                          description="Step when the field disappears", &
                          usage="END_TIME 2", &
                          default_i_val=-1)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_ramp_env_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_custom_env_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_custom_env_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",7089)
      CALL section_create(section, "CUSTOM_ENV", &
                          description="Parameters for a custom efield", &
                          n_keywords=2, n_subsections=1, repeats=.TRUE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, name="EFIELD_FILE_NAME", &
                          description="Specify file that contains the electric field [V/m].", &
                          usage="EFIELD_FILE_NAME filename", &
                          n_var=1, type_of_var=char_t, default_c_val="")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="TIMESTEP", &
                          description="The time step between the entries in the list with the electric field.", &
                          usage="TIME_STEP 1", &
                          unit_str="fs", &
                          default_r_val=1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_custom_env_section

! **************************************************************************************************
!> \brief      Create CP2K input section for the SCCS model
!> \param section ...
!> \par History:
!>      - Creation (10.10.2013,MK)
!> \author     Matthias Krack (MK)
!> \version    1.0
! **************************************************************************************************
   SUBROUTINE create_sccs_section(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_sccs_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",7131)

      CALL section_create(section, &
                          name="SCCS", &
                          description="Define the parameters for self-consistent continuum solvation (SCCS) model", &
                          citations=(/Fattebert2002, Andreussi2012/), &
                          n_keywords=8, &
                          n_subsections=2, &
                          repeats=.FALSE.)

      NULLIFY (keyword)

      CALL keyword_create(keyword, &
                          name="_SECTION_PARAMETERS_", &
                          description="Controls the activation of the SCCS section", &
                          usage="&SCCS ON", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="ALPHA", &
                          description="Solvent specific tunable parameter for the calculation of "// &
                          "the repulsion term <i>G<sup>rep</sup> = &alpha; S</i> "// &
                          "where <i>S</i> is the (quantum) surface of the cavity", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.0_dp, &
                          unit_str="mN/m")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="BETA", &
                          description="Solvent specific tunable parameter for the calculation of "// &
                          "the dispersion term <i>G<sup>dis</sup> = &beta; V</i> "// &
                          "where <i>V</i> is the (quantum) volume of the cavity", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.0_dp, &
                          unit_str="GPa")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="DELTA_RHO", &
                          description="Numerical increment for the calculation of the (quantum) "// &
                          "surface of the solute cavity", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=2.0E-5_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="DERIVATIVE_METHOD", &
                          description="Method for the calculation of the numerical derivatives on the real-space grids", &
                          usage="DERIVATIVE_METHOD cd5", &
                          repeats=.FALSE., &
                          n_var=1, &
                          default_i_val=sccs_derivative_fft, &
                          enum_c_vals=s2a("FFT", "CD3", "CD5", "CD7"), &
                          enum_i_vals=(/sccs_derivative_fft, &
                                        sccs_derivative_cd3, &
                                        sccs_derivative_cd5, &
                                        sccs_derivative_cd7/), &
                          enum_desc=s2a("Fast Fourier transformation", &
                                        "3-point stencil central differences", &
                                        "5-point stencil central differences", &
                                        "7-point stencil central differences"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="DIELECTRIC_CONSTANT", &
                          variants=(/"EPSILON_SOLVENT"/), &
                          description="Dielectric constant of the solvent", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=80.0_dp, &
                          usage="DIELECTRIC_CONSTANT 78.36")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="EPS_SCCS", &
                          variants=s2a("EPS_ITER", "TAU_POL"), &
                          description="Tolerance for the convergence of the polarisation density, "// &
                          "i.e. requested accuracy for the SCCS iteration cycle", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=1.0E-6_dp, &
                          usage="EPS_ITER 1.0E-7")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="EPS_SCF", &
                          description="The SCCS iteration cycle is activated only if the SCF iteration cycle "// &
                          "is converged to this threshold value", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.5_dp, &
                          usage="EPS_SCF 1.0E-2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="GAMMA", &
                          variants=s2a("SURFACE_TENSION"), &
                          description="Surface tension of the solvent used for the calculation of "// &
                          "the cavitation term <i>G<sup>cav</sup> = &gamma; S</i> "// &
                          "where <i>S</i> is the (quantum) surface of the cavity", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.0_dp, &
                          unit_str="mN/m")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="MAX_ITER", &
                          description="Maximum number of SCCS iteration steps performed to converge "// &
                          "within the given tolerance", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=integer_t, &
                          default_i_val=100, &
                          usage="MAX_ITER 50")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="METHOD", &
                          description="Method used for the smoothing of the dielectric function", &
                          usage="METHOD Fattebert-Gygi", &
                          default_i_val=sccs_andreussi, &
                          enum_c_vals=s2a("ANDREUSSI", "FATTEBERT-GYGI"), &
                          enum_i_vals=(/sccs_andreussi, sccs_fattebert_gygi/), &
                          enum_desc=s2a("Smoothing function proposed by Andreussi et al.", &
                                        "Smoothing function proposed by Fattebert and Gygi"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="MIXING", &
                          variants=(/"ETA"/), &
                          description="Mixing parameter (Hartree damping) employed during the iteration procedure", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.6_dp, &
                          usage="MIXING 0.2")
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      NULLIFY (subsection)

      CALL section_create(subsection, &
                          name="ANDREUSSI", &
                          description="Define the parameters of the dielectric smoothing function proposed by "// &
                          "Andreussi et al.", &
                          citations=(/Andreussi2012/), &
                          n_keywords=2, &
                          n_subsections=0, &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, &
                          name="RHO_MAX", &
                          description="Maximum density value used for the smoothing of the dielectric function", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.0035_dp, &
                          usage="RHO_MAX 0.01")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="RHO_MIN", &
                          description="Minimum density value used for the smoothing of the dielectric function", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.0001_dp, &
                          usage="RHO_MIN 0.0003")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL section_create(subsection, &
                          name="FATTEBERT-GYGI", &
                          description="Define the parameters of the dielectric smoothing function proposed by "// &
                          "Fattebert and Gygi", &
                          citations=(/Fattebert2002/), &
                          n_keywords=2, &
                          n_subsections=0, &
                          repeats=.FALSE.)

      CALL keyword_create(keyword, &
                          name="BETA", &
                          description="Parameter &beta; changes the width of the interface solute-solvent", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=1.7_dp, &
                          usage="BETA 1.3")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, &
                          name="RHO_ZERO", &
                          variants=(/"RHO0"/), &
                          description="Parameter &rho;<sub>0</sub> defines the critical density in the middle "// &
                          "of the interface solute-solvent", &
                          repeats=.FALSE., &
                          n_var=1, &
                          type_of_var=real_t, &
                          default_r_val=0.0006_dp, &
                          usage="RHO_ZERO 0.0004")
      CALL section_add_keyword(subsection, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_sccs_section

! **************************************************************************************************
!> \brief      Create CP2K input section for the calculation of an active space Hamiltonian
!> \param section ...
!> \par History:
!>      - Creation 06.04.2016
!> \author     JHU
! **************************************************************************************************
   SUBROUTINE create_active_space_section(section)

      TYPE(section_type), POINTER                        :: section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'create_active_space_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key, subsection

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",7386)

      CALL section_create(section, name="ACTIVE_SPACE", &
                          description="Define parameters and method to calculate an electronic active space", &
                          n_keywords=1, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword, subsection, print_key)

      CALL keyword_create(keyword, &
                          name="_SECTION_PARAMETERS_", &
                          description="Controls the activation of the ACTIVE_SPACE section", &
                          usage="&ACTIVE_SPACE ON", &
                          default_l_val=.FALSE., &
                          lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ACTIVE_ELECTRONS", &
                          description="The number of active electrons in the CAS space", &
                          usage="ACTIVE_ELECTRONS 4", n_var=1, default_i_val=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INACTIVE_ELECTRONS", &
                          description="The number of inactive electrons in the CAS space", &
                          usage="INACTIVE_ELECTRONS 2 {2}", n_var=-1, default_i_vals=(/-1, -1/), &
                          type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ACTIVE_ORBITALS", &
                          description="The number of active orbitals defining the CAS space.", &
                          usage="ACTIVE_ORBITALS 2 {2}", n_var=-1, default_i_vals=(/-1, -1/), &
                          type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="INACTIVE_ORBITALS", &
                          description="The number of nactive orbitals.", &
                          usage="INACTIVE_ORBITALS 2 {2}", n_var=-1, default_i_vals=(/-1, -1/), &
                          type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ISOLATED_SYSTEM", &
                          description="System is treated without any periodic boundary conditions.", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL cp_print_key_section_create(print_key, "FCIDUMP", &
                                       description="Controls the writing of a file in FCIDUMP format.", &
                                       print_level=high_print_level, filename="")
      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      CALL keyword_create(keyword, name="ORBITAL_SELECTION", &
                          description="Method used to select active space orbitals.", &
                          usage="ORBITAL_SELECTION CANONICAL", &
                          default_i_val=casci_canonical, &
                          enum_c_vals=s2a("CANONICAL", "WANNIER_PROJECTION", "MAO"), &
                          enum_i_vals=(/casci_canonical, wannier_projection, mao_projection/), &
                          enum_desc=s2a("Select orbitals using energy ordering of canoncial orbitals", &
                                        "Select orbitals from projected Wannier functions", &
                                        "Select orbitals from modified atomic orbitals"))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SUBSPACE_ATOM", &
                          description="Number of atom that defines the subspace to be projected on.", &
                          usage="SUBSPACE_ATOM x", default_i_val=-1, &
                          type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="SUBSPACE_SHELL", &
                          description="Shell definition for subsapce.", &
                          usage="SUBSPACE_SHELL 3d4s", default_c_val="X", &
                          type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL create_print_orb_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_eri_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_eri_gpw(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

      CALL create_localize_section(subsection)
      CALL section_add_subsection(section, subsection)
      CALL section_release(subsection)

   END SUBROUTINE create_active_space_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_print_orb_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_print_orb_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",7498)
      CALL section_create(section, "PRINT_ORBITAL_CUBES", &
                          description="Controls printing of active orbital cube files.", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="FILENAME", &
                          description="Body of Filename for the cube files.", &
                          usage="FILENAME {name}", default_c_val="ActiveOrbital", &
                          type_of_var=char_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="ALIST", &
                          description="List of alpha orbitals to be printed. -1 defaults to all values", &
                          usage="ALIST {1 2 3 ...}", n_var=-1, default_i_vals=(/-1/), &
                          lone_keyword_i_val=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="BLIST", &
                          description="List of beta orbitals to be printed. -1 defaults to all values", &
                          usage="ALIST {1 2 3 ...}", n_var=-1, default_i_vals=(/-1/), &
                          lone_keyword_i_val=-1, type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STRIDE", &
                          description="The stride (X,Y,Z) used to write the cube file"// &
                          "(larger values result in smaller cube files)."// &
                          " You can provide 3 numbers (for X,Y,Z) or 1 number valid for all components", &
                          usage="STRIDE {2 2 2}", n_var=-1, default_i_vals=(/2, 2, 2/), &
                          type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_print_orb_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_eri_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_eri_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",7548)
      CALL section_create(section, "ERI", &
                          description="Parameters for the electron repulsion integrals.", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="METHOD", &
                          description="Method used in ERI calculation.", &
                          usage="METHOD FULL_GPW", &
                          enum_c_vals=s2a("FULL_GPW", "GPW_HALF_TRANSFORM"), &
                          enum_i_vals=(/eri_method_full_gpw, eri_method_gpw_ht/), &
                          enum_desc=s2a("Use the GPW approach with MOs", &
                                        "Use the GPW approach for half-transformed MO ERIs"), &
                          default_i_val=eri_method_gpw_ht)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OPERATOR", &
                          description="Operator used in ERI calculation.", &
                          usage="OPERATOR <1/R>", &
                          enum_c_vals=s2a("<1/R>", "<EXP(-A*R)/R>", "<ERF(A*R)/R>", &
                                          "<ERFC(A*R)/R>", "<EXP(-A*R2)/R>"), &
                          enum_i_vals=(/eri_operator_coulomb, eri_operator_yukawa, &
                                        eri_operator_erf, eri_operator_erfc, eri_operator_gaussian/), &
                          enum_desc=s2a("Coulomb operator", &
                                        "Yukawa potential operator", &
                                        "Error function potential operator", &
                                        "Complementary error function potential operator", &
                                        "Gaussian potential operator"), &
                          default_i_val=eri_operator_coulomb)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="OPERATOR_PARAMETER", &
                          description="Range parameter for ERI operator.", &
                          usage="OPERATOR_PARAMETER 4.0", type_of_var=real_t, &
                          default_r_val=8.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PERIODICITY", &
                          description="Periodicity used for operators in ERI calclulation.", &
                          usage="PERIODICITY {1 1 1}", n_var=-1, default_i_vals=(/1, 1, 1/), &
                          type_of_var=integer_t)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUTOFF_RADIUS", &
                          description="Radius of operator interactions in non-periodic cases.", &
                          usage="CUTOFF_RADIUS 20", type_of_var=real_t, &
                          default_r_val=-1.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create( &
         keyword, name="EPS_INTEGRAL", &
         description="Accuracy of ERIs that will be stored.", &
         usage="EPS_FILTER 1.0E-10 ", type_of_var=real_t, &
         default_r_val=1.0E-12_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_eri_section

! **************************************************************************************************
!> \brief ...
!> \param section ...
! **************************************************************************************************
   SUBROUTINE create_eri_gpw(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_eri_gpw', routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",7623)
      CALL section_create(section, "ERI_GPW", &
                          description="Parameters for the GPW approach to electron repulsion integrals.", &
                          n_keywords=5, n_subsections=0, repeats=.FALSE.)

      NULLIFY (keyword)
      CALL keyword_create(keyword, name="EPS_GRID", &
                          description="Determines a threshold for the GPW based integration", &
                          usage="EPS_GRID 1.0E-9 ", type_of_var=real_t, &
                          default_r_val=1.0E-8_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="CUTOFF", &
                          description="The cutoff of the finest grid level in the GPW integration.", &
                          usage="CUTOFF 300", type_of_var=real_t, &
                          default_r_val=300.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="REL_CUTOFF", &
                          variants=(/"RELATIVE_CUTOFF"/), &
                          description="Determines the grid at which a Gaussian is mapped.", &
                          usage="REL_CUTOFF 50", type_of_var=real_t, &
                          default_r_val=50.0_dp)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="STORE_WFN", &
                          variants=(/"STORE_WAVEFUNCTION"/), &
                          description="Strore wavefunction in real space representation for integration.", &
                          usage="STORE_WFN  T", type_of_var=logical_t, &
                          default_l_val=.TRUE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

      CALL keyword_create(keyword, name="PRINT_LEVEL", &
                          variants=(/"IOLEVEL"/), &
                          description="How much output is written by the individual groups.", &
                          usage="PRINT_LEVEL HIGH", &
                          default_i_val=silent_print_level, enum_c_vals= &
                          s2a("SILENT", "LOW", "MEDIUM", "HIGH", "DEBUG"), &
                          enum_desc=s2a("Almost no output", &
                                        "Little output", "Quite some output", "Lots of output", &
                                        "Everything is written out, useful for debugging purposes only"), &
                          enum_i_vals=(/silent_print_level, low_print_level, medium_print_level, &
                                        high_print_level, debug_print_level/))
      CALL section_add_keyword(section, keyword)
      CALL keyword_release(keyword)

   END SUBROUTINE create_eri_gpw

! **************************************************************************************************
!> \brief Creates the section for cube files related to the implicit Poisson solver.
!> \param section the section to be created
!> \par History
!>      03.2016 refactored from create_print_dft_section [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE create_implicit_psolver_section(section)
      TYPE(section_type), POINTER                        :: section

      CHARACTER(len=*), PARAMETER :: routineN = 'create_implicit_psolver_section', &
         routineP = moduleN//':'//routineN

      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: print_key

      IF(.NOT.(.NOT. ASSOCIATED(section)))CALL cp__a("input_cp2k_dft.F",7691)
      CALL section_create(section, name="IMPLICIT_PSOLVER", &
                          description="Controls printing of cube files for data from the implicit "// &
                          "(generalized) Poisson solver.", &
                          citations=(/BaniHashemian2016/), &
                          n_keywords=0, n_subsections=3, repeats=.FALSE.)

      NULLIFY (keyword, print_key)

      ! dielectric constant function
      CALL cp_print_key_section_create(print_key, "DIELECTRIC_CUBE", &
                                       description="Controls the printing of a cube file with dielectric constant from "// &
                                       "the implicit (generalized) Poisson solver.", &
                                       print_level=high_print_level, filename="")

      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      ! dirichlet type constraints
      CALL cp_print_key_section_create( &
         print_key, "DIRICHLET_BC_CUBE", &
         description="Controls the printing of cube files with unit step functions (constraints) "// &
         "representing Dirichlet-type (boundary) regions defined in the implicit (generalized) Poisson "// &
         "solver section. The regions remain unchanged throughout the calculations. If the Dirichlet "// &
         " regions are relatively large and/or the number of partitions is quite high, in order to save memory,"// &
         " generate the cube files in early steps and perform the rest of the calculations with this keyword"// &
         " switched off.", &
         print_level=high_print_level, filename="")

      CALL keyword_create(keyword, name="TILE_CUBES", &
                          description="Print tiles that tessellate the Dirichlet regions into cube files. If TRUE, "// &
                          "generates cube files as many as the total number of tiles.", &
                          usage="TILE_CUBES <logical>", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

      ! charge introduced by Lagrange multipliers
      CALL cp_print_key_section_create(print_key, "DIRICHLET_CSTR_CHARGE_CUBE", &
                                       description="Controls the printing of cube files with penalty charges induced to "// &
                                       "Dirichlet regions by Lagrange multipliers (implicit Poisson solver).", &
                                       print_level=high_print_level, filename="")

      CALL keyword_create(keyword, name="stride", &
                          description="The stride (X,Y,Z) used to write the cube file "// &
                          "(larger values result in smaller cube files). You can provide 3 numbers (for X,Y,Z) or"// &
                          " 1 number valid for all components.", &
                          usage="STRIDE 2 2 2", n_var=-1, default_i_vals=(/2, 2, 2/), type_of_var=integer_t)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)
      CALL keyword_create(keyword, name="APPEND", &
                          description="append the cube files when they already exist", &
                          default_l_val=.FALSE., lone_keyword_l_val=.TRUE.)
      CALL section_add_keyword(print_key, keyword)
      CALL keyword_release(keyword)

      CALL section_add_subsection(section, print_key)
      CALL section_release(print_key)

   END SUBROUTINE create_implicit_psolver_section

END MODULE input_cp2k_dft
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief parameters that control an scf iteration
!> \note
!>       not in cp_control_types, to separate operator related parameters from
!>       method related parameters (as suggested by Matthias)
!> \par History
!>      09.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE scf_control_types

   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE input_constants,                 ONLY: &
        atomic_guess, core_guess, diag_ot, direct_p_mix, general_roks, high_spin_roks, &
        ot_algo_taylor_or_diag, outer_scf_becke_constraint, outer_scf_ddapc_constraint, &
        outer_scf_none, outer_scf_optimizer_bisect, outer_scf_optimizer_diis, &
        outer_scf_optimizer_none, outer_scf_optimizer_sd, outer_scf_s2_constraint, &
        smear_energy_window, smear_fermi_dirac, smear_list
   USE input_cp2k_dft,                  ONLY: create_scf_section
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: section_get_keyword,&
                                              section_release,&
                                              section_type,&
                                              section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE qs_ot_types,                     ONLY: ot_readwrite_input,&
                                              qs_ot_settings_init,&
                                              qs_ot_settings_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'scf_control_types'
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.

   INTEGER, SAVE :: last_scf_c_id_nr = 0

   ! Public data types

   PUBLIC :: scf_control_type, &
             smear_type

   ! Public subroutines

   PUBLIC :: scf_c_create, &
             scf_c_read_parameters, &
             scf_c_release, &
             scf_c_retain, &
             scf_c_write_parameters

! **************************************************************************************************
!> \brief contains the parameters needed by a scf run
!> \param density_guess how to choose the initial density
!>        (CORE,RANDOM,RESTART,ATOMIC,FROZEN)
!> \param eps_eigval wanted error on the eigenvalues
!> \param eps_scf whanted error on the whole scf
!> \param level_shift amount of level shift
!> \param p_mix how to mix the new and old densities in non diss iterations
!> \param eps_lumos error on the lumos calculated at the end of the scf
!> \param max_iter_lumus maxumum number of iterations used to calculate
!>        the lumos at the end of the scf
!> \param max_scf max scf iterations
!> \param id_nr unique number to identify an scf control
!> \param ref_count reference count (see cp2k/doc/ReferenceCounting.html)
!> \param added_mos additional number of MOs that might be used in the SCF
!> \par History
!>      09.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************

   TYPE outer_scf_control_type
      LOGICAL       :: have_scf
      INTEGER       :: max_scf
      REAL(KIND=dp) :: eps_scf, step_size
      INTEGER       :: TYPE
      INTEGER       :: optimizer
      INTEGER       :: diis_buffer_length
      INTEGER       :: extrapolation_order
      INTEGER       :: bisect_trust_count
   END TYPE outer_scf_control_type

   TYPE smear_type
      LOGICAL       :: do_smear
      LOGICAL       :: common_mu
      INTEGER       :: method
      REAL(KIND=dp) :: electronic_temperature, &
                       fixed_mag_mom, &
                       eps_fermi_dirac, &
                       window_size
      REAL(KIND=dp), DIMENSION(:), POINTER :: list
   END TYPE smear_type

   TYPE diagonalization_type
      INTEGER       :: method
      REAL(KIND=dp) :: eps_jacobi
      REAL(KIND=dp) :: jacobi_threshold
      INTEGER       :: max_iter, nkrylov, nblock_krylov
      ! Maximum Overlap Method
      LOGICAL       :: mom, mom_didguess
      INTEGER       :: mom_proj_formula
      ! indices of de-occupied and newly occupied alpha / beta molecular orbitals
      INTEGER, DIMENSION(:), POINTER :: mom_deoccA, mom_deoccB, mom_occA, mom_occB
      ! determines on SCF which iteration MOM will be switched on;
      ! since MOs from the previous iteration should be available, it might be at least
      !  1 when wave-function has been read from restart file, or
      !  2 when the atomic guess method has been used
      INTEGER       :: mom_start
      REAL(KIND=dp) :: eps_iter
      REAL(KIND=dp) :: eps_adapt
      TYPE(qs_ot_settings_type) :: ot_settings
   END TYPE diagonalization_type

   TYPE scf_control_type
      TYPE(outer_scf_control_type)          :: outer_scf
      TYPE(smear_type), POINTER             :: smear
      TYPE(diagonalization_type)            :: diagonalization
      INTEGER                               :: density_guess, mixing_method
      REAL(KIND=dp)                         :: eps_eigval, eps_scf, eps_scf_hist, &
                                               level_shift, &
                                               eps_lumos, eps_diis
      INTEGER                               :: max_iter_lumos, max_diis, nmixing
      INTEGER                               :: max_scf, max_scf_hist, &
                                               maxl, nkind
      LOGICAL                               :: do_diag_sub, &
                                               use_cholesky, use_ot, use_diag, do_outer_scf_reortho
      INTEGER                               :: ref_count, id_nr
      INTEGER, DIMENSION(2)                 :: added_mos
      INTEGER                               :: roks_scheme
      REAL(KIND=dp)                         :: roks_f
      REAL(KIND=dp), DIMENSION(0:2, 0:2, 1:2) :: roks_parameter
   END TYPE scf_control_type

CONTAINS

! **************************************************************************************************
!> \brief allocates and initializes an scf control object with the default values
!> \param scf_control the object to initialize
!> \par History
!>      09.2002 created [fawzi]
!>      - Default ROKS parameters added (05.04.06,MK)
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE scf_c_create(scf_control)

      TYPE(scf_control_type), POINTER                    :: scf_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scf_c_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      ALLOCATE (scf_control)

      ! Load the default values

      scf_control%density_guess = atomic_guess
      scf_control%eps_eigval = 1.0E-5_dp
      scf_control%eps_scf = 1.0E-5_dp
      scf_control%eps_scf_hist = 0.0_dp
      scf_control%eps_lumos = 1.0E-5_dp
      scf_control%max_iter_lumos = 2999
      scf_control%eps_diis = 0.1_dp
      scf_control%level_shift = 0.0_dp
      scf_control%max_diis = 4
      scf_control%max_scf = 50
      scf_control%nmixing = 2
      scf_control%use_cholesky = .TRUE.
      scf_control%use_diag = .TRUE.
      scf_control%do_diag_sub = .FALSE.
      scf_control%use_ot = .FALSE.
      scf_control%do_outer_scf_reortho = .TRUE.
      scf_control%max_diis = 4
      scf_control%eps_diis = 0.1_dp
      scf_control%ref_count = 1
      scf_control%id_nr = last_scf_c_id_nr
      scf_control%added_mos(:) = 0
      scf_control%max_scf_hist = 0

      !Mixing
      scf_control%mixing_method = direct_p_mix

      ! Diagonalization
      scf_control%diagonalization%method = 0
      scf_control%diagonalization%eps_jacobi = 0.0_dp
      scf_control%diagonalization%jacobi_threshold = 1.0E-7_dp
      scf_control%diagonalization%max_iter = 0
      scf_control%diagonalization%eps_iter = 0.0_dp
      scf_control%diagonalization%eps_adapt = 0.0_dp
      scf_control%diagonalization%nkrylov = 0
      scf_control%diagonalization%nblock_krylov = 0
      CALL qs_ot_settings_init(scf_control%diagonalization%ot_settings)

      scf_control%diagonalization%mom = .FALSE.
      scf_control%diagonalization%mom_didguess = .FALSE.
      scf_control%diagonalization%mom_proj_formula = 0
      NULLIFY (scf_control%diagonalization%mom_deoccA)
      NULLIFY (scf_control%diagonalization%mom_deoccB)
      NULLIFY (scf_control%diagonalization%mom_occA)
      NULLIFY (scf_control%diagonalization%mom_occB)
      scf_control%diagonalization%mom_start = 0

      ! ROKS

      scf_control%roks_scheme = high_spin_roks
      scf_control%roks_f = 0.5_dp

      ! Initialize the diagonal blocks with the default ROKS parameters
      ! 0 = v)irtual, 1 = o)pen shell, 2 = c)losed shell

      scf_control%roks_parameter(0, 0, 1) = 1.5_dp ! avv
      scf_control%roks_parameter(0, 0, 2) = -0.5_dp ! bvv
      scf_control%roks_parameter(1, 1, 1) = 0.5_dp ! aoo
      scf_control%roks_parameter(1, 1, 2) = 0.5_dp ! boo
      scf_control%roks_parameter(2, 2, 1) = -0.5_dp ! acc
      scf_control%roks_parameter(2, 2, 2) = 1.5_dp ! bcc

      ! Initialize off-diagonal blocks (fixed)

      scf_control%roks_parameter(0, 1, 1) = 1.0_dp ! avo
      scf_control%roks_parameter(0, 1, 2) = 0.0_dp ! bvo
      scf_control%roks_parameter(0, 2, 1) = 0.5_dp ! avc
      scf_control%roks_parameter(0, 2, 2) = 0.5_dp ! bvc
      scf_control%roks_parameter(1, 2, 1) = 0.0_dp ! aoc
      scf_control%roks_parameter(1, 2, 2) = 1.0_dp ! boc

      ! Symmetry enforces

      scf_control%roks_parameter(1, 0, 1) = scf_control%roks_parameter(0, 1, 1) ! aov
      scf_control%roks_parameter(1, 0, 2) = scf_control%roks_parameter(0, 1, 2) ! bov
      scf_control%roks_parameter(2, 0, 1) = scf_control%roks_parameter(0, 2, 1) ! acv
      scf_control%roks_parameter(2, 0, 2) = scf_control%roks_parameter(0, 2, 2) ! bcv
      scf_control%roks_parameter(2, 1, 1) = scf_control%roks_parameter(1, 2, 1) ! aco
      scf_control%roks_parameter(2, 1, 2) = scf_control%roks_parameter(1, 2, 2) ! bco

      ! Outer SCF default settings

      scf_control%outer_scf%have_scf = .FALSE.
      scf_control%outer_scf%max_scf = 0
      scf_control%outer_scf%eps_scf = 0.0_dp
      scf_control%outer_scf%step_size = 0.0_dp
      scf_control%outer_scf%type = -1
      scf_control%outer_scf%optimizer = -1
      scf_control%outer_scf%diis_buffer_length = -1

      ! Smearing of the MO occupations

      NULLIFY (scf_control%smear)

      last_scf_c_id_nr = last_scf_c_id_nr+1

      CALL timestop(handle)

   END SUBROUTINE scf_c_create

! **************************************************************************************************
!> \brief retains the given scf_control (see cp2k/doc/ReferenceCounting.html)
!> \param scf_control the object to retain
!> \par History
!>      09.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE scf_c_retain(scf_control)

      TYPE(scf_control_type), POINTER                    :: scf_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scf_c_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(scf_control)))CALL cp__a("scf_control_types.F",286)

      IF(.NOT.(scf_control%ref_count > 0))CALL cp__a("scf_control_types.F",288)
      scf_control%ref_count = scf_control%ref_count+1

   END SUBROUTINE scf_c_retain

! **************************************************************************************************
!> \brief releases the given scf_control (see cp2k/doc/ReferenceCounting.html)
!> \param scf_control the object to free
!> \par History
!>      09.2002 created [fawzi]
!> \author Fawzi Mohamed
!> \note
!>      at the moment does nothing
! **************************************************************************************************
   SUBROUTINE scf_c_release(scf_control)

      TYPE(scf_control_type), POINTER                    :: scf_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scf_c_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(scf_control)) THEN
         IF(.NOT.(scf_control%ref_count > 0))CALL cp__a("scf_control_types.F",309)
         scf_control%ref_count = scf_control%ref_count-1
         IF (scf_control%ref_count < 1) THEN
            IF (ASSOCIATED(scf_control%smear%list)) THEN
               DEALLOCATE (scf_control%smear%list)
            ENDIF
            DEALLOCATE (scf_control%smear)

            ! Maximum overlap method orbital indices lists
            ! mom_deoccA, mom_deoccB, mom_occA, mom_occB
            ! points to memory allocated by input file parser,
            ! so they do not have to be deallocated

            DEALLOCATE (scf_control)
         END IF
      END IF

      NULLIFY (scf_control)

   END SUBROUTINE scf_c_release

! **************************************************************************************************
!> \brief reads the parameters of the scf section into the given scf_control
!> \param scf_control the object that wil contain the values read
!> \param inp_section ...
!> \par History
!>      05.2001 created [Matthias]
!>      09.2002 creaded separated scf_control type [fawzi]
!> \author Matthias Krack
! **************************************************************************************************
   SUBROUTINE scf_c_read_parameters(scf_control, inp_section)

      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: inp_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scf_c_read_parameters', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: cholesky_flag, handle, ialgo
      INTEGER, DIMENSION(:), POINTER                     :: added_mos
      LOGICAL                                            :: do_mixing
      REAL(KIND=dp), DIMENSION(:), POINTER               :: roks_parameter
      TYPE(section_vals_type), POINTER                   :: mixing_section, outer_scf_section, &
                                                            scf_section, smear_section

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(scf_control)))CALL cp__a("scf_control_types.F",356)
      IF(.NOT.(scf_control%ref_count > 0))CALL cp__a("scf_control_types.F",357)
      scf_section => section_vals_get_subs_vals(inp_section, "SCF")
      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%_SECTION_PARAMETERS_", &
                                l_val=scf_control%use_diag)
      IF (scf_control%use_diag) THEN
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DIAG_SUB_SCF%_SECTION_PARAMETERS_", &
                                   l_val=scf_control%do_diag_sub)
      END IF
      CALL section_vals_val_get(scf_section, "OT%_SECTION_PARAMETERS_", l_val=scf_control%use_ot)
      IF (scf_control%use_diag .AND. scf_control%use_ot) THEN
         ! don't allow both options to be true
         CALL cp__b("scf_control_types.F",368,"Don't activate OT and Diagonaliztion together")
      ELSEIF (.NOT. (scf_control%use_diag .OR. scf_control%use_ot)) THEN
         ! set default to diagonalization
         scf_control%use_diag = .TRUE.
      END IF
      CALL section_vals_val_get(scf_section, "OT%ALGORITHM", i_val=ialgo)
      scf_control%do_outer_scf_reortho = ialgo .EQ. ot_algo_taylor_or_diag
      CALL section_vals_val_get(scf_section, "SCF_GUESS", i_val=scf_control%density_guess)
      CALL section_vals_val_get(scf_section, "EPS_DIIS", r_val=scf_control%eps_diis)
      CALL section_vals_val_get(scf_section, "eps_eigval", r_val=scf_control%eps_eigval)
      CALL section_vals_val_get(scf_section, "cholesky", i_val=cholesky_flag)
      IF (cholesky_flag > 0) THEN
         scf_control%use_cholesky = .TRUE.
      END IF
      CALL section_vals_val_get(scf_section, "eps_scf", r_val=scf_control%eps_scf)
      CALL section_vals_val_get(scf_section, "level_shift", r_val=scf_control%level_shift)
      CALL section_vals_val_get(scf_section, "max_diis", i_val=scf_control%max_diis)
      CALL section_vals_val_get(scf_section, "max_scf", i_val=scf_control%max_scf)

      ! Diagonaliztion section
      IF (scf_control%use_diag) THEN
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%ALGORITHM", &
                                   i_val=scf_control%diagonalization%method)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%EPS_JACOBI", &
                                   r_val=scf_control%diagonalization%eps_jacobi)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%JACOBI_THRESHOLD", &
                                   r_val=scf_control%diagonalization%jacobi_threshold)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%MAX_ITER", &
                                   i_val=scf_control%diagonalization%max_iter)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%EPS_ITER", &
                                   r_val=scf_control%diagonalization%eps_iter)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%EPS_ADAPT", &
                                   r_val=scf_control%diagonalization%eps_adapt)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%KRYLOV%NKRYLOV", &
                                   i_val=scf_control%diagonalization%nkrylov)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%KRYLOV%NBLOCK", &
                                   i_val=scf_control%diagonalization%nblock_krylov)
         IF (scf_control%diagonalization%method == diag_ot) THEN
            ! read OT section
            CALL ot_diag_read_input(scf_control%diagonalization%ot_settings, scf_section)
         END IF
         ! read maximum overlap method's parameters
         CALL section_vals_val_get(scf_section, "MOM%_SECTION_PARAMETERS_", &
                                   l_val=scf_control%diagonalization%MOM)
         IF (scf_control%diagonalization%mom) THEN
            CALL section_vals_val_get(scf_section, "MOM%START_ITER", &
                                      i_val=scf_control%diagonalization%mom_start)

            CALL section_vals_val_get(scf_section, "MOM%DEOCC_ALPHA", &
                                      i_vals=scf_control%diagonalization%mom_deoccA)

            CALL section_vals_val_get(scf_section, "MOM%DEOCC_BETA", &
                                      i_vals=scf_control%diagonalization%mom_deoccB)

            CALL section_vals_val_get(scf_section, "MOM%OCC_ALPHA", &
                                      i_vals=scf_control%diagonalization%mom_occA)

            CALL section_vals_val_get(scf_section, "MOM%OCC_BETA", &
                                      i_vals=scf_control%diagonalization%mom_occB)

            CALL section_vals_val_get(scf_section, "MOM%PROJ_FORMULA", &
                                      i_val=scf_control%diagonalization%mom_proj_formula)
         END IF
      END IF

      ! Read ROKS parameters
      CALL section_vals_val_get(scf_section, "ROKS_SCHEME", i_val=scf_control%roks_scheme)

      SELECT CASE (scf_control%roks_scheme)
      CASE (general_roks)
         ! Read parameters for the general ROKS scheme
         CALL section_vals_val_get(scf_section, "ROKS_F", r_val=scf_control%roks_f)
      CASE (high_spin_roks)
         ! Read high-spin ROKS parameters for the diagonal block
         ! 0 = v)irtual, 1 = o)pen shell, 2 = c)losed shell
         NULLIFY (roks_parameter)
         CALL section_vals_val_get(scf_section, "ROKS_PARAMETERS", r_vals=roks_parameter)
         IF (ASSOCIATED(roks_parameter)) THEN
            scf_control%roks_parameter(2, 2, 1) = roks_parameter(1) ! acc
            scf_control%roks_parameter(2, 2, 2) = roks_parameter(2) ! bcc
            scf_control%roks_parameter(1, 1, 1) = roks_parameter(3) ! aoo
            scf_control%roks_parameter(1, 1, 2) = roks_parameter(4) ! boo
            scf_control%roks_parameter(0, 0, 1) = roks_parameter(5) ! avv
            scf_control%roks_parameter(0, 0, 2) = roks_parameter(6) ! bvv
         END IF
      END SELECT

      ! should be moved to printkey
      CALL section_vals_val_get(scf_section, "eps_lumo", r_val=scf_control%eps_lumos)
      CALL section_vals_val_get(scf_section, "max_iter_lumo", i_val=scf_control%max_iter_lumos)

      ! Extra MOs, e.g. for smearing
      CALL section_vals_val_get(scf_section, "added_mos", i_vals=added_mos)
      IF(.NOT.(ASSOCIATED(added_mos)))CALL cp__a("scf_control_types.F",461)
      IF (SIZE(added_mos) > 0) THEN
         scf_control%added_mos(1) = added_mos(1)
         IF (SIZE(added_mos) > 1) THEN
            scf_control%added_mos(2) = added_mos(2)
         END IF
      END IF

      CALL section_vals_val_get(scf_section, "max_scf_history", i_val=scf_control%max_scf_hist)
      CALL section_vals_val_get(scf_section, "eps_scf_history", r_val=scf_control%eps_scf_hist)

      IF (scf_control%level_shift /= 0.0_dp) scf_control%use_cholesky = .FALSE.
      IF (scf_control%use_ot .AND. (scf_control%density_guess == core_guess)) THEN
         CALL cp__b("scf_control_types.F",474,"Use GUESS {ATOMIC,RESTART,RANDOM} with OT")
      END IF

      outer_scf_section => section_vals_get_subs_vals(scf_section, "OUTER_SCF")
      CALL section_vals_val_get(outer_scf_section, "_SECTION_PARAMETERS_", &
                                l_val=scf_control%outer_scf%have_scf)
      IF (scf_control%outer_scf%have_scf) THEN
         CALL section_vals_val_get(outer_scf_section, "EPS_SCF", &
                                   r_val=scf_control%outer_scf%eps_scf)
         CALL section_vals_val_get(outer_scf_section, "STEP_SIZE", &
                                   r_val=scf_control%outer_scf%step_size)
         CALL section_vals_val_get(outer_scf_section, "DIIS_BUFFER_LENGTH", &
                                   i_val=scf_control%outer_scf%diis_buffer_length)
         CALL section_vals_val_get(outer_scf_section, "BISECT_TRUST_COUNT", &
                                   i_val=scf_control%outer_scf%bisect_trust_count)
         CALL section_vals_val_get(outer_scf_section, "TYPE", &
                                   i_val=scf_control%outer_scf%type)
         CALL section_vals_val_get(outer_scf_section, "OPTIMIZER", &
                                   i_val=scf_control%outer_scf%optimizer)
         CALL section_vals_val_get(outer_scf_section, "MAX_SCF", &
                                   i_val=scf_control%outer_scf%max_scf)
         CALL section_vals_val_get(outer_scf_section, "EXTRAPOLATION_ORDER", &
                                   i_val=scf_control%outer_scf%extrapolation_order)
      END IF

      smear_section => section_vals_get_subs_vals(scf_section, "SMEAR")
      CALL init_smear(scf_control%smear)
      CALL read_smear_section(scf_control%smear, smear_section)

      do_mixing = .FALSE.
      mixing_section => section_vals_get_subs_vals(scf_section, "MIXING")
      CALL section_vals_val_get(mixing_section, "_SECTION_PARAMETERS_", &
                                l_val=do_mixing)
      IF (do_mixing) THEN
         CALL section_vals_val_get(mixing_section, "METHOD", &
                                   i_val=scf_control%mixing_method)
         CALL section_vals_val_get(mixing_section, "NMIXING", i_val=scf_control%nmixing)
      END IF ! do mixing

      CALL timestop(handle)

   END SUBROUTINE scf_c_read_parameters

! **************************************************************************************************
!> \brief ...
!> \param smear ...
! **************************************************************************************************
   SUBROUTINE init_smear(smear)
      TYPE(smear_type), POINTER                          :: smear

      CHARACTER(len=*), PARAMETER :: routineN = 'init_smear', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(smear)))CALL cp__a("scf_control_types.F",526)
      ALLOCATE (smear)
      smear%do_smear = .FALSE.
      smear%method = smear_energy_window
      smear%electronic_temperature = 0.0_dp
      smear%eps_fermi_dirac = 1.0E-5_dp
      smear%fixed_mag_mom = -100.0_dp
      smear%window_size = 0.0_dp
      NULLIFY (smear%list)
   END SUBROUTINE init_smear

! **************************************************************************************************
!> \brief ...
!> \param smear ...
!> \param smear_section ...
! **************************************************************************************************
   SUBROUTINE read_smear_section(smear, smear_section)
      TYPE(smear_type), POINTER                          :: smear
      TYPE(section_vals_type), POINTER                   :: smear_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_smear_section', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp), DIMENSION(:), POINTER               :: r_vals

      NULLIFY (r_vals)

      CALL section_vals_val_get(smear_section, "_SECTION_PARAMETERS_", &
                                l_val=smear%do_smear)
      IF (smear%do_smear) THEN
         CALL section_vals_val_get(smear_section, "METHOD", &
                                   i_val=smear%method)
         CALL section_vals_val_get(smear_section, "ELECTRONIC_TEMPERATURE", &
                                   r_val=smear%electronic_temperature)
         CALL section_vals_val_get(smear_section, "EPS_FERMI_DIRAC", &
                                   r_val=smear%eps_fermi_dirac)
         CALL section_vals_val_get(smear_section, "WINDOW_SIZE", &
                                   r_val=smear%window_size)
         IF (smear%method == smear_list) THEN
            CALL section_vals_val_get(smear_section, "LIST", &
                                      r_vals=r_vals)
            IF(.NOT.(ASSOCIATED(r_vals)))CALL cp__a("scf_control_types.F",567)
            ALLOCATE (smear%list(SIZE(r_vals)))
            smear%list = r_vals
         END IF
         CALL section_vals_val_get(smear_section, "FIXED_MAGNETIC_MOMENT", &
                                   r_val=smear%fixed_mag_mom)
      END IF ! do smear
   END SUBROUTINE read_smear_section

! **************************************************************************************************
!> \brief writes out the scf parameters
!> \param scf_control the object you want to print
!> \param dft_section ...
!> \par History
!>      05.2001 created [Matthias]
!>      09.2002 created separated scf_control type [fawzi]
!> \author Matthias Krack
! **************************************************************************************************
   SUBROUTINE scf_c_write_parameters(scf_control, dft_section)

      TYPE(scf_control_type), POINTER                    :: scf_control
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'scf_c_write_parameters', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, output_unit, roks_scheme
      LOGICAL                                            :: roks
      REAL(KIND=dp)                                      :: elec_temp
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: scf_section

      CALL timeset(routineN, handle)

      NULLIFY (logger)
      logger => cp_get_default_logger()

      NULLIFY (scf_section)
      NULLIFY (section)

      IF(.NOT.(ASSOCIATED(scf_control)))CALL cp__a("scf_control_types.F",610)
      IF(.NOT.(scf_control%ref_count > 0))CALL cp__a("scf_control_types.F",611)

      scf_section => section_vals_get_subs_vals(dft_section, "SCF")
      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".scfLog")

      IF (output_unit > 0) THEN

         IF (scf_control%max_scf > 0) THEN

            CALL create_scf_section(section)

            keyword => section_get_keyword(section, "SCF_GUESS")
            CALL keyword_get(keyword, enum=enum)

            WRITE (UNIT=output_unit, &
                   FMT="(/,/,T2,A,T25,A,T51,A30,/,T25,56('-'),3(/,T25,A,T76,I5),/,"// &
                   "T25,56('-'),4(/,T25,A,T72,ES9.2),/,T25,56('-'),"// &
                   "1(/,T25,A,T76,F5.2))") &
               "SCF PARAMETERS", &
               "Density guess:     ", ADJUSTR(TRIM(enum_i2c(enum, scf_control%density_guess))), &
               "max_scf:           ", scf_control%max_scf, &
               "max_scf_history:   ", scf_control%max_scf_hist, &
               "max_diis:          ", scf_control%max_diis, &
               "eps_scf:           ", scf_control%eps_scf, &
               "eps_scf_history:   ", scf_control%eps_scf_hist, &
               "eps_diis:          ", scf_control%eps_diis, &
               "eps_eigval:        ", scf_control%eps_eigval, &
               "level_shift [a.u.]:", scf_control%level_shift
            IF (SUM(scf_control%added_mos) > 0) THEN
               WRITE (UNIT=output_unit, FMT="(T25,A,T71,2I5)") &
                  "added MOs          ", scf_control%added_mos
            END IF

            IF (scf_control%mixing_method > 0 .AND. .NOT. scf_control%use_ot) THEN
               keyword => section_get_keyword(section, "MIXING%METHOD")
               CALL keyword_get(keyword, enum=enum)
               WRITE (UNIT=output_unit, FMT="(T25,A,/,T25,A,T51,A30)") &
                  REPEAT("-", 56), &
                  "Mixing method:      ", ADJUSTR(TRIM(enum_i2c(enum, scf_control%mixing_method)))
               IF (scf_control%mixing_method > 1) THEN
                  WRITE (UNIT=output_unit, FMT="(T47,A34)") "charge density mixing in g-space"
               END IF
            END IF
            IF (scf_control%smear%do_smear) THEN
               keyword => section_get_keyword(section, "SMEAR%METHOD")
               CALL keyword_get(keyword, enum=enum)
               WRITE (UNIT=output_unit, FMT="(T25,A,/,T25,A,T51,A30)") &
                  REPEAT("-", 56), &
                  "Smear method:      ", ADJUSTR(TRIM(enum_i2c(enum, scf_control%smear%method)))
               SELECT CASE (scf_control%smear%method)
               CASE (smear_fermi_dirac)
                  elec_temp = cp_unit_from_cp2k(scf_control%smear%electronic_temperature, &
                                                "K")
                  WRITE (UNIT=output_unit, FMT="(T25,A,T61,F20.1)") &
                     "Electronic temperature [K]:", elec_temp
                  WRITE (UNIT=output_unit, FMT="(T25,A,T71,ES10.2)") &
                     "Electronic temperature [a.u.]:", scf_control%smear%electronic_temperature, &
                     "Accuracy threshold:", scf_control%smear%eps_fermi_dirac
                  IF (scf_control%smear%fixed_mag_mom > 0.0_dp) WRITE (UNIT=output_unit, FMT="(T25,A,F10.5)") &
                     "Spin channel alpha and spin channel beta are smeared independently, keeping "// &
                     " fixed difference in number of electrons equal to ", scf_control%smear%fixed_mag_mom
               CASE (smear_energy_window)
                  WRITE (UNIT=output_unit, FMT="(T25,A,T71,F10.6)") &
                     "Smear window [a.u.]:       ", scf_control%smear%window_size
               END SELECT
            END IF

            CALL section_vals_val_get(dft_section, "ROKS", l_val=roks)
            IF (roks .AND. (.NOT. scf_control%use_ot)) THEN
               CALL section_vals_val_get(scf_section, "ROKS_SCHEME", &
                                         i_val=roks_scheme)
               keyword => section_get_keyword(section, "ROKS_SCHEME")
               CALL keyword_get(keyword, enum=enum)
               WRITE (UNIT=output_unit, FMT="(T25,A,/,T25,A,T51,A30)") &
                  REPEAT("-", 56), &
                  "ROKS scheme:", ADJUSTR(TRIM(enum_i2c(enum, roks_scheme)))
               SELECT CASE (roks_scheme)
               CASE (general_roks)
                  WRITE (UNIT=output_unit, FMT="(T25,A,T71,F10.6)") &
                     "ROKS parameter f:", scf_control%roks_f
               CASE (high_spin_roks)
                  WRITE (UNIT=output_unit, &
                         FMT="(T25,A,6(/,T25,A,T71,F10.6))") &
                     "ROKS parameters: a)lpha, b)eta; c)losed, o)pen, v)irtual", &
                     "acc", scf_control%roks_parameter(2, 2, 1), &
                     "bcc", scf_control%roks_parameter(2, 2, 2), &
                     "aoo", scf_control%roks_parameter(1, 1, 1), &
                     "boo", scf_control%roks_parameter(1, 1, 2), &
                     "avv", scf_control%roks_parameter(0, 0, 1), &
                     "bvv", scf_control%roks_parameter(0, 0, 2)
               END SELECT
            END IF
            CALL section_release(section)

            IF (scf_control%outer_scf%have_scf) THEN
               WRITE (output_unit, "(T25,56('-'),/,T25,A)") "Outer loop SCF in use "
               SELECT CASE (scf_control%outer_scf%type)
               CASE (outer_scf_none)
                  WRITE (output_unit, '(T25,A)') "No variables optimised in outer loop"
               CASE (outer_scf_ddapc_constraint)
                  WRITE (output_unit, '(T25,A)') "DDAPC constraint enforced"
               CASE (outer_scf_s2_constraint)
                  WRITE (output_unit, '(T25,A)') "S2 constraint enforced"
               CASE (outer_scf_becke_constraint)
                  WRITE (output_unit, '(T25,A)') "Becke weight population constraint enforced"
               CASE DEFAULT
                  CALL cp__b("scf_control_types.F",718,"")
               END SELECT
               WRITE (output_unit, '(T25,A,T72,ES9.2)') "eps_scf", scf_control%outer_scf%eps_scf
               WRITE (output_unit, '(T25,A,T72,I9)') "max_scf", scf_control%outer_scf%max_scf
               SELECT CASE (scf_control%outer_scf%optimizer)
               CASE (outer_scf_optimizer_none)
                  WRITE (output_unit, '(T25,A)') "No outer loop optimization"
               CASE (outer_scf_optimizer_sd)
                  WRITE (output_unit, '(T25,A)') "Steepest descent optimization"
               CASE (outer_scf_optimizer_bisect)
                  WRITE (output_unit, '(T25,A)') "Gradient bisection"
                  WRITE (output_unit, '(T25,A,T72,I9)') "bisect_trust_count", scf_control%outer_scf%bisect_trust_count
               CASE (outer_scf_optimizer_diis)
                  WRITE (output_unit, '(T25,A)') "DIIS optimization"
                  WRITE (output_unit, '(T25,A,T72,I9)') "DIIS buffer length", &
                     scf_control%outer_scf%diis_buffer_length
               CASE DEFAULT
                  CALL cp__b("scf_control_types.F",735,"")
               END SELECT
               WRITE (output_unit, '(T25,A,T72,ES9.2)') "step_size", scf_control%outer_scf%step_size
            ELSE
               WRITE (output_unit, "(T25,56('-'),/,T25,A)") "No outer SCF"
            END IF

         END IF ! max_scf > 0

      END IF ! output_unit > 0

      CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE scf_c_write_parameters

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param settings ...
!> \param scf_section ...
! **************************************************************************************************
   SUBROUTINE ot_diag_read_input(settings, scf_section)
      TYPE(qs_ot_settings_type)                          :: settings
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(len=*), PARAMETER :: routineN = 'ot_diag_read_input', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, output_unit
      LOGICAL                                            :: explicit
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER                   :: ot_section

      CALL timeset(routineN, handle)

      logger => cp_get_default_logger()
      output_unit = cp_print_key_unit_nr(logger, scf_section, "PRINT%PROGRAM_RUN_INFO", &
                                         extension=".log")

      ! decide default settings
      CALL qs_ot_settings_init(settings)

      ! use ot input new style
      ot_section => section_vals_get_subs_vals(scf_section, "DIAGONALIZATION%OT")
      CALL section_vals_get(ot_section, explicit=explicit)

      CALL ot_readwrite_input(settings, ot_section, output_unit)

      CALL cp_print_key_finished_output(output_unit, logger, scf_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

      CALL timestop(handle)

   END SUBROUTINE ot_diag_read_input

! **************************************************************************************************

END MODULE scf_control_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief The types needed for the calculation of active space Hamiltonians
!> \par History
!>      04.2016 created [JGH]
!> \author JGH
! **************************************************************************************************
MODULE qs_active_space_types

   USE cp_dbcsr_interface,              ONLY: cp_csr_destroy,&
                                              cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_p_type,&
                                              csr_p_type
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release
   USE kinds,                           ONLY: dp
   USE qs_mo_types,                     ONLY: deallocate_mo_set,&
                                              mo_set_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_active_space_types'

   PUBLIC :: active_space_type, eri_type
   PUBLIC :: create_active_space_type, release_active_space_type

! **************************************************************************************************
!> \brief Quantities needed for AS determination
!> \author JGH
! **************************************************************************************************
   TYPE eri_gpw_type
      LOGICAL                       :: redo_poisson
      REAL(KIND=dp)                 :: cutoff
      REAL(KIND=dp)                 :: rel_cutoff
      REAL(KIND=dp)                 :: eps_grid
      INTEGER                       :: print_level
      LOGICAL                       :: store_wfn
   END TYPE eri_gpw_type

   TYPE eri_type
      INTEGER                       :: method
      INTEGER                       :: OPERATOR
      REAL(KIND=dp)                 :: operator_parameter
      INTEGER, DIMENSION(3)         :: periodicity
      REAL(KIND=dp)                 :: cutoff_radius
      REAL(KIND=dp)                 :: eps_integral
      TYPE(eri_gpw_type)            :: eri_gpw
      TYPE(csr_p_type), &
         DIMENSION(:), POINTER      :: eri => NULL()
   END TYPE eri_type

   TYPE active_space_type
      INTEGER                       :: nactive
      INTEGER                       :: ninactive
      INTEGER, DIMENSION(2)         :: ninspin
      INTEGER, DIMENSION(2)         :: nelectrons
      INTEGER                       :: multiplicity
      INTEGER                       :: nspins
      LOGICAL                       :: molecule
      TYPE(mo_set_p_type), DIMENSION(:), POINTER  :: mos_active
      TYPE(mo_set_p_type), DIMENSION(:), POINTER  :: mos_inactive
      TYPE(eri_type)                :: eri
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER   :: p_ref
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER   :: ks_sub
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER   :: h_sub
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER:: pmat_inactive
   END TYPE active_space_type

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param active_space_env ...
! **************************************************************************************************
   SUBROUTINE create_active_space_type(active_space_env)
      TYPE(active_space_type), POINTER                   :: active_space_env

      CHARACTER(len=*), PARAMETER :: routineN = 'create_active_space_type', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(active_space_env)) THEN
         CALL release_active_space_type(active_space_env)
      END IF

      ALLOCATE (active_space_env)
      NULLIFY (active_space_env%mos_active, active_space_env%mos_inactive)
      NULLIFY (active_space_env%ks_sub, active_space_env%p_ref)
      NULLIFY (active_space_env%h_sub)
      NULLIFY (active_space_env%pmat_inactive)

   END SUBROUTINE create_active_space_type

! **************************************************************************************************
!> \brief ...
!> \param active_space_env ...
! **************************************************************************************************
   SUBROUTINE release_active_space_type(active_space_env)
      TYPE(active_space_type), POINTER                   :: active_space_env

      CHARACTER(len=*), PARAMETER :: routineN = 'release_active_space_type', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: imo, isp

      IF (ASSOCIATED(active_space_env)) THEN

         IF (ASSOCIATED(active_space_env%mos_active)) THEN
            DO imo = 1, SIZE(active_space_env%mos_active)
               CALL deallocate_mo_set(active_space_env%mos_active(imo)%mo_set)
            END DO
            DEALLOCATE (active_space_env%mos_active)
         END IF

         IF (ASSOCIATED(active_space_env%mos_inactive)) THEN
            DO imo = 1, SIZE(active_space_env%mos_inactive)
               CALL deallocate_mo_set(active_space_env%mos_inactive(imo)%mo_set)
            END DO
            DEALLOCATE (active_space_env%mos_inactive)
         END IF

         CALL release_eri_type(active_space_env%eri)

         IF (ASSOCIATED(active_space_env%p_ref)) THEN
            DO isp = 1, SIZE(active_space_env%p_ref)
               CALL cp_fm_release(active_space_env%p_ref(isp)%matrix)
            END DO
            DEALLOCATE (active_space_env%p_ref)
         END IF

         IF (ASSOCIATED(active_space_env%ks_sub)) THEN
            DO isp = 1, SIZE(active_space_env%ks_sub)
               CALL cp_fm_release(active_space_env%ks_sub(isp)%matrix)
            END DO
            DEALLOCATE (active_space_env%ks_sub)
         END IF

         IF (ASSOCIATED(active_space_env%h_sub)) THEN
            DO isp = 1, SIZE(active_space_env%h_sub)
               CALL cp_fm_release(active_space_env%h_sub(isp)%matrix)
            END DO
            DEALLOCATE (active_space_env%h_sub)
         END IF

         IF (ASSOCIATED(active_space_env%pmat_inactive)) &
            CALL cp_dbcsr_deallocate_matrix_set(active_space_env%pmat_inactive)

         DEALLOCATE (active_space_env)
      END IF

   END SUBROUTINE release_active_space_type

! **************************************************************************************************
!> \brief ...
!> \param eri_env ...
! **************************************************************************************************
   SUBROUTINE release_eri_type(eri_env)
      TYPE(eri_type)                                     :: eri_env

      CHARACTER(len=*), PARAMETER :: routineN = 'release_eri_type', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(eri_env%eri)) THEN

         DO i = 1, SIZE(eri_env%eri)
            CALL cp_csr_destroy(eri_env%eri(i)%csr_mat)
            DEALLOCATE (eri_env%eri(i)%csr_mat)
         END DO
         DEALLOCATE (eri_env%eri)

      END IF

   END SUBROUTINE release_eri_type

! **************************************************************************************************

END MODULE qs_active_space_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief collect pointers to a block of reals
!> \note
!>      needed to work around NEC compiler bug
!> \author fxcoudert [2006.12]
! **************************************************************************************************
MODULE block_p_types

   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: block_p_type
! **************************************************************************************************
   TYPE block_p_type
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: block
   END TYPE block_p_type
END MODULE block_p_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
!> \brief General overlap type integrals containers
!> \par History
!>      - rewrite of PPNL and OCE integrals
! **************************************************************************************************
MODULE sap_kind_types

   USE kinds,                           ONLY: dp
   USE util,                            ONLY: locate,&
                                              sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'sap_kind_types'

   TYPE clist_type
      INTEGER                                    :: catom, nsgf_cnt
      INTEGER, DIMENSION(:), POINTER             :: sgf_list
      INTEGER, DIMENSION(3)                      :: cell
      LOGICAL                                    :: sgf_soft_only
      REAL(KIND=dp)                            :: maxac, maxach
      REAL(KIND=dp), DIMENSION(3)              :: rac
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: acint
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: achint
   END TYPE clist_type

   TYPE alist_type
      INTEGER                                    :: aatom
      INTEGER                                    :: nclist
      TYPE(clist_type), DIMENSION(:), POINTER    :: clist
   END TYPE alist_type

   TYPE sap_int_type
      INTEGER                                    :: a_kind, p_kind
      INTEGER                                    :: nalist
      TYPE(alist_type), DIMENSION(:), POINTER    :: alist
      INTEGER, DIMENSION(:), POINTER             :: asort, aindex
   END TYPE sap_int_type

   PUBLIC :: sap_int_type, clist_type, alist_type, &
             release_sap_int, get_alist, alist_pre_align_blk, &
             alist_post_align_blk, sap_sort

CONTAINS

!==========================================================================================================

! **************************************************************************************************
!> \brief ...
!> \param sap_int ...
! **************************************************************************************************
   SUBROUTINE release_sap_int(sap_int)

      TYPE(sap_int_type), DIMENSION(:), POINTER          :: sap_int

      CHARACTER(LEN=*), PARAMETER :: routineN = 'release_sap_int', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, k
      TYPE(clist_type), POINTER                          :: clist

      IF(.NOT.(ASSOCIATED(sap_int)))CALL cp__a("sap_kind_types.F",69)

      DO i = 1, SIZE(sap_int)
         IF (ASSOCIATED(sap_int(i)%alist)) THEN
            DO j = 1, SIZE(sap_int(i)%alist)
               IF (ASSOCIATED(sap_int(i)%alist(j)%clist)) THEN
                  DO k = 1, SIZE(sap_int(i)%alist(j)%clist)
                     clist => sap_int(i)%alist(j)%clist(k)
                     IF (ASSOCIATED(clist%acint)) THEN
                        DEALLOCATE (clist%acint)
                     END IF
                     IF (ASSOCIATED(clist%sgf_list)) THEN
                        DEALLOCATE (clist%sgf_list)
                     END IF
                     IF (ASSOCIATED(clist%achint)) THEN
                        DEALLOCATE (clist%achint)
                     END IF
                  END DO
                  DEALLOCATE (sap_int(i)%alist(j)%clist)
               END IF
            END DO
            DEALLOCATE (sap_int(i)%alist)
         END IF
         IF (ASSOCIATED(sap_int(i)%asort)) THEN
            DEALLOCATE (sap_int(i)%asort)
         END IF
         IF (ASSOCIATED(sap_int(i)%aindex)) THEN
            DEALLOCATE (sap_int(i)%aindex)
         END IF
      END DO

      DEALLOCATE (sap_int)

   END SUBROUTINE release_sap_int

! **************************************************************************************************
!> \brief ...
!> \param sap_int ...
!> \param alist ...
!> \param atom ...
! **************************************************************************************************
   SUBROUTINE get_alist(sap_int, alist, atom)

      TYPE(sap_int_type), INTENT(IN)                     :: sap_int
      TYPE(alist_type), INTENT(OUT), POINTER             :: alist
      INTEGER, INTENT(IN)                                :: atom

      CHARACTER(LEN=*), PARAMETER :: routineN = 'get_alist', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      NULLIFY (alist)
      i = locate(sap_int%asort, atom)
      IF (i > 0 .AND. i <= SIZE(sap_int%alist)) THEN
         i = sap_int%aindex(i)
         alist => sap_int%alist(i)
      ELSE IF (i == 0) THEN
         NULLIFY (alist)
      ELSE
         CALL cp__b("sap_kind_types.F",128,"")
      END IF

   END SUBROUTINE get_alist

! **************************************************************************************************
!> \brief ...
!> \param blk_in ...
!> \param ldin ...
!> \param blk_out ...
!> \param ldout ...
!> \param ilist ...
!> \param in ...
!> \param jlist ...
!> \param jn ...
! **************************************************************************************************
   SUBROUTINE alist_pre_align_blk(blk_in, ldin, blk_out, ldout, ilist, in, jlist, jn)
      INTEGER, INTENT(IN)                                :: in, ilist(*), ldout
      REAL(dp), INTENT(INOUT)                            :: blk_out(ldout, *)
      INTEGER, INTENT(IN)                                :: ldin
      REAL(dp), INTENT(IN)                               :: blk_in(ldin, *)
      INTEGER, INTENT(IN)                                :: jlist(*), jn

      INTEGER                                            :: i, i0, i1, i2, i3, inn, inn1, j, j0

      inn = MOD(in, 4)
      inn1 = inn+1
      DO j = 1, jn
         j0 = jlist(j)
         DO i = 1, inn
            i0 = ilist(i)
            blk_out(i, j) = blk_in(i0, j0)
         ENDDO
         DO i = inn1, in, 4
            i0 = ilist(i)
            i1 = ilist(i+1)
            i2 = ilist(i+2)
            i3 = ilist(i+3)
            blk_out(i, j) = blk_in(i0, j0)
            blk_out(i+1, j) = blk_in(i1, j0)
            blk_out(i+2, j) = blk_in(i2, j0)
            blk_out(i+3, j) = blk_in(i3, j0)
         ENDDO
      ENDDO
   END SUBROUTINE alist_pre_align_blk

! **************************************************************************************************
!> \brief ...
!> \param blk_in ...
!> \param ldin ...
!> \param blk_out ...
!> \param ldout ...
!> \param ilist ...
!> \param in ...
!> \param jlist ...
!> \param jn ...
! **************************************************************************************************
   SUBROUTINE alist_post_align_blk(blk_in, ldin, blk_out, ldout, ilist, in, jlist, jn)
      INTEGER, INTENT(IN)                                :: in, ilist(*), ldout
      REAL(dp), INTENT(INOUT)                            :: blk_out(ldout, *)
      INTEGER, INTENT(IN)                                :: ldin
      REAL(dp), INTENT(IN)                               :: blk_in(ldin, *)
      INTEGER, INTENT(IN)                                :: jlist(*), jn

      INTEGER                                            :: i, i0, i1, i2, i3, inn, inn1, j, j0

      inn = MOD(in, 4)
      inn1 = inn+1
      DO j = 1, jn
         j0 = jlist(j)
         DO i = 1, inn
            i0 = ilist(i)
            blk_out(i0, j0) = blk_out(i0, j0)+blk_in(i, j)
         ENDDO
         DO i = inn1, in, 4
            i0 = ilist(i)
            i1 = ilist(i+1)
            i2 = ilist(i+2)
            i3 = ilist(i+3)
            blk_out(i0, j0) = blk_out(i0, j0)+blk_in(i, j)
            blk_out(i1, j0) = blk_out(i1, j0)+blk_in(i+1, j)
            blk_out(i2, j0) = blk_out(i2, j0)+blk_in(i+2, j)
            blk_out(i3, j0) = blk_out(i3, j0)+blk_in(i+3, j)
         ENDDO
      ENDDO
   END SUBROUTINE alist_post_align_blk

! **************************************************************************************************
!> \brief ...
!> \param sap_int ...
! **************************************************************************************************
   SUBROUTINE sap_sort(sap_int)
      TYPE(sap_int_type), DIMENSION(:), POINTER          :: sap_int

      CHARACTER(LEN=*), PARAMETER :: routineN = 'sap_sort', routineP = moduleN//':'//routineN

      INTEGER                                            :: iac, na

! *** Set up a sorting index

      DO iac = 1, SIZE(sap_int)
         IF (.NOT. ASSOCIATED(sap_int(iac)%alist)) CYCLE
         na = SIZE(sap_int(iac)%alist)
         ALLOCATE (sap_int(iac)%asort(na), sap_int(iac)%aindex(na))
         sap_int(iac)%asort(1:na) = sap_int(iac)%alist(1:na)%aatom
         CALL sort(sap_int(iac)%asort, na, sap_int(iac)%aindex)
      END DO

   END SUBROUTINE sap_sort

!==========================================================================================================

END MODULE sap_kind_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
MODULE qs_oce_types

   USE block_p_types,                   ONLY: block_p_type
   USE kinds,                           ONLY: dp
   USE sap_kind_types,                  ONLY: release_sap_int,&
                                              sap_int_type
   USE termination,                     ONLY: stop_memory
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_oce_types'

! *** Define a oce matrix type ***

! **************************************************************************************************
   TYPE qlist_type
      TYPE(block_p_type), DIMENSION(:), POINTER :: sac
      REAL(KIND=dp), DIMENSION(:), POINTER    :: r2
      REAL(KIND=dp), DIMENSION(:, :), POINTER  :: r
      INTEGER                                   :: n
      INTEGER, DIMENSION(:), POINTER            :: index_list, list
   END TYPE qlist_type

! **************************************************************************************************
   TYPE oce_matrix_type
      TYPE(sap_int_type), DIMENSION(:), POINTER        :: intac
   END TYPE

! *** Public data types ***

   PUBLIC :: oce_matrix_type

! *** Public subroutines ***

   PUBLIC :: allocate_oce_set, &
             create_oce_set, &
             deallocate_oce_set

CONTAINS

! **************************************************************************************************
!> \brief   Allocate and initialize the matrix set of oce coefficients.
!> \param oce_set ...
!> \param nkind ...
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE allocate_oce_set(oce_set, nkind)
      TYPE(oce_matrix_type), POINTER                     :: oce_set
      INTEGER, INTENT(IN)                                :: nkind

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_oce_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      ALLOCATE (oce_set%intac(nkind*nkind))
      DO i = 1, nkind*nkind
         NULLIFY (oce_set%intac(i)%alist)
         NULLIFY (oce_set%intac(i)%asort)
         NULLIFY (oce_set%intac(i)%aindex)
      END DO

   END SUBROUTINE allocate_oce_set

! **************************************************************************************************
!> \brief ...
!> \param oce_set ...
! **************************************************************************************************
   SUBROUTINE create_oce_set(oce_set)

      TYPE(oce_matrix_type), POINTER                     :: oce_set

      CHARACTER(len=*), PARAMETER :: routineN = 'create_oce_set', routineP = moduleN//':'//routineN

      INTEGER                                            :: istat

      IF (ASSOCIATED(oce_set)) CALL deallocate_oce_set(oce_set)

      ALLOCATE (oce_set, STAT=istat)
      IF (istat /= 0) CALL stop_memory(routineN, moduleN, 90, "oce_set", 0)

      NULLIFY (oce_set%intac)

   END SUBROUTINE create_oce_set

! **************************************************************************************************
!> \brief  Deallocate the matrix set of oce coefficients
!> \param oce_set ...
!> \date
!> \author
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE deallocate_oce_set(oce_set)
      TYPE(oce_matrix_type), POINTER                     :: oce_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_oce_set', &
         routineP = moduleN//':'//routineN

      IF (.NOT. ASSOCIATED(oce_set)) RETURN

      IF (ASSOCIATED(oce_set%intac)) CALL release_sap_int(oce_set%intac)

      DEALLOCATE (oce_set)

   END SUBROUTINE deallocate_oce_set

! **************************************************************************************************

END MODULE qs_oce_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Interface to the PEXSI library, providing wrappers for all PEXSI
!>        routines that are called inside CP2K. Tested for PEXSI versions
!>        v0.7.3 - v0.9.0
!> \par History
!>       2014.12 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
MODULE pexsi_interface

  USE kinds,                           ONLY: int_8,&
                                             real_8
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pexsi_interface'

   PUBLIC :: cp_pexsi_options, cp_pexsi_plan_initialize, &
             cp_pexsi_load_real_symmetric_hs_matrix, cp_pexsi_dft_driver, &
             cp_pexsi_retrieve_real_symmetric_dft_matrix, cp_pexsi_plan_finalize, &
             cp_pexsi_set_options, cp_pexsi_get_options, cp_pexsi_set_default_options

   TYPE cp_pexsi_options
      PRIVATE
      INTEGER :: unused = -1
   END TYPE cp_pexsi_options

CONTAINS

! **************************************************************************************************
!> \brief Set PEXSI internal options
!> \param pexsi_options ...
!> \param temperature ...
!> \param gap ...
!> \param deltaE ...
!> \param numPole ...
!> \param isInertiaCount ...
!> \param maxPEXSIIter ...
!> \param muMin0 ...
!> \param muMax0 ...
!> \param mu0 ...
!> \param muInertiaTolerance ...
!> \param muInertiaExpansion ...
!> \param muPEXSISafeGuard ...
!> \param numElectronPEXSITolerance ...
!> \param matrixType ...
!> \param isSymbolicFactorize ...
!> \param ordering ...
!> \param npSymbFact ...
!> \param verbosity ...
! **************************************************************************************************
   SUBROUTINE cp_pexsi_set_options(pexsi_options, temperature, gap, deltaE, numPole, &
                                   isInertiaCount, maxPEXSIIter, muMin0, muMax0, mu0, &
                                   muInertiaTolerance, muInertiaExpansion, &
                                   muPEXSISafeGuard, numElectronPEXSITolerance, &
                                   matrixType, isSymbolicFactorize, ordering, &
                                   npSymbFact, verbosity)

      TYPE(cp_pexsi_options), INTENT(INOUT)    :: pexsi_options
      REAL(KIND=real_8), INTENT(IN), OPTIONAL  :: temperature, gap, deltaE
      INTEGER, INTENT(IN), OPTIONAL            :: numPole, isInertiaCount, &
                                                  maxPEXSIIter
      REAL(KIND=real_8), INTENT(IN), OPTIONAL :: muMin0, muMax0, mu0, &
                                                 muInertiaTolerance, muInertiaExpansion, muPEXSISafeGuard, &
                                                 numElectronPEXSITolerance
      INTEGER, INTENT(IN), OPTIONAL            :: matrixType, &
                                                  isSymbolicFactorize, &
                                                  ordering, npSymbFact, &
                                                  verbosity

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_pexsi_set_options', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pexsi_options))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(temperature))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(gap))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(deltaE))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(numPole))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(isInertiaCount))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(maxPEXSIIter))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(muMin0))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(muMax0))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(mu0))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(muInertiaTolerance))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(muInertiaExpansion))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(muPEXSISafeGuard))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(numElectronPEXSITolerance))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(matrixType))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(isSymbolicFactorize))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(ordering))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(npSymbFact))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(verbosity))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("pexsi_interface.F",139,"Requires linking to the PEXSI library.")
   END SUBROUTINE cp_pexsi_set_options

! **************************************************************************************************
!> \brief Access PEXSI internal options
!> \param pexsi_options ...
!> \param temperature ...
!> \param gap ...
!> \param deltaE ...
!> \param numPole ...
!> \param isInertiaCount ...
!> \param maxPEXSIIter ...
!> \param muMin0 ...
!> \param muMax0 ...
!> \param mu0 ...
!> \param muInertiaTolerance ...
!> \param muInertiaExpansion ...
!> \param muPEXSISafeGuard ...
!> \param numElectronPEXSITolerance ...
!> \param matrixType ...
!> \param isSymbolicFactorize ...
!> \param ordering ...
!> \param npSymbFact ...
!> \param verbosity ...
! **************************************************************************************************
   SUBROUTINE cp_pexsi_get_options(pexsi_options, temperature, gap, deltaE, numPole, &
                                   isInertiaCount, maxPEXSIIter, muMin0, muMax0, mu0, &
                                   muInertiaTolerance, muInertiaExpansion, &
                                   muPEXSISafeGuard, numElectronPEXSITolerance, &
                                   matrixType, isSymbolicFactorize, ordering, &
                                   npSymbFact, verbosity)
      TYPE(cp_pexsi_options), INTENT(IN)       :: pexsi_options
      REAL(KIND=real_8), INTENT(OUT), OPTIONAL :: temperature, gap, deltaE
      INTEGER, INTENT(OUT), OPTIONAL           :: numPole, isInertiaCount, &
                                                  maxPEXSIIter
      REAL(KIND=real_8), INTENT(OUT), OPTIONAL :: muMin0, muMax0, mu0, &
                                                  muInertiaTolerance, muInertiaExpansion, muPEXSISafeGuard, &
                                                  numElectronPEXSITolerance
      INTEGER, INTENT(OUT), OPTIONAL           :: matrixType, &
                                                  isSymbolicFactorize, &
                                                  ordering, npSymbFact, &
                                                  verbosity

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_pexsi_get_options', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pexsi_options))==-1) EXIT ;  END DO ; ENDIF
      ! assign intent-out arguments to silence compiler warnings
      IF (PRESENT(temperature)) temperature = 0.0_real_8
      IF (PRESENT(gap)) gap = 0.0_real_8
      IF (PRESENT(deltaE)) deltaE = 0.0_real_8
      IF (PRESENT(numPole)) numPole = -1
      IF (PRESENT(isInertiaCount)) isInertiaCount = -1
      IF (PRESENT(maxPEXSIIter)) maxPEXSIIter = -1
      IF (PRESENT(muMin0)) muMin0 = 0.0_real_8
      IF (PRESENT(muMax0)) muMax0 = 0.0_real_8
      IF (PRESENT(mu0)) mu0 = 0.0_real_8
      IF (PRESENT(muInertiaTolerance)) muInertiaTolerance = 0.0_real_8
      IF (PRESENT(muInertiaExpansion)) muInertiaExpansion = 0.0_real_8
      IF (PRESENT(muPEXSISafeGuard)) muPEXSISafeGuard = 0.0_real_8
      IF (PRESENT(numElectronPEXSITolerance)) numElectronPEXSITolerance = 0.0_real_8
      IF (PRESENT(matrixType)) matrixType = -1
      IF (PRESENT(isSymbolicFactorize)) isSymbolicFactorize = -1
      IF (PRESENT(ordering)) ordering = -1
      IF (PRESENT(npSymbFact)) npSymbFact = -1
      IF (PRESENT(verbosity)) verbosity = -1
      CALL cp__b("pexsi_interface.F",231,"Requires linking to the PEXSI library.")
   END SUBROUTINE cp_pexsi_get_options

! **************************************************************************************************
!> \brief ...
!> \param pexsi_options ...
! **************************************************************************************************
   SUBROUTINE cp_pexsi_set_default_options(pexsi_options)
      TYPE(cp_pexsi_options), INTENT(OUT)      :: pexsi_options

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_pexsi_set_default_options', &
                                     routineP = moduleN//':'//routineN

      CALL cp__b("pexsi_interface.F",248,"Requires linking to the PEXSI library.")
   END SUBROUTINE cp_pexsi_set_default_options

! **************************************************************************************************
!> \brief ...
!> \param comm ...
!> \param numProcRow ...
!> \param numProcCol ...
!> \param outputFileIndex ...
!> \retval cp_pexsi_plan_initialize ...
! **************************************************************************************************
   FUNCTION cp_pexsi_plan_initialize(comm, numProcRow, numProcCol, outputFileIndex)
      INTEGER, INTENT(IN)                      :: comm, numProcRow, numProcCol, &
                                                  outputFileIndex
      INTEGER(KIND=int_8)                      :: cp_pexsi_plan_initialize

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_pexsi_plan_initialize', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(comm))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(numProcRow))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(numProcCol))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(outputFileIndex))==-1) EXIT ;  END DO ; ENDIF
      cp_pexsi_plan_initialize = 0
      CALL cp__b("pexsi_interface.F",282,"Requires linking to the PEXSI library.")
   END FUNCTION cp_pexsi_plan_initialize

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param pexsi_options ...
!> \param nrows ...
!> \param nnz ...
!> \param nnzLocal ...
!> \param numColLocal ...
!> \param colptrLocal ...
!> \param rowindLocal ...
!> \param HnzvalLocal ...
!> \param isSIdentity ...
!> \param SnzvalLocal ...
! **************************************************************************************************
   SUBROUTINE cp_pexsi_load_real_symmetric_hs_matrix(plan, pexsi_options, nrows, nnz, &
                                                     nnzLocal, numColLocal, colptrLocal, &
                                                     rowindLocal, HnzvalLocal, isSIdentity, &
                                                     SnzvalLocal)
      INTEGER(KIND=int_8), INTENT(IN)          :: plan
      TYPE(cp_pexsi_options), INTENT(IN)       :: pexsi_options
      INTEGER, INTENT(IN)                      :: nrows, nnz, nnzLocal, &
                                                  numColLocal, colptrLocal(*), &
                                                  rowindLocal(*)
      REAL(KIND=real_8), INTENT(IN)            :: HnzvalLocal(*)
      INTEGER, INTENT(IN)                      :: isSIdentity
      REAL(KIND=real_8), INTENT(IN)            :: SnzvalLocal(*)

      CHARACTER(LEN=*), PARAMETER :: &
         routineN = 'cp_pexsi_load_real_symmetric_hs_matrix', &
         routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pexsi_options))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nrows))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nnz))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(nnzLocal))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(numColLocal))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(isSIdentity))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("pexsi_interface.F",335,"Requires linking to the PEXSI library.")

      ! MARK_USED macro does not work on assumed shape variables
      IF (.FALSE.) THEN; DO
            IF (colptrLocal(1) > rowindLocal(1) .OR. HnzvalLocal(1) > SnzvalLocal(1)) EXIT
         ENDDO; ENDIF
   END SUBROUTINE cp_pexsi_load_real_symmetric_hs_matrix

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param pexsi_options ...
!> \param numElectronExact ...
!> \param muPEXSI ...
!> \param numElectronPEXSI ...
!> \param muMinInertia ...
!> \param muMaxInertia ...
!> \param numTotalInertiaIter ...
!> \param numTotalPEXSIIter ...
! **************************************************************************************************
   SUBROUTINE cp_pexsi_dft_driver(plan, pexsi_options, numElectronExact, muPEXSI, &
                                  numElectronPEXSI, muMinInertia, muMaxInertia, &
                                  numTotalInertiaIter, numTotalPEXSIIter)
      INTEGER(KIND=int_8), INTENT(IN)          :: plan
      TYPE(cp_pexsi_options), INTENT(IN)       :: pexsi_options
      REAL(KIND=real_8), INTENT(IN)            :: numElectronExact
      REAL(KIND=real_8), INTENT(out)           :: muPEXSI, numElectronPEXSI, &
                                                  muMinInertia, muMaxInertia
      INTEGER, INTENT(out)                     :: numTotalInertiaIter, &
                                                  numTotalPEXSIIter

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_pexsi_dft_driver', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(numelectronexact))==-1) EXIT ;  END DO ; ENDIF
      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(pexsi_options))==-1) EXIT ;  END DO ; ENDIF
      ! assign intent-out arguments to silence compiler warnings
      muPEXSI = 0.0_real_8
      numElectronPEXSI = 0.0_real_8
      muMinInertia = 0.0_real_8
      muMaxInertia = 0.0_real_8
      numTotalInertiaIter = -1
      numTotalPEXSIIter = -1
      CALL cp__b("pexsi_interface.F",391,"Requires linking to the PEXSI library.")
   END SUBROUTINE cp_pexsi_dft_driver

! **************************************************************************************************
!> \brief ...
!> \param plan ...
!> \param DMnzvalLocal ...
!> \param EDMnzvalLocal ...
!> \param FDMnzvalLocal ...
!> \param totalEnergyH ...
!> \param totalEnergyS ...
!> \param totalFreeEnergy ...
! **************************************************************************************************
   SUBROUTINE cp_pexsi_retrieve_real_symmetric_dft_matrix(plan, DMnzvalLocal, EDMnzvalLocal, &
                                                          FDMnzvalLocal, totalEnergyH, &
                                                          totalEnergyS, totalFreeEnergy)
      INTEGER(KIND=int_8), INTENT(IN)          :: plan
      REAL(KIND=real_8), INTENT(out) :: DMnzvalLocal(*), EDMnzvalLocal(*), &
                                        FDMnzvalLocal(*), totalEnergyH, totalEnergyS, totalFreeEnergy

      CHARACTER(LEN=*), PARAMETER :: &
         routineN = 'cp_pexsi_retrieve_real_symmetric_dft_matrix', &
         routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan))==-1) EXIT ;  END DO ; ENDIF
      ! assign intent-out arguments to silence compiler warnings
      DMnzvalLocal(1) = 0.0_real_8
      EDMnzvalLocal(1) = 0.0_real_8
      FDMnzvalLocal(1) = 0.0_real_8
      totalEnergyH = 0.0_real_8
      totalEnergyS = 0.0_real_8
      totalFreeEnergy = 0.0_real_8

      CALL cp__b("pexsi_interface.F",436,"Requires linking to the PEXSI library.")
   END SUBROUTINE cp_pexsi_retrieve_real_symmetric_dft_matrix

! **************************************************************************************************
!> \brief ...
!> \param plan ...
! **************************************************************************************************
   SUBROUTINE cp_pexsi_plan_finalize(plan)
      INTEGER(KIND=int_8), INTENT(IN)          :: plan

      CHARACTER(LEN=*), PARAMETER :: routineN = 'cp_pexsi_plan_finalize', &
                                     routineP = moduleN//':'//routineN

      IF(.FALSE.)THEN; DO ; IF(SIZE(SHAPE(plan))==-1) EXIT ;  END DO ; ENDIF
      CALL cp__b("pexsi_interface.F",459,"Requires linking to the PEXSI library.")
   END SUBROUTINE

END MODULE pexsi_interface
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Environment storing all data that is needed in order to call the DFT
!>        driver of the PEXSI library with data from the linear scaling quickstep
!>        SCF environment, mainly parameters and intermediate data for the matrix
!>        conversion between DBCSR and CSR format.
!> \par History
!>       2014.11 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************

MODULE pexsi_types
   USE bibliography,                    ONLY: Lin2009,&
                                              Lin2013,&
                                              cite_reference
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type,&
                                              cp_dbcsr_scale,&
                                              cp_dbcsr_type,&
                                              csr_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_unit_nr,&
                                              cp_logger_type
   USE kinds,                           ONLY: dp,&
                                              int_8
   USE message_passing,                 ONLY: mp_dims_create,&
                                              mp_environ
   USE pexsi_interface,                 ONLY: cp_pexsi_get_options,&
                                              cp_pexsi_options,&
                                              cp_pexsi_plan_finalize,&
                                              cp_pexsi_plan_initialize,&
                                              cp_pexsi_set_options
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'pexsi_types'

   LOGICAL, PARAMETER, PRIVATE          :: careful_mod = .FALSE.

   INTEGER, PARAMETER, PUBLIC           :: cp2k_to_pexsi = 1, pexsi_to_cp2k = 2

   PUBLIC :: lib_pexsi_env, lib_pexsi_init, lib_pexsi_finalize, &
             convert_nspin_cp2k_pexsi

! **************************************************************************************************
!> \brief All PEXSI related data
!> \param options                       PEXSI options
!> \param plan                          PEXSI plan
!> \param mp_group                      message-passing group ID
!> \param mp_dims                       dimensions of the MPI cartesian grid used
!>                                      for PEXSI
!> \param num_ranks_per_pole            number of MPI ranks per pole in PEXSI
!> \param kTS                           entropic energy contribution
!> \param matrix_w                      energy-weighted density matrix as needed
!>                                      for the forces
!> \param csr_mat*                      intermediate matrices in CSR format
!> \param dbcsr_template_matrix_sym     Symmetric template matrix fixing DBCSR
!>                                      sparsity pattern
!> \param dbcsr_template_matrix_nonsym  Nonsymmetric template matrix fixing
!>                                      DBCSR sparsity pattern
!> \param csr_sparsity                  DBCSR matrix defining CSR sparsity
!> \param csr_screening                 whether distance screening should be
!>                                      applied to CSR matrices
!> \param max_ev_vector                 eigenvector corresponding to the largest
!>                                      energy eigenvalue,
!>                                      returned by the Arnoldi method used to
!>                                      determine the spectral radius deltaE
!> \param nspin                         number of spins
!> \param do_adaptive_tol_nel           Whether or not to use adaptive threshold
!>                                      for PEXSI convergence
!> \param adaptive_nel_alpha            constants for adaptive thresholding
!> \param adaptive_nel_beta
!> \param tol_nel_initial               Initial convergence threshold (in number of electrons)
!> \param tol_nel_target                Target convergence threshold (in number of electrons)
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   TYPE lib_pexsi_env
      TYPE(cp_dbcsr_type)                      :: dbcsr_template_matrix_sym, &
                                                  dbcsr_template_matrix_nonsym
      TYPE(csr_type)                           :: csr_mat_p, csr_mat_ks, csr_mat_s, &
                                                  csr_mat_E, csr_mat_F
      TYPE(cp_pexsi_options)                   :: options
      REAL(KIND=dp), DIMENSION(:), POINTER     :: kTS => NULL()

      TYPE(cp_dbcsr_p_type), DIMENSION(:), &
         POINTER                                :: matrix_w => NULL()

      INTEGER(kind=int_8)                      :: plan
      INTEGER                                  :: nspin, num_ranks_per_pole, mp_group

      TYPE(cp_dbcsr_type), DIMENSION(:), &
         POINTER                                :: max_ev_vector
      TYPE(cp_dbcsr_type)                      :: csr_sparsity
      INTEGER, DIMENSION(2)                    :: mp_dims

      LOGICAL                                  :: csr_screening, do_adaptive_tol_nel
      REAL(KIND=dp)                            :: adaptive_nel_alpha, adaptive_nel_beta, &
                                                  tol_nel_initial, tol_nel_target
   END TYPE lib_pexsi_env

CONTAINS

! **************************************************************************************************
!> \brief Initialize PEXSI
!> \param pexsi_env All data needed by PEXSI
!> \param mp_group message-passing group ID
!> \param nspin number of spins
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE lib_pexsi_init(pexsi_env, mp_group, nspin)
      TYPE(lib_pexsi_env), INTENT(INOUT)                 :: pexsi_env
      INTEGER, INTENT(IN)                                :: mp_group, nspin

      CHARACTER(len=*), PARAMETER :: routineN = 'lib_pexsi_init', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, mynode, npSymbFact, &
                                                            numnodes, unit_nr
      TYPE(cp_logger_type), POINTER                      :: logger

      logger => cp_get_default_logger()
      IF (logger%para_env%mepos == logger%para_env%source) THEN
         unit_nr = cp_logger_get_default_unit_nr(logger, local=.TRUE.)
      ELSE
         unit_nr = -1
      ENDIF

      CALL timeset(routineN, handle)

      CALL cite_reference(Lin2009)
      CALL cite_reference(Lin2013)

      pexsi_env%mp_group = mp_group
      CALL mp_environ(numnodes, mynode, pexsi_env%mp_group)

      ! Use all nodes available per pole by default or if the user tries to use
      ! more nodes than MPI size
      IF ((pexsi_env%num_ranks_per_pole .GT. numnodes) &
          .OR. (pexsi_env%num_ranks_per_pole .EQ. 0)) THEN
         pexsi_env%num_ranks_per_pole = numnodes
      ENDIF

      ! Find num_ranks_per_pole from user input MIN_RANKS_PER_POLE s.t. num_ranks_per_pole
      ! is the smallest number greater or equal to min_ranks_per_pole that divides
      ! MPI size without remainder.
      DO WHILE (MOD(numnodes, pexsi_env%num_ranks_per_pole) .NE. 0)
         pexsi_env%num_ranks_per_pole = pexsi_env%num_ranks_per_pole+1
      ENDDO

      CALL cp_pexsi_get_options(pexsi_env%options, npSymbFact=npSymbFact)
      IF ((npSymbFact .GT. pexsi_env%num_ranks_per_pole) .OR. (npSymbFact .EQ. 0)) THEN
         ! Use maximum possible number of ranks for symbolic factorization
         CALL cp_pexsi_set_options(pexsi_env%options, npSymbFact=pexsi_env%num_ranks_per_pole)
      ENDIF

      ! Create dimensions for MPI cartesian grid for PEXSI
      pexsi_env%mp_dims = 0
      CALL mp_dims_create(pexsi_env%num_ranks_per_pole, pexsi_env%mp_dims)

      ! allocations with nspin
      pexsi_env%nspin = nspin
      ALLOCATE (pexsi_env%kTS(nspin))
      pexsi_env%kTS(:) = 0.0_dp
      ALLOCATE (pexsi_env%max_ev_vector(nspin))
      ALLOCATE (pexsi_env%matrix_w(nspin))
      DO ispin = 1, pexsi_env%nspin
         ALLOCATE (pexsi_env%matrix_w(ispin)%matrix)
      ENDDO

      ! Initialize PEXSI
      pexsi_env%plan = cp_pexsi_plan_initialize(pexsi_env%mp_group, pexsi_env%mp_dims(1), &
                                                pexsi_env%mp_dims(2), mynode)

      pexsi_env%do_adaptive_tol_nel = .FALSE.

      ! Print PEXSI infos
      IF (unit_nr > 0) CALL print_pexsi_info(pexsi_env, unit_nr)

      CALL timestop(handle)
   END SUBROUTINE lib_pexsi_init

! **************************************************************************************************
!> \brief Release all PEXSI data
!> \param pexsi_env ...
!> \par History
!>      11.2014 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE lib_pexsi_finalize(pexsi_env)
      TYPE(lib_pexsi_env), INTENT(INOUT)                 :: pexsi_env

      CHARACTER(len=*), PARAMETER :: routineN = 'lib_pexsi_finalize', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin

      CALL timeset(routineN, handle)
      CALL cp_pexsi_plan_finalize(pexsi_env%plan)
      DEALLOCATE (pexsi_env%kTS)
      DEALLOCATE (pexsi_env%max_ev_vector)
      DO ispin = 1, pexsi_env%nspin
         DEALLOCATE (pexsi_env%matrix_w(ispin)%matrix)
      ENDDO
      DEALLOCATE (pexsi_env%matrix_w)
      CALL timestop(handle)
   END SUBROUTINE lib_pexsi_finalize

! **************************************************************************************************
!> \brief Scale various quantities with factors of 2. This converts spin restricted
!>        DFT calculations (PEXSI) to the unrestricted case (as is the case where
!>        the density matrix method is called in the linear scaling code).
!> \param[in] direction ...
!> \param[in,out] numElectron ...
!> \param matrix_p ...
!> \param matrix_w ...
!> \param kTS ...
!> \par History
!>      01.2015 created [Patrick Seewald]
!> \author Patrick Seewald
! **************************************************************************************************
   SUBROUTINE convert_nspin_cp2k_pexsi(direction, numElectron, matrix_p, matrix_w, kTS)
      INTEGER, INTENT(IN)                                :: direction
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: numElectron
      TYPE(cp_dbcsr_type), INTENT(INOUT), OPTIONAL       :: matrix_p
      TYPE(cp_dbcsr_p_type), INTENT(INOUT), OPTIONAL     :: matrix_w
      REAL(KIND=dp), INTENT(INOUT), OPTIONAL             :: kTS

      CHARACTER(len=*), PARAMETER :: routineN = 'convert_nspin_cp2k_pexsi', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      REAL(KIND=dp)                                      :: scaling

      CALL timeset(routineN, handle)

      SELECT CASE (direction)
      CASE (cp2k_to_pexsi)
         scaling = 2.0_dp
      CASE (pexsi_to_cp2k)
         scaling = 0.5_dp
      END SELECT

      IF (PRESENT(numElectron)) numElectron = scaling*numElectron
      IF (PRESENT(matrix_p)) CALL cp_dbcsr_scale(matrix_p, scaling)
      IF (PRESENT(matrix_w)) CALL cp_dbcsr_scale(matrix_w%matrix, scaling)
      IF (PRESENT(kTS)) kTS = scaling*kTS

      CALL timestop(handle)
   END SUBROUTINE convert_nspin_cp2k_pexsi

! **************************************************************************************************
!> \brief Print relevant options of PEXSI
!> \param pexsi_env ...
!> \param unit_nr ...
! **************************************************************************************************
   SUBROUTINE print_pexsi_info(pexsi_env, unit_nr)
      TYPE(lib_pexsi_env), INTENT(IN)                    :: pexsi_env
      INTEGER, INTENT(IN)                                :: unit_nr

      INTEGER                                            :: mynode, npSymbFact, numnodes, numPole, &
                                                            ordering
      REAL(KIND=dp)                                      :: gap, muInertiaExpansion, &
                                                            muInertiaTolerance, muMax0, muMin0, &
                                                            muPEXSISafeGuard, temperature

      CALL mp_environ(numnodes, mynode, pexsi_env%mp_group)

      CALL cp_pexsi_get_options(pexsi_env%options, temperature=temperature, gap=gap, &
                                numPole=numPole, muMin0=muMin0, muMax0=muMax0, muInertiaTolerance= &
                                muInertiaTolerance, muInertiaExpansion=muInertiaExpansion, &
                                muPEXSISafeGuard=muPEXSISafeGuard, ordering=ordering, npSymbFact=npSymbFact)

      WRITE (unit_nr, '(/A)') " PEXSI| Initialized with parameters"
      WRITE (unit_nr, '(A,T61,E20.3)') " PEXSI|   Electronic temperature", temperature
      WRITE (unit_nr, '(A,T61,E20.3)') " PEXSI|   Spectral gap", gap
      WRITE (unit_nr, '(A,T61,I20)') " PEXSI|   Number of poles", numPole
      WRITE (unit_nr, '(A,T61,E20.3)') " PEXSI|   Target tolerance in number of electrons", &
         pexsi_env%tol_nel_target
      WRITE (unit_nr, '(A,T61,E20.3)') " PEXSI|   Convergence tolerance for inertia counting in mu", &
         muInertiaTolerance
      WRITE (unit_nr, '(A,T61,E20.3)') " PEXSI|   Restart tolerance for inertia counting in mu", &
         muPEXSISafeGuard
      WRITE (unit_nr, '(A,T61,E20.3)') " PEXSI|   Expansion of mu interval for inertia counting", &
         muInertiaExpansion

      WRITE (unit_nr, '(/A)') " PEXSI| Parallelization scheme"
      WRITE (unit_nr, '(A,T61,I20)') " PEXSI|   Number of poles processed in parallel", &
         numnodes/pexsi_env%num_ranks_per_pole
      WRITE (unit_nr, '(A,T61,I20)') " PEXSI|   Number of processors per pole", &
         pexsi_env%num_ranks_per_pole
      WRITE (unit_nr, '(A,T71,I5,T76,I5)') " PEXSI|   Process grid dimensions", &
         pexsi_env%mp_dims(1), pexsi_env%mp_dims(2)
      SELECT CASE (ordering)
      CASE (0)
         WRITE (unit_nr, '(A,T61,A20)') " PEXSI|   Ordering strategy", "parallel"
      CASE (1)
         WRITE (unit_nr, '(A,T61,A20)') " PEXSI|   Ordering strategy", "sequential"
      CASE (2)
         WRITE (unit_nr, '(A,T61,A20)') " PEXSI|   Ordering strategy", "multiple minimum degree"
      END SELECT
      IF (ordering .EQ. 0) WRITE (unit_nr, '(A,T61,I20/)') &
         " PEXSI|   Number of processors for symbolic factorization", npSymbFact

   END SUBROUTINE print_pexsi_info
END MODULE pexsi_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types needed for a linear scaling quickstep SCF run based on the density
!>        matrix
!> \par History
!>       2010.10 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
MODULE dm_ls_scf_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_release,&
                                              cp_dbcsr_type
   USE cp_para_env,                     ONLY: cp_para_env_release
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: ls_cluster_atomic,&
                                              ls_cluster_molecular
   USE input_section_types,             ONLY: section_vals_release,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE pao_types,                       ONLY: pao_env_type
   USE pexsi_types,                     ONLY: lib_pexsi_env,&
                                              lib_pexsi_finalize
   USE qs_density_mixing_types,         ONLY: mixing_storage_release,&
                                              mixing_storage_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'dm_ls_scf_types'

   PUBLIC :: ls_scf_env_type, ls_mstruct_type, ls_cluster_atomic, ls_cluster_molecular, &
             ls_scf_curvy_type

   TYPE ls_mstruct_type
      INTEGER :: cluster_type
      LOGICAL :: single_precision
      INTEGER, DIMENSION(:), ALLOCATABLE :: atom_to_molecule
      TYPE(cp_dbcsr_type)                :: matrix_A
      TYPE(cp_dbcsr_type)                :: matrix_B
   END TYPE

   TYPE ls_mat_history_type
      INTEGER :: istore, nstore
      TYPE(cp_dbcsr_type), DIMENSION(:, :), ALLOCATABLE :: matrix
   END TYPE

   TYPE ls_scf_curvy_type
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_dp
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_p
      TYPE(cp_dbcsr_type), DIMENSION(:, :), ALLOCATABLE :: matrix_psave
      TYPE(cp_dbcsr_type), DIMENSION(:, :), ALLOCATABLE :: matrix_BCH
      REAL(KIND=dp)                                    :: step_size(2)
      REAL(KIND=dp)                                    :: shift(2)
      REAL(KIND=dp)                                    :: cg_denom(2)
      REAL(KIND=dp)                                    :: cg_numer(2)
      REAL(KIND=dp), DIMENSION(6)                      :: energies
      INTEGER                                          :: line_search_step
      INTEGER                                          :: BCH_saved(2)
      LOGICAL                                          :: double_step_size
      LOGICAL                                          :: fix_shift(2)

      INTEGER                                          :: line_search_type
      INTEGER                                          :: n_bch_hist
      REAL(KIND=dp)                                    :: scale_filter
      REAL(KIND=dp)                                    :: filter_factor
      REAL(KIND=dp)                                    :: min_shift
      REAL(KIND=dp)                                    :: min_filter
   END TYPE

   TYPE chebyshev_type
      LOGICAL :: compute_chebyshev
      INTEGER :: n_chebyshev
      INTEGER :: n_gridpoint_dos
      REAL(KIND=dp), DIMENSION(:), POINTER :: min_energy => NULL()
      REAL(KIND=dp), DIMENSION(:), POINTER :: max_energy => NULL()
      TYPE(section_vals_type), POINTER :: print_key_dos => NULL()
      TYPE(section_vals_type), POINTER :: print_key_cube => NULL()
   END TYPE

   TYPE ls_scf_env_type
      INTEGER               :: nspins, natoms
      INTEGER               :: nelectron_total
      INTEGER, DIMENSION(2) :: nelectron_spin
      REAL(KIND=dp), DIMENSION(2) ::  mu_spin
      REAL(KIND=dp), DIMENSION(2) ::  homo_spin
      REAL(KIND=dp), DIMENSION(2) ::  lumo_spin

      TYPE(ls_mat_history_type) :: scf_history
      INTEGER :: extrapolation_order

      LOGICAL :: has_unit_metric

      LOGICAL :: curvy_steps
      INTEGER :: s_preconditioner_type
      INTEGER :: s_inversion_type
      INTEGER :: purification_method
      INTEGER :: sign_sqrt_order

      LOGICAL               :: needs_s_inv, has_s_preconditioner, fixed_mu, dynamic_threshold, check_s_inv
      LOGICAL               :: restart_read, restart_write, non_monotonic
      REAL(KIND=dp)         :: eps_filter, eps_scf

      REAL(KIND=dp)         :: eps_lanczos
      INTEGER               :: max_iter_lanczos

      REAL(KIND=dp)         :: mixing_fraction
      INTEGER               :: max_scf
      LOGICAL               :: ls_diis
      INTEGER               :: iter_ini_diis
      INTEGER               :: nmixing, max_diis
      REAL(KIND=dp)         :: eps_diis
      REAL(KIND=dp)         :: energy_init

      TYPE(cp_dbcsr_type)   :: matrix_s_inv
      TYPE(cp_dbcsr_type)   :: matrix_s
      TYPE(cp_dbcsr_type)   :: matrix_bs_sqrt, matrix_bs_sqrt_inv
      TYPE(cp_dbcsr_type)   :: matrix_s_sqrt, matrix_s_sqrt_inv
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_ks
      TYPE(cp_dbcsr_type), DIMENSION(:), ALLOCATABLE :: matrix_p

      LOGICAL  :: report_all_sparsities, perform_mu_scan, use_s_sqrt

      TYPE(ls_mstruct_type) :: ls_mstruct
      TYPE(ls_scf_curvy_type) :: curvy_data

      TYPE(chebyshev_type) :: chebyshev

      LOGICAL :: do_rho_mixing
      INTEGER :: density_mixing_method
      TYPE(mixing_storage_type), POINTER :: mixing_store

      LOGICAL :: do_transport
      LOGICAL :: do_pexsi

      TYPE(lib_pexsi_env) :: pexsi

      TYPE(cp_para_env_type), POINTER :: para_env
      LOGICAL                 :: do_pao = .FALSE.
      TYPE(pao_env_type)      :: pao_env
   END TYPE ls_scf_env_type

   PUBLIC :: ls_scf_release

CONTAINS

! **************************************************************************************************
!> \brief release the LS type.
!> \param ls_scf_env ...
!> \par History
!>       2012.11 created [Joost VandeVondele]
!> \author Joost VandeVondele
! **************************************************************************************************
   SUBROUTINE ls_scf_release(ls_scf_env)
      TYPE(ls_scf_env_type), POINTER                     :: ls_scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ls_scf_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ispin, istore

      CALL timeset(routineN, handle)

      CALL cp_para_env_release(ls_scf_env%para_env)

      DEALLOCATE (ls_scf_env%ls_mstruct%atom_to_molecule)

      ! set up the buffer for the history of matrices
      DO istore = 1, MIN(ls_scf_env%scf_history%istore, ls_scf_env%scf_history%nstore)
         DO ispin = 1, SIZE(ls_scf_env%scf_history%matrix, 1)
            CALL cp_dbcsr_release(ls_scf_env%scf_history%matrix(ispin, istore))
         ENDDO
      ENDDO
      DEALLOCATE (ls_scf_env%scf_history%matrix)

      IF (ASSOCIATED(ls_scf_env%chebyshev%print_key_dos)) &
         CALL section_vals_release(ls_scf_env%chebyshev%print_key_dos)
      IF (ASSOCIATED(ls_scf_env%chebyshev%print_key_cube)) &
         CALL section_vals_release(ls_scf_env%chebyshev%print_key_cube)
      IF (ASSOCIATED(ls_scf_env%chebyshev%min_energy)) THEN
         DEALLOCATE (ls_scf_env%chebyshev%min_energy)
      ENDIF
      IF (ASSOCIATED(ls_scf_env%chebyshev%max_energy)) THEN
         DEALLOCATE (ls_scf_env%chebyshev%max_energy)
      ENDIF

      IF (ASSOCIATED(ls_scf_env%mixing_store)) THEN
         CALL mixing_storage_release(ls_scf_env%mixing_store)
      ENDIF

      IF (ls_scf_env%do_pexsi) THEN
         CALL lib_pexsi_finalize(ls_scf_env%pexsi)
      ENDIF

      IF (ls_scf_env%do_pao) &
         CALL cp_dbcsr_release(ls_scf_env%pao_env%matrix_X)

      DEALLOCATE (ls_scf_env)

      CALL timestop(handle)

   END SUBROUTINE ls_scf_release

END MODULE dm_ls_scf_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief type for berry phase efield matrices. At the moment only used for
!>        cosmat and sinmat
!> \par History
!>      none
!> \author fschiff (06.2010)
! **************************************************************************************************

MODULE qs_period_efield_types

   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_p_type
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_period_efield_types'

   PUBLIC :: efield_berry_type, efield_berry_release, init_efield_matrices, &
             set_efield_matrices

   TYPE efield_berry_type
      REAL(KIND=dp)                                         :: field_energy
      REAL(KIND=dp), DIMENSION(3)                           :: polarisation
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER           :: cosmat
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER           :: sinmat
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER           :: dipmat
   END TYPE efield_berry_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param efield ...
! **************************************************************************************************
   SUBROUTINE init_efield_matrices(efield)
      TYPE(efield_berry_type), POINTER                   :: efield

      CHARACTER(len=*), PARAMETER :: routineN = 'init_efield_matrices', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: field_energy
      REAL(KIND=dp), DIMENSION(3)                        :: polarisation

! retain possible values for energy and polarisation

      IF (ASSOCIATED(efield)) THEN
         field_energy = efield%field_energy
         polarisation = efield%polarisation
         CALL efield_berry_release(efield)
      ELSE
         field_energy = 0.0_dp
         polarisation = 0.0_dp
      END IF

      ALLOCATE (efield)
      NULLIFY (efield%cosmat)
      NULLIFY (efield%sinmat)
      NULLIFY (efield%dipmat)

      efield%field_energy = field_energy
      efield%polarisation = polarisation

   END SUBROUTINE init_efield_matrices

! **************************************************************************************************
!> \brief ...
!> \param efield ...
!> \param sinmat ...
!> \param cosmat ...
!> \param dipmat ...
! **************************************************************************************************
   SUBROUTINE set_efield_matrices(efield, sinmat, cosmat, dipmat)

      TYPE(efield_berry_type), POINTER                   :: efield
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: sinmat, cosmat, dipmat

      CHARACTER(len=*), PARAMETER :: routineN = 'set_efield_matrices', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(cosmat)) efield%cosmat => cosmat
      IF (PRESENT(sinmat)) efield%sinmat => sinmat
      IF (PRESENT(dipmat)) efield%dipmat => dipmat

   END SUBROUTINE set_efield_matrices

! **************************************************************************************************
!> \brief ...
!> \param efield ...
! **************************************************************************************************
   SUBROUTINE efield_berry_release(efield)
      TYPE(efield_berry_type), POINTER                   :: efield

      CHARACTER(len=*), PARAMETER :: routineN = 'efield_berry_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(efield)) THEN
         IF (ASSOCIATED(efield%sinmat) .AND. ASSOCIATED(efield%cosmat)) THEN
            CALL cp_dbcsr_deallocate_matrix_set(efield%cosmat)
            CALL cp_dbcsr_deallocate_matrix_set(efield%sinmat)
         END IF
         IF (ASSOCIATED(efield%dipmat)) THEN
            CALL cp_dbcsr_deallocate_matrix_set(efield%dipmat)
         END IF
         DEALLOCATE (efield)
      END IF
   END SUBROUTINE efield_berry_release

END MODULE qs_period_efield_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief New version of the module for the localization of the molecular orbitals
!>      This should be able to use different definition of the spread functional
!>      It should also calculate the integrals analytically so that it can be
!>      used irrespective of the pw_env and the collocation of wfn on the grids
!>      It should also work with a selected set of states, instead than all of them,
!>      in this case one should check that the selected states have the same occupation number
!>      The spread functional can be only estimated, or also optimized by minimization
!>      and in principle also maximization should be available.
!>      This operations can be required irrespective of the printing requirements
!>      It would be highly desirable to do all this along a MD run every N steps,
!>      and have a trajectory of the centeroids of the localized wfn
!>      In addition these functions can be used for properties calculations
!>      like NMR and XAS. Therefore it is necessary that the rotated wfn are then copied
!>      in the mos fm matrix to be available for next use.
!> \author MI (05-2005)
! **************************************************************************************************
MODULE qs_loc_types

   USE cell_types,                      ONLY: cell_release,&
                                              cell_retain,&
                                              cell_type
   USE cp_array_utils_r,                ONLY: cp_2d_r_p_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_p_type
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE distribution_1d_types,           ONLY: distribution_1d_release,&
                                              distribution_1d_retain,&
                                              distribution_1d_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE particle_types,                  ONLY: particle_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters ***

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_loc_types'

!****t* qs_loc_types/qs_loc_env_new_type [1.0] *

! **************************************************************************************************
!> \brief contains all the info needed by quickstep to calculate
!>      the spread of a selectedset of orbitals and if required
!>      to minimize or mazimize the spread by rotation of the orbitals
!> \param para_env info for the distribution of the calculations
!> \param mo_coeff full matrix containing only the selected subset of orbitals
!> \param local_molecules molecules distributed
!> \param cell box that contains the system
!> \param localized_wfn_control variables and parameter that define the spread
!>                             functional and the optimization algorithm
!> \param particle_set position, type, ao_indexes etc for each atom
!> \param op_sm_set set of sparse matrices used to define the spread operator
!>                  when the functional is defined by the use operator acting on the
!>                  basis functions, e.g. the Berry phase definition
!>                  The matrix element of the type <a|O|b> are computed in initialization
!>                  of qs_loc_env
!> \param op_fm_set set of full matrices used to define the spread operator
!>                 when the functional has to be defined directly using the products of MOS
!>                 as in the case of the Pipek-Mezek definition.
!> \param weights for a spread defined as extension of the orbitral in the box, these
!>               factors renormilize with respect to the box size
!> \note
!>      this type should replace the previous set up for the localization of the wfn
!> \par History
!>      04-05 created
!> \author MI
! **************************************************************************************************
   TYPE qs_loc_env_new_type
      INTEGER :: ref_count
      LOGICAL :: molecular_states, do_localize, first_time
      LOGICAL :: wannier_states
      CHARACTER(LEN=default_string_length)        :: tag_mo
      TYPE(cp_para_env_type), POINTER          :: para_env
      TYPE(cp_fm_p_type), DIMENSION(:), &
         POINTER                                :: moloc_coeff
      TYPE(cp_fm_p_type), DIMENSION(:, :), &
         POINTER                                :: op_fm_set
      TYPE(distribution_1d_type), POINTER         :: local_molecules
      TYPE(cell_type), POINTER                 :: cell
      TYPE(localized_wfn_control_type), &
         POINTER                                :: localized_wfn_control
      TYPE(particle_type), DIMENSION(:), &
         POINTER                                :: particle_set
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), &
         POINTER                                :: op_sm_set

      REAL(KIND=dp)                            :: start_time, target_time
      REAL(KIND=dp)                            :: weights(6)
      INTEGER                                     :: dim_op
   END TYPE qs_loc_env_new_type

! **************************************************************************************************
!> \brief A type that holds controling information for the
!>    calculation of the spread of wfn and theoptimization of
!>    the spread functional
!> \param ref_count:
!> \param localization_method which algorithm is used for the optimization
!> \param operator_type how the spread is defined
!> \param nloc_states number of states on which the spread is computed
!> \param set_of_states how to choose the states
!> \param lu_bound_states lower and upper bounds of the set of states
!>    print_cubes:
!>    print_centers:
!>    print_spreads:
!> \param loc_states list of states on which the spread is computed
!> \param centers_set arrais containing centers and spreads of the selected wfn
!> \param centers_file_name output file names
!> \param spreads_file_name output file names
! **************************************************************************************************
   TYPE localized_wfn_control_type
      INTEGER                              :: ref_count
      INTEGER                              :: min_or_max
      INTEGER                              :: localization_method
      INTEGER                              :: operator_type
      INTEGER, DIMENSION(2)                :: nloc_states
      INTEGER                              :: set_of_states
      INTEGER, DIMENSION(2, 2)              :: lu_bound_states
      INTEGER                              :: max_iter
      INTEGER                              :: out_each
      REAL(KIND=dp)                        :: eps_localization
      REAL(KIND=dp)                        :: max_crazy_angle
      REAL(KIND=dp)                        :: crazy_scale
      REAL(KIND=dp)                        :: eps_occ
      REAL(KIND=dp), DIMENSION(2)          :: lu_ene_bound
      LOGICAL                              :: crazy_use_diag
      LOGICAL                              :: print_cubes, jacobi_fallback
      LOGICAL                              :: print_centers
      LOGICAL                              :: print_spreads
      LOGICAL                              :: do_homo
      LOGICAL                              :: loc_restart
      LOGICAL                              :: use_history
      INTEGER, POINTER, DIMENSION(:, :)       :: loc_states
      TYPE(cp_2d_r_p_type), DIMENSION(2)   :: centers_set
   END TYPE localized_wfn_control_type

! *** Public ***
   PUBLIC :: qs_loc_env_create, qs_loc_env_destroy, &
             qs_loc_env_release, qs_loc_env_retain, &
             get_qs_loc_env, set_qs_loc_env, &
             localized_wfn_control_create, localized_wfn_control_release
   PUBLIC :: qs_loc_env_new_type, localized_wfn_control_type

CONTAINS

!****f* qs_loc_types/qs_loc_env_create [1.0] *

! **************************************************************************************************
!> \brief ...
!> \param qs_loc_env ...
!> \par History
!>      04-05 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE qs_loc_env_create(qs_loc_env)

      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_loc_env_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(qs_loc_env)))CALL cp__a("qs_loc_types.F",174)

      ALLOCATE (qs_loc_env)

      qs_loc_env%ref_count = 1
      qs_loc_env%tag_mo = ""
      NULLIFY (qs_loc_env%para_env)
      NULLIFY (qs_loc_env%cell)
      NULLIFY (qs_loc_env%op_sm_set)
      NULLIFY (qs_loc_env%op_fm_set)
      NULLIFY (qs_loc_env%local_molecules)
      NULLIFY (qs_loc_env%moloc_coeff)
      NULLIFY (qs_loc_env%particle_set)
      NULLIFY (qs_loc_env%localized_wfn_control)
      qs_loc_env%weights = 0.0_dp

   END SUBROUTINE qs_loc_env_create

!****f* qs_loc_types/qs_loc_env_destroy [1.0] *

! **************************************************************************************************
!> \brief ...
!> \param qs_loc_env ...
!> \par History
!>      04-05 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE qs_loc_env_destroy(qs_loc_env)

      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_loc_env_destroy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      IF(.NOT.(ASSOCIATED(qs_loc_env)))CALL cp__a("qs_loc_types.F",210)

      IF (ASSOCIATED(qs_loc_env%cell)) CALL cell_release(qs_loc_env%cell)
      IF (ASSOCIATED(qs_loc_env%local_molecules)) &
         CALL distribution_1d_release(qs_loc_env%local_molecules)
      IF (ASSOCIATED(qs_loc_env%localized_wfn_control)) THEN
         CALL localized_wfn_control_release(qs_loc_env%localized_wfn_control)
      END IF
      IF (ASSOCIATED(qs_loc_env%para_env)) CALL cp_para_env_release(qs_loc_env%para_env)
      IF (ASSOCIATED(qs_loc_env%particle_set)) NULLIFY (qs_loc_env%particle_set)

      IF (ASSOCIATED(qs_loc_env%moloc_coeff)) THEN
         DO i = 1, SIZE(qs_loc_env%moloc_coeff, 1)
            CALL cp_fm_release(qs_loc_env%moloc_coeff(i)%matrix)
         END DO
         DEALLOCATE (qs_loc_env%moloc_coeff)
      END IF

      IF (ASSOCIATED(qs_loc_env%op_fm_set)) THEN
         DO i = 1, SIZE(qs_loc_env%op_fm_set, 2)
            DO j = 1, SIZE(qs_loc_env%op_fm_set, 1)
               CALL cp_fm_release(qs_loc_env%op_fm_set(j, i)%matrix)
            END DO
         END DO
         DEALLOCATE (qs_loc_env%op_fm_set)
      END IF

      IF (ASSOCIATED(qs_loc_env%op_sm_set)) THEN
         DO i = 1, SIZE(qs_loc_env%op_sm_set, 2)
            DO j = 1, SIZE(qs_loc_env%op_sm_set, 1)
               CALL cp_dbcsr_deallocate_matrix(qs_loc_env%op_sm_set(j, i)%matrix)
            ENDDO
         END DO
         DEALLOCATE (qs_loc_env%op_sm_set)
      END IF

      DEALLOCATE (qs_loc_env)

   END SUBROUTINE qs_loc_env_destroy

!****f* qs_loc_types/qs_loc_env_release [1.0] *

! **************************************************************************************************
!> \brief ...
!> \param qs_loc_env ...
!> \par History
!>      04-05 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE qs_loc_env_release(qs_loc_env)

      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_loc_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(qs_loc_env)) THEN
         IF(.NOT.(qs_loc_env%ref_count > 0))CALL cp__a("qs_loc_types.F",267)
         qs_loc_env%ref_count = qs_loc_env%ref_count-1
         IF (qs_loc_env%ref_count == 0) THEN
            CALL qs_loc_env_destroy(qs_loc_env)
         END IF
      END IF
   END SUBROUTINE qs_loc_env_release

!****f* qs_loc_types/qs_loc_env_retain [1.0] *

! **************************************************************************************************
!> \brief ...
!> \param qs_loc_env ...
!> \par History
!>      04-05 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE qs_loc_env_retain(qs_loc_env)

      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_loc_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(qs_loc_env)))CALL cp__a("qs_loc_types.F",291)
      IF(.NOT.(qs_loc_env%ref_count > 0))CALL cp__a("qs_loc_types.F",292)
      qs_loc_env%ref_count = qs_loc_env%ref_count+1
   END SUBROUTINE qs_loc_env_retain

! **************************************************************************************************
!> \brief create the localized_wfn_control_type
!> \param localized_wfn_control ...
!> \par History
!>      04.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE localized_wfn_control_create(localized_wfn_control)
      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control

      CHARACTER(len=*), PARAMETER :: routineN = 'localized_wfn_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(localized_wfn_control)))CALL cp__a("qs_loc_types.F",308)
      ALLOCATE (localized_wfn_control)

      localized_wfn_control%ref_count = 1
      localized_wfn_control%nloc_states = 0
      localized_wfn_control%lu_bound_states = 0
      localized_wfn_control%lu_ene_bound = 0.0_dp
      localized_wfn_control%print_cubes = .FALSE.
      localized_wfn_control%print_centers = .FALSE.
      localized_wfn_control%print_spreads = .FALSE.
      localized_wfn_control%do_homo = .TRUE.
      localized_wfn_control%use_history = .FALSE.
      NULLIFY (localized_wfn_control%loc_states)
      NULLIFY (localized_wfn_control%centers_set(1)%array)
      NULLIFY (localized_wfn_control%centers_set(2)%array)
   END SUBROUTINE localized_wfn_control_create

! **************************************************************************************************
!> \brief release the localized_wfn_control_type
!> \param localized_wfn_control ...
!> \par History
!>      04.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE localized_wfn_control_release(localized_wfn_control)

      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control

      CHARACTER(len=*), PARAMETER :: routineN = 'localized_wfn_control_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(localized_wfn_control)) THEN
         IF(.NOT.(localized_wfn_control%ref_count > 0))CALL cp__a("qs_loc_types.F",339)
         localized_wfn_control%ref_count = localized_wfn_control%ref_count-1
         IF (localized_wfn_control%ref_count == 0) THEN
            IF (ASSOCIATED(localized_wfn_control%loc_states)) THEN
               DEALLOCATE (localized_wfn_control%loc_states)
            ENDIF
            IF (ASSOCIATED(localized_wfn_control%centers_set(1)%array)) THEN
               DEALLOCATE (localized_wfn_control%centers_set(1)%array)
            ENDIF
            IF (ASSOCIATED(localized_wfn_control%centers_set(2)%array)) THEN
               DEALLOCATE (localized_wfn_control%centers_set(2)%array)
            ENDIF
            localized_wfn_control%ref_count = 0
            DEALLOCATE (localized_wfn_control)
         ENDIF
      END IF
   END SUBROUTINE localized_wfn_control_release

! **************************************************************************************************
!> \brief retain the localized_wfn_control_type
!> \param localized_wfn_control ...
!> \par History
!>      04.2005 created [MI]
! **************************************************************************************************
   SUBROUTINE localized_wfn_control_retain(localized_wfn_control)
      TYPE(localized_wfn_control_type), POINTER          :: localized_wfn_control

      CHARACTER(len=*), PARAMETER :: routineN = 'localized_wfn_control_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(localized_wfn_control)))CALL cp__a("qs_loc_types.F",369)

      localized_wfn_control%ref_count = localized_wfn_control%ref_count+1
   END SUBROUTINE localized_wfn_control_retain

!****f* qs_loc_types/get_qs_loc_env [1.0] *

! **************************************************************************************************
!> \brief ...
!> \param qs_loc_env ...
!> \param cell ...
!> \param local_molecules ...
!> \param localized_wfn_control ...
!> \param moloc_coeff ...
!> \param op_sm_set ...
!> \param op_fm_set ...
!> \param para_env ...
!> \param particle_set ...
!> \param weights ...
!> \param dim_op ...
!> \par History
!>      04-05 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE get_qs_loc_env(qs_loc_env, cell, local_molecules, localized_wfn_control, &
                             moloc_coeff, op_sm_set, op_fm_set, para_env, particle_set, weights, dim_op)

      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_molecules
      TYPE(localized_wfn_control_type), OPTIONAL, &
         POINTER                                         :: localized_wfn_control
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: moloc_coeff
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: op_sm_set
      TYPE(cp_fm_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: op_fm_set
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      REAL(dp), DIMENSION(6), OPTIONAL                   :: weights
      INTEGER, OPTIONAL                                  :: dim_op

      CHARACTER(len=*), PARAMETER :: routineN = 'get_qs_loc_env', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(qs_loc_env)))CALL cp__a("qs_loc_types.F",415)

      IF (PRESENT(cell)) cell => qs_loc_env%cell
      IF (PRESENT(moloc_coeff)) moloc_coeff => qs_loc_env%moloc_coeff
      IF (PRESENT(local_molecules)) local_molecules => qs_loc_env%local_molecules
      IF (PRESENT(localized_wfn_control)) &
         localized_wfn_control => qs_loc_env%localized_wfn_control
      IF (PRESENT(op_sm_set)) op_sm_set => qs_loc_env%op_sm_set
      IF (PRESENT(op_fm_set)) op_fm_set => qs_loc_env%op_fm_set
      IF (PRESENT(para_env)) para_env => qs_loc_env%para_env
      IF (PRESENT(particle_set)) particle_set => qs_loc_env%particle_set
      IF (PRESENT(weights)) weights(1:6) = qs_loc_env%weights(1:6)
      IF (PRESENT(dim_op)) dim_op = qs_loc_env%dim_op

   END SUBROUTINE get_qs_loc_env

!****f* qs_loc_types/set_qs_loc_env [1.0] *

! **************************************************************************************************
!> \brief ...
!> \param qs_loc_env ...
!> \param cell ...
!> \param local_molecules ...
!> \param localized_wfn_control ...
!> \param moloc_coeff ...
!> \param op_sm_set ...
!> \param op_fm_set ...
!> \param para_env ...
!> \param particle_set ...
!> \param weights ...
!> \param dim_op ...
!> \par History
!>      04-05 created
!> \author MI
! **************************************************************************************************
   SUBROUTINE set_qs_loc_env(qs_loc_env, cell, local_molecules, localized_wfn_control, &
                             moloc_coeff, op_sm_set, op_fm_set, para_env, particle_set, weights, dim_op)

      TYPE(qs_loc_env_new_type), POINTER                 :: qs_loc_env
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell
      TYPE(distribution_1d_type), OPTIONAL, POINTER      :: local_molecules
      TYPE(localized_wfn_control_type), OPTIONAL, &
         POINTER                                         :: localized_wfn_control
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: moloc_coeff
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: op_sm_set
      TYPE(cp_fm_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: op_fm_set
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      TYPE(particle_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: particle_set
      REAL(dp), DIMENSION(6), OPTIONAL                   :: weights
      INTEGER, OPTIONAL                                  :: dim_op

      CHARACTER(len=*), PARAMETER :: routineN = 'set_qs_loc_env', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF(.NOT.(ASSOCIATED(qs_loc_env)))CALL cp__a("qs_loc_types.F",474)
      IF (PRESENT(cell)) THEN
         CALL cell_retain(cell)
         CALL cell_release(qs_loc_env%cell)
         qs_loc_env%cell => cell
      END IF

      IF (PRESENT(local_molecules)) THEN
         CALL distribution_1d_retain(local_molecules)
         IF (ASSOCIATED(qs_loc_env%local_molecules)) &
            CALL distribution_1d_release(qs_loc_env%local_molecules)
         qs_loc_env%local_molecules => local_molecules
      END IF

      IF (PRESENT(localized_wfn_control)) THEN
         CALL localized_wfn_control_retain(localized_wfn_control)
         CALL localized_wfn_control_release(qs_loc_env%localized_wfn_control)
         qs_loc_env%localized_wfn_control => localized_wfn_control
      END IF
      IF (PRESENT(para_env)) THEN
         CALL cp_para_env_retain(para_env)
         CALL cp_para_env_release(qs_loc_env%para_env)
         qs_loc_env%para_env => para_env
      END IF
      IF (PRESENT(particle_set)) qs_loc_env%particle_set => particle_set
      IF (PRESENT(moloc_coeff)) THEN
         IF (ASSOCIATED(qs_loc_env%moloc_coeff)) THEN
            DO i = 1, SIZE(qs_loc_env%moloc_coeff, 1)
               CALL cp_fm_release(qs_loc_env%moloc_coeff(i)%matrix)
            END DO
            DEALLOCATE (qs_loc_env%moloc_coeff)
         END IF
         qs_loc_env%moloc_coeff => moloc_coeff
      END IF
      IF (PRESENT(op_sm_set)) THEN
         qs_loc_env%op_sm_set => op_sm_set
      END IF
      IF (PRESENT(op_fm_set)) THEN
         qs_loc_env%op_fm_set => op_fm_set
      END IF
      IF (PRESENT(weights)) THEN
         qs_loc_env%weights = weights
      END IF
      IF (PRESENT(dim_op)) THEN
         qs_loc_env%dim_op = dim_op
      END IF

   END SUBROUTINE set_qs_loc_env

END MODULE qs_loc_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Type definitiona for linear response calculations
!> \author MI
! **************************************************************************************************
MODULE qs_linres_types
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind,&
                                              get_atomic_kind_set
   USE basis_set_types,                 ONLY: get_gto_basis_set,&
                                              gto_basis_set_type
   USE cp_array_utils_i,                ONLY: cp_2d_i_p_type
   USE cp_array_utils_r,                ONLY: cp_2d_r_p_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_fm_types,                     ONLY: cp_fm_p_type
   USE kinds,                           ONLY: dp
   USE qs_grid_atom,                    ONLY: grid_atom_type
   USE qs_harmonics_atom,               ONLY: harmonics_atom_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE qs_loc_types,                    ONLY: qs_loc_env_new_type,&
                                              qs_loc_env_release
   USE qs_rho_atom_types,               ONLY: rho_atom_coeff,&
                                              rho_atom_type
   USE qs_rho_types,                    ONLY: qs_rho_p_type,&
                                              qs_rho_release
   USE realspace_grid_types,            ONLY: realspace_grid_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

!****s* qs_linres_types/linres_control_type

! **************************************************************************************************
!> \brief General settings for linear response calculations
!> \param property which quantity is to be calculated by LR
!> \param opt_method method to optimize the psi1 by minimization of the second order term of the energy
!> \param preconditioner which kind of preconditioner should be used, if any
!> \param localized_psi 0 : don't use the canonical psi0, but the maximally localized wavefunctions
!> \param do_kernel the kernel is zero if the rho1 is zero as for the magnetic field perturbation
!> \param tolerance convergence criterium for the optimization of the psi1
!> \author MI
! **************************************************************************************************
   TYPE linres_control_type
      INTEGER                                   :: ref_count
      INTEGER                                   :: property
      INTEGER                                   :: preconditioner_type
      INTEGER                                   :: restart_every
      REAL(dp)                                  :: energy_gap
      INTEGER                                   :: max_iter
      LOGICAL                                   :: localized_psi0
      LOGICAL                                   :: do_kernel
      LOGICAL                                   :: converged
      LOGICAL                                   :: linres_restart
      LOGICAL                                   :: lr_triplet
      REAL(KIND=dp)                             :: eps
      TYPE(qs_loc_env_new_type), POINTER        :: qs_loc_env
      CHARACTER(LEN=8)                          :: flag
   END TYPE linres_control_type

!****s* qs_linres_types/current_env_type

! **************************************************************************************************
!> \param ref_coun t
!> \param full_nmr true if the full correction is calculated
!> \param simplenmr_done , fullnmr_done : flags that indicate what has been
!>                    already calculated: used for restart
!> \param centers_set centers of the maximally localized psi0
!> \param spreads_set spreads of the maximally localized psi0
!> \param p_psi 0      : full matrixes, operator p applied to psi0
!> \param rxp_psi 0    : full matrixes, operator (r-d)xp applied to psi0
!> \param psi 1_p      : response wavefunctions to the perturbation given by
!>                    H1=p (xyz)  applied to psi0
!> \param psi 1_rxp    : response wavefunctions to the perturbation given by
!>                    H1=(r-d_i)xp applied to psi0_i where d_i is the center
!> \param psi 1_D      : response wavefunctions to the perturbation given by
!>                    H1=(d_j-d_i)xp applied to psi0_i where d_i is the center
!>                    and d_j is the center of psi0_j and psi1_D_j is the result
!>                    This operator has to be used in nstate scf calculations,
!>                    one for each psi1_D_j vector
!> \param chemical_shift the tensor for each atom
!> \param chi_tensor the susceptibility tensor
!> \param jrho 1_set   : current density on the global grid, if gapw this is only the soft part
!> \param jrho 1_atom_set : current density on the local atomic grids (only if gapw)
!> \author MI
! **************************************************************************************************
   TYPE realspaces_grid_p_type
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs
   END TYPE realspaces_grid_p_type

   TYPE current_env_type
      LOGICAL                                     :: full, simple_done(6), simple_converged(6), do_qmmm
      LOGICAL                                     :: use_old_gauge_atom, chi_pbc, do_selected_states, &
                                                     gauge_init, all_pert_op_done
      LOGICAL, DIMENSION(:, :), POINTER            :: full_done
      INTEGER                                     :: ref_count, nao, nstates(2), gauge, orb_center, nbr_center(2)
      INTEGER, DIMENSION(:), POINTER          :: list_cubes, selected_states_on_atom_list
      INTEGER, DIMENSION(:, :, :), POINTER          :: statetrueindex
      CHARACTER(LEN=30)                           :: gauge_name, orb_center_name
      REAL(dp)                                    :: chi_tensor(3, 3, 2), chi_tensor_loc(3, 3, 2), gauge_atom_radius
      REAL(dp)                                    :: selected_states_atom_radius
      REAL(dp), DIMENSION(:, :), POINTER           :: basisfun_center
      TYPE(cp_2d_i_p_type), DIMENSION(:), POINTER :: center_list
      TYPE(cp_2d_r_p_type), DIMENSION(:), POINTER :: centers_set
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER :: psi1_p, psi1_rxp, psi1_D, p_psi0, rxp_psi0
      TYPE(jrho_atom_type), DIMENSION(:), POINTER :: jrho1_atom_set
      TYPE(qs_rho_p_type), DIMENSION(:), POINTER :: jrho1_set
      TYPE(realspace_grid_p_type), DIMENSION(:), POINTER :: rs_buf
      TYPE(realspaces_grid_p_type), DIMENSION(:), POINTER :: rs_gauge
      !
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER :: psi0_order
   END TYPE current_env_type

!*********************************************************************************************************
! \param type for polarizability calculation using Berry operator

   TYPE polar_env_type
      INTEGER                             :: ref_count
      LOGICAL                             :: do_raman
      REAL(dp), DIMENSION(:, :), POINTER :: polar
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER :: psi1_dBerry, dBerry_psi0, mo_derivs
   END TYPE polar_env_type
! *********************************************************************************************************

   TYPE issc_env_type
      INTEGER                             :: ref_count
      INTEGER                             :: issc_natms
      INTEGER, DIMENSION(:), POINTER      :: issc_on_atom_list
      LOGICAL                             :: interpolate_issc
      LOGICAL                             :: do_fc, do_sd, do_pso, do_dso
      REAL(dp)                            :: issc_gapw_radius, issc_factor, issc_factor_gapw
      REAL(dp), DIMENSION(:, :, :, :, :), POINTER :: issc, issc_loc
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER :: psi1_efg, psi1_pso, efg_psi0, pso_psi0, dso_psi0, psi1_dso !last two not needed
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER :: psi1_fc
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER :: fc_psi0
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: matrix_efg, matrix_pso, matrix_dso, matrix_fc
   END TYPE issc_env_type

! *********************************************************************************************************
   TYPE nmr_env_type
      INTEGER                             :: ref_count, n_nics
      INTEGER, DIMENSION(:), POINTER      :: cs_atom_list
      INTEGER, DIMENSION(:), POINTER      :: do_calc_cs_atom
      LOGICAL                             :: do_nics, interpolate_shift
      REAL(dp)                            :: shift_gapw_radius, shift_factor, shift_factor_gapw, chi_factor, &
                                             chi_SI2shiftppm, chi_SI2ppmcgs
      REAL(dp), DIMENSION(:, :), POINTER :: r_nics
      REAL(dp), DIMENSION(:, :, :), POINTER :: chemical_shift, chemical_shift_loc, &
                                               chemical_shift_nics_loc, chemical_shift_nics
   END TYPE nmr_env_type

! **************************************************************************************************
   TYPE epr_env_type
      INTEGER                                     :: ref_count
      REAL(dp)                                    :: g_free_factor, g_soo_chicorr_factor, g_soo_factor, &
                                                     g_so_factor, g_so_factor_gapw, g_zke_factor, g_zke
      REAL(dp), DIMENSION(:, :), POINTER           :: g_total, g_so, g_soo
      TYPE(qs_rho_p_type), DIMENSION(:, :), POINTER :: nablavks_set
      TYPE(nablavks_atom_type), DIMENSION(:), POINTER :: nablavks_atom_set
      TYPE(qs_rho_p_type), DIMENSION(:, :), POINTER :: bind_set
      TYPE(rho_atom_coeff), DIMENSION(:, :), POINTER :: bind_atom_set
      TYPE(rho_atom_type), DIMENSION(:), POINTER :: vks_atom_set
   END TYPE epr_env_type

! **************************************************************************************************
   TYPE nablavks_atom_type
      TYPE(rho_atom_coeff), DIMENSION(:, :), &
         POINTER                                :: nablavks_vec_rad_h, &
                                                   nablavks_vec_rad_s
   END TYPE nablavks_atom_type

! **************************************************************************************************
   TYPE jrho_atom_p_type
      TYPE(jrho_atom_type), POINTER       :: jrho_atom
   END TYPE jrho_atom_p_type

! **************************************************************************************************
   TYPE jrho_atom_type
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER      :: cjc_h, cjc_s, cjc0_h, cjc0_s
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER      :: cjc_ii_h, cjc_ii_s
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER      :: cjc_iii_h, cjc_iii_s
      TYPE(rho_atom_coeff), DIMENSION(:, :), &
         POINTER      :: jrho_vec_rad_h, &
                         jrho_vec_rad_s
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER                               :: jrho_h, jrho_s
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER                               :: jrho_a_h, jrho_a_s
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER                               :: jrho_b_h, jrho_b_s
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER                               :: jrho_a_h_ii, jrho_a_s_ii
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER                               :: jrho_b_h_ii, jrho_b_s_ii
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER                               :: jrho_a_h_iii, jrho_a_s_iii
      TYPE(rho_atom_coeff), DIMENSION(:), &
         POINTER                               :: jrho_b_h_iii, jrho_b_s_iii
   END TYPE jrho_atom_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_linres_types'

! *** Public data types ***

   PUBLIC :: linres_control_type, &
             nmr_env_type, issc_env_type, jrho_atom_type, &
             epr_env_type, &
             nablavks_atom_type, current_env_type, &
             realspaces_grid_p_type, polar_env_type

! *** Public subroutines ***

   PUBLIC :: allocate_jrho_atom_rad, deallocate_jrho_atom_set, get_nmr_env, &
             get_current_env, allocate_jrho_coeff, &
             init_jrho_atom_set, init_nablavks_atom_set, linres_control_create, &
             linres_control_retain, linres_control_release, &
             set_epr_env, deallocate_nablavks_atom_set, &
             set2zero_jrho_atom_rad, get_epr_env, &
             nmr_env_create, epr_env_create, current_env_create, get_issc_env, &
             set_current_env, issc_env_create, get_polar_env, &
             polar_env_create

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param linres_control ...
! **************************************************************************************************
   SUBROUTINE linres_control_create(linres_control)

      TYPE(linres_control_type), POINTER                 :: linres_control

      CHARACTER(len=*), PARAMETER :: routineN = 'linres_control_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(linres_control)))CALL cp__a("qs_linres_types.F",245)
      ALLOCATE (linres_control)
      linres_control%ref_count = 1
      NULLIFY (linres_control%qs_loc_env)
      linres_control%property = HUGE(0) !is that used?
      linres_control%preconditioner_type = HUGE(0)
      linres_control%restart_every = HUGE(0)
      linres_control%energy_gap = HUGE(0.0_dp)
      linres_control%max_iter = HUGE(0)
      linres_control%localized_psi0 = .FALSE.
      linres_control%converged = .FALSE.
      linres_control%linres_restart = .FALSE.
      linres_control%eps = HUGE(0.0_dp)
      linres_control%flag = ""
      linres_control%do_kernel = .FALSE.
      linres_control%lr_triplet = .FALSE.

   END SUBROUTINE linres_control_create

! **************************************************************************************************
!> \brief ...
!> \param linres_control ...
! **************************************************************************************************
   SUBROUTINE linres_control_release(linres_control)

      TYPE(linres_control_type), POINTER                 :: linres_control

      CHARACTER(len=*), PARAMETER :: routineN = 'linres_control_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(linres_control)) THEN
         IF(.NOT.(linres_control%ref_count > 0))CALL cp__a("qs_linres_types.F",276)
         linres_control%ref_count = linres_control%ref_count-1
         IF (linres_control%ref_count < 1) THEN
            IF (ASSOCIATED(linres_control%qs_loc_env)) THEN
               CALL qs_loc_env_release(linres_control%qs_loc_env)
            END IF
            DEALLOCATE (linres_control)
         END IF
      END IF
      NULLIFY (linres_control)
   END SUBROUTINE linres_control_release

! **************************************************************************************************
!> \brief ...
!> \param linres_control ...
! **************************************************************************************************
   SUBROUTINE linres_control_retain(linres_control)

      TYPE(linres_control_type), POINTER                 :: linres_control

      CHARACTER(len=*), PARAMETER :: routineN = 'linres_control_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(linres_control)))CALL cp__a("qs_linres_types.F",299)
      IF(.NOT.(linres_control%ref_count > 0))CALL cp__a("qs_linres_types.F",300)
      linres_control%ref_count = linres_control%ref_count+1

   END SUBROUTINE linres_control_retain

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
! **************************************************************************************************
   SUBROUTINE current_env_create(current_env)

      TYPE(current_env_type)                             :: current_env

      CHARACTER(len=*), PARAMETER :: routineN = 'current_env_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(current_env%ref_count == 0))CALL cp__a("qs_linres_types.F",316)
      current_env%ref_count = 1
      current_env%nao = HUGE(1)
      current_env%gauge = HUGE(1)
      current_env%orb_center = HUGE(1)
      current_env%nstates(:) = HUGE(1)
      current_env%nbr_center(:) = HUGE(1)
      current_env%use_old_gauge_atom = .TRUE.
      current_env%chi_pbc = .FALSE.
      current_env%do_selected_states = .FALSE.
      current_env%gauge_init = .FALSE.
      NULLIFY (current_env%full_done)
      NULLIFY (current_env%list_cubes)
      NULLIFY (current_env%statetrueindex)
      NULLIFY (current_env%basisfun_center)
      NULLIFY (current_env%center_list)
      NULLIFY (current_env%centers_set)
      NULLIFY (current_env%psi1_p)
      NULLIFY (current_env%psi1_rxp)
      NULLIFY (current_env%psi1_D)
      NULLIFY (current_env%p_psi0)
      NULLIFY (current_env%rxp_psi0)
      NULLIFY (current_env%jrho1_atom_set)
      NULLIFY (current_env%jrho1_set)
      NULLIFY (current_env%rs_gauge)
      NULLIFY (current_env%rs_buf)
      NULLIFY (current_env%selected_states_on_atom_list)
      NULLIFY (current_env%psi0_order)

   END SUBROUTINE current_env_create
! **************************************************************************************************
!> \brief ...
!> \param nmr_env ...
! **************************************************************************************************
   SUBROUTINE nmr_env_create(nmr_env)

      TYPE(nmr_env_type)                                 :: nmr_env

      CHARACTER(len=*), PARAMETER :: routineN = 'nmr_env_create', routineP = moduleN//':'//routineN

      IF(.NOT.(nmr_env%ref_count == 0))CALL cp__a("qs_linres_types.F",356)
      nmr_env%ref_count = 1
      NULLIFY (nmr_env%chemical_shift)
      NULLIFY (nmr_env%chemical_shift_loc)
      NULLIFY (nmr_env%chemical_shift_nics_loc)
      NULLIFY (nmr_env%chemical_shift_nics)
      NULLIFY (nmr_env%r_nics)
      NULLIFY (nmr_env%cs_atom_list)
      NULLIFY (nmr_env%do_calc_cs_atom)

   END SUBROUTINE nmr_env_create

! **************************************************************************************************
!> \brief ...
!> \param issc_env ...
! **************************************************************************************************
   SUBROUTINE issc_env_create(issc_env)

      TYPE(issc_env_type)                                :: issc_env

      CHARACTER(len=*), PARAMETER :: routineN = 'issc_env_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(issc_env%ref_count == 0))CALL cp__a("qs_linres_types.F",379)
      issc_env%ref_count = 1
      NULLIFY (issc_env%issc)
      NULLIFY (issc_env%issc_loc)
      NULLIFY (issc_env%psi1_efg)
      NULLIFY (issc_env%psi1_fc)
      NULLIFY (issc_env%psi1_pso)
      NULLIFY (issc_env%psi1_dso)
      NULLIFY (issc_env%efg_psi0)
      NULLIFY (issc_env%pso_psi0)
      NULLIFY (issc_env%dso_psi0)
      NULLIFY (issc_env%fc_psi0)
      NULLIFY (issc_env%matrix_efg)
      NULLIFY (issc_env%matrix_pso)
      NULLIFY (issc_env%matrix_dso)
      NULLIFY (issc_env%matrix_fc)

   END SUBROUTINE issc_env_create

! **************************************************************************************************
!> \brief ...
!> \param epr_env ...
! **************************************************************************************************
   SUBROUTINE epr_env_create(epr_env)

      TYPE(epr_env_type)                                 :: epr_env

      CHARACTER(len=*), PARAMETER :: routineN = 'epr_env_create', routineP = moduleN//':'//routineN

      IF(.NOT.(epr_env%ref_count == 0))CALL cp__a("qs_linres_types.F",408)
      epr_env%ref_count = 1
      NULLIFY (epr_env%nablavks_set)
      NULLIFY (epr_env%nablavks_atom_set)
      NULLIFY (epr_env%bind_set)
      NULLIFY (epr_env%bind_atom_set)
      NULLIFY (epr_env%g_total)
      NULLIFY (epr_env%g_so)
      NULLIFY (epr_env%g_soo)
      NULLIFY (epr_env%vks_atom_set)

   END SUBROUTINE epr_env_create

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param simple_done ...
!> \param simple_converged ...
!> \param full_done ...
!> \param ref_count ...
!> \param nao ...
!> \param nstates ...
!> \param gauge ...
!> \param list_cubes ...
!> \param statetrueindex ...
!> \param gauge_name ...
!> \param basisfun_center ...
!> \param nbr_center ...
!> \param center_list ...
!> \param centers_set ...
!> \param psi1_p ...
!> \param psi1_rxp ...
!> \param psi1_D ...
!> \param p_psi0 ...
!> \param rxp_psi0 ...
!> \param jrho1_atom_set ...
!> \param jrho1_set ...
!> \param chi_tensor ...
!> \param chi_tensor_loc ...
!> \param gauge_atom_radius ...
!> \param rs_gauge ...
!> \param use_old_gauge_atom ...
!> \param chi_pbc ...
!> \param psi0_order ...
! **************************************************************************************************
   SUBROUTINE get_current_env(current_env, simple_done, simple_converged, full_done, ref_count, nao, &
                              nstates, gauge, list_cubes, statetrueindex, gauge_name, basisfun_center, &
                              nbr_center, center_list, centers_set, psi1_p, psi1_rxp, psi1_D, p_psi0, &
                              rxp_psi0, jrho1_atom_set, jrho1_set, chi_tensor, &
                              chi_tensor_loc, gauge_atom_radius, rs_gauge, use_old_gauge_atom, &
                              chi_pbc, psi0_order)

      TYPE(current_env_type), OPTIONAL                   :: current_env
      LOGICAL, OPTIONAL                                  :: simple_done(6), simple_converged(6)
      LOGICAL, DIMENSION(:, :), OPTIONAL, POINTER        :: full_done
      INTEGER, OPTIONAL                                  :: ref_count, nao, nstates(2), gauge
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: list_cubes
      INTEGER, DIMENSION(:, :, :), OPTIONAL, POINTER     :: statetrueindex
      CHARACTER(LEN=30), OPTIONAL                        :: gauge_name
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: basisfun_center
      INTEGER, OPTIONAL                                  :: nbr_center(2)
      TYPE(cp_2d_i_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: center_list
      TYPE(cp_2d_r_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: centers_set
      TYPE(cp_fm_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: psi1_p, psi1_rxp, psi1_D, p_psi0, &
                                                            rxp_psi0
      TYPE(jrho_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: jrho1_atom_set
      TYPE(qs_rho_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: jrho1_set
      REAL(dp), INTENT(OUT), OPTIONAL                    :: chi_tensor(3, 3, 2), &
                                                            chi_tensor_loc(3, 3, 2), &
                                                            gauge_atom_radius
      TYPE(realspaces_grid_p_type), DIMENSION(:), &
         OPTIONAL, POINTER                               :: rs_gauge
      LOGICAL, OPTIONAL                                  :: use_old_gauge_atom, chi_pbc
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: psi0_order

      CHARACTER(len=*), PARAMETER :: routineN = 'get_current_env', &
         routineP = moduleN//':'//routineN

!
!

      IF(.NOT.(current_env%ref_count > 0))CALL cp__a("qs_linres_types.F",495)

      IF (PRESENT(simple_done)) simple_done(1:6) = current_env%simple_done(1:6)
      IF (PRESENT(simple_converged)) simple_converged(1:6) = current_env%simple_converged(1:6)
      IF (PRESENT(full_done)) full_done => current_env%full_done
      IF (PRESENT(ref_count)) ref_count = current_env%ref_count
      IF (PRESENT(nao)) nao = current_env%nao
      IF (PRESENT(nstates)) nstates(1:2) = current_env%nstates(1:2)
      IF (PRESENT(gauge)) gauge = current_env%gauge
      IF (PRESENT(list_cubes)) list_cubes => current_env%list_cubes
      IF (PRESENT(statetrueindex)) statetrueindex => current_env%statetrueindex
      IF (PRESENT(gauge_name)) gauge_name = current_env%gauge_name
      IF (PRESENT(basisfun_center)) basisfun_center => current_env%basisfun_center
      IF (PRESENT(nbr_center)) nbr_center(1:2) = current_env%nbr_center(1:2)
      IF (PRESENT(center_list)) center_list => current_env%center_list
      IF (PRESENT(centers_set)) centers_set => current_env%centers_set
      IF (PRESENT(chi_tensor)) chi_tensor(:, :, :) = current_env%chi_tensor(:, :, :)
      IF (PRESENT(chi_tensor_loc)) chi_tensor_loc(:, :, :) = current_env%chi_tensor_loc(:, :, :)
      IF (PRESENT(psi1_p)) psi1_p => current_env%psi1_p
      IF (PRESENT(psi1_rxp)) psi1_rxp => current_env%psi1_rxp
      IF (PRESENT(psi1_D)) psi1_D => current_env%psi1_D
      IF (PRESENT(p_psi0)) p_psi0 => current_env%p_psi0
      IF (PRESENT(rxp_psi0)) rxp_psi0 => current_env%rxp_psi0
      IF (PRESENT(jrho1_atom_set)) jrho1_atom_set => current_env%jrho1_atom_set
      IF (PRESENT(jrho1_set)) jrho1_set => current_env%jrho1_set
      IF (PRESENT(rs_gauge)) rs_gauge => current_env%rs_gauge
      IF (PRESENT(psi0_order)) psi0_order => current_env%psi0_order
      IF (PRESENT(chi_pbc)) chi_pbc = current_env%chi_pbc
      IF (PRESENT(gauge_atom_radius)) gauge_atom_radius = current_env%gauge_atom_radius
      IF (PRESENT(use_old_gauge_atom)) use_old_gauge_atom = current_env%use_old_gauge_atom

   END SUBROUTINE get_current_env

! **************************************************************************************************
!> \brief ...
!> \param nmr_env ...
!> \param n_nics ...
!> \param cs_atom_list ...
!> \param do_calc_cs_atom ...
!> \param r_nics ...
!> \param chemical_shift ...
!> \param chemical_shift_loc ...
!> \param chemical_shift_nics_loc ...
!> \param chemical_shift_nics ...
!> \param shift_gapw_radius ...
!> \param do_nics ...
!> \param interpolate_shift ...
! **************************************************************************************************
   SUBROUTINE get_nmr_env(nmr_env, n_nics, cs_atom_list, do_calc_cs_atom, &
                          r_nics, chemical_shift, chemical_shift_loc, &
                          chemical_shift_nics_loc, chemical_shift_nics, &
                          shift_gapw_radius, do_nics, interpolate_shift)

      TYPE(nmr_env_type)                                 :: nmr_env
      INTEGER, INTENT(OUT), OPTIONAL                     :: n_nics
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: cs_atom_list, do_calc_cs_atom
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: r_nics
      REAL(dp), DIMENSION(:, :, :), OPTIONAL, POINTER    :: chemical_shift, chemical_shift_loc, &
                                                            chemical_shift_nics_loc, &
                                                            chemical_shift_nics
      REAL(dp), INTENT(OUT), OPTIONAL                    :: shift_gapw_radius
      LOGICAL, INTENT(OUT), OPTIONAL                     :: do_nics, interpolate_shift

      CHARACTER(len=*), PARAMETER :: routineN = 'get_nmr_env', routineP = moduleN//':'//routineN

      IF(.NOT.(nmr_env%ref_count > 0))CALL cp__a("qs_linres_types.F",560)

      IF (PRESENT(n_nics)) n_nics = nmr_env%n_nics
      IF (PRESENT(cs_atom_list)) cs_atom_list => nmr_env%cs_atom_list
      IF (PRESENT(do_calc_cs_atom)) do_calc_cs_atom => nmr_env%do_calc_cs_atom
      IF (PRESENT(chemical_shift)) chemical_shift => nmr_env%chemical_shift
      IF (PRESENT(chemical_shift_loc)) chemical_shift_loc => nmr_env%chemical_shift_loc
      IF (PRESENT(chemical_shift_nics)) chemical_shift_nics => nmr_env%chemical_shift_nics
      IF (PRESENT(r_nics)) r_nics => nmr_env%r_nics
      IF (PRESENT(chemical_shift_nics_loc)) chemical_shift_nics_loc => nmr_env%chemical_shift_nics_loc
      IF (PRESENT(shift_gapw_radius)) shift_gapw_radius = nmr_env%shift_gapw_radius
      IF (PRESENT(do_nics)) do_nics = nmr_env%do_nics
      IF (PRESENT(interpolate_shift)) interpolate_shift = nmr_env%interpolate_shift

   END SUBROUTINE get_nmr_env

! **************************************************************************************************
!> \brief ...
!> \param issc_env ...
!> \param issc_on_atom_list ...
!> \param issc_gapw_radius ...
!> \param issc_loc ...
!> \param do_fc ...
!> \param do_sd ...
!> \param do_pso ...
!> \param do_dso ...
!> \param issc ...
!> \param interpolate_issc ...
!> \param psi1_efg ...
!> \param psi1_pso ...
!> \param psi1_dso ...
!> \param psi1_fc ...
!> \param efg_psi0 ...
!> \param pso_psi0 ...
!> \param dso_psi0 ...
!> \param fc_psi0 ...
!> \param matrix_efg ...
!> \param matrix_pso ...
!> \param matrix_dso ...
!> \param matrix_fc ...
! **************************************************************************************************
   SUBROUTINE get_issc_env(issc_env, issc_on_atom_list, issc_gapw_radius, issc_loc, &
                           do_fc, do_sd, do_pso, do_dso, &
                           issc, interpolate_issc, psi1_efg, psi1_pso, psi1_dso, psi1_fc, efg_psi0, pso_psi0, dso_psi0, fc_psi0, &
                           matrix_efg, matrix_pso, matrix_dso, matrix_fc)

      TYPE(issc_env_type)                                :: issc_env
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: issc_on_atom_list
      REAL(dp), OPTIONAL                                 :: issc_gapw_radius
      REAL(dp), DIMENSION(:, :, :, :, :), OPTIONAL, &
         POINTER                                         :: issc_loc
      LOGICAL, OPTIONAL                                  :: do_fc, do_sd, do_pso, do_dso
      REAL(dp), DIMENSION(:, :, :, :, :), OPTIONAL, &
         POINTER                                         :: issc
      LOGICAL, OPTIONAL                                  :: interpolate_issc
      TYPE(cp_fm_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: psi1_efg, psi1_pso, psi1_dso
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: psi1_fc
      TYPE(cp_fm_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: efg_psi0, pso_psi0, dso_psi0
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: fc_psi0
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: matrix_efg, matrix_pso, matrix_dso, &
                                                            matrix_fc

      CHARACTER(len=*), PARAMETER :: routineN = 'get_issc_env', routineP = moduleN//':'//routineN

      IF(.NOT.(issc_env%ref_count > 0))CALL cp__a("qs_linres_types.F",629)

      IF (PRESENT(issc_on_atom_list)) issc_on_atom_list => issc_env%issc_on_atom_list
      IF (PRESENT(issc_gapw_radius)) issc_gapw_radius = issc_env%issc_gapw_radius
      IF (PRESENT(issc_loc)) issc_loc => issc_env%issc_loc
      IF (PRESENT(issc)) issc => issc_env%issc
      IF (PRESENT(interpolate_issc)) interpolate_issc = issc_env%interpolate_issc
      IF (PRESENT(psi1_efg)) psi1_efg => issc_env%psi1_efg
      IF (PRESENT(psi1_pso)) psi1_pso => issc_env%psi1_pso
      IF (PRESENT(psi1_dso)) psi1_dso => issc_env%psi1_dso
      IF (PRESENT(psi1_fc)) psi1_fc => issc_env%psi1_fc
      IF (PRESENT(efg_psi0)) efg_psi0 => issc_env%efg_psi0
      IF (PRESENT(pso_psi0)) pso_psi0 => issc_env%pso_psi0
      IF (PRESENT(dso_psi0)) dso_psi0 => issc_env%dso_psi0
      IF (PRESENT(fc_psi0)) fc_psi0 => issc_env%fc_psi0
      IF (PRESENT(matrix_efg)) matrix_efg => issc_env%matrix_efg
      IF (PRESENT(matrix_pso)) matrix_pso => issc_env%matrix_pso
      IF (PRESENT(matrix_fc)) matrix_fc => issc_env%matrix_fc
      IF (PRESENT(matrix_dso)) matrix_dso => issc_env%matrix_dso
      IF (PRESENT(do_fc)) do_fc = issc_env%do_fc
      IF (PRESENT(do_sd)) do_sd = issc_env%do_sd
      IF (PRESENT(do_pso)) do_pso = issc_env%do_pso
      IF (PRESENT(do_dso)) do_dso = issc_env%do_dso

   END SUBROUTINE get_issc_env

! **************************************************************************************************
!> \brief ...
!> \param current_env ...
!> \param jrho1_atom_set ...
!> \param jrho1_set ...
! **************************************************************************************************
   SUBROUTINE set_current_env(current_env, jrho1_atom_set, jrho1_set)

      TYPE(current_env_type)                             :: current_env
      TYPE(jrho_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: jrho1_atom_set
      TYPE(qs_rho_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: jrho1_set

      CHARACTER(len=*), PARAMETER :: routineN = 'set_current_env', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: idir

      IF(.NOT.(current_env%ref_count > 0))CALL cp__a("qs_linres_types.F",674)

      IF (PRESENT(jrho1_atom_set)) THEN
         IF (ASSOCIATED(current_env%jrho1_atom_set)) THEN
            CALL deallocate_jrho_atom_set(current_env%jrho1_atom_set)
         ENDIF
         current_env%jrho1_atom_set => jrho1_atom_set
      END IF

      IF (PRESENT(jrho1_set)) THEN
         IF (ASSOCIATED(current_env%jrho1_set)) THEN
            DO idir = 1, 3
               CALL qs_rho_release(current_env%jrho1_set(idir)%rho)
            END DO
         END IF
         current_env%jrho1_set => jrho1_set
      END IF

   END SUBROUTINE set_current_env

! **************************************************************************************************
!> \brief ...
!> \param epr_env ...
!> \param g_total ...
!> \param g_so ...
!> \param g_soo ...
!> \param nablavks_set ...
!> \param nablavks_atom_set ...
!> \param bind_set ...
!> \param bind_atom_set ...
! **************************************************************************************************
   SUBROUTINE get_epr_env(epr_env, g_total, g_so, g_soo, nablavks_set, nablavks_atom_set, &
                          bind_set, bind_atom_set)

      TYPE(epr_env_type)                                 :: epr_env
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: g_total, g_so, g_soo
      TYPE(qs_rho_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: nablavks_set
      TYPE(nablavks_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: nablavks_atom_set
      TYPE(qs_rho_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: bind_set
      TYPE(rho_atom_coeff), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: bind_atom_set

      CHARACTER(len=*), PARAMETER :: routineN = 'get_epr_env', routineP = moduleN//':'//routineN

      IF(.NOT.(epr_env%ref_count > 0))CALL cp__a("qs_linres_types.F",721)

      IF (PRESENT(g_total)) g_total => epr_env%g_total
      IF (PRESENT(g_so)) g_so => epr_env%g_so
      IF (PRESENT(g_soo)) g_soo => epr_env%g_soo
      IF (PRESENT(nablavks_set)) nablavks_set => epr_env%nablavks_set
      IF (PRESENT(nablavks_atom_set)) nablavks_atom_set => epr_env%nablavks_atom_set
      IF (PRESENT(bind_set)) bind_set => epr_env%bind_set
      IF (PRESENT(bind_atom_set)) bind_atom_set => epr_env%bind_atom_set

   END SUBROUTINE get_epr_env

! **************************************************************************************************
!> \brief ...
!> \param epr_env ...
!> \param g_free_factor ...
!> \param g_soo_chicorr_factor ...
!> \param g_soo_factor ...
!> \param g_so_factor ...
!> \param g_so_factor_gapw ...
!> \param g_zke_factor ...
!> \param nablavks_set ...
!> \param nablavks_atom_set ...
! **************************************************************************************************
   SUBROUTINE set_epr_env(epr_env, g_free_factor, g_soo_chicorr_factor, &
                          g_soo_factor, g_so_factor, g_so_factor_gapw, &
                          g_zke_factor, nablavks_set, nablavks_atom_set)

      TYPE(epr_env_type)                                 :: epr_env
      REAL(dp), INTENT(IN), OPTIONAL                     :: g_free_factor, g_soo_chicorr_factor, &
                                                            g_soo_factor, g_so_factor, &
                                                            g_so_factor_gapw, g_zke_factor
      TYPE(qs_rho_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: nablavks_set
      TYPE(nablavks_atom_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: nablavks_atom_set

      CHARACTER(len=*), PARAMETER :: routineN = 'set_epr_env', routineP = moduleN//':'//routineN

      INTEGER                                            :: idir, ispin

      IF(.NOT.(epr_env%ref_count > 0))CALL cp__a("qs_linres_types.F",762)

      IF (PRESENT(g_free_factor)) epr_env%g_free_factor = g_free_factor
      IF (PRESENT(g_zke_factor)) epr_env%g_zke_factor = g_zke_factor
      IF (PRESENT(g_so_factor)) epr_env%g_so_factor = g_so_factor
      IF (PRESENT(g_so_factor_gapw)) epr_env%g_so_factor_gapw = g_so_factor_gapw
      IF (PRESENT(g_soo_factor)) epr_env%g_soo_factor = g_soo_factor
      IF (PRESENT(g_soo_chicorr_factor)) epr_env%g_soo_chicorr_factor = g_soo_chicorr_factor

      IF (PRESENT(nablavks_set)) THEN
         IF (ASSOCIATED(epr_env%nablavks_set)) THEN
            DO ispin = 1, 2
               DO idir = 1, 3
                  CALL qs_rho_release(epr_env%nablavks_set(idir, ispin)%rho)
               END DO
            END DO
         END IF
         epr_env%nablavks_set => nablavks_set
      ENDIF

      IF (PRESENT(nablavks_atom_set)) THEN
         IF (ASSOCIATED(epr_env%nablavks_atom_set)) THEN
            CALL deallocate_nablavks_atom_set(epr_env%nablavks_atom_set)
         ENDIF
         epr_env%nablavks_atom_set => nablavks_atom_set
      ENDIF

   END SUBROUTINE set_epr_env

! **************************************************************************************************
!> \brief ...
!> \param nablavks_atom_set ...
!> \param natom ...
! **************************************************************************************************
   SUBROUTINE allocate_nablavks_atom_set(nablavks_atom_set, natom)

      TYPE(nablavks_atom_type), DIMENSION(:), POINTER    :: nablavks_atom_set
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_nablavks_atom_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iat

      ALLOCATE (nablavks_atom_set(natom))

      DO iat = 1, natom
         NULLIFY (nablavks_atom_set(iat)%nablavks_vec_rad_h)
         NULLIFY (nablavks_atom_set(iat)%nablavks_vec_rad_s)
      ENDDO
   END SUBROUTINE allocate_nablavks_atom_set

! **************************************************************************************************
!> \brief ...
!> \param nablavks_atom_set ...
! **************************************************************************************************
   SUBROUTINE deallocate_nablavks_atom_set(nablavks_atom_set)

      TYPE(nablavks_atom_type), DIMENSION(:), POINTER    :: nablavks_atom_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_nablavks_atom_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iat, idir, n, natom

      IF(.NOT.(ASSOCIATED(nablavks_atom_set)))CALL cp__a("qs_linres_types.F",827)
      natom = SIZE(nablavks_atom_set)

      DO iat = 1, natom
         IF (ASSOCIATED(nablavks_atom_set(iat)%nablavks_vec_rad_h)) THEN
            IF (ASSOCIATED(nablavks_atom_set(iat)%nablavks_vec_rad_h(1, 1)%r_coef)) THEN
               n = SIZE(nablavks_atom_set(iat)%nablavks_vec_rad_h, 2)
               DO i = 1, n
                  DO idir = 1, 3
                     DEALLOCATE (nablavks_atom_set(iat)%nablavks_vec_rad_h(idir, i)%r_coef)
                     DEALLOCATE (nablavks_atom_set(iat)%nablavks_vec_rad_s(idir, i)%r_coef)
                  ENDDO
               ENDDO
            ENDIF
            DEALLOCATE (nablavks_atom_set(iat)%nablavks_vec_rad_h)
            DEALLOCATE (nablavks_atom_set(iat)%nablavks_vec_rad_s)
         ENDIF
      ENDDO
      DEALLOCATE (nablavks_atom_set)
   END SUBROUTINE deallocate_nablavks_atom_set

! **************************************************************************************************
!> \brief ...
!> \param jrho_atom_set ...
!> \param natom ...
! **************************************************************************************************
   SUBROUTINE allocate_jrho_atom_set(jrho_atom_set, natom)

      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho_atom_set
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_jrho_atom_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iat

      ALLOCATE (jrho_atom_set(natom))

      DO iat = 1, natom
         NULLIFY (jrho_atom_set(iat)%cjc0_h)
         NULLIFY (jrho_atom_set(iat)%cjc0_s)
         NULLIFY (jrho_atom_set(iat)%cjc_h)
         NULLIFY (jrho_atom_set(iat)%cjc_s)
         NULLIFY (jrho_atom_set(iat)%cjc_ii_h)
         NULLIFY (jrho_atom_set(iat)%cjc_ii_s)
         NULLIFY (jrho_atom_set(iat)%cjc_iii_h)
         NULLIFY (jrho_atom_set(iat)%cjc_iii_s)
         NULLIFY (jrho_atom_set(iat)%jrho_vec_rad_h)
         NULLIFY (jrho_atom_set(iat)%jrho_vec_rad_s)
         NULLIFY (jrho_atom_set(iat)%jrho_h)
         NULLIFY (jrho_atom_set(iat)%jrho_s)
         NULLIFY (jrho_atom_set(iat)%jrho_a_h)
         NULLIFY (jrho_atom_set(iat)%jrho_a_s)
         NULLIFY (jrho_atom_set(iat)%jrho_b_h)
         NULLIFY (jrho_atom_set(iat)%jrho_b_s)
         NULLIFY (jrho_atom_set(iat)%jrho_a_h_ii)
         NULLIFY (jrho_atom_set(iat)%jrho_a_s_ii)
         NULLIFY (jrho_atom_set(iat)%jrho_b_h_ii)
         NULLIFY (jrho_atom_set(iat)%jrho_b_s_ii)
         NULLIFY (jrho_atom_set(iat)%jrho_a_h_iii)
         NULLIFY (jrho_atom_set(iat)%jrho_a_s_iii)
         NULLIFY (jrho_atom_set(iat)%jrho_b_h_iii)
         NULLIFY (jrho_atom_set(iat)%jrho_b_s_iii)
      ENDDO
   END SUBROUTINE allocate_jrho_atom_set

! **************************************************************************************************
!> \brief ...
!> \param jrho_atom_set ...
! **************************************************************************************************
   SUBROUTINE deallocate_jrho_atom_set(jrho_atom_set)

      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho_atom_set

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_jrho_atom_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iat, idir, n, natom

      IF(.NOT.(ASSOCIATED(jrho_atom_set)))CALL cp__a("qs_linres_types.F",906)
      natom = SIZE(jrho_atom_set)

      DO iat = 1, natom
         IF (ASSOCIATED(jrho_atom_set(iat)%cjc_h)) THEN
            IF (ASSOCIATED(jrho_atom_set(iat)%cjc_h(1)%r_coef)) THEN
               n = SIZE(jrho_atom_set(iat)%cjc_h)
               DO i = 1, n
                  !
                  ! size = (nsotot,nsotot) replicated
                  DEALLOCATE (jrho_atom_set(iat)%cjc0_h(i)%r_coef, &
                              jrho_atom_set(iat)%cjc0_s(i)%r_coef, &
                              jrho_atom_set(iat)%cjc_h(i)%r_coef, &
                              jrho_atom_set(iat)%cjc_s(i)%r_coef, &
                              jrho_atom_set(iat)%cjc_ii_h(i)%r_coef, &
                              jrho_atom_set(iat)%cjc_ii_s(i)%r_coef, &
                              jrho_atom_set(iat)%cjc_iii_h(i)%r_coef, &
                              jrho_atom_set(iat)%cjc_iii_s(i)%r_coef)
               END DO
            END IF
            DEALLOCATE (jrho_atom_set(iat)%cjc0_h, &
                        jrho_atom_set(iat)%cjc0_s, &
                        jrho_atom_set(iat)%cjc_h, &
                        jrho_atom_set(iat)%cjc_s, &
                        jrho_atom_set(iat)%cjc_ii_h, &
                        jrho_atom_set(iat)%cjc_ii_s, &
                        jrho_atom_set(iat)%cjc_iii_h, &
                        jrho_atom_set(iat)%cjc_iii_s)
         END IF

         IF (ASSOCIATED(jrho_atom_set(iat)%jrho_a_h)) THEN
            IF (ASSOCIATED(jrho_atom_set(iat)%jrho_a_h(1)%r_coef)) THEN
               n = SIZE(jrho_atom_set(iat)%jrho_a_h)
               DO i = 1, n
                  !
                  ! size = (nr,max_iso_not0) distributed
                  DEALLOCATE (jrho_atom_set(iat)%jrho_h(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_s(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_a_h(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_a_s(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_b_h(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_b_s(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_a_h_ii(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_a_s_ii(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_b_h_ii(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_b_s_ii(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_a_h_iii(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_a_s_iii(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_b_h_iii(i)%r_coef, &
                              jrho_atom_set(iat)%jrho_b_s_iii(i)%r_coef)
               END DO
            END IF
            DEALLOCATE (jrho_atom_set(iat)%jrho_h, &
                        jrho_atom_set(iat)%jrho_s, &
                        jrho_atom_set(iat)%jrho_a_h, &
                        jrho_atom_set(iat)%jrho_a_s, &
                        jrho_atom_set(iat)%jrho_b_h, &
                        jrho_atom_set(iat)%jrho_b_s, &
                        jrho_atom_set(iat)%jrho_a_h_ii, &
                        jrho_atom_set(iat)%jrho_a_s_ii, &
                        jrho_atom_set(iat)%jrho_b_h_ii, &
                        jrho_atom_set(iat)%jrho_b_s_ii, &
                        jrho_atom_set(iat)%jrho_a_h_iii, &
                        jrho_atom_set(iat)%jrho_a_s_iii, &
                        jrho_atom_set(iat)%jrho_b_h_iii, &
                        jrho_atom_set(iat)%jrho_b_s_iii)
         END IF

         IF (ASSOCIATED(jrho_atom_set(iat)%jrho_vec_rad_h)) THEN
            IF (ASSOCIATED(jrho_atom_set(iat)%jrho_vec_rad_h(1, 1)%r_coef)) THEN
               n = SIZE(jrho_atom_set(iat)%jrho_vec_rad_h, 2)
               DO i = 1, n
                  DO idir = 1, 3
                     !
                     ! size =(nr,na) distributed
                     DEALLOCATE (jrho_atom_set(iat)%jrho_vec_rad_h(idir, i)%r_coef, &
                                 jrho_atom_set(iat)%jrho_vec_rad_s(idir, i)%r_coef)
                  END DO
               END DO
            ENDIF
            DEALLOCATE (jrho_atom_set(iat)%jrho_vec_rad_h, &
                        jrho_atom_set(iat)%jrho_vec_rad_s)
         END IF
      END DO
      DEALLOCATE (jrho_atom_set)

   END SUBROUTINE deallocate_jrho_atom_set

! **************************************************************************************************
!> \brief ...
!> \param jrho1_atom ...
!> \param ispin ...
!> \param nr ...
!> \param na ...
!> \param max_iso_not0 ...
! **************************************************************************************************
   SUBROUTINE allocate_jrho_atom_rad(jrho1_atom, ispin, nr, na, max_iso_not0)

      TYPE(jrho_atom_type), POINTER                      :: jrho1_atom
      INTEGER, INTENT(IN)                                :: ispin, nr, na, max_iso_not0

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_jrho_atom_rad', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, idir

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(jrho1_atom)))CALL cp__a("qs_linres_types.F",1014)

      DO idir = 1, 3
         ALLOCATE (jrho1_atom%jrho_vec_rad_h(idir, ispin)%r_coef(nr, na), &
                   jrho1_atom%jrho_vec_rad_s(idir, ispin)%r_coef(nr, na))
         jrho1_atom%jrho_vec_rad_h(idir, ispin)%r_coef = 0.0_dp
         jrho1_atom%jrho_vec_rad_s(idir, ispin)%r_coef = 0.0_dp
      ENDDO

      ALLOCATE (jrho1_atom%jrho_h(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_s(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_a_h(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_a_s(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_b_h(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_b_s(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_a_h_ii(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_a_s_ii(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_b_h_ii(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_b_s_ii(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_a_h_iii(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_a_s_iii(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_b_h_iii(ispin)%r_coef(nr, max_iso_not0), &
                jrho1_atom%jrho_b_s_iii(ispin)%r_coef(nr, max_iso_not0))
      !
      jrho1_atom%jrho_h(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_s(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_a_h(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_a_s(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_h(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_s(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_a_h_ii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_a_s_ii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_h_ii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_s_ii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_a_h_iii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_a_s_iii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_h_iii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_s_iii(ispin)%r_coef = 0.0_dp
      CALL timestop(handle)

   END SUBROUTINE allocate_jrho_atom_rad

! **************************************************************************************************
!> \brief ...
!> \param jrho1_atom ...
!> \param ispin ...
! **************************************************************************************************
   SUBROUTINE set2zero_jrho_atom_rad(jrho1_atom, ispin)
      !
      TYPE(jrho_atom_type), POINTER                      :: jrho1_atom
      INTEGER, INTENT(IN)                                :: ispin

      CHARACTER(len=*), PARAMETER :: routineN = 'set2zero_jrho_atom_rad', &
         routineP = moduleN//':'//routineN

!

      IF(.NOT.(ASSOCIATED(jrho1_atom)))CALL cp__a("qs_linres_types.F",1071)
      !
      jrho1_atom%jrho_h(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_s(ispin)%r_coef = 0.0_dp
      !
      jrho1_atom%jrho_a_h(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_a_s(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_h(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_s(ispin)%r_coef = 0.0_dp
      !
      jrho1_atom%jrho_a_h_ii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_a_s_ii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_h_ii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_s_ii(ispin)%r_coef = 0.0_dp
      !
      jrho1_atom%jrho_a_h_iii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_a_s_iii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_h_iii(ispin)%r_coef = 0.0_dp
      jrho1_atom%jrho_b_s_iii(ispin)%r_coef = 0.0_dp
      !
   END SUBROUTINE set2zero_jrho_atom_rad

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param jrho1_atom_set ...
!> \param iatom ...
!> \param nsotot ...
! **************************************************************************************************
   SUBROUTINE allocate_jrho_coeff(jrho1_atom_set, iatom, nsotot)

      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho1_atom_set
      INTEGER, INTENT(IN)                                :: iatom, nsotot

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_jrho_coeff', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(jrho1_atom_set)))CALL cp__a("qs_linres_types.F",1112)
      DO i = 1, SIZE(jrho1_atom_set(iatom)%cjc0_h, 1)
         ALLOCATE (jrho1_atom_set(iatom)%cjc0_h(i)%r_coef(nsotot, nsotot), &
                   jrho1_atom_set(iatom)%cjc0_s(i)%r_coef(nsotot, nsotot), &
                   jrho1_atom_set(iatom)%cjc_h(i)%r_coef(nsotot, nsotot), &
                   jrho1_atom_set(iatom)%cjc_s(i)%r_coef(nsotot, nsotot), &
                   jrho1_atom_set(iatom)%cjc_ii_h(i)%r_coef(nsotot, nsotot), &
                   jrho1_atom_set(iatom)%cjc_ii_s(i)%r_coef(nsotot, nsotot), &
                   jrho1_atom_set(iatom)%cjc_iii_h(i)%r_coef(nsotot, nsotot), &
                   jrho1_atom_set(iatom)%cjc_iii_s(i)%r_coef(nsotot, nsotot))
         jrho1_atom_set(iatom)%cjc0_h(i)%r_coef = 0.0_dp
         jrho1_atom_set(iatom)%cjc0_s(i)%r_coef = 0.0_dp
         jrho1_atom_set(iatom)%cjc_h(i)%r_coef = 0.0_dp
         jrho1_atom_set(iatom)%cjc_s(i)%r_coef = 0.0_dp
         jrho1_atom_set(iatom)%cjc_ii_h(i)%r_coef = 0.0_dp
         jrho1_atom_set(iatom)%cjc_ii_s(i)%r_coef = 0.0_dp
         jrho1_atom_set(iatom)%cjc_iii_h(i)%r_coef = 0.0_dp
         jrho1_atom_set(iatom)%cjc_iii_s(i)%r_coef = 0.0_dp
      ENDDO
      CALL timestop(handle)
   END SUBROUTINE allocate_jrho_coeff

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param jrho1_atom_set ...
!> \param iatom ...
! **************************************************************************************************
   SUBROUTINE deallocate_jrho_coeff(jrho1_atom_set, iatom)

      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho1_atom_set
      INTEGER, INTENT(IN)                                :: iatom

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_jrho_coeff', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)
      IF(.NOT.(ASSOCIATED(jrho1_atom_set)))CALL cp__a("qs_linres_types.F",1152)
      DO i = 1, SIZE(jrho1_atom_set(iatom)%cjc0_h, 1)
         DEALLOCATE (jrho1_atom_set(iatom)%cjc0_h(i)%r_coef, &
                     jrho1_atom_set(iatom)%cjc0_s(i)%r_coef, &
                     jrho1_atom_set(iatom)%cjc_h(i)%r_coef, &
                     jrho1_atom_set(iatom)%cjc_s(i)%r_coef, &
                     jrho1_atom_set(iatom)%cjc_ii_h(i)%r_coef, &
                     jrho1_atom_set(iatom)%cjc_ii_s(i)%r_coef, &
                     jrho1_atom_set(iatom)%cjc_iii_h(i)%r_coef, &
                     jrho1_atom_set(iatom)%cjc_iii_s(i)%r_coef)
      ENDDO
      CALL timestop(handle)
   END SUBROUTINE deallocate_jrho_coeff

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param jrho1_atom_set ...
!> \param iatom ...
!> \param cjc_h ...
!> \param cjc_s ...
!> \param cjc_ii_h ...
!> \param cjc_ii_s ...
!> \param cjc_iii_h ...
!> \param cjc_iii_s ...
!> \param jrho_vec_rad_h ...
!> \param jrho_vec_rad_s ...
! **************************************************************************************************
   SUBROUTINE get_jrho_atom(jrho1_atom_set, iatom, cjc_h, cjc_s, cjc_ii_h, cjc_ii_s, &
                            cjc_iii_h, cjc_iii_s, jrho_vec_rad_h, jrho_vec_rad_s)

      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho1_atom_set
      INTEGER, INTENT(IN)                                :: iatom
      TYPE(rho_atom_coeff), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: cjc_h, cjc_s, cjc_ii_h, cjc_ii_s, &
                                                            cjc_iii_h, cjc_iii_s
      TYPE(rho_atom_coeff), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: jrho_vec_rad_h, jrho_vec_rad_s

      CHARACTER(len=*), PARAMETER :: routineN = 'get_jrho_atom', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(jrho1_atom_set)))CALL cp__a("qs_linres_types.F",1194)

      IF (PRESENT(cjc_h)) cjc_h => jrho1_atom_set(iatom)%cjc_h
      IF (PRESENT(cjc_s)) cjc_s => jrho1_atom_set(iatom)%cjc_s
      IF (PRESENT(cjc_ii_h)) cjc_ii_h => jrho1_atom_set(iatom)%cjc_ii_h
      IF (PRESENT(cjc_ii_s)) cjc_ii_s => jrho1_atom_set(iatom)%cjc_ii_s
      IF (PRESENT(cjc_iii_h)) cjc_iii_h => jrho1_atom_set(iatom)%cjc_iii_h
      IF (PRESENT(cjc_iii_s)) cjc_iii_s => jrho1_atom_set(iatom)%cjc_iii_s
      IF (PRESENT(jrho_vec_rad_h)) jrho_vec_rad_h => jrho1_atom_set(iatom)%jrho_vec_rad_h
      IF (PRESENT(jrho_vec_rad_s)) jrho_vec_rad_s => jrho1_atom_set(iatom)%jrho_vec_rad_s

   END SUBROUTINE get_jrho_atom

! **************************************************************************************************
!> \brief ...
!> \param jrho1_atom_set ...
!> \param atomic_kind_set ...
!> \param nspins ...
! **************************************************************************************************
   SUBROUTINE init_jrho_atom_set(jrho1_atom_set, atomic_kind_set, nspins)
      TYPE(jrho_atom_type), DIMENSION(:), POINTER        :: jrho1_atom_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      INTEGER, INTENT(IN)                                :: nspins

      CHARACTER(len=*), PARAMETER :: routineN = 'init_jrho_atom_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iat, iatom, idir, ikind, ispin, &
                                                            nat, natom, nkind
      INTEGER, DIMENSION(:), POINTER                     :: atom_list

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(atomic_kind_set)))CALL cp__a("qs_linres_types.F",1227)

      IF (ASSOCIATED(jrho1_atom_set)) THEN
         CALL deallocate_jrho_atom_set(jrho1_atom_set)
      END IF

      CALL get_atomic_kind_set(atomic_kind_set, natom=natom)

      CALL allocate_jrho_atom_set(jrho1_atom_set, natom)

      nkind = SIZE(atomic_kind_set)

      DO ikind = 1, nkind

         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=nat)

         DO iat = 1, nat
            iatom = atom_list(iat)

            !*** allocate the radial density for each LM,for each atom ***
            ALLOCATE (jrho1_atom_set(iatom)%jrho_vec_rad_h(3, nspins), &
                      jrho1_atom_set(iatom)%jrho_vec_rad_s(3, nspins), &
                      jrho1_atom_set(iatom)%jrho_h(nspins), &
                      jrho1_atom_set(iatom)%jrho_s(nspins), &
                      jrho1_atom_set(iatom)%jrho_a_h(nspins), &
                      jrho1_atom_set(iatom)%jrho_a_s(nspins), &
                      jrho1_atom_set(iatom)%jrho_b_h(nspins), &
                      jrho1_atom_set(iatom)%jrho_b_s(nspins), &
                      jrho1_atom_set(iatom)%jrho_a_h_ii(nspins), &
                      jrho1_atom_set(iatom)%jrho_a_s_ii(nspins), &
                      jrho1_atom_set(iatom)%jrho_b_s_ii(nspins), &
                      jrho1_atom_set(iatom)%jrho_b_h_ii(nspins), &
                      jrho1_atom_set(iatom)%jrho_a_h_iii(nspins), &
                      jrho1_atom_set(iatom)%jrho_a_s_iii(nspins), &
                      jrho1_atom_set(iatom)%jrho_b_s_iii(nspins), &
                      jrho1_atom_set(iatom)%jrho_b_h_iii(nspins), &
                      jrho1_atom_set(iatom)%cjc0_h(nspins), &
                      jrho1_atom_set(iatom)%cjc0_s(nspins), &
                      jrho1_atom_set(iatom)%cjc_h(nspins), &
                      jrho1_atom_set(iatom)%cjc_s(nspins), &
                      jrho1_atom_set(iatom)%cjc_ii_h(nspins), &
                      jrho1_atom_set(iatom)%cjc_ii_s(nspins), &
                      jrho1_atom_set(iatom)%cjc_iii_h(nspins), &
                      jrho1_atom_set(iatom)%cjc_iii_s(nspins))

            DO ispin = 1, nspins
               DO idir = 1, 3
                  NULLIFY (jrho1_atom_set(iatom)%jrho_vec_rad_h(idir, ispin)%r_coef)
                  NULLIFY (jrho1_atom_set(iatom)%jrho_vec_rad_s(idir, ispin)%r_coef)
               END DO
               NULLIFY (jrho1_atom_set(iatom)%jrho_h(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_s(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_a_h(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_a_s(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_b_h(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_b_s(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_a_h_ii(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_a_s_ii(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_b_h_ii(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_b_s_ii(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_a_h_iii(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_a_s_iii(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_b_h_iii(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%jrho_b_s_iii(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%cjc0_h(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%cjc0_s(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%cjc_h(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%cjc_s(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%cjc_ii_h(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%cjc_ii_s(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%cjc_iii_h(ispin)%r_coef)
               NULLIFY (jrho1_atom_set(iatom)%cjc_iii_s(ispin)%r_coef)

            ENDDO ! ispin

         END DO ! iat

      END DO ! ikind

      CALL timestop(handle)

   END SUBROUTINE init_jrho_atom_set

! **************************************************************************************************
!> \brief ...
!> \param nablavks_atom_set ...
!> \param atomic_kind_set ...
!> \param qs_kind_set ...
!> \param nspins ...
! **************************************************************************************************
   SUBROUTINE init_nablavks_atom_set(nablavks_atom_set, atomic_kind_set, qs_kind_set, nspins)

      TYPE(nablavks_atom_type), DIMENSION(:), POINTER    :: nablavks_atom_set
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set
      TYPE(qs_kind_type), DIMENSION(:), POINTER          :: qs_kind_set
      INTEGER, INTENT(IN)                                :: nspins

      CHARACTER(len=*), PARAMETER :: routineN = 'init_nablavks_atom_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, iat, iatom, idir, ikind, ispin, &
                                                            max_iso_not0, maxso, na, nat, natom, &
                                                            nkind, nr, nset, nsotot
      INTEGER, DIMENSION(:), POINTER                     :: atom_list
      TYPE(grid_atom_type), POINTER                      :: grid_atom
      TYPE(gto_basis_set_type), POINTER                  :: orb_basis_set
      TYPE(harmonics_atom_type), POINTER                 :: harmonics

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(qs_kind_set)))CALL cp__a("qs_linres_types.F",1337)

      IF (ASSOCIATED(nablavks_atom_set)) THEN
         CALL deallocate_nablavks_atom_set(nablavks_atom_set)
      END IF

      CALL get_atomic_kind_set(atomic_kind_set, natom=natom)

      CALL allocate_nablavks_atom_set(nablavks_atom_set, natom)

      nkind = SIZE(atomic_kind_set)

      DO ikind = 1, nkind
         CALL get_atomic_kind(atomic_kind_set(ikind), atom_list=atom_list, natom=nat)
         CALL get_qs_kind(qs_kind_set(ikind), &
                          basis_set=orb_basis_set, &
                          harmonics=harmonics, &
                          grid_atom=grid_atom)

         na = grid_atom%ng_sphere
         nr = grid_atom%nr

         CALL get_gto_basis_set(gto_basis_set=orb_basis_set, &
                                maxso=maxso, nset=nset)
         nsotot = maxso*nset
         max_iso_not0 = harmonics%max_iso_not0
         DO iat = 1, nat
            iatom = atom_list(iat)
            !*** allocate the radial density for each LM,for each atom ***

            ALLOCATE (nablavks_atom_set(iatom)%nablavks_vec_rad_h(3, nspins))
            ALLOCATE (nablavks_atom_set(iatom)%nablavks_vec_rad_s(3, nspins))
            DO ispin = 1, nspins
               DO idir = 1, 3
                  NULLIFY (nablavks_atom_set(iatom)%nablavks_vec_rad_h(idir, ispin)%r_coef)
                  NULLIFY (nablavks_atom_set(iatom)%nablavks_vec_rad_s(idir, ispin)%r_coef)
                  ALLOCATE (nablavks_atom_set(iatom)%nablavks_vec_rad_h(idir, ispin)%r_coef(nr, na))
                  ALLOCATE (nablavks_atom_set(iatom)%nablavks_vec_rad_s(idir, ispin)%r_coef(nr, na))
               END DO
            END DO ! ispin
         END DO ! iat

      END DO ! ikind

      CALL timestop(handle)

   END SUBROUTINE init_nablavks_atom_set
! **************************************************************************************************
!> \brief ...
!> \param polar_env ...
! **************************************************************************************************
   SUBROUTINE polar_env_create(polar_env)

      TYPE(polar_env_type)                               :: polar_env

      CHARACTER(len=*), PARAMETER :: routineN = 'polar_env_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(polar_env%ref_count == 0))CALL cp__a("qs_linres_types.F",1395)
      polar_env%ref_count = 1
!        polar_env%do_raman=.FALSE.
      NULLIFY (polar_env%polar)
      NULLIFY (polar_env%psi1_dBerry)
      NULLIFY (polar_env%dBerry_psi0)
      NULLIFY (polar_env%mo_derivs)

   END SUBROUTINE polar_env_create

! **************************************************************************************************
!> \brief ...
!> \param polar_env ...
!> \param do_raman ...
!> \param dBerry_psi0 ...
!> \param polar ...
!> \param psi1_dBerry ...
!> \param mo_derivs ...
! **************************************************************************************************
   SUBROUTINE get_polar_env(polar_env, do_raman, dBerry_psi0, polar, &
                            psi1_dBerry, mo_derivs)

      TYPE(polar_env_type)                               :: polar_env
      LOGICAL, OPTIONAL                                  :: do_raman
      TYPE(cp_fm_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: dBerry_psi0
      REAL(dp), DIMENSION(:, :), OPTIONAL, POINTER       :: polar
      TYPE(cp_fm_p_type), DIMENSION(:, :), OPTIONAL, &
         POINTER                                         :: psi1_dBerry, mo_derivs

      CHARACTER(len=*), PARAMETER :: routineN = 'get_polar_env', routineP = moduleN//':'//routineN

      IF(.NOT.(polar_env%ref_count > 0))CALL cp__a("qs_linres_types.F",1427)

      IF (PRESENT(polar)) polar => polar_env%polar
      IF (PRESENT(psi1_dBerry)) psi1_dBerry => polar_env%psi1_dBerry
      IF (PRESENT(dBerry_psi0)) dBerry_psi0 => polar_env%dBerry_psi0
      IF (PRESENT(mo_derivs)) mo_derivs => polar_env%mo_derivs
      IF (PRESENT(do_raman)) do_raman = polar_env%do_raman
   END SUBROUTINE get_polar_env

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param polar_env ...
! **************************************************************************************************
   SUBROUTINE set_polar_env(polar_env)

      TYPE(polar_env_type)                               :: polar_env

      CHARACTER(len=*), PARAMETER :: routineN = 'set_polar_env', routineP = moduleN//':'//routineN

      IF(.NOT.(polar_env%ref_count > 0))CALL cp__a("qs_linres_types.F",1448)

   END SUBROUTINE set_polar_env
! **************************************************************************************************

END MODULE qs_linres_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!
! **************************************************************************************************
MODULE hartree_local_types

   USE kinds,                           ONLY: dp
   USE qs_rho_atom_types,               ONLY: rho_atom_coeff
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! *** Global parameters (only in this module)

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'hartree_local_types'

! *** Define the ecoul_1center_type ***

! **************************************************************************************************
   TYPE ecoul_1center_type
      TYPE(rho_atom_coeff), POINTER   :: Vh1_h, Vh1_s
      REAL(dp)                        :: ecoul_1_h, ecoul_1_s, &
                                         ecoul_1_z, ecoul_1_0
   END TYPE ecoul_1center_type

! **************************************************************************************************
   TYPE hartree_local_type
      TYPE(ecoul_1center_type), &
         DIMENSION(:), POINTER   :: ecoul_1c
   END TYPE hartree_local_type

! *** Public subroutines ***

   PUBLIC :: allocate_ecoul_1center, &
             get_hartree_local, hartree_local_create, &
             hartree_local_release, set_ecoul_1c, &
             set_hartree_local

! *** Public data types ***

   PUBLIC :: ecoul_1center_type, hartree_local_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param ecoul_1c ...
!> \param natom ...
! **************************************************************************************************
   SUBROUTINE allocate_ecoul_1center(ecoul_1c, natom)

      TYPE(ecoul_1center_type), DIMENSION(:), POINTER    :: ecoul_1c
      INTEGER, INTENT(IN)                                :: natom

      CHARACTER(len=*), PARAMETER :: routineN = 'allocate_ecoul_1center', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iat

      IF (ASSOCIATED(ecoul_1c)) THEN
         CALL deallocate_ecoul_1center(ecoul_1c)
      END IF

      ALLOCATE (ecoul_1c(natom))

      DO iat = 1, natom
         ALLOCATE (ecoul_1c(iat)%Vh1_h)
         NULLIFY (ecoul_1c(iat)%Vh1_h%r_coef)
         ALLOCATE (ecoul_1c(iat)%Vh1_s)
         NULLIFY (ecoul_1c(iat)%Vh1_s%r_coef)
      END DO

   END SUBROUTINE allocate_ecoul_1center

! **************************************************************************************************
!> \brief ...
!> \param ecoul_1c ...
! **************************************************************************************************
   SUBROUTINE deallocate_ecoul_1center(ecoul_1c)

      TYPE(ecoul_1center_type), DIMENSION(:), POINTER    :: ecoul_1c

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_ecoul_1center', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iat, natom

      natom = SIZE(ecoul_1c, 1)

      DO iat = 1, natom
         IF (ASSOCIATED(ecoul_1c(iat)%Vh1_h%r_coef)) THEN
            DEALLOCATE (ecoul_1c(iat)%Vh1_h%r_coef)
         ENDIF
         DEALLOCATE (ecoul_1c(iat)%Vh1_h)

         IF (ASSOCIATED(ecoul_1c(iat)%Vh1_s%r_coef)) THEN
            DEALLOCATE (ecoul_1c(iat)%Vh1_s%r_coef)
         ENDIF
         DEALLOCATE (ecoul_1c(iat)%Vh1_s)

      END DO

      DEALLOCATE (ecoul_1c)

   END SUBROUTINE deallocate_ecoul_1center

! **************************************************************************************************
!> \brief ...
!> \param hartree_local ...
!> \param ecoul_1c ...
! **************************************************************************************************
   SUBROUTINE get_hartree_local(hartree_local, ecoul_1c)

      TYPE(hartree_local_type), POINTER                  :: hartree_local
      TYPE(ecoul_1center_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: ecoul_1c

      CHARACTER(len=*), PARAMETER :: routineN = 'get_hartree_local', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(ecoul_1c)) ecoul_1c => hartree_local%ecoul_1c

   END SUBROUTINE get_hartree_local

! **************************************************************************************************
!> \brief ...
!> \param hartree_local ...
! **************************************************************************************************
   SUBROUTINE hartree_local_create(hartree_local)

      TYPE(hartree_local_type), POINTER                  :: hartree_local

      CHARACTER(len=*), PARAMETER :: routineN = 'hartree_local_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (hartree_local)

      NULLIFY (hartree_local%ecoul_1c)

   END SUBROUTINE hartree_local_create

! **************************************************************************************************
!> \brief ...
!> \param hartree_local ...
! **************************************************************************************************
   SUBROUTINE hartree_local_release(hartree_local)

      TYPE(hartree_local_type), POINTER                  :: hartree_local

      CHARACTER(len=*), PARAMETER :: routineN = 'hartree_local_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(hartree_local)) THEN
         IF (ASSOCIATED(hartree_local%ecoul_1c)) THEN
            CALL deallocate_ecoul_1center(hartree_local%ecoul_1c)
         END IF

         DEALLOCATE (hartree_local)
      END IF

   END SUBROUTINE hartree_local_release

! **************************************************************************************************
!> \brief ...
!> \param ecoul_1c ...
!> \param iatom ...
!> \param ecoul_1_h ...
!> \param ecoul_1_s ...
!> \param ecoul_1_z ...
!> \param ecoul_1_0 ...
! **************************************************************************************************
   SUBROUTINE set_ecoul_1c(ecoul_1c, iatom, ecoul_1_h, ecoul_1_s, ecoul_1_z, ecoul_1_0)

      TYPE(ecoul_1center_type), DIMENSION(:), POINTER    :: ecoul_1c
      INTEGER, INTENT(IN), OPTIONAL                      :: iatom
      REAL(dp), INTENT(IN), OPTIONAL                     :: ecoul_1_h, ecoul_1_s, ecoul_1_z, &
                                                            ecoul_1_0

      IF (PRESENT(iatom)) THEN
         IF (PRESENT(ecoul_1_h)) ecoul_1c(iatom)%ecoul_1_h = ecoul_1_h
         IF (PRESENT(ecoul_1_s)) ecoul_1c(iatom)%ecoul_1_s = ecoul_1_s
         IF (PRESENT(ecoul_1_0)) ecoul_1c(iatom)%ecoul_1_0 = ecoul_1_0
         IF (PRESENT(ecoul_1_z)) ecoul_1c(iatom)%ecoul_1_z = ecoul_1_z
      END IF

   END SUBROUTINE set_ecoul_1c

! **************************************************************************************************
!> \brief ...
!> \param hartree_local ...
!> \param ecoul_1c ...
! **************************************************************************************************
   SUBROUTINE set_hartree_local(hartree_local, ecoul_1c)

      TYPE(hartree_local_type), POINTER                  :: hartree_local
      TYPE(ecoul_1center_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: ecoul_1c

      CHARACTER(len=*), PARAMETER :: routineN = 'set_hartree_local', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(ecoul_1c)) hartree_local%ecoul_1c => ecoul_1c

   END SUBROUTINE set_hartree_local

END MODULE hartree_local_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief routine to handle vectors of full matrixes
!> \par History
!>      08.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE cp_fm_vect
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_retain
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_fm_vect'

   PUBLIC :: cp_fm_vect_dealloc, cp_fm_vect_copy
!***
CONTAINS

! **************************************************************************************************
!> \brief deallocate an array of pointers to blacs matrixes
!> \param matrixes the array of matrixes to deallocate
!> \par History
!>      07.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_fm_vect_dealloc(matrixes)
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: matrixes

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_vect_dealloc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(matrixes)) THEN
         DO i = 1, SIZE(matrixes)
            CALL cp_fm_release(matrixes(i)%matrix)
         END DO
         DEALLOCATE (matrixes)
      END IF
   END SUBROUTINE cp_fm_vect_dealloc

! **************************************************************************************************
!> \brief Does a shallow copy of an array of full matrices (i.e. just retains
!>      the matrices)
!> \param matrixes the matrixes to copy
!> \param copy ...
!> \par History
!>      09.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE cp_fm_vect_copy(matrixes, copy)
      TYPE(cp_fm_p_type), DIMENSION(:), INTENT(in)       :: matrixes
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER          :: copy

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_fm_vect_copy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      ALLOCATE (copy(SIZE(matrixes)))
      DO i = 1, SIZE(matrixes)
         copy(i)%matrix => matrixes(i)%matrix
         CALL cp_fm_retain(matrixes(i)%matrix)
      END DO
   END SUBROUTINE cp_fm_vect_copy

END MODULE cp_fm_vect
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types and set_get for real time propagation
!>        depending on runtype and diagonalization method different
!>        matrices are allocated
!>        exp_H_old, exp_H_new, mos_new, mos_old contain always
!>        real and imaginary parts of the matrices
!>        odd index = real part (alpha, beta spin)
!>        even index= imaginary part (alpha, beta spin)
!> \par History
!>      02.2014 switched to dbcsr matrices [Samuel Andermatt]
!> \author Florian Schiffmann 02.09
! **************************************************************************************************

MODULE rt_propagation_types

   USE bibliography,                    ONLY: Kunert2003,&
                                              cite_reference
   USE cp_control_types,                ONLY: dft_control_type,&
                                              rtp_control_type
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_allocate_matrix_set,&
                                              cp_dbcsr_create,&
                                              cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_init,&
                                              cp_dbcsr_init_p,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_fm_pool_types,                ONLY: cp_fm_pool_p_type,&
                                              fm_pool_get_el_struct
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_get,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type,&
                                              cp_fm_release
   USE cp_fm_vect,                      ONLY: cp_fm_vect_dealloc
   USE cp_log_handling,                 ONLY: cp_to_string
   USE kinds,                           ONLY: dp
   USE qs_matrix_pools,                 ONLY: mpools_get,&
                                              qs_matrix_pools_type
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rt_propagation_types'

   TYPE rtp_rho_type
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:) :: new
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:) :: old
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:) :: next
   END TYPE rtp_rho_type

   TYPE rtp_history_type
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:, :)               :: rho_history
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:)                 :: s_history
      TYPE(cp_fm_p_type), POINTER, DIMENSION(:, :)                :: mo_history
   END TYPE rtp_history_type

   TYPE rtp_mos_type
      TYPE(cp_fm_p_type), POINTER, DIMENSION(:)     :: new
      TYPE(cp_fm_p_type), POINTER, DIMENSION(:)     :: old
      TYPE(cp_fm_p_type), POINTER, DIMENSION(:)     :: next
      TYPE(cp_fm_p_type), POINTER, DIMENSION(:)     :: admm
   END TYPE rtp_mos_type

   TYPE rt_prop_type
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:)  :: exp_H_old
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:)  :: exp_H_new
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:)  :: H_last_iter
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:)  :: propagator_matrix
      TYPE(cp_dbcsr_type), POINTER                 :: S_inv
      TYPE(cp_dbcsr_type), POINTER                 :: S_half
      TYPE(cp_dbcsr_type), POINTER                 :: S_minus_half
      TYPE(cp_dbcsr_type), POINTER                 :: B_mat
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:)  :: C_mat
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:)  :: S_der
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:)  :: SinvH
      TYPE(cp_dbcsr_p_type), POINTER, DIMENSION(:) :: SinvB
      TYPE(rtp_rho_type), POINTER                  :: rho
      TYPE(rtp_mos_type), POINTER                  :: mos
      REAL(KIND=dp)                               :: energy_old
      REAL(KIND=dp)                               :: energy_new
      REAL(KIND=dp)                               :: dt
      REAL(KIND=dp)                               :: delta_iter
      REAL(KIND=dp)                               :: delta_iter_old
      REAL(KIND=dp)                               :: filter_eps
      REAL(KIND=dp)                               :: filter_eps_small
      REAL(KIND=dp)                               :: mixing_factor
      LOGICAL                                     :: mixing
      LOGICAL                                     :: do_hfx
      INTEGER, DIMENSION(:, :), ALLOCATABLE          :: orders
      INTEGER                                     :: nsteps, istep, i_start
      INTEGER                                     :: iter
      INTEGER                                     :: narn_old
      LOGICAL                                     :: converged
      LOGICAL                                     :: matrix_update
      LOGICAL                                     :: write_restart
      TYPE(rtp_history_type), POINTER              :: history
      TYPE(cp_fm_struct_type), POINTER            :: ao_ao_fmstruct
      INTEGER                                     :: lanzcos_max_iter
      REAL(KIND=dp)                               :: lanzcos_threshold
      INTEGER                                     :: newton_schulz_order
      LOGICAL                                     :: linear_scaling
   END TYPE rt_prop_type

! *** Public data types ***

   PUBLIC :: rt_prop_type

! *** Public subroutines ***

   PUBLIC :: rt_prop_create, &
             get_rtp, &
             rt_prop_release, &
             rtp_history_create
CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param rtp ...
!> \param mos ...
!> \param mpools ...
!> \param dft_control ...
!> \param template ...
!> \param linear_scaling ...
!> \param mos_aux ...
! **************************************************************************************************
   SUBROUTINE rt_prop_create(rtp, mos, mpools, dft_control, template, linear_scaling, mos_aux)

      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos
      TYPE(qs_matrix_pools_type), POINTER                :: mpools
      TYPE(dft_control_type), POINTER                    :: dft_control
      TYPE(cp_dbcsr_type), POINTER                       :: template
      LOGICAL                                            :: linear_scaling
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos_aux

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_prop_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, nao, nrow_block, nspin
      TYPE(cp_fm_pool_p_type), DIMENSION(:), POINTER     :: ao_mo_fm_pools
      TYPE(cp_fm_struct_type), POINTER                   :: ao_mo_fmstruct
      TYPE(rtp_control_type), POINTER                    :: rtp_control

      CALL cite_reference(Kunert2003)

      NULLIFY (rtp_control)

      rtp_control => dft_control%rtp_control

      nspin = dft_control%nspins

      NULLIFY (rtp%mos, rtp%rho)
      rtp%linear_scaling = linear_scaling

      IF (rtp%linear_scaling) THEN
         ALLOCATE (rtp%rho)
         NULLIFY (rtp%rho%old)
         CALL cp_dbcsr_allocate_matrix_set(rtp%rho%old, 2*nspin)
         NULLIFY (rtp%rho%next)
         CALL cp_dbcsr_allocate_matrix_set(rtp%rho%next, 2*nspin)
         NULLIFY (rtp%rho%new)
         CALL cp_dbcsr_allocate_matrix_set(rtp%rho%new, 2*nspin)
         DO i = 1, 2*nspin
            CALL cp_dbcsr_init_p(rtp%rho%old(i)%matrix)
            CALL cp_dbcsr_create(rtp%rho%old(i)%matrix, template=template, matrix_type="N")
            CALL cp_dbcsr_init_p(rtp%rho%next(i)%matrix)
            CALL cp_dbcsr_create(rtp%rho%next(i)%matrix, template=template, matrix_type="N")
            CALL cp_dbcsr_init_p(rtp%rho%new(i)%matrix)
            CALL cp_dbcsr_create(rtp%rho%new(i)%matrix, template=template, matrix_type="N")
         END DO
      ELSE
         CALL mpools_get(mpools, ao_mo_fm_pools=ao_mo_fm_pools)

         ao_mo_fmstruct => fm_pool_get_el_struct(ao_mo_fm_pools(1)%pool)
         CALL cp_fm_struct_get(ao_mo_fmstruct, nrow_block=nrow_block)
         CALL get_mo_set(mos(1)%mo_set, nao=nao)

         CALL cp_fm_struct_create(fmstruct=rtp%ao_ao_fmstruct, &
                                  nrow_block=nrow_block, ncol_block=nrow_block, &
                                  nrow_global=nao, ncol_global=nao, &
                                  template_fmstruct=ao_mo_fmstruct)
         ALLOCATE (rtp%mos)
         ALLOCATE (rtp%mos%old(2*nspin))
         ALLOCATE (rtp%mos%new(2*nspin))
         ALLOCATE (rtp%mos%next(2*nspin))
         NULLIFY (rtp%mos%admm)
         IF (dft_control%do_admm) THEN
            ALLOCATE (rtp%mos%admm(2*nspin))
         END IF
         DO i = 1, nspin
            DO j = 1, 2
               NULLIFY (rtp%mos%old(2*(i-1)+j)%matrix)
               NULLIFY (rtp%mos%new(2*(i-1)+j)%matrix)
               NULLIFY (rtp%mos%next(2*(i-1)+j)%matrix)
               CALL cp_fm_create(rtp%mos%old(2*(i-1)+j)%matrix, &
                                 matrix_struct=mos(i)%mo_set%mo_coeff%matrix_struct, &
                                 name="mos_old"//TRIM(ADJUSTL(cp_to_string(2*(i-1)+j))))
               CALL cp_fm_create(rtp%mos%new(2*(i-1)+j)%matrix, &
                                 matrix_struct=mos(i)%mo_set%mo_coeff%matrix_struct, &
                                 name="mos_new"//TRIM(ADJUSTL(cp_to_string(2*(i-1)+j))))
               CALL cp_fm_create(rtp%mos%next(2*(i-1)+j)%matrix, &
                                 matrix_struct=mos(i)%mo_set%mo_coeff%matrix_struct, &
                                 name="mos_next"//TRIM(ADJUSTL(cp_to_string(2*(i-1)+j))))
               IF (dft_control%do_admm) THEN
                  NULLIFY (rtp%mos%admm(2*(i-1)+j)%matrix)
                  CALL cp_fm_create(rtp%mos%admm(2*(i-1)+j)%matrix, &
                                    matrix_struct=mos_aux(i)%mo_set%mo_coeff%matrix_struct, &
                                    name="mos_admm"//TRIM(ADJUSTL(cp_to_string(2*(i-1)+j))))
               END IF
            END DO
         END DO
      END IF

      NULLIFY (rtp%exp_H_old)
      NULLIFY (rtp%exp_H_new)
      NULLIFY (rtp%H_last_iter)
      NULLIFY (rtp%propagator_matrix)
      CALL cp_dbcsr_allocate_matrix_set(rtp%exp_H_old, 2*nspin)
      CALL cp_dbcsr_allocate_matrix_set(rtp%exp_H_new, 2*nspin)
      CALL cp_dbcsr_allocate_matrix_set(rtp%H_last_iter, 2*nspin)
      CALL cp_dbcsr_allocate_matrix_set(rtp%propagator_matrix, 2*nspin)
      DO i = 1, 2*nspin
         CALL cp_dbcsr_init_p(rtp%exp_H_old(i)%matrix)
         CALL cp_dbcsr_create(rtp%exp_H_old(i)%matrix, template=template, matrix_type="N")
         CALL cp_dbcsr_init_p(rtp%exp_H_new(i)%matrix)
         CALL cp_dbcsr_create(rtp%exp_H_new(i)%matrix, template=template, matrix_type="N")
         CALL cp_dbcsr_init_p(rtp%H_last_iter(i)%matrix)
         CALL cp_dbcsr_create(rtp%H_last_iter(i)%matrix, template=template, matrix_type="N")
         CALL cp_dbcsr_init_p(rtp%propagator_matrix(i)%matrix)
         CALL cp_dbcsr_create(rtp%propagator_matrix(i)%matrix, template=template, matrix_type="N")
      END DO
      NULLIFY (rtp%S_inv)
      ALLOCATE (rtp%S_inv)
      CALL cp_dbcsr_init(rtp%S_inv)
      CALL cp_dbcsr_create(rtp%S_inv, template=template, matrix_type="S")
      NULLIFY (rtp%S_half)
      ALLOCATE (rtp%S_half)
      CALL cp_dbcsr_init(rtp%S_half)
      CALL cp_dbcsr_create(rtp%S_half, template=template, matrix_type="S")
      NULLIFY (rtp%S_minus_half)
      ALLOCATE (rtp%S_minus_half)
      CALL cp_dbcsr_init(rtp%S_minus_half)
      CALL cp_dbcsr_create(rtp%S_minus_half, template=template, matrix_type="S")
      NULLIFY (rtp%B_mat)
      NULLIFY (rtp%C_mat)
      NULLIFY (rtp%S_der)
      NULLIFY (rtp%SinvH)
      NULLIFY (rtp%SinvB)
      IF (.NOT. rtp_control%fixed_ions) THEN
         ALLOCATE (rtp%B_mat)
         CALL cp_dbcsr_init(rtp%B_mat)
         CALL cp_dbcsr_create(rtp%B_mat, template=template, matrix_type="N")
         CALL cp_dbcsr_allocate_matrix_set(rtp%C_mat, 3)
         CALL cp_dbcsr_allocate_matrix_set(rtp%S_der, 9)
         CALL cp_dbcsr_allocate_matrix_set(rtp%SinvH, nspin)
         CALL cp_dbcsr_allocate_matrix_set(rtp%SinvB, nspin)
         DO i = 1, nspin
            CALL cp_dbcsr_init_p(rtp%SinvH(i)%matrix)
            CALL cp_dbcsr_create(rtp%SinvH(i)%matrix, template=template, matrix_type="N")
            CALL cp_dbcsr_init_p(rtp%SinvB(i)%matrix)
            CALL cp_dbcsr_create(rtp%SinvB(i)%matrix, template=template, matrix_type="N")
         END DO
         DO i = 1, 3
            CALL cp_dbcsr_init_p(rtp%C_mat(i)%matrix)
            CALL cp_dbcsr_create(rtp%C_mat(i)%matrix, template=template, matrix_type="N")
         END DO
         DO i = 1, 9
            CALL cp_dbcsr_init_p(rtp%S_der(i)%matrix)
            CALL cp_dbcsr_create(rtp%S_der(i)%matrix, template=template, matrix_type="N")
         END DO
      END IF
      ALLOCATE (rtp%orders(2, nspin))
      rtp_control%converged = .FALSE.
      rtp%matrix_update = .TRUE.
      rtp%narn_old = 0
      rtp%istep = 0
      rtp%iter = 0
      rtp%do_hfx = .FALSE.

   END SUBROUTINE rt_prop_create

! **************************************************************************************************
!> \brief ...
!> \param rtp ...
!> \param exp_H_old ...
!> \param exp_H_new ...
!> \param H_last_iter ...
!> \param rho_old ...
!> \param rho_next ...
!> \param rho_new ...
!> \param mos ...
!> \param mos_new ...
!> \param mos_old ...
!> \param mos_next ...
!> \param S_inv ...
!> \param S_half ...
!> \param S_minus_half ...
!> \param B_mat ...
!> \param C_mat ...
!> \param propagator_matrix ...
!> \param mixing ...
!> \param mixing_factor ...
!> \param S_der ...
!> \param dt ...
!> \param nsteps ...
!> \param SinvH ...
!> \param SinvB ...
!> \param admm_mos ...
! **************************************************************************************************
   SUBROUTINE get_rtp(rtp, exp_H_old, exp_H_new, H_last_iter, rho_old, rho_next, rho_new, mos, mos_new, mos_old, mos_next, &
                      S_inv, S_half, S_minus_half, B_mat, C_mat, propagator_matrix, mixing, mixing_factor, &
                      S_der, dt, nsteps, SinvH, SinvB, admm_mos)

      TYPE(rt_prop_type), POINTER                        :: rtp
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: exp_H_old, exp_H_new, H_last_iter, &
                                                            rho_old, rho_next, rho_new
      TYPE(rtp_mos_type), OPTIONAL, POINTER              :: mos
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: mos_new, mos_old, mos_next
      TYPE(cp_dbcsr_type), OPTIONAL, POINTER             :: S_inv, S_half, S_minus_half, B_mat
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: C_mat, propagator_matrix
      LOGICAL, OPTIONAL                                  :: mixing
      REAL(dp), INTENT(out), OPTIONAL                    :: mixing_factor
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: S_der
      REAL(dp), INTENT(out), OPTIONAL                    :: dt
      INTEGER, INTENT(out), OPTIONAL                     :: nsteps
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: SinvH, SinvB
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: admm_mos

      CHARACTER(len=*), PARAMETER :: routineN = 'get_rtp', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(rtp)))CALL cp__a("rt_propagation_types.F",348)
      IF (PRESENT(exp_H_old)) exp_H_old => rtp%exp_H_old
      IF (PRESENT(exp_H_new)) exp_H_new => rtp%exp_H_new
      IF (PRESENT(H_last_iter)) H_last_iter => rtp%H_last_iter
      IF (PRESENT(propagator_matrix)) propagator_matrix => rtp%propagator_matrix

      IF (PRESENT(rho_old)) rho_old => rtp%rho%old
      IF (PRESENT(rho_next)) rho_next => rtp%rho%next
      IF (PRESENT(rho_new)) rho_new => rtp%rho%new
      IF (PRESENT(mos)) mos => rtp%mos
      IF (PRESENT(mos_old)) mos_old => rtp%mos%old
      IF (PRESENT(mos_new)) mos_new => rtp%mos%new
      IF (PRESENT(mos_next)) mos_next => rtp%mos%next
      IF (PRESENT(admm_mos)) admm_mos => rtp%mos%admm

      IF (PRESENT(S_inv)) S_inv => rtp%S_inv
      IF (PRESENT(S_half)) S_half => rtp%S_half
      IF (PRESENT(S_minus_half)) S_minus_half => rtp%S_minus_half
      IF (PRESENT(B_mat)) B_mat => rtp%B_mat
      IF (PRESENT(C_mat)) C_mat => rtp%C_mat
      IF (PRESENT(SinvH)) SinvH => rtp%SinvH
      IF (PRESENT(SinvB)) SinvB => rtp%SinvB
      IF (PRESENT(S_der)) S_der => rtp%S_der

      IF (PRESENT(dt)) dt = rtp%dt
      IF (PRESENT(mixing)) mixing = rtp%mixing
      IF (PRESENT(mixing_factor)) mixing_factor = rtp%mixing_factor
      IF (PRESENT(nsteps)) nsteps = rtp%nsteps

   END SUBROUTINE get_rtp

! **************************************************************************************************
!> \brief ...
!> \param rtp ...
! **************************************************************************************************
   SUBROUTINE rt_prop_release(rtp)
      TYPE(rt_prop_type), INTENT(inout)                  :: rtp

      CHARACTER(len=*), PARAMETER :: routineN = 'rt_prop_release', &
         routineP = moduleN//':'//routineN

      CALL cp_dbcsr_deallocate_matrix_set(rtp%exp_H_old)
      CALL cp_dbcsr_deallocate_matrix_set(rtp%exp_H_new)
      CALL cp_dbcsr_deallocate_matrix_set(rtp%H_last_iter)
      CALL cp_dbcsr_deallocate_matrix_set(rtp%propagator_matrix)
      IF (ASSOCIATED(rtp%rho)) THEN
         IF (ASSOCIATED(rtp%rho%old)) &
            CALL cp_dbcsr_deallocate_matrix_set(rtp%rho%old)
         IF (ASSOCIATED(rtp%rho%next)) &
            CALL cp_dbcsr_deallocate_matrix_set(rtp%rho%next)
         IF (ASSOCIATED(rtp%rho%new)) &
            CALL cp_dbcsr_deallocate_matrix_set(rtp%rho%new)
         DEALLOCATE (rtp%rho)
      ENDIF
      IF (ASSOCIATED(rtp%mos)) THEN
         IF (ASSOCIATED(rtp%mos%old)) &
            CALL cp_fm_vect_dealloc(rtp%mos%old)
         IF (ASSOCIATED(rtp%mos%new)) &
            CALL cp_fm_vect_dealloc(rtp%mos%new)
         IF (ASSOCIATED(rtp%mos%next)) &
            CALL cp_fm_vect_dealloc(rtp%mos%next)
         IF (ASSOCIATED(rtp%mos%admm)) &
            CALL cp_fm_vect_dealloc(rtp%mos%admm)
         DEALLOCATE (rtp%mos)
      END IF
      CALL cp_dbcsr_deallocate_matrix(rtp%S_inv)
      CALL cp_dbcsr_deallocate_matrix(rtp%S_half)
      CALL cp_dbcsr_deallocate_matrix(rtp%S_minus_half)
      IF (ASSOCIATED(rtp%B_mat)) &
         CALL cp_dbcsr_deallocate_matrix(rtp%B_mat)
      IF (ASSOCIATED(rtp%C_mat)) &
         CALL cp_dbcsr_deallocate_matrix_set(rtp%C_mat)
      IF (ASSOCIATED(rtp%S_der)) &
         CALL cp_dbcsr_deallocate_matrix_set(rtp%S_der)
      IF (ASSOCIATED(rtp%SinvH)) &
         CALL cp_dbcsr_deallocate_matrix_set(rtp%SinvH)
      IF (ASSOCIATED(rtp%SinvB)) &
         CALL cp_dbcsr_deallocate_matrix_set(rtp%SinvB)
      IF (ASSOCIATED(rtp%history)) &
         CALL rtp_history_release(rtp)
      DEALLOCATE (rtp%orders)
      IF (.NOT. rtp%linear_scaling) CALL cp_fm_struct_release(rtp%ao_ao_fmstruct)
   END SUBROUTINE rt_prop_release

! **************************************************************************************************
!> \brief ...
!> \param rtp ...
!> \param aspc_order ...
! **************************************************************************************************
   SUBROUTINE rtp_history_create(rtp, aspc_order)
      TYPE(rt_prop_type), INTENT(inout)                  :: rtp
      INTEGER, INTENT(in)                                :: aspc_order

      CHARACTER(len=*), PARAMETER :: routineN = 'rtp_history_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j, nmat
      TYPE(rtp_history_type), POINTER                    :: history

      NULLIFY (history)
      ALLOCATE (rtp%history)
      history => rtp%history

      NULLIFY (history%rho_history, history%mo_history, history%s_history)
      IF (aspc_order .GT. 0) THEN
         IF (rtp%linear_scaling) THEN
            nmat = SIZE(rtp%rho%new)
            CALL cp_dbcsr_allocate_matrix_set(history%rho_history, nmat, aspc_order)
            DO i = 1, nmat
               DO j = 1, aspc_order
                  CALL cp_dbcsr_init_p(history%rho_history(i, j)%matrix)
                  CALL cp_dbcsr_create(history%rho_history(i, j)%matrix, &
                                       name="rho_hist"//TRIM(ADJUSTL(cp_to_string(i))), &
                                       template=rtp%rho%new(1)%matrix)
               END DO
            END DO
         ELSE
            nmat = SIZE(rtp%mos%old)
            ALLOCATE (history%mo_history(nmat, aspc_order))
            DO i = 1, nmat
               DO j = 1, aspc_order
                  NULLIFY (history%mo_history(i, j)%matrix)
                  CALL cp_fm_create(history%mo_history(i, j)%matrix, &
                                    matrix_struct=rtp%mos%new(i)%matrix%matrix_struct, &
                                    name="mo_hist"//TRIM(ADJUSTL(cp_to_string(i))))
               END DO
            END DO
            ALLOCATE (history%s_history(aspc_order))
            DO i = 1, aspc_order
               NULLIFY (history%s_history(i)%matrix)
            END DO
         END IF
      END IF

   END SUBROUTINE rtp_history_create

! **************************************************************************************************
!> \brief ...
!> \param rtp ...
! **************************************************************************************************
   SUBROUTINE rtp_history_release(rtp)
      TYPE(rt_prop_type), INTENT(inout)                  :: rtp

      CHARACTER(len=*), PARAMETER :: routineN = 'rtp_history_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      IF (ASSOCIATED(rtp%history%rho_history)) THEN
         CALL cp_dbcsr_deallocate_matrix_set(rtp%history%rho_history)
      END IF

      IF (ASSOCIATED(rtp%history%mo_history)) THEN
         DO i = 1, SIZE(rtp%history%mo_history, 1)
            DO j = 1, SIZE(rtp%history%mo_history, 2)
               CALL cp_fm_release(rtp%history%mo_history(i, j)%matrix)
            END DO
         END DO
         DEALLOCATE (rtp%history%mo_history)
      END IF
      IF (ASSOCIATED(rtp%history%s_history)) THEN
         DO i = 1, SIZE(rtp%history%s_history)
            IF (ASSOCIATED(rtp%history%s_history(i)%matrix)) &
               CALL cp_dbcsr_deallocate_matrix(rtp%history%s_history(i)%matrix)
         END DO
         DEALLOCATE (rtp%history%s_history)
      END IF
      DEALLOCATE (rtp%history)

   END SUBROUTINE rtp_history_release

END MODULE rt_propagation_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Definition and initialisation of the et_coupling data type.
!> \author Florian Schiffmann (01.2007,fschiff)
! **************************************************************************************************
MODULE et_coupling_types

   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'et_coupling_types'

! *** Public data types ***

   PUBLIC :: et_coupling_type

! *** Public subroutines ***

   PUBLIC :: et_coupling_create, &
             et_coupling_release, &
             set_et_coupling_type

! **************************************************************************************************
!> \par History
!>      01.2007 created [Florian Schiffmann]
!> \author fschiff
! **************************************************************************************************
   TYPE et_coupling_type
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER           :: et_mo_coeff
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER     :: rest_mat
      LOGICAL                                            :: first_run
      LOGICAL                                            :: keep_matrix
      REAL(KIND=dp)                                    :: energy, e1, order_p
   END TYPE

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param et_coupling ...
! **************************************************************************************************
   SUBROUTINE et_coupling_create(et_coupling)
      TYPE(et_coupling_type), POINTER                    :: et_coupling

      CHARACTER(len=*), PARAMETER :: routineN = 'et_coupling_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (et_coupling)

      NULLIFY (et_coupling%et_mo_coeff)
      NULLIFY (et_coupling%rest_mat)
      et_coupling%first_run = .TRUE.
      et_coupling%keep_matrix = .FALSE.
      ALLOCATE (et_coupling%rest_mat(2))

   END SUBROUTINE et_coupling_create

! **************************************************************************************************
!> \brief ...
!> \param et_coupling ...
!> \param et_mo_coeff ...
!> \param rest_mat ...
! **************************************************************************************************
   SUBROUTINE get_et_coupling_type(et_coupling, et_mo_coeff, rest_mat)
      TYPE(et_coupling_type), POINTER                    :: et_coupling
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: et_mo_coeff
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rest_mat

      CHARACTER(len=*), PARAMETER :: routineN = 'get_et_coupling_type', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(et_mo_coeff)) et_mo_coeff => et_coupling%et_mo_coeff
      IF (PRESENT(rest_mat)) rest_mat => et_coupling%rest_mat

   END SUBROUTINE get_et_coupling_type

! **************************************************************************************************
!> \brief ...
!> \param et_coupling ...
!> \param et_mo_coeff ...
!> \param rest_mat ...
! **************************************************************************************************
   SUBROUTINE set_et_coupling_type(et_coupling, et_mo_coeff, rest_mat)
      TYPE(et_coupling_type), POINTER                    :: et_coupling
      TYPE(cp_fm_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: et_mo_coeff
      TYPE(cp_dbcsr_p_type), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: rest_mat

      CHARACTER(len=*), PARAMETER :: routineN = 'set_et_coupling_type', &
         routineP = moduleN//':'//routineN

      IF (PRESENT(et_mo_coeff)) et_coupling%et_mo_coeff = et_mo_coeff
      IF (PRESENT(rest_mat)) et_coupling%rest_mat => rest_mat

   END SUBROUTINE set_et_coupling_type

! **************************************************************************************************
!> \brief ...
!> \param et_coupling ...
! **************************************************************************************************
   SUBROUTINE et_coupling_release(et_coupling)
      TYPE(et_coupling_type), POINTER                    :: et_coupling

      CHARACTER(LEN=*), PARAMETER :: routineN = 'et_coupling_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(et_coupling%et_mo_coeff)) THEN
         DO i = 1, SIZE(et_coupling%et_mo_coeff)
            CALL cp_fm_release(et_coupling%et_mo_coeff(i)%matrix)
         END DO
         DEALLOCATE (et_coupling%et_mo_coeff)
      END IF
      IF (ASSOCIATED(et_coupling%rest_mat)) THEN
!         CALL deallocate_matrix_set(et_coupling%rest_mat)
         DEALLOCATE (et_coupling%rest_mat)
      END IF

      DEALLOCATE (et_coupling)
   END SUBROUTINE et_coupling_release

END MODULE et_coupling_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Setting up the Spline coefficients used to Interpolate the G-Term
!>      in Ewald sums
!> \par History
!>      12.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
MODULE ewald_spline_util
   USE cell_types,                      ONLY: cell_create,&
                                              cell_release,&
                                              cell_type
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE message_passing,                 ONLY: mp_sum
   USE pw_grid_types,                   ONLY: HALFSPACE,&
                                              pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_create,&
                                              pw_grid_setup
   USE pw_methods,                      ONLY: pw_zero
   USE pw_pool_types,                   ONLY: pw_pool_create,&
                                              pw_pool_create_pw,&
                                              pw_pool_give_back_pw,&
                                              pw_pool_type
   USE pw_spline_utils,                 ONLY: &
        Eval_Interp_Spl3_pbc, Eval_d_Interp_Spl3_pbc, find_coeffs, pw_spline_do_precond, &
        pw_spline_precond_create, pw_spline_precond_release, pw_spline_precond_set_kind, &
        pw_spline_precond_type, spl3_pbc
   USE pw_types,                        ONLY: REALDATA3D,&
                                              REALSPACE,&
                                              pw_type
!NB parallelization
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ewald_spline_util'
   PUBLIC :: Setup_Ewald_Spline

CONTAINS

! **************************************************************************************************
!> \brief Setup of the G-space Ewald Term Spline Coefficients
!> \param pw_grid ...
!> \param pw_pool ...
!> \param coeff ...
!> \param LG ...
!> \param gx ...
!> \param gy ...
!> \param gz ...
!> \param hmat ...
!> \param npts ...
!> \param param_section ...
!> \param tag ...
!> \param print_section ...
!> \param para_env ...
!> \par History
!>      12.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE Setup_Ewald_Spline(pw_grid, pw_pool, coeff, LG, gx, gy, gz, hmat, npts, &
                                 param_section, tag, print_section, para_env)
      TYPE(pw_grid_type), POINTER                        :: pw_grid
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(pw_type), POINTER                             :: coeff
      REAL(KIND=dp), DIMENSION(:), POINTER               :: LG, gx, gy, gz
      REAL(KIND=dp), INTENT(IN)                          :: hmat(3, 3)
      INTEGER, INTENT(IN)                                :: npts(3)
      TYPE(section_vals_type), POINTER                   :: param_section
      CHARACTER(LEN=*), INTENT(IN)                       :: tag
      TYPE(section_vals_type), POINTER                   :: print_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'Setup_Ewald_Spline', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: bo(2, 3), iounit
      TYPE(cell_type), POINTER                           :: cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(pw_type), POINTER                             :: pw

!
! Setting Up Fit Procedure
!

      IF(.NOT.(.NOT. ASSOCIATED(pw_grid)))CALL cp__a("ewald_spline_util.F",99)
      IF(.NOT.(.NOT. ASSOCIATED(pw_pool)))CALL cp__a("ewald_spline_util.F",100)
      IF(.NOT.(.NOT. ASSOCIATED(coeff)))CALL cp__a("ewald_spline_util.F",101)
      NULLIFY (cell, pw)

      CALL cell_create(cell, hmat=hmat, periodic=(/1, 1, 1/))
      CALL pw_grid_create(pw_grid, para_env%group, local=.TRUE.)
      logger => cp_get_default_logger()
      iounit = cp_print_key_unit_nr(logger, print_section, "", &
                                    extension=".Log")
      bo(1, 1:3) = 0
      bo(2, 1:3) = npts(1:3)-1
      CALL pw_grid_setup(cell%hmat, pw_grid, grid_span=HALFSPACE, bounds=bo, iounit=iounit)

      CALL cp_print_key_finished_output(iounit, logger, print_section, &
                                        "")
      ! pw_pool initialized
      CALL pw_pool_create(pw_pool, pw_grid=pw_grid)
      CALL pw_pool_create_pw(pw_pool, pw, use_data=REALDATA3D, in_space=REALSPACE)
      CALL pw_pool_create_pw(pw_pool, coeff, use_data=REALDATA3D, in_space=REALSPACE)
      ! Evaluate function on grid
      CALL eval_pw_TabLR(pw, pw_pool, coeff, Lg, gx, gy, gz, hmat_mm=hmat, &
                         param_section=param_section, tag=tag)
      CALL pw_pool_give_back_pw(pw_pool, pw)
      CALL cell_release(cell)

   END SUBROUTINE Setup_Ewald_Spline

! **************************************************************************************************
!> \brief Evaluates the function G-Term in reciprocal space on the grid
!>      and find the coefficients of the Splines
!> \param grid ...
!> \param pw_pool ...
!> \param TabLR ...
!> \param Lg ...
!> \param gx ...
!> \param gy ...
!> \param gz ...
!> \param hmat_mm ...
!> \param param_section ...
!> \param tag ...
!> \par History
!>      12.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE eval_pw_TabLR(grid, pw_pool, TabLR, Lg, gx, gy, gz, hmat_mm, &
                            param_section, tag)
      TYPE(pw_type), POINTER                             :: grid
      TYPE(pw_pool_type), POINTER                        :: pw_pool
      TYPE(pw_type), POINTER                             :: TabLR
      REAL(KIND=dp), DIMENSION(:), POINTER               :: Lg, gx, gy, gz
      REAL(KIND=dp), DIMENSION(3, 3)                     :: hmat_mm
      TYPE(section_vals_type), POINTER                   :: param_section
      CHARACTER(LEN=*), INTENT(IN)                       :: tag

      CHARACTER(len=*), PARAMETER :: routineN = 'eval_pw_TabLR', routineP = moduleN//':'//routineN

      INTEGER :: act_nx, act_ny, aint_precond, handle, i, iii, is, j, js, k, ks, Lg_loc_max, &
         Lg_loc_min, max_iter, my_i, my_j, my_k, n1, n2, n3, n_extra, NLg_loc, nxlim, nylim, &
         nzlim, precond_kind
      INTEGER, DIMENSION(2, 3)                           :: gbo
      LOGICAL                                            :: success
      REAL(KIND=dp)                                      :: dr1, dr2, dr3, eps_r, eps_x, xs1, xs2, &
                                                            xs3
      REAL(KIND=dp), ALLOCATABLE                         :: cos_gx(:, :), cos_gy(:, :), &
                                                            cos_gz(:, :), lhs(:, :), rhs(:, :), &
                                                            sin_gx(:, :), sin_gy(:, :), &
                                                            sin_gz(:, :)
      TYPE(pw_spline_precond_type), POINTER              :: precond
      TYPE(section_vals_type), POINTER                   :: interp_section

!NB pull expensive Cos() out of inner looop
!NB temporaries for holding stuff so that dgemm can be used

      EXTERNAL :: DGEMM

      CALL timeset(routineN, handle)
      n1 = grid%pw_grid%npts(1)
      n2 = grid%pw_grid%npts(2)
      n3 = grid%pw_grid%npts(3)
      dr1 = grid%pw_grid%dr(1)
      dr2 = grid%pw_grid%dr(2)
      dr3 = grid%pw_grid%dr(3)
      gbo = grid%pw_grid%bounds
      nxlim = FLOOR(REAL(n1, KIND=dp)/2.0_dp)
      nylim = FLOOR(REAL(n2, KIND=dp)/2.0_dp)
      nzlim = FLOOR(REAL(n3, KIND=dp)/2.0_dp)
      is = 0
      js = 0
      ks = 0
      IF (2*nxlim /= n1) is = 1
      IF (2*nylim /= n2) js = 1
      IF (2*nzlim /= n3) ks = 1
      CALL pw_zero(grid)

      ! Used the full symmetry to reduce the evaluation to 1/64th
      !NB parallelization
      iii = 0
      !NB allocate temporaries for Cos refactoring
      ALLOCATE (cos_gx(SIZE(Lg), gbo(1, 1):gbo(2, 1)))
      ALLOCATE (sin_gx(SIZE(Lg), gbo(1, 1):gbo(2, 1)))
      ALLOCATE (cos_gy(SIZE(Lg), gbo(1, 2):gbo(2, 2)))
      ALLOCATE (sin_gy(SIZE(Lg), gbo(1, 2):gbo(2, 2)))
      ALLOCATE (cos_gz(SIZE(Lg), gbo(1, 3):gbo(2, 3)))
      ALLOCATE (sin_gz(SIZE(Lg), gbo(1, 3):gbo(2, 3)))
      !NB precalculate Cos(gx*xs1) etc for Cos refactoring
      DO k = gbo(1, 3), gbo(2, 3)
         my_k = k-gbo(1, 3)
         xs3 = REAL(my_k, dp)*dr3
         IF (k > nzlim) CYCLE
         cos_gz(1:SIZE(Lg), k) = COS(gz(1:SIZE(Lg))*xs3)
         sin_gz(1:SIZE(Lg), k) = SIN(gz(1:SIZE(Lg))*xs3)
      END DO ! k
      xs2 = 0.0_dp
      DO j = gbo(1, 2), gbo(2, 2)
         IF (j > nylim) CYCLE
         cos_gy(1:SIZE(Lg), j) = COS(gy(1:SIZE(Lg))*xs2)
         sin_gy(1:SIZE(Lg), j) = SIN(gy(1:SIZE(Lg))*xs2)
         xs2 = xs2+dr2
      END DO ! j
      xs1 = 0.0_dp
      DO i = gbo(1, 1), gbo(2, 1)
         IF (i > nxlim) CYCLE
         cos_gx(1:SIZE(Lg), i) = COS(gx(1:SIZE(Lg))*xs1)
         sin_gx(1:SIZE(Lg), i) = SIN(gx(1:SIZE(Lg))*xs1)
         xs1 = xs1+dr1
      END DO ! i

      !NB use DGEMM to compute sum over kg for each i, j, k
      ! number of elements per node, round down
      NLg_loc = SIZE(Lg)/grid%pw_grid%para%group_size
      ! number of extra elements not yet accounted for
      n_extra = MOD(SIZE(Lg), grid%pw_grid%para%group_size)
      ! first n_extra nodes get NLg_loc+1, remaining get NLg_loc
      IF (grid%pw_grid%para%my_pos < n_extra) THEN
         Lg_loc_min = (NLg_loc+1)*grid%pw_grid%para%my_pos+1
         Lg_loc_max = Lg_loc_min+(NLg_loc+1)-1
      ELSE
         Lg_loc_min = (NLg_loc+1)*n_extra+NLg_loc*(grid%pw_grid%para%my_pos-n_extra)+1
         Lg_loc_max = Lg_loc_min+NLg_loc-1
      END IF
      ! shouldn't be necessary
      Lg_loc_max = MIN(SIZE(Lg), Lg_loc_max)
      NLg_loc = Lg_loc_max-Lg_loc_min+1

      IF (NLg_loc > 0) THEN ! some work for this node

         act_nx = MIN(gbo(2, 1), nxlim)-gbo(1, 1)+1
         act_ny = MIN(gbo(2, 2), nylim)-gbo(1, 2)+1
         !NB temporaries for DGEMM use
         ALLOCATE (lhs(act_nx, NLg_loc))
         ALLOCATE (rhs(act_ny, NLg_loc))

         ! do cos(gx) cos(gy+gz) term
         DO i = gbo(1, 1), gbo(2, 1)
            IF (i > nxlim) CYCLE
            lhs(i-gbo(1, 1)+1, 1:NLg_loc) = lg(Lg_loc_min:Lg_loc_max)*cos_gx(Lg_loc_min:Lg_loc_max, i)
         END DO
         DO k = gbo(1, 3), gbo(2, 3)
            IF (k > nzlim) CYCLE
            DO j = gbo(1, 2), gbo(2, 2)
               IF (j > nylim) CYCLE
               rhs(j-gbo(1, 2)+1, 1:NLg_loc) = cos_gy(Lg_loc_min:Lg_loc_max, j)*cos_gz(Lg_loc_min:Lg_loc_max, k)- &
                                               sin_gy(Lg_loc_min:Lg_loc_max, j)*sin_gz(Lg_loc_min:Lg_loc_max, k)
            END DO
            CALL DGEMM('N', 'T', act_nx, act_ny, NLg_loc, 1.0D0, lhs(1, 1), act_nx, rhs(1, 1), act_ny, 0.0D0, &
                       grid%cr3d(gbo(1, 1), gbo(1, 2), k), SIZE(grid%cr3d, 1))
         END DO

         ! do sin(gx) sin(gy+gz) term
         DO i = gbo(1, 1), gbo(2, 1)
            IF (i > nxlim) CYCLE
            lhs(i-gbo(1, 1)+1, 1:NLg_loc) = -lg(Lg_loc_min:Lg_loc_max)*sin_gx(Lg_loc_min:Lg_loc_max, i)
         END DO
         DO k = gbo(1, 3), gbo(2, 3)
            IF (k > nzlim) CYCLE
            DO j = gbo(1, 2), gbo(2, 2)
               IF (j > nylim) CYCLE
               rhs(j-gbo(1, 2)+1, 1:NLg_loc) = cos_gy(Lg_loc_min:Lg_loc_max, j)*sin_gz(Lg_loc_min:Lg_loc_max, k)+ &
                                               sin_gy(Lg_loc_min:Lg_loc_max, j)*cos_gz(Lg_loc_min:Lg_loc_max, k)
            END DO
            CALL DGEMM('N', 'T', act_nx, act_ny, NLg_loc, 1.0D0, lhs(1, 1), act_nx, rhs(1, 1), act_ny, 1.0D0, &
                       grid%cr3d(gbo(1, 1), gbo(1, 2), k), SIZE(grid%cr3d, 1))
         END DO

         !NB deallocate temporaries for DGEMM use
         DEALLOCATE (lhs)
         DEALLOCATE (rhs)
         !NB deallocate temporaries for Cos refactoring
         DEALLOCATE (cos_gx)
         DEALLOCATE (sin_gx)
         DEALLOCATE (cos_gy)
         DEALLOCATE (sin_gy)
         DEALLOCATE (cos_gz)
         DEALLOCATE (sin_gz)
         !NB parallelization
      ELSE ! no work for this node, just zero contribution
         grid%cr3d(gbo(1, 1):nxlim, gbo(1, 2):nylim, gbo(1, 3):nzlim) = 0.0_dp
      ENDIF ! NLg_loc > 0

      CALL mp_sum(grid%cr3d(gbo(1, 1):nxlim, gbo(1, 2):nylim, gbo(1, 3):nzlim), grid%pw_grid%para%group)

      Fake_LoopOnGrid: DO k = gbo(1, 3), gbo(2, 3)
         my_k = k
         IF (k > nzlim) my_k = nzlim-ABS(nzlim-k)+ks
         DO j = gbo(1, 2), gbo(2, 2)
            my_j = j
            IF (j > nylim) my_j = nylim-ABS(nylim-j)+js
            DO i = gbo(1, 1), gbo(2, 1)
               my_i = i
               IF (i > nxlim) my_i = nxlim-ABS(nxlim-i)+is
               grid%cr3d(i, j, k) = grid%cr3d(my_i, my_j, my_k)
            END DO
         END DO
      END DO Fake_LoopOnGrid
      !
      ! Solve for spline coefficients
      !
      interp_section => section_vals_get_subs_vals(param_section, "INTERPOLATOR")
      CALL section_vals_val_get(interp_section, "aint_precond", i_val=aint_precond)
      CALL section_vals_val_get(interp_section, "precond", i_val=precond_kind)
      CALL section_vals_val_get(interp_section, "max_iter", i_val=max_iter)
      CALL section_vals_val_get(interp_section, "eps_r", r_val=eps_r)
      CALL section_vals_val_get(interp_section, "eps_x", r_val=eps_x)
      !
      ! Solve for spline coefficients
      !
      CALL pw_spline_precond_create(precond, precond_kind=aint_precond, &
                                    pool=pw_pool, pbc=.TRUE., transpose=.FALSE.)
      CALL pw_spline_do_precond(precond, grid, TabLR)
      CALL pw_spline_precond_set_kind(precond, precond_kind)
      success = find_coeffs(values=grid, coeffs=TabLR, &
                            linOp=spl3_pbc, preconditioner=precond, pool=pw_pool, &
                            eps_r=eps_r, eps_x=eps_x, &
                            max_iter=max_iter)
      IF(.NOT.(success))CALL cp__a("ewald_spline_util.F",334)
      CALL pw_spline_precond_release(precond)
      !
      ! Check for the interpolation Spline
      !
      CALL check_spline_interp_TabLR(hmat_mm, Lg, gx, gy, gz, TabLR, param_section, &
                                     tag)
      CALL timestop(handle)
   END SUBROUTINE eval_pw_TabLR

! **************************************************************************************************
!> \brief Routine to check the accuracy for the Spline Interpolation
!> \param hmat_mm ...
!> \param Lg ...
!> \param gx ...
!> \param gy ...
!> \param gz ...
!> \param TabLR ...
!> \param param_section ...
!> \param tag ...
!> \par History
!>      12.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE check_spline_interp_TabLR(hmat_mm, Lg, gx, gy, gz, TabLR, &
                                        param_section, tag)
      REAL(KIND=dp), DIMENSION(3, 3)                     :: hmat_mm
      REAL(KIND=dp), DIMENSION(:), POINTER               :: Lg, gx, gy, gz
      TYPE(pw_type), POINTER                             :: TabLR
      TYPE(section_vals_type), POINTER                   :: param_section
      CHARACTER(LEN=*), INTENT(IN)                       :: tag

      CHARACTER(len=*), PARAMETER :: routineN = 'check_spline_interp_TabLR', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i, iw, kg, npoints
      REAL(KIND=dp)                                      :: dn(3), dr1, dr2, dr3, dxTerm, dyTerm, &
                                                            dzTerm, errd, errf, Fterm, maxerrord, &
                                                            maxerrorf, Na, Nn, Term, tmp1, tmp2, &
                                                            vec(3), xs1, xs2, xs3
      TYPE(cp_logger_type), POINTER                      :: logger

      NULLIFY (logger)
      logger => cp_get_default_logger()
      iw = cp_print_key_unit_nr(logger, param_section, "check_spline", &
                                extension="."//TRIM(tag)//"Log")
      CALL timeset(routineN, handle)
      IF (iw > 0) THEN
         npoints = 100
         errf = 0.0_dp
         maxerrorf = 0.0_dp
         errd = 0.0_dp
         maxerrord = 0.0_dp
         dr1 = hmat_mm(1, 1)/REAL(npoints, KIND=dp)
         dr2 = hmat_mm(2, 2)/REAL(npoints, KIND=dp)
         dr3 = hmat_mm(3, 3)/REAL(npoints, KIND=dp)
         xs1 = 0.0_dp
         xs2 = 0.0_dp
         xs3 = 0.0_dp
         WRITE (iw, '(A,T5,A15,4X,A17,T50,4X,A,5X,A,T80,A,T85,A15,4X,A17,T130,4X,A,5X,A)') &
            "#", "Analytical Term", "Interpolated Term", "Error", "MaxError", &
            "*", " Analyt Deriv  ", "Interp Deriv Mod ", "Error", "MaxError"
         DO i = 1, npoints+1
            Term = 0.0_dp
            dxTerm = 0.0_dp
            dyTerm = 0.0_dp
            dzTerm = 0.0_dp
            ! Sum over k vectors
            DO kg = 1, SIZE(Lg)
               vec = (/REAL(gx(kg), KIND=dp), REAL(gy(kg), KIND=dp), REAL(gz(kg), KIND=dp)/)
               Term = Term+lg(kg)*COS(vec(1)*xs1+vec(2)*xs2+vec(3)*xs3)
               dxTerm = dxTerm-lg(kg)*SIN(vec(1)*xs1+vec(2)*xs2+vec(3)*xs3)*vec(1)
               dyTerm = dyTerm-lg(kg)*SIN(vec(1)*xs1+vec(2)*xs2+vec(3)*xs3)*vec(2)
               dzTerm = dzTerm-lg(kg)*SIN(vec(1)*xs1+vec(2)*xs2+vec(3)*xs3)*vec(3)
            END DO
            Na = SQRT(dxTerm*dxTerm+dyTerm*dyTerm+dzTerm*dzTerm)
            dn = Eval_d_Interp_Spl3_pbc((/xs1, xs2, xs3/), TabLR)
            Nn = SQRT(DOT_PRODUCT(dn, dn))
            Fterm = Eval_Interp_Spl3_pbc((/xs1, xs2, xs3/), TabLR)
            tmp1 = ABS(Term-Fterm)
            tmp2 = SQRT(DOT_PRODUCT(dn-(/dxTerm, dyTerm, dzTerm/), dn-(/dxTerm, dyTerm, dzTerm/)))
            errf = errf+tmp1
            maxerrorf = MAX(maxerrorf, tmp1)
            errd = errd+tmp2
            maxerrord = MAX(maxerrord, tmp2)
            WRITE (iw, '(T5,F15.10,5X,F15.10,T50,2F12.9,T80,A,T85,F15.10,5X,F15.10,T130,2F12.9)') &
               Term, Fterm, tmp1, maxerrorf, "*", Na, Nn, tmp2, maxerrord
            xs1 = xs1+dr1
            xs2 = xs2+dr2
            xs3 = xs3+dr3
         END DO
         WRITE (iw, '(A,T5,A,T50,F12.9,T130,F12.9)') "#", "Averages", errf/REAL(npoints, kind=dp), &
            errd/REAL(npoints, kind=dp)
      END IF
      CALL timestop(handle)
      CALL cp_print_key_finished_output(iw, logger, param_section, "check_spline")

   END SUBROUTINE check_spline_interp_TabLR

END MODULE ewald_spline_util

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Handles all functions related to the CELL
!> \par History
!>      11.2008 Teodoro Laino [tlaino] - deeply cleaning cell_type from units
!>      10.2014 Moved many routines to cell_types.F.
!> \author Matthias KracK (16.01.2002, based on a earlier version of CJM, JGH)
! **************************************************************************************************
MODULE cell_methods
   USE cell_types,                      ONLY: &
        cell_clone, cell_create, cell_sym_none, cell_type, get_cell, init_cell, set_cell_param, &
        use_perd_none, use_perd_x, use_perd_xy, use_perd_xyz, use_perd_xz, use_perd_y, &
        use_perd_yz, use_perd_z
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_parser_methods,               ONLY: parser_get_next_line
   USE cp_parser_types,                 ONLY: cp_parser_type,&
                                              parser_create,&
                                              parser_release
   USE cp_units,                        ONLY: cp_unit_from_cp2k,&
                                              cp_unit_to_cp2k
   USE input_constants,                 ONLY: do_cell_cp2k,&
                                              do_cell_xsc
   USE input_cp2k_subsys,               ONLY: create_cell_section
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: &
        section_get_keyword, section_release, section_type, section_vals_get, &
        section_vals_get_subs_vals, section_vals_type, section_vals_val_get, section_vals_val_set, &
        section_vals_val_unset
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE mathconstants,                   ONLY: degree
   USE mathlib,                         ONLY: angle
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cell_methods'

   ! Public subroutines
   PUBLIC :: read_cell, write_cell

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cell ...
!> \param cell_ref ...
!> \param use_ref_cell ...
!> \param cell_section ...
!> \param check_for_ref ...
!> \param para_env ...
!> \par History
!>      03.2005 created [teo]
!> \author Teodoro Laino
! **************************************************************************************************
   RECURSIVE SUBROUTINE read_cell(cell, cell_ref, use_ref_cell, cell_section, &
                                  check_for_ref, para_env)
      TYPE(cell_type), POINTER                           :: cell, cell_ref
      LOGICAL, INTENT(OUT), OPTIONAL                     :: use_ref_cell
      TYPE(section_vals_type), OPTIONAL, POINTER         :: cell_section
      LOGICAL, INTENT(IN), OPTIONAL                      :: check_for_ref
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'read_cell', routineP = moduleN//':'//routineN

      INTEGER                                            :: my_per
      INTEGER, DIMENSION(:), POINTER                     :: multiple_unit_cell
      LOGICAL                                            :: cell_read_a, cell_read_abc, cell_read_b, &
                                                            cell_read_c, cell_read_file, check, &
                                                            my_check
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cell_angles, cell_par
      TYPE(section_vals_type), POINTER                   :: cell_ref_section

      my_check = .TRUE.
      NULLIFY (cell_ref_section, cell_par, multiple_unit_cell)
      IF (.NOT. ASSOCIATED(cell)) CALL cell_create(cell)
      IF (.NOT. ASSOCIATED(cell_ref)) CALL cell_create(cell_ref)
      IF (PRESENT(check_for_ref)) my_check = check_for_ref

      cell%deth = 0.0_dp
      cell%orthorhombic = .FALSE.
      cell%perd(:) = 1
      cell%symmetry_id = cell_sym_none
      cell%hmat(:, :) = 0.0_dp
      cell%h_inv(:, :) = 0.0_dp
      cell_read_file = .FALSE.
      cell_read_a = .FALSE.
      cell_read_b = .FALSE.
      cell_read_c = .FALSE.
      ! Trying to read cell info from file
      CALL section_vals_val_get(cell_section, "CELL_FILE_NAME", explicit=cell_read_file)
      IF (cell_read_file) CALL read_cell_from_external_file(cell_section, para_env)

      ! Trying to read cell info from the separate A, B, C vectors
      ! If cell information is provided through file A,B,C contain the file information..
      ! a print warning is shown on screen..
      CALL section_vals_val_get(cell_section, "A", explicit=cell_read_a)
      IF (cell_read_a) THEN
         CALL section_vals_val_get(cell_section, "A", r_vals=cell_par)
         cell%hmat(:, 1) = cell_par(:)
      END IF
      CALL section_vals_val_get(cell_section, "B", explicit=cell_read_b)
      IF (cell_read_b) THEN
         CALL section_vals_val_get(cell_section, "B", r_vals=cell_par)
         cell%hmat(:, 2) = cell_par(:)
      END IF
      CALL section_vals_val_get(cell_section, "C", explicit=cell_read_c)
      IF (cell_read_c) THEN
         CALL section_vals_val_get(cell_section, "C", r_vals=cell_par)
         cell%hmat(:, 3) = cell_par(:)
      END IF
      check = ((cell_read_a .EQV. cell_read_b) .AND. (cell_read_b .EQV. cell_read_c))
      IF (.NOT. check) &
         CALL cp_warn(cp__l("cell_methods.F",128), &
                      "Cell Information provided through vectors A, B and C. Not all three "// &
                      "vectors were provided! Cell setup may be incomplete!")

      ! Very last option.. Trying to read cell info from ABC keyword
      CALL section_vals_val_get(cell_section, "ABC", explicit=cell_read_abc)
      IF (cell_read_abc) THEN
         check = (cell_read_a .OR. cell_read_b .OR. cell_read_c)
         IF (check) &
            CALL cp_warn(cp__l("cell_methods.F",137), &
                         "Cell Information provided through vectors A, B and C in conjunction with ABC."// &
                         " The definition of the ABC keyword will override the one provided by A,B and C.")
         cell%hmat = 0.0_dp
         CALL section_vals_val_get(cell_section, "ABC", r_vals=cell_par)
         CALL section_vals_val_get(cell_section, "ALPHA_BETA_GAMMA", r_vals=cell_angles)
         CALL set_cell_param(cell, cell_par, cell_angles, do_init_cell=.FALSE.)
      END IF

      ! Multiple unit cell
      CALL section_vals_val_get(cell_section, "MULTIPLE_UNIT_CELL", i_vals=multiple_unit_cell)
      IF (ANY(multiple_unit_cell /= 1)) CALL set_multiple_unit_cell(cell, multiple_unit_cell)

      CALL section_vals_val_get(cell_section, "PERIODIC", i_val=my_per)
      SELECT CASE (my_per)
      CASE (use_perd_x)
         cell%perd = (/1, 0, 0/)
      CASE (use_perd_y)
         cell%perd = (/0, 1, 0/)
      CASE (use_perd_z)
         cell%perd = (/0, 0, 1/)
      CASE (use_perd_xy)
         cell%perd = (/1, 1, 0/)
      CASE (use_perd_xz)
         cell%perd = (/1, 0, 1/)
      CASE (use_perd_yz)
         cell%perd = (/0, 1, 1/)
      CASE (use_perd_xyz)
         cell%perd = (/1, 1, 1/)
      CASE (use_perd_none)
         cell%perd = (/0, 0, 0/)
      CASE DEFAULT
         CALL cp__b("cell_methods.F",169,"")
      END SELECT

      ! Load requested cell symmetry
      CALL section_vals_val_get(cell_section, "SYMMETRY", i_val=cell%symmetry_id)

      ! Initialize cell
      CALL init_cell(cell)

      IF (.NOT. my_check) RETURN
      cell_ref_section => section_vals_get_subs_vals(cell_section, &
                                                     "CELL_REF")
      IF (parsed_cp2k_input(cell_ref_section, check_this_section=.TRUE.)) THEN
         IF (PRESENT(use_ref_cell)) use_ref_cell = .TRUE.
         CALL read_cell(cell_ref, cell_ref, use_ref_cell, cell_section=cell_ref_section, &
                        check_for_ref=.FALSE., para_env=para_env)
      ELSE
         CALL cell_clone(cell, cell_ref)
         IF (PRESENT(use_ref_cell)) use_ref_cell = .FALSE.
      END IF

   END SUBROUTINE read_cell

! **************************************************************************************************
!> \brief utility function to ease the transition to the new input.
!>      returns true if the new input was parsed
!> \param input_file the parsed input file
!> \param check_this_section ...
!> \retval res ...
!> \author fawzi
! **************************************************************************************************
   FUNCTION parsed_cp2k_input(input_file, check_this_section) RESULT(res)
      TYPE(section_vals_type), POINTER                   :: input_file
      LOGICAL, INTENT(IN), OPTIONAL                      :: check_this_section
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'parsed_cp2k_input', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: my_check
      TYPE(section_vals_type), POINTER                   :: glob_section

      my_check = .FALSE.
      IF (PRESENT(check_this_section)) my_check = check_this_section
      res = ASSOCIATED(input_file)
      IF (res) THEN
         IF(.NOT.(input_file%ref_count > 0))CALL cp__a("cell_methods.F",215)
         IF (.NOT. my_check) THEN
            glob_section => section_vals_get_subs_vals(input_file, "GLOBAL")
            CALL section_vals_get(glob_section, explicit=res)
         ELSE
            CALL section_vals_get(input_file, explicit=res)
         END IF
      END IF
   END FUNCTION parsed_cp2k_input

! **************************************************************************************************
!> \brief   Setup of the multiple unit_cell
!> \param cell ...
!> \param multiple_unit_cell ...
!> \date    05.2009
!> \author  Teodoro Laino [tlaino]
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE set_multiple_unit_cell(cell, multiple_unit_cell)

      TYPE(cell_type), POINTER                           :: cell
      INTEGER, DIMENSION(:), POINTER                     :: multiple_unit_cell

      CHARACTER(len=*), PARAMETER :: routineN = 'set_multiple_unit_cell', &
         routineP = moduleN//':'//routineN

! Fail is one of the value is set to zero..

      IF (ANY(multiple_unit_cell <= 0)) &
         CALL cp_abort(cp__l("cell_methods.F",244), &
                       "CELL%MULTIPLE_UNIT_CELL accepts only integer values larger than 0! "// &
                       "A value of 0 or negative is meaningless!")

      ! scale abc accordingly user request
      cell%hmat(:, 1) = cell%hmat(:, 1)*multiple_unit_cell(1)
      cell%hmat(:, 2) = cell%hmat(:, 2)*multiple_unit_cell(2)
      cell%hmat(:, 3) = cell%hmat(:, 3)*multiple_unit_cell(3)

   END SUBROUTINE set_multiple_unit_cell

! **************************************************************************************************
!> \brief   Read cell information from an external file
!> \param cell_section ...
!> \param para_env ...
!> \date    02.2008
!> \author  Teodoro Laino [tlaino] - University of Zurich
!> \version 1.0
! **************************************************************************************************
   SUBROUTINE read_cell_from_external_file(cell_section, para_env)

      TYPE(section_vals_type), POINTER                   :: cell_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'read_cell_from_external_file', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_path_length)                 :: cell_file_name
      INTEGER                                            :: i, idum, j, my_format, n_rep
      LOGICAL                                            :: explicit, my_end
      REAL(KIND=dp)                                      :: xdum
      REAL(KIND=dp), DIMENSION(3, 3)                     :: hmat
      REAL(KIND=dp), DIMENSION(:), POINTER               :: cell_par
      TYPE(cp_parser_type), POINTER                      :: parser

      NULLIFY (parser)
      CALL section_vals_val_get(cell_section, "CELL_FILE_NAME", c_val=cell_file_name)
      CALL section_vals_val_get(cell_section, "CELL_FILE_FORMAT", i_val=my_format)
      CALL parser_create(parser, cell_file_name, para_env=para_env)
      CALL parser_get_next_line(parser, 1)
      SELECT CASE (my_format)
      CASE (do_cell_cp2k)
         my_end = .FALSE.
         DO WHILE (.NOT. my_end)
            READ (parser%input_line, *) idum, xdum, hmat(:, 1), hmat(:, 2), hmat(:, 3)
            CALL parser_get_next_line(parser, 1, at_end=my_end)
         END DO
      CASE (do_cell_xsc)
         READ (parser%input_line, *) idum, hmat(:, 1), hmat(:, 2), hmat(:, 3)
      END SELECT
      CALL parser_release(parser)
      CALL section_vals_val_unset(cell_section, "CELL_FILE_NAME")
      CALL section_vals_val_unset(cell_section, "CELL_FILE_FORMAT")
      ! Conver to CP2K units
      DO i = 1, 3
         DO j = 1, 3
            hmat(j, i) = cp_unit_to_cp2k(hmat(j, i), "angstrom")
         END DO
      END DO
      ! Check if the cell was already defined
      explicit = .FALSE.
      CALL section_vals_val_get(cell_section, "A", n_rep_val=n_rep)
      explicit = explicit .OR. (n_rep == 1)
      CALL section_vals_val_get(cell_section, "B", n_rep_val=n_rep)
      explicit = explicit .OR. (n_rep == 1)
      CALL section_vals_val_get(cell_section, "C", n_rep_val=n_rep)
      explicit = explicit .OR. (n_rep == 1)
      CALL section_vals_val_get(cell_section, "ABC", n_rep_val=n_rep)
      explicit = explicit .OR. (n_rep == 1)
      ! Possibly print a warning
      IF (explicit) &
         CALL cp_warn(cp__l("cell_methods.F",315), &
                      "Cell specification (A,B,C or ABC) provided together with the external "// &
                      "cell setup! Ignoring (A,B,C or ABC) and proceeding with info read from the "// &
                      "external file! ")
      ! Copy cell information in the A, B, C fields..(we may need them later on..)
      ALLOCATE (cell_par(3))
      cell_par = hmat(:, 1)
      CALL section_vals_val_set(cell_section, "A", r_vals_ptr=cell_par)
      ALLOCATE (cell_par(3))
      cell_par = hmat(:, 2)
      CALL section_vals_val_set(cell_section, "B", r_vals_ptr=cell_par)
      ALLOCATE (cell_par(3))
      cell_par = hmat(:, 3)
      CALL section_vals_val_set(cell_section, "C", r_vals_ptr=cell_par)
      ! Unset possible keywords
      CALL section_vals_val_unset(cell_section, "ABC")
      CALL section_vals_val_unset(cell_section, "ALPHA_BETA_GAMMA")

   END SUBROUTINE read_cell_from_external_file

! **************************************************************************************************
!> \brief   Write the cell parameters to the output unit.
!> \param cell ...
!> \param subsys_section ...
!> \param cell_ref ...
!> \param label ...
!> \date    02.06.2000
!> \par     History
!>    - 11.2008 Teodoro Laino [tlaino] - rewrite and enabling user driven units
!> \author  Matthias Krack
!> \version 1.0
! **************************************************************************************************
   RECURSIVE SUBROUTINE write_cell(cell, subsys_section, cell_ref, label)

      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: subsys_section
      TYPE(cell_type), OPTIONAL, POINTER                 :: cell_ref
      CHARACTER(LEN=*), INTENT(IN), OPTIONAL             :: label

      CHARACTER(len=*), PARAMETER :: routineN = 'write_cell', routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: my_label, unit_str
      INTEGER                                            :: output_unit
      REAL(KIND=dp)                                      :: alpha, beta, gamma, val
      REAL(KIND=dp), DIMENSION(3)                        :: abc
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section

      NULLIFY (enum)
      NULLIFY (keyword)
      NULLIFY (logger)
      NULLIFY (section)
      logger => cp_get_default_logger()
      my_label = "CELL|"
      IF (PRESENT(label)) my_label = TRIM(label)
      output_unit = cp_print_key_unit_nr(logger, subsys_section, "PRINT%CELL", &
                                         extension=".Log")
      CALL section_vals_val_get(subsys_section, "PRINT%CELL%UNIT", c_val=unit_str)
      IF (output_unit > 0) THEN
         CALL get_cell(cell=cell, abc=abc, alpha=alpha, beta=beta, gamma=gamma)
         WRITE (UNIT=output_unit, FMT='( )')
         val = cp_unit_from_cp2k(cell%deth, TRIM(unit_str)//"^3")
         WRITE (UNIT=output_unit, FMT="(T2,A,T61,F20.3)") &
            TRIM(my_label)//" Volume ["//TRIM(unit_str)//"^3]:", val
         val = cp_unit_from_cp2k(1.0_dp, TRIM(unit_str))
         WRITE (UNIT=output_unit, FMT="(T2,A,T30,3F10.3,4X,A6,F11.3)") &
            TRIM(my_label)//" Vector a ["//TRIM(unit_str)//"]:", cell%hmat(:, 1)*val, &
            "|a| = ", abc(1)*val, &
            TRIM(my_label)//" Vector b ["//TRIM(unit_str)//"]:", cell%hmat(:, 2)*val, &
            "|b| = ", abc(2)*val, &
            TRIM(my_label)//" Vector c ["//TRIM(unit_str)//"]:", cell%hmat(:, 3)*val, &
            "|c| = ", abc(3)*val
         WRITE (UNIT=output_unit, FMT="(T2,A,T70,F11.3)") &
            TRIM(my_label)//" Angle (b,c), alpha [degree]: ", alpha, &
            TRIM(my_label)//" Angle (a,c), beta  [degree]: ", beta, &
            TRIM(my_label)//" Angle (a,b), gamma [degree]: ", gamma
         IF (cell%symmetry_id /= cell_sym_none) THEN
            CALL create_cell_section(section)
            keyword => section_get_keyword(section, "SYMMETRY")
            CALL keyword_get(keyword, enum=enum)
            WRITE (UNIT=output_unit, FMT="(T2,A,T61,A20)") &
               TRIM(my_label)//" Requested initial symmetry: ", &
               ADJUSTR(TRIM(enum_i2c(enum, cell%symmetry_id)))
            CALL section_release(section)
         END IF
         IF (cell%orthorhombic) THEN
            WRITE (UNIT=output_unit, FMT="(T2,A,T78,A3)") &
               TRIM(my_label)//" Numerically orthorhombic: ", "YES"
         ELSE
            WRITE (UNIT=output_unit, FMT="(T2,A,T78,A3)") &
               TRIM(my_label)//" Numerically orthorhombic: ", " NO"
         END IF
      END IF
      CALL cp_print_key_finished_output(output_unit, logger, subsys_section, &
                                        "PRINT%CELL")

      IF (PRESENT(cell_ref)) THEN
         CALL write_cell(cell_ref, subsys_section, label="CELL_REF|")
      END IF

   END SUBROUTINE write_cell

END MODULE cell_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains information regarding the decoupling/recoupling method of Bloechl
!> \author Teodoro Laino
! **************************************************************************************************
MODULE cp_ddapc_methods
   USE cell_types,                      ONLY: cell_type
   USE cp_log_handling,                 ONLY: cp_logger_get_default_io_unit
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE erf_fn,                          ONLY: erfc
   USE input_constants,                 ONLY: weight_type_mass,&
                                              weight_type_unit
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get,&
                                              section_vals_val_set
   USE kahan_sum,                       ONLY: accurate_sum
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: fourpi,&
                                              oorootpi,&
                                              pi,&
                                              twopi
   USE mathlib,                         ONLY: diamat_all,&
                                              invert_matrix
!NB for paralelizaion of ewald_ddapc_pot()
   USE message_passing,                 ONLY: mp_sum
   USE particle_types,                  ONLY: particle_type
   USE pw_spline_utils,                 ONLY: Eval_Interp_Spl3_pbc
   USE pw_types,                        ONLY: pw_type
   USE spherical_harmonics,             ONLY: legendre
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .FALSE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_ddapc_methods'
   PUBLIC :: ddapc_eval_gfunc, &
             build_b_vector, &
             build_der_b_vector, &
             build_A_matrix, &
             build_der_A_matrix_rows, &
             prep_g_dot_rvec_sin_cos, &
             cleanup_g_dot_rvec_sin_cos, &
             ddapc_eval_AmI, &
             ewald_ddapc_pot, &
             solvation_ddapc_pot

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param gfunc ...
!> \param w ...
!> \param gcut ...
!> \param rho_tot_g ...
!> \param radii ...
! **************************************************************************************************
   SUBROUTINE ddapc_eval_gfunc(gfunc, w, gcut, rho_tot_g, radii)
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: gfunc
      REAL(kind=dp), DIMENSION(:), POINTER               :: w
      REAL(KIND=dp), INTENT(IN)                          :: gcut
      TYPE(pw_type), POINTER                             :: rho_tot_g
      REAL(kind=dp), DIMENSION(:), POINTER               :: radii

      CHARACTER(len=*), PARAMETER :: routineN = 'ddapc_eval_gfunc', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: e_dim, handle, ig, igauss, s_dim
      REAL(KIND=dp)                                      :: g2, gcut2, rc, rc2

      CALL timeset(routineN, handle)
      gcut2 = gcut*gcut
      !
      s_dim = rho_tot_g%pw_grid%first_gne0
      e_dim = rho_tot_g%pw_grid%ngpts_cut_local
      ALLOCATE (gfunc(s_dim:e_dim, SIZE(radii)))
      ALLOCATE (w(s_dim:e_dim))
      gfunc = 0.0_dp
      w = 0.0_dp
      DO igauss = 1, SIZE(radii)
         rc = radii(igauss)
         rc2 = rc*rc
         DO ig = s_dim, e_dim
            g2 = rho_tot_g%pw_grid%gsq(ig)
            IF (g2 > gcut2) EXIT
            gfunc(ig, igauss) = EXP(-g2*rc2/4.0_dp)
         END DO
      END DO
      DO ig = s_dim, e_dim
         g2 = rho_tot_g%pw_grid%gsq(ig)
         IF (g2 > gcut2) EXIT
         w(ig) = fourpi*(g2-gcut2)**2/(g2*gcut2)
      END DO
      CALL timestop(handle)
   END SUBROUTINE ddapc_eval_gfunc

! **************************************************************************************************
!> \brief Computes the B vector for the solution of the linear system
!> \param bv ...
!> \param gfunc ...
!> \param w ...
!> \param particle_set ...
!> \param radii ...
!> \param rho_tot_g ...
!> \param gcut ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE build_b_vector(bv, gfunc, w, particle_set, radii, rho_tot_g, gcut)
      REAL(KIND=dp), DIMENSION(:), INTENT(INOUT)         :: bv
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: gfunc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: w
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii
      TYPE(pw_type), POINTER                             :: rho_tot_g
      REAL(KIND=dp), INTENT(IN)                          :: gcut

      CHARACTER(len=*), PARAMETER :: routineN = 'build_b_vector', routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: phase
      INTEGER                                            :: e_dim, handle, idim, ig, igauss, igmax, &
                                                            iparticle, s_dim
      REAL(KIND=dp)                                      :: arg, g2, gcut2, gvec(3), rvec(3)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: my_bv, my_bvw

      CALL timeset(routineN, handle)
      NULLIFY (my_bv, my_bvw)
      gcut2 = gcut*gcut
      s_dim = rho_tot_g%pw_grid%first_gne0
      e_dim = rho_tot_g%pw_grid%ngpts_cut_local
      igmax = 0
      DO ig = s_dim, e_dim
         g2 = rho_tot_g%pw_grid%gsq(ig)
         IF (g2 > gcut2) EXIT
         igmax = ig
      ENDDO
      IF (igmax .GE. s_dim) THEN
         ALLOCATE (my_bv(s_dim:igmax))
         ALLOCATE (my_bvw(s_dim:igmax))
         !
         DO iparticle = 1, SIZE(particle_set)
            rvec = particle_set(iparticle)%r
            my_bv = 0.0_dp
            DO ig = s_dim, igmax
               gvec = rho_tot_g%pw_grid%g(:, ig)
               arg = DOT_PRODUCT(gvec, rvec)
               phase = CMPLX(COS(arg), -SIN(arg), KIND=dp)
               my_bv(ig) = w(ig)*REAL(CONJG(rho_tot_g%cc(ig))*phase, KIND=dp)
            END DO
            DO igauss = 1, SIZE(radii)
               idim = (iparticle-1)*SIZE(radii)+igauss
               DO ig = s_dim, igmax
                  my_bvw(ig) = my_bv(ig)*gfunc(ig, igauss)
               END DO
               bv(idim) = accurate_sum(my_bvw)
            END DO
         END DO
         DEALLOCATE (my_bvw)
         DEALLOCATE (my_bv)
      ELSE
         DO iparticle = 1, SIZE(particle_set)
            DO igauss = 1, SIZE(radii)
               idim = (iparticle-1)*SIZE(radii)+igauss
               bv(idim) = 0.0_dp
            END DO
         END DO
      END IF
      CALL timestop(handle)
   END SUBROUTINE build_b_vector

! **************************************************************************************************
!> \brief Computes the A matrix for the solution of the linear system
!> \param Am ...
!> \param gfunc ...
!> \param w ...
!> \param particle_set ...
!> \param radii ...
!> \param rho_tot_g ...
!> \param gcut ...
!> \param g_dot_rvec_sin ...
!> \param g_dot_rvec_cos ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
!> \note NB accept g_dot_rvec_* arrays
! **************************************************************************************************
   SUBROUTINE build_A_matrix(Am, gfunc, w, particle_set, radii, rho_tot_g, gcut, g_dot_rvec_sin, g_dot_rvec_cos)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: Am
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: gfunc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: w
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii
      TYPE(pw_type), POINTER                             :: rho_tot_g
      REAL(KIND=dp), INTENT(IN)                          :: gcut
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: g_dot_rvec_sin, g_dot_rvec_cos

      CHARACTER(len=*), PARAMETER :: routineN = 'build_A_matrix', routineP = moduleN//':'//routineN

      INTEGER                                            :: e_dim, handle, idim1, idim2, ig, &
                                                            igauss1, igauss2, igmax, iparticle1, &
                                                            iparticle2, istart_g, s_dim
      REAL(KIND=dp)                                      :: g2, gcut2, tmp
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: my_Am, my_Amw
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :, :)     :: gfunc_sq(:, :, :)

!NB precalculate as many things outside of the innermost loop as possible, in particular w(ig)*gfunc(ig,igauus1)*gfunc(ig,igauss2)

      CALL timeset(routineN, handle)
      gcut2 = gcut*gcut
      s_dim = rho_tot_g%pw_grid%first_gne0
      e_dim = rho_tot_g%pw_grid%ngpts_cut_local
      igmax = 0
      DO ig = s_dim, e_dim
         g2 = rho_tot_g%pw_grid%gsq(ig)
         IF (g2 > gcut2) EXIT
         igmax = ig
      ENDDO
      IF (igmax .GE. s_dim) THEN
         ALLOCATE (my_Am(s_dim:igmax))
         ALLOCATE (my_Amw(s_dim:igmax))
         ALLOCATE (gfunc_sq(s_dim:igmax, SIZE(radii), SIZE(radii)))

         DO igauss1 = 1, SIZE(radii)
            DO igauss2 = 1, SIZE(radii)
               gfunc_sq(s_dim:igmax, igauss1, igauss2) = w(s_dim:igmax)*gfunc(s_dim:igmax, igauss1)*gfunc(s_dim:igmax, igauss2)
            END DO
         END DO

         DO iparticle1 = 1, SIZE(particle_set)
            DO iparticle2 = iparticle1, SIZE(particle_set)
               DO ig = s_dim, igmax
                  !NB replace explicit dot product and cosine with cos(A+B) formula - much faster
                  my_Am(ig) = (g_dot_rvec_cos(ig-s_dim+1, iparticle1)*g_dot_rvec_cos(ig-s_dim+1, iparticle2)+ &
                               g_dot_rvec_sin(ig-s_dim+1, iparticle1)*g_dot_rvec_sin(ig-s_dim+1, iparticle2))
               END DO
               DO igauss1 = 1, SIZE(radii)
                  idim1 = (iparticle1-1)*SIZE(radii)+igauss1
                  istart_g = 1
                  IF (iparticle2 == iparticle1) istart_g = igauss1
                  DO igauss2 = istart_g, SIZE(radii)
                     idim2 = (iparticle2-1)*SIZE(radii)+igauss2
                     my_Amw(s_dim:igmax) = my_Am(s_dim:igmax)*gfunc_sq(s_dim:igmax, igauss1, igauss2)
                     !NB no loss of accuracy in my test cases
                     !tmp = accurate_sum(my_Amw)
                     tmp = SUM(my_Amw)
                     Am(idim2, idim1) = tmp
                     Am(idim1, idim2) = tmp
                  END DO
               END DO
            END DO
         END DO
         DEALLOCATE (gfunc_sq)
         DEALLOCATE (my_Amw)
         DEALLOCATE (my_Am)
      END IF
      CALL timestop(handle)
   END SUBROUTINE build_A_matrix

! **************************************************************************************************
!> \brief Computes the derivative of B vector for the evaluation of the Pulay forces
!> \param dbv ...
!> \param gfunc ...
!> \param w ...
!> \param particle_set ...
!> \param radii ...
!> \param rho_tot_g ...
!> \param gcut ...
!> \param iparticle0 ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE build_der_b_vector(dbv, gfunc, w, particle_set, radii, rho_tot_g, gcut, iparticle0)
      REAL(KIND=dp), DIMENSION(:, :), INTENT(INOUT)      :: dbv
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: gfunc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: w
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: radii
      TYPE(pw_type), POINTER                             :: rho_tot_g
      REAL(KIND=dp), INTENT(IN)                          :: gcut
      INTEGER, INTENT(IN)                                :: iparticle0

      CHARACTER(len=*), PARAMETER :: routineN = 'build_der_b_vector', &
         routineP = moduleN//':'//routineN

      COMPLEX(KIND=dp)                                   :: dphase
      INTEGER                                            :: e_dim, handle, idim, ig, igauss, igmax, &
                                                            iparticle, s_dim
      REAL(KIND=dp)                                      :: arg, g2, gcut2
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: my_dbvw
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: my_dbv
      REAL(KIND=dp), DIMENSION(3)                        :: gvec, rvec

      CALL timeset(routineN, handle)
      gcut2 = gcut*gcut
      s_dim = rho_tot_g%pw_grid%first_gne0
      e_dim = rho_tot_g%pw_grid%ngpts_cut_local
      igmax = 0
      DO ig = s_dim, e_dim
         g2 = rho_tot_g%pw_grid%gsq(ig)
         IF (g2 > gcut2) EXIT
         igmax = ig
      ENDDO
      IF (igmax .GE. s_dim) THEN
         ALLOCATE (my_dbv(3, s_dim:igmax))
         ALLOCATE (my_dbvw(s_dim:igmax))
         DO iparticle = 1, SIZE(particle_set)
            IF (iparticle /= iparticle0) CYCLE
            rvec = particle_set(iparticle)%r
            DO ig = s_dim, igmax
               gvec = rho_tot_g%pw_grid%g(:, ig)
               arg = DOT_PRODUCT(gvec, rvec)
               dphase = -CMPLX(SIN(arg), COS(arg), KIND=dp)
               my_dbv(:, ig) = w(ig)*REAL(CONJG(rho_tot_g%cc(ig))*dphase, KIND=dp)*gvec(:)
            END DO
            DO igauss = 1, SIZE(radii)
               idim = (iparticle-1)*SIZE(radii)+igauss
               DO ig = s_dim, igmax
                  my_dbvw(ig) = my_dbv(1, ig)*gfunc(ig, igauss)
               END DO
               dbv(idim, 1) = accurate_sum(my_dbvw)
               DO ig = s_dim, igmax
                  my_dbvw(ig) = my_dbv(2, ig)*gfunc(ig, igauss)
               END DO
               dbv(idim, 2) = accurate_sum(my_dbvw)
               DO ig = s_dim, igmax
                  my_dbvw(ig) = my_dbv(3, ig)*gfunc(ig, igauss)
               END DO
               dbv(idim, 3) = accurate_sum(my_dbvw)
            END DO
         END DO
         DEALLOCATE (my_dbvw)
         DEALLOCATE (my_dbv)
      ELSE
         DO iparticle = 1, SIZE(particle_set)
            IF (iparticle /= iparticle0) CYCLE
            DO igauss = 1, SIZE(radii)
               idim = (iparticle-1)*SIZE(radii)+igauss
               dbv(idim, 1:3) = 0.0_dp
            END DO
         END DO
      END IF
      CALL timestop(handle)
   END SUBROUTINE build_der_b_vector

! **************************************************************************************************
!> \brief Computes the derivative of the A matrix for the evaluation of the
!>      Pulay forces
!> \param dAm ...
!> \param gfunc ...
!> \param w ...
!> \param particle_set ...
!> \param radii ...
!> \param rho_tot_g ...
!> \param gcut ...
!> \param iparticle0 ...
!> \param nparticles ...
!> \param g_dot_rvec_sin ...
!> \param g_dot_rvec_cos ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
!> \note NB accept g_dot_rvec_* arrays
! **************************************************************************************************
   SUBROUTINE build_der_A_matrix_rows(dAm, gfunc, w, particle_set, radii, &
                                      rho_tot_g, gcut, iparticle0, nparticles, g_dot_rvec_sin, g_dot_rvec_cos)
      REAL(KIND=dp), DIMENSION(:, :, :), INTENT(INOUT)   :: dAm
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: gfunc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: w
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii
      TYPE(pw_type), POINTER                             :: rho_tot_g
      REAL(KIND=dp), INTENT(IN)                          :: gcut
      INTEGER, INTENT(IN)                                :: iparticle0, nparticles
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: g_dot_rvec_sin, g_dot_rvec_cos

      CHARACTER(len=*), PARAMETER :: routineN = 'build_der_A_matrix_rows', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: e_dim, handle, ig, igauss2, igmax, &
                                                            iparticle1, iparticle2, s_dim
      REAL(KIND=dp)                                      :: g2, gcut2

!NB calculate derivatives for a block of particles, just the row parts (since derivative matrix is symmetric)
!NB Use DGEMM to speed up calculation, can't do accurate_sum() anymore because dgemm does the sum over g

      EXTERNAL DGEMM
      REAL(KIND=dp), ALLOCATABLE :: lhs(:, :), rhs(:, :)
      INTEGER :: Nr, Np, Ng, icomp, ipp

      CALL timeset(routineN, handle)
      gcut2 = gcut*gcut
      s_dim = rho_tot_g%pw_grid%first_gne0
      e_dim = rho_tot_g%pw_grid%ngpts_cut_local
      igmax = 0
      DO ig = s_dim, e_dim
         g2 = rho_tot_g%pw_grid%gsq(ig)
         IF (g2 > gcut2) EXIT
         igmax = ig
      ENDDO

      Nr = SIZE(radii)
      Np = SIZE(particle_set)
      Ng = igmax-s_dim+1
      IF (igmax .GE. s_dim) THEN
         ALLOCATE (lhs(nparticles*Nr, Ng))
         ALLOCATE (rhs(Ng, Np*Nr))

         ! rhs with first term of sin(g.(rvec1-rvec2))
         ! rhs has all parts that depend on iparticle2
         DO iparticle2 = 1, Np
            DO igauss2 = 1, Nr
               rhs(1:Ng, (iparticle2-1)*Nr+igauss2) = g_dot_rvec_sin(1:Ng, iparticle2)*gfunc(s_dim:igmax, igauss2)
            END DO
         END DO
         DO icomp = 1, 3
            ! create lhs, which has all parts that depend on iparticle1
            DO ipp = 1, nparticles
               iparticle1 = iparticle0+ipp-1
               DO ig = s_dim, igmax
                  lhs((ipp-1)*Nr+1:(ipp-1)*Nr+Nr, ig-s_dim+1) = w(ig)*rho_tot_g%pw_grid%g(icomp, ig)* &
                                                                gfunc(ig, 1:Nr)*g_dot_rvec_cos(ig-s_dim+1, iparticle1)
               END DO
            END DO ! ipp
            ! do main multiply
            CALL DGEMM('N', 'N', nparticles*Nr, Np*Nr, Ng, 1.0D0, lhs(1, 1), nparticles*Nr, rhs(1, 1), &
                       Ng, 0.0D0, dAm((iparticle0-1)*Nr+1, 1, icomp), Np*Nr)
            ! do extra multiplies to compensate for missing factor of 2
            DO ipp = 1, nparticles
               iparticle1 = iparticle0+ipp-1
               CALL DGEMM('N', 'N', Nr, Nr, Ng, 1.0D0, lhs((ipp-1)*Nr+1, 1), nparticles*Nr, rhs(1, (iparticle1-1)*Nr+1), &
                          Ng, 1.0D0, dAm((iparticle1-1)*Nr+1, (iparticle1-1)*Nr+1, icomp), Np*Nr)
            END DO
            ! now extra columns to account for factor of 2 in some rhs columns
         END DO ! icomp

         ! rhs with second term of sin(g.(rvec1-rvec2))
         ! rhs has all parts that depend on iparticle2
         DO iparticle2 = 1, Np
            DO igauss2 = 1, Nr
               rhs(1:Ng, (iparticle2-1)*Nr+igauss2) = -g_dot_rvec_cos(1:Ng, iparticle2)*gfunc(s_dim:igmax, igauss2)
            END DO
         END DO
         DO icomp = 1, 3
            ! create lhs, which has all parts that depend on iparticle1
            DO ipp = 1, nparticles
               iparticle1 = iparticle0+ipp-1
               DO ig = s_dim, igmax
                  lhs((ipp-1)*Nr+1:(ipp-1)*Nr+Nr, ig-s_dim+1) = w(ig)*rho_tot_g%pw_grid%g(icomp, ig)*gfunc(ig, 1:Nr)* &
                                                                g_dot_rvec_sin(ig-s_dim+1, iparticle1)
               END DO
            ENDDO
            ! do main multiply
            CALL DGEMM('N', 'N', nparticles*Nr, Np*Nr, Ng, 1.0D0, lhs(1, 1), nparticles*Nr, rhs(1, 1), &
                       Ng, 1.0D0, dAm((iparticle0-1)*Nr+1, 1, icomp), Np*Nr)
            ! do extra multiples to compensate for missing factor of 2
            DO ipp = 1, nparticles
               iparticle1 = iparticle0+ipp-1
               CALL DGEMM('N', 'N', Nr, Nr, Ng, 1.0D0, lhs((ipp-1)*Nr+1, 1), nparticles*Nr, rhs(1, (iparticle1-1)*Nr+1), &
                          Ng, 1.0D0, dAm((iparticle1-1)*Nr+1, (iparticle1-1)*Nr+1, icomp), Np*Nr)
            END DO
         END DO

         DEALLOCATE (rhs)
         DEALLOCATE (lhs)
      END IF
      CALL timestop(handle)
   END SUBROUTINE build_der_A_matrix_rows

! **************************************************************************************************
!> \brief deallocate g_dot_rvec_* arrays
!> \param g_dot_rvec_sin ...
!> \param g_dot_rvec_cos ...
! **************************************************************************************************
   SUBROUTINE cleanup_g_dot_rvec_sin_cos(g_dot_rvec_sin, g_dot_rvec_cos)
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: g_dot_rvec_sin, g_dot_rvec_cos

      CHARACTER(len=*), PARAMETER :: routineN = 'cleanup_g_dot_rvec_sin_cos', &
         routineP = moduleN//':'//routineN

      IF (ALLOCATED(g_dot_rvec_sin)) DEALLOCATE (g_dot_rvec_sin)
      IF (ALLOCATED(g_dot_rvec_cos)) DEALLOCATE (g_dot_rvec_cos)
   END SUBROUTINE cleanup_g_dot_rvec_sin_cos

! **************************************************************************************************
!> \brief precompute sin(g.r) and cos(g.r) for quicker evaluations of sin(g.(r1-r2)) and cos(g.(r1-r2))
!> \param rho_tot_g ...
!> \param particle_set ...
!> \param gcut ...
!> \param g_dot_rvec_sin ...
!> \param g_dot_rvec_cos ...
! **************************************************************************************************
   SUBROUTINE prep_g_dot_rvec_sin_cos(rho_tot_g, particle_set, gcut, g_dot_rvec_sin, g_dot_rvec_cos)
      TYPE(pw_type), POINTER                             :: rho_tot_g
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), INTENT(IN)                          :: gcut
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: g_dot_rvec_sin, g_dot_rvec_cos

      CHARACTER(len=*), PARAMETER :: routineN = 'prep_g_dot_rvec_sin_cos', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: e_dim, ig, igmax, iparticle, s_dim
      REAL(KIND=dp)                                      :: g2, g_dot_rvec, gcut2, rvec(3)

      gcut2 = gcut*gcut
      s_dim = rho_tot_g%pw_grid%first_gne0
      e_dim = rho_tot_g%pw_grid%ngpts_cut_local
      igmax = 0
      DO ig = s_dim, e_dim
         g2 = rho_tot_g%pw_grid%gsq(ig)
         IF (g2 > gcut2) EXIT
         igmax = ig
      ENDDO

      IF (igmax .GE. s_dim) THEN
         ALLOCATE (g_dot_rvec_sin(1:igmax-s_dim+1, SIZE(particle_set)))
         ALLOCATE (g_dot_rvec_cos(1:igmax-s_dim+1, SIZE(particle_set)))

         DO iparticle = 1, SIZE(particle_set)
            rvec = particle_set(iparticle)%r
            DO ig = s_dim, igmax
               g_dot_rvec = DOT_PRODUCT(rho_tot_g%pw_grid%g(:, ig), rvec)
               g_dot_rvec_sin(ig-s_dim+1, iparticle) = SIN(g_dot_rvec)
               g_dot_rvec_cos(ig-s_dim+1, iparticle) = COS(g_dot_rvec)
            END DO
         END DO
      END IF

   END SUBROUTINE prep_g_dot_rvec_sin_cos

! **************************************************************************************************
!> \brief Computes the inverse AmI of the Am matrix
!> \param GAmI ...
!> \param c0 ...
!> \param gfunc ...
!> \param w ...
!> \param particle_set ...
!> \param gcut ...
!> \param rho_tot_g ...
!> \param radii ...
!> \param iw ...
!> \param Vol ...
!> \par History
!>      12.2005 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE ddapc_eval_AmI(GAmI, c0, gfunc, w, particle_set, gcut, &
                             rho_tot_g, radii, iw, Vol)
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: GAmI
      REAL(KIND=dp), INTENT(OUT)                         :: c0
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: gfunc
      REAL(KIND=dp), DIMENSION(:), POINTER               :: w
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), INTENT(IN)                          :: gcut
      TYPE(pw_type), POINTER                             :: rho_tot_g
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii
      INTEGER, INTENT(IN)                                :: iw
      REAL(KIND=dp), INTENT(IN)                          :: Vol

      CHARACTER(len=*), PARAMETER :: routineN = 'ddapc_eval_AmI', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ndim
      REAL(KIND=dp)                                      :: condition_number, inv_error
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:)           :: AmE, cv
      REAL(KIND=dp), ALLOCATABLE, DIMENSION(:, :)        :: Am, AmI, Amw, g_dot_rvec_cos, &
                                                            g_dot_rvec_sin

!NB for precomputation of sin(g.r) and cos(g.r)

      CALL timeset(routineN, handle)
      ndim = SIZE(particle_set)*SIZE(radii)
      ALLOCATE (Am(ndim, ndim))
      ALLOCATE (AmI(ndim, ndim))
      ALLOCATE (GAmI(ndim, ndim))
      ALLOCATE (cv(ndim))
      Am = 0.0_dp
      AmI = 0.0_dp
      cv = 1.0_dp/Vol
      !NB precompute sin(g.r) and cos(g.r) for faster evaluation of cos(g.(r1-r2)) in build_A_matrix()
      CALL prep_g_dot_rvec_sin_cos(rho_tot_g, particle_set, gcut, g_dot_rvec_sin, g_dot_rvec_cos)
      CALL build_A_matrix(Am, gfunc, w, particle_set, radii, rho_tot_g, gcut, g_dot_rvec_sin, g_dot_rvec_cos)
      CALL cleanup_g_dot_rvec_sin_cos(g_dot_rvec_sin, g_dot_rvec_cos)
      Am(:, :) = Am(:, :)/(Vol*Vol)
      CALL mp_sum(Am, rho_tot_g%pw_grid%para%group)
      IF (iw > 0) THEN
         ! Checking conditions numbers and eigenvalues
         ALLOCATE (Amw(ndim, ndim))
         ALLOCATE (AmE(ndim))
         Amw(:, :) = Am
         CALL diamat_all(Amw, AmE)
         condition_number = MAXVAL(ABS(AmE))/MINVAL(ABS(AmE))
         WRITE (iw, '(T3,A)') " Eigenvalues of Matrix A:"
         WRITE (iw, '(T3,4E15.8)') AmE
         WRITE (iw, '(T3,A,1E15.9)') " Condition number:", condition_number
         IF (condition_number > 1.0E12_dp) THEN
            WRITE (iw, FMT="(/,T2,A)") &
               "WARNING: high condition number => possibly ill-conditioned matrix"
         END IF
         DEALLOCATE (Amw)
         DEALLOCATE (AmE)
      END IF
      CALL invert_matrix(Am, AmI, inv_error, "N", improve=.FALSE.)
      IF (iw > 0) THEN
         WRITE (iw, '(T3,A,F15.9)') " Error inverting the A matrix: ", inv_error
      END IF
      c0 = DOT_PRODUCT(cv, MATMUL(AmI, cv))
      DEALLOCATE (Am)
      DEALLOCATE (cv)
      GAmI = AmI
      DEALLOCATE (AmI)
      CALL timestop(handle)
   END SUBROUTINE ddapc_eval_AmI

! **************************************************************************************************
!> \brief Evaluates the Ewald term E2 and E3 energy term for the decoupling/coupling
!>      of periodic images
!> \param cp_para_env ...
!> \param coeff ...
!> \param factor ...
!> \param cell ...
!> \param multipole_section ...
!> \param particle_set ...
!> \param M ...
!> \param radii ...
!> \par History
!>      08.2005 created [tlaino]
!> \author Teodoro Laino
!> \note NB receive cp_para_env for parallelization
! **************************************************************************************************
   RECURSIVE SUBROUTINE ewald_ddapc_pot(cp_para_env, coeff, factor, cell, multipole_section, &
                                        particle_set, M, radii)
      TYPE(cp_para_env_type), POINTER                    :: cp_para_env
      TYPE(pw_type), POINTER                             :: coeff
      REAL(KIND=dp), INTENT(IN)                          :: factor
      TYPE(cell_type), POINTER                           :: cell
      TYPE(section_vals_type), POINTER                   :: multipole_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: M
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_ddapc_pot', &
         routineP = moduleN//':'//routineN

      INTEGER :: ewmdim, handle, idim, idim1, idim2, idimo, igauss1, igauss2, ip1, ip2, &
         iparticle1, iparticle2, istart_g, k1, k2, k3, n_rep, ndim, nmax1, nmax2, nmax3, r1, r2, &
         r3, rmax1, rmax2, rmax3
      LOGICAL                                            :: analyt
      REAL(KIND=dp) :: alpha, eps, ew_neut, fac, fac3, fs, g_ewald, galpha, gsq, gsqi, ij_fac, &
         my_val, r, r2tmp, r_ewald, rc1, rc12, rc2, rc22, rcut, rcut2, t1, tol, tol1
      REAL(KIND=dp), DIMENSION(3)                        :: fvec, gvec, ra, rvec
      REAL(KIND=dp), DIMENSION(:), POINTER               :: EwM

      NULLIFY (EwM)
      CALL timeset(routineN, handle)
      IF(.NOT.(.NOT. ASSOCIATED(M)))CALL cp__a("cp_ddapc_methods.F",659)
      IF(.NOT.(ASSOCIATED(radii)))CALL cp__a("cp_ddapc_methods.F",660)
      IF(.NOT.(cell%orthorhombic))CALL cp__a("cp_ddapc_methods.F",661)
      rcut = MIN(cell%hmat(1, 1), cell%hmat(2, 2), cell%hmat(3, 3))/2.0_dp
      CALL section_vals_val_get(multipole_section, "RCUT", n_rep_val=n_rep)
      IF (n_rep == 1) CALL section_vals_val_get(multipole_section, "RCUT", r_val=rcut)
      CALL section_vals_val_get(multipole_section, "EWALD_PRECISION", r_val=eps)
      CALL section_vals_val_get(multipole_section, "ANALYTICAL_GTERM", l_val=analyt)
      rcut2 = rcut**2
      !
      ! Setting-up parameters for Ewald summation
      !
      eps = MIN(ABS(eps), 0.5_dp)
      tol = SQRT(ABS(LOG(eps*rcut)))
      alpha = SQRT(ABS(LOG(eps*rcut*tol)))/rcut
      galpha = 1.0_dp/(4.0_dp*alpha*alpha)
      tol1 = SQRT(-LOG(eps*rcut*(2.0_dp*tol*alpha)**2))
      nmax1 = NINT(0.25_dp+cell%hmat(1, 1)*alpha*tol1/pi)
      nmax2 = NINT(0.25_dp+cell%hmat(2, 2)*alpha*tol1/pi)
      nmax3 = NINT(0.25_dp+cell%hmat(3, 3)*alpha*tol1/pi)

      rmax1 = CEILING(rcut/cell%hmat(1, 1))
      rmax2 = CEILING(rcut/cell%hmat(2, 2))
      rmax3 = CEILING(rcut/cell%hmat(3, 3))
      fac = 1.e0_dp/cell%deth
      fac3 = fac*pi
      fvec = twopi/(/cell%hmat(1, 1), cell%hmat(2, 2), cell%hmat(3, 3)/)
      ew_neut = -fac*pi/alpha**2
      !
      ewmdim = SIZE(particle_set)*(SIZE(particle_set)+1)/2
      ndim = SIZE(particle_set)*SIZE(radii)
      ALLOCATE (EwM(ewmdim))
      ALLOCATE (M(ndim, ndim))
      M = 0.0_dp
      !
      idim = 0
      !NB zero EwM so mp_sum(EwM) will give right answer
      EwM = 0.0_dp
      DO iparticle1 = 1, SIZE(particle_set)
         ip1 = (iparticle1-1)*SIZE(radii)
         DO iparticle2 = 1, iparticle1
            ij_fac = 1.0_dp
            IF (iparticle1 == iparticle2) ij_fac = 0.5_dp

            ip2 = (iparticle2-1)*SIZE(radii)
            idim = idim+1
            !NB parallelization, done here so indexing is right
            IF (MOD(iparticle1, cp_para_env%num_pe) /= cp_para_env%mepos) CYCLE
            !
            ! Real-Space Contribution
            !
            my_val = 0.0_dp
            rvec = particle_set(iparticle1)%r-particle_set(iparticle2)%r
            r_ewald = 0.0_dp
            IF (iparticle1 /= iparticle2) THEN
               ra = rvec
               r2tmp = DOT_PRODUCT(ra, ra)
               IF (r2tmp <= rcut2) THEN
                  r = SQRT(r2tmp)
                  t1 = erfc(alpha*r)/r
                  r_ewald = t1
               END IF
            END IF
            DO r1 = -rmax1, rmax1
               DO r2 = -rmax2, rmax2
                  DO r3 = -rmax3, rmax3
                     IF ((r1 == 0) .AND. (r2 == 0) .AND. (r3 == 0)) CYCLE
                     ra(1) = rvec(1)+cell%hmat(1, 1)*r1
                     ra(2) = rvec(2)+cell%hmat(2, 2)*r2
                     ra(3) = rvec(3)+cell%hmat(3, 3)*r3
                     r2tmp = DOT_PRODUCT(ra, ra)
                     IF (r2tmp <= rcut2) THEN
                        r = SQRT(r2tmp)
                        t1 = erfc(alpha*r)/r
                        r_ewald = r_ewald+t1*ij_fac
                     END IF
                  END DO
               END DO
            END DO
            !
            ! G-space Contribution
            !
            IF (analyt) THEN
               g_ewald = 0.0_dp
               DO k1 = 0, nmax1
                  DO k2 = -nmax2, nmax2
                     DO k3 = -nmax3, nmax3
                        IF (k1 == 0 .AND. k2 == 0 .AND. k3 == 0) CYCLE
                        fs = 2.0_dp; IF (k1 == 0) fs = 1.0_dp
                        gvec = fvec*(/REAL(k1, KIND=dp), REAL(k2, KIND=dp), REAL(k3, KIND=dp)/)
                        gsq = DOT_PRODUCT(gvec, gvec)
                        gsqi = fs/gsq
                        t1 = fac*gsqi*EXP(-galpha*gsq)
                        g_ewald = g_ewald+t1*COS(DOT_PRODUCT(gvec, rvec))
                     END DO
                  END DO
               END DO
            ELSE
               g_ewald = Eval_Interp_Spl3_pbc(rvec, coeff)
            END IF
            !
            ! G-EWALD, R-EWALD
            !
            g_ewald = r_ewald+fourpi*g_ewald
            !
            ! Self Contribution
            !
            IF (iparticle1 == iparticle2) THEN
               g_ewald = g_ewald-2.0_dp*alpha*oorootpi
            END IF
            !
            IF (iparticle1 /= iparticle2) THEN
               ra = rvec
               r = SQRT(DOT_PRODUCT(ra, ra))
               my_val = factor/r
            END IF
            EwM(idim) = my_val-factor*g_ewald
         END DO ! iparticle2
      END DO ! iparticle1
      !NB sum over parallelized contributions of different nodes
      CALL mp_sum(EwM, cp_para_env%group)
      idim = 0
      DO iparticle2 = 1, SIZE(particle_set)
         ip2 = (iparticle2-1)*SIZE(radii)
         idimo = (iparticle2-1)
         idimo = idimo*(idimo+1)/2
         DO igauss2 = 1, SIZE(radii)
            idim2 = ip2+igauss2
            rc2 = radii(igauss2)
            rc22 = rc2*rc2
            DO iparticle1 = 1, iparticle2
               ip1 = (iparticle1-1)*SIZE(radii)
               idim = idimo+iparticle1
               istart_g = 1
               IF (iparticle1 == iparticle2) istart_g = igauss2
               DO igauss1 = istart_g, SIZE(radii)
                  idim1 = ip1+igauss1
                  rc1 = radii(igauss1)
                  rc12 = rc1*rc1
                  M(idim1, idim2) = EwM(idim)-factor*ew_neut-factor*fac3*(rc12+rc22)
                  M(idim2, idim1) = M(idim1, idim2)
               END DO
            END DO
         END DO ! iparticle2
      END DO ! iparticle1
      DEALLOCATE (EwM)
      CALL timestop(handle)
   END SUBROUTINE ewald_ddapc_pot

! **************************************************************************************************
!> \brief Evaluates the electrostatic potential due to a simple solvation model
!>      Spherical cavity in a dieletric medium
!> \param solvation_section ...
!> \param particle_set ...
!> \param M ...
!> \param radii ...
!> \par History
!>      08.2006 created [tlaino]
!> \author Teodoro Laino
! **************************************************************************************************
   SUBROUTINE solvation_ddapc_pot(solvation_section, particle_set, M, radii)
      TYPE(section_vals_type), POINTER                   :: solvation_section
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: M
      REAL(KIND=dp), DIMENSION(:), POINTER               :: radii

      CHARACTER(len=*), PARAMETER :: routineN = 'solvation_ddapc_pot', &
         routineP = moduleN//':'//routineN

      INTEGER :: i, idim, idim1, idim2, igauss1, igauss2, ip1, ip2, iparticle1, iparticle2, &
         istart_g, j, l, lmax, n_rep1, n_rep2, ndim, output_unit, weight
      INTEGER, DIMENSION(:), POINTER                     :: list
      LOGICAL                                            :: fixed_center
      REAL(KIND=dp)                                      :: center(3), eps_in, eps_out, factor, &
                                                            mass, mycos, r1, r2, Rs, rvec(3)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: pos, R0
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: cost, LocP

      fixed_center = .FALSE.
      output_unit = cp_logger_get_default_io_unit()
      ndim = SIZE(particle_set)*SIZE(radii)
      ALLOCATE (M(ndim, ndim))
      M = 0.0_dp
      eps_in = 1.0_dp
      CALL section_vals_val_get(solvation_section, "EPS_OUT", r_val=eps_out)
      CALL section_vals_val_get(solvation_section, "LMAX", i_val=lmax)
      CALL section_vals_val_get(solvation_section, "SPHERE%RADIUS", r_val=Rs)
      CALL section_vals_val_get(solvation_section, "SPHERE%CENTER%XYZ", n_rep_val=n_rep1)
      IF (n_rep1 /= 0) THEN
         CALL section_vals_val_get(solvation_section, "SPHERE%CENTER%XYZ", r_vals=R0)
         center = R0
      ELSE
         CALL section_vals_val_get(solvation_section, "SPHERE%CENTER%ATOM_LIST", &
                                   n_rep_val=n_rep2)
         IF (n_rep2 /= 0) THEN
            CALL section_vals_val_get(solvation_section, "SPHERE%CENTER%ATOM_LIST", i_vals=list)
            CALL section_vals_val_get(solvation_section, "SPHERE%CENTER%WEIGHT_TYPE", i_val=weight)
            ALLOCATE (R0(3))
            SELECT CASE (weight)
            CASE (weight_type_unit)
               R0 = 0.0_dp
               DO i = 1, SIZE(list)
                  R0 = R0+particle_set(list(i))%r
               END DO
               R0 = R0/REAL(SIZE(list), KIND=dp)
            CASE (weight_type_mass)
               R0 = 0.0_dp
               mass = 0.0_dp
               DO i = 1, SIZE(list)
                  R0 = R0+particle_set(list(i))%r*particle_set(list(i))%atomic_kind%mass
                  mass = mass+particle_set(list(i))%atomic_kind%mass
               END DO
               R0 = R0/mass
            END SELECT
            center = R0
            CALL section_vals_val_get(solvation_section, "SPHERE%CENTER%FIXED", l_val=fixed_center)
            IF (fixed_center) THEN
               CALL section_vals_val_set(solvation_section, "SPHERE%CENTER%XYZ", &
                                         r_vals_ptr=R0)
            ELSE
               DEALLOCATE (R0)
            END IF
         END IF
      END IF
      IF(.NOT.(n_rep1 /= 0 .OR. n_rep2 /= 0))CALL cp__a("cp_ddapc_methods.F",883)
      ! Potential calculation
      ALLOCATE (LocP(0:lmax, SIZE(particle_set)))
      ALLOCATE (pos(SIZE(particle_set)))
      ALLOCATE (cost(SIZE(particle_set), SIZE(particle_set)))
      ! Determining the single atomic contribution to the dielectric dipole
      DO i = 1, SIZE(particle_set)
         rvec = particle_set(i)%r-center
         r2 = DOT_PRODUCT(rvec, rvec)
         r1 = SQRT(r2)
         IF (r1 >= Rs) THEN
            IF (output_unit > 0) THEN
               WRITE (output_unit, '(A,I6,A)') "Atom number :: ", i, " is out of the solvation sphere"
               WRITE (output_unit, '(2(A,F12.6))') "Distance from the center::", r1, " Radius of the sphere::", rs
            END IF
            CALL cp__b("cp_ddapc_methods.F",898,"")
         END IF
         LocP(:, i) = 0.0_dp
         IF (r1 /= 0.0_dp) THEN
            DO l = 0, lmax
               LocP(l, i) = (r1**l*REAL(l+1, KIND=dp)*(eps_in-eps_out))/ &
                            (Rs**(2*l+1)*eps_in*(REAL(l, KIND=dp)*eps_in+REAL(l+1, KIND=dp)*eps_out))
            END DO
         ELSE
            ! limit for r->0
            LocP(0, i) = (eps_in-eps_out)/(Rs*eps_in*eps_out)
         END IF
         pos(i) = r1
      END DO
      ! Particle-Particle potential energy matrix
      cost = 0.0_dp
      DO i = 1, SIZE(particle_set)
         DO j = 1, i
            factor = 0.0_dp
            IF (pos(i)*pos(j) /= 0.0_dp) THEN
               mycos = DOT_PRODUCT(particle_set(i)%r-center, particle_set(j)%r-center)/(pos(i)*pos(j))
               IF (ABS(mycos) > 1.0_dp) mycos = SIGN(1.0_dp, mycos)
               DO l = 0, lmax
                  factor = factor+LocP(l, i)*pos(j)**l*legendre(mycos, l, 0)
               END DO
            ELSE
               factor = LocP(0, i)
            END IF
            cost(i, j) = factor
            cost(j, i) = factor
         END DO
      END DO
      ! Computes the full potential energy matrix
      idim = 0
      DO iparticle2 = 1, SIZE(particle_set)
         ip2 = (iparticle2-1)*SIZE(radii)
         DO igauss2 = 1, SIZE(radii)
            idim2 = ip2+igauss2
            DO iparticle1 = 1, iparticle2
               ip1 = (iparticle1-1)*SIZE(radii)
               istart_g = 1
               IF (iparticle1 == iparticle2) istart_g = igauss2
               DO igauss1 = istart_g, SIZE(radii)
                  idim1 = ip1+igauss1
                  M(idim1, idim2) = cost(iparticle1, iparticle2)
                  M(idim2, idim1) = M(idim1, idim2)
               END DO
            END DO
         END DO
      END DO
      DEALLOCATE (cost)
      DEALLOCATE (pos)
      DEALLOCATE (LocP)
   END SUBROUTINE solvation_ddapc_pot

END MODULE cp_ddapc_methods
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains information regarding the decoupling/recoupling method of Bloechl
!> \author Teodoro Laino
! **************************************************************************************************
MODULE cp_ddapc_types
   USE cell_methods,                    ONLY: read_cell
   USE cell_types,                      ONLY: cell_release,&
                                              cell_type
   USE cp_ddapc_methods,                ONLY: ddapc_eval_AmI,&
                                              ddapc_eval_gfunc,&
                                              ewald_ddapc_pot,&
                                              solvation_ddapc_pot
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_get_default_io_unit,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_printkey_is_on
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE ewald_spline_util,               ONLY: Setup_Ewald_Spline
   USE input_section_types,             ONLY: section_vals_get,&
                                              section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE mathconstants,                   ONLY: pi
   USE particle_types,                  ONLY: particle_type
   USE pw_grid_types,                   ONLY: pw_grid_type
   USE pw_grids,                        ONLY: pw_grid_release
   USE pw_poisson_types,                ONLY: pw_poisson_multipole
   USE pw_pool_types,                   ONLY: pw_pool_give_back_pw,&
                                              pw_pool_release,&
                                              pw_pool_type
   USE pw_types,                        ONLY: pw_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   LOGICAL, PRIVATE, PARAMETER          :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'cp_ddapc_types'
   INTEGER, PRIVATE, SAVE               :: last_cp_ddapc_id = 0

   PUBLIC :: cp_ddapc_type, cp_ddapc_create, cp_ddapc_retain, cp_ddapc_release
   PUBLIC :: cp_ddapc_ewald_type, cp_ddapc_ewald_create, cp_ddapc_ewald_release

! **************************************************************************************************
!> \author Teodoro Laino
! **************************************************************************************************
   TYPE cp_ddapc_type
      INTEGER :: ref_count, id_nr
      REAL(KIND=dp) :: c0
      REAL(KIND=dp), DIMENSION(:, :), POINTER     :: AmI
      REAL(KIND=dp), DIMENSION(:, :), POINTER     :: Md ! decoupling
      REAL(KIND=dp), DIMENSION(:, :), POINTER     :: Mr ! recoupling
      REAL(KIND=dp), DIMENSION(:, :), POINTER     :: Mt ! decoupling+recoupling
      REAL(KIND=dp), DIMENSION(:, :), POINTER     :: Ms ! solvation
      REAL(KIND=dp), POINTER, DIMENSION(:, :)  :: gfunc
      REAL(KIND=dp), POINTER, DIMENSION(:)    :: w
   END TYPE cp_ddapc_type

! **************************************************************************************************
   TYPE cp_ddapc_ewald_type
      LOGICAL                                    :: do_decoupling
      LOGICAL                                    :: do_qmmm_periodic_decpl
      LOGICAL                                    :: do_solvation
      LOGICAL                                    :: do_property
      LOGICAL                                    :: do_restraint
      TYPE(section_vals_type), POINTER :: ewald_section
      TYPE(pw_pool_type), POINTER :: pw_pool_qm, pw_pool_mm
      TYPE(pw_grid_type), POINTER :: pw_grid_qm, pw_grid_mm
      TYPE(pw_type), POINTER :: coeff_qm, coeff_mm
   END TYPE cp_ddapc_ewald_type

CONTAINS

! **************************************************************************************************
!> \brief ...
!> \param cp_para_env ...
!> \param cp_ddapc_env ...
!> \param cp_ddapc_ewald ...
!> \param particle_set ...
!> \param radii ...
!> \param cell ...
!> \param super_cell ...
!> \param rho_tot_g ...
!> \param gcut ...
!> \param iw2 ...
!> \param Vol ...
!> \param force_env_section ...
!> \author Tedoro Laino
!> \note NB receive cp_para_env to pass down to parallelized ewald_ddapc_pot()
! **************************************************************************************************
   SUBROUTINE cp_ddapc_create(cp_para_env, cp_ddapc_env, cp_ddapc_ewald, &
                              particle_set, radii, cell, super_cell, rho_tot_g, gcut, iw2, Vol, &
                              force_env_section)
      TYPE(cp_para_env_type), POINTER                    :: cp_para_env
      TYPE(cp_ddapc_type), POINTER                       :: cp_ddapc_env
      TYPE(cp_ddapc_ewald_type), POINTER                 :: cp_ddapc_ewald
      TYPE(particle_type), DIMENSION(:), POINTER         :: particle_set
      REAL(kind=dp), DIMENSION(:), POINTER               :: radii
      TYPE(cell_type), POINTER                           :: cell, super_cell
      TYPE(pw_type), POINTER                             :: rho_tot_g
      REAL(KIND=dp), INTENT(IN)                          :: gcut
      INTEGER, INTENT(IN)                                :: iw2
      REAL(KIND=dp), INTENT(IN)                          :: Vol
      TYPE(section_vals_type), POINTER                   :: force_env_section

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_ddapc_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle
      TYPE(section_vals_type), POINTER                   :: param_section, solvation_section

      CALL timeset(routineN, handle)
      ALLOCATE (cp_ddapc_env)
      cp_ddapc_env%ref_count = 1
      last_cp_ddapc_id = last_cp_ddapc_id+1
      cp_ddapc_env%id_nr = last_cp_ddapc_id
      NULLIFY (cp_ddapc_env%AmI, &
               cp_ddapc_env%Md, &
               cp_ddapc_env%Mt, &
               cp_ddapc_env%Mr, &
               cp_ddapc_env%Ms, &
               cp_ddapc_env%gfunc, &
               cp_ddapc_env%w)
      ! Evaluates gfunc and AmI
      CALL ddapc_eval_gfunc(cp_ddapc_env%gfunc, cp_ddapc_env%w, gcut, rho_tot_g, radii)
      CALL ddapc_eval_AmI(cp_ddapc_env%AmI, &
                          cp_ddapc_env%c0, &
                          cp_ddapc_env%gfunc, &
                          cp_ddapc_env%w, &
                          particle_set, &
                          gcut, &
                          rho_tot_g, &
                          radii, &
                          iw2, &
                          Vol)
      IF (cp_ddapc_ewald%do_qmmm_periodic_decpl .OR. &
          cp_ddapc_ewald%do_decoupling) THEN
         !
         ! Evaluate the matrix for the Classical contribution to the coupling/decoupling scheme
         !
         param_section => cp_ddapc_ewald%ewald_section
         !NB parallelized ewald_ddapc_pot() needs cp_para_env
         CALL ewald_ddapc_pot(cp_para_env, cp_ddapc_ewald%coeff_qm, &
                              1.0_dp, &
                              cell, &
                              param_section, &
                              particle_set, &
                              cp_ddapc_env%Md, &
                              radii)
         IF (cp_ddapc_ewald%do_qmmm_periodic_decpl .OR. cp_ddapc_ewald%do_decoupling) THEN
            ALLOCATE (cp_ddapc_env%Mt(SIZE(cp_ddapc_env%Md, 1), SIZE(cp_ddapc_env%Md, 2)))
            IF (cp_ddapc_ewald%do_decoupling) THEN
               ! Just decoupling
               cp_ddapc_env%Mt = cp_ddapc_env%Md
            ELSE
               ! QMMM periodic calculation
               !NB parallelized ewald_ddapc_pot() needs cp_para_env
               CALL ewald_ddapc_pot(cp_para_env, cp_ddapc_ewald%coeff_mm, -1.0_dp, super_cell, param_section, &
                                    particle_set, cp_ddapc_env%Mr, radii)
               cp_ddapc_env%Mt = cp_ddapc_env%Md+cp_ddapc_env%Mr
            END IF
         END IF
      END IF
      IF (cp_ddapc_ewald%do_solvation) THEN
         ! Spherical Solvation model
         solvation_section => section_vals_get_subs_vals(force_env_section, "DFT%SCRF")
         CALL solvation_ddapc_pot(solvation_section, &
                                  particle_set, cp_ddapc_env%Ms, radii)
      END IF
      CALL timestop(handle)
   END SUBROUTINE cp_ddapc_create

! **************************************************************************************************
!> \brief ...
!> \param cp_ddapc_env ...
!> \par History
!>      none
!> \author Teodoro Laino - [tlaino]
! **************************************************************************************************
   SUBROUTINE cp_ddapc_retain(cp_ddapc_env)
      TYPE(cp_ddapc_type), POINTER                       :: cp_ddapc_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_ddapc_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(cp_ddapc_env)))CALL cp__a("cp_ddapc_types.F",191)
      IF(.NOT.(cp_ddapc_env%ref_count > 0))CALL cp__a("cp_ddapc_types.F",192)
      cp_ddapc_env%ref_count = cp_ddapc_env%ref_count+1
   END SUBROUTINE cp_ddapc_retain

! **************************************************************************************************
!> \brief ...
!> \param cp_ddapc_env ...
!> \par History
!>      none
!> \author Teodoro Laino - [tlaino]
! **************************************************************************************************
   SUBROUTINE cp_ddapc_release(cp_ddapc_env)
      TYPE(cp_ddapc_type), POINTER                       :: cp_ddapc_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_ddapc_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(cp_ddapc_env)) THEN
         IF(.NOT.(cp_ddapc_env%ref_count > 0))CALL cp__a("cp_ddapc_types.F",210)
         cp_ddapc_env%ref_count = cp_ddapc_env%ref_count-1
         IF (cp_ddapc_env%ref_count == 0) THEN
            IF (ASSOCIATED(cp_ddapc_env%AmI)) THEN
               DEALLOCATE (cp_ddapc_env%AmI)
            END IF
            IF (ASSOCIATED(cp_ddapc_env%Mt)) THEN
               DEALLOCATE (cp_ddapc_env%Mt)
            END IF
            IF (ASSOCIATED(cp_ddapc_env%Md)) THEN
               DEALLOCATE (cp_ddapc_env%Md)
            END IF
            IF (ASSOCIATED(cp_ddapc_env%Mr)) THEN
               DEALLOCATE (cp_ddapc_env%Mr)
            END IF
            IF (ASSOCIATED(cp_ddapc_env%Ms)) THEN
               DEALLOCATE (cp_ddapc_env%Ms)
            END IF
            IF (ASSOCIATED(cp_ddapc_env%gfunc)) THEN
               DEALLOCATE (cp_ddapc_env%gfunc)
            END IF
            IF (ASSOCIATED(cp_ddapc_env%w)) THEN
               DEALLOCATE (cp_ddapc_env%w)
            END IF
            DEALLOCATE (cp_ddapc_env)
         END IF
      END IF
   END SUBROUTINE cp_ddapc_release

! **************************************************************************************************
!> \brief ...
!> \param cp_ddapc_ewald ...
!> \param qmmm_decoupl ...
!> \param qm_cell ...
!> \param force_env_section ...
!> \param subsys_section ...
!> \param para_env ...
!> \par History
!>      none
!> \author Teodoro Laino - [tlaino]
! **************************************************************************************************
   SUBROUTINE cp_ddapc_ewald_create(cp_ddapc_ewald, qmmm_decoupl, qm_cell, &
                                    force_env_section, subsys_section, para_env)
      TYPE(cp_ddapc_ewald_type), POINTER                 :: cp_ddapc_ewald
      LOGICAL, INTENT(IN)                                :: qmmm_decoupl
      TYPE(cell_type), POINTER                           :: qm_cell
      TYPE(section_vals_type), POINTER                   :: force_env_section, subsys_section
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_ddapc_ewald_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: my_val, npts(3), output_unit
      INTEGER, DIMENSION(:), POINTER                     :: ngrids
      LOGICAL                                            :: analyt, decoupling, &
                                                            do_qmmm_periodic_decpl, do_restraint, &
                                                            do_restraintB, do_solvation
      REAL(KIND=dp)                                      :: hmat(3, 3)
      REAL(KIND=dp), DIMENSION(:), POINTER               :: gx, gy, gz, LG
      TYPE(cell_type), POINTER                           :: dummy_cell, mm_cell
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(section_vals_type), POINTER :: cell_section, grid_print_section, multipole_section, &
         poisson_section, printC_section, qmmm_per_section, restraint_section, restraint_sectionB, &
         solvation_section

      logger => cp_get_default_logger()
      output_unit = cp_logger_get_default_io_unit(logger)
      IF(.NOT.(.NOT. ASSOCIATED(cp_ddapc_ewald)))CALL cp__a("cp_ddapc_types.F",277)
      ALLOCATE (cp_ddapc_ewald)
      NULLIFY (cp_ddapc_ewald%pw_grid_mm, &
               cp_ddapc_ewald%pw_grid_qm, &
               cp_ddapc_ewald%ewald_section, &
               cp_ddapc_ewald%pw_pool_mm, &
               cp_ddapc_ewald%pw_pool_qm, &
               cp_ddapc_ewald%coeff_mm, &
               cp_ddapc_ewald%coeff_qm)
      NULLIFY (multipole_section)

      poisson_section => section_vals_get_subs_vals(force_env_section, "DFT%POISSON")
      solvation_section => section_vals_get_subs_vals(force_env_section, "DFT%SCRF")
      qmmm_per_section => section_vals_get_subs_vals(force_env_section, "QMMM%PERIODIC")
      printC_section => section_vals_get_subs_vals(force_env_section, "PROPERTIES%FIT_CHARGE")
      restraint_section => section_vals_get_subs_vals(force_env_section, "DFT%QS%DDAPC_RESTRAINT")
      restraint_sectionB => section_vals_get_subs_vals(force_env_section, &
                                                       "PROPERTIES%ET_COUPLING%DDAPC_RESTRAINT_A")
      CALL section_vals_get(solvation_section, explicit=do_solvation)
      CALL section_vals_get(poisson_section, explicit=decoupling)
      CALL section_vals_get(restraint_section, explicit=do_restraint)
      CALL section_vals_get(restraint_sectionB, explicit=do_restraintB)
      do_qmmm_periodic_decpl = qmmm_decoupl
      cp_ddapc_ewald%do_solvation = do_solvation
      cp_ddapc_ewald%do_qmmm_periodic_decpl = do_qmmm_periodic_decpl
      cp_ddapc_ewald%do_property = cp_printkey_is_on(logger%iter_info, printC_section)
      cp_ddapc_ewald%do_restraint = do_restraint .OR. do_restraintB
      ! Determining the tasks and further check
      IF (do_qmmm_periodic_decpl .AND. decoupling) THEN
         ! Check than an additional POISSON section has not been defined. In case write a warning
         IF (output_unit > 0) &
            WRITE (output_unit, '(T2,"WARNING",A)') &
            "A calculation with the QMMM periodic model has been requested.", &
            "The explicit POISSON section in DFT section will be IGNORED.", &
            "QM Electrostatic controlled only by the PERIODIC section in QMMM section"
         decoupling = .FALSE.
      END IF
      IF (decoupling) THEN
         ! Simple decoupling technique
         CALL section_vals_val_get(poisson_section, "POISSON_SOLVER", i_val=my_val)
         SELECT CASE (my_val)
         CASE (pw_poisson_multipole)
            multipole_section => section_vals_get_subs_vals(poisson_section, "MULTIPOLE")
         CASE DEFAULT
            decoupling = .FALSE.
         END SELECT
      END IF
      cp_ddapc_ewald%do_decoupling = decoupling
      IF (cp_ddapc_ewald%do_qmmm_periodic_decpl) THEN
         ! QMMM periodic
         multipole_section => section_vals_get_subs_vals(qmmm_per_section, "MULTIPOLE")
      END IF
      cp_ddapc_ewald%ewald_section => multipole_section
      IF (cp_ddapc_ewald%do_decoupling .OR. cp_ddapc_ewald%do_qmmm_periodic_decpl) THEN
         ! Do we do the calculation analytically or interpolating the g-space factor?
         CALL section_vals_val_get(multipole_section, "ANALYTICAL_GTERM", l_val=analyt)
         IF (.NOT. analyt) THEN
            CALL section_vals_val_get(multipole_section, "ngrids", i_vals=ngrids)
            npts = ngrids

            NULLIFY (LG, gx, gy, gz)
            hmat = qm_cell%hmat
            CALL eval_lg(multipole_section, hmat, qm_cell%deth, LG, gx, gy, gz)
            grid_print_section => section_vals_get_subs_vals(force_env_section, "PRINT%GRID_INFORMATION")
            CALL Setup_Ewald_Spline(pw_grid=cp_ddapc_ewald%pw_grid_qm, pw_pool=cp_ddapc_ewald%pw_pool_qm, &
                                    coeff=cp_ddapc_ewald%coeff_qm, LG=LG, gx=gx, gy=gy, gz=gz, hmat=hmat, npts=npts, &
                                    param_section=multipole_section, tag="ddapc", &
                                    para_env=para_env, print_section=grid_print_section)
            DEALLOCATE (LG)
            DEALLOCATE (gx)
            DEALLOCATE (gy)
            DEALLOCATE (gz)
            IF (cp_ddapc_ewald%do_qmmm_periodic_decpl) THEN
               NULLIFY (mm_cell, dummy_cell)
               cell_section => section_vals_get_subs_vals(subsys_section, "CELL")
               CALL read_cell(mm_cell, dummy_cell, cell_section=cell_section, para_env=para_env)
               hmat = mm_cell%hmat
               CALL eval_lg(multipole_section, hmat, mm_cell%deth, LG, gx, gy, gz)
               grid_print_section => section_vals_get_subs_vals(force_env_section, "PRINT%GRID_INFORMATION")
               CALL Setup_Ewald_Spline(pw_grid=cp_ddapc_ewald%pw_grid_mm, pw_pool=cp_ddapc_ewald%pw_pool_mm, &
                                       coeff=cp_ddapc_ewald%coeff_mm, LG=LG, gx=gx, gy=gy, gz=gz, hmat=hmat, npts=npts, &
                                       param_section=multipole_section, tag="ddapc", para_env=para_env, &
                                       print_section=grid_print_section)
               DEALLOCATE (LG)
               DEALLOCATE (gx)
               DEALLOCATE (gy)
               DEALLOCATE (gz)
               CALL cell_release(dummy_cell)
               CALL cell_release(mm_cell)
            END IF
         END IF
      END IF
   END SUBROUTINE cp_ddapc_ewald_create

! **************************************************************************************************
!> \brief ...
!> \param multipole_section ...
!> \param hmat ...
!> \param deth ...
!> \param LG ...
!> \param gx ...
!> \param gy ...
!> \param gz ...
!> \par History
!>      none
!> \author Teodoro Laino - [tlaino]
! **************************************************************************************************
   SUBROUTINE eval_lg(multipole_section, hmat, deth, LG, gx, gy, gz)
      TYPE(section_vals_type), POINTER                   :: multipole_section
      REAL(KIND=dp), INTENT(IN)                          :: hmat(3, 3), deth
      REAL(KIND=dp), DIMENSION(:), POINTER               :: LG, gx, gy, gz

      CHARACTER(len=*), PARAMETER :: routineN = 'eval_lg', routineP = moduleN//':'//routineN

      INTEGER                                            :: i, k1, k2, k3, n_rep, ndim, nmax1, &
                                                            nmax2, nmax3
      REAL(KIND=dp)                                      :: alpha, eps, fac, fs, fvec(3), galpha, &
                                                            gsq, gsqi, rcut, tol, tol1

      rcut = MIN(hmat(1, 1), hmat(2, 2), hmat(3, 3))/2.0_dp
      CALL section_vals_val_get(multipole_section, "RCUT", n_rep_val=n_rep)
      IF (n_rep == 1) CALL section_vals_val_get(multipole_section, "RCUT", r_val=rcut)
      CALL section_vals_val_get(multipole_section, "EWALD_PRECISION", r_val=eps)
      eps = MIN(ABS(eps), 0.5_dp)
      tol = SQRT(ABS(LOG(eps*rcut)))
      alpha = SQRT(ABS(LOG(eps*rcut*tol)))/rcut
      galpha = 1.0_dp/(4.0_dp*alpha*alpha)
      tol1 = SQRT(-LOG(eps*rcut*(2.0_dp*tol*alpha)**2))
      nmax1 = NINT(0.25_dp+hmat(1, 1)*alpha*tol1/pi)
      nmax2 = NINT(0.25_dp+hmat(2, 2)*alpha*tol1/pi)
      nmax3 = NINT(0.25_dp+hmat(3, 3)*alpha*tol1/pi)
      fac = 1.e0_dp/deth
      fvec = 2.0_dp*pi/(/hmat(1, 1), hmat(2, 2), hmat(3, 3)/)
      ndim = (nmax1+1)*(2*nmax2+1)*(2*nmax3+1)-1
      ALLOCATE (LG(ndim))
      ALLOCATE (gx(ndim))
      ALLOCATE (gy(ndim))
      ALLOCATE (gz(ndim))

      i = 0
      DO k1 = 0, nmax1
         DO k2 = -nmax2, nmax2
            DO k3 = -nmax3, nmax3
               IF (k1 == 0 .AND. k2 == 0 .AND. k3 == 0) CYCLE
               i = i+1
               fs = 2.0_dp; IF (k1 == 0) fs = 1.0_dp
               gx(i) = fvec(1)*REAL(k1, KIND=dp)
               gy(i) = fvec(2)*REAL(k2, KIND=dp)
               gz(i) = fvec(3)*REAL(k3, KIND=dp)
               gsq = gx(i)*gx(i)+gy(i)*gy(i)+gz(i)*gz(i)
               gsqi = fs/gsq
               LG(i) = fac*gsqi*EXP(-galpha*gsq)
            END DO
         END DO
      END DO

   END SUBROUTINE eval_lg

! **************************************************************************************************
!> \brief ...
!> \param cp_ddapc_ewald ...
!> \par History
!>      none
!> \author Teodoro Laino - [tlaino]
! **************************************************************************************************
   SUBROUTINE cp_ddapc_ewald_release(cp_ddapc_ewald)
      TYPE(cp_ddapc_ewald_type), POINTER                 :: cp_ddapc_ewald

      CHARACTER(len=*), PARAMETER :: routineN = 'cp_ddapc_ewald_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(cp_ddapc_ewald)) THEN
         IF (ASSOCIATED(cp_ddapc_ewald%coeff_qm)) THEN
            CALL pw_pool_give_back_pw(cp_ddapc_ewald%pw_pool_qm, cp_ddapc_ewald%coeff_qm)
         END IF
         IF (ASSOCIATED(cp_ddapc_ewald%coeff_mm)) THEN
            CALL pw_pool_give_back_pw(cp_ddapc_ewald%pw_pool_mm, cp_ddapc_ewald%coeff_mm)
         END IF
         IF (ASSOCIATED(cp_ddapc_ewald%pw_pool_qm)) THEN
            CALL pw_pool_release(cp_ddapc_ewald%pw_pool_qm)
            IF(.NOT.(.NOT. ASSOCIATED(cp_ddapc_ewald%pw_pool_qm)))CALL cp__a("cp_ddapc_types.F",457)
         END IF
         IF (ASSOCIATED(cp_ddapc_ewald%pw_pool_mm)) THEN
            CALL pw_pool_release(cp_ddapc_ewald%pw_pool_mm)
            IF(.NOT.(.NOT. ASSOCIATED(cp_ddapc_ewald%pw_pool_mm)))CALL cp__a("cp_ddapc_types.F",461)
         END IF
         IF (ASSOCIATED(cp_ddapc_ewald%pw_grid_qm)) THEN
            CALL pw_grid_release(cp_ddapc_ewald%pw_grid_qm)
            IF(.NOT.(.NOT. ASSOCIATED(cp_ddapc_ewald%pw_grid_qm)))CALL cp__a("cp_ddapc_types.F",465)
         END IF
         IF (ASSOCIATED(cp_ddapc_ewald%pw_grid_mm)) THEN
            CALL pw_grid_release(cp_ddapc_ewald%pw_grid_mm)
            IF(.NOT.(.NOT. ASSOCIATED(cp_ddapc_ewald%pw_grid_mm)))CALL cp__a("cp_ddapc_types.F",469)
         END IF
         DEALLOCATE (cp_ddapc_ewald)
      END IF

   END SUBROUTINE cp_ddapc_ewald_release

END MODULE cp_ddapc_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_fb_atomic_halo_types

   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE cell_types,                      ONLY: cell_type,&
                                              pbc
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_p_file,&
                                              cp_print_key_finished_output,&
                                              cp_print_key_should_output,&
                                              cp_print_key_unit_nr
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_section_types,             ONLY: section_vals_type
   USE kinds,                           ONLY: default_string_length,&
                                              dp
   USE memory_utilities,                ONLY: reallocate
   USE message_passing,                 ONLY: mp_max,&
                                              mp_min,&
                                              mp_sum
   USE particle_types,                  ONLY: particle_type
   USE qs_kind_types,                   ONLY: get_qs_kind,&
                                              qs_kind_type
   USE string_utilities,                ONLY: compress
   USE util,                            ONLY: locate,&
                                              sort
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! public types
   PUBLIC :: fb_atomic_halo_obj, &
             fb_atomic_halo_list_obj

! public methods
   PUBLIC :: fb_atomic_halo_release, &
             fb_atomic_halo_nullify, &
             fb_atomic_halo_has_data, &
             fb_atomic_halo_create, &
             fb_atomic_halo_init, &
             fb_atomic_halo_get, &
             fb_atomic_halo_set, &
             fb_atomic_halo_sort, &
             fb_atomic_halo_atom_global2halo, &
             fb_atomic_halo_nelectrons_estimate_Z, &
             fb_atomic_halo_cost, &
             fb_atomic_halo_build_halo_atoms, &
             fb_atomic_halo_list_retain, &
             fb_atomic_halo_list_release, &
             fb_atomic_halo_list_nullify, &
             fb_atomic_halo_list_has_data, &
             fb_atomic_halo_list_associate, &
             fb_atomic_halo_list_create, &
             fb_atomic_halo_list_get, &
             fb_atomic_halo_list_set, &
             fb_atomic_halo_list_write_info, &
             fb_build_pair_radii

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_atomic_halo_types'
   INTEGER, PRIVATE, SAVE :: last_fb_atomic_halo_id = 0
   INTEGER, PRIVATE, SAVE :: last_fb_atomic_halo_list_id = 0

! **************************************************************************************************
!> \brief derived type containing the list of atoms in an atomic halo,
!>        used by filtered-basis diagonalisation method
!> \param id_nr            : unique id for the object
!> \param ref_count        : reference count on the object
!> \param owner_atom       : global atomic id of the atom this halo belongs to
!> \param owner_id_in_halo : index of the owner_atom in the halo_atoms array
!> \param natoms           : number of atoms in the halo
!> \param nelectrons       : estimate of total number of electrons in halo
!> \param halo_atoms       : the list of global id of atoms in the halo
!> \param sorted           : whether the halo_atoms list is sorted or not
!> \param cost             : computational cost for the atomic matrix associated
!>                           to this atomic halo
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_atomic_halo_data
      INTEGER :: id_nr, ref_count
      INTEGER :: owner_atom
      INTEGER :: owner_id_in_halo
      INTEGER :: natoms
      INTEGER :: nelectrons
      INTEGER, DIMENSION(:), POINTER :: halo_atoms
      LOGICAL :: sorted
      REAL(KIND=dp) :: cost
   END TYPE fb_atomic_halo_data

! **************************************************************************************************
!> \brief defines a fb_atomic_halo object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_atomic_halo_obj
      TYPE(fb_atomic_halo_data), POINTER, PRIVATE :: obj
   END TYPE fb_atomic_halo_obj

! **************************************************************************************************
!> \brief derived type describing an atomic halo list used by
!>        filtered-basis diagonalisation method
!> \param id_nr     : unique id for the object
!> \param ref_count : reference count on the object
!> \param nhalos  : number of halos in the list
!> \param max_nhalos : maximum of the number of halos amongst all of the procs
!> \param halos   : halos(ihalo) gives the ihalo-th fb_atomic_halo object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_atomic_halo_list_data
      INTEGER :: id_nr, ref_count
      INTEGER :: nhalos
      INTEGER :: max_nhalos
      TYPE(fb_atomic_halo_obj), DIMENSION(:), POINTER :: halos
   END TYPE fb_atomic_halo_list_data

! **************************************************************************************************
!> \brief defines a fb_atomic_halo_list object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_atomic_halo_list_obj
      TYPE(fb_atomic_halo_list_data), POINTER, PRIVATE :: obj
   END TYPE fb_atomic_halo_list_obj

CONTAINS

! **************************************************************************************************
!> \brief Retains an fb_atomic_halo object
!> \param atomic_halo the fb_atomic_halo object, its content must
!>                     not be NULL or UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_retain(atomic_halo)
      ! note INTENT(IN) is okay because the obj pointer contained in the
      ! obj type will not be changed
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atomic_halo%obj)))CALL cp__a("qs_fb_atomic_halo_types.F",145)
      IF(.NOT.(atomic_halo%obj%ref_count > 0))CALL cp__a("qs_fb_atomic_halo_types.F",146)
      atomic_halo%obj%ref_count = atomic_halo%obj%ref_count+1
   END SUBROUTINE fb_atomic_halo_retain

! **************************************************************************************************
!> \brief Releases an fb_atomic_halo object
!> \param atomic_halo the fb_atomic_halo object, its content must
!>                     not be UNDEFINED, and the subroutine does nothing
!>                     if the content points to NULL
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_release(atomic_halo)
      TYPE(fb_atomic_halo_obj), INTENT(INOUT)            :: atomic_halo

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(atomic_halo%obj)) THEN
         IF(.NOT.(atomic_halo%obj%ref_count > 0))CALL cp__a("qs_fb_atomic_halo_types.F",164)
         atomic_halo%obj%ref_count = atomic_halo%obj%ref_count-1
         IF (atomic_halo%obj%ref_count == 0) THEN
            atomic_halo%obj%ref_count = 1
            IF (ASSOCIATED(atomic_halo%obj%halo_atoms)) THEN
               ! note that if there are other pointers associated to the memory pointed
               ! by atomic_halo%obj%halo_atoms, their behaviour becomes undefined per
               ! FORTRAN standard (and thus becomes compiler dependent and unreliable)
               ! after the following DEALLOCATE
               DEALLOCATE (atomic_halo%obj%halo_atoms)
            END IF
            atomic_halo%obj%ref_count = 0
            DEALLOCATE (atomic_halo%obj)
         END IF
      ELSE
         NULLIFY (atomic_halo%obj)
      END IF
   END SUBROUTINE fb_atomic_halo_release

! **************************************************************************************************
!> \brief Nullifies a fb_atomic_halo object, note that it does not
!>        release the original object. This procedure is used to nullify
!>        the pointer contained in the object which is used to associate
!>        to the actual object content
!> \param atomic_halo the fb_atomic_halo object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_nullify(atomic_halo)
      TYPE(fb_atomic_halo_obj), INTENT(INOUT)            :: atomic_halo

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_nullify', &
         routineP = moduleN//':'//routineN

      NULLIFY (atomic_halo%obj)
   END SUBROUTINE fb_atomic_halo_nullify

! **************************************************************************************************
!> \brief Associates one fb_atomic_halo object to another
!> \param a the fb_atomic_halo object to be associated
!> \param b the fb_atomic_halo object that a is to be associated to
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_associate(a, b)
      TYPE(fb_atomic_halo_obj), INTENT(OUT)              :: a
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: b

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_associate', &
         routineP = moduleN//':'//routineN

      a%obj => b%obj
   END SUBROUTINE fb_atomic_halo_associate

! **************************************************************************************************
!> \brief Checks if a fb_atomic_halo object is associated with an actual
!>        data content or not
!> \param atomic_halo the fb_atomic_halo object
!> \retval res ...
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   FUNCTION fb_atomic_halo_has_data(atomic_halo) RESULT(res)
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_has_data', &
         routineP = moduleN//':'//routineN

      res = ASSOCIATED(atomic_halo%obj)
   END FUNCTION fb_atomic_halo_has_data

! **************************************************************************************************
!> \brief Creates and initialises an empty fb_atomic_halo object
!> \param atomic_halo the fb_atomic_halo object, its content must
!>                     be NULL and cannot be UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_create(atomic_halo)
      TYPE(fb_atomic_halo_obj), INTENT(INOUT)            :: atomic_halo

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(atomic_halo%obj)))CALL cp__a("qs_fb_atomic_halo_types.F",245)
      ALLOCATE (atomic_halo%obj)
      atomic_halo%obj%owner_atom = 0
      atomic_halo%obj%owner_id_in_halo = 0
      atomic_halo%obj%natoms = 0
      atomic_halo%obj%nelectrons = 0
      atomic_halo%obj%sorted = .FALSE.
      atomic_halo%obj%cost = 0.0_dp
      NULLIFY (atomic_halo%obj%halo_atoms)
      atomic_halo%obj%ref_count = 1
      atomic_halo%obj%id_nr = last_fb_atomic_halo_id+1
      last_fb_atomic_halo_id = atomic_halo%obj%id_nr
   END SUBROUTINE fb_atomic_halo_create

! **************************************************************************************************
!> \brief Initialises an fb_atomic_halo object, and makes it empty
!> \param atomic_halo the fb_atomic_halo object, its content must
!>                     not be NULL or UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_init(atomic_halo)
      TYPE(fb_atomic_halo_obj), INTENT(INOUT)            :: atomic_halo

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_init', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atomic_halo%obj)))CALL cp__a("qs_fb_atomic_halo_types.F",271)
      ! if halo_atoms are associated, then deallocate and de-associate
      IF (ASSOCIATED(atomic_halo%obj%halo_atoms)) THEN
         DEALLOCATE (atomic_halo%obj%halo_atoms)
      END IF
      atomic_halo%obj%owner_atom = 0
      atomic_halo%obj%owner_id_in_halo = 0
      atomic_halo%obj%natoms = 0
      atomic_halo%obj%nelectrons = 0
      atomic_halo%obj%sorted = .FALSE.
      atomic_halo%obj%cost = 0.0_dp
   END SUBROUTINE fb_atomic_halo_init

! **************************************************************************************************
!> \brief Gets attributes from a fb_atomic_halo object, one should
!>        only access the data content in a fb_atomic_halo outside
!>        this module via this procedure.
!> \param atomic_halo the fb_atomic_halo object, its content must
!>                     not be NULL or UNDEFINED
!> \param owner_atom [OPTIONAL]: if present, outputs atmic_halo%obj%owner_atom
!> \param owner_id_in_halo ...
!> \param natoms [OPTIONAL]: if present, outputs atomic_halo%obj%natoms
!> \param nelectrons [OPTIONAL]: if present, outputs atomic_halo%obj%nelectrons
!> \param halo_atoms [OPTIONAL]: if present, outputs pointer
!>                               atomic_halo%obj%halo_atoms
!> \param sorted [OPTIONAL]: if present, outputs atomic_halo%obj%sorted
!> \param cost [OPTIONAL]: if present, outputs atomic_halo%obj%cost
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_get(atomic_halo, &
                                 owner_atom, &
                                 owner_id_in_halo, &
                                 natoms, &
                                 nelectrons, &
                                 halo_atoms, &
                                 sorted, &
                                 cost)
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      INTEGER, INTENT(OUT), OPTIONAL                     :: owner_atom, owner_id_in_halo, natoms, &
                                                            nelectrons
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: halo_atoms
      LOGICAL, INTENT(OUT), OPTIONAL                     :: sorted
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: cost

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_get', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atomic_halo%obj)))CALL cp__a("qs_fb_atomic_halo_types.F",318)
      IF (PRESENT(owner_atom)) owner_atom = atomic_halo%obj%owner_atom
      IF (PRESENT(owner_id_in_halo)) owner_id_in_halo = atomic_halo%obj%owner_id_in_halo
      IF (PRESENT(natoms)) natoms = atomic_halo%obj%natoms
      IF (PRESENT(nelectrons)) nelectrons = atomic_halo%obj%nelectrons
      IF (PRESENT(halo_atoms)) halo_atoms => atomic_halo%obj%halo_atoms
      IF (PRESENT(sorted)) sorted = atomic_halo%obj%sorted
      IF (PRESENT(cost)) cost = atomic_halo%obj%cost
   END SUBROUTINE fb_atomic_halo_get

! **************************************************************************************************
!> \brief Sets attributes in a fb_atomic_halo object, one should
!>        only set the data content in a fb_atomic_halo from outside
!>        this module via this procedure.
!> \param atomic_halo the fb_atomic_halo object, its content must
!>                     not be NULL or UNDEFINED
!> \param owner_atom [OPTIONAL]: if present, sets
!>                               atmic_halo%obj%owner_atom = owner_atom
!> \param owner_id_in_halo ...
!> \param natoms [OPTIONAL]: if present, sets atomic_halo%obj%natoms = natoms
!> \param nelectrons [OPTIONAL]: if present, sets atomic_halo%obj%nelectrons = nelectrons
!> \param halo_atoms [OPTIONAL]: if present, reallocates atomic_halo%obj%halo_atoms
!>                               to the size of halo_atoms, and copies
!>                               contents of halo_atoms to atomic_halo%obj%halo_atoms
!> \param sorted [OPTIONAL]: if present, sets atomic_halo%obj%sorted = sorted
!> \param cost [OPTIONAL]: if present, sets atomic_halo%obj%cost = cost
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_set(atomic_halo, &
                                 owner_atom, &
                                 owner_id_in_halo, &
                                 natoms, &
                                 nelectrons, &
                                 halo_atoms, &
                                 sorted, &
                                 cost)
      TYPE(fb_atomic_halo_obj), INTENT(INOUT)            :: atomic_halo
      INTEGER, INTENT(IN), OPTIONAL                      :: owner_atom, owner_id_in_halo, natoms, &
                                                            nelectrons
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: halo_atoms
      LOGICAL, INTENT(IN), OPTIONAL                      :: sorted
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: cost

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_set', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atomic_halo%obj)))CALL cp__a("qs_fb_atomic_halo_types.F",364)
      IF (PRESENT(owner_atom)) atomic_halo%obj%owner_atom = owner_atom
      IF (PRESENT(owner_id_in_halo)) atomic_halo%obj%owner_id_in_halo = owner_id_in_halo
      IF (PRESENT(natoms)) atomic_halo%obj%natoms = natoms
      IF (PRESENT(nelectrons)) atomic_halo%obj%nelectrons = nelectrons
      IF (PRESENT(halo_atoms)) THEN
         IF (ASSOCIATED(atomic_halo%obj%halo_atoms)) THEN
            DEALLOCATE (atomic_halo%obj%halo_atoms)
         END IF
         atomic_halo%obj%halo_atoms => halo_atoms
      END IF
      IF (PRESENT(nelectrons)) atomic_halo%obj%nelectrons = nelectrons
      IF (PRESENT(sorted)) atomic_halo%obj%sorted = sorted
      IF (PRESENT(cost)) atomic_halo%obj%cost = cost
   END SUBROUTINE fb_atomic_halo_set

! **************************************************************************************************
!> \brief Sort the list of atomic indices in the halo in accending order.
!>        The atomic_halo must not be empty
!> \param atomic_halo  the atomic_halo object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_sort(atomic_halo)
      TYPE(fb_atomic_halo_obj), INTENT(INOUT)            :: atomic_halo

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_sort', &
         routineP = moduleN//':'//routineN

      INTEGER, ALLOCATABLE, DIMENSION(:)                 :: tmp_index

      IF(.NOT.(SIZE(atomic_halo%obj%halo_atoms) > 0))CALL cp__a("qs_fb_atomic_halo_types.F",394)
      ALLOCATE (tmp_index(atomic_halo%obj%natoms))
      CALL sort(atomic_halo%obj%halo_atoms, atomic_halo%obj%natoms, tmp_index)
      DEALLOCATE (tmp_index)
      atomic_halo%obj%sorted = .TRUE.
   END SUBROUTINE fb_atomic_halo_sort

! **************************************************************************************************
!> \brief Given a global atomic index, convert it to its index in a
!>        given atomic halo, if found.
!>        The atomic_halo object must already have been sorted
!> \param atomic_halo  the atomic_halo object
!> \param iatom_global  the global atomic index
!> \param iatom_halo  the atomic index inside the halo
!> \param found  returns true if given atom is in the halo, otherwise
!>                false
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_atom_global2halo(atomic_halo, &
                                              iatom_global, &
                                              iatom_halo, &
                                              found)
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      INTEGER, INTENT(IN)                                :: iatom_global
      INTEGER, INTENT(OUT)                               :: iatom_halo
      LOGICAL, INTENT(OUT)                               :: found

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_atom_global2halo', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(atomic_halo%obj%sorted))CALL cp__a("qs_fb_atomic_halo_types.F",428)
      iatom_halo = locate(atomic_halo%obj%halo_atoms, iatom_global)
      IF (iatom_halo == 0) THEN
         found = .FALSE.
      ELSE
         found = .TRUE.
      END IF

      CALL timestop(handle)

   END SUBROUTINE fb_atomic_halo_atom_global2halo

! **************************************************************************************************
!> \brief Estimates the total number of electrons in a halo using atomic
!>        numbers
!> \param atomic_halo the atomic_halo object
!> \param particle_set an array of cp2k particle set objects (this
!>                       gives atomic information)
!> \retval nelectrons estimate of electron number
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   FUNCTION fb_atomic_halo_nelectrons_estimate_Z(atomic_halo, particle_set) RESULT(nelectrons)
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: particle_set
      INTEGER                                            :: nelectrons

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_nelectrons_estimate_Z', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom_global, iatom_halo, z
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      nelectrons = 0
      IF (ASSOCIATED(atomic_halo%obj)) THEN
         DO iatom_halo = 1, atomic_halo%obj%natoms
            iatom_global = atomic_halo%obj%halo_atoms(iatom_halo)
            atomic_kind => particle_set(iatom_global)%atomic_kind
            CALL get_atomic_kind(atomic_kind=atomic_kind, &
                                 z=z)
            nelectrons = nelectrons+z
         END DO
      END IF
   END FUNCTION fb_atomic_halo_nelectrons_estimate_Z

! **************************************************************************************************
!> \brief Estimates the computational cost with respect to the filter matrix
!>        calculation associated to an atomic halo. Given the bottle neck
!>        of the filter matrix generation will be the diagoanlisation of the
!>        atomic matrices (each consists of atoms in an atomic halo), the cost
!>        can be estimated by counting the total number of contracted gaussians
!>        in the halo
!> \param atomic_halo  : the atomic_halo object in question
!> \param particle_set : an array of cp2k particle set objects, this
!>                       provides atomic information
!> \param qs_kind_set  : cp2k qs_kind objects, provides information on the
!>                       number of contracted gaussian functions each kind
!>                       has
!> \retval cost        : computation cost w.r.t. the filter matrix
!>                       calculation for this atomic halo
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   FUNCTION fb_atomic_halo_cost(atomic_halo, &
                                particle_set, &
                                qs_kind_set) &
      RESULT(cost)
      TYPE(fb_atomic_halo_obj), INTENT(IN)               :: atomic_halo
      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: particle_set
      TYPE(qs_kind_type), DIMENSION(:), INTENT(IN)       :: qs_kind_set
      REAL(KIND=dp)                                      :: cost

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_cost', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, ii, ikind, ncgf

      cost = 0.0_dp
      DO ii = 1, atomic_halo%obj%natoms
         iatom = atomic_halo%obj%halo_atoms(ii)
         CALL get_atomic_kind(atomic_kind=particle_set(iatom)%atomic_kind, &
                              kind_number=ikind)
         CALL get_qs_kind(qs_kind=qs_kind_set(ikind), &
                          ncgf=ncgf)
         cost = cost+REAL(ncgf, dp)
      END DO
      ! diagonalisation is N**3 process, so cost must reflect that
      cost = cost**3
   END FUNCTION fb_atomic_halo_cost

! **************************************************************************************************
!> \brief Builds halo atoms for a given (owner) atom
!> \param owner_atom   : the atom the halo is going to be built for
!> \param particle_set : an array of cp2k particle set objects, this
!>                       provides atomic information
!> \param cell         : cp2k cell object, used for resolving periodic
!>                       boundary conditions
!> \param pair_radii   : 2D array storing interaction radii between two kinds
!> \param halo_atoms   : must be NULL pointer on input, and outputs an
!>                       array of halo atoms corresponding to the owner atom
!> \param nhalo_atoms  : outputs number of halo atoms
!> \param owner_id_in_halo : the index of the owner atom in the halo_atoms list
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_build_halo_atoms(owner_atom, &
                                              particle_set, &
                                              cell, &
                                              pair_radii, &
                                              halo_atoms, &
                                              nhalo_atoms, &
                                              owner_id_in_halo)
      INTEGER, INTENT(IN)                                :: owner_atom
      TYPE(particle_type), DIMENSION(:), INTENT(IN)      :: particle_set
      TYPE(cell_type), POINTER                           :: cell
      REAL(KIND=dp), DIMENSION(:, :), INTENT(IN)         :: pair_radii
      INTEGER, DIMENSION(:), POINTER                     :: halo_atoms
      INTEGER, INTENT(OUT)                               :: nhalo_atoms, owner_id_in_halo

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_build_halo_atoms', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, ikind, jatom, jkind, natoms_global
      LOGICAL                                            :: check_ok
      REAL(KIND=dp)                                      :: rij
      REAL(KIND=dp), DIMENSION(3)                        :: ri, rij_pbc, rj
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind

      check_ok = .NOT. ASSOCIATED(halo_atoms)
      IF(.NOT.(check_ok))CALL cp__a("qs_fb_atomic_halo_types.F",554)

      NULLIFY (atomic_kind)

      iatom = owner_atom
      atomic_kind => particle_set(iatom)%atomic_kind
      CALL get_atomic_kind(atomic_kind=atomic_kind, &
                           kind_number=ikind)
      natoms_global = SIZE(particle_set)
      ALLOCATE (halo_atoms(natoms_global))
      owner_id_in_halo = 0
      nhalo_atoms = 0
      DO jatom = 1, natoms_global
         atomic_kind => particle_set(jatom)%atomic_kind
         CALL get_atomic_kind(atomic_kind=atomic_kind, &
                              kind_number=jkind)
         ! calculate the minimum distance between iatom and
         ! jatom, taking account of the periodic boundary
         ! conditions
         ri(1:3) = particle_set(iatom)%r(1:3)
         rj(1:3) = particle_set(jatom)%r(1:3)
         rij_pbc = pbc(ri, rj, cell)
         rij = rij_pbc(1)*rij_pbc(1)+ &
               rij_pbc(2)*rij_pbc(2)+ &
               rij_pbc(3)*rij_pbc(3)
         rij = SQRT(rij)
         IF (rij .LE. pair_radii(ikind, jkind)) THEN
            ! jatom is in iatom's halo
            nhalo_atoms = nhalo_atoms+1
            halo_atoms(nhalo_atoms) = jatom
            IF (jatom == iatom) owner_id_in_halo = nhalo_atoms
         END IF
      END DO
      CALL reallocate(halo_atoms, 1, nhalo_atoms)
   END SUBROUTINE fb_atomic_halo_build_halo_atoms

! **************************************************************************************************
!> \brief Retains an fb_atomic_halo_list object
!> \param atomic_halos the fb_atomic_halo object, its content must
!>                      not be NULL or UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_list_retain(atomic_halos)
      TYPE(fb_atomic_halo_list_obj), INTENT(IN)          :: atomic_halos

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_list_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atomic_halos%obj)))CALL cp__a("qs_fb_atomic_halo_types.F",602)
      IF(.NOT.(atomic_halos%obj%ref_count > 0))CALL cp__a("qs_fb_atomic_halo_types.F",603)
      atomic_halos%obj%ref_count = atomic_halos%obj%ref_count+1
   END SUBROUTINE fb_atomic_halo_list_retain

! **************************************************************************************************
!> \brief Releases an fb_atomic_halo_list object
!> \param atomic_halos the fb_atomic_halo object, its content must
!>                      not be UNDEFINED, and does nothing if it is NULL
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_list_release(atomic_halos)
      TYPE(fb_atomic_halo_list_obj), INTENT(INOUT)       :: atomic_halos

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_list_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii

      IF (ASSOCIATED(atomic_halos%obj)) THEN
         IF(.NOT.(atomic_halos%obj%ref_count > 0))CALL cp__a("qs_fb_atomic_halo_types.F",622)
         atomic_halos%obj%ref_count = atomic_halos%obj%ref_count-1
         IF (atomic_halos%obj%ref_count == 0) THEN
            atomic_halos%obj%ref_count = 1
            IF (ASSOCIATED(atomic_halos%obj%halos)) THEN
               DO ii = 1, SIZE(atomic_halos%obj%halos)
                  CALL fb_atomic_halo_release(atomic_halos%obj%halos(ii))
               END DO
               DEALLOCATE (atomic_halos%obj%halos)
            END IF
            atomic_halos%obj%ref_count = 0
            DEALLOCATE (atomic_halos%obj)
         END IF
      ELSE
         NULLIFY (atomic_halos%obj)
      END IF
   END SUBROUTINE fb_atomic_halo_list_release

! **************************************************************************************************
!> \brief Nullifies a fb_atomic_halo_list object, note that it does
!>        not release the original object. This procedure is used to
!>        nullify the pointer contained in the object which is used to
!>        associate to the actual object content
!> \param atomic_halos the fb_atomic_halo_list object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_list_nullify(atomic_halos)
      TYPE(fb_atomic_halo_list_obj), INTENT(INOUT)       :: atomic_halos

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_list_nullify', &
         routineP = moduleN//':'//routineN

      NULLIFY (atomic_halos%obj)
   END SUBROUTINE fb_atomic_halo_list_nullify

! **************************************************************************************************
!> \brief Checks if a fb_atomic_halo_list object is associated with
!>        an actual data content or not
!> \param atomic_halos the fb_atomic_halo_list object
!> \retval res ...
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   FUNCTION fb_atomic_halo_list_has_data(atomic_halos) RESULT(res)
      TYPE(fb_atomic_halo_list_obj), INTENT(IN)          :: atomic_halos
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_list_has_data', &
         routineP = moduleN//':'//routineN

      res = ASSOCIATED(atomic_halos%obj)
   END FUNCTION fb_atomic_halo_list_has_data

! **************************************************************************************************
!> \brief Associates one fb_atomic_halo_list object to another
!> \param a the fb_atomic_halo_list object to be associated
!> \param b the fb_atomic_halo_list object that a is to be associated to
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_list_associate(a, b)
      TYPE(fb_atomic_halo_list_obj), INTENT(OUT)         :: a
      TYPE(fb_atomic_halo_list_obj), INTENT(IN)          :: b

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_list_associate', &
         routineP = moduleN//':'//routineN

      a%obj => b%obj
   END SUBROUTINE fb_atomic_halo_list_associate

! **************************************************************************************************
!> \brief Creates and initialises an empty fb_atomic_halo_list object
!> \param atomic_halos the fb_atomic_halo object, its content must
!>                      not be NULL or UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_list_create(atomic_halos)
      TYPE(fb_atomic_halo_list_obj), INTENT(INOUT)       :: atomic_halos

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_list_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(atomic_halos%obj)))CALL cp__a("qs_fb_atomic_halo_types.F",702)
      ALLOCATE (atomic_halos%obj)
      atomic_halos%obj%nhalos = 0
      atomic_halos%obj%max_nhalos = 0
      NULLIFY (atomic_halos%obj%halos)
      atomic_halos%obj%ref_count = 1
      atomic_halos%obj%id_nr = last_fb_atomic_halo_list_id+1
      last_fb_atomic_halo_list_id = atomic_halos%obj%id_nr
   END SUBROUTINE fb_atomic_halo_list_create

! **************************************************************************************************
!> \brief Initialises an fb_atomic_halo_list object and make it empty
!> \param atomic_halos the fb_atomic_halo object, its content must
!>                      not be NULL or UNDEFINED
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_list_init(atomic_halos)
      TYPE(fb_atomic_halo_list_obj), INTENT(INOUT)       :: atomic_halos

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_list_init', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ii

      IF(.NOT.(ASSOCIATED(atomic_halos%obj)))CALL cp__a("qs_fb_atomic_halo_types.F",726)
      ! if the arrays are associated, then deallocate and de-associate
      IF (ASSOCIATED(atomic_halos%obj%halos)) THEN
         DO ii = 1, SIZE(atomic_halos%obj%halos)
            CALL fb_atomic_halo_release(atomic_halos%obj%halos(ii))
         END DO
         DEALLOCATE (atomic_halos%obj%halos)
      END IF
      atomic_halos%obj%nhalos = 0
      atomic_halos%obj%max_nhalos = 0
   END SUBROUTINE fb_atomic_halo_list_init

! **************************************************************************************************
!> \brief Gets attributes from an fb_atomic_halo_list object, one should
!>        only access the data content in a fb_atomic_halo_list outside
!>        this module via this procedure.
!> \param atomic_halos the fb_atomic_halo object, its content must
!>                      not be NULL or UNDEFINED
!> \param nhalos [OPTIONAL]: if present, gives nhalos = atomic_halos%obj%nhalos
!> \param max_nhalos [OPTIONAL]: if present, gives max_nhalos = atomic_halos%obj%max_nhalos
!> \param halos [OPTIONAL]: if present, gives halos => atomic_halos%obj%halos
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_list_get(atomic_halos, nhalos, max_nhalos, halos)
      TYPE(fb_atomic_halo_list_obj), INTENT(IN)          :: atomic_halos
      INTEGER, INTENT(OUT), OPTIONAL                     :: nhalos, max_nhalos
      TYPE(fb_atomic_halo_obj), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: halos

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_list_get', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(atomic_halos%obj)))CALL cp__a("qs_fb_atomic_halo_types.F",758)
      IF (PRESENT(nhalos)) nhalos = atomic_halos%obj%nhalos
      IF (PRESENT(max_nhalos)) max_nhalos = atomic_halos%obj%max_nhalos
      IF (PRESENT(halos)) halos => atomic_halos%obj%halos
   END SUBROUTINE fb_atomic_halo_list_get

! **************************************************************************************************
!> \brief Sets attributes from an fb_atomic_halo_list object, one should
!>        only set the data content in a fb_atomic_halo_list outside
!>        this module via this procedure.
!> \param atomic_halos the fb_atomic_halo object, its content must
!>                      not be NULL or UNDEFINED
!> \param nhalos [OPTIONAL]: if present, sets atomic_halos%obj%nhalos = nhalos
!> \param max_nhalos [OPTIONAL]: if present, sets atomic_halos%obj%max_nhalos = max_nhalos
!> \param halos [OPTIONAL]: if present, reallocates atomic_halos%obj%halos
!>                          to the size of halos
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_list_set(atomic_halos, nhalos, max_nhalos, halos)
      TYPE(fb_atomic_halo_list_obj), INTENT(INOUT)       :: atomic_halos
      INTEGER, INTENT(IN), OPTIONAL                      :: nhalos, max_nhalos
      TYPE(fb_atomic_halo_obj), DIMENSION(:), OPTIONAL, &
         POINTER                                         :: halos

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_list_set', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ihalo

      IF(.NOT.(ASSOCIATED(atomic_halos%obj)))CALL cp__a("qs_fb_atomic_halo_types.F",787)
      IF (PRESENT(nhalos)) atomic_halos%obj%nhalos = nhalos
      IF (PRESENT(max_nhalos)) atomic_halos%obj%max_nhalos = max_nhalos
      IF (PRESENT(halos)) THEN
         IF (ASSOCIATED(atomic_halos%obj%halos)) THEN
            DO ihalo = 1, SIZE(atomic_halos%obj%halos)
               CALL fb_atomic_halo_release(atomic_halos%obj%halos(ihalo))
            END DO
            DEALLOCATE (atomic_halos%obj%halos)
         END IF
         atomic_halos%obj%halos => halos
      END IF
   END SUBROUTINE fb_atomic_halo_list_set

! **************************************************************************************************
!> \brief Writes out the atomic halo list from an fb_atomic_halo_list
!>        object using information
!> \param atomic_halos the fb_atomic_halo object
!> \param para_env pointer to a para_env_type object containing MPI info
!> \param fb_section pointer to the input section to filtered basis method
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_list_write(atomic_halos, para_env, fb_section)
      TYPE(fb_atomic_halo_list_obj), INTENT(IN)          :: atomic_halos
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: fb_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_list_write', &
         routineP = moduleN//':'//routineN

      CHARACTER(LEN=default_string_length)               :: string
      INTEGER                                            :: ihalo, jatom, mype, nhalo_atoms, nhalos, &
                                                            owner_atom, print_unit
      INTEGER, DIMENSION(:), POINTER                     :: halo_atoms
      LOGICAL                                            :: new_file
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(fb_atomic_halo_obj), DIMENSION(:), POINTER    :: halos

      NULLIFY (logger)
      logger => cp_get_default_logger()

      IF (BTEST(cp_print_key_should_output(logger%iter_info, fb_section, &
                                           "PRINT%ATOMIC_HALOS"), &
                cp_p_file)) THEN
         print_unit = cp_print_key_unit_nr(logger=logger, &
                                           basis_section=fb_section, &
                                           print_key_path="PRINT%ATOMIC_HALOS", &
                                           extension=".out", &
                                           local=.TRUE., &
                                           log_filename=.FALSE., &
                                           file_position="REWIND", &
                                           file_action="WRITE", &
                                           is_new_file=new_file)
         mype = para_env%mepos
         ! print headline
         string = ""
         WRITE (UNIT=string, FMT="(A,I5,A)") &
            "ATOMIC HALOS IN (PROCESS ", mype, ")"
         CALL compress(string)
         IF (print_unit > 0) THEN
            WRITE (UNIT=print_unit, FMT="(/,/,T2,A)") TRIM(string)
            WRITE (UNIT=print_unit, FMT="(/,T2,A)") &
               "atom : list of atoms in the atomic halo"
         END IF
         ! print content
         CALL fb_atomic_halo_list_get(atomic_halos=atomic_halos, &
                                      nhalos=nhalos, &
                                      halos=halos)
         DO ihalo = 1, nhalos
            CALL fb_atomic_halo_get(halos(ihalo), &
                                    owner_atom=owner_atom, &
                                    natoms=nhalo_atoms, &
                                    halo_atoms=halo_atoms)
            WRITE (UNIT=print_unit, FMT="(2X,I6,A)", ADVANCE="no") &
               owner_atom, " : "
            DO jatom = 1, nhalo_atoms
               WRITE (UNIT=print_unit, FMT="(I6)", ADVANCE="no") &
                  halo_atoms(jatom)
            END DO
            WRITE (UNIT=print_unit) ""
         END DO
         ! finish
         CALL cp_print_key_finished_output(print_unit, logger, fb_section, &
                                           "PRINT%ATOMIC_HALOS")
      END IF
   END SUBROUTINE fb_atomic_halo_list_write

! **************************************************************************************************
!> \brief Writes out the atomic halo list summary, no detailed neighbour lists,
!>        just average, min and max number of halo atoms in the halo list
!> \param atomic_halos : the fb_atomic_halo object
!> \param para_env     : pointer to a para_env_type object containing MPI info
!> \param scf_section  : pointer to the scf input section
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_atomic_halo_list_write_info(atomic_halos, para_env, scf_section)
      TYPE(fb_atomic_halo_list_obj), INTENT(IN)          :: atomic_halos
      TYPE(cp_para_env_type), POINTER                    :: para_env
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_atomic_halo_list_write_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ihalo, max_natoms, min_natoms, &
                                                            nhalo_atoms, nhalos, &
                                                            total_n_halo_atoms, total_n_halos, &
                                                            unit_nr
      REAL(KIND=dp)                                      :: ave_natoms
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(fb_atomic_halo_obj), DIMENSION(:), POINTER    :: halos

      NULLIFY (logger, halos)
      logger => cp_get_default_logger()
      unit_nr = cp_print_key_unit_nr(logger, scf_section, &
                                     "PRINT%FILTER_MATRIX", &
                                     extension="")

      ! obtain data
      CALL fb_atomic_halo_list_get(atomic_halos=atomic_halos, &
                                   halos=halos, &
                                   nhalos=nhalos)
      max_natoms = 0
      min_natoms = HUGE(0)
      total_n_halo_atoms = 0
      total_n_halos = nhalos
      DO ihalo = 1, nhalos
         CALL fb_atomic_halo_get(atomic_halo=halos(ihalo), &
                                 natoms=nhalo_atoms)
         total_n_halo_atoms = total_n_halo_atoms+nhalo_atoms
         max_natoms = MAX(max_natoms, nhalo_atoms)
         min_natoms = MIN(min_natoms, nhalo_atoms)
      END DO
      CALL mp_max(max_natoms, para_env%group)
      CALL mp_min(min_natoms, para_env%group)
      CALL mp_sum(total_n_halos, para_env%group)
      CALL mp_sum(total_n_halo_atoms, para_env%group)
      ave_natoms = REAL(total_n_halo_atoms, dp)/REAL(total_n_halos, dp)
      ! write info
      IF (unit_nr > 0) THEN
         WRITE (UNIT=unit_nr, FMT="(/,A)") &
            " FILTER_MAT_DIAG| Atomic matrix neighbor lists information:"
         WRITE (UNIT=unit_nr, FMT="(A,I10)") &
            " FILTER_MAT_DIAG|   Number of atomic matrices: ", &
            total_n_halos
         WRITE (UNIT=unit_nr, &
                FMT="(A,T45,A,T57,A,T69,A,T81,A)") &
            " FILTER_MAT_DIAG| ", "Average", "Max", "Min"
         WRITE (UNIT=unit_nr, &
                FMT="(A,T45,F10.1,T57,I10,T69,I10,T81,I10)") &
            " FILTER_MAT_DIAG|   N neighbors per atom:", &
            ave_natoms, max_natoms, min_natoms
      END IF
      ! finish
      CALL cp_print_key_finished_output(unit_nr, logger, scf_section, &
                                        "PRINT%FILTER_MATRIX")
   END SUBROUTINE fb_atomic_halo_list_write_info

! **************************************************************************************************
!> \brief Builds the required pair_radii array required for building the
!>        halo atoms from a given set of cut off radii
!> \param rcut   : rcut(ikind) is the cutoff radii for determining the halo
!>                 corresponding to atomic kind ikind
!> \param nkinds : total number of atomic kinds in rcut
!> \param pair_radii : output array pair_radii(ikind,jkind) gives the
!>                     corresponding interaction range between a pair of atoms
!>                     of kinds ikind and jkind
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   PURE SUBROUTINE fb_build_pair_radii(rcut, nkinds, pair_radii)
      REAL(KIND=dp), DIMENSION(:), INTENT(IN)            :: rcut
      INTEGER, INTENT(IN)                                :: nkinds
      REAL(KIND=dp), DIMENSION(:, :), INTENT(OUT)        :: pair_radii

      INTEGER                                            :: ii, jj

      pair_radii = 0.0_dp
      DO ii = 1, nkinds
         DO jj = 1, nkinds
            pair_radii(ii, jj) = rcut(ii)+rcut(jj)
         END DO
      END DO
   END SUBROUTINE fb_build_pair_radii

END MODULE qs_fb_atomic_halo_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_fb_trial_fns_types

! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler
   IMPLICIT NONE

   PRIVATE

! public types
   PUBLIC :: fb_trial_fns_obj

! public methods
   PUBLIC :: fb_trial_fns_retain, &
             fb_trial_fns_release, &
             fb_trial_fns_nullify, &
             fb_trial_fns_associate, &
             fb_trial_fns_has_data, &
             fb_trial_fns_create, &
             fb_trial_fns_get, &
             fb_trial_fns_set

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_trial_fns_types'
   INTEGER, PRIVATE, SAVE :: last_fb_trial_fns_id = 0

! **************************************************************************************************
!> \brief data containing information on trial functions used by filter
!>        matrix diagonalisation method
!> \param nfunctions : nfunctions(ikind) = number of trial functions for
!>                     atomic kind ikind
!> \param functions  : functions(itrial,ikind) = the index of the
!>                     GTO atomic orbital corresponding to itrial-th trial
!>                     function for kind ikind
!> \param id_nr      : unique id for the object
!> \param ref_count  : reference counter for the object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_trial_fns_data
      INTEGER :: id_nr, ref_count
      INTEGER, DIMENSION(:), POINTER :: nfunctions
      INTEGER, DIMENSION(:, :), POINTER :: functions
   END TYPE fb_trial_fns_data

! **************************************************************************************************
!> \brief the object container which allows for the creation of an array
!>        of pointers to fb_trial_fns objects
!> \param obj : pointer to the fb_trial_fns object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_trial_fns_obj
      TYPE(fb_trial_fns_data), POINTER, PRIVATE :: obj
   END TYPE fb_trial_fns_obj

CONTAINS

! **************************************************************************************************
!> \brief retains given object
!> \brief ...
!> \param trial_fns : the fb_trial_fns object in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_trial_fns_retain(trial_fns)
      ! note INTENT(IN) is okay because the obj pointer contained in the
      ! obj type will not be changed
      TYPE(fb_trial_fns_obj), INTENT(IN)                 :: trial_fns

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_trial_fns_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(trial_fns%obj)))CALL cp__a("qs_fb_trial_fns_types.F",73)
      IF(.NOT.(trial_fns%obj%ref_count > 0))CALL cp__a("qs_fb_trial_fns_types.F",74)
      trial_fns%obj%ref_count = trial_fns%obj%ref_count+1
   END SUBROUTINE fb_trial_fns_retain

! **************************************************************************************************
!> \brief releases given object
!> \brief ...
!> \param trial_fns : the fb_trial_fns object in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_trial_fns_release(trial_fns)
      TYPE(fb_trial_fns_obj), INTENT(INOUT)              :: trial_fns

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_trial_fns_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(trial_fns%obj)) THEN
         IF(.NOT.(trial_fns%obj%ref_count > 0))CALL cp__a("qs_fb_trial_fns_types.F",91)
         trial_fns%obj%ref_count = trial_fns%obj%ref_count-1
         IF (trial_fns%obj%ref_count == 0) THEN
            trial_fns%obj%ref_count = 1
            IF (ASSOCIATED(trial_fns%obj%nfunctions)) THEN
               DEALLOCATE (trial_fns%obj%nfunctions)
            END IF
            IF (ASSOCIATED(trial_fns%obj%functions)) THEN
               DEALLOCATE (trial_fns%obj%functions)
            END IF
            trial_fns%obj%ref_count = 0
            DEALLOCATE (trial_fns%obj)
         END IF
      ELSE
         NULLIFY (trial_fns%obj)
      END IF
   END SUBROUTINE fb_trial_fns_release

! **************************************************************************************************
!> \brief nullifies the content of given object
!> \param trial_fns : the fb_trial_fns object in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_trial_fns_nullify(trial_fns)
      TYPE(fb_trial_fns_obj), INTENT(INOUT)              :: trial_fns

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_trial_fns_nullify', &
         routineP = moduleN//':'//routineN

      NULLIFY (trial_fns%obj)
   END SUBROUTINE fb_trial_fns_nullify

! **************************************************************************************************
!> \brief associates the content of an object to that of another object
!>        of the same type
!> \param a : the output object
!> \param b : the input object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_trial_fns_associate(a, b)
      TYPE(fb_trial_fns_obj), INTENT(OUT)                :: a
      TYPE(fb_trial_fns_obj), INTENT(IN)                 :: b

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_trial_fns_associate', &
         routineP = moduleN//':'//routineN

      a%obj => b%obj
   END SUBROUTINE fb_trial_fns_associate

! **************************************************************************************************
!> \brief check if the object has data associated to it
!> \param trial_fns : the fb_trial_fns object in question
!> \retval res : true if trial_fns%obj is associated, false otherwise
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   FUNCTION fb_trial_fns_has_data(trial_fns) RESULT(res)
      TYPE(fb_trial_fns_obj), INTENT(IN)                 :: trial_fns
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_trial_fns_has_data', &
         routineP = moduleN//':'//routineN

      res = ASSOCIATED(trial_fns%obj)
   END FUNCTION fb_trial_fns_has_data

! **************************************************************************************************
!> \brief creates an fb_trial_fns object and initialises it
!> \param trial_fns : the fb_trial_fns object in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_trial_fns_create(trial_fns)
      TYPE(fb_trial_fns_obj), INTENT(INOUT)              :: trial_fns

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_trial_fns_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(trial_fns%obj)))CALL cp__a("qs_fb_trial_fns_types.F",167)
      ALLOCATE (trial_fns%obj)
      NULLIFY (trial_fns%obj%nfunctions)
      NULLIFY (trial_fns%obj%functions)
      trial_fns%obj%ref_count = 1
      trial_fns%obj%id_nr = last_fb_trial_fns_id+1
      last_fb_trial_fns_id = trial_fns%obj%id_nr
   END SUBROUTINE fb_trial_fns_create

! **************************************************************************************************
!> \brief initialises an fb_trial_fns object
!> \param trial_fns : the fb_trial_fns object in question
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_trial_fns_init(trial_fns)
      TYPE(fb_trial_fns_obj), INTENT(INOUT)              :: trial_fns

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_trial_fns_init', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(trial_fns%obj)))CALL cp__a("qs_fb_trial_fns_types.F",187)
      ! if halo_atoms are associated, then deallocate and de-associate
      IF (ASSOCIATED(trial_fns%obj%nfunctions)) THEN
         DEALLOCATE (trial_fns%obj%nfunctions)
      END IF
      IF (ASSOCIATED(trial_fns%obj%functions)) THEN
         DEALLOCATE (trial_fns%obj%functions)
      END IF
   END SUBROUTINE fb_trial_fns_init

! **************************************************************************************************
!> \brief get values of the attributes of a fb_trial_fns object
!> \param trial_fns  : the fb_trial_fns object in question
!> \param nfunctions : outputs pointer to trial_fns%obj%nfunctions
!> \param functions  : outputs pointer to trial_fns%obj%functions
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_trial_fns_get(trial_fns, &
                               nfunctions, &
                               functions)
      TYPE(fb_trial_fns_obj), INTENT(IN)                 :: trial_fns
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: nfunctions
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: functions

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_trial_fns_get', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(trial_fns%obj)))CALL cp__a("qs_fb_trial_fns_types.F",214)
      IF (PRESENT(nfunctions)) nfunctions => trial_fns%obj%nfunctions
      IF (PRESENT(functions)) functions => trial_fns%obj%functions
   END SUBROUTINE fb_trial_fns_get

! **************************************************************************************************
!> \brief sets the attributes of a fb_trial_fns object
!> \param trial_fns  : the fb_trial_fns object in question
!> \param nfunctions : associates trial_fns%obj%nfunctions to this pointer
!> \param functions  : associates trial_fns%obj%nfunctions to this pointer
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_trial_fns_set(trial_fns, &
                               nfunctions, &
                               functions)
      TYPE(fb_trial_fns_obj), INTENT(INOUT)              :: trial_fns
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: nfunctions
      INTEGER, DIMENSION(:, :), OPTIONAL, POINTER        :: functions

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_trial_fns_set', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(trial_fns%obj)))CALL cp__a("qs_fb_trial_fns_types.F",236)
      IF (PRESENT(nfunctions)) THEN
         IF (ASSOCIATED(trial_fns%obj%nfunctions)) THEN
            DEALLOCATE (trial_fns%obj%nfunctions)
         END IF
         trial_fns%obj%nfunctions => nfunctions
      END IF
      IF (PRESENT(functions)) THEN
         IF (ASSOCIATED(trial_fns%obj%functions)) THEN
            DEALLOCATE (trial_fns%obj%functions)
         END IF
         trial_fns%obj%functions => functions
      END IF
   END SUBROUTINE fb_trial_fns_set

END MODULE qs_fb_trial_fns_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

MODULE qs_fb_env_types

   USE kinds,                           ONLY: dp
   USE qs_fb_atomic_halo_types,         ONLY: fb_atomic_halo_list_associate,&
                                              fb_atomic_halo_list_has_data,&
                                              fb_atomic_halo_list_nullify,&
                                              fb_atomic_halo_list_obj,&
                                              fb_atomic_halo_list_release,&
                                              fb_atomic_halo_list_retain
   USE qs_fb_trial_fns_types,           ONLY: fb_trial_fns_associate,&
                                              fb_trial_fns_has_data,&
                                              fb_trial_fns_nullify,&
                                              fb_trial_fns_obj,&
                                              fb_trial_fns_release,&
                                              fb_trial_fns_retain
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! public types
   PUBLIC :: fb_env_obj

! public methods
   PUBLIC :: fb_env_release, &
             fb_env_nullify, &
             fb_env_has_data, &
             fb_env_create, &
             fb_env_get, &
             fb_env_set

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_fb_env_types'
   INTEGER, PRIVATE, SAVE :: last_fb_env_id = 0

! **********************************************************************
!> \brief wrapper to the simulation parameters used for filtered basis
!>        method
!> \param rcut   : cutoff for included filtered basis set centred at
!>                 each atom. These defines the ranges of the atomic
!>                 halos. rcut(ikind) gives the range for atom of
!>                 global kind ikind
!> \param atomic_halos  : stores information on the neighbors of each
!>                        atom ii, which are defined by rcut
!> \param filter_temperature : parameter controlling the smoothness of
!>                             the filter function during the construction
!>                             of the filter matrix
!> \param auto_cutoff_scale  : scale multiplied to max atomic orbital
!>                             radii used for automatic construction of
!>                             rcut
!> \param eps_default        : anything less than it is regarded as zero
!> \param collective_com     : whether the MPI communications are
!>                             to be done collectively together
!>                             at the start and end of each
!>                             filter matrix calculation. This makes
!>                             communication more efficient in the
!>                             expense of larger memory usage
!> \param local_atoms        : atoms corresponding to the
!>                             atomic halos responsible by this processor
!> \param id_nr : unique id of this object
!> \param ref_count : reference counter of this object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **********************************************************************
   TYPE fb_env_data
      INTEGER :: id_nr, ref_count
      REAL(KIND=dp), DIMENSION(:), POINTER :: rcut
      TYPE(fb_atomic_halo_list_obj) :: atomic_halos
      TYPE(fb_trial_fns_obj) :: trial_fns
      REAL(KIND=dp) :: filter_temperature
      REAL(KIND=dp) :: auto_cutoff_scale
      REAL(KIND=dp) :: eps_default
      LOGICAL :: collective_com
      INTEGER, DIMENSION(:), POINTER :: local_atoms
      INTEGER :: nlocal_atoms
   END TYPE fb_env_data

! **************************************************************************************************
!> \brief the object container which allows for the creation of an array of
!>        pointers to fb_env
!> \param obj : pointer to a filtered basis environment
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   TYPE fb_env_obj
      TYPE(fb_env_data), POINTER, PRIVATE :: obj
   END TYPE fb_env_obj

CONTAINS

! **********************************************************************
!> \brief retains the given fb_env
!> \param fb_env : the fb_env to retain
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_retain(fb_env)
      TYPE(fb_env_obj), INTENT(IN)                       :: fb_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(fb_env%obj)))CALL cp__a("qs_fb_env_types.F",104)
      IF(.NOT.(fb_env%obj%ref_count > 0))CALL cp__a("qs_fb_env_types.F",105)
      fb_env%obj%ref_count = fb_env%obj%ref_count+1
   END SUBROUTINE fb_env_retain

! **********************************************************************
!> \brief releases a given fb_env
!> \brief ...
!> \param fb_env : the fb_env to release
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_release(fb_env)
      TYPE(fb_env_obj), INTENT(INOUT)                    :: fb_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(fb_env%obj)) THEN
         IF(.NOT.(fb_env%obj%ref_count > 0))CALL cp__a("qs_fb_env_types.F",121)
         fb_env%obj%ref_count = fb_env%obj%ref_count-1
         IF (fb_env%obj%ref_count == 0) THEN
            fb_env%obj%ref_count = 1
            IF (ASSOCIATED(fb_env%obj%rcut)) THEN
               DEALLOCATE (fb_env%obj%rcut)
            END IF
            IF (ASSOCIATED(fb_env%obj%local_atoms)) THEN
               DEALLOCATE (fb_env%obj%local_atoms)
            END IF
            CALL fb_atomic_halo_list_release(fb_env%obj%atomic_halos)
            CALL fb_trial_fns_release(fb_env%obj%trial_fns)
            fb_env%obj%ref_count = 0
            DEALLOCATE (fb_env%obj)
         END IF
      ELSE
         NULLIFY (fb_env%obj)
      END IF
   END SUBROUTINE fb_env_release

! **********************************************************************
!> \brief nullifies a fb_env object, note that this does not
!>        release the original object. This procedure is used mainly
!>        to nullify the pointer inside the object which is used to
!>        point to the actual data content of the object.
!> \param fb_env : its content must be a NULL fb_env pointer on input,
!>                 and the output returns an empty fb_env object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_nullify(fb_env)
      TYPE(fb_env_obj), INTENT(INOUT)                    :: fb_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_nullify', routineP = moduleN//':'//routineN

      NULLIFY (fb_env%obj)
   END SUBROUTINE fb_env_nullify

! **********************************************************************
!> \brief Associates one fb_env object to another
!> \param a the fb_env object to be associated
!> \param b the fb_env object that a is to be associated to
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_associate(a, b)
      TYPE(fb_env_obj), INTENT(OUT)                      :: a
      TYPE(fb_env_obj), INTENT(IN)                       :: b

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_associate', &
         routineP = moduleN//':'//routineN

      a%obj => b%obj
   END SUBROUTINE fb_env_associate

! **********************************************************************
!> \brief Checks if a fb_env object is associated with an actual
!>        data content or not
!> \param fb_env the fb_env object
!> \retval res ...
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   FUNCTION fb_env_has_data(fb_env) RESULT(res)
      TYPE(fb_env_obj), INTENT(IN)                       :: fb_env
      LOGICAL                                            :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_has_data', &
         routineP = moduleN//':'//routineN

      res = ASSOCIATED(fb_env%obj)
   END FUNCTION fb_env_has_data

! **********************************************************************
!> \brief creates an empty fb_env object
!> \param fb_env : its content must be a NULL fb_env pointer on input,
!>                 and the output returns an empty fb_env object
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_create(fb_env)
      TYPE(fb_env_obj), INTENT(INOUT)                    :: fb_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_create', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(fb_env%obj)))CALL cp__a("qs_fb_env_types.F",202)
      ALLOCATE (fb_env%obj)
      NULLIFY (fb_env%obj%rcut)
      CALL fb_atomic_halo_list_nullify(fb_env%obj%atomic_halos)
      CALL fb_trial_fns_nullify(fb_env%obj%trial_fns)
      fb_env%obj%filter_temperature = 0.0_dp
      fb_env%obj%auto_cutoff_scale = 1.0_dp
      fb_env%obj%eps_default = 0.0_dp
      fb_env%obj%collective_com = .TRUE.
      NULLIFY (fb_env%obj%local_atoms)
      fb_env%obj%nlocal_atoms = 0
      fb_env%obj%ref_count = 1
      fb_env%obj%id_nr = last_fb_env_id+1
      last_fb_env_id = fb_env%obj%id_nr
   END SUBROUTINE fb_env_create

! **********************************************************************
!> \brief initialises a fb_env object to become empty
!> \brief ...
!> \param fb_env : the fb_env object, which must not be NULL or
!>                 UNDEFINED upon entry
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_init(fb_env)
      TYPE(fb_env_obj), INTENT(INOUT)                    :: fb_env

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_init', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(fb_env%obj)))CALL cp__a("qs_fb_env_types.F",230)
      IF (ASSOCIATED(fb_env%obj%rcut)) THEN
         DEALLOCATE (fb_env%obj%rcut)
      END IF
      CALL fb_atomic_halo_list_release(fb_env%obj%atomic_halos)
      CALL fb_trial_fns_release(fb_env%obj%trial_fns)
      fb_env%obj%filter_temperature = 0.0_dp
      fb_env%obj%auto_cutoff_scale = 1.0_dp
      fb_env%obj%eps_default = 0.0_dp
      fb_env%obj%collective_com = .TRUE.
      IF (ASSOCIATED(fb_env%obj%local_atoms)) THEN
         DEALLOCATE (fb_env%obj%local_atoms)
      END IF
      fb_env%obj%nlocal_atoms = 0
   END SUBROUTINE fb_env_init

! **********************************************************************
!> \brief method to get attributes from a given fb_env object
!> \brief ...
!> \param fb_env : the fb_env object in question
!> \param rcut   : outputs pointer to rcut attribute of fb_env (optional)
!> \param filter_temperature : outputs filter_temperature attribute
!>                             of fb_env (optional)
!> \param auto_cutoff_scale  : outputs auto_cutoff_scale attribute
!>                             of fb_env (optional)
!> \param eps_default        : outputs eps_default attribute
!>                             of fb_env (optional)
!> \param atomic_halos       : outputs pointer to atomic_halos
!>                             attribute of fb_env (optional)
!> \param trial_fns          : outputs pointer to trial_fns
!>                             attribute of fb_env (optional)
!> \param collective_com     : outputs pointer to trial_fns
!> \param local_atoms        : outputs pointer to local_atoms
!> \param nlocal_atoms       : outputs pointer to nlocal_atoms
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_get(fb_env, &
                         rcut, &
                         filter_temperature, &
                         auto_cutoff_scale, &
                         eps_default, &
                         atomic_halos, &
                         trial_fns, &
                         collective_com, &
                         local_atoms, &
                         nlocal_atoms)
      TYPE(fb_env_obj), INTENT(IN)                       :: fb_env
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: rcut
      REAL(KIND=dp), INTENT(OUT), OPTIONAL               :: filter_temperature, auto_cutoff_scale, &
                                                            eps_default
      TYPE(fb_atomic_halo_list_obj), INTENT(OUT), &
         OPTIONAL                                        :: atomic_halos
      TYPE(fb_trial_fns_obj), INTENT(OUT), OPTIONAL      :: trial_fns
      LOGICAL, INTENT(OUT), OPTIONAL                     :: collective_com
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: local_atoms
      INTEGER, INTENT(OUT), OPTIONAL                     :: nlocal_atoms

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_get', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(fb_env%obj)))CALL cp__a("qs_fb_env_types.F",289)
      IF(.NOT.(fb_env%obj%ref_count > 0))CALL cp__a("qs_fb_env_types.F",290)
      IF (PRESENT(rcut)) &
         rcut => fb_env%obj%rcut
      IF (PRESENT(filter_temperature)) &
         filter_temperature = fb_env%obj%filter_temperature
      IF (PRESENT(auto_cutoff_scale)) &
         auto_cutoff_scale = fb_env%obj%auto_cutoff_scale
      IF (PRESENT(eps_default)) &
         eps_default = fb_env%obj%eps_default
      IF (PRESENT(atomic_halos)) &
         CALL fb_atomic_halo_list_associate(atomic_halos, fb_env%obj%atomic_halos)
      IF (PRESENT(trial_fns)) &
         CALL fb_trial_fns_associate(trial_fns, fb_env%obj%trial_fns)
      IF (PRESENT(collective_com)) &
         collective_com = fb_env%obj%collective_com
      IF (PRESENT(local_atoms)) &
         local_atoms => fb_env%obj%local_atoms
      IF (PRESENT(nlocal_atoms)) &
         nlocal_atoms = fb_env%obj%nlocal_atoms
   END SUBROUTINE fb_env_get

! **********************************************************************
!> \brief method to set attributes from a given fb_env object
!> \brief ...
!> \param fb_env : the fb_env object in question
!> \param rcut   : sets rcut attribute of fb_env (optional)
!> \param filter_temperature : sets filter_temperature attribute of fb_env (optional)
!> \param auto_cutoff_scale  : sets auto_cutoff_scale attribute of fb_env (optional)
!> \param eps_default        : sets eps_default attribute of fb_env (optional)
!> \param atomic_halos       : sets atomic_halos attribute of fb_env (optional)
!> \param trial_fns          : sets trial_fns attribute of fb_env (optional)
!> \param collective_com     : sets collective_com attribute of fb_env (optional)
!> \param local_atoms        : sets local_atoms attribute of fb_env (optional)
!> \param nlocal_atoms       : sets nlocal_atoms attribute of fb_env (optional)
!> \author Lianheng Tong (LT) lianheng.tong@kcl.ac.uk
! **************************************************************************************************
   SUBROUTINE fb_env_set(fb_env, &
                         rcut, &
                         filter_temperature, &
                         auto_cutoff_scale, &
                         eps_default, &
                         atomic_halos, &
                         trial_fns, &
                         collective_com, &
                         local_atoms, &
                         nlocal_atoms)
      TYPE(fb_env_obj), INTENT(INOUT)                    :: fb_env
      REAL(KIND=dp), DIMENSION(:), OPTIONAL, POINTER     :: rcut
      REAL(KIND=dp), INTENT(IN), OPTIONAL                :: filter_temperature, auto_cutoff_scale, &
                                                            eps_default
      TYPE(fb_atomic_halo_list_obj), INTENT(IN), &
         OPTIONAL                                        :: atomic_halos
      TYPE(fb_trial_fns_obj), INTENT(IN), OPTIONAL       :: trial_fns
      LOGICAL, INTENT(IN), OPTIONAL                      :: collective_com
      INTEGER, DIMENSION(:), OPTIONAL, POINTER           :: local_atoms
      INTEGER, INTENT(IN), OPTIONAL                      :: nlocal_atoms

      CHARACTER(len=*), PARAMETER :: routineN = 'fb_env_set', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(fb_env%obj)))CALL cp__a("qs_fb_env_types.F",349)
      IF (PRESENT(rcut)) THEN
         IF (ASSOCIATED(fb_env%obj%rcut)) THEN
            DEALLOCATE (fb_env%obj%rcut)
         END IF
         fb_env%obj%rcut => rcut
      END IF
      IF (PRESENT(filter_temperature)) &
         fb_env%obj%filter_temperature = filter_temperature
      IF (PRESENT(auto_cutoff_scale)) &
         fb_env%obj%auto_cutoff_scale = auto_cutoff_scale
      IF (PRESENT(eps_default)) &
         fb_env%obj%eps_default = eps_default
      IF (PRESENT(atomic_halos)) THEN
         IF (fb_atomic_halo_list_has_data(atomic_halos)) &
            CALL fb_atomic_halo_list_retain(atomic_halos)
         CALL fb_atomic_halo_list_release(fb_env%obj%atomic_halos)
         CALL fb_atomic_halo_list_associate(fb_env%obj%atomic_halos, atomic_halos)
      END IF
      IF (PRESENT(trial_fns)) THEN
         IF (fb_trial_fns_has_data(trial_fns)) &
            CALL fb_trial_fns_retain(trial_fns)
         CALL fb_trial_fns_release(fb_env%obj%trial_fns)
         CALL fb_trial_fns_associate(fb_env%obj%trial_fns, trial_fns)
      END IF
      IF (PRESENT(collective_com)) &
         fb_env%obj%collective_com = collective_com
      IF (PRESENT(local_atoms)) THEN
         IF (ASSOCIATED(fb_env%obj%local_atoms)) THEN
            DEALLOCATE (fb_env%obj%local_atoms)
         END IF
         fb_env%obj%local_atoms => local_atoms
      END IF
      IF (PRESENT(nlocal_atoms)) &
         fb_env%obj%nlocal_atoms = nlocal_atoms
   END SUBROUTINE fb_env_set

END MODULE qs_fb_env_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief buffer for the diis of the scf
!> \par History
!>      02.2003 rewamped [fawzi]
!> \author Matthias Krack
! **************************************************************************************************
MODULE qs_diis_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_p_type,&
                                              cp_dbcsr_release
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_diis_types'

   PUBLIC :: qs_diis_buffer_type
   PUBLIC :: qs_diis_b_release
   PUBLIC :: qs_diis_buffer_type_sparse, &
             qs_diis_b_release_sparse

! **************************************************************************************************
!> \brief keeps a buffer with the previous values of s,p,k
!> \par History
!>      02.2003 rewamped [fawzi]
!> \author Matthias Krack
! **************************************************************************************************
   TYPE qs_diis_buffer_type
      INTEGER                                          :: nbuffer, ncall, &
                                                          id_nr, ref_count
      TYPE(cp_fm_p_type), DIMENSION(:, :), POINTER :: error, PARAMETER
      REAL(KIND=dp), DIMENSION(:, :), POINTER                :: b_matrix
   END TYPE qs_diis_buffer_type

! **************************************************************************************************
!> \brief build array of pointers to diis buffers
!> \param diis_buffer the diis buffer pointer
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE qs_diis_buffer_p_type
      TYPE(qs_diis_buffer_type), POINTER :: diis_buffer
   END TYPE qs_diis_buffer_p_type

! **************************************************************************************************
!> \brief build array of pointers to diis buffers for sparse matrix case
!> \param diis_buffer the diis buffer pointer
!> \par History
!>      10.2014 Modified from non-sparse case by Fredy W. Aquino
!> \author fwaq
! **************************************************************************************************
   TYPE qs_diis_buffer_type_sparse
      INTEGER                                          :: nbuffer, ncall, &
                                                          id_nr, ref_count
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER   :: error, PARAMETER
      REAL(KIND=dp), DIMENSION(:, :), POINTER         :: b_matrix
   END TYPE qs_diis_buffer_type_sparse

   TYPE qs_diis_buffer_p_type_sparse
      TYPE(qs_diis_buffer_type_sparse), POINTER       :: diis_buffer
   END TYPE qs_diis_buffer_p_type_sparse

CONTAINS

! **************************************************************************************************
!> \brief retains a diis buffer (see doc/ReferenceCounting.html)
!> \param diis_buffer the buffer to retain
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE qs_diis_b_retain(diis_buffer)
      TYPE(qs_diis_buffer_type), POINTER                 :: diis_buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_diis_b_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(diis_buffer)))CALL cp__a("qs_diis_types.F",88)
      IF(.NOT.(diis_buffer%ref_count > 0))CALL cp__a("qs_diis_types.F",89)
      diis_buffer%ref_count = diis_buffer%ref_count+1
   END SUBROUTINE qs_diis_b_retain

! **************************************************************************************************
!> \brief releases the given diis buffer (see doc/ReferenceCounting.html)
!> \param diis_buffer the buffer to release
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE qs_diis_b_release(diis_buffer)
      TYPE(qs_diis_buffer_type), POINTER                 :: diis_buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_diis_b_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      IF (ASSOCIATED(diis_buffer)) THEN
         IF(.NOT.(diis_buffer%ref_count > 0))CALL cp__a("qs_diis_types.F",109)
         diis_buffer%ref_count = diis_buffer%ref_count-1
         IF (diis_buffer%ref_count < 1) THEN
            IF (ASSOCIATED(diis_buffer%b_matrix)) THEN
               DEALLOCATE (diis_buffer%b_matrix)
            END IF
            IF (ASSOCIATED(diis_buffer%error)) THEN
               DO j = 1, SIZE(diis_buffer%error, 2)
                  DO i = 1, SIZE(diis_buffer%error, 1)
                     CALL cp_fm_release(diis_buffer%error(i, j)%matrix)
                  END DO
               END DO
               DEALLOCATE (diis_buffer%error)
            END IF
            IF (ASSOCIATED(diis_buffer%parameter)) THEN
               DO j = 1, SIZE(diis_buffer%parameter, 2)
                  DO i = 1, SIZE(diis_buffer%parameter, 1)
                     CALL cp_fm_release(diis_buffer%parameter(i, j)%matrix)
                  END DO
               END DO
               DEALLOCATE (diis_buffer%parameter)
            END IF
            DEALLOCATE (diis_buffer)
         END IF
      END IF
   END SUBROUTINE qs_diis_b_release

! **************************************************************************************************
!> \brief releases the given diis buffer (see doc/ReferenceCounting.html)
!> \param diis_buffer the buffer to release
!> \par History
!>      10-11-14 created [FA] modified from qs_diis_b_release
!> \author Fredy W. Aquino
! **************************************************************************************************
   SUBROUTINE qs_diis_b_release_sparse(diis_buffer)

      TYPE(qs_diis_buffer_type_sparse), POINTER          :: diis_buffer

      CHARACTER(len=*), PARAMETER :: routineN = 'qs_diis_b_release_sparse', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      IF (ASSOCIATED(diis_buffer)) THEN
         IF (ASSOCIATED(diis_buffer%b_matrix)) THEN
            DEALLOCATE (diis_buffer%b_matrix)
         END IF
         IF (ASSOCIATED(diis_buffer%error)) THEN
            DO j = 1, SIZE(diis_buffer%error, 2)
               DO i = 1, SIZE(diis_buffer%error, 1)
                  CALL cp_dbcsr_release(diis_buffer%error(i, j)%matrix)
                  DEALLOCATE (diis_buffer%error(i, j)%matrix)
               END DO
            END DO
            DEALLOCATE (diis_buffer%error)
         END IF
         IF (ASSOCIATED(diis_buffer%parameter)) THEN
            DO j = 1, SIZE(diis_buffer%parameter, 2)
               DO i = 1, SIZE(diis_buffer%parameter, 1)
                  CALL cp_dbcsr_release(diis_buffer%parameter(i, j)%matrix)
                  DEALLOCATE (diis_buffer%parameter(i, j)%matrix)
               END DO
            END DO
            DEALLOCATE (diis_buffer%parameter)
         END IF
         DEALLOCATE (diis_buffer)
      END IF
   END SUBROUTINE qs_diis_b_release_sparse

END MODULE qs_diis_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module that contains the algorithms to perform an itrative
!>         diagonalization by the block-Davidson approach
!>         P. Blaha, et al J. Comp. Physics, 229, (2010), 453-460
!>         \Iterative diagonalization in augmented plane wave based
!>              methods in electronic structure calculations!> \par History
!>      05.2011 created [MI]
!> \author MI
! **************************************************************************************************
MODULE qs_block_davidson_types

   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE input_section_types,             ONLY: section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_block_davidson_types'

   PUBLIC :: block_davidson_allocate, block_davidson_deallocate, block_davidson_release, &
             block_davidson_env_create, davidson_type

   TYPE davidson_type
      INTEGER :: max_iter, prec_type, solver_type, niter_new_prec, first_prec
      LOGICAL :: use_sparse_mos
      REAL(KIND=dp) :: conv_percent, energy_gap, eps_iter
      TYPE(cp_fm_type), POINTER :: H_block_mat, H_block_vec, &
                                   matrix_z, matrix_pz, S_block_mat, W_block_mat
   END TYPE davidson_type

CONTAINS

! **************************************************************************************************

! **************************************************************************************************
!> \brief ...
!> \param bdav_env ...
!> \param nspins ...
!> \param scf_section ...
! **************************************************************************************************
   SUBROUTINE block_davidson_env_create(bdav_env, nspins, scf_section)

      TYPE(davidson_type), DIMENSION(:), POINTER         :: bdav_env
      INTEGER, INTENT(IN)                                :: nspins
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(len=*), PARAMETER :: routineN = 'block_davidson_env_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin

      IF(.NOT.(.NOT. ASSOCIATED(bdav_env)))CALL cp__a("qs_block_davidson_types.F",65)
      ALLOCATE (bdav_env(nspins))
      DO ispin = 1, nspins
         NULLIFY (bdav_env(ispin)%H_block_mat)
         NULLIFY (bdav_env(ispin)%H_block_vec)
         NULLIFY (bdav_env(ispin)%S_block_mat)
         NULLIFY (bdav_env(ispin)%W_block_mat)
         NULLIFY (bdav_env(ispin)%matrix_z)
         NULLIFY (bdav_env(ispin)%matrix_pz)

         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DAVIDSON%PRECONDITIONER", &
                                   i_val=bdav_env(ispin)%prec_type)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DAVIDSON%PRECOND_SOLVER", &
                                   i_val=bdav_env(ispin)%solver_type)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DAVIDSON%ENERGY_GAP", &
                                   r_val=bdav_env(ispin)%energy_gap)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DAVIDSON%NEW_PREC_EACH", &
                                   i_val=bdav_env(ispin)%niter_new_prec)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%MAX_ITER", &
                                   i_val=bdav_env(ispin)%max_iter)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%EPS_ITER", &
                                   r_val=bdav_env(ispin)%eps_iter)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DAVIDSON%FIRST_PREC", &
                                   i_val=bdav_env(ispin)%first_prec)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DAVIDSON%CONV_MOS_PERCENT", &
                                   r_val=bdav_env(ispin)%conv_percent)
         CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DAVIDSON%SPARSE_MOS", &
                                   l_val=bdav_env(ispin)%use_sparse_mos)

      END DO

   END SUBROUTINE block_davidson_env_create

! **************************************************************************************************
!> \brief ...
!> \param bdav_env ...
!> \param mo_coeff ...
!> \param nao ...
!> \param nmo ...
! **************************************************************************************************
   SUBROUTINE block_davidson_allocate(bdav_env, mo_coeff, nao, nmo)

      TYPE(davidson_type)                                :: bdav_env
      TYPE(cp_fm_type), POINTER                          :: mo_coeff
      INTEGER, INTENT(IN)                                :: nao, nmo

      CHARACTER(len=*), PARAMETER :: routineN = 'block_davidson_allocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, nmox2
      TYPE(cp_fm_struct_type), POINTER                   :: fm_struct_tmp

      CALL timeset(routineN, handle)
      NULLIFY (fm_struct_tmp)

      nmox2 = 2*nmo
!    CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nmox2, ncol_global=nmox2,&
!                             para_env=mo_coeff%matrix_struct%para_env, &
!                             context=mo_coeff%matrix_struct%context)
!    CALL cp_fm_create(bdav_env%H_block_mat,fm_struct_tmp,name="H_dav")
!    CALL cp_fm_create(bdav_env%S_block_mat,fm_struct_tmp,name="S_dav")
!    CALL cp_fm_create(bdav_env%H_block_vec,fm_struct_tmp,name="C_dav")
!    CALL cp_fm_create(bdav_env%W_block_mat,fm_struct_tmp,name="W_dav")
!    CALL cp_fm_struct_release(fm_struct_tmp)

      CALL cp_fm_struct_create(fm_struct_tmp, nrow_global=nao, ncol_global=nmo, &
                               para_env=mo_coeff%matrix_struct%para_env, &
                               context=mo_coeff%matrix_struct%context)
      CALL cp_fm_create(bdav_env%matrix_z, fm_struct_tmp, name="Z_mat")
      CALL cp_fm_create(bdav_env%matrix_pz, fm_struct_tmp, name="Z_mat")
      CALL cp_fm_struct_release(fm_struct_tmp)

      CALL timestop(handle)

   END SUBROUTINE block_davidson_allocate

! **************************************************************************************************
!> \brief ...
!> \param bdav_env ...
! **************************************************************************************************
   SUBROUTINE block_davidson_deallocate(bdav_env)

      TYPE(davidson_type), DIMENSION(:), POINTER         :: bdav_env

      CHARACTER(len=*), PARAMETER :: routineN = 'block_davidson_deallocate', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, nspins

      IF (ASSOCIATED(bdav_env)) THEN

         nspins = SIZE(bdav_env)
         DO ispin = 1, nspins

            CALL cp_fm_release(bdav_env(ispin)%matrix_z)
            CALL cp_fm_release(bdav_env(ispin)%matrix_pz)
!        CALL cp_fm_release(bdav_env(ispin)%H_block_mat)
!        CALL cp_fm_release(bdav_env(ispin)%S_block_mat)
!        CALL cp_fm_release(bdav_env(ispin)%W_block_mat)
!        CALL cp_fm_release(bdav_env(ispin)%H_block_vec)

         END DO

      END IF

   END SUBROUTINE block_davidson_deallocate

! **************************************************************************************************
!> \brief ...
!> \param bdav_env ...
! **************************************************************************************************
   SUBROUTINE block_davidson_release(bdav_env)

      TYPE(davidson_type), DIMENSION(:), POINTER         :: bdav_env

      CHARACTER(len=*), PARAMETER :: routineN = 'block_davidson_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin, nspins

      IF (ASSOCIATED(bdav_env)) THEN

         nspins = SIZE(bdav_env)
         DO ispin = 1, nspins

            IF (ASSOCIATED(bdav_env(ispin)%matrix_z)) THEN
               CALL cp_fm_release(bdav_env(ispin)%matrix_z)
               CALL cp_fm_release(bdav_env(ispin)%matrix_pz)
            END IF
!        CALL cp_fm_release(bdav_env(ispin)%H_block_mat)
!        CALL cp_fm_release(bdav_env(ispin)%S_block_mat)
!        CALL cp_fm_release(bdav_env(ispin)%W_block_mat)
!        CALL cp_fm_release(bdav_env(ispin)%H_block_vec)

         END DO
         DEALLOCATE (bdav_env)

      END IF

   END SUBROUTINE block_davidson_release

END MODULE qs_block_davidson_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief module that contains the definitions of the scf types
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
MODULE qs_scf_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_fm_vect,                      ONLY: cp_fm_vect_dealloc
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE preconditioner_types,            ONLY: destroy_preconditioner,&
                                              preconditioner_p_type
   USE qs_block_davidson_types,         ONLY: block_davidson_release,&
                                              davidson_type
   USE qs_density_mixing_types,         ONLY: direct_mixing_nr,&
                                              mixing_storage_create,&
                                              mixing_storage_release,&
                                              mixing_storage_type
   USE qs_diis_types,                   ONLY: qs_diis_b_release,&
                                              qs_diis_buffer_type
   USE qs_fb_env_types,                 ONLY: fb_env_has_data,&
                                              fb_env_nullify,&
                                              fb_env_obj,&
                                              fb_env_release
   USE qs_ot_types,                     ONLY: qs_ot_destroy,&
                                              qs_ot_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_scf_types'

   INTEGER, PARAMETER, PUBLIC :: general_diag_method_nr = 1, &
                                 special_diag_method_nr = 2, &
                                 ot_diag_method_nr = 3, &
                                 block_krylov_diag_method_nr = 4, &
                                 block_davidson_diag_method_nr = 5, &
                                 ot_method_nr = 10, &
                                 filter_matrix_diag_method_nr = 20

   INTEGER, SAVE, PRIVATE :: last_scf_env_id = 0

   PUBLIC :: qs_scf_env_type
   PUBLIC :: scf_env_retain, scf_env_release, scf_env_create, scf_env_did_change
   PUBLIC :: krylov_space_type, krylov_space_create, subspace_env_type
   PUBLIC :: diag_subspace_env_create

! **************************************************************************************************
!> \brief wrapper for temporary and cached objects used in the scf iteration
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE qs_outer_scf_type
      INTEGER :: iter_count

      ! these are the variable of outer loop.
      ! right now, we assume that they can be easily written as
      ! small arrays, but we might want to go the cp_fm_types
      ! at a later stage
      ! also, we just store the full iteration history
      REAL(KIND=dp), DIMENSION(:), POINTER :: energy
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: variables
      REAL(KIND=dp), DIMENSION(:, :), POINTER :: gradient
      INTEGER, DIMENSION(:), POINTER :: count
   END TYPE qs_outer_scf_type

   TYPE krylov_space_type
      INTEGER :: nkrylov, nblock, nmo_conv, nmo_nc, max_iter
      LOGICAL :: always_check_conv
      REAL(dp) :: eps_std_diag, eps_conv, eps_adapt, max_res_norm, min_res_norm
      REAL(dp), DIMENSION(:), POINTER :: c_eval, t_eval
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER :: v_mat, mo_conv, mo_refine
      TYPE(cp_fm_type), POINTER ::  tmp_mat
!NOTE: the following matrices are small and could be used as standard array rather than distributed fm
      TYPE(cp_fm_type), POINTER :: block1_mat, block2_mat, block3_mat, block4_mat, block5_mat
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER ::  c_vec, chc_mat
   END TYPE krylov_space_type

   TYPE subspace_env_type
      INTEGER :: max_iter, mixing_method
      REAL(dp) :: eps_diag_sub, eps_ene, eps_adapt
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: p_matrix_store
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER :: p_matrix_mix
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER :: chc_mat, c_vec, c0
      TYPE(mixing_storage_type), POINTER :: mixing_store
   END TYPE subspace_env_type

   ! **************************************************************************************************
   TYPE qs_scf_env_type
      TYPE(qs_outer_scf_type) :: outer_scf
      INTEGER :: ref_count, id_nr, print_count, iter_count
      INTEGER :: cholesky_method, nelectron, method, mixing_method, nskip_mixing
      REAL(KIND=dp) :: iter_param, iter_delta, p_mix_alpha
      CHARACTER(len=15) :: iter_method
      COMPLEX(KIND=dp), DIMENSION(:, :, :), POINTER :: cc_buffer
      LOGICAL :: print_iter_line, skip_mixing, skip_diis, needs_ortho
      TYPE(mixing_storage_type), POINTER :: mixing_store
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER :: scf_work1
      TYPE(cp_fm_type), POINTER  :: scf_work2, ortho, ortho_m1, s_half, s_minus_one
      TYPE(krylov_space_type), POINTER :: krylov_space
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER :: p_delta, p_mix_new
      TYPE(cp_dbcsr_type), POINTER :: ortho_dbcsr, buf1_dbcsr, buf2_dbcsr
      TYPE(preconditioner_p_type), DIMENSION(:), POINTER :: ot_preconditioner
      TYPE(qs_ot_type), POINTER, DIMENSION(:)  :: qs_ot_env
      TYPE(qs_diis_buffer_type), POINTER :: scf_diis_buffer
      TYPE(subspace_env_type), POINTER :: subspace_env
      TYPE(davidson_type), POINTER, DIMENSION(:) :: block_davidson_env
      TYPE(fb_env_obj) :: filter_matrix_env
   END TYPE qs_scf_env_type

CONTAINS

! **************************************************************************************************
!> \brief allocates and initialize an scf_env
!> \param scf_env the scf env to initialize
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE scf_env_create(scf_env)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'scf_env_create', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      ALLOCATE (scf_env)
      scf_env%ref_count = 1
      scf_env%print_count = 0
      last_scf_env_id = last_scf_env_id+1
      scf_env%id_nr = last_scf_env_id
      scf_env%print_count = 0
      scf_env%iter_count = 0
      scf_env%nelectron = 0
      scf_env%iter_param = 0.0_dp
      scf_env%iter_delta = 0.0_dp
      scf_env%iter_method = ""
      scf_env%print_iter_line = .TRUE.
      scf_env%skip_mixing = .FALSE.
      scf_env%skip_diis = .FALSE.
      scf_env%needs_ortho = .FALSE.
      scf_env%method = -1
      scf_env%mixing_method = -1
      scf_env%p_mix_alpha = 1.0_dp
      scf_env%cholesky_method = -1
      scf_env%outer_scf%iter_count = 0
      NULLIFY (scf_env%outer_scf%variables)
      NULLIFY (scf_env%outer_scf%gradient)
      NULLIFY (scf_env%outer_scf%energy)
      NULLIFY (scf_env%outer_scf%count)
      NULLIFY (scf_env%scf_work1)
      NULLIFY (scf_env%scf_work2)
      NULLIFY (scf_env%ortho)
      NULLIFY (scf_env%ortho_dbcsr)
      NULLIFY (scf_env%ortho_m1)
      NULLIFY (scf_env%p_mix_new)
      NULLIFY (scf_env%ot_preconditioner)
      NULLIFY (scf_env%qs_ot_env)
      NULLIFY (scf_env%scf_diis_buffer)
      NULLIFY (scf_env%buf1_dbcsr)
      NULLIFY (scf_env%buf2_dbcsr)
      NULLIFY (scf_env%s_half)
      NULLIFY (scf_env%p_delta)
      NULLIFY (scf_env%s_minus_one)
      NULLIFY (scf_env%cc_buffer)
      NULLIFY (scf_env%mixing_store)
      NULLIFY (scf_env%krylov_space)
      NULLIFY (scf_env%subspace_env)
      NULLIFY (scf_env%block_davidson_env)
      CALL fb_env_nullify(scf_env%filter_matrix_env)

      CALL timestop(handle)

   END SUBROUTINE scf_env_create

! **************************************************************************************************
!> \brief retains an scf_env (see doc/ReferenceCounting.html)
!> \param scf_env the environment to retain
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE scf_env_retain(scf_env)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'scf_env_retain', routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(scf_env)))CALL cp__a("qs_scf_types.F",213)
      IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("qs_scf_types.F",214)
      scf_env%ref_count = scf_env%ref_count+1

      CALL timestop(handle)

   END SUBROUTINE scf_env_retain

! **************************************************************************************************
!> \brief function to be called to inform the scf_env about changes
!> \param scf_env the scf env to inform
!> \par History
!>      03.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE scf_env_did_change(scf_env)
      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'scf_env_did_change', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(scf_env)))CALL cp__a("qs_scf_types.F",238)
      IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("qs_scf_types.F",239)

      IF (ASSOCIATED(scf_env%p_mix_new)) THEN
         CALL cp_dbcsr_deallocate_matrix_set(scf_env%p_mix_new)
      END IF
      IF (ASSOCIATED(scf_env%p_delta)) THEN
         CALL cp_dbcsr_deallocate_matrix_set(scf_env%p_delta)
      END IF

      CALL timestop(handle)

   END SUBROUTINE scf_env_did_change

! **************************************************************************************************
!> \brief releases an scf_env (see doc/ReferenceCounting.html)
!> \param scf_env the environment to release
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE scf_env_release(scf_env)

      TYPE(qs_scf_env_type), POINTER                     :: scf_env

      CHARACTER(len=*), PARAMETER :: routineN = 'scf_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, i

      CALL timeset(routineN, handle)

      IF (ASSOCIATED(scf_env)) THEN
         IF(.NOT.(scf_env%ref_count > 0))CALL cp__a("qs_scf_types.F",271)
         scf_env%ref_count = scf_env%ref_count-1
         IF (scf_env%ref_count < 1) THEN
            CALL cp_fm_vect_dealloc(scf_env%scf_work1)
            CALL cp_fm_release(scf_env%scf_work2)
            CALL cp_fm_release(scf_env%ortho)
            CALL cp_fm_release(scf_env%ortho_m1)
            IF (ASSOCIATED(scf_env%ortho_dbcsr)) THEN
               ! we should not end up here, and give back using the pools
               IF(.NOT.(.TRUE.))CALL cp__a("qs_scf_types.F",280)
               CALL cp_dbcsr_deallocate_matrix(scf_env%ortho_dbcsr)
            END IF
            IF (ASSOCIATED(scf_env%buf1_dbcsr)) THEN
               ! we should not end up here, and give back using the pools
               IF(.NOT.(.TRUE.))CALL cp__a("qs_scf_types.F",285)
               CALL cp_dbcsr_deallocate_matrix(scf_env%buf1_dbcsr)
            END IF
            IF (ASSOCIATED(scf_env%buf2_dbcsr)) THEN
               ! we should not end up here, and give back using the pools
               IF(.NOT.(.TRUE.))CALL cp__a("qs_scf_types.F",290)
               CALL cp_dbcsr_deallocate_matrix(scf_env%buf2_dbcsr)
            END IF
            CALL cp_fm_release(scf_env%s_half)
            CALL cp_fm_release(scf_env%s_minus_one)
            IF (ASSOCIATED(scf_env%p_mix_new)) THEN
               ! we should not end up here, and give back using the pools
               IF(.NOT.(.TRUE.))CALL cp__a("qs_scf_types.F",297)
               CALL cp_dbcsr_deallocate_matrix_set(scf_env%p_mix_new)
            END IF
            IF (ASSOCIATED(scf_env%p_delta)) THEN
               ! we should not end up here, and give back using the pools
               IF(.NOT.(.TRUE.))CALL cp__a("qs_scf_types.F",302)
               CALL cp_dbcsr_deallocate_matrix_set(scf_env%p_delta)
            END IF
            IF (ASSOCIATED(scf_env%ot_preconditioner)) THEN
               DO i = 1, SIZE(scf_env%ot_preconditioner)
                  CALL destroy_preconditioner(scf_env%ot_preconditioner(i)%preconditioner)
                  DEALLOCATE (scf_env%ot_preconditioner(i)%preconditioner)
               END DO
               DEALLOCATE (scf_env%ot_preconditioner)
            END IF
            IF (ASSOCIATED(scf_env%qs_ot_env)) THEN
               DO i = 1, SIZE(scf_env%qs_ot_env)
                  CALL qs_ot_destroy(scf_env%qs_ot_env(i))
               END DO
               DEALLOCATE (scf_env%qs_ot_env)
            END IF
            CALL qs_diis_b_release(scf_env%scf_diis_buffer)
            IF (ASSOCIATED(scf_env%outer_scf%variables)) THEN
               DEALLOCATE (scf_env%outer_scf%variables)
            END IF
            IF (ASSOCIATED(scf_env%outer_scf%count)) THEN
               DEALLOCATE (scf_env%outer_scf%count)
            END IF
            IF (ASSOCIATED(scf_env%outer_scf%gradient)) THEN
               DEALLOCATE (scf_env%outer_scf%gradient)
            END IF
            IF (ASSOCIATED(scf_env%outer_scf%energy)) THEN
               DEALLOCATE (scf_env%outer_scf%energy)
            END IF
            IF (ASSOCIATED(scf_env%cc_buffer)) THEN
               DEALLOCATE (scf_env%cc_buffer)
            END IF
            IF (ASSOCIATED(scf_env%mixing_store)) THEN
               CALL mixing_storage_release(scf_env%mixing_store)
            END IF
            IF (ASSOCIATED(scf_env%krylov_space)) THEN
               CALL krylov_space_release(scf_env%krylov_space)
            END IF
            IF (ASSOCIATED(scf_env%subspace_env)) THEN
               CALL diag_subspace_env_release(scf_env%subspace_env)
            END IF
            IF (ASSOCIATED(scf_env%block_davidson_env)) THEN
               CALL block_davidson_release(scf_env%block_davidson_env)
            END IF
            IF (fb_env_has_data(scf_env%filter_matrix_env)) THEN
               CALL fb_env_release(scf_env%filter_matrix_env)
            END IF
            DEALLOCATE (scf_env)
         END IF
      END IF

      NULLIFY (scf_env)

      CALL timestop(handle)

   END SUBROUTINE scf_env_release

! **************************************************************************************************
!> \brief  creates krylov space
!> \param krylov_space ...
!> \param scf_section ...
!> \par History
!>      05.2009 created [MI]
!> \author [MI]
! **************************************************************************************************
   SUBROUTINE krylov_space_create(krylov_space, scf_section)

      TYPE(krylov_space_type), POINTER                   :: krylov_space
      TYPE(section_vals_type), POINTER                   :: scf_section

      CHARACTER(len=*), PARAMETER :: routineN = 'krylov_space_create', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(krylov_space)))CALL cp__a("qs_scf_types.F",375)
      ALLOCATE (krylov_space)

      NULLIFY (krylov_space%c_eval, krylov_space%t_eval)
      NULLIFY (krylov_space%v_mat)
      NULLIFY (krylov_space%mo_conv, krylov_space%mo_refine)
      NULLIFY (krylov_space%chc_mat, krylov_space%c_vec)
      NULLIFY (krylov_space%tmp_mat)
      NULLIFY (krylov_space%block1_mat, krylov_space%block2_mat)
      NULLIFY (krylov_space%block3_mat, krylov_space%block4_mat, krylov_space%block5_mat)

      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%MAX_ITER", &
                                i_val=krylov_space%max_iter)
      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%KRYLOV%NKRYLOV", &
                                i_val=krylov_space%nkrylov)
      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%KRYLOV%NBLOCK", &
                                i_val=krylov_space%nblock)
      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%KRYLOV%EPS_KRYLOV", &
                                r_val=krylov_space%eps_conv)
      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%KRYLOV%EPS_STD_DIAG", &
                                r_val=krylov_space%eps_std_diag)
      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%EPS_ADAPT", &
                                r_val=krylov_space%eps_adapt)
      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%KRYLOV%CHECK_MOS_CONV", &
                                l_val=krylov_space%always_check_conv)

   END SUBROUTINE krylov_space_create

! **************************************************************************************************
!> \brief releases krylov space
!> \param krylov_space ...
!> \par History
!>      05.2009 created [MI]
!> \author [MI]
! **************************************************************************************************
   SUBROUTINE krylov_space_release(krylov_space)
      TYPE(krylov_space_type), POINTER                   :: krylov_space

      CHARACTER(len=*), PARAMETER :: routineN = 'krylov_space_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(krylov_space)) THEN

         DEALLOCATE (krylov_space%c_eval)
         DEALLOCATE (krylov_space%t_eval)

         DO i = 1, SIZE(krylov_space%v_mat)
            CALL cp_fm_release(krylov_space%v_mat(i)%matrix)
         END DO
         DEALLOCATE (krylov_space%v_mat)

         DO i = 1, SIZE(krylov_space%mo_conv)
            CALL cp_fm_release(krylov_space%mo_conv(i)%matrix)
         END DO
         DEALLOCATE (krylov_space%mo_conv)

         DO i = 1, SIZE(krylov_space%mo_refine)
            CALL cp_fm_release(krylov_space%mo_refine(i)%matrix)
         END DO
         DEALLOCATE (krylov_space%mo_refine)

         DO i = 1, SIZE(krylov_space%chc_mat)
            CALL cp_fm_release(krylov_space%chc_mat(i)%matrix)
         END DO
         DEALLOCATE (krylov_space%chc_mat)
         DO i = 1, SIZE(krylov_space%c_vec)
            CALL cp_fm_release(krylov_space%c_vec(i)%matrix)
         END DO
         DEALLOCATE (krylov_space%c_vec)
         CALL cp_fm_release(krylov_space%tmp_mat)
         CALL cp_fm_release(krylov_space%block1_mat)
         CALL cp_fm_release(krylov_space%block2_mat)
         CALL cp_fm_release(krylov_space%block3_mat)
         CALL cp_fm_release(krylov_space%block4_mat)
         CALL cp_fm_release(krylov_space%block5_mat)

         DEALLOCATE (krylov_space)

         NULLIFY (krylov_space)
      END IF

   END SUBROUTINE krylov_space_release

! **************************************************************************************************
!> \brief creates subspace-rotation environment
!> \param subspace_env ...
!> \param scf_section ...
!> \param ecut ...
!> \par History
!>      09.2009 created [MI]
!> \author [MI]
! **************************************************************************************************
   SUBROUTINE diag_subspace_env_create(subspace_env, scf_section, ecut)

      TYPE(subspace_env_type), POINTER                   :: subspace_env
      TYPE(section_vals_type), POINTER                   :: scf_section
      REAL(dp), INTENT(IN)                               :: ecut

      CHARACTER(len=*), PARAMETER :: routineN = 'diag_subspace_env_create', &
         routineP = moduleN//':'//routineN

      LOGICAL                                            :: do_mixing
      TYPE(section_vals_type), POINTER                   :: mixing_section

      IF(.NOT.(.NOT. ASSOCIATED(subspace_env)))CALL cp__a("qs_scf_types.F",481)
      ALLOCATE (subspace_env)

      NULLIFY (subspace_env%p_matrix_store)
      NULLIFY (subspace_env%p_matrix_mix)
      NULLIFY (subspace_env%chc_mat)
      NULLIFY (subspace_env%c_vec)
      NULLIFY (subspace_env%c0)
      NULLIFY (subspace_env%mixing_store)
      NULLIFY (mixing_section)

      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DIAG_SUB_SCF%MAX_ITER", &
                                i_val=subspace_env%max_iter)
      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DIAG_SUB_SCF%EPS_ENE", &
                                r_val=subspace_env%eps_ene)
      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DIAG_SUB_SCF%EPS_SKIP_SUB_DIAG", &
                                r_val=subspace_env%eps_diag_sub)
      CALL section_vals_val_get(scf_section, "DIAGONALIZATION%DIAG_SUB_SCF%EPS_ADAPT_SCF", &
                                r_val=subspace_env%eps_adapt)
      subspace_env%mixing_method = 0
      do_mixing = .FALSE.
      mixing_section => section_vals_get_subs_vals(scf_section, "DIAGONALIZATION%DIAG_SUB_SCF%MIXING")
      CALL section_vals_val_get(mixing_section, "_SECTION_PARAMETERS_", &
                                l_val=do_mixing)
      IF (do_mixing) THEN
         CALL section_vals_val_get(mixing_section, "METHOD", &
                                   i_val=subspace_env%mixing_method)
         IF (subspace_env%mixing_method >= direct_mixing_nr) &
            CALL mixing_storage_create(subspace_env%mixing_store, mixing_section, &
                                       subspace_env%mixing_method, ecut=ecut)
      END IF

   END SUBROUTINE diag_subspace_env_create

! **************************************************************************************************
!> \brief releases subspace-rotation environment
!> \param subspace_env ...
!> \par History
!>      09.2009 created [MI]
!> \author [MI]
! **************************************************************************************************
   SUBROUTINE diag_subspace_env_release(subspace_env)
      TYPE(subspace_env_type), POINTER                   :: subspace_env

      CHARACTER(len=*), PARAMETER :: routineN = 'diag_subspace_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(subspace_env)) THEN

         IF (ASSOCIATED(subspace_env%p_matrix_store)) THEN

            IF(.NOT.(.TRUE.))CALL cp__a("qs_scf_types.F",534)
            CALL cp_dbcsr_deallocate_matrix_set(subspace_env%p_matrix_store)
         ENDIF
         DO i = 1, SIZE(subspace_env%chc_mat)
            CALL cp_fm_release(subspace_env%chc_mat(i)%matrix)
         END DO
         DEALLOCATE (subspace_env%chc_mat)
         DO i = 1, SIZE(subspace_env%c_vec)
            CALL cp_fm_release(subspace_env%c_vec(i)%matrix)
         END DO
         DEALLOCATE (subspace_env%c_vec)
         DO i = 1, SIZE(subspace_env%c0)
            CALL cp_fm_release(subspace_env%c0(i)%matrix)
         END DO
         DEALLOCATE (subspace_env%c0)

         IF (ASSOCIATED(subspace_env%mixing_store)) THEN
            CALL mixing_storage_release(subspace_env%mixing_store)
         END IF

         DEALLOCATE (subspace_env)
      END IF

   END SUBROUTINE diag_subspace_env_release

END MODULE qs_scf_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief contains the types and subroutines for dealing with the lri_env
!>        lri : local resolution of the identity
!> \par History
!>      created JGH [08.2012]
!>      Dorothea Golze [02.2014] (1) extended, re-structured, cleaned
!>                               (2) debugged
!> \authors JGH
!>          Dorothea Golze
! **************************************************************************************************
MODULE lri_environment_types
   USE atomic_kind_types,               ONLY: atomic_kind_type,&
                                              get_atomic_kind
   USE basis_set_types,                 ONLY: deallocate_gto_basis_set,&
                                              gto_basis_set_p_type,&
                                              gto_basis_set_type
   USE kinds,                           ONLY: dp
   USE qs_neighbor_list_types,          ONLY: deallocate_neighbor_list_set,&
                                              get_iterator_info,&
                                              neighbor_list_iterate,&
                                              neighbor_list_iterator_create,&
                                              neighbor_list_iterator_p_type,&
                                              neighbor_list_iterator_release,&
                                              neighbor_list_set_p_type
   USE virial_types,                    ONLY: virial_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

! **************************************************************************************************
   TYPE lri_rhoab_type
      ! number of spherical basis functions (a)
      INTEGER                                                 :: nba
      ! number of spherical basis functions (b)
      INTEGER                                                 :: nbb
      ! number of spherical fit basis functions (ai)
      INTEGER                                                 :: nfa
      ! number of spherical fit basis functions (bi)
      INTEGER                                                 :: nfb
      ! expansion coeffs for RI density
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: avec
      ! projection coeffs for RI density: SUM_ab (ab,i)*Pab
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: tvec
      ! integral (ai) * sinv * tvec
      REAL(KIND=dp)                                           :: nst
      ! Lagrange parameter
      REAL(KIND=dp)                                           :: lambda
      ! Charge of pair density
      REAL(KIND=dp)                                           :: charge
   END TYPE lri_rhoab_type

! **************************************************************************************************

   TYPE lri_int_type
      ! whether to calculate force for pair
      LOGICAL                                                 :: calc_force_pair
      ! number of spherical basis functions (a)
      INTEGER                                                 :: nba
      ! number of spherical basis functions (b)
      INTEGER                                                 :: nbb
      ! number of spherical fit basis functions (ai)
      INTEGER                                                 :: nfa
      ! number of spherical fit basis functions (bi)
      INTEGER                                                 :: nfb
      ! condition number of overlap matrix
      REAL(KIND=dp)                                           :: cond_num
      ! integrals (a,b,ai)
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER                :: abaint
      ! integrals (a,b,bi)
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER                :: abbint
      ! integrals (da/dA,b,dai/dA)
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER              :: dabdaint
      ! integrals (da/dA,b,bi)
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER              :: dabbint
      ! integrals (a,b)
      REAL(KIND=dp), DIMENSION(:, :), POINTER                  :: soo
      ! derivative d(a,b)/dA
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER                :: dsoo
      ! integrals (ai,bi)
      REAL(KIND=dp), DIMENSION(:, :), POINTER                  :: sab
      ! derivative d(ai,bi)/dA
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER                :: dsab
      ! derivative of fit coeff dacoef/dpmatrix
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER                :: dacoef
      ! inverse of integrals (ai,bi)
      REAL(KIND=dp), DIMENSION(:, :), POINTER                  :: sinv
      ! integral (ai) / (bi), dim(1..nfa,nfa+1..nfa+nfb)
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: n
      ! sinv * (ai)
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: sn
      ! (ai) * sinv * (ai)
      REAL(KIND=dp)                                           :: nsn
      ! dmax: max deviation for integrals of primitive gtos; for debugging
      ! dmax for overlap integrals (ai,bi); fit bas
      REAL(KIND=dp)                                           :: dmax_ab
      ! dmax for overlap integrals (a,b); orb bas
      REAL(KIND=dp)                                           :: dmax_oo
      ! dmax for integrals (a,b,ai)
      REAL(KIND=dp)                                           :: dmax_aba
      ! dmax for integrals (a,b,bi)
      REAL(KIND=dp)                                           :: dmax_abb
   END TYPE lri_int_type

   TYPE lri_int_rho_type
      ! integrals (aa,bb), orb basis
      REAL(KIND=dp), DIMENSION(:, :, :, :), POINTER              :: soaabb
      ! dmax for (aa,bb) integrals; for debugging
      REAL(KIND=dp)                                           :: dmax_aabb
   END TYPE lri_int_rho_type

   TYPE lri_node_type
      INTEGER                                                 :: nnode
      TYPE(lri_int_type), DIMENSION(:), POINTER               :: lri_int
      TYPE(lri_int_rho_type), DIMENSION(:), POINTER           :: lri_int_rho
      TYPE(lri_rhoab_type), DIMENSION(:), POINTER             :: lri_rhoab
   END TYPE lri_node_type

   TYPE lri_atom_type
      INTEGER                                                 :: natom
      TYPE(lri_node_type), DIMENSION(:), POINTER              :: lri_node
   END TYPE lri_atom_type

   TYPE lri_list_type
      INTEGER                                                 :: nkind
      TYPE(lri_atom_type), DIMENSION(:), POINTER              :: lri_atom
   END TYPE lri_list_type

   TYPE lri_list_p_type
      TYPE(lri_list_type), POINTER                             :: lri_list
   END TYPE lri_list_p_type

! **************************************************************************************************

   TYPE lri_bas_type
      ! integral of ri basis fbas
      REAL(KIND=dp), DIMENSION(:), POINTER                    :: int_fbas
      ! self overlap ri basis
      REAL(KIND=dp), DIMENSION(:, :), POINTER                  :: ri_ovlp
      ! self overlap orb basis
      REAL(KIND=dp), DIMENSION(:, :), POINTER                  :: orb_ovlp
      ! contraction matrix for SHG integrals ri basis
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER                :: scon_ri
      ! contraction matrix for SHG integrals orb basis
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER                :: scon_orb
   END TYPE lri_bas_type

! **************************************************************************************************

   TYPE lri_clebsch_gordon_type
      ! Clebsch-Gordon (CG) coefficients
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER                :: cg_coeff
      ! list of non-zero CG coefficients
      INTEGER, DIMENSION(:, :, :), POINTER                      :: cg_none0_list
      ! number of non-zero CG coefficients
      INTEGER, DIMENSION(:, :), POINTER                        :: ncg_none0
   END TYPE lri_clebsch_gordon_type

! **************************************************************************************************

   TYPE lri_environment_type
      INTEGER                                                 :: id_nr, ref_count, in_use
      ! parameter for (pseudo)inverse of overlap
      INTEGER                                                 :: lri_overlap_inv
      ! flag for debugging lri integrals
      LOGICAL                                                 :: debug
      ! flag for shg (solid haromonic Gaussian) integrals
      LOGICAL                                                 :: use_shg_integrals
      ! parameter for inversion (autoselect); maximal condition
      ! number up to where inversion is legal
      REAL(KIND=dp)                                           :: cond_max
      ! parameter for checking distance between atom pairs
      REAL(KIND=dp)                                           :: delta
      ! orbital basis set
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER       :: orb_basis
      ! lri (fit) basis set
      TYPE(gto_basis_set_p_type), DIMENSION(:), POINTER       :: ri_basis
      ! orb_basis neighborlist, LRI integrals
      TYPE(neighbor_list_set_p_type), DIMENSION(:), POINTER   :: soo_list
      ! local RI integrals
      TYPE(lri_list_type), POINTER                            :: lri_ints
      ! local integral of rho**2; for optimization
      TYPE(lri_list_type), POINTER                            :: lri_ints_rho
      ! properties of orb and aux basis
      TYPE(lri_bas_type), DIMENSION(:), POINTER               :: bas_prop
      ! Clebsch-Gordon for solid harmonics
      TYPE(lri_clebsch_gordon_type), POINTER                  :: cg_shg
   END TYPE lri_environment_type

! **************************************************************************************************

   TYPE lri_kind_type
      ! expansion coeff for lri density dim(natom,nsgf)
      REAL(KIND=dp), DIMENSION(:, :), POINTER                   :: acoef
      ! integrals V*fbas (potential*fit basis) dim(natom,nsgf)
      REAL(KIND=dp), DIMENSION(:, :), POINTER                   :: v_int
      ! SUM_i integral(V*fbas_i)*davec/dR dim(natom,3)
      REAL(KIND=dp), DIMENSION(:, :), POINTER                   :: v_dadr
      ! integrals V*dfbas/dR
      REAL(KIND=dp), DIMENSION(:, :), POINTER                   :: v_dfdr
   END TYPE lri_kind_type

   TYPE lri_spin_type
      TYPE(lri_kind_type), DIMENSION(:), POINTER              :: lri_kinds
   END TYPE lri_spin_type

! **************************************************************************************************

   TYPE lri_force_type
      REAL(KIND=dp), DIMENSION(:), POINTER                     :: st
      REAL(KIND=dp), DIMENSION(:, :), POINTER                   :: dssn, &
                                                                   sdssn, &
                                                                   dsst, &
                                                                   sdsst, &
                                                                   sdt
      ! derivative dtvec/dR
      REAL(KIND=dp), DIMENSION(:, :), POINTER                   :: dtvec
      ! derivative davec/dR
      REAL(KIND=dp), DIMENSION(:, :), POINTER                   :: davec
      ! derivative overlap matrix dS/dR
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER                 :: ds
   END TYPE lri_force_type

! **************************************************************************************************

   TYPE lri_density_type
      INTEGER :: id_nr, ref_count, in_use
      INTEGER                                                 :: nspin
      ! pair density expansion (nspin)
      TYPE(lri_list_p_type), DIMENSION(:), POINTER             :: lri_rhos
      ! coefficients of RI expansion and gradients (nspin)
      TYPE(lri_spin_type), DIMENSION(:), POINTER               :: lri_coefs
      TYPE(lri_force_type), POINTER                           :: lri_force
   END TYPE lri_density_type

! **************************************************************************************************

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'lri_environment_types'
   INTEGER, SAVE, PRIVATE :: last_lri_env_id = 0
   INTEGER, SAVE, PRIVATE :: last_lri_density_id = 0

   PUBLIC :: lri_environment_type, &
             lri_force_type, lri_list_type, &
             lri_int_type, lri_int_rho_type, lri_density_type, &
             lri_kind_type, lri_rhoab_type
   PUBLIC :: lri_env_create, lri_env_release, allocate_lri_coefs, &
             allocate_lri_ints, allocate_lri_ints_rho, lri_density_create, &
             lri_density_release, allocate_lri_rhos, allocate_lri_force_components, &
             deallocate_lri_ints, deallocate_lri_ints_rho, &
             deallocate_lri_force_components, deallocate_bas_properties

! **************************************************************************************************

CONTAINS

! **************************************************************************************************
!> \brief creates and initializes an lri_env
!> \param lri_env the lri_environment you want to create
! **************************************************************************************************
   SUBROUTINE lri_env_create(lri_env)

      TYPE(lri_environment_type), POINTER                :: lri_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_env_create', routineP = moduleN//':'//routineN

      ALLOCATE (lri_env)

      last_lri_env_id = last_lri_env_id+1
      lri_env%id_nr = last_lri_env_id
      lri_env%ref_count = 1
      lri_env%in_use = 0

      lri_env%debug = .FALSE.
      lri_env%delta = 1.E-6_dp

      NULLIFY (lri_env%orb_basis)
      NULLIFY (lri_env%ri_basis)

      NULLIFY (lri_env%soo_list)
      NULLIFY (lri_env%lri_ints)
      NULLIFY (lri_env%lri_ints_rho)
      NULLIFY (lri_env%bas_prop)
      NULLIFY (lri_env%cg_shg)

      ALLOCATE (lri_env%cg_shg)

   END SUBROUTINE lri_env_create

! **************************************************************************************************
!> \brief releases the given lri_env
!> \param lri_env the lri environment to release
! **************************************************************************************************
   SUBROUTINE lri_env_release(lri_env)

      TYPE(lri_environment_type), POINTER                :: lri_env

      CHARACTER(len=*), PARAMETER :: routineN = 'lri_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, ikind, nkind

      IF (ASSOCIATED(lri_env)) THEN
         lri_env%ref_count = 0

         ! deallocate basis sets
         IF (ASSOCIATED(lri_env%orb_basis)) THEN
            nkind = SIZE(lri_env%orb_basis)
            DO ikind = 1, nkind
               CALL deallocate_gto_basis_set(lri_env%orb_basis(ikind)%gto_basis_set)
            END DO
            DEALLOCATE (lri_env%orb_basis)
         END IF
         IF (ASSOCIATED(lri_env%ri_basis)) THEN
            nkind = SIZE(lri_env%ri_basis)
            DO ikind = 1, nkind
               CALL deallocate_gto_basis_set(lri_env%ri_basis(ikind)%gto_basis_set)
            END DO
            DEALLOCATE (lri_env%ri_basis)
         END IF
         IF (ASSOCIATED(lri_env%soo_list)) THEN
            DO i = 1, SIZE(lri_env%soo_list)
               CALL deallocate_neighbor_list_set(lri_env%soo_list(i)%neighbor_list_set)
            END DO
            DEALLOCATE (lri_env%soo_list)
         END IF
         IF (ASSOCIATED(lri_env%lri_ints)) THEN
            CALL deallocate_lri_ints(lri_env%lri_ints)
         END IF
         IF (ASSOCIATED(lri_env%lri_ints_rho)) THEN
            CALL deallocate_lri_ints_rho(lri_env%lri_ints_rho)
         END IF
         CALL deallocate_bas_properties(lri_env)
         IF (ASSOCIATED(lri_env%cg_shg)) THEN
            DEALLOCATE (lri_env%cg_shg%cg_coeff)
            DEALLOCATE (lri_env%cg_shg%cg_none0_list)
            DEALLOCATE (lri_env%cg_shg%ncg_none0)
            DEALLOCATE (lri_env%cg_shg)
         ENDIF
         DEALLOCATE (lri_env)
      END IF
      NULLIFY (lri_env)

   END SUBROUTINE lri_env_release

! **************************************************************************************************
!> \brief creates and initializes an lri_density environment
!> \param lri_density the lri_density environment you want to create
! **************************************************************************************************
   SUBROUTINE lri_density_create(lri_density)

      TYPE(lri_density_type), POINTER                    :: lri_density

      CHARACTER(LEN=*), PARAMETER :: routineN = 'lri_density_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (lri_density)

      last_lri_density_id = last_lri_density_id+1
      lri_density%id_nr = last_lri_density_id
      lri_density%ref_count = 1
      lri_density%in_use = 0

      lri_density%nspin = 0

      NULLIFY (lri_density%lri_rhos)
      NULLIFY (lri_density%lri_coefs)
      NULLIFY (lri_density%lri_force)

   END SUBROUTINE lri_density_create

! **************************************************************************************************
!> \brief releases the given lri_density
!> \param lri_density the lri_density to release
! **************************************************************************************************
   SUBROUTINE lri_density_release(lri_density)
      TYPE(lri_density_type), POINTER                    :: lri_density

      CHARACTER(len=*), PARAMETER :: routineN = 'lri_density_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(lri_density)) THEN
         lri_density%ref_count = 0

         CALL deallocate_lri_rhos(lri_density%lri_rhos)
         CALL deallocate_lri_coefs(lri_density%lri_coefs)
         CALL deallocate_lri_force_components(lri_density%lri_force)

         DEALLOCATE (lri_density)
      END IF
      NULLIFY (lri_density)

   END SUBROUTINE lri_density_release

! **************************************************************************************************
!> \brief allocate lri_ints, matrices that store LRI integrals
!> \param lri_env ...
!> \param lri_ints structure storing the LRI integrals
!> \param nkind number of atom kinds
!> \param calculate_forces ...
!> \param virial ...
! **************************************************************************************************
   SUBROUTINE allocate_lri_ints(lri_env, lri_ints, nkind, calculate_forces, &
                                virial)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_list_type), POINTER                       :: lri_ints
      INTEGER, INTENT(IN)                                :: nkind
      LOGICAL, INTENT(IN)                                :: calculate_forces
      TYPE(virial_type), POINTER                         :: virial

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_lri_ints', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iac, iatom, ikind, ilist, jatom, &
                                                            jkind, jneighbor, nba, nbb, nfa, nfb, &
                                                            nlist, nn, nneighbor
      LOGICAL                                            :: use_virial
      REAL(KIND=dp)                                      :: dab, rab(3)
      TYPE(gto_basis_set_type), POINTER                  :: fbasa, fbasb, obasa, obasb
      TYPE(lri_int_type), POINTER                        :: lrii
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator

      NULLIFY (fbasa, fbasb, lrii, nl_iterator, obasa, obasb)

      use_virial = virial%pv_availability .AND. (.NOT. virial%pv_numer)

      ALLOCATE (lri_ints)

      lri_ints%nkind = nkind
      ALLOCATE (lri_ints%lri_atom(nkind*nkind))

      DO i = 1, nkind*nkind
         NULLIFY (lri_ints%lri_atom(i)%lri_node)
         lri_ints%lri_atom(i)%natom = 0
      END DO

      CALL neighbor_list_iterator_create(nl_iterator, lri_env%soo_list)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)

         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                nlist=nlist, ilist=ilist, nnode=nneighbor, inode=jneighbor, &
                                iatom=iatom, jatom=jatom, r=rab)

         iac = ikind+nkind*(jkind-1)
         dab = SQRT(SUM(rab*rab))

         obasa => lri_env%orb_basis(ikind)%gto_basis_set
         obasb => lri_env%orb_basis(jkind)%gto_basis_set
         fbasa => lri_env%ri_basis(ikind)%gto_basis_set
         fbasb => lri_env%ri_basis(jkind)%gto_basis_set

         IF (.NOT. ASSOCIATED(obasa)) CYCLE
         IF (.NOT. ASSOCIATED(obasb)) CYCLE

         IF (.NOT. ASSOCIATED(lri_ints%lri_atom(iac)%lri_node)) THEN
            lri_ints%lri_atom(iac)%natom = nlist
            ALLOCATE (lri_ints%lri_atom(iac)%lri_node(nlist))
            DO i = 1, nlist
               NULLIFY (lri_ints%lri_atom(iac)%lri_node(i)%lri_int)
               lri_ints%lri_atom(iac)%lri_node(i)%nnode = 0
            END DO
         END IF
         IF (.NOT. ASSOCIATED(lri_ints%lri_atom(iac)%lri_node(ilist)%lri_int)) THEN
            lri_ints%lri_atom(iac)%lri_node(ilist)%nnode = nneighbor
            ALLOCATE (lri_ints%lri_atom(iac)%lri_node(ilist)%lri_int(nneighbor))
         END IF

         lrii => lri_ints%lri_atom(iac)%lri_node(ilist)%lri_int(jneighbor)

         nba = obasa%nsgf
         nbb = obasb%nsgf
         nfa = fbasa%nsgf
         nfb = fbasb%nsgf
         nn = nfa+nfb

         ALLOCATE (lrii%abaint(nba, nbb, nfa), &
                   lrii%abbint(nba, nbb, nfb))
         lrii%abaint = 0._dp
         lrii%abbint = 0._dp

         ALLOCATE (lrii%dabdaint(nba, nbb, nfa, 3), &
                   lrii%dabbint(nba, nbb, nfb, 3))
         lrii%dabdaint = 0._dp
         lrii%dabbint = 0._dp

         ALLOCATE (lrii%sab(nfa, nfb), lrii%dsab(nfa, nfb, 3))
         lrii%sab = 0._dp
         lrii%dsab = 0._dp

         IF (iatom == jatom .AND. dab < lri_env%delta) THEN
            ALLOCATE (lrii%sinv(nfa, nfa))
         ELSE
            ALLOCATE (lrii%sinv(nn, nn))
         ENDIF
         lrii%sinv = 0._dp

         IF (iatom == jatom .AND. dab < lri_env%delta) THEN
            ALLOCATE (lrii%n(nfa), lrii%sn(nfa))
         ELSE
            ALLOCATE (lrii%n(nn), lrii%sn(nn))
         ENDIF
         lrii%n = 0._dp
         lrii%sn = 0._dp

         ALLOCATE (lrii%soo(nba, nbb), lrii%dsoo(nba, nbb, 3))
         lrii%soo = 0._dp
         lrii%dsoo = 0._dp

         IF (iatom == jatom .AND. dab < lri_env%delta) THEN
            ALLOCATE (lrii%dacoef(nba, nbb, nfa))
         ELSE
            ALLOCATE (lrii%dacoef(nba, nbb, nn))
         ENDIF
         lrii%dacoef = 0._dp

         lrii%dmax_ab = 0._dp
         lrii%dmax_oo = 0._dp
         lrii%dmax_aba = 0._dp
         lrii%dmax_abb = 0._dp

         lrii%calc_force_pair = .FALSE.

         ! forces: not every pair is giving contributions
         IF (calculate_forces) THEN
            !no forces for self-pair aa
            IF (iatom == jatom .AND. dab < lri_env%delta) THEN
               lrii%calc_force_pair = .FALSE.
            ELSE
               !forces for periodic self-pair aa' required for virial
               IF (iatom == jatom .AND. .NOT. use_virial) THEN
                  lrii%calc_force_pair = .FALSE.
               ELSE
                  lrii%calc_force_pair = .TRUE.
               ENDIF
            ENDIF
         ENDIF

      ENDDO

      CALL neighbor_list_iterator_release(nl_iterator)

   END SUBROUTINE allocate_lri_ints

! **************************************************************************************************
!> \brief allocate lri_ints_rho, storing integral for the exact density
!> \param lri_env ...
!> \param lri_ints_rho structure storing the integrals (aa,bb)
!> \param nkind number of atom kinds
! **************************************************************************************************
   SUBROUTINE allocate_lri_ints_rho(lri_env, lri_ints_rho, nkind)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_list_type), POINTER                       :: lri_ints_rho
      INTEGER, INTENT(IN)                                :: nkind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_lri_ints_rho', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iac, iatom, ikind, ilist, jatom, &
                                                            jkind, jneighbor, nba, nbb, nlist, &
                                                            nneighbor
      TYPE(gto_basis_set_type), POINTER                  :: obasa, obasb
      TYPE(lri_int_rho_type), POINTER                    :: lriir
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator

      ALLOCATE (lri_ints_rho)

      lri_ints_rho%nkind = nkind
      ALLOCATE (lri_ints_rho%lri_atom(nkind*nkind))

      DO i = 1, nkind*nkind
         NULLIFY (lri_ints_rho%lri_atom(i)%lri_node)
         lri_ints_rho%lri_atom(i)%natom = 0
      ENDDO

      CALL neighbor_list_iterator_create(nl_iterator, lri_env%soo_list)
      DO WHILE (neighbor_list_iterate(nl_iterator) == 0)

         CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                nlist=nlist, ilist=ilist, nnode=nneighbor, inode=jneighbor, &
                                iatom=iatom, jatom=jatom)

         iac = ikind+nkind*(jkind-1)

         obasa => lri_env%orb_basis(ikind)%gto_basis_set
         obasb => lri_env%orb_basis(jkind)%gto_basis_set

         IF (.NOT. ASSOCIATED(obasa)) CYCLE
         IF (.NOT. ASSOCIATED(obasb)) CYCLE

         IF (.NOT. ASSOCIATED(lri_ints_rho%lri_atom(iac)%lri_node)) THEN
            lri_ints_rho%lri_atom(iac)%natom = nlist
            ALLOCATE (lri_ints_rho%lri_atom(iac)%lri_node(nlist))
            DO i = 1, nlist
               NULLIFY (lri_ints_rho%lri_atom(iac)%lri_node(i)%lri_int_rho)
               lri_ints_rho%lri_atom(iac)%lri_node(i)%nnode = 0
            END DO
         END IF
         IF (.NOT. ASSOCIATED(lri_ints_rho%lri_atom(iac)%lri_node(ilist)%lri_int_rho)) THEN
            lri_ints_rho%lri_atom(iac)%lri_node(ilist)%nnode = nneighbor
            ALLOCATE (lri_ints_rho%lri_atom(iac)%lri_node(ilist)%lri_int_rho(nneighbor))
         END IF

         lriir => lri_ints_rho%lri_atom(iac)%lri_node(ilist)%lri_int_rho(jneighbor)

         nba = obasa%nsgf
         nbb = obasb%nsgf

         ALLOCATE (lriir%soaabb(nba, nba, nbb, nbb))
         lriir%soaabb = 0._dp
         lriir%dmax_aabb = 0._dp

      ENDDO

      CALL neighbor_list_iterator_release(nl_iterator)

   END SUBROUTINE allocate_lri_ints_rho

! **************************************************************************************************
!> \brief creates and initializes lri_rhos
!> \param lri_env ...
!> \param lri_rhos structure storing tvec and avec
!> \param nspin ...
!> \param nkind number of atom kinds
! **************************************************************************************************
   SUBROUTINE allocate_lri_rhos(lri_env, lri_rhos, nspin, nkind)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_list_p_type), DIMENSION(:), POINTER       :: lri_rhos
      INTEGER, INTENT(IN)                                :: nspin, nkind

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_lri_rhos', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iac, iatom, ikind, ilist, ispin, &
                                                            jatom, jkind, jneighbor, nfa, nfb, &
                                                            nlist, nn, nneighbor
      REAL(KIND=dp)                                      :: dab, rab(3)
      TYPE(lri_int_type), POINTER                        :: lrii
      TYPE(lri_list_type), POINTER                       :: lri_rho
      TYPE(lri_rhoab_type), POINTER                      :: lrho
      TYPE(neighbor_list_iterator_p_type), &
         DIMENSION(:), POINTER                           :: nl_iterator

      NULLIFY (lri_rho, lrho, lrii, nl_iterator)

      ALLOCATE (lri_rhos(nspin))

      DO ispin = 1, nspin

         ALLOCATE (lri_rhos(ispin)%lri_list)

         lri_rhos(ispin)%lri_list%nkind = nkind
         ALLOCATE (lri_rhos(ispin)%lri_list%lri_atom(nkind*nkind))

         DO i = 1, nkind*nkind
            NULLIFY (lri_rhos(ispin)%lri_list%lri_atom(i)%lri_node)
            lri_rhos(ispin)%lri_list%lri_atom(i)%natom = 0
         END DO

         lri_rho => lri_rhos(ispin)%lri_list

         CALL neighbor_list_iterator_create(nl_iterator, lri_env%soo_list)
         DO WHILE (neighbor_list_iterate(nl_iterator) == 0)
            CALL get_iterator_info(nl_iterator, ikind=ikind, jkind=jkind, &
                                   iatom=iatom, jatom=jatom, nlist=nlist, ilist=ilist, &
                                   nnode=nneighbor, inode=jneighbor, r=rab)

            iac = ikind+nkind*(jkind-1)
            dab = SQRT(SUM(rab*rab))

            IF (.NOT. ASSOCIATED(lri_env%lri_ints%lri_atom(iac)%lri_node)) CYCLE

            IF (.NOT. ASSOCIATED(lri_rho%lri_atom(iac)%lri_node)) THEN
               lri_rho%lri_atom(iac)%natom = nlist
               ALLOCATE (lri_rho%lri_atom(iac)%lri_node(nlist))
               DO i = 1, nlist
                  NULLIFY (lri_rho%lri_atom(iac)%lri_node(i)%lri_rhoab)
                  lri_rho%lri_atom(iac)%lri_node(i)%nnode = 0
               END DO
            END IF
            IF (.NOT. ASSOCIATED(lri_rho%lri_atom(iac)%lri_node(ilist)%lri_rhoab)) THEN
               lri_rho%lri_atom(iac)%lri_node(ilist)%nnode = nneighbor
               ALLOCATE (lri_rho%lri_atom(iac)%lri_node(ilist)%lri_rhoab(nneighbor))
            END IF

            lrho => lri_rho%lri_atom(iac)%lri_node(ilist)%lri_rhoab(jneighbor)
            lrii => lri_env%lri_ints%lri_atom(iac)%lri_node(ilist)%lri_int(jneighbor)

            lrho%nba = lrii%nba
            lrho%nbb = lrii%nbb
            lrho%nfa = lrii%nfa
            lrho%nfb = lrii%nfb

            nfa = lrho%nfa
            nfb = lrho%nfb
            nn = nfa+nfb

            IF (iatom == jatom .AND. dab < lri_env%delta) THEN
               ALLOCATE (lrho%avec(nfa))
               ALLOCATE (lrho%tvec(nfa))
            ELSE
               ALLOCATE (lrho%avec(nn))
               ALLOCATE (lrho%tvec(nn))
            ENDIF
            lrho%avec = 0._dp
            lrho%tvec = 0._dp

         ENDDO

         CALL neighbor_list_iterator_release(nl_iterator)

      ENDDO

   END SUBROUTINE allocate_lri_rhos

! **************************************************************************************************
!> \brief creates and initializes lri_coefs
!> \param lri_env ...
!> \param lri_density ...
!> \param atomic_kind_set ...
! **************************************************************************************************
   SUBROUTINE allocate_lri_coefs(lri_env, lri_density, atomic_kind_set)

      TYPE(lri_environment_type), POINTER                :: lri_env
      TYPE(lri_density_type), POINTER                    :: lri_density
      TYPE(atomic_kind_type), DIMENSION(:), POINTER      :: atomic_kind_set

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_lri_coefs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ikind, ispin, natom, nkind, nsgf, nspin
      TYPE(atomic_kind_type), POINTER                    :: atomic_kind
      TYPE(gto_basis_set_type), POINTER                  :: fbas
      TYPE(lri_spin_type), DIMENSION(:), POINTER         :: lri_coefs

      NULLIFY (atomic_kind, fbas, lri_coefs)
      nkind = SIZE(atomic_kind_set)
      nspin = lri_density%nspin

      ALLOCATE (lri_density%lri_coefs(nspin))
      lri_coefs => lri_density%lri_coefs

      DO ispin = 1, nspin
         ALLOCATE (lri_density%lri_coefs(ispin)%lri_kinds(nkind))
         DO ikind = 1, nkind
            NULLIFY (lri_coefs(ispin)%lri_kinds(ikind)%acoef)
            NULLIFY (lri_coefs(ispin)%lri_kinds(ikind)%v_int)
            NULLIFY (lri_coefs(ispin)%lri_kinds(ikind)%v_dadr)
            NULLIFY (lri_coefs(ispin)%lri_kinds(ikind)%v_dfdr)
            atomic_kind => atomic_kind_set(ikind)
            CALL get_atomic_kind(atomic_kind=atomic_kind, natom=natom)
            fbas => lri_env%ri_basis(ikind)%gto_basis_set
            nsgf = fbas%nsgf
            ALLOCATE (lri_coefs(ispin)%lri_kinds(ikind)%acoef(natom, nsgf))
            lri_coefs(ispin)%lri_kinds(ikind)%acoef = 0._dp
            ALLOCATE (lri_coefs(ispin)%lri_kinds(ikind)%v_int(natom, nsgf))
            lri_coefs(ispin)%lri_kinds(ikind)%v_int = 0._dp
            ALLOCATE (lri_coefs(ispin)%lri_kinds(ikind)%v_dadr(natom, 3))
            lri_coefs(ispin)%lri_kinds(ikind)%v_dadr = 0._dp
            ALLOCATE (lri_coefs(ispin)%lri_kinds(ikind)%v_dfdr(natom, 3))
            lri_coefs(ispin)%lri_kinds(ikind)%v_dfdr = 0._dp
         END DO
      ENDDO

   END SUBROUTINE allocate_lri_coefs

! **************************************************************************************************
!> \brief creates and initializes lri_force
!> \param lri_force ...
!> \param nfa and nfb number of fit functions on a/b
!> \param nfb ...
! **************************************************************************************************
   SUBROUTINE allocate_lri_force_components(lri_force, nfa, nfb)

      TYPE(lri_force_type), POINTER                      :: lri_force
      INTEGER, INTENT(IN)                                :: nfa, nfb

      CHARACTER(LEN=*), PARAMETER :: routineN = 'allocate_lri_force_components', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: nn

      nn = nfa+nfb

      IF (.NOT. ASSOCIATED(lri_force)) THEN
         ALLOCATE (lri_force)

         ALLOCATE (lri_force%ds(nn, nn, 3))
         lri_force%ds = 0._dp
         ALLOCATE (lri_force%st(nn))
         lri_force%st = 0._dp
         ALLOCATE (lri_force%dsst(nn, 3))
         lri_force%dsst = 0._dp
         ALLOCATE (lri_force%sdsst(nn, 3))
         lri_force%sdsst = 0._dp
         ALLOCATE (lri_force%dssn(nn, 3))
         lri_force%dssn = 0._dp
         ALLOCATE (lri_force%sdssn(nn, 3))
         lri_force%sdssn = 0._dp
         ALLOCATE (lri_force%sdt(nn, 3))
         lri_force%sdt = 0._dp
         ALLOCATE (lri_force%davec(nn, 3))
         lri_force%davec = 0._dp
         ALLOCATE (lri_force%dtvec(nn, 3))
         lri_force%dtvec = 0._dp
      ENDIF

   END SUBROUTINE allocate_lri_force_components

! **************************************************************************************************
!> \brief deallocates one-center overlap integrals, integral of ri basis
!>        and scon matrices
!> \param lri_env ...
! **************************************************************************************************
   SUBROUTINE deallocate_bas_properties(lri_env)

      TYPE(lri_environment_type), POINTER                :: lri_env

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_bas_properties', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(lri_env%bas_prop)) THEN
         DO i = 1, SIZE(lri_env%bas_prop)
            DEALLOCATE (lri_env%bas_prop(i)%int_fbas)
            DEALLOCATE (lri_env%bas_prop(i)%ri_ovlp)
            DEALLOCATE (lri_env%bas_prop(i)%orb_ovlp)
            DEALLOCATE (lri_env%bas_prop(i)%scon_ri)
            DEALLOCATE (lri_env%bas_prop(i)%scon_orb)
         END DO
         DEALLOCATE (lri_env%bas_prop)
      END IF

   END SUBROUTINE deallocate_bas_properties

! **************************************************************************************************
!> \brief deallocates the given lri_ints
!> \param lri_ints ...
! **************************************************************************************************
   SUBROUTINE deallocate_lri_ints(lri_ints)

      TYPE(lri_list_type), POINTER                       :: lri_ints

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_lri_ints', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, ijkind, inode, natom, nkind, nnode

      IF(.NOT.(ASSOCIATED(lri_ints)))CALL cp__a("lri_environment_types.F",859)
      nkind = lri_ints%nkind

      IF (nkind > 0) THEN
         DO ijkind = 1, SIZE(lri_ints%lri_atom)
            natom = lri_ints%lri_atom(ijkind)%natom
            IF (natom > 0) THEN
               DO iatom = 1, natom
                  nnode = lri_ints%lri_atom(ijkind)%lri_node(iatom)%nnode
                  IF (nnode > 0) THEN
                     IF (ASSOCIATED(lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int)) THEN
                        DO inode = 1, nnode
                           DEALLOCATE (lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%abaint, &
                                       lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%abbint, &
                                       lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%dabdaint, &
                                       lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%dabbint, &
                                       lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%soo, &
                                       lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%dsoo, &
                                       lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%sab, &
                                       lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%dsab, &
                                       lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%dacoef, &
                                       lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%sinv, &
                                       lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%n, &
                                       lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int(inode)%sn)
                        END DO
                        DEALLOCATE (lri_ints%lri_atom(ijkind)%lri_node(iatom)%lri_int)
                     END IF
                  END IF
               END DO
               DEALLOCATE (lri_ints%lri_atom(ijkind)%lri_node)
            END IF
         END DO
         DEALLOCATE (lri_ints%lri_atom)
      END IF
      DEALLOCATE (lri_ints)

   END SUBROUTINE deallocate_lri_ints

! **************************************************************************************************
!> \brief deallocates the given lri_ints_rho
!> \param lri_ints_rho ...
! **************************************************************************************************
   SUBROUTINE deallocate_lri_ints_rho(lri_ints_rho)

      TYPE(lri_list_type), POINTER                       :: lri_ints_rho

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_lri_ints_rho', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iatom, ijkind, inode, natom, nkind, nnode

      IF(.NOT.(ASSOCIATED(lri_ints_rho)))CALL cp__a("lri_environment_types.F",910)
      nkind = lri_ints_rho%nkind

      IF (nkind > 0) THEN
         DO ijkind = 1, SIZE(lri_ints_rho%lri_atom)
            natom = lri_ints_rho%lri_atom(ijkind)%natom
            IF (natom > 0) THEN
               DO iatom = 1, natom
                  nnode = lri_ints_rho%lri_atom(ijkind)%lri_node(iatom)%nnode
                  IF (nnode > 0) THEN
                     IF (ASSOCIATED(lri_ints_rho%lri_atom(ijkind)%lri_node(iatom)%lri_int_rho)) THEN
                        DO inode = 1, nnode
                           DEALLOCATE (lri_ints_rho%lri_atom(ijkind)%lri_node(iatom)%lri_int_rho(inode)%soaabb)
                        END DO
                        DEALLOCATE (lri_ints_rho%lri_atom(ijkind)%lri_node(iatom)%lri_int_rho)
                     END IF
                  END IF
               END DO
               DEALLOCATE (lri_ints_rho%lri_atom(ijkind)%lri_node)
            END IF
         END DO
         DEALLOCATE (lri_ints_rho%lri_atom)
      END IF
      DEALLOCATE (lri_ints_rho)

   END SUBROUTINE deallocate_lri_ints_rho

! **************************************************************************************************
!> \brief deallocates the given lri_rhos
!> \param lri_rhos ...
! **************************************************************************************************
   SUBROUTINE deallocate_lri_rhos(lri_rhos)

      TYPE(lri_list_p_type), DIMENSION(:), POINTER       :: lri_rhos

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_lri_rhos', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iatom, ijkind, inode, natom, nkind, &
                                                            nnode
      TYPE(lri_list_type), POINTER                       :: lri_rho

      NULLIFY (lri_rho)

      IF (ASSOCIATED(lri_rhos)) THEN

         DO i = 1, SIZE(lri_rhos)

            lri_rho => lri_rhos(i)%lri_list
            IF(.NOT.(ASSOCIATED(lri_rho)))CALL cp__a("lri_environment_types.F",959)
            nkind = lri_rho%nkind

            IF (nkind > 0) THEN
               DO ijkind = 1, SIZE(lri_rho%lri_atom)
                  natom = lri_rho%lri_atom(ijkind)%natom
                  IF (natom > 0) THEN
                     DO iatom = 1, natom
                        nnode = lri_rho%lri_atom(ijkind)%lri_node(iatom)%nnode
                        IF (nnode > 0) THEN
                           IF (ASSOCIATED(lri_rho%lri_atom(ijkind)%lri_node(iatom)%lri_rhoab)) THEN
                              DO inode = 1, nnode
                                 DEALLOCATE (lri_rho%lri_atom(ijkind)%lri_node(iatom)%lri_rhoab(inode)%avec, &
                                             lri_rho%lri_atom(ijkind)%lri_node(iatom)%lri_rhoab(inode)%tvec)
                              END DO
                              DEALLOCATE (lri_rho%lri_atom(ijkind)%lri_node(iatom)%lri_rhoab)
                           END IF
                        END IF
                     END DO
                     DEALLOCATE (lri_rho%lri_atom(ijkind)%lri_node)
                  END IF
               END DO
               DEALLOCATE (lri_rho%lri_atom)
            END IF
            DEALLOCATE (lri_rho)
         END DO

         DEALLOCATE (lri_rhos)

      END IF

   END SUBROUTINE deallocate_lri_rhos

! **************************************************************************************************
!> \brief releases the given lri_coefs
!> \param lri_coefs the integral storage environment that is released
! **************************************************************************************************
   SUBROUTINE deallocate_lri_coefs(lri_coefs)
      TYPE(lri_spin_type), DIMENSION(:), POINTER         :: lri_coefs

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_lri_coefs', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, j

      IF (ASSOCIATED(lri_coefs)) THEN
         DO i = 1, SIZE(lri_coefs)
            DO j = 1, SIZE(lri_coefs(i)%lri_kinds)
               IF (ASSOCIATED(lri_coefs(i)%lri_kinds(j)%acoef)) THEN
                  DEALLOCATE (lri_coefs(i)%lri_kinds(j)%acoef)
               ENDIF
               IF (ASSOCIATED(lri_coefs(i)%lri_kinds(j)%v_int)) THEN
                  DEALLOCATE (lri_coefs(i)%lri_kinds(j)%v_int)
               ENDIF
               IF (ASSOCIATED(lri_coefs(i)%lri_kinds(j)%v_dadr)) THEN
                  DEALLOCATE (lri_coefs(i)%lri_kinds(j)%v_dadr)
               ENDIF
               IF (ASSOCIATED(lri_coefs(i)%lri_kinds(j)%v_dfdr)) THEN
                  DEALLOCATE (lri_coefs(i)%lri_kinds(j)%v_dfdr)
               ENDIF
            ENDDO
            DEALLOCATE (lri_coefs(i)%lri_kinds)
         END DO
         DEALLOCATE (lri_coefs)
      END IF
      NULLIFY (lri_coefs)

   END SUBROUTINE deallocate_lri_coefs

! **************************************************************************************************
!> \brief releases the given lri_force_type
!> \param lri_force the integral storage environment that is released
! **************************************************************************************************
   SUBROUTINE deallocate_lri_force_components(lri_force)

      TYPE(lri_force_type), POINTER                      :: lri_force

      CHARACTER(len=*), PARAMETER :: routineN = 'deallocate_lri_force_components', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(lri_force)) THEN

         IF (ASSOCIATED(lri_force%st)) THEN
            DEALLOCATE (lri_force%st)
         ENDIF
         IF (ASSOCIATED(lri_force%dssn)) THEN
            DEALLOCATE (lri_force%dssn)
         ENDIF
         IF (ASSOCIATED(lri_force%sdssn)) THEN
            DEALLOCATE (lri_force%sdssn)
         ENDIF
         IF (ASSOCIATED(lri_force%dsst)) THEN
            DEALLOCATE (lri_force%dsst)
         ENDIF
         IF (ASSOCIATED(lri_force%sdsst)) THEN
            DEALLOCATE (lri_force%sdsst)
         ENDIF
         IF (ASSOCIATED(lri_force%sdt)) THEN
            DEALLOCATE (lri_force%sdt)
         ENDIF
         IF (ASSOCIATED(lri_force%dtvec)) THEN
            DEALLOCATE (lri_force%dtvec)
         ENDIF
         IF (ASSOCIATED(lri_force%davec)) THEN
            DEALLOCATE (lri_force%davec)
         ENDIF
         IF (ASSOCIATED(lri_force%ds)) THEN
            DEALLOCATE (lri_force%ds)
         ENDIF

         DEALLOCATE (lri_force)

         NULLIFY (lri_force)
      ENDIF

   END SUBROUTINE deallocate_lri_force_components

END MODULE lri_environment_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types and set/get functions for auxiliary denisty matrix methods
!> \par History
!>      05.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
MODULE admm_types
   USE bibliography,                    ONLY: Guidon2010,&
                                              cite_reference
   USE cp_control_types,                ONLY: admm_control_type
   USE cp_fm_struct,                    ONLY: cp_fm_struct_create,&
                                              cp_fm_struct_release,&
                                              cp_fm_struct_type
   USE cp_fm_types,                     ONLY: cp_fm_create,&
                                              cp_fm_p_type,&
                                              cp_fm_release,&
                                              cp_fm_type
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE input_constants,                 ONLY: do_admm_blocked_projection,&
                                              do_admm_blocking_purify_full,&
                                              do_admm_charge_constrained_projection
   USE input_section_types,             ONLY: section_vals_release,&
                                              section_vals_type
   USE kinds,                           ONLY: dp
   USE qs_mo_types,                     ONLY: get_mo_set,&
                                              mo_set_p_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE
   PUBLIC :: admm_env_create, admm_env_release, admm_type

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'admm_types'

   TYPE eigvals_type
      REAL(dp), DIMENSION(:), POINTER          :: DATA
   END TYPE

   TYPE eigvals_p_type
      TYPE(eigvals_type), POINTER              :: eigvals
   END TYPE

! **************************************************************************************************
!> \brief stores some data used in wavefunction fitting
!> \param S overlap matrix for auxiliary fit basis set
!> \param P overlap matrix for mixed aux_fit/orb basis set
!> \param A contains inv(S)*P
!> \param B contains transpose(P)*inv(S)*P = transpose(P)*A
!> \param lambda contains transpose(mo_coeff_aux_fit)*B*mo_coeff_aux_fit
!> \param lambda_inv_sqrt contains inv(SQRT(lambda))
!> \param R contains eigenvectors of lambda
!> \param work_aux_aux temporary matrix
!> \param work_orb_nmo temporary matrix
!> \param work_nmo_nmo1 temporary matrix
!> \param work_nmo_nmo2 temporary matrix
!> \param work_aux_nmo temporary matrix
!> \param H contains KS_matrix * mo_coeff for auxiliary basis set
!> \param K contains KS matrix for auxiliary basis set
!> \param M contains matrix holding the 2nd order residues
!> \param nao_orb number of atomic orbitals in orb basis set
!> \param nao_aux_fit number of atomic orbitals in aux basis set
!> \param nmo number of molecular orbitals per spin
!> \param eigvals_lamda eigenvalues of lambda matrix
!> \param gsi contains ratio N_dens_m/N_aux_dens_m
!> \par History
!>      05.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   TYPE admm_type
      TYPE(cp_fm_type), POINTER                :: S_inv => Null(), &
                                                  S => Null(), &
                                                  Q => Null(), &
                                                  A => Null(), &
                                                  B => Null(), &
                                                  work_orb_orb => Null(), &
                                                  work_orb_orb2 => Null(), &
                                                  work_orb_orb3 => Null(), &
                                                  work_aux_orb => Null(), &
                                                  work_aux_orb2 => Null(), &
                                                  work_aux_orb3 => Null(), &
                                                  work_aux_aux => Null(), &
                                                  work_aux_aux2 => Null(), &
                                                  work_aux_aux3 => Null(), &
                                                  work_aux_aux4 => Null(), &
                                                  work_aux_aux5 => Null()

      TYPE(cp_fm_p_type), DIMENSION(:), &
         POINTER                                :: lambda => Null(), &
                                                   lambda_inv => Null(), &
                                                   lambda_inv_sqrt => Null(), &
                                                   R => Null(), &
                                                   R_purify => Null(), &
                                                   work_orb_nmo => Null(), &
                                                   work_nmo_nmo1 => Null(), &
                                                   R_schur_R_t => Null(), &
                                                   work_nmo_nmo2 => Null(), &
                                                   work_aux_nmo => Null(), &
                                                   work_aux_nmo2 => Null(), &
                                                   H => Null(), &
                                                   H_corr => Null(), &
                                                   mo_derivs_tmp => Null(), &
                                                   K => Null(), &
                                                   M => Null(), &
                                                   M_purify => Null(), &
                                                   P_to_be_purified => Null(), &
                                                   lambda_inv2 => Null(), &
                                                   C_hat => Null(), &
                                                   P_tilde => Null(), &
                                                   ks_to_be_merged => Null()
      TYPE(eigvals_p_type), DIMENSION(:), &
         POINTER                                :: eigvals_lambda => Null(), &
                                                   eigvals_P_to_be_purified => Null()
      TYPE(section_vals_type), POINTER         :: xc_section_primary => Null(), &
                                                  xc_section_aux => Null()
      REAL(KIND=dp)                            :: gsi(3), &
                                                  lambda_merlot(2), &
                                                  n_large_basis(3)
      INTEGER                                  :: nao_orb, nao_aux_fit, nmo(2)
      INTEGER                                  :: purification_method
      LOGICAL                                  :: charge_constrain = .FALSE.
      INTEGER                                  :: scaling_model, &
                                                  aux_exch_func
      LOGICAL                                  :: block_dm = .FALSE.
      LOGICAL                                  :: block_fit = .FALSE.
      INTEGER, DIMENSION(:, :), POINTER         :: block_map => Null()
   END TYPE

CONTAINS

! **************************************************************************************************
!> \brief creates ADDE environment, initializes the basic types
!>
!> \param admm_env The ADMM env
!> \param admm_control ...
!> \param mos the MO's of the orbital basis set
!> \param mos_aux_fit the MO's of the auxiliary fitting basis set
!> \param para_env The parallel env
!> \param natoms ...
!> \par History
!>      05.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE admm_env_create(admm_env, admm_control, mos, mos_aux_fit, para_env, natoms)
      TYPE(admm_type), POINTER                           :: admm_env
      TYPE(admm_control_type), POINTER                   :: admm_control
      TYPE(mo_set_p_type), DIMENSION(:), POINTER         :: mos, mos_aux_fit
      TYPE(cp_para_env_type), POINTER                    :: para_env
      INTEGER, INTENT(IN)                                :: natoms

      CHARACTER(LEN=*), PARAMETER :: routineN = 'admm_env_create', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: i, iatom, iblock, ispin, j, jatom, &
                                                            nao_aux_fit, nao_orb, nmo, nspins
      TYPE(cp_fm_struct_type), POINTER :: fm_struct_aux_aux, fm_struct_aux_nmo, fm_struct_aux_orb, &
         fm_struct_nmo_nmo, fm_struct_orb_aux, fm_struct_orb_nmo, fm_struct_orb_orb
      TYPE(cp_fm_type), POINTER                          :: mo_coeff

      CALL cite_reference(Guidon2010)

      ALLOCATE (admm_env)

      nspins = SIZE(mos)
      CALL get_mo_set(mos(1)%mo_set, mo_coeff=mo_coeff, nmo=nmo, nao=nao_orb)
      CALL get_mo_set(mos_aux_fit(1)%mo_set, nao=nao_aux_fit)
      admm_env%nmo = 0
      admm_env%nao_aux_fit = nao_aux_fit
      admm_env%nao_orb = nao_orb
      CALL cp_fm_struct_create(fm_struct_aux_aux, &
                               context=mo_coeff%matrix_struct%context, &
                               nrow_global=nao_aux_fit, &
                               ncol_global=nao_aux_fit, &
                               para_env=para_env)
      CALL cp_fm_struct_create(fm_struct_aux_orb, &
                               context=mo_coeff%matrix_struct%context, &
                               nrow_global=nao_aux_fit, &
                               ncol_global=nao_orb, &
                               para_env=para_env)
      CALL cp_fm_struct_create(fm_struct_orb_aux, &
                               context=mo_coeff%matrix_struct%context, &
                               nrow_global=nao_orb, &
                               ncol_global=nao_aux_fit, &
                               para_env=para_env)
      CALL cp_fm_struct_create(fm_struct_orb_orb, &
                               context=mo_coeff%matrix_struct%context, &
                               nrow_global=nao_orb, &
                               ncol_global=nao_orb, &
                               para_env=para_env)

      CALL cp_fm_create(admm_env%S, fm_struct_aux_aux, name="aux_fit_overlap")
      CALL cp_fm_create(admm_env%S_inv, fm_struct_aux_aux, name="aux_fit_overlap_inv")
      CALL cp_fm_create(admm_env%Q, fm_struct_aux_orb, name="mixed_overlap")
      CALL cp_fm_create(admm_env%A, fm_struct_aux_orb, name="work_A")
      CALL cp_fm_create(admm_env%B, fm_struct_orb_orb, name="work_B")
      CALL cp_fm_create(admm_env%work_orb_orb, fm_struct_orb_orb, name="work_orb_orb")
      CALL cp_fm_create(admm_env%work_orb_orb2, fm_struct_orb_orb, name="work_orb_orb")
      CALL cp_fm_create(admm_env%work_orb_orb3, fm_struct_orb_orb, name="work_orb_orb3")
      CALL cp_fm_create(admm_env%work_aux_orb, fm_struct_aux_orb, name="work_aux_orb")
      CALL cp_fm_create(admm_env%work_aux_orb2, fm_struct_aux_orb, name="work_aux_orb2")
      CALL cp_fm_create(admm_env%work_aux_orb3, fm_struct_aux_orb, name="work_aux_orb3")
      CALL cp_fm_create(admm_env%work_aux_aux, fm_struct_aux_aux, name="work_aux_aux")
      CALL cp_fm_create(admm_env%work_aux_aux2, fm_struct_aux_aux, name="work_aux_aux2")
      CALL cp_fm_create(admm_env%work_aux_aux3, fm_struct_aux_aux, name="work_aux_aux3")
      CALL cp_fm_create(admm_env%work_aux_aux4, fm_struct_aux_aux, name="work_aux_aux4")
      CALL cp_fm_create(admm_env%work_aux_aux5, fm_struct_aux_aux, name="work_aux_aux5")

      ALLOCATE (admm_env%lambda_inv(nspins))
      ALLOCATE (admm_env%lambda(nspins))
      ALLOCATE (admm_env%lambda_inv_sqrt(nspins))
      ALLOCATE (admm_env%R(nspins))
      ALLOCATE (admm_env%R_purify(nspins))
      ALLOCATE (admm_env%work_orb_nmo(nspins))
      ALLOCATE (admm_env%work_nmo_nmo1(nspins))
      ALLOCATE (admm_env%R_schur_R_t(nspins))
      ALLOCATE (admm_env%work_nmo_nmo2(nspins))
      ALLOCATE (admm_env%eigvals_lambda(nspins))
      ALLOCATE (admm_env%eigvals_P_to_be_purified(nspins))
      ALLOCATE (admm_env%H(nspins))
      ALLOCATE (admm_env%K(nspins))
      ALLOCATE (admm_env%M(nspins))
      ALLOCATE (admm_env%M_purify(nspins))
      ALLOCATE (admm_env%P_to_be_purified(nspins))
      ALLOCATE (admm_env%work_aux_nmo(nspins))
      ALLOCATE (admm_env%work_aux_nmo2(nspins))
      ALLOCATE (admm_env%mo_derivs_tmp(nspins))
      ALLOCATE (admm_env%H_corr(nspins))
      ALLOCATE (admm_env%ks_to_be_merged(nspins))
      ALLOCATE (admm_env%lambda_inv2(nspins))
      ALLOCATE (admm_env%C_hat(nspins))
      ALLOCATE (admm_env%P_tilde(nspins))

      DO ispin = 1, nspins
         CALL get_mo_set(mos(ispin)%mo_set, mo_coeff=mo_coeff, nmo=nmo)
         admm_env%nmo(ispin) = nmo
         CALL cp_fm_struct_create(fm_struct_aux_nmo, &
                                  context=mo_coeff%matrix_struct%context, &
                                  nrow_global=nao_aux_fit, &
                                  ncol_global=nmo, &
                                  para_env=para_env)
         CALL cp_fm_struct_create(fm_struct_orb_nmo, &
                                  context=mo_coeff%matrix_struct%context, &
                                  nrow_global=nao_orb, &
                                  ncol_global=nmo, &
                                  para_env=para_env)
         CALL cp_fm_struct_create(fm_struct_nmo_nmo, &
                                  context=mo_coeff%matrix_struct%context, &
                                  nrow_global=nmo, &
                                  ncol_global=nmo, &
                                  para_env=para_env)

         CALL cp_fm_create(admm_env%work_orb_nmo(ispin)%matrix, fm_struct_orb_nmo, name="work_orb_nmo")
         CALL cp_fm_create(admm_env%work_nmo_nmo1(ispin)%matrix, fm_struct_nmo_nmo, name="work_nmo_nmo1")
         CALL cp_fm_create(admm_env%R_schur_R_t(ispin)%matrix, fm_struct_nmo_nmo, name="R_schur_R_t")
         CALL cp_fm_create(admm_env%work_nmo_nmo2(ispin)%matrix, fm_struct_nmo_nmo, name="work_nmo_nmo2")
         CALL cp_fm_create(admm_env%lambda(ispin)%matrix, fm_struct_nmo_nmo, name="lambda")
         CALL cp_fm_create(admm_env%lambda_inv(ispin)%matrix, fm_struct_nmo_nmo, name="lambda_inv")
         CALL cp_fm_create(admm_env%lambda_inv_sqrt(ispin)%matrix, fm_struct_nmo_nmo, name="lambda_inv_sqrt")
         CALL cp_fm_create(admm_env%R(ispin)%matrix, fm_struct_nmo_nmo, name="R")
         CALL cp_fm_create(admm_env%R_purify(ispin)%matrix, fm_struct_aux_aux, name="R_purify")
         CALL cp_fm_create(admm_env%K(ispin)%matrix, fm_struct_aux_aux, name="K")
         CALL cp_fm_create(admm_env%H(ispin)%matrix, fm_struct_aux_nmo, name="H")
         CALL cp_fm_create(admm_env%H_corr(ispin)%matrix, fm_struct_orb_orb, name="H_corr")
         CALL cp_fm_create(admm_env%M(ispin)%matrix, fm_struct_nmo_nmo, name="M")
         CALL cp_fm_create(admm_env%M_purify(ispin)%matrix, fm_struct_aux_aux, name="M aux")
         CALL cp_fm_create(admm_env%P_to_be_purified(ispin)%matrix, fm_struct_aux_aux, name="P_to_be_purified")
         CALL cp_fm_create(admm_env%work_aux_nmo(ispin)%matrix, fm_struct_aux_nmo, name="work_aux_nmo")
         CALL cp_fm_create(admm_env%work_aux_nmo2(ispin)%matrix, fm_struct_aux_nmo, name="work_aux_nmo2")
         CALL cp_fm_create(admm_env%mo_derivs_tmp(ispin)%matrix, fm_struct_orb_nmo, name="mo_derivs_tmp")

         CALL cp_fm_create(admm_env%lambda_inv2(ispin)%matrix, fm_struct_nmo_nmo, name="lambda_inv2")
         CALL cp_fm_create(admm_env%C_hat(ispin)%matrix, fm_struct_aux_nmo, name="C_hat")
         CALL cp_fm_create(admm_env%P_tilde(ispin)%matrix, fm_struct_aux_aux, name="P_tilde")

         CALL cp_fm_create(admm_env%ks_to_be_merged(ispin)%matrix, fm_struct_orb_orb, name="KS_to_be_merged ")

         ALLOCATE (admm_env%eigvals_lambda(ispin)%eigvals)
         ALLOCATE (admm_env%eigvals_P_to_be_purified(ispin)%eigvals)
         ALLOCATE (admm_env%eigvals_lambda(ispin)%eigvals%data(nmo))
         ALLOCATE (admm_env%eigvals_P_to_be_purified(ispin)%eigvals%data(nao_aux_fit))
         admm_env%eigvals_lambda(ispin)%eigvals%data = 0.0_dp
         admm_env%eigvals_P_to_be_purified(ispin)%eigvals%data = 0.0_dp
         CALL cp_fm_struct_release(fm_struct_aux_nmo)
         CALL cp_fm_struct_release(fm_struct_orb_nmo)
         CALL cp_fm_struct_release(fm_struct_nmo_nmo)
      END DO

      CALL cp_fm_struct_release(fm_struct_aux_aux)
      CALL cp_fm_struct_release(fm_struct_aux_orb)
      CALL cp_fm_struct_release(fm_struct_orb_aux)
      CALL cp_fm_struct_release(fm_struct_orb_orb)

      ! copy settings from admm_control
      admm_env%purification_method = admm_control%purification_method
      admm_env%scaling_model = admm_control%scaling_model
      admm_env%aux_exch_func = admm_control%aux_exch_func
      admm_env%charge_constrain = (admm_control%method == do_admm_charge_constrained_projection)
      admm_env%block_dm = ((admm_control%method == do_admm_blocking_purify_full) .OR. &
                           (admm_control%method == do_admm_blocked_projection))
      admm_env%block_fit = admm_control%method == do_admm_blocked_projection

      IF ((admm_control%method == do_admm_blocking_purify_full) .OR. &
          (admm_control%method == do_admm_blocked_projection)) THEN
         ! create block map
         ALLOCATE (admm_env%block_map(natoms, natoms))
         admm_env%block_map(:, :) = 0
         DO iblock = 1, SIZE(admm_control%blocks)
            DO i = 1, SIZE(admm_control%blocks(iblock)%list)
               iatom = admm_control%blocks(iblock)%list(i)
               DO j = 1, SIZE(admm_control%blocks(iblock)%list)
                  jatom = admm_control%blocks(iblock)%list(j)
                  admm_env%block_map(iatom, jatom) = 1
               END DO
            END DO
         END DO
      ENDIF
   END SUBROUTINE admm_env_create

! **************************************************************************************************
!> \brief releases the ADMM environment, cleans up all types
!>
!> \param admm_env The ADMM env
!> \par History
!>      05.2008 created [Manuel Guidon]
!> \author Manuel Guidon
! **************************************************************************************************
   SUBROUTINE admm_env_release(admm_env)
      TYPE(admm_type), POINTER                           :: admm_env

      CHARACTER(LEN=*), PARAMETER :: routineN = 'admm_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ispin

      CALL cp_fm_release(admm_env%S)
      CALL cp_fm_release(admm_env%S_inv)
      CALL cp_fm_release(admm_env%Q)
      CALL cp_fm_release(admm_env%A)
      CALL cp_fm_release(admm_env%B)
      CALL cp_fm_release(admm_env%work_orb_orb)
      CALL cp_fm_release(admm_env%work_orb_orb2)
      CALL cp_fm_release(admm_env%work_orb_orb3)
      CALL cp_fm_release(admm_env%work_aux_aux)
      CALL cp_fm_release(admm_env%work_aux_aux2)
      CALL cp_fm_release(admm_env%work_aux_aux3)
      CALL cp_fm_release(admm_env%work_aux_aux4)
      CALL cp_fm_release(admm_env%work_aux_aux5)
      CALL cp_fm_release(admm_env%work_aux_orb)
      CALL cp_fm_release(admm_env%work_aux_orb2)
      CALL cp_fm_release(admm_env%work_aux_orb3)
      DO ispin = 1, SIZE(admm_env%lambda)
         CALL cp_fm_release(admm_env%lambda(ispin)%matrix)
         CALL cp_fm_release(admm_env%lambda_inv(ispin)%matrix)
         CALL cp_fm_release(admm_env%lambda_inv_sqrt(ispin)%matrix)
         CALL cp_fm_release(admm_env%lambda_inv2(ispin)%matrix)
         CALL cp_fm_release(admm_env%C_hat(ispin)%matrix)
         CALL cp_fm_release(admm_env%P_tilde(ispin)%matrix)
         CALL cp_fm_release(admm_env%R(ispin)%matrix)
         CALL cp_fm_release(admm_env%R_purify(ispin)%matrix)
         CALL cp_fm_release(admm_env%H(ispin)%matrix)
         CALL cp_fm_release(admm_env%H_corr(ispin)%matrix)
         CALL cp_fm_release(admm_env%K(ispin)%matrix)
         CALL cp_fm_release(admm_env%M(ispin)%matrix)
         CALL cp_fm_release(admm_env%M_purify(ispin)%matrix)
         CALL cp_fm_release(admm_env%P_to_be_purified(ispin)%matrix)
         CALL cp_fm_release(admm_env%work_orb_nmo(ispin)%matrix)
         CALL cp_fm_release(admm_env%work_nmo_nmo1(ispin)%matrix)
         CALL cp_fm_release(admm_env%R_schur_R_t(ispin)%matrix)
         CALL cp_fm_release(admm_env%work_nmo_nmo2(ispin)%matrix)
         CALL cp_fm_release(admm_env%work_aux_nmo(ispin)%matrix)
         CALL cp_fm_release(admm_env%work_aux_nmo2(ispin)%matrix)
         CALL cp_fm_release(admm_env%mo_derivs_tmp(ispin)%matrix)
         CALL cp_fm_release(admm_env%ks_to_be_merged(ispin)%matrix)
         CALL cp_fm_release(admm_env%lambda_inv2(ispin)%matrix)
         DEALLOCATE (admm_env%eigvals_lambda(ispin)%eigvals%data)
         DEALLOCATE (admm_env%eigvals_P_to_be_purified(ispin)%eigvals%data)
         DEALLOCATE (admm_env%eigvals_lambda(ispin)%eigvals)
         DEALLOCATE (admm_env%eigvals_P_to_be_purified(ispin)%eigvals)
      END DO
      DEALLOCATE (admm_env%eigvals_lambda)
      DEALLOCATE (admm_env%eigvals_P_to_be_purified)
      DEALLOCATE (admm_env%lambda)
      DEALLOCATE (admm_env%lambda_inv)
      DEALLOCATE (admm_env%lambda_inv_sqrt)
      DEALLOCATE (admm_env%R)
      DEALLOCATE (admm_env%R_purify)
      DEALLOCATE (admm_env%M)
      DEALLOCATE (admm_env%M_purify)
      DEALLOCATE (admm_env%P_to_be_purified)
      DEALLOCATE (admm_env%H)
      DEALLOCATE (admm_env%H_corr)
      DEALLOCATE (admm_env%K)
      DEALLOCATE (admm_env%work_orb_nmo)
      DEALLOCATE (admm_env%work_nmo_nmo1)
      DEALLOCATE (admm_env%R_schur_R_t)
      DEALLOCATE (admm_env%work_nmo_nmo2)
      DEALLOCATE (admm_env%work_aux_nmo)
      DEALLOCATE (admm_env%work_aux_nmo2)
      DEALLOCATE (admm_env%mo_derivs_tmp)
      DEALLOCATE (admm_env%ks_to_be_merged)
      DEALLOCATE (admm_env%lambda_inv2)
      DEALLOCATE (admm_env%C_hat)
      DEALLOCATE (admm_env%P_tilde)

      IF (ASSOCIATED(admm_env%block_map)) &
         DEALLOCATE (admm_env%block_map)

      IF (ASSOCIATED(admm_env%xc_section_primary)) &
         CALL section_vals_release(admm_env%xc_section_primary)
      IF (ASSOCIATED(admm_env%xc_section_aux)) &
         CALL section_vals_release(admm_env%xc_section_aux)

      DEALLOCATE (admm_env)

   END SUBROUTINE admm_env_release

END MODULE admm_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief represent the global information of a run: the input file,
!>      parallel environment (and partially output).
!>      Use sparingly, try not to send it too deep in your structures.
!> \par History
!>      - print keys, basis_set_file name and potential_file_name added to the
!>        global type (27.02.2001, MK)
!>      - JGH (28.11.2001) : added pp_library_path to type
!>      - Merged with MODULE print_keys (17.01.2002, MK)
!>      -  reference counting, create (08.2004, fawzi)
!>      - new (parallel) random number generator (11.03.06,MK)
!> \author JGH,MK,fawzi
! **************************************************************************************************
MODULE global_types

   USE cp_blacs_env,                    ONLY: BLACS_GRID_SQUARE
   USE kinds,                           ONLY: default_path_length,&
                                              default_string_length,&
                                              dp
   USE machine,                         ONLY: m_walltime
   USE parallel_rng_types,              ONLY: delete_rng_stream,&
                                              rng_stream_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   ! Global parameters

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'global_types'

   INTEGER, PARAMETER :: SILENT = 0, &
                         LOW = 1, &
                         MEDIUM = 2, &
                         HIGH = 3, &
                         DEBUG = 4

   INTEGER, SAVE :: last_globenv_id = 0

   ! Public parameters

   ! Public data types
   PUBLIC :: global_environment_type

   ! Public subroutines
   PUBLIC :: globenv_create, &
             globenv_retain, &
             globenv_release

! **************************************************************************************************
!> \brief contains the initally parsed file and the initial parallel environment
!> \param id_nr identification number (unique)
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \param handle handle with the total time of the computation
!>
!>      Personally I think that all the other attributes should go away
!>      (and maybe add  a logger)[fawzi]
!> \note
!>      This is not but really should be passed as pointer and use reference
!>      counting. Use it accordingly wherever possible.
! **************************************************************************************************
   TYPE global_environment_type
      INTEGER :: id_nr, ref_count
      TYPE(rng_stream_type), POINTER          :: gaussian_rng_stream
      CHARACTER(LEN=default_string_length)    :: diag_library
      CHARACTER(LEN=default_string_length)    :: default_fft_library
      CHARACTER(LEN=default_path_length)      :: fftw_wisdom_file_name

      INTEGER :: fft_pool_scratch_limit !! limit used for fft scratches
      INTEGER :: fftw_plan_type !! which kind of planning to use with fftw
      INTEGER :: idum !! random number seed
      INTEGER :: prog_name_id !! index to define the type of program
      INTEGER :: run_type_id !! index to define the run_tupe
      INTEGER :: blacs_grid_layout !! will store the user preference for the blacs grid
      INTEGER :: k_elpa !! optimized kernel for the ELPA diagonalization library
      LOGICAL :: blacs_repeatable !! will store the user preference for the repeatability of blacs collectives
      REAL(KIND=dp) :: cp2k_start_time, cp2k_target_time
      INTEGER :: handle
   END TYPE global_environment_type

CONTAINS

! **************************************************************************************************
!> \brief creates a globenv
!> \param globenv the globenv to create
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE globenv_create(globenv)
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'globenv_create', routineP = moduleN//':'//routineN

      IF(.NOT.(.NOT. ASSOCIATED(globenv)))CALL cp__a("global_types.F",99)
      ALLOCATE (globenv)
      last_globenv_id = last_globenv_id+1
      globenv%id_nr = last_globenv_id
      globenv%ref_count = 1
      globenv%run_type_id = 0
      globenv%diag_library = "SL"
      globenv%k_elpa = 1
      globenv%default_fft_library = "FFTSG"
      globenv%fftw_wisdom_file_name = "/etc/fftw/wisdom"
      globenv%prog_name_id = 0
      globenv%idum = 0 !! random number seed
      globenv%blacs_grid_layout = BLACS_GRID_SQUARE
      globenv%cp2k_start_time = m_walltime()
      NULLIFY (globenv%gaussian_rng_stream)
   END SUBROUTINE globenv_create

! **************************************************************************************************
!> \brief retains the global environment
!> \param globenv the global environment to retain
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE globenv_retain(globenv)
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'globenv_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(globenv)))CALL cp__a("global_types.F",126)
      IF(.NOT.(globenv%ref_count > 0))CALL cp__a("global_types.F",127)
      globenv%ref_count = globenv%ref_count+1
   END SUBROUTINE globenv_retain

! **************************************************************************************************
!> \brief releases the global environment
!> \param globenv the global environment to release
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE globenv_release(globenv)
      TYPE(global_environment_type), POINTER             :: globenv

      CHARACTER(len=*), PARAMETER :: routineN = 'globenv_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(globenv)) THEN
         IF(.NOT.(globenv%ref_count > 0))CALL cp__a("global_types.F",143)
         globenv%ref_count = globenv%ref_count-1
         IF (globenv%ref_count == 0) THEN
            IF (ASSOCIATED(globenv%gaussian_rng_stream)) THEN
               CALL delete_rng_stream(globenv%gaussian_rng_stream)
            END IF
            DEALLOCATE (globenv)
         END IF
      END IF
      NULLIFY (globenv)
   END SUBROUTINE globenv_release

END MODULE global_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief CP2K transport environment and related C-interoperable types
!> \par History
!>       05.2013 created C-interoperable matrices [Hossein Bani-Hashemian]
!>       07.2013 created transport_env [Hossein Bani-Hashemian]
!>       11.2014 revised into CSR matrices [Hossein Bani-Hashemian]
!>       12.2014 merged csr_interop and transport [Hossein Bani-Hashemian]
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
MODULE transport_env_types

   USE ISO_C_BINDING,                   ONLY: C_BOOL,&
                                              C_DOUBLE,&
                                              C_FUNPTR,&
                                              C_F_POINTER,&
                                              C_INT,&
                                              C_NULL_FUNPTR,&
                                              C_NULL_PTR,&
                                              C_PTR
   USE cp_dbcsr_interface,              ONLY: cp_csr_destroy,&
                                              cp_dbcsr_release,&
                                              cp_dbcsr_type,&
                                              csr_type
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'transport_env_types'

   PUBLIC :: transport_env_type, cp2k_transport_parameters
   PUBLIC :: cp2k_csr_interop_type

! DO NOT change the ORDERING or the NAMES in the following data type
   TYPE, BIND(C) :: cp2k_transport_parameters
      INTEGER(C_INT)  :: n_occ
      INTEGER(C_INT)  :: n_atoms
      REAL(C_DOUBLE)  :: energy_diff
      REAL(C_DOUBLE)  :: evoltfactor
      INTEGER(C_INT)  :: method
      INTEGER(C_INT)  :: injection_method
      INTEGER(C_INT)  :: rlaxis_integration_method
      INTEGER(C_INT)  :: linear_solver
      INTEGER(C_INT)  :: n_abscissae
      INTEGER(C_INT)  :: n_kpoint
      INTEGER(C_INT)  :: num_interval
      INTEGER(C_INT)  :: num_contacts
      INTEGER(C_INT)  :: tasks_per_point
      INTEGER(C_INT)  :: gpus_per_point
      INTEGER(C_INT)  :: n_blocks
      INTEGER(C_INT)  :: n_points_beyn
      INTEGER(C_INT)  :: ncrc_beyn
      INTEGER(C_INT)  :: cutout(2)
      REAL(C_DOUBLE)  :: colzero_threshold
      REAL(C_DOUBLE)  :: eps_limit
      REAL(C_DOUBLE)  :: eps_limit_cc
      REAL(C_DOUBLE)  :: eps_decay
      REAL(C_DOUBLE)  :: eps_singularity_curvatures
      REAL(C_DOUBLE)  :: eps_mu
      REAL(C_DOUBLE)  :: eps_eigval_degen
      REAL(C_DOUBLE)  :: eps_fermi
      REAL(C_DOUBLE)  :: energy_interval
      REAL(C_DOUBLE)  :: min_interval
      REAL(C_DOUBLE)  :: temperature
      REAL(C_DOUBLE)  :: n_rand_beyn
      REAL(C_DOUBLE)  :: n_rand_cc_beyn
      REAL(C_DOUBLE)  :: svd_cutoff
      TYPE(C_PTR)     :: contacts_data
      TYPE(C_PTR)     :: nsgf
      TYPE(C_PTR)     :: zeff
      TYPE(C_PTR)     :: tridiag_blocks
      LOGICAL(C_BOOL) :: extra_scf
   END TYPE cp2k_transport_parameters

   TYPE transport_env_type
      TYPE(C_FUNPTR)                   :: ext_c_method_ptr = C_NULL_FUNPTR
      TYPE(cp2k_transport_parameters)  :: params
      TYPE(cp_dbcsr_type)              :: template_matrix_sym
      TYPE(cp_dbcsr_type)              :: template_matrix_nosym
      TYPE(cp_dbcsr_type)              :: csr_sparsity
      TYPE(csr_type)                   :: s_matrix
      TYPE(csr_type)                   :: ks_matrix
      TYPE(csr_type)                   :: p_matrix
      LOGICAL                          :: csr_screening
      INTEGER, DIMENSION(:), POINTER   :: tridiag_blocks => NULL()
      INTEGER, DIMENSION(:), POINTER   :: contacts_data => NULL()
      INTEGER, DIMENSION(:), POINTER   :: nsgf => NULL()
      REAL(dp), DIMENSION(:), POINTER  :: zeff => NULL()
   END TYPE transport_env_type

! DO NOT change the ORDERING or the NAMES in the following data type
   TYPE, BIND(C) :: cp2k_csr_interop_type
      INTEGER(C_INT) :: nrows_total
      INTEGER(C_INT) :: ncols_total
      INTEGER(C_INT) :: nze_total
      INTEGER(C_INT) :: nze_local
      INTEGER(C_INT) :: nrows_local
      INTEGER(C_INT) :: data_type
      INTEGER(C_INT) :: first_row
      TYPE(C_PTR)    :: rowptr_local
      TYPE(C_PTR)    :: colind_local
      TYPE(C_PTR)    :: nzerow_local
      TYPE(C_PTR)    :: nzvals_local
   ENDTYPE cp2k_csr_interop_type

   PUBLIC :: csr_interop_nullify, &
             csr_interop_matrix_get_info
   PUBLIC :: transport_env_release

CONTAINS

! **************************************************************************************************
!> \brief releases the transport_env
!> \param[inout] transport_env the transport_env to be released
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE transport_env_release(transport_env)
      TYPE(transport_env_type), POINTER                  :: transport_env

      CHARACTER(len=*), PARAMETER :: routineN = 'transport_env_release', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF(.NOT.(ASSOCIATED(transport_env)))CALL cp__a("transport_env_types.F",134)

      CALL cp_csr_destroy(transport_env%s_matrix)
      CALL cp_csr_destroy(transport_env%ks_matrix)
      CALL cp_csr_destroy(transport_env%p_matrix)
      CALL cp_dbcsr_release(transport_env%template_matrix_sym)
      CALL cp_dbcsr_release(transport_env%template_matrix_nosym)
      CALL cp_dbcsr_release(transport_env%csr_sparsity)

      transport_env%ext_c_method_ptr = C_NULL_FUNPTR

      IF (ASSOCIATED(transport_env%tridiag_blocks)) DEALLOCATE (transport_env%tridiag_blocks)
      IF (ASSOCIATED(transport_env%contacts_data)) DEALLOCATE (transport_env%contacts_data)
      IF (ASSOCIATED(transport_env%nsgf)) DEALLOCATE (transport_env%nsgf)
      IF (ASSOCIATED(transport_env%zeff)) DEALLOCATE (transport_env%zeff)

      DEALLOCATE (transport_env)

      CALL timestop(handle)

   END SUBROUTINE transport_env_release

! **************************************************************************************************
!> \brief nullifies (and zeroizes) a C-interoperable CSR matrix
!> \param[inout] csr_interop_mat the matrix to be nullified
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE csr_interop_nullify(csr_interop_mat)

      TYPE(cp2k_csr_interop_type), INTENT(INOUT)         :: csr_interop_mat

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_interop_nullify', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      csr_interop_mat%nrows_total = 0
      csr_interop_mat%ncols_total = 0
      csr_interop_mat%nze_total = 0
      csr_interop_mat%nze_local = 0
      csr_interop_mat%nrows_local = 0
      csr_interop_mat%data_type = 0
      csr_interop_mat%first_row = 0
      csr_interop_mat%rowptr_local = C_NULL_PTR
      csr_interop_mat%colind_local = C_NULL_PTR
      csr_interop_mat%nzerow_local = C_NULL_PTR
      csr_interop_mat%nzvals_local = C_NULL_PTR

      CALL timestop(handle)

   END SUBROUTINE csr_interop_nullify

! **************************************************************************************************
!> \brief gets the fields of a C-interoperable CSR matrix
!> \param[in] csr_interop_mat C-interoperable CSR matrix
!> \param[out] nrows_total     total number of rows
!> \param[out] ncols_total     total number of columns
!> \param[out] nze_local       number of local nonzero elements
!> \param[out] nze_total       total number of nonzero elements
!> \param[out] nrows_local     number of local rows
!> \param[out] data_type       data type
!> \param[out] first_row       index of the first row (C indexing)
!> \param[out] rowptr_local    row pointer (local - Fortran indexing)
!> \param[out] colind_local    column index (local - Fortran indexing)
!> \param[out] nzerow_local    number of nunzeros per row (index-i, local - Fortran indexing)
!> \param[out] nzvals_local    nonzero elements (local)
!> \author Mohammad Hossein Bani-Hashemian
! **************************************************************************************************
   SUBROUTINE csr_interop_matrix_get_info(csr_interop_mat, &
                                          nrows_total, ncols_total, nze_local, nze_total, nrows_local, data_type, &
                                          first_row, rowptr_local, colind_local, nzerow_local, nzvals_local)

      TYPE(cp2k_csr_interop_type), INTENT(IN)            :: csr_interop_mat
      INTEGER, INTENT(OUT), OPTIONAL                     :: nrows_total, ncols_total, nze_local, &
                                                            nze_total, nrows_local, data_type, &
                                                            first_row
      INTEGER, DIMENSION(:), INTENT(OUT), OPTIONAL, &
         POINTER                                         :: rowptr_local, colind_local, nzerow_local
      REAL(dp), DIMENSION(:), INTENT(OUT), OPTIONAL, &
         POINTER                                         :: nzvals_local

      CHARACTER(LEN=*), PARAMETER :: routineN = 'csr_interop_matrix_get_info', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle

      CALL timeset(routineN, handle)

      IF (PRESENT(nrows_total)) nrows_total = csr_interop_mat%nrows_total
      IF (PRESENT(ncols_total)) ncols_total = csr_interop_mat%ncols_total
      IF (PRESENT(nze_local)) nze_local = csr_interop_mat%nze_local
      IF (PRESENT(nze_total)) nze_total = csr_interop_mat%nze_total
      IF (PRESENT(nrows_local)) nrows_local = csr_interop_mat%nrows_local
      IF (PRESENT(data_type)) data_type = csr_interop_mat%data_type
      IF (PRESENT(first_row)) first_row = csr_interop_mat%first_row

      IF (PRESENT(rowptr_local)) CALL C_F_POINTER(csr_interop_mat%rowptr_local, rowptr_local, [nrows_local+1])
      IF (PRESENT(colind_local)) CALL C_F_POINTER(csr_interop_mat%colind_local, colind_local, [nze_local])
      IF (PRESENT(nzerow_local)) CALL C_F_POINTER(csr_interop_mat%nzerow_local, nzerow_local, [nrows_local])
      IF (PRESENT(nzvals_local)) CALL C_F_POINTER(csr_interop_mat%nzvals_local, nzvals_local, [nze_local])

      CALL timestop(handle)

   END SUBROUTINE csr_interop_matrix_get_info

END MODULE transport_env_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \par History
!>      JGH FEB-13-2007 : Distributed/replicated realspace grids
!>      Teodoro Laino [tlaino] - University of Zurich - 12.2007
!> \author CJM NOV-30-2003
! **************************************************************************************************
MODULE ewald_environment_types
   USE cp_log_handling,                 ONLY: cp_get_default_logger,&
                                              cp_logger_type
   USE cp_output_handling,              ONLY: cp_print_key_finished_output,&
                                              cp_print_key_unit_nr
   USE cp_para_env,                     ONLY: cp_para_env_release,&
                                              cp_para_env_retain
   USE cp_para_types,                   ONLY: cp_para_env_type
   USE cp_units,                        ONLY: cp_unit_from_cp2k
   USE input_cp2k_poisson,              ONLY: create_ewald_section
   USE input_enumeration_types,         ONLY: enum_i2c,&
                                              enumeration_type
   USE input_keyword_types,             ONLY: keyword_get,&
                                              keyword_type
   USE input_section_types,             ONLY: section_get_keyword,&
                                              section_release,&
                                              section_type,&
                                              section_vals_get_subs_vals,&
                                              section_vals_release,&
                                              section_vals_retain,&
                                              section_vals_type,&
                                              section_vals_val_get
   USE kinds,                           ONLY: dp
   USE pw_poisson_types,                ONLY: do_ewald_ewald,&
                                              do_ewald_none,&
                                              do_ewald_pme,&
                                              do_ewald_spme
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

! **************************************************************************************************
!> \brief to build arrays of pointers
!> \param ewald_env the pointer to the ewald_env
!> \par History
!>      11/03
!> \author CJM
! **************************************************************************************************
   TYPE ewald_environment_type
      PRIVATE
      INTEGER   :: id_nr, ref_count
      LOGICAL   :: do_multipoles ! Flag for using the multipole code
      INTEGER   :: do_ipol ! Solver for induced dipoles
      INTEGER   :: max_multipole ! max expansion in the multipoles
      INTEGER   :: max_ipol_iter ! max number of interation for induced dipoles
      INTEGER   :: ewald_type ! type of ewald
      INTEGER   :: gmax(3) ! max Miller index
      INTEGER   :: ns_max ! # grid points for small grid (PME)
      INTEGER   :: o_spline ! order of spline (SPME)
      REAL(KIND=dp) :: precs ! precision achieved when evaluating the real-space part
      REAL(KIND=dp) :: alpha, rcut ! ewald alpha and real-space cutoff
      REAL(KIND=dp) :: epsilon ! tolerance for small grid (PME)
      REAL(KIND=dp) :: eps_pol ! tolerance for convergence of induced dipoles
      TYPE(cp_para_env_type), POINTER          :: para_env
      TYPE(section_vals_type), POINTER         :: poisson_section
      ! interaction cutoff is required to make the electrostatic interaction
      ! continuous at a pair distance equal to rcut. this is ignored by the
      ! multipole code and is otherwise only active when SHIFT_CUTOFF is used.
      REAL(KIND=dp), DIMENSION(:, :, :), POINTER :: interaction_cutoffs
      ! store current cell, used to rebuild lazily.
      REAL(KIND=dp), DIMENSION(3, 3)          :: cell_hmat = -1.0_dp
   END TYPE ewald_environment_type

! **************************************************************************************************
!> \brief to build arrays of pointers
!> \param ewald_env the pointer to the ewald_env
!> \par History
!>      11/03
!> \author CJM
! **************************************************************************************************
   TYPE ewald_environment_p_type
      TYPE(ewald_environment_type), POINTER :: ewald_env
   END TYPE ewald_environment_p_type

! *** Public data types ***
   PUBLIC :: ewald_environment_type

! *** Public subroutines ***
   PUBLIC :: ewald_env_get, &
             ewald_env_set, &
             ewald_env_create, &
             ewald_env_retain, &
             ewald_env_release, &
             read_ewald_section

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'ewald_environment_types'
   INTEGER, PRIVATE, SAVE :: last_ewald_env_id_nr = 0

CONTAINS

! **************************************************************************************************
!> \brief Purpose: Get the EWALD environment.
!> \param ewald_env the pointer to the ewald_env
!> \param ewald_type ...
!> \param alpha ...
!> \param eps_pol ...
!> \param epsilon ...
!> \param gmax ...
!> \param ns_max ...
!> \param o_spline ...
!> \param group ...
!> \param para_env ...
!> \param id_nr ...
!> \param poisson_section ...
!> \param precs ...
!> \param rcut ...
!> \param do_multipoles ...
!> \param max_multipole ...
!> \param do_ipol ...
!> \param max_ipol_iter ...
!> \param interaction_cutoffs ...
!> \param cell_hmat ...
!> \par History
!>      11/03
!> \author CJM
! **************************************************************************************************
   SUBROUTINE ewald_env_get(ewald_env, ewald_type, alpha, eps_pol, epsilon, &
                            gmax, ns_max, o_spline, group, para_env, id_nr, poisson_section, precs, &
                            rcut, do_multipoles, max_multipole, do_ipol, max_ipol_iter, &
                            interaction_cutoffs, cell_hmat)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      INTEGER, OPTIONAL                                  :: ewald_type
      REAL(KIND=dp), OPTIONAL                            :: alpha, eps_pol, epsilon
      INTEGER, OPTIONAL                                  :: gmax(3), ns_max, o_spline, group
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      INTEGER, INTENT(OUT), OPTIONAL                     :: id_nr
      TYPE(section_vals_type), OPTIONAL, POINTER         :: poisson_section
      REAL(KIND=dp), OPTIONAL                            :: precs, rcut
      LOGICAL, INTENT(OUT), OPTIONAL                     :: do_multipoles
      INTEGER, INTENT(OUT), OPTIONAL                     :: max_multipole, do_ipol, max_ipol_iter
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: interaction_cutoffs
      REAL(KIND=dp), DIMENSION(3, 3), OPTIONAL           :: cell_hmat

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_env_get', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ewald_env)))CALL cp__a("ewald_environment_types.F",149)

      IF (PRESENT(id_nr)) id_nr = ewald_env%id_nr
      IF (PRESENT(ewald_type)) ewald_type = ewald_env%ewald_type
      IF (PRESENT(do_multipoles)) do_multipoles = ewald_env%do_multipoles
      IF (PRESENT(do_ipol)) do_ipol = ewald_env%do_ipol
      IF (PRESENT(max_multipole)) max_multipole = ewald_env%max_multipole
      IF (PRESENT(max_ipol_iter)) max_ipol_iter = ewald_env%max_ipol_iter
      IF (PRESENT(alpha)) alpha = ewald_env%alpha
      IF (PRESENT(precs)) precs = ewald_env%precs
      IF (PRESENT(rcut)) rcut = ewald_env%rcut
      IF (PRESENT(epsilon)) epsilon = ewald_env%epsilon
      IF (PRESENT(eps_pol)) eps_pol = ewald_env%eps_pol
      IF (PRESENT(gmax)) gmax = ewald_env%gmax
      IF (PRESENT(ns_max)) ns_max = ewald_env%ns_max
      IF (PRESENT(o_spline)) o_spline = ewald_env%o_spline
      IF (PRESENT(group)) group = ewald_env%para_env%group
      IF (PRESENT(para_env)) para_env => ewald_env%para_env
      IF (PRESENT(poisson_section)) poisson_section => ewald_env%poisson_section
      IF (PRESENT(interaction_cutoffs)) interaction_cutoffs => &
         ewald_env%interaction_cutoffs
      IF (PRESENT(cell_hmat)) cell_hmat = ewald_env%cell_hmat
   END SUBROUTINE ewald_env_get

! **************************************************************************************************
!> \brief Purpose: Set the EWALD environment.
!> \param ewald_env the pointer to the ewald_env
!> \param ewald_type ...
!> \param alpha ...
!> \param epsilon ...
!> \param eps_pol ...
!> \param gmax ...
!> \param ns_max ...
!> \param precs ...
!> \param o_spline ...
!> \param para_env ...
!> \param id_nr ...
!> \param poisson_section ...
!> \param interaction_cutoffs ...
!> \param cell_hmat ...
!> \par History
!>      11/03
!> \author CJM
! **************************************************************************************************
   SUBROUTINE ewald_env_set(ewald_env, ewald_type, alpha, epsilon, eps_pol, &
                            gmax, ns_max, precs, o_spline, para_env, id_nr, poisson_section, &
                            interaction_cutoffs, cell_hmat)

      TYPE(ewald_environment_type), POINTER              :: ewald_env
      INTEGER, OPTIONAL                                  :: ewald_type
      REAL(KIND=dp), OPTIONAL                            :: alpha, epsilon, eps_pol
      INTEGER, OPTIONAL                                  :: gmax(3), ns_max
      REAL(KIND=dp), OPTIONAL                            :: precs
      INTEGER, OPTIONAL                                  :: o_spline
      TYPE(cp_para_env_type), OPTIONAL, POINTER          :: para_env
      INTEGER, INTENT(IN), OPTIONAL                      :: id_nr
      TYPE(section_vals_type), OPTIONAL, POINTER         :: poisson_section
      REAL(KIND=dp), DIMENSION(:, :, :), OPTIONAL, &
         POINTER                                         :: interaction_cutoffs
      REAL(KIND=dp), DIMENSION(3, 3), OPTIONAL           :: cell_hmat

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_env_set', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ewald_env)))CALL cp__a("ewald_environment_types.F",212)

      IF (PRESENT(id_nr)) ewald_env%id_nr = id_nr
      IF (PRESENT(ewald_type)) ewald_env%ewald_type = ewald_type
      IF (PRESENT(alpha)) ewald_env%alpha = alpha
      IF (PRESENT(precs)) ewald_env%precs = precs
      IF (PRESENT(epsilon)) ewald_env%epsilon = epsilon
      IF (PRESENT(eps_pol)) ewald_env%eps_pol = eps_pol
      IF (PRESENT(gmax)) ewald_env%gmax = gmax
      IF (PRESENT(ns_max)) ewald_env%ns_max = ns_max
      IF (PRESENT(o_spline)) ewald_env%o_spline = o_spline
      IF (PRESENT(para_env)) ewald_env%para_env => para_env
      IF (PRESENT(poisson_section)) THEN
         CALL section_vals_retain(poisson_section)
         CALL section_vals_release(ewald_env%poisson_section)
         ewald_env%poisson_section => poisson_section
      END IF
      IF (PRESENT(interaction_cutoffs)) ewald_env%interaction_cutoffs => &
         interaction_cutoffs
      IF (PRESENT(cell_hmat)) ewald_env%cell_hmat = cell_hmat
   END SUBROUTINE ewald_env_set

! **************************************************************************************************
!> \brief allocates and intitializes a ewald_env
!> \param ewald_env the object to create
!> \param para_env ...
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE ewald_env_create(ewald_env, para_env)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(cp_para_env_type), POINTER                    :: para_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_env_create', &
         routineP = moduleN//':'//routineN

      ALLOCATE (ewald_env)
      ewald_env%ref_count = 1
      last_ewald_env_id_nr = last_ewald_env_id_nr+1
      ewald_env%id_nr = last_ewald_env_id_nr
      NULLIFY (ewald_env%poisson_section)
      CALL cp_para_env_retain(para_env)
      ewald_env%para_env => para_env
      NULLIFY (ewald_env%interaction_cutoffs) ! allocated and initialized later
   END SUBROUTINE ewald_env_create

! **************************************************************************************************
!> \brief retains the given ewald_env (see doc/ReferenceCounting.html)
!> \param ewald_env the object to retain
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE ewald_env_retain(ewald_env)
      TYPE(ewald_environment_type), POINTER              :: ewald_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_env_retain', &
         routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(ewald_env)))CALL cp__a("ewald_environment_types.F",272)
      IF(.NOT.(ewald_env%ref_count > 0))CALL cp__a("ewald_environment_types.F",273)
      ewald_env%ref_count = ewald_env%ref_count+1
   END SUBROUTINE ewald_env_retain

! **************************************************************************************************
!> \brief releases the given ewald_env (see doc/ReferenceCounting.html)
!> \param ewald_env the object to release
!> \par History
!>      12.2002 created [fawzi]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE ewald_env_release(ewald_env)
      TYPE(ewald_environment_type), POINTER              :: ewald_env

      CHARACTER(len=*), PARAMETER :: routineN = 'ewald_env_release', &
         routineP = moduleN//':'//routineN

      IF (ASSOCIATED(ewald_env)) THEN
         IF(.NOT.(ewald_env%ref_count > 0))CALL cp__a("ewald_environment_types.F",291)
         ewald_env%ref_count = ewald_env%ref_count-1
         IF (ewald_env%ref_count < 1) THEN
            CALL cp_para_env_release(ewald_env%para_env)
            CALL section_vals_release(ewald_env%poisson_section)
            IF (ASSOCIATED(ewald_env%interaction_cutoffs)) THEN
               DEALLOCATE (ewald_env%interaction_cutoffs)
            ENDIF
            DEALLOCATE (ewald_env)
         ENDIF
      END IF
      NULLIFY (ewald_env)
   END SUBROUTINE ewald_env_release

! **************************************************************************************************
!> \brief Purpose: read the EWALD section
!> \param ewald_env the pointer to the ewald_env
!> \param ewald_section ...
!> \author Teodoro Laino [tlaino] -University of Zurich - 2005
! **************************************************************************************************
   SUBROUTINE read_ewald_section(ewald_env, ewald_section)
      TYPE(ewald_environment_type), POINTER              :: ewald_env
      TYPE(section_vals_type), POINTER                   :: ewald_section

      CHARACTER(len=*), PARAMETER :: routineN = 'read_ewald_section', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: iw
      INTEGER, DIMENSION(:), POINTER                     :: gmax_read
      LOGICAL                                            :: explicit
      REAL(KIND=dp)                                      :: dummy
      TYPE(cp_logger_type), POINTER                      :: logger
      TYPE(enumeration_type), POINTER                    :: enum
      TYPE(keyword_type), POINTER                        :: keyword
      TYPE(section_type), POINTER                        :: section
      TYPE(section_vals_type), POINTER                   :: multipole_section

      NULLIFY (enum, keyword, section, multipole_section)
      logger => cp_get_default_logger()
      CALL section_vals_val_get(ewald_section, "EWALD_TYPE", i_val=ewald_env%ewald_type)
      CALL section_vals_val_get(ewald_section, "ALPHA", r_val=ewald_env%alpha)
      CALL section_vals_val_get(ewald_section, "EWALD_ACCURACY", r_val=ewald_env%precs)

      IF (ewald_env%ewald_type == do_ewald_none) THEN
         ewald_env%rcut = 0.0_dp
      ELSE
         CALL section_vals_val_get(ewald_section, "RCUT", explicit=explicit)
         IF (explicit) THEN
            CALL section_vals_val_get(ewald_section, "RCUT", r_val=ewald_env%rcut)
         ELSE
            ewald_env%rcut = find_ewald_optimal_value(ewald_env%precs)/ewald_env%alpha
         ENDIF
      END IF
      ! we have no defaults for gmax, gmax is only needed for ewald and spme
      SELECT CASE (ewald_env%ewald_type)
      CASE (do_ewald_ewald, do_ewald_spme)
         CALL section_vals_val_get(ewald_section, "GMAX", i_vals=gmax_read)
         SELECT CASE (SIZE (gmax_read, 1))
         CASE (1)
            ewald_env%gmax = gmax_read(1)
         CASE (3)
            ewald_env%gmax = gmax_read
         CASE DEFAULT
            CALL cp__b("ewald_environment_types.F",354,"")
         END SELECT
         IF (ewald_env%ewald_type == do_ewald_spme) THEN
            CALL section_vals_val_get(ewald_section, "O_SPLINE", i_val=ewald_env%o_spline)
         END IF
      CASE (do_ewald_pme)
         CALL section_vals_val_get(ewald_section, "NS_MAX", i_val=ewald_env%ns_max)
         CALL section_vals_val_get(ewald_section, "EPSILON", r_val=ewald_env%epsilon)
      CASE DEFAULT
         ! this should not be used for do_ewald_none
         ewald_env%gmax = HUGE(0)
         ewald_env%ns_max = HUGE(0)
      END SELECT

      ! Multipoles
      multipole_section => section_vals_get_subs_vals(ewald_section, "MULTIPOLES")
      CALL section_vals_val_get(multipole_section, "_SECTION_PARAMETERS_", l_val=ewald_env%do_multipoles)
      CALL section_vals_val_get(multipole_section, "POL_SCF", i_val=ewald_env%do_ipol)
      CALL section_vals_val_get(multipole_section, "EPS_POL", r_val=ewald_env%eps_pol)
      IF (ewald_env%do_multipoles) THEN
         SELECT CASE (ewald_env%ewald_type)
         CASE (do_ewald_ewald)
            CALL section_vals_val_get(multipole_section, "MAX_MULTIPOLE_EXPANSION", i_val=ewald_env%max_multipole)
            CALL section_vals_val_get(multipole_section, "MAX_IPOL_ITER", i_val=ewald_env%max_ipol_iter)
         CASE DEFAULT
            CALL cp__b("ewald_environment_types.F",379,"Multipole code works at the moment only with standard EWALD sums.")
         END SELECT
      END IF

      iw = cp_print_key_unit_nr(logger, ewald_section, "PRINT%PROGRAM_RUN_INFO", &
                                extension=".log")
      IF (iw > 0) THEN
         NULLIFY (keyword, enum)
         CALL create_ewald_section(section)
         IF (ewald_env%ewald_type /= do_ewald_none) THEN
            keyword => section_get_keyword(section, "EWALD_TYPE")
            CALL keyword_get(keyword, enum=enum)
            WRITE (iw, '(/,T2,"EWALD| ",A,T67,A14 )') 'Summation is done by:', &
               ADJUSTR(TRIM(enum_i2c(enum, ewald_env%ewald_type)))
            IF (ewald_env%do_multipoles) THEN
               NULLIFY (keyword, enum)
               keyword => section_get_keyword(section, "MULTIPOLES%MAX_MULTIPOLE_EXPANSION")
               CALL keyword_get(keyword, enum=enum)
               WRITE (iw, '( T2,"EWALD| ",A )') 'Enabled Multipole Method'
               WRITE (iw, '( T2,"EWALD| ",A,T67,A14 )') 'Max Term in Multipole Expansion :', &
                  ADJUSTR(TRIM(enum_i2c(enum, ewald_env%max_multipole)))
               WRITE (iw, '( T2,"EWALD| ",A,T67,3I10 )') 'Max number Iterations for IPOL :', &
                  ewald_env%max_ipol_iter
            END IF
            dummy = cp_unit_from_cp2k(ewald_env%alpha, "angstrom^-1")
            WRITE (iw, '( T2,"EWALD| ",A,A18,A,T71,F10.4 )') &
               'Alpha parameter [', 'ANGSTROM^-1', ']', dummy
            dummy = cp_unit_from_cp2k(ewald_env%rcut, "angstrom")
            WRITE (iw, '( T2,"EWALD| ",A,A18,A,T71,F10.4 )') &
               'Real Space Cutoff [', 'ANGSTROM', ']', dummy

            SELECT CASE (ewald_env%ewald_type)
            CASE (do_ewald_ewald)
               WRITE (iw, '( T2,"EWALD| ",A,T51,3I10 )') &
                  'G-space max. Miller index', ewald_env%gmax
            CASE (do_ewald_pme)
               WRITE (iw, '( T2,"EWALD| ",A,T71,I10 )') &
                  'Max small-grid points (input) ', ewald_env%ns_max
               WRITE (iw, '( T2,"EWALD| ",A,T71,E10.4 )') &
                  'Gaussian tolerance (input) ', ewald_env%epsilon
            CASE (do_ewald_spme)
               WRITE (iw, '( T2,"EWALD| ",A,T51,3I10 )') &
                  'G-space max. Miller index', ewald_env%gmax
               WRITE (iw, '( T2,"EWALD| ",A,T71,I10 )') &
                  'Spline interpolation order ', ewald_env%o_spline
            CASE DEFAULT
               CALL cp__b("ewald_environment_types.F",425,"")
            END SELECT
         ELSE
            WRITE (iw, '( T2,"EWALD| ",T73, A )') 'not used'
         END IF
         CALL section_release(section)
      END IF
      CALL cp_print_key_finished_output(iw, logger, ewald_section, &
                                        "PRINT%PROGRAM_RUN_INFO")

   END SUBROUTINE read_ewald_section

! **************************************************************************************************
!> \brief triggers (by bisection) the optimal value for EWALD parameter x
!>      EXP(-x^2)/x^2 = EWALD_ACCURACY
!> \param precs ...
!> \retval value ...
!> \author Teodoro Laino [tlaino] - University of Zurich - 12.2007
! **************************************************************************************************
   FUNCTION find_ewald_optimal_value(precs) RESULT(value)
      REAL(KIND=dp)                                      :: precs, value

      CHARACTER(len=*), PARAMETER :: routineN = 'find_ewald_optimal_value', &
         routineP = moduleN//':'//routineN

      REAL(KIND=dp)                                      :: func, func1, func2, s, s1, s2

      s = 0.1_dp
      func = EXP(-s**2)/s**2-precs
      IF(.NOT.(func > 0.0_dp))CALL cp__a("ewald_environment_types.F",454)
      DO WHILE (func > 0.0_dp)
         s = s+0.1_dp
         func = EXP(-s**2)/s**2-precs
      END DO
      s2 = s
      s1 = s-0.1_dp
      ! Start bisection
      DO WHILE (.TRUE.)
         func2 = EXP(-s2**2)/s2**2-precs
         func1 = EXP(-s1**2)/s1**2-precs
         IF(.NOT.(func1 >= 0))CALL cp__a("ewald_environment_types.F",465)
         IF(.NOT.(func2 <= 0))CALL cp__a("ewald_environment_types.F",466)
         s = 0.5_dp*(s1+s2)
         func = EXP(-s**2)/s**2-precs
         IF (func > 0.0_dp) THEN
            s1 = s
         ELSE IF (func < 0.0_dp) THEN
            s2 = s
         END IF
         IF (ABS(func) < 100.0_dp*EPSILON(0.0_dp)) EXIT
      END DO
      value = s
   END FUNCTION find_ewald_optimal_value

END MODULE ewald_environment_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief Types used to generate the molecular SCF guess
!> \par History
!>       10.2014 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
MODULE mscfg_types
   USE cp_dbcsr_interface,              ONLY: &
        cp_dbcsr_add, cp_dbcsr_col_block_sizes, cp_dbcsr_complete_redistribute, cp_dbcsr_create, &
        cp_dbcsr_distribution, cp_dbcsr_distribution_new, cp_dbcsr_distribution_release, &
        cp_dbcsr_finalize, cp_dbcsr_init, cp_dbcsr_iterator, cp_dbcsr_iterator_blocks_left, &
        cp_dbcsr_iterator_next_block, cp_dbcsr_iterator_start, cp_dbcsr_iterator_stop, &
        cp_dbcsr_nfullcols_total, cp_dbcsr_nfullrows_total, cp_dbcsr_release, &
        cp_dbcsr_reserve_block2d, cp_dbcsr_row_block_sizes, cp_dbcsr_set, cp_dbcsr_type, &
        cp_dbcsr_work_create, dbcsr_distribution_col_clusters, dbcsr_distribution_col_dist, &
        dbcsr_distribution_mp, dbcsr_distribution_obj, dbcsr_distribution_row_clusters, &
        dbcsr_distribution_row_dist, dbcsr_type_no_symmetry
!cp_dbcsr_copy, &
!cp_dbcsr_filter, &
!, cp_dbcsr_nblkcols_total, &
!cp_dbcsr_nblkrows_total, cp_dbcsr_print, &
!dbcsr_mp_group, dbcsr_mp_numnodes, dbcsr_norm_maxabsnorm, &
   USE kinds,                           ONLY: dp
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'mscfg_types'

   INTEGER, PARAMETER, PUBLIC               :: mscfg_max_moset_size = 2

   ! Public types
   PUBLIC :: molecular_scf_guess_env_type

   ! Public subroutines
   PUBLIC :: molecular_scf_guess_env_init, &
             molecular_scf_guess_env_destroy, &
             get_matrix_from_submatrices

   ! Contains data pertaining to molecular_scf_guess calculations
   TYPE molecular_scf_guess_env_type

      ! Useful flags to pass around
      LOGICAL                                           :: is_fast_dirty, &
                                                           is_crystal

      ! Real data
      INTEGER                                           :: nfrags
      REAL(KIND=dp), DIMENSION(:), ALLOCATABLE          :: energy_of_frag
      INTEGER, DIMENSION(:), ALLOCATABLE                :: nmosets_of_frag
      TYPE(cp_dbcsr_type), DIMENSION(:, :), ALLOCATABLE  :: mos_of_frag

   END TYPE

CONTAINS

! **************************************************************************************************
!> \brief Allocates data
!> \param env ...
!> \param nfrags - number of entries
!> \par History
!>       2014.10 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE molecular_scf_guess_env_init(env, nfrags)

      TYPE(molecular_scf_guess_env_type)                 :: env
      INTEGER, INTENT(IN)                                :: nfrags

      CHARACTER(len=*), PARAMETER :: routineN = 'molecular_scf_guess_env_init', &
         routineP = moduleN//':'//routineN

! check if the number of fragments is already set
!IF (env%nfrags.ne.0) THEN
!   ! do not allow re-initialization
!   ! to prevent recursive calls
!   CPPostcondition(.FALSE.,cp_failure_level,routineP,failure)
!ENDIF

      env%nfrags = nfrags
      IF (nfrags .GT. 0) THEN
         ALLOCATE (env%energy_of_frag(nfrags))
         ALLOCATE (env%nmosets_of_frag(nfrags))
         ALLOCATE (env%mos_of_frag(nfrags, mscfg_max_moset_size))
      ENDIF

   END SUBROUTINE molecular_scf_guess_env_init

! **************************************************************************************************
!> \brief Destroyes both data and environment
!> \param env ...
!> \par History
!>       2014.10 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE molecular_scf_guess_env_destroy(env)

      TYPE(molecular_scf_guess_env_type)                 :: env

      CHARACTER(len=*), PARAMETER :: routineN = 'molecular_scf_guess_env_destroy', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: ifrag, jfrag

      IF (ALLOCATED(env%mos_of_frag)) THEN
         DO ifrag = 1, SIZE(env%mos_of_frag, 1)
            DO jfrag = 1, env%nmosets_of_frag(ifrag)
               CALL cp_dbcsr_release(env%mos_of_frag(ifrag, jfrag))
            ENDDO
         ENDDO
         DEALLOCATE (env%mos_of_frag)
      ENDIF
      IF (ALLOCATED(env%energy_of_frag)) DEALLOCATE (env%energy_of_frag)
      IF (ALLOCATED(env%nmosets_of_frag)) DEALLOCATE (env%nmosets_of_frag)

      env%nfrags = 0

   END SUBROUTINE molecular_scf_guess_env_destroy

! **************************************************************************************************
!> \brief Creates a distributed matrix from MOs on fragments
!> \param mscfg_env - env containing MOs of fragments
!> \param matrix_out - all existing blocks will be deleted!
!> \param iset - which set of MOs in mscfg_env has to be converted (e.g. spin)
!> \par History
!>       10.2014 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE get_matrix_from_submatrices(mscfg_env, matrix_out, iset)

      TYPE(molecular_scf_guess_env_type), INTENT(IN)     :: mscfg_env
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_out
      INTEGER, INTENT(IN)                                :: iset

      CHARACTER(len=*), PARAMETER :: routineN = 'get_matrix_from_submatrices', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: handle, ifrag
      INTEGER, DIMENSION(2)                              :: matrix_size, offset, submatrix_size
      TYPE(cp_dbcsr_type)                                :: matrix_temp

      CALL timeset(routineN, handle)

      IF(.NOT.(iset .LE. mscfg_max_moset_size))CALL cp__a("mscfg_types.F",151)

      CALL cp_dbcsr_init(matrix_temp)
      CALL cp_dbcsr_create(matrix_temp, &
                           template=matrix_out, &
                           matrix_type=dbcsr_type_no_symmetry)
      CALL cp_dbcsr_set(matrix_out, 0.0_dp)

      matrix_size(1) = cp_dbcsr_nfullrows_total(matrix_out)
      matrix_size(2) = cp_dbcsr_nfullcols_total(matrix_out)

      ! assume that the initial offset is zero
      offset(1) = 0
      offset(2) = 0

      DO ifrag = 1, mscfg_env%nfrags

         IF(.NOT.(iset .LE. mscfg_env%nmosets_of_frag(ifrag)))CALL cp__a("mscfg_types.F",168)

         submatrix_size(1) = cp_dbcsr_nfullrows_total(mscfg_env%mos_of_frag(ifrag, iset))
         submatrix_size(2) = cp_dbcsr_nfullcols_total(mscfg_env%mos_of_frag(ifrag, iset))

         CALL copy_submatrix_into_matrix(mscfg_env%mos_of_frag(ifrag, iset), &
                                         matrix_temp, offset, submatrix_size, matrix_size)

         CALL cp_dbcsr_add(matrix_out, matrix_temp, 1.0_dp, 1.0_dp)

         offset(1) = offset(1)+submatrix_size(1)
         offset(2) = offset(2)+submatrix_size(2)

      ENDDO

      ! Check that the accumulated size of submatrices
      ! is exactly the same as the size of the big matrix
      ! This is to prevent unexpected conversion errors
      ! If however such conversion is intended - remove these safeguards
      IF(.NOT.(offset(1) .EQ. matrix_size(1)))CALL cp__a("mscfg_types.F",187)
      IF(.NOT.(offset(2) .EQ. matrix_size(2)))CALL cp__a("mscfg_types.F",188)

      CALL cp_dbcsr_release(matrix_temp)

      CALL timestop(handle)

   END SUBROUTINE get_matrix_from_submatrices

! **************************************************************************************************
!> \brief Copies a distributed dbcsr submatrix into a distributed dbcsr matrix
!> \param submatrix_in ...
!> \param matrix_out - all existing blocks will be deleted!
!> \param offset ...
!> \param submatrix_size ...
!> \param matrix_size ...
!> \par History
!>       10.2014 created [Rustam Z Khaliullin]
!> \author Rustam Z Khaliullin
! **************************************************************************************************
   SUBROUTINE copy_submatrix_into_matrix(submatrix_in, matrix_out, &
                                         offset, submatrix_size, matrix_size)

      TYPE(cp_dbcsr_type), INTENT(IN)                    :: submatrix_in
      TYPE(cp_dbcsr_type), INTENT(INOUT)                 :: matrix_out
      INTEGER, DIMENSION(2), INTENT(IN)                  :: offset, submatrix_size, matrix_size

      CHARACTER(len=*), PARAMETER :: routineN = 'copy_submatrix_into_matrix', &
         routineP = moduleN//':'//routineN

      INTEGER                                            :: add_blocks_after, dimen, iblock_col, &
                                                            iblock_row, iblock_size, nblocks, &
                                                            nblocks_new, start_index, trailing_size
      INTEGER, DIMENSION(2)                              :: add_blocks_before
      INTEGER, DIMENSION(:), POINTER :: blk_distr, blk_sizes, block_sizes_new, cluster_distr, &
         cluster_distr_new, col_cluster_new, col_distr_new, col_sizes_new, distr_new_array, &
         row_cluster_new, row_distr_new, row_sizes_new
      REAL(KIND=dp), DIMENSION(:, :), POINTER            :: data_p, p_new_block
      TYPE(cp_dbcsr_iterator)                            :: iter
      TYPE(cp_dbcsr_type)                                :: matrix_new
      TYPE(dbcsr_distribution_obj)                       :: dist_new, dist_qs

! obtain distribution of the submatrix

      dist_qs = cp_dbcsr_distribution(submatrix_in)

      DO dimen = 1, 2 ! 1 - row, 2 - column dimension

         add_blocks_before(dimen) = 0
         add_blocks_after = 0
         start_index = 1
         trailing_size = matrix_size(dimen)-offset(dimen)-submatrix_size(dimen)
         IF (offset(dimen) .GT. 0) THEN
            add_blocks_before(dimen) = add_blocks_before(dimen)+1
            start_index = 2
         ENDIF
         IF (trailing_size .GT. 0) THEN
            add_blocks_after = add_blocks_after+1
         ENDIF

         NULLIFY (cluster_distr_new)

         IF (dimen == 1) THEN !rows
            blk_distr => dbcsr_distribution_row_dist(dist_qs)
            blk_sizes => cp_dbcsr_row_block_sizes(submatrix_in)
            cluster_distr => dbcsr_distribution_row_clusters(dist_qs)
         ELSE !columns
            blk_distr => dbcsr_distribution_col_dist(dist_qs)
            blk_sizes => cp_dbcsr_col_block_sizes(submatrix_in)
            cluster_distr => dbcsr_distribution_col_clusters(dist_qs)
         ENDIF
         nblocks = SIZE(blk_sizes) ! number of blocks in the small matrix

         nblocks_new = nblocks+add_blocks_before(dimen)+add_blocks_after
         ALLOCATE (block_sizes_new(nblocks_new))
         ALLOCATE (distr_new_array(nblocks_new))
         IF (ASSOCIATED(cluster_distr)) THEN
            ALLOCATE (cluster_distr_new(nblocks_new))
         END IF
         IF (add_blocks_before(dimen) .GT. 0) THEN
            block_sizes_new(1) = offset(dimen)
            distr_new_array(1) = 0
            IF (ASSOCIATED(cluster_distr)) THEN
               cluster_distr_new(1) = 0
            END IF
         ENDIF
         block_sizes_new(start_index:nblocks+start_index-1) = blk_sizes(1:nblocks)
         distr_new_array(start_index:nblocks+start_index-1) = blk_distr(1:nblocks)
         IF (ASSOCIATED(cluster_distr)) THEN
            cluster_distr_new(start_index:nblocks+start_index-1) = cluster_distr(1:nblocks)
         END IF
         IF (add_blocks_after .GT. 0) THEN
            block_sizes_new(nblocks_new) = trailing_size
            distr_new_array(nblocks_new) = 0
            IF (ASSOCIATED(cluster_distr)) THEN
               cluster_distr_new(nblocks_new) = 0
            END IF
         ENDIF

         ! create final arrays
         IF (dimen == 1) THEN !rows
            row_sizes_new => block_sizes_new
            row_distr_new => distr_new_array
            row_cluster_new => cluster_distr_new
         ELSE !columns
            col_sizes_new => block_sizes_new
            col_distr_new => distr_new_array
            col_cluster_new => cluster_distr_new
         ENDIF
      ENDDO ! both rows and columns are done

      ! Create the distribution
      CALL cp_dbcsr_distribution_new(dist_new, &
                                     dbcsr_distribution_mp(dist_qs), row_distr_new, col_distr_new, &
                                     row_cluster_new, col_cluster_new, &
                                     reuse_arrays=.TRUE.)

      ! Create big the matrix
      CALL cp_dbcsr_init(matrix_new)
      CALL cp_dbcsr_create(matrix_new, "BIG_AND_FAKE", &
                           dist_new, dbcsr_type_no_symmetry, &
                           row_sizes_new, col_sizes_new, &
                           reuse_arrays=.TRUE.)
      CALL cp_dbcsr_distribution_release(dist_new)

      !CALL cp_dbcsr_finalize(matrix_new)

      ! copy blocks of the small matrix to the big matrix
      !mynode = dbcsr_mp_mynode(dbcsr_distribution_mp(cp_dbcsr_distribution(matrix_new)))
      CALL cp_dbcsr_work_create(matrix_new, work_mutable=.TRUE.)

      ! iterate over local blocks of the small matrix
      CALL cp_dbcsr_iterator_start(iter, submatrix_in)

      DO WHILE (cp_dbcsr_iterator_blocks_left(iter))

         CALL cp_dbcsr_iterator_next_block(iter, iblock_row, iblock_col, data_p, row_size=iblock_size)

         ! it is important that distribution of the big matrix is the same as
         ! that of the small matrix but has the same number of columns and rows
         ! as the super-system matrix. this is necessary for complete redistribute
         ! to work
         NULLIFY (p_new_block)
         CALL cp_dbcsr_reserve_block2d(matrix_new, &
                                       iblock_row+add_blocks_before(1), &
                                       iblock_col+add_blocks_before(2), &
                                       p_new_block)

         IF(.NOT.(ASSOCIATED(p_new_block)))CALL cp__a("mscfg_types.F",335)
         IF(.NOT.(SIZE(p_new_block, 1) .EQ. SIZE(data_p, 1)))CALL cp__a("mscfg_types.F",336)
         IF(.NOT.(SIZE(p_new_block, 2) .EQ. SIZE(data_p, 2)))CALL cp__a("mscfg_types.F",337)

         p_new_block(:, :) = data_p(:, :)

      ENDDO
      CALL cp_dbcsr_iterator_stop(iter)

      CALL cp_dbcsr_finalize(matrix_new)

      ! finally call complete redistribute to get the matrix of the entire system
      CALL cp_dbcsr_set(matrix_out, 0.0_dp)
      CALL cp_dbcsr_complete_redistribute(matrix_new, matrix_out)
      CALL cp_dbcsr_release(matrix_new)

   END SUBROUTINE copy_submatrix_into_matrix

END MODULE mscfg_types

!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief parameters that control a relativistic calculation
!> \par History
!>      09.2002 created [fawzi] (as scf_control_types.F)
!>      10.2008 modifed for relativistic control types (Jens Thar)
!> \author Fawzi Mohamed
! **************************************************************************************************
MODULE rel_control_types

   USE input_constants,                 ONLY: rel_none,&
                                              rel_pot_full,&
                                              rel_trans_full,&
                                              rel_zora_full
   USE input_section_types,             ONLY: section_vals_get_subs_vals,&
                                              section_vals_type,&
                                              section_vals_val_get
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE

   PRIVATE

   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'rel_control_types'
   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.

   ! Public data types

   PUBLIC :: rel_control_type

   ! Public subroutines

   PUBLIC :: rel_c_create, &
             rel_c_read_parameters, &
             rel_c_release, &
             rel_c_retain

! **************************************************************************************************
!> \brief contains the parameters needed by a relativistic calculation
!> \param method used relativistic method (NONE, DKH)
!> \param DKH_order order of the DKH transformation (2,3)
!> \param transformation used blocks of the full matrix (FULL, MOLECULE, ATOM)
!> \param z_cutoff considered range of the Coulomb interaction
!> \param potential nuclear electron Coulomb potential (FULL, ERFC)
!> \par History
!>      09.2002 created [fawzi] for scf_control_type
!>      10.2008 copied to rel_control_type [JT]
!> \author Fawzi Mohamed
! **************************************************************************************************
   TYPE rel_control_type
      INTEGER                               :: rel_method
      INTEGER                               :: rel_DKH_order
      INTEGER                               :: rel_ZORA_type
      INTEGER                               :: rel_transformation
      INTEGER                               :: rel_z_cutoff
      INTEGER                               :: rel_potential
      INTEGER                               :: ref_count
   END TYPE rel_control_type

CONTAINS

! **************************************************************************************************
!> \brief allocates and initializes an rel control object with the default values
!> \param rel_control the object to initialize
!> \par History
!>      09.2002 created [fawzi] for scf_control_type
!>      10.2008 copied to rel_control_type [JT]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE rel_c_create(rel_control)

      TYPE(rel_control_type), POINTER                    :: rel_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rel_c_create', routineP = moduleN//':'//routineN

      ALLOCATE (rel_control)

      ! Load the default values

      rel_control%rel_method = rel_none
      rel_control%rel_DKH_order = 2
      rel_control%rel_ZORA_type = rel_zora_full
      rel_control%rel_transformation = rel_trans_full
      rel_control%rel_z_cutoff = 1
      rel_control%rel_potential = rel_pot_full
      rel_control%ref_count = 1

   END SUBROUTINE rel_c_create

! **************************************************************************************************
!> \brief retains the given rel_control (see cp2k/doc/ReferenceCounting.html)
!> \param rel_control the object to retain
!> \par History
!>      09.2002 created [fawzi] for scf_control_type
!>      10.2008 copied to rel_control_type [JT]
!> \author Fawzi Mohamed
! **************************************************************************************************
   SUBROUTINE rel_c_retain(rel_control)

      TYPE(rel_control_type), POINTER                    :: rel_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rel_c_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(rel_control)))CALL cp__a("rel_control_types.F",108)

      IF(.NOT.(rel_control%ref_count > 0))CALL cp__a("rel_control_types.F",110)
      rel_control%ref_count = rel_control%ref_count+1

   END SUBROUTINE rel_c_retain

! **************************************************************************************************
!> \brief releases the given rel_control (see cp2k/doc/ReferenceCounting.html)
!> \param rel_control the object to free
!> \par History
!>      09.2002 created [fawzi] for scf_control_type
!>      10.2008 copied to rel_control_type [JT]
!> \author Fawzi Mohamed
!> \note
!>      at the moment does nothing
! **************************************************************************************************
   SUBROUTINE rel_c_release(rel_control)

      TYPE(rel_control_type), POINTER                    :: rel_control

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rel_c_release', routineP = moduleN//':'//routineN

      IF (ASSOCIATED(rel_control)) THEN
         IF(.NOT.(rel_control%ref_count > 0))CALL cp__a("rel_control_types.F",132)
         rel_control%ref_count = rel_control%ref_count-1
         IF (rel_control%ref_count < 1) THEN
            DEALLOCATE (rel_control)
         END IF
      END IF

      NULLIFY (rel_control)

   END SUBROUTINE rel_c_release

! **************************************************************************************************
!> \brief reads the parameters of the relativistic section into the given rel_control
!> \param rel_control the object that wil contain the values read
!> \param dft_section ...
!> \par History
!>      05.2001 created [Matthias] for scf_control_type
!>      09.2002 created separated scf_control type [fawzi]
!>      10.2008 copied to rel_control_type [JT]
!> \author Matthias Krack
! **************************************************************************************************
   SUBROUTINE rel_c_read_parameters(rel_control, dft_section)

      TYPE(rel_control_type), POINTER                    :: rel_control
      TYPE(section_vals_type), POINTER                   :: dft_section

      CHARACTER(LEN=*), PARAMETER :: routineN = 'rel_c_read_parameters', &
         routineP = moduleN//':'//routineN

      TYPE(section_vals_type), POINTER                   :: rel_section

      IF(.NOT.(ASSOCIATED(rel_control)))CALL cp__a("rel_control_types.F",163)
      IF(.NOT.(rel_control%ref_count > 0))CALL cp__a("rel_control_types.F",164)
      IF(.NOT.(ASSOCIATED(dft_section)))CALL cp__a("rel_control_types.F",165)

      rel_section => section_vals_get_subs_vals(dft_section, "RELATIVISTIC")
      CALL section_vals_val_get(rel_section, "method", &
                                i_val=rel_control%rel_method)
      CALL section_vals_val_get(rel_section, "DKH_order", &
                                i_val=rel_control%rel_DKH_order)
      CALL section_vals_val_get(rel_section, "ZORA_TYPE", &
                                i_val=rel_control%rel_zora_type)
      CALL section_vals_val_get(rel_section, "transformation", &
                                i_val=rel_control%rel_transformation)
      CALL section_vals_val_get(rel_section, "z_cutoff", &
                                i_val=rel_control%rel_z_cutoff)
      CALL section_vals_val_get(rel_section, "potential", &
                                i_val=rel_control%rel_potential)

   END SUBROUTINE rel_c_read_parameters

END MODULE rel_control_types
!--------------------------------------------------------------------------------------------------!
!   CP2K: A general program to perform molecular dynamics simulations                              !
!   Copyright (C) 2000 - 2016  CP2K developers group                                               !
!--------------------------------------------------------------------------------------------------!

! **************************************************************************************************
!> \brief interpolate the wavefunctions to speed up the convergence when
!>      doing MD
!> \par History
!>      12.2002 created [fawzi]
!>      02.2005 wf_mol added [MI]
!> \author fawzi
! **************************************************************************************************
MODULE qs_wf_history_types
   USE cp_dbcsr_interface,              ONLY: cp_dbcsr_deallocate_matrix,&
                                              cp_dbcsr_deallocate_matrix_set,&
                                              cp_dbcsr_p_type,&
                                              cp_dbcsr_type
   USE cp_fm_types,                     ONLY: cp_fm_p_type,&
                                              cp_fm_release
   USE kinds,                           ONLY: dp
   USE pw_types,                        ONLY: pw_p_type
   USE qs_rho_types,                    ONLY: qs_rho_release,&
                                              qs_rho_type
! Basic use statements and preprocessor macros
! should be included in the use statements

  USE base_hooks,                      ONLY: cp__a,&
                                             cp__b,&
                                             cp__w,&
                                             cp__l,&
                                             cp_abort,&
                                             cp_warn,&
                                             timeset,&
                                             timestop


! Dangerous: Full path can be arbitrarily long and might overflow Fortran line.


! The MARK_USED macro can be used to mark an argument/variable as used.
! It is intended to make it possible to switch on -Werror=unused-dummy-argument,
! but deal elegantly with e.g. library wrapper routines that take arguments only used if the library is linked in. 
! This code should be valid for any Fortran variable, is always standard conforming,
! and will be optimized away completely by the compiler

   IMPLICIT NONE
   PRIVATE

   LOGICAL, PRIVATE, PARAMETER :: debug_this_module = .TRUE.
   CHARACTER(len=*), PARAMETER, PRIVATE :: moduleN = 'qs_wf_history_types'

   PUBLIC :: qs_wf_snapshot_type, &
             qs_wf_history_type, qs_wf_history_p_type
   PUBLIC :: wfi_retain, wfi_release, wfi_get_snapshot

! **************************************************************************************************
!> \brief represent a past snapshot of the wavefunction.
!>      some elements might not be associated (to spare memory)
!>      depending on how the snapshot was taken
!> \param wf the wavefunctions
!> \param rho_r the density in r space
!> \param rho_g the density in g space
!> \param rho_ao the density in ao space
!> \param overlap the overlap matrix
!> \param rho_frozen the frozen density structure
!> \param dt the time of the snapshot (wrf to te previous snapshot!)
!> \param id_nr unique identification number
!> \param ref_count reference count (see doc/ReferenceCounting.html)
!> \note
!>      keep trak also of occupation numbers and energies?
!> \par History
!>      02.2003 created [fawzi]
!>      02.2005 wf_mol added [MI]
!> \author fawzi
! **************************************************************************************************
   TYPE qs_wf_snapshot_type
      TYPE(cp_fm_p_type), DIMENSION(:), POINTER :: wf
      TYPE(pw_p_type), DIMENSION(:), POINTER :: rho_r
      TYPE(pw_p_type), DIMENSION(:), POINTER :: rho_g
      TYPE(cp_dbcsr_p_type), DIMENSION(:), POINTER :: rho_ao
      TYPE(cp_dbcsr_p_type), DIMENSION(:, :), POINTER :: rho_ao_kp
      TYPE(cp_dbcsr_type), POINTER :: overlap
      TYPE(qs_rho_type), POINTER :: rho_frozen
      REAL(KIND=dp) :: dt
      INTEGER :: id_nr, ref_count
   END TYPE qs_wf_snapshot_type

! **************************************************************************************************
!> \brief pointer to a snapshot
!> \param snapshot the pointer to the snapshot
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE qs_wf_snapshot_p_type
      TYPE(qs_wf_snapshot_type), POINTER :: snapshot
   END TYPE qs_wf_snapshot_p_type

! **************************************************************************************************
!> \brief keeps track of the previous wavefunctions and can extrapolate them
!>      for the next step of md
!> \param id_nr unique identification number
!> \param ref_cont reference count (see doc/ReferenceCounting.html)
!> \param memory_depth how many snapshots should be stored
!> \param last_state_index index of the latest snapshot
!> \param past_states array with the past states (index starts at
!>        last_state_index)
!> \param interpolation_method_nr the tag of the method used to
!>        extrapolate the new start state for qs
!> \param snapshot_count number of snapshot taken so far (cumulative,
!>        can be bigger than the history depth)
!> \note
!>      use a linked list for the past states ?
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   TYPE qs_wf_history_type
      INTEGER :: id_nr, ref_count, memory_depth, last_state_index, &
                 interpolation_method_nr, snapshot_count
      LOGICAL :: store_wf, store_rho_r, store_rho_g, store_rho_ao, store_rho_ao_kp, &
                 store_overlap, store_frozen_density
      TYPE(qs_wf_snapshot_p_type), DIMENSION(:), POINTER :: past_states
   END TYPE qs_wf_history_type

! **************************************************************************************************
!> \brief to create arrays of pointers to qs_wf_history_type
!> \param wf_hist the pointer to the wf history
!> \author fawzi
! **************************************************************************************************
   TYPE qs_wf_history_p_type
      TYPE(qs_wf_history_type), POINTER :: wf_history
   END TYPE qs_wf_history_p_type

CONTAINS

! **************************************************************************************************
!> \brief releases a snapshot of a wavefunction (see doc/ReferenceCounting.html)
!> \param snapshot the snapshot to release
!> \par History
!>      02.2003 created [fawzi]
!>      02.2005 wf_mol added [MI]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE wfs_release(snapshot)
      TYPE(qs_wf_snapshot_type), POINTER                 :: snapshot

      CHARACTER(len=*), PARAMETER :: routineN = 'wfs_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(snapshot)) THEN
         IF(.NOT.(snapshot%ref_count > 0))CALL cp__a("qs_wf_history_types.F",134)
         snapshot%ref_count = snapshot%ref_count-1
         IF (snapshot%ref_count == 0) THEN
            IF (ASSOCIATED(snapshot%wf)) THEN
               DO i = 1, SIZE(snapshot%wf)
                  CALL cp_fm_release(snapshot%wf(i)%matrix)
               END DO
               DEALLOCATE (snapshot%wf)
            END IF
            ! snapshot%rho_r & snapshot%rho_g is deallocated in wfs_update
            ! of qs_wf_history_methods, in case you wonder about it.
            IF (ASSOCIATED(snapshot%rho_ao)) THEN
               CALL cp_dbcsr_deallocate_matrix_set(snapshot%rho_ao)
            END IF
            IF (ASSOCIATED(snapshot%rho_ao_kp)) THEN
               CALL cp_dbcsr_deallocate_matrix_set(snapshot%rho_ao_kp)
            END IF
            IF (ASSOCIATED(snapshot%overlap)) THEN
               CALL cp_dbcsr_deallocate_matrix(snapshot%overlap)
            END IF
            IF (ASSOCIATED(snapshot%rho_frozen)) THEN
               CALL qs_rho_release(snapshot%rho_frozen)
            END IF
            DEALLOCATE (snapshot)
         END IF
      END IF
      NULLIFY (snapshot)
   END SUBROUTINE wfs_release

! **************************************************************************************************
!> \brief retains a wf history (see doc/ReferenceCounting.html)
!> \param wf_history the wf_history to retain
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE wfi_retain(wf_history)
      TYPE(qs_wf_history_type), POINTER                  :: wf_history

      CHARACTER(len=*), PARAMETER :: routineN = 'wfi_retain', routineP = moduleN//':'//routineN

      IF(.NOT.(ASSOCIATED(wf_history)))CALL cp__a("qs_wf_history_types.F",175)
      wf_history%ref_count = wf_history%ref_count+1

   END SUBROUTINE wfi_retain

! **************************************************************************************************
!> \brief releases a wf_history of a wavefunction
!>      (see doc/ReferenceCounting.html)
!> \param wf_history the wf_history to release
!> \par History
!>      02.2003 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   SUBROUTINE wfi_release(wf_history)
      TYPE(qs_wf_history_type), POINTER                  :: wf_history

      CHARACTER(len=*), PARAMETER :: routineN = 'wfi_release', routineP = moduleN//':'//routineN

      INTEGER                                            :: i

      IF (ASSOCIATED(wf_history)) THEN
         IF(.NOT.(wf_history%ref_count > 0))CALL cp__a("qs_wf_history_types.F",196)
         wf_history%ref_count = wf_history%ref_count-1
         IF (wf_history%ref_count == 0) THEN
            IF (ASSOCIATED(wf_history%past_states)) THEN
               DO i = 1, SIZE(wf_history%past_states)
                  CALL wfs_release(wf_history%past_states(i)%snapshot)
               END DO
               DEALLOCATE (wf_history%past_states)
            END IF
            DEALLOCATE (wf_history)
         END IF
      END IF
      NULLIFY (wf_history)
   END SUBROUTINE wfi_release

! **************************************************************************************************
!> \brief returns a snapshot, the first being the latest snapshot
!> \param wf_history the plage where to get the snapshot
!> \param index the index of the snapshot you want
!> \retval res ...
!> \par History
!>      12.2002 created [fawzi]
!> \author fawzi
! **************************************************************************************************
   FUNCTION wfi_get_snapshot(wf_history, index) RESULT(res)
      TYPE(qs_wf_history_type), POINTER                  :: wf_history
      INTEGER, INTENT(in)                                :: index
      TYPE(qs_wf_snapshot_type), POINTER                 :: res

      CHARACTER(len=*), PARAMETER :: routineN = 'wfi_get_snapshot', &
         routineP = moduleN//':'//routineN

      NULLIFY (res)

      IF(.NOT.(ASSOCIATED(wf_history)))CALL cp__a("qs_wf_history_types.F",230)
      IF(.NOT.(ASSOCIATED(wf_history%past_states)))CALL cp__a("qs_wf_history_types.F",231)
      IF (index > wf_history%memory_depth .OR. index > wf_history%snapshot_count) THEN
         CALL cp__b("qs_wf_history_types.F",233,"")
      END IF
      res => wf_history%past_states( &
             MODULO(wf_history%snapshot_count+1-index, &
                    wf_history%memory_depth)+1)%snapshot
   END FUNCTION wfi_get_snapshot

END MODULE qs_wf_history_types